// Copyright (c) 2019-2023 Intel Corporation
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in all
// copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
// SOFTWARE.

#include <set>

#include "av1_scd.h"

#if defined(MFX_ENABLE_ADAPTIVE_ENCODE)

namespace aenc {
    mfxU16 AGOPSelect1(mfxU32 Rs, mfxU32 CsDiff, mfxU32 TSC, mfxI32 diffAFD, mfxU32 Scindex, mfxU32 RsDiff, mfxI32 diffMVdiffVal, mfxU32 RsCsDiff, mfxU32 TSCindex, mfxI16 tcor, mfxU32 refDCval, mfxU32 MVDiff, mfxU32 AFD, mfxI32 diffTSC, mfxI32 diffRsCsdiff, mfxU32 posBalance, mfxU32 SC, mfxU32 mu_mv_mag_sq, mfxU32 negBalance, mfxI16 mcTcor, mfxU32 gchDC, mfxU32 Cs, mfxU32 ssDCval) {
        if (MVDiff < 6) {
            if (posBalance < 53) {
                if (negBalance < 3) {
                    if (Rs < 4736) {
                        if (ssDCval < 103) {
                            if (Rs < 1064) {
                                if (SC < 2273) {
                                    return 8;
                                }
                                else if (SC >= 2273) {
                                    if (mu_mv_mag_sq < 116) {
                                        if (RsCsDiff < 1) {
                                            return 4;
                                        }
                                        else if (RsCsDiff >= 1) {
                                            return 2;
                                        }
                                    }
                                    else if (mu_mv_mag_sq >= 116) {
                                        return 2;
                                    }
                                }
                            }
                            else if (Rs >= 1064) {
                                if (SC < 3689) {
                                    return 8;
                                }
                                else if (SC >= 3689) {
                                    if (Cs < 3985) {
                                        if (AFD < 1) {
                                            if (Cs < 1891) {
                                                return 1;
                                            }
                                            else if (Cs >= 1891) {
                                                if (Cs < 3881) {
                                                    return 8;
                                                }
                                                else if (Cs >= 3881) {
                                                    if (Cs < 3944) {
                                                        return 1;
                                                    }
                                                    else if (Cs >= 3944) {
                                                        if (Rs < 3755) {
                                                            return 8;
                                                        }
                                                        else if (Rs >= 3755) {
                                                            if (Rs < 3781) {
                                                                return 1;
                                                            }
                                                            else if (Rs >= 3781) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (AFD >= 1) {
                                            return 8;
                                        }
                                    }
                                    else if (Cs >= 3985) {
                                        return 8;
                                    }
                                }
                            }
                        }
                        else if (ssDCval >= 103) {
                            if (Rs < 3779) {
                                if (Cs < 4001) {
                                    if (mcTcor < 99) {
                                        if (mu_mv_mag_sq < 124) {
                                            if (diffTSC < 1) {
                                                return 8;
                                            }
                                            else if (diffTSC >= 1) {
                                                return 1;
                                            }
                                        }
                                        else if (mu_mv_mag_sq >= 124) {
                                            if (SC < 3004) {
                                                return 1;
                                            }
                                            else if (SC >= 3004) {
                                                if (posBalance < 22) {
                                                    if (diffRsCsdiff < -31) {
                                                        return 1;
                                                    }
                                                    else if (diffRsCsdiff >= -31) {
                                                        return 8;
                                                    }
                                                }
                                                else if (posBalance >= 22) {
                                                    return 1;
                                                }
                                            }
                                        }
                                    }
                                    else if (mcTcor >= 99) {
                                        if (Cs < 3849) {
                                            if (SC < 7308) {
                                                return 8;
                                            }
                                            else if (SC >= 7308) {
                                                return 1;
                                            }
                                        }
                                        else if (Cs >= 3849) {
                                            if (Rs < 3774) {
                                                return 8;
                                            }
                                            else if (Rs >= 3774) {
                                                return 1;
                                            }
                                        }
                                    }
                                }
                                else if (Cs >= 4001) {
                                    return 1;
                                }
                            }
                            else if (Rs >= 3779) {
                                return 8;
                            }
                        }
                    }
                    else if (Rs >= 4736) {
                        if (Cs < 5084) {
                            return 1;
                        }
                        else if (Cs >= 5084) {
                            return 8;
                        }
                    }
                }
                else if (negBalance >= 3) {
                    if (Rs < 712) {
                        if (refDCval < 161) {
                            if (AFD < 5) {
                                if (refDCval < 98) {
                                    if (TSCindex < 2) {
                                        if (tcor < 100) {
                                            if (Cs < 310) {
                                                return 8;
                                            }
                                            else if (Cs >= 310) {
                                                if (ssDCval < 40) {
                                                    return 8;
                                                }
                                                else if (ssDCval >= 40) {
                                                    if (TSC < 1) {
                                                        if (Rs < 184) {
                                                            return 8;
                                                        }
                                                        else if (Rs >= 184) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (TSC >= 1) {
                                                        if (Cs < 537) {
                                                            return 8;
                                                        }
                                                        else if (Cs >= 537) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (tcor >= 100) {
                                            return 8;
                                        }
                                    }
                                    else if (TSCindex >= 2) {
                                        if (SC < 436) {
                                            if (Rs < 77) {
                                                return 8;
                                            }
                                            else if (Rs >= 77) {
                                                return 1;
                                            }
                                        }
                                        else if (SC >= 436) {
                                            if (Rs < 592) {
                                                if (SC < 1278) {
                                                    if (refDCval < 71) {
                                                        if (AFD < 2) {
                                                            return 8;
                                                        }
                                                        else if (AFD >= 2) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (refDCval >= 71) {
                                                        return 4;
                                                    }
                                                }
                                                else if (SC >= 1278) {
                                                    if (Rs < 392) {
                                                        return 4;
                                                    }
                                                    else if (Rs >= 392) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                            else if (Rs >= 592) {
                                                if (Rs < 687) {
                                                    return 8;
                                                }
                                                else if (Rs >= 687) {
                                                    return 4;
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (refDCval >= 98) {
                                    if (mu_mv_mag_sq < 89) {
                                        if (Cs < 1405) {
                                            if (mu_mv_mag_sq < 33) {
                                                return 2;
                                            }
                                            else if (mu_mv_mag_sq >= 33) {
                                                if (negBalance < 46) {
                                                    if (Cs < 446) {
                                                        return 8;
                                                    }
                                                    else if (Cs >= 446) {
                                                        return 2;
                                                    }
                                                }
                                                else if (negBalance >= 46) {
                                                    return 8;
                                                }
                                            }
                                        }
                                        else if (Cs >= 1405) {
                                            return 4;
                                        }
                                    }
                                    else if (mu_mv_mag_sq >= 89) {
                                        if (TSCindex < 5) {
                                            return 8;
                                        }
                                        else if (TSCindex >= 5) {
                                            if (Cs < 522) {
                                                return 4;
                                            }
                                            else if (Cs >= 522) {
                                                return 8;
                                            }
                                        }
                                    }
                                }
                            }
                            else if (AFD >= 5) {
                                if (SC < 1274) {
                                    if (refDCval < 87) {
                                        return 4;
                                    }
                                    else if (refDCval >= 87) {
                                        if (posBalance < 14) {
                                            return 8;
                                        }
                                        else if (posBalance >= 14) {
                                            if (SC < 1134) {
                                                return 2;
                                            }
                                            else if (SC >= 1134) {
                                                return 4;
                                            }
                                        }
                                    }
                                }
                                else if (SC >= 1274) {
                                    return 4;
                                }
                            }
                        }
                        else if (refDCval >= 161) {
                            if (posBalance < 26) {
                                return 4;
                            }
                            else if (posBalance >= 26) {
                                if (Rs < 440) {
                                    return 4;
                                }
                                else if (Rs >= 440) {
                                    if (SC < 995) {
                                        return 2;
                                    }
                                    else if (SC >= 995) {
                                        return 4;
                                    }
                                }
                            }
                        }
                    }
                    else if (Rs >= 712) {
                        if (mu_mv_mag_sq < 99) {
                            if (ssDCval < 54) {
                                if (Rs < 1150) {
                                    if (Cs < 962) {
                                        if (AFD < 6) {
                                            return 4;
                                        }
                                        else if (AFD >= 6) {
                                            if (refDCval < 49) {
                                                return 1;
                                            }
                                            else if (refDCval >= 49) {
                                                return 4;
                                            }
                                        }
                                    }
                                    else if (Cs >= 962) {
                                        if (AFD < 1) {
                                            if (Cs < 1265) {
                                                return 4;
                                            }
                                            else if (Cs >= 1265) {
                                                return 2;
                                            }
                                        }
                                        else if (AFD >= 1) {
                                            return 8;
                                        }
                                    }
                                }
                                else if (Rs >= 1150) {
                                    if (ssDCval < 46) {
                                        return 4;
                                    }
                                    else if (ssDCval >= 46) {
                                        if (ssDCval < 52) {
                                            return 1;
                                        }
                                        else if (ssDCval >= 52) {
                                            if (tcor < 93) {
                                                if (mu_mv_mag_sq < 56) {
                                                    return 1;
                                                }
                                                else if (mu_mv_mag_sq >= 56) {
                                                    return 4;
                                                }
                                            }
                                            else if (tcor >= 93) {
                                                if (refDCval < 52) {
                                                    return 4;
                                                }
                                                else if (refDCval >= 52) {
                                                    return 1;
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                            else if (ssDCval >= 54) {
                                if (Cs < 829) {
                                    if (SC < 848) {
                                        if (Cs < 16) {
                                            return 1;
                                        }
                                        else if (Cs >= 16) {
                                            return 8;
                                        }
                                    }
                                    else if (SC >= 848) {
                                        if (Rs < 2164) {
                                            if (Rs < 745) {
                                                if (MVDiff < 2) {
                                                    return 2;
                                                }
                                                else if (MVDiff >= 2) {
                                                    return 8;
                                                }
                                            }
                                            else if (Rs >= 745) {
                                                return 8;
                                            }
                                        }
                                        else if (Rs >= 2164) {
                                            if (MVDiff < 2) {
                                                return 8;
                                            }
                                            else if (MVDiff >= 2) {
                                                return 2;
                                            }
                                        }
                                    }
                                }
                                else if (Cs >= 829) {
                                    if (negBalance < 24) {
                                        if (Rs < 1444) {
                                            return 8;
                                        }
                                        else if (Rs >= 1444) {
                                            return 1;
                                        }
                                    }
                                    else if (negBalance >= 24) {
                                        if (SC < 5202) {
                                            if (posBalance < 40) {
                                                if (CsDiff < 4) {
                                                    if (refDCval < 90) {
                                                        return 4;
                                                    }
                                                    else if (refDCval >= 90) {
                                                        if (Rs < 2747) {
                                                            return 8;
                                                        }
                                                        else if (Rs >= 2747) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                                else if (CsDiff >= 4) {
                                                    return 4;
                                                }
                                            }
                                            else if (posBalance >= 40) {
                                                if (mu_mv_mag_sq < 76) {
                                                    if (posBalance < 47) {
                                                        if (ssDCval < 93) {
                                                            return 4;
                                                        }
                                                        else if (ssDCval >= 93) {
                                                            return 2;
                                                        }
                                                    }
                                                    else if (posBalance >= 47) {
                                                        if (gchDC < 1) {
                                                            if (refDCval < 136) {
                                                                if (negBalance < 42) {
                                                                    return 1;
                                                                }
                                                                else if (negBalance >= 42) {
                                                                    if (Rs < 2335) {
                                                                        return 4;
                                                                    }
                                                                    else if (Rs >= 2335) {
                                                                        return 1;
                                                                    }
                                                                }
                                                            }
                                                            else if (refDCval >= 136) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (gchDC >= 1) {
                                                            if (ssDCval < 55) {
                                                                return 1;
                                                            }
                                                            else if (ssDCval >= 55) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (mu_mv_mag_sq >= 76) {
                                                    if (Rs < 2639) {
                                                        if (negBalance < 42) {
                                                            if (diffRsCsdiff < -32) {
                                                                return 4;
                                                            }
                                                            else if (diffRsCsdiff >= -32) {
                                                                if (mu_mv_mag_sq < 82) {
                                                                    if (Rs < 1570) {
                                                                        return 8;
                                                                    }
                                                                    else if (Rs >= 1570) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (mu_mv_mag_sq >= 82) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                        else if (negBalance >= 42) {
                                                            if (SC < 4171) {
                                                                return 4;
                                                            }
                                                            else if (SC >= 4171) {
                                                                return 2;
                                                            }
                                                        }
                                                    }
                                                    else if (Rs >= 2639) {
                                                        if (Cs < 1638) {
                                                            return 1;
                                                        }
                                                        else if (Cs >= 1638) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (SC >= 5202) {
                                            if (refDCval < 120) {
                                                if (ssDCval < 111) {
                                                    if (mcTcor < 78) {
                                                        if (Cs < 6240) {
                                                            return 8;
                                                        }
                                                        else if (Cs >= 6240) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (mcTcor >= 78) {
                                                        return 8;
                                                    }
                                                }
                                                else if (ssDCval >= 111) {
                                                    if (SC < 8103) {
                                                        return 8;
                                                    }
                                                    else if (SC >= 8103) {
                                                        if (Cs < 3122) {
                                                            return 2;
                                                        }
                                                        else if (Cs >= 3122) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (refDCval >= 120) {
                                                if (Rs < 4527) {
                                                    if (RsDiff < 5) {
                                                        return 8;
                                                    }
                                                    else if (RsDiff >= 5) {
                                                        if (tcor < 75) {
                                                            if (Rs < 3928) {
                                                                return 1;
                                                            }
                                                            else if (Rs >= 3928) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (tcor >= 75) {
                                                            if (Cs < 3253) {
                                                                return 4;
                                                            }
                                                            else if (Cs >= 3253) {
                                                                if (SC < 7986) {
                                                                    return 8;
                                                                }
                                                                else if (SC >= 7986) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (Rs >= 4527) {
                                                    if (Rs < 5048) {
                                                        return 2;
                                                    }
                                                    else if (Rs >= 5048) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                        }
                        else if (mu_mv_mag_sq >= 99) {
                            if (mu_mv_mag_sq < 125) {
                                if (ssDCval < 67) {
                                    if (TSCindex < 1) {
                                        if (Cs < 794) {
                                            return 4;
                                        }
                                        else if (Cs >= 794) {
                                            if (Cs < 1191) {
                                                return 8;
                                            }
                                            else if (Cs >= 1191) {
                                                if (refDCval < 32) {
                                                    return 4;
                                                }
                                                else if (refDCval >= 32) {
                                                    return 8;
                                                }
                                            }
                                        }
                                    }
                                    else if (TSCindex >= 1) {
                                        if (CsDiff < 4) {
                                            if (diffTSC < -1) {
                                                if (diffMVdiffVal < -3) {
                                                    if (diffAFD < -2) {
                                                        return 1;
                                                    }
                                                    else if (diffAFD >= -2) {
                                                        return 8;
                                                    }
                                                }
                                                else if (diffMVdiffVal >= -3) {
                                                    return 8;
                                                }
                                            }
                                            else if (diffTSC >= -1) {
                                                return 8;
                                            }
                                        }
                                        else if (CsDiff >= 4) {
                                            if (posBalance < 35) {
                                                if (negBalance < 41) {
                                                    return 8;
                                                }
                                                else if (negBalance >= 41) {
                                                    if (Rs < 2366) {
                                                        return 4;
                                                    }
                                                    else if (Rs >= 2366) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                            else if (posBalance >= 35) {
                                                if (TSCindex < 5) {
                                                    if (Cs < 1562) {
                                                        return 8;
                                                    }
                                                    else if (Cs >= 1562) {
                                                        return 4;
                                                    }
                                                }
                                                else if (TSCindex >= 5) {
                                                    return 8;
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (ssDCval >= 67) {
                                    if (negBalance < 63) {
                                        if (Cs < 3085) {
                                            if (Cs < 2030) {
                                                if (mcTcor < 66) {
                                                    return 1;
                                                }
                                                else if (mcTcor >= 66) {
                                                    if (CsDiff < 3) {
                                                        if (Rs < 900) {
                                                            if (SC < 1589) {
                                                                return 8;
                                                            }
                                                            else if (SC >= 1589) {
                                                                return 1;
                                                            }
                                                        }
                                                        else if (Rs >= 900) {
                                                            if (ssDCval < 74) {
                                                                if (Rs < 1801) {
                                                                    return 8;
                                                                }
                                                                else if (Rs >= 1801) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (ssDCval >= 74) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                    else if (CsDiff >= 3) {
                                                        if (negBalance < 33) {
                                                            if (mu_mv_mag_sq < 116) {
                                                                return 8;
                                                            }
                                                            else if (mu_mv_mag_sq >= 116) {
                                                                if (Rs < 1328) {
                                                                    return 4;
                                                                }
                                                                else if (Rs >= 1328) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                        else if (negBalance >= 33) {
                                                            if (Rs < 1374) {
                                                                return 4;
                                                            }
                                                            else if (Rs >= 1374) {
                                                                if (Cs < 1592) {
                                                                    return 8;
                                                                }
                                                                else if (Cs >= 1592) {
                                                                    if (mu_mv_mag_sq < 107) {
                                                                        return 8;
                                                                    }
                                                                    else if (mu_mv_mag_sq >= 107) {
                                                                        return 4;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                            else if (Cs >= 2030) {
                                                if (SC < 4666) {
                                                    if (RsCsDiff < 38) {
                                                        if (ssDCval < 81) {
                                                            return 1;
                                                        }
                                                        else if (ssDCval >= 81) {
                                                            if (negBalance < 33) {
                                                                if (Rs < 2331) {
                                                                    return 8;
                                                                }
                                                                else if (Rs >= 2331) {
                                                                    if (refDCval < 116) {
                                                                        return 8;
                                                                    }
                                                                    else if (refDCval >= 116) {
                                                                        return 4;
                                                                    }
                                                                }
                                                            }
                                                            else if (negBalance >= 33) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                    else if (RsCsDiff >= 38) {
                                                        if (diffAFD < -3) {
                                                            return 8;
                                                        }
                                                        else if (diffAFD >= -3) {
                                                            if (posBalance < 46) {
                                                                if (SC < 4477) {
                                                                    return 1;
                                                                }
                                                                else if (SC >= 4477) {
                                                                    if (SC < 4501) {
                                                                        return 8;
                                                                    }
                                                                    else if (SC >= 4501) {
                                                                        return 1;
                                                                    }
                                                                }
                                                            }
                                                            else if (posBalance >= 46) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (SC >= 4666) {
                                                    if (tcor < 98) {
                                                        if (ssDCval < 116) {
                                                            if (Rs < 4118) {
                                                                if (ssDCval < 87) {
                                                                    return 4;
                                                                }
                                                                else if (ssDCval >= 87) {
                                                                    if (diffRsCsdiff < 96) {
                                                                        return 8;
                                                                    }
                                                                    else if (diffRsCsdiff >= 96) {
                                                                        if (Rs < 3074) {
                                                                            return 1;
                                                                        }
                                                                        else if (Rs >= 3074) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                            else if (Rs >= 4118) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (ssDCval >= 116) {
                                                            if (refDCval < 120) {
                                                                if (Rs < 2431) {
                                                                    return 8;
                                                                }
                                                                else if (Rs >= 2431) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (refDCval >= 120) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                    else if (tcor >= 98) {
                                                        if (Rs < 3158) {
                                                            if (RsCsDiff < 29) {
                                                                return 8;
                                                            }
                                                            else if (RsCsDiff >= 29) {
                                                                if (ssDCval < 108) {
                                                                    return 8;
                                                                }
                                                                else if (ssDCval >= 108) {
                                                                    if (refDCval < 109) {
                                                                        if (Rs < 3097) {
                                                                            return 1;
                                                                        }
                                                                        else if (Rs >= 3097) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                    else if (refDCval >= 109) {
                                                                        return 1;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (Rs >= 3158) {
                                                            if (refDCval < 108) {
                                                                return 8;
                                                            }
                                                            else if (refDCval >= 108) {
                                                                if (Rs < 3390) {
                                                                    if (diffRsCsdiff < -22) {
                                                                        return 8;
                                                                    }
                                                                    else if (diffRsCsdiff >= -22) {
                                                                        return 1;
                                                                    }
                                                                }
                                                                else if (Rs >= 3390) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (Cs >= 3085) {
                                            return 8;
                                        }
                                    }
                                    else if (negBalance >= 63) {
                                        if (ssDCval < 91) {
                                            if (SC < 15470) {
                                                return 4;
                                            }
                                            else if (SC >= 15470) {
                                                return 2;
                                            }
                                        }
                                        else if (ssDCval >= 91) {
                                            return 4;
                                        }
                                    }
                                }
                            }
                            else if (mu_mv_mag_sq >= 125) {
                                if (Cs < 7570) {
                                    if (negBalance < 36) {
                                        if (RsDiff < 1) {
                                            if (negBalance < 12) {
                                                return 8;
                                            }
                                            else if (negBalance >= 12) {
                                                if (posBalance < 34) {
                                                    if (negBalance < 13) {
                                                        return 4;
                                                    }
                                                    else if (negBalance >= 13) {
                                                        return 8;
                                                    }
                                                }
                                                else if (posBalance >= 34) {
                                                    return 8;
                                                }
                                            }
                                        }
                                        else if (RsDiff >= 1) {
                                            return 8;
                                        }
                                    }
                                    else if (negBalance >= 36) {
                                        if (AFD < 7) {
                                            if (CsDiff < 8) {
                                                if (Cs < 2180) {
                                                    return 8;
                                                }
                                                else if (Cs >= 2180) {
                                                    if (RsDiff < 4) {
                                                        return 4;
                                                    }
                                                    else if (RsDiff >= 4) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                            else if (CsDiff >= 8) {
                                                return 4;
                                            }
                                        }
                                        else if (AFD >= 7) {
                                            return 8;
                                        }
                                    }
                                }
                                else if (Cs >= 7570) {
                                    return 2;
                                }
                            }
                        }
                    }
                }
            }
            else if (posBalance >= 53) {
                if (mu_mv_mag_sq < 109) {
                    if (Cs < 2472) {
                        if (negBalance < 44) {
                            if (MVDiff < 1) {
                                if (RsDiff < 1) {
                                    return 8;
                                }
                                else if (RsDiff >= 1) {
                                    if (TSC < 7) {
                                        return 4;
                                    }
                                    else if (TSC >= 7) {
                                        if (SC < 3546) {
                                            return 2;
                                        }
                                        else if (SC >= 3546) {
                                            return 4;
                                        }
                                    }
                                }
                            }
                            else if (MVDiff >= 1) {
                                if (TSC < 7) {
                                    if (refDCval < 104) {
                                        if (ssDCval < 73) {
                                            if (ssDCval < 54) {
                                                if (mcTcor < 95) {
                                                    if (Rs < 1159) {
                                                        return 4;
                                                    }
                                                    else if (Rs >= 1159) {
                                                        if (gchDC < 2) {
                                                            if (mu_mv_mag_sq < 3) {
                                                                if (Cs < 929) {
                                                                    return 1;
                                                                }
                                                                else if (Cs >= 929) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (mu_mv_mag_sq >= 3) {
                                                                return 1;
                                                            }
                                                        }
                                                        else if (gchDC >= 2) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                                else if (mcTcor >= 95) {
                                                    if (tcor < 93) {
                                                        if (SC < 2035) {
                                                            if (MVDiff < 5) {
                                                                return 4;
                                                            }
                                                            else if (MVDiff >= 5) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (SC >= 2035) {
                                                            return 1;
                                                        }
                                                    }
                                                    else if (tcor >= 93) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                            else if (ssDCval >= 54) {
                                                if (SC < 3757) {
                                                    if (diffMVdiffVal < -4) {
                                                        if (Rs < 1098) {
                                                            return 8;
                                                        }
                                                        else if (Rs >= 1098) {
                                                            return 1;
                                                        }
                                                    }
                                                    else if (diffMVdiffVal >= -4) {
                                                        return 8;
                                                    }
                                                }
                                                else if (SC >= 3757) {
                                                    return 4;
                                                }
                                            }
                                        }
                                        else if (ssDCval >= 73) {
                                            if (refDCval < 83) {
                                                return 4;
                                            }
                                            else if (refDCval >= 83) {
                                                if (TSC < 4) {
                                                    if (posBalance < 57) {
                                                        return 2;
                                                    }
                                                    else if (posBalance >= 57) {
                                                        return 4;
                                                    }
                                                }
                                                else if (TSC >= 4) {
                                                    if (posBalance < 54) {
                                                        return 1;
                                                    }
                                                    else if (posBalance >= 54) {
                                                        if (RsCsDiff < 38) {
                                                            return 8;
                                                        }
                                                        else if (RsCsDiff >= 38) {
                                                            if (Rs < 1901) {
                                                                return 8;
                                                            }
                                                            else if (Rs >= 1901) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (refDCval >= 104) {
                                        if (mu_mv_mag_sq < 44) {
                                            if (Rs < 1157) {
                                                if (Rs < 837) {
                                                    return 1;
                                                }
                                                else if (Rs >= 837) {
                                                    if (CsDiff < 2) {
                                                        return 8;
                                                    }
                                                    else if (CsDiff >= 2) {
                                                        return 1;
                                                    }
                                                }
                                            }
                                            else if (Rs >= 1157) {
                                                return 4;
                                            }
                                        }
                                        else if (mu_mv_mag_sq >= 44) {
                                            return 8;
                                        }
                                    }
                                }
                                else if (TSC >= 7) {
                                    if (ssDCval < 104) {
                                        if (AFD < 11) {
                                            if (refDCval < 85) {
                                                return 8;
                                            }
                                            else if (refDCval >= 85) {
                                                return 2;
                                            }
                                        }
                                        else if (AFD >= 11) {
                                            return 1;
                                        }
                                    }
                                    else if (ssDCval >= 104) {
                                        if (Rs < 1243) {
                                            if (Cs < 675) {
                                                return 2;
                                            }
                                            else if (Cs >= 675) {
                                                return 4;
                                            }
                                        }
                                        else if (Rs >= 1243) {
                                            return 8;
                                        }
                                    }
                                }
                            }
                        }
                        else if (negBalance >= 44) {
                            if (Rs < 1972) {
                                if (mcTcor < 88) {
                                    if (mu_mv_mag_sq < 81) {
                                        if (TSCindex < 8) {
                                            return 2;
                                        }
                                        else if (TSCindex >= 8) {
                                            if (MVDiff < 3) {
                                                return 2;
                                            }
                                            else if (MVDiff >= 3) {
                                                return 4;
                                            }
                                        }
                                    }
                                    else if (mu_mv_mag_sq >= 81) {
                                        if (Cs < 1298) {
                                            if (Rs < 949) {
                                                return 8;
                                            }
                                            else if (Rs >= 949) {
                                                return 4;
                                            }
                                        }
                                        else if (Cs >= 1298) {
                                            return 8;
                                        }
                                    }
                                }
                                else if (mcTcor >= 88) {
                                    if (AFD < 10) {
                                        if (mu_mv_mag_sq < 83) {
                                            if (Cs < 1222) {
                                                if (Rs < 1390) {
                                                    if (refDCval < 93) {
                                                        return 2;
                                                    }
                                                    else if (refDCval >= 93) {
                                                        if (Rs < 429) {
                                                            return 2;
                                                        }
                                                        else if (Rs >= 429) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                                else if (Rs >= 1390) {
                                                    if (SC < 2786) {
                                                        return 8;
                                                    }
                                                    else if (SC >= 2786) {
                                                        return 1;
                                                    }
                                                }
                                            }
                                            else if (Cs >= 1222) {
                                                if (ssDCval < 78) {
                                                    return 4;
                                                }
                                                else if (ssDCval >= 78) {
                                                    if (ssDCval < 88) {
                                                        return 2;
                                                    }
                                                    else if (ssDCval >= 88) {
                                                        if (Rs < 1515) {
                                                            return 8;
                                                        }
                                                        else if (Rs >= 1515) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (mu_mv_mag_sq >= 83) {
                                            if (MVDiff < 4) {
                                                return 2;
                                            }
                                            else if (MVDiff >= 4) {
                                                if (gchDC < 1) {
                                                    if (ssDCval < 97) {
                                                        return 8;
                                                    }
                                                    else if (ssDCval >= 97) {
                                                        return 2;
                                                    }
                                                }
                                                else if (gchDC >= 1) {
                                                    return 4;
                                                }
                                            }
                                        }
                                    }
                                    else if (AFD >= 10) {
                                        if (ssDCval < 89) {
                                            if (diffMVdiffVal < -5) {
                                                if (Rs < 1334) {
                                                    return 4;
                                                }
                                                else if (Rs >= 1334) {
                                                    return 1;
                                                }
                                            }
                                            else if (diffMVdiffVal >= -5) {
                                                return 2;
                                            }
                                        }
                                        else if (ssDCval >= 89) {
                                            if (Rs < 1713) {
                                                return 4;
                                            }
                                            else if (Rs >= 1713) {
                                                if (Rs < 1803) {
                                                    return 2;
                                                }
                                                else if (Rs >= 1803) {
                                                    return 1;
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                            else if (Rs >= 1972) {
                                if (ssDCval < 99) {
                                    if (ssDCval < 88) {
                                        if (RsDiff < 7) {
                                            if (SC < 4111) {
                                                if (Cs < 1928) {
                                                    return 2;
                                                }
                                                else if (Cs >= 1928) {
                                                    return 8;
                                                }
                                            }
                                            else if (SC >= 4111) {
                                                return 4;
                                            }
                                        }
                                        else if (RsDiff >= 7) {
                                            if (posBalance < 56) {
                                                if (tcor < 89) {
                                                    return 4;
                                                }
                                                else if (tcor >= 89) {
                                                    if (SC < 4203) {
                                                        return 1;
                                                    }
                                                    else if (SC >= 4203) {
                                                        return 2;
                                                    }
                                                }
                                            }
                                            else if (posBalance >= 56) {
                                                if (Cs < 2368) {
                                                    return 8;
                                                }
                                                else if (Cs >= 2368) {
                                                    if (gchDC < 1) {
                                                        if (RsCsDiff < 124) {
                                                            return 4;
                                                        }
                                                        else if (RsCsDiff >= 124) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (gchDC >= 1) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (ssDCval >= 88) {
                                        if (ssDCval < 91) {
                                            if (mcTcor < 97) {
                                                if (tcor < 89) {
                                                    return 8;
                                                }
                                                else if (tcor >= 89) {
                                                    if (posBalance < 59) {
                                                        if (SC < 4229) {
                                                            return 2;
                                                        }
                                                        else if (SC >= 4229) {
                                                            return 1;
                                                        }
                                                    }
                                                    else if (posBalance >= 59) {
                                                        return 2;
                                                    }
                                                }
                                            }
                                            else if (mcTcor >= 97) {
                                                if (SC < 4271) {
                                                    if (Rs < 2345) {
                                                        return 4;
                                                    }
                                                    else if (Rs >= 2345) {
                                                        return 8;
                                                    }
                                                }
                                                else if (SC >= 4271) {
                                                    return 1;
                                                }
                                            }
                                        }
                                        else if (ssDCval >= 91) {
                                            if (posBalance < 55) {
                                                return 4;
                                            }
                                            else if (posBalance >= 55) {
                                                if (negBalance < 52) {
                                                    if (refDCval < 94) {
                                                        if (TSCindex < 7) {
                                                            return 4;
                                                        }
                                                        else if (TSCindex >= 7) {
                                                            if (Rs < 2508) {
                                                                return 2;
                                                            }
                                                            else if (Rs >= 2508) {
                                                                return 1;
                                                            }
                                                        }
                                                    }
                                                    else if (refDCval >= 94) {
                                                        if (SC < 4018) {
                                                            return 8;
                                                        }
                                                        else if (SC >= 4018) {
                                                            if (Cs < 2322) {
                                                                return 4;
                                                            }
                                                            else if (Cs >= 2322) {
                                                                return 2;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (negBalance >= 52) {
                                                    if (ssDCval < 93) {
                                                        if (Cs < 2178) {
                                                            return 8;
                                                        }
                                                        else if (Cs >= 2178) {
                                                            return 2;
                                                        }
                                                    }
                                                    else if (ssDCval >= 93) {
                                                        if (Rs < 2011) {
                                                            return 8;
                                                        }
                                                        else if (Rs >= 2011) {
                                                            if (SC < 5344) {
                                                                return 2;
                                                            }
                                                            else if (SC >= 5344) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (ssDCval >= 99) {
                                    if (mu_mv_mag_sq < 57) {
                                        if (AFD < 12) {
                                            return 4;
                                        }
                                        else if (AFD >= 12) {
                                            if (tcor < 76) {
                                                return 1;
                                            }
                                            else if (tcor >= 76) {
                                                if (ssDCval < 107) {
                                                    if (Rs < 2855) {
                                                        return 8;
                                                    }
                                                    else if (Rs >= 2855) {
                                                        return 1;
                                                    }
                                                }
                                                else if (ssDCval >= 107) {
                                                    if (Rs < 2677) {
                                                        return 4;
                                                    }
                                                    else if (Rs >= 2677) {
                                                        return 2;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (mu_mv_mag_sq >= 57) {
                                        if (posBalance < 56) {
                                            if (tcor < 89) {
                                                return 1;
                                            }
                                            else if (tcor >= 89) {
                                                if (Cs < 2162) {
                                                    return 4;
                                                }
                                                else if (Cs >= 2162) {
                                                    return 8;
                                                }
                                            }
                                        }
                                        else if (posBalance >= 56) {
                                            if (mu_mv_mag_sq < 89) {
                                                if (Rs < 2836) {
                                                    if (ssDCval < 111) {
                                                        return 4;
                                                    }
                                                    else if (ssDCval >= 111) {
                                                        if (mu_mv_mag_sq < 79) {
                                                            if (Cs < 1713) {
                                                                return 4;
                                                            }
                                                            else if (Cs >= 1713) {
                                                                if (SC < 4739) {
                                                                    if (Cs < 2401) {
                                                                        if (MVDiff < 5) {
                                                                            return 8;
                                                                        }
                                                                        else if (MVDiff >= 5) {
                                                                            if (Cs < 1807) {
                                                                                return 8;
                                                                            }
                                                                            else if (Cs >= 1807) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (Cs >= 2401) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (SC >= 4739) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                        else if (mu_mv_mag_sq >= 79) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                                else if (Rs >= 2836) {
                                                    return 4;
                                                }
                                            }
                                            else if (mu_mv_mag_sq >= 89) {
                                                if (refDCval < 109) {
                                                    return 4;
                                                }
                                                else if (refDCval >= 109) {
                                                    return 8;
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                        }
                    }
                    else if (Cs >= 2472) {
                        if (Cs < 4129) {
                            if (ssDCval < 127) {
                                if (refDCval < 106) {
                                    if (TSC < 13) {
                                        if (negBalance < 48) {
                                            if (CsDiff < 4) {
                                                if (mcTcor < 97) {
                                                    if (mu_mv_mag_sq < 96) {
                                                        if (ssDCval < 55) {
                                                            if (Rs < 3485) {
                                                                return 8;
                                                            }
                                                            else if (Rs >= 3485) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (ssDCval >= 55) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (mu_mv_mag_sq >= 96) {
                                                        return 8;
                                                    }
                                                }
                                                else if (mcTcor >= 97) {
                                                    return 8;
                                                }
                                            }
                                            else if (CsDiff >= 4) {
                                                if (SC < 5232) {
                                                    if (AFD < 9) {
                                                        return 8;
                                                    }
                                                    else if (AFD >= 9) {
                                                        return 2;
                                                    }
                                                }
                                                else if (SC >= 5232) {
                                                    if (AFD < 17) {
                                                        if (SC < 6346) {
                                                            return 8;
                                                        }
                                                        else if (SC >= 6346) {
                                                            if (negBalance < 39) {
                                                                if (SC < 7327) {
                                                                    return 2;
                                                                }
                                                                else if (SC >= 7327) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (negBalance >= 39) {
                                                                if (Cs < 2992) {
                                                                    return 4;
                                                                }
                                                                else if (Cs >= 2992) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (AFD >= 17) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                        }
                                        else if (negBalance >= 48) {
                                            if (ssDCval < 96) {
                                                if (Rs < 3330) {
                                                    if (ssDCval < 88) {
                                                        return 4;
                                                    }
                                                    else if (ssDCval >= 88) {
                                                        if (Rs < 2782) {
                                                            if (SC < 5130) {
                                                                return 2;
                                                            }
                                                            else if (SC >= 5130) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (Rs >= 2782) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                                else if (Rs >= 3330) {
                                                    if (AFD < 8) {
                                                        return 8;
                                                    }
                                                    else if (AFD >= 8) {
                                                        return 2;
                                                    }
                                                }
                                            }
                                            else if (ssDCval >= 96) {
                                                if (SC < 5313) {
                                                    if (Rs < 2698) {
                                                        return 4;
                                                    }
                                                    else if (Rs >= 2698) {
                                                        return 8;
                                                    }
                                                }
                                                else if (SC >= 5313) {
                                                    return 4;
                                                }
                                            }
                                        }
                                    }
                                    else if (TSC >= 13) {
                                        return 8;
                                    }
                                }
                                else if (refDCval >= 106) {
                                    if (RsDiff < 11) {
                                        if (posBalance < 55) {
                                            if (TSC < 11) {
                                                return 8;
                                            }
                                            else if (TSC >= 11) {
                                                if (Rs < 3289) {
                                                    return 8;
                                                }
                                                else if (Rs >= 3289) {
                                                    return 1;
                                                }
                                            }
                                        }
                                        else if (posBalance >= 55) {
                                            return 8;
                                        }
                                    }
                                    else if (RsDiff >= 11) {
                                        if (AFD < 13) {
                                            if (Cs < 3426) {
                                                return 4;
                                            }
                                            else if (Cs >= 3426) {
                                                return 2;
                                            }
                                        }
                                        else if (AFD >= 13) {
                                            if (RsCsDiff < 326) {
                                                return 8;
                                            }
                                            else if (RsCsDiff >= 326) {
                                                if (negBalance < 49) {
                                                    if (AFD < 17) {
                                                        return 1;
                                                    }
                                                    else if (AFD >= 17) {
                                                        return 4;
                                                    }
                                                }
                                                else if (negBalance >= 49) {
                                                    return 8;
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                            else if (ssDCval >= 127) {
                                if (ssDCval < 141) {
                                    return 1;
                                }
                                else if (ssDCval >= 141) {
                                    if (Rs < 7393) {
                                        return 8;
                                    }
                                    else if (Rs >= 7393) {
                                        return 4;
                                    }
                                }
                            }
                        }
                        else if (Cs >= 4129) {
                            if (Cs < 7278) {
                                if (mu_mv_mag_sq < 3) {
                                    if (TSC < 14) {
                                        if (Rs < 5200) {
                                            return 4;
                                        }
                                        else if (Rs >= 5200) {
                                            return 1;
                                        }
                                    }
                                    else if (TSC >= 14) {
                                        return 8;
                                    }
                                }
                                else if (mu_mv_mag_sq >= 3) {
                                    if (Rs < 4465) {
                                        if (SC < 9111) {
                                            return 8;
                                        }
                                        else if (SC >= 9111) {
                                            return 1;
                                        }
                                    }
                                    else if (Rs >= 4465) {
                                        return 8;
                                    }
                                }
                            }
                            else if (Cs >= 7278) {
                                if (tcor < 76) {
                                    if (Cs < 11438) {
                                        if (TSC < 18) {
                                            return 8;
                                        }
                                        else if (TSC >= 18) {
                                            if (negBalance < 67) {
                                                return 4;
                                            }
                                            else if (negBalance >= 67) {
                                                return 2;
                                            }
                                        }
                                    }
                                    else if (Cs >= 11438) {
                                        if (Cs < 12135) {
                                            return 8;
                                        }
                                        else if (Cs >= 12135) {
                                            return 4;
                                        }
                                    }
                                }
                                else if (tcor >= 76) {
                                    return 8;
                                }
                            }
                        }
                    }
                }
                else if (mu_mv_mag_sq >= 109) {
                    if (mu_mv_mag_sq < 128) {
                        if (Cs < 3618) {
                            if (mcTcor < 96) {
                                if (Rs < 3408) {
                                    if (SC < 2626) {
                                        if (diffAFD < -2) {
                                            return 8;
                                        }
                                        else if (diffAFD >= -2) {
                                            if (negBalance < 47) {
                                                return 8;
                                            }
                                            else if (negBalance >= 47) {
                                                return 2;
                                            }
                                        }
                                    }
                                    else if (SC >= 2626) {
                                        if (TSC < 6) {
                                            if (Rs < 1792) {
                                                return 8;
                                            }
                                            else if (Rs >= 1792) {
                                                if (tcor < 95) {
                                                    if (posBalance < 56) {
                                                        if (Cs < 1933) {
                                                            return 8;
                                                        }
                                                        else if (Cs >= 1933) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (posBalance >= 56) {
                                                        return 8;
                                                    }
                                                }
                                                else if (tcor >= 95) {
                                                    if (mu_mv_mag_sq < 123) {
                                                        if (negBalance < 44) {
                                                            return 4;
                                                        }
                                                        else if (negBalance >= 44) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (mu_mv_mag_sq >= 123) {
                                                        if (ssDCval < 75) {
                                                            return 2;
                                                        }
                                                        else if (ssDCval >= 75) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (TSC >= 6) {
                                            if (ssDCval < 77) {
                                                if (ssDCval < 75) {
                                                    if (MVDiff < 1) {
                                                        if (posBalance < 58) {
                                                            if (TSC < 7) {
                                                                if (refDCval < 72) {
                                                                    return 2;
                                                                }
                                                                else if (refDCval >= 72) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (TSC >= 7) {
                                                                return 2;
                                                            }
                                                        }
                                                        else if (posBalance >= 58) {
                                                            if (AFD < 7) {
                                                                if (CsDiff < 5) {
                                                                    if (Rs < 1968) {
                                                                        return 2;
                                                                    }
                                                                    else if (Rs >= 1968) {
                                                                        return 8;
                                                                    }
                                                                }
                                                                else if (CsDiff >= 5) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (AFD >= 7) {
                                                                return 2;
                                                            }
                                                        }
                                                    }
                                                    else if (MVDiff >= 1) {
                                                        if (RsDiff < 4) {
                                                            if (Cs < 1588) {
                                                                return 2;
                                                            }
                                                            else if (Cs >= 1588) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (RsDiff >= 4) {
                                                            if (mcTcor < 94) {
                                                                return 8;
                                                            }
                                                            else if (mcTcor >= 94) {
                                                                if (refDCval < 73) {
                                                                    return 2;
                                                                }
                                                                else if (refDCval >= 73) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (ssDCval >= 75) {
                                                    if (Rs < 1837) {
                                                        if (Cs < 1504) {
                                                            return 4;
                                                        }
                                                        else if (Cs >= 1504) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (Rs >= 1837) {
                                                        if (TSCindex < 8) {
                                                            return 4;
                                                        }
                                                        else if (TSCindex >= 8) {
                                                            if (posBalance < 60) {
                                                                return 2;
                                                            }
                                                            else if (posBalance >= 60) {
                                                                if (Rs < 1927) {
                                                                    return 4;
                                                                }
                                                                else if (Rs >= 1927) {
                                                                    return 2;
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                            else if (ssDCval >= 77) {
                                                if (Rs < 2248) {
                                                    return 8;
                                                }
                                                else if (Rs >= 2248) {
                                                    if (AFD < 8) {
                                                        if (Rs < 2875) {
                                                            if (posBalance < 58) {
                                                                return 4;
                                                            }
                                                            else if (posBalance >= 58) {
                                                                if (mu_mv_mag_sq < 118) {
                                                                    if (Cs < 2340) {
                                                                        return 4;
                                                                    }
                                                                    else if (Cs >= 2340) {
                                                                        return 8;
                                                                    }
                                                                }
                                                                else if (mu_mv_mag_sq >= 118) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                        else if (Rs >= 2875) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (AFD >= 8) {
                                                        if (mu_mv_mag_sq < 120) {
                                                            return 8;
                                                        }
                                                        else if (mu_mv_mag_sq >= 120) {
                                                            if (Rs < 3183) {
                                                                return 8;
                                                            }
                                                            else if (Rs >= 3183) {
                                                                return 2;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (Rs >= 3408) {
                                    if (Rs < 3898) {
                                        if (Cs < 2887) {
                                            if (Cs < 2806) {
                                                return 8;
                                            }
                                            else if (Cs >= 2806) {
                                                return 4;
                                            }
                                        }
                                        else if (Cs >= 2887) {
                                            return 2;
                                        }
                                    }
                                    else if (Rs >= 3898) {
                                        if (ssDCval < 123) {
                                            if (Cs < 3170) {
                                                return 4;
                                            }
                                            else if (Cs >= 3170) {
                                                if (mu_mv_mag_sq < 111) {
                                                    return 2;
                                                }
                                                else if (mu_mv_mag_sq >= 111) {
                                                    if (Cs < 3236) {
                                                        return 2;
                                                    }
                                                    else if (Cs >= 3236) {
                                                        if (Rs < 4874) {
                                                            return 4;
                                                        }
                                                        else if (Rs >= 4874) {
                                                            if (Rs < 6221) {
                                                                return 2;
                                                            }
                                                            else if (Rs >= 6221) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (ssDCval >= 123) {
                                            return 8;
                                        }
                                    }
                                }
                            }
                            else if (mcTcor >= 96) {
                                if (refDCval < 83) {
                                    if (AFD < 6) {
                                        if (SC < 3588) {
                                            if (refDCval < 73) {
                                                return 8;
                                            }
                                            else if (refDCval >= 73) {
                                                if (tcor < 97) {
                                                    if (ssDCval < 74) {
                                                        return 2;
                                                    }
                                                    else if (ssDCval >= 74) {
                                                        if (SC < 3488) {
                                                            return 8;
                                                        }
                                                        else if (SC >= 3488) {
                                                            if (Rs < 1800) {
                                                                return 8;
                                                            }
                                                            else if (Rs >= 1800) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (tcor >= 97) {
                                                    if (AFD < 4) {
                                                        if (MVDiff < 2) {
                                                            return 8;
                                                        }
                                                        else if (MVDiff >= 2) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (AFD >= 4) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                        }
                                        else if (SC >= 3588) {
                                            if (MVDiff < 1) {
                                                if (ssDCval < 79) {
                                                    return 4;
                                                }
                                                else if (ssDCval >= 79) {
                                                    if (Cs < 2108) {
                                                        return 8;
                                                    }
                                                    else if (Cs >= 2108) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                            else if (MVDiff >= 1) {
                                                return 8;
                                            }
                                        }
                                    }
                                    else if (AFD >= 6) {
                                        if (RsCsDiff < 22) {
                                            return 2;
                                        }
                                        else if (RsCsDiff >= 22) {
                                            return 4;
                                        }
                                    }
                                }
                                else if (refDCval >= 83) {
                                    if (posBalance < 58) {
                                        if (mu_mv_mag_sq < 123) {
                                            if (Rs < 2470) {
                                                if (Cs < 1534) {
                                                    return 8;
                                                }
                                                else if (Cs >= 1534) {
                                                    return 4;
                                                }
                                            }
                                            else if (Rs >= 2470) {
                                                if (Cs < 2945) {
                                                    return 8;
                                                }
                                                else if (Cs >= 2945) {
                                                    if (ssDCval < 112) {
                                                        return 8;
                                                    }
                                                    else if (ssDCval >= 112) {
                                                        if (negBalance < 48) {
                                                            if (Cs < 3073) {
                                                                if (refDCval < 113) {
                                                                    if (SC < 5999) {
                                                                        return 4;
                                                                    }
                                                                    else if (SC >= 5999) {
                                                                        return 8;
                                                                    }
                                                                }
                                                                else if (refDCval >= 113) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (Cs >= 3073) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (negBalance >= 48) {
                                                            if (Rs < 2885) {
                                                                if (Cs < 3104) {
                                                                    if (CsDiff < 4) {
                                                                        return 8;
                                                                    }
                                                                    else if (CsDiff >= 4) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (Cs >= 3104) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (Rs >= 2885) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (mu_mv_mag_sq >= 123) {
                                            if (diffRsCsdiff < -8) {
                                                if (refDCval < 112) {
                                                    return 8;
                                                }
                                                else if (refDCval >= 112) {
                                                    if (Cs < 3077) {
                                                        if (SC < 5757) {
                                                            return 8;
                                                        }
                                                        else if (SC >= 5757) {
                                                            if (AFD < 6) {
                                                                return 8;
                                                            }
                                                            else if (AFD >= 6) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                    else if (Cs >= 3077) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                            else if (diffRsCsdiff >= -8) {
                                                if (TSC < 4) {
                                                    return 8;
                                                }
                                                else if (TSC >= 4) {
                                                    if (Rs < 2398) {
                                                        return 4;
                                                    }
                                                    else if (Rs >= 2398) {
                                                        if (mu_mv_mag_sq < 124) {
                                                            if (Rs < 2801) {
                                                                return 8;
                                                            }
                                                            else if (Rs >= 2801) {
                                                                if (Rs < 2803) {
                                                                    return 4;
                                                                }
                                                                else if (Rs >= 2803) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                        else if (mu_mv_mag_sq >= 124) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (posBalance >= 58) {
                                        if (Rs < 5629) {
                                            if (tcor < 96) {
                                                return 4;
                                            }
                                            else if (tcor >= 96) {
                                                if (mu_mv_mag_sq < 123) {
                                                    if (Rs < 2735) {
                                                        return 8;
                                                    }
                                                    else if (Rs >= 2735) {
                                                        if (CsDiff < 5) {
                                                            if (Rs < 2738) {
                                                                return 4;
                                                            }
                                                            else if (Rs >= 2738) {
                                                                if (negBalance < 48) {
                                                                    return 8;
                                                                }
                                                                else if (negBalance >= 48) {
                                                                    if (Rs < 3026) {
                                                                        if (Cs < 3046) {
                                                                            return 8;
                                                                        }
                                                                        else if (Cs >= 3046) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                    else if (Rs >= 3026) {
                                                                        return 8;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (CsDiff >= 5) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                                else if (mu_mv_mag_sq >= 123) {
                                                    return 8;
                                                }
                                            }
                                        }
                                        else if (Rs >= 5629) {
                                            return 4;
                                        }
                                    }
                                }
                            }
                        }
                        else if (Cs >= 3618) {
                            if (Rs < 7267) {
                                if (Cs < 4952) {
                                    return 8;
                                }
                                else if (Cs >= 4952) {
                                    if (ssDCval < 106) {
                                        if (Rs < 5205) {
                                            if (Rs < 4978) {
                                                if (diffRsCsdiff < 25) {
                                                    return 8;
                                                }
                                                else if (diffRsCsdiff >= 25) {
                                                    return 4;
                                                }
                                            }
                                            else if (Rs >= 4978) {
                                                return 4;
                                            }
                                        }
                                        else if (Rs >= 5205) {
                                            return 8;
                                        }
                                    }
                                    else if (ssDCval >= 106) {
                                        if (Cs < 5857) {
                                            if (ssDCval < 109) {
                                                return 4;
                                            }
                                            else if (ssDCval >= 109) {
                                                return 8;
                                            }
                                        }
                                        else if (Cs >= 5857) {
                                            return 8;
                                        }
                                    }
                                }
                            }
                            else if (Rs >= 7267) {
                                if (CsDiff < 21) {
                                    return 4;
                                }
                                else if (CsDiff >= 21) {
                                    if (Rs < 8163) {
                                        return 2;
                                    }
                                    else if (Rs >= 8163) {
                                        return 4;
                                    }
                                }
                            }
                        }
                    }
                    else if (mu_mv_mag_sq >= 128) {
                        if (ssDCval < 104) {
                            if (ssDCval < 82) {
                                if (Cs < 4877) {
                                    return 8;
                                }
                                else if (Cs >= 4877) {
                                    if (SC < 10176) {
                                        if (RsCsDiff < 137) {
                                            if (CsDiff < 8) {
                                                return 8;
                                            }
                                            else if (CsDiff >= 8) {
                                                if (diffAFD < 1) {
                                                    if (diffMVdiffVal < -37) {
                                                        return 8;
                                                    }
                                                    else if (diffMVdiffVal >= -37) {
                                                        if (ssDCval < 80) {
                                                            if (Rs < 4777) {
                                                                return 4;
                                                            }
                                                            else if (Rs >= 4777) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (ssDCval >= 80) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                                else if (diffAFD >= 1) {
                                                    if (Rs < 4841) {
                                                        return 8;
                                                    }
                                                    else if (Rs >= 4841) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                        }
                                        else if (RsCsDiff >= 137) {
                                            if (diffTSC < 1) {
                                                if (Cs < 5027) {
                                                    if (Cs < 4985) {
                                                        return 8;
                                                    }
                                                    else if (Cs >= 4985) {
                                                        if (AFD < 7) {
                                                            if (SC < 9966) {
                                                                return 8;
                                                            }
                                                            else if (SC >= 9966) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (AFD >= 7) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                                else if (Cs >= 5027) {
                                                    return 8;
                                                }
                                            }
                                            else if (diffTSC >= 1) {
                                                return 4;
                                            }
                                        }
                                    }
                                    else if (SC >= 10176) {
                                        return 4;
                                    }
                                }
                            }
                            else if (ssDCval >= 82) {
                                return 8;
                            }
                        }
                        else if (ssDCval >= 104) {
                            if (AFD < 7) {
                                return 8;
                            }
                            else if (AFD >= 7) {
                                if (SC < 6654) {
                                    return 8;
                                }
                                else if (SC >= 6654) {
                                    return 4;
                                }
                            }
                        }
                    }
                }
            }
        }
        else if (MVDiff >= 6) {
            if (Cs < 3274) {
                if (mu_mv_mag_sq < 28) {
                    if (tcor < 80) {
                        if (tcor < 63) {
                            if (ssDCval < 128) {
                                if (refDCval < 93) {
                                    if (SC < 4655) {
                                        if (AFD < 1) {
                                            if (MVDiff < 134) {
                                                if (MVDiff < 65) {
                                                    return 4;
                                                }
                                                else if (MVDiff >= 65) {
                                                    if (MVDiff < 69) {
                                                        return 8;
                                                    }
                                                    else if (MVDiff >= 69) {
                                                        if (diffMVdiffVal < 33) {
                                                            return 4;
                                                        }
                                                        else if (diffMVdiffVal >= 33) {
                                                            return 1;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (MVDiff >= 134) {
                                                if (diffAFD < -1) {
                                                    if (posBalance < 94) {
                                                        if (posBalance < 87) {
                                                            return 2;
                                                        }
                                                        else if (posBalance >= 87) {
                                                            if (posBalance < 91) {
                                                                return 8;
                                                            }
                                                            else if (posBalance >= 91) {
                                                                return 2;
                                                            }
                                                        }
                                                    }
                                                    else if (posBalance >= 94) {
                                                        return 4;
                                                    }
                                                }
                                                else if (diffAFD >= -1) {
                                                    return 4;
                                                }
                                            }
                                        }
                                        else if (AFD >= 1) {
                                            if (diffAFD < 33) {
                                                if (MVDiff < 41) {
                                                    if (Rs < 595) {
                                                        if (mcTcor < 90) {
                                                            if (TSC < 6) {
                                                                return 2;
                                                            }
                                                            else if (TSC >= 6) {
                                                                return 1;
                                                            }
                                                        }
                                                        else if (mcTcor >= 90) {
                                                            if (TSC < 3) {
                                                                return 2;
                                                            }
                                                            else if (TSC >= 3) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                    else if (Rs >= 595) {
                                                        if (diffMVdiffVal < -56) {
                                                            return 2;
                                                        }
                                                        else if (diffMVdiffVal >= -56) {
                                                            if (tcor < 50) {
                                                                if (Cs < 732) {
                                                                    return 1;
                                                                }
                                                                else if (Cs >= 732) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (tcor >= 50) {
                                                                if (diffRsCsdiff < 30) {
                                                                    return 1;
                                                                }
                                                                else if (diffRsCsdiff >= 30) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (MVDiff >= 41) {
                                                    if (mcTcor < 43) {
                                                        if (diffMVdiffVal < -151) {
                                                            return 2;
                                                        }
                                                        else if (diffMVdiffVal >= -151) {
                                                            if (diffMVdiffVal < 26) {
                                                                return 1;
                                                            }
                                                            else if (diffMVdiffVal >= 26) {
                                                                if (refDCval < 18) {
                                                                    return 1;
                                                                }
                                                                else if (refDCval >= 18) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (mcTcor >= 43) {
                                                        if (posBalance < 59) {
                                                            if (MVDiff < 124) {
                                                                if (mu_mv_mag_sq < 10) {
                                                                    return 2;
                                                                }
                                                                else if (mu_mv_mag_sq >= 10) {
                                                                    if (Cs < 619) {
                                                                        return 1;
                                                                    }
                                                                    else if (Cs >= 619) {
                                                                        return 4;
                                                                    }
                                                                }
                                                            }
                                                            else if (MVDiff >= 124) {
                                                                if (SC < 381) {
                                                                    return 2;
                                                                }
                                                                else if (SC >= 381) {
                                                                    return 1;
                                                                }
                                                            }
                                                        }
                                                        else if (posBalance >= 59) {
                                                            if (tcor < 51) {
                                                                if (mu_mv_mag_sq < 26) {
                                                                    if (diffRsCsdiff < 99) {
                                                                        return 1;
                                                                    }
                                                                    else if (diffRsCsdiff >= 99) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (mu_mv_mag_sq >= 26) {
                                                                    return 2;
                                                                }
                                                            }
                                                            else if (tcor >= 51) {
                                                                if (MVDiff < 129) {
                                                                    if (diffAFD < 1) {
                                                                        if (Cs < 881) {
                                                                            return 4;
                                                                        }
                                                                        else if (Cs >= 881) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                    else if (diffAFD >= 1) {
                                                                        return 1;
                                                                    }
                                                                }
                                                                else if (MVDiff >= 129) {
                                                                    if (TSC < 3) {
                                                                        return 4;
                                                                    }
                                                                    else if (TSC >= 3) {
                                                                        return 2;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                            else if (diffAFD >= 33) {
                                                if (gchDC < 53) {
                                                    return 4;
                                                }
                                                else if (gchDC >= 53) {
                                                    return 2;
                                                }
                                            }
                                        }
                                    }
                                    else if (SC >= 4655) {
                                        if (RsDiff < 14) {
                                            return 4;
                                        }
                                        else if (RsDiff >= 14) {
                                            if (AFD < 28) {
                                                if (SC < 5329) {
                                                    return 8;
                                                }
                                                else if (SC >= 5329) {
                                                    return 1;
                                                }
                                            }
                                            else if (AFD >= 28) {
                                                return 2;
                                            }
                                        }
                                    }
                                }
                                else if (refDCval >= 93) {
                                    if (Cs < 709) {
                                        return 2;
                                    }
                                    else if (Cs >= 709) {
                                        if (RsDiff < 13) {
                                            if (Cs < 2514) {
                                                if (negBalance < 62) {
                                                    return 1;
                                                }
                                                else if (negBalance >= 62) {
                                                    if (Cs < 849) {
                                                        return 2;
                                                    }
                                                    else if (Cs >= 849) {
                                                        return 1;
                                                    }
                                                }
                                            }
                                            else if (Cs >= 2514) {
                                                return 8;
                                            }
                                        }
                                        else if (RsDiff >= 13) {
                                            if (Rs < 4850) {
                                                if (diffAFD < 5) {
                                                    if (RsCsDiff < 474) {
                                                        if (ssDCval < 114) {
                                                            if (diffRsCsdiff < 20) {
                                                                if (Rs < 3061) {
                                                                    return 2;
                                                                }
                                                                else if (Rs >= 3061) {
                                                                    return 1;
                                                                }
                                                            }
                                                            else if (diffRsCsdiff >= 20) {
                                                                if (diffTSC < 1) {
                                                                    if (SC < 5105) {
                                                                        return 2;
                                                                    }
                                                                    else if (SC >= 5105) {
                                                                        return 1;
                                                                    }
                                                                }
                                                                else if (diffTSC >= 1) {
                                                                    return 1;
                                                                }
                                                            }
                                                        }
                                                        else if (ssDCval >= 114) {
                                                            return 1;
                                                        }
                                                    }
                                                    else if (RsCsDiff >= 474) {
                                                        return 2;
                                                    }
                                                }
                                                else if (diffAFD >= 5) {
                                                    return 1;
                                                }
                                            }
                                            else if (Rs >= 4850) {
                                                return 1;
                                            }
                                        }
                                    }
                                }
                            }
                            else if (ssDCval >= 128) {
                                if (Rs < 2093) {
                                    if (CsDiff < 7) {
                                        return 2;
                                    }
                                    else if (CsDiff >= 7) {
                                        if (posBalance < 60) {
                                            if (mcTcor < 86) {
                                                if (AFD < 31) {
                                                    return 1;
                                                }
                                                else if (AFD >= 31) {
                                                    return 8;
                                                }
                                            }
                                            else if (mcTcor >= 86) {
                                                return 4;
                                            }
                                        }
                                        else if (posBalance >= 60) {
                                            if (MVDiff < 530) {
                                                return 8;
                                            }
                                            else if (MVDiff >= 530) {
                                                return 1;
                                            }
                                        }
                                    }
                                }
                                else if (Rs >= 2093) {
                                    if (AFD < 29) {
                                        return 1;
                                    }
                                    else if (AFD >= 29) {
                                        return 4;
                                    }
                                }
                            }
                        }
                        else if (tcor >= 63) {
                            if (mcTcor < 92) {
                                if (posBalance < 59) {
                                    if (ssDCval < 124) {
                                        if (Cs < 1880) {
                                            if (Cs < 158) {
                                                if (TSCindex < 1) {
                                                    return 8;
                                                }
                                                else if (TSCindex >= 1) {
                                                    if (MVDiff < 25) {
                                                        return 2;
                                                    }
                                                    else if (MVDiff >= 25) {
                                                        return 1;
                                                    }
                                                }
                                            }
                                            else if (Cs >= 158) {
                                                if (ssDCval < 84) {
                                                    if (RsCsDiff < 76) {
                                                        if (mcTcor < 86) {
                                                            if (Cs < 555) {
                                                                return 4;
                                                            }
                                                            else if (Cs >= 555) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (mcTcor >= 86) {
                                                            return 2;
                                                        }
                                                    }
                                                    else if (RsCsDiff >= 76) {
                                                        if (RsDiff < 10) {
                                                            if (CsDiff < 8) {
                                                                if (SC < 2532) {
                                                                    if (AFD < 11) {
                                                                        return 1;
                                                                    }
                                                                    else if (AFD >= 11) {
                                                                        return 2;
                                                                    }
                                                                }
                                                                else if (SC >= 2532) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (CsDiff >= 8) {
                                                                return 1;
                                                            }
                                                        }
                                                        else if (RsDiff >= 10) {
                                                            if (MVDiff < 104) {
                                                                if (diffRsCsdiff < 36) {
                                                                    return 2;
                                                                }
                                                                else if (diffRsCsdiff >= 36) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (MVDiff >= 104) {
                                                                return 1;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (ssDCval >= 84) {
                                                    if (Cs < 1586) {
                                                        if (AFD < 13) {
                                                            if (RsCsDiff < 23) {
                                                                if (TSC < 6) {
                                                                    if (AFD < 9) {
                                                                        return 2;
                                                                    }
                                                                    else if (AFD >= 9) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (TSC >= 6) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (RsCsDiff >= 23) {
                                                                if (posBalance < 58) {
                                                                    if (Rs < 1569) {
                                                                        return 2;
                                                                    }
                                                                    else if (Rs >= 1569) {
                                                                        return 8;
                                                                    }
                                                                }
                                                                else if (posBalance >= 58) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                        else if (AFD >= 13) {
                                                            if (MVDiff < 108) {
                                                                return 2;
                                                            }
                                                            else if (MVDiff >= 108) {
                                                                return 1;
                                                            }
                                                        }
                                                    }
                                                    else if (Cs >= 1586) {
                                                        if (tcor < 73) {
                                                            if (SC < 3943) {
                                                                return 2;
                                                            }
                                                            else if (SC >= 3943) {
                                                                return 1;
                                                            }
                                                        }
                                                        else if (tcor >= 73) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (Cs >= 1880) {
                                            if (mcTcor < 88) {
                                                if (AFD < 16) {
                                                    if (Rs < 2593) {
                                                        return 4;
                                                    }
                                                    else if (Rs >= 2593) {
                                                        if (Cs < 3060) {
                                                            return 4;
                                                        }
                                                        else if (Cs >= 3060) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                                else if (AFD >= 16) {
                                                    if (refDCval < 122) {
                                                        if (mu_mv_mag_sq < 11) {
                                                            if (diffRsCsdiff < 95) {
                                                                return 8;
                                                            }
                                                            else if (diffRsCsdiff >= 95) {
                                                                if (Rs < 2994) {
                                                                    return 8;
                                                                }
                                                                else if (Rs >= 2994) {
                                                                    return 1;
                                                                }
                                                            }
                                                        }
                                                        else if (mu_mv_mag_sq >= 11) {
                                                            if (Rs < 3115) {
                                                                if (AFD < 18) {
                                                                    return 8;
                                                                }
                                                                else if (AFD >= 18) {
                                                                    return 1;
                                                                }
                                                            }
                                                            else if (Rs >= 3115) {
                                                                if (diffMVdiffVal < 8) {
                                                                    return 8;
                                                                }
                                                                else if (diffMVdiffVal >= 8) {
                                                                    if (SC < 5818) {
                                                                        return 4;
                                                                    }
                                                                    else if (SC >= 5818) {
                                                                        return 1;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (refDCval >= 122) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                            else if (mcTcor >= 88) {
                                                if (MVDiff < 11) {
                                                    if (Cs < 2393) {
                                                        return 2;
                                                    }
                                                    else if (Cs >= 2393) {
                                                        if (Rs < 2907) {
                                                            return 4;
                                                        }
                                                        else if (Rs >= 2907) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                                else if (MVDiff >= 11) {
                                                    if (refDCval < 77) {
                                                        return 8;
                                                    }
                                                    else if (refDCval >= 77) {
                                                        return 1;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (ssDCval >= 124) {
                                        if (ssDCval < 141) {
                                            if (MVDiff < 22) {
                                                if (SC < 3125) {
                                                    if (diffRsCsdiff < 15) {
                                                        if (Rs < 1320) {
                                                            return 8;
                                                        }
                                                        else if (Rs >= 1320) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (diffRsCsdiff >= 15) {
                                                        return 1;
                                                    }
                                                }
                                                else if (SC >= 3125) {
                                                    if (Cs < 2836) {
                                                        return 1;
                                                    }
                                                    else if (Cs >= 2836) {
                                                        if (gchDC < 1) {
                                                            return 1;
                                                        }
                                                        else if (gchDC >= 1) {
                                                            if (AFD < 24) {
                                                                return 8;
                                                            }
                                                            else if (AFD >= 24) {
                                                                return 1;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                            else if (MVDiff >= 22) {
                                                if (MVDiff < 63) {
                                                    return 4;
                                                }
                                                else if (MVDiff >= 63) {
                                                    if (Cs < 803) {
                                                        return 2;
                                                    }
                                                    else if (Cs >= 803) {
                                                        return 1;
                                                    }
                                                }
                                            }
                                        }
                                        else if (ssDCval >= 141) {
                                            if (Rs < 903) {
                                                if (Cs < 427) {
                                                    return 2;
                                                }
                                                else if (Cs >= 427) {
                                                    return 1;
                                                }
                                            }
                                            else if (Rs >= 903) {
                                                if (diffRsCsdiff < -396) {
                                                    return 2;
                                                }
                                                else if (diffRsCsdiff >= -396) {
                                                    if (posBalance < 33) {
                                                        return 1;
                                                    }
                                                    else if (posBalance >= 33) {
                                                        if (mcTcor < 80) {
                                                            if (Rs < 1269) {
                                                                return 4;
                                                            }
                                                            else if (Rs >= 1269) {
                                                                return 1;
                                                            }
                                                        }
                                                        else if (mcTcor >= 80) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (posBalance >= 59) {
                                    if (ssDCval < 102) {
                                        if (MVDiff < 17) {
                                            if (ssDCval < 71) {
                                                if (RsCsDiff < 99) {
                                                    if (posBalance < 65) {
                                                        return 8;
                                                    }
                                                    else if (posBalance >= 65) {
                                                        if (MVDiff < 10) {
                                                            if (Cs < 1402) {
                                                                return 4;
                                                            }
                                                            else if (Cs >= 1402) {
                                                                return 2;
                                                            }
                                                        }
                                                        else if (MVDiff >= 10) {
                                                            return 2;
                                                        }
                                                    }
                                                }
                                                else if (RsCsDiff >= 99) {
                                                    return 1;
                                                }
                                            }
                                            else if (ssDCval >= 71) {
                                                if (Rs < 606) {
                                                    if (Cs < 409) {
                                                        if (Cs < 366) {
                                                            return 2;
                                                        }
                                                        else if (Cs >= 366) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (Cs >= 409) {
                                                        if (gchDC < 1) {
                                                            return 8;
                                                        }
                                                        else if (gchDC >= 1) {
                                                            if (Cs < 487) {
                                                                return 8;
                                                            }
                                                            else if (Cs >= 487) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (Rs >= 606) {
                                                    if (SC < 2249) {
                                                        return 4;
                                                    }
                                                    else if (SC >= 2249) {
                                                        if (mcTcor < 85) {
                                                            return 4;
                                                        }
                                                        else if (mcTcor >= 85) {
                                                            if (refDCval < 99) {
                                                                return 2;
                                                            }
                                                            else if (refDCval >= 99) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (MVDiff >= 17) {
                                            if (AFD < 9) {
                                                if (Rs < 292) {
                                                    if (tcor < 70) {
                                                        if (SC < 88) {
                                                            return 8;
                                                        }
                                                        else if (SC >= 88) {
                                                            return 2;
                                                        }
                                                    }
                                                    else if (tcor >= 70) {
                                                        if (MVDiff < 56) {
                                                            return 2;
                                                        }
                                                        else if (MVDiff >= 56) {
                                                            if (MVDiff < 84) {
                                                                return 4;
                                                            }
                                                            else if (MVDiff >= 84) {
                                                                return 2;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (Rs >= 292) {
                                                    if (Rs < 956) {
                                                        return 4;
                                                    }
                                                    else if (Rs >= 956) {
                                                        return 2;
                                                    }
                                                }
                                            }
                                            else if (AFD >= 9) {
                                                if (SC < 832) {
                                                    return 1;
                                                }
                                                else if (SC >= 832) {
                                                    if (refDCval < 64) {
                                                        if (posBalance < 61) {
                                                            return 2;
                                                        }
                                                        else if (posBalance >= 61) {
                                                            return 1;
                                                        }
                                                    }
                                                    else if (refDCval >= 64) {
                                                        if (AFD < 22) {
                                                            if (negBalance < 52) {
                                                                if (Rs < 634) {
                                                                    return 8;
                                                                }
                                                                else if (Rs >= 634) {
                                                                    if (tcor < 71) {
                                                                        return 2;
                                                                    }
                                                                    else if (tcor >= 71) {
                                                                        if (posBalance < 66) {
                                                                            return 4;
                                                                        }
                                                                        else if (posBalance >= 66) {
                                                                            return 2;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                            else if (negBalance >= 52) {
                                                                if (ssDCval < 83) {
                                                                    return 8;
                                                                }
                                                                else if (ssDCval >= 83) {
                                                                    if (RsDiff < 3) {
                                                                        return 8;
                                                                    }
                                                                    else if (RsDiff >= 3) {
                                                                        return 4;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (AFD >= 22) {
                                                            if (diffAFD < 1) {
                                                                return 1;
                                                            }
                                                            else if (diffAFD >= 1) {
                                                                if (Rs < 2227) {
                                                                    return 8;
                                                                }
                                                                else if (Rs >= 2227) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (ssDCval >= 102) {
                                        if (negBalance < 52) {
                                            if (TSC < 11) {
                                                if (Cs < 1318) {
                                                    return 1;
                                                }
                                                else if (Cs >= 1318) {
                                                    return 4;
                                                }
                                            }
                                            else if (TSC >= 11) {
                                                if (mcTcor < 90) {
                                                    return 1;
                                                }
                                                else if (mcTcor >= 90) {
                                                    if (negBalance < 48) {
                                                        return 2;
                                                    }
                                                    else if (negBalance >= 48) {
                                                        return 1;
                                                    }
                                                }
                                            }
                                        }
                                        else if (negBalance >= 52) {
                                            if (Cs < 1410) {
                                                if (Cs < 331) {
                                                    return 1;
                                                }
                                                else if (Cs >= 331) {
                                                    return 2;
                                                }
                                            }
                                            else if (Cs >= 1410) {
                                                if (ssDCval < 114) {
                                                    return 1;
                                                }
                                                else if (ssDCval >= 114) {
                                                    if (mu_mv_mag_sq < 16) {
                                                        if (AFD < 24) {
                                                            if (diffTSC < 1) {
                                                                if (mcTcor < 89) {
                                                                    return 4;
                                                                }
                                                                else if (mcTcor >= 89) {
                                                                    if (SC < 6766) {
                                                                        return 1;
                                                                    }
                                                                    else if (SC >= 6766) {
                                                                        return 4;
                                                                    }
                                                                }
                                                            }
                                                            else if (diffTSC >= 1) {
                                                                if (Cs < 2594) {
                                                                    if (posBalance < 66) {
                                                                        return 4;
                                                                    }
                                                                    else if (posBalance >= 66) {
                                                                        return 2;
                                                                    }
                                                                }
                                                                else if (Cs >= 2594) {
                                                                    if (Rs < 4055) {
                                                                        return 1;
                                                                    }
                                                                    else if (Rs >= 4055) {
                                                                        return 2;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (AFD >= 24) {
                                                            if (AFD < 26) {
                                                                if (Rs < 2718) {
                                                                    return 2;
                                                                }
                                                                else if (Rs >= 2718) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (AFD >= 26) {
                                                                return 1;
                                                            }
                                                        }
                                                    }
                                                    else if (mu_mv_mag_sq >= 16) {
                                                        return 1;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                            else if (mcTcor >= 92) {
                                if (TSC < 6) {
                                    if (TSC < 4) {
                                        return 2;
                                    }
                                    else if (TSC >= 4) {
                                        if (tcor < 74) {
                                            return 4;
                                        }
                                        else if (tcor >= 74) {
                                            if (Cs < 1272) {
                                                return 1;
                                            }
                                            else if (Cs >= 1272) {
                                                return 4;
                                            }
                                        }
                                    }
                                }
                                else if (TSC >= 6) {
                                    if (refDCval < 63) {
                                        if (Cs < 796) {
                                            return 1;
                                        }
                                        else if (Cs >= 796) {
                                            return 8;
                                        }
                                    }
                                    else if (refDCval >= 63) {
                                        if (posBalance < 56) {
                                            if (Rs < 3957) {
                                                if (tcor < 78) {
                                                    return 2;
                                                }
                                                else if (tcor >= 78) {
                                                    if (negBalance < 48) {
                                                        if (Cs < 1463) {
                                                            return 2;
                                                        }
                                                        else if (Cs >= 1463) {
                                                            return 1;
                                                        }
                                                    }
                                                    else if (negBalance >= 48) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                            else if (Rs >= 3957) {
                                                return 4;
                                            }
                                        }
                                        else if (posBalance >= 56) {
                                            if (mcTcor < 93) {
                                                if (MVDiff < 20) {
                                                    if (MVDiff < 12) {
                                                        return 4;
                                                    }
                                                    else if (MVDiff >= 12) {
                                                        return 1;
                                                    }
                                                }
                                                else if (MVDiff >= 20) {
                                                    if (AFD < 13) {
                                                        return 1;
                                                    }
                                                    else if (AFD >= 13) {
                                                        return 2;
                                                    }
                                                }
                                            }
                                            else if (mcTcor >= 93) {
                                                if (negBalance < 37) {
                                                    return 4;
                                                }
                                                else if (negBalance >= 37) {
                                                    if (RsDiff < 5) {
                                                        if (SC < 1653) {
                                                            return 2;
                                                        }
                                                        else if (SC >= 1653) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (RsDiff >= 5) {
                                                        return 2;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                        }
                    }
                    else if (tcor >= 80) {
                        if (refDCval < 104) {
                            if (Cs < 1497) {
                                if (CsDiff < 2) {
                                    if (refDCval < 84) {
                                        if (Rs < 1) {
                                            if (mu_mv_mag_sq < 8) {
                                                if (negBalance < 108) {
                                                    return 2;
                                                }
                                                else if (negBalance >= 108) {
                                                    return 8;
                                                }
                                            }
                                            else if (mu_mv_mag_sq >= 8) {
                                                if (TSC < 2) {
                                                    if (gchDC < 1) {
                                                        return 8;
                                                    }
                                                    else if (gchDC >= 1) {
                                                        if (tcor < 93) {
                                                            if (MVDiff < 587) {
                                                                return 8;
                                                            }
                                                            else if (MVDiff >= 587) {
                                                                return 2;
                                                            }
                                                        }
                                                        else if (tcor >= 93) {
                                                            if (posBalance < 3) {
                                                                return 4;
                                                            }
                                                            else if (posBalance >= 3) {
                                                                if (AFD < 1) {
                                                                    return 4;
                                                                }
                                                                else if (AFD >= 1) {
                                                                    return 2;
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (TSC >= 2) {
                                                    if (MVDiff < 32) {
                                                        return 4;
                                                    }
                                                    else if (MVDiff >= 32) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                        }
                                        else if (Rs >= 1) {
                                            if (MVDiff < 38) {
                                                if (mu_mv_mag_sq < 3) {
                                                    if (SC < 473) {
                                                        if (refDCval < 32) {
                                                            return 4;
                                                        }
                                                        else if (refDCval >= 32) {
                                                            return 2;
                                                        }
                                                    }
                                                    else if (SC >= 473) {
                                                        if (SC < 2776) {
                                                            return 4;
                                                        }
                                                        else if (SC >= 2776) {
                                                            if (AFD < 6) {
                                                                return 4;
                                                            }
                                                            else if (AFD >= 6) {
                                                                return 2;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (mu_mv_mag_sq >= 3) {
                                                    if (diffRsCsdiff < -13) {
                                                        return 4;
                                                    }
                                                    else if (diffRsCsdiff >= -13) {
                                                        return 2;
                                                    }
                                                }
                                            }
                                            else if (MVDiff >= 38) {
                                                if (tcor < 101) {
                                                    if (AFD < 3) {
                                                        return 2;
                                                    }
                                                    else if (AFD >= 3) {
                                                        return 1;
                                                    }
                                                }
                                                else if (tcor >= 101) {
                                                    return 2;
                                                }
                                            }
                                        }
                                    }
                                    else if (refDCval >= 84) {
                                        if (AFD < 6) {
                                            return 2;
                                        }
                                        else if (AFD >= 6) {
                                            return 1;
                                        }
                                    }
                                }
                                else if (CsDiff >= 2) {
                                    if (SC < 1721) {
                                        if (ssDCval < 71) {
                                            if (diffTSC < -5) {
                                                return 8;
                                            }
                                            else if (diffTSC >= -5) {
                                                if (Cs < 501) {
                                                    if (Cs < 475) {
                                                        return 2;
                                                    }
                                                    else if (Cs >= 475) {
                                                        return 8;
                                                    }
                                                }
                                                else if (Cs >= 501) {
                                                    return 2;
                                                }
                                            }
                                        }
                                        else if (ssDCval >= 71) {
                                            if (negBalance < 53) {
                                                if (CsDiff < 3) {
                                                    return 8;
                                                }
                                                else if (CsDiff >= 3) {
                                                    return 2;
                                                }
                                            }
                                            else if (negBalance >= 53) {
                                                if (Cs < 565) {
                                                    return 2;
                                                }
                                                else if (Cs >= 565) {
                                                    return 4;
                                                }
                                            }
                                        }
                                    }
                                    else if (SC >= 1721) {
                                        if (ssDCval < 62) {
                                            if (SC < 1977) {
                                                if (AFD < 9) {
                                                    return 4;
                                                }
                                                else if (AFD >= 9) {
                                                    return 1;
                                                }
                                            }
                                            else if (SC >= 1977) {
                                                if (Rs < 1200) {
                                                    if (AFD < 8) {
                                                        return 1;
                                                    }
                                                    else if (AFD >= 8) {
                                                        return 8;
                                                    }
                                                }
                                                else if (Rs >= 1200) {
                                                    return 1;
                                                }
                                            }
                                        }
                                        else if (ssDCval >= 62) {
                                            if (MVDiff < 12) {
                                                if (Rs < 1403) {
                                                    if (AFD < 8) {
                                                        return 2;
                                                    }
                                                    else if (AFD >= 8) {
                                                        return 4;
                                                    }
                                                }
                                                else if (Rs >= 1403) {
                                                    if (RsCsDiff < 135) {
                                                        return 1;
                                                    }
                                                    else if (RsCsDiff >= 135) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                            else if (MVDiff >= 12) {
                                                if (tcor < 87) {
                                                    if (RsDiff < 10) {
                                                        if (Rs < 1052) {
                                                            return 2;
                                                        }
                                                        else if (Rs >= 1052) {
                                                            if (negBalance < 53) {
                                                                if (AFD < 13) {
                                                                    return 4;
                                                                }
                                                                else if (AFD >= 13) {
                                                                    return 2;
                                                                }
                                                            }
                                                            else if (negBalance >= 53) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                    else if (RsDiff >= 10) {
                                                        return 2;
                                                    }
                                                }
                                                else if (tcor >= 87) {
                                                    if (mcTcor < 93) {
                                                        return 2;
                                                    }
                                                    else if (mcTcor >= 93) {
                                                        if (mcTcor < 97) {
                                                            return 1;
                                                        }
                                                        else if (mcTcor >= 97) {
                                                            if (Rs < 2392) {
                                                                return 2;
                                                            }
                                                            else if (Rs >= 2392) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                            else if (Cs >= 1497) {
                                if (posBalance < 52) {
                                    if (CsDiff < 10) {
                                        if (diffRsCsdiff < -62) {
                                            return 1;
                                        }
                                        else if (diffRsCsdiff >= -62) {
                                            if (diffRsCsdiff < 40) {
                                                if (refDCval < 97) {
                                                    if (AFD < 15) {
                                                        return 8;
                                                    }
                                                    else if (AFD >= 15) {
                                                        if (posBalance < 50) {
                                                            return 1;
                                                        }
                                                        else if (posBalance >= 50) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                                else if (refDCval >= 97) {
                                                    return 4;
                                                }
                                            }
                                            else if (diffRsCsdiff >= 40) {
                                                if (Cs < 2314) {
                                                    return 1;
                                                }
                                                else if (Cs >= 2314) {
                                                    return 4;
                                                }
                                            }
                                        }
                                    }
                                    else if (CsDiff >= 10) {
                                        if (mcTcor < 95) {
                                            return 1;
                                        }
                                        else if (mcTcor >= 95) {
                                            if (SC < 3967) {
                                                return 2;
                                            }
                                            else if (SC >= 3967) {
                                                return 4;
                                            }
                                        }
                                    }
                                }
                                else if (posBalance >= 52) {
                                    if (negBalance < 52) {
                                        if (Rs < 2509) {
                                            if (CsDiff < 9) {
                                                if (Cs < 2550) {
                                                    if (RsDiff < 10) {
                                                        return 4;
                                                    }
                                                    else if (RsDiff >= 10) {
                                                        return 2;
                                                    }
                                                }
                                                else if (Cs >= 2550) {
                                                    return 2;
                                                }
                                            }
                                            else if (CsDiff >= 9) {
                                                if (RsCsDiff < 212) {
                                                    if (posBalance < 54) {
                                                        if (Rs < 2288) {
                                                            if (AFD < 17) {
                                                                return 8;
                                                            }
                                                            else if (AFD >= 17) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (Rs >= 2288) {
                                                            return 1;
                                                        }
                                                    }
                                                    else if (posBalance >= 54) {
                                                        if (SC < 3317) {
                                                            return 1;
                                                        }
                                                        else if (SC >= 3317) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                                else if (RsCsDiff >= 212) {
                                                    if (MVDiff < 20) {
                                                        return 4;
                                                    }
                                                    else if (MVDiff >= 20) {
                                                        if (posBalance < 58) {
                                                            return 1;
                                                        }
                                                        else if (posBalance >= 58) {
                                                            if (Rs < 2369) {
                                                                return 1;
                                                            }
                                                            else if (Rs >= 2369) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (Rs >= 2509) {
                                            if (RsCsDiff < 367) {
                                                if (ssDCval < 84) {
                                                    if (diffAFD < 2) {
                                                        if (MVDiff < 21) {
                                                            if (Cs < 2376) {
                                                                return 4;
                                                            }
                                                            else if (Cs >= 2376) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (MVDiff >= 21) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (diffAFD >= 2) {
                                                        if (CsDiff < 6) {
                                                            return 2;
                                                        }
                                                        else if (CsDiff >= 6) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                                else if (ssDCval >= 84) {
                                                    if (AFD < 14) {
                                                        if (Cs < 2272) {
                                                            return 1;
                                                        }
                                                        else if (Cs >= 2272) {
                                                            return 2;
                                                        }
                                                    }
                                                    else if (AFD >= 14) {
                                                        if (posBalance < 55) {
                                                            if (refDCval < 86) {
                                                                if (TSC < 11) {
                                                                    return 1;
                                                                }
                                                                else if (TSC >= 11) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (refDCval >= 86) {
                                                                if (SC < 5282) {
                                                                    return 1;
                                                                }
                                                                else if (SC >= 5282) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                        else if (posBalance >= 55) {
                                                            if (ssDCval < 101) {
                                                                return 4;
                                                            }
                                                            else if (ssDCval >= 101) {
                                                                if (posBalance < 61) {
                                                                    return 8;
                                                                }
                                                                else if (posBalance >= 61) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                            else if (RsCsDiff >= 367) {
                                                return 1;
                                            }
                                        }
                                    }
                                    else if (negBalance >= 52) {
                                        if (RsCsDiff < 183) {
                                            if (posBalance < 58) {
                                                if (ssDCval < 81) {
                                                    if (ssDCval < 49) {
                                                        return 2;
                                                    }
                                                    else if (ssDCval >= 49) {
                                                        return 8;
                                                    }
                                                }
                                                else if (ssDCval >= 81) {
                                                    return 2;
                                                }
                                            }
                                            else if (posBalance >= 58) {
                                                if (MVDiff < 16) {
                                                    if (ssDCval < 82) {
                                                        return 2;
                                                    }
                                                    else if (ssDCval >= 82) {
                                                        return 1;
                                                    }
                                                }
                                                else if (MVDiff >= 16) {
                                                    return 2;
                                                }
                                            }
                                        }
                                        else if (RsCsDiff >= 183) {
                                            if (SC < 5311) {
                                                if (Cs < 2396) {
                                                    return 2;
                                                }
                                                else if (Cs >= 2396) {
                                                    if (Cs < 2464) {
                                                        return 4;
                                                    }
                                                    else if (Cs >= 2464) {
                                                        return 2;
                                                    }
                                                }
                                            }
                                            else if (SC >= 5311) {
                                                return 4;
                                            }
                                        }
                                    }
                                }
                            }
                        }
                        else if (refDCval >= 104) {
                            if (negBalance < 44) {
                                if (Cs < 1276) {
                                    if (posBalance < 61) {
                                        if (Rs < 1335) {
                                            return 1;
                                        }
                                        else if (Rs >= 1335) {
                                            if (SC < 3101) {
                                                return 2;
                                            }
                                            else if (SC >= 3101) {
                                                return 8;
                                            }
                                        }
                                    }
                                    else if (posBalance >= 61) {
                                        if (Cs < 267) {
                                            return 1;
                                        }
                                        else if (Cs >= 267) {
                                            return 4;
                                        }
                                    }
                                }
                                else if (Cs >= 1276) {
                                    if (refDCval < 119) {
                                        if (SC < 5053) {
                                            if (negBalance < 40) {
                                                if (mcTcor < 94) {
                                                    return 8;
                                                }
                                                else if (mcTcor >= 94) {
                                                    if (RsCsDiff < 47) {
                                                        if (diffRsCsdiff < -8) {
                                                            if (Rs < 3195) {
                                                                return 4;
                                                            }
                                                            else if (Rs >= 3195) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (diffRsCsdiff >= -8) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (RsCsDiff >= 47) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                            else if (negBalance >= 40) {
                                                if (Rs < 1599) {
                                                    return 2;
                                                }
                                                else if (Rs >= 1599) {
                                                    return 4;
                                                }
                                            }
                                        }
                                        else if (SC >= 5053) {
                                            return 8;
                                        }
                                    }
                                    else if (refDCval >= 119) {
                                        return 4;
                                    }
                                }
                            }
                            else if (negBalance >= 44) {
                                if (RsDiff < 8) {
                                    if (mu_mv_mag_sq < 3) {
                                        if (Rs < 1786) {
                                            if (RsCsDiff < 9) {
                                                return 4;
                                            }
                                            else if (RsCsDiff >= 9) {
                                                if (MVDiff < 29) {
                                                    if (AFD < 10) {
                                                        if (Rs < 461) {
                                                            return 4;
                                                        }
                                                        else if (Rs >= 461) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (AFD >= 10) {
                                                        return 2;
                                                    }
                                                }
                                                else if (MVDiff >= 29) {
                                                    return 1;
                                                }
                                            }
                                        }
                                        else if (Rs >= 1786) {
                                            return 4;
                                        }
                                    }
                                    else if (mu_mv_mag_sq >= 3) {
                                        if (negBalance < 104) {
                                            if (posBalance < 46) {
                                                if (Cs < 1525) {
                                                    return 8;
                                                }
                                                else if (Cs >= 1525) {
                                                    return 2;
                                                }
                                            }
                                            else if (posBalance >= 46) {
                                                if (negBalance < 60) {
                                                    if (tcor < 81) {
                                                        return 1;
                                                    }
                                                    else if (tcor >= 81) {
                                                        return 2;
                                                    }
                                                }
                                                else if (negBalance >= 60) {
                                                    return 1;
                                                }
                                            }
                                        }
                                        else if (negBalance >= 104) {
                                            if (TSC < 12) {
                                                return 8;
                                            }
                                            else if (TSC >= 12) {
                                                return 4;
                                            }
                                        }
                                    }
                                }
                                else if (RsDiff >= 8) {
                                    if (ssDCval < 126) {
                                        if (SC < 3444) {
                                            if (MVDiff < 16) {
                                                if (Cs < 1454) {
                                                    return 4;
                                                }
                                                else if (Cs >= 1454) {
                                                    return 2;
                                                }
                                            }
                                            else if (MVDiff >= 16) {
                                                return 2;
                                            }
                                        }
                                        else if (SC >= 3444) {
                                            if (TSC < 13) {
                                                if (CsDiff < 7) {
                                                    if (negBalance < 53) {
                                                        if (Rs < 2866) {
                                                            return 8;
                                                        }
                                                        else if (Rs >= 2866) {
                                                            return 2;
                                                        }
                                                    }
                                                    else if (negBalance >= 53) {
                                                        return 4;
                                                    }
                                                }
                                                else if (CsDiff >= 7) {
                                                    if (tcor < 87) {
                                                        if (ssDCval < 103) {
                                                            return 4;
                                                        }
                                                        else if (ssDCval >= 103) {
                                                            if (ssDCval < 117) {
                                                                return 2;
                                                            }
                                                            else if (ssDCval >= 117) {
                                                                if (Rs < 3074) {
                                                                    if (Rs < 1980) {
                                                                        return 2;
                                                                    }
                                                                    else if (Rs >= 1980) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (Rs >= 3074) {
                                                                    return 2;
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (tcor >= 87) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                            else if (TSC >= 13) {
                                                if (Cs < 1790) {
                                                    if (MVDiff < 30) {
                                                        return 4;
                                                    }
                                                    else if (MVDiff >= 30) {
                                                        if (Cs < 1652) {
                                                            return 4;
                                                        }
                                                        else if (Cs >= 1652) {
                                                            return 2;
                                                        }
                                                    }
                                                }
                                                else if (Cs >= 1790) {
                                                    if (RsDiff < 13) {
                                                        return 1;
                                                    }
                                                    else if (RsDiff >= 13) {
                                                        if (gchDC < 3) {
                                                            if (AFD < 22) {
                                                                if (Cs < 1856) {
                                                                    return 8;
                                                                }
                                                                else if (Cs >= 1856) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (AFD >= 22) {
                                                                return 1;
                                                            }
                                                        }
                                                        else if (gchDC >= 3) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (ssDCval >= 126) {
                                        if (RsDiff < 11) {
                                            if (negBalance < 52) {
                                                if (refDCval < 129) {
                                                    return 8;
                                                }
                                                else if (refDCval >= 129) {
                                                    return 1;
                                                }
                                            }
                                            else if (negBalance >= 52) {
                                                return 2;
                                            }
                                        }
                                        else if (RsDiff >= 11) {
                                            return 2;
                                        }
                                    }
                                }
                            }
                        }
                    }
                }
                else if (mu_mv_mag_sq >= 28) {
                    if (AFD < 12) {
                        if (refDCval < 94) {
                            if (mu_mv_mag_sq < 90) {
                                if (SC < 1091) {
                                    if (AFD < 5) {
                                        if (tcor < 48) {
                                            if (diffMVdiffVal < 17) {
                                                if (MVDiff < 62) {
                                                    if (diffMVdiffVal < -7) {
                                                        return 8;
                                                    }
                                                    else if (diffMVdiffVal >= -7) {
                                                        if (posBalance < 27) {
                                                            return 1;
                                                        }
                                                        else if (posBalance >= 27) {
                                                            if (diffMVdiffVal < 2) {
                                                                return 4;
                                                            }
                                                            else if (diffMVdiffVal >= 2) {
                                                                return 1;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (MVDiff >= 62) {
                                                    if (MVDiff < 253) {
                                                        if (tcor < 41) {
                                                            if (MVDiff < 64) {
                                                                return 8;
                                                            }
                                                            else if (MVDiff >= 64) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (tcor >= 41) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (MVDiff >= 253) {
                                                        return 1;
                                                    }
                                                }
                                            }
                                            else if (diffMVdiffVal >= 17) {
                                                return 8;
                                            }
                                        }
                                        else if (tcor >= 48) {
                                            if (Cs < 385) {
                                                if (posBalance < 6) {
                                                    if (MVDiff < 43) {
                                                        return 8;
                                                    }
                                                    else if (MVDiff >= 43) {
                                                        if (diffMVdiffVal < -21) {
                                                            return 1;
                                                        }
                                                        else if (diffMVdiffVal >= -21) {
                                                            return 2;
                                                        }
                                                    }
                                                }
                                                else if (posBalance >= 6) {
                                                    if (Rs < 401) {
                                                        if (posBalance < 61) {
                                                            if (tcor < 100) {
                                                                if (MVDiff < 98) {
                                                                    if (mcTcor < 90) {
                                                                        if (Cs < 334) {
                                                                            if (MVDiff < 27) {
                                                                                if (RsDiff < 1) {
                                                                                    return 4;
                                                                                }
                                                                                else if (RsDiff >= 1) {
                                                                                    return 2;
                                                                                }
                                                                            }
                                                                            else if (MVDiff >= 27) {
                                                                                if (Rs < 379) {
                                                                                    if (refDCval < 21) {
                                                                                        return 4;
                                                                                    }
                                                                                    else if (refDCval >= 21) {
                                                                                        return 8;
                                                                                    }
                                                                                }
                                                                                else if (Rs >= 379) {
                                                                                    return 4;
                                                                                }
                                                                            }
                                                                        }
                                                                        else if (Cs >= 334) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                    else if (mcTcor >= 90) {
                                                                        if (Rs < 265) {
                                                                            return 2;
                                                                        }
                                                                        else if (Rs >= 265) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                }
                                                                else if (MVDiff >= 98) {
                                                                    if (AFD < 2) {
                                                                        return 2;
                                                                    }
                                                                    else if (AFD >= 2) {
                                                                        return 1;
                                                                    }
                                                                }
                                                            }
                                                            else if (tcor >= 100) {
                                                                if (ssDCval < 23) {
                                                                    return 8;
                                                                }
                                                                else if (ssDCval >= 23) {
                                                                    if (SC < 455) {
                                                                        return 4;
                                                                    }
                                                                    else if (SC >= 455) {
                                                                        return 2;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (posBalance >= 61) {
                                                            if (ssDCval < 37) {
                                                                return 4;
                                                            }
                                                            else if (ssDCval >= 37) {
                                                                if (SC < 146) {
                                                                    return 8;
                                                                }
                                                                else if (SC >= 146) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (Rs >= 401) {
                                                        if (ssDCval < 38) {
                                                            if (Cs < 366) {
                                                                if (tcor < 79) {
                                                                    if (Rs < 426) {
                                                                        return 8;
                                                                    }
                                                                    else if (Rs >= 426) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (tcor >= 79) {
                                                                    if (Rs < 451) {
                                                                        return 1;
                                                                    }
                                                                    else if (Rs >= 451) {
                                                                        return 2;
                                                                    }
                                                                }
                                                            }
                                                            else if (Cs >= 366) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (ssDCval >= 38) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (Cs >= 385) {
                                                if (tcor < 87) {
                                                    if (posBalance < 54) {
                                                        return 4;
                                                    }
                                                    else if (posBalance >= 54) {
                                                        if (tcor < 83) {
                                                            if (SC < 862) {
                                                                return 4;
                                                            }
                                                            else if (SC >= 862) {
                                                                return 2;
                                                            }
                                                        }
                                                        else if (tcor >= 83) {
                                                            return 2;
                                                        }
                                                    }
                                                }
                                                else if (tcor >= 87) {
                                                    if (Rs < 245) {
                                                        return 2;
                                                    }
                                                    else if (Rs >= 245) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (AFD >= 5) {
                                        if (refDCval < 85) {
                                            if (Cs < 385) {
                                                if (mu_mv_mag_sq < 33) {
                                                    if (diffRsCsdiff < -1) {
                                                        return 2;
                                                    }
                                                    else if (diffRsCsdiff >= -1) {
                                                        if (Rs < 69) {
                                                            return 1;
                                                        }
                                                        else if (Rs >= 69) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                                else if (mu_mv_mag_sq >= 33) {
                                                    if (tcor < 81) {
                                                        if (TSCindex < 6) {
                                                            if (negBalance < 22) {
                                                                return 4;
                                                            }
                                                            else if (negBalance >= 22) {
                                                                if (Cs < 320) {
                                                                    if (negBalance < 45) {
                                                                        return 1;
                                                                    }
                                                                    else if (negBalance >= 45) {
                                                                        if (Cs < 241) {
                                                                            return 1;
                                                                        }
                                                                        else if (Cs >= 241) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                }
                                                                else if (Cs >= 320) {
                                                                    if (Cs < 330) {
                                                                        return 4;
                                                                    }
                                                                    else if (Cs >= 330) {
                                                                        if (MVDiff < 17) {
                                                                            return 4;
                                                                        }
                                                                        else if (MVDiff >= 17) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (TSCindex >= 6) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (tcor >= 81) {
                                                        if (Rs < 508) {
                                                            if (AFD < 6) {
                                                                return 4;
                                                            }
                                                            else if (AFD >= 6) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (Rs >= 508) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (Cs >= 385) {
                                                if (refDCval < 39) {
                                                    if (mu_mv_mag_sq < 45) {
                                                        if (SC < 861) {
                                                            return 4;
                                                        }
                                                        else if (SC >= 861) {
                                                            if (Rs < 501) {
                                                                return 2;
                                                            }
                                                            else if (Rs >= 501) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                    else if (mu_mv_mag_sq >= 45) {
                                                        return 2;
                                                    }
                                                }
                                                else if (refDCval >= 39) {
                                                    if (Rs < 453) {
                                                        if (SC < 762) {
                                                            return 1;
                                                        }
                                                        else if (SC >= 762) {
                                                            if (TSC < 7) {
                                                                return 8;
                                                            }
                                                            else if (TSC >= 7) {
                                                                return 2;
                                                            }
                                                        }
                                                    }
                                                    else if (Rs >= 453) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                        }
                                        else if (refDCval >= 85) {
                                            if (Rs < 354) {
                                                return 2;
                                            }
                                            else if (Rs >= 354) {
                                                return 1;
                                            }
                                        }
                                    }
                                }
                                else if (SC >= 1091) {
                                    if (Cs < 1677) {
                                        if (gchDC < 2) {
                                            if (SC < 3344) {
                                                if (ssDCval < 39) {
                                                    if (TSCindex < 6) {
                                                        if (mu_mv_mag_sq < 76) {
                                                            if (tcor < 71) {
                                                                return 8;
                                                            }
                                                            else if (tcor >= 71) {
                                                                if (tcor < 96) {
                                                                    if (diffAFD < -2) {
                                                                        return 1;
                                                                    }
                                                                    else if (diffAFD >= -2) {
                                                                        return 2;
                                                                    }
                                                                }
                                                                else if (tcor >= 96) {
                                                                    if (Rs < 1024) {
                                                                        return 1;
                                                                    }
                                                                    else if (Rs >= 1024) {
                                                                        return 8;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (mu_mv_mag_sq >= 76) {
                                                            if (RsDiff < 2) {
                                                                if (Rs < 1060) {
                                                                    if (negBalance < 45) {
                                                                        return 4;
                                                                    }
                                                                    else if (negBalance >= 45) {
                                                                        return 8;
                                                                    }
                                                                }
                                                                else if (Rs >= 1060) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (RsDiff >= 2) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                    else if (TSCindex >= 6) {
                                                        if (TSC < 6) {
                                                            if (Rs < 855) {
                                                                if (mu_mv_mag_sq < 53) {
                                                                    if (tcor < 78) {
                                                                        if (RsCsDiff < 7) {
                                                                            return 4;
                                                                        }
                                                                        else if (RsCsDiff >= 7) {
                                                                            if (Rs < 755) {
                                                                                return 4;
                                                                            }
                                                                            else if (Rs >= 755) {
                                                                                return 1;
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (tcor >= 78) {
                                                                        if (mcTcor < 80) {
                                                                            return 4;
                                                                        }
                                                                        else if (mcTcor >= 80) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                }
                                                                else if (mu_mv_mag_sq >= 53) {
                                                                    if (posBalance < 58) {
                                                                        return 4;
                                                                    }
                                                                    else if (posBalance >= 58) {
                                                                        return 1;
                                                                    }
                                                                }
                                                            }
                                                            else if (Rs >= 855) {
                                                                if (Rs < 1030) {
                                                                    return 1;
                                                                }
                                                                else if (Rs >= 1030) {
                                                                    return 2;
                                                                }
                                                            }
                                                        }
                                                        else if (TSC >= 6) {
                                                            if (negBalance < 49) {
                                                                return 4;
                                                            }
                                                            else if (negBalance >= 49) {
                                                                return 1;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (ssDCval >= 39) {
                                                    if (Cs < 1232) {
                                                        if (Rs < 1221) {
                                                            if (Rs < 699) {
                                                                if (CsDiff < 2) {
                                                                    return 8;
                                                                }
                                                                else if (CsDiff >= 2) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (Rs >= 699) {
                                                                if (Cs < 836) {
                                                                    if (refDCval < 56) {
                                                                        if (negBalance < 44) {
                                                                            if (tcor < 90) {
                                                                                if (Cs < 825) {
                                                                                    if (diffMVdiffVal < 6) {
                                                                                        if (diffMVdiffVal < -3) {
                                                                                            if (Rs < 1100) {
                                                                                                return 1;
                                                                                            }
                                                                                            else if (Rs >= 1100) {
                                                                                                return 8;
                                                                                            }
                                                                                        }
                                                                                        else if (diffMVdiffVal >= -3) {
                                                                                            return 4;
                                                                                        }
                                                                                    }
                                                                                    else if (diffMVdiffVal >= 6) {
                                                                                        return 8;
                                                                                    }
                                                                                }
                                                                                else if (Cs >= 825) {
                                                                                    return 1;
                                                                                }
                                                                            }
                                                                            else if (tcor >= 90) {
                                                                                if (AFD < 4) {
                                                                                    return 1;
                                                                                }
                                                                                else if (AFD >= 4) {
                                                                                    if (MVDiff < 7) {
                                                                                        if (Cs < 798) {
                                                                                            return 4;
                                                                                        }
                                                                                        else if (Cs >= 798) {
                                                                                            return 1;
                                                                                        }
                                                                                    }
                                                                                    else if (MVDiff >= 7) {
                                                                                        return 4;
                                                                                    }
                                                                                }
                                                                            }
                                                                        }
                                                                        else if (negBalance >= 44) {
                                                                            if (Scindex < 4) {
                                                                                if (mu_mv_mag_sq < 47) {
                                                                                    if (mu_mv_mag_sq < 33) {
                                                                                        if (Rs < 827) {
                                                                                            if (Rs < 724) {
                                                                                                return 4;
                                                                                            }
                                                                                            else if (Rs >= 724) {
                                                                                                if (Rs < 744) {
                                                                                                    return 1;
                                                                                                }
                                                                                                else if (Rs >= 744) {
                                                                                                    if (diffAFD < -1) {
                                                                                                        return 4;
                                                                                                    }
                                                                                                    else if (diffAFD >= -1) {
                                                                                                        if (SC < 1334) {
                                                                                                            return 4;
                                                                                                        }
                                                                                                        else if (SC >= 1334) {
                                                                                                            return 1;
                                                                                                        }
                                                                                                    }
                                                                                                }
                                                                                            }
                                                                                        }
                                                                                        else if (Rs >= 827) {
                                                                                            return 4;
                                                                                        }
                                                                                    }
                                                                                    else if (mu_mv_mag_sq >= 33) {
                                                                                        if (TSC < 4) {
                                                                                            return 4;
                                                                                        }
                                                                                        else if (TSC >= 4) {
                                                                                            if (MVDiff < 44) {
                                                                                                if (RsCsDiff < 4) {
                                                                                                    if (diffMVdiffVal < -3) {
                                                                                                        return 4;
                                                                                                    }
                                                                                                    else if (diffMVdiffVal >= -3) {
                                                                                                        return 1;
                                                                                                    }
                                                                                                }
                                                                                                else if (RsCsDiff >= 4) {
                                                                                                    return 1;
                                                                                                }
                                                                                            }
                                                                                            else if (MVDiff >= 44) {
                                                                                                if (Rs < 875) {
                                                                                                    return 4;
                                                                                                }
                                                                                                else if (Rs >= 875) {
                                                                                                    return 1;
                                                                                                }
                                                                                            }
                                                                                        }
                                                                                    }
                                                                                }
                                                                                else if (mu_mv_mag_sq >= 47) {
                                                                                    return 4;
                                                                                }
                                                                            }
                                                                            else if (Scindex >= 4) {
                                                                                if (Cs < 607) {
                                                                                    return 8;
                                                                                }
                                                                                else if (Cs >= 607) {
                                                                                    if (RsCsDiff < 22) {
                                                                                        return 1;
                                                                                    }
                                                                                    else if (RsCsDiff >= 22) {
                                                                                        return 4;
                                                                                    }
                                                                                }
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (refDCval >= 56) {
                                                                        if (ssDCval < 63) {
                                                                            if (Rs < 938) {
                                                                                return 8;
                                                                            }
                                                                            else if (Rs >= 938) {
                                                                                if (Cs < 705) {
                                                                                    return 8;
                                                                                }
                                                                                else if (Cs >= 705) {
                                                                                    return 1;
                                                                                }
                                                                            }
                                                                        }
                                                                        else if (ssDCval >= 63) {
                                                                            if (Cs < 441) {
                                                                                return 1;
                                                                            }
                                                                            else if (Cs >= 441) {
                                                                                if (Rs < 996) {
                                                                                    return 2;
                                                                                }
                                                                                else if (Rs >= 996) {
                                                                                    return 1;
                                                                                }
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                                else if (Cs >= 836) {
                                                                    if (ssDCval < 49) {
                                                                        if (SC < 1906) {
                                                                            return 2;
                                                                        }
                                                                        else if (SC >= 1906) {
                                                                            if (Cs < 1012) {
                                                                                return 1;
                                                                            }
                                                                            else if (Cs >= 1012) {
                                                                                return 8;
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (ssDCval >= 49) {
                                                                        if (Rs < 1136) {
                                                                            return 4;
                                                                        }
                                                                        else if (Rs >= 1136) {
                                                                            if (Rs < 1187) {
                                                                                if (AFD < 9) {
                                                                                    return 1;
                                                                                }
                                                                                else if (AFD >= 9) {
                                                                                    return 4;
                                                                                }
                                                                            }
                                                                            else if (Rs >= 1187) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (Rs >= 1221) {
                                                            if (Cs < 609) {
                                                                if (negBalance < 36) {
                                                                    return 4;
                                                                }
                                                                else if (negBalance >= 36) {
                                                                    return 2;
                                                                }
                                                            }
                                                            else if (Cs >= 609) {
                                                                if (tcor < 86) {
                                                                    if (ssDCval < 54) {
                                                                        if (mcTcor < 88) {
                                                                            if (diffRsCsdiff < 17) {
                                                                                return 1;
                                                                            }
                                                                            else if (diffRsCsdiff >= 17) {
                                                                                if (Rs < 1363) {
                                                                                    return 8;
                                                                                }
                                                                                else if (Rs >= 1363) {
                                                                                    return 4;
                                                                                }
                                                                            }
                                                                        }
                                                                        else if (mcTcor >= 88) {
                                                                            if (mcTcor < 90) {
                                                                                return 8;
                                                                            }
                                                                            else if (mcTcor >= 90) {
                                                                                if (Rs < 1432) {
                                                                                    return 4;
                                                                                }
                                                                                else if (Rs >= 1432) {
                                                                                    return 1;
                                                                                }
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (ssDCval >= 54) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (tcor >= 86) {
                                                                    if (RsCsDiff < 12) {
                                                                        if (refDCval < 58) {
                                                                            if (mcTcor < 92) {
                                                                                if (ssDCval < 51) {
                                                                                    return 8;
                                                                                }
                                                                                else if (ssDCval >= 51) {
                                                                                    return 1;
                                                                                }
                                                                            }
                                                                            else if (mcTcor >= 92) {
                                                                                if (negBalance < 49) {
                                                                                    return 1;
                                                                                }
                                                                                else if (negBalance >= 49) {
                                                                                    return 4;
                                                                                }
                                                                            }
                                                                        }
                                                                        else if (refDCval >= 58) {
                                                                            if (TSC < 5) {
                                                                                return 8;
                                                                            }
                                                                            else if (TSC >= 5) {
                                                                                return 1;
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (RsCsDiff >= 12) {
                                                                        if (posBalance < 57) {
                                                                            if (ssDCval < 55) {
                                                                                if (CsDiff < 3) {
                                                                                    if (Rs < 1342) {
                                                                                        if (Rs < 1237) {
                                                                                            return 4;
                                                                                        }
                                                                                        else if (Rs >= 1237) {
                                                                                            return 1;
                                                                                        }
                                                                                    }
                                                                                    else if (Rs >= 1342) {
                                                                                        if (Cs < 974) {
                                                                                            return 4;
                                                                                        }
                                                                                        else if (Cs >= 974) {
                                                                                            return 1;
                                                                                        }
                                                                                    }
                                                                                }
                                                                                else if (CsDiff >= 3) {
                                                                                    return 1;
                                                                                }
                                                                            }
                                                                            else if (ssDCval >= 55) {
                                                                                if (SC < 3010) {
                                                                                    return 4;
                                                                                }
                                                                                else if (SC >= 3010) {
                                                                                    return 1;
                                                                                }
                                                                            }
                                                                        }
                                                                        else if (posBalance >= 57) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (Cs >= 1232) {
                                                        if (tcor < 83) {
                                                            return 2;
                                                        }
                                                        else if (tcor >= 83) {
                                                            if (diffRsCsdiff < -2) {
                                                                return 4;
                                                            }
                                                            else if (diffRsCsdiff >= -2) {
                                                                if (Rs < 1729) {
                                                                    return 8;
                                                                }
                                                                else if (Rs >= 1729) {
                                                                    if (MVDiff < 20) {
                                                                        return 4;
                                                                    }
                                                                    else if (MVDiff >= 20) {
                                                                        if (posBalance < 56) {
                                                                            return 8;
                                                                        }
                                                                        else if (posBalance >= 56) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                            else if (SC >= 3344) {
                                                if (mu_mv_mag_sq < 63) {
                                                    if (Cs < 1545) {
                                                        if (Cs < 1525) {
                                                            return 4;
                                                        }
                                                        else if (Cs >= 1525) {
                                                            if (negBalance < 46) {
                                                                return 1;
                                                            }
                                                            else if (negBalance >= 46) {
                                                                if (Rs < 2163) {
                                                                    return 4;
                                                                }
                                                                else if (Rs >= 2163) {
                                                                    return 1;
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (Cs >= 1545) {
                                                        if (diffRsCsdiff < -2) {
                                                            if (TSC < 5) {
                                                                return 1;
                                                            }
                                                            else if (TSC >= 5) {
                                                                if (Rs < 2184) {
                                                                    return 4;
                                                                }
                                                                else if (Rs >= 2184) {
                                                                    return 1;
                                                                }
                                                            }
                                                        }
                                                        else if (diffRsCsdiff >= -2) {
                                                            return 1;
                                                        }
                                                    }
                                                }
                                                else if (mu_mv_mag_sq >= 63) {
                                                    if (Cs < 1559) {
                                                        return 4;
                                                    }
                                                    else if (Cs >= 1559) {
                                                        if (mu_mv_mag_sq < 72) {
                                                            return 1;
                                                        }
                                                        else if (mu_mv_mag_sq >= 72) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (gchDC >= 2) {
                                            if (diffRsCsdiff < 12) {
                                                if (posBalance < 94) {
                                                    if (Cs < 661) {
                                                        return 8;
                                                    }
                                                    else if (Cs >= 661) {
                                                        if (refDCval < 53) {
                                                            if (MVDiff < 42) {
                                                                if (MVDiff < 19) {
                                                                    if (RsCsDiff < 13) {
                                                                        if (Cs < 678) {
                                                                            return 4;
                                                                        }
                                                                        else if (Cs >= 678) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                    else if (RsCsDiff >= 13) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (MVDiff >= 19) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (MVDiff >= 42) {
                                                                return 2;
                                                            }
                                                        }
                                                        else if (refDCval >= 53) {
                                                            if (mu_mv_mag_sq < 48) {
                                                                if (gchDC < 4) {
                                                                    if (Cs < 776) {
                                                                        return 8;
                                                                    }
                                                                    else if (Cs >= 776) {
                                                                        return 2;
                                                                    }
                                                                }
                                                                else if (gchDC >= 4) {
                                                                    return 1;
                                                                }
                                                            }
                                                            else if (mu_mv_mag_sq >= 48) {
                                                                if (mu_mv_mag_sq < 82) {
                                                                    return 4;
                                                                }
                                                                else if (mu_mv_mag_sq >= 82) {
                                                                    if (SC < 2074) {
                                                                        return 2;
                                                                    }
                                                                    else if (SC >= 2074) {
                                                                        return 4;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (posBalance >= 94) {
                                                    return 4;
                                                }
                                            }
                                            else if (diffRsCsdiff >= 12) {
                                                if (posBalance < 62) {
                                                    if (MVDiff < 8) {
                                                        return 2;
                                                    }
                                                    else if (MVDiff >= 8) {
                                                        return 4;
                                                    }
                                                }
                                                else if (posBalance >= 62) {
                                                    if (posBalance < 77) {
                                                        return 1;
                                                    }
                                                    else if (posBalance >= 77) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (Cs >= 1677) {
                                        if (refDCval < 82) {
                                            if (mu_mv_mag_sq < 40) {
                                                if (mcTcor < 95) {
                                                    if (ssDCval < 54) {
                                                        if (diffMVdiffVal < -9) {
                                                            if (diffMVdiffVal < -18) {
                                                                return 8;
                                                            }
                                                            else if (diffMVdiffVal >= -18) {
                                                                return 2;
                                                            }
                                                        }
                                                        else if (diffMVdiffVal >= -9) {
                                                            if (negBalance < 52) {
                                                                return 8;
                                                            }
                                                            else if (negBalance >= 52) {
                                                                return 2;
                                                            }
                                                        }
                                                    }
                                                    else if (ssDCval >= 54) {
                                                        return 2;
                                                    }
                                                }
                                                else if (mcTcor >= 95) {
                                                    if (MVDiff < 84) {
                                                        if (negBalance < 51) {
                                                            if (diffAFD < -1) {
                                                                return 8;
                                                            }
                                                            else if (diffAFD >= -1) {
                                                                if (diffMVdiffVal < -5) {
                                                                    return 8;
                                                                }
                                                                else if (diffMVdiffVal >= -5) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                        else if (negBalance >= 51) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (MVDiff >= 84) {
                                                        return 2;
                                                    }
                                                }
                                            }
                                            else if (mu_mv_mag_sq >= 40) {
                                                if (ssDCval < 52) {
                                                    if (Rs < 2117) {
                                                        return 8;
                                                    }
                                                    else if (Rs >= 2117) {
                                                        return 4;
                                                    }
                                                }
                                                else if (ssDCval >= 52) {
                                                    if (negBalance < 49) {
                                                        return 8;
                                                    }
                                                    else if (negBalance >= 49) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                        }
                                        else if (refDCval >= 82) {
                                            if (Rs < 2493) {
                                                if (posBalance < 56) {
                                                    if (refDCval < 87) {
                                                        return 1;
                                                    }
                                                    else if (refDCval >= 87) {
                                                        return 8;
                                                    }
                                                }
                                                else if (posBalance >= 56) {
                                                    if (refDCval < 93) {
                                                        if (Rs < 1548) {
                                                            return 4;
                                                        }
                                                        else if (Rs >= 1548) {
                                                            if (AFD < 10) {
                                                                return 8;
                                                            }
                                                            else if (AFD >= 10) {
                                                                return 2;
                                                            }
                                                        }
                                                    }
                                                    else if (refDCval >= 93) {
                                                        return 2;
                                                    }
                                                }
                                            }
                                            else if (Rs >= 2493) {
                                                if (Cs < 2218) {
                                                    return 1;
                                                }
                                                else if (Cs >= 2218) {
                                                    if (SC < 5369) {
                                                        if (RsDiff < 10) {
                                                            return 2;
                                                        }
                                                        else if (RsDiff >= 10) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (SC >= 5369) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                            else if (mu_mv_mag_sq >= 90) {
                                if (SC < 1786) {
                                    if (mu_mv_mag_sq < 103) {
                                        if (Cs < 613) {
                                            if (diffRsCsdiff < -1) {
                                                return 8;
                                            }
                                            else if (diffRsCsdiff >= -1) {
                                                return 4;
                                            }
                                        }
                                        else if (Cs >= 613) {
                                            if (Rs < 844) {
                                                return 4;
                                            }
                                            else if (Rs >= 844) {
                                                return 8;
                                            }
                                        }
                                    }
                                    else if (mu_mv_mag_sq >= 103) {
                                        if (diffRsCsdiff < -4) {
                                            if (TSCindex < 1) {
                                                return 4;
                                            }
                                            else if (TSCindex >= 1) {
                                                return 8;
                                            }
                                        }
                                        else if (diffRsCsdiff >= -4) {
                                            if (posBalance < 55) {
                                                if (Cs < 1) {
                                                    if (refDCval < 20) {
                                                        return 8;
                                                    }
                                                    else if (refDCval >= 20) {
                                                        return 4;
                                                    }
                                                }
                                                else if (Cs >= 1) {
                                                    if (Rs < 941) {
                                                        return 1;
                                                    }
                                                    else if (Rs >= 941) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                            else if (posBalance >= 55) {
                                                if (Rs < 162) {
                                                    return 2;
                                                }
                                                else if (Rs >= 162) {
                                                    return 8;
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (SC >= 1786) {
                                    if (ssDCval < 92) {
                                        if (RsCsDiff < 109) {
                                            if (RsCsDiff < 4) {
                                                if (mu_mv_mag_sq < 98) {
                                                    if (Cs < 1114) {
                                                        return 1;
                                                    }
                                                    else if (Cs >= 1114) {
                                                        return 4;
                                                    }
                                                }
                                                else if (mu_mv_mag_sq >= 98) {
                                                    if (posBalance < 20) {
                                                        if (SC < 2207) {
                                                            return 4;
                                                        }
                                                        else if (SC >= 2207) {
                                                            if (MVDiff < 86) {
                                                                return 8;
                                                            }
                                                            else if (MVDiff >= 86) {
                                                                return 2;
                                                            }
                                                        }
                                                    }
                                                    else if (posBalance >= 20) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                            else if (RsCsDiff >= 4) {
                                                if (diffAFD < -32) {
                                                    return 4;
                                                }
                                                else if (diffAFD >= -32) {
                                                    if (RsDiff < 6) {
                                                        if (gchDC < 1) {
                                                            return 8;
                                                        }
                                                        else if (gchDC >= 1) {
                                                            if (CsDiff < 2) {
                                                                if (Rs < 1520) {
                                                                    return 1;
                                                                }
                                                                else if (Rs >= 1520) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (CsDiff >= 2) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                    else if (RsDiff >= 6) {
                                                        if (RsCsDiff < 57) {
                                                            return 4;
                                                        }
                                                        else if (RsCsDiff >= 57) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (RsCsDiff >= 109) {
                                            if (RsDiff < 12) {
                                                return 4;
                                            }
                                            else if (RsDiff >= 12) {
                                                return 8;
                                            }
                                        }
                                    }
                                    else if (ssDCval >= 92) {
                                        if (RsCsDiff < 51) {
                                            return 8;
                                        }
                                        else if (RsCsDiff >= 51) {
                                            if (Rs < 2433) {
                                                return 2;
                                            }
                                            else if (Rs >= 2433) {
                                                return 8;
                                            }
                                        }
                                    }
                                }
                            }
                        }
                        else if (refDCval >= 94) {
                            if (refDCval < 116) {
                                if (RsDiff < 6) {
                                    if (Rs < 343) {
                                        if (MVDiff < 17) {
                                            return 4;
                                        }
                                        else if (MVDiff >= 17) {
                                            if (refDCval < 110) {
                                                return 8;
                                            }
                                            else if (refDCval >= 110) {
                                                return 4;
                                            }
                                        }
                                    }
                                    else if (Rs >= 343) {
                                        if (mu_mv_mag_sq < 45) {
                                            if (RsDiff < 4) {
                                                if (RsDiff < 1) {
                                                    return 1;
                                                }
                                                else if (RsDiff >= 1) {
                                                    return 8;
                                                }
                                            }
                                            else if (RsDiff >= 4) {
                                                if (ssDCval < 98) {
                                                    return 8;
                                                }
                                                else if (ssDCval >= 98) {
                                                    if (Cs < 1619) {
                                                        if (AFD < 11) {
                                                            return 4;
                                                        }
                                                        else if (AFD >= 11) {
                                                            return 2;
                                                        }
                                                    }
                                                    else if (Cs >= 1619) {
                                                        if (Cs < 1807) {
                                                            if (Rs < 3737) {
                                                                return 8;
                                                            }
                                                            else if (Rs >= 3737) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (Cs >= 1807) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (mu_mv_mag_sq >= 45) {
                                            if (diffRsCsdiff < -42) {
                                                if (diffAFD < -53) {
                                                    return 4;
                                                }
                                                else if (diffAFD >= -53) {
                                                    if (AFD < 6) {
                                                        return 8;
                                                    }
                                                    else if (AFD >= 6) {
                                                        return 2;
                                                    }
                                                }
                                            }
                                            else if (diffRsCsdiff >= -42) {
                                                if (refDCval < 99) {
                                                    if (TSCindex < 9) {
                                                        return 8;
                                                    }
                                                    else if (TSCindex >= 9) {
                                                        if (diffRsCsdiff < -9) {
                                                            return 2;
                                                        }
                                                        else if (diffRsCsdiff >= -9) {
                                                            if (mcTcor < 86) {
                                                                return 8;
                                                            }
                                                            else if (mcTcor >= 86) {
                                                                return 2;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (refDCval >= 99) {
                                                    if (MVDiff < 26) {
                                                        if (mu_mv_mag_sq < 62) {
                                                            if (AFD < 8) {
                                                                if (posBalance < 45) {
                                                                    if (ssDCval < 107) {
                                                                        return 8;
                                                                    }
                                                                    else if (ssDCval >= 107) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (posBalance >= 45) {
                                                                    if (Rs < 581) {
                                                                        return 4;
                                                                    }
                                                                    else if (Rs >= 581) {
                                                                        return 8;
                                                                    }
                                                                }
                                                            }
                                                            else if (AFD >= 8) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (mu_mv_mag_sq >= 62) {
                                                            if (Rs < 506) {
                                                                return 4;
                                                            }
                                                            else if (Rs >= 506) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                    else if (MVDiff >= 26) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (RsDiff >= 6) {
                                    if (refDCval < 98) {
                                        if (negBalance < 52) {
                                            if (AFD < 11) {
                                                return 8;
                                            }
                                            else if (AFD >= 11) {
                                                if (Rs < 2571) {
                                                    if (Cs < 2120) {
                                                        return 4;
                                                    }
                                                    else if (Cs >= 2120) {
                                                        return 8;
                                                    }
                                                }
                                                else if (Rs >= 2571) {
                                                    if (Rs < 2768) {
                                                        return 1;
                                                    }
                                                    else if (Rs >= 2768) {
                                                        return 2;
                                                    }
                                                }
                                            }
                                        }
                                        else if (negBalance >= 52) {
                                            if (TSCindex < 9) {
                                                if (diffRsCsdiff < 6) {
                                                    return 2;
                                                }
                                                else if (diffRsCsdiff >= 6) {
                                                    if (Rs < 2024) {
                                                        return 8;
                                                    }
                                                    else if (Rs >= 2024) {
                                                        return 2;
                                                    }
                                                }
                                            }
                                            else if (TSCindex >= 9) {
                                                return 2;
                                            }
                                        }
                                    }
                                    else if (refDCval >= 98) {
                                        if (Cs < 2703) {
                                            if (mcTcor < 96) {
                                                if (refDCval < 107) {
                                                    if (diffAFD < -3) {
                                                        return 1;
                                                    }
                                                    else if (diffAFD >= -3) {
                                                        if (Rs < 2139) {
                                                            return 4;
                                                        }
                                                        else if (Rs >= 2139) {
                                                            if (RsDiff < 8) {
                                                                if (Scindex < 6) {
                                                                    if (gchDC < 1) {
                                                                        if (Rs < 3038) {
                                                                            return 4;
                                                                        }
                                                                        else if (Rs >= 3038) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                    else if (gchDC >= 1) {
                                                                        if (Rs < 3111) {
                                                                            return 8;
                                                                        }
                                                                        else if (Rs >= 3111) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                }
                                                                else if (Scindex >= 6) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (RsDiff >= 8) {
                                                                if (negBalance < 35) {
                                                                    return 4;
                                                                }
                                                                else if (negBalance >= 35) {
                                                                    if (TSC < 7) {
                                                                        return 8;
                                                                    }
                                                                    else if (TSC >= 7) {
                                                                        if (Rs < 3030) {
                                                                            return 8;
                                                                        }
                                                                        else if (Rs >= 3030) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (refDCval >= 107) {
                                                    if (mu_mv_mag_sq < 72) {
                                                        if (gchDC < 3) {
                                                            if (Cs < 2592) {
                                                                if (diffRsCsdiff < -14) {
                                                                    if (tcor < 93) {
                                                                        return 4;
                                                                    }
                                                                    else if (tcor >= 93) {
                                                                        return 2;
                                                                    }
                                                                }
                                                                else if (diffRsCsdiff >= -14) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (Cs >= 2592) {
                                                                if (Cs < 2646) {
                                                                    return 8;
                                                                }
                                                                else if (Cs >= 2646) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                        else if (gchDC >= 3) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (mu_mv_mag_sq >= 72) {
                                                        if (RsCsDiff < 73) {
                                                            return 8;
                                                        }
                                                        else if (RsCsDiff >= 73) {
                                                            if (posBalance < 42) {
                                                                if (diffRsCsdiff < 24) {
                                                                    return 2;
                                                                }
                                                                else if (diffRsCsdiff >= 24) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (posBalance >= 42) {
                                                                if (Cs < 1584) {
                                                                    return 8;
                                                                }
                                                                else if (Cs >= 1584) {
                                                                    if (Rs < 4573) {
                                                                        return 4;
                                                                    }
                                                                    else if (Rs >= 4573) {
                                                                        return 2;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                            else if (mcTcor >= 96) {
                                                if (AFD < 8) {
                                                    if (AFD < 6) {
                                                        return 8;
                                                    }
                                                    else if (AFD >= 6) {
                                                        return 4;
                                                    }
                                                }
                                                else if (AFD >= 8) {
                                                    if (ssDCval < 101) {
                                                        return 4;
                                                    }
                                                    else if (ssDCval >= 101) {
                                                        if (Cs < 2015) {
                                                            return 4;
                                                        }
                                                        else if (Cs >= 2015) {
                                                            if (Rs < 4156) {
                                                                return 2;
                                                            }
                                                            else if (Rs >= 4156) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (Cs >= 2703) {
                                            if (Rs < 4683) {
                                                if (Rs < 4135) {
                                                    return 8;
                                                }
                                                else if (Rs >= 4135) {
                                                    if (refDCval < 109) {
                                                        if (Rs < 4405) {
                                                            return 1;
                                                        }
                                                        else if (Rs >= 4405) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (refDCval >= 109) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                            else if (Rs >= 4683) {
                                                if (ssDCval < 112) {
                                                    if (mcTcor < 86) {
                                                        return 2;
                                                    }
                                                    else if (mcTcor >= 86) {
                                                        if (mu_mv_mag_sq < 74) {
                                                            return 4;
                                                        }
                                                        else if (mu_mv_mag_sq >= 74) {
                                                            if (SC < 7990) {
                                                                return 2;
                                                            }
                                                            else if (SC >= 7990) {
                                                                if (posBalance < 48) {
                                                                    if (diffRsCsdiff < 20) {
                                                                        return 8;
                                                                    }
                                                                    else if (diffRsCsdiff >= 20) {
                                                                        return 2;
                                                                    }
                                                                }
                                                                else if (posBalance >= 48) {
                                                                    return 2;
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (ssDCval >= 112) {
                                                    return 8;
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                            else if (refDCval >= 116) {
                                if (ssDCval < 128) {
                                    if (negBalance < 40) {
                                        if (ssDCval < 119) {
                                            if (Rs < 2892) {
                                                return 8;
                                            }
                                            else if (Rs >= 2892) {
                                                if (SC < 7146) {
                                                    return 2;
                                                }
                                                else if (SC >= 7146) {
                                                    return 8;
                                                }
                                            }
                                        }
                                        else if (ssDCval >= 119) {
                                            if (SC < 7200) {
                                                if (SC < 4990) {
                                                    if (RsCsDiff < 51) {
                                                        if (Rs < 347) {
                                                            return 2;
                                                        }
                                                        else if (Rs >= 347) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (RsCsDiff >= 51) {
                                                        return 1;
                                                    }
                                                }
                                                else if (SC >= 4990) {
                                                    if (CsDiff < 5) {
                                                        return 8;
                                                    }
                                                    else if (CsDiff >= 5) {
                                                        if (ssDCval < 121) {
                                                            if (AFD < 8) {
                                                                return 4;
                                                            }
                                                            else if (AFD >= 8) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (ssDCval >= 121) {
                                                            if (diffTSC < 2) {
                                                                if (MVDiff < 7) {
                                                                    if (Rs < 4097) {
                                                                        return 8;
                                                                    }
                                                                    else if (Rs >= 4097) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (MVDiff >= 7) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (diffTSC >= 2) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                            else if (SC >= 7200) {
                                                if (refDCval < 126) {
                                                    if (Rs < 5101) {
                                                        return 8;
                                                    }
                                                    else if (Rs >= 5101) {
                                                        return 4;
                                                    }
                                                }
                                                else if (refDCval >= 126) {
                                                    if (AFD < 9) {
                                                        return 2;
                                                    }
                                                    else if (AFD >= 9) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (negBalance >= 40) {
                                        if (SC < 4461) {
                                            if (MVDiff < 8) {
                                                if (gchDC < 1) {
                                                    if (RsCsDiff < 107) {
                                                        if (AFD < 10) {
                                                            return 4;
                                                        }
                                                        else if (AFD >= 10) {
                                                            if (SC < 4339) {
                                                                return 4;
                                                            }
                                                            else if (SC >= 4339) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                    else if (RsCsDiff >= 107) {
                                                        return 8;
                                                    }
                                                }
                                                else if (gchDC >= 1) {
                                                    return 8;
                                                }
                                            }
                                            else if (MVDiff >= 8) {
                                                if (Cs < 1619) {
                                                    if (diffTSC < -1) {
                                                        return 8;
                                                    }
                                                    else if (diffTSC >= -1) {
                                                        if (Rs < 1182) {
                                                            return 8;
                                                        }
                                                        else if (Rs >= 1182) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                                else if (Cs >= 1619) {
                                                    return 4;
                                                }
                                            }
                                        }
                                        else if (SC >= 4461) {
                                            if (mu_mv_mag_sq < 62) {
                                                return 4;
                                            }
                                            else if (mu_mv_mag_sq >= 62) {
                                                if (ssDCval < 117) {
                                                    return 4;
                                                }
                                                else if (ssDCval >= 117) {
                                                    if (Rs < 2861) {
                                                        return 4;
                                                    }
                                                    else if (Rs >= 2861) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (ssDCval >= 128) {
                                    if (SC < 6714) {
                                        if (ssDCval < 141) {
                                            if (SC < 2796) {
                                                if (Cs < 626) {
                                                    return 8;
                                                }
                                                else if (Cs >= 626) {
                                                    return 1;
                                                }
                                            }
                                            else if (SC >= 2796) {
                                                if (posBalance < 47) {
                                                    if (RsCsDiff < 22) {
                                                        return 8;
                                                    }
                                                    else if (RsCsDiff >= 22) {
                                                        return 4;
                                                    }
                                                }
                                                else if (posBalance >= 47) {
                                                    if (Cs < 2902) {
                                                        return 1;
                                                    }
                                                    else if (Cs >= 2902) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                        }
                                        else if (ssDCval >= 141) {
                                            if (diffTSC < -1) {
                                                if (MVDiff < 20) {
                                                    if (CsDiff < 1) {
                                                        return 2;
                                                    }
                                                    else if (CsDiff >= 1) {
                                                        return 8;
                                                    }
                                                }
                                                else if (MVDiff >= 20) {
                                                    if (Rs < 370) {
                                                        return 1;
                                                    }
                                                    else if (Rs >= 370) {
                                                        if (Cs < 374) {
                                                            return 8;
                                                        }
                                                        else if (Cs >= 374) {
                                                            if (AFD < 9) {
                                                                return 4;
                                                            }
                                                            else if (AFD >= 9) {
                                                                return 2;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                            else if (diffTSC >= -1) {
                                                if (posBalance < 38) {
                                                    if (RsCsDiff < 4) {
                                                        return 4;
                                                    }
                                                    else if (RsCsDiff >= 4) {
                                                        return 8;
                                                    }
                                                }
                                                else if (posBalance >= 38) {
                                                    if (tcor < 89) {
                                                        return 4;
                                                    }
                                                    else if (tcor >= 89) {
                                                        if (ssDCval < 161) {
                                                            if (AFD < 7) {
                                                                if (Rs < 797) {
                                                                    return 4;
                                                                }
                                                                else if (Rs >= 797) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (AFD >= 7) {
                                                                return 2;
                                                            }
                                                        }
                                                        else if (ssDCval >= 161) {
                                                            return 2;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (SC >= 6714) {
                                        if (TSC < 9) {
                                            return 2;
                                        }
                                        else if (TSC >= 9) {
                                            if (SC < 7279) {
                                                return 2;
                                            }
                                            else if (SC >= 7279) {
                                                return 4;
                                            }
                                        }
                                    }
                                }
                            }
                        }
                    }
                    else if (AFD >= 12) {
                        if (AFD < 17) {
                            if (refDCval < 129) {
                                if (ssDCval < 76) {
                                    if (CsDiff < 9) {
                                        if (Cs < 1039) {
                                            if (mu_mv_mag_sq < 36) {
                                                if (TSC < 8) {
                                                    if (TSC < 7) {
                                                        return 1;
                                                    }
                                                    else if (TSC >= 7) {
                                                        return 8;
                                                    }
                                                }
                                                else if (TSC >= 8) {
                                                    return 1;
                                                }
                                            }
                                            else if (mu_mv_mag_sq >= 36) {
                                                if (refDCval < 65) {
                                                    if (Rs < 249) {
                                                        return 2;
                                                    }
                                                    else if (Rs >= 249) {
                                                        return 4;
                                                    }
                                                }
                                                else if (refDCval >= 65) {
                                                    return 2;
                                                }
                                            }
                                        }
                                        else if (Cs >= 1039) {
                                            if (negBalance < 41) {
                                                if (RsDiff < 8) {
                                                    return 4;
                                                }
                                                else if (RsDiff >= 8) {
                                                    return 8;
                                                }
                                            }
                                            else if (negBalance >= 41) {
                                                if (negBalance < 56) {
                                                    return 8;
                                                }
                                                else if (negBalance >= 56) {
                                                    if (Rs < 2132) {
                                                        return 4;
                                                    }
                                                    else if (Rs >= 2132) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (CsDiff >= 9) {
                                        if (negBalance < 47) {
                                            return 1;
                                        }
                                        else if (negBalance >= 47) {
                                            if (ssDCval < 75) {
                                                return 8;
                                            }
                                            else if (ssDCval >= 75) {
                                                return 4;
                                            }
                                        }
                                    }
                                }
                                else if (ssDCval >= 76) {
                                    if (Rs < 4604) {
                                        if (Cs < 2333) {
                                            if (MVDiff < 20) {
                                                if (Cs < 1772) {
                                                    if (negBalance < 28) {
                                                        if (Rs < 2096) {
                                                            return 4;
                                                        }
                                                        else if (Rs >= 2096) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (negBalance >= 28) {
                                                        if (RsDiff < 6) {
                                                            return 4;
                                                        }
                                                        else if (RsDiff >= 6) {
                                                            return 2;
                                                        }
                                                    }
                                                }
                                                else if (Cs >= 1772) {
                                                    if (RsDiff < 8) {
                                                        if (Rs < 2091) {
                                                            return 8;
                                                        }
                                                        else if (Rs >= 2091) {
                                                            if (AFD < 14) {
                                                                if (Cs < 2267) {
                                                                    if (RsCsDiff < 56) {
                                                                        return 2;
                                                                    }
                                                                    else if (RsCsDiff >= 56) {
                                                                        if (gchDC < 1) {
                                                                            return 1;
                                                                        }
                                                                        else if (gchDC >= 1) {
                                                                            if (Cs < 1969) {
                                                                                if (refDCval < 85) {
                                                                                    return 1;
                                                                                }
                                                                                else if (refDCval >= 85) {
                                                                                    if (Cs < 1862) {
                                                                                        return 1;
                                                                                    }
                                                                                    else if (Cs >= 1862) {
                                                                                        return 2;
                                                                                    }
                                                                                }
                                                                            }
                                                                            else if (Cs >= 1969) {
                                                                                return 1;
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                                else if (Cs >= 2267) {
                                                                    if (ssDCval < 92) {
                                                                        if (Rs < 2883) {
                                                                            return 4;
                                                                        }
                                                                        else if (Rs >= 2883) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                    else if (ssDCval >= 92) {
                                                                        return 1;
                                                                    }
                                                                }
                                                            }
                                                            else if (AFD >= 14) {
                                                                if (tcor < 84) {
                                                                    return 1;
                                                                }
                                                                else if (tcor >= 84) {
                                                                    return 2;
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (RsDiff >= 8) {
                                                        if (Rs < 2846) {
                                                            if (negBalance < 46) {
                                                                if (refDCval < 96) {
                                                                    if (posBalance < 54) {
                                                                        if (refDCval < 82) {
                                                                            return 8;
                                                                        }
                                                                        else if (refDCval >= 82) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                    else if (posBalance >= 54) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (refDCval >= 96) {
                                                                    return 1;
                                                                }
                                                            }
                                                            else if (negBalance >= 46) {
                                                                if (refDCval < 92) {
                                                                    if (RsDiff < 9) {
                                                                        if (refDCval < 87) {
                                                                            if (mcTcor < 92) {
                                                                                return 1;
                                                                            }
                                                                            else if (mcTcor >= 92) {
                                                                                if (diffTSC < 1) {
                                                                                    return 2;
                                                                                }
                                                                                else if (diffTSC >= 1) {
                                                                                    return 1;
                                                                                }
                                                                            }
                                                                        }
                                                                        else if (refDCval >= 87) {
                                                                            return 2;
                                                                        }
                                                                    }
                                                                    else if (RsDiff >= 9) {
                                                                        return 2;
                                                                    }
                                                                }
                                                                else if (refDCval >= 92) {
                                                                    if (negBalance < 52) {
                                                                        return 1;
                                                                    }
                                                                    else if (negBalance >= 52) {
                                                                        if (AFD < 13) {
                                                                            return 4;
                                                                        }
                                                                        else if (AFD >= 13) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (Rs >= 2846) {
                                                            if (Rs < 3145) {
                                                                if (Cs < 2330) {
                                                                    return 1;
                                                                }
                                                                else if (Cs >= 2330) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (Rs >= 3145) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                            else if (MVDiff >= 20) {
                                                if (diffAFD < -1) {
                                                    if (negBalance < 48) {
                                                        if (mu_mv_mag_sq < 42) {
                                                            return 8;
                                                        }
                                                        else if (mu_mv_mag_sq >= 42) {
                                                            if (mu_mv_mag_sq < 62) {
                                                                return 4;
                                                            }
                                                            else if (mu_mv_mag_sq >= 62) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                    else if (negBalance >= 48) {
                                                        if (Rs < 2994) {
                                                            if (ssDCval < 85) {
                                                                if (Cs < 2062) {
                                                                    return 1;
                                                                }
                                                                else if (Cs >= 2062) {
                                                                    return 2;
                                                                }
                                                            }
                                                            else if (ssDCval >= 85) {
                                                                if (Cs < 1086) {
                                                                    return 4;
                                                                }
                                                                else if (Cs >= 1086) {
                                                                    return 2;
                                                                }
                                                            }
                                                        }
                                                        else if (Rs >= 2994) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                                else if (diffAFD >= -1) {
                                                    if (Rs < 3512) {
                                                        if (tcor < 79) {
                                                            if (SC < 3124) {
                                                                return 1;
                                                            }
                                                            else if (SC >= 3124) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (tcor >= 79) {
                                                            if (RsDiff < 3) {
                                                                return 4;
                                                            }
                                                            else if (RsDiff >= 3) {
                                                                if (Rs < 2876) {
                                                                    if (ssDCval < 89) {
                                                                        if (negBalance < 54) {
                                                                            return 2;
                                                                        }
                                                                        else if (negBalance >= 54) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                    else if (ssDCval >= 89) {
                                                                        return 1;
                                                                    }
                                                                }
                                                                else if (Rs >= 2876) {
                                                                    if (diffAFD < 1) {
                                                                        if (Rs < 3268) {
                                                                            return 4;
                                                                        }
                                                                        else if (Rs >= 3268) {
                                                                            return 2;
                                                                        }
                                                                    }
                                                                    else if (diffAFD >= 1) {
                                                                        return 2;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (Rs >= 3512) {
                                                        if (Rs < 4087) {
                                                            return 4;
                                                        }
                                                        else if (Rs >= 4087) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (Cs >= 2333) {
                                            if (posBalance < 51) {
                                                if (Rs < 4083) {
                                                    if (negBalance < 39) {
                                                        if (tcor < 91) {
                                                            if (SC < 5799) {
                                                                return 1;
                                                            }
                                                            else if (SC >= 5799) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (tcor >= 91) {
                                                            if (Rs < 2969) {
                                                                return 8;
                                                            }
                                                            else if (Rs >= 2969) {
                                                                return 2;
                                                            }
                                                        }
                                                    }
                                                    else if (negBalance >= 39) {
                                                        return 8;
                                                    }
                                                }
                                                else if (Rs >= 4083) {
                                                    if (SC < 7122) {
                                                        if (Rs < 4510) {
                                                            return 4;
                                                        }
                                                        else if (Rs >= 4510) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (SC >= 7122) {
                                                        if (ssDCval < 108) {
                                                            return 4;
                                                        }
                                                        else if (ssDCval >= 108) {
                                                            return 1;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (posBalance >= 51) {
                                                if (RsCsDiff < 163) {
                                                    if (posBalance < 59) {
                                                        if (Rs < 2962) {
                                                            if (TSC < 11) {
                                                                if (Cs < 2376) {
                                                                    return 1;
                                                                }
                                                                else if (Cs >= 2376) {
                                                                    if (Cs < 2403) {
                                                                        return 8;
                                                                    }
                                                                    else if (Cs >= 2403) {
                                                                        return 4;
                                                                    }
                                                                }
                                                            }
                                                            else if (TSC >= 11) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (Rs >= 2962) {
                                                            if (SC < 6479) {
                                                                if (Cs < 2469) {
                                                                    if (diffRsCsdiff < 16) {
                                                                        if (diffRsCsdiff < -16) {
                                                                            if (Rs < 3292) {
                                                                                return 8;
                                                                            }
                                                                            else if (Rs >= 3292) {
                                                                                return 2;
                                                                            }
                                                                        }
                                                                        else if (diffRsCsdiff >= -16) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                    else if (diffRsCsdiff >= 16) {
                                                                        if (Rs < 2997) {
                                                                            return 8;
                                                                        }
                                                                        else if (Rs >= 2997) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                }
                                                                else if (Cs >= 2469) {
                                                                    return 1;
                                                                }
                                                            }
                                                            else if (SC >= 6479) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                    else if (posBalance >= 59) {
                                                        if (SC < 4925) {
                                                            return 2;
                                                        }
                                                        else if (SC >= 4925) {
                                                            return 1;
                                                        }
                                                    }
                                                }
                                                else if (RsCsDiff >= 163) {
                                                    if (mcTcor < 91) {
                                                        if (ssDCval < 124) {
                                                            if (ssDCval < 108) {
                                                                if (mcTcor < 90) {
                                                                    return 4;
                                                                }
                                                                else if (mcTcor >= 90) {
                                                                    if (Rs < 3318) {
                                                                        return 4;
                                                                    }
                                                                    else if (Rs >= 3318) {
                                                                        return 1;
                                                                    }
                                                                }
                                                            }
                                                            else if (ssDCval >= 108) {
                                                                if (SC < 7104) {
                                                                    return 4;
                                                                }
                                                                else if (SC >= 7104) {
                                                                    return 2;
                                                                }
                                                            }
                                                        }
                                                        else if (ssDCval >= 124) {
                                                            return 1;
                                                        }
                                                    }
                                                    else if (mcTcor >= 91) {
                                                        if (negBalance < 51) {
                                                            return 8;
                                                        }
                                                        else if (negBalance >= 51) {
                                                            return 2;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (Rs >= 4604) {
                                        if (refDCval < 105) {
                                            return 4;
                                        }
                                        else if (refDCval >= 105) {
                                            if (TSC < 11) {
                                                if (Cs < 2821) {
                                                    if (RsDiff < 13) {
                                                        return 8;
                                                    }
                                                    else if (RsDiff >= 13) {
                                                        return 4;
                                                    }
                                                }
                                                else if (Cs >= 2821) {
                                                    if (MVDiff < 11) {
                                                        return 4;
                                                    }
                                                    else if (MVDiff >= 11) {
                                                        return 2;
                                                    }
                                                }
                                            }
                                            else if (TSC >= 11) {
                                                return 2;
                                            }
                                        }
                                    }
                                }
                            }
                            else if (refDCval >= 129) {
                                if (posBalance < 58) {
                                    if (mcTcor < 89) {
                                        if (ssDCval < 140) {
                                            if (posBalance < 47) {
                                                return 8;
                                            }
                                            else if (posBalance >= 47) {
                                                return 1;
                                            }
                                        }
                                        else if (ssDCval >= 140) {
                                            return 8;
                                        }
                                    }
                                    else if (mcTcor >= 89) {
                                        if (Cs < 1705) {
                                            if (Cs < 709) {
                                                return 4;
                                            }
                                            else if (Cs >= 709) {
                                                return 2;
                                            }
                                        }
                                        else if (Cs >= 1705) {
                                            return 8;
                                        }
                                    }
                                }
                                else if (posBalance >= 58) {
                                    if (ssDCval < 134) {
                                        if (MVDiff < 9) {
                                            return 1;
                                        }
                                        else if (MVDiff >= 9) {
                                            return 8;
                                        }
                                    }
                                    else if (ssDCval >= 134) {
                                        if (mu_mv_mag_sq < 43) {
                                            return 2;
                                        }
                                        else if (mu_mv_mag_sq >= 43) {
                                            if (Cs < 1624) {
                                                return 4;
                                            }
                                            else if (Cs >= 1624) {
                                                return 1;
                                            }
                                        }
                                    }
                                }
                            }
                        }
                        else if (AFD >= 17) {
                            if (CsDiff < 9) {
                                if (diffAFD < 3) {
                                    if (AFD < 31) {
                                        if (mu_mv_mag_sq < 170) {
                                            if (TSC < 10) {
                                                if (Cs < 2037) {
                                                    if (TSCindex < 7) {
                                                        if (ssDCval < 117) {
                                                            return 2;
                                                        }
                                                        else if (ssDCval >= 117) {
                                                            if (Rs < 1250) {
                                                                return 2;
                                                            }
                                                            else if (Rs >= 1250) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                    else if (TSCindex >= 7) {
                                                        return 2;
                                                    }
                                                }
                                                else if (Cs >= 2037) {
                                                    if (Rs < 2932) {
                                                        return 8;
                                                    }
                                                    else if (Rs >= 2932) {
                                                        return 2;
                                                    }
                                                }
                                            }
                                            else if (TSC >= 10) {
                                                if (Cs < 1683) {
                                                    if (ssDCval < 137) {
                                                        if (AFD < 20) {
                                                            return 4;
                                                        }
                                                        else if (AFD >= 20) {
                                                            return 2;
                                                        }
                                                    }
                                                    else if (ssDCval >= 137) {
                                                        if (Rs < 3376) {
                                                            return 2;
                                                        }
                                                        else if (Rs >= 3376) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                                else if (Cs >= 1683) {
                                                    if (CsDiff < 8) {
                                                        if (Rs < 3596) {
                                                            return 8;
                                                        }
                                                        else if (Rs >= 3596) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (CsDiff >= 8) {
                                                        if (SC < 5947) {
                                                            if (diffRsCsdiff < -9) {
                                                                if (Cs < 2394) {
                                                                    return 2;
                                                                }
                                                                else if (Cs >= 2394) {
                                                                    return 1;
                                                                }
                                                            }
                                                            else if (diffRsCsdiff >= -9) {
                                                                return 1;
                                                            }
                                                        }
                                                        else if (SC >= 5947) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (mu_mv_mag_sq >= 170) {
                                            if (Cs < 260) {
                                                return 2;
                                            }
                                            else if (Cs >= 260) {
                                                return 1;
                                            }
                                        }
                                    }
                                    else if (AFD >= 31) {
                                        return 2;
                                    }
                                }
                                else if (diffAFD >= 3) {
                                    if (RsDiff < 17) {
                                        if (Cs < 350) {
                                            if (diffMVdiffVal < 351) {
                                                if (gchDC < 26) {
                                                    return 2;
                                                }
                                                else if (gchDC >= 26) {
                                                    return 1;
                                                }
                                            }
                                            else if (diffMVdiffVal >= 351) {
                                                return 8;
                                            }
                                        }
                                        else if (Cs >= 350) {
                                            if (diffTSC < 3) {
                                                if (posBalance < 54) {
                                                    if (Rs < 2175) {
                                                        return 2;
                                                    }
                                                    else if (Rs >= 2175) {
                                                        return 8;
                                                    }
                                                }
                                                else if (posBalance >= 54) {
                                                    return 1;
                                                }
                                            }
                                            else if (diffTSC >= 3) {
                                                if (mu_mv_mag_sq < 51) {
                                                    return 1;
                                                }
                                                else if (mu_mv_mag_sq >= 51) {
                                                    return 4;
                                                }
                                            }
                                        }
                                    }
                                    else if (RsDiff >= 17) {
                                        return 2;
                                    }
                                }
                            }
                            else if (CsDiff >= 9) {
                                if (AFD < 24) {
                                    if (Rs < 3661) {
                                        if (tcor < 81) {
                                            if (refDCval < 64) {
                                                if (Rs < 2500) {
                                                    return 4;
                                                }
                                                else if (Rs >= 2500) {
                                                    return 8;
                                                }
                                            }
                                            else if (refDCval >= 64) {
                                                if (RsDiff < 5) {
                                                    return 8;
                                                }
                                                else if (RsDiff >= 5) {
                                                    if (negBalance < 56) {
                                                        return 1;
                                                    }
                                                    else if (negBalance >= 56) {
                                                        if (MVDiff < 16) {
                                                            if (SC < 5887) {
                                                                return 4;
                                                            }
                                                            else if (SC >= 5887) {
                                                                return 1;
                                                            }
                                                        }
                                                        else if (MVDiff >= 16) {
                                                            return 1;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (tcor >= 81) {
                                            if (diffRsCsdiff < -36) {
                                                if (MVDiff < 12) {
                                                    return 2;
                                                }
                                                else if (MVDiff >= 12) {
                                                    return 8;
                                                }
                                            }
                                            else if (diffRsCsdiff >= -36) {
                                                if (Rs < 3123) {
                                                    if (RsCsDiff < 322) {
                                                        return 1;
                                                    }
                                                    else if (RsCsDiff >= 322) {
                                                        return 8;
                                                    }
                                                }
                                                else if (Rs >= 3123) {
                                                    return 2;
                                                }
                                            }
                                        }
                                    }
                                    else if (Rs >= 3661) {
                                        if (mcTcor < 80) {
                                            if (diffAFD < 8) {
                                                return 1;
                                            }
                                            else if (diffAFD >= 8) {
                                                return 8;
                                            }
                                        }
                                        else if (mcTcor >= 80) {
                                            if (AFD < 18) {
                                                if (MVDiff < 18) {
                                                    return 1;
                                                }
                                                else if (MVDiff >= 18) {
                                                    return 4;
                                                }
                                            }
                                            else if (AFD >= 18) {
                                                return 4;
                                            }
                                        }
                                    }
                                }
                                else if (AFD >= 24) {
                                    if (mcTcor < 85) {
                                        if (TSC < 13) {
                                            if (RsCsDiff < 191) {
                                                return 2;
                                            }
                                            else if (RsCsDiff >= 191) {
                                                return 1;
                                            }
                                        }
                                        else if (TSC >= 13) {
                                            if (AFD < 36) {
                                                if (CsDiff < 18) {
                                                    return 8;
                                                }
                                                else if (CsDiff >= 18) {
                                                    if (SC < 3620) {
                                                        return 1;
                                                    }
                                                    else if (SC >= 3620) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                            else if (AFD >= 36) {
                                                if (AFD < 62) {
                                                    if (diffAFD < 26) {
                                                        if (Rs < 1836) {
                                                            return 1;
                                                        }
                                                        else if (Rs >= 1836) {
                                                            return 2;
                                                        }
                                                    }
                                                    else if (diffAFD >= 26) {
                                                        if (ssDCval < 70) {
                                                            if (Rs < 981) {
                                                                return 8;
                                                            }
                                                            else if (Rs >= 981) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (ssDCval >= 70) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                                else if (AFD >= 62) {
                                                    if (Cs < 956) {
                                                        return 1;
                                                    }
                                                    else if (Cs >= 956) {
                                                        return 2;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (mcTcor >= 85) {
                                        if (SC < 6002) {
                                            if (ssDCval < 147) {
                                                if (Rs < 1487) {
                                                    return 8;
                                                }
                                                else if (Rs >= 1487) {
                                                    if (RsCsDiff < 191) {
                                                        return 1;
                                                    }
                                                    else if (RsCsDiff >= 191) {
                                                        return 2;
                                                    }
                                                }
                                            }
                                            else if (ssDCval >= 147) {
                                                if (TSC < 9) {
                                                    return 2;
                                                }
                                                else if (TSC >= 9) {
                                                    if (Rs < 2088) {
                                                        return 2;
                                                    }
                                                    else if (Rs >= 2088) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                        }
                                        else if (SC >= 6002) {
                                            return 4;
                                        }
                                    }
                                }
                            }
                        }
                    }
                }
            }
            else if (Cs >= 3274) {
                if (mu_mv_mag_sq < 60) {
                    if (AFD < 15) {
                        if (RsCsDiff < 46) {
                            if (TSCindex < 8) {
                                if (tcor < 90) {
                                    if (ssDCval < 63) {
                                        if (refDCval < 53) {
                                            if (diffRsCsdiff < 3) {
                                                return 8;
                                            }
                                            else if (diffRsCsdiff >= 3) {
                                                return 4;
                                            }
                                        }
                                        else if (refDCval >= 53) {
                                            return 8;
                                        }
                                    }
                                    else if (ssDCval >= 63) {
                                        return 4;
                                    }
                                }
                                else if (tcor >= 90) {
                                    if (posBalance < 57) {
                                        if (Cs < 3823) {
                                            if (Cs < 3306) {
                                                return 8;
                                            }
                                            else if (Cs >= 3306) {
                                                return 4;
                                            }
                                        }
                                        else if (Cs >= 3823) {
                                            return 2;
                                        }
                                    }
                                    else if (posBalance >= 57) {
                                        if (MVDiff < 18) {
                                            if (SC < 6779) {
                                                return 4;
                                            }
                                            else if (SC >= 6779) {
                                                return 8;
                                            }
                                        }
                                        else if (MVDiff >= 18) {
                                            return 8;
                                        }
                                    }
                                }
                            }
                            else if (TSCindex >= 8) {
                                if (MVDiff < 18) {
                                    if (RsDiff < 4) {
                                        if (posBalance < 57) {
                                            return 2;
                                        }
                                        else if (posBalance >= 57) {
                                            if (Rs < 3631) {
                                                return 4;
                                            }
                                            else if (Rs >= 3631) {
                                                return 2;
                                            }
                                        }
                                    }
                                    else if (RsDiff >= 4) {
                                        if (TSC < 8) {
                                            return 4;
                                        }
                                        else if (TSC >= 8) {
                                            if (Cs < 3943) {
                                                return 4;
                                            }
                                            else if (Cs >= 3943) {
                                                return 2;
                                            }
                                        }
                                    }
                                }
                                else if (MVDiff >= 18) {
                                    if (ssDCval < 58) {
                                        if (Rs < 3510) {
                                            if (refDCval < 52) {
                                                if (mu_mv_mag_sq < 37) {
                                                    if (Cs < 3410) {
                                                        return 2;
                                                    }
                                                    else if (Cs >= 3410) {
                                                        return 8;
                                                    }
                                                }
                                                else if (mu_mv_mag_sq >= 37) {
                                                    return 8;
                                                }
                                            }
                                            else if (refDCval >= 52) {
                                                return 8;
                                            }
                                        }
                                        else if (Rs >= 3510) {
                                            if (tcor < 86) {
                                                return 8;
                                            }
                                            else if (tcor >= 86) {
                                                if (mu_mv_mag_sq < 46) {
                                                    return 2;
                                                }
                                                else if (mu_mv_mag_sq >= 46) {
                                                    return 8;
                                                }
                                            }
                                        }
                                    }
                                    else if (ssDCval >= 58) {
                                        if (tcor < 89) {
                                            if (Cs < 3878) {
                                                return 4;
                                            }
                                            else if (Cs >= 3878) {
                                                return 8;
                                            }
                                        }
                                        else if (tcor >= 89) {
                                            return 8;
                                        }
                                    }
                                }
                            }
                        }
                        else if (RsCsDiff >= 46) {
                            if (mcTcor < 95) {
                                if (posBalance < 56) {
                                    if (mu_mv_mag_sq < 40) {
                                        return 8;
                                    }
                                    else if (mu_mv_mag_sq >= 40) {
                                        if (diffTSC < -2) {
                                            return 8;
                                        }
                                        else if (diffTSC >= -2) {
                                            if (MVDiff < 52) {
                                                if (negBalance < 45) {
                                                    if (Cs < 4070) {
                                                        return 1;
                                                    }
                                                    else if (Cs >= 4070) {
                                                        return 8;
                                                    }
                                                }
                                                else if (negBalance >= 45) {
                                                    return 4;
                                                }
                                            }
                                            else if (MVDiff >= 52) {
                                                return 8;
                                            }
                                        }
                                    }
                                }
                                else if (posBalance >= 56) {
                                    if (Cs < 3365) {
                                        return 8;
                                    }
                                    else if (Cs >= 3365) {
                                        if (mcTcor < 92) {
                                            return 4;
                                        }
                                        else if (mcTcor >= 92) {
                                            if (mu_mv_mag_sq < 45) {
                                                if (SC < 7558) {
                                                    return 8;
                                                }
                                                else if (SC >= 7558) {
                                                    if (Cs < 4042) {
                                                        return 4;
                                                    }
                                                    else if (Cs >= 4042) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                            else if (mu_mv_mag_sq >= 45) {
                                                return 4;
                                            }
                                        }
                                    }
                                }
                            }
                            else if (mcTcor >= 95) {
                                return 8;
                            }
                        }
                    }
                    else if (AFD >= 15) {
                        if (ssDCval < 153) {
                            if (Rs < 7219) {
                                if (Rs < 4103) {
                                    if (refDCval < 124) {
                                        if (CsDiff < 20) {
                                            if (CsDiff < 17) {
                                                if (Rs < 3999) {
                                                    return 8;
                                                }
                                                else if (Rs >= 3999) {
                                                    return 4;
                                                }
                                            }
                                            else if (CsDiff >= 17) {
                                                if (ssDCval < 78) {
                                                    return 1;
                                                }
                                                else if (ssDCval >= 78) {
                                                    if (AFD < 28) {
                                                        return 8;
                                                    }
                                                    else if (AFD >= 28) {
                                                        if (Rs < 2733) {
                                                            return 4;
                                                        }
                                                        else if (Rs >= 2733) {
                                                            return 2;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (CsDiff >= 20) {
                                            return 4;
                                        }
                                    }
                                    else if (refDCval >= 124) {
                                        return 1;
                                    }
                                }
                                else if (Rs >= 4103) {
                                    if (mu_mv_mag_sq < 3) {
                                        if (ssDCval < 106) {
                                            if (CsDiff < 16) {
                                                return 8;
                                            }
                                            else if (CsDiff >= 16) {
                                                return 1;
                                            }
                                        }
                                        else if (ssDCval >= 106) {
                                            if (mcTcor < 82) {
                                                if (SC < 9998) {
                                                    return 4;
                                                }
                                                else if (SC >= 9998) {
                                                    return 1;
                                                }
                                            }
                                            else if (mcTcor >= 82) {
                                                if (tcor < 77) {
                                                    if (Cs < 5099) {
                                                        if (ssDCval < 149) {
                                                            if (Cs < 3482) {
                                                                return 8;
                                                            }
                                                            else if (Cs >= 3482) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (ssDCval >= 149) {
                                                            return 2;
                                                        }
                                                    }
                                                    else if (Cs >= 5099) {
                                                        return 8;
                                                    }
                                                }
                                                else if (tcor >= 77) {
                                                    return 8;
                                                }
                                            }
                                        }
                                    }
                                    else if (mu_mv_mag_sq >= 3) {
                                        if (Cs < 3420) {
                                            if (ssDCval < 111) {
                                                if (Rs < 5339) {
                                                    return 2;
                                                }
                                                else if (Rs >= 5339) {
                                                    return 4;
                                                }
                                            }
                                            else if (ssDCval >= 111) {
                                                if (AFD < 21) {
                                                    if (SC < 8066) {
                                                        return 1;
                                                    }
                                                    else if (SC >= 8066) {
                                                        return 8;
                                                    }
                                                }
                                                else if (AFD >= 21) {
                                                    return 8;
                                                }
                                            }
                                        }
                                        else if (Cs >= 3420) {
                                            if (RsDiff < 36) {
                                                if (Cs < 4327) {
                                                    if (Rs < 5285) {
                                                        if (AFD < 18) {
                                                            if (ssDCval < 95) {
                                                                return 8;
                                                            }
                                                            else if (ssDCval >= 95) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (AFD >= 18) {
                                                            if (AFD < 31) {
                                                                return 8;
                                                            }
                                                            else if (AFD >= 31) {
                                                                if (Rs < 4377) {
                                                                    return 2;
                                                                }
                                                                else if (Rs >= 4377) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (Rs >= 5285) {
                                                        if (RsDiff < 20) {
                                                            return 8;
                                                        }
                                                        else if (RsDiff >= 20) {
                                                            if (ssDCval < 107) {
                                                                return 8;
                                                            }
                                                            else if (ssDCval >= 107) {
                                                                if (diffTSC < 3) {
                                                                    if (Rs < 5933) {
                                                                        return 4;
                                                                    }
                                                                    else if (Rs >= 5933) {
                                                                        if (SC < 10150) {
                                                                            return 8;
                                                                        }
                                                                        else if (SC >= 10150) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                }
                                                                else if (diffTSC >= 3) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (Cs >= 4327) {
                                                    if (refDCval < 60) {
                                                        return 4;
                                                    }
                                                    else if (refDCval >= 60) {
                                                        if (tcor < 84) {
                                                            if (Rs < 6965) {
                                                                return 8;
                                                            }
                                                            else if (Rs >= 6965) {
                                                                if (tcor < 76) {
                                                                    return 8;
                                                                }
                                                                else if (tcor >= 76) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                        else if (tcor >= 84) {
                                                            if (Rs < 6336) {
                                                                return 8;
                                                            }
                                                            else if (Rs >= 6336) {
                                                                if (diffAFD < 1) {
                                                                    return 4;
                                                                }
                                                                else if (diffAFD >= 1) {
                                                                    return 1;
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                            else if (RsDiff >= 36) {
                                                if (Cs < 5179) {
                                                    if (TSC < 21) {
                                                        return 2;
                                                    }
                                                    else if (TSC >= 21) {
                                                        return 8;
                                                    }
                                                }
                                                else if (Cs >= 5179) {
                                                    return 8;
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                            else if (Rs >= 7219) {
                                if (ssDCval < 79) {
                                    return 4;
                                }
                                else if (ssDCval >= 79) {
                                    if (Rs < 14475) {
                                        if (MVDiff < 23) {
                                            if (ssDCval < 93) {
                                                if (mcTcor < 87) {
                                                    return 4;
                                                }
                                                else if (mcTcor >= 87) {
                                                    return 2;
                                                }
                                            }
                                            else if (ssDCval >= 93) {
                                                if (TSC < 22) {
                                                    if (mcTcor < 91) {
                                                        if (tcor < 81) {
                                                            return 8;
                                                        }
                                                        else if (tcor >= 81) {
                                                            if (Cs < 6669) {
                                                                return 4;
                                                            }
                                                            else if (Cs >= 6669) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                    else if (mcTcor >= 91) {
                                                        if (Rs < 9940) {
                                                            return 4;
                                                        }
                                                        else if (Rs >= 9940) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                                else if (TSC >= 22) {
                                                    if (tcor < 62) {
                                                        if (Rs < 9829) {
                                                            return 8;
                                                        }
                                                        else if (Rs >= 9829) {
                                                            return 2;
                                                        }
                                                    }
                                                    else if (tcor >= 62) {
                                                        if (diffMVdiffVal < -2) {
                                                            if (SC < 22051) {
                                                                return 4;
                                                            }
                                                            else if (SC >= 22051) {
                                                                if (Rs < 13754) {
                                                                    return 2;
                                                                }
                                                                else if (Rs >= 13754) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                        else if (diffMVdiffVal >= -2) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (MVDiff >= 23) {
                                            if (diffTSC < -3) {
                                                return 8;
                                            }
                                            else if (diffTSC >= -3) {
                                                return 2;
                                            }
                                        }
                                    }
                                    else if (Rs >= 14475) {
                                        return 8;
                                    }
                                }
                            }
                        }
                        else if (ssDCval >= 153) {
                            if (SC < 12944) {
                                if (posBalance < 58) {
                                    if (AFD < 28) {
                                        return 8;
                                    }
                                    else if (AFD >= 28) {
                                        if (Cs < 4439) {
                                            return 4;
                                        }
                                        else if (Cs >= 4439) {
                                            if (AFD < 32) {
                                                return 4;
                                            }
                                            else if (AFD >= 32) {
                                                return 8;
                                            }
                                        }
                                    }
                                }
                                else if (posBalance >= 58) {
                                    if (mu_mv_mag_sq < 3) {
                                        if (Rs < 6096) {
                                            if (refDCval < 159) {
                                                return 4;
                                            }
                                            else if (refDCval >= 159) {
                                                return 2;
                                            }
                                        }
                                        else if (Rs >= 6096) {
                                            return 2;
                                        }
                                    }
                                    else if (mu_mv_mag_sq >= 3) {
                                        if (AFD < 25) {
                                            return 4;
                                        }
                                        else if (AFD >= 25) {
                                            return 8;
                                        }
                                    }
                                }
                            }
                            else if (SC >= 12944) {
                                if (AFD < 33) {
                                    return 8;
                                }
                                else if (AFD >= 33) {
                                    if (Cs < 6689) {
                                        return 2;
                                    }
                                    else if (Cs >= 6689) {
                                        return 8;
                                    }
                                }
                            }
                        }
                    }
                }
                else if (mu_mv_mag_sq >= 60) {
                    if (refDCval < 66) {
                        if (CsDiff < 4) {
                            if (ssDCval < 54) {
                                if (negBalance < 44) {
                                    return 8;
                                }
                                else if (negBalance >= 44) {
                                    if (posBalance < 56) {
                                        return 4;
                                    }
                                    else if (posBalance >= 56) {
                                        return 8;
                                    }
                                }
                            }
                            else if (ssDCval >= 54) {
                                if (ssDCval < 56) {
                                    return 8;
                                }
                                else if (ssDCval >= 56) {
                                    if (posBalance < 56) {
                                        return 8;
                                    }
                                    else if (posBalance >= 56) {
                                        return 4;
                                    }
                                }
                            }
                        }
                        else if (CsDiff >= 4) {
                            if (AFD < 12) {
                                if (TSC < 7) {
                                    if (ssDCval < 53) {
                                        return 4;
                                    }
                                    else if (ssDCval >= 53) {
                                        if (Rs < 3938) {
                                            return 8;
                                        }
                                        else if (Rs >= 3938) {
                                            if (Cs < 3992) {
                                                return 4;
                                            }
                                            else if (Cs >= 3992) {
                                                return 8;
                                            }
                                        }
                                    }
                                }
                                else if (TSC >= 7) {
                                    return 4;
                                }
                            }
                            else if (AFD >= 12) {
                                if (Cs < 3523) {
                                    return 4;
                                }
                                else if (Cs >= 3523) {
                                    return 8;
                                }
                            }
                        }
                    }
                    else if (refDCval >= 66) {
                        if (ssDCval < 136) {
                            if (Cs < 4118) {
                                if (mcTcor < 82) {
                                    if (AFD < 17) {
                                        if (TSC < 12) {
                                            return 8;
                                        }
                                        else if (TSC >= 12) {
                                            return 4;
                                        }
                                    }
                                    else if (AFD >= 17) {
                                        return 8;
                                    }
                                }
                                else if (mcTcor >= 82) {
                                    if (negBalance < 44) {
                                        if (ssDCval < 108) {
                                            return 8;
                                        }
                                        else if (ssDCval >= 108) {
                                            if (SC < 7715) {
                                                return 8;
                                            }
                                            else if (SC >= 7715) {
                                                return 2;
                                            }
                                        }
                                    }
                                    else if (negBalance >= 44) {
                                        if (MVDiff < 90) {
                                            return 8;
                                        }
                                        else if (MVDiff >= 90) {
                                            if (Rs < 5012) {
                                                return 4;
                                            }
                                            else if (Rs >= 5012) {
                                                return 8;
                                            }
                                        }
                                    }
                                }
                            }
                            else if (Cs >= 4118) {
                                if (mu_mv_mag_sq < 63) {
                                    if (TSC < 14) {
                                        return 4;
                                    }
                                    else if (TSC >= 14) {
                                        return 8;
                                    }
                                }
                                else if (mu_mv_mag_sq >= 63) {
                                    return 8;
                                }
                            }
                        }
                        else if (ssDCval >= 136) {
                            if (Rs < 4496) {
                                if (TSC < 13) {
                                    if (Rs < 4094) {
                                        return 1;
                                    }
                                    else if (Rs >= 4094) {
                                        return 8;
                                    }
                                }
                                else if (TSC >= 13) {
                                    return 1;
                                }
                            }
                            else if (Rs >= 4496) {
                                return 8;
                            }
                        }
                    }
                }
            }
        }
        return 4;
    }
    mfxU16 AGOPSelect2(mfxU32 Rs, mfxU32 CsDiff, mfxU32 TSC, mfxI32 diffAFD, mfxU32 Scindex, mfxU32 RsDiff, mfxI32 diffMVdiffVal, mfxU32 RsCsDiff, mfxU32 TSCindex, mfxI16 tcor, mfxU32 refDCval, mfxU32 MVDiff, mfxU32 AFD, mfxI32 diffTSC, mfxI32 diffRsCsdiff, mfxU32 posBalance, mfxU32 SC, mfxU32 mu_mv_mag_sq, mfxU32 negBalance, mfxI16 mcTcor, mfxU32 gchDC, mfxU32 Cs, mfxU32 ssDCval) {
        if (tcor < 95) {
            if (mu_mv_mag_sq < 74) {
                if (mu_mv_mag_sq < 1) {
                    if (SC < 3490) {
                        return 8;
                    }
                    else if (SC >= 3490) {
                        if (SC < 9327) {
                            if (Rs < 2662) {
                                return 1;
                            }
                            else if (Rs >= 2662) {
                                if (Cs < 3985) {
                                    if (Rs < 3755) {
                                        return 8;
                                    }
                                    else if (Rs >= 3755) {
                                        if (Cs < 3951) {
                                            return 1;
                                        }
                                        else if (Cs >= 3951) {
                                            if (SC < 7761) {
                                                return 1;
                                            }
                                            else if (SC >= 7761) {
                                                return 8;
                                            }
                                        }
                                    }
                                }
                                else if (Cs >= 3985) {
                                    return 8;
                                }
                            }
                        }
                        else if (SC >= 9327) {
                            if (Rs < 5376) {
                                return 1;
                            }
                            else if (Rs >= 5376) {
                                return 8;
                            }
                        }
                    }
                }
                else if (mu_mv_mag_sq >= 1) {
                    if (Cs < 3273) {
                        if (Cs < 1801) {
                            if (TSCindex < 9) {
                                if (MVDiff < 15) {
                                    if (negBalance < 46) {
                                        if (Cs < 624) {
                                            if (posBalance < 52) {
                                                if (ssDCval < 61) {
                                                    return 4;
                                                }
                                                else if (ssDCval >= 61) {
                                                    if (Rs < 1000) {
                                                        if (Cs < 452) {
                                                            return 2;
                                                        }
                                                        else if (Cs >= 452) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (Rs >= 1000) {
                                                        return 2;
                                                    }
                                                }
                                            }
                                            else if (posBalance >= 52) {
                                                if (ssDCval < 40) {
                                                    if (SC < 1259) {
                                                        if (mcTcor < 86) {
                                                            return 4;
                                                        }
                                                        else if (mcTcor >= 86) {
                                                            if (Cs < 184) {
                                                                return 8;
                                                            }
                                                            else if (Cs >= 184) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                    else if (SC >= 1259) {
                                                        return 1;
                                                    }
                                                }
                                                else if (ssDCval >= 40) {
                                                    if (Rs < 215) {
                                                        if (AFD < 5) {
                                                            return 8;
                                                        }
                                                        else if (AFD >= 5) {
                                                            return 2;
                                                        }
                                                    }
                                                    else if (Rs >= 215) {
                                                        if (Rs < 581) {
                                                            if (Cs < 343) {
                                                                return 8;
                                                            }
                                                            else if (Cs >= 343) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (Rs >= 581) {
                                                            if (CsDiff < 3) {
                                                                if (diffRsCsdiff < -3) {
                                                                    if (Rs < 640) {
                                                                        return 8;
                                                                    }
                                                                    else if (Rs >= 640) {
                                                                        if (AFD < 8) {
                                                                            return 4;
                                                                        }
                                                                        else if (AFD >= 8) {
                                                                            return 2;
                                                                        }
                                                                    }
                                                                }
                                                                else if (diffRsCsdiff >= -3) {
                                                                    if (mu_mv_mag_sq < 5) {
                                                                        if (SC < 1081) {
                                                                            return 4;
                                                                        }
                                                                        else if (SC >= 1081) {
                                                                            if (Rs < 1117) {
                                                                                return 8;
                                                                            }
                                                                            else if (Rs >= 1117) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (mu_mv_mag_sq >= 5) {
                                                                        return 8;
                                                                    }
                                                                }
                                                            }
                                                            else if (CsDiff >= 3) {
                                                                if (Rs < 2009) {
                                                                    return 2;
                                                                }
                                                                else if (Rs >= 2009) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (Cs >= 624) {
                                            if (Rs < 2153) {
                                                if (SC < 2018) {
                                                    if (diffRsCsdiff < 12) {
                                                        if (ssDCval < 54) {
                                                            if (mcTcor < 94) {
                                                                if (Cs < 716) {
                                                                    if (Rs < 841) {
                                                                        return 4;
                                                                    }
                                                                    else if (Rs >= 841) {
                                                                        return 1;
                                                                    }
                                                                }
                                                                else if (Cs >= 716) {
                                                                    if (diffAFD < -1) {
                                                                        if (Rs < 1018) {
                                                                            return 4;
                                                                        }
                                                                        else if (Rs >= 1018) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                    else if (diffAFD >= -1) {
                                                                        return 4;
                                                                    }
                                                                }
                                                            }
                                                            else if (mcTcor >= 94) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (ssDCval >= 54) {
                                                            if (Rs < 943) {
                                                                return 4;
                                                            }
                                                            else if (Rs >= 943) {
                                                                if (refDCval < 54) {
                                                                    return 4;
                                                                }
                                                                else if (refDCval >= 54) {
                                                                    return 1;
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (diffRsCsdiff >= 12) {
                                                        if (Rs < 546) {
                                                            return 2;
                                                        }
                                                        else if (Rs >= 546) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                                else if (SC >= 2018) {
                                                    if (refDCval < 140) {
                                                        if (SC < 2433) {
                                                            if (CsDiff < 4) {
                                                                if (negBalance < 36) {
                                                                    return 1;
                                                                }
                                                                else if (negBalance >= 36) {
                                                                    if (MVDiff < 9) {
                                                                        if (Rs < 1133) {
                                                                            return 4;
                                                                        }
                                                                        else if (Rs >= 1133) {
                                                                            if (posBalance < 49) {
                                                                                if (Cs < 872) {
                                                                                    return 1;
                                                                                }
                                                                                else if (Cs >= 872) {
                                                                                    if (refDCval < 50) {
                                                                                        return 1;
                                                                                    }
                                                                                    else if (refDCval >= 50) {
                                                                                        return 4;
                                                                                    }
                                                                                }
                                                                            }
                                                                            else if (posBalance >= 49) {
                                                                                if (negBalance < 37) {
                                                                                    return 4;
                                                                                }
                                                                                else if (negBalance >= 37) {
                                                                                    if (AFD < 6) {
                                                                                        if (Rs < 1405) {
                                                                                            if (Rs < 1292) {
                                                                                                return 1;
                                                                                            }
                                                                                            else if (Rs >= 1292) {
                                                                                                return 4;
                                                                                            }
                                                                                        }
                                                                                        else if (Rs >= 1405) {
                                                                                            return 1;
                                                                                        }
                                                                                    }
                                                                                    else if (AFD >= 6) {
                                                                                        return 1;
                                                                                    }
                                                                                }
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (MVDiff >= 9) {
                                                                        if (Cs < 997) {
                                                                            return 1;
                                                                        }
                                                                        else if (Cs >= 997) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                            else if (CsDiff >= 4) {
                                                                if (refDCval < 59) {
                                                                    if (Cs < 835) {
                                                                        return 1;
                                                                    }
                                                                    else if (Cs >= 835) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (refDCval >= 59) {
                                                                    if (mcTcor < 94) {
                                                                        return 1;
                                                                    }
                                                                    else if (mcTcor >= 94) {
                                                                        if (RsDiff < 4) {
                                                                            return 1;
                                                                        }
                                                                        else if (RsDiff >= 4) {
                                                                            return 2;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (SC >= 2433) {
                                                            if (mcTcor < 88) {
                                                                if (SC < 2484) {
                                                                    return 2;
                                                                }
                                                                else if (SC >= 2484) {
                                                                    return 1;
                                                                }
                                                            }
                                                            else if (mcTcor >= 88) {
                                                                if (ssDCval < 57) {
                                                                    if (SC < 2510) {
                                                                        if (posBalance < 51) {
                                                                            return 1;
                                                                        }
                                                                        else if (posBalance >= 51) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                    else if (SC >= 2510) {
                                                                        if (SC < 2568) {
                                                                            if (TSC < 5) {
                                                                                return 1;
                                                                            }
                                                                            else if (TSC >= 5) {
                                                                                return 8;
                                                                            }
                                                                        }
                                                                        else if (SC >= 2568) {
                                                                            if (diffMVdiffVal < -3) {
                                                                                if (RsDiff < 9) {
                                                                                    return 1;
                                                                                }
                                                                                else if (RsDiff >= 9) {
                                                                                    return 4;
                                                                                }
                                                                            }
                                                                            else if (diffMVdiffVal >= -3) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                                else if (ssDCval >= 57) {
                                                                    if (ssDCval < 60) {
                                                                        return 4;
                                                                    }
                                                                    else if (ssDCval >= 60) {
                                                                        if (refDCval < 73) {
                                                                            return 1;
                                                                        }
                                                                        else if (refDCval >= 73) {
                                                                            if (SC < 3257) {
                                                                                if (SC < 2645) {
                                                                                    if (refDCval < 120) {
                                                                                        return 2;
                                                                                    }
                                                                                    else if (refDCval >= 120) {
                                                                                        return 1;
                                                                                    }
                                                                                }
                                                                                else if (SC >= 2645) {
                                                                                    if (posBalance < 38) {
                                                                                        return 2;
                                                                                    }
                                                                                    else if (posBalance >= 38) {
                                                                                        if (TSC < 7) {
                                                                                            return 4;
                                                                                        }
                                                                                        else if (TSC >= 7) {
                                                                                            if (Rs < 1266) {
                                                                                                return 4;
                                                                                            }
                                                                                            else if (Rs >= 1266) {
                                                                                                return 1;
                                                                                            }
                                                                                        }
                                                                                    }
                                                                                }
                                                                            }
                                                                            else if (SC >= 3257) {
                                                                                return 8;
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (refDCval >= 140) {
                                                        if (posBalance < 38) {
                                                            if (MVDiff < 5) {
                                                                return 8;
                                                            }
                                                            else if (MVDiff >= 5) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (posBalance >= 38) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (Rs >= 2153) {
                                                if (posBalance < 47) {
                                                    if (Cs < 1043) {
                                                        if (AFD < 7) {
                                                            return 8;
                                                        }
                                                        else if (AFD >= 7) {
                                                            if (RsDiff < 11) {
                                                                return 2;
                                                            }
                                                            else if (RsDiff >= 11) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                    else if (Cs >= 1043) {
                                                        if (AFD < 11) {
                                                            if (diffMVdiffVal < 3) {
                                                                return 4;
                                                            }
                                                            else if (diffMVdiffVal >= 3) {
                                                                if (negBalance < 37) {
                                                                    return 8;
                                                                }
                                                                else if (negBalance >= 37) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                        else if (AFD >= 11) {
                                                            return 2;
                                                        }
                                                    }
                                                }
                                                else if (posBalance >= 47) {
                                                    if (ssDCval < 103) {
                                                        return 4;
                                                    }
                                                    else if (ssDCval >= 103) {
                                                        if (negBalance < 35) {
                                                            if (Cs < 1162) {
                                                                return 8;
                                                            }
                                                            else if (Cs >= 1162) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (negBalance >= 35) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (negBalance >= 46) {
                                        if (negBalance < 52) {
                                            if (ssDCval < 88) {
                                                if (Rs < 530) {
                                                    return 2;
                                                }
                                                else if (Rs >= 530) {
                                                    if (ssDCval < 55) {
                                                        if (posBalance < 52) {
                                                            if (SC < 2341) {
                                                                if (gchDC < 2) {
                                                                    return 4;
                                                                }
                                                                else if (gchDC >= 2) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (SC >= 2341) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (posBalance >= 52) {
                                                            if (Rs < 1652) {
                                                                if (AFD < 8) {
                                                                    return 1;
                                                                }
                                                                else if (AFD >= 8) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (Rs >= 1652) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                    else if (ssDCval >= 55) {
                                                        if (AFD < 11) {
                                                            if (mcTcor < 93) {
                                                                return 4;
                                                            }
                                                            else if (mcTcor >= 93) {
                                                                if (MVDiff < 9) {
                                                                    if (CsDiff < 3) {
                                                                        if (Rs < 1296) {
                                                                            return 2;
                                                                        }
                                                                        else if (Rs >= 1296) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                    else if (CsDiff >= 3) {
                                                                        return 2;
                                                                    }
                                                                }
                                                                else if (MVDiff >= 9) {
                                                                    if (MVDiff < 12) {
                                                                        return 4;
                                                                    }
                                                                    else if (MVDiff >= 12) {
                                                                        return 1;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (AFD >= 11) {
                                                            return 2;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (ssDCval >= 88) {
                                                if (refDCval < 126) {
                                                    if (posBalance < 60) {
                                                        if (Rs < 1059) {
                                                            if (Cs < 275) {
                                                                return 2;
                                                            }
                                                            else if (Cs >= 275) {
                                                                if (TSC < 7) {
                                                                    return 4;
                                                                }
                                                                else if (TSC >= 7) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                        else if (Rs >= 1059) {
                                                            if (refDCval < 120) {
                                                                return 2;
                                                            }
                                                            else if (refDCval >= 120) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                    else if (posBalance >= 60) {
                                                        if (TSCindex < 8) {
                                                            return 4;
                                                        }
                                                        else if (TSCindex >= 8) {
                                                            if (diffRsCsdiff < -1) {
                                                                return 4;
                                                            }
                                                            else if (diffRsCsdiff >= -1) {
                                                                if (Rs < 673) {
                                                                    if (Cs < 428) {
                                                                        return 4;
                                                                    }
                                                                    else if (Cs >= 428) {
                                                                        return 8;
                                                                    }
                                                                }
                                                                else if (Rs >= 673) {
                                                                    if (Cs < 963) {
                                                                        return 1;
                                                                    }
                                                                    else if (Cs >= 963) {
                                                                        if (Rs < 1784) {
                                                                            return 4;
                                                                        }
                                                                        else if (Rs >= 1784) {
                                                                            return 2;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (refDCval >= 126) {
                                                    if (Rs < 2292) {
                                                        return 1;
                                                    }
                                                    else if (Rs >= 2292) {
                                                        return 2;
                                                    }
                                                }
                                            }
                                        }
                                        else if (negBalance >= 52) {
                                            if (refDCval < 97) {
                                                if (CsDiff < 3) {
                                                    if (Rs < 630) {
                                                        return 8;
                                                    }
                                                    else if (Rs >= 630) {
                                                        if (Rs < 708) {
                                                            if (Rs < 650) {
                                                                return 4;
                                                            }
                                                            else if (Rs >= 650) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (Rs >= 708) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                                else if (CsDiff >= 3) {
                                                    if (AFD < 9) {
                                                        return 4;
                                                    }
                                                    else if (AFD >= 9) {
                                                        if (ssDCval < 60) {
                                                            if (Cs < 753) {
                                                                return 4;
                                                            }
                                                            else if (Cs >= 753) {
                                                                return 1;
                                                            }
                                                        }
                                                        else if (ssDCval >= 60) {
                                                            if (mcTcor < 82) {
                                                                return 8;
                                                            }
                                                            else if (mcTcor >= 82) {
                                                                if (refDCval < 62) {
                                                                    return 8;
                                                                }
                                                                else if (refDCval >= 62) {
                                                                    if (posBalance < 57) {
                                                                        if (Rs < 1355) {
                                                                            return 2;
                                                                        }
                                                                        else if (Rs >= 1355) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                    else if (posBalance >= 57) {
                                                                        if (posBalance < 59) {
                                                                            if (MVDiff < 5) {
                                                                                return 8;
                                                                            }
                                                                            else if (MVDiff >= 5) {
                                                                                return 2;
                                                                            }
                                                                        }
                                                                        else if (posBalance >= 59) {
                                                                            return 2;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                            else if (refDCval >= 97) {
                                                if (SC < 1085) {
                                                    if (TSC < 6) {
                                                        return 8;
                                                    }
                                                    else if (TSC >= 6) {
                                                        if (Rs < 485) {
                                                            return 4;
                                                        }
                                                        else if (Rs >= 485) {
                                                            return 1;
                                                        }
                                                    }
                                                }
                                                else if (SC >= 1085) {
                                                    if (posBalance < 59) {
                                                        if (AFD < 10) {
                                                            if (tcor < 77) {
                                                                return 8;
                                                            }
                                                            else if (tcor >= 77) {
                                                                return 2;
                                                            }
                                                        }
                                                        else if (AFD >= 10) {
                                                            if (TSC < 6) {
                                                                if (RsDiff < 7) {
                                                                    return 8;
                                                                }
                                                                else if (RsDiff >= 7) {
                                                                    return 2;
                                                                }
                                                            }
                                                            else if (TSC >= 6) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                    else if (posBalance >= 59) {
                                                        if (AFD < 20) {
                                                            return 4;
                                                        }
                                                        else if (AFD >= 20) {
                                                            return 2;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (MVDiff >= 15) {
                                    if (Cs < 270) {
                                        if (negBalance < 33) {
                                            if (tcor < 67) {
                                                if (ssDCval < 30) {
                                                    if (tcor < 54) {
                                                        if (diffRsCsdiff < -1341) {
                                                            return 1;
                                                        }
                                                        else if (diffRsCsdiff >= -1341) {
                                                            if (diffMVdiffVal < 314) {
                                                                if (diffMVdiffVal < -1) {
                                                                    if (diffMVdiffVal < -595) {
                                                                        return 4;
                                                                    }
                                                                    else if (diffMVdiffVal >= -595) {
                                                                        if (diffAFD < -15) {
                                                                            return 2;
                                                                        }
                                                                        else if (diffAFD >= -15) {
                                                                            if (AFD < 3) {
                                                                                if (AFD < 1) {
                                                                                    if (posBalance < 28) {
                                                                                        if (mcTcor < 50) {
                                                                                            return 1;
                                                                                        }
                                                                                        else if (mcTcor >= 50) {
                                                                                            if (diffMVdiffVal < -47) {
                                                                                                return 8;
                                                                                            }
                                                                                            else if (diffMVdiffVal >= -47) {
                                                                                                return 4;
                                                                                            }
                                                                                        }
                                                                                    }
                                                                                    else if (posBalance >= 28) {
                                                                                        return 4;
                                                                                    }
                                                                                }
                                                                                else if (AFD >= 1) {
                                                                                    return 1;
                                                                                }
                                                                            }
                                                                            else if (AFD >= 3) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                                else if (diffMVdiffVal >= -1) {
                                                                    if (MVDiff < 61) {
                                                                        if (MVDiff < 59) {
                                                                            if (posBalance < 25) {
                                                                                return 1;
                                                                            }
                                                                            else if (posBalance >= 25) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                        else if (MVDiff >= 59) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                    else if (MVDiff >= 61) {
                                                                        if (posBalance < 110) {
                                                                            if (tcor < 28) {
                                                                                return 4;
                                                                            }
                                                                            else if (tcor >= 28) {
                                                                                if (mcTcor < 56) {
                                                                                    if (posBalance < 27) {
                                                                                        return 8;
                                                                                    }
                                                                                    else if (posBalance >= 27) {
                                                                                        if (posBalance < 28) {
                                                                                            return 1;
                                                                                        }
                                                                                        else if (posBalance >= 28) {
                                                                                            if (MVDiff < 85) {
                                                                                                if (tcor < 36) {
                                                                                                    return 8;
                                                                                                }
                                                                                                else if (tcor >= 36) {
                                                                                                    return 4;
                                                                                                }
                                                                                            }
                                                                                            else if (MVDiff >= 85) {
                                                                                                return 8;
                                                                                            }
                                                                                        }
                                                                                    }
                                                                                }
                                                                                else if (mcTcor >= 56) {
                                                                                    return 4;
                                                                                }
                                                                            }
                                                                        }
                                                                        else if (posBalance >= 110) {
                                                                            if (diffAFD < 0) {
                                                                                return 1;
                                                                            }
                                                                            else if (diffAFD >= 0) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                            else if (diffMVdiffVal >= 314) {
                                                                if (TSC < 1) {
                                                                    return 2;
                                                                }
                                                                else if (TSC >= 1) {
                                                                    if (SC < 30) {
                                                                        return 2;
                                                                    }
                                                                    else if (SC >= 30) {
                                                                        return 1;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (tcor >= 54) {
                                                        if (mu_mv_mag_sq < 9) {
                                                            if (Cs < 2) {
                                                                return 1;
                                                            }
                                                            else if (Cs >= 2) {
                                                                return 2;
                                                            }
                                                        }
                                                        else if (mu_mv_mag_sq >= 9) {
                                                            if (ssDCval < 29) {
                                                                if (diffTSC < 1) {
                                                                    return 4;
                                                                }
                                                                else if (diffTSC >= 1) {
                                                                    if (posBalance < 89) {
                                                                        return 2;
                                                                    }
                                                                    else if (posBalance >= 89) {
                                                                        return 4;
                                                                    }
                                                                }
                                                            }
                                                            else if (ssDCval >= 29) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (ssDCval >= 30) {
                                                    if (MVDiff < 26) {
                                                        return 2;
                                                    }
                                                    else if (MVDiff >= 26) {
                                                        if (posBalance < 110) {
                                                            return 1;
                                                        }
                                                        else if (posBalance >= 110) {
                                                            if (Rs < 25) {
                                                                return 2;
                                                            }
                                                            else if (Rs >= 25) {
                                                                return 1;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                            else if (tcor >= 67) {
                                                if (mcTcor < 76) {
                                                    if (MVDiff < 149) {
                                                        return 8;
                                                    }
                                                    else if (MVDiff >= 149) {
                                                        return 2;
                                                    }
                                                }
                                                else if (mcTcor >= 76) {
                                                    if (negBalance < 23) {
                                                        if (mu_mv_mag_sq < 33) {
                                                            if (posBalance < 80) {
                                                                return 2;
                                                            }
                                                            else if (posBalance >= 80) {
                                                                if (posBalance < 111) {
                                                                    if (MVDiff < 84) {
                                                                        if (Cs < 3) {
                                                                            return 4;
                                                                        }
                                                                        else if (Cs >= 3) {
                                                                            if (Rs < 161) {
                                                                                return 8;
                                                                            }
                                                                            else if (Rs >= 161) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (MVDiff >= 84) {
                                                                        return 8;
                                                                    }
                                                                }
                                                                else if (posBalance >= 111) {
                                                                    if (Cs < 9) {
                                                                        return 2;
                                                                    }
                                                                    else if (Cs >= 9) {
                                                                        return 4;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (mu_mv_mag_sq >= 33) {
                                                            if (Cs < 66) {
                                                                return 4;
                                                            }
                                                            else if (Cs >= 66) {
                                                                if (Rs < 79) {
                                                                    return 2;
                                                                }
                                                                else if (Rs >= 79) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (negBalance >= 23) {
                                                        return 2;
                                                    }
                                                }
                                            }
                                        }
                                        else if (negBalance >= 33) {
                                            if (TSC < 5) {
                                                if (ssDCval < 65) {
                                                    if (AFD < 6) {
                                                        if (diffRsCsdiff < -3) {
                                                            if (Cs < 139) {
                                                                return 8;
                                                            }
                                                            else if (Cs >= 139) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (diffRsCsdiff >= -3) {
                                                            if (Rs < 5) {
                                                                if (AFD < 2) {
                                                                    return 2;
                                                                }
                                                                else if (AFD >= 2) {
                                                                    if (diffMVdiffVal < 166) {
                                                                        return 1;
                                                                    }
                                                                    else if (diffMVdiffVal >= 166) {
                                                                        return 2;
                                                                    }
                                                                }
                                                            }
                                                            else if (Rs >= 5) {
                                                                return 2;
                                                            }
                                                        }
                                                    }
                                                    else if (AFD >= 6) {
                                                        return 1;
                                                    }
                                                }
                                                else if (ssDCval >= 65) {
                                                    if (Cs < 141) {
                                                        return 2;
                                                    }
                                                    else if (Cs >= 141) {
                                                        return 1;
                                                    }
                                                }
                                            }
                                            else if (TSC >= 5) {
                                                if (refDCval < 75) {
                                                    if (Scindex < 2) {
                                                        return 1;
                                                    }
                                                    else if (Scindex >= 2) {
                                                        if (RsDiff < 2) {
                                                            return 1;
                                                        }
                                                        else if (RsDiff >= 2) {
                                                            return 2;
                                                        }
                                                    }
                                                }
                                                else if (refDCval >= 75) {
                                                    if (ssDCval < 108) {
                                                        if (diffMVdiffVal < 17) {
                                                            if (tcor < 65) {
                                                                return 2;
                                                            }
                                                            else if (tcor >= 65) {
                                                                return 1;
                                                            }
                                                        }
                                                        else if (diffMVdiffVal >= 17) {
                                                            return 2;
                                                        }
                                                    }
                                                    else if (ssDCval >= 108) {
                                                        if (Cs < 153) {
                                                            return 2;
                                                        }
                                                        else if (Cs >= 153) {
                                                            if (SC < 579) {
                                                                return 4;
                                                            }
                                                            else if (SC >= 579) {
                                                                return 1;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (Cs >= 270) {
                                        if (CsDiff < 3) {
                                            if (AFD < 6) {
                                                if (Rs < 1767) {
                                                    if (Cs < 383) {
                                                        if (mcTcor < 85) {
                                                            if (SC < 730) {
                                                                if (posBalance < 56) {
                                                                    if (diffTSC < 1) {
                                                                        return 8;
                                                                    }
                                                                    else if (diffTSC >= 1) {
                                                                        return 1;
                                                                    }
                                                                }
                                                                else if (posBalance >= 56) {
                                                                    if (tcor < 80) {
                                                                        return 8;
                                                                    }
                                                                    else if (tcor >= 80) {
                                                                        return 4;
                                                                    }
                                                                }
                                                            }
                                                            else if (SC >= 730) {
                                                                if (diffRsCsdiff < 1) {
                                                                    if (Cs < 372) {
                                                                        if (Rs < 402) {
                                                                            return 4;
                                                                        }
                                                                        else if (Rs >= 402) {
                                                                            if (negBalance < 41) {
                                                                                return 4;
                                                                            }
                                                                            else if (negBalance >= 41) {
                                                                                if (SC < 802) {
                                                                                    return 1;
                                                                                }
                                                                                else if (SC >= 802) {
                                                                                    if (Rs < 449) {
                                                                                        return 4;
                                                                                    }
                                                                                    else if (Rs >= 449) {
                                                                                        return 1;
                                                                                    }
                                                                                }
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (Cs >= 372) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (diffRsCsdiff >= 1) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                        else if (mcTcor >= 85) {
                                                            if (Rs < 520) {
                                                                return 4;
                                                            }
                                                            else if (Rs >= 520) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                    else if (Cs >= 383) {
                                                        if (posBalance < 54) {
                                                            if (CsDiff < 2) {
                                                                return 4;
                                                            }
                                                            else if (CsDiff >= 2) {
                                                                if (Rs < 984) {
                                                                    if (Cs < 726) {
                                                                        return 2;
                                                                    }
                                                                    else if (Cs >= 726) {
                                                                        return 8;
                                                                    }
                                                                }
                                                                else if (Rs >= 984) {
                                                                    return 1;
                                                                }
                                                            }
                                                        }
                                                        else if (posBalance >= 54) {
                                                            if (posBalance < 58) {
                                                                if (TSC < 5) {
                                                                    if (Cs < 400) {
                                                                        if (posBalance < 56) {
                                                                            if (SC < 846) {
                                                                                if (negBalance < 43) {
                                                                                    return 2;
                                                                                }
                                                                                else if (negBalance >= 43) {
                                                                                    return 4;
                                                                                }
                                                                            }
                                                                            else if (SC >= 846) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                        else if (posBalance >= 56) {
                                                                            return 2;
                                                                        }
                                                                    }
                                                                    else if (Cs >= 400) {
                                                                        if (RsCsDiff < 11) {
                                                                            if (diffMVdiffVal < 13) {
                                                                                return 2;
                                                                            }
                                                                            else if (diffMVdiffVal >= 13) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                        else if (RsCsDiff >= 11) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                }
                                                                else if (TSC >= 5) {
                                                                    if (Rs < 677) {
                                                                        if (Rs < 445) {
                                                                            return 4;
                                                                        }
                                                                        else if (Rs >= 445) {
                                                                            return 2;
                                                                        }
                                                                    }
                                                                    else if (Rs >= 677) {
                                                                        if (negBalance < 47) {
                                                                            return 4;
                                                                        }
                                                                        else if (negBalance >= 47) {
                                                                            if (diffTSC < 1) {
                                                                                return 1;
                                                                            }
                                                                            else if (diffTSC >= 1) {
                                                                                if (SC < 2133) {
                                                                                    return 4;
                                                                                }
                                                                                else if (SC >= 2133) {
                                                                                    return 8;
                                                                                }
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                            else if (posBalance >= 58) {
                                                                if (diffRsCsdiff < -2) {
                                                                    if (Cs < 876) {
                                                                        return 1;
                                                                    }
                                                                    else if (Cs >= 876) {
                                                                        return 2;
                                                                    }
                                                                }
                                                                else if (diffRsCsdiff >= -2) {
                                                                    if (diffMVdiffVal < -4) {
                                                                        if (SC < 1371) {
                                                                            return 1;
                                                                        }
                                                                        else if (SC >= 1371) {
                                                                            return 2;
                                                                        }
                                                                    }
                                                                    else if (diffMVdiffVal >= -4) {
                                                                        return 2;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (Rs >= 1767) {
                                                    if (mu_mv_mag_sq < 65) {
                                                        if (Cs < 1546) {
                                                            if (negBalance < 45) {
                                                                if (SC < 3364) {
                                                                    return 1;
                                                                }
                                                                else if (SC >= 3364) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (negBalance >= 45) {
                                                                if (Rs < 1855) {
                                                                    if (diffRsCsdiff < -2) {
                                                                        if (Cs < 1507) {
                                                                            return 1;
                                                                        }
                                                                        else if (Cs >= 1507) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                    else if (diffRsCsdiff >= -2) {
                                                                        if (SC < 3381) {
                                                                            return 4;
                                                                        }
                                                                        else if (SC >= 3381) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                }
                                                                else if (Rs >= 1855) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                        else if (Cs >= 1546) {
                                                            return 1;
                                                        }
                                                    }
                                                    else if (mu_mv_mag_sq >= 65) {
                                                        if (Cs < 1503) {
                                                            if (Rs < 1884) {
                                                                return 8;
                                                            }
                                                            else if (Rs >= 1884) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (Cs >= 1503) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (AFD >= 6) {
                                                if (SC < 1143) {
                                                    if (refDCval < 110) {
                                                        if (ssDCval < 98) {
                                                            if (tcor < 90) {
                                                                if (Cs < 447) {
                                                                    if (negBalance < 46) {
                                                                        if (negBalance < 34) {
                                                                            if (Cs < 349) {
                                                                                return 4;
                                                                            }
                                                                            else if (Cs >= 349) {
                                                                                return 2;
                                                                            }
                                                                        }
                                                                        else if (negBalance >= 34) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                    else if (negBalance >= 46) {
                                                                        if (Rs < 183) {
                                                                            return 4;
                                                                        }
                                                                        else if (Rs >= 183) {
                                                                            return 2;
                                                                        }
                                                                    }
                                                                }
                                                                else if (Cs >= 447) {
                                                                    if (Rs < 570) {
                                                                        return 4;
                                                                    }
                                                                    else if (Rs >= 570) {
                                                                        return 8;
                                                                    }
                                                                }
                                                            }
                                                            else if (tcor >= 90) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (ssDCval >= 98) {
                                                            if (RsDiff < 3) {
                                                                return 8;
                                                            }
                                                            else if (RsDiff >= 3) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                    else if (refDCval >= 110) {
                                                        if (ssDCval < 108) {
                                                            return 2;
                                                        }
                                                        else if (ssDCval >= 108) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                                else if (SC >= 1143) {
                                                    if (ssDCval < 63) {
                                                        if (posBalance < 54) {
                                                            if (Cs < 879) {
                                                                if (mcTcor < 93) {
                                                                    return 8;
                                                                }
                                                                else if (mcTcor >= 93) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (Cs >= 879) {
                                                                return 1;
                                                            }
                                                        }
                                                        else if (posBalance >= 54) {
                                                            if (MVDiff < 28) {
                                                                if (posBalance < 57) {
                                                                    if (mcTcor < 78) {
                                                                        if (Rs < 793) {
                                                                            return 1;
                                                                        }
                                                                        else if (Rs >= 793) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                    else if (mcTcor >= 78) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (posBalance >= 57) {
                                                                    if (ssDCval < 56) {
                                                                        if (Rs < 809) {
                                                                            if (mu_mv_mag_sq < 54) {
                                                                                if (SC < 1434) {
                                                                                    if (diffMVdiffVal < -3) {
                                                                                        return 1;
                                                                                    }
                                                                                    else if (diffMVdiffVal >= -3) {
                                                                                        if (Cs < 573) {
                                                                                            return 4;
                                                                                        }
                                                                                        else if (Cs >= 573) {
                                                                                            return 1;
                                                                                        }
                                                                                    }
                                                                                }
                                                                                else if (SC >= 1434) {
                                                                                    return 4;
                                                                                }
                                                                            }
                                                                            else if (mu_mv_mag_sq >= 54) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                        else if (Rs >= 809) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                    else if (ssDCval >= 56) {
                                                                        return 1;
                                                                    }
                                                                }
                                                            }
                                                            else if (MVDiff >= 28) {
                                                                if (mu_mv_mag_sq < 28) {
                                                                    return 4;
                                                                }
                                                                else if (mu_mv_mag_sq >= 28) {
                                                                    if (TSC < 6) {
                                                                        if (mcTcor < 81) {
                                                                            if (negBalance < 51) {
                                                                                if (mu_mv_mag_sq < 33) {
                                                                                    return 4;
                                                                                }
                                                                                else if (mu_mv_mag_sq >= 33) {
                                                                                    if (mu_mv_mag_sq < 42) {
                                                                                        if (AFD < 7) {
                                                                                            return 1;
                                                                                        }
                                                                                        else if (AFD >= 7) {
                                                                                            return 4;
                                                                                        }
                                                                                    }
                                                                                    else if (mu_mv_mag_sq >= 42) {
                                                                                        return 4;
                                                                                    }
                                                                                }
                                                                            }
                                                                            else if (negBalance >= 51) {
                                                                                return 1;
                                                                            }
                                                                        }
                                                                        else if (mcTcor >= 81) {
                                                                            if (posBalance < 57) {
                                                                                if (TSC < 5) {
                                                                                    return 1;
                                                                                }
                                                                                else if (TSC >= 5) {
                                                                                    return 4;
                                                                                }
                                                                            }
                                                                            else if (posBalance >= 57) {
                                                                                if (ssDCval < 40) {
                                                                                    return 1;
                                                                                }
                                                                                else if (ssDCval >= 40) {
                                                                                    return 4;
                                                                                }
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (TSC >= 6) {
                                                                        if (mcTcor < 74) {
                                                                            if (Cs < 563) {
                                                                                return 4;
                                                                            }
                                                                            else if (Cs >= 563) {
                                                                                return 1;
                                                                            }
                                                                        }
                                                                        else if (mcTcor >= 74) {
                                                                            if (negBalance < 49) {
                                                                                if (SC < 1367) {
                                                                                    return 1;
                                                                                }
                                                                                else if (SC >= 1367) {
                                                                                    return 4;
                                                                                }
                                                                            }
                                                                            else if (negBalance >= 49) {
                                                                                if (diffMVdiffVal < -10) {
                                                                                    if (Cs < 559) {
                                                                                        return 4;
                                                                                    }
                                                                                    else if (Cs >= 559) {
                                                                                        return 1;
                                                                                    }
                                                                                }
                                                                                else if (diffMVdiffVal >= -10) {
                                                                                    return 1;
                                                                                }
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (ssDCval >= 63) {
                                                        if (Cs < 363) {
                                                            return 1;
                                                        }
                                                        else if (Cs >= 363) {
                                                            if (TSC < 5) {
                                                                return 2;
                                                            }
                                                            else if (TSC >= 5) {
                                                                if (RsDiff < 4) {
                                                                    return 4;
                                                                }
                                                                else if (RsDiff >= 4) {
                                                                    if (tcor < 70) {
                                                                        return 2;
                                                                    }
                                                                    else if (tcor >= 70) {
                                                                        return 4;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (CsDiff >= 3) {
                                            if (Rs < 995) {
                                                if (AFD < 9) {
                                                    if (TSCindex < 6) {
                                                        if (MVDiff < 23) {
                                                            if (negBalance < 35) {
                                                                return 2;
                                                            }
                                                            else if (negBalance >= 35) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (MVDiff >= 23) {
                                                            return 2;
                                                        }
                                                    }
                                                    else if (TSCindex >= 6) {
                                                        if (Rs < 760) {
                                                            return 8;
                                                        }
                                                        else if (Rs >= 760) {
                                                            if (SC < 1469) {
                                                                return 2;
                                                            }
                                                            else if (SC >= 1469) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (AFD >= 9) {
                                                    if (SC < 1677) {
                                                        return 2;
                                                    }
                                                    else if (SC >= 1677) {
                                                        if (RsDiff < 5) {
                                                            if (SC < 1683) {
                                                                return 1;
                                                            }
                                                            else if (SC >= 1683) {
                                                                if (Cs < 864) {
                                                                    return 8;
                                                                }
                                                                else if (Cs >= 864) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                        else if (RsDiff >= 5) {
                                                            if (MVDiff < 16) {
                                                                return 1;
                                                            }
                                                            else if (MVDiff >= 16) {
                                                                return 2;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                            else if (Rs >= 995) {
                                                if (Cs < 1052) {
                                                    if (ssDCval < 62) {
                                                        if (ssDCval < 49) {
                                                            return 2;
                                                        }
                                                        else if (ssDCval >= 49) {
                                                            return 1;
                                                        }
                                                    }
                                                    else if (ssDCval >= 62) {
                                                        if (RsDiff < 5) {
                                                            if (Rs < 1101) {
                                                                if (gchDC < 1) {
                                                                    return 4;
                                                                }
                                                                else if (gchDC >= 1) {
                                                                    return 1;
                                                                }
                                                            }
                                                            else if (Rs >= 1101) {
                                                                if (ssDCval < 83) {
                                                                    return 4;
                                                                }
                                                                else if (ssDCval >= 83) {
                                                                    return 2;
                                                                }
                                                            }
                                                        }
                                                        else if (RsDiff >= 5) {
                                                            if (negBalance < 46) {
                                                                if (Cs < 595) {
                                                                    return 2;
                                                                }
                                                                else if (Cs >= 595) {
                                                                    return 1;
                                                                }
                                                            }
                                                            else if (negBalance >= 46) {
                                                                return 2;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (Cs >= 1052) {
                                                    if (mcTcor < 89) {
                                                        if (refDCval < 80) {
                                                            if (TSC < 7) {
                                                                return 1;
                                                            }
                                                            else if (TSC >= 7) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (refDCval >= 80) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (mcTcor >= 89) {
                                                        if (Cs < 1383) {
                                                            if (Cs < 1126) {
                                                                return 8;
                                                            }
                                                            else if (Cs >= 1126) {
                                                                return 2;
                                                            }
                                                        }
                                                        else if (Cs >= 1383) {
                                                            if (posBalance < 46) {
                                                                return 4;
                                                            }
                                                            else if (posBalance >= 46) {
                                                                if (refDCval < 147) {
                                                                    if (diffAFD < -1) {
                                                                        return 1;
                                                                    }
                                                                    else if (diffAFD >= -1) {
                                                                        if (diffTSC < -1) {
                                                                            return 8;
                                                                        }
                                                                        else if (diffTSC >= -1) {
                                                                            if (diffAFD < 1) {
                                                                                return 4;
                                                                            }
                                                                            else if (diffAFD >= 1) {
                                                                                return 8;
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                                else if (refDCval >= 147) {
                                                                    return 2;
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                            else if (TSCindex >= 9) {
                                if (Rs < 1719) {
                                    if (RsDiff < 7) {
                                        if (MVDiff < 30) {
                                            if (gchDC < 2) {
                                                if (mu_mv_mag_sq < 5) {
                                                    if (tcor < 88) {
                                                        if (AFD < 13) {
                                                            if (Cs < 888) {
                                                                return 8;
                                                            }
                                                            else if (Cs >= 888) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (AFD >= 13) {
                                                            if (Rs < 1082) {
                                                                return 2;
                                                            }
                                                            else if (Rs >= 1082) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                    else if (tcor >= 88) {
                                                        if (Rs < 1591) {
                                                            return 2;
                                                        }
                                                        else if (Rs >= 1591) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                                else if (mu_mv_mag_sq >= 5) {
                                                    if (SC < 3418) {
                                                        if (mcTcor < 71) {
                                                            if (Rs < 1177) {
                                                                return 4;
                                                            }
                                                            else if (Rs >= 1177) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (mcTcor >= 71) {
                                                            if (mu_mv_mag_sq < 44) {
                                                                if (MVDiff < 14) {
                                                                    if (RsCsDiff < 33) {
                                                                        return 8;
                                                                    }
                                                                    else if (RsCsDiff >= 33) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (MVDiff >= 14) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (mu_mv_mag_sq >= 44) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                    else if (SC >= 3418) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                            else if (gchDC >= 2) {
                                                if (ssDCval < 95) {
                                                    if (RsCsDiff < 39) {
                                                        return 1;
                                                    }
                                                    else if (RsCsDiff >= 39) {
                                                        return 2;
                                                    }
                                                }
                                                else if (ssDCval >= 95) {
                                                    return 4;
                                                }
                                            }
                                        }
                                        else if (MVDiff >= 30) {
                                            if (refDCval < 69) {
                                                if (refDCval < 63) {
                                                    if (SC < 1060) {
                                                        if (ssDCval < 67) {
                                                            return 1;
                                                        }
                                                        else if (ssDCval >= 67) {
                                                            return 2;
                                                        }
                                                    }
                                                    else if (SC >= 1060) {
                                                        if (Rs < 485) {
                                                            return 4;
                                                        }
                                                        else if (Rs >= 485) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                                else if (refDCval >= 63) {
                                                    return 8;
                                                }
                                            }
                                            else if (refDCval >= 69) {
                                                if (gchDC < 6) {
                                                    if (ssDCval < 138) {
                                                        return 1;
                                                    }
                                                    else if (ssDCval >= 138) {
                                                        return 2;
                                                    }
                                                }
                                                else if (gchDC >= 6) {
                                                    if (Cs < 401) {
                                                        if (posBalance < 21) {
                                                            return 1;
                                                        }
                                                        else if (posBalance >= 21) {
                                                            return 2;
                                                        }
                                                    }
                                                    else if (Cs >= 401) {
                                                        if (TSC < 11) {
                                                            return 8;
                                                        }
                                                        else if (TSC >= 11) {
                                                            if (TSC < 18) {
                                                                return 1;
                                                            }
                                                            else if (TSC >= 18) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (RsDiff >= 7) {
                                        if (SC < 2754) {
                                            if (diffAFD < 5) {
                                                if (Rs < 836) {
                                                    return 1;
                                                }
                                                else if (Rs >= 836) {
                                                    if (ssDCval < 73) {
                                                        if (mu_mv_mag_sq < 6) {
                                                            if (TSC < 10) {
                                                                return 4;
                                                            }
                                                            else if (TSC >= 10) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (mu_mv_mag_sq >= 6) {
                                                            if (Rs < 1415) {
                                                                return 2;
                                                            }
                                                            else if (Rs >= 1415) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                    else if (ssDCval >= 73) {
                                                        return 2;
                                                    }
                                                }
                                            }
                                            else if (diffAFD >= 5) {
                                                if (diffRsCsdiff < 130) {
                                                    if (AFD < 18) {
                                                        return 8;
                                                    }
                                                    else if (AFD >= 18) {
                                                        return 1;
                                                    }
                                                }
                                                else if (diffRsCsdiff >= 130) {
                                                    if (RsCsDiff < 229) {
                                                        return 2;
                                                    }
                                                    else if (RsCsDiff >= 229) {
                                                        if (ssDCval < 99) {
                                                            return 8;
                                                        }
                                                        else if (ssDCval >= 99) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (SC >= 2754) {
                                            if (CsDiff < 11) {
                                                if (SC < 3253) {
                                                    if (AFD < 15) {
                                                        if (mu_mv_mag_sq < 3) {
                                                            if (MVDiff < 13) {
                                                                return 1;
                                                            }
                                                            else if (MVDiff >= 13) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (mu_mv_mag_sq >= 3) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (AFD >= 15) {
                                                        if (mu_mv_mag_sq < 5) {
                                                            if (TSC < 12) {
                                                                return 2;
                                                            }
                                                            else if (TSC >= 12) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (mu_mv_mag_sq >= 5) {
                                                            if (RsDiff < 10) {
                                                                return 4;
                                                            }
                                                            else if (RsDiff >= 10) {
                                                                return 1;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (SC >= 3253) {
                                                    return 4;
                                                }
                                            }
                                            else if (CsDiff >= 11) {
                                                if (negBalance < 58) {
                                                    return 1;
                                                }
                                                else if (negBalance >= 58) {
                                                    return 2;
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (Rs >= 1719) {
                                    if (RsCsDiff < 108) {
                                        if (ssDCval < 112) {
                                            if (Rs < 1847) {
                                                if (RsCsDiff < 94) {
                                                    if (Rs < 1736) {
                                                        return 2;
                                                    }
                                                    else if (Rs >= 1736) {
                                                        return 8;
                                                    }
                                                }
                                                else if (RsCsDiff >= 94) {
                                                    if (SC < 3207) {
                                                        return 2;
                                                    }
                                                    else if (SC >= 3207) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                            else if (Rs >= 1847) {
                                                if (Rs < 2397) {
                                                    if (TSC < 10) {
                                                        if (SC < 3319) {
                                                            return 1;
                                                        }
                                                        else if (SC >= 3319) {
                                                            if (posBalance < 58) {
                                                                if (Rs < 2052) {
                                                                    return 2;
                                                                }
                                                                else if (Rs >= 2052) {
                                                                    return 1;
                                                                }
                                                            }
                                                            else if (posBalance >= 58) {
                                                                return 2;
                                                            }
                                                        }
                                                    }
                                                    else if (TSC >= 10) {
                                                        return 1;
                                                    }
                                                }
                                                else if (Rs >= 2397) {
                                                    if (ssDCval < 86) {
                                                        return 4;
                                                    }
                                                    else if (ssDCval >= 86) {
                                                        return 1;
                                                    }
                                                }
                                            }
                                        }
                                        else if (ssDCval >= 112) {
                                            if (mcTcor < 86) {
                                                if (ssDCval < 123) {
                                                    return 4;
                                                }
                                                else if (ssDCval >= 123) {
                                                    return 1;
                                                }
                                            }
                                            else if (mcTcor >= 86) {
                                                if (Rs < 3056) {
                                                    if (RsDiff < 6) {
                                                        return 8;
                                                    }
                                                    else if (RsDiff >= 6) {
                                                        if (MVDiff < 8) {
                                                            if (Rs < 2804) {
                                                                if (ssDCval < 124) {
                                                                    if (RsCsDiff < 99) {
                                                                        return 4;
                                                                    }
                                                                    else if (RsCsDiff >= 99) {
                                                                        if (Rs < 2317) {
                                                                            return 4;
                                                                        }
                                                                        else if (Rs >= 2317) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                }
                                                                else if (ssDCval >= 124) {
                                                                    if (Cs < 1733) {
                                                                        return 4;
                                                                    }
                                                                    else if (Cs >= 1733) {
                                                                        return 8;
                                                                    }
                                                                }
                                                            }
                                                            else if (Rs >= 2804) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (MVDiff >= 8) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                                else if (Rs >= 3056) {
                                                    return 2;
                                                }
                                            }
                                        }
                                    }
                                    else if (RsCsDiff >= 108) {
                                        if (Cs < 1479) {
                                            if (ssDCval < 85) {
                                                if (ssDCval < 79) {
                                                    return 4;
                                                }
                                                else if (ssDCval >= 79) {
                                                    if (Rs < 1931) {
                                                        return 2;
                                                    }
                                                    else if (Rs >= 1931) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                            else if (ssDCval >= 85) {
                                                if (tcor < 78) {
                                                    if (refDCval < 135) {
                                                        if (AFD < 29) {
                                                            if (Rs < 2056) {
                                                                if (refDCval < 115) {
                                                                    return 2;
                                                                }
                                                                else if (refDCval >= 115) {
                                                                    return 1;
                                                                }
                                                            }
                                                            else if (Rs >= 2056) {
                                                                if (TSC < 12) {
                                                                    return 1;
                                                                }
                                                                else if (TSC >= 12) {
                                                                    if (RsCsDiff < 150) {
                                                                        return 1;
                                                                    }
                                                                    else if (RsCsDiff >= 150) {
                                                                        if (diffRsCsdiff < 36) {
                                                                            return 2;
                                                                        }
                                                                        else if (diffRsCsdiff >= 36) {
                                                                            if (Cs < 727) {
                                                                                return 2;
                                                                            }
                                                                            else if (Cs >= 727) {
                                                                                return 1;
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (AFD >= 29) {
                                                            if (mu_mv_mag_sq < 3) {
                                                                if (Rs < 2830) {
                                                                    return 1;
                                                                }
                                                                else if (Rs >= 2830) {
                                                                    return 2;
                                                                }
                                                            }
                                                            else if (mu_mv_mag_sq >= 3) {
                                                                return 2;
                                                            }
                                                        }
                                                    }
                                                    else if (refDCval >= 135) {
                                                        if (tcor < 53) {
                                                            if (Rs < 2598) {
                                                                return 4;
                                                            }
                                                            else if (Rs >= 2598) {
                                                                return 1;
                                                            }
                                                        }
                                                        else if (tcor >= 53) {
                                                            if (gchDC < 6) {
                                                                if (Cs < 503) {
                                                                    return 4;
                                                                }
                                                                else if (Cs >= 503) {
                                                                    if (Rs < 3567) {
                                                                        return 2;
                                                                    }
                                                                    else if (Rs >= 3567) {
                                                                        return 4;
                                                                    }
                                                                }
                                                            }
                                                            else if (gchDC >= 6) {
                                                                return 2;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (tcor >= 78) {
                                                    if (mu_mv_mag_sq < 59) {
                                                        return 2;
                                                    }
                                                    else if (mu_mv_mag_sq >= 59) {
                                                        if (SC < 3227) {
                                                            return 4;
                                                        }
                                                        else if (SC >= 3227) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (Cs >= 1479) {
                                            if (mcTcor < 83) {
                                                if (Cs < 1781) {
                                                    return 1;
                                                }
                                                else if (Cs >= 1781) {
                                                    return 2;
                                                }
                                            }
                                            else if (mcTcor >= 83) {
                                                if (mcTcor < 92) {
                                                    if (tcor < 74) {
                                                        if (SC < 4053) {
                                                            if (negBalance < 59) {
                                                                return 1;
                                                            }
                                                            else if (negBalance >= 59) {
                                                                return 2;
                                                            }
                                                        }
                                                        else if (SC >= 4053) {
                                                            if (ssDCval < 137) {
                                                                return 2;
                                                            }
                                                            else if (ssDCval >= 137) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                    else if (tcor >= 74) {
                                                        if (TSC < 12) {
                                                            if (Rs < 2648) {
                                                                return 4;
                                                            }
                                                            else if (Rs >= 2648) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (TSC >= 12) {
                                                            if (tcor < 82) {
                                                                if (AFD < 18) {
                                                                    return 2;
                                                                }
                                                                else if (AFD >= 18) {
                                                                    if (ssDCval < 124) {
                                                                        return 4;
                                                                    }
                                                                    else if (ssDCval >= 124) {
                                                                        return 2;
                                                                    }
                                                                }
                                                            }
                                                            else if (tcor >= 82) {
                                                                if (CsDiff < 8) {
                                                                    return 8;
                                                                }
                                                                else if (CsDiff >= 8) {
                                                                    return 2;
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (mcTcor >= 92) {
                                                    if (SC < 5237) {
                                                        if (CsDiff < 7) {
                                                            if (diffMVdiffVal < 2) {
                                                                if (negBalance < 49) {
                                                                    return 2;
                                                                }
                                                                else if (negBalance >= 49) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (diffMVdiffVal >= 2) {
                                                                if (RsCsDiff < 170) {
                                                                    return 4;
                                                                }
                                                                else if (RsCsDiff >= 170) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                        else if (CsDiff >= 7) {
                                                            if (diffAFD < 5) {
                                                                if (Rs < 1825) {
                                                                    if (AFD < 17) {
                                                                        return 8;
                                                                    }
                                                                    else if (AFD >= 17) {
                                                                        return 2;
                                                                    }
                                                                }
                                                                else if (Rs >= 1825) {
                                                                    if (diffAFD < -4) {
                                                                        return 8;
                                                                    }
                                                                    else if (diffAFD >= -4) {
                                                                        return 2;
                                                                    }
                                                                }
                                                            }
                                                            else if (diffAFD >= 5) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                    else if (SC >= 5237) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                        }
                        else if (Cs >= 1801) {
                            if (tcor < 79) {
                                if (refDCval < 123) {
                                    if (mcTcor < 87) {
                                        if (TSC < 12) {
                                            if (mu_mv_mag_sq < 5) {
                                                if (Cs < 1829) {
                                                    return 1;
                                                }
                                                else if (Cs >= 1829) {
                                                    if (CsDiff < 14) {
                                                        return 4;
                                                    }
                                                    else if (CsDiff >= 14) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                            else if (mu_mv_mag_sq >= 5) {
                                                if (refDCval < 99) {
                                                    if (negBalance < 43) {
                                                        if (diffRsCsdiff < 28) {
                                                            if (Rs < 3308) {
                                                                return 1;
                                                            }
                                                            else if (Rs >= 3308) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (diffRsCsdiff >= 28) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (negBalance >= 43) {
                                                        if (Cs < 1864) {
                                                            return 2;
                                                        }
                                                        else if (Cs >= 1864) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                                else if (refDCval >= 99) {
                                                    if (posBalance < 47) {
                                                        if (SC < 7095) {
                                                            if (RsCsDiff < 221) {
                                                                return 8;
                                                            }
                                                            else if (RsCsDiff >= 221) {
                                                                return 2;
                                                            }
                                                        }
                                                        else if (SC >= 7095) {
                                                            return 1;
                                                        }
                                                    }
                                                    else if (posBalance >= 47) {
                                                        if (SC < 8516) {
                                                            return 4;
                                                        }
                                                        else if (SC >= 8516) {
                                                            if (Rs < 5424) {
                                                                return 2;
                                                            }
                                                            else if (Rs >= 5424) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (TSC >= 12) {
                                            if (Rs < 4350) {
                                                if (RsDiff < 19) {
                                                    if (mcTcor < 69) {
                                                        if (Rs < 2972) {
                                                            if (SC < 3604) {
                                                                return 1;
                                                            }
                                                            else if (SC >= 3604) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (Rs >= 2972) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (mcTcor >= 69) {
                                                        if (ssDCval < 119) {
                                                            if (diffMVdiffVal < -9) {
                                                                return 1;
                                                            }
                                                            else if (diffMVdiffVal >= -9) {
                                                                if (ssDCval < 80) {
                                                                    if (Rs < 2905) {
                                                                        return 8;
                                                                    }
                                                                    else if (Rs >= 2905) {
                                                                        if (refDCval < 76) {
                                                                            if (AFD < 17) {
                                                                                return 8;
                                                                            }
                                                                            else if (AFD >= 17) {
                                                                                return 1;
                                                                            }
                                                                        }
                                                                        else if (refDCval >= 76) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                }
                                                                else if (ssDCval >= 80) {
                                                                    if (RsDiff < 13) {
                                                                        return 1;
                                                                    }
                                                                    else if (RsDiff >= 13) {
                                                                        if (refDCval < 113) {
                                                                            return 8;
                                                                        }
                                                                        else if (refDCval >= 113) {
                                                                            if (Cs < 1940) {
                                                                                return 8;
                                                                            }
                                                                            else if (Cs >= 1940) {
                                                                                return 1;
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (ssDCval >= 119) {
                                                            if (tcor < 72) {
                                                                if (RsDiff < 16) {
                                                                    return 8;
                                                                }
                                                                else if (RsDiff >= 16) {
                                                                    return 1;
                                                                }
                                                            }
                                                            else if (tcor >= 72) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (RsDiff >= 19) {
                                                    if (mcTcor < 72) {
                                                        return 8;
                                                    }
                                                    else if (mcTcor >= 72) {
                                                        if (SC < 4917) {
                                                            return 8;
                                                        }
                                                        else if (SC >= 4917) {
                                                            return 2;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (Rs >= 4350) {
                                                if (mu_mv_mag_sq < 51) {
                                                    if (Rs < 5430) {
                                                        if (posBalance < 47) {
                                                            return 1;
                                                        }
                                                        else if (posBalance >= 47) {
                                                            return 2;
                                                        }
                                                    }
                                                    else if (Rs >= 5430) {
                                                        return 4;
                                                    }
                                                }
                                                else if (mu_mv_mag_sq >= 51) {
                                                    return 4;
                                                }
                                            }
                                        }
                                    }
                                    else if (mcTcor >= 87) {
                                        if (Rs < 4160) {
                                            if (Cs < 2280) {
                                                if (TSC < 11) {
                                                    if (refDCval < 82) {
                                                        if (Rs < 1820) {
                                                            return 1;
                                                        }
                                                        else if (Rs >= 1820) {
                                                            if (RsDiff < 12) {
                                                                return 4;
                                                            }
                                                            else if (RsDiff >= 12) {
                                                                return 2;
                                                            }
                                                        }
                                                    }
                                                    else if (refDCval >= 82) {
                                                        return 2;
                                                    }
                                                }
                                                else if (TSC >= 11) {
                                                    if (mu_mv_mag_sq < 7) {
                                                        return 8;
                                                    }
                                                    else if (mu_mv_mag_sq >= 7) {
                                                        if (negBalance < 59) {
                                                            return 1;
                                                        }
                                                        else if (negBalance >= 59) {
                                                            if (Cs < 2097) {
                                                                if (Cs < 2068) {
                                                                    return 8;
                                                                }
                                                                else if (Cs >= 2068) {
                                                                    return 2;
                                                                }
                                                            }
                                                            else if (Cs >= 2097) {
                                                                return 1;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                            else if (Cs >= 2280) {
                                                if (mu_mv_mag_sq < 19) {
                                                    if (refDCval < 82) {
                                                        if (mcTcor < 89) {
                                                            return 8;
                                                        }
                                                        else if (mcTcor >= 89) {
                                                            if (Rs < 2746) {
                                                                return 4;
                                                            }
                                                            else if (Rs >= 2746) {
                                                                return 1;
                                                            }
                                                        }
                                                    }
                                                    else if (refDCval >= 82) {
                                                        if (Rs < 2682) {
                                                            return 4;
                                                        }
                                                        else if (Rs >= 2682) {
                                                            if (posBalance < 66) {
                                                                if (tcor < 72) {
                                                                    return 8;
                                                                }
                                                                else if (tcor >= 72) {
                                                                    return 1;
                                                                }
                                                            }
                                                            else if (posBalance >= 66) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (mu_mv_mag_sq >= 19) {
                                                    if (Rs < 3682) {
                                                        return 1;
                                                    }
                                                    else if (Rs >= 3682) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                        }
                                        else if (Rs >= 4160) {
                                            if (Cs < 3015) {
                                                return 4;
                                            }
                                            else if (Cs >= 3015) {
                                                return 2;
                                            }
                                        }
                                    }
                                }
                                else if (refDCval >= 123) {
                                    if (ssDCval < 141) {
                                        if (negBalance < 57) {
                                            if (posBalance < 63) {
                                                if (AFD < 12) {
                                                    if (Rs < 3294) {
                                                        return 4;
                                                    }
                                                    else if (Rs >= 3294) {
                                                        return 1;
                                                    }
                                                }
                                                else if (AFD >= 12) {
                                                    if (ssDCval < 130) {
                                                        if (RsCsDiff < 289) {
                                                            if (Rs < 3624) {
                                                                return 1;
                                                            }
                                                            else if (Rs >= 3624) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (RsCsDiff >= 289) {
                                                            if (Cs < 3214) {
                                                                return 1;
                                                            }
                                                            else if (Cs >= 3214) {
                                                                if (Rs < 3699) {
                                                                    return 8;
                                                                }
                                                                else if (Rs >= 3699) {
                                                                    return 1;
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (ssDCval >= 130) {
                                                        return 1;
                                                    }
                                                }
                                            }
                                            else if (posBalance >= 63) {
                                                if (Cs < 2345) {
                                                    if (RsDiff < 13) {
                                                        return 1;
                                                    }
                                                    else if (RsDiff >= 13) {
                                                        return 2;
                                                    }
                                                }
                                                else if (Cs >= 2345) {
                                                    return 4;
                                                }
                                            }
                                        }
                                        else if (negBalance >= 57) {
                                            if (MVDiff < 24) {
                                                if (Cs < 1974) {
                                                    if (Rs < 2342) {
                                                        return 1;
                                                    }
                                                    else if (Rs >= 2342) {
                                                        return 4;
                                                    }
                                                }
                                                else if (Cs >= 1974) {
                                                    return 1;
                                                }
                                            }
                                            else if (MVDiff >= 24) {
                                                if (AFD < 24) {
                                                    if (Rs < 2940) {
                                                        return 4;
                                                    }
                                                    else if (Rs >= 2940) {
                                                        return 1;
                                                    }
                                                }
                                                else if (AFD >= 24) {
                                                    if (Rs < 2648) {
                                                        return 2;
                                                    }
                                                    else if (Rs >= 2648) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (ssDCval >= 141) {
                                        if (Rs < 2040) {
                                            return 8;
                                        }
                                        else if (Rs >= 2040) {
                                            if (TSC < 10) {
                                                if (SC < 4554) {
                                                    return 4;
                                                }
                                                else if (SC >= 4554) {
                                                    return 2;
                                                }
                                            }
                                            else if (TSC >= 10) {
                                                if (RsDiff < 14) {
                                                    if (diffRsCsdiff < -16) {
                                                        return 4;
                                                    }
                                                    else if (diffRsCsdiff >= -16) {
                                                        if (Cs < 2525) {
                                                            return 8;
                                                        }
                                                        else if (Cs >= 2525) {
                                                            return 2;
                                                        }
                                                    }
                                                }
                                                else if (RsDiff >= 14) {
                                                    if (SC < 4674) {
                                                        return 2;
                                                    }
                                                    else if (SC >= 4674) {
                                                        if (mu_mv_mag_sq < 5) {
                                                            if (negBalance < 57) {
                                                                return 2;
                                                            }
                                                            else if (negBalance >= 57) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (mu_mv_mag_sq >= 5) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                            else if (tcor >= 79) {
                                if (ssDCval < 94) {
                                    if (negBalance < 48) {
                                        if (CsDiff < 10) {
                                            if (posBalance < 54) {
                                                if (MVDiff < 26) {
                                                    if (TSC < 6) {
                                                        if (MVDiff < 13) {
                                                            return 4;
                                                        }
                                                        else if (MVDiff >= 13) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (TSC >= 6) {
                                                        if (CsDiff < 9) {
                                                            if (diffTSC < 2) {
                                                                if (mcTcor < 93) {
                                                                    return 8;
                                                                }
                                                                else if (mcTcor >= 93) {
                                                                    if (Scindex < 5) {
                                                                        return 8;
                                                                    }
                                                                    else if (Scindex >= 5) {
                                                                        if (SC < 4413) {
                                                                            if (MVDiff < 6) {
                                                                                return 8;
                                                                            }
                                                                            else if (MVDiff >= 6) {
                                                                                return 1;
                                                                            }
                                                                        }
                                                                        else if (SC >= 4413) {
                                                                            if (negBalance < 46) {
                                                                                if (Rs < 2295) {
                                                                                    if (ssDCval < 88) {
                                                                                        return 1;
                                                                                    }
                                                                                    else if (ssDCval >= 88) {
                                                                                        return 8;
                                                                                    }
                                                                                }
                                                                                else if (Rs >= 2295) {
                                                                                    return 8;
                                                                                }
                                                                            }
                                                                            else if (negBalance >= 46) {
                                                                                return 1;
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                            else if (diffTSC >= 2) {
                                                                if (Rs < 2633) {
                                                                    return 1;
                                                                }
                                                                else if (Rs >= 2633) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                        else if (CsDiff >= 9) {
                                                            if (mu_mv_mag_sq < 44) {
                                                                if (tcor < 89) {
                                                                    if (refDCval < 80) {
                                                                        return 8;
                                                                    }
                                                                    else if (refDCval >= 80) {
                                                                        if (TSC < 13) {
                                                                            return 1;
                                                                        }
                                                                        else if (TSC >= 13) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                }
                                                                else if (tcor >= 89) {
                                                                    if (RsDiff < 10) {
                                                                        return 8;
                                                                    }
                                                                    else if (RsDiff >= 10) {
                                                                        return 1;
                                                                    }
                                                                }
                                                            }
                                                            else if (mu_mv_mag_sq >= 44) {
                                                                return 1;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (MVDiff >= 26) {
                                                    if (posBalance < 47) {
                                                        if (Rs < 2267) {
                                                            return 1;
                                                        }
                                                        else if (Rs >= 2267) {
                                                            return 2;
                                                        }
                                                    }
                                                    else if (posBalance >= 47) {
                                                        if (Rs < 2576) {
                                                            if (mcTcor < 91) {
                                                                return 4;
                                                            }
                                                            else if (mcTcor >= 91) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (Rs >= 2576) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (posBalance >= 54) {
                                                if (mu_mv_mag_sq < 51) {
                                                    if (Rs < 3122) {
                                                        if (refDCval < 91) {
                                                            if (tcor < 82) {
                                                                return 2;
                                                            }
                                                            else if (tcor >= 82) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (refDCval >= 91) {
                                                            return 2;
                                                        }
                                                    }
                                                    else if (Rs >= 3122) {
                                                        return 8;
                                                    }
                                                }
                                                else if (mu_mv_mag_sq >= 51) {
                                                    if (Rs < 2641) {
                                                        if (Cs < 2007) {
                                                            if (CsDiff < 7) {
                                                                return 1;
                                                            }
                                                            else if (CsDiff >= 7) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (Cs >= 2007) {
                                                            if (RsCsDiff < 129) {
                                                                return 4;
                                                            }
                                                            else if (RsCsDiff >= 129) {
                                                                return 2;
                                                            }
                                                        }
                                                    }
                                                    else if (Rs >= 2641) {
                                                        if (refDCval < 87) {
                                                            return 8;
                                                        }
                                                        else if (refDCval >= 87) {
                                                            return 1;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (CsDiff >= 10) {
                                            if (posBalance < 55) {
                                                if (Cs < 2917) {
                                                    if (mu_mv_mag_sq < 28) {
                                                        if (Cs < 1944) {
                                                            if (Cs < 1856) {
                                                                return 1;
                                                            }
                                                            else if (Cs >= 1856) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (Cs >= 1944) {
                                                            return 1;
                                                        }
                                                    }
                                                    else if (mu_mv_mag_sq >= 28) {
                                                        if (posBalance < 53) {
                                                            if (tcor < 90) {
                                                                return 1;
                                                            }
                                                            else if (tcor >= 90) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (posBalance >= 53) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                                else if (Cs >= 2917) {
                                                    if (RsCsDiff < 232) {
                                                        if (SC < 6137) {
                                                            return 1;
                                                        }
                                                        else if (SC >= 6137) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (RsCsDiff >= 232) {
                                                        if (Rs < 3218) {
                                                            return 8;
                                                        }
                                                        else if (Rs >= 3218) {
                                                            return 1;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (posBalance >= 55) {
                                                if (refDCval < 86) {
                                                    return 8;
                                                }
                                                else if (refDCval >= 86) {
                                                    if (mcTcor < 93) {
                                                        if (SC < 5256) {
                                                            return 4;
                                                        }
                                                        else if (SC >= 5256) {
                                                            return 1;
                                                        }
                                                    }
                                                    else if (mcTcor >= 93) {
                                                        return 1;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (negBalance >= 48) {
                                        if (ssDCval < 78) {
                                            if (Rs < 3218) {
                                                if (mu_mv_mag_sq < 14) {
                                                    if (Cs < 2058) {
                                                        return 8;
                                                    }
                                                    else if (Cs >= 2058) {
                                                        return 4;
                                                    }
                                                }
                                                else if (mu_mv_mag_sq >= 14) {
                                                    if (tcor < 93) {
                                                        return 8;
                                                    }
                                                    else if (tcor >= 93) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                            else if (Rs >= 3218) {
                                                if (refDCval < 55) {
                                                    if (MVDiff < 20) {
                                                        return 4;
                                                    }
                                                    else if (MVDiff >= 20) {
                                                        if (TSCindex < 8) {
                                                            if (AFD < 9) {
                                                                if (SC < 6514) {
                                                                    return 8;
                                                                }
                                                                else if (SC >= 6514) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (AFD >= 9) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (TSCindex >= 8) {
                                                            if (MVDiff < 48) {
                                                                return 2;
                                                            }
                                                            else if (MVDiff >= 48) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (refDCval >= 55) {
                                                    if (TSC < 8) {
                                                        return 2;
                                                    }
                                                    else if (TSC >= 8) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                        }
                                        else if (ssDCval >= 78) {
                                            if (Cs < 2224) {
                                                if (RsCsDiff < 107) {
                                                    if (Rs < 1889) {
                                                        if (ssDCval < 88) {
                                                            return 8;
                                                        }
                                                        else if (ssDCval >= 88) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (Rs >= 1889) {
                                                        if (RsCsDiff < 57) {
                                                            if (Rs < 2410) {
                                                                return 2;
                                                            }
                                                            else if (Rs >= 2410) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (RsCsDiff >= 57) {
                                                            if (MVDiff < 25) {
                                                                if (RsDiff < 6) {
                                                                    return 2;
                                                                }
                                                                else if (RsDiff >= 6) {
                                                                    if (mu_mv_mag_sq < 48) {
                                                                        if (MVDiff < 14) {
                                                                            if (diffMVdiffVal < -6) {
                                                                                return 2;
                                                                            }
                                                                            else if (diffMVdiffVal >= -6) {
                                                                                return 1;
                                                                            }
                                                                        }
                                                                        else if (MVDiff >= 14) {
                                                                            if (refDCval < 82) {
                                                                                return 1;
                                                                            }
                                                                            else if (refDCval >= 82) {
                                                                                if (SC < 4906) {
                                                                                    return 2;
                                                                                }
                                                                                else if (SC >= 4906) {
                                                                                    return 1;
                                                                                }
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (mu_mv_mag_sq >= 48) {
                                                                        return 1;
                                                                    }
                                                                }
                                                            }
                                                            else if (MVDiff >= 25) {
                                                                return 2;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (RsCsDiff >= 107) {
                                                    if (AFD < 15) {
                                                        if (mu_mv_mag_sq < 46) {
                                                            if (Rs < 2689) {
                                                                if (mcTcor < 92) {
                                                                    if (Rs < 2327) {
                                                                        return 2;
                                                                    }
                                                                    else if (Rs >= 2327) {
                                                                        return 1;
                                                                    }
                                                                }
                                                                else if (mcTcor >= 92) {
                                                                    return 2;
                                                                }
                                                            }
                                                            else if (Rs >= 2689) {
                                                                return 1;
                                                            }
                                                        }
                                                        else if (mu_mv_mag_sq >= 46) {
                                                            return 1;
                                                        }
                                                    }
                                                    else if (AFD >= 15) {
                                                        if (MVDiff < 25) {
                                                            return 2;
                                                        }
                                                        else if (MVDiff >= 25) {
                                                            if (SC < 4889) {
                                                                if (Rs < 2586) {
                                                                    return 8;
                                                                }
                                                                else if (Rs >= 2586) {
                                                                    return 1;
                                                                }
                                                            }
                                                            else if (SC >= 4889) {
                                                                return 2;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                            else if (Cs >= 2224) {
                                                if (SC < 5381) {
                                                    if (refDCval < 80) {
                                                        return 8;
                                                    }
                                                    else if (refDCval >= 80) {
                                                        if (diffAFD < -2) {
                                                            if (posBalance < 50) {
                                                                if (Cs < 2303) {
                                                                    return 1;
                                                                }
                                                                else if (Cs >= 2303) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (posBalance >= 50) {
                                                                return 2;
                                                            }
                                                        }
                                                        else if (diffAFD >= -2) {
                                                            if (posBalance < 55) {
                                                                if (RsCsDiff < 201) {
                                                                    return 4;
                                                                }
                                                                else if (RsCsDiff >= 201) {
                                                                    if (Rs < 2756) {
                                                                        if (Rs < 2615) {
                                                                            return 4;
                                                                        }
                                                                        else if (Rs >= 2615) {
                                                                            return 2;
                                                                        }
                                                                    }
                                                                    else if (Rs >= 2756) {
                                                                        return 1;
                                                                    }
                                                                }
                                                            }
                                                            else if (posBalance >= 55) {
                                                                if (SC < 5090) {
                                                                    if (TSC < 11) {
                                                                        if (ssDCval < 86) {
                                                                            return 4;
                                                                        }
                                                                        else if (ssDCval >= 86) {
                                                                            if (RsCsDiff < 173) {
                                                                                return 2;
                                                                            }
                                                                            else if (RsCsDiff >= 173) {
                                                                                if (RsDiff < 11) {
                                                                                    return 4;
                                                                                }
                                                                                else if (RsDiff >= 11) {
                                                                                    return 2;
                                                                                }
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (TSC >= 11) {
                                                                        if (MVDiff < 27) {
                                                                            if (SC < 5067) {
                                                                                return 2;
                                                                            }
                                                                            else if (SC >= 5067) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                        else if (MVDiff >= 27) {
                                                                            if (Rs < 2537) {
                                                                                return 4;
                                                                            }
                                                                            else if (Rs >= 2537) {
                                                                                return 8;
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                                else if (SC >= 5090) {
                                                                    if (SC < 5184) {
                                                                        if (Rs < 2820) {
                                                                            return 4;
                                                                        }
                                                                        else if (Rs >= 2820) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                    else if (SC >= 5184) {
                                                                        if (Rs < 2940) {
                                                                            return 4;
                                                                        }
                                                                        else if (Rs >= 2940) {
                                                                            if (diffMVdiffVal < 0) {
                                                                                return 1;
                                                                            }
                                                                            else if (diffMVdiffVal >= 0) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (SC >= 5381) {
                                                    if (RsDiff < 9) {
                                                        if (mu_mv_mag_sq < 69) {
                                                            if (Cs < 2373) {
                                                                return 8;
                                                            }
                                                            else if (Cs >= 2373) {
                                                                return 1;
                                                            }
                                                        }
                                                        else if (mu_mv_mag_sq >= 69) {
                                                            if (SC < 5525) {
                                                                return 4;
                                                            }
                                                            else if (SC >= 5525) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                    else if (RsDiff >= 9) {
                                                        if (negBalance < 54) {
                                                            if (ssDCval < 85) {
                                                                if (Cs < 2822) {
                                                                    if (AFD < 13) {
                                                                        return 8;
                                                                    }
                                                                    else if (AFD >= 13) {
                                                                        return 1;
                                                                    }
                                                                }
                                                                else if (Cs >= 2822) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (ssDCval >= 85) {
                                                                if (SC < 5543) {
                                                                    return 4;
                                                                }
                                                                else if (SC >= 5543) {
                                                                    if (MVDiff < 11) {
                                                                        return 4;
                                                                    }
                                                                    else if (MVDiff >= 11) {
                                                                        return 1;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (negBalance >= 54) {
                                                            if (SC < 5716) {
                                                                if (Rs < 3105) {
                                                                    if (Cs < 2395) {
                                                                        return 4;
                                                                    }
                                                                    else if (Cs >= 2395) {
                                                                        return 1;
                                                                    }
                                                                }
                                                                else if (Rs >= 3105) {
                                                                    if (Cs < 2498) {
                                                                        return 2;
                                                                    }
                                                                    else if (Cs >= 2498) {
                                                                        return 1;
                                                                    }
                                                                }
                                                            }
                                                            else if (SC >= 5716) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (ssDCval >= 94) {
                                    if (refDCval < 129) {
                                        if (TSC < 13) {
                                            if (ssDCval < 118) {
                                                if (posBalance < 57) {
                                                    if (SC < 4875) {
                                                        if (Rs < 2316) {
                                                            if (negBalance < 40) {
                                                                return 8;
                                                            }
                                                            else if (negBalance >= 40) {
                                                                if (MVDiff < 20) {
                                                                    if (AFD < 12) {
                                                                        return 4;
                                                                    }
                                                                    else if (AFD >= 12) {
                                                                        return 2;
                                                                    }
                                                                }
                                                                else if (MVDiff >= 20) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                        else if (Rs >= 2316) {
                                                            if (ssDCval < 103) {
                                                                if (MVDiff < 7) {
                                                                    return 4;
                                                                }
                                                                else if (MVDiff >= 7) {
                                                                    return 1;
                                                                }
                                                            }
                                                            else if (ssDCval >= 103) {
                                                                if (TSC < 10) {
                                                                    return 2;
                                                                }
                                                                else if (TSC >= 10) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (SC >= 4875) {
                                                        if (Rs < 4519) {
                                                            if (Cs < 2530) {
                                                                if (MVDiff < 13) {
                                                                    if (negBalance < 47) {
                                                                        if (refDCval < 108) {
                                                                            if (SC < 5000) {
                                                                                return 1;
                                                                            }
                                                                            else if (SC >= 5000) {
                                                                                if (diffMVdiffVal < -5) {
                                                                                    return 4;
                                                                                }
                                                                                else if (diffMVdiffVal >= -5) {
                                                                                    if (diffRsCsdiff < -14) {
                                                                                        if (Rs < 3292) {
                                                                                            return 8;
                                                                                        }
                                                                                        else if (Rs >= 3292) {
                                                                                            return 4;
                                                                                        }
                                                                                    }
                                                                                    else if (diffRsCsdiff >= -14) {
                                                                                        return 8;
                                                                                    }
                                                                                }
                                                                            }
                                                                        }
                                                                        else if (refDCval >= 108) {
                                                                            if (Rs < 3183) {
                                                                                return 8;
                                                                            }
                                                                            else if (Rs >= 3183) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (negBalance >= 47) {
                                                                        if (diffTSC < 1) {
                                                                            return 4;
                                                                        }
                                                                        else if (diffTSC >= 1) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                }
                                                                else if (MVDiff >= 13) {
                                                                    if (mcTcor < 94) {
                                                                        if (AFD < 11) {
                                                                            return 4;
                                                                        }
                                                                        else if (AFD >= 11) {
                                                                            if (mu_mv_mag_sq < 24) {
                                                                                return 4;
                                                                            }
                                                                            else if (mu_mv_mag_sq >= 24) {
                                                                                return 8;
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (mcTcor >= 94) {
                                                                        if (Rs < 3145) {
                                                                            if (Cs < 2299) {
                                                                                return 2;
                                                                            }
                                                                            else if (Cs >= 2299) {
                                                                                return 8;
                                                                            }
                                                                        }
                                                                        else if (Rs >= 3145) {
                                                                            if (posBalance < 47) {
                                                                                if (Scindex < 6) {
                                                                                    if (Cs < 2408) {
                                                                                        if (mu_mv_mag_sq < 70) {
                                                                                            return 4;
                                                                                        }
                                                                                        else if (mu_mv_mag_sq >= 70) {
                                                                                            if (ssDCval < 109) {
                                                                                                return 2;
                                                                                            }
                                                                                            else if (ssDCval >= 109) {
                                                                                                return 4;
                                                                                            }
                                                                                        }
                                                                                    }
                                                                                    else if (Cs >= 2408) {
                                                                                        return 2;
                                                                                    }
                                                                                }
                                                                                else if (Scindex >= 6) {
                                                                                    if (MVDiff < 22) {
                                                                                        return 2;
                                                                                    }
                                                                                    else if (MVDiff >= 22) {
                                                                                        return 4;
                                                                                    }
                                                                                }
                                                                            }
                                                                            else if (posBalance >= 47) {
                                                                                if (SC < 5415) {
                                                                                    return 8;
                                                                                }
                                                                                else if (SC >= 5415) {
                                                                                    return 2;
                                                                                }
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                            else if (Cs >= 2530) {
                                                                if (negBalance < 38) {
                                                                    if (tcor < 82) {
                                                                        return 1;
                                                                    }
                                                                    else if (tcor >= 82) {
                                                                        if (Rs < 3294) {
                                                                            return 4;
                                                                        }
                                                                        else if (Rs >= 3294) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                }
                                                                else if (negBalance >= 38) {
                                                                    if (ssDCval < 100) {
                                                                        return 4;
                                                                    }
                                                                    else if (ssDCval >= 100) {
                                                                        if (TSC < 12) {
                                                                            return 8;
                                                                        }
                                                                        else if (TSC >= 12) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (Rs >= 4519) {
                                                            if (TSC < 10) {
                                                                if (posBalance < 51) {
                                                                    return 4;
                                                                }
                                                                else if (posBalance >= 51) {
                                                                    return 2;
                                                                }
                                                            }
                                                            else if (TSC >= 10) {
                                                                if (CsDiff < 8) {
                                                                    if (MVDiff < 10) {
                                                                        return 4;
                                                                    }
                                                                    else if (MVDiff >= 10) {
                                                                        if (Rs < 5089) {
                                                                            return 2;
                                                                        }
                                                                        else if (Rs >= 5089) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                }
                                                                else if (CsDiff >= 8) {
                                                                    return 2;
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (posBalance >= 57) {
                                                    if (posBalance < 61) {
                                                        if (TSC < 9) {
                                                            if (diffMVdiffVal < 2) {
                                                                if (Rs < 1963) {
                                                                    return 8;
                                                                }
                                                                else if (Rs >= 1963) {
                                                                    return 2;
                                                                }
                                                            }
                                                            else if (diffMVdiffVal >= 2) {
                                                                if (Rs < 3964) {
                                                                    if (diffAFD < -1) {
                                                                        return 2;
                                                                    }
                                                                    else if (diffAFD >= -1) {
                                                                        return 8;
                                                                    }
                                                                }
                                                                else if (Rs >= 3964) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                        else if (TSC >= 9) {
                                                            if (AFD < 12) {
                                                                if (negBalance < 50) {
                                                                    return 8;
                                                                }
                                                                else if (negBalance >= 50) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (AFD >= 12) {
                                                                if (MVDiff < 5) {
                                                                    if (Rs < 3604) {
                                                                        if (RsCsDiff < 121) {
                                                                            return 2;
                                                                        }
                                                                        else if (RsCsDiff >= 121) {
                                                                            if (SC < 5116) {
                                                                                return 2;
                                                                            }
                                                                            else if (SC >= 5116) {
                                                                                return 8;
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (Rs >= 3604) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (MVDiff >= 5) {
                                                                    if (refDCval < 112) {
                                                                        if (Rs < 2535) {
                                                                            if (AFD < 18) {
                                                                                return 4;
                                                                            }
                                                                            else if (AFD >= 18) {
                                                                                return 2;
                                                                            }
                                                                        }
                                                                        else if (Rs >= 2535) {
                                                                            if (negBalance < 45) {
                                                                                if (Cs < 2328) {
                                                                                    return 1;
                                                                                }
                                                                                else if (Cs >= 2328) {
                                                                                    if (SC < 5524) {
                                                                                        return 8;
                                                                                    }
                                                                                    else if (SC >= 5524) {
                                                                                        return 2;
                                                                                    }
                                                                                }
                                                                            }
                                                                            else if (negBalance >= 45) {
                                                                                return 1;
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (refDCval >= 112) {
                                                                        if (CsDiff < 8) {
                                                                            return 4;
                                                                        }
                                                                        else if (CsDiff >= 8) {
                                                                            return 2;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (posBalance >= 61) {
                                                        if (tcor < 91) {
                                                            if (mu_mv_mag_sq < 22) {
                                                                if (Cs < 2763) {
                                                                    if (tcor < 90) {
                                                                        return 2;
                                                                    }
                                                                    else if (tcor >= 90) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (Cs >= 2763) {
                                                                    if (Rs < 2191) {
                                                                        return 2;
                                                                    }
                                                                    else if (Rs >= 2191) {
                                                                        return 4;
                                                                    }
                                                                }
                                                            }
                                                            else if (mu_mv_mag_sq >= 22) {
                                                                if (RsDiff < 8) {
                                                                    return 1;
                                                                }
                                                                else if (RsDiff >= 8) {
                                                                    if (Rs < 4793) {
                                                                        if (diffAFD < -3) {
                                                                            return 2;
                                                                        }
                                                                        else if (diffAFD >= -3) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                    else if (Rs >= 4793) {
                                                                        return 2;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (tcor >= 91) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (ssDCval >= 118) {
                                                if (AFD < 13) {
                                                    if (refDCval < 122) {
                                                        if (TSC < 7) {
                                                            if (SC < 6300) {
                                                                return 2;
                                                            }
                                                            else if (SC >= 6300) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (TSC >= 7) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (refDCval >= 122) {
                                                        if (negBalance < 52) {
                                                            if (RsCsDiff < 107) {
                                                                if (Cs < 1899) {
                                                                    return 4;
                                                                }
                                                                else if (Cs >= 1899) {
                                                                    if (Rs < 2693) {
                                                                        return 8;
                                                                    }
                                                                    else if (Rs >= 2693) {
                                                                        if (posBalance < 59) {
                                                                            return 8;
                                                                        }
                                                                        else if (posBalance >= 59) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                            else if (RsCsDiff >= 107) {
                                                                if (Rs < 4676) {
                                                                    return 4;
                                                                }
                                                                else if (Rs >= 4676) {
                                                                    if (Rs < 4900) {
                                                                        return 2;
                                                                    }
                                                                    else if (Rs >= 4900) {
                                                                        if (Rs < 5126) {
                                                                            return 4;
                                                                        }
                                                                        else if (Rs >= 5126) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (negBalance >= 52) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                                else if (AFD >= 13) {
                                                    if (tcor < 84) {
                                                        if (Cs < 1978) {
                                                            return 4;
                                                        }
                                                        else if (Cs >= 1978) {
                                                            if (Cs < 2546) {
                                                                return 2;
                                                            }
                                                            else if (Cs >= 2546) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                    else if (tcor >= 84) {
                                                        if (diffRsCsdiff < 71) {
                                                            return 8;
                                                        }
                                                        else if (diffRsCsdiff >= 71) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (TSC >= 13) {
                                            if (ssDCval < 105) {
                                                if (Cs < 2364) {
                                                    return 2;
                                                }
                                                else if (Cs >= 2364) {
                                                    return 4;
                                                }
                                            }
                                            else if (ssDCval >= 105) {
                                                if (RsCsDiff < 189) {
                                                    return 1;
                                                }
                                                else if (RsCsDiff >= 189) {
                                                    if (refDCval < 124) {
                                                        if (ssDCval < 119) {
                                                            if (Rs < 2945) {
                                                                return 1;
                                                            }
                                                            else if (Rs >= 2945) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (ssDCval >= 119) {
                                                            if (CsDiff < 12) {
                                                                return 1;
                                                            }
                                                            else if (CsDiff >= 12) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                    else if (refDCval >= 124) {
                                                        if (mcTcor < 89) {
                                                            return 4;
                                                        }
                                                        else if (mcTcor >= 89) {
                                                            if (Rs < 2518) {
                                                                return 2;
                                                            }
                                                            else if (Rs >= 2518) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (refDCval >= 129) {
                                        if (AFD < 19) {
                                            if (TSC < 10) {
                                                if (Cs < 2205) {
                                                    return 4;
                                                }
                                                else if (Cs >= 2205) {
                                                    if (SC < 5811) {
                                                        return 8;
                                                    }
                                                    else if (SC >= 5811) {
                                                        return 2;
                                                    }
                                                }
                                            }
                                            else if (TSC >= 10) {
                                                if (Rs < 5885) {
                                                    return 8;
                                                }
                                                else if (Rs >= 5885) {
                                                    return 1;
                                                }
                                            }
                                        }
                                        else if (AFD >= 19) {
                                            if (RsCsDiff < 379) {
                                                if (TSC < 16) {
                                                    return 2;
                                                }
                                                else if (TSC >= 16) {
                                                    return 1;
                                                }
                                            }
                                            else if (RsCsDiff >= 379) {
                                                if (TSC < 10) {
                                                    return 2;
                                                }
                                                else if (TSC >= 10) {
                                                    return 4;
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                        }
                    }
                    else if (Cs >= 3273) {
                        if (RsDiff < 9) {
                            if (RsCsDiff < 46) {
                                if (MVDiff < 18) {
                                    if (TSCindex < 8) {
                                        if (negBalance < 48) {
                                            if (ssDCval < 55) {
                                                return 4;
                                            }
                                            else if (ssDCval >= 55) {
                                                if (ssDCval < 62) {
                                                    return 8;
                                                }
                                                else if (ssDCval >= 62) {
                                                    return 4;
                                                }
                                            }
                                        }
                                        else if (negBalance >= 48) {
                                            if (Cs < 3847) {
                                                return 4;
                                            }
                                            else if (Cs >= 3847) {
                                                return 2;
                                            }
                                        }
                                    }
                                    else if (TSCindex >= 8) {
                                        if (Rs < 3726) {
                                            if (mu_mv_mag_sq < 35) {
                                                return 4;
                                            }
                                            else if (mu_mv_mag_sq >= 35) {
                                                if (tcor < 90) {
                                                    return 4;
                                                }
                                                else if (tcor >= 90) {
                                                    if (ssDCval < 55) {
                                                        return 2;
                                                    }
                                                    else if (ssDCval >= 55) {
                                                        if (AFD < 9) {
                                                            return 2;
                                                        }
                                                        else if (AFD >= 9) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (Rs >= 3726) {
                                            return 4;
                                        }
                                    }
                                }
                                else if (MVDiff >= 18) {
                                    if (mu_mv_mag_sq < 48) {
                                        if (TSCindex < 8) {
                                            if (tcor < 93) {
                                                return 8;
                                            }
                                            else if (tcor >= 93) {
                                                if (Cs < 3304) {
                                                    return 8;
                                                }
                                                else if (Cs >= 3304) {
                                                    return 4;
                                                }
                                            }
                                        }
                                        else if (TSCindex >= 8) {
                                            if (posBalance < 63) {
                                                if (RsDiff < 4) {
                                                    if (MVDiff < 23) {
                                                        return 2;
                                                    }
                                                    else if (MVDiff >= 23) {
                                                        if (tcor < 89) {
                                                            return 8;
                                                        }
                                                        else if (tcor >= 89) {
                                                            if (Rs < 3478) {
                                                                return 8;
                                                            }
                                                            else if (Rs >= 3478) {
                                                                if (posBalance < 53) {
                                                                    return 8;
                                                                }
                                                                else if (posBalance >= 53) {
                                                                    if (Rs < 3588) {
                                                                        return 2;
                                                                    }
                                                                    else if (Rs >= 3588) {
                                                                        if (AFD < 11) {
                                                                            return 8;
                                                                        }
                                                                        else if (AFD >= 11) {
                                                                            return 2;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (RsDiff >= 4) {
                                                    if (tcor < 86) {
                                                        if (Cs < 3703) {
                                                            return 4;
                                                        }
                                                        else if (Cs >= 3703) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (tcor >= 86) {
                                                        if (Rs < 3485) {
                                                            return 8;
                                                        }
                                                        else if (Rs >= 3485) {
                                                            if (ssDCval < 57) {
                                                                return 2;
                                                            }
                                                            else if (ssDCval >= 57) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                            else if (posBalance >= 63) {
                                                return 8;
                                            }
                                        }
                                    }
                                    else if (mu_mv_mag_sq >= 48) {
                                        if (Cs < 3536) {
                                            return 8;
                                        }
                                        else if (Cs >= 3536) {
                                            if (negBalance < 48) {
                                                return 4;
                                            }
                                            else if (negBalance >= 48) {
                                                if (ssDCval < 59) {
                                                    return 8;
                                                }
                                                else if (ssDCval >= 59) {
                                                    return 4;
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                            else if (RsCsDiff >= 46) {
                                if (mcTcor < 94) {
                                    if (ssDCval < 53) {
                                        return 8;
                                    }
                                    else if (ssDCval >= 53) {
                                        if (MVDiff < 31) {
                                            if (refDCval < 70) {
                                                return 4;
                                            }
                                            else if (refDCval >= 70) {
                                                return 8;
                                            }
                                        }
                                        else if (MVDiff >= 31) {
                                            if (mcTcor < 92) {
                                                return 4;
                                            }
                                            else if (mcTcor >= 92) {
                                                if (Rs < 3808) {
                                                    return 8;
                                                }
                                                else if (Rs >= 3808) {
                                                    return 4;
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (mcTcor >= 94) {
                                    if (AFD < 13) {
                                        return 4;
                                    }
                                    else if (AFD >= 13) {
                                        if (negBalance < 40) {
                                            return 4;
                                        }
                                        else if (negBalance >= 40) {
                                            return 8;
                                        }
                                    }
                                }
                            }
                        }
                        else if (RsDiff >= 9) {
                            if (Rs < 6691) {
                                if (Cs < 4358) {
                                    if (refDCval < 160) {
                                        if (tcor < 73) {
                                            if (RsDiff < 17) {
                                                if (tcor < 46) {
                                                    return 4;
                                                }
                                                else if (tcor >= 46) {
                                                    if (ssDCval < 75) {
                                                        return 8;
                                                    }
                                                    else if (ssDCval >= 75) {
                                                        if (ssDCval < 127) {
                                                            if (MVDiff < 19) {
                                                                if (SC < 7357) {
                                                                    return 8;
                                                                }
                                                                else if (SC >= 7357) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (MVDiff >= 19) {
                                                                if (AFD < 35) {
                                                                    return 1;
                                                                }
                                                                else if (AFD >= 35) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                        else if (ssDCval >= 127) {
                                                            return 1;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (RsDiff >= 17) {
                                                if (AFD < 32) {
                                                    if (ssDCval < 132) {
                                                        return 8;
                                                    }
                                                    else if (ssDCval >= 132) {
                                                        if (Rs < 4966) {
                                                            return 1;
                                                        }
                                                        else if (Rs >= 4966) {
                                                            if (AFD < 21) {
                                                                return 8;
                                                            }
                                                            else if (AFD >= 21) {
                                                                if (RsCsDiff < 673) {
                                                                    if (CsDiff < 15) {
                                                                        return 4;
                                                                    }
                                                                    else if (CsDiff >= 15) {
                                                                        if (RsCsDiff < 638) {
                                                                            return 4;
                                                                        }
                                                                        else if (RsCsDiff >= 638) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                }
                                                                else if (RsCsDiff >= 673) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (AFD >= 32) {
                                                    if (TSC < 20) {
                                                        return 2;
                                                    }
                                                    else if (TSC >= 20) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                        }
                                        else if (tcor >= 73) {
                                            if (mu_mv_mag_sq < 51) {
                                                if (Rs < 5773) {
                                                    if (AFD < 17) {
                                                        if (Rs < 4187) {
                                                            return 8;
                                                        }
                                                        else if (Rs >= 4187) {
                                                            if (diffRsCsdiff < 37) {
                                                                if (SC < 8775) {
                                                                    return 4;
                                                                }
                                                                else if (SC >= 8775) {
                                                                    if (Rs < 5103) {
                                                                        return 8;
                                                                    }
                                                                    else if (Rs >= 5103) {
                                                                        return 4;
                                                                    }
                                                                }
                                                            }
                                                            else if (diffRsCsdiff >= 37) {
                                                                if (Cs < 3712) {
                                                                    return 2;
                                                                }
                                                                else if (Cs >= 3712) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (AFD >= 17) {
                                                        if (ssDCval < 154) {
                                                            if (refDCval < 121) {
                                                                if (negBalance < 54) {
                                                                    if (CsDiff < 12) {
                                                                        return 8;
                                                                    }
                                                                    else if (CsDiff >= 12) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (negBalance >= 54) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (refDCval >= 121) {
                                                                if (negBalance < 56) {
                                                                    return 8;
                                                                }
                                                                else if (negBalance >= 56) {
                                                                    if (tcor < 83) {
                                                                        return 8;
                                                                    }
                                                                    else if (tcor >= 83) {
                                                                        if (SC < 8869) {
                                                                            return 8;
                                                                        }
                                                                        else if (SC >= 8869) {
                                                                            if (SC < 9319) {
                                                                                return 4;
                                                                            }
                                                                            else if (SC >= 9319) {
                                                                                return 8;
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (ssDCval >= 154) {
                                                            if (SC < 8682) {
                                                                return 8;
                                                            }
                                                            else if (SC >= 8682) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (Rs >= 5773) {
                                                    if (Rs < 6306) {
                                                        return 4;
                                                    }
                                                    else if (Rs >= 6306) {
                                                        return 2;
                                                    }
                                                }
                                            }
                                            else if (mu_mv_mag_sq >= 51) {
                                                if (diffRsCsdiff < 64) {
                                                    return 8;
                                                }
                                                else if (diffRsCsdiff >= 64) {
                                                    if (AFD < 15) {
                                                        return 4;
                                                    }
                                                    else if (AFD >= 15) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (refDCval >= 160) {
                                        return 2;
                                    }
                                }
                                else if (Cs >= 4358) {
                                    if (TSC < 16) {
                                        if (mu_mv_mag_sq < 3) {
                                            if (tcor < 82) {
                                                if (SC < 8573) {
                                                    return 1;
                                                }
                                                else if (SC >= 8573) {
                                                    if (RsDiff < 18) {
                                                        return 8;
                                                    }
                                                    else if (RsDiff >= 18) {
                                                        return 2;
                                                    }
                                                }
                                            }
                                            else if (tcor >= 82) {
                                                return 1;
                                            }
                                        }
                                        else if (mu_mv_mag_sq >= 3) {
                                            if (TSC < 11) {
                                                if (Rs < 6214) {
                                                    return 8;
                                                }
                                                else if (Rs >= 6214) {
                                                    return 4;
                                                }
                                            }
                                            else if (TSC >= 11) {
                                                if (Rs < 6455) {
                                                    return 8;
                                                }
                                                else if (Rs >= 6455) {
                                                    if (tcor < 73) {
                                                        return 8;
                                                    }
                                                    else if (tcor >= 73) {
                                                        return 1;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (TSC >= 16) {
                                        if (Rs < 4164) {
                                            if (SC < 8373) {
                                                return 8;
                                            }
                                            else if (SC >= 8373) {
                                                return 1;
                                            }
                                        }
                                        else if (Rs >= 4164) {
                                            return 8;
                                        }
                                    }
                                }
                            }
                            else if (Rs >= 6691) {
                                if (negBalance < 52) {
                                    if (refDCval < 79) {
                                        if (RsCsDiff < 1088) {
                                            if (AFD < 17) {
                                                return 4;
                                            }
                                            else if (AFD >= 17) {
                                                if (negBalance < 47) {
                                                    return 8;
                                                }
                                                else if (negBalance >= 47) {
                                                    if (Cs < 5543) {
                                                        return 8;
                                                    }
                                                    else if (Cs >= 5543) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                        }
                                        else if (RsCsDiff >= 1088) {
                                            return 4;
                                        }
                                    }
                                    else if (refDCval >= 79) {
                                        if (Cs < 4439) {
                                            if (Rs < 7187) {
                                                return 1;
                                            }
                                            else if (Rs >= 7187) {
                                                return 4;
                                            }
                                        }
                                        else if (Cs >= 4439) {
                                            if (mcTcor < 87) {
                                                return 8;
                                            }
                                            else if (mcTcor >= 87) {
                                                if (SC < 15175) {
                                                    return 4;
                                                }
                                                else if (SC >= 15175) {
                                                    if (Rs < 10013) {
                                                        return 2;
                                                    }
                                                    else if (Rs >= 10013) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (negBalance >= 52) {
                                    if (Cs < 6243) {
                                        if (posBalance < 60) {
                                            if (RsDiff < 34) {
                                                if (MVDiff < 23) {
                                                    if (Cs < 5936) {
                                                        return 4;
                                                    }
                                                    else if (Cs >= 5936) {
                                                        if (TSC < 18) {
                                                            return 4;
                                                        }
                                                        else if (TSC >= 18) {
                                                            if (Rs < 7741) {
                                                                return 8;
                                                            }
                                                            else if (Rs >= 7741) {
                                                                return 2;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (MVDiff >= 23) {
                                                    if (ssDCval < 76) {
                                                        return 4;
                                                    }
                                                    else if (ssDCval >= 76) {
                                                        return 2;
                                                    }
                                                }
                                            }
                                            else if (RsDiff >= 34) {
                                                if (SC < 15728) {
                                                    return 2;
                                                }
                                                else if (SC >= 15728) {
                                                    return 4;
                                                }
                                            }
                                        }
                                        else if (posBalance >= 60) {
                                            if (ssDCval < 153) {
                                                if (SC < 14990) {
                                                    return 8;
                                                }
                                                else if (SC >= 14990) {
                                                    if (Cs < 6156) {
                                                        return 4;
                                                    }
                                                    else if (Cs >= 6156) {
                                                        return 2;
                                                    }
                                                }
                                            }
                                            else if (ssDCval >= 153) {
                                                if (Rs < 7804) {
                                                    return 2;
                                                }
                                                else if (Rs >= 7804) {
                                                    if (Cs < 5547) {
                                                        return 8;
                                                    }
                                                    else if (Cs >= 5547) {
                                                        return 2;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (Cs >= 6243) {
                                        if (RsDiff < 57) {
                                            if (gchDC < 1) {
                                                if (mcTcor < 84) {
                                                    if (ssDCval < 129) {
                                                        return 4;
                                                    }
                                                    else if (ssDCval >= 129) {
                                                        if (Rs < 9533) {
                                                            return 8;
                                                        }
                                                        else if (Rs >= 9533) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                                else if (mcTcor >= 84) {
                                                    return 8;
                                                }
                                            }
                                            else if (gchDC >= 1) {
                                                if (gchDC < 4) {
                                                    if (posBalance < 57) {
                                                        if (MVDiff < 17) {
                                                            return 8;
                                                        }
                                                        else if (MVDiff >= 17) {
                                                            if (posBalance < 55) {
                                                                return 8;
                                                            }
                                                            else if (posBalance >= 55) {
                                                                return 2;
                                                            }
                                                        }
                                                    }
                                                    else if (posBalance >= 57) {
                                                        if (mu_mv_mag_sq < 63) {
                                                            if (AFD < 42) {
                                                                if (Rs < 9617) {
                                                                    if (AFD < 30) {
                                                                        return 8;
                                                                    }
                                                                    else if (AFD >= 30) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (Rs >= 9617) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (AFD >= 42) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (mu_mv_mag_sq >= 63) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                                else if (gchDC >= 4) {
                                                    if (Rs < 7409) {
                                                        return 8;
                                                    }
                                                    else if (Rs >= 7409) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                        }
                                        else if (RsDiff >= 57) {
                                            if (Rs < 13169) {
                                                return 2;
                                            }
                                            else if (Rs >= 13169) {
                                                return 4;
                                            }
                                        }
                                    }
                                }
                            }
                        }
                    }
                }
            }
            else if (mu_mv_mag_sq >= 74) {
                if (Cs < 4158) {
                    if (TSC < 6) {
                        if (Scindex < 4) {
                            if (negBalance < 48) {
                                if (mu_mv_mag_sq < 91) {
                                    if (SC < 804) {
                                        if (MVDiff < 19) {
                                            if (Rs < 160) {
                                                return 4;
                                            }
                                            else if (Rs >= 160) {
                                                return 8;
                                            }
                                        }
                                        else if (MVDiff >= 19) {
                                            if (Rs < 420) {
                                                return 4;
                                            }
                                            else if (Rs >= 420) {
                                                return 1;
                                            }
                                        }
                                    }
                                    else if (SC >= 804) {
                                        if (SC < 1457) {
                                            return 4;
                                        }
                                        else if (SC >= 1457) {
                                            return 8;
                                        }
                                    }
                                }
                                else if (mu_mv_mag_sq >= 91) {
                                    if (posBalance < 54) {
                                        if (negBalance < 29) {
                                            if (MVDiff < 9) {
                                                return 8;
                                            }
                                            else if (MVDiff >= 9) {
                                                if (MVDiff < 13) {
                                                    return 1;
                                                }
                                                else if (MVDiff >= 13) {
                                                    if (Rs < 250) {
                                                        return 1;
                                                    }
                                                    else if (Rs >= 250) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                        }
                                        else if (negBalance >= 29) {
                                            if (SC < 1269) {
                                                return 4;
                                            }
                                            else if (SC >= 1269) {
                                                return 8;
                                            }
                                        }
                                    }
                                    else if (posBalance >= 54) {
                                        if (RsDiff < 4) {
                                            if (AFD < 2) {
                                                return 8;
                                            }
                                            else if (AFD >= 2) {
                                                if (posBalance < 97) {
                                                    return 4;
                                                }
                                                else if (posBalance >= 97) {
                                                    return 8;
                                                }
                                            }
                                        }
                                        else if (RsDiff >= 4) {
                                            if (SC < 789) {
                                                return 2;
                                            }
                                            else if (SC >= 789) {
                                                return 1;
                                            }
                                        }
                                    }
                                }
                            }
                            else if (negBalance >= 48) {
                                if (Rs < 370) {
                                    if (Rs < 27) {
                                        return 4;
                                    }
                                    else if (Rs >= 27) {
                                        return 2;
                                    }
                                }
                                else if (Rs >= 370) {
                                    return 1;
                                }
                            }
                        }
                        else if (Scindex >= 4) {
                            if (refDCval < 116) {
                                if (AFD < 7) {
                                    if (MVDiff < 6) {
                                        if (AFD < 6) {
                                            if (mcTcor < 92) {
                                                if (diffTSC < 1) {
                                                    return 8;
                                                }
                                                else if (diffTSC >= 1) {
                                                    if (Rs < 1429) {
                                                        return 1;
                                                    }
                                                    else if (Rs >= 1429) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                            else if (mcTcor >= 92) {
                                                if (refDCval < 53) {
                                                    if (mu_mv_mag_sq < 96) {
                                                        if (negBalance < 34) {
                                                            return 1;
                                                        }
                                                        else if (negBalance >= 34) {
                                                            if (Cs < 897) {
                                                                return 1;
                                                            }
                                                            else if (Cs >= 897) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                    else if (mu_mv_mag_sq >= 96) {
                                                        if (AFD < 5) {
                                                            if (posBalance < 33) {
                                                                return 8;
                                                            }
                                                            else if (posBalance >= 33) {
                                                                if (Rs < 1893) {
                                                                    return 8;
                                                                }
                                                                else if (Rs >= 1893) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                        else if (AFD >= 5) {
                                                            if (SC < 4408) {
                                                                return 4;
                                                            }
                                                            else if (SC >= 4408) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (refDCval >= 53) {
                                                    if (mu_mv_mag_sq < 106) {
                                                        if (MVDiff < 2) {
                                                            return 1;
                                                        }
                                                        else if (MVDiff >= 2) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (mu_mv_mag_sq >= 106) {
                                                        if (tcor < 88) {
                                                            if (ssDCval < 116) {
                                                                return 8;
                                                            }
                                                            else if (ssDCval >= 116) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (tcor >= 88) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (AFD >= 6) {
                                            if (Rs < 1845) {
                                                if (SC < 2423) {
                                                    return 1;
                                                }
                                                else if (SC >= 2423) {
                                                    return 8;
                                                }
                                            }
                                            else if (Rs >= 1845) {
                                                if (Rs < 2104) {
                                                    if (mu_mv_mag_sq < 112) {
                                                        return 4;
                                                    }
                                                    else if (mu_mv_mag_sq >= 112) {
                                                        return 8;
                                                    }
                                                }
                                                else if (Rs >= 2104) {
                                                    if (tcor < 94) {
                                                        return 8;
                                                    }
                                                    else if (tcor >= 94) {
                                                        if (RsCsDiff < 30) {
                                                            return 4;
                                                        }
                                                        else if (RsCsDiff >= 30) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (MVDiff >= 6) {
                                        if (negBalance < 40) {
                                            if (posBalance < 15) {
                                                if (Cs < 1076) {
                                                    return 4;
                                                }
                                                else if (Cs >= 1076) {
                                                    return 8;
                                                }
                                            }
                                            else if (posBalance >= 15) {
                                                if (posBalance < 83) {
                                                    if (mu_mv_mag_sq < 83) {
                                                        if (Rs < 2396) {
                                                            if (Cs < 996) {
                                                                return 1;
                                                            }
                                                            else if (Cs >= 996) {
                                                                if (Cs < 1105) {
                                                                    return 4;
                                                                }
                                                                else if (Cs >= 1105) {
                                                                    return 2;
                                                                }
                                                            }
                                                        }
                                                        else if (Rs >= 2396) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (mu_mv_mag_sq >= 83) {
                                                        if (diffMVdiffVal < 21) {
                                                            if (diffTSC < -1) {
                                                                if (MVDiff < 8) {
                                                                    return 1;
                                                                }
                                                                else if (MVDiff >= 8) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (diffTSC >= -1) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (diffMVdiffVal >= 21) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                                else if (posBalance >= 83) {
                                                    return 4;
                                                }
                                            }
                                        }
                                        else if (negBalance >= 40) {
                                            if (SC < 3342) {
                                                if (refDCval < 53) {
                                                    return 8;
                                                }
                                                else if (refDCval >= 53) {
                                                    if (Rs < 1142) {
                                                        return 8;
                                                    }
                                                    else if (Rs >= 1142) {
                                                        return 1;
                                                    }
                                                }
                                            }
                                            else if (SC >= 3342) {
                                                if (mu_mv_mag_sq < 98) {
                                                    if (posBalance < 53) {
                                                        return 1;
                                                    }
                                                    else if (posBalance >= 53) {
                                                        return 4;
                                                    }
                                                }
                                                else if (mu_mv_mag_sq >= 98) {
                                                    return 8;
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (AFD >= 7) {
                                    if (CsDiff < 6) {
                                        if (Rs < 2163) {
                                            return 4;
                                        }
                                        else if (Rs >= 2163) {
                                            return 8;
                                        }
                                    }
                                    else if (CsDiff >= 6) {
                                        return 1;
                                    }
                                }
                            }
                            else if (refDCval >= 116) {
                                if (diffMVdiffVal < -3) {
                                    if (Cs < 1535) {
                                        if (ssDCval < 141) {
                                            return 1;
                                        }
                                        else if (ssDCval >= 141) {
                                            if (RsDiff < 7) {
                                                return 8;
                                            }
                                            else if (RsDiff >= 7) {
                                                return 1;
                                            }
                                        }
                                    }
                                    else if (Cs >= 1535) {
                                        return 4;
                                    }
                                }
                                else if (diffMVdiffVal >= -3) {
                                    if (refDCval < 123) {
                                        if (diffRsCsdiff < 5) {
                                            if (diffRsCsdiff < -9) {
                                                return 8;
                                            }
                                            else if (diffRsCsdiff >= -9) {
                                                if (ssDCval < 117) {
                                                    if (Rs < 2360) {
                                                        return 8;
                                                    }
                                                    else if (Rs >= 2360) {
                                                        return 4;
                                                    }
                                                }
                                                else if (ssDCval >= 117) {
                                                    return 4;
                                                }
                                            }
                                        }
                                        else if (diffRsCsdiff >= 5) {
                                            return 8;
                                        }
                                    }
                                    else if (refDCval >= 123) {
                                        if (RsCsDiff < 67) {
                                            if (MVDiff < 7) {
                                                if (diffRsCsdiff < -35) {
                                                    if (RsCsDiff < 11) {
                                                        return 1;
                                                    }
                                                    else if (RsCsDiff >= 11) {
                                                        if (Rs < 1148) {
                                                            return 4;
                                                        }
                                                        else if (Rs >= 1148) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                                else if (diffRsCsdiff >= -35) {
                                                    if (diffAFD < 1) {
                                                        return 8;
                                                    }
                                                    else if (diffAFD >= 1) {
                                                        if (tcor < 78) {
                                                            if (RsCsDiff < 56) {
                                                                return 1;
                                                            }
                                                            else if (RsCsDiff >= 56) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (tcor >= 78) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (MVDiff >= 7) {
                                                if (CsDiff < 3) {
                                                    if (Rs < 881) {
                                                        return 8;
                                                    }
                                                    else if (Rs >= 881) {
                                                        return 1;
                                                    }
                                                }
                                                else if (CsDiff >= 3) {
                                                    if (diffRsCsdiff < 8) {
                                                        return 8;
                                                    }
                                                    else if (diffRsCsdiff >= 8) {
                                                        return 1;
                                                    }
                                                }
                                            }
                                        }
                                        else if (RsCsDiff >= 67) {
                                            if (mu_mv_mag_sq < 90) {
                                                return 4;
                                            }
                                            else if (mu_mv_mag_sq >= 90) {
                                                return 1;
                                            }
                                        }
                                    }
                                }
                            }
                        }
                    }
                    else if (TSC >= 6) {
                        if (RsCsDiff < 33) {
                            if (posBalance < 58) {
                                if (SC < 3789) {
                                    if (mcTcor < 95) {
                                        if (diffRsCsdiff < -1) {
                                            return 8;
                                        }
                                        else if (diffRsCsdiff >= -1) {
                                            if (ssDCval < 72) {
                                                return 2;
                                            }
                                            else if (ssDCval >= 72) {
                                                if (diffAFD < -1) {
                                                    return 2;
                                                }
                                                else if (diffAFD >= -1) {
                                                    if (Rs < 1834) {
                                                        return 8;
                                                    }
                                                    else if (Rs >= 1834) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (mcTcor >= 95) {
                                        if (SC < 3503) {
                                            return 2;
                                        }
                                        else if (SC >= 3503) {
                                            return 4;
                                        }
                                    }
                                }
                                else if (SC >= 3789) {
                                    if (TSCindex < 8) {
                                        if (Rs < 3661) {
                                            if (diffMVdiffVal < -2) {
                                                return 4;
                                            }
                                            else if (diffMVdiffVal >= -2) {
                                                if (SC < 4532) {
                                                    return 4;
                                                }
                                                else if (SC >= 4532) {
                                                    return 8;
                                                }
                                            }
                                        }
                                        else if (Rs >= 3661) {
                                            return 8;
                                        }
                                    }
                                    else if (TSCindex >= 8) {
                                        return 4;
                                    }
                                }
                            }
                            else if (posBalance >= 58) {
                                if (RsCsDiff < 23) {
                                    return 2;
                                }
                                else if (RsCsDiff >= 23) {
                                    if (refDCval < 75) {
                                        if (mu_mv_mag_sq < 113) {
                                            return 2;
                                        }
                                        else if (mu_mv_mag_sq >= 113) {
                                            if (tcor < 91) {
                                                if (RsDiff < 4) {
                                                    return 2;
                                                }
                                                else if (RsDiff >= 4) {
                                                    return 8;
                                                }
                                            }
                                            else if (tcor >= 91) {
                                                if (MVDiff < 1) {
                                                    return 2;
                                                }
                                                else if (MVDiff >= 1) {
                                                    return 8;
                                                }
                                            }
                                        }
                                    }
                                    else if (refDCval >= 75) {
                                        if (ssDCval < 90) {
                                            if (AFD < 7) {
                                                if (negBalance < 44) {
                                                    return 8;
                                                }
                                                else if (negBalance >= 44) {
                                                    if (Rs < 1777) {
                                                        return 8;
                                                    }
                                                    else if (Rs >= 1777) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                            else if (AFD >= 7) {
                                                if (tcor < 94) {
                                                    if (diffAFD < -1) {
                                                        return 4;
                                                    }
                                                    else if (diffAFD >= -1) {
                                                        if (MVDiff < 2) {
                                                            return 2;
                                                        }
                                                        else if (MVDiff >= 2) {
                                                            if (posBalance < 59) {
                                                                return 2;
                                                            }
                                                            else if (posBalance >= 59) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (tcor >= 94) {
                                                    return 2;
                                                }
                                            }
                                        }
                                        else if (ssDCval >= 90) {
                                            if (RsDiff < 5) {
                                                return 8;
                                            }
                                            else if (RsDiff >= 5) {
                                                return 2;
                                            }
                                        }
                                    }
                                }
                            }
                        }
                        else if (RsCsDiff >= 33) {
                            if (mu_mv_mag_sq < 127) {
                                if (mu_mv_mag_sq < 103) {
                                    if (refDCval < 86) {
                                        if (posBalance < 54) {
                                            if (TSC < 8) {
                                                if (ssDCval < 66) {
                                                    if (Cs < 1669) {
                                                        return 4;
                                                    }
                                                    else if (Cs >= 1669) {
                                                        return 8;
                                                    }
                                                }
                                                else if (ssDCval >= 66) {
                                                    if (RsCsDiff < 73) {
                                                        if (Rs < 2016) {
                                                            return 1;
                                                        }
                                                        else if (Rs >= 2016) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (RsCsDiff >= 73) {
                                                        return 1;
                                                    }
                                                }
                                            }
                                            else if (TSC >= 8) {
                                                if (SC < 7087) {
                                                    return 8;
                                                }
                                                else if (SC >= 7087) {
                                                    return 4;
                                                }
                                            }
                                        }
                                        else if (posBalance >= 54) {
                                            if (SC < 3614) {
                                                if (Cs < 1028) {
                                                    if (Cs < 876) {
                                                        if (Cs < 448) {
                                                            return 1;
                                                        }
                                                        else if (Cs >= 448) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (Cs >= 876) {
                                                        return 8;
                                                    }
                                                }
                                                else if (Cs >= 1028) {
                                                    if (SC < 2820) {
                                                        if (Rs < 1387) {
                                                            return 2;
                                                        }
                                                        else if (Rs >= 1387) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (SC >= 2820) {
                                                        return 2;
                                                    }
                                                }
                                            }
                                            else if (SC >= 3614) {
                                                if (ssDCval < 56) {
                                                    return 8;
                                                }
                                                else if (ssDCval >= 56) {
                                                    if (posBalance < 60) {
                                                        if (ssDCval < 83) {
                                                            if (diffRsCsdiff < 29) {
                                                                return 4;
                                                            }
                                                            else if (diffRsCsdiff >= 29) {
                                                                if (Cs < 2764) {
                                                                    return 8;
                                                                }
                                                                else if (Cs >= 2764) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                        else if (ssDCval >= 83) {
                                                            if (refDCval < 85) {
                                                                return 2;
                                                            }
                                                            else if (refDCval >= 85) {
                                                                if (mu_mv_mag_sq < 97) {
                                                                    return 4;
                                                                }
                                                                else if (mu_mv_mag_sq >= 97) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (posBalance >= 60) {
                                                        if (mcTcor < 82) {
                                                            return 4;
                                                        }
                                                        else if (mcTcor >= 82) {
                                                            if (Rs < 1979) {
                                                                return 4;
                                                            }
                                                            else if (Rs >= 1979) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (refDCval >= 86) {
                                        if (ssDCval < 98) {
                                            if (tcor < 93) {
                                                if (negBalance < 49) {
                                                    if (tcor < 89) {
                                                        if (SC < 5514) {
                                                            return 1;
                                                        }
                                                        else if (SC >= 5514) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (tcor >= 89) {
                                                        return 8;
                                                    }
                                                }
                                                else if (negBalance >= 49) {
                                                    if (diffAFD < -3) {
                                                        return 4;
                                                    }
                                                    else if (diffAFD >= -3) {
                                                        if (refDCval < 93) {
                                                            if (tcor < 90) {
                                                                if (Cs < 1167) {
                                                                    return 4;
                                                                }
                                                                else if (Cs >= 1167) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (tcor >= 90) {
                                                                if (diffMVdiffVal < -3) {
                                                                    return 8;
                                                                }
                                                                else if (diffMVdiffVal >= -3) {
                                                                    return 2;
                                                                }
                                                            }
                                                        }
                                                        else if (refDCval >= 93) {
                                                            if (Rs < 2089) {
                                                                return 8;
                                                            }
                                                            else if (Rs >= 2089) {
                                                                if (SC < 4736) {
                                                                    return 2;
                                                                }
                                                                else if (SC >= 4736) {
                                                                    if (Rs < 2991) {
                                                                        return 8;
                                                                    }
                                                                    else if (Rs >= 2991) {
                                                                        return 2;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                            else if (tcor >= 93) {
                                                if (posBalance < 48) {
                                                    if (SC < 4745) {
                                                        return 1;
                                                    }
                                                    else if (SC >= 4745) {
                                                        return 8;
                                                    }
                                                }
                                                else if (posBalance >= 48) {
                                                    if (posBalance < 55) {
                                                        if (Rs < 2366) {
                                                            return 2;
                                                        }
                                                        else if (Rs >= 2366) {
                                                            return 1;
                                                        }
                                                    }
                                                    else if (posBalance >= 55) {
                                                        return 2;
                                                    }
                                                }
                                            }
                                        }
                                        else if (ssDCval >= 98) {
                                            if (Rs < 2962) {
                                                if (AFD < 12) {
                                                    if (SC < 4769) {
                                                        if (ssDCval < 109) {
                                                            return 4;
                                                        }
                                                        else if (ssDCval >= 109) {
                                                            if (diffRsCsdiff < -34) {
                                                                if (diffTSC < -2) {
                                                                    return 4;
                                                                }
                                                                else if (diffTSC >= -2) {
                                                                    if (AFD < 9) {
                                                                        return 4;
                                                                    }
                                                                    else if (AFD >= 9) {
                                                                        if (Cs < 1761) {
                                                                            return 8;
                                                                        }
                                                                        else if (Cs >= 1761) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                            else if (diffRsCsdiff >= -34) {
                                                                if (RsDiff < 10) {
                                                                    if (negBalance < 50) {
                                                                        return 8;
                                                                    }
                                                                    else if (negBalance >= 50) {
                                                                        if (MVDiff < 6) {
                                                                            return 8;
                                                                        }
                                                                        else if (MVDiff >= 6) {
                                                                            if (RsCsDiff < 99) {
                                                                                return 8;
                                                                            }
                                                                            else if (RsCsDiff >= 99) {
                                                                                if (AFD < 11) {
                                                                                    return 4;
                                                                                }
                                                                                else if (AFD >= 11) {
                                                                                    return 8;
                                                                                }
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                                else if (RsDiff >= 10) {
                                                                    if (Cs < 1683) {
                                                                        return 4;
                                                                    }
                                                                    else if (Cs >= 1683) {
                                                                        return 1;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (SC >= 4769) {
                                                        if (MVDiff < 7) {
                                                            return 8;
                                                        }
                                                        else if (MVDiff >= 7) {
                                                            if (refDCval < 98) {
                                                                return 8;
                                                            }
                                                            else if (refDCval >= 98) {
                                                                return 1;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (AFD >= 12) {
                                                    if (refDCval < 133) {
                                                        if (RsDiff < 16) {
                                                            return 1;
                                                        }
                                                        else if (RsDiff >= 16) {
                                                            return 2;
                                                        }
                                                    }
                                                    else if (refDCval >= 133) {
                                                        if (AFD < 19) {
                                                            return 8;
                                                        }
                                                        else if (AFD >= 19) {
                                                            return 2;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (Rs >= 2962) {
                                                if (Cs < 3119) {
                                                    if (ssDCval < 127) {
                                                        if (TSC < 13) {
                                                            if (negBalance < 43) {
                                                                if (refDCval < 121) {
                                                                    if (mu_mv_mag_sq < 76) {
                                                                        if (posBalance < 42) {
                                                                            if (TSC < 7) {
                                                                                return 2;
                                                                            }
                                                                            else if (TSC >= 7) {
                                                                                return 8;
                                                                            }
                                                                        }
                                                                        else if (posBalance >= 42) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                    else if (mu_mv_mag_sq >= 76) {
                                                                        if (negBalance < 38) {
                                                                            if (MVDiff < 13) {
                                                                                if (TSC < 9) {
                                                                                    if (ssDCval < 109) {
                                                                                        if (posBalance < 45) {
                                                                                            return 8;
                                                                                        }
                                                                                        else if (posBalance >= 45) {
                                                                                            if (Rs < 3880) {
                                                                                                return 8;
                                                                                            }
                                                                                            else if (Rs >= 3880) {
                                                                                                return 4;
                                                                                            }
                                                                                        }
                                                                                    }
                                                                                    else if (ssDCval >= 109) {
                                                                                        if (RsDiff < 8) {
                                                                                            if (Cs < 2981) {
                                                                                                if (Cs < 2821) {
                                                                                                    if (Rs < 4278) {
                                                                                                        return 2;
                                                                                                    }
                                                                                                    else if (Rs >= 4278) {
                                                                                                        return 8;
                                                                                                    }
                                                                                                }
                                                                                                else if (Cs >= 2821) {
                                                                                                    return 2;
                                                                                                }
                                                                                            }
                                                                                            else if (Cs >= 2981) {
                                                                                                return 8;
                                                                                            }
                                                                                        }
                                                                                        else if (RsDiff >= 8) {
                                                                                            return 8;
                                                                                        }
                                                                                    }
                                                                                }
                                                                                else if (TSC >= 9) {
                                                                                    return 2;
                                                                                }
                                                                            }
                                                                            else if (MVDiff >= 13) {
                                                                                if (tcor < 92) {
                                                                                    return 4;
                                                                                }
                                                                                else if (tcor >= 92) {
                                                                                    return 2;
                                                                                }
                                                                            }
                                                                        }
                                                                        else if (negBalance >= 38) {
                                                                            if (SC < 7514) {
                                                                                return 8;
                                                                            }
                                                                            else if (SC >= 7514) {
                                                                                if (Rs < 4821) {
                                                                                    return 4;
                                                                                }
                                                                                else if (Rs >= 4821) {
                                                                                    if (Cs < 3031) {
                                                                                        return 8;
                                                                                    }
                                                                                    else if (Cs >= 3031) {
                                                                                        return 2;
                                                                                    }
                                                                                }
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                                else if (refDCval >= 121) {
                                                                    if (mu_mv_mag_sq < 84) {
                                                                        if (Cs < 2485) {
                                                                            if (ssDCval < 120) {
                                                                                return 8;
                                                                            }
                                                                            else if (ssDCval >= 120) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                        else if (Cs >= 2485) {
                                                                            if (ssDCval < 125) {
                                                                                return 8;
                                                                            }
                                                                            else if (ssDCval >= 125) {
                                                                                return 2;
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (mu_mv_mag_sq >= 84) {
                                                                        return 4;
                                                                    }
                                                                }
                                                            }
                                                            else if (negBalance >= 43) {
                                                                if (RsCsDiff < 147) {
                                                                    if (Rs < 3278) {
                                                                        if (Rs < 3081) {
                                                                            if (Rs < 2984) {
                                                                                return 8;
                                                                            }
                                                                            else if (Rs >= 2984) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                        else if (Rs >= 3081) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                    else if (Rs >= 3278) {
                                                                        return 1;
                                                                    }
                                                                }
                                                                else if (RsCsDiff >= 147) {
                                                                    if (SC < 5878) {
                                                                        return 8;
                                                                    }
                                                                    else if (SC >= 5878) {
                                                                        return 4;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (TSC >= 13) {
                                                            return 2;
                                                        }
                                                    }
                                                    else if (ssDCval >= 127) {
                                                        if (SC < 7008) {
                                                            if (Rs < 4117) {
                                                                if (AFD < 21) {
                                                                    return 1;
                                                                }
                                                                else if (AFD >= 21) {
                                                                    return 2;
                                                                }
                                                            }
                                                            else if (Rs >= 4117) {
                                                                if (SC < 6339) {
                                                                    return 4;
                                                                }
                                                                else if (SC >= 6339) {
                                                                    if (SC < 6853) {
                                                                        return 8;
                                                                    }
                                                                    else if (SC >= 6853) {
                                                                        return 2;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (SC >= 7008) {
                                                            return 2;
                                                        }
                                                    }
                                                }
                                                else if (Cs >= 3119) {
                                                    if (posBalance < 59) {
                                                        if (mu_mv_mag_sq < 91) {
                                                            if (AFD < 12) {
                                                                if (RsDiff < 10) {
                                                                    if (refDCval < 111) {
                                                                        return 2;
                                                                    }
                                                                    else if (refDCval >= 111) {
                                                                        return 8;
                                                                    }
                                                                }
                                                                else if (RsDiff >= 10) {
                                                                    return 2;
                                                                }
                                                            }
                                                            else if (AFD >= 12) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (mu_mv_mag_sq >= 91) {
                                                            if (mcTcor < 90) {
                                                                if (Rs < 4584) {
                                                                    return 1;
                                                                }
                                                                else if (Rs >= 4584) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (mcTcor >= 90) {
                                                                if (Cs < 3667) {
                                                                    return 4;
                                                                }
                                                                else if (Cs >= 3667) {
                                                                    return 2;
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (posBalance >= 59) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (mu_mv_mag_sq >= 103) {
                                    if (negBalance < 55) {
                                        if (Cs < 3165) {
                                            if (RsDiff < 11) {
                                                if (SC < 6531) {
                                                    if (SC < 4249) {
                                                        if (mu_mv_mag_sq < 105) {
                                                            if (RsDiff < 7) {
                                                                return 2;
                                                            }
                                                            else if (RsDiff >= 7) {
                                                                if (Rs < 2244) {
                                                                    return 8;
                                                                }
                                                                else if (Rs >= 2244) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                        else if (mu_mv_mag_sq >= 105) {
                                                            if (refDCval < 76) {
                                                                if (mu_mv_mag_sq < 112) {
                                                                    return 2;
                                                                }
                                                                else if (mu_mv_mag_sq >= 112) {
                                                                    if (posBalance < 60) {
                                                                        if (ssDCval < 75) {
                                                                            return 8;
                                                                        }
                                                                        else if (ssDCval >= 75) {
                                                                            if (Cs < 1609) {
                                                                                return 4;
                                                                            }
                                                                            else if (Cs >= 1609) {
                                                                                return 8;
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (posBalance >= 60) {
                                                                        if (mcTcor < 93) {
                                                                            if (Rs < 825) {
                                                                                return 1;
                                                                            }
                                                                            else if (Rs >= 825) {
                                                                                return 8;
                                                                            }
                                                                        }
                                                                        else if (mcTcor >= 93) {
                                                                            return 2;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                            else if (refDCval >= 76) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                    else if (SC >= 4249) {
                                                        if (TSC < 7) {
                                                            if (MVDiff < 1) {
                                                                if (mu_mv_mag_sq < 120) {
                                                                    if (diffTSC < -1) {
                                                                        if (Cs < 2115) {
                                                                            return 4;
                                                                        }
                                                                        else if (Cs >= 2115) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                    else if (diffTSC >= -1) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (mu_mv_mag_sq >= 120) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (MVDiff >= 1) {
                                                                if (diffMVdiffVal < 1) {
                                                                    return 8;
                                                                }
                                                                else if (diffMVdiffVal >= 1) {
                                                                    if (Rs < 2586) {
                                                                        return 4;
                                                                    }
                                                                    else if (Rs >= 2586) {
                                                                        return 8;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (TSC >= 7) {
                                                            if (mu_mv_mag_sq < 106) {
                                                                if (RsCsDiff < 56) {
                                                                    return 2;
                                                                }
                                                                else if (RsCsDiff >= 56) {
                                                                    if (diffRsCsdiff < 69) {
                                                                        if (RsDiff < 7) {
                                                                            if (MVDiff < 2) {
                                                                                return 4;
                                                                            }
                                                                            else if (MVDiff >= 2) {
                                                                                return 8;
                                                                            }
                                                                        }
                                                                        else if (RsDiff >= 7) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                    else if (diffRsCsdiff >= 69) {
                                                                        return 4;
                                                                    }
                                                                }
                                                            }
                                                            else if (mu_mv_mag_sq >= 106) {
                                                                if (Cs < 3045) {
                                                                    return 8;
                                                                }
                                                                else if (Cs >= 3045) {
                                                                    if (CsDiff < 8) {
                                                                        return 2;
                                                                    }
                                                                    else if (CsDiff >= 8) {
                                                                        return 8;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (SC >= 6531) {
                                                    if (TSC < 9) {
                                                        if (AFD < 9) {
                                                            return 2;
                                                        }
                                                        else if (AFD >= 9) {
                                                            return 1;
                                                        }
                                                    }
                                                    else if (TSC >= 9) {
                                                        if (SC < 6796) {
                                                            return 8;
                                                        }
                                                        else if (SC >= 6796) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (RsDiff >= 11) {
                                                if (posBalance < 54) {
                                                    if (Rs < 2669) {
                                                        return 1;
                                                    }
                                                    else if (Rs >= 2669) {
                                                        return 2;
                                                    }
                                                }
                                                else if (posBalance >= 54) {
                                                    return 4;
                                                }
                                            }
                                        }
                                        else if (Cs >= 3165) {
                                            if (Cs < 3579) {
                                                if (MVDiff < 1) {
                                                    if (CsDiff < 9) {
                                                        return 2;
                                                    }
                                                    else if (CsDiff >= 9) {
                                                        if (refDCval < 102) {
                                                            return 2;
                                                        }
                                                        else if (refDCval >= 102) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                                else if (MVDiff >= 1) {
                                                    if (ssDCval < 115) {
                                                        if (Rs < 4531) {
                                                            if (Rs < 3759) {
                                                                return 8;
                                                            }
                                                            else if (Rs >= 3759) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (Rs >= 4531) {
                                                            return 2;
                                                        }
                                                    }
                                                    else if (ssDCval >= 115) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                            else if (Cs >= 3579) {
                                                if (Rs < 5128) {
                                                    return 8;
                                                }
                                                else if (Rs >= 5128) {
                                                    return 2;
                                                }
                                            }
                                        }
                                    }
                                    else if (negBalance >= 55) {
                                        if (Rs < 4507) {
                                            if (RsDiff < 15) {
                                                return 8;
                                            }
                                            else if (RsDiff >= 15) {
                                                return 2;
                                            }
                                        }
                                        else if (Rs >= 4507) {
                                            return 4;
                                        }
                                    }
                                }
                            }
                            else if (mu_mv_mag_sq >= 127) {
                                if (Cs < 2155) {
                                    if (SC < 3272) {
                                        if (AFD < 34) {
                                            if (Cs < 242) {
                                                return 2;
                                            }
                                            else if (Cs >= 242) {
                                                return 1;
                                            }
                                        }
                                        else if (AFD >= 34) {
                                            if (Rs < 768) {
                                                return 4;
                                            }
                                            else if (Rs >= 768) {
                                                return 8;
                                            }
                                        }
                                    }
                                    else if (SC >= 3272) {
                                        if (AFD < 31) {
                                            if (AFD < 17) {
                                                return 8;
                                            }
                                            else if (AFD >= 17) {
                                                return 1;
                                            }
                                        }
                                        else if (AFD >= 31) {
                                            if (refDCval < 128) {
                                                return 2;
                                            }
                                            else if (refDCval >= 128) {
                                                return 4;
                                            }
                                        }
                                    }
                                }
                                else if (Cs >= 2155) {
                                    if (Cs < 4150) {
                                        return 4;
                                    }
                                    else if (Cs >= 4150) {
                                        return 8;
                                    }
                                }
                            }
                        }
                    }
                }
                else if (Cs >= 4158) {
                    if (negBalance < 64) {
                        if (mu_mv_mag_sq < 128) {
                            if (ssDCval < 106) {
                                if (Rs < 5170) {
                                    if (ssDCval < 61) {
                                        return 4;
                                    }
                                    else if (ssDCval >= 61) {
                                        if (Cs < 4876) {
                                            return 8;
                                        }
                                        else if (Cs >= 4876) {
                                            if (TSC < 9) {
                                                if (Rs < 4978) {
                                                    if (CsDiff < 9) {
                                                        if (Rs < 4888) {
                                                            return 8;
                                                        }
                                                        else if (Rs >= 4888) {
                                                            if (Cs < 5008) {
                                                                return 4;
                                                            }
                                                            else if (Cs >= 5008) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                    else if (CsDiff >= 9) {
                                                        return 4;
                                                    }
                                                }
                                                else if (Rs >= 4978) {
                                                    return 4;
                                                }
                                            }
                                            else if (TSC >= 9) {
                                                return 8;
                                            }
                                        }
                                    }
                                }
                                else if (Rs >= 5170) {
                                    return 8;
                                }
                            }
                            else if (ssDCval >= 106) {
                                if (mu_mv_mag_sq < 123) {
                                    if (mcTcor < 73) {
                                        return 1;
                                    }
                                    else if (mcTcor >= 73) {
                                        if (MVDiff < 1) {
                                            if (posBalance < 57) {
                                                return 8;
                                            }
                                            else if (posBalance >= 57) {
                                                if (Cs < 5542) {
                                                    return 8;
                                                }
                                                else if (Cs >= 5542) {
                                                    return 4;
                                                }
                                            }
                                        }
                                        else if (MVDiff >= 1) {
                                            return 8;
                                        }
                                    }
                                }
                                else if (mu_mv_mag_sq >= 123) {
                                    if (Rs < 6264) {
                                        if (posBalance < 60) {
                                            return 4;
                                        }
                                        else if (posBalance >= 60) {
                                            if (Rs < 5862) {
                                                return 8;
                                            }
                                            else if (Rs >= 5862) {
                                                return 4;
                                            }
                                        }
                                    }
                                    else if (Rs >= 6264) {
                                        if (Rs < 7852) {
                                            return 8;
                                        }
                                        else if (Rs >= 7852) {
                                            return 4;
                                        }
                                    }
                                }
                            }
                        }
                        else if (mu_mv_mag_sq >= 128) {
                            if (negBalance < 44) {
                                return 8;
                            }
                            else if (negBalance >= 44) {
                                if (Rs < 5304) {
                                    if (SC < 10178) {
                                        if (ssDCval < 80) {
                                            if (Rs < 4664) {
                                                return 4;
                                            }
                                            else if (Rs >= 4664) {
                                                return 8;
                                            }
                                        }
                                        else if (ssDCval >= 80) {
                                            if (Cs < 5123) {
                                                if (Rs < 4775) {
                                                    if (Rs < 4692) {
                                                        if (Rs < 4623) {
                                                            return 8;
                                                        }
                                                        else if (Rs >= 4623) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (Rs >= 4692) {
                                                        return 8;
                                                    }
                                                }
                                                else if (Rs >= 4775) {
                                                    if (CsDiff < 9) {
                                                        if (Rs < 4821) {
                                                            if (Rs < 4805) {
                                                                return 4;
                                                            }
                                                            else if (Rs >= 4805) {
                                                                if (SC < 9732) {
                                                                    return 4;
                                                                }
                                                                else if (SC >= 9732) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                        else if (Rs >= 4821) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (CsDiff >= 9) {
                                                        if (Cs < 5006) {
                                                            return 4;
                                                        }
                                                        else if (Cs >= 5006) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (Cs >= 5123) {
                                                return 8;
                                            }
                                        }
                                    }
                                    else if (SC >= 10178) {
                                        return 4;
                                    }
                                }
                                else if (Rs >= 5304) {
                                    if (refDCval < 105) {
                                        return 8;
                                    }
                                    else if (refDCval >= 105) {
                                        return 4;
                                    }
                                }
                            }
                        }
                    }
                    else if (negBalance >= 64) {
                        if (Cs < 7581) {
                            if (Cs < 7009) {
                                return 8;
                            }
                            else if (Cs >= 7009) {
                                return 4;
                            }
                        }
                        else if (Cs >= 7581) {
                            if (RsCsDiff < 947) {
                                return 4;
                            }
                            else if (RsCsDiff >= 947) {
                                if (RsCsDiff < 1036) {
                                    return 2;
                                }
                                else if (RsCsDiff >= 1036) {
                                    if (Cs < 8188) {
                                        return 4;
                                    }
                                    else if (Cs >= 8188) {
                                        return 2;
                                    }
                                }
                            }
                        }
                    }
                }
            }
        }
        else if (tcor >= 95) {
            if (MVDiff < 5) {
                if (tcor < 98) {
                    if (TSC < 9) {
                        if (Cs < 2779) {
                            if (negBalance < 35) {
                                if (Rs < 599) {
                                    if (SC < 782) {
                                        if (SC < 484) {
                                            if (refDCval < 60) {
                                                return 1;
                                            }
                                            else if (refDCval >= 60) {
                                                return 4;
                                            }
                                        }
                                        else if (SC >= 484) {
                                            return 8;
                                        }
                                    }
                                    else if (SC >= 782) {
                                        return 4;
                                    }
                                }
                                else if (Rs >= 599) {
                                    if (SC < 3894) {
                                        if (gchDC < 3) {
                                            if (mcTcor < 98) {
                                                if (diffTSC < -1) {
                                                    if (mu_mv_mag_sq < 117) {
                                                        if (Rs < 1321) {
                                                            if (Cs < 600) {
                                                                return 8;
                                                            }
                                                            else if (Cs >= 600) {
                                                                return 1;
                                                            }
                                                        }
                                                        else if (Rs >= 1321) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (mu_mv_mag_sq >= 117) {
                                                        return 8;
                                                    }
                                                }
                                                else if (diffTSC >= -1) {
                                                    return 8;
                                                }
                                            }
                                            else if (mcTcor >= 98) {
                                                if (ssDCval < 112) {
                                                    return 8;
                                                }
                                                else if (ssDCval >= 112) {
                                                    if (posBalance < 16) {
                                                        return 8;
                                                    }
                                                    else if (posBalance >= 16) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                        }
                                        else if (gchDC >= 3) {
                                            return 2;
                                        }
                                    }
                                    else if (SC >= 3894) {
                                        if (mu_mv_mag_sq < 94) {
                                            return 4;
                                        }
                                        else if (mu_mv_mag_sq >= 94) {
                                            if (SC < 4625) {
                                                if (Rs < 2026) {
                                                    return 1;
                                                }
                                                else if (Rs >= 2026) {
                                                    if (refDCval < 93) {
                                                        if (Cs < 1429) {
                                                            return 1;
                                                        }
                                                        else if (Cs >= 1429) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (refDCval >= 93) {
                                                        if (mu_mv_mag_sq < 116) {
                                                            if (Rs < 2289) {
                                                                return 8;
                                                            }
                                                            else if (Rs >= 2289) {
                                                                return 1;
                                                            }
                                                        }
                                                        else if (mu_mv_mag_sq >= 116) {
                                                            return 1;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (SC >= 4625) {
                                                return 8;
                                            }
                                        }
                                    }
                                }
                            }
                            else if (negBalance >= 35) {
                                if (Cs < 1650) {
                                    if (ssDCval < 76) {
                                        if (TSC < 6) {
                                            if (posBalance < 50) {
                                                if (Cs < 509) {
                                                    return 4;
                                                }
                                                else if (Cs >= 509) {
                                                    return 2;
                                                }
                                            }
                                            else if (posBalance >= 50) {
                                                if (Cs < 1571) {
                                                    if (Cs < 450) {
                                                        return 4;
                                                    }
                                                    else if (Cs >= 450) {
                                                        if (diffTSC < -1) {
                                                            if (Cs < 1516) {
                                                                return 8;
                                                            }
                                                            else if (Cs >= 1516) {
                                                                if (Rs < 1812) {
                                                                    return 2;
                                                                }
                                                                else if (Rs >= 1812) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                        else if (diffTSC >= -1) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                                else if (Cs >= 1571) {
                                                    if (SC < 3495) {
                                                        if (SC < 3455) {
                                                            return 8;
                                                        }
                                                        else if (SC >= 3455) {
                                                            if (Rs < 1899) {
                                                                if (SC < 3477) {
                                                                    return 4;
                                                                }
                                                                else if (SC >= 3477) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (Rs >= 1899) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                    else if (SC >= 3495) {
                                                        if (refDCval < 74) {
                                                            return 2;
                                                        }
                                                        else if (refDCval >= 74) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (TSC >= 6) {
                                            if (RsCsDiff < 29) {
                                                return 2;
                                            }
                                            else if (RsCsDiff >= 29) {
                                                return 8;
                                            }
                                        }
                                    }
                                    else if (ssDCval >= 76) {
                                        if (Rs < 2591) {
                                            if (mu_mv_mag_sq < 122) {
                                                if (MVDiff < 3) {
                                                    if (Rs < 1741) {
                                                        if (SC < 2885) {
                                                            if (Rs < 418) {
                                                                return 2;
                                                            }
                                                            else if (Rs >= 418) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (SC >= 2885) {
                                                            if (SC < 3092) {
                                                                return 8;
                                                            }
                                                            else if (SC >= 3092) {
                                                                if (Rs < 1654) {
                                                                    return 8;
                                                                }
                                                                else if (Rs >= 1654) {
                                                                    return 2;
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (Rs >= 1741) {
                                                        if (TSC < 7) {
                                                            return 4;
                                                        }
                                                        else if (TSC >= 7) {
                                                            return 2;
                                                        }
                                                    }
                                                }
                                                else if (MVDiff >= 3) {
                                                    if (ssDCval < 104) {
                                                        return 4;
                                                    }
                                                    else if (ssDCval >= 104) {
                                                        if (Cs < 753) {
                                                            if (SC < 2194) {
                                                                return 4;
                                                            }
                                                            else if (SC >= 2194) {
                                                                return 2;
                                                            }
                                                        }
                                                        else if (Cs >= 753) {
                                                            if (SC < 3298) {
                                                                return 8;
                                                            }
                                                            else if (SC >= 3298) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                            else if (mu_mv_mag_sq >= 122) {
                                                return 8;
                                            }
                                        }
                                        else if (Rs >= 2591) {
                                            if (mu_mv_mag_sq < 101) {
                                                return 1;
                                            }
                                            else if (mu_mv_mag_sq >= 101) {
                                                return 8;
                                            }
                                        }
                                    }
                                }
                                else if (Cs >= 1650) {
                                    if (refDCval < 89) {
                                        if (TSCindex < 5) {
                                            return 8;
                                        }
                                        else if (TSCindex >= 5) {
                                            if (SC < 3494) {
                                                return 8;
                                            }
                                            else if (SC >= 3494) {
                                                if (TSCindex < 7) {
                                                    if (mcTcor < 96) {
                                                        if (posBalance < 57) {
                                                            return 4;
                                                        }
                                                        else if (posBalance >= 57) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (mcTcor >= 96) {
                                                        return 4;
                                                    }
                                                }
                                                else if (TSCindex >= 7) {
                                                    if (refDCval < 86) {
                                                        if (RsDiff < 7) {
                                                            return 4;
                                                        }
                                                        else if (RsDiff >= 7) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (refDCval >= 86) {
                                                        return 2;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (refDCval >= 89) {
                                        if (Rs < 3629) {
                                            if (MVDiff < 3) {
                                                if (mu_mv_mag_sq < 103) {
                                                    if (SC < 4361) {
                                                        return 1;
                                                    }
                                                    else if (SC >= 4361) {
                                                        return 4;
                                                    }
                                                }
                                                else if (mu_mv_mag_sq >= 103) {
                                                    if (mu_mv_mag_sq < 110) {
                                                        if (Cs < 2373) {
                                                            return 2;
                                                        }
                                                        else if (Cs >= 2373) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (mu_mv_mag_sq >= 110) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                            else if (MVDiff >= 3) {
                                                if (Rs < 2504) {
                                                    if (CsDiff < 6) {
                                                        if (Cs < 1713) {
                                                            return 8;
                                                        }
                                                        else if (Cs >= 1713) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (CsDiff >= 6) {
                                                        return 2;
                                                    }
                                                }
                                                else if (Rs >= 2504) {
                                                    if (Rs < 3066) {
                                                        return 1;
                                                    }
                                                    else if (Rs >= 3066) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                        }
                                        else if (Rs >= 3629) {
                                            return 4;
                                        }
                                    }
                                }
                            }
                        }
                        else if (Cs >= 2779) {
                            if (refDCval < 92) {
                                if (tcor < 96) {
                                    if (SC < 7347) {
                                        if (TSC < 7) {
                                            if (RsCsDiff < 11) {
                                                return 8;
                                            }
                                            else if (RsCsDiff >= 11) {
                                                return 4;
                                            }
                                        }
                                        else if (TSC >= 7) {
                                            return 2;
                                        }
                                    }
                                    else if (SC >= 7347) {
                                        return 8;
                                    }
                                }
                                else if (tcor >= 96) {
                                    if (ssDCval < 54) {
                                        return 4;
                                    }
                                    else if (ssDCval >= 54) {
                                        return 8;
                                    }
                                }
                            }
                            else if (refDCval >= 92) {
                                if (Cs < 3153) {
                                    if (refDCval < 112) {
                                        if (TSC < 8) {
                                            if (Cs < 3049) {
                                                return 8;
                                            }
                                            else if (Cs >= 3049) {
                                                if (Cs < 3051) {
                                                    if (AFD < 4) {
                                                        return 8;
                                                    }
                                                    else if (AFD >= 4) {
                                                        return 4;
                                                    }
                                                }
                                                else if (Cs >= 3051) {
                                                    return 8;
                                                }
                                            }
                                        }
                                        else if (TSC >= 8) {
                                            if (Rs < 3460) {
                                                return 8;
                                            }
                                            else if (Rs >= 3460) {
                                                return 4;
                                            }
                                        }
                                    }
                                    else if (refDCval >= 112) {
                                        if (Cs < 2944) {
                                            if (mu_mv_mag_sq < 128) {
                                                if (Cs < 2878) {
                                                    return 8;
                                                }
                                                else if (Cs >= 2878) {
                                                    if (diffRsCsdiff < -4) {
                                                        if (mu_mv_mag_sq < 122) {
                                                            if (ssDCval < 113) {
                                                                return 8;
                                                            }
                                                            else if (ssDCval >= 113) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (mu_mv_mag_sq >= 122) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (diffRsCsdiff >= -4) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                            else if (mu_mv_mag_sq >= 128) {
                                                if (Cs < 2835) {
                                                    return 4;
                                                }
                                                else if (Cs >= 2835) {
                                                    return 8;
                                                }
                                            }
                                        }
                                        else if (Cs >= 2944) {
                                            if (negBalance < 48) {
                                                if (negBalance < 47) {
                                                    return 8;
                                                }
                                                else if (negBalance >= 47) {
                                                    if (mu_mv_mag_sq < 123) {
                                                        if (Cs < 3012) {
                                                            return 8;
                                                        }
                                                        else if (Cs >= 3012) {
                                                            if (Rs < 2899) {
                                                                if (mu_mv_mag_sq < 121) {
                                                                    return 8;
                                                                }
                                                                else if (mu_mv_mag_sq >= 121) {
                                                                    if (Rs < 2702) {
                                                                        return 8;
                                                                    }
                                                                    else if (Rs >= 2702) {
                                                                        return 4;
                                                                    }
                                                                }
                                                            }
                                                            else if (Rs >= 2899) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                    else if (mu_mv_mag_sq >= 123) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                            else if (negBalance >= 48) {
                                                if (Rs < 2703) {
                                                    return 8;
                                                }
                                                else if (Rs >= 2703) {
                                                    if (Rs < 2806) {
                                                        if (TSC < 6) {
                                                            return 8;
                                                        }
                                                        else if (TSC >= 6) {
                                                            if (ssDCval < 112) {
                                                                return 8;
                                                            }
                                                            else if (ssDCval >= 112) {
                                                                if (Cs < 3025) {
                                                                    return 4;
                                                                }
                                                                else if (Cs >= 3025) {
                                                                    if (SC < 5818) {
                                                                        return 8;
                                                                    }
                                                                    else if (SC >= 5818) {
                                                                        return 4;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (Rs >= 2806) {
                                                        if (Cs < 3145) {
                                                            if (Cs < 3078) {
                                                                if (Cs < 3064) {
                                                                    if (mu_mv_mag_sq < 122) {
                                                                        return 4;
                                                                    }
                                                                    else if (mu_mv_mag_sq >= 122) {
                                                                        return 8;
                                                                    }
                                                                }
                                                                else if (Cs >= 3064) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (Cs >= 3078) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (Cs >= 3145) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (Cs >= 3153) {
                                    if (RsDiff < 12) {
                                        if (Rs < 8460) {
                                            return 8;
                                        }
                                        else if (Rs >= 8460) {
                                            return 4;
                                        }
                                    }
                                    else if (RsDiff >= 12) {
                                        if (Rs < 8412) {
                                            return 8;
                                        }
                                        else if (Rs >= 8412) {
                                            return 4;
                                        }
                                    }
                                }
                            }
                        }
                    }
                    else if (TSC >= 9) {
                        if (negBalance < 54) {
                            if (refDCval < 106) {
                                if (Cs < 3615) {
                                    return 4;
                                }
                                else if (Cs >= 3615) {
                                    return 8;
                                }
                            }
                            else if (refDCval >= 106) {
                                if (refDCval < 111) {
                                    if (ssDCval < 106) {
                                        return 8;
                                    }
                                    else if (ssDCval >= 106) {
                                        return 4;
                                    }
                                }
                                else if (refDCval >= 111) {
                                    return 8;
                                }
                            }
                        }
                        else if (negBalance >= 54) {
                            return 4;
                        }
                    }
                }
                else if (tcor >= 98) {
                    if (mu_mv_mag_sq < 121) {
                        if (refDCval < 154) {
                            if (ssDCval < 100) {
                                if (refDCval < 73) {
                                    if (Cs < 708) {
                                        if (Cs < 692) {
                                            if (CsDiff < 1) {
                                                return 8;
                                            }
                                            else if (CsDiff >= 1) {
                                                if (Cs < 396) {
                                                    return 4;
                                                }
                                                else if (Cs >= 396) {
                                                    if (Rs < 1446) {
                                                        return 8;
                                                    }
                                                    else if (Rs >= 1446) {
                                                        return 2;
                                                    }
                                                }
                                            }
                                        }
                                        else if (Cs >= 692) {
                                            return 4;
                                        }
                                    }
                                    else if (Cs >= 708) {
                                        if (gchDC < 2) {
                                            if (SC < 2379) {
                                                if (SC < 2179) {
                                                    return 8;
                                                }
                                                else if (SC >= 2179) {
                                                    if (Rs < 1047) {
                                                        return 4;
                                                    }
                                                    else if (Rs >= 1047) {
                                                        return 2;
                                                    }
                                                }
                                            }
                                            else if (SC >= 2379) {
                                                return 8;
                                            }
                                        }
                                        else if (gchDC >= 2) {
                                            if (diffMVdiffVal < -5) {
                                                return 8;
                                            }
                                            else if (diffMVdiffVal >= -5) {
                                                return 4;
                                            }
                                        }
                                    }
                                }
                                else if (refDCval >= 73) {
                                    if (ssDCval < 76) {
                                        if (SC < 379) {
                                            return 8;
                                        }
                                        else if (SC >= 379) {
                                            return 4;
                                        }
                                    }
                                    else if (ssDCval >= 76) {
                                        if (Rs < 2431) {
                                            if (Cs < 595) {
                                                return 8;
                                            }
                                            else if (Cs >= 595) {
                                                if (Cs < 1726) {
                                                    if (RsDiff < 1) {
                                                        if (AFD < 3) {
                                                            return 4;
                                                        }
                                                        else if (AFD >= 3) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (RsDiff >= 1) {
                                                        return 4;
                                                    }
                                                }
                                                else if (Cs >= 1726) {
                                                    if (SC < 3737) {
                                                        return 8;
                                                    }
                                                    else if (SC >= 3737) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                        }
                                        else if (Rs >= 2431) {
                                            if (AFD < 6) {
                                                if (negBalance < 40) {
                                                    return 8;
                                                }
                                                else if (negBalance >= 40) {
                                                    if (Cs < 4102) {
                                                        return 1;
                                                    }
                                                    else if (Cs >= 4102) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                            else if (AFD >= 6) {
                                                if (Rs < 3302) {
                                                    return 1;
                                                }
                                                else if (Rs >= 3302) {
                                                    return 8;
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                            else if (ssDCval >= 100) {
                                if (refDCval < 111) {
                                    if (refDCval < 108) {
                                        if (TSC < 5) {
                                            return 8;
                                        }
                                        else if (TSC >= 5) {
                                            return 4;
                                        }
                                    }
                                    else if (refDCval >= 108) {
                                        if (Cs < 17) {
                                            return 1;
                                        }
                                        else if (Cs >= 17) {
                                            if (RsCsDiff < 22) {
                                                if (mu_mv_mag_sq < 117) {
                                                    return 8;
                                                }
                                                else if (mu_mv_mag_sq >= 117) {
                                                    return 1;
                                                }
                                            }
                                            else if (RsCsDiff >= 22) {
                                                if (mu_mv_mag_sq < 114) {
                                                    return 8;
                                                }
                                                else if (mu_mv_mag_sq >= 114) {
                                                    if (negBalance < 29) {
                                                        if (Cs < 3085) {
                                                            return 1;
                                                        }
                                                        else if (Cs >= 3085) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (negBalance >= 29) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (refDCval >= 111) {
                                    if (refDCval < 119) {
                                        if (AFD < 4) {
                                            return 8;
                                        }
                                        else if (AFD >= 4) {
                                            if (RsDiff < 3) {
                                                return 4;
                                            }
                                            else if (RsDiff >= 3) {
                                                return 8;
                                            }
                                        }
                                    }
                                    else if (refDCval >= 119) {
                                        return 8;
                                    }
                                }
                            }
                        }
                        else if (refDCval >= 154) {
                            if (negBalance < 9) {
                                return 4;
                            }
                            else if (negBalance >= 9) {
                                if (Cs < 437) {
                                    return 4;
                                }
                                else if (Cs >= 437) {
                                    if (ssDCval < 163) {
                                        return 4;
                                    }
                                    else if (ssDCval >= 163) {
                                        return 2;
                                    }
                                }
                            }
                        }
                    }
                    else if (mu_mv_mag_sq >= 121) {
                        if (Cs < 689) {
                            if (SC < 923) {
                                if (negBalance < 16) {
                                    return 8;
                                }
                                else if (negBalance >= 16) {
                                    if (Scindex < 3) {
                                        if (TSCindex < 2) {
                                            if (Cs < 308) {
                                                return 8;
                                            }
                                            else if (Cs >= 308) {
                                                return 4;
                                            }
                                        }
                                        else if (TSCindex >= 2) {
                                            return 4;
                                        }
                                    }
                                    else if (Scindex >= 3) {
                                        return 8;
                                    }
                                }
                            }
                            else if (SC >= 923) {
                                if (Rs < 673) {
                                    return 4;
                                }
                                else if (Rs >= 673) {
                                    return 8;
                                }
                            }
                        }
                        else if (Cs >= 689) {
                            if (posBalance < 23) {
                                if (ssDCval < 103) {
                                    return 8;
                                }
                                else if (ssDCval >= 103) {
                                    if (Rs < 3779) {
                                        if (Rs < 3774) {
                                            if (negBalance < 3) {
                                                if (posBalance < 22) {
                                                    if (Cs < 1340) {
                                                        return 1;
                                                    }
                                                    else if (Cs >= 1340) {
                                                        if (Cs < 3849) {
                                                            if (SC < 7310) {
                                                                return 8;
                                                            }
                                                            else if (SC >= 7310) {
                                                                return 1;
                                                            }
                                                        }
                                                        else if (Cs >= 3849) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                                else if (posBalance >= 22) {
                                                    if (diffRsCsdiff < 1) {
                                                        if (negBalance < 1) {
                                                            return 1;
                                                        }
                                                        else if (negBalance >= 1) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (diffRsCsdiff >= 1) {
                                                        return 1;
                                                    }
                                                }
                                            }
                                            else if (negBalance >= 3) {
                                                return 8;
                                            }
                                        }
                                        else if (Rs >= 3774) {
                                            return 1;
                                        }
                                    }
                                    else if (Rs >= 3779) {
                                        return 8;
                                    }
                                }
                            }
                            else if (posBalance >= 23) {
                                if (SC < 4334) {
                                    if (Cs < 2298) {
                                        if (RsCsDiff < 1) {
                                            if (Rs < 1422) {
                                                if (Rs < 1339) {
                                                    return 8;
                                                }
                                                else if (Rs >= 1339) {
                                                    if (Cs < 1284) {
                                                        return 4;
                                                    }
                                                    else if (Cs >= 1284) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                            else if (Rs >= 1422) {
                                                return 8;
                                            }
                                        }
                                        else if (RsCsDiff >= 1) {
                                            return 8;
                                        }
                                    }
                                    else if (Cs >= 2298) {
                                        return 4;
                                    }
                                }
                                else if (SC >= 4334) {
                                    if (mu_mv_mag_sq < 125) {
                                        if (Cs < 3056) {
                                            if (SC < 6070) {
                                                return 8;
                                            }
                                            else if (SC >= 6070) {
                                                if (negBalance < 26) {
                                                    if (refDCval < 108) {
                                                        return 8;
                                                    }
                                                    else if (refDCval >= 108) {
                                                        if (SC < 6300) {
                                                            return 1;
                                                        }
                                                        else if (SC >= 6300) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                                else if (negBalance >= 26) {
                                                    return 1;
                                                }
                                            }
                                        }
                                        else if (Cs >= 3056) {
                                            return 8;
                                        }
                                    }
                                    else if (mu_mv_mag_sq >= 125) {
                                        return 8;
                                    }
                                }
                            }
                        }
                    }
                }
            }
            else if (MVDiff >= 5) {
                if (posBalance < 43) {
                    if (mu_mv_mag_sq < 91) {
                        if (MVDiff < 27) {
                            if (ssDCval < 108) {
                                if (posBalance < 5) {
                                    return 8;
                                }
                                else if (posBalance >= 5) {
                                    if (ssDCval < 82) {
                                        if (mu_mv_mag_sq < 60) {
                                            if (RsCsDiff < 4) {
                                                if (TSCindex < 5) {
                                                    if (ssDCval < 34) {
                                                        return 2;
                                                    }
                                                    else if (ssDCval >= 34) {
                                                        return 4;
                                                    }
                                                }
                                                else if (TSCindex >= 5) {
                                                    if (Rs < 1595) {
                                                        return 8;
                                                    }
                                                    else if (Rs >= 1595) {
                                                        return 2;
                                                    }
                                                }
                                            }
                                            else if (RsCsDiff >= 4) {
                                                if (diffAFD < 1) {
                                                    if (negBalance < 33) {
                                                        return 4;
                                                    }
                                                    else if (negBalance >= 33) {
                                                        return 2;
                                                    }
                                                }
                                                else if (diffAFD >= 1) {
                                                    if (MVDiff < 8) {
                                                        return 2;
                                                    }
                                                    else if (MVDiff >= 8) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                        }
                                        else if (mu_mv_mag_sq >= 60) {
                                            if (posBalance < 37) {
                                                return 4;
                                            }
                                            else if (posBalance >= 37) {
                                                if (AFD < 4) {
                                                    return 1;
                                                }
                                                else if (AFD >= 4) {
                                                    return 4;
                                                }
                                            }
                                        }
                                    }
                                    else if (ssDCval >= 82) {
                                        if (Rs < 754) {
                                            return 4;
                                        }
                                        else if (Rs >= 754) {
                                            if (TSC < 8) {
                                                return 8;
                                            }
                                            else if (TSC >= 8) {
                                                return 4;
                                            }
                                        }
                                    }
                                }
                            }
                            else if (ssDCval >= 108) {
                                if (posBalance < 13) {
                                    return 8;
                                }
                                else if (posBalance >= 13) {
                                    if (Cs < 2064) {
                                        if (SC < 3660) {
                                            if (Cs < 1534) {
                                                if (posBalance < 26) {
                                                    return 4;
                                                }
                                                else if (posBalance >= 26) {
                                                    return 2;
                                                }
                                            }
                                            else if (Cs >= 1534) {
                                                return 8;
                                            }
                                        }
                                        else if (SC >= 3660) {
                                            return 4;
                                        }
                                    }
                                    else if (Cs >= 2064) {
                                        if (refDCval < 118) {
                                            if (SC < 6266) {
                                                return 2;
                                            }
                                            else if (SC >= 6266) {
                                                return 4;
                                            }
                                        }
                                        else if (refDCval >= 118) {
                                            if (Cs < 2526) {
                                                return 4;
                                            }
                                            else if (Cs >= 2526) {
                                                return 8;
                                            }
                                        }
                                    }
                                }
                            }
                        }
                        else if (MVDiff >= 27) {
                            if (AFD < 18) {
                                if (diffTSC < -2) {
                                    if (tcor < 105) {
                                        if (diffAFD < -47) {
                                            return 2;
                                        }
                                        else if (diffAFD >= -47) {
                                            return 4;
                                        }
                                    }
                                    else if (tcor >= 105) {
                                        return 2;
                                    }
                                }
                                else if (diffTSC >= -2) {
                                    if (mu_mv_mag_sq < 9) {
                                        if (diffMVdiffVal < -113) {
                                            return 1;
                                        }
                                        else if (diffMVdiffVal >= -113) {
                                            if (refDCval < 20) {
                                                return 1;
                                            }
                                            else if (refDCval >= 20) {
                                                return 2;
                                            }
                                        }
                                    }
                                    else if (mu_mv_mag_sq >= 9) {
                                        if (diffMVdiffVal < -10) {
                                            if (refDCval < 112) {
                                                return 2;
                                            }
                                            else if (refDCval >= 112) {
                                                if (Cs < 320) {
                                                    return 1;
                                                }
                                                else if (Cs >= 320) {
                                                    return 8;
                                                }
                                            }
                                        }
                                        else if (diffMVdiffVal >= -10) {
                                            if (Cs < 6) {
                                                if (diffAFD < -1) {
                                                    return 2;
                                                }
                                                else if (diffAFD >= -1) {
                                                    if (tcor < 102) {
                                                        return 8;
                                                    }
                                                    else if (tcor >= 102) {
                                                        if (diffMVdiffVal < 34) {
                                                            if (negBalance < 8) {
                                                                return 8;
                                                            }
                                                            else if (negBalance >= 8) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (diffMVdiffVal >= 34) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (Cs >= 6) {
                                                if (mu_mv_mag_sq < 43) {
                                                    if (Cs < 1376) {
                                                        return 2;
                                                    }
                                                    else if (Cs >= 1376) {
                                                        if (MVDiff < 82) {
                                                            return 8;
                                                        }
                                                        else if (MVDiff >= 82) {
                                                            return 2;
                                                        }
                                                    }
                                                }
                                                else if (mu_mv_mag_sq >= 43) {
                                                    return 8;
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                            else if (AFD >= 18) {
                                if (SC < 1934) {
                                    if (Cs < 36) {
                                        return 8;
                                    }
                                    else if (Cs >= 36) {
                                        return 1;
                                    }
                                }
                                else if (SC >= 1934) {
                                    return 4;
                                }
                            }
                        }
                    }
                    else if (mu_mv_mag_sq >= 91) {
                        if (SC < 1870) {
                            if (Cs < 360) {
                                if (TSCindex < 2) {
                                    return 8;
                                }
                                else if (TSCindex >= 2) {
                                    return 1;
                                }
                            }
                            else if (Cs >= 360) {
                                if (posBalance < 32) {
                                    if (Cs < 832) {
                                        if (posBalance < 23) {
                                            return 4;
                                        }
                                        else if (posBalance >= 23) {
                                            if (RsDiff < 1) {
                                                if (mu_mv_mag_sq < 93) {
                                                    return 2;
                                                }
                                                else if (mu_mv_mag_sq >= 93) {
                                                    if (diffMVdiffVal < 1) {
                                                        if (posBalance < 27) {
                                                            return 4;
                                                        }
                                                        else if (posBalance >= 27) {
                                                            return 2;
                                                        }
                                                    }
                                                    else if (diffMVdiffVal >= 1) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                            else if (RsDiff >= 1) {
                                                return 2;
                                            }
                                        }
                                    }
                                    else if (Cs >= 832) {
                                        if (MVDiff < 26) {
                                            return 1;
                                        }
                                        else if (MVDiff >= 26) {
                                            return 4;
                                        }
                                    }
                                }
                                else if (posBalance >= 32) {
                                    if (MVDiff < 22) {
                                        return 2;
                                    }
                                    else if (MVDiff >= 22) {
                                        if (SC < 1028) {
                                            return 4;
                                        }
                                        else if (SC >= 1028) {
                                            return 8;
                                        }
                                    }
                                }
                            }
                        }
                        else if (SC >= 1870) {
                            if (diffMVdiffVal < 13) {
                                if (diffMVdiffVal < -6) {
                                    if (AFD < 3) {
                                        return 1;
                                    }
                                    else if (AFD >= 3) {
                                        if (TSC < 5) {
                                            return 8;
                                        }
                                        else if (TSC >= 5) {
                                            return 4;
                                        }
                                    }
                                }
                                else if (diffMVdiffVal >= -6) {
                                    if (negBalance < 3) {
                                        if (diffRsCsdiff < -4) {
                                            if (diffAFD < -9) {
                                                return 2;
                                            }
                                            else if (diffAFD >= -9) {
                                                return 8;
                                            }
                                        }
                                        else if (diffRsCsdiff >= -4) {
                                            return 4;
                                        }
                                    }
                                    else if (negBalance >= 3) {
                                        if (CsDiff < 5) {
                                            return 8;
                                        }
                                        else if (CsDiff >= 5) {
                                            if (mu_mv_mag_sq < 100) {
                                                if (RsDiff < 6) {
                                                    return 4;
                                                }
                                                else if (RsDiff >= 6) {
                                                    return 8;
                                                }
                                            }
                                            else if (mu_mv_mag_sq >= 100) {
                                                return 8;
                                            }
                                        }
                                    }
                                }
                            }
                            else if (diffMVdiffVal >= 13) {
                                if (Rs < 1603) {
                                    return 4;
                                }
                                else if (Rs >= 1603) {
                                    return 2;
                                }
                            }
                        }
                    }
                }
                else if (posBalance >= 43) {
                    if (mu_mv_mag_sq < 82) {
                        if (Rs < 3000) {
                            if (gchDC < 2) {
                                if (negBalance < 14) {
                                    if (MVDiff < 7) {
                                        return 8;
                                    }
                                    else if (MVDiff >= 7) {
                                        if (AFD < 1) {
                                            return 4;
                                        }
                                        else if (AFD >= 1) {
                                            return 2;
                                        }
                                    }
                                }
                                else if (negBalance >= 14) {
                                    if (Rs < 995) {
                                        if (mu_mv_mag_sq < 45) {
                                            if (Cs < 286) {
                                                if (negBalance < 57) {
                                                    return 1;
                                                }
                                                else if (negBalance >= 57) {
                                                    return 2;
                                                }
                                            }
                                            else if (Cs >= 286) {
                                                if (Rs < 704) {
                                                    return 2;
                                                }
                                                else if (Rs >= 704) {
                                                    return 8;
                                                }
                                            }
                                        }
                                        else if (mu_mv_mag_sq >= 45) {
                                            if (Rs < 961) {
                                                if (Cs < 1274) {
                                                    return 8;
                                                }
                                                else if (Cs >= 1274) {
                                                    return 4;
                                                }
                                            }
                                            else if (Rs >= 961) {
                                                return 4;
                                            }
                                        }
                                    }
                                    else if (Rs >= 995) {
                                        if (posBalance < 54) {
                                            if (Rs < 1752) {
                                                return 1;
                                            }
                                            else if (Rs >= 1752) {
                                                if (Cs < 1629) {
                                                    return 4;
                                                }
                                                else if (Cs >= 1629) {
                                                    return 1;
                                                }
                                            }
                                        }
                                        else if (posBalance >= 54) {
                                            return 1;
                                        }
                                    }
                                }
                            }
                            else if (gchDC >= 2) {
                                if (RsCsDiff < 7) {
                                    if (Rs < 284) {
                                        return 1;
                                    }
                                    else if (Rs >= 284) {
                                        return 4;
                                    }
                                }
                                else if (RsCsDiff >= 7) {
                                    if (diffMVdiffVal < 7) {
                                        if (diffRsCsdiff < 1) {
                                            if (Rs < 1497) {
                                                return 8;
                                            }
                                            else if (Rs >= 1497) {
                                                if (Rs < 2195) {
                                                    return 4;
                                                }
                                                else if (Rs >= 2195) {
                                                    return 8;
                                                }
                                            }
                                        }
                                        else if (diffRsCsdiff >= 1) {
                                            return 4;
                                        }
                                    }
                                    else if (diffMVdiffVal >= 7) {
                                        if (Cs < 392) {
                                            return 2;
                                        }
                                        else if (Cs >= 392) {
                                            if (Rs < 908) {
                                                return 8;
                                            }
                                            else if (Rs >= 908) {
                                                return 1;
                                            }
                                        }
                                    }
                                }
                            }
                        }
                        else if (Rs >= 3000) {
                            if (RsCsDiff < 83) {
                                if (CsDiff < 5) {
                                    if (diffMVdiffVal < -2) {
                                        if (refDCval < 55) {
                                            return 4;
                                        }
                                        else if (refDCval >= 55) {
                                            if (diffRsCsdiff < 9) {
                                                return 8;
                                            }
                                            else if (diffRsCsdiff >= 9) {
                                                return 4;
                                            }
                                        }
                                    }
                                    else if (diffMVdiffVal >= -2) {
                                        if (diffRsCsdiff < -17) {
                                            if (Rs < 3401) {
                                                return 4;
                                            }
                                            else if (Rs >= 3401) {
                                                return 8;
                                            }
                                        }
                                        else if (diffRsCsdiff >= -17) {
                                            if (refDCval < 52) {
                                                return 4;
                                            }
                                            else if (refDCval >= 52) {
                                                return 8;
                                            }
                                        }
                                    }
                                }
                                else if (CsDiff >= 5) {
                                    if (ssDCval < 108) {
                                        return 8;
                                    }
                                    else if (ssDCval >= 108) {
                                        return 4;
                                    }
                                }
                            }
                            else if (RsCsDiff >= 83) {
                                if (Cs < 2655) {
                                    return 2;
                                }
                                else if (Cs >= 2655) {
                                    return 8;
                                }
                            }
                        }
                    }
                    else if (mu_mv_mag_sq >= 82) {
                        if (Cs < 748) {
                            if (Rs < 1115) {
                                if (ssDCval < 103) {
                                    if (refDCval < 83) {
                                        if (ssDCval < 73) {
                                            if (Cs < 652) {
                                                return 8;
                                            }
                                            else if (Cs >= 652) {
                                                return 2;
                                            }
                                        }
                                        else if (ssDCval >= 73) {
                                            return 4;
                                        }
                                    }
                                    else if (refDCval >= 83) {
                                        return 8;
                                    }
                                }
                                else if (ssDCval >= 103) {
                                    if (Cs < 516) {
                                        if (Rs < 323) {
                                            return 2;
                                        }
                                        else if (Rs >= 323) {
                                            return 4;
                                        }
                                    }
                                    else if (Cs >= 516) {
                                        return 2;
                                    }
                                }
                            }
                            else if (Rs >= 1115) {
                                return 1;
                            }
                        }
                        else if (Cs >= 748) {
                            if (diffAFD < -7) {
                                if (diffMVdiffVal < -5) {
                                    return 8;
                                }
                                else if (diffMVdiffVal >= -5) {
                                    if (Rs < 3967) {
                                        return 4;
                                    }
                                    else if (Rs >= 3967) {
                                        return 8;
                                    }
                                }
                            }
                            else if (diffAFD >= -7) {
                                if (refDCval < 54) {
                                    return 4;
                                }
                                else if (refDCval >= 54) {
                                    if (TSC < 8) {
                                        if (negBalance < 41) {
                                            return 8;
                                        }
                                        else if (negBalance >= 41) {
                                            if (tcor < 98) {
                                                return 8;
                                            }
                                            else if (tcor >= 98) {
                                                if (Rs < 1775) {
                                                    return 8;
                                                }
                                                else if (Rs >= 1775) {
                                                    if (Rs < 3478) {
                                                        return 4;
                                                    }
                                                    else if (Rs >= 3478) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (TSC >= 8) {
                                        if (SC < 4404) {
                                            return 1;
                                        }
                                        else if (SC >= 4404) {
                                            return 8;
                                        }
                                    }
                                }
                            }
                        }
                    }
                }
            }
        }
        return 4;
    }
    mfxU16 AGOPSelect3(mfxU32 Rs, mfxU32 CsDiff, mfxU32 TSC, mfxI32 diffAFD, mfxU32 Scindex, mfxU32 RsDiff, mfxI32 diffMVdiffVal, mfxU32 RsCsDiff, mfxU32 TSCindex, mfxI16 tcor, mfxU32 refDCval, mfxU32 MVDiff, mfxU32 AFD, mfxI32 diffTSC, mfxI32 diffRsCsdiff, mfxU32 posBalance, mfxU32 SC, mfxU32 mu_mv_mag_sq, mfxU32 negBalance, mfxI16 mcTcor, mfxU32 gchDC, mfxU32 Cs, mfxU32 ssDCval) {
        if (AFD < 7) {
            if (mu_mv_mag_sq < 108) {
                if (MVDiff < 10) {
                    if (refDCval < 26) {
                        if (Cs < 1787) {
                            if (Cs < 88) {
                                return 4;
                            }
                            else if (Cs >= 88) {
                                return 8;
                            }
                        }
                        else if (Cs >= 1787) {
                            if (Rs < 4675) {
                                if (Cs < 1891) {
                                    return 1;
                                }
                                else if (Cs >= 1891) {
                                    if (Rs < 3780) {
                                        if (Cs < 3973) {
                                            return 8;
                                        }
                                        else if (Cs >= 3973) {
                                            if (Cs < 3985) {
                                                return 1;
                                            }
                                            else if (Cs >= 3985) {
                                                return 8;
                                            }
                                        }
                                    }
                                    else if (Rs >= 3780) {
                                        if (Cs < 3951) {
                                            return 1;
                                        }
                                        else if (Cs >= 3951) {
                                            return 8;
                                        }
                                    }
                                }
                            }
                            else if (Rs >= 4675) {
                                if (Rs < 5376) {
                                    return 1;
                                }
                                else if (Rs >= 5376) {
                                    return 8;
                                }
                            }
                        }
                    }
                    else if (refDCval >= 26) {
                        if (ssDCval < 55) {
                            if (Rs < 1134) {
                                if (Cs < 1265) {
                                    if (Cs < 667) {
                                        if (Rs < 778) {
                                            if (RsCsDiff < 2) {
                                                if (mu_mv_mag_sq < 76) {
                                                    if (ssDCval < 39) {
                                                        return 4;
                                                    }
                                                    else if (ssDCval >= 39) {
                                                        return 8;
                                                    }
                                                }
                                                else if (mu_mv_mag_sq >= 76) {
                                                    return 4;
                                                }
                                            }
                                            else if (RsCsDiff >= 2) {
                                                if (SC < 1090) {
                                                    return 4;
                                                }
                                                else if (SC >= 1090) {
                                                    if (posBalance < 51) {
                                                        return 4;
                                                    }
                                                    else if (posBalance >= 51) {
                                                        if (Cs < 553) {
                                                            return 8;
                                                        }
                                                        else if (Cs >= 553) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (Rs >= 778) {
                                            if (posBalance < 51) {
                                                return 4;
                                            }
                                            else if (posBalance >= 51) {
                                                if (diffTSC < -1) {
                                                    if (Rs < 955) {
                                                        return 4;
                                                    }
                                                    else if (Rs >= 955) {
                                                        return 8;
                                                    }
                                                }
                                                else if (diffTSC >= -1) {
                                                    return 8;
                                                }
                                            }
                                        }
                                    }
                                    else if (Cs >= 667) {
                                        if (TSC < 2) {
                                            if (AFD < 1) {
                                                if (Rs < 888) {
                                                    return 1;
                                                }
                                                else if (Rs >= 888) {
                                                    return 4;
                                                }
                                            }
                                            else if (AFD >= 1) {
                                                return 8;
                                            }
                                        }
                                        else if (TSC >= 2) {
                                            if (negBalance < 43) {
                                                if (ssDCval < 49) {
                                                    if (Rs < 992) {
                                                        return 4;
                                                    }
                                                    else if (Rs >= 992) {
                                                        return 1;
                                                    }
                                                }
                                                else if (ssDCval >= 49) {
                                                    return 4;
                                                }
                                            }
                                            else if (negBalance >= 43) {
                                                if (Cs < 757) {
                                                    return 1;
                                                }
                                                else if (Cs >= 757) {
                                                    return 4;
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (Cs >= 1265) {
                                    if (RsDiff < 1) {
                                        return 2;
                                    }
                                    else if (RsDiff >= 1) {
                                        return 8;
                                    }
                                }
                            }
                            else if (Rs >= 1134) {
                                if (SC < 2599) {
                                    if (CsDiff < 2) {
                                        if (mu_mv_mag_sq < 92) {
                                            return 1;
                                        }
                                        else if (mu_mv_mag_sq >= 92) {
                                            return 8;
                                        }
                                    }
                                    else if (CsDiff >= 2) {
                                        if (posBalance < 53) {
                                            if (MVDiff < 8) {
                                                if (mcTcor < 92) {
                                                    return 1;
                                                }
                                                else if (mcTcor >= 92) {
                                                    if (refDCval < 51) {
                                                        if (Rs < 1395) {
                                                            return 1;
                                                        }
                                                        else if (Rs >= 1395) {
                                                            if (Rs < 1434) {
                                                                return 4;
                                                            }
                                                            else if (Rs >= 1434) {
                                                                return 1;
                                                            }
                                                        }
                                                    }
                                                    else if (refDCval >= 51) {
                                                        if (tcor < 94) {
                                                            if (mu_mv_mag_sq < 61) {
                                                                if (ssDCval < 52) {
                                                                    return 1;
                                                                }
                                                                else if (ssDCval >= 52) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (mu_mv_mag_sq >= 61) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (tcor >= 94) {
                                                            return 1;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (MVDiff >= 8) {
                                                if (diffRsCsdiff < -4) {
                                                    return 1;
                                                }
                                                else if (diffRsCsdiff >= -4) {
                                                    if (ssDCval < 49) {
                                                        return 8;
                                                    }
                                                    else if (ssDCval >= 49) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                        }
                                        else if (posBalance >= 53) {
                                            if (Rs < 1357) {
                                                return 1;
                                            }
                                            else if (Rs >= 1357) {
                                                if (SC < 2486) {
                                                    return 8;
                                                }
                                                else if (SC >= 2486) {
                                                    return 1;
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (SC >= 2599) {
                                    if (Rs < 1807) {
                                        if (ssDCval < 43) {
                                            return 8;
                                        }
                                        else if (ssDCval >= 43) {
                                            if (Rs < 1658) {
                                                return 4;
                                            }
                                            else if (Rs >= 1658) {
                                                return 1;
                                            }
                                        }
                                    }
                                    else if (Rs >= 1807) {
                                        if (refDCval < 54) {
                                            if (negBalance < 35) {
                                                if (TSC < 5) {
                                                    return 4;
                                                }
                                                else if (TSC >= 5) {
                                                    return 8;
                                                }
                                            }
                                            else if (negBalance >= 35) {
                                                if (Rs < 3574) {
                                                    if (SC < 7004) {
                                                        if (diffTSC < -1) {
                                                            if (Rs < 3536) {
                                                                return 8;
                                                            }
                                                            else if (Rs >= 3536) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (diffTSC >= -1) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (SC >= 7004) {
                                                        return 8;
                                                    }
                                                }
                                                else if (Rs >= 3574) {
                                                    return 4;
                                                }
                                            }
                                        }
                                        else if (refDCval >= 54) {
                                            return 8;
                                        }
                                    }
                                }
                            }
                        }
                        else if (ssDCval >= 55) {
                            if (SC < 4942) {
                                if (mcTcor < 95) {
                                    if (Rs < 288) {
                                        if (ssDCval < 88) {
                                            return 2;
                                        }
                                        else if (ssDCval >= 88) {
                                            if (Cs < 277) {
                                                return 2;
                                            }
                                            else if (Cs >= 277) {
                                                return 4;
                                            }
                                        }
                                    }
                                    else if (Rs >= 288) {
                                        if (MVDiff < 6) {
                                            if (Rs < 1641) {
                                                if (negBalance < 38) {
                                                    if (TSC < 3) {
                                                        if (CsDiff < 2) {
                                                            return 8;
                                                        }
                                                        else if (CsDiff >= 2) {
                                                            return 1;
                                                        }
                                                    }
                                                    else if (TSC >= 3) {
                                                        return 8;
                                                    }
                                                }
                                                else if (negBalance >= 38) {
                                                    if (AFD < 6) {
                                                        return 8;
                                                    }
                                                    else if (AFD >= 6) {
                                                        return 1;
                                                    }
                                                }
                                            }
                                            else if (Rs >= 1641) {
                                                if (mu_mv_mag_sq < 105) {
                                                    return 1;
                                                }
                                                else if (mu_mv_mag_sq >= 105) {
                                                    if (Cs < 1801) {
                                                        return 4;
                                                    }
                                                    else if (Cs >= 1801) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                        }
                                        else if (MVDiff >= 6) {
                                            if (posBalance < 18) {
                                                return 1;
                                            }
                                            else if (posBalance >= 18) {
                                                if (mu_mv_mag_sq < 73) {
                                                    if (TSC < 5) {
                                                        if (MVDiff < 7) {
                                                            return 1;
                                                        }
                                                        else if (MVDiff >= 7) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (TSC >= 5) {
                                                        return 8;
                                                    }
                                                }
                                                else if (mu_mv_mag_sq >= 73) {
                                                    if (diffMVdiffVal < -8) {
                                                        return 1;
                                                    }
                                                    else if (diffMVdiffVal >= -8) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (mcTcor >= 95) {
                                    if (Cs < 238) {
                                        if (gchDC < 1) {
                                            if (Rs < 867) {
                                                if (RsDiff < 1) {
                                                    return 1;
                                                }
                                                else if (RsDiff >= 1) {
                                                    return 8;
                                                }
                                            }
                                            else if (Rs >= 867) {
                                                return 8;
                                            }
                                        }
                                        else if (gchDC >= 1) {
                                            return 8;
                                        }
                                    }
                                    else if (Cs >= 238) {
                                        if (MVDiff < 1) {
                                            if (posBalance < 25) {
                                                return 8;
                                            }
                                            else if (posBalance >= 25) {
                                                if (AFD < 5) {
                                                    if (diffMVdiffVal < -1) {
                                                        if (ssDCval < 88) {
                                                            if (posBalance < 43) {
                                                                return 4;
                                                            }
                                                            else if (posBalance >= 43) {
                                                                return 1;
                                                            }
                                                        }
                                                        else if (ssDCval >= 88) {
                                                            if (Cs < 603) {
                                                                if (Rs < 370) {
                                                                    if (Rs < 230) {
                                                                        return 8;
                                                                    }
                                                                    else if (Rs >= 230) {
                                                                        return 2;
                                                                    }
                                                                }
                                                                else if (Rs >= 370) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (Cs >= 603) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                    else if (diffMVdiffVal >= -1) {
                                                        if (tcor < 98) {
                                                            if (Cs < 291) {
                                                                return 4;
                                                            }
                                                            else if (Cs >= 291) {
                                                                if (Cs < 390) {
                                                                    return 1;
                                                                }
                                                                else if (Cs >= 390) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                        else if (tcor >= 98) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                                else if (AFD >= 5) {
                                                    if (TSCindex < 7) {
                                                        return 4;
                                                    }
                                                    else if (TSCindex >= 7) {
                                                        if (SC < 3980) {
                                                            return 2;
                                                        }
                                                        else if (SC >= 3980) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (MVDiff >= 1) {
                                            if (ssDCval < 152) {
                                                if (refDCval < 104) {
                                                    if (mu_mv_mag_sq < 56) {
                                                        if (ssDCval < 64) {
                                                            if (Cs < 564) {
                                                                return 8;
                                                            }
                                                            else if (Cs >= 564) {
                                                                if (MVDiff < 2) {
                                                                    return 4;
                                                                }
                                                                else if (MVDiff >= 2) {
                                                                    return 1;
                                                                }
                                                            }
                                                        }
                                                        else if (ssDCval >= 64) {
                                                            if (negBalance < 34) {
                                                                return 4;
                                                            }
                                                            else if (negBalance >= 34) {
                                                                return 2;
                                                            }
                                                        }
                                                    }
                                                    else if (mu_mv_mag_sq >= 56) {
                                                        if (SC < 4222) {
                                                            if (refDCval < 72) {
                                                                if (AFD < 5) {
                                                                    return 8;
                                                                }
                                                                else if (AFD >= 5) {
                                                                    if (SC < 2545) {
                                                                        if (ssDCval < 66) {
                                                                            return 8;
                                                                        }
                                                                        else if (ssDCval >= 66) {
                                                                            if (SC < 1992) {
                                                                                return 4;
                                                                            }
                                                                            else if (SC >= 1992) {
                                                                                return 2;
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (SC >= 2545) {
                                                                        return 4;
                                                                    }
                                                                }
                                                            }
                                                            else if (refDCval >= 72) {
                                                                if (TSC < 6) {
                                                                    if (posBalance < 12) {
                                                                        return 8;
                                                                    }
                                                                    else if (posBalance >= 12) {
                                                                        if (mu_mv_mag_sq < 93) {
                                                                            if (Cs < 283) {
                                                                                return 8;
                                                                            }
                                                                            else if (Cs >= 283) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                        else if (mu_mv_mag_sq >= 93) {
                                                                            if (tcor < 97) {
                                                                                return 4;
                                                                            }
                                                                            else if (tcor >= 97) {
                                                                                if (SC < 591) {
                                                                                    return 4;
                                                                                }
                                                                                else if (SC >= 591) {
                                                                                    if (negBalance < 49) {
                                                                                        return 8;
                                                                                    }
                                                                                    else if (negBalance >= 49) {
                                                                                        return 4;
                                                                                    }
                                                                                }
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                                else if (TSC >= 6) {
                                                                    return 2;
                                                                }
                                                            }
                                                        }
                                                        else if (SC >= 4222) {
                                                            if (RsDiff < 5) {
                                                                if (SC < 4343) {
                                                                    return 8;
                                                                }
                                                                else if (SC >= 4343) {
                                                                    return 1;
                                                                }
                                                            }
                                                            else if (RsDiff >= 5) {
                                                                return 1;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (refDCval >= 104) {
                                                    if (Rs < 2669) {
                                                        if (SC < 1973) {
                                                            if (Rs < 222) {
                                                                return 8;
                                                            }
                                                            else if (Rs >= 222) {
                                                                if (negBalance < 30) {
                                                                    if (Cs < 733) {
                                                                        return 8;
                                                                    }
                                                                    else if (Cs >= 733) {
                                                                        return 1;
                                                                    }
                                                                }
                                                                else if (negBalance >= 30) {
                                                                    if (Rs < 452) {
                                                                        return 2;
                                                                    }
                                                                    else if (Rs >= 452) {
                                                                        return 4;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (SC >= 1973) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (Rs >= 2669) {
                                                        if (RsDiff < 5) {
                                                            return 8;
                                                        }
                                                        else if (RsDiff >= 5) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (ssDCval >= 152) {
                                                if (Rs < 750) {
                                                    if (SC < 879) {
                                                        return 4;
                                                    }
                                                    else if (SC >= 879) {
                                                        if (tcor < 97) {
                                                            return 4;
                                                        }
                                                        else if (tcor >= 97) {
                                                            if (posBalance < 23) {
                                                                return 4;
                                                            }
                                                            else if (posBalance >= 23) {
                                                                if (Rs < 539) {
                                                                    if (ssDCval < 162) {
                                                                        return 4;
                                                                    }
                                                                    else if (ssDCval >= 162) {
                                                                        return 2;
                                                                    }
                                                                }
                                                                else if (Rs >= 539) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (Rs >= 750) {
                                                    return 8;
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                            else if (SC >= 4942) {
                                if (CsDiff < 6) {
                                    if (SC < 7540) {
                                        return 8;
                                    }
                                    else if (SC >= 7540) {
                                        if (SC < 7547) {
                                            return 4;
                                        }
                                        else if (SC >= 7547) {
                                            return 8;
                                        }
                                    }
                                }
                                else if (CsDiff >= 6) {
                                    return 4;
                                }
                            }
                        }
                    }
                }
                else if (MVDiff >= 10) {
                    if (posBalance < 45) {
                        if (Rs < 1013) {
                            if (mu_mv_mag_sq < 77) {
                                if (tcor < 125) {
                                    if (posBalance < 29) {
                                        if (gchDC < 4) {
                                            if (negBalance < 30) {
                                                if (negBalance < 2) {
                                                    if (posBalance < 28) {
                                                        if (MVDiff < 274) {
                                                            return 1;
                                                        }
                                                        else if (MVDiff >= 274) {
                                                            return 2;
                                                        }
                                                    }
                                                    else if (posBalance >= 28) {
                                                        return 8;
                                                    }
                                                }
                                                else if (negBalance >= 2) {
                                                    if (MVDiff < 61) {
                                                        if (tcor < 103) {
                                                            if (MVDiff < 46) {
                                                                if (Rs < 395) {
                                                                    if (posBalance < 8) {
                                                                        return 4;
                                                                    }
                                                                    else if (posBalance >= 8) {
                                                                        return 8;
                                                                    }
                                                                }
                                                                else if (Rs >= 395) {
                                                                    return 1;
                                                                }
                                                            }
                                                            else if (MVDiff >= 46) {
                                                                if (diffMVdiffVal < 17) {
                                                                    return 1;
                                                                }
                                                                else if (diffMVdiffVal >= 17) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                        else if (tcor >= 103) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (MVDiff >= 61) {
                                                        if (mu_mv_mag_sq < 32) {
                                                            if (mu_mv_mag_sq < 25) {
                                                                if (posBalance < 25) {
                                                                    if (mcTcor < 24) {
                                                                        return 4;
                                                                    }
                                                                    else if (mcTcor >= 24) {
                                                                        return 8;
                                                                    }
                                                                }
                                                                else if (posBalance >= 25) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (mu_mv_mag_sq >= 25) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (mu_mv_mag_sq >= 32) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (negBalance >= 30) {
                                                if (MVDiff < 58) {
                                                    if (mcTcor < 90) {
                                                        if (Cs < 39) {
                                                            return 4;
                                                        }
                                                        else if (Cs >= 39) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (mcTcor >= 90) {
                                                        if (negBalance < 54) {
                                                            if (diffTSC < -1) {
                                                                return 2;
                                                            }
                                                            else if (diffTSC >= -1) {
                                                                if (Cs < 8) {
                                                                    return 4;
                                                                }
                                                                else if (Cs >= 8) {
                                                                    return 2;
                                                                }
                                                            }
                                                        }
                                                        else if (negBalance >= 54) {
                                                            if (AFD < 2) {
                                                                return 8;
                                                            }
                                                            else if (AFD >= 2) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (MVDiff >= 58) {
                                                    if (posBalance < 5) {
                                                        if (diffMVdiffVal < 179) {
                                                            if (MVDiff < 98) {
                                                                return 2;
                                                            }
                                                            else if (MVDiff >= 98) {
                                                                return 1;
                                                            }
                                                        }
                                                        else if (diffMVdiffVal >= 179) {
                                                            if (MVDiff < 347) {
                                                                return 8;
                                                            }
                                                            else if (MVDiff >= 347) {
                                                                if (TSC < 1) {
                                                                    if (negBalance < 70) {
                                                                        return 2;
                                                                    }
                                                                    else if (negBalance >= 70) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (TSC >= 1) {
                                                                    return 2;
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (posBalance >= 5) {
                                                        return 1;
                                                    }
                                                }
                                            }
                                        }
                                        else if (gchDC >= 4) {
                                            if (CsDiff < 3) {
                                                return 8;
                                            }
                                            else if (CsDiff >= 3) {
                                                return 2;
                                            }
                                        }
                                    }
                                    else if (posBalance >= 29) {
                                        if (negBalance < 62) {
                                            if (posBalance < 31) {
                                                return 4;
                                            }
                                            else if (posBalance >= 31) {
                                                if (MVDiff < 17) {
                                                    if (Cs < 416) {
                                                        if (Rs < 653) {
                                                            return 2;
                                                        }
                                                        else if (Rs >= 653) {
                                                            return 1;
                                                        }
                                                    }
                                                    else if (Cs >= 416) {
                                                        return 4;
                                                    }
                                                }
                                                else if (MVDiff >= 17) {
                                                    return 2;
                                                }
                                            }
                                        }
                                        else if (negBalance >= 62) {
                                            return 1;
                                        }
                                    }
                                }
                                else if (tcor >= 125) {
                                    if (diffAFD < -1) {
                                        return 2;
                                    }
                                    else if (diffAFD >= -1) {
                                        if (posBalance < 2) {
                                            if (ssDCval < 22) {
                                                if (AFD < 3) {
                                                    return 2;
                                                }
                                                else if (AFD >= 3) {
                                                    return 4;
                                                }
                                            }
                                            else if (ssDCval >= 22) {
                                                return 8;
                                            }
                                        }
                                        else if (posBalance >= 2) {
                                            return 4;
                                        }
                                    }
                                }
                            }
                            else if (mu_mv_mag_sq >= 77) {
                                if (SC < 1470) {
                                    if (ssDCval < 164) {
                                        if (diffAFD < -43) {
                                            return 2;
                                        }
                                        else if (diffAFD >= -43) {
                                            if (diffMVdiffVal < -14) {
                                                return 2;
                                            }
                                            else if (diffMVdiffVal >= -14) {
                                                return 4;
                                            }
                                        }
                                    }
                                    else if (ssDCval >= 164) {
                                        return 2;
                                    }
                                }
                                else if (SC >= 1470) {
                                    if (negBalance < 17) {
                                        return 1;
                                    }
                                    else if (negBalance >= 17) {
                                        if (SC < 2062) {
                                            return 4;
                                        }
                                        else if (SC >= 2062) {
                                            return 8;
                                        }
                                    }
                                }
                            }
                        }
                        else if (Rs >= 1013) {
                            if (refDCval < 109) {
                                if (Cs < 707) {
                                    if (diffRsCsdiff < -1) {
                                        if (Rs < 1872) {
                                            return 8;
                                        }
                                        else if (Rs >= 1872) {
                                            if (Cs < 346) {
                                                return 2;
                                            }
                                            else if (Cs >= 346) {
                                                return 4;
                                            }
                                        }
                                    }
                                    else if (diffRsCsdiff >= -1) {
                                        return 4;
                                    }
                                }
                                else if (Cs >= 707) {
                                    if (SC < 2746) {
                                        if (Cs < 839) {
                                            if (TSCindex < 6) {
                                                return 8;
                                            }
                                            else if (TSCindex >= 6) {
                                                return 4;
                                            }
                                        }
                                        else if (Cs >= 839) {
                                            if (Rs < 1387) {
                                                if (Cs < 1272) {
                                                    return 4;
                                                }
                                                else if (Cs >= 1272) {
                                                    return 8;
                                                }
                                            }
                                            else if (Rs >= 1387) {
                                                return 1;
                                            }
                                        }
                                    }
                                    else if (SC >= 2746) {
                                        if (MVDiff < 89) {
                                            return 8;
                                        }
                                        else if (MVDiff >= 89) {
                                            if (Rs < 2542) {
                                                return 2;
                                            }
                                            else if (Rs >= 2542) {
                                                return 4;
                                            }
                                        }
                                    }
                                }
                            }
                            else if (refDCval >= 109) {
                                if (diffMVdiffVal < 5) {
                                    return 4;
                                }
                                else if (diffMVdiffVal >= 5) {
                                    if (diffMVdiffVal < 11) {
                                        if (MVDiff < 13) {
                                            return 8;
                                        }
                                        else if (MVDiff >= 13) {
                                            if (Rs < 3033) {
                                                return 8;
                                            }
                                            else if (Rs >= 3033) {
                                                return 4;
                                            }
                                        }
                                    }
                                    else if (diffMVdiffVal >= 11) {
                                        return 4;
                                    }
                                }
                            }
                        }
                    }
                    else if (posBalance >= 45) {
                        if (Cs < 526) {
                            if (MVDiff < 44) {
                                if (Rs < 291) {
                                    if (negBalance < 10) {
                                        if (posBalance < 113) {
                                            if (tcor < 63) {
                                                if (Rs < 3) {
                                                    return 2;
                                                }
                                                else if (Rs >= 3) {
                                                    return 4;
                                                }
                                            }
                                            else if (tcor >= 63) {
                                                if (MVDiff < 12) {
                                                    return 8;
                                                }
                                                else if (MVDiff >= 12) {
                                                    return 4;
                                                }
                                            }
                                        }
                                        else if (posBalance >= 113) {
                                            return 2;
                                        }
                                    }
                                    else if (negBalance >= 10) {
                                        if (refDCval < 88) {
                                            if (posBalance < 96) {
                                                if (MVDiff < 36) {
                                                    return 2;
                                                }
                                                else if (MVDiff >= 36) {
                                                    return 1;
                                                }
                                            }
                                            else if (posBalance >= 96) {
                                                return 1;
                                            }
                                        }
                                        else if (refDCval >= 88) {
                                            return 4;
                                        }
                                    }
                                }
                                else if (Rs >= 291) {
                                    if (mu_mv_mag_sq < 64) {
                                        if (Cs < 382) {
                                            if (tcor < 81) {
                                                if (Rs < 392) {
                                                    if (Cs < 327) {
                                                        return 8;
                                                    }
                                                    else if (Cs >= 327) {
                                                        return 4;
                                                    }
                                                }
                                                else if (Rs >= 392) {
                                                    if (diffRsCsdiff < 1) {
                                                        if (ssDCval < 38) {
                                                            if (Rs < 436) {
                                                                if (MVDiff < 28) {
                                                                    return 1;
                                                                }
                                                                else if (MVDiff >= 28) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (Rs >= 436) {
                                                                return 1;
                                                            }
                                                        }
                                                        else if (ssDCval >= 38) {
                                                            return 1;
                                                        }
                                                    }
                                                    else if (diffRsCsdiff >= 1) {
                                                        if (Cs < 347) {
                                                            return 4;
                                                        }
                                                        else if (Cs >= 347) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (tcor >= 81) {
                                                if (Rs < 572) {
                                                    if (refDCval < 38) {
                                                        if (Rs < 428) {
                                                            if (mu_mv_mag_sq < 58) {
                                                                return 1;
                                                            }
                                                            else if (mu_mv_mag_sq >= 58) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (Rs >= 428) {
                                                            return 1;
                                                        }
                                                    }
                                                    else if (refDCval >= 38) {
                                                        return 4;
                                                    }
                                                }
                                                else if (Rs >= 572) {
                                                    return 1;
                                                }
                                            }
                                        }
                                        else if (Cs >= 382) {
                                            if (SC < 1107) {
                                                if (ssDCval < 40) {
                                                    if (negBalance < 47) {
                                                        return 2;
                                                    }
                                                    else if (negBalance >= 47) {
                                                        if (Cs < 413) {
                                                            if (posBalance < 58) {
                                                                return 4;
                                                            }
                                                            else if (posBalance >= 58) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (Cs >= 413) {
                                                            return 2;
                                                        }
                                                    }
                                                }
                                                else if (ssDCval >= 40) {
                                                    return 4;
                                                }
                                            }
                                            else if (SC >= 1107) {
                                                if (diffAFD < -1) {
                                                    return 8;
                                                }
                                                else if (diffAFD >= -1) {
                                                    if (diffTSC < -1) {
                                                        return 1;
                                                    }
                                                    else if (diffTSC >= -1) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (mu_mv_mag_sq >= 64) {
                                        if (Cs < 403) {
                                            if (SC < 737) {
                                                if (posBalance < 54) {
                                                    return 4;
                                                }
                                                else if (posBalance >= 54) {
                                                    return 8;
                                                }
                                            }
                                            else if (SC >= 737) {
                                                if (RsDiff < 1) {
                                                    return 8;
                                                }
                                                else if (RsDiff >= 1) {
                                                    return 4;
                                                }
                                            }
                                        }
                                        else if (Cs >= 403) {
                                            if (mcTcor < 83) {
                                                return 4;
                                            }
                                            else if (mcTcor >= 83) {
                                                if (ssDCval < 39) {
                                                    return 2;
                                                }
                                                else if (ssDCval >= 39) {
                                                    if (RsCsDiff < 1) {
                                                        return 2;
                                                    }
                                                    else if (RsCsDiff >= 1) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                            else if (MVDiff >= 44) {
                                if (mu_mv_mag_sq < 8) {
                                    if (ssDCval < 20) {
                                        if (AFD < 1) {
                                            return 4;
                                        }
                                        else if (AFD >= 1) {
                                            return 8;
                                        }
                                    }
                                    else if (ssDCval >= 20) {
                                        if (refDCval < 22) {
                                            if (mcTcor < 82) {
                                                return 1;
                                            }
                                            else if (mcTcor >= 82) {
                                                return 8;
                                            }
                                        }
                                        else if (refDCval >= 22) {
                                            return 1;
                                        }
                                    }
                                }
                                else if (mu_mv_mag_sq >= 8) {
                                    if (negBalance < 2) {
                                        if (posBalance < 110) {
                                            if (posBalance < 90) {
                                                if (tcor < 68) {
                                                    if (ssDCval < 21) {
                                                        if (diffMVdiffVal < 0) {
                                                            return 2;
                                                        }
                                                        else if (diffMVdiffVal >= 0) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (ssDCval >= 21) {
                                                        return 4;
                                                    }
                                                }
                                                else if (tcor >= 68) {
                                                    if (Rs < 15) {
                                                        return 8;
                                                    }
                                                    else if (Rs >= 15) {
                                                        return 2;
                                                    }
                                                }
                                            }
                                            else if (posBalance >= 90) {
                                                if (TSC < 1) {
                                                    if (mcTcor < 91) {
                                                        return 2;
                                                    }
                                                    else if (mcTcor >= 91) {
                                                        return 8;
                                                    }
                                                }
                                                else if (TSC >= 1) {
                                                    return 8;
                                                }
                                            }
                                        }
                                        else if (posBalance >= 110) {
                                            if (diffTSC < 1) {
                                                if (TSC < 3) {
                                                    return 4;
                                                }
                                                else if (TSC >= 3) {
                                                    return 2;
                                                }
                                            }
                                            else if (diffTSC >= 1) {
                                                if (TSC < 2) {
                                                    if (posBalance < 119) {
                                                        return 4;
                                                    }
                                                    else if (posBalance >= 119) {
                                                        return 1;
                                                    }
                                                }
                                                else if (TSC >= 2) {
                                                    return 1;
                                                }
                                            }
                                        }
                                    }
                                    else if (negBalance >= 2) {
                                        if (Cs < 224) {
                                            if (gchDC < 4) {
                                                return 1;
                                            }
                                            else if (gchDC >= 4) {
                                                if (Cs < 101) {
                                                    return 8;
                                                }
                                                else if (Cs >= 101) {
                                                    return 1;
                                                }
                                            }
                                        }
                                        else if (Cs >= 224) {
                                            if (Rs < 690) {
                                                return 8;
                                            }
                                            else if (Rs >= 690) {
                                                return 4;
                                            }
                                        }
                                    }
                                }
                            }
                        }
                        else if (Cs >= 526) {
                            if (mu_mv_mag_sq < 63) {
                                if (mu_mv_mag_sq < 27) {
                                    if (ssDCval < 50) {
                                        return 2;
                                    }
                                    else if (ssDCval >= 50) {
                                        if (tcor < 90) {
                                            return 1;
                                        }
                                        else if (tcor >= 90) {
                                            return 8;
                                        }
                                    }
                                }
                                else if (mu_mv_mag_sq >= 27) {
                                    if (ssDCval < 39) {
                                        if (SC < 1485) {
                                            if (RsCsDiff < 7) {
                                                if (MVDiff < 34) {
                                                    if (mcTcor < 81) {
                                                        return 4;
                                                    }
                                                    else if (mcTcor >= 81) {
                                                        if (diffMVdiffVal < 7) {
                                                            if (negBalance < 47) {
                                                                return 4;
                                                            }
                                                            else if (negBalance >= 47) {
                                                                return 1;
                                                            }
                                                        }
                                                        else if (diffMVdiffVal >= 7) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                                else if (MVDiff >= 34) {
                                                    if (posBalance < 57) {
                                                        return 4;
                                                    }
                                                    else if (posBalance >= 57) {
                                                        return 1;
                                                    }
                                                }
                                            }
                                            else if (RsCsDiff >= 7) {
                                                if (SC < 1400) {
                                                    if (diffAFD < 1) {
                                                        return 1;
                                                    }
                                                    else if (diffAFD >= 1) {
                                                        if (Rs < 781) {
                                                            return 1;
                                                        }
                                                        else if (Rs >= 781) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                                else if (SC >= 1400) {
                                                    return 4;
                                                }
                                            }
                                        }
                                        else if (SC >= 1485) {
                                            if (Rs < 1821) {
                                                if (diffMVdiffVal < 24) {
                                                    return 2;
                                                }
                                                else if (diffMVdiffVal >= 24) {
                                                    if (Cs < 1038) {
                                                        return 1;
                                                    }
                                                    else if (Cs >= 1038) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                            else if (Rs >= 1821) {
                                                return 1;
                                            }
                                        }
                                    }
                                    else if (ssDCval >= 39) {
                                        if (Cs < 1541) {
                                            if (mu_mv_mag_sq < 40) {
                                                if (MVDiff < 15) {
                                                    if (negBalance < 43) {
                                                        return 4;
                                                    }
                                                    else if (negBalance >= 43) {
                                                        return 1;
                                                    }
                                                }
                                                else if (MVDiff >= 15) {
                                                    return 1;
                                                }
                                            }
                                            else if (mu_mv_mag_sq >= 40) {
                                                if (Cs < 1283) {
                                                    if (Rs < 1157) {
                                                        if (SC < 1381) {
                                                            return 1;
                                                        }
                                                        else if (SC >= 1381) {
                                                            if (diffMVdiffVal < -5) {
                                                                if (SC < 1405) {
                                                                    return 4;
                                                                }
                                                                else if (SC >= 1405) {
                                                                    return 1;
                                                                }
                                                            }
                                                            else if (diffMVdiffVal >= -5) {
                                                                if (RsCsDiff < 16) {
                                                                    if (diffTSC < -1) {
                                                                        if (Rs < 1037) {
                                                                            return 1;
                                                                        }
                                                                        else if (Rs >= 1037) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                    else if (diffTSC >= -1) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (RsCsDiff >= 16) {
                                                                    return 1;
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (Rs >= 1157) {
                                                        if (Cs < 1012) {
                                                            return 1;
                                                        }
                                                        else if (Cs >= 1012) {
                                                            if (TSC < 5) {
                                                                return 4;
                                                            }
                                                            else if (TSC >= 5) {
                                                                return 1;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (Cs >= 1283) {
                                                    return 4;
                                                }
                                            }
                                        }
                                        else if (Cs >= 1541) {
                                            if (RsDiff < 4) {
                                                if (gchDC < 1) {
                                                    return 1;
                                                }
                                                else if (gchDC >= 1) {
                                                    return 4;
                                                }
                                            }
                                            else if (RsDiff >= 4) {
                                                return 4;
                                            }
                                        }
                                    }
                                }
                            }
                            else if (mu_mv_mag_sq >= 63) {
                                if (Cs < 1595) {
                                    if (Rs < 255) {
                                        return 2;
                                    }
                                    else if (Rs >= 255) {
                                        if (diffRsCsdiff < -6) {
                                            if (ssDCval < 71) {
                                                if (SC < 1751) {
                                                    return 1;
                                                }
                                                else if (SC >= 1751) {
                                                    return 8;
                                                }
                                            }
                                            else if (ssDCval >= 71) {
                                                return 4;
                                            }
                                        }
                                        else if (diffRsCsdiff >= -6) {
                                            if (ssDCval < 43) {
                                                if (diffMVdiffVal < -8) {
                                                    if (Rs < 1852) {
                                                        return 4;
                                                    }
                                                    else if (Rs >= 1852) {
                                                        return 1;
                                                    }
                                                }
                                                else if (diffMVdiffVal >= -8) {
                                                    if (diffRsCsdiff < 5) {
                                                        if (Cs < 566) {
                                                            if (Rs < 768) {
                                                                return 1;
                                                            }
                                                            else if (Rs >= 768) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (Cs >= 566) {
                                                            if (SC < 3340) {
                                                                if (Rs < 941) {
                                                                    return 4;
                                                                }
                                                                else if (Rs >= 941) {
                                                                    if (MVDiff < 20) {
                                                                        return 8;
                                                                    }
                                                                    else if (MVDiff >= 20) {
                                                                        return 4;
                                                                    }
                                                                }
                                                            }
                                                            else if (SC >= 3340) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                    else if (diffRsCsdiff >= 5) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                            else if (ssDCval >= 43) {
                                                if (refDCval < 61) {
                                                    if (diffTSC < -1) {
                                                        return 4;
                                                    }
                                                    else if (diffTSC >= -1) {
                                                        return 1;
                                                    }
                                                }
                                                else if (refDCval >= 61) {
                                                    return 4;
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (Cs >= 1595) {
                                    if (negBalance < 44) {
                                        if (CsDiff < 5) {
                                            return 8;
                                        }
                                        else if (CsDiff >= 5) {
                                            return 4;
                                        }
                                    }
                                    else if (negBalance >= 44) {
                                        return 4;
                                    }
                                }
                            }
                        }
                    }
                }
            }
            else if (mu_mv_mag_sq >= 108) {
                if (negBalance < 3) {
                    if (refDCval < 103) {
                        if (diffMVdiffVal < 73) {
                            if (diffMVdiffVal < -2) {
                                if (Rs < 1078) {
                                    return 4;
                                }
                                else if (Rs >= 1078) {
                                    return 8;
                                }
                            }
                            else if (diffMVdiffVal >= -2) {
                                if (Rs < 1077) {
                                    if (MVDiff < 3) {
                                        return 8;
                                    }
                                    else if (MVDiff >= 3) {
                                        if (Rs < 1045) {
                                            return 8;
                                        }
                                        else if (Rs >= 1045) {
                                            return 2;
                                        }
                                    }
                                }
                                else if (Rs >= 1077) {
                                    return 8;
                                }
                            }
                        }
                        else if (diffMVdiffVal >= 73) {
                            if (AFD < 3) {
                                return 1;
                            }
                            else if (AFD >= 3) {
                                return 4;
                            }
                        }
                    }
                    else if (refDCval >= 103) {
                        if (Rs < 3779) {
                            if (SC < 7764) {
                                if (TSCindex < 1) {
                                    if (posBalance < 1) {
                                        return 1;
                                    }
                                    else if (posBalance >= 1) {
                                        if (SC < 7310) {
                                            return 8;
                                        }
                                        else if (SC >= 7310) {
                                            if (Cs < 3849) {
                                                return 1;
                                            }
                                            else if (Cs >= 3849) {
                                                return 8;
                                            }
                                        }
                                    }
                                }
                                else if (TSCindex >= 1) {
                                    if (Cs < 3876) {
                                        if (mu_mv_mag_sq < 124) {
                                            if (diffRsCsdiff < 9) {
                                                return 8;
                                            }
                                            else if (diffRsCsdiff >= 9) {
                                                return 1;
                                            }
                                        }
                                        else if (mu_mv_mag_sq >= 124) {
                                            if (Rs < 3569) {
                                                if (SC < 3004) {
                                                    return 1;
                                                }
                                                else if (SC >= 3004) {
                                                    return 8;
                                                }
                                            }
                                            else if (Rs >= 3569) {
                                                return 1;
                                            }
                                        }
                                    }
                                    else if (Cs >= 3876) {
                                        return 8;
                                    }
                                }
                            }
                            else if (SC >= 7764) {
                                return 1;
                            }
                        }
                        else if (Rs >= 3779) {
                            if (Cs < 2780) {
                                return 2;
                            }
                            else if (Cs >= 2780) {
                                return 8;
                            }
                        }
                    }
                }
                else if (negBalance >= 3) {
                    if (Rs < 584) {
                        if (SC < 1276) {
                            if (Cs < 516) {
                                if (Rs < 173) {
                                    return 4;
                                }
                                else if (Rs >= 173) {
                                    if (negBalance < 23) {
                                        if (mu_mv_mag_sq < 110) {
                                            return 4;
                                        }
                                        else if (mu_mv_mag_sq >= 110) {
                                            if (Scindex < 3) {
                                                if (Cs < 308) {
                                                    return 8;
                                                }
                                                else if (Cs >= 308) {
                                                    if (SC < 485) {
                                                        return 8;
                                                    }
                                                    else if (SC >= 485) {
                                                        if (negBalance < 12) {
                                                            return 8;
                                                        }
                                                        else if (negBalance >= 12) {
                                                            if (posBalance < 30) {
                                                                if (posBalance < 26) {
                                                                    return 4;
                                                                }
                                                                else if (posBalance >= 26) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (posBalance >= 30) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                            else if (Scindex >= 3) {
                                                return 8;
                                            }
                                        }
                                    }
                                    else if (negBalance >= 23) {
                                        if (Rs < 498) {
                                            return 4;
                                        }
                                        else if (Rs >= 498) {
                                            return 8;
                                        }
                                    }
                                }
                            }
                            else if (Cs >= 516) {
                                if (gchDC < 1) {
                                    if (SC < 922) {
                                        return 8;
                                    }
                                    else if (SC >= 922) {
                                        return 4;
                                    }
                                }
                                else if (gchDC >= 1) {
                                    if (ssDCval < 73) {
                                        return 4;
                                    }
                                    else if (ssDCval >= 73) {
                                        return 8;
                                    }
                                }
                            }
                        }
                        else if (SC >= 1276) {
                            return 8;
                        }
                    }
                    else if (Rs >= 584) {
                        if (negBalance < 35) {
                            if (ssDCval < 28) {
                                if (negBalance < 7) {
                                    if (SC < 1620) {
                                        return 8;
                                    }
                                    else if (SC >= 1620) {
                                        return 4;
                                    }
                                }
                                else if (negBalance >= 7) {
                                    return 8;
                                }
                            }
                            else if (ssDCval >= 28) {
                                if (posBalance < 5) {
                                    if (RsDiff < 4) {
                                        if (ssDCval < 75) {
                                            return 8;
                                        }
                                        else if (ssDCval >= 75) {
                                            return 4;
                                        }
                                    }
                                    else if (RsDiff >= 4) {
                                        if (mu_mv_mag_sq < 114) {
                                            return 8;
                                        }
                                        else if (mu_mv_mag_sq >= 114) {
                                            return 1;
                                        }
                                    }
                                }
                                else if (posBalance >= 5) {
                                    if (mu_mv_mag_sq < 125) {
                                        if (refDCval < 70) {
                                            if (mcTcor < 102) {
                                                if (ssDCval < 51) {
                                                    if (Cs < 1771) {
                                                        if (diffTSC < -1) {
                                                            if (diffMVdiffVal < -2) {
                                                                if (posBalance < 26) {
                                                                    return 8;
                                                                }
                                                                else if (posBalance >= 26) {
                                                                    return 1;
                                                                }
                                                            }
                                                            else if (diffMVdiffVal >= -2) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (diffTSC >= -1) {
                                                            if (MVDiff < 3) {
                                                                return 8;
                                                            }
                                                            else if (MVDiff >= 3) {
                                                                if (posBalance < 22) {
                                                                    return 8;
                                                                }
                                                                else if (posBalance >= 22) {
                                                                    if (CsDiff < 1) {
                                                                        return 4;
                                                                    }
                                                                    else if (CsDiff >= 1) {
                                                                        return 8;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (Cs >= 1771) {
                                                        if (RsDiff < 4) {
                                                            if (Rs < 1959) {
                                                                return 4;
                                                            }
                                                            else if (Rs >= 1959) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (RsDiff >= 4) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                                else if (ssDCval >= 51) {
                                                    return 8;
                                                }
                                            }
                                            else if (mcTcor >= 102) {
                                                return 4;
                                            }
                                        }
                                        else if (refDCval >= 70) {
                                            if (Cs < 2030) {
                                                if (Cs < 709) {
                                                    if (diffRsCsdiff < -1) {
                                                        if (SC < 1624) {
                                                            return 1;
                                                        }
                                                        else if (SC >= 1624) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (diffRsCsdiff >= -1) {
                                                        return 8;
                                                    }
                                                }
                                                else if (Cs >= 709) {
                                                    if (Rs < 796) {
                                                        if (Rs < 730) {
                                                            return 8;
                                                        }
                                                        else if (Rs >= 730) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (Rs >= 796) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                            else if (Cs >= 2030) {
                                                if (SC < 6384) {
                                                    if (RsCsDiff < 7) {
                                                        if (Cs < 2185) {
                                                            return 8;
                                                        }
                                                        else if (Cs >= 2185) {
                                                            if (negBalance < 9) {
                                                                if (SC < 4864) {
                                                                    if (mcTcor < 94) {
                                                                        if (Rs < 2459) {
                                                                            return 4;
                                                                        }
                                                                        else if (Rs >= 2459) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                    else if (mcTcor >= 94) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (SC >= 4864) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (negBalance >= 9) {
                                                                if (ssDCval < 117) {
                                                                    return 8;
                                                                }
                                                                else if (ssDCval >= 117) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (RsCsDiff >= 7) {
                                                        if (Rs < 2464) {
                                                            if (SC < 4612) {
                                                                return 1;
                                                            }
                                                            else if (SC >= 4612) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (Rs >= 2464) {
                                                            if (Rs < 3162) {
                                                                if (CsDiff < 8) {
                                                                    if (Rs < 2813) {
                                                                        if (Rs < 2733) {
                                                                            return 8;
                                                                        }
                                                                        else if (Rs >= 2733) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                    else if (Rs >= 2813) {
                                                                        return 8;
                                                                    }
                                                                }
                                                                else if (CsDiff >= 8) {
                                                                    return 1;
                                                                }
                                                            }
                                                            else if (Rs >= 3162) {
                                                                if (refDCval < 108) {
                                                                    if (Cs < 3106) {
                                                                        return 8;
                                                                    }
                                                                    else if (Cs >= 3106) {
                                                                        return 1;
                                                                    }
                                                                }
                                                                else if (refDCval >= 108) {
                                                                    if (SC < 6290) {
                                                                        return 1;
                                                                    }
                                                                    else if (SC >= 6290) {
                                                                        if (tcor < 99) {
                                                                            if (SC < 6363) {
                                                                                return 8;
                                                                            }
                                                                            else if (SC >= 6363) {
                                                                                return 1;
                                                                            }
                                                                        }
                                                                        else if (tcor >= 99) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (SC >= 6384) {
                                                    if (ssDCval < 114) {
                                                        return 8;
                                                    }
                                                    else if (ssDCval >= 114) {
                                                        if (ssDCval < 120) {
                                                            return 4;
                                                        }
                                                        else if (ssDCval >= 120) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (mu_mv_mag_sq >= 125) {
                                        if (diffMVdiffVal < 12) {
                                            if (TSC < 3) {
                                                return 8;
                                            }
                                            else if (TSC >= 3) {
                                                if (mcTcor < 99) {
                                                    return 8;
                                                }
                                                else if (mcTcor >= 99) {
                                                    if (mu_mv_mag_sq < 126) {
                                                        if (ssDCval < 97) {
                                                            return 8;
                                                        }
                                                        else if (ssDCval >= 97) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (mu_mv_mag_sq >= 126) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                        }
                                        else if (diffMVdiffVal >= 12) {
                                            return 4;
                                        }
                                    }
                                }
                            }
                        }
                        else if (negBalance >= 35) {
                            if (CsDiff < 8) {
                                if (refDCval < 84) {
                                    if (Rs < 1781) {
                                        if (TSC < 6) {
                                            if (SC < 3630) {
                                                if (SC < 3444) {
                                                    return 8;
                                                }
                                                else if (SC >= 3444) {
                                                    if (tcor < 96) {
                                                        return 8;
                                                    }
                                                    else if (tcor >= 96) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                            else if (SC >= 3630) {
                                                if (SC < 3639) {
                                                    return 4;
                                                }
                                                else if (SC >= 3639) {
                                                    if (diffAFD < 1) {
                                                        return 8;
                                                    }
                                                    else if (diffAFD >= 1) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                        }
                                        else if (TSC >= 6) {
                                            if (RsDiff < 4) {
                                                if (ssDCval < 74) {
                                                    if (Cs < 1054) {
                                                        return 8;
                                                    }
                                                    else if (Cs >= 1054) {
                                                        return 2;
                                                    }
                                                }
                                                else if (ssDCval >= 74) {
                                                    return 8;
                                                }
                                            }
                                            else if (RsDiff >= 4) {
                                                if (SC < 2597) {
                                                    return 2;
                                                }
                                                else if (SC >= 2597) {
                                                    if (diffTSC < -1) {
                                                        return 2;
                                                    }
                                                    else if (diffTSC >= -1) {
                                                        if (Cs < 1558) {
                                                            if (Cs < 1548) {
                                                                return 8;
                                                            }
                                                            else if (Cs >= 1548) {
                                                                return 2;
                                                            }
                                                        }
                                                        else if (Cs >= 1558) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (Rs >= 1781) {
                                        if (SC < 4920) {
                                            if (mcTcor < 95) {
                                                if (TSC < 6) {
                                                    return 8;
                                                }
                                                else if (TSC >= 6) {
                                                    if (posBalance < 59) {
                                                        if (Rs < 1967) {
                                                            if (SC < 3331) {
                                                                return 4;
                                                            }
                                                            else if (SC >= 3331) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (Rs >= 1967) {
                                                            return 2;
                                                        }
                                                    }
                                                    else if (posBalance >= 59) {
                                                        if (refDCval < 74) {
                                                            return 2;
                                                        }
                                                        else if (refDCval >= 74) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (mcTcor >= 95) {
                                                if (SC < 3542) {
                                                    if (mu_mv_mag_sq < 122) {
                                                        if (ssDCval < 73) {
                                                            if (gchDC < 1) {
                                                                if (Cs < 1429) {
                                                                    return 8;
                                                                }
                                                                else if (Cs >= 1429) {
                                                                    return 2;
                                                                }
                                                            }
                                                            else if (gchDC >= 1) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (ssDCval >= 73) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (mu_mv_mag_sq >= 122) {
                                                        if (refDCval < 74) {
                                                            return 2;
                                                        }
                                                        else if (refDCval >= 74) {
                                                            if (Cs < 1589) {
                                                                return 8;
                                                            }
                                                            else if (Cs >= 1589) {
                                                                if (RsCsDiff < 29) {
                                                                    if (SC < 3490) {
                                                                        return 8;
                                                                    }
                                                                    else if (SC >= 3490) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (RsCsDiff >= 29) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (SC >= 3542) {
                                                    if (tcor < 94) {
                                                        return 8;
                                                    }
                                                    else if (tcor >= 94) {
                                                        if (Rs < 2442) {
                                                            return 4;
                                                        }
                                                        else if (Rs >= 2442) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (SC >= 4920) {
                                            if (Cs < 5036) {
                                                if (diffAFD < -2) {
                                                    if (AFD < 5) {
                                                        return 8;
                                                    }
                                                    else if (AFD >= 5) {
                                                        return 4;
                                                    }
                                                }
                                                else if (diffAFD >= -2) {
                                                    return 8;
                                                }
                                            }
                                            else if (Cs >= 5036) {
                                                return 4;
                                            }
                                        }
                                    }
                                }
                                else if (refDCval >= 84) {
                                    if (mu_mv_mag_sq < 123) {
                                        if (SC < 4940) {
                                            if (Cs < 1923) {
                                                if (posBalance < 50) {
                                                    if (Cs < 1467) {
                                                        return 4;
                                                    }
                                                    else if (Cs >= 1467) {
                                                        return 8;
                                                    }
                                                }
                                                else if (posBalance >= 50) {
                                                    return 8;
                                                }
                                            }
                                            else if (Cs >= 1923) {
                                                if (negBalance < 44) {
                                                    return 4;
                                                }
                                                else if (negBalance >= 44) {
                                                    if (RsDiff < 5) {
                                                        return 8;
                                                    }
                                                    else if (RsDiff >= 5) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                        }
                                        else if (SC >= 4940) {
                                            if (Rs < 2966) {
                                                if (SC < 5670) {
                                                    if (RsCsDiff < 29) {
                                                        if (RsDiff < 4) {
                                                            return 8;
                                                        }
                                                        else if (RsDiff >= 4) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (RsCsDiff >= 29) {
                                                        return 8;
                                                    }
                                                }
                                                else if (SC >= 5670) {
                                                    if (posBalance < 57) {
                                                        if (ssDCval < 112) {
                                                            return 8;
                                                        }
                                                        else if (ssDCval >= 112) {
                                                            if (mcTcor < 98) {
                                                                return 4;
                                                            }
                                                            else if (mcTcor >= 98) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                    else if (posBalance >= 57) {
                                                        if (refDCval < 115) {
                                                            if (CsDiff < 5) {
                                                                if (Cs < 3038) {
                                                                    if (Rs < 2752) {
                                                                        if (gchDC < 1) {
                                                                            if (Cs < 2984) {
                                                                                return 8;
                                                                            }
                                                                            else if (Cs >= 2984) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                        else if (gchDC >= 1) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                    else if (Rs >= 2752) {
                                                                        return 8;
                                                                    }
                                                                }
                                                                else if (Cs >= 3038) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (CsDiff >= 5) {
                                                                if (SC < 5702) {
                                                                    if (SC < 5694) {
                                                                        return 8;
                                                                    }
                                                                    else if (SC >= 5694) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (SC >= 5702) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                        else if (refDCval >= 115) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (Rs >= 2966) {
                                                return 8;
                                            }
                                        }
                                    }
                                    else if (mu_mv_mag_sq >= 123) {
                                        if (posBalance < 48) {
                                            if (Cs < 2262) {
                                                return 8;
                                            }
                                            else if (Cs >= 2262) {
                                                return 4;
                                            }
                                        }
                                        else if (posBalance >= 48) {
                                            if (Rs < 1751) {
                                                if (SC < 3839) {
                                                    return 8;
                                                }
                                                else if (SC >= 3839) {
                                                    return 4;
                                                }
                                            }
                                            else if (Rs >= 1751) {
                                                if (TSC < 6) {
                                                    return 8;
                                                }
                                                else if (TSC >= 6) {
                                                    if (refDCval < 113) {
                                                        return 8;
                                                    }
                                                    else if (refDCval >= 113) {
                                                        if (mu_mv_mag_sq < 124) {
                                                            if (Rs < 2796) {
                                                                return 8;
                                                            }
                                                            else if (Rs >= 2796) {
                                                                if (Rs < 2807) {
                                                                    return 4;
                                                                }
                                                                else if (Rs >= 2807) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                        else if (mu_mv_mag_sq >= 124) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                            else if (CsDiff >= 8) {
                                if (tcor < 95) {
                                    if (ssDCval < 80) {
                                        if (tcor < 93) {
                                            return 8;
                                        }
                                        else if (tcor >= 93) {
                                            return 4;
                                        }
                                    }
                                    else if (ssDCval >= 80) {
                                        if (diffRsCsdiff < 23) {
                                            if (gchDC < 1) {
                                                if (Rs < 4744) {
                                                    if (RsDiff < 8) {
                                                        return 8;
                                                    }
                                                    else if (RsDiff >= 8) {
                                                        return 4;
                                                    }
                                                }
                                                else if (Rs >= 4744) {
                                                    return 4;
                                                }
                                            }
                                            else if (gchDC >= 1) {
                                                if (mu_mv_mag_sq < 128) {
                                                    return 8;
                                                }
                                                else if (mu_mv_mag_sq >= 128) {
                                                    return 4;
                                                }
                                            }
                                        }
                                        else if (diffRsCsdiff >= 23) {
                                            if (CsDiff < 9) {
                                                return 8;
                                            }
                                            else if (CsDiff >= 9) {
                                                return 4;
                                            }
                                        }
                                    }
                                }
                                else if (tcor >= 95) {
                                    return 8;
                                }
                            }
                        }
                    }
                }
            }
        }
        else if (AFD >= 7) {
            if (SC < 7760) {
                if (MVDiff < 4) {
                    if (Rs < 1805) {
                        if (tcor < 83) {
                            if (refDCval < 74) {
                                return 4;
                            }
                            else if (refDCval >= 74) {
                                if (negBalance < 52) {
                                    return 8;
                                }
                                else if (negBalance >= 52) {
                                    return 4;
                                }
                            }
                        }
                        else if (tcor >= 83) {
                            if (negBalance < 47) {
                                if (ssDCval < 59) {
                                    return 1;
                                }
                                else if (ssDCval >= 59) {
                                    if (MVDiff < 1) {
                                        if (Rs < 1787) {
                                            return 2;
                                        }
                                        else if (Rs >= 1787) {
                                            return 8;
                                        }
                                    }
                                    else if (MVDiff >= 1) {
                                        if (Cs < 847) {
                                            if (Cs < 584) {
                                                return 2;
                                            }
                                            else if (Cs >= 584) {
                                                return 4;
                                            }
                                        }
                                        else if (Cs >= 847) {
                                            if (negBalance < 45) {
                                                if (Rs < 1689) {
                                                    return 8;
                                                }
                                                else if (Rs >= 1689) {
                                                    return 4;
                                                }
                                            }
                                            else if (negBalance >= 45) {
                                                if (mu_mv_mag_sq < 111) {
                                                    if (Rs < 1485) {
                                                        return 1;
                                                    }
                                                    else if (Rs >= 1485) {
                                                        return 2;
                                                    }
                                                }
                                                else if (mu_mv_mag_sq >= 111) {
                                                    return 8;
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                            else if (negBalance >= 47) {
                                if (mcTcor < 97) {
                                    if (Rs < 1770) {
                                        if (negBalance < 49) {
                                            if (posBalance < 56) {
                                                return 8;
                                            }
                                            else if (posBalance >= 56) {
                                                return 2;
                                            }
                                        }
                                        else if (negBalance >= 49) {
                                            return 2;
                                        }
                                    }
                                    else if (Rs >= 1770) {
                                        if (negBalance < 49) {
                                            if (AFD < 8) {
                                                return 8;
                                            }
                                            else if (AFD >= 8) {
                                                return 2;
                                            }
                                        }
                                        else if (negBalance >= 49) {
                                            return 2;
                                        }
                                    }
                                }
                                else if (mcTcor >= 97) {
                                    if (negBalance < 51) {
                                        return 2;
                                    }
                                    else if (negBalance >= 51) {
                                        return 4;
                                    }
                                }
                            }
                        }
                    }
                    else if (Rs >= 1805) {
                        if (ssDCval < 100) {
                            if (mcTcor < 93) {
                                if (posBalance < 58) {
                                    if (posBalance < 51) {
                                        if (Rs < 2119) {
                                            return 1;
                                        }
                                        else if (Rs >= 2119) {
                                            return 8;
                                        }
                                    }
                                    else if (posBalance >= 51) {
                                        if (mcTcor < 91) {
                                            if (Cs < 2054) {
                                                return 2;
                                            }
                                            else if (Cs >= 2054) {
                                                if (mu_mv_mag_sq < 100) {
                                                    return 2;
                                                }
                                                else if (mu_mv_mag_sq >= 100) {
                                                    return 8;
                                                }
                                            }
                                        }
                                        else if (mcTcor >= 91) {
                                            if (AFD < 17) {
                                                if (MVDiff < 1) {
                                                    return 2;
                                                }
                                                else if (MVDiff >= 1) {
                                                    return 4;
                                                }
                                            }
                                            else if (AFD >= 17) {
                                                return 2;
                                            }
                                        }
                                    }
                                }
                                else if (posBalance >= 58) {
                                    if (refDCval < 93) {
                                        if (tcor < 91) {
                                            return 8;
                                        }
                                        else if (tcor >= 91) {
                                            if (AFD < 8) {
                                                return 8;
                                            }
                                            else if (AFD >= 8) {
                                                if (negBalance < 51) {
                                                    return 2;
                                                }
                                                else if (negBalance >= 51) {
                                                    return 8;
                                                }
                                            }
                                        }
                                    }
                                    else if (refDCval >= 93) {
                                        if (Rs < 2866) {
                                            if (Rs < 2089) {
                                                return 8;
                                            }
                                            else if (Rs >= 2089) {
                                                return 2;
                                            }
                                        }
                                        else if (Rs >= 2866) {
                                            return 8;
                                        }
                                    }
                                }
                            }
                            else if (mcTcor >= 93) {
                                if (TSC < 7) {
                                    if (mu_mv_mag_sq < 112) {
                                        if (negBalance < 38) {
                                            if (Rs < 2339) {
                                                return 1;
                                            }
                                            else if (Rs >= 2339) {
                                                if (Rs < 3014) {
                                                    return 8;
                                                }
                                                else if (Rs >= 3014) {
                                                    return 2;
                                                }
                                            }
                                        }
                                        else if (negBalance >= 38) {
                                            if (Cs < 1705) {
                                                if (SC < 3971) {
                                                    return 2;
                                                }
                                                else if (SC >= 3971) {
                                                    if (Cs < 1063) {
                                                        return 4;
                                                    }
                                                    else if (Cs >= 1063) {
                                                        return 1;
                                                    }
                                                }
                                            }
                                            else if (Cs >= 1705) {
                                                if (diffAFD < -1) {
                                                    return 2;
                                                }
                                                else if (diffAFD >= -1) {
                                                    return 4;
                                                }
                                            }
                                        }
                                    }
                                    else if (mu_mv_mag_sq >= 112) {
                                        if (mu_mv_mag_sq < 116) {
                                            if (Rs < 2584) {
                                                return 4;
                                            }
                                            else if (Rs >= 2584) {
                                                return 8;
                                            }
                                        }
                                        else if (mu_mv_mag_sq >= 116) {
                                            if (posBalance < 60) {
                                                return 8;
                                            }
                                            else if (posBalance >= 60) {
                                                return 2;
                                            }
                                        }
                                    }
                                }
                                else if (TSC >= 7) {
                                    if (negBalance < 41) {
                                        if (TSCindex < 9) {
                                            return 8;
                                        }
                                        else if (TSCindex >= 9) {
                                            return 1;
                                        }
                                    }
                                    else if (negBalance >= 41) {
                                        if (refDCval < 96) {
                                            if (Scindex < 6) {
                                                if (mcTcor < 97) {
                                                    if (mcTcor < 94) {
                                                        if (diffAFD < -1) {
                                                            if (refDCval < 90) {
                                                                return 4;
                                                            }
                                                            else if (refDCval >= 90) {
                                                                return 2;
                                                            }
                                                        }
                                                        else if (diffAFD >= -1) {
                                                            if (Cs < 1777) {
                                                                return 2;
                                                            }
                                                            else if (Cs >= 1777) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                    else if (mcTcor >= 94) {
                                                        if (negBalance < 52) {
                                                            if (RsDiff < 6) {
                                                                if (ssDCval < 81) {
                                                                    if (SC < 3757) {
                                                                        return 2;
                                                                    }
                                                                    else if (SC >= 3757) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (ssDCval >= 81) {
                                                                    return 2;
                                                                }
                                                            }
                                                            else if (RsDiff >= 6) {
                                                                if (Cs < 1784) {
                                                                    return 1;
                                                                }
                                                                else if (Cs >= 1784) {
                                                                    if (Rs < 2381) {
                                                                        if (ssDCval < 94) {
                                                                            if (Rs < 2082) {
                                                                                return 4;
                                                                            }
                                                                            else if (Rs >= 2082) {
                                                                                return 2;
                                                                            }
                                                                        }
                                                                        else if (ssDCval >= 94) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                    else if (Rs >= 2381) {
                                                                        return 4;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (negBalance >= 52) {
                                                            return 2;
                                                        }
                                                    }
                                                }
                                                else if (mcTcor >= 97) {
                                                    return 1;
                                                }
                                            }
                                            else if (Scindex >= 6) {
                                                if (SC < 7632) {
                                                    return 2;
                                                }
                                                else if (SC >= 7632) {
                                                    return 8;
                                                }
                                            }
                                        }
                                        else if (refDCval >= 96) {
                                            return 4;
                                        }
                                    }
                                }
                            }
                        }
                        else if (ssDCval >= 100) {
                            if (posBalance < 59) {
                                if (SC < 4997) {
                                    if (ssDCval < 120) {
                                        if (mu_mv_mag_sq < 112) {
                                            if (ssDCval < 109) {
                                                if (SC < 4693) {
                                                    return 4;
                                                }
                                                else if (SC >= 4693) {
                                                    return 8;
                                                }
                                            }
                                            else if (ssDCval >= 109) {
                                                return 8;
                                            }
                                        }
                                        else if (mu_mv_mag_sq >= 112) {
                                            if (RsCsDiff < 63) {
                                                return 4;
                                            }
                                            else if (RsCsDiff >= 63) {
                                                if (Rs < 2306) {
                                                    return 4;
                                                }
                                                else if (Rs >= 2306) {
                                                    return 8;
                                                }
                                            }
                                        }
                                    }
                                    else if (ssDCval >= 120) {
                                        if (RsCsDiff < 46) {
                                            if (MVDiff < 3) {
                                                return 8;
                                            }
                                            else if (MVDiff >= 3) {
                                                return 2;
                                            }
                                        }
                                        else if (RsCsDiff >= 46) {
                                            if (TSC < 9) {
                                                return 4;
                                            }
                                            else if (TSC >= 9) {
                                                if (RsDiff < 7) {
                                                    return 8;
                                                }
                                                else if (RsDiff >= 7) {
                                                    if (SC < 4604) {
                                                        return 1;
                                                    }
                                                    else if (SC >= 4604) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (SC >= 4997) {
                                    if (Rs < 3558) {
                                        if (Rs < 2760) {
                                            if (SC < 5807) {
                                                if (AFD < 14) {
                                                    return 8;
                                                }
                                                else if (AFD >= 14) {
                                                    return 1;
                                                }
                                            }
                                            else if (SC >= 5807) {
                                                return 4;
                                            }
                                        }
                                        else if (Rs >= 2760) {
                                            if (Cs < 2601) {
                                                if (Cs < 2584) {
                                                    return 8;
                                                }
                                                else if (Cs >= 2584) {
                                                    if (AFD < 11) {
                                                        return 8;
                                                    }
                                                    else if (AFD >= 11) {
                                                        return 1;
                                                    }
                                                }
                                            }
                                            else if (Cs >= 2601) {
                                                return 8;
                                            }
                                        }
                                    }
                                    else if (Rs >= 3558) {
                                        if (AFD < 13) {
                                            if (RsDiff < 9) {
                                                if (TSC < 9) {
                                                    return 8;
                                                }
                                                else if (TSC >= 9) {
                                                    return 1;
                                                }
                                            }
                                            else if (RsDiff >= 9) {
                                                return 4;
                                            }
                                        }
                                        else if (AFD >= 13) {
                                            return 1;
                                        }
                                    }
                                }
                            }
                            else if (posBalance >= 59) {
                                if (negBalance < 50) {
                                    if (tcor < 96) {
                                        if (ssDCval < 109) {
                                            if (SC < 6565) {
                                                return 4;
                                            }
                                            else if (SC >= 6565) {
                                                return 8;
                                            }
                                        }
                                        else if (ssDCval >= 109) {
                                            if (mcTcor < 96) {
                                                return 8;
                                            }
                                            else if (mcTcor >= 96) {
                                                return 4;
                                            }
                                        }
                                    }
                                    else if (tcor >= 96) {
                                        return 8;
                                    }
                                }
                                else if (negBalance >= 50) {
                                    if (Cs < 1014) {
                                        if (TSC < 7) {
                                            return 4;
                                        }
                                        else if (TSC >= 7) {
                                            return 2;
                                        }
                                    }
                                    else if (Cs >= 1014) {
                                        if (negBalance < 52) {
                                            if (CsDiff < 7) {
                                                return 4;
                                            }
                                            else if (CsDiff >= 7) {
                                                if (Cs < 2310) {
                                                    if (ssDCval < 122) {
                                                        return 4;
                                                    }
                                                    else if (ssDCval >= 122) {
                                                        if (Rs < 2640) {
                                                            return 8;
                                                        }
                                                        else if (Rs >= 2640) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                                else if (Cs >= 2310) {
                                                    return 8;
                                                }
                                            }
                                        }
                                        else if (negBalance >= 52) {
                                            if (CsDiff < 10) {
                                                return 4;
                                            }
                                            else if (CsDiff >= 10) {
                                                return 1;
                                            }
                                        }
                                    }
                                }
                            }
                        }
                    }
                }
                else if (MVDiff >= 4) {
                    if (mu_mv_mag_sq < 27) {
                        if (Cs < 273) {
                            if (TSC < 6) {
                                if (RsDiff < 3) {
                                    if (Scindex < 3) {
                                        if (negBalance < 48) {
                                            if (MVDiff < 13) {
                                                return 2;
                                            }
                                            else if (MVDiff >= 13) {
                                                return 1;
                                            }
                                        }
                                        else if (negBalance >= 48) {
                                            if (Rs < 31) {
                                                return 1;
                                            }
                                            else if (Rs >= 31) {
                                                return 2;
                                            }
                                        }
                                    }
                                    else if (Scindex >= 3) {
                                        return 1;
                                    }
                                }
                                else if (RsDiff >= 3) {
                                    if (AFD < 9) {
                                        return 1;
                                    }
                                    else if (AFD >= 9) {
                                        return 2;
                                    }
                                }
                            }
                            else if (TSC >= 6) {
                                if (ssDCval < 108) {
                                    if (gchDC < 7) {
                                        if (Cs < 252) {
                                            return 1;
                                        }
                                        else if (Cs >= 252) {
                                            if (Cs < 261) {
                                                return 2;
                                            }
                                            else if (Cs >= 261) {
                                                return 1;
                                            }
                                        }
                                    }
                                    else if (gchDC >= 7) {
                                        if (TSC < 7) {
                                            if (Rs < 14) {
                                                return 2;
                                            }
                                            else if (Rs >= 14) {
                                                return 1;
                                            }
                                        }
                                        else if (TSC >= 7) {
                                            return 2;
                                        }
                                    }
                                }
                                else if (ssDCval >= 108) {
                                    if (AFD < 9) {
                                        return 4;
                                    }
                                    else if (AFD >= 9) {
                                        return 2;
                                    }
                                }
                            }
                        }
                        else if (Cs >= 273) {
                            if (AFD < 12) {
                                if (Rs < 629) {
                                    if (Rs < 507) {
                                        if (CsDiff < 4) {
                                            if (ssDCval < 67) {
                                                if (MVDiff < 29) {
                                                    return 2;
                                                }
                                                else if (MVDiff >= 29) {
                                                    return 1;
                                                }
                                            }
                                            else if (ssDCval >= 67) {
                                                if (TSCindex < 7) {
                                                    return 4;
                                                }
                                                else if (TSCindex >= 7) {
                                                    if (Rs < 337) {
                                                        return 4;
                                                    }
                                                    else if (Rs >= 337) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                        }
                                        else if (CsDiff >= 4) {
                                            return 2;
                                        }
                                    }
                                    else if (Rs >= 507) {
                                        if (posBalance < 60) {
                                            if (Cs < 680) {
                                                return 4;
                                            }
                                            else if (Cs >= 680) {
                                                return 2;
                                            }
                                        }
                                        else if (posBalance >= 60) {
                                            if (ssDCval < 144) {
                                                return 8;
                                            }
                                            else if (ssDCval >= 144) {
                                                return 4;
                                            }
                                        }
                                    }
                                }
                                else if (Rs >= 629) {
                                    if (mu_mv_mag_sq < 5) {
                                        if (posBalance < 59) {
                                            if (ssDCval < 116) {
                                                if (ssDCval < 90) {
                                                    if (CsDiff < 6) {
                                                        if (mu_mv_mag_sq < 2) {
                                                            if (Cs < 761) {
                                                                if (gchDC < 2) {
                                                                    return 2;
                                                                }
                                                                else if (gchDC >= 2) {
                                                                    if (Cs < 534) {
                                                                        return 4;
                                                                    }
                                                                    else if (Cs >= 534) {
                                                                        return 8;
                                                                    }
                                                                }
                                                            }
                                                            else if (Cs >= 761) {
                                                                if (RsCsDiff < 37) {
                                                                    if (MVDiff < 12) {
                                                                        if (ssDCval < 63) {
                                                                            return 8;
                                                                        }
                                                                        else if (ssDCval >= 63) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                    else if (MVDiff >= 12) {
                                                                        return 1;
                                                                    }
                                                                }
                                                                else if (RsCsDiff >= 37) {
                                                                    if (Cs < 1011) {
                                                                        return 1;
                                                                    }
                                                                    else if (Cs >= 1011) {
                                                                        return 4;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (mu_mv_mag_sq >= 2) {
                                                            if (RsCsDiff < 77) {
                                                                if (Rs < 1686) {
                                                                    return 2;
                                                                }
                                                                else if (Rs >= 1686) {
                                                                    return 1;
                                                                }
                                                            }
                                                            else if (RsCsDiff >= 77) {
                                                                return 1;
                                                            }
                                                        }
                                                    }
                                                    else if (CsDiff >= 6) {
                                                        if (TSC < 9) {
                                                            return 4;
                                                        }
                                                        else if (TSC >= 9) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                                else if (ssDCval >= 90) {
                                                    if (mu_mv_mag_sq < 2) {
                                                        if (gchDC < 2) {
                                                            if (diffMVdiffVal < 5) {
                                                                if (CsDiff < 3) {
                                                                    return 4;
                                                                }
                                                                else if (CsDiff >= 3) {
                                                                    if (tcor < 93) {
                                                                        if (MVDiff < 5) {
                                                                            return 4;
                                                                        }
                                                                        else if (MVDiff >= 5) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                    else if (tcor >= 93) {
                                                                        if (posBalance < 45) {
                                                                            return 4;
                                                                        }
                                                                        else if (posBalance >= 45) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                            else if (diffMVdiffVal >= 5) {
                                                                if (Cs < 1228) {
                                                                    return 2;
                                                                }
                                                                else if (Cs >= 1228) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                        else if (gchDC >= 2) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (mu_mv_mag_sq >= 2) {
                                                        if (diffRsCsdiff < 5) {
                                                            if (AFD < 8) {
                                                                return 8;
                                                            }
                                                            else if (AFD >= 8) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (diffRsCsdiff >= 5) {
                                                            if (Rs < 3349) {
                                                                if (SC < 4412) {
                                                                    return 2;
                                                                }
                                                                else if (SC >= 4412) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (Rs >= 3349) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                            else if (ssDCval >= 116) {
                                                if (RsDiff < 5) {
                                                    if (TSC < 5) {
                                                        return 4;
                                                    }
                                                    else if (TSC >= 5) {
                                                        return 1;
                                                    }
                                                }
                                                else if (RsDiff >= 5) {
                                                    if (diffRsCsdiff < -106) {
                                                        return 1;
                                                    }
                                                    else if (diffRsCsdiff >= -106) {
                                                        if (Cs < 2076) {
                                                            if (diffAFD < 2) {
                                                                if (diffAFD < -2) {
                                                                    return 8;
                                                                }
                                                                else if (diffAFD >= -2) {
                                                                    if (SC < 3766) {
                                                                        return 4;
                                                                    }
                                                                    else if (SC >= 3766) {
                                                                        if (refDCval < 127) {
                                                                            return 4;
                                                                        }
                                                                        else if (refDCval >= 127) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                            else if (diffAFD >= 2) {
                                                                return 1;
                                                            }
                                                        }
                                                        else if (Cs >= 2076) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (posBalance >= 59) {
                                            if (mcTcor < 93) {
                                                if (negBalance < 47) {
                                                    if (tcor < 82) {
                                                        if (negBalance < 39) {
                                                            return 4;
                                                        }
                                                        else if (negBalance >= 39) {
                                                            return 1;
                                                        }
                                                    }
                                                    else if (tcor >= 82) {
                                                        return 1;
                                                    }
                                                }
                                                else if (negBalance >= 47) {
                                                    if (refDCval < 63) {
                                                        return 8;
                                                    }
                                                    else if (refDCval >= 63) {
                                                        if (posBalance < 60) {
                                                            if (diffMVdiffVal < -4) {
                                                                if (AFD < 9) {
                                                                    return 2;
                                                                }
                                                                else if (AFD >= 9) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (diffMVdiffVal >= -4) {
                                                                if (SC < 4489) {
                                                                    if (diffMVdiffVal < 5) {
                                                                        if (MVDiff < 18) {
                                                                            return 4;
                                                                        }
                                                                        else if (MVDiff >= 18) {
                                                                            return 2;
                                                                        }
                                                                    }
                                                                    else if (diffMVdiffVal >= 5) {
                                                                        if (Rs < 1291) {
                                                                            return 4;
                                                                        }
                                                                        else if (Rs >= 1291) {
                                                                            if (Rs < 1541) {
                                                                                return 2;
                                                                            }
                                                                            else if (Rs >= 1541) {
                                                                                return 8;
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                                else if (SC >= 4489) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                        else if (posBalance >= 60) {
                                                            if (RsCsDiff < 9) {
                                                                if (diffAFD < 1) {
                                                                    return 4;
                                                                }
                                                                else if (diffAFD >= 1) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (RsCsDiff >= 9) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                            else if (mcTcor >= 93) {
                                                if (AFD < 11) {
                                                    if (tcor < 91) {
                                                        return 1;
                                                    }
                                                    else if (tcor >= 91) {
                                                        return 2;
                                                    }
                                                }
                                                else if (AFD >= 11) {
                                                    if (SC < 2937) {
                                                        return 4;
                                                    }
                                                    else if (SC >= 2937) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (mu_mv_mag_sq >= 5) {
                                        if (tcor < 81) {
                                            if (RsCsDiff < 29) {
                                                return 2;
                                            }
                                            else if (RsCsDiff >= 29) {
                                                if (gchDC < 2) {
                                                    if (diffAFD < -5) {
                                                        return 2;
                                                    }
                                                    else if (diffAFD >= -5) {
                                                        return 1;
                                                    }
                                                }
                                                else if (gchDC >= 2) {
                                                    return 2;
                                                }
                                            }
                                        }
                                        else if (tcor >= 81) {
                                            if (mcTcor < 94) {
                                                return 8;
                                            }
                                            else if (mcTcor >= 94) {
                                                if (Rs < 3238) {
                                                    return 2;
                                                }
                                                else if (Rs >= 3238) {
                                                    if (SC < 6559) {
                                                        return 8;
                                                    }
                                                    else if (SC >= 6559) {
                                                        return 2;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                            else if (AFD >= 12) {
                                if (Cs < 1447) {
                                    if (tcor < 78) {
                                        if (RsDiff < 5) {
                                            if (AFD < 17) {
                                                if (negBalance < 49) {
                                                    if (SC < 1259) {
                                                        return 2;
                                                    }
                                                    else if (SC >= 1259) {
                                                        return 1;
                                                    }
                                                }
                                                else if (negBalance >= 49) {
                                                    if (Cs < 621) {
                                                        return 1;
                                                    }
                                                    else if (Cs >= 621) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                            else if (AFD >= 17) {
                                                if (CsDiff < 3) {
                                                    return 8;
                                                }
                                                else if (CsDiff >= 3) {
                                                    return 1;
                                                }
                                            }
                                        }
                                        else if (RsDiff >= 5) {
                                            if (posBalance < 55) {
                                                if (CsDiff < 10) {
                                                    if (gchDC < 54) {
                                                        if (RsDiff < 6) {
                                                            if (negBalance < 53) {
                                                                return 8;
                                                            }
                                                            else if (negBalance >= 53) {
                                                                return 2;
                                                            }
                                                        }
                                                        else if (RsDiff >= 6) {
                                                            return 2;
                                                        }
                                                    }
                                                    else if (gchDC >= 54) {
                                                        return 1;
                                                    }
                                                }
                                                else if (CsDiff >= 10) {
                                                    if (SC < 2898) {
                                                        if (Cs < 1264) {
                                                            return 2;
                                                        }
                                                        else if (Cs >= 1264) {
                                                            return 1;
                                                        }
                                                    }
                                                    else if (SC >= 2898) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                            else if (posBalance >= 55) {
                                                if (SC < 2944) {
                                                    if (RsDiff < 11) {
                                                        if (SC < 2238) {
                                                            if (Cs < 871) {
                                                                if (TSC < 10) {
                                                                    if (AFD < 14) {
                                                                        return 1;
                                                                    }
                                                                    else if (AFD >= 14) {
                                                                        return 2;
                                                                    }
                                                                }
                                                                else if (TSC >= 10) {
                                                                    if (Rs < 762) {
                                                                        return 1;
                                                                    }
                                                                    else if (Rs >= 762) {
                                                                        return 4;
                                                                    }
                                                                }
                                                            }
                                                            else if (Cs >= 871) {
                                                                if (posBalance < 62) {
                                                                    if (refDCval < 74) {
                                                                        if (AFD < 13) {
                                                                            if (Cs < 971) {
                                                                                return 8;
                                                                            }
                                                                            else if (Cs >= 971) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                        else if (AFD >= 13) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                    else if (refDCval >= 74) {
                                                                        return 1;
                                                                    }
                                                                }
                                                                else if (posBalance >= 62) {
                                                                    return 1;
                                                                }
                                                            }
                                                        }
                                                        else if (SC >= 2238) {
                                                            if (posBalance < 63) {
                                                                if (CsDiff < 5) {
                                                                    return 1;
                                                                }
                                                                else if (CsDiff >= 5) {
                                                                    if (ssDCval < 58) {
                                                                        return 1;
                                                                    }
                                                                    else if (ssDCval >= 58) {
                                                                        return 8;
                                                                    }
                                                                }
                                                            }
                                                            else if (posBalance >= 63) {
                                                                if (refDCval < 123) {
                                                                    return 1;
                                                                }
                                                                else if (refDCval >= 123) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (RsDiff >= 11) {
                                                        if (diffMVdiffVal < 111) {
                                                            if (TSC < 5) {
                                                                return 8;
                                                            }
                                                            else if (TSC >= 5) {
                                                                return 2;
                                                            }
                                                        }
                                                        else if (diffMVdiffVal >= 111) {
                                                            if (Rs < 2191) {
                                                                if (MVDiff < 530) {
                                                                    return 8;
                                                                }
                                                                else if (MVDiff >= 530) {
                                                                    return 1;
                                                                }
                                                            }
                                                            else if (Rs >= 2191) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (SC >= 2944) {
                                                    if (SC < 4071) {
                                                        if (ssDCval < 116) {
                                                            if (refDCval < 107) {
                                                                if (Cs < 1286) {
                                                                    if (tcor < 49) {
                                                                        return 2;
                                                                    }
                                                                    else if (tcor >= 49) {
                                                                        if (AFD < 31) {
                                                                            return 1;
                                                                        }
                                                                        else if (AFD >= 31) {
                                                                            return 2;
                                                                        }
                                                                    }
                                                                }
                                                                else if (Cs >= 1286) {
                                                                    return 2;
                                                                }
                                                            }
                                                            else if (refDCval >= 107) {
                                                                return 2;
                                                            }
                                                        }
                                                        else if (ssDCval >= 116) {
                                                            return 1;
                                                        }
                                                    }
                                                    else if (SC >= 4071) {
                                                        if (ssDCval < 136) {
                                                            return 2;
                                                        }
                                                        else if (ssDCval >= 136) {
                                                            if (SC < 5137) {
                                                                return 4;
                                                            }
                                                            else if (SC >= 5137) {
                                                                return 1;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (tcor >= 78) {
                                        if (RsCsDiff < 93) {
                                            if (mu_mv_mag_sq < 10) {
                                                if (Cs < 751) {
                                                    if (ssDCval < 107) {
                                                        if (negBalance < 56) {
                                                            return 2;
                                                        }
                                                        else if (negBalance >= 56) {
                                                            if (Rs < 827) {
                                                                return 2;
                                                            }
                                                            else if (Rs >= 827) {
                                                                return 1;
                                                            }
                                                        }
                                                    }
                                                    else if (ssDCval >= 107) {
                                                        if (Rs < 342) {
                                                            return 1;
                                                        }
                                                        else if (Rs >= 342) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                                else if (Cs >= 751) {
                                                    if (mcTcor < 91) {
                                                        if (Rs < 1284) {
                                                            return 4;
                                                        }
                                                        else if (Rs >= 1284) {
                                                            return 2;
                                                        }
                                                    }
                                                    else if (mcTcor >= 91) {
                                                        if (refDCval < 89) {
                                                            if (RsDiff < 7) {
                                                                return 1;
                                                            }
                                                            else if (RsDiff >= 7) {
                                                                return 2;
                                                            }
                                                        }
                                                        else if (refDCval >= 89) {
                                                            if (Cs < 1288) {
                                                                return 1;
                                                            }
                                                            else if (Cs >= 1288) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                            else if (mu_mv_mag_sq >= 10) {
                                                if (gchDC < 13) {
                                                    if (refDCval < 117) {
                                                        return 2;
                                                    }
                                                    else if (refDCval >= 117) {
                                                        if (CsDiff < 6) {
                                                            if (Rs < 816) {
                                                                return 2;
                                                            }
                                                            else if (Rs >= 816) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (CsDiff >= 6) {
                                                            return 2;
                                                        }
                                                    }
                                                }
                                                else if (gchDC >= 13) {
                                                    return 8;
                                                }
                                            }
                                        }
                                        else if (RsCsDiff >= 93) {
                                            if (refDCval < 75) {
                                                if (RsDiff < 8) {
                                                    if (mcTcor < 94) {
                                                        return 1;
                                                    }
                                                    else if (mcTcor >= 94) {
                                                        return 2;
                                                    }
                                                }
                                                else if (RsDiff >= 8) {
                                                    return 4;
                                                }
                                            }
                                            else if (refDCval >= 75) {
                                                if (refDCval < 151) {
                                                    if (negBalance < 40) {
                                                        return 8;
                                                    }
                                                    else if (negBalance >= 40) {
                                                        if (TSC < 9) {
                                                            if (diffRsCsdiff < 44) {
                                                                return 2;
                                                            }
                                                            else if (diffRsCsdiff >= 44) {
                                                                return 1;
                                                            }
                                                        }
                                                        else if (TSC >= 9) {
                                                            return 2;
                                                        }
                                                    }
                                                }
                                                else if (refDCval >= 151) {
                                                    if (Rs < 977) {
                                                        return 2;
                                                    }
                                                    else if (Rs >= 977) {
                                                        return 1;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (Cs >= 1447) {
                                    if (Rs < 3153) {
                                        if (posBalance < 55) {
                                            if (refDCval < 93) {
                                                if (Rs < 1729) {
                                                    if (SC < 2180) {
                                                        return 2;
                                                    }
                                                    else if (SC >= 2180) {
                                                        return 1;
                                                    }
                                                }
                                                else if (Rs >= 1729) {
                                                    if (RsCsDiff < 176) {
                                                        if (SC < 4514) {
                                                            if (diffRsCsdiff < -42) {
                                                                if (Rs < 2010) {
                                                                    return 1;
                                                                }
                                                                else if (Rs >= 2010) {
                                                                    return 2;
                                                                }
                                                            }
                                                            else if (diffRsCsdiff >= -42) {
                                                                if (Cs < 2142) {
                                                                    if (refDCval < 92) {
                                                                        if (SC < 4395) {
                                                                            return 8;
                                                                        }
                                                                        else if (SC >= 4395) {
                                                                            if (SC < 4453) {
                                                                                return 1;
                                                                            }
                                                                            else if (SC >= 4453) {
                                                                                return 8;
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (refDCval >= 92) {
                                                                        return 1;
                                                                    }
                                                                }
                                                                else if (Cs >= 2142) {
                                                                    if (posBalance < 51) {
                                                                        return 8;
                                                                    }
                                                                    else if (posBalance >= 51) {
                                                                        return 1;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (SC >= 4514) {
                                                            if (refDCval < 86) {
                                                                if (CsDiff < 8) {
                                                                    return 8;
                                                                }
                                                                else if (CsDiff >= 8) {
                                                                    if (AFD < 14) {
                                                                        return 8;
                                                                    }
                                                                    else if (AFD >= 14) {
                                                                        if (TSC < 13) {
                                                                            return 4;
                                                                        }
                                                                        else if (TSC >= 13) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                            else if (refDCval >= 86) {
                                                                if (Rs < 2544) {
                                                                    return 4;
                                                                }
                                                                else if (Rs >= 2544) {
                                                                    return 1;
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (RsCsDiff >= 176) {
                                                        if (SC < 5724) {
                                                            if (Rs < 2844) {
                                                                if (mcTcor < 90) {
                                                                    return 2;
                                                                }
                                                                else if (mcTcor >= 90) {
                                                                    if (RsCsDiff < 183) {
                                                                        if (negBalance < 47) {
                                                                            return 1;
                                                                        }
                                                                        else if (negBalance >= 47) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                    else if (RsCsDiff >= 183) {
                                                                        return 1;
                                                                    }
                                                                }
                                                            }
                                                            else if (Rs >= 2844) {
                                                                if (TSC < 9) {
                                                                    return 4;
                                                                }
                                                                else if (TSC >= 9) {
                                                                    return 2;
                                                                }
                                                            }
                                                        }
                                                        else if (SC >= 5724) {
                                                            if (RsDiff < 18) {
                                                                return 8;
                                                            }
                                                            else if (RsDiff >= 18) {
                                                                return 1;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                            else if (refDCval >= 93) {
                                                if (refDCval < 112) {
                                                    if (tcor < 75) {
                                                        return 2;
                                                    }
                                                    else if (tcor >= 75) {
                                                        if (mcTcor < 93) {
                                                            return 4;
                                                        }
                                                        else if (mcTcor >= 93) {
                                                            if (SC < 4956) {
                                                                return 1;
                                                            }
                                                            else if (SC >= 4956) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (refDCval >= 112) {
                                                    if (ssDCval < 120) {
                                                        if (Rs < 1240) {
                                                            return 2;
                                                        }
                                                        else if (Rs >= 1240) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (ssDCval >= 120) {
                                                        if (posBalance < 45) {
                                                            return 4;
                                                        }
                                                        else if (posBalance >= 45) {
                                                            if (AFD < 28) {
                                                                return 1;
                                                            }
                                                            else if (AFD >= 28) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (posBalance >= 55) {
                                            if (tcor < 80) {
                                                if (MVDiff < 19) {
                                                    if (AFD < 18) {
                                                        if (negBalance < 51) {
                                                            if (mcTcor < 86) {
                                                                return 1;
                                                            }
                                                            else if (mcTcor >= 86) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (negBalance >= 51) {
                                                            if (ssDCval < 124) {
                                                                if (diffRsCsdiff < -48) {
                                                                    if (Rs < 2099) {
                                                                        return 2;
                                                                    }
                                                                    else if (Rs >= 2099) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (diffRsCsdiff >= -48) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (ssDCval >= 124) {
                                                                return 1;
                                                            }
                                                        }
                                                    }
                                                    else if (AFD >= 18) {
                                                        if (refDCval < 112) {
                                                            if (TSC < 12) {
                                                                if (SC < 5074) {
                                                                    if (refDCval < 83) {
                                                                        return 4;
                                                                    }
                                                                    else if (refDCval >= 83) {
                                                                        return 2;
                                                                    }
                                                                }
                                                                else if (SC >= 5074) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (TSC >= 12) {
                                                                return 1;
                                                            }
                                                        }
                                                        else if (refDCval >= 112) {
                                                            if (mu_mv_mag_sq < 15) {
                                                                if (ssDCval < 125) {
                                                                    if (TSC < 14) {
                                                                        return 1;
                                                                    }
                                                                    else if (TSC >= 14) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (ssDCval >= 125) {
                                                                    return 1;
                                                                }
                                                            }
                                                            else if (mu_mv_mag_sq >= 15) {
                                                                if (Cs < 2383) {
                                                                    return 2;
                                                                }
                                                                else if (Cs >= 2383) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (MVDiff >= 19) {
                                                    if (refDCval < 139) {
                                                        if (mu_mv_mag_sq < 7) {
                                                            if (RsDiff < 8) {
                                                                if (Rs < 1366) {
                                                                    return 1;
                                                                }
                                                                else if (Rs >= 1366) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (RsDiff >= 8) {
                                                                if (negBalance < 55) {
                                                                    if (AFD < 19) {
                                                                        if (Rs < 1516) {
                                                                            return 8;
                                                                        }
                                                                        else if (Rs >= 1516) {
                                                                            if (Cs < 2016) {
                                                                                return 4;
                                                                            }
                                                                            else if (Cs >= 2016) {
                                                                                return 8;
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (AFD >= 19) {
                                                                        if (CsDiff < 12) {
                                                                            if (Cs < 2000) {
                                                                                return 1;
                                                                            }
                                                                            else if (Cs >= 2000) {
                                                                                if (Cs < 3146) {
                                                                                    return 8;
                                                                                }
                                                                                else if (Cs >= 3146) {
                                                                                    return 1;
                                                                                }
                                                                            }
                                                                        }
                                                                        else if (CsDiff >= 12) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                }
                                                                else if (negBalance >= 55) {
                                                                    if (RsDiff < 14) {
                                                                        if (ssDCval < 128) {
                                                                            if (MVDiff < 20) {
                                                                                if (Cs < 2500) {
                                                                                    return 8;
                                                                                }
                                                                                else if (Cs >= 2500) {
                                                                                    return 4;
                                                                                }
                                                                            }
                                                                            else if (MVDiff >= 20) {
                                                                                if (diffAFD < -3) {
                                                                                    return 1;
                                                                                }
                                                                                else if (diffAFD >= -3) {
                                                                                    if (Cs < 3766) {
                                                                                        return 4;
                                                                                    }
                                                                                    else if (Cs >= 3766) {
                                                                                        if (Rs < 2101) {
                                                                                            return 1;
                                                                                        }
                                                                                        else if (Rs >= 2101) {
                                                                                            return 4;
                                                                                        }
                                                                                    }
                                                                                }
                                                                            }
                                                                        }
                                                                        else if (ssDCval >= 128) {
                                                                            if (ssDCval < 132) {
                                                                                if (SC < 4076) {
                                                                                    return 2;
                                                                                }
                                                                                else if (SC >= 4076) {
                                                                                    return 4;
                                                                                }
                                                                            }
                                                                            else if (ssDCval >= 132) {
                                                                                return 1;
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (RsDiff >= 14) {
                                                                        if (SC < 4957) {
                                                                            return 2;
                                                                        }
                                                                        else if (SC >= 4957) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (mu_mv_mag_sq >= 7) {
                                                            if (CsDiff < 10) {
                                                                if (MVDiff < 22) {
                                                                    return 4;
                                                                }
                                                                else if (MVDiff >= 22) {
                                                                    if (diffMVdiffVal < 13) {
                                                                        return 8;
                                                                    }
                                                                    else if (diffMVdiffVal >= 13) {
                                                                        if (AFD < 34) {
                                                                            return 1;
                                                                        }
                                                                        else if (AFD >= 34) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                            else if (CsDiff >= 10) {
                                                                if (diffRsCsdiff < 56) {
                                                                    if (refDCval < 53) {
                                                                        return 2;
                                                                    }
                                                                    else if (refDCval >= 53) {
                                                                        if (negBalance < 63) {
                                                                            if (MVDiff < 133) {
                                                                                if (diffAFD < 2) {
                                                                                    return 1;
                                                                                }
                                                                                else if (diffAFD >= 2) {
                                                                                    if (Rs < 2518) {
                                                                                        return 4;
                                                                                    }
                                                                                    else if (Rs >= 2518) {
                                                                                        return 1;
                                                                                    }
                                                                                }
                                                                            }
                                                                            else if (MVDiff >= 133) {
                                                                                return 8;
                                                                            }
                                                                        }
                                                                        else if (negBalance >= 63) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                }
                                                                else if (diffRsCsdiff >= 56) {
                                                                    if (AFD < 26) {
                                                                        if (diffMVdiffVal < 7) {
                                                                            return 8;
                                                                        }
                                                                        else if (diffMVdiffVal >= 7) {
                                                                            if (SC < 5626) {
                                                                                return 4;
                                                                            }
                                                                            else if (SC >= 5626) {
                                                                                return 1;
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (AFD >= 26) {
                                                                        return 1;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (refDCval >= 139) {
                                                        if (MVDiff < 24) {
                                                            return 2;
                                                        }
                                                        else if (MVDiff >= 24) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (tcor >= 80) {
                                                if (CsDiff < 7) {
                                                    if (negBalance < 53) {
                                                        if (Cs < 1681) {
                                                            return 8;
                                                        }
                                                        else if (Cs >= 1681) {
                                                            return 1;
                                                        }
                                                    }
                                                    else if (negBalance >= 53) {
                                                        if (mcTcor < 92) {
                                                            return 4;
                                                        }
                                                        else if (mcTcor >= 92) {
                                                            if (Rs < 2299) {
                                                                return 4;
                                                            }
                                                            else if (Rs >= 2299) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (CsDiff >= 7) {
                                                    if (negBalance < 52) {
                                                        if (posBalance < 58) {
                                                            if (RsDiff < 10) {
                                                                if (refDCval < 98) {
                                                                    if (refDCval < 79) {
                                                                        if (Rs < 2608) {
                                                                            return 4;
                                                                        }
                                                                        else if (Rs >= 2608) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                    else if (refDCval >= 79) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (refDCval >= 98) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (RsDiff >= 10) {
                                                                if (posBalance < 56) {
                                                                    if (CsDiff < 12) {
                                                                        if (negBalance < 51) {
                                                                            return 8;
                                                                        }
                                                                        else if (negBalance >= 51) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                    else if (CsDiff >= 12) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (posBalance >= 56) {
                                                                    if (mcTcor < 92) {
                                                                        return 4;
                                                                    }
                                                                    else if (mcTcor >= 92) {
                                                                        if (Cs < 1999) {
                                                                            return 2;
                                                                        }
                                                                        else if (Cs >= 1999) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (posBalance >= 58) {
                                                            if (RsCsDiff < 409) {
                                                                if (mcTcor < 93) {
                                                                    if (MVDiff < 10) {
                                                                        if (SC < 5228) {
                                                                            return 2;
                                                                        }
                                                                        else if (SC >= 5228) {
                                                                            if (Rs < 2894) {
                                                                                return 8;
                                                                            }
                                                                            else if (Rs >= 2894) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (MVDiff >= 10) {
                                                                        if (ssDCval < 104) {
                                                                            return 4;
                                                                        }
                                                                        else if (ssDCval >= 104) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                }
                                                                else if (mcTcor >= 93) {
                                                                    if (diffAFD < 3) {
                                                                        if (tcor < 90) {
                                                                            return 2;
                                                                        }
                                                                        else if (tcor >= 90) {
                                                                            if (Rs < 2200) {
                                                                                return 4;
                                                                            }
                                                                            else if (Rs >= 2200) {
                                                                                return 2;
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (diffAFD >= 3) {
                                                                        if (Rs < 2475) {
                                                                            return 4;
                                                                        }
                                                                        else if (Rs >= 2475) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                            else if (RsCsDiff >= 409) {
                                                                return 1;
                                                            }
                                                        }
                                                    }
                                                    else if (negBalance >= 52) {
                                                        if (Rs < 2680) {
                                                            if (TSC < 15) {
                                                                if (ssDCval < 78) {
                                                                    if (Rs < 2516) {
                                                                        return 8;
                                                                    }
                                                                    else if (Rs >= 2516) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (ssDCval >= 78) {
                                                                    if (ssDCval < 116) {
                                                                        if (gchDC < 2) {
                                                                            return 2;
                                                                        }
                                                                        else if (gchDC >= 2) {
                                                                            if (RsCsDiff < 147) {
                                                                                return 1;
                                                                            }
                                                                            else if (RsCsDiff >= 147) {
                                                                                return 2;
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (ssDCval >= 116) {
                                                                        if (mcTcor < 87) {
                                                                            return 4;
                                                                        }
                                                                        else if (mcTcor >= 87) {
                                                                            if (TSC < 13) {
                                                                                if (Cs < 1605) {
                                                                                    return 2;
                                                                                }
                                                                                else if (Cs >= 1605) {
                                                                                    return 8;
                                                                                }
                                                                            }
                                                                            else if (TSC >= 13) {
                                                                                return 2;
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                            else if (TSC >= 15) {
                                                                return 1;
                                                            }
                                                        }
                                                        else if (Rs >= 2680) {
                                                            if (CsDiff < 11) {
                                                                if (mcTcor < 90) {
                                                                    if (Cs < 2055) {
                                                                        return 4;
                                                                    }
                                                                    else if (Cs >= 2055) {
                                                                        return 8;
                                                                    }
                                                                }
                                                                else if (mcTcor >= 90) {
                                                                    if (AFD < 22) {
                                                                        return 1;
                                                                    }
                                                                    else if (AFD >= 22) {
                                                                        return 8;
                                                                    }
                                                                }
                                                            }
                                                            else if (CsDiff >= 11) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (Rs >= 3153) {
                                        if (MVDiff < 20) {
                                            if (mcTcor < 88) {
                                                if (refDCval < 115) {
                                                    if (Rs < 4337) {
                                                        return 8;
                                                    }
                                                    else if (Rs >= 4337) {
                                                        return 2;
                                                    }
                                                }
                                                else if (refDCval >= 115) {
                                                    if (CsDiff < 18) {
                                                        if (refDCval < 127) {
                                                            if (Rs < 3324) {
                                                                return 8;
                                                            }
                                                            else if (Rs >= 3324) {
                                                                return 1;
                                                            }
                                                        }
                                                        else if (refDCval >= 127) {
                                                            return 1;
                                                        }
                                                    }
                                                    else if (CsDiff >= 18) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                            else if (mcTcor >= 88) {
                                                if (CsDiff < 11) {
                                                    if (gchDC < 1) {
                                                        if (tcor < 86) {
                                                            return 2;
                                                        }
                                                        else if (tcor >= 86) {
                                                            if (Rs < 3824) {
                                                                return 1;
                                                            }
                                                            else if (Rs >= 3824) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                    else if (gchDC >= 1) {
                                                        if (Rs < 4458) {
                                                            return 1;
                                                        }
                                                        else if (Rs >= 4458) {
                                                            return 2;
                                                        }
                                                    }
                                                }
                                                else if (CsDiff >= 11) {
                                                    if (RsDiff < 16) {
                                                        if (diffMVdiffVal < 1) {
                                                            return 8;
                                                        }
                                                        else if (diffMVdiffVal >= 1) {
                                                            if (SC < 6754) {
                                                                return 1;
                                                            }
                                                            else if (SC >= 6754) {
                                                                if (Rs < 3950) {
                                                                    return 8;
                                                                }
                                                                else if (Rs >= 3950) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (RsDiff >= 16) {
                                                        if (negBalance < 54) {
                                                            if (Cs < 2074) {
                                                                return 4;
                                                            }
                                                            else if (Cs >= 2074) {
                                                                if (Cs < 2771) {
                                                                    return 2;
                                                                }
                                                                else if (Cs >= 2771) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                        else if (negBalance >= 54) {
                                                            return 2;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (MVDiff >= 20) {
                                            if (mcTcor < 75) {
                                                if (SC < 7473) {
                                                    return 1;
                                                }
                                                else if (SC >= 7473) {
                                                    return 8;
                                                }
                                            }
                                            else if (mcTcor >= 75) {
                                                if (ssDCval < 136) {
                                                    if (mu_mv_mag_sq < 15) {
                                                        if (Rs < 3221) {
                                                            return 1;
                                                        }
                                                        else if (Rs >= 3221) {
                                                            return 2;
                                                        }
                                                    }
                                                    else if (mu_mv_mag_sq >= 15) {
                                                        if (ssDCval < 112) {
                                                            if (RsDiff < 18) {
                                                                return 8;
                                                            }
                                                            else if (RsDiff >= 18) {
                                                                return 2;
                                                            }
                                                        }
                                                        else if (ssDCval >= 112) {
                                                            return 1;
                                                        }
                                                    }
                                                }
                                                else if (ssDCval >= 136) {
                                                    if (CsDiff < 9) {
                                                        return 2;
                                                    }
                                                    else if (CsDiff >= 9) {
                                                        if (RsCsDiff < 373) {
                                                            return 2;
                                                        }
                                                        else if (RsCsDiff >= 373) {
                                                            if (Rs < 3741) {
                                                                return 2;
                                                            }
                                                            else if (Rs >= 3741) {
                                                                if (diffMVdiffVal < 8) {
                                                                    return 4;
                                                                }
                                                                else if (diffMVdiffVal >= 8) {
                                                                    if (Cs < 2348) {
                                                                        return 4;
                                                                    }
                                                                    else if (Cs >= 2348) {
                                                                        return 2;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                        }
                    }
                    else if (mu_mv_mag_sq >= 27) {
                        if (tcor < 74) {
                            if (AFD < 24) {
                                if (Cs < 356) {
                                    if (negBalance < 46) {
                                        if (Cs < 105) {
                                            return 2;
                                        }
                                        else if (Cs >= 105) {
                                            if (SC < 737) {
                                                return 1;
                                            }
                                            else if (SC >= 737) {
                                                if (RsDiff < 3) {
                                                    return 4;
                                                }
                                                else if (RsDiff >= 3) {
                                                    return 8;
                                                }
                                            }
                                        }
                                    }
                                    else if (negBalance >= 46) {
                                        return 2;
                                    }
                                }
                                else if (Cs >= 356) {
                                    if (ssDCval < 127) {
                                        if (Rs < 2437) {
                                            if (negBalance < 54) {
                                                if (Cs < 2303) {
                                                    if (refDCval < 41) {
                                                        if (SC < 1334) {
                                                            if (Cs < 539) {
                                                                if (mcTcor < 74) {
                                                                    return 4;
                                                                }
                                                                else if (mcTcor >= 74) {
                                                                    return 1;
                                                                }
                                                            }
                                                            else if (Cs >= 539) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (SC >= 1334) {
                                                            return 1;
                                                        }
                                                    }
                                                    else if (refDCval >= 41) {
                                                        if (refDCval < 47) {
                                                            return 8;
                                                        }
                                                        else if (refDCval >= 47) {
                                                            if (diffRsCsdiff < 20) {
                                                                if (Rs < 1257) {
                                                                    if (diffMVdiffVal < -3) {
                                                                        if (Rs < 1199) {
                                                                            return 1;
                                                                        }
                                                                        else if (Rs >= 1199) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                    else if (diffMVdiffVal >= -3) {
                                                                        if (diffTSC < -2) {
                                                                            return 1;
                                                                        }
                                                                        else if (diffTSC >= -2) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                }
                                                                else if (Rs >= 1257) {
                                                                    if (diffMVdiffVal < 3) {
                                                                        return 1;
                                                                    }
                                                                    else if (diffMVdiffVal >= 3) {
                                                                        if (Rs < 1883) {
                                                                            return 8;
                                                                        }
                                                                        else if (Rs >= 1883) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                            else if (diffRsCsdiff >= 20) {
                                                                if (RsCsDiff < 87) {
                                                                    if (diffAFD < 11) {
                                                                        return 8;
                                                                    }
                                                                    else if (diffAFD >= 11) {
                                                                        return 1;
                                                                    }
                                                                }
                                                                else if (RsCsDiff >= 87) {
                                                                    if (posBalance < 45) {
                                                                        if (Rs < 2195) {
                                                                            return 2;
                                                                        }
                                                                        else if (Rs >= 2195) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                    else if (posBalance >= 45) {
                                                                        if (mcTcor < 91) {
                                                                            return 4;
                                                                        }
                                                                        else if (mcTcor >= 91) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (Cs >= 2303) {
                                                    return 1;
                                                }
                                            }
                                            else if (negBalance >= 54) {
                                                if (posBalance < 59) {
                                                    if (RsCsDiff < 43) {
                                                        return 8;
                                                    }
                                                    else if (RsCsDiff >= 43) {
                                                        if (MVDiff < 36) {
                                                            return 8;
                                                        }
                                                        else if (MVDiff >= 36) {
                                                            return 1;
                                                        }
                                                    }
                                                }
                                                else if (posBalance >= 59) {
                                                    return 2;
                                                }
                                            }
                                        }
                                        else if (Rs >= 2437) {
                                            if (ssDCval < 104) {
                                                if (AFD < 17) {
                                                    return 8;
                                                }
                                                else if (AFD >= 17) {
                                                    if (mcTcor < 85) {
                                                        return 8;
                                                    }
                                                    else if (mcTcor >= 85) {
                                                        if (RsCsDiff < 242) {
                                                            return 8;
                                                        }
                                                        else if (RsCsDiff >= 242) {
                                                            return 1;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (ssDCval >= 104) {
                                                if (AFD < 18) {
                                                    if (mu_mv_mag_sq < 47) {
                                                        if (ssDCval < 109) {
                                                            return 4;
                                                        }
                                                        else if (ssDCval >= 109) {
                                                            return 1;
                                                        }
                                                    }
                                                    else if (mu_mv_mag_sq >= 47) {
                                                        if (AFD < 14) {
                                                            return 4;
                                                        }
                                                        else if (AFD >= 14) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                                else if (AFD >= 18) {
                                                    if (RsDiff < 17) {
                                                        if (Cs < 3753) {
                                                            return 4;
                                                        }
                                                        else if (Cs >= 3753) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (RsDiff >= 17) {
                                                        if (TSC < 15) {
                                                            return 2;
                                                        }
                                                        else if (TSC >= 15) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (ssDCval >= 127) {
                                        if (ssDCval < 141) {
                                            if (CsDiff < 5) {
                                                return 4;
                                            }
                                            else if (CsDiff >= 5) {
                                                if (ssDCval < 130) {
                                                    if (Rs < 3510) {
                                                        return 1;
                                                    }
                                                    else if (Rs >= 3510) {
                                                        return 4;
                                                    }
                                                }
                                                else if (ssDCval >= 130) {
                                                    return 1;
                                                }
                                            }
                                        }
                                        else if (ssDCval >= 141) {
                                            if (Rs < 1380) {
                                                return 8;
                                            }
                                            else if (Rs >= 1380) {
                                                if (Cs < 1509) {
                                                    return 2;
                                                }
                                                else if (Cs >= 1509) {
                                                    return 4;
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                            else if (AFD >= 24) {
                                if (tcor < 33) {
                                    if (MVDiff < 684) {
                                        if (refDCval < 98) {
                                            if (MVDiff < 116) {
                                                return 1;
                                            }
                                            else if (MVDiff >= 116) {
                                                if (gchDC < 9) {
                                                    return 8;
                                                }
                                                else if (gchDC >= 9) {
                                                    return 4;
                                                }
                                            }
                                        }
                                        else if (refDCval >= 98) {
                                            if (Cs < 2145) {
                                                return 1;
                                            }
                                            else if (Cs >= 2145) {
                                                return 8;
                                            }
                                        }
                                    }
                                    else if (MVDiff >= 684) {
                                        return 8;
                                    }
                                }
                                else if (tcor >= 33) {
                                    if (refDCval < 91) {
                                        if (RsCsDiff < 285) {
                                            return 2;
                                        }
                                        else if (RsCsDiff >= 285) {
                                            if (Cs < 3248) {
                                                return 8;
                                            }
                                            else if (Cs >= 3248) {
                                                return 4;
                                            }
                                        }
                                    }
                                    else if (refDCval >= 91) {
                                        if (SC < 3943) {
                                            if (posBalance < 63) {
                                                if (RsCsDiff < 474) {
                                                    if (Cs < 569) {
                                                        return 8;
                                                    }
                                                    else if (Cs >= 569) {
                                                        return 2;
                                                    }
                                                }
                                                else if (RsCsDiff >= 474) {
                                                    if (Cs < 1111) {
                                                        return 1;
                                                    }
                                                    else if (Cs >= 1111) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                            else if (posBalance >= 63) {
                                                if (CsDiff < 8) {
                                                    if (SC < 3692) {
                                                        return 1;
                                                    }
                                                    else if (SC >= 3692) {
                                                        return 2;
                                                    }
                                                }
                                                else if (CsDiff >= 8) {
                                                    return 2;
                                                }
                                            }
                                        }
                                        else if (SC >= 3943) {
                                            if (negBalance < 47) {
                                                if (Cs < 1496) {
                                                    if (Rs < 2915) {
                                                        return 4;
                                                    }
                                                    else if (Rs >= 2915) {
                                                        return 2;
                                                    }
                                                }
                                                else if (Cs >= 1496) {
                                                    return 4;
                                                }
                                            }
                                            else if (negBalance >= 47) {
                                                return 2;
                                            }
                                        }
                                    }
                                }
                            }
                        }
                        else if (tcor >= 74) {
                            if (Rs < 3325) {
                                if (refDCval < 94) {
                                    if (mu_mv_mag_sq < 75) {
                                        if (ssDCval < 71) {
                                            if (TSCindex < 9) {
                                                if (Cs < 2695) {
                                                    if (refDCval < 65) {
                                                        if (negBalance < 27) {
                                                            if (ssDCval < 51) {
                                                                if (Rs < 826) {
                                                                    return 2;
                                                                }
                                                                else if (Rs >= 826) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (ssDCval >= 51) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (negBalance >= 27) {
                                                            if (gchDC < 3) {
                                                                if (diffMVdiffVal < 1) {
                                                                    if (Cs < 1088) {
                                                                        if (Rs < 1181) {
                                                                            if (Rs < 988) {
                                                                                if (diffMVdiffVal < -8) {
                                                                                    if (Cs < 634) {
                                                                                        return 1;
                                                                                    }
                                                                                    else if (Cs >= 634) {
                                                                                        return 8;
                                                                                    }
                                                                                }
                                                                                else if (diffMVdiffVal >= -8) {
                                                                                    return 1;
                                                                                }
                                                                            }
                                                                            else if (Rs >= 988) {
                                                                                if (TSCindex < 5) {
                                                                                    return 1;
                                                                                }
                                                                                else if (TSCindex >= 5) {
                                                                                    if (Cs < 722) {
                                                                                        return 1;
                                                                                    }
                                                                                    else if (Cs >= 722) {
                                                                                        return 4;
                                                                                    }
                                                                                }
                                                                            }
                                                                        }
                                                                        else if (Rs >= 1181) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                    else if (Cs >= 1088) {
                                                                        return 8;
                                                                    }
                                                                }
                                                                else if (diffMVdiffVal >= 1) {
                                                                    if (mu_mv_mag_sq < 32) {
                                                                        if (TSC < 6) {
                                                                            return 8;
                                                                        }
                                                                        else if (TSC >= 6) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                    else if (mu_mv_mag_sq >= 32) {
                                                                        if (diffTSC < 2) {
                                                                            if (posBalance < 55) {
                                                                                if (TSCindex < 6) {
                                                                                    return 4;
                                                                                }
                                                                                else if (TSCindex >= 6) {
                                                                                    if (SC < 2537) {
                                                                                        if (posBalance < 48) {
                                                                                            if (RsCsDiff < 14) {
                                                                                                return 8;
                                                                                            }
                                                                                            else if (RsCsDiff >= 14) {
                                                                                                return 4;
                                                                                            }
                                                                                        }
                                                                                        else if (posBalance >= 48) {
                                                                                            return 8;
                                                                                        }
                                                                                    }
                                                                                    else if (SC >= 2537) {
                                                                                        return 4;
                                                                                    }
                                                                                }
                                                                            }
                                                                            else if (posBalance >= 55) {
                                                                                if (tcor < 78) {
                                                                                    return 4;
                                                                                }
                                                                                else if (tcor >= 78) {
                                                                                    if (AFD < 8) {
                                                                                        if (Rs < 809) {
                                                                                            return 4;
                                                                                        }
                                                                                        else if (Rs >= 809) {
                                                                                            return 1;
                                                                                        }
                                                                                    }
                                                                                    else if (AFD >= 8) {
                                                                                        return 4;
                                                                                    }
                                                                                }
                                                                            }
                                                                        }
                                                                        else if (diffTSC >= 2) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                            else if (gchDC >= 3) {
                                                                if (Cs < 703) {
                                                                    if (Rs < 211) {
                                                                        return 1;
                                                                    }
                                                                    else if (Rs >= 211) {
                                                                        return 8;
                                                                    }
                                                                }
                                                                else if (Cs >= 703) {
                                                                    if (SC < 1796) {
                                                                        return 4;
                                                                    }
                                                                    else if (SC >= 1796) {
                                                                        if (Rs < 981) {
                                                                            return 2;
                                                                        }
                                                                        else if (Rs >= 981) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (refDCval >= 65) {
                                                        if (Cs < 804) {
                                                            if (negBalance < 38) {
                                                                if (SC < 2105) {
                                                                    return 4;
                                                                }
                                                                else if (SC >= 2105) {
                                                                    return 2;
                                                                }
                                                            }
                                                            else if (negBalance >= 38) {
                                                                if (negBalance < 59) {
                                                                    return 2;
                                                                }
                                                                else if (negBalance >= 59) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                        else if (Cs >= 804) {
                                                            if (diffAFD < 1) {
                                                                return 1;
                                                            }
                                                            else if (diffAFD >= 1) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (Cs >= 2695) {
                                                    if (refDCval < 55) {
                                                        if (tcor < 93) {
                                                            if (SC < 6578) {
                                                                if (negBalance < 51) {
                                                                    return 8;
                                                                }
                                                                else if (negBalance >= 51) {
                                                                    if (Rs < 3252) {
                                                                        return 8;
                                                                    }
                                                                    else if (Rs >= 3252) {
                                                                        return 2;
                                                                    }
                                                                }
                                                            }
                                                            else if (SC >= 6578) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (tcor >= 93) {
                                                            if (TSC < 7) {
                                                                return 4;
                                                            }
                                                            else if (TSC >= 7) {
                                                                if (AFD < 9) {
                                                                    return 2;
                                                                }
                                                                else if (AFD >= 9) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (refDCval >= 55) {
                                                        return 2;
                                                    }
                                                }
                                            }
                                            else if (TSCindex >= 9) {
                                                if (CsDiff < 12) {
                                                    if (mcTcor < 102) {
                                                        if (RsCsDiff < 23) {
                                                            return 4;
                                                        }
                                                        else if (RsCsDiff >= 23) {
                                                            if (ssDCval < 69) {
                                                                return 8;
                                                            }
                                                            else if (ssDCval >= 69) {
                                                                if (TSC < 12) {
                                                                    return 4;
                                                                }
                                                                else if (TSC >= 12) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (mcTcor >= 102) {
                                                        if (Cs < 823) {
                                                            return 2;
                                                        }
                                                        else if (Cs >= 823) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                                else if (CsDiff >= 12) {
                                                    if (Rs < 2789) {
                                                        return 1;
                                                    }
                                                    else if (Rs >= 2789) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                        }
                                        else if (ssDCval >= 71) {
                                            if (posBalance < 54) {
                                                if (CsDiff < 9) {
                                                    if (SC < 4234) {
                                                        if (SC < 3622) {
                                                            if (negBalance < 48) {
                                                                if (Rs < 1888) {
                                                                    return 4;
                                                                }
                                                                else if (Rs >= 1888) {
                                                                    return 2;
                                                                }
                                                            }
                                                            else if (negBalance >= 48) {
                                                                if (Rs < 1746) {
                                                                    return 8;
                                                                }
                                                                else if (Rs >= 1746) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                        else if (SC >= 3622) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (SC >= 4234) {
                                                        if (Cs < 2231) {
                                                            if (refDCval < 84) {
                                                                return 8;
                                                            }
                                                            else if (refDCval >= 84) {
                                                                if (diffAFD < -2) {
                                                                    return 8;
                                                                }
                                                                else if (diffAFD >= -2) {
                                                                    return 1;
                                                                }
                                                            }
                                                        }
                                                        else if (Cs >= 2231) {
                                                            if (mcTcor < 90) {
                                                                return 4;
                                                            }
                                                            else if (mcTcor >= 90) {
                                                                if (MVDiff < 40) {
                                                                    return 8;
                                                                }
                                                                else if (MVDiff >= 40) {
                                                                    return 2;
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (CsDiff >= 9) {
                                                    if (refDCval < 80) {
                                                        return 1;
                                                    }
                                                    else if (refDCval >= 80) {
                                                        if (posBalance < 53) {
                                                            if (diffRsCsdiff < 40) {
                                                                if (diffMVdiffVal < 4) {
                                                                    return 1;
                                                                }
                                                                else if (diffMVdiffVal >= 4) {
                                                                    if (SC < 5525) {
                                                                        if (Rs < 2531) {
                                                                            return 8;
                                                                        }
                                                                        else if (Rs >= 2531) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                    else if (SC >= 5525) {
                                                                        return 1;
                                                                    }
                                                                }
                                                            }
                                                            else if (diffRsCsdiff >= 40) {
                                                                if (Cs < 2830) {
                                                                    return 1;
                                                                }
                                                                else if (Cs >= 2830) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                        else if (posBalance >= 53) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (posBalance >= 54) {
                                                if (TSC < 8) {
                                                    if (gchDC < 3) {
                                                        if (Cs < 622) {
                                                            return 1;
                                                        }
                                                        else if (Cs >= 622) {
                                                            if (ssDCval < 73) {
                                                                return 1;
                                                            }
                                                            else if (ssDCval >= 73) {
                                                                if (mcTcor < 95) {
                                                                    return 4;
                                                                }
                                                                else if (mcTcor >= 95) {
                                                                    if (TSC < 7) {
                                                                        return 4;
                                                                    }
                                                                    else if (TSC >= 7) {
                                                                        return 1;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (gchDC >= 3) {
                                                        if (Cs < 416) {
                                                            return 8;
                                                        }
                                                        else if (Cs >= 416) {
                                                            if (diffAFD < -2) {
                                                                return 2;
                                                            }
                                                            else if (diffAFD >= -2) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (TSC >= 8) {
                                                    if (mu_mv_mag_sq < 48) {
                                                        if (SC < 5093) {
                                                            if (tcor < 84) {
                                                                if (ssDCval < 77) {
                                                                    if (CsDiff < 11) {
                                                                        return 8;
                                                                    }
                                                                    else if (CsDiff >= 11) {
                                                                        return 1;
                                                                    }
                                                                }
                                                                else if (ssDCval >= 77) {
                                                                    return 1;
                                                                }
                                                            }
                                                            else if (tcor >= 84) {
                                                                if (AFD < 14) {
                                                                    if (tcor < 89) {
                                                                        return 1;
                                                                    }
                                                                    else if (tcor >= 89) {
                                                                        if (MVDiff < 13) {
                                                                            if (mu_mv_mag_sq < 37) {
                                                                                return 2;
                                                                            }
                                                                            else if (mu_mv_mag_sq >= 37) {
                                                                                if (tcor < 90) {
                                                                                    if (Rs < 2447) {
                                                                                        return 2;
                                                                                    }
                                                                                    else if (Rs >= 2447) {
                                                                                        return 1;
                                                                                    }
                                                                                }
                                                                                else if (tcor >= 90) {
                                                                                    return 1;
                                                                                }
                                                                            }
                                                                        }
                                                                        else if (MVDiff >= 13) {
                                                                            if (SC < 4801) {
                                                                                return 2;
                                                                            }
                                                                            else if (SC >= 4801) {
                                                                                return 1;
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                                else if (AFD >= 14) {
                                                                    if (negBalance < 47) {
                                                                        return 4;
                                                                    }
                                                                    else if (negBalance >= 47) {
                                                                        if (tcor < 91) {
                                                                            return 2;
                                                                        }
                                                                        else if (tcor >= 91) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (SC >= 5093) {
                                                            if (Rs < 2705) {
                                                                return 8;
                                                            }
                                                            else if (Rs >= 2705) {
                                                                if (TSC < 12) {
                                                                    if (Rs < 3048) {
                                                                        return 1;
                                                                    }
                                                                    else if (Rs >= 3048) {
                                                                        return 2;
                                                                    }
                                                                }
                                                                else if (TSC >= 12) {
                                                                    return 1;
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (mu_mv_mag_sq >= 48) {
                                                        if (tcor < 93) {
                                                            if (MVDiff < 8) {
                                                                if (Rs < 2301) {
                                                                    if (Rs < 1849) {
                                                                        return 8;
                                                                    }
                                                                    else if (Rs >= 1849) {
                                                                        return 2;
                                                                    }
                                                                }
                                                                else if (Rs >= 2301) {
                                                                    if (Rs < 3031) {
                                                                        return 4;
                                                                    }
                                                                    else if (Rs >= 3031) {
                                                                        if (diffAFD < 1) {
                                                                            return 8;
                                                                        }
                                                                        else if (diffAFD >= 1) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                            else if (MVDiff >= 8) {
                                                                if (ssDCval < 90) {
                                                                    if (mu_mv_mag_sq < 67) {
                                                                        if (diffRsCsdiff < -34) {
                                                                            if (ssDCval < 85) {
                                                                                return 1;
                                                                            }
                                                                            else if (ssDCval >= 85) {
                                                                                return 2;
                                                                            }
                                                                        }
                                                                        else if (diffRsCsdiff >= -34) {
                                                                            if (tcor < 92) {
                                                                                return 1;
                                                                            }
                                                                            else if (tcor >= 92) {
                                                                                if (Rs < 2515) {
                                                                                    return 2;
                                                                                }
                                                                                else if (Rs >= 2515) {
                                                                                    return 1;
                                                                                }
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (mu_mv_mag_sq >= 67) {
                                                                        return 8;
                                                                    }
                                                                }
                                                                else if (ssDCval >= 90) {
                                                                    if (ssDCval < 94) {
                                                                        if (SC < 5114) {
                                                                            return 1;
                                                                        }
                                                                        else if (SC >= 5114) {
                                                                            if (Rs < 2873) {
                                                                                if (TSC < 11) {
                                                                                    return 1;
                                                                                }
                                                                                else if (TSC >= 11) {
                                                                                    if (posBalance < 60) {
                                                                                        return 4;
                                                                                    }
                                                                                    else if (posBalance >= 60) {
                                                                                        return 1;
                                                                                    }
                                                                                }
                                                                            }
                                                                            else if (Rs >= 2873) {
                                                                                if (posBalance < 55) {
                                                                                    return 4;
                                                                                }
                                                                                else if (posBalance >= 55) {
                                                                                    if (Cs < 2833) {
                                                                                        if (posBalance < 58) {
                                                                                            if (TSC < 11) {
                                                                                                return 1;
                                                                                            }
                                                                                            else if (TSC >= 11) {
                                                                                                if (Rs < 3038) {
                                                                                                    return 4;
                                                                                                }
                                                                                                else if (Rs >= 3038) {
                                                                                                    return 1;
                                                                                                }
                                                                                            }
                                                                                        }
                                                                                        else if (posBalance >= 58) {
                                                                                            return 1;
                                                                                        }
                                                                                    }
                                                                                    else if (Cs >= 2833) {
                                                                                        return 4;
                                                                                    }
                                                                                }
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (ssDCval >= 94) {
                                                                        return 4;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (tcor >= 93) {
                                                            if (MVDiff < 18) {
                                                                return 1;
                                                            }
                                                            else if (MVDiff >= 18) {
                                                                return 2;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (mu_mv_mag_sq >= 75) {
                                        if (mcTcor < 94) {
                                            if (Cs < 2366) {
                                                if (Cs < 1258) {
                                                    if (RsDiff < 5) {
                                                        return 4;
                                                    }
                                                    else if (RsDiff >= 5) {
                                                        if (SC < 2563) {
                                                            return 8;
                                                        }
                                                        else if (SC >= 2563) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                                else if (Cs >= 1258) {
                                                    if (refDCval < 93) {
                                                        return 8;
                                                    }
                                                    else if (refDCval >= 93) {
                                                        return 2;
                                                    }
                                                }
                                            }
                                            else if (Cs >= 2366) {
                                                if (negBalance < 49) {
                                                    if (SC < 4811) {
                                                        return 1;
                                                    }
                                                    else if (SC >= 4811) {
                                                        return 8;
                                                    }
                                                }
                                                else if (negBalance >= 49) {
                                                    if (MVDiff < 6) {
                                                        if (SC < 5486) {
                                                            if (Rs < 2704) {
                                                                return 8;
                                                            }
                                                            else if (Rs >= 2704) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (SC >= 5486) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (MVDiff >= 6) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                        }
                                        else if (mcTcor >= 94) {
                                            if (refDCval < 88) {
                                                if (negBalance < 39) {
                                                    if (Cs < 1735) {
                                                        return 4;
                                                    }
                                                    else if (Cs >= 1735) {
                                                        if (diffAFD < -5) {
                                                            return 1;
                                                        }
                                                        else if (diffAFD >= -5) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                                else if (negBalance >= 39) {
                                                    if (RsCsDiff < 99) {
                                                        if (ssDCval < 87) {
                                                            return 4;
                                                        }
                                                        else if (ssDCval >= 87) {
                                                            return 1;
                                                        }
                                                    }
                                                    else if (RsCsDiff >= 99) {
                                                        if (Rs < 3232) {
                                                            return 8;
                                                        }
                                                        else if (Rs >= 3232) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (refDCval >= 88) {
                                                if (negBalance < 36) {
                                                    if (MVDiff < 5) {
                                                        return 4;
                                                    }
                                                    else if (MVDiff >= 5) {
                                                        if (diffMVdiffVal < 1) {
                                                            return 8;
                                                        }
                                                        else if (diffMVdiffVal >= 1) {
                                                            if (Rs < 2339) {
                                                                return 1;
                                                            }
                                                            else if (Rs >= 2339) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (negBalance >= 36) {
                                                    if (SC < 4174) {
                                                        if (AFD < 10) {
                                                            return 8;
                                                        }
                                                        else if (AFD >= 10) {
                                                            return 2;
                                                        }
                                                    }
                                                    else if (SC >= 4174) {
                                                        return 1;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (refDCval >= 94) {
                                    if (AFD < 11) {
                                        if (tcor < 90) {
                                            if (refDCval < 98) {
                                                if (SC < 3751) {
                                                    if (MVDiff < 10) {
                                                        if (mu_mv_mag_sq < 64) {
                                                            if (diffMVdiffVal < -3) {
                                                                return 8;
                                                            }
                                                            else if (diffMVdiffVal >= -3) {
                                                                return 2;
                                                            }
                                                        }
                                                        else if (mu_mv_mag_sq >= 64) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (MVDiff >= 10) {
                                                        return 8;
                                                    }
                                                }
                                                else if (SC >= 3751) {
                                                    if (Cs < 2626) {
                                                        if (mcTcor < 82) {
                                                            return 8;
                                                        }
                                                        else if (mcTcor >= 82) {
                                                            return 2;
                                                        }
                                                    }
                                                    else if (Cs >= 2626) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                            else if (refDCval >= 98) {
                                                if (posBalance < 53) {
                                                    if (negBalance < 28) {
                                                        if (Rs < 687) {
                                                            return 4;
                                                        }
                                                        else if (Rs >= 687) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (negBalance >= 28) {
                                                        if (MVDiff < 5) {
                                                            return 2;
                                                        }
                                                        else if (MVDiff >= 5) {
                                                            return 1;
                                                        }
                                                    }
                                                }
                                                else if (posBalance >= 53) {
                                                    if (tcor < 84) {
                                                        if (tcor < 80) {
                                                            if (ssDCval < 124) {
                                                                return 4;
                                                            }
                                                            else if (ssDCval >= 124) {
                                                                if (AFD < 10) {
                                                                    return 4;
                                                                }
                                                                else if (AFD >= 10) {
                                                                    return 1;
                                                                }
                                                            }
                                                        }
                                                        else if (tcor >= 80) {
                                                            if (mu_mv_mag_sq < 79) {
                                                                if (Cs < 750) {
                                                                    if (Rs < 336) {
                                                                        return 4;
                                                                    }
                                                                    else if (Rs >= 336) {
                                                                        return 8;
                                                                    }
                                                                }
                                                                else if (Cs >= 750) {
                                                                    if (gchDC < 1) {
                                                                        if (SC < 4523) {
                                                                            return 4;
                                                                        }
                                                                        else if (SC >= 4523) {
                                                                            if (Rs < 2467) {
                                                                                if (ssDCval < 119) {
                                                                                    return 8;
                                                                                }
                                                                                else if (ssDCval >= 119) {
                                                                                    return 4;
                                                                                }
                                                                            }
                                                                            else if (Rs >= 2467) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (gchDC >= 1) {
                                                                        if (Rs < 2741) {
                                                                            return 8;
                                                                        }
                                                                        else if (Rs >= 2741) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                            else if (mu_mv_mag_sq >= 79) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                    else if (tcor >= 84) {
                                                        if (RsCsDiff < 7) {
                                                            return 4;
                                                        }
                                                        else if (RsCsDiff >= 7) {
                                                            if (Rs < 1214) {
                                                                if (Rs < 337) {
                                                                    return 4;
                                                                }
                                                                else if (Rs >= 337) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (Rs >= 1214) {
                                                                if (Cs < 600) {
                                                                    return 2;
                                                                }
                                                                else if (Cs >= 600) {
                                                                    if (mu_mv_mag_sq < 70) {
                                                                        if (Cs < 1407) {
                                                                            if (AFD < 9) {
                                                                                return 4;
                                                                            }
                                                                            else if (AFD >= 9) {
                                                                                return 8;
                                                                            }
                                                                        }
                                                                        else if (Cs >= 1407) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                    else if (mu_mv_mag_sq >= 70) {
                                                                        return 8;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (tcor >= 90) {
                                            if (RsDiff < 6) {
                                                if (mcTcor < 96) {
                                                    return 8;
                                                }
                                                else if (mcTcor >= 96) {
                                                    if (posBalance < 57) {
                                                        if (diffRsCsdiff < -2) {
                                                            return 8;
                                                        }
                                                        else if (diffRsCsdiff >= -2) {
                                                            if (Cs < 2118) {
                                                                return 4;
                                                            }
                                                            else if (Cs >= 2118) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                    else if (posBalance >= 57) {
                                                        if (CsDiff < 4) {
                                                            return 8;
                                                        }
                                                        else if (CsDiff >= 4) {
                                                            return 2;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (RsDiff >= 6) {
                                                if (posBalance < 49) {
                                                    if (refDCval < 102) {
                                                        if (RsDiff < 7) {
                                                            return 8;
                                                        }
                                                        else if (RsDiff >= 7) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (refDCval >= 102) {
                                                        return 4;
                                                    }
                                                }
                                                else if (posBalance >= 49) {
                                                    if (tcor < 92) {
                                                        if (gchDC < 1) {
                                                            return 8;
                                                        }
                                                        else if (gchDC >= 1) {
                                                            if (ssDCval < 117) {
                                                                return 4;
                                                            }
                                                            else if (ssDCval >= 117) {
                                                                if (Cs < 1544) {
                                                                    return 4;
                                                                }
                                                                else if (Cs >= 1544) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (tcor >= 92) {
                                                        if (mu_mv_mag_sq < 110) {
                                                            return 8;
                                                        }
                                                        else if (mu_mv_mag_sq >= 110) {
                                                            if (Rs < 2777) {
                                                                return 4;
                                                            }
                                                            else if (Rs >= 2777) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (AFD >= 11) {
                                        if (MVDiff < 22) {
                                            if (Cs < 838) {
                                                if (SC < 3339) {
                                                    if (mu_mv_mag_sq < 47) {
                                                        return 2;
                                                    }
                                                    else if (mu_mv_mag_sq >= 47) {
                                                        if (RsCsDiff < 112) {
                                                            return 8;
                                                        }
                                                        else if (RsCsDiff >= 112) {
                                                            return 2;
                                                        }
                                                    }
                                                }
                                                else if (SC >= 3339) {
                                                    return 8;
                                                }
                                            }
                                            else if (Cs >= 838) {
                                                if (posBalance < 58) {
                                                    if (posBalance < 50) {
                                                        if (SC < 4666) {
                                                            return 1;
                                                        }
                                                        else if (SC >= 4666) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (posBalance >= 50) {
                                                        if (negBalance < 48) {
                                                            return 1;
                                                        }
                                                        else if (negBalance >= 48) {
                                                            if (RsCsDiff < 187) {
                                                                if (AFD < 15) {
                                                                    if (tcor < 92) {
                                                                        if (ssDCval < 94) {
                                                                            return 1;
                                                                        }
                                                                        else if (ssDCval >= 94) {
                                                                            if (diffAFD < -2) {
                                                                                return 8;
                                                                            }
                                                                            else if (diffAFD >= -2) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (tcor >= 92) {
                                                                        return 1;
                                                                    }
                                                                }
                                                                else if (AFD >= 15) {
                                                                    return 1;
                                                                }
                                                            }
                                                            else if (RsCsDiff >= 187) {
                                                                if (negBalance < 56) {
                                                                    return 8;
                                                                }
                                                                else if (negBalance >= 56) {
                                                                    if (ssDCval < 118) {
                                                                        return 4;
                                                                    }
                                                                    else if (ssDCval >= 118) {
                                                                        return 8;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (posBalance >= 58) {
                                                    if (TSC < 11) {
                                                        if (ssDCval < 117) {
                                                            if (CsDiff < 10) {
                                                                if (negBalance < 45) {
                                                                    if (Rs < 2050) {
                                                                        return 4;
                                                                    }
                                                                    else if (Rs >= 2050) {
                                                                        if (TSC < 9) {
                                                                            return 2;
                                                                        }
                                                                        else if (TSC >= 9) {
                                                                            if (Rs < 2908) {
                                                                                return 1;
                                                                            }
                                                                            else if (Rs >= 2908) {
                                                                                return 8;
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                                else if (negBalance >= 45) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (CsDiff >= 10) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (ssDCval >= 117) {
                                                            if (negBalance < 53) {
                                                                if (SC < 2486) {
                                                                    return 4;
                                                                }
                                                                else if (SC >= 2486) {
                                                                    if (diffRsCsdiff < 72) {
                                                                        return 8;
                                                                    }
                                                                    else if (diffRsCsdiff >= 72) {
                                                                        return 4;
                                                                    }
                                                                }
                                                            }
                                                            else if (negBalance >= 53) {
                                                                if (CsDiff < 8) {
                                                                    if (ssDCval < 119) {
                                                                        return 8;
                                                                    }
                                                                    else if (ssDCval >= 119) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (CsDiff >= 8) {
                                                                    if (Rs < 3018) {
                                                                        return 2;
                                                                    }
                                                                    else if (Rs >= 3018) {
                                                                        return 8;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (TSC >= 11) {
                                                        if (tcor < 86) {
                                                            if (ssDCval < 104) {
                                                                return 4;
                                                            }
                                                            else if (ssDCval >= 104) {
                                                                return 1;
                                                            }
                                                        }
                                                        else if (tcor >= 86) {
                                                            if (posBalance < 61) {
                                                                return 8;
                                                            }
                                                            else if (posBalance >= 61) {
                                                                if (diffRsCsdiff < 21) {
                                                                    return 4;
                                                                }
                                                                else if (diffRsCsdiff >= 21) {
                                                                    return 1;
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (MVDiff >= 22) {
                                            if (SC < 1831) {
                                                if (diffAFD < -2) {
                                                    if (diffAFD < -3) {
                                                        return 1;
                                                    }
                                                    else if (diffAFD >= -3) {
                                                        return 2;
                                                    }
                                                }
                                                else if (diffAFD >= -2) {
                                                    if (MVDiff < 39) {
                                                        return 4;
                                                    }
                                                    else if (MVDiff >= 39) {
                                                        if (Rs < 1106) {
                                                            if (diffMVdiffVal < -192) {
                                                                return 2;
                                                            }
                                                            else if (diffMVdiffVal >= -192) {
                                                                return 1;
                                                            }
                                                        }
                                                        else if (Rs >= 1106) {
                                                            return 2;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (SC >= 1831) {
                                                if (diffMVdiffVal < -7) {
                                                    if (SC < 4243) {
                                                        if (Rs < 1196) {
                                                            return 8;
                                                        }
                                                        else if (Rs >= 1196) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (SC >= 4243) {
                                                        if (RsCsDiff < 185) {
                                                            if (Rs < 2641) {
                                                                return 1;
                                                            }
                                                            else if (Rs >= 2641) {
                                                                return 2;
                                                            }
                                                        }
                                                        else if (RsCsDiff >= 185) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                                else if (diffMVdiffVal >= -7) {
                                                    if (mcTcor < 88) {
                                                        if (Rs < 2469) {
                                                            return 4;
                                                        }
                                                        else if (Rs >= 2469) {
                                                            if (Cs < 2376) {
                                                                return 8;
                                                            }
                                                            else if (Cs >= 2376) {
                                                                return 1;
                                                            }
                                                        }
                                                    }
                                                    else if (mcTcor >= 88) {
                                                        if (diffAFD < -5) {
                                                            return 4;
                                                        }
                                                        else if (diffAFD >= -5) {
                                                            if (TSC < 13) {
                                                                if (diffMVdiffVal < 12) {
                                                                    return 2;
                                                                }
                                                                else if (diffMVdiffVal >= 12) {
                                                                    if (TSC < 9) {
                                                                        return 2;
                                                                    }
                                                                    else if (TSC >= 9) {
                                                                        return 4;
                                                                    }
                                                                }
                                                            }
                                                            else if (TSC >= 13) {
                                                                if (AFD < 18) {
                                                                    return 8;
                                                                }
                                                                else if (AFD >= 18) {
                                                                    return 1;
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                            else if (Rs >= 3325) {
                                if (TSC < 10) {
                                    if (TSC < 7) {
                                        if (Cs < 2072) {
                                            if (posBalance < 52) {
                                                if (gchDC < 3) {
                                                    return 4;
                                                }
                                                else if (gchDC >= 3) {
                                                    return 8;
                                                }
                                            }
                                            else if (posBalance >= 52) {
                                                if (ssDCval < 108) {
                                                    if (CsDiff < 5) {
                                                        if (MVDiff < 12) {
                                                            return 4;
                                                        }
                                                        else if (MVDiff >= 12) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (CsDiff >= 5) {
                                                        return 8;
                                                    }
                                                }
                                                else if (ssDCval >= 108) {
                                                    return 4;
                                                }
                                            }
                                        }
                                        else if (Cs >= 2072) {
                                            if (SC < 6199) {
                                                return 2;
                                            }
                                            else if (SC >= 6199) {
                                                if (tcor < 91) {
                                                    if (ssDCval < 54) {
                                                        if (Rs < 3450) {
                                                            return 8;
                                                        }
                                                        else if (Rs >= 3450) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (ssDCval >= 54) {
                                                        return 8;
                                                    }
                                                }
                                                else if (tcor >= 91) {
                                                    if (mu_mv_mag_sq < 36) {
                                                        return 8;
                                                    }
                                                    else if (mu_mv_mag_sq >= 36) {
                                                        if (MVDiff < 9) {
                                                            if (CsDiff < 6) {
                                                                if (mu_mv_mag_sq < 69) {
                                                                    return 4;
                                                                }
                                                                else if (mu_mv_mag_sq >= 69) {
                                                                    if (negBalance < 32) {
                                                                        return 4;
                                                                    }
                                                                    else if (negBalance >= 32) {
                                                                        return 8;
                                                                    }
                                                                }
                                                            }
                                                            else if (CsDiff >= 6) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (MVDiff >= 9) {
                                                            if (refDCval < 54) {
                                                                if (diffMVdiffVal < 5) {
                                                                    return 4;
                                                                }
                                                                else if (diffMVdiffVal >= 5) {
                                                                    if (Rs < 3598) {
                                                                        return 4;
                                                                    }
                                                                    else if (Rs >= 3598) {
                                                                        return 2;
                                                                    }
                                                                }
                                                            }
                                                            else if (refDCval >= 54) {
                                                                if (Rs < 4084) {
                                                                    if (TSC < 6) {
                                                                        if (AFD < 8) {
                                                                            return 4;
                                                                        }
                                                                        else if (AFD >= 8) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                    else if (TSC >= 6) {
                                                                        if (Cs < 3374) {
                                                                            return 2;
                                                                        }
                                                                        else if (Cs >= 3374) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                }
                                                                else if (Rs >= 4084) {
                                                                    if (SC < 7200) {
                                                                        if (ssDCval < 118) {
                                                                            return 2;
                                                                        }
                                                                        else if (ssDCval >= 118) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                    else if (SC >= 7200) {
                                                                        if (Cs < 2474) {
                                                                            return 8;
                                                                        }
                                                                        else if (Cs >= 2474) {
                                                                            return 2;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (TSC >= 7) {
                                        if (mcTcor < 88) {
                                            if (ssDCval < 107) {
                                                if (ssDCval < 104) {
                                                    if (Rs < 3677) {
                                                        return 8;
                                                    }
                                                    else if (Rs >= 3677) {
                                                        return 4;
                                                    }
                                                }
                                                else if (ssDCval >= 104) {
                                                    if (Cs < 2925) {
                                                        return 8;
                                                    }
                                                    else if (Cs >= 2925) {
                                                        if (Rs < 4083) {
                                                            return 8;
                                                        }
                                                        else if (Rs >= 4083) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (ssDCval >= 107) {
                                                if (Rs < 4475) {
                                                    if (AFD < 12) {
                                                        return 1;
                                                    }
                                                    else if (AFD >= 12) {
                                                        return 8;
                                                    }
                                                }
                                                else if (Rs >= 4475) {
                                                    return 2;
                                                }
                                            }
                                        }
                                        else if (mcTcor >= 88) {
                                            if (refDCval < 52) {
                                                if (AFD < 12) {
                                                    if (gchDC < 2) {
                                                        return 2;
                                                    }
                                                    else if (gchDC >= 2) {
                                                        return 8;
                                                    }
                                                }
                                                else if (AFD >= 12) {
                                                    if (Rs < 3991) {
                                                        return 8;
                                                    }
                                                    else if (Rs >= 3991) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                            else if (refDCval >= 52) {
                                                if (mcTcor < 92) {
                                                    if (negBalance < 43) {
                                                        if (RsCsDiff < 92) {
                                                            return 8;
                                                        }
                                                        else if (RsCsDiff >= 92) {
                                                            if (AFD < 11) {
                                                                return 4;
                                                            }
                                                            else if (AFD >= 11) {
                                                                if (CsDiff < 8) {
                                                                    return 8;
                                                                }
                                                                else if (CsDiff >= 8) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (negBalance >= 43) {
                                                        if (CsDiff < 5) {
                                                            if (TSC < 8) {
                                                                return 4;
                                                            }
                                                            else if (TSC >= 8) {
                                                                return 2;
                                                            }
                                                        }
                                                        else if (CsDiff >= 5) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                                else if (mcTcor >= 92) {
                                                    if (mu_mv_mag_sq < 95) {
                                                        if (AFD < 10) {
                                                            if (MVDiff < 30) {
                                                                if (Cs < 2194) {
                                                                    if (negBalance < 34) {
                                                                        if (SC < 5843) {
                                                                            if (RsDiff < 10) {
                                                                                return 2;
                                                                            }
                                                                            else if (RsDiff >= 10) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                        else if (SC >= 5843) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                    else if (negBalance >= 34) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (Cs >= 2194) {
                                                                    if (mu_mv_mag_sq < 85) {
                                                                        if (Cs < 3710) {
                                                                            if (MVDiff < 11) {
                                                                                if (mu_mv_mag_sq < 78) {
                                                                                    return 8;
                                                                                }
                                                                                else if (mu_mv_mag_sq >= 78) {
                                                                                    if (gchDC < 1) {
                                                                                        return 2;
                                                                                    }
                                                                                    else if (gchDC >= 1) {
                                                                                        if (Cs < 2688) {
                                                                                            if (SC < 6502) {
                                                                                                return 2;
                                                                                            }
                                                                                            else if (SC >= 6502) {
                                                                                                return 8;
                                                                                            }
                                                                                        }
                                                                                        else if (Cs >= 2688) {
                                                                                            return 2;
                                                                                        }
                                                                                    }
                                                                                }
                                                                            }
                                                                            else if (MVDiff >= 11) {
                                                                                if (posBalance < 59) {
                                                                                    if (Rs < 3938) {
                                                                                        return 2;
                                                                                    }
                                                                                    else if (Rs >= 3938) {
                                                                                        if (Cs < 2297) {
                                                                                            if (diffMVdiffVal < 1) {
                                                                                                return 2;
                                                                                            }
                                                                                            else if (diffMVdiffVal >= 1) {
                                                                                                if (SC < 6299) {
                                                                                                    if (Cs < 2253) {
                                                                                                        return 2;
                                                                                                    }
                                                                                                    else if (Cs >= 2253) {
                                                                                                        return 4;
                                                                                                    }
                                                                                                }
                                                                                                else if (SC >= 6299) {
                                                                                                    return 4;
                                                                                                }
                                                                                            }
                                                                                        }
                                                                                        else if (Cs >= 2297) {
                                                                                            if (Rs < 4848) {
                                                                                                return 2;
                                                                                            }
                                                                                            else if (Rs >= 4848) {
                                                                                                return 4;
                                                                                            }
                                                                                        }
                                                                                    }
                                                                                }
                                                                                else if (posBalance >= 59) {
                                                                                    if (Rs < 3464) {
                                                                                        return 8;
                                                                                    }
                                                                                    else if (Rs >= 3464) {
                                                                                        return 2;
                                                                                    }
                                                                                }
                                                                            }
                                                                        }
                                                                        else if (Cs >= 3710) {
                                                                            if (MVDiff < 14) {
                                                                                if (Rs < 3773) {
                                                                                    return 2;
                                                                                }
                                                                                else if (Rs >= 3773) {
                                                                                    return 4;
                                                                                }
                                                                            }
                                                                            else if (MVDiff >= 14) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (mu_mv_mag_sq >= 85) {
                                                                        if (ssDCval < 127) {
                                                                            return 4;
                                                                        }
                                                                        else if (ssDCval >= 127) {
                                                                            return 2;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                            else if (MVDiff >= 30) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (AFD >= 10) {
                                                            if (negBalance < 53) {
                                                                if (Rs < 3567) {
                                                                    if (ssDCval < 108) {
                                                                        if (diffRsCsdiff < -3) {
                                                                            return 2;
                                                                        }
                                                                        else if (diffRsCsdiff >= -3) {
                                                                            if (AFD < 12) {
                                                                                if (mu_mv_mag_sq < 37) {
                                                                                    if (posBalance < 59) {
                                                                                        return 2;
                                                                                    }
                                                                                    else if (posBalance >= 59) {
                                                                                        return 4;
                                                                                    }
                                                                                }
                                                                                else if (mu_mv_mag_sq >= 37) {
                                                                                    return 4;
                                                                                }
                                                                            }
                                                                            else if (AFD >= 12) {
                                                                                return 2;
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (ssDCval >= 108) {
                                                                        if (RsCsDiff < 215) {
                                                                            return 8;
                                                                        }
                                                                        else if (RsCsDiff >= 215) {
                                                                            return 2;
                                                                        }
                                                                    }
                                                                }
                                                                else if (Rs >= 3567) {
                                                                    if (SC < 7395) {
                                                                        if (mu_mv_mag_sq < 71) {
                                                                            if (MVDiff < 24) {
                                                                                if (diffRsCsdiff < 30) {
                                                                                    if (Cs < 3532) {
                                                                                        return 4;
                                                                                    }
                                                                                    else if (Cs >= 3532) {
                                                                                        if (Cs < 3543) {
                                                                                            return 2;
                                                                                        }
                                                                                        else if (Cs >= 3543) {
                                                                                            return 4;
                                                                                        }
                                                                                    }
                                                                                }
                                                                                else if (diffRsCsdiff >= 30) {
                                                                                    if (ssDCval < 128) {
                                                                                        return 4;
                                                                                    }
                                                                                    else if (ssDCval >= 128) {
                                                                                        return 2;
                                                                                    }
                                                                                }
                                                                            }
                                                                            else if (MVDiff >= 24) {
                                                                                if (negBalance < 45) {
                                                                                    return 4;
                                                                                }
                                                                                else if (negBalance >= 45) {
                                                                                    if (diffRsCsdiff < -13) {
                                                                                        return 4;
                                                                                    }
                                                                                    else if (diffRsCsdiff >= -13) {
                                                                                        return 8;
                                                                                    }
                                                                                }
                                                                            }
                                                                        }
                                                                        else if (mu_mv_mag_sq >= 71) {
                                                                            if (TSC < 9) {
                                                                                return 8;
                                                                            }
                                                                            else if (TSC >= 9) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (SC >= 7395) {
                                                                        if (RsDiff < 5) {
                                                                            if (ssDCval < 61) {
                                                                                return 2;
                                                                            }
                                                                            else if (ssDCval >= 61) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                        else if (RsDiff >= 5) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                            else if (negBalance >= 53) {
                                                                if (gchDC < 2) {
                                                                    return 2;
                                                                }
                                                                else if (gchDC >= 2) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (mu_mv_mag_sq >= 95) {
                                                        if (TSCindex < 9) {
                                                            if (CsDiff < 5) {
                                                                return 8;
                                                            }
                                                            else if (CsDiff >= 5) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (TSCindex >= 9) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (TSC >= 10) {
                                    if (tcor < 86) {
                                        if (mu_mv_mag_sq < 79) {
                                            if (CsDiff < 14) {
                                                if (negBalance < 53) {
                                                    if (mcTcor < 85) {
                                                        if (refDCval < 123) {
                                                            if (posBalance < 45) {
                                                                return 1;
                                                            }
                                                            else if (posBalance >= 45) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (refDCval >= 123) {
                                                            return 1;
                                                        }
                                                    }
                                                    else if (mcTcor >= 85) {
                                                        if (Cs < 2510) {
                                                            return 4;
                                                        }
                                                        else if (Cs >= 2510) {
                                                            if (Rs < 3692) {
                                                                return 1;
                                                            }
                                                            else if (Rs >= 3692) {
                                                                return 2;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (negBalance >= 53) {
                                                    if (AFD < 17) {
                                                        if (SC < 5124) {
                                                            return 2;
                                                        }
                                                        else if (SC >= 5124) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (AFD >= 17) {
                                                        return 1;
                                                    }
                                                }
                                            }
                                            else if (CsDiff >= 14) {
                                                return 8;
                                            }
                                        }
                                        else if (mu_mv_mag_sq >= 79) {
                                            if (negBalance < 48) {
                                                if (mu_mv_mag_sq < 86) {
                                                    return 4;
                                                }
                                                else if (mu_mv_mag_sq >= 86) {
                                                    return 8;
                                                }
                                            }
                                            else if (negBalance >= 48) {
                                                return 4;
                                            }
                                        }
                                    }
                                    else if (tcor >= 86) {
                                        if (refDCval < 113) {
                                            if (SC < 6885) {
                                                if (Cs < 2908) {
                                                    if (Rs < 3775) {
                                                        if (Cs < 2230) {
                                                            return 8;
                                                        }
                                                        else if (Cs >= 2230) {
                                                            return 1;
                                                        }
                                                    }
                                                    else if (Rs >= 3775) {
                                                        return 4;
                                                    }
                                                }
                                                else if (Cs >= 2908) {
                                                    return 8;
                                                }
                                            }
                                            else if (SC >= 6885) {
                                                if (Rs < 4401) {
                                                    return 4;
                                                }
                                                else if (Rs >= 4401) {
                                                    return 2;
                                                }
                                            }
                                        }
                                        else if (refDCval >= 113) {
                                            return 8;
                                        }
                                    }
                                }
                            }
                        }
                    }
                }
            }
            else if (SC >= 7760) {
                if (CsDiff < 9) {
                    if (TSC < 9) {
                        if (RsCsDiff < 94) {
                            if (mu_mv_mag_sq < 43) {
                                if (TSCindex < 8) {
                                    if (Cs < 4010) {
                                        return 8;
                                    }
                                    else if (Cs >= 4010) {
                                        return 2;
                                    }
                                }
                                else if (TSCindex >= 8) {
                                    if (Rs < 4022) {
                                        return 2;
                                    }
                                    else if (Rs >= 4022) {
                                        return 4;
                                    }
                                }
                            }
                            else if (mu_mv_mag_sq >= 43) {
                                if (negBalance < 39) {
                                    if (Cs < 2985) {
                                        return 2;
                                    }
                                    else if (Cs >= 2985) {
                                        return 8;
                                    }
                                }
                                else if (negBalance >= 39) {
                                    if (TSCindex < 8) {
                                        if (AFD < 10) {
                                            if (Cs < 3924) {
                                                return 8;
                                            }
                                            else if (Cs >= 3924) {
                                                return 4;
                                            }
                                        }
                                        else if (AFD >= 10) {
                                            return 4;
                                        }
                                    }
                                    else if (TSCindex >= 8) {
                                        if (Cs < 4468) {
                                            return 4;
                                        }
                                        else if (Cs >= 4468) {
                                            return 8;
                                        }
                                    }
                                }
                            }
                        }
                        else if (RsCsDiff >= 94) {
                            if (tcor < 92) {
                                if (diffAFD < -2) {
                                    return 8;
                                }
                                else if (diffAFD >= -2) {
                                    if (mcTcor < 91) {
                                        if (AFD < 10) {
                                            return 8;
                                        }
                                        else if (AFD >= 10) {
                                            return 4;
                                        }
                                    }
                                    else if (mcTcor >= 91) {
                                        if (diffRsCsdiff < -9) {
                                            if (SC < 9971) {
                                                return 8;
                                            }
                                            else if (SC >= 9971) {
                                                return 4;
                                            }
                                        }
                                        else if (diffRsCsdiff >= -9) {
                                            return 4;
                                        }
                                    }
                                }
                            }
                            else if (tcor >= 92) {
                                if (CsDiff < 7) {
                                    return 4;
                                }
                                else if (CsDiff >= 7) {
                                    if (mcTcor < 94) {
                                        if (diffRsCsdiff < -9) {
                                            return 8;
                                        }
                                        else if (diffRsCsdiff >= -9) {
                                            if (AFD < 8) {
                                                return 4;
                                            }
                                            else if (AFD >= 8) {
                                                return 2;
                                            }
                                        }
                                    }
                                    else if (mcTcor >= 94) {
                                        return 8;
                                    }
                                }
                            }
                        }
                    }
                    else if (TSC >= 9) {
                        if (negBalance < 50) {
                            if (mu_mv_mag_sq < 73) {
                                if (Cs < 4160) {
                                    if (ssDCval < 106) {
                                        return 4;
                                    }
                                    else if (ssDCval >= 106) {
                                        if (SC < 8015) {
                                            return 2;
                                        }
                                        else if (SC >= 8015) {
                                            if (AFD < 14) {
                                                return 4;
                                            }
                                            else if (AFD >= 14) {
                                                return 2;
                                            }
                                        }
                                    }
                                }
                                else if (Cs >= 4160) {
                                    if (refDCval < 66) {
                                        if (SC < 8947) {
                                            return 8;
                                        }
                                        else if (SC >= 8947) {
                                            return 4;
                                        }
                                    }
                                    else if (refDCval >= 66) {
                                        return 8;
                                    }
                                }
                            }
                            else if (mu_mv_mag_sq >= 73) {
                                if (mcTcor < 86) {
                                    return 2;
                                }
                                else if (mcTcor >= 86) {
                                    if (Rs < 5104) {
                                        return 2;
                                    }
                                    else if (Rs >= 5104) {
                                        if (SC < 8515) {
                                            return 8;
                                        }
                                        else if (SC >= 8515) {
                                            return 2;
                                        }
                                    }
                                }
                            }
                        }
                        else if (negBalance >= 50) {
                            if (AFD < 17) {
                                return 4;
                            }
                            else if (AFD >= 17) {
                                return 8;
                            }
                        }
                    }
                }
                else if (CsDiff >= 9) {
                    if (mu_mv_mag_sq < 61) {
                        if (TSC < 17) {
                            if (ssDCval < 153) {
                                if (Cs < 3333) {
                                    if (SC < 9179) {
                                        if (Cs < 2574) {
                                            return 1;
                                        }
                                        else if (Cs >= 2574) {
                                            if (Rs < 4747) {
                                                if (Rs < 4632) {
                                                    if (diffAFD < 4) {
                                                        if (SC < 7780) {
                                                            return 2;
                                                        }
                                                        else if (SC >= 7780) {
                                                            return 1;
                                                        }
                                                    }
                                                    else if (diffAFD >= 4) {
                                                        return 4;
                                                    }
                                                }
                                                else if (Rs >= 4632) {
                                                    return 8;
                                                }
                                            }
                                            else if (Rs >= 4747) {
                                                if (TSC < 12) {
                                                    return 4;
                                                }
                                                else if (TSC >= 12) {
                                                    if (ssDCval < 106) {
                                                        return 4;
                                                    }
                                                    else if (ssDCval >= 106) {
                                                        return 2;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (SC >= 9179) {
                                        return 1;
                                    }
                                }
                                else if (Cs >= 3333) {
                                    if (Rs < 6989) {
                                        if (CsDiff < 18) {
                                            if (AFD < 17) {
                                                if (diffRsCsdiff < 91) {
                                                    if (Cs < 5197) {
                                                        if (Rs < 3968) {
                                                            return 8;
                                                        }
                                                        else if (Rs >= 3968) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (Cs >= 5197) {
                                                        return 8;
                                                    }
                                                }
                                                else if (diffRsCsdiff >= 91) {
                                                    return 1;
                                                }
                                            }
                                            else if (AFD >= 17) {
                                                if (Cs < 3987) {
                                                    if (AFD < 20) {
                                                        if (mcTcor < 84) {
                                                            if (posBalance < 54) {
                                                                return 8;
                                                            }
                                                            else if (posBalance >= 54) {
                                                                if (Rs < 5714) {
                                                                    return 4;
                                                                }
                                                                else if (Rs >= 5714) {
                                                                    return 1;
                                                                }
                                                            }
                                                        }
                                                        else if (mcTcor >= 84) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (AFD >= 20) {
                                                        if (gchDC < 3) {
                                                            if (tcor < 69) {
                                                                if (SC < 8510) {
                                                                    return 8;
                                                                }
                                                                else if (SC >= 8510) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (tcor >= 69) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (gchDC >= 3) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                                else if (Cs >= 3987) {
                                                    if (posBalance < 57) {
                                                        if (Rs < 4895) {
                                                            return 4;
                                                        }
                                                        else if (Rs >= 4895) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (posBalance >= 57) {
                                                        if (Cs < 5186) {
                                                            if (tcor < 66) {
                                                                if (Cs < 4298) {
                                                                    return 1;
                                                                }
                                                                else if (Cs >= 4298) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (tcor >= 66) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (Cs >= 5186) {
                                                            if (Rs < 5994) {
                                                                return 8;
                                                            }
                                                            else if (Rs >= 5994) {
                                                                return 1;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (CsDiff >= 18) {
                                            if (mcTcor < 91) {
                                                if (RsCsDiff < 597) {
                                                    return 1;
                                                }
                                                else if (RsCsDiff >= 597) {
                                                    if (AFD < 33) {
                                                        return 8;
                                                    }
                                                    else if (AFD >= 33) {
                                                        return 2;
                                                    }
                                                }
                                            }
                                            else if (mcTcor >= 91) {
                                                if (mu_mv_mag_sq < 35) {
                                                    return 1;
                                                }
                                                else if (mu_mv_mag_sq >= 35) {
                                                    if (Rs < 6567) {
                                                        return 8;
                                                    }
                                                    else if (Rs >= 6567) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (Rs >= 6989) {
                                        if (MVDiff < 7) {
                                            return 8;
                                        }
                                        else if (MVDiff >= 7) {
                                            return 4;
                                        }
                                    }
                                }
                            }
                            else if (ssDCval >= 153) {
                                if (negBalance < 56) {
                                    if (mu_mv_mag_sq < 3) {
                                        if (posBalance < 61) {
                                            return 8;
                                        }
                                        else if (posBalance >= 61) {
                                            return 2;
                                        }
                                    }
                                    else if (mu_mv_mag_sq >= 3) {
                                        if (Cs < 4064) {
                                            return 4;
                                        }
                                        else if (Cs >= 4064) {
                                            return 8;
                                        }
                                    }
                                }
                                else if (negBalance >= 56) {
                                    if (Rs < 5068) {
                                        return 4;
                                    }
                                    else if (Rs >= 5068) {
                                        if (MVDiff < 27) {
                                            return 2;
                                        }
                                        else if (MVDiff >= 27) {
                                            return 4;
                                        }
                                    }
                                }
                            }
                        }
                        else if (TSC >= 17) {
                            if (Cs < 6358) {
                                if (SC < 12053) {
                                    if (mu_mv_mag_sq < 2) {
                                        if (RsCsDiff < 1181) {
                                            if (TSC < 18) {
                                                if (Cs < 3918) {
                                                    if (RsCsDiff < 494) {
                                                        return 2;
                                                    }
                                                    else if (RsCsDiff >= 494) {
                                                        return 4;
                                                    }
                                                }
                                                else if (Cs >= 3918) {
                                                    return 2;
                                                }
                                            }
                                            else if (TSC >= 18) {
                                                return 8;
                                            }
                                        }
                                        else if (RsCsDiff >= 1181) {
                                            if (RsCsDiff < 1502) {
                                                return 8;
                                            }
                                            else if (RsCsDiff >= 1502) {
                                                return 4;
                                            }
                                        }
                                    }
                                    else if (mu_mv_mag_sq >= 2) {
                                        if (Cs < 4323) {
                                            if (Rs < 5186) {
                                                if (Rs < 4162) {
                                                    return 1;
                                                }
                                                else if (Rs >= 4162) {
                                                    if (ssDCval < 113) {
                                                        if (Rs < 4589) {
                                                            return 2;
                                                        }
                                                        else if (Rs >= 4589) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (ssDCval >= 113) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                            else if (Rs >= 5186) {
                                                if (MVDiff < 13) {
                                                    if (Rs < 5890) {
                                                        return 8;
                                                    }
                                                    else if (Rs >= 5890) {
                                                        return 4;
                                                    }
                                                }
                                                else if (MVDiff >= 13) {
                                                    return 4;
                                                }
                                            }
                                        }
                                        else if (Cs >= 4323) {
                                            if (Rs < 6399) {
                                                if (negBalance < 52) {
                                                    if (diffRsCsdiff < 215) {
                                                        return 8;
                                                    }
                                                    else if (diffRsCsdiff >= 215) {
                                                        if (AFD < 30) {
                                                            return 4;
                                                        }
                                                        else if (AFD >= 30) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                                else if (negBalance >= 52) {
                                                    return 8;
                                                }
                                            }
                                            else if (Rs >= 6399) {
                                                if (Rs < 6471) {
                                                    return 2;
                                                }
                                                else if (Rs >= 6471) {
                                                    return 8;
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (SC >= 12053) {
                                    if (mu_mv_mag_sq < 12) {
                                        if (mcTcor < 91) {
                                            if (SC < 12354) {
                                                return 2;
                                            }
                                            else if (SC >= 12354) {
                                                if (Cs < 4439) {
                                                    return 4;
                                                }
                                                else if (Cs >= 4439) {
                                                    if (ssDCval < 95) {
                                                        return 4;
                                                    }
                                                    else if (ssDCval >= 95) {
                                                        if (SC < 12529) {
                                                            if (SC < 12502) {
                                                                return 8;
                                                            }
                                                            else if (SC >= 12502) {
                                                                return 2;
                                                            }
                                                        }
                                                        else if (SC >= 12529) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (mcTcor >= 91) {
                                            return 4;
                                        }
                                    }
                                    else if (mu_mv_mag_sq >= 12) {
                                        if (mcTcor < 87) {
                                            return 4;
                                        }
                                        else if (mcTcor >= 87) {
                                            if (refDCval < 136) {
                                                if (MVDiff < 13) {
                                                    return 8;
                                                }
                                                else if (MVDiff >= 13) {
                                                    return 2;
                                                }
                                            }
                                            else if (refDCval >= 136) {
                                                if (MVDiff < 30) {
                                                    return 4;
                                                }
                                                else if (MVDiff >= 30) {
                                                    return 2;
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                            else if (Cs >= 6358) {
                                if (Rs < 15116) {
                                    if (CsDiff < 31) {
                                        if (refDCval < 91) {
                                            if (RsDiff < 29) {
                                                return 4;
                                            }
                                            else if (RsDiff >= 29) {
                                                return 2;
                                            }
                                        }
                                        else if (refDCval >= 91) {
                                            if (RsDiff < 34) {
                                                if (gchDC < 4) {
                                                    return 8;
                                                }
                                                else if (gchDC >= 4) {
                                                    return 4;
                                                }
                                            }
                                            else if (RsDiff >= 34) {
                                                if (MVDiff < 10) {
                                                    return 8;
                                                }
                                                else if (MVDiff >= 10) {
                                                    if (MVDiff < 25) {
                                                        return 4;
                                                    }
                                                    else if (MVDiff >= 25) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (CsDiff >= 31) {
                                        if (negBalance < 54) {
                                            return 8;
                                        }
                                        else if (negBalance >= 54) {
                                            if (ssDCval < 129) {
                                                return 4;
                                            }
                                            else if (ssDCval >= 129) {
                                                if (gchDC < 2) {
                                                    if (refDCval < 147) {
                                                        return 4;
                                                    }
                                                    else if (refDCval >= 147) {
                                                        return 2;
                                                    }
                                                }
                                                else if (gchDC >= 2) {
                                                    return 8;
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (Rs >= 15116) {
                                    return 8;
                                }
                            }
                        }
                    }
                    else if (mu_mv_mag_sq >= 61) {
                        if (Rs < 7236) {
                            if (refDCval < 105) {
                                if (AFD < 8) {
                                    if (ssDCval < 83) {
                                        if (Cs < 5169) {
                                            if (posBalance < 58) {
                                                if (Cs < 4914) {
                                                    return 8;
                                                }
                                                else if (Cs >= 4914) {
                                                    return 4;
                                                }
                                            }
                                            else if (posBalance >= 58) {
                                                return 8;
                                            }
                                        }
                                        else if (Cs >= 5169) {
                                            return 4;
                                        }
                                    }
                                    else if (ssDCval >= 83) {
                                        return 8;
                                    }
                                }
                                else if (AFD >= 8) {
                                    if (Cs < 3354) {
                                        return 4;
                                    }
                                    else if (Cs >= 3354) {
                                        if (Cs < 7200) {
                                            return 8;
                                        }
                                        else if (Cs >= 7200) {
                                            return 4;
                                        }
                                    }
                                }
                            }
                            else if (refDCval >= 105) {
                                if (ssDCval < 121) {
                                    if (Cs < 3943) {
                                        if (negBalance < 47) {
                                            if (ssDCval < 115) {
                                                if (Rs < 5378) {
                                                    return 2;
                                                }
                                                else if (Rs >= 5378) {
                                                    return 4;
                                                }
                                            }
                                            else if (ssDCval >= 115) {
                                                return 4;
                                            }
                                        }
                                        else if (negBalance >= 47) {
                                            return 4;
                                        }
                                    }
                                    else if (Cs >= 3943) {
                                        if (mu_mv_mag_sq < 122) {
                                            if (Cs < 4118) {
                                                if (ssDCval < 117) {
                                                    return 8;
                                                }
                                                else if (ssDCval >= 117) {
                                                    return 4;
                                                }
                                            }
                                            else if (Cs >= 4118) {
                                                return 8;
                                            }
                                        }
                                        else if (mu_mv_mag_sq >= 122) {
                                            if (negBalance < 48) {
                                                if (diffTSC < -1) {
                                                    if (posBalance < 56) {
                                                        return 8;
                                                    }
                                                    else if (posBalance >= 56) {
                                                        return 4;
                                                    }
                                                }
                                                else if (diffTSC >= -1) {
                                                    return 8;
                                                }
                                            }
                                            else if (negBalance >= 48) {
                                                if (CsDiff < 14) {
                                                    if (refDCval < 106) {
                                                        return 8;
                                                    }
                                                    else if (refDCval >= 106) {
                                                        if (refDCval < 107) {
                                                            if (Rs < 5796) {
                                                                return 8;
                                                            }
                                                            else if (Rs >= 5796) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (refDCval >= 107) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                                else if (CsDiff >= 14) {
                                                    if (diffRsCsdiff < 41) {
                                                        return 8;
                                                    }
                                                    else if (diffRsCsdiff >= 41) {
                                                        if (mu_mv_mag_sq < 127) {
                                                            if (SC < 11082) {
                                                                return 4;
                                                            }
                                                            else if (SC >= 11082) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (mu_mv_mag_sq >= 127) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (ssDCval >= 121) {
                                    if (mcTcor < 85) {
                                        if (Rs < 4923) {
                                            return 1;
                                        }
                                        else if (Rs >= 4923) {
                                            if (AFD < 13) {
                                                return 1;
                                            }
                                            else if (AFD >= 13) {
                                                return 8;
                                            }
                                        }
                                    }
                                    else if (mcTcor >= 85) {
                                        if (Cs < 3098) {
                                            return 2;
                                        }
                                        else if (Cs >= 3098) {
                                            return 8;
                                        }
                                    }
                                }
                            }
                        }
                        else if (Rs >= 7236) {
                            if (negBalance < 61) {
                                if (Cs < 6375) {
                                    return 4;
                                }
                                else if (Cs >= 6375) {
                                    if (Cs < 12169) {
                                        return 8;
                                    }
                                    else if (Cs >= 12169) {
                                        return 4;
                                    }
                                }
                            }
                            else if (negBalance >= 61) {
                                if (mu_mv_mag_sq < 120) {
                                    if (negBalance < 67) {
                                        return 4;
                                    }
                                    else if (negBalance >= 67) {
                                        if (Cs < 7581) {
                                            return 4;
                                        }
                                        else if (Cs >= 7581) {
                                            if (AFD < 20) {
                                                return 4;
                                            }
                                            else if (AFD >= 20) {
                                                return 2;
                                            }
                                        }
                                    }
                                }
                                else if (mu_mv_mag_sq >= 120) {
                                    if (negBalance < 66) {
                                        return 2;
                                    }
                                    else if (negBalance >= 66) {
                                        if (tcor < 77) {
                                            if (Cs < 8194) {
                                                return 4;
                                            }
                                            else if (Cs >= 8194) {
                                                return 2;
                                            }
                                        }
                                        else if (tcor >= 77) {
                                            return 2;
                                        }
                                    }
                                }
                            }
                        }
                    }
                }
            }
        }
        return 4;
    }
    mfxU16 AGOPSelect4(mfxU32 Rs, mfxU32 CsDiff, mfxU32 TSC, mfxI32 diffAFD, mfxU32 Scindex, mfxU32 RsDiff, mfxI32 diffMVdiffVal, mfxU32 RsCsDiff, mfxU32 TSCindex, mfxI16 tcor, mfxU32 refDCval, mfxU32 MVDiff, mfxU32 AFD, mfxI32 diffTSC, mfxI32 diffRsCsdiff, mfxU32 posBalance, mfxU32 SC, mfxU32 mu_mv_mag_sq, mfxU32 negBalance, mfxI16 mcTcor, mfxU32 gchDC, mfxU32 Cs, mfxU32 ssDCval) {
        if (tcor < 95) {
            if (SC < 8102) {
                if (MVDiff < 5) {
                    if (negBalance < 47) {
                        if (negBalance < 30) {
                            if (mcTcor < 96) {
                                if (ssDCval < 70) {
                                    if (mu_mv_mag_sq < 94) {
                                        if (Cs < 1787) {
                                            if (CsDiff < 1) {
                                                if (Rs < 53) {
                                                    return 4;
                                                }
                                                else if (Rs >= 53) {
                                                    return 8;
                                                }
                                            }
                                            else if (CsDiff >= 1) {
                                                if (Cs < 644) {
                                                    return 4;
                                                }
                                                else if (Cs >= 644) {
                                                    return 1;
                                                }
                                            }
                                        }
                                        else if (Cs >= 1787) {
                                            if (SC < 4597) {
                                                return 1;
                                            }
                                            else if (SC >= 4597) {
                                                if (Rs < 3755) {
                                                    return 8;
                                                }
                                                else if (Rs >= 3755) {
                                                    if (SC < 7755) {
                                                        return 1;
                                                    }
                                                    else if (SC >= 7755) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (mu_mv_mag_sq >= 94) {
                                        if (posBalance < 34) {
                                            if (negBalance < 1) {
                                                return 2;
                                            }
                                            else if (negBalance >= 1) {
                                                return 8;
                                            }
                                        }
                                        else if (posBalance >= 34) {
                                            if (mcTcor < 90) {
                                                if (RsDiff < 6) {
                                                    if (Rs < 1913) {
                                                        if (Rs < 1232) {
                                                            return 4;
                                                        }
                                                        else if (Rs >= 1232) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (Rs >= 1913) {
                                                        return 4;
                                                    }
                                                }
                                                else if (RsDiff >= 6) {
                                                    return 8;
                                                }
                                            }
                                            else if (mcTcor >= 90) {
                                                if (ssDCval < 56) {
                                                    if (negBalance < 18) {
                                                        return 8;
                                                    }
                                                    else if (negBalance >= 18) {
                                                        if (Cs < 1489) {
                                                            return 1;
                                                        }
                                                        else if (Cs >= 1489) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                                else if (ssDCval >= 56) {
                                                    return 8;
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (ssDCval >= 70) {
                                    if (refDCval < 75) {
                                        if (Rs < 915) {
                                            return 2;
                                        }
                                        else if (Rs >= 915) {
                                            return 1;
                                        }
                                    }
                                    else if (refDCval >= 75) {
                                        if (tcor < 80) {
                                            if (mu_mv_mag_sq < 124) {
                                                if (Rs < 1261) {
                                                    return 8;
                                                }
                                                else if (Rs >= 1261) {
                                                    if (Cs < 1481) {
                                                        return 1;
                                                    }
                                                    else if (Cs >= 1481) {
                                                        if (negBalance < 4) {
                                                            if (ssDCval < 124) {
                                                                return 4;
                                                            }
                                                            else if (ssDCval >= 124) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (negBalance >= 4) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (mu_mv_mag_sq >= 124) {
                                                return 1;
                                            }
                                        }
                                        else if (tcor >= 80) {
                                            if (mu_mv_mag_sq < 81) {
                                                if (RsDiff < 2) {
                                                    return 2;
                                                }
                                                else if (RsDiff >= 2) {
                                                    if (diffTSC < 1) {
                                                        return 8;
                                                    }
                                                    else if (diffTSC >= 1) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                            else if (mu_mv_mag_sq >= 81) {
                                                if (posBalance < 6) {
                                                    if (Scindex < 5) {
                                                        return 8;
                                                    }
                                                    else if (Scindex >= 5) {
                                                        if (Rs < 2310) {
                                                            return 1;
                                                        }
                                                        else if (Rs >= 2310) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                                else if (posBalance >= 6) {
                                                    if (mcTcor < 95) {
                                                        if (TSC < 2) {
                                                            if (Rs < 2462) {
                                                                return 4;
                                                            }
                                                            else if (Rs >= 2462) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (TSC >= 2) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (mcTcor >= 95) {
                                                        if (diffRsCsdiff < -13) {
                                                            return 4;
                                                        }
                                                        else if (diffRsCsdiff >= -13) {
                                                            if (tcor < 89) {
                                                                if (Rs < 1640) {
                                                                    return 8;
                                                                }
                                                                else if (Rs >= 1640) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (tcor >= 89) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                            else if (mcTcor >= 96) {
                                if (posBalance < 35) {
                                    if (Rs < 2326) {
                                        return 8;
                                    }
                                    else if (Rs >= 2326) {
                                        if (ssDCval < 116) {
                                            if (AFD < 4) {
                                                if (SC < 4668) {
                                                    return 4;
                                                }
                                                else if (SC >= 4668) {
                                                    return 8;
                                                }
                                            }
                                            else if (AFD >= 4) {
                                                return 4;
                                            }
                                        }
                                        else if (ssDCval >= 116) {
                                            if (ssDCval < 117) {
                                                if (SC < 4789) {
                                                    return 8;
                                                }
                                                else if (SC >= 4789) {
                                                    return 4;
                                                }
                                            }
                                            else if (ssDCval >= 117) {
                                                return 4;
                                            }
                                        }
                                    }
                                }
                                else if (posBalance >= 35) {
                                    return 4;
                                }
                            }
                        }
                        else if (negBalance >= 30) {
                            if (MVDiff < 3) {
                                if (mcTcor < 95) {
                                    if (refDCval < 56) {
                                        if (Rs < 2193) {
                                            if (diffMVdiffVal < 2) {
                                                if (Rs < 1123) {
                                                    return 4;
                                                }
                                                else if (Rs >= 1123) {
                                                    if (negBalance < 37) {
                                                        if (Cs < 999) {
                                                            return 1;
                                                        }
                                                        else if (Cs >= 999) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (negBalance >= 37) {
                                                        return 1;
                                                    }
                                                }
                                            }
                                            else if (diffMVdiffVal >= 2) {
                                                return 8;
                                            }
                                        }
                                        else if (Rs >= 2193) {
                                            return 8;
                                        }
                                    }
                                    else if (refDCval >= 56) {
                                        if (Rs < 3377) {
                                            if (posBalance < 59) {
                                                if (mcTcor < 94) {
                                                    if (mcTcor < 90) {
                                                        if (refDCval < 99) {
                                                            return 8;
                                                        }
                                                        else if (refDCval >= 99) {
                                                            if (Cs < 2252) {
                                                                return 4;
                                                            }
                                                            else if (Cs >= 2252) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                    else if (mcTcor >= 90) {
                                                        return 8;
                                                    }
                                                }
                                                else if (mcTcor >= 94) {
                                                    if (AFD < 6) {
                                                        return 8;
                                                    }
                                                    else if (AFD >= 6) {
                                                        if (Cs < 2476) {
                                                            if (Rs < 1936) {
                                                                if (mu_mv_mag_sq < 115) {
                                                                    return 4;
                                                                }
                                                                else if (mu_mv_mag_sq >= 115) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (Rs >= 1936) {
                                                                if (Cs < 2002) {
                                                                    return 2;
                                                                }
                                                                else if (Cs >= 2002) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                        else if (Cs >= 2476) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (posBalance >= 59) {
                                                if (ssDCval < 85) {
                                                    if (Rs < 1848) {
                                                        if (SC < 3246) {
                                                            if (diffAFD < 1) {
                                                                return 2;
                                                            }
                                                            else if (diffAFD >= 1) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (SC >= 3246) {
                                                            if (AFD < 8) {
                                                                return 8;
                                                            }
                                                            else if (AFD >= 8) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                    else if (Rs >= 1848) {
                                                        if (refDCval < 75) {
                                                            if (CsDiff < 5) {
                                                                return 2;
                                                            }
                                                            else if (CsDiff >= 5) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (refDCval >= 75) {
                                                            if (ssDCval < 77) {
                                                                return 4;
                                                            }
                                                            else if (ssDCval >= 77) {
                                                                if (TSC < 7) {
                                                                    return 8;
                                                                }
                                                                else if (TSC >= 7) {
                                                                    return 2;
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (ssDCval >= 85) {
                                                    return 8;
                                                }
                                            }
                                        }
                                        else if (Rs >= 3377) {
                                            if (ssDCval < 93) {
                                                if (TSC < 7) {
                                                    if (RsDiff < 7) {
                                                        return 8;
                                                    }
                                                    else if (RsDiff >= 7) {
                                                        return 1;
                                                    }
                                                }
                                                else if (TSC >= 7) {
                                                    return 2;
                                                }
                                            }
                                            else if (ssDCval >= 93) {
                                                if (mu_mv_mag_sq < 105) {
                                                    if (diffMVdiffVal < -5) {
                                                        if (AFD < 11) {
                                                            return 2;
                                                        }
                                                        else if (AFD >= 11) {
                                                            return 1;
                                                        }
                                                    }
                                                    else if (diffMVdiffVal >= -5) {
                                                        if (Rs < 3497) {
                                                            return 8;
                                                        }
                                                        else if (Rs >= 3497) {
                                                            if (Cs < 3240) {
                                                                return 4;
                                                            }
                                                            else if (Cs >= 3240) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (mu_mv_mag_sq >= 105) {
                                                    return 8;
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (mcTcor >= 95) {
                                    if (TSCindex < 7) {
                                        if (ssDCval < 85) {
                                            if (mu_mv_mag_sq < 111) {
                                                if (ssDCval < 52) {
                                                    if (posBalance < 50) {
                                                        return 4;
                                                    }
                                                    else if (posBalance >= 50) {
                                                        return 8;
                                                    }
                                                }
                                                else if (ssDCval >= 52) {
                                                    return 4;
                                                }
                                            }
                                            else if (mu_mv_mag_sq >= 111) {
                                                return 8;
                                            }
                                        }
                                        else if (ssDCval >= 85) {
                                            if (mcTcor < 96) {
                                                return 8;
                                            }
                                            else if (mcTcor >= 96) {
                                                return 4;
                                            }
                                        }
                                    }
                                    else if (TSCindex >= 7) {
                                        if (Cs < 1909) {
                                            if (ssDCval < 80) {
                                                if (Cs < 1740) {
                                                    if (Cs < 1496) {
                                                        return 8;
                                                    }
                                                    else if (Cs >= 1496) {
                                                        return 2;
                                                    }
                                                }
                                                else if (Cs >= 1740) {
                                                    return 8;
                                                }
                                            }
                                            else if (ssDCval >= 80) {
                                                return 2;
                                            }
                                        }
                                        else if (Cs >= 1909) {
                                            if (TSC < 7) {
                                                return 4;
                                            }
                                            else if (TSC >= 7) {
                                                if (Rs < 3423) {
                                                    return 8;
                                                }
                                                else if (Rs >= 3423) {
                                                    return 2;
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                            else if (MVDiff >= 3) {
                                if (mu_mv_mag_sq < 80) {
                                    if (posBalance < 47) {
                                        if (refDCval < 79) {
                                            return 4;
                                        }
                                        else if (refDCval >= 79) {
                                            return 2;
                                        }
                                    }
                                    else if (posBalance >= 47) {
                                        if (SC < 5251) {
                                            if (Cs < 729) {
                                                if (Rs < 943) {
                                                    if (diffMVdiffVal < -1) {
                                                        return 4;
                                                    }
                                                    else if (diffMVdiffVal >= -1) {
                                                        return 8;
                                                    }
                                                }
                                                else if (Rs >= 943) {
                                                    if (TSC < 5) {
                                                        return 2;
                                                    }
                                                    else if (TSC >= 5) {
                                                        return 1;
                                                    }
                                                }
                                            }
                                            else if (Cs >= 729) {
                                                if (posBalance < 61) {
                                                    if (Rs < 1113) {
                                                        if (ssDCval < 55) {
                                                            return 4;
                                                        }
                                                        else if (ssDCval >= 55) {
                                                            return 1;
                                                        }
                                                    }
                                                    else if (Rs >= 1113) {
                                                        if (diffMVdiffVal < -1) {
                                                            if (MVDiff < 4) {
                                                                return 1;
                                                            }
                                                            else if (MVDiff >= 4) {
                                                                if (posBalance < 55) {
                                                                    if (Rs < 1283) {
                                                                        return 1;
                                                                    }
                                                                    else if (Rs >= 1283) {
                                                                        if (AFD < 9) {
                                                                            return 8;
                                                                        }
                                                                        else if (AFD >= 9) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                }
                                                                else if (posBalance >= 55) {
                                                                    return 1;
                                                                }
                                                            }
                                                        }
                                                        else if (diffMVdiffVal >= -1) {
                                                            if (mu_mv_mag_sq < 61) {
                                                                if (diffAFD < 2) {
                                                                    return 1;
                                                                }
                                                                else if (diffAFD >= 2) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (mu_mv_mag_sq >= 61) {
                                                                if (Cs < 962) {
                                                                    return 4;
                                                                }
                                                                else if (Cs >= 962) {
                                                                    if (ssDCval < 55) {
                                                                        return 1;
                                                                    }
                                                                    else if (ssDCval >= 55) {
                                                                        if (Rs < 2422) {
                                                                            return 8;
                                                                        }
                                                                        else if (Rs >= 2422) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (posBalance >= 61) {
                                                    return 4;
                                                }
                                            }
                                        }
                                        else if (SC >= 5251) {
                                            if (Rs < 3642) {
                                                return 8;
                                            }
                                            else if (Rs >= 3642) {
                                                if (Cs < 3248) {
                                                    return 1;
                                                }
                                                else if (Cs >= 3248) {
                                                    return 4;
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (mu_mv_mag_sq >= 80) {
                                    if (SC < 6456) {
                                        if (Cs < 2066) {
                                            if (Cs < 1012) {
                                                return 8;
                                            }
                                            else if (Cs >= 1012) {
                                                if (Cs < 1031) {
                                                    return 4;
                                                }
                                                else if (Cs >= 1031) {
                                                    if (RsDiff < 2) {
                                                        return 4;
                                                    }
                                                    else if (RsDiff >= 2) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                        }
                                        else if (Cs >= 2066) {
                                            if (TSC < 10) {
                                                if (Cs < 2242) {
                                                    return 1;
                                                }
                                                else if (Cs >= 2242) {
                                                    return 8;
                                                }
                                            }
                                            else if (TSC >= 10) {
                                                return 1;
                                            }
                                        }
                                    }
                                    else if (SC >= 6456) {
                                        if (mu_mv_mag_sq < 91) {
                                            if (SC < 7254) {
                                                return 4;
                                            }
                                            else if (SC >= 7254) {
                                                if (Cs < 3816) {
                                                    return 8;
                                                }
                                                else if (Cs >= 3816) {
                                                    return 4;
                                                }
                                            }
                                        }
                                        else if (mu_mv_mag_sq >= 91) {
                                            return 4;
                                        }
                                    }
                                }
                            }
                        }
                    }
                    else if (negBalance >= 47) {
                        if (RsCsDiff < 33) {
                            if (SC < 3504) {
                                if (mu_mv_mag_sq < 120) {
                                    if (MVDiff < 3) {
                                        if (TSC < 7) {
                                            if (mcTcor < 85) {
                                                return 8;
                                            }
                                            else if (mcTcor >= 85) {
                                                if (diffTSC < 1) {
                                                    if (diffRsCsdiff < 10) {
                                                        return 2;
                                                    }
                                                    else if (diffRsCsdiff >= 10) {
                                                        if (SC < 3351) {
                                                            return 2;
                                                        }
                                                        else if (SC >= 3351) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                                else if (diffTSC >= 1) {
                                                    if (Rs < 1021) {
                                                        return 2;
                                                    }
                                                    else if (Rs >= 1021) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                        }
                                        else if (TSC >= 7) {
                                            return 2;
                                        }
                                    }
                                    else if (MVDiff >= 3) {
                                        if (Cs < 541) {
                                            if (diffRsCsdiff < -1) {
                                                if (Rs < 802) {
                                                    return 4;
                                                }
                                                else if (Rs >= 802) {
                                                    return 2;
                                                }
                                            }
                                            else if (diffRsCsdiff >= -1) {
                                                return 2;
                                            }
                                        }
                                        else if (Cs >= 541) {
                                            return 4;
                                        }
                                    }
                                }
                                else if (mu_mv_mag_sq >= 120) {
                                    if (AFD < 7) {
                                        return 8;
                                    }
                                    else if (AFD >= 7) {
                                        return 2;
                                    }
                                }
                            }
                            else if (SC >= 3504) {
                                if (Cs < 1675) {
                                    if (posBalance < 57) {
                                        return 8;
                                    }
                                    else if (posBalance >= 57) {
                                        return 2;
                                    }
                                }
                                else if (Cs >= 1675) {
                                    if (TSCindex < 8) {
                                        return 8;
                                    }
                                    else if (TSCindex >= 8) {
                                        return 4;
                                    }
                                }
                            }
                        }
                        else if (RsCsDiff >= 33) {
                            if (refDCval < 98) {
                                if (mu_mv_mag_sq < 106) {
                                    if (mcTcor < 91) {
                                        if (TSC < 10) {
                                            if (negBalance < 53) {
                                                if (diffAFD < 1) {
                                                    if (TSCindex < 9) {
                                                        if (Rs < 1075) {
                                                            return 4;
                                                        }
                                                        else if (Rs >= 1075) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (TSCindex >= 9) {
                                                        if (mcTcor < 90) {
                                                            if (posBalance < 61) {
                                                                return 2;
                                                            }
                                                            else if (posBalance >= 61) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (mcTcor >= 90) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                                else if (diffAFD >= 1) {
                                                    return 4;
                                                }
                                            }
                                            else if (negBalance >= 53) {
                                                if (mu_mv_mag_sq < 99) {
                                                    return 2;
                                                }
                                                else if (mu_mv_mag_sq >= 99) {
                                                    return 8;
                                                }
                                            }
                                        }
                                        else if (TSC >= 10) {
                                            return 8;
                                        }
                                    }
                                    else if (mcTcor >= 91) {
                                        if (posBalance < 58) {
                                            if (SC < 4461) {
                                                if (Rs < 2317) {
                                                    return 2;
                                                }
                                                else if (Rs >= 2317) {
                                                    if (Cs < 1761) {
                                                        return 1;
                                                    }
                                                    else if (Cs >= 1761) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                            else if (SC >= 4461) {
                                                if (diffAFD < -2) {
                                                    if (SC < 4716) {
                                                        return 2;
                                                    }
                                                    else if (SC >= 4716) {
                                                        if (Cs < 2358) {
                                                            return 8;
                                                        }
                                                        else if (Cs >= 2358) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                                else if (diffAFD >= -2) {
                                                    if (Rs < 3330) {
                                                        return 4;
                                                    }
                                                    else if (Rs >= 3330) {
                                                        if (SC < 7241) {
                                                            return 2;
                                                        }
                                                        else if (SC >= 7241) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (posBalance >= 58) {
                                            if (posBalance < 62) {
                                                if (RsDiff < 9) {
                                                    if (refDCval < 86) {
                                                        if (Rs < 2232) {
                                                            return 2;
                                                        }
                                                        else if (Rs >= 2232) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (refDCval >= 86) {
                                                        return 2;
                                                    }
                                                }
                                                else if (RsDiff >= 9) {
                                                    return 8;
                                                }
                                            }
                                            else if (posBalance >= 62) {
                                                if (RsCsDiff < 129) {
                                                    if (tcor < 89) {
                                                        if (Rs < 1270) {
                                                            return 4;
                                                        }
                                                        else if (Rs >= 1270) {
                                                            return 2;
                                                        }
                                                    }
                                                    else if (tcor >= 89) {
                                                        return 4;
                                                    }
                                                }
                                                else if (RsCsDiff >= 129) {
                                                    return 2;
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (mu_mv_mag_sq >= 106) {
                                    if (mcTcor < 93) {
                                        if (mcTcor < 92) {
                                            if (posBalance < 62) {
                                                return 8;
                                            }
                                            else if (posBalance >= 62) {
                                                return 2;
                                            }
                                        }
                                        else if (mcTcor >= 92) {
                                            if (AFD < 8) {
                                                return 8;
                                            }
                                            else if (AFD >= 8) {
                                                if (MVDiff < 1) {
                                                    return 2;
                                                }
                                                else if (MVDiff >= 1) {
                                                    return 8;
                                                }
                                            }
                                        }
                                    }
                                    else if (mcTcor >= 93) {
                                        if (Rs < 2933) {
                                            if (ssDCval < 77) {
                                                if (Rs < 1689) {
                                                    return 8;
                                                }
                                                else if (Rs >= 1689) {
                                                    return 2;
                                                }
                                            }
                                            else if (ssDCval >= 77) {
                                                return 8;
                                            }
                                        }
                                        else if (Rs >= 2933) {
                                            return 2;
                                        }
                                    }
                                }
                            }
                            else if (refDCval >= 98) {
                                if (AFD < 10) {
                                    if (mu_mv_mag_sq < 115) {
                                        if (refDCval < 111) {
                                            return 4;
                                        }
                                        else if (refDCval >= 111) {
                                            if (MVDiff < 4) {
                                                if (mu_mv_mag_sq < 70) {
                                                    return 4;
                                                }
                                                else if (mu_mv_mag_sq >= 70) {
                                                    return 8;
                                                }
                                            }
                                            else if (MVDiff >= 4) {
                                                if (CsDiff < 7) {
                                                    if (ssDCval < 117) {
                                                        return 8;
                                                    }
                                                    else if (ssDCval >= 117) {
                                                        if (tcor < 92) {
                                                            return 4;
                                                        }
                                                        else if (tcor >= 92) {
                                                            if (ssDCval < 118) {
                                                                return 4;
                                                            }
                                                            else if (ssDCval >= 118) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (CsDiff >= 7) {
                                                    return 8;
                                                }
                                            }
                                        }
                                    }
                                    else if (mu_mv_mag_sq >= 115) {
                                        if (tcor < 88) {
                                            return 8;
                                        }
                                        else if (tcor >= 88) {
                                            if (Cs < 1641) {
                                                return 8;
                                            }
                                            else if (Cs >= 1641) {
                                                return 4;
                                            }
                                        }
                                    }
                                }
                                else if (AFD >= 10) {
                                    if (refDCval < 127) {
                                        if (AFD < 16) {
                                            if (negBalance < 52) {
                                                if (ssDCval < 110) {
                                                    if (TSC < 8) {
                                                        return 8;
                                                    }
                                                    else if (TSC >= 8) {
                                                        return 4;
                                                    }
                                                }
                                                else if (ssDCval >= 110) {
                                                    if (mu_mv_mag_sq < 64) {
                                                        if (Rs < 2979) {
                                                            if (Cs < 1732) {
                                                                return 4;
                                                            }
                                                            else if (Cs >= 1732) {
                                                                if (posBalance < 60) {
                                                                    return 8;
                                                                }
                                                                else if (posBalance >= 60) {
                                                                    if (Cs < 1777) {
                                                                        return 8;
                                                                    }
                                                                    else if (Cs >= 1777) {
                                                                        return 4;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (Rs >= 2979) {
                                                            return 2;
                                                        }
                                                    }
                                                    else if (mu_mv_mag_sq >= 64) {
                                                        if (ssDCval < 120) {
                                                            return 8;
                                                        }
                                                        else if (ssDCval >= 120) {
                                                            if (RsCsDiff < 121) {
                                                                return 8;
                                                            }
                                                            else if (RsCsDiff >= 121) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                            else if (negBalance >= 52) {
                                                if (diffMVdiffVal < -6) {
                                                    return 8;
                                                }
                                                else if (diffMVdiffVal >= -6) {
                                                    if (MVDiff < 3) {
                                                        return 4;
                                                    }
                                                    else if (MVDiff >= 3) {
                                                        if (ssDCval < 124) {
                                                            return 4;
                                                        }
                                                        else if (ssDCval >= 124) {
                                                            if (Rs < 2715) {
                                                                return 4;
                                                            }
                                                            else if (Rs >= 2715) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (AFD >= 16) {
                                            if (Rs < 2728) {
                                                if (Rs < 2406) {
                                                    return 2;
                                                }
                                                else if (Rs >= 2406) {
                                                    return 4;
                                                }
                                            }
                                            else if (Rs >= 2728) {
                                                if (TSC < 13) {
                                                    if (refDCval < 103) {
                                                        return 1;
                                                    }
                                                    else if (refDCval >= 103) {
                                                        return 8;
                                                    }
                                                }
                                                else if (TSC >= 13) {
                                                    return 1;
                                                }
                                            }
                                        }
                                    }
                                    else if (refDCval >= 127) {
                                        if (tcor < 77) {
                                            return 1;
                                        }
                                        else if (tcor >= 77) {
                                            if (SC < 4100) {
                                                return 2;
                                            }
                                            else if (SC >= 4100) {
                                                return 8;
                                            }
                                        }
                                    }
                                }
                            }
                        }
                    }
                }
                else if (MVDiff >= 5) {
                    if (AFD < 15) {
                        if (mu_mv_mag_sq < 42) {
                            if (Cs < 313) {
                                if (TSC < 5) {
                                    if (negBalance < 33) {
                                        if (tcor < 68) {
                                            if (ssDCval < 20) {
                                                if (posBalance < 23) {
                                                    return 8;
                                                }
                                                else if (posBalance >= 23) {
                                                    if (negBalance < 2) {
                                                        if (posBalance < 28) {
                                                            return 1;
                                                        }
                                                        else if (posBalance >= 28) {
                                                            if (diffMVdiffVal < -816) {
                                                                return 2;
                                                            }
                                                            else if (diffMVdiffVal >= -816) {
                                                                if (MVDiff < 210) {
                                                                    if (diffMVdiffVal < 14) {
                                                                        if (MVDiff < 62) {
                                                                            return 8;
                                                                        }
                                                                        else if (MVDiff >= 62) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                    else if (diffMVdiffVal >= 14) {
                                                                        return 8;
                                                                    }
                                                                }
                                                                else if (MVDiff >= 210) {
                                                                    if (TSC < 1) {
                                                                        if (posBalance < 46) {
                                                                            return 2;
                                                                        }
                                                                        else if (posBalance >= 46) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                    else if (TSC >= 1) {
                                                                        return 1;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (negBalance >= 2) {
                                                        if (MVDiff < 58) {
                                                            return 1;
                                                        }
                                                        else if (MVDiff >= 58) {
                                                            if (mu_mv_mag_sq < 33) {
                                                                if (mcTcor < 48) {
                                                                    return 4;
                                                                }
                                                                else if (mcTcor >= 48) {
                                                                    if (mcTcor < 49) {
                                                                        return 8;
                                                                    }
                                                                    else if (mcTcor >= 49) {
                                                                        return 4;
                                                                    }
                                                                }
                                                            }
                                                            else if (mu_mv_mag_sq >= 33) {
                                                                if (MVDiff < 59) {
                                                                    return 4;
                                                                }
                                                                else if (MVDiff >= 59) {
                                                                    return 1;
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                            else if (ssDCval >= 20) {
                                                if (AFD < 1) {
                                                    return 2;
                                                }
                                                else if (AFD >= 1) {
                                                    if (MVDiff < 94) {
                                                        if (Rs < 8) {
                                                            if (AFD < 3) {
                                                                if (refDCval < 27) {
                                                                    if (posBalance < 101) {
                                                                        return 2;
                                                                    }
                                                                    else if (posBalance >= 101) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (refDCval >= 27) {
                                                                    return 1;
                                                                }
                                                            }
                                                            else if (AFD >= 3) {
                                                                return 2;
                                                            }
                                                        }
                                                        else if (Rs >= 8) {
                                                            if (tcor < 57) {
                                                                return 1;
                                                            }
                                                            else if (tcor >= 57) {
                                                                if (Cs < 27) {
                                                                    return 4;
                                                                }
                                                                else if (Cs >= 27) {
                                                                    return 1;
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (MVDiff >= 94) {
                                                        if (mu_mv_mag_sq < 17) {
                                                            return 1;
                                                        }
                                                        else if (mu_mv_mag_sq >= 17) {
                                                            if (refDCval < 22) {
                                                                if (AFD < 3) {
                                                                    return 4;
                                                                }
                                                                else if (AFD >= 3) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (refDCval >= 22) {
                                                                return 1;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (tcor >= 68) {
                                            if (Cs < 36) {
                                                if (diffMVdiffVal < -5) {
                                                    if (tcor < 84) {
                                                        if (mu_mv_mag_sq < 20) {
                                                            if (AFD < 4) {
                                                                return 2;
                                                            }
                                                            else if (AFD >= 4) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (mu_mv_mag_sq >= 20) {
                                                            if (ssDCval < 35) {
                                                                return 4;
                                                            }
                                                            else if (ssDCval >= 35) {
                                                                if (Rs < 64) {
                                                                    return 4;
                                                                }
                                                                else if (Rs >= 64) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (tcor >= 84) {
                                                        return 8;
                                                    }
                                                }
                                                else if (diffMVdiffVal >= -5) {
                                                    if (negBalance < 2) {
                                                        if (TSC < 1) {
                                                            if (MVDiff < 197) {
                                                                return 4;
                                                            }
                                                            else if (MVDiff >= 197) {
                                                                if (MVDiff < 653) {
                                                                    return 8;
                                                                }
                                                                else if (MVDiff >= 653) {
                                                                    return 2;
                                                                }
                                                            }
                                                        }
                                                        else if (TSC >= 1) {
                                                            return 2;
                                                        }
                                                    }
                                                    else if (negBalance >= 2) {
                                                        if (ssDCval < 19) {
                                                            return 4;
                                                        }
                                                        else if (ssDCval >= 19) {
                                                            if (refDCval < 21) {
                                                                if (negBalance < 10) {
                                                                    return 8;
                                                                }
                                                                else if (negBalance >= 10) {
                                                                    return 2;
                                                                }
                                                            }
                                                            else if (refDCval >= 21) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                            else if (Cs >= 36) {
                                                if (Scindex < 3) {
                                                    if (posBalance < 55) {
                                                        return 1;
                                                    }
                                                    else if (posBalance >= 55) {
                                                        return 2;
                                                    }
                                                }
                                                else if (Scindex >= 3) {
                                                    if (AFD < 7) {
                                                        return 4;
                                                    }
                                                    else if (AFD >= 7) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (negBalance >= 33) {
                                        if (negBalance < 70) {
                                            if (Rs < 587) {
                                                if (Cs < 134) {
                                                    return 2;
                                                }
                                                else if (Cs >= 134) {
                                                    if (SC < 432) {
                                                        return 1;
                                                    }
                                                    else if (SC >= 432) {
                                                        if (SC < 464) {
                                                            return 4;
                                                        }
                                                        else if (SC >= 464) {
                                                            if (Cs < 281) {
                                                                return 2;
                                                            }
                                                            else if (Cs >= 281) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                            else if (Rs >= 587) {
                                                if (RsDiff < 6) {
                                                    return 4;
                                                }
                                                else if (RsDiff >= 6) {
                                                    return 1;
                                                }
                                            }
                                        }
                                        else if (negBalance >= 70) {
                                            if (TSC < 1) {
                                                return 4;
                                            }
                                            else if (TSC >= 1) {
                                                return 1;
                                            }
                                        }
                                    }
                                }
                                else if (TSC >= 5) {
                                    if (ssDCval < 108) {
                                        if (ssDCval < 82) {
                                            if (refDCval < 67) {
                                                if (tcor < 74) {
                                                    if (Rs < 4) {
                                                        return 2;
                                                    }
                                                    else if (Rs >= 4) {
                                                        return 1;
                                                    }
                                                }
                                                else if (tcor >= 74) {
                                                    return 2;
                                                }
                                            }
                                            else if (refDCval >= 67) {
                                                return 8;
                                            }
                                        }
                                        else if (ssDCval >= 82) {
                                            if (tcor < 65) {
                                                return 2;
                                            }
                                            else if (tcor >= 65) {
                                                return 1;
                                            }
                                        }
                                    }
                                    else if (ssDCval >= 108) {
                                        if (negBalance < 54) {
                                            return 4;
                                        }
                                        else if (negBalance >= 54) {
                                            return 2;
                                        }
                                    }
                                }
                            }
                            else if (Cs >= 313) {
                                if (Rs < 3045) {
                                    if (posBalance < 59) {
                                        if (negBalance < 46) {
                                            if (ssDCval < 40) {
                                                if (SC < 1382) {
                                                    if (Rs < 405) {
                                                        return 4;
                                                    }
                                                    else if (Rs >= 405) {
                                                        return 1;
                                                    }
                                                }
                                                else if (SC >= 1382) {
                                                    if (Rs < 1051) {
                                                        return 2;
                                                    }
                                                    else if (Rs >= 1051) {
                                                        if (Rs < 1148) {
                                                            return 1;
                                                        }
                                                        else if (Rs >= 1148) {
                                                            return 2;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (ssDCval >= 40) {
                                                if (TSCindex < 8) {
                                                    if (Rs < 1815) {
                                                        if (refDCval < 141) {
                                                            if (Rs < 867) {
                                                                if (posBalance < 52) {
                                                                    return 4;
                                                                }
                                                                else if (posBalance >= 52) {
                                                                    if (Rs < 626) {
                                                                        return 4;
                                                                    }
                                                                    else if (Rs >= 626) {
                                                                        return 8;
                                                                    }
                                                                }
                                                            }
                                                            else if (Rs >= 867) {
                                                                if (MVDiff < 7) {
                                                                    if (CsDiff < 4) {
                                                                        if (Cs < 593) {
                                                                            if (Rs < 1108) {
                                                                                return 4;
                                                                            }
                                                                            else if (Rs >= 1108) {
                                                                                return 2;
                                                                            }
                                                                        }
                                                                        else if (Cs >= 593) {
                                                                            if (diffMVdiffVal < -1) {
                                                                                return 1;
                                                                            }
                                                                            else if (diffMVdiffVal >= -1) {
                                                                                if (posBalance < 54) {
                                                                                    if (Cs < 867) {
                                                                                        return 4;
                                                                                    }
                                                                                    else if (Cs >= 867) {
                                                                                        return 1;
                                                                                    }
                                                                                }
                                                                                else if (posBalance >= 54) {
                                                                                    return 8;
                                                                                }
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (CsDiff >= 4) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (MVDiff >= 7) {
                                                                    if (negBalance < 36) {
                                                                        return 1;
                                                                    }
                                                                    else if (negBalance >= 36) {
                                                                        if (ssDCval < 67) {
                                                                            if (diffMVdiffVal < 1) {
                                                                                if (diffMVdiffVal < -2) {
                                                                                    if (TSCindex < 7) {
                                                                                        return 1;
                                                                                    }
                                                                                    else if (TSCindex >= 7) {
                                                                                        return 8;
                                                                                    }
                                                                                }
                                                                                else if (diffMVdiffVal >= -2) {
                                                                                    if (diffTSC < 1) {
                                                                                        return 8;
                                                                                    }
                                                                                    else if (diffTSC >= 1) {
                                                                                        if (AFD < 9) {
                                                                                            return 1;
                                                                                        }
                                                                                        else if (AFD >= 9) {
                                                                                            return 4;
                                                                                        }
                                                                                    }
                                                                                }
                                                                            }
                                                                            else if (diffMVdiffVal >= 1) {
                                                                                return 1;
                                                                            }
                                                                        }
                                                                        else if (ssDCval >= 67) {
                                                                            if (TSC < 6) {
                                                                                return 2;
                                                                            }
                                                                            else if (TSC >= 6) {
                                                                                if (SC < 2795) {
                                                                                    if (Rs < 1162) {
                                                                                        return 1;
                                                                                    }
                                                                                    else if (Rs >= 1162) {
                                                                                        return 8;
                                                                                    }
                                                                                }
                                                                                else if (SC >= 2795) {
                                                                                    return 4;
                                                                                }
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (refDCval >= 141) {
                                                            if (ssDCval < 149) {
                                                                return 4;
                                                            }
                                                            else if (ssDCval >= 149) {
                                                                if (RsDiff < 9) {
                                                                    if (SC < 2245) {
                                                                        return 4;
                                                                    }
                                                                    else if (SC >= 2245) {
                                                                        return 1;
                                                                    }
                                                                }
                                                                else if (RsDiff >= 9) {
                                                                    return 1;
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (Rs >= 1815) {
                                                        if (SC < 3289) {
                                                            if (ssDCval < 62) {
                                                                return 4;
                                                            }
                                                            else if (ssDCval >= 62) {
                                                                return 2;
                                                            }
                                                        }
                                                        else if (SC >= 3289) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                                else if (TSCindex >= 8) {
                                                    if (RsDiff < 8) {
                                                        if (tcor < 87) {
                                                            if (TSC < 9) {
                                                                return 1;
                                                            }
                                                            else if (TSC >= 9) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (tcor >= 87) {
                                                            if (AFD < 14) {
                                                                return 8;
                                                            }
                                                            else if (AFD >= 14) {
                                                                return 2;
                                                            }
                                                        }
                                                    }
                                                    else if (RsDiff >= 8) {
                                                        if (refDCval < 82) {
                                                            if (ssDCval < 50) {
                                                                return 4;
                                                            }
                                                            else if (ssDCval >= 50) {
                                                                if (Cs < 2146) {
                                                                    return 4;
                                                                }
                                                                else if (Cs >= 2146) {
                                                                    if (SC < 4208) {
                                                                        return 1;
                                                                    }
                                                                    else if (SC >= 4208) {
                                                                        return 8;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (refDCval >= 82) {
                                                            if (TSC < 11) {
                                                                if (mu_mv_mag_sq < 3) {
                                                                    return 1;
                                                                }
                                                                else if (mu_mv_mag_sq >= 3) {
                                                                    if (MVDiff < 19) {
                                                                        if (posBalance < 53) {
                                                                            return 1;
                                                                        }
                                                                        else if (posBalance >= 53) {
                                                                            if (Cs < 1907) {
                                                                                return 1;
                                                                            }
                                                                            else if (Cs >= 1907) {
                                                                                return 8;
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (MVDiff >= 19) {
                                                                        if (Rs < 2316) {
                                                                            if (SC < 4130) {
                                                                                return 8;
                                                                            }
                                                                            else if (SC >= 4130) {
                                                                                if (SC < 4344) {
                                                                                    return 1;
                                                                                }
                                                                                else if (SC >= 4344) {
                                                                                    return 8;
                                                                                }
                                                                            }
                                                                        }
                                                                        else if (Rs >= 2316) {
                                                                            if (ssDCval < 89) {
                                                                                return 8;
                                                                            }
                                                                            else if (ssDCval >= 89) {
                                                                                return 1;
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                            else if (TSC >= 11) {
                                                                if (gchDC < 2) {
                                                                    return 8;
                                                                }
                                                                else if (gchDC >= 2) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (negBalance >= 46) {
                                            if (tcor < 78) {
                                                if (Cs < 680) {
                                                    if (Rs < 675) {
                                                        if (mcTcor < 79) {
                                                            if (Rs < 477) {
                                                                return 8;
                                                            }
                                                            else if (Rs >= 477) {
                                                                return 2;
                                                            }
                                                        }
                                                        else if (mcTcor >= 79) {
                                                            if (negBalance < 50) {
                                                                if (TSC < 3) {
                                                                    return 2;
                                                                }
                                                                else if (TSC >= 3) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (negBalance >= 50) {
                                                                return 2;
                                                            }
                                                        }
                                                    }
                                                    else if (Rs >= 675) {
                                                        if (SC < 1423) {
                                                            if (Cs < 501) {
                                                                return 4;
                                                            }
                                                            else if (Cs >= 501) {
                                                                if (RsDiff < 2) {
                                                                    return 4;
                                                                }
                                                                else if (RsDiff >= 2) {
                                                                    if (Cs < 623) {
                                                                        return 1;
                                                                    }
                                                                    else if (Cs >= 623) {
                                                                        return 8;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (SC >= 1423) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                                else if (Cs >= 680) {
                                                    if (Rs < 1866) {
                                                        if (mu_mv_mag_sq < 2) {
                                                            return 8;
                                                        }
                                                        else if (mu_mv_mag_sq >= 2) {
                                                            if (mcTcor < 82) {
                                                                if (mu_mv_mag_sq < 35) {
                                                                    return 1;
                                                                }
                                                                else if (mu_mv_mag_sq >= 35) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (mcTcor >= 82) {
                                                                if (mu_mv_mag_sq < 16) {
                                                                    return 2;
                                                                }
                                                                else if (mu_mv_mag_sq >= 16) {
                                                                    if (Rs < 1362) {
                                                                        return 1;
                                                                    }
                                                                    else if (Rs >= 1362) {
                                                                        return 4;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (Rs >= 1866) {
                                                        if (ssDCval < 122) {
                                                            if (Cs < 1865) {
                                                                return 4;
                                                            }
                                                            else if (Cs >= 1865) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (ssDCval >= 122) {
                                                            return 1;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (tcor >= 78) {
                                                if (MVDiff < 11) {
                                                    if (Cs < 732) {
                                                        if (ssDCval < 58) {
                                                            return 4;
                                                        }
                                                        else if (ssDCval >= 58) {
                                                            return 2;
                                                        }
                                                    }
                                                    else if (Cs >= 732) {
                                                        if (ssDCval < 102) {
                                                            if (CsDiff < 8) {
                                                                if (SC < 3947) {
                                                                    if (refDCval < 91) {
                                                                        if (AFD < 13) {
                                                                            if (MVDiff < 10) {
                                                                                return 1;
                                                                            }
                                                                            else if (MVDiff >= 10) {
                                                                                if (Cs < 1033) {
                                                                                    return 8;
                                                                                }
                                                                                else if (Cs >= 1033) {
                                                                                    return 1;
                                                                                }
                                                                            }
                                                                        }
                                                                        else if (AFD >= 13) {
                                                                            return 2;
                                                                        }
                                                                    }
                                                                    else if (refDCval >= 91) {
                                                                        return 2;
                                                                    }
                                                                }
                                                                else if (SC >= 3947) {
                                                                    return 1;
                                                                }
                                                            }
                                                            else if (CsDiff >= 8) {
                                                                if (Cs < 2363) {
                                                                    return 2;
                                                                }
                                                                else if (Cs >= 2363) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                        else if (ssDCval >= 102) {
                                                            if (RsDiff < 4) {
                                                                return 8;
                                                            }
                                                            else if (RsDiff >= 4) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (MVDiff >= 11) {
                                                    if (mu_mv_mag_sq < 27) {
                                                        if (mcTcor < 92) {
                                                            if (diffTSC < -2) {
                                                                if (Rs < 1030) {
                                                                    return 8;
                                                                }
                                                                else if (Rs >= 1030) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (diffTSC >= -2) {
                                                                if (refDCval < 114) {
                                                                    if (refDCval < 70) {
                                                                        if (Rs < 901) {
                                                                            return 2;
                                                                        }
                                                                        else if (Rs >= 901) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                    else if (refDCval >= 70) {
                                                                        return 2;
                                                                    }
                                                                }
                                                                else if (refDCval >= 114) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                        else if (mcTcor >= 92) {
                                                            if (SC < 1995) {
                                                                if (AFD < 7) {
                                                                    return 4;
                                                                }
                                                                else if (AFD >= 7) {
                                                                    return 2;
                                                                }
                                                            }
                                                            else if (SC >= 1995) {
                                                                if (tcor < 88) {
                                                                    if (RsDiff < 7) {
                                                                        return 1;
                                                                    }
                                                                    else if (RsDiff >= 7) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (tcor >= 88) {
                                                                    if (Cs < 1385) {
                                                                        return 2;
                                                                    }
                                                                    else if (Cs >= 1385) {
                                                                        return 1;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (mu_mv_mag_sq >= 27) {
                                                        if (posBalance < 51) {
                                                            if (ssDCval < 50) {
                                                                return 4;
                                                            }
                                                            else if (ssDCval >= 50) {
                                                                if (RsCsDiff < 19) {
                                                                    if (Rs < 1889) {
                                                                        return 1;
                                                                    }
                                                                    else if (Rs >= 1889) {
                                                                        return 8;
                                                                    }
                                                                }
                                                                else if (RsCsDiff >= 19) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                        else if (posBalance >= 51) {
                                                            if (Cs < 657) {
                                                                if (Rs < 750) {
                                                                    return 4;
                                                                }
                                                                else if (Rs >= 750) {
                                                                    return 1;
                                                                }
                                                            }
                                                            else if (Cs >= 657) {
                                                                if (MVDiff < 133) {
                                                                    if (ssDCval < 96) {
                                                                        if (posBalance < 56) {
                                                                            return 1;
                                                                        }
                                                                        else if (posBalance >= 56) {
                                                                            if (diffTSC < -1) {
                                                                                return 1;
                                                                            }
                                                                            else if (diffTSC >= -1) {
                                                                                if (mu_mv_mag_sq < 39) {
                                                                                    if (RsCsDiff < 79) {
                                                                                        return 1;
                                                                                    }
                                                                                    else if (RsCsDiff >= 79) {
                                                                                        if (diffRsCsdiff < -21) {
                                                                                            return 1;
                                                                                        }
                                                                                        else if (diffRsCsdiff >= -21) {
                                                                                            return 2;
                                                                                        }
                                                                                    }
                                                                                }
                                                                                else if (mu_mv_mag_sq >= 39) {
                                                                                    if (diffAFD < -1) {
                                                                                        return 1;
                                                                                    }
                                                                                    else if (diffAFD >= -1) {
                                                                                        return 2;
                                                                                    }
                                                                                }
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (ssDCval >= 96) {
                                                                        if (MVDiff < 19) {
                                                                            return 8;
                                                                        }
                                                                        else if (MVDiff >= 19) {
                                                                            return 2;
                                                                        }
                                                                    }
                                                                }
                                                                else if (MVDiff >= 133) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (posBalance >= 59) {
                                        if (mu_mv_mag_sq < 5) {
                                            if (RsCsDiff < 9) {
                                                if (Rs < 262) {
                                                    return 2;
                                                }
                                                else if (Rs >= 262) {
                                                    if (negBalance < 50) {
                                                        return 4;
                                                    }
                                                    else if (negBalance >= 50) {
                                                        if (Rs < 629) {
                                                            if (posBalance < 60) {
                                                                return 4;
                                                            }
                                                            else if (posBalance >= 60) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (Rs >= 629) {
                                                            if (diffAFD < 1) {
                                                                if (Cs < 588) {
                                                                    return 4;
                                                                }
                                                                else if (Cs >= 588) {
                                                                    if (SC < 1280) {
                                                                        return 8;
                                                                    }
                                                                    else if (SC >= 1280) {
                                                                        return 4;
                                                                    }
                                                                }
                                                            }
                                                            else if (diffAFD >= 1) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                            else if (RsCsDiff >= 9) {
                                                if (MVDiff < 22) {
                                                    if (negBalance < 50) {
                                                        if (RsDiff < 9) {
                                                            if (Rs < 909) {
                                                                if (ssDCval < 81) {
                                                                    if (MVDiff < 20) {
                                                                        return 2;
                                                                    }
                                                                    else if (MVDiff >= 20) {
                                                                        return 1;
                                                                    }
                                                                }
                                                                else if (ssDCval >= 81) {
                                                                    if (CsDiff < 4) {
                                                                        return 4;
                                                                    }
                                                                    else if (CsDiff >= 4) {
                                                                        return 8;
                                                                    }
                                                                }
                                                            }
                                                            else if (Rs >= 909) {
                                                                if (negBalance < 39) {
                                                                    if (Rs < 1943) {
                                                                        return 4;
                                                                    }
                                                                    else if (Rs >= 1943) {
                                                                        return 8;
                                                                    }
                                                                }
                                                                else if (negBalance >= 39) {
                                                                    if (Cs < 1509) {
                                                                        return 1;
                                                                    }
                                                                    else if (Cs >= 1509) {
                                                                        return 4;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (RsDiff >= 9) {
                                                            if (diffAFD < -1) {
                                                                return 2;
                                                            }
                                                            else if (diffAFD >= -1) {
                                                                if (Rs < 2144) {
                                                                    return 8;
                                                                }
                                                                else if (Rs >= 2144) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (negBalance >= 50) {
                                                        if (tcor < 87) {
                                                            if (ssDCval < 79) {
                                                                if (posBalance < 60) {
                                                                    if (Rs < 1020) {
                                                                        return 8;
                                                                    }
                                                                    else if (Rs >= 1020) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (posBalance >= 60) {
                                                                    if (Rs < 1433) {
                                                                        return 8;
                                                                    }
                                                                    else if (Rs >= 1433) {
                                                                        return 1;
                                                                    }
                                                                }
                                                            }
                                                            else if (ssDCval >= 79) {
                                                                if (Rs < 1104) {
                                                                    return 4;
                                                                }
                                                                else if (Rs >= 1104) {
                                                                    if (ssDCval < 111) {
                                                                        if (Rs < 1695) {
                                                                            if (TSC < 6) {
                                                                                return 2;
                                                                            }
                                                                            else if (TSC >= 6) {
                                                                                if (TSC < 8) {
                                                                                    return 4;
                                                                                }
                                                                                else if (TSC >= 8) {
                                                                                    return 8;
                                                                                }
                                                                            }
                                                                        }
                                                                        else if (Rs >= 1695) {
                                                                            if (Cs < 1822) {
                                                                                return 2;
                                                                            }
                                                                            else if (Cs >= 1822) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (ssDCval >= 111) {
                                                                        return 4;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (tcor >= 87) {
                                                            if (mcTcor < 94) {
                                                                if (diffMVdiffVal < 4) {
                                                                    return 4;
                                                                }
                                                                else if (diffMVdiffVal >= 4) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (mcTcor >= 94) {
                                                                return 1;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (MVDiff >= 22) {
                                                    if (RsCsDiff < 46) {
                                                        return 4;
                                                    }
                                                    else if (RsCsDiff >= 46) {
                                                        if (Rs < 2810) {
                                                            return 2;
                                                        }
                                                        else if (Rs >= 2810) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (mu_mv_mag_sq >= 5) {
                                            if (Cs < 452) {
                                                if (mu_mv_mag_sq < 29) {
                                                    if (refDCval < 68) {
                                                        if (Rs < 328) {
                                                            return 1;
                                                        }
                                                        else if (Rs >= 328) {
                                                            return 2;
                                                        }
                                                    }
                                                    else if (refDCval >= 68) {
                                                        if (mu_mv_mag_sq < 10) {
                                                            return 8;
                                                        }
                                                        else if (mu_mv_mag_sq >= 10) {
                                                            return 2;
                                                        }
                                                    }
                                                }
                                                else if (mu_mv_mag_sq >= 29) {
                                                    if (tcor < 84) {
                                                        return 8;
                                                    }
                                                    else if (tcor >= 84) {
                                                        if (Rs < 1231) {
                                                            return 4;
                                                        }
                                                        else if (Rs >= 1231) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (Cs >= 452) {
                                                if (mcTcor < 94) {
                                                    if (posBalance < 60) {
                                                        if (diffMVdiffVal < 6) {
                                                            if (gchDC < 1) {
                                                                if (AFD < 11) {
                                                                    if (diffAFD < -1) {
                                                                        return 4;
                                                                    }
                                                                    else if (diffAFD >= -1) {
                                                                        if (TSC < 5) {
                                                                            return 4;
                                                                        }
                                                                        else if (TSC >= 5) {
                                                                            if (mcTcor < 75) {
                                                                                if (SC < 1346) {
                                                                                    return 4;
                                                                                }
                                                                                else if (SC >= 1346) {
                                                                                    return 1;
                                                                                }
                                                                            }
                                                                            else if (mcTcor >= 75) {
                                                                                return 1;
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                                else if (AFD >= 11) {
                                                                    return 1;
                                                                }
                                                            }
                                                            else if (gchDC >= 1) {
                                                                if (mu_mv_mag_sq < 26) {
                                                                    if (AFD < 14) {
                                                                        if (negBalance < 48) {
                                                                            return 1;
                                                                        }
                                                                        else if (negBalance >= 48) {
                                                                            return 2;
                                                                        }
                                                                    }
                                                                    else if (AFD >= 14) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (mu_mv_mag_sq >= 26) {
                                                                    return 1;
                                                                }
                                                            }
                                                        }
                                                        else if (diffMVdiffVal >= 6) {
                                                            if (tcor < 76) {
                                                                if (RsCsDiff < 7) {
                                                                    return 1;
                                                                }
                                                                else if (RsCsDiff >= 7) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (tcor >= 76) {
                                                                return 1;
                                                            }
                                                        }
                                                    }
                                                    else if (posBalance >= 60) {
                                                        if (mu_mv_mag_sq < 25) {
                                                            if (Rs < 991) {
                                                                if (diffAFD < 1) {
                                                                    return 2;
                                                                }
                                                                else if (diffAFD >= 1) {
                                                                    if (RsCsDiff < 4) {
                                                                        return 2;
                                                                    }
                                                                    else if (RsCsDiff >= 4) {
                                                                        return 1;
                                                                    }
                                                                }
                                                            }
                                                            else if (Rs >= 991) {
                                                                if (CsDiff < 3) {
                                                                    return 2;
                                                                }
                                                                else if (CsDiff >= 3) {
                                                                    return 1;
                                                                }
                                                            }
                                                        }
                                                        else if (mu_mv_mag_sq >= 25) {
                                                            if (ssDCval < 83) {
                                                                if (posBalance < 62) {
                                                                    if (TSC < 11) {
                                                                        if (tcor < 86) {
                                                                            return 1;
                                                                        }
                                                                        else if (tcor >= 86) {
                                                                            if (Rs < 1986) {
                                                                                return 4;
                                                                            }
                                                                            else if (Rs >= 1986) {
                                                                                return 1;
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (TSC >= 11) {
                                                                        return 8;
                                                                    }
                                                                }
                                                                else if (posBalance >= 62) {
                                                                    if (diffMVdiffVal < 7) {
                                                                        if (AFD < 3) {
                                                                            return 2;
                                                                        }
                                                                        else if (AFD >= 3) {
                                                                            if (Cs < 1144) {
                                                                                return 4;
                                                                            }
                                                                            else if (Cs >= 1144) {
                                                                                return 1;
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (diffMVdiffVal >= 7) {
                                                                        if (Rs < 991) {
                                                                            return 8;
                                                                        }
                                                                        else if (Rs >= 991) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                            else if (ssDCval >= 83) {
                                                                if (ssDCval < 104) {
                                                                    return 4;
                                                                }
                                                                else if (ssDCval >= 104) {
                                                                    if (refDCval < 115) {
                                                                        return 8;
                                                                    }
                                                                    else if (refDCval >= 115) {
                                                                        if (Rs < 2131) {
                                                                            return 4;
                                                                        }
                                                                        else if (Rs >= 2131) {
                                                                            return 2;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (mcTcor >= 94) {
                                                    if (negBalance < 47) {
                                                        if (diffMVdiffVal < 4) {
                                                            if (TSC < 4) {
                                                                return 2;
                                                            }
                                                            else if (TSC >= 4) {
                                                                if (Rs < 735) {
                                                                    return 2;
                                                                }
                                                                else if (Rs >= 735) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                        else if (diffMVdiffVal >= 4) {
                                                            return 1;
                                                        }
                                                    }
                                                    else if (negBalance >= 47) {
                                                        return 2;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (Rs >= 3045) {
                                    if (ssDCval < 133) {
                                        if (TSC < 7) {
                                            if (mu_mv_mag_sq < 36) {
                                                if (MVDiff < 12) {
                                                    if (Cs < 1816) {
                                                        if (mu_mv_mag_sq < 3) {
                                                            if (refDCval < 106) {
                                                                return 4;
                                                            }
                                                            else if (refDCval >= 106) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (mu_mv_mag_sq >= 3) {
                                                            if (diffRsCsdiff < 27) {
                                                                return 4;
                                                            }
                                                            else if (diffRsCsdiff >= 27) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                    else if (Cs >= 1816) {
                                                        return 8;
                                                    }
                                                }
                                                else if (MVDiff >= 12) {
                                                    if (diffTSC < -1) {
                                                        if (AFD < 9) {
                                                            if (Rs < 3493) {
                                                                return 8;
                                                            }
                                                            else if (Rs >= 3493) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (AFD >= 9) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (diffTSC >= -1) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                            else if (mu_mv_mag_sq >= 36) {
                                                if (Cs < 3897) {
                                                    if (Rs < 3394) {
                                                        if (Rs < 3337) {
                                                            if (AFD < 8) {
                                                                return 8;
                                                            }
                                                            else if (AFD >= 8) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (Rs >= 3337) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (Rs >= 3394) {
                                                        if (Cs < 3631) {
                                                            if (diffRsCsdiff < -14) {
                                                                if (SC < 5439) {
                                                                    return 8;
                                                                }
                                                                else if (SC >= 5439) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (diffRsCsdiff >= -14) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (Cs >= 3631) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                                else if (Cs >= 3897) {
                                                    return 2;
                                                }
                                            }
                                        }
                                        else if (TSC >= 7) {
                                            if (RsDiff < 5) {
                                                if (Cs < 3479) {
                                                    if (Rs < 3328) {
                                                        if (Cs < 3182) {
                                                            if (Cs < 3095) {
                                                                return 8;
                                                            }
                                                            else if (Cs >= 3095) {
                                                                return 2;
                                                            }
                                                        }
                                                        else if (Cs >= 3182) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (Rs >= 3328) {
                                                        if (Cs < 3414) {
                                                            return 2;
                                                        }
                                                        else if (Cs >= 3414) {
                                                            if (ssDCval < 48) {
                                                                return 8;
                                                            }
                                                            else if (ssDCval >= 48) {
                                                                if (Cs < 3424) {
                                                                    return 8;
                                                                }
                                                                else if (Cs >= 3424) {
                                                                    if (Rs < 3518) {
                                                                        return 8;
                                                                    }
                                                                    else if (Rs >= 3518) {
                                                                        return 2;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (Cs >= 3479) {
                                                    if (diffRsCsdiff < -6) {
                                                        if (Cs < 3521) {
                                                            return 2;
                                                        }
                                                        else if (Cs >= 3521) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (diffRsCsdiff >= -6) {
                                                        if (posBalance < 54) {
                                                            return 4;
                                                        }
                                                        else if (posBalance >= 54) {
                                                            if (mu_mv_mag_sq < 33) {
                                                                if (MVDiff < 21) {
                                                                    return 2;
                                                                }
                                                                else if (MVDiff >= 21) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (mu_mv_mag_sq >= 33) {
                                                                if (mu_mv_mag_sq < 40) {
                                                                    if (Rs < 3695) {
                                                                        if (diffAFD < 1) {
                                                                            return 8;
                                                                        }
                                                                        else if (diffAFD >= 1) {
                                                                            return 2;
                                                                        }
                                                                    }
                                                                    else if (Rs >= 3695) {
                                                                        return 2;
                                                                    }
                                                                }
                                                                else if (mu_mv_mag_sq >= 40) {
                                                                    if (refDCval < 55) {
                                                                        return 4;
                                                                    }
                                                                    else if (refDCval >= 55) {
                                                                        return 2;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                            else if (RsDiff >= 5) {
                                                if (Cs < 2304) {
                                                    if (Cs < 1537) {
                                                        return 4;
                                                    }
                                                    else if (Cs >= 1537) {
                                                        if (MVDiff < 14) {
                                                            return 2;
                                                        }
                                                        else if (MVDiff >= 14) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                                else if (Cs >= 2304) {
                                                    if (SC < 6777) {
                                                        if (ssDCval < 102) {
                                                            if (CsDiff < 7) {
                                                                return 8;
                                                            }
                                                            else if (CsDiff >= 7) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (ssDCval >= 102) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (SC >= 6777) {
                                                        if (posBalance < 52) {
                                                            if (Rs < 4108) {
                                                                return 8;
                                                            }
                                                            else if (Rs >= 4108) {
                                                                if (MVDiff < 17) {
                                                                    return 4;
                                                                }
                                                                else if (MVDiff >= 17) {
                                                                    return 1;
                                                                }
                                                            }
                                                        }
                                                        else if (posBalance >= 52) {
                                                            if (AFD < 10) {
                                                                return 2;
                                                            }
                                                            else if (AFD >= 10) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (ssDCval >= 133) {
                                        if (Cs < 1792) {
                                            return 2;
                                        }
                                        else if (Cs >= 1792) {
                                            return 1;
                                        }
                                    }
                                }
                            }
                        }
                        else if (mu_mv_mag_sq >= 42) {
                            if (TSC < 10) {
                                if (mu_mv_mag_sq < 96) {
                                    if (ssDCval < 75) {
                                        if (Rs < 2181) {
                                            if (refDCval < 39) {
                                                if (SC < 813) {
                                                    if (posBalance < 44) {
                                                        if (Cs < 6) {
                                                            if (MVDiff < 48) {
                                                                return 8;
                                                            }
                                                            else if (MVDiff >= 48) {
                                                                if (MVDiff < 258) {
                                                                    return 1;
                                                                }
                                                                else if (MVDiff >= 258) {
                                                                    return 2;
                                                                }
                                                            }
                                                        }
                                                        else if (Cs >= 6) {
                                                            if (MVDiff < 16) {
                                                                return 4;
                                                            }
                                                            else if (MVDiff >= 16) {
                                                                return 2;
                                                            }
                                                        }
                                                    }
                                                    else if (posBalance >= 44) {
                                                        if (tcor < 85) {
                                                            if (AFD < 2) {
                                                                return 4;
                                                            }
                                                            else if (AFD >= 2) {
                                                                if (mcTcor < 81) {
                                                                    if (Rs < 394) {
                                                                        if (TSC < 4) {
                                                                            return 1;
                                                                        }
                                                                        else if (TSC >= 4) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                    else if (Rs >= 394) {
                                                                        if (tcor < 79) {
                                                                            if (ssDCval < 38) {
                                                                                if (Cs < 229) {
                                                                                    return 1;
                                                                                }
                                                                                else if (Cs >= 229) {
                                                                                    if (AFD < 5) {
                                                                                        return 8;
                                                                                    }
                                                                                    else if (AFD >= 5) {
                                                                                        return 4;
                                                                                    }
                                                                                }
                                                                            }
                                                                            else if (ssDCval >= 38) {
                                                                                return 1;
                                                                            }
                                                                        }
                                                                        else if (tcor >= 79) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                }
                                                                else if (mcTcor >= 81) {
                                                                    if (MVDiff < 28) {
                                                                        if (Rs < 395) {
                                                                            if (negBalance < 34) {
                                                                                if (posBalance < 108) {
                                                                                    return 4;
                                                                                }
                                                                                else if (posBalance >= 108) {
                                                                                    return 8;
                                                                                }
                                                                            }
                                                                            else if (negBalance >= 34) {
                                                                                return 8;
                                                                            }
                                                                        }
                                                                        else if (Rs >= 395) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                    else if (MVDiff >= 28) {
                                                                        if (RsCsDiff < 6) {
                                                                            if (Cs < 311) {
                                                                                return 4;
                                                                            }
                                                                            else if (Cs >= 311) {
                                                                                if (Rs < 436) {
                                                                                    return 1;
                                                                                }
                                                                                else if (Rs >= 436) {
                                                                                    return 4;
                                                                                }
                                                                            }
                                                                        }
                                                                        else if (RsCsDiff >= 6) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (tcor >= 85) {
                                                            if (ssDCval < 38) {
                                                                if (RsDiff < 1) {
                                                                    return 4;
                                                                }
                                                                else if (RsDiff >= 1) {
                                                                    return 1;
                                                                }
                                                            }
                                                            else if (ssDCval >= 38) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (SC >= 813) {
                                                    if (SC < 1153) {
                                                        if (Cs < 380) {
                                                            if (negBalance < 43) {
                                                                return 4;
                                                            }
                                                            else if (negBalance >= 43) {
                                                                if (Cs < 361) {
                                                                    return 2;
                                                                }
                                                                else if (Cs >= 361) {
                                                                    return 1;
                                                                }
                                                            }
                                                        }
                                                        else if (Cs >= 380) {
                                                            if (mu_mv_mag_sq < 68) {
                                                                if (Rs < 452) {
                                                                    if (Rs < 450) {
                                                                        return 2;
                                                                    }
                                                                    else if (Rs >= 450) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (Rs >= 452) {
                                                                    return 2;
                                                                }
                                                            }
                                                            else if (mu_mv_mag_sq >= 68) {
                                                                if (Rs < 491) {
                                                                    return 2;
                                                                }
                                                                else if (Rs >= 491) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (SC >= 1153) {
                                                        if (SC < 1758) {
                                                            if (Cs < 683) {
                                                                if (negBalance < 46) {
                                                                    return 4;
                                                                }
                                                                else if (negBalance >= 46) {
                                                                    if (tcor < 73) {
                                                                        return 1;
                                                                    }
                                                                    else if (tcor >= 73) {
                                                                        if (mcTcor < 81) {
                                                                            return 4;
                                                                        }
                                                                        else if (mcTcor >= 81) {
                                                                            if (mu_mv_mag_sq < 54) {
                                                                                return 1;
                                                                            }
                                                                            else if (mu_mv_mag_sq >= 54) {
                                                                                if (diffMVdiffVal < 13) {
                                                                                    return 4;
                                                                                }
                                                                                else if (diffMVdiffVal >= 13) {
                                                                                    return 1;
                                                                                }
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                            else if (Cs >= 683) {
                                                                if (TSC < 2) {
                                                                    return 1;
                                                                }
                                                                else if (TSC >= 2) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                        else if (SC >= 1758) {
                                                            if (Cs < 855) {
                                                                return 8;
                                                            }
                                                            else if (Cs >= 855) {
                                                                if (Rs < 1851) {
                                                                    return 2;
                                                                }
                                                                else if (Rs >= 1851) {
                                                                    if (SC < 4297) {
                                                                        return 1;
                                                                    }
                                                                    else if (SC >= 4297) {
                                                                        return 8;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                            else if (refDCval >= 39) {
                                                if (Rs < 1200) {
                                                    if (mu_mv_mag_sq < 82) {
                                                        if (Cs < 1044) {
                                                            if (posBalance < 49) {
                                                                if (negBalance < 45) {
                                                                    if (MVDiff < 8) {
                                                                        if (mu_mv_mag_sq < 60) {
                                                                            return 1;
                                                                        }
                                                                        else if (mu_mv_mag_sq >= 60) {
                                                                            if (ssDCval < 55) {
                                                                                return 4;
                                                                            }
                                                                            else if (ssDCval >= 55) {
                                                                                return 1;
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (MVDiff >= 8) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (negBalance >= 45) {
                                                                    if (AFD < 8) {
                                                                        return 8;
                                                                    }
                                                                    else if (AFD >= 8) {
                                                                        if (Rs < 537) {
                                                                            return 1;
                                                                        }
                                                                        else if (Rs >= 537) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                            else if (posBalance >= 49) {
                                                                if (mu_mv_mag_sq < 61) {
                                                                    if (negBalance < 38) {
                                                                        if (mcTcor < 92) {
                                                                            if (refDCval < 47) {
                                                                                return 8;
                                                                            }
                                                                            else if (refDCval >= 47) {
                                                                                if (TSC < 6) {
                                                                                    return 1;
                                                                                }
                                                                                else if (TSC >= 6) {
                                                                                    return 4;
                                                                                }
                                                                            }
                                                                        }
                                                                        else if (mcTcor >= 92) {
                                                                            if (Cs < 585) {
                                                                                return 8;
                                                                            }
                                                                            else if (Cs >= 585) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (negBalance >= 38) {
                                                                        if (refDCval < 59) {
                                                                            if (mu_mv_mag_sq < 43) {
                                                                                if (Rs < 1002) {
                                                                                    return 1;
                                                                                }
                                                                                else if (Rs >= 1002) {
                                                                                    return 4;
                                                                                }
                                                                            }
                                                                            else if (mu_mv_mag_sq >= 43) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                        else if (refDCval >= 59) {
                                                                            if (Cs < 802) {
                                                                                return 2;
                                                                            }
                                                                            else if (Cs >= 802) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                                else if (mu_mv_mag_sq >= 61) {
                                                                    if (ssDCval < 53) {
                                                                        return 4;
                                                                    }
                                                                    else if (ssDCval >= 53) {
                                                                        if (negBalance < 39) {
                                                                            return 4;
                                                                        }
                                                                        else if (negBalance >= 39) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (Cs >= 1044) {
                                                            return 2;
                                                        }
                                                    }
                                                    else if (mu_mv_mag_sq >= 82) {
                                                        return 8;
                                                    }
                                                }
                                                else if (Rs >= 1200) {
                                                    if (Rs < 1713) {
                                                        if (RsCsDiff < 29) {
                                                            if (ssDCval < 44) {
                                                                return 8;
                                                            }
                                                            else if (ssDCval >= 44) {
                                                                if (SC < 2307) {
                                                                    if (AFD < 6) {
                                                                        if (mcTcor < 94) {
                                                                            if (Cs < 944) {
                                                                                if (Cs < 869) {
                                                                                    return 1;
                                                                                }
                                                                                else if (Cs >= 869) {
                                                                                    return 4;
                                                                                }
                                                                            }
                                                                            else if (Cs >= 944) {
                                                                                return 1;
                                                                            }
                                                                        }
                                                                        else if (mcTcor >= 94) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                    else if (AFD >= 6) {
                                                                        return 1;
                                                                    }
                                                                }
                                                                else if (SC >= 2307) {
                                                                    if (SC < 2357) {
                                                                        if (Cs < 969) {
                                                                            return 4;
                                                                        }
                                                                        else if (Cs >= 969) {
                                                                            if (Rs < 1314) {
                                                                                if (Cs < 1043) {
                                                                                    return 4;
                                                                                }
                                                                                else if (Cs >= 1043) {
                                                                                    return 8;
                                                                                }
                                                                            }
                                                                            else if (Rs >= 1314) {
                                                                                return 1;
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (SC >= 2357) {
                                                                        if (mu_mv_mag_sq < 82) {
                                                                            if (tcor < 91) {
                                                                                if (MVDiff < 9) {
                                                                                    if (Cs < 1086) {
                                                                                        return 8;
                                                                                    }
                                                                                    else if (Cs >= 1086) {
                                                                                        return 4;
                                                                                    }
                                                                                }
                                                                                else if (MVDiff >= 9) {
                                                                                    return 1;
                                                                                }
                                                                            }
                                                                            else if (tcor >= 91) {
                                                                                if (ssDCval < 54) {
                                                                                    return 1;
                                                                                }
                                                                                else if (ssDCval >= 54) {
                                                                                    if (TSC < 6) {
                                                                                        if (Rs < 1482) {
                                                                                            return 8;
                                                                                        }
                                                                                        else if (Rs >= 1482) {
                                                                                            return 4;
                                                                                        }
                                                                                    }
                                                                                    else if (TSC >= 6) {
                                                                                        return 1;
                                                                                    }
                                                                                }
                                                                            }
                                                                        }
                                                                        else if (mu_mv_mag_sq >= 82) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (RsCsDiff >= 29) {
                                                            if (negBalance < 41) {
                                                                if (Rs < 1635) {
                                                                    return 4;
                                                                }
                                                                else if (Rs >= 1635) {
                                                                    return 1;
                                                                }
                                                            }
                                                            else if (negBalance >= 41) {
                                                                if (posBalance < 58) {
                                                                    return 8;
                                                                }
                                                                else if (posBalance >= 58) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (Rs >= 1713) {
                                                        if (AFD < 7) {
                                                            if (Cs < 1550) {
                                                                if (Cs < 1493) {
                                                                    if (SC < 3346) {
                                                                        if (RsDiff < 2) {
                                                                            if (Rs < 1834) {
                                                                                return 4;
                                                                            }
                                                                            else if (Rs >= 1834) {
                                                                                return 8;
                                                                            }
                                                                        }
                                                                        else if (RsDiff >= 2) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                    else if (SC >= 3346) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (Cs >= 1493) {
                                                                    if (mu_mv_mag_sq < 64) {
                                                                        if (tcor < 91) {
                                                                            return 4;
                                                                        }
                                                                        else if (tcor >= 91) {
                                                                            if (Rs < 1838) {
                                                                                if (SC < 3323) {
                                                                                    return 4;
                                                                                }
                                                                                else if (SC >= 3323) {
                                                                                    return 1;
                                                                                }
                                                                            }
                                                                            else if (Rs >= 1838) {
                                                                                if (MVDiff < 19) {
                                                                                    return 1;
                                                                                }
                                                                                else if (MVDiff >= 19) {
                                                                                    return 4;
                                                                                }
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (mu_mv_mag_sq >= 64) {
                                                                        return 4;
                                                                    }
                                                                }
                                                            }
                                                            else if (Cs >= 1550) {
                                                                if (mu_mv_mag_sq < 75) {
                                                                    if (Cs < 1565) {
                                                                        if (Rs < 1916) {
                                                                            return 1;
                                                                        }
                                                                        else if (Rs >= 1916) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                    else if (Cs >= 1565) {
                                                                        return 1;
                                                                    }
                                                                }
                                                                else if (mu_mv_mag_sq >= 75) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                        else if (AFD >= 7) {
                                                            if (mcTcor < 84) {
                                                                return 8;
                                                            }
                                                            else if (mcTcor >= 84) {
                                                                if (Rs < 1860) {
                                                                    if (RsDiff < 4) {
                                                                        if (SC < 2923) {
                                                                            return 1;
                                                                        }
                                                                        else if (SC >= 2923) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                    else if (RsDiff >= 4) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (Rs >= 1860) {
                                                                    return 1;
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (Rs >= 2181) {
                                            if (Cs < 3707) {
                                                if (SC < 3760) {
                                                    return 2;
                                                }
                                                else if (SC >= 3760) {
                                                    if (ssDCval < 52) {
                                                        if (mcTcor < 93) {
                                                            if (SC < 6487) {
                                                                return 8;
                                                            }
                                                            else if (SC >= 6487) {
                                                                return 2;
                                                            }
                                                        }
                                                        else if (mcTcor >= 93) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (ssDCval >= 52) {
                                                        if (RsDiff < 11) {
                                                            if (TSC < 9) {
                                                                if (tcor < 90) {
                                                                    if (refDCval < 64) {
                                                                        if (negBalance < 52) {
                                                                            return 8;
                                                                        }
                                                                        else if (negBalance >= 52) {
                                                                            if (Rs < 3823) {
                                                                                return 8;
                                                                            }
                                                                            else if (Rs >= 3823) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (refDCval >= 64) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (tcor >= 90) {
                                                                    if (diffRsCsdiff < 7) {
                                                                        if (TSC < 8) {
                                                                            if (posBalance < 54) {
                                                                                if (MVDiff < 8) {
                                                                                    return 8;
                                                                                }
                                                                                else if (MVDiff >= 8) {
                                                                                    return 4;
                                                                                }
                                                                            }
                                                                            else if (posBalance >= 54) {
                                                                                if (ssDCval < 53) {
                                                                                    return 4;
                                                                                }
                                                                                else if (ssDCval >= 53) {
                                                                                    if (diffTSC < 1) {
                                                                                        return 8;
                                                                                    }
                                                                                    else if (diffTSC >= 1) {
                                                                                        if (SC < 6895) {
                                                                                            return 8;
                                                                                        }
                                                                                        else if (SC >= 6895) {
                                                                                            return 4;
                                                                                        }
                                                                                    }
                                                                                }
                                                                            }
                                                                        }
                                                                        else if (TSC >= 8) {
                                                                            if (SC < 7363) {
                                                                                return 4;
                                                                            }
                                                                            else if (SC >= 7363) {
                                                                                return 2;
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (diffRsCsdiff >= 7) {
                                                                        return 4;
                                                                    }
                                                                }
                                                            }
                                                            else if (TSC >= 9) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (RsDiff >= 11) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (Cs >= 3707) {
                                                if (negBalance < 52) {
                                                    if (ssDCval < 50) {
                                                        if (Rs < 3909) {
                                                            return 4;
                                                        }
                                                        else if (Rs >= 3909) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (ssDCval >= 50) {
                                                        if (RsCsDiff < 22) {
                                                            if (SC < 7609) {
                                                                return 4;
                                                            }
                                                            else if (SC >= 7609) {
                                                                if (Rs < 3853) {
                                                                    return 2;
                                                                }
                                                                else if (Rs >= 3853) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                        else if (RsCsDiff >= 22) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                                else if (negBalance >= 52) {
                                                    if (RsCsDiff < 51) {
                                                        if (Rs < 3600) {
                                                            return 4;
                                                        }
                                                        else if (Rs >= 3600) {
                                                            return 2;
                                                        }
                                                    }
                                                    else if (RsCsDiff >= 51) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (ssDCval >= 75) {
                                        if (Rs < 265) {
                                            return 2;
                                        }
                                        else if (Rs >= 265) {
                                            if (Rs < 1776) {
                                                if (TSC < 7) {
                                                    if (posBalance < 23) {
                                                        if (Cs < 756) {
                                                            return 1;
                                                        }
                                                        else if (Cs >= 756) {
                                                            if (CsDiff < 2) {
                                                                return 1;
                                                            }
                                                            else if (CsDiff >= 2) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                    else if (posBalance >= 23) {
                                                        if (RsCsDiff < 29) {
                                                            if (SC < 815) {
                                                                if (Rs < 322) {
                                                                    return 2;
                                                                }
                                                                else if (Rs >= 322) {
                                                                    if (Rs < 415) {
                                                                        return 1;
                                                                    }
                                                                    else if (Rs >= 415) {
                                                                        return 8;
                                                                    }
                                                                }
                                                            }
                                                            else if (SC >= 815) {
                                                                if (SC < 3118) {
                                                                    return 4;
                                                                }
                                                                else if (SC >= 3118) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                        else if (RsCsDiff >= 29) {
                                                            if (ssDCval < 115) {
                                                                if (MVDiff < 12) {
                                                                    return 4;
                                                                }
                                                                else if (MVDiff >= 12) {
                                                                    return 2;
                                                                }
                                                            }
                                                            else if (ssDCval >= 115) {
                                                                if (posBalance < 52) {
                                                                    return 8;
                                                                }
                                                                else if (posBalance >= 52) {
                                                                    return 2;
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (TSC >= 7) {
                                                    if (SC < 766) {
                                                        if (ssDCval < 106) {
                                                            return 1;
                                                        }
                                                        else if (ssDCval >= 106) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (SC >= 766) {
                                                        if (refDCval < 103) {
                                                            if (refDCval < 98) {
                                                                if (tcor < 90) {
                                                                    return 8;
                                                                }
                                                                else if (tcor >= 90) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (refDCval >= 98) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (refDCval >= 103) {
                                                            if (ssDCval < 117) {
                                                                if (MVDiff < 24) {
                                                                    return 8;
                                                                }
                                                                else if (MVDiff >= 24) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (ssDCval >= 117) {
                                                                if (Cs < 2427) {
                                                                    return 4;
                                                                }
                                                                else if (Cs >= 2427) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                            else if (Rs >= 1776) {
                                                if (refDCval < 102) {
                                                    if (negBalance < 49) {
                                                        if (posBalance < 53) {
                                                            if (negBalance < 42) {
                                                                if (SC < 4584) {
                                                                    if (Scindex < 5) {
                                                                        return 8;
                                                                    }
                                                                    else if (Scindex >= 5) {
                                                                        if (mcTcor < 93) {
                                                                            return 8;
                                                                        }
                                                                        else if (mcTcor >= 93) {
                                                                            if (MVDiff < 8) {
                                                                                return 8;
                                                                            }
                                                                            else if (MVDiff >= 8) {
                                                                                return 1;
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                                else if (SC >= 4584) {
                                                                    if (RsCsDiff < 155) {
                                                                        return 8;
                                                                    }
                                                                    else if (RsCsDiff >= 155) {
                                                                        if (Rs < 2555) {
                                                                            return 8;
                                                                        }
                                                                        else if (Rs >= 2555) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                            else if (negBalance >= 42) {
                                                                if (Cs < 1748) {
                                                                    return 4;
                                                                }
                                                                else if (Cs >= 1748) {
                                                                    if (refDCval < 86) {
                                                                        return 8;
                                                                    }
                                                                    else if (refDCval >= 86) {
                                                                        return 1;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (posBalance >= 53) {
                                                            if (ssDCval < 87) {
                                                                if (negBalance < 21) {
                                                                    return 8;
                                                                }
                                                                else if (negBalance >= 21) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (ssDCval >= 87) {
                                                                if (Cs < 2298) {
                                                                    if (posBalance < 60) {
                                                                        return 1;
                                                                    }
                                                                    else if (posBalance >= 60) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (Cs >= 2298) {
                                                                    if (diffAFD < -6) {
                                                                        return 8;
                                                                    }
                                                                    else if (diffAFD >= -6) {
                                                                        return 2;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (negBalance >= 49) {
                                                        if (Rs < 2466) {
                                                            if (Rs < 1999) {
                                                                if (Cs < 1694) {
                                                                    if (mu_mv_mag_sq < 64) {
                                                                        return 2;
                                                                    }
                                                                    else if (mu_mv_mag_sq >= 64) {
                                                                        return 8;
                                                                    }
                                                                }
                                                                else if (Cs >= 1694) {
                                                                    if (negBalance < 54) {
                                                                        return 8;
                                                                    }
                                                                    else if (negBalance >= 54) {
                                                                        return 2;
                                                                    }
                                                                }
                                                            }
                                                            else if (Rs >= 1999) {
                                                                if (diffMVdiffVal < 3) {
                                                                    if (ssDCval < 83) {
                                                                        if (Cs < 1741) {
                                                                            return 2;
                                                                        }
                                                                        else if (Cs >= 1741) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                    else if (ssDCval >= 83) {
                                                                        return 2;
                                                                    }
                                                                }
                                                                else if (diffMVdiffVal >= 3) {
                                                                    if (Cs < 2029) {
                                                                        return 2;
                                                                    }
                                                                    else if (Cs >= 2029) {
                                                                        return 8;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (Rs >= 2466) {
                                                            if (Cs < 2086) {
                                                                if (RsCsDiff < 56) {
                                                                    return 2;
                                                                }
                                                                else if (RsCsDiff >= 56) {
                                                                    return 1;
                                                                }
                                                            }
                                                            else if (Cs >= 2086) {
                                                                if (ssDCval < 87) {
                                                                    return 4;
                                                                }
                                                                else if (ssDCval >= 87) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (refDCval >= 102) {
                                                    if (ssDCval < 125) {
                                                        if (Cs < 2497) {
                                                            if (CsDiff < 7) {
                                                                if (ssDCval < 107) {
                                                                    if (RsDiff < 8) {
                                                                        if (RsDiff < 6) {
                                                                            return 8;
                                                                        }
                                                                        else if (RsDiff >= 6) {
                                                                            if (Rs < 3211) {
                                                                                return 4;
                                                                            }
                                                                            else if (Rs >= 3211) {
                                                                                if (ssDCval < 105) {
                                                                                    if (Rs < 3916) {
                                                                                        return 4;
                                                                                    }
                                                                                    else if (Rs >= 3916) {
                                                                                        return 8;
                                                                                    }
                                                                                }
                                                                                else if (ssDCval >= 105) {
                                                                                    return 8;
                                                                                }
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (RsDiff >= 8) {
                                                                        if (negBalance < 35) {
                                                                            return 4;
                                                                        }
                                                                        else if (negBalance >= 35) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                }
                                                                else if (ssDCval >= 107) {
                                                                    if (mu_mv_mag_sq < 74) {
                                                                        if (Cs < 1050) {
                                                                            return 2;
                                                                        }
                                                                        else if (Cs >= 1050) {
                                                                            if (diffTSC < -1) {
                                                                                if (ssDCval < 117) {
                                                                                    return 4;
                                                                                }
                                                                                else if (ssDCval >= 117) {
                                                                                    if (Rs < 2485) {
                                                                                        return 4;
                                                                                    }
                                                                                    else if (Rs >= 2485) {
                                                                                        return 8;
                                                                                    }
                                                                                }
                                                                            }
                                                                            else if (diffTSC >= -1) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (mu_mv_mag_sq >= 74) {
                                                                        if (posBalance < 57) {
                                                                            if (Rs < 4700) {
                                                                                if (posBalance < 42) {
                                                                                    if (Cs < 2237) {
                                                                                        if (Cs < 2114) {
                                                                                            return 2;
                                                                                        }
                                                                                        else if (Cs >= 2114) {
                                                                                            return 8;
                                                                                        }
                                                                                    }
                                                                                    else if (Cs >= 2237) {
                                                                                        return 4;
                                                                                    }
                                                                                }
                                                                                else if (posBalance >= 42) {
                                                                                    return 4;
                                                                                }
                                                                            }
                                                                            else if (Rs >= 4700) {
                                                                                if (RsCsDiff < 80) {
                                                                                    return 4;
                                                                                }
                                                                                else if (RsCsDiff >= 80) {
                                                                                    return 8;
                                                                                }
                                                                            }
                                                                        }
                                                                        else if (posBalance >= 57) {
                                                                            if (SC < 4669) {
                                                                                if (RsDiff < 11) {
                                                                                    return 8;
                                                                                }
                                                                                else if (RsDiff >= 11) {
                                                                                    return 4;
                                                                                }
                                                                            }
                                                                            else if (SC >= 4669) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                            else if (CsDiff >= 7) {
                                                                if (MVDiff < 13) {
                                                                    if (ssDCval < 119) {
                                                                        if (refDCval < 114) {
                                                                            return 4;
                                                                        }
                                                                        else if (refDCval >= 114) {
                                                                            if (MVDiff < 8) {
                                                                                if (diffAFD < 1) {
                                                                                    return 8;
                                                                                }
                                                                                else if (diffAFD >= 1) {
                                                                                    return 4;
                                                                                }
                                                                            }
                                                                            else if (MVDiff >= 8) {
                                                                                if (TSC < 8) {
                                                                                    if (TSC < 7) {
                                                                                        return 2;
                                                                                    }
                                                                                    else if (TSC >= 7) {
                                                                                        return 8;
                                                                                    }
                                                                                }
                                                                                else if (TSC >= 8) {
                                                                                    return 4;
                                                                                }
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (ssDCval >= 119) {
                                                                        if (gchDC < 1) {
                                                                            return 4;
                                                                        }
                                                                        else if (gchDC >= 1) {
                                                                            if (Cs < 1961) {
                                                                                return 8;
                                                                            }
                                                                            else if (Cs >= 1961) {
                                                                                if (diffAFD < -1) {
                                                                                    return 8;
                                                                                }
                                                                                else if (diffAFD >= -1) {
                                                                                    return 4;
                                                                                }
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                                else if (MVDiff >= 13) {
                                                                    if (Rs < 4166) {
                                                                        if (negBalance < 30) {
                                                                            return 4;
                                                                        }
                                                                        else if (negBalance >= 30) {
                                                                            if (SC < 4639) {
                                                                                return 4;
                                                                            }
                                                                            else if (SC >= 4639) {
                                                                                if (mu_mv_mag_sq < 62) {
                                                                                    if (Rs < 3259) {
                                                                                        return 2;
                                                                                    }
                                                                                    else if (Rs >= 3259) {
                                                                                        return 4;
                                                                                    }
                                                                                }
                                                                                else if (mu_mv_mag_sq >= 62) {
                                                                                    if (Cs < 2194) {
                                                                                        if (posBalance < 42) {
                                                                                            if (gchDC < 2) {
                                                                                                return 4;
                                                                                            }
                                                                                            else if (gchDC >= 2) {
                                                                                                return 2;
                                                                                            }
                                                                                        }
                                                                                        else if (posBalance >= 42) {
                                                                                            return 2;
                                                                                        }
                                                                                    }
                                                                                    else if (Cs >= 2194) {
                                                                                        return 2;
                                                                                    }
                                                                                }
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (Rs >= 4166) {
                                                                        if (tcor < 93) {
                                                                            return 4;
                                                                        }
                                                                        else if (tcor >= 93) {
                                                                            if (Cs < 2297) {
                                                                                return 4;
                                                                            }
                                                                            else if (Cs >= 2297) {
                                                                                return 2;
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (Cs >= 2497) {
                                                            if (RsDiff < 11) {
                                                                if (Cs < 2862) {
                                                                    if (Rs < 2032) {
                                                                        return 4;
                                                                    }
                                                                    else if (Rs >= 2032) {
                                                                        return 8;
                                                                    }
                                                                }
                                                                else if (Cs >= 2862) {
                                                                    if (MVDiff < 10) {
                                                                        if (RsCsDiff < 99) {
                                                                            return 2;
                                                                        }
                                                                        else if (RsCsDiff >= 99) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                    else if (MVDiff >= 10) {
                                                                        if (AFD < 11) {
                                                                            return 8;
                                                                        }
                                                                        else if (AFD >= 11) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                            else if (RsDiff >= 11) {
                                                                if (AFD < 13) {
                                                                    return 4;
                                                                }
                                                                else if (AFD >= 13) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (ssDCval >= 125) {
                                                        if (SC < 6843) {
                                                            if (tcor < 82) {
                                                                return 1;
                                                            }
                                                            else if (tcor >= 82) {
                                                                if (MVDiff < 6) {
                                                                    if (Rs < 2556) {
                                                                        return 1;
                                                                    }
                                                                    else if (Rs >= 2556) {
                                                                        return 8;
                                                                    }
                                                                }
                                                                else if (MVDiff >= 6) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                        else if (SC >= 6843) {
                                                            if (refDCval < 128) {
                                                                if (posBalance < 39) {
                                                                    if (Rs < 4655) {
                                                                        return 4;
                                                                    }
                                                                    else if (Rs >= 4655) {
                                                                        return 2;
                                                                    }
                                                                }
                                                                else if (posBalance >= 39) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (refDCval >= 128) {
                                                                if (posBalance < 53) {
                                                                    return 2;
                                                                }
                                                                else if (posBalance >= 53) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (mu_mv_mag_sq >= 96) {
                                    if (posBalance < 53) {
                                        if (mu_mv_mag_sq < 107) {
                                            if (mcTcor < 78) {
                                                if (tcor < 52) {
                                                    if (MVDiff < 8) {
                                                        return 8;
                                                    }
                                                    else if (MVDiff >= 8) {
                                                        return 1;
                                                    }
                                                }
                                                else if (tcor >= 52) {
                                                    return 4;
                                                }
                                            }
                                            else if (mcTcor >= 78) {
                                                if (ssDCval < 128) {
                                                    return 8;
                                                }
                                                else if (ssDCval >= 128) {
                                                    if (refDCval < 142) {
                                                        if (RsCsDiff < 7) {
                                                            return 8;
                                                        }
                                                        else if (RsCsDiff >= 7) {
                                                            return 1;
                                                        }
                                                    }
                                                    else if (refDCval >= 142) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                        }
                                        else if (mu_mv_mag_sq >= 107) {
                                            if (tcor < 62) {
                                                if (diffAFD < -8) {
                                                    if (Cs < 801) {
                                                        return 1;
                                                    }
                                                    else if (Cs >= 801) {
                                                        return 4;
                                                    }
                                                }
                                                else if (diffAFD >= -8) {
                                                    if (MVDiff < 6) {
                                                        return 1;
                                                    }
                                                    else if (MVDiff >= 6) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                            else if (tcor >= 62) {
                                                return 8;
                                            }
                                        }
                                    }
                                    else if (posBalance >= 53) {
                                        if (tcor < 83) {
                                            if (diffTSC < -1) {
                                                return 8;
                                            }
                                            else if (diffTSC >= -1) {
                                                if (AFD < 7) {
                                                    return 4;
                                                }
                                                else if (AFD >= 7) {
                                                    return 2;
                                                }
                                            }
                                        }
                                        else if (tcor >= 83) {
                                            if (refDCval < 93) {
                                                if (mcTcor < 94) {
                                                    return 8;
                                                }
                                                else if (mcTcor >= 94) {
                                                    if (AFD < 7) {
                                                        if (RsCsDiff < 4) {
                                                            return 4;
                                                        }
                                                        else if (RsCsDiff >= 4) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (AFD >= 7) {
                                                        if (SC < 5574) {
                                                            return 8;
                                                        }
                                                        else if (SC >= 5574) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (refDCval >= 93) {
                                                if (mcTcor < 92) {
                                                    if (ssDCval < 113) {
                                                        return 2;
                                                    }
                                                    else if (ssDCval >= 113) {
                                                        if (Cs < 2464) {
                                                            return 8;
                                                        }
                                                        else if (Cs >= 2464) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                                else if (mcTcor >= 92) {
                                                    return 8;
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                            else if (TSC >= 10) {
                                if (CsDiff < 10) {
                                    if (Rs < 2030) {
                                        if (Rs < 1868) {
                                            return 8;
                                        }
                                        else if (Rs >= 1868) {
                                            if (Rs < 1880) {
                                                return 4;
                                            }
                                            else if (Rs >= 1880) {
                                                return 8;
                                            }
                                        }
                                    }
                                    else if (Rs >= 2030) {
                                        if (ssDCval < 118) {
                                            if (RsCsDiff < 163) {
                                                if (mu_mv_mag_sq < 72) {
                                                    if (mu_mv_mag_sq < 47) {
                                                        if (diffMVdiffVal < -1) {
                                                            if (CsDiff < 6) {
                                                                if (SC < 6316) {
                                                                    return 2;
                                                                }
                                                                else if (SC >= 6316) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (CsDiff >= 6) {
                                                                if (MVDiff < 11) {
                                                                    return 1;
                                                                }
                                                                else if (MVDiff >= 11) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                        else if (diffMVdiffVal >= -1) {
                                                            if (gchDC < 1) {
                                                                return 1;
                                                            }
                                                            else if (gchDC >= 1) {
                                                                if (AFD < 14) {
                                                                    return 1;
                                                                }
                                                                else if (AFD >= 14) {
                                                                    return 2;
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (mu_mv_mag_sq >= 47) {
                                                        if (Cs < 2267) {
                                                            if (diffAFD < -2) {
                                                                if (RsCsDiff < 89) {
                                                                    return 1;
                                                                }
                                                                else if (RsCsDiff >= 89) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (diffAFD >= -2) {
                                                                if (Rs < 2104) {
                                                                    return 4;
                                                                }
                                                                else if (Rs >= 2104) {
                                                                    if (diffAFD < 1) {
                                                                        return 1;
                                                                    }
                                                                    else if (diffAFD >= 1) {
                                                                        if (TSC < 12) {
                                                                            return 1;
                                                                        }
                                                                        else if (TSC >= 12) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (Cs >= 2267) {
                                                            if (SC < 5543) {
                                                                if (MVDiff < 12) {
                                                                    if (negBalance < 49) {
                                                                        if (SC < 5346) {
                                                                            return 1;
                                                                        }
                                                                        else if (SC >= 5346) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                    else if (negBalance >= 49) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (MVDiff >= 12) {
                                                                    if (Cs < 2579) {
                                                                        if (Rs < 2809) {
                                                                            return 4;
                                                                        }
                                                                        else if (Rs >= 2809) {
                                                                            if (CsDiff < 8) {
                                                                                return 1;
                                                                            }
                                                                            else if (CsDiff >= 8) {
                                                                                if (SC < 5373) {
                                                                                    return 4;
                                                                                }
                                                                                else if (SC >= 5373) {
                                                                                    return 1;
                                                                                }
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (Cs >= 2579) {
                                                                        return 4;
                                                                    }
                                                                }
                                                            }
                                                            else if (SC >= 5543) {
                                                                if (Rs < 4685) {
                                                                    if (Cs < 3178) {
                                                                        return 1;
                                                                    }
                                                                    else if (Cs >= 3178) {
                                                                        if (Rs < 3798) {
                                                                            return 8;
                                                                        }
                                                                        else if (Rs >= 3798) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                }
                                                                else if (Rs >= 4685) {
                                                                    if (CsDiff < 7) {
                                                                        return 2;
                                                                    }
                                                                    else if (CsDiff >= 7) {
                                                                        return 4;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (mu_mv_mag_sq >= 72) {
                                                    if (posBalance < 60) {
                                                        if (ssDCval < 86) {
                                                            if (refDCval < 86) {
                                                                if (RsCsDiff < 107) {
                                                                    return 8;
                                                                }
                                                                else if (RsCsDiff >= 107) {
                                                                    if (Cs < 2437) {
                                                                        return 4;
                                                                    }
                                                                    else if (Cs >= 2437) {
                                                                        return 8;
                                                                    }
                                                                }
                                                            }
                                                            else if (refDCval >= 86) {
                                                                return 2;
                                                            }
                                                        }
                                                        else if (ssDCval >= 86) {
                                                            if (ssDCval < 91) {
                                                                return 8;
                                                            }
                                                            else if (ssDCval >= 91) {
                                                                if (AFD < 12) {
                                                                    if (Rs < 3786) {
                                                                        return 8;
                                                                    }
                                                                    else if (Rs >= 3786) {
                                                                        return 2;
                                                                    }
                                                                }
                                                                else if (AFD >= 12) {
                                                                    return 1;
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (posBalance >= 60) {
                                                        return 1;
                                                    }
                                                }
                                            }
                                            else if (RsCsDiff >= 163) {
                                                if (refDCval < 105) {
                                                    if (refDCval < 84) {
                                                        if (Rs < 2755) {
                                                            return 4;
                                                        }
                                                        else if (Rs >= 2755) {
                                                            return 1;
                                                        }
                                                    }
                                                    else if (refDCval >= 84) {
                                                        return 4;
                                                    }
                                                }
                                                else if (refDCval >= 105) {
                                                    if (refDCval < 109) {
                                                        if (Cs < 3091) {
                                                            return 2;
                                                        }
                                                        else if (Cs >= 3091) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (refDCval >= 109) {
                                                        if (Rs < 3256) {
                                                            return 4;
                                                        }
                                                        else if (Rs >= 3256) {
                                                            if (CsDiff < 8) {
                                                                return 8;
                                                            }
                                                            else if (CsDiff >= 8) {
                                                                return 1;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (ssDCval >= 118) {
                                            if (tcor < 84) {
                                                if (posBalance < 50) {
                                                    return 8;
                                                }
                                                else if (posBalance >= 50) {
                                                    if (AFD < 14) {
                                                        return 1;
                                                    }
                                                    else if (AFD >= 14) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                            else if (tcor >= 84) {
                                                return 8;
                                            }
                                        }
                                    }
                                }
                                else if (CsDiff >= 10) {
                                    if (mcTcor < 87) {
                                        if (ssDCval < 108) {
                                            if (refDCval < 74) {
                                                if (MVDiff < 18) {
                                                    return 8;
                                                }
                                                else if (MVDiff >= 18) {
                                                    return 1;
                                                }
                                            }
                                            else if (refDCval >= 74) {
                                                return 4;
                                            }
                                        }
                                        else if (ssDCval >= 108) {
                                            if (SC < 4995) {
                                                return 8;
                                            }
                                            else if (SC >= 4995) {
                                                return 1;
                                            }
                                        }
                                    }
                                    else if (mcTcor >= 87) {
                                        if (negBalance < 47) {
                                            if (Cs < 3255) {
                                                if (Rs < 2976) {
                                                    return 1;
                                                }
                                                else if (Rs >= 2976) {
                                                    return 4;
                                                }
                                            }
                                            else if (Cs >= 3255) {
                                                return 8;
                                            }
                                        }
                                        else if (negBalance >= 47) {
                                            return 8;
                                        }
                                    }
                                }
                            }
                        }
                    }
                    else if (AFD >= 15) {
                        if (ssDCval < 88) {
                            if (negBalance < 46) {
                                if (ssDCval < 84) {
                                    if (MVDiff < 40) {
                                        if (tcor < 81) {
                                            if (diffAFD < 6) {
                                                return 1;
                                            }
                                            else if (diffAFD >= 6) {
                                                return 4;
                                            }
                                        }
                                        else if (tcor >= 81) {
                                            if (Scindex < 5) {
                                                return 1;
                                            }
                                            else if (Scindex >= 5) {
                                                if (Cs < 3121) {
                                                    if (Rs < 2356) {
                                                        if (Cs < 2397) {
                                                            return 8;
                                                        }
                                                        else if (Cs >= 2397) {
                                                            return 1;
                                                        }
                                                    }
                                                    else if (Rs >= 2356) {
                                                        if (SC < 4950) {
                                                            return 4;
                                                        }
                                                        else if (SC >= 4950) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                                else if (Cs >= 3121) {
                                                    return 1;
                                                }
                                            }
                                        }
                                    }
                                    else if (MVDiff >= 40) {
                                        if (diffAFD < 9) {
                                            if (Scindex < 5) {
                                                if (Cs < 1661) {
                                                    return 2;
                                                }
                                                else if (Cs >= 1661) {
                                                    return 1;
                                                }
                                            }
                                            else if (Scindex >= 5) {
                                                return 8;
                                            }
                                        }
                                        else if (diffAFD >= 9) {
                                            if (Cs < 628) {
                                                return 8;
                                            }
                                            else if (Cs >= 628) {
                                                return 1;
                                            }
                                        }
                                    }
                                }
                                else if (ssDCval >= 84) {
                                    if (posBalance < 55) {
                                        if (posBalance < 48) {
                                            return 8;
                                        }
                                        else if (posBalance >= 48) {
                                            return 1;
                                        }
                                    }
                                    else if (posBalance >= 55) {
                                        if (gchDC < 2) {
                                            if (SC < 4226) {
                                                return 2;
                                            }
                                            else if (SC >= 4226) {
                                                return 4;
                                            }
                                        }
                                        else if (gchDC >= 2) {
                                            if (Rs < 1891) {
                                                return 1;
                                            }
                                            else if (Rs >= 1891) {
                                                return 8;
                                            }
                                        }
                                    }
                                }
                            }
                            else if (negBalance >= 46) {
                                if (Cs < 1129) {
                                    if (TSC < 9) {
                                        if (gchDC < 9) {
                                            return 2;
                                        }
                                        else if (gchDC >= 9) {
                                            return 1;
                                        }
                                    }
                                    else if (TSC >= 9) {
                                        if (AFD < 18) {
                                            if (RsCsDiff < 11) {
                                                return 4;
                                            }
                                            else if (RsCsDiff >= 11) {
                                                if (Rs < 1394) {
                                                    return 8;
                                                }
                                                else if (Rs >= 1394) {
                                                    return 2;
                                                }
                                            }
                                        }
                                        else if (AFD >= 18) {
                                            if (RsCsDiff < 89) {
                                                if (diffRsCsdiff < 49) {
                                                    return 1;
                                                }
                                                else if (diffRsCsdiff >= 49) {
                                                    return 4;
                                                }
                                            }
                                            else if (RsCsDiff >= 89) {
                                                if (Rs < 1931) {
                                                    if (Rs < 408) {
                                                        return 4;
                                                    }
                                                    else if (Rs >= 408) {
                                                        if (refDCval < 58) {
                                                            return 4;
                                                        }
                                                        else if (refDCval >= 58) {
                                                            return 2;
                                                        }
                                                    }
                                                }
                                                else if (Rs >= 1931) {
                                                    return 8;
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (Cs >= 1129) {
                                    if (refDCval < 69) {
                                        if (gchDC < 2) {
                                            if (diffRsCsdiff < -98) {
                                                return 2;
                                            }
                                            else if (diffRsCsdiff >= -98) {
                                                return 8;
                                            }
                                        }
                                        else if (gchDC >= 2) {
                                            if (gchDC < 5) {
                                                if (posBalance < 57) {
                                                    return 4;
                                                }
                                                else if (posBalance >= 57) {
                                                    return 8;
                                                }
                                            }
                                            else if (gchDC >= 5) {
                                                if (SC < 4972) {
                                                    return 2;
                                                }
                                                else if (SC >= 4972) {
                                                    return 8;
                                                }
                                            }
                                        }
                                    }
                                    else if (refDCval >= 69) {
                                        if (ssDCval < 80) {
                                            if (MVDiff < 25) {
                                                if (posBalance < 58) {
                                                    if (Cs < 1869) {
                                                        return 4;
                                                    }
                                                    else if (Cs >= 1869) {
                                                        if (Cs < 2317) {
                                                            if (TSC < 10) {
                                                                return 4;
                                                            }
                                                            else if (TSC >= 10) {
                                                                if (MVDiff < 22) {
                                                                    return 8;
                                                                }
                                                                else if (MVDiff >= 22) {
                                                                    if (Rs < 2178) {
                                                                        return 8;
                                                                    }
                                                                    else if (Rs >= 2178) {
                                                                        if (SC < 4598) {
                                                                            return 4;
                                                                        }
                                                                        else if (SC >= 4598) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (Cs >= 2317) {
                                                            if (negBalance < 49) {
                                                                return 1;
                                                            }
                                                            else if (negBalance >= 49) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (posBalance >= 58) {
                                                    if (Cs < 1202) {
                                                        return 8;
                                                    }
                                                    else if (Cs >= 1202) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                            else if (MVDiff >= 25) {
                                                if (tcor < 48) {
                                                    return 4;
                                                }
                                                else if (tcor >= 48) {
                                                    if (RsCsDiff < 236) {
                                                        if (SC < 3579) {
                                                            if (CsDiff < 8) {
                                                                if (diffAFD < -4) {
                                                                    if (Rs < 1321) {
                                                                        return 1;
                                                                    }
                                                                    else if (Rs >= 1321) {
                                                                        return 2;
                                                                    }
                                                                }
                                                                else if (diffAFD >= -4) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (CsDiff >= 8) {
                                                                if (diffMVdiffVal < 5) {
                                                                    return 1;
                                                                }
                                                                else if (diffMVdiffVal >= 5) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                        else if (SC >= 3579) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (RsCsDiff >= 236) {
                                                        if (posBalance < 61) {
                                                            if (mu_mv_mag_sq < 11) {
                                                                if (Cs < 1904) {
                                                                    if (SC < 3783) {
                                                                        return 1;
                                                                    }
                                                                    else if (SC >= 3783) {
                                                                        return 2;
                                                                    }
                                                                }
                                                                else if (Cs >= 1904) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (mu_mv_mag_sq >= 11) {
                                                                if (diffAFD < 1) {
                                                                    if (refDCval < 74) {
                                                                        return 1;
                                                                    }
                                                                    else if (refDCval >= 74) {
                                                                        if (diffRsCsdiff < -62) {
                                                                            return 8;
                                                                        }
                                                                        else if (diffRsCsdiff >= -62) {
                                                                            if (MVDiff < 169) {
                                                                                return 1;
                                                                            }
                                                                            else if (MVDiff >= 169) {
                                                                                return 8;
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                                else if (diffAFD >= 1) {
                                                                    return 1;
                                                                }
                                                            }
                                                        }
                                                        else if (posBalance >= 61) {
                                                            return 2;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (ssDCval >= 80) {
                                            if (MVDiff < 12) {
                                                if (Cs < 2294) {
                                                    if (Rs < 2394) {
                                                        return 2;
                                                    }
                                                    else if (Rs >= 2394) {
                                                        return 1;
                                                    }
                                                }
                                                else if (Cs >= 2294) {
                                                    return 4;
                                                }
                                            }
                                            else if (MVDiff >= 12) {
                                                if (posBalance < 57) {
                                                    if (RsDiff < 10) {
                                                        if (Rs < 2508) {
                                                            if (Rs < 1432) {
                                                                return 1;
                                                            }
                                                            else if (Rs >= 1432) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (Rs >= 2508) {
                                                            if (Cs < 2158) {
                                                                return 8;
                                                            }
                                                            else if (Cs >= 2158) {
                                                                return 1;
                                                            }
                                                        }
                                                    }
                                                    else if (RsDiff >= 10) {
                                                        if (posBalance < 55) {
                                                            if (TSC < 13) {
                                                                if (CsDiff < 9) {
                                                                    return 8;
                                                                }
                                                                else if (CsDiff >= 9) {
                                                                    return 1;
                                                                }
                                                            }
                                                            else if (TSC >= 13) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (posBalance >= 55) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                                else if (posBalance >= 57) {
                                                    if (tcor < 45) {
                                                        return 4;
                                                    }
                                                    else if (tcor >= 45) {
                                                        if (TSC < 12) {
                                                            if (Rs < 2559) {
                                                                return 4;
                                                            }
                                                            else if (Rs >= 2559) {
                                                                if (Cs < 2138) {
                                                                    return 1;
                                                                }
                                                                else if (Cs >= 2138) {
                                                                    return 2;
                                                                }
                                                            }
                                                        }
                                                        else if (TSC >= 12) {
                                                            if (RsDiff < 14) {
                                                                return 1;
                                                            }
                                                            else if (RsDiff >= 14) {
                                                                if (RsDiff < 18) {
                                                                    return 4;
                                                                }
                                                                else if (RsDiff >= 18) {
                                                                    return 2;
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                        }
                        else if (ssDCval >= 88) {
                            if (CsDiff < 10) {
                                if (RsCsDiff < 149) {
                                    if (mu_mv_mag_sq < 10) {
                                        if (AFD < 18) {
                                            if (refDCval < 102) {
                                                if (posBalance < 52) {
                                                    if (TSC < 11) {
                                                        return 2;
                                                    }
                                                    else if (TSC >= 11) {
                                                        return 8;
                                                    }
                                                }
                                                else if (posBalance >= 52) {
                                                    return 4;
                                                }
                                            }
                                            else if (refDCval >= 102) {
                                                if (negBalance < 53) {
                                                    return 1;
                                                }
                                                else if (negBalance >= 53) {
                                                    if (mcTcor < 90) {
                                                        if (AFD < 16) {
                                                            return 2;
                                                        }
                                                        else if (AFD >= 16) {
                                                            if (Cs < 717) {
                                                                return 4;
                                                            }
                                                            else if (Cs >= 717) {
                                                                if (SC < 4520) {
                                                                    return 2;
                                                                }
                                                                else if (SC >= 4520) {
                                                                    return 1;
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (mcTcor >= 90) {
                                                        if (Cs < 1175) {
                                                            if (negBalance < 54) {
                                                                return 4;
                                                            }
                                                            else if (negBalance >= 54) {
                                                                return 2;
                                                            }
                                                        }
                                                        else if (Cs >= 1175) {
                                                            if (SC < 3871) {
                                                                return 4;
                                                            }
                                                            else if (SC >= 3871) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (AFD >= 18) {
                                            if (Rs < 2278) {
                                                if (negBalance < 60) {
                                                    if (Cs < 639) {
                                                        if (Rs < 905) {
                                                            return 2;
                                                        }
                                                        else if (Rs >= 905) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (Cs >= 639) {
                                                        if (AFD < 22) {
                                                            return 2;
                                                        }
                                                        else if (AFD >= 22) {
                                                            return 1;
                                                        }
                                                    }
                                                }
                                                else if (negBalance >= 60) {
                                                    return 2;
                                                }
                                            }
                                            else if (Rs >= 2278) {
                                                return 1;
                                            }
                                        }
                                    }
                                    else if (mu_mv_mag_sq >= 10) {
                                        if (MVDiff < 57) {
                                            if (TSC < 12) {
                                                if (MVDiff < 27) {
                                                    if (mcTcor < 95) {
                                                        if (Rs < 2852) {
                                                            return 2;
                                                        }
                                                        else if (Rs >= 2852) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (mcTcor >= 95) {
                                                        if (diffTSC < -1) {
                                                            if (Rs < 1937) {
                                                                return 2;
                                                            }
                                                            else if (Rs >= 1937) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (diffTSC >= -1) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                                else if (MVDiff >= 27) {
                                                    if (Rs < 1765) {
                                                        if (diffTSC < 2) {
                                                            return 2;
                                                        }
                                                        else if (diffTSC >= 2) {
                                                            return 1;
                                                        }
                                                    }
                                                    else if (Rs >= 1765) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                            else if (TSC >= 12) {
                                                if (MVDiff < 13) {
                                                    return 1;
                                                }
                                                else if (MVDiff >= 13) {
                                                    if (Cs < 2381) {
                                                        if (RsDiff < 10) {
                                                            if (CsDiff < 8) {
                                                                return 1;
                                                            }
                                                            else if (CsDiff >= 8) {
                                                                if (Rs < 2742) {
                                                                    return 2;
                                                                }
                                                                else if (Rs >= 2742) {
                                                                    return 1;
                                                                }
                                                            }
                                                        }
                                                        else if (RsDiff >= 10) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (Cs >= 2381) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                        }
                                        else if (MVDiff >= 57) {
                                            if (RsDiff < 6) {
                                                if (Rs < 1236) {
                                                    return 2;
                                                }
                                                else if (Rs >= 1236) {
                                                    return 4;
                                                }
                                            }
                                            else if (RsDiff >= 6) {
                                                if (diffAFD < 42) {
                                                    if (RsCsDiff < 85) {
                                                        if (TSC < 7) {
                                                            return 1;
                                                        }
                                                        else if (TSC >= 7) {
                                                            if (Rs < 442) {
                                                                return 1;
                                                            }
                                                            else if (Rs >= 442) {
                                                                return 2;
                                                            }
                                                        }
                                                    }
                                                    else if (RsCsDiff >= 85) {
                                                        return 1;
                                                    }
                                                }
                                                else if (diffAFD >= 42) {
                                                    return 2;
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (RsCsDiff >= 149) {
                                    if (Cs < 1411) {
                                        if (tcor < 67) {
                                            if (TSC < 10) {
                                                if (RsCsDiff < 261) {
                                                    return 1;
                                                }
                                                else if (RsCsDiff >= 261) {
                                                    return 8;
                                                }
                                            }
                                            else if (TSC >= 10) {
                                                if (AFD < 33) {
                                                    if (mcTcor < 78) {
                                                        if (Cs < 753) {
                                                            if (diffRsCsdiff < 6) {
                                                                return 1;
                                                            }
                                                            else if (diffRsCsdiff >= 6) {
                                                                return 2;
                                                            }
                                                        }
                                                        else if (Cs >= 753) {
                                                            return 2;
                                                        }
                                                    }
                                                    else if (mcTcor >= 78) {
                                                        if (mu_mv_mag_sq < 17) {
                                                            return 1;
                                                        }
                                                        else if (mu_mv_mag_sq >= 17) {
                                                            if (TSC < 13) {
                                                                if (tcor < 56) {
                                                                    return 1;
                                                                }
                                                                else if (tcor >= 56) {
                                                                    return 2;
                                                                }
                                                            }
                                                            else if (TSC >= 13) {
                                                                return 2;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (AFD >= 33) {
                                                    return 2;
                                                }
                                            }
                                        }
                                        else if (tcor >= 67) {
                                            if (Cs < 435) {
                                                return 4;
                                            }
                                            else if (Cs >= 435) {
                                                return 2;
                                            }
                                        }
                                    }
                                    else if (Cs >= 1411) {
                                        if (RsCsDiff < 268) {
                                            if (CsDiff < 8) {
                                                if (mu_mv_mag_sq < 15) {
                                                    if (diffRsCsdiff < 43) {
                                                        if (Rs < 2851) {
                                                            return 1;
                                                        }
                                                        else if (Rs >= 2851) {
                                                            return 2;
                                                        }
                                                    }
                                                    else if (diffRsCsdiff >= 43) {
                                                        return 4;
                                                    }
                                                }
                                                else if (mu_mv_mag_sq >= 15) {
                                                    if (Cs < 2266) {
                                                        return 8;
                                                    }
                                                    else if (Cs >= 2266) {
                                                        return 1;
                                                    }
                                                }
                                            }
                                            else if (CsDiff >= 8) {
                                                if (TSC < 14) {
                                                    if (mcTcor < 84) {
                                                        if (refDCval < 124) {
                                                            if (Rs < 4397) {
                                                                return 4;
                                                            }
                                                            else if (Rs >= 4397) {
                                                                if (Cs < 2892) {
                                                                    return 4;
                                                                }
                                                                else if (Cs >= 2892) {
                                                                    return 2;
                                                                }
                                                            }
                                                        }
                                                        else if (refDCval >= 124) {
                                                            return 1;
                                                        }
                                                    }
                                                    else if (mcTcor >= 84) {
                                                        if (mcTcor < 93) {
                                                            if (gchDC < 2) {
                                                                if (tcor < 74) {
                                                                    return 1;
                                                                }
                                                                else if (tcor >= 74) {
                                                                    if (RsCsDiff < 183) {
                                                                        if (refDCval < 91) {
                                                                            return 2;
                                                                        }
                                                                        else if (refDCval >= 91) {
                                                                            if (diffRsCsdiff < 18) {
                                                                                return 1;
                                                                            }
                                                                            else if (diffRsCsdiff >= 18) {
                                                                                if (TSC < 13) {
                                                                                    return 2;
                                                                                }
                                                                                else if (TSC >= 13) {
                                                                                    return 1;
                                                                                }
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (RsCsDiff >= 183) {
                                                                        return 2;
                                                                    }
                                                                }
                                                            }
                                                            else if (gchDC >= 2) {
                                                                if (ssDCval < 110) {
                                                                    if (Cs < 1787) {
                                                                        return 2;
                                                                    }
                                                                    else if (Cs >= 1787) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (ssDCval >= 110) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                        else if (mcTcor >= 93) {
                                                            if (diffRsCsdiff < -34) {
                                                                return 4;
                                                            }
                                                            else if (diffRsCsdiff >= -34) {
                                                                if (diffAFD < 6) {
                                                                    return 2;
                                                                }
                                                                else if (diffAFD >= 6) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (TSC >= 14) {
                                                    return 1;
                                                }
                                            }
                                        }
                                        else if (RsCsDiff >= 268) {
                                            if (Rs < 2828) {
                                                return 4;
                                            }
                                            else if (Rs >= 2828) {
                                                if (MVDiff < 12) {
                                                    if (posBalance < 50) {
                                                        return 4;
                                                    }
                                                    else if (posBalance >= 50) {
                                                        if (Rs < 3289) {
                                                            return 1;
                                                        }
                                                        else if (Rs >= 3289) {
                                                            return 2;
                                                        }
                                                    }
                                                }
                                                else if (MVDiff >= 12) {
                                                    if (posBalance < 58) {
                                                        if (refDCval < 156) {
                                                            if (Cs < 1740) {
                                                                return 2;
                                                            }
                                                            else if (Cs >= 1740) {
                                                                if (Rs < 3012) {
                                                                    return 8;
                                                                }
                                                                else if (Rs >= 3012) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                        else if (refDCval >= 156) {
                                                            return 2;
                                                        }
                                                    }
                                                    else if (posBalance >= 58) {
                                                        if (refDCval < 112) {
                                                            if (Cs < 1954) {
                                                                return 2;
                                                            }
                                                            else if (Cs >= 1954) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (refDCval >= 112) {
                                                            return 2;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                            else if (CsDiff >= 10) {
                                if (tcor < 69) {
                                    if (ssDCval < 116) {
                                        if (MVDiff < 52) {
                                            if (mcTcor < 85) {
                                                if (negBalance < 56) {
                                                    if (TSC < 15) {
                                                        if (diffTSC < 1) {
                                                            return 1;
                                                        }
                                                        else if (diffTSC >= 1) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (TSC >= 15) {
                                                        return 2;
                                                    }
                                                }
                                                else if (negBalance >= 56) {
                                                    if (MVDiff < 17) {
                                                        return 8;
                                                    }
                                                    else if (MVDiff >= 17) {
                                                        if (diffRsCsdiff < 494) {
                                                            return 2;
                                                        }
                                                        else if (diffRsCsdiff >= 494) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (mcTcor >= 85) {
                                                if (tcor < 65) {
                                                    return 2;
                                                }
                                                else if (tcor >= 65) {
                                                    if (AFD < 31) {
                                                        if (SC < 5167) {
                                                            return 1;
                                                        }
                                                        else if (SC >= 5167) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (AFD >= 31) {
                                                        return 2;
                                                    }
                                                }
                                            }
                                        }
                                        else if (MVDiff >= 52) {
                                            if (SC < 4925) {
                                                if (TSC < 24) {
                                                    return 1;
                                                }
                                                else if (TSC >= 24) {
                                                    return 4;
                                                }
                                            }
                                            else if (SC >= 4925) {
                                                return 8;
                                            }
                                        }
                                    }
                                    else if (ssDCval >= 116) {
                                        if (ssDCval < 140) {
                                            if (Rs < 1461) {
                                                if (Cs < 1396) {
                                                    if (gchDC < 30) {
                                                        return 1;
                                                    }
                                                    else if (gchDC >= 30) {
                                                        if (Rs < 879) {
                                                            return 4;
                                                        }
                                                        else if (Rs >= 879) {
                                                            return 2;
                                                        }
                                                    }
                                                }
                                                else if (Cs >= 1396) {
                                                    if (diffTSC < 2) {
                                                        if (SC < 2703) {
                                                            return 8;
                                                        }
                                                        else if (SC >= 2703) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (diffTSC >= 2) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                            else if (Rs >= 1461) {
                                                if (posBalance < 65) {
                                                    if (negBalance < 58) {
                                                        return 1;
                                                    }
                                                    else if (negBalance >= 58) {
                                                        if (RsDiff < 20) {
                                                            if (ssDCval < 126) {
                                                                return 1;
                                                            }
                                                            else if (ssDCval >= 126) {
                                                                if (RsDiff < 16) {
                                                                    return 1;
                                                                }
                                                                else if (RsDiff >= 16) {
                                                                    return 2;
                                                                }
                                                            }
                                                        }
                                                        else if (RsDiff >= 20) {
                                                            if (SC < 4945) {
                                                                return 4;
                                                            }
                                                            else if (SC >= 4945) {
                                                                return 2;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (posBalance >= 65) {
                                                    if (negBalance < 51) {
                                                        if (Rs < 2967) {
                                                            return 1;
                                                        }
                                                        else if (Rs >= 2967) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (negBalance >= 51) {
                                                        return 2;
                                                    }
                                                }
                                            }
                                        }
                                        else if (ssDCval >= 140) {
                                            if (diffAFD < 3) {
                                                if (MVDiff < 23) {
                                                    if (AFD < 34) {
                                                        return 2;
                                                    }
                                                    else if (AFD >= 34) {
                                                        return 4;
                                                    }
                                                }
                                                else if (MVDiff >= 23) {
                                                    if (Cs < 2509) {
                                                        return 4;
                                                    }
                                                    else if (Cs >= 2509) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                            else if (diffAFD >= 3) {
                                                if (ssDCval < 151) {
                                                    if (Rs < 819) {
                                                        if (Cs < 564) {
                                                            return 8;
                                                        }
                                                        else if (Cs >= 564) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (Rs >= 819) {
                                                        return 1;
                                                    }
                                                }
                                                else if (ssDCval >= 151) {
                                                    return 8;
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (tcor >= 69) {
                                    if (mcTcor < 94) {
                                        if (SC < 6294) {
                                            if (refDCval < 125) {
                                                if (mu_mv_mag_sq < 7) {
                                                    if (Rs < 2191) {
                                                        if (Cs < 1969) {
                                                            if (Rs < 1983) {
                                                                return 4;
                                                            }
                                                            else if (Rs >= 1983) {
                                                                return 2;
                                                            }
                                                        }
                                                        else if (Cs >= 1969) {
                                                            if (Rs < 2096) {
                                                                return 8;
                                                            }
                                                            else if (Rs >= 2096) {
                                                                return 1;
                                                            }
                                                        }
                                                    }
                                                    else if (Rs >= 2191) {
                                                        if (SC < 5036) {
                                                            if (diffAFD < 2) {
                                                                if (diffMVdiffVal < 10) {
                                                                    if (TSC < 10) {
                                                                        if (SC < 4831) {
                                                                            return 2;
                                                                        }
                                                                        else if (SC >= 4831) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                    else if (TSC >= 10) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (diffMVdiffVal >= 10) {
                                                                    return 1;
                                                                }
                                                            }
                                                            else if (diffAFD >= 2) {
                                                                if (Rs < 2482) {
                                                                    return 4;
                                                                }
                                                                else if (Rs >= 2482) {
                                                                    return 1;
                                                                }
                                                            }
                                                        }
                                                        else if (SC >= 5036) {
                                                            if (Cs < 2566) {
                                                                return 1;
                                                            }
                                                            else if (Cs >= 2566) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (mu_mv_mag_sq >= 7) {
                                                    if (Cs < 1863) {
                                                        if (AFD < 18) {
                                                            return 2;
                                                        }
                                                        else if (AFD >= 18) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (Cs >= 1863) {
                                                        if (mu_mv_mag_sq < 48) {
                                                            if (diffMVdiffVal < 2) {
                                                                return 1;
                                                            }
                                                            else if (diffMVdiffVal >= 2) {
                                                                if (posBalance < 60) {
                                                                    if (diffAFD < 2) {
                                                                        if (Cs < 2510) {
                                                                            return 1;
                                                                        }
                                                                        else if (Cs >= 2510) {
                                                                            if (SC < 5723) {
                                                                                return 4;
                                                                            }
                                                                            else if (SC >= 5723) {
                                                                                return 1;
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (diffAFD >= 2) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (posBalance >= 60) {
                                                                    return 1;
                                                                }
                                                            }
                                                        }
                                                        else if (mu_mv_mag_sq >= 48) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (refDCval >= 125) {
                                                if (tcor < 80) {
                                                    if (mcTcor < 84) {
                                                        if (posBalance < 54) {
                                                            return 1;
                                                        }
                                                        else if (posBalance >= 54) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (mcTcor >= 84) {
                                                        if (posBalance < 61) {
                                                            if (negBalance < 47) {
                                                                if (diffRsCsdiff < -30) {
                                                                    return 2;
                                                                }
                                                                else if (diffRsCsdiff >= -30) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (negBalance >= 47) {
                                                                if (CsDiff < 11) {
                                                                    if (RsCsDiff < 197) {
                                                                        return 1;
                                                                    }
                                                                    else if (RsCsDiff >= 197) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (CsDiff >= 11) {
                                                                    return 1;
                                                                }
                                                            }
                                                        }
                                                        else if (posBalance >= 61) {
                                                            if (Cs < 2276) {
                                                                return 2;
                                                            }
                                                            else if (Cs >= 2276) {
                                                                return 1;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (tcor >= 80) {
                                                    if (ssDCval < 126) {
                                                        return 4;
                                                    }
                                                    else if (ssDCval >= 126) {
                                                        if (Cs < 2032) {
                                                            return 2;
                                                        }
                                                        else if (Cs >= 2032) {
                                                            return 1;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (SC >= 6294) {
                                            if (negBalance < 54) {
                                                if (mu_mv_mag_sq < 41) {
                                                    if (SC < 7699) {
                                                        if (Cs < 2102) {
                                                            return 4;
                                                        }
                                                        else if (Cs >= 2102) {
                                                            if (RsCsDiff < 243) {
                                                                return 4;
                                                            }
                                                            else if (RsCsDiff >= 243) {
                                                                return 1;
                                                            }
                                                        }
                                                    }
                                                    else if (SC >= 7699) {
                                                        return 4;
                                                    }
                                                }
                                                else if (mu_mv_mag_sq >= 41) {
                                                    if (mcTcor < 84) {
                                                        if (SC < 7610) {
                                                            return 1;
                                                        }
                                                        else if (SC >= 7610) {
                                                            return 2;
                                                        }
                                                    }
                                                    else if (mcTcor >= 84) {
                                                        if (CsDiff < 11) {
                                                            if (SC < 6956) {
                                                                return 2;
                                                            }
                                                            else if (SC >= 6956) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (CsDiff >= 11) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (negBalance >= 54) {
                                                if (MVDiff < 12) {
                                                    if (refDCval < 105) {
                                                        return 4;
                                                    }
                                                    else if (refDCval >= 105) {
                                                        return 8;
                                                    }
                                                }
                                                else if (MVDiff >= 12) {
                                                    if (posBalance < 62) {
                                                        if (Cs < 3352) {
                                                            if (AFD < 18) {
                                                                return 8;
                                                            }
                                                            else if (AFD >= 18) {
                                                                if (Rs < 5097) {
                                                                    return 4;
                                                                }
                                                                else if (Rs >= 5097) {
                                                                    if (Cs < 2725) {
                                                                        return 2;
                                                                    }
                                                                    else if (Cs >= 2725) {
                                                                        return 4;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (Cs >= 3352) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (posBalance >= 62) {
                                                        if (AFD < 22) {
                                                            return 2;
                                                        }
                                                        else if (AFD >= 22) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (mcTcor >= 94) {
                                        if (diffRsCsdiff < -95) {
                                            if (Rs < 2937) {
                                                return 2;
                                            }
                                            else if (Rs >= 2937) {
                                                if (Cs < 2722) {
                                                    return 4;
                                                }
                                                else if (Cs >= 2722) {
                                                    return 8;
                                                }
                                            }
                                        }
                                        else if (diffRsCsdiff >= -95) {
                                            return 2;
                                        }
                                    }
                                }
                            }
                        }
                    }
                }
            }
            else if (SC >= 8102) {
                if (ssDCval < 64) {
                    if (ssDCval < 52) {
                        if (Cs < 4574) {
                            if (Cs < 4281) {
                                return 8;
                            }
                            else if (Cs >= 4281) {
                                return 1;
                            }
                        }
                        else if (Cs >= 4574) {
                            return 8;
                        }
                    }
                    else if (ssDCval >= 52) {
                        if (diffTSC < -1) {
                            if (RsDiff < 5) {
                                return 2;
                            }
                            else if (RsDiff >= 5) {
                                return 4;
                            }
                        }
                        else if (diffTSC >= -1) {
                            if (tcor < 74) {
                                return 8;
                            }
                            else if (tcor >= 74) {
                                if (TSC < 7) {
                                    return 8;
                                }
                                else if (TSC >= 7) {
                                    return 4;
                                }
                            }
                        }
                    }
                }
                else if (ssDCval >= 64) {
                    if (mu_mv_mag_sq < 128) {
                        if (Cs < 3293) {
                            if (Rs < 5763) {
                                if (CsDiff < 8) {
                                    if (Cs < 3029) {
                                        return 4;
                                    }
                                    else if (Cs >= 3029) {
                                        if (Rs < 5116) {
                                            if (SC < 8116) {
                                                return 8;
                                            }
                                            else if (SC >= 8116) {
                                                return 2;
                                            }
                                        }
                                        else if (Rs >= 5116) {
                                            if (SC < 8457) {
                                                return 8;
                                            }
                                            else if (SC >= 8457) {
                                                return 2;
                                            }
                                        }
                                    }
                                }
                                else if (CsDiff >= 8) {
                                    if (refDCval < 104) {
                                        return 4;
                                    }
                                    else if (refDCval >= 104) {
                                        if (Cs < 3115) {
                                            if (mu_mv_mag_sq < 47) {
                                                if (Cs < 2843) {
                                                    return 4;
                                                }
                                                else if (Cs >= 2843) {
                                                    return 2;
                                                }
                                            }
                                            else if (mu_mv_mag_sq >= 47) {
                                                return 4;
                                            }
                                        }
                                        else if (Cs >= 3115) {
                                            return 2;
                                        }
                                    }
                                }
                            }
                            else if (Rs >= 5763) {
                                if (mcTcor < 86) {
                                    if (ssDCval < 116) {
                                        return 4;
                                    }
                                    else if (ssDCval >= 116) {
                                        return 1;
                                    }
                                }
                                else if (mcTcor >= 86) {
                                    if (mcTcor < 89) {
                                        if (Rs < 7007) {
                                            return 1;
                                        }
                                        else if (Rs >= 7007) {
                                            return 4;
                                        }
                                    }
                                    else if (mcTcor >= 89) {
                                        return 4;
                                    }
                                }
                            }
                        }
                        else if (Cs >= 3293) {
                            if (mu_mv_mag_sq < 61) {
                                if (refDCval < 149) {
                                    if (Rs < 6988) {
                                        if (CsDiff < 17) {
                                            if (diffRsCsdiff < 25) {
                                                if (Cs < 4313) {
                                                    if (tcor < 89) {
                                                        if (RsDiff < 20) {
                                                            if (Rs < 5464) {
                                                                if (mcTcor < 88) {
                                                                    if (Cs < 3382) {
                                                                        return 8;
                                                                    }
                                                                    else if (Cs >= 3382) {
                                                                        if (MVDiff < 27) {
                                                                            return 4;
                                                                        }
                                                                        else if (MVDiff >= 27) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                }
                                                                else if (mcTcor >= 88) {
                                                                    if (SC < 8169) {
                                                                        return 4;
                                                                    }
                                                                    else if (SC >= 8169) {
                                                                        return 8;
                                                                    }
                                                                }
                                                            }
                                                            else if (Rs >= 5464) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (RsDiff >= 20) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (tcor >= 89) {
                                                        return 4;
                                                    }
                                                }
                                                else if (Cs >= 4313) {
                                                    if (refDCval < 69) {
                                                        return 4;
                                                    }
                                                    else if (refDCval >= 69) {
                                                        if (diffMVdiffVal < 3) {
                                                            return 8;
                                                        }
                                                        else if (diffMVdiffVal >= 3) {
                                                            if (SC < 10621) {
                                                                return 8;
                                                            }
                                                            else if (SC >= 10621) {
                                                                if (AFD < 16) {
                                                                    return 4;
                                                                }
                                                                else if (AFD >= 16) {
                                                                    return 1;
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                            else if (diffRsCsdiff >= 25) {
                                                if (mcTcor < 91) {
                                                    if (Rs < 4836) {
                                                        if (Cs < 3874) {
                                                            return 8;
                                                        }
                                                        else if (Cs >= 3874) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (Rs >= 4836) {
                                                        if (RsCsDiff < 611) {
                                                            return 8;
                                                        }
                                                        else if (RsCsDiff >= 611) {
                                                            if (Cs < 3957) {
                                                                return 4;
                                                            }
                                                            else if (Cs >= 3957) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (mcTcor >= 91) {
                                                    return 8;
                                                }
                                            }
                                        }
                                        else if (CsDiff >= 17) {
                                            if (mcTcor < 93) {
                                                if (refDCval < 137) {
                                                    if (RsDiff < 30) {
                                                        if (SC < 12159) {
                                                            return 8;
                                                        }
                                                        else if (SC >= 12159) {
                                                            if (Cs < 5642) {
                                                                return 4;
                                                            }
                                                            else if (Cs >= 5642) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                    else if (RsDiff >= 30) {
                                                        if (TSC < 20) {
                                                            return 2;
                                                        }
                                                        else if (TSC >= 20) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                                else if (refDCval >= 137) {
                                                    if (mcTcor < 75) {
                                                        if (Rs < 4773) {
                                                            return 1;
                                                        }
                                                        else if (Rs >= 4773) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (mcTcor >= 75) {
                                                        if (negBalance < 62) {
                                                            if (SC < 11378) {
                                                                return 8;
                                                            }
                                                            else if (SC >= 11378) {
                                                                if (ssDCval < 140) {
                                                                    return 8;
                                                                }
                                                                else if (ssDCval >= 140) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                        else if (negBalance >= 62) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (mcTcor >= 93) {
                                                if (refDCval < 116) {
                                                    if (Cs < 4386) {
                                                        return 4;
                                                    }
                                                    else if (Cs >= 4386) {
                                                        return 1;
                                                    }
                                                }
                                                else if (refDCval >= 116) {
                                                    return 8;
                                                }
                                            }
                                        }
                                    }
                                    else if (Rs >= 6988) {
                                        if (ssDCval < 77) {
                                            return 4;
                                        }
                                        else if (ssDCval >= 77) {
                                            if (MVDiff < 26) {
                                                if (Rs < 14727) {
                                                    if (SC < 17742) {
                                                        if (ssDCval < 103) {
                                                            if (tcor < 76) {
                                                                if (RsDiff < 22) {
                                                                    return 8;
                                                                }
                                                                else if (RsDiff >= 22) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (tcor >= 76) {
                                                                if (Cs < 5803) {
                                                                    return 4;
                                                                }
                                                                else if (Cs >= 5803) {
                                                                    return 2;
                                                                }
                                                            }
                                                        }
                                                        else if (ssDCval >= 103) {
                                                            if (Cs < 6107) {
                                                                return 4;
                                                            }
                                                            else if (Cs >= 6107) {
                                                                if (diffMVdiffVal < 4) {
                                                                    return 8;
                                                                }
                                                                else if (diffMVdiffVal >= 4) {
                                                                    if (mcTcor < 87) {
                                                                        return 8;
                                                                    }
                                                                    else if (mcTcor >= 87) {
                                                                        if (Cs < 6557) {
                                                                            return 4;
                                                                        }
                                                                        else if (Cs >= 6557) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (SC >= 17742) {
                                                        if (refDCval < 116) {
                                                            if (SC < 25977) {
                                                                return 8;
                                                            }
                                                            else if (SC >= 25977) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (refDCval >= 116) {
                                                            if (RsDiff < 56) {
                                                                return 4;
                                                            }
                                                            else if (RsDiff >= 56) {
                                                                return 2;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (Rs >= 14727) {
                                                    if (RsDiff < 66) {
                                                        return 8;
                                                    }
                                                    else if (RsDiff >= 66) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                            else if (MVDiff >= 26) {
                                                if (AFD < 56) {
                                                    if (Cs < 6354) {
                                                        return 2;
                                                    }
                                                    else if (Cs >= 6354) {
                                                        return 8;
                                                    }
                                                }
                                                else if (AFD >= 56) {
                                                    return 8;
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (refDCval >= 149) {
                                    if (Rs < 7690) {
                                        if (MVDiff < 30) {
                                            if (posBalance < 56) {
                                                if (Rs < 6633) {
                                                    return 8;
                                                }
                                                else if (Rs >= 6633) {
                                                    return 4;
                                                }
                                            }
                                            else if (posBalance >= 56) {
                                                return 2;
                                            }
                                        }
                                        else if (MVDiff >= 30) {
                                            if (AFD < 23) {
                                                if (RsDiff < 19) {
                                                    return 2;
                                                }
                                                else if (RsDiff >= 19) {
                                                    return 8;
                                                }
                                            }
                                            else if (AFD >= 23) {
                                                if (Cs < 4719) {
                                                    return 4;
                                                }
                                                else if (Cs >= 4719) {
                                                    return 2;
                                                }
                                            }
                                        }
                                    }
                                    else if (Rs >= 7690) {
                                        if (mcTcor < 87) {
                                            if (TSC < 23) {
                                                if (Rs < 10034) {
                                                    return 8;
                                                }
                                                else if (Rs >= 10034) {
                                                    return 2;
                                                }
                                            }
                                            else if (TSC >= 23) {
                                                if (SC < 19200) {
                                                    return 4;
                                                }
                                                else if (SC >= 19200) {
                                                    return 2;
                                                }
                                            }
                                        }
                                        else if (mcTcor >= 87) {
                                            if (MVDiff < 28) {
                                                return 4;
                                            }
                                            else if (MVDiff >= 28) {
                                                return 2;
                                            }
                                        }
                                    }
                                }
                            }
                            else if (mu_mv_mag_sq >= 61) {
                                if (MVDiff < 2) {
                                    if (negBalance < 63) {
                                        if (ssDCval < 106) {
                                            if (Scindex < 7) {
                                                if (Cs < 5220) {
                                                    if (ssDCval < 105) {
                                                        if (mu_mv_mag_sq < 127) {
                                                            return 8;
                                                        }
                                                        else if (mu_mv_mag_sq >= 127) {
                                                            if (negBalance < 46) {
                                                                return 8;
                                                            }
                                                            else if (negBalance >= 46) {
                                                                if (AFD < 7) {
                                                                    if (SC < 9871) {
                                                                        return 8;
                                                                    }
                                                                    else if (SC >= 9871) {
                                                                        if (refDCval < 80) {
                                                                            return 8;
                                                                        }
                                                                        else if (refDCval >= 80) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                }
                                                                else if (AFD >= 7) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (ssDCval >= 105) {
                                                        if (RsCsDiff < 302) {
                                                            return 8;
                                                        }
                                                        else if (RsCsDiff >= 302) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                                else if (Cs >= 5220) {
                                                    if (Rs < 5270) {
                                                        return 4;
                                                    }
                                                    else if (Rs >= 5270) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                            else if (Scindex >= 7) {
                                                return 8;
                                            }
                                        }
                                        else if (ssDCval >= 106) {
                                            if (refDCval < 120) {
                                                if (MVDiff < 1) {
                                                    if (posBalance < 54) {
                                                        return 8;
                                                    }
                                                    else if (posBalance >= 54) {
                                                        if (SC < 13254) {
                                                            if (refDCval < 106) {
                                                                return 8;
                                                            }
                                                            else if (refDCval >= 106) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (SC >= 13254) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                                else if (MVDiff >= 1) {
                                                    if (Rs < 5039) {
                                                        if (Rs < 4804) {
                                                            return 4;
                                                        }
                                                        else if (Rs >= 4804) {
                                                            return 2;
                                                        }
                                                    }
                                                    else if (Rs >= 5039) {
                                                        if (Rs < 10734) {
                                                            return 8;
                                                        }
                                                        else if (Rs >= 10734) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (refDCval >= 120) {
                                                return 8;
                                            }
                                        }
                                    }
                                    else if (negBalance >= 63) {
                                        if (SC < 15713) {
                                            if (Cs < 7114) {
                                                return 8;
                                            }
                                            else if (Cs >= 7114) {
                                                return 4;
                                            }
                                        }
                                        else if (SC >= 15713) {
                                            if (RsCsDiff < 1059) {
                                                return 2;
                                            }
                                            else if (RsCsDiff >= 1059) {
                                                if (Rs < 8347) {
                                                    return 4;
                                                }
                                                else if (Rs >= 8347) {
                                                    return 2;
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (MVDiff >= 2) {
                                    if (mcTcor < 70) {
                                        if (ssDCval < 136) {
                                            return 8;
                                        }
                                        else if (ssDCval >= 136) {
                                            return 1;
                                        }
                                    }
                                    else if (mcTcor >= 70) {
                                        if (Cs < 3905) {
                                            if (diffAFD < -1) {
                                                if (ssDCval < 121) {
                                                    if (RsDiff < 9) {
                                                        return 8;
                                                    }
                                                    else if (RsDiff >= 9) {
                                                        return 2;
                                                    }
                                                }
                                                else if (ssDCval >= 121) {
                                                    return 8;
                                                }
                                            }
                                            else if (diffAFD >= -1) {
                                                return 8;
                                            }
                                        }
                                        else if (Cs >= 3905) {
                                            if (TSC < 28) {
                                                if (mu_mv_mag_sq < 72) {
                                                    if (AFD < 16) {
                                                        if (tcor < 88) {
                                                            return 4;
                                                        }
                                                        else if (tcor >= 88) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (AFD >= 16) {
                                                        return 8;
                                                    }
                                                }
                                                else if (mu_mv_mag_sq >= 72) {
                                                    return 8;
                                                }
                                            }
                                            else if (TSC >= 28) {
                                                return 4;
                                            }
                                        }
                                    }
                                }
                            }
                        }
                    }
                    else if (mu_mv_mag_sq >= 128) {
                        if (refDCval < 105) {
                            if (Rs < 5258) {
                                if (ssDCval < 80) {
                                    if (tcor < 93) {
                                        return 8;
                                    }
                                    else if (tcor >= 93) {
                                        if (Cs < 4919) {
                                            return 8;
                                        }
                                        else if (Cs >= 4919) {
                                            return 4;
                                        }
                                    }
                                }
                                else if (ssDCval >= 80) {
                                    if (Rs < 4824) {
                                        if (AFD < 7) {
                                            if (mcTcor < 92) {
                                                if (RsDiff < 8) {
                                                    return 4;
                                                }
                                                else if (RsDiff >= 8) {
                                                    if (SC < 9787) {
                                                        return 8;
                                                    }
                                                    else if (SC >= 9787) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                            else if (mcTcor >= 92) {
                                                return 4;
                                            }
                                        }
                                        else if (AFD >= 7) {
                                            return 8;
                                        }
                                    }
                                    else if (Rs >= 4824) {
                                        if (SC < 10085) {
                                            if (CsDiff < 9) {
                                                if (CsDiff < 8) {
                                                    return 8;
                                                }
                                                else if (CsDiff >= 8) {
                                                    return 4;
                                                }
                                            }
                                            else if (CsDiff >= 9) {
                                                if (AFD < 7) {
                                                    return 8;
                                                }
                                                else if (AFD >= 7) {
                                                    if (posBalance < 59) {
                                                        return 4;
                                                    }
                                                    else if (posBalance >= 59) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                        }
                                        else if (SC >= 10085) {
                                            return 4;
                                        }
                                    }
                                }
                            }
                            else if (Rs >= 5258) {
                                if (Cs < 7359) {
                                    return 8;
                                }
                                else if (Cs >= 7359) {
                                    return 2;
                                }
                            }
                        }
                        else if (refDCval >= 105) {
                            return 4;
                        }
                    }
                }
            }
        }
        else if (tcor >= 95) {
            if (Cs < 708) {
                if (MVDiff < 5) {
                    if (refDCval < 160) {
                        if (refDCval < 92) {
                            if (refDCval < 91) {
                                if (Rs < 1041) {
                                    if (SC < 991) {
                                        if (Cs < 600) {
                                            if (TSCindex < 2) {
                                                if (Cs < 309) {
                                                    return 8;
                                                }
                                                else if (Cs >= 309) {
                                                    if (mu_mv_mag_sq < 128) {
                                                        return 8;
                                                    }
                                                    else if (mu_mv_mag_sq >= 128) {
                                                        if (negBalance < 13) {
                                                            return 8;
                                                        }
                                                        else if (negBalance >= 13) {
                                                            if (Rs < 185) {
                                                                if (SC < 490) {
                                                                    return 4;
                                                                }
                                                                else if (SC >= 490) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (Rs >= 185) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                            else if (TSCindex >= 2) {
                                                if (ssDCval < 74) {
                                                    if (refDCval < 62) {
                                                        if (RsDiff < 1) {
                                                            return 1;
                                                        }
                                                        else if (RsDiff >= 1) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (refDCval >= 62) {
                                                        return 8;
                                                    }
                                                }
                                                else if (ssDCval >= 74) {
                                                    if (MVDiff < 2) {
                                                        return 4;
                                                    }
                                                    else if (MVDiff >= 2) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                        }
                                        else if (Cs >= 600) {
                                            return 4;
                                        }
                                    }
                                    else if (SC >= 991) {
                                        if (AFD < 4) {
                                            if (Rs < 727) {
                                                if (TSC < 1) {
                                                    return 4;
                                                }
                                                else if (TSC >= 1) {
                                                    if (RsDiff < 1) {
                                                        return 8;
                                                    }
                                                    else if (RsDiff >= 1) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                            else if (Rs >= 727) {
                                                return 8;
                                            }
                                        }
                                        else if (AFD >= 4) {
                                            if (ssDCval < 50) {
                                                return 4;
                                            }
                                            else if (ssDCval >= 50) {
                                                return 8;
                                            }
                                        }
                                    }
                                }
                                else if (Rs >= 1041) {
                                    if (Rs < 1193) {
                                        return 4;
                                    }
                                    else if (Rs >= 1193) {
                                        if (Cs < 612) {
                                            return 2;
                                        }
                                        else if (Cs >= 612) {
                                            return 8;
                                        }
                                    }
                                }
                            }
                            else if (refDCval >= 91) {
                                return 4;
                            }
                        }
                        else if (refDCval >= 92) {
                            if (negBalance < 27) {
                                if (Cs < 11) {
                                    if (Rs < 857) {
                                        return 1;
                                    }
                                    else if (Rs >= 857) {
                                        return 8;
                                    }
                                }
                                else if (Cs >= 11) {
                                    return 8;
                                }
                            }
                            else if (negBalance >= 27) {
                                if (mu_mv_mag_sq < 30) {
                                    if (Rs < 660) {
                                        return 2;
                                    }
                                    else if (Rs >= 660) {
                                        return 8;
                                    }
                                }
                                else if (mu_mv_mag_sq >= 30) {
                                    if (AFD < 5) {
                                        return 8;
                                    }
                                    else if (AFD >= 5) {
                                        if (ssDCval < 96) {
                                            return 8;
                                        }
                                        else if (ssDCval >= 96) {
                                            if (Rs < 206) {
                                                return 8;
                                            }
                                            else if (Rs >= 206) {
                                                return 2;
                                            }
                                        }
                                    }
                                }
                            }
                        }
                    }
                    else if (refDCval >= 160) {
                        if (Cs < 462) {
                            return 4;
                        }
                        else if (Cs >= 462) {
                            if (Cs < 490) {
                                return 2;
                            }
                            else if (Cs >= 490) {
                                return 4;
                            }
                        }
                    }
                }
                else if (MVDiff >= 5) {
                    if (ssDCval < 120) {
                        if (ssDCval < 83) {
                            if (negBalance < 92) {
                                if (AFD < 5) {
                                    if (MVDiff < 25) {
                                        if (posBalance < 5) {
                                            return 8;
                                        }
                                        else if (posBalance >= 5) {
                                            if (diffRsCsdiff < -4) {
                                                if (Cs < 224) {
                                                    return 4;
                                                }
                                                else if (Cs >= 224) {
                                                    return 2;
                                                }
                                            }
                                            else if (diffRsCsdiff >= -4) {
                                                if (MVDiff < 6) {
                                                    if (TSC < 3) {
                                                        return 8;
                                                    }
                                                    else if (TSC >= 3) {
                                                        return 4;
                                                    }
                                                }
                                                else if (MVDiff >= 6) {
                                                    return 4;
                                                }
                                            }
                                        }
                                    }
                                    else if (MVDiff >= 25) {
                                        if (MVDiff < 146) {
                                            if (negBalance < 27) {
                                                if (AFD < 3) {
                                                    if (MVDiff < 64) {
                                                        return 8;
                                                    }
                                                    else if (MVDiff >= 64) {
                                                        if (MVDiff < 82) {
                                                            return 4;
                                                        }
                                                        else if (MVDiff >= 82) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                                else if (AFD >= 3) {
                                                    return 2;
                                                }
                                            }
                                            else if (negBalance >= 27) {
                                                if (negBalance < 45) {
                                                    if (diffMVdiffVal < -1) {
                                                        return 2;
                                                    }
                                                    else if (diffMVdiffVal >= -1) {
                                                        if (ssDCval < 22) {
                                                            return 2;
                                                        }
                                                        else if (ssDCval >= 22) {
                                                            if (MVDiff < 39) {
                                                                return 2;
                                                            }
                                                            else if (MVDiff >= 39) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (negBalance >= 45) {
                                                    if (gchDC < 2) {
                                                        return 4;
                                                    }
                                                    else if (gchDC >= 2) {
                                                        if (MVDiff < 37) {
                                                            return 4;
                                                        }
                                                        else if (MVDiff >= 37) {
                                                            return 2;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (MVDiff >= 146) {
                                            return 1;
                                        }
                                    }
                                }
                                else if (AFD >= 5) {
                                    if (Cs < 570) {
                                        if (negBalance < 80) {
                                            if (ssDCval < 71) {
                                                return 2;
                                            }
                                            else if (ssDCval >= 71) {
                                                return 4;
                                            }
                                        }
                                        else if (negBalance >= 80) {
                                            if (AFD < 19) {
                                                return 4;
                                            }
                                            else if (AFD >= 19) {
                                                return 1;
                                            }
                                        }
                                    }
                                    else if (Cs >= 570) {
                                        if (gchDC < 1) {
                                            return 1;
                                        }
                                        else if (gchDC >= 1) {
                                            if (TSC < 6) {
                                                return 8;
                                            }
                                            else if (TSC >= 6) {
                                                return 2;
                                            }
                                        }
                                    }
                                }
                            }
                            else if (negBalance >= 92) {
                                if (diffAFD < -1) {
                                    if (Rs < 284) {
                                        return 2;
                                    }
                                    else if (Rs >= 284) {
                                        return 8;
                                    }
                                }
                                else if (diffAFD >= -1) {
                                    if (mcTcor < 86) {
                                        return 1;
                                    }
                                    else if (mcTcor >= 86) {
                                        if (negBalance < 109) {
                                            return 8;
                                        }
                                        else if (negBalance >= 109) {
                                            if (MVDiff < 140) {
                                                return 8;
                                            }
                                            else if (MVDiff >= 140) {
                                                return 1;
                                            }
                                        }
                                    }
                                }
                            }
                        }
                        else if (ssDCval >= 83) {
                            if (mu_mv_mag_sq < 43) {
                                if (SC < 928) {
                                    if (negBalance < 57) {
                                        return 1;
                                    }
                                    else if (negBalance >= 57) {
                                        return 2;
                                    }
                                }
                                else if (SC >= 928) {
                                    if (TSC < 12) {
                                        return 8;
                                    }
                                    else if (TSC >= 12) {
                                        return 4;
                                    }
                                }
                            }
                            else if (mu_mv_mag_sq >= 43) {
                                if (MVDiff < 149) {
                                    return 8;
                                }
                                else if (MVDiff >= 149) {
                                    return 2;
                                }
                            }
                        }
                    }
                    else if (ssDCval >= 120) {
                        if (ssDCval < 182) {
                            if (mcTcor < 99) {
                                if (posBalance < 24) {
                                    return 4;
                                }
                                else if (posBalance >= 24) {
                                    if (mu_mv_mag_sq < 83) {
                                        if (Rs < 1512) {
                                            return 4;
                                        }
                                        else if (Rs >= 1512) {
                                            return 2;
                                        }
                                    }
                                    else if (mu_mv_mag_sq >= 83) {
                                        return 2;
                                    }
                                }
                            }
                            else if (mcTcor >= 99) {
                                if (SC < 873) {
                                    return 4;
                                }
                                else if (SC >= 873) {
                                    if (AFD < 2) {
                                        if (MVDiff < 6) {
                                            if (SC < 950) {
                                                return 2;
                                            }
                                            else if (SC >= 950) {
                                                return 4;
                                            }
                                        }
                                        else if (MVDiff >= 6) {
                                            return 2;
                                        }
                                    }
                                    else if (AFD >= 2) {
                                        return 4;
                                    }
                                }
                            }
                        }
                        else if (ssDCval >= 182) {
                            return 1;
                        }
                    }
                }
            }
            else if (Cs >= 708) {
                if (negBalance < 29) {
                    if (MVDiff < 3) {
                        if (negBalance < 3) {
                            if (Cs < 4016) {
                                if (SC < 7308) {
                                    if (diffMVdiffVal < -2) {
                                        if (diffMVdiffVal < -3) {
                                            return 8;
                                        }
                                        else if (diffMVdiffVal >= -3) {
                                            if (Rs < 1078) {
                                                return 4;
                                            }
                                            else if (Rs >= 1078) {
                                                return 8;
                                            }
                                        }
                                    }
                                    else if (diffMVdiffVal >= -2) {
                                        return 8;
                                    }
                                }
                                else if (SC >= 7308) {
                                    if (Cs < 4008) {
                                        if (Cs < 3871) {
                                            if (RsDiff < 1) {
                                                if (Rs < 3638) {
                                                    return 1;
                                                }
                                                else if (Rs >= 3638) {
                                                    return 8;
                                                }
                                            }
                                            else if (RsDiff >= 1) {
                                                return 1;
                                            }
                                        }
                                        else if (Cs >= 3871) {
                                            if (Cs < 3987) {
                                                return 8;
                                            }
                                            else if (Cs >= 3987) {
                                                if (SC < 7774) {
                                                    if (Cs < 3991) {
                                                        return 1;
                                                    }
                                                    else if (Cs >= 3991) {
                                                        return 8;
                                                    }
                                                }
                                                else if (SC >= 7774) {
                                                    return 8;
                                                }
                                            }
                                        }
                                    }
                                    else if (Cs >= 4008) {
                                        return 1;
                                    }
                                }
                            }
                            else if (Cs >= 4016) {
                                return 8;
                            }
                        }
                        else if (negBalance >= 3) {
                            if (RsCsDiff < 18) {
                                if (diffMVdiffVal < -2) {
                                    if (diffAFD < 1) {
                                        if (diffTSC < -1) {
                                            if (ssDCval < 56) {
                                                return 1;
                                            }
                                            else if (ssDCval >= 56) {
                                                return 8;
                                            }
                                        }
                                        else if (diffTSC >= -1) {
                                            return 8;
                                        }
                                    }
                                    else if (diffAFD >= 1) {
                                        return 1;
                                    }
                                }
                                else if (diffMVdiffVal >= -2) {
                                    if (RsCsDiff < 12) {
                                        if (diffRsCsdiff < -3) {
                                            if (ssDCval < 105) {
                                                return 8;
                                            }
                                            else if (ssDCval >= 105) {
                                                if (refDCval < 105) {
                                                    return 4;
                                                }
                                                else if (refDCval >= 105) {
                                                    if (tcor < 96) {
                                                        if (Rs < 3718) {
                                                            return 4;
                                                        }
                                                        else if (Rs >= 3718) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (tcor >= 96) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                        }
                                        else if (diffRsCsdiff >= -3) {
                                            if (Cs < 1284) {
                                                if (Rs < 1394) {
                                                    return 8;
                                                }
                                                else if (Rs >= 1394) {
                                                    if (Cs < 1283) {
                                                        return 8;
                                                    }
                                                    else if (Cs >= 1283) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                            else if (Cs >= 1284) {
                                                return 8;
                                            }
                                        }
                                    }
                                    else if (RsCsDiff >= 12) {
                                        if (MVDiff < 2) {
                                            if (mu_mv_mag_sq < 123) {
                                                if (mu_mv_mag_sq < 121) {
                                                    return 8;
                                                }
                                                else if (mu_mv_mag_sq >= 121) {
                                                    if (SC < 6714) {
                                                        return 1;
                                                    }
                                                    else if (SC >= 6714) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                            else if (mu_mv_mag_sq >= 123) {
                                                return 8;
                                            }
                                        }
                                        else if (MVDiff >= 2) {
                                            return 1;
                                        }
                                    }
                                }
                            }
                            else if (RsCsDiff >= 18) {
                                if (mu_mv_mag_sq < 94) {
                                    if (RsCsDiff < 38) {
                                        return 8;
                                    }
                                    else if (RsCsDiff >= 38) {
                                        return 4;
                                    }
                                }
                                else if (mu_mv_mag_sq >= 94) {
                                    if (TSC < 6) {
                                        if (Cs < 3092) {
                                            if (TSC < 3) {
                                                if (Rs < 2352) {
                                                    if (SC < 3820) {
                                                        return 8;
                                                    }
                                                    else if (SC >= 3820) {
                                                        return 1;
                                                    }
                                                }
                                                else if (Rs >= 2352) {
                                                    if (posBalance < 42) {
                                                        if (tcor < 99) {
                                                            return 8;
                                                        }
                                                        else if (tcor >= 99) {
                                                            if (ssDCval < 108) {
                                                                return 8;
                                                            }
                                                            else if (ssDCval >= 108) {
                                                                return 1;
                                                            }
                                                        }
                                                    }
                                                    else if (posBalance >= 42) {
                                                        if (posBalance < 43) {
                                                            return 1;
                                                        }
                                                        else if (posBalance >= 43) {
                                                            if (AFD < 3) {
                                                                return 8;
                                                            }
                                                            else if (AFD >= 3) {
                                                                if (SC < 6188) {
                                                                    return 1;
                                                                }
                                                                else if (SC >= 6188) {
                                                                    if (Rs < 3171) {
                                                                        return 8;
                                                                    }
                                                                    else if (Rs >= 3171) {
                                                                        return 1;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                            else if (TSC >= 3) {
                                                if (diffRsCsdiff < -49) {
                                                    if (CsDiff < 4) {
                                                        return 8;
                                                    }
                                                    else if (CsDiff >= 4) {
                                                        return 1;
                                                    }
                                                }
                                                else if (diffRsCsdiff >= -49) {
                                                    if (mu_mv_mag_sq < 96) {
                                                        if (Rs < 2962) {
                                                            return 8;
                                                        }
                                                        else if (Rs >= 2962) {
                                                            return 1;
                                                        }
                                                    }
                                                    else if (mu_mv_mag_sq >= 96) {
                                                        if (Rs < 3102) {
                                                            return 8;
                                                        }
                                                        else if (Rs >= 3102) {
                                                            if (SC < 6174) {
                                                                return 1;
                                                            }
                                                            else if (SC >= 6174) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (Cs >= 3092) {
                                            return 8;
                                        }
                                    }
                                    else if (TSC >= 6) {
                                        if (MVDiff < 2) {
                                            return 1;
                                        }
                                        else if (MVDiff >= 2) {
                                            return 8;
                                        }
                                    }
                                }
                            }
                        }
                    }
                    else if (MVDiff >= 3) {
                        if (SC < 2503) {
                            if (ssDCval < 28) {
                                return 8;
                            }
                            else if (ssDCval >= 28) {
                                if (Cs < 1069) {
                                    if (mcTcor < 97) {
                                        if (TSC < 3) {
                                            return 1;
                                        }
                                        else if (TSC >= 3) {
                                            return 8;
                                        }
                                    }
                                    else if (mcTcor >= 97) {
                                        return 1;
                                    }
                                }
                                else if (Cs >= 1069) {
                                    if (RsDiff < 1) {
                                        if (SC < 2304) {
                                            if (refDCval < 102) {
                                                return 4;
                                            }
                                            else if (refDCval >= 102) {
                                                return 8;
                                            }
                                        }
                                        else if (SC >= 2304) {
                                            return 2;
                                        }
                                    }
                                    else if (RsDiff >= 1) {
                                        return 8;
                                    }
                                }
                            }
                        }
                        else if (SC >= 2503) {
                            if (CsDiff < 5) {
                                if (negBalance < 1) {
                                    return 2;
                                }
                                else if (negBalance >= 1) {
                                    if (ssDCval < 94) {
                                        if (negBalance < 22) {
                                            return 8;
                                        }
                                        else if (negBalance >= 22) {
                                            if (Rs < 2995) {
                                                if (Cs < 863) {
                                                    return 8;
                                                }
                                                else if (Cs >= 863) {
                                                    return 1;
                                                }
                                            }
                                            else if (Rs >= 2995) {
                                                return 8;
                                            }
                                        }
                                    }
                                    else if (ssDCval >= 94) {
                                        if (RsCsDiff < 43) {
                                            return 8;
                                        }
                                        else if (RsCsDiff >= 43) {
                                            return 4;
                                        }
                                    }
                                }
                            }
                            else if (CsDiff >= 5) {
                                if (Rs < 1372) {
                                    return 8;
                                }
                                else if (Rs >= 1372) {
                                    if (TSC < 6) {
                                        return 4;
                                    }
                                    else if (TSC >= 6) {
                                        if (Rs < 3798) {
                                            return 2;
                                        }
                                        else if (Rs >= 3798) {
                                            return 4;
                                        }
                                    }
                                }
                            }
                        }
                    }
                }
                else if (negBalance >= 29) {
                    if (mu_mv_mag_sq < 103) {
                        if (refDCval < 93) {
                            if (RsDiff < 5) {
                                if (AFD < 6) {
                                    if (ssDCval < 20) {
                                        return 2;
                                    }
                                    else if (ssDCval >= 20) {
                                        if (negBalance < 34) {
                                            if (Cs < 855) {
                                                if (ssDCval < 54) {
                                                    return 4;
                                                }
                                                else if (ssDCval >= 54) {
                                                    return 8;
                                                }
                                            }
                                            else if (Cs >= 855) {
                                                if (ssDCval < 54) {
                                                    return 1;
                                                }
                                                else if (ssDCval >= 54) {
                                                    if (CsDiff < 4) {
                                                        if (SC < 2031) {
                                                            return 4;
                                                        }
                                                        else if (SC >= 2031) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (CsDiff >= 4) {
                                                        return 1;
                                                    }
                                                }
                                            }
                                        }
                                        else if (negBalance >= 34) {
                                            if (TSC < 5) {
                                                if (mcTcor < 95) {
                                                    return 1;
                                                }
                                                else if (mcTcor >= 95) {
                                                    if (Cs < 759) {
                                                        return 4;
                                                    }
                                                    else if (Cs >= 759) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                            else if (TSC >= 5) {
                                                if (Rs < 2736) {
                                                    if (MVDiff < 17) {
                                                        return 1;
                                                    }
                                                    else if (MVDiff >= 17) {
                                                        return 8;
                                                    }
                                                }
                                                else if (Rs >= 2736) {
                                                    if (Cs < 3341) {
                                                        return 8;
                                                    }
                                                    else if (Cs >= 3341) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (AFD >= 6) {
                                    if (ssDCval < 89) {
                                        if (Scindex < 6) {
                                            if (diffAFD < 1) {
                                                if (mu_mv_mag_sq < 64) {
                                                    if (Cs < 764) {
                                                        return 1;
                                                    }
                                                    else if (Cs >= 764) {
                                                        if (Cs < 1026) {
                                                            return 4;
                                                        }
                                                        else if (Cs >= 1026) {
                                                            return 2;
                                                        }
                                                    }
                                                }
                                                else if (mu_mv_mag_sq >= 64) {
                                                    if (diffMVdiffVal < 1) {
                                                        return 4;
                                                    }
                                                    else if (diffMVdiffVal >= 1) {
                                                        if (CsDiff < 4) {
                                                            return 4;
                                                        }
                                                        else if (CsDiff >= 4) {
                                                            if (Rs < 1415) {
                                                                return 4;
                                                            }
                                                            else if (Rs >= 1415) {
                                                                return 2;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                            else if (diffAFD >= 1) {
                                                if (CsDiff < 2) {
                                                    return 4;
                                                }
                                                else if (CsDiff >= 2) {
                                                    return 2;
                                                }
                                            }
                                        }
                                        else if (Scindex >= 6) {
                                            if (refDCval < 54) {
                                                if (Rs < 3498) {
                                                    return 8;
                                                }
                                                else if (Rs >= 3498) {
                                                    return 4;
                                                }
                                            }
                                            else if (refDCval >= 54) {
                                                if (posBalance < 52) {
                                                    return 4;
                                                }
                                                else if (posBalance >= 52) {
                                                    return 8;
                                                }
                                            }
                                        }
                                    }
                                    else if (ssDCval >= 89) {
                                        if (negBalance < 42) {
                                            return 8;
                                        }
                                        else if (negBalance >= 42) {
                                            if (posBalance < 54) {
                                                return 1;
                                            }
                                            else if (posBalance >= 54) {
                                                return 8;
                                            }
                                        }
                                    }
                                }
                            }
                            else if (RsDiff >= 5) {
                                if (Rs < 2574) {
                                    if (TSCindex < 9) {
                                        if (posBalance < 30) {
                                            return 8;
                                        }
                                        else if (posBalance >= 30) {
                                            if (negBalance < 45) {
                                                return 4;
                                            }
                                            else if (negBalance >= 45) {
                                                if (Rs < 2266) {
                                                    return 4;
                                                }
                                                else if (Rs >= 2266) {
                                                    return 2;
                                                }
                                            }
                                        }
                                    }
                                    else if (TSCindex >= 9) {
                                        if (mcTcor < 99) {
                                            if (SC < 4421) {
                                                return 1;
                                            }
                                            else if (SC >= 4421) {
                                                return 2;
                                            }
                                        }
                                        else if (mcTcor >= 99) {
                                            return 4;
                                        }
                                    }
                                }
                                else if (Rs >= 2574) {
                                    if (MVDiff < 2) {
                                        if (Cs < 1842) {
                                            return 1;
                                        }
                                        else if (Cs >= 1842) {
                                            return 4;
                                        }
                                    }
                                    else if (MVDiff >= 2) {
                                        return 1;
                                    }
                                }
                            }
                        }
                        else if (refDCval >= 93) {
                            if (CsDiff < 5) {
                                if (Cs < 1769) {
                                    if (refDCval < 104) {
                                        return 4;
                                    }
                                    else if (refDCval >= 104) {
                                        if (Rs < 3569) {
                                            if (negBalance < 37) {
                                                if (refDCval < 119) {
                                                    return 8;
                                                }
                                                else if (refDCval >= 119) {
                                                    return 4;
                                                }
                                            }
                                            else if (negBalance >= 37) {
                                                if (SC < 2863) {
                                                    if (diffAFD < -5) {
                                                        return 4;
                                                    }
                                                    else if (diffAFD >= -5) {
                                                        return 8;
                                                    }
                                                }
                                                else if (SC >= 2863) {
                                                    return 4;
                                                }
                                            }
                                        }
                                        else if (Rs >= 3569) {
                                            if (Cs < 1673) {
                                                return 4;
                                            }
                                            else if (Cs >= 1673) {
                                                return 8;
                                            }
                                        }
                                    }
                                }
                                else if (Cs >= 1769) {
                                    return 8;
                                }
                            }
                            else if (CsDiff >= 5) {
                                if (mu_mv_mag_sq < 93) {
                                    if (refDCval < 119) {
                                        if (RsCsDiff < 79) {
                                            if (Cs < 1794) {
                                                return 4;
                                            }
                                            else if (Cs >= 1794) {
                                                if (diffMVdiffVal < 14) {
                                                    return 8;
                                                }
                                                else if (diffMVdiffVal >= 14) {
                                                    return 4;
                                                }
                                            }
                                        }
                                        else if (RsCsDiff >= 79) {
                                            if (TSC < 8) {
                                                if (Cs < 2130) {
                                                    return 4;
                                                }
                                                else if (Cs >= 2130) {
                                                    return 2;
                                                }
                                            }
                                            else if (TSC >= 8) {
                                                return 4;
                                            }
                                        }
                                    }
                                    else if (refDCval >= 119) {
                                        return 4;
                                    }
                                }
                                else if (mu_mv_mag_sq >= 93) {
                                    return 8;
                                }
                            }
                        }
                    }
                    else if (mu_mv_mag_sq >= 103) {
                        if (ssDCval < 123) {
                            if (negBalance < 53) {
                                if (Rs < 2437) {
                                    if (Cs < 1718) {
                                        if (AFD < 6) {
                                            if (ssDCval < 72) {
                                                return 8;
                                            }
                                            else if (ssDCval >= 72) {
                                                if (ssDCval < 74) {
                                                    if (refDCval < 73) {
                                                        return 8;
                                                    }
                                                    else if (refDCval >= 73) {
                                                        if (Rs < 1908) {
                                                            return 4;
                                                        }
                                                        else if (Rs >= 1908) {
                                                            return 2;
                                                        }
                                                    }
                                                }
                                                else if (ssDCval >= 74) {
                                                    if (mu_mv_mag_sq < 120) {
                                                        if (MVDiff < 1) {
                                                            if (negBalance < 33) {
                                                                return 8;
                                                            }
                                                            else if (negBalance >= 33) {
                                                                if (diffTSC < -1) {
                                                                    if (Rs < 1830) {
                                                                        return 8;
                                                                    }
                                                                    else if (Rs >= 1830) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (diffTSC >= -1) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                        else if (MVDiff >= 1) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (mu_mv_mag_sq >= 120) {
                                                        if (SC < 3380) {
                                                            return 8;
                                                        }
                                                        else if (SC >= 3380) {
                                                            if (posBalance < 58) {
                                                                if (mu_mv_mag_sq < 125) {
                                                                    if (Rs < 1898) {
                                                                        if (SC < 3391) {
                                                                            return 4;
                                                                        }
                                                                        else if (SC >= 3391) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                    else if (Rs >= 1898) {
                                                                        if (RsDiff < 4) {
                                                                            return 4;
                                                                        }
                                                                        else if (RsDiff >= 4) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                }
                                                                else if (mu_mv_mag_sq >= 125) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (posBalance >= 58) {
                                                                if (SC < 3490) {
                                                                    return 8;
                                                                }
                                                                else if (SC >= 3490) {
                                                                    return 2;
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (AFD >= 6) {
                                            if (Rs < 1787) {
                                                if (mu_mv_mag_sq < 117) {
                                                    return 2;
                                                }
                                                else if (mu_mv_mag_sq >= 117) {
                                                    return 8;
                                                }
                                            }
                                            else if (Rs >= 1787) {
                                                return 4;
                                            }
                                        }
                                    }
                                    else if (Cs >= 1718) {
                                        if (ssDCval < 89) {
                                            if (SC < 3532) {
                                                return 8;
                                            }
                                            else if (SC >= 3532) {
                                                if (MVDiff < 3) {
                                                    if (posBalance < 58) {
                                                        return 4;
                                                    }
                                                    else if (posBalance >= 58) {
                                                        if (Cs < 1799) {
                                                            return 4;
                                                        }
                                                        else if (Cs >= 1799) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                                else if (MVDiff >= 3) {
                                                    return 8;
                                                }
                                            }
                                        }
                                        else if (ssDCval >= 89) {
                                            if (ssDCval < 103) {
                                                return 8;
                                            }
                                            else if (ssDCval >= 103) {
                                                if (Rs < 2401) {
                                                    return 4;
                                                }
                                                else if (Rs >= 2401) {
                                                    return 8;
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (Rs >= 2437) {
                                    if (mcTcor < 96) {
                                        if (ssDCval < 93) {
                                            if (Cs < 3703) {
                                                if (TSC < 7) {
                                                    if (SC < 5267) {
                                                        return 8;
                                                    }
                                                    else if (SC >= 5267) {
                                                        return 4;
                                                    }
                                                }
                                                else if (TSC >= 7) {
                                                    return 2;
                                                }
                                            }
                                            else if (Cs >= 3703) {
                                                return 8;
                                            }
                                        }
                                        else if (ssDCval >= 93) {
                                            if (ssDCval < 106) {
                                                if (MVDiff < 1) {
                                                    return 8;
                                                }
                                                else if (MVDiff >= 1) {
                                                    if (posBalance < 54) {
                                                        return 8;
                                                    }
                                                    else if (posBalance >= 54) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                            else if (ssDCval >= 106) {
                                                if (posBalance < 57) {
                                                    return 8;
                                                }
                                                else if (posBalance >= 57) {
                                                    return 4;
                                                }
                                            }
                                        }
                                    }
                                    else if (mcTcor >= 96) {
                                        if (negBalance < 48) {
                                            if (Cs < 3073) {
                                                if (mu_mv_mag_sq < 123) {
                                                    if (refDCval < 89) {
                                                        if (refDCval < 77) {
                                                            return 8;
                                                        }
                                                        else if (refDCval >= 77) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (refDCval >= 89) {
                                                        if (Rs < 2713) {
                                                            return 8;
                                                        }
                                                        else if (Rs >= 2713) {
                                                            if (refDCval < 112) {
                                                                if (Cs < 3049) {
                                                                    return 8;
                                                                }
                                                                else if (Cs >= 3049) {
                                                                    if (Rs < 2812) {
                                                                        return 4;
                                                                    }
                                                                    else if (Rs >= 2812) {
                                                                        return 8;
                                                                    }
                                                                }
                                                            }
                                                            else if (refDCval >= 112) {
                                                                if (Rs < 2740) {
                                                                    return 4;
                                                                }
                                                                else if (Rs >= 2740) {
                                                                    if (SC < 5894) {
                                                                        return 8;
                                                                    }
                                                                    else if (SC >= 5894) {
                                                                        if (SC < 5999) {
                                                                            return 4;
                                                                        }
                                                                        else if (SC >= 5999) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (mu_mv_mag_sq >= 123) {
                                                    return 8;
                                                }
                                            }
                                            else if (Cs >= 3073) {
                                                if (ssDCval < 57) {
                                                    return 4;
                                                }
                                                else if (ssDCval >= 57) {
                                                    return 8;
                                                }
                                            }
                                        }
                                        else if (negBalance >= 48) {
                                            if (refDCval < 112) {
                                                return 8;
                                            }
                                            else if (refDCval >= 112) {
                                                if (Rs < 2703) {
                                                    if (CsDiff < 4) {
                                                        if (MVDiff < 1) {
                                                            if (Cs < 2884) {
                                                                return 8;
                                                            }
                                                            else if (Cs >= 2884) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (MVDiff >= 1) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (CsDiff >= 4) {
                                                        return 8;
                                                    }
                                                }
                                                else if (Rs >= 2703) {
                                                    if (SC < 6152) {
                                                        if (SC < 5750) {
                                                            if (refDCval < 113) {
                                                                return 8;
                                                            }
                                                            else if (refDCval >= 113) {
                                                                if (mu_mv_mag_sq < 121) {
                                                                    return 4;
                                                                }
                                                                else if (mu_mv_mag_sq >= 121) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                        else if (SC >= 5750) {
                                                            if (diffTSC < 1) {
                                                                if (Rs < 2806) {
                                                                    return 4;
                                                                }
                                                                else if (Rs >= 2806) {
                                                                    if (Cs < 3054) {
                                                                        return 8;
                                                                    }
                                                                    else if (Cs >= 3054) {
                                                                        if (RsCsDiff < 37) {
                                                                            if (ssDCval < 112) {
                                                                                return 4;
                                                                            }
                                                                            else if (ssDCval >= 112) {
                                                                                return 8;
                                                                            }
                                                                        }
                                                                        else if (RsCsDiff >= 37) {
                                                                            if (ssDCval < 114) {
                                                                                if (RsCsDiff < 46) {
                                                                                    return 4;
                                                                                }
                                                                                else if (RsCsDiff >= 46) {
                                                                                    return 8;
                                                                                }
                                                                            }
                                                                            else if (ssDCval >= 114) {
                                                                                return 8;
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                            else if (diffTSC >= 1) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                    else if (SC >= 6152) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                            else if (negBalance >= 53) {
                                if (Rs < 5675) {
                                    if (refDCval < 107) {
                                        return 8;
                                    }
                                    else if (refDCval >= 107) {
                                        return 4;
                                    }
                                }
                                else if (Rs >= 5675) {
                                    return 4;
                                }
                            }
                        }
                        else if (ssDCval >= 123) {
                            return 8;
                        }
                    }
                }
            }
        }
        return 4;
    }
    mfxU16 AGOPSelect5(mfxU32 Rs, mfxU32 CsDiff, mfxU32 TSC, mfxI32 diffAFD, mfxU32 Scindex, mfxU32 RsDiff, mfxI32 diffMVdiffVal, mfxU32 RsCsDiff, mfxU32 TSCindex, mfxI16 tcor, mfxU32 refDCval, mfxU32 MVDiff, mfxU32 AFD, mfxI32 diffTSC, mfxI32 diffRsCsdiff, mfxU32 posBalance, mfxU32 SC, mfxU32 mu_mv_mag_sq, mfxU32 negBalance, mfxI16 mcTcor, mfxU32 gchDC, mfxU32 Cs, mfxU32 ssDCval) {
        if (AFD < 7) {
            if (MVDiff < 5) {
                if (mu_mv_mag_sq < 110) {
                    if (Scindex < 2) {
                        if (Cs < 1787) {
                            if (Cs < 56) {
                                return 4;
                            }
                            else if (Cs >= 56) {
                                return 8;
                            }
                        }
                        else if (Cs >= 1787) {
                            if (Rs < 4729) {
                                if (Rs < 2707) {
                                    return 1;
                                }
                                else if (Rs >= 2707) {
                                    if (Cs < 3881) {
                                        return 8;
                                    }
                                    else if (Cs >= 3881) {
                                        if (Rs < 3791) {
                                            if (Rs < 3755) {
                                                return 8;
                                            }
                                            else if (Rs >= 3755) {
                                                if (SC < 7755) {
                                                    return 1;
                                                }
                                                else if (SC >= 7755) {
                                                    return 8;
                                                }
                                            }
                                        }
                                        else if (Rs >= 3791) {
                                            return 8;
                                        }
                                    }
                                }
                            }
                            else if (Rs >= 4729) {
                                if (Rs < 5376) {
                                    return 1;
                                }
                                else if (Rs >= 5376) {
                                    return 8;
                                }
                            }
                        }
                    }
                    else if (Scindex >= 2) {
                        if (refDCval < 101) {
                            if (MVDiff < 1) {
                                if (negBalance < 19) {
                                    if (AFD < 2) {
                                        return 8;
                                    }
                                    else if (AFD >= 2) {
                                        return 1;
                                    }
                                }
                                else if (negBalance >= 19) {
                                    if (negBalance < 44) {
                                        if (diffMVdiffVal < -1) {
                                            if (Rs < 584) {
                                                return 4;
                                            }
                                            else if (Rs >= 584) {
                                                if (refDCval < 79) {
                                                    return 1;
                                                }
                                                else if (refDCval >= 79) {
                                                    if (Rs < 1556) {
                                                        return 8;
                                                    }
                                                    else if (Rs >= 1556) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                        }
                                        else if (diffMVdiffVal >= -1) {
                                            return 4;
                                        }
                                    }
                                    else if (negBalance >= 44) {
                                        if (AFD < 6) {
                                            return 4;
                                        }
                                        else if (AFD >= 6) {
                                            if (Cs < 1942) {
                                                return 2;
                                            }
                                            else if (Cs >= 1942) {
                                                return 4;
                                            }
                                        }
                                    }
                                }
                            }
                            else if (MVDiff >= 1) {
                                if (Rs < 1137) {
                                    if (ssDCval < 85) {
                                        if (mu_mv_mag_sq < 88) {
                                            if (CsDiff < 1) {
                                                if (AFD < 5) {
                                                    return 8;
                                                }
                                                else if (AFD >= 5) {
                                                    return 4;
                                                }
                                            }
                                            else if (CsDiff >= 1) {
                                                if (posBalance < 48) {
                                                    if (negBalance < 27) {
                                                        return 1;
                                                    }
                                                    else if (negBalance >= 27) {
                                                        return 4;
                                                    }
                                                }
                                                else if (posBalance >= 48) {
                                                    if (posBalance < 57) {
                                                        if (mcTcor < 90) {
                                                            if (SC < 1396) {
                                                                return 4;
                                                            }
                                                            else if (SC >= 1396) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (mcTcor >= 90) {
                                                            if (ssDCval < 54) {
                                                                if (SC < 1201) {
                                                                    if (ssDCval < 51) {
                                                                        return 8;
                                                                    }
                                                                    else if (ssDCval >= 51) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (SC >= 1201) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (ssDCval >= 54) {
                                                                if (refDCval < 55) {
                                                                    return 8;
                                                                }
                                                                else if (refDCval >= 55) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (posBalance >= 57) {
                                                        if (RsCsDiff < 4) {
                                                            return 8;
                                                        }
                                                        else if (RsCsDiff >= 4) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (mu_mv_mag_sq >= 88) {
                                            if (Cs < 1263) {
                                                if (Rs < 547) {
                                                    return 4;
                                                }
                                                else if (Rs >= 547) {
                                                    return 8;
                                                }
                                            }
                                            else if (Cs >= 1263) {
                                                if (RsDiff < 1) {
                                                    return 2;
                                                }
                                                else if (RsDiff >= 1) {
                                                    return 8;
                                                }
                                            }
                                        }
                                    }
                                    else if (ssDCval >= 85) {
                                        if (mu_mv_mag_sq < 32) {
                                            if (tcor < 60) {
                                                return 4;
                                            }
                                            else if (tcor >= 60) {
                                                return 2;
                                            }
                                        }
                                        else if (mu_mv_mag_sq >= 32) {
                                            if (AFD < 2) {
                                                return 8;
                                            }
                                            else if (AFD >= 2) {
                                                return 4;
                                            }
                                        }
                                    }
                                }
                                else if (Rs >= 1137) {
                                    if (SC < 2359) {
                                        if (mu_mv_mag_sq < 100) {
                                            if (Cs < 934) {
                                                if (Rs < 1364) {
                                                    if (refDCval < 53) {
                                                        return 1;
                                                    }
                                                    else if (refDCval >= 53) {
                                                        return 4;
                                                    }
                                                }
                                                else if (Rs >= 1364) {
                                                    if (negBalance < 39) {
                                                        return 4;
                                                    }
                                                    else if (negBalance >= 39) {
                                                        return 1;
                                                    }
                                                }
                                            }
                                            else if (Cs >= 934) {
                                                return 1;
                                            }
                                        }
                                        else if (mu_mv_mag_sq >= 100) {
                                            if (Rs < 1257) {
                                                return 2;
                                            }
                                            else if (Rs >= 1257) {
                                                return 8;
                                            }
                                        }
                                    }
                                    else if (SC >= 2359) {
                                        if (AFD < 5) {
                                            if (mcTcor < 95) {
                                                return 1;
                                            }
                                            else if (mcTcor >= 95) {
                                                if (MVDiff < 2) {
                                                    if (Rs < 2893) {
                                                        return 8;
                                                    }
                                                    else if (Rs >= 2893) {
                                                        return 1;
                                                    }
                                                }
                                                else if (MVDiff >= 2) {
                                                    if (Rs < 1377) {
                                                        if (SC < 2376) {
                                                            return 8;
                                                        }
                                                        else if (SC >= 2376) {
                                                            return 1;
                                                        }
                                                    }
                                                    else if (Rs >= 1377) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                        }
                                        else if (AFD >= 5) {
                                            if (posBalance < 54) {
                                                if (RsCsDiff < 16) {
                                                    if (mu_mv_mag_sq < 72) {
                                                        if (Cs < 749) {
                                                            return 2;
                                                        }
                                                        else if (Cs >= 749) {
                                                            return 1;
                                                        }
                                                    }
                                                    else if (mu_mv_mag_sq >= 72) {
                                                        if (diffMVdiffVal < -3) {
                                                            if (TSC < 5) {
                                                                return 4;
                                                            }
                                                            else if (TSC >= 5) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (diffMVdiffVal >= -3) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                                else if (RsCsDiff >= 16) {
                                                    if (mu_mv_mag_sq < 103) {
                                                        if (ssDCval < 50) {
                                                            if (Rs < 2351) {
                                                                return 4;
                                                            }
                                                            else if (Rs >= 2351) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (ssDCval >= 50) {
                                                            if (negBalance < 46) {
                                                                return 1;
                                                            }
                                                            else if (negBalance >= 46) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                    else if (mu_mv_mag_sq >= 103) {
                                                        if (SC < 3943) {
                                                            return 4;
                                                        }
                                                        else if (SC >= 3943) {
                                                            if (Cs < 2435) {
                                                                return 8;
                                                            }
                                                            else if (Cs >= 2435) {
                                                                return 1;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                            else if (posBalance >= 54) {
                                                if (negBalance < 43) {
                                                    return 8;
                                                }
                                                else if (negBalance >= 43) {
                                                    if (Rs < 3569) {
                                                        if (Rs < 1291) {
                                                            return 2;
                                                        }
                                                        else if (Rs >= 1291) {
                                                            if (mcTcor < 97) {
                                                                return 8;
                                                            }
                                                            else if (mcTcor >= 97) {
                                                                return 2;
                                                            }
                                                        }
                                                    }
                                                    else if (Rs >= 3569) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                        }
                        else if (refDCval >= 101) {
                            if (ssDCval < 161) {
                                if (Cs < 15) {
                                    if (SC < 911) {
                                        if (SC < 690) {
                                            return 8;
                                        }
                                        else if (SC >= 690) {
                                            return 1;
                                        }
                                    }
                                    else if (SC >= 911) {
                                        return 8;
                                    }
                                }
                                else if (Cs >= 15) {
                                    if (Rs < 2407) {
                                        if (negBalance < 32) {
                                            if (AFD < 6) {
                                                return 8;
                                            }
                                            else if (AFD >= 6) {
                                                if (Rs < 601) {
                                                    return 4;
                                                }
                                                else if (Rs >= 601) {
                                                    return 8;
                                                }
                                            }
                                        }
                                        else if (negBalance >= 32) {
                                            if (ssDCval < 118) {
                                                if (refDCval < 103) {
                                                    if (Cs < 1188) {
                                                        return 8;
                                                    }
                                                    else if (Cs >= 1188) {
                                                        return 4;
                                                    }
                                                }
                                                else if (refDCval >= 103) {
                                                    return 8;
                                                }
                                            }
                                            else if (ssDCval >= 118) {
                                                if (SC < 1204) {
                                                    return 2;
                                                }
                                                else if (SC >= 1204) {
                                                    if (negBalance < 40) {
                                                        return 4;
                                                    }
                                                    else if (negBalance >= 40) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (Rs >= 2407) {
                                        if (tcor < 97) {
                                            if (Cs < 1194) {
                                                return 8;
                                            }
                                            else if (Cs >= 1194) {
                                                if (Rs < 2540) {
                                                    return 1;
                                                }
                                                else if (Rs >= 2540) {
                                                    if (Cs < 2282) {
                                                        return 4;
                                                    }
                                                    else if (Cs >= 2282) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                        }
                                        else if (tcor >= 97) {
                                            if (RsDiff < 7) {
                                                return 8;
                                            }
                                            else if (RsDiff >= 7) {
                                                if (SC < 5413) {
                                                    return 8;
                                                }
                                                else if (SC >= 5413) {
                                                    return 1;
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                            else if (ssDCval >= 161) {
                                if (Rs < 747) {
                                    if (refDCval < 165) {
                                        return 4;
                                    }
                                    else if (refDCval >= 165) {
                                        if (SC < 851) {
                                            return 4;
                                        }
                                        else if (SC >= 851) {
                                            return 2;
                                        }
                                    }
                                }
                                else if (Rs >= 747) {
                                    return 8;
                                }
                            }
                        }
                    }
                }
                else if (mu_mv_mag_sq >= 110) {
                    if (negBalance < 3) {
                        if (SC < 7800) {
                            if (Cs < 4003) {
                                if (ssDCval < 97) {
                                    if (diffRsCsdiff < -2) {
                                        if (SC < 2980) {
                                            return 2;
                                        }
                                        else if (SC >= 2980) {
                                            return 8;
                                        }
                                    }
                                    else if (diffRsCsdiff >= -2) {
                                        if (MVDiff < 3) {
                                            return 8;
                                        }
                                        else if (MVDiff >= 3) {
                                            if (Rs < 1026) {
                                                return 4;
                                            }
                                            else if (Rs >= 1026) {
                                                return 8;
                                            }
                                        }
                                    }
                                }
                                else if (ssDCval >= 97) {
                                    if (diffRsCsdiff < 2) {
                                        if (posBalance < 2) {
                                            if (tcor < 70) {
                                                return 8;
                                            }
                                            else if (tcor >= 70) {
                                                if (mu_mv_mag_sq < 123) {
                                                    return 8;
                                                }
                                                else if (mu_mv_mag_sq >= 123) {
                                                    if (Rs < 1423) {
                                                        return 1;
                                                    }
                                                    else if (Rs >= 1423) {
                                                        if (Cs < 1930) {
                                                            return 8;
                                                        }
                                                        else if (Cs >= 1930) {
                                                            return 1;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (posBalance >= 2) {
                                            if (ssDCval < 104) {
                                                if (Rs < 3701) {
                                                    if (negBalance < 1) {
                                                        return 1;
                                                    }
                                                    else if (negBalance >= 1) {
                                                        if (posBalance < 21) {
                                                            return 1;
                                                        }
                                                        else if (posBalance >= 21) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                                else if (Rs >= 3701) {
                                                    if (negBalance < 2) {
                                                        return 8;
                                                    }
                                                    else if (negBalance >= 2) {
                                                        if (RsDiff < 1) {
                                                            return 1;
                                                        }
                                                        else if (RsDiff >= 1) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (ssDCval >= 104) {
                                                return 8;
                                            }
                                        }
                                    }
                                    else if (diffRsCsdiff >= 2) {
                                        if (posBalance < 22) {
                                            if (RsCsDiff < 7) {
                                                return 8;
                                            }
                                            else if (RsCsDiff >= 7) {
                                                if (Cs < 2757) {
                                                    if (diffTSC < 1) {
                                                        return 8;
                                                    }
                                                    else if (diffTSC >= 1) {
                                                        return 1;
                                                    }
                                                }
                                                else if (Cs >= 2757) {
                                                    return 1;
                                                }
                                            }
                                        }
                                        else if (posBalance >= 22) {
                                            return 1;
                                        }
                                    }
                                }
                            }
                            else if (Cs >= 4003) {
                                return 1;
                            }
                        }
                        else if (SC >= 7800) {
                            return 8;
                        }
                    }
                    else if (negBalance >= 3) {
                        if (Rs < 567) {
                            if (Rs < 548) {
                                if (tcor < 100) {
                                    if (SC < 1298) {
                                        if (Cs < 501) {
                                            if (TSCindex < 3) {
                                                if (Cs < 332) {
                                                    if (mu_mv_mag_sq < 127) {
                                                        return 4;
                                                    }
                                                    else if (mu_mv_mag_sq >= 127) {
                                                        if (AFD < 1) {
                                                            return 8;
                                                        }
                                                        else if (AFD >= 1) {
                                                            if (TSCindex < 2) {
                                                                return 8;
                                                            }
                                                            else if (TSCindex >= 2) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (Cs >= 332) {
                                                    return 8;
                                                }
                                            }
                                            else if (TSCindex >= 3) {
                                                return 4;
                                            }
                                        }
                                        else if (Cs >= 501) {
                                            if (gchDC < 1) {
                                                return 4;
                                            }
                                            else if (gchDC >= 1) {
                                                return 8;
                                            }
                                        }
                                    }
                                    else if (SC >= 1298) {
                                        return 8;
                                    }
                                }
                                else if (tcor >= 100) {
                                    return 8;
                                }
                            }
                            else if (Rs >= 548) {
                                return 4;
                            }
                        }
                        else if (Rs >= 567) {
                            if (mu_mv_mag_sq < 121) {
                                if (refDCval < 73) {
                                    if (TSCindex < 7) {
                                        if (TSCindex < 1) {
                                            if (Rs < 1411) {
                                                if (Rs < 950) {
                                                    return 8;
                                                }
                                                else if (Rs >= 950) {
                                                    return 4;
                                                }
                                            }
                                            else if (Rs >= 1411) {
                                                return 8;
                                            }
                                        }
                                        else if (TSCindex >= 1) {
                                            if (diffAFD < -2) {
                                                if (RsDiff < 4) {
                                                    return 1;
                                                }
                                                else if (RsDiff >= 4) {
                                                    return 8;
                                                }
                                            }
                                            else if (diffAFD >= -2) {
                                                if (AFD < 6) {
                                                    if (refDCval < 71) {
                                                        return 8;
                                                    }
                                                    else if (refDCval >= 71) {
                                                        if (AFD < 3) {
                                                            return 1;
                                                        }
                                                        else if (AFD >= 3) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                                else if (AFD >= 6) {
                                                    if (mu_mv_mag_sq < 112) {
                                                        if (Rs < 2192) {
                                                            return 4;
                                                        }
                                                        else if (Rs >= 2192) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (mu_mv_mag_sq >= 112) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (TSCindex >= 7) {
                                        if (tcor < 92) {
                                            if (Rs < 1495) {
                                                return 2;
                                            }
                                            else if (Rs >= 1495) {
                                                return 8;
                                            }
                                        }
                                        else if (tcor >= 92) {
                                            if (Rs < 1862) {
                                                return 2;
                                            }
                                            else if (Rs >= 1862) {
                                                return 8;
                                            }
                                        }
                                    }
                                }
                                else if (refDCval >= 73) {
                                    if (negBalance < 33) {
                                        if (negBalance < 12) {
                                            if (Rs < 2106) {
                                                if (diffMVdiffVal < -2) {
                                                    if (gchDC < 1) {
                                                        return 4;
                                                    }
                                                    else if (gchDC >= 1) {
                                                        return 1;
                                                    }
                                                }
                                                else if (diffMVdiffVal >= -2) {
                                                    if (Rs < 1622) {
                                                        return 8;
                                                    }
                                                    else if (Rs >= 1622) {
                                                        return 1;
                                                    }
                                                }
                                            }
                                            else if (Rs >= 2106) {
                                                if (refDCval < 117) {
                                                    if (posBalance < 9) {
                                                        if (negBalance < 9) {
                                                            return 4;
                                                        }
                                                        else if (negBalance >= 9) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (posBalance >= 9) {
                                                        return 8;
                                                    }
                                                }
                                                else if (refDCval >= 117) {
                                                    if (SC < 4391) {
                                                        return 8;
                                                    }
                                                    else if (SC >= 4391) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                        }
                                        else if (negBalance >= 12) {
                                            if (RsDiff < 4) {
                                                if (diffRsCsdiff < 4) {
                                                    return 8;
                                                }
                                                else if (diffRsCsdiff >= 4) {
                                                    if (tcor < 99) {
                                                        return 8;
                                                    }
                                                    else if (tcor >= 99) {
                                                        if (RsCsDiff < 16) {
                                                            return 8;
                                                        }
                                                        else if (RsCsDiff >= 16) {
                                                            return 1;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (RsDiff >= 4) {
                                                if (Cs < 3077) {
                                                    if (diffMVdiffVal < 1) {
                                                        if (Rs < 1954) {
                                                            return 8;
                                                        }
                                                        else if (Rs >= 1954) {
                                                            if (negBalance < 30) {
                                                                if (mu_mv_mag_sq < 115) {
                                                                    if (Rs < 2885) {
                                                                        return 1;
                                                                    }
                                                                    else if (Rs >= 2885) {
                                                                        return 8;
                                                                    }
                                                                }
                                                                else if (mu_mv_mag_sq >= 115) {
                                                                    return 1;
                                                                }
                                                            }
                                                            else if (negBalance >= 30) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                    else if (diffMVdiffVal >= 1) {
                                                        return 8;
                                                    }
                                                }
                                                else if (Cs >= 3077) {
                                                    return 8;
                                                }
                                            }
                                        }
                                    }
                                    else if (negBalance >= 33) {
                                        if (SC < 4918) {
                                            if (Cs < 2110) {
                                                if (Rs < 1831) {
                                                    if (mu_mv_mag_sq < 115) {
                                                        if (tcor < 94) {
                                                            return 8;
                                                        }
                                                        else if (tcor >= 94) {
                                                            if (SC < 2390) {
                                                                return 8;
                                                            }
                                                            else if (SC >= 2390) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                    else if (mu_mv_mag_sq >= 115) {
                                                        if (Cs < 1353) {
                                                            if (Cs < 986) {
                                                                return 8;
                                                            }
                                                            else if (Cs >= 986) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (Cs >= 1353) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                                else if (Rs >= 1831) {
                                                    if (tcor < 94) {
                                                        if (diffRsCsdiff < 4) {
                                                            return 8;
                                                        }
                                                        else if (diffRsCsdiff >= 4) {
                                                            if (Cs < 1611) {
                                                                return 8;
                                                            }
                                                            else if (Cs >= 1611) {
                                                                return 2;
                                                            }
                                                        }
                                                    }
                                                    else if (tcor >= 94) {
                                                        if (Rs < 2493) {
                                                            return 4;
                                                        }
                                                        else if (Rs >= 2493) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (Cs >= 2110) {
                                                if (SC < 3925) {
                                                    return 8;
                                                }
                                                else if (SC >= 3925) {
                                                    if (tcor < 94) {
                                                        if (RsDiff < 5) {
                                                            return 8;
                                                        }
                                                        else if (RsDiff >= 5) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (tcor >= 94) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                        }
                                        else if (SC >= 4918) {
                                            if (negBalance < 48) {
                                                if (Rs < 2738) {
                                                    if (SC < 5717) {
                                                        if (Cs < 2881) {
                                                            if (Cs < 2866) {
                                                                return 8;
                                                            }
                                                            else if (Cs >= 2866) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (Cs >= 2881) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (SC >= 5717) {
                                                        if (Cs < 3078) {
                                                            return 4;
                                                        }
                                                        else if (Cs >= 3078) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                                else if (Rs >= 2738) {
                                                    return 8;
                                                }
                                            }
                                            else if (negBalance >= 48) {
                                                if (refDCval < 112) {
                                                    return 8;
                                                }
                                                else if (refDCval >= 112) {
                                                    if (ssDCval < 115) {
                                                        if (posBalance < 58) {
                                                            if (Rs < 2650) {
                                                                return 8;
                                                            }
                                                            else if (Rs >= 2650) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (posBalance >= 58) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (ssDCval >= 115) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                            else if (mu_mv_mag_sq >= 121) {
                                if (RsCsDiff < 109) {
                                    if (negBalance < 33) {
                                        if (TSC < 6) {
                                            if (posBalance < 3) {
                                                return 4;
                                            }
                                            else if (posBalance >= 3) {
                                                if (SC < 1254) {
                                                    if (Rs < 667) {
                                                        return 4;
                                                    }
                                                    else if (Rs >= 667) {
                                                        return 8;
                                                    }
                                                }
                                                else if (SC >= 1254) {
                                                    if (negBalance < 12) {
                                                        return 8;
                                                    }
                                                    else if (negBalance >= 12) {
                                                        if (ssDCval < 108) {
                                                            if (refDCval < 51) {
                                                                if (RsCsDiff < 38) {
                                                                    return 8;
                                                                }
                                                                else if (RsCsDiff >= 38) {
                                                                    if (SC < 3990) {
                                                                        return 8;
                                                                    }
                                                                    else if (SC >= 3990) {
                                                                        return 4;
                                                                    }
                                                                }
                                                            }
                                                            else if (refDCval >= 51) {
                                                                if (posBalance < 52) {
                                                                    return 8;
                                                                }
                                                                else if (posBalance >= 52) {
                                                                    if (diffAFD < -1) {
                                                                        if (refDCval < 89) {
                                                                            return 8;
                                                                        }
                                                                        else if (refDCval >= 89) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                    else if (diffAFD >= -1) {
                                                                        return 8;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (ssDCval >= 108) {
                                                            if (negBalance < 13) {
                                                                if (Rs < 1863) {
                                                                    return 4;
                                                                }
                                                                else if (Rs >= 1863) {
                                                                    if (RsCsDiff < 11) {
                                                                        return 8;
                                                                    }
                                                                    else if (RsCsDiff >= 11) {
                                                                        return 4;
                                                                    }
                                                                }
                                                            }
                                                            else if (negBalance >= 13) {
                                                                if (ssDCval < 109) {
                                                                    if (Rs < 3464) {
                                                                        return 1;
                                                                    }
                                                                    else if (Rs >= 3464) {
                                                                        return 8;
                                                                    }
                                                                }
                                                                else if (ssDCval >= 109) {
                                                                    if (RsCsDiff < 29) {
                                                                        return 8;
                                                                    }
                                                                    else if (RsCsDiff >= 29) {
                                                                        if (RsCsDiff < 33) {
                                                                            return 1;
                                                                        }
                                                                        else if (RsCsDiff >= 33) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (TSC >= 6) {
                                            return 1;
                                        }
                                    }
                                    else if (negBalance >= 33) {
                                        if (negBalance < 40) {
                                            if (SC < 4849) {
                                                if (CsDiff < 3) {
                                                    if (Cs < 2298) {
                                                        return 8;
                                                    }
                                                    else if (Cs >= 2298) {
                                                        return 4;
                                                    }
                                                }
                                                else if (CsDiff >= 3) {
                                                    if (tcor < 95) {
                                                        return 8;
                                                    }
                                                    else if (tcor >= 95) {
                                                        if (Cs < 1820) {
                                                            if (TSC < 5) {
                                                                return 8;
                                                            }
                                                            else if (TSC >= 5) {
                                                                if (posBalance < 56) {
                                                                    return 4;
                                                                }
                                                                else if (posBalance >= 56) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                        else if (Cs >= 1820) {
                                                            if (posBalance < 51) {
                                                                if (negBalance < 37) {
                                                                    return 8;
                                                                }
                                                                else if (negBalance >= 37) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (posBalance >= 51) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                            else if (SC >= 4849) {
                                                return 8;
                                            }
                                        }
                                        else if (negBalance >= 40) {
                                            if (refDCval < 74) {
                                                if (refDCval < 73) {
                                                    if (Cs < 1611) {
                                                        return 8;
                                                    }
                                                    else if (Cs >= 1611) {
                                                        if (SC < 3668) {
                                                            return 2;
                                                        }
                                                        else if (SC >= 3668) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                                else if (refDCval >= 73) {
                                                    return 2;
                                                }
                                            }
                                            else if (refDCval >= 74) {
                                                if (AFD < 6) {
                                                    if (Rs < 1791) {
                                                        if (Cs < 2108) {
                                                            return 8;
                                                        }
                                                        else if (Cs >= 2108) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (Rs >= 1791) {
                                                        if (ssDCval < 77) {
                                                            if (SC < 3552) {
                                                                return 8;
                                                            }
                                                            else if (SC >= 3552) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (ssDCval >= 77) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                                else if (AFD >= 6) {
                                                    if (ssDCval < 79) {
                                                        if (Rs < 1912) {
                                                            if (negBalance < 48) {
                                                                return 8;
                                                            }
                                                            else if (negBalance >= 48) {
                                                                if (Rs < 1771) {
                                                                    return 8;
                                                                }
                                                                else if (Rs >= 1771) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                        else if (Rs >= 1912) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (ssDCval >= 79) {
                                                        if (SC < 5670) {
                                                            return 8;
                                                        }
                                                        else if (SC >= 5670) {
                                                            if (mu_mv_mag_sq < 123) {
                                                                if (Cs < 3153) {
                                                                    if (posBalance < 59) {
                                                                        if (ssDCval < 111) {
                                                                            return 8;
                                                                        }
                                                                        else if (ssDCval >= 111) {
                                                                            if (ssDCval < 114) {
                                                                                if (negBalance < 48) {
                                                                                    if (SC < 5846) {
                                                                                        if (Cs < 2992) {
                                                                                            return 8;
                                                                                        }
                                                                                        else if (Cs >= 2992) {
                                                                                            if (Cs < 3016) {
                                                                                                return 4;
                                                                                            }
                                                                                            else if (Cs >= 3016) {
                                                                                                return 8;
                                                                                            }
                                                                                        }
                                                                                    }
                                                                                    else if (SC >= 5846) {
                                                                                        return 4;
                                                                                    }
                                                                                }
                                                                                else if (negBalance >= 48) {
                                                                                    if (ssDCval < 112) {
                                                                                        return 8;
                                                                                    }
                                                                                    else if (ssDCval >= 112) {
                                                                                        if (tcor < 97) {
                                                                                            return 8;
                                                                                        }
                                                                                        else if (tcor >= 97) {
                                                                                            return 4;
                                                                                        }
                                                                                    }
                                                                                }
                                                                            }
                                                                            else if (ssDCval >= 114) {
                                                                                if (CsDiff < 5) {
                                                                                    return 8;
                                                                                }
                                                                                else if (CsDiff >= 5) {
                                                                                    return 4;
                                                                                }
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (posBalance >= 59) {
                                                                        return 8;
                                                                    }
                                                                }
                                                                else if (Cs >= 3153) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (mu_mv_mag_sq >= 123) {
                                                                if (SC < 9308) {
                                                                    if (negBalance < 48) {
                                                                        return 8;
                                                                    }
                                                                    else if (negBalance >= 48) {
                                                                        if (RsDiff < 5) {
                                                                            if (SC < 5750) {
                                                                                return 8;
                                                                            }
                                                                            else if (SC >= 5750) {
                                                                                if (ssDCval < 113) {
                                                                                    if (gchDC < 1) {
                                                                                        return 8;
                                                                                    }
                                                                                    else if (gchDC >= 1) {
                                                                                        if (Rs < 2995) {
                                                                                            return 4;
                                                                                        }
                                                                                        else if (Rs >= 2995) {
                                                                                            return 8;
                                                                                        }
                                                                                    }
                                                                                }
                                                                                else if (ssDCval >= 113) {
                                                                                    if (diffTSC < 1) {
                                                                                        return 4;
                                                                                    }
                                                                                    else if (diffTSC >= 1) {
                                                                                        return 8;
                                                                                    }
                                                                                }
                                                                            }
                                                                        }
                                                                        else if (RsDiff >= 5) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                }
                                                                else if (SC >= 9308) {
                                                                    if (ssDCval < 84) {
                                                                        if (refDCval < 80) {
                                                                            return 8;
                                                                        }
                                                                        else if (refDCval >= 80) {
                                                                            if (Cs < 5031) {
                                                                                return 4;
                                                                            }
                                                                            else if (Cs >= 5031) {
                                                                                return 8;
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (ssDCval >= 84) {
                                                                        return 8;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (RsCsDiff >= 109) {
                                    if (Rs < 5251) {
                                        if (SC < 10029) {
                                            if (CsDiff < 8) {
                                                return 8;
                                            }
                                            else if (CsDiff >= 8) {
                                                if (Rs < 4618) {
                                                    return 8;
                                                }
                                                else if (Rs >= 4618) {
                                                    if (Rs < 4715) {
                                                        return 4;
                                                    }
                                                    else if (Rs >= 4715) {
                                                        if (SC < 9669) {
                                                            if (Rs < 4881) {
                                                                return 8;
                                                            }
                                                            else if (Rs >= 4881) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (SC >= 9669) {
                                                            if (ssDCval < 80) {
                                                                return 8;
                                                            }
                                                            else if (ssDCval >= 80) {
                                                                if (RsDiff < 8) {
                                                                    return 4;
                                                                }
                                                                else if (RsDiff >= 8) {
                                                                    if (mcTcor < 92) {
                                                                        if (Rs < 4899) {
                                                                            if (Cs < 4938) {
                                                                                return 8;
                                                                            }
                                                                            else if (Cs >= 4938) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                        else if (Rs >= 4899) {
                                                                            if (Rs < 4953) {
                                                                                return 8;
                                                                            }
                                                                            else if (Rs >= 4953) {
                                                                                if (RsCsDiff < 137) {
                                                                                    return 8;
                                                                                }
                                                                                else if (RsCsDiff >= 137) {
                                                                                    return 4;
                                                                                }
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (mcTcor >= 92) {
                                                                        if (mu_mv_mag_sq < 128) {
                                                                            if (Rs < 4858) {
                                                                                return 8;
                                                                            }
                                                                            else if (Rs >= 4858) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                        else if (mu_mv_mag_sq >= 128) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (SC >= 10029) {
                                            return 4;
                                        }
                                    }
                                    else if (Rs >= 5251) {
                                        return 8;
                                    }
                                }
                            }
                        }
                    }
                }
            }
            else if (MVDiff >= 5) {
                if (Cs < 1670) {
                    if (mu_mv_mag_sq < 100) {
                        if (posBalance < 14) {
                            if (mu_mv_mag_sq < 79) {
                                if (diffAFD < -1) {
                                    if (mu_mv_mag_sq < 49) {
                                        if (posBalance < 13) {
                                            if (AFD < 2) {
                                                if (MVDiff < 30) {
                                                    return 4;
                                                }
                                                else if (MVDiff >= 30) {
                                                    return 2;
                                                }
                                            }
                                            else if (AFD >= 2) {
                                                return 2;
                                            }
                                        }
                                        else if (posBalance >= 13) {
                                            return 4;
                                        }
                                    }
                                    else if (mu_mv_mag_sq >= 49) {
                                        if (posBalance < 12) {
                                            return 8;
                                        }
                                        else if (posBalance >= 12) {
                                            return 1;
                                        }
                                    }
                                }
                                else if (diffAFD >= -1) {
                                    if (Rs < 320) {
                                        if (TSCindex < 3) {
                                            if (negBalance < 33) {
                                                if (negBalance < 11) {
                                                    if (tcor < 59) {
                                                        return 4;
                                                    }
                                                    else if (tcor >= 59) {
                                                        return 8;
                                                    }
                                                }
                                                else if (negBalance >= 11) {
                                                    if (negBalance < 23) {
                                                        if (Cs < 44) {
                                                            if (mcTcor < 36) {
                                                                return 2;
                                                            }
                                                            else if (mcTcor >= 36) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (Cs >= 44) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (negBalance >= 23) {
                                                        if (mcTcor < 80) {
                                                            return 8;
                                                        }
                                                        else if (mcTcor >= 80) {
                                                            return 2;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (negBalance >= 33) {
                                                if (MVDiff < 39) {
                                                    return 4;
                                                }
                                                else if (MVDiff >= 39) {
                                                    if (diffTSC < -2) {
                                                        return 4;
                                                    }
                                                    else if (diffTSC >= -2) {
                                                        if (mcTcor < 64) {
                                                            if (TSC < 2) {
                                                                return 1;
                                                            }
                                                            else if (TSC >= 2) {
                                                                if (MVDiff < 282) {
                                                                    return 1;
                                                                }
                                                                else if (MVDiff >= 282) {
                                                                    return 2;
                                                                }
                                                            }
                                                        }
                                                        else if (mcTcor >= 64) {
                                                            return 2;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (TSCindex >= 3) {
                                            if (MVDiff < 57) {
                                                if (negBalance < 88) {
                                                    return 4;
                                                }
                                                else if (negBalance >= 88) {
                                                    return 8;
                                                }
                                            }
                                            else if (MVDiff >= 57) {
                                                return 8;
                                            }
                                        }
                                    }
                                    else if (Rs >= 320) {
                                        if (AFD < 1) {
                                            return 1;
                                        }
                                        else if (AFD >= 1) {
                                            return 8;
                                        }
                                    }
                                }
                            }
                            else if (mu_mv_mag_sq >= 79) {
                                if (Rs < 3) {
                                    return 2;
                                }
                                else if (Rs >= 3) {
                                    return 4;
                                }
                            }
                        }
                        else if (posBalance >= 14) {
                            if (SC < 1148) {
                                if (Cs < 374) {
                                    if (Rs < 230) {
                                        if (tcor < 58) {
                                            if (mcTcor < 61) {
                                                if (MVDiff < 140) {
                                                    if (diffMVdiffVal < -55) {
                                                        if (TSC < 1) {
                                                            if (MVDiff < 83) {
                                                                return 8;
                                                            }
                                                            else if (MVDiff >= 83) {
                                                                return 2;
                                                            }
                                                        }
                                                        else if (TSC >= 1) {
                                                            if (mcTcor < 52) {
                                                                return 8;
                                                            }
                                                            else if (mcTcor >= 52) {
                                                                return 1;
                                                            }
                                                        }
                                                    }
                                                    else if (diffMVdiffVal >= -55) {
                                                        if (MVDiff < 94) {
                                                            if (mu_mv_mag_sq < 38) {
                                                                if (negBalance < 2) {
                                                                    if (AFD < 1) {
                                                                        if (mu_mv_mag_sq < 32) {
                                                                            return 1;
                                                                        }
                                                                        else if (mu_mv_mag_sq >= 32) {
                                                                            if (diffMVdiffVal < -3) {
                                                                                return 8;
                                                                            }
                                                                            else if (diffMVdiffVal >= -3) {
                                                                                if (MVDiff < 60) {
                                                                                    return 1;
                                                                                }
                                                                                else if (MVDiff >= 60) {
                                                                                    return 4;
                                                                                }
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (AFD >= 1) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (negBalance >= 2) {
                                                                    if (Rs < 5) {
                                                                        if (mcTcor < 48) {
                                                                            return 4;
                                                                        }
                                                                        else if (mcTcor >= 48) {
                                                                            if (diffMVdiffVal < -1) {
                                                                                return 4;
                                                                            }
                                                                            else if (diffMVdiffVal >= -1) {
                                                                                if (diffMVdiffVal < 14) {
                                                                                    return 8;
                                                                                }
                                                                                else if (diffMVdiffVal >= 14) {
                                                                                    return 4;
                                                                                }
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (Rs >= 5) {
                                                                        return 1;
                                                                    }
                                                                }
                                                            }
                                                            else if (mu_mv_mag_sq >= 38) {
                                                                return 1;
                                                            }
                                                        }
                                                        else if (MVDiff >= 94) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                                else if (MVDiff >= 140) {
                                                    if (posBalance < 105) {
                                                        if (ssDCval < 19) {
                                                            return 2;
                                                        }
                                                        else if (ssDCval >= 19) {
                                                            if (diffMVdiffVal < -816) {
                                                                return 2;
                                                            }
                                                            else if (diffMVdiffVal >= -816) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                    else if (posBalance >= 105) {
                                                        if (refDCval < 18) {
                                                            return 1;
                                                        }
                                                        else if (refDCval >= 18) {
                                                            if (diffMVdiffVal < 57) {
                                                                if (diffRsCsdiff < -7) {
                                                                    return 2;
                                                                }
                                                                else if (diffRsCsdiff >= -7) {
                                                                    return 1;
                                                                }
                                                            }
                                                            else if (diffMVdiffVal >= 57) {
                                                                if (AFD < 3) {
                                                                    return 4;
                                                                }
                                                                else if (AFD >= 3) {
                                                                    return 2;
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                            else if (mcTcor >= 61) {
                                                return 1;
                                            }
                                        }
                                        else if (tcor >= 58) {
                                            if (MVDiff < 46) {
                                                if (tcor < 80) {
                                                    if (mu_mv_mag_sq < 3) {
                                                        if (ssDCval < 88) {
                                                            if (ssDCval < 27) {
                                                                if (TSC < 1) {
                                                                    return 4;
                                                                }
                                                                else if (TSC >= 1) {
                                                                    return 2;
                                                                }
                                                            }
                                                            else if (ssDCval >= 27) {
                                                                return 2;
                                                            }
                                                        }
                                                        else if (ssDCval >= 88) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (mu_mv_mag_sq >= 3) {
                                                        if (posBalance < 85) {
                                                            if (ssDCval < 88) {
                                                                if (diffRsCsdiff < 1) {
                                                                    if (tcor < 62) {
                                                                        return 4;
                                                                    }
                                                                    else if (tcor >= 62) {
                                                                        return 2;
                                                                    }
                                                                }
                                                                else if (diffRsCsdiff >= 1) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (ssDCval >= 88) {
                                                                if (Cs < 286) {
                                                                    return 2;
                                                                }
                                                                else if (Cs >= 286) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                        else if (posBalance >= 85) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                                else if (tcor >= 80) {
                                                    if (mcTcor < 90) {
                                                        if (ssDCval < 32) {
                                                            return 8;
                                                        }
                                                        else if (ssDCval >= 32) {
                                                            if (mu_mv_mag_sq < 30) {
                                                                if (Cs < 47) {
                                                                    if (Rs < 71) {
                                                                        return 2;
                                                                    }
                                                                    else if (Rs >= 71) {
                                                                        return 8;
                                                                    }
                                                                }
                                                                else if (Cs >= 47) {
                                                                    return 2;
                                                                }
                                                            }
                                                            else if (mu_mv_mag_sq >= 30) {
                                                                if (MVDiff < 16) {
                                                                    return 8;
                                                                }
                                                                else if (MVDiff >= 16) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (mcTcor >= 90) {
                                                        return 2;
                                                    }
                                                }
                                            }
                                            else if (MVDiff >= 46) {
                                                if (negBalance < 5) {
                                                    if (mcTcor < 66) {
                                                        return 1;
                                                    }
                                                    else if (mcTcor >= 66) {
                                                        if (posBalance < 80) {
                                                            if (posBalance < 61) {
                                                                return 2;
                                                            }
                                                            else if (posBalance >= 61) {
                                                                if (diffMVdiffVal < -479) {
                                                                    return 2;
                                                                }
                                                                else if (diffMVdiffVal >= -479) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                        else if (posBalance >= 80) {
                                                            if (Cs < 7) {
                                                                if (posBalance < 90) {
                                                                    if (diffMVdiffVal < -71) {
                                                                        return 4;
                                                                    }
                                                                    else if (diffMVdiffVal >= -71) {
                                                                        return 8;
                                                                    }
                                                                }
                                                                else if (posBalance >= 90) {
                                                                    if (refDCval < 27) {
                                                                        return 8;
                                                                    }
                                                                    else if (refDCval >= 27) {
                                                                        if (Cs < 3) {
                                                                            return 4;
                                                                        }
                                                                        else if (Cs >= 3) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                            else if (Cs >= 7) {
                                                                if (posBalance < 117) {
                                                                    return 2;
                                                                }
                                                                else if (posBalance >= 117) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (negBalance >= 5) {
                                                    return 1;
                                                }
                                            }
                                        }
                                    }
                                    else if (Rs >= 230) {
                                        if (tcor < 85) {
                                            if (Rs < 405) {
                                                if (Cs < 293) {
                                                    if (Cs < 190) {
                                                        return 1;
                                                    }
                                                    else if (Cs >= 190) {
                                                        return 4;
                                                    }
                                                }
                                                else if (Cs >= 293) {
                                                    if (gchDC < 1) {
                                                        if (mcTcor < 77) {
                                                            if (Rs < 393) {
                                                                return 8;
                                                            }
                                                            else if (Rs >= 393) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (mcTcor >= 77) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (gchDC >= 1) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                            else if (Rs >= 405) {
                                                if (diffRsCsdiff < 1) {
                                                    if (mcTcor < 82) {
                                                        return 1;
                                                    }
                                                    else if (mcTcor >= 82) {
                                                        if (tcor < 82) {
                                                            return 4;
                                                        }
                                                        else if (tcor >= 82) {
                                                            if (ssDCval < 38) {
                                                                return 1;
                                                            }
                                                            else if (ssDCval >= 38) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (diffRsCsdiff >= 1) {
                                                    if (MVDiff < 19) {
                                                        return 4;
                                                    }
                                                    else if (MVDiff >= 19) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                        }
                                        else if (tcor >= 85) {
                                            if (refDCval < 84) {
                                                if (mcTcor < 86) {
                                                    if (AFD < 5) {
                                                        return 1;
                                                    }
                                                    else if (AFD >= 5) {
                                                        return 4;
                                                    }
                                                }
                                                else if (mcTcor >= 86) {
                                                    if (refDCval < 38) {
                                                        if (MVDiff < 18) {
                                                            return 4;
                                                        }
                                                        else if (MVDiff >= 18) {
                                                            if (Rs < 400) {
                                                                return 8;
                                                            }
                                                            else if (Rs >= 400) {
                                                                if (RsDiff < 1) {
                                                                    return 4;
                                                                }
                                                                else if (RsDiff >= 1) {
                                                                    return 2;
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (refDCval >= 38) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                            else if (refDCval >= 84) {
                                                if (MVDiff < 8) {
                                                    if (Cs < 292) {
                                                        if (posBalance < 46) {
                                                            return 1;
                                                        }
                                                        else if (posBalance >= 46) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (Cs >= 292) {
                                                        return 4;
                                                    }
                                                }
                                                else if (MVDiff >= 8) {
                                                    if (Rs < 777) {
                                                        return 1;
                                                    }
                                                    else if (Rs >= 777) {
                                                        return 2;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (Cs >= 374) {
                                    if (tcor < 85) {
                                        if (negBalance < 49) {
                                            if (refDCval < 24) {
                                                return 1;
                                            }
                                            else if (refDCval >= 24) {
                                                if (Rs < 461) {
                                                    if (Rs < 439) {
                                                        if (mcTcor < 81) {
                                                            if (Rs < 227) {
                                                                return 2;
                                                            }
                                                            else if (Rs >= 227) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (mcTcor >= 81) {
                                                            return 2;
                                                        }
                                                    }
                                                    else if (Rs >= 439) {
                                                        if (negBalance < 43) {
                                                            if (MVDiff < 15) {
                                                                return 4;
                                                            }
                                                            else if (MVDiff >= 15) {
                                                                if (mu_mv_mag_sq < 65) {
                                                                    return 2;
                                                                }
                                                                else if (mu_mv_mag_sq >= 65) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                        else if (negBalance >= 43) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                                else if (Rs >= 461) {
                                                    if (ssDCval < 39) {
                                                        if (SC < 870) {
                                                            return 4;
                                                        }
                                                        else if (SC >= 870) {
                                                            if (mu_mv_mag_sq < 73) {
                                                                return 2;
                                                            }
                                                            else if (mu_mv_mag_sq >= 73) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                    else if (ssDCval >= 39) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                        }
                                        else if (negBalance >= 49) {
                                            if (ssDCval < 62) {
                                                return 8;
                                            }
                                            else if (ssDCval >= 62) {
                                                return 2;
                                            }
                                        }
                                    }
                                    else if (tcor >= 85) {
                                        if (mu_mv_mag_sq < 61) {
                                            if (MVDiff < 6) {
                                                return 8;
                                            }
                                            else if (MVDiff >= 6) {
                                                if (SC < 1087) {
                                                    return 4;
                                                }
                                                else if (SC >= 1087) {
                                                    return 8;
                                                }
                                            }
                                        }
                                        else if (mu_mv_mag_sq >= 61) {
                                            if (Cs < 400) {
                                                return 4;
                                            }
                                            else if (Cs >= 400) {
                                                if (tcor < 97) {
                                                    if (refDCval < 39) {
                                                        return 2;
                                                    }
                                                    else if (refDCval >= 39) {
                                                        if (RsDiff < 2) {
                                                            if (diffTSC < 1) {
                                                                return 4;
                                                            }
                                                            else if (diffTSC >= 1) {
                                                                return 2;
                                                            }
                                                        }
                                                        else if (RsDiff >= 2) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                                else if (tcor >= 97) {
                                                    if (negBalance < 9) {
                                                        return 4;
                                                    }
                                                    else if (negBalance >= 9) {
                                                        if (posBalance < 25) {
                                                            if (Rs < 504) {
                                                                return 2;
                                                            }
                                                            else if (Rs >= 504) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (posBalance >= 25) {
                                                            return 2;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                            else if (SC >= 1148) {
                                if (refDCval < 55) {
                                    if (ssDCval < 39) {
                                        if (Rs < 1060) {
                                            if (mu_mv_mag_sq < 53) {
                                                if (TSC < 5) {
                                                    if (mcTcor < 76) {
                                                        if (Cs < 690) {
                                                            return 8;
                                                        }
                                                        else if (Cs >= 690) {
                                                            return 1;
                                                        }
                                                    }
                                                    else if (mcTcor >= 76) {
                                                        if (SC < 1376) {
                                                            return 4;
                                                        }
                                                        else if (SC >= 1376) {
                                                            return 2;
                                                        }
                                                    }
                                                }
                                                else if (TSC >= 5) {
                                                    if (Cs < 491) {
                                                        return 4;
                                                    }
                                                    else if (Cs >= 491) {
                                                        if (Rs < 804) {
                                                            if (posBalance < 56) {
                                                                return 4;
                                                            }
                                                            else if (posBalance >= 56) {
                                                                return 1;
                                                            }
                                                        }
                                                        else if (Rs >= 804) {
                                                            if (mcTcor < 77) {
                                                                return 4;
                                                            }
                                                            else if (mcTcor >= 77) {
                                                                if (diffMVdiffVal < 4) {
                                                                    return 1;
                                                                }
                                                                else if (diffMVdiffVal >= 4) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                            else if (mu_mv_mag_sq >= 53) {
                                                if (mcTcor < 77) {
                                                    return 1;
                                                }
                                                else if (mcTcor >= 77) {
                                                    if (refDCval < 30) {
                                                        return 2;
                                                    }
                                                    else if (refDCval >= 30) {
                                                        if (diffAFD < -2) {
                                                            return 1;
                                                        }
                                                        else if (diffAFD >= -2) {
                                                            if (MVDiff < 11) {
                                                                if (Cs < 972) {
                                                                    return 8;
                                                                }
                                                                else if (Cs >= 972) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (MVDiff >= 11) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (Rs >= 1060) {
                                            if (Rs < 1821) {
                                                if (negBalance < 45) {
                                                    return 8;
                                                }
                                                else if (negBalance >= 45) {
                                                    return 2;
                                                }
                                            }
                                            else if (Rs >= 1821) {
                                                return 1;
                                            }
                                        }
                                    }
                                    else if (ssDCval >= 39) {
                                        if (Rs < 1793) {
                                            if (Rs < 1200) {
                                                if (negBalance < 41) {
                                                    if (refDCval < 53) {
                                                        if (SC < 1342) {
                                                            if (Cs < 822) {
                                                                return 8;
                                                            }
                                                            else if (Cs >= 822) {
                                                                return 2;
                                                            }
                                                        }
                                                        else if (SC >= 1342) {
                                                            if (diffMVdiffVal < 14) {
                                                                if (diffAFD < -1) {
                                                                    if (mu_mv_mag_sq < 62) {
                                                                        return 1;
                                                                    }
                                                                    else if (mu_mv_mag_sq >= 62) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (diffAFD >= -1) {
                                                                    if (Rs < 1132) {
                                                                        return 4;
                                                                    }
                                                                    else if (Rs >= 1132) {
                                                                        if (Rs < 1147) {
                                                                            return 1;
                                                                        }
                                                                        else if (Rs >= 1147) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                            else if (diffMVdiffVal >= 14) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                    else if (refDCval >= 53) {
                                                        if (SC < 1515) {
                                                            return 8;
                                                        }
                                                        else if (SC >= 1515) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                                else if (negBalance >= 41) {
                                                    if (mcTcor < 81) {
                                                        if (posBalance < 59) {
                                                            return 4;
                                                        }
                                                        else if (posBalance >= 59) {
                                                            return 1;
                                                        }
                                                    }
                                                    else if (mcTcor >= 81) {
                                                        if (mu_mv_mag_sq < 81) {
                                                            if (SC < 1716) {
                                                                if (posBalance < 47) {
                                                                    return 4;
                                                                }
                                                                else if (posBalance >= 47) {
                                                                    if (diffRsCsdiff < -24) {
                                                                        return 2;
                                                                    }
                                                                    else if (diffRsCsdiff >= -24) {
                                                                        if (Cs < 601) {
                                                                            if (Rs < 845) {
                                                                                return 1;
                                                                            }
                                                                            else if (Rs >= 845) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                        else if (Cs >= 601) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                            else if (SC >= 1716) {
                                                                if (tcor < 89) {
                                                                    if (Rs < 1125) {
                                                                        return 8;
                                                                    }
                                                                    else if (Rs >= 1125) {
                                                                        return 1;
                                                                    }
                                                                }
                                                                else if (tcor >= 89) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                        else if (mu_mv_mag_sq >= 81) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (Rs >= 1200) {
                                                if (ssDCval < 44) {
                                                    if (tcor < 90) {
                                                        return 4;
                                                    }
                                                    else if (tcor >= 90) {
                                                        return 8;
                                                    }
                                                }
                                                else if (ssDCval >= 44) {
                                                    if (tcor < 93) {
                                                        if (SC < 2308) {
                                                            if (AFD < 6) {
                                                                if (posBalance < 48) {
                                                                    return 4;
                                                                }
                                                                else if (posBalance >= 48) {
                                                                    if (diffMVdiffVal < -1) {
                                                                        if (Cs < 930) {
                                                                            return 4;
                                                                        }
                                                                        else if (Cs >= 930) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                    else if (diffMVdiffVal >= -1) {
                                                                        return 1;
                                                                    }
                                                                }
                                                            }
                                                            else if (AFD >= 6) {
                                                                return 1;
                                                            }
                                                        }
                                                        else if (SC >= 2308) {
                                                            if (refDCval < 53) {
                                                                if (negBalance < 38) {
                                                                    if (mu_mv_mag_sq < 76) {
                                                                        return 1;
                                                                    }
                                                                    else if (mu_mv_mag_sq >= 76) {
                                                                        if (Rs < 1538) {
                                                                            return 4;
                                                                        }
                                                                        else if (Rs >= 1538) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                }
                                                                else if (negBalance >= 38) {
                                                                    if (AFD < 6) {
                                                                        return 4;
                                                                    }
                                                                    else if (AFD >= 6) {
                                                                        return 8;
                                                                    }
                                                                }
                                                            }
                                                            else if (refDCval >= 53) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                    else if (tcor >= 93) {
                                                        return 1;
                                                    }
                                                }
                                            }
                                        }
                                        else if (Rs >= 1793) {
                                            if (mu_mv_mag_sq < 63) {
                                                if (Rs < 1959) {
                                                    if (MVDiff < 17) {
                                                        return 1;
                                                    }
                                                    else if (MVDiff >= 17) {
                                                        if (Cs < 1542) {
                                                            if (mu_mv_mag_sq < 58) {
                                                                return 4;
                                                            }
                                                            else if (mu_mv_mag_sq >= 58) {
                                                                if (Rs < 1893) {
                                                                    return 1;
                                                                }
                                                                else if (Rs >= 1893) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                        else if (Cs >= 1542) {
                                                            if (refDCval < 40) {
                                                                if (gchDC < 1) {
                                                                    return 1;
                                                                }
                                                                else if (gchDC >= 1) {
                                                                    if (Rs < 1916) {
                                                                        return 1;
                                                                    }
                                                                    else if (Rs >= 1916) {
                                                                        return 4;
                                                                    }
                                                                }
                                                            }
                                                            else if (refDCval >= 40) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (Rs >= 1959) {
                                                    return 1;
                                                }
                                            }
                                            else if (mu_mv_mag_sq >= 63) {
                                                if (SC < 3340) {
                                                    return 8;
                                                }
                                                else if (SC >= 3340) {
                                                    if (posBalance < 53) {
                                                        return 1;
                                                    }
                                                    else if (posBalance >= 53) {
                                                        if (Cs < 1565) {
                                                            return 4;
                                                        }
                                                        else if (Cs >= 1565) {
                                                            if (mu_mv_mag_sq < 76) {
                                                                return 1;
                                                            }
                                                            else if (mu_mv_mag_sq >= 76) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (refDCval >= 55) {
                                    if (mu_mv_mag_sq < 84) {
                                        if (ssDCval < 72) {
                                            if (Rs < 2006) {
                                                if (Rs < 1478) {
                                                    if (refDCval < 64) {
                                                        if (mu_mv_mag_sq < 65) {
                                                            if (Cs < 692) {
                                                                return 8;
                                                            }
                                                            else if (Cs >= 692) {
                                                                return 1;
                                                            }
                                                        }
                                                        else if (mu_mv_mag_sq >= 65) {
                                                            if (Cs < 976) {
                                                                return 4;
                                                            }
                                                            else if (Cs >= 976) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                    else if (refDCval >= 64) {
                                                        return 4;
                                                    }
                                                }
                                                else if (Rs >= 1478) {
                                                    return 4;
                                                }
                                            }
                                            else if (Rs >= 2006) {
                                                return 2;
                                            }
                                        }
                                        else if (ssDCval >= 72) {
                                            if (mcTcor < 96) {
                                                if (negBalance < 37) {
                                                    return 8;
                                                }
                                                else if (negBalance >= 37) {
                                                    return 4;
                                                }
                                            }
                                            else if (mcTcor >= 96) {
                                                if (RsDiff < 2) {
                                                    return 4;
                                                }
                                                else if (RsDiff >= 2) {
                                                    if (TSC < 4) {
                                                        if (Rs < 2685) {
                                                            return 8;
                                                        }
                                                        else if (Rs >= 2685) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (TSC >= 4) {
                                                        if (MVDiff < 11) {
                                                            if (diffTSC < 1) {
                                                                if (diffMVdiffVal < 4) {
                                                                    return 4;
                                                                }
                                                                else if (diffMVdiffVal >= 4) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (diffTSC >= 1) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (MVDiff >= 11) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (mu_mv_mag_sq >= 84) {
                                        if (posBalance < 19) {
                                            if (SC < 1999) {
                                                if (refDCval < 140) {
                                                    if (Rs < 867) {
                                                        return 8;
                                                    }
                                                    else if (Rs >= 867) {
                                                        return 1;
                                                    }
                                                }
                                                else if (refDCval >= 140) {
                                                    return 1;
                                                }
                                            }
                                            else if (SC >= 1999) {
                                                return 4;
                                            }
                                        }
                                        else if (posBalance >= 19) {
                                            if (RsDiff < 2) {
                                                return 4;
                                            }
                                            else if (RsDiff >= 2) {
                                                if (posBalance < 53) {
                                                    if (Rs < 3606) {
                                                        return 8;
                                                    }
                                                    else if (Rs >= 3606) {
                                                        if (Rs < 3717) {
                                                            return 4;
                                                        }
                                                        else if (Rs >= 3717) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                                else if (posBalance >= 53) {
                                                    return 1;
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                        }
                    }
                    else if (mu_mv_mag_sq >= 100) {
                        if (Cs < 708) {
                            if (mcTcor < 99) {
                                if (posBalance < 3) {
                                    return 4;
                                }
                                else if (posBalance >= 3) {
                                    if (SC < 1726) {
                                        if (negBalance < 12) {
                                            if (negBalance < 1) {
                                                return 1;
                                            }
                                            else if (negBalance >= 1) {
                                                return 8;
                                            }
                                        }
                                        else if (negBalance >= 12) {
                                            if (posBalance < 42) {
                                                return 1;
                                            }
                                            else if (posBalance >= 42) {
                                                return 8;
                                            }
                                        }
                                    }
                                    else if (SC >= 1726) {
                                        if (TSC < 3) {
                                            return 4;
                                        }
                                        else if (TSC >= 3) {
                                            return 8;
                                        }
                                    }
                                }
                            }
                            else if (mcTcor >= 99) {
                                if (Cs < 365) {
                                    return 8;
                                }
                                else if (Cs >= 365) {
                                    return 4;
                                }
                            }
                        }
                        else if (Cs >= 708) {
                            if (posBalance < 10) {
                                if (Cs < 1490) {
                                    if (diffRsCsdiff < -23) {
                                        return 8;
                                    }
                                    else if (diffRsCsdiff >= -23) {
                                        return 1;
                                    }
                                }
                                else if (Cs >= 1490) {
                                    return 4;
                                }
                            }
                            else if (posBalance >= 10) {
                                if (MVDiff < 22) {
                                    if (Cs < 1321) {
                                        if (refDCval < 106) {
                                            return 8;
                                        }
                                        else if (refDCval >= 106) {
                                            if (diffTSC < -1) {
                                                return 1;
                                            }
                                            else if (diffTSC >= -1) {
                                                return 8;
                                            }
                                        }
                                    }
                                    else if (Cs >= 1321) {
                                        if (Rs < 1803) {
                                            return 8;
                                        }
                                        else if (Rs >= 1803) {
                                            if (Rs < 2418) {
                                                return 4;
                                            }
                                            else if (Rs >= 2418) {
                                                return 8;
                                            }
                                        }
                                    }
                                }
                                else if (MVDiff >= 22) {
                                    if (negBalance < 12) {
                                        return 4;
                                    }
                                    else if (negBalance >= 12) {
                                        if (SC < 3056) {
                                            return 8;
                                        }
                                        else if (SC >= 3056) {
                                            return 4;
                                        }
                                    }
                                }
                            }
                        }
                    }
                }
                else if (Cs >= 1670) {
                    if (mu_mv_mag_sq < 86) {
                        if (negBalance < 43) {
                            if (RsCsDiff < 69) {
                                if (Rs < 530) {
                                    return 4;
                                }
                                else if (Rs >= 530) {
                                    if (CsDiff < 5) {
                                        return 8;
                                    }
                                    else if (CsDiff >= 5) {
                                        if (posBalance < 45) {
                                            if (RsDiff < 5) {
                                                return 8;
                                            }
                                            else if (RsDiff >= 5) {
                                                return 4;
                                            }
                                        }
                                        else if (posBalance >= 45) {
                                            return 8;
                                        }
                                    }
                                }
                            }
                            else if (RsCsDiff >= 69) {
                                return 4;
                            }
                        }
                        else if (negBalance >= 43) {
                            if (CsDiff < 1) {
                                return 8;
                            }
                            else if (CsDiff >= 1) {
                                if (RsCsDiff < 16) {
                                    if (negBalance < 47) {
                                        return 4;
                                    }
                                    else if (negBalance >= 47) {
                                        return 8;
                                    }
                                }
                                else if (RsCsDiff >= 16) {
                                    return 4;
                                }
                            }
                        }
                    }
                    else if (mu_mv_mag_sq >= 86) {
                        if (diffRsCsdiff < -38) {
                            if (posBalance < 26) {
                                if (diffRsCsdiff < -63) {
                                    return 8;
                                }
                                else if (diffRsCsdiff >= -63) {
                                    return 1;
                                }
                            }
                            else if (posBalance >= 26) {
                                if (TSC < 2) {
                                    return 2;
                                }
                                else if (TSC >= 2) {
                                    if (Cs < 2438) {
                                        if (MVDiff < 8) {
                                            return 8;
                                        }
                                        else if (MVDiff >= 8) {
                                            if (Cs < 1850) {
                                                return 8;
                                            }
                                            else if (Cs >= 1850) {
                                                return 4;
                                            }
                                        }
                                    }
                                    else if (Cs >= 2438) {
                                        return 8;
                                    }
                                }
                            }
                        }
                        else if (diffRsCsdiff >= -38) {
                            return 8;
                        }
                    }
                }
            }
        }
        else if (AFD >= 7) {
            if (Rs < 4380) {
                if (MVDiff < 4) {
                    if (SC < 3686) {
                        if (tcor < 86) {
                            if (AFD < 9) {
                                if (Cs < 1415) {
                                    if (Cs < 947) {
                                        return 1;
                                    }
                                    else if (Cs >= 947) {
                                        if (SC < 2820) {
                                            return 4;
                                        }
                                        else if (SC >= 2820) {
                                            return 8;
                                        }
                                    }
                                }
                                else if (Cs >= 1415) {
                                    return 8;
                                }
                            }
                            else if (AFD >= 9) {
                                if (mu_mv_mag_sq < 3) {
                                    if (posBalance < 63) {
                                        return 4;
                                    }
                                    else if (posBalance >= 63) {
                                        return 1;
                                    }
                                }
                                else if (mu_mv_mag_sq >= 3) {
                                    if (RsDiff < 6) {
                                        if (AFD < 11) {
                                            return 8;
                                        }
                                        else if (AFD >= 11) {
                                            return 4;
                                        }
                                    }
                                    else if (RsDiff >= 6) {
                                        return 2;
                                    }
                                }
                            }
                        }
                        else if (tcor >= 86) {
                            if (posBalance < 54) {
                                if (mu_mv_mag_sq < 80) {
                                    if (negBalance < 44) {
                                        return 2;
                                    }
                                    else if (negBalance >= 44) {
                                        return 8;
                                    }
                                }
                                else if (mu_mv_mag_sq >= 80) {
                                    if (gchDC < 2) {
                                        return 8;
                                    }
                                    else if (gchDC >= 2) {
                                        return 4;
                                    }
                                }
                            }
                            else if (posBalance >= 54) {
                                if (negBalance < 47) {
                                    if (mu_mv_mag_sq < 113) {
                                        if (Cs < 1811) {
                                            if (ssDCval < 127) {
                                                return 2;
                                            }
                                            else if (ssDCval >= 127) {
                                                return 8;
                                            }
                                        }
                                        else if (Cs >= 1811) {
                                            if (SC < 3224) {
                                                return 8;
                                            }
                                            else if (SC >= 3224) {
                                                return 4;
                                            }
                                        }
                                    }
                                    else if (mu_mv_mag_sq >= 113) {
                                        if (Rs < 1796) {
                                            return 8;
                                        }
                                        else if (Rs >= 1796) {
                                            if (mcTcor < 94) {
                                                if (Cs < 1661) {
                                                    return 8;
                                                }
                                                else if (Cs >= 1661) {
                                                    return 4;
                                                }
                                            }
                                            else if (mcTcor >= 94) {
                                                return 2;
                                            }
                                        }
                                    }
                                }
                                else if (negBalance >= 47) {
                                    if (AFD < 8) {
                                        if (RsCsDiff < 37) {
                                            if (mu_mv_mag_sq < 116) {
                                                return 2;
                                            }
                                            else if (mu_mv_mag_sq >= 116) {
                                                if (Rs < 1828) {
                                                    return 2;
                                                }
                                                else if (Rs >= 1828) {
                                                    if (mu_mv_mag_sq < 122) {
                                                        return 4;
                                                    }
                                                    else if (mu_mv_mag_sq >= 122) {
                                                        return 2;
                                                    }
                                                }
                                            }
                                        }
                                        else if (RsCsDiff >= 37) {
                                            if (diffRsCsdiff < 5) {
                                                return 8;
                                            }
                                            else if (diffRsCsdiff >= 5) {
                                                return 2;
                                            }
                                        }
                                    }
                                    else if (AFD >= 8) {
                                        return 2;
                                    }
                                }
                            }
                        }
                    }
                    else if (SC >= 3686) {
                        if (ssDCval < 97) {
                            if (TSC < 7) {
                                if (negBalance < 28) {
                                    if (SC < 4020) {
                                        return 8;
                                    }
                                    else if (SC >= 4020) {
                                        return 1;
                                    }
                                }
                                else if (negBalance >= 28) {
                                    if (mu_mv_mag_sq < 116) {
                                        if (SC < 5007) {
                                            if (diffAFD < -1) {
                                                if (RsCsDiff < 46) {
                                                    if (Cs < 1754) {
                                                        return 1;
                                                    }
                                                    else if (Cs >= 1754) {
                                                        return 2;
                                                    }
                                                }
                                                else if (RsCsDiff >= 46) {
                                                    return 4;
                                                }
                                            }
                                            else if (diffAFD >= -1) {
                                                if (Rs < 1783) {
                                                    return 8;
                                                }
                                                else if (Rs >= 1783) {
                                                    return 4;
                                                }
                                            }
                                        }
                                        else if (SC >= 5007) {
                                            return 8;
                                        }
                                    }
                                    else if (mu_mv_mag_sq >= 116) {
                                        return 8;
                                    }
                                }
                            }
                            else if (TSC >= 7) {
                                if (RsCsDiff < 154) {
                                    if (mu_mv_mag_sq < 106) {
                                        if (posBalance < 58) {
                                            if (Cs < 2247) {
                                                if (negBalance < 48) {
                                                    if (Rs < 2348) {
                                                        return 8;
                                                    }
                                                    else if (Rs >= 2348) {
                                                        if (SC < 4297) {
                                                            return 8;
                                                        }
                                                        else if (SC >= 4297) {
                                                            return 1;
                                                        }
                                                    }
                                                }
                                                else if (negBalance >= 48) {
                                                    if (TSC < 12) {
                                                        return 2;
                                                    }
                                                    else if (TSC >= 12) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                            else if (Cs >= 2247) {
                                                if (SC < 6096) {
                                                    return 4;
                                                }
                                                else if (SC >= 6096) {
                                                    return 2;
                                                }
                                            }
                                        }
                                        else if (posBalance >= 58) {
                                            if (tcor < 87) {
                                                if (diffMVdiffVal < -5) {
                                                    return 2;
                                                }
                                                else if (diffMVdiffVal >= -5) {
                                                    if (Rs < 2292) {
                                                        return 8;
                                                    }
                                                    else if (Rs >= 2292) {
                                                        if (Rs < 2667) {
                                                            return 2;
                                                        }
                                                        else if (Rs >= 2667) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (tcor >= 87) {
                                                if (SC < 4496) {
                                                    if (mu_mv_mag_sq < 80) {
                                                        if (AFD < 10) {
                                                            if (Cs < 1732) {
                                                                return 1;
                                                            }
                                                            else if (Cs >= 1732) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (AFD >= 10) {
                                                            if (Rs < 2147) {
                                                                return 2;
                                                            }
                                                            else if (Rs >= 2147) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                    else if (mu_mv_mag_sq >= 80) {
                                                        return 2;
                                                    }
                                                }
                                                else if (SC >= 4496) {
                                                    return 2;
                                                }
                                            }
                                        }
                                    }
                                    else if (mu_mv_mag_sq >= 106) {
                                        if (Cs < 2785) {
                                            return 8;
                                        }
                                        else if (Cs >= 2785) {
                                            if (Cs < 3834) {
                                                return 2;
                                            }
                                            else if (Cs >= 3834) {
                                                return 8;
                                            }
                                        }
                                    }
                                }
                                else if (RsCsDiff >= 154) {
                                    if (Cs < 2626) {
                                        return 4;
                                    }
                                    else if (Cs >= 2626) {
                                        return 8;
                                    }
                                }
                            }
                        }
                        else if (ssDCval >= 97) {
                            if (SC < 5160) {
                                if (mu_mv_mag_sq < 31) {
                                    if (Rs < 2990) {
                                        if (Cs < 1932) {
                                            return 4;
                                        }
                                        else if (Cs >= 1932) {
                                            if (diffAFD < 1) {
                                                if (ssDCval < 124) {
                                                    return 4;
                                                }
                                                else if (ssDCval >= 124) {
                                                    return 8;
                                                }
                                            }
                                            else if (diffAFD >= 1) {
                                                return 8;
                                            }
                                        }
                                    }
                                    else if (Rs >= 2990) {
                                        return 2;
                                    }
                                }
                                else if (mu_mv_mag_sq >= 31) {
                                    if (refDCval < 109) {
                                        if (SC < 4731) {
                                            return 4;
                                        }
                                        else if (SC >= 4731) {
                                            return 8;
                                        }
                                    }
                                    else if (refDCval >= 109) {
                                        if (tcor < 92) {
                                            if (mu_mv_mag_sq < 70) {
                                                if (TSC < 8) {
                                                    return 4;
                                                }
                                                else if (TSC >= 8) {
                                                    if (SC < 4805) {
                                                        return 8;
                                                    }
                                                    else if (SC >= 4805) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                            else if (mu_mv_mag_sq >= 70) {
                                                return 8;
                                            }
                                        }
                                        else if (tcor >= 92) {
                                            if (refDCval < 115) {
                                                return 8;
                                            }
                                            else if (refDCval >= 115) {
                                                if (TSC < 8) {
                                                    return 4;
                                                }
                                                else if (TSC >= 8) {
                                                    return 8;
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                            else if (SC >= 5160) {
                                if (TSC < 11) {
                                    if (Rs < 3497) {
                                        if (negBalance < 48) {
                                            return 8;
                                        }
                                        else if (negBalance >= 48) {
                                            if (diffRsCsdiff < -45) {
                                                return 4;
                                            }
                                            else if (diffRsCsdiff >= -45) {
                                                if (diffMVdiffVal < 1) {
                                                    return 8;
                                                }
                                                else if (diffMVdiffVal >= 1) {
                                                    if (Cs < 3035) {
                                                        return 8;
                                                    }
                                                    else if (Cs >= 3035) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (Rs >= 3497) {
                                        if (SC < 7383) {
                                            if (RsDiff < 7) {
                                                return 8;
                                            }
                                            else if (RsDiff >= 7) {
                                                if (negBalance < 41) {
                                                    return 8;
                                                }
                                                else if (negBalance >= 41) {
                                                    return 4;
                                                }
                                            }
                                        }
                                        else if (SC >= 7383) {
                                            return 8;
                                        }
                                    }
                                }
                                else if (TSC >= 11) {
                                    if (Cs < 3065) {
                                        if (SC < 6615) {
                                            if (CsDiff < 8) {
                                                return 1;
                                            }
                                            else if (CsDiff >= 8) {
                                                return 8;
                                            }
                                        }
                                        else if (SC >= 6615) {
                                            return 4;
                                        }
                                    }
                                    else if (Cs >= 3065) {
                                        return 8;
                                    }
                                }
                            }
                        }
                    }
                }
                else if (MVDiff >= 4) {
                    if (MVDiff < 17) {
                        if (refDCval < 133) {
                            if (TSC < 8) {
                                if (Rs < 2838) {
                                    if (negBalance < 50) {
                                        if (tcor < 83) {
                                            if (RsCsDiff < 77) {
                                                if (refDCval < 114) {
                                                    if (tcor < 59) {
                                                        return 1;
                                                    }
                                                    else if (tcor >= 59) {
                                                        if (SC < 2589) {
                                                            if (RsCsDiff < 23) {
                                                                if (Rs < 626) {
                                                                    if (ssDCval < 88) {
                                                                        return 2;
                                                                    }
                                                                    else if (ssDCval >= 88) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (Rs >= 626) {
                                                                    if (Rs < 1087) {
                                                                        return 4;
                                                                    }
                                                                    else if (Rs >= 1087) {
                                                                        if (Cs < 1061) {
                                                                            return 8;
                                                                        }
                                                                        else if (Cs >= 1061) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                            else if (RsCsDiff >= 23) {
                                                                if (AFD < 15) {
                                                                    if (TSCindex < 6) {
                                                                        if (Cs < 1199) {
                                                                            if (Rs < 1429) {
                                                                                return 4;
                                                                            }
                                                                            else if (Rs >= 1429) {
                                                                                return 2;
                                                                            }
                                                                        }
                                                                        else if (Cs >= 1199) {
                                                                            return 2;
                                                                        }
                                                                    }
                                                                    else if (TSCindex >= 6) {
                                                                        if (Cs < 567) {
                                                                            return 2;
                                                                        }
                                                                        else if (Cs >= 567) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                }
                                                                else if (AFD >= 15) {
                                                                    return 2;
                                                                }
                                                            }
                                                        }
                                                        else if (SC >= 2589) {
                                                            return 2;
                                                        }
                                                    }
                                                }
                                                else if (refDCval >= 114) {
                                                    if (Cs < 539) {
                                                        return 8;
                                                    }
                                                    else if (Cs >= 539) {
                                                        if (Cs < 1457) {
                                                            return 1;
                                                        }
                                                        else if (Cs >= 1457) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (RsCsDiff >= 77) {
                                                if (mu_mv_mag_sq < 10) {
                                                    if (mcTcor < 89) {
                                                        if (Cs < 1985) {
                                                            if (posBalance < 50) {
                                                                return 4;
                                                            }
                                                            else if (posBalance >= 50) {
                                                                return 1;
                                                            }
                                                        }
                                                        else if (Cs >= 1985) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (mcTcor >= 89) {
                                                        if (mcTcor < 93) {
                                                            return 4;
                                                        }
                                                        else if (mcTcor >= 93) {
                                                            if (Cs < 1740) {
                                                                if (Rs < 783) {
                                                                    return 2;
                                                                }
                                                                else if (Rs >= 783) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (Cs >= 1740) {
                                                                return 1;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (mu_mv_mag_sq >= 10) {
                                                    if (SC < 4628) {
                                                        return 1;
                                                    }
                                                    else if (SC >= 4628) {
                                                        if (Rs < 2280) {
                                                            return 1;
                                                        }
                                                        else if (Rs >= 2280) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (tcor >= 83) {
                                            if (Cs < 793) {
                                                if (diffMVdiffVal < -8) {
                                                    if (MVDiff < 12) {
                                                        if (Rs < 521) {
                                                            return 4;
                                                        }
                                                        else if (Rs >= 521) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (MVDiff >= 12) {
                                                        return 1;
                                                    }
                                                }
                                                else if (diffMVdiffVal >= -8) {
                                                    if (mcTcor < 95) {
                                                        if (ssDCval < 69) {
                                                            if (AFD < 8) {
                                                                if (ssDCval < 58) {
                                                                    return 4;
                                                                }
                                                                else if (ssDCval >= 58) {
                                                                    return 1;
                                                                }
                                                            }
                                                            else if (AFD >= 8) {
                                                                return 2;
                                                            }
                                                        }
                                                        else if (ssDCval >= 69) {
                                                            if (MVDiff < 7) {
                                                                if (RsDiff < 4) {
                                                                    return 4;
                                                                }
                                                                else if (RsDiff >= 4) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (MVDiff >= 7) {
                                                                if (Rs < 278) {
                                                                    return 2;
                                                                }
                                                                else if (Rs >= 278) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (mcTcor >= 95) {
                                                        if (Rs < 690) {
                                                            if (Cs < 359) {
                                                                return 2;
                                                            }
                                                            else if (Cs >= 359) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (Rs >= 690) {
                                                            if (posBalance < 38) {
                                                                return 4;
                                                            }
                                                            else if (posBalance >= 38) {
                                                                if (gchDC < 3) {
                                                                    return 2;
                                                                }
                                                                else if (gchDC >= 3) {
                                                                    if (RsDiff < 5) {
                                                                        return 2;
                                                                    }
                                                                    else if (RsDiff >= 5) {
                                                                        return 4;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                            else if (Cs >= 793) {
                                                if (refDCval < 90) {
                                                    if (posBalance < 58) {
                                                        if (refDCval < 57) {
                                                            if (refDCval < 50) {
                                                                if (mu_mv_mag_sq < 38) {
                                                                    return 1;
                                                                }
                                                                else if (mu_mv_mag_sq >= 38) {
                                                                    if (Rs < 2545) {
                                                                        return 4;
                                                                    }
                                                                    else if (Rs >= 2545) {
                                                                        if (Rs < 2685) {
                                                                            return 8;
                                                                        }
                                                                        else if (Rs >= 2685) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                            else if (refDCval >= 50) {
                                                                if (posBalance < 54) {
                                                                    if (CsDiff < 3) {
                                                                        return 8;
                                                                    }
                                                                    else if (CsDiff >= 3) {
                                                                        if (SC < 2284) {
                                                                            return 1;
                                                                        }
                                                                        else if (SC >= 2284) {
                                                                            if (diffMVdiffVal < -2) {
                                                                                return 1;
                                                                            }
                                                                            else if (diffMVdiffVal >= -2) {
                                                                                return 8;
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                                else if (posBalance >= 54) {
                                                                    if (TSC < 6) {
                                                                        if (diffAFD < 1) {
                                                                            if (ssDCval < 50) {
                                                                                return 8;
                                                                            }
                                                                            else if (ssDCval >= 50) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                        else if (diffAFD >= 1) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                    else if (TSC >= 6) {
                                                                        return 8;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (refDCval >= 57) {
                                                            if (SC < 3765) {
                                                                if (AFD < 11) {
                                                                    return 4;
                                                                }
                                                                else if (AFD >= 11) {
                                                                    return 2;
                                                                }
                                                            }
                                                            else if (SC >= 3765) {
                                                                if (tcor < 99) {
                                                                    return 1;
                                                                }
                                                                else if (tcor >= 99) {
                                                                    if (Cs < 1816) {
                                                                        return 4;
                                                                    }
                                                                    else if (Cs >= 1816) {
                                                                        return 8;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (posBalance >= 58) {
                                                        if (ssDCval < 75) {
                                                            if (tcor < 88) {
                                                                if (Cs < 842) {
                                                                    return 8;
                                                                }
                                                                else if (Cs >= 842) {
                                                                    if (Rs < 1178) {
                                                                        return 4;
                                                                    }
                                                                    else if (Rs >= 1178) {
                                                                        return 1;
                                                                    }
                                                                }
                                                            }
                                                            else if (tcor >= 88) {
                                                                return 1;
                                                            }
                                                        }
                                                        else if (ssDCval >= 75) {
                                                            if (SC < 3058) {
                                                                return 4;
                                                            }
                                                            else if (SC >= 3058) {
                                                                if (mu_mv_mag_sq < 72) {
                                                                    return 4;
                                                                }
                                                                else if (mu_mv_mag_sq >= 72) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (refDCval >= 90) {
                                                    if (negBalance < 40) {
                                                        if (Rs < 1253) {
                                                            if (SC < 2531) {
                                                                return 1;
                                                            }
                                                            else if (SC >= 2531) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (Rs >= 1253) {
                                                            if (TSC < 6) {
                                                                return 4;
                                                            }
                                                            else if (TSC >= 6) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                    else if (negBalance >= 40) {
                                                        if (Cs < 2683) {
                                                            if (RsCsDiff < 37) {
                                                                return 1;
                                                            }
                                                            else if (RsCsDiff >= 37) {
                                                                if (posBalance < 42) {
                                                                    return 8;
                                                                }
                                                                else if (posBalance >= 42) {
                                                                    if (Cs < 2331) {
                                                                        return 4;
                                                                    }
                                                                    else if (Cs >= 2331) {
                                                                        if (Cs < 2435) {
                                                                            return 8;
                                                                        }
                                                                        else if (Cs >= 2435) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (Cs >= 2683) {
                                                            return 2;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (negBalance >= 50) {
                                        if (ssDCval < 98) {
                                            if (RsCsDiff < 31) {
                                                if (RsDiff < 3) {
                                                    if (TSC < 6) {
                                                        if (Rs < 654) {
                                                            if (MVDiff < 9) {
                                                                if (mcTcor < 90) {
                                                                    if (Rs < 620) {
                                                                        return 8;
                                                                    }
                                                                    else if (Rs >= 620) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (mcTcor >= 90) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (MVDiff >= 9) {
                                                                if (Cs < 772) {
                                                                    if (Rs < 589) {
                                                                        return 8;
                                                                    }
                                                                    else if (Rs >= 589) {
                                                                        if (tcor < 78) {
                                                                            return 8;
                                                                        }
                                                                        else if (tcor >= 78) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                }
                                                                else if (Cs >= 772) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                        else if (Rs >= 654) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (TSC >= 6) {
                                                        if (Cs < 706) {
                                                            return 1;
                                                        }
                                                        else if (Cs >= 706) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                                else if (RsDiff >= 3) {
                                                    if (Cs < 582) {
                                                        if (AFD < 10) {
                                                            return 8;
                                                        }
                                                        else if (AFD >= 10) {
                                                            return 2;
                                                        }
                                                    }
                                                    else if (Cs >= 582) {
                                                        if (Rs < 1156) {
                                                            return 4;
                                                        }
                                                        else if (Rs >= 1156) {
                                                            if (SC < 2633) {
                                                                if (SC < 2306) {
                                                                    return 1;
                                                                }
                                                                else if (SC >= 2306) {
                                                                    return 2;
                                                                }
                                                            }
                                                            else if (SC >= 2633) {
                                                                if (SC < 2792) {
                                                                    return 8;
                                                                }
                                                                else if (SC >= 2792) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                            else if (RsCsDiff >= 31) {
                                                if (posBalance < 61) {
                                                    if (Cs < 838) {
                                                        if (AFD < 11) {
                                                            return 4;
                                                        }
                                                        else if (AFD >= 11) {
                                                            return 2;
                                                        }
                                                    }
                                                    else if (Cs >= 838) {
                                                        if (SC < 3313) {
                                                            if (negBalance < 56) {
                                                                if (RsCsDiff < 56) {
                                                                    if (posBalance < 60) {
                                                                        return 8;
                                                                    }
                                                                    else if (posBalance >= 60) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (RsCsDiff >= 56) {
                                                                    return 1;
                                                                }
                                                            }
                                                            else if (negBalance >= 56) {
                                                                return 1;
                                                            }
                                                        }
                                                        else if (SC >= 3313) {
                                                            if (refDCval < 96) {
                                                                if (diffRsCsdiff < 1) {
                                                                    if (refDCval < 92) {
                                                                        return 8;
                                                                    }
                                                                    else if (refDCval >= 92) {
                                                                        if (negBalance < 53) {
                                                                            if (diffMVdiffVal < -1) {
                                                                                return 8;
                                                                            }
                                                                            else if (diffMVdiffVal >= -1) {
                                                                                return 2;
                                                                            }
                                                                        }
                                                                        else if (negBalance >= 53) {
                                                                            return 2;
                                                                        }
                                                                    }
                                                                }
                                                                else if (diffRsCsdiff >= 1) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (refDCval >= 96) {
                                                                return 2;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (posBalance >= 61) {
                                                    if (refDCval < 83) {
                                                        return 2;
                                                    }
                                                    else if (refDCval >= 83) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                        }
                                        else if (ssDCval >= 98) {
                                            if (SC < 1083) {
                                                if (SC < 755) {
                                                    if (AFD < 9) {
                                                        return 4;
                                                    }
                                                    else if (AFD >= 9) {
                                                        if (Rs < 620) {
                                                            return 2;
                                                        }
                                                        else if (Rs >= 620) {
                                                            return 1;
                                                        }
                                                    }
                                                }
                                                else if (SC >= 755) {
                                                    return 8;
                                                }
                                            }
                                            else if (SC >= 1083) {
                                                if (SC < 2174) {
                                                    if (ssDCval < 105) {
                                                        return 4;
                                                    }
                                                    else if (ssDCval >= 105) {
                                                        return 8;
                                                    }
                                                }
                                                else if (SC >= 2174) {
                                                    if (TSC < 6) {
                                                        if (Rs < 2348) {
                                                            if (AFD < 9) {
                                                                return 2;
                                                            }
                                                            else if (AFD >= 9) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (Rs >= 2348) {
                                                            return 2;
                                                        }
                                                    }
                                                    else if (TSC >= 6) {
                                                        if (SC < 4405) {
                                                            if (AFD < 9) {
                                                                return 2;
                                                            }
                                                            else if (AFD >= 9) {
                                                                if (RsDiff < 10) {
                                                                    if (ssDCval < 99) {
                                                                        return 2;
                                                                    }
                                                                    else if (ssDCval >= 99) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (RsDiff >= 10) {
                                                                    return 1;
                                                                }
                                                            }
                                                        }
                                                        else if (SC >= 4405) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (Rs >= 2838) {
                                    if (refDCval < 110) {
                                        if (TSCindex < 8) {
                                            if (mu_mv_mag_sq < 75) {
                                                if (refDCval < 108) {
                                                    if (refDCval < 54) {
                                                        if (MVDiff < 5) {
                                                            return 8;
                                                        }
                                                        else if (MVDiff >= 5) {
                                                            if (diffMVdiffVal < 5) {
                                                                return 4;
                                                            }
                                                            else if (diffMVdiffVal >= 5) {
                                                                if (MVDiff < 14) {
                                                                    return 4;
                                                                }
                                                                else if (MVDiff >= 14) {
                                                                    return 2;
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (refDCval >= 54) {
                                                        if (posBalance < 46) {
                                                            return 4;
                                                        }
                                                        else if (posBalance >= 46) {
                                                            if (diffAFD < 1) {
                                                                if (Cs < 3484) {
                                                                    if (mcTcor < 96) {
                                                                        if (Cs < 1505) {
                                                                            return 4;
                                                                        }
                                                                        else if (Cs >= 1505) {
                                                                            if (mcTcor < 94) {
                                                                                return 4;
                                                                            }
                                                                            else if (mcTcor >= 94) {
                                                                                return 8;
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (mcTcor >= 96) {
                                                                        if (tcor < 95) {
                                                                            return 4;
                                                                        }
                                                                        else if (tcor >= 95) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                }
                                                                else if (Cs >= 3484) {
                                                                    if (CsDiff < 4) {
                                                                        return 4;
                                                                    }
                                                                    else if (CsDiff >= 4) {
                                                                        return 8;
                                                                    }
                                                                }
                                                            }
                                                            else if (diffAFD >= 1) {
                                                                if (SC < 7980) {
                                                                    return 8;
                                                                }
                                                                else if (SC >= 7980) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (refDCval >= 108) {
                                                    return 4;
                                                }
                                            }
                                            else if (mu_mv_mag_sq >= 75) {
                                                if (MVDiff < 11) {
                                                    if (SC < 7369) {
                                                        return 8;
                                                    }
                                                    else if (SC >= 7369) {
                                                        if (diffRsCsdiff < 3) {
                                                            return 4;
                                                        }
                                                        else if (diffRsCsdiff >= 3) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                                else if (MVDiff >= 11) {
                                                    return 4;
                                                }
                                            }
                                        }
                                        else if (TSCindex >= 8) {
                                            if (RsCsDiff < 29) {
                                                if (SC < 7534) {
                                                    if (diffMVdiffVal < -4) {
                                                        return 2;
                                                    }
                                                    else if (diffMVdiffVal >= -4) {
                                                        if (CsDiff < 4) {
                                                            return 4;
                                                        }
                                                        else if (CsDiff >= 4) {
                                                            return 2;
                                                        }
                                                    }
                                                }
                                                else if (SC >= 7534) {
                                                    return 4;
                                                }
                                            }
                                            else if (RsCsDiff >= 29) {
                                                if (ssDCval < 92) {
                                                    return 4;
                                                }
                                                else if (ssDCval >= 92) {
                                                    if (RsDiff < 8) {
                                                        if (MVDiff < 12) {
                                                            return 8;
                                                        }
                                                        else if (MVDiff >= 12) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (RsDiff >= 8) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (refDCval >= 110) {
                                        if (refDCval < 118) {
                                            if (refDCval < 111) {
                                                if (Rs < 3605) {
                                                    return 2;
                                                }
                                                else if (Rs >= 3605) {
                                                    return 4;
                                                }
                                            }
                                            else if (refDCval >= 111) {
                                                if (AFD < 8) {
                                                    if (Rs < 3621) {
                                                        return 4;
                                                    }
                                                    else if (Rs >= 3621) {
                                                        return 2;
                                                    }
                                                }
                                                else if (AFD >= 8) {
                                                    return 2;
                                                }
                                            }
                                        }
                                        else if (refDCval >= 118) {
                                            if (RsCsDiff < 46) {
                                                if (TSCindex < 7) {
                                                    return 4;
                                                }
                                                else if (TSCindex >= 7) {
                                                    return 8;
                                                }
                                            }
                                            else if (RsCsDiff >= 46) {
                                                if (mu_mv_mag_sq < 100) {
                                                    return 4;
                                                }
                                                else if (mu_mv_mag_sq >= 100) {
                                                    if (TSC < 7) {
                                                        return 8;
                                                    }
                                                    else if (TSC >= 7) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                            else if (TSC >= 8) {
                                if (mu_mv_mag_sq < 46) {
                                    if (refDCval < 124) {
                                        if (mcTcor < 85) {
                                            if (mu_mv_mag_sq < 2) {
                                                if (TSC < 9) {
                                                    if (MVDiff < 13) {
                                                        if (diffAFD < 1) {
                                                            return 8;
                                                        }
                                                        else if (diffAFD >= 1) {
                                                            if (Rs < 2288) {
                                                                return 2;
                                                            }
                                                            else if (Rs >= 2288) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                    else if (MVDiff >= 13) {
                                                        return 4;
                                                    }
                                                }
                                                else if (TSC >= 9) {
                                                    if (negBalance < 58) {
                                                        return 4;
                                                    }
                                                    else if (negBalance >= 58) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                            else if (mu_mv_mag_sq >= 2) {
                                                if (negBalance < 38) {
                                                    return 1;
                                                }
                                                else if (negBalance >= 38) {
                                                    if (ssDCval < 85) {
                                                        if (AFD < 12) {
                                                            return 4;
                                                        }
                                                        else if (AFD >= 12) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (ssDCval >= 85) {
                                                        if (posBalance < 47) {
                                                            return 8;
                                                        }
                                                        else if (posBalance >= 47) {
                                                            if (AFD < 24) {
                                                                return 4;
                                                            }
                                                            else if (AFD >= 24) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (mcTcor >= 85) {
                                            if (negBalance < 46) {
                                                if (posBalance < 56) {
                                                    if (ssDCval < 93) {
                                                        if (posBalance < 47) {
                                                            return 8;
                                                        }
                                                        else if (posBalance >= 47) {
                                                            if (tcor < 80) {
                                                                return 1;
                                                            }
                                                            else if (tcor >= 80) {
                                                                if (MVDiff < 12) {
                                                                    return 1;
                                                                }
                                                                else if (MVDiff >= 12) {
                                                                    if (mu_mv_mag_sq < 19) {
                                                                        if (refDCval < 80) {
                                                                            return 8;
                                                                        }
                                                                        else if (refDCval >= 80) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                    else if (mu_mv_mag_sq >= 19) {
                                                                        if (posBalance < 49) {
                                                                            return 1;
                                                                        }
                                                                        else if (posBalance >= 49) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (ssDCval >= 93) {
                                                        if (AFD < 17) {
                                                            return 8;
                                                        }
                                                        else if (AFD >= 17) {
                                                            return 2;
                                                        }
                                                    }
                                                }
                                                else if (posBalance >= 56) {
                                                    if (Scindex < 5) {
                                                        return 2;
                                                    }
                                                    else if (Scindex >= 5) {
                                                        if (SC < 4428) {
                                                            return 1;
                                                        }
                                                        else if (SC >= 4428) {
                                                            if (Cs < 2428) {
                                                                return 4;
                                                            }
                                                            else if (Cs >= 2428) {
                                                                if (Rs < 2850) {
                                                                    return 8;
                                                                }
                                                                else if (Rs >= 2850) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                            else if (negBalance >= 46) {
                                                if (ssDCval < 115) {
                                                    if (ssDCval < 103) {
                                                        if (refDCval < 66) {
                                                            if (RsCsDiff < 46) {
                                                                if (Cs < 3801) {
                                                                    if (Rs < 2533) {
                                                                        return 8;
                                                                    }
                                                                    else if (Rs >= 2533) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (Cs >= 3801) {
                                                                    return 2;
                                                                }
                                                            }
                                                            else if (RsCsDiff >= 46) {
                                                                if (tcor < 78) {
                                                                    return 8;
                                                                }
                                                                else if (tcor >= 78) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                        else if (refDCval >= 66) {
                                                            if (tcor < 90) {
                                                                if (RsDiff < 6) {
                                                                    if (Cs < 1170) {
                                                                        return 1;
                                                                    }
                                                                    else if (Cs >= 1170) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (RsDiff >= 6) {
                                                                    if (RsCsDiff < 99) {
                                                                        if (Cs < 910) {
                                                                            return 2;
                                                                        }
                                                                        else if (Cs >= 910) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                    else if (RsCsDiff >= 99) {
                                                                        if (SC < 5093) {
                                                                            if (negBalance < 50) {
                                                                                if (Cs < 1713) {
                                                                                    return 1;
                                                                                }
                                                                                else if (Cs >= 1713) {
                                                                                    return 4;
                                                                                }
                                                                            }
                                                                            else if (negBalance >= 50) {
                                                                                if (ssDCval < 78) {
                                                                                    return 4;
                                                                                }
                                                                                else if (ssDCval >= 78) {
                                                                                    if (Cs < 2340) {
                                                                                        if (mcTcor < 86) {
                                                                                            if (SC < 3979) {
                                                                                                return 4;
                                                                                            }
                                                                                            else if (SC >= 3979) {
                                                                                                return 2;
                                                                                            }
                                                                                        }
                                                                                        else if (mcTcor >= 86) {
                                                                                            return 2;
                                                                                        }
                                                                                    }
                                                                                    else if (Cs >= 2340) {
                                                                                        if (SC < 4877) {
                                                                                            if (Rs < 1620) {
                                                                                                return 2;
                                                                                            }
                                                                                            else if (Rs >= 1620) {
                                                                                                return 4;
                                                                                            }
                                                                                        }
                                                                                        else if (SC >= 4877) {
                                                                                            return 2;
                                                                                        }
                                                                                    }
                                                                                }
                                                                            }
                                                                        }
                                                                        else if (SC >= 5093) {
                                                                            if (TSC < 12) {
                                                                                if (CsDiff < 11) {
                                                                                    if (Cs < 2235) {
                                                                                        return 1;
                                                                                    }
                                                                                    else if (Cs >= 2235) {
                                                                                        if (refDCval < 94) {
                                                                                            return 2;
                                                                                        }
                                                                                        else if (refDCval >= 94) {
                                                                                            if (RsCsDiff < 201) {
                                                                                                return 4;
                                                                                            }
                                                                                            else if (RsCsDiff >= 201) {
                                                                                                return 2;
                                                                                            }
                                                                                        }
                                                                                    }
                                                                                }
                                                                                else if (CsDiff >= 11) {
                                                                                    return 4;
                                                                                }
                                                                            }
                                                                            else if (TSC >= 12) {
                                                                                if (TSC < 13) {
                                                                                    return 1;
                                                                                }
                                                                                else if (TSC >= 13) {
                                                                                    return 4;
                                                                                }
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                            else if (tcor >= 90) {
                                                                if (refDCval < 81) {
                                                                    if (diffAFD < 2) {
                                                                        return 2;
                                                                    }
                                                                    else if (diffAFD >= 2) {
                                                                        return 1;
                                                                    }
                                                                }
                                                                else if (refDCval >= 81) {
                                                                    if (RsCsDiff < 63) {
                                                                        if (SC < 4521) {
                                                                            return 2;
                                                                        }
                                                                        else if (SC >= 4521) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                    else if (RsCsDiff >= 63) {
                                                                        if (Cs < 2244) {
                                                                            if (gchDC < 5) {
                                                                                return 1;
                                                                            }
                                                                            else if (gchDC >= 5) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                        else if (Cs >= 2244) {
                                                                            return 2;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (ssDCval >= 103) {
                                                        if (posBalance < 63) {
                                                            if (posBalance < 59) {
                                                                if (ssDCval < 111) {
                                                                    if (Cs < 2382) {
                                                                        if (Rs < 1347) {
                                                                            return 8;
                                                                        }
                                                                        else if (Rs >= 1347) {
                                                                            return 2;
                                                                        }
                                                                    }
                                                                    else if (Cs >= 2382) {
                                                                        return 8;
                                                                    }
                                                                }
                                                                else if (ssDCval >= 111) {
                                                                    return 2;
                                                                }
                                                            }
                                                            else if (posBalance >= 59) {
                                                                if (TSC < 11) {
                                                                    if (Cs < 1720) {
                                                                        return 8;
                                                                    }
                                                                    else if (Cs >= 1720) {
                                                                        return 2;
                                                                    }
                                                                }
                                                                else if (TSC >= 11) {
                                                                    if (Cs < 1914) {
                                                                        return 8;
                                                                    }
                                                                    else if (Cs >= 1914) {
                                                                        return 1;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (posBalance >= 63) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                                else if (ssDCval >= 115) {
                                                    if (CsDiff < 8) {
                                                        if (mcTcor < 94) {
                                                            if (refDCval < 123) {
                                                                return 4;
                                                            }
                                                            else if (refDCval >= 123) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (mcTcor >= 94) {
                                                            return 2;
                                                        }
                                                    }
                                                    else if (CsDiff >= 8) {
                                                        if (mcTcor < 94) {
                                                            if (RsCsDiff < 327) {
                                                                if (AFD < 17) {
                                                                    return 2;
                                                                }
                                                                else if (AFD >= 17) {
                                                                    return 1;
                                                                }
                                                            }
                                                            else if (RsCsDiff >= 327) {
                                                                if (Cs < 2927) {
                                                                    return 1;
                                                                }
                                                                else if (Cs >= 2927) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                        else if (mcTcor >= 94) {
                                                            return 2;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (refDCval >= 124) {
                                        if (Cs < 2838) {
                                            if (tcor < 80) {
                                                return 1;
                                            }
                                            else if (tcor >= 80) {
                                                return 4;
                                            }
                                        }
                                        else if (Cs >= 2838) {
                                            if (RsCsDiff < 339) {
                                                if (MVDiff < 11) {
                                                    return 1;
                                                }
                                                else if (MVDiff >= 11) {
                                                    return 4;
                                                }
                                            }
                                            else if (RsCsDiff >= 339) {
                                                if (diffRsCsdiff < -60) {
                                                    return 8;
                                                }
                                                else if (diffRsCsdiff >= -60) {
                                                    if (Cs < 3541) {
                                                        return 1;
                                                    }
                                                    else if (Cs >= 3541) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (mu_mv_mag_sq >= 46) {
                                    if (Rs < 2527) {
                                        if (ssDCval < 103) {
                                            if (ssDCval < 93) {
                                                if (posBalance < 43) {
                                                    if (Cs < 1928) {
                                                        return 4;
                                                    }
                                                    else if (Cs >= 1928) {
                                                        return 8;
                                                    }
                                                }
                                                else if (posBalance >= 43) {
                                                    if (mu_mv_mag_sq < 58) {
                                                        if (mcTcor < 95) {
                                                            if (mu_mv_mag_sq < 50) {
                                                                if (diffAFD < 2) {
                                                                    return 1;
                                                                }
                                                                else if (diffAFD >= 2) {
                                                                    if (AFD < 16) {
                                                                        return 8;
                                                                    }
                                                                    else if (AFD >= 16) {
                                                                        return 1;
                                                                    }
                                                                }
                                                            }
                                                            else if (mu_mv_mag_sq >= 50) {
                                                                if (diffAFD < 3) {
                                                                    if (CsDiff < 9) {
                                                                        return 8;
                                                                    }
                                                                    else if (CsDiff >= 9) {
                                                                        return 1;
                                                                    }
                                                                }
                                                                else if (diffAFD >= 3) {
                                                                    return 1;
                                                                }
                                                            }
                                                        }
                                                        else if (mcTcor >= 95) {
                                                            if (Cs < 1604) {
                                                                return 1;
                                                            }
                                                            else if (Cs >= 1604) {
                                                                return 2;
                                                            }
                                                        }
                                                    }
                                                    else if (mu_mv_mag_sq >= 58) {
                                                        if (CsDiff < 8) {
                                                            if (refDCval < 93) {
                                                                if (mcTcor < 93) {
                                                                    return 8;
                                                                }
                                                                else if (mcTcor >= 93) {
                                                                    if (Cs < 1932) {
                                                                        return 2;
                                                                    }
                                                                    else if (Cs >= 1932) {
                                                                        return 8;
                                                                    }
                                                                }
                                                            }
                                                            else if (refDCval >= 93) {
                                                                if (Cs < 2122) {
                                                                    return 1;
                                                                }
                                                                else if (Cs >= 2122) {
                                                                    return 2;
                                                                }
                                                            }
                                                        }
                                                        else if (CsDiff >= 8) {
                                                            if (AFD < 12) {
                                                                return 8;
                                                            }
                                                            else if (AFD >= 12) {
                                                                if (Cs < 1922) {
                                                                    return 4;
                                                                }
                                                                else if (Cs >= 1922) {
                                                                    return 2;
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                            else if (ssDCval >= 93) {
                                                if (Rs < 2067) {
                                                    if (tcor < 82) {
                                                        if (mcTcor < 83) {
                                                            if (RsDiff < 6) {
                                                                return 8;
                                                            }
                                                            else if (RsDiff >= 6) {
                                                                return 2;
                                                            }
                                                        }
                                                        else if (mcTcor >= 83) {
                                                            return 2;
                                                        }
                                                    }
                                                    else if (tcor >= 82) {
                                                        return 8;
                                                    }
                                                }
                                                else if (Rs >= 2067) {
                                                    if (MVDiff < 12) {
                                                        return 2;
                                                    }
                                                    else if (MVDiff >= 12) {
                                                        return 1;
                                                    }
                                                }
                                            }
                                        }
                                        else if (ssDCval >= 103) {
                                            if (RsDiff < 6) {
                                                if (Cs < 484) {
                                                    return 2;
                                                }
                                                else if (Cs >= 484) {
                                                    return 8;
                                                }
                                            }
                                            else if (RsDiff >= 6) {
                                                if (Cs < 2501) {
                                                    if (AFD < 13) {
                                                        if (negBalance < 49) {
                                                            if (diffAFD < -2) {
                                                                return 4;
                                                            }
                                                            else if (diffAFD >= -2) {
                                                                if (refDCval < 128) {
                                                                    return 8;
                                                                }
                                                                else if (refDCval >= 128) {
                                                                    return 1;
                                                                }
                                                            }
                                                        }
                                                        else if (negBalance >= 49) {
                                                            if (Cs < 1689) {
                                                                if (MVDiff < 6) {
                                                                    return 4;
                                                                }
                                                                else if (MVDiff >= 6) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (Cs >= 1689) {
                                                                if (mu_mv_mag_sq < 77) {
                                                                    return 4;
                                                                }
                                                                else if (mu_mv_mag_sq >= 77) {
                                                                    if (Rs < 2101) {
                                                                        return 8;
                                                                    }
                                                                    else if (Rs >= 2101) {
                                                                        return 4;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (AFD >= 13) {
                                                        return 1;
                                                    }
                                                }
                                                else if (Cs >= 2501) {
                                                    return 8;
                                                }
                                            }
                                        }
                                    }
                                    else if (Rs >= 2527) {
                                        if (Rs < 3392) {
                                            if (mu_mv_mag_sq < 83) {
                                                if (tcor < 93) {
                                                    if (Cs < 1890) {
                                                        if (refDCval < 117) {
                                                            if (Cs < 1371) {
                                                                return 2;
                                                            }
                                                            else if (Cs >= 1371) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (refDCval >= 117) {
                                                            if (SC < 4570) {
                                                                if (mu_mv_mag_sq < 57) {
                                                                    return 4;
                                                                }
                                                                else if (mu_mv_mag_sq >= 57) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (SC >= 4570) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                    else if (Cs >= 1890) {
                                                        if (SC < 6208) {
                                                            if (AFD < 12) {
                                                                if (mu_mv_mag_sq < 74) {
                                                                    if (diffRsCsdiff < -30) {
                                                                        if (Cs < 2232) {
                                                                            return 4;
                                                                        }
                                                                        else if (Cs >= 2232) {
                                                                            return 2;
                                                                        }
                                                                    }
                                                                    else if (diffRsCsdiff >= -30) {
                                                                        if (posBalance < 56) {
                                                                            if (Cs < 2111) {
                                                                                return 1;
                                                                            }
                                                                            else if (Cs >= 2111) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                        else if (posBalance >= 56) {
                                                                            if (diffMVdiffVal < 1) {
                                                                                return 4;
                                                                            }
                                                                            else if (diffMVdiffVal >= 1) {
                                                                                if (MVDiff < 6) {
                                                                                    if (SC < 4720) {
                                                                                        return 8;
                                                                                    }
                                                                                    else if (SC >= 4720) {
                                                                                        if (Rs < 2583) {
                                                                                            return 8;
                                                                                        }
                                                                                        else if (Rs >= 2583) {
                                                                                            return 4;
                                                                                        }
                                                                                    }
                                                                                }
                                                                                else if (MVDiff >= 6) {
                                                                                    return 4;
                                                                                }
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                                else if (mu_mv_mag_sq >= 74) {
                                                                    if (SC < 5196) {
                                                                        if (SC < 4684) {
                                                                            return 8;
                                                                        }
                                                                        else if (SC >= 4684) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                    else if (SC >= 5196) {
                                                                        return 8;
                                                                    }
                                                                }
                                                            }
                                                            else if (AFD >= 12) {
                                                                if (refDCval < 126) {
                                                                    if (refDCval < 91) {
                                                                        if (ssDCval < 71) {
                                                                            return 8;
                                                                        }
                                                                        else if (ssDCval >= 71) {
                                                                            if (mu_mv_mag_sq < 63) {
                                                                                if (tcor < 84) {
                                                                                    return 4;
                                                                                }
                                                                                else if (tcor >= 84) {
                                                                                    return 1;
                                                                                }
                                                                            }
                                                                            else if (mu_mv_mag_sq >= 63) {
                                                                                if (TSC < 11) {
                                                                                    return 8;
                                                                                }
                                                                                else if (TSC >= 11) {
                                                                                    if (SC < 5462) {
                                                                                        if (Cs < 2250) {
                                                                                            return 1;
                                                                                        }
                                                                                        else if (Cs >= 2250) {
                                                                                            return 8;
                                                                                        }
                                                                                    }
                                                                                    else if (SC >= 5462) {
                                                                                        return 4;
                                                                                    }
                                                                                }
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (refDCval >= 91) {
                                                                        if (RsCsDiff < 201) {
                                                                            if (AFD < 13) {
                                                                                return 1;
                                                                            }
                                                                            else if (AFD >= 13) {
                                                                                if (ssDCval < 106) {
                                                                                    if (Cs < 2454) {
                                                                                        if (Cs < 2167) {
                                                                                            return 1;
                                                                                        }
                                                                                        else if (Cs >= 2167) {
                                                                                            if (CsDiff < 8) {
                                                                                                return 4;
                                                                                            }
                                                                                            else if (CsDiff >= 8) {
                                                                                                if (Rs < 2706) {
                                                                                                    return 1;
                                                                                                }
                                                                                                else if (Rs >= 2706) {
                                                                                                    if (Cs < 2426) {
                                                                                                        return 4;
                                                                                                    }
                                                                                                    else if (Cs >= 2426) {
                                                                                                        return 1;
                                                                                                    }
                                                                                                }
                                                                                            }
                                                                                        }
                                                                                    }
                                                                                    else if (Cs >= 2454) {
                                                                                        if (Cs < 2828) {
                                                                                            return 1;
                                                                                        }
                                                                                        else if (Cs >= 2828) {
                                                                                            return 4;
                                                                                        }
                                                                                    }
                                                                                }
                                                                                else if (ssDCval >= 106) {
                                                                                    return 1;
                                                                                }
                                                                            }
                                                                        }
                                                                        else if (RsCsDiff >= 201) {
                                                                            if (RsDiff < 14) {
                                                                                return 4;
                                                                            }
                                                                            else if (RsDiff >= 14) {
                                                                                return 8;
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                                else if (refDCval >= 126) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                        else if (SC >= 6208) {
                                                            if (RsCsDiff < 137) {
                                                                return 1;
                                                            }
                                                            else if (RsCsDiff >= 137) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (tcor >= 93) {
                                                    if (AFD < 9) {
                                                        return 4;
                                                    }
                                                    else if (AFD >= 9) {
                                                        if (Cs < 1659) {
                                                            return 8;
                                                        }
                                                        else if (Cs >= 1659) {
                                                            return 1;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (mu_mv_mag_sq >= 83) {
                                                if (ssDCval < 87) {
                                                    if (SC < 5841) {
                                                        if (posBalance < 53) {
                                                            return 2;
                                                        }
                                                        else if (posBalance >= 53) {
                                                            if (mu_mv_mag_sq < 98) {
                                                                return 4;
                                                            }
                                                            else if (mu_mv_mag_sq >= 98) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                    else if (SC >= 5841) {
                                                        return 8;
                                                    }
                                                }
                                                else if (ssDCval >= 87) {
                                                    if (Cs < 1805) {
                                                        if (Rs < 3189) {
                                                            return 4;
                                                        }
                                                        else if (Rs >= 3189) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (Cs >= 1805) {
                                                        if (Rs < 2580) {
                                                            return 2;
                                                        }
                                                        else if (Rs >= 2580) {
                                                            if (diffMVdiffVal < -5) {
                                                                return 4;
                                                            }
                                                            else if (diffMVdiffVal >= -5) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (Rs >= 3392) {
                                            if (ssDCval < 78) {
                                                if (tcor < 76) {
                                                    return 8;
                                                }
                                                else if (tcor >= 76) {
                                                    return 4;
                                                }
                                            }
                                            else if (ssDCval >= 78) {
                                                if (posBalance < 59) {
                                                    if (Rs < 3751) {
                                                        if (AFD < 13) {
                                                            return 8;
                                                        }
                                                        else if (AFD >= 13) {
                                                            if (Cs < 2987) {
                                                                if (SC < 5727) {
                                                                    return 8;
                                                                }
                                                                else if (SC >= 5727) {
                                                                    return 1;
                                                                }
                                                            }
                                                            else if (Cs >= 2987) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                    else if (Rs >= 3751) {
                                                        if (Cs < 3498) {
                                                            if (Rs < 4171) {
                                                                if (AFD < 14) {
                                                                    if (Cs < 2461) {
                                                                        return 2;
                                                                    }
                                                                    else if (Cs >= 2461) {
                                                                        if (Rs < 3762) {
                                                                            return 4;
                                                                        }
                                                                        else if (Rs >= 3762) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                }
                                                                else if (AFD >= 14) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (Rs >= 4171) {
                                                                if (AFD < 11) {
                                                                    if (diffMVdiffVal < 1) {
                                                                        if (RsDiff < 9) {
                                                                            return 4;
                                                                        }
                                                                        else if (RsDiff >= 9) {
                                                                            return 2;
                                                                        }
                                                                    }
                                                                    else if (diffMVdiffVal >= 1) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (AFD >= 11) {
                                                                    if (Cs < 3011) {
                                                                        return 1;
                                                                    }
                                                                    else if (Cs >= 3011) {
                                                                        return 2;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (Cs >= 3498) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                                else if (posBalance >= 59) {
                                                    return 8;
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                        }
                        else if (refDCval >= 133) {
                            if (refDCval < 140) {
                                if (tcor < 81) {
                                    if (posBalance < 62) {
                                        if (negBalance < 58) {
                                            return 1;
                                        }
                                        else if (negBalance >= 58) {
                                            if (Rs < 3987) {
                                                return 1;
                                            }
                                            else if (Rs >= 3987) {
                                                return 8;
                                            }
                                        }
                                    }
                                    else if (posBalance >= 62) {
                                        if (AFD < 18) {
                                            return 2;
                                        }
                                        else if (AFD >= 18) {
                                            return 1;
                                        }
                                    }
                                }
                                else if (tcor >= 81) {
                                    if (Cs < 1464) {
                                        if (SC < 2836) {
                                            return 4;
                                        }
                                        else if (SC >= 2836) {
                                            return 2;
                                        }
                                    }
                                    else if (Cs >= 1464) {
                                        if (CsDiff < 9) {
                                            if (Rs < 3437) {
                                                return 4;
                                            }
                                            else if (Rs >= 3437) {
                                                return 1;
                                            }
                                        }
                                        else if (CsDiff >= 9) {
                                            return 8;
                                        }
                                    }
                                }
                            }
                            else if (refDCval >= 140) {
                                if (mu_mv_mag_sq < 7) {
                                    if (diffTSC < 2) {
                                        if (diffMVdiffVal < -3) {
                                            if (RsDiff < 8) {
                                                return 4;
                                            }
                                            else if (RsDiff >= 8) {
                                                return 8;
                                            }
                                        }
                                        else if (diffMVdiffVal >= -3) {
                                            return 4;
                                        }
                                    }
                                    else if (diffTSC >= 2) {
                                        if (SC < 2326) {
                                            return 4;
                                        }
                                        else if (SC >= 2326) {
                                            if (Cs < 1965) {
                                                return 1;
                                            }
                                            else if (Cs >= 1965) {
                                                return 2;
                                            }
                                        }
                                    }
                                }
                                else if (mu_mv_mag_sq >= 7) {
                                    if (posBalance < 59) {
                                        if (Cs < 662) {
                                            return 2;
                                        }
                                        else if (Cs >= 662) {
                                            if (mcTcor < 95) {
                                                return 8;
                                            }
                                            else if (mcTcor >= 95) {
                                                if (gchDC < 1) {
                                                    return 8;
                                                }
                                                else if (gchDC >= 1) {
                                                    return 4;
                                                }
                                            }
                                        }
                                    }
                                    else if (posBalance >= 59) {
                                        if (RsDiff < 4) {
                                            return 4;
                                        }
                                        else if (RsDiff >= 4) {
                                            return 2;
                                        }
                                    }
                                }
                            }
                        }
                    }
                    else if (MVDiff >= 17) {
                        if (Cs < 1490) {
                            if (RsDiff < 7) {
                                if (SC < 872) {
                                    if (TSCindex < 6) {
                                        if (refDCval < 73) {
                                            if (posBalance < 98) {
                                                if (posBalance < 13) {
                                                    return 2;
                                                }
                                                else if (posBalance >= 13) {
                                                    return 1;
                                                }
                                            }
                                            else if (posBalance >= 98) {
                                                return 4;
                                            }
                                        }
                                        else if (refDCval >= 73) {
                                            if (ssDCval < 186) {
                                                if (Cs < 414) {
                                                    return 2;
                                                }
                                                else if (Cs >= 414) {
                                                    if (SC < 790) {
                                                        return 1;
                                                    }
                                                    else if (SC >= 790) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                            else if (ssDCval >= 186) {
                                                return 1;
                                            }
                                        }
                                    }
                                    else if (TSCindex >= 6) {
                                        if (refDCval < 108) {
                                            if (ssDCval < 84) {
                                                if (AFD < 29) {
                                                    if (Rs < 383) {
                                                        if (mcTcor < 102) {
                                                            if (posBalance < 121) {
                                                                if (diffMVdiffVal < 73) {
                                                                    if (SC < 100) {
                                                                        return 2;
                                                                    }
                                                                    else if (SC >= 100) {
                                                                        return 1;
                                                                    }
                                                                }
                                                                else if (diffMVdiffVal >= 73) {
                                                                    return 2;
                                                                }
                                                            }
                                                            else if (posBalance >= 121) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (mcTcor >= 102) {
                                                            return 2;
                                                        }
                                                    }
                                                    else if (Rs >= 383) {
                                                        if (refDCval < 74) {
                                                            return 2;
                                                        }
                                                        else if (refDCval >= 74) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                                else if (AFD >= 29) {
                                                    if (RsDiff < 6) {
                                                        return 8;
                                                    }
                                                    else if (RsDiff >= 6) {
                                                        return 1;
                                                    }
                                                }
                                            }
                                            else if (ssDCval >= 84) {
                                                if (tcor < 64) {
                                                    if (Rs < 172) {
                                                        return 1;
                                                    }
                                                    else if (Rs >= 172) {
                                                        return 2;
                                                    }
                                                }
                                                else if (tcor >= 64) {
                                                    if (tcor < 95) {
                                                        return 1;
                                                    }
                                                    else if (tcor >= 95) {
                                                        if (AFD < 9) {
                                                            return 1;
                                                        }
                                                        else if (AFD >= 9) {
                                                            return 2;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (refDCval >= 108) {
                                            if (RsDiff < 3) {
                                                if (MVDiff < 123) {
                                                    if (ssDCval < 110) {
                                                        return 8;
                                                    }
                                                    else if (ssDCval >= 110) {
                                                        return 4;
                                                    }
                                                }
                                                else if (MVDiff >= 123) {
                                                    return 1;
                                                }
                                            }
                                            else if (RsDiff >= 3) {
                                                if (CsDiff < 2) {
                                                    return 2;
                                                }
                                                else if (CsDiff >= 2) {
                                                    if (Rs < 365) {
                                                        return 1;
                                                    }
                                                    else if (Rs >= 365) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (SC >= 872) {
                                    if (negBalance < 54) {
                                        if (ssDCval < 44) {
                                            if (CsDiff < 3) {
                                                if (mu_mv_mag_sq < 29) {
                                                    return 4;
                                                }
                                                else if (mu_mv_mag_sq >= 29) {
                                                    if (TSC < 6) {
                                                        if (diffMVdiffVal < -4) {
                                                            return 1;
                                                        }
                                                        else if (diffMVdiffVal >= -4) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (TSC >= 6) {
                                                        if (diffMVdiffVal < -10) {
                                                            if (Cs < 559) {
                                                                return 4;
                                                            }
                                                            else if (Cs >= 559) {
                                                                return 1;
                                                            }
                                                        }
                                                        else if (diffMVdiffVal >= -10) {
                                                            if (MVDiff < 47) {
                                                                return 1;
                                                            }
                                                            else if (MVDiff >= 47) {
                                                                if (Rs < 735) {
                                                                    return 4;
                                                                }
                                                                else if (Rs >= 735) {
                                                                    return 1;
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                            else if (CsDiff >= 3) {
                                                if (negBalance < 50) {
                                                    return 2;
                                                }
                                                else if (negBalance >= 50) {
                                                    return 4;
                                                }
                                            }
                                        }
                                        else if (ssDCval >= 44) {
                                            if (mcTcor < 93) {
                                                if (CsDiff < 4) {
                                                    if (Rs < 649) {
                                                        if (refDCval < 134) {
                                                            if (tcor < 70) {
                                                                if (AFD < 8) {
                                                                    return 4;
                                                                }
                                                                else if (AFD >= 8) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (tcor >= 70) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (refDCval >= 134) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (Rs >= 649) {
                                                        if (tcor < 80) {
                                                            if (tcor < 66) {
                                                                if (negBalance < 43) {
                                                                    return 8;
                                                                }
                                                                else if (negBalance >= 43) {
                                                                    return 2;
                                                                }
                                                            }
                                                            else if (tcor >= 66) {
                                                                if (SC < 1905) {
                                                                    if (mcTcor < 91) {
                                                                        return 4;
                                                                    }
                                                                    else if (mcTcor >= 91) {
                                                                        return 8;
                                                                    }
                                                                }
                                                                else if (SC >= 1905) {
                                                                    if (ssDCval < 75) {
                                                                        return 1;
                                                                    }
                                                                    else if (ssDCval >= 75) {
                                                                        return 2;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (tcor >= 80) {
                                                            if (MVDiff < 23) {
                                                                if (Rs < 1600) {
                                                                    return 8;
                                                                }
                                                                else if (Rs >= 1600) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (MVDiff >= 23) {
                                                                if (posBalance < 53) {
                                                                    if (Rs < 1327) {
                                                                        return 2;
                                                                    }
                                                                    else if (Rs >= 1327) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (posBalance >= 53) {
                                                                    return 1;
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (CsDiff >= 4) {
                                                    if (SC < 2498) {
                                                        if (diffRsCsdiff < 11) {
                                                            if (mu_mv_mag_sq < 46) {
                                                                if (mu_mv_mag_sq < 3) {
                                                                    if (Rs < 1030) {
                                                                        return 8;
                                                                    }
                                                                    else if (Rs >= 1030) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (mu_mv_mag_sq >= 3) {
                                                                    if (RsDiff < 4) {
                                                                        return 2;
                                                                    }
                                                                    else if (RsDiff >= 4) {
                                                                        if (Cs < 526) {
                                                                            return 2;
                                                                        }
                                                                        else if (Cs >= 526) {
                                                                            if (Cs < 1051) {
                                                                                return 1;
                                                                            }
                                                                            else if (Cs >= 1051) {
                                                                                return 8;
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                            else if (mu_mv_mag_sq >= 46) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (diffRsCsdiff >= 11) {
                                                            if (refDCval < 50) {
                                                                if (Cs < 841) {
                                                                    return 1;
                                                                }
                                                                else if (Cs >= 841) {
                                                                    if (Rs < 714) {
                                                                        return 4;
                                                                    }
                                                                    else if (Rs >= 714) {
                                                                        return 8;
                                                                    }
                                                                }
                                                            }
                                                            else if (refDCval >= 50) {
                                                                return 1;
                                                            }
                                                        }
                                                    }
                                                    else if (SC >= 2498) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                            else if (mcTcor >= 93) {
                                                if (ssDCval < 74) {
                                                    if (CsDiff < 3) {
                                                        if (Cs < 779) {
                                                            return 8;
                                                        }
                                                        else if (Cs >= 779) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (CsDiff >= 3) {
                                                        if (mu_mv_mag_sq < 25) {
                                                            return 2;
                                                        }
                                                        else if (mu_mv_mag_sq >= 25) {
                                                            if (Cs < 931) {
                                                                return 1;
                                                            }
                                                            else if (Cs >= 931) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (ssDCval >= 74) {
                                                    if (TSCindex < 6) {
                                                        if (Rs < 1416) {
                                                            return 4;
                                                        }
                                                        else if (Rs >= 1416) {
                                                            return 2;
                                                        }
                                                    }
                                                    else if (TSCindex >= 6) {
                                                        if (posBalance < 77) {
                                                            return 1;
                                                        }
                                                        else if (posBalance >= 77) {
                                                            if (Rs < 1067) {
                                                                return 8;
                                                            }
                                                            else if (Rs >= 1067) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (negBalance >= 54) {
                                        if (Rs < 898) {
                                            if (TSC < 13) {
                                                if (CsDiff < 3) {
                                                    if (ssDCval < 73) {
                                                        if (gchDC < 4) {
                                                            return 2;
                                                        }
                                                        else if (gchDC >= 4) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (ssDCval >= 73) {
                                                        if (AFD < 14) {
                                                            if (AFD < 11) {
                                                                return 1;
                                                            }
                                                            else if (AFD >= 11) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (AFD >= 14) {
                                                            return 2;
                                                        }
                                                    }
                                                }
                                                else if (CsDiff >= 3) {
                                                    if (negBalance < 70) {
                                                        return 2;
                                                    }
                                                    else if (negBalance >= 70) {
                                                        if (MVDiff < 38) {
                                                            return 2;
                                                        }
                                                        else if (MVDiff >= 38) {
                                                            return 1;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (TSC >= 13) {
                                                return 1;
                                            }
                                        }
                                        else if (Rs >= 898) {
                                            if (AFD < 13) {
                                                if (RsCsDiff < 17) {
                                                    return 8;
                                                }
                                                else if (RsCsDiff >= 17) {
                                                    if (mcTcor < 88) {
                                                        return 4;
                                                    }
                                                    else if (mcTcor >= 88) {
                                                        return 2;
                                                    }
                                                }
                                            }
                                            else if (AFD >= 13) {
                                                if (SC < 2067) {
                                                    if (MVDiff < 105) {
                                                        return 4;
                                                    }
                                                    else if (MVDiff >= 105) {
                                                        return 8;
                                                    }
                                                }
                                                else if (SC >= 2067) {
                                                    if (tcor < 76) {
                                                        return 8;
                                                    }
                                                    else if (tcor >= 76) {
                                                        if (Rs < 1235) {
                                                            return 8;
                                                        }
                                                        else if (Rs >= 1235) {
                                                            return 1;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                            else if (RsDiff >= 7) {
                                if (mcTcor < 85) {
                                    if (diffTSC < 7) {
                                        if (Rs < 1579) {
                                            if (tcor < 45) {
                                                return 1;
                                            }
                                            else if (tcor >= 45) {
                                                if (diffMVdiffVal < 8) {
                                                    if (negBalance < 37) {
                                                        return 1;
                                                    }
                                                    else if (negBalance >= 37) {
                                                        if (CsDiff < 6) {
                                                            return 2;
                                                        }
                                                        else if (CsDiff >= 6) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                                else if (diffMVdiffVal >= 8) {
                                                    if (Rs < 1379) {
                                                        return 1;
                                                    }
                                                    else if (Rs >= 1379) {
                                                        return 2;
                                                    }
                                                }
                                            }
                                        }
                                        else if (Rs >= 1579) {
                                            if (AFD < 31) {
                                                if (MVDiff < 41) {
                                                    if (MVDiff < 23) {
                                                        if (AFD < 20) {
                                                            return 2;
                                                        }
                                                        else if (AFD >= 20) {
                                                            return 1;
                                                        }
                                                    }
                                                    else if (MVDiff >= 23) {
                                                        if (Cs < 503) {
                                                            return 4;
                                                        }
                                                        else if (Cs >= 503) {
                                                            return 2;
                                                        }
                                                    }
                                                }
                                                else if (MVDiff >= 41) {
                                                    if (refDCval < 61) {
                                                        return 4;
                                                    }
                                                    else if (refDCval >= 61) {
                                                        if (RsDiff < 25) {
                                                            if (mu_mv_mag_sq < 16) {
                                                                if (AFD < 26) {
                                                                    if (tcor < 64) {
                                                                        if (Cs < 1369) {
                                                                            return 2;
                                                                        }
                                                                        else if (Cs >= 1369) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                    else if (tcor >= 64) {
                                                                        return 1;
                                                                    }
                                                                }
                                                                else if (AFD >= 26) {
                                                                    if (diffAFD < 3) {
                                                                        if (TSC < 16) {
                                                                            return 1;
                                                                        }
                                                                        else if (TSC >= 16) {
                                                                            if (diffAFD < -2) {
                                                                                return 2;
                                                                            }
                                                                            else if (diffAFD >= -2) {
                                                                                return 1;
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (diffAFD >= 3) {
                                                                        return 2;
                                                                    }
                                                                }
                                                            }
                                                            else if (mu_mv_mag_sq >= 16) {
                                                                return 1;
                                                            }
                                                        }
                                                        else if (RsDiff >= 25) {
                                                            return 2;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (AFD >= 31) {
                                                return 2;
                                            }
                                        }
                                    }
                                    else if (diffTSC >= 7) {
                                        if (SC < 2155) {
                                            if (Cs < 937) {
                                                if (MVDiff < 131) {
                                                    return 4;
                                                }
                                                else if (MVDiff >= 131) {
                                                    if (Cs < 604) {
                                                        return 2;
                                                    }
                                                    else if (Cs >= 604) {
                                                        return 1;
                                                    }
                                                }
                                            }
                                            else if (Cs >= 937) {
                                                if (TSC < 14) {
                                                    return 2;
                                                }
                                                else if (TSC >= 14) {
                                                    return 8;
                                                }
                                            }
                                        }
                                        else if (SC >= 2155) {
                                            if (ssDCval < 136) {
                                                if (mcTcor < 67) {
                                                    if (Rs < 1303) {
                                                        if (AFD < 52) {
                                                            return 1;
                                                        }
                                                        else if (AFD >= 52) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (Rs >= 1303) {
                                                        return 4;
                                                    }
                                                }
                                                else if (mcTcor >= 67) {
                                                    return 8;
                                                }
                                            }
                                            else if (ssDCval >= 136) {
                                                if (Rs < 1733) {
                                                    return 1;
                                                }
                                                else if (Rs >= 1733) {
                                                    return 2;
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (mcTcor >= 85) {
                                    if (negBalance < 51) {
                                        if (RsDiff < 9) {
                                            if (SC < 3139) {
                                                if (diffMVdiffVal < -3) {
                                                    if (Cs < 1035) {
                                                        return 2;
                                                    }
                                                    else if (Cs >= 1035) {
                                                        return 1;
                                                    }
                                                }
                                                else if (diffMVdiffVal >= -3) {
                                                    if (Rs < 1958) {
                                                        return 1;
                                                    }
                                                    else if (Rs >= 1958) {
                                                        return 2;
                                                    }
                                                }
                                            }
                                            else if (SC >= 3139) {
                                                if (AFD < 17) {
                                                    return 8;
                                                }
                                                else if (AFD >= 17) {
                                                    return 4;
                                                }
                                            }
                                        }
                                        else if (RsDiff >= 9) {
                                            if (Cs < 1258) {
                                                if (tcor < 83) {
                                                    if (Cs < 569) {
                                                        return 8;
                                                    }
                                                    else if (Cs >= 569) {
                                                        return 2;
                                                    }
                                                }
                                                else if (tcor >= 83) {
                                                    return 1;
                                                }
                                            }
                                            else if (Cs >= 1258) {
                                                if (AFD < 18) {
                                                    if (AFD < 16) {
                                                        return 4;
                                                    }
                                                    else if (AFD >= 16) {
                                                        return 8;
                                                    }
                                                }
                                                else if (AFD >= 18) {
                                                    if (mu_mv_mag_sq < 105) {
                                                        if (SC < 5347) {
                                                            return 2;
                                                        }
                                                        else if (SC >= 5347) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (mu_mv_mag_sq >= 105) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (negBalance >= 51) {
                                        if (MVDiff < 39) {
                                            if (RsCsDiff < 90) {
                                                if (Rs < 1944) {
                                                    if (refDCval < 71) {
                                                        return 4;
                                                    }
                                                    else if (refDCval >= 71) {
                                                        return 2;
                                                    }
                                                }
                                                else if (Rs >= 1944) {
                                                    return 4;
                                                }
                                            }
                                            else if (RsCsDiff >= 90) {
                                                return 2;
                                            }
                                        }
                                        else if (MVDiff >= 39) {
                                            if (refDCval < 103) {
                                                if (SC < 3952) {
                                                    if (Cs < 897) {
                                                        return 2;
                                                    }
                                                    else if (Cs >= 897) {
                                                        if (SC < 2973) {
                                                            return 2;
                                                        }
                                                        else if (SC >= 2973) {
                                                            return 1;
                                                        }
                                                    }
                                                }
                                                else if (SC >= 3952) {
                                                    return 2;
                                                }
                                            }
                                            else if (refDCval >= 103) {
                                                if (Cs < 830) {
                                                    if (refDCval < 138) {
                                                        if (Rs < 2139) {
                                                            return 2;
                                                        }
                                                        else if (Rs >= 2139) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (refDCval >= 138) {
                                                        if (Rs < 786) {
                                                            return 8;
                                                        }
                                                        else if (Rs >= 786) {
                                                            return 1;
                                                        }
                                                    }
                                                }
                                                else if (Cs >= 830) {
                                                    return 2;
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                        }
                        else if (Cs >= 1490) {
                            if (RsCsDiff < 69) {
                                if (tcor < 85) {
                                    return 8;
                                }
                                else if (tcor >= 85) {
                                    if (TSCindex < 8) {
                                        if (refDCval < 50) {
                                            if (SC < 6438) {
                                                return 4;
                                            }
                                            else if (SC >= 6438) {
                                                if (MVDiff < 31) {
                                                    return 4;
                                                }
                                                else if (MVDiff >= 31) {
                                                    return 8;
                                                }
                                            }
                                        }
                                        else if (refDCval >= 50) {
                                            if (diffAFD < -2) {
                                                if (Cs < 3480) {
                                                    if (posBalance < 55) {
                                                        return 4;
                                                    }
                                                    else if (posBalance >= 55) {
                                                        if (Rs < 2266) {
                                                            return 2;
                                                        }
                                                        else if (Rs >= 2266) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                                else if (Cs >= 3480) {
                                                    return 8;
                                                }
                                            }
                                            else if (diffAFD >= -2) {
                                                if (tcor < 94) {
                                                    return 8;
                                                }
                                                else if (tcor >= 94) {
                                                    return 2;
                                                }
                                            }
                                        }
                                    }
                                    else if (TSCindex >= 8) {
                                        if (diffRsCsdiff < -16) {
                                            if (Cs < 3498) {
                                                if (RsDiff < 6) {
                                                    return 8;
                                                }
                                                else if (RsDiff >= 6) {
                                                    if (SC < 3314) {
                                                        return 8;
                                                    }
                                                    else if (SC >= 3314) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                            else if (Cs >= 3498) {
                                                if (gchDC < 2) {
                                                    return 4;
                                                }
                                                else if (gchDC >= 2) {
                                                    return 8;
                                                }
                                            }
                                        }
                                        else if (diffRsCsdiff >= -16) {
                                            if (SC < 7134) {
                                                if (Scindex < 5) {
                                                    if (SC < 3793) {
                                                        return 8;
                                                    }
                                                    else if (SC >= 3793) {
                                                        return 1;
                                                    }
                                                }
                                                else if (Scindex >= 5) {
                                                    if (Cs < 3410) {
                                                        if (MVDiff < 55) {
                                                            if (negBalance < 48) {
                                                                if (Rs < 3323) {
                                                                    return 8;
                                                                }
                                                                else if (Rs >= 3323) {
                                                                    return 2;
                                                                }
                                                            }
                                                            else if (negBalance >= 48) {
                                                                return 2;
                                                            }
                                                        }
                                                        else if (MVDiff >= 55) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (Cs >= 3410) {
                                                        if (diffMVdiffVal < -3) {
                                                            return 2;
                                                        }
                                                        else if (diffMVdiffVal >= -3) {
                                                            if (Rs < 3556) {
                                                                return 8;
                                                            }
                                                            else if (Rs >= 3556) {
                                                                if (Rs < 3614) {
                                                                    return 2;
                                                                }
                                                                else if (Rs >= 3614) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                            else if (SC >= 7134) {
                                                if (RsDiff < 5) {
                                                    if (ssDCval < 58) {
                                                        if (posBalance < 59) {
                                                            return 2;
                                                        }
                                                        else if (posBalance >= 59) {
                                                            if (Rs < 3793) {
                                                                return 8;
                                                            }
                                                            else if (Rs >= 3793) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                    else if (ssDCval >= 58) {
                                                        if (Rs < 3653) {
                                                            return 4;
                                                        }
                                                        else if (Rs >= 3653) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                                else if (RsDiff >= 5) {
                                                    if (AFD < 10) {
                                                        return 2;
                                                    }
                                                    else if (AFD >= 10) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                            else if (RsCsDiff >= 69) {
                                if (Cs < 2348) {
                                    if (refDCval < 94) {
                                        if (TSC < 13) {
                                            if (posBalance < 53) {
                                                if (AFD < 15) {
                                                    if (MVDiff < 18) {
                                                        if (Rs < 1859) {
                                                            return 1;
                                                        }
                                                        else if (Rs >= 1859) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (MVDiff >= 18) {
                                                        if (negBalance < 36) {
                                                            return 1;
                                                        }
                                                        else if (negBalance >= 36) {
                                                            if (tcor < 89) {
                                                                if (ssDCval < 91) {
                                                                    return 8;
                                                                }
                                                                else if (ssDCval >= 91) {
                                                                    return 1;
                                                                }
                                                            }
                                                            else if (tcor >= 89) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (AFD >= 15) {
                                                    if (refDCval < 87) {
                                                        if (negBalance < 71) {
                                                            if (Rs < 1977) {
                                                                return 1;
                                                            }
                                                            else if (Rs >= 1977) {
                                                                if (RsCsDiff < 193) {
                                                                    if (refDCval < 86) {
                                                                        return 8;
                                                                    }
                                                                    else if (refDCval >= 86) {
                                                                        return 1;
                                                                    }
                                                                }
                                                                else if (RsCsDiff >= 193) {
                                                                    return 1;
                                                                }
                                                            }
                                                        }
                                                        else if (negBalance >= 71) {
                                                            return 2;
                                                        }
                                                    }
                                                    else if (refDCval >= 87) {
                                                        if (Cs < 2271) {
                                                            return 1;
                                                        }
                                                        else if (Cs >= 2271) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (posBalance >= 53) {
                                                if (mu_mv_mag_sq < 6) {
                                                    if (CsDiff < 7) {
                                                        if (Rs < 2224) {
                                                            return 8;
                                                        }
                                                        else if (Rs >= 2224) {
                                                            return 1;
                                                        }
                                                    }
                                                    else if (CsDiff >= 7) {
                                                        if (RsDiff < 10) {
                                                            if (MVDiff < 28) {
                                                                return 4;
                                                            }
                                                            else if (MVDiff >= 28) {
                                                                if (RsDiff < 7) {
                                                                    return 8;
                                                                }
                                                                else if (RsDiff >= 7) {
                                                                    if (MVDiff < 30) {
                                                                        return 8;
                                                                    }
                                                                    else if (MVDiff >= 30) {
                                                                        return 4;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (RsDiff >= 10) {
                                                            if (diffMVdiffVal < 4) {
                                                                if (Rs < 2423) {
                                                                    return 4;
                                                                }
                                                                else if (Rs >= 2423) {
                                                                    return 2;
                                                                }
                                                            }
                                                            else if (diffMVdiffVal >= 4) {
                                                                return 1;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (mu_mv_mag_sq >= 6) {
                                                    if (RsCsDiff < 129) {
                                                        if (Scindex < 5) {
                                                            if (TSC < 12) {
                                                                if (SC < 3858) {
                                                                    if (SC < 3597) {
                                                                        return 4;
                                                                    }
                                                                    else if (SC >= 3597) {
                                                                        if (Rs < 1991) {
                                                                            return 4;
                                                                        }
                                                                        else if (Rs >= 1991) {
                                                                            return 2;
                                                                        }
                                                                    }
                                                                }
                                                                else if (SC >= 3858) {
                                                                    if (Rs < 2167) {
                                                                        return 8;
                                                                    }
                                                                    else if (Rs >= 2167) {
                                                                        return 1;
                                                                    }
                                                                }
                                                            }
                                                            else if (TSC >= 12) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (Scindex >= 5) {
                                                            if (refDCval < 89) {
                                                                if (mcTcor < 91) {
                                                                    if (Cs < 2124) {
                                                                        return 1;
                                                                    }
                                                                    else if (Cs >= 2124) {
                                                                        return 8;
                                                                    }
                                                                }
                                                                else if (mcTcor >= 91) {
                                                                    return 2;
                                                                }
                                                            }
                                                            else if (refDCval >= 89) {
                                                                return 1;
                                                            }
                                                        }
                                                    }
                                                    else if (RsCsDiff >= 129) {
                                                        if (ssDCval < 50) {
                                                            return 2;
                                                        }
                                                        else if (ssDCval >= 50) {
                                                            if (mu_mv_mag_sq < 59) {
                                                                if (SC < 4653) {
                                                                    return 1;
                                                                }
                                                                else if (SC >= 4653) {
                                                                    if (refDCval < 85) {
                                                                        return 1;
                                                                    }
                                                                    else if (refDCval >= 85) {
                                                                        if (Rs < 2558) {
                                                                            return 4;
                                                                        }
                                                                        else if (Rs >= 2558) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                            else if (mu_mv_mag_sq >= 59) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (TSC >= 13) {
                                            if (CsDiff < 10) {
                                                if (MVDiff < 26) {
                                                    if (diffRsCsdiff < -18) {
                                                        return 8;
                                                    }
                                                    else if (diffRsCsdiff >= -18) {
                                                        if (ssDCval < 79) {
                                                            if (TSC < 14) {
                                                                return 4;
                                                            }
                                                            else if (TSC >= 14) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (ssDCval >= 79) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                                else if (MVDiff >= 26) {
                                                    return 8;
                                                }
                                            }
                                            else if (CsDiff >= 10) {
                                                if (diffAFD < -1) {
                                                    return 8;
                                                }
                                                else if (diffAFD >= -1) {
                                                    if (AFD < 29) {
                                                        return 1;
                                                    }
                                                    else if (AFD >= 29) {
                                                        return 2;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (refDCval >= 94) {
                                        if (tcor < 68) {
                                            if (refDCval < 110) {
                                                return 2;
                                            }
                                            else if (refDCval >= 110) {
                                                if (diffTSC < 2) {
                                                    if (ssDCval < 147) {
                                                        if (RsDiff < 17) {
                                                            if (posBalance < 49) {
                                                                return 4;
                                                            }
                                                            else if (posBalance >= 49) {
                                                                return 1;
                                                            }
                                                        }
                                                        else if (RsDiff >= 17) {
                                                            if (MVDiff < 31) {
                                                                return 1;
                                                            }
                                                            else if (MVDiff >= 31) {
                                                                return 2;
                                                            }
                                                        }
                                                    }
                                                    else if (ssDCval >= 147) {
                                                        if (RsCsDiff < 434) {
                                                            if (Rs < 2192) {
                                                                return 8;
                                                            }
                                                            else if (Rs >= 2192) {
                                                                return 2;
                                                            }
                                                        }
                                                        else if (RsCsDiff >= 434) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                                else if (diffTSC >= 2) {
                                                    if (ssDCval < 104) {
                                                        return 4;
                                                    }
                                                    else if (ssDCval >= 104) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                        }
                                        else if (tcor >= 68) {
                                            if (mcTcor < 84) {
                                                if (diffMVdiffVal < -11) {
                                                    return 1;
                                                }
                                                else if (diffMVdiffVal >= -11) {
                                                    return 4;
                                                }
                                            }
                                            else if (mcTcor >= 84) {
                                                if (mcTcor < 95) {
                                                    if (Cs < 1889) {
                                                        if (RsDiff < 13) {
                                                            if (tcor < 85) {
                                                                if (diffRsCsdiff < -12) {
                                                                    if (diffRsCsdiff < -24) {
                                                                        if (diffTSC < 1) {
                                                                            return 4;
                                                                        }
                                                                        else if (diffTSC >= 1) {
                                                                            if (Cs < 1647) {
                                                                                return 8;
                                                                            }
                                                                            else if (Cs >= 1647) {
                                                                                return 2;
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (diffRsCsdiff >= -24) {
                                                                        return 2;
                                                                    }
                                                                }
                                                                else if (diffRsCsdiff >= -12) {
                                                                    if (gchDC < 4) {
                                                                        return 4;
                                                                    }
                                                                    else if (gchDC >= 4) {
                                                                        if (Rs < 2168) {
                                                                            return 4;
                                                                        }
                                                                        else if (Rs >= 2168) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                            else if (tcor >= 85) {
                                                                if (RsDiff < 12) {
                                                                    if (diffMVdiffVal < 10) {
                                                                        return 8;
                                                                    }
                                                                    else if (diffMVdiffVal >= 10) {
                                                                        if (Rs < 2652) {
                                                                            return 4;
                                                                        }
                                                                        else if (Rs >= 2652) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                }
                                                                else if (RsDiff >= 12) {
                                                                    return 2;
                                                                }
                                                            }
                                                        }
                                                        else if (RsDiff >= 13) {
                                                            if (tcor < 80) {
                                                                if (TSC < 15) {
                                                                    if (CsDiff < 11) {
                                                                        return 4;
                                                                    }
                                                                    else if (CsDiff >= 11) {
                                                                        return 2;
                                                                    }
                                                                }
                                                                else if (TSC >= 15) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (tcor >= 80) {
                                                                if (ssDCval < 126) {
                                                                    if (TSC < 13) {
                                                                        return 2;
                                                                    }
                                                                    else if (TSC >= 13) {
                                                                        if (Rs < 2779) {
                                                                            if (Rs < 2380) {
                                                                                return 2;
                                                                            }
                                                                            else if (Rs >= 2380) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                        else if (Rs >= 2779) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                }
                                                                else if (ssDCval >= 126) {
                                                                    return 2;
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (Cs >= 1889) {
                                                        if (ssDCval < 123) {
                                                            if (Rs < 2863) {
                                                                if (TSC < 13) {
                                                                    if (Rs < 2443) {
                                                                        if (negBalance < 40) {
                                                                            return 8;
                                                                        }
                                                                        else if (negBalance >= 40) {
                                                                            if (negBalance < 54) {
                                                                                return 4;
                                                                            }
                                                                            else if (negBalance >= 54) {
                                                                                return 2;
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (Rs >= 2443) {
                                                                        if (SC < 4688) {
                                                                            return 1;
                                                                        }
                                                                        else if (SC >= 4688) {
                                                                            if (Rs < 2706) {
                                                                                return 8;
                                                                            }
                                                                            else if (Rs >= 2706) {
                                                                                return 1;
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                                else if (TSC >= 13) {
                                                                    if (Cs < 1902) {
                                                                        return 8;
                                                                    }
                                                                    else if (Cs >= 1902) {
                                                                        return 1;
                                                                    }
                                                                }
                                                            }
                                                            else if (Rs >= 2863) {
                                                                if (diffMVdiffVal < -2) {
                                                                    if (tcor < 79) {
                                                                        if (Rs < 3091) {
                                                                            return 8;
                                                                        }
                                                                        else if (Rs >= 3091) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                    else if (tcor >= 79) {
                                                                        if (SC < 5365) {
                                                                            return 2;
                                                                        }
                                                                        else if (SC >= 5365) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                }
                                                                else if (diffMVdiffVal >= -2) {
                                                                    if (diffRsCsdiff < 19) {
                                                                        if (Cs < 2253) {
                                                                            return 2;
                                                                        }
                                                                        else if (Cs >= 2253) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                    else if (diffRsCsdiff >= 19) {
                                                                        return 4;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (ssDCval >= 123) {
                                                            if (Cs < 2077) {
                                                                if (RsCsDiff < 198) {
                                                                    if (Cs < 1958) {
                                                                        return 1;
                                                                    }
                                                                    else if (Cs >= 1958) {
                                                                        return 8;
                                                                    }
                                                                }
                                                                else if (RsCsDiff >= 198) {
                                                                    if (SC < 5938) {
                                                                        if (RsDiff < 13) {
                                                                            return 4;
                                                                        }
                                                                        else if (RsDiff >= 13) {
                                                                            return 2;
                                                                        }
                                                                    }
                                                                    else if (SC >= 5938) {
                                                                        return 4;
                                                                    }
                                                                }
                                                            }
                                                            else if (Cs >= 2077) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (mcTcor >= 95) {
                                                    if (CsDiff < 9) {
                                                        if (Rs < 3195) {
                                                            if (Rs < 1704) {
                                                                return 4;
                                                            }
                                                            else if (Rs >= 1704) {
                                                                return 2;
                                                            }
                                                        }
                                                        else if (Rs >= 3195) {
                                                            if (Cs < 2093) {
                                                                return 4;
                                                            }
                                                            else if (Cs >= 2093) {
                                                                if (tcor < 93) {
                                                                    if (RsDiff < 12) {
                                                                        return 4;
                                                                    }
                                                                    else if (RsDiff >= 12) {
                                                                        return 2;
                                                                    }
                                                                }
                                                                else if (tcor >= 93) {
                                                                    if (negBalance < 32) {
                                                                        if (posBalance < 48) {
                                                                            return 4;
                                                                        }
                                                                        else if (posBalance >= 48) {
                                                                            return 2;
                                                                        }
                                                                    }
                                                                    else if (negBalance >= 32) {
                                                                        return 2;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (CsDiff >= 9) {
                                                        return 2;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (Cs >= 2348) {
                                    if (AFD < 14) {
                                        if (tcor < 90) {
                                            if (posBalance < 49) {
                                                if (TSC < 10) {
                                                    if (SC < 6946) {
                                                        if (MVDiff < 32) {
                                                            return 8;
                                                        }
                                                        else if (MVDiff >= 32) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (SC >= 6946) {
                                                        if (AFD < 12) {
                                                            return 1;
                                                        }
                                                        else if (AFD >= 12) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                                else if (TSC >= 10) {
                                                    if (negBalance < 42) {
                                                        return 1;
                                                    }
                                                    else if (negBalance >= 42) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                            else if (posBalance >= 49) {
                                                if (MVDiff < 32) {
                                                    if (Rs < 3252) {
                                                        if (Cs < 2411) {
                                                            return 4;
                                                        }
                                                        else if (Cs >= 2411) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (Rs >= 3252) {
                                                        return 4;
                                                    }
                                                }
                                                else if (MVDiff >= 32) {
                                                    return 8;
                                                }
                                            }
                                        }
                                        else if (tcor >= 90) {
                                            if (negBalance < 42) {
                                                if (Cs < 2512) {
                                                    return 2;
                                                }
                                                else if (Cs >= 2512) {
                                                    if (Rs < 4228) {
                                                        return 4;
                                                    }
                                                    else if (Rs >= 4228) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                            else if (negBalance >= 42) {
                                                return 8;
                                            }
                                        }
                                    }
                                    else if (AFD >= 14) {
                                        if (Scindex < 6) {
                                            if (posBalance < 55) {
                                                if (diffMVdiffVal < 4) {
                                                    if (ssDCval < 97) {
                                                        if (negBalance < 48) {
                                                            return 1;
                                                        }
                                                        else if (negBalance >= 48) {
                                                            if (Rs < 2495) {
                                                                return 4;
                                                            }
                                                            else if (Rs >= 2495) {
                                                                return 1;
                                                            }
                                                        }
                                                    }
                                                    else if (ssDCval >= 97) {
                                                        if (Cs < 2485) {
                                                            return 4;
                                                        }
                                                        else if (Cs >= 2485) {
                                                            if (Cs < 2633) {
                                                                return 2;
                                                            }
                                                            else if (Cs >= 2633) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (diffMVdiffVal >= 4) {
                                                    if (Cs < 2478) {
                                                        if (Cs < 2431) {
                                                            return 8;
                                                        }
                                                        else if (Cs >= 2431) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (Cs >= 2478) {
                                                        if (Cs < 2822) {
                                                            return 1;
                                                        }
                                                        else if (Cs >= 2822) {
                                                            if (CsDiff < 15) {
                                                                return 8;
                                                            }
                                                            else if (CsDiff >= 15) {
                                                                return 1;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                            else if (posBalance >= 55) {
                                                if (Rs < 3161) {
                                                    if (Rs < 2349) {
                                                        if (Rs < 1495) {
                                                            if (MVDiff < 187) {
                                                                return 1;
                                                            }
                                                            else if (MVDiff >= 187) {
                                                                return 2;
                                                            }
                                                        }
                                                        else if (Rs >= 1495) {
                                                            if (mcTcor < 69) {
                                                                return 4;
                                                            }
                                                            else if (mcTcor >= 69) {
                                                                if (Rs < 2275) {
                                                                    return 1;
                                                                }
                                                                else if (Rs >= 2275) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (Rs >= 2349) {
                                                        if (ssDCval < 86) {
                                                            if (tcor < 64) {
                                                                if (Rs < 2915) {
                                                                    return 4;
                                                                }
                                                                else if (Rs >= 2915) {
                                                                    return 1;
                                                                }
                                                            }
                                                            else if (tcor >= 64) {
                                                                if (RsCsDiff < 191) {
                                                                    if (Cs < 2444) {
                                                                        return 2;
                                                                    }
                                                                    else if (Cs >= 2444) {
                                                                        return 1;
                                                                    }
                                                                }
                                                                else if (RsCsDiff >= 191) {
                                                                    if (diffRsCsdiff < -194) {
                                                                        return 1;
                                                                    }
                                                                    else if (diffRsCsdiff >= -194) {
                                                                        if (posBalance < 60) {
                                                                            return 8;
                                                                        }
                                                                        else if (posBalance >= 60) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (ssDCval >= 86) {
                                                            if (AFD < 35) {
                                                                if (mu_mv_mag_sq < 10) {
                                                                    return 4;
                                                                }
                                                                else if (mu_mv_mag_sq >= 10) {
                                                                    if (RsDiff < 14) {
                                                                        if (CsDiff < 10) {
                                                                            if (Rs < 3068) {
                                                                                return 1;
                                                                            }
                                                                            else if (Rs >= 3068) {
                                                                                return 2;
                                                                            }
                                                                        }
                                                                        else if (CsDiff >= 10) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                    else if (RsDiff >= 14) {
                                                                        if (SC < 5386) {
                                                                            return 4;
                                                                        }
                                                                        else if (SC >= 5386) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                            else if (AFD >= 35) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (Rs >= 3161) {
                                                    return 1;
                                                }
                                            }
                                        }
                                        else if (Scindex >= 6) {
                                            if (gchDC < 5) {
                                                if (mu_mv_mag_sq < 23) {
                                                    if (Rs < 3943) {
                                                        if (tcor < 44) {
                                                            return 2;
                                                        }
                                                        else if (tcor >= 44) {
                                                            if (CsDiff < 10) {
                                                                return 2;
                                                            }
                                                            else if (CsDiff >= 10) {
                                                                if (posBalance < 55) {
                                                                    if (SC < 6195) {
                                                                        return 1;
                                                                    }
                                                                    else if (SC >= 6195) {
                                                                        return 8;
                                                                    }
                                                                }
                                                                else if (posBalance >= 55) {
                                                                    return 1;
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (Rs >= 3943) {
                                                        if (mcTcor < 73) {
                                                            return 1;
                                                        }
                                                        else if (mcTcor >= 73) {
                                                            if (posBalance < 57) {
                                                                return 8;
                                                            }
                                                            else if (posBalance >= 57) {
                                                                if (diffMVdiffVal < 10) {
                                                                    return 2;
                                                                }
                                                                else if (diffMVdiffVal >= 10) {
                                                                    if (SC < 7225) {
                                                                        return 2;
                                                                    }
                                                                    else if (SC >= 7225) {
                                                                        return 1;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (mu_mv_mag_sq >= 23) {
                                                    if (CsDiff < 10) {
                                                        if (Cs < 3038) {
                                                            return 4;
                                                        }
                                                        else if (Cs >= 3038) {
                                                            if (Rs < 3894) {
                                                                return 8;
                                                            }
                                                            else if (Rs >= 3894) {
                                                                if (Rs < 4087) {
                                                                    return 4;
                                                                }
                                                                else if (Rs >= 4087) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (CsDiff >= 10) {
                                                        if (Cs < 3284) {
                                                            if (gchDC < 2) {
                                                                if (ssDCval < 102) {
                                                                    if (RsCsDiff < 291) {
                                                                        return 1;
                                                                    }
                                                                    else if (RsCsDiff >= 291) {
                                                                        return 8;
                                                                    }
                                                                }
                                                                else if (ssDCval >= 102) {
                                                                    return 1;
                                                                }
                                                            }
                                                            else if (gchDC >= 2) {
                                                                if (CsDiff < 16) {
                                                                    return 8;
                                                                }
                                                                else if (CsDiff >= 16) {
                                                                    return 2;
                                                                }
                                                            }
                                                        }
                                                        else if (Cs >= 3284) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (gchDC >= 5) {
                                                return 8;
                                            }
                                        }
                                    }
                                }
                            }
                        }
                    }
                }
            }
            else if (Rs >= 4380) {
                if (Cs < 4180) {
                    if (mu_mv_mag_sq < 127) {
                        if (mcTcor < 85) {
                            if (refDCval < 122) {
                                if (ssDCval < 106) {
                                    if (ssDCval < 94) {
                                        if (AFD < 28) {
                                            if (Rs < 4503) {
                                                return 4;
                                            }
                                            else if (Rs >= 4503) {
                                                return 8;
                                            }
                                        }
                                        else if (AFD >= 28) {
                                            return 2;
                                        }
                                    }
                                    else if (ssDCval >= 94) {
                                        return 4;
                                    }
                                }
                                else if (ssDCval >= 106) {
                                    if (ssDCval < 113) {
                                        if (posBalance < 44) {
                                            return 1;
                                        }
                                        else if (posBalance >= 44) {
                                            if (Cs < 2914) {
                                                if (Cs < 2748) {
                                                    return 2;
                                                }
                                                else if (Cs >= 2748) {
                                                    return 4;
                                                }
                                            }
                                            else if (Cs >= 2914) {
                                                return 2;
                                            }
                                        }
                                    }
                                    else if (ssDCval >= 113) {
                                        if (SC < 8082) {
                                            return 8;
                                        }
                                        else if (SC >= 8082) {
                                            return 4;
                                        }
                                    }
                                }
                            }
                            else if (refDCval >= 122) {
                                if (ssDCval < 138) {
                                    if (TSC < 16) {
                                        if (posBalance < 53) {
                                            if (CsDiff < 13) {
                                                return 8;
                                            }
                                            else if (CsDiff >= 13) {
                                                return 4;
                                            }
                                        }
                                        else if (posBalance >= 53) {
                                            return 1;
                                        }
                                    }
                                    else if (TSC >= 16) {
                                        if (Cs < 3557) {
                                            return 1;
                                        }
                                        else if (Cs >= 3557) {
                                            if (SC < 9407) {
                                                return 8;
                                            }
                                            else if (SC >= 9407) {
                                                return 4;
                                            }
                                        }
                                    }
                                }
                                else if (ssDCval >= 138) {
                                    if (AFD < 23) {
                                        if (Cs < 3984) {
                                            return 2;
                                        }
                                        else if (Cs >= 3984) {
                                            return 8;
                                        }
                                    }
                                    else if (AFD >= 23) {
                                        if (Rs < 5213) {
                                            return 8;
                                        }
                                        else if (Rs >= 5213) {
                                            return 4;
                                        }
                                    }
                                }
                            }
                        }
                        else if (mcTcor >= 85) {
                            if (ssDCval < 154) {
                                if (refDCval < 127) {
                                    if (negBalance < 55) {
                                        if (MVDiff < 13) {
                                            if (Rs < 4730) {
                                                if (mcTcor < 95) {
                                                    if (ssDCval < 111) {
                                                        if (diffAFD < -1) {
                                                            return 4;
                                                        }
                                                        else if (diffAFD >= -1) {
                                                            if (diffMVdiffVal < -4) {
                                                                if (RsDiff < 11) {
                                                                    return 2;
                                                                }
                                                                else if (RsDiff >= 11) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (diffMVdiffVal >= -4) {
                                                                if (MVDiff < 11) {
                                                                    return 8;
                                                                }
                                                                else if (MVDiff >= 11) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (ssDCval >= 111) {
                                                        if (posBalance < 43) {
                                                            if (Rs < 4630) {
                                                                return 8;
                                                            }
                                                            else if (Rs >= 4630) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (posBalance >= 43) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                                else if (mcTcor >= 95) {
                                                    if (refDCval < 121) {
                                                        return 8;
                                                    }
                                                    else if (refDCval >= 121) {
                                                        if (Cs < 2531) {
                                                            return 4;
                                                        }
                                                        else if (Cs >= 2531) {
                                                            return 2;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (Rs >= 4730) {
                                                if (refDCval < 115) {
                                                    if (tcor < 90) {
                                                        if (mu_mv_mag_sq < 80) {
                                                            if (SC < 8007) {
                                                                return 2;
                                                            }
                                                            else if (SC >= 8007) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (mu_mv_mag_sq >= 80) {
                                                            if (mcTcor < 89) {
                                                                if (negBalance < 38) {
                                                                    if (ssDCval < 112) {
                                                                        return 2;
                                                                    }
                                                                    else if (ssDCval >= 112) {
                                                                        return 8;
                                                                    }
                                                                }
                                                                else if (negBalance >= 38) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (mcTcor >= 89) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                    else if (tcor >= 90) {
                                                        if (diffMVdiffVal < 2) {
                                                            return 2;
                                                        }
                                                        else if (diffMVdiffVal >= 2) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                                else if (refDCval >= 115) {
                                                    if (ssDCval < 115) {
                                                        return 4;
                                                    }
                                                    else if (ssDCval >= 115) {
                                                        if (SC < 7252) {
                                                            if (SC < 7068) {
                                                                return 8;
                                                            }
                                                            else if (SC >= 7068) {
                                                                if (negBalance < 34) {
                                                                    return 4;
                                                                }
                                                                else if (negBalance >= 34) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                        else if (SC >= 7252) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (MVDiff >= 13) {
                                            if (refDCval < 122) {
                                                if (mu_mv_mag_sq < 17) {
                                                    return 2;
                                                }
                                                else if (mu_mv_mag_sq >= 17) {
                                                    if (mu_mv_mag_sq < 72) {
                                                        return 4;
                                                    }
                                                    else if (mu_mv_mag_sq >= 72) {
                                                        if (Rs < 4957) {
                                                            return 2;
                                                        }
                                                        else if (Rs >= 4957) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (refDCval >= 122) {
                                                if (MVDiff < 17) {
                                                    return 4;
                                                }
                                                else if (MVDiff >= 17) {
                                                    return 8;
                                                }
                                            }
                                        }
                                    }
                                    else if (negBalance >= 55) {
                                        if (AFD < 14) {
                                            return 4;
                                        }
                                        else if (AFD >= 14) {
                                            if (diffAFD < 1) {
                                                if (Rs < 4472) {
                                                    return 2;
                                                }
                                                else if (Rs >= 4472) {
                                                    return 8;
                                                }
                                            }
                                            else if (diffAFD >= 1) {
                                                if (negBalance < 60) {
                                                    return 4;
                                                }
                                                else if (negBalance >= 60) {
                                                    return 8;
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (refDCval >= 127) {
                                    if (negBalance < 36) {
                                        return 2;
                                    }
                                    else if (negBalance >= 36) {
                                        if (RsDiff < 23) {
                                            if (Cs < 2911) {
                                                if (Cs < 1930) {
                                                    return 2;
                                                }
                                                else if (Cs >= 1930) {
                                                    if (Cs < 2631) {
                                                        return 1;
                                                    }
                                                    else if (Cs >= 2631) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                            else if (Cs >= 2911) {
                                                if (Rs < 6210) {
                                                    if (posBalance < 43) {
                                                        return 4;
                                                    }
                                                    else if (posBalance >= 43) {
                                                        if (diffRsCsdiff < -340) {
                                                            if (Rs < 5368) {
                                                                return 8;
                                                            }
                                                            else if (Rs >= 5368) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (diffRsCsdiff >= -340) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                                else if (Rs >= 6210) {
                                                    return 1;
                                                }
                                            }
                                        }
                                        else if (RsDiff >= 23) {
                                            if (AFD < 32) {
                                                return 4;
                                            }
                                            else if (AFD >= 32) {
                                                return 2;
                                            }
                                        }
                                    }
                                }
                            }
                            else if (ssDCval >= 154) {
                                if (AFD < 22) {
                                    if (MVDiff < 25) {
                                        return 2;
                                    }
                                    else if (MVDiff >= 25) {
                                        if (Rs < 4846) {
                                            return 8;
                                        }
                                        else if (Rs >= 4846) {
                                            if (AFD < 21) {
                                                return 2;
                                            }
                                            else if (AFD >= 21) {
                                                return 4;
                                            }
                                        }
                                    }
                                }
                                else if (AFD >= 22) {
                                    if (mu_mv_mag_sq < 3) {
                                        if (tcor < 77) {
                                            if (AFD < 23) {
                                                return 2;
                                            }
                                            else if (AFD >= 23) {
                                                return 4;
                                            }
                                        }
                                        else if (tcor >= 77) {
                                            return 2;
                                        }
                                    }
                                    else if (mu_mv_mag_sq >= 3) {
                                        return 4;
                                    }
                                }
                            }
                        }
                    }
                    else if (mu_mv_mag_sq >= 127) {
                        if (ssDCval < 125) {
                            return 4;
                        }
                        else if (ssDCval >= 125) {
                            return 8;
                        }
                    }
                }
                else if (Cs >= 4180) {
                    if (mu_mv_mag_sq < 63) {
                        if (Rs < 6902) {
                            if (mcTcor < 90) {
                                if (mu_mv_mag_sq < 2) {
                                    if (refDCval < 149) {
                                        if (RsDiff < 30) {
                                            if (Rs < 4968) {
                                                return 4;
                                            }
                                            else if (Rs >= 4968) {
                                                return 8;
                                            }
                                        }
                                        else if (RsDiff >= 30) {
                                            return 4;
                                        }
                                    }
                                    else if (refDCval >= 149) {
                                        return 2;
                                    }
                                }
                                else if (mu_mv_mag_sq >= 2) {
                                    if (RsDiff < 35) {
                                        if (RsDiff < 11) {
                                            return 4;
                                        }
                                        else if (RsDiff >= 11) {
                                            if (Cs < 4301) {
                                                if (refDCval < 134) {
                                                    return 8;
                                                }
                                                else if (refDCval >= 134) {
                                                    if (MVDiff < 15) {
                                                        return 8;
                                                    }
                                                    else if (MVDiff >= 15) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                            else if (Cs >= 4301) {
                                                if (negBalance < 52) {
                                                    if (refDCval < 135) {
                                                        if (SC < 12208) {
                                                            return 8;
                                                        }
                                                        else if (SC >= 12208) {
                                                            if (Rs < 6533) {
                                                                return 8;
                                                            }
                                                            else if (Rs >= 6533) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                    else if (refDCval >= 135) {
                                                        return 4;
                                                    }
                                                }
                                                else if (negBalance >= 52) {
                                                    return 8;
                                                }
                                            }
                                        }
                                    }
                                    else if (RsDiff >= 35) {
                                        if (Cs < 5104) {
                                            return 2;
                                        }
                                        else if (Cs >= 5104) {
                                            return 8;
                                        }
                                    }
                                }
                            }
                            else if (mcTcor >= 90) {
                                if (mu_mv_mag_sq < 32) {
                                    if (ssDCval < 111) {
                                        if (AFD < 19) {
                                            return 8;
                                        }
                                        else if (AFD >= 19) {
                                            return 1;
                                        }
                                    }
                                    else if (ssDCval >= 111) {
                                        if (Rs < 6383) {
                                            return 8;
                                        }
                                        else if (Rs >= 6383) {
                                            return 2;
                                        }
                                    }
                                }
                                else if (mu_mv_mag_sq >= 32) {
                                    if (refDCval < 68) {
                                        return 4;
                                    }
                                    else if (refDCval >= 68) {
                                        if (Rs < 6455) {
                                            if (CsDiff < 5) {
                                                return 4;
                                            }
                                            else if (CsDiff >= 5) {
                                                return 8;
                                            }
                                        }
                                        else if (Rs >= 6455) {
                                            if (SC < 11778) {
                                                return 4;
                                            }
                                            else if (SC >= 11778) {
                                                return 1;
                                            }
                                        }
                                    }
                                }
                            }
                        }
                        else if (Rs >= 6902) {
                            if (ssDCval < 77) {
                                return 4;
                            }
                            else if (ssDCval >= 77) {
                                if (MVDiff < 23) {
                                    if (negBalance < 51) {
                                        if (tcor < 84) {
                                            if (diffMVdiffVal < 9) {
                                                return 8;
                                            }
                                            else if (diffMVdiffVal >= 9) {
                                                if (Rs < 8157) {
                                                    return 8;
                                                }
                                                else if (Rs >= 8157) {
                                                    return 4;
                                                }
                                            }
                                        }
                                        else if (tcor >= 84) {
                                            return 4;
                                        }
                                    }
                                    else if (negBalance >= 51) {
                                        if (mcTcor < 84) {
                                            if (tcor < 66) {
                                                if (MVDiff < 17) {
                                                    return 8;
                                                }
                                                else if (MVDiff >= 17) {
                                                    return 4;
                                                }
                                            }
                                            else if (tcor >= 66) {
                                                if (Rs < 14784) {
                                                    return 4;
                                                }
                                                else if (Rs >= 14784) {
                                                    return 8;
                                                }
                                            }
                                        }
                                        else if (mcTcor >= 84) {
                                            if (RsCsDiff < 1161) {
                                                if (MVDiff < 12) {
                                                    return 4;
                                                }
                                                else if (MVDiff >= 12) {
                                                    if (TSC < 14) {
                                                        return 4;
                                                    }
                                                    else if (TSC >= 14) {
                                                        return 2;
                                                    }
                                                }
                                            }
                                            else if (RsCsDiff >= 1161) {
                                                if (diffAFD < 4) {
                                                    if (Rs < 8431) {
                                                        if (negBalance < 55) {
                                                            if (Rs < 7200) {
                                                                return 8;
                                                            }
                                                            else if (Rs >= 7200) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (negBalance >= 55) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (Rs >= 8431) {
                                                        if (ssDCval < 135) {
                                                            if (Rs < 9456) {
                                                                return 2;
                                                            }
                                                            else if (Rs >= 9456) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (ssDCval >= 135) {
                                                            if (AFD < 27) {
                                                                return 8;
                                                            }
                                                            else if (AFD >= 27) {
                                                                if (AFD < 35) {
                                                                    if (posBalance < 50) {
                                                                        return 8;
                                                                    }
                                                                    else if (posBalance >= 50) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (AFD >= 35) {
                                                                    return 2;
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (diffAFD >= 4) {
                                                    return 4;
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (MVDiff >= 23) {
                                    if (Cs < 6621) {
                                        if (mcTcor < 85) {
                                            if (Rs < 7572) {
                                                return 2;
                                            }
                                            else if (Rs >= 7572) {
                                                if (Cs < 5692) {
                                                    return 8;
                                                }
                                                else if (Cs >= 5692) {
                                                    return 4;
                                                }
                                            }
                                        }
                                        else if (mcTcor >= 85) {
                                            return 2;
                                        }
                                    }
                                    else if (Cs >= 6621) {
                                        return 8;
                                    }
                                }
                            }
                        }
                    }
                    else if (mu_mv_mag_sq >= 63) {
                        if (Rs < 7263) {
                            if (ssDCval < 106) {
                                if (TSC < 8) {
                                    if (tcor < 92) {
                                        if (Rs < 4829) {
                                            return 8;
                                        }
                                        else if (Rs >= 4829) {
                                            return 4;
                                        }
                                    }
                                    else if (tcor >= 92) {
                                        return 8;
                                    }
                                }
                                else if (TSC >= 8) {
                                    if (mu_mv_mag_sq < 66) {
                                        if (posBalance < 58) {
                                            if (negBalance < 49) {
                                                return 8;
                                            }
                                            else if (negBalance >= 49) {
                                                return 4;
                                            }
                                        }
                                        else if (posBalance >= 58) {
                                            return 8;
                                        }
                                    }
                                    else if (mu_mv_mag_sq >= 66) {
                                        return 8;
                                    }
                                }
                            }
                            else if (ssDCval >= 106) {
                                if (mu_mv_mag_sq < 122) {
                                    if (mu_mv_mag_sq < 84) {
                                        if (mcTcor < 74) {
                                            return 1;
                                        }
                                        else if (mcTcor >= 74) {
                                            return 8;
                                        }
                                    }
                                    else if (mu_mv_mag_sq >= 84) {
                                        return 8;
                                    }
                                }
                                else if (mu_mv_mag_sq >= 122) {
                                    if (posBalance < 54) {
                                        return 8;
                                    }
                                    else if (posBalance >= 54) {
                                        if (posBalance < 60) {
                                            return 4;
                                        }
                                        else if (posBalance >= 60) {
                                            if (negBalance < 53) {
                                                return 4;
                                            }
                                            else if (negBalance >= 53) {
                                                return 8;
                                            }
                                        }
                                    }
                                }
                            }
                        }
                        else if (Rs >= 7263) {
                            if (refDCval < 93) {
                                if (mu_mv_mag_sq < 103) {
                                    if (mu_mv_mag_sq < 71) {
                                        return 8;
                                    }
                                    else if (mu_mv_mag_sq >= 71) {
                                        return 4;
                                    }
                                }
                                else if (mu_mv_mag_sq >= 103) {
                                    if (mcTcor < 79) {
                                        if (Rs < 7697) {
                                            if (AFD < 21) {
                                                return 4;
                                            }
                                            else if (AFD >= 21) {
                                                return 2;
                                            }
                                        }
                                        else if (Rs >= 7697) {
                                            if (RsCsDiff < 1036) {
                                                return 2;
                                            }
                                            else if (RsCsDiff >= 1036) {
                                                if (posBalance < 52) {
                                                    return 2;
                                                }
                                                else if (posBalance >= 52) {
                                                    return 4;
                                                }
                                            }
                                        }
                                    }
                                    else if (mcTcor >= 79) {
                                        return 2;
                                    }
                                }
                            }
                            else if (refDCval >= 93) {
                                if (RsDiff < 14) {
                                    return 4;
                                }
                                else if (RsDiff >= 14) {
                                    return 8;
                                }
                            }
                        }
                    }
                }
            }
        }
        return 4;
    }
    mfxU16 AGOPSelect6(mfxU32 Rs, mfxU32 CsDiff, mfxU32 TSC, mfxI32 diffAFD, mfxU32 Scindex, mfxU32 RsDiff, mfxI32 diffMVdiffVal, mfxU32 RsCsDiff, mfxU32 TSCindex, mfxI16 tcor, mfxU32 refDCval, mfxU32 MVDiff, mfxU32 AFD, mfxI32 diffTSC, mfxI32 diffRsCsdiff, mfxU32 posBalance, mfxU32 SC, mfxU32 mu_mv_mag_sq, mfxU32 negBalance, mfxI16 mcTcor, mfxU32 gchDC, mfxU32 Cs, mfxU32 ssDCval) {
        if (negBalance < 32) {
            if (Cs < 708) {
                if (MVDiff < 6) {
                    if (Rs < 629) {
                        if (Cs < 533) {
                            if (ssDCval < 161) {
                                if (TSC < 5) {
                                    if (ssDCval < 96) {
                                        if (negBalance < 25) {
                                            if (posBalance < 30) {
                                                return 8;
                                            }
                                            else if (posBalance >= 30) {
                                                if (Rs < 183) {
                                                    if (negBalance < 15) {
                                                        if (mu_mv_mag_sq < 127) {
                                                            return 4;
                                                        }
                                                        else if (mu_mv_mag_sq >= 127) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (negBalance >= 15) {
                                                        return 4;
                                                    }
                                                }
                                                else if (Rs >= 183) {
                                                    if (TSC < 2) {
                                                        return 8;
                                                    }
                                                    else if (TSC >= 2) {
                                                        if (AFD < 3) {
                                                            if (TSCindex < 3) {
                                                                return 8;
                                                            }
                                                            else if (TSCindex >= 3) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (AFD >= 3) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (negBalance >= 25) {
                                            return 4;
                                        }
                                    }
                                    else if (ssDCval >= 96) {
                                        return 8;
                                    }
                                }
                                else if (TSC >= 5) {
                                    return 2;
                                }
                            }
                            else if (ssDCval >= 161) {
                                if (Rs < 443) {
                                    return 4;
                                }
                                else if (Rs >= 443) {
                                    if (SC < 998) {
                                        if (posBalance < 24) {
                                            return 4;
                                        }
                                        else if (posBalance >= 24) {
                                            if (Rs < 512) {
                                                return 2;
                                            }
                                            else if (Rs >= 512) {
                                                return 4;
                                            }
                                        }
                                    }
                                    else if (SC >= 998) {
                                        return 4;
                                    }
                                }
                            }
                        }
                        else if (Cs >= 533) {
                            if (gchDC < 1) {
                                return 4;
                            }
                            else if (gchDC >= 1) {
                                if (RsDiff < 1) {
                                    return 8;
                                }
                                else if (RsDiff >= 1) {
                                    if (posBalance < 30) {
                                        return 8;
                                    }
                                    else if (posBalance >= 30) {
                                        return 4;
                                    }
                                }
                            }
                        }
                    }
                    else if (Rs >= 629) {
                        if (Cs < 703) {
                            if (diffMVdiffVal < -145) {
                                return 2;
                            }
                            else if (diffMVdiffVal >= -145) {
                                if (SC < 884) {
                                    return 1;
                                }
                                else if (SC >= 884) {
                                    if (gchDC < 3) {
                                        if (ssDCval < 54) {
                                            if (TSC < 3) {
                                                return 8;
                                            }
                                            else if (TSC >= 3) {
                                                return 4;
                                            }
                                        }
                                        else if (ssDCval >= 54) {
                                            if (diffMVdiffVal < -4) {
                                                if (Cs < 638) {
                                                    return 8;
                                                }
                                                else if (Cs >= 638) {
                                                    return 1;
                                                }
                                            }
                                            else if (diffMVdiffVal >= -4) {
                                                return 8;
                                            }
                                        }
                                    }
                                    else if (gchDC >= 3) {
                                        if (RsCsDiff < 12) {
                                            return 8;
                                        }
                                        else if (RsCsDiff >= 12) {
                                            return 4;
                                        }
                                    }
                                }
                            }
                        }
                        else if (Cs >= 703) {
                            if (Rs < 997) {
                                if (Cs < 706) {
                                    return 8;
                                }
                                else if (Cs >= 706) {
                                    return 1;
                                }
                            }
                            else if (Rs >= 997) {
                                return 4;
                            }
                        }
                    }
                }
                else if (MVDiff >= 6) {
                    if (mu_mv_mag_sq < 4) {
                        if (mcTcor < 63) {
                            return 1;
                        }
                        else if (mcTcor >= 63) {
                            if (tcor < 95) {
                                if (diffAFD < 1) {
                                    return 2;
                                }
                                else if (diffAFD >= 1) {
                                    if (mcTcor < 80) {
                                        return 2;
                                    }
                                    else if (mcTcor >= 80) {
                                        if (ssDCval < 38) {
                                            return 4;
                                        }
                                        else if (ssDCval >= 38) {
                                            if (SC < 246) {
                                                return 2;
                                            }
                                            else if (SC >= 246) {
                                                if (AFD < 5) {
                                                    return 2;
                                                }
                                                else if (AFD >= 5) {
                                                    return 4;
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                            else if (tcor >= 95) {
                                if (AFD < 3) {
                                    return 1;
                                }
                                else if (AFD >= 3) {
                                    return 4;
                                }
                            }
                        }
                    }
                    else if (mu_mv_mag_sq >= 4) {
                        if (MVDiff < 41) {
                            if (mcTcor < 86) {
                                if (diffMVdiffVal < 23) {
                                    if (mu_mv_mag_sq < 29) {
                                        if (Rs < 1) {
                                            return 4;
                                        }
                                        else if (Rs >= 1) {
                                            if (ssDCval < 38) {
                                                return 2;
                                            }
                                            else if (ssDCval >= 38) {
                                                return 8;
                                            }
                                        }
                                    }
                                    else if (mu_mv_mag_sq >= 29) {
                                        if (negBalance < 29) {
                                            return 4;
                                        }
                                        else if (negBalance >= 29) {
                                            if (Rs < 60) {
                                                return 8;
                                            }
                                            else if (Rs >= 60) {
                                                return 4;
                                            }
                                        }
                                    }
                                }
                                else if (diffMVdiffVal >= 23) {
                                    return 8;
                                }
                            }
                            else if (mcTcor >= 86) {
                                if (ssDCval < 153) {
                                    if (Rs < 290) {
                                        if (mu_mv_mag_sq < 49) {
                                            if (Cs < 44) {
                                                return 8;
                                            }
                                            else if (Cs >= 44) {
                                                return 2;
                                            }
                                        }
                                        else if (mu_mv_mag_sq >= 49) {
                                            if (MVDiff < 31) {
                                                if (gchDC < 4) {
                                                    if (AFD < 1) {
                                                        if (ssDCval < 23) {
                                                            return 8;
                                                        }
                                                        else if (ssDCval >= 23) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (AFD >= 1) {
                                                        return 8;
                                                    }
                                                }
                                                else if (gchDC >= 4) {
                                                    return 4;
                                                }
                                            }
                                            else if (MVDiff >= 31) {
                                                return 8;
                                            }
                                        }
                                    }
                                    else if (Rs >= 290) {
                                        if (negBalance < 13) {
                                            if (posBalance < 94) {
                                                if (AFD < 1) {
                                                    return 4;
                                                }
                                                else if (AFD >= 1) {
                                                    if (diffMVdiffVal < 8) {
                                                        return 8;
                                                    }
                                                    else if (diffMVdiffVal >= 8) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                            else if (posBalance >= 94) {
                                                return 4;
                                            }
                                        }
                                        else if (negBalance >= 13) {
                                            if (refDCval < 105) {
                                                if (RsDiff < 4) {
                                                    if (RsCsDiff < 2) {
                                                        return 8;
                                                    }
                                                    else if (RsCsDiff >= 2) {
                                                        return 4;
                                                    }
                                                }
                                                else if (RsDiff >= 4) {
                                                    return 2;
                                                }
                                            }
                                            else if (refDCval >= 105) {
                                                if (SC < 1759) {
                                                    if (AFD < 6) {
                                                        return 1;
                                                    }
                                                    else if (AFD >= 6) {
                                                        return 8;
                                                    }
                                                }
                                                else if (SC >= 1759) {
                                                    return 8;
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (ssDCval >= 153) {
                                    if (negBalance < 12) {
                                        if (Rs < 485) {
                                            return 4;
                                        }
                                        else if (Rs >= 485) {
                                            return 2;
                                        }
                                    }
                                    else if (negBalance >= 12) {
                                        return 4;
                                    }
                                }
                            }
                        }
                        else if (MVDiff >= 41) {
                            if (ssDCval < 160) {
                                if (tcor < 63) {
                                    if (refDCval < 22) {
                                        if (TSCindex < 1) {
                                            if (MVDiff < 62) {
                                                if (negBalance < 2) {
                                                    if (tcor < 37) {
                                                        return 8;
                                                    }
                                                    else if (tcor >= 37) {
                                                        return 1;
                                                    }
                                                }
                                                else if (negBalance >= 2) {
                                                    return 1;
                                                }
                                            }
                                            else if (MVDiff >= 62) {
                                                if (diffTSC < -1) {
                                                    if (diffMVdiffVal < 118) {
                                                        if (MVDiff < 178) {
                                                            return 1;
                                                        }
                                                        else if (MVDiff >= 178) {
                                                            if (AFD < 1) {
                                                                return 4;
                                                            }
                                                            else if (AFD >= 1) {
                                                                return 1;
                                                            }
                                                        }
                                                    }
                                                    else if (diffMVdiffVal >= 118) {
                                                        return 8;
                                                    }
                                                }
                                                else if (diffTSC >= -1) {
                                                    if (MVDiff < 174) {
                                                        if (MVDiff < 65) {
                                                            return 4;
                                                        }
                                                        else if (MVDiff >= 65) {
                                                            if (mu_mv_mag_sq < 25) {
                                                                if (tcor < 36) {
                                                                    return 1;
                                                                }
                                                                else if (tcor >= 36) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (mu_mv_mag_sq >= 25) {
                                                                if (diffMVdiffVal < 4) {
                                                                    if (posBalance < 27) {
                                                                        if (mu_mv_mag_sq < 32) {
                                                                            return 8;
                                                                        }
                                                                        else if (mu_mv_mag_sq >= 32) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                    else if (posBalance >= 27) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (diffMVdiffVal >= 4) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (MVDiff >= 174) {
                                                        if (tcor < 57) {
                                                            return 4;
                                                        }
                                                        else if (tcor >= 57) {
                                                            return 2;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (TSCindex >= 1) {
                                            if (diffAFD < -1) {
                                                if (AFD < 3) {
                                                    return 1;
                                                }
                                                else if (AFD >= 3) {
                                                    return 4;
                                                }
                                            }
                                            else if (diffAFD >= -1) {
                                                if (posBalance < 83) {
                                                    if (ssDCval < 19) {
                                                        return 8;
                                                    }
                                                    else if (ssDCval >= 19) {
                                                        return 2;
                                                    }
                                                }
                                                else if (posBalance >= 83) {
                                                    if (posBalance < 122) {
                                                        return 4;
                                                    }
                                                    else if (posBalance >= 122) {
                                                        return 2;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (refDCval >= 22) {
                                        if (AFD < 7) {
                                            return 1;
                                        }
                                        else if (AFD >= 7) {
                                            if (tcor < 25) {
                                                if (Rs < 464) {
                                                    return 1;
                                                }
                                                else if (Rs >= 464) {
                                                    return 8;
                                                }
                                            }
                                            else if (tcor >= 25) {
                                                return 2;
                                            }
                                        }
                                    }
                                }
                                else if (tcor >= 63) {
                                    if (diffMVdiffVal < 43) {
                                        if (mu_mv_mag_sq < 101) {
                                            if (Cs < 11) {
                                                if (diffAFD < -1) {
                                                    return 2;
                                                }
                                                else if (diffAFD >= -1) {
                                                    if (MVDiff < 82) {
                                                        if (MVDiff < 64) {
                                                            if (AFD < 1) {
                                                                return 4;
                                                            }
                                                            else if (AFD >= 1) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (MVDiff >= 64) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (MVDiff >= 82) {
                                                        if (MVDiff < 113) {
                                                            return 8;
                                                        }
                                                        else if (MVDiff >= 113) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (Cs >= 11) {
                                                if (TSC < 3) {
                                                    return 2;
                                                }
                                                else if (TSC >= 3) {
                                                    if (CsDiff < 2) {
                                                        return 4;
                                                    }
                                                    else if (CsDiff >= 2) {
                                                        if (AFD < 8) {
                                                            return 4;
                                                        }
                                                        else if (AFD >= 8) {
                                                            return 2;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (mu_mv_mag_sq >= 101) {
                                            if (MVDiff < 160) {
                                                return 8;
                                            }
                                            else if (MVDiff >= 160) {
                                                return 4;
                                            }
                                        }
                                    }
                                    else if (diffMVdiffVal >= 43) {
                                        if (negBalance < 20) {
                                            if (refDCval < 37) {
                                                if (diffMVdiffVal < 76) {
                                                    return 8;
                                                }
                                                else if (diffMVdiffVal >= 76) {
                                                    if (MVDiff < 248) {
                                                        return 2;
                                                    }
                                                    else if (MVDiff >= 248) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                            else if (refDCval >= 37) {
                                                return 2;
                                            }
                                        }
                                        else if (negBalance >= 20) {
                                            if (Cs < 294) {
                                                return 1;
                                            }
                                            else if (Cs >= 294) {
                                                return 8;
                                            }
                                        }
                                    }
                                }
                            }
                            else if (ssDCval >= 160) {
                                if (RsDiff < 7) {
                                    return 1;
                                }
                                else if (RsDiff >= 7) {
                                    return 4;
                                }
                            }
                        }
                    }
                }
            }
            else if (Cs >= 708) {
                if (MVDiff < 5) {
                    if (mu_mv_mag_sq < 94) {
                        if (SC < 3301) {
                            if (ssDCval < 41) {
                                return 8;
                            }
                            else if (ssDCval >= 41) {
                                if (Cs < 862) {
                                    if (TSC < 3) {
                                        return 4;
                                    }
                                    else if (TSC >= 3) {
                                        return 8;
                                    }
                                }
                                else if (Cs >= 862) {
                                    if (Cs < 1071) {
                                        if (ssDCval < 79) {
                                            return 1;
                                        }
                                        else if (ssDCval >= 79) {
                                            return 8;
                                        }
                                    }
                                    else if (Cs >= 1071) {
                                        if (refDCval < 102) {
                                            return 4;
                                        }
                                        else if (refDCval >= 102) {
                                            return 8;
                                        }
                                    }
                                }
                            }
                        }
                        else if (SC >= 3301) {
                            if (refDCval < 80) {
                                if (Rs < 4675) {
                                    if (SC < 4939) {
                                        if (Cs < 1547) {
                                            return 4;
                                        }
                                        else if (Cs >= 1547) {
                                            return 1;
                                        }
                                    }
                                    else if (SC >= 4939) {
                                        if (SC < 7709) {
                                            return 8;
                                        }
                                        else if (SC >= 7709) {
                                            if (Rs < 3791) {
                                                if (Cs < 3985) {
                                                    return 1;
                                                }
                                                else if (Cs >= 3985) {
                                                    return 8;
                                                }
                                            }
                                            else if (Rs >= 3791) {
                                                return 8;
                                            }
                                        }
                                    }
                                }
                                else if (Rs >= 4675) {
                                    if (Rs < 5376) {
                                        return 1;
                                    }
                                    else if (Rs >= 5376) {
                                        return 8;
                                    }
                                }
                            }
                            else if (refDCval >= 80) {
                                if (Cs < 2433) {
                                    if (refDCval < 107) {
                                        if (gchDC < 3) {
                                            if (diffAFD < -1) {
                                                return 4;
                                            }
                                            else if (diffAFD >= -1) {
                                                return 8;
                                            }
                                        }
                                        else if (gchDC >= 3) {
                                            return 4;
                                        }
                                    }
                                    else if (refDCval >= 107) {
                                        return 4;
                                    }
                                }
                                else if (Cs >= 2433) {
                                    return 8;
                                }
                            }
                        }
                    }
                    else if (mu_mv_mag_sq >= 94) {
                        if (refDCval < 103) {
                            if (mu_mv_mag_sq < 121) {
                                if (refDCval < 50) {
                                    if (Rs < 1065) {
                                        if (Rs < 1037) {
                                            if (SC < 2174) {
                                                return 8;
                                            }
                                            else if (SC >= 2174) {
                                                if (Cs < 1234) {
                                                    return 4;
                                                }
                                                else if (Cs >= 1234) {
                                                    return 8;
                                                }
                                            }
                                        }
                                        else if (Rs >= 1037) {
                                            if (diffMVdiffVal < -1) {
                                                return 4;
                                            }
                                            else if (diffMVdiffVal >= -1) {
                                                return 2;
                                            }
                                        }
                                    }
                                    else if (Rs >= 1065) {
                                        return 8;
                                    }
                                }
                                else if (refDCval >= 50) {
                                    if (SC < 4641) {
                                        if (diffAFD < -2) {
                                            if (Cs < 2150) {
                                                return 1;
                                            }
                                            else if (Cs >= 2150) {
                                                return 8;
                                            }
                                        }
                                        else if (diffAFD >= -2) {
                                            if (SC < 4243) {
                                                if (MVDiff < 3) {
                                                    return 8;
                                                }
                                                else if (MVDiff >= 3) {
                                                    if (Cs < 1444) {
                                                        return 1;
                                                    }
                                                    else if (Cs >= 1444) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                            else if (SC >= 4243) {
                                                if (Rs < 2134) {
                                                    return 1;
                                                }
                                                else if (Rs >= 2134) {
                                                    if (SC < 4291) {
                                                        return 1;
                                                    }
                                                    else if (SC >= 4291) {
                                                        if (Cs < 1634) {
                                                            return 1;
                                                        }
                                                        else if (Cs >= 1634) {
                                                            if (RsCsDiff < 37) {
                                                                if (Rs < 2299) {
                                                                    return 8;
                                                                }
                                                                else if (Rs >= 2299) {
                                                                    return 1;
                                                                }
                                                            }
                                                            else if (RsCsDiff >= 37) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (SC >= 4641) {
                                        return 8;
                                    }
                                }
                            }
                            else if (mu_mv_mag_sq >= 121) {
                                if (TSC < 6) {
                                    if (CsDiff < 4) {
                                        return 8;
                                    }
                                    else if (CsDiff >= 4) {
                                        if (refDCval < 51) {
                                            if (posBalance < 32) {
                                                return 8;
                                            }
                                            else if (posBalance >= 32) {
                                                if (SC < 4621) {
                                                    return 4;
                                                }
                                                else if (SC >= 4621) {
                                                    return 8;
                                                }
                                            }
                                        }
                                        else if (refDCval >= 51) {
                                            return 8;
                                        }
                                    }
                                }
                                else if (TSC >= 6) {
                                    if (Cs < 2902) {
                                        return 1;
                                    }
                                    else if (Cs >= 2902) {
                                        return 8;
                                    }
                                }
                            }
                        }
                        else if (refDCval >= 103) {
                            if (negBalance < 3) {
                                if (Rs < 3779) {
                                    if (SC < 7764) {
                                        if (Rs < 3730) {
                                            if (diffRsCsdiff < 2) {
                                                if (refDCval < 140) {
                                                    if (posBalance < 1) {
                                                        return 1;
                                                    }
                                                    else if (posBalance >= 1) {
                                                        if (Rs < 3596) {
                                                            if (diffRsCsdiff < -42) {
                                                                if (Rs < 2844) {
                                                                    return 1;
                                                                }
                                                                else if (Rs >= 2844) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (diffRsCsdiff >= -42) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (Rs >= 3596) {
                                                            if (Rs < 3632) {
                                                                return 1;
                                                            }
                                                            else if (Rs >= 3632) {
                                                                if (Cs < 3849) {
                                                                    if (Rs < 3676) {
                                                                        return 8;
                                                                    }
                                                                    else if (Rs >= 3676) {
                                                                        return 1;
                                                                    }
                                                                }
                                                                else if (Cs >= 3849) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (refDCval >= 140) {
                                                    if (Rs < 1050) {
                                                        return 8;
                                                    }
                                                    else if (Rs >= 1050) {
                                                        return 1;
                                                    }
                                                }
                                            }
                                            else if (diffRsCsdiff >= 2) {
                                                if (Rs < 3538) {
                                                    if (Cs < 1401) {
                                                        if (Rs < 1145) {
                                                            return 8;
                                                        }
                                                        else if (Rs >= 1145) {
                                                            return 1;
                                                        }
                                                    }
                                                    else if (Cs >= 1401) {
                                                        return 8;
                                                    }
                                                }
                                                else if (Rs >= 3538) {
                                                    return 1;
                                                }
                                            }
                                        }
                                        else if (Rs >= 3730) {
                                            return 8;
                                        }
                                    }
                                    else if (SC >= 7764) {
                                        return 1;
                                    }
                                }
                                else if (Rs >= 3779) {
                                    return 8;
                                }
                            }
                            else if (negBalance >= 3) {
                                if (AFD < 2) {
                                    if (diffMVdiffVal < 1) {
                                        if (Cs < 1340) {
                                            if (Rs < 1354) {
                                                return 8;
                                            }
                                            else if (Rs >= 1354) {
                                                if (Rs < 1424) {
                                                    return 4;
                                                }
                                                else if (Rs >= 1424) {
                                                    return 8;
                                                }
                                            }
                                        }
                                        else if (Cs >= 1340) {
                                            return 8;
                                        }
                                    }
                                    else if (diffMVdiffVal >= 1) {
                                        if (SC < 4816) {
                                            return 8;
                                        }
                                        else if (SC >= 4816) {
                                            return 4;
                                        }
                                    }
                                }
                                else if (AFD >= 2) {
                                    if (negBalance < 7) {
                                        if (Rs < 1261) {
                                            return 8;
                                        }
                                        else if (Rs >= 1261) {
                                            if (RsDiff < 4) {
                                                if (MVDiff < 1) {
                                                    return 8;
                                                }
                                                else if (MVDiff >= 1) {
                                                    return 4;
                                                }
                                            }
                                            else if (RsDiff >= 4) {
                                                if (Rs < 1307) {
                                                    if (Rs < 1287) {
                                                        return 1;
                                                    }
                                                    else if (Rs >= 1287) {
                                                        return 8;
                                                    }
                                                }
                                                else if (Rs >= 1307) {
                                                    return 1;
                                                }
                                            }
                                        }
                                    }
                                    else if (negBalance >= 7) {
                                        if (Cs < 2938) {
                                            if (ssDCval < 120) {
                                                if (tcor < 97) {
                                                    if (refDCval < 117) {
                                                        if (TSCindex < 2) {
                                                            if (SC < 4666) {
                                                                return 4;
                                                            }
                                                            else if (SC >= 4666) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (TSCindex >= 2) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (refDCval >= 117) {
                                                        if (mu_mv_mag_sq < 125) {
                                                            return 4;
                                                        }
                                                        else if (mu_mv_mag_sq >= 125) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                                else if (tcor >= 97) {
                                                    if (RsDiff < 6) {
                                                        if (posBalance < 51) {
                                                            return 8;
                                                        }
                                                        else if (posBalance >= 51) {
                                                            if (Rs < 2179) {
                                                                return 4;
                                                            }
                                                            else if (Rs >= 2179) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                    else if (RsDiff >= 6) {
                                                        return 1;
                                                    }
                                                }
                                            }
                                            else if (ssDCval >= 120) {
                                                return 8;
                                            }
                                        }
                                        else if (Cs >= 2938) {
                                            if (SC < 6455) {
                                                if (ssDCval < 109) {
                                                    if (posBalance < 42) {
                                                        if (mu_mv_mag_sq < 120) {
                                                            return 8;
                                                        }
                                                        else if (mu_mv_mag_sq >= 120) {
                                                            if (refDCval < 108) {
                                                                return 8;
                                                            }
                                                            else if (refDCval >= 108) {
                                                                return 1;
                                                            }
                                                        }
                                                    }
                                                    else if (posBalance >= 42) {
                                                        if (mu_mv_mag_sq < 115) {
                                                            return 8;
                                                        }
                                                        else if (mu_mv_mag_sq >= 115) {
                                                            return 1;
                                                        }
                                                    }
                                                }
                                                else if (ssDCval >= 109) {
                                                    return 8;
                                                }
                                            }
                                            else if (SC >= 6455) {
                                                return 8;
                                            }
                                        }
                                    }
                                }
                            }
                        }
                    }
                }
                else if (MVDiff >= 5) {
                    if (SC < 2754) {
                        if (posBalance < 33) {
                            if (ssDCval < 29) {
                                return 8;
                            }
                            else if (ssDCval >= 29) {
                                if (mcTcor < 98) {
                                    if (RsCsDiff < 92) {
                                        if (refDCval < 80) {
                                            return 8;
                                        }
                                        else if (refDCval >= 80) {
                                            if (MVDiff < 9) {
                                                if (Cs < 810) {
                                                    if (ssDCval < 150) {
                                                        return 8;
                                                    }
                                                    else if (ssDCval >= 150) {
                                                        return 4;
                                                    }
                                                }
                                                else if (Cs >= 810) {
                                                    if (AFD < 2) {
                                                        return 8;
                                                    }
                                                    else if (AFD >= 2) {
                                                        return 1;
                                                    }
                                                }
                                            }
                                            else if (MVDiff >= 9) {
                                                if (TSC < 4) {
                                                    return 1;
                                                }
                                                else if (TSC >= 4) {
                                                    return 8;
                                                }
                                            }
                                        }
                                    }
                                    else if (RsCsDiff >= 92) {
                                        return 4;
                                    }
                                }
                                else if (mcTcor >= 98) {
                                    if (posBalance < 11) {
                                        return 1;
                                    }
                                    else if (posBalance >= 11) {
                                        if (ssDCval < 30) {
                                            return 8;
                                        }
                                        else if (ssDCval >= 30) {
                                            if (Rs < 1500) {
                                                return 4;
                                            }
                                            else if (Rs >= 1500) {
                                                return 1;
                                            }
                                        }
                                    }
                                }
                            }
                        }
                        else if (posBalance >= 33) {
                            if (Rs < 943) {
                                if (RsCsDiff < 15) {
                                    if (Cs < 750) {
                                        return 4;
                                    }
                                    else if (Cs >= 750) {
                                        if (tcor < 84) {
                                            return 1;
                                        }
                                        else if (tcor >= 84) {
                                            if (Rs < 858) {
                                                return 8;
                                            }
                                            else if (Rs >= 858) {
                                                if (Cs < 1105) {
                                                    return 4;
                                                }
                                                else if (Cs >= 1105) {
                                                    return 2;
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (RsCsDiff >= 15) {
                                    if (refDCval < 67) {
                                        if (mcTcor < 61) {
                                            if (Rs < 333) {
                                                return 4;
                                            }
                                            else if (Rs >= 333) {
                                                return 1;
                                            }
                                        }
                                        else if (mcTcor >= 61) {
                                            if (Cs < 790) {
                                                return 4;
                                            }
                                            else if (Cs >= 790) {
                                                return 2;
                                            }
                                        }
                                    }
                                    else if (refDCval >= 67) {
                                        if (Rs < 893) {
                                            if (SC < 1888) {
                                                return 8;
                                            }
                                            else if (SC >= 1888) {
                                                return 4;
                                            }
                                        }
                                        else if (Rs >= 893) {
                                            return 1;
                                        }
                                    }
                                }
                            }
                            else if (Rs >= 943) {
                                if (SC < 2179) {
                                    if (posBalance < 38) {
                                        return 4;
                                    }
                                    else if (posBalance >= 38) {
                                        if (Cs < 812) {
                                            return 1;
                                        }
                                        else if (Cs >= 812) {
                                            if (CsDiff < 5) {
                                                return 4;
                                            }
                                            else if (CsDiff >= 5) {
                                                return 1;
                                            }
                                        }
                                    }
                                }
                                else if (SC >= 2179) {
                                    if (Cs < 1072) {
                                        if (refDCval < 52) {
                                            return 1;
                                        }
                                        else if (refDCval >= 52) {
                                            if (ssDCval < 139) {
                                                if (Rs < 1440) {
                                                    return 2;
                                                }
                                                else if (Rs >= 1440) {
                                                    return 8;
                                                }
                                            }
                                            else if (ssDCval >= 139) {
                                                return 1;
                                            }
                                        }
                                    }
                                    else if (Cs >= 1072) {
                                        if (refDCval < 63) {
                                            if (CsDiff < 4) {
                                                return 8;
                                            }
                                            else if (CsDiff >= 4) {
                                                return 4;
                                            }
                                        }
                                        else if (refDCval >= 63) {
                                            if (mu_mv_mag_sq < 38) {
                                                return 1;
                                            }
                                            else if (mu_mv_mag_sq >= 38) {
                                                if (SC < 2503) {
                                                    if (Rs < 1100) {
                                                        return 1;
                                                    }
                                                    else if (Rs >= 1100) {
                                                        return 8;
                                                    }
                                                }
                                                else if (SC >= 2503) {
                                                    return 4;
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                        }
                    }
                    else if (SC >= 2754) {
                        if (mu_mv_mag_sq < 90) {
                            if (Rs < 3095) {
                                if (MVDiff < 9) {
                                    if (ssDCval < 77) {
                                        return 1;
                                    }
                                    else if (ssDCval >= 77) {
                                        return 8;
                                    }
                                }
                                else if (MVDiff >= 9) {
                                    if (AFD < 17) {
                                        if (Cs < 1587) {
                                            return 4;
                                        }
                                        else if (Cs >= 1587) {
                                            if (ssDCval < 72) {
                                                if (TSC < 5) {
                                                    return 8;
                                                }
                                                else if (TSC >= 5) {
                                                    return 1;
                                                }
                                            }
                                            else if (ssDCval >= 72) {
                                                return 8;
                                            }
                                        }
                                    }
                                    else if (AFD >= 17) {
                                        if (diffTSC < -1) {
                                            return 4;
                                        }
                                        else if (diffTSC >= -1) {
                                            if (diffTSC < 1) {
                                                return 8;
                                            }
                                            else if (diffTSC >= 1) {
                                                if (Rs < 2708) {
                                                    return 2;
                                                }
                                                else if (Rs >= 2708) {
                                                    return 8;
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                            else if (Rs >= 3095) {
                                if (CsDiff < 3) {
                                    return 8;
                                }
                                else if (CsDiff >= 3) {
                                    if (diffTSC < 2) {
                                        if (Cs < 2095) {
                                            if (MVDiff < 8) {
                                                return 8;
                                            }
                                            else if (MVDiff >= 8) {
                                                return 4;
                                            }
                                        }
                                        else if (Cs >= 2095) {
                                            if (CsDiff < 7) {
                                                if (Cs < 2368) {
                                                    if (CsDiff < 5) {
                                                        return 8;
                                                    }
                                                    else if (CsDiff >= 5) {
                                                        return 2;
                                                    }
                                                }
                                                else if (Cs >= 2368) {
                                                    if (Cs < 2709) {
                                                        return 8;
                                                    }
                                                    else if (Cs >= 2709) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                            else if (CsDiff >= 7) {
                                                if (Rs < 3622) {
                                                    return 2;
                                                }
                                                else if (Rs >= 3622) {
                                                    if (ssDCval < 128) {
                                                        return 4;
                                                    }
                                                    else if (ssDCval >= 128) {
                                                        return 2;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (diffTSC >= 2) {
                                        if (Cs < 2772) {
                                            return 2;
                                        }
                                        else if (Cs >= 2772) {
                                            return 8;
                                        }
                                    }
                                }
                            }
                        }
                        else if (mu_mv_mag_sq >= 90) {
                            if (Rs < 2679) {
                                if (Cs < 2340) {
                                    if (ssDCval < 130) {
                                        if (AFD < 8) {
                                            return 8;
                                        }
                                        else if (AFD >= 8) {
                                            if (ssDCval < 93) {
                                                if (SC < 4168) {
                                                    return 4;
                                                }
                                                else if (SC >= 4168) {
                                                    return 8;
                                                }
                                            }
                                            else if (ssDCval >= 93) {
                                                return 1;
                                            }
                                        }
                                    }
                                    else if (ssDCval >= 130) {
                                        if (RsCsDiff < 37) {
                                            return 4;
                                        }
                                        else if (RsCsDiff >= 37) {
                                            return 8;
                                        }
                                    }
                                }
                                else if (Cs >= 2340) {
                                    if (Rs < 1819) {
                                        if (Rs < 1745) {
                                            return 8;
                                        }
                                        else if (Rs >= 1745) {
                                            return 2;
                                        }
                                    }
                                    else if (Rs >= 1819) {
                                        return 1;
                                    }
                                }
                            }
                            else if (Rs >= 2679) {
                                if (diffMVdiffVal < -7) {
                                    if (SC < 6210) {
                                        return 8;
                                    }
                                    else if (SC >= 6210) {
                                        return 4;
                                    }
                                }
                                else if (diffMVdiffVal >= -7) {
                                    if (RsDiff < 1) {
                                        if (Cs < 3541) {
                                            return 2;
                                        }
                                        else if (Cs >= 3541) {
                                            return 8;
                                        }
                                    }
                                    else if (RsDiff >= 1) {
                                        return 8;
                                    }
                                }
                            }
                        }
                    }
                }
            }
        }
        else if (negBalance >= 32) {
            if (SC < 8065) {
                if (mu_mv_mag_sq < 108) {
                    if (Cs < 1450) {
                        if (TSC < 8) {
                            if (Cs < 255) {
                                if (RsCsDiff < 3) {
                                    if (Rs < 272) {
                                        if (mcTcor < 100) {
                                            if (diffMVdiffVal < -20) {
                                                return 1;
                                            }
                                            else if (diffMVdiffVal >= -20) {
                                                if (posBalance < 18) {
                                                    if (diffAFD < -1) {
                                                        if (refDCval < 26) {
                                                            if (AFD < 2) {
                                                                if (TSC < 1) {
                                                                    return 2;
                                                                }
                                                                else if (TSC >= 1) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (AFD >= 2) {
                                                                return 1;
                                                            }
                                                        }
                                                        else if (refDCval >= 26) {
                                                            return 2;
                                                        }
                                                    }
                                                    else if (diffAFD >= -1) {
                                                        if (TSCindex < 3) {
                                                            if (tcor < 51) {
                                                                return 8;
                                                            }
                                                            else if (tcor >= 51) {
                                                                if (mu_mv_mag_sq < 7) {
                                                                    if (Rs < 45) {
                                                                        return 2;
                                                                    }
                                                                    else if (Rs >= 45) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (mu_mv_mag_sq >= 7) {
                                                                    if (Rs < 50) {
                                                                        if (TSC < 2) {
                                                                            return 4;
                                                                        }
                                                                        else if (TSC >= 2) {
                                                                            return 2;
                                                                        }
                                                                    }
                                                                    else if (Rs >= 50) {
                                                                        return 2;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (TSCindex >= 3) {
                                                            if (MVDiff < 25) {
                                                                return 4;
                                                            }
                                                            else if (MVDiff >= 25) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (posBalance >= 18) {
                                                    if (gchDC < 2) {
                                                        if (Rs < 178) {
                                                            if (MVDiff < 79) {
                                                                return 2;
                                                            }
                                                            else if (MVDiff >= 79) {
                                                                return 1;
                                                            }
                                                        }
                                                        else if (Rs >= 178) {
                                                            return 1;
                                                        }
                                                    }
                                                    else if (gchDC >= 2) {
                                                        return 1;
                                                    }
                                                }
                                            }
                                        }
                                        else if (mcTcor >= 100) {
                                            if (posBalance < 4) {
                                                if (Cs < 112) {
                                                    return 8;
                                                }
                                                else if (Cs >= 112) {
                                                    return 2;
                                                }
                                            }
                                            else if (posBalance >= 4) {
                                                if (mu_mv_mag_sq < 36) {
                                                    if (tcor < 112) {
                                                        return 1;
                                                    }
                                                    else if (tcor >= 112) {
                                                        return 2;
                                                    }
                                                }
                                                else if (mu_mv_mag_sq >= 36) {
                                                    return 4;
                                                }
                                            }
                                        }
                                    }
                                    else if (Rs >= 272) {
                                        if (TSC < 2) {
                                            return 8;
                                        }
                                        else if (TSC >= 2) {
                                            return 4;
                                        }
                                    }
                                }
                                else if (RsCsDiff >= 3) {
                                    if (TSC < 6) {
                                        if (AFD < 10) {
                                            if (ssDCval < 31) {
                                                return 4;
                                            }
                                            else if (ssDCval >= 31) {
                                                if (Cs < 51) {
                                                    return 2;
                                                }
                                                else if (Cs >= 51) {
                                                    return 1;
                                                }
                                            }
                                        }
                                        else if (AFD >= 10) {
                                            return 2;
                                        }
                                    }
                                    else if (TSC >= 6) {
                                        if (refDCval < 108) {
                                            if (AFD < 13) {
                                                if (tcor < 95) {
                                                    return 1;
                                                }
                                                else if (tcor >= 95) {
                                                    if (refDCval < 105) {
                                                        return 1;
                                                    }
                                                    else if (refDCval >= 105) {
                                                        return 2;
                                                    }
                                                }
                                            }
                                            else if (AFD >= 13) {
                                                if (Cs < 127) {
                                                    return 1;
                                                }
                                                else if (Cs >= 127) {
                                                    return 2;
                                                }
                                            }
                                        }
                                        else if (refDCval >= 108) {
                                            if (AFD < 9) {
                                                return 4;
                                            }
                                            else if (AFD >= 9) {
                                                return 2;
                                            }
                                        }
                                    }
                                }
                            }
                            else if (Cs >= 255) {
                                if (mu_mv_mag_sq < 83) {
                                    if (refDCval < 88) {
                                        if (refDCval < 61) {
                                            if (SC < 1253) {
                                                if (Rs < 431) {
                                                    if (gchDC < 2) {
                                                        if (Cs < 305) {
                                                            if (gchDC < 1) {
                                                                if (Rs < 378) {
                                                                    return 8;
                                                                }
                                                                else if (Rs >= 378) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (gchDC >= 1) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (Cs >= 305) {
                                                            if (Rs < 403) {
                                                                if (MVDiff < 24) {
                                                                    if (Cs < 334) {
                                                                        return 1;
                                                                    }
                                                                    else if (Cs >= 334) {
                                                                        return 8;
                                                                    }
                                                                }
                                                                else if (MVDiff >= 24) {
                                                                    if (SC < 708) {
                                                                        if (Rs < 390) {
                                                                            return 8;
                                                                        }
                                                                        else if (Rs >= 390) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                    else if (SC >= 708) {
                                                                        return 4;
                                                                    }
                                                                }
                                                            }
                                                            else if (Rs >= 403) {
                                                                if (tcor < 85) {
                                                                    if (Cs < 373) {
                                                                        if (tcor < 79) {
                                                                            if (Rs < 421) {
                                                                                return 8;
                                                                            }
                                                                            else if (Rs >= 421) {
                                                                                return 1;
                                                                            }
                                                                        }
                                                                        else if (tcor >= 79) {
                                                                            if (negBalance < 41) {
                                                                                return 4;
                                                                            }
                                                                            else if (negBalance >= 41) {
                                                                                return 1;
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (Cs >= 373) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (tcor >= 85) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (gchDC >= 2) {
                                                        if (TSC < 5) {
                                                            if (Cs < 499) {
                                                                return 2;
                                                            }
                                                            else if (Cs >= 499) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (TSC >= 5) {
                                                            return 2;
                                                        }
                                                    }
                                                }
                                                else if (Rs >= 431) {
                                                    if (mcTcor < 96) {
                                                        if (RsCsDiff < 4) {
                                                            if (Cs < 382) {
                                                                if (refDCval < 38) {
                                                                    if (Cs < 374) {
                                                                        return 1;
                                                                    }
                                                                    else if (Cs >= 374) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (refDCval >= 38) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (Cs >= 382) {
                                                                if (posBalance < 54) {
                                                                    return 4;
                                                                }
                                                                else if (posBalance >= 54) {
                                                                    if (ssDCval < 39) {
                                                                        if (Cs < 475) {
                                                                            if (MVDiff < 16) {
                                                                                return 4;
                                                                            }
                                                                            else if (MVDiff >= 16) {
                                                                                if (posBalance < 58) {
                                                                                    if (SC < 829) {
                                                                                        if (MVDiff < 28) {
                                                                                            return 4;
                                                                                        }
                                                                                        else if (MVDiff >= 28) {
                                                                                            return 2;
                                                                                        }
                                                                                    }
                                                                                    else if (SC >= 829) {
                                                                                        return 2;
                                                                                    }
                                                                                }
                                                                                else if (posBalance >= 58) {
                                                                                    return 8;
                                                                                }
                                                                            }
                                                                        }
                                                                        else if (Cs >= 475) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                    else if (ssDCval >= 39) {
                                                                        return 4;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (RsCsDiff >= 4) {
                                                            if (SC < 1110) {
                                                                if (TSC < 4) {
                                                                    return 4;
                                                                }
                                                                else if (TSC >= 4) {
                                                                    if (SC < 865) {
                                                                        return 8;
                                                                    }
                                                                    else if (SC >= 865) {
                                                                        return 2;
                                                                    }
                                                                }
                                                            }
                                                            else if (SC >= 1110) {
                                                                if (gchDC < 1) {
                                                                    if (mcTcor < 74) {
                                                                        return 4;
                                                                    }
                                                                    else if (mcTcor >= 74) {
                                                                        if (AFD < 13) {
                                                                            return 1;
                                                                        }
                                                                        else if (AFD >= 13) {
                                                                            return 2;
                                                                        }
                                                                    }
                                                                }
                                                                else if (gchDC >= 1) {
                                                                    if (Rs < 671) {
                                                                        if (tcor < 69) {
                                                                            return 4;
                                                                        }
                                                                        else if (tcor >= 69) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                    else if (Rs >= 671) {
                                                                        if (Cs < 482) {
                                                                            return 2;
                                                                        }
                                                                        else if (Cs >= 482) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (mcTcor >= 96) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                            else if (SC >= 1253) {
                                                if (refDCval < 39) {
                                                    if (Cs < 625) {
                                                        if (CsDiff < 2) {
                                                            if (MVDiff < 35) {
                                                                if (diffAFD < -1) {
                                                                    return 1;
                                                                }
                                                                else if (diffAFD >= -1) {
                                                                    if (Cs < 567) {
                                                                        if (Rs < 757) {
                                                                            return 4;
                                                                        }
                                                                        else if (Rs >= 757) {
                                                                            if (mcTcor < 82) {
                                                                                if (AFD < 7) {
                                                                                    return 4;
                                                                                }
                                                                                else if (AFD >= 7) {
                                                                                    return 1;
                                                                                }
                                                                            }
                                                                            else if (mcTcor >= 82) {
                                                                                return 1;
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (Cs >= 567) {
                                                                        return 4;
                                                                    }
                                                                }
                                                            }
                                                            else if (MVDiff >= 35) {
                                                                return 1;
                                                            }
                                                        }
                                                        else if (CsDiff >= 2) {
                                                            if (mu_mv_mag_sq < 59) {
                                                                if (MVDiff < 34) {
                                                                    return 1;
                                                                }
                                                                else if (MVDiff >= 34) {
                                                                    if (SC < 1372) {
                                                                        return 8;
                                                                    }
                                                                    else if (SC >= 1372) {
                                                                        return 4;
                                                                    }
                                                                }
                                                            }
                                                            else if (mu_mv_mag_sq >= 59) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                    else if (Cs >= 625) {
                                                        if (TSC < 6) {
                                                            if (MVDiff < 15) {
                                                                return 8;
                                                            }
                                                            else if (MVDiff >= 15) {
                                                                return 2;
                                                            }
                                                        }
                                                        else if (TSC >= 6) {
                                                            return 1;
                                                        }
                                                    }
                                                }
                                                else if (refDCval >= 39) {
                                                    if (AFD < 10) {
                                                        if (Rs < 1177) {
                                                            if (mu_mv_mag_sq < 42) {
                                                                if (mu_mv_mag_sq < 28) {
                                                                    if (tcor < 90) {
                                                                        if (RsCsDiff < 16) {
                                                                            return 4;
                                                                        }
                                                                        else if (RsCsDiff >= 16) {
                                                                            if (Rs < 924) {
                                                                                return 4;
                                                                            }
                                                                            else if (Rs >= 924) {
                                                                                return 1;
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (tcor >= 90) {
                                                                        if (diffRsCsdiff < -3) {
                                                                            return 4;
                                                                        }
                                                                        else if (diffRsCsdiff >= -3) {
                                                                            if (MVDiff < 1) {
                                                                                return 4;
                                                                            }
                                                                            else if (MVDiff >= 1) {
                                                                                return 8;
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                                else if (mu_mv_mag_sq >= 28) {
                                                                    if (AFD < 5) {
                                                                        if (mcTcor < 96) {
                                                                            return 4;
                                                                        }
                                                                        else if (mcTcor >= 96) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                    else if (AFD >= 5) {
                                                                        if (negBalance < 42) {
                                                                            if (Cs < 803) {
                                                                                if (Rs < 807) {
                                                                                    return 4;
                                                                                }
                                                                                else if (Rs >= 807) {
                                                                                    return 8;
                                                                                }
                                                                            }
                                                                            else if (Cs >= 803) {
                                                                                if (AFD < 8) {
                                                                                    return 1;
                                                                                }
                                                                                else if (AFD >= 8) {
                                                                                    return 4;
                                                                                }
                                                                            }
                                                                        }
                                                                        else if (negBalance >= 42) {
                                                                            if (tcor < 76) {
                                                                                if (mcTcor < 75) {
                                                                                    return 1;
                                                                                }
                                                                                else if (mcTcor >= 75) {
                                                                                    if (mu_mv_mag_sq < 34) {
                                                                                        return 4;
                                                                                    }
                                                                                    else if (mu_mv_mag_sq >= 34) {
                                                                                        if (RsDiff < 2) {
                                                                                            return 4;
                                                                                        }
                                                                                        else if (RsDiff >= 2) {
                                                                                            if (diffMVdiffVal < -4) {
                                                                                                return 4;
                                                                                            }
                                                                                            else if (diffMVdiffVal >= -4) {
                                                                                                return 1;
                                                                                            }
                                                                                        }
                                                                                    }
                                                                                }
                                                                            }
                                                                            else if (tcor >= 76) {
                                                                                if (negBalance < 51) {
                                                                                    return 1;
                                                                                }
                                                                                else if (negBalance >= 51) {
                                                                                    if (Cs < 603) {
                                                                                        return 1;
                                                                                    }
                                                                                    else if (Cs >= 603) {
                                                                                        return 4;
                                                                                    }
                                                                                }
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                            else if (mu_mv_mag_sq >= 42) {
                                                                if (posBalance < 43) {
                                                                    if (SC < 1842) {
                                                                        if (gchDC < 4) {
                                                                            return 8;
                                                                        }
                                                                        else if (gchDC >= 4) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                    else if (SC >= 1842) {
                                                                        return 2;
                                                                    }
                                                                }
                                                                else if (posBalance >= 43) {
                                                                    if (Cs < 547) {
                                                                        return 8;
                                                                    }
                                                                    else if (Cs >= 547) {
                                                                        if (refDCval < 57) {
                                                                            if (TSC < 5) {
                                                                                if (ssDCval < 46) {
                                                                                    return 1;
                                                                                }
                                                                                else if (ssDCval >= 46) {
                                                                                    return 4;
                                                                                }
                                                                            }
                                                                            else if (TSC >= 5) {
                                                                                if (gchDC < 2) {
                                                                                    if (Rs < 864) {
                                                                                        return 4;
                                                                                    }
                                                                                    else if (Rs >= 864) {
                                                                                        if (SC < 1733) {
                                                                                            return 1;
                                                                                        }
                                                                                        else if (SC >= 1733) {
                                                                                            if (ssDCval < 50) {
                                                                                                return 1;
                                                                                            }
                                                                                            else if (ssDCval >= 50) {
                                                                                                return 4;
                                                                                            }
                                                                                        }
                                                                                    }
                                                                                }
                                                                                else if (gchDC >= 2) {
                                                                                    if (refDCval < 52) {
                                                                                        return 8;
                                                                                    }
                                                                                    else if (refDCval >= 52) {
                                                                                        return 4;
                                                                                    }
                                                                                }
                                                                            }
                                                                        }
                                                                        else if (refDCval >= 57) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (Rs >= 1177) {
                                                            if (RsDiff < 4) {
                                                                if (SC < 2310) {
                                                                    if (Rs < 1202) {
                                                                        if (SC < 2176) {
                                                                            return 4;
                                                                        }
                                                                        else if (SC >= 2176) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                    else if (Rs >= 1202) {
                                                                        if (posBalance < 50) {
                                                                            if (negBalance < 39) {
                                                                                if (diffRsCsdiff < -6) {
                                                                                    if (AFD < 5) {
                                                                                        return 1;
                                                                                    }
                                                                                    else if (AFD >= 5) {
                                                                                        return 4;
                                                                                    }
                                                                                }
                                                                                else if (diffRsCsdiff >= -6) {
                                                                                    return 1;
                                                                                }
                                                                            }
                                                                            else if (negBalance >= 39) {
                                                                                return 1;
                                                                            }
                                                                        }
                                                                        else if (posBalance >= 50) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                }
                                                                else if (SC >= 2310) {
                                                                    if (posBalance < 52) {
                                                                        if (SC < 2555) {
                                                                            if (SC < 2346) {
                                                                                if (Rs < 1333) {
                                                                                    return 1;
                                                                                }
                                                                                else if (Rs >= 1333) {
                                                                                    return 4;
                                                                                }
                                                                            }
                                                                            else if (SC >= 2346) {
                                                                                if (posBalance < 51) {
                                                                                    return 1;
                                                                                }
                                                                                else if (posBalance >= 51) {
                                                                                    if (AFD < 6) {
                                                                                        return 1;
                                                                                    }
                                                                                    else if (AFD >= 6) {
                                                                                        return 8;
                                                                                    }
                                                                                }
                                                                            }
                                                                        }
                                                                        else if (SC >= 2555) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                    else if (posBalance >= 52) {
                                                                        if (ssDCval < 57) {
                                                                            if (diffMVdiffVal < -4) {
                                                                                return 1;
                                                                            }
                                                                            else if (diffMVdiffVal >= -4) {
                                                                                if (Rs < 1433) {
                                                                                    if (AFD < 8) {
                                                                                        if (posBalance < 56) {
                                                                                            if (MVDiff < 18) {
                                                                                                return 1;
                                                                                            }
                                                                                            else if (MVDiff >= 18) {
                                                                                                return 8;
                                                                                            }
                                                                                        }
                                                                                        else if (posBalance >= 56) {
                                                                                            return 8;
                                                                                        }
                                                                                    }
                                                                                    else if (AFD >= 8) {
                                                                                        return 8;
                                                                                    }
                                                                                }
                                                                                else if (Rs >= 1433) {
                                                                                    return 8;
                                                                                }
                                                                            }
                                                                        }
                                                                        else if (ssDCval >= 57) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                            else if (RsDiff >= 4) {
                                                                if (posBalance < 47) {
                                                                    return 4;
                                                                }
                                                                else if (posBalance >= 47) {
                                                                    if (negBalance < 40) {
                                                                        if (refDCval < 52) {
                                                                            if (SC < 2964) {
                                                                                if (TSC < 6) {
                                                                                    if (CsDiff < 3) {
                                                                                        return 4;
                                                                                    }
                                                                                    else if (CsDiff >= 3) {
                                                                                        return 1;
                                                                                    }
                                                                                }
                                                                                else if (TSC >= 6) {
                                                                                    return 4;
                                                                                }
                                                                            }
                                                                            else if (SC >= 2964) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                        else if (refDCval >= 52) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                    else if (negBalance >= 40) {
                                                                        if (refDCval < 52) {
                                                                            if (AFD < 6) {
                                                                                return 8;
                                                                            }
                                                                            else if (AFD >= 6) {
                                                                                return 1;
                                                                            }
                                                                        }
                                                                        else if (refDCval >= 52) {
                                                                            if (MVDiff < 8) {
                                                                                return 8;
                                                                            }
                                                                            else if (MVDiff >= 8) {
                                                                                if (ssDCval < 55) {
                                                                                    return 1;
                                                                                }
                                                                                else if (ssDCval >= 55) {
                                                                                    if (Rs < 1542) {
                                                                                        return 8;
                                                                                    }
                                                                                    else if (Rs >= 1542) {
                                                                                        return 4;
                                                                                    }
                                                                                }
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (AFD >= 10) {
                                                        if (Rs < 977) {
                                                            if (refDCval < 50) {
                                                                return 8;
                                                            }
                                                            else if (refDCval >= 50) {
                                                                return 2;
                                                            }
                                                        }
                                                        else if (Rs >= 977) {
                                                            if (SC < 2419) {
                                                                if (TSC < 5) {
                                                                    return 4;
                                                                }
                                                                else if (TSC >= 5) {
                                                                    if (ssDCval < 49) {
                                                                        return 8;
                                                                    }
                                                                    else if (ssDCval >= 49) {
                                                                        return 1;
                                                                    }
                                                                }
                                                            }
                                                            else if (SC >= 2419) {
                                                                if (Cs < 1228) {
                                                                    if (RsDiff < 6) {
                                                                        return 8;
                                                                    }
                                                                    else if (RsDiff >= 6) {
                                                                        if (Rs < 1522) {
                                                                            return 1;
                                                                        }
                                                                        else if (Rs >= 1522) {
                                                                            return 2;
                                                                        }
                                                                    }
                                                                }
                                                                else if (Cs >= 1228) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (refDCval >= 61) {
                                            if (Rs < 1411) {
                                                if (mcTcor < 99) {
                                                    if (negBalance < 49) {
                                                        if (SC < 1918) {
                                                            if (Rs < 756) {
                                                                if (Rs < 566) {
                                                                    if (Rs < 502) {
                                                                        if (MVDiff < 15) {
                                                                            return 4;
                                                                        }
                                                                        else if (MVDiff >= 15) {
                                                                            return 2;
                                                                        }
                                                                    }
                                                                    else if (Rs >= 502) {
                                                                        return 1;
                                                                    }
                                                                }
                                                                else if (Rs >= 566) {
                                                                    if (posBalance < 48) {
                                                                        if (TSC < 5) {
                                                                            return 2;
                                                                        }
                                                                        else if (TSC >= 5) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                    else if (posBalance >= 48) {
                                                                        if (ssDCval < 69) {
                                                                            if (AFD < 6) {
                                                                                return 8;
                                                                            }
                                                                            else if (AFD >= 6) {
                                                                                return 2;
                                                                            }
                                                                        }
                                                                        else if (ssDCval >= 69) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                            else if (Rs >= 756) {
                                                                return 2;
                                                            }
                                                        }
                                                        else if (SC >= 1918) {
                                                            if (MVDiff < 14) {
                                                                return 2;
                                                            }
                                                            else if (MVDiff >= 14) {
                                                                return 1;
                                                            }
                                                        }
                                                    }
                                                    else if (negBalance >= 49) {
                                                        if (negBalance < 60) {
                                                            if (ssDCval < 59) {
                                                                return 1;
                                                            }
                                                            else if (ssDCval >= 59) {
                                                                if (diffAFD < -3) {
                                                                    return 4;
                                                                }
                                                                else if (diffAFD >= -3) {
                                                                    if (diffAFD < 2) {
                                                                        return 2;
                                                                    }
                                                                    else if (diffAFD >= 2) {
                                                                        if (Rs < 900) {
                                                                            return 2;
                                                                        }
                                                                        else if (Rs >= 900) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (negBalance >= 60) {
                                                            if (Cs < 785) {
                                                                return 8;
                                                            }
                                                            else if (Cs >= 785) {
                                                                return 2;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (mcTcor >= 99) {
                                                    if (posBalance < 15) {
                                                        return 8;
                                                    }
                                                    else if (posBalance >= 15) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                            else if (Rs >= 1411) {
                                                if (posBalance < 49) {
                                                    if (MVDiff < 23) {
                                                        if (posBalance < 38) {
                                                            if (Rs < 2431) {
                                                                return 2;
                                                            }
                                                            else if (Rs >= 2431) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (posBalance >= 38) {
                                                            if (Rs < 3317) {
                                                                return 2;
                                                            }
                                                            else if (Rs >= 3317) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                    else if (MVDiff >= 23) {
                                                        return 4;
                                                    }
                                                }
                                                else if (posBalance >= 49) {
                                                    if (ssDCval < 79) {
                                                        if (gchDC < 2) {
                                                            if (ssDCval < 73) {
                                                                return 1;
                                                            }
                                                            else if (ssDCval >= 73) {
                                                                if (ssDCval < 74) {
                                                                    if (Rs < 1863) {
                                                                        return 4;
                                                                    }
                                                                    else if (Rs >= 1863) {
                                                                        return 1;
                                                                    }
                                                                }
                                                                else if (ssDCval >= 74) {
                                                                    return 1;
                                                                }
                                                            }
                                                        }
                                                        else if (gchDC >= 2) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (ssDCval >= 79) {
                                                        if (Rs < 1607) {
                                                            return 1;
                                                        }
                                                        else if (Rs >= 1607) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (refDCval >= 88) {
                                        if (RsDiff < 3) {
                                            if (SC < 704) {
                                                if (MVDiff < 4) {
                                                    return 2;
                                                }
                                                else if (MVDiff >= 4) {
                                                    return 4;
                                                }
                                            }
                                            else if (SC >= 704) {
                                                if (posBalance < 64) {
                                                    if (MVDiff < 23) {
                                                        if (Rs < 623) {
                                                            if (Cs < 379) {
                                                                return 4;
                                                            }
                                                            else if (Cs >= 379) {
                                                                if (diffAFD < -1) {
                                                                    if (SC < 955) {
                                                                        return 4;
                                                                    }
                                                                    else if (SC >= 955) {
                                                                        return 8;
                                                                    }
                                                                }
                                                                else if (diffAFD >= -1) {
                                                                    if (negBalance < 44) {
                                                                        return 4;
                                                                    }
                                                                    else if (negBalance >= 44) {
                                                                        return 8;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (Rs >= 623) {
                                                            if (tcor < 90) {
                                                                if (diffMVdiffVal < 3) {
                                                                    return 4;
                                                                }
                                                                else if (diffMVdiffVal >= 3) {
                                                                    if (Cs < 568) {
                                                                        return 4;
                                                                    }
                                                                    else if (Cs >= 568) {
                                                                        if (Cs < 620) {
                                                                            return 8;
                                                                        }
                                                                        else if (Cs >= 620) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                            else if (tcor >= 90) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                    else if (MVDiff >= 23) {
                                                        if (CsDiff < 2) {
                                                            return 8;
                                                        }
                                                        else if (CsDiff >= 2) {
                                                            if (Cs < 436) {
                                                                return 2;
                                                            }
                                                            else if (Cs >= 436) {
                                                                if (Rs < 647) {
                                                                    return 4;
                                                                }
                                                                else if (Rs >= 647) {
                                                                    return 2;
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (posBalance >= 64) {
                                                    if (TSC < 6) {
                                                        return 4;
                                                    }
                                                    else if (TSC >= 6) {
                                                        return 1;
                                                    }
                                                }
                                            }
                                        }
                                        else if (RsDiff >= 3) {
                                            if (negBalance < 50) {
                                                if (Cs < 1043) {
                                                    if (mu_mv_mag_sq < 15) {
                                                        if (Rs < 2030) {
                                                            if (negBalance < 36) {
                                                                return 4;
                                                            }
                                                            else if (negBalance >= 36) {
                                                                return 1;
                                                            }
                                                        }
                                                        else if (Rs >= 2030) {
                                                            return 2;
                                                        }
                                                    }
                                                    else if (mu_mv_mag_sq >= 15) {
                                                        if (CsDiff < 3) {
                                                            return 8;
                                                        }
                                                        else if (CsDiff >= 3) {
                                                            if (tcor < 79) {
                                                                if (Cs < 756) {
                                                                    return 8;
                                                                }
                                                                else if (Cs >= 756) {
                                                                    return 2;
                                                                }
                                                            }
                                                            else if (tcor >= 79) {
                                                                return 2;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (Cs >= 1043) {
                                                    if (AFD < 12) {
                                                        if (refDCval < 98) {
                                                            return 2;
                                                        }
                                                        else if (refDCval >= 98) {
                                                            if (TSC < 6) {
                                                                return 4;
                                                            }
                                                            else if (TSC >= 6) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                    else if (AFD >= 12) {
                                                        if (RsDiff < 8) {
                                                            if (Rs < 1067) {
                                                                return 4;
                                                            }
                                                            else if (Rs >= 1067) {
                                                                return 1;
                                                            }
                                                        }
                                                        else if (RsDiff >= 8) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (negBalance >= 50) {
                                                if (posBalance < 59) {
                                                    if (MVDiff < 10) {
                                                        if (TSCindex < 7) {
                                                            return 8;
                                                        }
                                                        else if (TSCindex >= 7) {
                                                            if (AFD < 10) {
                                                                if (diffTSC < 1) {
                                                                    if (AFD < 8) {
                                                                        return 4;
                                                                    }
                                                                    else if (AFD >= 8) {
                                                                        return 8;
                                                                    }
                                                                }
                                                                else if (diffTSC >= 1) {
                                                                    return 2;
                                                                }
                                                            }
                                                            else if (AFD >= 10) {
                                                                if (refDCval < 98) {
                                                                    if (Rs < 1281) {
                                                                        return 4;
                                                                    }
                                                                    else if (Rs >= 1281) {
                                                                        return 8;
                                                                    }
                                                                }
                                                                else if (refDCval >= 98) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (MVDiff >= 10) {
                                                        if (refDCval < 116) {
                                                            if (Rs < 1174) {
                                                                return 4;
                                                            }
                                                            else if (Rs >= 1174) {
                                                                if (MVDiff < 13) {
                                                                    return 2;
                                                                }
                                                                else if (MVDiff >= 13) {
                                                                    if (Rs < 1306) {
                                                                        return 2;
                                                                    }
                                                                    else if (Rs >= 1306) {
                                                                        return 8;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (refDCval >= 116) {
                                                            if (posBalance < 32) {
                                                                return 8;
                                                            }
                                                            else if (posBalance >= 32) {
                                                                return 2;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (posBalance >= 59) {
                                                    if (ssDCval < 130) {
                                                        if (Cs < 408) {
                                                            return 2;
                                                        }
                                                        else if (Cs >= 408) {
                                                            if (diffAFD < -1) {
                                                                if (MVDiff < 15) {
                                                                    return 4;
                                                                }
                                                                else if (MVDiff >= 15) {
                                                                    return 2;
                                                                }
                                                            }
                                                            else if (diffAFD >= -1) {
                                                                if (AFD < 9) {
                                                                    if (Rs < 1221) {
                                                                        return 4;
                                                                    }
                                                                    else if (Rs >= 1221) {
                                                                        return 2;
                                                                    }
                                                                }
                                                                else if (AFD >= 9) {
                                                                    if (Rs < 644) {
                                                                        if (SC < 1290) {
                                                                            return 8;
                                                                        }
                                                                        else if (SC >= 1290) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                    else if (Rs >= 644) {
                                                                        return 4;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (ssDCval >= 130) {
                                                        return 2;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (mu_mv_mag_sq >= 83) {
                                    if (TSC < 6) {
                                        if (Rs < 864) {
                                            if (refDCval < 119) {
                                                if (ssDCval < 59) {
                                                    if (Rs < 806) {
                                                        if (mu_mv_mag_sq < 85) {
                                                            if (SC < 1043) {
                                                                return 8;
                                                            }
                                                            else if (SC >= 1043) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (mu_mv_mag_sq >= 85) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (Rs >= 806) {
                                                        return 8;
                                                    }
                                                }
                                                else if (ssDCval >= 59) {
                                                    if (MVDiff < 1) {
                                                        return 4;
                                                    }
                                                    else if (MVDiff >= 1) {
                                                        if (ssDCval < 83) {
                                                            if (Rs < 483) {
                                                                return 4;
                                                            }
                                                            else if (Rs >= 483) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (ssDCval >= 83) {
                                                            if (Rs < 287) {
                                                                return 2;
                                                            }
                                                            else if (Rs >= 287) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                            else if (refDCval >= 119) {
                                                if (RsCsDiff < 2) {
                                                    return 2;
                                                }
                                                else if (RsCsDiff >= 2) {
                                                    return 4;
                                                }
                                            }
                                        }
                                        else if (Rs >= 864) {
                                            if (ssDCval < 73) {
                                                if (MVDiff < 17) {
                                                    if (diffTSC < -1) {
                                                        if (Cs < 735) {
                                                            if (Rs < 1151) {
                                                                return 8;
                                                            }
                                                            else if (Rs >= 1151) {
                                                                return 1;
                                                            }
                                                        }
                                                        else if (Cs >= 735) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (diffTSC >= -1) {
                                                        return 8;
                                                    }
                                                }
                                                else if (MVDiff >= 17) {
                                                    return 4;
                                                }
                                            }
                                            else if (ssDCval >= 73) {
                                                if (RsDiff < 3) {
                                                    return 4;
                                                }
                                                else if (RsDiff >= 3) {
                                                    return 1;
                                                }
                                            }
                                        }
                                    }
                                    else if (TSC >= 6) {
                                        if (tcor < 86) {
                                            if (SC < 2486) {
                                                if (SC < 2230) {
                                                    return 2;
                                                }
                                                else if (SC >= 2230) {
                                                    return 4;
                                                }
                                            }
                                            else if (SC >= 2486) {
                                                return 8;
                                            }
                                        }
                                        else if (tcor >= 86) {
                                            if (ssDCval < 65) {
                                                if (Rs < 1270) {
                                                    return 8;
                                                }
                                                else if (Rs >= 1270) {
                                                    return 4;
                                                }
                                            }
                                            else if (ssDCval >= 65) {
                                                return 2;
                                            }
                                        }
                                    }
                                }
                            }
                        }
                        else if (TSC >= 8) {
                            if (Rs < 1448) {
                                if (AFD < 10) {
                                    if (Rs < 1404) {
                                        if (refDCval < 91) {
                                            if (Rs < 1247) {
                                                return 4;
                                            }
                                            else if (Rs >= 1247) {
                                                return 2;
                                            }
                                        }
                                        else if (refDCval >= 91) {
                                            return 8;
                                        }
                                    }
                                    else if (Rs >= 1404) {
                                        if (CsDiff < 5) {
                                            return 2;
                                        }
                                        else if (CsDiff >= 5) {
                                            return 8;
                                        }
                                    }
                                }
                                else if (AFD >= 10) {
                                    if (Rs < 808) {
                                        if (diffTSC < 4) {
                                            if (refDCval < 131) {
                                                if (mcTcor < 102) {
                                                    if (mu_mv_mag_sq < 72) {
                                                        return 1;
                                                    }
                                                    else if (mu_mv_mag_sq >= 72) {
                                                        if (AFD < 27) {
                                                            return 2;
                                                        }
                                                        else if (AFD >= 27) {
                                                            return 1;
                                                        }
                                                    }
                                                }
                                                else if (mcTcor >= 102) {
                                                    if (AFD < 24) {
                                                        return 2;
                                                    }
                                                    else if (AFD >= 24) {
                                                        return 1;
                                                    }
                                                }
                                            }
                                            else if (refDCval >= 131) {
                                                if (AFD < 19) {
                                                    return 2;
                                                }
                                                else if (AFD >= 19) {
                                                    return 1;
                                                }
                                            }
                                        }
                                        else if (diffTSC >= 4) {
                                            if (AFD < 35) {
                                                return 8;
                                            }
                                            else if (AFD >= 35) {
                                                if (gchDC < 14) {
                                                    if (TSC < 21) {
                                                        return 2;
                                                    }
                                                    else if (TSC >= 21) {
                                                        return 8;
                                                    }
                                                }
                                                else if (gchDC >= 14) {
                                                    if (AFD < 45) {
                                                        if (Cs < 1076) {
                                                            return 4;
                                                        }
                                                        else if (Cs >= 1076) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (AFD >= 45) {
                                                        return 1;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (Rs >= 808) {
                                        if (Cs < 760) {
                                            if (ssDCval < 74) {
                                                if (Rs < 1153) {
                                                    return 8;
                                                }
                                                else if (Rs >= 1153) {
                                                    return 2;
                                                }
                                            }
                                            else if (ssDCval >= 74) {
                                                return 2;
                                            }
                                        }
                                        else if (Cs >= 760) {
                                            if (ssDCval < 74) {
                                                if (diffMVdiffVal < -2) {
                                                    return 8;
                                                }
                                                else if (diffMVdiffVal >= -2) {
                                                    if (refDCval < 57) {
                                                        if (SC < 1823) {
                                                            return 4;
                                                        }
                                                        else if (SC >= 1823) {
                                                            return 1;
                                                        }
                                                    }
                                                    else if (refDCval >= 57) {
                                                        if (Cs < 1050) {
                                                            if (MVDiff < 11) {
                                                                return 4;
                                                            }
                                                            else if (MVDiff >= 11) {
                                                                if (gchDC < 1) {
                                                                    if (Rs < 988) {
                                                                        return 8;
                                                                    }
                                                                    else if (Rs >= 988) {
                                                                        if (Cs < 947) {
                                                                            return 4;
                                                                        }
                                                                        else if (Cs >= 947) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                }
                                                                else if (gchDC >= 1) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                        else if (Cs >= 1050) {
                                                            if (Rs < 1174) {
                                                                return 2;
                                                            }
                                                            else if (Rs >= 1174) {
                                                                if (AFD < 27) {
                                                                    return 8;
                                                                }
                                                                else if (AFD >= 27) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                            else if (ssDCval >= 74) {
                                                if (refDCval < 108) {
                                                    if (AFD < 40) {
                                                        if (Rs < 1046) {
                                                            return 1;
                                                        }
                                                        else if (Rs >= 1046) {
                                                            return 2;
                                                        }
                                                    }
                                                    else if (AFD >= 40) {
                                                        if (TSC < 32) {
                                                            return 4;
                                                        }
                                                        else if (TSC >= 32) {
                                                            return 1;
                                                        }
                                                    }
                                                }
                                                else if (refDCval >= 108) {
                                                    if (AFD < 19) {
                                                        return 8;
                                                    }
                                                    else if (AFD >= 19) {
                                                        return 1;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                            else if (Rs >= 1448) {
                                if (negBalance < 53) {
                                    if (AFD < 12) {
                                        if (mu_mv_mag_sq < 67) {
                                            if (diffMVdiffVal < -1) {
                                                return 8;
                                            }
                                            else if (diffMVdiffVal >= -1) {
                                                return 4;
                                            }
                                        }
                                        else if (mu_mv_mag_sq >= 67) {
                                            return 8;
                                        }
                                    }
                                    else if (AFD >= 12) {
                                        if (refDCval < 71) {
                                            if (Rs < 1849) {
                                                return 8;
                                            }
                                            else if (Rs >= 1849) {
                                                return 4;
                                            }
                                        }
                                        else if (refDCval >= 71) {
                                            if (mcTcor < 85) {
                                                if (MVDiff < 22) {
                                                    return 1;
                                                }
                                                else if (MVDiff >= 22) {
                                                    if (SC < 2981) {
                                                        return 2;
                                                    }
                                                    else if (SC >= 2981) {
                                                        return 1;
                                                    }
                                                }
                                            }
                                            else if (mcTcor >= 85) {
                                                if (diffRsCsdiff < -62) {
                                                    if (posBalance < 60) {
                                                        return 4;
                                                    }
                                                    else if (posBalance >= 60) {
                                                        return 2;
                                                    }
                                                }
                                                else if (diffRsCsdiff >= -62) {
                                                    if (MVDiff < 10) {
                                                        return 1;
                                                    }
                                                    else if (MVDiff >= 10) {
                                                        if (tcor < 91) {
                                                            if (diffMVdiffVal < 26) {
                                                                return 2;
                                                            }
                                                            else if (diffMVdiffVal >= 26) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (tcor >= 91) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (negBalance >= 53) {
                                    if (tcor < 70) {
                                        if (refDCval < 84) {
                                            if (negBalance < 61) {
                                                return 8;
                                            }
                                            else if (negBalance >= 61) {
                                                if (negBalance < 89) {
                                                    return 4;
                                                }
                                                else if (negBalance >= 89) {
                                                    return 2;
                                                }
                                            }
                                        }
                                        else if (refDCval >= 84) {
                                            if (AFD < 31) {
                                                if (ssDCval < 94) {
                                                    return 2;
                                                }
                                                else if (ssDCval >= 94) {
                                                    if (refDCval < 103) {
                                                        if (SC < 4064) {
                                                            return 1;
                                                        }
                                                        else if (SC >= 4064) {
                                                            if (MVDiff < 80) {
                                                                return 1;
                                                            }
                                                            else if (MVDiff >= 80) {
                                                                if (posBalance < 62) {
                                                                    return 2;
                                                                }
                                                                else if (posBalance >= 62) {
                                                                    return 1;
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (refDCval >= 103) {
                                                        if (tcor < 54) {
                                                            if (mcTcor < 77) {
                                                                return 2;
                                                            }
                                                            else if (mcTcor >= 77) {
                                                                return 1;
                                                            }
                                                        }
                                                        else if (tcor >= 54) {
                                                            return 2;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (AFD >= 31) {
                                                if (CsDiff < 11) {
                                                    return 2;
                                                }
                                                else if (CsDiff >= 11) {
                                                    if (Cs < 592) {
                                                        return 4;
                                                    }
                                                    else if (Cs >= 592) {
                                                        return 1;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (tcor >= 70) {
                                        if (refDCval < 76) {
                                            if (Cs < 1076) {
                                                return 1;
                                            }
                                            else if (Cs >= 1076) {
                                                return 4;
                                            }
                                        }
                                        else if (refDCval >= 76) {
                                            if (MVDiff < 44) {
                                                if (diffAFD < 3) {
                                                    if (ssDCval < 117) {
                                                        return 2;
                                                    }
                                                    else if (ssDCval >= 117) {
                                                        if (Rs < 2890) {
                                                            if (RsCsDiff < 108) {
                                                                return 4;
                                                            }
                                                            else if (RsCsDiff >= 108) {
                                                                if (Cs < 510) {
                                                                    return 4;
                                                                }
                                                                else if (Cs >= 510) {
                                                                    return 2;
                                                                }
                                                            }
                                                        }
                                                        else if (Rs >= 2890) {
                                                            return 2;
                                                        }
                                                    }
                                                }
                                                else if (diffAFD >= 3) {
                                                    if (Cs < 1351) {
                                                        return 1;
                                                    }
                                                    else if (Cs >= 1351) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                            else if (MVDiff >= 44) {
                                                if (Rs < 2356) {
                                                    if (Cs < 856) {
                                                        return 4;
                                                    }
                                                    else if (Cs >= 856) {
                                                        return 2;
                                                    }
                                                }
                                                else if (Rs >= 2356) {
                                                    return 4;
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                        }
                    }
                    else if (Cs >= 1450) {
                        if (TSC < 8) {
                            if (TSC < 7) {
                                if (Rs < 3029) {
                                    if (MVDiff < 1) {
                                        if (negBalance < 44) {
                                            return 4;
                                        }
                                        else if (negBalance >= 44) {
                                            if (Rs < 1970) {
                                                if (Rs < 1446) {
                                                    return 8;
                                                }
                                                else if (Rs >= 1446) {
                                                    if (Cs < 1917) {
                                                        return 2;
                                                    }
                                                    else if (Cs >= 1917) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                            else if (Rs >= 1970) {
                                                return 4;
                                            }
                                        }
                                    }
                                    else if (MVDiff >= 1) {
                                        if (mu_mv_mag_sq < 99) {
                                            if (posBalance < 39) {
                                                if (Cs < 1660) {
                                                    if (ssDCval < 61) {
                                                        return 8;
                                                    }
                                                    else if (ssDCval >= 61) {
                                                        return 4;
                                                    }
                                                }
                                                else if (Cs >= 1660) {
                                                    if (Cs < 2359) {
                                                        if (mu_mv_mag_sq < 18) {
                                                            return 2;
                                                        }
                                                        else if (mu_mv_mag_sq >= 18) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (Cs >= 2359) {
                                                        return 2;
                                                    }
                                                }
                                            }
                                            else if (posBalance >= 39) {
                                                if (CsDiff < 13) {
                                                    if (refDCval < 91) {
                                                        if (negBalance < 40) {
                                                            if (ssDCval < 59) {
                                                                if (Rs < 2222) {
                                                                    return 4;
                                                                }
                                                                else if (Rs >= 2222) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (ssDCval >= 59) {
                                                                return 1;
                                                            }
                                                        }
                                                        else if (negBalance >= 40) {
                                                            if (Cs < 1537) {
                                                                if (Rs < 2388) {
                                                                    if (MVDiff < 19) {
                                                                        if (SC < 3348) {
                                                                            if (diffMVdiffVal < -6) {
                                                                                return 4;
                                                                            }
                                                                            else if (diffMVdiffVal >= -6) {
                                                                                if (Rs < 1844) {
                                                                                    return 1;
                                                                                }
                                                                                else if (Rs >= 1844) {
                                                                                    return 8;
                                                                                }
                                                                            }
                                                                        }
                                                                        else if (SC >= 3348) {
                                                                            if (mu_mv_mag_sq < 63) {
                                                                                if (Cs < 1530) {
                                                                                    return 1;
                                                                                }
                                                                                else if (Cs >= 1530) {
                                                                                    return 4;
                                                                                }
                                                                            }
                                                                            else if (mu_mv_mag_sq >= 63) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (MVDiff >= 19) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (Rs >= 2388) {
                                                                    return 1;
                                                                }
                                                            }
                                                            else if (Cs >= 1537) {
                                                                if (Cs < 1798) {
                                                                    if (mu_mv_mag_sq < 65) {
                                                                        if (negBalance < 50) {
                                                                            if (ssDCval < 40) {
                                                                                return 1;
                                                                            }
                                                                            else if (ssDCval >= 40) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                        else if (negBalance >= 50) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                    else if (mu_mv_mag_sq >= 65) {
                                                                        if (SC < 4071) {
                                                                            if (MVDiff < 13) {
                                                                                return 4;
                                                                            }
                                                                            else if (MVDiff >= 13) {
                                                                                if (MVDiff < 14) {
                                                                                    return 1;
                                                                                }
                                                                                else if (MVDiff >= 14) {
                                                                                    if (Cs < 1559) {
                                                                                        return 4;
                                                                                    }
                                                                                    else if (Cs >= 1559) {
                                                                                        if (Rs < 1929) {
                                                                                            return 1;
                                                                                        }
                                                                                        else if (Rs >= 1929) {
                                                                                            return 4;
                                                                                        }
                                                                                    }
                                                                                }
                                                                            }
                                                                        }
                                                                        else if (SC >= 4071) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                }
                                                                else if (Cs >= 1798) {
                                                                    if (Rs < 1369) {
                                                                        return 8;
                                                                    }
                                                                    else if (Rs >= 1369) {
                                                                        return 4;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (refDCval >= 91) {
                                                        if (mu_mv_mag_sq < 17) {
                                                            if (diffTSC < -2) {
                                                                return 1;
                                                            }
                                                            else if (diffTSC >= -2) {
                                                                if (ssDCval < 133) {
                                                                    return 4;
                                                                }
                                                                else if (ssDCval >= 133) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                        else if (mu_mv_mag_sq >= 17) {
                                                            if (refDCval < 111) {
                                                                return 4;
                                                            }
                                                            else if (refDCval >= 111) {
                                                                if (MVDiff < 21) {
                                                                    return 8;
                                                                }
                                                                else if (MVDiff >= 21) {
                                                                    return 2;
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (CsDiff >= 13) {
                                                    if (Rs < 1215) {
                                                        return 2;
                                                    }
                                                    else if (Rs >= 1215) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                        }
                                        else if (mu_mv_mag_sq >= 99) {
                                            if (Cs < 1515) {
                                                if (Cs < 1481) {
                                                    return 8;
                                                }
                                                else if (Cs >= 1481) {
                                                    return 2;
                                                }
                                            }
                                            else if (Cs >= 1515) {
                                                if (negBalance < 38) {
                                                    return 8;
                                                }
                                                else if (negBalance >= 38) {
                                                    if (ssDCval < 84) {
                                                        if (tcor < 95) {
                                                            if (TSC < 6) {
                                                                return 8;
                                                            }
                                                            else if (TSC >= 6) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (tcor >= 95) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (ssDCval >= 84) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (Rs >= 3029) {
                                    if (refDCval < 108) {
                                        if (refDCval < 54) {
                                            if (mu_mv_mag_sq < 36) {
                                                if (diffAFD < 1) {
                                                    return 8;
                                                }
                                                else if (diffAFD >= 1) {
                                                    if (Cs < 3240) {
                                                        return 8;
                                                    }
                                                    else if (Cs >= 3240) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                            else if (mu_mv_mag_sq >= 36) {
                                                if (SC < 7671) {
                                                    if (RsDiff < 3) {
                                                        if (diffAFD < 1) {
                                                            if (tcor < 90) {
                                                                return 8;
                                                            }
                                                            else if (tcor >= 90) {
                                                                if (MVDiff < 5) {
                                                                    if (Rs < 3510) {
                                                                        return 8;
                                                                    }
                                                                    else if (Rs >= 3510) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (MVDiff >= 5) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                        else if (diffAFD >= 1) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (RsDiff >= 3) {
                                                        return 4;
                                                    }
                                                }
                                                else if (SC >= 7671) {
                                                    if (Rs < 3899) {
                                                        return 2;
                                                    }
                                                    else if (Rs >= 3899) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                        }
                                        else if (refDCval >= 54) {
                                            if (negBalance < 55) {
                                                if (AFD < 7) {
                                                    if (negBalance < 37) {
                                                        return 8;
                                                    }
                                                    else if (negBalance >= 37) {
                                                        if (diffTSC < -1) {
                                                            if (posBalance < 55) {
                                                                return 4;
                                                            }
                                                            else if (posBalance >= 55) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (diffTSC >= -1) {
                                                            if (diffRsCsdiff < 9) {
                                                                return 8;
                                                            }
                                                            else if (diffRsCsdiff >= 9) {
                                                                if (TSC < 6) {
                                                                    return 4;
                                                                }
                                                                else if (TSC >= 6) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (AFD >= 7) {
                                                    if (negBalance < 34) {
                                                        if (Cs < 2411) {
                                                            return 4;
                                                        }
                                                        else if (Cs >= 2411) {
                                                            return 2;
                                                        }
                                                    }
                                                    else if (negBalance >= 34) {
                                                        if (negBalance < 47) {
                                                            if (posBalance < 57) {
                                                                if (diffMVdiffVal < -1) {
                                                                    if (tcor < 93) {
                                                                        if (refDCval < 104) {
                                                                            if (diffRsCsdiff < -6) {
                                                                                return 4;
                                                                            }
                                                                            else if (diffRsCsdiff >= -6) {
                                                                                if (negBalance < 41) {
                                                                                    return 4;
                                                                                }
                                                                                else if (negBalance >= 41) {
                                                                                    return 8;
                                                                                }
                                                                            }
                                                                        }
                                                                        else if (refDCval >= 104) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                    else if (tcor >= 93) {
                                                                        if (diffRsCsdiff < -7) {
                                                                            return 8;
                                                                        }
                                                                        else if (diffRsCsdiff >= -7) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                }
                                                                else if (diffMVdiffVal >= -1) {
                                                                    if (SC < 7987) {
                                                                        if (diffMVdiffVal < 8) {
                                                                            if (tcor < 93) {
                                                                                return 8;
                                                                            }
                                                                            else if (tcor >= 93) {
                                                                                if (diffAFD < 1) {
                                                                                    return 4;
                                                                                }
                                                                                else if (diffAFD >= 1) {
                                                                                    return 8;
                                                                                }
                                                                            }
                                                                        }
                                                                        else if (diffMVdiffVal >= 8) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                    else if (SC >= 7987) {
                                                                        return 4;
                                                                    }
                                                                }
                                                            }
                                                            else if (posBalance >= 57) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (negBalance >= 47) {
                                                            if (diffRsCsdiff < -8) {
                                                                if (AFD < 8) {
                                                                    return 4;
                                                                }
                                                                else if (AFD >= 8) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (diffRsCsdiff >= -8) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                            else if (negBalance >= 55) {
                                                if (RsDiff < 4) {
                                                    return 2;
                                                }
                                                else if (RsDiff >= 4) {
                                                    return 8;
                                                }
                                            }
                                        }
                                    }
                                    else if (refDCval >= 108) {
                                        if (RsDiff < 5) {
                                            if (SC < 5200) {
                                                return 4;
                                            }
                                            else if (SC >= 5200) {
                                                return 8;
                                            }
                                        }
                                        else if (RsDiff >= 5) {
                                            if (Cs < 2072) {
                                                if (mu_mv_mag_sq < 3) {
                                                    if (posBalance < 43) {
                                                        return 4;
                                                    }
                                                    else if (posBalance >= 43) {
                                                        return 8;
                                                    }
                                                }
                                                else if (mu_mv_mag_sq >= 3) {
                                                    if (tcor < 96) {
                                                        return 4;
                                                    }
                                                    else if (tcor >= 96) {
                                                        if (Cs < 1662) {
                                                            return 4;
                                                        }
                                                        else if (Cs >= 1662) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (Cs >= 2072) {
                                                if (ssDCval < 119) {
                                                    return 2;
                                                }
                                                else if (ssDCval >= 119) {
                                                    if (Rs < 4821) {
                                                        if (RsDiff < 6) {
                                                            return 8;
                                                        }
                                                        else if (RsDiff >= 6) {
                                                            if (diffTSC < -1) {
                                                                if (Rs < 4489) {
                                                                    return 4;
                                                                }
                                                                else if (Rs >= 4489) {
                                                                    return 2;
                                                                }
                                                            }
                                                            else if (diffTSC >= -1) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                    else if (Rs >= 4821) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                            else if (TSC >= 7) {
                                if (CsDiff < 10) {
                                    if (mcTcor < 93) {
                                        if (Rs < 2418) {
                                            if (AFD < 10) {
                                                if (ssDCval < 79) {
                                                    return 2;
                                                }
                                                else if (ssDCval >= 79) {
                                                    if (posBalance < 61) {
                                                        if (CsDiff < 6) {
                                                            if (Rs < 2048) {
                                                                if (diffAFD < 1) {
                                                                    if (Rs < 1845) {
                                                                        return 8;
                                                                    }
                                                                    else if (Rs >= 1845) {
                                                                        if (SC < 3597) {
                                                                            return 2;
                                                                        }
                                                                        else if (SC >= 3597) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                }
                                                                else if (diffAFD >= 1) {
                                                                    return 2;
                                                                }
                                                            }
                                                            else if (Rs >= 2048) {
                                                                return 2;
                                                            }
                                                        }
                                                        else if (CsDiff >= 6) {
                                                            if (AFD < 9) {
                                                                return 8;
                                                            }
                                                            else if (AFD >= 9) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                    else if (posBalance >= 61) {
                                                        if (ssDCval < 89) {
                                                            return 8;
                                                        }
                                                        else if (ssDCval >= 89) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (AFD >= 10) {
                                                if (ssDCval < 92) {
                                                    if (diffAFD < -2) {
                                                        return 1;
                                                    }
                                                    else if (diffAFD >= -2) {
                                                        return 4;
                                                    }
                                                }
                                                else if (ssDCval >= 92) {
                                                    if (Rs < 2248) {
                                                        return 2;
                                                    }
                                                    else if (Rs >= 2248) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                        }
                                        else if (Rs >= 2418) {
                                            if (SC < 4364) {
                                                return 4;
                                            }
                                            else if (SC >= 4364) {
                                                if (mu_mv_mag_sq < 83) {
                                                    if (SC < 7477) {
                                                        if (gchDC < 1) {
                                                            if (RsDiff < 5) {
                                                                return 8;
                                                            }
                                                            else if (RsDiff >= 5) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (gchDC >= 1) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (SC >= 7477) {
                                                        return 4;
                                                    }
                                                }
                                                else if (mu_mv_mag_sq >= 83) {
                                                    return 8;
                                                }
                                            }
                                        }
                                    }
                                    else if (mcTcor >= 93) {
                                        if (Rs < 3615) {
                                            if (posBalance < 55) {
                                                if (RsDiff < 7) {
                                                    if (ssDCval < 89) {
                                                        if (Cs < 1843) {
                                                            if (posBalance < 53) {
                                                                if (Rs < 1856) {
                                                                    return 1;
                                                                }
                                                                else if (Rs >= 1856) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (posBalance >= 53) {
                                                                if (Cs < 1684) {
                                                                    return 2;
                                                                }
                                                                else if (Cs >= 1684) {
                                                                    return 1;
                                                                }
                                                            }
                                                        }
                                                        else if (Cs >= 1843) {
                                                            if (mcTcor < 94) {
                                                                if (Rs < 2860) {
                                                                    return 1;
                                                                }
                                                                else if (Rs >= 2860) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (mcTcor >= 94) {
                                                                if (CsDiff < 4) {
                                                                    if (MVDiff < 63) {
                                                                        return 2;
                                                                    }
                                                                    else if (MVDiff >= 63) {
                                                                        return 8;
                                                                    }
                                                                }
                                                                else if (CsDiff >= 4) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (ssDCval >= 89) {
                                                        if (tcor < 97) {
                                                            return 8;
                                                        }
                                                        else if (tcor >= 97) {
                                                            return 1;
                                                        }
                                                    }
                                                }
                                                else if (RsDiff >= 7) {
                                                    if (mcTcor < 94) {
                                                        return 1;
                                                    }
                                                    else if (mcTcor >= 94) {
                                                        if (negBalance < 35) {
                                                            if (RsDiff < 10) {
                                                                return 2;
                                                            }
                                                            else if (RsDiff >= 10) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (negBalance >= 35) {
                                                            if (MVDiff < 26) {
                                                                return 4;
                                                            }
                                                            else if (MVDiff >= 26) {
                                                                return 2;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                            else if (posBalance >= 55) {
                                                if (SC < 5125) {
                                                    if (Cs < 1918) {
                                                        if (SC < 4218) {
                                                            if (gchDC < 3) {
                                                                return 2;
                                                            }
                                                            else if (gchDC >= 3) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (SC >= 4218) {
                                                            if (RsDiff < 7) {
                                                                if (RsCsDiff < 33) {
                                                                    return 4;
                                                                }
                                                                else if (RsCsDiff >= 33) {
                                                                    return 1;
                                                                }
                                                            }
                                                            else if (RsDiff >= 7) {
                                                                return 2;
                                                            }
                                                        }
                                                    }
                                                    else if (Cs >= 1918) {
                                                        if (Rs < 2173) {
                                                            if (SC < 3252) {
                                                                return 8;
                                                            }
                                                            else if (SC >= 3252) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (Rs >= 2173) {
                                                            if (ssDCval < 87) {
                                                                return 4;
                                                            }
                                                            else if (ssDCval >= 87) {
                                                                return 2;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (SC >= 5125) {
                                                    if (ssDCval < 58) {
                                                        if (Cs < 3418) {
                                                            if (ssDCval < 50) {
                                                                if (Cs < 3182) {
                                                                    return 2;
                                                                }
                                                                else if (Cs >= 3182) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (ssDCval >= 50) {
                                                                return 2;
                                                            }
                                                        }
                                                        else if (Cs >= 3418) {
                                                            if (Cs < 3514) {
                                                                return 8;
                                                            }
                                                            else if (Cs >= 3514) {
                                                                if (SC < 7046) {
                                                                    return 4;
                                                                }
                                                                else if (SC >= 7046) {
                                                                    return 2;
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (ssDCval >= 58) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                        }
                                        else if (Rs >= 3615) {
                                            if (refDCval < 127) {
                                                if (RsDiff < 8) {
                                                    if (MVDiff < 23) {
                                                        if (RsDiff < 3) {
                                                            return 2;
                                                        }
                                                        else if (RsDiff >= 3) {
                                                            if (AFD < 8) {
                                                                return 8;
                                                            }
                                                            else if (AFD >= 8) {
                                                                if (diffTSC < 1) {
                                                                    if (ssDCval < 56) {
                                                                        if (diffRsCsdiff < 4) {
                                                                            return 4;
                                                                        }
                                                                        else if (diffRsCsdiff >= 4) {
                                                                            if (Rs < 3672) {
                                                                                return 2;
                                                                            }
                                                                            else if (Rs >= 3672) {
                                                                                if (posBalance < 59) {
                                                                                    return 4;
                                                                                }
                                                                                else if (posBalance >= 59) {
                                                                                    return 2;
                                                                                }
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (ssDCval >= 56) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (diffTSC >= 1) {
                                                                    return 2;
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (MVDiff >= 23) {
                                                        if (MVDiff < 25) {
                                                            if (Rs < 3723) {
                                                                return 8;
                                                            }
                                                            else if (Rs >= 3723) {
                                                                return 2;
                                                            }
                                                        }
                                                        else if (MVDiff >= 25) {
                                                            if (tcor < 88) {
                                                                return 4;
                                                            }
                                                            else if (tcor >= 88) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (RsDiff >= 8) {
                                                    if (SC < 6745) {
                                                        if (Cs < 2227) {
                                                            if (negBalance < 33) {
                                                                if (tcor < 93) {
                                                                    return 4;
                                                                }
                                                                else if (tcor >= 93) {
                                                                    return 2;
                                                                }
                                                            }
                                                            else if (negBalance >= 33) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (Cs >= 2227) {
                                                            return 2;
                                                        }
                                                    }
                                                    else if (SC >= 6745) {
                                                        if (posBalance < 39) {
                                                            if (SC < 7143) {
                                                                return 4;
                                                            }
                                                            else if (SC >= 7143) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (posBalance >= 39) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (refDCval >= 127) {
                                                if (Cs < 2185) {
                                                    return 4;
                                                }
                                                else if (Cs >= 2185) {
                                                    return 2;
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (CsDiff >= 10) {
                                    if (ssDCval < 69) {
                                        if (Cs < 2322) {
                                            return 8;
                                        }
                                        else if (Cs >= 2322) {
                                            return 4;
                                        }
                                    }
                                    else if (ssDCval >= 69) {
                                        if (CsDiff < 13) {
                                            return 1;
                                        }
                                        else if (CsDiff >= 13) {
                                            return 8;
                                        }
                                    }
                                }
                            }
                        }
                        else if (TSC >= 8) {
                            if (mu_mv_mag_sq < 47) {
                                if (mcTcor < 94) {
                                    if (CsDiff < 10) {
                                        if (refDCval < 78) {
                                            if (tcor < 83) {
                                                if (Rs < 1666) {
                                                    return 8;
                                                }
                                                else if (Rs >= 1666) {
                                                    if (AFD < 14) {
                                                        return 4;
                                                    }
                                                    else if (AFD >= 14) {
                                                        if (MVDiff < 22) {
                                                            if (TSC < 12) {
                                                                return 4;
                                                            }
                                                            else if (TSC >= 12) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (MVDiff >= 22) {
                                                            if (tcor < 80) {
                                                                if (diffRsCsdiff < -36) {
                                                                    return 4;
                                                                }
                                                                else if (diffRsCsdiff >= -36) {
                                                                    if (Rs < 1687) {
                                                                        return 4;
                                                                    }
                                                                    else if (Rs >= 1687) {
                                                                        return 8;
                                                                    }
                                                                }
                                                            }
                                                            else if (tcor >= 80) {
                                                                if (ssDCval < 74) {
                                                                    return 4;
                                                                }
                                                                else if (ssDCval >= 74) {
                                                                    if (diffRsCsdiff < -21) {
                                                                        return 8;
                                                                    }
                                                                    else if (diffRsCsdiff >= -21) {
                                                                        return 4;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                            else if (tcor >= 83) {
                                                if (RsCsDiff < 46) {
                                                    if (ssDCval < 58) {
                                                        if (mcTcor < 93) {
                                                            if (diffMVdiffVal < 7) {
                                                                return 2;
                                                            }
                                                            else if (diffMVdiffVal >= 7) {
                                                                if (TSC < 9) {
                                                                    if (Rs < 3802) {
                                                                        return 8;
                                                                    }
                                                                    else if (Rs >= 3802) {
                                                                        return 2;
                                                                    }
                                                                }
                                                                else if (TSC >= 9) {
                                                                    return 2;
                                                                }
                                                            }
                                                        }
                                                        else if (mcTcor >= 93) {
                                                            if (ssDCval < 56) {
                                                                return 2;
                                                            }
                                                            else if (ssDCval >= 56) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                    else if (ssDCval >= 58) {
                                                        return 8;
                                                    }
                                                }
                                                else if (RsCsDiff >= 46) {
                                                    return 4;
                                                }
                                            }
                                        }
                                        else if (refDCval >= 78) {
                                            if (AFD < 13) {
                                                if (posBalance < 56) {
                                                    if (refDCval < 107) {
                                                        if (tcor < 91) {
                                                            if (diffTSC < -1) {
                                                                return 4;
                                                            }
                                                            else if (diffTSC >= -1) {
                                                                if (diffAFD < -3) {
                                                                    return 4;
                                                                }
                                                                else if (diffAFD >= -3) {
                                                                    if (posBalance < 47) {
                                                                        return 8;
                                                                    }
                                                                    else if (posBalance >= 47) {
                                                                        if (Rs < 3559) {
                                                                            return 8;
                                                                        }
                                                                        else if (Rs >= 3559) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (tcor >= 91) {
                                                            if (gchDC < 2) {
                                                                return 1;
                                                            }
                                                            else if (gchDC >= 2) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                    else if (refDCval >= 107) {
                                                        if (tcor < 86) {
                                                            return 1;
                                                        }
                                                        else if (tcor >= 86) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                                else if (posBalance >= 56) {
                                                    if (refDCval < 98) {
                                                        if (diffRsCsdiff < -16) {
                                                            return 2;
                                                        }
                                                        else if (diffRsCsdiff >= -16) {
                                                            if (ssDCval < 91) {
                                                                if (AFD < 11) {
                                                                    return 4;
                                                                }
                                                                else if (AFD >= 11) {
                                                                    return 1;
                                                                }
                                                            }
                                                            else if (ssDCval >= 91) {
                                                                if (RsDiff < 7) {
                                                                    return 8;
                                                                }
                                                                else if (RsDiff >= 7) {
                                                                    return 2;
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (refDCval >= 98) {
                                                        if (refDCval < 123) {
                                                            return 4;
                                                        }
                                                        else if (refDCval >= 123) {
                                                            if (tcor < 81) {
                                                                return 4;
                                                            }
                                                            else if (tcor >= 81) {
                                                                if (negBalance < 52) {
                                                                    return 8;
                                                                }
                                                                else if (negBalance >= 52) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                            else if (AFD >= 13) {
                                                if (posBalance < 56) {
                                                    if (MVDiff < 21) {
                                                        if (diffRsCsdiff < -18) {
                                                            if (posBalance < 41) {
                                                                return 2;
                                                            }
                                                            else if (posBalance >= 41) {
                                                                if (negBalance < 60) {
                                                                    if (MVDiff < 8) {
                                                                        return 4;
                                                                    }
                                                                    else if (MVDiff >= 8) {
                                                                        if (TSC < 10) {
                                                                            return 1;
                                                                        }
                                                                        else if (TSC >= 10) {
                                                                            if (CsDiff < 9) {
                                                                                return 4;
                                                                            }
                                                                            else if (CsDiff >= 9) {
                                                                                return 1;
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                                else if (negBalance >= 60) {
                                                                    return 2;
                                                                }
                                                            }
                                                        }
                                                        else if (diffRsCsdiff >= -18) {
                                                            if (refDCval < 80) {
                                                                return 8;
                                                            }
                                                            else if (refDCval >= 80) {
                                                                if (TSC < 11) {
                                                                    if (SC < 4715) {
                                                                        if (Cs < 1590) {
                                                                            return 8;
                                                                        }
                                                                        else if (Cs >= 1590) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                    else if (SC >= 4715) {
                                                                        if (Rs < 3056) {
                                                                            return 8;
                                                                        }
                                                                        else if (Rs >= 3056) {
                                                                            if (Cs < 2488) {
                                                                                return 1;
                                                                            }
                                                                            else if (Cs >= 2488) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                                else if (TSC >= 11) {
                                                                    if (Rs < 2362) {
                                                                        if (RsCsDiff < 154) {
                                                                            return 4;
                                                                        }
                                                                        else if (RsCsDiff >= 154) {
                                                                            if (Rs < 2228) {
                                                                                return 2;
                                                                            }
                                                                            else if (Rs >= 2228) {
                                                                                return 1;
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (Rs >= 2362) {
                                                                        if (Cs < 2361) {
                                                                            if (mu_mv_mag_sq < 33) {
                                                                                if (Cs < 2070) {
                                                                                    return 1;
                                                                                }
                                                                                else if (Cs >= 2070) {
                                                                                    return 8;
                                                                                }
                                                                            }
                                                                            else if (mu_mv_mag_sq >= 33) {
                                                                                return 1;
                                                                            }
                                                                        }
                                                                        else if (Cs >= 2361) {
                                                                            if (diffMVdiffVal < -1) {
                                                                                return 2;
                                                                            }
                                                                            else if (diffMVdiffVal >= -1) {
                                                                                if (Rs < 2878) {
                                                                                    return 8;
                                                                                }
                                                                                else if (Rs >= 2878) {
                                                                                    return 4;
                                                                                }
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (MVDiff >= 21) {
                                                        if (Rs < 2518) {
                                                            if (TSC < 10) {
                                                                if (negBalance < 40) {
                                                                    return 8;
                                                                }
                                                                else if (negBalance >= 40) {
                                                                    return 1;
                                                                }
                                                            }
                                                            else if (TSC >= 10) {
                                                                if (mcTcor < 92) {
                                                                    if (RsDiff < 12) {
                                                                        if (gchDC < 2) {
                                                                            return 4;
                                                                        }
                                                                        else if (gchDC >= 2) {
                                                                            if (MVDiff < 25) {
                                                                                return 8;
                                                                            }
                                                                            else if (MVDiff >= 25) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (RsDiff >= 12) {
                                                                        return 1;
                                                                    }
                                                                }
                                                                else if (mcTcor >= 92) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                        else if (Rs >= 2518) {
                                                            if (SC < 4972) {
                                                                return 8;
                                                            }
                                                            else if (SC >= 4972) {
                                                                if (RsCsDiff < 278) {
                                                                    if (TSC < 11) {
                                                                        return 4;
                                                                    }
                                                                    else if (TSC >= 11) {
                                                                        if (Rs < 3331) {
                                                                            return 1;
                                                                        }
                                                                        else if (Rs >= 3331) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                }
                                                                else if (RsCsDiff >= 278) {
                                                                    if (RsCsDiff < 493) {
                                                                        return 2;
                                                                    }
                                                                    else if (RsCsDiff >= 493) {
                                                                        return 4;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (posBalance >= 56) {
                                                    if (refDCval < 118) {
                                                        if (TSC < 13) {
                                                            if (MVDiff < 3) {
                                                                if (Cs < 2391) {
                                                                    return 2;
                                                                }
                                                                else if (Cs >= 2391) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (MVDiff >= 3) {
                                                                if (negBalance < 47) {
                                                                    if (TSC < 11) {
                                                                        if (MVDiff < 13) {
                                                                            return 2;
                                                                        }
                                                                        else if (MVDiff >= 13) {
                                                                            if (MVDiff < 51) {
                                                                                return 4;
                                                                            }
                                                                            else if (MVDiff >= 51) {
                                                                                return 8;
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (TSC >= 11) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (negBalance >= 47) {
                                                                    if (mcTcor < 84) {
                                                                        return 4;
                                                                    }
                                                                    else if (mcTcor >= 84) {
                                                                        if (refDCval < 81) {
                                                                            if (RsDiff < 8) {
                                                                                return 1;
                                                                            }
                                                                            else if (RsDiff >= 8) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                        else if (refDCval >= 81) {
                                                                            if (RsCsDiff < 181) {
                                                                                if (posBalance < 59) {
                                                                                    if (RsDiff < 11) {
                                                                                        if (negBalance < 50) {
                                                                                            if (refDCval < 107) {
                                                                                                if (SC < 4711) {
                                                                                                    return 4;
                                                                                                }
                                                                                                else if (SC >= 4711) {
                                                                                                    return 1;
                                                                                                }
                                                                                            }
                                                                                            else if (refDCval >= 107) {
                                                                                                return 2;
                                                                                            }
                                                                                        }
                                                                                        else if (negBalance >= 50) {
                                                                                            return 2;
                                                                                        }
                                                                                    }
                                                                                    else if (RsDiff >= 11) {
                                                                                        if (Cs < 1834) {
                                                                                            return 8;
                                                                                        }
                                                                                        else if (Cs >= 1834) {
                                                                                            return 4;
                                                                                        }
                                                                                    }
                                                                                }
                                                                                else if (posBalance >= 59) {
                                                                                    if (negBalance < 54) {
                                                                                        if (ssDCval < 92) {
                                                                                            if (Cs < 2005) {
                                                                                                return 1;
                                                                                            }
                                                                                            else if (Cs >= 2005) {
                                                                                                return 2;
                                                                                            }
                                                                                        }
                                                                                        else if (ssDCval >= 92) {
                                                                                            if (Rs < 2244) {
                                                                                                return 4;
                                                                                            }
                                                                                            else if (Rs >= 2244) {
                                                                                                return 1;
                                                                                            }
                                                                                        }
                                                                                    }
                                                                                    else if (negBalance >= 54) {
                                                                                        if (MVDiff < 6) {
                                                                                            return 2;
                                                                                        }
                                                                                        else if (MVDiff >= 6) {
                                                                                            return 1;
                                                                                        }
                                                                                    }
                                                                                }
                                                                            }
                                                                            else if (RsCsDiff >= 181) {
                                                                                if (MVDiff < 7) {
                                                                                    return 8;
                                                                                }
                                                                                else if (MVDiff >= 7) {
                                                                                    return 2;
                                                                                }
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (TSC >= 13) {
                                                            if (SC < 4870) {
                                                                if (Rs < 2377) {
                                                                    if (AFD < 19) {
                                                                        return 4;
                                                                    }
                                                                    else if (AFD >= 19) {
                                                                        return 2;
                                                                    }
                                                                }
                                                                else if (Rs >= 2377) {
                                                                    return 1;
                                                                }
                                                            }
                                                            else if (SC >= 4870) {
                                                                if (ssDCval < 88) {
                                                                    return 1;
                                                                }
                                                                else if (ssDCval >= 88) {
                                                                    if (RsCsDiff < 381) {
                                                                        return 8;
                                                                    }
                                                                    else if (RsCsDiff >= 381) {
                                                                        return 2;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (refDCval >= 118) {
                                                        if (tcor < 74) {
                                                            if (RsDiff < 16) {
                                                                return 1;
                                                            }
                                                            else if (RsDiff >= 16) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (tcor >= 74) {
                                                            if (AFD < 19) {
                                                                if (mcTcor < 89) {
                                                                    return 4;
                                                                }
                                                                else if (mcTcor >= 89) {
                                                                    if (tcor < 87) {
                                                                        if (AFD < 18) {
                                                                            if (Rs < 3033) {
                                                                                return 4;
                                                                            }
                                                                            else if (Rs >= 3033) {
                                                                                return 2;
                                                                            }
                                                                        }
                                                                        else if (AFD >= 18) {
                                                                            return 2;
                                                                        }
                                                                    }
                                                                    else if (tcor >= 87) {
                                                                        if (AFD < 14) {
                                                                            if (Cs < 1601) {
                                                                                return 2;
                                                                            }
                                                                            else if (Cs >= 1601) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                        else if (AFD >= 14) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                            else if (AFD >= 19) {
                                                                if (RsDiff < 13) {
                                                                    if (TSC < 13) {
                                                                        if (SC < 3648) {
                                                                            return 4;
                                                                        }
                                                                        else if (SC >= 3648) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                    else if (TSC >= 13) {
                                                                        return 1;
                                                                    }
                                                                }
                                                                else if (RsDiff >= 13) {
                                                                    if (Rs < 2770) {
                                                                        return 4;
                                                                    }
                                                                    else if (Rs >= 2770) {
                                                                        return 2;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (CsDiff >= 10) {
                                        if (posBalance < 57) {
                                            if (negBalance < 50) {
                                                if (MVDiff < 4) {
                                                    return 8;
                                                }
                                                else if (MVDiff >= 4) {
                                                    if (ssDCval < 86) {
                                                        if (Rs < 3067) {
                                                            if (posBalance < 54) {
                                                                if (AFD < 17) {
                                                                    if (Cs < 2916) {
                                                                        return 1;
                                                                    }
                                                                    else if (Cs >= 2916) {
                                                                        if (MVDiff < 12) {
                                                                            return 1;
                                                                        }
                                                                        else if (MVDiff >= 12) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                }
                                                                else if (AFD >= 17) {
                                                                    return 1;
                                                                }
                                                            }
                                                            else if (posBalance >= 54) {
                                                                if (TSC < 10) {
                                                                    return 1;
                                                                }
                                                                else if (TSC >= 10) {
                                                                    if (SC < 5667) {
                                                                        return 8;
                                                                    }
                                                                    else if (SC >= 5667) {
                                                                        return 1;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (Rs >= 3067) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (ssDCval >= 86) {
                                                        if (ssDCval < 144) {
                                                            if (gchDC < 2) {
                                                                if (RsDiff < 10) {
                                                                    if (posBalance < 52) {
                                                                        return 1;
                                                                    }
                                                                    else if (posBalance >= 52) {
                                                                        if (negBalance < 42) {
                                                                            if (Rs < 2471) {
                                                                                return 8;
                                                                            }
                                                                            else if (Rs >= 2471) {
                                                                                return 1;
                                                                            }
                                                                        }
                                                                        else if (negBalance >= 42) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                }
                                                                else if (RsDiff >= 10) {
                                                                    if (diffAFD < -4) {
                                                                        return 4;
                                                                    }
                                                                    else if (diffAFD >= -4) {
                                                                        if (mu_mv_mag_sq < 45) {
                                                                            return 1;
                                                                        }
                                                                        else if (mu_mv_mag_sq >= 45) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                            else if (gchDC >= 2) {
                                                                if (ssDCval < 93) {
                                                                    return 1;
                                                                }
                                                                else if (ssDCval >= 93) {
                                                                    if (Cs < 2812) {
                                                                        return 4;
                                                                    }
                                                                    else if (Cs >= 2812) {
                                                                        return 8;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (ssDCval >= 144) {
                                                            if (Rs < 1712) {
                                                                return 8;
                                                            }
                                                            else if (Rs >= 1712) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                            else if (negBalance >= 50) {
                                                if (tcor < 75) {
                                                    if (refDCval < 122) {
                                                        if (tcor < 66) {
                                                            if (Cs < 3077) {
                                                                if (gchDC < 2) {
                                                                    if (Rs < 2154) {
                                                                        return 1;
                                                                    }
                                                                    else if (Rs >= 2154) {
                                                                        return 8;
                                                                    }
                                                                }
                                                                else if (gchDC >= 2) {
                                                                    if (posBalance < 28) {
                                                                        return 8;
                                                                    }
                                                                    else if (posBalance >= 28) {
                                                                        return 2;
                                                                    }
                                                                }
                                                            }
                                                            else if (Cs >= 3077) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (tcor >= 66) {
                                                            if (RsCsDiff < 600) {
                                                                if (Rs < 2458) {
                                                                    if (TSC < 9) {
                                                                        return 8;
                                                                    }
                                                                    else if (TSC >= 9) {
                                                                        return 1;
                                                                    }
                                                                }
                                                                else if (Rs >= 2458) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (RsCsDiff >= 600) {
                                                                if (Rs < 3693) {
                                                                    return 1;
                                                                }
                                                                else if (Rs >= 3693) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (refDCval >= 122) {
                                                        if (posBalance < 50) {
                                                            return 8;
                                                        }
                                                        else if (posBalance >= 50) {
                                                            if (ssDCval < 124) {
                                                                if (Rs < 2551) {
                                                                    return 1;
                                                                }
                                                                else if (Rs >= 2551) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (ssDCval >= 124) {
                                                                return 1;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (tcor >= 75) {
                                                    if (MVDiff < 14) {
                                                        if (mu_mv_mag_sq < 25) {
                                                            if (refDCval < 102) {
                                                                return 4;
                                                            }
                                                            else if (refDCval >= 102) {
                                                                if (Rs < 4147) {
                                                                    return 8;
                                                                }
                                                                else if (Rs >= 4147) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                        else if (mu_mv_mag_sq >= 25) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (MVDiff >= 14) {
                                                        if (Rs < 2877) {
                                                            if (tcor < 78) {
                                                                return 4;
                                                            }
                                                            else if (tcor >= 78) {
                                                                if (gchDC < 1) {
                                                                    return 2;
                                                                }
                                                                else if (gchDC >= 1) {
                                                                    return 1;
                                                                }
                                                            }
                                                        }
                                                        else if (Rs >= 2877) {
                                                            if (Rs < 4587) {
                                                                return 8;
                                                            }
                                                            else if (Rs >= 4587) {
                                                                return 2;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (posBalance >= 57) {
                                            if (Rs < 3015) {
                                                if (Cs < 2066) {
                                                    if (TSC < 16) {
                                                        if (mu_mv_mag_sq < 2) {
                                                            if (RsCsDiff < 329) {
                                                                if (tcor < 81) {
                                                                    if (Cs < 2004) {
                                                                        if (Rs < 2302) {
                                                                            return 2;
                                                                        }
                                                                        else if (Rs >= 2302) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                    else if (Cs >= 2004) {
                                                                        return 2;
                                                                    }
                                                                }
                                                                else if (tcor >= 81) {
                                                                    return 2;
                                                                }
                                                            }
                                                            else if (RsCsDiff >= 329) {
                                                                return 2;
                                                            }
                                                        }
                                                        else if (mu_mv_mag_sq >= 2) {
                                                            if (negBalance < 54) {
                                                                if (mcTcor < 84) {
                                                                    return 1;
                                                                }
                                                                else if (mcTcor >= 84) {
                                                                    if (Rs < 1918) {
                                                                        return 1;
                                                                    }
                                                                    else if (Rs >= 1918) {
                                                                        return 4;
                                                                    }
                                                                }
                                                            }
                                                            else if (negBalance >= 54) {
                                                                if (ssDCval < 133) {
                                                                    if (AFD < 22) {
                                                                        return 2;
                                                                    }
                                                                    else if (AFD >= 22) {
                                                                        if (AFD < 31) {
                                                                            if (Rs < 2230) {
                                                                                return 1;
                                                                            }
                                                                            else if (Rs >= 2230) {
                                                                                return 2;
                                                                            }
                                                                        }
                                                                        else if (AFD >= 31) {
                                                                            return 2;
                                                                        }
                                                                    }
                                                                }
                                                                else if (ssDCval >= 133) {
                                                                    if (TSC < 13) {
                                                                        return 1;
                                                                    }
                                                                    else if (TSC >= 13) {
                                                                        return 4;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (TSC >= 16) {
                                                        if (Cs < 1883) {
                                                            return 8;
                                                        }
                                                        else if (Cs >= 1883) {
                                                            if (AFD < 22) {
                                                                return 4;
                                                            }
                                                            else if (AFD >= 22) {
                                                                return 1;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (Cs >= 2066) {
                                                    if (mu_mv_mag_sq < 14) {
                                                        if (ssDCval < 89) {
                                                            if (Rs < 1949) {
                                                                if (mcTcor < 69) {
                                                                    return 4;
                                                                }
                                                                else if (mcTcor >= 69) {
                                                                    return 1;
                                                                }
                                                            }
                                                            else if (Rs >= 1949) {
                                                                if (tcor < 73) {
                                                                    return 4;
                                                                }
                                                                else if (tcor >= 73) {
                                                                    if (SC < 5013) {
                                                                        return 4;
                                                                    }
                                                                    else if (SC >= 5013) {
                                                                        return 8;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (ssDCval >= 89) {
                                                            if (refDCval < 136) {
                                                                if (tcor < 84) {
                                                                    if (TSC < 11) {
                                                                        if (Cs < 2334) {
                                                                            return 2;
                                                                        }
                                                                        else if (Cs >= 2334) {
                                                                            if (TSC < 10) {
                                                                                return 2;
                                                                            }
                                                                            else if (TSC >= 10) {
                                                                                return 1;
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (TSC >= 11) {
                                                                        if (diffRsCsdiff < 10) {
                                                                            return 1;
                                                                        }
                                                                        else if (diffRsCsdiff >= 10) {
                                                                            if (ssDCval < 127) {
                                                                                if (negBalance < 61) {
                                                                                    if (mu_mv_mag_sq < 7) {
                                                                                        return 4;
                                                                                    }
                                                                                    else if (mu_mv_mag_sq >= 7) {
                                                                                        return 1;
                                                                                    }
                                                                                }
                                                                                else if (negBalance >= 61) {
                                                                                    return 8;
                                                                                }
                                                                            }
                                                                            else if (ssDCval >= 127) {
                                                                                return 1;
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                                else if (tcor >= 84) {
                                                                    if (TSC < 11) {
                                                                        return 8;
                                                                    }
                                                                    else if (TSC >= 11) {
                                                                        return 4;
                                                                    }
                                                                }
                                                            }
                                                            else if (refDCval >= 136) {
                                                                if (Cs < 2621) {
                                                                    return 4;
                                                                }
                                                                else if (Cs >= 2621) {
                                                                    return 1;
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (mu_mv_mag_sq >= 14) {
                                                        if (refDCval < 61) {
                                                            return 8;
                                                        }
                                                        else if (refDCval >= 61) {
                                                            if (SC < 5027) {
                                                                return 1;
                                                            }
                                                            else if (SC >= 5027) {
                                                                if (mcTcor < 91) {
                                                                    if (CsDiff < 12) {
                                                                        if (Cs < 2443) {
                                                                            return 8;
                                                                        }
                                                                        else if (Cs >= 2443) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                    else if (CsDiff >= 12) {
                                                                        if (gchDC < 1) {
                                                                            if (Rs < 2798) {
                                                                                return 1;
                                                                            }
                                                                            else if (Rs >= 2798) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                        else if (gchDC >= 1) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                }
                                                                else if (mcTcor >= 91) {
                                                                    if (Cs < 2479) {
                                                                        return 2;
                                                                    }
                                                                    else if (Cs >= 2479) {
                                                                        return 1;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                            else if (Rs >= 3015) {
                                                if (AFD < 30) {
                                                    if (tcor < 68) {
                                                        if (posBalance < 64) {
                                                            return 1;
                                                        }
                                                        else if (posBalance >= 64) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (tcor >= 68) {
                                                        if (mcTcor < 91) {
                                                            if (refDCval < 140) {
                                                                if (SC < 7381) {
                                                                    if (RsDiff < 12) {
                                                                        if (SC < 6197) {
                                                                            if (ssDCval < 108) {
                                                                                return 1;
                                                                            }
                                                                            else if (ssDCval >= 108) {
                                                                                return 2;
                                                                            }
                                                                        }
                                                                        else if (SC >= 6197) {
                                                                            if (Rs < 3847) {
                                                                                return 4;
                                                                            }
                                                                            else if (Rs >= 3847) {
                                                                                return 1;
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (RsDiff >= 12) {
                                                                        if (AFD < 16) {
                                                                            if (Rs < 3225) {
                                                                                return 8;
                                                                            }
                                                                            else if (Rs >= 3225) {
                                                                                if (Cs < 2770) {
                                                                                    return 1;
                                                                                }
                                                                                else if (Cs >= 2770) {
                                                                                    return 2;
                                                                                }
                                                                            }
                                                                        }
                                                                        else if (AFD >= 16) {
                                                                            if (mu_mv_mag_sq < 18) {
                                                                                return 1;
                                                                            }
                                                                            else if (mu_mv_mag_sq >= 18) {
                                                                                if (mu_mv_mag_sq < 20) {
                                                                                    return 8;
                                                                                }
                                                                                else if (mu_mv_mag_sq >= 20) {
                                                                                    if (mu_mv_mag_sq < 31) {
                                                                                        if (SC < 6140) {
                                                                                            return 1;
                                                                                        }
                                                                                        else if (SC >= 6140) {
                                                                                            return 4;
                                                                                        }
                                                                                    }
                                                                                    else if (mu_mv_mag_sq >= 31) {
                                                                                        return 1;
                                                                                    }
                                                                                }
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                                else if (SC >= 7381) {
                                                                    if (SC < 7683) {
                                                                        return 8;
                                                                    }
                                                                    else if (SC >= 7683) {
                                                                        return 4;
                                                                    }
                                                                }
                                                            }
                                                            else if (refDCval >= 140) {
                                                                if (Rs < 4673) {
                                                                    if (Cs < 2686) {
                                                                        return 4;
                                                                    }
                                                                    else if (Cs >= 2686) {
                                                                        return 2;
                                                                    }
                                                                }
                                                                else if (Rs >= 4673) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                        else if (mcTcor >= 91) {
                                                            if (Rs < 3786) {
                                                                if (Rs < 3258) {
                                                                    if (AFD < 15) {
                                                                        return 8;
                                                                    }
                                                                    else if (AFD >= 15) {
                                                                        return 1;
                                                                    }
                                                                }
                                                                else if (Rs >= 3258) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (Rs >= 3786) {
                                                                if (CsDiff < 11) {
                                                                    return 4;
                                                                }
                                                                else if (CsDiff >= 11) {
                                                                    if (AFD < 19) {
                                                                        return 8;
                                                                    }
                                                                    else if (AFD >= 19) {
                                                                        return 2;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (AFD >= 30) {
                                                    if (Scindex < 6) {
                                                        return 1;
                                                    }
                                                    else if (Scindex >= 6) {
                                                        return 2;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (mcTcor >= 94) {
                                    if (tcor < 90) {
                                        if (posBalance < 58) {
                                            if (mu_mv_mag_sq < 5) {
                                                if (negBalance < 51) {
                                                    return 8;
                                                }
                                                else if (negBalance >= 51) {
                                                    if (Rs < 2564) {
                                                        return 2;
                                                    }
                                                    else if (Rs >= 2564) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                            else if (mu_mv_mag_sq >= 5) {
                                                if (RsCsDiff < 489) {
                                                    if (diffMVdiffVal < -1) {
                                                        if (ssDCval < 94) {
                                                            return 1;
                                                        }
                                                        else if (ssDCval >= 94) {
                                                            return 2;
                                                        }
                                                    }
                                                    else if (diffMVdiffVal >= -1) {
                                                        return 2;
                                                    }
                                                }
                                                else if (RsCsDiff >= 489) {
                                                    if (RsDiff < 20) {
                                                        return 8;
                                                    }
                                                    else if (RsDiff >= 20) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                        }
                                        else if (posBalance >= 58) {
                                            if (diffAFD < 5) {
                                                return 2;
                                            }
                                            else if (diffAFD >= 5) {
                                                return 4;
                                            }
                                        }
                                    }
                                    else if (tcor >= 90) {
                                        if (SC < 4080) {
                                            if (negBalance < 45) {
                                                return 8;
                                            }
                                            else if (negBalance >= 45) {
                                                if (tcor < 97) {
                                                    return 2;
                                                }
                                                else if (tcor >= 97) {
                                                    if (Rs < 1133) {
                                                        return 2;
                                                    }
                                                    else if (Rs >= 1133) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                        }
                                        else if (SC >= 4080) {
                                            if (posBalance < 61) {
                                                if (Cs < 2224) {
                                                    if (MVDiff < 26) {
                                                        if (diffMVdiffVal < -4) {
                                                            if (AFD < 13) {
                                                                return 2;
                                                            }
                                                            else if (AFD >= 13) {
                                                                return 1;
                                                            }
                                                        }
                                                        else if (diffMVdiffVal >= -4) {
                                                            return 1;
                                                        }
                                                    }
                                                    else if (MVDiff >= 26) {
                                                        return 2;
                                                    }
                                                }
                                                else if (Cs >= 2224) {
                                                    if (Cs < 2894) {
                                                        return 8;
                                                    }
                                                    else if (Cs >= 2894) {
                                                        return 2;
                                                    }
                                                }
                                            }
                                            else if (posBalance >= 61) {
                                                if (diffRsCsdiff < -15) {
                                                    if (Cs < 2276) {
                                                        return 4;
                                                    }
                                                    else if (Cs >= 2276) {
                                                        return 2;
                                                    }
                                                }
                                                else if (diffRsCsdiff >= -15) {
                                                    return 4;
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                            else if (mu_mv_mag_sq >= 47) {
                                if (ssDCval < 110) {
                                    if (mcTcor < 89) {
                                        if (posBalance < 58) {
                                            if (Rs < 4176) {
                                                if (RsCsDiff < 124) {
                                                    if (negBalance < 53) {
                                                        return 8;
                                                    }
                                                    else if (negBalance >= 53) {
                                                        if (diffRsCsdiff < 17) {
                                                            return 8;
                                                        }
                                                        else if (diffRsCsdiff >= 17) {
                                                            return 2;
                                                        }
                                                    }
                                                }
                                                else if (RsCsDiff >= 124) {
                                                    if (tcor < 86) {
                                                        if (negBalance < 45) {
                                                            if (SC < 6894) {
                                                                return 1;
                                                            }
                                                            else if (SC >= 6894) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (negBalance >= 45) {
                                                            if (AFD < 17) {
                                                                return 8;
                                                            }
                                                            else if (AFD >= 17) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                    else if (tcor >= 86) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                            else if (Rs >= 4176) {
                                                if (RsDiff < 10) {
                                                    return 1;
                                                }
                                                else if (RsDiff >= 10) {
                                                    if (tcor < 81) {
                                                        if (AFD < 14) {
                                                            return 4;
                                                        }
                                                        else if (AFD >= 14) {
                                                            if (Cs < 3077) {
                                                                return 2;
                                                            }
                                                            else if (Cs >= 3077) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                    else if (tcor >= 81) {
                                                        return 2;
                                                    }
                                                }
                                            }
                                        }
                                        else if (posBalance >= 58) {
                                            if (RsCsDiff < 92) {
                                                if (refDCval < 93) {
                                                    if (Rs < 1481) {
                                                        return 2;
                                                    }
                                                    else if (Rs >= 1481) {
                                                        if (mu_mv_mag_sq < 55) {
                                                            if (Rs < 2055) {
                                                                return 8;
                                                            }
                                                            else if (Rs >= 2055) {
                                                                return 1;
                                                            }
                                                        }
                                                        else if (mu_mv_mag_sq >= 55) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                                else if (refDCval >= 93) {
                                                    if (refDCval < 103) {
                                                        if (RsDiff < 6) {
                                                            if (mcTcor < 86) {
                                                                return 8;
                                                            }
                                                            else if (mcTcor >= 86) {
                                                                return 2;
                                                            }
                                                        }
                                                        else if (RsDiff >= 6) {
                                                            return 2;
                                                        }
                                                    }
                                                    else if (refDCval >= 103) {
                                                        if (Cs < 2217) {
                                                            return 4;
                                                        }
                                                        else if (Cs >= 2217) {
                                                            return 1;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (RsCsDiff >= 92) {
                                                if (ssDCval < 93) {
                                                    if (TSC < 13) {
                                                        if (RsDiff < 10) {
                                                            if (Rs < 2279) {
                                                                return 8;
                                                            }
                                                            else if (Rs >= 2279) {
                                                                return 1;
                                                            }
                                                        }
                                                        else if (RsDiff >= 10) {
                                                            if (refDCval < 66) {
                                                                return 8;
                                                            }
                                                            else if (refDCval >= 66) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                    else if (TSC >= 13) {
                                                        return 8;
                                                    }
                                                }
                                                else if (ssDCval >= 93) {
                                                    if (MVDiff < 10) {
                                                        if (tcor < 86) {
                                                            return 4;
                                                        }
                                                        else if (tcor >= 86) {
                                                            return 2;
                                                        }
                                                    }
                                                    else if (MVDiff >= 10) {
                                                        if (Rs < 3013) {
                                                            return 4;
                                                        }
                                                        else if (Rs >= 3013) {
                                                            return 2;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (mcTcor >= 89) {
                                        if (SC < 5214) {
                                            if (mu_mv_mag_sq < 81) {
                                                if (Rs < 2455) {
                                                    if (mu_mv_mag_sq < 61) {
                                                        if (refDCval < 80) {
                                                            if (diffRsCsdiff < -28) {
                                                                if (AFD < 16) {
                                                                    return 8;
                                                                }
                                                                else if (AFD >= 16) {
                                                                    return 1;
                                                                }
                                                            }
                                                            else if (diffRsCsdiff >= -28) {
                                                                return 1;
                                                            }
                                                        }
                                                        else if (refDCval >= 80) {
                                                            if (negBalance < 47) {
                                                                if (MVDiff < 19) {
                                                                    if (diffMVdiffVal < -2) {
                                                                        if (CsDiff < 8) {
                                                                            return 1;
                                                                        }
                                                                        else if (CsDiff >= 8) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                    else if (diffMVdiffVal >= -2) {
                                                                        if (Rs < 2446) {
                                                                            return 8;
                                                                        }
                                                                        else if (Rs >= 2446) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                }
                                                                else if (MVDiff >= 19) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (negBalance >= 47) {
                                                                if (Rs < 1788) {
                                                                    return 8;
                                                                }
                                                                else if (Rs >= 1788) {
                                                                    return 2;
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (mu_mv_mag_sq >= 61) {
                                                        if (mcTcor < 93) {
                                                            if (ssDCval < 91) {
                                                                return 8;
                                                            }
                                                            else if (ssDCval >= 91) {
                                                                return 2;
                                                            }
                                                        }
                                                        else if (mcTcor >= 93) {
                                                            if (posBalance < 48) {
                                                                return 4;
                                                            }
                                                            else if (posBalance >= 48) {
                                                                if (refDCval < 82) {
                                                                    return 4;
                                                                }
                                                                else if (refDCval >= 82) {
                                                                    return 2;
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (Rs >= 2455) {
                                                    if (refDCval < 100) {
                                                        if (Rs < 2635) {
                                                            if (ssDCval < 91) {
                                                                if (refDCval < 85) {
                                                                    if (AFD < 10) {
                                                                        return 4;
                                                                    }
                                                                    else if (AFD >= 10) {
                                                                        return 1;
                                                                    }
                                                                }
                                                                else if (refDCval >= 85) {
                                                                    if (posBalance < 55) {
                                                                        return 1;
                                                                    }
                                                                    else if (posBalance >= 55) {
                                                                        if (CsDiff < 5) {
                                                                            return 1;
                                                                        }
                                                                        else if (CsDiff >= 5) {
                                                                            return 2;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                            else if (ssDCval >= 91) {
                                                                return 1;
                                                            }
                                                        }
                                                        else if (Rs >= 2635) {
                                                            if (negBalance < 49) {
                                                                return 1;
                                                            }
                                                            else if (negBalance >= 49) {
                                                                if (AFD < 10) {
                                                                    return 4;
                                                                }
                                                                else if (AFD >= 10) {
                                                                    if (Cs < 2267) {
                                                                        if (TSC < 12) {
                                                                            return 1;
                                                                        }
                                                                        else if (TSC >= 12) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                    else if (Cs >= 2267) {
                                                                        if (SC < 5171) {
                                                                            return 4;
                                                                        }
                                                                        else if (SC >= 5171) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (refDCval >= 100) {
                                                        if (Cs < 2266) {
                                                            if (gchDC < 1) {
                                                                return 8;
                                                            }
                                                            else if (gchDC >= 1) {
                                                                return 2;
                                                            }
                                                        }
                                                        else if (Cs >= 2266) {
                                                            if (Cs < 2394) {
                                                                return 1;
                                                            }
                                                            else if (Cs >= 2394) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                            else if (mu_mv_mag_sq >= 81) {
                                                if (mu_mv_mag_sq < 98) {
                                                    if (tcor < 93) {
                                                        if (ssDCval < 80) {
                                                            if (Rs < 2286) {
                                                                return 8;
                                                            }
                                                            else if (Rs >= 2286) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (ssDCval >= 80) {
                                                            if (posBalance < 61) {
                                                                if (mu_mv_mag_sq < 94) {
                                                                    return 2;
                                                                }
                                                                else if (mu_mv_mag_sq >= 94) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (posBalance >= 61) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                    else if (tcor >= 93) {
                                                        if (CsDiff < 7) {
                                                            if (ssDCval < 86) {
                                                                return 4;
                                                            }
                                                            else if (ssDCval >= 86) {
                                                                if (ssDCval < 94) {
                                                                    return 2;
                                                                }
                                                                else if (ssDCval >= 94) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                        else if (CsDiff >= 7) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                                else if (mu_mv_mag_sq >= 98) {
                                                    if (negBalance < 40) {
                                                        if (posBalance < 47) {
                                                            return 1;
                                                        }
                                                        else if (posBalance >= 47) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (negBalance >= 40) {
                                                        if (MVDiff < 3) {
                                                            if (diffTSC < 1) {
                                                                return 2;
                                                            }
                                                            else if (diffTSC >= 1) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (MVDiff >= 3) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (SC >= 5214) {
                                            if (mu_mv_mag_sq < 69) {
                                                if (MVDiff < 24) {
                                                    if (Rs < 3409) {
                                                        if (SC < 5391) {
                                                            if (negBalance < 47) {
                                                                return 1;
                                                            }
                                                            else if (negBalance >= 47) {
                                                                if (ssDCval < 92) {
                                                                    return 1;
                                                                }
                                                                else if (ssDCval >= 92) {
                                                                    if (Rs < 2963) {
                                                                        return 4;
                                                                    }
                                                                    else if (Rs >= 2963) {
                                                                        return 1;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (SC >= 5391) {
                                                            if (ssDCval < 97) {
                                                                if (ssDCval < 87) {
                                                                    if (Cs < 3135) {
                                                                        if (mcTcor < 92) {
                                                                            return 1;
                                                                        }
                                                                        else if (mcTcor >= 92) {
                                                                            if (posBalance < 59) {
                                                                                if (Rs < 3047) {
                                                                                    return 8;
                                                                                }
                                                                                else if (Rs >= 3047) {
                                                                                    return 1;
                                                                                }
                                                                            }
                                                                            else if (posBalance >= 59) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (Cs >= 3135) {
                                                                        return 8;
                                                                    }
                                                                }
                                                                else if (ssDCval >= 87) {
                                                                    return 1;
                                                                }
                                                            }
                                                            else if (ssDCval >= 97) {
                                                                if (AFD < 13) {
                                                                    return 8;
                                                                }
                                                                else if (AFD >= 13) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (Rs >= 3409) {
                                                        if (Rs < 3582) {
                                                            return 8;
                                                        }
                                                        else if (Rs >= 3582) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                                else if (MVDiff >= 24) {
                                                    if (diffAFD < 3) {
                                                        if (mu_mv_mag_sq < 61) {
                                                            if (refDCval < 53) {
                                                                return 2;
                                                            }
                                                            else if (refDCval >= 53) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (mu_mv_mag_sq >= 61) {
                                                            return 2;
                                                        }
                                                    }
                                                    else if (diffAFD >= 3) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                            else if (mu_mv_mag_sq >= 69) {
                                                if (refDCval < 106) {
                                                    if (MVDiff < 1) {
                                                        if (diffMVdiffVal < -1) {
                                                            if (Rs < 4042) {
                                                                return 8;
                                                            }
                                                            else if (Rs >= 4042) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (diffMVdiffVal >= -1) {
                                                            if (Cs < 2767) {
                                                                return 8;
                                                            }
                                                            else if (Cs >= 2767) {
                                                                return 2;
                                                            }
                                                        }
                                                    }
                                                    else if (MVDiff >= 1) {
                                                        if (negBalance < 49) {
                                                            if (SC < 5608) {
                                                                if (AFD < 12) {
                                                                    return 8;
                                                                }
                                                                else if (AFD >= 12) {
                                                                    return 1;
                                                                }
                                                            }
                                                            else if (SC >= 5608) {
                                                                if (negBalance < 44) {
                                                                    if (Rs < 3398) {
                                                                        return 8;
                                                                    }
                                                                    else if (Rs >= 3398) {
                                                                        if (Cs < 2801) {
                                                                            return 2;
                                                                        }
                                                                        else if (Cs >= 2801) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                }
                                                                else if (negBalance >= 44) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                        else if (negBalance >= 49) {
                                                            if (TSC < 14) {
                                                                if (diffRsCsdiff < -9) {
                                                                    return 4;
                                                                }
                                                                else if (diffRsCsdiff >= -9) {
                                                                    if (TSC < 10) {
                                                                        return 4;
                                                                    }
                                                                    else if (TSC >= 10) {
                                                                        if (TSC < 11) {
                                                                            if (SC < 5502) {
                                                                                return 4;
                                                                            }
                                                                            else if (SC >= 5502) {
                                                                                return 8;
                                                                            }
                                                                        }
                                                                        else if (TSC >= 11) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                            else if (TSC >= 14) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (refDCval >= 106) {
                                                    if (MVDiff < 22) {
                                                        return 8;
                                                    }
                                                    else if (MVDiff >= 22) {
                                                        return 2;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (ssDCval >= 110) {
                                    if (tcor < 78) {
                                        if (gchDC < 2) {
                                            if (ssDCval < 124) {
                                                if (Rs < 3241) {
                                                    return 4;
                                                }
                                                else if (Rs >= 3241) {
                                                    return 1;
                                                }
                                            }
                                            else if (ssDCval >= 124) {
                                                if (refDCval < 140) {
                                                    return 1;
                                                }
                                                else if (refDCval >= 140) {
                                                    if (AFD < 20) {
                                                        return 8;
                                                    }
                                                    else if (AFD >= 20) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                        }
                                        else if (gchDC >= 2) {
                                            if (SC < 5702) {
                                                return 2;
                                            }
                                            else if (SC >= 5702) {
                                                if (TSC < 12) {
                                                    return 4;
                                                }
                                                else if (TSC >= 12) {
                                                    return 8;
                                                }
                                            }
                                        }
                                    }
                                    else if (tcor >= 78) {
                                        if (posBalance < 51) {
                                            if (AFD < 10) {
                                                if (Rs < 3325) {
                                                    return 8;
                                                }
                                                else if (Rs >= 3325) {
                                                    if (MVDiff < 5) {
                                                        return 8;
                                                    }
                                                    else if (MVDiff >= 5) {
                                                        return 2;
                                                    }
                                                }
                                            }
                                            else if (AFD >= 10) {
                                                if (mu_mv_mag_sq < 79) {
                                                    if (mcTcor < 87) {
                                                        return 2;
                                                    }
                                                    else if (mcTcor >= 87) {
                                                        if (diffMVdiffVal < 4) {
                                                            if (MVDiff < 12) {
                                                                if (RsDiff < 8) {
                                                                    return 4;
                                                                }
                                                                else if (RsDiff >= 8) {
                                                                    if (Cs < 2488) {
                                                                        return 4;
                                                                    }
                                                                    else if (Cs >= 2488) {
                                                                        if (refDCval < 124) {
                                                                            return 8;
                                                                        }
                                                                        else if (refDCval >= 124) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                            else if (MVDiff >= 12) {
                                                                if (refDCval < 121) {
                                                                    return 4;
                                                                }
                                                                else if (refDCval >= 121) {
                                                                    if (Cs < 2727) {
                                                                        return 2;
                                                                    }
                                                                    else if (Cs >= 2727) {
                                                                        return 8;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (diffMVdiffVal >= 4) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                                else if (mu_mv_mag_sq >= 79) {
                                                    return 8;
                                                }
                                            }
                                        }
                                        else if (posBalance >= 51) {
                                            if (ssDCval < 127) {
                                                if (AFD < 12) {
                                                    if (negBalance < 51) {
                                                        if (Rs < 3812) {
                                                            if (posBalance < 60) {
                                                                if (diffAFD < -5) {
                                                                    return 4;
                                                                }
                                                                else if (diffAFD >= -5) {
                                                                    if (ssDCval < 121) {
                                                                        return 8;
                                                                    }
                                                                    else if (ssDCval >= 121) {
                                                                        if (Rs < 3065) {
                                                                            return 8;
                                                                        }
                                                                        else if (Rs >= 3065) {
                                                                            if (Cs < 2558) {
                                                                                return 4;
                                                                            }
                                                                            else if (Cs >= 2558) {
                                                                                return 8;
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                            else if (posBalance >= 60) {
                                                                if (ssDCval < 114) {
                                                                    return 8;
                                                                }
                                                                else if (ssDCval >= 114) {
                                                                    if (ssDCval < 117) {
                                                                        return 4;
                                                                    }
                                                                    else if (ssDCval >= 117) {
                                                                        if (Rs < 2581) {
                                                                            return 4;
                                                                        }
                                                                        else if (Rs >= 2581) {
                                                                            if (Cs < 2122) {
                                                                                return 8;
                                                                            }
                                                                            else if (Cs >= 2122) {
                                                                                if (ssDCval < 121) {
                                                                                    return 8;
                                                                                }
                                                                                else if (ssDCval >= 121) {
                                                                                    return 4;
                                                                                }
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (Rs >= 3812) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (negBalance >= 51) {
                                                        if (mu_mv_mag_sq < 58) {
                                                            return 4;
                                                        }
                                                        else if (mu_mv_mag_sq >= 58) {
                                                            if (RsCsDiff < 110) {
                                                                if (Rs < 2813) {
                                                                    if (Rs < 2483) {
                                                                        if (Cs < 2189) {
                                                                            return 4;
                                                                        }
                                                                        else if (Cs >= 2189) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                    else if (Rs >= 2483) {
                                                                        if (tcor < 85) {
                                                                            if (Cs < 2030) {
                                                                                return 8;
                                                                            }
                                                                            else if (Cs >= 2030) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                        else if (tcor >= 85) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                }
                                                                else if (Rs >= 2813) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (RsCsDiff >= 110) {
                                                                if (Cs < 1857) {
                                                                    return 8;
                                                                }
                                                                else if (Cs >= 1857) {
                                                                    if (Cs < 2227) {
                                                                        return 4;
                                                                    }
                                                                    else if (Cs >= 2227) {
                                                                        return 8;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (AFD >= 12) {
                                                    if (Cs < 2935) {
                                                        if (Cs < 2054) {
                                                            if (Cs < 1866) {
                                                                return 4;
                                                            }
                                                            else if (Cs >= 1866) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (Cs >= 2054) {
                                                            if (Rs < 3788) {
                                                                return 1;
                                                            }
                                                            else if (Rs >= 3788) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                    else if (Cs >= 2935) {
                                                        if (Rs < 4018) {
                                                            return 8;
                                                        }
                                                        else if (Rs >= 4018) {
                                                            return 2;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (ssDCval >= 127) {
                                                if (TSC < 10) {
                                                    return 1;
                                                }
                                                else if (TSC >= 10) {
                                                    return 8;
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                        }
                    }
                }
                else if (mu_mv_mag_sq >= 108) {
                    if (TSCindex < 8) {
                        if (Cs < 1676) {
                            if (TSCindex < 7) {
                                if (mu_mv_mag_sq < 149) {
                                    if (Rs < 1791) {
                                        if (SC < 1279) {
                                            if (CsDiff < 2) {
                                                return 4;
                                            }
                                            else if (CsDiff >= 2) {
                                                return 8;
                                            }
                                        }
                                        else if (SC >= 1279) {
                                            if (refDCval < 112) {
                                                return 8;
                                            }
                                            else if (refDCval >= 112) {
                                                if (Rs < 1369) {
                                                    return 4;
                                                }
                                                else if (Rs >= 1369) {
                                                    return 8;
                                                }
                                            }
                                        }
                                    }
                                    else if (Rs >= 1791) {
                                        if (refDCval < 80) {
                                            if (tcor < 96) {
                                                if (mu_mv_mag_sq < 124) {
                                                    if (mu_mv_mag_sq < 116) {
                                                        if (Cs < 1614) {
                                                            return 8;
                                                        }
                                                        else if (Cs >= 1614) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (mu_mv_mag_sq >= 116) {
                                                        return 8;
                                                    }
                                                }
                                                else if (mu_mv_mag_sq >= 124) {
                                                    if (refDCval < 74) {
                                                        return 2;
                                                    }
                                                    else if (refDCval >= 74) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                            else if (tcor >= 96) {
                                                if (SC < 3396) {
                                                    if (refDCval < 73) {
                                                        return 8;
                                                    }
                                                    else if (refDCval >= 73) {
                                                        if (SC < 3233) {
                                                            return 4;
                                                        }
                                                        else if (SC >= 3233) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                                else if (SC >= 3396) {
                                                    if (RsDiff < 4) {
                                                        if (CsDiff < 4) {
                                                            return 2;
                                                        }
                                                        else if (CsDiff >= 4) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (RsDiff >= 4) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                        }
                                        else if (refDCval >= 80) {
                                            return 8;
                                        }
                                    }
                                }
                                else if (mu_mv_mag_sq >= 149) {
                                    if (Cs < 235) {
                                        return 2;
                                    }
                                    else if (Cs >= 235) {
                                        return 1;
                                    }
                                }
                            }
                            else if (TSCindex >= 7) {
                                if (mcTcor < 94) {
                                    if (Rs < 821) {
                                        return 1;
                                    }
                                    else if (Rs >= 821) {
                                        if (Cs < 1329) {
                                            if (Cs < 1062) {
                                                return 8;
                                            }
                                            else if (Cs >= 1062) {
                                                return 2;
                                            }
                                        }
                                        else if (Cs >= 1329) {
                                            if (tcor < 92) {
                                                return 8;
                                            }
                                            else if (tcor >= 92) {
                                                if (mu_mv_mag_sq < 116) {
                                                    return 2;
                                                }
                                                else if (mu_mv_mag_sq >= 116) {
                                                    return 8;
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (mcTcor >= 94) {
                                    if (RsCsDiff < 22) {
                                        if (SC < 1690) {
                                            return 4;
                                        }
                                        else if (SC >= 1690) {
                                            if (ssDCval < 75) {
                                                return 2;
                                            }
                                            else if (ssDCval >= 75) {
                                                return 4;
                                            }
                                        }
                                    }
                                    else if (RsCsDiff >= 22) {
                                        if (Rs < 1838) {
                                            if (AFD < 7) {
                                                if (gchDC < 1) {
                                                    return 8;
                                                }
                                                else if (gchDC >= 1) {
                                                    if (Cs < 1544) {
                                                        return 8;
                                                    }
                                                    else if (Cs >= 1544) {
                                                        return 2;
                                                    }
                                                }
                                            }
                                            else if (AFD >= 7) {
                                                return 2;
                                            }
                                        }
                                        else if (Rs >= 1838) {
                                            if (refDCval < 75) {
                                                if (Rs < 2038) {
                                                    return 2;
                                                }
                                                else if (Rs >= 2038) {
                                                    return 4;
                                                }
                                            }
                                            else if (refDCval >= 75) {
                                                if (Rs < 2084) {
                                                    if (CsDiff < 5) {
                                                        return 4;
                                                    }
                                                    else if (CsDiff >= 5) {
                                                        return 8;
                                                    }
                                                }
                                                else if (Rs >= 2084) {
                                                    return 8;
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                        }
                        else if (Cs >= 1676) {
                            if (refDCval < 88) {
                                if (tcor < 95) {
                                    if (SC < 3901) {
                                        return 8;
                                    }
                                    else if (SC >= 3901) {
                                        if (tcor < 94) {
                                            if (AFD < 7) {
                                                return 8;
                                            }
                                            else if (AFD >= 7) {
                                                if (posBalance < 56) {
                                                    return 4;
                                                }
                                                else if (posBalance >= 56) {
                                                    return 8;
                                                }
                                            }
                                        }
                                        else if (tcor >= 94) {
                                            if (mu_mv_mag_sq < 114) {
                                                return 4;
                                            }
                                            else if (mu_mv_mag_sq >= 114) {
                                                if (Cs < 2398) {
                                                    return 8;
                                                }
                                                else if (Cs >= 2398) {
                                                    return 4;
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (tcor >= 95) {
                                    if (RsCsDiff < 16) {
                                        return 8;
                                    }
                                    else if (RsCsDiff >= 16) {
                                        if (MVDiff < 1) {
                                            if (SC < 3530) {
                                                if (negBalance < 40) {
                                                    return 4;
                                                }
                                                else if (negBalance >= 40) {
                                                    return 8;
                                                }
                                            }
                                            else if (SC >= 3530) {
                                                if (Cs < 3046) {
                                                    return 4;
                                                }
                                                else if (Cs >= 3046) {
                                                    return 8;
                                                }
                                            }
                                        }
                                        else if (MVDiff >= 1) {
                                            if (diffTSC < -2) {
                                                if (Rs < 2741) {
                                                    return 8;
                                                }
                                                else if (Rs >= 2741) {
                                                    return 4;
                                                }
                                            }
                                            else if (diffTSC >= -2) {
                                                return 8;
                                            }
                                        }
                                    }
                                }
                            }
                            else if (refDCval >= 88) {
                                if (posBalance < 58) {
                                    if (Rs < 1777) {
                                        if (Rs < 1649) {
                                            return 8;
                                        }
                                        else if (Rs >= 1649) {
                                            return 4;
                                        }
                                    }
                                    else if (Rs >= 1777) {
                                        if (negBalance < 48) {
                                            if (mu_mv_mag_sq < 123) {
                                                if (SC < 4630) {
                                                    if (Rs < 2173) {
                                                        return 8;
                                                    }
                                                    else if (Rs >= 2173) {
                                                        return 4;
                                                    }
                                                }
                                                else if (SC >= 4630) {
                                                    if (Cs < 3031) {
                                                        return 8;
                                                    }
                                                    else if (Cs >= 3031) {
                                                        if (Cs < 3085) {
                                                            if (Cs < 3069) {
                                                                if (Cs < 3047) {
                                                                    return 4;
                                                                }
                                                                else if (Cs >= 3047) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (Cs >= 3069) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (Cs >= 3085) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (mu_mv_mag_sq >= 123) {
                                                return 8;
                                            }
                                        }
                                        else if (negBalance >= 48) {
                                            if (Cs < 2943) {
                                                if (mu_mv_mag_sq < 118) {
                                                    if (RsDiff < 6) {
                                                        return 4;
                                                    }
                                                    else if (RsDiff >= 6) {
                                                        return 8;
                                                    }
                                                }
                                                else if (mu_mv_mag_sq >= 118) {
                                                    if (mu_mv_mag_sq < 120) {
                                                        if (Rs < 2650) {
                                                            return 8;
                                                        }
                                                        else if (Rs >= 2650) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (mu_mv_mag_sq >= 120) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                            else if (Cs >= 2943) {
                                                if (Rs < 2879) {
                                                    if (refDCval < 112) {
                                                        return 8;
                                                    }
                                                    else if (refDCval >= 112) {
                                                        if (Rs < 2698) {
                                                            return 8;
                                                        }
                                                        else if (Rs >= 2698) {
                                                            if (diffRsCsdiff < -8) {
                                                                return 4;
                                                            }
                                                            else if (diffRsCsdiff >= -8) {
                                                                if (RsCsDiff < 37) {
                                                                    if (Cs < 2958) {
                                                                        return 4;
                                                                    }
                                                                    else if (Cs >= 2958) {
                                                                        return 8;
                                                                    }
                                                                }
                                                                else if (RsCsDiff >= 37) {
                                                                    if (mu_mv_mag_sq < 123) {
                                                                        return 4;
                                                                    }
                                                                    else if (mu_mv_mag_sq >= 123) {
                                                                        return 8;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (Rs >= 2879) {
                                                    if (Scindex < 6) {
                                                        if (Rs < 2967) {
                                                            return 8;
                                                        }
                                                        else if (Rs >= 2967) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (Scindex >= 6) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (posBalance >= 58) {
                                    if (mu_mv_mag_sq < 123) {
                                        if (negBalance < 47) {
                                            if (diffRsCsdiff < 5) {
                                                return 8;
                                            }
                                            else if (diffRsCsdiff >= 5) {
                                                if (Cs < 2997) {
                                                    if (SC < 5722) {
                                                        return 8;
                                                    }
                                                    else if (SC >= 5722) {
                                                        return 4;
                                                    }
                                                }
                                                else if (Cs >= 2997) {
                                                    return 8;
                                                }
                                            }
                                        }
                                        else if (negBalance >= 47) {
                                            if (Cs < 3163) {
                                                if (SC < 5848) {
                                                    if (Rs < 2428) {
                                                        if (SC < 4445) {
                                                            return 8;
                                                        }
                                                        else if (SC >= 4445) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (Rs >= 2428) {
                                                        if (Cs < 2867) {
                                                            return 8;
                                                        }
                                                        else if (Cs >= 2867) {
                                                            if (Cs < 2882) {
                                                                return 4;
                                                            }
                                                            else if (Cs >= 2882) {
                                                                if (Cs < 3002) {
                                                                    return 8;
                                                                }
                                                                else if (Cs >= 3002) {
                                                                    if (Rs < 2702) {
                                                                        return 8;
                                                                    }
                                                                    else if (Rs >= 2702) {
                                                                        return 4;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (SC >= 5848) {
                                                    if (mu_mv_mag_sq < 122) {
                                                        return 8;
                                                    }
                                                    else if (mu_mv_mag_sq >= 122) {
                                                        if (Cs < 3031) {
                                                            return 8;
                                                        }
                                                        else if (Cs >= 3031) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (Cs >= 3163) {
                                                return 8;
                                            }
                                        }
                                    }
                                    else if (mu_mv_mag_sq >= 123) {
                                        return 8;
                                    }
                                }
                            }
                        }
                    }
                    else if (TSCindex >= 8) {
                        if (Rs < 4601) {
                            if (SC < 3672) {
                                if (refDCval < 116) {
                                    if (RsCsDiff < 46) {
                                        if (diffAFD < -1) {
                                            if (Cs < 1493) {
                                                return 8;
                                            }
                                            else if (Cs >= 1493) {
                                                return 4;
                                            }
                                        }
                                        else if (diffAFD >= -1) {
                                            if (negBalance < 43) {
                                                return 8;
                                            }
                                            else if (negBalance >= 43) {
                                                return 2;
                                            }
                                        }
                                    }
                                    else if (RsCsDiff >= 46) {
                                        if (Cs < 1217) {
                                            if (AFD < 31) {
                                                if (Rs < 1256) {
                                                    return 2;
                                                }
                                                else if (Rs >= 1256) {
                                                    return 1;
                                                }
                                            }
                                            else if (AFD >= 31) {
                                                return 2;
                                            }
                                        }
                                        else if (Cs >= 1217) {
                                            if (SC < 3541) {
                                                return 8;
                                            }
                                            else if (SC >= 3541) {
                                                return 2;
                                            }
                                        }
                                    }
                                }
                                else if (refDCval >= 116) {
                                    if (TSC < 18) {
                                        if (MVDiff < 32) {
                                            return 8;
                                        }
                                        else if (MVDiff >= 32) {
                                            return 1;
                                        }
                                    }
                                    else if (TSC >= 18) {
                                        if (TSC < 21) {
                                            return 8;
                                        }
                                        else if (TSC >= 21) {
                                            return 4;
                                        }
                                    }
                                }
                            }
                            else if (SC >= 3672) {
                                if (SC < 6372) {
                                    if (AFD < 29) {
                                        if (Cs < 2055) {
                                            if (posBalance < 48) {
                                                return 1;
                                            }
                                            else if (posBalance >= 48) {
                                                if (RsCsDiff < 77) {
                                                    if (Cs < 1982) {
                                                        if (posBalance < 56) {
                                                            return 4;
                                                        }
                                                        else if (posBalance >= 56) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (Cs >= 1982) {
                                                        return 4;
                                                    }
                                                }
                                                else if (RsCsDiff >= 77) {
                                                    if (Rs < 3199) {
                                                        return 8;
                                                    }
                                                    else if (Rs >= 3199) {
                                                        return 1;
                                                    }
                                                }
                                            }
                                        }
                                        else if (Cs >= 2055) {
                                            return 8;
                                        }
                                    }
                                    else if (AFD >= 29) {
                                        if (ssDCval < 135) {
                                            return 2;
                                        }
                                        else if (ssDCval >= 135) {
                                            if (SC < 4020) {
                                                return 2;
                                            }
                                            else if (SC >= 4020) {
                                                return 4;
                                            }
                                        }
                                    }
                                }
                                else if (SC >= 6372) {
                                    if (ssDCval < 93) {
                                        if (TSC < 11) {
                                            return 2;
                                        }
                                        else if (TSC >= 11) {
                                            return 8;
                                        }
                                    }
                                    else if (ssDCval >= 93) {
                                        if (Rs < 3766) {
                                            return 8;
                                        }
                                        else if (Rs >= 3766) {
                                            if (SC < 7652) {
                                                if (AFD < 10) {
                                                    return 4;
                                                }
                                                else if (AFD >= 10) {
                                                    return 8;
                                                }
                                            }
                                            else if (SC >= 7652) {
                                                return 8;
                                            }
                                        }
                                    }
                                }
                            }
                        }
                        else if (Rs >= 4601) {
                            return 4;
                        }
                    }
                }
            }
            else if (SC >= 8065) {
                if (CsDiff < 10) {
                    if (AFD < 9) {
                        if (ssDCval < 122) {
                            if (ssDCval < 113) {
                                if (ssDCval < 82) {
                                    if (Cs < 4874) {
                                        if (RsDiff < 5) {
                                            if (AFD < 7) {
                                                return 8;
                                            }
                                            else if (AFD >= 7) {
                                                return 4;
                                            }
                                        }
                                        else if (RsDiff >= 5) {
                                            if (AFD < 8) {
                                                if (diffRsCsdiff < -16) {
                                                    if (SC < 9112) {
                                                        return 8;
                                                    }
                                                    else if (SC >= 9112) {
                                                        return 4;
                                                    }
                                                }
                                                else if (diffRsCsdiff >= -16) {
                                                    return 8;
                                                }
                                            }
                                            else if (AFD >= 8) {
                                                return 4;
                                            }
                                        }
                                    }
                                    else if (Cs >= 4874) {
                                        if (diffAFD < -2) {
                                            return 8;
                                        }
                                        else if (diffAFD >= -2) {
                                            if (SC < 10074) {
                                                if (ssDCval < 80) {
                                                    if (diffRsCsdiff < 33) {
                                                        return 8;
                                                    }
                                                    else if (diffRsCsdiff >= 33) {
                                                        return 4;
                                                    }
                                                }
                                                else if (ssDCval >= 80) {
                                                    if (CsDiff < 8) {
                                                        return 8;
                                                    }
                                                    else if (CsDiff >= 8) {
                                                        if (mcTcor < 92) {
                                                            if (RsCsDiff < 137) {
                                                                if (SC < 9669) {
                                                                    return 8;
                                                                }
                                                                else if (SC >= 9669) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (RsCsDiff >= 137) {
                                                                if (negBalance < 47) {
                                                                    if (RsCsDiff < 154) {
                                                                        if (gchDC < 1) {
                                                                            if (diffTSC < 1) {
                                                                                return 8;
                                                                            }
                                                                            else if (diffTSC >= 1) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                        else if (gchDC >= 1) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                    else if (RsCsDiff >= 154) {
                                                                        return 8;
                                                                    }
                                                                }
                                                                else if (negBalance >= 47) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                        else if (mcTcor >= 92) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (SC >= 10074) {
                                                return 4;
                                            }
                                        }
                                    }
                                }
                                else if (ssDCval >= 82) {
                                    if (Cs < 2996) {
                                        return 2;
                                    }
                                    else if (Cs >= 2996) {
                                        return 8;
                                    }
                                }
                            }
                            else if (ssDCval >= 113) {
                                if (Cs < 4757) {
                                    if (MVDiff < 1) {
                                        return 4;
                                    }
                                    else if (MVDiff >= 1) {
                                        return 2;
                                    }
                                }
                                else if (Cs >= 4757) {
                                    return 8;
                                }
                            }
                        }
                        else if (ssDCval >= 122) {
                            return 8;
                        }
                    }
                    else if (AFD >= 9) {
                        if (negBalance < 53) {
                            if (ssDCval < 67) {
                                if (gchDC < 3) {
                                    return 4;
                                }
                                else if (gchDC >= 3) {
                                    if (Rs < 4659) {
                                        return 8;
                                    }
                                    else if (Rs >= 4659) {
                                        return 4;
                                    }
                                }
                            }
                            else if (ssDCval >= 67) {
                                if (ssDCval < 106) {
                                    if (Cs < 4310) {
                                        if (posBalance < 56) {
                                            return 4;
                                        }
                                        else if (posBalance >= 56) {
                                            if (Rs < 4883) {
                                                return 8;
                                            }
                                            else if (Rs >= 4883) {
                                                return 4;
                                            }
                                        }
                                    }
                                    else if (Cs >= 4310) {
                                        return 8;
                                    }
                                }
                                else if (ssDCval >= 106) {
                                    if (mcTcor < 85) {
                                        return 2;
                                    }
                                    else if (mcTcor >= 85) {
                                        if (ssDCval < 112) {
                                            if (MVDiff < 7) {
                                                if (Cs < 3110) {
                                                    if (TSC < 9) {
                                                        return 8;
                                                    }
                                                    else if (TSC >= 9) {
                                                        return 2;
                                                    }
                                                }
                                                else if (Cs >= 3110) {
                                                    return 8;
                                                }
                                            }
                                            else if (MVDiff >= 7) {
                                                if (Cs < 3113) {
                                                    return 4;
                                                }
                                                else if (Cs >= 3113) {
                                                    return 2;
                                                }
                                            }
                                        }
                                        else if (ssDCval >= 112) {
                                            if (SC < 8525) {
                                                return 8;
                                            }
                                            else if (SC >= 8525) {
                                                if (Rs < 5356) {
                                                    return 2;
                                                }
                                                else if (Rs >= 5356) {
                                                    return 8;
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                        }
                        else if (negBalance >= 53) {
                            if (Rs < 4714) {
                                if (CsDiff < 8) {
                                    return 4;
                                }
                                else if (CsDiff >= 8) {
                                    return 8;
                                }
                            }
                            else if (Rs >= 4714) {
                                return 4;
                            }
                        }
                    }
                }
                else if (CsDiff >= 10) {
                    if (mu_mv_mag_sq < 63) {
                        if (Cs < 3291) {
                            if (refDCval < 119) {
                                if (gchDC < 1) {
                                    return 2;
                                }
                                else if (gchDC >= 1) {
                                    if (Rs < 5210) {
                                        return 2;
                                    }
                                    else if (Rs >= 5210) {
                                        return 4;
                                    }
                                }
                            }
                            else if (refDCval >= 119) {
                                if (refDCval < 141) {
                                    return 1;
                                }
                                else if (refDCval >= 141) {
                                    if (ssDCval < 153) {
                                        return 2;
                                    }
                                    else if (ssDCval >= 153) {
                                        return 4;
                                    }
                                }
                            }
                        }
                        else if (Cs >= 3291) {
                            if (refDCval < 150) {
                                if (Rs < 6988) {
                                    if (CsDiff < 17) {
                                        if (AFD < 21) {
                                            if (RsCsDiff < 367) {
                                                if (Cs < 5248) {
                                                    if (AFD < 17) {
                                                        return 4;
                                                    }
                                                    else if (AFD >= 17) {
                                                        if (diffAFD < 1) {
                                                            return 4;
                                                        }
                                                        else if (diffAFD >= 1) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                                else if (Cs >= 5248) {
                                                    return 8;
                                                }
                                            }
                                            else if (RsCsDiff >= 367) {
                                                if (negBalance < 56) {
                                                    if (diffMVdiffVal < -6) {
                                                        if (Rs < 4939) {
                                                            return 4;
                                                        }
                                                        else if (Rs >= 4939) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (diffMVdiffVal >= -6) {
                                                        if (Cs < 5106) {
                                                            return 8;
                                                        }
                                                        else if (Cs >= 5106) {
                                                            if (AFD < 18) {
                                                                return 8;
                                                            }
                                                            else if (AFD >= 18) {
                                                                return 1;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (negBalance >= 56) {
                                                    if (Rs < 5700) {
                                                        return 8;
                                                    }
                                                    else if (Rs >= 5700) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                        }
                                        else if (AFD >= 21) {
                                            if (Rs < 5274) {
                                                if (refDCval < 121) {
                                                    return 4;
                                                }
                                                else if (refDCval >= 121) {
                                                    return 8;
                                                }
                                            }
                                            else if (Rs >= 5274) {
                                                if (diffAFD < 14) {
                                                    return 4;
                                                }
                                                else if (diffAFD >= 14) {
                                                    return 8;
                                                }
                                            }
                                        }
                                    }
                                    else if (CsDiff >= 17) {
                                        if (mu_mv_mag_sq < 3) {
                                            if (tcor < 83) {
                                                if (posBalance < 53) {
                                                    return 4;
                                                }
                                                else if (posBalance >= 53) {
                                                    if (AFD < 22) {
                                                        return 1;
                                                    }
                                                    else if (AFD >= 22) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                            else if (tcor >= 83) {
                                                if (Cs < 4344) {
                                                    return 4;
                                                }
                                                else if (Cs >= 4344) {
                                                    return 1;
                                                }
                                            }
                                        }
                                        else if (mu_mv_mag_sq >= 3) {
                                            if (MVDiff < 67) {
                                                if (RsCsDiff < 525) {
                                                    return 1;
                                                }
                                                else if (RsCsDiff >= 525) {
                                                    if (CsDiff < 18) {
                                                        if (RsCsDiff < 796) {
                                                            return 8;
                                                        }
                                                        else if (RsCsDiff >= 796) {
                                                            if (Cs < 4542) {
                                                                return 4;
                                                            }
                                                            else if (Cs >= 4542) {
                                                                return 1;
                                                            }
                                                        }
                                                    }
                                                    else if (CsDiff >= 18) {
                                                        if (diffRsCsdiff < -239) {
                                                            if (Rs < 6701) {
                                                                if (Cs < 4355) {
                                                                    return 4;
                                                                }
                                                                else if (Cs >= 4355) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (Rs >= 6701) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (diffRsCsdiff >= -239) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (MVDiff >= 67) {
                                                if (CsDiff < 29) {
                                                    return 8;
                                                }
                                                else if (CsDiff >= 29) {
                                                    return 2;
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (Rs >= 6988) {
                                    if (ssDCval < 79) {
                                        return 4;
                                    }
                                    else if (ssDCval >= 79) {
                                        if (MVDiff < 22) {
                                            if (mcTcor < 82) {
                                                if (tcor < 65) {
                                                    if (RsDiff < 58) {
                                                        return 8;
                                                    }
                                                    else if (RsDiff >= 58) {
                                                        return 4;
                                                    }
                                                }
                                                else if (tcor >= 65) {
                                                    if (Cs < 7037) {
                                                        return 8;
                                                    }
                                                    else if (Cs >= 7037) {
                                                        if (Rs < 15296) {
                                                            return 4;
                                                        }
                                                        else if (Rs >= 15296) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (mcTcor >= 82) {
                                                if (SC < 23998) {
                                                    if (mu_mv_mag_sq < 13) {
                                                        if (tcor < 84) {
                                                            if (Rs < 11705) {
                                                                if (CsDiff < 32) {
                                                                    return 8;
                                                                }
                                                                else if (CsDiff >= 32) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (Rs >= 11705) {
                                                                return 2;
                                                            }
                                                        }
                                                        else if (tcor >= 84) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (mu_mv_mag_sq >= 13) {
                                                        if (RsCsDiff < 1662) {
                                                            if (SC < 15542) {
                                                                if (Rs < 8367) {
                                                                    if (mu_mv_mag_sq < 57) {
                                                                        return 4;
                                                                    }
                                                                    else if (mu_mv_mag_sq >= 57) {
                                                                        return 8;
                                                                    }
                                                                }
                                                                else if (Rs >= 8367) {
                                                                    return 2;
                                                                }
                                                            }
                                                            else if (SC >= 15542) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (RsCsDiff >= 1662) {
                                                            if (CsDiff < 39) {
                                                                return 4;
                                                            }
                                                            else if (CsDiff >= 39) {
                                                                return 2;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (SC >= 23998) {
                                                    return 8;
                                                }
                                            }
                                        }
                                        else if (MVDiff >= 22) {
                                            if (gchDC < 14) {
                                                if (ssDCval < 145) {
                                                    return 2;
                                                }
                                                else if (ssDCval >= 145) {
                                                    return 8;
                                                }
                                            }
                                            else if (gchDC >= 14) {
                                                return 8;
                                            }
                                        }
                                    }
                                }
                            }
                            else if (refDCval >= 150) {
                                if (Cs < 3822) {
                                    if (tcor < 78) {
                                        if (RsCsDiff < 583) {
                                            return 2;
                                        }
                                        else if (RsCsDiff >= 583) {
                                            return 4;
                                        }
                                    }
                                    else if (tcor >= 78) {
                                        if (Cs < 3458) {
                                            return 8;
                                        }
                                        else if (Cs >= 3458) {
                                            return 2;
                                        }
                                    }
                                }
                                else if (Cs >= 3822) {
                                    if (CsDiff < 23) {
                                        if (posBalance < 58) {
                                            if (AFD < 31) {
                                                if (tcor < 70) {
                                                    return 4;
                                                }
                                                else if (tcor >= 70) {
                                                    return 8;
                                                }
                                            }
                                            else if (AFD >= 31) {
                                                return 4;
                                            }
                                        }
                                        else if (posBalance >= 58) {
                                            if (SC < 12840) {
                                                return 2;
                                            }
                                            else if (SC >= 12840) {
                                                return 8;
                                            }
                                        }
                                    }
                                    else if (CsDiff >= 23) {
                                        if (TSC < 20) {
                                            if (diffMVdiffVal < -6) {
                                                if (AFD < 28) {
                                                    return 8;
                                                }
                                                else if (AFD >= 28) {
                                                    return 4;
                                                }
                                            }
                                            else if (diffMVdiffVal >= -6) {
                                                return 8;
                                            }
                                        }
                                        else if (TSC >= 20) {
                                            if (RsDiff < 36) {
                                                if (Rs < 9063) {
                                                    return 8;
                                                }
                                                else if (Rs >= 9063) {
                                                    return 4;
                                                }
                                            }
                                            else if (RsDiff >= 36) {
                                                if (negBalance < 50) {
                                                    return 8;
                                                }
                                                else if (negBalance >= 50) {
                                                    return 2;
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                        }
                    }
                    else if (mu_mv_mag_sq >= 63) {
                        if (Cs < 7392) {
                            if (ssDCval < 105) {
                                if (AFD < 8) {
                                    return 4;
                                }
                                else if (AFD >= 8) {
                                    return 8;
                                }
                            }
                            else if (ssDCval >= 105) {
                                if (mu_mv_mag_sq < 122) {
                                    if (tcor < 81) {
                                        if (refDCval < 128) {
                                            if (mu_mv_mag_sq < 72) {
                                                if (Cs < 4295) {
                                                    return 4;
                                                }
                                                else if (Cs >= 4295) {
                                                    return 8;
                                                }
                                            }
                                            else if (mu_mv_mag_sq >= 72) {
                                                return 8;
                                            }
                                        }
                                        else if (refDCval >= 128) {
                                            if (negBalance < 45) {
                                                return 8;
                                            }
                                            else if (negBalance >= 45) {
                                                return 1;
                                            }
                                        }
                                    }
                                    else if (tcor >= 81) {
                                        if (SC < 8474) {
                                            if (diffAFD < -2) {
                                                if (MVDiff < 3) {
                                                    return 4;
                                                }
                                                else if (MVDiff >= 3) {
                                                    if (Rs < 4326) {
                                                        return 8;
                                                    }
                                                    else if (Rs >= 4326) {
                                                        return 2;
                                                    }
                                                }
                                            }
                                            else if (diffAFD >= -2) {
                                                if (diffRsCsdiff < 83) {
                                                    if (refDCval < 116) {
                                                        if (Cs < 3823) {
                                                            return 4;
                                                        }
                                                        else if (Cs >= 3823) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (refDCval >= 116) {
                                                        return 8;
                                                    }
                                                }
                                                else if (diffRsCsdiff >= 83) {
                                                    return 2;
                                                }
                                            }
                                        }
                                        else if (SC >= 8474) {
                                            if (Cs < 3985) {
                                                if (SC < 9010) {
                                                    return 8;
                                                }
                                                else if (SC >= 9010) {
                                                    return 2;
                                                }
                                            }
                                            else if (Cs >= 3985) {
                                                return 8;
                                            }
                                        }
                                    }
                                }
                                else if (mu_mv_mag_sq >= 122) {
                                    if (ssDCval < 106) {
                                        return 8;
                                    }
                                    else if (ssDCval >= 106) {
                                        if (refDCval < 116) {
                                            if (SC < 12077) {
                                                if (Cs < 5289) {
                                                    if (RsCsDiff < 289) {
                                                        return 4;
                                                    }
                                                    else if (RsCsDiff >= 289) {
                                                        return 8;
                                                    }
                                                }
                                                else if (Cs >= 5289) {
                                                    return 4;
                                                }
                                            }
                                            else if (SC >= 12077) {
                                                if (refDCval < 107) {
                                                    return 4;
                                                }
                                                else if (refDCval >= 107) {
                                                    return 8;
                                                }
                                            }
                                        }
                                        else if (refDCval >= 116) {
                                            return 8;
                                        }
                                    }
                                }
                            }
                        }
                        else if (Cs >= 7392) {
                            if (refDCval < 93) {
                                if (SC < 15437) {
                                    return 4;
                                }
                                else if (SC >= 15437) {
                                    if (posBalance < 54) {
                                        if (RsCsDiff < 1036) {
                                            return 2;
                                        }
                                        else if (RsCsDiff >= 1036) {
                                            if (diffRsCsdiff < 69) {
                                                return 4;
                                            }
                                            else if (diffRsCsdiff >= 69) {
                                                return 2;
                                            }
                                        }
                                    }
                                    else if (posBalance >= 54) {
                                        if (Cs < 7996) {
                                            return 4;
                                        }
                                        else if (Cs >= 7996) {
                                            return 8;
                                        }
                                    }
                                }
                            }
                            else if (refDCval >= 93) {
                                return 8;
                            }
                        }
                    }
                }
            }
        }
        return 4;
    }
    mfxU16 AGOPSelect7(mfxU32 Rs, mfxU32 CsDiff, mfxU32 TSC, mfxI32 diffAFD, mfxU32 Scindex, mfxU32 RsDiff, mfxI32 diffMVdiffVal, mfxU32 RsCsDiff, mfxU32 TSCindex, mfxI16 tcor, mfxU32 refDCval, mfxU32 MVDiff, mfxU32 AFD, mfxI32 diffTSC, mfxI32 diffRsCsdiff, mfxU32 posBalance, mfxU32 SC, mfxU32 mu_mv_mag_sq, mfxU32 negBalance, mfxI16 mcTcor, mfxU32 gchDC, mfxU32 Cs, mfxU32 ssDCval) {
        if (tcor < 95) {
            if (mu_mv_mag_sq < 83) {
                if (refDCval < 9) {
                    if (Cs < 1787) {
                        if (Rs < 366) {
                            return 4;
                        }
                        else if (Rs >= 366) {
                            return 8;
                        }
                    }
                    else if (Cs >= 1787) {
                        if (Rs < 4729) {
                            if (SC < 4597) {
                                return 1;
                            }
                            else if (SC >= 4597) {
                                if (SC < 7709) {
                                    return 8;
                                }
                                else if (SC >= 7709) {
                                    if (SC < 7755) {
                                        return 1;
                                    }
                                    else if (SC >= 7755) {
                                        return 8;
                                    }
                                }
                            }
                        }
                        else if (Rs >= 4729) {
                            if (Rs < 5376) {
                                return 1;
                            }
                            else if (Rs >= 5376) {
                                return 8;
                            }
                        }
                    }
                }
                else if (refDCval >= 9) {
                    if (Cs < 3266) {
                        if (AFD < 12) {
                            if (refDCval < 94) {
                                if (SC < 933) {
                                    if (TSC < 6) {
                                        if (ssDCval < 56) {
                                            if (Cs < 382) {
                                                if (mcTcor < 92) {
                                                    if (Rs < 381) {
                                                        if (MVDiff < 47) {
                                                            if (posBalance < 60) {
                                                                if (mcTcor < 84) {
                                                                    return 8;
                                                                }
                                                                else if (mcTcor >= 84) {
                                                                    if (MVDiff < 22) {
                                                                        return 4;
                                                                    }
                                                                    else if (MVDiff >= 22) {
                                                                        return 1;
                                                                    }
                                                                }
                                                            }
                                                            else if (posBalance >= 60) {
                                                                if (mu_mv_mag_sq < 23) {
                                                                    if (tcor < 53) {
                                                                        return 1;
                                                                    }
                                                                    else if (tcor >= 53) {
                                                                        if (posBalance < 89) {
                                                                            return 2;
                                                                        }
                                                                        else if (posBalance >= 89) {
                                                                            if (posBalance < 111) {
                                                                                if (SC < 134) {
                                                                                    return 4;
                                                                                }
                                                                                else if (SC >= 134) {
                                                                                    return 2;
                                                                                }
                                                                            }
                                                                            else if (posBalance >= 111) {
                                                                                return 2;
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                                else if (mu_mv_mag_sq >= 23) {
                                                                    if (tcor < 80) {
                                                                        if (Cs < 29) {
                                                                            if (MVDiff < 43) {
                                                                                return 4;
                                                                            }
                                                                            else if (MVDiff >= 43) {
                                                                                return 8;
                                                                            }
                                                                        }
                                                                        else if (Cs >= 29) {
                                                                            if (Rs < 192) {
                                                                                return 2;
                                                                            }
                                                                            else if (Rs >= 192) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (tcor >= 80) {
                                                                        if (diffMVdiffVal < 5) {
                                                                            return 8;
                                                                        }
                                                                        else if (diffMVdiffVal >= 5) {
                                                                            if (AFD < 5) {
                                                                                return 4;
                                                                            }
                                                                            else if (AFD >= 5) {
                                                                                return 2;
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (MVDiff >= 47) {
                                                            if (negBalance < 1) {
                                                                if (diffAFD < -1) {
                                                                    if (mu_mv_mag_sq < 8) {
                                                                        return 1;
                                                                    }
                                                                    else if (mu_mv_mag_sq >= 8) {
                                                                        if (ssDCval < 19) {
                                                                            if (MVDiff < 274) {
                                                                                return 1;
                                                                            }
                                                                            else if (MVDiff >= 274) {
                                                                                return 2;
                                                                            }
                                                                        }
                                                                        else if (ssDCval >= 19) {
                                                                            if (tcor < 39) {
                                                                                if (MVDiff < 203) {
                                                                                    return 2;
                                                                                }
                                                                                else if (MVDiff >= 203) {
                                                                                    return 4;
                                                                                }
                                                                            }
                                                                            else if (tcor >= 39) {
                                                                                return 2;
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                                else if (diffAFD >= -1) {
                                                                    if (refDCval < 22) {
                                                                        if (MVDiff < 140) {
                                                                            if (MVDiff < 95) {
                                                                                if (TSC < 2) {
                                                                                    return 4;
                                                                                }
                                                                                else if (TSC >= 2) {
                                                                                    return 2;
                                                                                }
                                                                            }
                                                                            else if (MVDiff >= 95) {
                                                                                if (posBalance < 83) {
                                                                                    return 2;
                                                                                }
                                                                                else if (posBalance >= 83) {
                                                                                    return 8;
                                                                                }
                                                                            }
                                                                        }
                                                                        else if (MVDiff >= 140) {
                                                                            if (MVDiff < 696) {
                                                                                return 4;
                                                                            }
                                                                            else if (MVDiff >= 696) {
                                                                                return 2;
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (refDCval >= 22) {
                                                                        if (mcTcor < 69) {
                                                                            return 1;
                                                                        }
                                                                        else if (mcTcor >= 69) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                            else if (negBalance >= 1) {
                                                                if (AFD < 2) {
                                                                    if (tcor < 59) {
                                                                        if (TSCindex < 1) {
                                                                            if (MVDiff < 776) {
                                                                                if (diffMVdiffVal < -7) {
                                                                                    if (posBalance < 23) {
                                                                                        return 4;
                                                                                    }
                                                                                    else if (posBalance >= 23) {
                                                                                        return 8;
                                                                                    }
                                                                                }
                                                                                else if (diffMVdiffVal >= -7) {
                                                                                    if (MVDiff < 62) {
                                                                                        if (posBalance < 28) {
                                                                                            if (posBalance < 27) {
                                                                                                return 1;
                                                                                            }
                                                                                            else if (posBalance >= 27) {
                                                                                                if (MVDiff < 59) {
                                                                                                    return 4;
                                                                                                }
                                                                                                else if (MVDiff >= 59) {
                                                                                                    return 1;
                                                                                                }
                                                                                            }
                                                                                        }
                                                                                        else if (posBalance >= 28) {
                                                                                            return 4;
                                                                                        }
                                                                                    }
                                                                                    else if (MVDiff >= 62) {
                                                                                        if (mu_mv_mag_sq < 26) {
                                                                                            if (MVDiff < 72) {
                                                                                                return 4;
                                                                                            }
                                                                                            else if (MVDiff >= 72) {
                                                                                                return 1;
                                                                                            }
                                                                                        }
                                                                                        else if (mu_mv_mag_sq >= 26) {
                                                                                            if (mu_mv_mag_sq < 35) {
                                                                                                return 8;
                                                                                            }
                                                                                            else if (mu_mv_mag_sq >= 35) {
                                                                                                if (posBalance < 35) {
                                                                                                    return 4;
                                                                                                }
                                                                                                else if (posBalance >= 35) {
                                                                                                    return 8;
                                                                                                }
                                                                                            }
                                                                                        }
                                                                                    }
                                                                                }
                                                                            }
                                                                            else if (MVDiff >= 776) {
                                                                                if (MVDiff < 1221) {
                                                                                    return 4;
                                                                                }
                                                                                else if (MVDiff >= 1221) {
                                                                                    return 2;
                                                                                }
                                                                            }
                                                                        }
                                                                        else if (TSCindex >= 1) {
                                                                            if (ssDCval < 22) {
                                                                                if (mcTcor < 39) {
                                                                                    return 8;
                                                                                }
                                                                                else if (mcTcor >= 39) {
                                                                                    return 2;
                                                                                }
                                                                            }
                                                                            else if (ssDCval >= 22) {
                                                                                return 1;
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (tcor >= 59) {
                                                                        if (negBalance < 17) {
                                                                            if (MVDiff < 82) {
                                                                                if (AFD < 1) {
                                                                                    if (MVDiff < 77) {
                                                                                        return 8;
                                                                                    }
                                                                                    else if (MVDiff >= 77) {
                                                                                        return 4;
                                                                                    }
                                                                                }
                                                                                else if (AFD >= 1) {
                                                                                    return 4;
                                                                                }
                                                                            }
                                                                            else if (MVDiff >= 82) {
                                                                                return 8;
                                                                            }
                                                                        }
                                                                        else if (negBalance >= 17) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                }
                                                                else if (AFD >= 2) {
                                                                    if (posBalance < 91) {
                                                                        return 1;
                                                                    }
                                                                    else if (posBalance >= 91) {
                                                                        if (diffMVdiffVal < -46) {
                                                                            if (Rs < 17) {
                                                                                return 4;
                                                                            }
                                                                            else if (Rs >= 17) {
                                                                                return 8;
                                                                            }
                                                                        }
                                                                        else if (diffMVdiffVal >= -46) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (Rs >= 381) {
                                                        if (mu_mv_mag_sq < 58) {
                                                            if (diffRsCsdiff < 1) {
                                                                if (tcor < 84) {
                                                                    if (Cs < 376) {
                                                                        if (mu_mv_mag_sq < 36) {
                                                                            return 4;
                                                                        }
                                                                        else if (mu_mv_mag_sq >= 36) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                    else if (Cs >= 376) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (tcor >= 84) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (diffRsCsdiff >= 1) {
                                                                if (posBalance < 56) {
                                                                    return 1;
                                                                }
                                                                else if (posBalance >= 56) {
                                                                    if (TSCindex < 5) {
                                                                        return 4;
                                                                    }
                                                                    else if (TSCindex >= 5) {
                                                                        return 8;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (mu_mv_mag_sq >= 58) {
                                                            if (ssDCval < 38) {
                                                                if (mcTcor < 82) {
                                                                    return 1;
                                                                }
                                                                else if (mcTcor >= 82) {
                                                                    if (Rs < 428) {
                                                                        if (Cs < 334) {
                                                                            return 4;
                                                                        }
                                                                        else if (Cs >= 334) {
                                                                            if (Rs < 409) {
                                                                                return 8;
                                                                            }
                                                                            else if (Rs >= 409) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (Rs >= 428) {
                                                                        if (Cs < 366) {
                                                                            return 1;
                                                                        }
                                                                        else if (Cs >= 366) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                            else if (ssDCval >= 38) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (mcTcor >= 92) {
                                                    if (MVDiff < 59) {
                                                        return 2;
                                                    }
                                                    else if (MVDiff >= 59) {
                                                        if (Rs < 6) {
                                                            return 4;
                                                        }
                                                        else if (Rs >= 6) {
                                                            return 1;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (Cs >= 382) {
                                                if (refDCval < 39) {
                                                    if (Rs < 466) {
                                                        if (Cs < 405) {
                                                            if (mu_mv_mag_sq < 68) {
                                                                if (Rs < 454) {
                                                                    if (diffTSC < 1) {
                                                                        return 2;
                                                                    }
                                                                    else if (diffTSC >= 1) {
                                                                        if (Cs < 384) {
                                                                            return 4;
                                                                        }
                                                                        else if (Cs >= 384) {
                                                                            return 2;
                                                                        }
                                                                    }
                                                                }
                                                                else if (Rs >= 454) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (mu_mv_mag_sq >= 68) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (Cs >= 405) {
                                                            if (AFD < 5) {
                                                                if (Rs < 457) {
                                                                    return 2;
                                                                }
                                                                else if (Rs >= 457) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (AFD >= 5) {
                                                                return 2;
                                                            }
                                                        }
                                                    }
                                                    else if (Rs >= 466) {
                                                        if (posBalance < 58) {
                                                            return 2;
                                                        }
                                                        else if (posBalance >= 58) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                                else if (refDCval >= 39) {
                                                    return 4;
                                                }
                                            }
                                        }
                                        else if (ssDCval >= 56) {
                                            if (TSC < 5) {
                                                if (negBalance < 49) {
                                                    if (Rs < 243) {
                                                        if (posBalance < 60) {
                                                            return 4;
                                                        }
                                                        else if (posBalance >= 60) {
                                                            if (gchDC < 7) {
                                                                if (AFD < 9) {
                                                                    return 2;
                                                                }
                                                                else if (AFD >= 9) {
                                                                    return 1;
                                                                }
                                                            }
                                                            else if (gchDC >= 7) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                    else if (Rs >= 243) {
                                                        return 8;
                                                    }
                                                }
                                                else if (negBalance >= 49) {
                                                    if (Rs < 325) {
                                                        return 2;
                                                    }
                                                    else if (Rs >= 325) {
                                                        if (Rs < 357) {
                                                            return 4;
                                                        }
                                                        else if (Rs >= 357) {
                                                            return 2;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (TSC >= 5) {
                                                if (SC < 660) {
                                                    if (diffMVdiffVal < 6) {
                                                        if (tcor < 77) {
                                                            return 2;
                                                        }
                                                        else if (tcor >= 77) {
                                                            return 1;
                                                        }
                                                    }
                                                    else if (diffMVdiffVal >= 6) {
                                                        return 2;
                                                    }
                                                }
                                                else if (SC >= 660) {
                                                    if (RsCsDiff < 7) {
                                                        if (Rs < 367) {
                                                            return 2;
                                                        }
                                                        else if (Rs >= 367) {
                                                            return 1;
                                                        }
                                                    }
                                                    else if (RsCsDiff >= 7) {
                                                        return 2;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (TSC >= 6) {
                                        if (refDCval < 81) {
                                            if (SC < 725) {
                                                if (ssDCval < 74) {
                                                    return 1;
                                                }
                                                else if (ssDCval >= 74) {
                                                    return 2;
                                                }
                                            }
                                            else if (SC >= 725) {
                                                if (mcTcor < 94) {
                                                    return 8;
                                                }
                                                else if (mcTcor >= 94) {
                                                    if (Rs < 591) {
                                                        return 8;
                                                    }
                                                    else if (Rs >= 591) {
                                                        return 2;
                                                    }
                                                }
                                            }
                                        }
                                        else if (refDCval >= 81) {
                                            if (RsDiff < 1) {
                                                return 2;
                                            }
                                            else if (RsDiff >= 1) {
                                                return 1;
                                            }
                                        }
                                    }
                                }
                                else if (SC >= 933) {
                                    if (TSC < 7) {
                                        if (ssDCval < 39) {
                                            if (Scindex < 4) {
                                                if (TSC < 5) {
                                                    if (posBalance < 14) {
                                                        return 8;
                                                    }
                                                    else if (posBalance >= 14) {
                                                        if (mcTcor < 72) {
                                                            return 8;
                                                        }
                                                        else if (mcTcor >= 72) {
                                                            if (tcor < 85) {
                                                                if (Cs < 349) {
                                                                    if (Rs < 829) {
                                                                        return 1;
                                                                    }
                                                                    else if (Rs >= 829) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (Cs >= 349) {
                                                                    if (mu_mv_mag_sq < 71) {
                                                                        return 2;
                                                                    }
                                                                    else if (mu_mv_mag_sq >= 71) {
                                                                        return 4;
                                                                    }
                                                                }
                                                            }
                                                            else if (tcor >= 85) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (TSC >= 5) {
                                                    if (TSC < 6) {
                                                        if (SC < 1101) {
                                                            return 2;
                                                        }
                                                        else if (SC >= 1101) {
                                                            if (negBalance < 51) {
                                                                if (tcor < 81) {
                                                                    if (diffTSC < -1) {
                                                                        if (RsDiff < 2) {
                                                                            return 4;
                                                                        }
                                                                        else if (RsDiff >= 2) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                    else if (diffTSC >= -1) {
                                                                        if (mu_mv_mag_sq < 38) {
                                                                            if (Rs < 757) {
                                                                                return 4;
                                                                            }
                                                                            else if (Rs >= 757) {
                                                                                return 1;
                                                                            }
                                                                        }
                                                                        else if (mu_mv_mag_sq >= 38) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                }
                                                                else if (tcor >= 81) {
                                                                    if (AFD < 6) {
                                                                        if (Cs < 569) {
                                                                            return 1;
                                                                        }
                                                                        else if (Cs >= 569) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                    else if (AFD >= 6) {
                                                                        return 1;
                                                                    }
                                                                }
                                                            }
                                                            else if (negBalance >= 51) {
                                                                return 1;
                                                            }
                                                        }
                                                    }
                                                    else if (TSC >= 6) {
                                                        if (mcTcor < 74) {
                                                            return 4;
                                                        }
                                                        else if (mcTcor >= 74) {
                                                            return 1;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (Scindex >= 4) {
                                                if (mu_mv_mag_sq < 57) {
                                                    if (negBalance < 49) {
                                                        if (Cs < 1550) {
                                                            return 2;
                                                        }
                                                        else if (Cs >= 1550) {
                                                            return 1;
                                                        }
                                                    }
                                                    else if (negBalance >= 49) {
                                                        if (TSC < 5) {
                                                            if (Rs < 1672) {
                                                                return 2;
                                                            }
                                                            else if (Rs >= 1672) {
                                                                return 1;
                                                            }
                                                        }
                                                        else if (TSC >= 5) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                                else if (mu_mv_mag_sq >= 57) {
                                                    if (Cs < 873) {
                                                        return 8;
                                                    }
                                                    else if (Cs >= 873) {
                                                        if (Cs < 2315) {
                                                            if (AFD < 1) {
                                                                return 2;
                                                            }
                                                            else if (AFD >= 1) {
                                                                return 1;
                                                            }
                                                        }
                                                        else if (Cs >= 2315) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (ssDCval >= 39) {
                                            if (RsCsDiff < 23) {
                                                if (SC < 1963) {
                                                    if (negBalance < 43) {
                                                        if (SC < 1639) {
                                                            if (mu_mv_mag_sq < 41) {
                                                                if (mcTcor < 96) {
                                                                    if (TSC < 5) {
                                                                        if (RsCsDiff < 9) {
                                                                            if (Rs < 790) {
                                                                                return 4;
                                                                            }
                                                                            else if (Rs >= 790) {
                                                                                if (Rs < 890) {
                                                                                    return 8;
                                                                                }
                                                                                else if (Rs >= 890) {
                                                                                    return 4;
                                                                                }
                                                                            }
                                                                        }
                                                                        else if (RsCsDiff >= 9) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                    else if (TSC >= 5) {
                                                                        return 8;
                                                                    }
                                                                }
                                                                else if (mcTcor >= 96) {
                                                                    if (MVDiff < 1) {
                                                                        return 4;
                                                                    }
                                                                    else if (MVDiff >= 1) {
                                                                        return 8;
                                                                    }
                                                                }
                                                            }
                                                            else if (mu_mv_mag_sq >= 41) {
                                                                if (RsCsDiff < 7) {
                                                                    if (ssDCval < 54) {
                                                                        if (diffRsCsdiff < 1) {
                                                                            return 4;
                                                                        }
                                                                        else if (diffRsCsdiff >= 1) {
                                                                            if (Rs < 779) {
                                                                                return 8;
                                                                            }
                                                                            else if (Rs >= 779) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (ssDCval >= 54) {
                                                                        if (ssDCval < 67) {
                                                                            return 8;
                                                                        }
                                                                        else if (ssDCval >= 67) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                }
                                                                else if (RsCsDiff >= 7) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                        else if (SC >= 1639) {
                                                            if (MVDiff < 20) {
                                                                if (posBalance < 50) {
                                                                    if (mcTcor < 94) {
                                                                        return 1;
                                                                    }
                                                                    else if (mcTcor >= 94) {
                                                                        if (SC < 1747) {
                                                                            if (Rs < 952) {
                                                                                return 4;
                                                                            }
                                                                            else if (Rs >= 952) {
                                                                                return 1;
                                                                            }
                                                                        }
                                                                        else if (SC >= 1747) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                }
                                                                else if (posBalance >= 50) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (MVDiff >= 20) {
                                                                if (Rs < 795) {
                                                                    return 2;
                                                                }
                                                                else if (Rs >= 795) {
                                                                    if (Rs < 964) {
                                                                        return 8;
                                                                    }
                                                                    else if (Rs >= 964) {
                                                                        return 1;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (negBalance >= 43) {
                                                        if (Rs < 718) {
                                                            if (AFD < 9) {
                                                                if (diffRsCsdiff < -2) {
                                                                    if (mu_mv_mag_sq < 29) {
                                                                        return 4;
                                                                    }
                                                                    else if (mu_mv_mag_sq >= 29) {
                                                                        if (Cs < 440) {
                                                                            return 4;
                                                                        }
                                                                        else if (Cs >= 440) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                }
                                                                else if (diffRsCsdiff >= -2) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (AFD >= 9) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (Rs >= 718) {
                                                            if (mcTcor < 84) {
                                                                if (RsCsDiff < 11) {
                                                                    if (MVDiff < 28) {
                                                                        return 4;
                                                                    }
                                                                    else if (MVDiff >= 28) {
                                                                        if (mu_mv_mag_sq < 33) {
                                                                            if (Rs < 817) {
                                                                                if (diffMVdiffVal < 3) {
                                                                                    if (SC < 1334) {
                                                                                        return 4;
                                                                                    }
                                                                                    else if (SC >= 1334) {
                                                                                        return 1;
                                                                                    }
                                                                                }
                                                                                else if (diffMVdiffVal >= 3) {
                                                                                    return 1;
                                                                                }
                                                                            }
                                                                            else if (Rs >= 817) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                        else if (mu_mv_mag_sq >= 33) {
                                                                            if (mcTcor < 79) {
                                                                                if (mu_mv_mag_sq < 35) {
                                                                                    if (Cs < 590) {
                                                                                        return 1;
                                                                                    }
                                                                                    else if (Cs >= 590) {
                                                                                        return 4;
                                                                                    }
                                                                                }
                                                                                else if (mu_mv_mag_sq >= 35) {
                                                                                    if (Rs < 764) {
                                                                                        return 4;
                                                                                    }
                                                                                    else if (Rs >= 764) {
                                                                                        if (MVDiff < 31) {
                                                                                            return 4;
                                                                                        }
                                                                                        else if (MVDiff >= 31) {
                                                                                            return 1;
                                                                                        }
                                                                                    }
                                                                                }
                                                                            }
                                                                            else if (mcTcor >= 79) {
                                                                                return 1;
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                                else if (RsCsDiff >= 11) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (mcTcor >= 84) {
                                                                if (diffAFD < -1) {
                                                                    if (Cs < 609) {
                                                                        return 2;
                                                                    }
                                                                    else if (Cs >= 609) {
                                                                        if (posBalance < 52) {
                                                                            if (diffRsCsdiff < -8) {
                                                                                return 8;
                                                                            }
                                                                            else if (diffRsCsdiff >= -8) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                        else if (posBalance >= 52) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                }
                                                                else if (diffAFD >= -1) {
                                                                    if (RsCsDiff < 7) {
                                                                        return 1;
                                                                    }
                                                                    else if (RsCsDiff >= 7) {
                                                                        if (refDCval < 53) {
                                                                            return 1;
                                                                        }
                                                                        else if (refDCval >= 53) {
                                                                            if (negBalance < 49) {
                                                                                if (SC < 1477) {
                                                                                    return 2;
                                                                                }
                                                                                else if (SC >= 1477) {
                                                                                    return 4;
                                                                                }
                                                                            }
                                                                            else if (negBalance >= 49) {
                                                                                if (TSC < 6) {
                                                                                    if (Rs < 930) {
                                                                                        return 4;
                                                                                    }
                                                                                    else if (Rs >= 930) {
                                                                                        return 8;
                                                                                    }
                                                                                }
                                                                                else if (TSC >= 6) {
                                                                                    return 1;
                                                                                }
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (SC >= 1963) {
                                                    if (Cs < 1188) {
                                                        if (mcTcor < 94) {
                                                            if (Rs < 1349) {
                                                                if (SC < 2150) {
                                                                    if (mu_mv_mag_sq < 62) {
                                                                        if (posBalance < 59) {
                                                                            if (mcTcor < 92) {
                                                                                return 1;
                                                                            }
                                                                            else if (mcTcor >= 92) {
                                                                                if (refDCval < 53) {
                                                                                    if (posBalance < 51) {
                                                                                        return 8;
                                                                                    }
                                                                                    else if (posBalance >= 51) {
                                                                                        if (AFD < 8) {
                                                                                            return 1;
                                                                                        }
                                                                                        else if (AFD >= 8) {
                                                                                            return 4;
                                                                                        }
                                                                                    }
                                                                                }
                                                                                else if (refDCval >= 53) {
                                                                                    if (RsDiff < 3) {
                                                                                        return 8;
                                                                                    }
                                                                                    else if (RsDiff >= 3) {
                                                                                        return 4;
                                                                                    }
                                                                                }
                                                                            }
                                                                        }
                                                                        else if (posBalance >= 59) {
                                                                            if (Rs < 1195) {
                                                                                return 4;
                                                                            }
                                                                            else if (Rs >= 1195) {
                                                                                return 8;
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (mu_mv_mag_sq >= 62) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (SC >= 2150) {
                                                                    if (AFD < 8) {
                                                                        if (posBalance < 48) {
                                                                            if (Cs < 978) {
                                                                                return 4;
                                                                            }
                                                                            else if (Cs >= 978) {
                                                                                return 1;
                                                                            }
                                                                        }
                                                                        else if (posBalance >= 48) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                    else if (AFD >= 8) {
                                                                        return 8;
                                                                    }
                                                                }
                                                            }
                                                            else if (Rs >= 1349) {
                                                                if (Cs < 1023) {
                                                                    if (mu_mv_mag_sq < 21) {
                                                                        if (Rs < 1408) {
                                                                            return 4;
                                                                        }
                                                                        else if (Rs >= 1408) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                    else if (mu_mv_mag_sq >= 21) {
                                                                        if (posBalance < 51) {
                                                                            if (Rs < 1411) {
                                                                                if (RsCsDiff < 17) {
                                                                                    return 4;
                                                                                }
                                                                                else if (RsCsDiff >= 17) {
                                                                                    return 1;
                                                                                }
                                                                            }
                                                                            else if (Rs >= 1411) {
                                                                                return 1;
                                                                            }
                                                                        }
                                                                        else if (posBalance >= 51) {
                                                                            if (MVDiff < 8) {
                                                                                return 4;
                                                                            }
                                                                            else if (MVDiff >= 8) {
                                                                                return 1;
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                                else if (Cs >= 1023) {
                                                                    if (Rs < 1500) {
                                                                        if (Cs < 1103) {
                                                                            return 8;
                                                                        }
                                                                        else if (Cs >= 1103) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                    else if (Rs >= 1500) {
                                                                        return 1;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (mcTcor >= 94) {
                                                            if (Cs < 569) {
                                                                if (Rs < 2152) {
                                                                    return 8;
                                                                }
                                                                else if (Rs >= 2152) {
                                                                    return 2;
                                                                }
                                                            }
                                                            else if (Cs >= 569) {
                                                                if (Rs < 1395) {
                                                                    return 1;
                                                                }
                                                                else if (Rs >= 1395) {
                                                                    if (Cs < 780) {
                                                                        return 4;
                                                                    }
                                                                    else if (Cs >= 780) {
                                                                        if (posBalance < 55) {
                                                                            if (ssDCval < 56) {
                                                                                if (Rs < 1408) {
                                                                                    return 4;
                                                                                }
                                                                                else if (Rs >= 1408) {
                                                                                    return 1;
                                                                                }
                                                                            }
                                                                            else if (ssDCval >= 56) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                        else if (posBalance >= 55) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (Cs >= 1188) {
                                                        if (SC < 4897) {
                                                            if (Cs < 1552) {
                                                                if (Rs < 1692) {
                                                                    return 8;
                                                                }
                                                                else if (Rs >= 1692) {
                                                                    if (AFD < 8) {
                                                                        if (TSC < 5) {
                                                                            if (Rs < 1867) {
                                                                                if (Cs < 1493) {
                                                                                    if (Rs < 1834) {
                                                                                        return 4;
                                                                                    }
                                                                                    else if (Rs >= 1834) {
                                                                                        return 8;
                                                                                    }
                                                                                }
                                                                                else if (Cs >= 1493) {
                                                                                    if (SC < 3323) {
                                                                                        return 4;
                                                                                    }
                                                                                    else if (SC >= 3323) {
                                                                                        if (posBalance < 57) {
                                                                                            if (negBalance < 46) {
                                                                                                if (MVDiff < 23) {
                                                                                                    if (diffAFD < 1) {
                                                                                                        return 1;
                                                                                                    }
                                                                                                    else if (diffAFD >= 1) {
                                                                                                        return 4;
                                                                                                    }
                                                                                                }
                                                                                                else if (MVDiff >= 23) {
                                                                                                    return 4;
                                                                                                }
                                                                                            }
                                                                                            else if (negBalance >= 46) {
                                                                                                return 4;
                                                                                            }
                                                                                        }
                                                                                        else if (posBalance >= 57) {
                                                                                            return 1;
                                                                                        }
                                                                                    }
                                                                                }
                                                                            }
                                                                            else if (Rs >= 1867) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                        else if (TSC >= 5) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                    else if (AFD >= 8) {
                                                                        return 1;
                                                                    }
                                                                }
                                                            }
                                                            else if (Cs >= 1552) {
                                                                if (diffMVdiffVal < -1) {
                                                                    if (diffMVdiffVal < -2) {
                                                                        if (refDCval < 40) {
                                                                            return 1;
                                                                        }
                                                                        else if (refDCval >= 40) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                    else if (diffMVdiffVal >= -2) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (diffMVdiffVal >= -1) {
                                                                    return 1;
                                                                }
                                                            }
                                                        }
                                                        else if (SC >= 4897) {
                                                            if (negBalance < 55) {
                                                                if (tcor < 93) {
                                                                    if (negBalance < 50) {
                                                                        if (Cs < 3204) {
                                                                            return 8;
                                                                        }
                                                                        else if (Cs >= 3204) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                    else if (negBalance >= 50) {
                                                                        return 8;
                                                                    }
                                                                }
                                                                else if (tcor >= 93) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (negBalance >= 55) {
                                                                return 2;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                            else if (RsCsDiff >= 23) {
                                                if (diffAFD < -1) {
                                                    if (SC < 1695) {
                                                        if (gchDC < 3) {
                                                            if (SC < 1619) {
                                                                return 2;
                                                            }
                                                            else if (SC >= 1619) {
                                                                if (Cs < 750) {
                                                                    return 4;
                                                                }
                                                                else if (Cs >= 750) {
                                                                    return 2;
                                                                }
                                                            }
                                                        }
                                                        else if (gchDC >= 3) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (SC >= 1695) {
                                                        if (ssDCval < 62) {
                                                            if (posBalance < 52) {
                                                                if (MVDiff < 26) {
                                                                    return 4;
                                                                }
                                                                else if (MVDiff >= 26) {
                                                                    if (Cs < 1148) {
                                                                        return 8;
                                                                    }
                                                                    else if (Cs >= 1148) {
                                                                        return 2;
                                                                    }
                                                                }
                                                            }
                                                            else if (posBalance >= 52) {
                                                                if (Rs < 1310) {
                                                                    return 4;
                                                                }
                                                                else if (Rs >= 1310) {
                                                                    return 1;
                                                                }
                                                            }
                                                        }
                                                        else if (ssDCval >= 62) {
                                                            if (RsCsDiff < 51) {
                                                                if (ssDCval < 88) {
                                                                    return 2;
                                                                }
                                                                else if (ssDCval >= 88) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (RsCsDiff >= 51) {
                                                                if (RsDiff < 3) {
                                                                    return 8;
                                                                }
                                                                else if (RsDiff >= 3) {
                                                                    return 1;
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (diffAFD >= -1) {
                                                    if (ssDCval < 73) {
                                                        if (SC < 2153) {
                                                            if (Scindex < 4) {
                                                                return 2;
                                                            }
                                                            else if (Scindex >= 4) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (SC >= 2153) {
                                                            if (mu_mv_mag_sq < 3) {
                                                                if (negBalance < 40) {
                                                                    if (MVDiff < 16) {
                                                                        return 4;
                                                                    }
                                                                    else if (MVDiff >= 16) {
                                                                        return 2;
                                                                    }
                                                                }
                                                                else if (negBalance >= 40) {
                                                                    return 1;
                                                                }
                                                            }
                                                            else if (mu_mv_mag_sq >= 3) {
                                                                if (refDCval < 65) {
                                                                    if (ssDCval < 53) {
                                                                        if (SC < 2486) {
                                                                            return 1;
                                                                        }
                                                                        else if (SC >= 2486) {
                                                                            if (Rs < 1764) {
                                                                                if (SC < 2525) {
                                                                                    return 8;
                                                                                }
                                                                                else if (SC >= 2525) {
                                                                                    return 1;
                                                                                }
                                                                            }
                                                                            else if (Rs >= 1764) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (ssDCval >= 53) {
                                                                        if (Cs < 1377) {
                                                                            if (MVDiff < 11) {
                                                                                return 4;
                                                                            }
                                                                            else if (MVDiff >= 11) {
                                                                                return 8;
                                                                            }
                                                                        }
                                                                        else if (Cs >= 1377) {
                                                                            if (SC < 3681) {
                                                                                return 1;
                                                                            }
                                                                            else if (SC >= 3681) {
                                                                                return 8;
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                                else if (refDCval >= 65) {
                                                                    if (SC < 3813) {
                                                                        return 2;
                                                                    }
                                                                    else if (SC >= 3813) {
                                                                        return 1;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (ssDCval >= 73) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (TSC >= 7) {
                                        if (MVDiff < 4) {
                                            if (Rs < 2369) {
                                                if (Cs < 1903) {
                                                    return 2;
                                                }
                                                else if (Cs >= 1903) {
                                                    if (Rs < 1955) {
                                                        return 4;
                                                    }
                                                    else if (Rs >= 1955) {
                                                        if (Rs < 2109) {
                                                            return 8;
                                                        }
                                                        else if (Rs >= 2109) {
                                                            return 2;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (Rs >= 2369) {
                                                if (SC < 4148) {
                                                    return 1;
                                                }
                                                else if (SC >= 4148) {
                                                    return 4;
                                                }
                                            }
                                        }
                                        else if (MVDiff >= 4) {
                                            if (mcTcor < 93) {
                                                if (refDCval < 89) {
                                                    if (Cs < 1460) {
                                                        if (MVDiff < 12) {
                                                            if (SC < 1875) {
                                                                if (refDCval < 57) {
                                                                    return 2;
                                                                }
                                                                else if (refDCval >= 57) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (SC >= 1875) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (MVDiff >= 12) {
                                                            if (diffRsCsdiff < 7) {
                                                                if (RsCsDiff < 43) {
                                                                    if (MVDiff < 26) {
                                                                        if (AFD < 11) {
                                                                            return 4;
                                                                        }
                                                                        else if (AFD >= 11) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                    else if (MVDiff >= 26) {
                                                                        return 8;
                                                                    }
                                                                }
                                                                else if (RsCsDiff >= 43) {
                                                                    return 2;
                                                                }
                                                            }
                                                            else if (diffRsCsdiff >= 7) {
                                                                if (ssDCval < 52) {
                                                                    return 8;
                                                                }
                                                                else if (ssDCval >= 52) {
                                                                    return 1;
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (Cs >= 1460) {
                                                        if (Cs < 2504) {
                                                            if (tcor < 92) {
                                                                if (Rs < 2446) {
                                                                    return 8;
                                                                }
                                                                else if (Rs >= 2446) {
                                                                    if (CsDiff < 7) {
                                                                        if (AFD < 10) {
                                                                            return 4;
                                                                        }
                                                                        else if (AFD >= 10) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                    else if (CsDiff >= 7) {
                                                                        if (Rs < 2692) {
                                                                            return 4;
                                                                        }
                                                                        else if (Rs >= 2692) {
                                                                            return 2;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                            else if (tcor >= 92) {
                                                                return 1;
                                                            }
                                                        }
                                                        else if (Cs >= 2504) {
                                                            if (Rs < 2137) {
                                                                return 1;
                                                            }
                                                            else if (Rs >= 2137) {
                                                                if (MVDiff < 8) {
                                                                    return 8;
                                                                }
                                                                else if (MVDiff >= 8) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (refDCval >= 89) {
                                                    if (Rs < 2144) {
                                                        return 4;
                                                    }
                                                    else if (Rs >= 2144) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                            else if (mcTcor >= 93) {
                                                if (refDCval < 67) {
                                                    if (mu_mv_mag_sq < 38) {
                                                        if (CsDiff < 4) {
                                                            if (posBalance < 62) {
                                                                return 2;
                                                            }
                                                            else if (posBalance >= 62) {
                                                                return 1;
                                                            }
                                                        }
                                                        else if (CsDiff >= 4) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (mu_mv_mag_sq >= 38) {
                                                        if (Cs < 3055) {
                                                            return 4;
                                                        }
                                                        else if (Cs >= 3055) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                                else if (refDCval >= 67) {
                                                    if (posBalance < 52) {
                                                        if (refDCval < 87) {
                                                            if (refDCval < 83) {
                                                                if (Cs < 1821) {
                                                                    return 4;
                                                                }
                                                                else if (Cs >= 1821) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (refDCval >= 83) {
                                                                return 1;
                                                            }
                                                        }
                                                        else if (refDCval >= 87) {
                                                            if (Rs < 2747) {
                                                                return 8;
                                                            }
                                                            else if (Rs >= 2747) {
                                                                return 2;
                                                            }
                                                        }
                                                    }
                                                    else if (posBalance >= 52) {
                                                        if (RsCsDiff < 80) {
                                                            if (negBalance < 47) {
                                                                return 1;
                                                            }
                                                            else if (negBalance >= 47) {
                                                                if (AFD < 10) {
                                                                    if (MVDiff < 16) {
                                                                        return 4;
                                                                    }
                                                                    else if (MVDiff >= 16) {
                                                                        return 1;
                                                                    }
                                                                }
                                                                else if (AFD >= 10) {
                                                                    if (MVDiff < 7) {
                                                                        if (Rs < 2355) {
                                                                            return 2;
                                                                        }
                                                                        else if (Rs >= 2355) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                    else if (MVDiff >= 7) {
                                                                        return 1;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (RsCsDiff >= 80) {
                                                            if (Cs < 1876) {
                                                                return 4;
                                                            }
                                                            else if (Cs >= 1876) {
                                                                if (ssDCval < 89) {
                                                                    if (Cs < 2123) {
                                                                        return 2;
                                                                    }
                                                                    else if (Cs >= 2123) {
                                                                        return 8;
                                                                    }
                                                                }
                                                                else if (ssDCval >= 89) {
                                                                    return 1;
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                            else if (refDCval >= 94) {
                                if (mu_mv_mag_sq < 5) {
                                    if (MVDiff < 19) {
                                        if (SC < 5127) {
                                            if (ssDCval < 99) {
                                                if (Rs < 1180) {
                                                    if (Cs < 613) {
                                                        if (mcTcor < 94) {
                                                            if (MVDiff < 10) {
                                                                if (negBalance < 54) {
                                                                    if (Rs < 619) {
                                                                        return 8;
                                                                    }
                                                                    else if (Rs >= 619) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (negBalance >= 54) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (MVDiff >= 10) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (mcTcor >= 94) {
                                                            if (Rs < 489) {
                                                                return 2;
                                                            }
                                                            else if (Rs >= 489) {
                                                                return 1;
                                                            }
                                                        }
                                                    }
                                                    else if (Cs >= 613) {
                                                        if (MVDiff < 17) {
                                                            return 4;
                                                        }
                                                        else if (MVDiff >= 17) {
                                                            return 2;
                                                        }
                                                    }
                                                }
                                                else if (Rs >= 1180) {
                                                    if (negBalance < 53) {
                                                        if (CsDiff < 6) {
                                                            return 2;
                                                        }
                                                        else if (CsDiff >= 6) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (negBalance >= 53) {
                                                        if (SC < 3420) {
                                                            return 8;
                                                        }
                                                        else if (SC >= 3420) {
                                                            return 2;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (ssDCval >= 99) {
                                                if (negBalance < 40) {
                                                    if (posBalance < 45) {
                                                        if (MVDiff < 14) {
                                                            if (RsDiff < 9) {
                                                                return 4;
                                                            }
                                                            else if (RsDiff >= 9) {
                                                                return 1;
                                                            }
                                                        }
                                                        else if (MVDiff >= 14) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (posBalance >= 45) {
                                                        if (Rs < 1205) {
                                                            if (Rs < 510) {
                                                                return 2;
                                                            }
                                                            else if (Rs >= 510) {
                                                                if (diffAFD < -4) {
                                                                    return 4;
                                                                }
                                                                else if (diffAFD >= -4) {
                                                                    return 1;
                                                                }
                                                            }
                                                        }
                                                        else if (Rs >= 1205) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                                else if (negBalance >= 40) {
                                                    if (Rs < 665) {
                                                        if (refDCval < 109) {
                                                            if (SC < 995) {
                                                                return 1;
                                                            }
                                                            else if (SC >= 995) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (refDCval >= 109) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (Rs >= 665) {
                                                        if (ssDCval < 123) {
                                                            return 4;
                                                        }
                                                        else if (ssDCval >= 123) {
                                                            if (Cs < 1968) {
                                                                if (Rs < 1936) {
                                                                    return 1;
                                                                }
                                                                else if (Rs >= 1936) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (Cs >= 1968) {
                                                                if (Cs < 2010) {
                                                                    return 8;
                                                                }
                                                                else if (Cs >= 2010) {
                                                                    if (AFD < 10) {
                                                                        return 8;
                                                                    }
                                                                    else if (AFD >= 10) {
                                                                        return 4;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (SC >= 5127) {
                                            if (refDCval < 108) {
                                                return 8;
                                            }
                                            else if (refDCval >= 108) {
                                                if (Cs < 2397) {
                                                    return 4;
                                                }
                                                else if (Cs >= 2397) {
                                                    return 8;
                                                }
                                            }
                                        }
                                    }
                                    else if (MVDiff >= 19) {
                                        if (Rs < 623) {
                                            if (SC < 826) {
                                                if (ssDCval < 99) {
                                                    return 1;
                                                }
                                                else if (ssDCval >= 99) {
                                                    return 2;
                                                }
                                            }
                                            else if (SC >= 826) {
                                                if (Rs < 500) {
                                                    return 4;
                                                }
                                                else if (Rs >= 500) {
                                                    return 8;
                                                }
                                            }
                                        }
                                        else if (Rs >= 623) {
                                            if (tcor < 80) {
                                                if (Rs < 1100) {
                                                    return 4;
                                                }
                                                else if (Rs >= 1100) {
                                                    if (Rs < 1191) {
                                                        return 2;
                                                    }
                                                    else if (Rs >= 1191) {
                                                        return 1;
                                                    }
                                                }
                                            }
                                            else if (tcor >= 80) {
                                                return 1;
                                            }
                                        }
                                    }
                                }
                                else if (mu_mv_mag_sq >= 5) {
                                    if (mu_mv_mag_sq < 30) {
                                        if (TSC < 7) {
                                            if (RsCsDiff < 8) {
                                                if (RsDiff < 2) {
                                                    return 1;
                                                }
                                                else if (RsDiff >= 2) {
                                                    return 8;
                                                }
                                            }
                                            else if (RsCsDiff >= 8) {
                                                if (Cs < 1218) {
                                                    if (TSC < 6) {
                                                        return 2;
                                                    }
                                                    else if (TSC >= 6) {
                                                        if (refDCval < 152) {
                                                            if (Cs < 774) {
                                                                return 1;
                                                            }
                                                            else if (Cs >= 774) {
                                                                return 2;
                                                            }
                                                        }
                                                        else if (refDCval >= 152) {
                                                            return 2;
                                                        }
                                                    }
                                                }
                                                else if (Cs >= 1218) {
                                                    return 1;
                                                }
                                            }
                                        }
                                        else if (TSC >= 7) {
                                            return 1;
                                        }
                                    }
                                    else if (mu_mv_mag_sq >= 30) {
                                        if (ssDCval < 125) {
                                            if (refDCval < 98) {
                                                if (MVDiff < 11) {
                                                    if (diffMVdiffVal < -4) {
                                                        return 8;
                                                    }
                                                    else if (diffMVdiffVal >= -4) {
                                                        if (mu_mv_mag_sq < 76) {
                                                            if (Rs < 1720) {
                                                                return 8;
                                                            }
                                                            else if (Rs >= 1720) {
                                                                if (posBalance < 52) {
                                                                    return 8;
                                                                }
                                                                else if (posBalance >= 52) {
                                                                    return 2;
                                                                }
                                                            }
                                                        }
                                                        else if (mu_mv_mag_sq >= 76) {
                                                            if (Scindex < 5) {
                                                                return 8;
                                                            }
                                                            else if (Scindex >= 5) {
                                                                if (Rs < 2533) {
                                                                    return 2;
                                                                }
                                                                else if (Rs >= 2533) {
                                                                    return 1;
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (MVDiff >= 11) {
                                                    if (SC < 4198) {
                                                        return 8;
                                                    }
                                                    else if (SC >= 4198) {
                                                        return 1;
                                                    }
                                                }
                                            }
                                            else if (refDCval >= 98) {
                                                if (Cs < 1685) {
                                                    if (TSC < 8) {
                                                        if (refDCval < 105) {
                                                            if (Cs < 737) {
                                                                if (Rs < 1893) {
                                                                    return 2;
                                                                }
                                                                else if (Rs >= 1893) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (Cs >= 737) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (refDCval >= 105) {
                                                            if (Cs < 349) {
                                                                return 4;
                                                            }
                                                            else if (Cs >= 349) {
                                                                if (TSC < 7) {
                                                                    if (mcTcor < 96) {
                                                                        if (Rs < 3138) {
                                                                            return 4;
                                                                        }
                                                                        else if (Rs >= 3138) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                    else if (mcTcor >= 96) {
                                                                        if (Cs < 967) {
                                                                            if (SC < 2483) {
                                                                                return 2;
                                                                            }
                                                                            else if (SC >= 2483) {
                                                                                return 8;
                                                                            }
                                                                        }
                                                                        else if (Cs >= 967) {
                                                                            if (TSC < 5) {
                                                                                return 8;
                                                                            }
                                                                            else if (TSC >= 5) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                                else if (TSC >= 7) {
                                                                    if (AFD < 9) {
                                                                        if (Cs < 421) {
                                                                            if (negBalance < 51) {
                                                                                return 4;
                                                                            }
                                                                            else if (negBalance >= 51) {
                                                                                return 8;
                                                                            }
                                                                        }
                                                                        else if (Cs >= 421) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                    else if (AFD >= 9) {
                                                                        return 4;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (TSC >= 8) {
                                                        if (SC < 3185) {
                                                            if (negBalance < 52) {
                                                                return 8;
                                                            }
                                                            else if (negBalance >= 52) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (SC >= 3185) {
                                                            if (AFD < 11) {
                                                                if (RsDiff < 6) {
                                                                    return 8;
                                                                }
                                                                else if (RsDiff >= 6) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (AFD >= 11) {
                                                                if (Rs < 2946) {
                                                                    return 8;
                                                                }
                                                                else if (Rs >= 2946) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (Cs >= 1685) {
                                                    if (ssDCval < 107) {
                                                        if (posBalance < 51) {
                                                            if (SC < 6062) {
                                                                if (tcor < 93) {
                                                                    if (refDCval < 106) {
                                                                        if (ssDCval < 102) {
                                                                            return 8;
                                                                        }
                                                                        else if (ssDCval >= 102) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                    else if (refDCval >= 106) {
                                                                        return 2;
                                                                    }
                                                                }
                                                                else if (tcor >= 93) {
                                                                    if (AFD < 10) {
                                                                        return 8;
                                                                    }
                                                                    else if (AFD >= 10) {
                                                                        return 2;
                                                                    }
                                                                }
                                                            }
                                                            else if (SC >= 6062) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (posBalance >= 51) {
                                                            if (mu_mv_mag_sq < 72) {
                                                                if (gchDC < 1) {
                                                                    if (SC < 6672) {
                                                                        return 8;
                                                                    }
                                                                    else if (SC >= 6672) {
                                                                        return 2;
                                                                    }
                                                                }
                                                                else if (gchDC >= 1) {
                                                                    if (TSC < 6) {
                                                                        return 8;
                                                                    }
                                                                    else if (TSC >= 6) {
                                                                        return 4;
                                                                    }
                                                                }
                                                            }
                                                            else if (mu_mv_mag_sq >= 72) {
                                                                if (Cs < 2516) {
                                                                    return 1;
                                                                }
                                                                else if (Cs >= 2516) {
                                                                    return 2;
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (ssDCval >= 107) {
                                                        if (Cs < 2069) {
                                                            if (MVDiff < 7) {
                                                                if (ssDCval < 117) {
                                                                    return 4;
                                                                }
                                                                else if (ssDCval >= 117) {
                                                                    if (SC < 4739) {
                                                                        if (diffRsCsdiff < -20) {
                                                                            if (Cs < 1797) {
                                                                                return 8;
                                                                            }
                                                                            else if (Cs >= 1797) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                        else if (diffRsCsdiff >= -20) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                    else if (SC >= 4739) {
                                                                        return 4;
                                                                    }
                                                                }
                                                            }
                                                            else if (MVDiff >= 7) {
                                                                if (RsDiff < 10) {
                                                                    return 4;
                                                                }
                                                                else if (RsDiff >= 10) {
                                                                    if (ssDCval < 109) {
                                                                        return 8;
                                                                    }
                                                                    else if (ssDCval >= 109) {
                                                                        return 4;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (Cs >= 2069) {
                                                            if (tcor < 93) {
                                                                if (MVDiff < 19) {
                                                                    if (Cs < 2597) {
                                                                        if (mu_mv_mag_sq < 68) {
                                                                            return 4;
                                                                        }
                                                                        else if (mu_mv_mag_sq >= 68) {
                                                                            if (tcor < 89) {
                                                                                if (Rs < 2271) {
                                                                                    if (Rs < 1983) {
                                                                                        return 4;
                                                                                    }
                                                                                    else if (Rs >= 1983) {
                                                                                        if (AFD < 11) {
                                                                                            return 8;
                                                                                        }
                                                                                        else if (AFD >= 11) {
                                                                                            if (Rs < 2123) {
                                                                                                return 8;
                                                                                            }
                                                                                            else if (Rs >= 2123) {
                                                                                                return 4;
                                                                                            }
                                                                                        }
                                                                                    }
                                                                                }
                                                                                else if (Rs >= 2271) {
                                                                                    return 4;
                                                                                }
                                                                            }
                                                                            else if (tcor >= 89) {
                                                                                if (ssDCval < 114) {
                                                                                    if (Rs < 4237) {
                                                                                        return 2;
                                                                                    }
                                                                                    else if (Rs >= 4237) {
                                                                                        return 4;
                                                                                    }
                                                                                }
                                                                                else if (ssDCval >= 114) {
                                                                                    return 8;
                                                                                }
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (Cs >= 2597) {
                                                                        if (Rs < 4627) {
                                                                            if (mu_mv_mag_sq < 59) {
                                                                                if (MVDiff < 12) {
                                                                                    return 8;
                                                                                }
                                                                                else if (MVDiff >= 12) {
                                                                                    return 1;
                                                                                }
                                                                            }
                                                                            else if (mu_mv_mag_sq >= 59) {
                                                                                if (RsDiff < 11) {
                                                                                    return 8;
                                                                                }
                                                                                else if (RsDiff >= 11) {
                                                                                    return 4;
                                                                                }
                                                                            }
                                                                        }
                                                                        else if (Rs >= 4627) {
                                                                            if (Rs < 5143) {
                                                                                return 2;
                                                                            }
                                                                            else if (Rs >= 5143) {
                                                                                if (posBalance < 47) {
                                                                                    return 8;
                                                                                }
                                                                                else if (posBalance >= 47) {
                                                                                    if (Cs < 3134) {
                                                                                        return 4;
                                                                                    }
                                                                                    else if (Cs >= 3134) {
                                                                                        return 2;
                                                                                    }
                                                                                }
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                                else if (MVDiff >= 19) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (tcor >= 93) {
                                                                if (Rs < 4359) {
                                                                    if (AFD < 8) {
                                                                        return 4;
                                                                    }
                                                                    else if (AFD >= 8) {
                                                                        if (MVDiff < 19) {
                                                                            return 2;
                                                                        }
                                                                        else if (MVDiff >= 19) {
                                                                            if (diffRsCsdiff < 16) {
                                                                                if (Rs < 3526) {
                                                                                    return 2;
                                                                                }
                                                                                else if (Rs >= 3526) {
                                                                                    return 4;
                                                                                }
                                                                            }
                                                                            else if (diffRsCsdiff >= 16) {
                                                                                return 2;
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                                else if (Rs >= 4359) {
                                                                    if (Cs < 2531) {
                                                                        if (CsDiff < 6) {
                                                                            return 8;
                                                                        }
                                                                        else if (CsDiff >= 6) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                    else if (Cs >= 2531) {
                                                                        return 8;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (ssDCval >= 125) {
                                            if (Rs < 4605) {
                                                if (mcTcor < 96) {
                                                    if (diffAFD < -2) {
                                                        return 1;
                                                    }
                                                    else if (diffAFD >= -2) {
                                                        if (TSC < 10) {
                                                            if (mcTcor < 84) {
                                                                return 1;
                                                            }
                                                            else if (mcTcor >= 84) {
                                                                if (SC < 3759) {
                                                                    if (refDCval < 136) {
                                                                        return 1;
                                                                    }
                                                                    else if (refDCval >= 136) {
                                                                        if (Rs < 684) {
                                                                            return 4;
                                                                        }
                                                                        else if (Rs >= 684) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                }
                                                                else if (SC >= 3759) {
                                                                    if (diffRsCsdiff < -43) {
                                                                        return 2;
                                                                    }
                                                                    else if (diffRsCsdiff >= -43) {
                                                                        if (mu_mv_mag_sq < 78) {
                                                                            return 4;
                                                                        }
                                                                        else if (mu_mv_mag_sq >= 78) {
                                                                            return 2;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (TSC >= 10) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                                else if (mcTcor >= 96) {
                                                    if (AFD < 11) {
                                                        return 4;
                                                    }
                                                    else if (AFD >= 11) {
                                                        if (Cs < 1087) {
                                                            return 4;
                                                        }
                                                        else if (Cs >= 1087) {
                                                            return 2;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (Rs >= 4605) {
                                                if (SC < 7101) {
                                                    return 4;
                                                }
                                                else if (SC >= 7101) {
                                                    return 2;
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                        }
                        else if (AFD >= 12) {
                            if (Cs < 1420) {
                                if (RsDiff < 6) {
                                    if (MVDiff < 16) {
                                        if (negBalance < 50) {
                                            if (Rs < 763) {
                                                return 4;
                                            }
                                            else if (Rs >= 763) {
                                                if (ssDCval < 118) {
                                                    return 2;
                                                }
                                                else if (ssDCval >= 118) {
                                                    return 1;
                                                }
                                            }
                                        }
                                        else if (negBalance >= 50) {
                                            if (RsDiff < 5) {
                                                return 4;
                                            }
                                            else if (RsDiff >= 5) {
                                                if (SC < 2829) {
                                                    if (AFD < 13) {
                                                        if (Rs < 1120) {
                                                            return 2;
                                                        }
                                                        else if (Rs >= 1120) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (AFD >= 13) {
                                                        return 8;
                                                    }
                                                }
                                                else if (SC >= 2829) {
                                                    return 1;
                                                }
                                            }
                                        }
                                    }
                                    else if (MVDiff >= 16) {
                                        if (Rs < 690) {
                                            if (tcor < 43) {
                                                if (TSC < 21) {
                                                    return 1;
                                                }
                                                else if (TSC >= 21) {
                                                    if (Rs < 234) {
                                                        return 8;
                                                    }
                                                    else if (Rs >= 234) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                            else if (tcor >= 43) {
                                                if (TSC < 7) {
                                                    if (diffAFD < 5) {
                                                        return 2;
                                                    }
                                                    else if (diffAFD >= 5) {
                                                        return 4;
                                                    }
                                                }
                                                else if (TSC >= 7) {
                                                    if (ssDCval < 76) {
                                                        if (TSC < 10) {
                                                            return 2;
                                                        }
                                                        else if (TSC >= 10) {
                                                            return 1;
                                                        }
                                                    }
                                                    else if (ssDCval >= 76) {
                                                        if (ssDCval < 111) {
                                                            if (refDCval < 111) {
                                                                return 1;
                                                            }
                                                            else if (refDCval >= 111) {
                                                                return 2;
                                                            }
                                                        }
                                                        else if (ssDCval >= 111) {
                                                            if (ssDCval < 176) {
                                                                return 2;
                                                            }
                                                            else if (ssDCval >= 176) {
                                                                return 1;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (Rs >= 690) {
                                            if (mu_mv_mag_sq < 40) {
                                                if (mu_mv_mag_sq < 10) {
                                                    if (Rs < 1389) {
                                                        if (Cs < 562) {
                                                            return 2;
                                                        }
                                                        else if (Cs >= 562) {
                                                            if (AFD < 17) {
                                                                if (Rs < 913) {
                                                                    return 2;
                                                                }
                                                                else if (Rs >= 913) {
                                                                    return 1;
                                                                }
                                                            }
                                                            else if (AFD >= 17) {
                                                                return 2;
                                                            }
                                                        }
                                                    }
                                                    else if (Rs >= 1389) {
                                                        return 8;
                                                    }
                                                }
                                                else if (mu_mv_mag_sq >= 10) {
                                                    if (Cs < 789) {
                                                        return 8;
                                                    }
                                                    else if (Cs >= 789) {
                                                        if (Cs < 950) {
                                                            return 1;
                                                        }
                                                        else if (Cs >= 950) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (mu_mv_mag_sq >= 40) {
                                                if (Cs < 842) {
                                                    return 8;
                                                }
                                                else if (Cs >= 842) {
                                                    return 4;
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (RsDiff >= 6) {
                                    if (MVDiff < 50) {
                                        if (mcTcor < 85) {
                                            if (Rs < 1583) {
                                                if (ssDCval < 64) {
                                                    if (SC < 2548) {
                                                        return 2;
                                                    }
                                                    else if (SC >= 2548) {
                                                        return 1;
                                                    }
                                                }
                                                else if (ssDCval >= 64) {
                                                    return 8;
                                                }
                                            }
                                            else if (Rs >= 1583) {
                                                if (TSC < 11) {
                                                    if (Cs < 661) {
                                                        return 4;
                                                    }
                                                    else if (Cs >= 661) {
                                                        return 1;
                                                    }
                                                }
                                                else if (TSC >= 11) {
                                                    if (Cs < 503) {
                                                        return 4;
                                                    }
                                                    else if (Cs >= 503) {
                                                        if (ssDCval < 125) {
                                                            return 2;
                                                        }
                                                        else if (ssDCval >= 125) {
                                                            if (AFD < 20) {
                                                                return 2;
                                                            }
                                                            else if (AFD >= 20) {
                                                                return 1;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (mcTcor >= 85) {
                                            if (negBalance < 45) {
                                                if (mcTcor < 88) {
                                                    return 1;
                                                }
                                                else if (mcTcor >= 88) {
                                                    if (Rs < 2204) {
                                                        if (ssDCval < 65) {
                                                            return 1;
                                                        }
                                                        else if (ssDCval >= 65) {
                                                            if (posBalance < 64) {
                                                                return 2;
                                                            }
                                                            else if (posBalance >= 64) {
                                                                if (Cs < 1051) {
                                                                    return 2;
                                                                }
                                                                else if (Cs >= 1051) {
                                                                    if (Rs < 1757) {
                                                                        return 8;
                                                                    }
                                                                    else if (Rs >= 1757) {
                                                                        return 4;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (Rs >= 2204) {
                                                        if (RsCsDiff < 53) {
                                                            return 2;
                                                        }
                                                        else if (RsCsDiff >= 53) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (negBalance >= 45) {
                                                if (mcTcor < 96) {
                                                    if (Cs < 1401) {
                                                        if (tcor < 77) {
                                                            if (mu_mv_mag_sq < 7) {
                                                                if (MVDiff < 41) {
                                                                    return 1;
                                                                }
                                                                else if (MVDiff >= 41) {
                                                                    return 2;
                                                                }
                                                            }
                                                            else if (mu_mv_mag_sq >= 7) {
                                                                if (diffTSC < -3) {
                                                                    return 1;
                                                                }
                                                                else if (diffTSC >= -3) {
                                                                    return 2;
                                                                }
                                                            }
                                                        }
                                                        else if (tcor >= 77) {
                                                            if (mu_mv_mag_sq < 43) {
                                                                if (refDCval < 75) {
                                                                    if (ssDCval < 73) {
                                                                        if (Cs < 816) {
                                                                            return 2;
                                                                        }
                                                                        else if (Cs >= 816) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                    else if (ssDCval >= 73) {
                                                                        return 1;
                                                                    }
                                                                }
                                                                else if (refDCval >= 75) {
                                                                    return 2;
                                                                }
                                                            }
                                                            else if (mu_mv_mag_sq >= 43) {
                                                                if (Cs < 638) {
                                                                    return 4;
                                                                }
                                                                else if (Cs >= 638) {
                                                                    return 2;
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (Cs >= 1401) {
                                                        if (diffMVdiffVal < -1) {
                                                            return 2;
                                                        }
                                                        else if (diffMVdiffVal >= -1) {
                                                            if (TSC < 9) {
                                                                return 2;
                                                            }
                                                            else if (TSC >= 9) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (mcTcor >= 96) {
                                                    if (Rs < 3284) {
                                                        if (diffAFD < -2) {
                                                            return 8;
                                                        }
                                                        else if (diffAFD >= -2) {
                                                            if (MVDiff < 9) {
                                                                return 4;
                                                            }
                                                            else if (MVDiff >= 9) {
                                                                return 2;
                                                            }
                                                        }
                                                    }
                                                    else if (Rs >= 3284) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (MVDiff >= 50) {
                                        if (ssDCval < 85) {
                                            if (ssDCval < 58) {
                                                if (mu_mv_mag_sq < 11) {
                                                    return 2;
                                                }
                                                else if (mu_mv_mag_sq >= 11) {
                                                    if (Rs < 1047) {
                                                        return 2;
                                                    }
                                                    else if (Rs >= 1047) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                            else if (ssDCval >= 58) {
                                                if (diffRsCsdiff < -34) {
                                                    if (Cs < 1210) {
                                                        return 1;
                                                    }
                                                    else if (Cs >= 1210) {
                                                        if (AFD < 14) {
                                                            return 4;
                                                        }
                                                        else if (AFD >= 14) {
                                                            return 2;
                                                        }
                                                    }
                                                }
                                                else if (diffRsCsdiff >= -34) {
                                                    if (AFD < 38) {
                                                        if (mcTcor < 81) {
                                                            return 8;
                                                        }
                                                        else if (mcTcor >= 81) {
                                                            if (Rs < 1025) {
                                                                return 1;
                                                            }
                                                            else if (Rs >= 1025) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                    else if (AFD >= 38) {
                                                        if (Rs < 1210) {
                                                            return 1;
                                                        }
                                                        else if (Rs >= 1210) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (ssDCval >= 85) {
                                            if (Cs < 660) {
                                                if (Rs < 434) {
                                                    if (Rs < 367) {
                                                        return 1;
                                                    }
                                                    else if (Rs >= 367) {
                                                        return 8;
                                                    }
                                                }
                                                else if (Rs >= 434) {
                                                    if (AFD < 19) {
                                                        return 4;
                                                    }
                                                    else if (AFD >= 19) {
                                                        if (tcor < 23) {
                                                            return 4;
                                                        }
                                                        else if (tcor >= 23) {
                                                            return 2;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (Cs >= 660) {
                                                if (mu_mv_mag_sq < 13) {
                                                    if (posBalance < 70) {
                                                        if (mcTcor < 73) {
                                                            if (TSC < 17) {
                                                                return 2;
                                                            }
                                                            else if (TSC >= 17) {
                                                                return 1;
                                                            }
                                                        }
                                                        else if (mcTcor >= 73) {
                                                            if (Cs < 1108) {
                                                                return 1;
                                                            }
                                                            else if (Cs >= 1108) {
                                                                if (tcor < 73) {
                                                                    return 1;
                                                                }
                                                                else if (tcor >= 73) {
                                                                    return 2;
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (posBalance >= 70) {
                                                        return 2;
                                                    }
                                                }
                                                else if (mu_mv_mag_sq >= 13) {
                                                    if (Rs < 1235) {
                                                        if (tcor < 7) {
                                                            return 4;
                                                        }
                                                        else if (tcor >= 7) {
                                                            return 1;
                                                        }
                                                    }
                                                    else if (Rs >= 1235) {
                                                        if (mu_mv_mag_sq < 28) {
                                                            if (negBalance < 66) {
                                                                if (diffMVdiffVal < -8) {
                                                                    if (ssDCval < 100) {
                                                                        return 2;
                                                                    }
                                                                    else if (ssDCval >= 100) {
                                                                        return 1;
                                                                    }
                                                                }
                                                                else if (diffMVdiffVal >= -8) {
                                                                    if (diffTSC < -1) {
                                                                        return 1;
                                                                    }
                                                                    else if (diffTSC >= -1) {
                                                                        return 2;
                                                                    }
                                                                }
                                                            }
                                                            else if (negBalance >= 66) {
                                                                return 1;
                                                            }
                                                        }
                                                        else if (mu_mv_mag_sq >= 28) {
                                                            if (AFD < 21) {
                                                                return 4;
                                                            }
                                                            else if (AFD >= 21) {
                                                                return 2;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                            else if (Cs >= 1420) {
                                if (refDCval < 81) {
                                    if (negBalance < 43) {
                                        if (gchDC < 3) {
                                            if (ssDCval < 78) {
                                                if (mcTcor < 82) {
                                                    return 8;
                                                }
                                                else if (mcTcor >= 82) {
                                                    return 1;
                                                }
                                            }
                                            else if (ssDCval >= 78) {
                                                if (TSC < 9) {
                                                    return 1;
                                                }
                                                else if (TSC >= 9) {
                                                    if (tcor < 88) {
                                                        if (AFD < 17) {
                                                            return 8;
                                                        }
                                                        else if (AFD >= 17) {
                                                            return 1;
                                                        }
                                                    }
                                                    else if (tcor >= 88) {
                                                        return 1;
                                                    }
                                                }
                                            }
                                        }
                                        else if (gchDC >= 3) {
                                            if (SC < 4366) {
                                                if (Rs < 1520) {
                                                    return 2;
                                                }
                                                else if (Rs >= 1520) {
                                                    if (Rs < 2028) {
                                                        return 1;
                                                    }
                                                    else if (Rs >= 2028) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                            else if (SC >= 4366) {
                                                if (diffRsCsdiff < 16) {
                                                    return 4;
                                                }
                                                else if (diffRsCsdiff >= 16) {
                                                    return 8;
                                                }
                                            }
                                        }
                                    }
                                    else if (negBalance >= 43) {
                                        if (mu_mv_mag_sq < 6) {
                                            if (TSC < 12) {
                                                if (negBalance < 48) {
                                                    if (Rs < 1927) {
                                                        if (Cs < 1777) {
                                                            return 4;
                                                        }
                                                        else if (Cs >= 1777) {
                                                            return 1;
                                                        }
                                                    }
                                                    else if (Rs >= 1927) {
                                                        if (Cs < 2524) {
                                                            return 8;
                                                        }
                                                        else if (Cs >= 2524) {
                                                            return 1;
                                                        }
                                                    }
                                                }
                                                else if (negBalance >= 48) {
                                                    if (CsDiff < 7) {
                                                        if (SC < 3072) {
                                                            return 8;
                                                        }
                                                        else if (SC >= 3072) {
                                                            return 1;
                                                        }
                                                    }
                                                    else if (CsDiff >= 7) {
                                                        if (tcor < 90) {
                                                            if (diffAFD < -4) {
                                                                return 2;
                                                            }
                                                            else if (diffAFD >= -4) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (tcor >= 90) {
                                                            return 2;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (TSC >= 12) {
                                                if (tcor < 66) {
                                                    if (RsDiff < 15) {
                                                        return 4;
                                                    }
                                                    else if (RsDiff >= 15) {
                                                        return 2;
                                                    }
                                                }
                                                else if (tcor >= 66) {
                                                    if (negBalance < 52) {
                                                        if (MVDiff < 23) {
                                                            return 4;
                                                        }
                                                        else if (MVDiff >= 23) {
                                                            if (Rs < 2518) {
                                                                if (Cs < 2032) {
                                                                    return 8;
                                                                }
                                                                else if (Cs >= 2032) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (Rs >= 2518) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                    else if (negBalance >= 52) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                        }
                                        else if (mu_mv_mag_sq >= 6) {
                                            if (refDCval < 78) {
                                                if (Rs < 2971) {
                                                    if (AFD < 18) {
                                                        if (diffMVdiffVal < 5) {
                                                            return 8;
                                                        }
                                                        else if (diffMVdiffVal >= 5) {
                                                            if (RsCsDiff < 67) {
                                                                return 8;
                                                            }
                                                            else if (RsCsDiff >= 67) {
                                                                if (SC < 3955) {
                                                                    return 4;
                                                                }
                                                                else if (SC >= 3955) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (AFD >= 18) {
                                                        if (MVDiff < 48) {
                                                            if (Rs < 1731) {
                                                                return 4;
                                                            }
                                                            else if (Rs >= 1731) {
                                                                if (TSC < 11) {
                                                                    return 4;
                                                                }
                                                                else if (TSC >= 11) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                        else if (MVDiff >= 48) {
                                                            if (SC < 4334) {
                                                                return 1;
                                                            }
                                                            else if (SC >= 4334) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (Rs >= 2971) {
                                                    if (AFD < 17) {
                                                        if (negBalance < 54) {
                                                            return 4;
                                                        }
                                                        else if (negBalance >= 54) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (AFD >= 17) {
                                                        if (tcor < 57) {
                                                            return 2;
                                                        }
                                                        else if (tcor >= 57) {
                                                            return 1;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (refDCval >= 78) {
                                                if (negBalance < 47) {
                                                    if (RsDiff < 12) {
                                                        return 8;
                                                    }
                                                    else if (RsDiff >= 12) {
                                                        return 1;
                                                    }
                                                }
                                                else if (negBalance >= 47) {
                                                    if (mcTcor < 94) {
                                                        if (mcTcor < 87) {
                                                            if (Rs < 2166) {
                                                                return 1;
                                                            }
                                                            else if (Rs >= 2166) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (mcTcor >= 87) {
                                                            if (Rs < 2446) {
                                                                return 1;
                                                            }
                                                            else if (Rs >= 2446) {
                                                                if (SC < 5144) {
                                                                    if (AFD < 14) {
                                                                        return 1;
                                                                    }
                                                                    else if (AFD >= 14) {
                                                                        return 2;
                                                                    }
                                                                }
                                                                else if (SC >= 5144) {
                                                                    return 1;
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (mcTcor >= 94) {
                                                        return 2;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (refDCval >= 81) {
                                    if (tcor < 84) {
                                        if (tcor < 69) {
                                            if (AFD < 30) {
                                                if (posBalance < 47) {
                                                    if (mcTcor < 81) {
                                                        if (SC < 7460) {
                                                            return 8;
                                                        }
                                                        else if (SC >= 7460) {
                                                            return 1;
                                                        }
                                                    }
                                                    else if (mcTcor >= 81) {
                                                        if (Cs < 1825) {
                                                            return 4;
                                                        }
                                                        else if (Cs >= 1825) {
                                                            return 2;
                                                        }
                                                    }
                                                }
                                                else if (posBalance >= 47) {
                                                    if (refDCval < 112) {
                                                        if (Cs < 2589) {
                                                            if (refDCval < 103) {
                                                                if (TSC < 15) {
                                                                    return 1;
                                                                }
                                                                else if (TSC >= 15) {
                                                                    if (Cs < 2322) {
                                                                        return 8;
                                                                    }
                                                                    else if (Cs >= 2322) {
                                                                        return 4;
                                                                    }
                                                                }
                                                            }
                                                            else if (refDCval >= 103) {
                                                                if (tcor < 63) {
                                                                    return 2;
                                                                }
                                                                else if (tcor >= 63) {
                                                                    if (AFD < 23) {
                                                                        if (Cs < 1778) {
                                                                            return 2;
                                                                        }
                                                                        else if (Cs >= 1778) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                    else if (AFD >= 23) {
                                                                        return 1;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (Cs >= 2589) {
                                                            if (Rs < 3123) {
                                                                return 4;
                                                            }
                                                            else if (Rs >= 3123) {
                                                                if (Rs < 4779) {
                                                                    return 8;
                                                                }
                                                                else if (Rs >= 4779) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (refDCval >= 112) {
                                                        if (posBalance < 66) {
                                                            if (refDCval < 140) {
                                                                if (ssDCval < 128) {
                                                                    if (diffMVdiffVal < 11) {
                                                                        if (mu_mv_mag_sq < 33) {
                                                                            return 1;
                                                                        }
                                                                        else if (mu_mv_mag_sq >= 33) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                    else if (diffMVdiffVal >= 11) {
                                                                        return 2;
                                                                    }
                                                                }
                                                                else if (ssDCval >= 128) {
                                                                    return 1;
                                                                }
                                                            }
                                                            else if (refDCval >= 140) {
                                                                if (TSC < 11) {
                                                                    return 2;
                                                                }
                                                                else if (TSC >= 11) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                        else if (posBalance >= 66) {
                                                            if (Cs < 1612) {
                                                                return 8;
                                                            }
                                                            else if (Cs >= 1612) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                            else if (AFD >= 30) {
                                                if (TSC < 17) {
                                                    if (refDCval < 149) {
                                                        if (gchDC < 7) {
                                                            return 2;
                                                        }
                                                        else if (gchDC >= 7) {
                                                            return 1;
                                                        }
                                                    }
                                                    else if (refDCval >= 149) {
                                                        if (MVDiff < 88) {
                                                            return 4;
                                                        }
                                                        else if (MVDiff >= 88) {
                                                            if (Rs < 3184) {
                                                                return 4;
                                                            }
                                                            else if (Rs >= 3184) {
                                                                return 2;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (TSC >= 17) {
                                                    if (gchDC < 6) {
                                                        return 1;
                                                    }
                                                    else if (gchDC >= 6) {
                                                        if (tcor < 6) {
                                                            if (Rs < 2959) {
                                                                return 4;
                                                            }
                                                            else if (Rs >= 2959) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (tcor >= 6) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (tcor >= 69) {
                                            if (refDCval < 139) {
                                                if (refDCval < 84) {
                                                    if (SC < 4328) {
                                                        if (TSC < 10) {
                                                            return 1;
                                                        }
                                                        else if (TSC >= 10) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (SC >= 4328) {
                                                        if (Rs < 2555) {
                                                            if (TSC < 10) {
                                                                return 8;
                                                            }
                                                            else if (TSC >= 10) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (Rs >= 2555) {
                                                            if (diffAFD < 1) {
                                                                return 1;
                                                            }
                                                            else if (diffAFD >= 1) {
                                                                if (AFD < 18) {
                                                                    return 1;
                                                                }
                                                                else if (AFD >= 18) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (refDCval >= 84) {
                                                    if (ssDCval < 133) {
                                                        if (CsDiff < 12) {
                                                            if (Rs < 4557) {
                                                                if (tcor < 80) {
                                                                    if (TSC < 10) {
                                                                        if (ssDCval < 91) {
                                                                            return 2;
                                                                        }
                                                                        else if (ssDCval >= 91) {
                                                                            if (mcTcor < 91) {
                                                                                if (ssDCval < 122) {
                                                                                    if (posBalance < 46) {
                                                                                        return 8;
                                                                                    }
                                                                                    else if (posBalance >= 46) {
                                                                                        return 4;
                                                                                    }
                                                                                }
                                                                                else if (ssDCval >= 122) {
                                                                                    if (Rs < 1599) {
                                                                                        return 8;
                                                                                    }
                                                                                    else if (Rs >= 1599) {
                                                                                        return 1;
                                                                                    }
                                                                                }
                                                                            }
                                                                            else if (mcTcor >= 91) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (TSC >= 10) {
                                                                        if (mu_mv_mag_sq < 3) {
                                                                            if (negBalance < 51) {
                                                                                if (refDCval < 119) {
                                                                                    if (Rs < 3267) {
                                                                                        return 4;
                                                                                    }
                                                                                    else if (Rs >= 3267) {
                                                                                        return 2;
                                                                                    }
                                                                                }
                                                                                else if (refDCval >= 119) {
                                                                                    return 1;
                                                                                }
                                                                            }
                                                                            else if (negBalance >= 51) {
                                                                                if (mcTcor < 88) {
                                                                                    return 4;
                                                                                }
                                                                                else if (mcTcor >= 88) {
                                                                                    if (diffAFD < -1) {
                                                                                        return 1;
                                                                                    }
                                                                                    else if (diffAFD >= -1) {
                                                                                        return 2;
                                                                                    }
                                                                                }
                                                                            }
                                                                        }
                                                                        else if (mu_mv_mag_sq >= 3) {
                                                                            if (AFD < 19) {
                                                                                if (negBalance < 39) {
                                                                                    if (Cs < 2612) {
                                                                                        return 4;
                                                                                    }
                                                                                    else if (Cs >= 2612) {
                                                                                        return 1;
                                                                                    }
                                                                                }
                                                                                else if (negBalance >= 39) {
                                                                                    if (ssDCval < 88) {
                                                                                        return 1;
                                                                                    }
                                                                                    else if (ssDCval >= 88) {
                                                                                        if (mu_mv_mag_sq < 7) {
                                                                                            if (Rs < 2177) {
                                                                                                return 4;
                                                                                            }
                                                                                            else if (Rs >= 2177) {
                                                                                                if (Rs < 2867) {
                                                                                                    return 1;
                                                                                                }
                                                                                                else if (Rs >= 2867) {
                                                                                                    return 2;
                                                                                                }
                                                                                            }
                                                                                        }
                                                                                        else if (mu_mv_mag_sq >= 7) {
                                                                                            if (CsDiff < 11) {
                                                                                                return 4;
                                                                                            }
                                                                                            else if (CsDiff >= 11) {
                                                                                                return 1;
                                                                                            }
                                                                                        }
                                                                                    }
                                                                                }
                                                                            }
                                                                            else if (AFD >= 19) {
                                                                                if (RsDiff < 16) {
                                                                                    if (SC < 6448) {
                                                                                        if (Cs < 1724) {
                                                                                            return 4;
                                                                                        }
                                                                                        else if (Cs >= 1724) {
                                                                                            return 1;
                                                                                        }
                                                                                    }
                                                                                    else if (SC >= 6448) {
                                                                                        return 4;
                                                                                    }
                                                                                }
                                                                                else if (RsDiff >= 16) {
                                                                                    return 8;
                                                                                }
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                                else if (tcor >= 80) {
                                                                    if (TSC < 15) {
                                                                        if (ssDCval < 113) {
                                                                            if (refDCval < 106) {
                                                                                if (posBalance < 55) {
                                                                                    if (MVDiff < 9) {
                                                                                        return 4;
                                                                                    }
                                                                                    else if (MVDiff >= 9) {
                                                                                        if (mcTcor < 87) {
                                                                                            return 2;
                                                                                        }
                                                                                        else if (mcTcor >= 87) {
                                                                                            if (ssDCval < 98) {
                                                                                                return 1;
                                                                                            }
                                                                                            else if (ssDCval >= 98) {
                                                                                                return 4;
                                                                                            }
                                                                                        }
                                                                                    }
                                                                                }
                                                                                else if (posBalance >= 55) {
                                                                                    if (negBalance < 55) {
                                                                                        if (refDCval < 92) {
                                                                                            if (negBalance < 51) {
                                                                                                if (Rs < 2933) {
                                                                                                    return 2;
                                                                                                }
                                                                                                else if (Rs >= 2933) {
                                                                                                    return 1;
                                                                                                }
                                                                                            }
                                                                                            else if (negBalance >= 51) {
                                                                                                if (refDCval < 87) {
                                                                                                    return 4;
                                                                                                }
                                                                                                else if (refDCval >= 87) {
                                                                                                    if (Cs < 2545) {
                                                                                                        return 2;
                                                                                                    }
                                                                                                    else if (Cs >= 2545) {
                                                                                                        return 4;
                                                                                                    }
                                                                                                }
                                                                                            }
                                                                                        }
                                                                                        else if (refDCval >= 92) {
                                                                                            if (diffAFD < -1) {
                                                                                                if (SC < 5588) {
                                                                                                    return 1;
                                                                                                }
                                                                                                else if (SC >= 5588) {
                                                                                                    return 4;
                                                                                                }
                                                                                            }
                                                                                            else if (diffAFD >= -1) {
                                                                                                return 4;
                                                                                            }
                                                                                        }
                                                                                    }
                                                                                    else if (negBalance >= 55) {
                                                                                        return 2;
                                                                                    }
                                                                                }
                                                                            }
                                                                            else if (refDCval >= 106) {
                                                                                if (MVDiff < 13) {
                                                                                    return 1;
                                                                                }
                                                                                else if (MVDiff >= 13) {
                                                                                    if (mcTcor < 93) {
                                                                                        return 8;
                                                                                    }
                                                                                    else if (mcTcor >= 93) {
                                                                                        return 2;
                                                                                    }
                                                                                }
                                                                            }
                                                                        }
                                                                        else if (ssDCval >= 113) {
                                                                            if (ssDCval < 117) {
                                                                                if (RsCsDiff < 192) {
                                                                                    if (TSC < 10) {
                                                                                        return 2;
                                                                                    }
                                                                                    else if (TSC >= 10) {
                                                                                        return 1;
                                                                                    }
                                                                                }
                                                                                else if (RsCsDiff >= 192) {
                                                                                    return 2;
                                                                                }
                                                                            }
                                                                            else if (ssDCval >= 117) {
                                                                                if (RsDiff < 14) {
                                                                                    if (AFD < 21) {
                                                                                        if (Rs < 3037) {
                                                                                            if (tcor < 83) {
                                                                                                return 4;
                                                                                            }
                                                                                            else if (tcor >= 83) {
                                                                                                if (Cs < 1756) {
                                                                                                    return 4;
                                                                                                }
                                                                                                else if (Cs >= 1756) {
                                                                                                    return 1;
                                                                                                }
                                                                                            }
                                                                                        }
                                                                                        else if (Rs >= 3037) {
                                                                                            if (Rs < 3802) {
                                                                                                return 2;
                                                                                            }
                                                                                            else if (Rs >= 3802) {
                                                                                                return 4;
                                                                                            }
                                                                                        }
                                                                                    }
                                                                                    else if (AFD >= 21) {
                                                                                        return 2;
                                                                                    }
                                                                                }
                                                                                else if (RsDiff >= 14) {
                                                                                    if (AFD < 19) {
                                                                                        if (mcTcor < 89) {
                                                                                            return 4;
                                                                                        }
                                                                                        else if (mcTcor >= 89) {
                                                                                            return 2;
                                                                                        }
                                                                                    }
                                                                                    else if (AFD >= 19) {
                                                                                        return 2;
                                                                                    }
                                                                                }
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (TSC >= 15) {
                                                                        return 1;
                                                                    }
                                                                }
                                                            }
                                                            else if (Rs >= 4557) {
                                                                if (ssDCval < 106) {
                                                                    if (Rs < 4598) {
                                                                        return 2;
                                                                    }
                                                                    else if (Rs >= 4598) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (ssDCval >= 106) {
                                                                    if (TSC < 11) {
                                                                        if (diffRsCsdiff < 4) {
                                                                            if (RsCsDiff < 244) {
                                                                                if (diffRsCsdiff < -11) {
                                                                                    return 2;
                                                                                }
                                                                                else if (diffRsCsdiff >= -11) {
                                                                                    if (Rs < 5112) {
                                                                                        return 4;
                                                                                    }
                                                                                    else if (Rs >= 5112) {
                                                                                        return 2;
                                                                                    }
                                                                                }
                                                                            }
                                                                            else if (RsCsDiff >= 244) {
                                                                                if (Cs < 2624) {
                                                                                    return 4;
                                                                                }
                                                                                else if (Cs >= 2624) {
                                                                                    return 1;
                                                                                }
                                                                            }
                                                                        }
                                                                        else if (diffRsCsdiff >= 4) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                    else if (TSC >= 11) {
                                                                        if (Rs < 6048) {
                                                                            return 2;
                                                                        }
                                                                        else if (Rs >= 6048) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (CsDiff >= 12) {
                                                            if (SC < 4596) {
                                                                if (gchDC < 2) {
                                                                    return 2;
                                                                }
                                                                else if (gchDC >= 2) {
                                                                    if (RsDiff < 9) {
                                                                        return 2;
                                                                    }
                                                                    else if (RsDiff >= 9) {
                                                                        return 1;
                                                                    }
                                                                }
                                                            }
                                                            else if (SC >= 4596) {
                                                                if (Cs < 2826) {
                                                                    if (TSC < 12) {
                                                                        if (Rs < 2770) {
                                                                            if (Rs < 2369) {
                                                                                return 1;
                                                                            }
                                                                            else if (Rs >= 2369) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                        else if (Rs >= 2770) {
                                                                            if (Cs < 2465) {
                                                                                return 2;
                                                                            }
                                                                            else if (Cs >= 2465) {
                                                                                return 1;
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (TSC >= 12) {
                                                                        if (negBalance < 52) {
                                                                            return 1;
                                                                        }
                                                                        else if (negBalance >= 52) {
                                                                            if (TSC < 16) {
                                                                                return 4;
                                                                            }
                                                                            else if (TSC >= 16) {
                                                                                return 1;
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                                else if (Cs >= 2826) {
                                                                    if (ssDCval < 112) {
                                                                        if (MVDiff < 11) {
                                                                            return 8;
                                                                        }
                                                                        else if (MVDiff >= 11) {
                                                                            if (Rs < 3885) {
                                                                                return 1;
                                                                            }
                                                                            else if (Rs >= 3885) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (ssDCval >= 112) {
                                                                        if (ssDCval < 114) {
                                                                            return 2;
                                                                        }
                                                                        else if (ssDCval >= 114) {
                                                                            if (negBalance < 57) {
                                                                                return 1;
                                                                            }
                                                                            else if (negBalance >= 57) {
                                                                                return 8;
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (ssDCval >= 133) {
                                                        if (MVDiff < 19) {
                                                            if (RsDiff < 14) {
                                                                return 1;
                                                            }
                                                            else if (RsDiff >= 14) {
                                                                if (refDCval < 137) {
                                                                    return 1;
                                                                }
                                                                else if (refDCval >= 137) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                        else if (MVDiff >= 19) {
                                                            if (mcTcor < 84) {
                                                                return 4;
                                                            }
                                                            else if (mcTcor >= 84) {
                                                                if (SC < 4070) {
                                                                    return 1;
                                                                }
                                                                else if (SC >= 4070) {
                                                                    return 2;
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                            else if (refDCval >= 139) {
                                                if (ssDCval < 157) {
                                                    if (diffAFD < 2) {
                                                        if (posBalance < 47) {
                                                            if (Rs < 3669) {
                                                                return 8;
                                                            }
                                                            else if (Rs >= 3669) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (posBalance >= 47) {
                                                            if (mcTcor < 93) {
                                                                return 4;
                                                            }
                                                            else if (mcTcor >= 93) {
                                                                if (posBalance < 53) {
                                                                    return 4;
                                                                }
                                                                else if (posBalance >= 53) {
                                                                    return 2;
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (diffAFD >= 2) {
                                                        if (RsDiff < 9) {
                                                            return 1;
                                                        }
                                                        else if (RsDiff >= 9) {
                                                            if (TSC < 9) {
                                                                return 2;
                                                            }
                                                            else if (TSC >= 9) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (ssDCval >= 157) {
                                                    if (posBalance < 48) {
                                                        if (CsDiff < 9) {
                                                            if (Cs < 2342) {
                                                                return 2;
                                                            }
                                                            else if (Cs >= 2342) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (CsDiff >= 9) {
                                                            if (Rs < 1251) {
                                                                return 8;
                                                            }
                                                            else if (Rs >= 1251) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                    else if (posBalance >= 48) {
                                                        if (refDCval < 157) {
                                                            return 4;
                                                        }
                                                        else if (refDCval >= 157) {
                                                            return 2;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (tcor >= 84) {
                                        if (ssDCval < 129) {
                                            if (MVDiff < 5) {
                                                if (Cs < 2367) {
                                                    if (diffMVdiffVal < -3) {
                                                        if (Rs < 2756) {
                                                            return 4;
                                                        }
                                                        else if (Rs >= 2756) {
                                                            return 1;
                                                        }
                                                    }
                                                    else if (diffMVdiffVal >= -3) {
                                                        if (diffRsCsdiff < 45) {
                                                            if (diffMVdiffVal < -2) {
                                                                if (gchDC < 1) {
                                                                    return 4;
                                                                }
                                                                else if (gchDC >= 1) {
                                                                    return 2;
                                                                }
                                                            }
                                                            else if (diffMVdiffVal >= -2) {
                                                                return 2;
                                                            }
                                                        }
                                                        else if (diffRsCsdiff >= 45) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                                else if (Cs >= 2367) {
                                                    if (ssDCval < 101) {
                                                        if (negBalance < 55) {
                                                            if (AFD < 13) {
                                                                if (Rs < 3530) {
                                                                    return 8;
                                                                }
                                                                else if (Rs >= 3530) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (AFD >= 13) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (negBalance >= 55) {
                                                            if (Rs < 2891) {
                                                                return 2;
                                                            }
                                                            else if (Rs >= 2891) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                    else if (ssDCval >= 101) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                            else if (MVDiff >= 5) {
                                                if (refDCval < 101) {
                                                    if (RsCsDiff < 87) {
                                                        if (refDCval < 89) {
                                                            if (CsDiff < 6) {
                                                                if (SC < 4578) {
                                                                    return 2;
                                                                }
                                                                else if (SC >= 4578) {
                                                                    if (SC < 4844) {
                                                                        return 1;
                                                                    }
                                                                    else if (SC >= 4844) {
                                                                        return 2;
                                                                    }
                                                                }
                                                            }
                                                            else if (CsDiff >= 6) {
                                                                if (MVDiff < 19) {
                                                                    if (Rs < 2439) {
                                                                        if (Rs < 2235) {
                                                                            return 1;
                                                                        }
                                                                        else if (Rs >= 2235) {
                                                                            return 2;
                                                                        }
                                                                    }
                                                                    else if (Rs >= 2439) {
                                                                        return 1;
                                                                    }
                                                                }
                                                                else if (MVDiff >= 19) {
                                                                    if (Rs < 2464) {
                                                                        return 1;
                                                                    }
                                                                    else if (Rs >= 2464) {
                                                                        return 2;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (refDCval >= 89) {
                                                            if (TSC < 11) {
                                                                if (Rs < 2916) {
                                                                    return 1;
                                                                }
                                                                else if (Rs >= 2916) {
                                                                    if (Cs < 2338) {
                                                                        return 1;
                                                                    }
                                                                    else if (Cs >= 2338) {
                                                                        return 4;
                                                                    }
                                                                }
                                                            }
                                                            else if (TSC >= 11) {
                                                                if (ssDCval < 92) {
                                                                    if (diffRsCsdiff < 7) {
                                                                        return 4;
                                                                    }
                                                                    else if (diffRsCsdiff >= 7) {
                                                                        return 1;
                                                                    }
                                                                }
                                                                else if (ssDCval >= 92) {
                                                                    return 1;
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (RsCsDiff >= 87) {
                                                        if (posBalance < 56) {
                                                            if (Rs < 2532) {
                                                                if (mcTcor < 92) {
                                                                    if (refDCval < 86) {
                                                                        if (AFD < 17) {
                                                                            return 8;
                                                                        }
                                                                        else if (AFD >= 17) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                    else if (refDCval >= 86) {
                                                                        if (gchDC < 2) {
                                                                            return 4;
                                                                        }
                                                                        else if (gchDC >= 2) {
                                                                            if (Cs < 2033) {
                                                                                return 4;
                                                                            }
                                                                            else if (Cs >= 2033) {
                                                                                return 8;
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                                else if (mcTcor >= 92) {
                                                                    if (negBalance < 47) {
                                                                        if (RsDiff < 10) {
                                                                            if (posBalance < 48) {
                                                                                return 1;
                                                                            }
                                                                            else if (posBalance >= 48) {
                                                                                if (TSC < 9) {
                                                                                    return 8;
                                                                                }
                                                                                else if (TSC >= 9) {
                                                                                    if (ssDCval < 93) {
                                                                                        if (diffRsCsdiff < -10) {
                                                                                            return 8;
                                                                                        }
                                                                                        else if (diffRsCsdiff >= -10) {
                                                                                            if (Cs < 2046) {
                                                                                                return 1;
                                                                                            }
                                                                                            else if (Cs >= 2046) {
                                                                                                return 8;
                                                                                            }
                                                                                        }
                                                                                    }
                                                                                    else if (ssDCval >= 93) {
                                                                                        return 1;
                                                                                    }
                                                                                }
                                                                            }
                                                                        }
                                                                        else if (RsDiff >= 10) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                    else if (negBalance >= 47) {
                                                                        if (Rs < 2290) {
                                                                            return 2;
                                                                        }
                                                                        else if (Rs >= 2290) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                            else if (Rs >= 2532) {
                                                                if (RsCsDiff < 121) {
                                                                    if (SC < 5196) {
                                                                        return 1;
                                                                    }
                                                                    else if (SC >= 5196) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (RsCsDiff >= 121) {
                                                                    if (refDCval < 84) {
                                                                        return 8;
                                                                    }
                                                                    else if (refDCval >= 84) {
                                                                        if (MVDiff < 6) {
                                                                            return 8;
                                                                        }
                                                                        else if (MVDiff >= 6) {
                                                                            if (TSC < 9) {
                                                                                return 2;
                                                                            }
                                                                            else if (TSC >= 9) {
                                                                                if (negBalance < 54) {
                                                                                    return 1;
                                                                                }
                                                                                else if (negBalance >= 54) {
                                                                                    return 2;
                                                                                }
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (posBalance >= 56) {
                                                            if (negBalance < 51) {
                                                                if (SC < 5321) {
                                                                    if (Rs < 2467) {
                                                                        if (Cs < 2509) {
                                                                            if (MVDiff < 11) {
                                                                                return 2;
                                                                            }
                                                                            else if (MVDiff >= 11) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                        else if (Cs >= 2509) {
                                                                            return 2;
                                                                        }
                                                                    }
                                                                    else if (Rs >= 2467) {
                                                                        if (Cs < 2322) {
                                                                            if (diffRsCsdiff < -29) {
                                                                                if (Rs < 2580) {
                                                                                    return 4;
                                                                                }
                                                                                else if (Rs >= 2580) {
                                                                                    return 2;
                                                                                }
                                                                            }
                                                                            else if (diffRsCsdiff >= -29) {
                                                                                if (diffRsCsdiff < 74) {
                                                                                    return 1;
                                                                                }
                                                                                else if (diffRsCsdiff >= 74) {
                                                                                    return 4;
                                                                                }
                                                                            }
                                                                        }
                                                                        else if (Cs >= 2322) {
                                                                            if (refDCval < 91) {
                                                                                return 2;
                                                                            }
                                                                            else if (refDCval >= 91) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                                else if (SC >= 5321) {
                                                                    if (mu_mv_mag_sq < 72) {
                                                                        if (diffMVdiffVal < 1) {
                                                                            return 1;
                                                                        }
                                                                        else if (diffMVdiffVal >= 1) {
                                                                            if (Rs < 2977) {
                                                                                return 4;
                                                                            }
                                                                            else if (Rs >= 2977) {
                                                                                if (ssDCval < 94) {
                                                                                    return 1;
                                                                                }
                                                                                else if (ssDCval >= 94) {
                                                                                    return 4;
                                                                                }
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (mu_mv_mag_sq >= 72) {
                                                                        if (Cs < 2418) {
                                                                            return 4;
                                                                        }
                                                                        else if (Cs >= 2418) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                            else if (negBalance >= 51) {
                                                                if (tcor < 90) {
                                                                    if (Rs < 2851) {
                                                                        if (mcTcor < 92) {
                                                                            if (RsCsDiff < 129) {
                                                                                if (AFD < 15) {
                                                                                    return 1;
                                                                                }
                                                                                else if (AFD >= 15) {
                                                                                    return 2;
                                                                                }
                                                                            }
                                                                            else if (RsCsDiff >= 129) {
                                                                                return 2;
                                                                            }
                                                                        }
                                                                        else if (mcTcor >= 92) {
                                                                            return 2;
                                                                        }
                                                                    }
                                                                    else if (Rs >= 2851) {
                                                                        if (MVDiff < 9) {
                                                                            return 4;
                                                                        }
                                                                        else if (MVDiff >= 9) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                }
                                                                else if (tcor >= 90) {
                                                                    if (Rs < 2955) {
                                                                        if (ssDCval < 93) {
                                                                            return 1;
                                                                        }
                                                                        else if (ssDCval >= 93) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                    else if (Rs >= 2955) {
                                                                        return 4;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (refDCval >= 101) {
                                                    if (MVDiff < 25) {
                                                        if (Cs < 1895) {
                                                            if (ssDCval < 117) {
                                                                if (Cs < 1814) {
                                                                    return 8;
                                                                }
                                                                else if (Cs >= 1814) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (ssDCval >= 117) {
                                                                if (MVDiff < 19) {
                                                                    return 4;
                                                                }
                                                                else if (MVDiff >= 19) {
                                                                    if (TSC < 12) {
                                                                        if (AFD < 18) {
                                                                            return 4;
                                                                        }
                                                                        else if (AFD >= 18) {
                                                                            return 2;
                                                                        }
                                                                    }
                                                                    else if (TSC >= 12) {
                                                                        return 8;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (Cs >= 1895) {
                                                            if (TSC < 11) {
                                                                if (SC < 5275) {
                                                                    if (TSC < 10) {
                                                                        return 2;
                                                                    }
                                                                    else if (TSC >= 10) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (SC >= 5275) {
                                                                    if (diffRsCsdiff < -16) {
                                                                        if (Rs < 5387) {
                                                                            return 2;
                                                                        }
                                                                        else if (Rs >= 5387) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                    else if (diffRsCsdiff >= -16) {
                                                                        if (diffMVdiffVal < -3) {
                                                                            return 4;
                                                                        }
                                                                        else if (diffMVdiffVal >= -3) {
                                                                            if (Rs < 5545) {
                                                                                return 8;
                                                                            }
                                                                            else if (Rs >= 5545) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                            else if (TSC >= 11) {
                                                                if (RsCsDiff < 189) {
                                                                    if (ssDCval < 105) {
                                                                        return 2;
                                                                    }
                                                                    else if (ssDCval >= 105) {
                                                                        return 1;
                                                                    }
                                                                }
                                                                else if (RsCsDiff >= 189) {
                                                                    if (mcTcor < 91) {
                                                                        if (ssDCval < 119) {
                                                                            return 4;
                                                                        }
                                                                        else if (ssDCval >= 119) {
                                                                            if (Rs < 3332) {
                                                                                return 8;
                                                                            }
                                                                            else if (Rs >= 3332) {
                                                                                return 1;
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (mcTcor >= 91) {
                                                                        if (TSC < 14) {
                                                                            return 8;
                                                                        }
                                                                        else if (TSC >= 14) {
                                                                            if (Rs < 3093) {
                                                                                return 1;
                                                                            }
                                                                            else if (Rs >= 3093) {
                                                                                return 8;
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (MVDiff >= 25) {
                                                        if (diffRsCsdiff < -1) {
                                                            if (posBalance < 58) {
                                                                if (diffRsCsdiff < -6) {
                                                                    if (Rs < 3588) {
                                                                        return 2;
                                                                    }
                                                                    else if (Rs >= 3588) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (diffRsCsdiff >= -6) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (posBalance >= 58) {
                                                                if (SC < 3652) {
                                                                    return 4;
                                                                }
                                                                else if (SC >= 3652) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                        else if (diffRsCsdiff >= -1) {
                                                            if (diffTSC < 3) {
                                                                return 2;
                                                            }
                                                            else if (diffTSC >= 3) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (ssDCval >= 129) {
                                            if (negBalance < 55) {
                                                if (CsDiff < 6) {
                                                    return 2;
                                                }
                                                else if (CsDiff >= 6) {
                                                    return 8;
                                                }
                                            }
                                            else if (negBalance >= 55) {
                                                return 2;
                                            }
                                        }
                                    }
                                }
                            }
                        }
                    }
                    else if (Cs >= 3266) {
                        if (refDCval < 68) {
                            if (RsCsDiff < 46) {
                                if (TSC < 7) {
                                    if (tcor < 90) {
                                        if (ssDCval < 63) {
                                            if (ssDCval < 54) {
                                                if (diffAFD < 1) {
                                                    return 8;
                                                }
                                                else if (diffAFD >= 1) {
                                                    return 4;
                                                }
                                            }
                                            else if (ssDCval >= 54) {
                                                return 8;
                                            }
                                        }
                                        else if (ssDCval >= 63) {
                                            return 4;
                                        }
                                    }
                                    else if (tcor >= 90) {
                                        if (posBalance < 54) {
                                            if (Cs < 4012) {
                                                if (Cs < 3355) {
                                                    return 8;
                                                }
                                                else if (Cs >= 3355) {
                                                    return 4;
                                                }
                                            }
                                            else if (Cs >= 4012) {
                                                return 2;
                                            }
                                        }
                                        else if (posBalance >= 54) {
                                            if (ssDCval < 55) {
                                                if (MVDiff < 24) {
                                                    if (diffMVdiffVal < -6) {
                                                        return 8;
                                                    }
                                                    else if (diffMVdiffVal >= -6) {
                                                        if (mu_mv_mag_sq < 71) {
                                                            return 4;
                                                        }
                                                        else if (mu_mv_mag_sq >= 71) {
                                                            if (tcor < 94) {
                                                                return 4;
                                                            }
                                                            else if (tcor >= 94) {
                                                                if (RsCsDiff < 16) {
                                                                    return 8;
                                                                }
                                                                else if (RsCsDiff >= 16) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (MVDiff >= 24) {
                                                    if (Cs < 3529) {
                                                        return 8;
                                                    }
                                                    else if (Cs >= 3529) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                            else if (ssDCval >= 55) {
                                                if (diffTSC < 1) {
                                                    return 8;
                                                }
                                                else if (diffTSC >= 1) {
                                                    return 4;
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (TSC >= 7) {
                                    if (MVDiff < 19) {
                                        if (negBalance < 50) {
                                            if (Cs < 4183) {
                                                return 4;
                                            }
                                            else if (Cs >= 4183) {
                                                return 2;
                                            }
                                        }
                                        else if (negBalance >= 50) {
                                            if (diffMVdiffVal < -3) {
                                                return 2;
                                            }
                                            else if (diffMVdiffVal >= -3) {
                                                if (MVDiff < 17) {
                                                    if (CsDiff < 4) {
                                                        return 4;
                                                    }
                                                    else if (CsDiff >= 4) {
                                                        if (refDCval < 55) {
                                                            return 2;
                                                        }
                                                        else if (refDCval >= 55) {
                                                            if (TSC < 9) {
                                                                return 4;
                                                            }
                                                            else if (TSC >= 9) {
                                                                return 2;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (MVDiff >= 17) {
                                                    return 2;
                                                }
                                            }
                                        }
                                    }
                                    else if (MVDiff >= 19) {
                                        if (ssDCval < 58) {
                                            if (Rs < 3490) {
                                                if (RsCsDiff < 23) {
                                                    if (Rs < 3381) {
                                                        return 8;
                                                    }
                                                    else if (Rs >= 3381) {
                                                        return 2;
                                                    }
                                                }
                                                else if (RsCsDiff >= 23) {
                                                    return 8;
                                                }
                                            }
                                            else if (Rs >= 3490) {
                                                if (AFD < 12) {
                                                    if (mcTcor < 94) {
                                                        return 2;
                                                    }
                                                    else if (mcTcor >= 94) {
                                                        if (SC < 7394) {
                                                            if (Cs < 3532) {
                                                                return 2;
                                                            }
                                                            else if (Cs >= 3532) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (SC >= 7394) {
                                                            return 2;
                                                        }
                                                    }
                                                }
                                                else if (AFD >= 12) {
                                                    if (SC < 7476) {
                                                        return 8;
                                                    }
                                                    else if (SC >= 7476) {
                                                        return 2;
                                                    }
                                                }
                                            }
                                        }
                                        else if (ssDCval >= 58) {
                                            if (Rs < 3787) {
                                                return 8;
                                            }
                                            else if (Rs >= 3787) {
                                                return 4;
                                            }
                                        }
                                    }
                                }
                            }
                            else if (RsCsDiff >= 46) {
                                if (mcTcor < 82) {
                                    return 8;
                                }
                                else if (mcTcor >= 82) {
                                    if (MVDiff < 40) {
                                        if (mu_mv_mag_sq < 40) {
                                            if (negBalance < 50) {
                                                return 8;
                                            }
                                            else if (negBalance >= 50) {
                                                if (mcTcor < 94) {
                                                    return 4;
                                                }
                                                else if (mcTcor >= 94) {
                                                    return 8;
                                                }
                                            }
                                        }
                                        else if (mu_mv_mag_sq >= 40) {
                                            return 4;
                                        }
                                    }
                                    else if (MVDiff >= 40) {
                                        if (SC < 8161) {
                                            return 8;
                                        }
                                        else if (SC >= 8161) {
                                            return 4;
                                        }
                                    }
                                }
                            }
                        }
                        else if (refDCval >= 68) {
                            if (SC < 12051) {
                                if (TSC < 18) {
                                    if (refDCval < 148) {
                                        if (Rs < 3669) {
                                            if (TSC < 13) {
                                                if (SC < 6337) {
                                                    return 1;
                                                }
                                                else if (SC >= 6337) {
                                                    return 8;
                                                }
                                            }
                                            else if (TSC >= 13) {
                                                return 1;
                                            }
                                        }
                                        else if (Rs >= 3669) {
                                            if (mu_mv_mag_sq < 3) {
                                                if (posBalance < 53) {
                                                    if (AFD < 22) {
                                                        return 8;
                                                    }
                                                    else if (AFD >= 22) {
                                                        return 4;
                                                    }
                                                }
                                                else if (posBalance >= 53) {
                                                    if (Cs < 4380) {
                                                        if (ssDCval < 123) {
                                                            return 4;
                                                        }
                                                        else if (ssDCval >= 123) {
                                                            if (AFD < 21) {
                                                                return 1;
                                                            }
                                                            else if (AFD >= 21) {
                                                                if (Rs < 5095) {
                                                                    return 8;
                                                                }
                                                                else if (Rs >= 5095) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (Cs >= 4380) {
                                                        if (ssDCval < 103) {
                                                            return 1;
                                                        }
                                                        else if (ssDCval >= 103) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (mu_mv_mag_sq >= 3) {
                                                if (RsDiff < 7) {
                                                    if (RsCsDiff < 67) {
                                                        return 4;
                                                    }
                                                    else if (RsCsDiff >= 67) {
                                                        return 8;
                                                    }
                                                }
                                                else if (RsDiff >= 7) {
                                                    if (refDCval < 95) {
                                                        if (AFD < 33) {
                                                            if (Rs < 6437) {
                                                                return 8;
                                                            }
                                                            else if (Rs >= 6437) {
                                                                if (tcor < 76) {
                                                                    return 8;
                                                                }
                                                                else if (tcor >= 76) {
                                                                    if (Rs < 6748) {
                                                                        return 1;
                                                                    }
                                                                    else if (Rs >= 6748) {
                                                                        return 4;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (AFD >= 33) {
                                                            return 2;
                                                        }
                                                    }
                                                    else if (refDCval >= 95) {
                                                        if (ssDCval < 122) {
                                                            if (RsCsDiff < 454) {
                                                                if (negBalance < 39) {
                                                                    if (RsCsDiff < 156) {
                                                                        return 2;
                                                                    }
                                                                    else if (RsCsDiff >= 156) {
                                                                        return 1;
                                                                    }
                                                                }
                                                                else if (negBalance >= 39) {
                                                                    if (Rs < 4049) {
                                                                        return 8;
                                                                    }
                                                                    else if (Rs >= 4049) {
                                                                        if (SC < 8796) {
                                                                            return 4;
                                                                        }
                                                                        else if (SC >= 8796) {
                                                                            return 2;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                            else if (RsCsDiff >= 454) {
                                                                if (RsCsDiff < 713) {
                                                                    return 8;
                                                                }
                                                                else if (RsCsDiff >= 713) {
                                                                    if (AFD < 29) {
                                                                        return 1;
                                                                    }
                                                                    else if (AFD >= 29) {
                                                                        return 2;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (ssDCval >= 122) {
                                                            if (tcor < 70) {
                                                                if (AFD < 21) {
                                                                    if (Rs < 5150) {
                                                                        return 1;
                                                                    }
                                                                    else if (Rs >= 5150) {
                                                                        return 8;
                                                                    }
                                                                }
                                                                else if (AFD >= 21) {
                                                                    if (CsDiff < 18) {
                                                                        if (diffRsCsdiff < -102) {
                                                                            return 8;
                                                                        }
                                                                        else if (diffRsCsdiff >= -102) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                    else if (CsDiff >= 18) {
                                                                        return 8;
                                                                    }
                                                                }
                                                            }
                                                            else if (tcor >= 70) {
                                                                if (negBalance < 61) {
                                                                    if (diffAFD < -7) {
                                                                        if (Rs < 5113) {
                                                                            return 8;
                                                                        }
                                                                        else if (Rs >= 5113) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                    else if (diffAFD >= -7) {
                                                                        if (posBalance < 62) {
                                                                            return 8;
                                                                        }
                                                                        else if (posBalance >= 62) {
                                                                            if (Rs < 4080) {
                                                                                return 1;
                                                                            }
                                                                            else if (Rs >= 4080) {
                                                                                return 8;
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                                else if (negBalance >= 61) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (refDCval >= 148) {
                                        if (mcTcor < 89) {
                                            if (RsDiff < 19) {
                                                return 2;
                                            }
                                            else if (RsDiff >= 19) {
                                                if (MVDiff < 26) {
                                                    if (Rs < 7406) {
                                                        return 2;
                                                    }
                                                    else if (Rs >= 7406) {
                                                        return 4;
                                                    }
                                                }
                                                else if (MVDiff >= 26) {
                                                    return 4;
                                                }
                                            }
                                        }
                                        else if (mcTcor >= 89) {
                                            if (tcor < 79) {
                                                if (Rs < 6860) {
                                                    return 2;
                                                }
                                                else if (Rs >= 6860) {
                                                    return 4;
                                                }
                                            }
                                            else if (tcor >= 79) {
                                                return 8;
                                            }
                                        }
                                    }
                                }
                                else if (TSC >= 18) {
                                    if (Rs < 3095) {
                                        return 4;
                                    }
                                    else if (Rs >= 3095) {
                                        if (refDCval < 136) {
                                            return 8;
                                        }
                                        else if (refDCval >= 136) {
                                            if (CsDiff < 17) {
                                                if (Rs < 5659) {
                                                    return 2;
                                                }
                                                else if (Rs >= 5659) {
                                                    return 4;
                                                }
                                            }
                                            else if (CsDiff >= 17) {
                                                if (Cs < 4085) {
                                                    return 4;
                                                }
                                                else if (Cs >= 4085) {
                                                    if (negBalance < 52) {
                                                        return 4;
                                                    }
                                                    else if (negBalance >= 52) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                            else if (SC >= 12051) {
                                if (MVDiff < 23) {
                                    if (mu_mv_mag_sq < 63) {
                                        if (refDCval < 91) {
                                            if (AFD < 27) {
                                                if (tcor < 63) {
                                                    return 8;
                                                }
                                                else if (tcor >= 63) {
                                                    if (Cs < 7376) {
                                                        return 4;
                                                    }
                                                    else if (Cs >= 7376) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                            else if (AFD >= 27) {
                                                if (mcTcor < 87) {
                                                    return 4;
                                                }
                                                else if (mcTcor >= 87) {
                                                    return 2;
                                                }
                                            }
                                        }
                                        else if (refDCval >= 91) {
                                            if (Rs < 7338) {
                                                if (TSC < 18) {
                                                    if (ssDCval < 150) {
                                                        return 4;
                                                    }
                                                    else if (ssDCval >= 150) {
                                                        return 2;
                                                    }
                                                }
                                                else if (TSC >= 18) {
                                                    if (ssDCval < 142) {
                                                        return 8;
                                                    }
                                                    else if (ssDCval >= 142) {
                                                        if (Rs < 7115) {
                                                            return 4;
                                                        }
                                                        else if (Rs >= 7115) {
                                                            return 2;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (Rs >= 7338) {
                                                if (RsCsDiff < 2061) {
                                                    if (mcTcor < 91) {
                                                        if (gchDC < 4) {
                                                            if (SC < 12627) {
                                                                return 4;
                                                            }
                                                            else if (SC >= 12627) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (gchDC >= 4) {
                                                            if (Rs < 7731) {
                                                                return 4;
                                                            }
                                                            else if (Rs >= 7731) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                    else if (mcTcor >= 91) {
                                                        if (Rs < 9956) {
                                                            return 4;
                                                        }
                                                        else if (Rs >= 9956) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                                else if (RsCsDiff >= 2061) {
                                                    if (Rs < 14703) {
                                                        if (negBalance < 53) {
                                                            if (Cs < 4831) {
                                                                return 4;
                                                            }
                                                            else if (Cs >= 4831) {
                                                                if (RsCsDiff < 2132) {
                                                                    return 4;
                                                                }
                                                                else if (RsCsDiff >= 2132) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                        else if (negBalance >= 53) {
                                                            if (ssDCval < 136) {
                                                                return 4;
                                                            }
                                                            else if (ssDCval >= 136) {
                                                                if (Rs < 12465) {
                                                                    return 4;
                                                                }
                                                                else if (Rs >= 12465) {
                                                                    return 2;
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (Rs >= 14703) {
                                                        if (SC < 28089) {
                                                            return 8;
                                                        }
                                                        else if (SC >= 28089) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (mu_mv_mag_sq >= 63) {
                                        if (posBalance < 55) {
                                            if (RsDiff < 24) {
                                                return 4;
                                            }
                                            else if (RsDiff >= 24) {
                                                return 8;
                                            }
                                        }
                                        else if (posBalance >= 55) {
                                            return 8;
                                        }
                                    }
                                }
                                else if (MVDiff >= 23) {
                                    if (Cs < 6354) {
                                        if (AFD < 24) {
                                            return 4;
                                        }
                                        else if (AFD >= 24) {
                                            if (refDCval < 159) {
                                                if (mcTcor < 84) {
                                                    if (CsDiff < 20) {
                                                        return 2;
                                                    }
                                                    else if (CsDiff >= 20) {
                                                        return 8;
                                                    }
                                                }
                                                else if (mcTcor >= 84) {
                                                    return 2;
                                                }
                                            }
                                            else if (refDCval >= 159) {
                                                if (Rs < 9010) {
                                                    return 8;
                                                }
                                                else if (Rs >= 9010) {
                                                    return 2;
                                                }
                                            }
                                        }
                                    }
                                    else if (Cs >= 6354) {
                                        return 8;
                                    }
                                }
                            }
                        }
                    }
                }
            }
            else if (mu_mv_mag_sq >= 83) {
                if (TSC < 6) {
                    if (RsCsDiff < 5) {
                        if (mu_mv_mag_sq < 110) {
                            if (diffRsCsdiff < -2) {
                                if (MVDiff < 23) {
                                    return 8;
                                }
                                else if (MVDiff >= 23) {
                                    if (SC < 2386) {
                                        return 4;
                                    }
                                    else if (SC >= 2386) {
                                        return 1;
                                    }
                                }
                            }
                            else if (diffRsCsdiff >= -2) {
                                if (MVDiff < 10) {
                                    if (mcTcor < 96) {
                                        if (refDCval < 46) {
                                            if (negBalance < 42) {
                                                return 4;
                                            }
                                            else if (negBalance >= 42) {
                                                return 8;
                                            }
                                        }
                                        else if (refDCval >= 46) {
                                            return 8;
                                        }
                                    }
                                    else if (mcTcor >= 96) {
                                        if (MVDiff < 8) {
                                            return 4;
                                        }
                                        else if (MVDiff >= 8) {
                                            return 8;
                                        }
                                    }
                                }
                                else if (MVDiff >= 10) {
                                    if (diffMVdiffVal < -5) {
                                        if (AFD < 5) {
                                            return 4;
                                        }
                                        else if (AFD >= 5) {
                                            if (Rs < 1324) {
                                                return 4;
                                            }
                                            else if (Rs >= 1324) {
                                                return 8;
                                            }
                                        }
                                    }
                                    else if (diffMVdiffVal >= -5) {
                                        return 4;
                                    }
                                }
                            }
                        }
                        else if (mu_mv_mag_sq >= 110) {
                            if (Scindex < 5) {
                                if (AFD < 11) {
                                    return 8;
                                }
                                else if (AFD >= 11) {
                                    return 2;
                                }
                            }
                            else if (Scindex >= 5) {
                                if (ssDCval < 117) {
                                    if (Cs < 2387) {
                                        if (AFD < 3) {
                                            if (posBalance < 9) {
                                                return 4;
                                            }
                                            else if (posBalance >= 9) {
                                                return 8;
                                            }
                                        }
                                        else if (AFD >= 3) {
                                            return 8;
                                        }
                                    }
                                    else if (Cs >= 2387) {
                                        return 8;
                                    }
                                }
                                else if (ssDCval >= 117) {
                                    if (SC < 4443) {
                                        return 1;
                                    }
                                    else if (SC >= 4443) {
                                        return 4;
                                    }
                                }
                            }
                        }
                    }
                    else if (RsCsDiff >= 5) {
                        if (ssDCval < 96) {
                            if (MVDiff < 24) {
                                if (mu_mv_mag_sq < 111) {
                                    if (Cs < 1520) {
                                        if (diffTSC < -1) {
                                            if (TSC < 4) {
                                                return 1;
                                            }
                                            else if (TSC >= 4) {
                                                return 8;
                                            }
                                        }
                                        else if (diffTSC >= -1) {
                                            if (posBalance < 15) {
                                                if (Rs < 1063) {
                                                    return 8;
                                                }
                                                else if (Rs >= 1063) {
                                                    return 4;
                                                }
                                            }
                                            else if (posBalance >= 15) {
                                                if (Rs < 851) {
                                                    if (Rs < 819) {
                                                        return 8;
                                                    }
                                                    else if (Rs >= 819) {
                                                        return 4;
                                                    }
                                                }
                                                else if (Rs >= 851) {
                                                    return 8;
                                                }
                                            }
                                        }
                                    }
                                    else if (Cs >= 1520) {
                                        if (posBalance < 41) {
                                            if (negBalance < 22) {
                                                return 1;
                                            }
                                            else if (negBalance >= 22) {
                                                return 8;
                                            }
                                        }
                                        else if (posBalance >= 41) {
                                            if (RsCsDiff < 23) {
                                                if (mu_mv_mag_sq < 86) {
                                                    return 4;
                                                }
                                                else if (mu_mv_mag_sq >= 86) {
                                                    if (diffAFD < -1) {
                                                        if (Rs < 2863) {
                                                            return 4;
                                                        }
                                                        else if (Rs >= 2863) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (diffAFD >= -1) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                            else if (RsCsDiff >= 23) {
                                                if (diffRsCsdiff < 10) {
                                                    return 4;
                                                }
                                                else if (diffRsCsdiff >= 10) {
                                                    return 8;
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (mu_mv_mag_sq >= 111) {
                                    if (ssDCval < 51) {
                                        if (negBalance < 20) {
                                            return 8;
                                        }
                                        else if (negBalance >= 20) {
                                            if (ssDCval < 50) {
                                                return 8;
                                            }
                                            else if (ssDCval >= 50) {
                                                return 4;
                                            }
                                        }
                                    }
                                    else if (ssDCval >= 51) {
                                        if (mu_mv_mag_sq < 112) {
                                            if (Cs < 2110) {
                                                return 8;
                                            }
                                            else if (Cs >= 2110) {
                                                if (Cs < 2307) {
                                                    return 4;
                                                }
                                                else if (Cs >= 2307) {
                                                    return 8;
                                                }
                                            }
                                        }
                                        else if (mu_mv_mag_sq >= 112) {
                                            return 8;
                                        }
                                    }
                                }
                            }
                            else if (MVDiff >= 24) {
                                if (MVDiff < 26) {
                                    return 4;
                                }
                                else if (MVDiff >= 26) {
                                    if (CsDiff < 2) {
                                        return 1;
                                    }
                                    else if (CsDiff >= 2) {
                                        return 8;
                                    }
                                }
                            }
                        }
                        else if (ssDCval >= 96) {
                            if (diffMVdiffVal < -3) {
                                if (mu_mv_mag_sq < 123) {
                                    if (TSCindex < 5) {
                                        if (ssDCval < 130) {
                                            if (SC < 3014) {
                                                return 1;
                                            }
                                            else if (SC >= 3014) {
                                                return 8;
                                            }
                                        }
                                        else if (ssDCval >= 130) {
                                            if (Cs < 713) {
                                                return 1;
                                            }
                                            else if (Cs >= 713) {
                                                return 4;
                                            }
                                        }
                                    }
                                    else if (TSCindex >= 5) {
                                        return 8;
                                    }
                                }
                                else if (mu_mv_mag_sq >= 123) {
                                    return 1;
                                }
                            }
                            else if (diffMVdiffVal >= -3) {
                                if (RsCsDiff < 7) {
                                    if (refDCval < 125) {
                                        if (SC < 4720) {
                                            return 4;
                                        }
                                        else if (SC >= 4720) {
                                            return 8;
                                        }
                                    }
                                    else if (refDCval >= 125) {
                                        if (ssDCval < 159) {
                                            return 8;
                                        }
                                        else if (ssDCval >= 159) {
                                            return 4;
                                        }
                                    }
                                }
                                else if (RsCsDiff >= 7) {
                                    if (MVDiff < 6) {
                                        if (mcTcor < 80) {
                                            if (Rs < 1307) {
                                                return 8;
                                            }
                                            else if (Rs >= 1307) {
                                                return 1;
                                            }
                                        }
                                        else if (mcTcor >= 80) {
                                            return 8;
                                        }
                                    }
                                    else if (MVDiff >= 6) {
                                        if (Cs < 1609) {
                                            if (diffAFD < -2) {
                                                return 4;
                                            }
                                            else if (diffAFD >= -2) {
                                                if (Cs < 626) {
                                                    return 8;
                                                }
                                                else if (Cs >= 626) {
                                                    if (posBalance < 6) {
                                                        return 8;
                                                    }
                                                    else if (posBalance >= 6) {
                                                        if (tcor < 92) {
                                                            if (mu_mv_mag_sq < 96) {
                                                                if (Cs < 749) {
                                                                    return 1;
                                                                }
                                                                else if (Cs >= 749) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (mu_mv_mag_sq >= 96) {
                                                                return 1;
                                                            }
                                                        }
                                                        else if (tcor >= 92) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (Cs >= 1609) {
                                            return 8;
                                        }
                                    }
                                }
                            }
                        }
                    }
                }
                else if (TSC >= 6) {
                    if (Cs < 4004) {
                        if (CsDiff < 4) {
                            if (Rs < 1882) {
                                if (RsDiff < 6) {
                                    if (mu_mv_mag_sq < 127) {
                                        return 2;
                                    }
                                    else if (mu_mv_mag_sq >= 127) {
                                        return 8;
                                    }
                                }
                                else if (RsDiff >= 6) {
                                    return 8;
                                }
                            }
                            else if (Rs >= 1882) {
                                if (Cs < 803) {
                                    return 2;
                                }
                                else if (Cs >= 803) {
                                    return 8;
                                }
                            }
                        }
                        else if (CsDiff >= 4) {
                            if (mu_mv_mag_sq < 127) {
                                if (Cs < 2027) {
                                    if (MVDiff < 19) {
                                        if (refDCval < 72) {
                                            if (negBalance < 44) {
                                                if (MVDiff < 12) {
                                                    return 8;
                                                }
                                                else if (MVDiff >= 12) {
                                                    return 4;
                                                }
                                            }
                                            else if (negBalance >= 44) {
                                                if (RsDiff < 5) {
                                                    return 2;
                                                }
                                                else if (RsDiff >= 5) {
                                                    if (MVDiff < 1) {
                                                        return 2;
                                                    }
                                                    else if (MVDiff >= 1) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                        }
                                        else if (refDCval >= 72) {
                                            if (mu_mv_mag_sq < 109) {
                                                if (refDCval < 95) {
                                                    if (mcTcor < 92) {
                                                        if (ssDCval < 92) {
                                                            if (Cs < 1289) {
                                                                return 4;
                                                            }
                                                            else if (Cs >= 1289) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (ssDCval >= 92) {
                                                            return 2;
                                                        }
                                                    }
                                                    else if (mcTcor >= 92) {
                                                        if (negBalance < 42) {
                                                            if (Rs < 2149) {
                                                                return 4;
                                                            }
                                                            else if (Rs >= 2149) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (negBalance >= 42) {
                                                            return 2;
                                                        }
                                                    }
                                                }
                                                else if (refDCval >= 95) {
                                                    if (diffRsCsdiff < -38) {
                                                        return 4;
                                                    }
                                                    else if (diffRsCsdiff >= -38) {
                                                        if (Rs < 2205) {
                                                            return 8;
                                                        }
                                                        else if (Rs >= 2205) {
                                                            if (diffMVdiffVal < -2) {
                                                                if (mcTcor < 90) {
                                                                    return 4;
                                                                }
                                                                else if (mcTcor >= 90) {
                                                                    if (Cs < 1576) {
                                                                        return 4;
                                                                    }
                                                                    else if (Cs >= 1576) {
                                                                        return 8;
                                                                    }
                                                                }
                                                            }
                                                            else if (diffMVdiffVal >= -2) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                            else if (mu_mv_mag_sq >= 109) {
                                                if (ssDCval < 77) {
                                                    if (SC < 3384) {
                                                        if (MVDiff < 1) {
                                                            if (posBalance < 58) {
                                                                return 8;
                                                            }
                                                            else if (posBalance >= 58) {
                                                                if (ssDCval < 75) {
                                                                    return 2;
                                                                }
                                                                else if (ssDCval >= 75) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                        else if (MVDiff >= 1) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (SC >= 3384) {
                                                        if (tcor < 92) {
                                                            return 8;
                                                        }
                                                        else if (tcor >= 92) {
                                                            if (posBalance < 58) {
                                                                if (Rs < 1855) {
                                                                    return 8;
                                                                }
                                                                else if (Rs >= 1855) {
                                                                    if (diffMVdiffVal < -1) {
                                                                        return 8;
                                                                    }
                                                                    else if (diffMVdiffVal >= -1) {
                                                                        return 2;
                                                                    }
                                                                }
                                                            }
                                                            else if (posBalance >= 58) {
                                                                if (ssDCval < 75) {
                                                                    return 2;
                                                                }
                                                                else if (ssDCval >= 75) {
                                                                    if (AFD < 7) {
                                                                        return 4;
                                                                    }
                                                                    else if (AFD >= 7) {
                                                                        if (Cs < 1591) {
                                                                            return 4;
                                                                        }
                                                                        else if (Cs >= 1591) {
                                                                            return 2;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (ssDCval >= 77) {
                                                    if (AFD < 21) {
                                                        if (ssDCval < 118) {
                                                            return 8;
                                                        }
                                                        else if (ssDCval >= 118) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (AFD >= 21) {
                                                        return 2;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (MVDiff >= 19) {
                                        if (negBalance < 54) {
                                            if (RsDiff < 15) {
                                                if (Rs < 1207) {
                                                    return 1;
                                                }
                                                else if (Rs >= 1207) {
                                                    return 2;
                                                }
                                            }
                                            else if (RsDiff >= 15) {
                                                if (refDCval < 151) {
                                                    return 4;
                                                }
                                                else if (refDCval >= 151) {
                                                    return 2;
                                                }
                                            }
                                        }
                                        else if (negBalance >= 54) {
                                            return 2;
                                        }
                                    }
                                }
                                else if (Cs >= 2027) {
                                    if (SC < 6386) {
                                        if (ssDCval < 86) {
                                            if (Cs < 2225) {
                                                if (tcor < 94) {
                                                    return 8;
                                                }
                                                else if (tcor >= 94) {
                                                    return 4;
                                                }
                                            }
                                            else if (Cs >= 2225) {
                                                if (posBalance < 53) {
                                                    if (SC < 4407) {
                                                        return 1;
                                                    }
                                                    else if (SC >= 4407) {
                                                        return 8;
                                                    }
                                                }
                                                else if (posBalance >= 53) {
                                                    if (MVDiff < 1) {
                                                        return 4;
                                                    }
                                                    else if (MVDiff >= 1) {
                                                        if (RsCsDiff < 94) {
                                                            if (diffAFD < -1) {
                                                                if (TSC < 8) {
                                                                    return 4;
                                                                }
                                                                else if (TSC >= 8) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (diffAFD >= -1) {
                                                                if (CsDiff < 7) {
                                                                    return 4;
                                                                }
                                                                else if (CsDiff >= 7) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                        else if (RsCsDiff >= 94) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (ssDCval >= 86) {
                                            if (SC < 4772) {
                                                if (ssDCval < 97) {
                                                    if (ssDCval < 92) {
                                                        if (diffAFD < -2) {
                                                            if (tcor < 91) {
                                                                return 8;
                                                            }
                                                            else if (tcor >= 91) {
                                                                return 2;
                                                            }
                                                        }
                                                        else if (diffAFD >= -2) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (ssDCval >= 92) {
                                                        if (posBalance < 51) {
                                                            return 1;
                                                        }
                                                        else if (posBalance >= 51) {
                                                            if (Cs < 2066) {
                                                                return 8;
                                                            }
                                                            else if (Cs >= 2066) {
                                                                if (refDCval < 93) {
                                                                    return 8;
                                                                }
                                                                else if (refDCval >= 93) {
                                                                    return 2;
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (ssDCval >= 97) {
                                                    if (Rs < 2094) {
                                                        return 8;
                                                    }
                                                    else if (Rs >= 2094) {
                                                        if (ssDCval < 112) {
                                                            return 4;
                                                        }
                                                        else if (ssDCval >= 112) {
                                                            if (posBalance < 56) {
                                                                if (SC < 4561) {
                                                                    return 4;
                                                                }
                                                                else if (SC >= 4561) {
                                                                    return 1;
                                                                }
                                                            }
                                                            else if (posBalance >= 56) {
                                                                if (Rs < 2113) {
                                                                    return 4;
                                                                }
                                                                else if (Rs >= 2113) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                            else if (SC >= 4772) {
                                                if (Rs < 3335) {
                                                    if (diffMVdiffVal < -2) {
                                                        if (diffRsCsdiff < -49) {
                                                            return 4;
                                                        }
                                                        else if (diffRsCsdiff >= -49) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (diffMVdiffVal >= -2) {
                                                        return 8;
                                                    }
                                                }
                                                else if (Rs >= 3335) {
                                                    if (AFD < 10) {
                                                        return 8;
                                                    }
                                                    else if (AFD >= 10) {
                                                        return 1;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (SC >= 6386) {
                                        if (SC < 7755) {
                                            if (MVDiff < 1) {
                                                if (Rs < 3939) {
                                                    if (Cs < 3727) {
                                                        return 2;
                                                    }
                                                    else if (Cs >= 3727) {
                                                        return 8;
                                                    }
                                                }
                                                else if (Rs >= 3939) {
                                                    if (Rs < 4350) {
                                                        return 8;
                                                    }
                                                    else if (Rs >= 4350) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                            else if (MVDiff >= 1) {
                                                if (SC < 6886) {
                                                    if (tcor < 84) {
                                                        if (Rs < 3769) {
                                                            if (Cs < 2905) {
                                                                return 1;
                                                            }
                                                            else if (Cs >= 2905) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (Rs >= 3769) {
                                                            return 1;
                                                        }
                                                    }
                                                    else if (tcor >= 84) {
                                                        if (Cs < 2380) {
                                                            return 4;
                                                        }
                                                        else if (Cs >= 2380) {
                                                            if (diffAFD < 1) {
                                                                if (diffMVdiffVal < 3) {
                                                                    return 8;
                                                                }
                                                                else if (diffMVdiffVal >= 3) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (diffAFD >= 1) {
                                                                if (mu_mv_mag_sq < 97) {
                                                                    if (SC < 6394) {
                                                                        return 1;
                                                                    }
                                                                    else if (SC >= 6394) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (mu_mv_mag_sq >= 97) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (SC >= 6886) {
                                                    if (ssDCval < 127) {
                                                        if (Cs < 3370) {
                                                            if (Rs < 5011) {
                                                                return 4;
                                                            }
                                                            else if (Rs >= 5011) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (Cs >= 3370) {
                                                            if (CsDiff < 5) {
                                                                return 4;
                                                            }
                                                            else if (CsDiff >= 5) {
                                                                if (CsDiff < 13) {
                                                                    return 8;
                                                                }
                                                                else if (CsDiff >= 13) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (ssDCval >= 127) {
                                                        if (Cs < 3053) {
                                                            return 2;
                                                        }
                                                        else if (Cs >= 3053) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (SC >= 7755) {
                                            if (negBalance < 45) {
                                                if (Rs < 5085) {
                                                    if (refDCval < 115) {
                                                        if (AFD < 9) {
                                                            if (SC < 7992) {
                                                                return 2;
                                                            }
                                                            else if (SC >= 7992) {
                                                                if (RsDiff < 10) {
                                                                    return 8;
                                                                }
                                                                else if (RsDiff >= 10) {
                                                                    return 2;
                                                                }
                                                            }
                                                        }
                                                        else if (AFD >= 9) {
                                                            return 2;
                                                        }
                                                    }
                                                    else if (refDCval >= 115) {
                                                        if (Rs < 4921) {
                                                            return 4;
                                                        }
                                                        else if (Rs >= 4921) {
                                                            return 2;
                                                        }
                                                    }
                                                }
                                                else if (Rs >= 5085) {
                                                    if (AFD < 11) {
                                                        return 8;
                                                    }
                                                    else if (AFD >= 11) {
                                                        if (Rs < 5506) {
                                                            return 2;
                                                        }
                                                        else if (Rs >= 5506) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (negBalance >= 45) {
                                                if (Cs < 3093) {
                                                    return 4;
                                                }
                                                else if (Cs >= 3093) {
                                                    if (RsCsDiff < 56) {
                                                        return 4;
                                                    }
                                                    else if (RsCsDiff >= 56) {
                                                        if (diffAFD < -3) {
                                                            return 4;
                                                        }
                                                        else if (diffAFD >= -3) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                            else if (mu_mv_mag_sq >= 127) {
                                if (MVDiff < 21) {
                                    if (Rs < 3485) {
                                        return 8;
                                    }
                                    else if (Rs >= 3485) {
                                        return 4;
                                    }
                                }
                                else if (MVDiff >= 21) {
                                    if (AFD < 33) {
                                        if (refDCval < 95) {
                                            return 2;
                                        }
                                        else if (refDCval >= 95) {
                                            return 1;
                                        }
                                    }
                                    else if (AFD >= 33) {
                                        if (MVDiff < 218) {
                                            return 2;
                                        }
                                        else if (MVDiff >= 218) {
                                            if (MVDiff < 541) {
                                                return 4;
                                            }
                                            else if (MVDiff >= 541) {
                                                return 8;
                                            }
                                        }
                                    }
                                }
                            }
                        }
                    }
                    else if (Cs >= 4004) {
                        if (mu_mv_mag_sq < 127) {
                            if (Rs < 7393) {
                                if (RsCsDiff < 56) {
                                    if (AFD < 7) {
                                        return 8;
                                    }
                                    else if (AFD >= 7) {
                                        return 4;
                                    }
                                }
                                else if (RsCsDiff >= 56) {
                                    if (ssDCval < 106) {
                                        return 8;
                                    }
                                    else if (ssDCval >= 106) {
                                        if (mu_mv_mag_sq < 122) {
                                            if (mcTcor < 79) {
                                                if (RsDiff < 16) {
                                                    return 1;
                                                }
                                                else if (RsDiff >= 16) {
                                                    return 8;
                                                }
                                            }
                                            else if (mcTcor >= 79) {
                                                return 8;
                                            }
                                        }
                                        else if (mu_mv_mag_sq >= 122) {
                                            if (ssDCval < 109) {
                                                if (CsDiff < 14) {
                                                    return 4;
                                                }
                                                else if (CsDiff >= 14) {
                                                    if (Rs < 5641) {
                                                        return 4;
                                                    }
                                                    else if (Rs >= 5641) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                            else if (ssDCval >= 109) {
                                                return 8;
                                            }
                                        }
                                    }
                                }
                            }
                            else if (Rs >= 7393) {
                                if (ssDCval < 96) {
                                    if (CsDiff < 22) {
                                        if (Rs < 7770) {
                                            return 4;
                                        }
                                        else if (Rs >= 7770) {
                                            return 2;
                                        }
                                    }
                                    else if (CsDiff >= 22) {
                                        if (diffRsCsdiff < -1) {
                                            if (Cs < 7967) {
                                                return 4;
                                            }
                                            else if (Cs >= 7967) {
                                                return 2;
                                            }
                                        }
                                        else if (diffRsCsdiff >= -1) {
                                            return 2;
                                        }
                                    }
                                }
                                else if (ssDCval >= 96) {
                                    return 8;
                                }
                            }
                        }
                        else if (mu_mv_mag_sq >= 127) {
                            if (Rs < 7364) {
                                if (ssDCval < 85) {
                                    if (Rs < 4966) {
                                        if (ssDCval < 80) {
                                            if (diffRsCsdiff < 33) {
                                                if (Cs < 4829) {
                                                    return 4;
                                                }
                                                else if (Cs >= 4829) {
                                                    return 8;
                                                }
                                            }
                                            else if (diffRsCsdiff >= 33) {
                                                return 4;
                                            }
                                        }
                                        else if (ssDCval >= 80) {
                                            if (Cs < 4876) {
                                                return 8;
                                            }
                                            else if (Cs >= 4876) {
                                                if (tcor < 92) {
                                                    if (negBalance < 47) {
                                                        if (diffRsCsdiff < -26) {
                                                            return 4;
                                                        }
                                                        else if (diffRsCsdiff >= -26) {
                                                            if (diffAFD < -1) {
                                                                if (SC < 9942) {
                                                                    return 4;
                                                                }
                                                                else if (SC >= 9942) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (diffAFD >= -1) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                    else if (negBalance >= 47) {
                                                        if (Rs < 4924) {
                                                            if (tcor < 91) {
                                                                return 8;
                                                            }
                                                            else if (tcor >= 91) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (Rs >= 4924) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                                else if (tcor >= 92) {
                                                    if (mu_mv_mag_sq < 128) {
                                                        if (negBalance < 47) {
                                                            return 4;
                                                        }
                                                        else if (negBalance >= 47) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (mu_mv_mag_sq >= 128) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (Rs >= 4966) {
                                        if (Rs < 5035) {
                                            if (negBalance < 47) {
                                                return 4;
                                            }
                                            else if (negBalance >= 47) {
                                                if (SC < 10117) {
                                                    return 4;
                                                }
                                                else if (SC >= 10117) {
                                                    return 8;
                                                }
                                            }
                                        }
                                        else if (Rs >= 5035) {
                                            return 4;
                                        }
                                    }
                                }
                                else if (ssDCval >= 85) {
                                    if (refDCval < 106) {
                                        return 8;
                                    }
                                    else if (refDCval >= 106) {
                                        if (Cs < 4969) {
                                            return 8;
                                        }
                                        else if (Cs >= 4969) {
                                            return 4;
                                        }
                                    }
                                }
                            }
                            else if (Rs >= 7364) {
                                if (SC < 14897) {
                                    return 4;
                                }
                                else if (SC >= 14897) {
                                    return 2;
                                }
                            }
                        }
                    }
                }
            }
        }
        else if (tcor >= 95) {
            if (MVDiff < 3) {
                if (AFD < 4) {
                    if (Cs < 708) {
                        if (Rs < 592) {
                            if (Cs < 501) {
                                if (ssDCval < 161) {
                                    if (TSCindex < 2) {
                                        if (mu_mv_mag_sq < 125) {
                                            return 8;
                                        }
                                        else if (mu_mv_mag_sq >= 125) {
                                            if (Cs < 308) {
                                                return 8;
                                            }
                                            else if (Cs >= 308) {
                                                if (Rs < 183) {
                                                    if (Cs < 309) {
                                                        return 4;
                                                    }
                                                    else if (Cs >= 309) {
                                                        return 8;
                                                    }
                                                }
                                                else if (Rs >= 183) {
                                                    if (Rs < 184) {
                                                        return 4;
                                                    }
                                                    else if (Rs >= 184) {
                                                        if (SC < 501) {
                                                            return 4;
                                                        }
                                                        else if (SC >= 501) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (TSCindex >= 2) {
                                        if (negBalance < 20) {
                                            if (Rs < 217) {
                                                if (Rs < 113) {
                                                    return 1;
                                                }
                                                else if (Rs >= 113) {
                                                    return 4;
                                                }
                                            }
                                            else if (Rs >= 217) {
                                                return 8;
                                            }
                                        }
                                        else if (negBalance >= 20) {
                                            if (Rs < 572) {
                                                return 4;
                                            }
                                            else if (Rs >= 572) {
                                                return 8;
                                            }
                                        }
                                    }
                                }
                                else if (ssDCval >= 161) {
                                    return 4;
                                }
                            }
                            else if (Cs >= 501) {
                                if (TSC < 3) {
                                    return 4;
                                }
                                else if (TSC >= 3) {
                                    return 8;
                                }
                            }
                        }
                        else if (Rs >= 592) {
                            if (ssDCval < 33) {
                                if (Cs < 650) {
                                    return 8;
                                }
                                else if (Cs >= 650) {
                                    return 4;
                                }
                            }
                            else if (ssDCval >= 33) {
                                if (ssDCval < 108) {
                                    return 8;
                                }
                                else if (ssDCval >= 108) {
                                    if (Cs < 17) {
                                        return 1;
                                    }
                                    else if (Cs >= 17) {
                                        if (ssDCval < 130) {
                                            return 8;
                                        }
                                        else if (ssDCval >= 130) {
                                            if (Cs < 684) {
                                                return 8;
                                            }
                                            else if (Cs >= 684) {
                                                return 1;
                                            }
                                        }
                                    }
                                }
                            }
                        }
                    }
                    else if (Cs >= 708) {
                        if (Cs < 2146) {
                            if (diffMVdiffVal < -2) {
                                if (mcTcor < 98) {
                                    if (Rs < 1107) {
                                        return 8;
                                    }
                                    else if (Rs >= 1107) {
                                        return 1;
                                    }
                                }
                                else if (mcTcor >= 98) {
                                    if (CsDiff < 1) {
                                        if (Cs < 1179) {
                                            return 8;
                                        }
                                        else if (Cs >= 1179) {
                                            if (ssDCval < 114) {
                                                return 4;
                                            }
                                            else if (ssDCval >= 114) {
                                                return 8;
                                            }
                                        }
                                    }
                                    else if (CsDiff >= 1) {
                                        return 8;
                                    }
                                }
                            }
                            else if (diffMVdiffVal >= -2) {
                                if (RsCsDiff < 1) {
                                    if (ssDCval < 110) {
                                        return 8;
                                    }
                                    else if (ssDCval >= 110) {
                                        if (negBalance < 11) {
                                            return 8;
                                        }
                                        else if (negBalance >= 11) {
                                            if (Rs < 1009) {
                                                return 8;
                                            }
                                            else if (Rs >= 1009) {
                                                return 4;
                                            }
                                        }
                                    }
                                }
                                else if (RsCsDiff >= 1) {
                                    return 8;
                                }
                            }
                        }
                        else if (Cs >= 2146) {
                            if (Rs < 3779) {
                                if (SC < 7764) {
                                    if (Rs < 2011) {
                                        if (refDCval < 85) {
                                            if (Cs < 2538) {
                                                return 1;
                                            }
                                            else if (Cs >= 2538) {
                                                return 8;
                                            }
                                        }
                                        else if (refDCval >= 85) {
                                            if (posBalance < 41) {
                                                return 8;
                                            }
                                            else if (posBalance >= 41) {
                                                return 4;
                                            }
                                        }
                                    }
                                    else if (Rs >= 2011) {
                                        if (Cs < 3716) {
                                            if (mu_mv_mag_sq < 123) {
                                                if (SC < 6273) {
                                                    if (Rs < 3162) {
                                                        if (diffRsCsdiff < -42) {
                                                            return 1;
                                                        }
                                                        else if (diffRsCsdiff >= -42) {
                                                            if (mu_mv_mag_sq < 121) {
                                                                return 8;
                                                            }
                                                            else if (mu_mv_mag_sq >= 121) {
                                                                if (AFD < 3) {
                                                                    return 8;
                                                                }
                                                                else if (AFD >= 3) {
                                                                    return 1;
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (Rs >= 3162) {
                                                        if (Cs < 3046) {
                                                            if (Rs < 3199) {
                                                                return 8;
                                                            }
                                                            else if (Rs >= 3199) {
                                                                return 1;
                                                            }
                                                        }
                                                        else if (Cs >= 3046) {
                                                            return 1;
                                                        }
                                                    }
                                                }
                                                else if (SC >= 6273) {
                                                    if (diffAFD < -1) {
                                                        if (ssDCval < 109) {
                                                            return 8;
                                                        }
                                                        else if (ssDCval >= 109) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (diffAFD >= -1) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                            else if (mu_mv_mag_sq >= 123) {
                                                return 8;
                                            }
                                        }
                                        else if (Cs >= 3716) {
                                            if (Cs < 3877) {
                                                if (RsDiff < 1) {
                                                    if (Rs < 3638) {
                                                        return 1;
                                                    }
                                                    else if (Rs >= 3638) {
                                                        return 8;
                                                    }
                                                }
                                                else if (RsDiff >= 1) {
                                                    return 1;
                                                }
                                            }
                                            else if (Cs >= 3877) {
                                                return 8;
                                            }
                                        }
                                    }
                                }
                                else if (SC >= 7764) {
                                    return 1;
                                }
                            }
                            else if (Rs >= 3779) {
                                return 8;
                            }
                        }
                    }
                }
                else if (AFD >= 4) {
                    if (negBalance < 53) {
                        if (SC < 4848) {
                            if (SC < 3536) {
                                if (Cs < 1194) {
                                    if (posBalance < 61) {
                                        if (AFD < 6) {
                                            if (ssDCval < 98) {
                                                if (diffAFD < -2) {
                                                    if (Cs < 528) {
                                                        return 8;
                                                    }
                                                    else if (Cs >= 528) {
                                                        return 4;
                                                    }
                                                }
                                                else if (diffAFD >= -2) {
                                                    return 4;
                                                }
                                            }
                                            else if (ssDCval >= 98) {
                                                if (Rs < 398) {
                                                    return 2;
                                                }
                                                else if (Rs >= 398) {
                                                    if (Rs < 881) {
                                                        return 4;
                                                    }
                                                    else if (Rs >= 881) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                        }
                                        else if (AFD >= 6) {
                                            return 2;
                                        }
                                    }
                                    else if (posBalance >= 61) {
                                        return 8;
                                    }
                                }
                                else if (Cs >= 1194) {
                                    if (TSC < 6) {
                                        if (Rs < 1775) {
                                            if (mcTcor < 98) {
                                                if (mu_mv_mag_sq < 100) {
                                                    if (SC < 3260) {
                                                        return 8;
                                                    }
                                                    else if (SC >= 3260) {
                                                        return 4;
                                                    }
                                                }
                                                else if (mu_mv_mag_sq >= 100) {
                                                    return 8;
                                                }
                                            }
                                            else if (mcTcor >= 98) {
                                                if (SC < 2769) {
                                                    return 4;
                                                }
                                                else if (SC >= 2769) {
                                                    if (Cs < 1693) {
                                                        return 8;
                                                    }
                                                    else if (Cs >= 1693) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                        }
                                        else if (Rs >= 1775) {
                                            if (mu_mv_mag_sq < 117) {
                                                return 4;
                                            }
                                            else if (mu_mv_mag_sq >= 117) {
                                                if (diffRsCsdiff < -12) {
                                                    return 2;
                                                }
                                                else if (diffRsCsdiff >= -12) {
                                                    if (Cs < 1564) {
                                                        return 8;
                                                    }
                                                    else if (Cs >= 1564) {
                                                        if (Cs < 1606) {
                                                            if (mu_mv_mag_sq < 125) {
                                                                return 4;
                                                            }
                                                            else if (mu_mv_mag_sq >= 125) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (Cs >= 1606) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (TSC >= 6) {
                                        if (refDCval < 78) {
                                            if (RsDiff < 4) {
                                                return 4;
                                            }
                                            else if (RsDiff >= 4) {
                                                return 8;
                                            }
                                        }
                                        else if (refDCval >= 78) {
                                            if (Rs < 1588) {
                                                return 8;
                                            }
                                            else if (Rs >= 1588) {
                                                return 2;
                                            }
                                        }
                                    }
                                }
                            }
                            else if (SC >= 3536) {
                                if (refDCval < 89) {
                                    if (negBalance < 34) {
                                        if (Rs < 2876) {
                                            return 8;
                                        }
                                        else if (Rs >= 2876) {
                                            return 1;
                                        }
                                    }
                                    else if (negBalance >= 34) {
                                        if (negBalance < 43) {
                                            return 4;
                                        }
                                        else if (negBalance >= 43) {
                                            if (TSC < 7) {
                                                if (SC < 3667) {
                                                    if (AFD < 6) {
                                                        if (Rs < 1960) {
                                                            return 8;
                                                        }
                                                        else if (Rs >= 1960) {
                                                            return 2;
                                                        }
                                                    }
                                                    else if (AFD >= 6) {
                                                        return 4;
                                                    }
                                                }
                                                else if (SC >= 3667) {
                                                    return 4;
                                                }
                                            }
                                            else if (TSC >= 7) {
                                                return 2;
                                            }
                                        }
                                    }
                                }
                                else if (refDCval >= 89) {
                                    if (ssDCval < 94) {
                                        if (AFD < 6) {
                                            if (MVDiff < 1) {
                                                return 8;
                                            }
                                            else if (MVDiff >= 1) {
                                                if (RsDiff < 4) {
                                                    return 8;
                                                }
                                                else if (RsDiff >= 4) {
                                                    return 1;
                                                }
                                            }
                                        }
                                        else if (AFD >= 6) {
                                            if (SC < 4525) {
                                                if (Rs < 2299) {
                                                    return 8;
                                                }
                                                else if (Rs >= 2299) {
                                                    if (gchDC < 1) {
                                                        return 1;
                                                    }
                                                    else if (gchDC >= 1) {
                                                        if (Rs < 2504) {
                                                            return 8;
                                                        }
                                                        else if (Rs >= 2504) {
                                                            return 1;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (SC >= 4525) {
                                                return 4;
                                            }
                                        }
                                    }
                                    else if (ssDCval >= 94) {
                                        if (Rs < 2723) {
                                            if (AFD < 8) {
                                                if (ssDCval < 151) {
                                                    if (tcor < 99) {
                                                        return 8;
                                                    }
                                                    else if (tcor >= 99) {
                                                        if (Rs < 1895) {
                                                            return 4;
                                                        }
                                                        else if (Rs >= 1895) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                                else if (ssDCval >= 151) {
                                                    return 4;
                                                }
                                            }
                                            else if (AFD >= 8) {
                                                return 4;
                                            }
                                        }
                                        else if (Rs >= 2723) {
                                            if (Rs < 2873) {
                                                return 4;
                                            }
                                            else if (Rs >= 2873) {
                                                return 8;
                                            }
                                        }
                                    }
                                }
                            }
                        }
                        else if (SC >= 4848) {
                            if (ssDCval < 92) {
                                if (Cs < 3577) {
                                    if (negBalance < 43) {
                                        return 8;
                                    }
                                    else if (negBalance >= 43) {
                                        if (RsCsDiff < 67) {
                                            return 4;
                                        }
                                        else if (RsCsDiff >= 67) {
                                            return 2;
                                        }
                                    }
                                }
                                else if (Cs >= 3577) {
                                    return 8;
                                }
                            }
                            else if (ssDCval >= 92) {
                                if (mu_mv_mag_sq < 123) {
                                    if (Cs < 3163) {
                                        if (Rs < 2669) {
                                            return 8;
                                        }
                                        else if (Rs >= 2669) {
                                            if (refDCval < 112) {
                                                if (mu_mv_mag_sq < 114) {
                                                    if (Rs < 3440) {
                                                        if (CsDiff < 6) {
                                                            return 8;
                                                        }
                                                        else if (CsDiff >= 6) {
                                                            return 1;
                                                        }
                                                    }
                                                    else if (Rs >= 3440) {
                                                        return 4;
                                                    }
                                                }
                                                else if (mu_mv_mag_sq >= 114) {
                                                    if (Cs < 3049) {
                                                        return 8;
                                                    }
                                                    else if (Cs >= 3049) {
                                                        if (RsCsDiff < 37) {
                                                            return 4;
                                                        }
                                                        else if (RsCsDiff >= 37) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (refDCval >= 112) {
                                                if (negBalance < 48) {
                                                    if (ssDCval < 114) {
                                                        if (posBalance < 59) {
                                                            if (SC < 5611) {
                                                                return 8;
                                                            }
                                                            else if (SC >= 5611) {
                                                                if (AFD < 7) {
                                                                    if (tcor < 97) {
                                                                        return 8;
                                                                    }
                                                                    else if (tcor >= 97) {
                                                                        if (AFD < 6) {
                                                                            return 8;
                                                                        }
                                                                        else if (AFD >= 6) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                }
                                                                else if (AFD >= 7) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                        else if (posBalance >= 59) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (ssDCval >= 114) {
                                                        return 8;
                                                    }
                                                }
                                                else if (negBalance >= 48) {
                                                    if (Cs < 2868) {
                                                        return 8;
                                                    }
                                                    else if (Cs >= 2868) {
                                                        if (Scindex < 6) {
                                                            if (mu_mv_mag_sq < 122) {
                                                                return 4;
                                                            }
                                                            else if (mu_mv_mag_sq >= 122) {
                                                                if (Cs < 3004) {
                                                                    if (Rs < 2733) {
                                                                        return 4;
                                                                    }
                                                                    else if (Rs >= 2733) {
                                                                        return 8;
                                                                    }
                                                                }
                                                                else if (Cs >= 3004) {
                                                                    if (RsCsDiff < 46) {
                                                                        if (MVDiff < 1) {
                                                                            return 4;
                                                                        }
                                                                        else if (MVDiff >= 1) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                    else if (RsCsDiff >= 46) {
                                                                        return 8;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (Scindex >= 6) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (Cs >= 3163) {
                                        return 8;
                                    }
                                }
                                else if (mu_mv_mag_sq >= 123) {
                                    if (RsCsDiff < 191) {
                                        if (negBalance < 48) {
                                            return 8;
                                        }
                                        else if (negBalance >= 48) {
                                            if (ssDCval < 123) {
                                                if (Rs < 5755) {
                                                    if (SC < 5750) {
                                                        return 8;
                                                    }
                                                    else if (SC >= 5750) {
                                                        if (Rs < 2872) {
                                                            if (Rs < 2861) {
                                                                if (Rs < 2809) {
                                                                    if (Rs < 2802) {
                                                                        if (gchDC < 1) {
                                                                            if (SC < 5761) {
                                                                                return 4;
                                                                            }
                                                                            else if (SC >= 5761) {
                                                                                return 8;
                                                                            }
                                                                        }
                                                                        else if (gchDC >= 1) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                    else if (Rs >= 2802) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (Rs >= 2809) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (Rs >= 2861) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (Rs >= 2872) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                                else if (Rs >= 5755) {
                                                    return 4;
                                                }
                                            }
                                            else if (ssDCval >= 123) {
                                                return 8;
                                            }
                                        }
                                    }
                                    else if (RsCsDiff >= 191) {
                                        if (refDCval < 106) {
                                            return 8;
                                        }
                                        else if (refDCval >= 106) {
                                            if (refDCval < 115) {
                                                if (Rs < 5436) {
                                                    return 8;
                                                }
                                                else if (Rs >= 5436) {
                                                    return 4;
                                                }
                                            }
                                            else if (refDCval >= 115) {
                                                return 8;
                                            }
                                        }
                                    }
                                }
                            }
                        }
                    }
                    else if (negBalance >= 53) {
                        if (mu_mv_mag_sq < 128) {
                            if (Rs < 5675) {
                                if (Cs < 5373) {
                                    if (diffAFD < 1) {
                                        if (Rs < 328) {
                                            return 8;
                                        }
                                        else if (Rs >= 328) {
                                            if (Cs < 2827) {
                                                return 4;
                                            }
                                            else if (Cs >= 2827) {
                                                if (SC < 10895) {
                                                    return 8;
                                                }
                                                else if (SC >= 10895) {
                                                    if (Cs < 5311) {
                                                        return 4;
                                                    }
                                                    else if (Cs >= 5311) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (diffAFD >= 1) {
                                        return 8;
                                    }
                                }
                                else if (Cs >= 5373) {
                                    return 4;
                                }
                            }
                            else if (Rs >= 5675) {
                                return 4;
                            }
                        }
                        else if (mu_mv_mag_sq >= 128) {
                            if (RsDiff < 5) {
                                return 8;
                            }
                            else if (RsDiff >= 5) {
                                return 4;
                            }
                        }
                    }
                }
            }
            else if (MVDiff >= 3) {
                if (mu_mv_mag_sq < 93) {
                    if (Rs < 3000) {
                        if (posBalance < 43) {
                            if (refDCval < 71) {
                                if (negBalance < 91) {
                                    if (mu_mv_mag_sq < 56) {
                                        if (AFD < 2) {
                                            if (mcTcor < 85) {
                                                if (MVDiff < 111) {
                                                    if (diffMVdiffVal < 25) {
                                                        return 4;
                                                    }
                                                    else if (diffMVdiffVal >= 25) {
                                                        return 2;
                                                    }
                                                }
                                                else if (MVDiff >= 111) {
                                                    if (refDCval < 22) {
                                                        return 1;
                                                    }
                                                    else if (refDCval >= 22) {
                                                        return 2;
                                                    }
                                                }
                                            }
                                            else if (mcTcor >= 85) {
                                                if (MVDiff < 31) {
                                                    return 8;
                                                }
                                                else if (MVDiff >= 31) {
                                                    if (negBalance < 25) {
                                                        return 8;
                                                    }
                                                    else if (negBalance >= 25) {
                                                        return 2;
                                                    }
                                                }
                                            }
                                        }
                                        else if (AFD >= 2) {
                                            if (negBalance < 50) {
                                                if (Cs < 752) {
                                                    return 2;
                                                }
                                                else if (Cs >= 752) {
                                                    if (Cs < 976) {
                                                        return 4;
                                                    }
                                                    else if (Cs >= 976) {
                                                        return 2;
                                                    }
                                                }
                                            }
                                            else if (negBalance >= 50) {
                                                if (CsDiff < 2) {
                                                    if (mcTcor < 83) {
                                                        return 1;
                                                    }
                                                    else if (mcTcor >= 83) {
                                                        if (diffAFD < -2) {
                                                            return 2;
                                                        }
                                                        else if (diffAFD >= -2) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                                else if (CsDiff >= 2) {
                                                    if (SC < 1525) {
                                                        return 2;
                                                    }
                                                    else if (SC >= 1525) {
                                                        return 1;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (mu_mv_mag_sq >= 56) {
                                        if (SC < 1655) {
                                            if (refDCval < 23) {
                                                return 8;
                                            }
                                            else if (refDCval >= 23) {
                                                return 4;
                                            }
                                        }
                                        else if (SC >= 1655) {
                                            if (Cs < 1063) {
                                                if (AFD < 4) {
                                                    return 1;
                                                }
                                                else if (AFD >= 4) {
                                                    return 4;
                                                }
                                            }
                                            else if (Cs >= 1063) {
                                                return 4;
                                            }
                                        }
                                    }
                                }
                                else if (negBalance >= 91) {
                                    if (tcor < 119) {
                                        return 8;
                                    }
                                    else if (tcor >= 119) {
                                        if (posBalance < 2) {
                                            if (AFD < 4) {
                                                return 2;
                                            }
                                            else if (AFD >= 4) {
                                                return 8;
                                            }
                                        }
                                        else if (posBalance >= 2) {
                                            if (MVDiff < 84) {
                                                return 2;
                                            }
                                            else if (MVDiff >= 84) {
                                                return 1;
                                            }
                                        }
                                    }
                                }
                            }
                            else if (refDCval >= 71) {
                                if (Cs < 220) {
                                    if (negBalance < 40) {
                                        return 1;
                                    }
                                    else if (negBalance >= 40) {
                                        return 8;
                                    }
                                }
                                else if (Cs >= 220) {
                                    if (tcor < 103) {
                                        if (AFD < 15) {
                                            if (mcTcor < 99) {
                                                if (Cs < 758) {
                                                    if (mu_mv_mag_sq < 90) {
                                                        if (SC < 3022) {
                                                            return 4;
                                                        }
                                                        else if (SC >= 3022) {
                                                            return 2;
                                                        }
                                                    }
                                                    else if (mu_mv_mag_sq >= 90) {
                                                        return 2;
                                                    }
                                                }
                                                else if (Cs >= 758) {
                                                    if (refDCval < 106) {
                                                        if (posBalance < 23) {
                                                            return 8;
                                                        }
                                                        else if (posBalance >= 23) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (refDCval >= 106) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                            else if (mcTcor >= 99) {
                                                return 4;
                                            }
                                        }
                                        else if (AFD >= 15) {
                                            if (CsDiff < 9) {
                                                if (CsDiff < 3) {
                                                    return 4;
                                                }
                                                else if (CsDiff >= 3) {
                                                    return 1;
                                                }
                                            }
                                            else if (CsDiff >= 9) {
                                                return 2;
                                            }
                                        }
                                    }
                                    else if (tcor >= 103) {
                                        if (diffTSC < -2) {
                                            if (SC < 1017) {
                                                return 1;
                                            }
                                            else if (SC >= 1017) {
                                                if (TSC < 8) {
                                                    return 4;
                                                }
                                                else if (TSC >= 8) {
                                                    return 2;
                                                }
                                            }
                                        }
                                        else if (diffTSC >= -2) {
                                            if (AFD < 19) {
                                                return 8;
                                            }
                                            else if (AFD >= 19) {
                                                return 4;
                                            }
                                        }
                                    }
                                }
                            }
                        }
                        else if (posBalance >= 43) {
                            if (Rs < 2306) {
                                if (mu_mv_mag_sq < 66) {
                                    if (SC < 887) {
                                        if (posBalance < 76) {
                                            if (RsCsDiff < 1) {
                                                return 1;
                                            }
                                            else if (RsCsDiff >= 1) {
                                                if (MVDiff < 29) {
                                                    if (diffMVdiffVal < -1) {
                                                        return 1;
                                                    }
                                                    else if (diffMVdiffVal >= -1) {
                                                        return 2;
                                                    }
                                                }
                                                else if (MVDiff >= 29) {
                                                    return 1;
                                                }
                                            }
                                        }
                                        else if (posBalance >= 76) {
                                            if (SC < 6) {
                                                return 4;
                                            }
                                            else if (SC >= 6) {
                                                if (AFD < 9) {
                                                    return 8;
                                                }
                                                else if (AFD >= 9) {
                                                    return 1;
                                                }
                                            }
                                        }
                                    }
                                    else if (SC >= 887) {
                                        if (Rs < 1189) {
                                            if (SC < 1050) {
                                                return 8;
                                            }
                                            else if (SC >= 1050) {
                                                if (TSC < 4) {
                                                    return 4;
                                                }
                                                else if (TSC >= 4) {
                                                    if (diffMVdiffVal < 4) {
                                                        return 8;
                                                    }
                                                    else if (diffMVdiffVal >= 4) {
                                                        if (SC < 1645) {
                                                            return 8;
                                                        }
                                                        else if (SC >= 1645) {
                                                            return 1;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (Rs >= 1189) {
                                            if (Cs < 699) {
                                                return 4;
                                            }
                                            else if (Cs >= 699) {
                                                return 1;
                                            }
                                        }
                                    }
                                }
                                else if (mu_mv_mag_sq >= 66) {
                                    if (refDCval < 65) {
                                        if (SC < 1398) {
                                            if (Rs < 664) {
                                                return 8;
                                            }
                                            else if (Rs >= 664) {
                                                return 4;
                                            }
                                        }
                                        else if (SC >= 1398) {
                                            if (Cs < 960) {
                                                if (mu_mv_mag_sq < 82) {
                                                    return 1;
                                                }
                                                else if (mu_mv_mag_sq >= 82) {
                                                    if (diffMVdiffVal < 2) {
                                                        if (Rs < 1456) {
                                                            return 1;
                                                        }
                                                        else if (Rs >= 1456) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (diffMVdiffVal >= 2) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                            else if (Cs >= 960) {
                                                return 8;
                                            }
                                        }
                                    }
                                    else if (refDCval >= 65) {
                                        if (refDCval < 83) {
                                            if (ssDCval < 72) {
                                                return 2;
                                            }
                                            else if (ssDCval >= 72) {
                                                return 4;
                                            }
                                        }
                                        else if (refDCval >= 83) {
                                            if (refDCval < 161) {
                                                if (posBalance < 60) {
                                                    if (Cs < 1430) {
                                                        if (negBalance < 32) {
                                                            return 4;
                                                        }
                                                        else if (negBalance >= 32) {
                                                            if (RsDiff < 1) {
                                                                return 2;
                                                            }
                                                            else if (RsDiff >= 1) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                    else if (Cs >= 1430) {
                                                        return 4;
                                                    }
                                                }
                                                else if (posBalance >= 60) {
                                                    return 8;
                                                }
                                            }
                                            else if (refDCval >= 161) {
                                                return 2;
                                            }
                                        }
                                    }
                                }
                            }
                            else if (Rs >= 2306) {
                                if (Cs < 2392) {
                                    if (Cs < 1559) {
                                        return 8;
                                    }
                                    else if (Cs >= 1559) {
                                        return 1;
                                    }
                                }
                                else if (Cs >= 2392) {
                                    return 8;
                                }
                            }
                        }
                    }
                    else if (Rs >= 3000) {
                        if (ssDCval < 109) {
                            if (refDCval < 54) {
                                if (Cs < 3316) {
                                    return 8;
                                }
                                else if (Cs >= 3316) {
                                    return 4;
                                }
                            }
                            else if (refDCval >= 54) {
                                if (mu_mv_mag_sq < 70) {
                                    if (SC < 6965) {
                                        if (ssDCval < 108) {
                                            return 8;
                                        }
                                        else if (ssDCval >= 108) {
                                            if (Rs < 3388) {
                                                if (CsDiff < 4) {
                                                    return 8;
                                                }
                                                else if (CsDiff >= 4) {
                                                    return 4;
                                                }
                                            }
                                            else if (Rs >= 3388) {
                                                return 8;
                                            }
                                        }
                                    }
                                    else if (SC >= 6965) {
                                        return 4;
                                    }
                                }
                                else if (mu_mv_mag_sq >= 70) {
                                    return 8;
                                }
                            }
                        }
                        else if (ssDCval >= 109) {
                            if (Rs < 3488) {
                                if (Cs < 2269) {
                                    return 4;
                                }
                                else if (Cs >= 2269) {
                                    return 8;
                                }
                            }
                            else if (Rs >= 3488) {
                                if (refDCval < 116) {
                                    if (diffMVdiffVal < -5) {
                                        return 4;
                                    }
                                    else if (diffMVdiffVal >= -5) {
                                        return 2;
                                    }
                                }
                                else if (refDCval >= 116) {
                                    if (Cs < 2272) {
                                        return 4;
                                    }
                                    else if (Cs >= 2272) {
                                        if (Cs < 2330) {
                                            return 2;
                                        }
                                        else if (Cs >= 2330) {
                                            return 4;
                                        }
                                    }
                                }
                            }
                        }
                    }
                }
                else if (mu_mv_mag_sq >= 93) {
                    if (SC < 2327) {
                        if (ssDCval < 160) {
                            if (SC < 2285) {
                                if (posBalance < 10) {
                                    if (refDCval < 29) {
                                        return 8;
                                    }
                                    else if (refDCval >= 29) {
                                        return 1;
                                    }
                                }
                                else if (posBalance >= 10) {
                                    if (tcor < 98) {
                                        if (diffAFD < -1) {
                                            if (Rs < 1095) {
                                                if (Rs < 641) {
                                                    return 4;
                                                }
                                                else if (Rs >= 641) {
                                                    return 8;
                                                }
                                            }
                                            else if (Rs >= 1095) {
                                                return 1;
                                            }
                                        }
                                        else if (diffAFD >= -1) {
                                            return 8;
                                        }
                                    }
                                    else if (tcor >= 98) {
                                        if (ssDCval < 26) {
                                            return 8;
                                        }
                                        else if (ssDCval >= 26) {
                                            if (Rs < 917) {
                                                if (gchDC < 2) {
                                                    if (diffTSC < -7) {
                                                        if (diffAFD < -20) {
                                                            return 4;
                                                        }
                                                        else if (diffAFD >= -20) {
                                                            return 1;
                                                        }
                                                    }
                                                    else if (diffTSC >= -7) {
                                                        return 8;
                                                    }
                                                }
                                                else if (gchDC >= 2) {
                                                    if (Cs < 121) {
                                                        return 2;
                                                    }
                                                    else if (Cs >= 121) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                            else if (Rs >= 917) {
                                                return 4;
                                            }
                                        }
                                    }
                                }
                            }
                            else if (SC >= 2285) {
                                return 2;
                            }
                        }
                        else if (ssDCval >= 160) {
                            if (RsDiff < 1) {
                                if (SC < 885) {
                                    return 4;
                                }
                                else if (SC >= 885) {
                                    if (negBalance < 9) {
                                        return 4;
                                    }
                                    else if (negBalance >= 9) {
                                        if (refDCval < 164) {
                                            return 4;
                                        }
                                        else if (refDCval >= 164) {
                                            return 2;
                                        }
                                    }
                                }
                            }
                            else if (RsDiff >= 1) {
                                return 2;
                            }
                        }
                    }
                    else if (SC >= 2327) {
                        if (SC < 4574) {
                            if (ssDCval < 73) {
                                if (CsDiff < 6) {
                                    return 8;
                                }
                                else if (CsDiff >= 6) {
                                    if (Rs < 2187) {
                                        return 4;
                                    }
                                    else if (Rs >= 2187) {
                                        return 8;
                                    }
                                }
                            }
                            else if (ssDCval >= 73) {
                                if (posBalance < 53) {
                                    if (Cs < 2083) {
                                        if (diffTSC < -6) {
                                            return 4;
                                        }
                                        else if (diffTSC >= -6) {
                                            if (refDCval < 90) {
                                                if (Cs < 1682) {
                                                    return 4;
                                                }
                                                else if (Cs >= 1682) {
                                                    return 8;
                                                }
                                            }
                                            else if (refDCval >= 90) {
                                                return 8;
                                            }
                                        }
                                    }
                                    else if (Cs >= 2083) {
                                        if (Rs < 1790) {
                                            return 4;
                                        }
                                        else if (Rs >= 1790) {
                                            return 1;
                                        }
                                    }
                                }
                                else if (posBalance >= 53) {
                                    if (CsDiff < 3) {
                                        return 4;
                                    }
                                    else if (CsDiff >= 3) {
                                        if (posBalance < 54) {
                                            if (Cs < 1697) {
                                                return 1;
                                            }
                                            else if (Cs >= 1697) {
                                                return 4;
                                            }
                                        }
                                        else if (posBalance >= 54) {
                                            return 8;
                                        }
                                    }
                                }
                            }
                        }
                        else if (SC >= 4574) {
                            if (SC < 4645) {
                                if (CsDiff < 4) {
                                    return 8;
                                }
                                else if (CsDiff >= 4) {
                                    return 2;
                                }
                            }
                            else if (SC >= 4645) {
                                return 8;
                            }
                        }
                    }
                }
            }
        }
        return 4;
    }
    mfxU16 AGOPSelect8(mfxU32 Rs, mfxU32 CsDiff, mfxU32 TSC, mfxI32 diffAFD, mfxU32 Scindex, mfxU32 RsDiff, mfxI32 diffMVdiffVal, mfxU32 RsCsDiff, mfxU32 TSCindex, mfxI16 tcor, mfxU32 refDCval, mfxU32 MVDiff, mfxU32 AFD, mfxI32 diffTSC, mfxI32 diffRsCsdiff, mfxU32 posBalance, mfxU32 SC, mfxU32 mu_mv_mag_sq, mfxU32 negBalance, mfxI16 mcTcor, mfxU32 gchDC, mfxU32 Cs, mfxU32 ssDCval) {
        if (mu_mv_mag_sq < 83) {
            if (ssDCval < 8) {
                if (Rs < 1843) {
                    if (Rs < 366) {
                        return 4;
                    }
                    else if (Rs >= 366) {
                        return 8;
                    }
                }
                else if (Rs >= 1843) {
                    if (Cs < 4248) {
                        if (Rs < 1997) {
                            return 1;
                        }
                        else if (Rs >= 1997) {
                            if (SC < 7709) {
                                return 8;
                            }
                            else if (SC >= 7709) {
                                if (Rs < 3770) {
                                    return 1;
                                }
                                else if (Rs >= 3770) {
                                    if (Cs < 3951) {
                                        return 1;
                                    }
                                    else if (Cs >= 3951) {
                                        return 8;
                                    }
                                }
                            }
                        }
                    }
                    else if (Cs >= 4248) {
                        if (SC < 9327) {
                            return 8;
                        }
                        else if (SC >= 9327) {
                            if (Cs < 5084) {
                                return 1;
                            }
                            else if (Cs >= 5084) {
                                return 8;
                            }
                        }
                    }
                }
            }
            else if (ssDCval >= 8) {
                if (Cs < 3288) {
                    if (Cs < 1800) {
                        if (AFD < 13) {
                            if (MVDiff < 16) {
                                if (RsCsDiff < 2) {
                                    if (Rs < 541) {
                                        if (mu_mv_mag_sq < 36) {
                                            if (tcor < 61) {
                                                return 4;
                                            }
                                            else if (tcor >= 61) {
                                                if (ssDCval < 57) {
                                                    if (TSCindex < 5) {
                                                        if (mu_mv_mag_sq < 29) {
                                                            return 4;
                                                        }
                                                        else if (mu_mv_mag_sq >= 29) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (TSCindex >= 5) {
                                                        return 2;
                                                    }
                                                }
                                                else if (ssDCval >= 57) {
                                                    return 2;
                                                }
                                            }
                                        }
                                        else if (mu_mv_mag_sq >= 36) {
                                            if (posBalance < 6) {
                                                return 8;
                                            }
                                            else if (posBalance >= 6) {
                                                if (ssDCval < 102) {
                                                    if (posBalance < 107) {
                                                        return 4;
                                                    }
                                                    else if (posBalance >= 107) {
                                                        return 8;
                                                    }
                                                }
                                                else if (ssDCval >= 102) {
                                                    if (TSC < 2) {
                                                        return 2;
                                                    }
                                                    else if (TSC >= 2) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (Rs >= 541) {
                                        if (Rs < 1467) {
                                            if (negBalance < 17) {
                                                if (Rs < 867) {
                                                    if (TSCindex < 3) {
                                                        return 1;
                                                    }
                                                    else if (TSCindex >= 3) {
                                                        if (SC < 1005) {
                                                            return 8;
                                                        }
                                                        else if (SC >= 1005) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                                else if (Rs >= 867) {
                                                    return 8;
                                                }
                                            }
                                            else if (negBalance >= 17) {
                                                if (mcTcor < 96) {
                                                    if (Cs < 562) {
                                                        return 8;
                                                    }
                                                    else if (Cs >= 562) {
                                                        return 4;
                                                    }
                                                }
                                                else if (mcTcor >= 96) {
                                                    return 8;
                                                }
                                            }
                                        }
                                        else if (Rs >= 1467) {
                                            if (Cs < 1049) {
                                                return 1;
                                            }
                                            else if (Cs >= 1049) {
                                                return 4;
                                            }
                                        }
                                    }
                                }
                                else if (RsCsDiff >= 2) {
                                    if (negBalance < 46) {
                                        if (Rs < 2715) {
                                            if (Cs < 622) {
                                                if (tcor < 83) {
                                                    if (RsDiff < 6) {
                                                        return 4;
                                                    }
                                                    else if (RsDiff >= 6) {
                                                        return 1;
                                                    }
                                                }
                                                else if (tcor >= 83) {
                                                    if (SC < 2912) {
                                                        if (posBalance < 50) {
                                                            if (mcTcor < 98) {
                                                                return 4;
                                                            }
                                                            else if (mcTcor >= 98) {
                                                                if (AFD < 4) {
                                                                    return 8;
                                                                }
                                                                else if (AFD >= 4) {
                                                                    if (RsCsDiff < 4) {
                                                                        if (diffRsCsdiff < -2) {
                                                                            if (Rs < 1596) {
                                                                                return 2;
                                                                            }
                                                                            else if (Rs >= 1596) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                        else if (diffRsCsdiff >= -2) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                    else if (RsCsDiff >= 4) {
                                                                        return 2;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (posBalance >= 50) {
                                                            if (mu_mv_mag_sq < 5) {
                                                                if (refDCval < 55) {
                                                                    if (refDCval < 54) {
                                                                        if (AFD < 5) {
                                                                            return 8;
                                                                        }
                                                                        else if (AFD >= 5) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                    else if (refDCval >= 54) {
                                                                        return 8;
                                                                    }
                                                                }
                                                                else if (refDCval >= 55) {
                                                                    if (SC < 805) {
                                                                        return 2;
                                                                    }
                                                                    else if (SC >= 805) {
                                                                        return 4;
                                                                    }
                                                                }
                                                            }
                                                            else if (mu_mv_mag_sq >= 5) {
                                                                if (SC < 1515) {
                                                                    if (diffMVdiffVal < -5) {
                                                                        return 2;
                                                                    }
                                                                    else if (diffMVdiffVal >= -5) {
                                                                        if (MVDiff < 11) {
                                                                            return 8;
                                                                        }
                                                                        else if (MVDiff >= 11) {
                                                                            if (RsDiff < 2) {
                                                                                if (MVDiff < 15) {
                                                                                    return 4;
                                                                                }
                                                                                else if (MVDiff >= 15) {
                                                                                    return 8;
                                                                                }
                                                                            }
                                                                            else if (RsDiff >= 2) {
                                                                                return 8;
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                                else if (SC >= 1515) {
                                                                    if (Cs < 415) {
                                                                        return 8;
                                                                    }
                                                                    else if (Cs >= 415) {
                                                                        if (SC < 2415) {
                                                                            return 2;
                                                                        }
                                                                        else if (SC >= 2415) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (SC >= 2912) {
                                                        if (MVDiff < 2) {
                                                            return 8;
                                                        }
                                                        else if (MVDiff >= 2) {
                                                            return 2;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (Cs >= 622) {
                                                if (refDCval < 92) {
                                                    if (Rs < 1133) {
                                                        if (CsDiff < 7) {
                                                            if (negBalance < 42) {
                                                                if (posBalance < 50) {
                                                                    if (SC < 1655) {
                                                                        return 4;
                                                                    }
                                                                    else if (SC >= 1655) {
                                                                        if (mu_mv_mag_sq < 69) {
                                                                            if (negBalance < 39) {
                                                                                if (tcor < 94) {
                                                                                    return 1;
                                                                                }
                                                                                else if (tcor >= 94) {
                                                                                    return 4;
                                                                                }
                                                                            }
                                                                            else if (negBalance >= 39) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                        else if (mu_mv_mag_sq >= 69) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                }
                                                                else if (posBalance >= 50) {
                                                                    if (diffRsCsdiff < -15) {
                                                                        if (Cs < 749) {
                                                                            return 4;
                                                                        }
                                                                        else if (Cs >= 749) {
                                                                            return 2;
                                                                        }
                                                                    }
                                                                    else if (diffRsCsdiff >= -15) {
                                                                        return 4;
                                                                    }
                                                                }
                                                            }
                                                            else if (negBalance >= 42) {
                                                                if (posBalance < 48) {
                                                                    if (Cs < 725) {
                                                                        return 1;
                                                                    }
                                                                    else if (Cs >= 725) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (posBalance >= 48) {
                                                                    if (mcTcor < 92) {
                                                                        return 8;
                                                                    }
                                                                    else if (mcTcor >= 92) {
                                                                        return 1;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (CsDiff >= 7) {
                                                            if (Rs < 1060) {
                                                                return 2;
                                                            }
                                                            else if (Rs >= 1060) {
                                                                return 1;
                                                            }
                                                        }
                                                    }
                                                    else if (Rs >= 1133) {
                                                        if (Cs < 1011) {
                                                            if (Rs < 1944) {
                                                                if (ssDCval < 52) {
                                                                    if (posBalance < 54) {
                                                                        if (diffMVdiffVal < -8) {
                                                                            return 4;
                                                                        }
                                                                        else if (diffMVdiffVal >= -8) {
                                                                            if (SC < 2308) {
                                                                                return 1;
                                                                            }
                                                                            else if (SC >= 2308) {
                                                                                if (negBalance < 44) {
                                                                                    if (RsDiff < 4) {
                                                                                        return 1;
                                                                                    }
                                                                                    else if (RsDiff >= 4) {
                                                                                        if (SC < 2393) {
                                                                                            return 4;
                                                                                        }
                                                                                        else if (SC >= 2393) {
                                                                                            return 1;
                                                                                        }
                                                                                    }
                                                                                }
                                                                                else if (negBalance >= 44) {
                                                                                    return 4;
                                                                                }
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (posBalance >= 54) {
                                                                        if (negBalance < 38) {
                                                                            return 1;
                                                                        }
                                                                        else if (negBalance >= 38) {
                                                                            if (SC < 2385) {
                                                                                if (Cs < 850) {
                                                                                    if (Cs < 769) {
                                                                                        return 1;
                                                                                    }
                                                                                    else if (Cs >= 769) {
                                                                                        return 8;
                                                                                    }
                                                                                }
                                                                                else if (Cs >= 850) {
                                                                                    return 1;
                                                                                }
                                                                            }
                                                                            else if (SC >= 2385) {
                                                                                return 8;
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                                else if (ssDCval >= 52) {
                                                                    if (mcTcor < 95) {
                                                                        if (posBalance < 52) {
                                                                            if (mu_mv_mag_sq < 47) {
                                                                                return 1;
                                                                            }
                                                                            else if (mu_mv_mag_sq >= 47) {
                                                                                if (diffMVdiffVal < -5) {
                                                                                    return 1;
                                                                                }
                                                                                else if (diffMVdiffVal >= -5) {
                                                                                    if (tcor < 94) {
                                                                                        return 4;
                                                                                    }
                                                                                    else if (tcor >= 94) {
                                                                                        return 1;
                                                                                    }
                                                                                }
                                                                            }
                                                                        }
                                                                        else if (posBalance >= 52) {
                                                                            if (Rs < 1192) {
                                                                                return 4;
                                                                            }
                                                                            else if (Rs >= 1192) {
                                                                                return 1;
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (mcTcor >= 95) {
                                                                        return 1;
                                                                    }
                                                                }
                                                            }
                                                            else if (Rs >= 1944) {
                                                                if (Rs < 2624) {
                                                                    return 4;
                                                                }
                                                                else if (Rs >= 2624) {
                                                                    return 2;
                                                                }
                                                            }
                                                        }
                                                        else if (Cs >= 1011) {
                                                            if (Cs < 1094) {
                                                                if (Rs < 1498) {
                                                                    if (Rs < 1348) {
                                                                        if (mcTcor < 92) {
                                                                            return 4;
                                                                        }
                                                                        else if (mcTcor >= 92) {
                                                                            if (Rs < 1219) {
                                                                                return 8;
                                                                            }
                                                                            else if (Rs >= 1219) {
                                                                                return 1;
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (Rs >= 1348) {
                                                                        if (Rs < 1477) {
                                                                            return 8;
                                                                        }
                                                                        else if (Rs >= 1477) {
                                                                            if (Cs < 1062) {
                                                                                return 1;
                                                                            }
                                                                            else if (Cs >= 1062) {
                                                                                return 8;
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                                else if (Rs >= 1498) {
                                                                    if (posBalance < 55) {
                                                                        if (refDCval < 52) {
                                                                            return 1;
                                                                        }
                                                                        else if (refDCval >= 52) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                    else if (posBalance >= 55) {
                                                                        return 1;
                                                                    }
                                                                }
                                                            }
                                                            else if (Cs >= 1094) {
                                                                if (Rs < 2367) {
                                                                    if (refDCval < 72) {
                                                                        if (ssDCval < 60) {
                                                                            if (AFD < 7) {
                                                                                if (mu_mv_mag_sq < 67) {
                                                                                    return 1;
                                                                                }
                                                                                else if (mu_mv_mag_sq >= 67) {
                                                                                    if (MVDiff < 6) {
                                                                                        return 1;
                                                                                    }
                                                                                    else if (MVDiff >= 6) {
                                                                                        if (diffMVdiffVal < -3) {
                                                                                            if (SC < 3444) {
                                                                                                return 4;
                                                                                            }
                                                                                            else if (SC >= 3444) {
                                                                                                return 1;
                                                                                            }
                                                                                        }
                                                                                        else if (diffMVdiffVal >= -3) {
                                                                                            return 4;
                                                                                        }
                                                                                    }
                                                                                }
                                                                            }
                                                                            else if (AFD >= 7) {
                                                                                if (negBalance < 40) {
                                                                                    return 4;
                                                                                }
                                                                                else if (negBalance >= 40) {
                                                                                    if (Cs < 1115) {
                                                                                        return 4;
                                                                                    }
                                                                                    else if (Cs >= 1115) {
                                                                                        return 1;
                                                                                    }
                                                                                }
                                                                            }
                                                                        }
                                                                        else if (ssDCval >= 60) {
                                                                            if (TSC < 8) {
                                                                                return 1;
                                                                            }
                                                                            else if (TSC >= 8) {
                                                                                return 8;
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (refDCval >= 72) {
                                                                        if (Cs < 1627) {
                                                                            return 4;
                                                                        }
                                                                        else if (Cs >= 1627) {
                                                                            if (Rs < 1994) {
                                                                                return 2;
                                                                            }
                                                                            else if (Rs >= 1994) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                                else if (Rs >= 2367) {
                                                                    return 1;
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (refDCval >= 92) {
                                                    if (refDCval < 123) {
                                                        if (TSC < 6) {
                                                            if (posBalance < 42) {
                                                                if (Cs < 1194) {
                                                                    return 8;
                                                                }
                                                                else if (Cs >= 1194) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (posBalance >= 42) {
                                                                if (AFD < 10) {
                                                                    return 4;
                                                                }
                                                                else if (AFD >= 10) {
                                                                    return 2;
                                                                }
                                                            }
                                                        }
                                                        else if (TSC >= 6) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (refDCval >= 123) {
                                                        if (ssDCval < 142) {
                                                            if (Cs < 1452) {
                                                                if (posBalance < 41) {
                                                                    if (Rs < 1913) {
                                                                        return 8;
                                                                    }
                                                                    else if (Rs >= 1913) {
                                                                        return 2;
                                                                    }
                                                                }
                                                                else if (posBalance >= 41) {
                                                                    if (RsDiff < 10) {
                                                                        return 1;
                                                                    }
                                                                    else if (RsDiff >= 10) {
                                                                        return 8;
                                                                    }
                                                                }
                                                            }
                                                            else if (Cs >= 1452) {
                                                                if (mcTcor < 91) {
                                                                    return 1;
                                                                }
                                                                else if (mcTcor >= 91) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                        else if (ssDCval >= 142) {
                                                            if (MVDiff < 11) {
                                                                return 4;
                                                            }
                                                            else if (MVDiff >= 11) {
                                                                if (SC < 2293) {
                                                                    return 4;
                                                                }
                                                                else if (SC >= 2293) {
                                                                    return 1;
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (Rs >= 2715) {
                                            if (SC < 3931) {
                                                if (ssDCval < 87) {
                                                    return 2;
                                                }
                                                else if (ssDCval >= 87) {
                                                    return 8;
                                                }
                                            }
                                            else if (SC >= 3931) {
                                                if (posBalance < 41) {
                                                    return 4;
                                                }
                                                else if (posBalance >= 41) {
                                                    if (Cs < 1550) {
                                                        if (ssDCval < 112) {
                                                            if (mu_mv_mag_sq < 74) {
                                                                if (tcor < 96) {
                                                                    return 4;
                                                                }
                                                                else if (tcor >= 96) {
                                                                    if (Cs < 1477) {
                                                                        return 4;
                                                                    }
                                                                    else if (Cs >= 1477) {
                                                                        return 8;
                                                                    }
                                                                }
                                                            }
                                                            else if (mu_mv_mag_sq >= 74) {
                                                                return 1;
                                                            }
                                                        }
                                                        else if (ssDCval >= 112) {
                                                            return 2;
                                                        }
                                                    }
                                                    else if (Cs >= 1550) {
                                                        if (ssDCval < 108) {
                                                            if (MVDiff < 11) {
                                                                if (negBalance < 40) {
                                                                    return 8;
                                                                }
                                                                else if (negBalance >= 40) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (MVDiff >= 11) {
                                                                if (AFD < 7) {
                                                                    if (SC < 4802) {
                                                                        if (AFD < 6) {
                                                                            return 8;
                                                                        }
                                                                        else if (AFD >= 6) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                    else if (SC >= 4802) {
                                                                        return 8;
                                                                    }
                                                                }
                                                                else if (AFD >= 7) {
                                                                    if (ssDCval < 106) {
                                                                        return 8;
                                                                    }
                                                                    else if (ssDCval >= 106) {
                                                                        return 4;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (ssDCval >= 108) {
                                                            if (CsDiff < 5) {
                                                                if (Cs < 1675) {
                                                                    return 4;
                                                                }
                                                                else if (Cs >= 1675) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (CsDiff >= 5) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (negBalance >= 46) {
                                        if (refDCval < 97) {
                                            if (mcTcor < 92) {
                                                if (mu_mv_mag_sq < 2) {
                                                    if (posBalance < 59) {
                                                        if (diffRsCsdiff < 5) {
                                                            return 2;
                                                        }
                                                        else if (diffRsCsdiff >= 5) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (posBalance >= 59) {
                                                        if (SC < 3613) {
                                                            if (refDCval < 63) {
                                                                return 8;
                                                            }
                                                            else if (refDCval >= 63) {
                                                                if (refDCval < 93) {
                                                                    return 4;
                                                                }
                                                                else if (refDCval >= 93) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                        else if (SC >= 3613) {
                                                            return 2;
                                                        }
                                                    }
                                                }
                                                else if (mu_mv_mag_sq >= 2) {
                                                    if (Cs < 498) {
                                                        if (mcTcor < 88) {
                                                            return 2;
                                                        }
                                                        else if (mcTcor >= 88) {
                                                            if (Cs < 240) {
                                                                return 2;
                                                            }
                                                            else if (Cs >= 240) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                    else if (Cs >= 498) {
                                                        if (RsCsDiff < 46) {
                                                            if (refDCval < 92) {
                                                                if (posBalance < 57) {
                                                                    if (diffAFD < 1) {
                                                                        return 8;
                                                                    }
                                                                    else if (diffAFD >= 1) {
                                                                        if (Rs < 1206) {
                                                                            return 4;
                                                                        }
                                                                        else if (Rs >= 1206) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                }
                                                                else if (posBalance >= 57) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (refDCval >= 92) {
                                                                if (tcor < 81) {
                                                                    return 8;
                                                                }
                                                                else if (tcor >= 81) {
                                                                    if (Rs < 1309) {
                                                                        return 4;
                                                                    }
                                                                    else if (Rs >= 1309) {
                                                                        return 8;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (RsCsDiff >= 46) {
                                                            if (mu_mv_mag_sq < 24) {
                                                                if (Rs < 1467) {
                                                                    return 2;
                                                                }
                                                                else if (Rs >= 1467) {
                                                                    return 1;
                                                                }
                                                            }
                                                            else if (mu_mv_mag_sq >= 24) {
                                                                if (TSCindex < 9) {
                                                                    if (diffMVdiffVal < -3) {
                                                                        if (Rs < 1906) {
                                                                            return 8;
                                                                        }
                                                                        else if (Rs >= 1906) {
                                                                            return 2;
                                                                        }
                                                                    }
                                                                    else if (diffMVdiffVal >= -3) {
                                                                        if (Rs < 1750) {
                                                                            return 8;
                                                                        }
                                                                        else if (Rs >= 1750) {
                                                                            return 2;
                                                                        }
                                                                    }
                                                                }
                                                                else if (TSCindex >= 9) {
                                                                    if (RsDiff < 6) {
                                                                        return 8;
                                                                    }
                                                                    else if (RsDiff >= 6) {
                                                                        if (Rs < 1685) {
                                                                            return 8;
                                                                        }
                                                                        else if (Rs >= 1685) {
                                                                            return 2;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                            else if (mcTcor >= 92) {
                                                if (RsCsDiff < 9) {
                                                    if (posBalance < 55) {
                                                        if (tcor < 94) {
                                                            if (TSCindex < 6) {
                                                                return 4;
                                                            }
                                                            else if (TSCindex >= 6) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (tcor >= 94) {
                                                            if (Rs < 651) {
                                                                if (Rs < 595) {
                                                                    return 8;
                                                                }
                                                                else if (Rs >= 595) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (Rs >= 651) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                    else if (posBalance >= 55) {
                                                        return 1;
                                                    }
                                                }
                                                else if (RsCsDiff >= 9) {
                                                    if (SC < 2541) {
                                                        if (diffMVdiffVal < -3) {
                                                            if (Cs < 765) {
                                                                if (Cs < 430) {
                                                                    return 1;
                                                                }
                                                                else if (Cs >= 430) {
                                                                    return 2;
                                                                }
                                                            }
                                                            else if (Cs >= 765) {
                                                                return 1;
                                                            }
                                                        }
                                                        else if (diffMVdiffVal >= -3) {
                                                            if (ssDCval < 75) {
                                                                if (TSC < 5) {
                                                                    return 4;
                                                                }
                                                                else if (TSC >= 5) {
                                                                    if (tcor < 85) {
                                                                        if (Cs < 521) {
                                                                            return 2;
                                                                        }
                                                                        else if (Cs >= 521) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                    else if (tcor >= 85) {
                                                                        return 2;
                                                                    }
                                                                }
                                                            }
                                                            else if (ssDCval >= 75) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                    else if (SC >= 2541) {
                                                        if (MVDiff < 2) {
                                                            return 2;
                                                        }
                                                        else if (MVDiff >= 2) {
                                                            if (mu_mv_mag_sq < 40) {
                                                                if (Cs < 804) {
                                                                    return 4;
                                                                }
                                                                else if (Cs >= 804) {
                                                                    if (diffAFD < 2) {
                                                                        return 1;
                                                                    }
                                                                    else if (diffAFD >= 2) {
                                                                        return 2;
                                                                    }
                                                                }
                                                            }
                                                            else if (mu_mv_mag_sq >= 40) {
                                                                if (AFD < 11) {
                                                                    if (Cs < 1734) {
                                                                        if (negBalance < 50) {
                                                                            if (Rs < 1624) {
                                                                                return 4;
                                                                            }
                                                                            else if (Rs >= 1624) {
                                                                                return 1;
                                                                            }
                                                                        }
                                                                        else if (negBalance >= 50) {
                                                                            if (SC < 3852) {
                                                                                return 4;
                                                                            }
                                                                            else if (SC >= 3852) {
                                                                                return 1;
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (Cs >= 1734) {
                                                                        if (Rs < 2186) {
                                                                            return 2;
                                                                        }
                                                                        else if (Rs >= 2186) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                }
                                                                else if (AFD >= 11) {
                                                                    if (Cs < 1528) {
                                                                        return 4;
                                                                    }
                                                                    else if (Cs >= 1528) {
                                                                        return 2;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (refDCval >= 97) {
                                            if (tcor < 85) {
                                                if (AFD < 12) {
                                                    if (refDCval < 98) {
                                                        if (Rs < 1166) {
                                                            if (negBalance < 52) {
                                                                if (TSC < 5) {
                                                                    return 8;
                                                                }
                                                                else if (TSC >= 5) {
                                                                    return 2;
                                                                }
                                                            }
                                                            else if (negBalance >= 52) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (Rs >= 1166) {
                                                            if (negBalance < 53) {
                                                                if (tcor < 83) {
                                                                    return 2;
                                                                }
                                                                else if (tcor >= 83) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (negBalance >= 53) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                    else if (refDCval >= 98) {
                                                        if (Rs < 612) {
                                                            if (RsDiff < 2) {
                                                                return 4;
                                                            }
                                                            else if (RsDiff >= 2) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (Rs >= 612) {
                                                            if (Rs < 2552) {
                                                                if (diffMVdiffVal < 6) {
                                                                    if (ssDCval < 99) {
                                                                        if (Rs < 1238) {
                                                                            if (SC < 2225) {
                                                                                return 4;
                                                                            }
                                                                            else if (SC >= 2225) {
                                                                                return 2;
                                                                            }
                                                                        }
                                                                        else if (Rs >= 1238) {
                                                                            if (MVDiff < 11) {
                                                                                return 8;
                                                                            }
                                                                            else if (MVDiff >= 11) {
                                                                                return 2;
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (ssDCval >= 99) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (diffMVdiffVal >= 6) {
                                                                    return 2;
                                                                }
                                                            }
                                                            else if (Rs >= 2552) {
                                                                if (ssDCval < 119) {
                                                                    if (ssDCval < 117) {
                                                                        return 4;
                                                                    }
                                                                    else if (ssDCval >= 117) {
                                                                        if (posBalance < 60) {
                                                                            return 8;
                                                                        }
                                                                        else if (posBalance >= 60) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                }
                                                                else if (ssDCval >= 119) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (AFD >= 12) {
                                                    if (Cs < 621) {
                                                        return 1;
                                                    }
                                                    else if (Cs >= 621) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                            else if (tcor >= 85) {
                                                if (mu_mv_mag_sq < 41) {
                                                    if (RsDiff < 4) {
                                                        if (CsDiff < 1) {
                                                            return 1;
                                                        }
                                                        else if (CsDiff >= 1) {
                                                            if (SC < 1003) {
                                                                return 2;
                                                            }
                                                            else if (SC >= 1003) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                    else if (RsDiff >= 4) {
                                                        if (ssDCval < 126) {
                                                            return 4;
                                                        }
                                                        else if (ssDCval >= 126) {
                                                            return 2;
                                                        }
                                                    }
                                                }
                                                else if (mu_mv_mag_sq >= 41) {
                                                    if (Cs < 388) {
                                                        return 4;
                                                    }
                                                    else if (Cs >= 388) {
                                                        if (Cs < 1333) {
                                                            return 8;
                                                        }
                                                        else if (Cs >= 1333) {
                                                            if (AFD < 10) {
                                                                return 4;
                                                            }
                                                            else if (AFD >= 10) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                            else if (MVDiff >= 16) {
                                if (TSC < 6) {
                                    if (SC < 2829) {
                                        if (AFD < 10) {
                                            if (refDCval < 106) {
                                                if (mu_mv_mag_sq < 54) {
                                                    if (refDCval < 94) {
                                                        if (Cs < 1000) {
                                                            if (posBalance < 19) {
                                                                if (negBalance < 88) {
                                                                    if (tcor < 119) {
                                                                        if (mcTcor < 95) {
                                                                            if (tcor < 86) {
                                                                                if (mcTcor < 30) {
                                                                                    if (MVDiff < 146) {
                                                                                        return 4;
                                                                                    }
                                                                                    else if (MVDiff >= 146) {
                                                                                        return 2;
                                                                                    }
                                                                                }
                                                                                else if (mcTcor >= 30) {
                                                                                    if (gchDC < 1) {
                                                                                        return 4;
                                                                                    }
                                                                                    else if (gchDC >= 1) {
                                                                                        if (refDCval < 20) {
                                                                                            if (tcor < 56) {
                                                                                                return 8;
                                                                                            }
                                                                                            else if (tcor >= 56) {
                                                                                                return 4;
                                                                                            }
                                                                                        }
                                                                                        else if (refDCval >= 20) {
                                                                                            return 8;
                                                                                        }
                                                                                    }
                                                                                }
                                                                            }
                                                                            else if (tcor >= 86) {
                                                                                if (mu_mv_mag_sq < 18) {
                                                                                    if (refDCval < 20) {
                                                                                        return 8;
                                                                                    }
                                                                                    else if (refDCval >= 20) {
                                                                                        return 2;
                                                                                    }
                                                                                }
                                                                                else if (mu_mv_mag_sq >= 18) {
                                                                                    return 8;
                                                                                }
                                                                            }
                                                                        }
                                                                        else if (mcTcor >= 95) {
                                                                            if (diffMVdiffVal < 1) {
                                                                                return 2;
                                                                            }
                                                                            else if (diffMVdiffVal >= 1) {
                                                                                if (CsDiff < 1) {
                                                                                    return 4;
                                                                                }
                                                                                else if (CsDiff >= 1) {
                                                                                    return 2;
                                                                                }
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (tcor >= 119) {
                                                                        if (ssDCval < 23) {
                                                                            if (refDCval < 22) {
                                                                                return 1;
                                                                            }
                                                                            else if (refDCval >= 22) {
                                                                                return 2;
                                                                            }
                                                                        }
                                                                        else if (ssDCval >= 23) {
                                                                            if (diffAFD < -1) {
                                                                                if (MVDiff < 30) {
                                                                                    return 4;
                                                                                }
                                                                                else if (MVDiff >= 30) {
                                                                                    return 2;
                                                                                }
                                                                            }
                                                                            else if (diffAFD >= -1) {
                                                                                if (Cs < 47) {
                                                                                    return 4;
                                                                                }
                                                                                else if (Cs >= 47) {
                                                                                    return 8;
                                                                                }
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                                else if (negBalance >= 88) {
                                                                    if (AFD < 5) {
                                                                        if (tcor < 82) {
                                                                            return 1;
                                                                        }
                                                                        else if (tcor >= 82) {
                                                                            return 2;
                                                                        }
                                                                    }
                                                                    else if (AFD >= 5) {
                                                                        if (diffTSC < -2) {
                                                                            return 2;
                                                                        }
                                                                        else if (diffTSC >= -2) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                            else if (posBalance >= 19) {
                                                                if (AFD < 6) {
                                                                    if (tcor < 75) {
                                                                        if (negBalance < 7) {
                                                                            if (MVDiff < 505) {
                                                                                if (ssDCval < 19) {
                                                                                    if (negBalance < 2) {
                                                                                        if (mu_mv_mag_sq < 32) {
                                                                                            return 1;
                                                                                        }
                                                                                        else if (mu_mv_mag_sq >= 32) {
                                                                                            if (tcor < 37) {
                                                                                                return 8;
                                                                                            }
                                                                                            else if (tcor >= 37) {
                                                                                                if (mcTcor < 46) {
                                                                                                    if (MVDiff < 59) {
                                                                                                        return 1;
                                                                                                    }
                                                                                                    else if (MVDiff >= 59) {
                                                                                                        return 4;
                                                                                                    }
                                                                                                }
                                                                                                else if (mcTcor >= 46) {
                                                                                                    return 1;
                                                                                                }
                                                                                            }
                                                                                        }
                                                                                    }
                                                                                    else if (negBalance >= 2) {
                                                                                        if (tcor < 41) {
                                                                                            if (tcor < 36) {
                                                                                                return 8;
                                                                                            }
                                                                                            else if (tcor >= 36) {
                                                                                                return 4;
                                                                                            }
                                                                                        }
                                                                                        else if (tcor >= 41) {
                                                                                            if (MVDiff < 63) {
                                                                                                return 1;
                                                                                            }
                                                                                            else if (MVDiff >= 63) {
                                                                                                if (mcTcor < 52) {
                                                                                                    return 8;
                                                                                                }
                                                                                                else if (mcTcor >= 52) {
                                                                                                    return 4;
                                                                                                }
                                                                                            }
                                                                                        }
                                                                                    }
                                                                                }
                                                                                else if (ssDCval >= 19) {
                                                                                    if (refDCval < 22) {
                                                                                        if (MVDiff < 208) {
                                                                                            if (diffMVdiffVal < -272) {
                                                                                                if (MVDiff < 78) {
                                                                                                    return 8;
                                                                                                }
                                                                                                else if (MVDiff >= 78) {
                                                                                                    return 2;
                                                                                                }
                                                                                            }
                                                                                            else if (diffMVdiffVal >= -272) {
                                                                                                if (mcTcor < 55) {
                                                                                                    if (tcor < 26) {
                                                                                                        return 4;
                                                                                                    }
                                                                                                    else if (tcor >= 26) {
                                                                                                        return 8;
                                                                                                    }
                                                                                                }
                                                                                                else if (mcTcor >= 55) {
                                                                                                    if (tcor < 60) {
                                                                                                        return 2;
                                                                                                    }
                                                                                                    else if (tcor >= 60) {
                                                                                                        if (posBalance < 99) {
                                                                                                            return 4;
                                                                                                        }
                                                                                                        else if (posBalance >= 99) {
                                                                                                            return 8;
                                                                                                        }
                                                                                                    }
                                                                                                }
                                                                                            }
                                                                                        }
                                                                                        else if (MVDiff >= 208) {
                                                                                            if (mcTcor < 26) {
                                                                                                return 1;
                                                                                            }
                                                                                            else if (mcTcor >= 26) {
                                                                                                return 4;
                                                                                            }
                                                                                        }
                                                                                    }
                                                                                    else if (refDCval >= 22) {
                                                                                        if (tcor < 59) {
                                                                                            return 1;
                                                                                        }
                                                                                        else if (tcor >= 59) {
                                                                                            if (ssDCval < 29) {
                                                                                                return 4;
                                                                                            }
                                                                                            else if (ssDCval >= 29) {
                                                                                                if (MVDiff < 89) {
                                                                                                    return 8;
                                                                                                }
                                                                                                else if (MVDiff >= 89) {
                                                                                                    if (Cs < 393) {
                                                                                                        return 4;
                                                                                                    }
                                                                                                    else if (Cs >= 393) {
                                                                                                        return 1;
                                                                                                    }
                                                                                                }
                                                                                            }
                                                                                        }
                                                                                    }
                                                                                }
                                                                            }
                                                                            else if (MVDiff >= 505) {
                                                                                if (ssDCval < 22) {
                                                                                    return 2;
                                                                                }
                                                                                else if (ssDCval >= 22) {
                                                                                    if (AFD < 4) {
                                                                                        return 1;
                                                                                    }
                                                                                    else if (AFD >= 4) {
                                                                                        return 2;
                                                                                    }
                                                                                }
                                                                            }
                                                                        }
                                                                        else if (negBalance >= 7) {
                                                                            if (Rs < 458) {
                                                                                if (ssDCval < 27) {
                                                                                    return 2;
                                                                                }
                                                                                else if (ssDCval >= 27) {
                                                                                    if (diffMVdiffVal < -1) {
                                                                                        return 1;
                                                                                    }
                                                                                    else if (diffMVdiffVal >= -1) {
                                                                                        if (SC < 711) {
                                                                                            return 8;
                                                                                        }
                                                                                        else if (SC >= 711) {
                                                                                            return 1;
                                                                                        }
                                                                                    }
                                                                                }
                                                                            }
                                                                            else if (Rs >= 458) {
                                                                                if (TSC < 5) {
                                                                                    return 8;
                                                                                }
                                                                                else if (TSC >= 5) {
                                                                                    return 2;
                                                                                }
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (tcor >= 75) {
                                                                        if (Cs < 310) {
                                                                            if (Cs < 284) {
                                                                                if (Cs < 37) {
                                                                                    if (ssDCval < 30) {
                                                                                        if (MVDiff < 98) {
                                                                                            return 4;
                                                                                        }
                                                                                        else if (MVDiff >= 98) {
                                                                                            if (MVDiff < 653) {
                                                                                                if (AFD < 2) {
                                                                                                    return 8;
                                                                                                }
                                                                                                else if (AFD >= 2) {
                                                                                                    return 1;
                                                                                                }
                                                                                            }
                                                                                            else if (MVDiff >= 653) {
                                                                                                return 2;
                                                                                            }
                                                                                        }
                                                                                    }
                                                                                    else if (ssDCval >= 30) {
                                                                                        if (Rs < 73) {
                                                                                            return 2;
                                                                                        }
                                                                                        else if (Rs >= 73) {
                                                                                            return 8;
                                                                                        }
                                                                                    }
                                                                                }
                                                                                else if (Cs >= 37) {
                                                                                    if (ssDCval < 34) {
                                                                                        if (MVDiff < 41) {
                                                                                            return 1;
                                                                                        }
                                                                                        else if (MVDiff >= 41) {
                                                                                            return 2;
                                                                                        }
                                                                                    }
                                                                                    else if (ssDCval >= 34) {
                                                                                        return 2;
                                                                                    }
                                                                                }
                                                                            }
                                                                            else if (Cs >= 284) {
                                                                                return 8;
                                                                            }
                                                                        }
                                                                        else if (Cs >= 310) {
                                                                            if (Cs < 387) {
                                                                                if (Rs < 442) {
                                                                                    if (Cs < 376) {
                                                                                        return 1;
                                                                                    }
                                                                                    else if (Cs >= 376) {
                                                                                        return 4;
                                                                                    }
                                                                                }
                                                                                else if (Rs >= 442) {
                                                                                    if (MVDiff < 38) {
                                                                                        if (diffMVdiffVal < -2) {
                                                                                            return 1;
                                                                                        }
                                                                                        else if (diffMVdiffVal >= -2) {
                                                                                            return 4;
                                                                                        }
                                                                                    }
                                                                                    else if (MVDiff >= 38) {
                                                                                        return 8;
                                                                                    }
                                                                                }
                                                                            }
                                                                            else if (Cs >= 387) {
                                                                                if (Rs < 1115) {
                                                                                    if (mcTcor < 93) {
                                                                                        if (MVDiff < 25) {
                                                                                            if (diffMVdiffVal < -10) {
                                                                                                return 4;
                                                                                            }
                                                                                            else if (diffMVdiffVal >= -10) {
                                                                                                if (negBalance < 45) {
                                                                                                    return 2;
                                                                                                }
                                                                                                else if (negBalance >= 45) {
                                                                                                    if (Rs < 508) {
                                                                                                        return 2;
                                                                                                    }
                                                                                                    else if (Rs >= 508) {
                                                                                                        return 1;
                                                                                                    }
                                                                                                }
                                                                                            }
                                                                                        }
                                                                                        else if (MVDiff >= 25) {
                                                                                            return 2;
                                                                                        }
                                                                                    }
                                                                                    else if (mcTcor >= 93) {
                                                                                        if (Rs < 880) {
                                                                                            return 4;
                                                                                        }
                                                                                        else if (Rs >= 880) {
                                                                                            return 1;
                                                                                        }
                                                                                    }
                                                                                }
                                                                                else if (Rs >= 1115) {
                                                                                    if (refDCval < 42) {
                                                                                        return 2;
                                                                                    }
                                                                                    else if (refDCval >= 42) {
                                                                                        return 1;
                                                                                    }
                                                                                }
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                                else if (AFD >= 6) {
                                                                    if (SC < 673) {
                                                                        if (Rs < 311) {
                                                                            if (ssDCval < 48) {
                                                                                return 1;
                                                                            }
                                                                            else if (ssDCval >= 48) {
                                                                                if (gchDC < 3) {
                                                                                    if (refDCval < 91) {
                                                                                        return 2;
                                                                                    }
                                                                                    else if (refDCval >= 91) {
                                                                                        return 1;
                                                                                    }
                                                                                }
                                                                                else if (gchDC >= 3) {
                                                                                    if (Rs < 131) {
                                                                                        return 1;
                                                                                    }
                                                                                    else if (Rs >= 131) {
                                                                                        return 4;
                                                                                    }
                                                                                }
                                                                            }
                                                                        }
                                                                        else if (Rs >= 311) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                    else if (SC >= 673) {
                                                                        if (mcTcor < 95) {
                                                                            if (Scindex < 4) {
                                                                                if (diffRsCsdiff < 2) {
                                                                                    if (tcor < 76) {
                                                                                        if (negBalance < 47) {
                                                                                            return 2;
                                                                                        }
                                                                                        else if (negBalance >= 47) {
                                                                                            if (mcTcor < 77) {
                                                                                                if (RsCsDiff < 7) {
                                                                                                    return 4;
                                                                                                }
                                                                                                else if (RsCsDiff >= 7) {
                                                                                                    return 1;
                                                                                                }
                                                                                            }
                                                                                            else if (mcTcor >= 77) {
                                                                                                return 4;
                                                                                            }
                                                                                        }
                                                                                    }
                                                                                    else if (tcor >= 76) {
                                                                                        if (negBalance < 47) {
                                                                                            if (RsDiff < 3) {
                                                                                                return 4;
                                                                                            }
                                                                                            else if (RsDiff >= 3) {
                                                                                                return 2;
                                                                                            }
                                                                                        }
                                                                                        else if (negBalance >= 47) {
                                                                                            if (SC < 1314) {
                                                                                                if (Rs < 710) {
                                                                                                    if (ssDCval < 38) {
                                                                                                        return 4;
                                                                                                    }
                                                                                                    else if (ssDCval >= 38) {
                                                                                                        return 1;
                                                                                                    }
                                                                                                }
                                                                                                else if (Rs >= 710) {
                                                                                                    return 4;
                                                                                                }
                                                                                            }
                                                                                            else if (SC >= 1314) {
                                                                                                return 1;
                                                                                            }
                                                                                        }
                                                                                    }
                                                                                }
                                                                                else if (diffRsCsdiff >= 2) {
                                                                                    if (Rs < 655) {
                                                                                        return 2;
                                                                                    }
                                                                                    else if (Rs >= 655) {
                                                                                        if (mu_mv_mag_sq < 43) {
                                                                                            if (mu_mv_mag_sq < 34) {
                                                                                                if (Cs < 392) {
                                                                                                    return 1;
                                                                                                }
                                                                                                else if (Cs >= 392) {
                                                                                                    if (Cs < 643) {
                                                                                                        return 4;
                                                                                                    }
                                                                                                    else if (Cs >= 643) {
                                                                                                        return 1;
                                                                                                    }
                                                                                                }
                                                                                            }
                                                                                            else if (mu_mv_mag_sq >= 34) {
                                                                                                return 1;
                                                                                            }
                                                                                        }
                                                                                        else if (mu_mv_mag_sq >= 43) {
                                                                                            if (Rs < 837) {
                                                                                                return 4;
                                                                                            }
                                                                                            else if (Rs >= 837) {
                                                                                                return 8;
                                                                                            }
                                                                                        }
                                                                                    }
                                                                                }
                                                                            }
                                                                            else if (Scindex >= 4) {
                                                                                if (posBalance < 55) {
                                                                                    if (RsCsDiff < 16) {
                                                                                        if (SC < 1827) {
                                                                                            return 8;
                                                                                        }
                                                                                        else if (SC >= 1827) {
                                                                                            return 4;
                                                                                        }
                                                                                    }
                                                                                    else if (RsCsDiff >= 16) {
                                                                                        if (Rs < 912) {
                                                                                            return 2;
                                                                                        }
                                                                                        else if (Rs >= 912) {
                                                                                            if (MVDiff < 22) {
                                                                                                return 4;
                                                                                            }
                                                                                            else if (MVDiff >= 22) {
                                                                                                return 1;
                                                                                            }
                                                                                        }
                                                                                    }
                                                                                }
                                                                                else if (posBalance >= 55) {
                                                                                    if (Cs < 690) {
                                                                                        return 8;
                                                                                    }
                                                                                    else if (Cs >= 690) {
                                                                                        return 1;
                                                                                    }
                                                                                }
                                                                            }
                                                                        }
                                                                        else if (mcTcor >= 95) {
                                                                            if (Cs < 536) {
                                                                                return 2;
                                                                            }
                                                                            else if (Cs >= 536) {
                                                                                return 8;
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (Cs >= 1000) {
                                                            if (refDCval < 54) {
                                                                if (RsDiff < 6) {
                                                                    return 2;
                                                                }
                                                                else if (RsDiff >= 6) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (refDCval >= 54) {
                                                                return 1;
                                                            }
                                                        }
                                                    }
                                                    else if (refDCval >= 94) {
                                                        if (RsDiff < 3) {
                                                            if (MVDiff < 18) {
                                                                if (RsCsDiff < 7) {
                                                                    return 4;
                                                                }
                                                                else if (RsCsDiff >= 7) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (MVDiff >= 18) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (RsDiff >= 3) {
                                                            if (Rs < 993) {
                                                                return 4;
                                                            }
                                                            else if (Rs >= 993) {
                                                                return 2;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (mu_mv_mag_sq >= 54) {
                                                    if (tcor < 89) {
                                                        if (tcor < 85) {
                                                            if (Rs < 704) {
                                                                if (negBalance < 42) {
                                                                    if (mu_mv_mag_sq < 68) {
                                                                        if (SC < 918) {
                                                                            if (Rs < 23) {
                                                                                return 8;
                                                                            }
                                                                            else if (Rs >= 23) {
                                                                                if (ssDCval < 38) {
                                                                                    if (Rs < 251) {
                                                                                        return 4;
                                                                                    }
                                                                                    else if (Rs >= 251) {
                                                                                        return 2;
                                                                                    }
                                                                                }
                                                                                else if (ssDCval >= 38) {
                                                                                    return 4;
                                                                                }
                                                                            }
                                                                        }
                                                                        else if (SC >= 918) {
                                                                            return 2;
                                                                        }
                                                                    }
                                                                    else if (mu_mv_mag_sq >= 68) {
                                                                        if (posBalance < 55) {
                                                                            return 2;
                                                                        }
                                                                        else if (posBalance >= 55) {
                                                                            if (Rs < 459) {
                                                                                return 8;
                                                                            }
                                                                            else if (Rs >= 459) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                                else if (negBalance >= 42) {
                                                                    if (Cs < 381) {
                                                                        if (Rs < 428) {
                                                                            if (diffTSC < 1) {
                                                                                return 4;
                                                                            }
                                                                            else if (diffTSC >= 1) {
                                                                                if (Rs < 389) {
                                                                                    return 4;
                                                                                }
                                                                                else if (Rs >= 389) {
                                                                                    return 1;
                                                                                }
                                                                            }
                                                                        }
                                                                        else if (Rs >= 428) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                    else if (Cs >= 381) {
                                                                        if (MVDiff < 24) {
                                                                            if (posBalance < 54) {
                                                                                return 4;
                                                                            }
                                                                            else if (posBalance >= 54) {
                                                                                if (diffAFD < 1) {
                                                                                    return 2;
                                                                                }
                                                                                else if (diffAFD >= 1) {
                                                                                    return 4;
                                                                                }
                                                                            }
                                                                        }
                                                                        else if (MVDiff >= 24) {
                                                                            return 2;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                            else if (Rs >= 704) {
                                                                if (Cs < 820) {
                                                                    if (Rs < 869) {
                                                                        if (mcTcor < 84) {
                                                                            return 4;
                                                                        }
                                                                        else if (mcTcor >= 84) {
                                                                            if (Rs < 815) {
                                                                                return 1;
                                                                            }
                                                                            else if (Rs >= 815) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (Rs >= 869) {
                                                                        return 1;
                                                                    }
                                                                }
                                                                else if (Cs >= 820) {
                                                                    return 1;
                                                                }
                                                            }
                                                        }
                                                        else if (tcor >= 85) {
                                                            if (diffAFD < -1) {
                                                                if (Cs < 400) {
                                                                    return 4;
                                                                }
                                                                else if (Cs >= 400) {
                                                                    if (posBalance < 24) {
                                                                        return 8;
                                                                    }
                                                                    else if (posBalance >= 24) {
                                                                        return 2;
                                                                    }
                                                                }
                                                            }
                                                            else if (diffAFD >= -1) {
                                                                if (ssDCval < 38) {
                                                                    if (Cs < 334) {
                                                                        return 4;
                                                                    }
                                                                    else if (Cs >= 334) {
                                                                        if (MVDiff < 22) {
                                                                            return 2;
                                                                        }
                                                                        else if (MVDiff >= 22) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                }
                                                                else if (ssDCval >= 38) {
                                                                    if (negBalance < 43) {
                                                                        return 4;
                                                                    }
                                                                    else if (negBalance >= 43) {
                                                                        if (MVDiff < 27) {
                                                                            return 4;
                                                                        }
                                                                        else if (MVDiff >= 27) {
                                                                            if (Cs < 359) {
                                                                                return 4;
                                                                            }
                                                                            else if (Cs >= 359) {
                                                                                return 2;
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (tcor >= 89) {
                                                        if (refDCval < 64) {
                                                            if (Rs < 1235) {
                                                                if (mu_mv_mag_sq < 74) {
                                                                    return 8;
                                                                }
                                                                else if (mu_mv_mag_sq >= 74) {
                                                                    if (ssDCval < 23) {
                                                                        return 8;
                                                                    }
                                                                    else if (ssDCval >= 23) {
                                                                        if (ssDCval < 26) {
                                                                            return 4;
                                                                        }
                                                                        else if (ssDCval >= 26) {
                                                                            return 2;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                            else if (Rs >= 1235) {
                                                                if (SC < 2500) {
                                                                    return 1;
                                                                }
                                                                else if (SC >= 2500) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                        else if (refDCval >= 64) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (refDCval >= 106) {
                                                if (SC < 817) {
                                                    return 1;
                                                }
                                                else if (SC >= 817) {
                                                    if (Rs < 827) {
                                                        return 4;
                                                    }
                                                    else if (Rs >= 827) {
                                                        return 1;
                                                    }
                                                }
                                            }
                                        }
                                        else if (AFD >= 10) {
                                            if (Rs < 996) {
                                                if (ssDCval < 107) {
                                                    if (posBalance < 23) {
                                                        return 1;
                                                    }
                                                    else if (posBalance >= 23) {
                                                        return 2;
                                                    }
                                                }
                                                else if (ssDCval >= 107) {
                                                    if (Rs < 245) {
                                                        return 1;
                                                    }
                                                    else if (Rs >= 245) {
                                                        if (Rs < 352) {
                                                            return 2;
                                                        }
                                                        else if (Rs >= 352) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (Rs >= 996) {
                                                if (Rs < 1743) {
                                                    return 1;
                                                }
                                                else if (Rs >= 1743) {
                                                    return 2;
                                                }
                                            }
                                        }
                                    }
                                    else if (SC >= 2829) {
                                        if (ssDCval < 39) {
                                            if (Cs < 1494) {
                                                return 2;
                                            }
                                            else if (Cs >= 1494) {
                                                return 1;
                                            }
                                        }
                                        else if (ssDCval >= 39) {
                                            if (Cs < 1549) {
                                                if (AFD < 8) {
                                                    if (Rs < 1822) {
                                                        if (Rs < 1821) {
                                                            return 4;
                                                        }
                                                        else if (Rs >= 1821) {
                                                            return 1;
                                                        }
                                                    }
                                                    else if (Rs >= 1822) {
                                                        return 4;
                                                    }
                                                }
                                                else if (AFD >= 8) {
                                                    return 1;
                                                }
                                            }
                                            else if (Cs >= 1549) {
                                                if (ssDCval < 40) {
                                                    if (mu_mv_mag_sq < 63) {
                                                        return 1;
                                                    }
                                                    else if (mu_mv_mag_sq >= 63) {
                                                        if (Rs < 1991) {
                                                            return 4;
                                                        }
                                                        else if (Rs >= 1991) {
                                                            return 1;
                                                        }
                                                    }
                                                }
                                                else if (ssDCval >= 40) {
                                                    return 4;
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (TSC >= 6) {
                                    if (Cs < 272) {
                                        if (Rs < 593) {
                                            if (mcTcor < 86) {
                                                if (MVDiff < 28) {
                                                    return 4;
                                                }
                                                else if (MVDiff >= 28) {
                                                    if (Cs < 10) {
                                                        return 2;
                                                    }
                                                    else if (Cs >= 10) {
                                                        return 1;
                                                    }
                                                }
                                            }
                                            else if (mcTcor >= 86) {
                                                return 1;
                                            }
                                        }
                                        else if (Rs >= 593) {
                                            if (Rs < 621) {
                                                if (mcTcor < 93) {
                                                    return 1;
                                                }
                                                else if (mcTcor >= 93) {
                                                    if (AFD < 12) {
                                                        return 2;
                                                    }
                                                    else if (AFD >= 12) {
                                                        return 1;
                                                    }
                                                }
                                            }
                                            else if (Rs >= 621) {
                                                if (Cs < 123) {
                                                    if (SC < 759) {
                                                        return 2;
                                                    }
                                                    else if (SC >= 759) {
                                                        return 1;
                                                    }
                                                }
                                                else if (Cs >= 123) {
                                                    return 1;
                                                }
                                            }
                                        }
                                    }
                                    else if (Cs >= 272) {
                                        if (CsDiff < 3) {
                                            if (refDCval < 104) {
                                                if (MVDiff < 70) {
                                                    if (tcor < 90) {
                                                        if (SC < 1369) {
                                                            if (MVDiff < 41) {
                                                                if (posBalance < 62) {
                                                                    if (Rs < 760) {
                                                                        return 1;
                                                                    }
                                                                    else if (Rs >= 760) {
                                                                        if (Cs < 564) {
                                                                            if (posBalance < 59) {
                                                                                if (Cs < 552) {
                                                                                    return 4;
                                                                                }
                                                                                else if (Cs >= 552) {
                                                                                    if (MVDiff < 30) {
                                                                                        return 1;
                                                                                    }
                                                                                    else if (MVDiff >= 30) {
                                                                                        return 4;
                                                                                    }
                                                                                }
                                                                            }
                                                                            else if (posBalance >= 59) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                        else if (Cs >= 564) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                }
                                                                else if (posBalance >= 62) {
                                                                    return 1;
                                                                }
                                                            }
                                                            else if (MVDiff >= 41) {
                                                                if (Cs < 514) {
                                                                    return 4;
                                                                }
                                                                else if (Cs >= 514) {
                                                                    if (mu_mv_mag_sq < 29) {
                                                                        return 4;
                                                                    }
                                                                    else if (mu_mv_mag_sq >= 29) {
                                                                        return 1;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (SC >= 1369) {
                                                            if (mcTcor < 76) {
                                                                return 1;
                                                            }
                                                            else if (mcTcor >= 76) {
                                                                if (diffAFD < 1) {
                                                                    return 4;
                                                                }
                                                                else if (diffAFD >= 1) {
                                                                    if (mu_mv_mag_sq < 41) {
                                                                        return 1;
                                                                    }
                                                                    else if (mu_mv_mag_sq >= 41) {
                                                                        return 4;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (tcor >= 90) {
                                                        if (Rs < 995) {
                                                            return 8;
                                                        }
                                                        else if (Rs >= 995) {
                                                            return 1;
                                                        }
                                                    }
                                                }
                                                else if (MVDiff >= 70) {
                                                    return 8;
                                                }
                                            }
                                            else if (refDCval >= 104) {
                                                if (diffAFD < 1) {
                                                    if (Cs < 382) {
                                                        return 4;
                                                    }
                                                    else if (Cs >= 382) {
                                                        return 8;
                                                    }
                                                }
                                                else if (diffAFD >= 1) {
                                                    return 4;
                                                }
                                            }
                                        }
                                        else if (CsDiff >= 3) {
                                            if (negBalance < 49) {
                                                if (negBalance < 44) {
                                                    if (negBalance < 34) {
                                                        if (ssDCval < 61) {
                                                            if (SC < 1234) {
                                                                return 8;
                                                            }
                                                            else if (SC >= 1234) {
                                                                return 1;
                                                            }
                                                        }
                                                        else if (ssDCval >= 61) {
                                                            if (diffMVdiffVal < -1) {
                                                                return 8;
                                                            }
                                                            else if (diffMVdiffVal >= -1) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                    else if (negBalance >= 34) {
                                                        if (AFD < 8) {
                                                            return 4;
                                                        }
                                                        else if (AFD >= 8) {
                                                            if (posBalance < 60) {
                                                                return 8;
                                                            }
                                                            else if (posBalance >= 60) {
                                                                if (Cs < 1119) {
                                                                    return 1;
                                                                }
                                                                else if (Cs >= 1119) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (negBalance >= 44) {
                                                    if (AFD < 10) {
                                                        if (Rs < 1563) {
                                                            return 8;
                                                        }
                                                        else if (Rs >= 1563) {
                                                            return 1;
                                                        }
                                                    }
                                                    else if (AFD >= 10) {
                                                        if (CsDiff < 6) {
                                                            return 1;
                                                        }
                                                        else if (CsDiff >= 6) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (negBalance >= 49) {
                                                if (Cs < 732) {
                                                    return 2;
                                                }
                                                else if (Cs >= 732) {
                                                    if (ssDCval < 139) {
                                                        if (refDCval < 73) {
                                                            if (TSCindex < 8) {
                                                                return 8;
                                                            }
                                                            else if (TSCindex >= 8) {
                                                                return 1;
                                                            }
                                                        }
                                                        else if (refDCval >= 73) {
                                                            if (Rs < 768) {
                                                                return 8;
                                                            }
                                                            else if (Rs >= 768) {
                                                                if (TSC < 10) {
                                                                    if (mcTcor < 82) {
                                                                        return 8;
                                                                    }
                                                                    else if (mcTcor >= 82) {
                                                                        if (diffMVdiffVal < -5) {
                                                                            if (SC < 2907) {
                                                                                return 2;
                                                                            }
                                                                            else if (SC >= 2907) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                        else if (diffMVdiffVal >= -5) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                }
                                                                else if (TSC >= 10) {
                                                                    if (Cs < 1401) {
                                                                        return 2;
                                                                    }
                                                                    else if (Cs >= 1401) {
                                                                        return 8;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (ssDCval >= 139) {
                                                        return 2;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                        }
                        else if (AFD >= 13) {
                            if (RsDiff < 8) {
                                if (SC < 2634) {
                                    if (Cs < 525) {
                                        if (gchDC < 18) {
                                            if (TSCindex < 9) {
                                                if (diffMVdiffVal < 15) {
                                                    return 2;
                                                }
                                                else if (diffMVdiffVal >= 15) {
                                                    if (MVDiff < 52) {
                                                        return 4;
                                                    }
                                                    else if (MVDiff >= 52) {
                                                        return 2;
                                                    }
                                                }
                                            }
                                            else if (TSCindex >= 9) {
                                                if (diffAFD < -2) {
                                                    return 1;
                                                }
                                                else if (diffAFD >= -2) {
                                                    if (posBalance < 55) {
                                                        return 2;
                                                    }
                                                    else if (posBalance >= 55) {
                                                        if (gchDC < 4) {
                                                            if (ssDCval < 137) {
                                                                return 1;
                                                            }
                                                            else if (ssDCval >= 137) {
                                                                return 2;
                                                            }
                                                        }
                                                        else if (gchDC >= 4) {
                                                            return 2;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (gchDC >= 18) {
                                            if (Cs < 1) {
                                                return 8;
                                            }
                                            else if (Cs >= 1) {
                                                return 1;
                                            }
                                        }
                                    }
                                    else if (Cs >= 525) {
                                        if (ssDCval < 92) {
                                            if (SC < 2233) {
                                                if (MVDiff < 32) {
                                                    if (ssDCval < 66) {
                                                        if (AFD < 14) {
                                                            if (MVDiff < 12) {
                                                                return 8;
                                                            }
                                                            else if (MVDiff >= 12) {
                                                                return 1;
                                                            }
                                                        }
                                                        else if (AFD >= 14) {
                                                            if (Rs < 824) {
                                                                return 2;
                                                            }
                                                            else if (Rs >= 824) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                    else if (ssDCval >= 66) {
                                                        if (TSC < 11) {
                                                            return 2;
                                                        }
                                                        else if (TSC >= 11) {
                                                            return 1;
                                                        }
                                                    }
                                                }
                                                else if (MVDiff >= 32) {
                                                    if (diffMVdiffVal < 43) {
                                                        if (gchDC < 24) {
                                                            return 1;
                                                        }
                                                        else if (gchDC >= 24) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (diffMVdiffVal >= 43) {
                                                        if (Cs < 698) {
                                                            return 8;
                                                        }
                                                        else if (Cs >= 698) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (SC >= 2233) {
                                                if (mcTcor < 86) {
                                                    if (AFD < 19) {
                                                        return 8;
                                                    }
                                                    else if (AFD >= 19) {
                                                        return 1;
                                                    }
                                                }
                                                else if (mcTcor >= 86) {
                                                    return 1;
                                                }
                                            }
                                        }
                                        else if (ssDCval >= 92) {
                                            if (refDCval < 134) {
                                                if (tcor < 79) {
                                                    if (SC < 1351) {
                                                        if (Rs < 509) {
                                                            return 4;
                                                        }
                                                        else if (Rs >= 509) {
                                                            return 1;
                                                        }
                                                    }
                                                    else if (SC >= 1351) {
                                                        return 4;
                                                    }
                                                }
                                                else if (tcor >= 79) {
                                                    if (SC < 1602) {
                                                        return 8;
                                                    }
                                                    else if (SC >= 1602) {
                                                        if (RsCsDiff < 51) {
                                                            return 4;
                                                        }
                                                        else if (RsCsDiff >= 51) {
                                                            return 2;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (refDCval >= 134) {
                                                if (posBalance < 52) {
                                                    if (gchDC < 17) {
                                                        if (RsDiff < 6) {
                                                            return 8;
                                                        }
                                                        else if (RsDiff >= 6) {
                                                            return 1;
                                                        }
                                                    }
                                                    else if (gchDC >= 17) {
                                                        return 1;
                                                    }
                                                }
                                                else if (posBalance >= 52) {
                                                    return 2;
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (SC >= 2634) {
                                    if (mcTcor < 85) {
                                        if (TSC < 10) {
                                            return 1;
                                        }
                                        else if (TSC >= 10) {
                                            return 8;
                                        }
                                    }
                                    else if (mcTcor >= 85) {
                                        if (ssDCval < 132) {
                                            if (RsDiff < 6) {
                                                return 8;
                                            }
                                            else if (RsDiff >= 6) {
                                                if (diffMVdiffVal < 4) {
                                                    if (mcTcor < 96) {
                                                        if (tcor < 78) {
                                                            if (negBalance < 50) {
                                                                return 1;
                                                            }
                                                            else if (negBalance >= 50) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (tcor >= 78) {
                                                            if (ssDCval < 79) {
                                                                if (SC < 3665) {
                                                                    return 4;
                                                                }
                                                                else if (SC >= 3665) {
                                                                    return 1;
                                                                }
                                                            }
                                                            else if (ssDCval >= 79) {
                                                                if (RsCsDiff < 67) {
                                                                    return 1;
                                                                }
                                                                else if (RsCsDiff >= 67) {
                                                                    return 2;
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (mcTcor >= 96) {
                                                        if (Cs < 1574) {
                                                            return 8;
                                                        }
                                                        else if (Cs >= 1574) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                                else if (diffMVdiffVal >= 4) {
                                                    if (MVDiff < 17) {
                                                        return 8;
                                                    }
                                                    else if (MVDiff >= 17) {
                                                        if (negBalance < 41) {
                                                            return 8;
                                                        }
                                                        else if (negBalance >= 41) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (ssDCval >= 132) {
                                            if (ssDCval < 156) {
                                                return 4;
                                            }
                                            else if (ssDCval >= 156) {
                                                return 2;
                                            }
                                        }
                                    }
                                }
                            }
                            else if (RsDiff >= 8) {
                                if (mcTcor < 85) {
                                    if (Cs < 1406) {
                                        if (CsDiff < 10) {
                                            if (Cs < 540) {
                                                if (TSC < 12) {
                                                    return 4;
                                                }
                                                else if (TSC >= 12) {
                                                    if (AFD < 42) {
                                                        return 2;
                                                    }
                                                    else if (AFD >= 42) {
                                                        if (Rs < 1496) {
                                                            return 1;
                                                        }
                                                        else if (Rs >= 1496) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (Cs >= 540) {
                                                if (TSC < 8) {
                                                    if (Cs < 1092) {
                                                        return 2;
                                                    }
                                                    else if (Cs >= 1092) {
                                                        return 8;
                                                    }
                                                }
                                                else if (TSC >= 8) {
                                                    if (Rs < 1377) {
                                                        if (Rs < 1214) {
                                                            return 1;
                                                        }
                                                        else if (Rs >= 1214) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (Rs >= 1377) {
                                                        if (tcor < 46) {
                                                            return 2;
                                                        }
                                                        else if (tcor >= 46) {
                                                            if (tcor < 72) {
                                                                if (RsDiff < 13) {
                                                                    if (SC < 2958) {
                                                                        if (Rs < 2058) {
                                                                            return 2;
                                                                        }
                                                                        else if (Rs >= 2058) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                    else if (SC >= 2958) {
                                                                        return 1;
                                                                    }
                                                                }
                                                                else if (RsDiff >= 13) {
                                                                    if (RsCsDiff < 293) {
                                                                        return 2;
                                                                    }
                                                                    else if (RsCsDiff >= 293) {
                                                                        if (TSC < 15) {
                                                                            return 1;
                                                                        }
                                                                        else if (TSC >= 15) {
                                                                            if (Cs < 937) {
                                                                                return 8;
                                                                            }
                                                                            else if (Cs >= 937) {
                                                                                return 2;
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                            else if (tcor >= 72) {
                                                                return 2;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (CsDiff >= 10) {
                                            if (diffRsCsdiff < 526) {
                                                if (diffAFD < 1) {
                                                    return 8;
                                                }
                                                else if (diffAFD >= 1) {
                                                    if (Rs < 893) {
                                                        return 8;
                                                    }
                                                    else if (Rs >= 893) {
                                                        return 1;
                                                    }
                                                }
                                            }
                                            else if (diffRsCsdiff >= 526) {
                                                if (AFD < 54) {
                                                    return 4;
                                                }
                                                else if (AFD >= 54) {
                                                    return 1;
                                                }
                                            }
                                        }
                                    }
                                    else if (Cs >= 1406) {
                                        if (tcor < 77) {
                                            if (negBalance < 29) {
                                                return 4;
                                            }
                                            else if (negBalance >= 29) {
                                                if (AFD < 31) {
                                                    return 1;
                                                }
                                                else if (AFD >= 31) {
                                                    return 8;
                                                }
                                            }
                                        }
                                        else if (tcor >= 77) {
                                            if (Rs < 2256) {
                                                return 4;
                                            }
                                            else if (Rs >= 2256) {
                                                return 2;
                                            }
                                        }
                                    }
                                }
                                else if (mcTcor >= 85) {
                                    if (Cs < 1407) {
                                        if (MVDiff < 318) {
                                            if (tcor < 64) {
                                                if (gchDC < 1) {
                                                    return 1;
                                                }
                                                else if (gchDC >= 1) {
                                                    if (SC < 4877) {
                                                        if (diffMVdiffVal < -3) {
                                                            if (mcTcor < 87) {
                                                                return 1;
                                                            }
                                                            else if (mcTcor >= 87) {
                                                                return 2;
                                                            }
                                                        }
                                                        else if (diffMVdiffVal >= -3) {
                                                            return 2;
                                                        }
                                                    }
                                                    else if (SC >= 4877) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                            else if (tcor >= 64) {
                                                if (Rs < 1068) {
                                                    if (mu_mv_mag_sq < 30) {
                                                        return 1;
                                                    }
                                                    else if (mu_mv_mag_sq >= 30) {
                                                        return 2;
                                                    }
                                                }
                                                else if (Rs >= 1068) {
                                                    if (refDCval < 74) {
                                                        return 4;
                                                    }
                                                    else if (refDCval >= 74) {
                                                        if (mu_mv_mag_sq < 33) {
                                                            if (posBalance < 56) {
                                                                if (Cs < 1256) {
                                                                    return 2;
                                                                }
                                                                else if (Cs >= 1256) {
                                                                    if (AFD < 15) {
                                                                        return 1;
                                                                    }
                                                                    else if (AFD >= 15) {
                                                                        return 2;
                                                                    }
                                                                }
                                                            }
                                                            else if (posBalance >= 56) {
                                                                return 2;
                                                            }
                                                        }
                                                        else if (mu_mv_mag_sq >= 33) {
                                                            if (mcTcor < 87) {
                                                                return 4;
                                                            }
                                                            else if (mcTcor >= 87) {
                                                                if (tcor < 87) {
                                                                    if (CsDiff < 3) {
                                                                        return 4;
                                                                    }
                                                                    else if (CsDiff >= 3) {
                                                                        return 2;
                                                                    }
                                                                }
                                                                else if (tcor >= 87) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (MVDiff >= 318) {
                                            if (Cs < 370) {
                                                return 8;
                                            }
                                            else if (Cs >= 370) {
                                                if (Rs < 1957) {
                                                    return 2;
                                                }
                                                else if (Rs >= 1957) {
                                                    return 4;
                                                }
                                            }
                                        }
                                    }
                                    else if (Cs >= 1407) {
                                        if (ssDCval < 92) {
                                            if (CsDiff < 11) {
                                                if (TSC < 9) {
                                                    if (Cs < 1696) {
                                                        return 1;
                                                    }
                                                    else if (Cs >= 1696) {
                                                        return 8;
                                                    }
                                                }
                                                else if (TSC >= 9) {
                                                    if (tcor < 82) {
                                                        if (Cs < 1495) {
                                                            return 2;
                                                        }
                                                        else if (Cs >= 1495) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (tcor >= 82) {
                                                        if (Rs < 1939) {
                                                            return 1;
                                                        }
                                                        else if (Rs >= 1939) {
                                                            return 2;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (CsDiff >= 11) {
                                                return 1;
                                            }
                                        }
                                        else if (ssDCval >= 92) {
                                            if (ssDCval < 117) {
                                                if (posBalance < 56) {
                                                    if (Cs < 1767) {
                                                        return 8;
                                                    }
                                                    else if (Cs >= 1767) {
                                                        return 4;
                                                    }
                                                }
                                                else if (posBalance >= 56) {
                                                    if (tcor < 86) {
                                                        if (CsDiff < 7) {
                                                            return 8;
                                                        }
                                                        else if (CsDiff >= 7) {
                                                            return 2;
                                                        }
                                                    }
                                                    else if (tcor >= 86) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                            else if (ssDCval >= 117) {
                                                if (refDCval < 122) {
                                                    if (tcor < 87) {
                                                        if (mcTcor < 95) {
                                                            return 4;
                                                        }
                                                        else if (mcTcor >= 95) {
                                                            return 2;
                                                        }
                                                    }
                                                    else if (tcor >= 87) {
                                                        return 8;
                                                    }
                                                }
                                                else if (refDCval >= 122) {
                                                    if (tcor < 71) {
                                                        if (AFD < 30) {
                                                            if (SC < 3498) {
                                                                if (AFD < 20) {
                                                                    if (AFD < 16) {
                                                                        return 8;
                                                                    }
                                                                    else if (AFD >= 16) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (AFD >= 20) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (SC >= 3498) {
                                                                if (RsCsDiff < 473) {
                                                                    return 1;
                                                                }
                                                                else if (RsCsDiff >= 473) {
                                                                    return 2;
                                                                }
                                                            }
                                                        }
                                                        else if (AFD >= 30) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (tcor >= 71) {
                                                        if (gchDC < 3) {
                                                            if (posBalance < 63) {
                                                                return 2;
                                                            }
                                                            else if (posBalance >= 63) {
                                                                if (Cs < 1542) {
                                                                    return 2;
                                                                }
                                                                else if (Cs >= 1542) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                        else if (gchDC >= 3) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                        }
                    }
                    else if (Cs >= 1800) {
                        if (tcor < 79) {
                            if (refDCval < 124) {
                                if (AFD < 17) {
                                    if (refDCval < 63) {
                                        if (ssDCval < 50) {
                                            if (diffMVdiffVal < 2) {
                                                return 4;
                                            }
                                            else if (diffMVdiffVal >= 2) {
                                                if (Rs < 1496) {
                                                    return 2;
                                                }
                                                else if (Rs >= 1496) {
                                                    return 8;
                                                }
                                            }
                                        }
                                        else if (ssDCval >= 50) {
                                            return 8;
                                        }
                                    }
                                    else if (refDCval >= 63) {
                                        if (negBalance < 39) {
                                            if (RsCsDiff < 123) {
                                                if (Rs < 4163) {
                                                    return 8;
                                                }
                                                else if (Rs >= 4163) {
                                                    return 1;
                                                }
                                            }
                                            else if (RsCsDiff >= 123) {
                                                return 1;
                                            }
                                        }
                                        else if (negBalance >= 39) {
                                            if (ssDCval < 82) {
                                                if (posBalance < 57) {
                                                    if (tcor < 78) {
                                                        return 8;
                                                    }
                                                    else if (tcor >= 78) {
                                                        if (Cs < 2344) {
                                                            return 4;
                                                        }
                                                        else if (Cs >= 2344) {
                                                            return 1;
                                                        }
                                                    }
                                                }
                                                else if (posBalance >= 57) {
                                                    return 4;
                                                }
                                            }
                                            else if (ssDCval >= 82) {
                                                if (posBalance < 55) {
                                                    if (AFD < 14) {
                                                        if (Rs < 3953) {
                                                            return 8;
                                                        }
                                                        else if (Rs >= 3953) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (AFD >= 14) {
                                                        if (mcTcor < 81) {
                                                            if (SC < 8188) {
                                                                return 4;
                                                            }
                                                            else if (SC >= 8188) {
                                                                return 2;
                                                            }
                                                        }
                                                        else if (mcTcor >= 81) {
                                                            return 2;
                                                        }
                                                    }
                                                }
                                                else if (posBalance >= 55) {
                                                    if (diffRsCsdiff < 62) {
                                                        return 4;
                                                    }
                                                    else if (diffRsCsdiff >= 62) {
                                                        if (CsDiff < 10) {
                                                            return 4;
                                                        }
                                                        else if (CsDiff >= 10) {
                                                            return 2;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (AFD >= 17) {
                                    if (mcTcor < 89) {
                                        if (RsCsDiff < 301) {
                                            if (RsCsDiff < 113) {
                                                return 8;
                                            }
                                            else if (RsCsDiff >= 113) {
                                                if (refDCval < 80) {
                                                    if (mu_mv_mag_sq < 29) {
                                                        return 8;
                                                    }
                                                    else if (mu_mv_mag_sq >= 29) {
                                                        return 1;
                                                    }
                                                }
                                                else if (refDCval >= 80) {
                                                    if (mu_mv_mag_sq < 7) {
                                                        if (CsDiff < 9) {
                                                            return 1;
                                                        }
                                                        else if (CsDiff >= 9) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (mu_mv_mag_sq >= 7) {
                                                        if (diffMVdiffVal < 21) {
                                                            if (SC < 6406) {
                                                                return 1;
                                                            }
                                                            else if (SC >= 6406) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (diffMVdiffVal >= 21) {
                                                            if (MVDiff < 200) {
                                                                if (Rs < 2097) {
                                                                    return 4;
                                                                }
                                                                else if (Rs >= 2097) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (MVDiff >= 200) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (RsCsDiff >= 301) {
                                            if (SC < 6471) {
                                                if (diffTSC < -2) {
                                                    if (tcor < 71) {
                                                        return 2;
                                                    }
                                                    else if (tcor >= 71) {
                                                        return 8;
                                                    }
                                                }
                                                else if (diffTSC >= -2) {
                                                    if (Cs < 2910) {
                                                        if (ssDCval < 116) {
                                                            if (posBalance < 50) {
                                                                return 2;
                                                            }
                                                            else if (posBalance >= 50) {
                                                                if (SC < 4977) {
                                                                    if (RsCsDiff < 423) {
                                                                        return 1;
                                                                    }
                                                                    else if (RsCsDiff >= 423) {
                                                                        if (Cs < 2398) {
                                                                            if (Cs < 2058) {
                                                                                return 2;
                                                                            }
                                                                            else if (Cs >= 2058) {
                                                                                return 8;
                                                                            }
                                                                        }
                                                                        else if (Cs >= 2398) {
                                                                            return 2;
                                                                        }
                                                                    }
                                                                }
                                                                else if (SC >= 4977) {
                                                                    if (diffAFD < -3) {
                                                                        return 1;
                                                                    }
                                                                    else if (diffAFD >= -3) {
                                                                        if (Cs < 2201) {
                                                                            if (Cs < 2117) {
                                                                                return 8;
                                                                            }
                                                                            else if (Cs >= 2117) {
                                                                                return 1;
                                                                            }
                                                                        }
                                                                        else if (Cs >= 2201) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (ssDCval >= 116) {
                                                            return 1;
                                                        }
                                                    }
                                                    else if (Cs >= 2910) {
                                                        if (refDCval < 78) {
                                                            return 1;
                                                        }
                                                        else if (refDCval >= 78) {
                                                            if (AFD < 23) {
                                                                return 8;
                                                            }
                                                            else if (AFD >= 23) {
                                                                if (Cs < 3199) {
                                                                    return 1;
                                                                }
                                                                else if (Cs >= 3199) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                            else if (SC >= 6471) {
                                                if (RsDiff < 20) {
                                                    if (Rs < 4335) {
                                                        return 8;
                                                    }
                                                    else if (Rs >= 4335) {
                                                        if (Cs < 2842) {
                                                            return 2;
                                                        }
                                                        else if (Cs >= 2842) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                                else if (RsDiff >= 20) {
                                                    if (RsDiff < 26) {
                                                        return 2;
                                                    }
                                                    else if (RsDiff >= 26) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (mcTcor >= 89) {
                                        if (negBalance < 43) {
                                            if (Rs < 3905) {
                                                return 1;
                                            }
                                            else if (Rs >= 3905) {
                                                return 4;
                                            }
                                        }
                                        else if (negBalance >= 43) {
                                            if (TSC < 10) {
                                                if (RsDiff < 10) {
                                                    return 4;
                                                }
                                                else if (RsDiff >= 10) {
                                                    return 2;
                                                }
                                            }
                                            else if (TSC >= 10) {
                                                if (MVDiff < 12) {
                                                    if (Cs < 2894) {
                                                        return 4;
                                                    }
                                                    else if (Cs >= 2894) {
                                                        return 8;
                                                    }
                                                }
                                                else if (MVDiff >= 12) {
                                                    if (SC < 6442) {
                                                        if (posBalance < 57) {
                                                            return 1;
                                                        }
                                                        else if (posBalance >= 57) {
                                                            if (RsDiff < 15) {
                                                                return 1;
                                                            }
                                                            else if (RsDiff >= 15) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                    else if (SC >= 6442) {
                                                        if (Cs < 3005) {
                                                            return 4;
                                                        }
                                                        else if (Cs >= 3005) {
                                                            return 2;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                            else if (refDCval >= 124) {
                                if (ssDCval < 139) {
                                    if (MVDiff < 34) {
                                        if (tcor < 68) {
                                            return 1;
                                        }
                                        else if (tcor >= 68) {
                                            if (TSC < 14) {
                                                if (negBalance < 52) {
                                                    if (Cs < 3189) {
                                                        return 1;
                                                    }
                                                    else if (Cs >= 3189) {
                                                        if (Rs < 5064) {
                                                            return 8;
                                                        }
                                                        else if (Rs >= 5064) {
                                                            return 1;
                                                        }
                                                    }
                                                }
                                                else if (negBalance >= 52) {
                                                    if (Rs < 2936) {
                                                        if (TSC < 13) {
                                                            return 1;
                                                        }
                                                        else if (TSC >= 13) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (Rs >= 2936) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                            else if (TSC >= 14) {
                                                if (MVDiff < 11) {
                                                    return 1;
                                                }
                                                else if (MVDiff >= 11) {
                                                    return 4;
                                                }
                                            }
                                        }
                                    }
                                    else if (MVDiff >= 34) {
                                        if (SC < 4542) {
                                            if (Cs < 1908) {
                                                return 1;
                                            }
                                            else if (Cs >= 1908) {
                                                return 2;
                                            }
                                        }
                                        else if (SC >= 4542) {
                                            if (Rs < 3094) {
                                                return 4;
                                            }
                                            else if (Rs >= 3094) {
                                                return 8;
                                            }
                                        }
                                    }
                                }
                                else if (ssDCval >= 139) {
                                    if (SC < 3929) {
                                        return 8;
                                    }
                                    else if (SC >= 3929) {
                                        if (MVDiff < 16) {
                                            if (refDCval < 141) {
                                                return 1;
                                            }
                                            else if (refDCval >= 141) {
                                                if (CsDiff < 12) {
                                                    if (Rs < 2966) {
                                                        return 8;
                                                    }
                                                    else if (Rs >= 2966) {
                                                        return 2;
                                                    }
                                                }
                                                else if (CsDiff >= 12) {
                                                    return 4;
                                                }
                                            }
                                        }
                                        else if (MVDiff >= 16) {
                                            if (mcTcor < 91) {
                                                if (Cs < 2872) {
                                                    return 4;
                                                }
                                                else if (Cs >= 2872) {
                                                    if (RsDiff < 24) {
                                                        return 2;
                                                    }
                                                    else if (RsDiff >= 24) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                            else if (mcTcor >= 91) {
                                                if (posBalance < 56) {
                                                    if (Rs < 3767) {
                                                        return 2;
                                                    }
                                                    else if (Rs >= 3767) {
                                                        return 4;
                                                    }
                                                }
                                                else if (posBalance >= 56) {
                                                    return 2;
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                        }
                        else if (tcor >= 79) {
                            if (ssDCval < 94) {
                                if (negBalance < 48) {
                                    if (MVDiff < 2) {
                                        if (CsDiff < 7) {
                                            return 4;
                                        }
                                        else if (CsDiff >= 7) {
                                            return 1;
                                        }
                                    }
                                    else if (MVDiff >= 2) {
                                        if (posBalance < 56) {
                                            if (RsCsDiff < 147) {
                                                if (tcor < 94) {
                                                    if (negBalance < 44) {
                                                        if (Rs < 2136) {
                                                            if (ssDCval < 87) {
                                                                if (RsCsDiff < 79) {
                                                                    if (AFD < 11) {
                                                                        return 8;
                                                                    }
                                                                    else if (AFD >= 11) {
                                                                        return 1;
                                                                    }
                                                                }
                                                                else if (RsCsDiff >= 79) {
                                                                    return 1;
                                                                }
                                                            }
                                                            else if (ssDCval >= 87) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (Rs >= 2136) {
                                                            if (ssDCval < 91) {
                                                                if (diffMVdiffVal < -2) {
                                                                    if (TSC < 10) {
                                                                        if (Rs < 2304) {
                                                                            if (CsDiff < 8) {
                                                                                return 8;
                                                                            }
                                                                            else if (CsDiff >= 8) {
                                                                                return 1;
                                                                            }
                                                                        }
                                                                        else if (Rs >= 2304) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                    else if (TSC >= 10) {
                                                                        if (Rs < 2421) {
                                                                            return 8;
                                                                        }
                                                                        else if (Rs >= 2421) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                }
                                                                else if (diffMVdiffVal >= -2) {
                                                                    if (TSC < 6) {
                                                                        return 4;
                                                                    }
                                                                    else if (TSC >= 6) {
                                                                        return 8;
                                                                    }
                                                                }
                                                            }
                                                            else if (ssDCval >= 91) {
                                                                if (diffMVdiffVal < -8) {
                                                                    return 8;
                                                                }
                                                                else if (diffMVdiffVal >= -8) {
                                                                    return 1;
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (negBalance >= 44) {
                                                        if (refDCval < 85) {
                                                            if (ssDCval < 54) {
                                                                return 4;
                                                            }
                                                            else if (ssDCval >= 54) {
                                                                if (MVDiff < 32) {
                                                                    return 8;
                                                                }
                                                                else if (MVDiff >= 32) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                        else if (refDCval >= 85) {
                                                            if (mu_mv_mag_sq < 45) {
                                                                if (Cs < 2071) {
                                                                    return 8;
                                                                }
                                                                else if (Cs >= 2071) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (mu_mv_mag_sq >= 45) {
                                                                if (Rs < 2504) {
                                                                    return 4;
                                                                }
                                                                else if (Rs >= 2504) {
                                                                    return 1;
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (tcor >= 94) {
                                                    if (Rs < 3218) {
                                                        return 1;
                                                    }
                                                    else if (Rs >= 3218) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                            else if (RsCsDiff >= 147) {
                                                if (MVDiff < 34) {
                                                    if (posBalance < 55) {
                                                        if (MVDiff < 23) {
                                                            if (mcTcor < 94) {
                                                                if (Cs < 2917) {
                                                                    if (mu_mv_mag_sq < 28) {
                                                                        return 1;
                                                                    }
                                                                    else if (mu_mv_mag_sq >= 28) {
                                                                        if (posBalance < 53) {
                                                                            if (posBalance < 46) {
                                                                                return 8;
                                                                            }
                                                                            else if (posBalance >= 46) {
                                                                                return 1;
                                                                            }
                                                                        }
                                                                        else if (posBalance >= 53) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                }
                                                                else if (Cs >= 2917) {
                                                                    if (ssDCval < 76) {
                                                                        return 1;
                                                                    }
                                                                    else if (ssDCval >= 76) {
                                                                        if (Rs < 3144) {
                                                                            if (AFD < 13) {
                                                                                return 1;
                                                                            }
                                                                            else if (AFD >= 13) {
                                                                                return 8;
                                                                            }
                                                                        }
                                                                        else if (Rs >= 3144) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                            else if (mcTcor >= 94) {
                                                                if (AFD < 15) {
                                                                    return 8;
                                                                }
                                                                else if (AFD >= 15) {
                                                                    return 1;
                                                                }
                                                            }
                                                        }
                                                        else if (MVDiff >= 23) {
                                                            return 1;
                                                        }
                                                    }
                                                    else if (posBalance >= 55) {
                                                        if (Cs < 2409) {
                                                            return 1;
                                                        }
                                                        else if (Cs >= 2409) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                                else if (MVDiff >= 34) {
                                                    if (SC < 4981) {
                                                        if (RsDiff < 10) {
                                                            return 4;
                                                        }
                                                        else if (RsDiff >= 10) {
                                                            return 1;
                                                        }
                                                    }
                                                    else if (SC >= 4981) {
                                                        if (Rs < 2624) {
                                                            return 4;
                                                        }
                                                        else if (Rs >= 2624) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (posBalance >= 56) {
                                            if (refDCval < 87) {
                                                if (RsDiff < 4) {
                                                    if (diffAFD < -1) {
                                                        return 2;
                                                    }
                                                    else if (diffAFD >= -1) {
                                                        return 8;
                                                    }
                                                }
                                                else if (RsDiff >= 4) {
                                                    if (mcTcor < 90) {
                                                        if (TSC < 11) {
                                                            return 2;
                                                        }
                                                        else if (TSC >= 11) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (mcTcor >= 90) {
                                                        if (AFD < 13) {
                                                            if (tcor < 91) {
                                                                return 8;
                                                            }
                                                            else if (tcor >= 91) {
                                                                if (gchDC < 6) {
                                                                    if (diffRsCsdiff < 1) {
                                                                        if (SC < 5972) {
                                                                            return 2;
                                                                        }
                                                                        else if (SC >= 5972) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                    else if (diffRsCsdiff >= 1) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (gchDC >= 6) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                        else if (AFD >= 13) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (refDCval >= 87) {
                                                if (Rs < 2572) {
                                                    if (MVDiff < 13) {
                                                        return 2;
                                                    }
                                                    else if (MVDiff >= 13) {
                                                        if (Cs < 2277) {
                                                            return 4;
                                                        }
                                                        else if (Cs >= 2277) {
                                                            return 1;
                                                        }
                                                    }
                                                }
                                                else if (Rs >= 2572) {
                                                    if (Rs < 2988) {
                                                        return 1;
                                                    }
                                                    else if (Rs >= 2988) {
                                                        if (Cs < 2351) {
                                                            return 8;
                                                        }
                                                        else if (Cs >= 2351) {
                                                            return 1;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (negBalance >= 48) {
                                    if (ssDCval < 77) {
                                        if (tcor < 89) {
                                            if (CsDiff < 9) {
                                                if (negBalance < 53) {
                                                    if (RsDiff < 9) {
                                                        return 8;
                                                    }
                                                    else if (RsDiff >= 9) {
                                                        if (ssDCval < 72) {
                                                            return 8;
                                                        }
                                                        else if (ssDCval >= 72) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                                else if (negBalance >= 53) {
                                                    return 4;
                                                }
                                            }
                                            else if (CsDiff >= 9) {
                                                if (Rs < 2013) {
                                                    return 1;
                                                }
                                                else if (Rs >= 2013) {
                                                    return 4;
                                                }
                                            }
                                        }
                                        else if (tcor >= 89) {
                                            if (SC < 6566) {
                                                if (TSC < 7) {
                                                    if (mu_mv_mag_sq < 38) {
                                                        if (posBalance < 57) {
                                                            return 8;
                                                        }
                                                        else if (posBalance >= 57) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (mu_mv_mag_sq >= 38) {
                                                        if (Cs < 3207) {
                                                            return 2;
                                                        }
                                                        else if (Cs >= 3207) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                                else if (TSC >= 7) {
                                                    if (SC < 6474) {
                                                        return 2;
                                                    }
                                                    else if (SC >= 6474) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                            else if (SC >= 6566) {
                                                return 2;
                                            }
                                        }
                                    }
                                    else if (ssDCval >= 77) {
                                        if (Rs < 2963) {
                                            if (MVDiff < 5) {
                                                if (SC < 4189) {
                                                    if (SC < 3469) {
                                                        return 4;
                                                    }
                                                    else if (SC >= 3469) {
                                                        return 2;
                                                    }
                                                }
                                                else if (SC >= 4189) {
                                                    if (diffAFD < -2) {
                                                        return 2;
                                                    }
                                                    else if (diffAFD >= -2) {
                                                        if (tcor < 92) {
                                                            return 4;
                                                        }
                                                        else if (tcor >= 92) {
                                                            if (posBalance < 60) {
                                                                return 4;
                                                            }
                                                            else if (posBalance >= 60) {
                                                                if (Rs < 2361) {
                                                                    return 2;
                                                                }
                                                                else if (Rs >= 2361) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                            else if (MVDiff >= 5) {
                                                if (RsCsDiff < 107) {
                                                    if (AFD < 14) {
                                                        if (SC < 3904) {
                                                            if (ssDCval < 89) {
                                                                if (TSC < 10) {
                                                                    return 8;
                                                                }
                                                                else if (TSC >= 10) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (ssDCval >= 89) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (SC >= 3904) {
                                                            if (Cs < 2285) {
                                                                if (RsCsDiff < 67) {
                                                                    return 2;
                                                                }
                                                                else if (RsCsDiff >= 67) {
                                                                    if (AFD < 10) {
                                                                        return 8;
                                                                    }
                                                                    else if (AFD >= 10) {
                                                                        return 1;
                                                                    }
                                                                }
                                                            }
                                                            else if (Cs >= 2285) {
                                                                if (Rs < 2671) {
                                                                    return 8;
                                                                }
                                                                else if (Rs >= 2671) {
                                                                    if (SC < 5317) {
                                                                        if (Rs < 2915) {
                                                                            if (negBalance < 50) {
                                                                                return 1;
                                                                            }
                                                                            else if (negBalance >= 50) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                        else if (Rs >= 2915) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                    else if (SC >= 5317) {
                                                                        return 4;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (AFD >= 14) {
                                                        if (refDCval < 87) {
                                                            if (diffAFD < 1) {
                                                                return 2;
                                                            }
                                                            else if (diffAFD >= 1) {
                                                                if (Rs < 2558) {
                                                                    return 2;
                                                                }
                                                                else if (Rs >= 2558) {
                                                                    return 1;
                                                                }
                                                            }
                                                        }
                                                        else if (refDCval >= 87) {
                                                            if (Cs < 2066) {
                                                                return 2;
                                                            }
                                                            else if (Cs >= 2066) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (RsCsDiff >= 107) {
                                                    if (MVDiff < 24) {
                                                        if (Rs < 2837) {
                                                            if (posBalance < 56) {
                                                                if (Rs < 2250) {
                                                                    return 2;
                                                                }
                                                                else if (Rs >= 2250) {
                                                                    if (AFD < 11) {
                                                                        return 8;
                                                                    }
                                                                    else if (AFD >= 11) {
                                                                        return 4;
                                                                    }
                                                                }
                                                            }
                                                            else if (posBalance >= 56) {
                                                                if (diffMVdiffVal < 5) {
                                                                    return 2;
                                                                }
                                                                else if (diffMVdiffVal >= 5) {
                                                                    if (SC < 4554) {
                                                                        return 1;
                                                                    }
                                                                    else if (SC >= 4554) {
                                                                        if (Cs < 2337) {
                                                                            return 2;
                                                                        }
                                                                        else if (Cs >= 2337) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (Rs >= 2837) {
                                                            if (MVDiff < 13) {
                                                                return 4;
                                                            }
                                                            else if (MVDiff >= 13) {
                                                                return 1;
                                                            }
                                                        }
                                                    }
                                                    else if (MVDiff >= 24) {
                                                        if (refDCval < 87) {
                                                            if (RsDiff < 11) {
                                                                if (Cs < 2111) {
                                                                    return 1;
                                                                }
                                                                else if (Cs >= 2111) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (RsDiff >= 11) {
                                                                return 1;
                                                            }
                                                        }
                                                        else if (refDCval >= 87) {
                                                            if (AFD < 16) {
                                                                return 2;
                                                            }
                                                            else if (AFD >= 16) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (Rs >= 2963) {
                                            if (TSC < 13) {
                                                if (MVDiff < 8) {
                                                    if (diffAFD < 1) {
                                                        return 8;
                                                    }
                                                    else if (diffAFD >= 1) {
                                                        if (Rs < 3066) {
                                                            return 8;
                                                        }
                                                        else if (Rs >= 3066) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                                else if (MVDiff >= 8) {
                                                    if (RsCsDiff < 163) {
                                                        if (AFD < 12) {
                                                            return 8;
                                                        }
                                                        else if (AFD >= 12) {
                                                            return 1;
                                                        }
                                                    }
                                                    else if (RsCsDiff >= 163) {
                                                        if (posBalance < 59) {
                                                            if (ssDCval < 84) {
                                                                return 4;
                                                            }
                                                            else if (ssDCval >= 84) {
                                                                return 2;
                                                            }
                                                        }
                                                        else if (posBalance >= 59) {
                                                            if (MVDiff < 11) {
                                                                return 4;
                                                            }
                                                            else if (MVDiff >= 11) {
                                                                return 1;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                            else if (TSC >= 13) {
                                                if (AFD < 18) {
                                                    return 8;
                                                }
                                                else if (AFD >= 18) {
                                                    if (Cs < 2751) {
                                                        return 1;
                                                    }
                                                    else if (Cs >= 2751) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                            else if (ssDCval >= 94) {
                                if (refDCval < 129) {
                                    if (RsCsDiff < 46) {
                                        if (Rs < 2884) {
                                            if (Rs < 2159) {
                                                return 4;
                                            }
                                            else if (Rs >= 2159) {
                                                return 1;
                                            }
                                        }
                                        else if (Rs >= 2884) {
                                            if (SC < 5299) {
                                                return 4;
                                            }
                                            else if (SC >= 5299) {
                                                return 8;
                                            }
                                        }
                                    }
                                    else if (RsCsDiff >= 46) {
                                        if (TSC < 12) {
                                            if (Cs < 2475) {
                                                if (refDCval < 101) {
                                                    if (negBalance < 51) {
                                                        if (AFD < 11) {
                                                            return 8;
                                                        }
                                                        else if (AFD >= 11) {
                                                            if (SC < 4320) {
                                                                if (AFD < 19) {
                                                                    return 4;
                                                                }
                                                                else if (AFD >= 19) {
                                                                    return 2;
                                                                }
                                                            }
                                                            else if (SC >= 4320) {
                                                                if (Cs < 2279) {
                                                                    if (posBalance < 61) {
                                                                        return 1;
                                                                    }
                                                                    else if (posBalance >= 61) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (Cs >= 2279) {
                                                                    if (refDCval < 99) {
                                                                        if (Rs < 3228) {
                                                                            return 4;
                                                                        }
                                                                        else if (Rs >= 3228) {
                                                                            return 2;
                                                                        }
                                                                    }
                                                                    else if (refDCval >= 99) {
                                                                        if (Rs < 3034) {
                                                                            return 8;
                                                                        }
                                                                        else if (Rs >= 3034) {
                                                                            return 2;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (negBalance >= 51) {
                                                        if (mcTcor < 83) {
                                                            return 8;
                                                        }
                                                        else if (mcTcor >= 83) {
                                                            if (diffRsCsdiff < -19) {
                                                                return 4;
                                                            }
                                                            else if (diffRsCsdiff >= -19) {
                                                                if (MVDiff < 12) {
                                                                    return 2;
                                                                }
                                                                else if (MVDiff >= 12) {
                                                                    return 1;
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (refDCval >= 101) {
                                                    if (CsDiff < 7) {
                                                        if (refDCval < 108) {
                                                            if (SC < 6177) {
                                                                if (TSC < 6) {
                                                                    return 8;
                                                                }
                                                                else if (TSC >= 6) {
                                                                    if (MVDiff < 17) {
                                                                        return 4;
                                                                    }
                                                                    else if (MVDiff >= 17) {
                                                                        return 8;
                                                                    }
                                                                }
                                                            }
                                                            else if (SC >= 6177) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (refDCval >= 108) {
                                                            if (negBalance < 48) {
                                                                if (Cs < 2104) {
                                                                    if (AFD < 12) {
                                                                        return 4;
                                                                    }
                                                                    else if (AFD >= 12) {
                                                                        return 2;
                                                                    }
                                                                }
                                                                else if (Cs >= 2104) {
                                                                    if (Rs < 4735) {
                                                                        if (tcor < 94) {
                                                                            if (diffMVdiffVal < -3) {
                                                                                if (TSC < 9) {
                                                                                    return 8;
                                                                                }
                                                                                else if (TSC >= 9) {
                                                                                    return 1;
                                                                                }
                                                                            }
                                                                            else if (diffMVdiffVal >= -3) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                        else if (tcor >= 94) {
                                                                            if (ssDCval < 119) {
                                                                                return 2;
                                                                            }
                                                                            else if (ssDCval >= 119) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (Rs >= 4735) {
                                                                        return 8;
                                                                    }
                                                                }
                                                            }
                                                            else if (negBalance >= 48) {
                                                                if (mu_mv_mag_sq < 46) {
                                                                    if (diffRsCsdiff < 12) {
                                                                        return 4;
                                                                    }
                                                                    else if (diffRsCsdiff >= 12) {
                                                                        if (TSC < 9) {
                                                                            if (Rs < 2647) {
                                                                                return 4;
                                                                            }
                                                                            else if (Rs >= 2647) {
                                                                                return 8;
                                                                            }
                                                                        }
                                                                        else if (TSC >= 9) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                }
                                                                else if (mu_mv_mag_sq >= 46) {
                                                                    if (RsCsDiff < 165) {
                                                                        if (refDCval < 120) {
                                                                            return 8;
                                                                        }
                                                                        else if (refDCval >= 120) {
                                                                            if (Rs < 2685) {
                                                                                return 8;
                                                                            }
                                                                            else if (Rs >= 2685) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (RsCsDiff >= 165) {
                                                                        return 4;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (CsDiff >= 7) {
                                                        if (MVDiff < 12) {
                                                            if (AFD < 12) {
                                                                if (RsCsDiff < 121) {
                                                                    if (diffAFD < -1) {
                                                                        if (diffRsCsdiff < 7) {
                                                                            return 8;
                                                                        }
                                                                        else if (diffRsCsdiff >= 7) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                    else if (diffAFD >= -1) {
                                                                        if (Cs < 1904) {
                                                                            return 4;
                                                                        }
                                                                        else if (Cs >= 1904) {
                                                                            if (diffRsCsdiff < -16) {
                                                                                return 4;
                                                                            }
                                                                            else if (diffRsCsdiff >= -16) {
                                                                                if (negBalance < 53) {
                                                                                    if (RsDiff < 7) {
                                                                                        if (negBalance < 52) {
                                                                                            return 4;
                                                                                        }
                                                                                        else if (negBalance >= 52) {
                                                                                            if (Cs < 2008) {
                                                                                                return 8;
                                                                                            }
                                                                                            else if (Cs >= 2008) {
                                                                                                return 4;
                                                                                            }
                                                                                        }
                                                                                    }
                                                                                    else if (RsDiff >= 7) {
                                                                                        if (mcTcor < 95) {
                                                                                            if (MVDiff < 8) {
                                                                                                return 8;
                                                                                            }
                                                                                            else if (MVDiff >= 8) {
                                                                                                if (Rs < 3830) {
                                                                                                    return 4;
                                                                                                }
                                                                                                else if (Rs >= 3830) {
                                                                                                    return 8;
                                                                                                }
                                                                                            }
                                                                                        }
                                                                                        else if (mcTcor >= 95) {
                                                                                            return 4;
                                                                                        }
                                                                                    }
                                                                                }
                                                                                else if (negBalance >= 53) {
                                                                                    return 4;
                                                                                }
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                                else if (RsCsDiff >= 121) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (AFD >= 12) {
                                                                if (Rs < 2814) {
                                                                    if (Cs < 2382) {
                                                                        return 1;
                                                                    }
                                                                    else if (Cs >= 2382) {
                                                                        return 8;
                                                                    }
                                                                }
                                                                else if (Rs >= 2814) {
                                                                    if (Rs < 2988) {
                                                                        return 4;
                                                                    }
                                                                    else if (Rs >= 2988) {
                                                                        return 2;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (MVDiff >= 12) {
                                                            if (Rs < 4315) {
                                                                if (mcTcor < 94) {
                                                                    if (ssDCval < 110) {
                                                                        if (Rs < 4119) {
                                                                            return 4;
                                                                        }
                                                                        else if (Rs >= 4119) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                    else if (ssDCval >= 110) {
                                                                        if (Rs < 3206) {
                                                                            if (RsCsDiff < 156) {
                                                                                return 4;
                                                                            }
                                                                            else if (RsCsDiff >= 156) {
                                                                                return 8;
                                                                            }
                                                                        }
                                                                        else if (Rs >= 3206) {
                                                                            if (Cs < 1894) {
                                                                                return 4;
                                                                            }
                                                                            else if (Cs >= 1894) {
                                                                                return 2;
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                                else if (mcTcor >= 94) {
                                                                    if (mu_mv_mag_sq < 72) {
                                                                        if (diffRsCsdiff < 12) {
                                                                            if (SC < 5367) {
                                                                                return 2;
                                                                            }
                                                                            else if (SC >= 5367) {
                                                                                if (Cs < 2307) {
                                                                                    if (refDCval < 103) {
                                                                                        return 2;
                                                                                    }
                                                                                    else if (refDCval >= 103) {
                                                                                        return 4;
                                                                                    }
                                                                                }
                                                                                else if (Cs >= 2307) {
                                                                                    return 2;
                                                                                }
                                                                            }
                                                                        }
                                                                        else if (diffRsCsdiff >= 12) {
                                                                            if (refDCval < 103) {
                                                                                if (Cs < 2359) {
                                                                                    return 2;
                                                                                }
                                                                                else if (Cs >= 2359) {
                                                                                    return 4;
                                                                                }
                                                                            }
                                                                            else if (refDCval >= 103) {
                                                                                return 2;
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (mu_mv_mag_sq >= 72) {
                                                                        return 2;
                                                                    }
                                                                }
                                                            }
                                                            else if (Rs >= 4315) {
                                                                if (negBalance < 34) {
                                                                    return 4;
                                                                }
                                                                else if (negBalance >= 34) {
                                                                    if (RsDiff < 8) {
                                                                        return 4;
                                                                    }
                                                                    else if (RsDiff >= 8) {
                                                                        if (AFD < 11) {
                                                                            if (Cs < 2383) {
                                                                                return 8;
                                                                            }
                                                                            else if (Cs >= 2383) {
                                                                                return 2;
                                                                            }
                                                                        }
                                                                        else if (AFD >= 11) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                            else if (Cs >= 2475) {
                                                if (SC < 7077) {
                                                    if (negBalance < 50) {
                                                        if (posBalance < 61) {
                                                            if (TSC < 11) {
                                                                if (Rs < 4303) {
                                                                    if (diffRsCsdiff < -45) {
                                                                        if (RsDiff < 8) {
                                                                            if (Cs < 2788) {
                                                                                return 4;
                                                                            }
                                                                            else if (Cs >= 2788) {
                                                                                return 2;
                                                                            }
                                                                        }
                                                                        else if (RsDiff >= 8) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                    else if (diffRsCsdiff >= -45) {
                                                                        if (diffRsCsdiff < 14) {
                                                                            return 8;
                                                                        }
                                                                        else if (diffRsCsdiff >= 14) {
                                                                            if (RsCsDiff < 137) {
                                                                                if (diffRsCsdiff < 22) {
                                                                                    return 1;
                                                                                }
                                                                                else if (diffRsCsdiff >= 22) {
                                                                                    return 8;
                                                                                }
                                                                            }
                                                                            else if (RsCsDiff >= 137) {
                                                                                return 8;
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                                else if (Rs >= 4303) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (TSC >= 11) {
                                                                if (CsDiff < 9) {
                                                                    return 1;
                                                                }
                                                                else if (CsDiff >= 9) {
                                                                    if (Cs < 2540) {
                                                                        return 2;
                                                                    }
                                                                    else if (Cs >= 2540) {
                                                                        return 8;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (posBalance >= 61) {
                                                            if (SC < 6002) {
                                                                if (AFD < 16) {
                                                                    return 8;
                                                                }
                                                                else if (AFD >= 16) {
                                                                    return 2;
                                                                }
                                                            }
                                                            else if (SC >= 6002) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                    else if (negBalance >= 50) {
                                                        if (ssDCval < 103) {
                                                            if (Rs < 2556) {
                                                                return 2;
                                                            }
                                                            else if (Rs >= 2556) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (ssDCval >= 103) {
                                                            if (MVDiff < 5) {
                                                                return 8;
                                                            }
                                                            else if (MVDiff >= 5) {
                                                                if (Rs < 3329) {
                                                                    return 4;
                                                                }
                                                                else if (Rs >= 3329) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (SC >= 7077) {
                                                    if (RsCsDiff < 205) {
                                                        if (mcTcor < 86) {
                                                            if (Cs < 2999) {
                                                                if (Rs < 4569) {
                                                                    return 1;
                                                                }
                                                                else if (Rs >= 4569) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (Cs >= 2999) {
                                                                return 2;
                                                            }
                                                        }
                                                        else if (mcTcor >= 86) {
                                                            if (Rs < 5111) {
                                                                if (mu_mv_mag_sq < 66) {
                                                                    if (Rs < 4636) {
                                                                        return 4;
                                                                    }
                                                                    else if (Rs >= 4636) {
                                                                        return 8;
                                                                    }
                                                                }
                                                                else if (mu_mv_mag_sq >= 66) {
                                                                    if (posBalance < 46) {
                                                                        if (ssDCval < 125) {
                                                                            if (Cs < 2531) {
                                                                                return 4;
                                                                            }
                                                                            else if (Cs >= 2531) {
                                                                                return 8;
                                                                            }
                                                                        }
                                                                        else if (ssDCval >= 125) {
                                                                            return 2;
                                                                        }
                                                                    }
                                                                    else if (posBalance >= 46) {
                                                                        return 2;
                                                                    }
                                                                }
                                                            }
                                                            else if (Rs >= 5111) {
                                                                if (mu_mv_mag_sq < 76) {
                                                                    return 4;
                                                                }
                                                                else if (mu_mv_mag_sq >= 76) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (RsCsDiff >= 205) {
                                                        if (mcTcor < 89) {
                                                            if (AFD < 13) {
                                                                return 4;
                                                            }
                                                            else if (AFD >= 13) {
                                                                return 2;
                                                            }
                                                        }
                                                        else if (mcTcor >= 89) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (TSC >= 12) {
                                            if (RsCsDiff < 189) {
                                                if (refDCval < 104) {
                                                    if (Cs < 2355) {
                                                        return 2;
                                                    }
                                                    else if (Cs >= 2355) {
                                                        if (SC < 5855) {
                                                            return 8;
                                                        }
                                                        else if (SC >= 5855) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                                else if (refDCval >= 104) {
                                                    if (Rs < 3559) {
                                                        if (Rs < 1210) {
                                                            return 2;
                                                        }
                                                        else if (Rs >= 1210) {
                                                            return 1;
                                                        }
                                                    }
                                                    else if (Rs >= 3559) {
                                                        if (Rs < 3874) {
                                                            return 4;
                                                        }
                                                        else if (Rs >= 3874) {
                                                            return 1;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (RsCsDiff >= 189) {
                                                if (ssDCval < 106) {
                                                    if (gchDC < 2) {
                                                        if (AFD < 19) {
                                                            if (Rs < 2971) {
                                                                return 8;
                                                            }
                                                            else if (Rs >= 2971) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (AFD >= 19) {
                                                            return 1;
                                                        }
                                                    }
                                                    else if (gchDC >= 2) {
                                                        return 4;
                                                    }
                                                }
                                                else if (ssDCval >= 106) {
                                                    if (mcTcor < 89) {
                                                        if (Rs < 3220) {
                                                            if (Rs < 2547) {
                                                                return 1;
                                                            }
                                                            else if (Rs >= 2547) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (Rs >= 3220) {
                                                            if (AFD < 19) {
                                                                return 2;
                                                            }
                                                            else if (AFD >= 19) {
                                                                return 1;
                                                            }
                                                        }
                                                    }
                                                    else if (mcTcor >= 89) {
                                                        if (tcor < 85) {
                                                            if (Rs < 2481) {
                                                                return 2;
                                                            }
                                                            else if (Rs >= 2481) {
                                                                if (Cs < 2757) {
                                                                    return 1;
                                                                }
                                                                else if (Cs >= 2757) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                        else if (tcor >= 85) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (refDCval >= 129) {
                                    if (tcor < 84) {
                                        if (RsCsDiff < 379) {
                                            if (posBalance < 59) {
                                                if (posBalance < 47) {
                                                    if (Rs < 1862) {
                                                        return 8;
                                                    }
                                                    else if (Rs >= 1862) {
                                                        return 2;
                                                    }
                                                }
                                                else if (posBalance >= 47) {
                                                    if (AFD < 24) {
                                                        return 1;
                                                    }
                                                    else if (AFD >= 24) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                            else if (posBalance >= 59) {
                                                return 2;
                                            }
                                        }
                                        else if (RsCsDiff >= 379) {
                                            return 4;
                                        }
                                    }
                                    else if (tcor >= 84) {
                                        if (AFD < 12) {
                                            if (Cs < 2304) {
                                                return 4;
                                            }
                                            else if (Cs >= 2304) {
                                                if (Rs < 3856) {
                                                    return 8;
                                                }
                                                else if (Rs >= 3856) {
                                                    return 2;
                                                }
                                            }
                                        }
                                        else if (AFD >= 12) {
                                            return 8;
                                        }
                                    }
                                }
                            }
                        }
                    }
                }
                else if (Cs >= 3288) {
                    if (ssDCval < 68) {
                        if (RsCsDiff < 46) {
                            if (TSC < 7) {
                                if (ssDCval < 54) {
                                    if (MVDiff < 20) {
                                        if (Cs < 3847) {
                                            if (SC < 6878) {
                                                if (Rs < 3475) {
                                                    if (MVDiff < 9) {
                                                        return 8;
                                                    }
                                                    else if (MVDiff >= 9) {
                                                        return 4;
                                                    }
                                                }
                                                else if (Rs >= 3475) {
                                                    return 8;
                                                }
                                            }
                                            else if (SC >= 6878) {
                                                return 4;
                                            }
                                        }
                                        else if (Cs >= 3847) {
                                            return 2;
                                        }
                                    }
                                    else if (MVDiff >= 20) {
                                        if (mu_mv_mag_sq < 36) {
                                            return 8;
                                        }
                                        else if (mu_mv_mag_sq >= 36) {
                                            if (posBalance < 53) {
                                                if (MVDiff < 68) {
                                                    return 8;
                                                }
                                                else if (MVDiff >= 68) {
                                                    return 4;
                                                }
                                            }
                                            else if (posBalance >= 53) {
                                                return 4;
                                            }
                                        }
                                    }
                                }
                                else if (ssDCval >= 54) {
                                    if (negBalance < 46) {
                                        if (refDCval < 62) {
                                            if (tcor < 93) {
                                                return 8;
                                            }
                                            else if (tcor >= 93) {
                                                if (mu_mv_mag_sq < 55) {
                                                    return 4;
                                                }
                                                else if (mu_mv_mag_sq >= 55) {
                                                    if (diffMVdiffVal < 4) {
                                                        return 8;
                                                    }
                                                    else if (diffMVdiffVal >= 4) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                        }
                                        else if (refDCval >= 62) {
                                            return 4;
                                        }
                                    }
                                    else if (negBalance >= 46) {
                                        return 8;
                                    }
                                }
                            }
                            else if (TSC >= 7) {
                                if (negBalance < 51) {
                                    if (SC < 7204) {
                                        if (MVDiff < 29) {
                                            if (Rs < 3432) {
                                                return 8;
                                            }
                                            else if (Rs >= 3432) {
                                                return 2;
                                            }
                                        }
                                        else if (MVDiff >= 29) {
                                            return 8;
                                        }
                                    }
                                    else if (SC >= 7204) {
                                        if (MVDiff < 20) {
                                            if (mu_mv_mag_sq < 47) {
                                                if (Cs < 3763) {
                                                    return 4;
                                                }
                                                else if (Cs >= 3763) {
                                                    if (posBalance < 59) {
                                                        return 2;
                                                    }
                                                    else if (posBalance >= 59) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                            else if (mu_mv_mag_sq >= 47) {
                                                return 4;
                                            }
                                        }
                                        else if (MVDiff >= 20) {
                                            if (posBalance < 58) {
                                                return 4;
                                            }
                                            else if (posBalance >= 58) {
                                                if (ssDCval < 57) {
                                                    return 2;
                                                }
                                                else if (ssDCval >= 57) {
                                                    return 8;
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (negBalance >= 51) {
                                    if (posBalance < 52) {
                                        return 8;
                                    }
                                    else if (posBalance >= 52) {
                                        if (posBalance < 59) {
                                            if (MVDiff < 16) {
                                                if (Cs < 3648) {
                                                    return 4;
                                                }
                                                else if (Cs >= 3648) {
                                                    if (posBalance < 57) {
                                                        return 2;
                                                    }
                                                    else if (posBalance >= 57) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                            else if (MVDiff >= 16) {
                                                if (mcTcor < 94) {
                                                    return 2;
                                                }
                                                else if (mcTcor >= 94) {
                                                    if (Rs < 3381) {
                                                        return 8;
                                                    }
                                                    else if (Rs >= 3381) {
                                                        if (Cs < 3643) {
                                                            return 8;
                                                        }
                                                        else if (Cs >= 3643) {
                                                            return 2;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (posBalance >= 59) {
                                            if (Cs < 3616) {
                                                if (TSC < 8) {
                                                    return 8;
                                                }
                                                else if (TSC >= 8) {
                                                    return 2;
                                                }
                                            }
                                            else if (Cs >= 3616) {
                                                return 4;
                                            }
                                        }
                                    }
                                }
                            }
                        }
                        else if (RsCsDiff >= 46) {
                            if (mcTcor < 83) {
                                return 8;
                            }
                            else if (mcTcor >= 83) {
                                if (mcTcor < 95) {
                                    if (MVDiff < 32) {
                                        if (TSC < 8) {
                                            if (AFD < 11) {
                                                return 4;
                                            }
                                            else if (AFD >= 11) {
                                                return 8;
                                            }
                                        }
                                        else if (TSC >= 8) {
                                            return 4;
                                        }
                                    }
                                    else if (MVDiff >= 32) {
                                        if (RsDiff < 9) {
                                            return 8;
                                        }
                                        else if (RsDiff >= 9) {
                                            if (SC < 8161) {
                                                return 8;
                                            }
                                            else if (SC >= 8161) {
                                                return 4;
                                            }
                                        }
                                    }
                                }
                                else if (mcTcor >= 95) {
                                    return 8;
                                }
                            }
                        }
                    }
                    else if (ssDCval >= 68) {
                        if (Rs < 4178) {
                            if (tcor < 72) {
                                if (ssDCval < 131) {
                                    if (SC < 6781) {
                                        if (Rs < 2351) {
                                            return 4;
                                        }
                                        else if (Rs >= 2351) {
                                            return 1;
                                        }
                                    }
                                    else if (SC >= 6781) {
                                        if (RsCsDiff < 451) {
                                            if (Rs < 3596) {
                                                return 8;
                                            }
                                            else if (Rs >= 3596) {
                                                return 4;
                                            }
                                        }
                                        else if (RsCsDiff >= 451) {
                                            return 8;
                                        }
                                    }
                                }
                                else if (ssDCval >= 131) {
                                    return 1;
                                }
                            }
                            else if (tcor >= 72) {
                                if (ssDCval < 70) {
                                    if (posBalance < 49) {
                                        return 4;
                                    }
                                    else if (posBalance >= 49) {
                                        if (Rs < 3962) {
                                            return 8;
                                        }
                                        else if (Rs >= 3962) {
                                            return 4;
                                        }
                                    }
                                }
                                else if (ssDCval >= 70) {
                                    return 8;
                                }
                            }
                        }
                        else if (Rs >= 4178) {
                            if (Cs < 4035) {
                                if (posBalance < 61) {
                                    if (tcor < 76) {
                                        if (SC < 8037) {
                                            if (MVDiff < 29) {
                                                if (Cs < 3496) {
                                                    return 1;
                                                }
                                                else if (Cs >= 3496) {
                                                    return 4;
                                                }
                                            }
                                            else if (MVDiff >= 29) {
                                                return 8;
                                            }
                                        }
                                        else if (SC >= 8037) {
                                            if (diffAFD < 3) {
                                                if (gchDC < 3) {
                                                    if (gchDC < 2) {
                                                        if (Cs < 3292) {
                                                            return 2;
                                                        }
                                                        else if (Cs >= 3292) {
                                                            if (diffTSC < -1) {
                                                                if (AFD < 32) {
                                                                    return 4;
                                                                }
                                                                else if (AFD >= 32) {
                                                                    return 2;
                                                                }
                                                            }
                                                            else if (diffTSC >= -1) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                    else if (gchDC >= 2) {
                                                        if (Rs < 5140) {
                                                            return 8;
                                                        }
                                                        else if (Rs >= 5140) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                                else if (gchDC >= 3) {
                                                    return 2;
                                                }
                                            }
                                            else if (diffAFD >= 3) {
                                                if (MVDiff < 23) {
                                                    return 8;
                                                }
                                                else if (MVDiff >= 23) {
                                                    if (Rs < 5068) {
                                                        return 8;
                                                    }
                                                    else if (Rs >= 5068) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (tcor >= 76) {
                                        if (refDCval < 122) {
                                            return 4;
                                        }
                                        else if (refDCval >= 122) {
                                            if (negBalance < 62) {
                                                return 8;
                                            }
                                            else if (negBalance >= 62) {
                                                return 4;
                                            }
                                        }
                                    }
                                }
                                else if (posBalance >= 61) {
                                    if (mu_mv_mag_sq < 22) {
                                        if (refDCval < 159) {
                                            if (CsDiff < 16) {
                                                return 4;
                                            }
                                            else if (CsDiff >= 16) {
                                                return 2;
                                            }
                                        }
                                        else if (refDCval >= 159) {
                                            return 2;
                                        }
                                    }
                                    else if (mu_mv_mag_sq >= 22) {
                                        return 8;
                                    }
                                }
                            }
                            else if (Cs >= 4035) {
                                if (Rs < 6604) {
                                    if (negBalance < 55) {
                                        if (Cs < 4327) {
                                            if (AFD < 22) {
                                                if (RsCsDiff < 371) {
                                                    if (negBalance < 48) {
                                                        return 4;
                                                    }
                                                    else if (negBalance >= 48) {
                                                        return 8;
                                                    }
                                                }
                                                else if (RsCsDiff >= 371) {
                                                    if (mcTcor < 69) {
                                                        return 1;
                                                    }
                                                    else if (mcTcor >= 69) {
                                                        if (AFD < 21) {
                                                            return 8;
                                                        }
                                                        else if (AFD >= 21) {
                                                            if (Rs < 5390) {
                                                                return 4;
                                                            }
                                                            else if (Rs >= 5390) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                            else if (AFD >= 22) {
                                                if (TSC < 17) {
                                                    return 8;
                                                }
                                                else if (TSC >= 17) {
                                                    return 4;
                                                }
                                            }
                                        }
                                        else if (Cs >= 4327) {
                                            if (tcor < 84) {
                                                if (Rs < 4441) {
                                                    if (Cs < 4845) {
                                                        return 8;
                                                    }
                                                    else if (Cs >= 4845) {
                                                        return 1;
                                                    }
                                                }
                                                else if (Rs >= 4441) {
                                                    if (Rs < 6258) {
                                                        return 8;
                                                    }
                                                    else if (Rs >= 6258) {
                                                        if (Cs < 4832) {
                                                            return 2;
                                                        }
                                                        else if (Cs >= 4832) {
                                                            if (RsDiff < 23) {
                                                                if (Rs < 6301) {
                                                                    return 1;
                                                                }
                                                                else if (Rs >= 6301) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (RsDiff >= 23) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                            else if (tcor >= 84) {
                                                if (mu_mv_mag_sq < 33) {
                                                    if (AFD < 19) {
                                                        return 8;
                                                    }
                                                    else if (AFD >= 19) {
                                                        return 1;
                                                    }
                                                }
                                                else if (mu_mv_mag_sq >= 33) {
                                                    if (mcTcor < 91) {
                                                        return 1;
                                                    }
                                                    else if (mcTcor >= 91) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (negBalance >= 55) {
                                        if (diffMVdiffVal < -36) {
                                            return 2;
                                        }
                                        else if (diffMVdiffVal >= -36) {
                                            if (refDCval < 157) {
                                                return 8;
                                            }
                                            else if (refDCval >= 157) {
                                                return 2;
                                            }
                                        }
                                    }
                                }
                                else if (Rs >= 6604) {
                                    if (MVDiff < 24) {
                                        if (SC < 12265) {
                                            if (ssDCval < 113) {
                                                if (AFD < 16) {
                                                    return 4;
                                                }
                                                else if (AFD >= 16) {
                                                    return 8;
                                                }
                                            }
                                            else if (ssDCval >= 113) {
                                                if (negBalance < 56) {
                                                    if (Cs < 4544) {
                                                        if (SC < 11555) {
                                                            return 2;
                                                        }
                                                        else if (SC >= 11555) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (Cs >= 4544) {
                                                        return 4;
                                                    }
                                                }
                                                else if (negBalance >= 56) {
                                                    return 2;
                                                }
                                            }
                                        }
                                        else if (SC >= 12265) {
                                            if (refDCval < 92) {
                                                if (Cs < 6422) {
                                                    return 4;
                                                }
                                                else if (Cs >= 6422) {
                                                    if (mcTcor < 86) {
                                                        return 4;
                                                    }
                                                    else if (mcTcor >= 86) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                            else if (refDCval >= 92) {
                                                if (negBalance < 54) {
                                                    if (MVDiff < 16) {
                                                        return 8;
                                                    }
                                                    else if (MVDiff >= 16) {
                                                        if (RsDiff < 36) {
                                                            return 8;
                                                        }
                                                        else if (RsDiff >= 36) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                                else if (negBalance >= 54) {
                                                    if (RsDiff < 44) {
                                                        if (ssDCval < 131) {
                                                            return 8;
                                                        }
                                                        else if (ssDCval >= 131) {
                                                            if (TSC < 22) {
                                                                if (CsDiff < 21) {
                                                                    return 4;
                                                                }
                                                                else if (CsDiff >= 21) {
                                                                    if (RsCsDiff < 1965) {
                                                                        return 8;
                                                                    }
                                                                    else if (RsCsDiff >= 1965) {
                                                                        if (Rs < 8401) {
                                                                            return 8;
                                                                        }
                                                                        else if (Rs >= 8401) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                            else if (TSC >= 22) {
                                                                if (AFD < 41) {
                                                                    return 4;
                                                                }
                                                                else if (AFD >= 41) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (RsDiff >= 44) {
                                                        if (Rs < 14703) {
                                                            if (ssDCval < 132) {
                                                                return 4;
                                                            }
                                                            else if (ssDCval >= 132) {
                                                                return 2;
                                                            }
                                                        }
                                                        else if (Rs >= 14703) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (MVDiff >= 24) {
                                        if (Cs < 6295) {
                                            if (refDCval < 80) {
                                                if (Cs < 5232) {
                                                    return 8;
                                                }
                                                else if (Cs >= 5232) {
                                                    return 4;
                                                }
                                            }
                                            else if (refDCval >= 80) {
                                                if (mcTcor < 83) {
                                                    return 8;
                                                }
                                                else if (mcTcor >= 83) {
                                                    return 2;
                                                }
                                            }
                                        }
                                        else if (Cs >= 6295) {
                                            if (mcTcor < 85) {
                                                return 8;
                                            }
                                            else if (mcTcor >= 85) {
                                                return 2;
                                            }
                                        }
                                    }
                                }
                            }
                        }
                    }
                }
            }
        }
        else if (mu_mv_mag_sq >= 83) {
            if (negBalance < 36) {
                if (Rs < 684) {
                    if (RsCsDiff < 1) {
                        if (mu_mv_mag_sq < 108) {
                            if (Cs < 443) {
                                if (mu_mv_mag_sq < 86) {
                                    if (Rs < 197) {
                                        if (Rs < 28) {
                                            return 1;
                                        }
                                        else if (Rs >= 28) {
                                            return 4;
                                        }
                                    }
                                    else if (Rs >= 197) {
                                        return 8;
                                    }
                                }
                                else if (mu_mv_mag_sq >= 86) {
                                    if (Rs < 144) {
                                        if (MVDiff < 4) {
                                            return 1;
                                        }
                                        else if (MVDiff >= 4) {
                                            return 4;
                                        }
                                    }
                                    else if (Rs >= 144) {
                                        return 4;
                                    }
                                }
                            }
                            else if (Cs >= 443) {
                                if (refDCval < 165) {
                                    if (posBalance < 38) {
                                        if (mcTcor < 99) {
                                            if (Rs < 488) {
                                                return 2;
                                            }
                                            else if (Rs >= 488) {
                                                return 4;
                                            }
                                        }
                                        else if (mcTcor >= 99) {
                                            return 4;
                                        }
                                    }
                                    else if (posBalance >= 38) {
                                        return 2;
                                    }
                                }
                                else if (refDCval >= 165) {
                                    if (negBalance < 9) {
                                        return 4;
                                    }
                                    else if (negBalance >= 9) {
                                        return 2;
                                    }
                                }
                            }
                        }
                        else if (mu_mv_mag_sq >= 108) {
                            if (Cs < 490) {
                                if (diffTSC < -1) {
                                    if (mu_mv_mag_sq < 115) {
                                        if (MVDiff < 77) {
                                            return 8;
                                        }
                                        else if (MVDiff >= 77) {
                                            return 1;
                                        }
                                    }
                                    else if (mu_mv_mag_sq >= 115) {
                                        return 8;
                                    }
                                }
                                else if (diffTSC >= -1) {
                                    if (Cs < 129) {
                                        if (AFD < 7) {
                                            return 4;
                                        }
                                        else if (AFD >= 7) {
                                            return 2;
                                        }
                                    }
                                    else if (Cs >= 129) {
                                        if (refDCval < 71) {
                                            return 8;
                                        }
                                        else if (refDCval >= 71) {
                                            if (Cs < 302) {
                                                return 8;
                                            }
                                            else if (Cs >= 302) {
                                                if (negBalance < 16) {
                                                    if (ssDCval < 124) {
                                                        if (SC < 498) {
                                                            return 8;
                                                        }
                                                        else if (SC >= 498) {
                                                            if (Cs < 311) {
                                                                return 4;
                                                            }
                                                            else if (Cs >= 311) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                    else if (ssDCval >= 124) {
                                                        return 4;
                                                    }
                                                }
                                                else if (negBalance >= 16) {
                                                    if (SC < 781) {
                                                        if (posBalance < 32) {
                                                            if (Cs < 318) {
                                                                return 4;
                                                            }
                                                            else if (Cs >= 318) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (posBalance >= 32) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (SC >= 781) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                            else if (Cs >= 490) {
                                if (Cs < 823) {
                                    return 4;
                                }
                                else if (Cs >= 823) {
                                    return 8;
                                }
                            }
                        }
                    }
                    else if (RsCsDiff >= 1) {
                        if (tcor < 99) {
                            if (mu_mv_mag_sq < 89) {
                                if (SC < 749) {
                                    return 8;
                                }
                                else if (SC >= 749) {
                                    return 4;
                                }
                            }
                            else if (mu_mv_mag_sq >= 89) {
                                if (negBalance < 22) {
                                    if (refDCval < 161) {
                                        if (Scindex < 3) {
                                            if (ssDCval < 111) {
                                                return 4;
                                            }
                                            else if (ssDCval >= 111) {
                                                return 8;
                                            }
                                        }
                                        else if (Scindex >= 3) {
                                            return 8;
                                        }
                                    }
                                    else if (refDCval >= 161) {
                                        return 4;
                                    }
                                }
                                else if (negBalance >= 22) {
                                    if (Rs < 248) {
                                        return 8;
                                    }
                                    else if (Rs >= 248) {
                                        if (ssDCval < 75) {
                                            return 8;
                                        }
                                        else if (ssDCval >= 75) {
                                            if (posBalance < 62) {
                                                return 4;
                                            }
                                            else if (posBalance >= 62) {
                                                if (Cs < 446) {
                                                    return 4;
                                                }
                                                else if (Cs >= 446) {
                                                    return 8;
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                        }
                        else if (tcor >= 99) {
                            if (ssDCval < 148) {
                                if (ssDCval < 84) {
                                    if (ssDCval < 75) {
                                        return 8;
                                    }
                                    else if (ssDCval >= 75) {
                                        return 4;
                                    }
                                }
                                else if (ssDCval >= 84) {
                                    return 8;
                                }
                            }
                            else if (ssDCval >= 148) {
                                return 2;
                            }
                        }
                    }
                }
                else if (Rs >= 684) {
                    if (MVDiff < 3) {
                        if (mu_mv_mag_sq < 121) {
                            if (Cs < 1992) {
                                if (Rs < 2862) {
                                    if (negBalance < 7) {
                                        if (Cs < 713) {
                                            if (Cs < 650) {
                                                return 8;
                                            }
                                            else if (Cs >= 650) {
                                                return 4;
                                            }
                                        }
                                        else if (Cs >= 713) {
                                            if (AFD < 3) {
                                                if (diffMVdiffVal < -2) {
                                                    if (Cs < 1183) {
                                                        return 8;
                                                    }
                                                    else if (Cs >= 1183) {
                                                        if (CsDiff < 1) {
                                                            return 4;
                                                        }
                                                        else if (CsDiff >= 1) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                                else if (diffMVdiffVal >= -2) {
                                                    return 8;
                                                }
                                            }
                                            else if (AFD >= 3) {
                                                return 4;
                                            }
                                        }
                                    }
                                    else if (negBalance >= 7) {
                                        if (negBalance < 33) {
                                            if (diffMVdiffVal < -2) {
                                                if (RsCsDiff < 4) {
                                                    if (Cs < 557) {
                                                        return 8;
                                                    }
                                                    else if (Cs >= 557) {
                                                        return 1;
                                                    }
                                                }
                                                else if (RsCsDiff >= 4) {
                                                    return 8;
                                                }
                                            }
                                            else if (diffMVdiffVal >= -2) {
                                                return 8;
                                            }
                                        }
                                        else if (negBalance >= 33) {
                                            if (posBalance < 44) {
                                                if (Rs < 1713) {
                                                    return 4;
                                                }
                                                else if (Rs >= 1713) {
                                                    return 8;
                                                }
                                            }
                                            else if (posBalance >= 44) {
                                                return 8;
                                            }
                                        }
                                    }
                                }
                                else if (Rs >= 2862) {
                                    if (SC < 4931) {
                                        return 1;
                                    }
                                    else if (SC >= 4931) {
                                        return 8;
                                    }
                                }
                            }
                            else if (Cs >= 1992) {
                                if (refDCval < 116) {
                                    if (Rs < 2134) {
                                        if (Rs < 1727) {
                                            if (Rs < 1272) {
                                                return 8;
                                            }
                                            else if (Rs >= 1272) {
                                                return 4;
                                            }
                                        }
                                        else if (Rs >= 1727) {
                                            return 1;
                                        }
                                    }
                                    else if (Rs >= 2134) {
                                        if (tcor < 95) {
                                            return 8;
                                        }
                                        else if (tcor >= 95) {
                                            if (Rs < 3419) {
                                                if (negBalance < 34) {
                                                    if (mu_mv_mag_sq < 116) {
                                                        if (CsDiff < 7) {
                                                            return 8;
                                                        }
                                                        else if (CsDiff >= 7) {
                                                            if (Rs < 2648) {
                                                                return 8;
                                                            }
                                                            else if (Rs >= 2648) {
                                                                return 1;
                                                            }
                                                        }
                                                    }
                                                    else if (mu_mv_mag_sq >= 116) {
                                                        if (RsDiff < 5) {
                                                            if (RsCsDiff < 22) {
                                                                if (diffRsCsdiff < 3) {
                                                                    return 8;
                                                                }
                                                                else if (diffRsCsdiff >= 3) {
                                                                    return 1;
                                                                }
                                                            }
                                                            else if (RsCsDiff >= 22) {
                                                                if (posBalance < 40) {
                                                                    return 8;
                                                                }
                                                                else if (posBalance >= 40) {
                                                                    if (negBalance < 29) {
                                                                        return 1;
                                                                    }
                                                                    else if (negBalance >= 29) {
                                                                        return 8;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (RsDiff >= 5) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                                else if (negBalance >= 34) {
                                                    if (ssDCval < 90) {
                                                        return 4;
                                                    }
                                                    else if (ssDCval >= 90) {
                                                        if (Rs < 2675) {
                                                            return 1;
                                                        }
                                                        else if (Rs >= 2675) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (Rs >= 3419) {
                                                return 8;
                                            }
                                        }
                                    }
                                }
                                else if (refDCval >= 116) {
                                    if (Rs < 4407) {
                                        if (Cs < 2274) {
                                            return 4;
                                        }
                                        else if (Cs >= 2274) {
                                            if (Rs < 2431) {
                                                return 8;
                                            }
                                            else if (Rs >= 2431) {
                                                return 4;
                                            }
                                        }
                                    }
                                    else if (Rs >= 4407) {
                                        return 8;
                                    }
                                }
                            }
                        }
                        else if (mu_mv_mag_sq >= 121) {
                            if (ssDCval < 103) {
                                if (negBalance < 34) {
                                    if (diffMVdiffVal < 1) {
                                        if (RsCsDiff < 38) {
                                            return 8;
                                        }
                                        else if (RsCsDiff >= 38) {
                                            if (refDCval < 51) {
                                                if (Scindex < 5) {
                                                    return 8;
                                                }
                                                else if (Scindex >= 5) {
                                                    return 4;
                                                }
                                            }
                                            else if (refDCval >= 51) {
                                                return 8;
                                            }
                                        }
                                    }
                                    else if (diffMVdiffVal >= 1) {
                                        if (refDCval < 93) {
                                            if (diffRsCsdiff < -17) {
                                                if (refDCval < 51) {
                                                    return 4;
                                                }
                                                else if (refDCval >= 51) {
                                                    return 8;
                                                }
                                            }
                                            else if (diffRsCsdiff >= -17) {
                                                return 8;
                                            }
                                        }
                                        else if (refDCval >= 93) {
                                            if (Cs < 2897) {
                                                return 1;
                                            }
                                            else if (Cs >= 2897) {
                                                return 8;
                                            }
                                        }
                                    }
                                }
                                else if (negBalance >= 34) {
                                    if (Rs < 2441) {
                                        if (SC < 3902) {
                                            return 8;
                                        }
                                        else if (SC >= 3902) {
                                            return 4;
                                        }
                                    }
                                    else if (Rs >= 2441) {
                                        return 8;
                                    }
                                }
                            }
                            else if (ssDCval >= 103) {
                                if (refDCval < 104) {
                                    if (SC < 7803) {
                                        if (Cs < 4003) {
                                            if (SC < 7610) {
                                                if (diffRsCsdiff < 2) {
                                                    if (posBalance < 21) {
                                                        return 1;
                                                    }
                                                    else if (posBalance >= 21) {
                                                        if (negBalance < 1) {
                                                            return 1;
                                                        }
                                                        else if (negBalance >= 1) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                                else if (diffRsCsdiff >= 2) {
                                                    if (Cs < 3543) {
                                                        return 8;
                                                    }
                                                    else if (Cs >= 3543) {
                                                        return 1;
                                                    }
                                                }
                                            }
                                            else if (SC >= 7610) {
                                                return 8;
                                            }
                                        }
                                        else if (Cs >= 4003) {
                                            return 1;
                                        }
                                    }
                                    else if (SC >= 7803) {
                                        return 8;
                                    }
                                }
                                else if (refDCval >= 104) {
                                    if (mu_mv_mag_sq < 125) {
                                        if (RsCsDiff < 7) {
                                            return 8;
                                        }
                                        else if (RsCsDiff >= 7) {
                                            if (SC < 6446) {
                                                if (Rs < 3139) {
                                                    if (SC < 2969) {
                                                        return 1;
                                                    }
                                                    else if (SC >= 2969) {
                                                        return 8;
                                                    }
                                                }
                                                else if (Rs >= 3139) {
                                                    if (Cs < 2864) {
                                                        return 8;
                                                    }
                                                    else if (Cs >= 2864) {
                                                        return 1;
                                                    }
                                                }
                                            }
                                            else if (SC >= 6446) {
                                                if (ssDCval < 116) {
                                                    return 8;
                                                }
                                                else if (ssDCval >= 116) {
                                                    if (posBalance < 27) {
                                                        return 8;
                                                    }
                                                    else if (posBalance >= 27) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (mu_mv_mag_sq >= 125) {
                                        if (diffMVdiffVal < -33) {
                                            return 4;
                                        }
                                        else if (diffMVdiffVal >= -33) {
                                            if (mcTcor < 95) {
                                                if (RsDiff < 2) {
                                                    return 1;
                                                }
                                                else if (RsDiff >= 2) {
                                                    if (Rs < 4411) {
                                                        return 8;
                                                    }
                                                    else if (Rs >= 4411) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                            else if (mcTcor >= 95) {
                                                return 8;
                                            }
                                        }
                                    }
                                }
                            }
                        }
                    }
                    else if (MVDiff >= 3) {
                        if (refDCval < 121) {
                            if (TSCindex < 1) {
                                if (SC < 2312) {
                                    if (refDCval < 26) {
                                        return 8;
                                    }
                                    else if (refDCval >= 26) {
                                        if (Rs < 828) {
                                            if (Rs < 797) {
                                                return 1;
                                            }
                                            else if (Rs >= 797) {
                                                return 8;
                                            }
                                        }
                                        else if (Rs >= 828) {
                                            if (RsDiff < 1) {
                                                return 4;
                                            }
                                            else if (RsDiff >= 1) {
                                                return 2;
                                            }
                                        }
                                    }
                                }
                                else if (SC >= 2312) {
                                    if (SC < 2373) {
                                        return 2;
                                    }
                                    else if (SC >= 2373) {
                                        if (posBalance < 4) {
                                            return 2;
                                        }
                                        else if (posBalance >= 4) {
                                            if (refDCval < 104) {
                                                return 8;
                                            }
                                            else if (refDCval >= 104) {
                                                return 2;
                                            }
                                        }
                                    }
                                }
                            }
                            else if (TSCindex >= 1) {
                                if (SC < 4706) {
                                    if (diffAFD < -1) {
                                        if (Cs < 2059) {
                                            if (RsCsDiff < 7) {
                                                if (SC < 1928) {
                                                    return 8;
                                                }
                                                else if (SC >= 1928) {
                                                    if (refDCval < 53) {
                                                        return 1;
                                                    }
                                                    else if (refDCval >= 53) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                            else if (RsCsDiff >= 7) {
                                                if (Cs < 855) {
                                                    if (AFD < 18) {
                                                        return 4;
                                                    }
                                                    else if (AFD >= 18) {
                                                        return 8;
                                                    }
                                                }
                                                else if (Cs >= 855) {
                                                    if (ssDCval < 46) {
                                                        if (Cs < 1835) {
                                                            return 8;
                                                        }
                                                        else if (Cs >= 1835) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (ssDCval >= 46) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                        }
                                        else if (Cs >= 2059) {
                                            if (refDCval < 79) {
                                                if (Cs < 2128) {
                                                    return 4;
                                                }
                                                else if (Cs >= 2128) {
                                                    return 8;
                                                }
                                            }
                                            else if (refDCval >= 79) {
                                                return 1;
                                            }
                                        }
                                    }
                                    else if (diffAFD >= -1) {
                                        if (diffMVdiffVal < 10) {
                                            if (refDCval < 73) {
                                                if (RsDiff < 2) {
                                                    if (diffMVdiffVal < -3) {
                                                        return 1;
                                                    }
                                                    else if (diffMVdiffVal >= -3) {
                                                        return 8;
                                                    }
                                                }
                                                else if (RsDiff >= 2) {
                                                    return 8;
                                                }
                                            }
                                            else if (refDCval >= 73) {
                                                if (RsDiff < 3) {
                                                    return 4;
                                                }
                                                else if (RsDiff >= 3) {
                                                    if (Cs < 2123) {
                                                        return 8;
                                                    }
                                                    else if (Cs >= 2123) {
                                                        return 1;
                                                    }
                                                }
                                            }
                                        }
                                        else if (diffMVdiffVal >= 10) {
                                            if (RsDiff < 12) {
                                                return 4;
                                            }
                                            else if (RsDiff >= 12) {
                                                return 8;
                                            }
                                        }
                                    }
                                }
                                else if (SC >= 4706) {
                                    if (negBalance < 34) {
                                        return 8;
                                    }
                                    else if (negBalance >= 34) {
                                        if (ssDCval < 110) {
                                            return 8;
                                        }
                                        else if (ssDCval >= 110) {
                                            if (diffRsCsdiff < 6) {
                                                if (RsDiff < 6) {
                                                    return 2;
                                                }
                                                else if (RsDiff >= 6) {
                                                    return 8;
                                                }
                                            }
                                            else if (diffRsCsdiff >= 6) {
                                                return 2;
                                            }
                                        }
                                    }
                                }
                            }
                        }
                        else if (refDCval >= 121) {
                            if (Rs < 4020) {
                                if (negBalance < 17) {
                                    if (diffTSC < -3) {
                                        if (AFD < 3) {
                                            return 8;
                                        }
                                        else if (AFD >= 3) {
                                            return 4;
                                        }
                                    }
                                    else if (diffTSC >= -3) {
                                        if (ssDCval < 141) {
                                            if (MVDiff < 6) {
                                                if (refDCval < 127) {
                                                    return 1;
                                                }
                                                else if (refDCval >= 127) {
                                                    if (RsDiff < 5) {
                                                        if (Rs < 1312) {
                                                            if (diffMVdiffVal < -4) {
                                                                return 1;
                                                            }
                                                            else if (diffMVdiffVal >= -4) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (Rs >= 1312) {
                                                            return 1;
                                                        }
                                                    }
                                                    else if (RsDiff >= 5) {
                                                        return 1;
                                                    }
                                                }
                                            }
                                            else if (MVDiff >= 6) {
                                                return 1;
                                            }
                                        }
                                        else if (ssDCval >= 141) {
                                            return 8;
                                        }
                                    }
                                }
                                else if (negBalance >= 17) {
                                    if (MVDiff < 25) {
                                        if (diffTSC < -1) {
                                            if (Cs < 1820) {
                                                return 4;
                                            }
                                            else if (Cs >= 1820) {
                                                return 8;
                                            }
                                        }
                                        else if (diffTSC >= -1) {
                                            return 8;
                                        }
                                    }
                                    else if (MVDiff >= 25) {
                                        if (AFD < 11) {
                                            return 4;
                                        }
                                        else if (AFD >= 11) {
                                            return 1;
                                        }
                                    }
                                }
                            }
                            else if (Rs >= 4020) {
                                if (ssDCval < 127) {
                                    if (CsDiff < 9) {
                                        if (RsCsDiff < 38) {
                                            return 8;
                                        }
                                        else if (RsCsDiff >= 38) {
                                            return 4;
                                        }
                                    }
                                    else if (CsDiff >= 9) {
                                        return 8;
                                    }
                                }
                                else if (ssDCval >= 127) {
                                    return 2;
                                }
                            }
                        }
                    }
                }
            }
            else if (negBalance >= 36) {
                if (Cs < 1676) {
                    if (AFD < 7) {
                        if (mu_mv_mag_sq < 107) {
                            if (negBalance < 38) {
                                if (ssDCval < 118) {
                                    if (posBalance < 44) {
                                        if (RsDiff < 3) {
                                            return 4;
                                        }
                                        else if (RsDiff >= 3) {
                                            return 1;
                                        }
                                    }
                                    else if (posBalance >= 44) {
                                        return 8;
                                    }
                                }
                                else if (ssDCval >= 118) {
                                    if (SC < 1343) {
                                        return 2;
                                    }
                                    else if (SC >= 1343) {
                                        return 4;
                                    }
                                }
                            }
                            else if (negBalance >= 38) {
                                if (Rs < 2399) {
                                    if (MVDiff < 1) {
                                        if (ssDCval < 75) {
                                            return 2;
                                        }
                                        else if (ssDCval >= 75) {
                                            return 4;
                                        }
                                    }
                                    else if (MVDiff >= 1) {
                                        if (Rs < 319) {
                                            if (Cs < 375) {
                                                if (refDCval < 84) {
                                                    if (diffMVdiffVal < -13) {
                                                        return 2;
                                                    }
                                                    else if (diffMVdiffVal >= -13) {
                                                        return 4;
                                                    }
                                                }
                                                else if (refDCval >= 84) {
                                                    return 8;
                                                }
                                            }
                                            else if (Cs >= 375) {
                                                return 2;
                                            }
                                        }
                                        else if (Rs >= 319) {
                                            if (gchDC < 2) {
                                                if (tcor < 94) {
                                                    if (diffRsCsdiff < -1) {
                                                        return 8;
                                                    }
                                                    else if (diffRsCsdiff >= -1) {
                                                        if (MVDiff < 23) {
                                                            if (SC < 1412) {
                                                                return 4;
                                                            }
                                                            else if (SC >= 1412) {
                                                                if (SC < 3350) {
                                                                    if (Rs < 951) {
                                                                        if (AFD < 5) {
                                                                            return 8;
                                                                        }
                                                                        else if (AFD >= 5) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                    else if (Rs >= 951) {
                                                                        return 8;
                                                                    }
                                                                }
                                                                else if (SC >= 3350) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                        else if (MVDiff >= 23) {
                                                            return 1;
                                                        }
                                                    }
                                                }
                                                else if (tcor >= 94) {
                                                    if (MVDiff < 6) {
                                                        return 8;
                                                    }
                                                    else if (MVDiff >= 6) {
                                                        if (RsDiff < 2) {
                                                            if (Rs < 783) {
                                                                return 4;
                                                            }
                                                            else if (Rs >= 783) {
                                                                return 1;
                                                            }
                                                        }
                                                        else if (RsDiff >= 2) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (gchDC >= 2) {
                                                return 4;
                                            }
                                        }
                                    }
                                }
                                else if (Rs >= 2399) {
                                    return 1;
                                }
                            }
                        }
                        else if (mu_mv_mag_sq >= 107) {
                            if (SC < 3085) {
                                if (TSC < 6) {
                                    if (refDCval < 95) {
                                        return 8;
                                    }
                                    else if (refDCval >= 95) {
                                        if (Rs < 609) {
                                            return 8;
                                        }
                                        else if (Rs >= 609) {
                                            return 4;
                                        }
                                    }
                                }
                                else if (TSC >= 6) {
                                    if (mcTcor < 94) {
                                        if (negBalance < 46) {
                                            return 8;
                                        }
                                        else if (negBalance >= 46) {
                                            return 2;
                                        }
                                    }
                                    else if (mcTcor >= 94) {
                                        return 2;
                                    }
                                }
                            }
                            else if (SC >= 3085) {
                                if (mcTcor < 97) {
                                    if (ssDCval < 75) {
                                        if (RsCsDiff < 22) {
                                            if (posBalance < 58) {
                                                return 8;
                                            }
                                            else if (posBalance >= 58) {
                                                return 2;
                                            }
                                        }
                                        else if (RsCsDiff >= 22) {
                                            if (Rs < 1923) {
                                                if (negBalance < 48) {
                                                    return 8;
                                                }
                                                else if (negBalance >= 48) {
                                                    if (Cs < 1557) {
                                                        return 2;
                                                    }
                                                    else if (Cs >= 1557) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                            else if (Rs >= 1923) {
                                                if (SC < 3623) {
                                                    return 2;
                                                }
                                                else if (SC >= 3623) {
                                                    return 8;
                                                }
                                            }
                                        }
                                    }
                                    else if (ssDCval >= 75) {
                                        if (AFD < 6) {
                                            if (tcor < 95) {
                                                return 8;
                                            }
                                            else if (tcor >= 95) {
                                                if (Cs < 1570) {
                                                    return 8;
                                                }
                                                else if (Cs >= 1570) {
                                                    if (SC < 3355) {
                                                        return 8;
                                                    }
                                                    else if (SC >= 3355) {
                                                        if (Cs < 1605) {
                                                            return 4;
                                                        }
                                                        else if (Cs >= 1605) {
                                                            if (Rs < 1839) {
                                                                return 8;
                                                            }
                                                            else if (Rs >= 1839) {
                                                                if (Rs < 1869) {
                                                                    return 4;
                                                                }
                                                                else if (Rs >= 1869) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (AFD >= 6) {
                                            if (RsDiff < 5) {
                                                if (SC < 3305) {
                                                    if (diffAFD < -1) {
                                                        return 2;
                                                    }
                                                    else if (diffAFD >= -1) {
                                                        return 8;
                                                    }
                                                }
                                                else if (SC >= 3305) {
                                                    if (refDCval < 75) {
                                                        return 2;
                                                    }
                                                    else if (refDCval >= 75) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                            else if (RsDiff >= 5) {
                                                return 8;
                                            }
                                        }
                                    }
                                }
                                else if (mcTcor >= 97) {
                                    if (ssDCval < 81) {
                                        if (SC < 3110) {
                                            if (Rs < 1834) {
                                                return 4;
                                            }
                                            else if (Rs >= 1834) {
                                                return 8;
                                            }
                                        }
                                        else if (SC >= 3110) {
                                            return 4;
                                        }
                                    }
                                    else if (ssDCval >= 81) {
                                        return 8;
                                    }
                                }
                            }
                        }
                    }
                    else if (AFD >= 7) {
                        if (RsDiff < 5) {
                            if (RsCsDiff < 23) {
                                if (Rs < 10) {
                                    return 1;
                                }
                                else if (Rs >= 10) {
                                    return 2;
                                }
                            }
                            else if (RsCsDiff >= 23) {
                                if (refDCval < 86) {
                                    if (MVDiff < 3) {
                                        if (posBalance < 59) {
                                            if (Rs < 1839) {
                                                if (mu_mv_mag_sq < 112) {
                                                    return 2;
                                                }
                                                else if (mu_mv_mag_sq >= 112) {
                                                    if (tcor < 92) {
                                                        return 8;
                                                    }
                                                    else if (tcor >= 92) {
                                                        return 2;
                                                    }
                                                }
                                            }
                                            else if (Rs >= 1839) {
                                                if (Rs < 1886) {
                                                    return 4;
                                                }
                                                else if (Rs >= 1886) {
                                                    return 2;
                                                }
                                            }
                                        }
                                        else if (posBalance >= 59) {
                                            return 2;
                                        }
                                    }
                                    else if (MVDiff >= 3) {
                                        if (AFD < 9) {
                                            return 8;
                                        }
                                        else if (AFD >= 9) {
                                            if (Rs < 707) {
                                                return 1;
                                            }
                                            else if (Rs >= 707) {
                                                return 4;
                                            }
                                        }
                                    }
                                }
                                else if (refDCval >= 86) {
                                    if (diffRsCsdiff < -5) {
                                        return 8;
                                    }
                                    else if (diffRsCsdiff >= -5) {
                                        if (mu_mv_mag_sq < 104) {
                                            if (negBalance < 47) {
                                                return 8;
                                            }
                                            else if (negBalance >= 47) {
                                                return 4;
                                            }
                                        }
                                        else if (mu_mv_mag_sq >= 104) {
                                            return 8;
                                        }
                                    }
                                }
                            }
                        }
                        else if (RsDiff >= 5) {
                            if (tcor < 38) {
                                if (MVDiff < 778) {
                                    if (SC < 780) {
                                        return 2;
                                    }
                                    else if (SC >= 780) {
                                        return 1;
                                    }
                                }
                                else if (MVDiff >= 778) {
                                    return 8;
                                }
                            }
                            else if (tcor >= 38) {
                                if (RsCsDiff < 135) {
                                    if (tcor < 95) {
                                        if (refDCval < 89) {
                                            if (mu_mv_mag_sq < 112) {
                                                if (AFD < 10) {
                                                    return 2;
                                                }
                                                else if (AFD >= 10) {
                                                    return 8;
                                                }
                                            }
                                            else if (mu_mv_mag_sq >= 112) {
                                                return 8;
                                            }
                                        }
                                        else if (refDCval >= 89) {
                                            if (AFD < 15) {
                                                if (TSC < 9) {
                                                    if (Rs < 2390) {
                                                        return 8;
                                                    }
                                                    else if (Rs >= 2390) {
                                                        if (Cs < 1641) {
                                                            return 8;
                                                        }
                                                        else if (Cs >= 1641) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                                else if (TSC >= 9) {
                                                    if (refDCval < 112) {
                                                        return 2;
                                                    }
                                                    else if (refDCval >= 112) {
                                                        if (posBalance < 59) {
                                                            return 8;
                                                        }
                                                        else if (posBalance >= 59) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (AFD >= 15) {
                                                if (mcTcor < 99) {
                                                    return 2;
                                                }
                                                else if (mcTcor >= 99) {
                                                    return 1;
                                                }
                                            }
                                        }
                                    }
                                    else if (tcor >= 95) {
                                        if (Rs < 2401) {
                                            return 4;
                                        }
                                        else if (Rs >= 2401) {
                                            return 1;
                                        }
                                    }
                                }
                                else if (RsCsDiff >= 135) {
                                    if (Rs < 2376) {
                                        if (refDCval < 137) {
                                            return 1;
                                        }
                                        else if (refDCval >= 137) {
                                            return 2;
                                        }
                                    }
                                    else if (Rs >= 2376) {
                                        if (AFD < 29) {
                                            return 1;
                                        }
                                        else if (AFD >= 29) {
                                            if (mcTcor < 93) {
                                                return 2;
                                            }
                                            else if (mcTcor >= 93) {
                                                if (Cs < 1509) {
                                                    return 2;
                                                }
                                                else if (Cs >= 1509) {
                                                    return 4;
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                        }
                    }
                }
                else if (Cs >= 1676) {
                    if (Rs < 7155) {
                        if (Cs < 3590) {
                            if (Rs < 5295) {
                                if (TSCindex < 8) {
                                    if (refDCval < 89) {
                                        if (mu_mv_mag_sq < 113) {
                                            if (SC < 4830) {
                                                if (MVDiff < 2) {
                                                    if (diffMVdiffVal < -1) {
                                                        if (posBalance < 57) {
                                                            return 4;
                                                        }
                                                        else if (posBalance >= 57) {
                                                            return 2;
                                                        }
                                                    }
                                                    else if (diffMVdiffVal >= -1) {
                                                        return 4;
                                                    }
                                                }
                                                else if (MVDiff >= 2) {
                                                    if (mu_mv_mag_sq < 92) {
                                                        return 4;
                                                    }
                                                    else if (mu_mv_mag_sq >= 92) {
                                                        if (SC < 3592) {
                                                            return 2;
                                                        }
                                                        else if (SC >= 3592) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (SC >= 4830) {
                                                if (negBalance < 44) {
                                                    return 8;
                                                }
                                                else if (negBalance >= 44) {
                                                    if (ssDCval < 54) {
                                                        if (Rs < 3457) {
                                                            return 8;
                                                        }
                                                        else if (Rs >= 3457) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (ssDCval >= 54) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                        }
                                        else if (mu_mv_mag_sq >= 113) {
                                            if (mcTcor < 95) {
                                                return 8;
                                            }
                                            else if (mcTcor >= 95) {
                                                if (SC < 3810) {
                                                    if (SC < 3588) {
                                                        if (mcTcor < 98) {
                                                            return 8;
                                                        }
                                                        else if (mcTcor >= 98) {
                                                            if (Rs < 1686) {
                                                                return 8;
                                                            }
                                                            else if (Rs >= 1686) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                    else if (SC >= 3588) {
                                                        if (Rs < 1845) {
                                                            if (Cs < 1823) {
                                                                return 4;
                                                            }
                                                            else if (Cs >= 1823) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (Rs >= 1845) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                                else if (SC >= 3810) {
                                                    if (MVDiff < 1) {
                                                        if (negBalance < 43) {
                                                            return 4;
                                                        }
                                                        else if (negBalance >= 43) {
                                                            if (Rs < 2016) {
                                                                return 8;
                                                            }
                                                            else if (Rs >= 2016) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                    else if (MVDiff >= 1) {
                                                        if (AFD < 6) {
                                                            return 8;
                                                        }
                                                        else if (AFD >= 6) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (refDCval >= 89) {
                                        if (Rs < 1773) {
                                            if (Cs < 2113) {
                                                if (Rs < 381) {
                                                    return 4;
                                                }
                                                else if (Rs >= 381) {
                                                    return 8;
                                                }
                                            }
                                            else if (Cs >= 2113) {
                                                return 4;
                                            }
                                        }
                                        else if (Rs >= 1773) {
                                            if (negBalance < 48) {
                                                if (mu_mv_mag_sq < 113) {
                                                    if (Cs < 2285) {
                                                        if (Rs < 2590) {
                                                            return 4;
                                                        }
                                                        else if (Rs >= 2590) {
                                                            if (Rs < 2921) {
                                                                return 1;
                                                            }
                                                            else if (Rs >= 2921) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                    else if (Cs >= 2285) {
                                                        return 8;
                                                    }
                                                }
                                                else if (mu_mv_mag_sq >= 113) {
                                                    if (mu_mv_mag_sq < 123) {
                                                        if (ssDCval < 114) {
                                                            if (Rs < 2832) {
                                                                if (Cs < 2987) {
                                                                    if (RsDiff < 6) {
                                                                        return 8;
                                                                    }
                                                                    else if (RsDiff >= 6) {
                                                                        if (Rs < 2072) {
                                                                            return 8;
                                                                        }
                                                                        else if (Rs >= 2072) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                }
                                                                else if (Cs >= 2987) {
                                                                    if (RsCsDiff < 37) {
                                                                        if (SC < 5716) {
                                                                            return 8;
                                                                        }
                                                                        else if (SC >= 5716) {
                                                                            if (negBalance < 46) {
                                                                                return 8;
                                                                            }
                                                                            else if (negBalance >= 46) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (RsCsDiff >= 37) {
                                                                        return 8;
                                                                    }
                                                                }
                                                            }
                                                            else if (Rs >= 2832) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (ssDCval >= 114) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (mu_mv_mag_sq >= 123) {
                                                        if (RsDiff < 2) {
                                                            return 4;
                                                        }
                                                        else if (RsDiff >= 2) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (negBalance >= 48) {
                                                if (mu_mv_mag_sq < 121) {
                                                    if (posBalance < 58) {
                                                        if (mu_mv_mag_sq < 118) {
                                                            return 4;
                                                        }
                                                        else if (mu_mv_mag_sq >= 118) {
                                                            if (posBalance < 57) {
                                                                if (ssDCval < 112) {
                                                                    return 8;
                                                                }
                                                                else if (ssDCval >= 112) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (posBalance >= 57) {
                                                                if (ssDCval < 113) {
                                                                    if (Cs < 3190) {
                                                                        return 4;
                                                                    }
                                                                    else if (Cs >= 3190) {
                                                                        return 8;
                                                                    }
                                                                }
                                                                else if (ssDCval >= 113) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (posBalance >= 58) {
                                                        return 8;
                                                    }
                                                }
                                                else if (mu_mv_mag_sq >= 121) {
                                                    if (Rs < 2718) {
                                                        return 8;
                                                    }
                                                    else if (Rs >= 2718) {
                                                        if (refDCval < 112) {
                                                            return 8;
                                                        }
                                                        else if (refDCval >= 112) {
                                                            if (Rs < 2975) {
                                                                if (diffTSC < 1) {
                                                                    if (diffRsCsdiff < -8) {
                                                                        return 4;
                                                                    }
                                                                    else if (diffRsCsdiff >= -8) {
                                                                        if (gchDC < 1) {
                                                                            if (Rs < 2947) {
                                                                                if (refDCval < 113) {
                                                                                    if (RsCsDiff < 37) {
                                                                                        return 8;
                                                                                    }
                                                                                    else if (RsCsDiff >= 37) {
                                                                                        if (CsDiff < 5) {
                                                                                            return 4;
                                                                                        }
                                                                                        else if (CsDiff >= 5) {
                                                                                            return 8;
                                                                                        }
                                                                                    }
                                                                                }
                                                                                else if (refDCval >= 113) {
                                                                                    if (ssDCval < 114) {
                                                                                        return 4;
                                                                                    }
                                                                                    else if (ssDCval >= 114) {
                                                                                        if (RsCsDiff < 37) {
                                                                                            return 8;
                                                                                        }
                                                                                        else if (RsCsDiff >= 37) {
                                                                                            if (RsDiff < 5) {
                                                                                                return 4;
                                                                                            }
                                                                                            else if (RsDiff >= 5) {
                                                                                                return 8;
                                                                                            }
                                                                                        }
                                                                                    }
                                                                                }
                                                                            }
                                                                            else if (Rs >= 2947) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                        else if (gchDC >= 1) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                }
                                                                else if (diffTSC >= 1) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (Rs >= 2975) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (TSCindex >= 8) {
                                    if (negBalance < 55) {
                                        if (MVDiff < 2) {
                                            if (SC < 6394) {
                                                if (SC < 4406) {
                                                    if (mu_mv_mag_sq < 99) {
                                                        if (refDCval < 95) {
                                                            return 2;
                                                        }
                                                        else if (refDCval >= 95) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (mu_mv_mag_sq >= 99) {
                                                        if (tcor < 86) {
                                                            return 4;
                                                        }
                                                        else if (tcor >= 86) {
                                                            if (RsDiff < 5) {
                                                                return 4;
                                                            }
                                                            else if (RsDiff >= 5) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (SC >= 4406) {
                                                    if (Cs < 2353) {
                                                        if (Cs < 2331) {
                                                            if (posBalance < 60) {
                                                                if (diffMVdiffVal < 1) {
                                                                    return 8;
                                                                }
                                                                else if (diffMVdiffVal >= 1) {
                                                                    if (Rs < 2505) {
                                                                        return 4;
                                                                    }
                                                                    else if (Rs >= 2505) {
                                                                        return 8;
                                                                    }
                                                                }
                                                            }
                                                            else if (posBalance >= 60) {
                                                                if (RsCsDiff < 56) {
                                                                    return 2;
                                                                }
                                                                else if (RsCsDiff >= 56) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                        else if (Cs >= 2331) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (Cs >= 2353) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                            else if (SC >= 6394) {
                                                if (Rs < 4009) {
                                                    if (negBalance < 42) {
                                                        return 8;
                                                    }
                                                    else if (negBalance >= 42) {
                                                        if (ssDCval < 94) {
                                                            return 2;
                                                        }
                                                        else if (ssDCval >= 94) {
                                                            if (Cs < 3035) {
                                                                return 4;
                                                            }
                                                            else if (Cs >= 3035) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (Rs >= 4009) {
                                                    if (refDCval < 122) {
                                                        if (diffRsCsdiff < -31) {
                                                            return 2;
                                                        }
                                                        else if (diffRsCsdiff >= -31) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (refDCval >= 122) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                        }
                                        else if (MVDiff >= 2) {
                                            if (ssDCval < 87) {
                                                if (Cs < 2371) {
                                                    if (mcTcor < 95) {
                                                        if (diffTSC < -3) {
                                                            return 4;
                                                        }
                                                        else if (diffTSC >= -3) {
                                                            if (RsDiff < 9) {
                                                                return 8;
                                                            }
                                                            else if (RsDiff >= 9) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                    else if (mcTcor >= 95) {
                                                        return 4;
                                                    }
                                                }
                                                else if (Cs >= 2371) {
                                                    return 4;
                                                }
                                            }
                                            else if (ssDCval >= 87) {
                                                if (ssDCval < 96) {
                                                    if (refDCval < 93) {
                                                        if (mu_mv_mag_sq < 88) {
                                                            if (Rs < 2351) {
                                                                return 2;
                                                            }
                                                            else if (Rs >= 2351) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (mu_mv_mag_sq >= 88) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (refDCval >= 93) {
                                                        if (diffTSC < 2) {
                                                            return 2;
                                                        }
                                                        else if (diffTSC >= 2) {
                                                            return 1;
                                                        }
                                                    }
                                                }
                                                else if (ssDCval >= 96) {
                                                    if (negBalance < 48) {
                                                        if (Rs < 4660) {
                                                            if (mu_mv_mag_sq < 95) {
                                                                if (ssDCval < 109) {
                                                                    if (Cs < 2775) {
                                                                        if (Rs < 3787) {
                                                                            return 8;
                                                                        }
                                                                        else if (Rs >= 3787) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                    else if (Cs >= 2775) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (ssDCval >= 109) {
                                                                    if (AFD < 14) {
                                                                        if (posBalance < 48) {
                                                                            return 2;
                                                                        }
                                                                        else if (posBalance >= 48) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                    else if (AFD >= 14) {
                                                                        return 8;
                                                                    }
                                                                }
                                                            }
                                                            else if (mu_mv_mag_sq >= 95) {
                                                                if (RsCsDiff < 211) {
                                                                    if (diffTSC < -2) {
                                                                        if (diffRsCsdiff < -129) {
                                                                            return 8;
                                                                        }
                                                                        else if (diffRsCsdiff >= -129) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                    else if (diffTSC >= -2) {
                                                                        return 8;
                                                                    }
                                                                }
                                                                else if (RsCsDiff >= 211) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                        else if (Rs >= 4660) {
                                                            if (Rs < 5039) {
                                                                return 2;
                                                            }
                                                            else if (Rs >= 5039) {
                                                                if (refDCval < 112) {
                                                                    if (TSC < 9) {
                                                                        return 8;
                                                                    }
                                                                    else if (TSC >= 9) {
                                                                        return 2;
                                                                    }
                                                                }
                                                                else if (refDCval >= 112) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (negBalance >= 48) {
                                                        if (posBalance < 57) {
                                                            if (mcTcor < 91) {
                                                                return 8;
                                                            }
                                                            else if (mcTcor >= 91) {
                                                                if (SC < 7188) {
                                                                    return 4;
                                                                }
                                                                else if (SC >= 7188) {
                                                                    if (Rs < 4107) {
                                                                        return 8;
                                                                    }
                                                                    else if (Rs >= 4107) {
                                                                        return 4;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (posBalance >= 57) {
                                                            if (mcTcor < 86) {
                                                                if (refDCval < 109) {
                                                                    return 4;
                                                                }
                                                                else if (refDCval >= 109) {
                                                                    if (refDCval < 119) {
                                                                        return 8;
                                                                    }
                                                                    else if (refDCval >= 119) {
                                                                        if (Cs < 2611) {
                                                                            return 4;
                                                                        }
                                                                        else if (Cs >= 2611) {
                                                                            if (Rs < 3489) {
                                                                                return 8;
                                                                            }
                                                                            else if (Rs >= 3489) {
                                                                                return 1;
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                            else if (mcTcor >= 86) {
                                                                if (tcor < 82) {
                                                                    if (Rs < 2873) {
                                                                        return 8;
                                                                    }
                                                                    else if (Rs >= 2873) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (tcor >= 82) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (negBalance >= 55) {
                                        if (Rs < 4432) {
                                            if (MVDiff < 8) {
                                                if (ssDCval < 85) {
                                                    return 4;
                                                }
                                                else if (ssDCval >= 85) {
                                                    return 8;
                                                }
                                            }
                                            else if (MVDiff >= 8) {
                                                return 2;
                                            }
                                        }
                                        else if (Rs >= 4432) {
                                            return 4;
                                        }
                                    }
                                }
                            }
                            else if (Rs >= 5295) {
                                return 4;
                            }
                        }
                        else if (Cs >= 3590) {
                            if (AFD < 13) {
                                if (ssDCval < 82) {
                                    if (CsDiff < 8) {
                                        if (TSC < 7) {
                                            if (mu_mv_mag_sq < 99) {
                                                if (Cs < 3731) {
                                                    if (Rs < 3737) {
                                                        return 4;
                                                    }
                                                    else if (Rs >= 3737) {
                                                        return 8;
                                                    }
                                                }
                                                else if (Cs >= 3731) {
                                                    return 8;
                                                }
                                            }
                                            else if (mu_mv_mag_sq >= 99) {
                                                if (Cs < 4633) {
                                                    return 8;
                                                }
                                                else if (Cs >= 4633) {
                                                    if (Rs < 4980) {
                                                        if (Rs < 4591) {
                                                            return 4;
                                                        }
                                                        else if (Rs >= 4591) {
                                                            if (mu_mv_mag_sq < 128) {
                                                                if (Cs < 4899) {
                                                                    return 8;
                                                                }
                                                                else if (Cs >= 4899) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (mu_mv_mag_sq >= 128) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                    else if (Rs >= 4980) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                        }
                                        else if (TSC >= 7) {
                                            return 4;
                                        }
                                    }
                                    else if (CsDiff >= 8) {
                                        if (TSC < 8) {
                                            if (Cs < 4856) {
                                                if (diffRsCsdiff < -9) {
                                                    return 4;
                                                }
                                                else if (diffRsCsdiff >= -9) {
                                                    return 8;
                                                }
                                            }
                                            else if (Cs >= 4856) {
                                                if (ssDCval < 80) {
                                                    if (Rs < 4709) {
                                                        return 4;
                                                    }
                                                    else if (Rs >= 4709) {
                                                        if (diffAFD < 1) {
                                                            return 8;
                                                        }
                                                        else if (diffAFD >= 1) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                                else if (ssDCval >= 80) {
                                                    if (Cs < 5157) {
                                                        if (SC < 10127) {
                                                            if (diffRsCsdiff < -61) {
                                                                return 8;
                                                            }
                                                            else if (diffRsCsdiff >= -61) {
                                                                if (CsDiff < 9) {
                                                                    if (Cs < 4916) {
                                                                        if (Rs < 4773) {
                                                                            if (Rs < 4687) {
                                                                                return 4;
                                                                            }
                                                                            else if (Rs >= 4687) {
                                                                                return 8;
                                                                            }
                                                                        }
                                                                        else if (Rs >= 4773) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                    else if (Cs >= 4916) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (CsDiff >= 9) {
                                                                    if (Cs < 5045) {
                                                                        if (Rs < 4920) {
                                                                            if (negBalance < 48) {
                                                                                return 4;
                                                                            }
                                                                            else if (negBalance >= 48) {
                                                                                return 8;
                                                                            }
                                                                        }
                                                                        else if (Rs >= 4920) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                    else if (Cs >= 5045) {
                                                                        return 8;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (SC >= 10127) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (Cs >= 5157) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                        }
                                        else if (TSC >= 8) {
                                            return 8;
                                        }
                                    }
                                }
                                else if (ssDCval >= 82) {
                                    if (AFD < 9) {
                                        return 8;
                                    }
                                    else if (AFD >= 9) {
                                        if (refDCval < 106) {
                                            return 8;
                                        }
                                        else if (refDCval >= 106) {
                                            if (refDCval < 117) {
                                                if (posBalance < 54) {
                                                    return 8;
                                                }
                                                else if (posBalance >= 54) {
                                                    if (MVDiff < 1) {
                                                        return 4;
                                                    }
                                                    else if (MVDiff >= 1) {
                                                        if (Cs < 4510) {
                                                            return 2;
                                                        }
                                                        else if (Cs >= 4510) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (refDCval >= 117) {
                                                return 8;
                                            }
                                        }
                                    }
                                }
                            }
                            else if (AFD >= 13) {
                                if (ssDCval < 137) {
                                    return 8;
                                }
                                else if (ssDCval >= 137) {
                                    if (AFD < 14) {
                                        if (CsDiff < 14) {
                                            return 1;
                                        }
                                        else if (CsDiff >= 14) {
                                            return 8;
                                        }
                                    }
                                    else if (AFD >= 14) {
                                        if (SC < 10472) {
                                            return 1;
                                        }
                                        else if (SC >= 10472) {
                                            return 8;
                                        }
                                    }
                                }
                            }
                        }
                    }
                    else if (Rs >= 7155) {
                        if (RsDiff < 15) {
                            return 4;
                        }
                        else if (RsDiff >= 15) {
                            if (negBalance < 60) {
                                return 8;
                            }
                            else if (negBalance >= 60) {
                                if (TSC < 19) {
                                    return 2;
                                }
                                else if (TSC >= 19) {
                                    if (RsCsDiff < 950) {
                                        return 4;
                                    }
                                    else if (RsCsDiff >= 950) {
                                        if (RsDiff < 24) {
                                            if (Cs < 7581) {
                                                return 4;
                                            }
                                            else if (Cs >= 7581) {
                                                if (diffRsCsdiff < -1) {
                                                    if (SC < 16065) {
                                                        return 2;
                                                    }
                                                    else if (SC >= 16065) {
                                                        return 4;
                                                    }
                                                }
                                                else if (diffRsCsdiff >= -1) {
                                                    return 2;
                                                }
                                            }
                                        }
                                        else if (RsDiff >= 24) {
                                            if (Cs < 8173) {
                                                return 4;
                                            }
                                            else if (Cs >= 8173) {
                                                if (MVDiff < 1) {
                                                    return 2;
                                                }
                                                else if (MVDiff >= 1) {
                                                    return 4;
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                        }
                    }
                }
            }
        }
        return 4;
    }
    mfxU16 AGOPSelect9(mfxU32 Rs, mfxU32 CsDiff, mfxU32 TSC, mfxI32 diffAFD, mfxU32 Scindex, mfxU32 RsDiff, mfxI32 diffMVdiffVal, mfxU32 RsCsDiff, mfxU32 TSCindex, mfxI16 tcor, mfxU32 refDCval, mfxU32 MVDiff, mfxU32 AFD, mfxI32 diffTSC, mfxI32 diffRsCsdiff, mfxU32 posBalance, mfxU32 SC, mfxU32 mu_mv_mag_sq, mfxU32 negBalance, mfxI16 mcTcor, mfxU32 gchDC, mfxU32 Cs, mfxU32 ssDCval) {
        if (MVDiff < 6) {
            if (negBalance < 36) {
                if (Rs < 695) {
                    if (ssDCval < 161) {
                        if (refDCval < 92) {
                            if (refDCval < 91) {
                                if (Rs < 539) {
                                    if (diffMVdiffVal < -10) {
                                        if (SC < 430) {
                                            return 1;
                                        }
                                        else if (SC >= 430) {
                                            if (Cs < 328) {
                                                return 8;
                                            }
                                            else if (Cs >= 328) {
                                                return 2;
                                            }
                                        }
                                    }
                                    else if (diffMVdiffVal >= -10) {
                                        if (tcor < 100) {
                                            if (posBalance < 64) {
                                                if (mu_mv_mag_sq < 122) {
                                                    return 4;
                                                }
                                                else if (mu_mv_mag_sq >= 122) {
                                                    if (posBalance < 39) {
                                                        if (Rs < 248) {
                                                            if (Cs < 312) {
                                                                if (Cs < 308) {
                                                                    return 8;
                                                                }
                                                                else if (Cs >= 308) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (Cs >= 312) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (Rs >= 248) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (posBalance >= 39) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                            else if (posBalance >= 64) {
                                                if (AFD < 3) {
                                                    return 4;
                                                }
                                                else if (AFD >= 3) {
                                                    return 8;
                                                }
                                            }
                                        }
                                        else if (tcor >= 100) {
                                            return 8;
                                        }
                                    }
                                }
                                else if (Rs >= 539) {
                                    if (AFD < 5) {
                                        return 8;
                                    }
                                    else if (AFD >= 5) {
                                        return 4;
                                    }
                                }
                            }
                            else if (refDCval >= 91) {
                                return 4;
                            }
                        }
                        else if (refDCval >= 92) {
                            if (SC < 1228) {
                                if (AFD < 6) {
                                    return 8;
                                }
                                else if (AFD >= 6) {
                                    return 2;
                                }
                            }
                            else if (SC >= 1228) {
                                if (AFD < 4) {
                                    return 8;
                                }
                                else if (AFD >= 4) {
                                    if (mu_mv_mag_sq < 99) {
                                        return 4;
                                    }
                                    else if (mu_mv_mag_sq >= 99) {
                                        return 8;
                                    }
                                }
                            }
                        }
                    }
                    else if (ssDCval >= 161) {
                        if (SC < 879) {
                            return 4;
                        }
                        else if (SC >= 879) {
                            if (ssDCval < 164) {
                                return 4;
                            }
                            else if (ssDCval >= 164) {
                                if (Rs < 519) {
                                    return 2;
                                }
                                else if (Rs >= 519) {
                                    return 4;
                                }
                            }
                        }
                    }
                }
                else if (Rs >= 695) {
                    if (posBalance < 2) {
                        if (Cs < 1810) {
                            if (mu_mv_mag_sq < 124) {
                                if (negBalance < 2) {
                                    return 8;
                                }
                                else if (negBalance >= 2) {
                                    return 1;
                                }
                            }
                            else if (mu_mv_mag_sq >= 124) {
                                return 1;
                            }
                        }
                        else if (Cs >= 1810) {
                            if (Rs < 4675) {
                                if (SC < 5245) {
                                    if (Rs < 2402) {
                                        return 1;
                                    }
                                    else if (Rs >= 2402) {
                                        return 4;
                                    }
                                }
                                else if (SC >= 5245) {
                                    if (SC < 7755) {
                                        if (SC < 7709) {
                                            return 8;
                                        }
                                        else if (SC >= 7709) {
                                            return 1;
                                        }
                                    }
                                    else if (SC >= 7755) {
                                        return 8;
                                    }
                                }
                            }
                            else if (Rs >= 4675) {
                                if (Cs < 5084) {
                                    return 1;
                                }
                                else if (Cs >= 5084) {
                                    return 8;
                                }
                            }
                        }
                    }
                    else if (posBalance >= 2) {
                        if (mu_mv_mag_sq < 92) {
                            if (RsCsDiff < 33) {
                                if (ssDCval < 54) {
                                    if (Cs < 804) {
                                        if (posBalance < 56) {
                                            return 4;
                                        }
                                        else if (posBalance >= 56) {
                                            return 8;
                                        }
                                    }
                                    else if (Cs >= 804) {
                                        if (TSC < 2) {
                                            if (Rs < 972) {
                                                return 8;
                                            }
                                            else if (Rs >= 972) {
                                                return 4;
                                            }
                                        }
                                        else if (TSC >= 2) {
                                            if (negBalance < 34) {
                                                return 1;
                                            }
                                            else if (negBalance >= 34) {
                                                if (Cs < 905) {
                                                    return 1;
                                                }
                                                else if (Cs >= 905) {
                                                    if (gchDC < 1) {
                                                        return 1;
                                                    }
                                                    else if (gchDC >= 1) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (ssDCval >= 54) {
                                    if (SC < 867) {
                                        return 1;
                                    }
                                    else if (SC >= 867) {
                                        if (diffMVdiffVal < -3) {
                                            if (Cs < 326) {
                                                return 2;
                                            }
                                            else if (Cs >= 326) {
                                                if (posBalance < 51) {
                                                    return 8;
                                                }
                                                else if (posBalance >= 51) {
                                                    if (SC < 1851) {
                                                        return 8;
                                                    }
                                                    else if (SC >= 1851) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                        }
                                        else if (diffMVdiffVal >= -3) {
                                            if (AFD < 8) {
                                                if (diffAFD < -1) {
                                                    if (diffRsCsdiff < -7) {
                                                        if (RsCsDiff < 7) {
                                                            return 2;
                                                        }
                                                        else if (RsCsDiff >= 7) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (diffRsCsdiff >= -7) {
                                                        return 8;
                                                    }
                                                }
                                                else if (diffAFD >= -1) {
                                                    if (diffMVdiffVal < 4) {
                                                        return 8;
                                                    }
                                                    else if (diffMVdiffVal >= 4) {
                                                        if (MVDiff < 5) {
                                                            return 1;
                                                        }
                                                        else if (MVDiff >= 5) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (AFD >= 8) {
                                                return 2;
                                            }
                                        }
                                    }
                                }
                            }
                            else if (RsCsDiff >= 33) {
                                if (ssDCval < 79) {
                                    if (refDCval < 64) {
                                        return 4;
                                    }
                                    else if (refDCval >= 64) {
                                        return 1;
                                    }
                                }
                                else if (ssDCval >= 79) {
                                    if (refDCval < 107) {
                                        if (Rs < 1936) {
                                            return 2;
                                        }
                                        else if (Rs >= 1936) {
                                            if (SC < 3758) {
                                                return 4;
                                            }
                                            else if (SC >= 3758) {
                                                return 8;
                                            }
                                        }
                                    }
                                    else if (refDCval >= 107) {
                                        if (negBalance < 35) {
                                            if (Rs < 4438) {
                                                return 4;
                                            }
                                            else if (Rs >= 4438) {
                                                return 2;
                                            }
                                        }
                                        else if (negBalance >= 35) {
                                            if (Cs < 1245) {
                                                return 2;
                                            }
                                            else if (Cs >= 1245) {
                                                return 8;
                                            }
                                        }
                                    }
                                }
                            }
                        }
                        else if (mu_mv_mag_sq >= 92) {
                            if (mu_mv_mag_sq < 121) {
                                if (SC < 3876) {
                                    if (AFD < 1) {
                                        if (Rs < 1111) {
                                            if (Rs < 1023) {
                                                return 8;
                                            }
                                            else if (Rs >= 1023) {
                                                if (ssDCval < 25) {
                                                    return 8;
                                                }
                                                else if (ssDCval >= 25) {
                                                    if (negBalance < 2) {
                                                        return 2;
                                                    }
                                                    else if (negBalance >= 2) {
                                                        if (SC < 2272) {
                                                            return 4;
                                                        }
                                                        else if (SC >= 2272) {
                                                            return 2;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (Rs >= 1111) {
                                            return 8;
                                        }
                                    }
                                    else if (AFD >= 1) {
                                        if (mcTcor < 98) {
                                            if (TSC < 4) {
                                                if (SC < 2653) {
                                                    if (Rs < 1220) {
                                                        if (diffMVdiffVal < -3) {
                                                            if (Cs < 1072) {
                                                                return 1;
                                                            }
                                                            else if (Cs >= 1072) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (diffMVdiffVal >= -3) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (Rs >= 1220) {
                                                        if (diffMVdiffVal < -2) {
                                                            return 1;
                                                        }
                                                        else if (diffMVdiffVal >= -2) {
                                                            if (Rs < 1274) {
                                                                if (RsDiff < 4) {
                                                                    return 1;
                                                                }
                                                                else if (RsDiff >= 4) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (Rs >= 1274) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (SC >= 2653) {
                                                    if (mcTcor < 80) {
                                                        if (SC < 2924) {
                                                            return 4;
                                                        }
                                                        else if (SC >= 2924) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (mcTcor >= 80) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                            else if (TSC >= 4) {
                                                return 8;
                                            }
                                        }
                                        else if (mcTcor >= 98) {
                                            if (negBalance < 30) {
                                                return 8;
                                            }
                                            else if (negBalance >= 30) {
                                                if (ssDCval < 73) {
                                                    return 8;
                                                }
                                                else if (ssDCval >= 73) {
                                                    if (Cs < 1510) {
                                                        return 4;
                                                    }
                                                    else if (Cs >= 1510) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (SC >= 3876) {
                                    if (ssDCval < 115) {
                                        if (ssDCval < 70) {
                                            if (Rs < 2082) {
                                                return 4;
                                            }
                                            else if (Rs >= 2082) {
                                                return 8;
                                            }
                                        }
                                        else if (ssDCval >= 70) {
                                            if (Rs < 3385) {
                                                if (Rs < 2134) {
                                                    if (posBalance < 42) {
                                                        return 1;
                                                    }
                                                    else if (posBalance >= 42) {
                                                        return 4;
                                                    }
                                                }
                                                else if (Rs >= 2134) {
                                                    if (negBalance < 25) {
                                                        return 8;
                                                    }
                                                    else if (negBalance >= 25) {
                                                        if (posBalance < 53) {
                                                            if (posBalance < 48) {
                                                                if (tcor < 94) {
                                                                    return 8;
                                                                }
                                                                else if (tcor >= 94) {
                                                                    if (mcTcor < 98) {
                                                                        if (SC < 4373) {
                                                                            if (SC < 4349) {
                                                                                if (RsDiff < 4) {
                                                                                    if (SC < 4290) {
                                                                                        return 1;
                                                                                    }
                                                                                    else if (SC >= 4290) {
                                                                                        return 8;
                                                                                    }
                                                                                }
                                                                                else if (RsDiff >= 4) {
                                                                                    return 1;
                                                                                }
                                                                            }
                                                                            else if (SC >= 4349) {
                                                                                return 8;
                                                                            }
                                                                        }
                                                                        else if (SC >= 4373) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                    else if (mcTcor >= 98) {
                                                                        if (Rs < 3167) {
                                                                            if (mu_mv_mag_sq < 113) {
                                                                                return 8;
                                                                            }
                                                                            else if (mu_mv_mag_sq >= 113) {
                                                                                if (SC < 5443) {
                                                                                    return 1;
                                                                                }
                                                                                else if (SC >= 5443) {
                                                                                    if (posBalance < 44) {
                                                                                        return 8;
                                                                                    }
                                                                                    else if (posBalance >= 44) {
                                                                                        if (Rs < 3015) {
                                                                                            return 8;
                                                                                        }
                                                                                        else if (Rs >= 3015) {
                                                                                            if (Rs < 3158) {
                                                                                                if (Rs < 3097) {
                                                                                                    return 1;
                                                                                                }
                                                                                                else if (Rs >= 3097) {
                                                                                                    return 8;
                                                                                                }
                                                                                            }
                                                                                            else if (Rs >= 3158) {
                                                                                                return 1;
                                                                                            }
                                                                                        }
                                                                                    }
                                                                                }
                                                                            }
                                                                        }
                                                                        else if (Rs >= 3167) {
                                                                            if (Cs < 3097) {
                                                                                if (RsCsDiff < 22) {
                                                                                    if (Cs < 3084) {
                                                                                        return 8;
                                                                                    }
                                                                                    else if (Cs >= 3084) {
                                                                                        return 1;
                                                                                    }
                                                                                }
                                                                                else if (RsCsDiff >= 22) {
                                                                                    return 1;
                                                                                }
                                                                            }
                                                                            else if (Cs >= 3097) {
                                                                                return 8;
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                            else if (posBalance >= 48) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (posBalance >= 53) {
                                                            if (Rs < 2773) {
                                                                return 4;
                                                            }
                                                            else if (Rs >= 2773) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                            else if (Rs >= 3385) {
                                                if (Rs < 3482) {
                                                    if (Rs < 3471) {
                                                        return 8;
                                                    }
                                                    else if (Rs >= 3471) {
                                                        return 2;
                                                    }
                                                }
                                                else if (Rs >= 3482) {
                                                    return 8;
                                                }
                                            }
                                        }
                                    }
                                    else if (ssDCval >= 115) {
                                        if (negBalance < 9) {
                                            if (Rs < 2496) {
                                                if (RsDiff < 6) {
                                                    if (Rs < 2430) {
                                                        return 4;
                                                    }
                                                    else if (Rs >= 2430) {
                                                        if (Rs < 2461) {
                                                            return 8;
                                                        }
                                                        else if (Rs >= 2461) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                                else if (RsDiff >= 6) {
                                                    return 1;
                                                }
                                            }
                                            else if (Rs >= 2496) {
                                                if (SC < 4814) {
                                                    return 8;
                                                }
                                                else if (SC >= 4814) {
                                                    if (Rs < 2623) {
                                                        return 4;
                                                    }
                                                    else if (Rs >= 2623) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                        }
                                        else if (negBalance >= 9) {
                                            if (tcor < 97) {
                                                if (refDCval < 125) {
                                                    if (RsCsDiff < 22) {
                                                        if (ssDCval < 117) {
                                                            return 8;
                                                        }
                                                        else if (ssDCval >= 117) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (RsCsDiff >= 22) {
                                                        return 8;
                                                    }
                                                }
                                                else if (refDCval >= 125) {
                                                    return 4;
                                                }
                                            }
                                            else if (tcor >= 97) {
                                                return 8;
                                            }
                                        }
                                    }
                                }
                            }
                            else if (mu_mv_mag_sq >= 121) {
                                if (negBalance < 3) {
                                    if (Rs < 3779) {
                                        if (Cs < 4001) {
                                            if (Cs < 3716) {
                                                return 8;
                                            }
                                            else if (Cs >= 3716) {
                                                if (Cs < 3876) {
                                                    if (RsDiff < 1) {
                                                        if (Cs < 3792) {
                                                            return 1;
                                                        }
                                                        else if (Cs >= 3792) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (RsDiff >= 1) {
                                                        return 1;
                                                    }
                                                }
                                                else if (Cs >= 3876) {
                                                    if (Rs < 3774) {
                                                        return 8;
                                                    }
                                                    else if (Rs >= 3774) {
                                                        return 1;
                                                    }
                                                }
                                            }
                                        }
                                        else if (Cs >= 4001) {
                                            return 1;
                                        }
                                    }
                                    else if (Rs >= 3779) {
                                        return 8;
                                    }
                                }
                                else if (negBalance >= 3) {
                                    if (mu_mv_mag_sq < 125) {
                                        if (refDCval < 92) {
                                            if (TSC < 3) {
                                                return 8;
                                            }
                                            else if (TSC >= 3) {
                                                if (refDCval < 51) {
                                                    if (negBalance < 19) {
                                                        return 8;
                                                    }
                                                    else if (negBalance >= 19) {
                                                        if (RsDiff < 4) {
                                                            return 8;
                                                        }
                                                        else if (RsDiff >= 4) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                                else if (refDCval >= 51) {
                                                    if (negBalance < 34) {
                                                        return 8;
                                                    }
                                                    else if (negBalance >= 34) {
                                                        if (Rs < 2477) {
                                                            return 4;
                                                        }
                                                        else if (Rs >= 2477) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (refDCval >= 92) {
                                            if (CsDiff < 3) {
                                                if (diffRsCsdiff < 4) {
                                                    return 8;
                                                }
                                                else if (diffRsCsdiff >= 4) {
                                                    if (ssDCval < 114) {
                                                        return 8;
                                                    }
                                                    else if (ssDCval >= 114) {
                                                        if (Rs < 2947) {
                                                            return 8;
                                                        }
                                                        else if (Rs >= 2947) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (CsDiff >= 3) {
                                                if (Rs < 3322) {
                                                    if (posBalance < 46) {
                                                        if (negBalance < 26) {
                                                            if (ssDCval < 108) {
                                                                return 8;
                                                            }
                                                            else if (ssDCval >= 108) {
                                                                if (Rs < 3149) {
                                                                    return 8;
                                                                }
                                                                else if (Rs >= 3149) {
                                                                    return 1;
                                                                }
                                                            }
                                                        }
                                                        else if (negBalance >= 26) {
                                                            if (posBalance < 44) {
                                                                return 1;
                                                            }
                                                            else if (posBalance >= 44) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                    else if (posBalance >= 46) {
                                                        return 8;
                                                    }
                                                }
                                                else if (Rs >= 3322) {
                                                    return 8;
                                                }
                                            }
                                        }
                                    }
                                    else if (mu_mv_mag_sq >= 125) {
                                        if (ssDCval < 104) {
                                            return 8;
                                        }
                                        else if (ssDCval >= 104) {
                                            if (SC < 4237) {
                                                if (ssDCval < 114) {
                                                    if (Cs < 1283) {
                                                        return 8;
                                                    }
                                                    else if (Cs >= 1283) {
                                                        return 4;
                                                    }
                                                }
                                                else if (ssDCval >= 114) {
                                                    return 8;
                                                }
                                            }
                                            else if (SC >= 4237) {
                                                return 8;
                                            }
                                        }
                                    }
                                }
                            }
                        }
                    }
                }
            }
            else if (negBalance >= 36) {
                if (SC < 2715) {
                    if (posBalance < 57) {
                        if (ssDCval < 56) {
                            if (SC < 1923) {
                                if (refDCval < 48) {
                                    if (Rs < 489) {
                                        return 4;
                                    }
                                    else if (Rs >= 489) {
                                        if (Rs < 909) {
                                            return 8;
                                        }
                                        else if (Rs >= 909) {
                                            return 1;
                                        }
                                    }
                                }
                                else if (refDCval >= 48) {
                                    if (mu_mv_mag_sq < 46) {
                                        return 4;
                                    }
                                    else if (mu_mv_mag_sq >= 46) {
                                        if (Rs < 1012) {
                                            if (posBalance < 49) {
                                                return 4;
                                            }
                                            else if (posBalance >= 49) {
                                                return 8;
                                            }
                                        }
                                        else if (Rs >= 1012) {
                                            return 4;
                                        }
                                    }
                                }
                            }
                            else if (SC >= 1923) {
                                if (mu_mv_mag_sq < 61) {
                                    if (RsCsDiff < 19) {
                                        if (Rs < 1434) {
                                            return 1;
                                        }
                                        else if (Rs >= 1434) {
                                            if (Cs < 972) {
                                                return 8;
                                            }
                                            else if (Cs >= 972) {
                                                return 1;
                                            }
                                        }
                                    }
                                    else if (RsCsDiff >= 19) {
                                        if (Cs < 917) {
                                            return 1;
                                        }
                                        else if (Cs >= 917) {
                                            if (posBalance < 54) {
                                                return 4;
                                            }
                                            else if (posBalance >= 54) {
                                                return 1;
                                            }
                                        }
                                    }
                                }
                                else if (mu_mv_mag_sq >= 61) {
                                    if (negBalance < 41) {
                                        if (TSC < 4) {
                                            return 1;
                                        }
                                        else if (TSC >= 4) {
                                            return 4;
                                        }
                                    }
                                    else if (negBalance >= 41) {
                                        return 1;
                                    }
                                }
                            }
                        }
                        else if (ssDCval >= 56) {
                            if (AFD < 4) {
                                return 8;
                            }
                            else if (AFD >= 4) {
                                if (TSC < 4) {
                                    if (gchDC < 3) {
                                        if (Rs < 410) {
                                            return 2;
                                        }
                                        else if (Rs >= 410) {
                                            return 4;
                                        }
                                    }
                                    else if (gchDC >= 3) {
                                        return 8;
                                    }
                                }
                                else if (TSC >= 4) {
                                    if (ssDCval < 91) {
                                        if (mu_mv_mag_sq < 58) {
                                            if (mu_mv_mag_sq < 57) {
                                                if (ssDCval < 66) {
                                                    return 4;
                                                }
                                                else if (ssDCval >= 66) {
                                                    return 2;
                                                }
                                            }
                                            else if (mu_mv_mag_sq >= 57) {
                                                return 1;
                                            }
                                        }
                                        else if (mu_mv_mag_sq >= 58) {
                                            if (diffMVdiffVal < 1) {
                                                if (mu_mv_mag_sq < 70) {
                                                    if (Rs < 979) {
                                                        return 8;
                                                    }
                                                    else if (Rs >= 979) {
                                                        return 1;
                                                    }
                                                }
                                                else if (mu_mv_mag_sq >= 70) {
                                                    return 8;
                                                }
                                            }
                                            else if (diffMVdiffVal >= 1) {
                                                if (ssDCval < 63) {
                                                    return 8;
                                                }
                                                else if (ssDCval >= 63) {
                                                    return 4;
                                                }
                                            }
                                        }
                                    }
                                    else if (ssDCval >= 91) {
                                        if (Rs < 1627) {
                                            return 4;
                                        }
                                        else if (Rs >= 1627) {
                                            return 8;
                                        }
                                    }
                                }
                            }
                        }
                    }
                    else if (posBalance >= 57) {
                        if (Rs < 1096) {
                            if (ssDCval < 88) {
                                if (tcor < 95) {
                                    if (negBalance < 44) {
                                        return 8;
                                    }
                                    else if (negBalance >= 44) {
                                        return 2;
                                    }
                                }
                                else if (tcor >= 95) {
                                    if (Rs < 321) {
                                        return 4;
                                    }
                                    else if (Rs >= 321) {
                                        return 8;
                                    }
                                }
                            }
                            else if (ssDCval >= 88) {
                                if (refDCval < 116) {
                                    return 4;
                                }
                                else if (refDCval >= 116) {
                                    return 1;
                                }
                            }
                        }
                        else if (Rs >= 1096) {
                            if (RsCsDiff < 38) {
                                if (mcTcor < 88) {
                                    if (tcor < 79) {
                                        return 8;
                                    }
                                    else if (tcor >= 79) {
                                        return 2;
                                    }
                                }
                                else if (mcTcor >= 88) {
                                    if (MVDiff < 5) {
                                        return 2;
                                    }
                                    else if (MVDiff >= 5) {
                                        return 4;
                                    }
                                }
                            }
                            else if (RsCsDiff >= 38) {
                                if (mu_mv_mag_sq < 62) {
                                    if (Cs < 1216) {
                                        return 4;
                                    }
                                    else if (Cs >= 1216) {
                                        return 1;
                                    }
                                }
                                else if (mu_mv_mag_sq >= 62) {
                                    return 8;
                                }
                            }
                        }
                    }
                }
                else if (SC >= 2715) {
                    if (Cs < 4017) {
                        if (AFD < 7) {
                            if (ssDCval < 84) {
                                if (SC < 3570) {
                                    if (mu_mv_mag_sq < 108) {
                                        if (SC < 3274) {
                                            if (Rs < 2084) {
                                                return 8;
                                            }
                                            else if (Rs >= 2084) {
                                                return 2;
                                            }
                                        }
                                        else if (SC >= 3274) {
                                            return 4;
                                        }
                                    }
                                    else if (mu_mv_mag_sq >= 108) {
                                        if (TSCindex < 7) {
                                            if (tcor < 97) {
                                                if (Cs < 1576) {
                                                    return 8;
                                                }
                                                else if (Cs >= 1576) {
                                                    if (tcor < 95) {
                                                        return 8;
                                                    }
                                                    else if (tcor >= 95) {
                                                        if (Rs < 1898) {
                                                            if (Cs < 1579) {
                                                                return 4;
                                                            }
                                                            else if (Cs >= 1579) {
                                                                if (Cs < 1625) {
                                                                    return 8;
                                                                }
                                                                else if (Cs >= 1625) {
                                                                    if (SC < 3462) {
                                                                        return 8;
                                                                    }
                                                                    else if (SC >= 3462) {
                                                                        if (Cs < 1631) {
                                                                            return 4;
                                                                        }
                                                                        else if (Cs >= 1631) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (Rs >= 1898) {
                                                            if (mu_mv_mag_sq < 125) {
                                                                return 4;
                                                            }
                                                            else if (mu_mv_mag_sq >= 125) {
                                                                return 2;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                            else if (tcor >= 97) {
                                                if (Cs < 1272) {
                                                    if (posBalance < 54) {
                                                        if (ssDCval < 73) {
                                                            return 8;
                                                        }
                                                        else if (ssDCval >= 73) {
                                                            if (AFD < 4) {
                                                                return 8;
                                                            }
                                                            else if (AFD >= 4) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                    else if (posBalance >= 54) {
                                                        return 8;
                                                    }
                                                }
                                                else if (Cs >= 1272) {
                                                    if (AFD < 5) {
                                                        if (Rs < 1783) {
                                                            return 8;
                                                        }
                                                        else if (Rs >= 1783) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (AFD >= 5) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                        }
                                        else if (TSCindex >= 7) {
                                            if (refDCval < 74) {
                                                if (mcTcor < 94) {
                                                    if (RsDiff < 4) {
                                                        if (Rs < 1836) {
                                                            return 2;
                                                        }
                                                        else if (Rs >= 1836) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (RsDiff >= 4) {
                                                        return 8;
                                                    }
                                                }
                                                else if (mcTcor >= 94) {
                                                    return 2;
                                                }
                                            }
                                            else if (refDCval >= 74) {
                                                if (Rs < 1845) {
                                                    if (RsDiff < 4) {
                                                        return 4;
                                                    }
                                                    else if (RsDiff >= 4) {
                                                        return 8;
                                                    }
                                                }
                                                else if (Rs >= 1845) {
                                                    return 4;
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (SC >= 3570) {
                                    if (tcor < 94) {
                                        if (ssDCval < 81) {
                                            if (Cs < 1676) {
                                                if (SC < 3798) {
                                                    return 2;
                                                }
                                                else if (SC >= 3798) {
                                                    return 8;
                                                }
                                            }
                                            else if (Cs >= 1676) {
                                                return 8;
                                            }
                                        }
                                        else if (ssDCval >= 81) {
                                            return 4;
                                        }
                                    }
                                    else if (tcor >= 94) {
                                        if (Rs < 2943) {
                                            if (tcor < 95) {
                                                if (Rs < 1837) {
                                                    return 8;
                                                }
                                                else if (Rs >= 1837) {
                                                    if (Rs < 2358) {
                                                        if (Cs < 1911) {
                                                            if (Rs < 1944) {
                                                                return 8;
                                                            }
                                                            else if (Rs >= 1944) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (Cs >= 1911) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (Rs >= 2358) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                            else if (tcor >= 95) {
                                                if (posBalance < 32) {
                                                    return 8;
                                                }
                                                else if (posBalance >= 32) {
                                                    if (posBalance < 58) {
                                                        return 4;
                                                    }
                                                    else if (posBalance >= 58) {
                                                        if (TSC < 6) {
                                                            if (Cs < 1799) {
                                                                return 4;
                                                            }
                                                            else if (Cs >= 1799) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (TSC >= 6) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (Rs >= 2943) {
                                            if (mu_mv_mag_sq < 94) {
                                                if (TSC < 5) {
                                                    return 8;
                                                }
                                                else if (TSC >= 5) {
                                                    if (negBalance < 47) {
                                                        if (ssDCval < 55) {
                                                            return 4;
                                                        }
                                                        else if (ssDCval >= 55) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (negBalance >= 47) {
                                                        if (RsDiff < 3) {
                                                            return 8;
                                                        }
                                                        else if (RsDiff >= 3) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (mu_mv_mag_sq >= 94) {
                                                return 8;
                                            }
                                        }
                                    }
                                }
                            }
                            else if (ssDCval >= 84) {
                                if (Cs < 3161) {
                                    if (mu_mv_mag_sq < 123) {
                                        if (MVDiff < 1) {
                                            if (mu_mv_mag_sq < 117) {
                                                if (Rs < 2722) {
                                                    if (Rs < 1808) {
                                                        return 8;
                                                    }
                                                    else if (Rs >= 1808) {
                                                        if (Cs < 3012) {
                                                            return 4;
                                                        }
                                                        else if (Cs >= 3012) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                                else if (Rs >= 2722) {
                                                    return 8;
                                                }
                                            }
                                            else if (mu_mv_mag_sq >= 117) {
                                                if (negBalance < 38) {
                                                    return 4;
                                                }
                                                else if (negBalance >= 38) {
                                                    if (posBalance < 59) {
                                                        if (SC < 5612) {
                                                            if (RsCsDiff < 46) {
                                                                if (Rs < 1771) {
                                                                    return 4;
                                                                }
                                                                else if (Rs >= 1771) {
                                                                    if (SC < 5553) {
                                                                        return 8;
                                                                    }
                                                                    else if (SC >= 5553) {
                                                                        if (RsCsDiff < 29) {
                                                                            if (SC < 5572) {
                                                                                return 4;
                                                                            }
                                                                            else if (SC >= 5572) {
                                                                                return 8;
                                                                            }
                                                                        }
                                                                        else if (RsCsDiff >= 29) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                            else if (RsCsDiff >= 46) {
                                                                if (diffAFD < -1) {
                                                                    return 4;
                                                                }
                                                                else if (diffAFD >= -1) {
                                                                    if (diffMVdiffVal < -1) {
                                                                        return 4;
                                                                    }
                                                                    else if (diffMVdiffVal >= -1) {
                                                                        return 8;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (SC >= 5612) {
                                                            if (refDCval < 114) {
                                                                if (tcor < 97) {
                                                                    return 8;
                                                                }
                                                                else if (tcor >= 97) {
                                                                    if (RsCsDiff < 46) {
                                                                        if (ssDCval < 113) {
                                                                            if (Rs < 2802) {
                                                                                return 8;
                                                                            }
                                                                            else if (Rs >= 2802) {
                                                                                if (Scindex < 6) {
                                                                                    if (ssDCval < 111) {
                                                                                        return 8;
                                                                                    }
                                                                                    else if (ssDCval >= 111) {
                                                                                        if (mu_mv_mag_sq < 122) {
                                                                                            if (SC < 5844) {
                                                                                                return 8;
                                                                                            }
                                                                                            else if (SC >= 5844) {
                                                                                                return 4;
                                                                                            }
                                                                                        }
                                                                                        else if (mu_mv_mag_sq >= 122) {
                                                                                            return 4;
                                                                                        }
                                                                                    }
                                                                                }
                                                                                else if (Scindex >= 6) {
                                                                                    return 8;
                                                                                }
                                                                            }
                                                                        }
                                                                        else if (ssDCval >= 113) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                    else if (RsCsDiff >= 46) {
                                                                        if (mu_mv_mag_sq < 120) {
                                                                            return 4;
                                                                        }
                                                                        else if (mu_mv_mag_sq >= 120) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                            else if (refDCval >= 114) {
                                                                if (RsCsDiff < 37) {
                                                                    return 8;
                                                                }
                                                                else if (RsCsDiff >= 37) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (posBalance >= 59) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                        }
                                        else if (MVDiff >= 1) {
                                            if (mcTcor < 98) {
                                                if (mcTcor < 93) {
                                                    if (Rs < 2288) {
                                                        return 8;
                                                    }
                                                    else if (Rs >= 2288) {
                                                        return 4;
                                                    }
                                                }
                                                else if (mcTcor >= 93) {
                                                    return 8;
                                                }
                                            }
                                            else if (mcTcor >= 98) {
                                                if (RsDiff < 4) {
                                                    return 8;
                                                }
                                                else if (RsDiff >= 4) {
                                                    if (mu_mv_mag_sq < 100) {
                                                        return 1;
                                                    }
                                                    else if (mu_mv_mag_sq >= 100) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (mu_mv_mag_sq >= 123) {
                                        if (Rs < 1791) {
                                            if (SC < 3848) {
                                                return 8;
                                            }
                                            else if (SC >= 3848) {
                                                return 4;
                                            }
                                        }
                                        else if (Rs >= 1791) {
                                            if (SC < 4738) {
                                                return 8;
                                            }
                                            else if (SC >= 4738) {
                                                if (Rs < 2359) {
                                                    return 4;
                                                }
                                                else if (Rs >= 2359) {
                                                    if (SC < 5750) {
                                                        return 8;
                                                    }
                                                    else if (SC >= 5750) {
                                                        if (posBalance < 58) {
                                                            if (mu_mv_mag_sq < 126) {
                                                                if (SC < 5769) {
                                                                    if (negBalance < 48) {
                                                                        return 8;
                                                                    }
                                                                    else if (negBalance >= 48) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (SC >= 5769) {
                                                                    if (diffRsCsdiff < -8) {
                                                                        if (Rs < 2868) {
                                                                            return 4;
                                                                        }
                                                                        else if (Rs >= 2868) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                    else if (diffRsCsdiff >= -8) {
                                                                        return 8;
                                                                    }
                                                                }
                                                            }
                                                            else if (mu_mv_mag_sq >= 126) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (posBalance >= 58) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (Cs >= 3161) {
                                    return 8;
                                }
                            }
                        }
                        else if (AFD >= 7) {
                            if (mu_mv_mag_sq < 128) {
                                if (tcor < 77) {
                                    if (refDCval < 122) {
                                        if (CsDiff < 6) {
                                            return 8;
                                        }
                                        else if (CsDiff >= 6) {
                                            if (diffRsCsdiff < -24) {
                                                return 2;
                                            }
                                            else if (diffRsCsdiff >= -24) {
                                                return 4;
                                            }
                                        }
                                    }
                                    else if (refDCval >= 122) {
                                        if (negBalance < 55) {
                                            return 1;
                                        }
                                        else if (negBalance >= 55) {
                                            if (AFD < 18) {
                                                return 8;
                                            }
                                            else if (AFD >= 18) {
                                                return 1;
                                            }
                                        }
                                    }
                                }
                                else if (tcor >= 77) {
                                    if (Rs < 2760) {
                                        if (refDCval < 99) {
                                            if (mu_mv_mag_sq < 104) {
                                                if (TSC < 7) {
                                                    if (Cs < 1859) {
                                                        if (TSC < 6) {
                                                            if (Cs < 857) {
                                                                return 2;
                                                            }
                                                            else if (Cs >= 857) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (TSC >= 6) {
                                                            if (SC < 4051) {
                                                                if (Cs < 1417) {
                                                                    return 4;
                                                                }
                                                                else if (Cs >= 1417) {
                                                                    return 2;
                                                                }
                                                            }
                                                            else if (SC >= 4051) {
                                                                return 1;
                                                            }
                                                        }
                                                    }
                                                    else if (Cs >= 1859) {
                                                        if (Rs < 1372) {
                                                            return 8;
                                                        }
                                                        else if (Rs >= 1372) {
                                                            if (Cs < 2567) {
                                                                return 4;
                                                            }
                                                            else if (Cs >= 2567) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (TSC >= 7) {
                                                    if (mcTcor < 91) {
                                                        if (Rs < 1757) {
                                                            if (diffMVdiffVal < -2) {
                                                                return 4;
                                                            }
                                                            else if (diffMVdiffVal >= -2) {
                                                                if (Rs < 1644) {
                                                                    return 8;
                                                                }
                                                                else if (Rs >= 1644) {
                                                                    return 1;
                                                                }
                                                            }
                                                        }
                                                        else if (Rs >= 1757) {
                                                            if (refDCval < 93) {
                                                                return 8;
                                                            }
                                                            else if (refDCval >= 93) {
                                                                if (CsDiff < 5) {
                                                                    return 8;
                                                                }
                                                                else if (CsDiff >= 5) {
                                                                    if (Rs < 1906) {
                                                                        if (Cs < 1682) {
                                                                            return 2;
                                                                        }
                                                                        else if (Cs >= 1682) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                    else if (Rs >= 1906) {
                                                                        return 2;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (mcTcor >= 91) {
                                                        if (negBalance < 46) {
                                                            if (ssDCval < 88) {
                                                                if (Cs < 2019) {
                                                                    if (RsDiff < 5) {
                                                                        return 2;
                                                                    }
                                                                    else if (RsDiff >= 5) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (Cs >= 2019) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (ssDCval >= 88) {
                                                                if (diffAFD < -2) {
                                                                    if (Rs < 2300) {
                                                                        return 2;
                                                                    }
                                                                    else if (Rs >= 2300) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (diffAFD >= -2) {
                                                                    if (SC < 4233) {
                                                                        if (AFD < 9) {
                                                                            return 1;
                                                                        }
                                                                        else if (AFD >= 9) {
                                                                            return 2;
                                                                        }
                                                                    }
                                                                    else if (SC >= 4233) {
                                                                        return 1;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (negBalance >= 46) {
                                                            if (SC < 4185) {
                                                                if (tcor < 95) {
                                                                    if (diffMVdiffVal < -2) {
                                                                        if (RsDiff < 7) {
                                                                            if (Rs < 2036) {
                                                                                return 4;
                                                                            }
                                                                            else if (Rs >= 2036) {
                                                                                return 1;
                                                                            }
                                                                        }
                                                                        else if (RsDiff >= 7) {
                                                                            return 2;
                                                                        }
                                                                    }
                                                                    else if (diffMVdiffVal >= -2) {
                                                                        return 2;
                                                                    }
                                                                }
                                                                else if (tcor >= 95) {
                                                                    if (SC < 3426) {
                                                                        return 1;
                                                                    }
                                                                    else if (SC >= 3426) {
                                                                        return 4;
                                                                    }
                                                                }
                                                            }
                                                            else if (SC >= 4185) {
                                                                if (mcTcor < 96) {
                                                                    if (TSC < 11) {
                                                                        if (tcor < 87) {
                                                                            return 4;
                                                                        }
                                                                        else if (tcor >= 87) {
                                                                            if (ssDCval < 87) {
                                                                                return 4;
                                                                            }
                                                                            else if (ssDCval >= 87) {
                                                                                if (diffTSC < 1) {
                                                                                    return 2;
                                                                                }
                                                                                else if (diffTSC >= 1) {
                                                                                    if (ssDCval < 93) {
                                                                                        return 2;
                                                                                    }
                                                                                    else if (ssDCval >= 93) {
                                                                                        return 4;
                                                                                    }
                                                                                }
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (TSC >= 11) {
                                                                        if (SC < 4492) {
                                                                            return 4;
                                                                        }
                                                                        else if (SC >= 4492) {
                                                                            return 2;
                                                                        }
                                                                    }
                                                                }
                                                                else if (mcTcor >= 96) {
                                                                    if (Rs < 2496) {
                                                                        if (negBalance < 51) {
                                                                            return 2;
                                                                        }
                                                                        else if (negBalance >= 51) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                    else if (Rs >= 2496) {
                                                                        return 1;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                            else if (mu_mv_mag_sq >= 104) {
                                                if (mcTcor < 93) {
                                                    if (ssDCval < 72) {
                                                        return 2;
                                                    }
                                                    else if (ssDCval >= 72) {
                                                        if (refDCval < 77) {
                                                            if (Cs < 1656) {
                                                                return 8;
                                                            }
                                                            else if (Cs >= 1656) {
                                                                return 2;
                                                            }
                                                        }
                                                        else if (refDCval >= 77) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                                else if (mcTcor >= 93) {
                                                    if (ssDCval < 75) {
                                                        if (tcor < 91) {
                                                            if (SC < 3288) {
                                                                return 2;
                                                            }
                                                            else if (SC >= 3288) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (tcor >= 91) {
                                                            if (Rs < 2024) {
                                                                return 2;
                                                            }
                                                            else if (Rs >= 2024) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                    else if (ssDCval >= 75) {
                                                        if (posBalance < 57) {
                                                            return 4;
                                                        }
                                                        else if (posBalance >= 57) {
                                                            if (Cs < 1781) {
                                                                if (diffAFD < -1) {
                                                                    return 4;
                                                                }
                                                                else if (diffAFD >= -1) {
                                                                    if (Cs < 1551) {
                                                                        return 4;
                                                                    }
                                                                    else if (Cs >= 1551) {
                                                                        return 2;
                                                                    }
                                                                }
                                                            }
                                                            else if (Cs >= 1781) {
                                                                if (ssDCval < 90) {
                                                                    return 8;
                                                                }
                                                                else if (ssDCval >= 90) {
                                                                    return 2;
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (refDCval >= 99) {
                                            if (mu_mv_mag_sq < 57) {
                                                if (posBalance < 58) {
                                                    if (tcor < 83) {
                                                        if (negBalance < 53) {
                                                            if (Rs < 2435) {
                                                                return 1;
                                                            }
                                                            else if (Rs >= 2435) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (negBalance >= 53) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (tcor >= 83) {
                                                        if (RsDiff < 8) {
                                                            if (Rs < 1976) {
                                                                return 8;
                                                            }
                                                            else if (Rs >= 1976) {
                                                                return 2;
                                                            }
                                                        }
                                                        else if (RsDiff >= 8) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                                else if (posBalance >= 58) {
                                                    if (SC < 4563) {
                                                        return 4;
                                                    }
                                                    else if (SC >= 4563) {
                                                        if (diffAFD < 1) {
                                                            return 4;
                                                        }
                                                        else if (diffAFD >= 1) {
                                                            if (CsDiff < 7) {
                                                                if (Rs < 2621) {
                                                                    return 8;
                                                                }
                                                                else if (Rs >= 2621) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (CsDiff >= 7) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                            else if (mu_mv_mag_sq >= 57) {
                                                if (ssDCval < 110) {
                                                    if (Rs < 2566) {
                                                        return 4;
                                                    }
                                                    else if (Rs >= 2566) {
                                                        if (MVDiff < 4) {
                                                            return 8;
                                                        }
                                                        else if (MVDiff >= 4) {
                                                            if (Cs < 2340) {
                                                                return 1;
                                                            }
                                                            else if (Cs >= 2340) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (ssDCval >= 110) {
                                                    if (diffRsCsdiff < -36) {
                                                        return 4;
                                                    }
                                                    else if (diffRsCsdiff >= -36) {
                                                        if (SC < 5807) {
                                                            if (Rs < 2265) {
                                                                return 8;
                                                            }
                                                            else if (Rs >= 2265) {
                                                                if (Rs < 2488) {
                                                                    if (posBalance < 57) {
                                                                        return 4;
                                                                    }
                                                                    else if (posBalance >= 57) {
                                                                        if (mu_mv_mag_sq < 72) {
                                                                            return 4;
                                                                        }
                                                                        else if (mu_mv_mag_sq >= 72) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                }
                                                                else if (Rs >= 2488) {
                                                                    if (diffRsCsdiff < 40) {
                                                                        return 8;
                                                                    }
                                                                    else if (diffRsCsdiff >= 40) {
                                                                        return 4;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (SC >= 5807) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (Rs >= 2760) {
                                        if (Rs < 3961) {
                                            if (ssDCval < 99) {
                                                if (RsCsDiff < 153) {
                                                    if (ssDCval < 88) {
                                                        if (Cs < 3434) {
                                                            if (refDCval < 86) {
                                                                if (ssDCval < 86) {
                                                                    if (negBalance < 52) {
                                                                        return 4;
                                                                    }
                                                                    else if (negBalance >= 52) {
                                                                        return 8;
                                                                    }
                                                                }
                                                                else if (ssDCval >= 86) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (refDCval >= 86) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (Cs >= 3434) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (ssDCval >= 88) {
                                                        if (MVDiff < 1) {
                                                            return 2;
                                                        }
                                                        else if (MVDiff >= 1) {
                                                            if (Cs < 2409) {
                                                                if (TSC < 10) {
                                                                    return 1;
                                                                }
                                                                else if (TSC >= 10) {
                                                                    return 2;
                                                                }
                                                            }
                                                            else if (Cs >= 2409) {
                                                                if (ssDCval < 96) {
                                                                    if (refDCval < 90) {
                                                                        return 2;
                                                                    }
                                                                    else if (refDCval >= 90) {
                                                                        return 8;
                                                                    }
                                                                }
                                                                else if (ssDCval >= 96) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (RsCsDiff >= 153) {
                                                    if (SC < 5324) {
                                                        return 4;
                                                    }
                                                    else if (SC >= 5324) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                            else if (ssDCval >= 99) {
                                                if (SC < 5243) {
                                                    if (mu_mv_mag_sq < 71) {
                                                        if (Rs < 3042) {
                                                            if (Cs < 2390) {
                                                                return 4;
                                                            }
                                                            else if (Cs >= 2390) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (Rs >= 3042) {
                                                            return 2;
                                                        }
                                                    }
                                                    else if (mu_mv_mag_sq >= 71) {
                                                        if (mu_mv_mag_sq < 87) {
                                                            if (ssDCval < 117) {
                                                                return 4;
                                                            }
                                                            else if (ssDCval >= 117) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (mu_mv_mag_sq >= 87) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                                else if (SC >= 5243) {
                                                    if (TSC < 11) {
                                                        if (RsDiff < 10) {
                                                            if (negBalance < 50) {
                                                                return 8;
                                                            }
                                                            else if (negBalance >= 50) {
                                                                if (tcor < 83) {
                                                                    if (Cs < 2746) {
                                                                        return 4;
                                                                    }
                                                                    else if (Cs >= 2746) {
                                                                        return 8;
                                                                    }
                                                                }
                                                                else if (tcor >= 83) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                        else if (RsDiff >= 10) {
                                                            if (ssDCval < 105) {
                                                                if (AFD < 14) {
                                                                    return 4;
                                                                }
                                                                else if (AFD >= 14) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (ssDCval >= 105) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                    else if (TSC >= 11) {
                                                        if (Cs < 2908) {
                                                            if (diffAFD < 2) {
                                                                return 1;
                                                            }
                                                            else if (diffAFD >= 2) {
                                                                if (diffMVdiffVal < -1) {
                                                                    return 4;
                                                                }
                                                                else if (diffMVdiffVal >= -1) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                        else if (Cs >= 2908) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (Rs >= 3961) {
                                            if (refDCval < 122) {
                                                if (mu_mv_mag_sq < 122) {
                                                    if (posBalance < 60) {
                                                        if (posBalance < 47) {
                                                            return 8;
                                                        }
                                                        else if (posBalance >= 47) {
                                                            if (mcTcor < 90) {
                                                                if (Rs < 4608) {
                                                                    return 8;
                                                                }
                                                                else if (Rs >= 4608) {
                                                                    return 2;
                                                                }
                                                            }
                                                            else if (mcTcor >= 90) {
                                                                if (Cs < 3783) {
                                                                    return 4;
                                                                }
                                                                else if (Cs >= 3783) {
                                                                    return 2;
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (posBalance >= 60) {
                                                        return 8;
                                                    }
                                                }
                                                else if (mu_mv_mag_sq >= 122) {
                                                    return 4;
                                                }
                                            }
                                            else if (refDCval >= 122) {
                                                if (SC < 6215) {
                                                    return 4;
                                                }
                                                else if (SC >= 6215) {
                                                    return 8;
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                            else if (mu_mv_mag_sq >= 128) {
                                if (SC < 7395) {
                                    if (Rs < 3599) {
                                        return 8;
                                    }
                                    else if (Rs >= 3599) {
                                        return 2;
                                    }
                                }
                                else if (SC >= 7395) {
                                    return 4;
                                }
                            }
                        }
                    }
                    else if (Cs >= 4017) {
                        if (mu_mv_mag_sq < 128) {
                            if (negBalance < 64) {
                                if (refDCval < 106) {
                                    if (RsCsDiff < 163) {
                                        if (mu_mv_mag_sq < 127) {
                                            if (ssDCval < 73) {
                                                return 4;
                                            }
                                            else if (ssDCval >= 73) {
                                                if (mu_mv_mag_sq < 126) {
                                                    return 8;
                                                }
                                                else if (mu_mv_mag_sq >= 126) {
                                                    if (posBalance < 58) {
                                                        return 8;
                                                    }
                                                    else if (posBalance >= 58) {
                                                        if (Cs < 4711) {
                                                            return 8;
                                                        }
                                                        else if (Cs >= 4711) {
                                                            if (Rs < 5991) {
                                                                return 4;
                                                            }
                                                            else if (Rs >= 5991) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (mu_mv_mag_sq >= 127) {
                                            if (ssDCval < 84) {
                                                if (CsDiff < 9) {
                                                    if (SC < 9883) {
                                                        return 8;
                                                    }
                                                    else if (SC >= 9883) {
                                                        if (refDCval < 80) {
                                                            return 8;
                                                        }
                                                        else if (refDCval >= 80) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                                else if (CsDiff >= 9) {
                                                    return 4;
                                                }
                                            }
                                            else if (ssDCval >= 84) {
                                                return 8;
                                            }
                                        }
                                    }
                                    else if (RsCsDiff >= 163) {
                                        return 8;
                                    }
                                }
                                else if (refDCval >= 106) {
                                    if (ssDCval < 123) {
                                        if (mu_mv_mag_sq < 124) {
                                            if (mu_mv_mag_sq < 63) {
                                                if (TSC < 25) {
                                                    if (Rs < 5020) {
                                                        return 4;
                                                    }
                                                    else if (Rs >= 5020) {
                                                        return 8;
                                                    }
                                                }
                                                else if (TSC >= 25) {
                                                    return 4;
                                                }
                                            }
                                            else if (mu_mv_mag_sq >= 63) {
                                                if (RsCsDiff < 379) {
                                                    if (posBalance < 58) {
                                                        if (Cs < 4079) {
                                                            return 4;
                                                        }
                                                        else if (Cs >= 4079) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (posBalance >= 58) {
                                                        if (Cs < 5064) {
                                                            return 8;
                                                        }
                                                        else if (Cs >= 5064) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                                else if (RsCsDiff >= 379) {
                                                    return 8;
                                                }
                                            }
                                        }
                                        else if (mu_mv_mag_sq >= 124) {
                                            if (ssDCval < 106) {
                                                return 8;
                                            }
                                            else if (ssDCval >= 106) {
                                                if (negBalance < 46) {
                                                    return 8;
                                                }
                                                else if (negBalance >= 46) {
                                                    if (Cs < 5999) {
                                                        return 4;
                                                    }
                                                    else if (Cs >= 5999) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (ssDCval >= 123) {
                                        if (mcTcor < 76) {
                                            return 1;
                                        }
                                        else if (mcTcor >= 76) {
                                            return 8;
                                        }
                                    }
                                }
                            }
                            else if (negBalance >= 64) {
                                if (Rs < 7247) {
                                    if (mcTcor < 77) {
                                        if (Cs < 7114) {
                                            return 8;
                                        }
                                        else if (Cs >= 7114) {
                                            return 4;
                                        }
                                    }
                                    else if (mcTcor >= 77) {
                                        return 8;
                                    }
                                }
                                else if (Rs >= 7247) {
                                    if (SC < 15622) {
                                        if (CsDiff < 22) {
                                            return 4;
                                        }
                                        else if (CsDiff >= 22) {
                                            return 2;
                                        }
                                    }
                                    else if (SC >= 15622) {
                                        if (RsCsDiff < 1036) {
                                            return 2;
                                        }
                                        else if (RsCsDiff >= 1036) {
                                            if (Cs < 8194) {
                                                return 4;
                                            }
                                            else if (Cs >= 8194) {
                                                if (SC < 16695) {
                                                    return 2;
                                                }
                                                else if (SC >= 16695) {
                                                    return 4;
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                        }
                        else if (mu_mv_mag_sq >= 128) {
                            if (mcTcor < 96) {
                                if (negBalance < 55) {
                                    if (ssDCval < 84) {
                                        if (ssDCval < 80) {
                                            if (Rs < 4576) {
                                                return 4;
                                            }
                                            else if (Rs >= 4576) {
                                                return 8;
                                            }
                                        }
                                        else if (ssDCval >= 80) {
                                            if (CsDiff < 9) {
                                                if (CsDiff < 8) {
                                                    if (Rs < 5033) {
                                                        return 8;
                                                    }
                                                    else if (Rs >= 5033) {
                                                        return 4;
                                                    }
                                                }
                                                else if (CsDiff >= 8) {
                                                    if (Cs < 4916) {
                                                        if (SC < 9676) {
                                                            if (SC < 9543) {
                                                                if (Rs < 4623) {
                                                                    return 8;
                                                                }
                                                                else if (Rs >= 4623) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (SC >= 9543) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (SC >= 9676) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (Cs >= 4916) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                            else if (CsDiff >= 9) {
                                                if (TSCindex < 8) {
                                                    if (Cs < 4995) {
                                                        return 4;
                                                    }
                                                    else if (Cs >= 4995) {
                                                        return 8;
                                                    }
                                                }
                                                else if (TSCindex >= 8) {
                                                    if (SC < 9816) {
                                                        return 8;
                                                    }
                                                    else if (SC >= 9816) {
                                                        if (posBalance < 59) {
                                                            return 4;
                                                        }
                                                        else if (posBalance >= 59) {
                                                            if (RsCsDiff < 155) {
                                                                return 8;
                                                            }
                                                            else if (RsCsDiff >= 155) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (ssDCval >= 84) {
                                        if (RsCsDiff < 328) {
                                            return 8;
                                        }
                                        else if (RsCsDiff >= 328) {
                                            if (ssDCval < 105) {
                                                return 8;
                                            }
                                            else if (ssDCval >= 105) {
                                                return 4;
                                            }
                                        }
                                    }
                                }
                                else if (negBalance >= 55) {
                                    if (Cs < 6323) {
                                        return 4;
                                    }
                                    else if (Cs >= 6323) {
                                        return 2;
                                    }
                                }
                            }
                            else if (mcTcor >= 96) {
                                return 8;
                            }
                        }
                    }
                }
            }
        }
        else if (MVDiff >= 6) {
            if (Cs < 3432) {
                if (tcor < 88) {
                    if (mcTcor < 94) {
                        if (CsDiff < 10) {
                            if (negBalance < 55) {
                                if (RsCsDiff < 194) {
                                    if (Cs < 270) {
                                        if (TSCindex < 7) {
                                            if (AFD < 9) {
                                                if (Rs < 342) {
                                                    if (AFD < 3) {
                                                        if (mcTcor < 71) {
                                                            if (posBalance < 42) {
                                                                if (tcor < 44) {
                                                                    if (MVDiff < 62) {
                                                                        if (posBalance < 28) {
                                                                            if (posBalance < 27) {
                                                                                return 1;
                                                                            }
                                                                            else if (posBalance >= 27) {
                                                                                if (MVDiff < 59) {
                                                                                    return 4;
                                                                                }
                                                                                else if (MVDiff >= 59) {
                                                                                    return 1;
                                                                                }
                                                                            }
                                                                        }
                                                                        else if (posBalance >= 28) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                    else if (MVDiff >= 62) {
                                                                        if (diffRsCsdiff < -1335) {
                                                                            return 1;
                                                                        }
                                                                        else if (diffRsCsdiff >= -1335) {
                                                                            if (mu_mv_mag_sq < 29) {
                                                                                if (MVDiff < 69) {
                                                                                    if (diffMVdiffVal < 1) {
                                                                                        return 4;
                                                                                    }
                                                                                    else if (diffMVdiffVal >= 1) {
                                                                                        return 8;
                                                                                    }
                                                                                }
                                                                                else if (MVDiff >= 69) {
                                                                                    return 4;
                                                                                }
                                                                            }
                                                                            else if (mu_mv_mag_sq >= 29) {
                                                                                return 8;
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                                else if (tcor >= 44) {
                                                                    if (mu_mv_mag_sq < 32) {
                                                                        if (mcTcor < 28) {
                                                                            return 4;
                                                                        }
                                                                        else if (mcTcor >= 28) {
                                                                            if (posBalance < 25) {
                                                                                return 8;
                                                                            }
                                                                            else if (posBalance >= 25) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (mu_mv_mag_sq >= 32) {
                                                                        return 4;
                                                                    }
                                                                }
                                                            }
                                                            else if (posBalance >= 42) {
                                                                if (posBalance < 87) {
                                                                    if (ssDCval < 18) {
                                                                        return 1;
                                                                    }
                                                                    else if (ssDCval >= 18) {
                                                                        if (AFD < 1) {
                                                                            if (posBalance < 47) {
                                                                                return 2;
                                                                            }
                                                                            else if (posBalance >= 47) {
                                                                                if (tcor < 48) {
                                                                                    return 4;
                                                                                }
                                                                                else if (tcor >= 48) {
                                                                                    return 2;
                                                                                }
                                                                            }
                                                                        }
                                                                        else if (AFD >= 1) {
                                                                            if (gchDC < 2) {
                                                                                return 2;
                                                                            }
                                                                            else if (gchDC >= 2) {
                                                                                return 1;
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                                else if (posBalance >= 87) {
                                                                    if (negBalance < 1) {
                                                                        if (TSCindex < 2) {
                                                                            if (MVDiff < 615) {
                                                                                if (diffAFD < 1) {
                                                                                    return 4;
                                                                                }
                                                                                else if (diffAFD >= 1) {
                                                                                    return 8;
                                                                                }
                                                                            }
                                                                            else if (MVDiff >= 615) {
                                                                                return 2;
                                                                            }
                                                                        }
                                                                        else if (TSCindex >= 2) {
                                                                            if (ssDCval < 26) {
                                                                                return 1;
                                                                            }
                                                                            else if (ssDCval >= 26) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (negBalance >= 1) {
                                                                        return 1;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (mcTcor >= 71) {
                                                            if (tcor < 58) {
                                                                return 1;
                                                            }
                                                            else if (tcor >= 58) {
                                                                if (mcTcor < 86) {
                                                                    if (negBalance < 1) {
                                                                        if (ssDCval < 23) {
                                                                            if (mu_mv_mag_sq < 16) {
                                                                                return 8;
                                                                            }
                                                                            else if (mu_mv_mag_sq >= 16) {
                                                                                return 2;
                                                                            }
                                                                        }
                                                                        else if (ssDCval >= 23) {
                                                                            if (Scindex < 1) {
                                                                                if (refDCval < 28) {
                                                                                    return 4;
                                                                                }
                                                                                else if (refDCval >= 28) {
                                                                                    return 2;
                                                                                }
                                                                            }
                                                                            else if (Scindex >= 1) {
                                                                                return 2;
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (negBalance >= 1) {
                                                                        if (mu_mv_mag_sq < 20) {
                                                                            return 2;
                                                                        }
                                                                        else if (mu_mv_mag_sq >= 20) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                }
                                                                else if (mcTcor >= 86) {
                                                                    if (MVDiff < 557) {
                                                                        return 8;
                                                                    }
                                                                    else if (MVDiff >= 557) {
                                                                        return 2;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (AFD >= 3) {
                                                        if (MVDiff < 23) {
                                                            if (mu_mv_mag_sq < 15) {
                                                                if (Cs < 23) {
                                                                    if (SC < 59) {
                                                                        return 2;
                                                                    }
                                                                    else if (SC >= 59) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (Cs >= 23) {
                                                                    return 2;
                                                                }
                                                            }
                                                            else if (mu_mv_mag_sq >= 15) {
                                                                if (Rs < 274) {
                                                                    return 8;
                                                                }
                                                                else if (Rs >= 274) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                        else if (MVDiff >= 23) {
                                                            if (diffTSC < 1) {
                                                                if (diffAFD < -2) {
                                                                    if (Rs < 9) {
                                                                        return 8;
                                                                    }
                                                                    else if (Rs >= 9) {
                                                                        return 1;
                                                                    }
                                                                }
                                                                else if (diffAFD >= -2) {
                                                                    if (diffMVdiffVal < -2) {
                                                                        if (Rs < 10) {
                                                                            return 4;
                                                                        }
                                                                        else if (Rs >= 10) {
                                                                            if (Rs < 131) {
                                                                                return 8;
                                                                            }
                                                                            else if (Rs >= 131) {
                                                                                return 2;
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (diffMVdiffVal >= -2) {
                                                                        if (refDCval < 90) {
                                                                            return 2;
                                                                        }
                                                                        else if (refDCval >= 90) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                            else if (diffTSC >= 1) {
                                                                if (MVDiff < 37) {
                                                                    if (MVDiff < 27) {
                                                                        return 1;
                                                                    }
                                                                    else if (MVDiff >= 27) {
                                                                        return 2;
                                                                    }
                                                                }
                                                                else if (MVDiff >= 37) {
                                                                    if (posBalance < 123) {
                                                                        return 1;
                                                                    }
                                                                    else if (posBalance >= 123) {
                                                                        return 2;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (Rs >= 342) {
                                                    if (posBalance < 68) {
                                                        return 1;
                                                    }
                                                    else if (posBalance >= 68) {
                                                        if (AFD < 6) {
                                                            return 4;
                                                        }
                                                        else if (AFD >= 6) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (AFD >= 9) {
                                                if (refDCval < 55) {
                                                    if (Cs < 113) {
                                                        return 2;
                                                    }
                                                    else if (Cs >= 113) {
                                                        return 1;
                                                    }
                                                }
                                                else if (refDCval >= 55) {
                                                    return 2;
                                                }
                                            }
                                        }
                                        else if (TSCindex >= 7) {
                                            if (refDCval < 108) {
                                                if (AFD < 14) {
                                                    return 1;
                                                }
                                                else if (AFD >= 14) {
                                                    if (diffAFD < -2) {
                                                        return 1;
                                                    }
                                                    else if (diffAFD >= -2) {
                                                        return 2;
                                                    }
                                                }
                                            }
                                            else if (refDCval >= 108) {
                                                if (AFD < 19) {
                                                    return 4;
                                                }
                                                else if (AFD >= 19) {
                                                    return 1;
                                                }
                                            }
                                        }
                                    }
                                    else if (Cs >= 270) {
                                        if (refDCval < 122) {
                                            if (TSCindex < 9) {
                                                if (refDCval < 88) {
                                                    if (posBalance < 34) {
                                                        if (diffRsCsdiff < 112) {
                                                            if (mu_mv_mag_sq < 32) {
                                                                return 2;
                                                            }
                                                            else if (mu_mv_mag_sq >= 32) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (diffRsCsdiff >= 112) {
                                                            if (Rs < 2083) {
                                                                return 4;
                                                            }
                                                            else if (Rs >= 2083) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                    else if (posBalance >= 34) {
                                                        if (refDCval < 85) {
                                                            if (Rs < 736) {
                                                                if (Rs < 430) {
                                                                    if (mu_mv_mag_sq < 25) {
                                                                        if (Rs < 319) {
                                                                            return 2;
                                                                        }
                                                                        else if (Rs >= 319) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                    else if (mu_mv_mag_sq >= 25) {
                                                                        if (ssDCval < 38) {
                                                                            if (Cs < 326) {
                                                                                if (Cs < 293) {
                                                                                    return 4;
                                                                                }
                                                                                else if (Cs >= 293) {
                                                                                    if (mcTcor < 82) {
                                                                                        return 8;
                                                                                    }
                                                                                    else if (mcTcor >= 82) {
                                                                                        if (Cs < 300) {
                                                                                            return 8;
                                                                                        }
                                                                                        else if (Cs >= 300) {
                                                                                            return 4;
                                                                                        }
                                                                                    }
                                                                                }
                                                                            }
                                                                            else if (Cs >= 326) {
                                                                                if (negBalance < 43) {
                                                                                    return 8;
                                                                                }
                                                                                else if (negBalance >= 43) {
                                                                                    if (SC < 725) {
                                                                                        return 1;
                                                                                    }
                                                                                    else if (SC >= 725) {
                                                                                        if (Rs < 413) {
                                                                                            return 4;
                                                                                        }
                                                                                        else if (Rs >= 413) {
                                                                                            return 1;
                                                                                        }
                                                                                    }
                                                                                }
                                                                            }
                                                                        }
                                                                        else if (ssDCval >= 38) {
                                                                            if (mcTcor < 83) {
                                                                                return 1;
                                                                            }
                                                                            else if (mcTcor >= 83) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                                else if (Rs >= 430) {
                                                                    if (Cs < 383) {
                                                                        if (TSC < 6) {
                                                                            if (negBalance < 43) {
                                                                                if (Rs < 679) {
                                                                                    if (mu_mv_mag_sq < 79) {
                                                                                        return 4;
                                                                                    }
                                                                                    else if (mu_mv_mag_sq >= 79) {
                                                                                        if (Cs < 371) {
                                                                                            return 1;
                                                                                        }
                                                                                        else if (Cs >= 371) {
                                                                                            return 4;
                                                                                        }
                                                                                    }
                                                                                }
                                                                                else if (Rs >= 679) {
                                                                                    return 1;
                                                                                }
                                                                            }
                                                                            else if (negBalance >= 43) {
                                                                                if (mcTcor < 83) {
                                                                                    if (negBalance < 45) {
                                                                                        return 1;
                                                                                    }
                                                                                    else if (negBalance >= 45) {
                                                                                        if (SC < 812) {
                                                                                            return 1;
                                                                                        }
                                                                                        else if (SC >= 812) {
                                                                                            if (Cs < 370) {
                                                                                                return 1;
                                                                                            }
                                                                                            else if (Cs >= 370) {
                                                                                                return 4;
                                                                                            }
                                                                                        }
                                                                                    }
                                                                                }
                                                                                else if (mcTcor >= 83) {
                                                                                    if (Rs < 457) {
                                                                                        return 4;
                                                                                    }
                                                                                    else if (Rs >= 457) {
                                                                                        return 2;
                                                                                    }
                                                                                }
                                                                            }
                                                                        }
                                                                        else if (TSC >= 6) {
                                                                            if (AFD < 13) {
                                                                                return 8;
                                                                            }
                                                                            else if (AFD >= 13) {
                                                                                return 2;
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (Cs >= 383) {
                                                                        if (TSC < 6) {
                                                                            if (refDCval < 39) {
                                                                                if (diffMVdiffVal < 14) {
                                                                                    if (negBalance < 47) {
                                                                                        if (mu_mv_mag_sq < 76) {
                                                                                            if (mu_mv_mag_sq < 60) {
                                                                                                return 2;
                                                                                            }
                                                                                            else if (mu_mv_mag_sq >= 60) {
                                                                                                if (mu_mv_mag_sq < 61) {
                                                                                                    if (Rs < 450) {
                                                                                                        return 2;
                                                                                                    }
                                                                                                    else if (Rs >= 450) {
                                                                                                        return 4;
                                                                                                    }
                                                                                                }
                                                                                                else if (mu_mv_mag_sq >= 61) {
                                                                                                    if (gchDC < 1) {
                                                                                                        return 2;
                                                                                                    }
                                                                                                    else if (gchDC >= 1) {
                                                                                                        return 4;
                                                                                                    }
                                                                                                }
                                                                                            }
                                                                                        }
                                                                                        else if (mu_mv_mag_sq >= 76) {
                                                                                            if (Cs < 673) {
                                                                                                return 4;
                                                                                            }
                                                                                            else if (Cs >= 673) {
                                                                                                return 8;
                                                                                            }
                                                                                        }
                                                                                    }
                                                                                    else if (negBalance >= 47) {
                                                                                        if (SC < 1241) {
                                                                                            return 4;
                                                                                        }
                                                                                        else if (SC >= 1241) {
                                                                                            return 1;
                                                                                        }
                                                                                    }
                                                                                }
                                                                                else if (diffMVdiffVal >= 14) {
                                                                                    return 8;
                                                                                }
                                                                            }
                                                                            else if (refDCval >= 39) {
                                                                                if (AFD < 7) {
                                                                                    if (diffAFD < 2) {
                                                                                        return 4;
                                                                                    }
                                                                                    else if (diffAFD >= 2) {
                                                                                        return 8;
                                                                                    }
                                                                                }
                                                                                else if (AFD >= 7) {
                                                                                    if (Rs < 463) {
                                                                                        return 4;
                                                                                    }
                                                                                    else if (Rs >= 463) {
                                                                                        return 2;
                                                                                    }
                                                                                }
                                                                            }
                                                                        }
                                                                        else if (TSC >= 6) {
                                                                            if (tcor < 72) {
                                                                                if (Cs < 471) {
                                                                                    return 2;
                                                                                }
                                                                                else if (Cs >= 471) {
                                                                                    return 4;
                                                                                }
                                                                            }
                                                                            else if (tcor >= 72) {
                                                                                if (SC < 1152) {
                                                                                    if (SC < 1079) {
                                                                                        return 4;
                                                                                    }
                                                                                    else if (SC >= 1079) {
                                                                                        return 8;
                                                                                    }
                                                                                }
                                                                                else if (SC >= 1152) {
                                                                                    return 1;
                                                                                }
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                            else if (Rs >= 736) {
                                                                if (AFD < 6) {
                                                                    if (Cs < 634) {
                                                                        if (negBalance < 46) {
                                                                            if (CsDiff < 2) {
                                                                                return 4;
                                                                            }
                                                                            else if (CsDiff >= 2) {
                                                                                return 2;
                                                                            }
                                                                        }
                                                                        else if (negBalance >= 46) {
                                                                            if (mu_mv_mag_sq < 56) {
                                                                                return 1;
                                                                            }
                                                                            else if (mu_mv_mag_sq >= 56) {
                                                                                if (diffMVdiffVal < 13) {
                                                                                    return 4;
                                                                                }
                                                                                else if (diffMVdiffVal >= 13) {
                                                                                    return 1;
                                                                                }
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (Cs >= 634) {
                                                                        if (MVDiff < 20) {
                                                                            if (Cs < 1307) {
                                                                                return 1;
                                                                            }
                                                                            else if (Cs >= 1307) {
                                                                                return 8;
                                                                            }
                                                                        }
                                                                        else if (MVDiff >= 20) {
                                                                            if (Rs < 783) {
                                                                                return 1;
                                                                            }
                                                                            else if (Rs >= 783) {
                                                                                return 2;
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                                else if (AFD >= 6) {
                                                                    if (mu_mv_mag_sq < 42) {
                                                                        if (diffAFD < 3) {
                                                                            if (Rs < 1954) {
                                                                                if (negBalance < 29) {
                                                                                    return 4;
                                                                                }
                                                                                else if (negBalance >= 29) {
                                                                                    if (MVDiff < 7) {
                                                                                        if (diffMVdiffVal < 3) {
                                                                                            return 4;
                                                                                        }
                                                                                        else if (diffMVdiffVal >= 3) {
                                                                                            return 1;
                                                                                        }
                                                                                    }
                                                                                    else if (MVDiff >= 7) {
                                                                                        if (diffMVdiffVal < 6) {
                                                                                            if (negBalance < 53) {
                                                                                                if (diffRsCsdiff < -2) {
                                                                                                    if (diffRsCsdiff < -13) {
                                                                                                        if (diffMVdiffVal < 5) {
                                                                                                            return 1;
                                                                                                        }
                                                                                                        else if (diffMVdiffVal >= 5) {
                                                                                                            return 2;
                                                                                                        }
                                                                                                    }
                                                                                                    else if (diffRsCsdiff >= -13) {
                                                                                                        if (refDCval < 52) {
                                                                                                            if (MVDiff < 34) {
                                                                                                                return 1;
                                                                                                            }
                                                                                                            else if (MVDiff >= 34) {
                                                                                                                if (SC < 1341) {
                                                                                                                    return 1;
                                                                                                                }
                                                                                                                else if (SC >= 1341) {
                                                                                                                    return 4;
                                                                                                                }
                                                                                                            }
                                                                                                        }
                                                                                                        else if (refDCval >= 52) {
                                                                                                            if (Cs < 737) {
                                                                                                                return 8;
                                                                                                            }
                                                                                                            else if (Cs >= 737) {
                                                                                                                if (Cs < 992) {
                                                                                                                    return 4;
                                                                                                                }
                                                                                                                else if (Cs >= 992) {
                                                                                                                    return 8;
                                                                                                                }
                                                                                                            }
                                                                                                        }
                                                                                                    }
                                                                                                }
                                                                                                else if (diffRsCsdiff >= -2) {
                                                                                                    if (mu_mv_mag_sq < 3) {
                                                                                                        if (TSC < 5) {
                                                                                                            return 1;
                                                                                                        }
                                                                                                        else if (TSC >= 5) {
                                                                                                            return 4;
                                                                                                        }
                                                                                                    }
                                                                                                    else if (mu_mv_mag_sq >= 3) {
                                                                                                        if (mcTcor < 75) {
                                                                                                            if (Rs < 784) {
                                                                                                                return 4;
                                                                                                            }
                                                                                                            else if (Rs >= 784) {
                                                                                                                return 1;
                                                                                                            }
                                                                                                        }
                                                                                                        else if (mcTcor >= 75) {
                                                                                                            return 1;
                                                                                                        }
                                                                                                    }
                                                                                                }
                                                                                            }
                                                                                            else if (negBalance >= 53) {
                                                                                                if (SC < 1568) {
                                                                                                    return 4;
                                                                                                }
                                                                                                else if (SC >= 1568) {
                                                                                                    return 8;
                                                                                                }
                                                                                            }
                                                                                        }
                                                                                        else if (diffMVdiffVal >= 6) {
                                                                                            if (refDCval < 51) {
                                                                                                if (Cs < 552) {
                                                                                                    if (Rs < 777) {
                                                                                                        return 1;
                                                                                                    }
                                                                                                    else if (Rs >= 777) {
                                                                                                        return 4;
                                                                                                    }
                                                                                                }
                                                                                                else if (Cs >= 552) {
                                                                                                    if (tcor < 78) {
                                                                                                        return 4;
                                                                                                    }
                                                                                                    else if (tcor >= 78) {
                                                                                                        if (diffRsCsdiff < 3) {
                                                                                                            return 1;
                                                                                                        }
                                                                                                        else if (diffRsCsdiff >= 3) {
                                                                                                            return 4;
                                                                                                        }
                                                                                                    }
                                                                                                }
                                                                                            }
                                                                                            else if (refDCval >= 51) {
                                                                                                if (AFD < 11) {
                                                                                                    if (mcTcor < 91) {
                                                                                                        return 1;
                                                                                                    }
                                                                                                    else if (mcTcor >= 91) {
                                                                                                        return 8;
                                                                                                    }
                                                                                                }
                                                                                                else if (AFD >= 11) {
                                                                                                    return 1;
                                                                                                }
                                                                                            }
                                                                                        }
                                                                                    }
                                                                                }
                                                                            }
                                                                            else if (Rs >= 1954) {
                                                                                if (Cs < 1712) {
                                                                                    return 4;
                                                                                }
                                                                                else if (Cs >= 1712) {
                                                                                    return 8;
                                                                                }
                                                                            }
                                                                        }
                                                                        else if (diffAFD >= 3) {
                                                                            if (Rs < 1187) {
                                                                                if (Rs < 868) {
                                                                                    if (Rs < 756) {
                                                                                        return 8;
                                                                                    }
                                                                                    else if (Rs >= 756) {
                                                                                        return 2;
                                                                                    }
                                                                                }
                                                                                else if (Rs >= 868) {
                                                                                    return 8;
                                                                                }
                                                                            }
                                                                            else if (Rs >= 1187) {
                                                                                return 1;
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (mu_mv_mag_sq >= 42) {
                                                                        if (posBalance < 63) {
                                                                            if (AFD < 8) {
                                                                                if (gchDC < 1) {
                                                                                    if (diffRsCsdiff < 5) {
                                                                                        if (mu_mv_mag_sq < 59) {
                                                                                            if (SC < 1373) {
                                                                                                return 4;
                                                                                            }
                                                                                            else if (SC >= 1373) {
                                                                                                if (Cs < 602) {
                                                                                                    return 1;
                                                                                                }
                                                                                                else if (Cs >= 602) {
                                                                                                    if (diffAFD < 1) {
                                                                                                        return 4;
                                                                                                    }
                                                                                                    else if (diffAFD >= 1) {
                                                                                                        return 1;
                                                                                                    }
                                                                                                }
                                                                                            }
                                                                                        }
                                                                                        else if (mu_mv_mag_sq >= 59) {
                                                                                            return 4;
                                                                                        }
                                                                                    }
                                                                                    else if (diffRsCsdiff >= 5) {
                                                                                        return 1;
                                                                                    }
                                                                                }
                                                                                else if (gchDC >= 1) {
                                                                                    if (refDCval < 51) {
                                                                                        if (Cs < 961) {
                                                                                            return 8;
                                                                                        }
                                                                                        else if (Cs >= 961) {
                                                                                            return 4;
                                                                                        }
                                                                                    }
                                                                                    else if (refDCval >= 51) {
                                                                                        return 1;
                                                                                    }
                                                                                }
                                                                            }
                                                                            else if (AFD >= 8) {
                                                                                if (gchDC < 2) {
                                                                                    if (Rs < 972) {
                                                                                        return 8;
                                                                                    }
                                                                                    else if (Rs >= 972) {
                                                                                        return 4;
                                                                                    }
                                                                                }
                                                                                else if (gchDC >= 2) {
                                                                                    if (AFD < 9) {
                                                                                        return 8;
                                                                                    }
                                                                                    else if (AFD >= 9) {
                                                                                        return 2;
                                                                                    }
                                                                                }
                                                                            }
                                                                        }
                                                                        else if (posBalance >= 63) {
                                                                            if (RsDiff < 8) {
                                                                                return 2;
                                                                            }
                                                                            else if (RsDiff >= 8) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (refDCval >= 85) {
                                                            if (mcTcor < 91) {
                                                                return 2;
                                                            }
                                                            else if (mcTcor >= 91) {
                                                                return 1;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (refDCval >= 88) {
                                                    if (RsDiff < 3) {
                                                        if (AFD < 8) {
                                                            return 4;
                                                        }
                                                        else if (AFD >= 8) {
                                                            if (AFD < 10) {
                                                                if (Cs < 379) {
                                                                    return 4;
                                                                }
                                                                else if (Cs >= 379) {
                                                                    if (Rs < 589) {
                                                                        return 8;
                                                                    }
                                                                    else if (Rs >= 589) {
                                                                        if (SC < 1114) {
                                                                            return 8;
                                                                        }
                                                                        else if (SC >= 1114) {
                                                                            if (negBalance < 54) {
                                                                                if (Rs < 687) {
                                                                                    return 8;
                                                                                }
                                                                                else if (Rs >= 687) {
                                                                                    return 4;
                                                                                }
                                                                            }
                                                                            else if (negBalance >= 54) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                            else if (AFD >= 10) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                    else if (RsDiff >= 3) {
                                                        if (Rs < 1325) {
                                                            if (Cs < 385) {
                                                                if (SC < 1143) {
                                                                    return 2;
                                                                }
                                                                else if (SC >= 1143) {
                                                                    return 1;
                                                                }
                                                            }
                                                            else if (Cs >= 385) {
                                                                if (negBalance < 52) {
                                                                    if (tcor < 70) {
                                                                        if (Rs < 704) {
                                                                            return 4;
                                                                        }
                                                                        else if (Rs >= 704) {
                                                                            return 2;
                                                                        }
                                                                    }
                                                                    else if (tcor >= 70) {
                                                                        if (SC < 2145) {
                                                                            return 4;
                                                                        }
                                                                        else if (SC >= 2145) {
                                                                            return 2;
                                                                        }
                                                                    }
                                                                }
                                                                else if (negBalance >= 52) {
                                                                    if (Rs < 622) {
                                                                        return 8;
                                                                    }
                                                                    else if (Rs >= 622) {
                                                                        return 4;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (Rs >= 1325) {
                                                            if (ssDCval < 99) {
                                                                if (RsDiff < 6) {
                                                                    if (SC < 2949) {
                                                                        if (negBalance < 54) {
                                                                            return 2;
                                                                        }
                                                                        else if (negBalance >= 54) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                    else if (SC >= 2949) {
                                                                        if (diffRsCsdiff < -16) {
                                                                            return 2;
                                                                        }
                                                                        else if (diffRsCsdiff >= -16) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                }
                                                                else if (RsDiff >= 6) {
                                                                    if (diffMVdiffVal < -1) {
                                                                        return 8;
                                                                    }
                                                                    else if (diffMVdiffVal >= -1) {
                                                                        return 2;
                                                                    }
                                                                }
                                                            }
                                                            else if (ssDCval >= 99) {
                                                                if (mcTcor < 89) {
                                                                    return 4;
                                                                }
                                                                else if (mcTcor >= 89) {
                                                                    if (TSC < 5) {
                                                                        return 2;
                                                                    }
                                                                    else if (TSC >= 5) {
                                                                        if (Rs < 4626) {
                                                                            return 8;
                                                                        }
                                                                        else if (Rs >= 4626) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                            else if (TSCindex >= 9) {
                                                if (RsCsDiff < 63) {
                                                    if (gchDC < 2) {
                                                        if (Rs < 1907) {
                                                            if (MVDiff < 11) {
                                                                if (mu_mv_mag_sq < 48) {
                                                                    if (diffRsCsdiff < -4) {
                                                                        if (RsDiff < 5) {
                                                                            return 4;
                                                                        }
                                                                        else if (RsDiff >= 5) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                    else if (diffRsCsdiff >= -4) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (mu_mv_mag_sq >= 48) {
                                                                    if (RsDiff < 6) {
                                                                        return 8;
                                                                    }
                                                                    else if (RsDiff >= 6) {
                                                                        return 2;
                                                                    }
                                                                }
                                                            }
                                                            else if (MVDiff >= 11) {
                                                                if (diffMVdiffVal < 23) {
                                                                    if (TSC < 9) {
                                                                        if (ssDCval < 93) {
                                                                            if (RsDiff < 5) {
                                                                                return 8;
                                                                            }
                                                                            else if (RsDiff >= 5) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                        else if (ssDCval >= 93) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                    else if (TSC >= 9) {
                                                                        return 8;
                                                                    }
                                                                }
                                                                else if (diffMVdiffVal >= 23) {
                                                                    if (Rs < 1011) {
                                                                        return 4;
                                                                    }
                                                                    else if (Rs >= 1011) {
                                                                        return 1;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (Rs >= 1907) {
                                                            if (negBalance < 53) {
                                                                if (posBalance < 58) {
                                                                    return 8;
                                                                }
                                                                else if (posBalance >= 58) {
                                                                    if (RsDiff < 6) {
                                                                        if (AFD < 10) {
                                                                            return 8;
                                                                        }
                                                                        else if (AFD >= 10) {
                                                                            return 2;
                                                                        }
                                                                    }
                                                                    else if (RsDiff >= 6) {
                                                                        return 2;
                                                                    }
                                                                }
                                                            }
                                                            else if (negBalance >= 53) {
                                                                return 2;
                                                            }
                                                        }
                                                    }
                                                    else if (gchDC >= 2) {
                                                        if (diffRsCsdiff < 18) {
                                                            if (MVDiff < 70) {
                                                                return 1;
                                                            }
                                                            else if (MVDiff >= 70) {
                                                                if (SC < 1219) {
                                                                    return 1;
                                                                }
                                                                else if (SC >= 1219) {
                                                                    if (Rs < 603) {
                                                                        return 1;
                                                                    }
                                                                    else if (Rs >= 603) {
                                                                        return 8;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (diffRsCsdiff >= 18) {
                                                            if (Rs < 389) {
                                                                return 2;
                                                            }
                                                            else if (Rs >= 389) {
                                                                if (Rs < 1307) {
                                                                    return 1;
                                                                }
                                                                else if (Rs >= 1307) {
                                                                    if (Rs < 2283) {
                                                                        return 4;
                                                                    }
                                                                    else if (Rs >= 2283) {
                                                                        return 8;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (RsCsDiff >= 63) {
                                                    if (Cs < 1407) {
                                                        if (refDCval < 84) {
                                                            if (diffTSC < 4) {
                                                                if (AFD < 15) {
                                                                    if (posBalance < 56) {
                                                                        return 8;
                                                                    }
                                                                    else if (posBalance >= 56) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (AFD >= 15) {
                                                                    return 2;
                                                                }
                                                            }
                                                            else if (diffTSC >= 4) {
                                                                if (Rs < 1287) {
                                                                    return 8;
                                                                }
                                                                else if (Rs >= 1287) {
                                                                    return 1;
                                                                }
                                                            }
                                                        }
                                                        else if (refDCval >= 84) {
                                                            if (Rs < 2065) {
                                                                if (Cs < 422) {
                                                                    return 4;
                                                                }
                                                                else if (Cs >= 422) {
                                                                    return 2;
                                                                }
                                                            }
                                                            else if (Rs >= 2065) {
                                                                if (Rs < 2162) {
                                                                    if (Rs < 2087) {
                                                                        return 1;
                                                                    }
                                                                    else if (Rs >= 2087) {
                                                                        return 2;
                                                                    }
                                                                }
                                                                else if (Rs >= 2162) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (Cs >= 1407) {
                                                        if (refDCval < 116) {
                                                            if (refDCval < 82) {
                                                                if (mu_mv_mag_sq < 7) {
                                                                    if (negBalance < 45) {
                                                                        if (Rs < 2025) {
                                                                            return 1;
                                                                        }
                                                                        else if (Rs >= 2025) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                    else if (negBalance >= 45) {
                                                                        if (SC < 3630) {
                                                                            return 4;
                                                                        }
                                                                        else if (SC >= 3630) {
                                                                            if (RsCsDiff < 121) {
                                                                                return 8;
                                                                            }
                                                                            else if (RsCsDiff >= 121) {
                                                                                if (MVDiff < 15) {
                                                                                    return 4;
                                                                                }
                                                                                else if (MVDiff >= 15) {
                                                                                    if (CsDiff < 9) {
                                                                                        return 4;
                                                                                    }
                                                                                    else if (CsDiff >= 9) {
                                                                                        if (Rs < 2469) {
                                                                                            return 8;
                                                                                        }
                                                                                        else if (Rs >= 2469) {
                                                                                            if (Rs < 2548) {
                                                                                                return 4;
                                                                                            }
                                                                                            else if (Rs >= 2548) {
                                                                                                if (Cs < 2308) {
                                                                                                    return 8;
                                                                                                }
                                                                                                else if (Cs >= 2308) {
                                                                                                    return 4;
                                                                                                }
                                                                                            }
                                                                                        }
                                                                                    }
                                                                                }
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                                else if (mu_mv_mag_sq >= 7) {
                                                                    if (mu_mv_mag_sq < 64) {
                                                                        if (diffTSC < 1) {
                                                                            if (mcTcor < 87) {
                                                                                return 8;
                                                                            }
                                                                            else if (mcTcor >= 87) {
                                                                                if (Rs < 2203) {
                                                                                    if (MVDiff < 18) {
                                                                                        if (negBalance < 51) {
                                                                                            return 8;
                                                                                        }
                                                                                        else if (negBalance >= 51) {
                                                                                            return 1;
                                                                                        }
                                                                                    }
                                                                                    else if (MVDiff >= 18) {
                                                                                        return 1;
                                                                                    }
                                                                                }
                                                                                else if (Rs >= 2203) {
                                                                                    if (MVDiff < 19) {
                                                                                        return 1;
                                                                                    }
                                                                                    else if (MVDiff >= 19) {
                                                                                        return 2;
                                                                                    }
                                                                                }
                                                                            }
                                                                        }
                                                                        else if (diffTSC >= 1) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                    else if (mu_mv_mag_sq >= 64) {
                                                                        if (mu_mv_mag_sq < 85) {
                                                                            return 8;
                                                                        }
                                                                        else if (mu_mv_mag_sq >= 85) {
                                                                            if (Rs < 2417) {
                                                                                return 8;
                                                                            }
                                                                            else if (Rs >= 2417) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                            else if (refDCval >= 82) {
                                                                if (Rs < 2453) {
                                                                    if (RsCsDiff < 92) {
                                                                        if (tcor < 85) {
                                                                            if (Cs < 1745) {
                                                                                return 4;
                                                                            }
                                                                            else if (Cs >= 1745) {
                                                                                return 2;
                                                                            }
                                                                        }
                                                                        else if (tcor >= 85) {
                                                                            if (diffMVdiffVal < -2) {
                                                                                return 2;
                                                                            }
                                                                            else if (diffMVdiffVal >= -2) {
                                                                                if (gchDC < 1) {
                                                                                    return 8;
                                                                                }
                                                                                else if (gchDC >= 1) {
                                                                                    if (Rs < 2192) {
                                                                                        return 8;
                                                                                    }
                                                                                    else if (Rs >= 2192) {
                                                                                        if (Rs < 2364) {
                                                                                            return 2;
                                                                                        }
                                                                                        else if (Rs >= 2364) {
                                                                                            return 1;
                                                                                        }
                                                                                    }
                                                                                }
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (RsCsDiff >= 92) {
                                                                        if (posBalance < 63) {
                                                                            if (mcTcor < 92) {
                                                                                if (diffTSC < 1) {
                                                                                    return 4;
                                                                                }
                                                                                else if (diffTSC >= 1) {
                                                                                    if (TSC < 10) {
                                                                                        return 8;
                                                                                    }
                                                                                    else if (TSC >= 10) {
                                                                                        return 4;
                                                                                    }
                                                                                }
                                                                            }
                                                                            else if (mcTcor >= 92) {
                                                                                if (posBalance < 56) {
                                                                                    if (Rs < 2336) {
                                                                                        return 8;
                                                                                    }
                                                                                    else if (Rs >= 2336) {
                                                                                        return 1;
                                                                                    }
                                                                                }
                                                                                else if (posBalance >= 56) {
                                                                                    return 4;
                                                                                }
                                                                            }
                                                                        }
                                                                        else if (posBalance >= 63) {
                                                                            if (Rs < 2034) {
                                                                                return 4;
                                                                            }
                                                                            else if (Rs >= 2034) {
                                                                                return 2;
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                                else if (Rs >= 2453) {
                                                                    if (TSC < 11) {
                                                                        if (posBalance < 47) {
                                                                            if (SC < 7008) {
                                                                                return 8;
                                                                            }
                                                                            else if (SC >= 7008) {
                                                                                if (MVDiff < 11) {
                                                                                    if (diffTSC < 1) {
                                                                                        if (TSC < 10) {
                                                                                            return 8;
                                                                                        }
                                                                                        else if (TSC >= 10) {
                                                                                            return 2;
                                                                                        }
                                                                                    }
                                                                                    else if (diffTSC >= 1) {
                                                                                        if (diffMVdiffVal < 1) {
                                                                                            if (Rs < 5230) {
                                                                                                return 8;
                                                                                            }
                                                                                            else if (Rs >= 5230) {
                                                                                                return 2;
                                                                                            }
                                                                                        }
                                                                                        else if (diffMVdiffVal >= 1) {
                                                                                            return 2;
                                                                                        }
                                                                                    }
                                                                                }
                                                                                else if (MVDiff >= 11) {
                                                                                    if (Cs < 2695) {
                                                                                        return 4;
                                                                                    }
                                                                                    else if (Cs >= 2695) {
                                                                                        if (TSC < 9) {
                                                                                            return 8;
                                                                                        }
                                                                                        else if (TSC >= 9) {
                                                                                            return 1;
                                                                                        }
                                                                                    }
                                                                                }
                                                                            }
                                                                        }
                                                                        else if (posBalance >= 47) {
                                                                            if (tcor < 79) {
                                                                                return 4;
                                                                            }
                                                                            else if (tcor >= 79) {
                                                                                if (diffRsCsdiff < -37) {
                                                                                    return 4;
                                                                                }
                                                                                else if (diffRsCsdiff >= -37) {
                                                                                    if (Rs < 2833) {
                                                                                        if (refDCval < 97) {
                                                                                            return 2;
                                                                                        }
                                                                                        else if (refDCval >= 97) {
                                                                                            if (Rs < 2568) {
                                                                                                return 8;
                                                                                            }
                                                                                            else if (Rs >= 2568) {
                                                                                                return 1;
                                                                                            }
                                                                                        }
                                                                                    }
                                                                                    else if (Rs >= 2833) {
                                                                                        if (RsCsDiff < 157) {
                                                                                            if (SC < 8238) {
                                                                                                if (ssDCval < 110) {
                                                                                                    return 4;
                                                                                                }
                                                                                                else if (ssDCval >= 110) {
                                                                                                    if (ssDCval < 116) {
                                                                                                        if (RsCsDiff < 133) {
                                                                                                            return 2;
                                                                                                        }
                                                                                                        else if (RsCsDiff >= 133) {
                                                                                                            return 4;
                                                                                                        }
                                                                                                    }
                                                                                                    else if (ssDCval >= 116) {
                                                                                                        return 4;
                                                                                                    }
                                                                                                }
                                                                                            }
                                                                                            else if (SC >= 8238) {
                                                                                                if (Cs < 3215) {
                                                                                                    return 8;
                                                                                                }
                                                                                                else if (Cs >= 3215) {
                                                                                                    return 2;
                                                                                                }
                                                                                            }
                                                                                        }
                                                                                        else if (RsCsDiff >= 157) {
                                                                                            return 2;
                                                                                        }
                                                                                    }
                                                                                }
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (TSC >= 11) {
                                                                        if (mu_mv_mag_sq < 23) {
                                                                            if (AFD < 17) {
                                                                                if (Rs < 3836) {
                                                                                    return 8;
                                                                                }
                                                                                else if (Rs >= 3836) {
                                                                                    return 2;
                                                                                }
                                                                            }
                                                                            else if (AFD >= 17) {
                                                                                return 1;
                                                                            }
                                                                        }
                                                                        else if (mu_mv_mag_sq >= 23) {
                                                                            if (Rs < 4577) {
                                                                                if (refDCval < 92) {
                                                                                    if (mu_mv_mag_sq < 46) {
                                                                                        if (ssDCval < 87) {
                                                                                            if (Rs < 2602) {
                                                                                                return 4;
                                                                                            }
                                                                                            else if (Rs >= 2602) {
                                                                                                if (RsCsDiff < 107) {
                                                                                                    return 2;
                                                                                                }
                                                                                                else if (RsCsDiff >= 107) {
                                                                                                    if (gchDC < 2) {
                                                                                                        return 1;
                                                                                                    }
                                                                                                    else if (gchDC >= 2) {
                                                                                                        return 2;
                                                                                                    }
                                                                                                }
                                                                                            }
                                                                                        }
                                                                                        else if (ssDCval >= 87) {
                                                                                            return 2;
                                                                                        }
                                                                                    }
                                                                                    else if (mu_mv_mag_sq >= 46) {
                                                                                        return 1;
                                                                                    }
                                                                                }
                                                                                else if (refDCval >= 92) {
                                                                                    if (MVDiff < 10) {
                                                                                        if (SC < 5203) {
                                                                                            return 1;
                                                                                        }
                                                                                        else if (SC >= 5203) {
                                                                                            return 4;
                                                                                        }
                                                                                    }
                                                                                    else if (MVDiff >= 10) {
                                                                                        return 1;
                                                                                    }
                                                                                }
                                                                            }
                                                                            else if (Rs >= 4577) {
                                                                                if (ssDCval < 107) {
                                                                                    return 4;
                                                                                }
                                                                                else if (ssDCval >= 107) {
                                                                                    return 2;
                                                                                }
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (refDCval >= 116) {
                                                            if (TSC < 13) {
                                                                if (mu_mv_mag_sq < 80) {
                                                                    if (mcTcor < 93) {
                                                                        if (MVDiff < 7) {
                                                                            if (negBalance < 53) {
                                                                                if (SC < 4210) {
                                                                                    return 8;
                                                                                }
                                                                                else if (SC >= 4210) {
                                                                                    if (RsCsDiff < 107) {
                                                                                        return 4;
                                                                                    }
                                                                                    else if (RsCsDiff >= 107) {
                                                                                        return 8;
                                                                                    }
                                                                                }
                                                                            }
                                                                            else if (negBalance >= 53) {
                                                                                return 8;
                                                                            }
                                                                        }
                                                                        else if (MVDiff >= 7) {
                                                                            if (MVDiff < 8) {
                                                                                if (diffMVdiffVal < 3) {
                                                                                    return 4;
                                                                                }
                                                                                else if (diffMVdiffVal >= 3) {
                                                                                    return 8;
                                                                                }
                                                                            }
                                                                            else if (MVDiff >= 8) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (mcTcor >= 93) {
                                                                        return 8;
                                                                    }
                                                                }
                                                                else if (mu_mv_mag_sq >= 80) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (TSC >= 13) {
                                                                if (Rs < 3098) {
                                                                    return 1;
                                                                }
                                                                else if (Rs >= 3098) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (refDCval >= 122) {
                                            if (TSC < 9) {
                                                if (refDCval < 139) {
                                                    if (diffRsCsdiff < -274) {
                                                        return 8;
                                                    }
                                                    else if (diffRsCsdiff >= -274) {
                                                        if (SC < 2643) {
                                                            return 1;
                                                        }
                                                        else if (SC >= 2643) {
                                                            if (mcTcor < 88) {
                                                                if (refDCval < 126) {
                                                                    if (posBalance < 58) {
                                                                        return 1;
                                                                    }
                                                                    else if (posBalance >= 58) {
                                                                        if (SC < 4987) {
                                                                            return 4;
                                                                        }
                                                                        else if (SC >= 4987) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                }
                                                                else if (refDCval >= 126) {
                                                                    return 1;
                                                                }
                                                            }
                                                            else if (mcTcor >= 88) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (refDCval >= 139) {
                                                    if (ssDCval < 142) {
                                                        if (Cs < 756) {
                                                            if (Rs < 816) {
                                                                return 8;
                                                            }
                                                            else if (Rs >= 816) {
                                                                return 1;
                                                            }
                                                        }
                                                        else if (Cs >= 756) {
                                                            if (mcTcor < 90) {
                                                                return 8;
                                                            }
                                                            else if (mcTcor >= 90) {
                                                                return 1;
                                                            }
                                                        }
                                                    }
                                                    else if (ssDCval >= 142) {
                                                        if (diffTSC < -2) {
                                                            if (Rs < 1046) {
                                                                return 8;
                                                            }
                                                            else if (Rs >= 1046) {
                                                                return 1;
                                                            }
                                                        }
                                                        else if (diffTSC >= -2) {
                                                            if (diffAFD < 2) {
                                                                if (SC < 3357) {
                                                                    if (TSC < 2) {
                                                                        return 8;
                                                                    }
                                                                    else if (TSC >= 2) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (SC >= 3357) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (diffAFD >= 2) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                            else if (TSC >= 9) {
                                                if (CsDiff < 5) {
                                                    if (SC < 2175) {
                                                        return 1;
                                                    }
                                                    else if (SC >= 2175) {
                                                        if (CsDiff < 3) {
                                                            return 4;
                                                        }
                                                        else if (CsDiff >= 3) {
                                                            return 2;
                                                        }
                                                    }
                                                }
                                                else if (CsDiff >= 5) {
                                                    if (ssDCval < 139) {
                                                        if (diffMVdiffVal < -5) {
                                                            return 8;
                                                        }
                                                        else if (diffMVdiffVal >= -5) {
                                                            return 1;
                                                        }
                                                    }
                                                    else if (ssDCval >= 139) {
                                                        if (Cs < 1467) {
                                                            return 2;
                                                        }
                                                        else if (Cs >= 1467) {
                                                            if (Cs < 1732) {
                                                                return 4;
                                                            }
                                                            else if (Cs >= 1732) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (RsCsDiff >= 194) {
                                    if (AFD < 14) {
                                        if (CsDiff < 9) {
                                            if (negBalance < 38) {
                                                if (Rs < 4697) {
                                                    return 8;
                                                }
                                                else if (Rs >= 4697) {
                                                    return 4;
                                                }
                                            }
                                            else if (negBalance >= 38) {
                                                return 4;
                                            }
                                        }
                                        else if (CsDiff >= 9) {
                                            if (refDCval < 129) {
                                                if (posBalance < 58) {
                                                    if (Cs < 3009) {
                                                        return 2;
                                                    }
                                                    else if (Cs >= 3009) {
                                                        return 4;
                                                    }
                                                }
                                                else if (posBalance >= 58) {
                                                    return 4;
                                                }
                                            }
                                            else if (refDCval >= 129) {
                                                return 1;
                                            }
                                        }
                                    }
                                    else if (AFD >= 14) {
                                        if (tcor < 79) {
                                            if (AFD < 28) {
                                                if (tcor < 69) {
                                                    if (refDCval < 107) {
                                                        if (SC < 3553) {
                                                            return 1;
                                                        }
                                                        else if (SC >= 3553) {
                                                            if (Rs < 2642) {
                                                                return 4;
                                                            }
                                                            else if (Rs >= 2642) {
                                                                return 2;
                                                            }
                                                        }
                                                    }
                                                    else if (refDCval >= 107) {
                                                        return 1;
                                                    }
                                                }
                                                else if (tcor >= 69) {
                                                    if (Cs < 1487) {
                                                        if (Rs < 1031) {
                                                            return 8;
                                                        }
                                                        else if (Rs >= 1031) {
                                                            return 2;
                                                        }
                                                    }
                                                    else if (Cs >= 1487) {
                                                        if (posBalance < 54) {
                                                            if (MVDiff < 14) {
                                                                return 2;
                                                            }
                                                            else if (MVDiff >= 14) {
                                                                if (CsDiff < 8) {
                                                                    return 2;
                                                                }
                                                                else if (CsDiff >= 8) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                        else if (posBalance >= 54) {
                                                            if (mcTcor < 84) {
                                                                return 4;
                                                            }
                                                            else if (mcTcor >= 84) {
                                                                if (Cs < 2176) {
                                                                    return 1;
                                                                }
                                                                else if (Cs >= 2176) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                            else if (AFD >= 28) {
                                                if (CsDiff < 9) {
                                                    return 2;
                                                }
                                                else if (CsDiff >= 9) {
                                                    if (SC < 5543) {
                                                        if (TSC < 11) {
                                                            return 2;
                                                        }
                                                        else if (TSC >= 11) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (SC >= 5543) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                        }
                                        else if (tcor >= 79) {
                                            if (posBalance < 70) {
                                                if (diffAFD < 3) {
                                                    if (RsDiff < 18) {
                                                        if (TSC < 13) {
                                                            return 2;
                                                        }
                                                        else if (TSC >= 13) {
                                                            if (Rs < 2536) {
                                                                return 2;
                                                            }
                                                            else if (Rs >= 2536) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                    else if (RsDiff >= 18) {
                                                        return 4;
                                                    }
                                                }
                                                else if (diffAFD >= 3) {
                                                    return 8;
                                                }
                                            }
                                            else if (posBalance >= 70) {
                                                if (Cs < 1547) {
                                                    return 8;
                                                }
                                                else if (Cs >= 1547) {
                                                    return 4;
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                            else if (negBalance >= 55) {
                                if (AFD < 14) {
                                    if (Rs < 792) {
                                        if (TSC < 6) {
                                            if (AFD < 10) {
                                                if (diffMVdiffVal < 213) {
                                                    return 1;
                                                }
                                                else if (diffMVdiffVal >= 213) {
                                                    if (mcTcor < 40) {
                                                        return 4;
                                                    }
                                                    else if (mcTcor >= 40) {
                                                        return 2;
                                                    }
                                                }
                                            }
                                            else if (AFD >= 10) {
                                                if (Rs < 84) {
                                                    return 1;
                                                }
                                                else if (Rs >= 84) {
                                                    return 2;
                                                }
                                            }
                                        }
                                        else if (TSC >= 6) {
                                            if (ssDCval < 140) {
                                                return 1;
                                            }
                                            else if (ssDCval >= 140) {
                                                return 2;
                                            }
                                        }
                                    }
                                    else if (Rs >= 792) {
                                        if (AFD < 12) {
                                            if (SC < 4451) {
                                                return 4;
                                            }
                                            else if (SC >= 4451) {
                                                return 8;
                                            }
                                        }
                                        else if (AFD >= 12) {
                                            if (mcTcor < 90) {
                                                if (diffRsCsdiff < -3) {
                                                    if (gchDC < 1) {
                                                        return 4;
                                                    }
                                                    else if (gchDC >= 1) {
                                                        return 2;
                                                    }
                                                }
                                                else if (diffRsCsdiff >= -3) {
                                                    if (MVDiff < 11) {
                                                        if (Cs < 976) {
                                                            return 4;
                                                        }
                                                        else if (Cs >= 976) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (MVDiff >= 11) {
                                                        if (Rs < 2749) {
                                                            return 1;
                                                        }
                                                        else if (Rs >= 2749) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (mcTcor >= 90) {
                                                if (TSCindex < 8) {
                                                    return 2;
                                                }
                                                else if (TSCindex >= 8) {
                                                    if (SC < 4830) {
                                                        return 1;
                                                    }
                                                    else if (SC >= 4830) {
                                                        return 2;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (AFD >= 14) {
                                    if (ssDCval < 75) {
                                        if (TSC < 9) {
                                            if (Cs < 1036) {
                                                return 2;
                                            }
                                            else if (Cs >= 1036) {
                                                return 8;
                                            }
                                        }
                                        else if (TSC >= 9) {
                                            if (RsDiff < 8) {
                                                if (TSC < 10) {
                                                    if (Rs < 965) {
                                                        return 8;
                                                    }
                                                    else if (Rs >= 965) {
                                                        return 4;
                                                    }
                                                }
                                                else if (TSC >= 10) {
                                                    if (mcTcor < 85) {
                                                        return 8;
                                                    }
                                                    else if (mcTcor >= 85) {
                                                        if (diffMVdiffVal < 2) {
                                                            return 8;
                                                        }
                                                        else if (diffMVdiffVal >= 2) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (RsDiff >= 8) {
                                                if (gchDC < 2) {
                                                    if (mcTcor < 82) {
                                                        return 8;
                                                    }
                                                    else if (mcTcor >= 82) {
                                                        if (RsDiff < 10) {
                                                            return 4;
                                                        }
                                                        else if (RsDiff >= 10) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                                else if (gchDC >= 2) {
                                                    if (Rs < 924) {
                                                        return 2;
                                                    }
                                                    else if (Rs >= 924) {
                                                        if (Cs < 1270) {
                                                            return 4;
                                                        }
                                                        else if (Cs >= 1270) {
                                                            return 2;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (ssDCval >= 75) {
                                        if (AFD < 17) {
                                            if (CsDiff < 5) {
                                                if (Cs < 514) {
                                                    if (SC < 1857) {
                                                        return 2;
                                                    }
                                                    else if (SC >= 1857) {
                                                        return 4;
                                                    }
                                                }
                                                else if (Cs >= 514) {
                                                    return 2;
                                                }
                                            }
                                            else if (CsDiff >= 5) {
                                                if (Cs < 1437) {
                                                    return 2;
                                                }
                                                else if (Cs >= 1437) {
                                                    if (diffRsCsdiff < -15) {
                                                        if (tcor < 86) {
                                                            if (RsCsDiff < 103) {
                                                                return 4;
                                                            }
                                                            else if (RsCsDiff >= 103) {
                                                                if (diffTSC < -1) {
                                                                    return 1;
                                                                }
                                                                else if (diffTSC >= -1) {
                                                                    return 2;
                                                                }
                                                            }
                                                        }
                                                        else if (tcor >= 86) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (diffRsCsdiff >= -15) {
                                                        if (mcTcor < 90) {
                                                            return 4;
                                                        }
                                                        else if (mcTcor >= 90) {
                                                            return 1;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (AFD >= 17) {
                                            if (RsDiff < 11) {
                                                if (ssDCval < 78) {
                                                    if (Cs < 1111) {
                                                        return 2;
                                                    }
                                                    else if (Cs >= 1111) {
                                                        return 4;
                                                    }
                                                }
                                                else if (ssDCval >= 78) {
                                                    if (TSC < 13) {
                                                        if (RsDiff < 10) {
                                                            if (mu_mv_mag_sq < 23) {
                                                                if (mu_mv_mag_sq < 2) {
                                                                    return 1;
                                                                }
                                                                else if (mu_mv_mag_sq >= 2) {
                                                                    return 2;
                                                                }
                                                            }
                                                            else if (mu_mv_mag_sq >= 23) {
                                                                return 1;
                                                            }
                                                        }
                                                        else if (RsDiff >= 10) {
                                                            if (AFD < 19) {
                                                                return 4;
                                                            }
                                                            else if (AFD >= 19) {
                                                                if (MVDiff < 16) {
                                                                    return 1;
                                                                }
                                                                else if (MVDiff >= 16) {
                                                                    return 2;
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (TSC >= 13) {
                                                        if (diffMVdiffVal < 4) {
                                                            return 1;
                                                        }
                                                        else if (diffMVdiffVal >= 4) {
                                                            if (Cs < 814) {
                                                                if (SC < 1892) {
                                                                    return 1;
                                                                }
                                                                else if (SC >= 1892) {
                                                                    return 2;
                                                                }
                                                            }
                                                            else if (Cs >= 814) {
                                                                return 1;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                            else if (RsDiff >= 11) {
                                                if (AFD < 31) {
                                                    if (SC < 4736) {
                                                        if (tcor < 65) {
                                                            if (tcor < 43) {
                                                                return 2;
                                                            }
                                                            else if (tcor >= 43) {
                                                                if (tcor < 52) {
                                                                    return 1;
                                                                }
                                                                else if (tcor >= 52) {
                                                                    if (Rs < 3051) {
                                                                        if (CsDiff < 4) {
                                                                            return 1;
                                                                        }
                                                                        else if (CsDiff >= 4) {
                                                                            if (diffRsCsdiff < -42) {
                                                                                if (AFD < 28) {
                                                                                    return 1;
                                                                                }
                                                                                else if (AFD >= 28) {
                                                                                    return 2;
                                                                                }
                                                                            }
                                                                            else if (diffRsCsdiff >= -42) {
                                                                                return 2;
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (Rs >= 3051) {
                                                                        return 1;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (tcor >= 65) {
                                                            if (Cs < 2014) {
                                                                if (AFD < 27) {
                                                                    if (ssDCval < 117) {
                                                                        return 2;
                                                                    }
                                                                    else if (ssDCval >= 117) {
                                                                        if (refDCval < 138) {
                                                                            return 4;
                                                                        }
                                                                        else if (refDCval >= 138) {
                                                                            return 2;
                                                                        }
                                                                    }
                                                                }
                                                                else if (AFD >= 27) {
                                                                    return 1;
                                                                }
                                                            }
                                                            else if (Cs >= 2014) {
                                                                return 1;
                                                            }
                                                        }
                                                    }
                                                    else if (SC >= 4736) {
                                                        if (Rs < 3686) {
                                                            if (Cs < 2321) {
                                                                if (MVDiff < 113) {
                                                                    return 8;
                                                                }
                                                                else if (MVDiff >= 113) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (Cs >= 2321) {
                                                                return 1;
                                                            }
                                                        }
                                                        else if (Rs >= 3686) {
                                                            return 2;
                                                        }
                                                    }
                                                }
                                                else if (AFD >= 31) {
                                                    if (MVDiff < 166) {
                                                        return 2;
                                                    }
                                                    else if (MVDiff >= 166) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                        }
                        else if (CsDiff >= 10) {
                            if (negBalance < 55) {
                                if (gchDC < 2) {
                                    if (refDCval < 145) {
                                        if (refDCval < 125) {
                                            if (Rs < 2200) {
                                                if (AFD < 11) {
                                                    return 2;
                                                }
                                                else if (AFD >= 11) {
                                                    if (ssDCval < 97) {
                                                        return 1;
                                                    }
                                                    else if (ssDCval >= 97) {
                                                        if (Rs < 1699) {
                                                            return 1;
                                                        }
                                                        else if (Rs >= 1699) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (Rs >= 2200) {
                                                if (posBalance < 56) {
                                                    if (SC < 8097) {
                                                        if (refDCval < 88) {
                                                            if (Rs < 3075) {
                                                                if (posBalance < 55) {
                                                                    if (mcTcor < 85) {
                                                                        return 8;
                                                                    }
                                                                    else if (mcTcor >= 85) {
                                                                        if (tcor < 79) {
                                                                            return 1;
                                                                        }
                                                                        else if (tcor >= 79) {
                                                                            if (diffRsCsdiff < 22) {
                                                                                return 1;
                                                                            }
                                                                            else if (diffRsCsdiff >= 22) {
                                                                                if (SC < 5430) {
                                                                                    return 1;
                                                                                }
                                                                                else if (SC >= 5430) {
                                                                                    if (CsDiff < 11) {
                                                                                        return 1;
                                                                                    }
                                                                                    else if (CsDiff >= 11) {
                                                                                        return 8;
                                                                                    }
                                                                                }
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                                else if (posBalance >= 55) {
                                                                    if (ssDCval < 75) {
                                                                        return 1;
                                                                    }
                                                                    else if (ssDCval >= 75) {
                                                                        return 8;
                                                                    }
                                                                }
                                                            }
                                                            else if (Rs >= 3075) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (refDCval >= 88) {
                                                            if (diffAFD < -4) {
                                                                return 4;
                                                            }
                                                            else if (diffAFD >= -4) {
                                                                if (negBalance < 48) {
                                                                    if (CsDiff < 11) {
                                                                        if (Rs < 2442) {
                                                                            return 4;
                                                                        }
                                                                        else if (Rs >= 2442) {
                                                                            if (mu_mv_mag_sq < 45) {
                                                                                if (posBalance < 55) {
                                                                                    return 1;
                                                                                }
                                                                                else if (posBalance >= 55) {
                                                                                    return 8;
                                                                                }
                                                                            }
                                                                            else if (mu_mv_mag_sq >= 45) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (CsDiff >= 11) {
                                                                        return 1;
                                                                    }
                                                                }
                                                                else if (negBalance >= 48) {
                                                                    if (Rs < 2680) {
                                                                        return 4;
                                                                    }
                                                                    else if (Rs >= 2680) {
                                                                        return 8;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (SC >= 8097) {
                                                        if (RsDiff < 16) {
                                                            return 2;
                                                        }
                                                        else if (RsDiff >= 16) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                                else if (posBalance >= 56) {
                                                    if (AFD < 18) {
                                                        if (ssDCval < 106) {
                                                            if (Cs < 2464) {
                                                                if (Rs < 2618) {
                                                                    if (Rs < 2209) {
                                                                        return 2;
                                                                    }
                                                                    else if (Rs >= 2209) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (Rs >= 2618) {
                                                                    return 2;
                                                                }
                                                            }
                                                            else if (Cs >= 2464) {
                                                                if (Cs < 3304) {
                                                                    return 4;
                                                                }
                                                                else if (Cs >= 3304) {
                                                                    if (RsDiff < 13) {
                                                                        return 8;
                                                                    }
                                                                    else if (RsDiff >= 13) {
                                                                        return 4;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (ssDCval >= 106) {
                                                            if (RsDiff < 16) {
                                                                return 2;
                                                            }
                                                            else if (RsDiff >= 16) {
                                                                return 1;
                                                            }
                                                        }
                                                    }
                                                    else if (AFD >= 18) {
                                                        if (ssDCval < 84) {
                                                            if (Rs < 2930) {
                                                                return 8;
                                                            }
                                                            else if (Rs >= 2930) {
                                                                return 1;
                                                            }
                                                        }
                                                        else if (ssDCval >= 84) {
                                                            if (Cs < 1925) {
                                                                if (MVDiff < 41) {
                                                                    return 4;
                                                                }
                                                                else if (MVDiff >= 41) {
                                                                    return 2;
                                                                }
                                                            }
                                                            else if (Cs >= 1925) {
                                                                if (refDCval < 90) {
                                                                    if (Cs < 2462) {
                                                                        if (ssDCval < 89) {
                                                                            return 4;
                                                                        }
                                                                        else if (ssDCval >= 89) {
                                                                            return 2;
                                                                        }
                                                                    }
                                                                    else if (Cs >= 2462) {
                                                                        return 1;
                                                                    }
                                                                }
                                                                else if (refDCval >= 90) {
                                                                    return 1;
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (refDCval >= 125) {
                                            if (tcor < 76) {
                                                if (diffMVdiffVal < -53) {
                                                    return 4;
                                                }
                                                else if (diffMVdiffVal >= -53) {
                                                    if (Rs < 1335) {
                                                        return 8;
                                                    }
                                                    else if (Rs >= 1335) {
                                                        if (Cs < 3392) {
                                                            if (ssDCval < 127) {
                                                                if (MVDiff < 15) {
                                                                    return 1;
                                                                }
                                                                else if (MVDiff >= 15) {
                                                                    if (Rs < 4598) {
                                                                        return 4;
                                                                    }
                                                                    else if (Rs >= 4598) {
                                                                        return 1;
                                                                    }
                                                                }
                                                            }
                                                            else if (ssDCval >= 127) {
                                                                return 1;
                                                            }
                                                        }
                                                        else if (Cs >= 3392) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (tcor >= 76) {
                                                if (Rs < 2692) {
                                                    return 2;
                                                }
                                                else if (Rs >= 2692) {
                                                    if (TSC < 12) {
                                                        return 1;
                                                    }
                                                    else if (TSC >= 12) {
                                                        if (Cs < 2789) {
                                                            return 1;
                                                        }
                                                        else if (Cs >= 2789) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (refDCval >= 145) {
                                        if (SC < 5580) {
                                            if (Rs < 1544) {
                                                return 8;
                                            }
                                            else if (Rs >= 1544) {
                                                return 2;
                                            }
                                        }
                                        else if (SC >= 5580) {
                                            if (posBalance < 58) {
                                                return 4;
                                            }
                                            else if (posBalance >= 58) {
                                                return 2;
                                            }
                                        }
                                    }
                                }
                                else if (gchDC >= 2) {
                                    if (negBalance < 36) {
                                        if (SC < 3741) {
                                            if (mu_mv_mag_sq < 31) {
                                                return 2;
                                            }
                                            else if (mu_mv_mag_sq >= 31) {
                                                if (diffAFD < 53) {
                                                    return 4;
                                                }
                                                else if (diffAFD >= 53) {
                                                    return 2;
                                                }
                                            }
                                        }
                                        else if (SC >= 3741) {
                                            if (diffAFD < -2) {
                                                return 4;
                                            }
                                            else if (diffAFD >= -2) {
                                                return 8;
                                            }
                                        }
                                    }
                                    else if (negBalance >= 36) {
                                        if (AFD < 24) {
                                            if (ssDCval < 91) {
                                                if (negBalance < 53) {
                                                    if (posBalance < 47) {
                                                        return 8;
                                                    }
                                                    else if (posBalance >= 47) {
                                                        return 1;
                                                    }
                                                }
                                                else if (negBalance >= 53) {
                                                    if (ssDCval < 74) {
                                                        return 8;
                                                    }
                                                    else if (ssDCval >= 74) {
                                                        if (AFD < 23) {
                                                            return 4;
                                                        }
                                                        else if (AFD >= 23) {
                                                            return 1;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (ssDCval >= 91) {
                                                if (ssDCval < 115) {
                                                    if (AFD < 21) {
                                                        if (negBalance < 39) {
                                                            if (Rs < 4428) {
                                                                return 8;
                                                            }
                                                            else if (Rs >= 4428) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (negBalance >= 39) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (AFD >= 21) {
                                                        if (Cs < 2676) {
                                                            return 1;
                                                        }
                                                        else if (Cs >= 2676) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                                else if (ssDCval >= 115) {
                                                    if (mcTcor < 89) {
                                                        if (AFD < 19) {
                                                            if (Rs < 1365) {
                                                                return 8;
                                                            }
                                                            else if (Rs >= 1365) {
                                                                return 1;
                                                            }
                                                        }
                                                        else if (AFD >= 19) {
                                                            if (negBalance < 53) {
                                                                return 1;
                                                            }
                                                            else if (negBalance >= 53) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                    else if (mcTcor >= 89) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                        }
                                        else if (AFD >= 24) {
                                            if (diffMVdiffVal < 7) {
                                                if (tcor < 65) {
                                                    if (MVDiff < 49) {
                                                        return 2;
                                                    }
                                                    else if (MVDiff >= 49) {
                                                        return 1;
                                                    }
                                                }
                                                else if (tcor >= 65) {
                                                    if (refDCval < 134) {
                                                        return 1;
                                                    }
                                                    else if (refDCval >= 134) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                            else if (diffMVdiffVal >= 7) {
                                                if (ssDCval < 120) {
                                                    return 4;
                                                }
                                                else if (ssDCval >= 120) {
                                                    if (Rs < 2131) {
                                                        return 1;
                                                    }
                                                    else if (Rs >= 2131) {
                                                        if (RsCsDiff < 395) {
                                                            return 8;
                                                        }
                                                        else if (RsCsDiff >= 395) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                            else if (negBalance >= 55) {
                                if (refDCval < 68) {
                                    if (Cs < 1146) {
                                        return 2;
                                    }
                                    else if (Cs >= 1146) {
                                        return 8;
                                    }
                                }
                                else if (refDCval >= 68) {
                                    if (ssDCval < 139) {
                                        if (Cs < 1944) {
                                            if (SC < 3632) {
                                                if (mcTcor < 57) {
                                                    if (AFD < 35) {
                                                        return 8;
                                                    }
                                                    else if (AFD >= 35) {
                                                        return 4;
                                                    }
                                                }
                                                else if (mcTcor >= 57) {
                                                    if (TSC < 7) {
                                                        if (Cs < 1568) {
                                                            return 8;
                                                        }
                                                        else if (Cs >= 1568) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (TSC >= 7) {
                                                        if (Rs < 589) {
                                                            if (Rs < 409) {
                                                                return 4;
                                                            }
                                                            else if (Rs >= 409) {
                                                                return 2;
                                                            }
                                                        }
                                                        else if (Rs >= 589) {
                                                            if (AFD < 18) {
                                                                return 2;
                                                            }
                                                            else if (AFD >= 18) {
                                                                if (mcTcor < 90) {
                                                                    return 1;
                                                                }
                                                                else if (mcTcor >= 90) {
                                                                    return 2;
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                            else if (SC >= 3632) {
                                                if (refDCval < 133) {
                                                    if (mcTcor < 88) {
                                                        if (RsCsDiff < 257) {
                                                            if (Rs < 2207) {
                                                                return 2;
                                                            }
                                                            else if (Rs >= 2207) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (RsCsDiff >= 257) {
                                                            if (diffTSC < -7) {
                                                                return 8;
                                                            }
                                                            else if (diffTSC >= -7) {
                                                                return 2;
                                                            }
                                                        }
                                                    }
                                                    else if (mcTcor >= 88) {
                                                        if (Cs < 1887) {
                                                            return 2;
                                                        }
                                                        else if (Cs >= 1887) {
                                                            if (TSC < 15) {
                                                                return 2;
                                                            }
                                                            else if (TSC >= 15) {
                                                                return 1;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (refDCval >= 133) {
                                                    return 4;
                                                }
                                            }
                                        }
                                        else if (Cs >= 1944) {
                                            if (RsDiff < 14) {
                                                if (mu_mv_mag_sq < 3) {
                                                    if (ssDCval < 97) {
                                                        return 4;
                                                    }
                                                    else if (ssDCval >= 97) {
                                                        if (SC < 5953) {
                                                            if (TSC < 13) {
                                                                return 1;
                                                            }
                                                            else if (TSC >= 13) {
                                                                if (Rs < 2577) {
                                                                    return 1;
                                                                }
                                                                else if (Rs >= 2577) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                        else if (SC >= 5953) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                                else if (mu_mv_mag_sq >= 3) {
                                                    if (mcTcor < 70) {
                                                        if (TSC < 15) {
                                                            return 1;
                                                        }
                                                        else if (TSC >= 15) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (mcTcor >= 70) {
                                                        if (posBalance < 51) {
                                                            if (AFD < 17) {
                                                                return 8;
                                                            }
                                                            else if (AFD >= 17) {
                                                                return 2;
                                                            }
                                                        }
                                                        else if (posBalance >= 51) {
                                                            if (tcor < 68) {
                                                                return 1;
                                                            }
                                                            else if (tcor >= 68) {
                                                                if (diffMVdiffVal < 1) {
                                                                    return 1;
                                                                }
                                                                else if (diffMVdiffVal >= 1) {
                                                                    if (gchDC < 1) {
                                                                        return 1;
                                                                    }
                                                                    else if (gchDC >= 1) {
                                                                        if (AFD < 23) {
                                                                            return 4;
                                                                        }
                                                                        else if (AFD >= 23) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                            else if (RsDiff >= 14) {
                                                if (posBalance < 56) {
                                                    if (AFD < 26) {
                                                        if (gchDC < 5) {
                                                            if (SC < 5437) {
                                                                if (SC < 4549) {
                                                                    return 8;
                                                                }
                                                                else if (SC >= 4549) {
                                                                    return 2;
                                                                }
                                                            }
                                                            else if (SC >= 5437) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (gchDC >= 5) {
                                                            if (Cs < 2175) {
                                                                return 1;
                                                            }
                                                            else if (Cs >= 2175) {
                                                                return 2;
                                                            }
                                                        }
                                                    }
                                                    else if (AFD >= 26) {
                                                        if (RsDiff < 24) {
                                                            if (Cs < 2907) {
                                                                return 1;
                                                            }
                                                            else if (Cs >= 2907) {
                                                                return 2;
                                                            }
                                                        }
                                                        else if (RsDiff >= 24) {
                                                            if (AFD < 30) {
                                                                return 4;
                                                            }
                                                            else if (AFD >= 30) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (posBalance >= 56) {
                                                    if (MVDiff < 57) {
                                                        if (refDCval < 122) {
                                                            if (Rs < 3706) {
                                                                if (TSC < 11) {
                                                                    return 2;
                                                                }
                                                                else if (TSC >= 11) {
                                                                    return 1;
                                                                }
                                                            }
                                                            else if (Rs >= 3706) {
                                                                return 2;
                                                            }
                                                        }
                                                        else if (refDCval >= 122) {
                                                            return 1;
                                                        }
                                                    }
                                                    else if (MVDiff >= 57) {
                                                        if (Cs < 2222) {
                                                            if (SC < 4867) {
                                                                return 2;
                                                            }
                                                            else if (SC >= 4867) {
                                                                return 1;
                                                            }
                                                        }
                                                        else if (Cs >= 2222) {
                                                            if (Cs < 2274) {
                                                                return 8;
                                                            }
                                                            else if (Cs >= 2274) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (ssDCval >= 139) {
                                        if (RsDiff < 11) {
                                            return 8;
                                        }
                                        else if (RsDiff >= 11) {
                                            if (ssDCval < 159) {
                                                if (TSC < 10) {
                                                    return 2;
                                                }
                                                else if (TSC >= 10) {
                                                    return 4;
                                                }
                                            }
                                            else if (ssDCval >= 159) {
                                                return 2;
                                            }
                                        }
                                    }
                                }
                            }
                        }
                    }
                    else if (mcTcor >= 94) {
                        if (RsCsDiff < 121) {
                            if (Rs < 2017) {
                                if (tcor < 86) {
                                    if (posBalance < 85) {
                                        if (SC < 845) {
                                            if (Rs < 551) {
                                                if (tcor < 84) {
                                                    return 2;
                                                }
                                                else if (tcor >= 84) {
                                                    if (TSC < 6) {
                                                        if (negBalance < 89) {
                                                            return 2;
                                                        }
                                                        else if (negBalance >= 89) {
                                                            return 1;
                                                        }
                                                    }
                                                    else if (TSC >= 6) {
                                                        return 1;
                                                    }
                                                }
                                            }
                                            else if (Rs >= 551) {
                                                return 1;
                                            }
                                        }
                                        else if (SC >= 845) {
                                            if (RsDiff < 3) {
                                                if (Rs < 274) {
                                                    return 2;
                                                }
                                                else if (Rs >= 274) {
                                                    return 4;
                                                }
                                            }
                                            else if (RsDiff >= 3) {
                                                if (tcor < 79) {
                                                    if (RsCsDiff < 63) {
                                                        return 4;
                                                    }
                                                    else if (RsCsDiff >= 63) {
                                                        return 2;
                                                    }
                                                }
                                                else if (tcor >= 79) {
                                                    if (AFD < 10) {
                                                        if (Cs < 946) {
                                                            return 4;
                                                        }
                                                        else if (Cs >= 946) {
                                                            if (SC < 1980) {
                                                                return 2;
                                                            }
                                                            else if (SC >= 1980) {
                                                                return 1;
                                                            }
                                                        }
                                                    }
                                                    else if (AFD >= 10) {
                                                        if (TSC < 11) {
                                                            return 2;
                                                        }
                                                        else if (TSC >= 11) {
                                                            return 1;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (posBalance >= 85) {
                                        return 8;
                                    }
                                }
                                else if (tcor >= 86) {
                                    if (mu_mv_mag_sq < 38) {
                                        if (Rs < 1022) {
                                            if (posBalance < 64) {
                                                if (Rs < 28) {
                                                    return 1;
                                                }
                                                else if (Rs >= 28) {
                                                    return 2;
                                                }
                                            }
                                            else if (posBalance >= 64) {
                                                return 1;
                                            }
                                        }
                                        else if (Rs >= 1022) {
                                            return 1;
                                        }
                                    }
                                    else if (mu_mv_mag_sq >= 38) {
                                        if (diffRsCsdiff < 2) {
                                            return 4;
                                        }
                                        else if (diffRsCsdiff >= 2) {
                                            if (Cs < 752) {
                                                return 1;
                                            }
                                            else if (Cs >= 752) {
                                                return 8;
                                            }
                                        }
                                    }
                                }
                            }
                            else if (Rs >= 2017) {
                                if (mcTcor < 95) {
                                    if (Rs < 2517) {
                                        return 4;
                                    }
                                    else if (Rs >= 2517) {
                                        if (Rs < 2579) {
                                            return 2;
                                        }
                                        else if (Rs >= 2579) {
                                            return 8;
                                        }
                                    }
                                }
                                else if (mcTcor >= 95) {
                                    return 8;
                                }
                            }
                        }
                        else if (RsCsDiff >= 121) {
                            if (RsDiff < 22) {
                                if (ssDCval < 83) {
                                    if (TSC < 6) {
                                        return 2;
                                    }
                                    else if (TSC >= 6) {
                                        if (Rs < 2165) {
                                            return 1;
                                        }
                                        else if (Rs >= 2165) {
                                            return 4;
                                        }
                                    }
                                }
                                else if (ssDCval >= 83) {
                                    if (TSC < 6) {
                                        if (RsCsDiff < 141) {
                                            return 2;
                                        }
                                        else if (RsCsDiff >= 141) {
                                            return 8;
                                        }
                                    }
                                    else if (TSC >= 6) {
                                        if (CsDiff < 15) {
                                            if (tcor < 86) {
                                                return 2;
                                            }
                                            else if (tcor >= 86) {
                                                if (AFD < 14) {
                                                    return 4;
                                                }
                                                else if (AFD >= 14) {
                                                    return 2;
                                                }
                                            }
                                        }
                                        else if (CsDiff >= 15) {
                                            if (Cs < 2821) {
                                                return 2;
                                            }
                                            else if (Cs >= 2821) {
                                                return 8;
                                            }
                                        }
                                    }
                                }
                            }
                            else if (RsDiff >= 22) {
                                if (diffRsCsdiff < 192) {
                                    return 4;
                                }
                                else if (diffRsCsdiff >= 192) {
                                    if (Rs < 1647) {
                                        return 8;
                                    }
                                    else if (Rs >= 1647) {
                                        return 2;
                                    }
                                }
                            }
                        }
                    }
                }
                else if (tcor >= 88) {
                    if (Rs < 3179) {
                        if (AFD < 10) {
                            if (CsDiff < 3) {
                                if (tcor < 98) {
                                    if (TSCindex < 7) {
                                        if (mu_mv_mag_sq < 85) {
                                            if (refDCval < 36) {
                                                if (gchDC < 1) {
                                                    return 2;
                                                }
                                                else if (gchDC >= 1) {
                                                    if (negBalance < 2) {
                                                        return 2;
                                                    }
                                                    else if (negBalance >= 2) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                            else if (refDCval >= 36) {
                                                if (Rs < 976) {
                                                    if (SC < 709) {
                                                        if (refDCval < 67) {
                                                            if (Rs < 236) {
                                                                return 2;
                                                            }
                                                            else if (Rs >= 236) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (refDCval >= 67) {
                                                            if (Rs < 250) {
                                                                return 2;
                                                            }
                                                            else if (Rs >= 250) {
                                                                return 1;
                                                            }
                                                        }
                                                    }
                                                    else if (SC >= 709) {
                                                        if (diffMVdiffVal < -7) {
                                                            if (ssDCval < 59) {
                                                                return 8;
                                                            }
                                                            else if (ssDCval >= 59) {
                                                                return 2;
                                                            }
                                                        }
                                                        else if (diffMVdiffVal >= -7) {
                                                            if (mu_mv_mag_sq < 37) {
                                                                if (Rs < 824) {
                                                                    return 4;
                                                                }
                                                                else if (Rs >= 824) {
                                                                    if (gchDC < 1) {
                                                                        return 8;
                                                                    }
                                                                    else if (gchDC >= 1) {
                                                                        return 1;
                                                                    }
                                                                }
                                                            }
                                                            else if (mu_mv_mag_sq >= 37) {
                                                                if (RsDiff < 3) {
                                                                    if (mu_mv_mag_sq < 84) {
                                                                        if (Cs < 705) {
                                                                            return 4;
                                                                        }
                                                                        else if (Cs >= 705) {
                                                                            if (MVDiff < 15) {
                                                                                return 4;
                                                                            }
                                                                            else if (MVDiff >= 15) {
                                                                                return 1;
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (mu_mv_mag_sq >= 84) {
                                                                        if (posBalance < 37) {
                                                                            return 4;
                                                                        }
                                                                        else if (posBalance >= 37) {
                                                                            return 2;
                                                                        }
                                                                    }
                                                                }
                                                                else if (RsDiff >= 3) {
                                                                    if (SC < 1593) {
                                                                        return 2;
                                                                    }
                                                                    else if (SC >= 1593) {
                                                                        if (Rs < 894) {
                                                                            return 8;
                                                                        }
                                                                        else if (Rs >= 894) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (Rs >= 976) {
                                                    if (Rs < 1753) {
                                                        if (ssDCval < 42) {
                                                            if (MVDiff < 23) {
                                                                if (SC < 1624) {
                                                                    return 4;
                                                                }
                                                                else if (SC >= 1624) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (MVDiff >= 23) {
                                                                return 2;
                                                            }
                                                        }
                                                        else if (ssDCval >= 42) {
                                                            if (TSC < 4) {
                                                                return 1;
                                                            }
                                                            else if (TSC >= 4) {
                                                                if (mcTcor < 94) {
                                                                    if (Rs < 1218) {
                                                                        if (tcor < 89) {
                                                                            if (SC < 2042) {
                                                                                return 4;
                                                                            }
                                                                            else if (SC >= 2042) {
                                                                                return 1;
                                                                            }
                                                                        }
                                                                        else if (tcor >= 89) {
                                                                            if (MVDiff < 8) {
                                                                                return 1;
                                                                            }
                                                                            else if (MVDiff >= 8) {
                                                                                if (diffAFD < 1) {
                                                                                    return 8;
                                                                                }
                                                                                else if (diffAFD >= 1) {
                                                                                    return 4;
                                                                                }
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (Rs >= 1218) {
                                                                        if (posBalance < 50) {
                                                                            if (Cs < 980) {
                                                                                if (Cs < 872) {
                                                                                    return 1;
                                                                                }
                                                                                else if (Cs >= 872) {
                                                                                    return 4;
                                                                                }
                                                                            }
                                                                            else if (Cs >= 980) {
                                                                                return 1;
                                                                            }
                                                                        }
                                                                        else if (posBalance >= 50) {
                                                                            if (AFD < 7) {
                                                                                return 1;
                                                                            }
                                                                            else if (AFD >= 7) {
                                                                                if (Rs < 1266) {
                                                                                    return 1;
                                                                                }
                                                                                else if (Rs >= 1266) {
                                                                                    return 8;
                                                                                }
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                                else if (mcTcor >= 94) {
                                                                    if (posBalance < 54) {
                                                                        if (MVDiff < 9) {
                                                                            if (Cs < 974) {
                                                                                if (AFD < 6) {
                                                                                    return 1;
                                                                                }
                                                                                else if (AFD >= 6) {
                                                                                    if (RsDiff < 3) {
                                                                                        return 4;
                                                                                    }
                                                                                    else if (RsDiff >= 3) {
                                                                                        return 1;
                                                                                    }
                                                                                }
                                                                            }
                                                                            else if (Cs >= 974) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                        else if (MVDiff >= 9) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                    else if (posBalance >= 54) {
                                                                        if (negBalance < 41) {
                                                                            if (Cs < 852) {
                                                                                return 4;
                                                                            }
                                                                            else if (Cs >= 852) {
                                                                                return 1;
                                                                            }
                                                                        }
                                                                        else if (negBalance >= 41) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (Rs >= 1753) {
                                                        if (Cs < 1549) {
                                                            if (Rs < 2203) {
                                                                if (Rs < 1815) {
                                                                    if (RsDiff < 3) {
                                                                        if (MVDiff < 23) {
                                                                            return 1;
                                                                        }
                                                                        else if (MVDiff >= 23) {
                                                                            return 2;
                                                                        }
                                                                    }
                                                                    else if (RsDiff >= 3) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (Rs >= 1815) {
                                                                    if (SC < 3344) {
                                                                        if (diffMVdiffVal < -1) {
                                                                            return 4;
                                                                        }
                                                                        else if (diffMVdiffVal >= -1) {
                                                                            if (Cs < 1494) {
                                                                                return 8;
                                                                            }
                                                                            else if (Cs >= 1494) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (SC >= 3344) {
                                                                        if (posBalance < 56) {
                                                                            if (tcor < 92) {
                                                                                if (diffRsCsdiff < -2) {
                                                                                    return 4;
                                                                                }
                                                                                else if (diffRsCsdiff >= -2) {
                                                                                    if (Cs < 1519) {
                                                                                        return 4;
                                                                                    }
                                                                                    else if (Cs >= 1519) {
                                                                                        return 1;
                                                                                    }
                                                                                }
                                                                            }
                                                                            else if (tcor >= 92) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                        else if (posBalance >= 56) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                            else if (Rs >= 2203) {
                                                                if (AFD < 6) {
                                                                    return 4;
                                                                }
                                                                else if (AFD >= 6) {
                                                                    return 2;
                                                                }
                                                            }
                                                        }
                                                        else if (Cs >= 1549) {
                                                            if (mu_mv_mag_sq < 75) {
                                                                if (negBalance < 50) {
                                                                    return 1;
                                                                }
                                                                else if (negBalance >= 50) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (mu_mv_mag_sq >= 75) {
                                                                if (MVDiff < 23) {
                                                                    return 4;
                                                                }
                                                                else if (MVDiff >= 23) {
                                                                    return 1;
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (mu_mv_mag_sq >= 85) {
                                            if (RsCsDiff < 4) {
                                                if (negBalance < 13) {
                                                    if (RsCsDiff < 1) {
                                                        if (AFD < 1) {
                                                            return 4;
                                                        }
                                                        else if (AFD >= 1) {
                                                            if (Rs < 583) {
                                                                return 2;
                                                            }
                                                            else if (Rs >= 583) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                    else if (RsCsDiff >= 1) {
                                                        return 8;
                                                    }
                                                }
                                                else if (negBalance >= 13) {
                                                    if (negBalance < 42) {
                                                        if (Rs < 286) {
                                                            return 2;
                                                        }
                                                        else if (Rs >= 286) {
                                                            if (diffTSC < -1) {
                                                                return 1;
                                                            }
                                                            else if (diffTSC >= -1) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                    else if (negBalance >= 42) {
                                                        if (Scindex < 4) {
                                                            return 4;
                                                        }
                                                        else if (Scindex >= 4) {
                                                            if (posBalance < 56) {
                                                                if (Cs < 1517) {
                                                                    return 8;
                                                                }
                                                                else if (Cs >= 1517) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (posBalance >= 56) {
                                                                if (tcor < 95) {
                                                                    return 8;
                                                                }
                                                                else if (tcor >= 95) {
                                                                    return 1;
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                            else if (RsCsDiff >= 4) {
                                                if (posBalance < 19) {
                                                    if (Cs < 846) {
                                                        return 1;
                                                    }
                                                    else if (Cs >= 846) {
                                                        return 8;
                                                    }
                                                }
                                                else if (posBalance >= 19) {
                                                    if (diffTSC < -2) {
                                                        if (Rs < 1352) {
                                                            return 8;
                                                        }
                                                        else if (Rs >= 1352) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (diffTSC >= -2) {
                                                        if (diffAFD < -2) {
                                                            return 1;
                                                        }
                                                        else if (diffAFD >= -2) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (TSCindex >= 7) {
                                        if (refDCval < 105) {
                                            if (SC < 2923) {
                                                if (Rs < 143) {
                                                    return 2;
                                                }
                                                else if (Rs >= 143) {
                                                    return 1;
                                                }
                                            }
                                            else if (SC >= 2923) {
                                                if (TSC < 7) {
                                                    if (Rs < 2472) {
                                                        return 4;
                                                    }
                                                    else if (Rs >= 2472) {
                                                        return 8;
                                                    }
                                                }
                                                else if (TSC >= 7) {
                                                    return 1;
                                                }
                                            }
                                        }
                                        else if (refDCval >= 105) {
                                            if (Rs < 440) {
                                                return 4;
                                            }
                                            else if (Rs >= 440) {
                                                if (Rs < 1159) {
                                                    return 8;
                                                }
                                                else if (Rs >= 1159) {
                                                    return 2;
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (tcor >= 98) {
                                    if (negBalance < 40) {
                                        if (TSCindex < 1) {
                                            if (mu_mv_mag_sq < 108) {
                                                if (Rs < 852) {
                                                    if (SC < 873) {
                                                        if (negBalance < 18) {
                                                            if (MVDiff < 31) {
                                                                return 4;
                                                            }
                                                            else if (MVDiff >= 31) {
                                                                if (mcTcor < 80) {
                                                                    return 4;
                                                                }
                                                                else if (mcTcor >= 80) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                        else if (negBalance >= 18) {
                                                            if (negBalance < 34) {
                                                                return 2;
                                                            }
                                                            else if (negBalance >= 34) {
                                                                if (negBalance < 36) {
                                                                    return 4;
                                                                }
                                                                else if (negBalance >= 36) {
                                                                    return 1;
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (SC >= 873) {
                                                        return 1;
                                                    }
                                                }
                                                else if (Rs >= 852) {
                                                    return 4;
                                                }
                                            }
                                            else if (mu_mv_mag_sq >= 108) {
                                                if (SC < 1870) {
                                                    if (diffAFD < -5) {
                                                        return 8;
                                                    }
                                                    else if (diffAFD >= -5) {
                                                        return 4;
                                                    }
                                                }
                                                else if (SC >= 1870) {
                                                    return 8;
                                                }
                                            }
                                        }
                                        else if (TSCindex >= 1) {
                                            if (Cs < 777) {
                                                if (Cs < 282) {
                                                    if (ssDCval < 48) {
                                                        if (mu_mv_mag_sq < 39) {
                                                            return 2;
                                                        }
                                                        else if (mu_mv_mag_sq >= 39) {
                                                            if (Cs < 60) {
                                                                return 4;
                                                            }
                                                            else if (Cs >= 60) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                    else if (ssDCval >= 48) {
                                                        return 1;
                                                    }
                                                }
                                                else if (Cs >= 282) {
                                                    if (Cs < 328) {
                                                        if (posBalance < 66) {
                                                            return 4;
                                                        }
                                                        else if (posBalance >= 66) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (Cs >= 328) {
                                                        if (MVDiff < 16) {
                                                            if (SC < 924) {
                                                                return 4;
                                                            }
                                                            else if (SC >= 924) {
                                                                if (Rs < 957) {
                                                                    return 2;
                                                                }
                                                                else if (Rs >= 957) {
                                                                    if (Rs < 1888) {
                                                                        return 8;
                                                                    }
                                                                    else if (Rs >= 1888) {
                                                                        return 2;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (MVDiff >= 16) {
                                                            if (Rs < 430) {
                                                                return 8;
                                                            }
                                                            else if (Rs >= 430) {
                                                                if (diffMVdiffVal < 0) {
                                                                    return 4;
                                                                }
                                                                else if (diffMVdiffVal >= 0) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                            else if (Cs >= 777) {
                                                if (Cs < 1025) {
                                                    return 1;
                                                }
                                                else if (Cs >= 1025) {
                                                    if (SC < 1882) {
                                                        return 4;
                                                    }
                                                    else if (SC >= 1882) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (negBalance >= 40) {
                                        if (Rs < 482) {
                                            if (diffAFD < -1) {
                                                if (MVDiff < 29) {
                                                    if (posBalance < 11) {
                                                        if (Rs < 272) {
                                                            return 2;
                                                        }
                                                        else if (Rs >= 272) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (posBalance >= 11) {
                                                        return 4;
                                                    }
                                                }
                                                else if (MVDiff >= 29) {
                                                    if (MVDiff < 93) {
                                                        return 2;
                                                    }
                                                    else if (MVDiff >= 93) {
                                                        return 1;
                                                    }
                                                }
                                            }
                                            else if (diffAFD >= -1) {
                                                if (negBalance < 90) {
                                                    if (ssDCval < 82) {
                                                        if (diffTSC < -1) {
                                                            if (Rs < 50) {
                                                                return 4;
                                                            }
                                                            else if (Rs >= 50) {
                                                                return 2;
                                                            }
                                                        }
                                                        else if (diffTSC >= -1) {
                                                            if (TSCindex < 2) {
                                                                return 8;
                                                            }
                                                            else if (TSCindex >= 2) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                    else if (ssDCval >= 82) {
                                                        return 8;
                                                    }
                                                }
                                                else if (negBalance >= 90) {
                                                    return 8;
                                                }
                                            }
                                        }
                                        else if (Rs >= 482) {
                                            if (diffMVdiffVal < 45) {
                                                return 8;
                                            }
                                            else if (diffMVdiffVal >= 45) {
                                                return 2;
                                            }
                                        }
                                    }
                                }
                            }
                            else if (CsDiff >= 3) {
                                if (mu_mv_mag_sq < 46) {
                                    if (Cs < 1196) {
                                        if (Cs < 540) {
                                            if (AFD < 7) {
                                                return 8;
                                            }
                                            else if (AFD >= 7) {
                                                if (gchDC < 1) {
                                                    return 2;
                                                }
                                                else if (gchDC >= 1) {
                                                    return 4;
                                                }
                                            }
                                        }
                                        else if (Cs >= 540) {
                                            return 2;
                                        }
                                    }
                                    else if (Cs >= 1196) {
                                        if (posBalance < 48) {
                                            return 4;
                                        }
                                        else if (posBalance >= 48) {
                                            if (SC < 4069) {
                                                return 4;
                                            }
                                            else if (SC >= 4069) {
                                                return 8;
                                            }
                                        }
                                    }
                                }
                                else if (mu_mv_mag_sq >= 46) {
                                    if (ssDCval < 94) {
                                        if (mu_mv_mag_sq < 75) {
                                            if (Rs < 2604) {
                                                if (Cs < 1670) {
                                                    if (tcor < 91) {
                                                        if (Cs < 1048) {
                                                            return 1;
                                                        }
                                                        else if (Cs >= 1048) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (tcor >= 91) {
                                                        if (ssDCval < 72) {
                                                            if (MVDiff < 12) {
                                                                return 1;
                                                            }
                                                            else if (MVDiff >= 12) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (ssDCval >= 72) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                                else if (Cs >= 1670) {
                                                    if (MVDiff < 16) {
                                                        if (Cs < 1929) {
                                                            return 8;
                                                        }
                                                        else if (Cs >= 1929) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (MVDiff >= 16) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                            else if (Rs >= 2604) {
                                                if (Cs < 1277) {
                                                    return 2;
                                                }
                                                else if (Cs >= 1277) {
                                                    return 1;
                                                }
                                            }
                                        }
                                        else if (mu_mv_mag_sq >= 75) {
                                            if (posBalance < 56) {
                                                if (refDCval < 57) {
                                                    return 8;
                                                }
                                                else if (refDCval >= 57) {
                                                    if (Rs < 1955) {
                                                        if (Cs < 1977) {
                                                            if (TSC < 6) {
                                                                return 4;
                                                            }
                                                            else if (TSC >= 6) {
                                                                if (SC < 2061) {
                                                                    return 2;
                                                                }
                                                                else if (SC >= 2061) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                        else if (Cs >= 1977) {
                                                            return 1;
                                                        }
                                                    }
                                                    else if (Rs >= 1955) {
                                                        if (SC < 4997) {
                                                            return 8;
                                                        }
                                                        else if (SC >= 4997) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (posBalance >= 56) {
                                                if (negBalance < 52) {
                                                    return 8;
                                                }
                                                else if (negBalance >= 52) {
                                                    if (Rs < 2308) {
                                                        return 2;
                                                    }
                                                    else if (Rs >= 2308) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (ssDCval >= 94) {
                                        if (mu_mv_mag_sq < 86) {
                                            if (mcTcor < 97) {
                                                if (diffTSC < -2) {
                                                    return 4;
                                                }
                                                else if (diffTSC >= -2) {
                                                    if (negBalance < 51) {
                                                        return 8;
                                                    }
                                                    else if (negBalance >= 51) {
                                                        if (SC < 4554) {
                                                            return 2;
                                                        }
                                                        else if (SC >= 4554) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (mcTcor >= 97) {
                                                if (Cs < 1453) {
                                                    return 4;
                                                }
                                                else if (Cs >= 1453) {
                                                    if (Rs < 2043) {
                                                        return 2;
                                                    }
                                                    else if (Rs >= 2043) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                        }
                                        else if (mu_mv_mag_sq >= 86) {
                                            if (Rs < 2225) {
                                                if (negBalance < 74) {
                                                    if (Cs < 869) {
                                                        return 1;
                                                    }
                                                    else if (Cs >= 869) {
                                                        return 8;
                                                    }
                                                }
                                                else if (negBalance >= 74) {
                                                    return 4;
                                                }
                                            }
                                            else if (Rs >= 2225) {
                                                if (Rs < 2790) {
                                                    return 4;
                                                }
                                                else if (Rs >= 2790) {
                                                    return 8;
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                        }
                        else if (AFD >= 10) {
                            if (SC < 3885) {
                                if (Rs < 1061) {
                                    if (MVDiff < 27) {
                                        if (posBalance < 63) {
                                            if (Rs < 994) {
                                                return 2;
                                            }
                                            else if (Rs >= 994) {
                                                if (SC < 1724) {
                                                    return 2;
                                                }
                                                else if (SC >= 1724) {
                                                    return 1;
                                                }
                                            }
                                        }
                                        else if (posBalance >= 63) {
                                            if (mu_mv_mag_sq < 7) {
                                                if (Rs < 940) {
                                                    return 4;
                                                }
                                                else if (Rs >= 940) {
                                                    return 2;
                                                }
                                            }
                                            else if (mu_mv_mag_sq >= 7) {
                                                return 8;
                                            }
                                        }
                                    }
                                    else if (MVDiff >= 27) {
                                        if (tcor < 100) {
                                            if (refDCval < 105) {
                                                return 1;
                                            }
                                            else if (refDCval >= 105) {
                                                if (RsDiff < 4) {
                                                    if (TSC < 8) {
                                                        return 2;
                                                    }
                                                    else if (TSC >= 8) {
                                                        return 8;
                                                    }
                                                }
                                                else if (RsDiff >= 4) {
                                                    if (MVDiff < 49) {
                                                        return 2;
                                                    }
                                                    else if (MVDiff >= 49) {
                                                        return 1;
                                                    }
                                                }
                                            }
                                        }
                                        else if (tcor >= 100) {
                                            if (MVDiff < 59) {
                                                return 2;
                                            }
                                            else if (MVDiff >= 59) {
                                                if (tcor < 106) {
                                                    return 8;
                                                }
                                                else if (tcor >= 106) {
                                                    if (Rs < 461) {
                                                        return 1;
                                                    }
                                                    else if (Rs >= 461) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (Rs >= 1061) {
                                    if (Cs < 1271) {
                                        if (refDCval < 115) {
                                            if (negBalance < 57) {
                                                if (diffAFD < 2) {
                                                    return 2;
                                                }
                                                else if (diffAFD >= 2) {
                                                    return 1;
                                                }
                                            }
                                            else if (negBalance >= 57) {
                                                if (diffRsCsdiff < 7) {
                                                    if (TSC < 5) {
                                                        return 8;
                                                    }
                                                    else if (TSC >= 5) {
                                                        return 4;
                                                    }
                                                }
                                                else if (diffRsCsdiff >= 7) {
                                                    return 2;
                                                }
                                            }
                                        }
                                        else if (refDCval >= 115) {
                                            if (RsCsDiff < 46) {
                                                if (diffAFD < -17) {
                                                    if (Rs < 1670) {
                                                        return 8;
                                                    }
                                                    else if (Rs >= 1670) {
                                                        return 2;
                                                    }
                                                }
                                                else if (diffAFD >= -17) {
                                                    return 4;
                                                }
                                            }
                                            else if (RsCsDiff >= 46) {
                                                return 8;
                                            }
                                        }
                                    }
                                    else if (Cs >= 1271) {
                                        if (Rs < 1647) {
                                            return 4;
                                        }
                                        else if (Rs >= 1647) {
                                            if (ssDCval < 95) {
                                                if (Cs < 1723) {
                                                    return 2;
                                                }
                                                else if (Cs >= 1723) {
                                                    return 8;
                                                }
                                            }
                                            else if (ssDCval >= 95) {
                                                if (RsDiff < 10) {
                                                    if (negBalance < 39) {
                                                        return 8;
                                                    }
                                                    else if (negBalance >= 39) {
                                                        return 4;
                                                    }
                                                }
                                                else if (RsDiff >= 10) {
                                                    if (TSC < 11) {
                                                        return 8;
                                                    }
                                                    else if (TSC >= 11) {
                                                        return 2;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                            else if (SC >= 3885) {
                                if (RsCsDiff < 80) {
                                    if (refDCval < 79) {
                                        if (SC < 4015) {
                                            return 1;
                                        }
                                        else if (SC >= 4015) {
                                            return 8;
                                        }
                                    }
                                    else if (refDCval >= 79) {
                                        if (mu_mv_mag_sq < 66) {
                                            if (posBalance < 58) {
                                                if (TSC < 8) {
                                                    return 2;
                                                }
                                                else if (TSC >= 8) {
                                                    if (MVDiff < 19) {
                                                        return 1;
                                                    }
                                                    else if (MVDiff >= 19) {
                                                        return 2;
                                                    }
                                                }
                                            }
                                            else if (posBalance >= 58) {
                                                if (tcor < 92) {
                                                    return 1;
                                                }
                                                else if (tcor >= 92) {
                                                    if (Rs < 2426) {
                                                        return 2;
                                                    }
                                                    else if (Rs >= 2426) {
                                                        return 1;
                                                    }
                                                }
                                            }
                                        }
                                        else if (mu_mv_mag_sq >= 66) {
                                            if (SC < 5235) {
                                                return 1;
                                            }
                                            else if (SC >= 5235) {
                                                return 4;
                                            }
                                        }
                                    }
                                }
                                else if (RsCsDiff >= 80) {
                                    if (Rs < 2452) {
                                        if (posBalance < 54) {
                                            if (MVDiff < 30) {
                                                if (gchDC < 2) {
                                                    if (MVDiff < 15) {
                                                        if (mu_mv_mag_sq < 47) {
                                                            return 1;
                                                        }
                                                        else if (mu_mv_mag_sq >= 47) {
                                                            if (diffTSC < -1) {
                                                                if (ssDCval < 82) {
                                                                    return 8;
                                                                }
                                                                else if (ssDCval >= 82) {
                                                                    return 1;
                                                                }
                                                            }
                                                            else if (diffTSC >= -1) {
                                                                if (diffRsCsdiff < -24) {
                                                                    return 1;
                                                                }
                                                                else if (diffRsCsdiff >= -24) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (MVDiff >= 15) {
                                                        return 8;
                                                    }
                                                }
                                                else if (gchDC >= 2) {
                                                    if (Rs < 1986) {
                                                        return 2;
                                                    }
                                                    else if (Rs >= 1986) {
                                                        return 1;
                                                    }
                                                }
                                            }
                                            else if (MVDiff >= 30) {
                                                return 4;
                                            }
                                        }
                                        else if (posBalance >= 54) {
                                            if (Cs < 2023) {
                                                if (Rs < 2353) {
                                                    return 2;
                                                }
                                                else if (Rs >= 2353) {
                                                    if (Rs < 2379) {
                                                        return 8;
                                                    }
                                                    else if (Rs >= 2379) {
                                                        return 1;
                                                    }
                                                }
                                            }
                                            else if (Cs >= 2023) {
                                                if (MVDiff < 10) {
                                                    return 2;
                                                }
                                                else if (MVDiff >= 10) {
                                                    return 4;
                                                }
                                            }
                                        }
                                    }
                                    else if (Rs >= 2452) {
                                        if (MVDiff < 26) {
                                            if (AFD < 12) {
                                                if (mcTcor < 94) {
                                                    if (tcor < 89) {
                                                        return 4;
                                                    }
                                                    else if (tcor >= 89) {
                                                        if (negBalance < 55) {
                                                            return 8;
                                                        }
                                                        else if (negBalance >= 55) {
                                                            return 2;
                                                        }
                                                    }
                                                }
                                                else if (mcTcor >= 94) {
                                                    if (SC < 5021) {
                                                        return 1;
                                                    }
                                                    else if (SC >= 5021) {
                                                        if (TSC < 9) {
                                                            return 2;
                                                        }
                                                        else if (TSC >= 9) {
                                                            if (Cs < 2738) {
                                                                return 4;
                                                            }
                                                            else if (Cs >= 2738) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                            else if (AFD >= 12) {
                                                if (RsDiff < 12) {
                                                    if (Cs < 2036) {
                                                        if (ssDCval < 83) {
                                                            return 2;
                                                        }
                                                        else if (ssDCval >= 83) {
                                                            return 1;
                                                        }
                                                    }
                                                    else if (Cs >= 2036) {
                                                        if (Rs < 2521) {
                                                            if (ssDCval < 88) {
                                                                return 4;
                                                            }
                                                            else if (ssDCval >= 88) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (Rs >= 2521) {
                                                            if (MVDiff < 10) {
                                                                if (diffAFD < 1) {
                                                                    if (mu_mv_mag_sq < 61) {
                                                                        if (TSC < 8) {
                                                                            return 4;
                                                                        }
                                                                        else if (TSC >= 8) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                    else if (mu_mv_mag_sq >= 61) {
                                                                        if (refDCval < 93) {
                                                                            if (Cs < 2450) {
                                                                                return 4;
                                                                            }
                                                                            else if (Cs >= 2450) {
                                                                                return 8;
                                                                            }
                                                                        }
                                                                        else if (refDCval >= 93) {
                                                                            if (tcor < 89) {
                                                                                return 4;
                                                                            }
                                                                            else if (tcor >= 89) {
                                                                                return 1;
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                                else if (diffAFD >= 1) {
                                                                    if (Cs < 2236) {
                                                                        return 1;
                                                                    }
                                                                    else if (Cs >= 2236) {
                                                                        return 8;
                                                                    }
                                                                }
                                                            }
                                                            else if (MVDiff >= 10) {
                                                                if (mu_mv_mag_sq < 61) {
                                                                    return 1;
                                                                }
                                                                else if (mu_mv_mag_sq >= 61) {
                                                                    if (Rs < 2962) {
                                                                        if (Cs < 2297) {
                                                                            return 1;
                                                                        }
                                                                        else if (Cs >= 2297) {
                                                                            if (diffMVdiffVal < 1) {
                                                                                if (Cs < 2380) {
                                                                                    return 1;
                                                                                }
                                                                                else if (Cs >= 2380) {
                                                                                    return 4;
                                                                                }
                                                                            }
                                                                            else if (diffMVdiffVal >= 1) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (Rs >= 2962) {
                                                                        return 1;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (RsDiff >= 12) {
                                                    if (AFD < 13) {
                                                        return 4;
                                                    }
                                                    else if (AFD >= 13) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                        }
                                        else if (MVDiff >= 26) {
                                            if (Cs < 2231) {
                                                return 2;
                                            }
                                            else if (Cs >= 2231) {
                                                if (MVDiff < 30) {
                                                    if (Cs < 2349) {
                                                        return 2;
                                                    }
                                                    else if (Cs >= 2349) {
                                                        return 1;
                                                    }
                                                }
                                                else if (MVDiff >= 30) {
                                                    return 4;
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                        }
                    }
                    else if (Rs >= 3179) {
                        if (CsDiff < 10) {
                            if (ssDCval < 108) {
                                if (AFD < 7) {
                                    if (posBalance < 55) {
                                        return 8;
                                    }
                                    else if (posBalance >= 55) {
                                        return 4;
                                    }
                                }
                                else if (AFD >= 7) {
                                    if (TSCindex < 8) {
                                        if (tcor < 91) {
                                            if (negBalance < 35) {
                                                if (CsDiff < 5) {
                                                    return 4;
                                                }
                                                else if (CsDiff >= 5) {
                                                    return 8;
                                                }
                                            }
                                            else if (negBalance >= 35) {
                                                return 8;
                                            }
                                        }
                                        else if (tcor >= 91) {
                                            if (diffAFD < -1) {
                                                if (diffMVdiffVal < 6) {
                                                    if (Cs < 3357) {
                                                        return 8;
                                                    }
                                                    else if (Cs >= 3357) {
                                                        return 4;
                                                    }
                                                }
                                                else if (diffMVdiffVal >= 6) {
                                                    return 4;
                                                }
                                            }
                                            else if (diffAFD >= -1) {
                                                if (RsCsDiff < 11) {
                                                    if (MVDiff < 26) {
                                                        return 4;
                                                    }
                                                    else if (MVDiff >= 26) {
                                                        return 2;
                                                    }
                                                }
                                                else if (RsCsDiff >= 11) {
                                                    if (RsCsDiff < 16) {
                                                        return 8;
                                                    }
                                                    else if (RsCsDiff >= 16) {
                                                        if (RsDiff < 4) {
                                                            return 4;
                                                        }
                                                        else if (RsDiff >= 4) {
                                                            if (gchDC < 1) {
                                                                if (RsCsDiff < 57) {
                                                                    return 8;
                                                                }
                                                                else if (RsCsDiff >= 57) {
                                                                    if (Cs < 1808) {
                                                                        return 4;
                                                                    }
                                                                    else if (Cs >= 1808) {
                                                                        return 8;
                                                                    }
                                                                }
                                                            }
                                                            else if (gchDC >= 1) {
                                                                if (posBalance < 53) {
                                                                    return 4;
                                                                }
                                                                else if (posBalance >= 53) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (TSCindex >= 8) {
                                        if (RsDiff < 5) {
                                            if (refDCval < 48) {
                                                return 8;
                                            }
                                            else if (refDCval >= 48) {
                                                if (ssDCval < 51) {
                                                    if (negBalance < 49) {
                                                        return 8;
                                                    }
                                                    else if (negBalance >= 49) {
                                                        if (posBalance < 52) {
                                                            return 8;
                                                        }
                                                        else if (posBalance >= 52) {
                                                            return 2;
                                                        }
                                                    }
                                                }
                                                else if (ssDCval >= 51) {
                                                    return 2;
                                                }
                                            }
                                        }
                                        else if (RsDiff >= 5) {
                                            if (diffMVdiffVal < -3) {
                                                if (Rs < 4213) {
                                                    if (Cs < 2187) {
                                                        return 4;
                                                    }
                                                    else if (Cs >= 2187) {
                                                        if (SC < 6650) {
                                                            return 2;
                                                        }
                                                        else if (SC >= 6650) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                                else if (Rs >= 4213) {
                                                    return 8;
                                                }
                                            }
                                            else if (diffMVdiffVal >= -3) {
                                                if (TSC < 11) {
                                                    if (negBalance < 37) {
                                                        return 4;
                                                    }
                                                    else if (negBalance >= 37) {
                                                        if (ssDCval < 91) {
                                                            if (Cs < 3176) {
                                                                return 4;
                                                            }
                                                            else if (Cs >= 3176) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (ssDCval >= 91) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                                else if (TSC >= 11) {
                                                    if (Rs < 3256) {
                                                        return 4;
                                                    }
                                                    else if (Rs >= 3256) {
                                                        return 1;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                            else if (ssDCval >= 108) {
                                if (AFD < 12) {
                                    if (Cs < 2068) {
                                        if (refDCval < 109) {
                                            if (MVDiff < 9) {
                                                return 8;
                                            }
                                            else if (MVDiff >= 9) {
                                                if (TSC < 7) {
                                                    if (Cs < 1610) {
                                                        if (diffMVdiffVal < 1) {
                                                            if (Rs < 3414) {
                                                                return 4;
                                                            }
                                                            else if (Rs >= 3414) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (diffMVdiffVal >= 1) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (Cs >= 1610) {
                                                        return 4;
                                                    }
                                                }
                                                else if (TSC >= 7) {
                                                    return 8;
                                                }
                                            }
                                        }
                                        else if (refDCval >= 109) {
                                            if (ssDCval < 109) {
                                                if (Cs < 1926) {
                                                    return 4;
                                                }
                                                else if (Cs >= 1926) {
                                                    return 2;
                                                }
                                            }
                                            else if (ssDCval >= 109) {
                                                return 4;
                                            }
                                        }
                                    }
                                    else if (Cs >= 2068) {
                                        if (ssDCval < 128) {
                                            if (ssDCval < 119) {
                                                if (mcTcor < 95) {
                                                    if (MVDiff < 12) {
                                                        if (refDCval < 112) {
                                                            if (Rs < 4781) {
                                                                return 8;
                                                            }
                                                            else if (Rs >= 4781) {
                                                                return 2;
                                                            }
                                                        }
                                                        else if (refDCval >= 112) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (MVDiff >= 12) {
                                                        if (refDCval < 116) {
                                                            return 4;
                                                        }
                                                        else if (refDCval >= 116) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                                else if (mcTcor >= 95) {
                                                    if (tcor < 93) {
                                                        if (Rs < 3846) {
                                                            return 2;
                                                        }
                                                        else if (Rs >= 3846) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (tcor >= 93) {
                                                        if (diffRsCsdiff < -74) {
                                                            return 4;
                                                        }
                                                        else if (diffRsCsdiff >= -74) {
                                                            if (mu_mv_mag_sq < 68) {
                                                                if (Cs < 2139) {
                                                                    return 4;
                                                                }
                                                                else if (Cs >= 2139) {
                                                                    return 2;
                                                                }
                                                            }
                                                            else if (mu_mv_mag_sq >= 68) {
                                                                return 2;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                            else if (ssDCval >= 119) {
                                                if (ssDCval < 125) {
                                                    if (Cs < 2458) {
                                                        if (Cs < 2247) {
                                                            return 4;
                                                        }
                                                        else if (Cs >= 2247) {
                                                            if (ssDCval < 121) {
                                                                return 8;
                                                            }
                                                            else if (ssDCval >= 121) {
                                                                if (SC < 6304) {
                                                                    return 8;
                                                                }
                                                                else if (SC >= 6304) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (Cs >= 2458) {
                                                        if (gchDC < 1) {
                                                            return 8;
                                                        }
                                                        else if (gchDC >= 1) {
                                                            if (SC < 7175) {
                                                                return 4;
                                                            }
                                                            else if (SC >= 7175) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (ssDCval >= 125) {
                                                    if (Rs < 4605) {
                                                        return 4;
                                                    }
                                                    else if (Rs >= 4605) {
                                                        return 2;
                                                    }
                                                }
                                            }
                                        }
                                        else if (ssDCval >= 128) {
                                            if (negBalance < 36) {
                                                return 2;
                                            }
                                            else if (negBalance >= 36) {
                                                if (negBalance < 44) {
                                                    if (Cs < 2824) {
                                                        return 2;
                                                    }
                                                    else if (Cs >= 2824) {
                                                        return 4;
                                                    }
                                                }
                                                else if (negBalance >= 44) {
                                                    return 8;
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (AFD >= 12) {
                                    if (mu_mv_mag_sq < 34) {
                                        return 2;
                                    }
                                    else if (mu_mv_mag_sq >= 34) {
                                        if (TSC < 9) {
                                            return 4;
                                        }
                                        else if (TSC >= 9) {
                                            return 8;
                                        }
                                    }
                                }
                            }
                        }
                        else if (CsDiff >= 10) {
                            if (negBalance < 47) {
                                return 4;
                            }
                            else if (negBalance >= 47) {
                                return 8;
                            }
                        }
                    }
                }
            }
            else if (Cs >= 3432) {
                if (ssDCval < 76) {
                    if (mcTcor < 80) {
                        if (Scindex < 6) {
                            return 4;
                        }
                        else if (Scindex >= 6) {
                            if (Cs < 5542) {
                                return 8;
                            }
                            else if (Cs >= 5542) {
                                return 4;
                            }
                        }
                    }
                    else if (mcTcor >= 80) {
                        if (Rs < 3991) {
                            if (mu_mv_mag_sq < 48) {
                                if (ssDCval < 60) {
                                    if (RsCsDiff < 46) {
                                        if (TSCindex < 9) {
                                            if (MVDiff < 17) {
                                                if (Cs < 3734) {
                                                    if (refDCval < 55) {
                                                        return 4;
                                                    }
                                                    else if (refDCval >= 55) {
                                                        if (Cs < 3517) {
                                                            return 2;
                                                        }
                                                        else if (Cs >= 3517) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                                else if (Cs >= 3734) {
                                                    if (refDCval < 55) {
                                                        return 2;
                                                    }
                                                    else if (refDCval >= 55) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                            else if (MVDiff >= 17) {
                                                if (diffRsCsdiff < -6) {
                                                    if (ssDCval < 50) {
                                                        return 4;
                                                    }
                                                    else if (ssDCval >= 50) {
                                                        if (SC < 7242) {
                                                            return 8;
                                                        }
                                                        else if (SC >= 7242) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                                else if (diffRsCsdiff >= -6) {
                                                    if (RsCsDiff < 37) {
                                                        if (negBalance < 43) {
                                                            return 8;
                                                        }
                                                        else if (negBalance >= 43) {
                                                            return 2;
                                                        }
                                                    }
                                                    else if (RsCsDiff >= 37) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                        }
                                        else if (TSCindex >= 9) {
                                            if (MVDiff < 15) {
                                                return 4;
                                            }
                                            else if (MVDiff >= 15) {
                                                if (negBalance < 47) {
                                                    return 8;
                                                }
                                                else if (negBalance >= 47) {
                                                    return 2;
                                                }
                                            }
                                        }
                                    }
                                    else if (RsCsDiff >= 46) {
                                        if (TSC < 7) {
                                            return 8;
                                        }
                                        else if (TSC >= 7) {
                                            if (MVDiff < 59) {
                                                return 4;
                                            }
                                            else if (MVDiff >= 59) {
                                                return 8;
                                            }
                                        }
                                    }
                                }
                                else if (ssDCval >= 60) {
                                    return 8;
                                }
                            }
                            else if (mu_mv_mag_sq >= 48) {
                                if (SC < 7151) {
                                    if (refDCval < 53) {
                                        return 4;
                                    }
                                    else if (refDCval >= 53) {
                                        return 8;
                                    }
                                }
                                else if (SC >= 7151) {
                                    if (TSC < 7) {
                                        if (mu_mv_mag_sq < 77) {
                                            if (tcor < 90) {
                                                return 8;
                                            }
                                            else if (tcor >= 90) {
                                                return 4;
                                            }
                                        }
                                        else if (mu_mv_mag_sq >= 77) {
                                            if (ssDCval < 54) {
                                                return 4;
                                            }
                                            else if (ssDCval >= 54) {
                                                if (diffTSC < -1) {
                                                    if (SC < 7437) {
                                                        return 4;
                                                    }
                                                    else if (SC >= 7437) {
                                                        if (ssDCval < 56) {
                                                            return 8;
                                                        }
                                                        else if (ssDCval >= 56) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                                else if (diffTSC >= -1) {
                                                    return 8;
                                                }
                                            }
                                        }
                                    }
                                    else if (TSC >= 7) {
                                        if (diffMVdiffVal < 11) {
                                            if (Rs < 3663) {
                                                if (Rs < 3606) {
                                                    return 4;
                                                }
                                                else if (Rs >= 3606) {
                                                    return 8;
                                                }
                                            }
                                            else if (Rs >= 3663) {
                                                return 4;
                                            }
                                        }
                                        else if (diffMVdiffVal >= 11) {
                                            return 8;
                                        }
                                    }
                                }
                            }
                        }
                        else if (Rs >= 3991) {
                            if (TSC < 7) {
                                if (mu_mv_mag_sq < 50) {
                                    return 8;
                                }
                                else if (mu_mv_mag_sq >= 50) {
                                    if (Cs < 3971) {
                                        return 4;
                                    }
                                    else if (Cs >= 3971) {
                                        return 8;
                                    }
                                }
                            }
                            else if (TSC >= 7) {
                                if (ssDCval < 68) {
                                    if (MVDiff < 73) {
                                        if (diffAFD < 5) {
                                            return 4;
                                        }
                                        else if (diffAFD >= 5) {
                                            if (Rs < 4109) {
                                                return 4;
                                            }
                                            else if (Rs >= 4109) {
                                                return 8;
                                            }
                                        }
                                    }
                                    else if (MVDiff >= 73) {
                                        return 8;
                                    }
                                }
                                else if (ssDCval >= 68) {
                                    if (AFD < 14) {
                                        return 4;
                                    }
                                    else if (AFD >= 14) {
                                        if (RsCsDiff < 358) {
                                            return 8;
                                        }
                                        else if (RsCsDiff >= 358) {
                                            if (mu_mv_mag_sq < 63) {
                                                if (Cs < 7255) {
                                                    if (Cs < 5495) {
                                                        if (Rs < 5191) {
                                                            return 4;
                                                        }
                                                        else if (Rs >= 5191) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (Cs >= 5495) {
                                                        return 4;
                                                    }
                                                }
                                                else if (Cs >= 7255) {
                                                    return 8;
                                                }
                                            }
                                            else if (mu_mv_mag_sq >= 63) {
                                                return 8;
                                            }
                                        }
                                    }
                                }
                            }
                        }
                    }
                }
                else if (ssDCval >= 76) {
                    if (mu_mv_mag_sq < 61) {
                        if (SC < 12042) {
                            if (refDCval < 157) {
                                if (RsCsDiff < 882) {
                                    if (mu_mv_mag_sq < 33) {
                                        if (Cs < 5032) {
                                            if (posBalance < 53) {
                                                if (mcTcor < 77) {
                                                    return 4;
                                                }
                                                else if (mcTcor >= 77) {
                                                    return 8;
                                                }
                                            }
                                            else if (posBalance >= 53) {
                                                if (diffAFD < 2) {
                                                    if (tcor < 77) {
                                                        if (Rs < 4223) {
                                                            if (tcor < 45) {
                                                                return 4;
                                                            }
                                                            else if (tcor >= 45) {
                                                                if (TSC < 14) {
                                                                    return 8;
                                                                }
                                                                else if (TSC >= 14) {
                                                                    return 1;
                                                                }
                                                            }
                                                        }
                                                        else if (Rs >= 4223) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (tcor >= 77) {
                                                        if (Rs < 5566) {
                                                            return 8;
                                                        }
                                                        else if (Rs >= 5566) {
                                                            return 1;
                                                        }
                                                    }
                                                }
                                                else if (diffAFD >= 2) {
                                                    if (Rs < 4116) {
                                                        return 1;
                                                    }
                                                    else if (Rs >= 4116) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                        }
                                        else if (Cs >= 5032) {
                                            return 1;
                                        }
                                    }
                                    else if (mu_mv_mag_sq >= 33) {
                                        if (SC < 7634) {
                                            if (AFD < 22) {
                                                return 1;
                                            }
                                            else if (AFD >= 22) {
                                                return 8;
                                            }
                                        }
                                        else if (SC >= 7634) {
                                            if (posBalance < 56) {
                                                if (Cs < 4075) {
                                                    if (AFD < 18) {
                                                        if (Rs < 5103) {
                                                            return 4;
                                                        }
                                                        else if (Rs >= 5103) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (AFD >= 18) {
                                                        if (Rs < 5323) {
                                                            return 8;
                                                        }
                                                        else if (Rs >= 5323) {
                                                            if (MVDiff < 22) {
                                                                return 4;
                                                            }
                                                            else if (MVDiff >= 22) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (Cs >= 4075) {
                                                    if (RsCsDiff < 794) {
                                                        return 8;
                                                    }
                                                    else if (RsCsDiff >= 794) {
                                                        if (Rs < 6677) {
                                                            if (SC < 10033) {
                                                                return 4;
                                                            }
                                                            else if (SC >= 10033) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (Rs >= 6677) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (posBalance >= 56) {
                                                if (RsCsDiff < 565) {
                                                    if (Cs < 4232) {
                                                        if (RsDiff < 16) {
                                                            return 4;
                                                        }
                                                        else if (RsDiff >= 16) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (Cs >= 4232) {
                                                        return 1;
                                                    }
                                                }
                                                else if (RsCsDiff >= 565) {
                                                    return 8;
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (RsCsDiff >= 882) {
                                    if (posBalance < 55) {
                                        if (mu_mv_mag_sq < 33) {
                                            if (MVDiff < 22) {
                                                if (TSC < 20) {
                                                    return 8;
                                                }
                                                else if (TSC >= 20) {
                                                    return 4;
                                                }
                                            }
                                            else if (MVDiff >= 22) {
                                                return 4;
                                            }
                                        }
                                        else if (mu_mv_mag_sq >= 33) {
                                            if (CsDiff < 19) {
                                                if (Cs < 4120) {
                                                    return 4;
                                                }
                                                else if (Cs >= 4120) {
                                                    return 8;
                                                }
                                            }
                                            else if (CsDiff >= 19) {
                                                return 8;
                                            }
                                        }
                                    }
                                    else if (posBalance >= 55) {
                                        if (Rs < 6395) {
                                            if (TSC < 16) {
                                                if (AFD < 31) {
                                                    if (Rs < 5504) {
                                                        return 8;
                                                    }
                                                    else if (Rs >= 5504) {
                                                        return 1;
                                                    }
                                                }
                                                else if (AFD >= 31) {
                                                    return 2;
                                                }
                                            }
                                            else if (TSC >= 16) {
                                                return 8;
                                            }
                                        }
                                        else if (Rs >= 6395) {
                                            return 2;
                                        }
                                    }
                                }
                            }
                            else if (refDCval >= 157) {
                                if (tcor < 76) {
                                    if (Cs < 3975) {
                                        return 4;
                                    }
                                    else if (Cs >= 3975) {
                                        return 2;
                                    }
                                }
                                else if (tcor >= 76) {
                                    if (Rs < 4900) {
                                        return 8;
                                    }
                                    else if (Rs >= 4900) {
                                        return 2;
                                    }
                                }
                            }
                        }
                        else if (SC >= 12042) {
                            if (mcTcor < 85) {
                                if (ssDCval < 93) {
                                    return 4;
                                }
                                else if (ssDCval >= 93) {
                                    if (MVDiff < 21) {
                                        if (mu_mv_mag_sq < 3) {
                                            if (RsCsDiff < 6195) {
                                                return 8;
                                            }
                                            else if (RsCsDiff >= 6195) {
                                                return 4;
                                            }
                                        }
                                        else if (mu_mv_mag_sq >= 3) {
                                            if (diffRsCsdiff < -442) {
                                                return 4;
                                            }
                                            else if (diffRsCsdiff >= -442) {
                                                if (diffMVdiffVal < 2) {
                                                    if (Cs < 6412) {
                                                        return 4;
                                                    }
                                                    else if (Cs >= 6412) {
                                                        return 8;
                                                    }
                                                }
                                                else if (diffMVdiffVal >= 2) {
                                                    if (RsDiff < 40) {
                                                        return 8;
                                                    }
                                                    else if (RsDiff >= 40) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (MVDiff >= 21) {
                                        if (negBalance < 58) {
                                            return 8;
                                        }
                                        else if (negBalance >= 58) {
                                            if (ssDCval < 154) {
                                                if (RsDiff < 49) {
                                                    return 8;
                                                }
                                                else if (RsDiff >= 49) {
                                                    return 2;
                                                }
                                            }
                                            else if (ssDCval >= 154) {
                                                if (Rs < 7779) {
                                                    return 2;
                                                }
                                                else if (Rs >= 7779) {
                                                    return 8;
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                            else if (mcTcor >= 85) {
                                if (RsDiff < 39) {
                                    if (RsCsDiff < 1000) {
                                        if (CsDiff < 19) {
                                            return 4;
                                        }
                                        else if (CsDiff >= 19) {
                                            if (refDCval < 149) {
                                                return 4;
                                            }
                                            else if (refDCval >= 149) {
                                                return 2;
                                            }
                                        }
                                    }
                                    else if (RsCsDiff >= 1000) {
                                        if (mu_mv_mag_sq < 7) {
                                            if (CsDiff < 24) {
                                                if (MVDiff < 12) {
                                                    return 4;
                                                }
                                                else if (MVDiff >= 12) {
                                                    if (Cs < 5278) {
                                                        if (Cs < 4645) {
                                                            return 4;
                                                        }
                                                        else if (Cs >= 4645) {
                                                            return 2;
                                                        }
                                                    }
                                                    else if (Cs >= 5278) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                            else if (CsDiff >= 24) {
                                                return 8;
                                            }
                                        }
                                        else if (mu_mv_mag_sq >= 7) {
                                            if (refDCval < 136) {
                                                if (TSC < 20) {
                                                    if (posBalance < 60) {
                                                        return 2;
                                                    }
                                                    else if (posBalance >= 60) {
                                                        return 4;
                                                    }
                                                }
                                                else if (TSC >= 20) {
                                                    return 8;
                                                }
                                            }
                                            else if (refDCval >= 136) {
                                                if (ssDCval < 141) {
                                                    if (TSC < 19) {
                                                        return 4;
                                                    }
                                                    else if (TSC >= 19) {
                                                        if (Rs < 7961) {
                                                            return 8;
                                                        }
                                                        else if (Rs >= 7961) {
                                                            return 2;
                                                        }
                                                    }
                                                }
                                                else if (ssDCval >= 141) {
                                                    return 4;
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (RsDiff >= 39) {
                                    return 2;
                                }
                            }
                        }
                    }
                    else if (mu_mv_mag_sq >= 61) {
                        if (refDCval < 136) {
                            if (Cs < 4110) {
                                if (ssDCval < 121) {
                                    if (mu_mv_mag_sq < 73) {
                                        return 4;
                                    }
                                    else if (mu_mv_mag_sq >= 73) {
                                        if (diffMVdiffVal < -1) {
                                            if (Rs < 3883) {
                                                return 8;
                                            }
                                            else if (Rs >= 3883) {
                                                return 4;
                                            }
                                        }
                                        else if (diffMVdiffVal >= -1) {
                                            return 8;
                                        }
                                    }
                                }
                                else if (ssDCval >= 121) {
                                    return 8;
                                }
                            }
                            else if (Cs >= 4110) {
                                return 8;
                            }
                        }
                        else if (refDCval >= 136) {
                            if (mcTcor < 80) {
                                if (Rs < 4972) {
                                    return 1;
                                }
                                else if (Rs >= 4972) {
                                    return 8;
                                }
                            }
                            else if (mcTcor >= 80) {
                                return 8;
                            }
                        }
                    }
                }
            }
        }
        return 4;
    }
    mfxU16 AGOPSelectA(mfxU32 Rs, mfxU32 CsDiff, mfxU32 TSC, mfxI32 diffAFD, mfxU32 Scindex, mfxU32 RsDiff, mfxI32 diffMVdiffVal, mfxU32 RsCsDiff, mfxU32 TSCindex, mfxI16 tcor, mfxU32 refDCval, mfxU32 MVDiff, mfxU32 AFD, mfxI32 diffTSC, mfxI32 diffRsCsdiff, mfxU32 posBalance, mfxU32 SC, mfxU32 mu_mv_mag_sq, mfxU32 negBalance, mfxI16 mcTcor, mfxU32 gchDC, mfxU32 Cs, mfxU32 ssDCval) {
        if (mu_mv_mag_sq < 88) {
            if (posBalance < 5) {
                if (refDCval < 18) {
                    if (AFD < 1) {
                        if (Cs < 1772) {
                            if (SC < 729) {
                                return 4;
                            }
                            else if (SC >= 729) {
                                return 8;
                            }
                        }
                        else if (Cs >= 1772) {
                            if (Rs < 4729) {
                                if (Rs < 2707) {
                                    return 1;
                                }
                                else if (Rs >= 2707) {
                                    if (Rs < 3755) {
                                        return 8;
                                    }
                                    else if (Rs >= 3755) {
                                        if (Cs < 3985) {
                                            if (Rs < 3791) {
                                                return 1;
                                            }
                                            else if (Rs >= 3791) {
                                                return 8;
                                            }
                                        }
                                        else if (Cs >= 3985) {
                                            return 8;
                                        }
                                    }
                                }
                            }
                            else if (Rs >= 4729) {
                                return 1;
                            }
                        }
                    }
                    else if (AFD >= 1) {
                        return 1;
                    }
                }
                else if (refDCval >= 18) {
                    if (MVDiff < 13) {
                        return 8;
                    }
                    else if (MVDiff >= 13) {
                        if (tcor < 118) {
                            if (MVDiff < 365) {
                                if (refDCval < 22) {
                                    if (negBalance < 55) {
                                        if (posBalance < 1) {
                                            return 4;
                                        }
                                        else if (posBalance >= 1) {
                                            if (mu_mv_mag_sq < 15) {
                                                return 8;
                                            }
                                            else if (mu_mv_mag_sq >= 15) {
                                                return 4;
                                            }
                                        }
                                    }
                                    else if (negBalance >= 55) {
                                        if (Rs < 885) {
                                            return 1;
                                        }
                                        else if (Rs >= 885) {
                                            return 2;
                                        }
                                    }
                                }
                                else if (refDCval >= 22) {
                                    return 8;
                                }
                            }
                            else if (MVDiff >= 365) {
                                if (MVDiff < 391) {
                                    return 2;
                                }
                                else if (MVDiff >= 391) {
                                    if (diffMVdiffVal < 828) {
                                        return 4;
                                    }
                                    else if (diffMVdiffVal >= 828) {
                                        return 2;
                                    }
                                }
                            }
                        }
                        else if (tcor >= 118) {
                            if (tcor < 132) {
                                if (mu_mv_mag_sq < 41) {
                                    if (posBalance < 3) {
                                        return 4;
                                    }
                                    else if (posBalance >= 3) {
                                        if (Rs < 17) {
                                            return 2;
                                        }
                                        else if (Rs >= 17) {
                                            if (Rs < 117) {
                                                return 4;
                                            }
                                            else if (Rs >= 117) {
                                                return 2;
                                            }
                                        }
                                    }
                                }
                                else if (mu_mv_mag_sq >= 41) {
                                    if (posBalance < 1) {
                                        return 2;
                                    }
                                    else if (posBalance >= 1) {
                                        return 1;
                                    }
                                }
                            }
                            else if (tcor >= 132) {
                                if (diffMVdiffVal < 3) {
                                    if (AFD < 1) {
                                        return 8;
                                    }
                                    else if (AFD >= 1) {
                                        return 2;
                                    }
                                }
                                else if (diffMVdiffVal >= 3) {
                                    if (ssDCval < 22) {
                                        return 4;
                                    }
                                    else if (ssDCval >= 22) {
                                        if (Cs < 2) {
                                            return 8;
                                        }
                                        else if (Cs >= 2) {
                                            if (SC < 22) {
                                                return 2;
                                            }
                                            else if (SC >= 22) {
                                                return 8;
                                            }
                                        }
                                    }
                                }
                            }
                        }
                    }
                }
            }
            else if (posBalance >= 5) {
                if (Scindex < 6) {
                    if (AFD < 12) {
                        if (ssDCval < 104) {
                            if (Cs < 272) {
                                if (TSC < 5) {
                                    if (negBalance < 24) {
                                        if (mcTcor < 68) {
                                            if (posBalance < 28) {
                                                if (negBalance < 2) {
                                                    if (mcTcor < 44) {
                                                        return 1;
                                                    }
                                                    else if (mcTcor >= 44) {
                                                        if (MVDiff < 62) {
                                                            return 1;
                                                        }
                                                        else if (MVDiff >= 62) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                                else if (negBalance >= 2) {
                                                    if (tcor < 43) {
                                                        if (MVDiff < 66) {
                                                            if (tcor < 41) {
                                                                return 4;
                                                            }
                                                            else if (tcor >= 41) {
                                                                return 1;
                                                            }
                                                        }
                                                        else if (MVDiff >= 66) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (tcor >= 43) {
                                                        if (diffMVdiffVal < -14) {
                                                            if (MVDiff < 67) {
                                                                return 4;
                                                            }
                                                            else if (MVDiff >= 67) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (diffMVdiffVal >= -14) {
                                                            if (MVDiff < 77) {
                                                                return 8;
                                                            }
                                                            else if (MVDiff >= 77) {
                                                                if (MVDiff < 82) {
                                                                    return 4;
                                                                }
                                                                else if (MVDiff >= 82) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                            else if (posBalance >= 28) {
                                                if (ssDCval < 20) {
                                                    if (tcor < 29) {
                                                        if (AFD < 2) {
                                                            if (mcTcor < 33) {
                                                                return 4;
                                                            }
                                                            else if (mcTcor >= 33) {
                                                                return 2;
                                                            }
                                                        }
                                                        else if (AFD >= 2) {
                                                            return 1;
                                                        }
                                                    }
                                                    else if (tcor >= 29) {
                                                        if (diffAFD < -1) {
                                                            if (MVDiff < 200) {
                                                                return 2;
                                                            }
                                                            else if (MVDiff >= 200) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (diffAFD >= -1) {
                                                            if (tcor < 36) {
                                                                if (MVDiff < 172) {
                                                                    return 8;
                                                                }
                                                                else if (MVDiff >= 172) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (tcor >= 36) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (ssDCval >= 20) {
                                                    if (mu_mv_mag_sq < 7) {
                                                        if (posBalance < 87) {
                                                            return 2;
                                                        }
                                                        else if (posBalance >= 87) {
                                                            return 1;
                                                        }
                                                    }
                                                    else if (mu_mv_mag_sq >= 7) {
                                                        if (ssDCval < 27) {
                                                            if (posBalance < 119) {
                                                                return 4;
                                                            }
                                                            else if (posBalance >= 119) {
                                                                return 1;
                                                            }
                                                        }
                                                        else if (ssDCval >= 27) {
                                                            return 1;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (mcTcor >= 68) {
                                            if (tcor < 83) {
                                                if (posBalance < 85) {
                                                    if (mcTcor < 75) {
                                                        return 2;
                                                    }
                                                    else if (mcTcor >= 75) {
                                                        if (mu_mv_mag_sq < 40) {
                                                            if (Cs < 36) {
                                                                if (ssDCval < 21) {
                                                                    return 2;
                                                                }
                                                                else if (ssDCval >= 21) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (Cs >= 36) {
                                                                return 2;
                                                            }
                                                        }
                                                        else if (mu_mv_mag_sq >= 40) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                                else if (posBalance >= 85) {
                                                    if (refDCval < 22) {
                                                        return 8;
                                                    }
                                                    else if (refDCval >= 22) {
                                                        if (mu_mv_mag_sq < 28) {
                                                            if (posBalance < 114) {
                                                                if (mcTcor < 79) {
                                                                    if (Rs < 18) {
                                                                        return 2;
                                                                    }
                                                                    else if (Rs >= 18) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (mcTcor >= 79) {
                                                                    if (negBalance < 4) {
                                                                        return 4;
                                                                    }
                                                                    else if (negBalance >= 4) {
                                                                        if (Rs < 186) {
                                                                            return 8;
                                                                        }
                                                                        else if (Rs >= 186) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                            else if (posBalance >= 114) {
                                                                return 2;
                                                            }
                                                        }
                                                        else if (mu_mv_mag_sq >= 28) {
                                                            if (AFD < 5) {
                                                                return 4;
                                                            }
                                                            else if (AFD >= 5) {
                                                                if (Cs < 83) {
                                                                    return 8;
                                                                }
                                                                else if (Cs >= 83) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                            else if (tcor >= 83) {
                                                if (posBalance < 99) {
                                                    if (diffRsCsdiff < 4) {
                                                        if (mu_mv_mag_sq < 59) {
                                                            if (mu_mv_mag_sq < 15) {
                                                                if (gchDC < 1) {
                                                                    return 8;
                                                                }
                                                                else if (gchDC >= 1) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (mu_mv_mag_sq >= 15) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (mu_mv_mag_sq >= 59) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (diffRsCsdiff >= 4) {
                                                        return 2;
                                                    }
                                                }
                                                else if (posBalance >= 99) {
                                                    return 8;
                                                }
                                            }
                                        }
                                    }
                                    else if (negBalance >= 24) {
                                        if (mcTcor < 80) {
                                            if (negBalance < 41) {
                                                if (ssDCval < 34) {
                                                    return 8;
                                                }
                                                else if (ssDCval >= 34) {
                                                    return 1;
                                                }
                                            }
                                            else if (negBalance >= 41) {
                                                return 1;
                                            }
                                        }
                                        else if (mcTcor >= 80) {
                                            if (tcor < 96) {
                                                if (MVDiff < 49) {
                                                    if (ssDCval < 35) {
                                                        if (Cs < 102) {
                                                            return 4;
                                                        }
                                                        else if (Cs >= 102) {
                                                            return 1;
                                                        }
                                                    }
                                                    else if (ssDCval >= 35) {
                                                        return 2;
                                                    }
                                                }
                                                else if (MVDiff >= 49) {
                                                    if (ssDCval < 70) {
                                                        return 1;
                                                    }
                                                    else if (ssDCval >= 70) {
                                                        return 2;
                                                    }
                                                }
                                            }
                                            else if (tcor >= 96) {
                                                if (MVDiff < 31) {
                                                    if (MVDiff < 5) {
                                                        return 8;
                                                    }
                                                    else if (MVDiff >= 5) {
                                                        if (diffTSC < -1) {
                                                            if (Cs < 123) {
                                                                return 4;
                                                            }
                                                            else if (Cs >= 123) {
                                                                return 2;
                                                            }
                                                        }
                                                        else if (diffTSC >= -1) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                                else if (MVDiff >= 31) {
                                                    return 2;
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (TSC >= 5) {
                                    if (MVDiff < 17) {
                                        if (ssDCval < 83) {
                                            if (refDCval < 72) {
                                                return 2;
                                            }
                                            else if (refDCval >= 72) {
                                                return 8;
                                            }
                                        }
                                        else if (ssDCval >= 83) {
                                            return 1;
                                        }
                                    }
                                    else if (MVDiff >= 17) {
                                        if (Cs < 251) {
                                            if (tcor < 111) {
                                                if (diffMVdiffVal < 24) {
                                                    return 1;
                                                }
                                                else if (diffMVdiffVal >= 24) {
                                                    return 2;
                                                }
                                            }
                                            else if (tcor >= 111) {
                                                return 2;
                                            }
                                        }
                                        else if (Cs >= 251) {
                                            if (Cs < 267) {
                                                return 2;
                                            }
                                            else if (Cs >= 267) {
                                                return 1;
                                            }
                                        }
                                    }
                                }
                            }
                            else if (Cs >= 272) {
                                if (Rs < 675) {
                                    if (AFD < 9) {
                                        if (Rs < 270) {
                                            if (mcTcor < 90) {
                                                if (posBalance < 51) {
                                                    return 4;
                                                }
                                                else if (posBalance >= 51) {
                                                    if (ssDCval < 88) {
                                                        return 2;
                                                    }
                                                    else if (ssDCval >= 88) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                            else if (mcTcor >= 90) {
                                                if (CsDiff < 2) {
                                                    if (Cs < 386) {
                                                        if (ssDCval < 34) {
                                                            return 2;
                                                        }
                                                        else if (ssDCval >= 34) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (Cs >= 386) {
                                                        return 2;
                                                    }
                                                }
                                                else if (CsDiff >= 2) {
                                                    return 2;
                                                }
                                            }
                                        }
                                        else if (Rs >= 270) {
                                            if (Cs < 383) {
                                                if (mcTcor < 86) {
                                                    if (Rs < 406) {
                                                        if (ssDCval < 38) {
                                                            if (mu_mv_mag_sq < 65) {
                                                                if (mu_mv_mag_sq < 54) {
                                                                    if (SC < 731) {
                                                                        return 8;
                                                                    }
                                                                    else if (SC >= 731) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (mu_mv_mag_sq >= 54) {
                                                                    if (diffTSC < 1) {
                                                                        return 4;
                                                                    }
                                                                    else if (diffTSC >= 1) {
                                                                        return 1;
                                                                    }
                                                                }
                                                            }
                                                            else if (mu_mv_mag_sq >= 65) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (ssDCval >= 38) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (Rs >= 406) {
                                                        if (mcTcor < 82) {
                                                            if (diffRsCsdiff < 1) {
                                                                if (diffTSC < -1) {
                                                                    return 4;
                                                                }
                                                                else if (diffTSC >= -1) {
                                                                    return 1;
                                                                }
                                                            }
                                                            else if (diffRsCsdiff >= 1) {
                                                                if (MVDiff < 17) {
                                                                    return 4;
                                                                }
                                                                else if (MVDiff >= 17) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                        else if (mcTcor >= 82) {
                                                            if (ssDCval < 38) {
                                                                if (posBalance < 55) {
                                                                    if (tcor < 85) {
                                                                        return 4;
                                                                    }
                                                                    else if (tcor >= 85) {
                                                                        return 1;
                                                                    }
                                                                }
                                                                else if (posBalance >= 55) {
                                                                    return 1;
                                                                }
                                                            }
                                                            else if (ssDCval >= 38) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (mcTcor >= 86) {
                                                    if (mu_mv_mag_sq < 79) {
                                                        if (refDCval < 52) {
                                                            if (refDCval < 38) {
                                                                return 2;
                                                            }
                                                            else if (refDCval >= 38) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (refDCval >= 52) {
                                                            return 2;
                                                        }
                                                    }
                                                    else if (mu_mv_mag_sq >= 79) {
                                                        if (Rs < 334) {
                                                            return 4;
                                                        }
                                                        else if (Rs >= 334) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (Cs >= 383) {
                                                if (tcor < 87) {
                                                    if (ssDCval < 39) {
                                                        if (Rs < 460) {
                                                            if (TSCindex < 6) {
                                                                if (SC < 1274) {
                                                                    if (MVDiff < 19) {
                                                                        return 4;
                                                                    }
                                                                    else if (MVDiff >= 19) {
                                                                        if (Cs < 398) {
                                                                            if (negBalance < 43) {
                                                                                return 2;
                                                                            }
                                                                            else if (negBalance >= 43) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                        else if (Cs >= 398) {
                                                                            return 2;
                                                                        }
                                                                    }
                                                                }
                                                                else if (SC >= 1274) {
                                                                    return 2;
                                                                }
                                                            }
                                                            else if (TSCindex >= 6) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (Rs >= 460) {
                                                            if (mu_mv_mag_sq < 78) {
                                                                if (Cs < 399) {
                                                                    if (SC < 889) {
                                                                        return 2;
                                                                    }
                                                                    else if (SC >= 889) {
                                                                        return 8;
                                                                    }
                                                                }
                                                                else if (Cs >= 399) {
                                                                    return 2;
                                                                }
                                                            }
                                                            else if (mu_mv_mag_sq >= 78) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                    else if (ssDCval >= 39) {
                                                        if (Cs < 841) {
                                                            if (Rs < 277) {
                                                                return 2;
                                                            }
                                                            else if (Rs >= 277) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (Cs >= 841) {
                                                            return 2;
                                                        }
                                                    }
                                                }
                                                else if (tcor >= 87) {
                                                    if (Rs < 339) {
                                                        if (AFD < 5) {
                                                            return 4;
                                                        }
                                                        else if (AFD >= 5) {
                                                            return 2;
                                                        }
                                                    }
                                                    else if (Rs >= 339) {
                                                        if (posBalance < 68) {
                                                            if (tcor < 92) {
                                                                if (Cs < 460) {
                                                                    return 4;
                                                                }
                                                                else if (Cs >= 460) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (tcor >= 92) {
                                                                if (mu_mv_mag_sq < 78) {
                                                                    return 8;
                                                                }
                                                                else if (mu_mv_mag_sq >= 78) {
                                                                    if (Cs < 940) {
                                                                        return 4;
                                                                    }
                                                                    else if (Cs >= 940) {
                                                                        return 8;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (posBalance >= 68) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (AFD >= 9) {
                                        if (Cs < 681) {
                                            if (negBalance < 41) {
                                                if (SC < 594) {
                                                    return 2;
                                                }
                                                else if (SC >= 594) {
                                                    return 4;
                                                }
                                            }
                                            else if (negBalance >= 41) {
                                                if (Rs < 671) {
                                                    if (Cs < 520) {
                                                        return 8;
                                                    }
                                                    else if (Cs >= 520) {
                                                        if (Rs < 589) {
                                                            return 8;
                                                        }
                                                        else if (Rs >= 589) {
                                                            if (MVDiff < 10) {
                                                                if (mcTcor < 90) {
                                                                    return 8;
                                                                }
                                                                else if (mcTcor >= 90) {
                                                                    if (negBalance < 53) {
                                                                        return 8;
                                                                    }
                                                                    else if (negBalance >= 53) {
                                                                        return 4;
                                                                    }
                                                                }
                                                            }
                                                            else if (MVDiff >= 10) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (Rs >= 671) {
                                                    return 2;
                                                }
                                            }
                                        }
                                        else if (Cs >= 681) {
                                            if (mu_mv_mag_sq < 45) {
                                                return 2;
                                            }
                                            else if (mu_mv_mag_sq >= 45) {
                                                return 4;
                                            }
                                        }
                                    }
                                }
                                else if (Rs >= 675) {
                                    if (TSCindex < 8) {
                                        if (refDCval < 72) {
                                            if (ssDCval < 64) {
                                                if (refDCval < 45) {
                                                    if (TSC < 5) {
                                                        if (MVDiff < 22) {
                                                            if (diffAFD < 2) {
                                                                if (diffMVdiffVal < -5) {
                                                                    if (mu_mv_mag_sq < 76) {
                                                                        return 1;
                                                                    }
                                                                    else if (mu_mv_mag_sq >= 76) {
                                                                        if (Cs < 1489) {
                                                                            return 8;
                                                                        }
                                                                        else if (Cs >= 1489) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                }
                                                                else if (diffMVdiffVal >= -5) {
                                                                    if (SC < 3357) {
                                                                        if (MVDiff < 9) {
                                                                            return 4;
                                                                        }
                                                                        else if (MVDiff >= 9) {
                                                                            if (tcor < 89) {
                                                                                if (RsDiff < 6) {
                                                                                    if (Cs < 630) {
                                                                                        return 4;
                                                                                    }
                                                                                    else if (Cs >= 630) {
                                                                                        return 2;
                                                                                    }
                                                                                }
                                                                                else if (RsDiff >= 6) {
                                                                                    if (Rs < 1562) {
                                                                                        return 1;
                                                                                    }
                                                                                    else if (Rs >= 1562) {
                                                                                        return 4;
                                                                                    }
                                                                                }
                                                                            }
                                                                            else if (tcor >= 89) {
                                                                                if (Cs < 1493) {
                                                                                    if (Rs < 821) {
                                                                                        return 1;
                                                                                    }
                                                                                    else if (Rs >= 821) {
                                                                                        return 8;
                                                                                    }
                                                                                }
                                                                                else if (Cs >= 1493) {
                                                                                    if (Cs < 1530) {
                                                                                        if (Cs < 1504) {
                                                                                            return 1;
                                                                                        }
                                                                                        else if (Cs >= 1504) {
                                                                                            return 4;
                                                                                        }
                                                                                    }
                                                                                    else if (Cs >= 1530) {
                                                                                        return 1;
                                                                                    }
                                                                                }
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (SC >= 3357) {
                                                                        if (negBalance < 44) {
                                                                            return 4;
                                                                        }
                                                                        else if (negBalance >= 44) {
                                                                            if (SC < 3446) {
                                                                                return 4;
                                                                            }
                                                                            else if (SC >= 3446) {
                                                                                if (Cs < 1553) {
                                                                                    return 4;
                                                                                }
                                                                                else if (Cs >= 1553) {
                                                                                    return 1;
                                                                                }
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                            else if (diffAFD >= 2) {
                                                                return 2;
                                                            }
                                                        }
                                                        else if (MVDiff >= 22) {
                                                            if (SC < 3247) {
                                                                if (TSC < 2) {
                                                                    if (Rs < 840) {
                                                                        return 1;
                                                                    }
                                                                    else if (Rs >= 840) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (TSC >= 2) {
                                                                    return 2;
                                                                }
                                                            }
                                                            else if (SC >= 3247) {
                                                                if (Cs < 1540) {
                                                                    return 4;
                                                                }
                                                                else if (Cs >= 1540) {
                                                                    if (AFD < 3) {
                                                                        return 8;
                                                                    }
                                                                    else if (AFD >= 3) {
                                                                        if (diffMVdiffVal < 2) {
                                                                            return 1;
                                                                        }
                                                                        else if (diffMVdiffVal >= 2) {
                                                                            if (Cs < 1581) {
                                                                                if (Rs < 1838) {
                                                                                    return 1;
                                                                                }
                                                                                else if (Rs >= 1838) {
                                                                                    return 4;
                                                                                }
                                                                            }
                                                                            else if (Cs >= 1581) {
                                                                                return 1;
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (TSC >= 5) {
                                                        if (posBalance < 57) {
                                                            if (mu_mv_mag_sq < 34) {
                                                                return 1;
                                                            }
                                                            else if (mu_mv_mag_sq >= 34) {
                                                                if (diffTSC < -1) {
                                                                    if (CsDiff < 9) {
                                                                        if (AFD < 8) {
                                                                            if (Cs < 574) {
                                                                                return 1;
                                                                            }
                                                                            else if (Cs >= 574) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                        else if (AFD >= 8) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                    else if (CsDiff >= 9) {
                                                                        return 8;
                                                                    }
                                                                }
                                                                else if (diffTSC >= -1) {
                                                                    if (MVDiff < 21) {
                                                                        if (Cs < 1325) {
                                                                            return 8;
                                                                        }
                                                                        else if (Cs >= 1325) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                    else if (MVDiff >= 21) {
                                                                        return 4;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (posBalance >= 57) {
                                                            if (mu_mv_mag_sq < 45) {
                                                                if (tcor < 75) {
                                                                    if (diffMVdiffVal < 14) {
                                                                        if (posBalance < 59) {
                                                                            if (Rs < 812) {
                                                                                if (mu_mv_mag_sq < 35) {
                                                                                    return 4;
                                                                                }
                                                                                else if (mu_mv_mag_sq >= 35) {
                                                                                    if (Cs < 575) {
                                                                                        return 4;
                                                                                    }
                                                                                    else if (Cs >= 575) {
                                                                                        return 1;
                                                                                    }
                                                                                }
                                                                            }
                                                                            else if (Rs >= 812) {
                                                                                return 1;
                                                                            }
                                                                        }
                                                                        else if (posBalance >= 59) {
                                                                            if (diffRsCsdiff < 2) {
                                                                                if (posBalance < 61) {
                                                                                    if (mu_mv_mag_sq < 28) {
                                                                                        return 4;
                                                                                    }
                                                                                    else if (mu_mv_mag_sq >= 28) {
                                                                                        return 1;
                                                                                    }
                                                                                }
                                                                                else if (posBalance >= 61) {
                                                                                    return 4;
                                                                                }
                                                                            }
                                                                            else if (diffRsCsdiff >= 2) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (diffMVdiffVal >= 14) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (tcor >= 75) {
                                                                    if (mu_mv_mag_sq < 39) {
                                                                        if (SC < 1241) {
                                                                            return 4;
                                                                        }
                                                                        else if (SC >= 1241) {
                                                                            if (diffRsCsdiff < -2) {
                                                                                if (AFD < 7) {
                                                                                    return 1;
                                                                                }
                                                                                else if (AFD >= 7) {
                                                                                    return 4;
                                                                                }
                                                                            }
                                                                            else if (diffRsCsdiff >= -2) {
                                                                                return 1;
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (mu_mv_mag_sq >= 39) {
                                                                        if (diffMVdiffVal < 4) {
                                                                            if (Rs < 1350) {
                                                                                return 1;
                                                                            }
                                                                            else if (Rs >= 1350) {
                                                                                if (Cs < 1528) {
                                                                                    return 4;
                                                                                }
                                                                                else if (Cs >= 1528) {
                                                                                    return 1;
                                                                                }
                                                                            }
                                                                        }
                                                                        else if (diffMVdiffVal >= 4) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                            else if (mu_mv_mag_sq >= 45) {
                                                                if (Cs < 1560) {
                                                                    if (SC < 1362) {
                                                                        if (diffTSC < -1) {
                                                                            return 4;
                                                                        }
                                                                        else if (diffTSC >= -1) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                    else if (SC >= 1362) {
                                                                        if (MVDiff < 35) {
                                                                            return 4;
                                                                        }
                                                                        else if (MVDiff >= 35) {
                                                                            if (Rs < 1001) {
                                                                                return 4;
                                                                            }
                                                                            else if (Rs >= 1001) {
                                                                                return 2;
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                                else if (Cs >= 1560) {
                                                                    return 1;
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (refDCval >= 45) {
                                                    if (Rs < 1139) {
                                                        if (SC < 1624) {
                                                            if (posBalance < 51) {
                                                                if (refDCval < 55) {
                                                                    return 4;
                                                                }
                                                                else if (refDCval >= 55) {
                                                                    if (Rs < 725) {
                                                                        return 4;
                                                                    }
                                                                    else if (Rs >= 725) {
                                                                        return 8;
                                                                    }
                                                                }
                                                            }
                                                            else if (posBalance >= 51) {
                                                                if (CsDiff < 3) {
                                                                    if (Rs < 821) {
                                                                        return 8;
                                                                    }
                                                                    else if (Rs >= 821) {
                                                                        if (Rs < 880) {
                                                                            if (AFD < 8) {
                                                                                return 4;
                                                                            }
                                                                            else if (AFD >= 8) {
                                                                                return 8;
                                                                            }
                                                                        }
                                                                        else if (Rs >= 880) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                }
                                                                else if (CsDiff >= 3) {
                                                                    return 2;
                                                                }
                                                            }
                                                        }
                                                        else if (SC >= 1624) {
                                                            if (posBalance < 47) {
                                                                if (mcTcor < 93) {
                                                                    if (diffAFD < 1) {
                                                                        if (Rs < 1112) {
                                                                            return 8;
                                                                        }
                                                                        else if (Rs >= 1112) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                    else if (diffAFD >= 1) {
                                                                        if (Rs < 1026) {
                                                                            return 4;
                                                                        }
                                                                        else if (Rs >= 1026) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                }
                                                                else if (mcTcor >= 93) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (posBalance >= 47) {
                                                                if (SC < 1746) {
                                                                    if (negBalance < 41) {
                                                                        if (ssDCval < 51) {
                                                                            if (ssDCval < 47) {
                                                                                if (SC < 1689) {
                                                                                    return 1;
                                                                                }
                                                                                else if (SC >= 1689) {
                                                                                    return 8;
                                                                                }
                                                                            }
                                                                            else if (ssDCval >= 47) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                        else if (ssDCval >= 51) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                    else if (negBalance >= 41) {
                                                                        return 1;
                                                                    }
                                                                }
                                                                else if (SC >= 1746) {
                                                                    if (refDCval < 57) {
                                                                        if (SC < 2110) {
                                                                            if (refDCval < 49) {
                                                                                if (CsDiff < 3) {
                                                                                    return 1;
                                                                                }
                                                                                else if (CsDiff >= 3) {
                                                                                    return 4;
                                                                                }
                                                                            }
                                                                            else if (refDCval >= 49) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                        else if (SC >= 2110) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                    else if (refDCval >= 57) {
                                                                        return 1;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (Rs >= 1139) {
                                                        if (RsDiff < 4) {
                                                            if (ssDCval < 54) {
                                                                if (TSCindex < 6) {
                                                                    if (ssDCval < 52) {
                                                                        if (Rs < 1142) {
                                                                            return 4;
                                                                        }
                                                                        else if (Rs >= 1142) {
                                                                            if (diffAFD < -1) {
                                                                                if (Rs < 1333) {
                                                                                    return 1;
                                                                                }
                                                                                else if (Rs >= 1333) {
                                                                                    if (SC < 2378) {
                                                                                        return 4;
                                                                                    }
                                                                                    else if (SC >= 2378) {
                                                                                        return 1;
                                                                                    }
                                                                                }
                                                                            }
                                                                            else if (diffAFD >= -1) {
                                                                                return 1;
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (ssDCval >= 52) {
                                                                        if (refDCval < 52) {
                                                                            return 4;
                                                                        }
                                                                        else if (refDCval >= 52) {
                                                                            if (mu_mv_mag_sq < 64) {
                                                                                return 1;
                                                                            }
                                                                            else if (mu_mv_mag_sq >= 64) {
                                                                                if (tcor < 93) {
                                                                                    return 4;
                                                                                }
                                                                                else if (tcor >= 93) {
                                                                                    return 1;
                                                                                }
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                                else if (TSCindex >= 6) {
                                                                    if (tcor < 88) {
                                                                        if (diffMVdiffVal < -1) {
                                                                            if (AFD < 8) {
                                                                                return 1;
                                                                            }
                                                                            else if (AFD >= 8) {
                                                                                return 8;
                                                                            }
                                                                        }
                                                                        else if (diffMVdiffVal >= -1) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                    else if (tcor >= 88) {
                                                                        if (Cs < 989) {
                                                                            return 1;
                                                                        }
                                                                        else if (Cs >= 989) {
                                                                            if (Rs < 1313) {
                                                                                return 4;
                                                                            }
                                                                            else if (Rs >= 1313) {
                                                                                return 1;
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                            else if (ssDCval >= 54) {
                                                                if (mu_mv_mag_sq < 74) {
                                                                    if (tcor < 91) {
                                                                        if (Rs < 1359) {
                                                                            return 1;
                                                                        }
                                                                        else if (Rs >= 1359) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                    else if (tcor >= 91) {
                                                                        return 1;
                                                                    }
                                                                }
                                                                else if (mu_mv_mag_sq >= 74) {
                                                                    if (Cs < 768) {
                                                                        return 1;
                                                                    }
                                                                    else if (Cs >= 768) {
                                                                        return 8;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (RsDiff >= 4) {
                                                            if (MVDiff < 15) {
                                                                if (negBalance < 40) {
                                                                    if (Cs < 2277) {
                                                                        if (ssDCval < 52) {
                                                                            if (tcor < 79) {
                                                                                return 4;
                                                                            }
                                                                            else if (tcor >= 79) {
                                                                                if (diffMVdiffVal < 1) {
                                                                                    return 1;
                                                                                }
                                                                                else if (diffMVdiffVal >= 1) {
                                                                                    return 4;
                                                                                }
                                                                            }
                                                                        }
                                                                        else if (ssDCval >= 52) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                    else if (Cs >= 2277) {
                                                                        return 8;
                                                                    }
                                                                }
                                                                else if (negBalance >= 40) {
                                                                    if (SC < 2500) {
                                                                        if (diffRsCsdiff < 10) {
                                                                            if (MVDiff < 13) {
                                                                                return 8;
                                                                            }
                                                                            else if (MVDiff >= 13) {
                                                                                return 1;
                                                                            }
                                                                        }
                                                                        else if (diffRsCsdiff >= 10) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                    else if (SC >= 2500) {
                                                                        if (tcor < 84) {
                                                                            if (Cs < 1100) {
                                                                                return 1;
                                                                            }
                                                                            else if (Cs >= 1100) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                        else if (tcor >= 84) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                            else if (MVDiff >= 15) {
                                                                if (diffAFD < -5) {
                                                                    return 8;
                                                                }
                                                                else if (diffAFD >= -5) {
                                                                    if (AFD < 7) {
                                                                        return 4;
                                                                    }
                                                                    else if (AFD >= 7) {
                                                                        return 1;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                            else if (ssDCval >= 64) {
                                                if (Cs < 730) {
                                                    if (gchDC < 3) {
                                                        return 2;
                                                    }
                                                    else if (gchDC >= 3) {
                                                        return 4;
                                                    }
                                                }
                                                else if (Cs >= 730) {
                                                    if (MVDiff < 18) {
                                                        if (negBalance < 47) {
                                                            if (Cs < 871) {
                                                                if (AFD < 10) {
                                                                    return 4;
                                                                }
                                                                else if (AFD >= 10) {
                                                                    return 1;
                                                                }
                                                            }
                                                            else if (Cs >= 871) {
                                                                return 1;
                                                            }
                                                        }
                                                        else if (negBalance >= 47) {
                                                            return 2;
                                                        }
                                                    }
                                                    else if (MVDiff >= 18) {
                                                        if (Cs < 1563) {
                                                            return 4;
                                                        }
                                                        else if (Cs >= 1563) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (refDCval >= 72) {
                                            if (Rs < 2584) {
                                                if (posBalance < 63) {
                                                    if (negBalance < 58) {
                                                        if (ssDCval < 101) {
                                                            if (AFD < 8) {
                                                                if (RsDiff < 6) {
                                                                    return 4;
                                                                }
                                                                else if (RsDiff >= 6) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (AFD >= 8) {
                                                                if (Cs < 787) {
                                                                    if (tcor < 64) {
                                                                        return 2;
                                                                    }
                                                                    else if (tcor >= 64) {
                                                                        if (negBalance < 51) {
                                                                            return 2;
                                                                        }
                                                                        else if (negBalance >= 51) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                }
                                                                else if (Cs >= 787) {
                                                                    if (SC < 3393) {
                                                                        if (Rs < 1104) {
                                                                            if (Cs < 1527) {
                                                                                return 4;
                                                                            }
                                                                            else if (Cs >= 1527) {
                                                                                return 8;
                                                                            }
                                                                        }
                                                                        else if (Rs >= 1104) {
                                                                            if (negBalance < 53) {
                                                                                if (gchDC < 3) {
                                                                                    return 2;
                                                                                }
                                                                                else if (gchDC >= 3) {
                                                                                    return 4;
                                                                                }
                                                                            }
                                                                            else if (negBalance >= 53) {
                                                                                return 8;
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (SC >= 3393) {
                                                                        return 4;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (ssDCval >= 101) {
                                                            return 2;
                                                        }
                                                    }
                                                    else if (negBalance >= 58) {
                                                        if (Cs < 861) {
                                                            return 8;
                                                        }
                                                        else if (Cs >= 861) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                                else if (posBalance >= 63) {
                                                    if (negBalance < 27) {
                                                        if (Cs < 607) {
                                                            if (RsCsDiff < 6) {
                                                                return 4;
                                                            }
                                                            else if (RsCsDiff >= 6) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (Cs >= 607) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (negBalance >= 27) {
                                                        return 1;
                                                    }
                                                }
                                            }
                                            else if (Rs >= 2584) {
                                                if (TSC < 6) {
                                                    return 8;
                                                }
                                                else if (TSC >= 6) {
                                                    if (posBalance < 51) {
                                                        return 4;
                                                    }
                                                    else if (posBalance >= 51) {
                                                        return 1;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (TSCindex >= 8) {
                                        if (negBalance < 50) {
                                            if (MVDiff < 3) {
                                                if (Rs < 2407) {
                                                    if (Cs < 1909) {
                                                        return 2;
                                                    }
                                                    else if (Cs >= 1909) {
                                                        if (posBalance < 59) {
                                                            return 2;
                                                        }
                                                        else if (posBalance >= 59) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                                else if (Rs >= 2407) {
                                                    return 4;
                                                }
                                            }
                                            else if (MVDiff >= 3) {
                                                if (posBalance < 51) {
                                                    if (RsCsDiff < 129) {
                                                        if (refDCval < 95) {
                                                            if (ssDCval < 87) {
                                                                if (ssDCval < 83) {
                                                                    if (TSC < 8) {
                                                                        if (Cs < 1827) {
                                                                            return 4;
                                                                        }
                                                                        else if (Cs >= 1827) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                    else if (TSC >= 8) {
                                                                        return 8;
                                                                    }
                                                                }
                                                                else if (ssDCval >= 83) {
                                                                    return 1;
                                                                }
                                                            }
                                                            else if (ssDCval >= 87) {
                                                                if (MVDiff < 42) {
                                                                    return 8;
                                                                }
                                                                else if (MVDiff >= 42) {
                                                                    return 2;
                                                                }
                                                            }
                                                        }
                                                        else if (refDCval >= 95) {
                                                            return 2;
                                                        }
                                                    }
                                                    else if (RsCsDiff >= 129) {
                                                        if (CsDiff < 9) {
                                                            return 4;
                                                        }
                                                        else if (CsDiff >= 9) {
                                                            if (RsDiff < 8) {
                                                                return 1;
                                                            }
                                                            else if (RsDiff >= 8) {
                                                                if (Rs < 2387) {
                                                                    return 8;
                                                                }
                                                                else if (Rs >= 2387) {
                                                                    return 2;
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (posBalance >= 51) {
                                                    if (mcTcor < 94) {
                                                        if (refDCval < 91) {
                                                            if (mu_mv_mag_sq < 72) {
                                                                if (MVDiff < 13) {
                                                                    if (TSC < 8) {
                                                                        return 4;
                                                                    }
                                                                    else if (TSC >= 8) {
                                                                        if (Cs < 1561) {
                                                                            return 8;
                                                                        }
                                                                        else if (Cs >= 1561) {
                                                                            if (diffAFD < -1) {
                                                                                return 8;
                                                                            }
                                                                            else if (diffAFD >= -1) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                                else if (MVDiff >= 13) {
                                                                    if (tcor < 74) {
                                                                        if (MVDiff < 33) {
                                                                            return 4;
                                                                        }
                                                                        else if (MVDiff >= 33) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                    else if (tcor >= 74) {
                                                                        if (Rs < 923) {
                                                                            return 8;
                                                                        }
                                                                        else if (Rs >= 923) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                            else if (mu_mv_mag_sq >= 72) {
                                                                if (TSC < 8) {
                                                                    return 4;
                                                                }
                                                                else if (TSC >= 8) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                        else if (refDCval >= 91) {
                                                            if (Cs < 2137) {
                                                                return 1;
                                                            }
                                                            else if (Cs >= 2137) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                    else if (mcTcor >= 94) {
                                                        if (mu_mv_mag_sq < 84) {
                                                            if (negBalance < 38) {
                                                                return 4;
                                                            }
                                                            else if (negBalance >= 38) {
                                                                if (mcTcor < 95) {
                                                                    if (RsCsDiff < 30) {
                                                                        return 1;
                                                                    }
                                                                    else if (RsCsDiff >= 30) {
                                                                        if (AFD < 10) {
                                                                            return 4;
                                                                        }
                                                                        else if (AFD >= 10) {
                                                                            if (Rs < 2566) {
                                                                                return 2;
                                                                            }
                                                                            else if (Rs >= 2566) {
                                                                                return 1;
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                                else if (mcTcor >= 95) {
                                                                    if (Cs < 2212) {
                                                                        return 1;
                                                                    }
                                                                    else if (Cs >= 2212) {
                                                                        return 4;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (mu_mv_mag_sq >= 84) {
                                                            if (RsCsDiff < 70) {
                                                                return 2;
                                                            }
                                                            else if (RsCsDiff >= 70) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (negBalance >= 50) {
                                            if (ssDCval < 98) {
                                                if (tcor < 78) {
                                                    if (Cs < 1085) {
                                                        if (negBalance < 52) {
                                                            return 8;
                                                        }
                                                        else if (negBalance >= 52) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (Cs >= 1085) {
                                                        return 8;
                                                    }
                                                }
                                                else if (tcor >= 78) {
                                                    if (Rs < 2393) {
                                                        if (mcTcor < 91) {
                                                            if (negBalance < 52) {
                                                                if (mu_mv_mag_sq < 3) {
                                                                    if (Rs < 1219) {
                                                                        return 8;
                                                                    }
                                                                    else if (Rs >= 1219) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (mu_mv_mag_sq >= 3) {
                                                                    if (gchDC < 1) {
                                                                        return 8;
                                                                    }
                                                                    else if (gchDC >= 1) {
                                                                        if (RsCsDiff < 57) {
                                                                            return 8;
                                                                        }
                                                                        else if (RsCsDiff >= 57) {
                                                                            return 2;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                            else if (negBalance >= 52) {
                                                                if (SC < 3747) {
                                                                    if (RsDiff < 6) {
                                                                        if (mu_mv_mag_sq < 56) {
                                                                            return 8;
                                                                        }
                                                                        else if (mu_mv_mag_sq >= 56) {
                                                                            if (SC < 3580) {
                                                                                return 2;
                                                                            }
                                                                            else if (SC >= 3580) {
                                                                                return 8;
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (RsDiff >= 6) {
                                                                        if (Cs < 1707) {
                                                                            return 2;
                                                                        }
                                                                        else if (Cs >= 1707) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                }
                                                                else if (SC >= 3747) {
                                                                    if (TSC < 9) {
                                                                        if (mcTcor < 82) {
                                                                            if (SC < 3903) {
                                                                                return 2;
                                                                            }
                                                                            else if (SC >= 3903) {
                                                                                return 8;
                                                                            }
                                                                        }
                                                                        else if (mcTcor >= 82) {
                                                                            return 2;
                                                                        }
                                                                    }
                                                                    else if (TSC >= 9) {
                                                                        return 8;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (mcTcor >= 91) {
                                                            if (negBalance < 56) {
                                                                if (diffMVdiffVal < -3) {
                                                                    return 4;
                                                                }
                                                                else if (diffMVdiffVal >= -3) {
                                                                    if (diffRsCsdiff < 25) {
                                                                        if (diffMVdiffVal < 5) {
                                                                            return 2;
                                                                        }
                                                                        else if (diffMVdiffVal >= 5) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                    else if (diffRsCsdiff >= 25) {
                                                                        if (RsDiff < 6) {
                                                                            return 1;
                                                                        }
                                                                        else if (RsDiff >= 6) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                            else if (negBalance >= 56) {
                                                                if (SC < 2727) {
                                                                    return 4;
                                                                }
                                                                else if (SC >= 2727) {
                                                                    return 1;
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (Rs >= 2393) {
                                                        if (mu_mv_mag_sq < 66) {
                                                            return 1;
                                                        }
                                                        else if (mu_mv_mag_sq >= 66) {
                                                            if (tcor < 91) {
                                                                if (RsDiff < 10) {
                                                                    return 2;
                                                                }
                                                                else if (RsDiff >= 10) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (tcor >= 91) {
                                                                if (SC < 5664) {
                                                                    return 4;
                                                                }
                                                                else if (SC >= 5664) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                            else if (ssDCval >= 98) {
                                                if (posBalance < 58) {
                                                    if (Rs < 2015) {
                                                        return 2;
                                                    }
                                                    else if (Rs >= 2015) {
                                                        return 8;
                                                    }
                                                }
                                                else if (posBalance >= 58) {
                                                    return 4;
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                        }
                        else if (ssDCval >= 104) {
                            if (refDCval < 120) {
                                if (SC < 3631) {
                                    if (SC < 869) {
                                        if (Rs < 487) {
                                            if (CsDiff < 2) {
                                                if (Cs < 444) {
                                                    return 4;
                                                }
                                                else if (Cs >= 444) {
                                                    return 8;
                                                }
                                            }
                                            else if (CsDiff >= 2) {
                                                if (posBalance < 63) {
                                                    if (Rs < 349) {
                                                        return 4;
                                                    }
                                                    else if (Rs >= 349) {
                                                        return 8;
                                                    }
                                                }
                                                else if (posBalance >= 63) {
                                                    return 8;
                                                }
                                            }
                                        }
                                        else if (Rs >= 487) {
                                            if (negBalance < 34) {
                                                if (RsCsDiff < 1) {
                                                    return 1;
                                                }
                                                else if (RsCsDiff >= 1) {
                                                    return 8;
                                                }
                                            }
                                            else if (negBalance >= 34) {
                                                if (Rs < 687) {
                                                    return 2;
                                                }
                                                else if (Rs >= 687) {
                                                    return 8;
                                                }
                                            }
                                        }
                                    }
                                    else if (SC >= 869) {
                                        if (ssDCval < 106) {
                                            if (SC < 2159) {
                                                return 2;
                                            }
                                            else if (SC >= 2159) {
                                                if (Cs < 898) {
                                                    return 8;
                                                }
                                                else if (Cs >= 898) {
                                                    return 4;
                                                }
                                            }
                                        }
                                        else if (ssDCval >= 106) {
                                            if (AFD < 10) {
                                                if (Rs < 400) {
                                                    if (posBalance < 28) {
                                                        return 8;
                                                    }
                                                    else if (posBalance >= 28) {
                                                        if (Rs < 358) {
                                                            return 2;
                                                        }
                                                        else if (Rs >= 358) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                                else if (Rs >= 400) {
                                                    return 8;
                                                }
                                            }
                                            else if (AFD >= 10) {
                                                if (diffMVdiffVal < -1) {
                                                    if (negBalance < 37) {
                                                        return 2;
                                                    }
                                                    else if (negBalance >= 37) {
                                                        return 4;
                                                    }
                                                }
                                                else if (diffMVdiffVal >= -1) {
                                                    if (RsDiff < 6) {
                                                        if (mu_mv_mag_sq < 19) {
                                                            return 4;
                                                        }
                                                        else if (mu_mv_mag_sq >= 19) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (RsDiff >= 6) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (SC >= 3631) {
                                    if (MVDiff < 14) {
                                        if (ssDCval < 107) {
                                            if (gchDC < 1) {
                                                if (SC < 4700) {
                                                    if (diffTSC < 1) {
                                                        if (Rs < 3051) {
                                                            return 4;
                                                        }
                                                        else if (Rs >= 3051) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (diffTSC >= 1) {
                                                        return 8;
                                                    }
                                                }
                                                else if (SC >= 4700) {
                                                    return 8;
                                                }
                                            }
                                            else if (gchDC >= 1) {
                                                if (TSC < 6) {
                                                    if (SC < 5092) {
                                                        if (Rs < 3166) {
                                                            if (refDCval < 107) {
                                                                if (Rs < 2578) {
                                                                    return 4;
                                                                }
                                                                else if (Rs >= 2578) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (refDCval >= 107) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (Rs >= 3166) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (SC >= 5092) {
                                                        return 8;
                                                    }
                                                }
                                                else if (TSC >= 6) {
                                                    return 4;
                                                }
                                            }
                                        }
                                        else if (ssDCval >= 107) {
                                            if (mu_mv_mag_sq < 73) {
                                                if (Scindex < 5) {
                                                    return 4;
                                                }
                                                else if (Scindex >= 5) {
                                                    if (posBalance < 46) {
                                                        if (diffMVdiffVal < 2) {
                                                            return 4;
                                                        }
                                                        else if (diffMVdiffVal >= 2) {
                                                            if (negBalance < 36) {
                                                                return 8;
                                                            }
                                                            else if (negBalance >= 36) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                    else if (posBalance >= 46) {
                                                        if (posBalance < 55) {
                                                            return 8;
                                                        }
                                                        else if (posBalance >= 55) {
                                                            if (SC < 4538) {
                                                                if (MVDiff < 7) {
                                                                    if (mu_mv_mag_sq < 66) {
                                                                        if (SC < 4275) {
                                                                            return 4;
                                                                        }
                                                                        else if (SC >= 4275) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                    else if (mu_mv_mag_sq >= 66) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (MVDiff >= 7) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (SC >= 4538) {
                                                                if (Rs < 3667) {
                                                                    if (TSC < 9) {
                                                                        return 8;
                                                                    }
                                                                    else if (TSC >= 9) {
                                                                        if (TSC < 10) {
                                                                            return 4;
                                                                        }
                                                                        else if (TSC >= 10) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                }
                                                                else if (Rs >= 3667) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                            else if (mu_mv_mag_sq >= 73) {
                                                if (SC < 5882) {
                                                    if (diffMVdiffVal < -4) {
                                                        if (Cs < 1897) {
                                                            return 8;
                                                        }
                                                        else if (Cs >= 1897) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (diffMVdiffVal >= -4) {
                                                        if (MVDiff < 5) {
                                                            return 8;
                                                        }
                                                        else if (MVDiff >= 5) {
                                                            if (SC < 5001) {
                                                                if (refDCval < 117) {
                                                                    if (negBalance < 50) {
                                                                        return 8;
                                                                    }
                                                                    else if (negBalance >= 50) {
                                                                        if (ssDCval < 117) {
                                                                            return 4;
                                                                        }
                                                                        else if (ssDCval >= 117) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                }
                                                                else if (refDCval >= 117) {
                                                                    if (Cs < 2381) {
                                                                        return 8;
                                                                    }
                                                                    else if (Cs >= 2381) {
                                                                        return 4;
                                                                    }
                                                                }
                                                            }
                                                            else if (SC >= 5001) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (SC >= 5882) {
                                                    return 2;
                                                }
                                            }
                                        }
                                    }
                                    else if (MVDiff >= 14) {
                                        if (Cs < 2040) {
                                            if (mu_mv_mag_sq < 65) {
                                                return 4;
                                            }
                                            else if (mu_mv_mag_sq >= 65) {
                                                if (CsDiff < 6) {
                                                    if (Rs < 3881) {
                                                        return 8;
                                                    }
                                                    else if (Rs >= 3881) {
                                                        return 4;
                                                    }
                                                }
                                                else if (CsDiff >= 6) {
                                                    return 4;
                                                }
                                            }
                                        }
                                        else if (Cs >= 2040) {
                                            if (mu_mv_mag_sq < 72) {
                                                if (SC < 5930) {
                                                    if (MVDiff < 28) {
                                                        return 2;
                                                    }
                                                    else if (MVDiff >= 28) {
                                                        return 4;
                                                    }
                                                }
                                                else if (SC >= 5930) {
                                                    if (Rs < 3808) {
                                                        return 8;
                                                    }
                                                    else if (Rs >= 3808) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                            else if (mu_mv_mag_sq >= 72) {
                                                if (Cs < 2613) {
                                                    return 2;
                                                }
                                                else if (Cs >= 2613) {
                                                    return 8;
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                            else if (refDCval >= 120) {
                                if (SC < 3398) {
                                    if (mcTcor < 95) {
                                        if (refDCval < 142) {
                                            if (posBalance < 58) {
                                                if (Rs < 1205) {
                                                    return 1;
                                                }
                                                else if (Rs >= 1205) {
                                                    if (Cs < 1242) {
                                                        return 8;
                                                    }
                                                    else if (Cs >= 1242) {
                                                        return 1;
                                                    }
                                                }
                                            }
                                            else if (posBalance >= 58) {
                                                if (posBalance < 64) {
                                                    if (Cs < 878) {
                                                        return 1;
                                                    }
                                                    else if (Cs >= 878) {
                                                        return 4;
                                                    }
                                                }
                                                else if (posBalance >= 64) {
                                                    if (Cs < 790) {
                                                        return 1;
                                                    }
                                                    else if (Cs >= 790) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                        }
                                        else if (refDCval >= 142) {
                                            if (Cs < 367) {
                                                return 8;
                                            }
                                            else if (Cs >= 367) {
                                                if (diffTSC < -2) {
                                                    return 1;
                                                }
                                                else if (diffTSC >= -2) {
                                                    return 4;
                                                }
                                            }
                                        }
                                    }
                                    else if (mcTcor >= 95) {
                                        if (SC < 646) {
                                            return 1;
                                        }
                                        else if (SC >= 646) {
                                            if (negBalance < 21) {
                                                if (negBalance < 11) {
                                                    return 2;
                                                }
                                                else if (negBalance >= 11) {
                                                    if (AFD < 3) {
                                                        if (Cs < 511) {
                                                            return 2;
                                                        }
                                                        else if (Cs >= 511) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (AFD >= 3) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                            else if (negBalance >= 21) {
                                                if (ssDCval < 131) {
                                                    if (tcor < 95) {
                                                        return 4;
                                                    }
                                                    else if (tcor >= 95) {
                                                        if (ssDCval < 117) {
                                                            return 4;
                                                        }
                                                        else if (ssDCval >= 117) {
                                                            if (mu_mv_mag_sq < 84) {
                                                                return 8;
                                                            }
                                                            else if (mu_mv_mag_sq >= 84) {
                                                                return 2;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (ssDCval >= 131) {
                                                    if (RsDiff < 3) {
                                                        return 8;
                                                    }
                                                    else if (RsDiff >= 3) {
                                                        return 2;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (SC >= 3398) {
                                    if (ssDCval < 123) {
                                        return 4;
                                    }
                                    else if (ssDCval >= 123) {
                                        if (Cs < 1935) {
                                            if (diffAFD < -3) {
                                                return 1;
                                            }
                                            else if (diffAFD >= -3) {
                                                if (diffTSC < 1) {
                                                    return 4;
                                                }
                                                else if (diffTSC >= 1) {
                                                    if (diffRsCsdiff < 6) {
                                                        if (Rs < 3334) {
                                                            return 8;
                                                        }
                                                        else if (Rs >= 3334) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (diffRsCsdiff >= 6) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                        }
                                        else if (Cs >= 1935) {
                                            if (tcor < 81) {
                                                if (Cs < 2053) {
                                                    return 1;
                                                }
                                                else if (Cs >= 2053) {
                                                    return 4;
                                                }
                                            }
                                            else if (tcor >= 81) {
                                                if (SC < 4644) {
                                                    if (mcTcor < 86) {
                                                        return 4;
                                                    }
                                                    else if (mcTcor >= 86) {
                                                        return 8;
                                                    }
                                                }
                                                else if (SC >= 4644) {
                                                    if (posBalance < 58) {
                                                        return 4;
                                                    }
                                                    else if (posBalance >= 58) {
                                                        if (diffAFD < -1) {
                                                            if (tcor < 83) {
                                                                return 4;
                                                            }
                                                            else if (tcor >= 83) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (diffAFD >= -1) {
                                                            if (Rs < 2691) {
                                                                return 8;
                                                            }
                                                            else if (Rs >= 2691) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                        }
                    }
                    else if (AFD >= 12) {
                        if (Cs < 1451) {
                            if (RsCsDiff < 98) {
                                if (Cs < 805) {
                                    if (Cs < 285) {
                                        if (tcor < 75) {
                                            if (ssDCval < 62) {
                                                if (Cs < 1) {
                                                    return 2;
                                                }
                                                else if (Cs >= 1) {
                                                    return 1;
                                                }
                                            }
                                            else if (ssDCval >= 62) {
                                                if (CsDiff < 3) {
                                                    return 2;
                                                }
                                                else if (CsDiff >= 3) {
                                                    return 1;
                                                }
                                            }
                                        }
                                        else if (tcor >= 75) {
                                            if (negBalance < 99) {
                                                if (Cs < 43) {
                                                    return 2;
                                                }
                                                else if (Cs >= 43) {
                                                    return 1;
                                                }
                                            }
                                            else if (negBalance >= 99) {
                                                return 2;
                                            }
                                        }
                                    }
                                    else if (Cs >= 285) {
                                        if (ssDCval < 95) {
                                            if (TSC < 9) {
                                                if (ssDCval < 60) {
                                                    if (Rs < 1139) {
                                                        return 2;
                                                    }
                                                    else if (Rs >= 1139) {
                                                        return 1;
                                                    }
                                                }
                                                else if (ssDCval >= 60) {
                                                    if (Rs < 526) {
                                                        return 4;
                                                    }
                                                    else if (Rs >= 526) {
                                                        return 2;
                                                    }
                                                }
                                            }
                                            else if (TSC >= 9) {
                                                if (mcTcor < 77) {
                                                    if (diffMVdiffVal < 78) {
                                                        return 1;
                                                    }
                                                    else if (diffMVdiffVal >= 78) {
                                                        return 8;
                                                    }
                                                }
                                                else if (mcTcor >= 77) {
                                                    if (mu_mv_mag_sq < 14) {
                                                        return 2;
                                                    }
                                                    else if (mu_mv_mag_sq >= 14) {
                                                        if (Rs < 820) {
                                                            return 8;
                                                        }
                                                        else if (Rs >= 820) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (ssDCval >= 95) {
                                            if (ssDCval < 115) {
                                                if (posBalance < 66) {
                                                    if (mu_mv_mag_sq < 3) {
                                                        return 4;
                                                    }
                                                    else if (mu_mv_mag_sq >= 3) {
                                                        if (gchDC < 17) {
                                                            return 2;
                                                        }
                                                        else if (gchDC >= 17) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                                else if (posBalance >= 66) {
                                                    if (TSC < 14) {
                                                        return 8;
                                                    }
                                                    else if (TSC >= 14) {
                                                        if (Rs < 905) {
                                                            return 2;
                                                        }
                                                        else if (Rs >= 905) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (ssDCval >= 115) {
                                                if (mu_mv_mag_sq < 18) {
                                                    return 1;
                                                }
                                                else if (mu_mv_mag_sq >= 18) {
                                                    if (CsDiff < 2) {
                                                        return 4;
                                                    }
                                                    else if (CsDiff >= 2) {
                                                        if (Rs < 479) {
                                                            return 1;
                                                        }
                                                        else if (Rs >= 479) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (Cs >= 805) {
                                    if (mcTcor < 87) {
                                        if (refDCval < 79) {
                                            return 8;
                                        }
                                        else if (refDCval >= 79) {
                                            return 1;
                                        }
                                    }
                                    else if (mcTcor >= 87) {
                                        if (MVDiff < 22) {
                                            if (SC < 2345) {
                                                if (CsDiff < 7) {
                                                    return 4;
                                                }
                                                else if (CsDiff >= 7) {
                                                    return 2;
                                                }
                                            }
                                            else if (SC >= 2345) {
                                                if (TSC < 9) {
                                                    if (mu_mv_mag_sq < 18) {
                                                        if (SC < 2509) {
                                                            if (Cs < 1195) {
                                                                return 1;
                                                            }
                                                            else if (Cs >= 1195) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (SC >= 2509) {
                                                            return 1;
                                                        }
                                                    }
                                                    else if (mu_mv_mag_sq >= 18) {
                                                        return 4;
                                                    }
                                                }
                                                else if (TSC >= 9) {
                                                    if (Rs < 3010) {
                                                        if (Cs < 1083) {
                                                            return 2;
                                                        }
                                                        else if (Cs >= 1083) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (Rs >= 3010) {
                                                        return 2;
                                                    }
                                                }
                                            }
                                        }
                                        else if (MVDiff >= 22) {
                                            if (MVDiff < 118) {
                                                if (TSC < 7) {
                                                    return 1;
                                                }
                                                else if (TSC >= 7) {
                                                    if (Cs < 1274) {
                                                        if (MVDiff < 59) {
                                                            return 2;
                                                        }
                                                        else if (MVDiff >= 59) {
                                                            return 1;
                                                        }
                                                    }
                                                    else if (Cs >= 1274) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                            else if (MVDiff >= 118) {
                                                if (AFD < 13) {
                                                    return 4;
                                                }
                                                else if (AFD >= 13) {
                                                    return 8;
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                            else if (RsCsDiff >= 98) {
                                if (tcor < 79) {
                                    if (AFD < 15) {
                                        if (ssDCval < 134) {
                                            if (ssDCval < 50) {
                                                return 2;
                                            }
                                            else if (ssDCval >= 50) {
                                                return 1;
                                            }
                                        }
                                        else if (ssDCval >= 134) {
                                            return 8;
                                        }
                                    }
                                    else if (AFD >= 15) {
                                        if (CsDiff < 11) {
                                            if (mcTcor < 89) {
                                                if (RsCsDiff < 117) {
                                                    if (AFD < 24) {
                                                        if (RsDiff < 9) {
                                                            return 8;
                                                        }
                                                        else if (RsDiff >= 9) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (AFD >= 24) {
                                                        return 1;
                                                    }
                                                }
                                                else if (RsCsDiff >= 117) {
                                                    if (tcor < 69) {
                                                        if (mcTcor < 73) {
                                                            if (diffTSC < 21) {
                                                                return 2;
                                                            }
                                                            else if (diffTSC >= 21) {
                                                                if (Rs < 1427) {
                                                                    return 4;
                                                                }
                                                                else if (Rs >= 1427) {
                                                                    return 2;
                                                                }
                                                            }
                                                        }
                                                        else if (mcTcor >= 73) {
                                                            if (AFD < 28) {
                                                                if (MVDiff < 113) {
                                                                    if (CsDiff < 6) {
                                                                        if (SC < 3210) {
                                                                            return 2;
                                                                        }
                                                                        else if (SC >= 3210) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                    else if (CsDiff >= 6) {
                                                                        return 1;
                                                                    }
                                                                }
                                                                else if (MVDiff >= 113) {
                                                                    return 2;
                                                                }
                                                            }
                                                            else if (AFD >= 28) {
                                                                if (Cs < 971) {
                                                                    if (diffRsCsdiff < 45) {
                                                                        if (Rs < 1097) {
                                                                            return 1;
                                                                        }
                                                                        else if (Rs >= 1097) {
                                                                            return 2;
                                                                        }
                                                                    }
                                                                    else if (diffRsCsdiff >= 45) {
                                                                        return 1;
                                                                    }
                                                                }
                                                                else if (Cs >= 971) {
                                                                    return 2;
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (tcor >= 69) {
                                                        if (CsDiff < 3) {
                                                            return 4;
                                                        }
                                                        else if (CsDiff >= 3) {
                                                            return 2;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (mcTcor >= 89) {
                                                if (RsCsDiff < 751) {
                                                    if (refDCval < 192) {
                                                        if (Rs < 798) {
                                                            return 8;
                                                        }
                                                        else if (Rs >= 798) {
                                                            return 2;
                                                        }
                                                    }
                                                    else if (refDCval >= 192) {
                                                        return 4;
                                                    }
                                                }
                                                else if (RsCsDiff >= 751) {
                                                    if (Cs < 1405) {
                                                        return 4;
                                                    }
                                                    else if (Cs >= 1405) {
                                                        return 2;
                                                    }
                                                }
                                            }
                                        }
                                        else if (CsDiff >= 11) {
                                            if (RsCsDiff < 410) {
                                                if (gchDC < 5) {
                                                    if (RsDiff < 13) {
                                                        return 8;
                                                    }
                                                    else if (RsDiff >= 13) {
                                                        return 1;
                                                    }
                                                }
                                                else if (gchDC >= 5) {
                                                    if (Cs < 856) {
                                                        return 4;
                                                    }
                                                    else if (Cs >= 856) {
                                                        return 1;
                                                    }
                                                }
                                            }
                                            else if (RsCsDiff >= 410) {
                                                if (Cs < 744) {
                                                    if (AFD < 59) {
                                                        return 8;
                                                    }
                                                    else if (AFD >= 59) {
                                                        return 1;
                                                    }
                                                }
                                                else if (Cs >= 744) {
                                                    if (AFD < 64) {
                                                        return 4;
                                                    }
                                                    else if (AFD >= 64) {
                                                        return 2;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (tcor >= 79) {
                                    if (gchDC < 3) {
                                        if (Rs < 1693) {
                                            if (Rs < 1641) {
                                                if (diffAFD < 2) {
                                                    return 2;
                                                }
                                                else if (diffAFD >= 2) {
                                                    return 1;
                                                }
                                            }
                                            else if (Rs >= 1641) {
                                                if (SC < 3079) {
                                                    return 4;
                                                }
                                                else if (SC >= 3079) {
                                                    return 1;
                                                }
                                            }
                                        }
                                        else if (Rs >= 1693) {
                                            if (refDCval < 75) {
                                                return 4;
                                            }
                                            else if (refDCval >= 75) {
                                                return 2;
                                            }
                                        }
                                    }
                                    else if (gchDC >= 3) {
                                        if (TSC < 10) {
                                            if (refDCval < 154) {
                                                return 2;
                                            }
                                            else if (refDCval >= 154) {
                                                if (Rs < 1639) {
                                                    return 1;
                                                }
                                                else if (Rs >= 1639) {
                                                    return 2;
                                                }
                                            }
                                        }
                                        else if (TSC >= 10) {
                                            if (AFD < 18) {
                                                return 8;
                                            }
                                            else if (AFD >= 18) {
                                                if (Rs < 1302) {
                                                    return 1;
                                                }
                                                else if (Rs >= 1302) {
                                                    return 4;
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                        }
                        else if (Cs >= 1451) {
                            if (mcTcor < 94) {
                                if (mu_mv_mag_sq < 2) {
                                    if (negBalance < 53) {
                                        if (posBalance < 60) {
                                            if (RsDiff < 11) {
                                                if (refDCval < 128) {
                                                    if (Rs < 2522) {
                                                        if (mcTcor < 93) {
                                                            if (posBalance < 51) {
                                                                if (Rs < 1716) {
                                                                    return 4;
                                                                }
                                                                else if (Rs >= 1716) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (posBalance >= 51) {
                                                                if (refDCval < 80) {
                                                                    if (SC < 3645) {
                                                                        if (Rs < 1638) {
                                                                            return 1;
                                                                        }
                                                                        else if (Rs >= 1638) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                    else if (SC >= 3645) {
                                                                        if (SC < 4422) {
                                                                            return 8;
                                                                        }
                                                                        else if (SC >= 4422) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                }
                                                                else if (refDCval >= 80) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                        else if (mcTcor >= 93) {
                                                            if (SC < 4410) {
                                                                if (SC < 3671) {
                                                                    return 8;
                                                                }
                                                                else if (SC >= 3671) {
                                                                    return 1;
                                                                }
                                                            }
                                                            else if (SC >= 4410) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                    else if (Rs >= 2522) {
                                                        if (refDCval < 101) {
                                                            if (TSC < 12) {
                                                                if (negBalance < 49) {
                                                                    if (diffRsCsdiff < 48) {
                                                                        return 1;
                                                                    }
                                                                    else if (diffRsCsdiff >= 48) {
                                                                        if (Rs < 2689) {
                                                                            return 4;
                                                                        }
                                                                        else if (Rs >= 2689) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                }
                                                                else if (negBalance >= 49) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (TSC >= 12) {
                                                                if (Rs < 2533) {
                                                                    return 1;
                                                                }
                                                                else if (Rs >= 2533) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                        else if (refDCval >= 101) {
                                                            if (ssDCval < 113) {
                                                                return 8;
                                                            }
                                                            else if (ssDCval >= 113) {
                                                                return 2;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (refDCval >= 128) {
                                                    return 1;
                                                }
                                            }
                                            else if (RsDiff >= 11) {
                                                if (ssDCval < 81) {
                                                    return 8;
                                                }
                                                else if (ssDCval >= 81) {
                                                    if (Rs < 3451) {
                                                        if (diffAFD < -3) {
                                                            if (MVDiff < 7) {
                                                                return 4;
                                                            }
                                                            else if (MVDiff >= 7) {
                                                                return 1;
                                                            }
                                                        }
                                                        else if (diffAFD >= -3) {
                                                            if (diffTSC < 2) {
                                                                if (AFD < 14) {
                                                                    return 2;
                                                                }
                                                                else if (AFD >= 14) {
                                                                    return 1;
                                                                }
                                                            }
                                                            else if (diffTSC >= 2) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                    else if (Rs >= 3451) {
                                                        if (Cs < 2449) {
                                                            return 2;
                                                        }
                                                        else if (Cs >= 2449) {
                                                            return 1;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (posBalance >= 60) {
                                            if (posBalance < 65) {
                                                if (SC < 4535) {
                                                    if (ssDCval < 122) {
                                                        return 4;
                                                    }
                                                    else if (ssDCval >= 122) {
                                                        return 1;
                                                    }
                                                }
                                                else if (SC >= 4535) {
                                                    if (mcTcor < 86) {
                                                        return 1;
                                                    }
                                                    else if (mcTcor >= 86) {
                                                        if (tcor < 90) {
                                                            if (Cs < 2380) {
                                                                return 2;
                                                            }
                                                            else if (Cs >= 2380) {
                                                                return 1;
                                                            }
                                                        }
                                                        else if (tcor >= 90) {
                                                            return 1;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (posBalance >= 65) {
                                                if (Rs < 2643) {
                                                    if (TSC < 9) {
                                                        return 4;
                                                    }
                                                    else if (TSC >= 9) {
                                                        return 2;
                                                    }
                                                }
                                                else if (Rs >= 2643) {
                                                    return 4;
                                                }
                                            }
                                        }
                                    }
                                    else if (negBalance >= 53) {
                                        if (refDCval < 94) {
                                            if (tcor < 74) {
                                                return 4;
                                            }
                                            else if (tcor >= 74) {
                                                if (diffRsCsdiff < 7) {
                                                    if (AFD < 16) {
                                                        return 2;
                                                    }
                                                    else if (AFD >= 16) {
                                                        return 8;
                                                    }
                                                }
                                                else if (diffRsCsdiff >= 7) {
                                                    if (Cs < 2325) {
                                                        return 2;
                                                    }
                                                    else if (Cs >= 2325) {
                                                        return 1;
                                                    }
                                                }
                                            }
                                        }
                                        else if (refDCval >= 94) {
                                            if (refDCval < 115) {
                                                if (TSC < 13) {
                                                    if (SC < 5425) {
                                                        return 2;
                                                    }
                                                    else if (SC >= 5425) {
                                                        return 8;
                                                    }
                                                }
                                                else if (TSC >= 13) {
                                                    if (CsDiff < 9) {
                                                        return 1;
                                                    }
                                                    else if (CsDiff >= 9) {
                                                        return 2;
                                                    }
                                                }
                                            }
                                            else if (refDCval >= 115) {
                                                if (RsCsDiff < 293) {
                                                    if (SC < 4905) {
                                                        if (diffMVdiffVal < 10) {
                                                            if (mcTcor < 91) {
                                                                return 4;
                                                            }
                                                            else if (mcTcor >= 91) {
                                                                if (Cs < 1641) {
                                                                    return 2;
                                                                }
                                                                else if (Cs >= 1641) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                        else if (diffMVdiffVal >= 10) {
                                                            if (CsDiff < 11) {
                                                                return 1;
                                                            }
                                                            else if (CsDiff >= 11) {
                                                                if (Rs < 2374) {
                                                                    return 2;
                                                                }
                                                                else if (Rs >= 2374) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (SC >= 4905) {
                                                        return 1;
                                                    }
                                                }
                                                else if (RsCsDiff >= 293) {
                                                    if (Rs < 2619) {
                                                        return 2;
                                                    }
                                                    else if (Rs >= 2619) {
                                                        if (RsCsDiff < 305) {
                                                            return 2;
                                                        }
                                                        else if (RsCsDiff >= 305) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (mu_mv_mag_sq >= 2) {
                                    if (RsCsDiff < 63) {
                                        if (mcTcor < 90) {
                                            if (ssDCval < 78) {
                                                return 8;
                                            }
                                            else if (ssDCval >= 78) {
                                                if (Cs < 1923) {
                                                    return 1;
                                                }
                                                else if (Cs >= 1923) {
                                                    return 8;
                                                }
                                            }
                                        }
                                        else if (mcTcor >= 90) {
                                            if (Cs < 1581) {
                                                return 4;
                                            }
                                            else if (Cs >= 1581) {
                                                if (Rs < 2069) {
                                                    return 2;
                                                }
                                                else if (Rs >= 2069) {
                                                    return 1;
                                                }
                                            }
                                        }
                                    }
                                    else if (RsCsDiff >= 63) {
                                        if (negBalance < 50) {
                                            if (gchDC < 2) {
                                                if (ssDCval < 85) {
                                                    if (CsDiff < 9) {
                                                        if (refDCval < 79) {
                                                            return 8;
                                                        }
                                                        else if (refDCval >= 79) {
                                                            if (negBalance < 43) {
                                                                return 1;
                                                            }
                                                            else if (negBalance >= 43) {
                                                                if (posBalance < 58) {
                                                                    if (Rs < 1900) {
                                                                        return 1;
                                                                    }
                                                                    else if (Rs >= 1900) {
                                                                        return 8;
                                                                    }
                                                                }
                                                                else if (posBalance >= 58) {
                                                                    return 1;
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (CsDiff >= 9) {
                                                        if (refDCval < 79) {
                                                            return 1;
                                                        }
                                                        else if (refDCval >= 79) {
                                                            if (TSC < 9) {
                                                                return 1;
                                                            }
                                                            else if (TSC >= 9) {
                                                                if (Cs < 2823) {
                                                                    if (SC < 5380) {
                                                                        return 1;
                                                                    }
                                                                    else if (SC >= 5380) {
                                                                        return 8;
                                                                    }
                                                                }
                                                                else if (Cs >= 2823) {
                                                                    if (posBalance < 45) {
                                                                        return 1;
                                                                    }
                                                                    else if (posBalance >= 45) {
                                                                        return 8;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (ssDCval >= 85) {
                                                    if (Rs < 1799) {
                                                        if (SC < 3180) {
                                                            return 4;
                                                        }
                                                        else if (SC >= 3180) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (Rs >= 1799) {
                                                        if (RsDiff < 12) {
                                                            if (tcor < 90) {
                                                                if (refDCval < 87) {
                                                                    if (posBalance < 53) {
                                                                        return 1;
                                                                    }
                                                                    else if (posBalance >= 53) {
                                                                        if (posBalance < 56) {
                                                                            if (mcTcor < 92) {
                                                                                if (Cs < 2307) {
                                                                                    return 4;
                                                                                }
                                                                                else if (Cs >= 2307) {
                                                                                    return 1;
                                                                                }
                                                                            }
                                                                            else if (mcTcor >= 92) {
                                                                                return 8;
                                                                            }
                                                                        }
                                                                        else if (posBalance >= 56) {
                                                                            if (Cs < 2488) {
                                                                                return 4;
                                                                            }
                                                                            else if (Cs >= 2488) {
                                                                                return 1;
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                                else if (refDCval >= 87) {
                                                                    if (posBalance < 56) {
                                                                        if (MVDiff < 22) {
                                                                            if (Cs < 2797) {
                                                                                if (ssDCval < 142) {
                                                                                    return 1;
                                                                                }
                                                                                else if (ssDCval >= 142) {
                                                                                    return 8;
                                                                                }
                                                                            }
                                                                            else if (Cs >= 2797) {
                                                                                return 8;
                                                                            }
                                                                        }
                                                                        else if (MVDiff >= 22) {
                                                                            if (refDCval < 104) {
                                                                                if (TSC < 12) {
                                                                                    return 1;
                                                                                }
                                                                                else if (TSC >= 12) {
                                                                                    return 4;
                                                                                }
                                                                            }
                                                                            else if (refDCval >= 104) {
                                                                                if (SC < 4471) {
                                                                                    return 8;
                                                                                }
                                                                                else if (SC >= 4471) {
                                                                                    return 4;
                                                                                }
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (posBalance >= 56) {
                                                                        if (Cs < 2673) {
                                                                            if (refDCval < 93) {
                                                                                if (posBalance < 59) {
                                                                                    return 4;
                                                                                }
                                                                                else if (posBalance >= 59) {
                                                                                    return 1;
                                                                                }
                                                                            }
                                                                            else if (refDCval >= 93) {
                                                                                if (posBalance < 58) {
                                                                                    if (MVDiff < 9) {
                                                                                        return 4;
                                                                                    }
                                                                                    else if (MVDiff >= 9) {
                                                                                        return 1;
                                                                                    }
                                                                                }
                                                                                else if (posBalance >= 58) {
                                                                                    return 1;
                                                                                }
                                                                            }
                                                                        }
                                                                        else if (Cs >= 2673) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                            else if (tcor >= 90) {
                                                                if (RsCsDiff < 122) {
                                                                    if (RsCsDiff < 99) {
                                                                        return 1;
                                                                    }
                                                                    else if (RsCsDiff >= 99) {
                                                                        if (posBalance < 58) {
                                                                            if (Rs < 2639) {
                                                                                return 8;
                                                                            }
                                                                            else if (Rs >= 2639) {
                                                                                return 1;
                                                                            }
                                                                        }
                                                                        else if (posBalance >= 58) {
                                                                            if (Rs < 2694) {
                                                                                return 2;
                                                                            }
                                                                            else if (Rs >= 2694) {
                                                                                return 8;
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                                else if (RsCsDiff >= 122) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                        else if (RsDiff >= 12) {
                                                            if (Cs < 2024) {
                                                                if (CsDiff < 9) {
                                                                    return 2;
                                                                }
                                                                else if (CsDiff >= 9) {
                                                                    if (SC < 5928) {
                                                                        return 1;
                                                                    }
                                                                    else if (SC >= 5928) {
                                                                        return 4;
                                                                    }
                                                                }
                                                            }
                                                            else if (Cs >= 2024) {
                                                                if (Cs < 2842) {
                                                                    if (diffAFD < -4) {
                                                                        if (Cs < 2363) {
                                                                            return 1;
                                                                        }
                                                                        else if (Cs >= 2363) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                    else if (diffAFD >= -4) {
                                                                        if (Rs < 3416) {
                                                                            if (posBalance < 59) {
                                                                                return 1;
                                                                            }
                                                                            else if (posBalance >= 59) {
                                                                                if (ssDCval < 90) {
                                                                                    return 4;
                                                                                }
                                                                                else if (ssDCval >= 90) {
                                                                                    return 1;
                                                                                }
                                                                            }
                                                                        }
                                                                        else if (Rs >= 3416) {
                                                                            return 2;
                                                                        }
                                                                    }
                                                                }
                                                                else if (Cs >= 2842) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                            else if (gchDC >= 2) {
                                                if (diffRsCsdiff < 66) {
                                                    if (TSC < 10) {
                                                        if (Cs < 2032) {
                                                            if (RsCsDiff < 256) {
                                                                if (diffAFD < -6) {
                                                                    return 8;
                                                                }
                                                                else if (diffAFD >= -6) {
                                                                    if (negBalance < 46) {
                                                                        return 4;
                                                                    }
                                                                    else if (negBalance >= 46) {
                                                                        return 2;
                                                                    }
                                                                }
                                                            }
                                                            else if (RsCsDiff >= 256) {
                                                                return 2;
                                                            }
                                                        }
                                                        else if (Cs >= 2032) {
                                                            if (AFD < 13) {
                                                                return 2;
                                                            }
                                                            else if (AFD >= 13) {
                                                                return 1;
                                                            }
                                                        }
                                                    }
                                                    else if (TSC >= 10) {
                                                        if (mu_mv_mag_sq < 22) {
                                                            if (ssDCval < 91) {
                                                                if (Cs < 1982) {
                                                                    return 4;
                                                                }
                                                                else if (Cs >= 1982) {
                                                                    return 1;
                                                                }
                                                            }
                                                            else if (ssDCval >= 91) {
                                                                if (AFD < 19) {
                                                                    return 4;
                                                                }
                                                                else if (AFD >= 19) {
                                                                    if (negBalance < 47) {
                                                                        return 4;
                                                                    }
                                                                    else if (negBalance >= 47) {
                                                                        return 1;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (mu_mv_mag_sq >= 22) {
                                                            if (mu_mv_mag_sq < 37) {
                                                                if (CsDiff < 12) {
                                                                    if (RsDiff < 16) {
                                                                        return 8;
                                                                    }
                                                                    else if (RsDiff >= 16) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (CsDiff >= 12) {
                                                                    return 1;
                                                                }
                                                            }
                                                            else if (mu_mv_mag_sq >= 37) {
                                                                if (Rs < 3021) {
                                                                    return 1;
                                                                }
                                                                else if (Rs >= 3021) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (diffRsCsdiff >= 66) {
                                                    if (CsDiff < 13) {
                                                        if (AFD < 16) {
                                                            if (Rs < 3262) {
                                                                if (Cs < 2183) {
                                                                    return 8;
                                                                }
                                                                else if (Cs >= 2183) {
                                                                    if (SC < 5464) {
                                                                        return 1;
                                                                    }
                                                                    else if (SC >= 5464) {
                                                                        return 8;
                                                                    }
                                                                }
                                                            }
                                                            else if (Rs >= 3262) {
                                                                return 2;
                                                            }
                                                        }
                                                        else if (AFD >= 16) {
                                                            if (Rs < 2188) {
                                                                return 1;
                                                            }
                                                            else if (Rs >= 2188) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                    else if (CsDiff >= 13) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                        }
                                        else if (negBalance >= 50) {
                                            if (refDCval < 78) {
                                                if (MVDiff < 39) {
                                                    if (mcTcor < 89) {
                                                        if (Scindex < 5) {
                                                            if (Rs < 1750) {
                                                                if (SC < 3256) {
                                                                    return 8;
                                                                }
                                                                else if (SC >= 3256) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (Rs >= 1750) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (Scindex >= 5) {
                                                            if (RsDiff < 17) {
                                                                if (diffAFD < -4) {
                                                                    return 2;
                                                                }
                                                                else if (diffAFD >= -4) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (RsDiff >= 17) {
                                                                return 2;
                                                            }
                                                        }
                                                    }
                                                    else if (mcTcor >= 89) {
                                                        if (AFD < 15) {
                                                            return 1;
                                                        }
                                                        else if (AFD >= 15) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                                else if (MVDiff >= 39) {
                                                    if (Rs < 2430) {
                                                        if (diffRsCsdiff < -73) {
                                                            if (Cs < 2512) {
                                                                return 1;
                                                            }
                                                            else if (Cs >= 2512) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (diffRsCsdiff >= -73) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (Rs >= 2430) {
                                                        if (Cs < 2893) {
                                                            return 8;
                                                        }
                                                        else if (Cs >= 2893) {
                                                            return 1;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (refDCval >= 78) {
                                                if (tcor < 91) {
                                                    if (refDCval < 139) {
                                                        if (Cs < 2265) {
                                                            if (mcTcor < 84) {
                                                                if (diffMVdiffVal < 25) {
                                                                    if (refDCval < 137) {
                                                                        if (CsDiff < 9) {
                                                                            if (Rs < 2269) {
                                                                                return 4;
                                                                            }
                                                                            else if (Rs >= 2269) {
                                                                                return 1;
                                                                            }
                                                                        }
                                                                        else if (CsDiff >= 9) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                    else if (refDCval >= 137) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (diffMVdiffVal >= 25) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (mcTcor >= 84) {
                                                                if (SC < 3662) {
                                                                    if (refDCval < 114) {
                                                                        return 1;
                                                                    }
                                                                    else if (refDCval >= 114) {
                                                                        if (Rs < 2027) {
                                                                            return 4;
                                                                        }
                                                                        else if (Rs >= 2027) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                }
                                                                else if (SC >= 3662) {
                                                                    if (ssDCval < 84) {
                                                                        if (ssDCval < 78) {
                                                                            if (Rs < 2502) {
                                                                                return 2;
                                                                            }
                                                                            else if (Rs >= 2502) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                        else if (ssDCval >= 78) {
                                                                            if (MVDiff < 21) {
                                                                                return 1;
                                                                            }
                                                                            else if (MVDiff >= 21) {
                                                                                if (Cs < 2140) {
                                                                                    return 1;
                                                                                }
                                                                                else if (Cs >= 2140) {
                                                                                    return 2;
                                                                                }
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (ssDCval >= 84) {
                                                                        if (refDCval < 108) {
                                                                            if (AFD < 14) {
                                                                                return 1;
                                                                            }
                                                                            else if (AFD >= 14) {
                                                                                if (diffMVdiffVal < -11) {
                                                                                    return 1;
                                                                                }
                                                                                else if (diffMVdiffVal >= -11) {
                                                                                    if (tcor < 89) {
                                                                                        return 2;
                                                                                    }
                                                                                    else if (tcor >= 89) {
                                                                                        if (Rs < 2609) {
                                                                                            return 2;
                                                                                        }
                                                                                        else if (Rs >= 2609) {
                                                                                            return 4;
                                                                                        }
                                                                                    }
                                                                                }
                                                                            }
                                                                        }
                                                                        else if (refDCval >= 108) {
                                                                            if (RsDiff < 12) {
                                                                                if (AFD < 18) {
                                                                                    if (Cs < 1859) {
                                                                                        if (Rs < 2311) {
                                                                                            return 1;
                                                                                        }
                                                                                        else if (Rs >= 2311) {
                                                                                            return 8;
                                                                                        }
                                                                                    }
                                                                                    else if (Cs >= 1859) {
                                                                                        if (Rs < 2164) {
                                                                                            return 4;
                                                                                        }
                                                                                        else if (Rs >= 2164) {
                                                                                            return 1;
                                                                                        }
                                                                                    }
                                                                                }
                                                                                else if (AFD >= 18) {
                                                                                    return 1;
                                                                                }
                                                                            }
                                                                            else if (RsDiff >= 12) {
                                                                                if (tcor < 77) {
                                                                                    if (refDCval < 114) {
                                                                                        return 2;
                                                                                    }
                                                                                    else if (refDCval >= 114) {
                                                                                        if (refDCval < 116) {
                                                                                            if (SC < 5047) {
                                                                                                return 8;
                                                                                            }
                                                                                            else if (SC >= 5047) {
                                                                                                return 1;
                                                                                            }
                                                                                        }
                                                                                        else if (refDCval >= 116) {
                                                                                            return 1;
                                                                                        }
                                                                                    }
                                                                                }
                                                                                else if (tcor >= 77) {
                                                                                    if (mu_mv_mag_sq < 24) {
                                                                                        if (MVDiff < 55) {
                                                                                            if (SC < 4028) {
                                                                                                if (AFD < 18) {
                                                                                                    return 2;
                                                                                                }
                                                                                                else if (AFD >= 18) {
                                                                                                    return 4;
                                                                                                }
                                                                                            }
                                                                                            else if (SC >= 4028) {
                                                                                                return 2;
                                                                                            }
                                                                                        }
                                                                                        else if (MVDiff >= 55) {
                                                                                            return 8;
                                                                                        }
                                                                                    }
                                                                                    else if (mu_mv_mag_sq >= 24) {
                                                                                        return 8;
                                                                                    }
                                                                                }
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (Cs >= 2265) {
                                                            if (Rs < 3035) {
                                                                if (posBalance < 49) {
                                                                    if (Cs < 2434) {
                                                                        if (Cs < 2282) {
                                                                            return 1;
                                                                        }
                                                                        else if (Cs >= 2282) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                    else if (Cs >= 2434) {
                                                                        return 2;
                                                                    }
                                                                }
                                                                else if (posBalance >= 49) {
                                                                    if (gchDC < 2) {
                                                                        if (TSC < 12) {
                                                                            if (negBalance < 53) {
                                                                                if (SC < 5485) {
                                                                                    if (MVDiff < 5) {
                                                                                        return 8;
                                                                                    }
                                                                                    else if (MVDiff >= 5) {
                                                                                        if (diffAFD < 3) {
                                                                                            return 4;
                                                                                        }
                                                                                        else if (diffAFD >= 3) {
                                                                                            return 8;
                                                                                        }
                                                                                    }
                                                                                }
                                                                                else if (SC >= 5485) {
                                                                                    return 1;
                                                                                }
                                                                            }
                                                                            else if (negBalance >= 53) {
                                                                                if (Rs < 2964) {
                                                                                    if (refDCval < 115) {
                                                                                        return 4;
                                                                                    }
                                                                                    else if (refDCval >= 115) {
                                                                                        return 1;
                                                                                    }
                                                                                }
                                                                                else if (Rs >= 2964) {
                                                                                    return 2;
                                                                                }
                                                                            }
                                                                        }
                                                                        else if (TSC >= 12) {
                                                                            if (tcor < 44) {
                                                                                return 4;
                                                                            }
                                                                            else if (tcor >= 44) {
                                                                                if (Rs < 2907) {
                                                                                    if (diffAFD < 1) {
                                                                                        return 1;
                                                                                    }
                                                                                    else if (diffAFD >= 1) {
                                                                                        if (SC < 5042) {
                                                                                            return 8;
                                                                                        }
                                                                                        else if (SC >= 5042) {
                                                                                            return 1;
                                                                                        }
                                                                                    }
                                                                                }
                                                                                else if (Rs >= 2907) {
                                                                                    if (tcor < 77) {
                                                                                        if (Rs < 2968) {
                                                                                            return 1;
                                                                                        }
                                                                                        else if (Rs >= 2968) {
                                                                                            return 8;
                                                                                        }
                                                                                    }
                                                                                    else if (tcor >= 77) {
                                                                                        return 4;
                                                                                    }
                                                                                }
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (gchDC >= 2) {
                                                                        return 4;
                                                                    }
                                                                }
                                                            }
                                                            else if (Rs >= 3035) {
                                                                if (mu_mv_mag_sq < 59) {
                                                                    if (TSC < 12) {
                                                                        if (diffTSC < 1) {
                                                                            if (negBalance < 54) {
                                                                                return 1;
                                                                            }
                                                                            else if (negBalance >= 54) {
                                                                                if (Rs < 3233) {
                                                                                    return 2;
                                                                                }
                                                                                else if (Rs >= 3233) {
                                                                                    return 1;
                                                                                }
                                                                            }
                                                                        }
                                                                        else if (diffTSC >= 1) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                    else if (TSC >= 12) {
                                                                        return 1;
                                                                    }
                                                                }
                                                                else if (mu_mv_mag_sq >= 59) {
                                                                    if (refDCval < 85) {
                                                                        return 8;
                                                                    }
                                                                    else if (refDCval >= 85) {
                                                                        if (Cs < 2394) {
                                                                            return 8;
                                                                        }
                                                                        else if (Cs >= 2394) {
                                                                            if (posBalance < 56) {
                                                                                return 1;
                                                                            }
                                                                            else if (posBalance >= 56) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (refDCval >= 139) {
                                                        if (posBalance < 52) {
                                                            return 8;
                                                        }
                                                        else if (posBalance >= 52) {
                                                            if (mcTcor < 88) {
                                                                if (Cs < 2696) {
                                                                    return 4;
                                                                }
                                                                else if (Cs >= 2696) {
                                                                    return 1;
                                                                }
                                                            }
                                                            else if (mcTcor >= 88) {
                                                                return 2;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (tcor >= 91) {
                                                    if (AFD < 14) {
                                                        return 8;
                                                    }
                                                    else if (AFD >= 14) {
                                                        return 2;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                            else if (mcTcor >= 94) {
                                if (AFD < 16) {
                                    if (refDCval < 116) {
                                        if (negBalance < 48) {
                                            if (ssDCval < 101) {
                                                if (CsDiff < 8) {
                                                    if (SC < 3930) {
                                                        if (RsDiff < 7) {
                                                            return 1;
                                                        }
                                                        else if (RsDiff >= 7) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (SC >= 3930) {
                                                        if (RsDiff < 6) {
                                                            return 2;
                                                        }
                                                        else if (RsDiff >= 6) {
                                                            return 1;
                                                        }
                                                    }
                                                }
                                                else if (CsDiff >= 8) {
                                                    if (MVDiff < 15) {
                                                        if (RsDiff < 7) {
                                                            return 8;
                                                        }
                                                        else if (RsDiff >= 7) {
                                                            return 1;
                                                        }
                                                    }
                                                    else if (MVDiff >= 15) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                            else if (ssDCval >= 101) {
                                                if (AFD < 14) {
                                                    if (SC < 5361) {
                                                        return 2;
                                                    }
                                                    else if (SC >= 5361) {
                                                        if (negBalance < 32) {
                                                            return 2;
                                                        }
                                                        else if (negBalance >= 32) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                                else if (AFD >= 14) {
                                                    if (MVDiff < 10) {
                                                        return 8;
                                                    }
                                                    else if (MVDiff >= 10) {
                                                        return 2;
                                                    }
                                                }
                                            }
                                        }
                                        else if (negBalance >= 48) {
                                            if (CsDiff < 7) {
                                                if (ssDCval < 88) {
                                                    if (MVDiff < 12) {
                                                        if (tcor < 90) {
                                                            return 2;
                                                        }
                                                        else if (tcor >= 90) {
                                                            return 1;
                                                        }
                                                    }
                                                    else if (MVDiff >= 12) {
                                                        return 2;
                                                    }
                                                }
                                                else if (ssDCval >= 88) {
                                                    if (tcor < 91) {
                                                        return 2;
                                                    }
                                                    else if (tcor >= 91) {
                                                        return 1;
                                                    }
                                                }
                                            }
                                            else if (CsDiff >= 7) {
                                                if (MVDiff < 8) {
                                                    if (diffMVdiffVal < 1) {
                                                        if (tcor < 88) {
                                                            return 4;
                                                        }
                                                        else if (tcor >= 88) {
                                                            if (ssDCval < 95) {
                                                                if (Cs < 2271) {
                                                                    return 2;
                                                                }
                                                                else if (Cs >= 2271) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (ssDCval >= 95) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                    else if (diffMVdiffVal >= 1) {
                                                        if (SC < 4381) {
                                                            return 4;
                                                        }
                                                        else if (SC >= 4381) {
                                                            if (ssDCval < 96) {
                                                                return 2;
                                                            }
                                                            else if (ssDCval >= 96) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (MVDiff >= 8) {
                                                    if (ssDCval < 94) {
                                                        if (Cs < 1788) {
                                                            return 2;
                                                        }
                                                        else if (Cs >= 1788) {
                                                            return 1;
                                                        }
                                                    }
                                                    else if (ssDCval >= 94) {
                                                        if (Cs < 2254) {
                                                            return 2;
                                                        }
                                                        else if (Cs >= 2254) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (refDCval >= 116) {
                                        return 8;
                                    }
                                }
                                else if (AFD >= 16) {
                                    if (SC < 5104) {
                                        if (TSC < 5) {
                                            return 4;
                                        }
                                        else if (TSC >= 5) {
                                            if (AFD < 17) {
                                                if (Scindex < 5) {
                                                    return 2;
                                                }
                                                else if (Scindex >= 5) {
                                                    if (Cs < 2538) {
                                                        return 4;
                                                    }
                                                    else if (Cs >= 2538) {
                                                        return 2;
                                                    }
                                                }
                                            }
                                            else if (AFD >= 17) {
                                                return 2;
                                            }
                                        }
                                    }
                                    else if (SC >= 5104) {
                                        if (posBalance < 59) {
                                            if (Cs < 2584) {
                                                if (CsDiff < 9) {
                                                    return 2;
                                                }
                                                else if (CsDiff >= 9) {
                                                    return 4;
                                                }
                                            }
                                            else if (Cs >= 2584) {
                                                return 8;
                                            }
                                        }
                                        else if (posBalance >= 59) {
                                            return 2;
                                        }
                                    }
                                }
                            }
                        }
                    }
                }
                else if (Scindex >= 6) {
                    if (tcor < 74) {
                        if (ssDCval < 98) {
                            if (Rs < 7033) {
                                if (Cs < 3394) {
                                    if (AFD < 30) {
                                        if (CsDiff < 17) {
                                            if (posBalance < 59) {
                                                if (Cs < 3099) {
                                                    if (Rs < 3611) {
                                                        return 1;
                                                    }
                                                    else if (Rs >= 3611) {
                                                        if (SC < 6938) {
                                                            return 4;
                                                        }
                                                        else if (SC >= 6938) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                                else if (Cs >= 3099) {
                                                    return 8;
                                                }
                                            }
                                            else if (posBalance >= 59) {
                                                return 4;
                                            }
                                        }
                                        else if (CsDiff >= 17) {
                                            return 1;
                                        }
                                    }
                                    else if (AFD >= 30) {
                                        if (AFD < 35) {
                                            return 2;
                                        }
                                        else if (AFD >= 35) {
                                            return 8;
                                        }
                                    }
                                }
                                else if (Cs >= 3394) {
                                    if (RsDiff < 33) {
                                        return 8;
                                    }
                                    else if (RsDiff >= 33) {
                                        return 2;
                                    }
                                }
                            }
                            else if (Rs >= 7033) {
                                return 4;
                            }
                        }
                        else if (ssDCval >= 98) {
                            if (Cs < 3849) {
                                if (ssDCval < 143) {
                                    if (ssDCval < 125) {
                                        if (negBalance < 57) {
                                            if (posBalance < 48) {
                                                if (ssDCval < 108) {
                                                    return 4;
                                                }
                                                else if (ssDCval >= 108) {
                                                    return 1;
                                                }
                                            }
                                            else if (posBalance >= 48) {
                                                if (diffRsCsdiff < 196) {
                                                    if (posBalance < 51) {
                                                        if (Rs < 4851) {
                                                            return 4;
                                                        }
                                                        else if (Rs >= 4851) {
                                                            return 2;
                                                        }
                                                    }
                                                    else if (posBalance >= 51) {
                                                        return 4;
                                                    }
                                                }
                                                else if (diffRsCsdiff >= 196) {
                                                    return 8;
                                                }
                                            }
                                        }
                                        else if (negBalance >= 57) {
                                            if (gchDC < 2) {
                                                if (SC < 6288) {
                                                    return 1;
                                                }
                                                else if (SC >= 6288) {
                                                    return 8;
                                                }
                                            }
                                            else if (gchDC >= 2) {
                                                if (Cs < 3512) {
                                                    return 2;
                                                }
                                                else if (Cs >= 3512) {
                                                    return 8;
                                                }
                                            }
                                        }
                                    }
                                    else if (ssDCval >= 125) {
                                        if (gchDC < 7) {
                                            if (tcor < 68) {
                                                return 1;
                                            }
                                            else if (tcor >= 68) {
                                                if (SC < 8881) {
                                                    if (negBalance < 52) {
                                                        if (Rs < 3378) {
                                                            return 8;
                                                        }
                                                        else if (Rs >= 3378) {
                                                            return 1;
                                                        }
                                                    }
                                                    else if (negBalance >= 52) {
                                                        if (RsCsDiff < 289) {
                                                            if (AFD < 18) {
                                                                return 8;
                                                            }
                                                            else if (AFD >= 18) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (RsCsDiff >= 289) {
                                                            if (MVDiff < 12) {
                                                                return 1;
                                                            }
                                                            else if (MVDiff >= 12) {
                                                                if (SC < 7824) {
                                                                    return 4;
                                                                }
                                                                else if (SC >= 7824) {
                                                                    return 1;
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (SC >= 8881) {
                                                    if (Rs < 6074) {
                                                        return 4;
                                                    }
                                                    else if (Rs >= 6074) {
                                                        return 1;
                                                    }
                                                }
                                            }
                                        }
                                        else if (gchDC >= 7) {
                                            return 8;
                                        }
                                    }
                                }
                                else if (ssDCval >= 143) {
                                    if (refDCval < 153) {
                                        return 2;
                                    }
                                    else if (refDCval >= 153) {
                                        return 4;
                                    }
                                }
                            }
                            else if (Cs >= 3849) {
                                if (RsDiff < 16) {
                                    if (ssDCval < 128) {
                                        if (AFD < 19) {
                                            return 4;
                                        }
                                        else if (AFD >= 19) {
                                            if (Cs < 4213) {
                                                return 4;
                                            }
                                            else if (Cs >= 4213) {
                                                return 8;
                                            }
                                        }
                                    }
                                    else if (ssDCval >= 128) {
                                        return 1;
                                    }
                                }
                                else if (RsDiff >= 16) {
                                    if (RsCsDiff < 2258) {
                                        if (posBalance < 56) {
                                            if (mcTcor < 78) {
                                                if (TSC < 17) {
                                                    return 8;
                                                }
                                                else if (TSC >= 17) {
                                                    if (Cs < 4628) {
                                                        return 4;
                                                    }
                                                    else if (Cs >= 4628) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                            else if (mcTcor >= 78) {
                                                if (mcTcor < 87) {
                                                    if (CsDiff < 16) {
                                                        if (RsDiff < 20) {
                                                            return 8;
                                                        }
                                                        else if (RsDiff >= 20) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (CsDiff >= 16) {
                                                        return 8;
                                                    }
                                                }
                                                else if (mcTcor >= 87) {
                                                    return 4;
                                                }
                                            }
                                        }
                                        else if (posBalance >= 56) {
                                            if (refDCval < 138) {
                                                return 8;
                                            }
                                            else if (refDCval >= 138) {
                                                if (RsCsDiff < 1265) {
                                                    if (Rs < 6200) {
                                                        if (Rs < 4656) {
                                                            return 1;
                                                        }
                                                        else if (Rs >= 4656) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (Rs >= 6200) {
                                                        return 2;
                                                    }
                                                }
                                                else if (RsCsDiff >= 1265) {
                                                    if (gchDC < 2) {
                                                        return 8;
                                                    }
                                                    else if (gchDC >= 2) {
                                                        if (AFD < 36) {
                                                            return 4;
                                                        }
                                                        else if (AFD >= 36) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (RsCsDiff >= 2258) {
                                        if (Rs < 14828) {
                                            if (refDCval < 126) {
                                                return 4;
                                            }
                                            else if (refDCval >= 126) {
                                                if (Cs < 4601) {
                                                    return 4;
                                                }
                                                else if (Cs >= 4601) {
                                                    if (refDCval < 132) {
                                                        if (AFD < 43) {
                                                            return 8;
                                                        }
                                                        else if (AFD >= 43) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (refDCval >= 132) {
                                                        if (Cs < 5299) {
                                                            return 8;
                                                        }
                                                        else if (Cs >= 5299) {
                                                            return 2;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (Rs >= 14828) {
                                            if (MVDiff < 17) {
                                                return 8;
                                            }
                                            else if (MVDiff >= 17) {
                                                return 4;
                                            }
                                        }
                                    }
                                }
                            }
                        }
                    }
                    else if (tcor >= 74) {
                        if (ssDCval < 72) {
                            if (RsCsDiff < 46) {
                                if (negBalance < 51) {
                                    if (MVDiff < 18) {
                                        if (negBalance < 48) {
                                            if (Cs < 3721) {
                                                if (refDCval < 54) {
                                                    if (Rs < 3518) {
                                                        return 8;
                                                    }
                                                    else if (Rs >= 3518) {
                                                        if (diffAFD < 1) {
                                                            return 4;
                                                        }
                                                        else if (diffAFD >= 1) {
                                                            if (RsCsDiff < 11) {
                                                                return 8;
                                                            }
                                                            else if (RsCsDiff >= 11) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (refDCval >= 54) {
                                                    if (TSC < 7) {
                                                        if (mcTcor < 97) {
                                                            return 8;
                                                        }
                                                        else if (mcTcor >= 97) {
                                                            if (AFD < 6) {
                                                                return 8;
                                                            }
                                                            else if (AFD >= 6) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                    else if (TSC >= 7) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                            else if (Cs >= 3721) {
                                                if (TSC < 7) {
                                                    if (diffRsCsdiff < 3) {
                                                        if (tcor < 86) {
                                                            return 8;
                                                        }
                                                        else if (tcor >= 86) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (diffRsCsdiff >= 3) {
                                                        return 8;
                                                    }
                                                }
                                                else if (TSC >= 7) {
                                                    return 4;
                                                }
                                            }
                                        }
                                        else if (negBalance >= 48) {
                                            if (mu_mv_mag_sq < 46) {
                                                if (diffMVdiffVal < -5) {
                                                    if (Cs < 3324) {
                                                        return 4;
                                                    }
                                                    else if (Cs >= 3324) {
                                                        return 2;
                                                    }
                                                }
                                                else if (diffMVdiffVal >= -5) {
                                                    if (ssDCval < 56) {
                                                        return 4;
                                                    }
                                                    else if (ssDCval >= 56) {
                                                        return 2;
                                                    }
                                                }
                                            }
                                            else if (mu_mv_mag_sq >= 46) {
                                                if (CsDiff < 3) {
                                                    return 8;
                                                }
                                                else if (CsDiff >= 3) {
                                                    if (MVDiff < 7) {
                                                        if (Rs < 3582) {
                                                            return 4;
                                                        }
                                                        else if (Rs >= 3582) {
                                                            if (Rs < 3754) {
                                                                return 8;
                                                            }
                                                            else if (Rs >= 3754) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                    else if (MVDiff >= 7) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (MVDiff >= 18) {
                                        if (tcor < 94) {
                                            if (mu_mv_mag_sq < 58) {
                                                if (MVDiff < 22) {
                                                    if (Cs < 3139) {
                                                        return 4;
                                                    }
                                                    else if (Cs >= 3139) {
                                                        return 8;
                                                    }
                                                }
                                                else if (MVDiff >= 22) {
                                                    if (posBalance < 58) {
                                                        return 8;
                                                    }
                                                    else if (posBalance >= 58) {
                                                        if (mcTcor < 93) {
                                                            return 8;
                                                        }
                                                        else if (mcTcor >= 93) {
                                                            if (TSCindex < 8) {
                                                                return 8;
                                                            }
                                                            else if (TSCindex >= 8) {
                                                                if (diffRsCsdiff < 8) {
                                                                    if (Cs < 3367) {
                                                                        if (Rs < 3331) {
                                                                            if (negBalance < 50) {
                                                                                return 8;
                                                                            }
                                                                            else if (negBalance >= 50) {
                                                                                return 2;
                                                                            }
                                                                        }
                                                                        else if (Rs >= 3331) {
                                                                            return 2;
                                                                        }
                                                                    }
                                                                    else if (Cs >= 3367) {
                                                                        return 8;
                                                                    }
                                                                }
                                                                else if (diffRsCsdiff >= 8) {
                                                                    return 2;
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                            else if (mu_mv_mag_sq >= 58) {
                                                return 4;
                                            }
                                        }
                                        else if (tcor >= 94) {
                                            return 4;
                                        }
                                    }
                                }
                                else if (negBalance >= 51) {
                                    if (TSC < 7) {
                                        if (mu_mv_mag_sq < 36) {
                                            return 8;
                                        }
                                        else if (mu_mv_mag_sq >= 36) {
                                            if (Rs < 3561) {
                                                return 4;
                                            }
                                            else if (Rs >= 3561) {
                                                if (Rs < 3824) {
                                                    return 2;
                                                }
                                                else if (Rs >= 3824) {
                                                    return 8;
                                                }
                                            }
                                        }
                                    }
                                    else if (TSC >= 7) {
                                        if (gchDC < 2) {
                                            if (diffMVdiffVal < 10) {
                                                if (MVDiff < 15) {
                                                    if (mcTcor < 94) {
                                                        return 4;
                                                    }
                                                    else if (mcTcor >= 94) {
                                                        return 2;
                                                    }
                                                }
                                                else if (MVDiff >= 15) {
                                                    if (posBalance < 52) {
                                                        return 8;
                                                    }
                                                    else if (posBalance >= 52) {
                                                        return 2;
                                                    }
                                                }
                                            }
                                            else if (diffMVdiffVal >= 10) {
                                                return 8;
                                            }
                                        }
                                        else if (gchDC >= 2) {
                                            if (SC < 6402) {
                                                return 2;
                                            }
                                            else if (SC >= 6402) {
                                                return 8;
                                            }
                                        }
                                    }
                                }
                            }
                            else if (RsCsDiff >= 46) {
                                if (mu_mv_mag_sq < 38) {
                                    if (TSC < 7) {
                                        return 8;
                                    }
                                    else if (TSC >= 7) {
                                        if (tcor < 80) {
                                            if (Rs < 6157) {
                                                return 8;
                                            }
                                            else if (Rs >= 6157) {
                                                return 4;
                                            }
                                        }
                                        else if (tcor >= 80) {
                                            if (diffMVdiffVal < -4) {
                                                if (RsDiff < 8) {
                                                    return 4;
                                                }
                                                else if (RsDiff >= 8) {
                                                    return 8;
                                                }
                                            }
                                            else if (diffMVdiffVal >= -4) {
                                                if (Cs < 4465) {
                                                    return 4;
                                                }
                                                else if (Cs >= 4465) {
                                                    return 8;
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (mu_mv_mag_sq >= 38) {
                                    if (ssDCval < 68) {
                                        if (Rs < 3564) {
                                            return 8;
                                        }
                                        else if (Rs >= 3564) {
                                            if (mcTcor < 80) {
                                                return 8;
                                            }
                                            else if (mcTcor >= 80) {
                                                return 4;
                                            }
                                        }
                                    }
                                    else if (ssDCval >= 68) {
                                        if (RsDiff < 15) {
                                            if (AFD < 14) {
                                                return 4;
                                            }
                                            else if (AFD >= 14) {
                                                return 8;
                                            }
                                        }
                                        else if (RsDiff >= 15) {
                                            return 8;
                                        }
                                    }
                                }
                            }
                        }
                        else if (ssDCval >= 72) {
                            if (ssDCval < 153) {
                                if (negBalance < 50) {
                                    if (negBalance < 37) {
                                        if (ssDCval < 128) {
                                            if (ssDCval < 109) {
                                                if (TSCindex < 8) {
                                                    return 8;
                                                }
                                                else if (TSCindex >= 8) {
                                                    if (gchDC < 1) {
                                                        return 1;
                                                    }
                                                    else if (gchDC >= 1) {
                                                        if (RsDiff < 7) {
                                                            return 8;
                                                        }
                                                        else if (RsDiff >= 7) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (ssDCval >= 109) {
                                                if (Rs < 4135) {
                                                    if (ssDCval < 120) {
                                                        if (MVDiff < 21) {
                                                            return 2;
                                                        }
                                                        else if (MVDiff >= 21) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (ssDCval >= 120) {
                                                        return 4;
                                                    }
                                                }
                                                else if (Rs >= 4135) {
                                                    if (SC < 7219) {
                                                        if (Cs < 2155) {
                                                            return 4;
                                                        }
                                                        else if (Cs >= 2155) {
                                                            if (SC < 6772) {
                                                                if (RsDiff < 10) {
                                                                    if (MVDiff < 14) {
                                                                        if (Cs < 2320) {
                                                                            return 4;
                                                                        }
                                                                        else if (Cs >= 2320) {
                                                                            return 2;
                                                                        }
                                                                    }
                                                                    else if (MVDiff >= 14) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (RsDiff >= 10) {
                                                                    return 2;
                                                                }
                                                            }
                                                            else if (SC >= 6772) {
                                                                if (diffAFD < -1) {
                                                                    return 8;
                                                                }
                                                                else if (diffAFD >= -1) {
                                                                    if (posBalance < 41) {
                                                                        return 4;
                                                                    }
                                                                    else if (posBalance >= 41) {
                                                                        if (tcor < 93) {
                                                                            if (negBalance < 34) {
                                                                                return 4;
                                                                            }
                                                                            else if (negBalance >= 34) {
                                                                                if (Rs < 4313) {
                                                                                    return 4;
                                                                                }
                                                                                else if (Rs >= 4313) {
                                                                                    return 8;
                                                                                }
                                                                            }
                                                                        }
                                                                        else if (tcor >= 93) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (SC >= 7219) {
                                                        if (RsDiff < 12) {
                                                            if (Cs < 2600) {
                                                                return 8;
                                                            }
                                                            else if (Cs >= 2600) {
                                                                if (gchDC < 1) {
                                                                    return 2;
                                                                }
                                                                else if (gchDC >= 1) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                        else if (RsDiff >= 12) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (ssDCval >= 128) {
                                            if (mu_mv_mag_sq < 80) {
                                                if (refDCval < 129) {
                                                    if (Rs < 4676) {
                                                        return 4;
                                                    }
                                                    else if (Rs >= 4676) {
                                                        if (Cs < 4239) {
                                                            return 2;
                                                        }
                                                        else if (Cs >= 4239) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                                else if (refDCval >= 129) {
                                                    return 2;
                                                }
                                            }
                                            else if (mu_mv_mag_sq >= 80) {
                                                return 2;
                                            }
                                        }
                                    }
                                    else if (negBalance >= 37) {
                                        if (Rs < 4091) {
                                            if (tcor < 79) {
                                                if (MVDiff < 14) {
                                                    return 1;
                                                }
                                                else if (MVDiff >= 14) {
                                                    if (Rs < 3376) {
                                                        return 1;
                                                    }
                                                    else if (Rs >= 3376) {
                                                        if (tcor < 78) {
                                                            if (ssDCval < 102) {
                                                                return 4;
                                                            }
                                                            else if (ssDCval >= 102) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (tcor >= 78) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (tcor >= 79) {
                                                if (RsDiff < 11) {
                                                    if (AFD < 13) {
                                                        return 8;
                                                    }
                                                    else if (AFD >= 13) {
                                                        if (mu_mv_mag_sq < 67) {
                                                            if (Cs < 2938) {
                                                                return 1;
                                                            }
                                                            else if (Cs >= 2938) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (mu_mv_mag_sq >= 67) {
                                                            if (Cs < 2908) {
                                                                return 1;
                                                            }
                                                            else if (Cs >= 2908) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (RsDiff >= 11) {
                                                    if (AFD < 15) {
                                                        if (refDCval < 106) {
                                                            return 4;
                                                        }
                                                        else if (refDCval >= 106) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (AFD >= 15) {
                                                        if (mu_mv_mag_sq < 49) {
                                                            if (ssDCval < 99) {
                                                                return 1;
                                                            }
                                                            else if (ssDCval >= 99) {
                                                                if (SC < 6795) {
                                                                    return 8;
                                                                }
                                                                else if (SC >= 6795) {
                                                                    return 2;
                                                                }
                                                            }
                                                        }
                                                        else if (mu_mv_mag_sq >= 49) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (Rs >= 4091) {
                                            if (SC < 8918) {
                                                if (SC < 7432) {
                                                    if (CsDiff < 8) {
                                                        if (refDCval < 113) {
                                                            if (posBalance < 46) {
                                                                if (RsCsDiff < 99) {
                                                                    if (Rs < 4198) {
                                                                        return 8;
                                                                    }
                                                                    else if (Rs >= 4198) {
                                                                        return 1;
                                                                    }
                                                                }
                                                                else if (RsCsDiff >= 99) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (posBalance >= 46) {
                                                                if (Rs < 4477) {
                                                                    if (Cs < 2314) {
                                                                        return 8;
                                                                    }
                                                                    else if (Cs >= 2314) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (Rs >= 4477) {
                                                                    return 2;
                                                                }
                                                            }
                                                        }
                                                        else if (refDCval >= 113) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (CsDiff >= 8) {
                                                        if (ssDCval < 107) {
                                                            return 4;
                                                        }
                                                        else if (ssDCval >= 107) {
                                                            if (mcTcor < 87) {
                                                                return 1;
                                                            }
                                                            else if (mcTcor >= 87) {
                                                                if (Rs < 4180) {
                                                                    return 2;
                                                                }
                                                                else if (Rs >= 4180) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (SC >= 7432) {
                                                    if (mu_mv_mag_sq < 75) {
                                                        if (refDCval < 129) {
                                                            if (refDCval < 105) {
                                                                return 4;
                                                            }
                                                            else if (refDCval >= 105) {
                                                                if (ssDCval < 112) {
                                                                    if (MVDiff < 10) {
                                                                        if (refDCval < 110) {
                                                                            return 4;
                                                                        }
                                                                        else if (refDCval >= 110) {
                                                                            return 2;
                                                                        }
                                                                    }
                                                                    else if (MVDiff >= 10) {
                                                                        if (MVDiff < 21) {
                                                                            return 2;
                                                                        }
                                                                        else if (MVDiff >= 21) {
                                                                            if (Rs < 5030) {
                                                                                return 1;
                                                                            }
                                                                            else if (Rs >= 5030) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                                else if (ssDCval >= 112) {
                                                                    if (refDCval < 114) {
                                                                        if (Rs < 4736) {
                                                                            return 2;
                                                                        }
                                                                        else if (Rs >= 4736) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                    else if (refDCval >= 114) {
                                                                        return 4;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (refDCval >= 129) {
                                                            if (SC < 7484) {
                                                                return 1;
                                                            }
                                                            else if (SC >= 7484) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                    else if (mu_mv_mag_sq >= 75) {
                                                        if (refDCval < 112) {
                                                            if (ssDCval < 111) {
                                                                if (CsDiff < 7) {
                                                                    return 8;
                                                                }
                                                                else if (CsDiff >= 7) {
                                                                    return 2;
                                                                }
                                                            }
                                                            else if (ssDCval >= 111) {
                                                                return 2;
                                                            }
                                                        }
                                                        else if (refDCval >= 112) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (SC >= 8918) {
                                                if (Cs < 3251) {
                                                    if (Rs < 6346) {
                                                        return 4;
                                                    }
                                                    else if (Rs >= 6346) {
                                                        return 1;
                                                    }
                                                }
                                                else if (Cs >= 3251) {
                                                    if (ssDCval < 143) {
                                                        if (ssDCval < 75) {
                                                            if (Rs < 7162) {
                                                                return 8;
                                                            }
                                                            else if (Rs >= 7162) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (ssDCval >= 75) {
                                                            if (negBalance < 49) {
                                                                return 8;
                                                            }
                                                            else if (negBalance >= 49) {
                                                                if (refDCval < 93) {
                                                                    return 1;
                                                                }
                                                                else if (refDCval >= 93) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (ssDCval >= 143) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (negBalance >= 50) {
                                    if (MVDiff < 22) {
                                        if (tcor < 88) {
                                            if (mcTcor < 93) {
                                                if (Cs < 3266) {
                                                    if (mcTcor < 88) {
                                                        if (MVDiff < 8) {
                                                            return 1;
                                                        }
                                                        else if (MVDiff >= 8) {
                                                            if (gchDC < 2) {
                                                                return 4;
                                                            }
                                                            else if (gchDC >= 2) {
                                                                return 1;
                                                            }
                                                        }
                                                    }
                                                    else if (mcTcor >= 88) {
                                                        if (posBalance < 58) {
                                                            if (refDCval < 118) {
                                                                return 4;
                                                            }
                                                            else if (refDCval >= 118) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (posBalance >= 58) {
                                                            if (Cs < 2682) {
                                                                return 2;
                                                            }
                                                            else if (Cs >= 2682) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (Cs >= 3266) {
                                                    if (refDCval < 136) {
                                                        if (ssDCval < 78) {
                                                            if (MVDiff < 6) {
                                                                return 8;
                                                            }
                                                            else if (MVDiff >= 6) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (ssDCval >= 78) {
                                                            if (gchDC < 2) {
                                                                if (mcTcor < 92) {
                                                                    return 8;
                                                                }
                                                                else if (mcTcor >= 92) {
                                                                    if (posBalance < 58) {
                                                                        return 8;
                                                                    }
                                                                    else if (posBalance >= 58) {
                                                                        return 1;
                                                                    }
                                                                }
                                                            }
                                                            else if (gchDC >= 2) {
                                                                if (MVDiff < 15) {
                                                                    return 8;
                                                                }
                                                                else if (MVDiff >= 15) {
                                                                    if (Rs < 7003) {
                                                                        if (gchDC < 5) {
                                                                            return 8;
                                                                        }
                                                                        else if (gchDC >= 5) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                    else if (Rs >= 7003) {
                                                                        if (Cs < 6608) {
                                                                            return 2;
                                                                        }
                                                                        else if (Cs >= 6608) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (refDCval >= 136) {
                                                        if (Rs < 6188) {
                                                            return 8;
                                                        }
                                                        else if (Rs >= 6188) {
                                                            if (negBalance < 56) {
                                                                return 4;
                                                            }
                                                            else if (negBalance >= 56) {
                                                                if (negBalance < 60) {
                                                                    if (RsDiff < 33) {
                                                                        return 8;
                                                                    }
                                                                    else if (RsDiff >= 33) {
                                                                        if (tcor < 82) {
                                                                            return 4;
                                                                        }
                                                                        else if (tcor >= 82) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                }
                                                                else if (negBalance >= 60) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                            else if (mcTcor >= 93) {
                                                if (ssDCval < 110) {
                                                    return 1;
                                                }
                                                else if (ssDCval >= 110) {
                                                    if (SC < 12317) {
                                                        return 8;
                                                    }
                                                    else if (SC >= 12317) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                        }
                                        else if (tcor >= 88) {
                                            if (ssDCval < 109) {
                                                return 4;
                                            }
                                            else if (ssDCval >= 109) {
                                                if (diffTSC < -3) {
                                                    if (Rs < 5190) {
                                                        return 8;
                                                    }
                                                    else if (Rs >= 5190) {
                                                        return 4;
                                                    }
                                                }
                                                else if (diffTSC >= -3) {
                                                    return 8;
                                                }
                                            }
                                        }
                                    }
                                    else if (MVDiff >= 22) {
                                        if (RsDiff < 32) {
                                            if (ssDCval < 122) {
                                                if (mu_mv_mag_sq < 36) {
                                                    if (Rs < 6913) {
                                                        return 8;
                                                    }
                                                    else if (Rs >= 6913) {
                                                        return 4;
                                                    }
                                                }
                                                else if (mu_mv_mag_sq >= 36) {
                                                    if (negBalance < 53) {
                                                        if (diffAFD < -21) {
                                                            return 4;
                                                        }
                                                        else if (diffAFD >= -21) {
                                                            return 2;
                                                        }
                                                    }
                                                    else if (negBalance >= 53) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                            else if (ssDCval >= 122) {
                                                if (AFD < 30) {
                                                    return 8;
                                                }
                                                else if (AFD >= 30) {
                                                    return 2;
                                                }
                                            }
                                        }
                                        else if (RsDiff >= 32) {
                                            return 2;
                                        }
                                    }
                                }
                            }
                            else if (ssDCval >= 153) {
                                if (RsCsDiff < 1092) {
                                    if (MVDiff < 27) {
                                        if (SC < 7950) {
                                            if (RsCsDiff < 367) {
                                                return 2;
                                            }
                                            else if (RsCsDiff >= 367) {
                                                if (TSC < 10) {
                                                    return 2;
                                                }
                                                else if (TSC >= 10) {
                                                    return 4;
                                                }
                                            }
                                        }
                                        else if (SC >= 7950) {
                                            return 2;
                                        }
                                    }
                                    else if (MVDiff >= 27) {
                                        if (ssDCval < 159) {
                                            if (Rs < 4654) {
                                                return 8;
                                            }
                                            else if (Rs >= 4654) {
                                                if (SC < 11171) {
                                                    return 4;
                                                }
                                                else if (SC >= 11171) {
                                                    return 2;
                                                }
                                            }
                                        }
                                        else if (ssDCval >= 159) {
                                            return 2;
                                        }
                                    }
                                }
                                else if (RsCsDiff >= 1092) {
                                    return 8;
                                }
                            }
                        }
                    }
                }
            }
        }
        else if (mu_mv_mag_sq >= 88) {
            if (AFD < 7) {
                if (Rs < 573) {
                    if (refDCval < 161) {
                        if (RsCsDiff < 1) {
                            if (negBalance < 19) {
                                if (Rs < 149) {
                                    if (Rs < 55) {
                                        if (SC < 1) {
                                            return 8;
                                        }
                                        else if (SC >= 1) {
                                            return 1;
                                        }
                                    }
                                    else if (Rs >= 55) {
                                        return 1;
                                    }
                                }
                                else if (Rs >= 149) {
                                    if (Cs < 423) {
                                        if (AFD < 2) {
                                            if (mu_mv_mag_sq < 128) {
                                                if (ssDCval < 71) {
                                                    return 8;
                                                }
                                                else if (ssDCval >= 71) {
                                                    if (Cs < 300) {
                                                        return 8;
                                                    }
                                                    else if (Cs >= 300) {
                                                        if (negBalance < 18) {
                                                            return 4;
                                                        }
                                                        else if (negBalance >= 18) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (mu_mv_mag_sq >= 128) {
                                                return 8;
                                            }
                                        }
                                        else if (AFD >= 2) {
                                            return 4;
                                        }
                                    }
                                    else if (Cs >= 423) {
                                        if (Cs < 717) {
                                            return 4;
                                        }
                                        else if (Cs >= 717) {
                                            return 8;
                                        }
                                    }
                                }
                            }
                            else if (negBalance >= 19) {
                                if (Cs < 1108) {
                                    if (diffMVdiffVal < -3) {
                                        return 8;
                                    }
                                    else if (diffMVdiffVal >= -3) {
                                        if (negBalance < 20) {
                                            if (Cs < 466) {
                                                return 8;
                                            }
                                            else if (Cs >= 466) {
                                                return 4;
                                            }
                                        }
                                        else if (negBalance >= 20) {
                                            return 4;
                                        }
                                    }
                                }
                                else if (Cs >= 1108) {
                                    return 8;
                                }
                            }
                        }
                        else if (RsCsDiff >= 1) {
                            if (negBalance < 25) {
                                if (posBalance < 68) {
                                    return 8;
                                }
                                else if (posBalance >= 68) {
                                    if (AFD < 3) {
                                        return 4;
                                    }
                                    else if (AFD >= 3) {
                                        return 8;
                                    }
                                }
                            }
                            else if (negBalance >= 25) {
                                if (ssDCval < 114) {
                                    if (ssDCval < 94) {
                                        if (mu_mv_mag_sq < 125) {
                                            if (Cs < 308) {
                                                if (negBalance < 56) {
                                                    if (Cs < 134) {
                                                        return 1;
                                                    }
                                                    else if (Cs >= 134) {
                                                        return 8;
                                                    }
                                                }
                                                else if (negBalance >= 56) {
                                                    return 4;
                                                }
                                            }
                                            else if (Cs >= 308) {
                                                if (TSC < 2) {
                                                    return 8;
                                                }
                                                else if (TSC >= 2) {
                                                    if (Rs < 263) {
                                                        return 2;
                                                    }
                                                    else if (Rs >= 263) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                        }
                                        else if (mu_mv_mag_sq >= 125) {
                                            return 8;
                                        }
                                    }
                                    else if (ssDCval >= 94) {
                                        if (MVDiff < 3) {
                                            return 8;
                                        }
                                        else if (MVDiff >= 3) {
                                            if (SC < 2115) {
                                                return 4;
                                            }
                                            else if (SC >= 2115) {
                                                return 8;
                                            }
                                        }
                                    }
                                }
                                else if (ssDCval >= 114) {
                                    if (Rs < 410) {
                                        return 2;
                                    }
                                    else if (Rs >= 410) {
                                        return 4;
                                    }
                                }
                            }
                        }
                    }
                    else if (refDCval >= 161) {
                        if (Rs < 437) {
                            return 4;
                        }
                        else if (Rs >= 437) {
                            if (refDCval < 165) {
                                if (MVDiff < 7) {
                                    return 4;
                                }
                                else if (MVDiff >= 7) {
                                    return 2;
                                }
                            }
                            else if (refDCval >= 165) {
                                if (SC < 995) {
                                    return 2;
                                }
                                else if (SC >= 995) {
                                    if (Cs < 494) {
                                        return 4;
                                    }
                                    else if (Cs >= 494) {
                                        return 2;
                                    }
                                }
                            }
                        }
                    }
                }
                else if (Rs >= 573) {
                    if (posBalance < 51) {
                        if (Rs < 3779) {
                            if (Cs < 3716) {
                                if (ssDCval < 32) {
                                    if (Cs < 1149) {
                                        if (Cs < 707) {
                                            if (refDCval < 27) {
                                                return 8;
                                            }
                                            else if (refDCval >= 27) {
                                                return 4;
                                            }
                                        }
                                        else if (Cs >= 707) {
                                            if (MVDiff < 22) {
                                                if (ssDCval < 31) {
                                                    return 8;
                                                }
                                                else if (ssDCval >= 31) {
                                                    if (Rs < 902) {
                                                        return 1;
                                                    }
                                                    else if (Rs >= 902) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                            else if (MVDiff >= 22) {
                                                return 4;
                                            }
                                        }
                                    }
                                    else if (Cs >= 1149) {
                                        if (Rs < 1045) {
                                            if (SC < 2174) {
                                                return 8;
                                            }
                                            else if (SC >= 2174) {
                                                return 4;
                                            }
                                        }
                                        else if (Rs >= 1045) {
                                            if (Rs < 1087) {
                                                return 2;
                                            }
                                            else if (Rs >= 1087) {
                                                return 8;
                                            }
                                        }
                                    }
                                }
                                else if (ssDCval >= 32) {
                                    if (mu_mv_mag_sq < 126) {
                                        if (refDCval < 65) {
                                            if (diffAFD < -1) {
                                                if (Rs < 1589) {
                                                    if (ssDCval < 56) {
                                                        if (Cs < 1096) {
                                                            return 1;
                                                        }
                                                        else if (Cs >= 1096) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (ssDCval >= 56) {
                                                        return 8;
                                                    }
                                                }
                                                else if (Rs >= 1589) {
                                                    if (ssDCval < 44) {
                                                        return 4;
                                                    }
                                                    else if (ssDCval >= 44) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                            else if (diffAFD >= -1) {
                                                if (refDCval < 32) {
                                                    return 4;
                                                }
                                                else if (refDCval >= 32) {
                                                    if (negBalance < 42) {
                                                        if (refDCval < 51) {
                                                            if (Cs < 1929) {
                                                                return 8;
                                                            }
                                                            else if (Cs >= 1929) {
                                                                if (RsCsDiff < 33) {
                                                                    return 8;
                                                                }
                                                                else if (RsCsDiff >= 33) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                        else if (refDCval >= 51) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (negBalance >= 42) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                        }
                                        else if (refDCval >= 65) {
                                            if (posBalance < 33) {
                                                if (Cs < 1527) {
                                                    if (posBalance < 2) {
                                                        if (Rs < 949) {
                                                            return 8;
                                                        }
                                                        else if (Rs >= 949) {
                                                            return 1;
                                                        }
                                                    }
                                                    else if (posBalance >= 2) {
                                                        if (MVDiff < 5) {
                                                            if (diffMVdiffVal < -3) {
                                                                return 1;
                                                            }
                                                            else if (diffMVdiffVal >= -3) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (MVDiff >= 5) {
                                                            if (negBalance < 20) {
                                                                if (tcor < 88) {
                                                                    if (SC < 2574) {
                                                                        if (MVDiff < 9) {
                                                                            return 8;
                                                                        }
                                                                        else if (MVDiff >= 9) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                    else if (SC >= 2574) {
                                                                        return 1;
                                                                    }
                                                                }
                                                                else if (tcor >= 88) {
                                                                    if (CsDiff < 2) {
                                                                        return 8;
                                                                    }
                                                                    else if (CsDiff >= 2) {
                                                                        return 1;
                                                                    }
                                                                }
                                                            }
                                                            else if (negBalance >= 20) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (Cs >= 1527) {
                                                    if (mu_mv_mag_sq < 121) {
                                                        if (SC < 4542) {
                                                            if (SC < 4198) {
                                                                if (CsDiff < 3) {
                                                                    return 8;
                                                                }
                                                                else if (CsDiff >= 3) {
                                                                    if (Cs < 1638) {
                                                                        return 4;
                                                                    }
                                                                    else if (Cs >= 1638) {
                                                                        return 8;
                                                                    }
                                                                }
                                                            }
                                                            else if (SC >= 4198) {
                                                                if (Rs < 2167) {
                                                                    return 1;
                                                                }
                                                                else if (Rs >= 2167) {
                                                                    if (diffRsCsdiff < 2) {
                                                                        return 8;
                                                                    }
                                                                    else if (diffRsCsdiff >= 2) {
                                                                        if (Cs < 2136) {
                                                                            return 1;
                                                                        }
                                                                        else if (Cs >= 2136) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (SC >= 4542) {
                                                            if (refDCval < 117) {
                                                                if (posBalance < 8) {
                                                                    return 4;
                                                                }
                                                                else if (posBalance >= 8) {
                                                                    if (mu_mv_mag_sq < 106) {
                                                                        return 1;
                                                                    }
                                                                    else if (mu_mv_mag_sq >= 106) {
                                                                        return 8;
                                                                    }
                                                                }
                                                            }
                                                            else if (refDCval >= 117) {
                                                                if (CsDiff < 6) {
                                                                    return 4;
                                                                }
                                                                else if (CsDiff >= 6) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (mu_mv_mag_sq >= 121) {
                                                        if (RsCsDiff < 2) {
                                                            if (Cs < 2234) {
                                                                return 8;
                                                            }
                                                            else if (Cs >= 2234) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (RsCsDiff >= 2) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (posBalance >= 33) {
                                                if (refDCval < 110) {
                                                    if (Rs < 3394) {
                                                        if (SC < 4252) {
                                                            if (negBalance < 42) {
                                                                if (TSC < 5) {
                                                                    return 8;
                                                                }
                                                                else if (TSC >= 5) {
                                                                    if (Rs < 2083) {
                                                                        return 4;
                                                                    }
                                                                    else if (Rs >= 2083) {
                                                                        return 8;
                                                                    }
                                                                }
                                                            }
                                                            else if (negBalance >= 42) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (SC >= 4252) {
                                                            if (ssDCval < 95) {
                                                                if (negBalance < 26) {
                                                                    return 8;
                                                                }
                                                                else if (negBalance >= 26) {
                                                                    if (Cs < 2260) {
                                                                        if (Rs < 2340) {
                                                                            if (MVDiff < 2) {
                                                                                return 1;
                                                                            }
                                                                            else if (MVDiff >= 2) {
                                                                                return 8;
                                                                            }
                                                                        }
                                                                        else if (Rs >= 2340) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                    else if (Cs >= 2260) {
                                                                        return 4;
                                                                    }
                                                                }
                                                            }
                                                            else if (ssDCval >= 95) {
                                                                if (posBalance < 40) {
                                                                    return 8;
                                                                }
                                                                else if (posBalance >= 40) {
                                                                    if (SC < 6064) {
                                                                        if (diffRsCsdiff < -52) {
                                                                            return 1;
                                                                        }
                                                                        else if (diffRsCsdiff >= -52) {
                                                                            if (Rs < 2752) {
                                                                                if (AFD < 5) {
                                                                                    return 1;
                                                                                }
                                                                                else if (AFD >= 5) {
                                                                                    return 8;
                                                                                }
                                                                            }
                                                                            else if (Rs >= 2752) {
                                                                                return 8;
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (SC >= 6064) {
                                                                        if (CsDiff < 4) {
                                                                            if (negBalance < 25) {
                                                                                return 8;
                                                                            }
                                                                            else if (negBalance >= 25) {
                                                                                return 1;
                                                                            }
                                                                        }
                                                                        else if (CsDiff >= 4) {
                                                                            if (negBalance < 27) {
                                                                                return 8;
                                                                            }
                                                                            else if (negBalance >= 27) {
                                                                                return 1;
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (Rs >= 3394) {
                                                        return 8;
                                                    }
                                                }
                                                else if (refDCval >= 110) {
                                                    if (negBalance < 34) {
                                                        return 8;
                                                    }
                                                    else if (negBalance >= 34) {
                                                        if (RsDiff < 3) {
                                                            return 4;
                                                        }
                                                        else if (RsDiff >= 3) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (mu_mv_mag_sq >= 126) {
                                        if (diffMVdiffVal < 11) {
                                            if (Rs < 583) {
                                                return 4;
                                            }
                                            else if (Rs >= 583) {
                                                if (diffAFD < -1) {
                                                    if (negBalance < 40) {
                                                        if (negBalance < 2) {
                                                            if (Rs < 2606) {
                                                                return 1;
                                                            }
                                                            else if (Rs >= 2606) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (negBalance >= 2) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (negBalance >= 40) {
                                                        if (Cs < 2293) {
                                                            return 8;
                                                        }
                                                        else if (Cs >= 2293) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                                else if (diffAFD >= -1) {
                                                    if (negBalance < 38) {
                                                        if (negBalance < 12) {
                                                            return 8;
                                                        }
                                                        else if (negBalance >= 12) {
                                                            if (TSC < 1) {
                                                                if (Cs < 1283) {
                                                                    return 8;
                                                                }
                                                                else if (Cs >= 1283) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (TSC >= 1) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                    else if (negBalance >= 38) {
                                                        if (Cs < 2356) {
                                                            return 8;
                                                        }
                                                        else if (Cs >= 2356) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (diffMVdiffVal >= 11) {
                                            return 4;
                                        }
                                    }
                                }
                            }
                            else if (Cs >= 3716) {
                                if (posBalance < 22) {
                                    if (SC < 7764) {
                                        if (Cs < 3849) {
                                            return 1;
                                        }
                                        else if (Cs >= 3849) {
                                            return 8;
                                        }
                                    }
                                    else if (SC >= 7764) {
                                        return 1;
                                    }
                                }
                                else if (posBalance >= 22) {
                                    if (AFD < 1) {
                                        if (diffRsCsdiff < 2) {
                                            if (Rs < 3665) {
                                                return 1;
                                            }
                                            else if (Rs >= 3665) {
                                                return 8;
                                            }
                                        }
                                        else if (diffRsCsdiff >= 2) {
                                            return 1;
                                        }
                                    }
                                    else if (AFD >= 1) {
                                        return 8;
                                    }
                                }
                            }
                        }
                        else if (Rs >= 3779) {
                            if (MVDiff < 9) {
                                return 8;
                            }
                            else if (MVDiff >= 9) {
                                if (refDCval < 114) {
                                    return 8;
                                }
                                else if (refDCval >= 114) {
                                    if (Rs < 4005) {
                                        return 8;
                                    }
                                    else if (Rs >= 4005) {
                                        return 4;
                                    }
                                }
                            }
                        }
                    }
                    else if (posBalance >= 51) {
                        if (ssDCval < 84) {
                            if (SC < 3545) {
                                if (MVDiff < 9) {
                                    if (AFD < 6) {
                                        if (Cs < 344) {
                                            return 4;
                                        }
                                        else if (Cs >= 344) {
                                            if (refDCval < 73) {
                                                return 8;
                                            }
                                            else if (refDCval >= 73) {
                                                if (refDCval < 74) {
                                                    if (Rs < 1880) {
                                                        return 4;
                                                    }
                                                    else if (Rs >= 1880) {
                                                        return 2;
                                                    }
                                                }
                                                else if (refDCval >= 74) {
                                                    if (SC < 3382) {
                                                        return 8;
                                                    }
                                                    else if (SC >= 3382) {
                                                        if (Cs < 1595) {
                                                            if (posBalance < 56) {
                                                                return 8;
                                                            }
                                                            else if (posBalance >= 56) {
                                                                if (Cs < 1584) {
                                                                    if (Rs < 1822) {
                                                                        return 4;
                                                                    }
                                                                    else if (Rs >= 1822) {
                                                                        return 8;
                                                                    }
                                                                }
                                                                else if (Cs >= 1584) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                        else if (Cs >= 1595) {
                                                            if (SC < 3536) {
                                                                return 8;
                                                            }
                                                            else if (SC >= 3536) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (AFD >= 6) {
                                        if (refDCval < 75) {
                                            if (ssDCval < 72) {
                                                if (negBalance < 46) {
                                                    return 8;
                                                }
                                                else if (negBalance >= 46) {
                                                    return 2;
                                                }
                                            }
                                            else if (ssDCval >= 72) {
                                                if (tcor < 93) {
                                                    return 8;
                                                }
                                                else if (tcor >= 93) {
                                                    if (SC < 3260) {
                                                        return 8;
                                                    }
                                                    else if (SC >= 3260) {
                                                        if (refDCval < 74) {
                                                            return 2;
                                                        }
                                                        else if (refDCval >= 74) {
                                                            if (Cs < 1544) {
                                                                return 2;
                                                            }
                                                            else if (Cs >= 1544) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (refDCval >= 75) {
                                            if (mu_mv_mag_sq < 121) {
                                                if (Rs < 1736) {
                                                    return 2;
                                                }
                                                else if (Rs >= 1736) {
                                                    return 4;
                                                }
                                            }
                                            else if (mu_mv_mag_sq >= 121) {
                                                return 8;
                                            }
                                        }
                                    }
                                }
                                else if (MVDiff >= 9) {
                                    if (Rs < 1368) {
                                        if (tcor < 87) {
                                            if (Cs < 746) {
                                                return 4;
                                            }
                                            else if (Cs >= 746) {
                                                return 8;
                                            }
                                        }
                                        else if (tcor >= 87) {
                                            if (TSC < 5) {
                                                return 8;
                                            }
                                            else if (TSC >= 5) {
                                                if (Cs < 674) {
                                                    return 8;
                                                }
                                                else if (Cs >= 674) {
                                                    return 1;
                                                }
                                            }
                                        }
                                    }
                                    else if (Rs >= 1368) {
                                        return 4;
                                    }
                                }
                            }
                            else if (SC >= 3545) {
                                if (SC < 4862) {
                                    if (Rs < 1821) {
                                        if (ssDCval < 79) {
                                            return 4;
                                        }
                                        else if (ssDCval >= 79) {
                                            return 8;
                                        }
                                    }
                                    else if (Rs >= 1821) {
                                        if (tcor < 94) {
                                            if (Rs < 1961) {
                                                return 8;
                                            }
                                            else if (Rs >= 1961) {
                                                if (Rs < 2161) {
                                                    return 2;
                                                }
                                                else if (Rs >= 2161) {
                                                    return 4;
                                                }
                                            }
                                        }
                                        else if (tcor >= 94) {
                                            if (Cs < 1597) {
                                                if (Cs < 1593) {
                                                    return 4;
                                                }
                                                else if (Cs >= 1593) {
                                                    return 2;
                                                }
                                            }
                                            else if (Cs >= 1597) {
                                                if (mcTcor < 95) {
                                                    if (Cs < 2007) {
                                                        return 8;
                                                    }
                                                    else if (Cs >= 2007) {
                                                        return 4;
                                                    }
                                                }
                                                else if (mcTcor >= 95) {
                                                    return 4;
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (SC >= 4862) {
                                    if (Cs < 4881) {
                                        if (diffRsCsdiff < -29) {
                                            return 4;
                                        }
                                        else if (diffRsCsdiff >= -29) {
                                            if (ssDCval < 54) {
                                                if (Rs < 3514) {
                                                    return 8;
                                                }
                                                else if (Rs >= 3514) {
                                                    return 4;
                                                }
                                            }
                                            else if (ssDCval >= 54) {
                                                if (CsDiff < 8) {
                                                    return 8;
                                                }
                                                else if (CsDiff >= 8) {
                                                    if (diffAFD < -1) {
                                                        return 4;
                                                    }
                                                    else if (diffAFD >= -1) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (Cs >= 4881) {
                                        if (ssDCval < 80) {
                                            if (mcTcor < 93) {
                                                return 8;
                                            }
                                            else if (mcTcor >= 93) {
                                                return 4;
                                            }
                                        }
                                        else if (ssDCval >= 80) {
                                            if (Cs < 5141) {
                                                if (Cs < 5119) {
                                                    if (RsDiff < 9) {
                                                        if (diffTSC < -2) {
                                                            return 8;
                                                        }
                                                        else if (diffTSC >= -2) {
                                                            if (refDCval < 80) {
                                                                if (Rs < 4935) {
                                                                    return 4;
                                                                }
                                                                else if (Rs >= 4935) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (refDCval >= 80) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                    else if (RsDiff >= 9) {
                                                        if (Rs < 4841) {
                                                            return 8;
                                                        }
                                                        else if (Rs >= 4841) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                                else if (Cs >= 5119) {
                                                    return 8;
                                                }
                                            }
                                            else if (Cs >= 5141) {
                                                return 4;
                                            }
                                        }
                                    }
                                }
                            }
                        }
                        else if (ssDCval >= 84) {
                            if (Cs < 3161) {
                                if (mu_mv_mag_sq < 123) {
                                    if (Rs < 2541) {
                                        if (SC < 3734) {
                                            if (SC < 2419) {
                                                if (AFD < 3) {
                                                    return 8;
                                                }
                                                else if (AFD >= 3) {
                                                    return 4;
                                                }
                                            }
                                            else if (SC >= 2419) {
                                                return 8;
                                            }
                                        }
                                        else if (SC >= 3734) {
                                            if (Cs < 2658) {
                                                if (negBalance < 44) {
                                                    return 4;
                                                }
                                                else if (negBalance >= 44) {
                                                    if (ssDCval < 99) {
                                                        return 8;
                                                    }
                                                    else if (ssDCval >= 99) {
                                                        if (RsDiff < 5) {
                                                            return 8;
                                                        }
                                                        else if (RsDiff >= 5) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (Cs >= 2658) {
                                                return 8;
                                            }
                                        }
                                    }
                                    else if (Rs >= 2541) {
                                        if (SC < 5698) {
                                            if (mu_mv_mag_sq < 99) {
                                                if (SC < 4846) {
                                                    return 1;
                                                }
                                                else if (SC >= 4846) {
                                                    return 8;
                                                }
                                            }
                                            else if (mu_mv_mag_sq >= 99) {
                                                if (negBalance < 49) {
                                                    if (RsCsDiff < 58) {
                                                        return 8;
                                                    }
                                                    else if (RsCsDiff >= 58) {
                                                        if (Cs < 1732) {
                                                            return 8;
                                                        }
                                                        else if (Cs >= 1732) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                                else if (negBalance >= 49) {
                                                    if (Cs < 2899) {
                                                        if (Rs < 2650) {
                                                            return 8;
                                                        }
                                                        else if (Rs >= 2650) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (Cs >= 2899) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                        }
                                        else if (SC >= 5698) {
                                            if (ssDCval < 115) {
                                                if (refDCval < 112) {
                                                    if (refDCval < 111) {
                                                        return 8;
                                                    }
                                                    else if (refDCval >= 111) {
                                                        if (Rs < 2807) {
                                                            if (Rs < 2802) {
                                                                return 8;
                                                            }
                                                            else if (Rs >= 2802) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (Rs >= 2807) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                                else if (refDCval >= 112) {
                                                    if (Cs < 2984) {
                                                        if (SC < 5704) {
                                                            return 4;
                                                        }
                                                        else if (SC >= 5704) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (Cs >= 2984) {
                                                        if (mcTcor < 97) {
                                                            return 8;
                                                        }
                                                        else if (mcTcor >= 97) {
                                                            if (RsCsDiff < 46) {
                                                                if (Cs < 3026) {
                                                                    if (refDCval < 113) {
                                                                        if (SC < 5838) {
                                                                            return 4;
                                                                        }
                                                                        else if (SC >= 5838) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                    else if (refDCval >= 113) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (Cs >= 3026) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (RsCsDiff >= 46) {
                                                                if (Rs < 2724) {
                                                                    return 8;
                                                                }
                                                                else if (Rs >= 2724) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                            else if (ssDCval >= 115) {
                                                return 8;
                                            }
                                        }
                                    }
                                }
                                else if (mu_mv_mag_sq >= 123) {
                                    if (gchDC < 1) {
                                        if (posBalance < 52) {
                                            if (mu_mv_mag_sq < 126) {
                                                return 4;
                                            }
                                            else if (mu_mv_mag_sq >= 126) {
                                                return 8;
                                            }
                                        }
                                        else if (posBalance >= 52) {
                                            if (negBalance < 48) {
                                                return 8;
                                            }
                                            else if (negBalance >= 48) {
                                                if (Rs < 2796) {
                                                    return 8;
                                                }
                                                else if (Rs >= 2796) {
                                                    if (ssDCval < 113) {
                                                        return 8;
                                                    }
                                                    else if (ssDCval >= 113) {
                                                        if (Rs < 2807) {
                                                            return 4;
                                                        }
                                                        else if (Rs >= 2807) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (gchDC >= 1) {
                                        if (Cs < 3040) {
                                            if (RsDiff < 4) {
                                                if (negBalance < 37) {
                                                    if (Cs < 2070) {
                                                        return 8;
                                                    }
                                                    else if (Cs >= 2070) {
                                                        return 4;
                                                    }
                                                }
                                                else if (negBalance >= 37) {
                                                    return 8;
                                                }
                                            }
                                            else if (RsDiff >= 4) {
                                                return 8;
                                            }
                                        }
                                        else if (Cs >= 3040) {
                                            if (diffRsCsdiff < -5) {
                                                if (Rs < 2995) {
                                                    return 4;
                                                }
                                                else if (Rs >= 2995) {
                                                    return 8;
                                                }
                                            }
                                            else if (diffRsCsdiff >= -5) {
                                                return 8;
                                            }
                                        }
                                    }
                                }
                            }
                            else if (Cs >= 3161) {
                                if (RsCsDiff < 123) {
                                    return 8;
                                }
                                else if (RsCsDiff >= 123) {
                                    if (Rs < 4936) {
                                        return 4;
                                    }
                                    else if (Rs >= 4936) {
                                        return 8;
                                    }
                                }
                            }
                        }
                    }
                }
            }
            else if (AFD >= 7) {
                if (Cs < 4178) {
                    if (Cs < 1695) {
                        if (RsCsDiff < 33) {
                            if (CsDiff < 4) {
                                if (diffAFD < 1) {
                                    return 2;
                                }
                                else if (diffAFD >= 1) {
                                    if (posBalance < 33) {
                                        if (Rs < 258) {
                                            return 1;
                                        }
                                        else if (Rs >= 258) {
                                            return 8;
                                        }
                                    }
                                    else if (posBalance >= 33) {
                                        if (RsCsDiff < 23) {
                                            return 2;
                                        }
                                        else if (RsCsDiff >= 23) {
                                            if (Cs < 1199) {
                                                return 8;
                                            }
                                            else if (Cs >= 1199) {
                                                return 2;
                                            }
                                        }
                                    }
                                }
                            }
                            else if (CsDiff >= 4) {
                                if (AFD < 8) {
                                    if (tcor < 90) {
                                        if (Rs < 1155) {
                                            return 4;
                                        }
                                        else if (Rs >= 1155) {
                                            return 8;
                                        }
                                    }
                                    else if (tcor >= 90) {
                                        if (refDCval < 85) {
                                            if (mu_mv_mag_sq < 116) {
                                                return 2;
                                            }
                                            else if (mu_mv_mag_sq >= 116) {
                                                if (MVDiff < 1) {
                                                    if (Rs < 1869) {
                                                        if (refDCval < 75) {
                                                            return 2;
                                                        }
                                                        else if (refDCval >= 75) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (Rs >= 1869) {
                                                        return 2;
                                                    }
                                                }
                                                else if (MVDiff >= 1) {
                                                    return 8;
                                                }
                                            }
                                        }
                                        else if (refDCval >= 85) {
                                            if (Cs < 1560) {
                                                return 8;
                                            }
                                            else if (Cs >= 1560) {
                                                return 4;
                                            }
                                        }
                                    }
                                }
                                else if (AFD >= 8) {
                                    return 2;
                                }
                            }
                        }
                        else if (RsCsDiff >= 33) {
                            if (mu_mv_mag_sq < 162) {
                                if (posBalance < 60) {
                                    if (Rs < 1821) {
                                        if (ssDCval < 72) {
                                            if (diffAFD < -1) {
                                                return 8;
                                            }
                                            else if (diffAFD >= -1) {
                                                return 2;
                                            }
                                        }
                                        else if (ssDCval >= 72) {
                                            if (gchDC < 3) {
                                                return 8;
                                            }
                                            else if (gchDC >= 3) {
                                                return 4;
                                            }
                                        }
                                    }
                                    else if (Rs >= 1821) {
                                        if (diffRsCsdiff < -10) {
                                            if (tcor < 58) {
                                                if (Rs < 2510) {
                                                    return 1;
                                                }
                                                else if (Rs >= 2510) {
                                                    return 2;
                                                }
                                            }
                                            else if (tcor >= 58) {
                                                return 4;
                                            }
                                        }
                                        else if (diffRsCsdiff >= -10) {
                                            if (negBalance < 48) {
                                                if (SC < 3983) {
                                                    if (SC < 3023) {
                                                        return 4;
                                                    }
                                                    else if (SC >= 3023) {
                                                        return 2;
                                                    }
                                                }
                                                else if (SC >= 3983) {
                                                    return 1;
                                                }
                                            }
                                            else if (negBalance >= 48) {
                                                if (RsCsDiff < 131) {
                                                    if (Cs < 1654) {
                                                        return 8;
                                                    }
                                                    else if (Cs >= 1654) {
                                                        return 2;
                                                    }
                                                }
                                                else if (RsCsDiff >= 131) {
                                                    if (RsDiff < 22) {
                                                        return 4;
                                                    }
                                                    else if (RsDiff >= 22) {
                                                        return 2;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (posBalance >= 60) {
                                    if (Cs < 1086) {
                                        if (TSC < 11) {
                                            return 8;
                                        }
                                        else if (TSC >= 11) {
                                            return 1;
                                        }
                                    }
                                    else if (Cs >= 1086) {
                                        if (mcTcor < 85) {
                                            if (Rs < 2347) {
                                                return 8;
                                            }
                                            else if (Rs >= 2347) {
                                                return 2;
                                            }
                                        }
                                        else if (mcTcor >= 85) {
                                            if (TSCindex < 8) {
                                                if (Rs < 1805) {
                                                    return 2;
                                                }
                                                else if (Rs >= 1805) {
                                                    return 8;
                                                }
                                            }
                                            else if (TSCindex >= 8) {
                                                return 2;
                                            }
                                        }
                                    }
                                }
                            }
                            else if (mu_mv_mag_sq >= 162) {
                                if (AFD < 33) {
                                    if (SC < 782) {
                                        return 2;
                                    }
                                    else if (SC >= 782) {
                                        if (Rs < 2816) {
                                            return 1;
                                        }
                                        else if (Rs >= 2816) {
                                            if (SC < 3781) {
                                                return 2;
                                            }
                                            else if (SC >= 3781) {
                                                return 1;
                                            }
                                        }
                                    }
                                }
                                else if (AFD >= 33) {
                                    if (CsDiff < 8) {
                                        return 2;
                                    }
                                    else if (CsDiff >= 8) {
                                        return 8;
                                    }
                                }
                            }
                        }
                    }
                    else if (Cs >= 1695) {
                        if (Rs < 5189) {
                            if (ssDCval < 109) {
                                if (ssDCval < 86) {
                                    if (posBalance < 44) {
                                        if (Rs < 2150) {
                                            return 1;
                                        }
                                        else if (Rs >= 2150) {
                                            return 8;
                                        }
                                    }
                                    else if (posBalance >= 44) {
                                        if (Rs < 1957) {
                                            if (refDCval < 81) {
                                                if (ssDCval < 77) {
                                                    if (Cs < 1858) {
                                                        return 4;
                                                    }
                                                    else if (Cs >= 1858) {
                                                        return 8;
                                                    }
                                                }
                                                else if (ssDCval >= 77) {
                                                    return 8;
                                                }
                                            }
                                            else if (refDCval >= 81) {
                                                return 2;
                                            }
                                        }
                                        else if (Rs >= 1957) {
                                            if (tcor < 92) {
                                                if (mcTcor < 80) {
                                                    return 4;
                                                }
                                                else if (mcTcor >= 80) {
                                                    if (tcor < 89) {
                                                        return 8;
                                                    }
                                                    else if (tcor >= 89) {
                                                        if (Rs < 2876) {
                                                            if (negBalance < 39) {
                                                                return 8;
                                                            }
                                                            else if (negBalance >= 39) {
                                                                if (posBalance < 57) {
                                                                    return 4;
                                                                }
                                                                else if (posBalance >= 57) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                        else if (Rs >= 2876) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (tcor >= 92) {
                                                if (Rs < 3121) {
                                                    return 4;
                                                }
                                                else if (Rs >= 3121) {
                                                    if (posBalance < 54) {
                                                        return 8;
                                                    }
                                                    else if (posBalance >= 54) {
                                                        if (Rs < 3126) {
                                                            return 8;
                                                        }
                                                        else if (Rs >= 3126) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (ssDCval >= 86) {
                                    if (refDCval < 97) {
                                        if (mcTcor < 92) {
                                            if (ssDCval < 93) {
                                                if (MVDiff < 8) {
                                                    return 8;
                                                }
                                                else if (MVDiff >= 8) {
                                                    if (Rs < 2011) {
                                                        return 8;
                                                    }
                                                    else if (Rs >= 2011) {
                                                        return 2;
                                                    }
                                                }
                                            }
                                            else if (ssDCval >= 93) {
                                                if (Cs < 1880) {
                                                    return 8;
                                                }
                                                else if (Cs >= 1880) {
                                                    return 2;
                                                }
                                            }
                                        }
                                        else if (mcTcor >= 92) {
                                            if (Cs < 2673) {
                                                if (posBalance < 46) {
                                                    return 1;
                                                }
                                                else if (posBalance >= 46) {
                                                    if (MVDiff < 2) {
                                                        if (AFD < 8) {
                                                            if (Cs < 1939) {
                                                                return 2;
                                                            }
                                                            else if (Cs >= 1939) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (AFD >= 8) {
                                                            return 2;
                                                        }
                                                    }
                                                    else if (MVDiff >= 2) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                            else if (Cs >= 2673) {
                                                if (Cs < 3834) {
                                                    return 2;
                                                }
                                                else if (Cs >= 3834) {
                                                    return 8;
                                                }
                                            }
                                        }
                                    }
                                    else if (refDCval >= 97) {
                                        if (negBalance < 44) {
                                            if (negBalance < 33) {
                                                if (Cs < 3498) {
                                                    return 2;
                                                }
                                                else if (Cs >= 3498) {
                                                    return 8;
                                                }
                                            }
                                            else if (negBalance >= 33) {
                                                return 8;
                                            }
                                        }
                                        else if (negBalance >= 44) {
                                            if (Cs < 3482) {
                                                if (RsDiff < 17) {
                                                    if (Cs < 3282) {
                                                        return 4;
                                                    }
                                                    else if (Cs >= 3282) {
                                                        if (Cs < 3327) {
                                                            return 8;
                                                        }
                                                        else if (Cs >= 3327) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                                else if (RsDiff >= 17) {
                                                    return 2;
                                                }
                                            }
                                            else if (Cs >= 3482) {
                                                return 8;
                                            }
                                        }
                                    }
                                }
                            }
                            else if (ssDCval >= 109) {
                                if (Rs < 3993) {
                                    if (diffMVdiffVal < -4) {
                                        if (mcTcor < 79) {
                                            return 1;
                                        }
                                        else if (mcTcor >= 79) {
                                            if (refDCval < 116) {
                                                if (Rs < 2620) {
                                                    return 4;
                                                }
                                                else if (Rs >= 2620) {
                                                    return 8;
                                                }
                                            }
                                            else if (refDCval >= 116) {
                                                return 4;
                                            }
                                        }
                                    }
                                    else if (diffMVdiffVal >= -4) {
                                        if (CsDiff < 12) {
                                            if (Rs < 2760) {
                                                if (tcor < 97) {
                                                    if (diffAFD < -4) {
                                                        return 4;
                                                    }
                                                    else if (diffAFD >= -4) {
                                                        if (diffMVdiffVal < -1) {
                                                            if (CsDiff < 6) {
                                                                return 4;
                                                            }
                                                            else if (CsDiff >= 6) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (diffMVdiffVal >= -1) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                                else if (tcor >= 97) {
                                                    return 4;
                                                }
                                            }
                                            else if (Rs >= 2760) {
                                                if (mu_mv_mag_sq < 92) {
                                                    if (tcor < 89) {
                                                        return 8;
                                                    }
                                                    else if (tcor >= 89) {
                                                        return 1;
                                                    }
                                                }
                                                else if (mu_mv_mag_sq >= 92) {
                                                    return 8;
                                                }
                                            }
                                        }
                                        else if (CsDiff >= 12) {
                                            return 1;
                                        }
                                    }
                                }
                                else if (Rs >= 3993) {
                                    if (ssDCval < 127) {
                                        if (refDCval < 115) {
                                            if (negBalance < 50) {
                                                if (Rs < 5079) {
                                                    if (mu_mv_mag_sq < 89) {
                                                        return 8;
                                                    }
                                                    else if (mu_mv_mag_sq >= 89) {
                                                        if (diffRsCsdiff < 12) {
                                                            return 2;
                                                        }
                                                        else if (diffRsCsdiff >= 12) {
                                                            if (SC < 8045) {
                                                                return 8;
                                                            }
                                                            else if (SC >= 8045) {
                                                                return 2;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (Rs >= 5079) {
                                                    return 8;
                                                }
                                            }
                                            else if (negBalance >= 50) {
                                                return 4;
                                            }
                                        }
                                        else if (refDCval >= 115) {
                                            if (Cs < 3822) {
                                                if (tcor < 96) {
                                                    return 4;
                                                }
                                                else if (tcor >= 96) {
                                                    return 8;
                                                }
                                            }
                                            else if (Cs >= 3822) {
                                                return 8;
                                            }
                                        }
                                    }
                                    else if (ssDCval >= 127) {
                                        if (negBalance < 41) {
                                            return 2;
                                        }
                                        else if (negBalance >= 41) {
                                            return 8;
                                        }
                                    }
                                }
                            }
                        }
                        else if (Rs >= 5189) {
                            if (MVDiff < 1) {
                                return 4;
                            }
                            else if (MVDiff >= 1) {
                                if (Cs < 3079) {
                                    return 2;
                                }
                                else if (Cs >= 3079) {
                                    return 8;
                                }
                            }
                        }
                    }
                }
                else if (Cs >= 4178) {
                    if (Rs < 7236) {
                        if (mu_mv_mag_sq < 122) {
                            if (SC < 8582) {
                                if (SC < 8470) {
                                    return 8;
                                }
                                else if (SC >= 8470) {
                                    return 1;
                                }
                            }
                            else if (SC >= 8582) {
                                return 8;
                            }
                        }
                        else if (mu_mv_mag_sq >= 122) {
                            if (refDCval < 106) {
                                if (ssDCval < 83) {
                                    if (Cs < 5186) {
                                        if (Rs < 4834) {
                                            return 8;
                                        }
                                        else if (Rs >= 4834) {
                                            if (AFD < 8) {
                                                return 4;
                                            }
                                            else if (AFD >= 8) {
                                                return 8;
                                            }
                                        }
                                    }
                                    else if (Cs >= 5186) {
                                        return 4;
                                    }
                                }
                                else if (ssDCval >= 83) {
                                    return 8;
                                }
                            }
                            else if (refDCval >= 106) {
                                if (Rs < 6342) {
                                    if (Cs < 5247) {
                                        return 8;
                                    }
                                    else if (Cs >= 5247) {
                                        if (MVDiff < 1) {
                                            return 4;
                                        }
                                        else if (MVDiff >= 1) {
                                            return 8;
                                        }
                                    }
                                }
                                else if (Rs >= 6342) {
                                    return 8;
                                }
                            }
                        }
                    }
                    else if (Rs >= 7236) {
                        if (negBalance < 60) {
                            if (Cs < 5522) {
                                return 4;
                            }
                            else if (Cs >= 5522) {
                                return 8;
                            }
                        }
                        else if (negBalance >= 60) {
                            if (Cs < 7701) {
                                return 4;
                            }
                            else if (Cs >= 7701) {
                                if (posBalance < 53) {
                                    return 2;
                                }
                                else if (posBalance >= 53) {
                                    if (Rs < 8151) {
                                        if (RsCsDiff < 950) {
                                            return 4;
                                        }
                                        else if (RsCsDiff >= 950) {
                                            return 2;
                                        }
                                    }
                                    else if (Rs >= 8151) {
                                        return 4;
                                    }
                                }
                            }
                        }
                    }
                }
            }
        }
        return 4;
    }
    mfxU16 AGOPSelectB(mfxU32 Rs, mfxU32 CsDiff, mfxU32 TSC, mfxI32 diffAFD, mfxU32 Scindex, mfxU32 RsDiff, mfxI32 diffMVdiffVal, mfxU32 RsCsDiff, mfxU32 TSCindex, mfxI16 tcor, mfxU32 refDCval, mfxU32 MVDiff, mfxU32 AFD, mfxI32 diffTSC, mfxI32 diffRsCsdiff, mfxU32 posBalance, mfxU32 SC, mfxU32 mu_mv_mag_sq, mfxU32 negBalance, mfxI16 mcTcor, mfxU32 gchDC, mfxU32 Cs, mfxU32 ssDCval) {
        if (mu_mv_mag_sq < 89) {
            if (mu_mv_mag_sq < 1) {
                if (Cs < 1772) {
                    return 8;
                }
                else if (Cs >= 1772) {
                    if (SC < 9327) {
                        if (Cs < 1891) {
                            return 1;
                        }
                        else if (Cs >= 1891) {
                            if (SC < 7755) {
                                if (Cs < 3881) {
                                    return 8;
                                }
                                else if (Cs >= 3881) {
                                    if (Rs < 3755) {
                                        return 8;
                                    }
                                    else if (Rs >= 3755) {
                                        return 1;
                                    }
                                }
                            }
                            else if (SC >= 7755) {
                                return 8;
                            }
                        }
                    }
                    else if (SC >= 9327) {
                        if (Rs < 5376) {
                            return 1;
                        }
                        else if (Rs >= 5376) {
                            return 8;
                        }
                    }
                }
            }
            else if (mu_mv_mag_sq >= 1) {
                if (SC < 7382) {
                    if (mu_mv_mag_sq < 28) {
                        if (mu_mv_mag_sq < 3) {
                            if (Cs < 317) {
                                if (RsDiff < 2) {
                                    if (MVDiff < 5) {
                                        return 8;
                                    }
                                    else if (MVDiff >= 5) {
                                        if (ssDCval < 88) {
                                            if (mcTcor < 60) {
                                                if (ssDCval < 20) {
                                                    return 8;
                                                }
                                                else if (ssDCval >= 20) {
                                                    return 1;
                                                }
                                            }
                                            else if (mcTcor >= 60) {
                                                if (diffRsCsdiff < -1) {
                                                    if (Cs < 208) {
                                                        return 1;
                                                    }
                                                    else if (Cs >= 208) {
                                                        return 2;
                                                    }
                                                }
                                                else if (diffRsCsdiff >= -1) {
                                                    if (SC < 550) {
                                                        if (posBalance < 87) {
                                                            return 2;
                                                        }
                                                        else if (posBalance >= 87) {
                                                            if (posBalance < 113) {
                                                                if (mcTcor < 69) {
                                                                    return 2;
                                                                }
                                                                else if (mcTcor >= 69) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (posBalance >= 113) {
                                                                return 2;
                                                            }
                                                        }
                                                    }
                                                    else if (SC >= 550) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                        }
                                        else if (ssDCval >= 88) {
                                            if (Cs < 82) {
                                                return 1;
                                            }
                                            else if (Cs >= 82) {
                                                return 4;
                                            }
                                        }
                                    }
                                }
                                else if (RsDiff >= 2) {
                                    if (ssDCval < 106) {
                                        return 1;
                                    }
                                    else if (ssDCval >= 106) {
                                        if (refDCval < 108) {
                                            return 1;
                                        }
                                        else if (refDCval >= 108) {
                                            return 2;
                                        }
                                    }
                                }
                            }
                            else if (Cs >= 317) {
                                if (refDCval < 126) {
                                    if (negBalance < 48) {
                                        if (posBalance < 57) {
                                            if (Cs < 1470) {
                                                if (posBalance < 48) {
                                                    if (ssDCval < 114) {
                                                        if (tcor < 96) {
                                                            if (AFD < 14) {
                                                                if (Cs < 1348) {
                                                                    return 4;
                                                                }
                                                                else if (Cs >= 1348) {
                                                                    return 2;
                                                                }
                                                            }
                                                            else if (AFD >= 14) {
                                                                return 2;
                                                            }
                                                        }
                                                        else if (tcor >= 96) {
                                                            if (MVDiff < 4) {
                                                                if (Cs < 814) {
                                                                    return 2;
                                                                }
                                                                else if (Cs >= 814) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (MVDiff >= 4) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                    else if (ssDCval >= 114) {
                                                        return 8;
                                                    }
                                                }
                                                else if (posBalance >= 48) {
                                                    if (Rs < 1083) {
                                                        if (mcTcor < 96) {
                                                            if (RsCsDiff < 18) {
                                                                return 4;
                                                            }
                                                            else if (RsCsDiff >= 18) {
                                                                return 2;
                                                            }
                                                        }
                                                        else if (mcTcor >= 96) {
                                                            if (Cs < 519) {
                                                                return 2;
                                                            }
                                                            else if (Cs >= 519) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                    else if (Rs >= 1083) {
                                                        if (RsDiff < 5) {
                                                            if (Cs < 956) {
                                                                if (SC < 2419) {
                                                                    return 1;
                                                                }
                                                                else if (SC >= 2419) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (Cs >= 956) {
                                                                if (posBalance < 53) {
                                                                    return 1;
                                                                }
                                                                else if (posBalance >= 53) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                        else if (RsDiff >= 5) {
                                                            if (SC < 2566) {
                                                                return 2;
                                                            }
                                                            else if (SC >= 2566) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                            else if (Cs >= 1470) {
                                                if (refDCval < 101) {
                                                    if (TSC < 11) {
                                                        if (refDCval < 80) {
                                                            if (MVDiff < 9) {
                                                                return 4;
                                                            }
                                                            else if (MVDiff >= 9) {
                                                                if (RsDiff < 8) {
                                                                    if (Rs < 1760) {
                                                                        return 4;
                                                                    }
                                                                    else if (Rs >= 1760) {
                                                                        if (Rs < 1960) {
                                                                            return 1;
                                                                        }
                                                                        else if (Rs >= 1960) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                }
                                                                else if (RsDiff >= 8) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                        else if (refDCval >= 80) {
                                                            if (SC < 3941) {
                                                                if (ssDCval < 90) {
                                                                    if (Rs < 1655) {
                                                                        return 2;
                                                                    }
                                                                    else if (Rs >= 1655) {
                                                                        return 1;
                                                                    }
                                                                }
                                                                else if (ssDCval >= 90) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (SC >= 3941) {
                                                                if (negBalance < 44) {
                                                                    return 1;
                                                                }
                                                                else if (negBalance >= 44) {
                                                                    if (Rs < 2600) {
                                                                        return 4;
                                                                    }
                                                                    else if (Rs >= 2600) {
                                                                        return 1;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (TSC >= 11) {
                                                        if (posBalance < 54) {
                                                            if (SC < 4175) {
                                                                return 4;
                                                            }
                                                            else if (SC >= 4175) {
                                                                if (diffRsCsdiff < -40) {
                                                                    if (Cs < 2108) {
                                                                        return 8;
                                                                    }
                                                                    else if (Cs >= 2108) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (diffRsCsdiff >= -40) {
                                                                    if (Cs < 2137) {
                                                                        return 8;
                                                                    }
                                                                    else if (Cs >= 2137) {
                                                                        if (SC < 5226) {
                                                                            return 1;
                                                                        }
                                                                        else if (SC >= 5226) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (posBalance >= 54) {
                                                            if (Rs < 2629) {
                                                                return 4;
                                                            }
                                                            else if (Rs >= 2629) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (refDCval >= 101) {
                                                    if (SC < 3009) {
                                                        return 2;
                                                    }
                                                    else if (SC >= 3009) {
                                                        if (Cs < 1745) {
                                                            if (Rs < 3159) {
                                                                return 8;
                                                            }
                                                            else if (Rs >= 3159) {
                                                                if (MVDiff < 11) {
                                                                    if (Rs < 3199) {
                                                                        return 4;
                                                                    }
                                                                    else if (Rs >= 3199) {
                                                                        return 8;
                                                                    }
                                                                }
                                                                else if (MVDiff >= 11) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                        else if (Cs >= 1745) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (posBalance >= 57) {
                                            if (TSC < 11) {
                                                if (AFD < 5) {
                                                    return 8;
                                                }
                                                else if (AFD >= 5) {
                                                    if (diffMVdiffVal < 2) {
                                                        if (CsDiff < 4) {
                                                            if (SC < 641) {
                                                                return 2;
                                                            }
                                                            else if (SC >= 641) {
                                                                if (posBalance < 58) {
                                                                    return 8;
                                                                }
                                                                else if (posBalance >= 58) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                        else if (CsDiff >= 4) {
                                                            if (SC < 5210) {
                                                                return 2;
                                                            }
                                                            else if (SC >= 5210) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                    else if (diffMVdiffVal >= 2) {
                                                        if (refDCval < 93) {
                                                            if (CsDiff < 7) {
                                                                if (Rs < 1087) {
                                                                    return 2;
                                                                }
                                                                else if (Rs >= 1087) {
                                                                    return 1;
                                                                }
                                                            }
                                                            else if (CsDiff >= 7) {
                                                                if (Cs < 1578) {
                                                                    return 2;
                                                                }
                                                                else if (Cs >= 1578) {
                                                                    return 1;
                                                                }
                                                            }
                                                        }
                                                        else if (refDCval >= 93) {
                                                            if (ssDCval < 121) {
                                                                return 8;
                                                            }
                                                            else if (ssDCval >= 121) {
                                                                if (Cs < 1598) {
                                                                    return 1;
                                                                }
                                                                else if (Cs >= 1598) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                            else if (TSC >= 11) {
                                                if (diffMVdiffVal < 2) {
                                                    if (Cs < 2776) {
                                                        if (posBalance < 58) {
                                                            return 4;
                                                        }
                                                        else if (posBalance >= 58) {
                                                            return 1;
                                                        }
                                                    }
                                                    else if (Cs >= 2776) {
                                                        return 4;
                                                    }
                                                }
                                                else if (diffMVdiffVal >= 2) {
                                                    return 1;
                                                }
                                            }
                                        }
                                    }
                                    else if (negBalance >= 48) {
                                        if (ssDCval < 116) {
                                            if (tcor < 77) {
                                                if (AFD < 18) {
                                                    if (RsCsDiff < 7) {
                                                        if (RsDiff < 2) {
                                                            if (ssDCval < 88) {
                                                                return 2;
                                                            }
                                                            else if (ssDCval >= 88) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (RsDiff >= 2) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (RsCsDiff >= 7) {
                                                        if (Rs < 1104) {
                                                            if (Rs < 610) {
                                                                if (mcTcor < 88) {
                                                                    if (MVDiff < 16) {
                                                                        return 8;
                                                                    }
                                                                    else if (MVDiff >= 16) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (mcTcor >= 88) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (Rs >= 610) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (Rs >= 1104) {
                                                            if (RsCsDiff < 92) {
                                                                if (refDCval < 98) {
                                                                    if (RsDiff < 5) {
                                                                        if (SC < 2204) {
                                                                            return 4;
                                                                        }
                                                                        else if (SC >= 2204) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                    else if (RsDiff >= 5) {
                                                                        return 8;
                                                                    }
                                                                }
                                                                else if (refDCval >= 98) {
                                                                    if (RsDiff < 4) {
                                                                        return 2;
                                                                    }
                                                                    else if (RsDiff >= 4) {
                                                                        if (TSC < 6) {
                                                                            return 8;
                                                                        }
                                                                        else if (TSC >= 6) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                            else if (RsCsDiff >= 92) {
                                                                if (MVDiff < 24) {
                                                                    if (CsDiff < 14) {
                                                                        if (diffRsCsdiff < -64) {
                                                                            return 2;
                                                                        }
                                                                        else if (diffRsCsdiff >= -64) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                    else if (CsDiff >= 14) {
                                                                        return 8;
                                                                    }
                                                                }
                                                                else if (MVDiff >= 24) {
                                                                    return 2;
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (AFD >= 18) {
                                                    if (RsCsDiff < 421) {
                                                        if (ssDCval < 86) {
                                                            if (Rs < 1773) {
                                                                if (SC < 1793) {
                                                                    return 1;
                                                                }
                                                                else if (SC >= 1793) {
                                                                    return 2;
                                                                }
                                                            }
                                                            else if (Rs >= 1773) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (ssDCval >= 86) {
                                                            if (Cs < 852) {
                                                                return 2;
                                                            }
                                                            else if (Cs >= 852) {
                                                                if (RsCsDiff < 267) {
                                                                    return 1;
                                                                }
                                                                else if (RsCsDiff >= 267) {
                                                                    if (Rs < 2062) {
                                                                        return 1;
                                                                    }
                                                                    else if (Rs >= 2062) {
                                                                        if (Cs < 2316) {
                                                                            return 2;
                                                                        }
                                                                        else if (Cs >= 2316) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (RsCsDiff >= 421) {
                                                        if (SC < 6607) {
                                                            return 4;
                                                        }
                                                        else if (SC >= 6607) {
                                                            return 2;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (tcor >= 77) {
                                                if (ssDCval < 104) {
                                                    if (tcor < 88) {
                                                        if (TSC < 13) {
                                                            if (Rs < 618) {
                                                                if (SC < 909) {
                                                                    return 2;
                                                                }
                                                                else if (SC >= 909) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (Rs >= 618) {
                                                                if (CsDiff < 3) {
                                                                    return 4;
                                                                }
                                                                else if (CsDiff >= 3) {
                                                                    if (SC < 5207) {
                                                                        if (negBalance < 54) {
                                                                            if (posBalance < 57) {
                                                                                if (SC < 3724) {
                                                                                    return 2;
                                                                                }
                                                                                else if (SC >= 3724) {
                                                                                    if (Rs < 2630) {
                                                                                        return 4;
                                                                                    }
                                                                                    else if (Rs >= 2630) {
                                                                                        return 1;
                                                                                    }
                                                                                }
                                                                            }
                                                                            else if (posBalance >= 57) {
                                                                                if (mcTcor < 86) {
                                                                                    if (diffAFD < -1) {
                                                                                        return 8;
                                                                                    }
                                                                                    else if (diffAFD >= -1) {
                                                                                        if (Cs < 1054) {
                                                                                            if (SC < 2009) {
                                                                                                return 8;
                                                                                            }
                                                                                            else if (SC >= 2009) {
                                                                                                return 4;
                                                                                            }
                                                                                        }
                                                                                        else if (Cs >= 1054) {
                                                                                            return 2;
                                                                                        }
                                                                                    }
                                                                                }
                                                                                else if (mcTcor >= 86) {
                                                                                    if (TSC < 9) {
                                                                                        if (posBalance < 59) {
                                                                                            if (TSC < 8) {
                                                                                                return 2;
                                                                                            }
                                                                                            else if (TSC >= 8) {
                                                                                                return 4;
                                                                                            }
                                                                                        }
                                                                                        else if (posBalance >= 59) {
                                                                                            if (AFD < 14) {
                                                                                                if (RsCsDiff < 105) {
                                                                                                    if (ssDCval < 62) {
                                                                                                        return 1;
                                                                                                    }
                                                                                                    else if (ssDCval >= 62) {
                                                                                                        return 4;
                                                                                                    }
                                                                                                }
                                                                                                else if (RsCsDiff >= 105) {
                                                                                                    return 1;
                                                                                                }
                                                                                            }
                                                                                            else if (AFD >= 14) {
                                                                                                return 2;
                                                                                            }
                                                                                        }
                                                                                    }
                                                                                    else if (TSC >= 9) {
                                                                                        if (TSC < 12) {
                                                                                            return 2;
                                                                                        }
                                                                                        else if (TSC >= 12) {
                                                                                            return 4;
                                                                                        }
                                                                                    }
                                                                                }
                                                                            }
                                                                        }
                                                                        else if (negBalance >= 54) {
                                                                            if (CsDiff < 4) {
                                                                                return 4;
                                                                            }
                                                                            else if (CsDiff >= 4) {
                                                                                if (negBalance < 69) {
                                                                                    return 2;
                                                                                }
                                                                                else if (negBalance >= 69) {
                                                                                    return 1;
                                                                                }
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (SC >= 5207) {
                                                                        if (posBalance < 59) {
                                                                            return 4;
                                                                        }
                                                                        else if (posBalance >= 59) {
                                                                            if (diffRsCsdiff < -18) {
                                                                                return 1;
                                                                            }
                                                                            else if (diffRsCsdiff >= -18) {
                                                                                if (Rs < 3123) {
                                                                                    return 8;
                                                                                }
                                                                                else if (Rs >= 3123) {
                                                                                    return 4;
                                                                                }
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (TSC >= 13) {
                                                            if (ssDCval < 77) {
                                                                if (TSC < 14) {
                                                                    return 4;
                                                                }
                                                                else if (TSC >= 14) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (ssDCval >= 77) {
                                                                if (Rs < 2468) {
                                                                    return 2;
                                                                }
                                                                else if (Rs >= 2468) {
                                                                    if (posBalance < 51) {
                                                                        return 1;
                                                                    }
                                                                    else if (posBalance >= 51) {
                                                                        return 8;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (tcor >= 88) {
                                                        if (negBalance < 53) {
                                                            if (Scindex < 5) {
                                                                if (ssDCval < 60) {
                                                                    if (Rs < 914) {
                                                                        return 4;
                                                                    }
                                                                    else if (Rs >= 914) {
                                                                        if (Cs < 955) {
                                                                            return 1;
                                                                        }
                                                                        else if (Cs >= 955) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                }
                                                                else if (ssDCval >= 60) {
                                                                    if (MVDiff < 10) {
                                                                        return 2;
                                                                    }
                                                                    else if (MVDiff >= 10) {
                                                                        return 4;
                                                                    }
                                                                }
                                                            }
                                                            else if (Scindex >= 5) {
                                                                if (refDCval < 94) {
                                                                    if (CsDiff < 8) {
                                                                        return 1;
                                                                    }
                                                                    else if (CsDiff >= 8) {
                                                                        return 2;
                                                                    }
                                                                }
                                                                else if (refDCval >= 94) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                        else if (negBalance >= 53) {
                                                            if (negBalance < 65) {
                                                                return 2;
                                                            }
                                                            else if (negBalance >= 65) {
                                                                if (Rs < 1195) {
                                                                    return 1;
                                                                }
                                                                else if (Rs >= 1195) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (ssDCval >= 104) {
                                                    if (MVDiff < 22) {
                                                        if (refDCval < 111) {
                                                            if (tcor < 86) {
                                                                return 8;
                                                            }
                                                            else if (tcor >= 86) {
                                                                if (negBalance < 51) {
                                                                    return 8;
                                                                }
                                                                else if (negBalance >= 51) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                        else if (refDCval >= 111) {
                                                            if (diffAFD < 2) {
                                                                return 2;
                                                            }
                                                            else if (diffAFD >= 2) {
                                                                if (Rs < 3297) {
                                                                    return 8;
                                                                }
                                                                else if (Rs >= 3297) {
                                                                    return 1;
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (MVDiff >= 22) {
                                                        if (Rs < 853) {
                                                            return 4;
                                                        }
                                                        else if (Rs >= 853) {
                                                            return 2;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (ssDCval >= 116) {
                                            if (Rs < 3127) {
                                                if (posBalance < 57) {
                                                    return 1;
                                                }
                                                else if (posBalance >= 57) {
                                                    if (posBalance < 59) {
                                                        if (tcor < 84) {
                                                            return 4;
                                                        }
                                                        else if (tcor >= 84) {
                                                            if (Rs < 2508) {
                                                                return 8;
                                                            }
                                                            else if (Rs >= 2508) {
                                                                if (Cs < 2012) {
                                                                    return 4;
                                                                }
                                                                else if (Cs >= 2012) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (posBalance >= 59) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                            else if (Rs >= 3127) {
                                                if (MVDiff < 14) {
                                                    if (Cs < 2879) {
                                                        return 1;
                                                    }
                                                    else if (Cs >= 2879) {
                                                        return 8;
                                                    }
                                                }
                                                else if (MVDiff >= 14) {
                                                    return 2;
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (refDCval >= 126) {
                                    if (posBalance < 60) {
                                        if (MVDiff < 27) {
                                            if (posBalance < 46) {
                                                if (Rs < 2903) {
                                                    return 4;
                                                }
                                                else if (Rs >= 2903) {
                                                    return 8;
                                                }
                                            }
                                            else if (posBalance >= 46) {
                                                if (mcTcor < 90) {
                                                    if (Cs < 1198) {
                                                        if (Cs < 1132) {
                                                            return 1;
                                                        }
                                                        else if (Cs >= 1132) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (Cs >= 1198) {
                                                        if (ssDCval < 128) {
                                                            if (refDCval < 127) {
                                                                return 8;
                                                            }
                                                            else if (refDCval >= 127) {
                                                                return 1;
                                                            }
                                                        }
                                                        else if (ssDCval >= 128) {
                                                            return 1;
                                                        }
                                                    }
                                                }
                                                else if (mcTcor >= 90) {
                                                    if (AFD < 12) {
                                                        if (Rs < 1104) {
                                                            return 1;
                                                        }
                                                        else if (Rs >= 1104) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (AFD >= 12) {
                                                        return 2;
                                                    }
                                                }
                                            }
                                        }
                                        else if (MVDiff >= 27) {
                                            if (SC < 4243) {
                                                if (MVDiff < 57) {
                                                    return 2;
                                                }
                                                else if (MVDiff >= 57) {
                                                    return 1;
                                                }
                                            }
                                            else if (SC >= 4243) {
                                                return 4;
                                            }
                                        }
                                    }
                                    else if (posBalance >= 60) {
                                        if (MVDiff < 21) {
                                            if (refDCval < 139) {
                                                if (tcor < 81) {
                                                    return 1;
                                                }
                                                else if (tcor >= 81) {
                                                    if (Cs < 855) {
                                                        return 1;
                                                    }
                                                    else if (Cs >= 855) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                            else if (refDCval >= 139) {
                                                if (TSC < 7) {
                                                    return 8;
                                                }
                                                else if (TSC >= 7) {
                                                    return 4;
                                                }
                                            }
                                        }
                                        else if (MVDiff >= 21) {
                                            if (SC < 4561) {
                                                if (diffMVdiffVal < 65) {
                                                    return 2;
                                                }
                                                else if (diffMVdiffVal >= 65) {
                                                    if (Cs < 537) {
                                                        return 4;
                                                    }
                                                    else if (Cs >= 537) {
                                                        return 1;
                                                    }
                                                }
                                            }
                                            else if (SC >= 4561) {
                                                if (Rs < 3429) {
                                                    if (tcor < 80) {
                                                        return 4;
                                                    }
                                                    else if (tcor >= 80) {
                                                        return 2;
                                                    }
                                                }
                                                else if (Rs >= 3429) {
                                                    if (AFD < 21) {
                                                        return 1;
                                                    }
                                                    else if (AFD >= 21) {
                                                        return 2;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                        }
                        else if (mu_mv_mag_sq >= 3) {
                            if (ssDCval < 138) {
                                if (Cs < 1904) {
                                    if (RsDiff < 11) {
                                        if (Cs < 1519) {
                                            if (posBalance < 63) {
                                                if (posBalance < 25) {
                                                    if (negBalance < 102) {
                                                        if (posBalance < 24) {
                                                            if (mcTcor < 37) {
                                                                if (negBalance < 47) {
                                                                    return 2;
                                                                }
                                                                else if (negBalance >= 47) {
                                                                    return 1;
                                                                }
                                                            }
                                                            else if (mcTcor >= 37) {
                                                                if (mu_mv_mag_sq < 8) {
                                                                    return 2;
                                                                }
                                                                else if (mu_mv_mag_sq >= 8) {
                                                                    if (diffMVdiffVal < 12) {
                                                                        if (MVDiff < 28) {
                                                                            return 4;
                                                                        }
                                                                        else if (MVDiff >= 28) {
                                                                            if (MVDiff < 45) {
                                                                                return 2;
                                                                            }
                                                                            else if (MVDiff >= 45) {
                                                                                if (AFD < 1) {
                                                                                    if (refDCval < 20) {
                                                                                        return 8;
                                                                                    }
                                                                                    else if (refDCval >= 20) {
                                                                                        return 2;
                                                                                    }
                                                                                }
                                                                                else if (AFD >= 1) {
                                                                                    return 4;
                                                                                }
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (diffMVdiffVal >= 12) {
                                                                        if (Cs < 88) {
                                                                            return 4;
                                                                        }
                                                                        else if (Cs >= 88) {
                                                                            return 2;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (posBalance >= 24) {
                                                            if (Rs < 321) {
                                                                return 8;
                                                            }
                                                            else if (Rs >= 321) {
                                                                return 2;
                                                            }
                                                        }
                                                    }
                                                    else if (negBalance >= 102) {
                                                        if (diffMVdiffVal < 78) {
                                                            if (MVDiff < 83) {
                                                                if (diffTSC < -1) {
                                                                    return 2;
                                                                }
                                                                else if (diffTSC >= -1) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (MVDiff >= 83) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (diffMVdiffVal >= 78) {
                                                            return 1;
                                                        }
                                                    }
                                                }
                                                else if (posBalance >= 25) {
                                                    if (Cs < 282) {
                                                        if (TSC < 5) {
                                                            if (diffAFD < -2) {
                                                                return 1;
                                                            }
                                                            else if (diffAFD >= -2) {
                                                                if (SC < 9) {
                                                                    if (diffMVdiffVal < 247) {
                                                                        if (Rs < 2) {
                                                                            return 4;
                                                                        }
                                                                        else if (Rs >= 2) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                    else if (diffMVdiffVal >= 247) {
                                                                        return 2;
                                                                    }
                                                                }
                                                                else if (SC >= 9) {
                                                                    return 2;
                                                                }
                                                            }
                                                        }
                                                        else if (TSC >= 5) {
                                                            if (ssDCval < 77) {
                                                                if (Rs < 151) {
                                                                    return 1;
                                                                }
                                                                else if (Rs >= 151) {
                                                                    return 2;
                                                                }
                                                            }
                                                            else if (ssDCval >= 77) {
                                                                if (mcTcor < 95) {
                                                                    return 1;
                                                                }
                                                                else if (mcTcor >= 95) {
                                                                    if (Rs < 623) {
                                                                        return 2;
                                                                    }
                                                                    else if (Rs >= 623) {
                                                                        return 1;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (Cs >= 282) {
                                                        if (tcor < 80) {
                                                            if (RsCsDiff < 33) {
                                                                if (tcor < 71) {
                                                                    if (RsDiff < 4) {
                                                                        if (diffTSC < 1) {
                                                                            return 4;
                                                                        }
                                                                        else if (diffTSC >= 1) {
                                                                            if (Rs < 680) {
                                                                                return 2;
                                                                            }
                                                                            else if (Rs >= 680) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (RsDiff >= 4) {
                                                                        return 2;
                                                                    }
                                                                }
                                                                else if (tcor >= 71) {
                                                                    if (negBalance < 53) {
                                                                        if (SC < 1206) {
                                                                            return 8;
                                                                        }
                                                                        else if (SC >= 1206) {
                                                                            if (tcor < 74) {
                                                                                return 8;
                                                                            }
                                                                            else if (tcor >= 74) {
                                                                                return 1;
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (negBalance >= 53) {
                                                                        if (Cs < 685) {
                                                                            return 2;
                                                                        }
                                                                        else if (Cs >= 685) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                            else if (RsCsDiff >= 33) {
                                                                if (MVDiff < 19) {
                                                                    if (mcTcor < 90) {
                                                                        if (mu_mv_mag_sq < 5) {
                                                                            if (SC < 2002) {
                                                                                return 8;
                                                                            }
                                                                            else if (SC >= 2002) {
                                                                                return 1;
                                                                            }
                                                                        }
                                                                        else if (mu_mv_mag_sq >= 5) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                    else if (mcTcor >= 90) {
                                                                        if (AFD < 15) {
                                                                            if (AFD < 11) {
                                                                                return 4;
                                                                            }
                                                                            else if (AFD >= 11) {
                                                                                return 1;
                                                                            }
                                                                        }
                                                                        else if (AFD >= 15) {
                                                                            if (TSC < 5) {
                                                                                return 8;
                                                                            }
                                                                            else if (TSC >= 5) {
                                                                                return 2;
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                                else if (MVDiff >= 19) {
                                                                    if (TSC < 10) {
                                                                        if (diffAFD < -8) {
                                                                            return 1;
                                                                        }
                                                                        else if (diffAFD >= -8) {
                                                                            if (mcTcor < 75) {
                                                                                return 1;
                                                                            }
                                                                            else if (mcTcor >= 75) {
                                                                                return 2;
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (TSC >= 10) {
                                                                        if (ssDCval < 75) {
                                                                            if (diffRsCsdiff < -92) {
                                                                                return 1;
                                                                            }
                                                                            else if (diffRsCsdiff >= -92) {
                                                                                return 8;
                                                                            }
                                                                        }
                                                                        else if (ssDCval >= 75) {
                                                                            if (Rs < 841) {
                                                                                return 1;
                                                                            }
                                                                            else if (Rs >= 841) {
                                                                                if (gchDC < 1) {
                                                                                    return 1;
                                                                                }
                                                                                else if (gchDC >= 1) {
                                                                                    return 2;
                                                                                }
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (tcor >= 80) {
                                                            if (mcTcor < 97) {
                                                                if (AFD < 14) {
                                                                    if (Rs < 1651) {
                                                                        if (RsCsDiff < 22) {
                                                                            if (Rs < 738) {
                                                                                return 2;
                                                                            }
                                                                            else if (Rs >= 738) {
                                                                                return 1;
                                                                            }
                                                                        }
                                                                        else if (RsCsDiff >= 22) {
                                                                            return 2;
                                                                        }
                                                                    }
                                                                    else if (Rs >= 1651) {
                                                                        if (refDCval < 80) {
                                                                            return 1;
                                                                        }
                                                                        else if (refDCval >= 80) {
                                                                            if (AFD < 9) {
                                                                                return 4;
                                                                            }
                                                                            else if (AFD >= 9) {
                                                                                return 2;
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                                else if (AFD >= 14) {
                                                                    if (ssDCval < 74) {
                                                                        return 4;
                                                                    }
                                                                    else if (ssDCval >= 74) {
                                                                        return 2;
                                                                    }
                                                                }
                                                            }
                                                            else if (mcTcor >= 97) {
                                                                if (CsDiff < 5) {
                                                                    if (diffTSC < -1) {
                                                                        if (Rs < 1908) {
                                                                            return 2;
                                                                        }
                                                                        else if (Rs >= 1908) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                    else if (diffTSC >= -1) {
                                                                        return 2;
                                                                    }
                                                                }
                                                                else if (CsDiff >= 5) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                            else if (posBalance >= 63) {
                                                if (MVDiff < 17) {
                                                    if (Cs < 908) {
                                                        if (tcor < 87) {
                                                            if (AFD < 11) {
                                                                if (Rs < 968) {
                                                                    return 1;
                                                                }
                                                                else if (Rs >= 968) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (AFD >= 11) {
                                                                return 2;
                                                            }
                                                        }
                                                        else if (tcor >= 87) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (Cs >= 908) {
                                                        return 4;
                                                    }
                                                }
                                                else if (MVDiff >= 17) {
                                                    if (ssDCval < 29) {
                                                        if (mu_mv_mag_sq < 9) {
                                                            if (MVDiff < 254) {
                                                                return 8;
                                                            }
                                                            else if (MVDiff >= 254) {
                                                                return 1;
                                                            }
                                                        }
                                                        else if (mu_mv_mag_sq >= 9) {
                                                            if (SC < 154) {
                                                                if (ssDCval < 22) {
                                                                    if (diffMVdiffVal < -261) {
                                                                        return 2;
                                                                    }
                                                                    else if (diffMVdiffVal >= -261) {
                                                                        if (tcor < 62) {
                                                                            if (diffMVdiffVal < 62) {
                                                                                if (refDCval < 20) {
                                                                                    return 2;
                                                                                }
                                                                                else if (refDCval >= 20) {
                                                                                    return 4;
                                                                                }
                                                                            }
                                                                            else if (diffMVdiffVal >= 62) {
                                                                                if (TSCindex < 1) {
                                                                                    return 8;
                                                                                }
                                                                                else if (TSCindex >= 1) {
                                                                                    return 4;
                                                                                }
                                                                            }
                                                                        }
                                                                        else if (tcor >= 62) {
                                                                            if (MVDiff < 98) {
                                                                                return 4;
                                                                            }
                                                                            else if (MVDiff >= 98) {
                                                                                return 8;
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                                else if (ssDCval >= 22) {
                                                                    if (MVDiff < 394) {
                                                                        return 4;
                                                                    }
                                                                    else if (MVDiff >= 394) {
                                                                        return 1;
                                                                    }
                                                                }
                                                            }
                                                            else if (SC >= 154) {
                                                                return 2;
                                                            }
                                                        }
                                                    }
                                                    else if (ssDCval >= 29) {
                                                        if (mcTcor < 96) {
                                                            if (ssDCval < 87) {
                                                                if (mcTcor < 90) {
                                                                    if (negBalance < 25) {
                                                                        if (refDCval < 38) {
                                                                            return 1;
                                                                        }
                                                                        else if (refDCval >= 38) {
                                                                            if (MVDiff < 25) {
                                                                                return 8;
                                                                            }
                                                                            else if (MVDiff >= 25) {
                                                                                return 2;
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (negBalance >= 25) {
                                                                        if (AFD < 35) {
                                                                            if (ssDCval < 40) {
                                                                                if (Rs < 359) {
                                                                                    return 1;
                                                                                }
                                                                                else if (Rs >= 359) {
                                                                                    return 4;
                                                                                }
                                                                            }
                                                                            else if (ssDCval >= 40) {
                                                                                return 1;
                                                                            }
                                                                        }
                                                                        else if (AFD >= 35) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                }
                                                                else if (mcTcor >= 90) {
                                                                    if (MVDiff < 19) {
                                                                        if (SC < 1087) {
                                                                            return 8;
                                                                        }
                                                                        else if (SC >= 1087) {
                                                                            return 2;
                                                                        }
                                                                    }
                                                                    else if (MVDiff >= 19) {
                                                                        return 2;
                                                                    }
                                                                }
                                                            }
                                                            else if (ssDCval >= 87) {
                                                                return 1;
                                                            }
                                                        }
                                                        else if (mcTcor >= 96) {
                                                            if (AFD < 21) {
                                                                return 2;
                                                            }
                                                            else if (AFD >= 21) {
                                                                if (Rs < 1180) {
                                                                    return 8;
                                                                }
                                                                else if (Rs >= 1180) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (Cs >= 1519) {
                                            if (TSC < 12) {
                                                if (refDCval < 100) {
                                                    if (mcTcor < 92) {
                                                        if (AFD < 16) {
                                                            if (posBalance < 55) {
                                                                return 2;
                                                            }
                                                            else if (posBalance >= 55) {
                                                                if (Rs < 1322) {
                                                                    return 1;
                                                                }
                                                                else if (Rs >= 1322) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                        else if (AFD >= 16) {
                                                            if (SC < 3279) {
                                                                return 1;
                                                            }
                                                            else if (SC >= 3279) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                    else if (mcTcor >= 92) {
                                                        if (AFD < 18) {
                                                            if (RsDiff < 7) {
                                                                return 8;
                                                            }
                                                            else if (RsDiff >= 7) {
                                                                return 1;
                                                            }
                                                        }
                                                        else if (AFD >= 18) {
                                                            return 2;
                                                        }
                                                    }
                                                }
                                                else if (refDCval >= 100) {
                                                    if (ssDCval < 129) {
                                                        if (AFD < 19) {
                                                            if (negBalance < 40) {
                                                                if (Cs < 1621) {
                                                                    return 8;
                                                                }
                                                                else if (Cs >= 1621) {
                                                                    if (Cs < 1726) {
                                                                        return 4;
                                                                    }
                                                                    else if (Cs >= 1726) {
                                                                        return 8;
                                                                    }
                                                                }
                                                            }
                                                            else if (negBalance >= 40) {
                                                                if (MVDiff < 8) {
                                                                    return 2;
                                                                }
                                                                else if (MVDiff >= 8) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                        else if (AFD >= 19) {
                                                            return 2;
                                                        }
                                                    }
                                                    else if (ssDCval >= 129) {
                                                        if (RsDiff < 10) {
                                                            if (Rs < 1266) {
                                                                return 4;
                                                            }
                                                            else if (Rs >= 1266) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (RsDiff >= 10) {
                                                            return 1;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (TSC >= 12) {
                                                if (AFD < 21) {
                                                    if (Cs < 1829) {
                                                        return 8;
                                                    }
                                                    else if (Cs >= 1829) {
                                                        if (Cs < 1842) {
                                                            return 1;
                                                        }
                                                        else if (Cs >= 1842) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                                else if (AFD >= 21) {
                                                    return 1;
                                                }
                                            }
                                        }
                                    }
                                    else if (RsDiff >= 11) {
                                        if (tcor < 82) {
                                            if (ssDCval < 132) {
                                                if (ssDCval < 113) {
                                                    if (MVDiff < 37) {
                                                        return 2;
                                                    }
                                                    else if (MVDiff >= 37) {
                                                        if (diffRsCsdiff < 87) {
                                                            if (Rs < 1766) {
                                                                return 1;
                                                            }
                                                            else if (Rs >= 1766) {
                                                                if (diffMVdiffVal < 7) {
                                                                    if (diffTSC < -2) {
                                                                        return 2;
                                                                    }
                                                                    else if (diffTSC >= -2) {
                                                                        return 1;
                                                                    }
                                                                }
                                                                else if (diffMVdiffVal >= 7) {
                                                                    if (Rs < 3094) {
                                                                        return 2;
                                                                    }
                                                                    else if (Rs >= 3094) {
                                                                        return 1;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (diffRsCsdiff >= 87) {
                                                            if (refDCval < 72) {
                                                                return 4;
                                                            }
                                                            else if (refDCval >= 72) {
                                                                return 2;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (ssDCval >= 113) {
                                                    if (CsDiff < 12) {
                                                        if (Cs < 1551) {
                                                            return 2;
                                                        }
                                                        else if (Cs >= 1551) {
                                                            if (AFD < 21) {
                                                                if (tcor < 80) {
                                                                    return 4;
                                                                }
                                                                else if (tcor >= 80) {
                                                                    if (Cs < 1618) {
                                                                        return 4;
                                                                    }
                                                                    else if (Cs >= 1618) {
                                                                        return 2;
                                                                    }
                                                                }
                                                            }
                                                            else if (AFD >= 21) {
                                                                if (Cs < 1817) {
                                                                    return 8;
                                                                }
                                                                else if (Cs >= 1817) {
                                                                    return 2;
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (CsDiff >= 12) {
                                                        if (AFD < 22) {
                                                            return 4;
                                                        }
                                                        else if (AFD >= 22) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (ssDCval >= 132) {
                                                if (AFD < 17) {
                                                    return 2;
                                                }
                                                else if (AFD >= 17) {
                                                    if (AFD < 32) {
                                                        return 1;
                                                    }
                                                    else if (AFD >= 32) {
                                                        return 2;
                                                    }
                                                }
                                            }
                                        }
                                        else if (tcor >= 82) {
                                            if (Cs < 1807) {
                                                if (mu_mv_mag_sq < 24) {
                                                    return 2;
                                                }
                                                else if (mu_mv_mag_sq >= 24) {
                                                    if (Rs < 2186) {
                                                        return 2;
                                                    }
                                                    else if (Rs >= 2186) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                            else if (Cs >= 1807) {
                                                return 8;
                                            }
                                        }
                                    }
                                }
                                else if (Cs >= 1904) {
                                    if (mcTcor < 94) {
                                        if (ssDCval < 121) {
                                            if (mcTcor < 89) {
                                                if (mcTcor < 69) {
                                                    if (gchDC < 5) {
                                                        return 4;
                                                    }
                                                    else if (gchDC >= 5) {
                                                        if (AFD < 50) {
                                                            return 8;
                                                        }
                                                        else if (AFD >= 50) {
                                                            return 2;
                                                        }
                                                    }
                                                }
                                                else if (mcTcor >= 69) {
                                                    if (ssDCval < 83) {
                                                        if (AFD < 22) {
                                                            if (SC < 3976) {
                                                                return 1;
                                                            }
                                                            else if (SC >= 3976) {
                                                                if (Cs < 3097) {
                                                                    return 8;
                                                                }
                                                                else if (Cs >= 3097) {
                                                                    return 1;
                                                                }
                                                            }
                                                        }
                                                        else if (AFD >= 22) {
                                                            if (posBalance < 56) {
                                                                if (AFD < 23) {
                                                                    return 8;
                                                                }
                                                                else if (AFD >= 23) {
                                                                    return 1;
                                                                }
                                                            }
                                                            else if (posBalance >= 56) {
                                                                if (Rs < 3546) {
                                                                    return 1;
                                                                }
                                                                else if (Rs >= 3546) {
                                                                    return 2;
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (ssDCval >= 83) {
                                                        if (diffRsCsdiff < 155) {
                                                            if (TSC < 12) {
                                                                return 2;
                                                            }
                                                            else if (TSC >= 12) {
                                                                if (Cs < 2792) {
                                                                    if (MVDiff < 70) {
                                                                        return 1;
                                                                    }
                                                                    else if (MVDiff >= 70) {
                                                                        return 8;
                                                                    }
                                                                }
                                                                else if (Cs >= 2792) {
                                                                    if (TSC < 15) {
                                                                        if (Rs < 3541) {
                                                                            return 2;
                                                                        }
                                                                        else if (Rs >= 3541) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                    else if (TSC >= 15) {
                                                                        return 8;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (diffRsCsdiff >= 155) {
                                                            if (SC < 5146) {
                                                                return 8;
                                                            }
                                                            else if (SC >= 5146) {
                                                                if (Rs < 3893) {
                                                                    return 4;
                                                                }
                                                                else if (Rs >= 3893) {
                                                                    return 2;
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                            else if (mcTcor >= 89) {
                                                if (mu_mv_mag_sq < 23) {
                                                    if (ssDCval < 96) {
                                                        if (posBalance < 54) {
                                                            if (Rs < 2988) {
                                                                return 1;
                                                            }
                                                            else if (Rs >= 2988) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (posBalance >= 54) {
                                                            if (CsDiff < 15) {
                                                                if (Rs < 2587) {
                                                                    if (diffMVdiffVal < 14) {
                                                                        if (mcTcor < 93) {
                                                                            return 4;
                                                                        }
                                                                        else if (mcTcor >= 93) {
                                                                            if (Cs < 2146) {
                                                                                return 4;
                                                                            }
                                                                            else if (Cs >= 2146) {
                                                                                return 1;
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (diffMVdiffVal >= 14) {
                                                                        return 1;
                                                                    }
                                                                }
                                                                else if (Rs >= 2587) {
                                                                    if (AFD < 18) {
                                                                        return 4;
                                                                    }
                                                                    else if (AFD >= 18) {
                                                                        return 1;
                                                                    }
                                                                }
                                                            }
                                                            else if (CsDiff >= 15) {
                                                                return 1;
                                                            }
                                                        }
                                                    }
                                                    else if (ssDCval >= 96) {
                                                        if (negBalance < 57) {
                                                            if (Cs < 2322) {
                                                                if (refDCval < 111) {
                                                                    if (Cs < 1958) {
                                                                        return 4;
                                                                    }
                                                                    else if (Cs >= 1958) {
                                                                        return 1;
                                                                    }
                                                                }
                                                                else if (refDCval >= 111) {
                                                                    if (Rs < 3376) {
                                                                        return 8;
                                                                    }
                                                                    else if (Rs >= 3376) {
                                                                        return 2;
                                                                    }
                                                                }
                                                            }
                                                            else if (Cs >= 2322) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (negBalance >= 57) {
                                                            if (RsDiff < 16) {
                                                                if (Cs < 2008) {
                                                                    return 2;
                                                                }
                                                                else if (Cs >= 2008) {
                                                                    return 1;
                                                                }
                                                            }
                                                            else if (RsDiff >= 16) {
                                                                return 2;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (mu_mv_mag_sq >= 23) {
                                                    if (RsCsDiff < 173) {
                                                        return 2;
                                                    }
                                                    else if (RsCsDiff >= 173) {
                                                        if (Cs < 2660) {
                                                            return 1;
                                                        }
                                                        else if (Cs >= 2660) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (ssDCval >= 121) {
                                            if (MVDiff < 24) {
                                                return 1;
                                            }
                                            else if (MVDiff >= 24) {
                                                if (refDCval < 133) {
                                                    if (tcor < 76) {
                                                        return 1;
                                                    }
                                                    else if (tcor >= 76) {
                                                        return 4;
                                                    }
                                                }
                                                else if (refDCval >= 133) {
                                                    return 4;
                                                }
                                            }
                                        }
                                    }
                                    else if (mcTcor >= 94) {
                                        if (ssDCval < 92) {
                                            if (AFD < 11) {
                                                return 2;
                                            }
                                            else if (AFD >= 11) {
                                                return 8;
                                            }
                                        }
                                        else if (ssDCval >= 92) {
                                            if (Rs < 2509) {
                                                return 2;
                                            }
                                            else if (Rs >= 2509) {
                                                return 4;
                                            }
                                        }
                                    }
                                }
                            }
                            else if (ssDCval >= 138) {
                                if (CsDiff < 6) {
                                    if (MVDiff < 73) {
                                        if (MVDiff < 13) {
                                            if (negBalance < 36) {
                                                return 4;
                                            }
                                            else if (negBalance >= 36) {
                                                return 2;
                                            }
                                        }
                                        else if (MVDiff >= 13) {
                                            return 2;
                                        }
                                    }
                                    else if (MVDiff >= 73) {
                                        if (negBalance < 82) {
                                            if (AFD < 10) {
                                                return 1;
                                            }
                                            else if (AFD >= 10) {
                                                if (Cs < 786) {
                                                    return 1;
                                                }
                                                else if (Cs >= 786) {
                                                    return 2;
                                                }
                                            }
                                        }
                                        else if (negBalance >= 82) {
                                            if (Rs < 702) {
                                                return 2;
                                            }
                                            else if (Rs >= 702) {
                                                return 8;
                                            }
                                        }
                                    }
                                }
                                else if (CsDiff >= 6) {
                                    if (diffTSC < 2) {
                                        if (mcTcor < 91) {
                                            if (Cs < 2624) {
                                                if (tcor < 55) {
                                                    return 1;
                                                }
                                                else if (tcor >= 55) {
                                                    if (diffTSC < -2) {
                                                        return 1;
                                                    }
                                                    else if (diffTSC >= -2) {
                                                        if (Cs < 1111) {
                                                            if (Rs < 1304) {
                                                                return 4;
                                                            }
                                                            else if (Rs >= 1304) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (Cs >= 1111) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (Cs >= 2624) {
                                                if (Cs < 3132) {
                                                    return 2;
                                                }
                                                else if (Cs >= 3132) {
                                                    return 1;
                                                }
                                            }
                                        }
                                        else if (mcTcor >= 91) {
                                            if (tcor < 65) {
                                                return 4;
                                            }
                                            else if (tcor >= 65) {
                                                if (gchDC < 3) {
                                                    return 2;
                                                }
                                                else if (gchDC >= 3) {
                                                    if (Cs < 904) {
                                                        return 1;
                                                    }
                                                    else if (Cs >= 904) {
                                                        return 2;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (diffTSC >= 2) {
                                        if (gchDC < 4) {
                                            if (Cs < 1067) {
                                                return 1;
                                            }
                                            else if (Cs >= 1067) {
                                                if (Rs < 1365) {
                                                    return 4;
                                                }
                                                else if (Rs >= 1365) {
                                                    return 2;
                                                }
                                            }
                                        }
                                        else if (gchDC >= 4) {
                                            if (AFD < 56) {
                                                return 8;
                                            }
                                            else if (AFD >= 56) {
                                                return 2;
                                            }
                                        }
                                    }
                                }
                            }
                        }
                    }
                    else if (mu_mv_mag_sq >= 28) {
                        if (AFD < 12) {
                            if (refDCval < 99) {
                                if (TSC < 7) {
                                    if (posBalance < 24) {
                                        if (Rs < 279) {
                                            if (mu_mv_mag_sq < 69) {
                                                if (diffMVdiffVal < 55) {
                                                    if (Cs < 3) {
                                                        if (negBalance < 2) {
                                                            return 1;
                                                        }
                                                        else if (negBalance >= 2) {
                                                            if (MVDiff < 91) {
                                                                if (mu_mv_mag_sq < 48) {
                                                                    if (diffMVdiffVal < -2) {
                                                                        return 4;
                                                                    }
                                                                    else if (diffMVdiffVal >= -2) {
                                                                        if (mu_mv_mag_sq < 32) {
                                                                            return 8;
                                                                        }
                                                                        else if (mu_mv_mag_sq >= 32) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                }
                                                                else if (mu_mv_mag_sq >= 48) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (MVDiff >= 91) {
                                                                return 1;
                                                            }
                                                        }
                                                    }
                                                    else if (Cs >= 3) {
                                                        if (MVDiff < 45) {
                                                            if (SC < 457) {
                                                                return 8;
                                                            }
                                                            else if (SC >= 457) {
                                                                return 2;
                                                            }
                                                        }
                                                        else if (MVDiff >= 45) {
                                                            if (Rs < 43) {
                                                                return 2;
                                                            }
                                                            else if (Rs >= 43) {
                                                                return 1;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (diffMVdiffVal >= 55) {
                                                    return 2;
                                                }
                                            }
                                            else if (mu_mv_mag_sq >= 69) {
                                                if (refDCval < 23) {
                                                    return 8;
                                                }
                                                else if (refDCval >= 23) {
                                                    if (MVDiff < 39) {
                                                        return 4;
                                                    }
                                                    else if (MVDiff >= 39) {
                                                        return 2;
                                                    }
                                                }
                                            }
                                        }
                                        else if (Rs >= 279) {
                                            if (mu_mv_mag_sq < 83) {
                                                if (SC < 4054) {
                                                    return 8;
                                                }
                                                else if (SC >= 4054) {
                                                    return 2;
                                                }
                                            }
                                            else if (mu_mv_mag_sq >= 83) {
                                                return 4;
                                            }
                                        }
                                    }
                                    else if (posBalance >= 24) {
                                        if (tcor < 87) {
                                            if (Rs < 652) {
                                                if (Cs < 400) {
                                                    if (negBalance < 50) {
                                                        if (mcTcor < 82) {
                                                            if (ssDCval < 33) {
                                                                if (tcor < 44) {
                                                                    if (posBalance < 27) {
                                                                        return 1;
                                                                    }
                                                                    else if (posBalance >= 27) {
                                                                        if (mu_mv_mag_sq < 34) {
                                                                            if (diffMVdiffVal < 3) {
                                                                                return 8;
                                                                            }
                                                                            else if (diffMVdiffVal >= 3) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                        else if (mu_mv_mag_sq >= 34) {
                                                                            if (AFD < 1) {
                                                                                if (mcTcor < 44) {
                                                                                    return 1;
                                                                                }
                                                                                else if (mcTcor >= 44) {
                                                                                    return 4;
                                                                                }
                                                                            }
                                                                            else if (AFD >= 1) {
                                                                                return 1;
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                                else if (tcor >= 44) {
                                                                    if (gchDC < 1) {
                                                                        return 2;
                                                                    }
                                                                    else if (gchDC >= 1) {
                                                                        if (Cs < 186) {
                                                                            if (diffMVdiffVal < -102) {
                                                                                if (MVDiff < 300) {
                                                                                    return 4;
                                                                                }
                                                                                else if (MVDiff >= 300) {
                                                                                    return 1;
                                                                                }
                                                                            }
                                                                            else if (diffMVdiffVal >= -102) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                        else if (Cs >= 186) {
                                                                            return 2;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                            else if (ssDCval >= 33) {
                                                                if (ssDCval < 38) {
                                                                    if (negBalance < 46) {
                                                                        if (tcor < 74) {
                                                                            if (SC < 681) {
                                                                                return 1;
                                                                            }
                                                                            else if (SC >= 681) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                        else if (tcor >= 74) {
                                                                            if (negBalance < 45) {
                                                                                if (negBalance < 44) {
                                                                                    if (Cs < 286) {
                                                                                        return 8;
                                                                                    }
                                                                                    else if (Cs >= 286) {
                                                                                        return 1;
                                                                                    }
                                                                                }
                                                                                else if (negBalance >= 44) {
                                                                                    return 8;
                                                                                }
                                                                            }
                                                                            else if (negBalance >= 45) {
                                                                                return 1;
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (negBalance >= 46) {
                                                                        if (diffMVdiffVal < -1) {
                                                                            return 4;
                                                                        }
                                                                        else if (diffMVdiffVal >= -1) {
                                                                            if (diffTSC < -1) {
                                                                                return 4;
                                                                            }
                                                                            else if (diffTSC >= -1) {
                                                                                return 8;
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                                else if (ssDCval >= 38) {
                                                                    if (Cs < 386) {
                                                                        return 1;
                                                                    }
                                                                    else if (Cs >= 386) {
                                                                        if (SC < 887) {
                                                                            return 2;
                                                                        }
                                                                        else if (SC >= 887) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (mcTcor >= 82) {
                                                            if (negBalance < 42) {
                                                                if (Rs < 430) {
                                                                    if (MVDiff < 23) {
                                                                        if (negBalance < 41) {
                                                                            if (MVDiff < 15) {
                                                                                if (ssDCval < 37) {
                                                                                    return 4;
                                                                                }
                                                                                else if (ssDCval >= 37) {
                                                                                    return 8;
                                                                                }
                                                                            }
                                                                            else if (MVDiff >= 15) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                        else if (negBalance >= 41) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                    else if (MVDiff >= 23) {
                                                                        if (negBalance < 10) {
                                                                            return 4;
                                                                        }
                                                                        else if (negBalance >= 10) {
                                                                            if (diffAFD < -1) {
                                                                                if (gchDC < 1) {
                                                                                    return 8;
                                                                                }
                                                                                else if (gchDC >= 1) {
                                                                                    return 1;
                                                                                }
                                                                            }
                                                                            else if (diffAFD >= -1) {
                                                                                if (Cs < 203) {
                                                                                    return 2;
                                                                                }
                                                                                else if (Cs >= 203) {
                                                                                    return 4;
                                                                                }
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                                else if (Rs >= 430) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (negBalance >= 42) {
                                                                if (SC < 810) {
                                                                    if (ssDCval < 38) {
                                                                        if (RsDiff < 1) {
                                                                            return 4;
                                                                        }
                                                                        else if (RsDiff >= 1) {
                                                                            if (SC < 757) {
                                                                                return 1;
                                                                            }
                                                                            else if (SC >= 757) {
                                                                                if (mcTcor < 84) {
                                                                                    return 4;
                                                                                }
                                                                                else if (mcTcor >= 84) {
                                                                                    return 1;
                                                                                }
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (ssDCval >= 38) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (SC >= 810) {
                                                                    if (Cs < 394) {
                                                                        if (MVDiff < 14) {
                                                                            return 4;
                                                                        }
                                                                        else if (MVDiff >= 14) {
                                                                            return 2;
                                                                        }
                                                                    }
                                                                    else if (Cs >= 394) {
                                                                        return 4;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (negBalance >= 50) {
                                                        return 2;
                                                    }
                                                }
                                                else if (Cs >= 400) {
                                                    if (MVDiff < 39) {
                                                        if (MVDiff < 22) {
                                                            if (SC < 909) {
                                                                if (Rs < 364) {
                                                                    return 2;
                                                                }
                                                                else if (Rs >= 364) {
                                                                    if (Rs < 468) {
                                                                        return 4;
                                                                    }
                                                                    else if (Rs >= 468) {
                                                                        return 2;
                                                                    }
                                                                }
                                                            }
                                                            else if (SC >= 909) {
                                                                if (Cs < 1020) {
                                                                    return 4;
                                                                }
                                                                else if (Cs >= 1020) {
                                                                    return 2;
                                                                }
                                                            }
                                                        }
                                                        else if (MVDiff >= 22) {
                                                            if (RsDiff < 4) {
                                                                return 2;
                                                            }
                                                            else if (RsDiff >= 4) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                    else if (MVDiff >= 39) {
                                                        if (mu_mv_mag_sq < 39) {
                                                            if (SC < 1092) {
                                                                return 2;
                                                            }
                                                            else if (SC >= 1092) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (mu_mv_mag_sq >= 39) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (Rs >= 652) {
                                                if (SC < 1511) {
                                                    if (mu_mv_mag_sq < 45) {
                                                        if (ssDCval < 42) {
                                                            if (Cs < 478) {
                                                                return 4;
                                                            }
                                                            else if (Cs >= 478) {
                                                                if (Cs < 659) {
                                                                    if (RsDiff < 2) {
                                                                        if (mu_mv_mag_sq < 40) {
                                                                            if (posBalance < 59) {
                                                                                return 4;
                                                                            }
                                                                            else if (posBalance >= 59) {
                                                                                return 1;
                                                                            }
                                                                        }
                                                                        else if (mu_mv_mag_sq >= 40) {
                                                                            if (mcTcor < 81) {
                                                                                if (MVDiff < 32) {
                                                                                    return 4;
                                                                                }
                                                                                else if (MVDiff >= 32) {
                                                                                    return 1;
                                                                                }
                                                                            }
                                                                            else if (mcTcor >= 81) {
                                                                                return 1;
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (RsDiff >= 2) {
                                                                        if (TSC < 5) {
                                                                            return 8;
                                                                        }
                                                                        else if (TSC >= 5) {
                                                                            if (MVDiff < 32) {
                                                                                if (mcTcor < 76) {
                                                                                    return 4;
                                                                                }
                                                                                else if (mcTcor >= 76) {
                                                                                    return 1;
                                                                                }
                                                                            }
                                                                            else if (MVDiff >= 32) {
                                                                                return 1;
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                                else if (Cs >= 659) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                        else if (ssDCval >= 42) {
                                                            if (TSC < 6) {
                                                                return 8;
                                                            }
                                                            else if (TSC >= 6) {
                                                                return 1;
                                                            }
                                                        }
                                                    }
                                                    else if (mu_mv_mag_sq >= 45) {
                                                        if (Rs < 715) {
                                                            if (SC < 1299) {
                                                                return 8;
                                                            }
                                                            else if (SC >= 1299) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (Rs >= 715) {
                                                            if (Cs < 576) {
                                                                if (Rs < 818) {
                                                                    if (AFD < 6) {
                                                                        if (negBalance < 46) {
                                                                            return 4;
                                                                        }
                                                                        else if (negBalance >= 46) {
                                                                            if (Rs < 761) {
                                                                                return 4;
                                                                            }
                                                                            else if (Rs >= 761) {
                                                                                return 1;
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (AFD >= 6) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (Rs >= 818) {
                                                                    return 1;
                                                                }
                                                            }
                                                            else if (Cs >= 576) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (SC >= 1511) {
                                                    if (MVDiff < 19) {
                                                        if (tcor < 79) {
                                                            if (diffAFD < 2) {
                                                                return 4;
                                                            }
                                                            else if (diffAFD >= 2) {
                                                                if (Rs < 1565) {
                                                                    return 2;
                                                                }
                                                                else if (Rs >= 1565) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                        else if (tcor >= 79) {
                                                            if (Rs < 1180) {
                                                                if (Rs < 895) {
                                                                    return 2;
                                                                }
                                                                else if (Rs >= 895) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (Rs >= 1180) {
                                                                if (diffRsCsdiff < -5) {
                                                                    if (Rs < 3074) {
                                                                        return 4;
                                                                    }
                                                                    else if (Rs >= 3074) {
                                                                        return 8;
                                                                    }
                                                                }
                                                                else if (diffRsCsdiff >= -5) {
                                                                    if (Cs < 808) {
                                                                        if (SC < 2565) {
                                                                            return 4;
                                                                        }
                                                                        else if (SC >= 2565) {
                                                                            return 2;
                                                                        }
                                                                    }
                                                                    else if (Cs >= 808) {
                                                                        if (diffMVdiffVal < 1) {
                                                                            if (posBalance < 51) {
                                                                                return 1;
                                                                            }
                                                                            else if (posBalance >= 51) {
                                                                                if (Cs < 955) {
                                                                                    return 1;
                                                                                }
                                                                                else if (Cs >= 955) {
                                                                                    return 8;
                                                                                }
                                                                            }
                                                                        }
                                                                        else if (diffMVdiffVal >= 1) {
                                                                            if (mu_mv_mag_sq < 44) {
                                                                                return 1;
                                                                            }
                                                                            else if (mu_mv_mag_sq >= 44) {
                                                                                if (MVDiff < 6) {
                                                                                    return 1;
                                                                                }
                                                                                else if (MVDiff >= 6) {
                                                                                    if (SC < 2373) {
                                                                                        return 1;
                                                                                    }
                                                                                    else if (SC >= 2373) {
                                                                                        return 4;
                                                                                    }
                                                                                }
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (MVDiff >= 19) {
                                                        if (TSCindex < 6) {
                                                            if (mu_mv_mag_sq < 46) {
                                                                return 2;
                                                            }
                                                            else if (mu_mv_mag_sq >= 46) {
                                                                if (Rs < 1744) {
                                                                    if (Cs < 995) {
                                                                        return 1;
                                                                    }
                                                                    else if (Cs >= 995) {
                                                                        return 2;
                                                                    }
                                                                }
                                                                else if (Rs >= 1744) {
                                                                    return 1;
                                                                }
                                                            }
                                                        }
                                                        else if (TSCindex >= 6) {
                                                            if (diffTSC < -1) {
                                                                if (mu_mv_mag_sq < 47) {
                                                                    if (Cs < 976) {
                                                                        return 1;
                                                                    }
                                                                    else if (Cs >= 976) {
                                                                        return 2;
                                                                    }
                                                                }
                                                                else if (mu_mv_mag_sq >= 47) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (diffTSC >= -1) {
                                                                return 1;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (tcor >= 87) {
                                            if (Cs < 2594) {
                                                if (Rs < 940) {
                                                    if (Cs < 212) {
                                                        if (diffMVdiffVal < 2) {
                                                            if (gchDC < 3) {
                                                                return 1;
                                                            }
                                                            else if (gchDC >= 3) {
                                                                if (SC < 373) {
                                                                    return 8;
                                                                }
                                                                else if (SC >= 373) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                        else if (diffMVdiffVal >= 2) {
                                                            return 2;
                                                        }
                                                    }
                                                    else if (Cs >= 212) {
                                                        if (SC < 953) {
                                                            if (RsDiff < 2) {
                                                                if (gchDC < 4) {
                                                                    return 4;
                                                                }
                                                                else if (gchDC >= 4) {
                                                                    if (SC < 750) {
                                                                        return 8;
                                                                    }
                                                                    else if (SC >= 750) {
                                                                        return 4;
                                                                    }
                                                                }
                                                            }
                                                            else if (RsDiff >= 2) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (SC >= 953) {
                                                            if (MVDiff < 8) {
                                                                if (mcTcor < 92) {
                                                                    return 8;
                                                                }
                                                                else if (mcTcor >= 92) {
                                                                    if (TSC < 3) {
                                                                        return 8;
                                                                    }
                                                                    else if (TSC >= 3) {
                                                                        if (ssDCval < 54) {
                                                                            if (posBalance < 57) {
                                                                                return 4;
                                                                            }
                                                                            else if (posBalance >= 57) {
                                                                                return 8;
                                                                            }
                                                                        }
                                                                        else if (ssDCval >= 54) {
                                                                            if (Rs < 679) {
                                                                                return 4;
                                                                            }
                                                                            else if (Rs >= 679) {
                                                                                return 8;
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                            else if (MVDiff >= 8) {
                                                                if (RsCsDiff < 9) {
                                                                    if (mu_mv_mag_sq < 31) {
                                                                        if (Cs < 634) {
                                                                            return 1;
                                                                        }
                                                                        else if (Cs >= 634) {
                                                                            return 2;
                                                                        }
                                                                    }
                                                                    else if (mu_mv_mag_sq >= 31) {
                                                                        if (AFD < 8) {
                                                                            if (posBalance < 53) {
                                                                                return 4;
                                                                            }
                                                                            else if (posBalance >= 53) {
                                                                                if (RsCsDiff < 2) {
                                                                                    return 4;
                                                                                }
                                                                                else if (RsCsDiff >= 2) {
                                                                                    if (Cs < 670) {
                                                                                        if (Rs < 725) {
                                                                                            return 8;
                                                                                        }
                                                                                        else if (Rs >= 725) {
                                                                                            return 4;
                                                                                        }
                                                                                    }
                                                                                    else if (Cs >= 670) {
                                                                                        return 8;
                                                                                    }
                                                                                }
                                                                            }
                                                                        }
                                                                        else if (AFD >= 8) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                }
                                                                else if (RsCsDiff >= 9) {
                                                                    if (Cs < 618) {
                                                                        return 2;
                                                                    }
                                                                    else if (Cs >= 618) {
                                                                        if (Cs < 1058) {
                                                                            return 4;
                                                                        }
                                                                        else if (Cs >= 1058) {
                                                                            return 2;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (Rs >= 940) {
                                                    if (Rs < 1800) {
                                                        if (ssDCval < 42) {
                                                            if (TSC < 2) {
                                                                return 4;
                                                            }
                                                            else if (TSC >= 2) {
                                                                if (mu_mv_mag_sq < 56) {
                                                                    if (RsDiff < 2) {
                                                                        return 2;
                                                                    }
                                                                    else if (RsDiff >= 2) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (mu_mv_mag_sq >= 56) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                        else if (ssDCval >= 42) {
                                                            if (mu_mv_mag_sq < 83) {
                                                                if (Rs < 1191) {
                                                                    if (negBalance < 42) {
                                                                        if (ssDCval < 46) {
                                                                            if (Cs < 723) {
                                                                                return 1;
                                                                            }
                                                                            else if (Cs >= 723) {
                                                                                return 8;
                                                                            }
                                                                        }
                                                                        else if (ssDCval >= 46) {
                                                                            if (mcTcor < 94) {
                                                                                if (ssDCval < 49) {
                                                                                    return 1;
                                                                                }
                                                                                else if (ssDCval >= 49) {
                                                                                    if (SC < 1901) {
                                                                                        if (SC < 1785) {
                                                                                            return 4;
                                                                                        }
                                                                                        else if (SC >= 1785) {
                                                                                            return 1;
                                                                                        }
                                                                                    }
                                                                                    else if (SC >= 1901) {
                                                                                        return 4;
                                                                                    }
                                                                                }
                                                                            }
                                                                            else if (mcTcor >= 94) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (negBalance >= 42) {
                                                                        if (mcTcor < 93) {
                                                                            if (Rs < 950) {
                                                                                return 1;
                                                                            }
                                                                            else if (Rs >= 950) {
                                                                                if (diffAFD < 1) {
                                                                                    return 8;
                                                                                }
                                                                                else if (diffAFD >= 1) {
                                                                                    return 4;
                                                                                }
                                                                            }
                                                                        }
                                                                        else if (mcTcor >= 93) {
                                                                            if (negBalance < 48) {
                                                                                if (SC < 1950) {
                                                                                    return 1;
                                                                                }
                                                                                else if (SC >= 1950) {
                                                                                    return 4;
                                                                                }
                                                                            }
                                                                            else if (negBalance >= 48) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                                else if (Rs >= 1191) {
                                                                    if (Rs < 1534) {
                                                                        if (refDCval < 54) {
                                                                            if (ssDCval < 49) {
                                                                                if (AFD < 4) {
                                                                                    return 1;
                                                                                }
                                                                                else if (AFD >= 4) {
                                                                                    return 8;
                                                                                }
                                                                            }
                                                                            else if (ssDCval >= 49) {
                                                                                if (Rs < 1355) {
                                                                                    if (mu_mv_mag_sq < 64) {
                                                                                        return 1;
                                                                                    }
                                                                                    else if (mu_mv_mag_sq >= 64) {
                                                                                        if (tcor < 90) {
                                                                                            if (Cs < 995) {
                                                                                                if (ssDCval < 53) {
                                                                                                    return 1;
                                                                                                }
                                                                                                else if (ssDCval >= 53) {
                                                                                                    return 4;
                                                                                                }
                                                                                            }
                                                                                            else if (Cs >= 995) {
                                                                                                return 4;
                                                                                            }
                                                                                        }
                                                                                        else if (tcor >= 90) {
                                                                                            return 1;
                                                                                        }
                                                                                    }
                                                                                }
                                                                                else if (Rs >= 1355) {
                                                                                    if (SC < 2339) {
                                                                                        if (Cs < 900) {
                                                                                            return 1;
                                                                                        }
                                                                                        else if (Cs >= 900) {
                                                                                            return 4;
                                                                                        }
                                                                                    }
                                                                                    else if (SC >= 2339) {
                                                                                        if (posBalance < 51) {
                                                                                            return 1;
                                                                                        }
                                                                                        else if (posBalance >= 51) {
                                                                                            if (Cs < 988) {
                                                                                                if (RsDiff < 3) {
                                                                                                    return 1;
                                                                                                }
                                                                                                else if (RsDiff >= 3) {
                                                                                                    return 4;
                                                                                                }
                                                                                            }
                                                                                            else if (Cs >= 988) {
                                                                                                return 1;
                                                                                            }
                                                                                        }
                                                                                    }
                                                                                }
                                                                            }
                                                                        }
                                                                        else if (refDCval >= 54) {
                                                                            if (negBalance < 40) {
                                                                                if (mcTcor < 95) {
                                                                                    if (diffAFD < -2) {
                                                                                        return 1;
                                                                                    }
                                                                                    else if (diffAFD >= -2) {
                                                                                        return 8;
                                                                                    }
                                                                                }
                                                                                else if (mcTcor >= 95) {
                                                                                    if (negBalance < 29) {
                                                                                        return 4;
                                                                                    }
                                                                                    else if (negBalance >= 29) {
                                                                                        if (SC < 2498) {
                                                                                            return 1;
                                                                                        }
                                                                                        else if (SC >= 2498) {
                                                                                            return 4;
                                                                                        }
                                                                                    }
                                                                                }
                                                                            }
                                                                            else if (negBalance >= 40) {
                                                                                if (posBalance < 49) {
                                                                                    return 4;
                                                                                }
                                                                                else if (posBalance >= 49) {
                                                                                    if (Rs < 1418) {
                                                                                        return 1;
                                                                                    }
                                                                                    else if (Rs >= 1418) {
                                                                                        return 8;
                                                                                    }
                                                                                }
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (Rs >= 1534) {
                                                                        if (negBalance < 34) {
                                                                            if (Rs < 1764) {
                                                                                return 1;
                                                                            }
                                                                            else if (Rs >= 1764) {
                                                                                return 8;
                                                                            }
                                                                        }
                                                                        else if (negBalance >= 34) {
                                                                            if (posBalance < 58) {
                                                                                return 4;
                                                                            }
                                                                            else if (posBalance >= 58) {
                                                                                return 1;
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                            else if (mu_mv_mag_sq >= 83) {
                                                                if (RsCsDiff < 16) {
                                                                    if (ssDCval < 49) {
                                                                        return 1;
                                                                    }
                                                                    else if (ssDCval >= 49) {
                                                                        return 8;
                                                                    }
                                                                }
                                                                else if (RsCsDiff >= 16) {
                                                                    if (SC < 3139) {
                                                                        return 2;
                                                                    }
                                                                    else if (SC >= 3139) {
                                                                        return 1;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (Rs >= 1800) {
                                                        if (Cs < 1538) {
                                                            if (Cs < 893) {
                                                                if (RsDiff < 2) {
                                                                    return 4;
                                                                }
                                                                else if (RsDiff >= 2) {
                                                                    return 2;
                                                                }
                                                            }
                                                            else if (Cs >= 893) {
                                                                if (Cs < 1220) {
                                                                    if (SC < 4670) {
                                                                        if (TSC < 6) {
                                                                            return 4;
                                                                        }
                                                                        else if (TSC >= 6) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                    else if (SC >= 4670) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (Cs >= 1220) {
                                                                    if (mu_mv_mag_sq < 61) {
                                                                        if (MVDiff < 19) {
                                                                            if (TSC < 5) {
                                                                                if (Cs < 1530) {
                                                                                    return 1;
                                                                                }
                                                                                else if (Cs >= 1530) {
                                                                                    return 4;
                                                                                }
                                                                            }
                                                                            else if (TSC >= 5) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                        else if (MVDiff >= 19) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                    else if (mu_mv_mag_sq >= 61) {
                                                                        if (mu_mv_mag_sq < 85) {
                                                                            return 4;
                                                                        }
                                                                        else if (mu_mv_mag_sq >= 85) {
                                                                            if (Cs < 1503) {
                                                                                return 8;
                                                                            }
                                                                            else if (Cs >= 1503) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (Cs >= 1538) {
                                                            if (MVDiff < 1) {
                                                                if (SC < 3715) {
                                                                    return 2;
                                                                }
                                                                else if (SC >= 3715) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (MVDiff >= 1) {
                                                                if (MVDiff < 30) {
                                                                    if (Cs < 1554) {
                                                                        if (SC < 3374) {
                                                                            return 1;
                                                                        }
                                                                        else if (SC >= 3374) {
                                                                            if (tcor < 92) {
                                                                                return 4;
                                                                            }
                                                                            else if (tcor >= 92) {
                                                                                if (Rs < 1874) {
                                                                                    return 4;
                                                                                }
                                                                                else if (Rs >= 1874) {
                                                                                    return 1;
                                                                                }
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (Cs >= 1554) {
                                                                        if (posBalance < 59) {
                                                                            if (diffRsCsdiff < -16) {
                                                                                return 4;
                                                                            }
                                                                            else if (diffRsCsdiff >= -16) {
                                                                                if (SC < 3969) {
                                                                                    if (mu_mv_mag_sq < 74) {
                                                                                        if (TSC < 5) {
                                                                                            return 1;
                                                                                        }
                                                                                        else if (TSC >= 5) {
                                                                                            if (Rs < 1946) {
                                                                                                return 1;
                                                                                            }
                                                                                            else if (Rs >= 1946) {
                                                                                                return 4;
                                                                                            }
                                                                                        }
                                                                                    }
                                                                                    else if (mu_mv_mag_sq >= 74) {
                                                                                        if (Rs < 1922) {
                                                                                            return 1;
                                                                                        }
                                                                                        else if (Rs >= 1922) {
                                                                                            return 4;
                                                                                        }
                                                                                    }
                                                                                }
                                                                                else if (SC >= 3969) {
                                                                                    return 1;
                                                                                }
                                                                            }
                                                                        }
                                                                        else if (posBalance >= 59) {
                                                                            if (Rs < 2148) {
                                                                                return 4;
                                                                            }
                                                                            else if (Rs >= 2148) {
                                                                                return 8;
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                                else if (MVDiff >= 30) {
                                                                    return 1;
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                            else if (Cs >= 2594) {
                                                if (refDCval < 54) {
                                                    if (Rs < 3467) {
                                                        if (MVDiff < 22) {
                                                            if (ssDCval < 53) {
                                                                if (Cs < 3073) {
                                                                    return 8;
                                                                }
                                                                else if (Cs >= 3073) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (ssDCval >= 53) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (MVDiff >= 22) {
                                                            if (tcor < 94) {
                                                                return 8;
                                                            }
                                                            else if (tcor >= 94) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                    else if (Rs >= 3467) {
                                                        if (mu_mv_mag_sq < 33) {
                                                            return 8;
                                                        }
                                                        else if (mu_mv_mag_sq >= 33) {
                                                            if (MVDiff < 5) {
                                                                if (Cs < 3357) {
                                                                    return 8;
                                                                }
                                                                else if (Cs >= 3357) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (MVDiff >= 5) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (refDCval >= 54) {
                                                    if (MVDiff < 27) {
                                                        if (refDCval < 55) {
                                                            if (SC < 7095) {
                                                                return 8;
                                                            }
                                                            else if (SC >= 7095) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (refDCval >= 55) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (MVDiff >= 27) {
                                                        return 2;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (TSC >= 7) {
                                    if (mcTcor < 95) {
                                        if (negBalance < 49) {
                                            if (posBalance < 60) {
                                                if (tcor < 83) {
                                                    if (Rs < 3002) {
                                                        return 8;
                                                    }
                                                    else if (Rs >= 3002) {
                                                        return 4;
                                                    }
                                                }
                                                else if (tcor >= 83) {
                                                    if (refDCval < 92) {
                                                        if (MVDiff < 10) {
                                                            if (posBalance < 58) {
                                                                return 8;
                                                            }
                                                            else if (posBalance >= 58) {
                                                                if (tcor < 92) {
                                                                    return 8;
                                                                }
                                                                else if (tcor >= 92) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                        else if (MVDiff >= 10) {
                                                            if (ssDCval < 88) {
                                                                if (ssDCval < 82) {
                                                                    if (SC < 3592) {
                                                                        return 1;
                                                                    }
                                                                    else if (SC >= 3592) {
                                                                        return 8;
                                                                    }
                                                                }
                                                                else if (ssDCval >= 82) {
                                                                    return 1;
                                                                }
                                                            }
                                                            else if (ssDCval >= 88) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                    else if (refDCval >= 92) {
                                                        if (SC < 4223) {
                                                            return 8;
                                                        }
                                                        else if (SC >= 4223) {
                                                            return 1;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (posBalance >= 60) {
                                                if (SC < 4937) {
                                                    if (Cs < 1646) {
                                                        if (diffRsCsdiff < -8) {
                                                            return 8;
                                                        }
                                                        else if (diffRsCsdiff >= -8) {
                                                            if (negBalance < 28) {
                                                                return 4;
                                                            }
                                                            else if (negBalance >= 28) {
                                                                return 1;
                                                            }
                                                        }
                                                    }
                                                    else if (Cs >= 1646) {
                                                        return 4;
                                                    }
                                                }
                                                else if (SC >= 4937) {
                                                    if (posBalance < 63) {
                                                        return 2;
                                                    }
                                                    else if (posBalance >= 63) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                        }
                                        else if (negBalance >= 49) {
                                            if (ssDCval < 86) {
                                                if (RsCsDiff < 56) {
                                                    if (MVDiff < 15) {
                                                        if (MVDiff < 3) {
                                                            return 2;
                                                        }
                                                        else if (MVDiff >= 3) {
                                                            if (RsDiff < 3) {
                                                                return 2;
                                                            }
                                                            else if (RsDiff >= 3) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                    else if (MVDiff >= 15) {
                                                        if (Rs < 2271) {
                                                            if (Rs < 1225) {
                                                                return 1;
                                                            }
                                                            else if (Rs >= 1225) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (Rs >= 2271) {
                                                            if (diffMVdiffVal < 6) {
                                                                if (mu_mv_mag_sq < 44) {
                                                                    return 2;
                                                                }
                                                                else if (mu_mv_mag_sq >= 44) {
                                                                    if (Rs < 3570) {
                                                                        return 8;
                                                                    }
                                                                    else if (Rs >= 3570) {
                                                                        return 2;
                                                                    }
                                                                }
                                                            }
                                                            else if (diffMVdiffVal >= 6) {
                                                                if (Rs < 3522) {
                                                                    return 8;
                                                                }
                                                                else if (Rs >= 3522) {
                                                                    if (posBalance < 58) {
                                                                        return 2;
                                                                    }
                                                                    else if (posBalance >= 58) {
                                                                        return 8;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (RsCsDiff >= 56) {
                                                    if (SC < 4161) {
                                                        return 1;
                                                    }
                                                    else if (SC >= 4161) {
                                                        if (diffMVdiffVal < -3) {
                                                            if (Cs < 2462) {
                                                                return 8;
                                                            }
                                                            else if (Cs >= 2462) {
                                                                return 2;
                                                            }
                                                        }
                                                        else if (diffMVdiffVal >= -3) {
                                                            if (Cs < 2152) {
                                                                return 8;
                                                            }
                                                            else if (Cs >= 2152) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                            else if (ssDCval >= 86) {
                                                if (RsDiff < 6) {
                                                    if (SC < 2406) {
                                                        return 4;
                                                    }
                                                    else if (SC >= 2406) {
                                                        if (Rs < 1754) {
                                                            return 8;
                                                        }
                                                        else if (Rs >= 1754) {
                                                            if (tcor < 85) {
                                                                if (mu_mv_mag_sq < 63) {
                                                                    if (MVDiff < 8) {
                                                                        if (diffRsCsdiff < -1) {
                                                                            return 2;
                                                                        }
                                                                        else if (diffRsCsdiff >= -1) {
                                                                            if (SC < 3720) {
                                                                                if (Rs < 1858) {
                                                                                    return 2;
                                                                                }
                                                                                else if (Rs >= 1858) {
                                                                                    return 8;
                                                                                }
                                                                            }
                                                                            else if (SC >= 3720) {
                                                                                return 2;
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (MVDiff >= 8) {
                                                                        if (diffRsCsdiff < -10) {
                                                                            return 2;
                                                                        }
                                                                        else if (diffRsCsdiff >= -10) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                }
                                                                else if (mu_mv_mag_sq >= 63) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (tcor >= 85) {
                                                                return 2;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (RsDiff >= 6) {
                                                    if (ssDCval < 91) {
                                                        if (Rs < 3209) {
                                                            if (mcTcor < 93) {
                                                                if (diffAFD < -1) {
                                                                    return 2;
                                                                }
                                                                else if (diffAFD >= -1) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (mcTcor >= 93) {
                                                                return 2;
                                                            }
                                                        }
                                                        else if (Rs >= 3209) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (ssDCval >= 91) {
                                                        return 2;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (mcTcor >= 95) {
                                        if (Rs < 2430) {
                                            if (negBalance < 42) {
                                                if (Rs < 2366) {
                                                    if (mu_mv_mag_sq < 54) {
                                                        if (SC < 3574) {
                                                            return 4;
                                                        }
                                                        else if (SC >= 3574) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (mu_mv_mag_sq >= 54) {
                                                        return 8;
                                                    }
                                                }
                                                else if (Rs >= 2366) {
                                                    return 1;
                                                }
                                            }
                                            else if (negBalance >= 42) {
                                                if (RsDiff < 7) {
                                                    if (ssDCval < 84) {
                                                        if (Cs < 1135) {
                                                            return 2;
                                                        }
                                                        else if (Cs >= 1135) {
                                                            if (RsCsDiff < 47) {
                                                                if (AFD < 8) {
                                                                    return 2;
                                                                }
                                                                else if (AFD >= 8) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (RsCsDiff >= 47) {
                                                                if (Cs < 1900) {
                                                                    return 1;
                                                                }
                                                                else if (Cs >= 1900) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (ssDCval >= 84) {
                                                        return 2;
                                                    }
                                                }
                                                else if (RsDiff >= 7) {
                                                    return 4;
                                                }
                                            }
                                        }
                                        else if (Rs >= 2430) {
                                            if (Rs < 3143) {
                                                if (ssDCval < 86) {
                                                    if (posBalance < 58) {
                                                        return 4;
                                                    }
                                                    else if (posBalance >= 58) {
                                                        if (ssDCval < 85) {
                                                            return 1;
                                                        }
                                                        else if (ssDCval >= 85) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                                else if (ssDCval >= 86) {
                                                    return 1;
                                                }
                                            }
                                            else if (Rs >= 3143) {
                                                if (MVDiff < 39) {
                                                    return 2;
                                                }
                                                else if (MVDiff >= 39) {
                                                    return 8;
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                            else if (refDCval >= 99) {
                                if (ssDCval < 119) {
                                    if (RsCsDiff < 69) {
                                        if (mu_mv_mag_sq < 49) {
                                            if (SC < 2329) {
                                                if (Cs < 17) {
                                                    return 1;
                                                }
                                                else if (Cs >= 17) {
                                                    if (SC < 766) {
                                                        if (Rs < 327) {
                                                            return 4;
                                                        }
                                                        else if (Rs >= 327) {
                                                            if (MVDiff < 15) {
                                                                return 4;
                                                            }
                                                            else if (MVDiff >= 15) {
                                                                if (posBalance < 62) {
                                                                    return 8;
                                                                }
                                                                else if (posBalance >= 62) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (SC >= 766) {
                                                        if (Rs < 1215) {
                                                            return 8;
                                                        }
                                                        else if (Rs >= 1215) {
                                                            return 2;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (SC >= 2329) {
                                                if (AFD < 7) {
                                                    return 4;
                                                }
                                                else if (AFD >= 7) {
                                                    if (Cs < 1758) {
                                                        if (negBalance < 38) {
                                                            if (SC < 5425) {
                                                                return 8;
                                                            }
                                                            else if (SC >= 5425) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (negBalance >= 38) {
                                                            if (Rs < 1202) {
                                                                return 8;
                                                            }
                                                            else if (Rs >= 1202) {
                                                                if (TSC < 5) {
                                                                    return 8;
                                                                }
                                                                else if (TSC >= 5) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (Cs >= 1758) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                        }
                                        else if (mu_mv_mag_sq >= 49) {
                                            if (ssDCval < 104) {
                                                if (Rs < 2482) {
                                                    if (tcor < 100) {
                                                        if (tcor < 96) {
                                                            if (mcTcor < 92) {
                                                                return 4;
                                                            }
                                                            else if (mcTcor >= 92) {
                                                                return 2;
                                                            }
                                                        }
                                                        else if (tcor >= 96) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (tcor >= 100) {
                                                        return 8;
                                                    }
                                                }
                                                else if (Rs >= 2482) {
                                                    return 8;
                                                }
                                            }
                                            else if (ssDCval >= 104) {
                                                if (SC < 3899) {
                                                    if (gchDC < 5) {
                                                        if (diffAFD < -26) {
                                                            return 2;
                                                        }
                                                        else if (diffAFD >= -26) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (gchDC >= 5) {
                                                        return 4;
                                                    }
                                                }
                                                else if (SC >= 3899) {
                                                    if (Cs < 1536) {
                                                        return 4;
                                                    }
                                                    else if (Cs >= 1536) {
                                                        if (SC < 4418) {
                                                            return 4;
                                                        }
                                                        else if (SC >= 4418) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (RsCsDiff >= 69) {
                                        if (mcTcor < 95) {
                                            if (RsDiff < 9) {
                                                if (negBalance < 51) {
                                                    if (tcor < 88) {
                                                        if (SC < 6989) {
                                                            if (Cs < 1900) {
                                                                if (RsDiff < 8) {
                                                                    return 4;
                                                                }
                                                                else if (RsDiff >= 8) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (Cs >= 1900) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (SC >= 6989) {
                                                            if (ssDCval < 108) {
                                                                return 8;
                                                            }
                                                            else if (ssDCval >= 108) {
                                                                if (refDCval < 109) {
                                                                    return 1;
                                                                }
                                                                else if (refDCval >= 109) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (tcor >= 88) {
                                                        if (ssDCval < 110) {
                                                            if (negBalance < 35) {
                                                                return 4;
                                                            }
                                                            else if (negBalance >= 35) {
                                                                if (MVDiff < 11) {
                                                                    if (diffRsCsdiff < 18) {
                                                                        return 8;
                                                                    }
                                                                    else if (diffRsCsdiff >= 18) {
                                                                        if (Rs < 4168) {
                                                                            return 4;
                                                                        }
                                                                        else if (Rs >= 4168) {
                                                                            if (SC < 7254) {
                                                                                return 8;
                                                                            }
                                                                            else if (SC >= 7254) {
                                                                                return 2;
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                                else if (MVDiff >= 11) {
                                                                    if (gchDC < 2) {
                                                                        return 4;
                                                                    }
                                                                    else if (gchDC >= 2) {
                                                                        return 2;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (ssDCval >= 110) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                                else if (negBalance >= 51) {
                                                    if (Rs < 2496) {
                                                        if (tcor < 82) {
                                                            return 4;
                                                        }
                                                        else if (tcor >= 82) {
                                                            if (Rs < 2267) {
                                                                return 8;
                                                            }
                                                            else if (Rs >= 2267) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                    else if (Rs >= 2496) {
                                                        if (mcTcor < 87) {
                                                            return 8;
                                                        }
                                                        else if (mcTcor >= 87) {
                                                            if (ssDCval < 118) {
                                                                if (MVDiff < 6) {
                                                                    return 8;
                                                                }
                                                                else if (MVDiff >= 6) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (ssDCval >= 118) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                            else if (RsDiff >= 9) {
                                                if (Cs < 2589) {
                                                    if (mu_mv_mag_sq < 61) {
                                                        if (ssDCval < 108) {
                                                            return 8;
                                                        }
                                                        else if (ssDCval >= 108) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (mu_mv_mag_sq >= 61) {
                                                        if (ssDCval < 116) {
                                                            if (posBalance < 60) {
                                                                return 4;
                                                            }
                                                            else if (posBalance >= 60) {
                                                                if (Cs < 2413) {
                                                                    return 4;
                                                                }
                                                                else if (Cs >= 2413) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                        else if (ssDCval >= 116) {
                                                            if (Rs < 2344) {
                                                                return 4;
                                                            }
                                                            else if (Rs >= 2344) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (Cs >= 2589) {
                                                    if (RsCsDiff < 178) {
                                                        return 2;
                                                    }
                                                    else if (RsCsDiff >= 178) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                        }
                                        else if (mcTcor >= 95) {
                                            if (mu_mv_mag_sq < 55) {
                                                if (ssDCval < 107) {
                                                    return 8;
                                                }
                                                else if (ssDCval >= 107) {
                                                    return 4;
                                                }
                                            }
                                            else if (mu_mv_mag_sq >= 55) {
                                                if (Cs < 2070) {
                                                    if (CsDiff < 5) {
                                                        return 8;
                                                    }
                                                    else if (CsDiff >= 5) {
                                                        if (posBalance < 37) {
                                                            return 2;
                                                        }
                                                        else if (posBalance >= 37) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                                else if (Cs >= 2070) {
                                                    if (Rs < 4246) {
                                                        if (refDCval < 105) {
                                                            if (diffMVdiffVal < 3) {
                                                                return 2;
                                                            }
                                                            else if (diffMVdiffVal >= 3) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (refDCval >= 105) {
                                                            return 2;
                                                        }
                                                    }
                                                    else if (Rs >= 4246) {
                                                        if (RsDiff < 11) {
                                                            return 4;
                                                        }
                                                        else if (RsDiff >= 11) {
                                                            return 2;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (ssDCval >= 119) {
                                    if (RsCsDiff < 2) {
                                        if (tcor < 97) {
                                            if (ssDCval < 139) {
                                                return 2;
                                            }
                                            else if (ssDCval >= 139) {
                                                return 4;
                                            }
                                        }
                                        else if (tcor >= 97) {
                                            if (MVDiff < 5) {
                                                if (Rs < 224) {
                                                    return 8;
                                                }
                                                else if (Rs >= 224) {
                                                    return 2;
                                                }
                                            }
                                            else if (MVDiff >= 5) {
                                                return 2;
                                            }
                                        }
                                    }
                                    else if (RsCsDiff >= 2) {
                                        if (SC < 7219) {
                                            if (tcor < 82) {
                                                if (ssDCval < 126) {
                                                    return 4;
                                                }
                                                else if (ssDCval >= 126) {
                                                    if (ssDCval < 139) {
                                                        return 1;
                                                    }
                                                    else if (ssDCval >= 139) {
                                                        if (Cs < 709) {
                                                            return 4;
                                                        }
                                                        else if (Cs >= 709) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (tcor >= 82) {
                                                if (Rs < 3540) {
                                                    if (Cs < 830) {
                                                        if (Cs < 352) {
                                                            if (Cs < 178) {
                                                                return 1;
                                                            }
                                                            else if (Cs >= 178) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (Cs >= 352) {
                                                            if (ssDCval < 141) {
                                                                return 1;
                                                            }
                                                            else if (ssDCval >= 141) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                    else if (Cs >= 830) {
                                                        if (Cs < 1631) {
                                                            if (ssDCval < 156) {
                                                                if (Cs < 1197) {
                                                                    return 8;
                                                                }
                                                                else if (Cs >= 1197) {
                                                                    if (AFD < 10) {
                                                                        return 1;
                                                                    }
                                                                    else if (AFD >= 10) {
                                                                        return 8;
                                                                    }
                                                                }
                                                            }
                                                            else if (ssDCval >= 156) {
                                                                return 2;
                                                            }
                                                        }
                                                        else if (Cs >= 1631) {
                                                            if (RsDiff < 4) {
                                                                return 8;
                                                            }
                                                            else if (RsDiff >= 4) {
                                                                if (RsCsDiff < 107) {
                                                                    if (Cs < 2034) {
                                                                        if (Cs < 1973) {
                                                                            if (negBalance < 51) {
                                                                                if (Rs < 2581) {
                                                                                    return 4;
                                                                                }
                                                                                else if (Rs >= 2581) {
                                                                                    return 8;
                                                                                }
                                                                            }
                                                                            else if (negBalance >= 51) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                        else if (Cs >= 1973) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                    else if (Cs >= 2034) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (RsCsDiff >= 107) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (Rs >= 3540) {
                                                    if (TSC < 7) {
                                                        return 4;
                                                    }
                                                    else if (TSC >= 7) {
                                                        if (Cs < 2113) {
                                                            return 4;
                                                        }
                                                        else if (Cs >= 2113) {
                                                            if (ssDCval < 127) {
                                                                if (CsDiff < 7) {
                                                                    if (mu_mv_mag_sq < 79) {
                                                                        return 8;
                                                                    }
                                                                    else if (mu_mv_mag_sq >= 79) {
                                                                        if (Rs < 4617) {
                                                                            return 4;
                                                                        }
                                                                        else if (Rs >= 4617) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                }
                                                                else if (CsDiff >= 7) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (ssDCval >= 127) {
                                                                return 2;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (SC >= 7219) {
                                            if (ssDCval < 125) {
                                                return 8;
                                            }
                                            else if (ssDCval >= 125) {
                                                return 2;
                                            }
                                        }
                                    }
                                }
                            }
                        }
                        else if (AFD >= 12) {
                            if (RsDiff < 18) {
                                if (Cs < 1830) {
                                    if (Cs < 1396) {
                                        if (SC < 3176) {
                                            if (tcor < 90) {
                                                if (Cs < 529) {
                                                    if (refDCval < 86) {
                                                        return 2;
                                                    }
                                                    else if (refDCval >= 86) {
                                                        if (RsDiff < 13) {
                                                            if (refDCval < 130) {
                                                                return 1;
                                                            }
                                                            else if (refDCval >= 130) {
                                                                return 2;
                                                            }
                                                        }
                                                        else if (RsDiff >= 13) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                                else if (Cs >= 529) {
                                                    if (ssDCval < 73) {
                                                        if (posBalance < 58) {
                                                            if (negBalance < 47) {
                                                                return 4;
                                                            }
                                                            else if (negBalance >= 47) {
                                                                if (SC < 1187) {
                                                                    return 4;
                                                                }
                                                                else if (SC >= 1187) {
                                                                    if (mu_mv_mag_sq < 47) {
                                                                        return 8;
                                                                    }
                                                                    else if (mu_mv_mag_sq >= 47) {
                                                                        return 2;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (posBalance >= 58) {
                                                            if (TSC < 11) {
                                                                if (Rs < 837) {
                                                                    return 8;
                                                                }
                                                                else if (Rs >= 837) {
                                                                    return 1;
                                                                }
                                                            }
                                                            else if (TSC >= 11) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                    else if (ssDCval >= 73) {
                                                        if (diffMVdiffVal < 24) {
                                                            if (Rs < 797) {
                                                                if (negBalance < 62) {
                                                                    if (Rs < 660) {
                                                                        return 2;
                                                                    }
                                                                    else if (Rs >= 660) {
                                                                        return 8;
                                                                    }
                                                                }
                                                                else if (negBalance >= 62) {
                                                                    return 1;
                                                                }
                                                            }
                                                            else if (Rs >= 797) {
                                                                if (MVDiff < 15) {
                                                                    if (Cs < 923) {
                                                                        return 1;
                                                                    }
                                                                    else if (Cs >= 923) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (MVDiff >= 15) {
                                                                    if (Cs < 1177) {
                                                                        return 2;
                                                                    }
                                                                    else if (Cs >= 1177) {
                                                                        return 1;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (diffMVdiffVal >= 24) {
                                                            if (AFD < 48) {
                                                                return 4;
                                                            }
                                                            else if (AFD >= 48) {
                                                                if (RsDiff < 14) {
                                                                    return 1;
                                                                }
                                                                else if (RsDiff >= 14) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                            else if (tcor >= 90) {
                                                if (Cs < 443) {
                                                    if (Rs < 102) {
                                                        return 8;
                                                    }
                                                    else if (Rs >= 102) {
                                                        return 1;
                                                    }
                                                }
                                                else if (Cs >= 443) {
                                                    if (SC < 2743) {
                                                        return 4;
                                                    }
                                                    else if (SC >= 2743) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                        }
                                        else if (SC >= 3176) {
                                            if (CsDiff < 3) {
                                                return 8;
                                            }
                                            else if (CsDiff >= 3) {
                                                if (mcTcor < 84) {
                                                    if (Cs < 891) {
                                                        return 2;
                                                    }
                                                    else if (Cs >= 891) {
                                                        return 1;
                                                    }
                                                }
                                                else if (mcTcor >= 84) {
                                                    return 2;
                                                }
                                            }
                                        }
                                    }
                                    else if (Cs >= 1396) {
                                        if (mcTcor < 93) {
                                            if (RsCsDiff < 67) {
                                                if (Rs < 2000) {
                                                    return 8;
                                                }
                                                else if (Rs >= 2000) {
                                                    return 1;
                                                }
                                            }
                                            else if (RsCsDiff >= 67) {
                                                if (SC < 4671) {
                                                    if (MVDiff < 17) {
                                                        return 1;
                                                    }
                                                    else if (MVDiff >= 17) {
                                                        if (Cs < 1497) {
                                                            return 8;
                                                        }
                                                        else if (Cs >= 1497) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                                else if (SC >= 4671) {
                                                    return 8;
                                                }
                                            }
                                        }
                                        else if (mcTcor >= 93) {
                                            if (Rs < 2147) {
                                                if (TSC < 7) {
                                                    return 4;
                                                }
                                                else if (TSC >= 7) {
                                                    if (MVDiff < 10) {
                                                        return 2;
                                                    }
                                                    else if (MVDiff >= 10) {
                                                        if (Rs < 1686) {
                                                            return 2;
                                                        }
                                                        else if (Rs >= 1686) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (Rs >= 2147) {
                                                if (MVDiff < 10) {
                                                    return 1;
                                                }
                                                else if (MVDiff >= 10) {
                                                    if (diffMVdiffVal < 3) {
                                                        return 2;
                                                    }
                                                    else if (diffMVdiffVal >= 3) {
                                                        if (Rs < 2554) {
                                                            return 2;
                                                        }
                                                        else if (Rs >= 2554) {
                                                            return 1;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (Cs >= 1830) {
                                    if (ssDCval < 128) {
                                        if (ssDCval < 60) {
                                            if (MVDiff < 19) {
                                                return 8;
                                            }
                                            else if (MVDiff >= 19) {
                                                if (Rs < 3921) {
                                                    if (Rs < 3760) {
                                                        return 2;
                                                    }
                                                    else if (Rs >= 3760) {
                                                        return 8;
                                                    }
                                                }
                                                else if (Rs >= 3921) {
                                                    return 4;
                                                }
                                            }
                                        }
                                        else if (ssDCval >= 60) {
                                            if (ssDCval < 85) {
                                                if (RsCsDiff < 154) {
                                                    if (refDCval < 78) {
                                                        if (SC < 5497) {
                                                            return 8;
                                                        }
                                                        else if (SC >= 5497) {
                                                            if (SC < 6448) {
                                                                return 1;
                                                            }
                                                            else if (SC >= 6448) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                    else if (refDCval >= 78) {
                                                        if (posBalance < 54) {
                                                            if (diffRsCsdiff < -28) {
                                                                if (TSC < 10) {
                                                                    return 8;
                                                                }
                                                                else if (TSC >= 10) {
                                                                    return 1;
                                                                }
                                                            }
                                                            else if (diffRsCsdiff >= -28) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (posBalance >= 54) {
                                                            if (MVDiff < 8) {
                                                                return 8;
                                                            }
                                                            else if (MVDiff >= 8) {
                                                                if (mu_mv_mag_sq < 48) {
                                                                    if (SC < 4782) {
                                                                        if (gchDC < 1) {
                                                                            return 1;
                                                                        }
                                                                        else if (gchDC >= 1) {
                                                                            if (SC < 4074) {
                                                                                return 1;
                                                                            }
                                                                            else if (SC >= 4074) {
                                                                                if (Cs < 1939) {
                                                                                    return 2;
                                                                                }
                                                                                else if (Cs >= 1939) {
                                                                                    return 1;
                                                                                }
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (SC >= 4782) {
                                                                        if (Rs < 2898) {
                                                                            return 2;
                                                                        }
                                                                        else if (Rs >= 2898) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                }
                                                                else if (mu_mv_mag_sq >= 48) {
                                                                    return 1;
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (RsCsDiff >= 154) {
                                                    if (mcTcor < 80) {
                                                        if (posBalance < 46) {
                                                            return 8;
                                                        }
                                                        else if (posBalance >= 46) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (mcTcor >= 80) {
                                                        if (ssDCval < 79) {
                                                            return 1;
                                                        }
                                                        else if (ssDCval >= 79) {
                                                            if (CsDiff < 11) {
                                                                if (mcTcor < 86) {
                                                                    return 4;
                                                                }
                                                                else if (mcTcor >= 86) {
                                                                    if (refDCval < 81) {
                                                                        if (Cs < 2679) {
                                                                            return 2;
                                                                        }
                                                                        else if (Cs >= 2679) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                    else if (refDCval >= 81) {
                                                                        return 1;
                                                                    }
                                                                }
                                                            }
                                                            else if (CsDiff >= 11) {
                                                                if (mu_mv_mag_sq < 37) {
                                                                    return 8;
                                                                }
                                                                else if (mu_mv_mag_sq >= 37) {
                                                                    if (Rs < 3008) {
                                                                        return 1;
                                                                    }
                                                                    else if (Rs >= 3008) {
                                                                        return 8;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                            else if (ssDCval >= 85) {
                                                if (Cs < 2890) {
                                                    if (MVDiff < 26) {
                                                        if (Rs < 2609) {
                                                            if (ssDCval < 94) {
                                                                if (negBalance < 42) {
                                                                    if (MVDiff < 22) {
                                                                        if (diffRsCsdiff < -63) {
                                                                            return 1;
                                                                        }
                                                                        else if (diffRsCsdiff >= -63) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                    else if (MVDiff >= 22) {
                                                                        if (Cs < 2059) {
                                                                            return 1;
                                                                        }
                                                                        else if (Cs >= 2059) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                }
                                                                else if (negBalance >= 42) {
                                                                    if (negBalance < 46) {
                                                                        return 4;
                                                                    }
                                                                    else if (negBalance >= 46) {
                                                                        if (tcor < 95) {
                                                                            if (mcTcor < 90) {
                                                                                return 4;
                                                                            }
                                                                            else if (mcTcor >= 90) {
                                                                                if (Rs < 2551) {
                                                                                    return 2;
                                                                                }
                                                                                else if (Rs >= 2551) {
                                                                                    if (tcor < 87) {
                                                                                        return 2;
                                                                                    }
                                                                                    else if (tcor >= 87) {
                                                                                        if (Cs < 1930) {
                                                                                            return 2;
                                                                                        }
                                                                                        else if (Cs >= 1930) {
                                                                                            return 1;
                                                                                        }
                                                                                    }
                                                                                }
                                                                            }
                                                                        }
                                                                        else if (tcor >= 95) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                            else if (ssDCval >= 94) {
                                                                if (SC < 4094) {
                                                                    return 4;
                                                                }
                                                                else if (SC >= 4094) {
                                                                    if (RsCsDiff < 221) {
                                                                        if (MVDiff < 7) {
                                                                            return 4;
                                                                        }
                                                                        else if (MVDiff >= 7) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                    else if (RsCsDiff >= 221) {
                                                                        return 8;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (Rs >= 2609) {
                                                            if (Rs < 2793) {
                                                                if (AFD < 14) {
                                                                    return 1;
                                                                }
                                                                else if (AFD >= 14) {
                                                                    if (ssDCval < 101) {
                                                                        if (diffTSC < 1) {
                                                                            return 2;
                                                                        }
                                                                        else if (diffTSC >= 1) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                    else if (ssDCval >= 101) {
                                                                        return 1;
                                                                    }
                                                                }
                                                            }
                                                            else if (Rs >= 2793) {
                                                                if (gchDC < 2) {
                                                                    if (tcor < 90) {
                                                                        if (SC < 5375) {
                                                                            if (ssDCval < 97) {
                                                                                if (diffAFD < 1) {
                                                                                    if (posBalance < 58) {
                                                                                        return 4;
                                                                                    }
                                                                                    else if (posBalance >= 58) {
                                                                                        if (diffMVdiffVal < 2) {
                                                                                            if (Cs < 2309) {
                                                                                                return 2;
                                                                                            }
                                                                                            else if (Cs >= 2309) {
                                                                                                return 1;
                                                                                            }
                                                                                        }
                                                                                        else if (diffMVdiffVal >= 2) {
                                                                                            if (Cs < 2278) {
                                                                                                return 1;
                                                                                            }
                                                                                            else if (Cs >= 2278) {
                                                                                                return 4;
                                                                                            }
                                                                                        }
                                                                                    }
                                                                                }
                                                                                else if (diffAFD >= 1) {
                                                                                    return 1;
                                                                                }
                                                                            }
                                                                            else if (ssDCval >= 97) {
                                                                                return 1;
                                                                            }
                                                                        }
                                                                        else if (SC >= 5375) {
                                                                            if (RsDiff < 10) {
                                                                                if (Cs < 2461) {
                                                                                    if (Rs < 3007) {
                                                                                        if (Rs < 2985) {
                                                                                            return 1;
                                                                                        }
                                                                                        else if (Rs >= 2985) {
                                                                                            return 8;
                                                                                        }
                                                                                    }
                                                                                    else if (Rs >= 3007) {
                                                                                        if (Rs < 3650) {
                                                                                            return 4;
                                                                                        }
                                                                                        else if (Rs >= 3650) {
                                                                                            return 8;
                                                                                        }
                                                                                    }
                                                                                }
                                                                                else if (Cs >= 2461) {
                                                                                    if (MVDiff < 3) {
                                                                                        return 8;
                                                                                    }
                                                                                    else if (MVDiff >= 3) {
                                                                                        return 1;
                                                                                    }
                                                                                }
                                                                            }
                                                                            else if (RsDiff >= 10) {
                                                                                if (AFD < 16) {
                                                                                    if (SC < 6821) {
                                                                                        if (Rs < 3025) {
                                                                                            return 1;
                                                                                        }
                                                                                        else if (Rs >= 3025) {
                                                                                            return 4;
                                                                                        }
                                                                                    }
                                                                                    else if (SC >= 6821) {
                                                                                        if (Cs < 2777) {
                                                                                            return 4;
                                                                                        }
                                                                                        else if (Cs >= 2777) {
                                                                                            return 2;
                                                                                        }
                                                                                    }
                                                                                }
                                                                                else if (AFD >= 16) {
                                                                                    if (MVDiff < 18) {
                                                                                        if (ssDCval < 112) {
                                                                                            if (Rs < 3157) {
                                                                                                if (AFD < 17) {
                                                                                                    return 4;
                                                                                                }
                                                                                                else if (AFD >= 17) {
                                                                                                    return 2;
                                                                                                }
                                                                                            }
                                                                                            else if (Rs >= 3157) {
                                                                                                return 1;
                                                                                            }
                                                                                        }
                                                                                        else if (ssDCval >= 112) {
                                                                                            return 1;
                                                                                        }
                                                                                    }
                                                                                    else if (MVDiff >= 18) {
                                                                                        return 4;
                                                                                    }
                                                                                }
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (tcor >= 90) {
                                                                        if (ssDCval < 88) {
                                                                            if (diffAFD < 1) {
                                                                                return 8;
                                                                            }
                                                                            else if (diffAFD >= 1) {
                                                                                if (Rs < 3074) {
                                                                                    return 8;
                                                                                }
                                                                                else if (Rs >= 3074) {
                                                                                    return 4;
                                                                                }
                                                                            }
                                                                        }
                                                                        else if (ssDCval >= 88) {
                                                                            if (CsDiff < 9) {
                                                                                if (ssDCval < 118) {
                                                                                    if (Rs < 2826) {
                                                                                        return 4;
                                                                                    }
                                                                                    else if (Rs >= 2826) {
                                                                                        if (diffAFD < -1) {
                                                                                            if (diffRsCsdiff < -42) {
                                                                                                return 4;
                                                                                            }
                                                                                            else if (diffRsCsdiff >= -42) {
                                                                                                return 1;
                                                                                            }
                                                                                        }
                                                                                        else if (diffAFD >= -1) {
                                                                                            return 1;
                                                                                        }
                                                                                    }
                                                                                }
                                                                                else if (ssDCval >= 118) {
                                                                                    return 8;
                                                                                }
                                                                            }
                                                                            else if (CsDiff >= 9) {
                                                                                if (Rs < 3592) {
                                                                                    return 8;
                                                                                }
                                                                                else if (Rs >= 3592) {
                                                                                    return 4;
                                                                                }
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                                else if (gchDC >= 2) {
                                                                    if (SC < 5879) {
                                                                        if (Cs < 2214) {
                                                                            return 8;
                                                                        }
                                                                        else if (Cs >= 2214) {
                                                                            if (Cs < 2651) {
                                                                                return 2;
                                                                            }
                                                                            else if (Cs >= 2651) {
                                                                                return 8;
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (SC >= 5879) {
                                                                        return 4;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (MVDiff >= 26) {
                                                        if (mu_mv_mag_sq < 38) {
                                                            if (diffTSC < 4) {
                                                                if (AFD < 18) {
                                                                    return 4;
                                                                }
                                                                else if (AFD >= 18) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (diffTSC >= 4) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (mu_mv_mag_sq >= 38) {
                                                            if (diffRsCsdiff < -1) {
                                                                if (mu_mv_mag_sq < 51) {
                                                                    return 4;
                                                                }
                                                                else if (mu_mv_mag_sq >= 51) {
                                                                    if (ssDCval < 107) {
                                                                        return 2;
                                                                    }
                                                                    else if (ssDCval >= 107) {
                                                                        return 4;
                                                                    }
                                                                }
                                                            }
                                                            else if (diffRsCsdiff >= -1) {
                                                                if (tcor < 91) {
                                                                    return 2;
                                                                }
                                                                else if (tcor >= 91) {
                                                                    if (RsCsDiff < 163) {
                                                                        return 2;
                                                                    }
                                                                    else if (RsCsDiff >= 163) {
                                                                        return 4;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (Cs >= 2890) {
                                                    if (SC < 7223) {
                                                        if (Rs < 3800) {
                                                            if (diffMVdiffVal < -1) {
                                                                if (Cs < 3012) {
                                                                    return 4;
                                                                }
                                                                else if (Cs >= 3012) {
                                                                    if (AFD < 18) {
                                                                        return 8;
                                                                    }
                                                                    else if (AFD >= 18) {
                                                                        return 1;
                                                                    }
                                                                }
                                                            }
                                                            else if (diffMVdiffVal >= -1) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (Rs >= 3800) {
                                                            if (posBalance < 46) {
                                                                return 8;
                                                            }
                                                            else if (posBalance >= 46) {
                                                                if (SC < 7168) {
                                                                    return 4;
                                                                }
                                                                else if (SC >= 7168) {
                                                                    if (AFD < 16) {
                                                                        return 1;
                                                                    }
                                                                    else if (AFD >= 16) {
                                                                        return 8;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (SC >= 7223) {
                                                        if (refDCval < 107) {
                                                            return 4;
                                                        }
                                                        else if (refDCval >= 107) {
                                                            return 1;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (ssDCval >= 128) {
                                        if (mcTcor < 89) {
                                            if (posBalance < 48) {
                                                return 8;
                                            }
                                            else if (posBalance >= 48) {
                                                if (diffMVdiffVal < -53) {
                                                    return 4;
                                                }
                                                else if (diffMVdiffVal >= -53) {
                                                    return 1;
                                                }
                                            }
                                        }
                                        else if (mcTcor >= 89) {
                                            if (MVDiff < 46) {
                                                if (AFD < 24) {
                                                    return 8;
                                                }
                                                else if (AFD >= 24) {
                                                    return 2;
                                                }
                                            }
                                            else if (MVDiff >= 46) {
                                                if (Rs < 2505) {
                                                    return 4;
                                                }
                                                else if (Rs >= 2505) {
                                                    return 2;
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                            else if (RsDiff >= 18) {
                                if (tcor < 72) {
                                    if (Scindex < 6) {
                                        if (Cs < 2091) {
                                            return 2;
                                        }
                                        else if (Cs >= 2091) {
                                            return 8;
                                        }
                                    }
                                    else if (Scindex >= 6) {
                                        if (MVDiff < 68) {
                                            if (AFD < 28) {
                                                return 1;
                                            }
                                            else if (AFD >= 28) {
                                                return 2;
                                            }
                                        }
                                        else if (MVDiff >= 68) {
                                            return 8;
                                        }
                                    }
                                }
                                else if (tcor >= 72) {
                                    return 4;
                                }
                            }
                        }
                    }
                }
                else if (SC >= 7382) {
                    if (Cs < 3290) {
                        if (tcor < 77) {
                            if (ssDCval < 107) {
                                if (posBalance < 64) {
                                    return 4;
                                }
                                else if (posBalance >= 64) {
                                    if (Rs < 4308) {
                                        return 8;
                                    }
                                    else if (Rs >= 4308) {
                                        return 2;
                                    }
                                }
                            }
                            else if (ssDCval >= 107) {
                                if (AFD < 28) {
                                    if (refDCval < 146) {
                                        if (posBalance < 52) {
                                            if (negBalance < 42) {
                                                return 1;
                                            }
                                            else if (negBalance >= 42) {
                                                return 2;
                                            }
                                        }
                                        else if (posBalance >= 52) {
                                            if (refDCval < 123) {
                                                return 8;
                                            }
                                            else if (refDCval >= 123) {
                                                return 1;
                                            }
                                        }
                                    }
                                    else if (refDCval >= 146) {
                                        return 4;
                                    }
                                }
                                else if (AFD >= 28) {
                                    if (RsCsDiff < 1081) {
                                        return 2;
                                    }
                                    else if (RsCsDiff >= 1081) {
                                        return 4;
                                    }
                                }
                            }
                        }
                        else if (tcor >= 77) {
                            if (CsDiff < 11) {
                                if (posBalance < 48) {
                                    if (refDCval < 128) {
                                        if (CsDiff < 8) {
                                            if (Rs < 4474) {
                                                return 4;
                                            }
                                            else if (Rs >= 4474) {
                                                if (ssDCval < 125) {
                                                    if (TSC < 10) {
                                                        return 8;
                                                    }
                                                    else if (TSC >= 10) {
                                                        if (Cs < 3055) {
                                                            return 8;
                                                        }
                                                        else if (Cs >= 3055) {
                                                            return 2;
                                                        }
                                                    }
                                                }
                                                else if (ssDCval >= 125) {
                                                    return 2;
                                                }
                                            }
                                        }
                                        else if (CsDiff >= 8) {
                                            if (tcor < 81) {
                                                return 2;
                                            }
                                            else if (tcor >= 81) {
                                                return 4;
                                            }
                                        }
                                    }
                                    else if (refDCval >= 128) {
                                        return 2;
                                    }
                                }
                                else if (posBalance >= 48) {
                                    if (diffAFD < -1) {
                                        if (posBalance < 61) {
                                            return 4;
                                        }
                                        else if (posBalance >= 61) {
                                            return 2;
                                        }
                                    }
                                    else if (diffAFD >= -1) {
                                        if (negBalance < 43) {
                                            if (tcor < 83) {
                                                if (AFD < 14) {
                                                    return 4;
                                                }
                                                else if (AFD >= 14) {
                                                    return 2;
                                                }
                                            }
                                            else if (tcor >= 83) {
                                                if (mu_mv_mag_sq < 61) {
                                                    return 4;
                                                }
                                                else if (mu_mv_mag_sq >= 61) {
                                                    return 2;
                                                }
                                            }
                                        }
                                        else if (negBalance >= 43) {
                                            return 2;
                                        }
                                    }
                                }
                            }
                            else if (CsDiff >= 11) {
                                if (refDCval < 130) {
                                    if (mcTcor < 88) {
                                        if (SC < 8556) {
                                            return 2;
                                        }
                                        else if (SC >= 8556) {
                                            return 1;
                                        }
                                    }
                                    else if (mcTcor >= 88) {
                                        return 4;
                                    }
                                }
                                else if (refDCval >= 130) {
                                    return 8;
                                }
                            }
                        }
                    }
                    else if (Cs >= 3290) {
                        if (refDCval < 68) {
                            if (TSC < 7) {
                                if (ssDCval < 54) {
                                    if (Rs < 3960) {
                                        if (Cs < 3899) {
                                            return 4;
                                        }
                                        else if (Cs >= 3899) {
                                            return 2;
                                        }
                                    }
                                    else if (Rs >= 3960) {
                                        return 8;
                                    }
                                }
                                else if (ssDCval >= 54) {
                                    if (SC < 7987) {
                                        if (refDCval < 56) {
                                            if (diffRsCsdiff < 8) {
                                                return 8;
                                            }
                                            else if (diffRsCsdiff >= 8) {
                                                return 4;
                                            }
                                        }
                                        else if (refDCval >= 56) {
                                            return 8;
                                        }
                                    }
                                    else if (SC >= 7987) {
                                        if (mu_mv_mag_sq < 45) {
                                            return 8;
                                        }
                                        else if (mu_mv_mag_sq >= 45) {
                                            return 4;
                                        }
                                    }
                                }
                            }
                            else if (TSC >= 7) {
                                if (TSC < 11) {
                                    if (RsDiff < 5) {
                                        if (negBalance < 51) {
                                            if (posBalance < 58) {
                                                return 4;
                                            }
                                            else if (posBalance >= 58) {
                                                if (posBalance < 60) {
                                                    if (Rs < 4009) {
                                                        return 2;
                                                    }
                                                    else if (Rs >= 4009) {
                                                        return 4;
                                                    }
                                                }
                                                else if (posBalance >= 60) {
                                                    return 8;
                                                }
                                            }
                                        }
                                        else if (negBalance >= 51) {
                                            if (CsDiff < 6) {
                                                return 2;
                                            }
                                            else if (CsDiff >= 6) {
                                                return 8;
                                            }
                                        }
                                    }
                                    else if (RsDiff >= 5) {
                                        if (tcor < 80) {
                                            if (MVDiff < 16) {
                                                return 4;
                                            }
                                            else if (MVDiff >= 16) {
                                                return 8;
                                            }
                                        }
                                        else if (tcor >= 80) {
                                            if (mcTcor < 87) {
                                                return 8;
                                            }
                                            else if (mcTcor >= 87) {
                                                return 4;
                                            }
                                        }
                                    }
                                }
                                else if (TSC >= 11) {
                                    if (RsDiff < 19) {
                                        if (AFD < 17) {
                                            if (RsDiff < 11) {
                                                return 4;
                                            }
                                            else if (RsDiff >= 11) {
                                                return 8;
                                            }
                                        }
                                        else if (AFD >= 17) {
                                            return 8;
                                        }
                                    }
                                    else if (RsDiff >= 19) {
                                        return 4;
                                    }
                                }
                            }
                        }
                        else if (refDCval >= 68) {
                            if (refDCval < 153) {
                                if (mu_mv_mag_sq < 63) {
                                    if (Rs < 6965) {
                                        if (Cs < 4307) {
                                            if (tcor < 66) {
                                                if (RsCsDiff < 569) {
                                                    if (Cs < 3640) {
                                                        return 8;
                                                    }
                                                    else if (Cs >= 3640) {
                                                        return 1;
                                                    }
                                                }
                                                else if (RsCsDiff >= 569) {
                                                    if (refDCval < 135) {
                                                        if (Rs < 4377) {
                                                            if (Rs < 4138) {
                                                                return 8;
                                                            }
                                                            else if (Rs >= 4138) {
                                                                return 2;
                                                            }
                                                        }
                                                        else if (Rs >= 4377) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (refDCval >= 135) {
                                                        if (TSC < 17) {
                                                            if (SC < 9763) {
                                                                return 2;
                                                            }
                                                            else if (SC >= 9763) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (TSC >= 17) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (tcor >= 66) {
                                                if (ssDCval < 139) {
                                                    if (mcTcor < 82) {
                                                        if (CsDiff < 10) {
                                                            if (Cs < 3367) {
                                                                return 1;
                                                            }
                                                            else if (Cs >= 3367) {
                                                                return 2;
                                                            }
                                                        }
                                                        else if (CsDiff >= 10) {
                                                            if (gchDC < 2) {
                                                                return 4;
                                                            }
                                                            else if (gchDC >= 2) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                    else if (mcTcor >= 82) {
                                                        if (RsCsDiff < 67) {
                                                            return 4;
                                                        }
                                                        else if (RsCsDiff >= 67) {
                                                            if (Rs < 5594) {
                                                                if (posBalance < 51) {
                                                                    return 8;
                                                                }
                                                                else if (posBalance >= 51) {
                                                                    if (diffMVdiffVal < -7) {
                                                                        return 8;
                                                                    }
                                                                    else if (diffMVdiffVal >= -7) {
                                                                        if (negBalance < 57) {
                                                                            if (TSC < 15) {
                                                                                if (ssDCval < 121) {
                                                                                    if (CsDiff < 9) {
                                                                                        return 8;
                                                                                    }
                                                                                    else if (CsDiff >= 9) {
                                                                                        return 4;
                                                                                    }
                                                                                }
                                                                                else if (ssDCval >= 121) {
                                                                                    return 8;
                                                                                }
                                                                            }
                                                                            else if (TSC >= 15) {
                                                                                if (SC < 9327) {
                                                                                    return 4;
                                                                                }
                                                                                else if (SC >= 9327) {
                                                                                    return 8;
                                                                                }
                                                                            }
                                                                        }
                                                                        else if (negBalance >= 57) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                            else if (Rs >= 5594) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (ssDCval >= 139) {
                                                    return 8;
                                                }
                                            }
                                        }
                                        else if (Cs >= 4307) {
                                            if (mcTcor < 90) {
                                                if (Rs < 4126) {
                                                    return 1;
                                                }
                                                else if (Rs >= 4126) {
                                                    if (MVDiff < 105) {
                                                        if (diffRsCsdiff < 215) {
                                                            if (diffTSC < -4) {
                                                                return 4;
                                                            }
                                                            else if (diffTSC >= -4) {
                                                                if (posBalance < 65) {
                                                                    return 8;
                                                                }
                                                                else if (posBalance >= 65) {
                                                                    if (Rs < 5945) {
                                                                        return 8;
                                                                    }
                                                                    else if (Rs >= 5945) {
                                                                        return 4;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (diffRsCsdiff >= 215) {
                                                            if (refDCval < 138) {
                                                                return 8;
                                                            }
                                                            else if (refDCval >= 138) {
                                                                if (posBalance < 55) {
                                                                    return 4;
                                                                }
                                                                else if (posBalance >= 55) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (MVDiff >= 105) {
                                                        return 2;
                                                    }
                                                }
                                            }
                                            else if (mcTcor >= 90) {
                                                if (mu_mv_mag_sq < 32) {
                                                    if (RsCsDiff < 1184) {
                                                        return 1;
                                                    }
                                                    else if (RsCsDiff >= 1184) {
                                                        if (Rs < 6317) {
                                                            return 8;
                                                        }
                                                        else if (Rs >= 6317) {
                                                            return 2;
                                                        }
                                                    }
                                                }
                                                else if (mu_mv_mag_sq >= 32) {
                                                    if (SC < 11761) {
                                                        return 8;
                                                    }
                                                    else if (SC >= 11761) {
                                                        if (TSC < 17) {
                                                            return 1;
                                                        }
                                                        else if (TSC >= 17) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (Rs >= 6965) {
                                        if (mu_mv_mag_sq < 11) {
                                            if (ssDCval < 75) {
                                                return 4;
                                            }
                                            else if (ssDCval >= 75) {
                                                if (tcor < 84) {
                                                    if (TSC < 28) {
                                                        if (mcTcor < 91) {
                                                            if (MVDiff < 1) {
                                                                return 4;
                                                            }
                                                            else if (MVDiff >= 1) {
                                                                if (diffTSC < -3) {
                                                                    if (diffMVdiffVal < 3) {
                                                                        return 8;
                                                                    }
                                                                    else if (diffMVdiffVal >= 3) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (diffTSC >= -3) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                        else if (mcTcor >= 91) {
                                                            if (Cs < 6010) {
                                                                return 4;
                                                            }
                                                            else if (Cs >= 6010) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                    else if (TSC >= 28) {
                                                        if (gchDC < 2) {
                                                            return 4;
                                                        }
                                                        else if (gchDC >= 2) {
                                                            if (Rs < 14379) {
                                                                if (Cs < 9842) {
                                                                    return 2;
                                                                }
                                                                else if (Cs >= 9842) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (Rs >= 14379) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (tcor >= 84) {
                                                    return 4;
                                                }
                                            }
                                        }
                                        else if (mu_mv_mag_sq >= 11) {
                                            if (Cs < 6367) {
                                                if (AFD < 27) {
                                                    return 4;
                                                }
                                                else if (AFD >= 27) {
                                                    if (negBalance < 61) {
                                                        if (RsDiff < 38) {
                                                            if (ssDCval < 139) {
                                                                if (gchDC < 1) {
                                                                    if (Rs < 8375) {
                                                                        return 4;
                                                                    }
                                                                    else if (Rs >= 8375) {
                                                                        return 2;
                                                                    }
                                                                }
                                                                else if (gchDC >= 1) {
                                                                    if (posBalance < 59) {
                                                                        return 2;
                                                                    }
                                                                    else if (posBalance >= 59) {
                                                                        return 8;
                                                                    }
                                                                }
                                                            }
                                                            else if (ssDCval >= 139) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (RsDiff >= 38) {
                                                            return 2;
                                                        }
                                                    }
                                                    else if (negBalance >= 61) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                            else if (Cs >= 6367) {
                                                if (tcor < 76) {
                                                    if (AFD < 37) {
                                                        if (posBalance < 55) {
                                                            if (Cs < 7204) {
                                                                return 4;
                                                            }
                                                            else if (Cs >= 7204) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (posBalance >= 55) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (AFD >= 37) {
                                                        if (gchDC < 2) {
                                                            if (SC < 23573) {
                                                                return 2;
                                                            }
                                                            else if (SC >= 23573) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (gchDC >= 2) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                                else if (tcor >= 76) {
                                                    if (mu_mv_mag_sq < 39) {
                                                        return 4;
                                                    }
                                                    else if (mu_mv_mag_sq >= 39) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (mu_mv_mag_sq >= 63) {
                                    if (SC < 9709) {
                                        if (refDCval < 137) {
                                            if (ssDCval < 124) {
                                                if (ssDCval < 105) {
                                                    if (Cs < 3544) {
                                                        if (Rs < 3972) {
                                                            return 4;
                                                        }
                                                        else if (Rs >= 3972) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (Cs >= 3544) {
                                                        return 8;
                                                    }
                                                }
                                                else if (ssDCval >= 105) {
                                                    if (posBalance < 49) {
                                                        if (Rs < 4863) {
                                                            return 8;
                                                        }
                                                        else if (Rs >= 4863) {
                                                            return 2;
                                                        }
                                                    }
                                                    else if (posBalance >= 49) {
                                                        if (TSC < 11) {
                                                            if (Cs < 3576) {
                                                                return 2;
                                                            }
                                                            else if (Cs >= 3576) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (TSC >= 11) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (ssDCval >= 124) {
                                                if (RsCsDiff < 657) {
                                                    return 8;
                                                }
                                                else if (RsCsDiff >= 657) {
                                                    return 1;
                                                }
                                            }
                                        }
                                        else if (refDCval >= 137) {
                                            if (negBalance < 53) {
                                                return 1;
                                            }
                                            else if (negBalance >= 53) {
                                                return 8;
                                            }
                                        }
                                    }
                                    else if (SC >= 9709) {
                                        if (Cs < 7228) {
                                            if (TSC < 11) {
                                                return 4;
                                            }
                                            else if (TSC >= 11) {
                                                return 8;
                                            }
                                        }
                                        else if (Cs >= 7228) {
                                            if (MVDiff < 1) {
                                                return 4;
                                            }
                                            else if (MVDiff >= 1) {
                                                return 8;
                                            }
                                        }
                                    }
                                }
                            }
                            else if (refDCval >= 153) {
                                if (CsDiff < 23) {
                                    if (negBalance < 54) {
                                        if (AFD < 27) {
                                            return 8;
                                        }
                                        else if (AFD >= 27) {
                                            return 4;
                                        }
                                    }
                                    else if (negBalance >= 54) {
                                        if (MVDiff < 30) {
                                            if (Rs < 5140) {
                                                return 8;
                                            }
                                            else if (Rs >= 5140) {
                                                if (MVDiff < 27) {
                                                    return 2;
                                                }
                                                else if (MVDiff >= 27) {
                                                    if (RsCsDiff < 1013) {
                                                        return 2;
                                                    }
                                                    else if (RsCsDiff >= 1013) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                        }
                                        else if (MVDiff >= 30) {
                                            if (RsCsDiff < 553) {
                                                return 2;
                                            }
                                            else if (RsCsDiff >= 553) {
                                                if (Rs < 6583) {
                                                    return 4;
                                                }
                                                else if (Rs >= 6583) {
                                                    return 2;
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (CsDiff >= 23) {
                                    if (negBalance < 54) {
                                        if (ssDCval < 166) {
                                            return 8;
                                        }
                                        else if (ssDCval >= 166) {
                                            return 4;
                                        }
                                    }
                                    else if (negBalance >= 54) {
                                        if (Rs < 6588) {
                                            return 8;
                                        }
                                        else if (Rs >= 6588) {
                                            return 2;
                                        }
                                    }
                                }
                            }
                        }
                    }
                }
            }
        }
        else if (mu_mv_mag_sq >= 89) {
            if (negBalance < 36) {
                if (SC < 1258) {
                    if (refDCval < 161) {
                        if (Cs < 547) {
                            if (SC < 579) {
                                if (AFD < 2) {
                                    if (mu_mv_mag_sq < 127) {
                                        if (CsDiff < 1) {
                                            return 4;
                                        }
                                        else if (CsDiff >= 1) {
                                            return 8;
                                        }
                                    }
                                    else if (mu_mv_mag_sq >= 127) {
                                        if (negBalance < 16) {
                                            return 8;
                                        }
                                        else if (negBalance >= 16) {
                                            if (Rs < 181) {
                                                return 4;
                                            }
                                            else if (Rs >= 181) {
                                                if (Cs < 316) {
                                                    return 8;
                                                }
                                                else if (Cs >= 316) {
                                                    if (Rs < 184) {
                                                        return 4;
                                                    }
                                                    else if (Rs >= 184) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (AFD >= 2) {
                                    if (Rs < 103) {
                                        if (diffAFD < 2) {
                                            return 8;
                                        }
                                        else if (diffAFD >= 2) {
                                            return 2;
                                        }
                                    }
                                    else if (Rs >= 103) {
                                        if (TSC < 16) {
                                            if (ssDCval < 121) {
                                                return 4;
                                            }
                                            else if (ssDCval >= 121) {
                                                return 8;
                                            }
                                        }
                                        else if (TSC >= 16) {
                                            return 1;
                                        }
                                    }
                                }
                            }
                            else if (SC >= 579) {
                                if (negBalance < 31) {
                                    if (diffMVdiffVal < 6) {
                                        if (negBalance < 26) {
                                            return 8;
                                        }
                                        else if (negBalance >= 26) {
                                            if (Cs < 524) {
                                                return 8;
                                            }
                                            else if (Cs >= 524) {
                                                return 4;
                                            }
                                        }
                                    }
                                    else if (diffMVdiffVal >= 6) {
                                        if (AFD < 6) {
                                            return 4;
                                        }
                                        else if (AFD >= 6) {
                                            return 8;
                                        }
                                    }
                                }
                                else if (negBalance >= 31) {
                                    return 4;
                                }
                            }
                        }
                        else if (Cs >= 547) {
                            if (AFD < 3) {
                                if (ssDCval < 40) {
                                    return 8;
                                }
                                else if (ssDCval >= 40) {
                                    return 4;
                                }
                            }
                            else if (AFD >= 3) {
                                if (SC < 1103) {
                                    return 8;
                                }
                                else if (SC >= 1103) {
                                    return 4;
                                }
                            }
                        }
                    }
                    else if (refDCval >= 161) {
                        if (SC < 879) {
                            return 4;
                        }
                        else if (SC >= 879) {
                            if (ssDCval < 164) {
                                return 4;
                            }
                            else if (ssDCval >= 164) {
                                if (Rs < 535) {
                                    return 2;
                                }
                                else if (Rs >= 535) {
                                    if (SC < 1133) {
                                        return 4;
                                    }
                                    else if (SC >= 1133) {
                                        return 1;
                                    }
                                }
                            }
                        }
                    }
                }
                else if (SC >= 1258) {
                    if (Rs < 3779) {
                        if (Rs < 3774) {
                            if (ssDCval < 70) {
                                if (ssDCval < 33) {
                                    if (Cs < 708) {
                                        return 4;
                                    }
                                    else if (Cs >= 708) {
                                        if (ssDCval < 31) {
                                            if (mu_mv_mag_sq < 105) {
                                                if (Cs < 922) {
                                                    return 8;
                                                }
                                                else if (Cs >= 922) {
                                                    return 4;
                                                }
                                            }
                                            else if (mu_mv_mag_sq >= 105) {
                                                return 8;
                                            }
                                        }
                                        else if (ssDCval >= 31) {
                                            if (Rs < 963) {
                                                if (tcor < 100) {
                                                    return 8;
                                                }
                                                else if (tcor >= 100) {
                                                    if (Cs < 987) {
                                                        return 1;
                                                    }
                                                    else if (Cs >= 987) {
                                                        if (Rs < 784) {
                                                            return 8;
                                                        }
                                                        else if (Rs >= 784) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (Rs >= 963) {
                                                if (negBalance < 2) {
                                                    return 2;
                                                }
                                                else if (negBalance >= 2) {
                                                    if (SC < 2272) {
                                                        return 4;
                                                    }
                                                    else if (SC >= 2272) {
                                                        if (Cs < 1492) {
                                                            return 2;
                                                        }
                                                        else if (Cs >= 1492) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (ssDCval >= 33) {
                                    if (MVDiff < 17) {
                                        if (Cs < 1006) {
                                            if (ssDCval < 56) {
                                                if (mu_mv_mag_sq < 104) {
                                                    return 1;
                                                }
                                                else if (mu_mv_mag_sq >= 104) {
                                                    if (posBalance < 38) {
                                                        return 8;
                                                    }
                                                    else if (posBalance >= 38) {
                                                        if (Rs < 1118) {
                                                            return 8;
                                                        }
                                                        else if (Rs >= 1118) {
                                                            return 1;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (ssDCval >= 56) {
                                                return 8;
                                            }
                                        }
                                        else if (Cs >= 1006) {
                                            return 8;
                                        }
                                    }
                                    else if (MVDiff >= 17) {
                                        if (Rs < 1352) {
                                            return 8;
                                        }
                                        else if (Rs >= 1352) {
                                            return 4;
                                        }
                                    }
                                }
                            }
                            else if (ssDCval >= 70) {
                                if (tcor < 93) {
                                    if (SC < 2821) {
                                        if (SC < 2591) {
                                            if (MVDiff < 8) {
                                                if (diffMVdiffVal < -4) {
                                                    return 1;
                                                }
                                                else if (diffMVdiffVal >= -4) {
                                                    if (Rs < 1277) {
                                                        if (diffAFD < -3) {
                                                            return 4;
                                                        }
                                                        else if (diffAFD >= -3) {
                                                            if (refDCval < 130) {
                                                                return 4;
                                                            }
                                                            else if (refDCval >= 130) {
                                                                if (MVDiff < 7) {
                                                                    return 8;
                                                                }
                                                                else if (MVDiff >= 7) {
                                                                    if (Rs < 883) {
                                                                        return 8;
                                                                    }
                                                                    else if (Rs >= 883) {
                                                                        return 1;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (Rs >= 1277) {
                                                        return 1;
                                                    }
                                                }
                                            }
                                            else if (MVDiff >= 8) {
                                                if (mcTcor < 89) {
                                                    return 8;
                                                }
                                                else if (mcTcor >= 89) {
                                                    if (Rs < 1530) {
                                                        return 1;
                                                    }
                                                    else if (Rs >= 1530) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                        }
                                        else if (SC >= 2591) {
                                            if (Rs < 1900) {
                                                return 1;
                                            }
                                            else if (Rs >= 1900) {
                                                return 8;
                                            }
                                        }
                                    }
                                    else if (SC >= 2821) {
                                        if (TSC < 3) {
                                            if (Rs < 1276) {
                                                return 4;
                                            }
                                            else if (Rs >= 1276) {
                                                if (SC < 4426) {
                                                    if (SC < 3827) {
                                                        return 8;
                                                    }
                                                    else if (SC >= 3827) {
                                                        return 1;
                                                    }
                                                }
                                                else if (SC >= 4426) {
                                                    if (ssDCval < 117) {
                                                        if (diffMVdiffVal < -2) {
                                                            return 4;
                                                        }
                                                        else if (diffMVdiffVal >= -2) {
                                                            if (SC < 4584) {
                                                                return 4;
                                                            }
                                                            else if (SC >= 4584) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                    else if (ssDCval >= 117) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                        }
                                        else if (TSC >= 3) {
                                            if (posBalance < 14) {
                                                return 1;
                                            }
                                            else if (posBalance >= 14) {
                                                if (mcTcor < 94) {
                                                    return 8;
                                                }
                                                else if (mcTcor >= 94) {
                                                    if (SC < 5493) {
                                                        return 1;
                                                    }
                                                    else if (SC >= 5493) {
                                                        return 2;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (tcor >= 93) {
                                    if (ssDCval < 110) {
                                        if (posBalance < 48) {
                                            if (mu_mv_mag_sq < 123) {
                                                if (Rs < 3394) {
                                                    if (SC < 3711) {
                                                        if (tcor < 98) {
                                                            return 4;
                                                        }
                                                        else if (tcor >= 98) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (SC >= 3711) {
                                                        if (posBalance < 28) {
                                                            return 1;
                                                        }
                                                        else if (posBalance >= 28) {
                                                            if (mu_mv_mag_sq < 116) {
                                                                if (RsDiff < 6) {
                                                                    if (negBalance < 32) {
                                                                        if (Cs < 1338) {
                                                                            return 1;
                                                                        }
                                                                        else if (Cs >= 1338) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                    else if (negBalance >= 32) {
                                                                        if (Cs < 2304) {
                                                                            return 1;
                                                                        }
                                                                        else if (Cs >= 2304) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                }
                                                                else if (RsDiff >= 6) {
                                                                    if (SC < 4291) {
                                                                        return 8;
                                                                    }
                                                                    else if (SC >= 4291) {
                                                                        return 1;
                                                                    }
                                                                }
                                                            }
                                                            else if (mu_mv_mag_sq >= 116) {
                                                                if (negBalance < 25) {
                                                                    return 8;
                                                                }
                                                                else if (negBalance >= 25) {
                                                                    if (Rs < 3227) {
                                                                        if (negBalance < 27) {
                                                                            if (diffAFD < -1) {
                                                                                if (Rs < 2999) {
                                                                                    return 8;
                                                                                }
                                                                                else if (Rs >= 2999) {
                                                                                    return 1;
                                                                                }
                                                                            }
                                                                            else if (diffAFD >= -1) {
                                                                                if (AFD < 3) {
                                                                                    return 8;
                                                                                }
                                                                                else if (AFD >= 3) {
                                                                                    if (Cs < 3069) {
                                                                                        return 1;
                                                                                    }
                                                                                    else if (Cs >= 3069) {
                                                                                        return 8;
                                                                                    }
                                                                                }
                                                                            }
                                                                        }
                                                                        else if (negBalance >= 27) {
                                                                            if (posBalance < 45) {
                                                                                if (RsCsDiff < 58) {
                                                                                    return 1;
                                                                                }
                                                                                else if (RsCsDiff >= 58) {
                                                                                    return 8;
                                                                                }
                                                                            }
                                                                            else if (posBalance >= 45) {
                                                                                return 8;
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (Rs >= 3227) {
                                                                        return 1;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (Rs >= 3394) {
                                                    return 8;
                                                }
                                            }
                                            else if (mu_mv_mag_sq >= 123) {
                                                if (Cs < 3716) {
                                                    if (posBalance < 41) {
                                                        return 8;
                                                    }
                                                    else if (posBalance >= 41) {
                                                        if (CsDiff < 3) {
                                                            return 8;
                                                        }
                                                        else if (CsDiff >= 3) {
                                                            if (Cs < 3010) {
                                                                if (posBalance < 45) {
                                                                    return 1;
                                                                }
                                                                else if (posBalance >= 45) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (Cs >= 3010) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (Cs >= 3716) {
                                                    if (SC < 7605) {
                                                        if (Cs < 3849) {
                                                            return 1;
                                                        }
                                                        else if (Cs >= 3849) {
                                                            if (Rs < 3720) {
                                                                return 8;
                                                            }
                                                            else if (Rs >= 3720) {
                                                                return 1;
                                                            }
                                                        }
                                                    }
                                                    else if (SC >= 7605) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                        }
                                        else if (posBalance >= 48) {
                                            if (negBalance < 35) {
                                                if (posBalance < 50) {
                                                    if (Rs < 1778) {
                                                        if (Rs < 1650) {
                                                            return 8;
                                                        }
                                                        else if (Rs >= 1650) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (Rs >= 1778) {
                                                        return 8;
                                                    }
                                                }
                                                else if (posBalance >= 50) {
                                                    return 8;
                                                }
                                            }
                                            else if (negBalance >= 35) {
                                                if (Rs < 2453) {
                                                    if (SC < 3899) {
                                                        if (Rs < 1780) {
                                                            return 8;
                                                        }
                                                        else if (Rs >= 1780) {
                                                            if (TSC < 4) {
                                                                return 4;
                                                            }
                                                            else if (TSC >= 4) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                    else if (SC >= 3899) {
                                                        return 4;
                                                    }
                                                }
                                                else if (Rs >= 2453) {
                                                    return 8;
                                                }
                                            }
                                        }
                                    }
                                    else if (ssDCval >= 110) {
                                        if (negBalance < 1) {
                                            return 1;
                                        }
                                        else if (negBalance >= 1) {
                                            if (mcTcor < 99) {
                                                if (ssDCval < 119) {
                                                    if (refDCval < 117) {
                                                        return 8;
                                                    }
                                                    else if (refDCval >= 117) {
                                                        if (TSC < 4) {
                                                            return 4;
                                                        }
                                                        else if (TSC >= 4) {
                                                            if (MVDiff < 1) {
                                                                return 4;
                                                            }
                                                            else if (MVDiff >= 1) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (ssDCval >= 119) {
                                                    if (SC < 1614) {
                                                        if (tcor < 95) {
                                                            return 8;
                                                        }
                                                        else if (tcor >= 95) {
                                                            if (AFD < 4) {
                                                                return 1;
                                                            }
                                                            else if (AFD >= 4) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                    else if (SC >= 1614) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                            else if (mcTcor >= 99) {
                                                if (Rs < 1419) {
                                                    if (Rs < 1414) {
                                                        return 8;
                                                    }
                                                    else if (Rs >= 1414) {
                                                        return 4;
                                                    }
                                                }
                                                else if (Rs >= 1419) {
                                                    return 8;
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                        }
                        else if (Rs >= 3774) {
                            return 1;
                        }
                    }
                    else if (Rs >= 3779) {
                        if (mu_mv_mag_sq < 94) {
                            if (ssDCval < 109) {
                                return 8;
                            }
                            else if (ssDCval >= 109) {
                                if (Rs < 4506) {
                                    return 4;
                                }
                                else if (Rs >= 4506) {
                                    return 2;
                                }
                            }
                        }
                        else if (mu_mv_mag_sq >= 94) {
                            if (refDCval < 114) {
                                if (diffTSC < -9) {
                                    if (Cs < 2881) {
                                        return 2;
                                    }
                                    else if (Cs >= 2881) {
                                        return 8;
                                    }
                                }
                                else if (diffTSC >= -9) {
                                    return 8;
                                }
                            }
                            else if (refDCval >= 114) {
                                if (MVDiff < 12) {
                                    if (negBalance < 35) {
                                        return 8;
                                    }
                                    else if (negBalance >= 35) {
                                        if (Rs < 5015) {
                                            return 4;
                                        }
                                        else if (Rs >= 5015) {
                                            return 8;
                                        }
                                    }
                                }
                                else if (MVDiff >= 12) {
                                    return 4;
                                }
                            }
                        }
                    }
                }
            }
            else if (negBalance >= 36) {
                if (Cs < 1675) {
                    if (TSCindex < 7) {
                        if (Rs < 528) {
                            if (AFD < 6) {
                                if (posBalance < 53) {
                                    if (Cs < 1068) {
                                        if (negBalance < 42) {
                                            if (Rs < 408) {
                                                return 2;
                                            }
                                            else if (Rs >= 408) {
                                                return 4;
                                            }
                                        }
                                        else if (negBalance >= 42) {
                                            return 4;
                                        }
                                    }
                                    else if (Cs >= 1068) {
                                        if (Rs < 404) {
                                            return 4;
                                        }
                                        else if (Rs >= 404) {
                                            return 8;
                                        }
                                    }
                                }
                                else if (posBalance >= 53) {
                                    if (RsCsDiff < 2) {
                                        return 8;
                                    }
                                    else if (RsCsDiff >= 2) {
                                        return 2;
                                    }
                                }
                            }
                            else if (AFD >= 6) {
                                if (SC < 327) {
                                    return 2;
                                }
                                else if (SC >= 327) {
                                    return 1;
                                }
                            }
                        }
                        else if (Rs >= 528) {
                            if (AFD < 6) {
                                if (MVDiff < 7) {
                                    if (SC < 3107) {
                                        if (ssDCval < 91) {
                                            if (negBalance < 37) {
                                                if (SC < 3059) {
                                                    return 8;
                                                }
                                                else if (SC >= 3059) {
                                                    return 4;
                                                }
                                            }
                                            else if (negBalance >= 37) {
                                                return 8;
                                            }
                                        }
                                        else if (ssDCval >= 91) {
                                            if (Rs < 623) {
                                                return 8;
                                            }
                                            else if (Rs >= 623) {
                                                return 4;
                                            }
                                        }
                                    }
                                    else if (SC >= 3107) {
                                        if (MVDiff < 3) {
                                            if (Rs < 1821) {
                                                return 8;
                                            }
                                            else if (Rs >= 1821) {
                                                if (refDCval < 82) {
                                                    if (mcTcor < 95) {
                                                        return 8;
                                                    }
                                                    else if (mcTcor >= 95) {
                                                        if (Cs < 1634) {
                                                            if (RsCsDiff < 29) {
                                                                if (negBalance < 41) {
                                                                    return 8;
                                                                }
                                                                else if (negBalance >= 41) {
                                                                    if (Rs < 1879) {
                                                                        if (SC < 3290) {
                                                                            return 4;
                                                                        }
                                                                        else if (SC >= 3290) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                    else if (Rs >= 1879) {
                                                                        return 4;
                                                                    }
                                                                }
                                                            }
                                                            else if (RsCsDiff >= 29) {
                                                                if (SC < 3553) {
                                                                    return 8;
                                                                }
                                                                else if (SC >= 3553) {
                                                                    return 2;
                                                                }
                                                            }
                                                        }
                                                        else if (Cs >= 1634) {
                                                            if (Cs < 1645) {
                                                                return 2;
                                                            }
                                                            else if (Cs >= 1645) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (refDCval >= 82) {
                                                    return 8;
                                                }
                                            }
                                        }
                                        else if (MVDiff >= 3) {
                                            if (Cs < 1415) {
                                                return 4;
                                            }
                                            else if (Cs >= 1415) {
                                                return 8;
                                            }
                                        }
                                    }
                                }
                                else if (MVDiff >= 7) {
                                    if (SC < 2978) {
                                        if (negBalance < 42) {
                                            return 8;
                                        }
                                        else if (negBalance >= 42) {
                                            if (SC < 1674) {
                                                return 4;
                                            }
                                            else if (SC >= 1674) {
                                                if (SC < 2085) {
                                                    return 1;
                                                }
                                                else if (SC >= 2085) {
                                                    return 8;
                                                }
                                            }
                                        }
                                    }
                                    else if (SC >= 2978) {
                                        return 4;
                                    }
                                }
                            }
                            else if (AFD >= 6) {
                                if (tcor < 94) {
                                    if (mu_mv_mag_sq < 192) {
                                        return 8;
                                    }
                                    else if (mu_mv_mag_sq >= 192) {
                                        return 1;
                                    }
                                }
                                else if (tcor >= 94) {
                                    if (Rs < 2387) {
                                        if (diffTSC < -1) {
                                            if (gchDC < 1) {
                                                return 8;
                                            }
                                            else if (gchDC >= 1) {
                                                return 4;
                                            }
                                        }
                                        else if (diffTSC >= -1) {
                                            return 4;
                                        }
                                    }
                                    else if (Rs >= 2387) {
                                        return 1;
                                    }
                                }
                            }
                        }
                    }
                    else if (TSCindex >= 7) {
                        if (refDCval < 115) {
                            if (Cs < 1348) {
                                if (Rs < 1112) {
                                    if (SC < 944) {
                                        if (SC < 789) {
                                            return 2;
                                        }
                                        else if (SC >= 789) {
                                            return 1;
                                        }
                                    }
                                    else if (SC >= 944) {
                                        if (negBalance < 47) {
                                            return 8;
                                        }
                                        else if (negBalance >= 47) {
                                            return 4;
                                        }
                                    }
                                }
                                else if (Rs >= 1112) {
                                    if (RsCsDiff < 23) {
                                        return 2;
                                    }
                                    else if (RsCsDiff >= 23) {
                                        if (negBalance < 50) {
                                            if (refDCval < 69) {
                                                return 2;
                                            }
                                            else if (refDCval >= 69) {
                                                if (Rs < 1265) {
                                                    return 2;
                                                }
                                                else if (Rs >= 1265) {
                                                    return 8;
                                                }
                                            }
                                        }
                                        else if (negBalance >= 50) {
                                            if (diffRsCsdiff < -134) {
                                                return 1;
                                            }
                                            else if (diffRsCsdiff >= -134) {
                                                if (mu_mv_mag_sq < 401) {
                                                    return 2;
                                                }
                                                else if (mu_mv_mag_sq >= 401) {
                                                    return 1;
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                            else if (Cs >= 1348) {
                                if (mcTcor < 94) {
                                    if (TSC < 7) {
                                        if (ssDCval < 72) {
                                            if (tcor < 91) {
                                                return 8;
                                            }
                                            else if (tcor >= 91) {
                                                return 2;
                                            }
                                        }
                                        else if (ssDCval >= 72) {
                                            return 8;
                                        }
                                    }
                                    else if (TSC >= 7) {
                                        if (ssDCval < 84) {
                                            return 2;
                                        }
                                        else if (ssDCval >= 84) {
                                            if (Rs < 2347) {
                                                return 8;
                                            }
                                            else if (Rs >= 2347) {
                                                return 2;
                                            }
                                        }
                                    }
                                }
                                else if (mcTcor >= 94) {
                                    if (posBalance < 56) {
                                        if (Scindex < 5) {
                                            return 4;
                                        }
                                        else if (Scindex >= 5) {
                                            if (Rs < 2786) {
                                                return 1;
                                            }
                                            else if (Rs >= 2786) {
                                                return 8;
                                            }
                                        }
                                    }
                                    else if (posBalance >= 56) {
                                        if (AFD < 7) {
                                            if (CsDiff < 5) {
                                                if (ssDCval < 75) {
                                                    if (SC < 3140) {
                                                        return 8;
                                                    }
                                                    else if (SC >= 3140) {
                                                        return 2;
                                                    }
                                                }
                                                else if (ssDCval >= 75) {
                                                    if (Rs < 1757) {
                                                        if (Cs < 1427) {
                                                            return 8;
                                                        }
                                                        else if (Cs >= 1427) {
                                                            return 2;
                                                        }
                                                    }
                                                    else if (Rs >= 1757) {
                                                        if (refDCval < 75) {
                                                            return 2;
                                                        }
                                                        else if (refDCval >= 75) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (CsDiff >= 5) {
                                                return 8;
                                            }
                                        }
                                        else if (AFD >= 7) {
                                            if (diffAFD < 1) {
                                                return 2;
                                            }
                                            else if (diffAFD >= 1) {
                                                if (Rs < 1828) {
                                                    return 2;
                                                }
                                                else if (Rs >= 1828) {
                                                    if (Rs < 1883) {
                                                        return 4;
                                                    }
                                                    else if (Rs >= 1883) {
                                                        return 2;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                        }
                        else if (refDCval >= 115) {
                            if (MVDiff < 50) {
                                if (RsDiff < 8) {
                                    if (ssDCval < 114) {
                                        return 4;
                                    }
                                    else if (ssDCval >= 114) {
                                        return 8;
                                    }
                                }
                                else if (RsDiff >= 8) {
                                    if (SC < 3639) {
                                        return 8;
                                    }
                                    else if (SC >= 3639) {
                                        return 4;
                                    }
                                }
                            }
                            else if (MVDiff >= 50) {
                                if (diffAFD < 28) {
                                    if (Rs < 2169) {
                                        return 1;
                                    }
                                    else if (Rs >= 2169) {
                                        return 2;
                                    }
                                }
                                else if (diffAFD >= 28) {
                                    return 8;
                                }
                            }
                        }
                    }
                }
                else if (Cs >= 1675) {
                    if (mu_mv_mag_sq < 128) {
                        if (Cs < 2760) {
                            if (ssDCval < 86) {
                                if (mu_mv_mag_sq < 114) {
                                    if (diffAFD < -2) {
                                        if (RsDiff < 8) {
                                            return 8;
                                        }
                                        else if (RsDiff >= 8) {
                                            return 4;
                                        }
                                    }
                                    else if (diffAFD >= -2) {
                                        if (refDCval < 70) {
                                            if (diffRsCsdiff < -14) {
                                                return 4;
                                            }
                                            else if (diffRsCsdiff >= -14) {
                                                return 8;
                                            }
                                        }
                                        else if (refDCval >= 70) {
                                            if (mcTcor < 90) {
                                                return 8;
                                            }
                                            else if (mcTcor >= 90) {
                                                if (Rs < 1760) {
                                                    return 2;
                                                }
                                                else if (Rs >= 1760) {
                                                    if (TSC < 7) {
                                                        return 4;
                                                    }
                                                    else if (TSC >= 7) {
                                                        if (Rs < 2093) {
                                                            return 2;
                                                        }
                                                        else if (Rs >= 2093) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (mu_mv_mag_sq >= 114) {
                                    if (tcor < 95) {
                                        if (Rs < 1904) {
                                            return 8;
                                        }
                                        else if (Rs >= 1904) {
                                            if (MVDiff < 1) {
                                                if (ssDCval < 74) {
                                                    return 8;
                                                }
                                                else if (ssDCval >= 74) {
                                                    return 4;
                                                }
                                            }
                                            else if (MVDiff >= 1) {
                                                return 8;
                                            }
                                        }
                                    }
                                    else if (tcor >= 95) {
                                        if (SC < 3540) {
                                            if (Cs < 1737) {
                                                return 4;
                                            }
                                            else if (Cs >= 1737) {
                                                return 8;
                                            }
                                        }
                                        else if (SC >= 3540) {
                                            return 4;
                                        }
                                    }
                                }
                            }
                            else if (ssDCval >= 86) {
                                if (Rs < 4482) {
                                    if (mcTcor < 97) {
                                        if (mu_mv_mag_sq < 107) {
                                            if (diffRsCsdiff < -25) {
                                                if (Cs < 2604) {
                                                    if (Rs < 2564) {
                                                        return 4;
                                                    }
                                                    else if (Rs >= 2564) {
                                                        if (Rs < 3272) {
                                                            return 8;
                                                        }
                                                        else if (Rs >= 3272) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                                else if (Cs >= 2604) {
                                                    return 1;
                                                }
                                            }
                                            else if (diffRsCsdiff >= -25) {
                                                if (SC < 4700) {
                                                    if (refDCval < 97) {
                                                        if (refDCval < 93) {
                                                            if (mcTcor < 91) {
                                                                return 8;
                                                            }
                                                            else if (mcTcor >= 91) {
                                                                return 2;
                                                            }
                                                        }
                                                        else if (refDCval >= 93) {
                                                            if (diffAFD < 2) {
                                                                return 2;
                                                            }
                                                            else if (diffAFD >= 2) {
                                                                return 1;
                                                            }
                                                        }
                                                    }
                                                    else if (refDCval >= 97) {
                                                        if (refDCval < 112) {
                                                            return 4;
                                                        }
                                                        else if (refDCval >= 112) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                                else if (SC >= 4700) {
                                                    if (ssDCval < 109) {
                                                        if (posBalance < 58) {
                                                            if (SC < 6454) {
                                                                return 8;
                                                            }
                                                            else if (SC >= 6454) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (posBalance >= 58) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (ssDCval >= 109) {
                                                        if (negBalance < 42) {
                                                            return 1;
                                                        }
                                                        else if (negBalance >= 42) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (mu_mv_mag_sq >= 107) {
                                            if (TSC < 11) {
                                                if (refDCval < 108) {
                                                    return 8;
                                                }
                                                else if (refDCval >= 108) {
                                                    if (Rs < 2521) {
                                                        if (ssDCval < 111) {
                                                            return 4;
                                                        }
                                                        else if (ssDCval >= 111) {
                                                            if (Rs < 2388) {
                                                                return 8;
                                                            }
                                                            else if (Rs >= 2388) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                    else if (Rs >= 2521) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                            else if (TSC >= 11) {
                                                if (Cs < 1831) {
                                                    return 2;
                                                }
                                                else if (Cs >= 1831) {
                                                    return 4;
                                                }
                                            }
                                        }
                                    }
                                    else if (mcTcor >= 97) {
                                        if (refDCval < 109) {
                                            if (SC < 3079) {
                                                return 8;
                                            }
                                            else if (SC >= 3079) {
                                                if (CsDiff < 6) {
                                                    return 4;
                                                }
                                                else if (CsDiff >= 6) {
                                                    return 8;
                                                }
                                            }
                                        }
                                        else if (refDCval >= 109) {
                                            return 8;
                                        }
                                    }
                                }
                                else if (Rs >= 4482) {
                                    return 4;
                                }
                            }
                        }
                        else if (Cs >= 2760) {
                            if (ssDCval < 92) {
                                if (ssDCval < 89) {
                                    if (Rs < 4018) {
                                        if (RsCsDiff < 277) {
                                            if (Cs < 3926) {
                                                if (negBalance < 45) {
                                                    return 8;
                                                }
                                                else if (negBalance >= 45) {
                                                    if (TSC < 6) {
                                                        if (refDCval < 56) {
                                                            return 4;
                                                        }
                                                        else if (refDCval >= 56) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (TSC >= 6) {
                                                        if (mu_mv_mag_sq < 98) {
                                                            if (diffMVdiffVal < -2) {
                                                                return 4;
                                                            }
                                                            else if (diffMVdiffVal >= -2) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (mu_mv_mag_sq >= 98) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (Cs >= 3926) {
                                                return 4;
                                            }
                                        }
                                        else if (RsCsDiff >= 277) {
                                            return 4;
                                        }
                                    }
                                    else if (Rs >= 4018) {
                                        if (mu_mv_mag_sq < 126) {
                                            if (refDCval < 66) {
                                                if (SC < 8112) {
                                                    return 8;
                                                }
                                                else if (SC >= 8112) {
                                                    return 4;
                                                }
                                            }
                                            else if (refDCval >= 66) {
                                                return 8;
                                            }
                                        }
                                        else if (mu_mv_mag_sq >= 126) {
                                            if (refDCval < 81) {
                                                if (Cs < 5057) {
                                                    if (SC < 9752) {
                                                        return 8;
                                                    }
                                                    else if (SC >= 9752) {
                                                        if (Cs < 4963) {
                                                            return 4;
                                                        }
                                                        else if (Cs >= 4963) {
                                                            if (CsDiff < 9) {
                                                                return 8;
                                                            }
                                                            else if (CsDiff >= 9) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (Cs >= 5057) {
                                                    return 4;
                                                }
                                            }
                                            else if (refDCval >= 81) {
                                                return 8;
                                            }
                                        }
                                    }
                                }
                                else if (ssDCval >= 89) {
                                    if (RsDiff < 9) {
                                        return 2;
                                    }
                                    else if (RsDiff >= 9) {
                                        if (Rs < 6948) {
                                            if (negBalance < 45) {
                                                return 2;
                                            }
                                            else if (negBalance >= 45) {
                                                return 8;
                                            }
                                        }
                                        else if (Rs >= 6948) {
                                            if (AFD < 19) {
                                                return 2;
                                            }
                                            else if (AFD >= 19) {
                                                if (SC < 15573) {
                                                    return 4;
                                                }
                                                else if (SC >= 15573) {
                                                    if (diffTSC < 1) {
                                                        if (Rs < 8095) {
                                                            if (Cs < 7802) {
                                                                return 4;
                                                            }
                                                            else if (Cs >= 7802) {
                                                                return 2;
                                                            }
                                                        }
                                                        else if (Rs >= 8095) {
                                                            if (Rs < 8366) {
                                                                return 4;
                                                            }
                                                            else if (Rs >= 8366) {
                                                                return 2;
                                                            }
                                                        }
                                                    }
                                                    else if (diffTSC >= 1) {
                                                        return 2;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                            else if (ssDCval >= 92) {
                                if (Cs < 3073) {
                                    if (Rs < 3529) {
                                        if (Cs < 2944) {
                                            if (diffMVdiffVal < -4) {
                                                return 4;
                                            }
                                            else if (diffMVdiffVal >= -4) {
                                                if (CsDiff < 4) {
                                                    if (mu_mv_mag_sq < 120) {
                                                        if (refDCval < 114) {
                                                            return 4;
                                                        }
                                                        else if (refDCval >= 114) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (mu_mv_mag_sq >= 120) {
                                                        return 8;
                                                    }
                                                }
                                                else if (CsDiff >= 4) {
                                                    return 8;
                                                }
                                            }
                                        }
                                        else if (Cs >= 2944) {
                                            if (mu_mv_mag_sq < 123) {
                                                if (tcor < 97) {
                                                    return 8;
                                                }
                                                else if (tcor >= 97) {
                                                    if (Rs < 2712) {
                                                        return 8;
                                                    }
                                                    else if (Rs >= 2712) {
                                                        if (refDCval < 111) {
                                                            return 8;
                                                        }
                                                        else if (refDCval >= 111) {
                                                            if (refDCval < 115) {
                                                                if (posBalance < 57) {
                                                                    return 4;
                                                                }
                                                                else if (posBalance >= 57) {
                                                                    if (Cs < 3004) {
                                                                        if (SC < 5749) {
                                                                            return 4;
                                                                        }
                                                                        else if (SC >= 5749) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                    else if (Cs >= 3004) {
                                                                        if (Rs < 2899) {
                                                                            if (posBalance < 59) {
                                                                                if (CsDiff < 5) {
                                                                                    return 4;
                                                                                }
                                                                                else if (CsDiff >= 5) {
                                                                                    return 8;
                                                                                }
                                                                            }
                                                                            else if (posBalance >= 59) {
                                                                                return 8;
                                                                            }
                                                                        }
                                                                        else if (Rs >= 2899) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                            else if (refDCval >= 115) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                            else if (mu_mv_mag_sq >= 123) {
                                                if (ssDCval < 113) {
                                                    return 8;
                                                }
                                                else if (ssDCval >= 113) {
                                                    if (SC < 5748) {
                                                        return 8;
                                                    }
                                                    else if (SC >= 5748) {
                                                        if (negBalance < 48) {
                                                            return 8;
                                                        }
                                                        else if (negBalance >= 48) {
                                                            if (SC < 5769) {
                                                                return 4;
                                                            }
                                                            else if (SC >= 5769) {
                                                                if (SC < 5814) {
                                                                    return 8;
                                                                }
                                                                else if (SC >= 5814) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (Rs >= 3529) {
                                        if (negBalance < 43) {
                                            if (ssDCval < 115) {
                                                if (RsDiff < 6) {
                                                    return 8;
                                                }
                                                else if (RsDiff >= 6) {
                                                    if (Rs < 5079) {
                                                        return 2;
                                                    }
                                                    else if (Rs >= 5079) {
                                                        if (AFD < 9) {
                                                            return 2;
                                                        }
                                                        else if (AFD >= 9) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (ssDCval >= 115) {
                                                return 4;
                                            }
                                        }
                                        else if (negBalance >= 43) {
                                            if (ssDCval < 110) {
                                                return 4;
                                            }
                                            else if (ssDCval >= 110) {
                                                if (Rs < 3722) {
                                                    return 1;
                                                }
                                                else if (Rs >= 3722) {
                                                    return 4;
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (Cs >= 3073) {
                                    if (tcor < 74) {
                                        if (Rs < 5031) {
                                            return 1;
                                        }
                                        else if (Rs >= 5031) {
                                            return 8;
                                        }
                                    }
                                    else if (tcor >= 74) {
                                        if (refDCval < 122) {
                                            if (ssDCval < 106) {
                                                if (Cs < 3587) {
                                                    if (Rs < 3954) {
                                                        return 8;
                                                    }
                                                    else if (Rs >= 3954) {
                                                        return 4;
                                                    }
                                                }
                                                else if (Cs >= 3587) {
                                                    return 8;
                                                }
                                            }
                                            else if (ssDCval >= 106) {
                                                if (refDCval < 109) {
                                                    if (negBalance < 48) {
                                                        return 8;
                                                    }
                                                    else if (negBalance >= 48) {
                                                        if (Cs < 5289) {
                                                            if (Cs < 3705) {
                                                                return 4;
                                                            }
                                                            else if (Cs >= 3705) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (Cs >= 5289) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                                else if (refDCval >= 109) {
                                                    if (Rs < 4092) {
                                                        if (mu_mv_mag_sq < 123) {
                                                            if (RsCsDiff < 46) {
                                                                if (Rs < 2944) {
                                                                    return 8;
                                                                }
                                                                else if (Rs >= 2944) {
                                                                    if (SC < 6212) {
                                                                        return 4;
                                                                    }
                                                                    else if (SC >= 6212) {
                                                                        return 8;
                                                                    }
                                                                }
                                                            }
                                                            else if (RsCsDiff >= 46) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (mu_mv_mag_sq >= 123) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (Rs >= 4092) {
                                                        if (posBalance < 59) {
                                                            if (SC < 9411) {
                                                                if (SC < 8439) {
                                                                    if (Rs < 4840) {
                                                                        return 4;
                                                                    }
                                                                    else if (Rs >= 4840) {
                                                                        return 2;
                                                                    }
                                                                }
                                                                else if (SC >= 8439) {
                                                                    return 2;
                                                                }
                                                            }
                                                            else if (SC >= 9411) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (posBalance >= 59) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (refDCval >= 122) {
                                            return 8;
                                        }
                                    }
                                }
                            }
                        }
                    }
                    else if (mu_mv_mag_sq >= 128) {
                        if (negBalance < 53) {
                            if (ssDCval < 84) {
                                if (posBalance < 59) {
                                    if (SC < 9973) {
                                        if (diffRsCsdiff < -16) {
                                            if (SC < 9694) {
                                                if (Rs < 4633) {
                                                    return 4;
                                                }
                                                else if (Rs >= 4633) {
                                                    return 8;
                                                }
                                            }
                                            else if (SC >= 9694) {
                                                return 4;
                                            }
                                        }
                                        else if (diffRsCsdiff >= -16) {
                                            if (Cs < 4877) {
                                                return 8;
                                            }
                                            else if (Cs >= 4877) {
                                                if (CsDiff < 9) {
                                                    if (Rs < 4876) {
                                                        return 4;
                                                    }
                                                    else if (Rs >= 4876) {
                                                        return 8;
                                                    }
                                                }
                                                else if (CsDiff >= 9) {
                                                    if (diffAFD < 1) {
                                                        return 8;
                                                    }
                                                    else if (diffAFD >= 1) {
                                                        if (SC < 9824) {
                                                            return 8;
                                                        }
                                                        else if (SC >= 9824) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (SC >= 9973) {
                                        if (diffRsCsdiff < 40) {
                                            return 4;
                                        }
                                        else if (diffRsCsdiff >= 40) {
                                            return 8;
                                        }
                                    }
                                }
                                else if (posBalance >= 59) {
                                    return 4;
                                }
                            }
                            else if (ssDCval >= 84) {
                                if (refDCval < 106) {
                                    if (diffRsCsdiff < -55) {
                                        if (diffRsCsdiff < -65) {
                                            return 8;
                                        }
                                        else if (diffRsCsdiff >= -65) {
                                            return 2;
                                        }
                                    }
                                    else if (diffRsCsdiff >= -55) {
                                        if (posBalance < 61) {
                                            return 8;
                                        }
                                        else if (posBalance >= 61) {
                                            if (mcTcor < 94) {
                                                return 8;
                                            }
                                            else if (mcTcor >= 94) {
                                                return 2;
                                            }
                                        }
                                    }
                                }
                                else if (refDCval >= 106) {
                                    if (tcor < 97) {
                                        if (diffRsCsdiff < -6) {
                                            return 8;
                                        }
                                        else if (diffRsCsdiff >= -6) {
                                            return 4;
                                        }
                                    }
                                    else if (tcor >= 97) {
                                        return 8;
                                    }
                                }
                            }
                        }
                        else if (negBalance >= 53) {
                            if (refDCval < 104) {
                                if (SC < 14769) {
                                    return 8;
                                }
                                else if (SC >= 14769) {
                                    return 2;
                                }
                            }
                            else if (refDCval >= 104) {
                                return 4;
                            }
                        }
                    }
                }
            }
        }
        return 4;
    }
    mfxU16 AGOPSelectC(mfxU32 Rs, mfxU32 CsDiff, mfxU32 TSC, mfxI32 diffAFD, mfxU32 Scindex, mfxU32 RsDiff, mfxI32 diffMVdiffVal, mfxU32 RsCsDiff, mfxU32 TSCindex, mfxI16 tcor, mfxU32 refDCval, mfxU32 MVDiff, mfxU32 AFD, mfxI32 diffTSC, mfxI32 diffRsCsdiff, mfxU32 posBalance, mfxU32 SC, mfxU32 mu_mv_mag_sq, mfxU32 negBalance, mfxI16 mcTcor, mfxU32 gchDC, mfxU32 Cs, mfxU32 ssDCval) {
        if (mu_mv_mag_sq < 89) {
            if (tcor < 1) {
                if (SC < 9327) {
                    if (SC < 3490) {
                        if (AFD < 18) {
                            if (Rs < 366) {
                                return 4;
                            }
                            else if (Rs >= 366) {
                                return 8;
                            }
                        }
                        else if (AFD >= 18) {
                            if (tcor < -4) {
                                return 1;
                            }
                            else if (tcor >= -4) {
                                return 4;
                            }
                        }
                    }
                    else if (SC >= 3490) {
                        if (Rs < 2662) {
                            if (Rs < 1823) {
                                return 4;
                            }
                            else if (Rs >= 1823) {
                                return 1;
                            }
                        }
                        else if (Rs >= 2662) {
                            if (AFD < 27) {
                                if (SC < 7755) {
                                    if (SC < 7709) {
                                        return 8;
                                    }
                                    else if (SC >= 7709) {
                                        return 1;
                                    }
                                }
                                else if (SC >= 7755) {
                                    return 8;
                                }
                            }
                            else if (AFD >= 27) {
                                if (Rs < 4814) {
                                    return 4;
                                }
                                else if (Rs >= 4814) {
                                    return 8;
                                }
                            }
                        }
                    }
                }
                else if (SC >= 9327) {
                    if (Cs < 5084) {
                        return 1;
                    }
                    else if (Cs >= 5084) {
                        return 8;
                    }
                }
            }
            else if (tcor >= 1) {
                if (Scindex < 6) {
                    if (AFD < 12) {
                        if (refDCval < 104) {
                            if (posBalance < 14) {
                                if (Rs < 280) {
                                    if (diffAFD < -1) {
                                        if (MVDiff < 27) {
                                            if (Cs < 189) {
                                                return 4;
                                            }
                                            else if (Cs >= 189) {
                                                return 2;
                                            }
                                        }
                                        else if (MVDiff >= 27) {
                                            if (diffTSC < -1) {
                                                return 2;
                                            }
                                            else if (diffTSC >= -1) {
                                                if (tcor < 96) {
                                                    if (AFD < 2) {
                                                        return 8;
                                                    }
                                                    else if (AFD >= 2) {
                                                        return 1;
                                                    }
                                                }
                                                else if (tcor >= 96) {
                                                    return 2;
                                                }
                                            }
                                        }
                                    }
                                    else if (diffAFD >= -1) {
                                        if (negBalance < 93) {
                                            if (tcor < 102) {
                                                if (posBalance < 2) {
                                                    if (negBalance < 70) {
                                                        return 2;
                                                    }
                                                    else if (negBalance >= 70) {
                                                        return 4;
                                                    }
                                                }
                                                else if (posBalance >= 2) {
                                                    if (diffMVdiffVal < -5) {
                                                        return 4;
                                                    }
                                                    else if (diffMVdiffVal >= -5) {
                                                        if (mcTcor < 24) {
                                                            return 4;
                                                        }
                                                        else if (mcTcor >= 24) {
                                                            if (ssDCval < 19) {
                                                                if (posBalance < 6) {
                                                                    return 8;
                                                                }
                                                                else if (posBalance >= 6) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (ssDCval >= 19) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                            else if (tcor >= 102) {
                                                if (gchDC < 1) {
                                                    if (negBalance < 10) {
                                                        return 8;
                                                    }
                                                    else if (negBalance >= 10) {
                                                        return 4;
                                                    }
                                                }
                                                else if (gchDC >= 1) {
                                                    if (mcTcor < 78) {
                                                        if (ssDCval < 21) {
                                                            if (refDCval < 21) {
                                                                return 4;
                                                            }
                                                            else if (refDCval >= 21) {
                                                                return 1;
                                                            }
                                                        }
                                                        else if (ssDCval >= 21) {
                                                            return 2;
                                                        }
                                                    }
                                                    else if (mcTcor >= 78) {
                                                        if (negBalance < 44) {
                                                            if (Rs < 6) {
                                                                if (diffMVdiffVal < 4) {
                                                                    return 4;
                                                                }
                                                                else if (diffMVdiffVal >= 4) {
                                                                    return 2;
                                                                }
                                                            }
                                                            else if (Rs >= 6) {
                                                                return 2;
                                                            }
                                                        }
                                                        else if (negBalance >= 44) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (negBalance >= 93) {
                                            if (negBalance < 113) {
                                                return 8;
                                            }
                                            else if (negBalance >= 113) {
                                                if (diffMVdiffVal < 166) {
                                                    return 1;
                                                }
                                                else if (diffMVdiffVal >= 166) {
                                                    return 2;
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (Rs >= 280) {
                                    if (mu_mv_mag_sq < 30) {
                                        return 4;
                                    }
                                    else if (mu_mv_mag_sq >= 30) {
                                        if (MVDiff < 82) {
                                            if (AFD < 1) {
                                                return 1;
                                            }
                                            else if (AFD >= 1) {
                                                if (posBalance < 9) {
                                                    return 8;
                                                }
                                                else if (posBalance >= 9) {
                                                    if (gchDC < 3) {
                                                        return 4;
                                                    }
                                                    else if (gchDC >= 3) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                        }
                                        else if (MVDiff >= 82) {
                                            return 2;
                                        }
                                    }
                                }
                            }
                            else if (posBalance >= 14) {
                                if (RsCsDiff < 19) {
                                    if (TSC < 6) {
                                        if (SC < 1112) {
                                            if (negBalance < 49) {
                                                if (Cs < 375) {
                                                    if (mu_mv_mag_sq < 29) {
                                                        if (negBalance < 7) {
                                                            if (mu_mv_mag_sq < 3) {
                                                                if (Rs < 12) {
                                                                    if (gchDC < 2) {
                                                                        return 1;
                                                                    }
                                                                    else if (gchDC >= 2) {
                                                                        return 2;
                                                                    }
                                                                }
                                                                else if (Rs >= 12) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (mu_mv_mag_sq >= 3) {
                                                                if (MVDiff < 87) {
                                                                    if (AFD < 3) {
                                                                        if (ssDCval < 20) {
                                                                            if (diffMVdiffVal < 4) {
                                                                                if (tcor < 36) {
                                                                                    return 8;
                                                                                }
                                                                                else if (tcor >= 36) {
                                                                                    if (MVDiff < 65) {
                                                                                        return 4;
                                                                                    }
                                                                                    else if (MVDiff >= 65) {
                                                                                        return 8;
                                                                                    }
                                                                                }
                                                                            }
                                                                            else if (diffMVdiffVal >= 4) {
                                                                                return 8;
                                                                            }
                                                                        }
                                                                        else if (ssDCval >= 20) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                    else if (AFD >= 3) {
                                                                        if (Rs < 10) {
                                                                            return 2;
                                                                        }
                                                                        else if (Rs >= 10) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                }
                                                                else if (MVDiff >= 87) {
                                                                    if (refDCval < 23) {
                                                                        if (posBalance < 36) {
                                                                            if (tcor < 32) {
                                                                                return 1;
                                                                            }
                                                                            else if (tcor >= 32) {
                                                                                if (posBalance < 23) {
                                                                                    return 2;
                                                                                }
                                                                                else if (posBalance >= 23) {
                                                                                    return 4;
                                                                                }
                                                                            }
                                                                        }
                                                                        else if (posBalance >= 36) {
                                                                            if (mu_mv_mag_sq < 7) {
                                                                                if (diffAFD < 0) {
                                                                                    return 1;
                                                                                }
                                                                                else if (diffAFD >= 0) {
                                                                                    return 8;
                                                                                }
                                                                            }
                                                                            else if (mu_mv_mag_sq >= 7) {
                                                                                if (MVDiff < 190) {
                                                                                    if (mu_mv_mag_sq < 12) {
                                                                                        return 4;
                                                                                    }
                                                                                    else if (mu_mv_mag_sq >= 12) {
                                                                                        return 2;
                                                                                    }
                                                                                }
                                                                                else if (MVDiff >= 190) {
                                                                                    if (MVDiff < 978) {
                                                                                        if (MVDiff < 208) {
                                                                                            return 8;
                                                                                        }
                                                                                        else if (MVDiff >= 208) {
                                                                                            return 4;
                                                                                        }
                                                                                    }
                                                                                    else if (MVDiff >= 978) {
                                                                                        return 2;
                                                                                    }
                                                                                }
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (refDCval >= 23) {
                                                                        return 1;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (negBalance >= 7) {
                                                            if (tcor < 57) {
                                                                return 1;
                                                            }
                                                            else if (tcor >= 57) {
                                                                if (Cs < 316) {
                                                                    if (diffTSC < -1) {
                                                                        return 1;
                                                                    }
                                                                    else if (diffTSC >= -1) {
                                                                        if (ssDCval < 74) {
                                                                            return 2;
                                                                        }
                                                                        else if (ssDCval >= 74) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                }
                                                                else if (Cs >= 316) {
                                                                    if (AFD < 7) {
                                                                        return 4;
                                                                    }
                                                                    else if (AFD >= 7) {
                                                                        return 2;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (mu_mv_mag_sq >= 29) {
                                                        if (mcTcor < 86) {
                                                            if (posBalance < 60) {
                                                                if (SC < 730) {
                                                                    if (posBalance < 28) {
                                                                        if (posBalance < 23) {
                                                                            if (MVDiff < 64) {
                                                                                return 8;
                                                                            }
                                                                            else if (MVDiff >= 64) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                        else if (posBalance >= 23) {
                                                                            if (MVDiff < 62) {
                                                                                return 1;
                                                                            }
                                                                            else if (MVDiff >= 62) {
                                                                                return 8;
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (posBalance >= 28) {
                                                                        if (Rs < 352) {
                                                                            if (tcor < 37) {
                                                                                return 8;
                                                                            }
                                                                            else if (tcor >= 37) {
                                                                                if (diffMVdiffVal < -37) {
                                                                                    return 2;
                                                                                }
                                                                                else if (diffMVdiffVal >= -37) {
                                                                                    if (Rs < 4) {
                                                                                        return 4;
                                                                                    }
                                                                                    else if (Rs >= 4) {
                                                                                        if (Rs < 175) {
                                                                                            return 2;
                                                                                        }
                                                                                        else if (Rs >= 175) {
                                                                                            return 4;
                                                                                        }
                                                                                    }
                                                                                }
                                                                            }
                                                                        }
                                                                        else if (Rs >= 352) {
                                                                            if (Cs < 292) {
                                                                                if (Rs < 370) {
                                                                                    return 1;
                                                                                }
                                                                                else if (Rs >= 370) {
                                                                                    return 4;
                                                                                }
                                                                            }
                                                                            else if (Cs >= 292) {
                                                                                if (MVDiff < 20) {
                                                                                    return 1;
                                                                                }
                                                                                else if (MVDiff >= 20) {
                                                                                    if (Rs < 391) {
                                                                                        return 8;
                                                                                    }
                                                                                    else if (Rs >= 391) {
                                                                                        return 4;
                                                                                    }
                                                                                }
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                                else if (SC >= 730) {
                                                                    if (mcTcor < 84) {
                                                                        if (diffRsCsdiff < 1) {
                                                                            if (MVDiff < 25) {
                                                                                if (MVDiff < 24) {
                                                                                    return 1;
                                                                                }
                                                                                else if (MVDiff >= 24) {
                                                                                    return 4;
                                                                                }
                                                                            }
                                                                            else if (MVDiff >= 25) {
                                                                                return 1;
                                                                            }
                                                                        }
                                                                        else if (diffRsCsdiff >= 1) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                    else if (mcTcor >= 84) {
                                                                        if (Rs < 439) {
                                                                            return 1;
                                                                        }
                                                                        else if (Rs >= 439) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                            else if (posBalance >= 60) {
                                                                if (TSCindex < 1) {
                                                                    if (Cs < 19) {
                                                                        return 4;
                                                                    }
                                                                    else if (Cs >= 19) {
                                                                        return 2;
                                                                    }
                                                                }
                                                                else if (TSCindex >= 1) {
                                                                    if (negBalance < 17) {
                                                                        if (tcor < 48) {
                                                                            return 1;
                                                                        }
                                                                        else if (tcor >= 48) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                    else if (negBalance >= 17) {
                                                                        return 1;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (mcTcor >= 86) {
                                                            if (negBalance < 16) {
                                                                if (posBalance < 99) {
                                                                    if (mu_mv_mag_sq < 35) {
                                                                        return 8;
                                                                    }
                                                                    else if (mu_mv_mag_sq >= 35) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (posBalance >= 99) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (negBalance >= 16) {
                                                                if (AFD < 6) {
                                                                    if (Cs < 164) {
                                                                        return 2;
                                                                    }
                                                                    else if (Cs >= 164) {
                                                                        if (ssDCval < 38) {
                                                                            if (Rs < 400) {
                                                                                return 8;
                                                                            }
                                                                            else if (Rs >= 400) {
                                                                                if (Rs < 431) {
                                                                                    return 4;
                                                                                }
                                                                                else if (Rs >= 431) {
                                                                                    return 2;
                                                                                }
                                                                            }
                                                                        }
                                                                        else if (ssDCval >= 38) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                }
                                                                else if (AFD >= 6) {
                                                                    return 2;
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (Cs >= 375) {
                                                    if (ssDCval < 39) {
                                                        if (SC < 860) {
                                                            if (tcor < 81) {
                                                                return 4;
                                                            }
                                                            else if (tcor >= 81) {
                                                                if (Rs < 450) {
                                                                    if (Cs < 385) {
                                                                        if (MVDiff < 27) {
                                                                            return 4;
                                                                        }
                                                                        else if (MVDiff >= 27) {
                                                                            return 2;
                                                                        }
                                                                    }
                                                                    else if (Cs >= 385) {
                                                                        return 2;
                                                                    }
                                                                }
                                                                else if (Rs >= 450) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                        else if (SC >= 860) {
                                                            if (MVDiff < 21) {
                                                                if (diffMVdiffVal < -6) {
                                                                    return 2;
                                                                }
                                                                else if (diffMVdiffVal >= -6) {
                                                                    if (Cs < 396) {
                                                                        return 2;
                                                                    }
                                                                    else if (Cs >= 396) {
                                                                        return 4;
                                                                    }
                                                                }
                                                            }
                                                            else if (MVDiff >= 21) {
                                                                if (negBalance < 47) {
                                                                    return 2;
                                                                }
                                                                else if (negBalance >= 47) {
                                                                    if (Rs < 491) {
                                                                        return 2;
                                                                    }
                                                                    else if (Rs >= 491) {
                                                                        return 8;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (ssDCval >= 39) {
                                                        if (Rs < 268) {
                                                            return 2;
                                                        }
                                                        else if (Rs >= 268) {
                                                            if (TSC < 3) {
                                                                if (Cs < 542) {
                                                                    return 8;
                                                                }
                                                                else if (Cs >= 542) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (TSC >= 3) {
                                                                if (TSC < 5) {
                                                                    return 4;
                                                                }
                                                                else if (TSC >= 5) {
                                                                    return 2;
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                            else if (negBalance >= 49) {
                                                if (SC < 781) {
                                                    if (tcor < 85) {
                                                        if (ssDCval < 86) {
                                                            if (mcTcor < 93) {
                                                                return 1;
                                                            }
                                                            else if (mcTcor >= 93) {
                                                                return 2;
                                                            }
                                                        }
                                                        else if (ssDCval >= 86) {
                                                            if (diffRsCsdiff < -2) {
                                                                if (Cs < 69) {
                                                                    return 1;
                                                                }
                                                                else if (Cs >= 69) {
                                                                    return 2;
                                                                }
                                                            }
                                                            else if (diffRsCsdiff >= -2) {
                                                                if (posBalance < 58) {
                                                                    if (negBalance < 50) {
                                                                        if (Cs < 286) {
                                                                            return 2;
                                                                        }
                                                                        else if (Cs >= 286) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                    else if (negBalance >= 50) {
                                                                        return 2;
                                                                    }
                                                                }
                                                                else if (posBalance >= 58) {
                                                                    return 2;
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (tcor >= 85) {
                                                        if (AFD < 5) {
                                                            if (Rs < 19) {
                                                                return 1;
                                                            }
                                                            else if (Rs >= 19) {
                                                                if (SC < 417) {
                                                                    return 2;
                                                                }
                                                                else if (SC >= 417) {
                                                                    if (Cs < 308) {
                                                                        return 4;
                                                                    }
                                                                    else if (Cs >= 308) {
                                                                        return 8;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (AFD >= 5) {
                                                            return 1;
                                                        }
                                                    }
                                                }
                                                else if (SC >= 781) {
                                                    if (tcor < 69) {
                                                        if (Rs < 500) {
                                                            return 4;
                                                        }
                                                        else if (Rs >= 500) {
                                                            return 2;
                                                        }
                                                    }
                                                    else if (tcor >= 69) {
                                                        if (Rs < 459) {
                                                            return 4;
                                                        }
                                                        else if (Rs >= 459) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (SC >= 1112) {
                                            if (Rs < 1132) {
                                                if (mu_mv_mag_sq < 28) {
                                                    if (tcor < 75) {
                                                        return 4;
                                                    }
                                                    else if (tcor >= 75) {
                                                        if (Rs < 629) {
                                                            return 8;
                                                        }
                                                        else if (Rs >= 629) {
                                                            if (CsDiff < 3) {
                                                                if (negBalance < 39) {
                                                                    if (SC < 1502) {
                                                                        return 8;
                                                                    }
                                                                    else if (SC >= 1502) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (negBalance >= 39) {
                                                                    if (MVDiff < 9) {
                                                                        return 4;
                                                                    }
                                                                    else if (MVDiff >= 9) {
                                                                        if (posBalance < 60) {
                                                                            return 8;
                                                                        }
                                                                        else if (posBalance >= 60) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                            else if (CsDiff >= 3) {
                                                                return 2;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (mu_mv_mag_sq >= 28) {
                                                    if (tcor < 84) {
                                                        if (SC < 1511) {
                                                            if (posBalance < 58) {
                                                                if (CsDiff < 2) {
                                                                    return 4;
                                                                }
                                                                else if (CsDiff >= 2) {
                                                                    if (tcor < 77) {
                                                                        if (MVDiff < 26) {
                                                                            return 4;
                                                                        }
                                                                        else if (MVDiff >= 26) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                    else if (tcor >= 77) {
                                                                        return 4;
                                                                    }
                                                                }
                                                            }
                                                            else if (posBalance >= 58) {
                                                                if (Cs < 615) {
                                                                    if (AFD < 7) {
                                                                        if (mu_mv_mag_sq < 42) {
                                                                            return 1;
                                                                        }
                                                                        else if (mu_mv_mag_sq >= 42) {
                                                                            if (mcTcor < 81) {
                                                                                return 4;
                                                                            }
                                                                            else if (mcTcor >= 81) {
                                                                                if (Cs < 587) {
                                                                                    return 1;
                                                                                }
                                                                                else if (Cs >= 587) {
                                                                                    return 4;
                                                                                }
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (AFD >= 7) {
                                                                        if (SC < 1300) {
                                                                            return 4;
                                                                        }
                                                                        else if (SC >= 1300) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                }
                                                                else if (Cs >= 615) {
                                                                    if (Cs < 717) {
                                                                        return 4;
                                                                    }
                                                                    else if (Cs >= 717) {
                                                                        return 1;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (SC >= 1511) {
                                                            if (ssDCval < 46) {
                                                                if (TSC < 3) {
                                                                    return 1;
                                                                }
                                                                else if (TSC >= 3) {
                                                                    if (MVDiff < 28) {
                                                                        return 2;
                                                                    }
                                                                    else if (MVDiff >= 28) {
                                                                        if (Rs < 973) {
                                                                            return 1;
                                                                        }
                                                                        else if (Rs >= 973) {
                                                                            return 2;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                            else if (ssDCval >= 46) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                    else if (tcor >= 84) {
                                                        if (mcTcor < 93) {
                                                            if (ssDCval < 54) {
                                                                if (mu_mv_mag_sq < 46) {
                                                                    if (RsDiff < 2) {
                                                                        return 2;
                                                                    }
                                                                    else if (RsDiff >= 2) {
                                                                        if (posBalance < 60) {
                                                                            if (mu_mv_mag_sq < 37) {
                                                                                return 1;
                                                                            }
                                                                            else if (mu_mv_mag_sq >= 37) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                        else if (posBalance >= 60) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                }
                                                                else if (mu_mv_mag_sq >= 46) {
                                                                    if (Rs < 984) {
                                                                        if (Rs < 734) {
                                                                            return 8;
                                                                        }
                                                                        else if (Rs >= 734) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                    else if (Rs >= 984) {
                                                                        if (negBalance < 43) {
                                                                            return 4;
                                                                        }
                                                                        else if (negBalance >= 43) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                            else if (ssDCval >= 54) {
                                                                if (posBalance < 48) {
                                                                    if (Cs < 735) {
                                                                        return 8;
                                                                    }
                                                                    else if (Cs >= 735) {
                                                                        return 1;
                                                                    }
                                                                }
                                                                else if (posBalance >= 48) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                        else if (mcTcor >= 93) {
                                                            if (Rs < 928) {
                                                                if (TSC < 4) {
                                                                    if (posBalance < 50) {
                                                                        return 4;
                                                                    }
                                                                    else if (posBalance >= 50) {
                                                                        if (Cs < 620) {
                                                                            return 8;
                                                                        }
                                                                        else if (Cs >= 620) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                }
                                                                else if (TSC >= 4) {
                                                                    if (diffRsCsdiff < 4) {
                                                                        if (SC < 1755) {
                                                                            if (diffAFD < 1) {
                                                                                if (diffMVdiffVal < 8) {
                                                                                    if (mu_mv_mag_sq < 38) {
                                                                                        return 8;
                                                                                    }
                                                                                    else if (mu_mv_mag_sq >= 38) {
                                                                                        return 4;
                                                                                    }
                                                                                }
                                                                                else if (diffMVdiffVal >= 8) {
                                                                                    return 8;
                                                                                }
                                                                            }
                                                                            else if (diffAFD >= 1) {
                                                                                return 8;
                                                                            }
                                                                        }
                                                                        else if (SC >= 1755) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                    else if (diffRsCsdiff >= 4) {
                                                                        return 2;
                                                                    }
                                                                }
                                                            }
                                                            else if (Rs >= 928) {
                                                                if (negBalance < 39) {
                                                                    if (Cs < 441) {
                                                                        return 8;
                                                                    }
                                                                    else if (Cs >= 441) {
                                                                        if (refDCval < 49) {
                                                                            if (diffAFD < -1) {
                                                                                return 1;
                                                                            }
                                                                            else if (diffAFD >= -1) {
                                                                                if (Rs < 1074) {
                                                                                    if (gchDC < 1) {
                                                                                        return 4;
                                                                                    }
                                                                                    else if (gchDC >= 1) {
                                                                                        if (MVDiff < 5) {
                                                                                            return 1;
                                                                                        }
                                                                                        else if (MVDiff >= 5) {
                                                                                            return 4;
                                                                                        }
                                                                                    }
                                                                                }
                                                                                else if (Rs >= 1074) {
                                                                                    return 1;
                                                                                }
                                                                            }
                                                                        }
                                                                        else if (refDCval >= 49) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                }
                                                                else if (negBalance >= 39) {
                                                                    if (ssDCval < 46) {
                                                                        return 8;
                                                                    }
                                                                    else if (ssDCval >= 46) {
                                                                        if (TSC < 5) {
                                                                            if (Rs < 1042) {
                                                                                if (SC < 1645) {
                                                                                    return 4;
                                                                                }
                                                                                else if (SC >= 1645) {
                                                                                    return 1;
                                                                                }
                                                                            }
                                                                            else if (Rs >= 1042) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                        else if (TSC >= 5) {
                                                                            if (Cs < 607) {
                                                                                return 8;
                                                                            }
                                                                            else if (Cs >= 607) {
                                                                                return 1;
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                            else if (Rs >= 1132) {
                                                if (SC < 2599) {
                                                    if (refDCval < 39) {
                                                        return 2;
                                                    }
                                                    else if (refDCval >= 39) {
                                                        if (Rs < 1374) {
                                                            if (gchDC < 2) {
                                                                if (mu_mv_mag_sq < 48) {
                                                                    return 1;
                                                                }
                                                                else if (mu_mv_mag_sq >= 48) {
                                                                    if (Rs < 1202) {
                                                                        return 4;
                                                                    }
                                                                    else if (Rs >= 1202) {
                                                                        if (AFD < 7) {
                                                                            if (refDCval < 56) {
                                                                                if (SC < 2316) {
                                                                                    if (MVDiff < 4) {
                                                                                        if (AFD < 6) {
                                                                                            return 1;
                                                                                        }
                                                                                        else if (AFD >= 6) {
                                                                                            return 4;
                                                                                        }
                                                                                    }
                                                                                    else if (MVDiff >= 4) {
                                                                                        return 1;
                                                                                    }
                                                                                }
                                                                                else if (SC >= 2316) {
                                                                                    if (negBalance < 35) {
                                                                                        return 1;
                                                                                    }
                                                                                    else if (negBalance >= 35) {
                                                                                        if (RsDiff < 3) {
                                                                                            return 4;
                                                                                        }
                                                                                        else if (RsDiff >= 3) {
                                                                                            return 1;
                                                                                        }
                                                                                    }
                                                                                }
                                                                            }
                                                                            else if (refDCval >= 56) {
                                                                                if (AFD < 6) {
                                                                                    return 8;
                                                                                }
                                                                                else if (AFD >= 6) {
                                                                                    return 1;
                                                                                }
                                                                            }
                                                                        }
                                                                        else if (AFD >= 7) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                            else if (gchDC >= 2) {
                                                                if (TSC < 5) {
                                                                    return 4;
                                                                }
                                                                else if (TSC >= 5) {
                                                                    return 1;
                                                                }
                                                            }
                                                        }
                                                        else if (Rs >= 1374) {
                                                            if (refDCval < 59) {
                                                                if (diffMVdiffVal < -2) {
                                                                    if (diffTSC < 1) {
                                                                        return 1;
                                                                    }
                                                                    else if (diffTSC >= 1) {
                                                                        return 8;
                                                                    }
                                                                }
                                                                else if (diffMVdiffVal >= -2) {
                                                                    if (mcTcor < 94) {
                                                                        if (posBalance < 49) {
                                                                            if (Rs < 1441) {
                                                                                return 4;
                                                                            }
                                                                            else if (Rs >= 1441) {
                                                                                return 1;
                                                                            }
                                                                        }
                                                                        else if (posBalance >= 49) {
                                                                            if (SC < 2341) {
                                                                                return 4;
                                                                            }
                                                                            else if (SC >= 2341) {
                                                                                if (diffTSC < -1) {
                                                                                    if (Rs < 1456) {
                                                                                        return 1;
                                                                                    }
                                                                                    else if (Rs >= 1456) {
                                                                                        return 8;
                                                                                    }
                                                                                }
                                                                                else if (diffTSC >= -1) {
                                                                                    return 8;
                                                                                }
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (mcTcor >= 94) {
                                                                        if (diffTSC < -1) {
                                                                            return 4;
                                                                        }
                                                                        else if (diffTSC >= -1) {
                                                                            if (mu_mv_mag_sq < 71) {
                                                                                return 1;
                                                                            }
                                                                            else if (mu_mv_mag_sq >= 71) {
                                                                                if (diffMVdiffVal < 1) {
                                                                                    if (Rs < 1437) {
                                                                                        return 4;
                                                                                    }
                                                                                    else if (Rs >= 1437) {
                                                                                        return 1;
                                                                                    }
                                                                                }
                                                                                else if (diffMVdiffVal >= 1) {
                                                                                    if (TSC < 2) {
                                                                                        return 1;
                                                                                    }
                                                                                    else if (TSC >= 2) {
                                                                                        return 8;
                                                                                    }
                                                                                }
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                            else if (refDCval >= 59) {
                                                                if (Cs < 775) {
                                                                    if (ssDCval < 87) {
                                                                        return 4;
                                                                    }
                                                                    else if (ssDCval >= 87) {
                                                                        return 8;
                                                                    }
                                                                }
                                                                else if (Cs >= 775) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (SC >= 2599) {
                                                    if (Cs < 1552) {
                                                        if (Cs < 1434) {
                                                            if (ssDCval < 72) {
                                                                if (negBalance < 32) {
                                                                    return 1;
                                                                }
                                                                else if (negBalance >= 32) {
                                                                    if (diffMVdiffVal < -2) {
                                                                        if (refDCval < 62) {
                                                                            if (RsDiff < 3) {
                                                                                if (Rs < 1586) {
                                                                                    return 2;
                                                                                }
                                                                                else if (Rs >= 1586) {
                                                                                    return 8;
                                                                                }
                                                                            }
                                                                            else if (RsDiff >= 3) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                        else if (refDCval >= 62) {
                                                                            return 2;
                                                                        }
                                                                    }
                                                                    else if (diffMVdiffVal >= -2) {
                                                                        return 2;
                                                                    }
                                                                }
                                                            }
                                                            else if (ssDCval >= 72) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (Cs >= 1434) {
                                                            if (MVDiff < 42) {
                                                                if (diffMVdiffVal < -10) {
                                                                    return 1;
                                                                }
                                                                else if (diffMVdiffVal >= -10) {
                                                                    if (Cs < 1538) {
                                                                        if (posBalance < 56) {
                                                                            if (MVDiff < 21) {
                                                                                return 4;
                                                                            }
                                                                            else if (MVDiff >= 21) {
                                                                                if (Rs < 1904) {
                                                                                    return 8;
                                                                                }
                                                                                else if (Rs >= 1904) {
                                                                                    return 4;
                                                                                }
                                                                            }
                                                                        }
                                                                        else if (posBalance >= 56) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                    else if (Cs >= 1538) {
                                                                        if (mu_mv_mag_sq < 64) {
                                                                            if (negBalance < 46) {
                                                                                return 1;
                                                                            }
                                                                            else if (negBalance >= 46) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                        else if (mu_mv_mag_sq >= 64) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                            else if (MVDiff >= 42) {
                                                                return 1;
                                                            }
                                                        }
                                                    }
                                                    else if (Cs >= 1552) {
                                                        if (Rs < 1987) {
                                                            if (negBalance < 50) {
                                                                if (Rs < 1964) {
                                                                    return 1;
                                                                }
                                                                else if (Rs >= 1964) {
                                                                    if (Rs < 1966) {
                                                                        return 4;
                                                                    }
                                                                    else if (Rs >= 1966) {
                                                                        return 1;
                                                                    }
                                                                }
                                                            }
                                                            else if (negBalance >= 50) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (Rs >= 1987) {
                                                            if (Rs < 2017) {
                                                                return 4;
                                                            }
                                                            else if (Rs >= 2017) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (TSC >= 6) {
                                        if (tcor < 81) {
                                            if (MVDiff < 19) {
                                                return 4;
                                            }
                                            else if (MVDiff >= 19) {
                                                if (Rs < 628) {
                                                    return 1;
                                                }
                                                else if (Rs >= 628) {
                                                    if (AFD < 9) {
                                                        if (MVDiff < 47) {
                                                            if (mu_mv_mag_sq < 43) {
                                                                if (Rs < 794) {
                                                                    if (tcor < 73) {
                                                                        if (RsDiff < 3) {
                                                                            if (Rs < 744) {
                                                                                if (Rs < 722) {
                                                                                    return 4;
                                                                                }
                                                                                else if (Rs >= 722) {
                                                                                    return 1;
                                                                                }
                                                                            }
                                                                            else if (Rs >= 744) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                        else if (RsDiff >= 3) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                    else if (tcor >= 73) {
                                                                        return 1;
                                                                    }
                                                                }
                                                                else if (Rs >= 794) {
                                                                    return 1;
                                                                }
                                                            }
                                                            else if (mu_mv_mag_sq >= 43) {
                                                                if (MVDiff < 24) {
                                                                    return 1;
                                                                }
                                                                else if (MVDiff >= 24) {
                                                                    if (RsCsDiff < 11) {
                                                                        return 4;
                                                                    }
                                                                    else if (RsCsDiff >= 11) {
                                                                        return 1;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (MVDiff >= 47) {
                                                            if (SC < 1280) {
                                                                return 4;
                                                            }
                                                            else if (SC >= 1280) {
                                                                if (Rs < 767) {
                                                                    return 1;
                                                                }
                                                                else if (Rs >= 767) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (AFD >= 9) {
                                                        if (Rs < 982) {
                                                            return 8;
                                                        }
                                                        else if (Rs >= 982) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (tcor >= 81) {
                                            if (Cs < 251) {
                                                return 1;
                                            }
                                            else if (Cs >= 251) {
                                                if (mu_mv_mag_sq < 76) {
                                                    if (RsCsDiff < 12) {
                                                        if (Rs < 690) {
                                                            if (AFD < 8) {
                                                                return 1;
                                                            }
                                                            else if (AFD >= 8) {
                                                                return 2;
                                                            }
                                                        }
                                                        else if (Rs >= 690) {
                                                            if (posBalance < 95) {
                                                                return 1;
                                                            }
                                                            else if (posBalance >= 95) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                    else if (RsCsDiff >= 12) {
                                                        if (Rs < 730) {
                                                            return 8;
                                                        }
                                                        else if (Rs >= 730) {
                                                            if (ssDCval < 77) {
                                                                if (Rs < 2129) {
                                                                    if (mcTcor < 94) {
                                                                        if (MVDiff < 5) {
                                                                            return 2;
                                                                        }
                                                                        else if (MVDiff >= 5) {
                                                                            if (diffRsCsdiff < 0) {
                                                                                return 8;
                                                                            }
                                                                            else if (diffRsCsdiff >= 0) {
                                                                                return 1;
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (mcTcor >= 94) {
                                                                        return 1;
                                                                    }
                                                                }
                                                                else if (Rs >= 2129) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (ssDCval >= 77) {
                                                                if (Cs < 479) {
                                                                    return 1;
                                                                }
                                                                else if (Cs >= 479) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (mu_mv_mag_sq >= 76) {
                                                    if (Rs < 871) {
                                                        return 4;
                                                    }
                                                    else if (Rs >= 871) {
                                                        return 2;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (RsCsDiff >= 19) {
                                    if (negBalance < 49) {
                                        if (TSC < 8) {
                                            if (refDCval < 75) {
                                                if (Rs < 1018) {
                                                    if (Cs < 422) {
                                                        if (AFD < 10) {
                                                            return 1;
                                                        }
                                                        else if (AFD >= 10) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (Cs >= 422) {
                                                        if (MVDiff < 16) {
                                                            if (negBalance < 43) {
                                                                if (Rs < 460) {
                                                                    return 2;
                                                                }
                                                                else if (Rs >= 460) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (negBalance >= 43) {
                                                                if (SC < 1684) {
                                                                    return 2;
                                                                }
                                                                else if (SC >= 1684) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                        else if (MVDiff >= 16) {
                                                            if (SC < 1244) {
                                                                return 8;
                                                            }
                                                            else if (SC >= 1244) {
                                                                return 2;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (Rs >= 1018) {
                                                    if (refDCval < 65) {
                                                        if (posBalance < 48) {
                                                            if (RsCsDiff < 233) {
                                                                return 4;
                                                            }
                                                            else if (RsCsDiff >= 233) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (posBalance >= 48) {
                                                            if (Rs < 1957) {
                                                                if (negBalance < 31) {
                                                                    return 4;
                                                                }
                                                                else if (negBalance >= 31) {
                                                                    if (MVDiff < 29) {
                                                                        if (mu_mv_mag_sq < 56) {
                                                                            if (ssDCval < 54) {
                                                                                if (Cs < 1022) {
                                                                                    return 1;
                                                                                }
                                                                                else if (Cs >= 1022) {
                                                                                    if (SC < 2527) {
                                                                                        if (ssDCval < 52) {
                                                                                            return 4;
                                                                                        }
                                                                                        else if (ssDCval >= 52) {
                                                                                            return 8;
                                                                                        }
                                                                                    }
                                                                                    else if (SC >= 2527) {
                                                                                        return 1;
                                                                                    }
                                                                                }
                                                                            }
                                                                            else if (ssDCval >= 54) {
                                                                                if (SC < 2484) {
                                                                                    return 4;
                                                                                }
                                                                                else if (SC >= 2484) {
                                                                                    if (AFD < 8) {
                                                                                        return 8;
                                                                                    }
                                                                                    else if (AFD >= 8) {
                                                                                        return 1;
                                                                                    }
                                                                                }
                                                                            }
                                                                        }
                                                                        else if (mu_mv_mag_sq >= 56) {
                                                                            if (mu_mv_mag_sq < 70) {
                                                                                return 4;
                                                                            }
                                                                            else if (mu_mv_mag_sq >= 70) {
                                                                                return 1;
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (MVDiff >= 29) {
                                                                        return 8;
                                                                    }
                                                                }
                                                            }
                                                            else if (Rs >= 1957) {
                                                                if (SC < 4950) {
                                                                    return 4;
                                                                }
                                                                else if (SC >= 4950) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (refDCval >= 65) {
                                                        if (Rs < 1791) {
                                                            return 4;
                                                        }
                                                        else if (Rs >= 1791) {
                                                            if (Cs < 841) {
                                                                return 2;
                                                            }
                                                            else if (Cs >= 841) {
                                                                return 1;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                            else if (refDCval >= 75) {
                                                if (AFD < 9) {
                                                    if (negBalance < 38) {
                                                        if (mcTcor < 94) {
                                                            return 4;
                                                        }
                                                        else if (mcTcor >= 94) {
                                                            if (Rs < 1659) {
                                                                if (Rs < 1324) {
                                                                    return 2;
                                                                }
                                                                else if (Rs >= 1324) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (Rs >= 1659) {
                                                                if (Cs < 1493) {
                                                                    return 1;
                                                                }
                                                                else if (Cs >= 1493) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (negBalance >= 38) {
                                                        if (posBalance < 56) {
                                                            if (Rs < 2262) {
                                                                return 4;
                                                            }
                                                            else if (Rs >= 2262) {
                                                                return 1;
                                                            }
                                                        }
                                                        else if (posBalance >= 56) {
                                                            if (TSCindex < 8) {
                                                                return 4;
                                                            }
                                                            else if (TSCindex >= 8) {
                                                                if (mu_mv_mag_sq < 87) {
                                                                    if (MVDiff < 2) {
                                                                        if (Rs < 1879) {
                                                                            return 2;
                                                                        }
                                                                        else if (Rs >= 1879) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                    else if (MVDiff >= 2) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (mu_mv_mag_sq >= 87) {
                                                                    return 2;
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (AFD >= 9) {
                                                    if (MVDiff < 9) {
                                                        if (refDCval < 82) {
                                                            if (Rs < 1755) {
                                                                return 2;
                                                            }
                                                            else if (Rs >= 1755) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (refDCval >= 82) {
                                                            if (negBalance < 43) {
                                                                if (mcTcor < 96) {
                                                                    if (SC < 2767) {
                                                                        return 2;
                                                                    }
                                                                    else if (SC >= 2767) {
                                                                        if (Cs < 1915) {
                                                                            if (Rs < 2449) {
                                                                                return 8;
                                                                            }
                                                                            else if (Rs >= 2449) {
                                                                                return 1;
                                                                            }
                                                                        }
                                                                        else if (Cs >= 1915) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                }
                                                                else if (mcTcor >= 96) {
                                                                    if (ssDCval < 101) {
                                                                        return 4;
                                                                    }
                                                                    else if (ssDCval >= 101) {
                                                                        return 2;
                                                                    }
                                                                }
                                                            }
                                                            else if (negBalance >= 43) {
                                                                return 2;
                                                            }
                                                        }
                                                    }
                                                    else if (MVDiff >= 9) {
                                                        if (SC < 5091) {
                                                            if (posBalance < 48) {
                                                                if (ssDCval < 88) {
                                                                    return 1;
                                                                }
                                                                else if (ssDCval >= 88) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (posBalance >= 48) {
                                                                if (RsCsDiff < 63) {
                                                                    return 8;
                                                                }
                                                                else if (RsCsDiff >= 63) {
                                                                    if (Cs < 2298) {
                                                                        return 4;
                                                                    }
                                                                    else if (Cs >= 2298) {
                                                                        return 8;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (SC >= 5091) {
                                                            return 2;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (TSC >= 8) {
                                            if (tcor < 94) {
                                                if (ssDCval < 66) {
                                                    if (Rs < 924) {
                                                        return 2;
                                                    }
                                                    else if (Rs >= 924) {
                                                        return 4;
                                                    }
                                                }
                                                else if (ssDCval >= 66) {
                                                    if (ssDCval < 82) {
                                                        return 8;
                                                    }
                                                    else if (ssDCval >= 82) {
                                                        if (diffMVdiffVal < 3) {
                                                            if (mu_mv_mag_sq < 60) {
                                                                if (refDCval < 88) {
                                                                    return 1;
                                                                }
                                                                else if (refDCval >= 88) {
                                                                    if (negBalance < 45) {
                                                                        return 8;
                                                                    }
                                                                    else if (negBalance >= 45) {
                                                                        return 1;
                                                                    }
                                                                }
                                                            }
                                                            else if (mu_mv_mag_sq >= 60) {
                                                                if (posBalance < 56) {
                                                                    return 1;
                                                                }
                                                                else if (posBalance >= 56) {
                                                                    if (Rs < 2655) {
                                                                        if (ssDCval < 83) {
                                                                            return 4;
                                                                        }
                                                                        else if (ssDCval >= 83) {
                                                                            return 2;
                                                                        }
                                                                    }
                                                                    else if (Rs >= 2655) {
                                                                        if (AFD < 11) {
                                                                            return 4;
                                                                        }
                                                                        else if (AFD >= 11) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (diffMVdiffVal >= 3) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (tcor >= 94) {
                                                if (AFD < 9) {
                                                    return 4;
                                                }
                                                else if (AFD >= 9) {
                                                    if (Rs < 2351) {
                                                        if (Rs < 2162) {
                                                            return 4;
                                                        }
                                                        else if (Rs >= 2162) {
                                                            return 2;
                                                        }
                                                    }
                                                    else if (Rs >= 2351) {
                                                        return 1;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (negBalance >= 49) {
                                        if (refDCval < 98) {
                                            if (mcTcor < 91) {
                                                if (SC < 3760) {
                                                    if (mu_mv_mag_sq < 3) {
                                                        if (posBalance < 59) {
                                                            if (mcTcor < 84) {
                                                                return 8;
                                                            }
                                                            else if (mcTcor >= 84) {
                                                                if (TSCindex < 8) {
                                                                    if (diffRsCsdiff < -6) {
                                                                        return 8;
                                                                    }
                                                                    else if (diffRsCsdiff >= -6) {
                                                                        return 2;
                                                                    }
                                                                }
                                                                else if (TSCindex >= 8) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                        else if (posBalance >= 59) {
                                                            if (ssDCval < 63) {
                                                                return 8;
                                                            }
                                                            else if (ssDCval >= 63) {
                                                                if (mcTcor < 81) {
                                                                    if (Cs < 1233) {
                                                                        return 4;
                                                                    }
                                                                    else if (Cs >= 1233) {
                                                                        return 8;
                                                                    }
                                                                }
                                                                else if (mcTcor >= 81) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (mu_mv_mag_sq >= 3) {
                                                        if (Cs < 1100) {
                                                            if (Cs < 905) {
                                                                if (ssDCval < 66) {
                                                                    if (diffMVdiffVal < 7) {
                                                                        return 8;
                                                                    }
                                                                    else if (diffMVdiffVal >= 7) {
                                                                        return 2;
                                                                    }
                                                                }
                                                                else if (ssDCval >= 66) {
                                                                    return 2;
                                                                }
                                                            }
                                                            else if (Cs >= 905) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (Cs >= 1100) {
                                                            if (tcor < 82) {
                                                                if (mu_mv_mag_sq < 52) {
                                                                    return 8;
                                                                }
                                                                else if (mu_mv_mag_sq >= 52) {
                                                                    if (mu_mv_mag_sq < 58) {
                                                                        if (MVDiff < 5) {
                                                                            return 8;
                                                                        }
                                                                        else if (MVDiff >= 5) {
                                                                            return 2;
                                                                        }
                                                                    }
                                                                    else if (mu_mv_mag_sq >= 58) {
                                                                        if (SC < 3597) {
                                                                            if (Rs < 1811) {
                                                                                return 8;
                                                                            }
                                                                            else if (Rs >= 1811) {
                                                                                return 2;
                                                                            }
                                                                        }
                                                                        else if (SC >= 3597) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                            else if (tcor >= 82) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (SC >= 3760) {
                                                    if (refDCval < 83) {
                                                        if (SC < 4780) {
                                                            return 8;
                                                        }
                                                        else if (SC >= 4780) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (refDCval >= 83) {
                                                        if (RsCsDiff < 46) {
                                                            return 8;
                                                        }
                                                        else if (RsCsDiff >= 46) {
                                                            if (ssDCval < 93) {
                                                                if (RsDiff < 7) {
                                                                    return 8;
                                                                }
                                                                else if (RsDiff >= 7) {
                                                                    return 2;
                                                                }
                                                            }
                                                            else if (ssDCval >= 93) {
                                                                return 2;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                            else if (mcTcor >= 91) {
                                                if (MVDiff < 4) {
                                                    if (Rs < 2339) {
                                                        return 2;
                                                    }
                                                    else if (Rs >= 2339) {
                                                        return 4;
                                                    }
                                                }
                                                else if (MVDiff >= 4) {
                                                    if (Cs < 749) {
                                                        if (ssDCval < 52) {
                                                            return 4;
                                                        }
                                                        else if (ssDCval >= 52) {
                                                            return 2;
                                                        }
                                                    }
                                                    else if (Cs >= 749) {
                                                        if (AFD < 10) {
                                                            if (refDCval < 92) {
                                                                if (refDCval < 54) {
                                                                    return 8;
                                                                }
                                                                else if (refDCval >= 54) {
                                                                    if (RsDiff < 6) {
                                                                        return 4;
                                                                    }
                                                                    else if (RsDiff >= 6) {
                                                                        return 8;
                                                                    }
                                                                }
                                                            }
                                                            else if (refDCval >= 92) {
                                                                return 2;
                                                            }
                                                        }
                                                        else if (AFD >= 10) {
                                                            if (MVDiff < 6) {
                                                                return 4;
                                                            }
                                                            else if (MVDiff >= 6) {
                                                                if (TSC < 10) {
                                                                    if (Rs < 1339) {
                                                                        if (Rs < 864) {
                                                                            return 2;
                                                                        }
                                                                        else if (Rs >= 864) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                    else if (Rs >= 1339) {
                                                                        return 1;
                                                                    }
                                                                }
                                                                else if (TSC >= 10) {
                                                                    if (refDCval < 90) {
                                                                        return 4;
                                                                    }
                                                                    else if (refDCval >= 90) {
                                                                        return 1;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (refDCval >= 98) {
                                            if (posBalance < 59) {
                                                if (RsCsDiff < 23) {
                                                    return 4;
                                                }
                                                else if (RsCsDiff >= 23) {
                                                    if (tcor < 75) {
                                                        return 8;
                                                    }
                                                    else if (tcor >= 75) {
                                                        if (SC < 3851) {
                                                            return 2;
                                                        }
                                                        else if (SC >= 3851) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (posBalance >= 59) {
                                                if (diffMVdiffVal < 4) {
                                                    return 4;
                                                }
                                                else if (diffMVdiffVal >= 4) {
                                                    if (MVDiff < 20) {
                                                        if (Cs < 1063) {
                                                            return 4;
                                                        }
                                                        else if (Cs >= 1063) {
                                                            return 2;
                                                        }
                                                    }
                                                    else if (MVDiff >= 20) {
                                                        return 2;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                        }
                        else if (refDCval >= 104) {
                            if (ssDCval < 119) {
                                if (RsDiff < 6) {
                                    if (mu_mv_mag_sq < 46) {
                                        if (TSC < 3) {
                                            if (Rs < 828) {
                                                return 1;
                                            }
                                            else if (Rs >= 828) {
                                                return 8;
                                            }
                                        }
                                        else if (TSC >= 3) {
                                            if (ssDCval < 110) {
                                                if (diffRsCsdiff < -6) {
                                                    if (SC < 705) {
                                                        return 2;
                                                    }
                                                    else if (SC >= 705) {
                                                        if (Rs < 3153) {
                                                            if (mu_mv_mag_sq < 44) {
                                                                return 4;
                                                            }
                                                            else if (mu_mv_mag_sq >= 44) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (Rs >= 3153) {
                                                            if (MVDiff < 11) {
                                                                return 8;
                                                            }
                                                            else if (MVDiff >= 11) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (diffRsCsdiff >= -6) {
                                                    if (Cs < 244) {
                                                        return 2;
                                                    }
                                                    else if (Cs >= 244) {
                                                        if (RsCsDiff < 38) {
                                                            if (refDCval < 107) {
                                                                if (SC < 3131) {
                                                                    return 4;
                                                                }
                                                                else if (SC >= 3131) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (refDCval >= 107) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (RsCsDiff >= 38) {
                                                            if (SC < 4790) {
                                                                if (Rs < 2202) {
                                                                    return 2;
                                                                }
                                                                else if (Rs >= 2202) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (SC >= 4790) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                            else if (ssDCval >= 110) {
                                                if (Rs < 1713) {
                                                    return 4;
                                                }
                                                else if (Rs >= 1713) {
                                                    if (Rs < 2936) {
                                                        return 8;
                                                    }
                                                    else if (Rs >= 2936) {
                                                        if (Rs < 3349) {
                                                            return 2;
                                                        }
                                                        else if (Rs >= 3349) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (mu_mv_mag_sq >= 46) {
                                        if (posBalance < 42) {
                                            if (mu_mv_mag_sq < 57) {
                                                return 4;
                                            }
                                            else if (mu_mv_mag_sq >= 57) {
                                                if (diffMVdiffVal < -11) {
                                                    return 2;
                                                }
                                                else if (diffMVdiffVal >= -11) {
                                                    if (TSCindex < 7) {
                                                        if (mu_mv_mag_sq < 58) {
                                                            if (Cs < 1058) {
                                                                return 8;
                                                            }
                                                            else if (Cs >= 1058) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (mu_mv_mag_sq >= 58) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (TSCindex >= 7) {
                                                        if (Rs < 1579) {
                                                            return 4;
                                                        }
                                                        else if (Rs >= 1579) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (posBalance >= 42) {
                                            if (MVDiff < 1) {
                                                return 4;
                                            }
                                            else if (MVDiff >= 1) {
                                                return 8;
                                            }
                                        }
                                    }
                                }
                                else if (RsDiff >= 6) {
                                    if (posBalance < 48) {
                                        if (negBalance < 39) {
                                            if (Rs < 2883) {
                                                return 4;
                                            }
                                            else if (Rs >= 2883) {
                                                if (AFD < 7) {
                                                    if (Rs < 3415) {
                                                        return 4;
                                                    }
                                                    else if (Rs >= 3415) {
                                                        return 8;
                                                    }
                                                }
                                                else if (AFD >= 7) {
                                                    if (ssDCval < 112) {
                                                        if (Rs < 3604) {
                                                            if (ssDCval < 106) {
                                                                if (mu_mv_mag_sq < 65) {
                                                                    return 4;
                                                                }
                                                                else if (mu_mv_mag_sq >= 65) {
                                                                    return 2;
                                                                }
                                                            }
                                                            else if (ssDCval >= 106) {
                                                                return 2;
                                                            }
                                                        }
                                                        else if (Rs >= 3604) {
                                                            if (RsDiff < 9) {
                                                                return 4;
                                                            }
                                                            else if (RsDiff >= 9) {
                                                                if (AFD < 10) {
                                                                    return 2;
                                                                }
                                                                else if (AFD >= 10) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (ssDCval >= 112) {
                                                        return 2;
                                                    }
                                                }
                                            }
                                        }
                                        else if (negBalance >= 39) {
                                            return 4;
                                        }
                                    }
                                    else if (posBalance >= 48) {
                                        if (negBalance < 51) {
                                            if (negBalance < 35) {
                                                if (TSC < 7) {
                                                    if (SC < 4369) {
                                                        return 8;
                                                    }
                                                    else if (SC >= 4369) {
                                                        return 4;
                                                    }
                                                }
                                                else if (TSC >= 7) {
                                                    if (Rs < 3107) {
                                                        return 8;
                                                    }
                                                    else if (Rs >= 3107) {
                                                        return 2;
                                                    }
                                                }
                                            }
                                            else if (negBalance >= 35) {
                                                if (SC < 4382) {
                                                    if (AFD < 10) {
                                                        return 8;
                                                    }
                                                    else if (AFD >= 10) {
                                                        if (diffRsCsdiff < -22) {
                                                            return 8;
                                                        }
                                                        else if (diffRsCsdiff >= -22) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                                else if (SC >= 4382) {
                                                    if (negBalance < 37) {
                                                        if (Cs < 1977) {
                                                            return 8;
                                                        }
                                                        else if (Cs >= 1977) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (negBalance >= 37) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                        }
                                        else if (negBalance >= 51) {
                                            if (refDCval < 110) {
                                                return 4;
                                            }
                                            else if (refDCval >= 110) {
                                                if (MVDiff < 8) {
                                                    if (Cs < 1557) {
                                                        return 4;
                                                    }
                                                    else if (Cs >= 1557) {
                                                        if (RsDiff < 9) {
                                                            if (ssDCval < 117) {
                                                                if (Cs < 2138) {
                                                                    return 4;
                                                                }
                                                                else if (Cs >= 2138) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (ssDCval >= 117) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (RsDiff >= 9) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                                else if (MVDiff >= 8) {
                                                    return 4;
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                            else if (ssDCval >= 119) {
                                if (SC < 3364) {
                                    if (mcTcor < 95) {
                                        if (RsDiff < 5) {
                                            if (tcor < 88) {
                                                if (mcTcor < 85) {
                                                    return 8;
                                                }
                                                else if (mcTcor >= 85) {
                                                    return 1;
                                                }
                                            }
                                            else if (tcor >= 88) {
                                                if (MVDiff < 14) {
                                                    return 8;
                                                }
                                                else if (MVDiff >= 14) {
                                                    return 4;
                                                }
                                            }
                                        }
                                        else if (RsDiff >= 5) {
                                            if (diffAFD < 2) {
                                                if (negBalance < 32) {
                                                    if (SC < 2245) {
                                                        return 4;
                                                    }
                                                    else if (SC >= 2245) {
                                                        if (MVDiff < 11) {
                                                            if (Cs < 1299) {
                                                                return 8;
                                                            }
                                                            else if (Cs >= 1299) {
                                                                return 1;
                                                            }
                                                        }
                                                        else if (MVDiff >= 11) {
                                                            return 1;
                                                        }
                                                    }
                                                }
                                                else if (negBalance >= 32) {
                                                    if (tcor < 86) {
                                                        return 4;
                                                    }
                                                    else if (tcor >= 86) {
                                                        if (ssDCval < 123) {
                                                            return 4;
                                                        }
                                                        else if (ssDCval >= 123) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (diffAFD >= 2) {
                                                if (SC < 2453) {
                                                    return 8;
                                                }
                                                else if (SC >= 2453) {
                                                    return 1;
                                                }
                                            }
                                        }
                                    }
                                    else if (mcTcor >= 95) {
                                        if (CsDiff < 4) {
                                            if (Rs < 336) {
                                                if (SC < 505) {
                                                    return 1;
                                                }
                                                else if (SC >= 505) {
                                                    if (AFD < 5) {
                                                        if (diffMVdiffVal < -25) {
                                                            return 2;
                                                        }
                                                        else if (diffMVdiffVal >= -25) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (AFD >= 5) {
                                                        return 2;
                                                    }
                                                }
                                            }
                                            else if (Rs >= 336) {
                                                if (SC < 1703) {
                                                    if (TSC < 2) {
                                                        if (Rs < 563) {
                                                            return 2;
                                                        }
                                                        else if (Rs >= 563) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (TSC >= 2) {
                                                        if (Cs < 304) {
                                                            return 1;
                                                        }
                                                        else if (Cs >= 304) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                                else if (SC >= 1703) {
                                                    if (posBalance < 58) {
                                                        if (RsCsDiff < 22) {
                                                            if (ssDCval < 173) {
                                                                return 8;
                                                            }
                                                            else if (ssDCval >= 173) {
                                                                return 2;
                                                            }
                                                        }
                                                        else if (RsCsDiff >= 22) {
                                                            return 2;
                                                        }
                                                    }
                                                    else if (posBalance >= 58) {
                                                        if (RsDiff < 9) {
                                                            return 4;
                                                        }
                                                        else if (RsDiff >= 9) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (CsDiff >= 4) {
                                            return 2;
                                        }
                                    }
                                }
                                else if (SC >= 3364) {
                                    if (ssDCval < 129) {
                                        if (mu_mv_mag_sq < 55) {
                                            if (refDCval < 123) {
                                                return 4;
                                            }
                                            else if (refDCval >= 123) {
                                                if (SC < 4563) {
                                                    return 4;
                                                }
                                                else if (SC >= 4563) {
                                                    if (Rs < 2607) {
                                                        return 8;
                                                    }
                                                    else if (Rs >= 2607) {
                                                        if (SC < 4888) {
                                                            if (negBalance < 52) {
                                                                return 8;
                                                            }
                                                            else if (negBalance >= 52) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (SC >= 4888) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (mu_mv_mag_sq >= 55) {
                                            if (negBalance < 48) {
                                                return 4;
                                            }
                                            else if (negBalance >= 48) {
                                                if (AFD < 10) {
                                                    if (diffTSC < 1) {
                                                        return 4;
                                                    }
                                                    else if (diffTSC >= 1) {
                                                        return 8;
                                                    }
                                                }
                                                else if (AFD >= 10) {
                                                    if (CsDiff < 8) {
                                                        if (negBalance < 53) {
                                                            return 8;
                                                        }
                                                        else if (negBalance >= 53) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (CsDiff >= 8) {
                                                        if (refDCval < 124) {
                                                            return 4;
                                                        }
                                                        else if (refDCval >= 124) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (ssDCval >= 129) {
                                        if (tcor < 85) {
                                            return 1;
                                        }
                                        else if (tcor >= 85) {
                                            if (Cs < 2453) {
                                                return 4;
                                            }
                                            else if (Cs >= 2453) {
                                                return 8;
                                            }
                                        }
                                    }
                                }
                            }
                        }
                    }
                    else if (AFD >= 12) {
                        if (Cs < 1464) {
                            if (RsDiff < 8) {
                                if (Cs < 533) {
                                    if (ssDCval < 97) {
                                        if (AFD < 30) {
                                            if (SC < 304) {
                                                if (RsCsDiff < 3) {
                                                    return 2;
                                                }
                                                else if (RsCsDiff >= 3) {
                                                    if (AFD < 19) {
                                                        return 1;
                                                    }
                                                    else if (AFD >= 19) {
                                                        return 2;
                                                    }
                                                }
                                            }
                                            else if (SC >= 304) {
                                                if (TSC < 10) {
                                                    if (mcTcor < 79) {
                                                        return 1;
                                                    }
                                                    else if (mcTcor >= 79) {
                                                        return 2;
                                                    }
                                                }
                                                else if (TSC >= 10) {
                                                    return 1;
                                                }
                                            }
                                        }
                                        else if (AFD >= 30) {
                                            return 8;
                                        }
                                    }
                                    else if (ssDCval >= 97) {
                                        if (tcor < 77) {
                                            if (Rs < 222) {
                                                return 1;
                                            }
                                            else if (Rs >= 222) {
                                                if (Rs < 1430) {
                                                    return 2;
                                                }
                                                else if (Rs >= 1430) {
                                                    return 8;
                                                }
                                            }
                                        }
                                        else if (tcor >= 77) {
                                            return 1;
                                        }
                                    }
                                }
                                else if (Cs >= 533) {
                                    if (RsDiff < 5) {
                                        if (MVDiff < 14) {
                                            if (ssDCval < 83) {
                                                if (AFD < 13) {
                                                    return 4;
                                                }
                                                else if (AFD >= 13) {
                                                    return 2;
                                                }
                                            }
                                            else if (ssDCval >= 83) {
                                                return 4;
                                            }
                                        }
                                        else if (MVDiff >= 14) {
                                            if (CsDiff < 3) {
                                                if (Rs < 878) {
                                                    return 8;
                                                }
                                                else if (Rs >= 878) {
                                                    return 4;
                                                }
                                            }
                                            else if (CsDiff >= 3) {
                                                if (Rs < 1257) {
                                                    if (posBalance < 59) {
                                                        if (mu_mv_mag_sq < 38) {
                                                            if (negBalance < 97) {
                                                                return 2;
                                                            }
                                                            else if (negBalance >= 97) {
                                                                if (Rs < 750) {
                                                                    return 1;
                                                                }
                                                                else if (Rs >= 750) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                        else if (mu_mv_mag_sq >= 38) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (posBalance >= 59) {
                                                        if (TSC < 7) {
                                                            if (negBalance < 49) {
                                                                return 1;
                                                            }
                                                            else if (negBalance >= 49) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (TSC >= 7) {
                                                            return 1;
                                                        }
                                                    }
                                                }
                                                else if (Rs >= 1257) {
                                                    return 8;
                                                }
                                            }
                                        }
                                    }
                                    else if (RsDiff >= 5) {
                                        if (TSC < 9) {
                                            if (tcor < 79) {
                                                if (negBalance < 30) {
                                                    return 8;
                                                }
                                                else if (negBalance >= 30) {
                                                    if (Cs < 873) {
                                                        if (mu_mv_mag_sq < 5) {
                                                            if (SC < 2025) {
                                                                return 8;
                                                            }
                                                            else if (SC >= 2025) {
                                                                return 2;
                                                            }
                                                        }
                                                        else if (mu_mv_mag_sq >= 5) {
                                                            return 1;
                                                        }
                                                    }
                                                    else if (Cs >= 873) {
                                                        return 1;
                                                    }
                                                }
                                            }
                                            else if (tcor >= 79) {
                                                if (Rs < 2085) {
                                                    if (Cs < 751) {
                                                        return 2;
                                                    }
                                                    else if (Cs >= 751) {
                                                        if (TSC < 7) {
                                                            if (MVDiff < 53) {
                                                                return 2;
                                                            }
                                                            else if (MVDiff >= 53) {
                                                                if (Cs < 938) {
                                                                    return 1;
                                                                }
                                                                else if (Cs >= 938) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                        else if (TSC >= 7) {
                                                            if (Rs < 1337) {
                                                                if (MVDiff < 28) {
                                                                    return 4;
                                                                }
                                                                else if (MVDiff >= 28) {
                                                                    return 2;
                                                                }
                                                            }
                                                            else if (Rs >= 1337) {
                                                                if (RsCsDiff < 63) {
                                                                    return 1;
                                                                }
                                                                else if (RsCsDiff >= 63) {
                                                                    if (Cs < 1101) {
                                                                        return 2;
                                                                    }
                                                                    else if (Cs >= 1101) {
                                                                        return 1;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (Rs >= 2085) {
                                                    if (tcor < 88) {
                                                        return 8;
                                                    }
                                                    else if (tcor >= 88) {
                                                        if (Rs < 2142) {
                                                            return 8;
                                                        }
                                                        else if (Rs >= 2142) {
                                                            return 2;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (TSC >= 9) {
                                            if (TSC < 12) {
                                                if (ssDCval < 97) {
                                                    if (posBalance < 62) {
                                                        if (mcTcor < 85) {
                                                            return 8;
                                                        }
                                                        else if (mcTcor >= 85) {
                                                            if (CsDiff < 5) {
                                                                return 8;
                                                            }
                                                            else if (CsDiff >= 5) {
                                                                if (AFD < 17) {
                                                                    return 4;
                                                                }
                                                                else if (AFD >= 17) {
                                                                    return 1;
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (posBalance >= 62) {
                                                        return 1;
                                                    }
                                                }
                                                else if (ssDCval >= 97) {
                                                    if (ssDCval < 122) {
                                                        return 2;
                                                    }
                                                    else if (ssDCval >= 122) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                            else if (TSC >= 12) {
                                                if (Rs < 1342) {
                                                    if (AFD < 37) {
                                                        return 1;
                                                    }
                                                    else if (AFD >= 37) {
                                                        return 4;
                                                    }
                                                }
                                                else if (Rs >= 1342) {
                                                    if (SC < 2610) {
                                                        return 4;
                                                    }
                                                    else if (SC >= 2610) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                            else if (RsDiff >= 8) {
                                if (mcTcor < 85) {
                                    if (SC < 3045) {
                                        if (mu_mv_mag_sq < 10) {
                                            if (MVDiff < 22) {
                                                return 8;
                                            }
                                            else if (MVDiff >= 22) {
                                                if (posBalance < 65) {
                                                    if (SC < 2973) {
                                                        return 2;
                                                    }
                                                    else if (SC >= 2973) {
                                                        if (Rs < 2025) {
                                                            return 2;
                                                        }
                                                        else if (Rs >= 2025) {
                                                            return 1;
                                                        }
                                                    }
                                                }
                                                else if (posBalance >= 65) {
                                                    return 8;
                                                }
                                            }
                                        }
                                        else if (mu_mv_mag_sq >= 10) {
                                            if (Cs < 1181) {
                                                if (TSC < 12) {
                                                    if (SC < 2433) {
                                                        return 1;
                                                    }
                                                    else if (SC >= 2433) {
                                                        return 4;
                                                    }
                                                }
                                                else if (TSC >= 12) {
                                                    if (tcor < 8) {
                                                        if (AFD < 45) {
                                                            return 8;
                                                        }
                                                        else if (AFD >= 45) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (tcor >= 8) {
                                                        if (TSC < 18) {
                                                            if (SC < 2396) {
                                                                return 2;
                                                            }
                                                            else if (SC >= 2396) {
                                                                if (Rs < 1731) {
                                                                    return 4;
                                                                }
                                                                else if (Rs >= 1731) {
                                                                    return 1;
                                                                }
                                                            }
                                                        }
                                                        else if (TSC >= 18) {
                                                            if (mcTcor < 58) {
                                                                if (Rs < 1379) {
                                                                    return 1;
                                                                }
                                                                else if (Rs >= 1379) {
                                                                    return 2;
                                                                }
                                                            }
                                                            else if (mcTcor >= 58) {
                                                                return 1;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                            else if (Cs >= 1181) {
                                                if (mcTcor < 81) {
                                                    return 8;
                                                }
                                                else if (mcTcor >= 81) {
                                                    return 1;
                                                }
                                            }
                                        }
                                    }
                                    else if (SC >= 3045) {
                                        if (AFD < 30) {
                                            if (AFD < 25) {
                                                if (TSC < 11) {
                                                    return 1;
                                                }
                                                else if (TSC >= 11) {
                                                    if (tcor < 57) {
                                                        return 1;
                                                    }
                                                    else if (tcor >= 57) {
                                                        if (MVDiff < 96) {
                                                            return 2;
                                                        }
                                                        else if (MVDiff >= 96) {
                                                            return 1;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (AFD >= 25) {
                                                return 1;
                                            }
                                        }
                                        else if (AFD >= 30) {
                                            if (refDCval < 89) {
                                                return 4;
                                            }
                                            else if (refDCval >= 89) {
                                                return 2;
                                            }
                                        }
                                    }
                                }
                                else if (mcTcor >= 85) {
                                    if (Rs < 1228) {
                                        if (RsCsDiff < 254) {
                                            if (mu_mv_mag_sq < 24) {
                                                return 1;
                                            }
                                            else if (mu_mv_mag_sq >= 24) {
                                                return 2;
                                            }
                                        }
                                        else if (RsCsDiff >= 254) {
                                            return 8;
                                        }
                                    }
                                    else if (Rs >= 1228) {
                                        if (TSC < 10) {
                                            if (MVDiff < 7) {
                                                if (diffTSC < 1) {
                                                    return 1;
                                                }
                                                else if (diffTSC >= 1) {
                                                    if (AFD < 13) {
                                                        return 4;
                                                    }
                                                    else if (AFD >= 13) {
                                                        return 2;
                                                    }
                                                }
                                            }
                                            else if (MVDiff >= 7) {
                                                if (mcTcor < 86) {
                                                    return 1;
                                                }
                                                else if (mcTcor >= 86) {
                                                    if (Cs < 952) {
                                                        return 2;
                                                    }
                                                    else if (Cs >= 952) {
                                                        if (refDCval < 82) {
                                                            if (MVDiff < 31) {
                                                                return 4;
                                                            }
                                                            else if (MVDiff >= 31) {
                                                                return 2;
                                                            }
                                                        }
                                                        else if (refDCval >= 82) {
                                                            if (AFD < 16) {
                                                                if (diffMVdiffVal < -4) {
                                                                    return 4;
                                                                }
                                                                else if (diffMVdiffVal >= -4) {
                                                                    return 2;
                                                                }
                                                            }
                                                            else if (AFD >= 16) {
                                                                return 2;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (TSC >= 10) {
                                            if (negBalance < 41) {
                                                if (MVDiff < 29) {
                                                    return 8;
                                                }
                                                else if (MVDiff >= 29) {
                                                    if (AFD < 21) {
                                                        return 4;
                                                    }
                                                    else if (AFD >= 21) {
                                                        return 2;
                                                    }
                                                }
                                            }
                                            else if (negBalance >= 41) {
                                                if (Cs < 653) {
                                                    if (AFD < 17) {
                                                        return 4;
                                                    }
                                                    else if (AFD >= 17) {
                                                        return 2;
                                                    }
                                                }
                                                else if (Cs >= 653) {
                                                    if (MVDiff < 39) {
                                                        return 2;
                                                    }
                                                    else if (MVDiff >= 39) {
                                                        if (MVDiff < 40) {
                                                            if (Rs < 1643) {
                                                                return 2;
                                                            }
                                                            else if (Rs >= 1643) {
                                                                return 1;
                                                            }
                                                        }
                                                        else if (MVDiff >= 40) {
                                                            return 2;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                        }
                        else if (Cs >= 1464) {
                            if (refDCval < 80) {
                                if (negBalance < 43) {
                                    if (mu_mv_mag_sq < 33) {
                                        if (SC < 4242) {
                                            if (Rs < 1063) {
                                                return 2;
                                            }
                                            else if (Rs >= 1063) {
                                                return 1;
                                            }
                                        }
                                        else if (SC >= 4242) {
                                            return 8;
                                        }
                                    }
                                    else if (mu_mv_mag_sq >= 33) {
                                        if (mcTcor < 83) {
                                            if (Rs < 2352) {
                                                return 4;
                                            }
                                            else if (Rs >= 2352) {
                                                return 8;
                                            }
                                        }
                                        else if (mcTcor >= 83) {
                                            if (RsCsDiff < 129) {
                                                if (Rs < 1656) {
                                                    return 1;
                                                }
                                                else if (Rs >= 1656) {
                                                    return 8;
                                                }
                                            }
                                            else if (RsCsDiff >= 129) {
                                                return 1;
                                            }
                                        }
                                    }
                                }
                                else if (negBalance >= 43) {
                                    if (mu_mv_mag_sq < 5) {
                                        if (tcor < 66) {
                                            return 4;
                                        }
                                        else if (tcor >= 66) {
                                            if (mcTcor < 88) {
                                                if (negBalance < 52) {
                                                    if (Cs < 2510) {
                                                        return 4;
                                                    }
                                                    else if (Cs >= 2510) {
                                                        return 8;
                                                    }
                                                }
                                                else if (negBalance >= 52) {
                                                    return 8;
                                                }
                                            }
                                            else if (mcTcor >= 88) {
                                                if (ssDCval < 78) {
                                                    if (Rs < 1857) {
                                                        return 1;
                                                    }
                                                    else if (Rs >= 1857) {
                                                        return 4;
                                                    }
                                                }
                                                else if (ssDCval >= 78) {
                                                    if (SC < 4233) {
                                                        if (SC < 3728) {
                                                            return 4;
                                                        }
                                                        else if (SC >= 3728) {
                                                            return 1;
                                                        }
                                                    }
                                                    else if (SC >= 4233) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (mu_mv_mag_sq >= 5) {
                                        if (ssDCval < 78) {
                                            if (AFD < 18) {
                                                if (tcor < 81) {
                                                    return 8;
                                                }
                                                else if (tcor >= 81) {
                                                    if (negBalance < 56) {
                                                        return 8;
                                                    }
                                                    else if (negBalance >= 56) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                            else if (AFD >= 18) {
                                                if (RsCsDiff < 258) {
                                                    if (SC < 3619) {
                                                        return 4;
                                                    }
                                                    else if (SC >= 3619) {
                                                        return 8;
                                                    }
                                                }
                                                else if (RsCsDiff >= 258) {
                                                    if (mu_mv_mag_sq < 11) {
                                                        if (TSC < 13) {
                                                            if (Rs < 2541) {
                                                                return 1;
                                                            }
                                                            else if (Rs >= 2541) {
                                                                return 2;
                                                            }
                                                        }
                                                        else if (TSC >= 13) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (mu_mv_mag_sq >= 11) {
                                                        if (AFD < 27) {
                                                            return 1;
                                                        }
                                                        else if (AFD >= 27) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (ssDCval >= 78) {
                                            if (tcor < 90) {
                                                if (negBalance < 45) {
                                                    if (RsDiff < 12) {
                                                        return 8;
                                                    }
                                                    else if (RsDiff >= 12) {
                                                        return 1;
                                                    }
                                                }
                                                else if (negBalance >= 45) {
                                                    if (SC < 6106) {
                                                        return 1;
                                                    }
                                                    else if (SC >= 6106) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                            else if (tcor >= 90) {
                                                return 2;
                                            }
                                        }
                                    }
                                }
                            }
                            else if (refDCval >= 80) {
                                if (mu_mv_mag_sq < 2) {
                                    if (negBalance < 51) {
                                        if (posBalance < 57) {
                                            if (Rs < 2518) {
                                                if (TSC < 11) {
                                                    if (Rs < 1657) {
                                                        if (Rs < 1481) {
                                                            if (Rs < 1070) {
                                                                return 2;
                                                            }
                                                            else if (Rs >= 1070) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (Rs >= 1481) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (Rs >= 1657) {
                                                        if (SC < 4569) {
                                                            return 1;
                                                        }
                                                        else if (SC >= 4569) {
                                                            if (MVDiff < 18) {
                                                                return 4;
                                                            }
                                                            else if (MVDiff >= 18) {
                                                                return 1;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (TSC >= 11) {
                                                    if (ssDCval < 89) {
                                                        return 4;
                                                    }
                                                    else if (ssDCval >= 89) {
                                                        if (Rs < 2387) {
                                                            return 4;
                                                        }
                                                        else if (Rs >= 2387) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (Rs >= 2518) {
                                                if (negBalance < 43) {
                                                    if (refDCval < 105) {
                                                        return 1;
                                                    }
                                                    else if (refDCval >= 105) {
                                                        return 8;
                                                    }
                                                }
                                                else if (negBalance >= 43) {
                                                    if (ssDCval < 119) {
                                                        if (posBalance < 51) {
                                                            return 1;
                                                        }
                                                        else if (posBalance >= 51) {
                                                            if (SC < 4734) {
                                                                if (Rs < 2532) {
                                                                    return 8;
                                                                }
                                                                else if (Rs >= 2532) {
                                                                    return 1;
                                                                }
                                                            }
                                                            else if (SC >= 4734) {
                                                                if (diffAFD < -4) {
                                                                    return 4;
                                                                }
                                                                else if (diffAFD >= -4) {
                                                                    if (negBalance < 44) {
                                                                        if (Rs < 2804) {
                                                                            return 4;
                                                                        }
                                                                        else if (Rs >= 2804) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                    else if (negBalance >= 44) {
                                                                        return 8;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (ssDCval >= 119) {
                                                        return 1;
                                                    }
                                                }
                                            }
                                        }
                                        else if (posBalance >= 57) {
                                            if (refDCval < 122) {
                                                if (Cs < 2412) {
                                                    if (Rs < 2226) {
                                                        return 4;
                                                    }
                                                    else if (Rs >= 2226) {
                                                        if (Rs < 2942) {
                                                            if (RsCsDiff < 121) {
                                                                return 1;
                                                            }
                                                            else if (RsCsDiff >= 121) {
                                                                if (MVDiff < 14) {
                                                                    return 2;
                                                                }
                                                                else if (MVDiff >= 14) {
                                                                    return 1;
                                                                }
                                                            }
                                                        }
                                                        else if (Rs >= 2942) {
                                                            if (Cs < 2281) {
                                                                return 2;
                                                            }
                                                            else if (Cs >= 2281) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (Cs >= 2412) {
                                                    if (posBalance < 62) {
                                                        if (Rs < 2552) {
                                                            return 2;
                                                        }
                                                        else if (Rs >= 2552) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (posBalance >= 62) {
                                                        if (SC < 5566) {
                                                            return 2;
                                                        }
                                                        else if (SC >= 5566) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (refDCval >= 122) {
                                                return 1;
                                            }
                                        }
                                    }
                                    else if (negBalance >= 51) {
                                        if (mcTcor < 87) {
                                            if (MVDiff < 19) {
                                                if (tcor < 65) {
                                                    return 1;
                                                }
                                                else if (tcor >= 65) {
                                                    if (posBalance < 55) {
                                                        return 1;
                                                    }
                                                    else if (posBalance >= 55) {
                                                        if (MVDiff < 18) {
                                                            return 4;
                                                        }
                                                        else if (MVDiff >= 18) {
                                                            if (Rs < 2828) {
                                                                return 1;
                                                            }
                                                            else if (Rs >= 2828) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                            else if (MVDiff >= 19) {
                                                if (Rs < 2528) {
                                                    if (posBalance < 61) {
                                                        if (SC < 3811) {
                                                            if (TSC < 14) {
                                                                return 2;
                                                            }
                                                            else if (TSC >= 14) {
                                                                return 1;
                                                            }
                                                        }
                                                        else if (SC >= 3811) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (posBalance >= 61) {
                                                        if (CsDiff < 9) {
                                                            if (Rs < 2054) {
                                                                return 4;
                                                            }
                                                            else if (Rs >= 2054) {
                                                                return 1;
                                                            }
                                                        }
                                                        else if (CsDiff >= 9) {
                                                            return 2;
                                                        }
                                                    }
                                                }
                                                else if (Rs >= 2528) {
                                                    return 4;
                                                }
                                            }
                                        }
                                        else if (mcTcor >= 87) {
                                            if (Cs < 2438) {
                                                if (CsDiff < 7) {
                                                    if (ssDCval < 105) {
                                                        if (Rs < 1779) {
                                                            return 4;
                                                        }
                                                        else if (Rs >= 1779) {
                                                            if (RsDiff < 8) {
                                                                return 2;
                                                            }
                                                            else if (RsDiff >= 8) {
                                                                return 1;
                                                            }
                                                        }
                                                    }
                                                    else if (ssDCval >= 105) {
                                                        return 4;
                                                    }
                                                }
                                                else if (CsDiff >= 7) {
                                                    if (Cs < 2396) {
                                                        if (negBalance < 52) {
                                                            if (SC < 4390) {
                                                                return 2;
                                                            }
                                                            else if (SC >= 4390) {
                                                                if (SC < 4929) {
                                                                    return 4;
                                                                }
                                                                else if (SC >= 4929) {
                                                                    return 2;
                                                                }
                                                            }
                                                        }
                                                        else if (negBalance >= 52) {
                                                            if (TSC < 13) {
                                                                if (ssDCval < 116) {
                                                                    if (refDCval < 81) {
                                                                        if (Rs < 2342) {
                                                                            return 2;
                                                                        }
                                                                        else if (Rs >= 2342) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                    else if (refDCval >= 81) {
                                                                        return 2;
                                                                    }
                                                                }
                                                                else if (ssDCval >= 116) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (TSC >= 13) {
                                                                if (MVDiff < 28) {
                                                                    if (Rs < 2377) {
                                                                        return 2;
                                                                    }
                                                                    else if (Rs >= 2377) {
                                                                        return 1;
                                                                    }
                                                                }
                                                                else if (MVDiff >= 28) {
                                                                    if (tcor < 77) {
                                                                        return 4;
                                                                    }
                                                                    else if (tcor >= 77) {
                                                                        return 2;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (Cs >= 2396) {
                                                        if (Rs < 2774) {
                                                            return 4;
                                                        }
                                                        else if (Rs >= 2774) {
                                                            return 1;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (Cs >= 2438) {
                                                if (Rs < 2770) {
                                                    if (AFD < 18) {
                                                        return 4;
                                                    }
                                                    else if (AFD >= 18) {
                                                        return 2;
                                                    }
                                                }
                                                else if (Rs >= 2770) {
                                                    return 8;
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (mu_mv_mag_sq >= 2) {
                                    if (mcTcor < 94) {
                                        if (tcor < 90) {
                                            if (ssDCval < 140) {
                                                if (Cs < 2823) {
                                                    if (Cs < 2282) {
                                                        if (tcor < 72) {
                                                            if (AFD < 34) {
                                                                if (diffMVdiffVal < 29) {
                                                                    if (diffRsCsdiff < -109) {
                                                                        if (MVDiff < 24) {
                                                                            return 1;
                                                                        }
                                                                        else if (MVDiff >= 24) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                    else if (diffRsCsdiff >= -109) {
                                                                        if (mcTcor < 85) {
                                                                            if (tcor < 69) {
                                                                                return 1;
                                                                            }
                                                                            else if (tcor >= 69) {
                                                                                if (posBalance < 59) {
                                                                                    return 1;
                                                                                }
                                                                                else if (posBalance >= 59) {
                                                                                    if (Rs < 2715) {
                                                                                        return 4;
                                                                                    }
                                                                                    else if (Rs >= 2715) {
                                                                                        return 1;
                                                                                    }
                                                                                }
                                                                            }
                                                                        }
                                                                        else if (mcTcor >= 85) {
                                                                            if (MVDiff < 12) {
                                                                                if (ssDCval < 117) {
                                                                                    return 2;
                                                                                }
                                                                                else if (ssDCval >= 117) {
                                                                                    return 1;
                                                                                }
                                                                            }
                                                                            else if (MVDiff >= 12) {
                                                                                return 1;
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                                else if (diffMVdiffVal >= 29) {
                                                                    if (negBalance < 30) {
                                                                        return 4;
                                                                    }
                                                                    else if (negBalance >= 30) {
                                                                        return 8;
                                                                    }
                                                                }
                                                            }
                                                            else if (AFD >= 34) {
                                                                if (Cs < 1999) {
                                                                    return 2;
                                                                }
                                                                else if (Cs >= 1999) {
                                                                    return 1;
                                                                }
                                                            }
                                                        }
                                                        else if (tcor >= 72) {
                                                            if (ssDCval < 84) {
                                                                if (SC < 4113) {
                                                                    if (Cs < 1616) {
                                                                        return 4;
                                                                    }
                                                                    else if (Cs >= 1616) {
                                                                        return 1;
                                                                    }
                                                                }
                                                                else if (SC >= 4113) {
                                                                    if (negBalance < 51) {
                                                                        if (Rs < 2652) {
                                                                            return 8;
                                                                        }
                                                                        else if (Rs >= 2652) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                    else if (negBalance >= 51) {
                                                                        if (tcor < 89) {
                                                                            if (MVDiff < 17) {
                                                                                return 1;
                                                                            }
                                                                            else if (MVDiff >= 17) {
                                                                                if (negBalance < 54) {
                                                                                    return 2;
                                                                                }
                                                                                else if (negBalance >= 54) {
                                                                                    return 1;
                                                                                }
                                                                            }
                                                                        }
                                                                        else if (tcor >= 89) {
                                                                            return 2;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                            else if (ssDCval >= 84) {
                                                                if (negBalance < 47) {
                                                                    if (Rs < 2093) {
                                                                        if (tcor < 84) {
                                                                            if (Cs < 1514) {
                                                                                return 1;
                                                                            }
                                                                            else if (Cs >= 1514) {
                                                                                if (Cs < 2001) {
                                                                                    return 4;
                                                                                }
                                                                                else if (Cs >= 2001) {
                                                                                    return 8;
                                                                                }
                                                                            }
                                                                        }
                                                                        else if (tcor >= 84) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                    else if (Rs >= 2093) {
                                                                        if (TSC < 9) {
                                                                            if (SC < 5929) {
                                                                                return 8;
                                                                            }
                                                                            else if (SC >= 5929) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                        else if (TSC >= 9) {
                                                                            if (ssDCval < 88) {
                                                                                if (negBalance < 41) {
                                                                                    return 1;
                                                                                }
                                                                                else if (negBalance >= 41) {
                                                                                    if (Rs < 2319) {
                                                                                        return 8;
                                                                                    }
                                                                                    else if (Rs >= 2319) {
                                                                                        return 4;
                                                                                    }
                                                                                }
                                                                            }
                                                                            else if (ssDCval >= 88) {
                                                                                return 1;
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                                else if (negBalance >= 47) {
                                                                    if (Cs < 1763) {
                                                                        if (SC < 3661) {
                                                                            if (Cs < 1627) {
                                                                                return 4;
                                                                            }
                                                                            else if (Cs >= 1627) {
                                                                                return 2;
                                                                            }
                                                                        }
                                                                        else if (SC >= 3661) {
                                                                            if (diffTSC < -1) {
                                                                                return 8;
                                                                            }
                                                                            else if (diffTSC >= -1) {
                                                                                if (refDCval < 123) {
                                                                                    if (AFD < 18) {
                                                                                        return 2;
                                                                                    }
                                                                                    else if (AFD >= 18) {
                                                                                        return 4;
                                                                                    }
                                                                                }
                                                                                else if (refDCval >= 123) {
                                                                                    return 2;
                                                                                }
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (Cs >= 1763) {
                                                                        if (mcTcor < 87) {
                                                                            if (posBalance < 59) {
                                                                                return 1;
                                                                            }
                                                                            else if (posBalance >= 59) {
                                                                                if (AFD < 25) {
                                                                                    return 4;
                                                                                }
                                                                                else if (AFD >= 25) {
                                                                                    return 1;
                                                                                }
                                                                            }
                                                                        }
                                                                        else if (mcTcor >= 87) {
                                                                            if (TSC < 12) {
                                                                                if (RsCsDiff < 284) {
                                                                                    if (mu_mv_mag_sq < 45) {
                                                                                        if (ssDCval < 127) {
                                                                                            return 2;
                                                                                        }
                                                                                        else if (ssDCval >= 127) {
                                                                                            return 1;
                                                                                        }
                                                                                    }
                                                                                    else if (mu_mv_mag_sq >= 45) {
                                                                                        if (Cs < 2107) {
                                                                                            return 4;
                                                                                        }
                                                                                        else if (Cs >= 2107) {
                                                                                            return 1;
                                                                                        }
                                                                                    }
                                                                                }
                                                                                else if (RsCsDiff >= 284) {
                                                                                    if (Cs < 1948) {
                                                                                        return 1;
                                                                                    }
                                                                                    else if (Cs >= 1948) {
                                                                                        return 4;
                                                                                    }
                                                                                }
                                                                            }
                                                                            else if (TSC >= 12) {
                                                                                if (CsDiff < 8) {
                                                                                    if (RsDiff < 9) {
                                                                                        if (AFD < 15) {
                                                                                            return 4;
                                                                                        }
                                                                                        else if (AFD >= 15) {
                                                                                            return 1;
                                                                                        }
                                                                                    }
                                                                                    else if (RsDiff >= 9) {
                                                                                        return 8;
                                                                                    }
                                                                                }
                                                                                else if (CsDiff >= 8) {
                                                                                    if (RsCsDiff < 264) {
                                                                                        if (refDCval < 104) {
                                                                                            if (refDCval < 88) {
                                                                                                return 8;
                                                                                            }
                                                                                            else if (refDCval >= 88) {
                                                                                                return 2;
                                                                                            }
                                                                                        }
                                                                                        else if (refDCval >= 104) {
                                                                                            if (RsDiff < 11) {
                                                                                                return 1;
                                                                                            }
                                                                                            else if (RsDiff >= 11) {
                                                                                                return 8;
                                                                                            }
                                                                                        }
                                                                                    }
                                                                                    else if (RsCsDiff >= 264) {
                                                                                        if (RsDiff < 15) {
                                                                                            return 2;
                                                                                        }
                                                                                        else if (RsDiff >= 15) {
                                                                                            return 8;
                                                                                        }
                                                                                    }
                                                                                }
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (Cs >= 2282) {
                                                        if (SC < 5403) {
                                                            if (posBalance < 54) {
                                                                if (mcTcor < 87) {
                                                                    if (CsDiff < 11) {
                                                                        return 4;
                                                                    }
                                                                    else if (CsDiff >= 11) {
                                                                        if (SC < 5066) {
                                                                            return 8;
                                                                        }
                                                                        else if (SC >= 5066) {
                                                                            return 2;
                                                                        }
                                                                    }
                                                                }
                                                                else if (mcTcor >= 87) {
                                                                    if (RsCsDiff < 115) {
                                                                        return 4;
                                                                    }
                                                                    else if (RsCsDiff >= 115) {
                                                                        if (diffMVdiffVal < 4) {
                                                                            return 1;
                                                                        }
                                                                        else if (diffMVdiffVal >= 4) {
                                                                            if (TSC < 12) {
                                                                                if (SC < 4777) {
                                                                                    return 8;
                                                                                }
                                                                                else if (SC >= 4777) {
                                                                                    return 4;
                                                                                }
                                                                            }
                                                                            else if (TSC >= 12) {
                                                                                return 1;
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                            else if (posBalance >= 54) {
                                                                if (MVDiff < 24) {
                                                                    if (ssDCval < 113) {
                                                                        if (refDCval < 92) {
                                                                            if (SC < 5122) {
                                                                                if (Cs < 2515) {
                                                                                    return 4;
                                                                                }
                                                                                else if (Cs >= 2515) {
                                                                                    return 8;
                                                                                }
                                                                            }
                                                                            else if (SC >= 5122) {
                                                                                if (SC < 5264) {
                                                                                    return 1;
                                                                                }
                                                                                else if (SC >= 5264) {
                                                                                    if (diffRsCsdiff < -2) {
                                                                                        return 8;
                                                                                    }
                                                                                    else if (diffRsCsdiff >= -2) {
                                                                                        return 1;
                                                                                    }
                                                                                }
                                                                            }
                                                                        }
                                                                        else if (refDCval >= 92) {
                                                                            if (Cs < 2333) {
                                                                                if (Rs < 2931) {
                                                                                    return 1;
                                                                                }
                                                                                else if (Rs >= 2931) {
                                                                                    return 4;
                                                                                }
                                                                            }
                                                                            else if (Cs >= 2333) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (ssDCval >= 113) {
                                                                        return 1;
                                                                    }
                                                                }
                                                                else if (MVDiff >= 24) {
                                                                    if (Rs < 2255) {
                                                                        if (TSC < 17) {
                                                                            return 1;
                                                                        }
                                                                        else if (TSC >= 17) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                    else if (Rs >= 2255) {
                                                                        if (mcTcor < 92) {
                                                                            if (Rs < 2777) {
                                                                                return 4;
                                                                            }
                                                                            else if (Rs >= 2777) {
                                                                                if (SC < 5203) {
                                                                                    return 1;
                                                                                }
                                                                                else if (SC >= 5203) {
                                                                                    return 4;
                                                                                }
                                                                            }
                                                                        }
                                                                        else if (mcTcor >= 92) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (SC >= 5403) {
                                                            if (tcor < 81) {
                                                                if (ssDCval < 82) {
                                                                    return 8;
                                                                }
                                                                else if (ssDCval >= 82) {
                                                                    return 1;
                                                                }
                                                            }
                                                            else if (tcor >= 81) {
                                                                if (Cs < 2453) {
                                                                    if (gchDC < 2) {
                                                                        if (Cs < 2438) {
                                                                            return 1;
                                                                        }
                                                                        else if (Cs >= 2438) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                    else if (gchDC >= 2) {
                                                                        return 2;
                                                                    }
                                                                }
                                                                else if (Cs >= 2453) {
                                                                    if (RsDiff < 10) {
                                                                        return 1;
                                                                    }
                                                                    else if (RsDiff >= 10) {
                                                                        if (mcTcor < 91) {
                                                                            if (Cs < 2610) {
                                                                                return 1;
                                                                            }
                                                                            else if (Cs >= 2610) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                        else if (mcTcor >= 91) {
                                                                            if (CsDiff < 10) {
                                                                                return 2;
                                                                            }
                                                                            else if (CsDiff >= 10) {
                                                                                return 1;
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (Cs >= 2823) {
                                                    if (tcor < 69) {
                                                        if (MVDiff < 38) {
                                                            return 1;
                                                        }
                                                        else if (MVDiff >= 38) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (tcor >= 69) {
                                                        if (posBalance < 61) {
                                                            if (tcor < 79) {
                                                                if (Rs < 2918) {
                                                                    return 1;
                                                                }
                                                                else if (Rs >= 2918) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (tcor >= 79) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (posBalance >= 61) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (ssDCval >= 140) {
                                                if (AFD < 17) {
                                                    return 8;
                                                }
                                                else if (AFD >= 17) {
                                                    if (diffAFD < 2) {
                                                        if (mcTcor < 93) {
                                                            return 4;
                                                        }
                                                        else if (mcTcor >= 93) {
                                                            if (SC < 5944) {
                                                                return 2;
                                                            }
                                                            else if (SC >= 5944) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                    else if (diffAFD >= 2) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                        }
                                        else if (tcor >= 90) {
                                            if (RsCsDiff < 129) {
                                                if (posBalance < 52) {
                                                    return 8;
                                                }
                                                else if (posBalance >= 52) {
                                                    if (Rs < 2536) {
                                                        if (TSC < 9) {
                                                            return 4;
                                                        }
                                                        else if (TSC >= 9) {
                                                            return 2;
                                                        }
                                                    }
                                                    else if (Rs >= 2536) {
                                                        if (ssDCval < 92) {
                                                            if (MVDiff < 11) {
                                                                if (Rs < 3072) {
                                                                    return 8;
                                                                }
                                                                else if (Rs >= 3072) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (MVDiff >= 11) {
                                                                if (SC < 5263) {
                                                                    return 1;
                                                                }
                                                                else if (SC >= 5263) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                        else if (ssDCval >= 92) {
                                                            return 1;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (RsCsDiff >= 129) {
                                                if (SC < 4884) {
                                                    return 4;
                                                }
                                                else if (SC >= 4884) {
                                                    if (refDCval < 86) {
                                                        if (TSC < 11) {
                                                            return 8;
                                                        }
                                                        else if (TSC >= 11) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (refDCval >= 86) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (mcTcor >= 94) {
                                        if (SC < 5411) {
                                            if (CsDiff < 8) {
                                                if (mu_mv_mag_sq < 54) {
                                                    if (MVDiff < 14) {
                                                        if (RsCsDiff < 107) {
                                                            if (mu_mv_mag_sq < 42) {
                                                                return 1;
                                                            }
                                                            else if (mu_mv_mag_sq >= 42) {
                                                                if (tcor < 90) {
                                                                    return 2;
                                                                }
                                                                else if (tcor >= 90) {
                                                                    return 1;
                                                                }
                                                            }
                                                        }
                                                        else if (RsCsDiff >= 107) {
                                                            return 2;
                                                        }
                                                    }
                                                    else if (MVDiff >= 14) {
                                                        if (Cs < 2008) {
                                                            return 2;
                                                        }
                                                        else if (Cs >= 2008) {
                                                            if (TSC < 8) {
                                                                return 4;
                                                            }
                                                            else if (TSC >= 8) {
                                                                return 1;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (mu_mv_mag_sq >= 54) {
                                                    if (diffMVdiffVal < -9) {
                                                        return 4;
                                                    }
                                                    else if (diffMVdiffVal >= -9) {
                                                        if (diffAFD < -2) {
                                                            return 4;
                                                        }
                                                        else if (diffAFD >= -2) {
                                                            return 1;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (CsDiff >= 8) {
                                                if (tcor < 89) {
                                                    if (TSCindex < 9) {
                                                        if (Rs < 2193) {
                                                            if (Rs < 1396) {
                                                                return 4;
                                                            }
                                                            else if (Rs >= 1396) {
                                                                return 1;
                                                            }
                                                        }
                                                        else if (Rs >= 2193) {
                                                            return 2;
                                                        }
                                                    }
                                                    else if (TSCindex >= 9) {
                                                        return 2;
                                                    }
                                                }
                                                else if (tcor >= 89) {
                                                    if (negBalance < 52) {
                                                        if (refDCval < 91) {
                                                            return 8;
                                                        }
                                                        else if (refDCval >= 91) {
                                                            if (tcor < 91) {
                                                                return 2;
                                                            }
                                                            else if (tcor >= 91) {
                                                                if (Cs < 2136) {
                                                                    return 8;
                                                                }
                                                                else if (Cs >= 2136) {
                                                                    return 2;
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (negBalance >= 52) {
                                                        if (SC < 4381) {
                                                            return 4;
                                                        }
                                                        else if (SC >= 4381) {
                                                            return 1;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (SC >= 5411) {
                                            if (tcor < 91) {
                                                if (tcor < 89) {
                                                    if (Rs < 3568) {
                                                        if (AFD < 13) {
                                                            return 4;
                                                        }
                                                        else if (AFD >= 13) {
                                                            return 2;
                                                        }
                                                    }
                                                    else if (Rs >= 3568) {
                                                        return 4;
                                                    }
                                                }
                                                else if (tcor >= 89) {
                                                    if (Cs < 2637) {
                                                        return 2;
                                                    }
                                                    else if (Cs >= 2637) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                            else if (tcor >= 91) {
                                                if (refDCval < 117) {
                                                    if (Rs < 3296) {
                                                        return 4;
                                                    }
                                                    else if (Rs >= 3296) {
                                                        return 2;
                                                    }
                                                }
                                                else if (refDCval >= 117) {
                                                    return 8;
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                        }
                    }
                }
                else if (Scindex >= 6) {
                    if (CsDiff < 10) {
                        if (tcor < 96) {
                            if (refDCval < 128) {
                                if (Rs < 3449) {
                                    if (RsCsDiff < 23) {
                                        if (TSCindex < 8) {
                                            if (posBalance < 58) {
                                                if (negBalance < 55) {
                                                    return 8;
                                                }
                                                else if (negBalance >= 55) {
                                                    if (Cs < 3180) {
                                                        return 8;
                                                    }
                                                    else if (Cs >= 3180) {
                                                        return 2;
                                                    }
                                                }
                                            }
                                            else if (posBalance >= 58) {
                                                if (MVDiff < 23) {
                                                    return 4;
                                                }
                                                else if (MVDiff >= 23) {
                                                    return 8;
                                                }
                                            }
                                        }
                                        else if (TSCindex >= 8) {
                                            if (Rs < 3323) {
                                                if (negBalance < 51) {
                                                    if (RsDiff < 3) {
                                                        return 2;
                                                    }
                                                    else if (RsDiff >= 3) {
                                                        return 8;
                                                    }
                                                }
                                                else if (negBalance >= 51) {
                                                    return 2;
                                                }
                                            }
                                            else if (Rs >= 3323) {
                                                if (Rs < 3431) {
                                                    return 2;
                                                }
                                                else if (Rs >= 3431) {
                                                    return 4;
                                                }
                                            }
                                        }
                                    }
                                    else if (RsCsDiff >= 23) {
                                        if (Cs < 2900) {
                                            return 1;
                                        }
                                        else if (Cs >= 2900) {
                                            if (SC < 6203) {
                                                return 4;
                                            }
                                            else if (SC >= 6203) {
                                                return 8;
                                            }
                                        }
                                    }
                                }
                                else if (Rs >= 3449) {
                                    if (mcTcor < 81) {
                                        if (refDCval < 107) {
                                            return 4;
                                        }
                                        else if (refDCval >= 107) {
                                            if (AFD < 14) {
                                                return 1;
                                            }
                                            else if (AFD >= 14) {
                                                if (AFD < 29) {
                                                    return 2;
                                                }
                                                else if (AFD >= 29) {
                                                    return 8;
                                                }
                                            }
                                        }
                                    }
                                    else if (mcTcor >= 81) {
                                        if (posBalance < 51) {
                                            if (ssDCval < 108) {
                                                if (negBalance < 41) {
                                                    if (diffMVdiffVal < -6) {
                                                        if (Rs < 4187) {
                                                            return 4;
                                                        }
                                                        else if (Rs >= 4187) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (diffMVdiffVal >= -6) {
                                                        if (Rs < 4469) {
                                                            return 8;
                                                        }
                                                        else if (Rs >= 4469) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                                else if (negBalance >= 41) {
                                                    if (ssDCval < 62) {
                                                        if (Rs < 4240) {
                                                            if (Rs < 3485) {
                                                                return 4;
                                                            }
                                                            else if (Rs >= 3485) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (Rs >= 4240) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (ssDCval >= 62) {
                                                        if (RsCsDiff < 67) {
                                                            return 4;
                                                        }
                                                        else if (RsCsDiff >= 67) {
                                                            if (RsCsDiff < 166) {
                                                                if (posBalance < 46) {
                                                                    return 8;
                                                                }
                                                                else if (posBalance >= 46) {
                                                                    if (ssDCval < 92) {
                                                                        return 8;
                                                                    }
                                                                    else if (ssDCval >= 92) {
                                                                        return 4;
                                                                    }
                                                                }
                                                            }
                                                            else if (RsCsDiff >= 166) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                            else if (ssDCval >= 108) {
                                                if (mcTcor < 86) {
                                                    if (Cs < 2944) {
                                                        return 1;
                                                    }
                                                    else if (Cs >= 2944) {
                                                        return 2;
                                                    }
                                                }
                                                else if (mcTcor >= 86) {
                                                    if (SC < 7214) {
                                                        if (SC < 6618) {
                                                            if (refDCval < 119) {
                                                                if (tcor < 92) {
                                                                    return 4;
                                                                }
                                                                else if (tcor >= 92) {
                                                                    if (MVDiff < 22) {
                                                                        return 2;
                                                                    }
                                                                    else if (MVDiff >= 22) {
                                                                        if (Cs < 2300) {
                                                                            return 4;
                                                                        }
                                                                        else if (Cs >= 2300) {
                                                                            return 2;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                            else if (refDCval >= 119) {
                                                                if (Rs < 3982) {
                                                                    return 8;
                                                                }
                                                                else if (Rs >= 3982) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                        else if (SC >= 6618) {
                                                            if (SC < 7101) {
                                                                if (SC < 6896) {
                                                                    return 4;
                                                                }
                                                                else if (SC >= 6896) {
                                                                    if (diffTSC < 1) {
                                                                        if (CsDiff < 8) {
                                                                            if (negBalance < 36) {
                                                                                if (diffRsCsdiff < 21) {
                                                                                    return 8;
                                                                                }
                                                                                else if (diffRsCsdiff >= 21) {
                                                                                    return 4;
                                                                                }
                                                                            }
                                                                            else if (negBalance >= 36) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                        else if (CsDiff >= 8) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                    else if (diffTSC >= 1) {
                                                                        return 4;
                                                                    }
                                                                }
                                                            }
                                                            else if (SC >= 7101) {
                                                                if (diffRsCsdiff < 14) {
                                                                    return 4;
                                                                }
                                                                else if (diffRsCsdiff >= 14) {
                                                                    if (posBalance < 40) {
                                                                        return 2;
                                                                    }
                                                                    else if (posBalance >= 40) {
                                                                        return 4;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (SC >= 7214) {
                                                        if (mu_mv_mag_sq < 73) {
                                                            if (mu_mv_mag_sq < 67) {
                                                                return 4;
                                                            }
                                                            else if (mu_mv_mag_sq >= 67) {
                                                                if (CsDiff < 8) {
                                                                    if (Cs < 2916) {
                                                                        return 8;
                                                                    }
                                                                    else if (Cs >= 2916) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (CsDiff >= 8) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                        else if (mu_mv_mag_sq >= 73) {
                                                            if (ssDCval < 125) {
                                                                if (refDCval < 112) {
                                                                    if (mcTcor < 89) {
                                                                        if (Rs < 5139) {
                                                                            return 2;
                                                                        }
                                                                        else if (Rs >= 5139) {
                                                                            if (RsCsDiff < 115) {
                                                                                return 2;
                                                                            }
                                                                            else if (RsCsDiff >= 115) {
                                                                                return 8;
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (mcTcor >= 89) {
                                                                        return 8;
                                                                    }
                                                                }
                                                                else if (refDCval >= 112) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (ssDCval >= 125) {
                                                                return 2;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (posBalance >= 51) {
                                            if (TSC < 7) {
                                                if (tcor < 89) {
                                                    return 8;
                                                }
                                                else if (tcor >= 89) {
                                                    if (refDCval < 54) {
                                                        if (MVDiff < 27) {
                                                            if (negBalance < 52) {
                                                                if (RsCsDiff < 11) {
                                                                    if (Cs < 3458) {
                                                                        if (MVDiff < 6) {
                                                                            return 8;
                                                                        }
                                                                        else if (MVDiff >= 6) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                    else if (Cs >= 3458) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (RsCsDiff >= 11) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (negBalance >= 52) {
                                                                return 2;
                                                            }
                                                        }
                                                        else if (MVDiff >= 27) {
                                                            if (CsDiff < 3) {
                                                                return 8;
                                                            }
                                                            else if (CsDiff >= 3) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                    else if (refDCval >= 54) {
                                                        if (Rs < 4027) {
                                                            if (Cs < 3383) {
                                                                return 4;
                                                            }
                                                            else if (Cs >= 3383) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (Rs >= 4027) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (TSC >= 7) {
                                                if (RsCsDiff < 46) {
                                                    if (SC < 7760) {
                                                        if (MVDiff < 17) {
                                                            if (posBalance < 60) {
                                                                if (RsDiff < 4) {
                                                                    if (diffMVdiffVal < -4) {
                                                                        return 2;
                                                                    }
                                                                    else if (diffMVdiffVal >= -4) {
                                                                        if (mcTcor < 93) {
                                                                            return 2;
                                                                        }
                                                                        else if (mcTcor >= 93) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                }
                                                                else if (RsDiff >= 4) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (posBalance >= 60) {
                                                                if (Rs < 3570) {
                                                                    return 8;
                                                                }
                                                                else if (Rs >= 3570) {
                                                                    return 2;
                                                                }
                                                            }
                                                        }
                                                        else if (MVDiff >= 17) {
                                                            if (tcor < 86) {
                                                                if (Cs < 3811) {
                                                                    return 4;
                                                                }
                                                                else if (Cs >= 3811) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (tcor >= 86) {
                                                                if (ssDCval < 58) {
                                                                    if (mu_mv_mag_sq < 45) {
                                                                        if (Cs < 3505) {
                                                                            if (gchDC < 1) {
                                                                                return 2;
                                                                            }
                                                                            else if (gchDC >= 1) {
                                                                                if (Rs < 3544) {
                                                                                    return 2;
                                                                                }
                                                                                else if (Rs >= 3544) {
                                                                                    return 8;
                                                                                }
                                                                            }
                                                                        }
                                                                        else if (Cs >= 3505) {
                                                                            return 2;
                                                                        }
                                                                    }
                                                                    else if (mu_mv_mag_sq >= 45) {
                                                                        if (AFD < 10) {
                                                                            return 8;
                                                                        }
                                                                        else if (AFD >= 10) {
                                                                            return 2;
                                                                        }
                                                                    }
                                                                }
                                                                else if (ssDCval >= 58) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (SC >= 7760) {
                                                        if (negBalance < 51) {
                                                            return 4;
                                                        }
                                                        else if (negBalance >= 51) {
                                                            return 2;
                                                        }
                                                    }
                                                }
                                                else if (RsCsDiff >= 46) {
                                                    if (ssDCval < 104) {
                                                        if (ssDCval < 66) {
                                                            if (MVDiff < 31) {
                                                                if (TSC < 8) {
                                                                    if (Rs < 4035) {
                                                                        return 8;
                                                                    }
                                                                    else if (Rs >= 4035) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (TSC >= 8) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (MVDiff >= 31) {
                                                                if (CsDiff < 7) {
                                                                    if (diffAFD < 1) {
                                                                        return 4;
                                                                    }
                                                                    else if (diffAFD >= 1) {
                                                                        return 8;
                                                                    }
                                                                }
                                                                else if (CsDiff >= 7) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                        else if (ssDCval >= 66) {
                                                            if (ssDCval < 95) {
                                                                if (SC < 10996) {
                                                                    return 8;
                                                                }
                                                                else if (SC >= 10996) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (ssDCval >= 95) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                    else if (ssDCval >= 104) {
                                                        if (ssDCval < 109) {
                                                            if (RsDiff < 10) {
                                                                if (Rs < 4531) {
                                                                    return 4;
                                                                }
                                                                else if (Rs >= 4531) {
                                                                    return 2;
                                                                }
                                                            }
                                                            else if (RsDiff >= 10) {
                                                                if (Rs < 5569) {
                                                                    return 2;
                                                                }
                                                                else if (Rs >= 5569) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                        else if (ssDCval >= 109) {
                                                            if (Cs < 2555) {
                                                                return 4;
                                                            }
                                                            else if (Cs >= 2555) {
                                                                if (mcTcor < 85) {
                                                                    return 4;
                                                                }
                                                                else if (mcTcor >= 85) {
                                                                    if (Cs < 2947) {
                                                                        return 1;
                                                                    }
                                                                    else if (Cs >= 2947) {
                                                                        return 8;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                            else if (refDCval >= 128) {
                                if (Rs < 4010) {
                                    if (MVDiff < 15) {
                                        return 1;
                                    }
                                    else if (MVDiff >= 15) {
                                        return 8;
                                    }
                                }
                                else if (Rs >= 4010) {
                                    if (AFD < 24) {
                                        if (Cs < 2928) {
                                            if (AFD < 7) {
                                                return 4;
                                            }
                                            else if (AFD >= 7) {
                                                return 2;
                                            }
                                        }
                                        else if (Cs >= 2928) {
                                            return 4;
                                        }
                                    }
                                    else if (AFD >= 24) {
                                        return 4;
                                    }
                                }
                            }
                        }
                        else if (tcor >= 96) {
                            if (negBalance < 44) {
                                return 8;
                            }
                            else if (negBalance >= 44) {
                                if (Rs < 3526) {
                                    return 8;
                                }
                                else if (Rs >= 3526) {
                                    if (negBalance < 47) {
                                        return 4;
                                    }
                                    else if (negBalance >= 47) {
                                        return 8;
                                    }
                                }
                            }
                        }
                    }
                    else if (CsDiff >= 10) {
                        if (Cs < 3530) {
                            if (mcTcor < 87) {
                                if (ssDCval < 127) {
                                    if (AFD < 22) {
                                        if (refDCval < 106) {
                                            if (mcTcor < 82) {
                                                if (ssDCval < 67) {
                                                    return 8;
                                                }
                                                else if (ssDCval >= 67) {
                                                    if (AFD < 19) {
                                                        return 4;
                                                    }
                                                    else if (AFD >= 19) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                            else if (mcTcor >= 82) {
                                                if (AFD < 17) {
                                                    return 4;
                                                }
                                                else if (AFD >= 17) {
                                                    return 1;
                                                }
                                            }
                                        }
                                        else if (refDCval >= 106) {
                                            if (ssDCval < 110) {
                                                return 2;
                                            }
                                            else if (ssDCval >= 110) {
                                                if (TSC < 14) {
                                                    if (Rs < 5096) {
                                                        return 1;
                                                    }
                                                    else if (Rs >= 5096) {
                                                        if (Cs < 2761) {
                                                            return 1;
                                                        }
                                                        else if (Cs >= 2761) {
                                                            return 2;
                                                        }
                                                    }
                                                }
                                                else if (TSC >= 14) {
                                                    if (TSC < 15) {
                                                        return 4;
                                                    }
                                                    else if (TSC >= 15) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (AFD >= 22) {
                                        if (tcor < 54) {
                                            if (AFD < 35) {
                                                return 2;
                                            }
                                            else if (AFD >= 35) {
                                                return 8;
                                            }
                                        }
                                        else if (tcor >= 54) {
                                            if (ssDCval < 78) {
                                                return 1;
                                            }
                                            else if (ssDCval >= 78) {
                                                if (refDCval < 126) {
                                                    return 8;
                                                }
                                                else if (refDCval >= 126) {
                                                    return 1;
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (ssDCval >= 127) {
                                    if (ssDCval < 150) {
                                        if (Cs < 3365) {
                                            if (CsDiff < 11) {
                                                if (posBalance < 49) {
                                                    return 8;
                                                }
                                                else if (posBalance >= 49) {
                                                    return 1;
                                                }
                                            }
                                            else if (CsDiff >= 11) {
                                                return 1;
                                            }
                                        }
                                        else if (Cs >= 3365) {
                                            if (Rs < 4654) {
                                                return 1;
                                            }
                                            else if (Rs >= 4654) {
                                                if (RsCsDiff < 625) {
                                                    return 8;
                                                }
                                                else if (RsCsDiff >= 625) {
                                                    return 4;
                                                }
                                            }
                                        }
                                    }
                                    else if (ssDCval >= 150) {
                                        if (Rs < 6374) {
                                            return 2;
                                        }
                                        else if (Rs >= 6374) {
                                            return 4;
                                        }
                                    }
                                }
                            }
                            else if (mcTcor >= 87) {
                                if (mu_mv_mag_sq < 37) {
                                    if (Rs < 3631) {
                                        if (ssDCval < 104) {
                                            if (MVDiff < 10) {
                                                return 4;
                                            }
                                            else if (MVDiff >= 10) {
                                                return 1;
                                            }
                                        }
                                        else if (ssDCval >= 104) {
                                            return 8;
                                        }
                                    }
                                    else if (Rs >= 3631) {
                                        if (RsCsDiff < 428) {
                                            if (refDCval < 157) {
                                                if (posBalance < 51) {
                                                    return 8;
                                                }
                                                else if (posBalance >= 51) {
                                                    if (posBalance < 62) {
                                                        if (SC < 6448) {
                                                            return 2;
                                                        }
                                                        else if (SC >= 6448) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (posBalance >= 62) {
                                                        return 1;
                                                    }
                                                }
                                            }
                                            else if (refDCval >= 157) {
                                                return 2;
                                            }
                                        }
                                        else if (RsCsDiff >= 428) {
                                            if (tcor < 60) {
                                                if (Rs < 5381) {
                                                    return 2;
                                                }
                                                else if (Rs >= 5381) {
                                                    if (SC < 9307) {
                                                        return 1;
                                                    }
                                                    else if (SC >= 9307) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                            else if (tcor >= 60) {
                                                if (CsDiff < 20) {
                                                    return 4;
                                                }
                                                else if (CsDiff >= 20) {
                                                    if (Rs < 4730) {
                                                        return 4;
                                                    }
                                                    else if (Rs >= 4730) {
                                                        return 2;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (mu_mv_mag_sq >= 37) {
                                    if (refDCval < 112) {
                                        if (Rs < 3879) {
                                            if (SC < 6243) {
                                                if (AFD < 13) {
                                                    return 8;
                                                }
                                                else if (AFD >= 13) {
                                                    return 4;
                                                }
                                            }
                                            else if (SC >= 6243) {
                                                return 8;
                                            }
                                        }
                                        else if (Rs >= 3879) {
                                            if (mcTcor < 88) {
                                                return 8;
                                            }
                                            else if (mcTcor >= 88) {
                                                return 4;
                                            }
                                        }
                                    }
                                    else if (refDCval >= 112) {
                                        if (MVDiff < 9) {
                                            if (Cs < 3478) {
                                                if (mu_mv_mag_sq < 49) {
                                                    return 2;
                                                }
                                                else if (mu_mv_mag_sq >= 49) {
                                                    return 8;
                                                }
                                            }
                                            else if (Cs >= 3478) {
                                                return 2;
                                            }
                                        }
                                        else if (MVDiff >= 9) {
                                            return 8;
                                        }
                                    }
                                }
                            }
                        }
                        else if (Cs >= 3530) {
                            if (refDCval < 153) {
                                if (mu_mv_mag_sq < 63) {
                                    if (SC < 12154) {
                                        if (Rs < 4237) {
                                            if (mcTcor < 71) {
                                                return 1;
                                            }
                                            else if (mcTcor >= 71) {
                                                if (refDCval < 135) {
                                                    if (CsDiff < 21) {
                                                        if (TSC < 15) {
                                                            return 8;
                                                        }
                                                        else if (TSC >= 15) {
                                                            if (ssDCval < 94) {
                                                                return 8;
                                                            }
                                                            else if (ssDCval >= 94) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                    else if (CsDiff >= 21) {
                                                        return 4;
                                                    }
                                                }
                                                else if (refDCval >= 135) {
                                                    return 1;
                                                }
                                            }
                                        }
                                        else if (Rs >= 4237) {
                                            if (mcTcor < 91) {
                                                if (Cs < 4327) {
                                                    if (refDCval < 110) {
                                                        if (gchDC < 3) {
                                                            return 8;
                                                        }
                                                        else if (gchDC >= 3) {
                                                            return 2;
                                                        }
                                                    }
                                                    else if (refDCval >= 110) {
                                                        if (ssDCval < 140) {
                                                            if (tcor < 78) {
                                                                if (gchDC < 4) {
                                                                    if (diffMVdiffVal < 2) {
                                                                        return 4;
                                                                    }
                                                                    else if (diffMVdiffVal >= 2) {
                                                                        if (posBalance < 52) {
                                                                            return 8;
                                                                        }
                                                                        else if (posBalance >= 52) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                }
                                                                else if (gchDC >= 4) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (tcor >= 78) {
                                                                if (Rs < 5582) {
                                                                    return 8;
                                                                }
                                                                else if (Rs >= 5582) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                        else if (ssDCval >= 140) {
                                                            if (Rs < 6406) {
                                                                return 8;
                                                            }
                                                            else if (Rs >= 6406) {
                                                                return 2;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (Cs >= 4327) {
                                                    if (RsDiff < 38) {
                                                        if (MVDiff < 32) {
                                                            if (AFD < 16) {
                                                                return 4;
                                                            }
                                                            else if (AFD >= 16) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (MVDiff >= 32) {
                                                            if (CsDiff < 20) {
                                                                return 4;
                                                            }
                                                            else if (CsDiff >= 20) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                    else if (RsDiff >= 38) {
                                                        return 2;
                                                    }
                                                }
                                            }
                                            else if (mcTcor >= 91) {
                                                if (MVDiff < 8) {
                                                    if (diffRsCsdiff < 78) {
                                                        return 8;
                                                    }
                                                    else if (diffRsCsdiff >= 78) {
                                                        if (CsDiff < 20) {
                                                            return 8;
                                                        }
                                                        else if (CsDiff >= 20) {
                                                            return 1;
                                                        }
                                                    }
                                                }
                                                else if (MVDiff >= 8) {
                                                    if (ssDCval < 113) {
                                                        if (AFD < 18) {
                                                            return 4;
                                                        }
                                                        else if (AFD >= 18) {
                                                            return 1;
                                                        }
                                                    }
                                                    else if (ssDCval >= 113) {
                                                        if (negBalance < 54) {
                                                            return 8;
                                                        }
                                                        else if (negBalance >= 54) {
                                                            if (posBalance < 50) {
                                                                return 4;
                                                            }
                                                            else if (posBalance >= 50) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (SC >= 12154) {
                                        if (refDCval < 91) {
                                            if (refDCval < 79) {
                                                return 4;
                                            }
                                            else if (refDCval >= 79) {
                                                if (mcTcor < 87) {
                                                    return 4;
                                                }
                                                else if (mcTcor >= 87) {
                                                    return 2;
                                                }
                                            }
                                        }
                                        else if (refDCval >= 91) {
                                            if (mcTcor < 91) {
                                                if (MVDiff < 22) {
                                                    if (RsDiff < 33) {
                                                        if (tcor < 82) {
                                                            if (MVDiff < 20) {
                                                                return 8;
                                                            }
                                                            else if (MVDiff >= 20) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (tcor >= 82) {
                                                            if (ssDCval < 138) {
                                                                return 8;
                                                            }
                                                            else if (ssDCval >= 138) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                    else if (RsDiff >= 33) {
                                                        if (negBalance < 54) {
                                                            if (refDCval < 135) {
                                                                return 8;
                                                            }
                                                            else if (refDCval >= 135) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (negBalance >= 54) {
                                                            if (Rs < 15255) {
                                                                if (diffMVdiffVal < -2) {
                                                                    if (AFD < 36) {
                                                                        if (CsDiff < 36) {
                                                                            return 4;
                                                                        }
                                                                        else if (CsDiff >= 36) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                    else if (AFD >= 36) {
                                                                        return 2;
                                                                    }
                                                                }
                                                                else if (diffMVdiffVal >= -2) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (Rs >= 15255) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (MVDiff >= 22) {
                                                    if (Rs < 8141) {
                                                        if (ssDCval < 119) {
                                                            return 2;
                                                        }
                                                        else if (ssDCval >= 119) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (Rs >= 8141) {
                                                        return 2;
                                                    }
                                                }
                                            }
                                            else if (mcTcor >= 91) {
                                                if (MVDiff < 21) {
                                                    return 4;
                                                }
                                                else if (MVDiff >= 21) {
                                                    return 2;
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (mu_mv_mag_sq >= 63) {
                                    if (mcTcor < 71) {
                                        if (ssDCval < 112) {
                                            return 8;
                                        }
                                        else if (ssDCval >= 112) {
                                            return 1;
                                        }
                                    }
                                    else if (mcTcor >= 71) {
                                        if (CsDiff < 14) {
                                            if (tcor < 77) {
                                                if (MVDiff < 19) {
                                                    return 4;
                                                }
                                                else if (MVDiff >= 19) {
                                                    return 8;
                                                }
                                            }
                                            else if (tcor >= 77) {
                                                if (RsCsDiff < 353) {
                                                    return 8;
                                                }
                                                else if (RsCsDiff >= 353) {
                                                    if (RsCsDiff < 380) {
                                                        return 4;
                                                    }
                                                    else if (RsCsDiff >= 380) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                        }
                                        else if (CsDiff >= 14) {
                                            if (negBalance < 64) {
                                                if (Rs < 3669) {
                                                    return 1;
                                                }
                                                else if (Rs >= 3669) {
                                                    return 8;
                                                }
                                            }
                                            else if (negBalance >= 64) {
                                                if (SC < 13988) {
                                                    return 8;
                                                }
                                                else if (SC >= 13988) {
                                                    return 4;
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                            else if (refDCval >= 153) {
                                if (CsDiff < 24) {
                                    if (posBalance < 58) {
                                        if (tcor < 71) {
                                            if (mcTcor < 86) {
                                                if (Cs < 4439) {
                                                    return 4;
                                                }
                                                else if (Cs >= 4439) {
                                                    return 8;
                                                }
                                            }
                                            else if (mcTcor >= 86) {
                                                return 4;
                                            }
                                        }
                                        else if (tcor >= 71) {
                                            return 8;
                                        }
                                    }
                                    else if (posBalance >= 58) {
                                        if (Rs < 7751) {
                                            if (Cs < 3744) {
                                                if (RsDiff < 19) {
                                                    return 2;
                                                }
                                                else if (RsDiff >= 19) {
                                                    return 4;
                                                }
                                            }
                                            else if (Cs >= 3744) {
                                                return 2;
                                            }
                                        }
                                        else if (Rs >= 7751) {
                                            return 8;
                                        }
                                    }
                                }
                                else if (CsDiff >= 24) {
                                    if (AFD < 34) {
                                        return 8;
                                    }
                                    else if (AFD >= 34) {
                                        return 2;
                                    }
                                }
                            }
                        }
                    }
                }
            }
        }
        else if (mu_mv_mag_sq >= 89) {
            if (AFD < 7) {
                if (SC < 1269) {
                    if (refDCval < 161) {
                        if (ssDCval < 94) {
                            if (refDCval < 90) {
                                if (SC < 481) {
                                    if (diffTSC < -5) {
                                        return 1;
                                    }
                                    else if (diffTSC >= -5) {
                                        if (diffRsCsdiff < -1) {
                                            return 8;
                                        }
                                        else if (diffRsCsdiff >= -1) {
                                            if (MVDiff < 1) {
                                                if (Cs < 303) {
                                                    return 1;
                                                }
                                                else if (Cs >= 303) {
                                                    return 4;
                                                }
                                            }
                                            else if (MVDiff >= 1) {
                                                return 4;
                                            }
                                        }
                                    }
                                }
                                else if (SC >= 481) {
                                    if (negBalance < 35) {
                                        if (mcTcor < 100) {
                                            if (Cs < 641) {
                                                if (Cs < 345) {
                                                    if (Cs < 309) {
                                                        if (Cs < 231) {
                                                            return 4;
                                                        }
                                                        else if (Cs >= 231) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (Cs >= 309) {
                                                        return 4;
                                                    }
                                                }
                                                else if (Cs >= 345) {
                                                    if (negBalance < 11) {
                                                        if (tcor < 99) {
                                                            return 8;
                                                        }
                                                        else if (tcor >= 99) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (negBalance >= 11) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                            else if (Cs >= 641) {
                                                return 4;
                                            }
                                        }
                                        else if (mcTcor >= 100) {
                                            return 8;
                                        }
                                    }
                                    else if (negBalance >= 35) {
                                        return 4;
                                    }
                                }
                            }
                            else if (refDCval >= 90) {
                                return 4;
                            }
                        }
                        else if (ssDCval >= 94) {
                            if (diffRsCsdiff < -237) {
                                if (Rs < 430) {
                                    return 8;
                                }
                                else if (Rs >= 430) {
                                    return 4;
                                }
                            }
                            else if (diffRsCsdiff >= -237) {
                                if (Cs < 697) {
                                    if (diffTSC < 2) {
                                        return 8;
                                    }
                                    else if (diffTSC >= 2) {
                                        return 4;
                                    }
                                }
                                else if (Cs >= 697) {
                                    if (Cs < 914) {
                                        return 2;
                                    }
                                    else if (Cs >= 914) {
                                        return 8;
                                    }
                                }
                            }
                        }
                    }
                    else if (refDCval >= 161) {
                        if (Cs < 443) {
                            if (TSCindex < 2) {
                                return 4;
                            }
                            else if (TSCindex >= 2) {
                                if (Rs < 452) {
                                    return 4;
                                }
                                else if (Rs >= 452) {
                                    return 2;
                                }
                            }
                        }
                        else if (Cs >= 443) {
                            if (MVDiff < 6) {
                                if (SC < 997) {
                                    if (ssDCval < 163) {
                                        return 4;
                                    }
                                    else if (ssDCval >= 163) {
                                        return 2;
                                    }
                                }
                                else if (SC >= 997) {
                                    return 4;
                                }
                            }
                            else if (MVDiff >= 6) {
                                return 2;
                            }
                        }
                    }
                }
                else if (SC >= 1269) {
                    if (AFD < 4) {
                        if (ssDCval < 103) {
                            if (mu_mv_mag_sq < 117) {
                                if (TSCindex < 1) {
                                    if (ssDCval < 31) {
                                        if (posBalance < 20) {
                                            if (Rs < 1413) {
                                                return 4;
                                            }
                                            else if (Rs >= 1413) {
                                                return 8;
                                            }
                                        }
                                        else if (posBalance >= 20) {
                                            return 8;
                                        }
                                    }
                                    else if (ssDCval >= 31) {
                                        if (Rs < 1047) {
                                            if (Rs < 813) {
                                                return 1;
                                            }
                                            else if (Rs >= 813) {
                                                return 4;
                                            }
                                        }
                                        else if (Rs >= 1047) {
                                            if (diffMVdiffVal < 2) {
                                                return 2;
                                            }
                                            else if (diffMVdiffVal >= 2) {
                                                return 4;
                                            }
                                        }
                                    }
                                }
                                else if (TSCindex >= 1) {
                                    if (Cs < 2370) {
                                        if (diffAFD < -1) {
                                            if (Cs < 818) {
                                                return 8;
                                            }
                                            else if (Cs >= 818) {
                                                if (Cs < 1037) {
                                                    return 1;
                                                }
                                                else if (Cs >= 1037) {
                                                    return 8;
                                                }
                                            }
                                        }
                                        else if (diffAFD >= -1) {
                                            return 8;
                                        }
                                    }
                                    else if (Cs >= 2370) {
                                        if (Rs < 2330) {
                                            if (Rs < 1781) {
                                                return 8;
                                            }
                                            else if (Rs >= 1781) {
                                                return 1;
                                            }
                                        }
                                        else if (Rs >= 2330) {
                                            return 8;
                                        }
                                    }
                                }
                            }
                            else if (mu_mv_mag_sq >= 117) {
                                if (diffMVdiffVal < 2) {
                                    if (mu_mv_mag_sq < 125) {
                                        if (MVDiff < 9) {
                                            if (refDCval < 51) {
                                                if (negBalance < 18) {
                                                    if (SC < 1823) {
                                                        return 4;
                                                    }
                                                    else if (SC >= 1823) {
                                                        return 8;
                                                    }
                                                }
                                                else if (negBalance >= 18) {
                                                    if (ssDCval < 49) {
                                                        return 8;
                                                    }
                                                    else if (ssDCval >= 49) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                            else if (refDCval >= 51) {
                                                return 8;
                                            }
                                        }
                                        else if (MVDiff >= 9) {
                                            if (ssDCval < 38) {
                                                return 8;
                                            }
                                            else if (ssDCval >= 38) {
                                                return 1;
                                            }
                                        }
                                    }
                                    else if (mu_mv_mag_sq >= 125) {
                                        return 8;
                                    }
                                }
                                else if (diffMVdiffVal >= 2) {
                                    if (diffAFD < -1) {
                                        if (AFD < 1) {
                                            if (Rs < 1454) {
                                                return 4;
                                            }
                                            else if (Rs >= 1454) {
                                                return 2;
                                            }
                                        }
                                        else if (AFD >= 1) {
                                            return 1;
                                        }
                                    }
                                    else if (diffAFD >= -1) {
                                        if (refDCval < 28) {
                                            return 4;
                                        }
                                        else if (refDCval >= 28) {
                                            return 8;
                                        }
                                    }
                                }
                            }
                        }
                        else if (ssDCval >= 103) {
                            if (negBalance < 3) {
                                if (SC < 7800) {
                                    if (Cs < 4003) {
                                        if (Rs < 3730) {
                                            if (Rs < 3721) {
                                                if (posBalance < 2) {
                                                    if (SC < 2916) {
                                                        if (Rs < 949) {
                                                            return 8;
                                                        }
                                                        else if (Rs >= 949) {
                                                            return 1;
                                                        }
                                                    }
                                                    else if (SC >= 2916) {
                                                        if (Cs < 1974) {
                                                            return 8;
                                                        }
                                                        else if (Cs >= 1974) {
                                                            return 1;
                                                        }
                                                    }
                                                }
                                                else if (posBalance >= 2) {
                                                    if (SC < 7308) {
                                                        return 8;
                                                    }
                                                    else if (SC >= 7308) {
                                                        if (Cs < 3849) {
                                                            return 1;
                                                        }
                                                        else if (Cs >= 3849) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (Rs >= 3721) {
                                                if (RsCsDiff < 2) {
                                                    return 8;
                                                }
                                                else if (RsCsDiff >= 2) {
                                                    return 1;
                                                }
                                            }
                                        }
                                        else if (Rs >= 3730) {
                                            if (Cs < 2780) {
                                                return 2;
                                            }
                                            else if (Cs >= 2780) {
                                                return 8;
                                            }
                                        }
                                    }
                                    else if (Cs >= 4003) {
                                        return 1;
                                    }
                                }
                                else if (SC >= 7800) {
                                    return 8;
                                }
                            }
                            else if (negBalance >= 3) {
                                if (CsDiff < 3) {
                                    if (posBalance < 18) {
                                        if (ssDCval < 120) {
                                            if (refDCval < 116) {
                                                return 8;
                                            }
                                            else if (refDCval >= 116) {
                                                if (MVDiff < 1) {
                                                    if (negBalance < 9) {
                                                        if (SC < 5682) {
                                                            return 4;
                                                        }
                                                        else if (SC >= 5682) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (negBalance >= 9) {
                                                        return 8;
                                                    }
                                                }
                                                else if (MVDiff >= 1) {
                                                    return 4;
                                                }
                                            }
                                        }
                                        else if (ssDCval >= 120) {
                                            if (diffMVdiffVal < -3) {
                                                return 1;
                                            }
                                            else if (diffMVdiffVal >= -3) {
                                                if (MVDiff < 8) {
                                                    if (diffMVdiffVal < -2) {
                                                        if (Cs < 754) {
                                                            return 1;
                                                        }
                                                        else if (Cs >= 754) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (diffMVdiffVal >= -2) {
                                                        return 8;
                                                    }
                                                }
                                                else if (MVDiff >= 8) {
                                                    return 1;
                                                }
                                            }
                                        }
                                    }
                                    else if (posBalance >= 18) {
                                        if (diffMVdiffVal < 12) {
                                            if (mcTcor < 97) {
                                                if (ssDCval < 120) {
                                                    if (diffTSC < 1) {
                                                        return 4;
                                                    }
                                                    else if (diffTSC >= 1) {
                                                        return 8;
                                                    }
                                                }
                                                else if (ssDCval >= 120) {
                                                    return 8;
                                                }
                                            }
                                            else if (mcTcor >= 97) {
                                                if (negBalance < 39) {
                                                    return 8;
                                                }
                                                else if (negBalance >= 39) {
                                                    if (Cs < 2179) {
                                                        return 8;
                                                    }
                                                    else if (Cs >= 2179) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                        }
                                        else if (diffMVdiffVal >= 12) {
                                            return 4;
                                        }
                                    }
                                }
                                else if (CsDiff >= 3) {
                                    if (SC < 6417) {
                                        if (Cs < 2935) {
                                            if (mcTcor < 70) {
                                                if (Cs < 1490) {
                                                    return 1;
                                                }
                                                else if (Cs >= 1490) {
                                                    return 4;
                                                }
                                            }
                                            else if (mcTcor >= 70) {
                                                if (CsDiff < 4) {
                                                    return 8;
                                                }
                                                else if (CsDiff >= 4) {
                                                    if (mu_mv_mag_sq < 114) {
                                                        return 8;
                                                    }
                                                    else if (mu_mv_mag_sq >= 114) {
                                                        if (Rs < 1706) {
                                                            return 4;
                                                        }
                                                        else if (Rs >= 1706) {
                                                            return 1;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (Cs >= 2935) {
                                            if (Rs < 3162) {
                                                if (Cs < 2980) {
                                                    return 1;
                                                }
                                                else if (Cs >= 2980) {
                                                    return 8;
                                                }
                                            }
                                            else if (Rs >= 3162) {
                                                if (SC < 6281) {
                                                    if (refDCval < 108) {
                                                        return 8;
                                                    }
                                                    else if (refDCval >= 108) {
                                                        if (ssDCval < 119) {
                                                            if (mu_mv_mag_sq < 115) {
                                                                return 8;
                                                            }
                                                            else if (mu_mv_mag_sq >= 115) {
                                                                return 1;
                                                            }
                                                        }
                                                        else if (ssDCval >= 119) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                                else if (SC >= 6281) {
                                                    if (Cs < 3064) {
                                                        return 1;
                                                    }
                                                    else if (Cs >= 3064) {
                                                        if (SC < 6362) {
                                                            return 8;
                                                        }
                                                        else if (SC >= 6362) {
                                                            return 1;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (SC >= 6417) {
                                        if (mcTcor < 96) {
                                            if (diffTSC < -1) {
                                                return 4;
                                            }
                                            else if (diffTSC >= -1) {
                                                return 8;
                                            }
                                        }
                                        else if (mcTcor >= 96) {
                                            return 8;
                                        }
                                    }
                                }
                            }
                        }
                    }
                    else if (AFD >= 4) {
                        if (ssDCval < 88) {
                            if (negBalance < 35) {
                                if (mu_mv_mag_sq < 109) {
                                    if (MVDiff < 2) {
                                        if (Rs < 1415) {
                                            return 8;
                                        }
                                        else if (Rs >= 1415) {
                                            if (TSC < 5) {
                                                return 1;
                                            }
                                            else if (TSC >= 5) {
                                                return 8;
                                            }
                                        }
                                    }
                                    else if (MVDiff >= 2) {
                                        if (Cs < 709) {
                                            return 4;
                                        }
                                        else if (Cs >= 709) {
                                            if (diffAFD < -1) {
                                                if (TSC < 4) {
                                                    if (Cs < 1424) {
                                                        return 1;
                                                    }
                                                    else if (Cs >= 1424) {
                                                        return 4;
                                                    }
                                                }
                                                else if (TSC >= 4) {
                                                    return 8;
                                                }
                                            }
                                            else if (diffAFD >= -1) {
                                                return 8;
                                            }
                                        }
                                    }
                                }
                                else if (mu_mv_mag_sq >= 109) {
                                    if (refDCval < 85) {
                                        return 8;
                                    }
                                    else if (refDCval >= 85) {
                                        return 4;
                                    }
                                }
                            }
                            else if (negBalance >= 35) {
                                if (Rs < 1762) {
                                    if (MVDiff < 8) {
                                        if (TSCindex < 7) {
                                            if (SC < 3441) {
                                                return 8;
                                            }
                                            else if (SC >= 3441) {
                                                if (Rs < 1708) {
                                                    return 8;
                                                }
                                                else if (Rs >= 1708) {
                                                    if (ssDCval < 78) {
                                                        return 4;
                                                    }
                                                    else if (ssDCval >= 78) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                        }
                                        else if (TSCindex >= 7) {
                                            if (ssDCval < 72) {
                                                if (posBalance < 56) {
                                                    return 8;
                                                }
                                                else if (posBalance >= 56) {
                                                    return 2;
                                                }
                                            }
                                            else if (ssDCval >= 72) {
                                                if (RsCsDiff < 29) {
                                                    return 2;
                                                }
                                                else if (RsCsDiff >= 29) {
                                                    if (diffTSC < -1) {
                                                        return 2;
                                                    }
                                                    else if (diffTSC >= -1) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (MVDiff >= 8) {
                                        if (SC < 1693) {
                                            if (Rs < 946) {
                                                return 4;
                                            }
                                            else if (Rs >= 946) {
                                                return 8;
                                            }
                                        }
                                        else if (SC >= 1693) {
                                            if (diffRsCsdiff < -3) {
                                                return 8;
                                            }
                                            else if (diffRsCsdiff >= -3) {
                                                if (RsDiff < 2) {
                                                    return 1;
                                                }
                                                else if (RsDiff >= 2) {
                                                    return 8;
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (Rs >= 1762) {
                                    if (Rs < 2558) {
                                        if (tcor < 96) {
                                            if (mcTcor < 95) {
                                                if (mu_mv_mag_sq < 112) {
                                                    if (diffTSC < 1) {
                                                        return 4;
                                                    }
                                                    else if (diffTSC >= 1) {
                                                        return 8;
                                                    }
                                                }
                                                else if (mu_mv_mag_sq >= 112) {
                                                    if (Rs < 1769) {
                                                        if (Rs < 1766) {
                                                            return 8;
                                                        }
                                                        else if (Rs >= 1766) {
                                                            return 2;
                                                        }
                                                    }
                                                    else if (Rs >= 1769) {
                                                        if (Rs < 1988) {
                                                            return 8;
                                                        }
                                                        else if (Rs >= 1988) {
                                                            if (SC < 3671) {
                                                                return 2;
                                                            }
                                                            else if (SC >= 3671) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                            else if (mcTcor >= 95) {
                                                if (refDCval < 75) {
                                                    if (RsCsDiff < 29) {
                                                        if (SC < 3758) {
                                                            return 2;
                                                        }
                                                        else if (SC >= 3758) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (RsCsDiff >= 29) {
                                                        if (negBalance < 44) {
                                                            return 4;
                                                        }
                                                        else if (negBalance >= 44) {
                                                            if (Rs < 1803) {
                                                                return 8;
                                                            }
                                                            else if (Rs >= 1803) {
                                                                return 2;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (refDCval >= 75) {
                                                    if (mu_mv_mag_sq < 115) {
                                                        return 4;
                                                    }
                                                    else if (mu_mv_mag_sq >= 115) {
                                                        if (Rs < 1844) {
                                                            return 8;
                                                        }
                                                        else if (Rs >= 1844) {
                                                            if (Cs < 1644) {
                                                                if (AFD < 6) {
                                                                    return 8;
                                                                }
                                                                else if (AFD >= 6) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (Cs >= 1644) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (tcor >= 96) {
                                            if (Cs < 1272) {
                                                if (posBalance < 54) {
                                                    if (Rs < 1852) {
                                                        return 4;
                                                    }
                                                    else if (Rs >= 1852) {
                                                        return 8;
                                                    }
                                                }
                                                else if (posBalance >= 54) {
                                                    return 8;
                                                }
                                            }
                                            else if (Cs >= 1272) {
                                                if (SC < 3448) {
                                                    if (MVDiff < 1) {
                                                        if (SC < 3294) {
                                                            return 4;
                                                        }
                                                        else if (SC >= 3294) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (MVDiff >= 1) {
                                                        if (negBalance < 36) {
                                                            return 8;
                                                        }
                                                        else if (negBalance >= 36) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                                else if (SC >= 3448) {
                                                    if (posBalance < 30) {
                                                        return 8;
                                                    }
                                                    else if (posBalance >= 30) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (Rs >= 2558) {
                                        if (RsCsDiff < 92) {
                                            if (ssDCval < 79) {
                                                if (diffTSC < -2) {
                                                    return 4;
                                                }
                                                else if (diffTSC >= -2) {
                                                    if (Cs < 1803) {
                                                        return 1;
                                                    }
                                                    else if (Cs >= 1803) {
                                                        if (ssDCval < 54) {
                                                            if (Rs < 3514) {
                                                                return 8;
                                                            }
                                                            else if (Rs >= 3514) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (ssDCval >= 54) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (ssDCval >= 79) {
                                                return 8;
                                            }
                                        }
                                        else if (RsCsDiff >= 92) {
                                            if (Cs < 4874) {
                                                if (refDCval < 80) {
                                                    return 8;
                                                }
                                                else if (refDCval >= 80) {
                                                    if (mu_mv_mag_sq < 128) {
                                                        return 8;
                                                    }
                                                    else if (mu_mv_mag_sq >= 128) {
                                                        if (Cs < 4799) {
                                                            return 4;
                                                        }
                                                        else if (Cs >= 4799) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (Cs >= 4874) {
                                                if (Rs < 5661) {
                                                    if (Cs < 5060) {
                                                        if (Cs < 4969) {
                                                            return 4;
                                                        }
                                                        else if (Cs >= 4969) {
                                                            if (refDCval < 80) {
                                                                if (RsDiff < 9) {
                                                                    return 8;
                                                                }
                                                                else if (RsDiff >= 9) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (refDCval >= 80) {
                                                                if (RsCsDiff < 137) {
                                                                    if (posBalance < 57) {
                                                                        return 8;
                                                                    }
                                                                    else if (posBalance >= 57) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (RsCsDiff >= 137) {
                                                                    if (tcor < 92) {
                                                                        return 8;
                                                                    }
                                                                    else if (tcor >= 92) {
                                                                        return 4;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (Cs >= 5060) {
                                                        return 4;
                                                    }
                                                }
                                                else if (Rs >= 5661) {
                                                    return 8;
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                        }
                        else if (ssDCval >= 88) {
                            if (RsDiff < 3) {
                                if (ssDCval < 130) {
                                    if (Rs < 2042) {
                                        return 4;
                                    }
                                    else if (Rs >= 2042) {
                                        if (Cs < 2228) {
                                            return 8;
                                        }
                                        else if (Cs >= 2228) {
                                            return 4;
                                        }
                                    }
                                }
                                else if (ssDCval >= 130) {
                                    if (ssDCval < 141) {
                                        if (diffMVdiffVal < -2) {
                                            return 1;
                                        }
                                        else if (diffMVdiffVal >= -2) {
                                            if (MVDiff < 6) {
                                                return 8;
                                            }
                                            else if (MVDiff >= 6) {
                                                if (refDCval < 140) {
                                                    return 8;
                                                }
                                                else if (refDCval >= 140) {
                                                    return 1;
                                                }
                                            }
                                        }
                                    }
                                    else if (ssDCval >= 141) {
                                        if (refDCval < 145) {
                                            return 4;
                                        }
                                        else if (refDCval >= 145) {
                                            return 8;
                                        }
                                    }
                                }
                            }
                            else if (RsDiff >= 3) {
                                if (mu_mv_mag_sq < 123) {
                                    if (MVDiff < 1) {
                                        if (mu_mv_mag_sq < 108) {
                                            return 4;
                                        }
                                        else if (mu_mv_mag_sq >= 108) {
                                            if (refDCval < 115) {
                                                if (refDCval < 113) {
                                                    if (posBalance < 46) {
                                                        return 1;
                                                    }
                                                    else if (posBalance >= 46) {
                                                        if (Cs < 3032) {
                                                            if (RsDiff < 5) {
                                                                if (AFD < 6) {
                                                                    if (Cs < 1997) {
                                                                        return 8;
                                                                    }
                                                                    else if (Cs >= 1997) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (AFD >= 6) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (RsDiff >= 5) {
                                                                if (SC < 4991) {
                                                                    if (Rs < 2063) {
                                                                        return 8;
                                                                    }
                                                                    else if (Rs >= 2063) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (SC >= 4991) {
                                                                    if (SC < 5804) {
                                                                        return 8;
                                                                    }
                                                                    else if (SC >= 5804) {
                                                                        if (Cs < 3009) {
                                                                            return 4;
                                                                        }
                                                                        else if (Cs >= 3009) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (Cs >= 3032) {
                                                            if (Rs < 2899) {
                                                                if (CsDiff < 5) {
                                                                    return 4;
                                                                }
                                                                else if (CsDiff >= 5) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (Rs >= 2899) {
                                                                if (RsCsDiff < 46) {
                                                                    if (SC < 6145) {
                                                                        return 4;
                                                                    }
                                                                    else if (SC >= 6145) {
                                                                        return 8;
                                                                    }
                                                                }
                                                                else if (RsCsDiff >= 46) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (refDCval >= 113) {
                                                    if (Rs < 2665) {
                                                        return 8;
                                                    }
                                                    else if (Rs >= 2665) {
                                                        if (mu_mv_mag_sq < 121) {
                                                            if (posBalance < 59) {
                                                                return 4;
                                                            }
                                                            else if (posBalance >= 59) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (mu_mv_mag_sq >= 121) {
                                                            if (negBalance < 48) {
                                                                if (Cs < 2996) {
                                                                    return 8;
                                                                }
                                                                else if (Cs >= 2996) {
                                                                    if (Rs < 2711) {
                                                                        return 8;
                                                                    }
                                                                    else if (Rs >= 2711) {
                                                                        return 4;
                                                                    }
                                                                }
                                                            }
                                                            else if (negBalance >= 48) {
                                                                if (Rs < 2765) {
                                                                    return 4;
                                                                }
                                                                else if (Rs >= 2765) {
                                                                    if (Rs < 2781) {
                                                                        return 8;
                                                                    }
                                                                    else if (Rs >= 2781) {
                                                                        return 4;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                            else if (refDCval >= 115) {
                                                return 8;
                                            }
                                        }
                                    }
                                    else if (MVDiff >= 1) {
                                        if (diffAFD < -52) {
                                            return 4;
                                        }
                                        else if (diffAFD >= -52) {
                                            if (ssDCval < 95) {
                                                if (diffRsCsdiff < -8) {
                                                    if (TSC < 6) {
                                                        if (mu_mv_mag_sq < 101) {
                                                            return 1;
                                                        }
                                                        else if (mu_mv_mag_sq >= 101) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (TSC >= 6) {
                                                        return 1;
                                                    }
                                                }
                                                else if (diffRsCsdiff >= -8) {
                                                    if (mu_mv_mag_sq < 99) {
                                                        if (tcor < 98) {
                                                            return 8;
                                                        }
                                                        else if (tcor >= 98) {
                                                            return 1;
                                                        }
                                                    }
                                                    else if (mu_mv_mag_sq >= 99) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                            else if (ssDCval >= 95) {
                                                if (posBalance < 18) {
                                                    if (MVDiff < 4) {
                                                        return 8;
                                                    }
                                                    else if (MVDiff >= 4) {
                                                        return 1;
                                                    }
                                                }
                                                else if (posBalance >= 18) {
                                                    if (diffRsCsdiff < -10) {
                                                        if (diffMVdiffVal < -5) {
                                                            return 4;
                                                        }
                                                        else if (diffMVdiffVal >= -5) {
                                                            if (tcor < 91) {
                                                                if (Rs < 2277) {
                                                                    return 8;
                                                                }
                                                                else if (Rs >= 2277) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (tcor >= 91) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                    else if (diffRsCsdiff >= -10) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (mu_mv_mag_sq >= 123) {
                                    if (CsDiff < 8) {
                                        if (MVDiff < 1) {
                                            if (TSCindex < 7) {
                                                return 8;
                                            }
                                            else if (TSCindex >= 7) {
                                                if (negBalance < 48) {
                                                    return 8;
                                                }
                                                else if (negBalance >= 48) {
                                                    if (diffRsCsdiff < -8) {
                                                        if (diffRsCsdiff < -10) {
                                                            return 8;
                                                        }
                                                        else if (diffRsCsdiff >= -10) {
                                                            if (SC < 5757) {
                                                                return 8;
                                                            }
                                                            else if (SC >= 5757) {
                                                                if (ssDCval < 112) {
                                                                    if (Rs < 2988) {
                                                                        return 4;
                                                                    }
                                                                    else if (Rs >= 2988) {
                                                                        return 8;
                                                                    }
                                                                }
                                                                else if (ssDCval >= 112) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (diffRsCsdiff >= -8) {
                                                        if (RsCsDiff < 37) {
                                                            if (SC < 5750) {
                                                                return 8;
                                                            }
                                                            else if (SC >= 5750) {
                                                                if (Rs < 2808) {
                                                                    return 4;
                                                                }
                                                                else if (Rs >= 2808) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                        else if (RsCsDiff >= 37) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (MVDiff >= 1) {
                                            if (negBalance < 29) {
                                                if (Cs < 2897) {
                                                    return 1;
                                                }
                                                else if (Cs >= 2897) {
                                                    return 8;
                                                }
                                            }
                                            else if (negBalance >= 29) {
                                                return 8;
                                            }
                                        }
                                    }
                                    else if (CsDiff >= 8) {
                                        if (ssDCval < 113) {
                                            return 8;
                                        }
                                        else if (ssDCval >= 113) {
                                            return 4;
                                        }
                                    }
                                }
                            }
                        }
                    }
                }
            }
            else if (AFD >= 7) {
                if (CsDiff < 9) {
                    if (SC < 7361) {
                        if (Cs < 1367) {
                            if (mu_mv_mag_sq < 170) {
                                if (negBalance < 41) {
                                    if (MVDiff < 36) {
                                        if (ssDCval < 63) {
                                            return 4;
                                        }
                                        else if (ssDCval >= 63) {
                                            return 8;
                                        }
                                    }
                                    else if (MVDiff >= 36) {
                                        if (Cs < 261) {
                                            return 1;
                                        }
                                        else if (Cs >= 261) {
                                            return 4;
                                        }
                                    }
                                }
                                else if (negBalance >= 41) {
                                    if (RsCsDiff < 31) {
                                        return 2;
                                    }
                                    else if (RsCsDiff >= 31) {
                                        if (SC < 2323) {
                                            if (Cs < 1108) {
                                                return 8;
                                            }
                                            else if (Cs >= 1108) {
                                                return 4;
                                            }
                                        }
                                        else if (SC >= 2323) {
                                            return 2;
                                        }
                                    }
                                }
                            }
                            else if (mu_mv_mag_sq >= 170) {
                                if (RsDiff < 12) {
                                    if (SC < 731) {
                                        if (ssDCval < 26) {
                                            return 1;
                                        }
                                        else if (ssDCval >= 26) {
                                            return 2;
                                        }
                                    }
                                    else if (SC >= 731) {
                                        return 1;
                                    }
                                }
                                else if (RsDiff >= 12) {
                                    if (AFD < 31) {
                                        return 1;
                                    }
                                    else if (AFD >= 31) {
                                        return 2;
                                    }
                                }
                            }
                        }
                        else if (Cs >= 1367) {
                            if (ssDCval < 100) {
                                if (mcTcor < 94) {
                                    if (ssDCval < 76) {
                                        if (negBalance < 45) {
                                            if (AFD < 13) {
                                                if (TSC < 6) {
                                                    return 1;
                                                }
                                                else if (TSC >= 6) {
                                                    return 8;
                                                }
                                            }
                                            else if (AFD >= 13) {
                                                return 4;
                                            }
                                        }
                                        else if (negBalance >= 45) {
                                            if (mu_mv_mag_sq < 113) {
                                                return 2;
                                            }
                                            else if (mu_mv_mag_sq >= 113) {
                                                if (refDCval < 75) {
                                                    if (SC < 3266) {
                                                        return 2;
                                                    }
                                                    else if (SC >= 3266) {
                                                        if (ssDCval < 75) {
                                                            return 8;
                                                        }
                                                        else if (ssDCval >= 75) {
                                                            return 2;
                                                        }
                                                    }
                                                }
                                                else if (refDCval >= 75) {
                                                    return 4;
                                                }
                                            }
                                        }
                                    }
                                    else if (ssDCval >= 76) {
                                        if (refDCval < 93) {
                                            if (RsCsDiff < 106) {
                                                if (Rs < 3296) {
                                                    if (mcTcor < 91) {
                                                        return 8;
                                                    }
                                                    else if (mcTcor >= 91) {
                                                        if (SC < 3332) {
                                                            return 2;
                                                        }
                                                        else if (SC >= 3332) {
                                                            if (negBalance < 53) {
                                                                return 8;
                                                            }
                                                            else if (negBalance >= 53) {
                                                                return 2;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (Rs >= 3296) {
                                                    return 2;
                                                }
                                            }
                                            else if (RsCsDiff >= 106) {
                                                if (Rs < 3199) {
                                                    if (Cs < 2283) {
                                                        return 8;
                                                    }
                                                    else if (Cs >= 2283) {
                                                        return 4;
                                                    }
                                                }
                                                else if (Rs >= 3199) {
                                                    return 2;
                                                }
                                            }
                                        }
                                        else if (refDCval >= 93) {
                                            if (SC < 4736) {
                                                if (Rs < 1945) {
                                                    return 8;
                                                }
                                                else if (Rs >= 1945) {
                                                    return 2;
                                                }
                                            }
                                            else if (SC >= 4736) {
                                                return 8;
                                            }
                                        }
                                    }
                                }
                                else if (mcTcor >= 94) {
                                    if (posBalance < 53) {
                                        if (Cs < 1764) {
                                            return 4;
                                        }
                                        else if (Cs >= 1764) {
                                            if (posBalance < 46) {
                                                if (TSC < 7) {
                                                    return 8;
                                                }
                                                else if (TSC >= 7) {
                                                    if (gchDC < 2) {
                                                        return 1;
                                                    }
                                                    else if (gchDC >= 2) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                            else if (posBalance >= 46) {
                                                if (negBalance < 37) {
                                                    return 8;
                                                }
                                                else if (negBalance >= 37) {
                                                    return 4;
                                                }
                                            }
                                        }
                                    }
                                    else if (posBalance >= 53) {
                                        if (Rs < 3253) {
                                            if (Cs < 2167) {
                                                if (mu_mv_mag_sq < 113) {
                                                    if (Rs < 2524) {
                                                        return 2;
                                                    }
                                                    else if (Rs >= 2524) {
                                                        if (Rs < 2720) {
                                                            return 8;
                                                        }
                                                        else if (Rs >= 2720) {
                                                            return 1;
                                                        }
                                                    }
                                                }
                                                else if (mu_mv_mag_sq >= 113) {
                                                    if (ssDCval < 76) {
                                                        return 2;
                                                    }
                                                    else if (ssDCval >= 76) {
                                                        if (SC < 3439) {
                                                            if (Cs < 1584) {
                                                                return 4;
                                                            }
                                                            else if (Cs >= 1584) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (SC >= 3439) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (Cs >= 2167) {
                                                if (diffAFD < -1) {
                                                    if (Rs < 2582) {
                                                        return 2;
                                                    }
                                                    else if (Rs >= 2582) {
                                                        if (Rs < 3103) {
                                                            return 4;
                                                        }
                                                        else if (Rs >= 3103) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                                else if (diffAFD >= -1) {
                                                    if (ssDCval < 62) {
                                                        return 8;
                                                    }
                                                    else if (ssDCval >= 62) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                        }
                                        else if (Rs >= 3253) {
                                            if (ssDCval < 94) {
                                                return 2;
                                            }
                                            else if (ssDCval >= 94) {
                                                return 4;
                                            }
                                        }
                                    }
                                }
                            }
                            else if (ssDCval >= 100) {
                                if (Rs < 4056) {
                                    if (SC < 4566) {
                                        if (Cs < 1888) {
                                            if (diffMVdiffVal < -1) {
                                                if (diffRsCsdiff < 23) {
                                                    if (Cs < 1713) {
                                                        return 8;
                                                    }
                                                    else if (Cs >= 1713) {
                                                        return 4;
                                                    }
                                                }
                                                else if (diffRsCsdiff >= 23) {
                                                    return 4;
                                                }
                                            }
                                            else if (diffMVdiffVal >= -1) {
                                                return 8;
                                            }
                                        }
                                        else if (Cs >= 1888) {
                                            if (refDCval < 116) {
                                                return 4;
                                            }
                                            else if (refDCval >= 116) {
                                                if (mcTcor < 92) {
                                                    return 8;
                                                }
                                                else if (mcTcor >= 92) {
                                                    return 4;
                                                }
                                            }
                                        }
                                    }
                                    else if (SC >= 4566) {
                                        if (Rs < 3764) {
                                            if (RsDiff < 10) {
                                                if (Rs < 2761) {
                                                    if (Rs < 2749) {
                                                        if (Cs < 2347) {
                                                            if (Rs < 2371) {
                                                                return 4;
                                                            }
                                                            else if (Rs >= 2371) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (Cs >= 2347) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (Rs >= 2749) {
                                                        return 4;
                                                    }
                                                }
                                                else if (Rs >= 2761) {
                                                    return 8;
                                                }
                                            }
                                            else if (RsDiff >= 10) {
                                                if (SC < 5617) {
                                                    return 8;
                                                }
                                                else if (SC >= 5617) {
                                                    return 4;
                                                }
                                            }
                                        }
                                        else if (Rs >= 3764) {
                                            if (Rs < 3773) {
                                                return 4;
                                            }
                                            else if (Rs >= 3773) {
                                                if (RsDiff < 8) {
                                                    return 8;
                                                }
                                                else if (RsDiff >= 8) {
                                                    return 1;
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (Rs >= 4056) {
                                    if (ssDCval < 127) {
                                        if (SC < 6677) {
                                            return 8;
                                        }
                                        else if (SC >= 6677) {
                                            return 4;
                                        }
                                    }
                                    else if (ssDCval >= 127) {
                                        if (Rs < 4392) {
                                            return 8;
                                        }
                                        else if (Rs >= 4392) {
                                            return 2;
                                        }
                                    }
                                }
                            }
                        }
                    }
                    else if (SC >= 7361) {
                        if (AFD < 9) {
                            if (CsDiff < 7) {
                                if (MVDiff < 2) {
                                    return 4;
                                }
                                else if (MVDiff >= 2) {
                                    if (Cs < 3112) {
                                        return 2;
                                    }
                                    else if (Cs >= 3112) {
                                        if (posBalance < 54) {
                                            return 8;
                                        }
                                        else if (posBalance >= 54) {
                                            return 4;
                                        }
                                    }
                                }
                            }
                            else if (CsDiff >= 7) {
                                if (ssDCval < 81) {
                                    if (SC < 9919) {
                                        return 8;
                                    }
                                    else if (SC >= 9919) {
                                        if (Rs < 5084) {
                                            return 4;
                                        }
                                        else if (Rs >= 5084) {
                                            return 8;
                                        }
                                    }
                                }
                                else if (ssDCval >= 81) {
                                    return 8;
                                }
                            }
                        }
                        else if (AFD >= 9) {
                            if (mu_mv_mag_sq < 111) {
                                if (tcor < 89) {
                                    if (SC < 8345) {
                                        return 8;
                                    }
                                    else if (SC >= 8345) {
                                        return 2;
                                    }
                                }
                                else if (tcor >= 89) {
                                    return 2;
                                }
                            }
                            else if (mu_mv_mag_sq >= 111) {
                                return 4;
                            }
                        }
                    }
                }
                else if (CsDiff >= 9) {
                    if (Cs < 2219) {
                        if (RsDiff < 16) {
                            if (mcTcor < 46) {
                                if (SC < 2790) {
                                    return 8;
                                }
                                else if (SC >= 2790) {
                                    return 4;
                                }
                            }
                            else if (mcTcor >= 46) {
                                if (tcor < 64) {
                                    return 1;
                                }
                                else if (tcor >= 64) {
                                    if (Rs < 1841) {
                                        return 2;
                                    }
                                    else if (Rs >= 1841) {
                                        return 8;
                                    }
                                }
                            }
                        }
                        else if (RsDiff >= 16) {
                            if (ssDCval < 108) {
                                return 2;
                            }
                            else if (ssDCval >= 108) {
                                return 4;
                            }
                        }
                    }
                    else if (Cs >= 2219) {
                        if (Rs < 7303) {
                            if (TSC < 13) {
                                if (SC < 7816) {
                                    if (SC < 6665) {
                                        if (ssDCval < 95) {
                                            return 2;
                                        }
                                        else if (ssDCval >= 95) {
                                            return 8;
                                        }
                                    }
                                    else if (SC >= 6665) {
                                        if (RsDiff < 9) {
                                            if (Rs < 3624) {
                                                return 8;
                                            }
                                            else if (Rs >= 3624) {
                                                return 2;
                                            }
                                        }
                                        else if (RsDiff >= 9) {
                                            if (ssDCval < 125) {
                                                if (Cs < 3211) {
                                                    return 4;
                                                }
                                                else if (Cs >= 3211) {
                                                    if (RsCsDiff < 201) {
                                                        return 8;
                                                    }
                                                    else if (RsCsDiff >= 201) {
                                                        if (AFD < 14) {
                                                            return 4;
                                                        }
                                                        else if (AFD >= 14) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (ssDCval >= 125) {
                                                if (posBalance < 57) {
                                                    return 1;
                                                }
                                                else if (posBalance >= 57) {
                                                    return 8;
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (SC >= 7816) {
                                    if (MVDiff < 1) {
                                        if (ssDCval < 106) {
                                            if (refDCval < 84) {
                                                if (SC < 10115) {
                                                    if (negBalance < 47) {
                                                        return 8;
                                                    }
                                                    else if (negBalance >= 47) {
                                                        if (Cs < 4876) {
                                                            return 8;
                                                        }
                                                        else if (Cs >= 4876) {
                                                            if (Cs < 5042) {
                                                                return 4;
                                                            }
                                                            else if (Cs >= 5042) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (SC >= 10115) {
                                                    return 4;
                                                }
                                            }
                                            else if (refDCval >= 84) {
                                                return 8;
                                            }
                                        }
                                        else if (ssDCval >= 106) {
                                            if (Rs < 5412) {
                                                if (ssDCval < 119) {
                                                    return 4;
                                                }
                                                else if (ssDCval >= 119) {
                                                    return 8;
                                                }
                                            }
                                            else if (Rs >= 5412) {
                                                if (SC < 12083) {
                                                    if (posBalance < 54) {
                                                        return 8;
                                                    }
                                                    else if (posBalance >= 54) {
                                                        if (CsDiff < 14) {
                                                            return 4;
                                                        }
                                                        else if (CsDiff >= 14) {
                                                            if (Rs < 5702) {
                                                                return 4;
                                                            }
                                                            else if (Rs >= 5702) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (SC >= 12083) {
                                                    return 8;
                                                }
                                            }
                                        }
                                    }
                                    else if (MVDiff >= 1) {
                                        if (Cs < 3918) {
                                            if (negBalance < 45) {
                                                if (Rs < 4686) {
                                                    return 4;
                                                }
                                                else if (Rs >= 4686) {
                                                    return 2;
                                                }
                                            }
                                            else if (negBalance >= 45) {
                                                return 8;
                                            }
                                        }
                                        else if (Cs >= 3918) {
                                            return 8;
                                        }
                                    }
                                }
                            }
                            else if (TSC >= 13) {
                                if (refDCval < 136) {
                                    return 8;
                                }
                                else if (refDCval >= 136) {
                                    if (Rs < 5153) {
                                        return 1;
                                    }
                                    else if (Rs >= 5153) {
                                        return 8;
                                    }
                                }
                            }
                        }
                        else if (Rs >= 7303) {
                            if (Rs < 9235) {
                                if (Cs < 7810) {
                                    return 4;
                                }
                                else if (Cs >= 7810) {
                                    if (Cs < 8857) {
                                        if (RsCsDiff < 1036) {
                                            return 2;
                                        }
                                        else if (RsCsDiff >= 1036) {
                                            if (Cs < 8185) {
                                                return 4;
                                            }
                                            else if (Cs >= 8185) {
                                                if (Cs < 8345) {
                                                    return 2;
                                                }
                                                else if (Cs >= 8345) {
                                                    return 4;
                                                }
                                            }
                                        }
                                    }
                                    else if (Cs >= 8857) {
                                        return 8;
                                    }
                                }
                            }
                            else if (Rs >= 9235) {
                                return 8;
                            }
                        }
                    }
                }
            }
        }
        return 4;
    }
    mfxU16 AGOPSelectD(mfxU32 Rs, mfxU32 CsDiff, mfxU32 TSC, mfxI32 diffAFD, mfxU32 Scindex, mfxU32 RsDiff, mfxI32 diffMVdiffVal, mfxU32 RsCsDiff, mfxU32 TSCindex, mfxI16 tcor, mfxU32 refDCval, mfxU32 MVDiff, mfxU32 AFD, mfxI32 diffTSC, mfxI32 diffRsCsdiff, mfxU32 posBalance, mfxU32 SC, mfxU32 mu_mv_mag_sq, mfxU32 negBalance, mfxI16 mcTcor, mfxU32 gchDC, mfxU32 Cs, mfxU32 ssDCval) {
        if (mu_mv_mag_sq < 84) {
            if (posBalance < 5) {
                if (SC < 3696) {
                    if (MVDiff < 13) {
                        if (Cs < 4) {
                            return 4;
                        }
                        else if (Cs >= 4) {
                            return 8;
                        }
                    }
                    else if (MVDiff >= 13) {
                        if (diffAFD < -1) {
                            if (MVDiff < 141) {
                                if (negBalance < 50) {
                                    return 4;
                                }
                                else if (negBalance >= 50) {
                                    return 2;
                                }
                            }
                            else if (MVDiff >= 141) {
                                if (Rs < 395) {
                                    return 1;
                                }
                                else if (Rs >= 395) {
                                    return 8;
                                }
                            }
                        }
                        else if (diffAFD >= -1) {
                            if (mcTcor < 88) {
                                if (diffTSC < 1) {
                                    if (mu_mv_mag_sq < 7) {
                                        if (MVDiff < 283) {
                                            return 2;
                                        }
                                        else if (MVDiff >= 283) {
                                            return 8;
                                        }
                                    }
                                    else if (mu_mv_mag_sq >= 7) {
                                        if (posBalance < 4) {
                                            if (negBalance < 10) {
                                                if (diffMVdiffVal < 13) {
                                                    return 8;
                                                }
                                                else if (diffMVdiffVal >= 13) {
                                                    return 4;
                                                }
                                            }
                                            else if (negBalance >= 10) {
                                                return 4;
                                            }
                                        }
                                        else if (posBalance >= 4) {
                                            return 2;
                                        }
                                    }
                                }
                                else if (diffTSC >= 1) {
                                    if (mu_mv_mag_sq < 37) {
                                        return 1;
                                    }
                                    else if (mu_mv_mag_sq >= 37) {
                                        if (Rs < 694) {
                                            return 2;
                                        }
                                        else if (Rs >= 694) {
                                            return 4;
                                        }
                                    }
                                }
                            }
                            else if (mcTcor >= 88) {
                                if (tcor < 141) {
                                    if (MVDiff < 388) {
                                        return 8;
                                    }
                                    else if (MVDiff >= 388) {
                                        return 4;
                                    }
                                }
                                else if (tcor >= 141) {
                                    if (diffTSC < -1) {
                                        return 4;
                                    }
                                    else if (diffTSC >= -1) {
                                        if (ssDCval < 23) {
                                            return 4;
                                        }
                                        else if (ssDCval >= 23) {
                                            return 8;
                                        }
                                    }
                                }
                            }
                        }
                    }
                }
                else if (SC >= 3696) {
                    if (SC < 9327) {
                        if (diffMVdiffVal < 44) {
                            if (AFD < 1) {
                                if (Rs < 2662) {
                                    return 1;
                                }
                                else if (Rs >= 2662) {
                                    if (SC < 7709) {
                                        return 8;
                                    }
                                    else if (SC >= 7709) {
                                        if (Rs < 3791) {
                                            if (Cs < 3985) {
                                                return 1;
                                            }
                                            else if (Cs >= 3985) {
                                                return 8;
                                            }
                                        }
                                        else if (Rs >= 3791) {
                                            return 8;
                                        }
                                    }
                                }
                            }
                            else if (AFD >= 1) {
                                return 8;
                            }
                        }
                        else if (diffMVdiffVal >= 44) {
                            return 2;
                        }
                    }
                    else if (SC >= 9327) {
                        if (Rs < 5376) {
                            return 1;
                        }
                        else if (Rs >= 5376) {
                            return 8;
                        }
                    }
                }
            }
            else if (posBalance >= 5) {
                if (CsDiff < 17) {
                    if (mu_mv_mag_sq < 38) {
                        if (Cs < 1489) {
                            if (AFD < 13) {
                                if (SC < 899) {
                                    if (TSCindex < 6) {
                                        if (mcTcor < 79) {
                                            if (negBalance < 9) {
                                                if (mu_mv_mag_sq < 8) {
                                                    if (MVDiff < 45) {
                                                        return 2;
                                                    }
                                                    else if (MVDiff >= 45) {
                                                        if (MVDiff < 2579) {
                                                            if (diffMVdiffVal < 23) {
                                                                return 1;
                                                            }
                                                            else if (diffMVdiffVal >= 23) {
                                                                if (diffTSC < 1) {
                                                                    return 1;
                                                                }
                                                                else if (diffTSC >= 1) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                        else if (MVDiff >= 2579) {
                                                            return 2;
                                                        }
                                                    }
                                                }
                                                else if (mu_mv_mag_sq >= 8) {
                                                    if (mcTcor < 47) {
                                                        if (gchDC < 1) {
                                                            if (diffMVdiffVal < 3) {
                                                                if (negBalance < 2) {
                                                                    return 1;
                                                                }
                                                                else if (negBalance >= 2) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (diffMVdiffVal >= 3) {
                                                                if (diffMVdiffVal < 247) {
                                                                    if (negBalance < 1) {
                                                                        return 2;
                                                                    }
                                                                    else if (negBalance >= 1) {
                                                                        if (diffAFD < -5) {
                                                                            return 1;
                                                                        }
                                                                        else if (diffAFD >= -5) {
                                                                            if (posBalance < 27) {
                                                                                return 1;
                                                                            }
                                                                            else if (posBalance >= 27) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                                else if (diffMVdiffVal >= 247) {
                                                                    return 2;
                                                                }
                                                            }
                                                        }
                                                        else if (gchDC >= 1) {
                                                            if (diffMVdiffVal < -160) {
                                                                return 8;
                                                            }
                                                            else if (diffMVdiffVal >= -160) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                    else if (mcTcor >= 47) {
                                                        if (AFD < 2) {
                                                            if (ssDCval < 20) {
                                                                if (mcTcor < 57) {
                                                                    if (mcTcor < 51) {
                                                                        if (diffMVdiffVal < 7) {
                                                                            return 8;
                                                                        }
                                                                        else if (diffMVdiffVal >= 7) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                    else if (mcTcor >= 51) {
                                                                        if (posBalance < 88) {
                                                                            return 4;
                                                                        }
                                                                        else if (posBalance >= 88) {
                                                                            if (MVDiff < 210) {
                                                                                return 8;
                                                                            }
                                                                            else if (MVDiff >= 210) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                                else if (mcTcor >= 57) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (ssDCval >= 20) {
                                                                if (posBalance < 85) {
                                                                    if (ssDCval < 23) {
                                                                        if (mu_mv_mag_sq < 11) {
                                                                            return 4;
                                                                        }
                                                                        else if (mu_mv_mag_sq >= 11) {
                                                                            return 2;
                                                                        }
                                                                    }
                                                                    else if (ssDCval >= 23) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (posBalance >= 85) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                        else if (AFD >= 2) {
                                                            if (gchDC < 5) {
                                                                if (mu_mv_mag_sq < 27) {
                                                                    return 1;
                                                                }
                                                                else if (mu_mv_mag_sq >= 27) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (gchDC >= 5) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                            else if (negBalance >= 9) {
                                                if (gchDC < 3) {
                                                    if (posBalance < 60) {
                                                        if (Rs < 406) {
                                                            if (mu_mv_mag_sq < 37) {
                                                                return 8;
                                                            }
                                                            else if (mu_mv_mag_sq >= 37) {
                                                                if (Rs < 352) {
                                                                    return 4;
                                                                }
                                                                else if (Rs >= 352) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                        else if (Rs >= 406) {
                                                            if (mu_mv_mag_sq < 34) {
                                                                return 8;
                                                            }
                                                            else if (mu_mv_mag_sq >= 34) {
                                                                return 1;
                                                            }
                                                        }
                                                    }
                                                    else if (posBalance >= 60) {
                                                        if (AFD < 6) {
                                                            return 2;
                                                        }
                                                        else if (AFD >= 6) {
                                                            return 1;
                                                        }
                                                    }
                                                }
                                                else if (gchDC >= 3) {
                                                    return 1;
                                                }
                                            }
                                        }
                                        else if (mcTcor >= 79) {
                                            if (ssDCval < 105) {
                                                if (posBalance < 79) {
                                                    if (refDCval < 90) {
                                                        if (SC < 450) {
                                                            if (mcTcor < 100) {
                                                                if (diffAFD < -2) {
                                                                    if (AFD < 5) {
                                                                        return 2;
                                                                    }
                                                                    else if (AFD >= 5) {
                                                                        return 1;
                                                                    }
                                                                }
                                                                else if (diffAFD >= -2) {
                                                                    if (Cs < 245) {
                                                                        if (AFD < 10) {
                                                                            return 2;
                                                                        }
                                                                        else if (AFD >= 10) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                    else if (Cs >= 245) {
                                                                        return 1;
                                                                    }
                                                                }
                                                            }
                                                            else if (mcTcor >= 100) {
                                                                if (RsCsDiff < 1) {
                                                                    return 4;
                                                                }
                                                                else if (RsCsDiff >= 1) {
                                                                    return 1;
                                                                }
                                                            }
                                                        }
                                                        else if (SC >= 450) {
                                                            if (AFD < 6) {
                                                                if (AFD < 5) {
                                                                    if (CsDiff < 1) {
                                                                        return 2;
                                                                    }
                                                                    else if (CsDiff >= 1) {
                                                                        if (SC < 513) {
                                                                            return 2;
                                                                        }
                                                                        else if (SC >= 513) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                }
                                                                else if (AFD >= 5) {
                                                                    if (MVDiff < 27) {
                                                                        return 2;
                                                                    }
                                                                    else if (MVDiff >= 27) {
                                                                        return 4;
                                                                    }
                                                                }
                                                            }
                                                            else if (AFD >= 6) {
                                                                if (Cs < 218) {
                                                                    return 1;
                                                                }
                                                                else if (Cs >= 218) {
                                                                    if (diffMVdiffVal < -6) {
                                                                        return 2;
                                                                    }
                                                                    else if (diffMVdiffVal >= -6) {
                                                                        return 4;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (refDCval >= 90) {
                                                        return 2;
                                                    }
                                                }
                                                else if (posBalance >= 79) {
                                                    if (MVDiff < 101) {
                                                        if (mcTcor < 84) {
                                                            if (posBalance < 98) {
                                                                return 8;
                                                            }
                                                            else if (posBalance >= 98) {
                                                                if (refDCval < 23) {
                                                                    return 4;
                                                                }
                                                                else if (refDCval >= 23) {
                                                                    return 2;
                                                                }
                                                            }
                                                        }
                                                        else if (mcTcor >= 84) {
                                                            if (Rs < 273) {
                                                                return 4;
                                                            }
                                                            else if (Rs >= 273) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                    else if (MVDiff >= 101) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                            else if (ssDCval >= 105) {
                                                if (RsDiff < 2) {
                                                    if (Rs < 832) {
                                                        return 1;
                                                    }
                                                    else if (Rs >= 832) {
                                                        return 8;
                                                    }
                                                }
                                                else if (RsDiff >= 2) {
                                                    return 4;
                                                }
                                            }
                                        }
                                    }
                                    else if (TSCindex >= 6) {
                                        if (MVDiff < 17) {
                                            if (Cs < 116) {
                                                if (ssDCval < 106) {
                                                    return 1;
                                                }
                                                else if (ssDCval >= 106) {
                                                    if (Rs < 397) {
                                                        return 4;
                                                    }
                                                    else if (Rs >= 397) {
                                                        return 2;
                                                    }
                                                }
                                            }
                                            else if (Cs >= 116) {
                                                if (negBalance < 47) {
                                                    return 2;
                                                }
                                                else if (negBalance >= 47) {
                                                    if (Cs < 244) {
                                                        if (Rs < 426) {
                                                            return 4;
                                                        }
                                                        else if (Rs >= 426) {
                                                            return 2;
                                                        }
                                                    }
                                                    else if (Cs >= 244) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                        }
                                        else if (MVDiff >= 17) {
                                            if (Cs < 272) {
                                                if (refDCval < 107) {
                                                    if (diffMVdiffVal < 19) {
                                                        if (refDCval < 74) {
                                                            if (tcor < 79) {
                                                                return 1;
                                                            }
                                                            else if (tcor >= 79) {
                                                                if (Cs < 49) {
                                                                    return 1;
                                                                }
                                                                else if (Cs >= 49) {
                                                                    return 2;
                                                                }
                                                            }
                                                        }
                                                        else if (refDCval >= 74) {
                                                            return 1;
                                                        }
                                                    }
                                                    else if (diffMVdiffVal >= 19) {
                                                        if (Cs < 196) {
                                                            return 2;
                                                        }
                                                        else if (Cs >= 196) {
                                                            return 1;
                                                        }
                                                    }
                                                }
                                                else if (refDCval >= 107) {
                                                    if (Rs < 593) {
                                                        if (Rs < 337) {
                                                            return 4;
                                                        }
                                                        else if (Rs >= 337) {
                                                            return 1;
                                                        }
                                                    }
                                                    else if (Rs >= 593) {
                                                        return 2;
                                                    }
                                                }
                                            }
                                            else if (Cs >= 272) {
                                                if (posBalance < 62) {
                                                    if (ssDCval < 99) {
                                                        if (AFD < 6) {
                                                            if (Rs < 446) {
                                                                return 4;
                                                            }
                                                            else if (Rs >= 446) {
                                                                return 2;
                                                            }
                                                        }
                                                        else if (AFD >= 6) {
                                                            return 2;
                                                        }
                                                    }
                                                    else if (ssDCval >= 99) {
                                                        return 4;
                                                    }
                                                }
                                                else if (posBalance >= 62) {
                                                    if (SC < 716) {
                                                        return 1;
                                                    }
                                                    else if (SC >= 716) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (SC >= 899) {
                                    if (RsCsDiff < 12) {
                                        if (Rs < 654) {
                                            if (tcor < 80) {
                                                if (CsDiff < 2) {
                                                    return 8;
                                                }
                                                else if (CsDiff >= 2) {
                                                    if (refDCval < 93) {
                                                        if (Cs < 457) {
                                                            return 2;
                                                        }
                                                        else if (Cs >= 457) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (refDCval >= 93) {
                                                        if (Rs < 589) {
                                                            return 8;
                                                        }
                                                        else if (Rs >= 589) {
                                                            if (MVDiff < 10) {
                                                                return 4;
                                                            }
                                                            else if (MVDiff >= 10) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                            else if (tcor >= 80) {
                                                if (diffRsCsdiff < 1) {
                                                    if (MVDiff < 9) {
                                                        if (SC < 1247) {
                                                            return 4;
                                                        }
                                                        else if (SC >= 1247) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (MVDiff >= 9) {
                                                        return 4;
                                                    }
                                                }
                                                else if (diffRsCsdiff >= 1) {
                                                    if (MVDiff < 15) {
                                                        if (Cs < 466) {
                                                            return 4;
                                                        }
                                                        else if (Cs >= 466) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (MVDiff >= 15) {
                                                        return 2;
                                                    }
                                                }
                                            }
                                        }
                                        else if (Rs >= 654) {
                                            if (MVDiff < 16) {
                                                if (mcTcor < 90) {
                                                    return 4;
                                                }
                                                else if (mcTcor >= 90) {
                                                    if (negBalance < 26) {
                                                        return 8;
                                                    }
                                                    else if (negBalance >= 26) {
                                                        if (posBalance < 53) {
                                                            if (Cs < 706) {
                                                                return 4;
                                                            }
                                                            else if (Cs >= 706) {
                                                                if (gchDC < 1) {
                                                                    return 1;
                                                                }
                                                                else if (gchDC >= 1) {
                                                                    if (SC < 1928) {
                                                                        return 4;
                                                                    }
                                                                    else if (SC >= 1928) {
                                                                        return 8;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (posBalance >= 53) {
                                                            if (mcTcor < 92) {
                                                                if (MVDiff < 10) {
                                                                    return 4;
                                                                }
                                                                else if (MVDiff >= 10) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (mcTcor >= 92) {
                                                                if (SC < 1191) {
                                                                    if (SC < 1068) {
                                                                        return 1;
                                                                    }
                                                                    else if (SC >= 1068) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (SC >= 1191) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                            else if (MVDiff >= 16) {
                                                if (tcor < 75) {
                                                    if (ssDCval < 46) {
                                                        if (mu_mv_mag_sq < 33) {
                                                            if (TSC < 6) {
                                                                return 4;
                                                            }
                                                            else if (TSC >= 6) {
                                                                if (SC < 1224) {
                                                                    return 4;
                                                                }
                                                                else if (SC >= 1224) {
                                                                    if (Cs < 554) {
                                                                        return 1;
                                                                    }
                                                                    else if (Cs >= 554) {
                                                                        if (Cs < 564) {
                                                                            return 4;
                                                                        }
                                                                        else if (Cs >= 564) {
                                                                            if (Cs < 581) {
                                                                                return 1;
                                                                            }
                                                                            else if (Cs >= 581) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (mu_mv_mag_sq >= 33) {
                                                            if (diffMVdiffVal < -7) {
                                                                return 4;
                                                            }
                                                            else if (diffMVdiffVal >= -7) {
                                                                return 1;
                                                            }
                                                        }
                                                    }
                                                    else if (ssDCval >= 46) {
                                                        if (refDCval < 66) {
                                                            return 8;
                                                        }
                                                        else if (refDCval >= 66) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                                else if (tcor >= 75) {
                                                    if (SC < 1579) {
                                                        if (SC < 1445) {
                                                            return 1;
                                                        }
                                                        else if (SC >= 1445) {
                                                            if (Cs < 611) {
                                                                return 4;
                                                            }
                                                            else if (Cs >= 611) {
                                                                return 1;
                                                            }
                                                        }
                                                    }
                                                    else if (SC >= 1579) {
                                                        if (Rs < 1697) {
                                                            return 8;
                                                        }
                                                        else if (Rs >= 1697) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (RsCsDiff >= 12) {
                                        if (mu_mv_mag_sq < 3) {
                                            if (SC < 3252) {
                                                if (negBalance < 49) {
                                                    if (TSCindex < 5) {
                                                        if (ssDCval < 91) {
                                                            return 2;
                                                        }
                                                        else if (ssDCval >= 91) {
                                                            if (AFD < 8) {
                                                                return 8;
                                                            }
                                                            else if (AFD >= 8) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                    else if (TSCindex >= 5) {
                                                        if (posBalance < 45) {
                                                            if (gchDC < 1) {
                                                                if (Rs < 1812) {
                                                                    return 4;
                                                                }
                                                                else if (Rs >= 1812) {
                                                                    return 2;
                                                                }
                                                            }
                                                            else if (gchDC >= 1) {
                                                                return 2;
                                                            }
                                                        }
                                                        else if (posBalance >= 45) {
                                                            if (SC < 1685) {
                                                                if (RsCsDiff < 15) {
                                                                    if (Cs < 705) {
                                                                        return 4;
                                                                    }
                                                                    else if (Cs >= 705) {
                                                                        return 2;
                                                                    }
                                                                }
                                                                else if (RsCsDiff >= 15) {
                                                                    if (diffMVdiffVal < -2) {
                                                                        if (Cs < 674) {
                                                                            return 1;
                                                                        }
                                                                        else if (Cs >= 674) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                    else if (diffMVdiffVal >= -2) {
                                                                        if (Cs < 468) {
                                                                            return 1;
                                                                        }
                                                                        else if (Cs >= 468) {
                                                                            return 2;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                            else if (SC >= 1685) {
                                                                if (Rs < 1841) {
                                                                    if (negBalance < 28) {
                                                                        return 4;
                                                                    }
                                                                    else if (negBalance >= 28) {
                                                                        if (ssDCval < 137) {
                                                                            if (SC < 2432) {
                                                                                return 1;
                                                                            }
                                                                            else if (SC >= 2432) {
                                                                                if (Cs < 1027) {
                                                                                    return 8;
                                                                                }
                                                                                else if (Cs >= 1027) {
                                                                                    if (posBalance < 47) {
                                                                                        return 4;
                                                                                    }
                                                                                    else if (posBalance >= 47) {
                                                                                        if (diffAFD < -5) {
                                                                                            return 4;
                                                                                        }
                                                                                        else if (diffAFD >= -5) {
                                                                                            return 1;
                                                                                        }
                                                                                    }
                                                                                }
                                                                            }
                                                                        }
                                                                        else if (ssDCval >= 137) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                }
                                                                else if (Rs >= 1841) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (negBalance >= 49) {
                                                    if (tcor < 85) {
                                                        if (Rs < 1168) {
                                                            if (refDCval < 63) {
                                                                if (AFD < 12) {
                                                                    return 8;
                                                                }
                                                                else if (AFD >= 12) {
                                                                    return 2;
                                                                }
                                                            }
                                                            else if (refDCval >= 63) {
                                                                if (SC < 1183) {
                                                                    if (SC < 1171) {
                                                                        return 4;
                                                                    }
                                                                    else if (SC >= 1171) {
                                                                        return 8;
                                                                    }
                                                                }
                                                                else if (SC >= 1183) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                        else if (Rs >= 1168) {
                                                            if (refDCval < 99) {
                                                                if (negBalance < 53) {
                                                                    if (Cs < 867) {
                                                                        return 4;
                                                                    }
                                                                    else if (Cs >= 867) {
                                                                        return 2;
                                                                    }
                                                                }
                                                                else if (negBalance >= 53) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (refDCval >= 99) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                    else if (tcor >= 85) {
                                                        if (RsDiff < 4) {
                                                            if (refDCval < 89) {
                                                                return 2;
                                                            }
                                                            else if (refDCval >= 89) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (RsDiff >= 4) {
                                                            if (TSC < 8) {
                                                                if (Cs < 1078) {
                                                                    if (tcor < 88) {
                                                                        if (Cs < 1011) {
                                                                            return 1;
                                                                        }
                                                                        else if (Cs >= 1011) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                    else if (tcor >= 88) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (Cs >= 1078) {
                                                                    if (Rs < 1376) {
                                                                        return 2;
                                                                    }
                                                                    else if (Rs >= 1376) {
                                                                        return 1;
                                                                    }
                                                                }
                                                            }
                                                            else if (TSC >= 8) {
                                                                if (RsCsDiff < 47) {
                                                                    return 4;
                                                                }
                                                                else if (RsCsDiff >= 47) {
                                                                    return 2;
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                            else if (SC >= 3252) {
                                                return 4;
                                            }
                                        }
                                        else if (mu_mv_mag_sq >= 3) {
                                            if (posBalance < 35) {
                                                if (mcTcor < 93) {
                                                    if (RsDiff < 9) {
                                                        if (Rs < 1326) {
                                                            return 4;
                                                        }
                                                        else if (Rs >= 1326) {
                                                            return 1;
                                                        }
                                                    }
                                                    else if (RsDiff >= 9) {
                                                        return 1;
                                                    }
                                                }
                                                else if (mcTcor >= 93) {
                                                    return 2;
                                                }
                                            }
                                            else if (posBalance >= 35) {
                                                if (mcTcor < 90) {
                                                    if (MVDiff < 14) {
                                                        return 1;
                                                    }
                                                    else if (MVDiff >= 14) {
                                                        if (diffRsCsdiff < 10) {
                                                            if (ssDCval < 112) {
                                                                if (ssDCval < 69) {
                                                                    if (refDCval < 47) {
                                                                        if (CsDiff < 3) {
                                                                            return 1;
                                                                        }
                                                                        else if (CsDiff >= 3) {
                                                                            return 2;
                                                                        }
                                                                    }
                                                                    else if (refDCval >= 47) {
                                                                        return 1;
                                                                    }
                                                                }
                                                                else if (ssDCval >= 69) {
                                                                    return 2;
                                                                }
                                                            }
                                                            else if (ssDCval >= 112) {
                                                                return 1;
                                                            }
                                                        }
                                                        else if (diffRsCsdiff >= 10) {
                                                            if (RsDiff < 4) {
                                                                if (Rs < 707) {
                                                                    return 8;
                                                                }
                                                                else if (Rs >= 707) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (RsDiff >= 4) {
                                                                if (SC < 2419) {
                                                                    if (diffTSC < -1) {
                                                                        return 2;
                                                                    }
                                                                    else if (diffTSC >= -1) {
                                                                        return 1;
                                                                    }
                                                                }
                                                                else if (SC >= 2419) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (mcTcor >= 90) {
                                                    if (posBalance < 65) {
                                                        if (Cs < 735) {
                                                            if (Cs < 254) {
                                                                return 1;
                                                            }
                                                            else if (Cs >= 254) {
                                                                if (diffRsCsdiff < -18) {
                                                                    if (Rs < 1227) {
                                                                        return 1;
                                                                    }
                                                                    else if (Rs >= 1227) {
                                                                        return 2;
                                                                    }
                                                                }
                                                                else if (diffRsCsdiff >= -18) {
                                                                    return 2;
                                                                }
                                                            }
                                                        }
                                                        else if (Cs >= 735) {
                                                            if (MVDiff < 14) {
                                                                if (posBalance < 46) {
                                                                    if (Rs < 2529) {
                                                                        return 8;
                                                                    }
                                                                    else if (Rs >= 2529) {
                                                                        return 2;
                                                                    }
                                                                }
                                                                else if (posBalance >= 46) {
                                                                    if (TSCindex < 9) {
                                                                        if (diffMVdiffVal < -1) {
                                                                            if (Cs < 1096) {
                                                                                return 1;
                                                                            }
                                                                            else if (Cs >= 1096) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                        else if (diffMVdiffVal >= -1) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                    else if (TSCindex >= 9) {
                                                                        if (Cs < 850) {
                                                                            return 2;
                                                                        }
                                                                        else if (Cs >= 850) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                            else if (MVDiff >= 14) {
                                                                if (Cs < 792) {
                                                                    if (Cs < 751) {
                                                                        return 2;
                                                                    }
                                                                    else if (Cs >= 751) {
                                                                        return 1;
                                                                    }
                                                                }
                                                                else if (Cs >= 792) {
                                                                    return 2;
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (posBalance >= 65) {
                                                        if (ssDCval < 66) {
                                                            return 2;
                                                        }
                                                        else if (ssDCval >= 66) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                            else if (AFD >= 13) {
                                if (RsDiff < 8) {
                                    if (mcTcor < 87) {
                                        if (MVDiff < 55) {
                                            if (gchDC < 2) {
                                                if (Rs < 1187) {
                                                    if (gchDC < 1) {
                                                        return 8;
                                                    }
                                                    else if (gchDC >= 1) {
                                                        return 1;
                                                    }
                                                }
                                                else if (Rs >= 1187) {
                                                    return 8;
                                                }
                                            }
                                            else if (gchDC >= 2) {
                                                return 1;
                                            }
                                        }
                                        else if (MVDiff >= 55) {
                                            if (Cs < 405) {
                                                if (diffAFD < -3) {
                                                    return 1;
                                                }
                                                else if (diffAFD >= -3) {
                                                    if (mcTcor < 67) {
                                                        if (Rs < 284) {
                                                            return 8;
                                                        }
                                                        else if (Rs >= 284) {
                                                            return 1;
                                                        }
                                                    }
                                                    else if (mcTcor >= 67) {
                                                        return 2;
                                                    }
                                                }
                                            }
                                            else if (Cs >= 405) {
                                                if (gchDC < 23) {
                                                    if (Rs < 1103) {
                                                        return 1;
                                                    }
                                                    else if (Rs >= 1103) {
                                                        if (AFD < 18) {
                                                            return 8;
                                                        }
                                                        else if (AFD >= 18) {
                                                            return 1;
                                                        }
                                                    }
                                                }
                                                else if (gchDC >= 23) {
                                                    if (SC < 1874) {
                                                        return 8;
                                                    }
                                                    else if (SC >= 1874) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (mcTcor >= 87) {
                                        if (refDCval < 91) {
                                            if (SC < 1782) {
                                                if (diffAFD < -8) {
                                                    return 8;
                                                }
                                                else if (diffAFD >= -8) {
                                                    if (Cs < 107) {
                                                        if (Cs < 26) {
                                                            return 2;
                                                        }
                                                        else if (Cs >= 26) {
                                                            return 1;
                                                        }
                                                    }
                                                    else if (Cs >= 107) {
                                                        return 2;
                                                    }
                                                }
                                            }
                                            else if (SC >= 1782) {
                                                if (refDCval < 66) {
                                                    return 1;
                                                }
                                                else if (refDCval >= 66) {
                                                    if (gchDC < 6) {
                                                        if (refDCval < 69) {
                                                            if (CsDiff < 8) {
                                                                return 4;
                                                            }
                                                            else if (CsDiff >= 8) {
                                                                return 2;
                                                            }
                                                        }
                                                        else if (refDCval >= 69) {
                                                            return 2;
                                                        }
                                                    }
                                                    else if (gchDC >= 6) {
                                                        if (RsDiff < 4) {
                                                            return 2;
                                                        }
                                                        else if (RsDiff >= 4) {
                                                            return 1;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (refDCval >= 91) {
                                            if (RsCsDiff < 57) {
                                                if (mcTcor < 97) {
                                                    if (Cs < 458) {
                                                        return 1;
                                                    }
                                                    else if (Cs >= 458) {
                                                        if (SC < 2487) {
                                                            return 4;
                                                        }
                                                        else if (SC >= 2487) {
                                                            return 1;
                                                        }
                                                    }
                                                }
                                                else if (mcTcor >= 97) {
                                                    if (Cs < 473) {
                                                        return 2;
                                                    }
                                                    else if (Cs >= 473) {
                                                        if (CsDiff < 2) {
                                                            return 4;
                                                        }
                                                        else if (CsDiff >= 2) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (RsCsDiff >= 57) {
                                                if (diffTSC < -2) {
                                                    return 1;
                                                }
                                                else if (diffTSC >= -2) {
                                                    if (diffAFD < 1) {
                                                        if (Rs < 1975) {
                                                            return 2;
                                                        }
                                                        else if (Rs >= 1975) {
                                                            if (Rs < 2569) {
                                                                return 8;
                                                            }
                                                            else if (Rs >= 2569) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                    else if (diffAFD >= 1) {
                                                        return 1;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (RsDiff >= 8) {
                                    if (mcTcor < 84) {
                                        if (Rs < 1959) {
                                            if (posBalance < 67) {
                                                if (ssDCval < 52) {
                                                    return 4;
                                                }
                                                else if (ssDCval >= 52) {
                                                    if (posBalance < 60) {
                                                        if (TSC < 10) {
                                                            if (SC < 2462) {
                                                                return 8;
                                                            }
                                                            else if (SC >= 2462) {
                                                                return 1;
                                                            }
                                                        }
                                                        else if (TSC >= 10) {
                                                            return 2;
                                                        }
                                                    }
                                                    else if (posBalance >= 60) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                            else if (posBalance >= 67) {
                                                if (Cs < 872) {
                                                    return 4;
                                                }
                                                else if (Cs >= 872) {
                                                    return 1;
                                                }
                                            }
                                        }
                                        else if (Rs >= 1959) {
                                            if (AFD < 20) {
                                                return 2;
                                            }
                                            else if (AFD >= 20) {
                                                if (tcor < 43) {
                                                    return 2;
                                                }
                                                else if (tcor >= 43) {
                                                    if (negBalance < 58) {
                                                        return 1;
                                                    }
                                                    else if (negBalance >= 58) {
                                                        if (Rs < 2266) {
                                                            return 2;
                                                        }
                                                        else if (Rs >= 2266) {
                                                            if (diffAFD < -2) {
                                                                if (Rs < 2604) {
                                                                    return 1;
                                                                }
                                                                else if (Rs >= 2604) {
                                                                    return 2;
                                                                }
                                                            }
                                                            else if (diffAFD >= -2) {
                                                                return 1;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (mcTcor >= 84) {
                                        if (negBalance < 51) {
                                            if (RsCsDiff < 513) {
                                                if (AFD < 15) {
                                                    if (SC < 3813) {
                                                        return 1;
                                                    }
                                                    else if (SC >= 3813) {
                                                        return 2;
                                                    }
                                                }
                                                else if (AFD >= 15) {
                                                    if (MVDiff < 114) {
                                                        if (TSC < 6) {
                                                            return 8;
                                                        }
                                                        else if (TSC >= 6) {
                                                            if (diffAFD < 4) {
                                                                return 2;
                                                            }
                                                            else if (diffAFD >= 4) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                    else if (MVDiff >= 114) {
                                                        if (Rs < 1302) {
                                                            return 1;
                                                        }
                                                        else if (Rs >= 1302) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (RsCsDiff >= 513) {
                                                return 4;
                                            }
                                        }
                                        else if (negBalance >= 51) {
                                            if (diffRsCsdiff < -30) {
                                                if (mcTcor < 87) {
                                                    if (tcor < 70) {
                                                        if (SC < 4947) {
                                                            return 1;
                                                        }
                                                        else if (SC >= 4947) {
                                                            return 2;
                                                        }
                                                    }
                                                    else if (tcor >= 70) {
                                                        return 2;
                                                    }
                                                }
                                                else if (mcTcor >= 87) {
                                                    if (ssDCval < 70) {
                                                        if (AFD < 15) {
                                                            return 4;
                                                        }
                                                        else if (AFD >= 15) {
                                                            return 2;
                                                        }
                                                    }
                                                    else if (ssDCval >= 70) {
                                                        if (Cs < 472) {
                                                            return 4;
                                                        }
                                                        else if (Cs >= 472) {
                                                            return 2;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (diffRsCsdiff >= -30) {
                                                if (tcor < 68) {
                                                    if (RsCsDiff < 239) {
                                                        return 1;
                                                    }
                                                    else if (RsCsDiff >= 239) {
                                                        return 2;
                                                    }
                                                }
                                                else if (tcor >= 68) {
                                                    if (diffAFD < 2) {
                                                        return 2;
                                                    }
                                                    else if (diffAFD >= 2) {
                                                        if (SC < 3353) {
                                                            return 1;
                                                        }
                                                        else if (SC >= 3353) {
                                                            return 2;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                        }
                        else if (Cs >= 1489) {
                            if (tcor < 71) {
                                if (refDCval < 116) {
                                    if (TSC < 12) {
                                        if (ssDCval < 93) {
                                            if (mu_mv_mag_sq < 5) {
                                                if (diffRsCsdiff < 54) {
                                                    return 4;
                                                }
                                                else if (diffRsCsdiff >= 54) {
                                                    if (Cs < 2208) {
                                                        return 1;
                                                    }
                                                    else if (Cs >= 2208) {
                                                        if (RsDiff < 12) {
                                                            return 4;
                                                        }
                                                        else if (RsDiff >= 12) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (mu_mv_mag_sq >= 5) {
                                                if (mcTcor < 84) {
                                                    return 8;
                                                }
                                                else if (mcTcor >= 84) {
                                                    if (MVDiff < 22) {
                                                        return 1;
                                                    }
                                                    else if (MVDiff >= 22) {
                                                        return 2;
                                                    }
                                                }
                                            }
                                        }
                                        else if (ssDCval >= 93) {
                                            if (negBalance < 49) {
                                                return 4;
                                            }
                                            else if (negBalance >= 49) {
                                                return 2;
                                            }
                                        }
                                    }
                                    else if (TSC >= 12) {
                                        if (Cs < 2714) {
                                            if (ssDCval < 73) {
                                                if (AFD < 24) {
                                                    return 8;
                                                }
                                                else if (AFD >= 24) {
                                                    return 1;
                                                }
                                            }
                                            else if (ssDCval >= 73) {
                                                if (RsDiff < 16) {
                                                    if (mcTcor < 93) {
                                                        return 1;
                                                    }
                                                    else if (mcTcor >= 93) {
                                                        return 2;
                                                    }
                                                }
                                                else if (RsDiff >= 16) {
                                                    if (Cs < 1992) {
                                                        return 2;
                                                    }
                                                    else if (Cs >= 1992) {
                                                        if (Rs < 3544) {
                                                            return 8;
                                                        }
                                                        else if (Rs >= 3544) {
                                                            return 2;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (Cs >= 2714) {
                                            if (Rs < 2947) {
                                                if (diffRsCsdiff < 24) {
                                                    return 8;
                                                }
                                                else if (diffRsCsdiff >= 24) {
                                                    return 4;
                                                }
                                            }
                                            else if (Rs >= 2947) {
                                                if (TSC < 15) {
                                                    if (diffRsCsdiff < -32) {
                                                        return 1;
                                                    }
                                                    else if (diffRsCsdiff >= -32) {
                                                        if (diffAFD < 6) {
                                                            return 8;
                                                        }
                                                        else if (diffAFD >= 6) {
                                                            return 1;
                                                        }
                                                    }
                                                }
                                                else if (TSC >= 15) {
                                                    return 8;
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (refDCval >= 116) {
                                    if (MVDiff < 20) {
                                        if (TSC < 14) {
                                            if (mcTcor < 94) {
                                                if (refDCval < 128) {
                                                    if (Cs < 2864) {
                                                        return 1;
                                                    }
                                                    else if (Cs >= 2864) {
                                                        if (RsDiff < 18) {
                                                            return 8;
                                                        }
                                                        else if (RsDiff >= 18) {
                                                            return 1;
                                                        }
                                                    }
                                                }
                                                else if (refDCval >= 128) {
                                                    return 1;
                                                }
                                            }
                                            else if (mcTcor >= 94) {
                                                return 4;
                                            }
                                        }
                                        else if (TSC >= 14) {
                                            if (mcTcor < 77) {
                                                return 1;
                                            }
                                            else if (mcTcor >= 77) {
                                                if (refDCval < 135) {
                                                    if (Cs < 3446) {
                                                        if (AFD < 21) {
                                                            return 4;
                                                        }
                                                        else if (AFD >= 21) {
                                                            return 1;
                                                        }
                                                    }
                                                    else if (Cs >= 3446) {
                                                        if (Cs < 4079) {
                                                            return 4;
                                                        }
                                                        else if (Cs >= 4079) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                                else if (refDCval >= 135) {
                                                    return 4;
                                                }
                                            }
                                        }
                                    }
                                    else if (MVDiff >= 20) {
                                        if (diffMVdiffVal < 13) {
                                            if (ssDCval < 139) {
                                                if (Cs < 3330) {
                                                    if (SC < 3149) {
                                                        return 4;
                                                    }
                                                    else if (SC >= 3149) {
                                                        if (refDCval < 136) {
                                                            return 1;
                                                        }
                                                        else if (refDCval >= 136) {
                                                            if (Rs < 2727) {
                                                                return 4;
                                                            }
                                                            else if (Rs >= 2727) {
                                                                return 1;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (Cs >= 3330) {
                                                    if (Cs < 3669) {
                                                        return 8;
                                                    }
                                                    else if (Cs >= 3669) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                            else if (ssDCval >= 139) {
                                                if (mcTcor < 92) {
                                                    if (SC < 3766) {
                                                        return 8;
                                                    }
                                                    else if (SC >= 3766) {
                                                        return 4;
                                                    }
                                                }
                                                else if (mcTcor >= 92) {
                                                    return 2;
                                                }
                                            }
                                        }
                                        else if (diffMVdiffVal >= 13) {
                                            if (gchDC < 2) {
                                                if (CsDiff < 13) {
                                                    return 1;
                                                }
                                                else if (CsDiff >= 13) {
                                                    return 2;
                                                }
                                            }
                                            else if (gchDC >= 2) {
                                                return 8;
                                            }
                                        }
                                    }
                                }
                            }
                            else if (tcor >= 71) {
                                if (posBalance < 57) {
                                    if (negBalance < 51) {
                                        if (refDCval < 93) {
                                            if (RsDiff < 10) {
                                                if (Rs < 1842) {
                                                    if (mcTcor < 88) {
                                                        if (AFD < 14) {
                                                            if (Rs < 1540) {
                                                                return 8;
                                                            }
                                                            else if (Rs >= 1540) {
                                                                return 2;
                                                            }
                                                        }
                                                        else if (AFD >= 14) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (mcTcor >= 88) {
                                                        if (tcor < 91) {
                                                            return 1;
                                                        }
                                                        else if (tcor >= 91) {
                                                            if (Cs < 1622) {
                                                                if (Cs < 1555) {
                                                                    return 1;
                                                                }
                                                                else if (Cs >= 1555) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (Cs >= 1622) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (Rs >= 1842) {
                                                    if (ssDCval < 90) {
                                                        if (posBalance < 51) {
                                                            if (SC < 4846) {
                                                                return 8;
                                                            }
                                                            else if (SC >= 4846) {
                                                                return 1;
                                                            }
                                                        }
                                                        else if (posBalance >= 51) {
                                                            if (diffRsCsdiff < -9) {
                                                                if (Rs < 2374) {
                                                                    if (Rs < 2291) {
                                                                        return 8;
                                                                    }
                                                                    else if (Rs >= 2291) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (Rs >= 2374) {
                                                                    if (diffTSC < 1) {
                                                                        if (MVDiff < 7) {
                                                                            return 4;
                                                                        }
                                                                        else if (MVDiff >= 7) {
                                                                            if (AFD < 17) {
                                                                                if (diffAFD < -6) {
                                                                                    return 4;
                                                                                }
                                                                                else if (diffAFD >= -6) {
                                                                                    return 8;
                                                                                }
                                                                            }
                                                                            else if (AFD >= 17) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (diffTSC >= 1) {
                                                                        return 4;
                                                                    }
                                                                }
                                                            }
                                                            else if (diffRsCsdiff >= -9) {
                                                                if (SC < 4402) {
                                                                    if (Cs < 1778) {
                                                                        return 1;
                                                                    }
                                                                    else if (Cs >= 1778) {
                                                                        return 8;
                                                                    }
                                                                }
                                                                else if (SC >= 4402) {
                                                                    if (Rs < 3889) {
                                                                        return 4;
                                                                    }
                                                                    else if (Rs >= 3889) {
                                                                        return 8;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (ssDCval >= 90) {
                                                        if (MVDiff < 18) {
                                                            return 1;
                                                        }
                                                        else if (MVDiff >= 18) {
                                                            if (mcTcor < 92) {
                                                                if (Cs < 2592) {
                                                                    return 4;
                                                                }
                                                                else if (Cs >= 2592) {
                                                                    return 1;
                                                                }
                                                            }
                                                            else if (mcTcor >= 92) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                            else if (RsDiff >= 10) {
                                                if (posBalance < 55) {
                                                    if (mcTcor < 88) {
                                                        if (Rs < 2899) {
                                                            return 1;
                                                        }
                                                        else if (Rs >= 2899) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (mcTcor >= 88) {
                                                        if (posBalance < 46) {
                                                            return 8;
                                                        }
                                                        else if (posBalance >= 46) {
                                                            if (RsCsDiff < 173) {
                                                                return 8;
                                                            }
                                                            else if (RsCsDiff >= 173) {
                                                                if (mu_mv_mag_sq < 28) {
                                                                    return 1;
                                                                }
                                                                else if (mu_mv_mag_sq >= 28) {
                                                                    if (diffMVdiffVal < -3) {
                                                                        if (Cs < 2646) {
                                                                            return 1;
                                                                        }
                                                                        else if (Cs >= 2646) {
                                                                            if (RsDiff < 12) {
                                                                                return 8;
                                                                            }
                                                                            else if (RsDiff >= 12) {
                                                                                return 1;
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (diffMVdiffVal >= -3) {
                                                                        return 1;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (posBalance >= 55) {
                                                    if (ssDCval < 77) {
                                                        if (Rs < 3933) {
                                                            return 1;
                                                        }
                                                        else if (Rs >= 3933) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (ssDCval >= 77) {
                                                        if (Cs < 2232) {
                                                            return 1;
                                                        }
                                                        else if (Cs >= 2232) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (refDCval >= 93) {
                                            if (tcor < 86) {
                                                if (tcor < 76) {
                                                    if (MVDiff < 15) {
                                                        if (Rs < 4249) {
                                                            return 1;
                                                        }
                                                        else if (Rs >= 4249) {
                                                            if (SC < 9797) {
                                                                return 4;
                                                            }
                                                            else if (SC >= 9797) {
                                                                return 1;
                                                            }
                                                        }
                                                    }
                                                    else if (MVDiff >= 15) {
                                                        if (MVDiff < 16) {
                                                            return 8;
                                                        }
                                                        else if (MVDiff >= 16) {
                                                            if (mcTcor < 92) {
                                                                return 4;
                                                            }
                                                            else if (mcTcor >= 92) {
                                                                return 2;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (tcor >= 76) {
                                                    if (Rs < 5268) {
                                                        if (diffMVdiffVal < -4) {
                                                            if (diffTSC < -1) {
                                                                return 2;
                                                            }
                                                            else if (diffTSC >= -1) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (diffMVdiffVal >= -4) {
                                                            if (gchDC < 2) {
                                                                if (tcor < 79) {
                                                                    if (MVDiff < 10) {
                                                                        return 1;
                                                                    }
                                                                    else if (MVDiff >= 10) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (tcor >= 79) {
                                                                    if (RsCsDiff < 278) {
                                                                        if (RsDiff < 7) {
                                                                            return 1;
                                                                        }
                                                                        else if (RsDiff >= 7) {
                                                                            if (Rs < 2712) {
                                                                                if (Rs < 2019) {
                                                                                    return 8;
                                                                                }
                                                                                else if (Rs >= 2019) {
                                                                                    if (SC < 5023) {
                                                                                        return 4;
                                                                                    }
                                                                                    else if (SC >= 5023) {
                                                                                        if (Cs < 2573) {
                                                                                            return 8;
                                                                                        }
                                                                                        else if (Cs >= 2573) {
                                                                                            return 4;
                                                                                        }
                                                                                    }
                                                                                }
                                                                            }
                                                                            else if (Rs >= 2712) {
                                                                                return 8;
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (RsCsDiff >= 278) {
                                                                        return 1;
                                                                    }
                                                                }
                                                            }
                                                            else if (gchDC >= 2) {
                                                                if (diffRsCsdiff < 54) {
                                                                    return 4;
                                                                }
                                                                else if (diffRsCsdiff >= 54) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (Rs >= 5268) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                            else if (tcor >= 86) {
                                                if (TSC < 8) {
                                                    if (negBalance < 38) {
                                                        if (Rs < 4000) {
                                                            return 8;
                                                        }
                                                        else if (Rs >= 4000) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (negBalance >= 38) {
                                                        if (CsDiff < 8) {
                                                            return 4;
                                                        }
                                                        else if (CsDiff >= 8) {
                                                            if (Rs < 1786) {
                                                                return 2;
                                                            }
                                                            else if (Rs >= 1786) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (TSC >= 8) {
                                                    if (MVDiff < 20) {
                                                        return 8;
                                                    }
                                                    else if (MVDiff >= 20) {
                                                        if (AFD < 13) {
                                                            return 8;
                                                        }
                                                        else if (AFD >= 13) {
                                                            if (Cs < 2283) {
                                                                return 1;
                                                            }
                                                            else if (Cs >= 2283) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (negBalance >= 51) {
                                        if (Rs < 3434) {
                                            if (tcor < 82) {
                                                if (Rs < 1815) {
                                                    return 1;
                                                }
                                                else if (Rs >= 1815) {
                                                    if (AFD < 17) {
                                                        if (MVDiff < 16) {
                                                            if (SC < 5117) {
                                                                return 2;
                                                            }
                                                            else if (SC >= 5117) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (MVDiff >= 16) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (AFD >= 17) {
                                                        if (refDCval < 79) {
                                                            if (MVDiff < 16) {
                                                                return 4;
                                                            }
                                                            else if (MVDiff >= 16) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (refDCval >= 79) {
                                                            if (AFD < 20) {
                                                                return 4;
                                                            }
                                                            else if (AFD >= 20) {
                                                                if (TSC < 11) {
                                                                    return 4;
                                                                }
                                                                else if (TSC >= 11) {
                                                                    if (Rs < 2834) {
                                                                        return 1;
                                                                    }
                                                                    else if (Rs >= 2834) {
                                                                        return 8;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                            else if (tcor >= 82) {
                                                if (MVDiff < 10) {
                                                    if (negBalance < 53) {
                                                        return 4;
                                                    }
                                                    else if (negBalance >= 53) {
                                                        if (Cs < 2421) {
                                                            if (Rs < 1847) {
                                                                return 4;
                                                            }
                                                            else if (Rs >= 1847) {
                                                                return 2;
                                                            }
                                                        }
                                                        else if (Cs >= 2421) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                                else if (MVDiff >= 10) {
                                                    if (mu_mv_mag_sq < 21) {
                                                        if (Cs < 2013) {
                                                            return 2;
                                                        }
                                                        else if (Cs >= 2013) {
                                                            if (Cs < 2210) {
                                                                if (RsDiff < 9) {
                                                                    return 2;
                                                                }
                                                                else if (RsDiff >= 9) {
                                                                    return 1;
                                                                }
                                                            }
                                                            else if (Cs >= 2210) {
                                                                return 2;
                                                            }
                                                        }
                                                    }
                                                    else if (mu_mv_mag_sq >= 21) {
                                                        if (Rs < 3222) {
                                                            if (Rs < 2523) {
                                                                return 1;
                                                            }
                                                            else if (Rs >= 2523) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (Rs >= 3222) {
                                                            if (diffRsCsdiff < -11) {
                                                                return 8;
                                                            }
                                                            else if (diffRsCsdiff >= -11) {
                                                                if (Cs < 3303) {
                                                                    return 2;
                                                                }
                                                                else if (Cs >= 3303) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (Rs >= 3434) {
                                            if (posBalance < 53) {
                                                if (mu_mv_mag_sq < 32) {
                                                    return 8;
                                                }
                                                else if (mu_mv_mag_sq >= 32) {
                                                    if (mcTcor < 94) {
                                                        return 4;
                                                    }
                                                    else if (mcTcor >= 94) {
                                                        if (Cs < 2475) {
                                                            return 4;
                                                        }
                                                        else if (Cs >= 2475) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (posBalance >= 53) {
                                                if (Cs < 3579) {
                                                    if (MVDiff < 21) {
                                                        if (gchDC < 3) {
                                                            return 4;
                                                        }
                                                        else if (gchDC >= 3) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (MVDiff >= 21) {
                                                        if (Cs < 3425) {
                                                            return 8;
                                                        }
                                                        else if (Cs >= 3425) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                                else if (Cs >= 3579) {
                                                    if (refDCval < 65) {
                                                        if (TSC < 9) {
                                                            return 2;
                                                        }
                                                        else if (TSC >= 9) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (refDCval >= 65) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (posBalance >= 57) {
                                    if (refDCval < 76) {
                                        if (RsDiff < 5) {
                                            if (ssDCval < 49) {
                                                if (AFD < 7) {
                                                    return 1;
                                                }
                                                else if (AFD >= 7) {
                                                    return 4;
                                                }
                                            }
                                            else if (ssDCval >= 49) {
                                                if (MVDiff < 20) {
                                                    if (AFD < 11) {
                                                        return 4;
                                                    }
                                                    else if (AFD >= 11) {
                                                        return 2;
                                                    }
                                                }
                                                else if (MVDiff >= 20) {
                                                    if (TSC < 7) {
                                                        return 8;
                                                    }
                                                    else if (TSC >= 7) {
                                                        if (tcor < 89) {
                                                            if (ssDCval < 52) {
                                                                return 2;
                                                            }
                                                            else if (ssDCval >= 52) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (tcor >= 89) {
                                                            if (posBalance < 63) {
                                                                if (MVDiff < 53) {
                                                                    if (Rs < 3609) {
                                                                        return 2;
                                                                    }
                                                                    else if (Rs >= 3609) {
                                                                        if (Cs < 3513) {
                                                                            return 8;
                                                                        }
                                                                        else if (Cs >= 3513) {
                                                                            return 2;
                                                                        }
                                                                    }
                                                                }
                                                                else if (MVDiff >= 53) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (posBalance >= 63) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (RsDiff >= 5) {
                                            if (TSC < 12) {
                                                if (negBalance < 48) {
                                                    if (Cs < 2920) {
                                                        return 1;
                                                    }
                                                    else if (Cs >= 2920) {
                                                        return 8;
                                                    }
                                                }
                                                else if (negBalance >= 48) {
                                                    if (Rs < 1629) {
                                                        return 8;
                                                    }
                                                    else if (Rs >= 1629) {
                                                        if (gchDC < 3) {
                                                            return 4;
                                                        }
                                                        else if (gchDC >= 3) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (TSC >= 12) {
                                                if (negBalance < 52) {
                                                    if (AFD < 19) {
                                                        if (tcor < 80) {
                                                            return 8;
                                                        }
                                                        else if (tcor >= 80) {
                                                            if (Cs < 3602) {
                                                                return 4;
                                                            }
                                                            else if (Cs >= 3602) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                    else if (AFD >= 19) {
                                                        return 1;
                                                    }
                                                }
                                                else if (negBalance >= 52) {
                                                    return 8;
                                                }
                                            }
                                        }
                                    }
                                    else if (refDCval >= 76) {
                                        if (AFD < 13) {
                                            if (tcor < 90) {
                                                if (SC < 4565) {
                                                    if (diffMVdiffVal < 5) {
                                                        if (mcTcor < 84) {
                                                            if (Rs < 1891) {
                                                                return 4;
                                                            }
                                                            else if (Rs >= 1891) {
                                                                return 2;
                                                            }
                                                        }
                                                        else if (mcTcor >= 84) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (diffMVdiffVal >= 5) {
                                                        if (Rs < 1525) {
                                                            return 4;
                                                        }
                                                        else if (Rs >= 1525) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                                else if (SC >= 4565) {
                                                    if (RsDiff < 8) {
                                                        if (ssDCval < 123) {
                                                            return 4;
                                                        }
                                                        else if (ssDCval >= 123) {
                                                            if (negBalance < 52) {
                                                                return 8;
                                                            }
                                                            else if (negBalance >= 52) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                    else if (RsDiff >= 8) {
                                                        if (diffAFD < -1) {
                                                            return 2;
                                                        }
                                                        else if (diffAFD >= -1) {
                                                            if (ssDCval < 129) {
                                                                return 4;
                                                            }
                                                            else if (ssDCval >= 129) {
                                                                return 2;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                            else if (tcor >= 90) {
                                                if (mcTcor < 96) {
                                                    return 2;
                                                }
                                                else if (mcTcor >= 96) {
                                                    if (Cs < 1657) {
                                                        return 8;
                                                    }
                                                    else if (Cs >= 1657) {
                                                        return 1;
                                                    }
                                                }
                                            }
                                        }
                                        else if (AFD >= 13) {
                                            if (MVDiff < 5) {
                                                if (Cs < 2479) {
                                                    if (ssDCval < 96) {
                                                        if (diffRsCsdiff < 27) {
                                                            return 2;
                                                        }
                                                        else if (diffRsCsdiff >= 27) {
                                                            if (AFD < 16) {
                                                                return 4;
                                                            }
                                                            else if (AFD >= 16) {
                                                                return 2;
                                                            }
                                                        }
                                                    }
                                                    else if (ssDCval >= 96) {
                                                        if (Rs < 2702) {
                                                            return 4;
                                                        }
                                                        else if (Rs >= 2702) {
                                                            return 2;
                                                        }
                                                    }
                                                }
                                                else if (Cs >= 2479) {
                                                    if (Rs < 3659) {
                                                        return 8;
                                                    }
                                                    else if (Rs >= 3659) {
                                                        if (SC < 7619) {
                                                            return 4;
                                                        }
                                                        else if (SC >= 7619) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (MVDiff >= 5) {
                                                if (ssDCval < 159) {
                                                    if (ssDCval < 137) {
                                                        if (tcor < 79) {
                                                            if (Rs < 3090) {
                                                                if (AFD < 23) {
                                                                    if (ssDCval < 110) {
                                                                        if (posBalance < 59) {
                                                                            return 2;
                                                                        }
                                                                        else if (posBalance >= 59) {
                                                                            if (TSC < 10) {
                                                                                return 2;
                                                                            }
                                                                            else if (TSC >= 10) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (ssDCval >= 110) {
                                                                        if (MVDiff < 59) {
                                                                            if (RsDiff < 11) {
                                                                                if (RsCsDiff < 92) {
                                                                                    return 4;
                                                                                }
                                                                                else if (RsCsDiff >= 92) {
                                                                                    return 1;
                                                                                }
                                                                            }
                                                                            else if (RsDiff >= 11) {
                                                                                if (RsCsDiff < 331) {
                                                                                    return 4;
                                                                                }
                                                                                else if (RsCsDiff >= 331) {
                                                                                    return 2;
                                                                                }
                                                                            }
                                                                        }
                                                                        else if (MVDiff >= 59) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                }
                                                                else if (AFD >= 23) {
                                                                    if (mu_mv_mag_sq < 5) {
                                                                        if (SC < 3550) {
                                                                            return 1;
                                                                        }
                                                                        else if (SC >= 3550) {
                                                                            return 2;
                                                                        }
                                                                    }
                                                                    else if (mu_mv_mag_sq >= 5) {
                                                                        if (Rs < 2819) {
                                                                            return 1;
                                                                        }
                                                                        else if (Rs >= 2819) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                            else if (Rs >= 3090) {
                                                                if (SC < 7453) {
                                                                    if (mcTcor < 82) {
                                                                        if (SC < 6602) {
                                                                            return 1;
                                                                        }
                                                                        else if (SC >= 6602) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                    else if (mcTcor >= 82) {
                                                                        return 1;
                                                                    }
                                                                }
                                                                else if (SC >= 7453) {
                                                                    if (SC < 9035) {
                                                                        return 4;
                                                                    }
                                                                    else if (SC >= 9035) {
                                                                        return 1;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (tcor >= 79) {
                                                            if (ssDCval < 126) {
                                                                if (Rs < 2332) {
                                                                    if (CsDiff < 8) {
                                                                        if (ssDCval < 85) {
                                                                            return 2;
                                                                        }
                                                                        else if (ssDCval >= 85) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                    else if (CsDiff >= 8) {
                                                                        if (mcTcor < 94) {
                                                                            if (AFD < 19) {
                                                                                if (TSC < 12) {
                                                                                    return 4;
                                                                                }
                                                                                else if (TSC >= 12) {
                                                                                    if (ssDCval < 116) {
                                                                                        return 2;
                                                                                    }
                                                                                    else if (ssDCval >= 116) {
                                                                                        return 4;
                                                                                    }
                                                                                }
                                                                            }
                                                                            else if (AFD >= 19) {
                                                                                return 1;
                                                                            }
                                                                        }
                                                                        else if (mcTcor >= 94) {
                                                                            if (RsCsDiff < 109) {
                                                                                return 4;
                                                                            }
                                                                            else if (RsCsDiff >= 109) {
                                                                                return 2;
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                                else if (Rs >= 2332) {
                                                                    if (RsCsDiff < 133) {
                                                                        if (tcor < 90) {
                                                                            if (TSC < 10) {
                                                                                return 2;
                                                                            }
                                                                            else if (TSC >= 10) {
                                                                                if (Rs < 2579) {
                                                                                    if (Rs < 2485) {
                                                                                        if (MVDiff < 13) {
                                                                                            return 2;
                                                                                        }
                                                                                        else if (MVDiff >= 13) {
                                                                                            return 1;
                                                                                        }
                                                                                    }
                                                                                    else if (Rs >= 2485) {
                                                                                        return 2;
                                                                                    }
                                                                                }
                                                                                else if (Rs >= 2579) {
                                                                                    return 1;
                                                                                }
                                                                            }
                                                                        }
                                                                        else if (tcor >= 90) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                    else if (RsCsDiff >= 133) {
                                                                        if (CsDiff < 10) {
                                                                            if (TSC < 12) {
                                                                                if (SC < 4534) {
                                                                                    if (posBalance < 60) {
                                                                                        if (RsCsDiff < 174) {
                                                                                            return 4;
                                                                                        }
                                                                                        else if (RsCsDiff >= 174) {
                                                                                            return 2;
                                                                                        }
                                                                                    }
                                                                                    else if (posBalance >= 60) {
                                                                                        if (RsDiff < 10) {
                                                                                            return 1;
                                                                                        }
                                                                                        else if (RsDiff >= 10) {
                                                                                            return 8;
                                                                                        }
                                                                                    }
                                                                                }
                                                                                else if (SC >= 4534) {
                                                                                    if (diffMVdiffVal < 6) {
                                                                                        if (mcTcor < 88) {
                                                                                            return 4;
                                                                                        }
                                                                                        else if (mcTcor >= 88) {
                                                                                            return 2;
                                                                                        }
                                                                                    }
                                                                                    else if (diffMVdiffVal >= 6) {
                                                                                        return 4;
                                                                                    }
                                                                                }
                                                                            }
                                                                            else if (TSC >= 12) {
                                                                                if (mu_mv_mag_sq < 3) {
                                                                                    if (RsCsDiff < 191) {
                                                                                        if (ssDCval < 92) {
                                                                                            if (Rs < 2461) {
                                                                                                return 2;
                                                                                            }
                                                                                            else if (Rs >= 2461) {
                                                                                                return 4;
                                                                                            }
                                                                                        }
                                                                                        else if (ssDCval >= 92) {
                                                                                            if (AFD < 16) {
                                                                                                return 4;
                                                                                            }
                                                                                            else if (AFD >= 16) {
                                                                                                return 1;
                                                                                            }
                                                                                        }
                                                                                    }
                                                                                    else if (RsCsDiff >= 191) {
                                                                                        return 2;
                                                                                    }
                                                                                }
                                                                                else if (mu_mv_mag_sq >= 3) {
                                                                                    if (Cs < 1876) {
                                                                                        return 8;
                                                                                    }
                                                                                    else if (Cs >= 1876) {
                                                                                        if (RsDiff < 13) {
                                                                                            if (diffAFD < 3) {
                                                                                                return 1;
                                                                                            }
                                                                                            else if (diffAFD >= 3) {
                                                                                                if (Rs < 2532) {
                                                                                                    return 1;
                                                                                                }
                                                                                                else if (Rs >= 2532) {
                                                                                                    return 2;
                                                                                                }
                                                                                            }
                                                                                        }
                                                                                        else if (RsDiff >= 13) {
                                                                                            return 8;
                                                                                        }
                                                                                    }
                                                                                }
                                                                            }
                                                                        }
                                                                        else if (CsDiff >= 10) {
                                                                            if (AFD < 21) {
                                                                                if (refDCval < 86) {
                                                                                    if (Cs < 2467) {
                                                                                        return 2;
                                                                                    }
                                                                                    else if (Cs >= 2467) {
                                                                                        return 1;
                                                                                    }
                                                                                }
                                                                                else if (refDCval >= 86) {
                                                                                    if (RsCsDiff < 267) {
                                                                                        if (AFD < 18) {
                                                                                            return 4;
                                                                                        }
                                                                                        else if (AFD >= 18) {
                                                                                            return 1;
                                                                                        }
                                                                                    }
                                                                                    else if (RsCsDiff >= 267) {
                                                                                        return 4;
                                                                                    }
                                                                                }
                                                                            }
                                                                            else if (AFD >= 21) {
                                                                                return 1;
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                            else if (ssDCval >= 126) {
                                                                if (Rs < 3660) {
                                                                    if (CsDiff < 7) {
                                                                        return 8;
                                                                    }
                                                                    else if (CsDiff >= 7) {
                                                                        return 2;
                                                                    }
                                                                }
                                                                else if (Rs >= 3660) {
                                                                    return 1;
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (ssDCval >= 137) {
                                                        if (TSC < 11) {
                                                            return 2;
                                                        }
                                                        else if (TSC >= 11) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                                else if (ssDCval >= 159) {
                                                    return 2;
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                        }
                    }
                    else if (mu_mv_mag_sq >= 38) {
                        if (ssDCval < 104) {
                            if (SC < 6283) {
                                if (Cs < 1588) {
                                    if (Cs < 547) {
                                        if (SC < 864) {
                                            if (refDCval < 85) {
                                                if (tcor < 84) {
                                                    if (mcTcor < 85) {
                                                        if (Rs < 403) {
                                                            if (gchDC < 1) {
                                                                if (posBalance < 33) {
                                                                    if (MVDiff < 65) {
                                                                        return 1;
                                                                    }
                                                                    else if (MVDiff >= 65) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (posBalance >= 33) {
                                                                    if (mcTcor < 82) {
                                                                        if (AFD < 5) {
                                                                            return 8;
                                                                        }
                                                                        else if (AFD >= 5) {
                                                                            if (Cs < 311) {
                                                                                return 8;
                                                                            }
                                                                            else if (Cs >= 311) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (mcTcor >= 82) {
                                                                        if (mu_mv_mag_sq < 62) {
                                                                            if (Cs < 312) {
                                                                                return 4;
                                                                            }
                                                                            else if (Cs >= 312) {
                                                                                if (Cs < 333) {
                                                                                    return 1;
                                                                                }
                                                                                else if (Cs >= 333) {
                                                                                    return 4;
                                                                                }
                                                                            }
                                                                        }
                                                                        else if (mu_mv_mag_sq >= 62) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                            else if (gchDC >= 1) {
                                                                if (ssDCval < 44) {
                                                                    if (MVDiff < 71) {
                                                                        return 4;
                                                                    }
                                                                    else if (MVDiff >= 71) {
                                                                        if (AFD < 1) {
                                                                            return 4;
                                                                        }
                                                                        else if (AFD >= 1) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                }
                                                                else if (ssDCval >= 44) {
                                                                    if (posBalance < 102) {
                                                                        return 1;
                                                                    }
                                                                    else if (posBalance >= 102) {
                                                                        return 4;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (Rs >= 403) {
                                                            if (Cs < 375) {
                                                                if (mu_mv_mag_sq < 54) {
                                                                    return 1;
                                                                }
                                                                else if (mu_mv_mag_sq >= 54) {
                                                                    if (negBalance < 41) {
                                                                        return 4;
                                                                    }
                                                                    else if (negBalance >= 41) {
                                                                        if (MVDiff < 16) {
                                                                            return 4;
                                                                        }
                                                                        else if (MVDiff >= 16) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                            else if (Cs >= 375) {
                                                                if (refDCval < 38) {
                                                                    return 4;
                                                                }
                                                                else if (refDCval >= 38) {
                                                                    if (Rs < 456) {
                                                                        return 2;
                                                                    }
                                                                    else if (Rs >= 456) {
                                                                        return 4;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (mcTcor >= 85) {
                                                        if (mu_mv_mag_sq < 57) {
                                                            if (AFD < 14) {
                                                                return 4;
                                                            }
                                                            else if (AFD >= 14) {
                                                                return 2;
                                                            }
                                                        }
                                                        else if (mu_mv_mag_sq >= 57) {
                                                            if (Cs < 355) {
                                                                return 8;
                                                            }
                                                            else if (Cs >= 355) {
                                                                return 2;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (tcor >= 84) {
                                                    if (MVDiff < 43) {
                                                        if (negBalance < 11) {
                                                            if (Rs < 286) {
                                                                return 8;
                                                            }
                                                            else if (Rs >= 286) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (negBalance >= 11) {
                                                            if (refDCval < 38) {
                                                                if (negBalance < 29) {
                                                                    if (MVDiff < 30) {
                                                                        return 4;
                                                                    }
                                                                    else if (MVDiff >= 30) {
                                                                        return 8;
                                                                    }
                                                                }
                                                                else if (negBalance >= 29) {
                                                                    if (mcTcor < 86) {
                                                                        if (Rs < 414) {
                                                                            return 8;
                                                                        }
                                                                        else if (Rs >= 414) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                    else if (mcTcor >= 86) {
                                                                        return 2;
                                                                    }
                                                                }
                                                            }
                                                            else if (refDCval >= 38) {
                                                                if (posBalance < 7) {
                                                                    return 8;
                                                                }
                                                                else if (posBalance >= 7) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (MVDiff >= 43) {
                                                        if (diffMVdiffVal < 17) {
                                                            return 2;
                                                        }
                                                        else if (diffMVdiffVal >= 17) {
                                                            return 1;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (refDCval >= 85) {
                                                if (RsCsDiff < 7) {
                                                    return 2;
                                                }
                                                else if (RsCsDiff >= 7) {
                                                    return 1;
                                                }
                                            }
                                        }
                                        else if (SC >= 864) {
                                            if (ssDCval < 39) {
                                                if (MVDiff < 21) {
                                                    if (mcTcor < 83) {
                                                        return 4;
                                                    }
                                                    else if (mcTcor >= 83) {
                                                        return 2;
                                                    }
                                                }
                                                else if (MVDiff >= 21) {
                                                    if (Rs < 649) {
                                                        return 2;
                                                    }
                                                    else if (Rs >= 649) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                            else if (ssDCval >= 39) {
                                                if (CsDiff < 2) {
                                                    if (Rs < 1039) {
                                                        if (ssDCval < 43) {
                                                            return 4;
                                                        }
                                                        else if (ssDCval >= 43) {
                                                            if (diffAFD < 1) {
                                                                if (Cs < 530) {
                                                                    if (Cs < 393) {
                                                                        return 8;
                                                                    }
                                                                    else if (Cs >= 393) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (Cs >= 530) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (diffAFD >= 1) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                    else if (Rs >= 1039) {
                                                        if (MVDiff < 11) {
                                                            if (Cs < 321) {
                                                                return 4;
                                                            }
                                                            else if (Cs >= 321) {
                                                                return 2;
                                                            }
                                                        }
                                                        else if (MVDiff >= 11) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                                else if (CsDiff >= 2) {
                                                    if (ssDCval < 52) {
                                                        if (Cs < 535) {
                                                            if (Rs < 1071) {
                                                                return 8;
                                                            }
                                                            else if (Rs >= 1071) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (Cs >= 535) {
                                                            return 1;
                                                        }
                                                    }
                                                    else if (ssDCval >= 52) {
                                                        return 2;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (Cs >= 547) {
                                        if (Cs < 1011) {
                                            if (Cs < 830) {
                                                if (negBalance < 53) {
                                                    if (ssDCval < 64) {
                                                        if (TSCindex < 6) {
                                                            if (Cs < 700) {
                                                                if (negBalance < 48) {
                                                                    if (RsCsDiff < 7) {
                                                                        if (refDCval < 53) {
                                                                            return 4;
                                                                        }
                                                                        else if (refDCval >= 53) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                    else if (RsCsDiff >= 7) {
                                                                        if (negBalance < 34) {
                                                                            return 4;
                                                                        }
                                                                        else if (negBalance >= 34) {
                                                                            if (Cs < 600) {
                                                                                return 4;
                                                                            }
                                                                            else if (Cs >= 600) {
                                                                                return 8;
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                                else if (negBalance >= 48) {
                                                                    return 2;
                                                                }
                                                            }
                                                            else if (Cs >= 700) {
                                                                if (mu_mv_mag_sq < 50) {
                                                                    if (posBalance < 51) {
                                                                        if (RsDiff < 4) {
                                                                            return 1;
                                                                        }
                                                                        else if (RsDiff >= 4) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                    else if (posBalance >= 51) {
                                                                        if (Cs < 741) {
                                                                            return 4;
                                                                        }
                                                                        else if (Cs >= 741) {
                                                                            if (SC < 1563) {
                                                                                return 1;
                                                                            }
                                                                            else if (SC >= 1563) {
                                                                                return 2;
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                                else if (mu_mv_mag_sq >= 50) {
                                                                    if (Rs < 1177) {
                                                                        if (TSC < 3) {
                                                                            return 1;
                                                                        }
                                                                        else if (TSC >= 3) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                    else if (Rs >= 1177) {
                                                                        return 8;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (TSCindex >= 6) {
                                                            if (MVDiff < 19) {
                                                                if (tcor < 93) {
                                                                    if (negBalance < 43) {
                                                                        return 4;
                                                                    }
                                                                    else if (negBalance >= 43) {
                                                                        if (tcor < 79) {
                                                                            return 4;
                                                                        }
                                                                        else if (tcor >= 79) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                }
                                                                else if (tcor >= 93) {
                                                                    return 1;
                                                                }
                                                            }
                                                            else if (MVDiff >= 19) {
                                                                if (diffMVdiffVal < 13) {
                                                                    if (negBalance < 49) {
                                                                        if (Rs < 742) {
                                                                            return 1;
                                                                        }
                                                                        else if (Rs >= 742) {
                                                                            if (Rs < 865) {
                                                                                if (AFD < 7) {
                                                                                    if (Cs < 560) {
                                                                                        if (CsDiff < 2) {
                                                                                            return 4;
                                                                                        }
                                                                                        else if (CsDiff >= 2) {
                                                                                            return 1;
                                                                                        }
                                                                                    }
                                                                                    else if (Cs >= 560) {
                                                                                        return 4;
                                                                                    }
                                                                                }
                                                                                else if (AFD >= 7) {
                                                                                    return 1;
                                                                                }
                                                                            }
                                                                            else if (Rs >= 865) {
                                                                                return 1;
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (negBalance >= 49) {
                                                                        if (MVDiff < 29) {
                                                                            return 1;
                                                                        }
                                                                        else if (MVDiff >= 29) {
                                                                            if (mcTcor < 79) {
                                                                                return 1;
                                                                            }
                                                                            else if (mcTcor >= 79) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                                else if (diffMVdiffVal >= 13) {
                                                                    if (Cs < 677) {
                                                                        return 1;
                                                                    }
                                                                    else if (Cs >= 677) {
                                                                        return 4;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (ssDCval >= 64) {
                                                        if (ssDCval < 69) {
                                                            if (mu_mv_mag_sq < 61) {
                                                                if (Rs < 605) {
                                                                    return 4;
                                                                }
                                                                else if (Rs >= 605) {
                                                                    return 2;
                                                                }
                                                            }
                                                            else if (mu_mv_mag_sq >= 61) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (ssDCval >= 69) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                                else if (negBalance >= 53) {
                                                    if (CsDiff < 3) {
                                                        if (mu_mv_mag_sq < 39) {
                                                            return 4;
                                                        }
                                                        else if (mu_mv_mag_sq >= 39) {
                                                            if (Cs < 759) {
                                                                return 8;
                                                            }
                                                            else if (Cs >= 759) {
                                                                if (Cs < 774) {
                                                                    return 4;
                                                                }
                                                                else if (Cs >= 774) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (CsDiff >= 3) {
                                                        if (Rs < 360) {
                                                            return 4;
                                                        }
                                                        else if (Rs >= 360) {
                                                            return 2;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (Cs >= 830) {
                                                if (Rs < 1184) {
                                                    if (refDCval < 42) {
                                                        if (AFD < 2) {
                                                            return 1;
                                                        }
                                                        else if (AFD >= 2) {
                                                            if (AFD < 4) {
                                                                return 8;
                                                            }
                                                            else if (AFD >= 4) {
                                                                return 2;
                                                            }
                                                        }
                                                    }
                                                    else if (refDCval >= 42) {
                                                        if (diffMVdiffVal < 7) {
                                                            if (gchDC < 4) {
                                                                if (Rs < 1010) {
                                                                    if (MVDiff < 6) {
                                                                        if (Rs < 1001) {
                                                                            return 4;
                                                                        }
                                                                        else if (Rs >= 1001) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                    else if (MVDiff >= 6) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (Rs >= 1010) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (gchDC >= 4) {
                                                                return 1;
                                                            }
                                                        }
                                                        else if (diffMVdiffVal >= 7) {
                                                            if (gchDC < 2) {
                                                                return 1;
                                                            }
                                                            else if (gchDC >= 2) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (Rs >= 1184) {
                                                    if (negBalance < 45) {
                                                        if (tcor < 92) {
                                                            if (ssDCval < 51) {
                                                                if (SC < 2408) {
                                                                    return 1;
                                                                }
                                                                else if (SC >= 2408) {
                                                                    if (Cs < 978) {
                                                                        return 8;
                                                                    }
                                                                    else if (Cs >= 978) {
                                                                        return 1;
                                                                    }
                                                                }
                                                            }
                                                            else if (ssDCval >= 51) {
                                                                if (posBalance < 48) {
                                                                    return 4;
                                                                }
                                                                else if (posBalance >= 48) {
                                                                    if (SC < 2273) {
                                                                        return 1;
                                                                    }
                                                                    else if (SC >= 2273) {
                                                                        if (Rs < 1387) {
                                                                            if (SC < 2349) {
                                                                                return 1;
                                                                            }
                                                                            else if (SC >= 2349) {
                                                                                return 8;
                                                                            }
                                                                        }
                                                                        else if (Rs >= 1387) {
                                                                            if (diffAFD < 1) {
                                                                                if (diffMVdiffVal < -2) {
                                                                                    return 1;
                                                                                }
                                                                                else if (diffMVdiffVal >= -2) {
                                                                                    return 4;
                                                                                }
                                                                            }
                                                                            else if (diffAFD >= 1) {
                                                                                return 8;
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (tcor >= 92) {
                                                            if (SC < 2348) {
                                                                if (posBalance < 51) {
                                                                    return 1;
                                                                }
                                                                else if (posBalance >= 51) {
                                                                    if (Rs < 1399) {
                                                                        return 1;
                                                                    }
                                                                    else if (Rs >= 1399) {
                                                                        return 4;
                                                                    }
                                                                }
                                                            }
                                                            else if (SC >= 2348) {
                                                                return 1;
                                                            }
                                                        }
                                                    }
                                                    else if (negBalance >= 45) {
                                                        if (posBalance < 52) {
                                                            if (SC < 2321) {
                                                                return 4;
                                                            }
                                                            else if (SC >= 2321) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (posBalance >= 52) {
                                                            if (SC < 2095) {
                                                                return 4;
                                                            }
                                                            else if (SC >= 2095) {
                                                                return 2;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (Cs >= 1011) {
                                            if (Rs < 1760) {
                                                if (Rs < 1337) {
                                                    if (ssDCval < 49) {
                                                        if (RsCsDiff < 1) {
                                                            return 4;
                                                        }
                                                        else if (RsCsDiff >= 1) {
                                                            return 2;
                                                        }
                                                    }
                                                    else if (ssDCval >= 49) {
                                                        if (MVDiff < 3) {
                                                            return 2;
                                                        }
                                                        else if (MVDiff >= 3) {
                                                            if (ssDCval < 53) {
                                                                if (diffRsCsdiff < -3) {
                                                                    return 4;
                                                                }
                                                                else if (diffRsCsdiff >= -3) {
                                                                    return 1;
                                                                }
                                                            }
                                                            else if (ssDCval >= 53) {
                                                                if (MVDiff < 17) {
                                                                    if (Cs < 1472) {
                                                                        if (SC < 2435) {
                                                                            return 4;
                                                                        }
                                                                        else if (SC >= 2435) {
                                                                            if (SC < 2485) {
                                                                                return 2;
                                                                            }
                                                                            else if (SC >= 2485) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (Cs >= 1472) {
                                                                        return 1;
                                                                    }
                                                                }
                                                                else if (MVDiff >= 17) {
                                                                    if (SC < 1770) {
                                                                        if (CsDiff < 4) {
                                                                            return 4;
                                                                        }
                                                                        else if (CsDiff >= 4) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                    else if (SC >= 1770) {
                                                                        return 8;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (Rs >= 1337) {
                                                    if (ssDCval < 39) {
                                                        return 2;
                                                    }
                                                    else if (ssDCval >= 39) {
                                                        if (TSCindex < 9) {
                                                            if (Rs < 1489) {
                                                                if (posBalance < 59) {
                                                                    if (Rs < 1434) {
                                                                        return 8;
                                                                    }
                                                                    else if (Rs >= 1434) {
                                                                        if (posBalance < 49) {
                                                                            return 1;
                                                                        }
                                                                        else if (posBalance >= 49) {
                                                                            if (mcTcor < 87) {
                                                                                return 1;
                                                                            }
                                                                            else if (mcTcor >= 87) {
                                                                                return 8;
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                                else if (posBalance >= 59) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (Rs >= 1489) {
                                                                if (negBalance < 44) {
                                                                    if (MVDiff < 12) {
                                                                        if (Rs < 1524) {
                                                                            return 1;
                                                                        }
                                                                        else if (Rs >= 1524) {
                                                                            if (AFD < 7) {
                                                                                return 4;
                                                                            }
                                                                            else if (AFD >= 7) {
                                                                                return 1;
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (MVDiff >= 12) {
                                                                        if (diffMVdiffVal < -2) {
                                                                            if (Cs < 1063) {
                                                                                return 2;
                                                                            }
                                                                            else if (Cs >= 1063) {
                                                                                return 8;
                                                                            }
                                                                        }
                                                                        else if (diffMVdiffVal >= -2) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                }
                                                                else if (negBalance >= 44) {
                                                                    if (Cs < 1288) {
                                                                        return 4;
                                                                    }
                                                                    else if (Cs >= 1288) {
                                                                        if (Rs < 1698) {
                                                                            return 8;
                                                                        }
                                                                        else if (Rs >= 1698) {
                                                                            if (Cs < 1473) {
                                                                                return 4;
                                                                            }
                                                                            else if (Cs >= 1473) {
                                                                                return 2;
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (TSCindex >= 9) {
                                                            if (Cs < 1217) {
                                                                return 4;
                                                            }
                                                            else if (Cs >= 1217) {
                                                                if (tcor < 90) {
                                                                    return 8;
                                                                }
                                                                else if (tcor >= 90) {
                                                                    return 2;
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                            else if (Rs >= 1760) {
                                                if (negBalance < 51) {
                                                    if (mcTcor < 91) {
                                                        return 4;
                                                    }
                                                    else if (mcTcor >= 91) {
                                                        if (SC < 3134) {
                                                            if (Cs < 1220) {
                                                                return 1;
                                                            }
                                                            else if (Cs >= 1220) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (SC >= 3134) {
                                                            if (Cs < 1538) {
                                                                if (MVDiff < 2) {
                                                                    return 1;
                                                                }
                                                                else if (MVDiff >= 2) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (Cs >= 1538) {
                                                                if (TSC < 5) {
                                                                    if (diffMVdiffVal < 2) {
                                                                        return 1;
                                                                    }
                                                                    else if (diffMVdiffVal >= 2) {
                                                                        if (Rs < 1852) {
                                                                            return 1;
                                                                        }
                                                                        else if (Rs >= 1852) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                }
                                                                else if (TSC >= 5) {
                                                                    if (MVDiff < 29) {
                                                                        if (SC < 3869) {
                                                                            return 4;
                                                                        }
                                                                        else if (SC >= 3869) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                    else if (MVDiff >= 29) {
                                                                        return 1;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (negBalance >= 51) {
                                                    if (MVDiff < 33) {
                                                        if (ssDCval < 88) {
                                                            return 4;
                                                        }
                                                        else if (ssDCval >= 88) {
                                                            return 2;
                                                        }
                                                    }
                                                    else if (MVDiff >= 33) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (Cs >= 1588) {
                                    if (posBalance < 56) {
                                        if (refDCval < 102) {
                                            if (SC < 4173) {
                                                if (mu_mv_mag_sq < 70) {
                                                    if (posBalance < 54) {
                                                        return 8;
                                                    }
                                                    else if (posBalance >= 54) {
                                                        if (Rs < 1961) {
                                                            if (CsDiff < 10) {
                                                                return 8;
                                                            }
                                                            else if (CsDiff >= 10) {
                                                                return 1;
                                                            }
                                                        }
                                                        else if (Rs >= 1961) {
                                                            return 1;
                                                        }
                                                    }
                                                }
                                                else if (mu_mv_mag_sq >= 70) {
                                                    if (tcor < 93) {
                                                        if (refDCval < 89) {
                                                            if (Cs < 1999) {
                                                                return 8;
                                                            }
                                                            else if (Cs >= 1999) {
                                                                return 1;
                                                            }
                                                        }
                                                        else if (refDCval >= 89) {
                                                            return 2;
                                                        }
                                                    }
                                                    else if (tcor >= 93) {
                                                        if (diffRsCsdiff < 7) {
                                                            if (Rs < 2252) {
                                                                return 4;
                                                            }
                                                            else if (Rs >= 2252) {
                                                                return 1;
                                                            }
                                                        }
                                                        else if (diffRsCsdiff >= 7) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (SC >= 4173) {
                                                if (TSC < 11) {
                                                    if (Rs < 2446) {
                                                        if (refDCval < 78) {
                                                            if (Cs < 2281) {
                                                                if (TSC < 6) {
                                                                    return 4;
                                                                }
                                                                else if (TSC >= 6) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (Cs >= 2281) {
                                                                if (RsCsDiff < 94) {
                                                                    if (Rs < 2413) {
                                                                        return 8;
                                                                    }
                                                                    else if (Rs >= 2413) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (RsCsDiff >= 94) {
                                                                    return 1;
                                                                }
                                                            }
                                                        }
                                                        else if (refDCval >= 78) {
                                                            if (posBalance < 54) {
                                                                if (ssDCval < 92) {
                                                                    if (refDCval < 87) {
                                                                        if (ssDCval < 83) {
                                                                            if (SC < 4845) {
                                                                                return 8;
                                                                            }
                                                                            else if (SC >= 4845) {
                                                                                return 1;
                                                                            }
                                                                        }
                                                                        else if (ssDCval >= 83) {
                                                                            if (SC < 4608) {
                                                                                return 1;
                                                                            }
                                                                            else if (SC >= 4608) {
                                                                                return 8;
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (refDCval >= 87) {
                                                                        return 8;
                                                                    }
                                                                }
                                                                else if (ssDCval >= 92) {
                                                                    return 1;
                                                                }
                                                            }
                                                            else if (posBalance >= 54) {
                                                                if (AFD < 13) {
                                                                    return 1;
                                                                }
                                                                else if (AFD >= 13) {
                                                                    return 2;
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (Rs >= 2446) {
                                                        if (TSC < 7) {
                                                            if (gchDC < 1) {
                                                                return 8;
                                                            }
                                                            else if (gchDC >= 1) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (TSC >= 7) {
                                                            if (negBalance < 53) {
                                                                if (diffMVdiffVal < -12) {
                                                                    return 8;
                                                                }
                                                                else if (diffMVdiffVal >= -12) {
                                                                    if (tcor < 91) {
                                                                        if (mcTcor < 95) {
                                                                            if (Rs < 2858) {
                                                                                if (AFD < 11) {
                                                                                    return 8;
                                                                                }
                                                                                else if (AFD >= 11) {
                                                                                    if (tcor < 90) {
                                                                                        return 1;
                                                                                    }
                                                                                    else if (tcor >= 90) {
                                                                                        if (CsDiff < 8) {
                                                                                            return 1;
                                                                                        }
                                                                                        else if (CsDiff >= 8) {
                                                                                            return 8;
                                                                                        }
                                                                                    }
                                                                                }
                                                                            }
                                                                            else if (Rs >= 2858) {
                                                                                if (CsDiff < 9) {
                                                                                    if (negBalance < 47) {
                                                                                        return 8;
                                                                                    }
                                                                                    else if (negBalance >= 47) {
                                                                                        if (RsDiff < 7) {
                                                                                            return 4;
                                                                                        }
                                                                                        else if (RsDiff >= 7) {
                                                                                            return 1;
                                                                                        }
                                                                                    }
                                                                                }
                                                                                else if (CsDiff >= 9) {
                                                                                    return 1;
                                                                                }
                                                                            }
                                                                        }
                                                                        else if (mcTcor >= 95) {
                                                                            return 2;
                                                                        }
                                                                    }
                                                                    else if (tcor >= 91) {
                                                                        return 1;
                                                                    }
                                                                }
                                                            }
                                                            else if (negBalance >= 53) {
                                                                if (SC < 4875) {
                                                                    if (AFD < 12) {
                                                                        return 4;
                                                                    }
                                                                    else if (AFD >= 12) {
                                                                        return 1;
                                                                    }
                                                                }
                                                                else if (SC >= 4875) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (TSC >= 11) {
                                                    if (ssDCval < 71) {
                                                        return 8;
                                                    }
                                                    else if (ssDCval >= 71) {
                                                        if (Cs < 2177) {
                                                            return 8;
                                                        }
                                                        else if (Cs >= 2177) {
                                                            if (Cs < 2655) {
                                                                if (refDCval < 99) {
                                                                    if (posBalance < 48) {
                                                                        if (Rs < 2338) {
                                                                            return 4;
                                                                        }
                                                                        else if (Rs >= 2338) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                    else if (posBalance >= 48) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (refDCval >= 99) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (Cs >= 2655) {
                                                                return 1;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (refDCval >= 102) {
                                            if (mcTcor < 94) {
                                                return 4;
                                            }
                                            else if (mcTcor >= 94) {
                                                if (negBalance < 39) {
                                                    return 2;
                                                }
                                                else if (negBalance >= 39) {
                                                    if (SC < 5218) {
                                                        return 2;
                                                    }
                                                    else if (SC >= 5218) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (posBalance >= 56) {
                                        if (MVDiff < 8) {
                                            if (mcTcor < 91) {
                                                if (negBalance < 51) {
                                                    if (TSC < 10) {
                                                        return 8;
                                                    }
                                                    else if (TSC >= 10) {
                                                        return 4;
                                                    }
                                                }
                                                else if (negBalance >= 51) {
                                                    if (Rs < 2011) {
                                                        if (Cs < 1672) {
                                                            return 2;
                                                        }
                                                        else if (Cs >= 1672) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (Rs >= 2011) {
                                                        return 2;
                                                    }
                                                }
                                            }
                                            else if (mcTcor >= 91) {
                                                if (Rs < 2404) {
                                                    if (negBalance < 45) {
                                                        if (MVDiff < 2) {
                                                            return 4;
                                                        }
                                                        else if (MVDiff >= 2) {
                                                            if (Cs < 2077) {
                                                                return 8;
                                                            }
                                                            else if (Cs >= 2077) {
                                                                return 2;
                                                            }
                                                        }
                                                    }
                                                    else if (negBalance >= 45) {
                                                        if (SC < 3223) {
                                                            return 4;
                                                        }
                                                        else if (SC >= 3223) {
                                                            return 2;
                                                        }
                                                    }
                                                }
                                                else if (Rs >= 2404) {
                                                    if (SC < 5042) {
                                                        if (ssDCval < 86) {
                                                            if (Cs < 1697) {
                                                                return 1;
                                                            }
                                                            else if (Cs >= 1697) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (ssDCval >= 86) {
                                                            if (AFD < 14) {
                                                                return 1;
                                                            }
                                                            else if (AFD >= 14) {
                                                                return 2;
                                                            }
                                                        }
                                                    }
                                                    else if (SC >= 5042) {
                                                        if (refDCval < 88) {
                                                            if (ssDCval < 86) {
                                                                if (ssDCval < 85) {
                                                                    return 8;
                                                                }
                                                                else if (ssDCval >= 85) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (ssDCval >= 86) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (refDCval >= 88) {
                                                            if (Cs < 2492) {
                                                                return 4;
                                                            }
                                                            else if (Cs >= 2492) {
                                                                return 1;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (MVDiff >= 8) {
                                            if (mcTcor < 88) {
                                                if (TSC < 9) {
                                                    if (Rs < 2074) {
                                                        if (SC < 3165) {
                                                            return 1;
                                                        }
                                                        else if (SC >= 3165) {
                                                            if (RsDiff < 6) {
                                                                return 8;
                                                            }
                                                            else if (RsDiff >= 6) {
                                                                if (mcTcor < 83) {
                                                                    return 2;
                                                                }
                                                                else if (mcTcor >= 83) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (Rs >= 2074) {
                                                        return 2;
                                                    }
                                                }
                                                else if (TSC >= 9) {
                                                    if (ssDCval < 75) {
                                                        return 8;
                                                    }
                                                    else if (ssDCval >= 75) {
                                                        if (TSC < 10) {
                                                            return 8;
                                                        }
                                                        else if (TSC >= 10) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (mcTcor >= 88) {
                                                if (Rs < 1852) {
                                                    if (RsCsDiff < 63) {
                                                        return 8;
                                                    }
                                                    else if (RsCsDiff >= 63) {
                                                        return 4;
                                                    }
                                                }
                                                else if (Rs >= 1852) {
                                                    if (Cs < 2307) {
                                                        if (AFD < 14) {
                                                            if (posBalance < 82) {
                                                                if (mu_mv_mag_sq < 79) {
                                                                    if (Rs < 2547) {
                                                                        if (mcTcor < 94) {
                                                                            return 1;
                                                                        }
                                                                        else if (mcTcor >= 94) {
                                                                            return 2;
                                                                        }
                                                                    }
                                                                    else if (Rs >= 2547) {
                                                                        return 1;
                                                                    }
                                                                }
                                                                else if (mu_mv_mag_sq >= 79) {
                                                                    return 2;
                                                                }
                                                            }
                                                            else if (posBalance >= 82) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (AFD >= 14) {
                                                            if (ssDCval < 83) {
                                                                if (SC < 4355) {
                                                                    return 8;
                                                                }
                                                                else if (SC >= 4355) {
                                                                    return 1;
                                                                }
                                                            }
                                                            else if (ssDCval >= 83) {
                                                                if (negBalance < 46) {
                                                                    if (Cs < 2175) {
                                                                        return 4;
                                                                    }
                                                                    else if (Cs >= 2175) {
                                                                        return 1;
                                                                    }
                                                                }
                                                                else if (negBalance >= 46) {
                                                                    if (tcor < 91) {
                                                                        return 2;
                                                                    }
                                                                    else if (tcor >= 91) {
                                                                        return 1;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (Cs >= 2307) {
                                                        if (RsCsDiff < 192) {
                                                            if (mu_mv_mag_sq < 43) {
                                                                return 2;
                                                            }
                                                            else if (mu_mv_mag_sq >= 43) {
                                                                if (mu_mv_mag_sq < 69) {
                                                                    if (Rs < 3012) {
                                                                        if (AFD < 13) {
                                                                            if (Cs < 2516) {
                                                                                return 1;
                                                                            }
                                                                            else if (Cs >= 2516) {
                                                                                return 2;
                                                                            }
                                                                        }
                                                                        else if (AFD >= 13) {
                                                                            if (Cs < 2443) {
                                                                                if (refDCval < 92) {
                                                                                    if (Cs < 2314) {
                                                                                        return 4;
                                                                                    }
                                                                                    else if (Cs >= 2314) {
                                                                                        if (SC < 5320) {
                                                                                            return 1;
                                                                                        }
                                                                                        else if (SC >= 5320) {
                                                                                            return 4;
                                                                                        }
                                                                                    }
                                                                                }
                                                                                else if (refDCval >= 92) {
                                                                                    return 4;
                                                                                }
                                                                            }
                                                                            else if (Cs >= 2443) {
                                                                                return 1;
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (Rs >= 3012) {
                                                                        return 1;
                                                                    }
                                                                }
                                                                else if (mu_mv_mag_sq >= 69) {
                                                                    if (TSC < 11) {
                                                                        if (ssDCval < 90) {
                                                                            return 8;
                                                                        }
                                                                        else if (ssDCval >= 90) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                    else if (TSC >= 11) {
                                                                        return 4;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (RsCsDiff >= 192) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                            else if (SC >= 6283) {
                                if (MVDiff < 9) {
                                    if (refDCval < 54) {
                                        if (AFD < 9) {
                                            if (Rs < 3518) {
                                                return 8;
                                            }
                                            else if (Rs >= 3518) {
                                                return 4;
                                            }
                                        }
                                        else if (AFD >= 9) {
                                            return 2;
                                        }
                                    }
                                    else if (refDCval >= 54) {
                                        if (TSC < 7) {
                                            if (refDCval < 55) {
                                                if (ssDCval < 54) {
                                                    if (Cs < 3411) {
                                                        return 8;
                                                    }
                                                    else if (Cs >= 3411) {
                                                        return 4;
                                                    }
                                                }
                                                else if (ssDCval >= 54) {
                                                    return 8;
                                                }
                                            }
                                            else if (refDCval >= 55) {
                                                return 8;
                                            }
                                        }
                                        else if (TSC >= 7) {
                                            if (ssDCval < 71) {
                                                return 4;
                                            }
                                            else if (ssDCval >= 71) {
                                                if (refDCval < 95) {
                                                    if (Rs < 7284) {
                                                        return 8;
                                                    }
                                                    else if (Rs >= 7284) {
                                                        return 4;
                                                    }
                                                }
                                                else if (refDCval >= 95) {
                                                    return 4;
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (MVDiff >= 9) {
                                    if (RsDiff < 5) {
                                        if (negBalance < 50) {
                                            if (MVDiff < 18) {
                                                if (TSC < 7) {
                                                    if (refDCval < 54) {
                                                        return 4;
                                                    }
                                                    else if (refDCval >= 54) {
                                                        if (refDCval < 62) {
                                                            if (ssDCval < 55) {
                                                                return 4;
                                                            }
                                                            else if (ssDCval >= 55) {
                                                                if (TSCindex < 7) {
                                                                    if (Cs < 3458) {
                                                                        return 8;
                                                                    }
                                                                    else if (Cs >= 3458) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (TSCindex >= 7) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                        else if (refDCval >= 62) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                                else if (TSC >= 7) {
                                                    if (mu_mv_mag_sq < 43) {
                                                        if (MVDiff < 16) {
                                                            return 2;
                                                        }
                                                        else if (MVDiff >= 16) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (mu_mv_mag_sq >= 43) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                            else if (MVDiff >= 18) {
                                                if (AFD < 8) {
                                                    return 4;
                                                }
                                                else if (AFD >= 8) {
                                                    return 8;
                                                }
                                            }
                                        }
                                        else if (negBalance >= 50) {
                                            if (refDCval < 54) {
                                                if (MVDiff < 26) {
                                                    if (SC < 7364) {
                                                        return 4;
                                                    }
                                                    else if (SC >= 7364) {
                                                        if (Rs < 3766) {
                                                            return 2;
                                                        }
                                                        else if (Rs >= 3766) {
                                                            if (SC < 7800) {
                                                                return 4;
                                                            }
                                                            else if (SC >= 7800) {
                                                                return 2;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (MVDiff >= 26) {
                                                    return 2;
                                                }
                                            }
                                            else if (refDCval >= 54) {
                                                if (posBalance < 57) {
                                                    if (MVDiff < 27) {
                                                        return 2;
                                                    }
                                                    else if (MVDiff >= 27) {
                                                        return 8;
                                                    }
                                                }
                                                else if (posBalance >= 57) {
                                                    return 4;
                                                }
                                            }
                                        }
                                    }
                                    else if (RsDiff >= 5) {
                                        if (TSC < 12) {
                                            if (SC < 6582) {
                                                return 8;
                                            }
                                            else if (SC >= 6582) {
                                                if (ssDCval < 53) {
                                                    if (AFD < 13) {
                                                        return 8;
                                                    }
                                                    else if (AFD >= 13) {
                                                        return 4;
                                                    }
                                                }
                                                else if (ssDCval >= 53) {
                                                    if (Cs < 4372) {
                                                        if (TSCindex < 8) {
                                                            if (refDCval < 63) {
                                                                return 8;
                                                            }
                                                            else if (refDCval >= 63) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (TSCindex >= 8) {
                                                            if (MVDiff < 43) {
                                                                if (mu_mv_mag_sq < 40) {
                                                                    if (diffRsCsdiff < 20) {
                                                                        return 4;
                                                                    }
                                                                    else if (diffRsCsdiff >= 20) {
                                                                        return 8;
                                                                    }
                                                                }
                                                                else if (mu_mv_mag_sq >= 40) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (MVDiff >= 43) {
                                                                if (Rs < 4069) {
                                                                    return 8;
                                                                }
                                                                else if (Rs >= 4069) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (Cs >= 4372) {
                                                        if (mu_mv_mag_sq < 48) {
                                                            if (refDCval < 68) {
                                                                return 4;
                                                            }
                                                            else if (refDCval >= 68) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (mu_mv_mag_sq >= 48) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (TSC >= 12) {
                                            if (mcTcor < 84) {
                                                if (Cs < 3153) {
                                                    return 4;
                                                }
                                                else if (Cs >= 3153) {
                                                    return 8;
                                                }
                                            }
                                            else if (mcTcor >= 84) {
                                                if (gchDC < 2) {
                                                    if (Rs < 4852) {
                                                        return 8;
                                                    }
                                                    else if (Rs >= 4852) {
                                                        return 1;
                                                    }
                                                }
                                                else if (gchDC >= 2) {
                                                    return 4;
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                        }
                        else if (ssDCval >= 104) {
                            if (tcor < 81) {
                                if (refDCval < 123) {
                                    if (AFD < 14) {
                                        if (posBalance < 46) {
                                            if (Rs < 4199) {
                                                return 8;
                                            }
                                            else if (Rs >= 4199) {
                                                if (SC < 8175) {
                                                    return 1;
                                                }
                                                else if (SC >= 8175) {
                                                    return 2;
                                                }
                                            }
                                        }
                                        else if (posBalance >= 46) {
                                            if (posBalance < 60) {
                                                if (diffMVdiffVal < 5) {
                                                    return 4;
                                                }
                                                else if (diffMVdiffVal >= 5) {
                                                    if (Rs < 4701) {
                                                        return 4;
                                                    }
                                                    else if (Rs >= 4701) {
                                                        return 2;
                                                    }
                                                }
                                            }
                                            else if (posBalance >= 60) {
                                                if (Rs < 2069) {
                                                    return 4;
                                                }
                                                else if (Rs >= 2069) {
                                                    if (RsDiff < 10) {
                                                        if (ssDCval < 109) {
                                                            return 4;
                                                        }
                                                        else if (ssDCval >= 109) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (RsDiff >= 10) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (AFD >= 14) {
                                        if (negBalance < 43) {
                                            if (MVDiff < 115) {
                                                if (AFD < 17) {
                                                    return 1;
                                                }
                                                else if (AFD >= 17) {
                                                    if (RsDiff < 24) {
                                                        return 8;
                                                    }
                                                    else if (RsDiff >= 24) {
                                                        return 1;
                                                    }
                                                }
                                            }
                                            else if (MVDiff >= 115) {
                                                if (RsCsDiff < 253) {
                                                    return 4;
                                                }
                                                else if (RsCsDiff >= 253) {
                                                    return 2;
                                                }
                                            }
                                        }
                                        else if (negBalance >= 43) {
                                            if (ssDCval < 117) {
                                                if (ssDCval < 107) {
                                                    if (Rs < 5416) {
                                                        if (Cs < 3147) {
                                                            return 2;
                                                        }
                                                        else if (Cs >= 3147) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (Rs >= 5416) {
                                                        return 4;
                                                    }
                                                }
                                                else if (ssDCval >= 107) {
                                                    return 2;
                                                }
                                            }
                                            else if (ssDCval >= 117) {
                                                if (diffMVdiffVal < 6) {
                                                    return 4;
                                                }
                                                else if (diffMVdiffVal >= 6) {
                                                    if (RsCsDiff < 375) {
                                                        return 1;
                                                    }
                                                    else if (RsCsDiff >= 375) {
                                                        if (Rs < 3528) {
                                                            return 2;
                                                        }
                                                        else if (Rs >= 3528) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (refDCval >= 123) {
                                    if (mcTcor < 89) {
                                        if (ssDCval < 140) {
                                            if (gchDC < 2) {
                                                if (Cs < 796) {
                                                    return 4;
                                                }
                                                else if (Cs >= 796) {
                                                    if (diffMVdiffVal < -53) {
                                                        return 4;
                                                    }
                                                    else if (diffMVdiffVal >= -53) {
                                                        if (AFD < 11) {
                                                            if (Rs < 2483) {
                                                                return 1;
                                                            }
                                                            else if (Rs >= 2483) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (AFD >= 11) {
                                                            if (diffRsCsdiff < -171) {
                                                                if (Cs < 3382) {
                                                                    return 1;
                                                                }
                                                                else if (Cs >= 3382) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (diffRsCsdiff >= -171) {
                                                                return 1;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                            else if (gchDC >= 2) {
                                                return 8;
                                            }
                                        }
                                        else if (ssDCval >= 140) {
                                            if (AFD < 26) {
                                                return 8;
                                            }
                                            else if (AFD >= 26) {
                                                if (Rs < 1343) {
                                                    return 1;
                                                }
                                                else if (Rs >= 1343) {
                                                    if (Rs < 2862) {
                                                        return 4;
                                                    }
                                                    else if (Rs >= 2862) {
                                                        return 2;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (mcTcor >= 89) {
                                        if (posBalance < 57) {
                                            if (Cs < 2382) {
                                                if (mcTcor < 96) {
                                                    if (ssDCval < 141) {
                                                        return 8;
                                                    }
                                                    else if (ssDCval >= 141) {
                                                        if (refDCval < 169) {
                                                            return 4;
                                                        }
                                                        else if (refDCval >= 169) {
                                                            if (Rs < 2208) {
                                                                return 4;
                                                            }
                                                            else if (Rs >= 2208) {
                                                                return 2;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (mcTcor >= 96) {
                                                    return 2;
                                                }
                                            }
                                            else if (Cs >= 2382) {
                                                return 8;
                                            }
                                        }
                                        else if (posBalance >= 57) {
                                            return 2;
                                        }
                                    }
                                }
                            }
                            else if (tcor >= 81) {
                                if (TSC < 11) {
                                    if (posBalance < 53) {
                                        if (RsCsDiff < 83) {
                                            if (Cs < 362) {
                                                if (SC < 878) {
                                                    if (mu_mv_mag_sq < 62) {
                                                        return 1;
                                                    }
                                                    else if (mu_mv_mag_sq >= 62) {
                                                        return 8;
                                                    }
                                                }
                                                else if (SC >= 878) {
                                                    return 8;
                                                }
                                            }
                                            else if (Cs >= 362) {
                                                if (refDCval < 107) {
                                                    if (mu_mv_mag_sq < 43) {
                                                        if (negBalance < 38) {
                                                            return 8;
                                                        }
                                                        else if (negBalance >= 38) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (mu_mv_mag_sq >= 43) {
                                                        return 8;
                                                    }
                                                }
                                                else if (refDCval >= 107) {
                                                    if (mcTcor < 95) {
                                                        if (Rs < 2511) {
                                                            if (Cs < 1507) {
                                                                if (diffTSC < -1) {
                                                                    return 8;
                                                                }
                                                                else if (diffTSC >= -1) {
                                                                    if (Rs < 805) {
                                                                        return 4;
                                                                    }
                                                                    else if (Rs >= 805) {
                                                                        return 1;
                                                                    }
                                                                }
                                                            }
                                                            else if (Cs >= 1507) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (Rs >= 2511) {
                                                            if (Rs < 4286) {
                                                                return 4;
                                                            }
                                                            else if (Rs >= 4286) {
                                                                return 2;
                                                            }
                                                        }
                                                    }
                                                    else if (mcTcor >= 95) {
                                                        if (mcTcor < 99) {
                                                            if (Cs < 2365) {
                                                                if (CsDiff < 5) {
                                                                    if (Cs < 1688) {
                                                                        if (diffTSC < 1) {
                                                                            if (diffTSC < -1) {
                                                                                if (AFD < 6) {
                                                                                    return 8;
                                                                                }
                                                                                else if (AFD >= 6) {
                                                                                    return 4;
                                                                                }
                                                                            }
                                                                            else if (diffTSC >= -1) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                        else if (diffTSC >= 1) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                    else if (Cs >= 1688) {
                                                                        if (AFD < 8) {
                                                                            return 8;
                                                                        }
                                                                        else if (AFD >= 8) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                }
                                                                else if (CsDiff >= 5) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (Cs >= 2365) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (mcTcor >= 99) {
                                                            if (AFD < 3) {
                                                                if (Rs < 568) {
                                                                    return 2;
                                                                }
                                                                else if (Rs >= 568) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (AFD >= 3) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (RsCsDiff >= 83) {
                                            if (Cs < 2528) {
                                                if (Rs < 4273) {
                                                    if (Cs < 2111) {
                                                        if (RsDiff < 9) {
                                                            if (Rs < 3624) {
                                                                if (CsDiff < 7) {
                                                                    return 2;
                                                                }
                                                                else if (CsDiff >= 7) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (Rs >= 3624) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (RsDiff >= 9) {
                                                            if (TSC < 8) {
                                                                if (Rs < 4157) {
                                                                    return 8;
                                                                }
                                                                else if (Rs >= 4157) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (TSC >= 8) {
                                                                return 2;
                                                            }
                                                        }
                                                    }
                                                    else if (Cs >= 2111) {
                                                        if (mcTcor < 95) {
                                                            return 4;
                                                        }
                                                        else if (mcTcor >= 95) {
                                                            if (diffRsCsdiff < -58) {
                                                                return 4;
                                                            }
                                                            else if (diffRsCsdiff >= -58) {
                                                                if (diffMVdiffVal < 6) {
                                                                    if (ssDCval < 119) {
                                                                        if (diffMVdiffVal < -9) {
                                                                            if (diffRsCsdiff < 14) {
                                                                                return 2;
                                                                            }
                                                                            else if (diffRsCsdiff >= 14) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                        else if (diffMVdiffVal >= -9) {
                                                                            return 2;
                                                                        }
                                                                    }
                                                                    else if (ssDCval >= 119) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (diffMVdiffVal >= 6) {
                                                                    if (SC < 5888) {
                                                                        if (MVDiff < 32) {
                                                                            return 4;
                                                                        }
                                                                        else if (MVDiff >= 32) {
                                                                            return 2;
                                                                        }
                                                                    }
                                                                    else if (SC >= 5888) {
                                                                        return 2;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (Rs >= 4273) {
                                                    if (SC < 7217) {
                                                        if (ssDCval < 109) {
                                                            return 8;
                                                        }
                                                        else if (ssDCval >= 109) {
                                                            if (Rs < 4512) {
                                                                return 4;
                                                            }
                                                            else if (Rs >= 4512) {
                                                                if (RsDiff < 11) {
                                                                    if (TSC < 7) {
                                                                        return 4;
                                                                    }
                                                                    else if (TSC >= 7) {
                                                                        if (SC < 6938) {
                                                                            return 8;
                                                                        }
                                                                        else if (SC >= 6938) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                }
                                                                else if (RsDiff >= 11) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (SC >= 7217) {
                                                        if (SC < 7243) {
                                                            return 2;
                                                        }
                                                        else if (SC >= 7243) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (Cs >= 2528) {
                                                if (SC < 7263) {
                                                    if (RsDiff < 10) {
                                                        if (SC < 5763) {
                                                            return 4;
                                                        }
                                                        else if (SC >= 5763) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (RsDiff >= 10) {
                                                        return 2;
                                                    }
                                                }
                                                else if (SC >= 7263) {
                                                    if (mu_mv_mag_sq < 73) {
                                                        if (Rs < 4535) {
                                                            return 8;
                                                        }
                                                        else if (Rs >= 4535) {
                                                            if (SC < 7396) {
                                                                return 2;
                                                            }
                                                            else if (SC >= 7396) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                    else if (mu_mv_mag_sq >= 73) {
                                                        if (Rs < 4928) {
                                                            return 2;
                                                        }
                                                        else if (Rs >= 4928) {
                                                            if (posBalance < 47) {
                                                                if (Cs < 3206) {
                                                                    return 8;
                                                                }
                                                                else if (Cs >= 3206) {
                                                                    return 2;
                                                                }
                                                            }
                                                            else if (posBalance >= 47) {
                                                                return 2;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (posBalance >= 53) {
                                        if (AFD < 12) {
                                            if (negBalance < 50) {
                                                if (Rs < 3715) {
                                                    if (posBalance < 87) {
                                                        if (AFD < 11) {
                                                            return 8;
                                                        }
                                                        else if (AFD >= 11) {
                                                            if (Cs < 1629) {
                                                                return 2;
                                                            }
                                                            else if (Cs >= 1629) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                    else if (posBalance >= 87) {
                                                        if (RsDiff < 7) {
                                                            return 4;
                                                        }
                                                        else if (RsDiff >= 7) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                                else if (Rs >= 3715) {
                                                    if (Cs < 3340) {
                                                        return 4;
                                                    }
                                                    else if (Cs >= 3340) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                            else if (negBalance >= 50) {
                                                if (Cs < 382) {
                                                    return 4;
                                                }
                                                else if (Cs >= 382) {
                                                    if (RsDiff < 4) {
                                                        return 8;
                                                    }
                                                    else if (RsDiff >= 4) {
                                                        if (AFD < 11) {
                                                            if (negBalance < 51) {
                                                                if (ssDCval < 119) {
                                                                    return 8;
                                                                }
                                                                else if (ssDCval >= 119) {
                                                                    if (mu_mv_mag_sq < 59) {
                                                                        return 8;
                                                                    }
                                                                    else if (mu_mv_mag_sq >= 59) {
                                                                        return 4;
                                                                    }
                                                                }
                                                            }
                                                            else if (negBalance >= 51) {
                                                                if (MVDiff < 42) {
                                                                    if (Rs < 2692) {
                                                                        if (Scindex < 5) {
                                                                            return 4;
                                                                        }
                                                                        else if (Scindex >= 5) {
                                                                            if (MVDiff < 6) {
                                                                                return 8;
                                                                            }
                                                                            else if (MVDiff >= 6) {
                                                                                if (Rs < 2617) {
                                                                                    return 4;
                                                                                }
                                                                                else if (Rs >= 2617) {
                                                                                    return 8;
                                                                                }
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (Rs >= 2692) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (MVDiff >= 42) {
                                                                    return 2;
                                                                }
                                                            }
                                                        }
                                                        else if (AFD >= 11) {
                                                            if (Cs < 1774) {
                                                                return 8;
                                                            }
                                                            else if (Cs >= 1774) {
                                                                if (Cs < 2128) {
                                                                    return 4;
                                                                }
                                                                else if (Cs >= 2128) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (AFD >= 12) {
                                            if (tcor < 90) {
                                                if (diffMVdiffVal < 1) {
                                                    if (diffAFD < -1) {
                                                        if (diffMVdiffVal < -2) {
                                                            return 4;
                                                        }
                                                        else if (diffMVdiffVal >= -2) {
                                                            return 1;
                                                        }
                                                    }
                                                    else if (diffAFD >= -1) {
                                                        return 2;
                                                    }
                                                }
                                                else if (diffMVdiffVal >= 1) {
                                                    return 2;
                                                }
                                            }
                                            else if (tcor >= 90) {
                                                if (mcTcor < 96) {
                                                    return 8;
                                                }
                                                else if (mcTcor >= 96) {
                                                    if (diffRsCsdiff < -55) {
                                                        return 4;
                                                    }
                                                    else if (diffRsCsdiff >= -55) {
                                                        if (Cs < 2710) {
                                                            return 2;
                                                        }
                                                        else if (Cs >= 2710) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (TSC >= 11) {
                                    if (RsCsDiff < 172) {
                                        if (SC < 5913) {
                                            return 1;
                                        }
                                        else if (SC >= 5913) {
                                            if (ssDCval < 117) {
                                                return 1;
                                            }
                                            else if (ssDCval >= 117) {
                                                return 8;
                                            }
                                        }
                                    }
                                    else if (RsCsDiff >= 172) {
                                        if (refDCval < 119) {
                                            if (negBalance < 47) {
                                                if (CsDiff < 11) {
                                                    return 2;
                                                }
                                                else if (CsDiff >= 11) {
                                                    return 8;
                                                }
                                            }
                                            else if (negBalance >= 47) {
                                                if (refDCval < 108) {
                                                    return 4;
                                                }
                                                else if (refDCval >= 108) {
                                                    if (Rs < 3225) {
                                                        return 4;
                                                    }
                                                    else if (Rs >= 3225) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                        }
                                        else if (refDCval >= 119) {
                                            if (RsCsDiff < 741) {
                                                return 8;
                                            }
                                            else if (RsCsDiff >= 741) {
                                                return 4;
                                            }
                                        }
                                    }
                                }
                            }
                        }
                    }
                }
                else if (CsDiff >= 17) {
                    if (mu_mv_mag_sq < 44) {
                        if (Rs < 6771) {
                            if (TSC < 16) {
                                if (refDCval < 143) {
                                    if (TSC < 9) {
                                        return 4;
                                    }
                                    else if (TSC >= 9) {
                                        if (posBalance < 50) {
                                            return 8;
                                        }
                                        else if (posBalance >= 50) {
                                            if (gchDC < 3) {
                                                if (posBalance < 62) {
                                                    if (diffRsCsdiff < -98) {
                                                        return 8;
                                                    }
                                                    else if (diffRsCsdiff >= -98) {
                                                        if (posBalance < 56) {
                                                            if (refDCval < 82) {
                                                                if (Cs < 2861) {
                                                                    return 8;
                                                                }
                                                                else if (Cs >= 2861) {
                                                                    return 1;
                                                                }
                                                            }
                                                            else if (refDCval >= 82) {
                                                                if (TSC < 13) {
                                                                    return 1;
                                                                }
                                                                else if (TSC >= 13) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                        else if (posBalance >= 56) {
                                                            return 1;
                                                        }
                                                    }
                                                }
                                                else if (posBalance >= 62) {
                                                    return 8;
                                                }
                                            }
                                            else if (gchDC >= 3) {
                                                if (Cs < 4866) {
                                                    return 2;
                                                }
                                                else if (Cs >= 4866) {
                                                    return 1;
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (refDCval >= 143) {
                                    if (diffAFD < -2) {
                                        if (Cs < 3517) {
                                            return 4;
                                        }
                                        else if (Cs >= 3517) {
                                            return 8;
                                        }
                                    }
                                    else if (diffAFD >= -2) {
                                        return 2;
                                    }
                                }
                            }
                            else if (TSC >= 16) {
                                if (RsDiff < 17) {
                                    if (tcor < 47) {
                                        if (SC < 4004) {
                                            return 2;
                                        }
                                        else if (SC >= 4004) {
                                            return 4;
                                        }
                                    }
                                    else if (tcor >= 47) {
                                        if (gchDC < 1) {
                                            return 1;
                                        }
                                        else if (gchDC >= 1) {
                                            if (Cs < 4569) {
                                                return 4;
                                            }
                                            else if (Cs >= 4569) {
                                                return 8;
                                            }
                                        }
                                    }
                                }
                                else if (RsDiff >= 17) {
                                    if (AFD < 31) {
                                        if (Cs < 4611) {
                                            if (Rs < 5298) {
                                                return 8;
                                            }
                                            else if (Rs >= 5298) {
                                                if (ssDCval < 124) {
                                                    return 8;
                                                }
                                                else if (ssDCval >= 124) {
                                                    return 4;
                                                }
                                            }
                                        }
                                        else if (Cs >= 4611) {
                                            return 8;
                                        }
                                    }
                                    else if (AFD >= 31) {
                                        if (TSC < 20) {
                                            if (ssDCval < 109) {
                                                return 2;
                                            }
                                            else if (ssDCval >= 109) {
                                                if (Rs < 6352) {
                                                    return 8;
                                                }
                                                else if (Rs >= 6352) {
                                                    return 2;
                                                }
                                            }
                                        }
                                        else if (TSC >= 20) {
                                            return 8;
                                        }
                                    }
                                }
                            }
                        }
                        else if (Rs >= 6771) {
                            if (refDCval < 76) {
                                return 4;
                            }
                            else if (refDCval >= 76) {
                                if (MVDiff < 21) {
                                    if (mu_mv_mag_sq < 14) {
                                        if (posBalance < 60) {
                                            if (mcTcor < 91) {
                                                if (Cs < 4716) {
                                                    return 4;
                                                }
                                                else if (Cs >= 4716) {
                                                    if (TSC < 28) {
                                                        if (MVDiff < 1) {
                                                            return 4;
                                                        }
                                                        else if (MVDiff >= 1) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (TSC >= 28) {
                                                        if (RsDiff < 51) {
                                                            return 4;
                                                        }
                                                        else if (RsDiff >= 51) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (mcTcor >= 91) {
                                                return 4;
                                            }
                                        }
                                        else if (posBalance >= 60) {
                                            if (SC < 12967) {
                                                return 2;
                                            }
                                            else if (SC >= 12967) {
                                                if (SC < 19620) {
                                                    return 8;
                                                }
                                                else if (SC >= 19620) {
                                                    if (Rs < 14313) {
                                                        return 2;
                                                    }
                                                    else if (Rs >= 14313) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (mu_mv_mag_sq >= 14) {
                                        if (Rs < 15414) {
                                            if (CsDiff < 43) {
                                                if (TSC < 20) {
                                                    if (Rs < 8200) {
                                                        return 4;
                                                    }
                                                    else if (Rs >= 8200) {
                                                        if (Cs < 4646) {
                                                            return 4;
                                                        }
                                                        else if (Cs >= 4646) {
                                                            return 2;
                                                        }
                                                    }
                                                }
                                                else if (TSC >= 20) {
                                                    if (RsDiff < 32) {
                                                        if (Rs < 8707) {
                                                            return 4;
                                                        }
                                                        else if (Rs >= 8707) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (RsDiff >= 32) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                            else if (CsDiff >= 43) {
                                                return 2;
                                            }
                                        }
                                        else if (Rs >= 15414) {
                                            return 8;
                                        }
                                    }
                                }
                                else if (MVDiff >= 21) {
                                    if (mcTcor < 85) {
                                        if (negBalance < 58) {
                                            return 8;
                                        }
                                        else if (negBalance >= 58) {
                                            if (tcor < 73) {
                                                if (RsDiff < 32) {
                                                    return 2;
                                                }
                                                else if (RsDiff >= 32) {
                                                    if (Rs < 10133) {
                                                        return 8;
                                                    }
                                                    else if (Rs >= 10133) {
                                                        return 2;
                                                    }
                                                }
                                            }
                                            else if (tcor >= 73) {
                                                return 8;
                                            }
                                        }
                                    }
                                    else if (mcTcor >= 85) {
                                        if (Cs < 6595) {
                                            return 2;
                                        }
                                        else if (Cs >= 6595) {
                                            return 8;
                                        }
                                    }
                                }
                            }
                        }
                    }
                    else if (mu_mv_mag_sq >= 44) {
                        if (AFD < 19) {
                            if (tcor < 69) {
                                return 1;
                            }
                            else if (tcor >= 69) {
                                return 8;
                            }
                        }
                        else if (AFD >= 19) {
                            if (SC < 12626) {
                                if (RsDiff < 16) {
                                    if (ssDCval < 113) {
                                        return 8;
                                    }
                                    else if (ssDCval >= 113) {
                                        if (Rs < 2507) {
                                            return 4;
                                        }
                                        else if (Rs >= 2507) {
                                            return 1;
                                        }
                                    }
                                }
                                else if (RsDiff >= 16) {
                                    return 8;
                                }
                            }
                            else if (SC >= 12626) {
                                if (tcor < 74) {
                                    if (SC < 13521) {
                                        return 8;
                                    }
                                    else if (SC >= 13521) {
                                        return 4;
                                    }
                                }
                                else if (tcor >= 74) {
                                    if (mu_mv_mag_sq < 63) {
                                        if (ssDCval < 68) {
                                            return 4;
                                        }
                                        else if (ssDCval >= 68) {
                                            if (negBalance < 54) {
                                                return 8;
                                            }
                                            else if (negBalance >= 54) {
                                                if (mcTcor < 83) {
                                                    return 4;
                                                }
                                                else if (mcTcor >= 83) {
                                                    if (Rs < 7239) {
                                                        return 4;
                                                    }
                                                    else if (Rs >= 7239) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (mu_mv_mag_sq >= 63) {
                                        return 8;
                                    }
                                }
                            }
                        }
                    }
                }
            }
        }
        else if (mu_mv_mag_sq >= 84) {
            if (negBalance < 32) {
                if (Scindex < 4) {
                    if (refDCval < 161) {
                        if (Rs < 601) {
                            if (RsCsDiff < 1) {
                                if (SC < 925) {
                                    if (Rs < 173) {
                                        if (negBalance < 19) {
                                            if (AFD < 3) {
                                                if (AFD < 1) {
                                                    return 8;
                                                }
                                                else if (AFD >= 1) {
                                                    return 1;
                                                }
                                            }
                                            else if (AFD >= 3) {
                                                if (TSC < 5) {
                                                    if (MVDiff < 456) {
                                                        return 8;
                                                    }
                                                    else if (MVDiff >= 456) {
                                                        return 4;
                                                    }
                                                }
                                                else if (TSC >= 5) {
                                                    return 2;
                                                }
                                            }
                                        }
                                        else if (negBalance >= 19) {
                                            return 4;
                                        }
                                    }
                                    else if (Rs >= 173) {
                                        if (negBalance < 21) {
                                            if (Cs < 145) {
                                                return 4;
                                            }
                                            else if (Cs >= 145) {
                                                if (SC < 498) {
                                                    return 8;
                                                }
                                                else if (SC >= 498) {
                                                    if (refDCval < 71) {
                                                        return 8;
                                                    }
                                                    else if (refDCval >= 71) {
                                                        if (TSCindex < 1) {
                                                            return 8;
                                                        }
                                                        else if (TSCindex >= 1) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (negBalance >= 21) {
                                            return 4;
                                        }
                                    }
                                }
                                else if (SC >= 925) {
                                    if (mu_mv_mag_sq < 125) {
                                        return 8;
                                    }
                                    else if (mu_mv_mag_sq >= 125) {
                                        return 4;
                                    }
                                }
                            }
                            else if (RsCsDiff >= 1) {
                                if (ssDCval < 94) {
                                    if (SC < 1257) {
                                        if (posBalance < 17) {
                                            return 8;
                                        }
                                        else if (posBalance >= 17) {
                                            if (RsCsDiff < 4) {
                                                return 4;
                                            }
                                            else if (RsCsDiff >= 4) {
                                                return 8;
                                            }
                                        }
                                    }
                                    else if (SC >= 1257) {
                                        return 8;
                                    }
                                }
                                else if (ssDCval >= 94) {
                                    if (TSC < 11) {
                                        if (Rs < 597) {
                                            return 8;
                                        }
                                        else if (Rs >= 597) {
                                            return 4;
                                        }
                                    }
                                    else if (TSC >= 11) {
                                        return 4;
                                    }
                                }
                            }
                        }
                        else if (Rs >= 601) {
                            if (MVDiff < 6) {
                                return 8;
                            }
                            else if (MVDiff >= 6) {
                                if (CsDiff < 2) {
                                    if (AFD < 5) {
                                        return 8;
                                    }
                                    else if (AFD >= 5) {
                                        if (Rs < 753) {
                                            return 4;
                                        }
                                        else if (Rs >= 753) {
                                            return 2;
                                        }
                                    }
                                }
                                else if (CsDiff >= 2) {
                                    if (tcor < 84) {
                                        return 8;
                                    }
                                    else if (tcor >= 84) {
                                        return 1;
                                    }
                                }
                            }
                        }
                    }
                    else if (refDCval >= 161) {
                        if (Cs < 410) {
                            if (SC < 604) {
                                return 8;
                            }
                            else if (SC >= 604) {
                                return 4;
                            }
                        }
                        else if (Cs >= 410) {
                            if (MVDiff < 7) {
                                if (refDCval < 165) {
                                    if (diffRsCsdiff < -1) {
                                        if (Rs < 500) {
                                            return 2;
                                        }
                                        else if (Rs >= 500) {
                                            return 4;
                                        }
                                    }
                                    else if (diffRsCsdiff >= -1) {
                                        return 4;
                                    }
                                }
                                else if (refDCval >= 165) {
                                    return 2;
                                }
                            }
                            else if (MVDiff >= 7) {
                                if (CsDiff < 1) {
                                    return 2;
                                }
                                else if (CsDiff >= 1) {
                                    if (TSC < 4) {
                                        return 4;
                                    }
                                    else if (TSC >= 4) {
                                        return 1;
                                    }
                                }
                            }
                        }
                    }
                }
                else if (Scindex >= 4) {
                    if (mu_mv_mag_sq < 121) {
                        if (TSCindex < 3) {
                            if (mu_mv_mag_sq < 94) {
                                if (posBalance < 40) {
                                    if (RsDiff < 1) {
                                        if (Rs < 1313) {
                                            return 4;
                                        }
                                        else if (Rs >= 1313) {
                                            return 1;
                                        }
                                    }
                                    else if (RsDiff >= 1) {
                                        return 8;
                                    }
                                }
                                else if (posBalance >= 40) {
                                    return 8;
                                }
                            }
                            else if (mu_mv_mag_sq >= 94) {
                                if (Cs < 707) {
                                    if (Rs < 1187) {
                                        if (Rs < 999) {
                                            return 1;
                                        }
                                        else if (Rs >= 999) {
                                            return 4;
                                        }
                                    }
                                    else if (Rs >= 1187) {
                                        return 8;
                                    }
                                }
                                else if (Cs >= 707) {
                                    if (posBalance < 11) {
                                        if (Cs < 1490) {
                                            if (mcTcor < 100) {
                                                if (Rs < 1277) {
                                                    return 8;
                                                }
                                                else if (Rs >= 1277) {
                                                    return 1;
                                                }
                                            }
                                            else if (mcTcor >= 100) {
                                                if (Rs < 915) {
                                                    return 1;
                                                }
                                                else if (Rs >= 915) {
                                                    return 2;
                                                }
                                            }
                                        }
                                        else if (Cs >= 1490) {
                                            if (posBalance < 4) {
                                                if (Rs < 2030) {
                                                    if (Rs < 1253) {
                                                        return 4;
                                                    }
                                                    else if (Rs >= 1253) {
                                                        return 8;
                                                    }
                                                }
                                                else if (Rs >= 2030) {
                                                    return 4;
                                                }
                                            }
                                            else if (posBalance >= 4) {
                                                return 4;
                                            }
                                        }
                                    }
                                    else if (posBalance >= 11) {
                                        if (diffMVdiffVal < -2) {
                                            if (Rs < 1061) {
                                                if (Cs < 1241) {
                                                    return 4;
                                                }
                                                else if (Cs >= 1241) {
                                                    if (Rs < 1038) {
                                                        return 8;
                                                    }
                                                    else if (Rs >= 1038) {
                                                        return 2;
                                                    }
                                                }
                                            }
                                            else if (Rs >= 1061) {
                                                if (SC < 3989) {
                                                    return 8;
                                                }
                                                else if (SC >= 3989) {
                                                    return 4;
                                                }
                                            }
                                        }
                                        else if (diffMVdiffVal >= -2) {
                                            if (Cs < 1193) {
                                                return 8;
                                            }
                                            else if (Cs >= 1193) {
                                                if (Cs < 1201) {
                                                    return 4;
                                                }
                                                else if (Cs >= 1201) {
                                                    if (TSCindex < 1) {
                                                        if (ssDCval < 33) {
                                                            return 2;
                                                        }
                                                        else if (ssDCval >= 33) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (TSCindex >= 1) {
                                                        if (AFD < 3) {
                                                            return 8;
                                                        }
                                                        else if (AFD >= 3) {
                                                            if (refDCval < 118) {
                                                                return 8;
                                                            }
                                                            else if (refDCval >= 118) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                        }
                        else if (TSCindex >= 3) {
                            if (SC < 6273) {
                                if (Cs < 2340) {
                                    if (SC < 2384) {
                                        if (ssDCval < 141) {
                                            if (mu_mv_mag_sq < 100) {
                                                if (Cs < 1271) {
                                                    if (Rs < 867) {
                                                        return 8;
                                                    }
                                                    else if (Rs >= 867) {
                                                        return 1;
                                                    }
                                                }
                                                else if (Cs >= 1271) {
                                                    return 8;
                                                }
                                            }
                                            else if (mu_mv_mag_sq >= 100) {
                                                if (mcTcor < 96) {
                                                    return 8;
                                                }
                                                else if (mcTcor >= 96) {
                                                    if (Rs < 1049) {
                                                        return 4;
                                                    }
                                                    else if (Rs >= 1049) {
                                                        return 1;
                                                    }
                                                }
                                            }
                                        }
                                        else if (ssDCval >= 141) {
                                            return 8;
                                        }
                                    }
                                    else if (SC >= 2384) {
                                        if (RsCsDiff < 38) {
                                            if (diffAFD < -6) {
                                                return 1;
                                            }
                                            else if (diffAFD >= -6) {
                                                if (diffAFD < 2) {
                                                    if (AFD < 6) {
                                                        return 8;
                                                    }
                                                    else if (AFD >= 6) {
                                                        if (Rs < 2299) {
                                                            return 8;
                                                        }
                                                        else if (Rs >= 2299) {
                                                            return 1;
                                                        }
                                                    }
                                                }
                                                else if (diffAFD >= 2) {
                                                    if (Cs < 2028) {
                                                        return 8;
                                                    }
                                                    else if (Cs >= 2028) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                        }
                                        else if (RsCsDiff >= 38) {
                                            if (ssDCval < 67) {
                                                if (RsCsDiff < 107) {
                                                    return 8;
                                                }
                                                else if (RsCsDiff >= 107) {
                                                    if (mu_mv_mag_sq < 106) {
                                                        return 4;
                                                    }
                                                    else if (mu_mv_mag_sq >= 106) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                            else if (ssDCval >= 67) {
                                                if (mu_mv_mag_sq < 116) {
                                                    if (posBalance < 38) {
                                                        if (ssDCval < 126) {
                                                            return 1;
                                                        }
                                                        else if (ssDCval >= 126) {
                                                            if (Rs < 1995) {
                                                                return 8;
                                                            }
                                                            else if (Rs >= 1995) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                    else if (posBalance >= 38) {
                                                        if (TSC < 5) {
                                                            if (Rs < 3481) {
                                                                return 1;
                                                            }
                                                            else if (Rs >= 3481) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (TSC >= 5) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                                else if (mu_mv_mag_sq >= 116) {
                                                    return 1;
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (Cs >= 2340) {
                                    if (Rs < 2137) {
                                        if (Rs < 1380) {
                                            return 8;
                                        }
                                        else if (Rs >= 1380) {
                                            return 1;
                                        }
                                    }
                                    else if (Rs >= 2137) {
                                        if (mcTcor < 98) {
                                            return 8;
                                        }
                                        else if (mcTcor >= 98) {
                                            if (CsDiff < 4) {
                                                if (diffTSC < -1) {
                                                    if (Rs < 3142) {
                                                        return 8;
                                                    }
                                                    else if (Rs >= 3142) {
                                                        return 1;
                                                    }
                                                }
                                                else if (diffTSC >= -1) {
                                                    return 8;
                                                }
                                            }
                                            else if (CsDiff >= 4) {
                                                if (posBalance < 56) {
                                                    if (MVDiff < 5) {
                                                        return 1;
                                                    }
                                                    else if (MVDiff >= 5) {
                                                        return 8;
                                                    }
                                                }
                                                else if (posBalance >= 56) {
                                                    return 8;
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                            else if (SC >= 6273) {
                                if (refDCval < 117) {
                                    return 8;
                                }
                                else if (refDCval >= 117) {
                                    if (mu_mv_mag_sq < 97) {
                                        if (TSC < 7) {
                                            return 4;
                                        }
                                        else if (TSC >= 7) {
                                            return 2;
                                        }
                                    }
                                    else if (mu_mv_mag_sq >= 97) {
                                        if (TSC < 4) {
                                            return 4;
                                        }
                                        else if (TSC >= 4) {
                                            return 8;
                                        }
                                    }
                                }
                            }
                        }
                    }
                    else if (mu_mv_mag_sq >= 121) {
                        if (negBalance < 3) {
                            if (Rs < 3779) {
                                if (SC < 7764) {
                                    if (diffMVdiffVal < -1) {
                                        if (Rs < 1377) {
                                            return 1;
                                        }
                                        else if (Rs >= 1377) {
                                            return 8;
                                        }
                                    }
                                    else if (diffMVdiffVal >= -1) {
                                        if (Scindex < 6) {
                                            if (diffAFD < -1) {
                                                if (diffRsCsdiff < -38) {
                                                    return 1;
                                                }
                                                else if (diffRsCsdiff >= -38) {
                                                    return 8;
                                                }
                                            }
                                            else if (diffAFD >= -1) {
                                                return 8;
                                            }
                                        }
                                        else if (Scindex >= 6) {
                                            if (posBalance < 22) {
                                                if (RsDiff < 2) {
                                                    if (SC < 7541) {
                                                        if (posBalance < 21) {
                                                            return 1;
                                                        }
                                                        else if (posBalance >= 21) {
                                                            if (Cs < 3717) {
                                                                return 8;
                                                            }
                                                            else if (Cs >= 3717) {
                                                                if (Cs < 3792) {
                                                                    return 1;
                                                                }
                                                                else if (Cs >= 3792) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (SC >= 7541) {
                                                        return 8;
                                                    }
                                                }
                                                else if (RsDiff >= 2) {
                                                    return 1;
                                                }
                                            }
                                            else if (posBalance >= 22) {
                                                if (Cs < 3891) {
                                                    return 1;
                                                }
                                                else if (Cs >= 3891) {
                                                    return 8;
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (SC >= 7764) {
                                    return 1;
                                }
                            }
                            else if (Rs >= 3779) {
                                return 8;
                            }
                        }
                        else if (negBalance >= 3) {
                            if (gchDC < 2) {
                                if (MVDiff < 4) {
                                    if (refDCval < 108) {
                                        if (tcor < 91) {
                                            if (negBalance < 18) {
                                                return 8;
                                            }
                                            else if (negBalance >= 18) {
                                                if (ssDCval < 52) {
                                                    return 4;
                                                }
                                                else if (ssDCval >= 52) {
                                                    return 8;
                                                }
                                            }
                                        }
                                        else if (tcor >= 91) {
                                            return 8;
                                        }
                                    }
                                    else if (refDCval >= 108) {
                                        if (negBalance < 24) {
                                            if (SC < 2703) {
                                                if (Rs < 1354) {
                                                    return 8;
                                                }
                                                else if (Rs >= 1354) {
                                                    return 4;
                                                }
                                            }
                                            else if (SC >= 2703) {
                                                if (posBalance < 4) {
                                                    return 4;
                                                }
                                                else if (posBalance >= 4) {
                                                    if (diffRsCsdiff < 4) {
                                                        return 8;
                                                    }
                                                    else if (diffRsCsdiff >= 4) {
                                                        if (negBalance < 13) {
                                                            if (posBalance < 34) {
                                                                return 8;
                                                            }
                                                            else if (posBalance >= 34) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (negBalance >= 13) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (negBalance >= 24) {
                                            if (mu_mv_mag_sq < 125) {
                                                if (SC < 6151) {
                                                    return 8;
                                                }
                                                else if (SC >= 6151) {
                                                    return 1;
                                                }
                                            }
                                            else if (mu_mv_mag_sq >= 125) {
                                                return 8;
                                            }
                                        }
                                    }
                                }
                                else if (MVDiff >= 4) {
                                    if (diffMVdiffVal < -3) {
                                        if (Cs < 1329) {
                                            return 1;
                                        }
                                        else if (Cs >= 1329) {
                                            return 8;
                                        }
                                    }
                                    else if (diffMVdiffVal >= -3) {
                                        if (Rs < 1889) {
                                            if (RsDiff < 1) {
                                                return 4;
                                            }
                                            else if (RsDiff >= 1) {
                                                return 8;
                                            }
                                        }
                                        else if (Rs >= 1889) {
                                            return 8;
                                        }
                                    }
                                }
                            }
                            else if (gchDC >= 2) {
                                if (AFD < 3) {
                                    return 4;
                                }
                                else if (AFD >= 3) {
                                    if (Rs < 2329) {
                                        return 1;
                                    }
                                    else if (Rs >= 2329) {
                                        return 8;
                                    }
                                }
                            }
                        }
                    }
                }
            }
            else if (negBalance >= 32) {
                if (CsDiff < 9) {
                    if (TSC < 7) {
                        if (mu_mv_mag_sq < 112) {
                            if (TSC < 6) {
                                if (MVDiff < 1) {
                                    if (Cs < 746) {
                                        if (SC < 1262) {
                                            return 2;
                                        }
                                        else if (SC >= 1262) {
                                            return 8;
                                        }
                                    }
                                    else if (Cs >= 746) {
                                        if (Cs < 2915) {
                                            return 4;
                                        }
                                        else if (Cs >= 2915) {
                                            return 8;
                                        }
                                    }
                                }
                                else if (MVDiff >= 1) {
                                    if (SC < 1424) {
                                        if (refDCval < 83) {
                                            if (tcor < 145) {
                                                if (mu_mv_mag_sq < 85) {
                                                    if (Rs < 381) {
                                                        return 4;
                                                    }
                                                    else if (Rs >= 381) {
                                                        return 8;
                                                    }
                                                }
                                                else if (mu_mv_mag_sq >= 85) {
                                                    return 4;
                                                }
                                            }
                                            else if (tcor >= 145) {
                                                return 2;
                                            }
                                        }
                                        else if (refDCval >= 83) {
                                            if (Cs < 423) {
                                                return 8;
                                            }
                                            else if (Cs >= 423) {
                                                if (Rs < 382) {
                                                    return 2;
                                                }
                                                else if (Rs >= 382) {
                                                    return 4;
                                                }
                                            }
                                        }
                                    }
                                    else if (SC >= 1424) {
                                        if (mu_mv_mag_sq < 96) {
                                            if (Cs < 1163) {
                                                if (tcor < 86) {
                                                    return 4;
                                                }
                                                else if (tcor >= 86) {
                                                    if (ssDCval < 65) {
                                                        return 8;
                                                    }
                                                    else if (ssDCval >= 65) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                            else if (Cs >= 1163) {
                                                if (ssDCval < 104) {
                                                    if (mu_mv_mag_sq < 87) {
                                                        if (Cs < 2100) {
                                                            if (MVDiff < 16) {
                                                                return 4;
                                                            }
                                                            else if (MVDiff >= 16) {
                                                                if (Cs < 1503) {
                                                                    return 8;
                                                                }
                                                                else if (Cs >= 1503) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                        else if (Cs >= 2100) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (mu_mv_mag_sq >= 87) {
                                                        if (SC < 6872) {
                                                            return 4;
                                                        }
                                                        else if (SC >= 6872) {
                                                            if (Rs < 3562) {
                                                                return 8;
                                                            }
                                                            else if (Rs >= 3562) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (ssDCval >= 104) {
                                                    return 8;
                                                }
                                            }
                                        }
                                        else if (mu_mv_mag_sq >= 96) {
                                            if (Rs < 1823) {
                                                if (Rs < 678) {
                                                    if (Rs < 526) {
                                                        return 8;
                                                    }
                                                    else if (Rs >= 526) {
                                                        return 4;
                                                    }
                                                }
                                                else if (Rs >= 678) {
                                                    return 8;
                                                }
                                            }
                                            else if (Rs >= 1823) {
                                                if (Scindex < 5) {
                                                    if (SC < 3844) {
                                                        return 4;
                                                    }
                                                    else if (SC >= 3844) {
                                                        return 8;
                                                    }
                                                }
                                                else if (Scindex >= 5) {
                                                    if (Cs < 1367) {
                                                        return 1;
                                                    }
                                                    else if (Cs >= 1367) {
                                                        if (Rs < 2716) {
                                                            if (refDCval < 90) {
                                                                if (diffRsCsdiff < 15) {
                                                                    return 8;
                                                                }
                                                                else if (diffRsCsdiff >= 15) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (refDCval >= 90) {
                                                                return 1;
                                                            }
                                                        }
                                                        else if (Rs >= 2716) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                            else if (TSC >= 6) {
                                if (Cs < 1849) {
                                    if (ssDCval < 88) {
                                        if (ssDCval < 66) {
                                            return 4;
                                        }
                                        else if (ssDCval >= 66) {
                                            if (diffTSC < 1) {
                                                return 2;
                                            }
                                            else if (diffTSC >= 1) {
                                                return 8;
                                            }
                                        }
                                    }
                                    else if (ssDCval >= 88) {
                                        if (mcTcor < 97) {
                                            if (Cs < 1281) {
                                                return 4;
                                            }
                                            else if (Cs >= 1281) {
                                                return 8;
                                            }
                                        }
                                        else if (mcTcor >= 97) {
                                            if (RsDiff < 4) {
                                                return 4;
                                            }
                                            else if (RsDiff >= 4) {
                                                if (Rs < 2789) {
                                                    return 1;
                                                }
                                                else if (Rs >= 2789) {
                                                    return 8;
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (Cs >= 1849) {
                                    if (MVDiff < 1) {
                                        return 4;
                                    }
                                    else if (MVDiff >= 1) {
                                        if (AFD < 7) {
                                            return 8;
                                        }
                                        else if (AFD >= 7) {
                                            if (RsCsDiff < 80) {
                                                if (negBalance < 39) {
                                                    return 8;
                                                }
                                                else if (negBalance >= 39) {
                                                    if (SC < 4372) {
                                                        return 4;
                                                    }
                                                    else if (SC >= 4372) {
                                                        if (Rs < 3926) {
                                                            return 8;
                                                        }
                                                        else if (Rs >= 3926) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (RsCsDiff >= 80) {
                                                if (refDCval < 122) {
                                                    if (SC < 6535) {
                                                        return 2;
                                                    }
                                                    else if (SC >= 6535) {
                                                        return 8;
                                                    }
                                                }
                                                else if (refDCval >= 122) {
                                                    return 4;
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                        }
                        else if (mu_mv_mag_sq >= 112) {
                            if (SC < 3584) {
                                if (negBalance < 45) {
                                    if (refDCval < 116) {
                                        if (Rs < 1723) {
                                            return 8;
                                        }
                                        else if (Rs >= 1723) {
                                            if (refDCval < 78) {
                                                if (RsCsDiff < 11) {
                                                    if (ssDCval < 73) {
                                                        return 8;
                                                    }
                                                    else if (ssDCval >= 73) {
                                                        return 4;
                                                    }
                                                }
                                                else if (RsCsDiff >= 11) {
                                                    if (refDCval < 77) {
                                                        if (tcor < 95) {
                                                            return 8;
                                                        }
                                                        else if (tcor >= 95) {
                                                            if (Rs < 1801) {
                                                                return 8;
                                                            }
                                                            else if (Rs >= 1801) {
                                                                if (AFD < 6) {
                                                                    if (Rs < 1809) {
                                                                        return 4;
                                                                    }
                                                                    else if (Rs >= 1809) {
                                                                        return 8;
                                                                    }
                                                                }
                                                                else if (AFD >= 6) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (refDCval >= 77) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                            else if (refDCval >= 78) {
                                                return 8;
                                            }
                                        }
                                    }
                                    else if (refDCval >= 116) {
                                        if (Cs < 248) {
                                            return 2;
                                        }
                                        else if (Cs >= 248) {
                                            return 4;
                                        }
                                    }
                                }
                                else if (negBalance >= 45) {
                                    if (MVDiff < 31) {
                                        if (ssDCval < 74) {
                                            if (negBalance < 48) {
                                                if (MVDiff < 1) {
                                                    if (Cs < 1074) {
                                                        return 8;
                                                    }
                                                    else if (Cs >= 1074) {
                                                        if (RsDiff < 5) {
                                                            if (negBalance < 46) {
                                                                if (Cs < 1512) {
                                                                    return 8;
                                                                }
                                                                else if (Cs >= 1512) {
                                                                    return 2;
                                                                }
                                                            }
                                                            else if (negBalance >= 46) {
                                                                return 2;
                                                            }
                                                        }
                                                        else if (RsDiff >= 5) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                                else if (MVDiff >= 1) {
                                                    if (Cs < 1271) {
                                                        return 2;
                                                    }
                                                    else if (Cs >= 1271) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                            else if (negBalance >= 48) {
                                                if (AFD < 5) {
                                                    return 4;
                                                }
                                                else if (AFD >= 5) {
                                                    return 2;
                                                }
                                            }
                                        }
                                        else if (ssDCval >= 74) {
                                            if (refDCval < 75) {
                                                if (Cs < 1633) {
                                                    if (AFD < 7) {
                                                        return 8;
                                                    }
                                                    else if (AFD >= 7) {
                                                        if (SC < 3401) {
                                                            return 2;
                                                        }
                                                        else if (SC >= 3401) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                                else if (Cs >= 1633) {
                                                    return 2;
                                                }
                                            }
                                            else if (refDCval >= 75) {
                                                if (ssDCval < 77) {
                                                    if (Rs < 1788) {
                                                        return 8;
                                                    }
                                                    else if (Rs >= 1788) {
                                                        if (CsDiff < 5) {
                                                            return 4;
                                                        }
                                                        else if (CsDiff >= 5) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                                else if (ssDCval >= 77) {
                                                    return 8;
                                                }
                                            }
                                        }
                                    }
                                    else if (MVDiff >= 31) {
                                        if (SC < 661) {
                                            return 2;
                                        }
                                        else if (SC >= 661) {
                                            return 1;
                                        }
                                    }
                                }
                            }
                            else if (SC >= 3584) {
                                if (refDCval < 88) {
                                    if (MVDiff < 1) {
                                        if (refDCval < 77) {
                                            if (posBalance < 57) {
                                                if (AFD < 4) {
                                                    return 8;
                                                }
                                                else if (AFD >= 4) {
                                                    return 4;
                                                }
                                            }
                                            else if (posBalance >= 57) {
                                                if (Cs < 1803) {
                                                    if (ssDCval < 75) {
                                                        return 8;
                                                    }
                                                    else if (ssDCval >= 75) {
                                                        return 4;
                                                    }
                                                }
                                                else if (Cs >= 1803) {
                                                    return 8;
                                                }
                                            }
                                        }
                                        else if (refDCval >= 77) {
                                            if (CsDiff < 8) {
                                                if (negBalance < 41) {
                                                    if (AFD < 4) {
                                                        return 8;
                                                    }
                                                    else if (AFD >= 4) {
                                                        if (mu_mv_mag_sq < 126) {
                                                            if (Rs < 3173) {
                                                                return 4;
                                                            }
                                                            else if (Rs >= 3173) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (mu_mv_mag_sq >= 126) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                                else if (negBalance >= 41) {
                                                    if (RsCsDiff < 92) {
                                                        if (tcor < 94) {
                                                            return 8;
                                                        }
                                                        else if (tcor >= 94) {
                                                            if (mu_mv_mag_sq < 118) {
                                                                return 4;
                                                            }
                                                            else if (mu_mv_mag_sq >= 118) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                    else if (RsCsDiff >= 92) {
                                                        if (SC < 9752) {
                                                            if (posBalance < 57) {
                                                                if (Rs < 4655) {
                                                                    return 4;
                                                                }
                                                                else if (Rs >= 4655) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (posBalance >= 57) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (SC >= 9752) {
                                                            if (ssDCval < 81) {
                                                                return 4;
                                                            }
                                                            else if (ssDCval >= 81) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                            else if (CsDiff >= 8) {
                                                if (ssDCval < 84) {
                                                    if (diffRsCsdiff < 23) {
                                                        if (mu_mv_mag_sq < 128) {
                                                            if (tcor < 93) {
                                                                if (diffRsCsdiff < -35) {
                                                                    if (Rs < 5012) {
                                                                        return 8;
                                                                    }
                                                                    else if (Rs >= 5012) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (diffRsCsdiff >= -35) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (tcor >= 93) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (mu_mv_mag_sq >= 128) {
                                                            if (tcor < 92) {
                                                                if (Cs < 4951) {
                                                                    if (RsDiff < 8) {
                                                                        return 4;
                                                                    }
                                                                    else if (RsDiff >= 8) {
                                                                        return 8;
                                                                    }
                                                                }
                                                                else if (Cs >= 4951) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (tcor >= 92) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                    else if (diffRsCsdiff >= 23) {
                                                        return 8;
                                                    }
                                                }
                                                else if (ssDCval >= 84) {
                                                    return 8;
                                                }
                                            }
                                        }
                                    }
                                    else if (MVDiff >= 1) {
                                        if (diffAFD < -3) {
                                            if (diffTSC < -3) {
                                                return 8;
                                            }
                                            else if (diffTSC >= -3) {
                                                return 4;
                                            }
                                        }
                                        else if (diffAFD >= -3) {
                                            return 8;
                                        }
                                    }
                                }
                                else if (refDCval >= 88) {
                                    if (Rs < 1767) {
                                        return 4;
                                    }
                                    else if (Rs >= 1767) {
                                        if (mu_mv_mag_sq < 125) {
                                            if (Rs < 2389) {
                                                if (Cs < 1903) {
                                                    if (Rs < 2269) {
                                                        return 8;
                                                    }
                                                    else if (Rs >= 2269) {
                                                        return 4;
                                                    }
                                                }
                                                else if (Cs >= 1903) {
                                                    return 4;
                                                }
                                            }
                                            else if (Rs >= 2389) {
                                                if (Rs < 2668) {
                                                    return 8;
                                                }
                                                else if (Rs >= 2668) {
                                                    if (Cs < 3161) {
                                                        if (negBalance < 48) {
                                                            if (RsCsDiff < 37) {
                                                                if (SC < 5727) {
                                                                    return 8;
                                                                }
                                                                else if (SC >= 5727) {
                                                                    if (mu_mv_mag_sq < 123) {
                                                                        if (refDCval < 114) {
                                                                            if (Rs < 2832) {
                                                                                if (Cs < 2981) {
                                                                                    return 8;
                                                                                }
                                                                                else if (Cs >= 2981) {
                                                                                    return 4;
                                                                                }
                                                                            }
                                                                            else if (Rs >= 2832) {
                                                                                if (SC < 5932) {
                                                                                    return 8;
                                                                                }
                                                                                else if (SC >= 5932) {
                                                                                    if (Rs < 2946) {
                                                                                        return 4;
                                                                                    }
                                                                                    else if (Rs >= 2946) {
                                                                                        return 8;
                                                                                    }
                                                                                }
                                                                            }
                                                                        }
                                                                        else if (refDCval >= 114) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                    else if (mu_mv_mag_sq >= 123) {
                                                                        return 8;
                                                                    }
                                                                }
                                                            }
                                                            else if (RsCsDiff >= 37) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (negBalance >= 48) {
                                                            if (refDCval < 113) {
                                                                if (Cs < 3144) {
                                                                    return 8;
                                                                }
                                                                else if (Cs >= 3144) {
                                                                    if (Rs < 3073) {
                                                                        return 4;
                                                                    }
                                                                    else if (Rs >= 3073) {
                                                                        return 8;
                                                                    }
                                                                }
                                                            }
                                                            else if (refDCval >= 113) {
                                                                if (ssDCval < 114) {
                                                                    if (Rs < 2711) {
                                                                        if (RsCsDiff < 29) {
                                                                            return 4;
                                                                        }
                                                                        else if (RsCsDiff >= 29) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                    else if (Rs >= 2711) {
                                                                        if (mcTcor < 98) {
                                                                            if (Rs < 2877) {
                                                                                return 4;
                                                                            }
                                                                            else if (Rs >= 2877) {
                                                                                return 8;
                                                                            }
                                                                        }
                                                                        else if (mcTcor >= 98) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                }
                                                                else if (ssDCval >= 114) {
                                                                    if (diffAFD < 1) {
                                                                        if (posBalance < 56) {
                                                                            return 4;
                                                                        }
                                                                        else if (posBalance >= 56) {
                                                                            if (diffRsCsdiff < -8) {
                                                                                return 4;
                                                                            }
                                                                            else if (diffRsCsdiff >= -8) {
                                                                                return 8;
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (diffAFD >= 1) {
                                                                        return 4;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (Cs >= 3161) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                        }
                                        else if (mu_mv_mag_sq >= 125) {
                                            if (RsDiff < 2) {
                                                return 4;
                                            }
                                            else if (RsDiff >= 2) {
                                                if (diffRsCsdiff < 17) {
                                                    return 8;
                                                }
                                                else if (diffRsCsdiff >= 17) {
                                                    if (diffRsCsdiff < 18) {
                                                        return 4;
                                                    }
                                                    else if (diffRsCsdiff >= 18) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                        }
                    }
                    else if (TSC >= 7) {
                        if (SC < 7392) {
                            if (RsCsDiff < 31) {
                                if (diffMVdiffVal < 3) {
                                    if (mu_mv_mag_sq < 120) {
                                        return 2;
                                    }
                                    else if (mu_mv_mag_sq >= 120) {
                                        return 8;
                                    }
                                }
                                else if (diffMVdiffVal >= 3) {
                                    if (Rs < 1845) {
                                        return 1;
                                    }
                                    else if (Rs >= 1845) {
                                        return 4;
                                    }
                                }
                            }
                            else if (RsCsDiff >= 31) {
                                if (refDCval < 105) {
                                    if (MVDiff < 1) {
                                        if (refDCval < 95) {
                                            if (mcTcor < 94) {
                                                if (ssDCval < 77) {
                                                    if (diffRsCsdiff < -14) {
                                                        return 4;
                                                    }
                                                    else if (diffRsCsdiff >= -14) {
                                                        return 2;
                                                    }
                                                }
                                                else if (ssDCval >= 77) {
                                                    if (Rs < 2968) {
                                                        if (SC < 3386) {
                                                            return 2;
                                                        }
                                                        else if (SC >= 3386) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (Rs >= 2968) {
                                                        return 2;
                                                    }
                                                }
                                            }
                                            else if (mcTcor >= 94) {
                                                return 2;
                                            }
                                        }
                                        else if (refDCval >= 95) {
                                            if (Cs < 2302) {
                                                return 4;
                                            }
                                            else if (Cs >= 2302) {
                                                return 8;
                                            }
                                        }
                                    }
                                    else if (MVDiff >= 1) {
                                        if (Cs < 2260) {
                                            if (refDCval < 93) {
                                                if (mcTcor < 91) {
                                                    if (Rs < 850) {
                                                        return 1;
                                                    }
                                                    else if (Rs >= 850) {
                                                        return 8;
                                                    }
                                                }
                                                else if (mcTcor >= 91) {
                                                    if (mcTcor < 99) {
                                                        if (diffMVdiffVal < -2) {
                                                            if (Rs < 2419) {
                                                                return 8;
                                                            }
                                                            else if (Rs >= 2419) {
                                                                if (Cs < 2208) {
                                                                    return 1;
                                                                }
                                                                else if (Cs >= 2208) {
                                                                    return 2;
                                                                }
                                                            }
                                                        }
                                                        else if (diffMVdiffVal >= -2) {
                                                            if (diffTSC < 1) {
                                                                if (MVDiff < 4) {
                                                                    return 2;
                                                                }
                                                                else if (MVDiff >= 4) {
                                                                    return 1;
                                                                }
                                                            }
                                                            else if (diffTSC >= 1) {
                                                                if (Rs < 1710) {
                                                                    return 2;
                                                                }
                                                                else if (Rs >= 1710) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (mcTcor >= 99) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                            else if (refDCval >= 93) {
                                                if (RsDiff < 6) {
                                                    if (negBalance < 38) {
                                                        return 1;
                                                    }
                                                    else if (negBalance >= 38) {
                                                        if (Rs < 2144) {
                                                            return 8;
                                                        }
                                                        else if (Rs >= 2144) {
                                                            if (MVDiff < 6) {
                                                                return 2;
                                                            }
                                                            else if (MVDiff >= 6) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (RsDiff >= 6) {
                                                    if (SC < 3344) {
                                                        return 1;
                                                    }
                                                    else if (SC >= 3344) {
                                                        if (diffMVdiffVal < 69) {
                                                            return 2;
                                                        }
                                                        else if (diffMVdiffVal >= 69) {
                                                            return 1;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (Cs >= 2260) {
                                            if (negBalance < 44) {
                                                return 8;
                                            }
                                            else if (negBalance >= 44) {
                                                if (Cs < 2426) {
                                                    if (mu_mv_mag_sq < 99) {
                                                        if (ssDCval < 87) {
                                                            return 4;
                                                        }
                                                        else if (ssDCval >= 87) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (mu_mv_mag_sq >= 99) {
                                                        return 8;
                                                    }
                                                }
                                                else if (Cs >= 2426) {
                                                    if (mu_mv_mag_sq < 86) {
                                                        return 2;
                                                    }
                                                    else if (mu_mv_mag_sq >= 86) {
                                                        if (AFD < 8) {
                                                            return 2;
                                                        }
                                                        else if (AFD >= 8) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (refDCval >= 105) {
                                    if (MVDiff < 6) {
                                        if (AFD < 12) {
                                            if (Cs < 2234) {
                                                if (negBalance < 49) {
                                                    if (mu_mv_mag_sq < 121) {
                                                        if (refDCval < 111) {
                                                            return 4;
                                                        }
                                                        else if (refDCval >= 111) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (mu_mv_mag_sq >= 121) {
                                                        if (ssDCval < 114) {
                                                            return 8;
                                                        }
                                                        else if (ssDCval >= 114) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                                else if (negBalance >= 49) {
                                                    if (SC < 4975) {
                                                        if (diffRsCsdiff < 5) {
                                                            if (Rs < 2496) {
                                                                return 4;
                                                            }
                                                            else if (Rs >= 2496) {
                                                                if (Rs < 2788) {
                                                                    return 8;
                                                                }
                                                                else if (Rs >= 2788) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                        else if (diffRsCsdiff >= 5) {
                                                            if (diffAFD < 1) {
                                                                return 8;
                                                            }
                                                            else if (diffAFD >= 1) {
                                                                if (Rs < 2610) {
                                                                    return 4;
                                                                }
                                                                else if (Rs >= 2610) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (SC >= 4975) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                            else if (Cs >= 2234) {
                                                if (Scindex < 6) {
                                                    return 8;
                                                }
                                                else if (Scindex >= 6) {
                                                    if (SC < 6705) {
                                                        if (diffAFD < 1) {
                                                            return 8;
                                                        }
                                                        else if (diffAFD >= 1) {
                                                            if (SC < 6320) {
                                                                return 8;
                                                            }
                                                            else if (SC >= 6320) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                    else if (SC >= 6705) {
                                                        if (Cs < 3100) {
                                                            return 4;
                                                        }
                                                        else if (Cs >= 3100) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (AFD >= 12) {
                                            return 1;
                                        }
                                    }
                                    else if (MVDiff >= 6) {
                                        if (SC < 6588) {
                                            if (MVDiff < 15) {
                                                if (RsCsDiff < 94) {
                                                    if (diffAFD < -5) {
                                                        return 4;
                                                    }
                                                    else if (diffAFD >= -5) {
                                                        return 8;
                                                    }
                                                }
                                                else if (RsCsDiff >= 94) {
                                                    if (Cs < 2593) {
                                                        return 4;
                                                    }
                                                    else if (Cs >= 2593) {
                                                        if (AFD < 11) {
                                                            return 1;
                                                        }
                                                        else if (AFD >= 11) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (MVDiff >= 15) {
                                                if (Rs < 1954) {
                                                    return 1;
                                                }
                                                else if (Rs >= 1954) {
                                                    return 2;
                                                }
                                            }
                                        }
                                        else if (SC >= 6588) {
                                            if (mu_mv_mag_sq < 87) {
                                                if (Rs < 4132) {
                                                    return 4;
                                                }
                                                else if (Rs >= 4132) {
                                                    return 2;
                                                }
                                            }
                                            else if (mu_mv_mag_sq >= 87) {
                                                if (SC < 7232) {
                                                    return 4;
                                                }
                                                else if (SC >= 7232) {
                                                    return 2;
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                        }
                        else if (SC >= 7392) {
                            if (negBalance < 53) {
                                if (SC < 8233) {
                                    if (RsDiff < 6) {
                                        return 4;
                                    }
                                    else if (RsDiff >= 6) {
                                        if (RsCsDiff < 140) {
                                            if (SC < 7680) {
                                                return 2;
                                            }
                                            else if (SC >= 7680) {
                                                if (mcTcor < 89) {
                                                    return 2;
                                                }
                                                else if (mcTcor >= 89) {
                                                    if (MVDiff < 5) {
                                                        return 8;
                                                    }
                                                    else if (MVDiff >= 5) {
                                                        if (diffMVdiffVal < 1) {
                                                            if (CsDiff < 6) {
                                                                return 2;
                                                            }
                                                            else if (CsDiff >= 6) {
                                                                if (RsCsDiff < 107) {
                                                                    return 8;
                                                                }
                                                                else if (RsCsDiff >= 107) {
                                                                    return 2;
                                                                }
                                                            }
                                                        }
                                                        else if (diffMVdiffVal >= 1) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (RsCsDiff >= 140) {
                                            return 4;
                                        }
                                    }
                                }
                                else if (SC >= 8233) {
                                    if (ssDCval < 82) {
                                        if (SC < 9934) {
                                            if (Cs < 4482) {
                                                return 4;
                                            }
                                            else if (Cs >= 4482) {
                                                return 8;
                                            }
                                        }
                                        else if (SC >= 9934) {
                                            if (Cs < 5213) {
                                                return 4;
                                            }
                                            else if (Cs >= 5213) {
                                                return 8;
                                            }
                                        }
                                    }
                                    else if (ssDCval >= 82) {
                                        if (Cs < 2936) {
                                            return 4;
                                        }
                                        else if (Cs >= 2936) {
                                            if (MVDiff < 9) {
                                                return 8;
                                            }
                                            else if (MVDiff >= 9) {
                                                return 2;
                                            }
                                        }
                                    }
                                }
                            }
                            else if (negBalance >= 53) {
                                return 4;
                            }
                        }
                    }
                }
                else if (CsDiff >= 9) {
                    if (SC < 4808) {
                        if (ssDCval < 151) {
                            if (SC < 3530) {
                                if (Rs < 621) {
                                    return 2;
                                }
                                else if (Rs >= 621) {
                                    if (MVDiff < 60) {
                                        return 2;
                                    }
                                    else if (MVDiff >= 60) {
                                        return 1;
                                    }
                                }
                            }
                            else if (SC >= 3530) {
                                if (AFD < 12) {
                                    if (Cs < 2237) {
                                        return 4;
                                    }
                                    else if (Cs >= 2237) {
                                        return 8;
                                    }
                                }
                                else if (AFD >= 12) {
                                    if (TSC < 24) {
                                        return 2;
                                    }
                                    else if (TSC >= 24) {
                                        return 4;
                                    }
                                }
                            }
                        }
                        else if (ssDCval >= 151) {
                            return 8;
                        }
                    }
                    else if (SC >= 4808) {
                        if (negBalance < 64) {
                            if (CsDiff < 15) {
                                if (tcor < 81) {
                                    if (Cs < 3595) {
                                        if (AFD < 14) {
                                            return 4;
                                        }
                                        else if (AFD >= 14) {
                                            if (Cs < 2433) {
                                                return 2;
                                            }
                                            else if (Cs >= 2433) {
                                                return 8;
                                            }
                                        }
                                    }
                                    else if (Cs >= 3595) {
                                        if (Rs < 3998) {
                                            return 1;
                                        }
                                        else if (Rs >= 3998) {
                                            return 8;
                                        }
                                    }
                                }
                                else if (tcor >= 81) {
                                    if (ssDCval < 106) {
                                        if (SC < 10460) {
                                            if (tcor < 90) {
                                                return 8;
                                            }
                                            else if (tcor >= 90) {
                                                if (SC < 6887) {
                                                    if (Rs < 3403) {
                                                        return 8;
                                                    }
                                                    else if (Rs >= 3403) {
                                                        return 2;
                                                    }
                                                }
                                                else if (SC >= 6887) {
                                                    if (posBalance < 60) {
                                                        if (tcor < 94) {
                                                            if (Cs < 5144) {
                                                                if (mu_mv_mag_sq < 114) {
                                                                    return 4;
                                                                }
                                                                else if (mu_mv_mag_sq >= 114) {
                                                                    if (Rs < 4704) {
                                                                        return 8;
                                                                    }
                                                                    else if (Rs >= 4704) {
                                                                        if (Cs < 5045) {
                                                                            if (SC < 9928) {
                                                                                if (Rs < 4859) {
                                                                                    if (posBalance < 58) {
                                                                                        return 8;
                                                                                    }
                                                                                    else if (posBalance >= 58) {
                                                                                        if (RsDiff < 9) {
                                                                                            return 4;
                                                                                        }
                                                                                        else if (RsDiff >= 9) {
                                                                                            return 8;
                                                                                        }
                                                                                    }
                                                                                }
                                                                                else if (Rs >= 4859) {
                                                                                    return 4;
                                                                                }
                                                                            }
                                                                            else if (SC >= 9928) {
                                                                                return 8;
                                                                            }
                                                                        }
                                                                        else if (Cs >= 5045) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                            else if (Cs >= 5144) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (tcor >= 94) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (posBalance >= 60) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                        }
                                        else if (SC >= 10460) {
                                            return 8;
                                        }
                                    }
                                    else if (ssDCval >= 106) {
                                        if (ssDCval < 109) {
                                            if (negBalance < 46) {
                                                return 8;
                                            }
                                            else if (negBalance >= 46) {
                                                if (diffMVdiffVal < -1) {
                                                    return 8;
                                                }
                                                else if (diffMVdiffVal >= -1) {
                                                    return 4;
                                                }
                                            }
                                        }
                                        else if (ssDCval >= 109) {
                                            if (ssDCval < 121) {
                                                if (refDCval < 114) {
                                                    return 8;
                                                }
                                                else if (refDCval >= 114) {
                                                    if (posBalance < 53) {
                                                        return 8;
                                                    }
                                                    else if (posBalance >= 53) {
                                                        if (Cs < 4319) {
                                                            if (ssDCval < 115) {
                                                                return 2;
                                                            }
                                                            else if (ssDCval >= 115) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (Cs >= 4319) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (ssDCval >= 121) {
                                                if (Cs < 3011) {
                                                    if (Rs < 3787) {
                                                        return 8;
                                                    }
                                                    else if (Rs >= 3787) {
                                                        return 2;
                                                    }
                                                }
                                                else if (Cs >= 3011) {
                                                    return 8;
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                            else if (CsDiff >= 15) {
                                if (Cs < 4292) {
                                    if (ssDCval < 133) {
                                        if (Cs < 3404) {
                                            return 4;
                                        }
                                        else if (Cs >= 3404) {
                                            return 8;
                                        }
                                    }
                                    else if (ssDCval >= 133) {
                                        return 1;
                                    }
                                }
                                else if (Cs >= 4292) {
                                    if (tcor < 63) {
                                        if (Rs < 4843) {
                                            return 1;
                                        }
                                        else if (Rs >= 4843) {
                                            return 8;
                                        }
                                    }
                                    else if (tcor >= 63) {
                                        if (diffTSC < -3) {
                                            if (Rs < 8032) {
                                                return 8;
                                            }
                                            else if (Rs >= 8032) {
                                                return 2;
                                            }
                                        }
                                        else if (diffTSC >= -3) {
                                            return 8;
                                        }
                                    }
                                }
                            }
                        }
                        else if (negBalance >= 64) {
                            if (SC < 15573) {
                                if (Cs < 7374) {
                                    return 8;
                                }
                                else if (Cs >= 7374) {
                                    if (posBalance < 52) {
                                        return 2;
                                    }
                                    else if (posBalance >= 52) {
                                        if (CsDiff < 22) {
                                            return 4;
                                        }
                                        else if (CsDiff >= 22) {
                                            return 2;
                                        }
                                    }
                                }
                            }
                            else if (SC >= 15573) {
                                if (posBalance < 54) {
                                    if (RsDiff < 25) {
                                        if (posBalance < 53) {
                                            return 2;
                                        }
                                        else if (posBalance >= 53) {
                                            if (Rs < 8163) {
                                                return 2;
                                            }
                                            else if (Rs >= 8163) {
                                                return 4;
                                            }
                                        }
                                    }
                                    else if (RsDiff >= 25) {
                                        if (Rs < 8298) {
                                            return 4;
                                        }
                                        else if (Rs >= 8298) {
                                            return 2;
                                        }
                                    }
                                }
                                else if (posBalance >= 54) {
                                    return 4;
                                }
                            }
                        }
                    }
                }
            }
        }
        return 4;
    }
    mfxU16 AGOPSelectE(mfxU32 Rs, mfxU32 CsDiff, mfxU32 TSC, mfxI32 diffAFD, mfxU32 Scindex, mfxU32 RsDiff, mfxI32 diffMVdiffVal, mfxU32 RsCsDiff, mfxU32 TSCindex, mfxI16 tcor, mfxU32 refDCval, mfxU32 MVDiff, mfxU32 AFD, mfxI32 diffTSC, mfxI32 diffRsCsdiff, mfxU32 posBalance, mfxU32 SC, mfxU32 mu_mv_mag_sq, mfxU32 negBalance, mfxI16 mcTcor, mfxU32 gchDC, mfxU32 Cs, mfxU32 ssDCval) {
        if (tcor < 96) {
            if (mu_mv_mag_sq < 82) {
                if (refDCval < 9) {
                    if (SC < 3490) {
                        return 8;
                    }
                    else if (SC >= 3490) {
                        if (SC < 9327) {
                            if (Rs < 2662) {
                                return 1;
                            }
                            else if (Rs >= 2662) {
                                if (SC < 7709) {
                                    return 8;
                                }
                                else if (SC >= 7709) {
                                    if (SC < 7755) {
                                        return 1;
                                    }
                                    else if (SC >= 7755) {
                                        return 8;
                                    }
                                }
                            }
                        }
                        else if (SC >= 9327) {
                            return 1;
                        }
                    }
                }
                else if (refDCval >= 9) {
                    if (Cs < 3269) {
                        if (AFD < 12) {
                            if (RsCsDiff < 19) {
                                if (RsCsDiff < 3) {
                                    if (ssDCval < 85) {
                                        if (mu_mv_mag_sq < 45) {
                                            if (MVDiff < 6) {
                                                if (mcTcor < 98) {
                                                    if (posBalance < 52) {
                                                        return 4;
                                                    }
                                                    else if (posBalance >= 52) {
                                                        if (refDCval < 43) {
                                                            return 4;
                                                        }
                                                        else if (refDCval >= 43) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                                else if (mcTcor >= 98) {
                                                    return 4;
                                                }
                                            }
                                            else if (MVDiff >= 6) {
                                                if (posBalance < 29) {
                                                    if (negBalance < 2) {
                                                        if (diffMVdiffVal < -2) {
                                                            if (posBalance < 25) {
                                                                return 1;
                                                            }
                                                            else if (posBalance >= 25) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (diffMVdiffVal >= -2) {
                                                            return 1;
                                                        }
                                                    }
                                                    else if (negBalance >= 2) {
                                                        if (AFD < 2) {
                                                            if (mu_mv_mag_sq < 32) {
                                                                if (mu_mv_mag_sq < 25) {
                                                                    if (tcor < 67) {
                                                                        if (mcTcor < 9) {
                                                                            return 2;
                                                                        }
                                                                        else if (mcTcor >= 9) {
                                                                            if (mu_mv_mag_sq < 5) {
                                                                                return 8;
                                                                            }
                                                                            else if (mu_mv_mag_sq >= 5) {
                                                                                if (diffMVdiffVal < -35) {
                                                                                    return 8;
                                                                                }
                                                                                else if (diffMVdiffVal >= -35) {
                                                                                    return 4;
                                                                                }
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (tcor >= 67) {
                                                                        return 8;
                                                                    }
                                                                }
                                                                else if (mu_mv_mag_sq >= 25) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (mu_mv_mag_sq >= 32) {
                                                                if (ssDCval < 20) {
                                                                    if (MVDiff < 57) {
                                                                        return 1;
                                                                    }
                                                                    else if (MVDiff >= 57) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (ssDCval >= 20) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                        else if (AFD >= 2) {
                                                            if (MVDiff < 59) {
                                                                if (Cs < 57) {
                                                                    return 2;
                                                                }
                                                                else if (Cs >= 57) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (MVDiff >= 59) {
                                                                return 1;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (posBalance >= 29) {
                                                    if (tcor < 75) {
                                                        if (AFD < 1) {
                                                            if (diffMVdiffVal < -42) {
                                                                if (gchDC < 1) {
                                                                    return 2;
                                                                }
                                                                else if (gchDC >= 1) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (diffMVdiffVal >= -42) {
                                                                if (tcor < 39) {
                                                                    if (MVDiff < 939) {
                                                                        return 4;
                                                                    }
                                                                    else if (MVDiff >= 939) {
                                                                        return 2;
                                                                    }
                                                                }
                                                                else if (tcor >= 39) {
                                                                    if (gchDC < 1) {
                                                                        if (refDCval < 20) {
                                                                            return 8;
                                                                        }
                                                                        else if (refDCval >= 20) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                    else if (gchDC >= 1) {
                                                                        return 2;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (AFD >= 1) {
                                                            if (tcor < 68) {
                                                                if (mu_mv_mag_sq < 17) {
                                                                    if (MVDiff < 37) {
                                                                        return 2;
                                                                    }
                                                                    else if (MVDiff >= 37) {
                                                                        if (mcTcor < 43) {
                                                                            if (diffTSC < -1) {
                                                                                return 1;
                                                                            }
                                                                            else if (diffTSC >= -1) {
                                                                                if (MVDiff < 105) {
                                                                                    return 8;
                                                                                }
                                                                                else if (MVDiff >= 105) {
                                                                                    return 2;
                                                                                }
                                                                            }
                                                                        }
                                                                        else if (mcTcor >= 43) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                }
                                                                else if (mu_mv_mag_sq >= 17) {
                                                                    if (ssDCval < 25) {
                                                                        return 4;
                                                                    }
                                                                    else if (ssDCval >= 25) {
                                                                        if (refDCval < 26) {
                                                                            return 8;
                                                                        }
                                                                        else if (refDCval >= 26) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                            else if (tcor >= 68) {
                                                                if (posBalance < 81) {
                                                                    if (Cs < 379) {
                                                                        return 1;
                                                                    }
                                                                    else if (Cs >= 379) {
                                                                        if (AFD < 7) {
                                                                            return 4;
                                                                        }
                                                                        else if (AFD >= 7) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                }
                                                                else if (posBalance >= 81) {
                                                                    if (TSC < 1) {
                                                                        return 8;
                                                                    }
                                                                    else if (TSC >= 1) {
                                                                        if (tcor < 70) {
                                                                            if (Cs < 3) {
                                                                                return 4;
                                                                            }
                                                                            else if (Cs >= 3) {
                                                                                return 8;
                                                                            }
                                                                        }
                                                                        else if (tcor >= 70) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (tcor >= 75) {
                                                        if (mu_mv_mag_sq < 33) {
                                                            if (tcor < 78) {
                                                                if (AFD < 4) {
                                                                    return 2;
                                                                }
                                                                else if (AFD >= 4) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (tcor >= 78) {
                                                                if (Cs < 1) {
                                                                    if (AFD < 1) {
                                                                        return 8;
                                                                    }
                                                                    else if (AFD >= 1) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (Cs >= 1) {
                                                                    if (Rs < 228) {
                                                                        if (ssDCval < 34) {
                                                                            return 1;
                                                                        }
                                                                        else if (ssDCval >= 34) {
                                                                            if (AFD < 5) {
                                                                                return 2;
                                                                            }
                                                                            else if (AFD >= 5) {
                                                                                if (mu_mv_mag_sq < 23) {
                                                                                    return 2;
                                                                                }
                                                                                else if (mu_mv_mag_sq >= 23) {
                                                                                    return 8;
                                                                                }
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (Rs >= 228) {
                                                                        if (Cs < 368) {
                                                                            if (CsDiff < 1) {
                                                                                return 4;
                                                                            }
                                                                            else if (CsDiff >= 1) {
                                                                                if (Cs < 234) {
                                                                                    return 2;
                                                                                }
                                                                                else if (Cs >= 234) {
                                                                                    return 4;
                                                                                }
                                                                            }
                                                                        }
                                                                        else if (Cs >= 368) {
                                                                            if (Rs < 888) {
                                                                                return 2;
                                                                            }
                                                                            else if (Rs >= 888) {
                                                                                return 8;
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (mu_mv_mag_sq >= 33) {
                                                            if (refDCval < 39) {
                                                                if (Rs < 808) {
                                                                    if (refDCval < 38) {
                                                                        if (Rs < 457) {
                                                                            if (MVDiff < 21) {
                                                                                return 8;
                                                                            }
                                                                            else if (MVDiff >= 21) {
                                                                                if (mcTcor < 77) {
                                                                                    return 1;
                                                                                }
                                                                                else if (mcTcor >= 77) {
                                                                                    if (tcor < 77) {
                                                                                        return 2;
                                                                                    }
                                                                                    else if (tcor >= 77) {
                                                                                        return 4;
                                                                                    }
                                                                                }
                                                                            }
                                                                        }
                                                                        else if (Rs >= 457) {
                                                                            return 2;
                                                                        }
                                                                    }
                                                                    else if (refDCval >= 38) {
                                                                        if (Rs < 377) {
                                                                            return 4;
                                                                        }
                                                                        else if (Rs >= 377) {
                                                                            if (Rs < 745) {
                                                                                return 1;
                                                                            }
                                                                            else if (Rs >= 745) {
                                                                                if (posBalance < 59) {
                                                                                    return 4;
                                                                                }
                                                                                else if (posBalance >= 59) {
                                                                                    return 1;
                                                                                }
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                                else if (Rs >= 808) {
                                                                    return 2;
                                                                }
                                                            }
                                                            else if (refDCval >= 39) {
                                                                if (ssDCval < 47) {
                                                                    if (Rs < 782) {
                                                                        return 4;
                                                                    }
                                                                    else if (Rs >= 782) {
                                                                        if (Rs < 1891) {
                                                                            return 1;
                                                                        }
                                                                        else if (Rs >= 1891) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                }
                                                                else if (ssDCval >= 47) {
                                                                    if (diffAFD < -1) {
                                                                        return 8;
                                                                    }
                                                                    else if (diffAFD >= -1) {
                                                                        return 4;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (mu_mv_mag_sq >= 45) {
                                            if (ssDCval < 39) {
                                                if (MVDiff < 23) {
                                                    if (ssDCval < 38) {
                                                        if (mu_mv_mag_sq < 66) {
                                                            if (tcor < 83) {
                                                                if (mcTcor < 82) {
                                                                    return 1;
                                                                }
                                                                else if (mcTcor >= 82) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (tcor >= 83) {
                                                                return 2;
                                                            }
                                                        }
                                                        else if (mu_mv_mag_sq >= 66) {
                                                            if (diffMVdiffVal < 7) {
                                                                return 4;
                                                            }
                                                            else if (diffMVdiffVal >= 7) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                    else if (ssDCval >= 38) {
                                                        if (mu_mv_mag_sq < 77) {
                                                            if (Rs < 1359) {
                                                                return 4;
                                                            }
                                                            else if (Rs >= 1359) {
                                                                return 1;
                                                            }
                                                        }
                                                        else if (mu_mv_mag_sq >= 77) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                                else if (MVDiff >= 23) {
                                                    if (Rs < 460) {
                                                        if (negBalance < 47) {
                                                            if (mcTcor < 82) {
                                                                if (TSC < 1) {
                                                                    return 4;
                                                                }
                                                                else if (TSC >= 1) {
                                                                    if (AFD < 5) {
                                                                        if (negBalance < 44) {
                                                                            if (MVDiff < 25) {
                                                                                return 8;
                                                                            }
                                                                            else if (MVDiff >= 25) {
                                                                                if (posBalance < 36) {
                                                                                    return 8;
                                                                                }
                                                                                else if (posBalance >= 36) {
                                                                                    return 1;
                                                                                }
                                                                            }
                                                                        }
                                                                        else if (negBalance >= 44) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                    else if (AFD >= 5) {
                                                                        if (posBalance < 57) {
                                                                            return 4;
                                                                        }
                                                                        else if (posBalance >= 57) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                            else if (mcTcor >= 82) {
                                                                if (MVDiff < 26) {
                                                                    if (mu_mv_mag_sq < 59) {
                                                                        if (diffAFD < -1) {
                                                                            return 1;
                                                                        }
                                                                        else if (diffAFD >= -1) {
                                                                            return 2;
                                                                        }
                                                                    }
                                                                    else if (mu_mv_mag_sq >= 59) {
                                                                        if (negBalance < 42) {
                                                                            return 4;
                                                                        }
                                                                        else if (negBalance >= 42) {
                                                                            if (Cs < 377) {
                                                                                return 1;
                                                                            }
                                                                            else if (Cs >= 377) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                                else if (MVDiff >= 26) {
                                                                    if (Cs < 386) {
                                                                        return 4;
                                                                    }
                                                                    else if (Cs >= 386) {
                                                                        return 2;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (negBalance >= 47) {
                                                            if (ssDCval < 18) {
                                                                return 1;
                                                            }
                                                            else if (ssDCval >= 18) {
                                                                return 2;
                                                            }
                                                        }
                                                    }
                                                    else if (Rs >= 460) {
                                                        if (TSC < 5) {
                                                            if (MVDiff < 213) {
                                                                if (mu_mv_mag_sq < 77) {
                                                                    return 2;
                                                                }
                                                                else if (mu_mv_mag_sq >= 77) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (MVDiff >= 213) {
                                                                return 1;
                                                            }
                                                        }
                                                        else if (TSC >= 5) {
                                                            if (diffMVdiffVal < 13) {
                                                                if (SC < 1079) {
                                                                    return 2;
                                                                }
                                                                else if (SC >= 1079) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (diffMVdiffVal >= 13) {
                                                                return 1;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                            else if (ssDCval >= 39) {
                                                if (ssDCval < 53) {
                                                    if (SC < 1607) {
                                                        if (diffAFD < 2) {
                                                            if (TSCindex < 4) {
                                                                if (Rs < 699) {
                                                                    return 8;
                                                                }
                                                                else if (Rs >= 699) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (TSCindex >= 4) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (diffAFD >= 2) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (SC >= 1607) {
                                                        if (TSC < 3) {
                                                            return 1;
                                                        }
                                                        else if (TSC >= 3) {
                                                            if (Cs < 1549) {
                                                                if (SC < 1677) {
                                                                    if (Rs < 929) {
                                                                        return 4;
                                                                    }
                                                                    else if (Rs >= 929) {
                                                                        return 1;
                                                                    }
                                                                }
                                                                else if (SC >= 1677) {
                                                                    if (Cs < 1526) {
                                                                        return 4;
                                                                    }
                                                                    else if (Cs >= 1526) {
                                                                        if (Rs < 1829) {
                                                                            return 1;
                                                                        }
                                                                        else if (Rs >= 1829) {
                                                                            if (mu_mv_mag_sq < 62) {
                                                                                if (Rs < 1850) {
                                                                                    return 4;
                                                                                }
                                                                                else if (Rs >= 1850) {
                                                                                    return 1;
                                                                                }
                                                                            }
                                                                            else if (mu_mv_mag_sq >= 62) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                            else if (Cs >= 1549) {
                                                                if (mu_mv_mag_sq < 62) {
                                                                    return 1;
                                                                }
                                                                else if (mu_mv_mag_sq >= 62) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (ssDCval >= 53) {
                                                    if (AFD < 5) {
                                                        if (SC < 656) {
                                                            return 8;
                                                        }
                                                        else if (SC >= 656) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (AFD >= 5) {
                                                        return 1;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (ssDCval >= 85) {
                                        if (Rs < 294) {
                                            if (TSC < 5) {
                                                if (ssDCval < 88) {
                                                    return 2;
                                                }
                                                else if (ssDCval >= 88) {
                                                    if (Cs < 277) {
                                                        if (Rs < 224) {
                                                            return 2;
                                                        }
                                                        else if (Rs >= 224) {
                                                            return 1;
                                                        }
                                                    }
                                                    else if (Cs >= 277) {
                                                        if (AFD < 7) {
                                                            return 4;
                                                        }
                                                        else if (AFD >= 7) {
                                                            return 2;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (TSC >= 5) {
                                                if (SC < 373) {
                                                    return 1;
                                                }
                                                else if (SC >= 373) {
                                                    return 2;
                                                }
                                            }
                                        }
                                        else if (Rs >= 294) {
                                            return 4;
                                        }
                                    }
                                }
                                else if (RsCsDiff >= 3) {
                                    if (SC < 760) {
                                        if (TSCindex < 6) {
                                            if (refDCval < 35) {
                                                if (Rs < 302) {
                                                    return 8;
                                                }
                                                else if (Rs >= 302) {
                                                    return 4;
                                                }
                                            }
                                            else if (refDCval >= 35) {
                                                if (diffAFD < 2) {
                                                    return 2;
                                                }
                                                else if (diffAFD >= 2) {
                                                    if (AFD < 8) {
                                                        return 2;
                                                    }
                                                    else if (AFD >= 8) {
                                                        return 1;
                                                    }
                                                }
                                            }
                                        }
                                        else if (TSCindex >= 6) {
                                            if (CsDiff < 1) {
                                                return 1;
                                            }
                                            else if (CsDiff >= 1) {
                                                if (ssDCval < 110) {
                                                    if (MVDiff < 17) {
                                                        if (mu_mv_mag_sq < 27) {
                                                            return 2;
                                                        }
                                                        else if (mu_mv_mag_sq >= 27) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (MVDiff >= 17) {
                                                        if (refDCval < 105) {
                                                            return 1;
                                                        }
                                                        else if (refDCval >= 105) {
                                                            if (Rs < 470) {
                                                                return 8;
                                                            }
                                                            else if (Rs >= 470) {
                                                                return 2;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (ssDCval >= 110) {
                                                    return 4;
                                                }
                                            }
                                        }
                                    }
                                    else if (SC >= 760) {
                                        if (ssDCval < 96) {
                                            if (Rs < 2180) {
                                                if (Rs < 1205) {
                                                    if (TSC < 5) {
                                                        if (MVDiff < 14) {
                                                            if (Rs < 930) {
                                                                if (RsCsDiff < 12) {
                                                                    if (posBalance < 52) {
                                                                        return 4;
                                                                    }
                                                                    else if (posBalance >= 52) {
                                                                        if (RsCsDiff < 7) {
                                                                            if (Rs < 757) {
                                                                                return 4;
                                                                            }
                                                                            else if (Rs >= 757) {
                                                                                return 8;
                                                                            }
                                                                        }
                                                                        else if (RsCsDiff >= 7) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                }
                                                                else if (RsCsDiff >= 12) {
                                                                    if (MVDiff < 5) {
                                                                        return 4;
                                                                    }
                                                                    else if (MVDiff >= 5) {
                                                                        return 2;
                                                                    }
                                                                }
                                                            }
                                                            else if (Rs >= 930) {
                                                                if (diffMVdiffVal < 3) {
                                                                    if (ssDCval < 57) {
                                                                        return 4;
                                                                    }
                                                                    else if (ssDCval >= 57) {
                                                                        if (Cs < 618) {
                                                                            return 4;
                                                                        }
                                                                        else if (Cs >= 618) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                }
                                                                else if (diffMVdiffVal >= 3) {
                                                                    if (MVDiff < 7) {
                                                                        return 4;
                                                                    }
                                                                    else if (MVDiff >= 7) {
                                                                        if (CsDiff < 2) {
                                                                            return 4;
                                                                        }
                                                                        else if (CsDiff >= 2) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (MVDiff >= 14) {
                                                            if (refDCval < 44) {
                                                                if (tcor < 89) {
                                                                    return 2;
                                                                }
                                                                else if (tcor >= 89) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (refDCval >= 44) {
                                                                if (diffRsCsdiff < -6) {
                                                                    if (diffMVdiffVal < 1) {
                                                                        return 8;
                                                                    }
                                                                    else if (diffMVdiffVal >= 1) {
                                                                        if (Rs < 869) {
                                                                            return 4;
                                                                        }
                                                                        else if (Rs >= 869) {
                                                                            return 2;
                                                                        }
                                                                    }
                                                                }
                                                                else if (diffRsCsdiff >= -6) {
                                                                    if (diffMVdiffVal < -9) {
                                                                        if (refDCval < 53) {
                                                                            return 1;
                                                                        }
                                                                        else if (refDCval >= 53) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                    else if (diffMVdiffVal >= -9) {
                                                                        if (diffMVdiffVal < 9) {
                                                                            return 4;
                                                                        }
                                                                        else if (diffMVdiffVal >= 9) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (TSC >= 5) {
                                                        if (gchDC < 2) {
                                                            if (Cs < 368) {
                                                                return 1;
                                                            }
                                                            else if (Cs >= 368) {
                                                                if (ssDCval < 61) {
                                                                    if (Cs < 510) {
                                                                        if (Rs < 656) {
                                                                            if (Cs < 494) {
                                                                                return 2;
                                                                            }
                                                                            else if (Cs >= 494) {
                                                                                return 8;
                                                                            }
                                                                        }
                                                                        else if (Rs >= 656) {
                                                                            if (SC < 1193) {
                                                                                if (SC < 1172) {
                                                                                    return 4;
                                                                                }
                                                                                else if (SC >= 1172) {
                                                                                    return 1;
                                                                                }
                                                                            }
                                                                            else if (SC >= 1193) {
                                                                                if (negBalance < 51) {
                                                                                    return 4;
                                                                                }
                                                                                else if (negBalance >= 51) {
                                                                                    if (mcTcor < 74) {
                                                                                        return 4;
                                                                                    }
                                                                                    else if (mcTcor >= 74) {
                                                                                        return 1;
                                                                                    }
                                                                                }
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (Cs >= 510) {
                                                                        if (diffMVdiffVal < 15) {
                                                                            if (negBalance < 49) {
                                                                                if (tcor < 78) {
                                                                                    if (MVDiff < 23) {
                                                                                        if (SC < 1391) {
                                                                                            return 1;
                                                                                        }
                                                                                        else if (SC >= 1391) {
                                                                                            return 4;
                                                                                        }
                                                                                    }
                                                                                    else if (MVDiff >= 23) {
                                                                                        return 4;
                                                                                    }
                                                                                }
                                                                                else if (tcor >= 78) {
                                                                                    if (SC < 2093) {
                                                                                        if (tcor < 87) {
                                                                                            if (Rs < 824) {
                                                                                                if (Rs < 808) {
                                                                                                    return 1;
                                                                                                }
                                                                                                else if (Rs >= 808) {
                                                                                                    return 4;
                                                                                                }
                                                                                            }
                                                                                            else if (Rs >= 824) {
                                                                                                return 1;
                                                                                            }
                                                                                        }
                                                                                        else if (tcor >= 87) {
                                                                                            if (mu_mv_mag_sq < 54) {
                                                                                                return 4;
                                                                                            }
                                                                                            else if (mu_mv_mag_sq >= 54) {
                                                                                                if (gchDC < 1) {
                                                                                                    return 1;
                                                                                                }
                                                                                                else if (gchDC >= 1) {
                                                                                                    return 4;
                                                                                                }
                                                                                            }
                                                                                        }
                                                                                    }
                                                                                    else if (SC >= 2093) {
                                                                                        return 2;
                                                                                    }
                                                                                }
                                                                            }
                                                                            else if (negBalance >= 49) {
                                                                                if (tcor < 71) {
                                                                                    if (Rs < 794) {
                                                                                        if (tcor < 69) {
                                                                                            return 1;
                                                                                        }
                                                                                        else if (tcor >= 69) {
                                                                                            return 4;
                                                                                        }
                                                                                    }
                                                                                    else if (Rs >= 794) {
                                                                                        return 1;
                                                                                    }
                                                                                }
                                                                                else if (tcor >= 71) {
                                                                                    if (Rs < 865) {
                                                                                        return 1;
                                                                                    }
                                                                                    else if (Rs >= 865) {
                                                                                        if (diffAFD < -1) {
                                                                                            return 1;
                                                                                        }
                                                                                        else if (diffAFD >= -1) {
                                                                                            return 4;
                                                                                        }
                                                                                    }
                                                                                }
                                                                            }
                                                                        }
                                                                        else if (diffMVdiffVal >= 15) {
                                                                            if (Rs < 895) {
                                                                                return 4;
                                                                            }
                                                                            else if (Rs >= 895) {
                                                                                return 8;
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                                else if (ssDCval >= 61) {
                                                                    if (ssDCval < 79) {
                                                                        if (AFD < 10) {
                                                                            return 2;
                                                                        }
                                                                        else if (AFD >= 10) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                    else if (ssDCval >= 79) {
                                                                        return 4;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (gchDC >= 2) {
                                                            if (ssDCval < 85) {
                                                                if (mu_mv_mag_sq < 46) {
                                                                    if (Cs < 747) {
                                                                        if (Cs < 266) {
                                                                            if (SC < 1107) {
                                                                                return 2;
                                                                            }
                                                                            else if (SC >= 1107) {
                                                                                return 8;
                                                                            }
                                                                        }
                                                                        else if (Cs >= 266) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                    else if (Cs >= 747) {
                                                                        if (SC < 1534) {
                                                                            return 2;
                                                                        }
                                                                        else if (SC >= 1534) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                }
                                                                else if (mu_mv_mag_sq >= 46) {
                                                                    if (negBalance < 33) {
                                                                        return 4;
                                                                    }
                                                                    else if (negBalance >= 33) {
                                                                        if (Cs < 830) {
                                                                            return 4;
                                                                        }
                                                                        else if (Cs >= 830) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                            else if (ssDCval >= 85) {
                                                                return 1;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (Rs >= 1205) {
                                                    if (mu_mv_mag_sq < 74) {
                                                        if (SC < 2912) {
                                                            if (Rs < 1436) {
                                                                if (negBalance < 52) {
                                                                    if (posBalance < 49) {
                                                                        if (MVDiff < 8) {
                                                                            if (Rs < 1230) {
                                                                                if (Rs < 1213) {
                                                                                    return 1;
                                                                                }
                                                                                else if (Rs >= 1213) {
                                                                                    return 4;
                                                                                }
                                                                            }
                                                                            else if (Rs >= 1230) {
                                                                                if (MVDiff < 7) {
                                                                                    return 1;
                                                                                }
                                                                                else if (MVDiff >= 7) {
                                                                                    if (Cs < 954) {
                                                                                        return 4;
                                                                                    }
                                                                                    else if (Cs >= 954) {
                                                                                        return 1;
                                                                                    }
                                                                                }
                                                                            }
                                                                        }
                                                                        else if (MVDiff >= 8) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                    else if (posBalance >= 49) {
                                                                        if (mu_mv_mag_sq < 69) {
                                                                            return 1;
                                                                        }
                                                                        else if (mu_mv_mag_sq >= 69) {
                                                                            if (Rs < 1294) {
                                                                                return 1;
                                                                            }
                                                                            else if (Rs >= 1294) {
                                                                                if (Rs < 1354) {
                                                                                    return 4;
                                                                                }
                                                                                else if (Rs >= 1354) {
                                                                                    return 1;
                                                                                }
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                                else if (negBalance >= 52) {
                                                                    if (Cs < 1002) {
                                                                        return 4;
                                                                    }
                                                                    else if (Cs >= 1002) {
                                                                        return 2;
                                                                    }
                                                                }
                                                            }
                                                            else if (Rs >= 1436) {
                                                                if (tcor < 91) {
                                                                    if (SC < 2550) {
                                                                        if (mcTcor < 94) {
                                                                            return 8;
                                                                        }
                                                                        else if (mcTcor >= 94) {
                                                                            if (TSC < 5) {
                                                                                return 1;
                                                                            }
                                                                            else if (TSC >= 5) {
                                                                                return 8;
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (SC >= 2550) {
                                                                        if (Rs < 1529) {
                                                                            return 1;
                                                                        }
                                                                        else if (Rs >= 1529) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                }
                                                                else if (tcor >= 91) {
                                                                    if (Cs < 831) {
                                                                        return 4;
                                                                    }
                                                                    else if (Cs >= 831) {
                                                                        if (tcor < 93) {
                                                                            if (MVDiff < 10) {
                                                                                return 1;
                                                                            }
                                                                            else if (MVDiff >= 10) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                        else if (tcor >= 93) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (SC >= 2912) {
                                                            if (Rs < 1830) {
                                                                return 4;
                                                            }
                                                            else if (Rs >= 1830) {
                                                                if (Cs < 1565) {
                                                                    if (posBalance < 58) {
                                                                        if (diffAFD < -1) {
                                                                            return 1;
                                                                        }
                                                                        else if (diffAFD >= -1) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                    else if (posBalance >= 58) {
                                                                        return 1;
                                                                    }
                                                                }
                                                                else if (Cs >= 1565) {
                                                                    return 1;
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (mu_mv_mag_sq >= 74) {
                                                        if (refDCval < 53) {
                                                            if (refDCval < 45) {
                                                                return 8;
                                                            }
                                                            else if (refDCval >= 45) {
                                                                if (SC < 2522) {
                                                                    return 1;
                                                                }
                                                                else if (SC >= 2522) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                        else if (refDCval >= 53) {
                                                            if (Rs < 1310) {
                                                                return 2;
                                                            }
                                                            else if (Rs >= 1310) {
                                                                if (SC < 2613) {
                                                                    return 8;
                                                                }
                                                                else if (SC >= 2613) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                            else if (Rs >= 2180) {
                                                if (Cs < 3226) {
                                                    if (ssDCval < 54) {
                                                        if (tcor < 93) {
                                                            return 8;
                                                        }
                                                        else if (tcor >= 93) {
                                                            if (Cs < 3184) {
                                                                return 4;
                                                            }
                                                            else if (Cs >= 3184) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                    else if (ssDCval >= 54) {
                                                        if (ssDCval < 69) {
                                                            if (Rs < 3343) {
                                                                return 2;
                                                            }
                                                            else if (Rs >= 3343) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (ssDCval >= 69) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                                else if (Cs >= 3226) {
                                                    if (TSC < 7) {
                                                        return 4;
                                                    }
                                                    else if (TSC >= 7) {
                                                        return 2;
                                                    }
                                                }
                                            }
                                        }
                                        else if (ssDCval >= 96) {
                                            if (Rs < 627) {
                                                if (ssDCval < 110) {
                                                    if (MVDiff < 10) {
                                                        if (Cs < 574) {
                                                            if (Cs < 480) {
                                                                return 8;
                                                            }
                                                            else if (Cs >= 480) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (Cs >= 574) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (MVDiff >= 10) {
                                                        return 8;
                                                    }
                                                }
                                                else if (ssDCval >= 110) {
                                                    return 4;
                                                }
                                            }
                                            else if (Rs >= 627) {
                                                if (Cs < 392) {
                                                    if (TSC < 6) {
                                                        return 2;
                                                    }
                                                    else if (TSC >= 6) {
                                                        return 1;
                                                    }
                                                }
                                                else if (Cs >= 392) {
                                                    if (SC < 2125) {
                                                        if (diffAFD < -2) {
                                                            return 1;
                                                        }
                                                        else if (diffAFD >= -2) {
                                                            if (SC < 1106) {
                                                                return 8;
                                                            }
                                                            else if (SC >= 1106) {
                                                                if (Rs < 1100) {
                                                                    return 4;
                                                                }
                                                                else if (Rs >= 1100) {
                                                                    if (Rs < 1112) {
                                                                        return 2;
                                                                    }
                                                                    else if (Rs >= 1112) {
                                                                        return 4;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (SC >= 2125) {
                                                        if (SC < 2203) {
                                                            if (Rs < 1058) {
                                                                return 1;
                                                            }
                                                            else if (Rs >= 1058) {
                                                                return 2;
                                                            }
                                                        }
                                                        else if (SC >= 2203) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                            else if (RsCsDiff >= 19) {
                                if (refDCval < 98) {
                                    if (TSC < 9) {
                                        if (ssDCval < 92) {
                                            if (Rs < 1264) {
                                                if (mcTcor < 88) {
                                                    if (negBalance < 35) {
                                                        if (CsDiff < 6) {
                                                            return 4;
                                                        }
                                                        else if (CsDiff >= 6) {
                                                            if (MVDiff < 9) {
                                                                return 4;
                                                            }
                                                            else if (MVDiff >= 9) {
                                                                return 2;
                                                            }
                                                        }
                                                    }
                                                    else if (negBalance >= 35) {
                                                        if (tcor < 64) {
                                                            return 1;
                                                        }
                                                        else if (tcor >= 64) {
                                                            if (refDCval < 63) {
                                                                if (mcTcor < 85) {
                                                                    return 8;
                                                                }
                                                                else if (mcTcor >= 85) {
                                                                    if (RsCsDiff < 38) {
                                                                        return 2;
                                                                    }
                                                                    else if (RsCsDiff >= 38) {
                                                                        return 8;
                                                                    }
                                                                }
                                                            }
                                                            else if (refDCval >= 63) {
                                                                if (Cs < 1164) {
                                                                    return 4;
                                                                }
                                                                else if (Cs >= 1164) {
                                                                    if (SC < 2456) {
                                                                        return 1;
                                                                    }
                                                                    else if (SC >= 2456) {
                                                                        return 2;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (mcTcor >= 88) {
                                                    if (negBalance < 45) {
                                                        if (tcor < 86) {
                                                            if (Rs < 862) {
                                                                return 2;
                                                            }
                                                            else if (Rs >= 862) {
                                                                if (diffMVdiffVal < -2) {
                                                                    return 2;
                                                                }
                                                                else if (diffMVdiffVal >= -2) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                        else if (tcor >= 86) {
                                                            if (negBalance < 40) {
                                                                return 4;
                                                            }
                                                            else if (negBalance >= 40) {
                                                                if (Rs < 1004) {
                                                                    return 8;
                                                                }
                                                                else if (Rs >= 1004) {
                                                                    if (posBalance < 49) {
                                                                        return 8;
                                                                    }
                                                                    else if (posBalance >= 49) {
                                                                        return 1;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (negBalance >= 45) {
                                                        if (ssDCval < 53) {
                                                            if (Rs < 889) {
                                                                return 2;
                                                            }
                                                            else if (Rs >= 889) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (ssDCval >= 53) {
                                                            if (diffRsCsdiff < 10) {
                                                                if (tcor < 85) {
                                                                    if (MVDiff < 17) {
                                                                        return 4;
                                                                    }
                                                                    else if (MVDiff >= 17) {
                                                                        return 2;
                                                                    }
                                                                }
                                                                else if (tcor >= 85) {
                                                                    return 2;
                                                                }
                                                            }
                                                            else if (diffRsCsdiff >= 10) {
                                                                if (MVDiff < 3) {
                                                                    return 2;
                                                                }
                                                                else if (MVDiff >= 3) {
                                                                    if (AFD < 9) {
                                                                        return 8;
                                                                    }
                                                                    else if (AFD >= 9) {
                                                                        return 4;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                            else if (Rs >= 1264) {
                                                if (Cs < 2022) {
                                                    if (ssDCval < 65) {
                                                        if (ssDCval < 57) {
                                                            if (SC < 2308) {
                                                                if (Cs < 553) {
                                                                    return 4;
                                                                }
                                                                else if (Cs >= 553) {
                                                                    return 1;
                                                                }
                                                            }
                                                            else if (SC >= 2308) {
                                                                if (Rs < 1944) {
                                                                    if (Rs < 1414) {
                                                                        if (MVDiff < 10) {
                                                                            return 4;
                                                                        }
                                                                        else if (MVDiff >= 10) {
                                                                            if (negBalance < 42) {
                                                                                if (Rs < 1315) {
                                                                                    return 4;
                                                                                }
                                                                                else if (Rs >= 1315) {
                                                                                    return 1;
                                                                                }
                                                                            }
                                                                            else if (negBalance >= 42) {
                                                                                return 8;
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (Rs >= 1414) {
                                                                        if (mu_mv_mag_sq < 51) {
                                                                            if (posBalance < 57) {
                                                                                if (diffTSC < 1) {
                                                                                    return 8;
                                                                                }
                                                                                else if (diffTSC >= 1) {
                                                                                    return 1;
                                                                                }
                                                                            }
                                                                            else if (posBalance >= 57) {
                                                                                return 1;
                                                                            }
                                                                        }
                                                                        else if (mu_mv_mag_sq >= 51) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                }
                                                                else if (Rs >= 1944) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                        else if (ssDCval >= 57) {
                                                            if (negBalance < 46) {
                                                                return 4;
                                                            }
                                                            else if (negBalance >= 46) {
                                                                if (Cs < 893) {
                                                                    return 1;
                                                                }
                                                                else if (Cs >= 893) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (ssDCval >= 65) {
                                                        if (tcor < 93) {
                                                            if (Cs < 881) {
                                                                return 2;
                                                            }
                                                            else if (Cs >= 881) {
                                                                if (MVDiff < 4) {
                                                                    if (negBalance < 49) {
                                                                        if (SC < 3642) {
                                                                            if (Cs < 1397) {
                                                                                return 1;
                                                                            }
                                                                            else if (Cs >= 1397) {
                                                                                return 2;
                                                                            }
                                                                        }
                                                                        else if (SC >= 3642) {
                                                                            if (RsCsDiff < 86) {
                                                                                return 4;
                                                                            }
                                                                            else if (RsCsDiff >= 86) {
                                                                                return 1;
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (negBalance >= 49) {
                                                                        return 2;
                                                                    }
                                                                }
                                                                else if (MVDiff >= 4) {
                                                                    if (TSC < 7) {
                                                                        if (SC < 4042) {
                                                                            return 2;
                                                                        }
                                                                        else if (SC >= 4042) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                    else if (TSC >= 7) {
                                                                        if (Rs < 1747) {
                                                                            if (mu_mv_mag_sq < 45) {
                                                                                if (Rs < 1686) {
                                                                                    return 4;
                                                                                }
                                                                                else if (Rs >= 1686) {
                                                                                    return 8;
                                                                                }
                                                                            }
                                                                            else if (mu_mv_mag_sq >= 45) {
                                                                                if (AFD < 9) {
                                                                                    return 4;
                                                                                }
                                                                                else if (AFD >= 9) {
                                                                                    return 8;
                                                                                }
                                                                            }
                                                                        }
                                                                        else if (Rs >= 1747) {
                                                                            if (diffRsCsdiff < -1) {
                                                                                if (MVDiff < 16) {
                                                                                    return 4;
                                                                                }
                                                                                else if (MVDiff >= 16) {
                                                                                    return 8;
                                                                                }
                                                                            }
                                                                            else if (diffRsCsdiff >= -1) {
                                                                                if (RsDiff < 11) {
                                                                                    if (mu_mv_mag_sq < 74) {
                                                                                        return 1;
                                                                                    }
                                                                                    else if (mu_mv_mag_sq >= 74) {
                                                                                        if (Rs < 2304) {
                                                                                            return 2;
                                                                                        }
                                                                                        else if (Rs >= 2304) {
                                                                                            return 4;
                                                                                        }
                                                                                    }
                                                                                }
                                                                                else if (RsDiff >= 11) {
                                                                                    return 4;
                                                                                }
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (tcor >= 93) {
                                                            if (Rs < 2544) {
                                                                if (TSC < 7) {
                                                                    if (Rs < 2322) {
                                                                        return 4;
                                                                    }
                                                                    else if (Rs >= 2322) {
                                                                        return 2;
                                                                    }
                                                                }
                                                                else if (TSC >= 7) {
                                                                    if (SC < 4181) {
                                                                        if (Cs < 1757) {
                                                                            if (RsCsDiff < 82) {
                                                                                return 1;
                                                                            }
                                                                            else if (RsCsDiff >= 82) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                        else if (Cs >= 1757) {
                                                                            return 2;
                                                                        }
                                                                    }
                                                                    else if (SC >= 4181) {
                                                                        return 4;
                                                                    }
                                                                }
                                                            }
                                                            else if (Rs >= 2544) {
                                                                return 1;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (Cs >= 2022) {
                                                    if (posBalance < 46) {
                                                        if (negBalance < 27) {
                                                            return 1;
                                                        }
                                                        else if (negBalance >= 27) {
                                                            if (mu_mv_mag_sq < 57) {
                                                                if (tcor < 90) {
                                                                    return 8;
                                                                }
                                                                else if (tcor >= 90) {
                                                                    return 1;
                                                                }
                                                            }
                                                            else if (mu_mv_mag_sq >= 57) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                    else if (posBalance >= 46) {
                                                        if (MVDiff < 9) {
                                                            if (gchDC < 1) {
                                                                if (diffRsCsdiff < -14) {
                                                                    return 1;
                                                                }
                                                                else if (diffRsCsdiff >= -14) {
                                                                    if (negBalance < 31) {
                                                                        return 1;
                                                                    }
                                                                    else if (negBalance >= 31) {
                                                                        return 4;
                                                                    }
                                                                }
                                                            }
                                                            else if (gchDC >= 1) {
                                                                if (negBalance < 39) {
                                                                    if (TSCindex < 9) {
                                                                        return 8;
                                                                    }
                                                                    else if (TSCindex >= 9) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (negBalance >= 39) {
                                                                    if (Cs < 2497) {
                                                                        return 4;
                                                                    }
                                                                    else if (Cs >= 2497) {
                                                                        return 2;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (MVDiff >= 9) {
                                                            if (mcTcor < 95) {
                                                                if (TSCindex < 7) {
                                                                    return 4;
                                                                }
                                                                else if (TSCindex >= 7) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (mcTcor >= 95) {
                                                                return 1;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (ssDCval >= 92) {
                                            if (posBalance < 55) {
                                                if (SC < 4324) {
                                                    return 8;
                                                }
                                                else if (SC >= 4324) {
                                                    if (AFD < 11) {
                                                        return 2;
                                                    }
                                                    else if (AFD >= 11) {
                                                        return 1;
                                                    }
                                                }
                                            }
                                            else if (posBalance >= 55) {
                                                if (Rs < 2001) {
                                                    if (Rs < 1220) {
                                                        if (Cs < 608) {
                                                            return 2;
                                                        }
                                                        else if (Cs >= 608) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (Rs >= 1220) {
                                                        if (tcor < 88) {
                                                            if (RsDiff < 6) {
                                                                if (Rs < 1412) {
                                                                    if (Rs < 1351) {
                                                                        return 8;
                                                                    }
                                                                    else if (Rs >= 1351) {
                                                                        return 2;
                                                                    }
                                                                }
                                                                else if (Rs >= 1412) {
                                                                    if (posBalance < 59) {
                                                                        return 8;
                                                                    }
                                                                    else if (posBalance >= 59) {
                                                                        if (Cs < 1674) {
                                                                            if (diffTSC < 1) {
                                                                                return 2;
                                                                            }
                                                                            else if (diffTSC >= 1) {
                                                                                return 8;
                                                                            }
                                                                        }
                                                                        else if (Cs >= 1674) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                            else if (RsDiff >= 6) {
                                                                return 2;
                                                            }
                                                        }
                                                        else if (tcor >= 88) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                                else if (Rs >= 2001) {
                                                    if (Cs < 1355) {
                                                        if (Cs < 813) {
                                                            return 8;
                                                        }
                                                        else if (Cs >= 813) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (Cs >= 1355) {
                                                        if (diffAFD < 1) {
                                                            return 2;
                                                        }
                                                        else if (diffAFD >= 1) {
                                                            if (Cs < 1853) {
                                                                return 2;
                                                            }
                                                            else if (Cs >= 1853) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (TSC >= 9) {
                                        if (ssDCval < 88) {
                                            if (mu_mv_mag_sq < 71) {
                                                if (negBalance < 49) {
                                                    if (SC < 4609) {
                                                        return 8;
                                                    }
                                                    else if (SC >= 4609) {
                                                        return 1;
                                                    }
                                                }
                                                else if (negBalance >= 49) {
                                                    if (Rs < 3498) {
                                                        return 2;
                                                    }
                                                    else if (Rs >= 3498) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                            else if (mu_mv_mag_sq >= 71) {
                                                return 8;
                                            }
                                        }
                                        else if (ssDCval >= 88) {
                                            return 1;
                                        }
                                    }
                                }
                                else if (refDCval >= 98) {
                                    if (ssDCval < 125) {
                                        if (mu_mv_mag_sq < 48) {
                                            if (refDCval < 108) {
                                                if (SC < 4711) {
                                                    if (RsDiff < 4) {
                                                        return 8;
                                                    }
                                                    else if (RsDiff >= 4) {
                                                        if (diffRsCsdiff < 6) {
                                                            if (Rs < 1328) {
                                                                if (tcor < 70) {
                                                                    if (CsDiff < 3) {
                                                                        return 2;
                                                                    }
                                                                    else if (CsDiff >= 3) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (tcor >= 70) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (Rs >= 1328) {
                                                                if (SC < 2574) {
                                                                    if (RsDiff < 5) {
                                                                        return 2;
                                                                    }
                                                                    else if (RsDiff >= 5) {
                                                                        if (Cs < 862) {
                                                                            return 2;
                                                                        }
                                                                        else if (Cs >= 862) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                }
                                                                else if (SC >= 2574) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                        else if (diffRsCsdiff >= 6) {
                                                            if (tcor < 70) {
                                                                return 8;
                                                            }
                                                            else if (tcor >= 70) {
                                                                if (RsCsDiff < 57) {
                                                                    if (diffTSC < -1) {
                                                                        if (Rs < 1178) {
                                                                            return 4;
                                                                        }
                                                                        else if (Rs >= 1178) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                    else if (diffTSC >= -1) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (RsCsDiff >= 57) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (SC >= 4711) {
                                                    if (ssDCval < 107) {
                                                        if (diffRsCsdiff < -41) {
                                                            return 4;
                                                        }
                                                        else if (diffRsCsdiff >= -41) {
                                                            if (SC < 4807) {
                                                                if (Rs < 3106) {
                                                                    return 8;
                                                                }
                                                                else if (Rs >= 3106) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (SC >= 4807) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                    else if (ssDCval >= 107) {
                                                        if (Rs < 3362) {
                                                            return 4;
                                                        }
                                                        else if (Rs >= 3362) {
                                                            if (posBalance < 56) {
                                                                return 8;
                                                            }
                                                            else if (posBalance >= 56) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                            else if (refDCval >= 108) {
                                                if (Cs < 912) {
                                                    return 8;
                                                }
                                                else if (Cs >= 912) {
                                                    if (SC < 6791) {
                                                        if (Cs < 1939) {
                                                            if (RsCsDiff < 51) {
                                                                if (Cs < 1659) {
                                                                    if (diffMVdiffVal < 4) {
                                                                        return 4;
                                                                    }
                                                                    else if (diffMVdiffVal >= 4) {
                                                                        if (Rs < 3205) {
                                                                            return 4;
                                                                        }
                                                                        else if (Rs >= 3205) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                }
                                                                else if (Cs >= 1659) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (RsCsDiff >= 51) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (Cs >= 1939) {
                                                            if (mcTcor < 87) {
                                                                return 4;
                                                            }
                                                            else if (mcTcor >= 87) {
                                                                if (Rs < 3158) {
                                                                    if (negBalance < 52) {
                                                                        return 8;
                                                                    }
                                                                    else if (negBalance >= 52) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (Rs >= 3158) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (SC >= 6791) {
                                                        if (Rs < 4511) {
                                                            return 1;
                                                        }
                                                        else if (Rs >= 4511) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (mu_mv_mag_sq >= 48) {
                                            if (RsDiff < 6) {
                                                if (Rs < 1535) {
                                                    return 8;
                                                }
                                                else if (Rs >= 1535) {
                                                    if (negBalance < 42) {
                                                        if (diffRsCsdiff < -8) {
                                                            if (posBalance < 44) {
                                                                return 4;
                                                            }
                                                            else if (posBalance >= 44) {
                                                                if (ssDCval < 103) {
                                                                    return 2;
                                                                }
                                                                else if (ssDCval >= 103) {
                                                                    if (RsDiff < 5) {
                                                                        return 8;
                                                                    }
                                                                    else if (RsDiff >= 5) {
                                                                        return 4;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (diffRsCsdiff >= -8) {
                                                            if (Cs < 1341) {
                                                                return 4;
                                                            }
                                                            else if (Cs >= 1341) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                    else if (negBalance >= 42) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                            else if (RsDiff >= 6) {
                                                if (negBalance < 40) {
                                                    if (CsDiff < 7) {
                                                        if (mcTcor < 96) {
                                                            if (AFD < 11) {
                                                                if (MVDiff < 13) {
                                                                    if (RsDiff < 7) {
                                                                        return 8;
                                                                    }
                                                                    else if (RsDiff >= 7) {
                                                                        if (SC < 6910) {
                                                                            if (Cs < 1764) {
                                                                                return 8;
                                                                            }
                                                                            else if (Cs >= 1764) {
                                                                                if (Cs < 2497) {
                                                                                    if (posBalance < 43) {
                                                                                        if (Rs < 4020) {
                                                                                            return 8;
                                                                                        }
                                                                                        else if (Rs >= 4020) {
                                                                                            return 4;
                                                                                        }
                                                                                    }
                                                                                    else if (posBalance >= 43) {
                                                                                        return 4;
                                                                                    }
                                                                                }
                                                                                else if (Cs >= 2497) {
                                                                                    return 8;
                                                                                }
                                                                            }
                                                                        }
                                                                        else if (SC >= 6910) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                }
                                                                else if (MVDiff >= 13) {
                                                                    if (Cs < 2190) {
                                                                        return 4;
                                                                    }
                                                                    else if (Cs >= 2190) {
                                                                        if (ssDCval < 108) {
                                                                            return 8;
                                                                        }
                                                                        else if (ssDCval >= 108) {
                                                                            if (refDCval < 121) {
                                                                                if (TSC < 8) {
                                                                                    return 8;
                                                                                }
                                                                                else if (TSC >= 8) {
                                                                                    return 4;
                                                                                }
                                                                            }
                                                                            else if (refDCval >= 121) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                            else if (AFD >= 11) {
                                                                if (Cs < 3020) {
                                                                    if (Rs < 4725) {
                                                                        return 1;
                                                                    }
                                                                    else if (Rs >= 4725) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (Cs >= 3020) {
                                                                    return 2;
                                                                }
                                                            }
                                                        }
                                                        else if (mcTcor >= 96) {
                                                            if (MVDiff < 10) {
                                                                if (ssDCval < 106) {
                                                                    return 8;
                                                                }
                                                                else if (ssDCval >= 106) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (MVDiff >= 10) {
                                                                if (refDCval < 119) {
                                                                    if (AFD < 7) {
                                                                        return 8;
                                                                    }
                                                                    else if (AFD >= 7) {
                                                                        if (Cs < 2007) {
                                                                            return 4;
                                                                        }
                                                                        else if (Cs >= 2007) {
                                                                            return 2;
                                                                        }
                                                                    }
                                                                }
                                                                else if (refDCval >= 119) {
                                                                    if (Rs < 4366) {
                                                                        return 4;
                                                                    }
                                                                    else if (Rs >= 4366) {
                                                                        return 8;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (CsDiff >= 7) {
                                                        if (SC < 6513) {
                                                            if (posBalance < 53) {
                                                                if (tcor < 93) {
                                                                    if (Rs < 3413) {
                                                                        if (Rs < 2695) {
                                                                            return 4;
                                                                        }
                                                                        else if (Rs >= 2695) {
                                                                            return 2;
                                                                        }
                                                                    }
                                                                    else if (Rs >= 3413) {
                                                                        if (mu_mv_mag_sq < 71) {
                                                                            return 4;
                                                                        }
                                                                        else if (mu_mv_mag_sq >= 71) {
                                                                            if (SC < 6332) {
                                                                                return 2;
                                                                            }
                                                                            else if (SC >= 6332) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                                else if (tcor >= 93) {
                                                                    if (diffRsCsdiff < -9) {
                                                                        if (Cs < 2095) {
                                                                            return 4;
                                                                        }
                                                                        else if (Cs >= 2095) {
                                                                            if (Rs < 4187) {
                                                                                return 2;
                                                                            }
                                                                            else if (Rs >= 4187) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (diffRsCsdiff >= -9) {
                                                                        return 2;
                                                                    }
                                                                }
                                                            }
                                                            else if (posBalance >= 53) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (SC >= 6513) {
                                                            if (tcor < 85) {
                                                                if (diffRsCsdiff < -8) {
                                                                    return 2;
                                                                }
                                                                else if (diffRsCsdiff >= -8) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (tcor >= 85) {
                                                                if (MVDiff < 12) {
                                                                    if (MVDiff < 9) {
                                                                        if (SC < 7203) {
                                                                            return 4;
                                                                        }
                                                                        else if (SC >= 7203) {
                                                                            return 2;
                                                                        }
                                                                    }
                                                                    else if (MVDiff >= 9) {
                                                                        return 8;
                                                                    }
                                                                }
                                                                else if (MVDiff >= 12) {
                                                                    if (ssDCval < 115) {
                                                                        return 4;
                                                                    }
                                                                    else if (ssDCval >= 115) {
                                                                        return 2;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (negBalance >= 40) {
                                                    if (mu_mv_mag_sq < 60) {
                                                        if (RsCsDiff < 99) {
                                                            if (negBalance < 52) {
                                                                if (Rs < 2764) {
                                                                    return 8;
                                                                }
                                                                else if (Rs >= 2764) {
                                                                    if (MVDiff < 14) {
                                                                        return 4;
                                                                    }
                                                                    else if (MVDiff >= 14) {
                                                                        return 8;
                                                                    }
                                                                }
                                                            }
                                                            else if (negBalance >= 52) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (RsCsDiff >= 99) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (mu_mv_mag_sq >= 60) {
                                                        if (negBalance < 51) {
                                                            if (RsDiff < 9) {
                                                                if (posBalance < 59) {
                                                                    return 8;
                                                                }
                                                                else if (posBalance >= 59) {
                                                                    if (Rs < 2062) {
                                                                        if (Rs < 1646) {
                                                                            return 8;
                                                                        }
                                                                        else if (Rs >= 1646) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                    else if (Rs >= 2062) {
                                                                        if (Rs < 2819) {
                                                                            return 8;
                                                                        }
                                                                        else if (Rs >= 2819) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                            else if (RsDiff >= 9) {
                                                                if (diffRsCsdiff < 10) {
                                                                    return 2;
                                                                }
                                                                else if (diffRsCsdiff >= 10) {
                                                                    if (mu_mv_mag_sq < 74) {
                                                                        return 4;
                                                                    }
                                                                    else if (mu_mv_mag_sq >= 74) {
                                                                        return 8;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (negBalance >= 51) {
                                                            if (diffRsCsdiff < 4) {
                                                                if (posBalance < 58) {
                                                                    return 8;
                                                                }
                                                                else if (posBalance >= 58) {
                                                                    if (Rs < 2417) {
                                                                        return 4;
                                                                    }
                                                                    else if (Rs >= 2417) {
                                                                        if (MVDiff < 6) {
                                                                            if (Rs < 2889) {
                                                                                return 8;
                                                                            }
                                                                            else if (Rs >= 2889) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                        else if (MVDiff >= 6) {
                                                                            if (refDCval < 118) {
                                                                                return 4;
                                                                            }
                                                                            else if (refDCval >= 118) {
                                                                                return 8;
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                            else if (diffRsCsdiff >= 4) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (ssDCval >= 125) {
                                        if (posBalance < 44) {
                                            if (posBalance < 37) {
                                                if (diffRsCsdiff < 74) {
                                                    return 4;
                                                }
                                                else if (diffRsCsdiff >= 74) {
                                                    return 1;
                                                }
                                            }
                                            else if (posBalance >= 37) {
                                                if (diffMVdiffVal < -2) {
                                                    if (refDCval < 129) {
                                                        if (mu_mv_mag_sq < 80) {
                                                            return 4;
                                                        }
                                                        else if (mu_mv_mag_sq >= 80) {
                                                            if (Rs < 4657) {
                                                                return 4;
                                                            }
                                                            else if (Rs >= 4657) {
                                                                return 2;
                                                            }
                                                        }
                                                    }
                                                    else if (refDCval >= 129) {
                                                        if (Rs < 4420) {
                                                            if (SC < 4178) {
                                                                return 8;
                                                            }
                                                            else if (SC >= 4178) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (Rs >= 4420) {
                                                            return 2;
                                                        }
                                                    }
                                                }
                                                else if (diffMVdiffVal >= -2) {
                                                    if (Rs < 4618) {
                                                        if (Rs < 3242) {
                                                            return 2;
                                                        }
                                                        else if (Rs >= 3242) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (Rs >= 4618) {
                                                        return 2;
                                                    }
                                                }
                                            }
                                        }
                                        else if (posBalance >= 44) {
                                            if (ssDCval < 144) {
                                                if (ssDCval < 129) {
                                                    if (gchDC < 1) {
                                                        return 4;
                                                    }
                                                    else if (gchDC >= 1) {
                                                        if (Rs < 3057) {
                                                            return 1;
                                                        }
                                                        else if (Rs >= 3057) {
                                                            if (SC < 6747) {
                                                                return 4;
                                                            }
                                                            else if (SC >= 6747) {
                                                                return 2;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (ssDCval >= 129) {
                                                    if (posBalance < 56) {
                                                        if (refDCval < 139) {
                                                            return 1;
                                                        }
                                                        else if (refDCval >= 139) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (posBalance >= 56) {
                                                        if (Rs < 1198) {
                                                            if (Cs < 1109) {
                                                                return 1;
                                                            }
                                                            else if (Cs >= 1109) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (Rs >= 1198) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (ssDCval >= 144) {
                                                return 2;
                                            }
                                        }
                                    }
                                }
                            }
                        }
                        else if (AFD >= 12) {
                            if (ssDCval < 139) {
                                if (tcor < 90) {
                                    if (ssDCval < 128) {
                                        if (Cs < 1420) {
                                            if (RsDiff < 8) {
                                                if (ssDCval < 92) {
                                                    if (mcTcor < 87) {
                                                        if (SC < 2497) {
                                                            if (posBalance < 60) {
                                                                if (mu_mv_mag_sq < 47) {
                                                                    if (ssDCval < 75) {
                                                                        if (TSC < 9) {
                                                                            if (Rs < 731) {
                                                                                return 2;
                                                                            }
                                                                            else if (Rs >= 731) {
                                                                                return 8;
                                                                            }
                                                                        }
                                                                        else if (TSC >= 9) {
                                                                            if (negBalance < 57) {
                                                                                return 4;
                                                                            }
                                                                            else if (negBalance >= 57) {
                                                                                return 8;
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (ssDCval >= 75) {
                                                                        return 1;
                                                                    }
                                                                }
                                                                else if (mu_mv_mag_sq >= 47) {
                                                                    if (refDCval < 74) {
                                                                        return 4;
                                                                    }
                                                                    else if (refDCval >= 74) {
                                                                        return 1;
                                                                    }
                                                                }
                                                            }
                                                            else if (posBalance >= 60) {
                                                                if (diffTSC < 20) {
                                                                    if (refDCval < 51) {
                                                                        if (Cs < 98) {
                                                                            return 2;
                                                                        }
                                                                        else if (Cs >= 98) {
                                                                            if (diffTSC < 6) {
                                                                                return 1;
                                                                            }
                                                                            else if (diffTSC >= 6) {
                                                                                return 2;
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (refDCval >= 51) {
                                                                        if (SC < 2213) {
                                                                            return 1;
                                                                        }
                                                                        else if (SC >= 2213) {
                                                                            if (RsDiff < 6) {
                                                                                return 8;
                                                                            }
                                                                            else if (RsDiff >= 6) {
                                                                                return 1;
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                                else if (diffTSC >= 20) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                        else if (SC >= 2497) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (mcTcor >= 87) {
                                                        if (SC < 1942) {
                                                            if (SC < 1922) {
                                                                if (Cs < 44) {
                                                                    return 1;
                                                                }
                                                                else if (Cs >= 44) {
                                                                    if (negBalance < 39) {
                                                                        if (RsDiff < 5) {
                                                                            return 4;
                                                                        }
                                                                        else if (RsDiff >= 5) {
                                                                            return 2;
                                                                        }
                                                                    }
                                                                    else if (negBalance >= 39) {
                                                                        return 2;
                                                                    }
                                                                }
                                                            }
                                                            else if (SC >= 1922) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (SC >= 1942) {
                                                            if (posBalance < 55) {
                                                                return 2;
                                                            }
                                                            else if (posBalance >= 55) {
                                                                if (MVDiff < 19) {
                                                                    return 1;
                                                                }
                                                                else if (MVDiff >= 19) {
                                                                    if (RsDiff < 7) {
                                                                        return 1;
                                                                    }
                                                                    else if (RsDiff >= 7) {
                                                                        return 2;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (ssDCval >= 92) {
                                                    if (Cs < 311) {
                                                        return 1;
                                                    }
                                                    else if (Cs >= 311) {
                                                        if (Rs < 1310) {
                                                            if (Cs < 609) {
                                                                if (AFD < 20) {
                                                                    return 2;
                                                                }
                                                                else if (AFD >= 20) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (Cs >= 609) {
                                                                if (ssDCval < 118) {
                                                                    return 4;
                                                                }
                                                                else if (ssDCval >= 118) {
                                                                    return 1;
                                                                }
                                                            }
                                                        }
                                                        else if (Rs >= 1310) {
                                                            if (Rs < 2040) {
                                                                if (ssDCval < 117) {
                                                                    if (SC < 2170) {
                                                                        return 4;
                                                                    }
                                                                    else if (SC >= 2170) {
                                                                        return 2;
                                                                    }
                                                                }
                                                                else if (ssDCval >= 117) {
                                                                    return 1;
                                                                }
                                                            }
                                                            else if (Rs >= 2040) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                            else if (RsDiff >= 8) {
                                                if (tcor < 75) {
                                                    if (diffRsCsdiff < 290) {
                                                        if (refDCval < 98) {
                                                            if (tcor < 33) {
                                                                return 4;
                                                            }
                                                            else if (tcor >= 33) {
                                                                if (RsDiff < 9) {
                                                                    if (mcTcor < 88) {
                                                                        if (SC < 2269) {
                                                                            return 1;
                                                                        }
                                                                        else if (SC >= 2269) {
                                                                            if (Cs < 1289) {
                                                                                return 8;
                                                                            }
                                                                            else if (Cs >= 1289) {
                                                                                return 1;
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (mcTcor >= 88) {
                                                                        return 2;
                                                                    }
                                                                }
                                                                else if (RsDiff >= 9) {
                                                                    if (ssDCval < 98) {
                                                                        if (ssDCval < 51) {
                                                                            if (Rs < 1570) {
                                                                                return 2;
                                                                            }
                                                                            else if (Rs >= 1570) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                        else if (ssDCval >= 51) {
                                                                            return 2;
                                                                        }
                                                                    }
                                                                    else if (ssDCval >= 98) {
                                                                        if (Cs < 646) {
                                                                            return 2;
                                                                        }
                                                                        else if (Cs >= 646) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (refDCval >= 98) {
                                                            if (ssDCval < 105) {
                                                                if (RsDiff < 25) {
                                                                    if (diffRsCsdiff < -94) {
                                                                        if (Rs < 2688) {
                                                                            return 2;
                                                                        }
                                                                        else if (Rs >= 2688) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                    else if (diffRsCsdiff >= -94) {
                                                                        return 1;
                                                                    }
                                                                }
                                                                else if (RsDiff >= 25) {
                                                                    return 2;
                                                                }
                                                            }
                                                            else if (ssDCval >= 105) {
                                                                if (Cs < 1352) {
                                                                    if (diffMVdiffVal < -2) {
                                                                        if (diffTSC < -2) {
                                                                            return 2;
                                                                        }
                                                                        else if (diffTSC >= -2) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                    else if (diffMVdiffVal >= -2) {
                                                                        return 2;
                                                                    }
                                                                }
                                                                else if (Cs >= 1352) {
                                                                    return 1;
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (diffRsCsdiff >= 290) {
                                                        if (diffTSC < 19) {
                                                            return 8;
                                                        }
                                                        else if (diffTSC >= 19) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                                else if (tcor >= 75) {
                                                    if (refDCval < 84) {
                                                        if (TSC < 8) {
                                                            return 2;
                                                        }
                                                        else if (TSC >= 8) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (refDCval >= 84) {
                                                        if (SC < 3894) {
                                                            return 2;
                                                        }
                                                        else if (SC >= 3894) {
                                                            if (Rs < 2991) {
                                                                return 8;
                                                            }
                                                            else if (Rs >= 2991) {
                                                                return 2;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (Cs >= 1420) {
                                            if (refDCval < 84) {
                                                if (negBalance < 44) {
                                                    if (TSC < 9) {
                                                        if (RsCsDiff < 87) {
                                                            if (Cs < 1502) {
                                                                return 8;
                                                            }
                                                            else if (Cs >= 1502) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (RsCsDiff >= 87) {
                                                            if (ssDCval < 56) {
                                                                if (Cs < 2537) {
                                                                    return 2;
                                                                }
                                                                else if (Cs >= 2537) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (ssDCval >= 56) {
                                                                return 1;
                                                            }
                                                        }
                                                    }
                                                    else if (TSC >= 9) {
                                                        if (refDCval < 73) {
                                                            if (negBalance < 33) {
                                                                if (Rs < 1752) {
                                                                    return 2;
                                                                }
                                                                else if (Rs >= 1752) {
                                                                    if (Cs < 2289) {
                                                                        return 4;
                                                                    }
                                                                    else if (Cs >= 2289) {
                                                                        return 8;
                                                                    }
                                                                }
                                                            }
                                                            else if (negBalance >= 33) {
                                                                return 1;
                                                            }
                                                        }
                                                        else if (refDCval >= 73) {
                                                            if (SC < 4114) {
                                                                return 1;
                                                            }
                                                            else if (SC >= 4114) {
                                                                if (Cs < 2666) {
                                                                    return 8;
                                                                }
                                                                else if (Cs >= 2666) {
                                                                    if (AFD < 18) {
                                                                        if (tcor < 87) {
                                                                            if (diffTSC < 1) {
                                                                                return 8;
                                                                            }
                                                                            else if (diffTSC >= 1) {
                                                                                if (SC < 5655) {
                                                                                    return 8;
                                                                                }
                                                                                else if (SC >= 5655) {
                                                                                    return 1;
                                                                                }
                                                                            }
                                                                        }
                                                                        else if (tcor >= 87) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                    else if (AFD >= 18) {
                                                                        return 1;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (negBalance >= 44) {
                                                    if (mu_mv_mag_sq < 3) {
                                                        if (MVDiff < 23) {
                                                            if (MVDiff < 17) {
                                                                if (Cs < 1993) {
                                                                    return 2;
                                                                }
                                                                else if (Cs >= 1993) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (MVDiff >= 17) {
                                                                if (posBalance < 54) {
                                                                    return 8;
                                                                }
                                                                else if (posBalance >= 54) {
                                                                    if (ssDCval < 72) {
                                                                        if (negBalance < 51) {
                                                                            return 4;
                                                                        }
                                                                        else if (negBalance >= 51) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                    else if (ssDCval >= 72) {
                                                                        if (Rs < 2466) {
                                                                            return 4;
                                                                        }
                                                                        else if (Rs >= 2466) {
                                                                            if (Rs < 2477) {
                                                                                return 8;
                                                                            }
                                                                            else if (Rs >= 2477) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (MVDiff >= 23) {
                                                            if (posBalance < 52) {
                                                                return 1;
                                                            }
                                                            else if (posBalance >= 52) {
                                                                if (negBalance < 55) {
                                                                    if (diffTSC < 1) {
                                                                        return 8;
                                                                    }
                                                                    else if (diffTSC >= 1) {
                                                                        if (SC < 5234) {
                                                                            return 4;
                                                                        }
                                                                        else if (SC >= 5234) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                }
                                                                else if (negBalance >= 55) {
                                                                    if (TSC < 17) {
                                                                        return 2;
                                                                    }
                                                                    else if (TSC >= 17) {
                                                                        return 4;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (mu_mv_mag_sq >= 3) {
                                                        if (RsCsDiff < 57) {
                                                            return 8;
                                                        }
                                                        else if (RsCsDiff >= 57) {
                                                            if (refDCval < 78) {
                                                                if (AFD < 22) {
                                                                    if (Rs < 3502) {
                                                                        if (Cs < 1988) {
                                                                            if (RsDiff < 8) {
                                                                                if (mcTcor < 87) {
                                                                                    return 8;
                                                                                }
                                                                                else if (mcTcor >= 87) {
                                                                                    if (AFD < 14) {
                                                                                        return 8;
                                                                                    }
                                                                                    else if (AFD >= 14) {
                                                                                        return 4;
                                                                                    }
                                                                                }
                                                                            }
                                                                            else if (RsDiff >= 8) {
                                                                                if (RsDiff < 11) {
                                                                                    return 4;
                                                                                }
                                                                                else if (RsDiff >= 11) {
                                                                                    return 1;
                                                                                }
                                                                            }
                                                                        }
                                                                        else if (Cs >= 1988) {
                                                                            if (negBalance < 50) {
                                                                                if (Rs < 2728) {
                                                                                    return 8;
                                                                                }
                                                                                else if (Rs >= 2728) {
                                                                                    return 1;
                                                                                }
                                                                            }
                                                                            else if (negBalance >= 50) {
                                                                                return 8;
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (Rs >= 3502) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (AFD >= 22) {
                                                                    if (RsDiff < 20) {
                                                                        if (MVDiff < 48) {
                                                                            if (Rs < 2840) {
                                                                                if (Rs < 2210) {
                                                                                    return 4;
                                                                                }
                                                                                else if (Rs >= 2210) {
                                                                                    return 8;
                                                                                }
                                                                            }
                                                                            else if (Rs >= 2840) {
                                                                                if (Rs < 2936) {
                                                                                    return 2;
                                                                                }
                                                                                else if (Rs >= 2936) {
                                                                                    return 1;
                                                                                }
                                                                            }
                                                                        }
                                                                        else if (MVDiff >= 48) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                    else if (RsDiff >= 20) {
                                                                        return 2;
                                                                    }
                                                                }
                                                            }
                                                            else if (refDCval >= 78) {
                                                                if (posBalance < 57) {
                                                                    if (Rs < 1950) {
                                                                        return 1;
                                                                    }
                                                                    else if (Rs >= 1950) {
                                                                        if (gchDC < 2) {
                                                                            if (negBalance < 52) {
                                                                                if (Rs < 2571) {
                                                                                    if (TSC < 10) {
                                                                                        return 8;
                                                                                    }
                                                                                    else if (TSC >= 10) {
                                                                                        return 1;
                                                                                    }
                                                                                }
                                                                                else if (Rs >= 2571) {
                                                                                    return 8;
                                                                                }
                                                                            }
                                                                            else if (negBalance >= 52) {
                                                                                if (diffRsCsdiff < 2) {
                                                                                    return 1;
                                                                                }
                                                                                else if (diffRsCsdiff >= 2) {
                                                                                    if (AFD < 17) {
                                                                                        return 4;
                                                                                    }
                                                                                    else if (AFD >= 17) {
                                                                                        return 8;
                                                                                    }
                                                                                }
                                                                            }
                                                                        }
                                                                        else if (gchDC >= 2) {
                                                                            if (Cs < 2193) {
                                                                                if (Rs < 2320) {
                                                                                    return 4;
                                                                                }
                                                                                else if (Rs >= 2320) {
                                                                                    return 2;
                                                                                }
                                                                            }
                                                                            else if (Cs >= 2193) {
                                                                                return 1;
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                                else if (posBalance >= 57) {
                                                                    if (mu_mv_mag_sq < 11) {
                                                                        if (RsDiff < 8) {
                                                                            return 2;
                                                                        }
                                                                        else if (RsDiff >= 8) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                    else if (mu_mv_mag_sq >= 11) {
                                                                        if (mcTcor < 92) {
                                                                            if (mu_mv_mag_sq < 35) {
                                                                                if (MVDiff < 18) {
                                                                                    return 2;
                                                                                }
                                                                                else if (MVDiff >= 18) {
                                                                                    return 1;
                                                                                }
                                                                            }
                                                                            else if (mu_mv_mag_sq >= 35) {
                                                                                return 1;
                                                                            }
                                                                        }
                                                                        else if (mcTcor >= 92) {
                                                                            if (SC < 4232) {
                                                                                return 1;
                                                                            }
                                                                            else if (SC >= 4232) {
                                                                                if (Rs < 2907) {
                                                                                    if (Rs < 2459) {
                                                                                        return 1;
                                                                                    }
                                                                                    else if (Rs >= 2459) {
                                                                                        return 2;
                                                                                    }
                                                                                }
                                                                                else if (Rs >= 2907) {
                                                                                    return 1;
                                                                                }
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                            else if (refDCval >= 84) {
                                                if (Cs < 1761) {
                                                    if (AFD < 19) {
                                                        if (refDCval < 116) {
                                                            if (RsCsDiff < 92) {
                                                                if (CsDiff < 7) {
                                                                    return 4;
                                                                }
                                                                else if (CsDiff >= 7) {
                                                                    if (Cs < 1504) {
                                                                        return 1;
                                                                    }
                                                                    else if (Cs >= 1504) {
                                                                        return 8;
                                                                    }
                                                                }
                                                            }
                                                            else if (RsCsDiff >= 92) {
                                                                if (CsDiff < 9) {
                                                                    if (diffAFD < 1) {
                                                                        if (Rs < 3175) {
                                                                            return 2;
                                                                        }
                                                                        else if (Rs >= 3175) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                    else if (diffAFD >= 1) {
                                                                        return 8;
                                                                    }
                                                                }
                                                                else if (CsDiff >= 9) {
                                                                    return 2;
                                                                }
                                                            }
                                                        }
                                                        else if (refDCval >= 116) {
                                                            if (tcor < 86) {
                                                                return 4;
                                                            }
                                                            else if (tcor >= 86) {
                                                                if (Rs < 2413) {
                                                                    if (posBalance < 58) {
                                                                        return 8;
                                                                    }
                                                                    else if (posBalance >= 58) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (Rs >= 2413) {
                                                                    return 2;
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (AFD >= 19) {
                                                        if (diffMVdiffVal < 17) {
                                                            if (CsDiff < 7) {
                                                                if (Rs < 2280) {
                                                                    return 8;
                                                                }
                                                                else if (Rs >= 2280) {
                                                                    return 1;
                                                                }
                                                            }
                                                            else if (CsDiff >= 7) {
                                                                if (diffMVdiffVal < -17) {
                                                                    return 4;
                                                                }
                                                                else if (diffMVdiffVal >= -17) {
                                                                    return 2;
                                                                }
                                                            }
                                                        }
                                                        else if (diffMVdiffVal >= 17) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                                else if (Cs >= 1761) {
                                                    if (MVDiff < 5) {
                                                        if (Cs < 2427) {
                                                            if (negBalance < 54) {
                                                                if (SC < 4970) {
                                                                    return 4;
                                                                }
                                                                else if (SC >= 4970) {
                                                                    if (Rs < 2906) {
                                                                        return 8;
                                                                    }
                                                                    else if (Rs >= 2906) {
                                                                        if (Cs < 2155) {
                                                                            return 2;
                                                                        }
                                                                        else if (Cs >= 2155) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                            else if (negBalance >= 54) {
                                                                return 2;
                                                            }
                                                        }
                                                        else if (Cs >= 2427) {
                                                            if (Rs < 2698) {
                                                                if (tcor < 87) {
                                                                    return 4;
                                                                }
                                                                else if (tcor >= 87) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (Rs >= 2698) {
                                                                if (SC < 6622) {
                                                                    return 8;
                                                                }
                                                                else if (SC >= 6622) {
                                                                    if (refDCval < 95) {
                                                                        return 8;
                                                                    }
                                                                    else if (refDCval >= 95) {
                                                                        if (refDCval < 106) {
                                                                            return 4;
                                                                        }
                                                                        else if (refDCval >= 106) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (MVDiff >= 5) {
                                                        if (mcTcor < 94) {
                                                            if (AFD < 18) {
                                                                if (SC < 7748) {
                                                                    if (posBalance < 55) {
                                                                        if (negBalance < 40) {
                                                                            if (Cs < 2379) {
                                                                                if (RsDiff < 9) {
                                                                                    return 1;
                                                                                }
                                                                                else if (RsDiff >= 9) {
                                                                                    return 8;
                                                                                }
                                                                            }
                                                                            else if (Cs >= 2379) {
                                                                                if (Rs < 4633) {
                                                                                    return 1;
                                                                                }
                                                                                else if (Rs >= 4633) {
                                                                                    return 2;
                                                                                }
                                                                            }
                                                                        }
                                                                        else if (negBalance >= 40) {
                                                                            if (SC < 5272) {
                                                                                if (MVDiff < 19) {
                                                                                    if (negBalance < 52) {
                                                                                        return 1;
                                                                                    }
                                                                                    else if (negBalance >= 52) {
                                                                                        if (CsDiff < 7) {
                                                                                            return 1;
                                                                                        }
                                                                                        else if (CsDiff >= 7) {
                                                                                            return 2;
                                                                                        }
                                                                                    }
                                                                                }
                                                                                else if (MVDiff >= 19) {
                                                                                    if (TSC < 12) {
                                                                                        if (ssDCval < 94) {
                                                                                            if (RsDiff < 9) {
                                                                                                return 4;
                                                                                            }
                                                                                            else if (RsDiff >= 9) {
                                                                                                return 1;
                                                                                            }
                                                                                        }
                                                                                        else if (ssDCval >= 94) {
                                                                                            if (ssDCval < 101) {
                                                                                                if (Rs < 2335) {
                                                                                                    return 4;
                                                                                                }
                                                                                                else if (Rs >= 2335) {
                                                                                                    return 1;
                                                                                                }
                                                                                            }
                                                                                            else if (ssDCval >= 101) {
                                                                                                return 4;
                                                                                            }
                                                                                        }
                                                                                    }
                                                                                    else if (TSC >= 12) {
                                                                                        if (negBalance < 46) {
                                                                                            return 8;
                                                                                        }
                                                                                        else if (negBalance >= 46) {
                                                                                            if (refDCval < 87) {
                                                                                                return 8;
                                                                                            }
                                                                                            else if (refDCval >= 87) {
                                                                                                return 4;
                                                                                            }
                                                                                        }
                                                                                    }
                                                                                }
                                                                            }
                                                                            else if (SC >= 5272) {
                                                                                if (posBalance < 50) {
                                                                                    if (Rs < 4176) {
                                                                                        return 8;
                                                                                    }
                                                                                    else if (Rs >= 4176) {
                                                                                        return 4;
                                                                                    }
                                                                                }
                                                                                else if (posBalance >= 50) {
                                                                                    if (RsDiff < 13) {
                                                                                        if (diffRsCsdiff < -15) {
                                                                                            if (TSC < 10) {
                                                                                                return 4;
                                                                                            }
                                                                                            else if (TSC >= 10) {
                                                                                                return 2;
                                                                                            }
                                                                                        }
                                                                                        else if (diffRsCsdiff >= -15) {
                                                                                            return 4;
                                                                                        }
                                                                                    }
                                                                                    else if (RsDiff >= 13) {
                                                                                        return 8;
                                                                                    }
                                                                                }
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (posBalance >= 55) {
                                                                        if (Cs < 2380) {
                                                                            if (MVDiff < 16) {
                                                                                if (refDCval < 116) {
                                                                                    if (mu_mv_mag_sq < 45) {
                                                                                        if (mcTcor < 87) {
                                                                                            return 4;
                                                                                        }
                                                                                        else if (mcTcor >= 87) {
                                                                                            if (MVDiff < 6) {
                                                                                                if (Cs < 2288) {
                                                                                                    return 1;
                                                                                                }
                                                                                                else if (Cs >= 2288) {
                                                                                                    return 2;
                                                                                                }
                                                                                            }
                                                                                            else if (MVDiff >= 6) {
                                                                                                if (negBalance < 43) {
                                                                                                    return 4;
                                                                                                }
                                                                                                else if (negBalance >= 43) {
                                                                                                    if (CsDiff < 9) {
                                                                                                        return 2;
                                                                                                    }
                                                                                                    else if (CsDiff >= 9) {
                                                                                                        if (mcTcor < 92) {
                                                                                                            return 2;
                                                                                                        }
                                                                                                        else if (mcTcor >= 92) {
                                                                                                            if (TSC < 12) {
                                                                                                                return 4;
                                                                                                            }
                                                                                                            else if (TSC >= 12) {
                                                                                                                return 2;
                                                                                                            }
                                                                                                        }
                                                                                                    }
                                                                                                }
                                                                                            }
                                                                                        }
                                                                                    }
                                                                                    else if (mu_mv_mag_sq >= 45) {
                                                                                        if (MVDiff < 10) {
                                                                                            if (RsCsDiff < 106) {
                                                                                                return 1;
                                                                                            }
                                                                                            else if (RsCsDiff >= 106) {
                                                                                                return 4;
                                                                                            }
                                                                                        }
                                                                                        else if (MVDiff >= 10) {
                                                                                            if (RsDiff < 10) {
                                                                                                return 1;
                                                                                            }
                                                                                            else if (RsDiff >= 10) {
                                                                                                return 4;
                                                                                            }
                                                                                        }
                                                                                    }
                                                                                }
                                                                                else if (refDCval >= 116) {
                                                                                    if (CsDiff < 9) {
                                                                                        return 4;
                                                                                    }
                                                                                    else if (CsDiff >= 9) {
                                                                                        return 1;
                                                                                    }
                                                                                }
                                                                            }
                                                                            else if (MVDiff >= 16) {
                                                                                if (Rs < 2585) {
                                                                                    if (diffTSC < -1) {
                                                                                        return 1;
                                                                                    }
                                                                                    else if (diffTSC >= -1) {
                                                                                        return 4;
                                                                                    }
                                                                                }
                                                                                else if (Rs >= 2585) {
                                                                                    if (CsDiff < 8) {
                                                                                        if (SC < 4757) {
                                                                                            if (Rs < 2618) {
                                                                                                return 2;
                                                                                            }
                                                                                            else if (Rs >= 2618) {
                                                                                                return 4;
                                                                                            }
                                                                                        }
                                                                                        else if (SC >= 4757) {
                                                                                            return 8;
                                                                                        }
                                                                                    }
                                                                                    else if (CsDiff >= 8) {
                                                                                        if (Rs < 3082) {
                                                                                            return 1;
                                                                                        }
                                                                                        else if (Rs >= 3082) {
                                                                                            return 2;
                                                                                        }
                                                                                    }
                                                                                }
                                                                            }
                                                                        }
                                                                        else if (Cs >= 2380) {
                                                                            if (Rs < 2971) {
                                                                                if (negBalance < 46) {
                                                                                    return 8;
                                                                                }
                                                                                else if (negBalance >= 46) {
                                                                                    return 4;
                                                                                }
                                                                            }
                                                                            else if (Rs >= 2971) {
                                                                                if (ssDCval < 96) {
                                                                                    if (RsCsDiff < 236) {
                                                                                        if (refDCval < 87) {
                                                                                            if (mcTcor < 92) {
                                                                                                return 1;
                                                                                            }
                                                                                            else if (mcTcor >= 92) {
                                                                                                if (Rs < 3132) {
                                                                                                    return 2;
                                                                                                }
                                                                                                else if (Rs >= 3132) {
                                                                                                    return 4;
                                                                                                }
                                                                                            }
                                                                                        }
                                                                                        else if (refDCval >= 87) {
                                                                                            return 1;
                                                                                        }
                                                                                    }
                                                                                    else if (RsCsDiff >= 236) {
                                                                                        return 4;
                                                                                    }
                                                                                }
                                                                                else if (ssDCval >= 96) {
                                                                                    if (CsDiff < 11) {
                                                                                        if (SC < 6027) {
                                                                                            if (Rs < 3116) {
                                                                                                return 4;
                                                                                            }
                                                                                            else if (Rs >= 3116) {
                                                                                                return 1;
                                                                                            }
                                                                                        }
                                                                                        else if (SC >= 6027) {
                                                                                            if (refDCval < 105) {
                                                                                                return 4;
                                                                                            }
                                                                                            else if (refDCval >= 105) {
                                                                                                if (ssDCval < 117) {
                                                                                                    return 2;
                                                                                                }
                                                                                                else if (ssDCval >= 117) {
                                                                                                    return 4;
                                                                                                }
                                                                                            }
                                                                                        }
                                                                                    }
                                                                                    else if (CsDiff >= 11) {
                                                                                        if (Rs < 4385) {
                                                                                            return 2;
                                                                                        }
                                                                                        else if (Rs >= 4385) {
                                                                                            return 1;
                                                                                        }
                                                                                    }
                                                                                }
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                                else if (SC >= 7748) {
                                                                    if (ssDCval < 106) {
                                                                        return 4;
                                                                    }
                                                                    else if (ssDCval >= 106) {
                                                                        if (MVDiff < 24) {
                                                                            if (TSC < 11) {
                                                                                return 4;
                                                                            }
                                                                            else if (TSC >= 11) {
                                                                                return 2;
                                                                            }
                                                                        }
                                                                        else if (MVDiff >= 24) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                            else if (AFD >= 18) {
                                                                if (MVDiff < 65) {
                                                                    if (Cs < 2925) {
                                                                        if (MVDiff < 28) {
                                                                            if (TSC < 12) {
                                                                                if (RsDiff < 14) {
                                                                                    if (CsDiff < 10) {
                                                                                        return 1;
                                                                                    }
                                                                                    else if (CsDiff >= 10) {
                                                                                        return 4;
                                                                                    }
                                                                                }
                                                                                else if (RsDiff >= 14) {
                                                                                    if (Rs < 4181) {
                                                                                        if (Cs < 2639) {
                                                                                            return 2;
                                                                                        }
                                                                                        else if (Cs >= 2639) {
                                                                                            return 1;
                                                                                        }
                                                                                    }
                                                                                    else if (Rs >= 4181) {
                                                                                        return 1;
                                                                                    }
                                                                                }
                                                                            }
                                                                            else if (TSC >= 12) {
                                                                                if (MVDiff < 7) {
                                                                                    if (diffMVdiffVal < -3) {
                                                                                        return 1;
                                                                                    }
                                                                                    else if (diffMVdiffVal >= -3) {
                                                                                        return 2;
                                                                                    }
                                                                                }
                                                                                else if (MVDiff >= 7) {
                                                                                    if (mcTcor < 87) {
                                                                                        if (tcor < 71) {
                                                                                            return 1;
                                                                                        }
                                                                                        else if (tcor >= 71) {
                                                                                            if (RsCsDiff < 198) {
                                                                                                return 1;
                                                                                            }
                                                                                            else if (RsCsDiff >= 198) {
                                                                                                if (refDCval < 126) {
                                                                                                    return 4;
                                                                                                }
                                                                                                else if (refDCval >= 126) {
                                                                                                    return 1;
                                                                                                }
                                                                                            }
                                                                                        }
                                                                                    }
                                                                                    else if (mcTcor >= 87) {
                                                                                        if (AFD < 32) {
                                                                                            if (CsDiff < 10) {
                                                                                                if (RsCsDiff < 226) {
                                                                                                    if (diffRsCsdiff < 45) {
                                                                                                        return 1;
                                                                                                    }
                                                                                                    else if (diffRsCsdiff >= 45) {
                                                                                                        return 2;
                                                                                                    }
                                                                                                }
                                                                                                else if (RsCsDiff >= 226) {
                                                                                                    return 2;
                                                                                                }
                                                                                            }
                                                                                            else if (CsDiff >= 10) {
                                                                                                if (tcor < 85) {
                                                                                                    return 1;
                                                                                                }
                                                                                                else if (tcor >= 85) {
                                                                                                    if (SC < 4851) {
                                                                                                        return 1;
                                                                                                    }
                                                                                                    else if (SC >= 4851) {
                                                                                                        return 4;
                                                                                                    }
                                                                                                }
                                                                                            }
                                                                                        }
                                                                                        else if (AFD >= 32) {
                                                                                            return 2;
                                                                                        }
                                                                                    }
                                                                                }
                                                                            }
                                                                        }
                                                                        else if (MVDiff >= 28) {
                                                                            if (tcor < 62) {
                                                                                if (RsDiff < 15) {
                                                                                    if (AFD < 30) {
                                                                                        return 4;
                                                                                    }
                                                                                    else if (AFD >= 30) {
                                                                                        return 1;
                                                                                    }
                                                                                }
                                                                                else if (RsDiff >= 15) {
                                                                                    if (SC < 5006) {
                                                                                        return 2;
                                                                                    }
                                                                                    else if (SC >= 5006) {
                                                                                        return 8;
                                                                                    }
                                                                                }
                                                                            }
                                                                            else if (tcor >= 62) {
                                                                                if (Rs < 3745) {
                                                                                    if (diffMVdiffVal < -23) {
                                                                                        return 8;
                                                                                    }
                                                                                    else if (diffMVdiffVal >= -23) {
                                                                                        if (AFD < 19) {
                                                                                            if (diffAFD < -1) {
                                                                                                return 1;
                                                                                            }
                                                                                            else if (diffAFD >= -1) {
                                                                                                return 4;
                                                                                            }
                                                                                        }
                                                                                        else if (AFD >= 19) {
                                                                                            if (refDCval < 125) {
                                                                                                if (diffRsCsdiff < 227) {
                                                                                                    if (diffMVdiffVal < 12) {
                                                                                                        return 1;
                                                                                                    }
                                                                                                    else if (diffMVdiffVal >= 12) {
                                                                                                        if (TSC < 12) {
                                                                                                            return 4;
                                                                                                        }
                                                                                                        else if (TSC >= 12) {
                                                                                                            return 1;
                                                                                                        }
                                                                                                    }
                                                                                                }
                                                                                                else if (diffRsCsdiff >= 227) {
                                                                                                    return 4;
                                                                                                }
                                                                                            }
                                                                                            else if (refDCval >= 125) {
                                                                                                return 4;
                                                                                            }
                                                                                        }
                                                                                    }
                                                                                }
                                                                                else if (Rs >= 3745) {
                                                                                    return 4;
                                                                                }
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (Cs >= 2925) {
                                                                        if (AFD < 28) {
                                                                            if (negBalance < 54) {
                                                                                if (SC < 6670) {
                                                                                    if (Cs < 3140) {
                                                                                        return 1;
                                                                                    }
                                                                                    else if (Cs >= 3140) {
                                                                                        return 8;
                                                                                    }
                                                                                }
                                                                                else if (SC >= 6670) {
                                                                                    return 1;
                                                                                }
                                                                            }
                                                                            else if (negBalance >= 54) {
                                                                                if (AFD < 22) {
                                                                                    if (Rs < 3623) {
                                                                                        return 8;
                                                                                    }
                                                                                    else if (Rs >= 3623) {
                                                                                        return 4;
                                                                                    }
                                                                                }
                                                                                else if (AFD >= 22) {
                                                                                    return 8;
                                                                                }
                                                                            }
                                                                        }
                                                                        else if (AFD >= 28) {
                                                                            return 2;
                                                                        }
                                                                    }
                                                                }
                                                                else if (MVDiff >= 65) {
                                                                    if (Rs < 1762) {
                                                                        return 1;
                                                                    }
                                                                    else if (Rs >= 1762) {
                                                                        if (diffMVdiffVal < 97) {
                                                                            if (posBalance < 80) {
                                                                                return 8;
                                                                            }
                                                                            else if (posBalance >= 80) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                        else if (diffMVdiffVal >= 97) {
                                                                            if (Cs < 2679) {
                                                                                return 4;
                                                                            }
                                                                            else if (Cs >= 2679) {
                                                                                return 8;
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (mcTcor >= 94) {
                                                            if (SC < 5709) {
                                                                if (tcor < 89) {
                                                                    if (posBalance < 42) {
                                                                        if (SC < 5227) {
                                                                            return 2;
                                                                        }
                                                                        else if (SC >= 5227) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                    else if (posBalance >= 42) {
                                                                        return 2;
                                                                    }
                                                                }
                                                                else if (tcor >= 89) {
                                                                    if (Cs < 1961) {
                                                                        return 1;
                                                                    }
                                                                    else if (Cs >= 1961) {
                                                                        if (gchDC < 1) {
                                                                            return 8;
                                                                        }
                                                                        else if (gchDC >= 1) {
                                                                            return 2;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                            else if (SC >= 5709) {
                                                                if (RsDiff < 8) {
                                                                    return 1;
                                                                }
                                                                else if (RsDiff >= 8) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (ssDCval >= 128) {
                                        if (posBalance < 61) {
                                            if (tcor < 80) {
                                                if (MVDiff < 29) {
                                                    if (posBalance < 47) {
                                                        if (Rs < 2496) {
                                                            return 4;
                                                        }
                                                        else if (Rs >= 2496) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (posBalance >= 47) {
                                                        return 1;
                                                    }
                                                }
                                                else if (MVDiff >= 29) {
                                                    if (Cs < 1628) {
                                                        if (CsDiff < 9) {
                                                            return 2;
                                                        }
                                                        else if (CsDiff >= 9) {
                                                            return 1;
                                                        }
                                                    }
                                                    else if (Cs >= 1628) {
                                                        if (tcor < 66) {
                                                            if (Rs < 3040) {
                                                                return 8;
                                                            }
                                                            else if (Rs >= 3040) {
                                                                return 1;
                                                            }
                                                        }
                                                        else if (tcor >= 66) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (tcor >= 80) {
                                                if (negBalance < 55) {
                                                    if (ssDCval < 130) {
                                                        if (SC < 8557) {
                                                            if (Rs < 3689) {
                                                                return 2;
                                                            }
                                                            else if (Rs >= 3689) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (SC >= 8557) {
                                                            return 1;
                                                        }
                                                    }
                                                    else if (ssDCval >= 130) {
                                                        if (Rs < 2407) {
                                                            return 2;
                                                        }
                                                        else if (Rs >= 2407) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                                else if (negBalance >= 55) {
                                                    if (Rs < 1971) {
                                                        return 1;
                                                    }
                                                    else if (Rs >= 1971) {
                                                        return 2;
                                                    }
                                                }
                                            }
                                        }
                                        else if (posBalance >= 61) {
                                            if (mcTcor < 85) {
                                                if (tcor < 68) {
                                                    if (MVDiff < 497) {
                                                        return 1;
                                                    }
                                                    else if (MVDiff >= 497) {
                                                        return 2;
                                                    }
                                                }
                                                else if (tcor >= 68) {
                                                    if (MVDiff < 23) {
                                                        if (Rs < 4198) {
                                                            return 2;
                                                        }
                                                        else if (Rs >= 4198) {
                                                            return 1;
                                                        }
                                                    }
                                                    else if (MVDiff >= 23) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                            else if (mcTcor >= 85) {
                                                if (negBalance < 52) {
                                                    if (Cs < 1187) {
                                                        if (AFD < 20) {
                                                            return 1;
                                                        }
                                                        else if (AFD >= 20) {
                                                            return 2;
                                                        }
                                                    }
                                                    else if (Cs >= 1187) {
                                                        if (MVDiff < 11) {
                                                            return 4;
                                                        }
                                                        else if (MVDiff >= 11) {
                                                            if (ssDCval < 132) {
                                                                return 8;
                                                            }
                                                            else if (ssDCval >= 132) {
                                                                if (Rs < 2639) {
                                                                    return 8;
                                                                }
                                                                else if (Rs >= 2639) {
                                                                    return 1;
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (negBalance >= 52) {
                                                    if (tcor < 60) {
                                                        return 1;
                                                    }
                                                    else if (tcor >= 60) {
                                                        if (Cs < 522) {
                                                            return 4;
                                                        }
                                                        else if (Cs >= 522) {
                                                            return 2;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (tcor >= 90) {
                                    if (Cs < 2376) {
                                        if (ssDCval < 110) {
                                            if (RsCsDiff < 99) {
                                                if (refDCval < 88) {
                                                    if (diffAFD < 4) {
                                                        if (Cs < 1968) {
                                                            if (Cs < 320) {
                                                                return 1;
                                                            }
                                                            else if (Cs >= 320) {
                                                                return 2;
                                                            }
                                                        }
                                                        else if (Cs >= 1968) {
                                                            return 1;
                                                        }
                                                    }
                                                    else if (diffAFD >= 4) {
                                                        return 1;
                                                    }
                                                }
                                                else if (refDCval >= 88) {
                                                    if (ssDCval < 93) {
                                                        return 1;
                                                    }
                                                    else if (ssDCval >= 93) {
                                                        if (MVDiff < 6) {
                                                            return 2;
                                                        }
                                                        else if (MVDiff >= 6) {
                                                            if (Rs < 334) {
                                                                return 2;
                                                            }
                                                            else if (Rs >= 334) {
                                                                return 1;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                            else if (RsCsDiff >= 99) {
                                                if (Rs < 2245) {
                                                    if (posBalance < 54) {
                                                        return 8;
                                                    }
                                                    else if (posBalance >= 54) {
                                                        if (refDCval < 92) {
                                                            return 2;
                                                        }
                                                        else if (refDCval >= 92) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                                else if (Rs >= 2245) {
                                                    if (diffRsCsdiff < 14) {
                                                        if (SC < 4557) {
                                                            if (gchDC < 2) {
                                                                if (Rs < 2464) {
                                                                    return 8;
                                                                }
                                                                else if (Rs >= 2464) {
                                                                    if (AFD < 14) {
                                                                        return 2;
                                                                    }
                                                                    else if (AFD >= 14) {
                                                                        return 1;
                                                                    }
                                                                }
                                                            }
                                                            else if (gchDC >= 2) {
                                                                return 1;
                                                            }
                                                        }
                                                        else if (SC >= 4557) {
                                                            if (CsDiff < 7) {
                                                                return 8;
                                                            }
                                                            else if (CsDiff >= 7) {
                                                                if (Rs < 3395) {
                                                                    return 2;
                                                                }
                                                                else if (Rs >= 3395) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (diffRsCsdiff >= 14) {
                                                        if (Rs < 3152) {
                                                            return 1;
                                                        }
                                                        else if (Rs >= 3152) {
                                                            return 2;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (ssDCval >= 110) {
                                            if (RsCsDiff < 161) {
                                                if (MVDiff < 6) {
                                                    return 8;
                                                }
                                                else if (MVDiff >= 6) {
                                                    return 2;
                                                }
                                            }
                                            else if (RsCsDiff >= 161) {
                                                if (Cs < 1866) {
                                                    if (Cs < 1673) {
                                                        return 8;
                                                    }
                                                    else if (Cs >= 1673) {
                                                        return 4;
                                                    }
                                                }
                                                else if (Cs >= 1866) {
                                                    return 8;
                                                }
                                            }
                                        }
                                    }
                                    else if (Cs >= 2376) {
                                        if (refDCval < 112) {
                                            if (MVDiff < 3) {
                                                return 8;
                                            }
                                            else if (MVDiff >= 3) {
                                                if (refDCval < 92) {
                                                    if (Rs < 3204) {
                                                        if (TSC < 11) {
                                                            return 8;
                                                        }
                                                        else if (TSC >= 11) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (Rs >= 3204) {
                                                        return 4;
                                                    }
                                                }
                                                else if (refDCval >= 92) {
                                                    if (TSC < 11) {
                                                        if (TSC < 10) {
                                                            return 2;
                                                        }
                                                        else if (TSC >= 10) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (TSC >= 11) {
                                                        if (CsDiff < 9) {
                                                            return 1;
                                                        }
                                                        else if (CsDiff >= 9) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (refDCval >= 112) {
                                            return 8;
                                        }
                                    }
                                }
                            }
                            else if (ssDCval >= 139) {
                                if (Cs < 1400) {
                                    if (Rs < 1524) {
                                        if (ssDCval < 154) {
                                            if (AFD < 15) {
                                                if (MVDiff < 13) {
                                                    return 4;
                                                }
                                                else if (MVDiff >= 13) {
                                                    return 8;
                                                }
                                            }
                                            else if (AFD >= 15) {
                                                if (Rs < 424) {
                                                    return 8;
                                                }
                                                else if (Rs >= 424) {
                                                    return 2;
                                                }
                                            }
                                        }
                                        else if (ssDCval >= 154) {
                                            if (AFD < 17) {
                                                if (MVDiff < 314) {
                                                    return 2;
                                                }
                                                else if (MVDiff >= 314) {
                                                    return 8;
                                                }
                                            }
                                            else if (AFD >= 17) {
                                                if (gchDC < 62) {
                                                    return 1;
                                                }
                                                else if (gchDC >= 62) {
                                                    if (TSC < 25) {
                                                        return 8;
                                                    }
                                                    else if (TSC >= 25) {
                                                        return 2;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (Rs >= 1524) {
                                        if (RsDiff < 24) {
                                            return 2;
                                        }
                                        else if (RsDiff >= 24) {
                                            return 4;
                                        }
                                    }
                                }
                                else if (Cs >= 1400) {
                                    if (RsDiff < 12) {
                                        if (posBalance < 53) {
                                            if (ssDCval < 179) {
                                                return 8;
                                            }
                                            else if (ssDCval >= 179) {
                                                return 4;
                                            }
                                        }
                                        else if (posBalance >= 53) {
                                            if (tcor < 71) {
                                                return 1;
                                            }
                                            else if (tcor >= 71) {
                                                if (TSC < 11) {
                                                    return 2;
                                                }
                                                else if (TSC >= 11) {
                                                    return 4;
                                                }
                                            }
                                        }
                                    }
                                    else if (RsDiff >= 12) {
                                        if (CsDiff < 11) {
                                            if (negBalance < 51) {
                                                if (ssDCval < 162) {
                                                    if (RsDiff < 16) {
                                                        return 1;
                                                    }
                                                    else if (RsDiff >= 16) {
                                                        return 4;
                                                    }
                                                }
                                                else if (ssDCval >= 162) {
                                                    if (Cs < 1641) {
                                                        return 4;
                                                    }
                                                    else if (Cs >= 1641) {
                                                        return 2;
                                                    }
                                                }
                                            }
                                            else if (negBalance >= 51) {
                                                if (Rs < 5068) {
                                                    return 2;
                                                }
                                                else if (Rs >= 5068) {
                                                    if (RsDiff < 17) {
                                                        return 2;
                                                    }
                                                    else if (RsDiff >= 17) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                        }
                                        else if (CsDiff >= 11) {
                                            if (AFD < 24) {
                                                if (AFD < 18) {
                                                    return 1;
                                                }
                                                else if (AFD >= 18) {
                                                    return 4;
                                                }
                                            }
                                            else if (AFD >= 24) {
                                                if (ssDCval < 154) {
                                                    if (SC < 4753) {
                                                        if (Rs < 2231) {
                                                            return 8;
                                                        }
                                                        else if (Rs >= 2231) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (SC >= 4753) {
                                                        return 2;
                                                    }
                                                }
                                                else if (ssDCval >= 154) {
                                                    return 4;
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                        }
                    }
                    else if (Cs >= 3269) {
                        if (RsCsDiff < 138) {
                            if (RsDiff < 5) {
                                if (TSC < 7) {
                                    if (mu_mv_mag_sq < 35) {
                                        if (diffMVdiffVal < 2) {
                                            return 8;
                                        }
                                        else if (diffMVdiffVal >= 2) {
                                            if (SC < 6728) {
                                                return 4;
                                            }
                                            else if (SC >= 6728) {
                                                if (mcTcor < 95) {
                                                    return 4;
                                                }
                                                else if (mcTcor >= 95) {
                                                    return 8;
                                                }
                                            }
                                        }
                                    }
                                    else if (mu_mv_mag_sq >= 35) {
                                        if (refDCval < 54) {
                                            if (negBalance < 52) {
                                                if (SC < 6792) {
                                                    if (AFD < 9) {
                                                        return 8;
                                                    }
                                                    else if (AFD >= 9) {
                                                        if (Rs < 3366) {
                                                            return 4;
                                                        }
                                                        else if (Rs >= 3366) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                                else if (SC >= 6792) {
                                                    if (MVDiff < 5) {
                                                        if (RsDiff < 3) {
                                                            return 8;
                                                        }
                                                        else if (RsDiff >= 3) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (MVDiff >= 5) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                            else if (negBalance >= 52) {
                                                return 2;
                                            }
                                        }
                                        else if (refDCval >= 54) {
                                            if (refDCval < 62) {
                                                if (diffTSC < -1) {
                                                    if (Cs < 3550) {
                                                        return 4;
                                                    }
                                                    else if (Cs >= 3550) {
                                                        return 8;
                                                    }
                                                }
                                                else if (diffTSC >= -1) {
                                                    if (AFD < 8) {
                                                        if (mu_mv_mag_sq < 68) {
                                                            if (Rs < 3597) {
                                                                if (Rs < 3542) {
                                                                    return 4;
                                                                }
                                                                else if (Rs >= 3542) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (Rs >= 3597) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (mu_mv_mag_sq >= 68) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (AFD >= 8) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                            else if (refDCval >= 62) {
                                                return 4;
                                            }
                                        }
                                    }
                                }
                                else if (TSC >= 7) {
                                    if (MVDiff < 19) {
                                        if (negBalance < 50) {
                                            if (mu_mv_mag_sq < 43) {
                                                if (diffAFD < 1) {
                                                    if (mu_mv_mag_sq < 39) {
                                                        return 4;
                                                    }
                                                    else if (mu_mv_mag_sq >= 39) {
                                                        return 2;
                                                    }
                                                }
                                                else if (diffAFD >= 1) {
                                                    return 4;
                                                }
                                            }
                                            else if (mu_mv_mag_sq >= 43) {
                                                return 4;
                                            }
                                        }
                                        else if (negBalance >= 50) {
                                            if (gchDC < 1) {
                                                if (refDCval < 53) {
                                                    return 4;
                                                }
                                                else if (refDCval >= 53) {
                                                    return 2;
                                                }
                                            }
                                            else if (gchDC >= 1) {
                                                if (Cs < 3786) {
                                                    return 4;
                                                }
                                                else if (Cs >= 3786) {
                                                    if (posBalance < 57) {
                                                        return 2;
                                                    }
                                                    else if (posBalance >= 57) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (MVDiff >= 19) {
                                        if (Rs < 3518) {
                                            if (RsCsDiff < 23) {
                                                if (ssDCval < 48) {
                                                    return 8;
                                                }
                                                else if (ssDCval >= 48) {
                                                    if (SC < 6669) {
                                                        return 8;
                                                    }
                                                    else if (SC >= 6669) {
                                                        return 2;
                                                    }
                                                }
                                            }
                                            else if (RsCsDiff >= 23) {
                                                if (MVDiff < 42) {
                                                    return 8;
                                                }
                                                else if (MVDiff >= 42) {
                                                    return 2;
                                                }
                                            }
                                        }
                                        else if (Rs >= 3518) {
                                            if (posBalance < 66) {
                                                if (ssDCval < 58) {
                                                    if (tcor < 86) {
                                                        if (SC < 7297) {
                                                            return 4;
                                                        }
                                                        else if (SC >= 7297) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (tcor >= 86) {
                                                        if (MVDiff < 26) {
                                                            if (MVDiff < 25) {
                                                                return 2;
                                                            }
                                                            else if (MVDiff >= 25) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (MVDiff >= 26) {
                                                            return 2;
                                                        }
                                                    }
                                                }
                                                else if (ssDCval >= 58) {
                                                    if (Cs < 3678) {
                                                        return 8;
                                                    }
                                                    else if (Cs >= 3678) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                            else if (posBalance >= 66) {
                                                return 8;
                                            }
                                        }
                                    }
                                }
                            }
                            else if (RsDiff >= 5) {
                                if (refDCval < 70) {
                                    if (mcTcor < 95) {
                                        if (Rs < 3554) {
                                            return 8;
                                        }
                                        else if (Rs >= 3554) {
                                            if (TSCindex < 8) {
                                                return 8;
                                            }
                                            else if (TSCindex >= 8) {
                                                if (tcor < 80) {
                                                    if (Cs < 3900) {
                                                        return 8;
                                                    }
                                                    else if (Cs >= 3900) {
                                                        return 4;
                                                    }
                                                }
                                                else if (tcor >= 80) {
                                                    return 4;
                                                }
                                            }
                                        }
                                    }
                                    else if (mcTcor >= 95) {
                                        if (AFD < 12) {
                                            if (refDCval < 63) {
                                                return 8;
                                            }
                                            else if (refDCval >= 63) {
                                                return 4;
                                            }
                                        }
                                        else if (AFD >= 12) {
                                            return 8;
                                        }
                                    }
                                }
                                else if (refDCval >= 70) {
                                    if (AFD < 13) {
                                        return 4;
                                    }
                                    else if (AFD >= 13) {
                                        return 8;
                                    }
                                }
                            }
                        }
                        else if (RsCsDiff >= 138) {
                            if (ssDCval < 149) {
                                if (Rs < 6989) {
                                    if (mu_mv_mag_sq < 64) {
                                        if (SC < 8269) {
                                            if (refDCval < 135) {
                                                if (Rs < 2639) {
                                                    if (mcTcor < 69) {
                                                        return 4;
                                                    }
                                                    else if (mcTcor >= 69) {
                                                        return 1;
                                                    }
                                                }
                                                else if (Rs >= 2639) {
                                                    if (Rs < 4048) {
                                                        if (ssDCval < 127) {
                                                            if (diffRsCsdiff < 154) {
                                                                if (CsDiff < 20) {
                                                                    return 8;
                                                                }
                                                                else if (CsDiff >= 20) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (diffRsCsdiff >= 154) {
                                                                return 1;
                                                            }
                                                        }
                                                        else if (ssDCval >= 127) {
                                                            return 1;
                                                        }
                                                    }
                                                    else if (Rs >= 4048) {
                                                        if (MVDiff < 18) {
                                                            if (posBalance < 59) {
                                                                if (ssDCval < 125) {
                                                                    return 4;
                                                                }
                                                                else if (ssDCval >= 125) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (posBalance >= 59) {
                                                                if (SC < 7575) {
                                                                    return 1;
                                                                }
                                                                else if (SC >= 7575) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                        else if (MVDiff >= 18) {
                                                            if (SC < 7672) {
                                                                return 2;
                                                            }
                                                            else if (SC >= 7672) {
                                                                if (RsDiff < 15) {
                                                                    return 1;
                                                                }
                                                                else if (RsDiff >= 15) {
                                                                    if (SC < 7791) {
                                                                        return 1;
                                                                    }
                                                                    else if (SC >= 7791) {
                                                                        return 8;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                            else if (refDCval >= 135) {
                                                if (mcTcor < 85) {
                                                    return 1;
                                                }
                                                else if (mcTcor >= 85) {
                                                    return 8;
                                                }
                                            }
                                        }
                                        else if (SC >= 8269) {
                                            if (CsDiff < 18) {
                                                if (RsDiff < 22) {
                                                    if (AFD < 17) {
                                                        if (Cs < 4466) {
                                                            if (negBalance < 45) {
                                                                if (Cs < 3712) {
                                                                    return 2;
                                                                }
                                                                else if (Cs >= 3712) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (negBalance >= 45) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (Cs >= 4466) {
                                                            if (AFD < 15) {
                                                                return 4;
                                                            }
                                                            else if (AFD >= 15) {
                                                                if (ssDCval < 66) {
                                                                    return 4;
                                                                }
                                                                else if (ssDCval >= 66) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (AFD >= 17) {
                                                        if (posBalance < 52) {
                                                            return 8;
                                                        }
                                                        else if (posBalance >= 52) {
                                                            if (diffTSC < 1) {
                                                                if (mcTcor < 82) {
                                                                    if (TSC < 14) {
                                                                        return 4;
                                                                    }
                                                                    else if (TSC >= 14) {
                                                                        if (RsCsDiff < 557) {
                                                                            return 8;
                                                                        }
                                                                        else if (RsCsDiff >= 557) {
                                                                            if (ssDCval < 109) {
                                                                                return 8;
                                                                            }
                                                                            else if (ssDCval >= 109) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                                else if (mcTcor >= 82) {
                                                                    if (tcor < 89) {
                                                                        return 8;
                                                                    }
                                                                    else if (tcor >= 89) {
                                                                        return 4;
                                                                    }
                                                                }
                                                            }
                                                            else if (diffTSC >= 1) {
                                                                if (Rs < 6334) {
                                                                    if (SC < 8320) {
                                                                        return 1;
                                                                    }
                                                                    else if (SC >= 8320) {
                                                                        return 8;
                                                                    }
                                                                }
                                                                else if (Rs >= 6334) {
                                                                    return 1;
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (RsDiff >= 22) {
                                                    if (TSC < 15) {
                                                        if (MVDiff < 17) {
                                                            return 1;
                                                        }
                                                        else if (MVDiff >= 17) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (TSC >= 15) {
                                                        if (ssDCval < 107) {
                                                            return 8;
                                                        }
                                                        else if (ssDCval >= 107) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (CsDiff >= 18) {
                                                if (mcTcor < 91) {
                                                    if (Rs < 4189) {
                                                        return 1;
                                                    }
                                                    else if (Rs >= 4189) {
                                                        if (tcor < 32) {
                                                            if (Rs < 6127) {
                                                                return 2;
                                                            }
                                                            else if (Rs >= 6127) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (tcor >= 32) {
                                                            if (diffTSC < -5) {
                                                                if (Rs < 6477) {
                                                                    return 8;
                                                                }
                                                                else if (Rs >= 6477) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (diffTSC >= -5) {
                                                                if (RsCsDiff < 1000) {
                                                                    return 8;
                                                                }
                                                                else if (RsCsDiff >= 1000) {
                                                                    if (CsDiff < 23) {
                                                                        if (RsCsDiff < 1269) {
                                                                            if (tcor < 81) {
                                                                                if (diffMVdiffVal < 11) {
                                                                                    return 8;
                                                                                }
                                                                                else if (diffMVdiffVal >= 11) {
                                                                                    return 4;
                                                                                }
                                                                            }
                                                                            else if (tcor >= 81) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                        else if (RsCsDiff >= 1269) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                    else if (CsDiff >= 23) {
                                                                        return 8;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (mcTcor >= 91) {
                                                    if (TSC < 15) {
                                                        if (Cs < 5604) {
                                                            return 1;
                                                        }
                                                        else if (Cs >= 5604) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (TSC >= 15) {
                                                        if (RsDiff < 24) {
                                                            return 8;
                                                        }
                                                        else if (RsDiff >= 24) {
                                                            if (SC < 10922) {
                                                                return 4;
                                                            }
                                                            else if (SC >= 10922) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (mu_mv_mag_sq >= 64) {
                                        if (ssDCval < 136) {
                                            if (posBalance < 59) {
                                                if (diffAFD < 1) {
                                                    if (diffRsCsdiff < -46) {
                                                        if (RsCsDiff < 352) {
                                                            return 4;
                                                        }
                                                        else if (RsCsDiff >= 352) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (diffRsCsdiff >= -46) {
                                                        if (posBalance < 57) {
                                                            if (AFD < 13) {
                                                                return 2;
                                                            }
                                                            else if (AFD >= 13) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (posBalance >= 57) {
                                                            if (Cs < 3888) {
                                                                return 4;
                                                            }
                                                            else if (Cs >= 3888) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (diffAFD >= 1) {
                                                    return 8;
                                                }
                                            }
                                            else if (posBalance >= 59) {
                                                return 8;
                                            }
                                        }
                                        else if (ssDCval >= 136) {
                                            if (Rs < 4859) {
                                                return 1;
                                            }
                                            else if (Rs >= 4859) {
                                                return 8;
                                            }
                                        }
                                    }
                                }
                                else if (Rs >= 6989) {
                                    if (Cs < 6091) {
                                        if (TSC < 18) {
                                            if (Cs < 4940) {
                                                return 8;
                                            }
                                            else if (Cs >= 4940) {
                                                return 4;
                                            }
                                        }
                                        else if (TSC >= 18) {
                                            if (mcTcor < 93) {
                                                return 2;
                                            }
                                            else if (mcTcor >= 93) {
                                                return 4;
                                            }
                                        }
                                    }
                                    else if (Cs >= 6091) {
                                        if (mu_mv_mag_sq < 63) {
                                            if (SC < 14588) {
                                                if (Rs < 7911) {
                                                    if (RsDiff < 24) {
                                                        return 4;
                                                    }
                                                    else if (RsDiff >= 24) {
                                                        if (negBalance < 55) {
                                                            if (AFD < 27) {
                                                                return 4;
                                                            }
                                                            else if (AFD >= 27) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (negBalance >= 55) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                                else if (Rs >= 7911) {
                                                    return 4;
                                                }
                                            }
                                            else if (SC >= 14588) {
                                                if (posBalance < 57) {
                                                    if (Rs < 13509) {
                                                        if (refDCval < 82) {
                                                            return 4;
                                                        }
                                                        else if (refDCval >= 82) {
                                                            if (mu_mv_mag_sq < 20) {
                                                                if (Rs < 11021) {
                                                                    return 8;
                                                                }
                                                                else if (Rs >= 11021) {
                                                                    return 2;
                                                                }
                                                            }
                                                            else if (mu_mv_mag_sq >= 20) {
                                                                if (Rs < 8969) {
                                                                    return 2;
                                                                }
                                                                else if (Rs >= 8969) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (Rs >= 13509) {
                                                        return 8;
                                                    }
                                                }
                                                else if (posBalance >= 57) {
                                                    if (refDCval < 125) {
                                                        if (AFD < 42) {
                                                            if (diffMVdiffVal < 8) {
                                                                if (CsDiff < 23) {
                                                                    if (Cs < 7265) {
                                                                        return 4;
                                                                    }
                                                                    else if (Cs >= 7265) {
                                                                        return 8;
                                                                    }
                                                                }
                                                                else if (CsDiff >= 23) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (diffMVdiffVal >= 8) {
                                                                return 2;
                                                            }
                                                        }
                                                        else if (AFD >= 42) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (refDCval >= 125) {
                                                        if (TSC < 22) {
                                                            return 8;
                                                        }
                                                        else if (TSC >= 22) {
                                                            if (ssDCval < 134) {
                                                                return 8;
                                                            }
                                                            else if (ssDCval >= 134) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (mu_mv_mag_sq >= 63) {
                                            return 8;
                                        }
                                    }
                                }
                            }
                            else if (ssDCval >= 149) {
                                if (negBalance < 54) {
                                    if (Cs < 4742) {
                                        if (ssDCval < 151) {
                                            if (TSC < 15) {
                                                return 8;
                                            }
                                            else if (TSC >= 15) {
                                                return 2;
                                            }
                                        }
                                        else if (ssDCval >= 151) {
                                            if (MVDiff < 20) {
                                                return 4;
                                            }
                                            else if (MVDiff >= 20) {
                                                return 8;
                                            }
                                        }
                                    }
                                    else if (Cs >= 4742) {
                                        if (AFD < 43) {
                                            if (negBalance < 45) {
                                                if (Rs < 6339) {
                                                    return 8;
                                                }
                                                else if (Rs >= 6339) {
                                                    return 4;
                                                }
                                            }
                                            else if (negBalance >= 45) {
                                                return 8;
                                            }
                                        }
                                        else if (AFD >= 43) {
                                            if (Rs < 8252) {
                                                return 8;
                                            }
                                            else if (Rs >= 8252) {
                                                return 2;
                                            }
                                        }
                                    }
                                }
                                else if (negBalance >= 54) {
                                    if (refDCval < 153) {
                                        if (Rs < 8636) {
                                            return 8;
                                        }
                                        else if (Rs >= 8636) {
                                            if (Cs < 7704) {
                                                return 4;
                                            }
                                            else if (Cs >= 7704) {
                                                return 2;
                                            }
                                        }
                                    }
                                    else if (refDCval >= 153) {
                                        if (Scindex < 7) {
                                            if (tcor < 76) {
                                                return 4;
                                            }
                                            else if (tcor >= 76) {
                                                if (ssDCval < 159) {
                                                    return 4;
                                                }
                                                else if (ssDCval >= 159) {
                                                    return 2;
                                                }
                                            }
                                        }
                                        else if (Scindex >= 7) {
                                            if (MVDiff < 7) {
                                                return 8;
                                            }
                                            else if (MVDiff >= 7) {
                                                if (diffRsCsdiff < -180) {
                                                    if (Cs < 4926) {
                                                        return 8;
                                                    }
                                                    else if (Cs >= 4926) {
                                                        return 2;
                                                    }
                                                }
                                                else if (diffRsCsdiff >= -180) {
                                                    return 2;
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                        }
                    }
                }
            }
            else if (mu_mv_mag_sq >= 82) {
                if (Cs < 4177) {
                    if (AFD < 7) {
                        if (MVDiff < 9) {
                            if (mcTcor < 95) {
                                if (refDCval < 107) {
                                    if (TSC < 6) {
                                        if (ssDCval < 43) {
                                            if (mu_mv_mag_sq < 111) {
                                                if (tcor < 91) {
                                                    if (Rs < 629) {
                                                        return 4;
                                                    }
                                                    else if (Rs >= 629) {
                                                        return 8;
                                                    }
                                                }
                                                else if (tcor >= 91) {
                                                    if (Cs < 1365) {
                                                        return 2;
                                                    }
                                                    else if (Cs >= 1365) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                            else if (mu_mv_mag_sq >= 111) {
                                                return 8;
                                            }
                                        }
                                        else if (ssDCval >= 43) {
                                            if (MVDiff < 6) {
                                                if (mcTcor < 94) {
                                                    if (Cs < 1778) {
                                                        return 8;
                                                    }
                                                    else if (Cs >= 1778) {
                                                        if (refDCval < 51) {
                                                            return 4;
                                                        }
                                                        else if (refDCval >= 51) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                                else if (mcTcor >= 94) {
                                                    if (diffRsCsdiff < 29) {
                                                        if (mu_mv_mag_sq < 108) {
                                                            if (refDCval < 68) {
                                                                return 8;
                                                            }
                                                            else if (refDCval >= 68) {
                                                                if (Rs < 1843) {
                                                                    return 1;
                                                                }
                                                                else if (Rs >= 1843) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                        else if (mu_mv_mag_sq >= 108) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (diffRsCsdiff >= 29) {
                                                        return 1;
                                                    }
                                                }
                                            }
                                            else if (MVDiff >= 6) {
                                                if (refDCval < 54) {
                                                    if (Rs < 1202) {
                                                        return 8;
                                                    }
                                                    else if (Rs >= 1202) {
                                                        return 1;
                                                    }
                                                }
                                                else if (refDCval >= 54) {
                                                    return 8;
                                                }
                                            }
                                        }
                                    }
                                    else if (TSC >= 6) {
                                        if (refDCval < 75) {
                                            if (posBalance < 58) {
                                                if (negBalance < 46) {
                                                    return 8;
                                                }
                                                else if (negBalance >= 46) {
                                                    if (diffRsCsdiff < 12) {
                                                        if (Rs < 1912) {
                                                            return 8;
                                                        }
                                                        else if (Rs >= 1912) {
                                                            if (CsDiff < 4) {
                                                                return 8;
                                                            }
                                                            else if (CsDiff >= 4) {
                                                                return 2;
                                                            }
                                                        }
                                                    }
                                                    else if (diffRsCsdiff >= 12) {
                                                        return 2;
                                                    }
                                                }
                                            }
                                            else if (posBalance >= 58) {
                                                if (SC < 3556) {
                                                    return 2;
                                                }
                                                else if (SC >= 3556) {
                                                    return 8;
                                                }
                                            }
                                        }
                                        else if (refDCval >= 75) {
                                            if (Cs < 1542) {
                                                return 4;
                                            }
                                            else if (Cs >= 1542) {
                                                if (posBalance < 55) {
                                                    if (Cs < 2368) {
                                                        return 8;
                                                    }
                                                    else if (Cs >= 2368) {
                                                        return 4;
                                                    }
                                                }
                                                else if (posBalance >= 55) {
                                                    return 8;
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (refDCval >= 107) {
                                    if (ssDCval < 143) {
                                        if (Rs < 2181) {
                                            if (mu_mv_mag_sq < 124) {
                                                if (negBalance < 3) {
                                                    if (diffAFD < 1) {
                                                        return 8;
                                                    }
                                                    else if (diffAFD >= 1) {
                                                        if (SC < 2848) {
                                                            return 1;
                                                        }
                                                        else if (SC >= 2848) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                                else if (negBalance >= 3) {
                                                    if (mu_mv_mag_sq < 114) {
                                                        if (MVDiff < 6) {
                                                            return 8;
                                                        }
                                                        else if (MVDiff >= 6) {
                                                            if (Rs < 1022) {
                                                                if (Cs < 770) {
                                                                    return 1;
                                                                }
                                                                else if (Cs >= 770) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (Rs >= 1022) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                    else if (mu_mv_mag_sq >= 114) {
                                                        if (AFD < 3) {
                                                            return 1;
                                                        }
                                                        else if (AFD >= 3) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (mu_mv_mag_sq >= 124) {
                                                if (AFD < 1) {
                                                    return 8;
                                                }
                                                else if (AFD >= 1) {
                                                    return 1;
                                                }
                                            }
                                        }
                                        else if (Rs >= 2181) {
                                            if (negBalance < 36) {
                                                if (Rs < 3154) {
                                                    if (mcTcor < 92) {
                                                        return 4;
                                                    }
                                                    else if (mcTcor >= 92) {
                                                        if (posBalance < 9) {
                                                            return 4;
                                                        }
                                                        else if (posBalance >= 9) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                                else if (Rs >= 3154) {
                                                    return 8;
                                                }
                                            }
                                            else if (negBalance >= 36) {
                                                return 4;
                                            }
                                        }
                                    }
                                    else if (ssDCval >= 143) {
                                        return 8;
                                    }
                                }
                            }
                            else if (mcTcor >= 95) {
                                if (AFD < 6) {
                                    if (mu_mv_mag_sq < 121) {
                                        if (refDCval < 72) {
                                            if (ssDCval < 71) {
                                                if (diffTSC < -1) {
                                                    if (Rs < 1786) {
                                                        if (diffRsCsdiff < -7) {
                                                            if (SC < 2507) {
                                                                return 1;
                                                            }
                                                            else if (SC >= 2507) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (diffRsCsdiff >= -7) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (Rs >= 1786) {
                                                        return 4;
                                                    }
                                                }
                                                else if (diffTSC >= -1) {
                                                    if (Rs < 650) {
                                                        return 4;
                                                    }
                                                    else if (Rs >= 650) {
                                                        if (diffRsCsdiff < -13) {
                                                            if (Cs < 1958) {
                                                                return 8;
                                                            }
                                                            else if (Cs >= 1958) {
                                                                if (Rs < 2222) {
                                                                    return 4;
                                                                }
                                                                else if (Rs >= 2222) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                        else if (diffRsCsdiff >= -13) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (ssDCval >= 71) {
                                                if (Cs < 1932) {
                                                    return 8;
                                                }
                                                else if (Cs >= 1932) {
                                                    return 1;
                                                }
                                            }
                                        }
                                        else if (refDCval >= 72) {
                                            if (ssDCval < 128) {
                                                if (refDCval < 117) {
                                                    if (Rs < 2536) {
                                                        if (SC < 4753) {
                                                            if (posBalance < 55) {
                                                                if (refDCval < 79) {
                                                                    return 4;
                                                                }
                                                                else if (refDCval >= 79) {
                                                                    if (Rs < 1085) {
                                                                        return 4;
                                                                    }
                                                                    else if (Rs >= 1085) {
                                                                        if (RsDiff < 4) {
                                                                            if (ssDCval < 117) {
                                                                                if (TSCindex < 2) {
                                                                                    return 4;
                                                                                }
                                                                                else if (TSCindex >= 2) {
                                                                                    return 8;
                                                                                }
                                                                            }
                                                                            else if (ssDCval >= 117) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                        else if (RsDiff >= 4) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                            else if (posBalance >= 55) {
                                                                if (Rs < 1805) {
                                                                    return 8;
                                                                }
                                                                else if (Rs >= 1805) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                        else if (SC >= 4753) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (Rs >= 2536) {
                                                        return 8;
                                                    }
                                                }
                                                else if (refDCval >= 117) {
                                                    return 4;
                                                }
                                            }
                                            else if (ssDCval >= 128) {
                                                if (diffAFD < -1) {
                                                    if (Cs < 1690) {
                                                        return 1;
                                                    }
                                                    else if (Cs >= 1690) {
                                                        return 8;
                                                    }
                                                }
                                                else if (diffAFD >= -1) {
                                                    if (mu_mv_mag_sq < 85) {
                                                        return 4;
                                                    }
                                                    else if (mu_mv_mag_sq >= 85) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (mu_mv_mag_sq >= 121) {
                                        if (gchDC < 1) {
                                            return 8;
                                        }
                                        else if (gchDC >= 1) {
                                            if (mu_mv_mag_sq < 124) {
                                                if (ssDCval < 81) {
                                                    if (Cs < 1953) {
                                                        if (Rs < 1767) {
                                                            return 8;
                                                        }
                                                        else if (Rs >= 1767) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (Cs >= 1953) {
                                                        return 8;
                                                    }
                                                }
                                                else if (ssDCval >= 81) {
                                                    return 4;
                                                }
                                            }
                                            else if (mu_mv_mag_sq >= 124) {
                                                return 8;
                                            }
                                        }
                                    }
                                }
                                else if (AFD >= 6) {
                                    if (MVDiff < 4) {
                                        if (ssDCval < 73) {
                                            if (Rs < 1940) {
                                                return 2;
                                            }
                                            else if (Rs >= 1940) {
                                                if (posBalance < 55) {
                                                    if (Rs < 3511) {
                                                        return 8;
                                                    }
                                                    else if (Rs >= 3511) {
                                                        return 4;
                                                    }
                                                }
                                                else if (posBalance >= 55) {
                                                    return 4;
                                                }
                                            }
                                        }
                                        else if (ssDCval >= 73) {
                                            if (mu_mv_mag_sq < 113) {
                                                return 4;
                                            }
                                            else if (mu_mv_mag_sq >= 113) {
                                                if (Rs < 1838) {
                                                    return 8;
                                                }
                                                else if (Rs >= 1838) {
                                                    if (negBalance < 42) {
                                                        return 8;
                                                    }
                                                    else if (negBalance >= 42) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (MVDiff >= 4) {
                                        if (mu_mv_mag_sq < 84) {
                                            return 4;
                                        }
                                        else if (mu_mv_mag_sq >= 84) {
                                            return 8;
                                        }
                                    }
                                }
                            }
                        }
                        else if (MVDiff >= 9) {
                            if (SC < 1789) {
                                if (RsCsDiff < 7) {
                                    if (mu_mv_mag_sq < 104) {
                                        if (posBalance < 58) {
                                            if (Rs < 940) {
                                                if (diffTSC < -1) {
                                                    if (Rs < 418) {
                                                        return 8;
                                                    }
                                                    else if (Rs >= 418) {
                                                        return 4;
                                                    }
                                                }
                                                else if (diffTSC >= -1) {
                                                    return 4;
                                                }
                                            }
                                            else if (Rs >= 940) {
                                                return 1;
                                            }
                                        }
                                        else if (posBalance >= 58) {
                                            if (Cs < 217) {
                                                return 1;
                                            }
                                            else if (Cs >= 217) {
                                                return 2;
                                            }
                                        }
                                    }
                                    else if (mu_mv_mag_sq >= 104) {
                                        if (diffMVdiffVal < 7) {
                                            return 8;
                                        }
                                        else if (diffMVdiffVal >= 7) {
                                            if (diffMVdiffVal < 140) {
                                                return 4;
                                            }
                                            else if (diffMVdiffVal >= 140) {
                                                if (AFD < 1) {
                                                    return 8;
                                                }
                                                else if (AFD >= 1) {
                                                    return 1;
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (RsCsDiff >= 7) {
                                    return 1;
                                }
                            }
                            else if (SC >= 1789) {
                                if (negBalance < 39) {
                                    if (diffRsCsdiff < 107) {
                                        if (refDCval < 115) {
                                            if (diffAFD < -4) {
                                                if (CsDiff < 2) {
                                                    return 4;
                                                }
                                                else if (CsDiff >= 2) {
                                                    return 8;
                                                }
                                            }
                                            else if (diffAFD >= -4) {
                                                return 8;
                                            }
                                        }
                                        else if (refDCval >= 115) {
                                            if (Rs < 1485) {
                                                return 8;
                                            }
                                            else if (Rs >= 1485) {
                                                if (Cs < 1768) {
                                                    return 1;
                                                }
                                                else if (Cs >= 1768) {
                                                    return 4;
                                                }
                                            }
                                        }
                                    }
                                    else if (diffRsCsdiff >= 107) {
                                        if (Cs < 1055) {
                                            return 4;
                                        }
                                        else if (Cs >= 1055) {
                                            return 8;
                                        }
                                    }
                                }
                                else if (negBalance >= 39) {
                                    if (posBalance < 53) {
                                        if (Cs < 1838) {
                                            return 1;
                                        }
                                        else if (Cs >= 1838) {
                                            return 8;
                                        }
                                    }
                                    else if (posBalance >= 53) {
                                        if (tcor < 94) {
                                            if (SC < 3340) {
                                                return 8;
                                            }
                                            else if (SC >= 3340) {
                                                return 4;
                                            }
                                        }
                                        else if (tcor >= 94) {
                                            if (tcor < 95) {
                                                return 8;
                                            }
                                            else if (tcor >= 95) {
                                                if (Cs < 2057) {
                                                    return 1;
                                                }
                                                else if (Cs >= 2057) {
                                                    return 4;
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                        }
                    }
                    else if (AFD >= 7) {
                        if (Cs < 1658) {
                            if (ssDCval < 112) {
                                if (tcor < 86) {
                                    if (tcor < 69) {
                                        if (posBalance < 48) {
                                            return 1;
                                        }
                                        else if (posBalance >= 48) {
                                            if (AFD < 33) {
                                                if (TSC < 10) {
                                                    return 2;
                                                }
                                                else if (TSC >= 10) {
                                                    if (ssDCval < 102) {
                                                        if (posBalance < 60) {
                                                            return 2;
                                                        }
                                                        else if (posBalance >= 60) {
                                                            return 1;
                                                        }
                                                    }
                                                    else if (ssDCval >= 102) {
                                                        return 2;
                                                    }
                                                }
                                            }
                                            else if (AFD >= 33) {
                                                return 2;
                                            }
                                        }
                                    }
                                    else if (tcor >= 69) {
                                        if (Cs < 1258) {
                                            if (diffRsCsdiff < -8) {
                                                return 8;
                                            }
                                            else if (diffRsCsdiff >= -8) {
                                                return 4;
                                            }
                                        }
                                        else if (Cs >= 1258) {
                                            return 8;
                                        }
                                    }
                                }
                                else if (tcor >= 86) {
                                    if (Rs < 1775) {
                                        if (MVDiff < 5) {
                                            return 2;
                                        }
                                        else if (MVDiff >= 5) {
                                            if (Cs < 934) {
                                                return 2;
                                            }
                                            else if (Cs >= 934) {
                                                return 8;
                                            }
                                        }
                                    }
                                    else if (Rs >= 1775) {
                                        if (SC < 3448) {
                                            if (mu_mv_mag_sq < 110) {
                                                if (RsDiff < 6) {
                                                    return 2;
                                                }
                                                else if (RsDiff >= 6) {
                                                    return 4;
                                                }
                                            }
                                            else if (mu_mv_mag_sq >= 110) {
                                                if (mcTcor < 94) {
                                                    return 8;
                                                }
                                                else if (mcTcor >= 94) {
                                                    if (SC < 3398) {
                                                        return 4;
                                                    }
                                                    else if (SC >= 3398) {
                                                        return 2;
                                                    }
                                                }
                                            }
                                        }
                                        else if (SC >= 3448) {
                                            return 2;
                                        }
                                    }
                                }
                            }
                            else if (ssDCval >= 112) {
                                if (mu_mv_mag_sq < 128) {
                                    if (RsCsDiff < 117) {
                                        return 8;
                                    }
                                    else if (RsCsDiff >= 117) {
                                        if (mcTcor < 79) {
                                            return 1;
                                        }
                                        else if (mcTcor >= 79) {
                                            if (ssDCval < 131) {
                                                if (CsDiff < 5) {
                                                    return 4;
                                                }
                                                else if (CsDiff >= 5) {
                                                    return 8;
                                                }
                                            }
                                            else if (ssDCval >= 131) {
                                                return 2;
                                            }
                                        }
                                    }
                                }
                                else if (mu_mv_mag_sq >= 128) {
                                    if (diffAFD < 28) {
                                        if (Rs < 2169) {
                                            if (Cs < 283) {
                                                return 2;
                                            }
                                            else if (Cs >= 283) {
                                                return 1;
                                            }
                                        }
                                        else if (Rs >= 2169) {
                                            return 2;
                                        }
                                    }
                                    else if (diffAFD >= 28) {
                                        return 8;
                                    }
                                }
                            }
                        }
                        else if (Cs >= 1658) {
                            if (Rs < 5295) {
                                if (ssDCval < 97) {
                                    if (refDCval < 86) {
                                        if (negBalance < 39) {
                                            if (posBalance < 41) {
                                                if (Rs < 2338) {
                                                    return 1;
                                                }
                                                else if (Rs >= 2338) {
                                                    return 8;
                                                }
                                            }
                                            else if (posBalance >= 41) {
                                                return 8;
                                            }
                                        }
                                        else if (negBalance >= 39) {
                                            if (SC < 3901) {
                                                if (Cs < 1764) {
                                                    if (mu_mv_mag_sq < 110) {
                                                        return 2;
                                                    }
                                                    else if (mu_mv_mag_sq >= 110) {
                                                        if (Cs < 1717) {
                                                            if (SC < 3584) {
                                                                return 2;
                                                            }
                                                            else if (SC >= 3584) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (Cs >= 1717) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                                else if (Cs >= 1764) {
                                                    if (mcTcor < 95) {
                                                        if (Rs < 1925) {
                                                            return 8;
                                                        }
                                                        else if (Rs >= 1925) {
                                                            return 2;
                                                        }
                                                    }
                                                    else if (mcTcor >= 95) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                            else if (SC >= 3901) {
                                                if (Rs < 4110) {
                                                    if (mu_mv_mag_sq < 84) {
                                                        if (negBalance < 48) {
                                                            return 8;
                                                        }
                                                        else if (negBalance >= 48) {
                                                            if (Cs < 2412) {
                                                                return 4;
                                                            }
                                                            else if (Cs >= 2412) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                    else if (mu_mv_mag_sq >= 84) {
                                                        if (mcTcor < 93) {
                                                            if (Rs < 2390) {
                                                                return 8;
                                                            }
                                                            else if (Rs >= 2390) {
                                                                if (ssDCval < 84) {
                                                                    return 4;
                                                                }
                                                                else if (ssDCval >= 84) {
                                                                    return 2;
                                                                }
                                                            }
                                                        }
                                                        else if (mcTcor >= 93) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                                else if (Rs >= 4110) {
                                                    return 8;
                                                }
                                            }
                                        }
                                    }
                                    else if (refDCval >= 86) {
                                        if (ssDCval < 89) {
                                            if (tcor < 93) {
                                                if (MVDiff < 9) {
                                                    if (ssDCval < 86) {
                                                        return 4;
                                                    }
                                                    else if (ssDCval >= 86) {
                                                        return 8;
                                                    }
                                                }
                                                else if (MVDiff >= 9) {
                                                    return 2;
                                                }
                                            }
                                            else if (tcor >= 93) {
                                                if (Rs < 2304) {
                                                    return 2;
                                                }
                                                else if (Rs >= 2304) {
                                                    return 8;
                                                }
                                            }
                                        }
                                        else if (ssDCval >= 89) {
                                            if (mcTcor < 90) {
                                                if (refDCval < 93) {
                                                    return 8;
                                                }
                                                else if (refDCval >= 93) {
                                                    if (RsCsDiff < 56) {
                                                        if (Cs < 1957) {
                                                            return 8;
                                                        }
                                                        else if (Cs >= 1957) {
                                                            if (Rs < 2444) {
                                                                return 2;
                                                            }
                                                            else if (Rs >= 2444) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                    else if (RsCsDiff >= 56) {
                                                        if (Rs < 1873) {
                                                            return 4;
                                                        }
                                                        else if (Rs >= 1873) {
                                                            return 2;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (mcTcor >= 90) {
                                                if (negBalance < 41) {
                                                    if (posBalance < 46) {
                                                        return 1;
                                                    }
                                                    else if (posBalance >= 46) {
                                                        return 8;
                                                    }
                                                }
                                                else if (negBalance >= 41) {
                                                    if (SC < 7632) {
                                                        if (Cs < 2214) {
                                                            if (Cs < 2047) {
                                                                return 2;
                                                            }
                                                            else if (Cs >= 2047) {
                                                                if (AFD < 9) {
                                                                    return 8;
                                                                }
                                                                else if (AFD >= 9) {
                                                                    return 2;
                                                                }
                                                            }
                                                        }
                                                        else if (Cs >= 2214) {
                                                            return 2;
                                                        }
                                                    }
                                                    else if (SC >= 7632) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (ssDCval >= 97) {
                                    if (refDCval < 128) {
                                        if (TSC < 11) {
                                            if (Rs < 3990) {
                                                if (SC < 5055) {
                                                    if (refDCval < 109) {
                                                        if (CsDiff < 6) {
                                                            if (Cs < 2033) {
                                                                return 8;
                                                            }
                                                            else if (Cs >= 2033) {
                                                                return 1;
                                                            }
                                                        }
                                                        else if (CsDiff >= 6) {
                                                            if (Rs < 2579) {
                                                                return 4;
                                                            }
                                                            else if (Rs >= 2579) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                    else if (refDCval >= 109) {
                                                        if (MVDiff < 4) {
                                                            if (mcTcor < 93) {
                                                                return 8;
                                                            }
                                                            else if (mcTcor >= 93) {
                                                                if (MVDiff < 2) {
                                                                    return 8;
                                                                }
                                                                else if (MVDiff >= 2) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                        else if (MVDiff >= 4) {
                                                            if (Rs < 2125) {
                                                                return 8;
                                                            }
                                                            else if (Rs >= 2125) {
                                                                if (mcTcor < 88) {
                                                                    if (Rs < 2956) {
                                                                        return 8;
                                                                    }
                                                                    else if (Rs >= 2956) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (mcTcor >= 88) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (SC >= 5055) {
                                                    if (Rs < 3444) {
                                                        if (mu_mv_mag_sq < 91) {
                                                            if (diffMVdiffVal < -2) {
                                                                return 4;
                                                            }
                                                            else if (diffMVdiffVal >= -2) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (mu_mv_mag_sq >= 91) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (Rs >= 3444) {
                                                        if (Cs < 3231) {
                                                            if (negBalance < 40) {
                                                                if (CsDiff < 6) {
                                                                    return 8;
                                                                }
                                                                else if (CsDiff >= 6) {
                                                                    if (TSC < 8) {
                                                                        if (SC < 6537) {
                                                                            return 2;
                                                                        }
                                                                        else if (SC >= 6537) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                    else if (TSC >= 8) {
                                                                        return 8;
                                                                    }
                                                                }
                                                            }
                                                            else if (negBalance >= 40) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (Cs >= 3231) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (Rs >= 3990) {
                                                if (ssDCval < 115) {
                                                    if (SC < 7898) {
                                                        if (SC < 7005) {
                                                            return 8;
                                                        }
                                                        else if (SC >= 7005) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (SC >= 7898) {
                                                        if (Rs < 5002) {
                                                            if (Cs < 3725) {
                                                                return 2;
                                                            }
                                                            else if (Cs >= 3725) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (Rs >= 5002) {
                                                            if (Cs < 3245) {
                                                                if (diffRsCsdiff < 19) {
                                                                    return 8;
                                                                }
                                                                else if (diffRsCsdiff >= 19) {
                                                                    if (SC < 8078) {
                                                                        return 8;
                                                                    }
                                                                    else if (SC >= 8078) {
                                                                        return 2;
                                                                    }
                                                                }
                                                            }
                                                            else if (Cs >= 3245) {
                                                                return 2;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (ssDCval >= 115) {
                                                    if (ssDCval < 122) {
                                                        if (MVDiff < 9) {
                                                            return 4;
                                                        }
                                                        else if (MVDiff >= 9) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (ssDCval >= 122) {
                                                        if (diffRsCsdiff < 14) {
                                                            return 4;
                                                        }
                                                        else if (diffRsCsdiff >= 14) {
                                                            if (diffTSC < -1) {
                                                                return 2;
                                                            }
                                                            else if (diffTSC >= -1) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (TSC >= 11) {
                                            if (mu_mv_mag_sq < 123) {
                                                if (Cs < 2853) {
                                                    if (SC < 5025) {
                                                        if (CsDiff < 14) {
                                                            return 2;
                                                        }
                                                        else if (CsDiff >= 14) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (SC >= 5025) {
                                                        return 1;
                                                    }
                                                }
                                                else if (Cs >= 2853) {
                                                    if (Rs < 3621) {
                                                        return 8;
                                                    }
                                                    else if (Rs >= 3621) {
                                                        if (negBalance < 48) {
                                                            if (Cs < 3275) {
                                                                return 4;
                                                            }
                                                            else if (Cs >= 3275) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (negBalance >= 48) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (mu_mv_mag_sq >= 123) {
                                                return 4;
                                            }
                                        }
                                    }
                                    else if (refDCval >= 128) {
                                        if (tcor < 85) {
                                            if (refDCval < 141) {
                                                return 1;
                                            }
                                            else if (refDCval >= 141) {
                                                if (AFD < 21) {
                                                    return 8;
                                                }
                                                else if (AFD >= 21) {
                                                    return 4;
                                                }
                                            }
                                        }
                                        else if (tcor >= 85) {
                                            if (Cs < 3095) {
                                                if (refDCval < 139) {
                                                    return 2;
                                                }
                                                else if (refDCval >= 139) {
                                                    return 8;
                                                }
                                            }
                                            else if (Cs >= 3095) {
                                                return 8;
                                            }
                                        }
                                    }
                                }
                            }
                            else if (Rs >= 5295) {
                                if (posBalance < 57) {
                                    return 8;
                                }
                                else if (posBalance >= 57) {
                                    return 4;
                                }
                            }
                        }
                    }
                }
                else if (Cs >= 4177) {
                    if (AFD < 13) {
                        if (ssDCval < 106) {
                            if (ssDCval < 81) {
                                if (Cs < 4874) {
                                    if (negBalance < 47) {
                                        return 8;
                                    }
                                    else if (negBalance >= 47) {
                                        if (RsDiff < 8) {
                                            return 4;
                                        }
                                        else if (RsDiff >= 8) {
                                            return 8;
                                        }
                                    }
                                }
                                else if (Cs >= 4874) {
                                    if (AFD < 8) {
                                        if (Rs < 5772) {
                                            if (SC < 10176) {
                                                if (RsCsDiff < 137) {
                                                    if (Cs < 5117) {
                                                        if (tcor < 93) {
                                                            if (ssDCval < 80) {
                                                                return 8;
                                                            }
                                                            else if (ssDCval >= 80) {
                                                                if (diffAFD < -2) {
                                                                    return 8;
                                                                }
                                                                else if (diffAFD >= -2) {
                                                                    if (gchDC < 1) {
                                                                        if (Rs < 4827) {
                                                                            if (Rs < 4813) {
                                                                                if (SC < 9669) {
                                                                                    if (Cs < 4915) {
                                                                                        return 8;
                                                                                    }
                                                                                    else if (Cs >= 4915) {
                                                                                        return 4;
                                                                                    }
                                                                                }
                                                                                else if (SC >= 9669) {
                                                                                    return 4;
                                                                                }
                                                                            }
                                                                            else if (Rs >= 4813) {
                                                                                return 8;
                                                                            }
                                                                        }
                                                                        else if (Rs >= 4827) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                    else if (gchDC >= 1) {
                                                                        if (Rs < 4935) {
                                                                            return 4;
                                                                        }
                                                                        else if (Rs >= 4935) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (tcor >= 93) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (Cs >= 5117) {
                                                        return 8;
                                                    }
                                                }
                                                else if (RsCsDiff >= 137) {
                                                    if (diffAFD < 1) {
                                                        if (diffRsCsdiff < 33) {
                                                            if (SC < 9628) {
                                                                return 4;
                                                            }
                                                            else if (SC >= 9628) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (diffRsCsdiff >= 33) {
                                                            if (Rs < 4966) {
                                                                return 4;
                                                            }
                                                            else if (Rs >= 4966) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                    else if (diffAFD >= 1) {
                                                        if (Rs < 4811) {
                                                            return 8;
                                                        }
                                                        else if (Rs >= 4811) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (SC >= 10176) {
                                                return 4;
                                            }
                                        }
                                        else if (Rs >= 5772) {
                                            return 8;
                                        }
                                    }
                                    else if (AFD >= 8) {
                                        return 8;
                                    }
                                }
                            }
                            else if (ssDCval >= 81) {
                                return 8;
                            }
                        }
                        else if (ssDCval >= 106) {
                            if (refDCval < 109) {
                                if (MVDiff < 1) {
                                    if (Rs < 6162) {
                                        if (refDCval < 106) {
                                            return 8;
                                        }
                                        else if (refDCval >= 106) {
                                            return 4;
                                        }
                                    }
                                    else if (Rs >= 6162) {
                                        if (RsCsDiff < 315) {
                                            return 4;
                                        }
                                        else if (RsCsDiff >= 315) {
                                            return 8;
                                        }
                                    }
                                }
                                else if (MVDiff >= 1) {
                                    return 8;
                                }
                            }
                            else if (refDCval >= 109) {
                                return 8;
                            }
                        }
                    }
                    else if (AFD >= 13) {
                        if (negBalance < 63) {
                            if (SC < 8529) {
                                if (TSC < 16) {
                                    return 8;
                                }
                                else if (TSC >= 16) {
                                    return 1;
                                }
                            }
                            else if (SC >= 8529) {
                                return 8;
                            }
                        }
                        else if (negBalance >= 63) {
                            if (Rs < 7770) {
                                if (Rs < 7325) {
                                    return 8;
                                }
                                else if (Rs >= 7325) {
                                    return 4;
                                }
                            }
                            else if (Rs >= 7770) {
                                if (RsCsDiff < 1036) {
                                    return 2;
                                }
                                else if (RsCsDiff >= 1036) {
                                    if (diffAFD < 1) {
                                        return 4;
                                    }
                                    else if (diffAFD >= 1) {
                                        return 2;
                                    }
                                }
                            }
                        }
                    }
                }
            }
        }
        else if (tcor >= 96) {
            if (Rs < 583) {
                if (MVDiff < 23) {
                    if (ssDCval < 146) {
                        if (ssDCval < 55) {
                            if (negBalance < 96) {
                                if (refDCval < 38) {
                                    if (MVDiff < 5) {
                                        return 8;
                                    }
                                    else if (MVDiff >= 5) {
                                        if (ssDCval < 23) {
                                            return 8;
                                        }
                                        else if (ssDCval >= 23) {
                                            if (CsDiff < 2) {
                                                return 4;
                                            }
                                            else if (CsDiff >= 2) {
                                                return 2;
                                            }
                                        }
                                    }
                                }
                                else if (refDCval >= 38) {
                                    return 4;
                                }
                            }
                            else if (negBalance >= 96) {
                                if (TSC < 5) {
                                    return 8;
                                }
                                else if (TSC >= 5) {
                                    if (Rs < 263) {
                                        return 2;
                                    }
                                    else if (Rs >= 263) {
                                        return 4;
                                    }
                                }
                            }
                        }
                        else if (ssDCval >= 55) {
                            if (posBalance < 29) {
                                if (Cs < 653) {
                                    if (posBalance < 20) {
                                        return 8;
                                    }
                                    else if (posBalance >= 20) {
                                        if (SC < 497) {
                                            return 8;
                                        }
                                        else if (SC >= 497) {
                                            if (Rs < 190) {
                                                return 4;
                                            }
                                            else if (Rs >= 190) {
                                                return 8;
                                            }
                                        }
                                    }
                                }
                                else if (Cs >= 653) {
                                    if (Cs < 880) {
                                        return 4;
                                    }
                                    else if (Cs >= 880) {
                                        return 8;
                                    }
                                }
                            }
                            else if (posBalance >= 29) {
                                if (ssDCval < 95) {
                                    if (ssDCval < 71) {
                                        if (Rs < 218) {
                                            return 1;
                                        }
                                        else if (Rs >= 218) {
                                            return 8;
                                        }
                                    }
                                    else if (ssDCval >= 71) {
                                        if (negBalance < 14) {
                                            if (SC < 488) {
                                                if (Rs < 179) {
                                                    return 4;
                                                }
                                                else if (Rs >= 179) {
                                                    return 8;
                                                }
                                            }
                                            else if (SC >= 488) {
                                                return 8;
                                            }
                                        }
                                        else if (negBalance >= 14) {
                                            if (Cs < 868) {
                                                if (TSC < 4) {
                                                    if (negBalance < 16) {
                                                        if (Rs < 335) {
                                                            return 8;
                                                        }
                                                        else if (Rs >= 335) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (negBalance >= 16) {
                                                        return 4;
                                                    }
                                                }
                                                else if (TSC >= 4) {
                                                    if (Cs < 285) {
                                                        return 8;
                                                    }
                                                    else if (Cs >= 285) {
                                                        if (posBalance < 44) {
                                                            return 8;
                                                        }
                                                        else if (posBalance >= 44) {
                                                            if (diffMVdiffVal < -46) {
                                                                return 8;
                                                            }
                                                            else if (diffMVdiffVal >= -46) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                            else if (Cs >= 868) {
                                                if (Rs < 373) {
                                                    return 4;
                                                }
                                                else if (Rs >= 373) {
                                                    return 8;
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (ssDCval >= 95) {
                                    if (Rs < 208) {
                                        return 8;
                                    }
                                    else if (Rs >= 208) {
                                        if (mu_mv_mag_sq < 93) {
                                            if (TSC < 3) {
                                                return 4;
                                            }
                                            else if (TSC >= 3) {
                                                return 2;
                                            }
                                        }
                                        else if (mu_mv_mag_sq >= 93) {
                                            if (ssDCval < 113) {
                                                return 8;
                                            }
                                            else if (ssDCval >= 113) {
                                                return 4;
                                            }
                                        }
                                    }
                                }
                            }
                        }
                    }
                    else if (ssDCval >= 146) {
                        if (Cs < 439) {
                            if (SC < 879) {
                                return 4;
                            }
                            else if (SC >= 879) {
                                if (Rs < 471) {
                                    return 2;
                                }
                                else if (Rs >= 471) {
                                    return 4;
                                }
                            }
                        }
                        else if (Cs >= 439) {
                            if (ssDCval < 162) {
                                return 4;
                            }
                            else if (ssDCval >= 162) {
                                if (negBalance < 9) {
                                    return 4;
                                }
                                else if (negBalance >= 9) {
                                    return 2;
                                }
                            }
                        }
                    }
                }
                else if (MVDiff >= 23) {
                    if (refDCval < 152) {
                        if (AFD < 6) {
                            if (negBalance < 10) {
                                return 8;
                            }
                            else if (negBalance >= 10) {
                                if (mcTcor < 98) {
                                    if (MVDiff < 80) {
                                        if (diffMVdiffVal < 1) {
                                            return 2;
                                        }
                                        else if (diffMVdiffVal >= 1) {
                                            return 4;
                                        }
                                    }
                                    else if (MVDiff >= 80) {
                                        if (MVDiff < 224) {
                                            return 2;
                                        }
                                        else if (MVDiff >= 224) {
                                            return 1;
                                        }
                                    }
                                }
                                else if (mcTcor >= 98) {
                                    if (diffMVdiffVal < 26) {
                                        if (Cs < 345) {
                                            if (mu_mv_mag_sq < 76) {
                                                if (diffAFD < -2) {
                                                    return 2;
                                                }
                                                else if (diffAFD >= -2) {
                                                    return 4;
                                                }
                                            }
                                            else if (mu_mv_mag_sq >= 76) {
                                                return 8;
                                            }
                                        }
                                        else if (Cs >= 345) {
                                            return 4;
                                        }
                                    }
                                    else if (diffMVdiffVal >= 26) {
                                        if (ssDCval < 113) {
                                            return 8;
                                        }
                                        else if (ssDCval >= 113) {
                                            return 2;
                                        }
                                    }
                                }
                            }
                        }
                        else if (AFD >= 6) {
                            if (diffAFD < -2) {
                                if (diffAFD < -4) {
                                    return 1;
                                }
                                else if (diffAFD >= -4) {
                                    return 2;
                                }
                            }
                            else if (diffAFD >= -2) {
                                if (mu_mv_mag_sq < 9) {
                                    return 1;
                                }
                                else if (mu_mv_mag_sq >= 9) {
                                    if (SC < 986) {
                                        if (refDCval < 93) {
                                            return 8;
                                        }
                                        else if (refDCval >= 93) {
                                            return 1;
                                        }
                                    }
                                    else if (SC >= 986) {
                                        return 2;
                                    }
                                }
                            }
                        }
                    }
                    else if (refDCval >= 152) {
                        if (Rs < 352) {
                            return 1;
                        }
                        else if (Rs >= 352) {
                            return 4;
                        }
                    }
                }
            }
            else if (Rs >= 583) {
                if (mu_mv_mag_sq < 98) {
                    if (refDCval < 96) {
                        if (mcTcor < 99) {
                            if (Cs < 689) {
                                if (Rs < 1735) {
                                    return 8;
                                }
                                else if (Rs >= 1735) {
                                    return 2;
                                }
                            }
                            else if (Cs >= 689) {
                                if (AFD < 6) {
                                    if (Cs < 1488) {
                                        if (Rs < 1192) {
                                            if (ssDCval < 34) {
                                                return 4;
                                            }
                                            else if (ssDCval >= 34) {
                                                if (Cs < 751) {
                                                    return 1;
                                                }
                                                else if (Cs >= 751) {
                                                    if (AFD < 5) {
                                                        return 8;
                                                    }
                                                    else if (AFD >= 5) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                        }
                                        else if (Rs >= 1192) {
                                            if (ssDCval < 55) {
                                                return 1;
                                            }
                                            else if (ssDCval >= 55) {
                                                if (Rs < 2542) {
                                                    return 8;
                                                }
                                                else if (Rs >= 2542) {
                                                    return 1;
                                                }
                                            }
                                        }
                                    }
                                    else if (Cs >= 1488) {
                                        return 8;
                                    }
                                }
                                else if (AFD >= 6) {
                                    if (Rs < 2523) {
                                        if (SC < 2266) {
                                            if (SC < 2002) {
                                                return 1;
                                            }
                                            else if (SC >= 2002) {
                                                return 2;
                                            }
                                        }
                                        else if (SC >= 2266) {
                                            if (posBalance < 45) {
                                                if (SC < 3799) {
                                                    return 4;
                                                }
                                                else if (SC >= 3799) {
                                                    return 8;
                                                }
                                            }
                                            else if (posBalance >= 45) {
                                                return 4;
                                            }
                                        }
                                    }
                                    else if (Rs >= 2523) {
                                        if (Cs < 2741) {
                                            if (Cs < 1548) {
                                                return 8;
                                            }
                                            else if (Cs >= 1548) {
                                                return 1;
                                            }
                                        }
                                        else if (Cs >= 2741) {
                                            if (posBalance < 57) {
                                                return 4;
                                            }
                                            else if (posBalance >= 57) {
                                                return 8;
                                            }
                                        }
                                    }
                                }
                            }
                        }
                        else if (mcTcor >= 99) {
                            if (posBalance < 10) {
                                if (Cs < 1467) {
                                    return 1;
                                }
                                else if (Cs >= 1467) {
                                    if (SC < 4001) {
                                        return 8;
                                    }
                                    else if (SC >= 4001) {
                                        return 2;
                                    }
                                }
                            }
                            else if (posBalance >= 10) {
                                if (Rs < 650) {
                                    if (TSC < 2) {
                                        return 4;
                                    }
                                    else if (TSC >= 2) {
                                        return 8;
                                    }
                                }
                                else if (Rs >= 650) {
                                    if (tcor < 99) {
                                        if (Rs < 2518) {
                                            if (RsDiff < 1) {
                                                return 1;
                                            }
                                            else if (RsDiff >= 1) {
                                                return 4;
                                            }
                                        }
                                        else if (Rs >= 2518) {
                                            return 2;
                                        }
                                    }
                                    else if (tcor >= 99) {
                                        if (mu_mv_mag_sq < 93) {
                                            if (SC < 1677) {
                                                if (Rs < 876) {
                                                    return 4;
                                                }
                                                else if (Rs >= 876) {
                                                    return 8;
                                                }
                                            }
                                            else if (SC >= 1677) {
                                                if (mu_mv_mag_sq < 48) {
                                                    if (ssDCval < 71) {
                                                        return 2;
                                                    }
                                                    else if (ssDCval >= 71) {
                                                        return 4;
                                                    }
                                                }
                                                else if (mu_mv_mag_sq >= 48) {
                                                    return 4;
                                                }
                                            }
                                        }
                                        else if (mu_mv_mag_sq >= 93) {
                                            if (Rs < 1111) {
                                                return 2;
                                            }
                                            else if (Rs >= 1111) {
                                                return 4;
                                            }
                                        }
                                    }
                                }
                            }
                        }
                    }
                    else if (refDCval >= 96) {
                        if (Rs < 840) {
                            if (mu_mv_mag_sq < 55) {
                                if (SC < 1069) {
                                    return 1;
                                }
                                else if (SC >= 1069) {
                                    return 8;
                                }
                            }
                            else if (mu_mv_mag_sq >= 55) {
                                if (Cs < 175) {
                                    return 8;
                                }
                                else if (Cs >= 175) {
                                    if (Rs < 715) {
                                        return 4;
                                    }
                                    else if (Rs >= 715) {
                                        return 2;
                                    }
                                }
                            }
                        }
                        else if (Rs >= 840) {
                            if (RsDiff < 5) {
                                if (MVDiff < 11) {
                                    if (negBalance < 95) {
                                        return 8;
                                    }
                                    else if (negBalance >= 95) {
                                        return 4;
                                    }
                                }
                                else if (MVDiff >= 11) {
                                    if (ssDCval < 109) {
                                        return 8;
                                    }
                                    else if (ssDCval >= 109) {
                                        if (RsDiff < 2) {
                                            return 8;
                                        }
                                        else if (RsDiff >= 2) {
                                            return 4;
                                        }
                                    }
                                }
                            }
                            else if (RsDiff >= 5) {
                                if (CsDiff < 4) {
                                    if (SC < 3035) {
                                        return 4;
                                    }
                                    else if (SC >= 3035) {
                                        return 8;
                                    }
                                }
                                else if (CsDiff >= 4) {
                                    if (Cs < 1702) {
                                        return 4;
                                    }
                                    else if (Cs >= 1702) {
                                        if (negBalance < 32) {
                                            if (TSC < 8) {
                                                return 4;
                                            }
                                            else if (TSC >= 8) {
                                                return 8;
                                            }
                                        }
                                        else if (negBalance >= 32) {
                                            return 8;
                                        }
                                    }
                                }
                            }
                        }
                    }
                }
                else if (mu_mv_mag_sq >= 98) {
                    if (negBalance < 35) {
                        if (refDCval < 33) {
                            if (Rs < 1112) {
                                if (SC < 1788) {
                                    if (MVDiff < 6) {
                                        return 8;
                                    }
                                    else if (MVDiff >= 6) {
                                        return 1;
                                    }
                                }
                                else if (SC >= 1788) {
                                    if (RsCsDiff < 1) {
                                        if (negBalance < 2) {
                                            if (Cs < 1056) {
                                                return 8;
                                            }
                                            else if (Cs >= 1056) {
                                                return 2;
                                            }
                                        }
                                        else if (negBalance >= 2) {
                                            if (Cs < 1258) {
                                                return 4;
                                            }
                                            else if (Cs >= 1258) {
                                                return 2;
                                            }
                                        }
                                    }
                                    else if (RsCsDiff >= 1) {
                                        return 8;
                                    }
                                }
                            }
                            else if (Rs >= 1112) {
                                if (Cs < 713) {
                                    return 4;
                                }
                                else if (Cs >= 713) {
                                    return 8;
                                }
                            }
                        }
                        else if (refDCval >= 33) {
                            if (Cs < 2144) {
                                if (diffTSC < -2) {
                                    if (diffTSC < -3) {
                                        return 8;
                                    }
                                    else if (diffTSC >= -3) {
                                        if (mcTcor < 98) {
                                            return 1;
                                        }
                                        else if (mcTcor >= 98) {
                                            return 8;
                                        }
                                    }
                                }
                                else if (diffTSC >= -2) {
                                    if (diffMVdiffVal < -34) {
                                        return 4;
                                    }
                                    else if (diffMVdiffVal >= -34) {
                                        if (CsDiff < 4) {
                                            if (diffMVdiffVal < -3) {
                                                if (TSC < 2) {
                                                    if (mu_mv_mag_sq < 125) {
                                                        if (Cs < 1136) {
                                                            return 1;
                                                        }
                                                        else if (Cs >= 1136) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (mu_mv_mag_sq >= 125) {
                                                        return 8;
                                                    }
                                                }
                                                else if (TSC >= 2) {
                                                    return 8;
                                                }
                                            }
                                            else if (diffMVdiffVal >= -3) {
                                                if (RsCsDiff < 1) {
                                                    if (negBalance < 12) {
                                                        return 8;
                                                    }
                                                    else if (negBalance >= 12) {
                                                        if (ssDCval < 110) {
                                                            return 8;
                                                        }
                                                        else if (ssDCval >= 110) {
                                                            if (Cs < 1283) {
                                                                return 8;
                                                            }
                                                            else if (Cs >= 1283) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (RsCsDiff >= 1) {
                                                    return 8;
                                                }
                                            }
                                        }
                                        else if (CsDiff >= 4) {
                                            if (Rs < 1315) {
                                                return 4;
                                            }
                                            else if (Rs >= 1315) {
                                                if (MVDiff < 1) {
                                                    if (SC < 3933) {
                                                        return 8;
                                                    }
                                                    else if (SC >= 3933) {
                                                        return 1;
                                                    }
                                                }
                                                else if (MVDiff >= 1) {
                                                    return 8;
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                            else if (Cs >= 2144) {
                                if (Rs < 3779) {
                                    if (SC < 7764) {
                                        if (negBalance < 3) {
                                            if (Cs < 3876) {
                                                if (RsCsDiff < 2) {
                                                    if (negBalance < 1) {
                                                        return 1;
                                                    }
                                                    else if (negBalance >= 1) {
                                                        if (posBalance < 21) {
                                                            return 1;
                                                        }
                                                        else if (posBalance >= 21) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                                else if (RsCsDiff >= 2) {
                                                    return 1;
                                                }
                                            }
                                            else if (Cs >= 3876) {
                                                return 8;
                                            }
                                        }
                                        else if (negBalance >= 3) {
                                            if (RsCsDiff < 7) {
                                                if (posBalance < 51) {
                                                    if (SC < 4730) {
                                                        if (AFD < 2) {
                                                            return 4;
                                                        }
                                                        else if (AFD >= 2) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (SC >= 4730) {
                                                        return 8;
                                                    }
                                                }
                                                else if (posBalance >= 51) {
                                                    return 4;
                                                }
                                            }
                                            else if (RsCsDiff >= 7) {
                                                if (mu_mv_mag_sq < 125) {
                                                    if (Cs < 3092) {
                                                        if (negBalance < 13) {
                                                            if (Cs < 2685) {
                                                                return 1;
                                                            }
                                                            else if (Cs >= 2685) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (negBalance >= 13) {
                                                            if (MVDiff < 1) {
                                                                if (SC < 6156) {
                                                                    if (diffTSC < 1) {
                                                                        if (diffAFD < -1) {
                                                                            if (Cs < 2786) {
                                                                                return 1;
                                                                            }
                                                                            else if (Cs >= 2786) {
                                                                                return 8;
                                                                            }
                                                                        }
                                                                        else if (diffAFD >= -1) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                    else if (diffTSC >= 1) {
                                                                        if (RsDiff < 4) {
                                                                            return 8;
                                                                        }
                                                                        else if (RsDiff >= 4) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                }
                                                                else if (SC >= 6156) {
                                                                    if (negBalance < 26) {
                                                                        if (Cs < 3010) {
                                                                            if (Cs < 2939) {
                                                                                return 8;
                                                                            }
                                                                            else if (Cs >= 2939) {
                                                                                return 1;
                                                                            }
                                                                        }
                                                                        else if (Cs >= 3010) {
                                                                            if (ssDCval < 109) {
                                                                                return 8;
                                                                            }
                                                                            else if (ssDCval >= 109) {
                                                                                return 1;
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (negBalance >= 26) {
                                                                        return 1;
                                                                    }
                                                                }
                                                            }
                                                            else if (MVDiff >= 1) {
                                                                if (TSC < 6) {
                                                                    return 8;
                                                                }
                                                                else if (TSC >= 6) {
                                                                    return 1;
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (Cs >= 3092) {
                                                        if (Rs < 3244) {
                                                            if (ssDCval < 108) {
                                                                if (Rs < 3228) {
                                                                    return 8;
                                                                }
                                                                else if (Rs >= 3228) {
                                                                    return 1;
                                                                }
                                                            }
                                                            else if (ssDCval >= 108) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (Rs >= 3244) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                                else if (mu_mv_mag_sq >= 125) {
                                                    return 8;
                                                }
                                            }
                                        }
                                    }
                                    else if (SC >= 7764) {
                                        return 1;
                                    }
                                }
                                else if (Rs >= 3779) {
                                    return 8;
                                }
                            }
                        }
                    }
                    else if (negBalance >= 35) {
                        if (refDCval < 83) {
                            if (SC < 3110) {
                                if (diffAFD < 1) {
                                    return 8;
                                }
                                else if (diffAFD >= 1) {
                                    if (Rs < 1748) {
                                        return 8;
                                    }
                                    else if (Rs >= 1748) {
                                        return 4;
                                    }
                                }
                            }
                            else if (SC >= 3110) {
                                if (AFD < 4) {
                                    return 8;
                                }
                                else if (AFD >= 4) {
                                    if (negBalance < 45) {
                                        if (Cs < 2697) {
                                            if (SC < 3510) {
                                                if (mcTcor < 97) {
                                                    if (refDCval < 77) {
                                                        return 8;
                                                    }
                                                    else if (refDCval >= 77) {
                                                        return 4;
                                                    }
                                                }
                                                else if (mcTcor >= 97) {
                                                    if (Cs < 1272) {
                                                        return 8;
                                                    }
                                                    else if (Cs >= 1272) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                            else if (SC >= 3510) {
                                                if (posBalance < 29) {
                                                    return 8;
                                                }
                                                else if (posBalance >= 29) {
                                                    return 4;
                                                }
                                            }
                                        }
                                        else if (Cs >= 2697) {
                                            return 8;
                                        }
                                    }
                                    else if (negBalance >= 45) {
                                        if (TSCindex < 7) {
                                            if (MVDiff < 10) {
                                                return 8;
                                            }
                                            else if (MVDiff >= 10) {
                                                return 4;
                                            }
                                        }
                                        else if (TSCindex >= 7) {
                                            if (Cs < 1588) {
                                                return 2;
                                            }
                                            else if (Cs >= 1588) {
                                                return 4;
                                            }
                                        }
                                    }
                                }
                            }
                        }
                        else if (refDCval >= 83) {
                            if (refDCval < 123) {
                                if (refDCval < 118) {
                                    if (mu_mv_mag_sq < 123) {
                                        if (Rs < 2478) {
                                            if (AFD < 4) {
                                                return 8;
                                            }
                                            else if (AFD >= 4) {
                                                if (ssDCval < 109) {
                                                    return 4;
                                                }
                                                else if (ssDCval >= 109) {
                                                    return 8;
                                                }
                                            }
                                        }
                                        else if (Rs >= 2478) {
                                            if (negBalance < 48) {
                                                if (Rs < 2713) {
                                                    return 8;
                                                }
                                                else if (Rs >= 2713) {
                                                    if (Rs < 2717) {
                                                        return 4;
                                                    }
                                                    else if (Rs >= 2717) {
                                                        if (RsCsDiff < 37) {
                                                            if (negBalance < 46) {
                                                                return 8;
                                                            }
                                                            else if (negBalance >= 46) {
                                                                if (refDCval < 114) {
                                                                    if (Cs < 3033) {
                                                                        if (Rs < 2747) {
                                                                            if (ssDCval < 112) {
                                                                                return 8;
                                                                            }
                                                                            else if (ssDCval >= 112) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                        else if (Rs >= 2747) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                    else if (Cs >= 3033) {
                                                                        if (SC < 5999) {
                                                                            return 4;
                                                                        }
                                                                        else if (SC >= 5999) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                }
                                                                else if (refDCval >= 114) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                        else if (RsCsDiff >= 37) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (negBalance >= 48) {
                                                if (Cs < 2886) {
                                                    return 8;
                                                }
                                                else if (Cs >= 2886) {
                                                    if (tcor < 97) {
                                                        return 8;
                                                    }
                                                    else if (tcor >= 97) {
                                                        if (CsDiff < 5) {
                                                            if (Rs < 2647) {
                                                                return 8;
                                                            }
                                                            else if (Rs >= 2647) {
                                                                if (refDCval < 115) {
                                                                    if (SC < 6217) {
                                                                        return 4;
                                                                    }
                                                                    else if (SC >= 6217) {
                                                                        return 8;
                                                                    }
                                                                }
                                                                else if (refDCval >= 115) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                        else if (CsDiff >= 5) {
                                                            if (ssDCval < 113) {
                                                                if (refDCval < 113) {
                                                                    if (RsCsDiff < 46) {
                                                                        if (Rs < 2935) {
                                                                            return 8;
                                                                        }
                                                                        else if (Rs >= 2935) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                    else if (RsCsDiff >= 46) {
                                                                        return 8;
                                                                    }
                                                                }
                                                                else if (refDCval >= 113) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (ssDCval >= 113) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (mu_mv_mag_sq >= 123) {
                                        if (Rs < 1828) {
                                            return 4;
                                        }
                                        else if (Rs >= 1828) {
                                            if (refDCval < 112) {
                                                return 8;
                                            }
                                            else if (refDCval >= 112) {
                                                if (negBalance < 48) {
                                                    return 8;
                                                }
                                                else if (negBalance >= 48) {
                                                    if (Rs < 2796) {
                                                        return 8;
                                                    }
                                                    else if (Rs >= 2796) {
                                                        if (CsDiff < 5) {
                                                            if (posBalance < 57) {
                                                                return 8;
                                                            }
                                                            else if (posBalance >= 57) {
                                                                if (Cs < 3078) {
                                                                    return 4;
                                                                }
                                                                else if (Cs >= 3078) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                        else if (CsDiff >= 5) {
                                                            if (Rs < 2807) {
                                                                return 4;
                                                            }
                                                            else if (Rs >= 2807) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (refDCval >= 118) {
                                    if (Cs < 4101) {
                                        return 4;
                                    }
                                    else if (Cs >= 4101) {
                                        return 8;
                                    }
                                }
                            }
                            else if (refDCval >= 123) {
                                return 8;
                            }
                        }
                    }
                }
            }
        }
        return 4;
    }
    mfxU16 AGOPSelectF(mfxU32 Rs, mfxU32 CsDiff, mfxU32 TSC, mfxI32 diffAFD, mfxU32 Scindex, mfxU32 RsDiff, mfxI32 diffMVdiffVal, mfxU32 RsCsDiff, mfxU32 TSCindex, mfxI16 tcor, mfxU32 refDCval, mfxU32 MVDiff, mfxU32 AFD, mfxI32 diffTSC, mfxI32 diffRsCsdiff, mfxU32 posBalance, mfxU32 SC, mfxU32 mu_mv_mag_sq, mfxU32 negBalance, mfxI16 mcTcor, mfxU32 gchDC, mfxU32 Cs, mfxU32 ssDCval) {
        if (MVDiff < 6) {
            if (AFD < 7) {
                if (mu_mv_mag_sq < 110) {
                    if (posBalance < 11) {
                        if (Cs < 1787) {
                            return 8;
                        }
                        else if (Cs >= 1787) {
                            if (Cs < 4248) {
                                if (Cs < 1891) {
                                    return 1;
                                }
                                else if (Cs >= 1891) {
                                    if (SC < 7709) {
                                        return 8;
                                    }
                                    else if (SC >= 7709) {
                                        if (Rs < 3770) {
                                            return 1;
                                        }
                                        else if (Rs >= 3770) {
                                            if (SC < 7736) {
                                                return 1;
                                            }
                                            else if (SC >= 7736) {
                                                return 8;
                                            }
                                        }
                                    }
                                }
                            }
                            else if (Cs >= 4248) {
                                if (Rs < 4729) {
                                    return 8;
                                }
                                else if (Rs >= 4729) {
                                    if (Rs < 5376) {
                                        return 1;
                                    }
                                    else if (Rs >= 5376) {
                                        return 8;
                                    }
                                }
                            }
                        }
                    }
                    else if (posBalance >= 11) {
                        if (Rs < 571) {
                            if (posBalance < 40) {
                                if (Rs < 461) {
                                    if (mcTcor < 98) {
                                        if (SC < 498) {
                                            return 1;
                                        }
                                        else if (SC >= 498) {
                                            return 4;
                                        }
                                    }
                                    else if (mcTcor >= 98) {
                                        return 4;
                                    }
                                }
                                else if (Rs >= 461) {
                                    if (refDCval < 165) {
                                        if (Rs < 467) {
                                            return 2;
                                        }
                                        else if (Rs >= 467) {
                                            if (posBalance < 35) {
                                                return 4;
                                            }
                                            else if (posBalance >= 35) {
                                                return 8;
                                            }
                                        }
                                    }
                                    else if (refDCval >= 165) {
                                        if (Rs < 525) {
                                            return 2;
                                        }
                                        else if (Rs >= 525) {
                                            return 4;
                                        }
                                    }
                                }
                            }
                            else if (posBalance >= 40) {
                                if (negBalance < 32) {
                                    if (Rs < 152) {
                                        if (Cs < 375) {
                                            return 1;
                                        }
                                        else if (Cs >= 375) {
                                            return 8;
                                        }
                                    }
                                    else if (Rs >= 152) {
                                        if (mcTcor < 96) {
                                            if (ssDCval < 47) {
                                                return 4;
                                            }
                                            else if (ssDCval >= 47) {
                                                return 8;
                                            }
                                        }
                                        else if (mcTcor >= 96) {
                                            if (posBalance < 59) {
                                                if (SC < 1975) {
                                                    return 4;
                                                }
                                                else if (SC >= 1975) {
                                                    return 8;
                                                }
                                            }
                                            else if (posBalance >= 59) {
                                                return 8;
                                            }
                                        }
                                    }
                                }
                                else if (negBalance >= 32) {
                                    if (TSCindex < 5) {
                                        if (Cs < 843) {
                                            return 2;
                                        }
                                        else if (Cs >= 843) {
                                            return 4;
                                        }
                                    }
                                    else if (TSCindex >= 5) {
                                        if (negBalance < 53) {
                                            if (posBalance < 62) {
                                                return 4;
                                            }
                                            else if (posBalance >= 62) {
                                                return 8;
                                            }
                                        }
                                        else if (negBalance >= 53) {
                                            return 8;
                                        }
                                    }
                                }
                            }
                        }
                        else if (Rs >= 571) {
                            if (MVDiff < 1) {
                                if (AFD < 5) {
                                    if (negBalance < 32) {
                                        if (CsDiff < 5) {
                                            return 8;
                                        }
                                        else if (CsDiff >= 5) {
                                            return 1;
                                        }
                                    }
                                    else if (negBalance >= 32) {
                                        return 4;
                                    }
                                }
                                else if (AFD >= 5) {
                                    if (SC < 1121) {
                                        return 8;
                                    }
                                    else if (SC >= 1121) {
                                        if (Cs < 2756) {
                                            if (SC < 3464) {
                                                if (TSC < 6) {
                                                    return 4;
                                                }
                                                else if (TSC >= 6) {
                                                    return 2;
                                                }
                                            }
                                            else if (SC >= 3464) {
                                                return 4;
                                            }
                                        }
                                        else if (Cs >= 2756) {
                                            return 1;
                                        }
                                    }
                                }
                            }
                            else if (MVDiff >= 1) {
                                if (ssDCval < 54) {
                                    if (refDCval < 47) {
                                        if (tcor < 86) {
                                            return 4;
                                        }
                                        else if (tcor >= 86) {
                                            if (mu_mv_mag_sq < 97) {
                                                if (SC < 2214) {
                                                    return 8;
                                                }
                                                else if (SC >= 2214) {
                                                    if (mu_mv_mag_sq < 94) {
                                                        return 4;
                                                    }
                                                    else if (mu_mv_mag_sq >= 94) {
                                                        return 2;
                                                    }
                                                }
                                            }
                                            else if (mu_mv_mag_sq >= 97) {
                                                return 8;
                                            }
                                        }
                                    }
                                    else if (refDCval >= 47) {
                                        if (SC < 1932) {
                                            if (Cs < 691) {
                                                if (posBalance < 56) {
                                                    if (Rs < 645) {
                                                        return 8;
                                                    }
                                                    else if (Rs >= 645) {
                                                        return 4;
                                                    }
                                                }
                                                else if (posBalance >= 56) {
                                                    if (Cs < 464) {
                                                        return 4;
                                                    }
                                                    else if (Cs >= 464) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                            else if (Cs >= 691) {
                                                if (mu_mv_mag_sq < 45) {
                                                    if (SC < 1733) {
                                                        return 1;
                                                    }
                                                    else if (SC >= 1733) {
                                                        return 4;
                                                    }
                                                }
                                                else if (mu_mv_mag_sq >= 45) {
                                                    return 4;
                                                }
                                            }
                                        }
                                        else if (SC >= 1932) {
                                            if (Rs < 1555) {
                                                if (mu_mv_mag_sq < 59) {
                                                    return 1;
                                                }
                                                else if (mu_mv_mag_sq >= 59) {
                                                    if (SC < 2060) {
                                                        if (Cs < 778) {
                                                            return 8;
                                                        }
                                                        else if (Cs >= 778) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (SC >= 2060) {
                                                        if (ssDCval < 51) {
                                                            return 1;
                                                        }
                                                        else if (ssDCval >= 51) {
                                                            if (SC < 2343) {
                                                                if (mcTcor < 96) {
                                                                    return 4;
                                                                }
                                                                else if (mcTcor >= 96) {
                                                                    return 1;
                                                                }
                                                            }
                                                            else if (SC >= 2343) {
                                                                if (Cs < 988) {
                                                                    if (Rs < 1444) {
                                                                        return 1;
                                                                    }
                                                                    else if (Rs >= 1444) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (Cs >= 988) {
                                                                    return 1;
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                            else if (Rs >= 1555) {
                                                if (ssDCval < 51) {
                                                    return 8;
                                                }
                                                else if (ssDCval >= 51) {
                                                    return 4;
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (ssDCval >= 54) {
                                    if (SC < 880) {
                                        return 1;
                                    }
                                    else if (SC >= 880) {
                                        if (ssDCval < 91) {
                                            if (AFD < 5) {
                                                if (refDCval < 73) {
                                                    if (mu_mv_mag_sq < 77) {
                                                        if (SC < 1579) {
                                                            return 8;
                                                        }
                                                        else if (SC >= 1579) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (mu_mv_mag_sq >= 77) {
                                                        return 8;
                                                    }
                                                }
                                                else if (refDCval >= 73) {
                                                    if (Rs < 2803) {
                                                        if (ssDCval < 81) {
                                                            return 4;
                                                        }
                                                        else if (ssDCval >= 81) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (Rs >= 2803) {
                                                        return 1;
                                                    }
                                                }
                                            }
                                            else if (AFD >= 5) {
                                                if (negBalance < 19) {
                                                    return 1;
                                                }
                                                else if (negBalance >= 19) {
                                                    if (Cs < 530) {
                                                        return 2;
                                                    }
                                                    else if (Cs >= 530) {
                                                        if (posBalance < 56) {
                                                            if (negBalance < 38) {
                                                                if (mu_mv_mag_sq < 73) {
                                                                    return 1;
                                                                }
                                                                else if (mu_mv_mag_sq >= 73) {
                                                                    if (RsCsDiff < 71) {
                                                                        if (mu_mv_mag_sq < 82) {
                                                                            if (ssDCval < 57) {
                                                                                return 8;
                                                                            }
                                                                            else if (ssDCval >= 57) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                        else if (mu_mv_mag_sq >= 82) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                    else if (RsCsDiff >= 71) {
                                                                        if (AFD < 6) {
                                                                            return 8;
                                                                        }
                                                                        else if (AFD >= 6) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                            else if (negBalance >= 38) {
                                                                if (RsDiff < 3) {
                                                                    if (posBalance < 48) {
                                                                        return 1;
                                                                    }
                                                                    else if (posBalance >= 48) {
                                                                        return 8;
                                                                    }
                                                                }
                                                                else if (RsDiff >= 3) {
                                                                    if (Cs < 1619) {
                                                                        if (mu_mv_mag_sq < 100) {
                                                                            if (Rs < 2185) {
                                                                                if (Rs < 1512) {
                                                                                    return 1;
                                                                                }
                                                                                else if (Rs >= 1512) {
                                                                                    return 4;
                                                                                }
                                                                            }
                                                                            else if (Rs >= 2185) {
                                                                                return 1;
                                                                            }
                                                                        }
                                                                        else if (mu_mv_mag_sq >= 100) {
                                                                            if (Rs < 2391) {
                                                                                return 4;
                                                                            }
                                                                            else if (Rs >= 2391) {
                                                                                return 8;
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (Cs >= 1619) {
                                                                        if (Cs < 1942) {
                                                                            return 4;
                                                                        }
                                                                        else if (Cs >= 1942) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (posBalance >= 56) {
                                                            if (negBalance < 48) {
                                                                if (TSC < 6) {
                                                                    return 8;
                                                                }
                                                                else if (TSC >= 6) {
                                                                    if (Cs < 1543) {
                                                                        return 2;
                                                                    }
                                                                    else if (Cs >= 1543) {
                                                                        return 8;
                                                                    }
                                                                }
                                                            }
                                                            else if (negBalance >= 48) {
                                                                return 2;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (ssDCval >= 91) {
                                            if (refDCval < 158) {
                                                if (posBalance < 36) {
                                                    if (RsCsDiff < 38) {
                                                        if (Rs < 1297) {
                                                            return 8;
                                                        }
                                                        else if (Rs >= 1297) {
                                                            if (Cs < 1224) {
                                                                return 8;
                                                            }
                                                            else if (Cs >= 1224) {
                                                                if (MVDiff < 3) {
                                                                    return 4;
                                                                }
                                                                else if (MVDiff >= 3) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (RsCsDiff >= 38) {
                                                        return 4;
                                                    }
                                                }
                                                else if (posBalance >= 36) {
                                                    if (gchDC < 2) {
                                                        return 8;
                                                    }
                                                    else if (gchDC >= 2) {
                                                        if (Rs < 2608) {
                                                            return 4;
                                                        }
                                                        else if (Rs >= 2608) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (refDCval >= 158) {
                                                if (Rs < 750) {
                                                    return 4;
                                                }
                                                else if (Rs >= 750) {
                                                    return 8;
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                        }
                    }
                }
                else if (mu_mv_mag_sq >= 110) {
                    if (negBalance < 34) {
                        if (Rs < 583) {
                            if (Cs < 501) {
                                if (TSC < 1) {
                                    return 8;
                                }
                                else if (TSC >= 1) {
                                    if (Cs < 347) {
                                        if (Cs < 307) {
                                            return 8;
                                        }
                                        else if (Cs >= 307) {
                                            if (Cs < 309) {
                                                return 4;
                                            }
                                            else if (Cs >= 309) {
                                                if (negBalance < 13) {
                                                    return 4;
                                                }
                                                else if (negBalance >= 13) {
                                                    if (Cs < 316) {
                                                        return 8;
                                                    }
                                                    else if (Cs >= 316) {
                                                        if (SC < 502) {
                                                            return 4;
                                                        }
                                                        else if (SC >= 502) {
                                                            if (AFD < 2) {
                                                                return 8;
                                                            }
                                                            else if (AFD >= 2) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (Cs >= 347) {
                                        return 8;
                                    }
                                }
                            }
                            else if (Cs >= 501) {
                                if (SC < 1265) {
                                    if (gchDC < 1) {
                                        return 4;
                                    }
                                    else if (gchDC >= 1) {
                                        return 8;
                                    }
                                }
                                else if (SC >= 1265) {
                                    return 8;
                                }
                            }
                        }
                        else if (Rs >= 583) {
                            if (negBalance < 3) {
                                if (Rs < 3779) {
                                    if (ssDCval < 97) {
                                        if (diffMVdiffVal < -2) {
                                            if (Cs < 1093) {
                                                return 8;
                                            }
                                            else if (Cs >= 1093) {
                                                return 4;
                                            }
                                        }
                                        else if (diffMVdiffVal >= -2) {
                                            if (diffRsCsdiff < -2) {
                                                if (Rs < 1476) {
                                                    return 2;
                                                }
                                                else if (Rs >= 1476) {
                                                    return 8;
                                                }
                                            }
                                            else if (diffRsCsdiff >= -2) {
                                                return 8;
                                            }
                                        }
                                    }
                                    else if (ssDCval >= 97) {
                                        if (Rs < 3774) {
                                            if (mcTcor < 99) {
                                                if (diffRsCsdiff < -16) {
                                                    return 1;
                                                }
                                                else if (diffRsCsdiff >= -16) {
                                                    if (mu_mv_mag_sq < 124) {
                                                        if (diffRsCsdiff < 12) {
                                                            return 8;
                                                        }
                                                        else if (diffRsCsdiff >= 12) {
                                                            return 1;
                                                        }
                                                    }
                                                    else if (mu_mv_mag_sq >= 124) {
                                                        if (Cs < 1545) {
                                                            return 1;
                                                        }
                                                        else if (Cs >= 1545) {
                                                            if (posBalance < 22) {
                                                                return 8;
                                                            }
                                                            else if (posBalance >= 22) {
                                                                return 1;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                            else if (mcTcor >= 99) {
                                                if (RsDiff < 2) {
                                                    if (negBalance < 1) {
                                                        if (mcTcor < 100) {
                                                            if (Rs < 3547) {
                                                                return 8;
                                                            }
                                                            else if (Rs >= 3547) {
                                                                return 1;
                                                            }
                                                        }
                                                        else if (mcTcor >= 100) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (negBalance >= 1) {
                                                        if (diffRsCsdiff < -1) {
                                                            if (RsDiff < 1) {
                                                                return 8;
                                                            }
                                                            else if (RsDiff >= 1) {
                                                                if (ssDCval < 104) {
                                                                    return 1;
                                                                }
                                                                else if (ssDCval >= 104) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                        else if (diffRsCsdiff >= -1) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                                else if (RsDiff >= 2) {
                                                    return 1;
                                                }
                                            }
                                        }
                                        else if (Rs >= 3774) {
                                            return 1;
                                        }
                                    }
                                }
                                else if (Rs >= 3779) {
                                    return 8;
                                }
                            }
                            else if (negBalance >= 3) {
                                if (CsDiff < 3) {
                                    if (mu_mv_mag_sq < 123) {
                                        if (posBalance < 20) {
                                            if (Rs < 1099) {
                                                if (diffMVdiffVal < -2) {
                                                    if (SC < 1882) {
                                                        return 1;
                                                    }
                                                    else if (SC >= 1882) {
                                                        return 8;
                                                    }
                                                }
                                                else if (diffMVdiffVal >= -2) {
                                                    return 8;
                                                }
                                            }
                                            else if (Rs >= 1099) {
                                                if (posBalance < 10) {
                                                    if (refDCval < 116) {
                                                        return 8;
                                                    }
                                                    else if (refDCval >= 116) {
                                                        if (Cs < 2274) {
                                                            return 4;
                                                        }
                                                        else if (Cs >= 2274) {
                                                            if (Rs < 2452) {
                                                                return 8;
                                                            }
                                                            else if (Rs >= 2452) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (posBalance >= 10) {
                                                    if (Rs < 1116) {
                                                        return 4;
                                                    }
                                                    else if (Rs >= 1116) {
                                                        if (refDCval < 117) {
                                                            return 8;
                                                        }
                                                        else if (refDCval >= 117) {
                                                            if (ssDCval < 120) {
                                                                return 4;
                                                            }
                                                            else if (ssDCval >= 120) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (posBalance >= 20) {
                                            if (diffMVdiffVal < -3) {
                                                if (Rs < 1283) {
                                                    return 8;
                                                }
                                                else if (Rs >= 1283) {
                                                    if (Cs < 2820) {
                                                        return 1;
                                                    }
                                                    else if (Cs >= 2820) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                            else if (diffMVdiffVal >= -3) {
                                                if (ssDCval < 107) {
                                                    return 8;
                                                }
                                                else if (ssDCval >= 107) {
                                                    if (SC < 4559) {
                                                        if (Rs < 1079) {
                                                            return 4;
                                                        }
                                                        else if (Rs >= 1079) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (SC >= 4559) {
                                                        if (diffRsCsdiff < -3) {
                                                            if (ssDCval < 108) {
                                                                return 1;
                                                            }
                                                            else if (ssDCval >= 108) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (diffRsCsdiff >= -3) {
                                                            if (SC < 4575) {
                                                                return 4;
                                                            }
                                                            else if (SC >= 4575) {
                                                                if (RsDiff < 2) {
                                                                    return 8;
                                                                }
                                                                else if (RsDiff >= 2) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (mu_mv_mag_sq >= 123) {
                                        return 8;
                                    }
                                }
                                else if (CsDiff >= 3) {
                                    if (mcTcor < 96) {
                                        if (ssDCval < 115) {
                                            if (posBalance < 34) {
                                                return 8;
                                            }
                                            else if (posBalance >= 34) {
                                                if (AFD < 4) {
                                                    if (SC < 3722) {
                                                        return 8;
                                                    }
                                                    else if (SC >= 3722) {
                                                        return 4;
                                                    }
                                                }
                                                else if (AFD >= 4) {
                                                    return 8;
                                                }
                                            }
                                        }
                                        else if (ssDCval >= 115) {
                                            if (diffMVdiffVal < -1) {
                                                if (SC < 2777) {
                                                    return 1;
                                                }
                                                else if (SC >= 2777) {
                                                    return 4;
                                                }
                                            }
                                            else if (diffMVdiffVal >= -1) {
                                                if (refDCval < 119) {
                                                    return 4;
                                                }
                                                else if (refDCval >= 119) {
                                                    return 8;
                                                }
                                            }
                                        }
                                    }
                                    else if (mcTcor >= 96) {
                                        if (Cs < 3139) {
                                            if (ssDCval < 110) {
                                                if (ssDCval < 70) {
                                                    return 8;
                                                }
                                                else if (ssDCval >= 70) {
                                                    if (negBalance < 30) {
                                                        if (Cs < 3076) {
                                                            if (posBalance < 48) {
                                                                if (RsCsDiff < 19) {
                                                                    if (SC < 6221) {
                                                                        if (SC < 4614) {
                                                                            return 1;
                                                                        }
                                                                        else if (SC >= 4614) {
                                                                            if (SC < 6158) {
                                                                                return 8;
                                                                            }
                                                                            else if (SC >= 6158) {
                                                                                if (AFD < 2) {
                                                                                    return 8;
                                                                                }
                                                                                else if (AFD >= 2) {
                                                                                    return 1;
                                                                                }
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (SC >= 6221) {
                                                                        if (Rs < 3498) {
                                                                            return 1;
                                                                        }
                                                                        else if (Rs >= 3498) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                }
                                                                else if (RsCsDiff >= 19) {
                                                                    if (mu_mv_mag_sq < 116) {
                                                                        if (Rs < 2112) {
                                                                            return 1;
                                                                        }
                                                                        else if (Rs >= 2112) {
                                                                            if (ssDCval < 109) {
                                                                                return 8;
                                                                            }
                                                                            else if (ssDCval >= 109) {
                                                                                return 1;
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (mu_mv_mag_sq >= 116) {
                                                                        return 1;
                                                                    }
                                                                }
                                                            }
                                                            else if (posBalance >= 48) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (Cs >= 3076) {
                                                            if (diffRsCsdiff < 16) {
                                                                return 8;
                                                            }
                                                            else if (diffRsCsdiff >= 16) {
                                                                return 1;
                                                            }
                                                        }
                                                    }
                                                    else if (negBalance >= 30) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                            else if (ssDCval >= 110) {
                                                return 8;
                                            }
                                        }
                                        else if (Cs >= 3139) {
                                            return 8;
                                        }
                                    }
                                }
                            }
                        }
                    }
                    else if (negBalance >= 34) {
                        if (refDCval < 84) {
                            if (SC < 3582) {
                                if (TSC < 6) {
                                    if (Rs < 1789) {
                                        if (negBalance < 63) {
                                            if (ssDCval < 76) {
                                                return 8;
                                            }
                                            else if (ssDCval >= 76) {
                                                if (gchDC < 1) {
                                                    return 8;
                                                }
                                                else if (gchDC >= 1) {
                                                    if (Rs < 1712) {
                                                        return 8;
                                                    }
                                                    else if (Rs >= 1712) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                        }
                                        else if (negBalance >= 63) {
                                            return 4;
                                        }
                                    }
                                    else if (Rs >= 1789) {
                                        if (MVDiff < 2) {
                                            if (SC < 3532) {
                                                if (Rs < 1793) {
                                                    return 2;
                                                }
                                                else if (Rs >= 1793) {
                                                    if (Cs < 1629) {
                                                        if (diffRsCsdiff < -4) {
                                                            if (Rs < 1902) {
                                                                return 8;
                                                            }
                                                            else if (Rs >= 1902) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (diffRsCsdiff >= -4) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (Cs >= 1629) {
                                                        if (SC < 3477) {
                                                            return 4;
                                                        }
                                                        else if (SC >= 3477) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (SC >= 3532) {
                                                if (mcTcor < 95) {
                                                    return 8;
                                                }
                                                else if (mcTcor >= 95) {
                                                    return 2;
                                                }
                                            }
                                        }
                                        else if (MVDiff >= 2) {
                                            if (ssDCval < 73) {
                                                if (refDCval < 73) {
                                                    return 8;
                                                }
                                                else if (refDCval >= 73) {
                                                    return 4;
                                                }
                                            }
                                            else if (ssDCval >= 73) {
                                                if (RsCsDiff < 19) {
                                                    return 4;
                                                }
                                                else if (RsCsDiff >= 19) {
                                                    return 8;
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (TSC >= 6) {
                                    if (refDCval < 74) {
                                        if (posBalance < 58) {
                                            if (mu_mv_mag_sq < 114) {
                                                if (diffRsCsdiff < 7) {
                                                    if (RsDiff < 4) {
                                                        return 2;
                                                    }
                                                    else if (RsDiff >= 4) {
                                                        return 8;
                                                    }
                                                }
                                                else if (diffRsCsdiff >= 7) {
                                                    return 2;
                                                }
                                            }
                                            else if (mu_mv_mag_sq >= 114) {
                                                return 8;
                                            }
                                        }
                                        else if (posBalance >= 58) {
                                            if (Rs < 1902) {
                                                return 2;
                                            }
                                            else if (Rs >= 1902) {
                                                return 8;
                                            }
                                        }
                                    }
                                    else if (refDCval >= 74) {
                                        if (CsDiff < 4) {
                                            return 4;
                                        }
                                        else if (CsDiff >= 4) {
                                            if (Rs < 1892) {
                                                return 8;
                                            }
                                            else if (Rs >= 1892) {
                                                return 4;
                                            }
                                        }
                                    }
                                }
                            }
                            else if (SC >= 3582) {
                                if (tcor < 95) {
                                    if (SC < 9308) {
                                        if (tcor < 94) {
                                            return 8;
                                        }
                                        else if (tcor >= 94) {
                                            if (Cs < 1731) {
                                                return 4;
                                            }
                                            else if (Cs >= 1731) {
                                                if (mu_mv_mag_sq < 115) {
                                                    return 4;
                                                }
                                                else if (mu_mv_mag_sq >= 115) {
                                                    return 8;
                                                }
                                            }
                                        }
                                    }
                                    else if (SC >= 9308) {
                                        if (Cs < 5031) {
                                            if (Rs < 4664) {
                                                return 4;
                                            }
                                            else if (Rs >= 4664) {
                                                if (ssDCval < 80) {
                                                    return 8;
                                                }
                                                else if (ssDCval >= 80) {
                                                    if (mu_mv_mag_sq < 128) {
                                                        if (Rs < 4888) {
                                                            return 8;
                                                        }
                                                        else if (Rs >= 4888) {
                                                            if (Cs < 5008) {
                                                                return 4;
                                                            }
                                                            else if (Cs >= 5008) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                    else if (mu_mv_mag_sq >= 128) {
                                                        if (diffRsCsdiff < 16) {
                                                            if (SC < 9676) {
                                                                return 8;
                                                            }
                                                            else if (SC >= 9676) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (diffRsCsdiff >= 16) {
                                                            if (RsDiff < 9) {
                                                                return 8;
                                                            }
                                                            else if (RsDiff >= 9) {
                                                                if (Rs < 4835) {
                                                                    return 8;
                                                                }
                                                                else if (Rs >= 4835) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (Cs >= 5031) {
                                            if (diffRsCsdiff < 39) {
                                                if (CsDiff < 9) {
                                                    return 4;
                                                }
                                                else if (CsDiff >= 9) {
                                                    if (Cs < 5139) {
                                                        return 8;
                                                    }
                                                    else if (Cs >= 5139) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                            else if (diffRsCsdiff >= 39) {
                                                if (SC < 10032) {
                                                    return 4;
                                                }
                                                else if (SC >= 10032) {
                                                    return 8;
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (tcor >= 95) {
                                    if (RsDiff < 3) {
                                        return 8;
                                    }
                                    else if (RsDiff >= 3) {
                                        if (SC < 4807) {
                                            if (posBalance < 58) {
                                                return 4;
                                            }
                                            else if (posBalance >= 58) {
                                                if (Cs < 1757) {
                                                    return 4;
                                                }
                                                else if (Cs >= 1757) {
                                                    return 8;
                                                }
                                            }
                                        }
                                        else if (SC >= 4807) {
                                            if (diffRsCsdiff < -61) {
                                                return 4;
                                            }
                                            else if (diffRsCsdiff >= -61) {
                                                return 8;
                                            }
                                        }
                                    }
                                }
                            }
                        }
                        else if (refDCval >= 84) {
                            if (Rs < 5017) {
                                if (mu_mv_mag_sq < 123) {
                                    if (posBalance < 59) {
                                        if (Cs < 3164) {
                                            if (negBalance < 38) {
                                                if (Rs < 2537) {
                                                    return 4;
                                                }
                                                else if (Rs >= 2537) {
                                                    return 8;
                                                }
                                            }
                                            else if (negBalance >= 38) {
                                                if (refDCval < 112) {
                                                    if (CsDiff < 6) {
                                                        if (SC < 1805) {
                                                            return 4;
                                                        }
                                                        else if (SC >= 1805) {
                                                            if (Cs < 3035) {
                                                                if (Rs < 1721) {
                                                                    if (SC < 3548) {
                                                                        return 8;
                                                                    }
                                                                    else if (SC >= 3548) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (Rs >= 1721) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (Cs >= 3035) {
                                                                if (Rs < 2856) {
                                                                    return 4;
                                                                }
                                                                else if (Rs >= 2856) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (CsDiff >= 6) {
                                                        if (SC < 5051) {
                                                            return 4;
                                                        }
                                                        else if (SC >= 5051) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                                else if (refDCval >= 112) {
                                                    if (SC < 5612) {
                                                        if (posBalance < 57) {
                                                            if (Rs < 2685) {
                                                                return 4;
                                                            }
                                                            else if (Rs >= 2685) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (posBalance >= 57) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (SC >= 5612) {
                                                        if (ssDCval < 114) {
                                                            if (Rs < 3023) {
                                                                if (tcor < 97) {
                                                                    return 8;
                                                                }
                                                                else if (tcor >= 97) {
                                                                    if (Cs < 2867) {
                                                                        return 8;
                                                                    }
                                                                    else if (Cs >= 2867) {
                                                                        if (posBalance < 57) {
                                                                            return 4;
                                                                        }
                                                                        else if (posBalance >= 57) {
                                                                            if (Rs < 2702) {
                                                                                return 8;
                                                                            }
                                                                            else if (Rs >= 2702) {
                                                                                if (Cs < 3004) {
                                                                                    if (SC < 5762) {
                                                                                        return 4;
                                                                                    }
                                                                                    else if (SC >= 5762) {
                                                                                        return 8;
                                                                                    }
                                                                                }
                                                                                else if (Cs >= 3004) {
                                                                                    return 4;
                                                                                }
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                            else if (Rs >= 3023) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (ssDCval >= 114) {
                                                            if (SC < 5708) {
                                                                if (Rs < 2705) {
                                                                    return 8;
                                                                }
                                                                else if (Rs >= 2705) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (SC >= 5708) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (Cs >= 3164) {
                                            return 8;
                                        }
                                    }
                                    else if (posBalance >= 59) {
                                        return 8;
                                    }
                                }
                                else if (mu_mv_mag_sq >= 123) {
                                    if (posBalance < 54) {
                                        if (RsCsDiff < 121) {
                                            if (Rs < 1796) {
                                                if (Rs < 1135) {
                                                    return 8;
                                                }
                                                else if (Rs >= 1135) {
                                                    return 4;
                                                }
                                            }
                                            else if (Rs >= 1796) {
                                                return 8;
                                            }
                                        }
                                        else if (RsCsDiff >= 121) {
                                            return 4;
                                        }
                                    }
                                    else if (posBalance >= 54) {
                                        if (negBalance < 48) {
                                            if (tcor < 96) {
                                                if (ssDCval < 86) {
                                                    return 4;
                                                }
                                                else if (ssDCval >= 86) {
                                                    return 8;
                                                }
                                            }
                                            else if (tcor >= 96) {
                                                return 8;
                                            }
                                        }
                                        else if (negBalance >= 48) {
                                            if (SC < 5750) {
                                                return 8;
                                            }
                                            else if (SC >= 5750) {
                                                if (SC < 5769) {
                                                    return 4;
                                                }
                                                else if (SC >= 5769) {
                                                    if (diffRsCsdiff < -5) {
                                                        if (Rs < 2950) {
                                                            if (Rs < 2802) {
                                                                return 8;
                                                            }
                                                            else if (Rs >= 2802) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (Rs >= 2950) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (diffRsCsdiff >= -5) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                            else if (Rs >= 5017) {
                                return 8;
                            }
                        }
                    }
                }
            }
            else if (AFD >= 7) {
                if (Cs < 4146) {
                    if (Rs < 5440) {
                        if (SC < 3720) {
                            if (mu_mv_mag_sq < 82) {
                                if (negBalance < 47) {
                                    if (Cs < 733) {
                                        if (tcor < 87) {
                                            if (negBalance < 19) {
                                                return 8;
                                            }
                                            else if (negBalance >= 19) {
                                                if (Rs < 768) {
                                                    return 4;
                                                }
                                                else if (Rs >= 768) {
                                                    return 1;
                                                }
                                            }
                                        }
                                        else if (tcor >= 87) {
                                            if (ssDCval < 147) {
                                                if (diffTSC < -1) {
                                                    return 4;
                                                }
                                                else if (diffTSC >= -1) {
                                                    if (mu_mv_mag_sq < 41) {
                                                        return 2;
                                                    }
                                                    else if (mu_mv_mag_sq >= 41) {
                                                        if (SC < 1349) {
                                                            return 2;
                                                        }
                                                        else if (SC >= 1349) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (ssDCval >= 147) {
                                                return 4;
                                            }
                                        }
                                    }
                                    else if (Cs >= 733) {
                                        if (negBalance < 37) {
                                            if (diffMVdiffVal < -9) {
                                                if (Cs < 844) {
                                                    return 4;
                                                }
                                                else if (Cs >= 844) {
                                                    return 8;
                                                }
                                            }
                                            else if (diffMVdiffVal >= -9) {
                                                return 4;
                                            }
                                        }
                                        else if (negBalance >= 37) {
                                            if (Rs < 1607) {
                                                if (TSC < 6) {
                                                    if (Rs < 1414) {
                                                        if (CsDiff < 4) {
                                                            return 1;
                                                        }
                                                        else if (CsDiff >= 4) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (Rs >= 1414) {
                                                        return 8;
                                                    }
                                                }
                                                else if (TSC >= 6) {
                                                    return 8;
                                                }
                                            }
                                            else if (Rs >= 1607) {
                                                if (negBalance < 43) {
                                                    return 1;
                                                }
                                                else if (negBalance >= 43) {
                                                    if (diffAFD < -1) {
                                                        return 1;
                                                    }
                                                    else if (diffAFD >= -1) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (negBalance >= 47) {
                                    if (refDCval < 88) {
                                        if (tcor < 95) {
                                            if (diffMVdiffVal < -9) {
                                                return 4;
                                            }
                                            else if (diffMVdiffVal >= -9) {
                                                return 2;
                                            }
                                        }
                                        else if (tcor >= 95) {
                                            if (Rs < 1986) {
                                                return 4;
                                            }
                                            else if (Rs >= 1986) {
                                                return 1;
                                            }
                                        }
                                    }
                                    else if (refDCval >= 88) {
                                        if (mu_mv_mag_sq < 2) {
                                            return 4;
                                        }
                                        else if (mu_mv_mag_sq >= 2) {
                                            if (Rs < 1329) {
                                                if (CsDiff < 2) {
                                                    return 8;
                                                }
                                                else if (CsDiff >= 2) {
                                                    return 4;
                                                }
                                            }
                                            else if (Rs >= 1329) {
                                                if (tcor < 78) {
                                                    return 1;
                                                }
                                                else if (tcor >= 78) {
                                                    if (posBalance < 61) {
                                                        if (diffTSC < 1) {
                                                            if (Rs < 1786) {
                                                                return 4;
                                                            }
                                                            else if (Rs >= 1786) {
                                                                if (Cs < 1142) {
                                                                    return 8;
                                                                }
                                                                else if (Cs >= 1142) {
                                                                    return 2;
                                                                }
                                                            }
                                                        }
                                                        else if (diffTSC >= 1) {
                                                            return 2;
                                                        }
                                                    }
                                                    else if (posBalance >= 61) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                            else if (mu_mv_mag_sq >= 82) {
                                if (mcTcor < 88) {
                                    if (Cs < 1289) {
                                        return 4;
                                    }
                                    else if (Cs >= 1289) {
                                        return 8;
                                    }
                                }
                                else if (mcTcor >= 88) {
                                    if (ssDCval < 112) {
                                        if (posBalance < 49) {
                                            return 4;
                                        }
                                        else if (posBalance >= 49) {
                                            if (negBalance < 49) {
                                                if (mu_mv_mag_sq < 113) {
                                                    if (Cs < 1817) {
                                                        return 2;
                                                    }
                                                    else if (Cs >= 1817) {
                                                        if (AFD < 8) {
                                                            return 4;
                                                        }
                                                        else if (AFD >= 8) {
                                                            return 2;
                                                        }
                                                    }
                                                }
                                                else if (mu_mv_mag_sq >= 113) {
                                                    if (ssDCval < 75) {
                                                        if (posBalance < 59) {
                                                            if (RsDiff < 4) {
                                                                return 2;
                                                            }
                                                            else if (RsDiff >= 4) {
                                                                if (Rs < 1840) {
                                                                    return 8;
                                                                }
                                                                else if (Rs >= 1840) {
                                                                    return 2;
                                                                }
                                                            }
                                                        }
                                                        else if (posBalance >= 59) {
                                                            return 2;
                                                        }
                                                    }
                                                    else if (ssDCval >= 75) {
                                                        if (Rs < 1827) {
                                                            return 8;
                                                        }
                                                        else if (Rs >= 1827) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (negBalance >= 49) {
                                                return 2;
                                            }
                                        }
                                    }
                                    else if (ssDCval >= 112) {
                                        if (negBalance < 42) {
                                            if (RsCsDiff < 14) {
                                                return 4;
                                            }
                                            else if (RsCsDiff >= 14) {
                                                return 8;
                                            }
                                        }
                                        else if (negBalance >= 42) {
                                            return 4;
                                        }
                                    }
                                }
                            }
                        }
                        else if (SC >= 3720) {
                            if (ssDCval < 99) {
                                if (mcTcor < 91) {
                                    if (ssDCval < 93) {
                                        if (refDCval < 85) {
                                            if (mu_mv_mag_sq < 87) {
                                                if (negBalance < 33) {
                                                    return 1;
                                                }
                                                else if (negBalance >= 33) {
                                                    if (MVDiff < 4) {
                                                        return 2;
                                                    }
                                                    else if (MVDiff >= 4) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                            else if (mu_mv_mag_sq >= 87) {
                                                if (tcor < 88) {
                                                    return 8;
                                                }
                                                else if (tcor >= 88) {
                                                    return 4;
                                                }
                                            }
                                        }
                                        else if (refDCval >= 85) {
                                            if (tcor < 77) {
                                                return 4;
                                            }
                                            else if (tcor >= 77) {
                                                return 8;
                                            }
                                        }
                                    }
                                    else if (ssDCval >= 93) {
                                        if (Rs < 3051) {
                                            if (CsDiff < 5) {
                                                return 8;
                                            }
                                            else if (CsDiff >= 5) {
                                                return 2;
                                            }
                                        }
                                        else if (Rs >= 3051) {
                                            return 8;
                                        }
                                    }
                                }
                                else if (mcTcor >= 91) {
                                    if (SC < 6188) {
                                        if (refDCval < 89) {
                                            if (tcor < 93) {
                                                if (negBalance < 24) {
                                                    return 1;
                                                }
                                                else if (negBalance >= 24) {
                                                    if (mu_mv_mag_sq < 76) {
                                                        if (Rs < 1868) {
                                                            return 2;
                                                        }
                                                        else if (Rs >= 1868) {
                                                            if (SC < 5454) {
                                                                return 4;
                                                            }
                                                            else if (SC >= 5454) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                    else if (mu_mv_mag_sq >= 76) {
                                                        if (Cs < 1885) {
                                                            if (CsDiff < 6) {
                                                                return 2;
                                                            }
                                                            else if (CsDiff >= 6) {
                                                                return 1;
                                                            }
                                                        }
                                                        else if (Cs >= 1885) {
                                                            if (Cs < 2497) {
                                                                if (refDCval < 78) {
                                                                    if (Rs < 2615) {
                                                                        return 4;
                                                                    }
                                                                    else if (Rs >= 2615) {
                                                                        return 8;
                                                                    }
                                                                }
                                                                else if (refDCval >= 78) {
                                                                    if (Cs < 2394) {
                                                                        return 8;
                                                                    }
                                                                    else if (Cs >= 2394) {
                                                                        if (SC < 5653) {
                                                                            return 4;
                                                                        }
                                                                        else if (SC >= 5653) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                            else if (Cs >= 2497) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                            else if (tcor >= 93) {
                                                if (Cs < 2112) {
                                                    if (RsCsDiff < 67) {
                                                        if (diffAFD < -1) {
                                                            if (SC < 3753) {
                                                                return 4;
                                                            }
                                                            else if (SC >= 3753) {
                                                                if (refDCval < 81) {
                                                                    if (AFD < 9) {
                                                                        return 4;
                                                                    }
                                                                    else if (AFD >= 9) {
                                                                        return 1;
                                                                    }
                                                                }
                                                                else if (refDCval >= 81) {
                                                                    return 2;
                                                                }
                                                            }
                                                        }
                                                        else if (diffAFD >= -1) {
                                                            if (Rs < 1959) {
                                                                if (Rs < 1829) {
                                                                    return 8;
                                                                }
                                                                else if (Rs >= 1829) {
                                                                    if (AFD < 9) {
                                                                        return 4;
                                                                    }
                                                                    else if (AFD >= 9) {
                                                                        return 2;
                                                                    }
                                                                }
                                                            }
                                                            else if (Rs >= 1959) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                    else if (RsCsDiff >= 67) {
                                                        return 8;
                                                    }
                                                }
                                                else if (Cs >= 2112) {
                                                    if (diffRsCsdiff < -34) {
                                                        return 8;
                                                    }
                                                    else if (diffRsCsdiff >= -34) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                        }
                                        else if (refDCval >= 89) {
                                            if (negBalance < 46) {
                                                if (SC < 4343) {
                                                    if (SC < 4141) {
                                                        return 4;
                                                    }
                                                    else if (SC >= 4141) {
                                                        if (Cs < 1937) {
                                                            return 2;
                                                        }
                                                        else if (Cs >= 1937) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                                else if (SC >= 4343) {
                                                    if (MVDiff < 2) {
                                                        return 2;
                                                    }
                                                    else if (MVDiff >= 2) {
                                                        if (diffAFD < -2) {
                                                            return 4;
                                                        }
                                                        else if (diffAFD >= -2) {
                                                            return 1;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (negBalance >= 46) {
                                                if (SC < 5419) {
                                                    if (Cs < 1870) {
                                                        if (Rs < 2422) {
                                                            return 2;
                                                        }
                                                        else if (Rs >= 2422) {
                                                            return 1;
                                                        }
                                                    }
                                                    else if (Cs >= 1870) {
                                                        if (TSC < 11) {
                                                            if (tcor < 86) {
                                                                return 4;
                                                            }
                                                            else if (tcor >= 86) {
                                                                if (diffTSC < -2) {
                                                                    if (Cs < 2328) {
                                                                        return 4;
                                                                    }
                                                                    else if (Cs >= 2328) {
                                                                        return 2;
                                                                    }
                                                                }
                                                                else if (diffTSC >= -2) {
                                                                    if (ssDCval < 97) {
                                                                        if (posBalance < 55) {
                                                                            if (Rs < 2233) {
                                                                                return 2;
                                                                            }
                                                                            else if (Rs >= 2233) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                        else if (posBalance >= 55) {
                                                                            return 2;
                                                                        }
                                                                    }
                                                                    else if (ssDCval >= 97) {
                                                                        return 4;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (TSC >= 11) {
                                                            return 2;
                                                        }
                                                    }
                                                }
                                                else if (SC >= 5419) {
                                                    return 8;
                                                }
                                            }
                                        }
                                    }
                                    else if (SC >= 6188) {
                                        if (MVDiff < 1) {
                                            if (SC < 7617) {
                                                return 2;
                                            }
                                            else if (SC >= 7617) {
                                                return 8;
                                            }
                                        }
                                        else if (MVDiff >= 1) {
                                            if (TSCindex < 8) {
                                                return 8;
                                            }
                                            else if (TSCindex >= 8) {
                                                if (Cs < 3201) {
                                                    if (SC < 6508) {
                                                        return 8;
                                                    }
                                                    else if (SC >= 6508) {
                                                        return 4;
                                                    }
                                                }
                                                else if (Cs >= 3201) {
                                                    if (Cs < 4032) {
                                                        return 2;
                                                    }
                                                    else if (Cs >= 4032) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                            else if (ssDCval >= 99) {
                                if (tcor < 77) {
                                    if (refDCval < 123) {
                                        return 4;
                                    }
                                    else if (refDCval >= 123) {
                                        if (posBalance < 49) {
                                            return 8;
                                        }
                                        else if (posBalance >= 49) {
                                            return 1;
                                        }
                                    }
                                }
                                else if (tcor >= 77) {
                                    if (Cs < 2250) {
                                        if (mu_mv_mag_sq < 57) {
                                            if (posBalance < 58) {
                                                if (Rs < 2636) {
                                                    return 1;
                                                }
                                                else if (Rs >= 2636) {
                                                    if (Cs < 1360) {
                                                        return 2;
                                                    }
                                                    else if (Cs >= 1360) {
                                                        if (Rs < 3562) {
                                                            return 2;
                                                        }
                                                        else if (Rs >= 3562) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (posBalance >= 58) {
                                                if (refDCval < 123) {
                                                    return 4;
                                                }
                                                else if (refDCval >= 123) {
                                                    if (SC < 4479) {
                                                        if (Cs < 1157) {
                                                            return 2;
                                                        }
                                                        else if (Cs >= 1157) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (SC >= 4479) {
                                                        if (Cs < 2149) {
                                                            if (negBalance < 52) {
                                                                return 8;
                                                            }
                                                            else if (negBalance >= 52) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (Cs >= 2149) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (mu_mv_mag_sq >= 57) {
                                            if (ssDCval < 109) {
                                                if (Rs < 2542) {
                                                    return 4;
                                                }
                                                else if (Rs >= 2542) {
                                                    if (Cs < 2204) {
                                                        return 1;
                                                    }
                                                    else if (Cs >= 2204) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                            else if (ssDCval >= 109) {
                                                if (mu_mv_mag_sq < 71) {
                                                    if (SC < 4797) {
                                                        if (mu_mv_mag_sq < 59) {
                                                            return 8;
                                                        }
                                                        else if (mu_mv_mag_sq >= 59) {
                                                            if (SC < 4285) {
                                                                return 4;
                                                            }
                                                            else if (SC >= 4285) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                    else if (SC >= 4797) {
                                                        return 4;
                                                    }
                                                }
                                                else if (mu_mv_mag_sq >= 71) {
                                                    if (mcTcor < 93) {
                                                        if (RsDiff < 10) {
                                                            return 8;
                                                        }
                                                        else if (RsDiff >= 10) {
                                                            return 1;
                                                        }
                                                    }
                                                    else if (mcTcor >= 93) {
                                                        if (Rs < 2528) {
                                                            if (RsCsDiff < 77) {
                                                                return 4;
                                                            }
                                                            else if (RsCsDiff >= 77) {
                                                                if (SC < 3792) {
                                                                    return 4;
                                                                }
                                                                else if (SC >= 3792) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                        else if (Rs >= 2528) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (Cs >= 2250) {
                                        if (negBalance < 54) {
                                            if (TSC < 13) {
                                                if (Rs < 3497) {
                                                    if (posBalance < 57) {
                                                        if (ssDCval < 100) {
                                                            return 1;
                                                        }
                                                        else if (ssDCval >= 100) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (posBalance >= 57) {
                                                        if (negBalance < 48) {
                                                            if (refDCval < 101) {
                                                                return 1;
                                                            }
                                                            else if (refDCval >= 101) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (negBalance >= 48) {
                                                            if (TSCindex < 8) {
                                                                if (Rs < 2834) {
                                                                    return 4;
                                                                }
                                                                else if (Rs >= 2834) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (TSCindex >= 8) {
                                                                if (negBalance < 53) {
                                                                    if (diffRsCsdiff < -50) {
                                                                        if (Cs < 2378) {
                                                                            return 8;
                                                                        }
                                                                        else if (Cs >= 2378) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                    else if (diffRsCsdiff >= -50) {
                                                                        if (RsCsDiff < 138) {
                                                                            return 8;
                                                                        }
                                                                        else if (RsCsDiff >= 138) {
                                                                            if (Cs < 2449) {
                                                                                if (SC < 5370) {
                                                                                    return 8;
                                                                                }
                                                                                else if (SC >= 5370) {
                                                                                    return 4;
                                                                                }
                                                                            }
                                                                            else if (Cs >= 2449) {
                                                                                return 8;
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                                else if (negBalance >= 53) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (Rs >= 3497) {
                                                    if (ssDCval < 125) {
                                                        if (RsDiff < 9) {
                                                            if (Cs < 2948) {
                                                                if (RsDiff < 6) {
                                                                    return 8;
                                                                }
                                                                else if (RsDiff >= 6) {
                                                                    if (ssDCval < 115) {
                                                                        if (SC < 6909) {
                                                                            return 1;
                                                                        }
                                                                        else if (SC >= 6909) {
                                                                            return 2;
                                                                        }
                                                                    }
                                                                    else if (ssDCval >= 115) {
                                                                        return 4;
                                                                    }
                                                                }
                                                            }
                                                            else if (Cs >= 2948) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (RsDiff >= 9) {
                                                            if (Cs < 3692) {
                                                                if (SC < 8088) {
                                                                    if (Cs < 3250) {
                                                                        return 4;
                                                                    }
                                                                    else if (Cs >= 3250) {
                                                                        return 8;
                                                                    }
                                                                }
                                                                else if (SC >= 8088) {
                                                                    if (refDCval < 115) {
                                                                        if (Cs < 3610) {
                                                                            if (ssDCval < 115) {
                                                                                return 2;
                                                                            }
                                                                            else if (ssDCval >= 115) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                        else if (Cs >= 3610) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                    else if (refDCval >= 115) {
                                                                        return 4;
                                                                    }
                                                                }
                                                            }
                                                            else if (Cs >= 3692) {
                                                                if (TSC < 10) {
                                                                    return 2;
                                                                }
                                                                else if (TSC >= 10) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (ssDCval >= 125) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                            else if (TSC >= 13) {
                                                if (SC < 6329) {
                                                    return 1;
                                                }
                                                else if (SC >= 6329) {
                                                    return 8;
                                                }
                                            }
                                        }
                                        else if (negBalance >= 54) {
                                            if (diffMVdiffVal < -1) {
                                                if (MVDiff < 1) {
                                                    return 4;
                                                }
                                                else if (MVDiff >= 1) {
                                                    if (Rs < 2724) {
                                                        return 2;
                                                    }
                                                    else if (Rs >= 2724) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                            else if (diffMVdiffVal >= -1) {
                                                if (Cs < 3729) {
                                                    return 4;
                                                }
                                                else if (Cs >= 3729) {
                                                    return 8;
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                        }
                    }
                    else if (Rs >= 5440) {
                        if (posBalance < 57) {
                            if (Cs < 3520) {
                                return 1;
                            }
                            else if (Cs >= 3520) {
                                if (SC < 11328) {
                                    return 8;
                                }
                                else if (SC >= 11328) {
                                    return 4;
                                }
                            }
                        }
                        else if (posBalance >= 57) {
                            return 4;
                        }
                    }
                }
                else if (Cs >= 4146) {
                    if (negBalance < 63) {
                        if (refDCval < 106) {
                            if (refDCval < 82) {
                                if (RsCsDiff < 173) {
                                    if (SC < 9941) {
                                        if (Rs < 4837) {
                                            if (diffRsCsdiff < -16) {
                                                return 4;
                                            }
                                            else if (diffRsCsdiff >= -16) {
                                                return 8;
                                            }
                                        }
                                        else if (Rs >= 4837) {
                                            return 4;
                                        }
                                    }
                                    else if (SC >= 9941) {
                                        return 4;
                                    }
                                }
                                else if (RsCsDiff >= 173) {
                                    if (mcTcor < 95) {
                                        return 8;
                                    }
                                    else if (mcTcor >= 95) {
                                        return 1;
                                    }
                                }
                            }
                            else if (refDCval >= 82) {
                                return 8;
                            }
                        }
                        else if (refDCval >= 106) {
                            if (MVDiff < 1) {
                                if (refDCval < 109) {
                                    if (Rs < 6270) {
                                        return 4;
                                    }
                                    else if (Rs >= 6270) {
                                        return 8;
                                    }
                                }
                                else if (refDCval >= 109) {
                                    if (Rs < 7869) {
                                        return 8;
                                    }
                                    else if (Rs >= 7869) {
                                        return 4;
                                    }
                                }
                            }
                            else if (MVDiff >= 1) {
                                if (TSC < 28) {
                                    if (Rs < 4842) {
                                        if (AFD < 17) {
                                            return 8;
                                        }
                                        else if (AFD >= 17) {
                                            if (SC < 9201) {
                                                return 4;
                                            }
                                            else if (SC >= 9201) {
                                                return 1;
                                            }
                                        }
                                    }
                                    else if (Rs >= 4842) {
                                        if (Cs < 12154) {
                                            return 8;
                                        }
                                        else if (Cs >= 12154) {
                                            return 4;
                                        }
                                    }
                                }
                                else if (TSC >= 28) {
                                    if (AFD < 31) {
                                        return 8;
                                    }
                                    else if (AFD >= 31) {
                                        return 4;
                                    }
                                }
                            }
                        }
                    }
                    else if (negBalance >= 63) {
                        if (SC < 14578) {
                            if (mcTcor < 77) {
                                if (SC < 13988) {
                                    return 8;
                                }
                                else if (SC >= 13988) {
                                    return 4;
                                }
                            }
                            else if (mcTcor >= 77) {
                                return 8;
                            }
                        }
                        else if (SC >= 14578) {
                            if (tcor < 77) {
                                if (Rs < 7707) {
                                    if (Rs < 7538) {
                                        if (Rs < 7412) {
                                            return 4;
                                        }
                                        else if (Rs >= 7412) {
                                            return 2;
                                        }
                                    }
                                    else if (Rs >= 7538) {
                                        return 4;
                                    }
                                }
                                else if (Rs >= 7707) {
                                    if (Rs < 8095) {
                                        return 2;
                                    }
                                    else if (Rs >= 8095) {
                                        if (diffRsCsdiff < 69) {
                                            return 4;
                                        }
                                        else if (diffRsCsdiff >= 69) {
                                            return 2;
                                        }
                                    }
                                }
                            }
                            else if (tcor >= 77) {
                                return 2;
                            }
                        }
                    }
                }
            }
        }
        else if (MVDiff >= 6) {
            if (Cs < 3274) {
                if (negBalance < 53) {
                    if (CsDiff < 10) {
                        if (Rs < 3316) {
                            if (mu_mv_mag_sq < 74) {
                                if (TSC < 7) {
                                    if (Rs < 671) {
                                        if (MVDiff < 15) {
                                            if (SC < 768) {
                                                if (TSC < 4) {
                                                    if (CsDiff < 1) {
                                                        if (ssDCval < 35) {
                                                            return 4;
                                                        }
                                                        else if (ssDCval >= 35) {
                                                            if (Cs < 118) {
                                                                return 8;
                                                            }
                                                            else if (Cs >= 118) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                    else if (CsDiff >= 1) {
                                                        if (Cs < 279) {
                                                            if (Cs < 220) {
                                                                return 1;
                                                            }
                                                            else if (Cs >= 220) {
                                                                return 2;
                                                            }
                                                        }
                                                        else if (Cs >= 279) {
                                                            if (AFD < 6) {
                                                                if (MVDiff < 10) {
                                                                    return 2;
                                                                }
                                                                else if (MVDiff >= 10) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (AFD >= 6) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (TSC >= 4) {
                                                    if (diffAFD < -1) {
                                                        if (Rs < 363) {
                                                            return 4;
                                                        }
                                                        else if (Rs >= 363) {
                                                            return 2;
                                                        }
                                                    }
                                                    else if (diffAFD >= -1) {
                                                        return 2;
                                                    }
                                                }
                                            }
                                            else if (SC >= 768) {
                                                if (Rs < 498) {
                                                    return 4;
                                                }
                                                else if (Rs >= 498) {
                                                    if (negBalance < 52) {
                                                        if (diffMVdiffVal < -3) {
                                                            if (AFD < 7) {
                                                                if (Rs < 603) {
                                                                    return 4;
                                                                }
                                                                else if (Rs >= 603) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (AFD >= 7) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (diffMVdiffVal >= -3) {
                                                            if (Cs < 357) {
                                                                return 2;
                                                            }
                                                            else if (Cs >= 357) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                    else if (negBalance >= 52) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                        }
                                        else if (MVDiff >= 15) {
                                            if (Cs < 373) {
                                                if (refDCval < 81) {
                                                    if (tcor < 61) {
                                                        if (refDCval < 22) {
                                                            if (negBalance < 1) {
                                                                if (MVDiff < 269) {
                                                                    if (mu_mv_mag_sq < 9) {
                                                                        if (MVDiff < 140) {
                                                                            return 8;
                                                                        }
                                                                        else if (MVDiff >= 140) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                    else if (mu_mv_mag_sq >= 9) {
                                                                        if (MVDiff < 200) {
                                                                            if (refDCval < 20) {
                                                                                if (MVDiff < 76) {
                                                                                    return 4;
                                                                                }
                                                                                else if (MVDiff >= 76) {
                                                                                    if (tcor < 38) {
                                                                                        return 2;
                                                                                    }
                                                                                    else if (tcor >= 38) {
                                                                                        if (MVDiff < 158) {
                                                                                            return 4;
                                                                                        }
                                                                                        else if (MVDiff >= 158) {
                                                                                            return 2;
                                                                                        }
                                                                                    }
                                                                                }
                                                                            }
                                                                            else if (refDCval >= 20) {
                                                                                if (AFD < 2) {
                                                                                    return 4;
                                                                                }
                                                                                else if (AFD >= 2) {
                                                                                    return 8;
                                                                                }
                                                                            }
                                                                        }
                                                                        else if (MVDiff >= 200) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                }
                                                                else if (MVDiff >= 269) {
                                                                    if (tcor < 44) {
                                                                        return 2;
                                                                    }
                                                                    else if (tcor >= 44) {
                                                                        if (AFD < 1) {
                                                                            return 2;
                                                                        }
                                                                        else if (AFD >= 1) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                            else if (negBalance >= 1) {
                                                                if (MVDiff < 61) {
                                                                    if (posBalance < 27) {
                                                                        return 1;
                                                                    }
                                                                    else if (posBalance >= 27) {
                                                                        return 8;
                                                                    }
                                                                }
                                                                else if (MVDiff >= 61) {
                                                                    if (tcor < 34) {
                                                                        if (diffAFD < -5) {
                                                                            return 1;
                                                                        }
                                                                        else if (diffAFD >= -5) {
                                                                            return 2;
                                                                        }
                                                                    }
                                                                    else if (tcor >= 34) {
                                                                        if (tcor < 42) {
                                                                            if (posBalance < 28) {
                                                                                if (posBalance < 27) {
                                                                                    if (mu_mv_mag_sq < 25) {
                                                                                        return 4;
                                                                                    }
                                                                                    else if (mu_mv_mag_sq >= 25) {
                                                                                        return 8;
                                                                                    }
                                                                                }
                                                                                else if (posBalance >= 27) {
                                                                                    return 8;
                                                                                }
                                                                            }
                                                                            else if (posBalance >= 28) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                        else if (tcor >= 42) {
                                                                            if (ssDCval < 19) {
                                                                                return 8;
                                                                            }
                                                                            else if (ssDCval >= 19) {
                                                                                if (AFD < 1) {
                                                                                    return 4;
                                                                                }
                                                                                else if (AFD >= 1) {
                                                                                    return 2;
                                                                                }
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (refDCval >= 22) {
                                                            if (tcor < 51) {
                                                                return 1;
                                                            }
                                                            else if (tcor >= 51) {
                                                                if (AFD < 3) {
                                                                    return 2;
                                                                }
                                                                else if (AFD >= 3) {
                                                                    if (AFD < 9) {
                                                                        return 1;
                                                                    }
                                                                    else if (AFD >= 9) {
                                                                        return 2;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (tcor >= 61) {
                                                        if (mu_mv_mag_sq < 35) {
                                                            if (Scindex < 2) {
                                                                if (MVDiff < 83) {
                                                                    if (ssDCval < 29) {
                                                                        if (MVDiff < 28) {
                                                                            return 4;
                                                                        }
                                                                        else if (MVDiff >= 28) {
                                                                            if (negBalance < 19) {
                                                                                if (TSC < 1) {
                                                                                    if (Rs < 15) {
                                                                                        return 4;
                                                                                    }
                                                                                    else if (Rs >= 15) {
                                                                                        return 2;
                                                                                    }
                                                                                }
                                                                                else if (TSC >= 1) {
                                                                                    return 2;
                                                                                }
                                                                            }
                                                                            else if (negBalance >= 19) {
                                                                                if (negBalance < 29) {
                                                                                    return 8;
                                                                                }
                                                                                else if (negBalance >= 29) {
                                                                                    return 2;
                                                                                }
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (ssDCval >= 29) {
                                                                        if (MVDiff < 33) {
                                                                            if (refDCval < 39) {
                                                                                return 2;
                                                                            }
                                                                            else if (refDCval >= 39) {
                                                                                return 8;
                                                                            }
                                                                        }
                                                                        else if (MVDiff >= 33) {
                                                                            if (refDCval < 40) {
                                                                                return 1;
                                                                            }
                                                                            else if (refDCval >= 40) {
                                                                                if (Rs < 46) {
                                                                                    return 2;
                                                                                }
                                                                                else if (Rs >= 46) {
                                                                                    return 4;
                                                                                }
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                                else if (MVDiff >= 83) {
                                                                    if (negBalance < 10) {
                                                                        if (diffMVdiffVal < -70) {
                                                                            if (diffMVdiffVal < -454) {
                                                                                return 2;
                                                                            }
                                                                            else if (diffMVdiffVal >= -454) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                        else if (diffMVdiffVal >= -70) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                    else if (negBalance >= 10) {
                                                                        return 2;
                                                                    }
                                                                }
                                                            }
                                                            else if (Scindex >= 2) {
                                                                if (SC < 689) {
                                                                    if (mcTcor < 79) {
                                                                        return 8;
                                                                    }
                                                                    else if (mcTcor >= 79) {
                                                                        return 2;
                                                                    }
                                                                }
                                                                else if (SC >= 689) {
                                                                    if (Cs < 265) {
                                                                        return 2;
                                                                    }
                                                                    else if (Cs >= 265) {
                                                                        if (posBalance < 49) {
                                                                            return 2;
                                                                        }
                                                                        else if (posBalance >= 49) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (mu_mv_mag_sq >= 35) {
                                                            if (SC < 733) {
                                                                if (diffMVdiffVal < -6) {
                                                                    if (RsDiff < 2) {
                                                                        if (posBalance < 5) {
                                                                            return 1;
                                                                        }
                                                                        else if (posBalance >= 5) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                    else if (RsDiff >= 2) {
                                                                        return 2;
                                                                    }
                                                                }
                                                                else if (diffMVdiffVal >= -6) {
                                                                    if (negBalance < 3) {
                                                                        if (posBalance < 65) {
                                                                            return 2;
                                                                        }
                                                                        else if (posBalance >= 65) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                    else if (negBalance >= 3) {
                                                                        if (posBalance < 44) {
                                                                            if (Rs < 90) {
                                                                                return 8;
                                                                            }
                                                                            else if (Rs >= 90) {
                                                                                return 2;
                                                                            }
                                                                        }
                                                                        else if (posBalance >= 44) {
                                                                            if (ssDCval < 38) {
                                                                                if (tcor < 74) {
                                                                                    return 4;
                                                                                }
                                                                                else if (tcor >= 74) {
                                                                                    if (MVDiff < 20) {
                                                                                        if (Cs < 335) {
                                                                                            return 1;
                                                                                        }
                                                                                        else if (Cs >= 335) {
                                                                                            return 8;
                                                                                        }
                                                                                    }
                                                                                    else if (MVDiff >= 20) {
                                                                                        return 8;
                                                                                    }
                                                                                }
                                                                            }
                                                                            else if (ssDCval >= 38) {
                                                                                if (Cs < 315) {
                                                                                    return 1;
                                                                                }
                                                                                else if (Cs >= 315) {
                                                                                    return 4;
                                                                                }
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                            else if (SC >= 733) {
                                                                if (mcTcor < 84) {
                                                                    if (mu_mv_mag_sq < 60) {
                                                                        return 1;
                                                                    }
                                                                    else if (mu_mv_mag_sq >= 60) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (mcTcor >= 84) {
                                                                    if (mcTcor < 85) {
                                                                        if (Rs < 439) {
                                                                            return 1;
                                                                        }
                                                                        else if (Rs >= 439) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                    else if (mcTcor >= 85) {
                                                                        return 4;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (refDCval >= 81) {
                                                    if (tcor < 77) {
                                                        return 2;
                                                    }
                                                    else if (tcor >= 77) {
                                                        if (Rs < 239) {
                                                            if (mu_mv_mag_sq < 44) {
                                                                return 1;
                                                            }
                                                            else if (mu_mv_mag_sq >= 44) {
                                                                return 2;
                                                            }
                                                        }
                                                        else if (Rs >= 239) {
                                                            return 1;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (Cs >= 373) {
                                                if (Rs < 519) {
                                                    if (refDCval < 88) {
                                                        if (Rs < 364) {
                                                            if (ssDCval < 29) {
                                                                return 8;
                                                            }
                                                            else if (ssDCval >= 29) {
                                                                return 2;
                                                            }
                                                        }
                                                        else if (Rs >= 364) {
                                                            if (ssDCval < 39) {
                                                                if (negBalance < 47) {
                                                                    if (Cs < 379) {
                                                                        return 4;
                                                                    }
                                                                    else if (Cs >= 379) {
                                                                        if (SC < 886) {
                                                                            if (MVDiff < 24) {
                                                                                if (negBalance < 43) {
                                                                                    return 2;
                                                                                }
                                                                                else if (negBalance >= 43) {
                                                                                    return 4;
                                                                                }
                                                                            }
                                                                            else if (MVDiff >= 24) {
                                                                                return 2;
                                                                            }
                                                                        }
                                                                        else if (SC >= 886) {
                                                                            return 2;
                                                                        }
                                                                    }
                                                                }
                                                                else if (negBalance >= 47) {
                                                                    if (Rs < 436) {
                                                                        return 4;
                                                                    }
                                                                    else if (Rs >= 436) {
                                                                        if (Cs < 406) {
                                                                            return 8;
                                                                        }
                                                                        else if (Cs >= 406) {
                                                                            return 2;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                            else if (ssDCval >= 39) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                    else if (refDCval >= 88) {
                                                        if (Rs < 489) {
                                                            return 4;
                                                        }
                                                        else if (Rs >= 489) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                                else if (Rs >= 519) {
                                                    if (negBalance < 51) {
                                                        if (diffTSC < -1) {
                                                            if (SC < 1071) {
                                                                return 4;
                                                            }
                                                            else if (SC >= 1071) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (diffTSC >= -1) {
                                                            if (posBalance < 58) {
                                                                return 2;
                                                            }
                                                            else if (posBalance >= 58) {
                                                                if (Cs < 749) {
                                                                    if (AFD < 13) {
                                                                        if (Cs < 435) {
                                                                            return 8;
                                                                        }
                                                                        else if (Cs >= 435) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                    else if (AFD >= 13) {
                                                                        return 2;
                                                                    }
                                                                }
                                                                else if (Cs >= 749) {
                                                                    if (Rs < 603) {
                                                                        return 1;
                                                                    }
                                                                    else if (Rs >= 603) {
                                                                        return 2;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (negBalance >= 51) {
                                                        if (Cs < 578) {
                                                            return 8;
                                                        }
                                                        else if (Cs >= 578) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (Rs >= 671) {
                                        if (Rs < 2038) {
                                            if (negBalance < 46) {
                                                if (Rs < 1007) {
                                                    if (mu_mv_mag_sq < 36) {
                                                        if (Cs < 623) {
                                                            if (negBalance < 13) {
                                                                if (Rs < 819) {
                                                                    return 4;
                                                                }
                                                                else if (Rs >= 819) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (negBalance >= 13) {
                                                                if (SC < 1495) {
                                                                    return 1;
                                                                }
                                                                else if (SC >= 1495) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                        else if (Cs >= 623) {
                                                            if (negBalance < 35) {
                                                                return 1;
                                                            }
                                                            else if (negBalance >= 35) {
                                                                if (diffAFD < 1) {
                                                                    return 2;
                                                                }
                                                                else if (diffAFD >= 1) {
                                                                    if (SC < 1721) {
                                                                        return 8;
                                                                    }
                                                                    else if (SC >= 1721) {
                                                                        if (SC < 2192) {
                                                                            return 1;
                                                                        }
                                                                        else if (SC >= 2192) {
                                                                            return 2;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (mu_mv_mag_sq >= 36) {
                                                        if (diffAFD < -3) {
                                                            if (Rs < 820) {
                                                                return 1;
                                                            }
                                                            else if (Rs >= 820) {
                                                                return 2;
                                                            }
                                                        }
                                                        else if (diffAFD >= -3) {
                                                            if (RsDiff < 3) {
                                                                if (diffTSC < 2) {
                                                                    if (ssDCval < 32) {
                                                                        return 1;
                                                                    }
                                                                    else if (ssDCval >= 32) {
                                                                        if (posBalance < 53) {
                                                                            if (Rs < 946) {
                                                                                if (Rs < 723) {
                                                                                    return 1;
                                                                                }
                                                                                else if (Rs >= 723) {
                                                                                    return 4;
                                                                                }
                                                                            }
                                                                            else if (Rs >= 946) {
                                                                                if (Cs < 698) {
                                                                                    if (ssDCval < 47) {
                                                                                        return 1;
                                                                                    }
                                                                                    else if (ssDCval >= 47) {
                                                                                        return 4;
                                                                                    }
                                                                                }
                                                                                else if (Cs >= 698) {
                                                                                    return 1;
                                                                                }
                                                                            }
                                                                        }
                                                                        else if (posBalance >= 53) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                }
                                                                else if (diffTSC >= 2) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (RsDiff >= 3) {
                                                                if (MVDiff < 19) {
                                                                    if (SC < 1516) {
                                                                        return 2;
                                                                    }
                                                                    else if (SC >= 1516) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (MVDiff >= 19) {
                                                                    if (Rs < 791) {
                                                                        return 2;
                                                                    }
                                                                    else if (Rs >= 791) {
                                                                        return 8;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (Rs >= 1007) {
                                                    if (RsDiff < 4) {
                                                        if (Rs < 1234) {
                                                            if (mu_mv_mag_sq < 41) {
                                                                if (MVDiff < 11) {
                                                                    if (SC < 1923) {
                                                                        if (Cs < 786) {
                                                                            return 4;
                                                                        }
                                                                        else if (Cs >= 786) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                    else if (SC >= 1923) {
                                                                        return 1;
                                                                    }
                                                                }
                                                                else if (MVDiff >= 11) {
                                                                    if (TSC < 5) {
                                                                        return 1;
                                                                    }
                                                                    else if (TSC >= 5) {
                                                                        return 8;
                                                                    }
                                                                }
                                                            }
                                                            else if (mu_mv_mag_sq >= 41) {
                                                                if (AFD < 7) {
                                                                    if (gchDC < 1) {
                                                                        if (mcTcor < 90) {
                                                                            return 4;
                                                                        }
                                                                        else if (mcTcor >= 90) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                    else if (gchDC >= 1) {
                                                                        if (RsCsDiff < 16) {
                                                                            return 4;
                                                                        }
                                                                        else if (RsCsDiff >= 16) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                }
                                                                else if (AFD >= 7) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                        else if (Rs >= 1234) {
                                                            if (SC < 2907) {
                                                                if (Cs < 602) {
                                                                    return 8;
                                                                }
                                                                else if (Cs >= 602) {
                                                                    if (TSC < 5) {
                                                                        if (ssDCval < 42) {
                                                                            return 2;
                                                                        }
                                                                        else if (ssDCval >= 42) {
                                                                            if (ssDCval < 56) {
                                                                                if (mcTcor < 91) {
                                                                                    if (Cs < 944) {
                                                                                        return 4;
                                                                                    }
                                                                                    else if (Cs >= 944) {
                                                                                        return 1;
                                                                                    }
                                                                                }
                                                                                else if (mcTcor >= 91) {
                                                                                    return 1;
                                                                                }
                                                                            }
                                                                            else if (ssDCval >= 56) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (TSC >= 5) {
                                                                        if (ssDCval < 54) {
                                                                            if (Cs < 954) {
                                                                                if (Cs < 884) {
                                                                                    return 8;
                                                                                }
                                                                                else if (Cs >= 884) {
                                                                                    return 1;
                                                                                }
                                                                            }
                                                                            else if (Cs >= 954) {
                                                                                return 8;
                                                                            }
                                                                        }
                                                                        else if (ssDCval >= 54) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                            else if (SC >= 2907) {
                                                                if (TSC < 5) {
                                                                    if (SC < 3352) {
                                                                        if (Cs < 1503) {
                                                                            return 8;
                                                                        }
                                                                        else if (Cs >= 1503) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                    else if (SC >= 3352) {
                                                                        if (Cs < 1525) {
                                                                            return 4;
                                                                        }
                                                                        else if (Cs >= 1525) {
                                                                            if (diffAFD < 1) {
                                                                                return 1;
                                                                            }
                                                                            else if (diffAFD >= 1) {
                                                                                if (MVDiff < 19) {
                                                                                    return 4;
                                                                                }
                                                                                else if (MVDiff >= 19) {
                                                                                    return 1;
                                                                                }
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                                else if (TSC >= 5) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (RsDiff >= 4) {
                                                        if (Rs < 1686) {
                                                            if (diffRsCsdiff < 10) {
                                                                if (Rs < 1210) {
                                                                    if (MVDiff < 10) {
                                                                        return 4;
                                                                    }
                                                                    else if (MVDiff >= 10) {
                                                                        if (diffTSC < 3) {
                                                                            if (refDCval < 58) {
                                                                                if (RsCsDiff < 45) {
                                                                                    return 4;
                                                                                }
                                                                                else if (RsCsDiff >= 45) {
                                                                                    return 1;
                                                                                }
                                                                            }
                                                                            else if (refDCval >= 58) {
                                                                                return 1;
                                                                            }
                                                                        }
                                                                        else if (diffTSC >= 3) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                }
                                                                else if (Rs >= 1210) {
                                                                    if (MVDiff < 7) {
                                                                        if (ssDCval < 71) {
                                                                            return 4;
                                                                        }
                                                                        else if (ssDCval >= 71) {
                                                                            return 2;
                                                                        }
                                                                    }
                                                                    else if (MVDiff >= 7) {
                                                                        if (gchDC < 1) {
                                                                            if (mu_mv_mag_sq < 42) {
                                                                                if (AFD < 10) {
                                                                                    return 1;
                                                                                }
                                                                                else if (AFD >= 10) {
                                                                                    if (Cs < 1650) {
                                                                                        return 8;
                                                                                    }
                                                                                    else if (Cs >= 1650) {
                                                                                        return 1;
                                                                                    }
                                                                                }
                                                                            }
                                                                            else if (mu_mv_mag_sq >= 42) {
                                                                                if (posBalance < 49) {
                                                                                    return 8;
                                                                                }
                                                                                else if (posBalance >= 49) {
                                                                                    if (tcor < 91) {
                                                                                        return 4;
                                                                                    }
                                                                                    else if (tcor >= 91) {
                                                                                        return 1;
                                                                                    }
                                                                                }
                                                                            }
                                                                        }
                                                                        else if (gchDC >= 1) {
                                                                            if (diffRsCsdiff < -19) {
                                                                                if (Rs < 1312) {
                                                                                    return 4;
                                                                                }
                                                                                else if (Rs >= 1312) {
                                                                                    if (MVDiff < 13) {
                                                                                        return 4;
                                                                                    }
                                                                                    else if (MVDiff >= 13) {
                                                                                        return 8;
                                                                                    }
                                                                                }
                                                                            }
                                                                            else if (diffRsCsdiff >= -19) {
                                                                                if (diffAFD < -2) {
                                                                                    return 1;
                                                                                }
                                                                                else if (diffAFD >= -2) {
                                                                                    if (refDCval < 82) {
                                                                                        if (RsCsDiff < 37) {
                                                                                            if (Rs < 1350) {
                                                                                                if (SC < 2291) {
                                                                                                    return 1;
                                                                                                }
                                                                                                else if (SC >= 2291) {
                                                                                                    return 4;
                                                                                                }
                                                                                            }
                                                                                            else if (Rs >= 1350) {
                                                                                                return 4;
                                                                                            }
                                                                                        }
                                                                                        else if (RsCsDiff >= 37) {
                                                                                            return 1;
                                                                                        }
                                                                                    }
                                                                                    else if (refDCval >= 82) {
                                                                                        if (TSC < 4) {
                                                                                            return 4;
                                                                                        }
                                                                                        else if (TSC >= 4) {
                                                                                            return 2;
                                                                                        }
                                                                                    }
                                                                                }
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                            else if (diffRsCsdiff >= 10) {
                                                                if (diffMVdiffVal < -1) {
                                                                    return 1;
                                                                }
                                                                else if (diffMVdiffVal >= -1) {
                                                                    if (posBalance < 47) {
                                                                        return 4;
                                                                    }
                                                                    else if (posBalance >= 47) {
                                                                        if (posBalance < 65) {
                                                                            if (MVDiff < 9) {
                                                                                if (Rs < 1369) {
                                                                                    return 1;
                                                                                }
                                                                                else if (Rs >= 1369) {
                                                                                    if (Cs < 1069) {
                                                                                        return 8;
                                                                                    }
                                                                                    else if (Cs >= 1069) {
                                                                                        return 4;
                                                                                    }
                                                                                }
                                                                            }
                                                                            else if (MVDiff >= 9) {
                                                                                if (diffRsCsdiff < 67) {
                                                                                    return 1;
                                                                                }
                                                                                else if (diffRsCsdiff >= 67) {
                                                                                    return 8;
                                                                                }
                                                                            }
                                                                        }
                                                                        else if (posBalance >= 65) {
                                                                            if (SC < 2785) {
                                                                                return 8;
                                                                            }
                                                                            else if (SC >= 2785) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (Rs >= 1686) {
                                                            if (diffTSC < -1) {
                                                                if (Rs < 1832) {
                                                                    return 4;
                                                                }
                                                                else if (Rs >= 1832) {
                                                                    return 2;
                                                                }
                                                            }
                                                            else if (diffTSC >= -1) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                            else if (negBalance >= 46) {
                                                if (RsCsDiff < 19) {
                                                    if (TSCindex < 6) {
                                                        if (SC < 3247) {
                                                            if (posBalance < 52) {
                                                                if (ssDCval < 52) {
                                                                    return 1;
                                                                }
                                                                else if (ssDCval >= 52) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (posBalance >= 52) {
                                                                if (Rs < 808) {
                                                                    return 4;
                                                                }
                                                                else if (Rs >= 808) {
                                                                    return 2;
                                                                }
                                                            }
                                                        }
                                                        else if (SC >= 3247) {
                                                            if (Cs < 1545) {
                                                                if (mcTcor < 94) {
                                                                    return 4;
                                                                }
                                                                else if (mcTcor >= 94) {
                                                                    return 1;
                                                                }
                                                            }
                                                            else if (Cs >= 1545) {
                                                                if (SC < 3657) {
                                                                    return 1;
                                                                }
                                                                else if (SC >= 3657) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (TSCindex >= 6) {
                                                        if (mu_mv_mag_sq < 46) {
                                                            if (refDCval < 61) {
                                                                if (MVDiff < 14) {
                                                                    return 8;
                                                                }
                                                                else if (MVDiff >= 14) {
                                                                    if (diffMVdiffVal < 6) {
                                                                        if (Cs < 483) {
                                                                            if (TSC < 6) {
                                                                                return 4;
                                                                            }
                                                                            else if (TSC >= 6) {
                                                                                return 1;
                                                                            }
                                                                        }
                                                                        else if (Cs >= 483) {
                                                                            if (tcor < 89) {
                                                                                if (diffMVdiffVal < -10) {
                                                                                    if (Cs < 559) {
                                                                                        return 4;
                                                                                    }
                                                                                    else if (Cs >= 559) {
                                                                                        return 1;
                                                                                    }
                                                                                }
                                                                                else if (diffMVdiffVal >= -10) {
                                                                                    if (SC < 1418) {
                                                                                        return 1;
                                                                                    }
                                                                                    else if (SC >= 1418) {
                                                                                        if (Cs < 615) {
                                                                                            if (AFD < 7) {
                                                                                                return 1;
                                                                                            }
                                                                                            else if (AFD >= 7) {
                                                                                                return 4;
                                                                                            }
                                                                                        }
                                                                                        else if (Cs >= 615) {
                                                                                            return 1;
                                                                                        }
                                                                                    }
                                                                                }
                                                                            }
                                                                            else if (tcor >= 89) {
                                                                                if (Rs < 1897) {
                                                                                    return 4;
                                                                                }
                                                                                else if (Rs >= 1897) {
                                                                                    if (RsDiff < 2) {
                                                                                        return 4;
                                                                                    }
                                                                                    else if (RsDiff >= 2) {
                                                                                        return 1;
                                                                                    }
                                                                                }
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (diffMVdiffVal >= 6) {
                                                                        if (diffRsCsdiff < -2) {
                                                                            return 1;
                                                                        }
                                                                        else if (diffRsCsdiff >= -2) {
                                                                            if (mcTcor < 80) {
                                                                                if (negBalance < 51) {
                                                                                    return 4;
                                                                                }
                                                                                else if (negBalance >= 51) {
                                                                                    if (Rs < 767) {
                                                                                        return 1;
                                                                                    }
                                                                                    else if (Rs >= 767) {
                                                                                        return 4;
                                                                                    }
                                                                                }
                                                                            }
                                                                            else if (mcTcor >= 80) {
                                                                                if (diffMVdiffVal < 11) {
                                                                                    return 1;
                                                                                }
                                                                                else if (diffMVdiffVal >= 11) {
                                                                                    if (SC < 3378) {
                                                                                        return 4;
                                                                                    }
                                                                                    else if (SC >= 3378) {
                                                                                        return 1;
                                                                                    }
                                                                                }
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                            else if (refDCval >= 61) {
                                                                if (mcTcor < 91) {
                                                                    if (SC < 1996) {
                                                                        if (Cs < 334) {
                                                                            return 2;
                                                                        }
                                                                        else if (Cs >= 334) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                    else if (SC >= 1996) {
                                                                        return 2;
                                                                    }
                                                                }
                                                                else if (mcTcor >= 91) {
                                                                    if (AFD < 8) {
                                                                        return 2;
                                                                    }
                                                                    else if (AFD >= 8) {
                                                                        return 1;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (mu_mv_mag_sq >= 46) {
                                                            if (mcTcor < 93) {
                                                                if (RsDiff < 2) {
                                                                    return 4;
                                                                }
                                                                else if (RsDiff >= 2) {
                                                                    if (tcor < 79) {
                                                                        if (Cs < 558) {
                                                                            if (diffTSC < 0) {
                                                                                return 1;
                                                                            }
                                                                            else if (diffTSC >= 0) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                        else if (Cs >= 558) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                    else if (tcor >= 79) {
                                                                        if (Rs < 1520) {
                                                                            if (SC < 2009) {
                                                                                if (Cs < 526) {
                                                                                    return 4;
                                                                                }
                                                                                else if (Cs >= 526) {
                                                                                    return 1;
                                                                                }
                                                                            }
                                                                            else if (SC >= 2009) {
                                                                                if (Cs < 956) {
                                                                                    return 2;
                                                                                }
                                                                                else if (Cs >= 956) {
                                                                                    return 8;
                                                                                }
                                                                            }
                                                                        }
                                                                        else if (Rs >= 1520) {
                                                                            if (Rs < 1969) {
                                                                                return 4;
                                                                            }
                                                                            else if (Rs >= 1969) {
                                                                                return 1;
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                            else if (mcTcor >= 93) {
                                                                if (gchDC < 2) {
                                                                    return 1;
                                                                }
                                                                else if (gchDC >= 2) {
                                                                    if (Rs < 990) {
                                                                        return 2;
                                                                    }
                                                                    else if (Rs >= 990) {
                                                                        return 4;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (RsCsDiff >= 19) {
                                                    if (Scindex < 4) {
                                                        if (diffRsCsdiff < -9) {
                                                            if (Cs < 603) {
                                                                return 1;
                                                            }
                                                            else if (Cs >= 603) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (diffRsCsdiff >= -9) {
                                                            return 2;
                                                        }
                                                    }
                                                    else if (Scindex >= 4) {
                                                        if (negBalance < 49) {
                                                            if (ssDCval < 120) {
                                                                return 2;
                                                            }
                                                            else if (ssDCval >= 120) {
                                                                return 1;
                                                            }
                                                        }
                                                        else if (negBalance >= 49) {
                                                            if (MVDiff < 20) {
                                                                if (SC < 2314) {
                                                                    return 4;
                                                                }
                                                                else if (SC >= 2314) {
                                                                    if (Rs < 1241) {
                                                                        return 4;
                                                                    }
                                                                    else if (Rs >= 1241) {
                                                                        return 2;
                                                                    }
                                                                }
                                                            }
                                                            else if (MVDiff >= 20) {
                                                                if (Rs < 864) {
                                                                    return 8;
                                                                }
                                                                else if (Rs >= 864) {
                                                                    return 2;
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (Rs >= 2038) {
                                            if (Scindex < 5) {
                                                if (negBalance < 34) {
                                                    if (diffAFD < 1) {
                                                        if (AFD < 6) {
                                                            return 4;
                                                        }
                                                        else if (AFD >= 6) {
                                                            if (RsDiff < 7) {
                                                                return 8;
                                                            }
                                                            else if (RsDiff >= 7) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                    else if (diffAFD >= 1) {
                                                        return 8;
                                                    }
                                                }
                                                else if (negBalance >= 34) {
                                                    if (AFD < 16) {
                                                        if (mu_mv_mag_sq < 9) {
                                                            if (Cs < 378) {
                                                                return 4;
                                                            }
                                                            else if (Cs >= 378) {
                                                                if (Rs < 2223) {
                                                                    return 4;
                                                                }
                                                                else if (Rs >= 2223) {
                                                                    return 2;
                                                                }
                                                            }
                                                        }
                                                        else if (mu_mv_mag_sq >= 9) {
                                                            return 2;
                                                        }
                                                    }
                                                    else if (AFD >= 16) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                            else if (Scindex >= 5) {
                                                if (posBalance < 47) {
                                                    if (mu_mv_mag_sq < 58) {
                                                        if (Rs < 3269) {
                                                            if (diffAFD < -1) {
                                                                if (RsDiff < 5) {
                                                                    return 8;
                                                                }
                                                                else if (RsDiff >= 5) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (diffAFD >= -1) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (Rs >= 3269) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (mu_mv_mag_sq >= 58) {
                                                        return 8;
                                                    }
                                                }
                                                else if (posBalance >= 47) {
                                                    if (CsDiff < 5) {
                                                        if (refDCval < 50) {
                                                            if (SC < 6438) {
                                                                return 4;
                                                            }
                                                            else if (SC >= 6438) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (refDCval >= 50) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (CsDiff >= 5) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (TSC >= 7) {
                                    if (Rs < 2607) {
                                        if (mu_mv_mag_sq < 23) {
                                            if (posBalance < 51) {
                                                if (SC < 3094) {
                                                    if (Cs < 884) {
                                                        return 1;
                                                    }
                                                    else if (Cs >= 884) {
                                                        return 2;
                                                    }
                                                }
                                                else if (SC >= 3094) {
                                                    if (refDCval < 87) {
                                                        if (mcTcor < 92) {
                                                            return 8;
                                                        }
                                                        else if (mcTcor >= 92) {
                                                            return 1;
                                                        }
                                                    }
                                                    else if (refDCval >= 87) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                            else if (posBalance >= 51) {
                                                if (Cs < 1513) {
                                                    if (RsCsDiff < 31) {
                                                        if (SC < 1605) {
                                                            if (posBalance < 62) {
                                                                if (Cs < 507) {
                                                                    return 4;
                                                                }
                                                                else if (Cs >= 507) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (posBalance >= 62) {
                                                                if (negBalance < 33) {
                                                                    if (tcor < 61) {
                                                                        if (Rs < 664) {
                                                                            return 1;
                                                                        }
                                                                        else if (Rs >= 664) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                    else if (tcor >= 61) {
                                                                        if (SC < 1101) {
                                                                            return 2;
                                                                        }
                                                                        else if (SC >= 1101) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                }
                                                                else if (negBalance >= 33) {
                                                                    if (RsDiff < 2) {
                                                                        return 4;
                                                                    }
                                                                    else if (RsDiff >= 2) {
                                                                        return 1;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (SC >= 1605) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (RsCsDiff >= 31) {
                                                        if (refDCval < 122) {
                                                            if (posBalance < 67) {
                                                                if (ssDCval < 63) {
                                                                    if (Cs < 639) {
                                                                        return 2;
                                                                    }
                                                                    else if (Cs >= 639) {
                                                                        return 8;
                                                                    }
                                                                }
                                                                else if (ssDCval >= 63) {
                                                                    if (AFD < 12) {
                                                                        if (posBalance < 57) {
                                                                            return 2;
                                                                        }
                                                                        else if (posBalance >= 57) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                    else if (AFD >= 12) {
                                                                        if (Rs < 2252) {
                                                                            return 2;
                                                                        }
                                                                        else if (Rs >= 2252) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                            else if (posBalance >= 67) {
                                                                if (MVDiff < 149) {
                                                                    if (ssDCval < 84) {
                                                                        if (TSC < 10) {
                                                                            return 2;
                                                                        }
                                                                        else if (TSC >= 10) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                    else if (ssDCval >= 84) {
                                                                        if (Rs < 1881) {
                                                                            return 1;
                                                                        }
                                                                        else if (Rs >= 1881) {
                                                                            return 2;
                                                                        }
                                                                    }
                                                                }
                                                                else if (MVDiff >= 149) {
                                                                    if (Rs < 1426) {
                                                                        return 8;
                                                                    }
                                                                    else if (Rs >= 1426) {
                                                                        return 4;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (refDCval >= 122) {
                                                            if (RsDiff < 11) {
                                                                return 1;
                                                            }
                                                            else if (RsDiff >= 11) {
                                                                return 2;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (Cs >= 1513) {
                                                    if (mcTcor < 83) {
                                                        if (RsDiff < 7) {
                                                            return 4;
                                                        }
                                                        else if (RsDiff >= 7) {
                                                            return 1;
                                                        }
                                                    }
                                                    else if (mcTcor >= 83) {
                                                        if (mcTcor < 92) {
                                                            if (refDCval < 86) {
                                                                if (TSC < 11) {
                                                                    if (gchDC < 1) {
                                                                        return 4;
                                                                    }
                                                                    else if (gchDC >= 1) {
                                                                        return 2;
                                                                    }
                                                                }
                                                                else if (TSC >= 11) {
                                                                    if (negBalance < 52) {
                                                                        if (ssDCval < 80) {
                                                                            return 4;
                                                                        }
                                                                        else if (ssDCval >= 80) {
                                                                            if (Rs < 2153) {
                                                                                return 4;
                                                                            }
                                                                            else if (Rs >= 2153) {
                                                                                return 8;
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (negBalance >= 52) {
                                                                        return 8;
                                                                    }
                                                                }
                                                            }
                                                            else if (refDCval >= 86) {
                                                                if (RsCsDiff < 133) {
                                                                    return 4;
                                                                }
                                                                else if (RsCsDiff >= 133) {
                                                                    if (Rs < 2371) {
                                                                        return 4;
                                                                    }
                                                                    else if (Rs >= 2371) {
                                                                        if (RsDiff < 11) {
                                                                            return 1;
                                                                        }
                                                                        else if (RsDiff >= 11) {
                                                                            return 2;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (mcTcor >= 92) {
                                                            if (SC < 4325) {
                                                                if (RsDiff < 6) {
                                                                    return 4;
                                                                }
                                                                else if (RsDiff >= 6) {
                                                                    if (gchDC < 1) {
                                                                        if (TSC < 9) {
                                                                            if (Cs < 2110) {
                                                                                return 1;
                                                                            }
                                                                            else if (Cs >= 2110) {
                                                                                return 8;
                                                                            }
                                                                        }
                                                                        else if (TSC >= 9) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                    else if (gchDC >= 1) {
                                                                        return 8;
                                                                    }
                                                                }
                                                            }
                                                            else if (SC >= 4325) {
                                                                if (ssDCval < 89) {
                                                                    return 4;
                                                                }
                                                                else if (ssDCval >= 89) {
                                                                    if (SC < 4508) {
                                                                        return 1;
                                                                    }
                                                                    else if (SC >= 4508) {
                                                                        return 2;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (mu_mv_mag_sq >= 23) {
                                            if (AFD < 20) {
                                                if (Rs < 1932) {
                                                    if (Cs < 422) {
                                                        if (diffMVdiffVal < -7) {
                                                            return 1;
                                                        }
                                                        else if (diffMVdiffVal >= -7) {
                                                            if (ssDCval < 110) {
                                                                if (diffMVdiffVal < 1) {
                                                                    return 8;
                                                                }
                                                                else if (diffMVdiffVal >= 1) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (ssDCval >= 110) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                    else if (Cs >= 422) {
                                                        if (posBalance < 59) {
                                                            if (diffMVdiffVal < 14) {
                                                                if (RsCsDiff < 129) {
                                                                    if (ssDCval < 117) {
                                                                        if (mcTcor < 95) {
                                                                            return 8;
                                                                        }
                                                                        else if (mcTcor >= 95) {
                                                                            if (AFD < 9) {
                                                                                return 1;
                                                                            }
                                                                            else if (AFD >= 9) {
                                                                                return 8;
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (ssDCval >= 117) {
                                                                        if (RsDiff < 6) {
                                                                            return 8;
                                                                        }
                                                                        else if (RsDiff >= 6) {
                                                                            if (Cs < 1634) {
                                                                                return 1;
                                                                            }
                                                                            else if (Cs >= 1634) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                                else if (RsCsDiff >= 129) {
                                                                    return 2;
                                                                }
                                                            }
                                                            else if (diffMVdiffVal >= 14) {
                                                                if (Cs < 925) {
                                                                    return 4;
                                                                }
                                                                else if (Cs >= 925) {
                                                                    if (Cs < 1265) {
                                                                        return 8;
                                                                    }
                                                                    else if (Cs >= 1265) {
                                                                        return 4;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (posBalance >= 59) {
                                                            if (Rs < 928) {
                                                                return 8;
                                                            }
                                                            else if (Rs >= 928) {
                                                                if (refDCval < 77) {
                                                                    if (SC < 2835) {
                                                                        if (Cs < 902) {
                                                                            return 1;
                                                                        }
                                                                        else if (Cs >= 902) {
                                                                            if (TSC < 8) {
                                                                                return 1;
                                                                            }
                                                                            else if (TSC >= 8) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (SC >= 2835) {
                                                                        return 8;
                                                                    }
                                                                }
                                                                else if (refDCval >= 77) {
                                                                    if (mu_mv_mag_sq < 45) {
                                                                        if (Cs < 727) {
                                                                            return 2;
                                                                        }
                                                                        else if (Cs >= 727) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                    else if (mu_mv_mag_sq >= 45) {
                                                                        if (TSCindex < 9) {
                                                                            if (posBalance < 60) {
                                                                                if (Rs < 1735) {
                                                                                    return 8;
                                                                                }
                                                                                else if (Rs >= 1735) {
                                                                                    return 2;
                                                                                }
                                                                            }
                                                                            else if (posBalance >= 60) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                        else if (TSCindex >= 9) {
                                                                            if (Rs < 1776) {
                                                                                if (RsDiff < 5) {
                                                                                    return 4;
                                                                                }
                                                                                else if (RsDiff >= 5) {
                                                                                    return 8;
                                                                                }
                                                                            }
                                                                            else if (Rs >= 1776) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (Rs >= 1932) {
                                                    if (posBalance < 56) {
                                                        if (ssDCval < 81) {
                                                            if (TSC < 8) {
                                                                return 2;
                                                            }
                                                            else if (TSC >= 8) {
                                                                if (SC < 4989) {
                                                                    return 8;
                                                                }
                                                                else if (SC >= 4989) {
                                                                    return 1;
                                                                }
                                                            }
                                                        }
                                                        else if (ssDCval >= 81) {
                                                            if (SC < 4535) {
                                                                if (negBalance < 32) {
                                                                    if (RsDiff < 9) {
                                                                        return 8;
                                                                    }
                                                                    else if (RsDiff >= 9) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (negBalance >= 32) {
                                                                    if (diffMVdiffVal < -8) {
                                                                        return 8;
                                                                    }
                                                                    else if (diffMVdiffVal >= -8) {
                                                                        if (posBalance < 43) {
                                                                            return 8;
                                                                        }
                                                                        else if (posBalance >= 43) {
                                                                            if (posBalance < 55) {
                                                                                return 1;
                                                                            }
                                                                            else if (posBalance >= 55) {
                                                                                if (Cs < 1977) {
                                                                                    return 1;
                                                                                }
                                                                                else if (Cs >= 1977) {
                                                                                    return 2;
                                                                                }
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                            else if (SC >= 4535) {
                                                                if (MVDiff < 27) {
                                                                    if (negBalance < 39) {
                                                                        return 8;
                                                                    }
                                                                    else if (negBalance >= 39) {
                                                                        if (RsDiff < 9) {
                                                                            return 1;
                                                                        }
                                                                        else if (RsDiff >= 9) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                }
                                                                else if (MVDiff >= 27) {
                                                                    if (Rs < 2538) {
                                                                        return 4;
                                                                    }
                                                                    else if (Rs >= 2538) {
                                                                        return 8;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (posBalance >= 56) {
                                                        if (mcTcor < 94) {
                                                            if (refDCval < 82) {
                                                                if (Cs < 1966) {
                                                                    return 1;
                                                                }
                                                                else if (Cs >= 1966) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (refDCval >= 82) {
                                                                if (RsDiff < 10) {
                                                                    if (RsCsDiff < 92) {
                                                                        if (Rs < 2024) {
                                                                            return 8;
                                                                        }
                                                                        else if (Rs >= 2024) {
                                                                            if (Cs < 2034) {
                                                                                return 2;
                                                                            }
                                                                            else if (Cs >= 2034) {
                                                                                return 1;
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (RsCsDiff >= 92) {
                                                                        if (ssDCval < 109) {
                                                                            if (refDCval < 88) {
                                                                                return 4;
                                                                            }
                                                                            else if (refDCval >= 88) {
                                                                                if (gchDC < 3) {
                                                                                    return 2;
                                                                                }
                                                                                else if (gchDC >= 3) {
                                                                                    return 1;
                                                                                }
                                                                            }
                                                                        }
                                                                        else if (ssDCval >= 109) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                }
                                                                else if (RsDiff >= 10) {
                                                                    if (ssDCval < 117) {
                                                                        if (SC < 4151) {
                                                                            return 2;
                                                                        }
                                                                        else if (SC >= 4151) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                    else if (ssDCval >= 117) {
                                                                        return 8;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (mcTcor >= 94) {
                                                            if (Rs < 2541) {
                                                                return 2;
                                                            }
                                                            else if (Rs >= 2541) {
                                                                if (Cs < 1832) {
                                                                    return 1;
                                                                }
                                                                else if (Cs >= 1832) {
                                                                    return 2;
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                            else if (AFD >= 20) {
                                                if (RsDiff < 8) {
                                                    return 1;
                                                }
                                                else if (RsDiff >= 8) {
                                                    if (SC < 3837) {
                                                        if (Rs < 612) {
                                                            return 8;
                                                        }
                                                        else if (Rs >= 612) {
                                                            return 2;
                                                        }
                                                    }
                                                    else if (SC >= 3837) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (Rs >= 2607) {
                                        if (refDCval < 101) {
                                            if (mu_mv_mag_sq < 46) {
                                                if (mcTcor < 90) {
                                                    if (MVDiff < 12) {
                                                        if (SC < 4373) {
                                                            return 2;
                                                        }
                                                        else if (SC >= 4373) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (MVDiff >= 12) {
                                                        return 8;
                                                    }
                                                }
                                                else if (mcTcor >= 90) {
                                                    if (Rs < 3253) {
                                                        if (tcor < 87) {
                                                            return 2;
                                                        }
                                                        else if (tcor >= 87) {
                                                            if (diffAFD < -1) {
                                                                if (Cs < 2081) {
                                                                    return 1;
                                                                }
                                                                else if (Cs >= 2081) {
                                                                    return 2;
                                                                }
                                                            }
                                                            else if (diffAFD >= -1) {
                                                                return 1;
                                                            }
                                                        }
                                                    }
                                                    else if (Rs >= 3253) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                            else if (mu_mv_mag_sq >= 46) {
                                                if (SC < 5169) {
                                                    if (TSC < 8) {
                                                        return 2;
                                                    }
                                                    else if (TSC >= 8) {
                                                        if (TSC < 12) {
                                                            return 1;
                                                        }
                                                        else if (TSC >= 12) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                                else if (SC >= 5169) {
                                                    if (diffAFD < 2) {
                                                        if (refDCval < 92) {
                                                            if (mcTcor < 87) {
                                                                return 4;
                                                            }
                                                            else if (mcTcor >= 87) {
                                                                if (SC < 6211) {
                                                                    if (MVDiff < 8) {
                                                                        if (ssDCval < 86) {
                                                                            return 8;
                                                                        }
                                                                        else if (ssDCval >= 86) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                    else if (MVDiff >= 8) {
                                                                        return 1;
                                                                    }
                                                                }
                                                                else if (SC >= 6211) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                        else if (refDCval >= 92) {
                                                            if (MVDiff < 11) {
                                                                if (Rs < 3047) {
                                                                    return 4;
                                                                }
                                                                else if (Rs >= 3047) {
                                                                    if (Rs < 3253) {
                                                                        return 1;
                                                                    }
                                                                    else if (Rs >= 3253) {
                                                                        return 4;
                                                                    }
                                                                }
                                                            }
                                                            else if (MVDiff >= 11) {
                                                                if (mu_mv_mag_sq < 63) {
                                                                    if (Cs < 2341) {
                                                                        if (Rs < 3004) {
                                                                            return 4;
                                                                        }
                                                                        else if (Rs >= 3004) {
                                                                            return 2;
                                                                        }
                                                                    }
                                                                    else if (Cs >= 2341) {
                                                                        if (gchDC < 1) {
                                                                            return 1;
                                                                        }
                                                                        else if (gchDC >= 1) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                }
                                                                else if (mu_mv_mag_sq >= 63) {
                                                                    return 1;
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (diffAFD >= 2) {
                                                        if (Cs < 2453) {
                                                            return 4;
                                                        }
                                                        else if (Cs >= 2453) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (refDCval >= 101) {
                                            if (Rs < 2797) {
                                                if (TSC < 9) {
                                                    if (Rs < 2791) {
                                                        return 4;
                                                    }
                                                    else if (Rs >= 2791) {
                                                        return 2;
                                                    }
                                                }
                                                else if (TSC >= 9) {
                                                    if (Cs < 1885) {
                                                        if (Cs < 1010) {
                                                            return 2;
                                                        }
                                                        else if (Cs >= 1010) {
                                                            if (Cs < 1640) {
                                                                return 4;
                                                            }
                                                            else if (Cs >= 1640) {
                                                                if (Cs < 1797) {
                                                                    return 1;
                                                                }
                                                                else if (Cs >= 1797) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (Cs >= 1885) {
                                                        return 1;
                                                    }
                                                }
                                            }
                                            else if (Rs >= 2797) {
                                                if (MVDiff < 16) {
                                                    if (ssDCval < 117) {
                                                        if (Cs < 2233) {
                                                            if (CsDiff < 6) {
                                                                return 4;
                                                            }
                                                            else if (CsDiff >= 6) {
                                                                return 2;
                                                            }
                                                        }
                                                        else if (Cs >= 2233) {
                                                            if (Rs < 3014) {
                                                                return 1;
                                                            }
                                                            else if (Rs >= 3014) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                    else if (ssDCval >= 117) {
                                                        if (SC < 4631) {
                                                            return 8;
                                                        }
                                                        else if (SC >= 4631) {
                                                            return 1;
                                                        }
                                                    }
                                                }
                                                else if (MVDiff >= 16) {
                                                    if (Cs < 2307) {
                                                        if (CsDiff < 7) {
                                                            if (diffRsCsdiff < 28) {
                                                                return 2;
                                                            }
                                                            else if (diffRsCsdiff >= 28) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (CsDiff >= 7) {
                                                            return 2;
                                                        }
                                                    }
                                                    else if (Cs >= 2307) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                            else if (mu_mv_mag_sq >= 74) {
                                if (SC < 1423) {
                                    if (RsDiff < 3) {
                                        if (ssDCval < 52) {
                                            if (negBalance < 28) {
                                                if (gchDC < 2) {
                                                    if (negBalance < 18) {
                                                        return 8;
                                                    }
                                                    else if (negBalance >= 18) {
                                                        if (Cs < 89) {
                                                            return 4;
                                                        }
                                                        else if (Cs >= 89) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                                else if (gchDC >= 2) {
                                                    if (diffAFD < -14) {
                                                        return 8;
                                                    }
                                                    else if (diffAFD >= -14) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                            else if (negBalance >= 28) {
                                                if (Cs < 366) {
                                                    if (Cs < 338) {
                                                        return 4;
                                                    }
                                                    else if (Cs >= 338) {
                                                        return 1;
                                                    }
                                                }
                                                else if (Cs >= 366) {
                                                    return 4;
                                                }
                                            }
                                        }
                                        else if (ssDCval >= 52) {
                                            if (mu_mv_mag_sq < 105) {
                                                if (Rs < 263) {
                                                    return 2;
                                                }
                                                else if (Rs >= 263) {
                                                    if (ssDCval < 164) {
                                                        if (ssDCval < 128) {
                                                            if (Rs < 292) {
                                                                return 4;
                                                            }
                                                            else if (Rs >= 292) {
                                                                if (AFD < 6) {
                                                                    return 8;
                                                                }
                                                                else if (AFD >= 6) {
                                                                    return 2;
                                                                }
                                                            }
                                                        }
                                                        else if (ssDCval >= 128) {
                                                            if (mu_mv_mag_sq < 85) {
                                                                return 4;
                                                            }
                                                            else if (mu_mv_mag_sq >= 85) {
                                                                if (diffMVdiffVal < 2) {
                                                                    return 2;
                                                                }
                                                                else if (diffMVdiffVal >= 2) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (ssDCval >= 164) {
                                                        if (Rs < 471) {
                                                            return 4;
                                                        }
                                                        else if (Rs >= 471) {
                                                            return 2;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (mu_mv_mag_sq >= 105) {
                                                if (Rs < 226) {
                                                    return 2;
                                                }
                                                else if (Rs >= 226) {
                                                    return 8;
                                                }
                                            }
                                        }
                                    }
                                    else if (RsDiff >= 3) {
                                        if (RsDiff < 5) {
                                            return 1;
                                        }
                                        else if (RsDiff >= 5) {
                                            if (SC < 1101) {
                                                return 2;
                                            }
                                            else if (SC >= 1101) {
                                                return 1;
                                            }
                                        }
                                    }
                                }
                                else if (SC >= 1423) {
                                    if (refDCval < 131) {
                                        if (negBalance < 52) {
                                            if (mu_mv_mag_sq < 92) {
                                                if (RsDiff < 2) {
                                                    if (ssDCval < 43) {
                                                        if (diffRsCsdiff < -5) {
                                                            if (SC < 3221) {
                                                                return 8;
                                                            }
                                                            else if (SC >= 3221) {
                                                                return 1;
                                                            }
                                                        }
                                                        else if (diffRsCsdiff >= -5) {
                                                            if (negBalance < 44) {
                                                                return 4;
                                                            }
                                                            else if (negBalance >= 44) {
                                                                if (Rs < 1860) {
                                                                    return 8;
                                                                }
                                                                else if (Rs >= 1860) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (ssDCval >= 43) {
                                                        if (ssDCval < 74) {
                                                            return 1;
                                                        }
                                                        else if (ssDCval >= 74) {
                                                            if (Rs < 468) {
                                                                return 4;
                                                            }
                                                            else if (Rs >= 468) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (RsDiff >= 2) {
                                                    if (SC < 4301) {
                                                        if (MVDiff < 19) {
                                                            if (ssDCval < 85) {
                                                                if (Cs < 987) {
                                                                    if (mu_mv_mag_sq < 85) {
                                                                        if (SC < 1911) {
                                                                            return 2;
                                                                        }
                                                                        else if (SC >= 1911) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                    else if (mu_mv_mag_sq >= 85) {
                                                                        return 8;
                                                                    }
                                                                }
                                                                else if (Cs >= 987) {
                                                                    if (negBalance < 45) {
                                                                        if (posBalance < 53) {
                                                                            return 4;
                                                                        }
                                                                        else if (posBalance >= 53) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                    else if (negBalance >= 45) {
                                                                        return 8;
                                                                    }
                                                                }
                                                            }
                                                            else if (ssDCval >= 85) {
                                                                if (SC < 4170) {
                                                                    return 8;
                                                                }
                                                                else if (SC >= 4170) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                        else if (MVDiff >= 19) {
                                                            if (SC < 3857) {
                                                                return 4;
                                                            }
                                                            else if (SC >= 3857) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                    else if (SC >= 4301) {
                                                        if (RsCsDiff < 87) {
                                                            if (refDCval < 76) {
                                                                return 8;
                                                            }
                                                            else if (refDCval >= 76) {
                                                                if (Cs < 2383) {
                                                                    if (Rs < 3077) {
                                                                        if (Rs < 2505) {
                                                                            return 8;
                                                                        }
                                                                        else if (Rs >= 2505) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                    else if (Rs >= 3077) {
                                                                        return 8;
                                                                    }
                                                                }
                                                                else if (Cs >= 2383) {
                                                                    if (MVDiff < 13) {
                                                                        return 8;
                                                                    }
                                                                    else if (MVDiff >= 13) {
                                                                        return 1;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (RsCsDiff >= 87) {
                                                            if (negBalance < 50) {
                                                                if (TSCindex < 8) {
                                                                    return 1;
                                                                }
                                                                else if (TSCindex >= 8) {
                                                                    if (refDCval < 98) {
                                                                        return 8;
                                                                    }
                                                                    else if (refDCval >= 98) {
                                                                        if (MVDiff < 9) {
                                                                            if (TSC < 8) {
                                                                                return 4;
                                                                            }
                                                                            else if (TSC >= 8) {
                                                                                return 8;
                                                                            }
                                                                        }
                                                                        else if (MVDiff >= 9) {
                                                                            if (mcTcor < 89) {
                                                                                return 1;
                                                                            }
                                                                            else if (mcTcor >= 89) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                            else if (negBalance >= 50) {
                                                                if (Cs < 2689) {
                                                                    return 4;
                                                                }
                                                                else if (Cs >= 2689) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                            else if (mu_mv_mag_sq >= 92) {
                                                if (ssDCval < 21) {
                                                    return 2;
                                                }
                                                else if (ssDCval >= 21) {
                                                    if (mu_mv_mag_sq < 149) {
                                                        if (tcor < 102) {
                                                            if (MVDiff < 237) {
                                                                if (RsCsDiff < 3) {
                                                                    if (MVDiff < 11) {
                                                                        if (diffMVdiffVal < -2) {
                                                                            if (Cs < 1058) {
                                                                                if (refDCval < 28) {
                                                                                    return 4;
                                                                                }
                                                                                else if (refDCval >= 28) {
                                                                                    return 8;
                                                                                }
                                                                            }
                                                                            else if (Cs >= 1058) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                        else if (diffMVdiffVal >= -2) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                    else if (MVDiff >= 11) {
                                                                        if (diffRsCsdiff < -161) {
                                                                            return 8;
                                                                        }
                                                                        else if (diffRsCsdiff >= -161) {
                                                                            if (Rs < 2459) {
                                                                                if (negBalance < 42) {
                                                                                    return 4;
                                                                                }
                                                                                else if (negBalance >= 42) {
                                                                                    return 8;
                                                                                }
                                                                            }
                                                                            else if (Rs >= 2459) {
                                                                                return 8;
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                                else if (RsCsDiff >= 3) {
                                                                    if (RsCsDiff < 110) {
                                                                        if (ssDCval < 130) {
                                                                            if (negBalance < 36) {
                                                                                return 8;
                                                                            }
                                                                            else if (negBalance >= 36) {
                                                                                if (refDCval < 44) {
                                                                                    return 4;
                                                                                }
                                                                                else if (refDCval >= 44) {
                                                                                    if (tcor < 93) {
                                                                                        return 8;
                                                                                    }
                                                                                    else if (tcor >= 93) {
                                                                                        if (ssDCval < 72) {
                                                                                            return 8;
                                                                                        }
                                                                                        else if (ssDCval >= 72) {
                                                                                            if (ssDCval < 92) {
                                                                                                if (posBalance < 30) {
                                                                                                    return 8;
                                                                                                }
                                                                                                else if (posBalance >= 30) {
                                                                                                    return 4;
                                                                                                }
                                                                                            }
                                                                                            else if (ssDCval >= 92) {
                                                                                                return 8;
                                                                                            }
                                                                                        }
                                                                                    }
                                                                                }
                                                                            }
                                                                        }
                                                                        else if (ssDCval >= 130) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                    else if (RsCsDiff >= 110) {
                                                                        if (RsDiff < 13) {
                                                                            if (refDCval < 115) {
                                                                                return 4;
                                                                            }
                                                                            else if (refDCval >= 115) {
                                                                                if (gchDC < 1) {
                                                                                    return 4;
                                                                                }
                                                                                else if (gchDC >= 1) {
                                                                                    return 8;
                                                                                }
                                                                            }
                                                                        }
                                                                        else if (RsDiff >= 13) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                            else if (MVDiff >= 237) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (tcor >= 102) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (mu_mv_mag_sq >= 149) {
                                                        return 1;
                                                    }
                                                }
                                            }
                                        }
                                        else if (negBalance >= 52) {
                                            if (Rs < 2339) {
                                                if (ssDCval < 92) {
                                                    return 8;
                                                }
                                                else if (ssDCval >= 92) {
                                                    if (Cs < 1820) {
                                                        return 8;
                                                    }
                                                    else if (Cs >= 1820) {
                                                        return 2;
                                                    }
                                                }
                                            }
                                            else if (Rs >= 2339) {
                                                if (Cs < 1611) {
                                                    return 8;
                                                }
                                                else if (Cs >= 1611) {
                                                    return 4;
                                                }
                                            }
                                        }
                                    }
                                    else if (refDCval >= 131) {
                                        if (ssDCval < 141) {
                                            if (Rs < 808) {
                                                return 8;
                                            }
                                            else if (Rs >= 808) {
                                                return 1;
                                            }
                                        }
                                        else if (ssDCval >= 141) {
                                            if (CsDiff < 5) {
                                                return 8;
                                            }
                                            else if (CsDiff >= 5) {
                                                if (Rs < 1861) {
                                                    return 1;
                                                }
                                                else if (Rs >= 1861) {
                                                    return 2;
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                        }
                        else if (Rs >= 3316) {
                            if (RsCsDiff < 48) {
                                if (negBalance < 46) {
                                    if (refDCval < 126) {
                                        if (SC < 7004) {
                                            if (posBalance < 54) {
                                                return 8;
                                            }
                                            else if (posBalance >= 54) {
                                                return 4;
                                            }
                                        }
                                        else if (SC >= 7004) {
                                            if (Rs < 4366) {
                                                return 8;
                                            }
                                            else if (Rs >= 4366) {
                                                return 2;
                                            }
                                        }
                                    }
                                    else if (refDCval >= 126) {
                                        return 4;
                                    }
                                }
                                else if (negBalance >= 46) {
                                    if (posBalance < 57) {
                                        if (AFD < 9) {
                                            return 4;
                                        }
                                        else if (AFD >= 9) {
                                            return 8;
                                        }
                                    }
                                    else if (posBalance >= 57) {
                                        return 2;
                                    }
                                }
                            }
                            else if (RsCsDiff >= 48) {
                                if (tcor < 81) {
                                    if (ssDCval < 107) {
                                        if (RsDiff < 9) {
                                            if (posBalance < 46) {
                                                return 8;
                                            }
                                            else if (posBalance >= 46) {
                                                return 4;
                                            }
                                        }
                                        else if (RsDiff >= 9) {
                                            return 4;
                                        }
                                    }
                                    else if (ssDCval >= 107) {
                                        if (Rs < 4747) {
                                            if (AFD < 17) {
                                                if (AFD < 14) {
                                                    if (refDCval < 108) {
                                                        if (RsDiff < 8) {
                                                            return 1;
                                                        }
                                                        else if (RsDiff >= 8) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (refDCval >= 108) {
                                                        return 1;
                                                    }
                                                }
                                                else if (AFD >= 14) {
                                                    if (Rs < 3756) {
                                                        return 1;
                                                    }
                                                    else if (Rs >= 3756) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                            else if (AFD >= 17) {
                                                if (Rs < 3849) {
                                                    return 2;
                                                }
                                                else if (Rs >= 3849) {
                                                    return 4;
                                                }
                                            }
                                        }
                                        else if (Rs >= 4747) {
                                            return 2;
                                        }
                                    }
                                }
                                else if (tcor >= 81) {
                                    if (Cs < 2069) {
                                        if (refDCval < 111) {
                                            if (TSC < 6) {
                                                if (ssDCval < 109) {
                                                    if (refDCval < 107) {
                                                        if (Cs < 1337) {
                                                            return 4;
                                                        }
                                                        else if (Cs >= 1337) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (refDCval >= 107) {
                                                        if (diffAFD < 1) {
                                                            if (Rs < 3733) {
                                                                return 8;
                                                            }
                                                            else if (Rs >= 3733) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (diffAFD >= 1) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                                else if (ssDCval >= 109) {
                                                    return 4;
                                                }
                                            }
                                            else if (TSC >= 6) {
                                                if (negBalance < 46) {
                                                    if (Rs < 4229) {
                                                        if (tcor < 94) {
                                                            return 4;
                                                        }
                                                        else if (tcor >= 94) {
                                                            if (Cs < 2027) {
                                                                return 4;
                                                            }
                                                            else if (Cs >= 2027) {
                                                                return 2;
                                                            }
                                                        }
                                                    }
                                                    else if (Rs >= 4229) {
                                                        return 8;
                                                    }
                                                }
                                                else if (negBalance >= 46) {
                                                    if (AFD < 11) {
                                                        return 8;
                                                    }
                                                    else if (AFD >= 11) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                        }
                                        else if (refDCval >= 111) {
                                            if (RsCsDiff < 262) {
                                                return 4;
                                            }
                                            else if (RsCsDiff >= 262) {
                                                return 2;
                                            }
                                        }
                                    }
                                    else if (Cs >= 2069) {
                                        if (refDCval < 130) {
                                            if (ssDCval < 107) {
                                                if (AFD < 11) {
                                                    if (SC < 5965) {
                                                        if (TSC < 7) {
                                                            return 2;
                                                        }
                                                        else if (TSC >= 7) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (SC >= 5965) {
                                                        return 8;
                                                    }
                                                }
                                                else if (AFD >= 11) {
                                                    if (ssDCval < 98) {
                                                        if (Rs < 3642) {
                                                            return 1;
                                                        }
                                                        else if (Rs >= 3642) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (ssDCval >= 98) {
                                                        if (Cs < 2335) {
                                                            if (posBalance < 44) {
                                                                return 4;
                                                            }
                                                            else if (posBalance >= 44) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (Cs >= 2335) {
                                                            if (SC < 7929) {
                                                                return 2;
                                                            }
                                                            else if (SC >= 7929) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                            else if (ssDCval >= 107) {
                                                if (CsDiff < 7) {
                                                    if (Rs < 4520) {
                                                        if (tcor < 92) {
                                                            if (gchDC < 1) {
                                                                return 4;
                                                            }
                                                            else if (gchDC >= 1) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (tcor >= 92) {
                                                            if (ssDCval < 119) {
                                                                if (tcor < 93) {
                                                                    return 4;
                                                                }
                                                                else if (tcor >= 93) {
                                                                    return 2;
                                                                }
                                                            }
                                                            else if (ssDCval >= 119) {
                                                                if (ssDCval < 127) {
                                                                    return 4;
                                                                }
                                                                else if (ssDCval >= 127) {
                                                                    return 2;
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (Rs >= 4520) {
                                                        if (RsCsDiff < 87) {
                                                            if (SC < 7199) {
                                                                return 4;
                                                            }
                                                            else if (SC >= 7199) {
                                                                if (Rs < 5104) {
                                                                    if (MVDiff < 10) {
                                                                        return 2;
                                                                    }
                                                                    else if (MVDiff >= 10) {
                                                                        return 8;
                                                                    }
                                                                }
                                                                else if (Rs >= 5104) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                        else if (RsCsDiff >= 87) {
                                                            if (mu_mv_mag_sq < 75) {
                                                                if (Rs < 5088) {
                                                                    if (TSCindex < 8) {
                                                                        return 4;
                                                                    }
                                                                    else if (TSCindex >= 8) {
                                                                        return 8;
                                                                    }
                                                                }
                                                                else if (Rs >= 5088) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (mu_mv_mag_sq >= 75) {
                                                                if (mcTcor < 86) {
                                                                    return 2;
                                                                }
                                                                else if (mcTcor >= 86) {
                                                                    if (ssDCval < 127) {
                                                                        return 8;
                                                                    }
                                                                    else if (ssDCval >= 127) {
                                                                        return 2;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (CsDiff >= 7) {
                                                    if (refDCval < 128) {
                                                        if (SC < 6411) {
                                                            if (Cs < 2233) {
                                                                if (Rs < 3821) {
                                                                    return 2;
                                                                }
                                                                else if (Rs >= 3821) {
                                                                    if (Cs < 2196) {
                                                                        return 4;
                                                                    }
                                                                    else if (Cs >= 2196) {
                                                                        if (diffRsCsdiff < 29) {
                                                                            return 2;
                                                                        }
                                                                        else if (diffRsCsdiff >= 29) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                            else if (Cs >= 2233) {
                                                                return 2;
                                                            }
                                                        }
                                                        else if (SC >= 6411) {
                                                            if (mcTcor < 88) {
                                                                if (Rs < 4398) {
                                                                    if (SC < 6851) {
                                                                        return 4;
                                                                    }
                                                                    else if (SC >= 6851) {
                                                                        return 1;
                                                                    }
                                                                }
                                                                else if (Rs >= 4398) {
                                                                    if (RsDiff < 10) {
                                                                        if (Cs < 2953) {
                                                                            return 2;
                                                                        }
                                                                        else if (Cs >= 2953) {
                                                                            if (Cs < 3078) {
                                                                                return 4;
                                                                            }
                                                                            else if (Cs >= 3078) {
                                                                                return 8;
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (RsDiff >= 10) {
                                                                        return 2;
                                                                    }
                                                                }
                                                            }
                                                            else if (mcTcor >= 88) {
                                                                if (MVDiff < 12) {
                                                                    if (ssDCval < 123) {
                                                                        if (AFD < 8) {
                                                                            if (Cs < 2424) {
                                                                                return 8;
                                                                            }
                                                                            else if (Cs >= 2424) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                        else if (AFD >= 8) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                    else if (ssDCval >= 123) {
                                                                        if (Rs < 4670) {
                                                                            return 4;
                                                                        }
                                                                        else if (Rs >= 4670) {
                                                                            if (Rs < 4676) {
                                                                                return 2;
                                                                            }
                                                                            else if (Rs >= 4676) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                                else if (MVDiff >= 12) {
                                                                    if (Rs < 4295) {
                                                                        if (SC < 6615) {
                                                                            return 4;
                                                                        }
                                                                        else if (SC >= 6615) {
                                                                            return 2;
                                                                        }
                                                                    }
                                                                    else if (Rs >= 4295) {
                                                                        return 4;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (refDCval >= 128) {
                                                        if (AFD < 11) {
                                                            return 2;
                                                        }
                                                        else if (AFD >= 11) {
                                                            if (diffMVdiffVal < 1) {
                                                                return 8;
                                                            }
                                                            else if (diffMVdiffVal >= 1) {
                                                                if (SC < 7279) {
                                                                    return 2;
                                                                }
                                                                else if (SC >= 7279) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (refDCval >= 130) {
                                            return 8;
                                        }
                                    }
                                }
                            }
                        }
                    }
                    else if (CsDiff >= 10) {
                        if (tcor < 90) {
                            if (ssDCval < 150) {
                                if (tcor < 83) {
                                    if (refDCval < 122) {
                                        if (refDCval < 93) {
                                            if (TSC < 11) {
                                                if (ssDCval < 60) {
                                                    if (Rs < 1410) {
                                                        return 2;
                                                    }
                                                    else if (Rs >= 1410) {
                                                        if (MVDiff < 10) {
                                                            return 4;
                                                        }
                                                        else if (MVDiff >= 10) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                                else if (ssDCval >= 60) {
                                                    if (negBalance < 51) {
                                                        if (refDCval < 83) {
                                                            if (Rs < 774) {
                                                                return 2;
                                                            }
                                                            else if (Rs >= 774) {
                                                                return 1;
                                                            }
                                                        }
                                                        else if (refDCval >= 83) {
                                                            if (mu_mv_mag_sq < 28) {
                                                                if (RsDiff < 13) {
                                                                    return 1;
                                                                }
                                                                else if (RsDiff >= 13) {
                                                                    if (Rs < 2369) {
                                                                        return 1;
                                                                    }
                                                                    else if (Rs >= 2369) {
                                                                        return 4;
                                                                    }
                                                                }
                                                            }
                                                            else if (mu_mv_mag_sq >= 28) {
                                                                if (Rs < 2105) {
                                                                    return 2;
                                                                }
                                                                else if (Rs >= 2105) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (negBalance >= 51) {
                                                        if (Cs < 2648) {
                                                            return 2;
                                                        }
                                                        else if (Cs >= 2648) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (TSC >= 11) {
                                                if (mu_mv_mag_sq < 73) {
                                                    if (mu_mv_mag_sq < 3) {
                                                        return 8;
                                                    }
                                                    else if (mu_mv_mag_sq >= 3) {
                                                        if (mcTcor < 89) {
                                                            if (ssDCval < 76) {
                                                                if (Cs < 1822) {
                                                                    return 4;
                                                                }
                                                                else if (Cs >= 1822) {
                                                                    return 1;
                                                                }
                                                            }
                                                            else if (ssDCval >= 76) {
                                                                if (Cs < 2157) {
                                                                    if (diffMVdiffVal < 109) {
                                                                        return 1;
                                                                    }
                                                                    else if (diffMVdiffVal >= 109) {
                                                                        if (Rs < 509) {
                                                                            return 4;
                                                                        }
                                                                        else if (Rs >= 509) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                }
                                                                else if (Cs >= 2157) {
                                                                    if (AFD < 16) {
                                                                        return 4;
                                                                    }
                                                                    else if (AFD >= 16) {
                                                                        return 8;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (mcTcor >= 89) {
                                                            if (TSC < 12) {
                                                                if (MVDiff < 45) {
                                                                    return 4;
                                                                }
                                                                else if (MVDiff >= 45) {
                                                                    return 1;
                                                                }
                                                            }
                                                            else if (TSC >= 12) {
                                                                return 1;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (mu_mv_mag_sq >= 73) {
                                                    return 4;
                                                }
                                            }
                                        }
                                        else if (refDCval >= 93) {
                                            if (Cs < 3093) {
                                                if (MVDiff < 23) {
                                                    if (diffRsCsdiff < 10) {
                                                        if (MVDiff < 8) {
                                                            return 4;
                                                        }
                                                        else if (MVDiff >= 8) {
                                                            if (RsCsDiff < 259) {
                                                                return 1;
                                                            }
                                                            else if (RsCsDiff >= 259) {
                                                                if (RsCsDiff < 329) {
                                                                    return 2;
                                                                }
                                                                else if (RsCsDiff >= 329) {
                                                                    return 1;
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (diffRsCsdiff >= 10) {
                                                        if (posBalance < 52) {
                                                            return 4;
                                                        }
                                                        else if (posBalance >= 52) {
                                                            return 2;
                                                        }
                                                    }
                                                }
                                                else if (MVDiff >= 23) {
                                                    if (CsDiff < 11) {
                                                        if (MVDiff < 27) {
                                                            if (Cs < 2458) {
                                                                return 8;
                                                            }
                                                            else if (Cs >= 2458) {
                                                                if (Cs < 2984) {
                                                                    return 4;
                                                                }
                                                                else if (Cs >= 2984) {
                                                                    return 1;
                                                                }
                                                            }
                                                        }
                                                        else if (MVDiff >= 27) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (CsDiff >= 11) {
                                                        if (Rs < 3326) {
                                                            if (AFD < 19) {
                                                                return 4;
                                                            }
                                                            else if (AFD >= 19) {
                                                                if (AFD < 52) {
                                                                    return 1;
                                                                }
                                                                else if (AFD >= 52) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                        else if (Rs >= 3326) {
                                                            if (TSC < 13) {
                                                                if (Rs < 3863) {
                                                                    return 2;
                                                                }
                                                                else if (Rs >= 3863) {
                                                                    return 1;
                                                                }
                                                            }
                                                            else if (TSC >= 13) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                            else if (Cs >= 3093) {
                                                if (MVDiff < 17) {
                                                    return 2;
                                                }
                                                else if (MVDiff >= 17) {
                                                    if (ssDCval < 110) {
                                                        return 4;
                                                    }
                                                    else if (ssDCval >= 110) {
                                                        return 1;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (refDCval >= 122) {
                                        if (refDCval < 142) {
                                            if (Cs < 1971) {
                                                if (posBalance < 65) {
                                                    return 1;
                                                }
                                                else if (posBalance >= 65) {
                                                    if (SC < 3564) {
                                                        if (Cs < 1324) {
                                                            return 1;
                                                        }
                                                        else if (Cs >= 1324) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (SC >= 3564) {
                                                        return 2;
                                                    }
                                                }
                                            }
                                            else if (Cs >= 1971) {
                                                if (diffMVdiffVal < 4) {
                                                    return 1;
                                                }
                                                else if (diffMVdiffVal >= 4) {
                                                    if (negBalance < 52) {
                                                        return 1;
                                                    }
                                                    else if (negBalance >= 52) {
                                                        if (Rs < 3146) {
                                                            return 8;
                                                        }
                                                        else if (Rs >= 3146) {
                                                            return 1;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (refDCval >= 142) {
                                            if (refDCval < 145) {
                                                if (Cs < 1363) {
                                                    return 8;
                                                }
                                                else if (Cs >= 1363) {
                                                    return 4;
                                                }
                                            }
                                            else if (refDCval >= 145) {
                                                return 2;
                                            }
                                        }
                                    }
                                }
                                else if (tcor >= 83) {
                                    if (Cs < 2699) {
                                        if (posBalance < 55) {
                                            if (refDCval < 101) {
                                                if (negBalance < 47) {
                                                    if (Cs < 2361) {
                                                        if (Cs < 2348) {
                                                            if (Cs < 1938) {
                                                                if (Rs < 1324) {
                                                                    return 2;
                                                                }
                                                                else if (Rs >= 1324) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (Cs >= 1938) {
                                                                return 1;
                                                            }
                                                        }
                                                        else if (Cs >= 2348) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (Cs >= 2361) {
                                                        return 1;
                                                    }
                                                }
                                                else if (negBalance >= 47) {
                                                    if (RsDiff < 10) {
                                                        return 4;
                                                    }
                                                    else if (RsDiff >= 10) {
                                                        if (Cs < 2386) {
                                                            return 1;
                                                        }
                                                        else if (Cs >= 2386) {
                                                            return 2;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (refDCval >= 101) {
                                                return 4;
                                            }
                                        }
                                        else if (posBalance >= 55) {
                                            if (mcTcor < 94) {
                                                if (mu_mv_mag_sq < 9) {
                                                    if (Cs < 2092) {
                                                        return 2;
                                                    }
                                                    else if (Cs >= 2092) {
                                                        return 4;
                                                    }
                                                }
                                                else if (mu_mv_mag_sq >= 9) {
                                                    if (mcTcor < 92) {
                                                        if (Cs < 2483) {
                                                            return 2;
                                                        }
                                                        else if (Cs >= 2483) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (mcTcor >= 92) {
                                                        if (ssDCval < 89) {
                                                            return 8;
                                                        }
                                                        else if (ssDCval >= 89) {
                                                            return 1;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (mcTcor >= 94) {
                                                return 2;
                                            }
                                        }
                                    }
                                    else if (Cs >= 2699) {
                                        if (ssDCval < 92) {
                                            if (Rs < 2619) {
                                                return 8;
                                            }
                                            else if (Rs >= 2619) {
                                                if (SC < 6284) {
                                                    return 1;
                                                }
                                                else if (SC >= 6284) {
                                                    return 8;
                                                }
                                            }
                                        }
                                        else if (ssDCval >= 92) {
                                            if (refDCval < 129) {
                                                if (diffTSC < 3) {
                                                    if (AFD < 18) {
                                                        return 4;
                                                    }
                                                    else if (AFD >= 18) {
                                                        return 1;
                                                    }
                                                }
                                                else if (diffTSC >= 3) {
                                                    return 8;
                                                }
                                            }
                                            else if (refDCval >= 129) {
                                                return 8;
                                            }
                                        }
                                    }
                                }
                            }
                            else if (ssDCval >= 150) {
                                if (Rs < 2343) {
                                    return 8;
                                }
                                else if (Rs >= 2343) {
                                    if (tcor < 78) {
                                        if (RsCsDiff < 462) {
                                            return 2;
                                        }
                                        else if (RsCsDiff >= 462) {
                                            return 4;
                                        }
                                    }
                                    else if (tcor >= 78) {
                                        return 8;
                                    }
                                }
                            }
                        }
                        else if (tcor >= 90) {
                            if (ssDCval < 122) {
                                if (ssDCval < 102) {
                                    return 8;
                                }
                                else if (ssDCval >= 102) {
                                    if (Cs < 2555) {
                                        return 2;
                                    }
                                    else if (Cs >= 2555) {
                                        return 4;
                                    }
                                }
                            }
                            else if (ssDCval >= 122) {
                                return 8;
                            }
                        }
                    }
                }
                else if (negBalance >= 53) {
                    if (mcTcor < 84) {
                        if (AFD < 18) {
                            if (posBalance < 56) {
                                if (Rs < 661) {
                                    if (TSC < 9) {
                                        if (refDCval < 20) {
                                            if (refDCval < 18) {
                                                return 1;
                                            }
                                            else if (refDCval >= 18) {
                                                return 4;
                                            }
                                        }
                                        else if (refDCval >= 20) {
                                            if (posBalance < 5) {
                                                if (tcor < 75) {
                                                    return 1;
                                                }
                                                else if (tcor >= 75) {
                                                    return 2;
                                                }
                                            }
                                            else if (posBalance >= 5) {
                                                return 1;
                                            }
                                        }
                                    }
                                    else if (TSC >= 9) {
                                        return 2;
                                    }
                                }
                                else if (Rs >= 661) {
                                    if (ssDCval < 105) {
                                        if (negBalance < 55) {
                                            if (Cs < 1954) {
                                                return 1;
                                            }
                                            else if (Cs >= 1954) {
                                                return 8;
                                            }
                                        }
                                        else if (negBalance >= 55) {
                                            return 8;
                                        }
                                    }
                                    else if (ssDCval >= 105) {
                                        if (refDCval < 138) {
                                            return 1;
                                        }
                                        else if (refDCval >= 138) {
                                            return 8;
                                        }
                                    }
                                }
                            }
                            else if (posBalance >= 56) {
                                if (Rs < 2092) {
                                    if (Rs < 1144) {
                                        if (mu_mv_mag_sq < 27) {
                                            return 4;
                                        }
                                        else if (mu_mv_mag_sq >= 27) {
                                            if (Rs < 769) {
                                                if (AFD < 10) {
                                                    return 4;
                                                }
                                                else if (AFD >= 10) {
                                                    return 2;
                                                }
                                            }
                                            else if (Rs >= 769) {
                                                return 1;
                                            }
                                        }
                                    }
                                    else if (Rs >= 1144) {
                                        if (Rs < 1897) {
                                            if (mcTcor < 83) {
                                                return 8;
                                            }
                                            else if (mcTcor >= 83) {
                                                if (MVDiff < 11) {
                                                    return 4;
                                                }
                                                else if (MVDiff >= 11) {
                                                    return 8;
                                                }
                                            }
                                        }
                                        else if (Rs >= 1897) {
                                            if (posBalance < 59) {
                                                return 2;
                                            }
                                            else if (posBalance >= 59) {
                                                return 8;
                                            }
                                        }
                                    }
                                }
                                else if (Rs >= 2092) {
                                    if (ssDCval < 65) {
                                        return 8;
                                    }
                                    else if (ssDCval >= 65) {
                                        if (ssDCval < 129) {
                                            return 4;
                                        }
                                        else if (ssDCval >= 129) {
                                            return 1;
                                        }
                                    }
                                }
                            }
                        }
                        else if (AFD >= 18) {
                            if (CsDiff < 8) {
                                if (RsDiff < 11) {
                                    if (Cs < 1115) {
                                        if (tcor < 24) {
                                            if (Cs < 600) {
                                                return 1;
                                            }
                                            else if (Cs >= 600) {
                                                return 4;
                                            }
                                        }
                                        else if (tcor >= 24) {
                                            if (mcTcor < 75) {
                                                return 1;
                                            }
                                            else if (mcTcor >= 75) {
                                                if (ssDCval < 91) {
                                                    if (posBalance < 20) {
                                                        return 1;
                                                    }
                                                    else if (posBalance >= 20) {
                                                        return 2;
                                                    }
                                                }
                                                else if (ssDCval >= 91) {
                                                    if (diffAFD < 2) {
                                                        return 1;
                                                    }
                                                    else if (diffAFD >= 2) {
                                                        if (RsDiff < 7) {
                                                            return 2;
                                                        }
                                                        else if (RsDiff >= 7) {
                                                            if (AFD < 25) {
                                                                return 2;
                                                            }
                                                            else if (AFD >= 25) {
                                                                return 1;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (Cs >= 1115) {
                                        if (Rs < 1674) {
                                            return 4;
                                        }
                                        else if (Rs >= 1674) {
                                            return 8;
                                        }
                                    }
                                }
                                else if (RsDiff >= 11) {
                                    if (AFD < 31) {
                                        if (Cs < 1122) {
                                            if (Rs < 3037) {
                                                return 2;
                                            }
                                            else if (Rs >= 3037) {
                                                if (tcor < 44) {
                                                    return 2;
                                                }
                                                else if (tcor >= 44) {
                                                    if (Rs < 3524) {
                                                        return 1;
                                                    }
                                                    else if (Rs >= 3524) {
                                                        return 2;
                                                    }
                                                }
                                            }
                                        }
                                        else if (Cs >= 1122) {
                                            return 1;
                                        }
                                    }
                                    else if (AFD >= 31) {
                                        return 2;
                                    }
                                }
                            }
                            else if (CsDiff >= 8) {
                                if (ssDCval < 139) {
                                    if (refDCval < 124) {
                                        if (Scindex < 6) {
                                            if (TSC < 18) {
                                                if (tcor < 74) {
                                                    if (MVDiff < 39) {
                                                        if (RsDiff < 9) {
                                                            return 1;
                                                        }
                                                        else if (RsDiff >= 9) {
                                                            if (ssDCval < 99) {
                                                                return 8;
                                                            }
                                                            else if (ssDCval >= 99) {
                                                                if (gchDC < 1) {
                                                                    return 1;
                                                                }
                                                                else if (gchDC >= 1) {
                                                                    return 2;
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (MVDiff >= 39) {
                                                        if (diffMVdiffVal < 42) {
                                                            return 1;
                                                        }
                                                        else if (diffMVdiffVal >= 42) {
                                                            if (SC < 1908) {
                                                                return 8;
                                                            }
                                                            else if (SC >= 1908) {
                                                                if (Cs < 1359) {
                                                                    return 2;
                                                                }
                                                                else if (Cs >= 1359) {
                                                                    if (SC < 3422) {
                                                                        return 1;
                                                                    }
                                                                    else if (SC >= 3422) {
                                                                        return 2;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (tcor >= 74) {
                                                    return 4;
                                                }
                                            }
                                            else if (TSC >= 18) {
                                                if (gchDC < 4) {
                                                    if (ssDCval < 86) {
                                                        return 4;
                                                    }
                                                    else if (ssDCval >= 86) {
                                                        if (Rs < 1909) {
                                                            return 1;
                                                        }
                                                        else if (Rs >= 1909) {
                                                            return 2;
                                                        }
                                                    }
                                                }
                                                else if (gchDC >= 4) {
                                                    if (diffAFD < 39) {
                                                        if (Rs < 2930) {
                                                            return 8;
                                                        }
                                                        else if (Rs >= 2930) {
                                                            return 1;
                                                        }
                                                    }
                                                    else if (diffAFD >= 39) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                        }
                                        else if (Scindex >= 6) {
                                            if (RsCsDiff < 532) {
                                                if (Cs < 2843) {
                                                    return 2;
                                                }
                                                else if (Cs >= 2843) {
                                                    return 8;
                                                }
                                            }
                                            else if (RsCsDiff >= 532) {
                                                return 2;
                                            }
                                        }
                                    }
                                    else if (refDCval >= 124) {
                                        if (TSC < 14) {
                                            if (ssDCval < 124) {
                                                return 4;
                                            }
                                            else if (ssDCval >= 124) {
                                                return 1;
                                            }
                                        }
                                        else if (TSC >= 14) {
                                            if (mcTcor < 79) {
                                                if (tcor < 28) {
                                                    if (diffRsCsdiff < 436) {
                                                        return 4;
                                                    }
                                                    else if (diffRsCsdiff >= 436) {
                                                        if (Rs < 2089) {
                                                            return 8;
                                                        }
                                                        else if (Rs >= 2089) {
                                                            return 1;
                                                        }
                                                    }
                                                }
                                                else if (tcor >= 28) {
                                                    return 1;
                                                }
                                            }
                                            else if (mcTcor >= 79) {
                                                if (MVDiff < 60) {
                                                    return 4;
                                                }
                                                else if (MVDiff >= 60) {
                                                    if (Cs < 1933) {
                                                        return 8;
                                                    }
                                                    else if (Cs >= 1933) {
                                                        return 2;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (ssDCval >= 139) {
                                    if (Cs < 1454) {
                                        return 8;
                                    }
                                    else if (Cs >= 1454) {
                                        return 4;
                                    }
                                }
                            }
                        }
                    }
                    else if (mcTcor >= 84) {
                        if (AFD < 14) {
                            if (MVDiff < 19) {
                                if (Rs < 1297) {
                                    if (RsDiff < 3) {
                                        if (mu_mv_mag_sq < 77) {
                                            if (mu_mv_mag_sq < 51) {
                                                if (Cs < 575) {
                                                    if (gchDC < 7) {
                                                        return 4;
                                                    }
                                                    else if (gchDC >= 7) {
                                                        return 2;
                                                    }
                                                }
                                                else if (Cs >= 575) {
                                                    if (refDCval < 97) {
                                                        if (mcTcor < 90) {
                                                            if (Rs < 616) {
                                                                return 8;
                                                            }
                                                            else if (Rs >= 616) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (mcTcor >= 90) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (refDCval >= 97) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                            else if (mu_mv_mag_sq >= 51) {
                                                return 8;
                                            }
                                        }
                                        else if (mu_mv_mag_sq >= 77) {
                                            return 4;
                                        }
                                    }
                                    else if (RsDiff >= 3) {
                                        if (tcor < 87) {
                                            if (CsDiff < 4) {
                                                return 4;
                                            }
                                            else if (CsDiff >= 4) {
                                                if (diffMVdiffVal < -2) {
                                                    return 8;
                                                }
                                                else if (diffMVdiffVal >= -2) {
                                                    return 4;
                                                }
                                            }
                                        }
                                        else if (tcor >= 87) {
                                            if (MVDiff < 15) {
                                                if (TSC < 5) {
                                                    return 4;
                                                }
                                                else if (TSC >= 5) {
                                                    return 1;
                                                }
                                            }
                                            else if (MVDiff >= 15) {
                                                return 2;
                                            }
                                        }
                                    }
                                }
                                else if (Rs >= 1297) {
                                    if (ssDCval < 97) {
                                        if (AFD < 12) {
                                            if (ssDCval < 92) {
                                                if (Cs < 1583) {
                                                    if (TSC < 7) {
                                                        return 4;
                                                    }
                                                    else if (TSC >= 7) {
                                                        if (Rs < 1882) {
                                                            return 2;
                                                        }
                                                        else if (Rs >= 1882) {
                                                            return 1;
                                                        }
                                                    }
                                                }
                                                else if (Cs >= 1583) {
                                                    if (RsDiff < 8) {
                                                        return 8;
                                                    }
                                                    else if (RsDiff >= 8) {
                                                        if (Rs < 2518) {
                                                            return 8;
                                                        }
                                                        else if (Rs >= 2518) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (ssDCval >= 92) {
                                                if (RsDiff < 6) {
                                                    return 8;
                                                }
                                                else if (RsDiff >= 6) {
                                                    if (TSC < 9) {
                                                        return 2;
                                                    }
                                                    else if (TSC >= 9) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                        }
                                        else if (AFD >= 12) {
                                            if (Cs < 969) {
                                                if (ssDCval < 68) {
                                                    return 1;
                                                }
                                                else if (ssDCval >= 68) {
                                                    return 2;
                                                }
                                            }
                                            else if (Cs >= 969) {
                                                if (SC < 4836) {
                                                    if (RsCsDiff < 106) {
                                                        return 1;
                                                    }
                                                    else if (RsCsDiff >= 106) {
                                                        return 2;
                                                    }
                                                }
                                                else if (SC >= 4836) {
                                                    return 4;
                                                }
                                            }
                                        }
                                    }
                                    else if (ssDCval >= 97) {
                                        if (Cs < 2229) {
                                            if (Rs < 3667) {
                                                if (ssDCval < 99) {
                                                    if (TSCindex < 8) {
                                                        return 8;
                                                    }
                                                    else if (TSCindex >= 8) {
                                                        return 2;
                                                    }
                                                }
                                                else if (ssDCval >= 99) {
                                                    if (Cs < 1072) {
                                                        return 2;
                                                    }
                                                    else if (Cs >= 1072) {
                                                        if (SC < 5043) {
                                                            if (mu_mv_mag_sq < 46) {
                                                                return 4;
                                                            }
                                                            else if (mu_mv_mag_sq >= 46) {
                                                                if (Cs < 1686) {
                                                                    if (TSC < 8) {
                                                                        return 4;
                                                                    }
                                                                    else if (TSC >= 8) {
                                                                        return 8;
                                                                    }
                                                                }
                                                                else if (Cs >= 1686) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                        else if (SC >= 5043) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (Rs >= 3667) {
                                                return 2;
                                            }
                                        }
                                        else if (Cs >= 2229) {
                                            return 8;
                                        }
                                    }
                                }
                            }
                            else if (MVDiff >= 19) {
                                if (Cs < 316) {
                                    if (TSCindex < 6) {
                                        if (refDCval < 35) {
                                            if (negBalance < 91) {
                                                if (MVDiff < 78) {
                                                    return 4;
                                                }
                                                else if (MVDiff >= 78) {
                                                    return 1;
                                                }
                                            }
                                            else if (negBalance >= 91) {
                                                if (AFD < 4) {
                                                    return 2;
                                                }
                                                else if (AFD >= 4) {
                                                    return 8;
                                                }
                                            }
                                        }
                                        else if (refDCval >= 35) {
                                            if (mu_mv_mag_sq < 15) {
                                                if (AFD < 5) {
                                                    return 2;
                                                }
                                                else if (AFD >= 5) {
                                                    if (SC < 16) {
                                                        return 2;
                                                    }
                                                    else if (SC >= 16) {
                                                        return 1;
                                                    }
                                                }
                                            }
                                            else if (mu_mv_mag_sq >= 15) {
                                                return 2;
                                            }
                                        }
                                    }
                                    else if (TSCindex >= 6) {
                                        if (ssDCval < 106) {
                                            if (ssDCval < 74) {
                                                if (Rs < 287) {
                                                    return 1;
                                                }
                                                else if (Rs >= 287) {
                                                    return 2;
                                                }
                                            }
                                            else if (ssDCval >= 74) {
                                                return 1;
                                            }
                                        }
                                        else if (ssDCval >= 106) {
                                            if (Rs < 592) {
                                                return 1;
                                            }
                                            else if (Rs >= 592) {
                                                return 2;
                                            }
                                        }
                                    }
                                }
                                else if (Cs >= 316) {
                                    if (RsCsDiff < 23) {
                                        if (TSC < 7) {
                                            if (Rs < 339) {
                                                if (AFD < 5) {
                                                    return 4;
                                                }
                                                else if (AFD >= 5) {
                                                    return 2;
                                                }
                                            }
                                            else if (Rs >= 339) {
                                                if (Cs < 942) {
                                                    if (SC < 1106) {
                                                        return 8;
                                                    }
                                                    else if (SC >= 1106) {
                                                        if (diffTSC < 1) {
                                                            if (MVDiff < 104) {
                                                                return 4;
                                                            }
                                                            else if (MVDiff >= 104) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (diffTSC >= 1) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                                else if (Cs >= 942) {
                                                    if (negBalance < 56) {
                                                        return 8;
                                                    }
                                                    else if (negBalance >= 56) {
                                                        if (gchDC < 5) {
                                                            if (Cs < 2793) {
                                                                return 2;
                                                            }
                                                            else if (Cs >= 2793) {
                                                                if (Cs < 3180) {
                                                                    return 8;
                                                                }
                                                                else if (Cs >= 3180) {
                                                                    return 2;
                                                                }
                                                            }
                                                        }
                                                        else if (gchDC >= 5) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (TSC >= 7) {
                                            if (AFD < 11) {
                                                return 2;
                                            }
                                            else if (AFD >= 11) {
                                                return 4;
                                            }
                                        }
                                    }
                                    else if (RsCsDiff >= 23) {
                                        if (Scindex < 5) {
                                            if (SC < 3329) {
                                                if (diffTSC < -5) {
                                                    return 1;
                                                }
                                                else if (diffTSC >= -5) {
                                                    if (CsDiff < 4) {
                                                        if (Cs < 743) {
                                                            return 2;
                                                        }
                                                        else if (Cs >= 743) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (CsDiff >= 4) {
                                                        return 2;
                                                    }
                                                }
                                            }
                                            else if (SC >= 3329) {
                                                if (Rs < 1951) {
                                                    return 8;
                                                }
                                                else if (Rs >= 1951) {
                                                    if (Rs < 2179) {
                                                        return 1;
                                                    }
                                                    else if (Rs >= 2179) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                        }
                                        else if (Scindex >= 5) {
                                            if (CsDiff < 7) {
                                                return 4;
                                            }
                                            else if (CsDiff >= 7) {
                                                if (Rs < 2499) {
                                                    return 1;
                                                }
                                                else if (Rs >= 2499) {
                                                    return 8;
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                        }
                        else if (AFD >= 14) {
                            if (AFD < 17) {
                                if (mu_mv_mag_sq < 42) {
                                    if (diffAFD < 2) {
                                        if (RsCsDiff < 87) {
                                            if (gchDC < 2) {
                                                if (diffTSC < -1) {
                                                    if (Rs < 1497) {
                                                        return 8;
                                                    }
                                                    else if (Rs >= 1497) {
                                                        return 2;
                                                    }
                                                }
                                                else if (diffTSC >= -1) {
                                                    if (diffRsCsdiff < -32) {
                                                        return 2;
                                                    }
                                                    else if (diffRsCsdiff >= -32) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                            else if (gchDC >= 2) {
                                                if (mu_mv_mag_sq < 18) {
                                                    return 2;
                                                }
                                                else if (mu_mv_mag_sq >= 18) {
                                                    if (AFD < 15) {
                                                        return 8;
                                                    }
                                                    else if (AFD >= 15) {
                                                        if (Cs < 346) {
                                                            return 2;
                                                        }
                                                        else if (Cs >= 346) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (RsCsDiff >= 87) {
                                            if (Cs < 1437) {
                                                if (RsCsDiff < 105) {
                                                    if (Rs < 2242) {
                                                        if (refDCval < 84) {
                                                            return 4;
                                                        }
                                                        else if (refDCval >= 84) {
                                                            return 2;
                                                        }
                                                    }
                                                    else if (Rs >= 2242) {
                                                        return 4;
                                                    }
                                                }
                                                else if (RsCsDiff >= 105) {
                                                    return 2;
                                                }
                                            }
                                            else if (Cs >= 1437) {
                                                if (ssDCval < 117) {
                                                    if (Rs < 1902) {
                                                        return 4;
                                                    }
                                                    else if (Rs >= 1902) {
                                                        if (Rs < 2844) {
                                                            return 2;
                                                        }
                                                        else if (Rs >= 2844) {
                                                            if (Cs < 2059) {
                                                                return 8;
                                                            }
                                                            else if (Cs >= 2059) {
                                                                return 1;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (ssDCval >= 117) {
                                                    if (mcTcor < 92) {
                                                        return 4;
                                                    }
                                                    else if (mcTcor >= 92) {
                                                        if (SC < 3773) {
                                                            return 4;
                                                        }
                                                        else if (SC >= 3773) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (diffAFD >= 2) {
                                        if (mu_mv_mag_sq < 13) {
                                            if (ssDCval < 98) {
                                                return 2;
                                            }
                                            else if (ssDCval >= 98) {
                                                return 4;
                                            }
                                        }
                                        else if (mu_mv_mag_sq >= 13) {
                                            if (RsDiff < 17) {
                                                if (SC < 5700) {
                                                    return 8;
                                                }
                                                else if (SC >= 5700) {
                                                    if (AFD < 16) {
                                                        return 8;
                                                    }
                                                    else if (AFD >= 16) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                            else if (RsDiff >= 17) {
                                                return 4;
                                            }
                                        }
                                    }
                                }
                                else if (mu_mv_mag_sq >= 42) {
                                    if (TSC < 9) {
                                        return 2;
                                    }
                                    else if (TSC >= 9) {
                                        if (SC < 3393) {
                                            if (RsCsDiff < 67) {
                                                return 8;
                                            }
                                            else if (RsCsDiff >= 67) {
                                                return 4;
                                            }
                                        }
                                        else if (SC >= 3393) {
                                            if (RsDiff < 9) {
                                                if (Cs < 1838) {
                                                    return 8;
                                                }
                                                else if (Cs >= 1838) {
                                                    return 1;
                                                }
                                            }
                                            else if (RsDiff >= 9) {
                                                if (diffTSC < 1) {
                                                    return 8;
                                                }
                                                else if (diffTSC >= 1) {
                                                    return 4;
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                            else if (AFD >= 17) {
                                if (RsDiff < 20) {
                                    if (MVDiff < 39) {
                                        if (mcTcor < 94) {
                                            if (MVDiff < 28) {
                                                if (refDCval < 80) {
                                                    if (TSC < 14) {
                                                        if (posBalance < 59) {
                                                            if (SC < 2733) {
                                                                return 2;
                                                            }
                                                            else if (SC >= 2733) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (posBalance >= 59) {
                                                            return 2;
                                                        }
                                                    }
                                                    else if (TSC >= 14) {
                                                        return 8;
                                                    }
                                                }
                                                else if (refDCval >= 80) {
                                                    if (refDCval < 138) {
                                                        if (Cs < 3028) {
                                                            if (ssDCval < 110) {
                                                                if (TSC < 13) {
                                                                    if (Cs < 2268) {
                                                                        if (refDCval < 82) {
                                                                            return 1;
                                                                        }
                                                                        else if (refDCval >= 82) {
                                                                            return 2;
                                                                        }
                                                                    }
                                                                    else if (Cs >= 2268) {
                                                                        if (refDCval < 90) {
                                                                            return 2;
                                                                        }
                                                                        else if (refDCval >= 90) {
                                                                            if (CsDiff < 11) {
                                                                                return 1;
                                                                            }
                                                                            else if (CsDiff >= 11) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                                else if (TSC >= 13) {
                                                                    return 1;
                                                                }
                                                            }
                                                            else if (ssDCval >= 110) {
                                                                if (mcTcor < 93) {
                                                                    if (MVDiff < 21) {
                                                                        return 1;
                                                                    }
                                                                    else if (MVDiff >= 21) {
                                                                        if (ssDCval < 123) {
                                                                            if (Cs < 1743) {
                                                                                return 4;
                                                                            }
                                                                            else if (Cs >= 1743) {
                                                                                return 1;
                                                                            }
                                                                        }
                                                                        else if (ssDCval >= 123) {
                                                                            if (tcor < 82) {
                                                                                return 4;
                                                                            }
                                                                            else if (tcor >= 82) {
                                                                                return 2;
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                                else if (mcTcor >= 93) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                        else if (Cs >= 3028) {
                                                            if (Rs < 4000) {
                                                                if (TSC < 16) {
                                                                    return 8;
                                                                }
                                                                else if (TSC >= 16) {
                                                                    return 1;
                                                                }
                                                            }
                                                            else if (Rs >= 4000) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                    else if (refDCval >= 138) {
                                                        if (RsDiff < 17) {
                                                            if (AFD < 28) {
                                                                return 2;
                                                            }
                                                            else if (AFD >= 28) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (RsDiff >= 17) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (MVDiff >= 28) {
                                                if (Cs < 1810) {
                                                    if (MVDiff < 33) {
                                                        if (diffRsCsdiff < 40) {
                                                            if (AFD < 19) {
                                                                if (AFD < 18) {
                                                                    return 2;
                                                                }
                                                                else if (AFD >= 18) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (AFD >= 19) {
                                                                return 2;
                                                            }
                                                        }
                                                        else if (diffRsCsdiff >= 40) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (MVDiff >= 33) {
                                                        if (Cs < 1196) {
                                                            if (RsCsDiff < 228) {
                                                                return 2;
                                                            }
                                                            else if (RsCsDiff >= 228) {
                                                                return 1;
                                                            }
                                                        }
                                                        else if (Cs >= 1196) {
                                                            return 2;
                                                        }
                                                    }
                                                }
                                                else if (Cs >= 1810) {
                                                    if (ssDCval < 123) {
                                                        if (diffAFD < 3) {
                                                            if (Cs < 2491) {
                                                                return 8;
                                                            }
                                                            else if (Cs >= 2491) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (diffAFD >= 3) {
                                                            return 1;
                                                        }
                                                    }
                                                    else if (ssDCval >= 123) {
                                                        if (RsDiff < 14) {
                                                            return 4;
                                                        }
                                                        else if (RsDiff >= 14) {
                                                            if (Cs < 2355) {
                                                                return 2;
                                                            }
                                                            else if (Cs >= 2355) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (mcTcor >= 94) {
                                            if (TSC < 5) {
                                                if (SC < 1719) {
                                                    return 2;
                                                }
                                                else if (SC >= 1719) {
                                                    return 8;
                                                }
                                            }
                                            else if (TSC >= 5) {
                                                return 2;
                                            }
                                        }
                                    }
                                    else if (MVDiff >= 39) {
                                        if (diffMVdiffVal < 187) {
                                            if (SC < 4731) {
                                                if (posBalance < 18) {
                                                    if (Cs < 348) {
                                                        return 1;
                                                    }
                                                    else if (Cs >= 348) {
                                                        if (TSC < 12) {
                                                            return 8;
                                                        }
                                                        else if (TSC >= 12) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                                else if (posBalance >= 18) {
                                                    if (ssDCval < 105) {
                                                        if (AFD < 34) {
                                                            if (TSC < 10) {
                                                                if (diffMVdiffVal < 9) {
                                                                    return 1;
                                                                }
                                                                else if (diffMVdiffVal >= 9) {
                                                                    return 2;
                                                                }
                                                            }
                                                            else if (TSC >= 10) {
                                                                return 1;
                                                            }
                                                        }
                                                        else if (AFD >= 34) {
                                                            return 2;
                                                        }
                                                    }
                                                    else if (ssDCval >= 105) {
                                                        if (Cs < 1926) {
                                                            if (refDCval < 135) {
                                                                if (AFD < 19) {
                                                                    return 4;
                                                                }
                                                                else if (AFD >= 19) {
                                                                    if (SC < 3895) {
                                                                        return 2;
                                                                    }
                                                                    else if (SC >= 3895) {
                                                                        if (TSC < 13) {
                                                                            return 4;
                                                                        }
                                                                        else if (TSC >= 13) {
                                                                            return 2;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                            else if (refDCval >= 135) {
                                                                if (Cs < 366) {
                                                                    return 2;
                                                                }
                                                                else if (Cs >= 366) {
                                                                    return 1;
                                                                }
                                                            }
                                                        }
                                                        else if (Cs >= 1926) {
                                                            return 1;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (SC >= 4731) {
                                                if (RsDiff < 18) {
                                                    return 8;
                                                }
                                                else if (RsDiff >= 18) {
                                                    if (Rs < 3941) {
                                                        return 1;
                                                    }
                                                    else if (Rs >= 3941) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                        }
                                        else if (diffMVdiffVal >= 187) {
                                            if (CsDiff < 2) {
                                                return 2;
                                            }
                                            else if (CsDiff >= 2) {
                                                return 8;
                                            }
                                        }
                                    }
                                }
                                else if (RsDiff >= 20) {
                                    if (Cs < 602) {
                                        return 8;
                                    }
                                    else if (Cs >= 602) {
                                        if (Rs < 4714) {
                                            return 2;
                                        }
                                        else if (Rs >= 4714) {
                                            return 4;
                                        }
                                    }
                                }
                            }
                        }
                    }
                }
            }
            else if (Cs >= 3274) {
                if (mu_mv_mag_sq < 63) {
                    if (ssDCval < 76) {
                        if (RsDiff < 5) {
                            if (MVDiff < 17) {
                                if (mcTcor < 93) {
                                    return 2;
                                }
                                else if (mcTcor >= 93) {
                                    if (ssDCval < 55) {
                                        if (Cs < 3734) {
                                            if (diffAFD < -2) {
                                                if (Cs < 3644) {
                                                    return 4;
                                                }
                                                else if (Cs >= 3644) {
                                                    return 2;
                                                }
                                            }
                                            else if (diffAFD >= -2) {
                                                return 4;
                                            }
                                        }
                                        else if (Cs >= 3734) {
                                            return 2;
                                        }
                                    }
                                    else if (ssDCval >= 55) {
                                        if (Cs < 3571) {
                                            if (diffRsCsdiff < 9) {
                                                return 8;
                                            }
                                            else if (diffRsCsdiff >= 9) {
                                                if (Cs < 3517) {
                                                    return 4;
                                                }
                                                else if (Cs >= 3517) {
                                                    return 8;
                                                }
                                            }
                                        }
                                        else if (Cs >= 3571) {
                                            if (MVDiff < 14) {
                                                return 4;
                                            }
                                            else if (MVDiff >= 14) {
                                                return 8;
                                            }
                                        }
                                    }
                                }
                            }
                            else if (MVDiff >= 17) {
                                if (TSC < 7) {
                                    if (AFD < 8) {
                                        return 4;
                                    }
                                    else if (AFD >= 8) {
                                        if (tcor < 93) {
                                            return 8;
                                        }
                                        else if (tcor >= 93) {
                                            if (ssDCval < 51) {
                                                return 4;
                                            }
                                            else if (ssDCval >= 51) {
                                                return 8;
                                            }
                                        }
                                    }
                                }
                                else if (TSC >= 7) {
                                    if (diffRsCsdiff < -18) {
                                        if (AFD < 11) {
                                            return 4;
                                        }
                                        else if (AFD >= 11) {
                                            return 8;
                                        }
                                    }
                                    else if (diffRsCsdiff >= -18) {
                                        if (Rs < 3515) {
                                            if (negBalance < 49) {
                                                return 8;
                                            }
                                            else if (negBalance >= 49) {
                                                if (ssDCval < 51) {
                                                    return 8;
                                                }
                                                else if (ssDCval >= 51) {
                                                    if (Cs < 3443) {
                                                        return 2;
                                                    }
                                                    else if (Cs >= 3443) {
                                                        if (Rs < 3431) {
                                                            return 2;
                                                        }
                                                        else if (Rs >= 3431) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (Rs >= 3515) {
                                            if (ssDCval < 58) {
                                                if (posBalance < 67) {
                                                    if (tcor < 86) {
                                                        return 8;
                                                    }
                                                    else if (tcor >= 86) {
                                                        if (SC < 7267) {
                                                            if (Rs < 3662) {
                                                                return 2;
                                                            }
                                                            else if (Rs >= 3662) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (SC >= 7267) {
                                                            return 2;
                                                        }
                                                    }
                                                }
                                                else if (posBalance >= 67) {
                                                    return 8;
                                                }
                                            }
                                            else if (ssDCval >= 58) {
                                                if (diffMVdiffVal < 7) {
                                                    return 8;
                                                }
                                                else if (diffMVdiffVal >= 7) {
                                                    return 4;
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                        }
                        else if (RsDiff >= 5) {
                            if (mcTcor < 80) {
                                if (Rs < 2593) {
                                    return 4;
                                }
                                else if (Rs >= 2593) {
                                    if (AFD < 23) {
                                        return 8;
                                    }
                                    else if (AFD >= 23) {
                                        if (Rs < 5719) {
                                            return 2;
                                        }
                                        else if (Rs >= 5719) {
                                            if (SC < 12441) {
                                                return 8;
                                            }
                                            else if (SC >= 12441) {
                                                return 4;
                                            }
                                        }
                                    }
                                }
                            }
                            else if (mcTcor >= 80) {
                                if (mcTcor < 93) {
                                    if (refDCval < 53) {
                                        return 8;
                                    }
                                    else if (refDCval >= 53) {
                                        if (Rs < 3554) {
                                            return 8;
                                        }
                                        else if (Rs >= 3554) {
                                            if (refDCval < 69) {
                                                if (TSC < 13) {
                                                    return 4;
                                                }
                                                else if (TSC >= 13) {
                                                    if (Rs < 6127) {
                                                        return 8;
                                                    }
                                                    else if (Rs >= 6127) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                            else if (refDCval >= 69) {
                                                if (Rs < 6045) {
                                                    return 8;
                                                }
                                                else if (Rs >= 6045) {
                                                    if (Cs < 7178) {
                                                        return 4;
                                                    }
                                                    else if (Cs >= 7178) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (mcTcor >= 93) {
                                    if (MVDiff < 15) {
                                        if (SC < 11481) {
                                            if (gchDC < 1) {
                                                if (diffRsCsdiff < 12) {
                                                    return 4;
                                                }
                                                else if (diffRsCsdiff >= 12) {
                                                    return 8;
                                                }
                                            }
                                            else if (gchDC >= 1) {
                                                return 4;
                                            }
                                        }
                                        else if (SC >= 11481) {
                                            return 8;
                                        }
                                    }
                                    else if (MVDiff >= 15) {
                                        if (negBalance < 44) {
                                            return 4;
                                        }
                                        else if (negBalance >= 44) {
                                            if (diffRsCsdiff < -53) {
                                                return 4;
                                            }
                                            else if (diffRsCsdiff >= -53) {
                                                if (AFD < 10) {
                                                    if (Rs < 3960) {
                                                        return 2;
                                                    }
                                                    else if (Rs >= 3960) {
                                                        return 8;
                                                    }
                                                }
                                                else if (AFD >= 10) {
                                                    return 8;
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                        }
                    }
                    else if (ssDCval >= 76) {
                        if (SC < 7813) {
                            if (ssDCval < 135) {
                                if (SC < 5659) {
                                    return 4;
                                }
                                else if (SC >= 5659) {
                                    if (RsDiff < 21) {
                                        if (diffAFD < -1) {
                                            if (Cs < 3324) {
                                                return 1;
                                            }
                                            else if (Cs >= 3324) {
                                                return 8;
                                            }
                                        }
                                        else if (diffAFD >= -1) {
                                            if (Cs < 3537) {
                                                return 1;
                                            }
                                            else if (Cs >= 3537) {
                                                if (Rs < 2906) {
                                                    return 1;
                                                }
                                                else if (Rs >= 2906) {
                                                    if (MVDiff < 11) {
                                                        return 8;
                                                    }
                                                    else if (MVDiff >= 11) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (RsDiff >= 21) {
                                        return 2;
                                    }
                                }
                            }
                            else if (ssDCval >= 135) {
                                return 1;
                            }
                        }
                        else if (SC >= 7813) {
                            if (refDCval < 153) {
                                if (Rs < 6759) {
                                    if (mcTcor < 93) {
                                        if (Cs < 4329) {
                                            if (mcTcor < 72) {
                                                if (diffRsCsdiff < -96) {
                                                    if (MVDiff < 47) {
                                                        return 4;
                                                    }
                                                    else if (MVDiff >= 47) {
                                                        if (SC < 8841) {
                                                            return 8;
                                                        }
                                                        else if (SC >= 8841) {
                                                            return 2;
                                                        }
                                                    }
                                                }
                                                else if (diffRsCsdiff >= -96) {
                                                    return 1;
                                                }
                                            }
                                            else if (mcTcor >= 72) {
                                                if (ssDCval < 121) {
                                                    if (AFD < 26) {
                                                        if (mu_mv_mag_sq < 42) {
                                                            return 4;
                                                        }
                                                        else if (mu_mv_mag_sq >= 42) {
                                                            if (SC < 8523) {
                                                                return 4;
                                                            }
                                                            else if (SC >= 8523) {
                                                                return 2;
                                                            }
                                                        }
                                                    }
                                                    else if (AFD >= 26) {
                                                        if (SC < 7959) {
                                                            return 2;
                                                        }
                                                        else if (SC >= 7959) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                                else if (ssDCval >= 121) {
                                                    if (RsCsDiff < 621) {
                                                        if (diffAFD < -8) {
                                                            if (Rs < 5032) {
                                                                return 8;
                                                            }
                                                            else if (Rs >= 5032) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (diffAFD >= -8) {
                                                            if (ssDCval < 122) {
                                                                return 4;
                                                            }
                                                            else if (ssDCval >= 122) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                    else if (RsCsDiff >= 621) {
                                                        if (AFD < 25) {
                                                            if (MVDiff < 12) {
                                                                return 8;
                                                            }
                                                            else if (MVDiff >= 12) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (AFD >= 25) {
                                                            if (Rs < 5151) {
                                                                return 8;
                                                            }
                                                            else if (Rs >= 5151) {
                                                                if (Rs < 6003) {
                                                                    return 4;
                                                                }
                                                                else if (Rs >= 6003) {
                                                                    return 2;
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (Cs >= 4329) {
                                            if (RsDiff < 30) {
                                                if (RsDiff < 15) {
                                                    return 1;
                                                }
                                                else if (RsDiff >= 15) {
                                                    if (ssDCval < 138) {
                                                        return 8;
                                                    }
                                                    else if (ssDCval >= 138) {
                                                        if (posBalance < 54) {
                                                            if (AFD < 25) {
                                                                return 4;
                                                            }
                                                            else if (AFD >= 25) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (posBalance >= 54) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (RsDiff >= 30) {
                                                if (SC < 11312) {
                                                    return 2;
                                                }
                                                else if (SC >= 11312) {
                                                    return 8;
                                                }
                                            }
                                        }
                                    }
                                    else if (mcTcor >= 93) {
                                        if (Rs < 5766) {
                                            return 8;
                                        }
                                        else if (Rs >= 5766) {
                                            if (SC < 10087) {
                                                return 4;
                                            }
                                            else if (SC >= 10087) {
                                                return 1;
                                            }
                                        }
                                    }
                                }
                                else if (Rs >= 6759) {
                                    if (MVDiff < 26) {
                                        if (Rs < 15123) {
                                            if (CsDiff < 31) {
                                                if (tcor < 81) {
                                                    if (gchDC < 3) {
                                                        if (mu_mv_mag_sq < 18) {
                                                            return 8;
                                                        }
                                                        else if (mu_mv_mag_sq >= 18) {
                                                            if (mu_mv_mag_sq < 42) {
                                                                if (RsCsDiff < 1577) {
                                                                    if (RsDiff < 27) {
                                                                        return 4;
                                                                    }
                                                                    else if (RsDiff >= 27) {
                                                                        return 2;
                                                                    }
                                                                }
                                                                else if (RsCsDiff >= 1577) {
                                                                    if (CsDiff < 26) {
                                                                        return 8;
                                                                    }
                                                                    else if (CsDiff >= 26) {
                                                                        return 4;
                                                                    }
                                                                }
                                                            }
                                                            else if (mu_mv_mag_sq >= 42) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                    else if (gchDC >= 3) {
                                                        if (refDCval < 137) {
                                                            if (AFD < 28) {
                                                                return 4;
                                                            }
                                                            else if (AFD >= 28) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (refDCval >= 137) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                                else if (tcor >= 81) {
                                                    if (Cs < 6019) {
                                                        if (CsDiff < 27) {
                                                            return 4;
                                                        }
                                                        else if (CsDiff >= 27) {
                                                            return 2;
                                                        }
                                                    }
                                                    else if (Cs >= 6019) {
                                                        if (Rs < 8307) {
                                                            return 8;
                                                        }
                                                        else if (Rs >= 8307) {
                                                            if (Rs < 8783) {
                                                                return 2;
                                                            }
                                                            else if (Rs >= 8783) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                            else if (CsDiff >= 31) {
                                                if (SC < 14616) {
                                                    return 8;
                                                }
                                                else if (SC >= 14616) {
                                                    if (refDCval < 138) {
                                                        return 4;
                                                    }
                                                    else if (refDCval >= 138) {
                                                        if (Rs < 11810) {
                                                            return 4;
                                                        }
                                                        else if (Rs >= 11810) {
                                                            return 2;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (Rs >= 15123) {
                                            if (TSC < 30) {
                                                return 8;
                                            }
                                            else if (TSC >= 30) {
                                                return 4;
                                            }
                                        }
                                    }
                                    else if (MVDiff >= 26) {
                                        if (mcTcor < 87) {
                                            return 8;
                                        }
                                        else if (mcTcor >= 87) {
                                            return 2;
                                        }
                                    }
                                }
                            }
                            else if (refDCval >= 153) {
                                if (negBalance < 54) {
                                    if (Cs < 4506) {
                                        return 4;
                                    }
                                    else if (Cs >= 4506) {
                                        if (AFD < 39) {
                                            return 8;
                                        }
                                        else if (AFD >= 39) {
                                            return 2;
                                        }
                                    }
                                }
                                else if (negBalance >= 54) {
                                    if (diffAFD < 2) {
                                        if (Cs < 3821) {
                                            if (ssDCval < 159) {
                                                return 4;
                                            }
                                            else if (ssDCval >= 159) {
                                                return 2;
                                            }
                                        }
                                        else if (Cs >= 3821) {
                                            return 2;
                                        }
                                    }
                                    else if (diffAFD >= 2) {
                                        if (AFD < 26) {
                                            return 8;
                                        }
                                        else if (AFD >= 26) {
                                            if (Rs < 7779) {
                                                return 2;
                                            }
                                            else if (Rs >= 7779) {
                                                return 8;
                                            }
                                        }
                                    }
                                }
                            }
                        }
                    }
                }
                else if (mu_mv_mag_sq >= 63) {
                    if (refDCval < 66) {
                        if (TSCindex < 8) {
                            if (refDCval < 54) {
                                if (posBalance < 55) {
                                    if (AFD < 4) {
                                        return 8;
                                    }
                                    else if (AFD >= 4) {
                                        return 4;
                                    }
                                }
                                else if (posBalance >= 55) {
                                    if (Cs < 3574) {
                                        return 8;
                                    }
                                    else if (Cs >= 3574) {
                                        return 4;
                                    }
                                }
                            }
                            else if (refDCval >= 54) {
                                if (diffTSC < -1) {
                                    if (ssDCval < 56) {
                                        return 8;
                                    }
                                    else if (ssDCval >= 56) {
                                        return 4;
                                    }
                                }
                                else if (diffTSC >= -1) {
                                    if (MVDiff < 7) {
                                        if (posBalance < 55) {
                                            return 8;
                                        }
                                        else if (posBalance >= 55) {
                                            return 4;
                                        }
                                    }
                                    else if (MVDiff >= 7) {
                                        return 8;
                                    }
                                }
                            }
                        }
                        else if (TSCindex >= 8) {
                            if (negBalance < 50) {
                                return 4;
                            }
                            else if (negBalance >= 50) {
                                return 8;
                            }
                        }
                    }
                    else if (refDCval >= 66) {
                        if (Cs < 4122) {
                            if (mcTcor < 82) {
                                if (AFD < 16) {
                                    if (negBalance < 43) {
                                        return 8;
                                    }
                                    else if (negBalance >= 43) {
                                        return 4;
                                    }
                                }
                                else if (AFD >= 16) {
                                    return 8;
                                }
                            }
                            else if (mcTcor >= 82) {
                                if (diffMVdiffVal < -10) {
                                    return 4;
                                }
                                else if (diffMVdiffVal >= -10) {
                                    if (TSC < 11) {
                                        if (CsDiff < 11) {
                                            return 8;
                                        }
                                        else if (CsDiff >= 11) {
                                            return 2;
                                        }
                                    }
                                    else if (TSC >= 11) {
                                        return 8;
                                    }
                                }
                            }
                        }
                        else if (Cs >= 4122) {
                            if (posBalance < 60) {
                                if (mcTcor < 75) {
                                    if (RsDiff < 16) {
                                        return 1;
                                    }
                                    else if (RsDiff >= 16) {
                                        return 8;
                                    }
                                }
                                else if (mcTcor >= 75) {
                                    if (ssDCval < 72) {
                                        return 4;
                                    }
                                    else if (ssDCval >= 72) {
                                        return 8;
                                    }
                                }
                            }
                            else if (posBalance >= 60) {
                                return 8;
                            }
                        }
                    }
                }
            }
        }
        return 4;
    }
    mfxU16 AGOPSelect10(mfxU32 Rs, mfxU32 CsDiff, mfxU32 TSC, mfxI32 diffAFD, mfxU32 Scindex, mfxU32 RsDiff, mfxI32 diffMVdiffVal, mfxU32 RsCsDiff, mfxU32 TSCindex, mfxI16 tcor, mfxU32 refDCval, mfxU32 MVDiff, mfxU32 AFD, mfxI32 diffTSC, mfxI32 diffRsCsdiff, mfxU32 posBalance, mfxU32 SC, mfxU32 mu_mv_mag_sq, mfxU32 negBalance, mfxI16 mcTcor, mfxU32 gchDC, mfxU32 Cs, mfxU32 ssDCval) {
        if (mu_mv_mag_sq < 88) {
            if (ssDCval < 17) {
                if (Cs < 1787) {
                    if (ssDCval < 8) {
                        return 8;
                    }
                    else if (ssDCval >= 8) {
                        if (AFD < 18) {
                            if (MVDiff < 363) {
                                return 1;
                            }
                            else if (MVDiff >= 363) {
                                return 4;
                            }
                        }
                        else if (AFD >= 18) {
                            return 8;
                        }
                    }
                }
                else if (Cs >= 1787) {
                    if (Rs < 4572) {
                        if (Rs < 2707) {
                            return 1;
                        }
                        else if (Rs >= 2707) {
                            if (SC < 7755) {
                                if (Rs < 3755) {
                                    return 8;
                                }
                                else if (Rs >= 3755) {
                                    return 1;
                                }
                            }
                            else if (SC >= 7755) {
                                return 8;
                            }
                        }
                    }
                    else if (Rs >= 4572) {
                        if (Rs < 5376) {
                            return 1;
                        }
                        else if (Rs >= 5376) {
                            return 8;
                        }
                    }
                }
            }
            else if (ssDCval >= 17) {
                if (Cs < 3269) {
                    if (AFD < 12) {
                        if (ssDCval < 97) {
                            if (posBalance < 14) {
                                if (ssDCval < 58) {
                                    if (MVDiff < 29) {
                                        if (Cs < 754) {
                                            if (refDCval < 39) {
                                                if (SC < 291) {
                                                    return 4;
                                                }
                                                else if (SC >= 291) {
                                                    if (Cs < 163) {
                                                        return 4;
                                                    }
                                                    else if (Cs >= 163) {
                                                        return 2;
                                                    }
                                                }
                                            }
                                            else if (refDCval >= 39) {
                                                if (negBalance < 93) {
                                                    if (diffTSC < -1) {
                                                        if (Cs < 417) {
                                                            return 2;
                                                        }
                                                        else if (Cs >= 417) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (diffTSC >= -1) {
                                                        return 4;
                                                    }
                                                }
                                                else if (negBalance >= 93) {
                                                    if (AFD < 8) {
                                                        return 8;
                                                    }
                                                    else if (AFD >= 8) {
                                                        return 2;
                                                    }
                                                }
                                            }
                                        }
                                        else if (Cs >= 754) {
                                            if (AFD < 1) {
                                                return 1;
                                            }
                                            else if (AFD >= 1) {
                                                return 8;
                                            }
                                        }
                                    }
                                    else if (MVDiff >= 29) {
                                        if (AFD < 1) {
                                            if (negBalance < 11) {
                                                if (tcor < 65) {
                                                    return 4;
                                                }
                                                else if (tcor >= 65) {
                                                    return 8;
                                                }
                                            }
                                            else if (negBalance >= 11) {
                                                if (Cs < 6) {
                                                    if (MVDiff < 85) {
                                                        if (diffAFD < -1) {
                                                            return 8;
                                                        }
                                                        else if (diffAFD >= -1) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (MVDiff >= 85) {
                                                        if (refDCval < 20) {
                                                            return 8;
                                                        }
                                                        else if (refDCval >= 20) {
                                                            return 2;
                                                        }
                                                    }
                                                }
                                                else if (Cs >= 6) {
                                                    return 8;
                                                }
                                            }
                                        }
                                        else if (AFD >= 1) {
                                            if (diffAFD < -1) {
                                                if (MVDiff < 137) {
                                                    return 2;
                                                }
                                                else if (MVDiff >= 137) {
                                                    return 1;
                                                }
                                            }
                                            else if (diffAFD >= -1) {
                                                if (negBalance < 90) {
                                                    if (MVDiff < 49) {
                                                        if (tcor < 112) {
                                                            return 8;
                                                        }
                                                        else if (tcor >= 112) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (MVDiff >= 49) {
                                                        if (negBalance < 33) {
                                                            return 8;
                                                        }
                                                        else if (negBalance >= 33) {
                                                            if (diffMVdiffVal < -20) {
                                                                return 1;
                                                            }
                                                            else if (diffMVdiffVal >= -20) {
                                                                if (refDCval < 25) {
                                                                    if (diffMVdiffVal < 450) {
                                                                        return 2;
                                                                    }
                                                                    else if (diffMVdiffVal >= 450) {
                                                                        if (MVDiff < 974) {
                                                                            return 4;
                                                                        }
                                                                        else if (MVDiff >= 974) {
                                                                            return 2;
                                                                        }
                                                                    }
                                                                }
                                                                else if (refDCval >= 25) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (negBalance >= 90) {
                                                    if (tcor < 98) {
                                                        if (negBalance < 117) {
                                                            return 1;
                                                        }
                                                        else if (negBalance >= 117) {
                                                            return 2;
                                                        }
                                                    }
                                                    else if (tcor >= 98) {
                                                        if (Cs < 618) {
                                                            return 8;
                                                        }
                                                        else if (Cs >= 618) {
                                                            return 2;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (ssDCval >= 58) {
                                    if (Cs < 878) {
                                        return 8;
                                    }
                                    else if (Cs >= 878) {
                                        if (TSC < 6) {
                                            return 4;
                                        }
                                        else if (TSC >= 6) {
                                            return 8;
                                        }
                                    }
                                }
                            }
                            else if (posBalance >= 14) {
                                if (AFD < 9) {
                                    if (Cs < 544) {
                                        if (AFD < 7) {
                                            if (Cs < 382) {
                                                if (refDCval < 63) {
                                                    if (Rs < 403) {
                                                        if (MVDiff < 43) {
                                                            if (mu_mv_mag_sq < 29) {
                                                                if (MVDiff < 24) {
                                                                    if (Cs < 2) {
                                                                        return 4;
                                                                    }
                                                                    else if (Cs >= 2) {
                                                                        if (SC < 164) {
                                                                            return 2;
                                                                        }
                                                                        else if (SC >= 164) {
                                                                            if (Rs < 186) {
                                                                                return 8;
                                                                            }
                                                                            else if (Rs >= 186) {
                                                                                if (ssDCval < 48) {
                                                                                    return 4;
                                                                                }
                                                                                else if (ssDCval >= 48) {
                                                                                    return 2;
                                                                                }
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                                else if (MVDiff >= 24) {
                                                                    if (Rs < 331) {
                                                                        if (mcTcor < 62) {
                                                                            return 1;
                                                                        }
                                                                        else if (mcTcor >= 62) {
                                                                            return 2;
                                                                        }
                                                                    }
                                                                    else if (Rs >= 331) {
                                                                        return 8;
                                                                    }
                                                                }
                                                            }
                                                            else if (mu_mv_mag_sq >= 29) {
                                                                if (Rs < 11) {
                                                                    return 4;
                                                                }
                                                                else if (Rs >= 11) {
                                                                    if (mcTcor < 92) {
                                                                        if (tcor < 85) {
                                                                            if (TSC < 5) {
                                                                                if (refDCval < 31) {
                                                                                    return 4;
                                                                                }
                                                                                else if (refDCval >= 31) {
                                                                                    if (Rs < 315) {
                                                                                        if (diffMVdiffVal < 5) {
                                                                                            return 8;
                                                                                        }
                                                                                        else if (diffMVdiffVal >= 5) {
                                                                                            return 4;
                                                                                        }
                                                                                    }
                                                                                    else if (Rs >= 315) {
                                                                                        if (gchDC < 1) {
                                                                                            if (MVDiff < 21) {
                                                                                                if (Rs < 389) {
                                                                                                    return 4;
                                                                                                }
                                                                                                else if (Rs >= 389) {
                                                                                                    return 1;
                                                                                                }
                                                                                            }
                                                                                            else if (MVDiff >= 21) {
                                                                                                if (tcor < 81) {
                                                                                                    return 8;
                                                                                                }
                                                                                                else if (tcor >= 81) {
                                                                                                    if (Cs < 294) {
                                                                                                        return 4;
                                                                                                    }
                                                                                                    else if (Cs >= 294) {
                                                                                                        return 8;
                                                                                                    }
                                                                                                }
                                                                                            }
                                                                                        }
                                                                                        else if (gchDC >= 1) {
                                                                                            return 4;
                                                                                        }
                                                                                    }
                                                                                }
                                                                            }
                                                                            else if (TSC >= 5) {
                                                                                return 1;
                                                                            }
                                                                        }
                                                                        else if (tcor >= 85) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                    else if (mcTcor >= 92) {
                                                                        if (SC < 413) {
                                                                            return 2;
                                                                        }
                                                                        else if (SC >= 413) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (MVDiff >= 43) {
                                                            if (negBalance < 4) {
                                                                if (mcTcor < 49) {
                                                                    if (mcTcor < 39) {
                                                                        if (mu_mv_mag_sq < 39) {
                                                                            if (mu_mv_mag_sq < 11) {
                                                                                return 2;
                                                                            }
                                                                            else if (mu_mv_mag_sq >= 11) {
                                                                                if (MVDiff < 850) {
                                                                                    return 4;
                                                                                }
                                                                                else if (MVDiff >= 850) {
                                                                                    return 2;
                                                                                }
                                                                            }
                                                                        }
                                                                        else if (mu_mv_mag_sq >= 39) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                    else if (mcTcor >= 39) {
                                                                        if (posBalance < 28) {
                                                                            if (MVDiff < 61) {
                                                                                return 1;
                                                                            }
                                                                            else if (MVDiff >= 61) {
                                                                                if (MVDiff < 69) {
                                                                                    if (negBalance < 2) {
                                                                                        if (MVDiff < 62) {
                                                                                            return 1;
                                                                                        }
                                                                                        else if (MVDiff >= 62) {
                                                                                            return 8;
                                                                                        }
                                                                                    }
                                                                                    else if (negBalance >= 2) {
                                                                                        return 8;
                                                                                    }
                                                                                }
                                                                                else if (MVDiff >= 69) {
                                                                                    if (tcor < 32) {
                                                                                        return 1;
                                                                                    }
                                                                                    else if (tcor >= 32) {
                                                                                        if (posBalance < 24) {
                                                                                            return 2;
                                                                                        }
                                                                                        else if (posBalance >= 24) {
                                                                                            return 4;
                                                                                        }
                                                                                    }
                                                                                }
                                                                            }
                                                                        }
                                                                        else if (posBalance >= 28) {
                                                                            if (negBalance < 2) {
                                                                                if (Cs < 18) {
                                                                                    return 8;
                                                                                }
                                                                                else if (Cs >= 18) {
                                                                                    return 1;
                                                                                }
                                                                            }
                                                                            else if (negBalance >= 2) {
                                                                                if (diffMVdiffVal < 3) {
                                                                                    return 8;
                                                                                }
                                                                                else if (diffMVdiffVal >= 3) {
                                                                                    return 4;
                                                                                }
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                                else if (mcTcor >= 49) {
                                                                    if (MVDiff < 304) {
                                                                        if (diffAFD < -1) {
                                                                            if (diffMVdiffVal < 101) {
                                                                                if (posBalance < 94) {
                                                                                    return 2;
                                                                                }
                                                                                else if (posBalance >= 94) {
                                                                                    return 4;
                                                                                }
                                                                            }
                                                                            else if (diffMVdiffVal >= 101) {
                                                                                return 8;
                                                                            }
                                                                        }
                                                                        else if (diffAFD >= -1) {
                                                                            if (MVDiff < 51) {
                                                                                return 2;
                                                                            }
                                                                            else if (MVDiff >= 51) {
                                                                                if (posBalance < 110) {
                                                                                    if (diffMVdiffVal < -71) {
                                                                                        if (mcTcor < 64) {
                                                                                            if (MVDiff < 156) {
                                                                                                return 2;
                                                                                            }
                                                                                            else if (MVDiff >= 156) {
                                                                                                return 4;
                                                                                            }
                                                                                        }
                                                                                        else if (mcTcor >= 64) {
                                                                                            return 4;
                                                                                        }
                                                                                    }
                                                                                    else if (diffMVdiffVal >= -71) {
                                                                                        if (diffMVdiffVal < -32) {
                                                                                            return 8;
                                                                                        }
                                                                                        else if (diffMVdiffVal >= -32) {
                                                                                            if (MVDiff < 101) {
                                                                                                if (posBalance < 24) {
                                                                                                    if (MVDiff < 67) {
                                                                                                        return 4;
                                                                                                    }
                                                                                                    else if (MVDiff >= 67) {
                                                                                                        return 8;
                                                                                                    }
                                                                                                }
                                                                                                else if (posBalance >= 24) {
                                                                                                    return 4;
                                                                                                }
                                                                                            }
                                                                                            else if (MVDiff >= 101) {
                                                                                                if (MVDiff < 111) {
                                                                                                    return 8;
                                                                                                }
                                                                                                else if (MVDiff >= 111) {
                                                                                                    return 2;
                                                                                                }
                                                                                            }
                                                                                        }
                                                                                    }
                                                                                }
                                                                                else if (posBalance >= 110) {
                                                                                    if (tcor < 63) {
                                                                                        return 1;
                                                                                    }
                                                                                    else if (tcor >= 63) {
                                                                                        return 4;
                                                                                    }
                                                                                }
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (MVDiff >= 304) {
                                                                        if (AFD < 1) {
                                                                            if (posBalance < 73) {
                                                                                return 2;
                                                                            }
                                                                            else if (posBalance >= 73) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                        else if (AFD >= 1) {
                                                                            if (MVDiff < 694) {
                                                                                return 1;
                                                                            }
                                                                            else if (MVDiff >= 694) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                            else if (negBalance >= 4) {
                                                                if (MVDiff < 49) {
                                                                    if (Rs < 27) {
                                                                        return 2;
                                                                    }
                                                                    else if (Rs >= 27) {
                                                                        return 1;
                                                                    }
                                                                }
                                                                else if (MVDiff >= 49) {
                                                                    return 1;
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (Rs >= 403) {
                                                        if (negBalance < 43) {
                                                            if (MVDiff < 31) {
                                                                return 4;
                                                            }
                                                            else if (MVDiff >= 31) {
                                                                if (Cs < 343) {
                                                                    if (Rs < 422) {
                                                                        return 1;
                                                                    }
                                                                    else if (Rs >= 422) {
                                                                        return 8;
                                                                    }
                                                                }
                                                                else if (Cs >= 343) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                        else if (negBalance >= 43) {
                                                            if (negBalance < 47) {
                                                                if (posBalance < 55) {
                                                                    return 2;
                                                                }
                                                                else if (posBalance >= 55) {
                                                                    if (mcTcor < 83) {
                                                                        return 1;
                                                                    }
                                                                    else if (mcTcor >= 83) {
                                                                        if (posBalance < 57) {
                                                                            return 4;
                                                                        }
                                                                        else if (posBalance >= 57) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                            else if (negBalance >= 47) {
                                                                if (SC < 811) {
                                                                    return 8;
                                                                }
                                                                else if (SC >= 811) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (refDCval >= 63) {
                                                    if (mu_mv_mag_sq < 51) {
                                                        if (negBalance < 50) {
                                                            if (refDCval < 72) {
                                                                return 2;
                                                            }
                                                            else if (refDCval >= 72) {
                                                                if (RsDiff < 1) {
                                                                    if (tcor < 60) {
                                                                        return 4;
                                                                    }
                                                                    else if (tcor >= 60) {
                                                                        return 2;
                                                                    }
                                                                }
                                                                else if (RsDiff >= 1) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                        else if (negBalance >= 50) {
                                                            return 2;
                                                        }
                                                    }
                                                    else if (mu_mv_mag_sq >= 51) {
                                                        if (Cs < 376) {
                                                            return 4;
                                                        }
                                                        else if (Cs >= 376) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (Cs >= 382) {
                                                if (TSC < 4) {
                                                    if (negBalance < 40) {
                                                        if (Cs < 474) {
                                                            if (posBalance < 51) {
                                                                return 8;
                                                            }
                                                            else if (posBalance >= 51) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (Cs >= 474) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (negBalance >= 40) {
                                                        return 4;
                                                    }
                                                }
                                                else if (TSC >= 4) {
                                                    if (ssDCval < 39) {
                                                        if (Rs < 647) {
                                                            if (MVDiff < 24) {
                                                                if (SC < 850) {
                                                                    if (MVDiff < 23) {
                                                                        return 2;
                                                                    }
                                                                    else if (MVDiff >= 23) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (SC >= 850) {
                                                                    if (Cs < 428) {
                                                                        return 2;
                                                                    }
                                                                    else if (Cs >= 428) {
                                                                        return 4;
                                                                    }
                                                                }
                                                            }
                                                            else if (MVDiff >= 24) {
                                                                if (gchDC < 1) {
                                                                    return 2;
                                                                }
                                                                else if (gchDC >= 1) {
                                                                    if (tcor < 77) {
                                                                        return 8;
                                                                    }
                                                                    else if (tcor >= 77) {
                                                                        return 2;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (Rs >= 647) {
                                                            if (Rs < 712) {
                                                                if (SC < 1196) {
                                                                    return 4;
                                                                }
                                                                else if (SC >= 1196) {
                                                                    return 1;
                                                                }
                                                            }
                                                            else if (Rs >= 712) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                    else if (ssDCval >= 39) {
                                                        if (refDCval < 60) {
                                                            return 4;
                                                        }
                                                        else if (refDCval >= 60) {
                                                            if (ssDCval < 88) {
                                                                if (gchDC < 3) {
                                                                    return 2;
                                                                }
                                                                else if (gchDC >= 3) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (ssDCval >= 88) {
                                                                if (CsDiff < 2) {
                                                                    return 8;
                                                                }
                                                                else if (CsDiff >= 2) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (AFD >= 7) {
                                            if (MVDiff < 13) {
                                                if (negBalance < 15) {
                                                    if (SC < 537) {
                                                        return 2;
                                                    }
                                                    else if (SC >= 537) {
                                                        return 8;
                                                    }
                                                }
                                                else if (negBalance >= 15) {
                                                    if (RsCsDiff < 43) {
                                                        return 2;
                                                    }
                                                    else if (RsCsDiff >= 43) {
                                                        return 1;
                                                    }
                                                }
                                            }
                                            else if (MVDiff >= 13) {
                                                if (tcor < 79) {
                                                    if (ssDCval < 74) {
                                                        if (diffAFD < 3) {
                                                            if (mu_mv_mag_sq < 31) {
                                                                if (Cs < 424) {
                                                                    return 1;
                                                                }
                                                                else if (Cs >= 424) {
                                                                    if (Rs < 721) {
                                                                        return 4;
                                                                    }
                                                                    else if (Rs >= 721) {
                                                                        if (Rs < 753) {
                                                                            return 1;
                                                                        }
                                                                        else if (Rs >= 753) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                            else if (mu_mv_mag_sq >= 31) {
                                                                if (Rs < 194) {
                                                                    return 4;
                                                                }
                                                                else if (Rs >= 194) {
                                                                    return 1;
                                                                }
                                                            }
                                                        }
                                                        else if (diffAFD >= 3) {
                                                            if (Rs < 999) {
                                                                return 8;
                                                            }
                                                            else if (Rs >= 999) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                    else if (ssDCval >= 74) {
                                                        if (Rs < 220) {
                                                            return 2;
                                                        }
                                                        else if (Rs >= 220) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                                else if (tcor >= 79) {
                                                    if (CsDiff < 2) {
                                                        return 1;
                                                    }
                                                    else if (CsDiff >= 2) {
                                                        if (RsDiff < 3) {
                                                            return 4;
                                                        }
                                                        else if (RsDiff >= 3) {
                                                            return 2;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (Cs >= 544) {
                                        if (negBalance < 25) {
                                            if (Rs < 746) {
                                                if (RsCsDiff < 11) {
                                                    if (TSC < 1) {
                                                        return 2;
                                                    }
                                                    else if (TSC >= 1) {
                                                        if (Cs < 749) {
                                                            return 4;
                                                        }
                                                        else if (Cs >= 749) {
                                                            return 1;
                                                        }
                                                    }
                                                }
                                                else if (RsCsDiff >= 11) {
                                                    return 2;
                                                }
                                            }
                                            else if (Rs >= 746) {
                                                if (negBalance < 19) {
                                                    if (diffMVdiffVal < 7) {
                                                        if (diffMVdiffVal < -14) {
                                                            return 2;
                                                        }
                                                        else if (diffMVdiffVal >= -14) {
                                                            if (SC < 3692) {
                                                                if (TSC < 2) {
                                                                    if (Cs < 1038) {
                                                                        return 1;
                                                                    }
                                                                    else if (Cs >= 1038) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (TSC >= 2) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (SC >= 3692) {
                                                                return 1;
                                                            }
                                                        }
                                                    }
                                                    else if (diffMVdiffVal >= 7) {
                                                        if (Rs < 809) {
                                                            return 1;
                                                        }
                                                        else if (Rs >= 809) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                                else if (negBalance >= 19) {
                                                    if (mu_mv_mag_sq < 73) {
                                                        if (AFD < 5) {
                                                            return 1;
                                                        }
                                                        else if (AFD >= 5) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (mu_mv_mag_sq >= 73) {
                                                        return 1;
                                                    }
                                                }
                                            }
                                        }
                                        else if (negBalance >= 25) {
                                            if (TSCindex < 8) {
                                                if (SC < 3042) {
                                                    if (SC < 1934) {
                                                        if (TSCindex < 6) {
                                                            if (Rs < 943) {
                                                                if (ssDCval < 50) {
                                                                    if (RsDiff < 3) {
                                                                        if (posBalance < 57) {
                                                                            if (diffAFD < 1) {
                                                                                return 4;
                                                                            }
                                                                            else if (diffAFD >= 1) {
                                                                                if (SC < 1586) {
                                                                                    return 4;
                                                                                }
                                                                                else if (SC >= 1586) {
                                                                                    return 8;
                                                                                }
                                                                            }
                                                                        }
                                                                        else if (posBalance >= 57) {
                                                                            if (Cs < 734) {
                                                                                return 2;
                                                                            }
                                                                            else if (Cs >= 734) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (RsDiff >= 3) {
                                                                        if (MVDiff < 30) {
                                                                            return 2;
                                                                        }
                                                                        else if (MVDiff >= 30) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                }
                                                                else if (ssDCval >= 50) {
                                                                    if (ssDCval < 65) {
                                                                        if (Cs < 562) {
                                                                            if (ssDCval < 52) {
                                                                                return 8;
                                                                            }
                                                                            else if (ssDCval >= 52) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                        else if (Cs >= 562) {
                                                                            if (negBalance < 39) {
                                                                                return 8;
                                                                            }
                                                                            else if (negBalance >= 39) {
                                                                                if (TSC < 4) {
                                                                                    return 4;
                                                                                }
                                                                                else if (TSC >= 4) {
                                                                                    return 8;
                                                                                }
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (ssDCval >= 65) {
                                                                        if (SC < 1592) {
                                                                            return 2;
                                                                        }
                                                                        else if (SC >= 1592) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                            else if (Rs >= 943) {
                                                                if (ssDCval < 46) {
                                                                    if (SC < 1690) {
                                                                        return 1;
                                                                    }
                                                                    else if (SC >= 1690) {
                                                                        return 8;
                                                                    }
                                                                }
                                                                else if (ssDCval >= 46) {
                                                                    if (Rs < 1085) {
                                                                        if (RsDiff < 3) {
                                                                            if (ssDCval < 48) {
                                                                                if (SC < 1636) {
                                                                                    return 4;
                                                                                }
                                                                                else if (SC >= 1636) {
                                                                                    return 1;
                                                                                }
                                                                            }
                                                                            else if (ssDCval >= 48) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                        else if (RsDiff >= 3) {
                                                                            if (mu_mv_mag_sq < 41) {
                                                                                if (Rs < 988) {
                                                                                    return 4;
                                                                                }
                                                                                else if (Rs >= 988) {
                                                                                    return 1;
                                                                                }
                                                                            }
                                                                            else if (mu_mv_mag_sq >= 41) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (Rs >= 1085) {
                                                                        if (MVDiff < 13) {
                                                                            return 4;
                                                                        }
                                                                        else if (MVDiff >= 13) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (TSCindex >= 6) {
                                                            if (refDCval < 55) {
                                                                if (tcor < 93) {
                                                                    if (mu_mv_mag_sq < 53) {
                                                                        if (Cs < 654) {
                                                                            if (SC < 1327) {
                                                                                if (Rs < 745) {
                                                                                    return 1;
                                                                                }
                                                                                else if (Rs >= 745) {
                                                                                    return 4;
                                                                                }
                                                                            }
                                                                            else if (SC >= 1327) {
                                                                                if (RsCsDiff < 7) {
                                                                                    if (negBalance < 49) {
                                                                                        if (MVDiff < 25) {
                                                                                            return 1;
                                                                                        }
                                                                                        else if (MVDiff >= 25) {
                                                                                            return 4;
                                                                                        }
                                                                                    }
                                                                                    else if (negBalance >= 49) {
                                                                                        return 1;
                                                                                    }
                                                                                }
                                                                                else if (RsCsDiff >= 7) {
                                                                                    return 1;
                                                                                }
                                                                            }
                                                                        }
                                                                        else if (Cs >= 654) {
                                                                            if (mcTcor < 77) {
                                                                                if (Cs < 786) {
                                                                                    return 1;
                                                                                }
                                                                                else if (Cs >= 786) {
                                                                                    return 8;
                                                                                }
                                                                            }
                                                                            else if (mcTcor >= 77) {
                                                                                if (MVDiff < 49) {
                                                                                    if (Rs < 839) {
                                                                                        return 1;
                                                                                    }
                                                                                    else if (Rs >= 839) {
                                                                                        return 4;
                                                                                    }
                                                                                }
                                                                                else if (MVDiff >= 49) {
                                                                                    return 1;
                                                                                }
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (mu_mv_mag_sq >= 53) {
                                                                        if (negBalance < 49) {
                                                                            return 4;
                                                                        }
                                                                        else if (negBalance >= 49) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                }
                                                                else if (tcor >= 93) {
                                                                    if (Rs < 985) {
                                                                        return 8;
                                                                    }
                                                                    else if (Rs >= 985) {
                                                                        return 1;
                                                                    }
                                                                }
                                                            }
                                                            else if (refDCval >= 55) {
                                                                if (refDCval < 78) {
                                                                    if (Cs < 657) {
                                                                        return 8;
                                                                    }
                                                                    else if (Cs >= 657) {
                                                                        if (RsCsDiff < 11) {
                                                                            return 8;
                                                                        }
                                                                        else if (RsCsDiff >= 11) {
                                                                            return 2;
                                                                        }
                                                                    }
                                                                }
                                                                else if (refDCval >= 78) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (SC >= 1934) {
                                                        if (ssDCval < 41) {
                                                            if (TSCindex < 6) {
                                                                return 2;
                                                            }
                                                            else if (TSCindex >= 6) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (ssDCval >= 41) {
                                                            if (ssDCval < 53) {
                                                                if (Rs < 1127) {
                                                                    if (tcor < 93) {
                                                                        return 4;
                                                                    }
                                                                    else if (tcor >= 93) {
                                                                        if (Rs < 1001) {
                                                                            return 2;
                                                                        }
                                                                        else if (Rs >= 1001) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                }
                                                                else if (Rs >= 1127) {
                                                                    if (mu_mv_mag_sq < 66) {
                                                                        if (TSCindex < 7) {
                                                                            if (negBalance < 37) {
                                                                                return 1;
                                                                            }
                                                                            else if (negBalance >= 37) {
                                                                                if (SC < 2385) {
                                                                                    if (refDCval < 49) {
                                                                                        if (Cs < 841) {
                                                                                            return 8;
                                                                                        }
                                                                                        else if (Cs >= 841) {
                                                                                            return 1;
                                                                                        }
                                                                                    }
                                                                                    else if (refDCval >= 49) {
                                                                                        if (MVDiff < 7) {
                                                                                            return 1;
                                                                                        }
                                                                                        else if (MVDiff >= 7) {
                                                                                            if (diffMVdiffVal < 1) {
                                                                                                return 1;
                                                                                            }
                                                                                            else if (diffMVdiffVal >= 1) {
                                                                                                if (negBalance < 40) {
                                                                                                    return 4;
                                                                                                }
                                                                                                else if (negBalance >= 40) {
                                                                                                    return 1;
                                                                                                }
                                                                                            }
                                                                                        }
                                                                                    }
                                                                                }
                                                                                else if (SC >= 2385) {
                                                                                    if (posBalance < 51) {
                                                                                        if (SC < 2787) {
                                                                                            return 1;
                                                                                        }
                                                                                        else if (SC >= 2787) {
                                                                                            return 4;
                                                                                        }
                                                                                    }
                                                                                    else if (posBalance >= 51) {
                                                                                        if (Cs < 1069) {
                                                                                            return 8;
                                                                                        }
                                                                                        else if (Cs >= 1069) {
                                                                                            return 1;
                                                                                        }
                                                                                    }
                                                                                }
                                                                            }
                                                                        }
                                                                        else if (TSCindex >= 7) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                    else if (mu_mv_mag_sq >= 66) {
                                                                        if (MVDiff < 8) {
                                                                            if (negBalance < 35) {
                                                                                return 1;
                                                                            }
                                                                            else if (negBalance >= 35) {
                                                                                if (Rs < 1444) {
                                                                                    if (diffMVdiffVal < -2) {
                                                                                        return 1;
                                                                                    }
                                                                                    else if (diffMVdiffVal >= -2) {
                                                                                        if (Rs < 1300) {
                                                                                            return 1;
                                                                                        }
                                                                                        else if (Rs >= 1300) {
                                                                                            return 4;
                                                                                        }
                                                                                    }
                                                                                }
                                                                                else if (Rs >= 1444) {
                                                                                    return 4;
                                                                                }
                                                                            }
                                                                        }
                                                                        else if (MVDiff >= 8) {
                                                                            if (ssDCval < 50) {
                                                                                return 1;
                                                                            }
                                                                            else if (ssDCval >= 50) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                            else if (ssDCval >= 53) {
                                                                if (posBalance < 56) {
                                                                    if (ssDCval < 57) {
                                                                        if (CsDiff < 4) {
                                                                            if (Cs < 1016) {
                                                                                if (mu_mv_mag_sq < 58) {
                                                                                    if (Cs < 988) {
                                                                                        return 1;
                                                                                    }
                                                                                    else if (Cs >= 988) {
                                                                                        return 8;
                                                                                    }
                                                                                }
                                                                                else if (mu_mv_mag_sq >= 58) {
                                                                                    if (MVDiff < 3) {
                                                                                        return 1;
                                                                                    }
                                                                                    else if (MVDiff >= 3) {
                                                                                        return 4;
                                                                                    }
                                                                                }
                                                                            }
                                                                            else if (Cs >= 1016) {
                                                                                if (MVDiff < 4) {
                                                                                    if (refDCval < 56) {
                                                                                        return 1;
                                                                                    }
                                                                                    else if (refDCval >= 56) {
                                                                                        return 8;
                                                                                    }
                                                                                }
                                                                                else if (MVDiff >= 4) {
                                                                                    return 8;
                                                                                }
                                                                            }
                                                                        }
                                                                        else if (CsDiff >= 4) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                    else if (ssDCval >= 57) {
                                                                        if (AFD < 5) {
                                                                            return 8;
                                                                        }
                                                                        else if (AFD >= 5) {
                                                                            if (ssDCval < 82) {
                                                                                return 4;
                                                                            }
                                                                            else if (ssDCval >= 82) {
                                                                                return 2;
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                                else if (posBalance >= 56) {
                                                                    if (tcor < 87) {
                                                                        if (AFD < 7) {
                                                                            return 8;
                                                                        }
                                                                        else if (AFD >= 7) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                    else if (tcor >= 87) {
                                                                        if (mu_mv_mag_sq < 79) {
                                                                            return 1;
                                                                        }
                                                                        else if (mu_mv_mag_sq >= 79) {
                                                                            if (Rs < 1272) {
                                                                                return 2;
                                                                            }
                                                                            else if (Rs >= 1272) {
                                                                                return 8;
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (SC >= 3042) {
                                                    if (SC < 4789) {
                                                        if (RsDiff < 4) {
                                                            if (Cs < 1554) {
                                                                if (Cs < 1527) {
                                                                    if (Cs < 1227) {
                                                                        return 1;
                                                                    }
                                                                    else if (Cs >= 1227) {
                                                                        if (mu_mv_mag_sq < 85) {
                                                                            if (Rs < 1841) {
                                                                                if (diffRsCsdiff < -2) {
                                                                                    return 1;
                                                                                }
                                                                                else if (diffRsCsdiff >= -2) {
                                                                                    return 4;
                                                                                }
                                                                            }
                                                                            else if (Rs >= 1841) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                        else if (mu_mv_mag_sq >= 85) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                }
                                                                else if (Cs >= 1527) {
                                                                    if (Rs < 1905) {
                                                                        if (mu_mv_mag_sq < 64) {
                                                                            if (Rs < 1850) {
                                                                                if (mu_mv_mag_sq < 48) {
                                                                                    return 1;
                                                                                }
                                                                                else if (mu_mv_mag_sq >= 48) {
                                                                                    if (Rs < 1829) {
                                                                                        return 1;
                                                                                    }
                                                                                    else if (Rs >= 1829) {
                                                                                        return 4;
                                                                                    }
                                                                                }
                                                                            }
                                                                            else if (Rs >= 1850) {
                                                                                return 1;
                                                                            }
                                                                        }
                                                                        else if (mu_mv_mag_sq >= 64) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                    else if (Rs >= 1905) {
                                                                        return 4;
                                                                    }
                                                                }
                                                            }
                                                            else if (Cs >= 1554) {
                                                                if (refDCval < 40) {
                                                                    if (Cs < 1565) {
                                                                        if (Rs < 1949) {
                                                                            return 1;
                                                                        }
                                                                        else if (Rs >= 1949) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                    else if (Cs >= 1565) {
                                                                        return 1;
                                                                    }
                                                                }
                                                                else if (refDCval >= 40) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                        else if (RsDiff >= 4) {
                                                            if (Scindex < 5) {
                                                                if (Rs < 2479) {
                                                                    if (SC < 4075) {
                                                                        return 4;
                                                                    }
                                                                    else if (SC >= 4075) {
                                                                        return 8;
                                                                    }
                                                                }
                                                                else if (Rs >= 2479) {
                                                                    return 2;
                                                                }
                                                            }
                                                            else if (Scindex >= 5) {
                                                                if (TSC < 6) {
                                                                    return 4;
                                                                }
                                                                else if (TSC >= 6) {
                                                                    if (refDCval < 78) {
                                                                        return 4;
                                                                    }
                                                                    else if (refDCval >= 78) {
                                                                        return 1;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (SC >= 4789) {
                                                        if (negBalance < 53) {
                                                            if (refDCval < 54) {
                                                                if (MVDiff < 32) {
                                                                    return 4;
                                                                }
                                                                else if (MVDiff >= 32) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (refDCval >= 54) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (negBalance >= 53) {
                                                            return 2;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (TSCindex >= 8) {
                                                if (negBalance < 47) {
                                                    if (Cs < 1840) {
                                                        if (negBalance < 44) {
                                                            if (RsDiff < 5) {
                                                                return 4;
                                                            }
                                                            else if (RsDiff >= 5) {
                                                                return 1;
                                                            }
                                                        }
                                                        else if (negBalance >= 44) {
                                                            return 1;
                                                        }
                                                    }
                                                    else if (Cs >= 1840) {
                                                        if (Rs < 1950) {
                                                            return 4;
                                                        }
                                                        else if (Rs >= 1950) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                                else if (negBalance >= 47) {
                                                    if (negBalance < 49) {
                                                        if (diffMVdiffVal < 1) {
                                                            if (RsDiff < 4) {
                                                                return 2;
                                                            }
                                                            else if (RsDiff >= 4) {
                                                                if (MVDiff < 2) {
                                                                    return 2;
                                                                }
                                                                else if (MVDiff >= 2) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                        else if (diffMVdiffVal >= 1) {
                                                            return 1;
                                                        }
                                                    }
                                                    else if (negBalance >= 49) {
                                                        if (mcTcor < 87) {
                                                            if (Cs < 1610) {
                                                                return 2;
                                                            }
                                                            else if (Cs >= 1610) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (mcTcor >= 87) {
                                                            if (Cs < 3182) {
                                                                if (diffRsCsdiff < -18) {
                                                                    return 4;
                                                                }
                                                                else if (diffRsCsdiff >= -18) {
                                                                    if (ssDCval < 72) {
                                                                        return 2;
                                                                    }
                                                                    else if (ssDCval >= 72) {
                                                                        if (Cs < 1843) {
                                                                            if (Rs < 1703) {
                                                                                return 2;
                                                                            }
                                                                            else if (Rs >= 1703) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                        else if (Cs >= 1843) {
                                                                            return 2;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                            else if (Cs >= 3182) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (AFD >= 9) {
                                    if (Rs < 543) {
                                        if (TSCindex < 7) {
                                            if (Rs < 99) {
                                                return 1;
                                            }
                                            else if (Rs >= 99) {
                                                if (negBalance < 42) {
                                                    if (Cs < 277) {
                                                        return 1;
                                                    }
                                                    else if (Cs >= 277) {
                                                        if (ssDCval < 64) {
                                                            return 2;
                                                        }
                                                        else if (ssDCval >= 64) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                                else if (negBalance >= 42) {
                                                    return 2;
                                                }
                                            }
                                        }
                                        else if (TSCindex >= 7) {
                                            if (MVDiff < 13) {
                                                if (Rs < 391) {
                                                    return 2;
                                                }
                                                else if (Rs >= 391) {
                                                    return 4;
                                                }
                                            }
                                            else if (MVDiff >= 13) {
                                                return 1;
                                            }
                                        }
                                    }
                                    else if (Rs >= 543) {
                                        if (negBalance < 48) {
                                            if (ssDCval < 84) {
                                                if (ssDCval < 75) {
                                                    if (Cs < 762) {
                                                        if (mcTcor < 84) {
                                                            if (SC < 1515) {
                                                                return 8;
                                                            }
                                                            else if (SC >= 1515) {
                                                                return 1;
                                                            }
                                                        }
                                                        else if (mcTcor >= 84) {
                                                            if (ssDCval < 68) {
                                                                if (RsDiff < 8) {
                                                                    return 2;
                                                                }
                                                                else if (RsDiff >= 8) {
                                                                    return 1;
                                                                }
                                                            }
                                                            else if (ssDCval >= 68) {
                                                                if (Rs < 642) {
                                                                    return 8;
                                                                }
                                                                else if (Rs >= 642) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (Cs >= 762) {
                                                        if (MVDiff < 29) {
                                                            if (Rs < 2259) {
                                                                if (TSC < 7) {
                                                                    if (mcTcor < 91) {
                                                                        if (mu_mv_mag_sq < 3) {
                                                                            if (negBalance < 40) {
                                                                                return 4;
                                                                            }
                                                                            else if (negBalance >= 40) {
                                                                                return 1;
                                                                            }
                                                                        }
                                                                        else if (mu_mv_mag_sq >= 3) {
                                                                            if (TSC < 5) {
                                                                                return 4;
                                                                            }
                                                                            else if (TSC >= 5) {
                                                                                return 1;
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (mcTcor >= 91) {
                                                                        if (diffRsCsdiff < -40) {
                                                                            return 1;
                                                                        }
                                                                        else if (diffRsCsdiff >= -40) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                }
                                                                else if (TSC >= 7) {
                                                                    if (TSC < 9) {
                                                                        return 1;
                                                                    }
                                                                    else if (TSC >= 9) {
                                                                        return 8;
                                                                    }
                                                                }
                                                            }
                                                            else if (Rs >= 2259) {
                                                                if (CsDiff < 9) {
                                                                    if (CsDiff < 3) {
                                                                        return 8;
                                                                    }
                                                                    else if (CsDiff >= 3) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (CsDiff >= 9) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                        else if (MVDiff >= 29) {
                                                            if (Cs < 1269) {
                                                                if (MVDiff < 43) {
                                                                    return 8;
                                                                }
                                                                else if (MVDiff >= 43) {
                                                                    return 1;
                                                                }
                                                            }
                                                            else if (Cs >= 1269) {
                                                                if (posBalance < 62) {
                                                                    return 2;
                                                                }
                                                                else if (posBalance >= 62) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (ssDCval >= 75) {
                                                    if (TSCindex < 9) {
                                                        if (RsDiff < 4) {
                                                            return 8;
                                                        }
                                                        else if (RsDiff >= 4) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (TSCindex >= 9) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                            else if (ssDCval >= 84) {
                                                if (posBalance < 52) {
                                                    if (ssDCval < 88) {
                                                        if (Rs < 2091) {
                                                            if (Rs < 1774) {
                                                                return 2;
                                                            }
                                                            else if (Rs >= 1774) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (Rs >= 2091) {
                                                            return 1;
                                                        }
                                                    }
                                                    else if (ssDCval >= 88) {
                                                        if (Rs < 2670) {
                                                            return 8;
                                                        }
                                                        else if (Rs >= 2670) {
                                                            return 2;
                                                        }
                                                    }
                                                }
                                                else if (posBalance >= 52) {
                                                    if (MVDiff < 3) {
                                                        if (Cs < 1698) {
                                                            return 1;
                                                        }
                                                        else if (Cs >= 1698) {
                                                            return 2;
                                                        }
                                                    }
                                                    else if (MVDiff >= 3) {
                                                        if (negBalance < 38) {
                                                            return 4;
                                                        }
                                                        else if (negBalance >= 38) {
                                                            if (Rs < 2229) {
                                                                if (Rs < 1681) {
                                                                    return 1;
                                                                }
                                                                else if (Rs >= 1681) {
                                                                    return 2;
                                                                }
                                                            }
                                                            else if (Rs >= 2229) {
                                                                return 1;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (negBalance >= 48) {
                                            if (Cs < 385) {
                                                return 1;
                                            }
                                            else if (Cs >= 385) {
                                                if (MVDiff < 4) {
                                                    if (RsDiff < 6) {
                                                        return 2;
                                                    }
                                                    else if (RsDiff >= 6) {
                                                        if (Rs < 2391) {
                                                            return 2;
                                                        }
                                                        else if (Rs >= 2391) {
                                                            if (SC < 4127) {
                                                                return 1;
                                                            }
                                                            else if (SC >= 4127) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (MVDiff >= 4) {
                                                    if (tcor < 93) {
                                                        if (SC < 3760) {
                                                            if (mcTcor < 91) {
                                                                if (RsDiff < 6) {
                                                                    if (diffAFD < -1) {
                                                                        return 4;
                                                                    }
                                                                    else if (diffAFD >= -1) {
                                                                        if (mu_mv_mag_sq < 48) {
                                                                            if (posBalance < 61) {
                                                                                if (mcTcor < 88) {
                                                                                    return 8;
                                                                                }
                                                                                else if (mcTcor >= 88) {
                                                                                    if (tcor < 80) {
                                                                                        return 8;
                                                                                    }
                                                                                    else if (tcor >= 80) {
                                                                                        return 4;
                                                                                    }
                                                                                }
                                                                            }
                                                                            else if (posBalance >= 61) {
                                                                                if (diffMVdiffVal < 3) {
                                                                                    return 4;
                                                                                }
                                                                                else if (diffMVdiffVal >= 3) {
                                                                                    return 8;
                                                                                }
                                                                            }
                                                                        }
                                                                        else if (mu_mv_mag_sq >= 48) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                }
                                                                else if (RsDiff >= 6) {
                                                                    if (Rs < 1840) {
                                                                        if (Rs < 1430) {
                                                                            return 2;
                                                                        }
                                                                        else if (Rs >= 1430) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                    else if (Rs >= 1840) {
                                                                        return 2;
                                                                    }
                                                                }
                                                            }
                                                            else if (mcTcor >= 91) {
                                                                if (SC < 2272) {
                                                                    if (RsDiff < 4) {
                                                                        if (gchDC < 1) {
                                                                            return 8;
                                                                        }
                                                                        else if (gchDC >= 1) {
                                                                            if (SC < 1442) {
                                                                                return 4;
                                                                            }
                                                                            else if (SC >= 1442) {
                                                                                return 1;
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (RsDiff >= 4) {
                                                                        if (ssDCval < 64) {
                                                                            if (Rs < 999) {
                                                                                return 2;
                                                                            }
                                                                            else if (Rs >= 999) {
                                                                                if (AFD < 11) {
                                                                                    return 4;
                                                                                }
                                                                                else if (AFD >= 11) {
                                                                                    return 1;
                                                                                }
                                                                            }
                                                                        }
                                                                        else if (ssDCval >= 64) {
                                                                            return 2;
                                                                        }
                                                                    }
                                                                }
                                                                else if (SC >= 2272) {
                                                                    if (TSC < 8) {
                                                                        if (AFD < 11) {
                                                                            return 4;
                                                                        }
                                                                        else if (AFD >= 11) {
                                                                            if (Cs < 1281) {
                                                                                return 1;
                                                                            }
                                                                            else if (Cs >= 1281) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (TSC >= 8) {
                                                                        if (Cs < 1395) {
                                                                            return 1;
                                                                        }
                                                                        else if (Cs >= 1395) {
                                                                            return 2;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (SC >= 3760) {
                                                            if (refDCval < 88) {
                                                                if (RsDiff < 4) {
                                                                    if (Rs < 3222) {
                                                                        return 8;
                                                                    }
                                                                    else if (Rs >= 3222) {
                                                                        return 2;
                                                                    }
                                                                }
                                                                else if (RsDiff >= 4) {
                                                                    if (mu_mv_mag_sq < 82) {
                                                                        if (diffAFD < -2) {
                                                                            if (Rs < 2599) {
                                                                                return 1;
                                                                            }
                                                                            else if (Rs >= 2599) {
                                                                                return 2;
                                                                            }
                                                                        }
                                                                        else if (diffAFD >= -2) {
                                                                            if (diffRsCsdiff < 33) {
                                                                                return 8;
                                                                            }
                                                                            else if (diffRsCsdiff >= 33) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (mu_mv_mag_sq >= 82) {
                                                                        if (RsDiff < 8) {
                                                                            return 8;
                                                                        }
                                                                        else if (RsDiff >= 8) {
                                                                            if (gchDC < 2) {
                                                                                return 4;
                                                                            }
                                                                            else if (gchDC >= 2) {
                                                                                return 2;
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                            else if (refDCval >= 88) {
                                                                return 2;
                                                            }
                                                        }
                                                    }
                                                    else if (tcor >= 93) {
                                                        if (Rs < 2443) {
                                                            if (Cs < 1874) {
                                                                return 4;
                                                            }
                                                            else if (Cs >= 1874) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (Rs >= 2443) {
                                                            return 1;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                        }
                        else if (ssDCval >= 97) {
                            if (mu_mv_mag_sq < 41) {
                                if (Cs < 315) {
                                    if (Rs < 880) {
                                        if (Rs < 619) {
                                            if (SC < 603) {
                                                if (MVDiff < 32) {
                                                    return 4;
                                                }
                                                else if (MVDiff >= 32) {
                                                    return 1;
                                                }
                                            }
                                            else if (SC >= 603) {
                                                return 2;
                                            }
                                        }
                                        else if (Rs >= 619) {
                                            if (MVDiff < 37) {
                                                return 1;
                                            }
                                            else if (MVDiff >= 37) {
                                                return 2;
                                            }
                                        }
                                    }
                                    else if (Rs >= 880) {
                                        if (refDCval < 103) {
                                            return 1;
                                        }
                                        else if (refDCval >= 103) {
                                            return 8;
                                        }
                                    }
                                }
                                else if (Cs >= 315) {
                                    if (refDCval < 126) {
                                        if (TSC < 6) {
                                            if (Cs < 1068) {
                                                if (SC < 800) {
                                                    return 2;
                                                }
                                                else if (SC >= 800) {
                                                    if (Rs < 627) {
                                                        if (MVDiff < 11) {
                                                            if (posBalance < 60) {
                                                                return 4;
                                                            }
                                                            else if (posBalance >= 60) {
                                                                if (Rs < 406) {
                                                                    return 4;
                                                                }
                                                                else if (Rs >= 406) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                        else if (MVDiff >= 11) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (Rs >= 627) {
                                                        if (Rs < 1068) {
                                                            if (AFD < 10) {
                                                                return 4;
                                                            }
                                                            else if (AFD >= 10) {
                                                                if (MVDiff < 13) {
                                                                    return 4;
                                                                }
                                                                else if (MVDiff >= 13) {
                                                                    return 2;
                                                                }
                                                            }
                                                        }
                                                        else if (Rs >= 1068) {
                                                            if (SC < 2333) {
                                                                if (negBalance < 52) {
                                                                    return 2;
                                                                }
                                                                else if (negBalance >= 52) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (SC >= 2333) {
                                                                if (RsDiff < 6) {
                                                                    return 8;
                                                                }
                                                                else if (RsDiff >= 6) {
                                                                    if (Rs < 2733) {
                                                                        return 4;
                                                                    }
                                                                    else if (Rs >= 2733) {
                                                                        return 2;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                            else if (Cs >= 1068) {
                                                if (Cs < 1775) {
                                                    if (ssDCval < 106) {
                                                        if (MVDiff < 12) {
                                                            if (diffRsCsdiff < 19) {
                                                                return 8;
                                                            }
                                                            else if (diffRsCsdiff >= 19) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (MVDiff >= 12) {
                                                            if (Rs < 3159) {
                                                                return 8;
                                                            }
                                                            else if (Rs >= 3159) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                    else if (ssDCval >= 106) {
                                                        if (SC < 4721) {
                                                            return 4;
                                                        }
                                                        else if (SC >= 4721) {
                                                            if (diffRsCsdiff < -24) {
                                                                return 8;
                                                            }
                                                            else if (diffRsCsdiff >= -24) {
                                                                if (MVDiff < 11) {
                                                                    if (SC < 5377) {
                                                                        return 8;
                                                                    }
                                                                    else if (SC >= 5377) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (MVDiff >= 11) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (Cs >= 1775) {
                                                    if (MVDiff < 13) {
                                                        return 8;
                                                    }
                                                    else if (MVDiff >= 13) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                        }
                                        else if (TSC >= 6) {
                                            if (refDCval < 117) {
                                                if (Rs < 1520) {
                                                    if (negBalance < 52) {
                                                        if (diffAFD < -1) {
                                                            return 2;
                                                        }
                                                        else if (diffAFD >= -1) {
                                                            if (RsDiff < 1) {
                                                                return 2;
                                                            }
                                                            else if (RsDiff >= 1) {
                                                                if (Rs < 807) {
                                                                    if (Rs < 336) {
                                                                        return 4;
                                                                    }
                                                                    else if (Rs >= 336) {
                                                                        return 8;
                                                                    }
                                                                }
                                                                else if (Rs >= 807) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (negBalance >= 52) {
                                                        if (Rs < 1337) {
                                                            return 4;
                                                        }
                                                        else if (Rs >= 1337) {
                                                            if (RsDiff < 5) {
                                                                if (Rs < 1351) {
                                                                    return 8;
                                                                }
                                                                else if (Rs >= 1351) {
                                                                    return 2;
                                                                }
                                                            }
                                                            else if (RsDiff >= 5) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (Rs >= 1520) {
                                                    if (Rs < 2592) {
                                                        if (Cs < 1736) {
                                                            return 8;
                                                        }
                                                        else if (Cs >= 1736) {
                                                            return 2;
                                                        }
                                                    }
                                                    else if (Rs >= 2592) {
                                                        if (Rs < 3571) {
                                                            return 4;
                                                        }
                                                        else if (Rs >= 3571) {
                                                            if (Rs < 3858) {
                                                                return 8;
                                                            }
                                                            else if (Rs >= 3858) {
                                                                if (MVDiff < 11) {
                                                                    return 8;
                                                                }
                                                                else if (MVDiff >= 11) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                            else if (refDCval >= 117) {
                                                if (SC < 4484) {
                                                    return 4;
                                                }
                                                else if (SC >= 4484) {
                                                    if (Rs < 2691) {
                                                        if (Cs < 1932) {
                                                            return 4;
                                                        }
                                                        else if (Cs >= 1932) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (Rs >= 2691) {
                                                        if (diffAFD < -1) {
                                                            if (Cs < 2014) {
                                                                return 4;
                                                            }
                                                            else if (Cs >= 2014) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (diffAFD >= -1) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (refDCval >= 126) {
                                        if (mcTcor < 95) {
                                            if (posBalance < 39) {
                                                if (SC < 2210) {
                                                    return 4;
                                                }
                                                else if (SC >= 2210) {
                                                    if (Cs < 1072) {
                                                        if (AFD < 11) {
                                                            return 1;
                                                        }
                                                        else if (AFD >= 11) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (Cs >= 1072) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                            else if (posBalance >= 39) {
                                                if (Cs < 1590) {
                                                    if (RsCsDiff < 99) {
                                                        return 1;
                                                    }
                                                    else if (RsCsDiff >= 99) {
                                                        return 8;
                                                    }
                                                }
                                                else if (Cs >= 1590) {
                                                    return 4;
                                                }
                                            }
                                        }
                                        else if (mcTcor >= 95) {
                                            if (CsDiff < 2) {
                                                return 4;
                                            }
                                            else if (CsDiff >= 2) {
                                                return 2;
                                            }
                                        }
                                    }
                                }
                            }
                            else if (mu_mv_mag_sq >= 41) {
                                if (RsDiff < 6) {
                                    if (refDCval < 128) {
                                        if (tcor < 84) {
                                            if (Cs < 758) {
                                                if (SC < 664) {
                                                    return 4;
                                                }
                                                else if (SC >= 664) {
                                                    return 8;
                                                }
                                            }
                                            else if (Cs >= 758) {
                                                if (refDCval < 98) {
                                                    if (SC < 3335) {
                                                        return 8;
                                                    }
                                                    else if (SC >= 3335) {
                                                        return 2;
                                                    }
                                                }
                                                else if (refDCval >= 98) {
                                                    return 4;
                                                }
                                            }
                                        }
                                        else if (tcor >= 84) {
                                            if (mu_mv_mag_sq < 48) {
                                                if (posBalance < 46) {
                                                    if (Rs < 1963) {
                                                        return 1;
                                                    }
                                                    else if (Rs >= 1963) {
                                                        return 4;
                                                    }
                                                }
                                                else if (posBalance >= 46) {
                                                    if (SC < 714) {
                                                        if (Cs < 54) {
                                                            return 1;
                                                        }
                                                        else if (Cs >= 54) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (SC >= 714) {
                                                        if (Rs < 3674) {
                                                            if (diffAFD < -1) {
                                                                if (SC < 861) {
                                                                    return 4;
                                                                }
                                                                else if (SC >= 861) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (diffAFD >= -1) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (Rs >= 3674) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (mu_mv_mag_sq >= 48) {
                                                if (Rs < 2667) {
                                                    if (refDCval < 102) {
                                                        if (CsDiff < 1) {
                                                            return 8;
                                                        }
                                                        else if (CsDiff >= 1) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (refDCval >= 102) {
                                                        if (mu_mv_mag_sq < 87) {
                                                            if (diffMVdiffVal < -109) {
                                                                return 2;
                                                            }
                                                            else if (diffMVdiffVal >= -109) {
                                                                if (gchDC < 6) {
                                                                    return 8;
                                                                }
                                                                else if (gchDC >= 6) {
                                                                    if (Rs < 853) {
                                                                        return 4;
                                                                    }
                                                                    else if (Rs >= 853) {
                                                                        return 8;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (mu_mv_mag_sq >= 87) {
                                                            return 2;
                                                        }
                                                    }
                                                }
                                                else if (Rs >= 2667) {
                                                    if (Cs < 1530) {
                                                        return 4;
                                                    }
                                                    else if (Cs >= 1530) {
                                                        if (CsDiff < 5) {
                                                            return 8;
                                                        }
                                                        else if (CsDiff >= 5) {
                                                            if (ssDCval < 115) {
                                                                if (SC < 5309) {
                                                                    return 4;
                                                                }
                                                                else if (SC >= 5309) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (ssDCval >= 115) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (refDCval >= 128) {
                                        if (negBalance < 12) {
                                            if (Rs < 360) {
                                                return 1;
                                            }
                                            else if (Rs >= 360) {
                                                if (MVDiff < 6) {
                                                    return 4;
                                                }
                                                else if (MVDiff >= 6) {
                                                    return 2;
                                                }
                                            }
                                        }
                                        else if (negBalance >= 12) {
                                            if (ssDCval < 167) {
                                                if (mu_mv_mag_sq < 73) {
                                                    if (ssDCval < 150) {
                                                        return 8;
                                                    }
                                                    else if (ssDCval >= 150) {
                                                        if (Rs < 818) {
                                                            return 4;
                                                        }
                                                        else if (Rs >= 818) {
                                                            return 2;
                                                        }
                                                    }
                                                }
                                                else if (mu_mv_mag_sq >= 73) {
                                                    if (Cs < 352) {
                                                        return 8;
                                                    }
                                                    else if (Cs >= 352) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                            else if (ssDCval >= 167) {
                                                if (Rs < 538) {
                                                    return 1;
                                                }
                                                else if (Rs >= 538) {
                                                    return 8;
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (RsDiff >= 6) {
                                    if (Cs < 2078) {
                                        if (ssDCval < 99) {
                                            if (AFD < 11) {
                                                return 2;
                                            }
                                            else if (AFD >= 11) {
                                                return 4;
                                            }
                                        }
                                        else if (ssDCval >= 99) {
                                            if (mcTcor < 96) {
                                                if (MVDiff < 10) {
                                                    if (mcTcor < 86) {
                                                        if (refDCval < 113) {
                                                            return 4;
                                                        }
                                                        else if (refDCval >= 113) {
                                                            if (ssDCval < 121) {
                                                                return 8;
                                                            }
                                                            else if (ssDCval >= 121) {
                                                                return 1;
                                                            }
                                                        }
                                                    }
                                                    else if (mcTcor >= 86) {
                                                        if (negBalance < 51) {
                                                            if (SC < 5879) {
                                                                if (negBalance < 38) {
                                                                    return 8;
                                                                }
                                                                else if (negBalance >= 38) {
                                                                    if (Rs < 2853) {
                                                                        return 8;
                                                                    }
                                                                    else if (Rs >= 2853) {
                                                                        if (SC < 4975) {
                                                                            return 4;
                                                                        }
                                                                        else if (SC >= 4975) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                            else if (SC >= 5879) {
                                                                if (Rs < 3914) {
                                                                    return 8;
                                                                }
                                                                else if (Rs >= 3914) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                        else if (negBalance >= 51) {
                                                            if (SC < 4707) {
                                                                if (AFD < 11) {
                                                                    if (MVDiff < 8) {
                                                                        if (refDCval < 119) {
                                                                            return 4;
                                                                        }
                                                                        else if (refDCval >= 119) {
                                                                            if (Rs < 2363) {
                                                                                if (Cs < 1629) {
                                                                                    return 8;
                                                                                }
                                                                                else if (Cs >= 1629) {
                                                                                    return 4;
                                                                                }
                                                                            }
                                                                            else if (Rs >= 2363) {
                                                                                return 8;
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (MVDiff >= 8) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (AFD >= 11) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (SC >= 4707) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (MVDiff >= 10) {
                                                    if (TSC < 6) {
                                                        return 8;
                                                    }
                                                    else if (TSC >= 6) {
                                                        if (diffRsCsdiff < -41) {
                                                            if (MVDiff < 18) {
                                                                return 8;
                                                            }
                                                            else if (MVDiff >= 18) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (diffRsCsdiff >= -41) {
                                                            if (mu_mv_mag_sq < 65) {
                                                                return 4;
                                                            }
                                                            else if (mu_mv_mag_sq >= 65) {
                                                                if (RsDiff < 10) {
                                                                    if (RsDiff < 7) {
                                                                        return 8;
                                                                    }
                                                                    else if (RsDiff >= 7) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (RsDiff >= 10) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                            else if (mcTcor >= 96) {
                                                if (Rs < 3833) {
                                                    if (posBalance < 44) {
                                                        return 4;
                                                    }
                                                    else if (posBalance >= 44) {
                                                        if (Rs < 3428) {
                                                            if (SC < 3631) {
                                                                return 2;
                                                            }
                                                            else if (SC >= 3631) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (Rs >= 3428) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                                else if (Rs >= 3833) {
                                                    return 4;
                                                }
                                            }
                                        }
                                    }
                                    else if (Cs >= 2078) {
                                        if (refDCval < 128) {
                                            if (mcTcor < 95) {
                                                if (Cs < 2576) {
                                                    if (refDCval < 107) {
                                                        return 8;
                                                    }
                                                    else if (refDCval >= 107) {
                                                        if (Rs < 2263) {
                                                            if (refDCval < 119) {
                                                                return 8;
                                                            }
                                                            else if (refDCval >= 119) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (Rs >= 2263) {
                                                            if (TSCindex < 9) {
                                                                if (Cs < 2282) {
                                                                    return 4;
                                                                }
                                                                else if (Cs >= 2282) {
                                                                    if (RsCsDiff < 157) {
                                                                        return 8;
                                                                    }
                                                                    else if (RsCsDiff >= 157) {
                                                                        return 4;
                                                                    }
                                                                }
                                                            }
                                                            else if (TSCindex >= 9) {
                                                                if (negBalance < 33) {
                                                                    if (Rs < 5042) {
                                                                        return 4;
                                                                    }
                                                                    else if (Rs >= 5042) {
                                                                        return 8;
                                                                    }
                                                                }
                                                                else if (negBalance >= 33) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (Cs >= 2576) {
                                                    if (mcTcor < 91) {
                                                        if (SC < 6923) {
                                                            return 8;
                                                        }
                                                        else if (SC >= 6923) {
                                                            if (Rs < 4334) {
                                                                if (Rs < 4113) {
                                                                    return 8;
                                                                }
                                                                else if (Rs >= 4113) {
                                                                    return 1;
                                                                }
                                                            }
                                                            else if (Rs >= 4334) {
                                                                if (posBalance < 47) {
                                                                    if (mcTcor < 86) {
                                                                        return 2;
                                                                    }
                                                                    else if (mcTcor >= 86) {
                                                                        if (mu_mv_mag_sq < 86) {
                                                                            return 8;
                                                                        }
                                                                        else if (mu_mv_mag_sq >= 86) {
                                                                            if (Rs < 4994) {
                                                                                return 2;
                                                                            }
                                                                            else if (Rs >= 4994) {
                                                                                return 8;
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                                else if (posBalance >= 47) {
                                                                    if (Cs < 3134) {
                                                                        if (tcor < 85) {
                                                                            if (ssDCval < 109) {
                                                                                return 4;
                                                                            }
                                                                            else if (ssDCval >= 109) {
                                                                                return 2;
                                                                            }
                                                                        }
                                                                        else if (tcor >= 85) {
                                                                            return 2;
                                                                        }
                                                                    }
                                                                    else if (Cs >= 3134) {
                                                                        if (SC < 8401) {
                                                                            return 8;
                                                                        }
                                                                        else if (SC >= 8401) {
                                                                            return 2;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (mcTcor >= 91) {
                                                        if (mcTcor < 94) {
                                                            if (tcor < 90) {
                                                                return 8;
                                                            }
                                                            else if (tcor >= 90) {
                                                                if (MVDiff < 2) {
                                                                    return 8;
                                                                }
                                                                else if (MVDiff >= 2) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                        else if (mcTcor >= 94) {
                                                            if (diffMVdiffVal < 2) {
                                                                if (Cs < 2718) {
                                                                    return 2;
                                                                }
                                                                else if (Cs >= 2718) {
                                                                    if (negBalance < 34) {
                                                                        return 2;
                                                                    }
                                                                    else if (negBalance >= 34) {
                                                                        return 4;
                                                                    }
                                                                }
                                                            }
                                                            else if (diffMVdiffVal >= 2) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                            else if (mcTcor >= 95) {
                                                if (CsDiff < 6) {
                                                    if (Cs < 2145) {
                                                        if (Rs < 4005) {
                                                            return 2;
                                                        }
                                                        else if (Rs >= 4005) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (Cs >= 2145) {
                                                        return 8;
                                                    }
                                                }
                                                else if (CsDiff >= 6) {
                                                    if (ssDCval < 119) {
                                                        if (gchDC < 1) {
                                                            return 2;
                                                        }
                                                        else if (gchDC >= 1) {
                                                            if (SC < 6665) {
                                                                if (TSC < 8) {
                                                                    if (diffRsCsdiff < -70) {
                                                                        return 4;
                                                                    }
                                                                    else if (diffRsCsdiff >= -70) {
                                                                        if (tcor < 93) {
                                                                            if (Rs < 3593) {
                                                                                return 2;
                                                                            }
                                                                            else if (Rs >= 3593) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                        else if (tcor >= 93) {
                                                                            return 2;
                                                                        }
                                                                    }
                                                                }
                                                                else if (TSC >= 8) {
                                                                    if (refDCval < 105) {
                                                                        return 4;
                                                                    }
                                                                    else if (refDCval >= 105) {
                                                                        if (SC < 5563) {
                                                                            if (SC < 5236) {
                                                                                return 2;
                                                                            }
                                                                            else if (SC >= 5236) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                        else if (SC >= 5563) {
                                                                            return 2;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                            else if (SC >= 6665) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                    else if (ssDCval >= 119) {
                                                        if (diffAFD < -1) {
                                                            if (Rs < 4457) {
                                                                return 4;
                                                            }
                                                            else if (Rs >= 4457) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (diffAFD >= -1) {
                                                            if (Rs < 4667) {
                                                                return 4;
                                                            }
                                                            else if (Rs >= 4667) {
                                                                return 2;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (refDCval >= 128) {
                                            if (refDCval < 130) {
                                                return 2;
                                            }
                                            else if (refDCval >= 130) {
                                                if (RsCsDiff < 147) {
                                                    if (Rs < 3151) {
                                                        return 4;
                                                    }
                                                    else if (Rs >= 3151) {
                                                        return 8;
                                                    }
                                                }
                                                else if (RsCsDiff >= 147) {
                                                    return 1;
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                        }
                    }
                    else if (AFD >= 12) {
                        if (refDCval < 139) {
                            if (Cs < 1474) {
                                if (RsDiff < 8) {
                                    if (refDCval < 85) {
                                        if (mcTcor < 88) {
                                            if (negBalance < 47) {
                                                if (AFD < 33) {
                                                    if (refDCval < 42) {
                                                        return 2;
                                                    }
                                                    else if (refDCval >= 42) {
                                                        if (mu_mv_mag_sq < 35) {
                                                            if (diffMVdiffVal < 55) {
                                                                return 1;
                                                            }
                                                            else if (diffMVdiffVal >= 55) {
                                                                if (Rs < 809) {
                                                                    return 2;
                                                                }
                                                                else if (Rs >= 809) {
                                                                    return 1;
                                                                }
                                                            }
                                                        }
                                                        else if (mu_mv_mag_sq >= 35) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                                else if (AFD >= 33) {
                                                    if (SC < 1242) {
                                                        return 8;
                                                    }
                                                    else if (SC >= 1242) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                            else if (negBalance >= 47) {
                                                if (Cs < 559) {
                                                    if (ssDCval < 72) {
                                                        if (posBalance < 66) {
                                                            return 2;
                                                        }
                                                        else if (posBalance >= 66) {
                                                            return 1;
                                                        }
                                                    }
                                                    else if (ssDCval >= 72) {
                                                        return 8;
                                                    }
                                                }
                                                else if (Cs >= 559) {
                                                    if (mcTcor < 85) {
                                                        if (RsDiff < 5) {
                                                            if (Cs < 640) {
                                                                return 8;
                                                            }
                                                            else if (Cs >= 640) {
                                                                if (Rs < 1269) {
                                                                    return 4;
                                                                }
                                                                else if (Rs >= 1269) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                        else if (RsDiff >= 5) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (mcTcor >= 85) {
                                                        if (refDCval < 56) {
                                                            return 1;
                                                        }
                                                        else if (refDCval >= 56) {
                                                            if (AFD < 14) {
                                                                return 8;
                                                            }
                                                            else if (AFD >= 14) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (mcTcor >= 88) {
                                            if (Cs < 930) {
                                                if (diffAFD < -3) {
                                                    if (Rs < 255) {
                                                        return 1;
                                                    }
                                                    else if (Rs >= 255) {
                                                        return 2;
                                                    }
                                                }
                                                else if (diffAFD >= -3) {
                                                    if (negBalance < 42) {
                                                        if (AFD < 15) {
                                                            return 4;
                                                        }
                                                        else if (AFD >= 15) {
                                                            return 2;
                                                        }
                                                    }
                                                    else if (negBalance >= 42) {
                                                        if (refDCval < 64) {
                                                            if (Cs < 695) {
                                                                return 2;
                                                            }
                                                            else if (Cs >= 695) {
                                                                if (Cs < 847) {
                                                                    return 1;
                                                                }
                                                                else if (Cs >= 847) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                        else if (refDCval >= 64) {
                                                            return 2;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (Cs >= 930) {
                                                if (ssDCval < 87) {
                                                    if (SC < 2215) {
                                                        return 2;
                                                    }
                                                    else if (SC >= 2215) {
                                                        return 1;
                                                    }
                                                }
                                                else if (ssDCval >= 87) {
                                                    return 8;
                                                }
                                            }
                                        }
                                    }
                                    else if (refDCval >= 85) {
                                        if (MVDiff < 12) {
                                            if (SC < 2468) {
                                                if (negBalance < 47) {
                                                    return 2;
                                                }
                                                else if (negBalance >= 47) {
                                                    return 4;
                                                }
                                            }
                                            else if (SC >= 2468) {
                                                if (tcor < 85) {
                                                    return 1;
                                                }
                                                else if (tcor >= 85) {
                                                    if (diffMVdiffVal < 2) {
                                                        if (Rs < 1693) {
                                                            return 4;
                                                        }
                                                        else if (Rs >= 1693) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (diffMVdiffVal >= 2) {
                                                        return 2;
                                                    }
                                                }
                                            }
                                        }
                                        else if (MVDiff >= 12) {
                                            if (mcTcor < 96) {
                                                if (SC < 2102) {
                                                    if (posBalance < 40) {
                                                        if (mu_mv_mag_sq < 46) {
                                                            if (Rs < 917) {
                                                                return 2;
                                                            }
                                                            else if (Rs >= 917) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (mu_mv_mag_sq >= 46) {
                                                            if (Rs < 182) {
                                                                return 1;
                                                            }
                                                            else if (Rs >= 182) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                    else if (posBalance >= 40) {
                                                        if (AFD < 38) {
                                                            if (Rs < 422) {
                                                                return 4;
                                                            }
                                                            else if (Rs >= 422) {
                                                                if (diffRsCsdiff < 10) {
                                                                    return 1;
                                                                }
                                                                else if (diffRsCsdiff >= 10) {
                                                                    if (AFD < 13) {
                                                                        return 4;
                                                                    }
                                                                    else if (AFD >= 13) {
                                                                        return 1;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (AFD >= 38) {
                                                            return 2;
                                                        }
                                                    }
                                                }
                                                else if (SC >= 2102) {
                                                    if (MVDiff < 28) {
                                                        if (ssDCval < 117) {
                                                            return 2;
                                                        }
                                                        else if (ssDCval >= 117) {
                                                            return 1;
                                                        }
                                                    }
                                                    else if (MVDiff >= 28) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                            else if (mcTcor >= 96) {
                                                if (Rs < 360) {
                                                    if (RsDiff < 2) {
                                                        return 2;
                                                    }
                                                    else if (RsDiff >= 2) {
                                                        return 1;
                                                    }
                                                }
                                                else if (Rs >= 360) {
                                                    if (diffRsCsdiff < -29) {
                                                        return 4;
                                                    }
                                                    else if (diffRsCsdiff >= -29) {
                                                        if (Cs < 676) {
                                                            if (refDCval < 109) {
                                                                return 8;
                                                            }
                                                            else if (refDCval >= 109) {
                                                                if (Rs < 1806) {
                                                                    return 4;
                                                                }
                                                                else if (Rs >= 1806) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                        else if (Cs >= 676) {
                                                            return 2;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (RsDiff >= 8) {
                                    if (mcTcor < 85) {
                                        if (diffTSC < 19) {
                                            if (refDCval < 83) {
                                                if (diffMVdiffVal < 93) {
                                                    if (Rs < 1230) {
                                                        return 1;
                                                    }
                                                    else if (Rs >= 1230) {
                                                        if (Cs < 1032) {
                                                            if (MVDiff < 52) {
                                                                return 1;
                                                            }
                                                            else if (MVDiff >= 52) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (Cs >= 1032) {
                                                            if (posBalance < 50) {
                                                                return 2;
                                                            }
                                                            else if (posBalance >= 50) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (diffMVdiffVal >= 93) {
                                                    if (ssDCval < 58) {
                                                        return 2;
                                                    }
                                                    else if (ssDCval >= 58) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                            else if (refDCval >= 83) {
                                                if (refDCval < 119) {
                                                    if (MVDiff < 52) {
                                                        return 2;
                                                    }
                                                    else if (MVDiff >= 52) {
                                                        if (diffMVdiffVal < 35) {
                                                            if (SC < 2978) {
                                                                if (posBalance < 65) {
                                                                    return 2;
                                                                }
                                                                else if (posBalance >= 65) {
                                                                    return 1;
                                                                }
                                                            }
                                                            else if (SC >= 2978) {
                                                                if (SC < 4341) {
                                                                    if (diffAFD < 3) {
                                                                        return 1;
                                                                    }
                                                                    else if (diffAFD >= 3) {
                                                                        return 2;
                                                                    }
                                                                }
                                                                else if (SC >= 4341) {
                                                                    return 2;
                                                                }
                                                            }
                                                        }
                                                        else if (diffMVdiffVal >= 35) {
                                                            return 2;
                                                        }
                                                    }
                                                }
                                                else if (refDCval >= 119) {
                                                    if (CsDiff < 6) {
                                                        if (AFD < 17) {
                                                            return 2;
                                                        }
                                                        else if (AFD >= 17) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (CsDiff >= 6) {
                                                        return 1;
                                                    }
                                                }
                                            }
                                        }
                                        else if (diffTSC >= 19) {
                                            if (ssDCval < 125) {
                                                return 4;
                                            }
                                            else if (ssDCval >= 125) {
                                                if (Rs < 1191) {
                                                    if (Cs < 907) {
                                                        return 4;
                                                    }
                                                    else if (Cs >= 907) {
                                                        return 1;
                                                    }
                                                }
                                                else if (Rs >= 1191) {
                                                    return 2;
                                                }
                                            }
                                        }
                                    }
                                    else if (mcTcor >= 85) {
                                        if (negBalance < 51) {
                                            if (AFD < 15) {
                                                if (Scindex < 5) {
                                                    if (Cs < 1040) {
                                                        return 2;
                                                    }
                                                    else if (Cs >= 1040) {
                                                        if (CsDiff < 6) {
                                                            return 8;
                                                        }
                                                        else if (CsDiff >= 6) {
                                                            return 1;
                                                        }
                                                    }
                                                }
                                                else if (Scindex >= 5) {
                                                    return 4;
                                                }
                                            }
                                            else if (AFD >= 15) {
                                                if (Rs < 866) {
                                                    if (Rs < 626) {
                                                        return 8;
                                                    }
                                                    else if (Rs >= 626) {
                                                        return 1;
                                                    }
                                                }
                                                else if (Rs >= 866) {
                                                    if (mu_mv_mag_sq < 54) {
                                                        if (CsDiff < 12) {
                                                            return 2;
                                                        }
                                                        else if (CsDiff >= 12) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (mu_mv_mag_sq >= 54) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                        }
                                        else if (negBalance >= 51) {
                                            if (mu_mv_mag_sq < 45) {
                                                if (posBalance < 28) {
                                                    return 4;
                                                }
                                                else if (posBalance >= 28) {
                                                    if (diffRsCsdiff < -32) {
                                                        if (AFD < 23) {
                                                            if (Cs < 1401) {
                                                                if (Cs < 677) {
                                                                    return 4;
                                                                }
                                                                else if (Cs >= 677) {
                                                                    if (ssDCval < 74) {
                                                                        if (Cs < 1194) {
                                                                            return 2;
                                                                        }
                                                                        else if (Cs >= 1194) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                    else if (ssDCval >= 74) {
                                                                        return 2;
                                                                    }
                                                                }
                                                            }
                                                            else if (Cs >= 1401) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (AFD >= 23) {
                                                            if (negBalance < 57) {
                                                                return 2;
                                                            }
                                                            else if (negBalance >= 57) {
                                                                return 1;
                                                            }
                                                        }
                                                    }
                                                    else if (diffRsCsdiff >= -32) {
                                                        if (MVDiff < 17) {
                                                            if (Rs < 2492) {
                                                                if (Rs < 1965) {
                                                                    if (Cs < 1388) {
                                                                        return 2;
                                                                    }
                                                                    else if (Cs >= 1388) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (Rs >= 1965) {
                                                                    return 1;
                                                                }
                                                            }
                                                            else if (Rs >= 2492) {
                                                                return 2;
                                                            }
                                                        }
                                                        else if (MVDiff >= 17) {
                                                            return 2;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (mu_mv_mag_sq >= 45) {
                                                if (AFD < 17) {
                                                    return 4;
                                                }
                                                else if (AFD >= 17) {
                                                    return 2;
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                            else if (Cs >= 1474) {
                                if (refDCval < 133) {
                                    if (refDCval < 129) {
                                        if (refDCval < 81) {
                                            if (negBalance < 45) {
                                                if (gchDC < 3) {
                                                    if (CsDiff < 9) {
                                                        if (tcor < 88) {
                                                            if (RsDiff < 8) {
                                                                if (SC < 5379) {
                                                                    return 1;
                                                                }
                                                                else if (SC >= 5379) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (RsDiff >= 8) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (tcor >= 88) {
                                                            return 1;
                                                        }
                                                    }
                                                    else if (CsDiff >= 9) {
                                                        if (ssDCval < 79) {
                                                            if (diffRsCsdiff < 79) {
                                                                return 1;
                                                            }
                                                            else if (diffRsCsdiff >= 79) {
                                                                if (AFD < 15) {
                                                                    return 8;
                                                                }
                                                                else if (AFD >= 15) {
                                                                    return 1;
                                                                }
                                                            }
                                                        }
                                                        else if (ssDCval >= 79) {
                                                            if (posBalance < 51) {
                                                                if (Rs < 2871) {
                                                                    return 1;
                                                                }
                                                                else if (Rs >= 2871) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (posBalance >= 51) {
                                                                if (Cs < 2101) {
                                                                    return 1;
                                                                }
                                                                else if (Cs >= 2101) {
                                                                    if (RsDiff < 12) {
                                                                        return 8;
                                                                    }
                                                                    else if (RsDiff >= 12) {
                                                                        return 1;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (gchDC >= 3) {
                                                    if (tcor < 85) {
                                                        if (SC < 3857) {
                                                            if (Rs < 1721) {
                                                                return 2;
                                                            }
                                                            else if (Rs >= 1721) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (SC >= 3857) {
                                                            if (Rs < 1959) {
                                                                return 1;
                                                            }
                                                            else if (Rs >= 1959) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                    else if (tcor >= 85) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                            else if (negBalance >= 45) {
                                                if (mu_mv_mag_sq < 6) {
                                                    if (MVDiff < 18) {
                                                        if (negBalance < 48) {
                                                            return 8;
                                                        }
                                                        else if (negBalance >= 48) {
                                                            if (ssDCval < 80) {
                                                                return 4;
                                                            }
                                                            else if (ssDCval >= 80) {
                                                                if (Cs < 2079) {
                                                                    return 2;
                                                                }
                                                                else if (Cs >= 2079) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (MVDiff >= 18) {
                                                        if (RsDiff < 16) {
                                                            if (mcTcor < 90) {
                                                                if (Rs < 2605) {
                                                                    if (negBalance < 52) {
                                                                        if (RsCsDiff < 92) {
                                                                            if (Rs < 1824) {
                                                                                return 4;
                                                                            }
                                                                            else if (Rs >= 1824) {
                                                                                return 8;
                                                                            }
                                                                        }
                                                                        else if (RsCsDiff >= 92) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                    else if (negBalance >= 52) {
                                                                        if (RsCsDiff < 163) {
                                                                            return 8;
                                                                        }
                                                                        else if (RsCsDiff >= 163) {
                                                                            if (SC < 4500) {
                                                                                if (Cs < 2210) {
                                                                                    return 8;
                                                                                }
                                                                                else if (Cs >= 2210) {
                                                                                    return 1;
                                                                                }
                                                                            }
                                                                            else if (SC >= 4500) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                                else if (Rs >= 2605) {
                                                                    if (MVDiff < 24) {
                                                                        return 4;
                                                                    }
                                                                    else if (MVDiff >= 24) {
                                                                        return 8;
                                                                    }
                                                                }
                                                            }
                                                            else if (mcTcor >= 90) {
                                                                if (Cs < 2295) {
                                                                    if (ssDCval < 78) {
                                                                        return 4;
                                                                    }
                                                                    else if (ssDCval >= 78) {
                                                                        return 8;
                                                                    }
                                                                }
                                                                else if (Cs >= 2295) {
                                                                    return 1;
                                                                }
                                                            }
                                                        }
                                                        else if (RsDiff >= 16) {
                                                            return 2;
                                                        }
                                                    }
                                                }
                                                else if (mu_mv_mag_sq >= 6) {
                                                    if (refDCval < 78) {
                                                        if (SC < 5964) {
                                                            if (gchDC < 2) {
                                                                if (Cs < 1646) {
                                                                    if (ssDCval < 69) {
                                                                        return 8;
                                                                    }
                                                                    else if (ssDCval >= 69) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (Cs >= 1646) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (gchDC >= 2) {
                                                                if (Cs < 1876) {
                                                                    return 4;
                                                                }
                                                                else if (Cs >= 1876) {
                                                                    if (Rs < 2069) {
                                                                        return 4;
                                                                    }
                                                                    else if (Rs >= 2069) {
                                                                        return 8;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (SC >= 5964) {
                                                            if (Rs < 3445) {
                                                                if (refDCval < 64) {
                                                                    return 8;
                                                                }
                                                                else if (refDCval >= 64) {
                                                                    return 1;
                                                                }
                                                            }
                                                            else if (Rs >= 3445) {
                                                                if (TSC < 14) {
                                                                    return 4;
                                                                }
                                                                else if (TSC >= 14) {
                                                                    return 2;
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (refDCval >= 78) {
                                                        if (mcTcor < 94) {
                                                            if (SC < 5250) {
                                                                if (tcor < 89) {
                                                                    if (posBalance < 41) {
                                                                        return 2;
                                                                    }
                                                                    else if (posBalance >= 41) {
                                                                        return 1;
                                                                    }
                                                                }
                                                                else if (tcor >= 89) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (SC >= 5250) {
                                                                if (negBalance < 51) {
                                                                    return 8;
                                                                }
                                                                else if (negBalance >= 51) {
                                                                    if (Cs < 2630) {
                                                                        return 4;
                                                                    }
                                                                    else if (Cs >= 2630) {
                                                                        return 1;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (mcTcor >= 94) {
                                                            return 2;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (refDCval >= 81) {
                                            if (MVDiff < 4) {
                                                if (Cs < 2407) {
                                                    if (negBalance < 55) {
                                                        if (AFD < 17) {
                                                            if (negBalance < 48) {
                                                                return 2;
                                                            }
                                                            else if (negBalance >= 48) {
                                                                if (negBalance < 52) {
                                                                    return 4;
                                                                }
                                                                else if (negBalance >= 52) {
                                                                    if (Rs < 2239) {
                                                                        return 2;
                                                                    }
                                                                    else if (Rs >= 2239) {
                                                                        return 4;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (AFD >= 17) {
                                                            if (Cs < 2118) {
                                                                return 2;
                                                            }
                                                            else if (Cs >= 2118) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                    else if (negBalance >= 55) {
                                                        return 2;
                                                    }
                                                }
                                                else if (Cs >= 2407) {
                                                    if (RsDiff < 12) {
                                                        if (mu_mv_mag_sq < 80) {
                                                            if (Rs < 2532) {
                                                                return 4;
                                                            }
                                                            else if (Rs >= 2532) {
                                                                if (Rs < 3879) {
                                                                    return 8;
                                                                }
                                                                else if (Rs >= 3879) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                        else if (mu_mv_mag_sq >= 80) {
                                                            return 1;
                                                        }
                                                    }
                                                    else if (RsDiff >= 12) {
                                                        if (Rs < 5495) {
                                                            return 4;
                                                        }
                                                        else if (Rs >= 5495) {
                                                            return 1;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (MVDiff >= 4) {
                                                if (Cs < 1798) {
                                                    if (refDCval < 108) {
                                                        if (mu_mv_mag_sq < 17) {
                                                            if (negBalance < 53) {
                                                                if (diffMVdiffVal < -3) {
                                                                    if (Cs < 1767) {
                                                                        return 2;
                                                                    }
                                                                    else if (Cs >= 1767) {
                                                                        if (TSC < 12) {
                                                                            return 4;
                                                                        }
                                                                        else if (TSC >= 12) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                }
                                                                else if (diffMVdiffVal >= -3) {
                                                                    return 1;
                                                                }
                                                            }
                                                            else if (negBalance >= 53) {
                                                                return 2;
                                                            }
                                                        }
                                                        else if (mu_mv_mag_sq >= 17) {
                                                            return 2;
                                                        }
                                                    }
                                                    else if (refDCval >= 108) {
                                                        if (tcor < 82) {
                                                            if (mu_mv_mag_sq < 10) {
                                                                return 4;
                                                            }
                                                            else if (mu_mv_mag_sq >= 10) {
                                                                if (diffMVdiffVal < 2) {
                                                                    if (SC < 3819) {
                                                                        if (Cs < 1583) {
                                                                            return 8;
                                                                        }
                                                                        else if (Cs >= 1583) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                    else if (SC >= 3819) {
                                                                        return 2;
                                                                    }
                                                                }
                                                                else if (diffMVdiffVal >= 2) {
                                                                    if (Rs < 2901) {
                                                                        return 4;
                                                                    }
                                                                    else if (Rs >= 2901) {
                                                                        return 2;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (tcor >= 82) {
                                                            if (CsDiff < 7) {
                                                                if (refDCval < 117) {
                                                                    if (gchDC < 1) {
                                                                        return 8;
                                                                    }
                                                                    else if (gchDC >= 1) {
                                                                        if (SC < 4809) {
                                                                            return 2;
                                                                        }
                                                                        else if (SC >= 4809) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                }
                                                                else if (refDCval >= 117) {
                                                                    if (posBalance < 63) {
                                                                        return 4;
                                                                    }
                                                                    else if (posBalance >= 63) {
                                                                        return 8;
                                                                    }
                                                                }
                                                            }
                                                            else if (CsDiff >= 7) {
                                                                if (AFD < 15) {
                                                                    if (Cs < 1727) {
                                                                        return 8;
                                                                    }
                                                                    else if (Cs >= 1727) {
                                                                        return 2;
                                                                    }
                                                                }
                                                                else if (AFD >= 15) {
                                                                    if (Rs < 1835) {
                                                                        return 8;
                                                                    }
                                                                    else if (Rs >= 1835) {
                                                                        if (MVDiff < 32) {
                                                                            return 2;
                                                                        }
                                                                        else if (MVDiff >= 32) {
                                                                            if (RsCsDiff < 194) {
                                                                                return 8;
                                                                            }
                                                                            else if (RsCsDiff >= 194) {
                                                                                return 2;
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (Cs >= 1798) {
                                                    if (SC < 7439) {
                                                        if (diffRsCsdiff < 176) {
                                                            if (mcTcor < 95) {
                                                                if (posBalance < 56) {
                                                                    if (mcTcor < 91) {
                                                                        if (RsCsDiff < 234) {
                                                                            if (ssDCval < 109) {
                                                                                if (refDCval < 87) {
                                                                                    if (AFD < 15) {
                                                                                        return 4;
                                                                                    }
                                                                                    else if (AFD >= 15) {
                                                                                        if (Cs < 2324) {
                                                                                            return 8;
                                                                                        }
                                                                                        else if (Cs >= 2324) {
                                                                                            return 1;
                                                                                        }
                                                                                    }
                                                                                }
                                                                                else if (refDCval >= 87) {
                                                                                    if (posBalance < 46) {
                                                                                        if (Rs < 2903) {
                                                                                            return 2;
                                                                                        }
                                                                                        else if (Rs >= 2903) {
                                                                                            return 8;
                                                                                        }
                                                                                    }
                                                                                    else if (posBalance >= 46) {
                                                                                        if (RsDiff < 10) {
                                                                                            return 4;
                                                                                        }
                                                                                        else if (RsDiff >= 10) {
                                                                                            if (ssDCval < 101) {
                                                                                                if (ssDCval < 92) {
                                                                                                    return 2;
                                                                                                }
                                                                                                else if (ssDCval >= 92) {
                                                                                                    if (Rs < 3530) {
                                                                                                        return 8;
                                                                                                    }
                                                                                                    else if (Rs >= 3530) {
                                                                                                        return 4;
                                                                                                    }
                                                                                                }
                                                                                            }
                                                                                            else if (ssDCval >= 101) {
                                                                                                if (SC < 7309) {
                                                                                                    return 4;
                                                                                                }
                                                                                                else if (SC >= 7309) {
                                                                                                    return 2;
                                                                                                }
                                                                                            }
                                                                                        }
                                                                                    }
                                                                                }
                                                                            }
                                                                            else if (ssDCval >= 109) {
                                                                                if (RsDiff < 11) {
                                                                                    return 1;
                                                                                }
                                                                                else if (RsDiff >= 11) {
                                                                                    return 4;
                                                                                }
                                                                            }
                                                                        }
                                                                        else if (RsCsDiff >= 234) {
                                                                            if (TSC < 12) {
                                                                                if (ssDCval < 94) {
                                                                                    if (CsDiff < 14) {
                                                                                        return 1;
                                                                                    }
                                                                                    else if (CsDiff >= 14) {
                                                                                        if (SC < 5850) {
                                                                                            return 8;
                                                                                        }
                                                                                        else if (SC >= 5850) {
                                                                                            return 1;
                                                                                        }
                                                                                    }
                                                                                }
                                                                                else if (ssDCval >= 94) {
                                                                                    if (CsDiff < 11) {
                                                                                        return 1;
                                                                                    }
                                                                                    else if (CsDiff >= 11) {
                                                                                        return 2;
                                                                                    }
                                                                                }
                                                                            }
                                                                            else if (TSC >= 12) {
                                                                                if (SC < 5599) {
                                                                                    if (negBalance < 51) {
                                                                                        return 1;
                                                                                    }
                                                                                    else if (negBalance >= 51) {
                                                                                        if (Cs < 2408) {
                                                                                            return 8;
                                                                                        }
                                                                                        else if (Cs >= 2408) {
                                                                                            return 4;
                                                                                        }
                                                                                    }
                                                                                }
                                                                                else if (SC >= 5599) {
                                                                                    if (refDCval < 113) {
                                                                                        return 8;
                                                                                    }
                                                                                    else if (refDCval >= 113) {
                                                                                        if (Cs < 3089) {
                                                                                            return 1;
                                                                                        }
                                                                                        else if (Cs >= 3089) {
                                                                                            return 8;
                                                                                        }
                                                                                    }
                                                                                }
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (mcTcor >= 91) {
                                                                        if (mcTcor < 94) {
                                                                            if (mu_mv_mag_sq < 19) {
                                                                                if (MVDiff < 12) {
                                                                                    if (MVDiff < 7) {
                                                                                        return 8;
                                                                                    }
                                                                                    else if (MVDiff >= 7) {
                                                                                        if (Rs < 2258) {
                                                                                            return 2;
                                                                                        }
                                                                                        else if (Rs >= 2258) {
                                                                                            return 4;
                                                                                        }
                                                                                    }
                                                                                }
                                                                                else if (MVDiff >= 12) {
                                                                                    if (TSC < 10) {
                                                                                        return 1;
                                                                                    }
                                                                                    else if (TSC >= 10) {
                                                                                        if (RsCsDiff < 154) {
                                                                                            if (mcTcor < 93) {
                                                                                                if (TSC < 12) {
                                                                                                    return 4;
                                                                                                }
                                                                                                else if (TSC >= 12) {
                                                                                                    if (MVDiff < 20) {
                                                                                                        if (Cs < 2112) {
                                                                                                            return 1;
                                                                                                        }
                                                                                                        else if (Cs >= 2112) {
                                                                                                            return 4;
                                                                                                        }
                                                                                                    }
                                                                                                    else if (MVDiff >= 20) {
                                                                                                        return 8;
                                                                                                    }
                                                                                                }
                                                                                            }
                                                                                            else if (mcTcor >= 93) {
                                                                                                return 1;
                                                                                            }
                                                                                        }
                                                                                        else if (RsCsDiff >= 154) {
                                                                                            if (negBalance < 44) {
                                                                                                if (mu_mv_mag_sq < 5) {
                                                                                                    return 1;
                                                                                                }
                                                                                                else if (mu_mv_mag_sq >= 5) {
                                                                                                    if (Cs < 2422) {
                                                                                                        return 4;
                                                                                                    }
                                                                                                    else if (Cs >= 2422) {
                                                                                                        return 1;
                                                                                                    }
                                                                                                }
                                                                                            }
                                                                                            else if (negBalance >= 44) {
                                                                                                return 1;
                                                                                            }
                                                                                        }
                                                                                    }
                                                                                }
                                                                            }
                                                                            else if (mu_mv_mag_sq >= 19) {
                                                                                if (Rs < 2644) {
                                                                                    if (SC < 4876) {
                                                                                        if (TSC < 11) {
                                                                                            if (negBalance < 39) {
                                                                                                if (posBalance < 53) {
                                                                                                    return 1;
                                                                                                }
                                                                                                else if (posBalance >= 53) {
                                                                                                    return 8;
                                                                                                }
                                                                                            }
                                                                                            else if (negBalance >= 39) {
                                                                                                return 1;
                                                                                            }
                                                                                        }
                                                                                        else if (TSC >= 11) {
                                                                                            if (Cs < 2177) {
                                                                                                return 8;
                                                                                            }
                                                                                            else if (Cs >= 2177) {
                                                                                                if (MVDiff < 20) {
                                                                                                    return 1;
                                                                                                }
                                                                                                else if (MVDiff >= 20) {
                                                                                                    return 4;
                                                                                                }
                                                                                            }
                                                                                        }
                                                                                    }
                                                                                    else if (SC >= 4876) {
                                                                                        return 8;
                                                                                    }
                                                                                }
                                                                                else if (Rs >= 2644) {
                                                                                    if (diffAFD < -2) {
                                                                                        if (AFD < 18) {
                                                                                            return 8;
                                                                                        }
                                                                                        else if (AFD >= 18) {
                                                                                            return 1;
                                                                                        }
                                                                                    }
                                                                                    else if (diffAFD >= -2) {
                                                                                        if (Rs < 2843) {
                                                                                            return 1;
                                                                                        }
                                                                                        else if (Rs >= 2843) {
                                                                                            if (Cs < 2549) {
                                                                                                if (mu_mv_mag_sq < 67) {
                                                                                                    if (mcTcor < 92) {
                                                                                                        return 1;
                                                                                                    }
                                                                                                    else if (mcTcor >= 92) {
                                                                                                        if (Rs < 2885) {
                                                                                                            if (Rs < 2848) {
                                                                                                                return 4;
                                                                                                            }
                                                                                                            else if (Rs >= 2848) {
                                                                                                                return 1;
                                                                                                            }
                                                                                                        }
                                                                                                        else if (Rs >= 2885) {
                                                                                                            return 4;
                                                                                                        }
                                                                                                    }
                                                                                                }
                                                                                                else if (mu_mv_mag_sq >= 67) {
                                                                                                    if (SC < 6275) {
                                                                                                        if (diffRsCsdiff < -14) {
                                                                                                            return 4;
                                                                                                        }
                                                                                                        else if (diffRsCsdiff >= -14) {
                                                                                                            return 1;
                                                                                                        }
                                                                                                    }
                                                                                                    else if (SC >= 6275) {
                                                                                                        return 8;
                                                                                                    }
                                                                                                }
                                                                                            }
                                                                                            else if (Cs >= 2549) {
                                                                                                return 1;
                                                                                            }
                                                                                        }
                                                                                    }
                                                                                }
                                                                            }
                                                                        }
                                                                        else if (mcTcor >= 94) {
                                                                            if (refDCval < 101) {
                                                                                if (MVDiff < 20) {
                                                                                    if (Cs < 2292) {
                                                                                        if (SC < 4370) {
                                                                                            if (TSC < 8) {
                                                                                                return 1;
                                                                                            }
                                                                                            else if (TSC >= 8) {
                                                                                                return 2;
                                                                                            }
                                                                                        }
                                                                                        else if (SC >= 4370) {
                                                                                            return 1;
                                                                                        }
                                                                                    }
                                                                                    else if (Cs >= 2292) {
                                                                                        if (Rs < 2394) {
                                                                                            return 4;
                                                                                        }
                                                                                        else if (Rs >= 2394) {
                                                                                            return 8;
                                                                                        }
                                                                                    }
                                                                                }
                                                                                else if (MVDiff >= 20) {
                                                                                    return 8;
                                                                                }
                                                                            }
                                                                            else if (refDCval >= 101) {
                                                                                if (Cs < 2357) {
                                                                                    if (mu_mv_mag_sq < 51) {
                                                                                        return 4;
                                                                                    }
                                                                                    else if (mu_mv_mag_sq >= 51) {
                                                                                        return 2;
                                                                                    }
                                                                                }
                                                                                else if (Cs >= 2357) {
                                                                                    return 4;
                                                                                }
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                                else if (posBalance >= 56) {
                                                                    if (Rs < 2705) {
                                                                        if (refDCval < 104) {
                                                                            if (tcor < 88) {
                                                                                if (mu_mv_mag_sq < 22) {
                                                                                    if (posBalance < 63) {
                                                                                        if (negBalance < 50) {
                                                                                            if (mcTcor < 93) {
                                                                                                return 4;
                                                                                            }
                                                                                            else if (mcTcor >= 93) {
                                                                                                return 1;
                                                                                            }
                                                                                        }
                                                                                        else if (negBalance >= 50) {
                                                                                            if (Rs < 2567) {
                                                                                                if (mcTcor < 86) {
                                                                                                    return 4;
                                                                                                }
                                                                                                else if (mcTcor >= 86) {
                                                                                                    if (MVDiff < 8) {
                                                                                                        return 4;
                                                                                                    }
                                                                                                    else if (MVDiff >= 8) {
                                                                                                        return 2;
                                                                                                    }
                                                                                                }
                                                                                            }
                                                                                            else if (Rs >= 2567) {
                                                                                                return 2;
                                                                                            }
                                                                                        }
                                                                                    }
                                                                                    else if (posBalance >= 63) {
                                                                                        if (Rs < 1833) {
                                                                                            return 1;
                                                                                        }
                                                                                        else if (Rs >= 1833) {
                                                                                            return 2;
                                                                                        }
                                                                                    }
                                                                                }
                                                                                else if (mu_mv_mag_sq >= 22) {
                                                                                    if (posBalance < 59) {
                                                                                        if (Cs < 1974) {
                                                                                            return 1;
                                                                                        }
                                                                                        else if (Cs >= 1974) {
                                                                                            return 2;
                                                                                        }
                                                                                    }
                                                                                    else if (posBalance >= 59) {
                                                                                        if (CsDiff < 8) {
                                                                                            if (Cs < 2073) {
                                                                                                return 1;
                                                                                            }
                                                                                            else if (Cs >= 2073) {
                                                                                                return 2;
                                                                                            }
                                                                                        }
                                                                                        else if (CsDiff >= 8) {
                                                                                            return 4;
                                                                                        }
                                                                                    }
                                                                                }
                                                                            }
                                                                            else if (tcor >= 88) {
                                                                                if (Rs < 2549) {
                                                                                    if (SC < 3914) {
                                                                                        return 4;
                                                                                    }
                                                                                    else if (SC >= 3914) {
                                                                                        if (mcTcor < 93) {
                                                                                            if (Cs < 2185) {
                                                                                                return 4;
                                                                                            }
                                                                                            else if (Cs >= 2185) {
                                                                                                return 2;
                                                                                            }
                                                                                        }
                                                                                        else if (mcTcor >= 93) {
                                                                                            return 2;
                                                                                        }
                                                                                    }
                                                                                }
                                                                                else if (Rs >= 2549) {
                                                                                    if (ssDCval < 85) {
                                                                                        return 2;
                                                                                    }
                                                                                    else if (ssDCval >= 85) {
                                                                                        return 1;
                                                                                    }
                                                                                }
                                                                            }
                                                                        }
                                                                        else if (refDCval >= 104) {
                                                                            if (refDCval < 125) {
                                                                                if (AFD < 13) {
                                                                                    return 4;
                                                                                }
                                                                                else if (AFD >= 13) {
                                                                                    if (TSC < 12) {
                                                                                        if (RsCsDiff < 207) {
                                                                                            return 1;
                                                                                        }
                                                                                        else if (RsCsDiff >= 207) {
                                                                                            if (TSC < 11) {
                                                                                                return 2;
                                                                                            }
                                                                                            else if (TSC >= 11) {
                                                                                                return 4;
                                                                                            }
                                                                                        }
                                                                                    }
                                                                                    else if (TSC >= 12) {
                                                                                        if (Rs < 2642) {
                                                                                            return 1;
                                                                                        }
                                                                                        else if (Rs >= 2642) {
                                                                                            if (Cs < 1941) {
                                                                                                return 4;
                                                                                            }
                                                                                            else if (Cs >= 1941) {
                                                                                                return 1;
                                                                                            }
                                                                                        }
                                                                                    }
                                                                                }
                                                                            }
                                                                            else if (refDCval >= 125) {
                                                                                if (diffAFD < -1) {
                                                                                    return 4;
                                                                                }
                                                                                else if (diffAFD >= -1) {
                                                                                    if (RsDiff < 12) {
                                                                                        return 1;
                                                                                    }
                                                                                    else if (RsDiff >= 12) {
                                                                                        if (Rs < 2483) {
                                                                                            return 2;
                                                                                        }
                                                                                        else if (Rs >= 2483) {
                                                                                            return 8;
                                                                                        }
                                                                                    }
                                                                                }
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (Rs >= 2705) {
                                                                        if (negBalance < 46) {
                                                                            if (CsDiff < 12) {
                                                                                if (SC < 5288) {
                                                                                    return 1;
                                                                                }
                                                                                else if (SC >= 5288) {
                                                                                    if (diffMVdiffVal < 3) {
                                                                                        return 4;
                                                                                    }
                                                                                    else if (diffMVdiffVal >= 3) {
                                                                                        if (Cs < 2434) {
                                                                                            return 4;
                                                                                        }
                                                                                        else if (Cs >= 2434) {
                                                                                            return 8;
                                                                                        }
                                                                                    }
                                                                                }
                                                                            }
                                                                            else if (CsDiff >= 12) {
                                                                                if (Rs < 3171) {
                                                                                    return 1;
                                                                                }
                                                                                else if (Rs >= 3171) {
                                                                                    return 2;
                                                                                }
                                                                            }
                                                                        }
                                                                        else if (negBalance >= 46) {
                                                                            if (Cs < 2895) {
                                                                                if (Cs < 2518) {
                                                                                    if (mu_mv_mag_sq < 46) {
                                                                                        if (TSC < 12) {
                                                                                            if (MVDiff < 10) {
                                                                                                if (TSC < 10) {
                                                                                                    return 2;
                                                                                                }
                                                                                                else if (TSC >= 10) {
                                                                                                    if (diffMVdiffVal < -2) {
                                                                                                        if (Cs < 1954) {
                                                                                                            return 4;
                                                                                                        }
                                                                                                        else if (Cs >= 1954) {
                                                                                                            return 1;
                                                                                                        }
                                                                                                    }
                                                                                                    else if (diffMVdiffVal >= -2) {
                                                                                                        if (Rs < 2729) {
                                                                                                            return 8;
                                                                                                        }
                                                                                                        else if (Rs >= 2729) {
                                                                                                            return 1;
                                                                                                        }
                                                                                                    }
                                                                                                }
                                                                                            }
                                                                                            else if (MVDiff >= 10) {
                                                                                                if (diffRsCsdiff < -30) {
                                                                                                    if (MVDiff < 25) {
                                                                                                        return 1;
                                                                                                    }
                                                                                                    else if (MVDiff >= 25) {
                                                                                                        return 2;
                                                                                                    }
                                                                                                }
                                                                                                else if (diffRsCsdiff >= -30) {
                                                                                                    if (diffAFD < 2) {
                                                                                                        return 2;
                                                                                                    }
                                                                                                    else if (diffAFD >= 2) {
                                                                                                        if (Rs < 2856) {
                                                                                                            return 2;
                                                                                                        }
                                                                                                        else if (Rs >= 2856) {
                                                                                                            return 1;
                                                                                                        }
                                                                                                    }
                                                                                                }
                                                                                            }
                                                                                        }
                                                                                        else if (TSC >= 12) {
                                                                                            if (SC < 5428) {
                                                                                                if (gchDC < 1) {
                                                                                                    if (SC < 5250) {
                                                                                                        if (refDCval < 120) {
                                                                                                            return 8;
                                                                                                        }
                                                                                                        else if (refDCval >= 120) {
                                                                                                            return 4;
                                                                                                        }
                                                                                                    }
                                                                                                    else if (SC >= 5250) {
                                                                                                        if (SC < 5343) {
                                                                                                            return 1;
                                                                                                        }
                                                                                                        else if (SC >= 5343) {
                                                                                                            return 4;
                                                                                                        }
                                                                                                    }
                                                                                                }
                                                                                                else if (gchDC >= 1) {
                                                                                                    if (ssDCval < 116) {
                                                                                                        if (Rs < 3133) {
                                                                                                            return 1;
                                                                                                        }
                                                                                                        else if (Rs >= 3133) {
                                                                                                            return 2;
                                                                                                        }
                                                                                                    }
                                                                                                    else if (ssDCval >= 116) {
                                                                                                        if (diffMVdiffVal < -2) {
                                                                                                            return 1;
                                                                                                        }
                                                                                                        else if (diffMVdiffVal >= -2) {
                                                                                                            return 4;
                                                                                                        }
                                                                                                    }
                                                                                                }
                                                                                            }
                                                                                            else if (SC >= 5428) {
                                                                                                return 1;
                                                                                            }
                                                                                        }
                                                                                    }
                                                                                    else if (mu_mv_mag_sq >= 46) {
                                                                                        if (MVDiff < 10) {
                                                                                            if (Cs < 2335) {
                                                                                                return 8;
                                                                                            }
                                                                                            else if (Cs >= 2335) {
                                                                                                if (ssDCval < 85) {
                                                                                                    return 8;
                                                                                                }
                                                                                                else if (ssDCval >= 85) {
                                                                                                    if (MVDiff < 5) {
                                                                                                        return 8;
                                                                                                    }
                                                                                                    else if (MVDiff >= 5) {
                                                                                                        return 4;
                                                                                                    }
                                                                                                }
                                                                                            }
                                                                                        }
                                                                                        else if (MVDiff >= 10) {
                                                                                            if (MVDiff < 13) {
                                                                                                if (Rs < 3009) {
                                                                                                    if (diffMVdiffVal < -3) {
                                                                                                        return 1;
                                                                                                    }
                                                                                                    else if (diffMVdiffVal >= -3) {
                                                                                                        if (gchDC < 1) {
                                                                                                            if (mu_mv_mag_sq < 56) {
                                                                                                                if (CsDiff < 8) {
                                                                                                                    return 1;
                                                                                                                }
                                                                                                                else if (CsDiff >= 8) {
                                                                                                                    return 4;
                                                                                                                }
                                                                                                            }
                                                                                                            else if (mu_mv_mag_sq >= 56) {
                                                                                                                return 4;
                                                                                                            }
                                                                                                        }
                                                                                                        else if (gchDC >= 1) {
                                                                                                            return 1;
                                                                                                        }
                                                                                                    }
                                                                                                }
                                                                                                else if (Rs >= 3009) {
                                                                                                    return 1;
                                                                                                }
                                                                                            }
                                                                                            else if (MVDiff >= 13) {
                                                                                                return 1;
                                                                                            }
                                                                                        }
                                                                                    }
                                                                                }
                                                                                else if (Cs >= 2518) {
                                                                                    if (Rs < 2932) {
                                                                                        if (Rs < 2812) {
                                                                                            return 4;
                                                                                        }
                                                                                        else if (Rs >= 2812) {
                                                                                            return 8;
                                                                                        }
                                                                                    }
                                                                                    else if (Rs >= 2932) {
                                                                                        if (posBalance < 57) {
                                                                                            if (gchDC < 1) {
                                                                                                return 4;
                                                                                            }
                                                                                            else if (gchDC >= 1) {
                                                                                                if (Rs < 3720) {
                                                                                                    return 1;
                                                                                                }
                                                                                                else if (Rs >= 3720) {
                                                                                                    return 4;
                                                                                                }
                                                                                            }
                                                                                        }
                                                                                        else if (posBalance >= 57) {
                                                                                            if (diffMVdiffVal < 2) {
                                                                                                return 1;
                                                                                            }
                                                                                            else if (diffMVdiffVal >= 2) {
                                                                                                if (MVDiff < 16) {
                                                                                                    return 1;
                                                                                                }
                                                                                                else if (MVDiff >= 16) {
                                                                                                    if (Cs < 2632) {
                                                                                                        return 1;
                                                                                                    }
                                                                                                    else if (Cs >= 2632) {
                                                                                                        return 4;
                                                                                                    }
                                                                                                }
                                                                                            }
                                                                                        }
                                                                                    }
                                                                                }
                                                                            }
                                                                            else if (Cs >= 2895) {
                                                                                if (gchDC < 2) {
                                                                                    if (CsDiff < 12) {
                                                                                        if (tcor < 86) {
                                                                                            return 4;
                                                                                        }
                                                                                        else if (tcor >= 86) {
                                                                                            if (Rs < 3322) {
                                                                                                return 4;
                                                                                            }
                                                                                            else if (Rs >= 3322) {
                                                                                                return 8;
                                                                                            }
                                                                                        }
                                                                                    }
                                                                                    else if (CsDiff >= 12) {
                                                                                        if (SC < 7306) {
                                                                                            return 8;
                                                                                        }
                                                                                        else if (SC >= 7306) {
                                                                                            return 1;
                                                                                        }
                                                                                    }
                                                                                }
                                                                                else if (gchDC >= 2) {
                                                                                    if (tcor < 74) {
                                                                                        if (Rs < 3809) {
                                                                                            return 4;
                                                                                        }
                                                                                        else if (Rs >= 3809) {
                                                                                            return 2;
                                                                                        }
                                                                                    }
                                                                                    else if (tcor >= 74) {
                                                                                        return 1;
                                                                                    }
                                                                                }
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                            else if (mcTcor >= 95) {
                                                                if (refDCval < 92) {
                                                                    if (Rs < 2098) {
                                                                        if (AFD < 15) {
                                                                            return 8;
                                                                        }
                                                                        else if (AFD >= 15) {
                                                                            return 2;
                                                                        }
                                                                    }
                                                                    else if (Rs >= 2098) {
                                                                        if (MVDiff < 16) {
                                                                            if (CsDiff < 9) {
                                                                                return 1;
                                                                            }
                                                                            else if (CsDiff >= 9) {
                                                                                return 2;
                                                                            }
                                                                        }
                                                                        else if (MVDiff >= 16) {
                                                                            return 2;
                                                                        }
                                                                    }
                                                                }
                                                                else if (refDCval >= 92) {
                                                                    if (CsDiff < 9) {
                                                                        if (refDCval < 105) {
                                                                            if (ssDCval < 108) {
                                                                                return 4;
                                                                            }
                                                                            else if (ssDCval >= 108) {
                                                                                return 2;
                                                                            }
                                                                        }
                                                                        else if (refDCval >= 105) {
                                                                            if (SC < 5443) {
                                                                                return 2;
                                                                            }
                                                                            else if (SC >= 5443) {
                                                                                return 8;
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (CsDiff >= 9) {
                                                                        return 2;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (diffRsCsdiff >= 176) {
                                                            if (tcor < 71) {
                                                                return 8;
                                                            }
                                                            else if (tcor >= 71) {
                                                                if (Cs < 2337) {
                                                                    return 4;
                                                                }
                                                                else if (Cs >= 2337) {
                                                                    if (Rs < 2864) {
                                                                        return 1;
                                                                    }
                                                                    else if (Rs >= 2864) {
                                                                        return 8;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (SC >= 7439) {
                                                        if (ssDCval < 112) {
                                                            if (ssDCval < 107) {
                                                                if (CsDiff < 17) {
                                                                    if (MVDiff < 13) {
                                                                        if (diffRsCsdiff < 4) {
                                                                            if (Rs < 5353) {
                                                                                return 2;
                                                                            }
                                                                            else if (Rs >= 5353) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                        else if (diffRsCsdiff >= 4) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                    else if (MVDiff >= 13) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (CsDiff >= 17) {
                                                                    return 2;
                                                                }
                                                            }
                                                            else if (ssDCval >= 107) {
                                                                if (TSC < 11) {
                                                                    if (diffTSC < -1) {
                                                                        return 2;
                                                                    }
                                                                    else if (diffTSC >= -1) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (TSC >= 11) {
                                                                    return 2;
                                                                }
                                                            }
                                                        }
                                                        else if (ssDCval >= 112) {
                                                            if (tcor < 78) {
                                                                return 1;
                                                            }
                                                            else if (tcor >= 78) {
                                                                if (Rs < 4681) {
                                                                    return 2;
                                                                }
                                                                else if (Rs >= 4681) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (refDCval >= 129) {
                                        if (tcor < 86) {
                                            if (mcTcor < 85) {
                                                if (AFD < 34) {
                                                    if (diffMVdiffVal < -53) {
                                                        return 4;
                                                    }
                                                    else if (diffMVdiffVal >= -53) {
                                                        return 1;
                                                    }
                                                }
                                                else if (AFD >= 34) {
                                                    return 8;
                                                }
                                            }
                                            else if (mcTcor >= 85) {
                                                if (RsCsDiff < 255) {
                                                    if (MVDiff < 8) {
                                                        return 1;
                                                    }
                                                    else if (MVDiff >= 8) {
                                                        return 4;
                                                    }
                                                }
                                                else if (RsCsDiff >= 255) {
                                                    if (tcor < 77) {
                                                        if (AFD < 24) {
                                                            return 1;
                                                        }
                                                        else if (AFD >= 24) {
                                                            return 2;
                                                        }
                                                    }
                                                    else if (tcor >= 77) {
                                                        return 2;
                                                    }
                                                }
                                            }
                                        }
                                        else if (tcor >= 86) {
                                            if (TSC < 9) {
                                                return 2;
                                            }
                                            else if (TSC >= 9) {
                                                return 8;
                                            }
                                        }
                                    }
                                }
                                else if (refDCval >= 133) {
                                    if (negBalance < 58) {
                                        if (gchDC < 3) {
                                            if (tcor < 86) {
                                                return 1;
                                            }
                                            else if (tcor >= 86) {
                                                return 8;
                                            }
                                        }
                                        else if (gchDC >= 3) {
                                            return 2;
                                        }
                                    }
                                    else if (negBalance >= 58) {
                                        if (RsCsDiff < 280) {
                                            if (Rs < 3331) {
                                                return 1;
                                            }
                                            else if (Rs >= 3331) {
                                                return 8;
                                            }
                                        }
                                        else if (RsCsDiff >= 280) {
                                            if (posBalance < 59) {
                                                return 1;
                                            }
                                            else if (posBalance >= 59) {
                                                return 4;
                                            }
                                        }
                                    }
                                }
                            }
                        }
                        else if (refDCval >= 139) {
                            if (Cs < 1401) {
                                if (Rs < 1433) {
                                    if (RsCsDiff < 18) {
                                        if (Rs < 662) {
                                            return 2;
                                        }
                                        else if (Rs >= 662) {
                                            return 4;
                                        }
                                    }
                                    else if (RsCsDiff >= 18) {
                                        if (refDCval < 144) {
                                            return 8;
                                        }
                                        else if (refDCval >= 144) {
                                            if (CsDiff < 4) {
                                                return 8;
                                            }
                                            else if (CsDiff >= 4) {
                                                if (posBalance < 53) {
                                                    return 1;
                                                }
                                                else if (posBalance >= 53) {
                                                    if (AFD < 27) {
                                                        return 2;
                                                    }
                                                    else if (AFD >= 27) {
                                                        return 1;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (Rs >= 1433) {
                                    if (TSC < 24) {
                                        return 2;
                                    }
                                    else if (TSC >= 24) {
                                        return 1;
                                    }
                                }
                            }
                            else if (Cs >= 1401) {
                                if (TSC < 11) {
                                    if (RsDiff < 12) {
                                        if (ssDCval < 170) {
                                            return 8;
                                        }
                                        else if (ssDCval >= 170) {
                                            if (diffMVdiffVal < -8) {
                                                return 4;
                                            }
                                            else if (diffMVdiffVal >= -8) {
                                                if (Cs < 1756) {
                                                    return 2;
                                                }
                                                else if (Cs >= 1756) {
                                                    return 4;
                                                }
                                            }
                                        }
                                    }
                                    else if (RsDiff >= 12) {
                                        if (posBalance < 54) {
                                            if (Rs < 3701) {
                                                if (TSC < 10) {
                                                    return 2;
                                                }
                                                else if (TSC >= 10) {
                                                    return 4;
                                                }
                                            }
                                            else if (Rs >= 3701) {
                                                return 4;
                                            }
                                        }
                                        else if (posBalance >= 54) {
                                            return 2;
                                        }
                                    }
                                }
                                else if (TSC >= 11) {
                                    if (ssDCval < 144) {
                                        if (Cs < 2538) {
                                            if (posBalance < 47) {
                                                if (Rs < 1254) {
                                                    return 1;
                                                }
                                                else if (Rs >= 1254) {
                                                    return 8;
                                                }
                                            }
                                            else if (posBalance >= 47) {
                                                if (gchDC < 3) {
                                                    return 4;
                                                }
                                                else if (gchDC >= 3) {
                                                    return 8;
                                                }
                                            }
                                        }
                                        else if (Cs >= 2538) {
                                            if (TSC < 24) {
                                                return 1;
                                            }
                                            else if (TSC >= 24) {
                                                return 8;
                                            }
                                        }
                                    }
                                    else if (ssDCval >= 144) {
                                        if (SC < 7591) {
                                            if (Cs < 2171) {
                                                if (SC < 5568) {
                                                    if (Rs < 2915) {
                                                        return 4;
                                                    }
                                                    else if (Rs >= 2915) {
                                                        return 2;
                                                    }
                                                }
                                                else if (SC >= 5568) {
                                                    return 4;
                                                }
                                            }
                                            else if (Cs >= 2171) {
                                                if (gchDC < 3) {
                                                    return 2;
                                                }
                                                else if (gchDC >= 3) {
                                                    return 8;
                                                }
                                            }
                                        }
                                        else if (SC >= 7591) {
                                            if (refDCval < 151) {
                                                return 2;
                                            }
                                            else if (refDCval >= 151) {
                                                return 4;
                                            }
                                        }
                                    }
                                }
                            }
                        }
                    }
                }
                else if (Cs >= 3269) {
                    if (ssDCval < 75) {
                        if (mcTcor < 81) {
                            if (tcor < 42) {
                                return 4;
                            }
                            else if (tcor >= 42) {
                                if (Rs < 7452) {
                                    return 8;
                                }
                                else if (Rs >= 7452) {
                                    return 4;
                                }
                            }
                        }
                        else if (mcTcor >= 81) {
                            if (RsCsDiff < 46) {
                                if (mu_mv_mag_sq < 48) {
                                    if (TSC < 7) {
                                        if (MVDiff < 21) {
                                            if (refDCval < 55) {
                                                if (negBalance < 51) {
                                                    if (posBalance < 57) {
                                                        return 4;
                                                    }
                                                    else if (posBalance >= 57) {
                                                        if (Cs < 3360) {
                                                            return 4;
                                                        }
                                                        else if (Cs >= 3360) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                                else if (negBalance >= 51) {
                                                    if (Cs < 3669) {
                                                        return 4;
                                                    }
                                                    else if (Cs >= 3669) {
                                                        return 2;
                                                    }
                                                }
                                            }
                                            else if (refDCval >= 55) {
                                                if (ssDCval < 62) {
                                                    return 8;
                                                }
                                                else if (ssDCval >= 62) {
                                                    return 4;
                                                }
                                            }
                                        }
                                        else if (MVDiff >= 21) {
                                            if (mu_mv_mag_sq < 36) {
                                                return 8;
                                            }
                                            else if (mu_mv_mag_sq >= 36) {
                                                if (Rs < 3786) {
                                                    return 4;
                                                }
                                                else if (Rs >= 3786) {
                                                    return 8;
                                                }
                                            }
                                        }
                                    }
                                    else if (TSC >= 7) {
                                        if (MVDiff < 22) {
                                            if (Cs < 3786) {
                                                if (SC < 7304) {
                                                    if (diffAFD < 1) {
                                                        return 2;
                                                    }
                                                    else if (diffAFD >= 1) {
                                                        if (Rs < 3507) {
                                                            return 4;
                                                        }
                                                        else if (Rs >= 3507) {
                                                            if (RsDiff < 4) {
                                                                return 2;
                                                            }
                                                            else if (RsDiff >= 4) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (SC >= 7304) {
                                                    return 4;
                                                }
                                            }
                                            else if (Cs >= 3786) {
                                                return 2;
                                            }
                                        }
                                        else if (MVDiff >= 22) {
                                            if (tcor < 87) {
                                                if (ssDCval < 53) {
                                                    return 4;
                                                }
                                                else if (ssDCval >= 53) {
                                                    return 8;
                                                }
                                            }
                                            else if (tcor >= 87) {
                                                if (ssDCval < 47) {
                                                    return 8;
                                                }
                                                else if (ssDCval >= 47) {
                                                    if (negBalance < 43) {
                                                        return 8;
                                                    }
                                                    else if (negBalance >= 43) {
                                                        if (posBalance < 52) {
                                                            return 8;
                                                        }
                                                        else if (posBalance >= 52) {
                                                            if (refDCval < 58) {
                                                                if (mcTcor < 94) {
                                                                    return 2;
                                                                }
                                                                else if (mcTcor >= 94) {
                                                                    if (CsDiff < 3) {
                                                                        return 8;
                                                                    }
                                                                    else if (CsDiff >= 3) {
                                                                        if (MVDiff < 26) {
                                                                            if (RsDiff < 4) {
                                                                                return 8;
                                                                            }
                                                                            else if (RsDiff >= 4) {
                                                                                return 2;
                                                                            }
                                                                        }
                                                                        else if (MVDiff >= 26) {
                                                                            return 2;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                            else if (refDCval >= 58) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (mu_mv_mag_sq >= 48) {
                                    if (AFD < 6) {
                                        return 8;
                                    }
                                    else if (AFD >= 6) {
                                        if (mcTcor < 97) {
                                            if (ssDCval < 54) {
                                                if (Rs < 3960) {
                                                    if (Cs < 3354) {
                                                        return 8;
                                                    }
                                                    else if (Cs >= 3354) {
                                                        return 4;
                                                    }
                                                }
                                                else if (Rs >= 3960) {
                                                    return 8;
                                                }
                                            }
                                            else if (ssDCval >= 54) {
                                                if (Cs < 3627) {
                                                    if (RsCsDiff < 11) {
                                                        if (Rs < 3722) {
                                                            return 4;
                                                        }
                                                        else if (Rs >= 3722) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (RsCsDiff >= 11) {
                                                        return 8;
                                                    }
                                                }
                                                else if (Cs >= 3627) {
                                                    if (TSC < 7) {
                                                        if (RsDiff < 4) {
                                                            if (SC < 7580) {
                                                                if (negBalance < 46) {
                                                                    return 4;
                                                                }
                                                                else if (negBalance >= 46) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (SC >= 7580) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (RsDiff >= 4) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (TSC >= 7) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                        }
                                        else if (mcTcor >= 97) {
                                            return 8;
                                        }
                                    }
                                }
                            }
                            else if (RsCsDiff >= 46) {
                                if (TSC < 7) {
                                    return 8;
                                }
                                else if (TSC >= 7) {
                                    if (ssDCval < 68) {
                                        if (MVDiff < 33) {
                                            if (gchDC < 3) {
                                                if (negBalance < 48) {
                                                    if (Rs < 4037) {
                                                        return 8;
                                                    }
                                                    else if (Rs >= 4037) {
                                                        return 4;
                                                    }
                                                }
                                                else if (negBalance >= 48) {
                                                    return 4;
                                                }
                                            }
                                            else if (gchDC >= 3) {
                                                if (Rs < 4501) {
                                                    return 8;
                                                }
                                                else if (Rs >= 4501) {
                                                    return 4;
                                                }
                                            }
                                        }
                                        else if (MVDiff >= 33) {
                                            if (SC < 8723) {
                                                if (Rs < 3652) {
                                                    if (Cs < 3518) {
                                                        return 8;
                                                    }
                                                    else if (Cs >= 3518) {
                                                        return 4;
                                                    }
                                                }
                                                else if (Rs >= 3652) {
                                                    return 8;
                                                }
                                            }
                                            else if (SC >= 8723) {
                                                return 4;
                                            }
                                        }
                                    }
                                    else if (ssDCval >= 68) {
                                        if (Rs < 6474) {
                                            if (posBalance < 54) {
                                                if (MVDiff < 20) {
                                                    return 4;
                                                }
                                                else if (MVDiff >= 20) {
                                                    return 8;
                                                }
                                            }
                                            else if (posBalance >= 54) {
                                                return 8;
                                            }
                                        }
                                        else if (Rs >= 6474) {
                                            if (Cs < 6980) {
                                                return 4;
                                            }
                                            else if (Cs >= 6980) {
                                                return 8;
                                            }
                                        }
                                    }
                                }
                            }
                        }
                    }
                    else if (ssDCval >= 75) {
                        if (ssDCval < 149) {
                            if (Rs < 6893) {
                                if (tcor < 62) {
                                    if (RsDiff < 17) {
                                        if (MVDiff < 39) {
                                            if (gchDC < 2) {
                                                return 1;
                                            }
                                            else if (gchDC >= 2) {
                                                return 8;
                                            }
                                        }
                                        else if (MVDiff >= 39) {
                                            return 4;
                                        }
                                    }
                                    else if (RsDiff >= 17) {
                                        if (Cs < 4635) {
                                            if (ssDCval < 138) {
                                                if (posBalance < 67) {
                                                    if (diffRsCsdiff < -52) {
                                                        if (Cs < 4472) {
                                                            return 2;
                                                        }
                                                        else if (Cs >= 4472) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (diffRsCsdiff >= -52) {
                                                        return 8;
                                                    }
                                                }
                                                else if (posBalance >= 67) {
                                                    if (Rs < 4237) {
                                                        return 4;
                                                    }
                                                    else if (Rs >= 4237) {
                                                        return 2;
                                                    }
                                                }
                                            }
                                            else if (ssDCval >= 138) {
                                                if (SC < 9255) {
                                                    return 8;
                                                }
                                                else if (SC >= 9255) {
                                                    return 4;
                                                }
                                            }
                                        }
                                        else if (Cs >= 4635) {
                                            return 8;
                                        }
                                    }
                                }
                                else if (tcor >= 62) {
                                    if (Cs < 4250) {
                                        if (mcTcor < 84) {
                                            if (refDCval < 137) {
                                                if (posBalance < 50) {
                                                    if (Rs < 4843) {
                                                        if (Rs < 4137) {
                                                            return 8;
                                                        }
                                                        else if (Rs >= 4137) {
                                                            return 1;
                                                        }
                                                    }
                                                    else if (Rs >= 4843) {
                                                        return 2;
                                                    }
                                                }
                                                else if (posBalance >= 50) {
                                                    if (posBalance < 61) {
                                                        if (RsDiff < 16) {
                                                            return 4;
                                                        }
                                                        else if (RsDiff >= 16) {
                                                            if (ssDCval < 132) {
                                                                if (diffAFD < -1) {
                                                                    return 4;
                                                                }
                                                                else if (diffAFD >= -1) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (ssDCval >= 132) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                    else if (posBalance >= 61) {
                                                        if (Rs < 4153) {
                                                            return 1;
                                                        }
                                                        else if (Rs >= 4153) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (refDCval >= 137) {
                                                if (Rs < 4700) {
                                                    return 1;
                                                }
                                                else if (Rs >= 4700) {
                                                    return 8;
                                                }
                                            }
                                        }
                                        else if (mcTcor >= 84) {
                                            if (RsCsDiff < 892) {
                                                if (SC < 6656) {
                                                    return 1;
                                                }
                                                else if (SC >= 6656) {
                                                    if (ssDCval < 112) {
                                                        if (RsCsDiff < 289) {
                                                            if (Rs < 4865) {
                                                                if (Rs < 4285) {
                                                                    return 8;
                                                                }
                                                                else if (Rs >= 4285) {
                                                                    if (Rs < 4533) {
                                                                        return 4;
                                                                    }
                                                                    else if (Rs >= 4533) {
                                                                        return 8;
                                                                    }
                                                                }
                                                            }
                                                            else if (Rs >= 4865) {
                                                                return 2;
                                                            }
                                                        }
                                                        else if (RsCsDiff >= 289) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (ssDCval >= 112) {
                                                        if (diffRsCsdiff < 146) {
                                                            return 8;
                                                        }
                                                        else if (diffRsCsdiff >= 146) {
                                                            if (Rs < 4242) {
                                                                return 1;
                                                            }
                                                            else if (Rs >= 4242) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                            else if (RsCsDiff >= 892) {
                                                if (Cs < 3638) {
                                                    return 2;
                                                }
                                                else if (Cs >= 3638) {
                                                    if (Cs < 4078) {
                                                        return 4;
                                                    }
                                                    else if (Cs >= 4078) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (Cs >= 4250) {
                                        if (mcTcor < 93) {
                                            if (Rs < 6443) {
                                                if (Rs < 4468) {
                                                    if (mcTcor < 73) {
                                                        return 1;
                                                    }
                                                    else if (mcTcor >= 73) {
                                                        return 8;
                                                    }
                                                }
                                                else if (Rs >= 4468) {
                                                    return 8;
                                                }
                                            }
                                            else if (Rs >= 6443) {
                                                if (diffAFD < -3) {
                                                    return 4;
                                                }
                                                else if (diffAFD >= -3) {
                                                    if (mcTcor < 90) {
                                                        return 8;
                                                    }
                                                    else if (mcTcor >= 90) {
                                                        if (Cs < 5820) {
                                                            if (Rs < 6509) {
                                                                return 2;
                                                            }
                                                            else if (Rs >= 6509) {
                                                                return 1;
                                                            }
                                                        }
                                                        else if (Cs >= 5820) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (mcTcor >= 93) {
                                            if (SC < 10077) {
                                                return 8;
                                            }
                                            else if (SC >= 10077) {
                                                return 1;
                                            }
                                        }
                                    }
                                }
                            }
                            else if (Rs >= 6893) {
                                if (mu_mv_mag_sq < 63) {
                                    if (MVDiff < 23) {
                                        if (posBalance < 57) {
                                            if (SC < 15346) {
                                                if (mu_mv_mag_sq < 18) {
                                                    if (mcTcor < 91) {
                                                        if (diffMVdiffVal < 3) {
                                                            if (Rs < 8229) {
                                                                return 8;
                                                            }
                                                            else if (Rs >= 8229) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (diffMVdiffVal >= 3) {
                                                            if (AFD < 29) {
                                                                return 4;
                                                            }
                                                            else if (AFD >= 29) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                    else if (mcTcor >= 91) {
                                                        return 4;
                                                    }
                                                }
                                                else if (mu_mv_mag_sq >= 18) {
                                                    if (mu_mv_mag_sq < 56) {
                                                        return 4;
                                                    }
                                                    else if (mu_mv_mag_sq >= 56) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                            else if (SC >= 15346) {
                                                return 8;
                                            }
                                        }
                                        else if (posBalance >= 57) {
                                            if (Cs < 8119) {
                                                if (CsDiff < 21) {
                                                    return 4;
                                                }
                                                else if (CsDiff >= 21) {
                                                    if (Rs < 8006) {
                                                        return 8;
                                                    }
                                                    else if (Rs >= 8006) {
                                                        if (posBalance < 59) {
                                                            if (diffAFD < -2) {
                                                                return 4;
                                                            }
                                                            else if (diffAFD >= -2) {
                                                                return 2;
                                                            }
                                                        }
                                                        else if (posBalance >= 59) {
                                                            if (Rs < 11000) {
                                                                return 8;
                                                            }
                                                            else if (Rs >= 11000) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                            else if (Cs >= 8119) {
                                                if (SC < 27093) {
                                                    return 4;
                                                }
                                                else if (SC >= 27093) {
                                                    return 8;
                                                }
                                            }
                                        }
                                    }
                                    else if (MVDiff >= 23) {
                                        return 2;
                                    }
                                }
                                else if (mu_mv_mag_sq >= 63) {
                                    if (MVDiff < 1) {
                                        return 4;
                                    }
                                    else if (MVDiff >= 1) {
                                        return 8;
                                    }
                                }
                            }
                        }
                        else if (ssDCval >= 149) {
                            if (mcTcor < 85) {
                                if (negBalance < 55) {
                                    return 8;
                                }
                                else if (negBalance >= 55) {
                                    if (refDCval < 154) {
                                        if (Cs < 5951) {
                                            return 8;
                                        }
                                        else if (Cs >= 5951) {
                                            return 4;
                                        }
                                    }
                                    else if (refDCval >= 154) {
                                        return 2;
                                    }
                                }
                            }
                            else if (mcTcor >= 85) {
                                if (posBalance < 58) {
                                    if (tcor < 70) {
                                        if (AFD < 35) {
                                            return 4;
                                        }
                                        else if (AFD >= 35) {
                                            return 2;
                                        }
                                    }
                                    else if (tcor >= 70) {
                                        if (TSC < 19) {
                                            return 8;
                                        }
                                        else if (TSC >= 19) {
                                            if (Rs < 8165) {
                                                return 4;
                                            }
                                            else if (Rs >= 8165) {
                                                return 2;
                                            }
                                        }
                                    }
                                }
                                else if (posBalance >= 58) {
                                    if (refDCval < 159) {
                                        if (RsCsDiff < 657) {
                                            if (Rs < 4848) {
                                                return 8;
                                            }
                                            else if (Rs >= 4848) {
                                                return 4;
                                            }
                                        }
                                        else if (RsCsDiff >= 657) {
                                            if (Rs < 6025) {
                                                if (Rs < 5547) {
                                                    return 8;
                                                }
                                                else if (Rs >= 5547) {
                                                    return 4;
                                                }
                                            }
                                            else if (Rs >= 6025) {
                                                if (Cs < 6202) {
                                                    if (mcTcor < 87) {
                                                        if (Rs < 7761) {
                                                            return 2;
                                                        }
                                                        else if (Rs >= 7761) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (mcTcor >= 87) {
                                                        return 2;
                                                    }
                                                }
                                                else if (Cs >= 6202) {
                                                    if (Rs < 11127) {
                                                        if (Cs < 6329) {
                                                            return 4;
                                                        }
                                                        else if (Cs >= 6329) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (Rs >= 11127) {
                                                        return 2;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (refDCval >= 159) {
                                        return 2;
                                    }
                                }
                            }
                        }
                    }
                }
            }
        }
        else if (mu_mv_mag_sq >= 88) {
            if (posBalance < 51) {
                if (SC < 1237) {
                    if (Rs < 590) {
                        if (ssDCval < 162) {
                            if (SC < 895) {
                                if (mu_mv_mag_sq < 129) {
                                    if (negBalance < 20) {
                                        if (SC < 447) {
                                            return 1;
                                        }
                                        else if (SC >= 447) {
                                            if (posBalance < 38) {
                                                if (posBalance < 36) {
                                                    if (Cs < 339) {
                                                        return 8;
                                                    }
                                                    else if (Cs >= 339) {
                                                        if (refDCval < 155) {
                                                            return 8;
                                                        }
                                                        else if (refDCval >= 155) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                                else if (posBalance >= 36) {
                                                    return 4;
                                                }
                                            }
                                            else if (posBalance >= 38) {
                                                return 8;
                                            }
                                        }
                                    }
                                    else if (negBalance >= 20) {
                                        if (SC < 568) {
                                            if (mcTcor < 120) {
                                                return 4;
                                            }
                                            else if (mcTcor >= 120) {
                                                return 8;
                                            }
                                        }
                                        else if (SC >= 568) {
                                            if (CsDiff < 1) {
                                                return 8;
                                            }
                                            else if (CsDiff >= 1) {
                                                return 2;
                                            }
                                        }
                                    }
                                }
                                else if (mu_mv_mag_sq >= 129) {
                                    if (mu_mv_mag_sq < 290) {
                                        return 2;
                                    }
                                    else if (mu_mv_mag_sq >= 290) {
                                        return 1;
                                    }
                                }
                            }
                            else if (SC >= 895) {
                                if (RsCsDiff < 1) {
                                    if (Cs < 423) {
                                        return 8;
                                    }
                                    else if (Cs >= 423) {
                                        return 4;
                                    }
                                }
                                else if (RsCsDiff >= 1) {
                                    if (tcor < 93) {
                                        if (TSC < 4) {
                                            return 8;
                                        }
                                        else if (TSC >= 4) {
                                            return 1;
                                        }
                                    }
                                    else if (tcor >= 93) {
                                        if (TSCindex < 4) {
                                            if (SC < 1057) {
                                                return 8;
                                            }
                                            else if (SC >= 1057) {
                                                return 4;
                                            }
                                        }
                                        else if (TSCindex >= 4) {
                                            return 4;
                                        }
                                    }
                                }
                            }
                        }
                        else if (ssDCval >= 162) {
                            if (SC < 879) {
                                return 4;
                            }
                            else if (SC >= 879) {
                                if (gchDC < 1) {
                                    if (diffTSC < 1) {
                                        if (Rs < 540) {
                                            return 2;
                                        }
                                        else if (Rs >= 540) {
                                            return 4;
                                        }
                                    }
                                    else if (diffTSC >= 1) {
                                        return 4;
                                    }
                                }
                                else if (gchDC >= 1) {
                                    return 2;
                                }
                            }
                        }
                    }
                    else if (Rs >= 590) {
                        return 8;
                    }
                }
                else if (SC >= 1237) {
                    if (Rs < 3779) {
                        if (SC < 7308) {
                            if (MVDiff < 3) {
                                if (ssDCval < 65) {
                                    if (ssDCval < 28) {
                                        if (Cs < 713) {
                                            return 4;
                                        }
                                        else if (Cs >= 713) {
                                            return 8;
                                        }
                                    }
                                    else if (ssDCval >= 28) {
                                        if (RsDiff < 4) {
                                            return 8;
                                        }
                                        else if (RsDiff >= 4) {
                                            if (Rs < 2215) {
                                                if (negBalance < 20) {
                                                    return 8;
                                                }
                                                else if (negBalance >= 20) {
                                                    if (negBalance < 23) {
                                                        return 4;
                                                    }
                                                    else if (negBalance >= 23) {
                                                        if (AFD < 5) {
                                                            return 8;
                                                        }
                                                        else if (AFD >= 5) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (Rs >= 2215) {
                                                return 8;
                                            }
                                        }
                                    }
                                }
                                else if (ssDCval >= 65) {
                                    if (RsCsDiff < 7) {
                                        if (mu_mv_mag_sq < 119) {
                                            if (tcor < 99) {
                                                if (AFD < 4) {
                                                    if (posBalance < 11) {
                                                        if (refDCval < 116) {
                                                            if (Rs < 2424) {
                                                                return 4;
                                                            }
                                                            else if (Rs >= 2424) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (refDCval >= 116) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (posBalance >= 11) {
                                                        if (diffTSC < -1) {
                                                            if (Rs < 996) {
                                                                return 1;
                                                            }
                                                            else if (Rs >= 996) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (diffTSC >= -1) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                                else if (AFD >= 4) {
                                                    return 4;
                                                }
                                            }
                                            else if (tcor >= 99) {
                                                return 8;
                                            }
                                        }
                                        else if (mu_mv_mag_sq >= 119) {
                                            if (Rs < 2165) {
                                                if (Rs < 1699) {
                                                    if (diffMVdiffVal < -2) {
                                                        if (Rs < 1199) {
                                                            if (Cs < 731) {
                                                                return 1;
                                                            }
                                                            else if (Cs >= 731) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (Rs >= 1199) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (diffMVdiffVal >= -2) {
                                                        return 8;
                                                    }
                                                }
                                                else if (Rs >= 1699) {
                                                    if (Rs < 1972) {
                                                        return 4;
                                                    }
                                                    else if (Rs >= 1972) {
                                                        return 1;
                                                    }
                                                }
                                            }
                                            else if (Rs >= 2165) {
                                                return 8;
                                            }
                                        }
                                    }
                                    else if (RsCsDiff >= 7) {
                                        if (posBalance < 45) {
                                            if (gchDC < 2) {
                                                if (SC < 6451) {
                                                    if (SC < 3852) {
                                                        if (mcTcor < 78) {
                                                            if (mcTcor < 74) {
                                                                if (Rs < 1277) {
                                                                    return 8;
                                                                }
                                                                else if (Rs >= 1277) {
                                                                    if (Rs < 1431) {
                                                                        return 1;
                                                                    }
                                                                    else if (Rs >= 1431) {
                                                                        return 8;
                                                                    }
                                                                }
                                                            }
                                                            else if (mcTcor >= 74) {
                                                                return 1;
                                                            }
                                                        }
                                                        else if (mcTcor >= 78) {
                                                            if (ssDCval < 122) {
                                                                if (SC < 2342) {
                                                                    return 4;
                                                                }
                                                                else if (SC >= 2342) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (ssDCval >= 122) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                    else if (SC >= 3852) {
                                                        if (mu_mv_mag_sq < 125) {
                                                            if (Rs < 2153) {
                                                                if (Cs < 2558) {
                                                                    return 1;
                                                                }
                                                                else if (Cs >= 2558) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (Rs >= 2153) {
                                                                if (Rs < 3162) {
                                                                    if (negBalance < 26) {
                                                                        return 8;
                                                                    }
                                                                    else if (negBalance >= 26) {
                                                                        if (Cs < 2980) {
                                                                            if (ssDCval < 131) {
                                                                                if (refDCval < 93) {
                                                                                    if (RsCsDiff < 37) {
                                                                                        return 1;
                                                                                    }
                                                                                    else if (RsCsDiff >= 37) {
                                                                                        return 8;
                                                                                    }
                                                                                }
                                                                                else if (refDCval >= 93) {
                                                                                    if (Cs < 1984) {
                                                                                        return 8;
                                                                                    }
                                                                                    else if (Cs >= 1984) {
                                                                                        if (CsDiff < 4) {
                                                                                            return 8;
                                                                                        }
                                                                                        else if (CsDiff >= 4) {
                                                                                            return 1;
                                                                                        }
                                                                                    }
                                                                                }
                                                                            }
                                                                            else if (ssDCval >= 131) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                        else if (Cs >= 2980) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                }
                                                                else if (Rs >= 3162) {
                                                                    if (posBalance < 40) {
                                                                        if (Cs < 1926) {
                                                                            return 1;
                                                                        }
                                                                        else if (Cs >= 1926) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                    else if (posBalance >= 40) {
                                                                        if (Rs < 3186) {
                                                                            if (SC < 6205) {
                                                                                return 8;
                                                                            }
                                                                            else if (SC >= 6205) {
                                                                                return 1;
                                                                            }
                                                                        }
                                                                        else if (Rs >= 3186) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (mu_mv_mag_sq >= 125) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                                else if (SC >= 6451) {
                                                    if (refDCval < 115) {
                                                        return 8;
                                                    }
                                                    else if (refDCval >= 115) {
                                                        if (refDCval < 119) {
                                                            if (SC < 6696) {
                                                                return 4;
                                                            }
                                                            else if (SC >= 6696) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (refDCval >= 119) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (gchDC >= 2) {
                                                if (CsDiff < 6) {
                                                    return 4;
                                                }
                                                else if (CsDiff >= 6) {
                                                    return 8;
                                                }
                                            }
                                        }
                                        else if (posBalance >= 45) {
                                            if (negBalance < 38) {
                                                if (diffRsCsdiff < 96) {
                                                    if (AFD < 4) {
                                                        if (diffRsCsdiff < -40) {
                                                            if (Cs < 2476) {
                                                                return 8;
                                                            }
                                                            else if (Cs >= 2476) {
                                                                return 1;
                                                            }
                                                        }
                                                        else if (diffRsCsdiff >= -40) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (AFD >= 4) {
                                                        return 8;
                                                    }
                                                }
                                                else if (diffRsCsdiff >= 96) {
                                                    return 1;
                                                }
                                            }
                                            else if (negBalance >= 38) {
                                                if (RsDiff < 3) {
                                                    return 4;
                                                }
                                                else if (RsDiff >= 3) {
                                                    if (tcor < 97) {
                                                        return 8;
                                                    }
                                                    else if (tcor >= 97) {
                                                        if (Rs < 2581) {
                                                            return 4;
                                                        }
                                                        else if (Rs >= 2581) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                            else if (MVDiff >= 3) {
                                if (mcTcor < 98) {
                                    if (RsCsDiff < 201) {
                                        if (refDCval < 78) {
                                            if (mcTcor < 95) {
                                                if (diffMVdiffVal < 10) {
                                                    if (diffMVdiffVal < -4) {
                                                        if (Rs < 1905) {
                                                            return 8;
                                                        }
                                                        else if (Rs >= 1905) {
                                                            if (AFD < 5) {
                                                                return 8;
                                                            }
                                                            else if (AFD >= 5) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                    else if (diffMVdiffVal >= -4) {
                                                        return 8;
                                                    }
                                                }
                                                else if (diffMVdiffVal >= 10) {
                                                    if (RsDiff < 12) {
                                                        return 4;
                                                    }
                                                    else if (RsDiff >= 12) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                            else if (mcTcor >= 95) {
                                                if (Cs < 1022) {
                                                    if (Cs < 856) {
                                                        return 8;
                                                    }
                                                    else if (Cs >= 856) {
                                                        return 1;
                                                    }
                                                }
                                                else if (Cs >= 1022) {
                                                    return 8;
                                                }
                                            }
                                        }
                                        else if (refDCval >= 78) {
                                            if (refDCval < 141) {
                                                if (Rs < 1134) {
                                                    if (negBalance < 29) {
                                                        if (tcor < 88) {
                                                            if (Cs < 716) {
                                                                return 1;
                                                            }
                                                            else if (Cs >= 716) {
                                                                if (SC < 1655) {
                                                                    return 8;
                                                                }
                                                                else if (SC >= 1655) {
                                                                    if (Rs < 972) {
                                                                        return 8;
                                                                    }
                                                                    else if (Rs >= 972) {
                                                                        return 1;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (tcor >= 88) {
                                                            if (SC < 1407) {
                                                                return 8;
                                                            }
                                                            else if (SC >= 1407) {
                                                                if (RsDiff < 2) {
                                                                    if (AFD < 2) {
                                                                        return 1;
                                                                    }
                                                                    else if (AFD >= 2) {
                                                                        return 8;
                                                                    }
                                                                }
                                                                else if (RsDiff >= 2) {
                                                                    return 1;
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (negBalance >= 29) {
                                                        if (RsDiff < 5) {
                                                            return 4;
                                                        }
                                                        else if (RsDiff >= 5) {
                                                            return 1;
                                                        }
                                                    }
                                                }
                                                else if (Rs >= 1134) {
                                                    if (RsDiff < 5) {
                                                        if (diffRsCsdiff < -38) {
                                                            if (MVDiff < 6) {
                                                                return 1;
                                                            }
                                                            else if (MVDiff >= 6) {
                                                                if (CsDiff < 4) {
                                                                    return 8;
                                                                }
                                                                else if (CsDiff >= 4) {
                                                                    if (Rs < 1813) {
                                                                        return 4;
                                                                    }
                                                                    else if (Rs >= 1813) {
                                                                        return 8;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (diffRsCsdiff >= -38) {
                                                            if (CsDiff < 1) {
                                                                return 4;
                                                            }
                                                            else if (CsDiff >= 1) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                    else if (RsDiff >= 5) {
                                                        if (Rs < 2662) {
                                                            if (mu_mv_mag_sq < 113) {
                                                                if (Cs < 2048) {
                                                                    return 8;
                                                                }
                                                                else if (Cs >= 2048) {
                                                                    if (AFD < 8) {
                                                                        return 8;
                                                                    }
                                                                    else if (AFD >= 8) {
                                                                        return 1;
                                                                    }
                                                                }
                                                            }
                                                            else if (mu_mv_mag_sq >= 113) {
                                                                return 1;
                                                            }
                                                        }
                                                        else if (Rs >= 2662) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (refDCval >= 141) {
                                                if (diffRsCsdiff < -26) {
                                                    return 4;
                                                }
                                                else if (diffRsCsdiff >= -26) {
                                                    return 8;
                                                }
                                            }
                                        }
                                    }
                                    else if (RsCsDiff >= 201) {
                                        if (RsCsDiff < 408) {
                                            return 1;
                                        }
                                        else if (RsCsDiff >= 408) {
                                            if (Cs < 1362) {
                                                return 2;
                                            }
                                            else if (Cs >= 1362) {
                                                return 4;
                                            }
                                        }
                                    }
                                }
                                else if (mcTcor >= 98) {
                                    if (SC < 2313) {
                                        if (ssDCval < 26) {
                                            return 8;
                                        }
                                        else if (ssDCval >= 26) {
                                            if (refDCval < 104) {
                                                if (CsDiff < 1) {
                                                    if (Rs < 813) {
                                                        return 1;
                                                    }
                                                    else if (Rs >= 813) {
                                                        return 4;
                                                    }
                                                }
                                                else if (CsDiff >= 1) {
                                                    if (Rs < 771) {
                                                        return 4;
                                                    }
                                                    else if (Rs >= 771) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                            else if (refDCval >= 104) {
                                                if (TSC < 4) {
                                                    return 8;
                                                }
                                                else if (TSC >= 4) {
                                                    return 1;
                                                }
                                            }
                                        }
                                    }
                                    else if (SC >= 2313) {
                                        if (SC < 2335) {
                                            return 2;
                                        }
                                        else if (SC >= 2335) {
                                            if (TSCindex < 7) {
                                                if (ssDCval < 22) {
                                                    return 2;
                                                }
                                                else if (ssDCval >= 22) {
                                                    if (diffMVdiffVal < 15) {
                                                        return 8;
                                                    }
                                                    else if (diffMVdiffVal >= 15) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                            else if (TSCindex >= 7) {
                                                return 4;
                                            }
                                        }
                                    }
                                }
                            }
                        }
                        else if (SC >= 7308) {
                            if (Rs < 3774) {
                                if (Cs < 3877) {
                                    if (RsCsDiff < 2) {
                                        if (Cs < 3849) {
                                            return 1;
                                        }
                                        else if (Cs >= 3849) {
                                            return 8;
                                        }
                                    }
                                    else if (RsCsDiff >= 2) {
                                        return 1;
                                    }
                                }
                                else if (Cs >= 3877) {
                                    return 8;
                                }
                            }
                            else if (Rs >= 3774) {
                                return 1;
                            }
                        }
                    }
                    else if (Rs >= 3779) {
                        if (refDCval < 110) {
                            if (negBalance < 56) {
                                return 8;
                            }
                            else if (negBalance >= 56) {
                                return 2;
                            }
                        }
                        else if (refDCval >= 110) {
                            if (mu_mv_mag_sq < 99) {
                                if (negBalance < 38) {
                                    if (SC < 7178) {
                                        if (SC < 6550) {
                                            return 8;
                                        }
                                        else if (SC >= 6550) {
                                            return 4;
                                        }
                                    }
                                    else if (SC >= 7178) {
                                        if (Rs < 5055) {
                                            return 2;
                                        }
                                        else if (Rs >= 5055) {
                                            if (ssDCval < 112) {
                                                if (MVDiff < 5) {
                                                    return 8;
                                                }
                                                else if (MVDiff >= 5) {
                                                    return 2;
                                                }
                                            }
                                            else if (ssDCval >= 112) {
                                                return 8;
                                            }
                                        }
                                    }
                                }
                                else if (negBalance >= 38) {
                                    return 8;
                                }
                            }
                            else if (mu_mv_mag_sq >= 99) {
                                if (ssDCval < 137) {
                                    return 8;
                                }
                                else if (ssDCval >= 137) {
                                    if (Rs < 4855) {
                                        return 1;
                                    }
                                    else if (Rs >= 4855) {
                                        return 8;
                                    }
                                }
                            }
                        }
                    }
                }
            }
            else if (posBalance >= 51) {
                if (SC < 2705) {
                    if (TSC < 6) {
                        if (Cs < 463) {
                            if (refDCval < 83) {
                                if (Cs < 16) {
                                    return 2;
                                }
                                else if (Cs >= 16) {
                                    return 4;
                                }
                            }
                            else if (refDCval >= 83) {
                                if (RsCsDiff < 2) {
                                    return 8;
                                }
                                else if (RsCsDiff >= 2) {
                                    return 2;
                                }
                            }
                        }
                        else if (Cs >= 463) {
                            if (mu_mv_mag_sq < 104) {
                                if (Scindex < 4) {
                                    if (Rs < 332) {
                                        return 8;
                                    }
                                    else if (Rs >= 332) {
                                        if (Rs < 803) {
                                            return 4;
                                        }
                                        else if (Rs >= 803) {
                                            if (AFD < 5) {
                                                return 8;
                                            }
                                            else if (AFD >= 5) {
                                                return 4;
                                            }
                                        }
                                    }
                                }
                                else if (Scindex >= 4) {
                                    if (diffMVdiffVal < -3) {
                                        if (Rs < 1241) {
                                            return 1;
                                        }
                                        else if (Rs >= 1241) {
                                            return 8;
                                        }
                                    }
                                    else if (diffMVdiffVal >= -3) {
                                        return 8;
                                    }
                                }
                            }
                            else if (mu_mv_mag_sq >= 104) {
                                if (ssDCval < 115) {
                                    return 8;
                                }
                                else if (ssDCval >= 115) {
                                    if (Rs < 332) {
                                        return 8;
                                    }
                                    else if (Rs >= 332) {
                                        if (AFD < 11) {
                                            return 4;
                                        }
                                        else if (AFD >= 11) {
                                            return 1;
                                        }
                                    }
                                }
                            }
                        }
                    }
                    else if (TSC >= 6) {
                        if (RsCsDiff < 37) {
                            if (negBalance < 44) {
                                return 8;
                            }
                            else if (negBalance >= 44) {
                                if (mcTcor < 84) {
                                    if (Rs < 796) {
                                        return 2;
                                    }
                                    else if (Rs >= 796) {
                                        if (Rs < 1153) {
                                            return 4;
                                        }
                                        else if (Rs >= 1153) {
                                            return 8;
                                        }
                                    }
                                }
                                else if (mcTcor >= 84) {
                                    if (mu_mv_mag_sq < 122) {
                                        return 2;
                                    }
                                    else if (mu_mv_mag_sq >= 122) {
                                        return 8;
                                    }
                                }
                            }
                        }
                        else if (RsCsDiff >= 37) {
                            if (AFD < 15) {
                                return 8;
                            }
                            else if (AFD >= 15) {
                                if (AFD < 28) {
                                    if (ssDCval < 152) {
                                        return 2;
                                    }
                                    else if (ssDCval >= 152) {
                                        return 1;
                                    }
                                }
                                else if (AFD >= 28) {
                                    if (RsDiff < 8) {
                                        return 4;
                                    }
                                    else if (RsDiff >= 8) {
                                        return 1;
                                    }
                                }
                            }
                        }
                    }
                }
                else if (SC >= 2705) {
                    if (Rs < 7103) {
                        if (SC < 10499) {
                            if (mcTcor < 97) {
                                if (Rs < 5576) {
                                    if (Rs < 1762) {
                                        if (AFD < 7) {
                                            if (negBalance < 45) {
                                                return 8;
                                            }
                                            else if (negBalance >= 45) {
                                                if (refDCval < 74) {
                                                    if (TSC < 6) {
                                                        return 8;
                                                    }
                                                    else if (TSC >= 6) {
                                                        if (mcTcor < 93) {
                                                            return 8;
                                                        }
                                                        else if (mcTcor >= 93) {
                                                            return 2;
                                                        }
                                                    }
                                                }
                                                else if (refDCval >= 74) {
                                                    return 8;
                                                }
                                            }
                                        }
                                        else if (AFD >= 7) {
                                            if (RsDiff < 5) {
                                                if (ssDCval < 87) {
                                                    if (SC < 3303) {
                                                        if (mu_mv_mag_sq < 119) {
                                                            return 2;
                                                        }
                                                        else if (mu_mv_mag_sq >= 119) {
                                                            if (RsCsDiff < 37) {
                                                                return 2;
                                                            }
                                                            else if (RsCsDiff >= 37) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                    else if (SC >= 3303) {
                                                        if (MVDiff < 1) {
                                                            return 2;
                                                        }
                                                        else if (MVDiff >= 1) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                                else if (ssDCval >= 87) {
                                                    return 8;
                                                }
                                            }
                                            else if (RsDiff >= 5) {
                                                if (AFD < 9) {
                                                    return 8;
                                                }
                                                else if (AFD >= 9) {
                                                    if (Cs < 1452) {
                                                        return 1;
                                                    }
                                                    else if (Cs >= 1452) {
                                                        return 2;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (Rs >= 1762) {
                                        if (ssDCval < 86) {
                                            if (SC < 3659) {
                                                if (SC < 3529) {
                                                    if (refDCval < 76) {
                                                        if (posBalance < 58) {
                                                            if (CsDiff < 2) {
                                                                return 4;
                                                            }
                                                            else if (CsDiff >= 2) {
                                                                if (TSC < 7) {
                                                                    if (ssDCval < 75) {
                                                                        return 8;
                                                                    }
                                                                    else if (ssDCval >= 75) {
                                                                        if (negBalance < 43) {
                                                                            return 8;
                                                                        }
                                                                        else if (negBalance >= 43) {
                                                                            if (negBalance < 44) {
                                                                                if (Rs < 1786) {
                                                                                    return 8;
                                                                                }
                                                                                else if (Rs >= 1786) {
                                                                                    if (Rs < 1849) {
                                                                                        return 4;
                                                                                    }
                                                                                    else if (Rs >= 1849) {
                                                                                        return 8;
                                                                                    }
                                                                                }
                                                                            }
                                                                            else if (negBalance >= 44) {
                                                                                return 8;
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                                else if (TSC >= 7) {
                                                                    return 2;
                                                                }
                                                            }
                                                        }
                                                        else if (posBalance >= 58) {
                                                            if (TSC < 6) {
                                                                if (Cs < 1521) {
                                                                    return 2;
                                                                }
                                                                else if (Cs >= 1521) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (TSC >= 6) {
                                                                if (diffAFD < 1) {
                                                                    if (mcTcor < 93) {
                                                                        return 8;
                                                                    }
                                                                    else if (mcTcor >= 93) {
                                                                        if (tcor < 94) {
                                                                            return 2;
                                                                        }
                                                                        else if (tcor >= 94) {
                                                                            if (Rs < 1815) {
                                                                                return 2;
                                                                            }
                                                                            else if (Rs >= 1815) {
                                                                                if (SC < 3431) {
                                                                                    return 4;
                                                                                }
                                                                                else if (SC >= 3431) {
                                                                                    return 2;
                                                                                }
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                                else if (diffAFD >= 1) {
                                                                    if (tcor < 92) {
                                                                        return 8;
                                                                    }
                                                                    else if (tcor >= 92) {
                                                                        if (negBalance < 46) {
                                                                            return 2;
                                                                        }
                                                                        else if (negBalance >= 46) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (refDCval >= 76) {
                                                        if (MVDiff < 1) {
                                                            return 4;
                                                        }
                                                        else if (MVDiff >= 1) {
                                                            return 2;
                                                        }
                                                    }
                                                }
                                                else if (SC >= 3529) {
                                                    if (negBalance < 45) {
                                                        if (tcor < 94) {
                                                            return 8;
                                                        }
                                                        else if (tcor >= 94) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (negBalance >= 45) {
                                                        if (Rs < 1851) {
                                                            if (SC < 3546) {
                                                                return 2;
                                                            }
                                                            else if (SC >= 3546) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (Rs >= 1851) {
                                                            if (ssDCval < 76) {
                                                                if (Cs < 1673) {
                                                                    return 2;
                                                                }
                                                                else if (Cs >= 1673) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (ssDCval >= 76) {
                                                                if (AFD < 7) {
                                                                    return 4;
                                                                }
                                                                else if (AFD >= 7) {
                                                                    return 2;
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                            else if (SC >= 3659) {
                                                if (SC < 7691) {
                                                    if (ssDCval < 56) {
                                                        if (diffMVdiffVal < -2) {
                                                            if (mu_mv_mag_sq < 101) {
                                                                return 4;
                                                            }
                                                            else if (mu_mv_mag_sq >= 101) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (diffMVdiffVal >= -2) {
                                                            if (diffTSC < -1) {
                                                                if (SC < 7220) {
                                                                    return 4;
                                                                }
                                                                else if (SC >= 7220) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (diffTSC >= -1) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                    else if (ssDCval >= 56) {
                                                        if (Rs < 1865) {
                                                            if (posBalance < 56) {
                                                                return 4;
                                                            }
                                                            else if (posBalance >= 56) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (Rs >= 1865) {
                                                            if (posBalance < 58) {
                                                                if (RsCsDiff < 22) {
                                                                    if (Rs < 3777) {
                                                                        return 8;
                                                                    }
                                                                    else if (Rs >= 3777) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (RsCsDiff >= 22) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (posBalance >= 58) {
                                                                if (mu_mv_mag_sq < 100) {
                                                                    return 4;
                                                                }
                                                                else if (mu_mv_mag_sq >= 100) {
                                                                    if (SC < 3780) {
                                                                        if (ssDCval < 74) {
                                                                            return 8;
                                                                        }
                                                                        else if (ssDCval >= 74) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                    else if (SC >= 3780) {
                                                                        return 8;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (SC >= 7691) {
                                                    if (Cs < 4876) {
                                                        if (RsDiff < 5) {
                                                            if (TSC < 7) {
                                                                return 8;
                                                            }
                                                            else if (TSC >= 7) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (RsDiff >= 5) {
                                                            if (gchDC < 1) {
                                                                return 8;
                                                            }
                                                            else if (gchDC >= 1) {
                                                                if (refDCval < 80) {
                                                                    return 8;
                                                                }
                                                                else if (refDCval >= 80) {
                                                                    if (SC < 9058) {
                                                                        return 8;
                                                                    }
                                                                    else if (SC >= 9058) {
                                                                        if (Cs < 4834) {
                                                                            return 4;
                                                                        }
                                                                        else if (Cs >= 4834) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (Cs >= 4876) {
                                                        if (AFD < 9) {
                                                            if (Rs < 4974) {
                                                                if (ssDCval < 80) {
                                                                    if (SC < 9773) {
                                                                        if (diffRsCsdiff < -16) {
                                                                            return 8;
                                                                        }
                                                                        else if (diffRsCsdiff >= -16) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                    else if (SC >= 9773) {
                                                                        return 8;
                                                                    }
                                                                }
                                                                else if (ssDCval >= 80) {
                                                                    if (mu_mv_mag_sq < 128) {
                                                                        if (Rs < 4770) {
                                                                            return 4;
                                                                        }
                                                                        else if (Rs >= 4770) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                    else if (mu_mv_mag_sq >= 128) {
                                                                        if (tcor < 92) {
                                                                            if (Rs < 4832) {
                                                                                if (diffAFD < -1) {
                                                                                    return 4;
                                                                                }
                                                                                else if (diffAFD >= -1) {
                                                                                    return 8;
                                                                                }
                                                                            }
                                                                            else if (Rs >= 4832) {
                                                                                if (Cs < 5055) {
                                                                                    return 4;
                                                                                }
                                                                                else if (Cs >= 5055) {
                                                                                    if (Rs < 4915) {
                                                                                        return 8;
                                                                                    }
                                                                                    else if (Rs >= 4915) {
                                                                                        return 4;
                                                                                    }
                                                                                }
                                                                            }
                                                                        }
                                                                        else if (tcor >= 92) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                            else if (Rs >= 4974) {
                                                                if (diffRsCsdiff < 41) {
                                                                    return 4;
                                                                }
                                                                else if (diffRsCsdiff >= 41) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                        else if (AFD >= 9) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (ssDCval >= 86) {
                                            if (refDCval < 97) {
                                                if (Rs < 2472) {
                                                    if (mu_mv_mag_sq < 106) {
                                                        if (ssDCval < 93) {
                                                            if (negBalance < 49) {
                                                                return 2;
                                                            }
                                                            else if (negBalance >= 49) {
                                                                if (refDCval < 93) {
                                                                    return 8;
                                                                }
                                                                else if (refDCval >= 93) {
                                                                    return 2;
                                                                }
                                                            }
                                                        }
                                                        else if (ssDCval >= 93) {
                                                            return 2;
                                                        }
                                                    }
                                                    else if (mu_mv_mag_sq >= 106) {
                                                        if (negBalance < 36) {
                                                            if (Rs < 2164) {
                                                                return 8;
                                                            }
                                                            else if (Rs >= 2164) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (negBalance >= 36) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                                else if (Rs >= 2472) {
                                                    if (Cs < 3598) {
                                                        if (Cs < 2725) {
                                                            if (ssDCval < 94) {
                                                                if (SC < 5572) {
                                                                    return 8;
                                                                }
                                                                else if (SC >= 5572) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (ssDCval >= 94) {
                                                                if (Cs < 1026) {
                                                                    return 1;
                                                                }
                                                                else if (Cs >= 1026) {
                                                                    return 2;
                                                                }
                                                            }
                                                        }
                                                        else if (Cs >= 2725) {
                                                            if (RsCsDiff < 183) {
                                                                return 2;
                                                            }
                                                            else if (RsCsDiff >= 183) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                    else if (Cs >= 3598) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                            else if (refDCval >= 97) {
                                                if (negBalance < 54) {
                                                    if (SC < 4742) {
                                                        if (ssDCval < 109) {
                                                            return 4;
                                                        }
                                                        else if (ssDCval >= 109) {
                                                            if (refDCval < 154) {
                                                                if (posBalance < 56) {
                                                                    return 4;
                                                                }
                                                                else if (posBalance >= 56) {
                                                                    if (Cs < 1601) {
                                                                        if (Cs < 1342) {
                                                                            return 8;
                                                                        }
                                                                        else if (Cs >= 1342) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                    else if (Cs >= 1601) {
                                                                        return 8;
                                                                    }
                                                                }
                                                            }
                                                            else if (refDCval >= 154) {
                                                                return 2;
                                                            }
                                                        }
                                                    }
                                                    else if (SC >= 4742) {
                                                        if (mcTcor < 84) {
                                                            if (AFD < 16) {
                                                                return 1;
                                                            }
                                                            else if (AFD >= 16) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (mcTcor >= 84) {
                                                            if (refDCval < 121) {
                                                                if (Rs < 3497) {
                                                                    if (AFD < 13) {
                                                                        return 8;
                                                                    }
                                                                    else if (AFD >= 13) {
                                                                        if (Rs < 3347) {
                                                                            return 4;
                                                                        }
                                                                        else if (Rs >= 3347) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                }
                                                                else if (Rs >= 3497) {
                                                                    if (Cs < 3804) {
                                                                        if (RsDiff < 13) {
                                                                            if (tcor < 88) {
                                                                                if (Cs < 3129) {
                                                                                    return 4;
                                                                                }
                                                                                else if (Cs >= 3129) {
                                                                                    return 8;
                                                                                }
                                                                            }
                                                                            else if (tcor >= 88) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                        else if (RsDiff >= 13) {
                                                                            if (Rs < 4610) {
                                                                                if (Rs < 4187) {
                                                                                    return 8;
                                                                                }
                                                                                else if (Rs >= 4187) {
                                                                                    return 4;
                                                                                }
                                                                            }
                                                                            else if (Rs >= 4610) {
                                                                                return 2;
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (Cs >= 3804) {
                                                                        return 8;
                                                                    }
                                                                }
                                                            }
                                                            else if (refDCval >= 121) {
                                                                if (Cs < 1966) {
                                                                    return 4;
                                                                }
                                                                else if (Cs >= 1966) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (negBalance >= 54) {
                                                    if (RsCsDiff < 231) {
                                                        return 4;
                                                    }
                                                    else if (RsCsDiff >= 231) {
                                                        if (AFD < 31) {
                                                            if (Cs < 2278) {
                                                                return 1;
                                                            }
                                                            else if (Cs >= 2278) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (AFD >= 31) {
                                                            return 2;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (Rs >= 5576) {
                                    if (Cs < 3894) {
                                        return 4;
                                    }
                                    else if (Cs >= 3894) {
                                        return 8;
                                    }
                                }
                            }
                            else if (mcTcor >= 97) {
                                if (mu_mv_mag_sq < 121) {
                                    if (Rs < 1771) {
                                        if (RsCsDiff < 19) {
                                            return 8;
                                        }
                                        else if (RsCsDiff >= 19) {
                                            if (Rs < 1607) {
                                                return 8;
                                            }
                                            else if (Rs >= 1607) {
                                                if (Rs < 1711) {
                                                    return 4;
                                                }
                                                else if (Rs >= 1711) {
                                                    return 2;
                                                }
                                            }
                                        }
                                    }
                                    else if (Rs >= 1771) {
                                        if (posBalance < 58) {
                                            if (ssDCval < 89) {
                                                if (MVDiff < 1) {
                                                    return 4;
                                                }
                                                else if (MVDiff >= 1) {
                                                    if (negBalance < 38) {
                                                        if (refDCval < 73) {
                                                            return 8;
                                                        }
                                                        else if (refDCval >= 73) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (negBalance >= 38) {
                                                        if (Rs < 1808) {
                                                            return 8;
                                                        }
                                                        else if (Rs >= 1808) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (ssDCval >= 89) {
                                                if (Cs < 1792) {
                                                    if (AFD < 6) {
                                                        return 8;
                                                    }
                                                    else if (AFD >= 6) {
                                                        if (MVDiff < 2) {
                                                            return 1;
                                                        }
                                                        else if (MVDiff >= 2) {
                                                            if (Cs < 1521) {
                                                                return 1;
                                                            }
                                                            else if (Cs >= 1521) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (Cs >= 1792) {
                                                    if (gchDC < 1) {
                                                        if (negBalance < 48) {
                                                            return 8;
                                                        }
                                                        else if (negBalance >= 48) {
                                                            if (ssDCval < 115) {
                                                                if (ssDCval < 111) {
                                                                    return 8;
                                                                }
                                                                else if (ssDCval >= 111) {
                                                                    if (Rs < 2634) {
                                                                        return 8;
                                                                    }
                                                                    else if (Rs >= 2634) {
                                                                        return 4;
                                                                    }
                                                                }
                                                            }
                                                            else if (ssDCval >= 115) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                    else if (gchDC >= 1) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                        }
                                        else if (posBalance >= 58) {
                                            if (Cs < 2288) {
                                                if (AFD < 4) {
                                                    return 8;
                                                }
                                                else if (AFD >= 4) {
                                                    return 4;
                                                }
                                            }
                                            else if (Cs >= 2288) {
                                                if (ssDCval < 113) {
                                                    return 8;
                                                }
                                                else if (ssDCval >= 113) {
                                                    if (posBalance < 59) {
                                                        if (diffRsCsdiff < 4) {
                                                            return 8;
                                                        }
                                                        else if (diffRsCsdiff >= 4) {
                                                            if (Rs < 2740) {
                                                                return 4;
                                                            }
                                                            else if (Rs >= 2740) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                    else if (posBalance >= 59) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (mu_mv_mag_sq >= 121) {
                                    if (negBalance < 52) {
                                        if (mu_mv_mag_sq < 123) {
                                            if (gchDC < 1) {
                                                if (ssDCval < 76) {
                                                    return 4;
                                                }
                                                else if (ssDCval >= 76) {
                                                    if (negBalance < 50) {
                                                        if (Cs < 2943) {
                                                            if (ssDCval < 97) {
                                                                if (Rs < 2142) {
                                                                    return 8;
                                                                }
                                                                else if (Rs >= 2142) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (ssDCval >= 97) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (Cs >= 2943) {
                                                            if (Rs < 2895) {
                                                                if (Cs < 2953) {
                                                                    return 4;
                                                                }
                                                                else if (Cs >= 2953) {
                                                                    if (SC < 5917) {
                                                                        if (CsDiff < 5) {
                                                                            if (RsDiff < 5) {
                                                                                if (Cs < 3007) {
                                                                                    return 8;
                                                                                }
                                                                                else if (Cs >= 3007) {
                                                                                    if (SC < 5863) {
                                                                                        return 4;
                                                                                    }
                                                                                    else if (SC >= 5863) {
                                                                                        if (refDCval < 112) {
                                                                                            return 8;
                                                                                        }
                                                                                        else if (refDCval >= 112) {
                                                                                            if (Rs < 2796) {
                                                                                                return 8;
                                                                                            }
                                                                                            else if (Rs >= 2796) {
                                                                                                return 4;
                                                                                            }
                                                                                        }
                                                                                    }
                                                                                }
                                                                            }
                                                                            else if (RsDiff >= 5) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                        else if (CsDiff >= 5) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                    else if (SC >= 5917) {
                                                                        return 4;
                                                                    }
                                                                }
                                                            }
                                                            else if (Rs >= 2895) {
                                                                if (Cs < 3153) {
                                                                    if (Cs < 3111) {
                                                                        return 8;
                                                                    }
                                                                    else if (Cs >= 3111) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (Cs >= 3153) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (negBalance >= 50) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                            else if (gchDC >= 1) {
                                                return 8;
                                            }
                                        }
                                        else if (mu_mv_mag_sq >= 123) {
                                            if (Rs < 1871) {
                                                if (RsCsDiff < 9) {
                                                    if (ssDCval < 94) {
                                                        return 8;
                                                    }
                                                    else if (ssDCval >= 94) {
                                                        return 4;
                                                    }
                                                }
                                                else if (RsCsDiff >= 9) {
                                                    return 4;
                                                }
                                            }
                                            else if (Rs >= 1871) {
                                                if (refDCval < 83) {
                                                    if (Rs < 2218) {
                                                        return 8;
                                                    }
                                                    else if (Rs >= 2218) {
                                                        return 4;
                                                    }
                                                }
                                                else if (refDCval >= 83) {
                                                    if (TSC < 6) {
                                                        return 8;
                                                    }
                                                    else if (TSC >= 6) {
                                                        if (Rs < 2863) {
                                                            if (Cs < 2949) {
                                                                return 8;
                                                            }
                                                            else if (Cs >= 2949) {
                                                                if (Rs < 2861) {
                                                                    if (diffRsCsdiff < -8) {
                                                                        if (SC < 5758) {
                                                                            return 8;
                                                                        }
                                                                        else if (SC >= 5758) {
                                                                            if (Cs < 3066) {
                                                                                if (ssDCval < 114) {
                                                                                    return 4;
                                                                                }
                                                                                else if (ssDCval >= 114) {
                                                                                    return 8;
                                                                                }
                                                                            }
                                                                            else if (Cs >= 3066) {
                                                                                return 8;
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (diffRsCsdiff >= -8) {
                                                                        if (mu_mv_mag_sq < 124) {
                                                                            if (Cs < 2952) {
                                                                                return 4;
                                                                            }
                                                                            else if (Cs >= 2952) {
                                                                                return 8;
                                                                            }
                                                                        }
                                                                        else if (mu_mv_mag_sq >= 124) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                }
                                                                else if (Rs >= 2861) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                        else if (Rs >= 2863) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (negBalance >= 52) {
                                        if (Rs < 5507) {
                                            return 8;
                                        }
                                        else if (Rs >= 5507) {
                                            return 4;
                                        }
                                    }
                                }
                            }
                        }
                        else if (SC >= 10499) {
                            if (ssDCval < 106) {
                                if (refDCval < 81) {
                                    if (negBalance < 49) {
                                        return 4;
                                    }
                                    else if (negBalance >= 49) {
                                        return 8;
                                    }
                                }
                                else if (refDCval >= 81) {
                                    return 8;
                                }
                            }
                            else if (ssDCval >= 106) {
                                if (ssDCval < 109) {
                                    if (mcTcor < 94) {
                                        if (diffRsCsdiff < 95) {
                                            return 8;
                                        }
                                        else if (diffRsCsdiff >= 95) {
                                            return 4;
                                        }
                                    }
                                    else if (mcTcor >= 94) {
                                        return 4;
                                    }
                                }
                                else if (ssDCval >= 109) {
                                    if (negBalance < 57) {
                                        return 8;
                                    }
                                    else if (negBalance >= 57) {
                                        return 4;
                                    }
                                }
                            }
                        }
                    }
                    else if (Rs >= 7103) {
                        if (AFD < 12) {
                            return 4;
                        }
                        else if (AFD >= 12) {
                            if (negBalance < 59) {
                                return 8;
                            }
                            else if (negBalance >= 59) {
                                if (TSC < 19) {
                                    return 2;
                                }
                                else if (TSC >= 19) {
                                    if (negBalance < 67) {
                                        if (mu_mv_mag_sq < 120) {
                                            return 4;
                                        }
                                        else if (mu_mv_mag_sq >= 120) {
                                            if (Cs < 7971) {
                                                if (Cs < 7828) {
                                                    return 2;
                                                }
                                                else if (Cs >= 7828) {
                                                    return 4;
                                                }
                                            }
                                            else if (Cs >= 7971) {
                                                return 2;
                                            }
                                        }
                                    }
                                    else if (negBalance >= 67) {
                                        if (AFD < 20) {
                                            return 4;
                                        }
                                        else if (AFD >= 20) {
                                            return 2;
                                        }
                                    }
                                }
                            }
                        }
                    }
                }
            }
        }
        return 4;
    }
    mfxU16 AGOPSelect11(mfxU32 Rs, mfxU32 CsDiff, mfxU32 TSC, mfxI32 diffAFD, mfxU32 Scindex, mfxU32 RsDiff, mfxI32 diffMVdiffVal, mfxU32 RsCsDiff, mfxU32 TSCindex, mfxI16 tcor, mfxU32 refDCval, mfxU32 MVDiff, mfxU32 AFD, mfxI32 diffTSC, mfxI32 diffRsCsdiff, mfxU32 posBalance, mfxU32 SC, mfxU32 mu_mv_mag_sq, mfxU32 negBalance, mfxI16 mcTcor, mfxU32 gchDC, mfxU32 Cs, mfxU32 ssDCval) {
        if (mu_mv_mag_sq < 89) {
            if (posBalance < 3) {
                if (SC < 3696) {
                    if (Rs < 12) {
                        if (mcTcor < 88) {
                            if (MVDiff < 111) {
                                if (AFD < 2) {
                                    return 4;
                                }
                                else if (AFD >= 2) {
                                    return 2;
                                }
                            }
                            else if (MVDiff >= 111) {
                                if (negBalance < 35) {
                                    return 8;
                                }
                                else if (negBalance >= 35) {
                                    if (AFD < 2) {
                                        if (tcor < 127) {
                                            if (negBalance < 65) {
                                                return 1;
                                            }
                                            else if (negBalance >= 65) {
                                                return 2;
                                            }
                                        }
                                        else if (tcor >= 127) {
                                            return 2;
                                        }
                                    }
                                    else if (AFD >= 2) {
                                        return 1;
                                    }
                                }
                            }
                        }
                        else if (mcTcor >= 88) {
                            if (diffTSC < -1) {
                                if (SC < 21) {
                                    return 4;
                                }
                                else if (SC >= 21) {
                                    return 8;
                                }
                            }
                            else if (diffTSC >= -1) {
                                return 8;
                            }
                        }
                    }
                    else if (Rs >= 12) {
                        return 8;
                    }
                }
                else if (SC >= 3696) {
                    if (Rs < 4729) {
                        if (SC < 3733) {
                            return 1;
                        }
                        else if (SC >= 3733) {
                            if (Cs < 3985) {
                                if (Rs < 3755) {
                                    return 8;
                                }
                                else if (Rs >= 3755) {
                                    if (Rs < 3791) {
                                        return 1;
                                    }
                                    else if (Rs >= 3791) {
                                        return 8;
                                    }
                                }
                            }
                            else if (Cs >= 3985) {
                                return 8;
                            }
                        }
                    }
                    else if (Rs >= 4729) {
                        return 1;
                    }
                }
            }
            else if (posBalance >= 3) {
                if (Scindex < 6) {
                    if (AFD < 12) {
                        if (refDCval < 104) {
                            if (mcTcor < 90) {
                                if (Cs < 1295) {
                                    if (MVDiff < 17) {
                                        if (Rs < 273) {
                                            if (AFD < 4) {
                                                if (refDCval < 39) {
                                                    return 4;
                                                }
                                                else if (refDCval >= 39) {
                                                    return 8;
                                                }
                                            }
                                            else if (AFD >= 4) {
                                                if (Cs < 72) {
                                                    return 8;
                                                }
                                                else if (Cs >= 72) {
                                                    return 2;
                                                }
                                            }
                                        }
                                        else if (Rs >= 273) {
                                            if (ssDCval < 55) {
                                                if (RsCsDiff < 7) {
                                                    if (posBalance < 53) {
                                                        return 8;
                                                    }
                                                    else if (posBalance >= 53) {
                                                        if (diffMVdiffVal < -10) {
                                                            return 2;
                                                        }
                                                        else if (diffMVdiffVal >= -10) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                                else if (RsCsDiff >= 7) {
                                                    if (gchDC < 2) {
                                                        if (mu_mv_mag_sq < 40) {
                                                            if (RsCsDiff < 16) {
                                                                if (Rs < 1067) {
                                                                    return 4;
                                                                }
                                                                else if (Rs >= 1067) {
                                                                    return 1;
                                                                }
                                                            }
                                                            else if (RsCsDiff >= 16) {
                                                                return 1;
                                                            }
                                                        }
                                                        else if (mu_mv_mag_sq >= 40) {
                                                            if (Rs < 1253) {
                                                                return 4;
                                                            }
                                                            else if (Rs >= 1253) {
                                                                if (AFD < 7) {
                                                                    return 1;
                                                                }
                                                                else if (AFD >= 7) {
                                                                    if (Rs < 1454) {
                                                                        return 4;
                                                                    }
                                                                    else if (Rs >= 1454) {
                                                                        return 8;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (gchDC >= 2) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                            else if (ssDCval >= 55) {
                                                if (ssDCval < 99) {
                                                    if (mu_mv_mag_sq < 82) {
                                                        if (SC < 2171) {
                                                            if (SC < 1288) {
                                                                if (Rs < 491) {
                                                                    return 4;
                                                                }
                                                                else if (Rs >= 491) {
                                                                    if (Cs < 425) {
                                                                        return 2;
                                                                    }
                                                                    else if (Cs >= 425) {
                                                                        return 8;
                                                                    }
                                                                }
                                                            }
                                                            else if (SC >= 1288) {
                                                                if (refDCval < 64) {
                                                                    return 8;
                                                                }
                                                                else if (refDCval >= 64) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                        else if (SC >= 2171) {
                                                            if (RsCsDiff < 37) {
                                                                if (AFD < 10) {
                                                                    if (Cs < 1029) {
                                                                        return 8;
                                                                    }
                                                                    else if (Cs >= 1029) {
                                                                        return 2;
                                                                    }
                                                                }
                                                                else if (AFD >= 10) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (RsCsDiff >= 37) {
                                                                if (ssDCval < 66) {
                                                                    return 4;
                                                                }
                                                                else if (ssDCval >= 66) {
                                                                    if (Cs < 1013) {
                                                                        return 2;
                                                                    }
                                                                    else if (Cs >= 1013) {
                                                                        return 8;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (mu_mv_mag_sq >= 82) {
                                                        return 8;
                                                    }
                                                }
                                                else if (ssDCval >= 99) {
                                                    return 4;
                                                }
                                            }
                                        }
                                    }
                                    else if (MVDiff >= 17) {
                                        if (TSC < 6) {
                                            if (ssDCval < 86) {
                                                if (Cs < 386) {
                                                    if (AFD < 6) {
                                                        if (Rs < 403) {
                                                            if (negBalance < 72) {
                                                                if (tcor < 69) {
                                                                    if (mcTcor < 42) {
                                                                        if (MVDiff < 307) {
                                                                            if (mcTcor < 37) {
                                                                                if (mu_mv_mag_sq < 26) {
                                                                                    if (diffMVdiffVal < -154) {
                                                                                        if (ssDCval < 20) {
                                                                                            return 8;
                                                                                        }
                                                                                        else if (ssDCval >= 20) {
                                                                                            return 2;
                                                                                        }
                                                                                    }
                                                                                    else if (diffMVdiffVal >= -154) {
                                                                                        return 4;
                                                                                    }
                                                                                }
                                                                                else if (mu_mv_mag_sq >= 26) {
                                                                                    return 1;
                                                                                }
                                                                            }
                                                                            else if (mcTcor >= 37) {
                                                                                if (diffTSC < -2) {
                                                                                    return 1;
                                                                                }
                                                                                else if (diffTSC >= -2) {
                                                                                    return 8;
                                                                                }
                                                                            }
                                                                        }
                                                                        else if (MVDiff >= 307) {
                                                                            return 2;
                                                                        }
                                                                    }
                                                                    else if (mcTcor >= 42) {
                                                                        if (MVDiff < 23) {
                                                                            return 2;
                                                                        }
                                                                        else if (MVDiff >= 23) {
                                                                            if (MVDiff < 61) {
                                                                                if (negBalance < 2) {
                                                                                    return 1;
                                                                                }
                                                                                else if (negBalance >= 2) {
                                                                                    if (negBalance < 6) {
                                                                                        if (posBalance < 26) {
                                                                                            return 1;
                                                                                        }
                                                                                        else if (posBalance >= 26) {
                                                                                            return 4;
                                                                                        }
                                                                                    }
                                                                                    else if (negBalance >= 6) {
                                                                                        return 1;
                                                                                    }
                                                                                }
                                                                            }
                                                                            else if (MVDiff >= 61) {
                                                                                if (refDCval < 22) {
                                                                                    if (diffAFD < -1) {
                                                                                        if (AFD < 1) {
                                                                                            return 2;
                                                                                        }
                                                                                        else if (AFD >= 1) {
                                                                                            return 1;
                                                                                        }
                                                                                    }
                                                                                    else if (diffAFD >= -1) {
                                                                                        if (MVDiff < 69) {
                                                                                            if (MVDiff < 67) {
                                                                                                if (MVDiff < 64) {
                                                                                                    if (diffMVdiffVal < -2) {
                                                                                                        return 8;
                                                                                                    }
                                                                                                    else if (diffMVdiffVal >= -2) {
                                                                                                        return 4;
                                                                                                    }
                                                                                                }
                                                                                                else if (MVDiff >= 64) {
                                                                                                    return 4;
                                                                                                }
                                                                                            }
                                                                                            else if (MVDiff >= 67) {
                                                                                                return 8;
                                                                                            }
                                                                                        }
                                                                                        else if (MVDiff >= 69) {
                                                                                            return 4;
                                                                                        }
                                                                                    }
                                                                                }
                                                                                else if (refDCval >= 22) {
                                                                                    if (SC < 11) {
                                                                                        if (MVDiff < 79) {
                                                                                            return 4;
                                                                                        }
                                                                                        else if (MVDiff >= 79) {
                                                                                            return 1;
                                                                                        }
                                                                                    }
                                                                                    else if (SC >= 11) {
                                                                                        if (Rs < 17) {
                                                                                            return 4;
                                                                                        }
                                                                                        else if (Rs >= 17) {
                                                                                            return 8;
                                                                                        }
                                                                                    }
                                                                                }
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                                else if (tcor >= 69) {
                                                                    if (mu_mv_mag_sq < 20) {
                                                                        if (ssDCval < 22) {
                                                                            if (negBalance < 10) {
                                                                                return 8;
                                                                            }
                                                                            else if (negBalance >= 10) {
                                                                                if (posBalance < 5) {
                                                                                    return 2;
                                                                                }
                                                                                else if (posBalance >= 5) {
                                                                                    return 4;
                                                                                }
                                                                            }
                                                                        }
                                                                        else if (ssDCval >= 22) {
                                                                            if (tcor < 78) {
                                                                                return 2;
                                                                            }
                                                                            else if (tcor >= 78) {
                                                                                if (diffTSC < 1) {
                                                                                    return 4;
                                                                                }
                                                                                else if (diffTSC >= 1) {
                                                                                    return 2;
                                                                                }
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (mu_mv_mag_sq >= 20) {
                                                                        if (negBalance < 1) {
                                                                            if (AFD < 1) {
                                                                                return 2;
                                                                            }
                                                                            else if (AFD >= 1) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                        else if (negBalance >= 1) {
                                                                            if (MVDiff < 18) {
                                                                                if (Rs < 196) {
                                                                                    return 4;
                                                                                }
                                                                                else if (Rs >= 196) {
                                                                                    return 1;
                                                                                }
                                                                            }
                                                                            else if (MVDiff >= 18) {
                                                                                if (ssDCval < 19) {
                                                                                    return 4;
                                                                                }
                                                                                else if (ssDCval >= 19) {
                                                                                    if (negBalance < 42) {
                                                                                        return 8;
                                                                                    }
                                                                                    else if (negBalance >= 42) {
                                                                                        if (SC < 727) {
                                                                                            if (tcor < 83) {
                                                                                                if (gchDC < 1) {
                                                                                                    return 8;
                                                                                                }
                                                                                                else if (gchDC >= 1) {
                                                                                                    if (TSC < 4) {
                                                                                                        return 8;
                                                                                                    }
                                                                                                    else if (TSC >= 4) {
                                                                                                        return 4;
                                                                                                    }
                                                                                                }
                                                                                            }
                                                                                            else if (tcor >= 83) {
                                                                                                return 4;
                                                                                            }
                                                                                        }
                                                                                        else if (SC >= 727) {
                                                                                            return 4;
                                                                                        }
                                                                                    }
                                                                                }
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                            else if (negBalance >= 72) {
                                                                return 1;
                                                            }
                                                        }
                                                        else if (Rs >= 403) {
                                                            if (mu_mv_mag_sq < 60) {
                                                                if (tcor < 84) {
                                                                    if (Cs < 376) {
                                                                        return 1;
                                                                    }
                                                                    else if (Cs >= 376) {
                                                                        if (SC < 833) {
                                                                            return 4;
                                                                        }
                                                                        else if (SC >= 833) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                }
                                                                else if (tcor >= 84) {
                                                                    if (SC < 844) {
                                                                        return 4;
                                                                    }
                                                                    else if (SC >= 844) {
                                                                        return 8;
                                                                    }
                                                                }
                                                            }
                                                            else if (mu_mv_mag_sq >= 60) {
                                                                if (refDCval < 38) {
                                                                    if (tcor < 86) {
                                                                        if (diffTSC < -1) {
                                                                            return 4;
                                                                        }
                                                                        else if (diffTSC >= -1) {
                                                                            if (Cs < 376) {
                                                                                return 1;
                                                                            }
                                                                            else if (Cs >= 376) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (tcor >= 86) {
                                                                        return 2;
                                                                    }
                                                                }
                                                                else if (refDCval >= 38) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (AFD >= 6) {
                                                        return 1;
                                                    }
                                                }
                                                else if (Cs >= 386) {
                                                    if (TSC < 5) {
                                                        if (refDCval < 39) {
                                                            if (TSC < 2) {
                                                                if (TSCindex < 1) {
                                                                    return 2;
                                                                }
                                                                else if (TSCindex >= 1) {
                                                                    if (mcTcor < 87) {
                                                                        return 1;
                                                                    }
                                                                    else if (mcTcor >= 87) {
                                                                        return 2;
                                                                    }
                                                                }
                                                            }
                                                            else if (TSC >= 2) {
                                                                if (mu_mv_mag_sq < 60) {
                                                                    if (mcTcor < 72) {
                                                                        return 8;
                                                                    }
                                                                    else if (mcTcor >= 72) {
                                                                        return 2;
                                                                    }
                                                                }
                                                                else if (mu_mv_mag_sq >= 60) {
                                                                    if (MVDiff < 22) {
                                                                        if (TSC < 4) {
                                                                            return 8;
                                                                        }
                                                                        else if (TSC >= 4) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                    else if (MVDiff >= 22) {
                                                                        if (mu_mv_mag_sq < 67) {
                                                                            if (diffAFD < 1) {
                                                                                return 2;
                                                                            }
                                                                            else if (diffAFD >= 1) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                        else if (mu_mv_mag_sq >= 67) {
                                                                            if (Cs < 405) {
                                                                                return 4;
                                                                            }
                                                                            else if (Cs >= 405) {
                                                                                return 2;
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (refDCval >= 39) {
                                                            if (RsCsDiff < 4) {
                                                                return 4;
                                                            }
                                                            else if (RsCsDiff >= 4) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                    else if (TSC >= 5) {
                                                        if (SC < 1166) {
                                                            if (Rs < 448) {
                                                                return 4;
                                                            }
                                                            else if (Rs >= 448) {
                                                                return 2;
                                                            }
                                                        }
                                                        else if (SC >= 1166) {
                                                            if (Cs < 675) {
                                                                if (AFD < 8) {
                                                                    if (diffAFD < -1) {
                                                                        return 1;
                                                                    }
                                                                    else if (diffAFD >= -1) {
                                                                        if (tcor < 77) {
                                                                            return 4;
                                                                        }
                                                                        else if (tcor >= 77) {
                                                                            if (mu_mv_mag_sq < 52) {
                                                                                if (SC < 1307) {
                                                                                    if (Cs < 528) {
                                                                                        return 1;
                                                                                    }
                                                                                    else if (Cs >= 528) {
                                                                                        return 4;
                                                                                    }
                                                                                }
                                                                                else if (SC >= 1307) {
                                                                                    return 1;
                                                                                }
                                                                            }
                                                                            else if (mu_mv_mag_sq >= 52) {
                                                                                if (Rs < 763) {
                                                                                    if (CsDiff < 2) {
                                                                                        if (Rs < 761) {
                                                                                            return 4;
                                                                                        }
                                                                                        else if (Rs >= 761) {
                                                                                            return 1;
                                                                                        }
                                                                                    }
                                                                                    else if (CsDiff >= 2) {
                                                                                        return 1;
                                                                                    }
                                                                                }
                                                                                else if (Rs >= 763) {
                                                                                    return 4;
                                                                                }
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                                else if (AFD >= 8) {
                                                                    if (Rs < 832) {
                                                                        return 2;
                                                                    }
                                                                    else if (Rs >= 832) {
                                                                        return 8;
                                                                    }
                                                                }
                                                            }
                                                            else if (Cs >= 675) {
                                                                if (diffTSC < 1) {
                                                                    if (AFD < 6) {
                                                                        return 8;
                                                                    }
                                                                    else if (AFD >= 6) {
                                                                        if (Rs < 815) {
                                                                            return 2;
                                                                        }
                                                                        else if (Rs >= 815) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                }
                                                                else if (diffTSC >= 1) {
                                                                    if (Cs < 853) {
                                                                        return 8;
                                                                    }
                                                                    else if (Cs >= 853) {
                                                                        return 2;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                            else if (ssDCval >= 86) {
                                                if (refDCval < 98) {
                                                    if (SC < 781) {
                                                        return 2;
                                                    }
                                                    else if (SC >= 781) {
                                                        if (ssDCval < 93) {
                                                            return 4;
                                                        }
                                                        else if (ssDCval >= 93) {
                                                            return 2;
                                                        }
                                                    }
                                                }
                                                else if (refDCval >= 98) {
                                                    if (Rs < 663) {
                                                        return 8;
                                                    }
                                                    else if (Rs >= 663) {
                                                        if (Cs < 692) {
                                                            return 4;
                                                        }
                                                        else if (Cs >= 692) {
                                                            return 2;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (TSC >= 6) {
                                            if (Cs < 423) {
                                                return 1;
                                            }
                                            else if (Cs >= 423) {
                                                if (mu_mv_mag_sq < 28) {
                                                    if (Rs < 1104) {
                                                        if (MVDiff < 23) {
                                                            if (MVDiff < 22) {
                                                                return 4;
                                                            }
                                                            else if (MVDiff >= 22) {
                                                                return 2;
                                                            }
                                                        }
                                                        else if (MVDiff >= 23) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (Rs >= 1104) {
                                                        if (Rs < 1262) {
                                                            return 2;
                                                        }
                                                        else if (Rs >= 1262) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                                else if (mu_mv_mag_sq >= 28) {
                                                    if (mu_mv_mag_sq < 42) {
                                                        if (TSCindex < 8) {
                                                            if (Rs < 688) {
                                                                if (Cs < 484) {
                                                                    return 4;
                                                                }
                                                                else if (Cs >= 484) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (Rs >= 688) {
                                                                if (mcTcor < 77) {
                                                                    if (posBalance < 59) {
                                                                        if (mcTcor < 75) {
                                                                            return 1;
                                                                        }
                                                                        else if (mcTcor >= 75) {
                                                                            if (RsDiff < 3) {
                                                                                return 4;
                                                                            }
                                                                            else if (RsDiff >= 3) {
                                                                                return 1;
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (posBalance >= 59) {
                                                                        if (Rs < 726) {
                                                                            return 4;
                                                                        }
                                                                        else if (Rs >= 726) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                }
                                                                else if (mcTcor >= 77) {
                                                                    return 1;
                                                                }
                                                            }
                                                        }
                                                        else if (TSCindex >= 8) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (mu_mv_mag_sq >= 42) {
                                                        if (RsDiff < 3) {
                                                            return 4;
                                                        }
                                                        else if (RsDiff >= 3) {
                                                            return 1;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (Cs >= 1295) {
                                    if (refDCval < 91) {
                                        if (TSC < 5) {
                                            return 2;
                                        }
                                        else if (TSC >= 5) {
                                            if (Cs < 1587) {
                                                if (mu_mv_mag_sq < 39) {
                                                    if (refDCval < 75) {
                                                        return 1;
                                                    }
                                                    else if (refDCval >= 75) {
                                                        return 2;
                                                    }
                                                }
                                                else if (mu_mv_mag_sq >= 39) {
                                                    if (Rs < 1684) {
                                                        return 8;
                                                    }
                                                    else if (Rs >= 1684) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                            else if (Cs >= 1587) {
                                                if (mu_mv_mag_sq < 20) {
                                                    if (TSC < 9) {
                                                        return 4;
                                                    }
                                                    else if (TSC >= 9) {
                                                        return 8;
                                                    }
                                                }
                                                else if (mu_mv_mag_sq >= 20) {
                                                    if (TSC < 9) {
                                                        return 8;
                                                    }
                                                    else if (TSC >= 9) {
                                                        if (Rs < 2211) {
                                                            return 8;
                                                        }
                                                        else if (Rs >= 2211) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (refDCval >= 91) {
                                        if (Rs < 2008) {
                                            if (refDCval < 96) {
                                                if (mu_mv_mag_sq < 64) {
                                                    if (Cs < 1712) {
                                                        return 2;
                                                    }
                                                    else if (Cs >= 1712) {
                                                        return 8;
                                                    }
                                                }
                                                else if (mu_mv_mag_sq >= 64) {
                                                    return 8;
                                                }
                                            }
                                            else if (refDCval >= 96) {
                                                return 8;
                                            }
                                        }
                                        else if (Rs >= 2008) {
                                            if (Cs < 2174) {
                                                if (Rs < 2031) {
                                                    if (AFD < 9) {
                                                        return 8;
                                                    }
                                                    else if (AFD >= 9) {
                                                        return 2;
                                                    }
                                                }
                                                else if (Rs >= 2031) {
                                                    return 2;
                                                }
                                            }
                                            else if (Cs >= 2174) {
                                                if (SC < 4581) {
                                                    return 2;
                                                }
                                                else if (SC >= 4581) {
                                                    return 8;
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                            else if (mcTcor >= 90) {
                                if (refDCval < 86) {
                                    if (SC < 1665) {
                                        if (tcor < 89) {
                                            if (mu_mv_mag_sq < 27) {
                                                if (refDCval < 78) {
                                                    if (MVDiff < 51) {
                                                        return 2;
                                                    }
                                                    else if (MVDiff >= 51) {
                                                        if (negBalance < 55) {
                                                            if (MVDiff < 587) {
                                                                if (gchDC < 1) {
                                                                    return 1;
                                                                }
                                                                else if (gchDC >= 1) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (MVDiff >= 587) {
                                                                return 2;
                                                            }
                                                        }
                                                        else if (negBalance >= 55) {
                                                            return 1;
                                                        }
                                                    }
                                                }
                                                else if (refDCval >= 78) {
                                                    return 1;
                                                }
                                            }
                                            else if (mu_mv_mag_sq >= 27) {
                                                if (ssDCval < 54) {
                                                    if (Rs < 220) {
                                                        if (Rs < 9) {
                                                            return 4;
                                                        }
                                                        else if (Rs >= 9) {
                                                            return 2;
                                                        }
                                                    }
                                                    else if (Rs >= 220) {
                                                        return 4;
                                                    }
                                                }
                                                else if (ssDCval >= 54) {
                                                    if (AFD < 9) {
                                                        if (RsDiff < 4) {
                                                            return 8;
                                                        }
                                                        else if (RsDiff >= 4) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (AFD >= 9) {
                                                        if (Rs < 836) {
                                                            return 4;
                                                        }
                                                        else if (Rs >= 836) {
                                                            return 2;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (tcor >= 89) {
                                            if (mu_mv_mag_sq < 33) {
                                                if (Rs < 266) {
                                                    if (MVDiff < 76) {
                                                        if (RsCsDiff < 3) {
                                                            return 2;
                                                        }
                                                        else if (RsCsDiff >= 3) {
                                                            if (SC < 291) {
                                                                return 4;
                                                            }
                                                            else if (SC >= 291) {
                                                                return 2;
                                                            }
                                                        }
                                                    }
                                                    else if (MVDiff >= 76) {
                                                        if (Rs < 13) {
                                                            return 4;
                                                        }
                                                        else if (Rs >= 13) {
                                                            return 1;
                                                        }
                                                    }
                                                }
                                                else if (Rs >= 266) {
                                                    if (RsCsDiff < 16) {
                                                        if (negBalance < 39) {
                                                            return 8;
                                                        }
                                                        else if (negBalance >= 39) {
                                                            if (TSC < 5) {
                                                                if (SC < 1293) {
                                                                    return 4;
                                                                }
                                                                else if (SC >= 1293) {
                                                                    if (Rs < 813) {
                                                                        return 8;
                                                                    }
                                                                    else if (Rs >= 813) {
                                                                        return 4;
                                                                    }
                                                                }
                                                            }
                                                            else if (TSC >= 5) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                    else if (RsCsDiff >= 16) {
                                                        if (SC < 1186) {
                                                            if (RsDiff < 5) {
                                                                return 8;
                                                            }
                                                            else if (RsDiff >= 5) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (SC >= 1186) {
                                                            return 2;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (mu_mv_mag_sq >= 33) {
                                                if (posBalance < 6) {
                                                    if (ssDCval < 55) {
                                                        if (MVDiff < 30) {
                                                            return 4;
                                                        }
                                                        else if (MVDiff >= 30) {
                                                            return 2;
                                                        }
                                                    }
                                                    else if (ssDCval >= 55) {
                                                        return 8;
                                                    }
                                                }
                                                else if (posBalance >= 6) {
                                                    if (diffRsCsdiff < -3) {
                                                        if (Rs < 350) {
                                                            if (Rs < 149) {
                                                                return 1;
                                                            }
                                                            else if (Rs >= 149) {
                                                                return 2;
                                                            }
                                                        }
                                                        else if (Rs >= 350) {
                                                            if (RsCsDiff < 7) {
                                                                if (diffAFD < -1) {
                                                                    if (TSC < 3) {
                                                                        return 4;
                                                                    }
                                                                    else if (TSC >= 3) {
                                                                        return 8;
                                                                    }
                                                                }
                                                                else if (diffAFD >= -1) {
                                                                    if (MVDiff < 4) {
                                                                        return 8;
                                                                    }
                                                                    else if (MVDiff >= 4) {
                                                                        return 4;
                                                                    }
                                                                }
                                                            }
                                                            else if (RsCsDiff >= 7) {
                                                                if (Rs < 887) {
                                                                    return 4;
                                                                }
                                                                else if (Rs >= 887) {
                                                                    return 2;
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (diffRsCsdiff >= -3) {
                                                        if (mcTcor < 98) {
                                                            if (TSCindex < 4) {
                                                                return 8;
                                                            }
                                                            else if (TSCindex >= 4) {
                                                                if (posBalance < 54) {
                                                                    if (negBalance < 50) {
                                                                        if (Cs < 524) {
                                                                            if (SC < 1240) {
                                                                                return 4;
                                                                            }
                                                                            else if (SC >= 1240) {
                                                                                return 8;
                                                                            }
                                                                        }
                                                                        else if (Cs >= 524) {
                                                                            if (mcTcor < 91) {
                                                                                return 8;
                                                                            }
                                                                            else if (mcTcor >= 91) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (negBalance >= 50) {
                                                                        return 8;
                                                                    }
                                                                }
                                                                else if (posBalance >= 54) {
                                                                    if (MVDiff < 7) {
                                                                        return 8;
                                                                    }
                                                                    else if (MVDiff >= 7) {
                                                                        if (Rs < 905) {
                                                                            return 4;
                                                                        }
                                                                        else if (Rs >= 905) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (mcTcor >= 98) {
                                                            if (SC < 560) {
                                                                return 4;
                                                            }
                                                            else if (SC >= 560) {
                                                                if (ssDCval < 82) {
                                                                    if (RsCsDiff < 8) {
                                                                        if (posBalance < 44) {
                                                                            if (refDCval < 46) {
                                                                                return 4;
                                                                            }
                                                                            else if (refDCval >= 46) {
                                                                                return 8;
                                                                            }
                                                                        }
                                                                        else if (posBalance >= 44) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                    else if (RsCsDiff >= 8) {
                                                                        return 2;
                                                                    }
                                                                }
                                                                else if (ssDCval >= 82) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (SC >= 1665) {
                                        if (negBalance < 47) {
                                            if (Cs < 620) {
                                                if (SC < 2944) {
                                                    if (ssDCval < 71) {
                                                        if (Cs < 432) {
                                                            return 2;
                                                        }
                                                        else if (Cs >= 432) {
                                                            if (RsCsDiff < 35) {
                                                                return 4;
                                                            }
                                                            else if (RsCsDiff >= 35) {
                                                                return 2;
                                                            }
                                                        }
                                                    }
                                                    else if (ssDCval >= 71) {
                                                        return 4;
                                                    }
                                                }
                                                else if (SC >= 2944) {
                                                    return 2;
                                                }
                                            }
                                            else if (Cs >= 620) {
                                                if (ssDCval < 72) {
                                                    if (refDCval < 44) {
                                                        if (mu_mv_mag_sq < 75) {
                                                            if (refDCval < 39) {
                                                                if (MVDiff < 24) {
                                                                    if (diffMVdiffVal < -6) {
                                                                        return 2;
                                                                    }
                                                                    else if (diffMVdiffVal >= -6) {
                                                                        return 1;
                                                                    }
                                                                }
                                                                else if (MVDiff >= 24) {
                                                                    if (MVDiff < 41) {
                                                                        return 2;
                                                                    }
                                                                    else if (MVDiff >= 41) {
                                                                        return 4;
                                                                    }
                                                                }
                                                            }
                                                            else if (refDCval >= 39) {
                                                                if (RsCsDiff < 29) {
                                                                    if (Cs < 1554) {
                                                                        if (MVDiff < 19) {
                                                                            if (mu_mv_mag_sq < 63) {
                                                                                return 1;
                                                                            }
                                                                            else if (mu_mv_mag_sq >= 63) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                        else if (MVDiff >= 19) {
                                                                            if (SC < 3346) {
                                                                                if (SC < 3329) {
                                                                                    return 4;
                                                                                }
                                                                                else if (SC >= 3329) {
                                                                                    return 8;
                                                                                }
                                                                            }
                                                                            else if (SC >= 3346) {
                                                                                if (diffMVdiffVal < -6) {
                                                                                    return 1;
                                                                                }
                                                                                else if (diffMVdiffVal >= -6) {
                                                                                    return 4;
                                                                                }
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (Cs >= 1554) {
                                                                        return 1;
                                                                    }
                                                                }
                                                                else if (RsCsDiff >= 29) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                        else if (mu_mv_mag_sq >= 75) {
                                                            if (SC < 2047) {
                                                                return 8;
                                                            }
                                                            else if (SC >= 2047) {
                                                                if (mu_mv_mag_sq < 77) {
                                                                    return 8;
                                                                }
                                                                else if (mu_mv_mag_sq >= 77) {
                                                                    if (Rs < 1858) {
                                                                        if (AFD < 5) {
                                                                            return 4;
                                                                        }
                                                                        else if (AFD >= 5) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                    else if (Rs >= 1858) {
                                                                        return 4;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (refDCval >= 44) {
                                                        if (Rs < 1181) {
                                                            if (Cs < 726) {
                                                                return 1;
                                                            }
                                                            else if (Cs >= 726) {
                                                                if (RsCsDiff < 45) {
                                                                    if (diffMVdiffVal < 10) {
                                                                        if (negBalance < 44) {
                                                                            if (refDCval < 49) {
                                                                                return 1;
                                                                            }
                                                                            else if (refDCval >= 49) {
                                                                                if (refDCval < 57) {
                                                                                    if (mu_mv_mag_sq < 16) {
                                                                                        if (Rs < 1097) {
                                                                                            return 4;
                                                                                        }
                                                                                        else if (Rs >= 1097) {
                                                                                            return 1;
                                                                                        }
                                                                                    }
                                                                                    else if (mu_mv_mag_sq >= 16) {
                                                                                        return 4;
                                                                                    }
                                                                                }
                                                                                else if (refDCval >= 57) {
                                                                                    if (Cs < 906) {
                                                                                        return 1;
                                                                                    }
                                                                                    else if (Cs >= 906) {
                                                                                        return 4;
                                                                                    }
                                                                                }
                                                                            }
                                                                        }
                                                                        else if (negBalance >= 44) {
                                                                            if (RsCsDiff < 9) {
                                                                                return 4;
                                                                            }
                                                                            else if (RsCsDiff >= 9) {
                                                                                return 1;
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (diffMVdiffVal >= 10) {
                                                                        if (TSC < 6) {
                                                                            return 1;
                                                                        }
                                                                        else if (TSC >= 6) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                }
                                                                else if (RsCsDiff >= 45) {
                                                                    return 2;
                                                                }
                                                            }
                                                        }
                                                        else if (Rs >= 1181) {
                                                            if (tcor < 83) {
                                                                if (Cs < 1020) {
                                                                    return 8;
                                                                }
                                                                else if (Cs >= 1020) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (tcor >= 83) {
                                                                if (mu_mv_mag_sq < 83) {
                                                                    if (RsCsDiff < 7) {
                                                                        return 1;
                                                                    }
                                                                    else if (RsCsDiff >= 7) {
                                                                        if (TSC < 5) {
                                                                            if (refDCval < 56) {
                                                                                if (ssDCval < 51) {
                                                                                    if (Cs < 822) {
                                                                                        return 8;
                                                                                    }
                                                                                    else if (Cs >= 822) {
                                                                                        if (mu_mv_mag_sq < 15) {
                                                                                            if (SC < 2301) {
                                                                                                return 1;
                                                                                            }
                                                                                            else if (SC >= 2301) {
                                                                                                return 8;
                                                                                            }
                                                                                        }
                                                                                        else if (mu_mv_mag_sq >= 15) {
                                                                                            if (mcTcor < 91) {
                                                                                                if (SC < 2223) {
                                                                                                    return 1;
                                                                                                }
                                                                                                else if (SC >= 2223) {
                                                                                                    return 4;
                                                                                                }
                                                                                            }
                                                                                            else if (mcTcor >= 91) {
                                                                                                return 1;
                                                                                            }
                                                                                        }
                                                                                    }
                                                                                }
                                                                                else if (ssDCval >= 51) {
                                                                                    if (tcor < 92) {
                                                                                        if (mu_mv_mag_sq < 62) {
                                                                                            if (SC < 2205) {
                                                                                                return 1;
                                                                                            }
                                                                                            else if (SC >= 2205) {
                                                                                                if (posBalance < 50) {
                                                                                                    if (Rs < 1450) {
                                                                                                        return 4;
                                                                                                    }
                                                                                                    else if (Rs >= 1450) {
                                                                                                        return 1;
                                                                                                    }
                                                                                                }
                                                                                                else if (posBalance >= 50) {
                                                                                                    return 1;
                                                                                                }
                                                                                            }
                                                                                        }
                                                                                        else if (mu_mv_mag_sq >= 62) {
                                                                                            return 4;
                                                                                        }
                                                                                    }
                                                                                    else if (tcor >= 92) {
                                                                                        if (MVDiff < 3) {
                                                                                            return 8;
                                                                                        }
                                                                                        else if (MVDiff >= 3) {
                                                                                            if (diffRsCsdiff < -4) {
                                                                                                if (Cs < 948) {
                                                                                                    return 1;
                                                                                                }
                                                                                                else if (Cs >= 948) {
                                                                                                    if (Rs < 1293) {
                                                                                                        return 1;
                                                                                                    }
                                                                                                    else if (Rs >= 1293) {
                                                                                                        return 4;
                                                                                                    }
                                                                                                }
                                                                                            }
                                                                                            else if (diffRsCsdiff >= -4) {
                                                                                                return 1;
                                                                                            }
                                                                                        }
                                                                                    }
                                                                                }
                                                                            }
                                                                            else if (refDCval >= 56) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                        else if (TSC >= 5) {
                                                                            if (ssDCval < 58) {
                                                                                if (Cs < 954) {
                                                                                    return 1;
                                                                                }
                                                                                else if (Cs >= 954) {
                                                                                    if (Cs < 1063) {
                                                                                        if (SC < 2537) {
                                                                                            return 8;
                                                                                        }
                                                                                        else if (SC >= 2537) {
                                                                                            return 4;
                                                                                        }
                                                                                    }
                                                                                    else if (Cs >= 1063) {
                                                                                        if (MVDiff < 4) {
                                                                                            return 8;
                                                                                        }
                                                                                        else if (MVDiff >= 4) {
                                                                                            if (Cs < 1117) {
                                                                                                return 1;
                                                                                            }
                                                                                            else if (Cs >= 1117) {
                                                                                                return 4;
                                                                                            }
                                                                                        }
                                                                                    }
                                                                                }
                                                                            }
                                                                            else if (ssDCval >= 58) {
                                                                                if (Rs < 2830) {
                                                                                    return 1;
                                                                                }
                                                                                else if (Rs >= 2830) {
                                                                                    return 4;
                                                                                }
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                                else if (mu_mv_mag_sq >= 83) {
                                                                    if (negBalance < 25) {
                                                                        return 1;
                                                                    }
                                                                    else if (negBalance >= 25) {
                                                                        return 8;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (ssDCval >= 72) {
                                                    if (negBalance < 36) {
                                                        if (MVDiff < 14) {
                                                            if (negBalance < 34) {
                                                                return 8;
                                                            }
                                                            else if (negBalance >= 34) {
                                                                if (Cs < 2651) {
                                                                    return 1;
                                                                }
                                                                else if (Cs >= 2651) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                        else if (MVDiff >= 14) {
                                                            if (Rs < 1749) {
                                                                return 4;
                                                            }
                                                            else if (Rs >= 1749) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                    else if (negBalance >= 36) {
                                                        if (CsDiff < 6) {
                                                            if (SC < 3136) {
                                                                if (mu_mv_mag_sq < 63) {
                                                                    if (SC < 3052) {
                                                                        return 8;
                                                                    }
                                                                    else if (SC >= 3052) {
                                                                        return 1;
                                                                    }
                                                                }
                                                                else if (mu_mv_mag_sq >= 63) {
                                                                    if (Cs < 1141) {
                                                                        return 1;
                                                                    }
                                                                    else if (Cs >= 1141) {
                                                                        return 4;
                                                                    }
                                                                }
                                                            }
                                                            else if (SC >= 3136) {
                                                                if (TSC < 8) {
                                                                    return 4;
                                                                }
                                                                else if (TSC >= 8) {
                                                                    if (Rs < 2232) {
                                                                        return 2;
                                                                    }
                                                                    else if (Rs >= 2232) {
                                                                        return 4;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (CsDiff >= 6) {
                                                            if (ssDCval < 80) {
                                                                if (Rs < 1753) {
                                                                    return 8;
                                                                }
                                                                else if (Rs >= 1753) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (ssDCval >= 80) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (negBalance >= 47) {
                                            if (Rs < 1802) {
                                                if (MVDiff < 3) {
                                                    return 2;
                                                }
                                                else if (MVDiff >= 3) {
                                                    if (RsCsDiff < 46) {
                                                        if (refDCval < 52) {
                                                            if (CsDiff < 2) {
                                                                return 2;
                                                            }
                                                            else if (CsDiff >= 2) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (refDCval >= 52) {
                                                            if (Cs < 722) {
                                                                if (SC < 1722) {
                                                                    return 8;
                                                                }
                                                                else if (SC >= 1722) {
                                                                    return 1;
                                                                }
                                                            }
                                                            else if (Cs >= 722) {
                                                                if (mu_mv_mag_sq < 48) {
                                                                    if (RsDiff < 4) {
                                                                        if (MVDiff < 13) {
                                                                            return 8;
                                                                        }
                                                                        else if (MVDiff >= 13) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                    else if (RsDiff >= 4) {
                                                                        if (Cs < 1025) {
                                                                            return 4;
                                                                        }
                                                                        else if (Cs >= 1025) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                }
                                                                else if (mu_mv_mag_sq >= 48) {
                                                                    if (Rs < 763) {
                                                                        return 8;
                                                                    }
                                                                    else if (Rs >= 763) {
                                                                        return 4;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (RsCsDiff >= 46) {
                                                        if (Cs < 883) {
                                                            return 2;
                                                        }
                                                        else if (Cs >= 883) {
                                                            return 1;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (Rs >= 1802) {
                                                if (Cs < 1694) {
                                                    if (Cs < 1545) {
                                                        if (TSC < 7) {
                                                            return 4;
                                                        }
                                                        else if (TSC >= 7) {
                                                            return 1;
                                                        }
                                                    }
                                                    else if (Cs >= 1545) {
                                                        if (refDCval < 40) {
                                                            return 1;
                                                        }
                                                        else if (refDCval >= 40) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                                else if (Cs >= 1694) {
                                                    if (SC < 5639) {
                                                        if (Rs < 2159) {
                                                            return 2;
                                                        }
                                                        else if (Rs >= 2159) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (SC >= 5639) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (refDCval >= 86) {
                                    if (negBalance < 43) {
                                        if (AFD < 8) {
                                            if (negBalance < 36) {
                                                if (Rs < 464) {
                                                    return 4;
                                                }
                                                else if (Rs >= 464) {
                                                    if (Rs < 4014) {
                                                        if (posBalance < 94) {
                                                            return 8;
                                                        }
                                                        else if (posBalance >= 94) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (Rs >= 4014) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                            else if (negBalance >= 36) {
                                                if (Rs < 265) {
                                                    return 2;
                                                }
                                                else if (Rs >= 265) {
                                                    if (TSC < 6) {
                                                        return 4;
                                                    }
                                                    else if (TSC >= 6) {
                                                        return 1;
                                                    }
                                                }
                                            }
                                        }
                                        else if (AFD >= 8) {
                                            if (ssDCval < 95) {
                                                if (ssDCval < 89) {
                                                    if (Cs < 1649) {
                                                        return 4;
                                                    }
                                                    else if (Cs >= 1649) {
                                                        return 1;
                                                    }
                                                }
                                                else if (ssDCval >= 89) {
                                                    if (negBalance < 38) {
                                                        return 8;
                                                    }
                                                    else if (negBalance >= 38) {
                                                        if (Rs < 2075) {
                                                            return 8;
                                                        }
                                                        else if (Rs >= 2075) {
                                                            if (Rs < 2771) {
                                                                return 1;
                                                            }
                                                            else if (Rs >= 2771) {
                                                                return 2;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                            else if (ssDCval >= 95) {
                                                if (RsCsDiff < 73) {
                                                    if (diffAFD < 2) {
                                                        if (negBalance < 42) {
                                                            return 2;
                                                        }
                                                        else if (negBalance >= 42) {
                                                            return 1;
                                                        }
                                                    }
                                                    else if (diffAFD >= 2) {
                                                        return 8;
                                                    }
                                                }
                                                else if (RsCsDiff >= 73) {
                                                    return 4;
                                                }
                                            }
                                        }
                                    }
                                    else if (negBalance >= 43) {
                                        if (mu_mv_mag_sq < 80) {
                                            if (TSC < 5) {
                                                if (Rs < 462) {
                                                    if (diffRsCsdiff < -1) {
                                                        if (Rs < 150) {
                                                            return 8;
                                                        }
                                                        else if (Rs >= 150) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (diffRsCsdiff >= -1) {
                                                        if (MVDiff < 10) {
                                                            if (diffMVdiffVal < -3) {
                                                                return 4;
                                                            }
                                                            else if (diffMVdiffVal >= -3) {
                                                                return 2;
                                                            }
                                                        }
                                                        else if (MVDiff >= 10) {
                                                            return 2;
                                                        }
                                                    }
                                                }
                                                else if (Rs >= 462) {
                                                    if (Rs < 626) {
                                                        return 8;
                                                    }
                                                    else if (Rs >= 626) {
                                                        if (Rs < 973) {
                                                            return 4;
                                                        }
                                                        else if (Rs >= 973) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (TSC >= 5) {
                                                if (Cs < 364) {
                                                    return 1;
                                                }
                                                else if (Cs >= 364) {
                                                    if (Rs < 2424) {
                                                        if (MVDiff < 5) {
                                                            if (AFD < 9) {
                                                                return 4;
                                                            }
                                                            else if (AFD >= 9) {
                                                                if (mcTcor < 93) {
                                                                    if (TSC < 8) {
                                                                        return 4;
                                                                    }
                                                                    else if (TSC >= 8) {
                                                                        return 8;
                                                                    }
                                                                }
                                                                else if (mcTcor >= 93) {
                                                                    return 2;
                                                                }
                                                            }
                                                        }
                                                        else if (MVDiff >= 5) {
                                                            if (TSC < 6) {
                                                                if (RsDiff < 3) {
                                                                    if (SC < 1218) {
                                                                        return 8;
                                                                    }
                                                                    else if (SC >= 1218) {
                                                                        if (Cs < 667) {
                                                                            return 4;
                                                                        }
                                                                        else if (Cs >= 667) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                }
                                                                else if (RsDiff >= 3) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (TSC >= 6) {
                                                                if (mu_mv_mag_sq < 73) {
                                                                    return 4;
                                                                }
                                                                else if (mu_mv_mag_sq >= 73) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (Rs >= 2424) {
                                                        if (tcor < 92) {
                                                            if (RsDiff < 8) {
                                                                return 8;
                                                            }
                                                            else if (RsDiff >= 8) {
                                                                return 2;
                                                            }
                                                        }
                                                        else if (tcor >= 92) {
                                                            return 1;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (mu_mv_mag_sq >= 80) {
                                            if (SC < 4243) {
                                                if (gchDC < 3) {
                                                    return 2;
                                                }
                                                else if (gchDC >= 3) {
                                                    return 4;
                                                }
                                            }
                                            else if (SC >= 4243) {
                                                if (RsDiff < 6) {
                                                    return 1;
                                                }
                                                else if (RsDiff >= 6) {
                                                    if (SC < 4894) {
                                                        return 2;
                                                    }
                                                    else if (SC >= 4894) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                        }
                        else if (refDCval >= 104) {
                            if (SC < 3261) {
                                if (mu_mv_mag_sq < 41) {
                                    if (negBalance < 31) {
                                        if (ssDCval < 125) {
                                            if (Rs < 872) {
                                                return 1;
                                            }
                                            else if (Rs >= 872) {
                                                return 8;
                                            }
                                        }
                                        else if (ssDCval >= 125) {
                                            if (MVDiff < 8) {
                                                return 4;
                                            }
                                            else if (MVDiff >= 8) {
                                                if (posBalance < 34) {
                                                    if (Rs < 1137) {
                                                        return 4;
                                                    }
                                                    else if (Rs >= 1137) {
                                                        return 1;
                                                    }
                                                }
                                                else if (posBalance >= 34) {
                                                    return 1;
                                                }
                                            }
                                        }
                                    }
                                    else if (negBalance >= 31) {
                                        if (tcor < 88) {
                                            if (refDCval < 110) {
                                                return 8;
                                            }
                                            else if (refDCval >= 110) {
                                                if (MVDiff < 42) {
                                                    if (ssDCval < 130) {
                                                        return 4;
                                                    }
                                                    else if (ssDCval >= 130) {
                                                        if (posBalance < 39) {
                                                            return 4;
                                                        }
                                                        else if (posBalance >= 39) {
                                                            if (refDCval < 139) {
                                                                if (diffAFD < -4) {
                                                                    return 4;
                                                                }
                                                                else if (diffAFD >= -4) {
                                                                    return 1;
                                                                }
                                                            }
                                                            else if (refDCval >= 139) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (MVDiff >= 42) {
                                                    if (SC < 779) {
                                                        return 1;
                                                    }
                                                    else if (SC >= 779) {
                                                        return 2;
                                                    }
                                                }
                                            }
                                        }
                                        else if (tcor >= 88) {
                                            if (Rs < 421) {
                                                if (Rs < 226) {
                                                    return 1;
                                                }
                                                else if (Rs >= 226) {
                                                    return 4;
                                                }
                                            }
                                            else if (Rs >= 421) {
                                                if (Cs < 1094) {
                                                    if (RsDiff < 3) {
                                                        if (Rs < 845) {
                                                            if (Cs < 870) {
                                                                return 4;
                                                            }
                                                            else if (Cs >= 870) {
                                                                return 2;
                                                            }
                                                        }
                                                        else if (Rs >= 845) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (RsDiff >= 3) {
                                                        if (mu_mv_mag_sq < 33) {
                                                            return 2;
                                                        }
                                                        else if (mu_mv_mag_sq >= 33) {
                                                            if (Rs < 1755) {
                                                                return 2;
                                                            }
                                                            else if (Rs >= 1755) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (Cs >= 1094) {
                                                    return 4;
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (mu_mv_mag_sq >= 41) {
                                    if (ssDCval < 147) {
                                        if (ssDCval < 119) {
                                            if (SC < 783) {
                                                if (TSC < 6) {
                                                    if (Rs < 734) {
                                                        return 8;
                                                    }
                                                    else if (Rs >= 734) {
                                                        return 1;
                                                    }
                                                }
                                                else if (TSC >= 6) {
                                                    if (Rs < 358) {
                                                        return 4;
                                                    }
                                                    else if (Rs >= 358) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                            else if (SC >= 783) {
                                                if (gchDC < 5) {
                                                    return 8;
                                                }
                                                else if (gchDC >= 5) {
                                                    if (TSC < 6) {
                                                        return 8;
                                                    }
                                                    else if (TSC >= 6) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                        }
                                        else if (ssDCval >= 119) {
                                            if (mcTcor < 95) {
                                                if (diffRsCsdiff < 8) {
                                                    return 1;
                                                }
                                                else if (diffRsCsdiff >= 8) {
                                                    return 8;
                                                }
                                            }
                                            else if (mcTcor >= 95) {
                                                if (ssDCval < 122) {
                                                    if (gchDC < 2) {
                                                        return 2;
                                                    }
                                                    else if (gchDC >= 2) {
                                                        return 8;
                                                    }
                                                }
                                                else if (ssDCval >= 122) {
                                                    return 8;
                                                }
                                            }
                                        }
                                    }
                                    else if (ssDCval >= 147) {
                                        if (AFD < 3) {
                                            if (Rs < 443) {
                                                return 8;
                                            }
                                            else if (Rs >= 443) {
                                                if (Rs < 563) {
                                                    return 2;
                                                }
                                                else if (Rs >= 563) {
                                                    return 4;
                                                }
                                            }
                                        }
                                        else if (AFD >= 3) {
                                            if (Rs < 370) {
                                                return 1;
                                            }
                                            else if (Rs >= 370) {
                                                return 4;
                                            }
                                        }
                                    }
                                }
                            }
                            else if (SC >= 3261) {
                                if (ssDCval < 119) {
                                    if (negBalance < 39) {
                                        if (CsDiff < 6) {
                                            if (ssDCval < 108) {
                                                if (TSC < 6) {
                                                    if (Cs < 1676) {
                                                        if (CsDiff < 4) {
                                                            return 8;
                                                        }
                                                        else if (CsDiff >= 4) {
                                                            if (refDCval < 106) {
                                                                return 8;
                                                            }
                                                            else if (refDCval >= 106) {
                                                                if (Rs < 3337) {
                                                                    return 4;
                                                                }
                                                                else if (Rs >= 3337) {
                                                                    if (diffMVdiffVal < 4) {
                                                                        return 8;
                                                                    }
                                                                    else if (diffMVdiffVal >= 4) {
                                                                        return 4;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (Cs >= 1676) {
                                                        if (RsDiff < 5) {
                                                            if (diffRsCsdiff < -15) {
                                                                return 4;
                                                            }
                                                            else if (diffRsCsdiff >= -15) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (RsDiff >= 5) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                                else if (TSC >= 6) {
                                                    return 4;
                                                }
                                            }
                                            else if (ssDCval >= 108) {
                                                if (RsDiff < 5) {
                                                    if (RsCsDiff < 9) {
                                                        return 4;
                                                    }
                                                    else if (RsCsDiff >= 9) {
                                                        return 8;
                                                    }
                                                }
                                                else if (RsDiff >= 5) {
                                                    if (TSCindex < 8) {
                                                        if (mu_mv_mag_sq < 61) {
                                                            return 4;
                                                        }
                                                        else if (mu_mv_mag_sq >= 61) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (TSCindex >= 8) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                        }
                                        else if (CsDiff >= 6) {
                                            if (tcor < 94) {
                                                if (SC < 5638) {
                                                    if (Cs < 1983) {
                                                        if (Rs < 3329) {
                                                            return 2;
                                                        }
                                                        else if (Rs >= 3329) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (Cs >= 1983) {
                                                        if (mu_mv_mag_sq < 65) {
                                                            return 4;
                                                        }
                                                        else if (mu_mv_mag_sq >= 65) {
                                                            if (SC < 5012) {
                                                                return 4;
                                                            }
                                                            else if (SC >= 5012) {
                                                                return 2;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (SC >= 5638) {
                                                    return 4;
                                                }
                                            }
                                            else if (tcor >= 94) {
                                                if (mu_mv_mag_sq < 68) {
                                                    if (diffAFD < 1) {
                                                        return 2;
                                                    }
                                                    else if (diffAFD >= 1) {
                                                        return 8;
                                                    }
                                                }
                                                else if (mu_mv_mag_sq >= 68) {
                                                    if (RsCsDiff < 107) {
                                                        return 2;
                                                    }
                                                    else if (RsCsDiff >= 107) {
                                                        if (Rs < 3633) {
                                                            return 2;
                                                        }
                                                        else if (Rs >= 3633) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (negBalance >= 39) {
                                        if (Cs < 2132) {
                                            if (ssDCval < 117) {
                                                if (diffRsCsdiff < -49) {
                                                    return 8;
                                                }
                                                else if (diffRsCsdiff >= -49) {
                                                    if (SC < 5463) {
                                                        return 4;
                                                    }
                                                    else if (SC >= 5463) {
                                                        if (Cs < 1869) {
                                                            return 8;
                                                        }
                                                        else if (Cs >= 1869) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (ssDCval >= 117) {
                                                if (Cs < 1593) {
                                                    return 4;
                                                }
                                                else if (Cs >= 1593) {
                                                    if (MVDiff < 8) {
                                                        if (mu_mv_mag_sq < 57) {
                                                            return 4;
                                                        }
                                                        else if (mu_mv_mag_sq >= 57) {
                                                            if (tcor < 92) {
                                                                return 8;
                                                            }
                                                            else if (tcor >= 92) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                    else if (MVDiff >= 8) {
                                                        if (Cs < 1600) {
                                                            return 8;
                                                        }
                                                        else if (Cs >= 1600) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (Cs >= 2132) {
                                            if (negBalance < 51) {
                                                return 8;
                                            }
                                            else if (negBalance >= 51) {
                                                if (TSC < 8) {
                                                    return 8;
                                                }
                                                else if (TSC >= 8) {
                                                    return 4;
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (ssDCval >= 119) {
                                    if (ssDCval < 131) {
                                        if (TSC < 8) {
                                            if (mcTcor < 88) {
                                                if (RsCsDiff < 87) {
                                                    return 1;
                                                }
                                                else if (RsCsDiff >= 87) {
                                                    return 4;
                                                }
                                            }
                                            else if (mcTcor >= 88) {
                                                return 4;
                                            }
                                        }
                                        else if (TSC >= 8) {
                                            if (tcor < 87) {
                                                if (negBalance < 52) {
                                                    if (refDCval < 123) {
                                                        return 4;
                                                    }
                                                    else if (refDCval >= 123) {
                                                        if (mcTcor < 86) {
                                                            return 4;
                                                        }
                                                        else if (mcTcor >= 86) {
                                                            if (Rs < 2434) {
                                                                return 4;
                                                            }
                                                            else if (Rs >= 2434) {
                                                                if (mu_mv_mag_sq < 66) {
                                                                    return 8;
                                                                }
                                                                else if (mu_mv_mag_sq >= 66) {
                                                                    if (gchDC < 1) {
                                                                        return 4;
                                                                    }
                                                                    else if (gchDC >= 1) {
                                                                        return 8;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (negBalance >= 52) {
                                                    if (diffAFD < 1) {
                                                        return 4;
                                                    }
                                                    else if (diffAFD >= 1) {
                                                        if (MVDiff < 4) {
                                                            if (Rs < 2700) {
                                                                return 4;
                                                            }
                                                            else if (Rs >= 2700) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (MVDiff >= 4) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (tcor >= 87) {
                                                return 8;
                                            }
                                        }
                                    }
                                    else if (ssDCval >= 131) {
                                        if (MVDiff < 6) {
                                            if (Cs < 1424) {
                                                return 2;
                                            }
                                            else if (Cs >= 1424) {
                                                return 4;
                                            }
                                        }
                                        else if (MVDiff >= 6) {
                                            if (MVDiff < 10) {
                                                return 1;
                                            }
                                            else if (MVDiff >= 10) {
                                                return 8;
                                            }
                                        }
                                    }
                                }
                            }
                        }
                    }
                    else if (AFD >= 12) {
                        if (Cs < 1451) {
                            if (RsCsDiff < 87) {
                                if (Scindex < 4) {
                                    if (SC < 845) {
                                        if (AFD < 13) {
                                            if (MVDiff < 22) {
                                                return 2;
                                            }
                                            else if (MVDiff >= 22) {
                                                return 1;
                                            }
                                        }
                                        else if (AFD >= 13) {
                                            if (MVDiff < 38) {
                                                return 2;
                                            }
                                            else if (MVDiff >= 38) {
                                                if (tcor < 74) {
                                                    if (mcTcor < 69) {
                                                        return 1;
                                                    }
                                                    else if (mcTcor >= 69) {
                                                        if (Rs < 222) {
                                                            if (CsDiff < 2) {
                                                                return 2;
                                                            }
                                                            else if (CsDiff >= 2) {
                                                                return 1;
                                                            }
                                                        }
                                                        else if (Rs >= 222) {
                                                            return 2;
                                                        }
                                                    }
                                                }
                                                else if (tcor >= 74) {
                                                    if (mu_mv_mag_sq < 69) {
                                                        if (RsDiff < 2) {
                                                            return 2;
                                                        }
                                                        else if (RsDiff >= 2) {
                                                            return 1;
                                                        }
                                                    }
                                                    else if (mu_mv_mag_sq >= 69) {
                                                        return 2;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (SC >= 845) {
                                        if (MVDiff < 43) {
                                            if (ssDCval < 89) {
                                                if (RsDiff < 4) {
                                                    if (CsDiff < 4) {
                                                        return 2;
                                                    }
                                                    else if (CsDiff >= 4) {
                                                        if (Rs < 541) {
                                                            return 4;
                                                        }
                                                        else if (Rs >= 541) {
                                                            return 1;
                                                        }
                                                    }
                                                }
                                                else if (RsDiff >= 4) {
                                                    return 2;
                                                }
                                            }
                                            else if (ssDCval >= 89) {
                                                if (SC < 1427) {
                                                    if (RsDiff < 3) {
                                                        return 4;
                                                    }
                                                    else if (RsDiff >= 3) {
                                                        return 1;
                                                    }
                                                }
                                                else if (SC >= 1427) {
                                                    return 2;
                                                }
                                            }
                                        }
                                        else if (MVDiff >= 43) {
                                            if (RsCsDiff < 49) {
                                                if (CsDiff < 4) {
                                                    if (SC < 1219) {
                                                        if (Cs < 484) {
                                                            return 2;
                                                        }
                                                        else if (Cs >= 484) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (SC >= 1219) {
                                                        return 8;
                                                    }
                                                }
                                                else if (CsDiff >= 4) {
                                                    if (AFD < 13) {
                                                        return 2;
                                                    }
                                                    else if (AFD >= 13) {
                                                        return 1;
                                                    }
                                                }
                                            }
                                            else if (RsCsDiff >= 49) {
                                                return 1;
                                            }
                                        }
                                    }
                                }
                                else if (Scindex >= 4) {
                                    if (RsDiff < 5) {
                                        if (refDCval < 89) {
                                            if (SC < 2053) {
                                                if (CsDiff < 3) {
                                                    return 4;
                                                }
                                                else if (CsDiff >= 3) {
                                                    return 1;
                                                }
                                            }
                                            else if (SC >= 2053) {
                                                if (Cs < 1221) {
                                                    return 4;
                                                }
                                                else if (Cs >= 1221) {
                                                    return 8;
                                                }
                                            }
                                        }
                                        else if (refDCval >= 89) {
                                            if (TSC < 5) {
                                                return 8;
                                            }
                                            else if (TSC >= 5) {
                                                return 4;
                                            }
                                        }
                                    }
                                    else if (RsDiff >= 5) {
                                        if (tcor < 80) {
                                            if (ssDCval < 125) {
                                                if (diffRsCsdiff < 14) {
                                                    return 8;
                                                }
                                                else if (diffRsCsdiff >= 14) {
                                                    if (Cs < 634) {
                                                        if (CsDiff < 4) {
                                                            return 4;
                                                        }
                                                        else if (CsDiff >= 4) {
                                                            return 2;
                                                        }
                                                    }
                                                    else if (Cs >= 634) {
                                                        if (posBalance < 63) {
                                                            if (tcor < 79) {
                                                                if (MVDiff < 11) {
                                                                    if (Rs < 1332) {
                                                                        return 4;
                                                                    }
                                                                    else if (Rs >= 1332) {
                                                                        return 8;
                                                                    }
                                                                }
                                                                else if (MVDiff >= 11) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (tcor >= 79) {
                                                                return 1;
                                                            }
                                                        }
                                                        else if (posBalance >= 63) {
                                                            return 1;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (ssDCval >= 125) {
                                                return 1;
                                            }
                                        }
                                        else if (tcor >= 80) {
                                            if (SC < 2681) {
                                                if (Cs < 782) {
                                                    if (TSC < 11) {
                                                        return 2;
                                                    }
                                                    else if (TSC >= 11) {
                                                        return 4;
                                                    }
                                                }
                                                else if (Cs >= 782) {
                                                    if (diffAFD < -1) {
                                                        return 1;
                                                    }
                                                    else if (diffAFD >= -1) {
                                                        if (TSCindex < 8) {
                                                            return 2;
                                                        }
                                                        else if (TSCindex >= 8) {
                                                            if (RsDiff < 7) {
                                                                return 4;
                                                            }
                                                            else if (RsDiff >= 7) {
                                                                return 2;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                            else if (SC >= 2681) {
                                                if (ssDCval < 112) {
                                                    if (Rs < 2012) {
                                                        if (TSC < 9) {
                                                            return 1;
                                                        }
                                                        else if (TSC >= 9) {
                                                            return 2;
                                                        }
                                                    }
                                                    else if (Rs >= 2012) {
                                                        return 2;
                                                    }
                                                }
                                                else if (ssDCval >= 112) {
                                                    return 8;
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                            else if (RsCsDiff >= 87) {
                                if (mcTcor < 85) {
                                    if (SC < 3045) {
                                        if (Cs < 878) {
                                            if (mu_mv_mag_sq < 9) {
                                                if (SC < 2973) {
                                                    return 2;
                                                }
                                                else if (SC >= 2973) {
                                                    return 1;
                                                }
                                            }
                                            else if (mu_mv_mag_sq >= 9) {
                                                if (AFD < 24) {
                                                    return 4;
                                                }
                                                else if (AFD >= 24) {
                                                    if (AFD < 37) {
                                                        if (posBalance < 62) {
                                                            return 2;
                                                        }
                                                        else if (posBalance >= 62) {
                                                            return 1;
                                                        }
                                                    }
                                                    else if (AFD >= 37) {
                                                        if (AFD < 47) {
                                                            if (CsDiff < 15) {
                                                                return 4;
                                                            }
                                                            else if (CsDiff >= 15) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (AFD >= 47) {
                                                            if (Rs < 446) {
                                                                return 4;
                                                            }
                                                            else if (Rs >= 446) {
                                                                return 1;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (Cs >= 878) {
                                            if (diffAFD < 44) {
                                                if (ssDCval < 58) {
                                                    if (Cs < 1032) {
                                                        return 1;
                                                    }
                                                    else if (Cs >= 1032) {
                                                        return 2;
                                                    }
                                                }
                                                else if (ssDCval >= 58) {
                                                    if (MVDiff < 367) {
                                                        if (mu_mv_mag_sq < 10) {
                                                            if (tcor < 71) {
                                                                return 8;
                                                            }
                                                            else if (tcor >= 71) {
                                                                if (Rs < 1704) {
                                                                    return 1;
                                                                }
                                                                else if (Rs >= 1704) {
                                                                    return 2;
                                                                }
                                                            }
                                                        }
                                                        else if (mu_mv_mag_sq >= 10) {
                                                            if (Rs < 701) {
                                                                return 8;
                                                            }
                                                            else if (Rs >= 701) {
                                                                return 1;
                                                            }
                                                        }
                                                    }
                                                    else if (MVDiff >= 367) {
                                                        if (Rs < 1674) {
                                                            return 1;
                                                        }
                                                        else if (Rs >= 1674) {
                                                            return 2;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (diffAFD >= 44) {
                                                if (MVDiff < 441) {
                                                    return 4;
                                                }
                                                else if (MVDiff >= 441) {
                                                    return 2;
                                                }
                                            }
                                        }
                                    }
                                    else if (SC >= 3045) {
                                        if (AFD < 29) {
                                            if (tcor < 70) {
                                                if (TSC < 15) {
                                                    return 1;
                                                }
                                                else if (TSC >= 15) {
                                                    if (diffRsCsdiff < -54) {
                                                        return 2;
                                                    }
                                                    else if (diffRsCsdiff >= -54) {
                                                        return 1;
                                                    }
                                                }
                                            }
                                            else if (tcor >= 70) {
                                                if (SC < 3140) {
                                                    return 1;
                                                }
                                                else if (SC >= 3140) {
                                                    return 2;
                                                }
                                            }
                                        }
                                        else if (AFD >= 29) {
                                            if (MVDiff < 157) {
                                                return 2;
                                            }
                                            else if (MVDiff >= 157) {
                                                return 1;
                                            }
                                        }
                                    }
                                }
                                else if (mcTcor >= 85) {
                                    if (Rs < 1211) {
                                        if (mcTcor < 91) {
                                            return 1;
                                        }
                                        else if (mcTcor >= 91) {
                                            if (ssDCval < 151) {
                                                if (Rs < 467) {
                                                    return 8;
                                                }
                                                else if (Rs >= 467) {
                                                    if (mu_mv_mag_sq < 3) {
                                                        return 1;
                                                    }
                                                    else if (mu_mv_mag_sq >= 3) {
                                                        return 2;
                                                    }
                                                }
                                            }
                                            else if (ssDCval >= 151) {
                                                return 1;
                                            }
                                        }
                                    }
                                    else if (Rs >= 1211) {
                                        if (negBalance < 50) {
                                            if (tcor < 86) {
                                                if (Cs < 1231) {
                                                    return 2;
                                                }
                                                else if (Cs >= 1231) {
                                                    if (MVDiff < 12) {
                                                        return 1;
                                                    }
                                                    else if (MVDiff >= 12) {
                                                        if (RsCsDiff < 513) {
                                                            if (RsCsDiff < 183) {
                                                                if (SC < 3363) {
                                                                    return 4;
                                                                }
                                                                else if (SC >= 3363) {
                                                                    return 2;
                                                                }
                                                            }
                                                            else if (RsCsDiff >= 183) {
                                                                return 2;
                                                            }
                                                        }
                                                        else if (RsCsDiff >= 513) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (tcor >= 86) {
                                                if (RsCsDiff < 205) {
                                                    return 4;
                                                }
                                                else if (RsCsDiff >= 205) {
                                                    return 8;
                                                }
                                            }
                                        }
                                        else if (negBalance >= 50) {
                                            if (tcor < 68) {
                                                if (AFD < 28) {
                                                    if (SC < 2830) {
                                                        return 2;
                                                    }
                                                    else if (SC >= 2830) {
                                                        return 1;
                                                    }
                                                }
                                                else if (AFD >= 28) {
                                                    return 2;
                                                }
                                            }
                                            else if (tcor >= 68) {
                                                if (posBalance < 28) {
                                                    return 4;
                                                }
                                                else if (posBalance >= 28) {
                                                    if (CsDiff < 3) {
                                                        if (Scindex < 5) {
                                                            return 4;
                                                        }
                                                        else if (Scindex >= 5) {
                                                            return 2;
                                                        }
                                                    }
                                                    else if (CsDiff >= 3) {
                                                        if (RsCsDiff < 103) {
                                                            if (mcTcor < 90) {
                                                                return 2;
                                                            }
                                                            else if (mcTcor >= 90) {
                                                                if (AFD < 14) {
                                                                    return 2;
                                                                }
                                                                else if (AFD >= 14) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                        else if (RsCsDiff >= 103) {
                                                            return 2;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                        }
                        else if (Cs >= 1451) {
                            if (ssDCval < 82) {
                                if (TSC < 11) {
                                    if (negBalance < 43) {
                                        if (RsCsDiff < 147) {
                                            if (Scindex < 5) {
                                                if (Rs < 1002) {
                                                    return 2;
                                                }
                                                else if (Rs >= 1002) {
                                                    return 1;
                                                }
                                            }
                                            else if (Scindex >= 5) {
                                                if (Rs < 2478) {
                                                    return 8;
                                                }
                                                else if (Rs >= 2478) {
                                                    if (Rs < 2832) {
                                                        return 1;
                                                    }
                                                    else if (Rs >= 2832) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                        }
                                        else if (RsCsDiff >= 147) {
                                            if (ssDCval < 57) {
                                                return 8;
                                            }
                                            else if (ssDCval >= 57) {
                                                return 1;
                                            }
                                        }
                                    }
                                    else if (negBalance >= 43) {
                                        if (mu_mv_mag_sq < 10) {
                                            if (posBalance < 55) {
                                                if (SC < 4230) {
                                                    return 1;
                                                }
                                                else if (SC >= 4230) {
                                                    if (mcTcor < 92) {
                                                        if (RsDiff < 13) {
                                                            return 8;
                                                        }
                                                        else if (RsDiff >= 13) {
                                                            return 2;
                                                        }
                                                    }
                                                    else if (mcTcor >= 92) {
                                                        if (Cs < 2707) {
                                                            return 4;
                                                        }
                                                        else if (Cs >= 2707) {
                                                            return 1;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (posBalance >= 55) {
                                                if (tcor < 68) {
                                                    return 4;
                                                }
                                                else if (tcor >= 68) {
                                                    if (posBalance < 58) {
                                                        return 4;
                                                    }
                                                    else if (posBalance >= 58) {
                                                        return 2;
                                                    }
                                                }
                                            }
                                        }
                                        else if (mu_mv_mag_sq >= 10) {
                                            if (mcTcor < 86) {
                                                return 8;
                                            }
                                            else if (mcTcor >= 86) {
                                                if (negBalance < 53) {
                                                    if (mcTcor < 94) {
                                                        if (mu_mv_mag_sq < 39) {
                                                            return 1;
                                                        }
                                                        else if (mu_mv_mag_sq >= 39) {
                                                            if (mcTcor < 92) {
                                                                return 8;
                                                            }
                                                            else if (mcTcor >= 92) {
                                                                if (SC < 4044) {
                                                                    return 1;
                                                                }
                                                                else if (SC >= 4044) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (mcTcor >= 94) {
                                                        if (TSC < 9) {
                                                            return 1;
                                                        }
                                                        else if (TSC >= 9) {
                                                            return 2;
                                                        }
                                                    }
                                                }
                                                else if (negBalance >= 53) {
                                                    return 1;
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (TSC >= 11) {
                                    if (refDCval < 70) {
                                        return 8;
                                    }
                                    else if (refDCval >= 70) {
                                        if (mcTcor < 71) {
                                            return 4;
                                        }
                                        else if (mcTcor >= 71) {
                                            if (CsDiff < 11) {
                                                if (SC < 3594) {
                                                    if (CsDiff < 6) {
                                                        return 8;
                                                    }
                                                    else if (CsDiff >= 6) {
                                                        return 4;
                                                    }
                                                }
                                                else if (SC >= 3594) {
                                                    if (refDCval < 81) {
                                                        if (gchDC < 1) {
                                                            if (negBalance < 55) {
                                                                return 8;
                                                            }
                                                            else if (negBalance >= 55) {
                                                                if (Rs < 2493) {
                                                                    return 4;
                                                                }
                                                                else if (Rs >= 2493) {
                                                                    return 1;
                                                                }
                                                            }
                                                        }
                                                        else if (gchDC >= 1) {
                                                            if (Cs < 2867) {
                                                                if (RsDiff < 9) {
                                                                    if (Rs < 2124) {
                                                                        return 8;
                                                                    }
                                                                    else if (Rs >= 2124) {
                                                                        if (AFD < 15) {
                                                                            return 1;
                                                                        }
                                                                        else if (AFD >= 15) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                }
                                                                else if (RsDiff >= 9) {
                                                                    if (diffRsCsdiff < 11) {
                                                                        if (diffRsCsdiff < -10) {
                                                                            return 8;
                                                                        }
                                                                        else if (diffRsCsdiff >= -10) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                    else if (diffRsCsdiff >= 11) {
                                                                        if (Cs < 2473) {
                                                                            return 8;
                                                                        }
                                                                        else if (Cs >= 2473) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                            else if (Cs >= 2867) {
                                                                return 1;
                                                            }
                                                        }
                                                    }
                                                    else if (refDCval >= 81) {
                                                        if (CsDiff < 8) {
                                                            return 1;
                                                        }
                                                        else if (CsDiff >= 8) {
                                                            if (Cs < 2017) {
                                                                return 1;
                                                            }
                                                            else if (Cs >= 2017) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                            else if (CsDiff >= 11) {
                                                if (MVDiff < 37) {
                                                    if (mcTcor < 86) {
                                                        if (TSC < 12) {
                                                            return 4;
                                                        }
                                                        else if (TSC >= 12) {
                                                            if (Cs < 2339) {
                                                                return 2;
                                                            }
                                                            else if (Cs >= 2339) {
                                                                return 1;
                                                            }
                                                        }
                                                    }
                                                    else if (mcTcor >= 86) {
                                                        if (Rs < 2563) {
                                                            if (Rs < 2528) {
                                                                return 1;
                                                            }
                                                            else if (Rs >= 2528) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (Rs >= 2563) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                                else if (MVDiff >= 37) {
                                                    if (ssDCval < 78) {
                                                        if (ssDCval < 73) {
                                                            return 1;
                                                        }
                                                        else if (ssDCval >= 73) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (ssDCval >= 78) {
                                                        return 1;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                            else if (ssDCval >= 82) {
                                if (SC < 5609) {
                                    if (ssDCval < 139) {
                                        if (posBalance < 55) {
                                            if (CsDiff < 10) {
                                                if (MVDiff < 25) {
                                                    if (negBalance < 48) {
                                                        if (mcTcor < 94) {
                                                            if (MVDiff < 19) {
                                                                if (Cs < 2577) {
                                                                    if (posBalance < 45) {
                                                                        return 4;
                                                                    }
                                                                    else if (posBalance >= 45) {
                                                                        if (diffAFD < 6) {
                                                                            if (tcor < 90) {
                                                                                return 1;
                                                                            }
                                                                            else if (tcor >= 90) {
                                                                                return 8;
                                                                            }
                                                                        }
                                                                        else if (diffAFD >= 6) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                }
                                                                else if (Cs >= 2577) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (MVDiff >= 19) {
                                                                if (Rs < 2498) {
                                                                    if (RsDiff < 10) {
                                                                        if (mcTcor < 92) {
                                                                            if (MVDiff < 24) {
                                                                                return 4;
                                                                            }
                                                                            else if (MVDiff >= 24) {
                                                                                return 8;
                                                                            }
                                                                        }
                                                                        else if (mcTcor >= 92) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                    else if (RsDiff >= 10) {
                                                                        return 1;
                                                                    }
                                                                }
                                                                else if (Rs >= 2498) {
                                                                    if (ssDCval < 85) {
                                                                        return 8;
                                                                    }
                                                                    else if (ssDCval >= 85) {
                                                                        return 1;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (mcTcor >= 94) {
                                                            if (mu_mv_mag_sq < 46) {
                                                                if (TSC < 10) {
                                                                    return 8;
                                                                }
                                                                else if (TSC >= 10) {
                                                                    return 1;
                                                                }
                                                            }
                                                            else if (mu_mv_mag_sq >= 46) {
                                                                if (SC < 4745) {
                                                                    return 1;
                                                                }
                                                                else if (SC >= 4745) {
                                                                    return 2;
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (negBalance >= 48) {
                                                        if (ssDCval < 117) {
                                                            if (RsCsDiff < 92) {
                                                                return 1;
                                                            }
                                                            else if (RsCsDiff >= 92) {
                                                                if (tcor < 85) {
                                                                    return 2;
                                                                }
                                                                else if (tcor >= 85) {
                                                                    if (Rs < 2242) {
                                                                        if (Rs < 1909) {
                                                                            return 4;
                                                                        }
                                                                        else if (Rs >= 1909) {
                                                                            return 2;
                                                                        }
                                                                    }
                                                                    else if (Rs >= 2242) {
                                                                        return 4;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (ssDCval >= 117) {
                                                            if (negBalance < 60) {
                                                                return 1;
                                                            }
                                                            else if (negBalance >= 60) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (MVDiff >= 25) {
                                                    if (mcTcor < 93) {
                                                        if (refDCval < 87) {
                                                            return 8;
                                                        }
                                                        else if (refDCval >= 87) {
                                                            if (tcor < 89) {
                                                                return 4;
                                                            }
                                                            else if (tcor >= 89) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                    else if (mcTcor >= 93) {
                                                        if (gchDC < 1) {
                                                            return 4;
                                                        }
                                                        else if (gchDC >= 1) {
                                                            return 2;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (CsDiff >= 10) {
                                                if (refDCval < 94) {
                                                    if (mu_mv_mag_sq < 28) {
                                                        if (MVDiff < 8) {
                                                            return 4;
                                                        }
                                                        else if (MVDiff >= 8) {
                                                            if (diffAFD < 2) {
                                                                return 1;
                                                            }
                                                            else if (diffAFD >= 2) {
                                                                if (posBalance < 53) {
                                                                    return 8;
                                                                }
                                                                else if (posBalance >= 53) {
                                                                    return 1;
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (mu_mv_mag_sq >= 28) {
                                                        if (ssDCval < 89) {
                                                            return 8;
                                                        }
                                                        else if (ssDCval >= 89) {
                                                            return 1;
                                                        }
                                                    }
                                                }
                                                else if (refDCval >= 94) {
                                                    if (TSC < 11) {
                                                        if (mcTcor < 89) {
                                                            if (ssDCval < 117) {
                                                                if (ssDCval < 98) {
                                                                    return 8;
                                                                }
                                                                else if (ssDCval >= 98) {
                                                                    return 2;
                                                                }
                                                            }
                                                            else if (ssDCval >= 117) {
                                                                return 1;
                                                            }
                                                        }
                                                        else if (mcTcor >= 89) {
                                                            if (diffAFD < 1) {
                                                                return 4;
                                                            }
                                                            else if (diffAFD >= 1) {
                                                                if (Rs < 1698) {
                                                                    return 4;
                                                                }
                                                                else if (Rs >= 1698) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (TSC >= 11) {
                                                        if (RsDiff < 6) {
                                                            return 2;
                                                        }
                                                        else if (RsDiff >= 6) {
                                                            if (SC < 5532) {
                                                                if (RsDiff < 14) {
                                                                    if (MVDiff < 5) {
                                                                        return 8;
                                                                    }
                                                                    else if (MVDiff >= 5) {
                                                                        return 1;
                                                                    }
                                                                }
                                                                else if (RsDiff >= 14) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (SC >= 5532) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (posBalance >= 55) {
                                            if (mu_mv_mag_sq < 2) {
                                                if (mcTcor < 87) {
                                                    if (negBalance < 51) {
                                                        return 1;
                                                    }
                                                    else if (negBalance >= 51) {
                                                        if (tcor < 65) {
                                                            if (TSC < 15) {
                                                                return 1;
                                                            }
                                                            else if (TSC >= 15) {
                                                                if (Cs < 2276) {
                                                                    return 2;
                                                                }
                                                                else if (Cs >= 2276) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                        else if (tcor >= 65) {
                                                            if (RsDiff < 12) {
                                                                if (TSC < 12) {
                                                                    return 4;
                                                                }
                                                                else if (TSC >= 12) {
                                                                    return 1;
                                                                }
                                                            }
                                                            else if (RsDiff >= 12) {
                                                                if (AFD < 18) {
                                                                    return 2;
                                                                }
                                                                else if (AFD >= 18) {
                                                                    if (RsCsDiff < 329) {
                                                                        return 4;
                                                                    }
                                                                    else if (RsCsDiff >= 329) {
                                                                        if (Cs < 2072) {
                                                                            return 2;
                                                                        }
                                                                        else if (Cs >= 2072) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (mcTcor >= 87) {
                                                    if (Cs < 2422) {
                                                        if (tcor < 90) {
                                                            if (AFD < 18) {
                                                                if (mcTcor < 94) {
                                                                    if (Cs < 2031) {
                                                                        if (CsDiff < 9) {
                                                                            return 4;
                                                                        }
                                                                        else if (CsDiff >= 9) {
                                                                            return 2;
                                                                        }
                                                                    }
                                                                    else if (Cs >= 2031) {
                                                                        if (ssDCval < 87) {
                                                                            if (Rs < 2320) {
                                                                                return 2;
                                                                            }
                                                                            else if (Rs >= 2320) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                        else if (ssDCval >= 87) {
                                                                            if (posBalance < 56) {
                                                                                return 4;
                                                                            }
                                                                            else if (posBalance >= 56) {
                                                                                return 2;
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                                else if (mcTcor >= 94) {
                                                                    return 2;
                                                                }
                                                            }
                                                            else if (AFD >= 18) {
                                                                if (SC < 4557) {
                                                                    if (diffMVdiffVal < -4) {
                                                                        if (Rs < 2135) {
                                                                            return 2;
                                                                        }
                                                                        else if (Rs >= 2135) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                    else if (diffMVdiffVal >= -4) {
                                                                        return 2;
                                                                    }
                                                                }
                                                                else if (SC >= 4557) {
                                                                    if (RsDiff < 14) {
                                                                        if (AFD < 19) {
                                                                            if (Cs < 2273) {
                                                                                return 2;
                                                                            }
                                                                            else if (Cs >= 2273) {
                                                                                return 1;
                                                                            }
                                                                        }
                                                                        else if (AFD >= 19) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                    else if (RsDiff >= 14) {
                                                                        return 2;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (tcor >= 90) {
                                                            if (ssDCval < 92) {
                                                                if (Cs < 2214) {
                                                                    if (SC < 3914) {
                                                                        return 2;
                                                                    }
                                                                    else if (SC >= 3914) {
                                                                        return 1;
                                                                    }
                                                                }
                                                                else if (Cs >= 2214) {
                                                                    return 2;
                                                                }
                                                            }
                                                            else if (ssDCval >= 92) {
                                                                if (Cs < 2299) {
                                                                    return 4;
                                                                }
                                                                else if (Cs >= 2299) {
                                                                    return 2;
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (Cs >= 2422) {
                                                        if (ssDCval < 100) {
                                                            if (Rs < 2891) {
                                                                return 2;
                                                            }
                                                            else if (Rs >= 2891) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (ssDCval >= 100) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (mu_mv_mag_sq >= 2) {
                                                if (mcTcor < 90) {
                                                    if (RsDiff < 17) {
                                                        if (MVDiff < 24) {
                                                            if (ssDCval < 120) {
                                                                if (AFD < 13) {
                                                                    return 4;
                                                                }
                                                                else if (AFD >= 13) {
                                                                    if (diffAFD < 2) {
                                                                        if (diffMVdiffVal < 6) {
                                                                            if (RsDiff < 12) {
                                                                                return 1;
                                                                            }
                                                                            else if (RsDiff >= 12) {
                                                                                if (Cs < 2118) {
                                                                                    return 2;
                                                                                }
                                                                                else if (Cs >= 2118) {
                                                                                    return 1;
                                                                                }
                                                                            }
                                                                        }
                                                                        else if (diffMVdiffVal >= 6) {
                                                                            if (Rs < 2693) {
                                                                                return 4;
                                                                            }
                                                                            else if (Rs >= 2693) {
                                                                                return 1;
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (diffAFD >= 2) {
                                                                        return 4;
                                                                    }
                                                                }
                                                            }
                                                            else if (ssDCval >= 120) {
                                                                return 1;
                                                            }
                                                        }
                                                        else if (MVDiff >= 24) {
                                                            if (tcor < 79) {
                                                                if (AFD < 20) {
                                                                    if (MVDiff < 28) {
                                                                        return 4;
                                                                    }
                                                                    else if (MVDiff >= 28) {
                                                                        return 8;
                                                                    }
                                                                }
                                                                else if (AFD >= 20) {
                                                                    if (TSC < 19) {
                                                                        return 1;
                                                                    }
                                                                    else if (TSC >= 19) {
                                                                        if (Cs < 2622) {
                                                                            return 8;
                                                                        }
                                                                        else if (Cs >= 2622) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                            else if (tcor >= 79) {
                                                                if (mu_mv_mag_sq < 22) {
                                                                    if (negBalance < 54) {
                                                                        return 4;
                                                                    }
                                                                    else if (negBalance >= 54) {
                                                                        if (AFD < 18) {
                                                                            return 2;
                                                                        }
                                                                        else if (AFD >= 18) {
                                                                            if (refDCval < 126) {
                                                                                return 4;
                                                                            }
                                                                            else if (refDCval >= 126) {
                                                                                return 2;
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                                else if (mu_mv_mag_sq >= 22) {
                                                                    if (Rs < 2649) {
                                                                        return 2;
                                                                    }
                                                                    else if (Rs >= 2649) {
                                                                        return 1;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (RsDiff >= 17) {
                                                        if (MVDiff < 74) {
                                                            if (ssDCval < 113) {
                                                                if (SC < 5432) {
                                                                    return 2;
                                                                }
                                                                else if (SC >= 5432) {
                                                                    return 1;
                                                                }
                                                            }
                                                            else if (ssDCval >= 113) {
                                                                return 1;
                                                            }
                                                        }
                                                        else if (MVDiff >= 74) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                                else if (mcTcor >= 90) {
                                                    if (SC < 5105) {
                                                        if (AFD < 14) {
                                                            if (ssDCval < 98) {
                                                                if (ssDCval < 90) {
                                                                    if (MVDiff < 16) {
                                                                        if (Rs < 2391) {
                                                                            return 2;
                                                                        }
                                                                        else if (Rs >= 2391) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                    else if (MVDiff >= 16) {
                                                                        return 2;
                                                                    }
                                                                }
                                                                else if (ssDCval >= 90) {
                                                                    if (Cs < 2266) {
                                                                        return 1;
                                                                    }
                                                                    else if (Cs >= 2266) {
                                                                        return 4;
                                                                    }
                                                                }
                                                            }
                                                            else if (ssDCval >= 98) {
                                                                if (negBalance < 38) {
                                                                    return 4;
                                                                }
                                                                else if (negBalance >= 38) {
                                                                    if (Cs < 2289) {
                                                                        return 8;
                                                                    }
                                                                    else if (Cs >= 2289) {
                                                                        return 2;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (AFD >= 14) {
                                                            if (mcTcor < 92) {
                                                                if (MVDiff < 33) {
                                                                    if (negBalance < 49) {
                                                                        if (Cs < 1798) {
                                                                            return 8;
                                                                        }
                                                                        else if (Cs >= 1798) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                    else if (negBalance >= 49) {
                                                                        if (CsDiff < 8) {
                                                                            if (Rs < 2563) {
                                                                                if (Rs < 2297) {
                                                                                    return 4;
                                                                                }
                                                                                else if (Rs >= 2297) {
                                                                                    return 2;
                                                                                }
                                                                            }
                                                                            else if (Rs >= 2563) {
                                                                                return 8;
                                                                            }
                                                                        }
                                                                        else if (CsDiff >= 8) {
                                                                            if (RsCsDiff < 243) {
                                                                                if (SC < 4765) {
                                                                                    return 1;
                                                                                }
                                                                                else if (SC >= 4765) {
                                                                                    return 2;
                                                                                }
                                                                            }
                                                                            else if (RsCsDiff >= 243) {
                                                                                return 2;
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                                else if (MVDiff >= 33) {
                                                                    if (TSC < 13) {
                                                                        if (Rs < 1824) {
                                                                            return 8;
                                                                        }
                                                                        else if (Rs >= 1824) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                    else if (TSC >= 13) {
                                                                        return 8;
                                                                    }
                                                                }
                                                            }
                                                            else if (mcTcor >= 92) {
                                                                if (CsDiff < 8) {
                                                                    if (TSC < 12) {
                                                                        if (tcor < 89) {
                                                                            return 2;
                                                                        }
                                                                        else if (tcor >= 89) {
                                                                            if (Cs < 2066) {
                                                                                return 2;
                                                                            }
                                                                            else if (Cs >= 2066) {
                                                                                return 1;
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (TSC >= 12) {
                                                                        return 8;
                                                                    }
                                                                }
                                                                else if (CsDiff >= 8) {
                                                                    if (negBalance < 44) {
                                                                        if (posBalance < 57) {
                                                                            if (TSC < 11) {
                                                                                return 8;
                                                                            }
                                                                            else if (TSC >= 11) {
                                                                                return 1;
                                                                            }
                                                                        }
                                                                        else if (posBalance >= 57) {
                                                                            if (Rs < 1227) {
                                                                                return 2;
                                                                            }
                                                                            else if (Rs >= 1227) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (negBalance >= 44) {
                                                                        if (SC < 4369) {
                                                                            if (diffAFD < -4) {
                                                                                if (Rs < 1734) {
                                                                                    return 2;
                                                                                }
                                                                                else if (Rs >= 1734) {
                                                                                    return 8;
                                                                                }
                                                                            }
                                                                            else if (diffAFD >= -4) {
                                                                                return 2;
                                                                            }
                                                                        }
                                                                        else if (SC >= 4369) {
                                                                            if (SC < 4876) {
                                                                                if (negBalance < 54) {
                                                                                    if (MVDiff < 11) {
                                                                                        return 2;
                                                                                    }
                                                                                    else if (MVDiff >= 11) {
                                                                                        return 1;
                                                                                    }
                                                                                }
                                                                                else if (negBalance >= 54) {
                                                                                    if (refDCval < 107) {
                                                                                        return 2;
                                                                                    }
                                                                                    else if (refDCval >= 107) {
                                                                                        return 1;
                                                                                    }
                                                                                }
                                                                            }
                                                                            else if (SC >= 4876) {
                                                                                return 2;
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (SC >= 5105) {
                                                        if (MVDiff < 9) {
                                                            if (diffMVdiffVal < -3) {
                                                                return 1;
                                                            }
                                                            else if (diffMVdiffVal >= -3) {
                                                                if (mu_mv_mag_sq < 72) {
                                                                    if (Cs < 2368) {
                                                                        if (Cs < 2350) {
                                                                            return 4;
                                                                        }
                                                                        else if (Cs >= 2350) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                    else if (Cs >= 2368) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (mu_mv_mag_sq >= 72) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                        else if (MVDiff >= 9) {
                                                            if (ssDCval < 103) {
                                                                if (Cs < 2587) {
                                                                    if (Rs < 2962) {
                                                                        if (Cs < 2380) {
                                                                            if (diffTSC < 1) {
                                                                                if (diffMVdiffVal < 3) {
                                                                                    return 1;
                                                                                }
                                                                                else if (diffMVdiffVal >= 3) {
                                                                                    return 4;
                                                                                }
                                                                            }
                                                                            else if (diffTSC >= 1) {
                                                                                return 1;
                                                                            }
                                                                        }
                                                                        else if (Cs >= 2380) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                    else if (Rs >= 2962) {
                                                                        if (negBalance < 52) {
                                                                            return 1;
                                                                        }
                                                                        else if (negBalance >= 52) {
                                                                            if (Cs < 2323) {
                                                                                return 1;
                                                                            }
                                                                            else if (Cs >= 2323) {
                                                                                return 2;
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                                else if (Cs >= 2587) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (ssDCval >= 103) {
                                                                if (SC < 5538) {
                                                                    if (SC < 5501) {
                                                                        if (Rs < 2713) {
                                                                            return 4;
                                                                        }
                                                                        else if (Rs >= 2713) {
                                                                            if (tcor < 84) {
                                                                                return 2;
                                                                            }
                                                                            else if (tcor >= 84) {
                                                                                return 8;
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (SC >= 5501) {
                                                                        return 8;
                                                                    }
                                                                }
                                                                else if (SC >= 5538) {
                                                                    return 1;
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (ssDCval >= 139) {
                                        if (mcTcor < 93) {
                                            if (negBalance < 56) {
                                                if (diffRsCsdiff < 28) {
                                                    if (MVDiff < 11) {
                                                        return 1;
                                                    }
                                                    else if (MVDiff >= 11) {
                                                        if (Rs < 1416) {
                                                            return 8;
                                                        }
                                                        else if (Rs >= 1416) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                                else if (diffRsCsdiff >= 28) {
                                                    if (SC < 5308) {
                                                        return 8;
                                                    }
                                                    else if (SC >= 5308) {
                                                        if (Rs < 3363) {
                                                            return 1;
                                                        }
                                                        else if (Rs >= 3363) {
                                                            return 2;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (negBalance >= 56) {
                                                return 4;
                                            }
                                        }
                                        else if (mcTcor >= 93) {
                                            if (TSC < 10) {
                                                if (tcor < 84) {
                                                    return 2;
                                                }
                                                else if (tcor >= 84) {
                                                    return 8;
                                                }
                                            }
                                            else if (TSC >= 10) {
                                                if (Cs < 1617) {
                                                    return 4;
                                                }
                                                else if (Cs >= 1617) {
                                                    return 2;
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (SC >= 5609) {
                                    if (mcTcor < 93) {
                                        if (refDCval < 119) {
                                            if (ssDCval < 92) {
                                                if (Cs < 2498) {
                                                    if (Cs < 2479) {
                                                        return 1;
                                                    }
                                                    else if (Cs >= 2479) {
                                                        return 2;
                                                    }
                                                }
                                                else if (Cs >= 2498) {
                                                    if (tcor < 90) {
                                                        if (AFD < 25) {
                                                            return 1;
                                                        }
                                                        else if (AFD >= 25) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (tcor >= 90) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                            else if (ssDCval >= 92) {
                                                if (diffTSC < 1) {
                                                    if (Rs < 3274) {
                                                        if (negBalance < 54) {
                                                            if (Cs < 2715) {
                                                                return 1;
                                                            }
                                                            else if (Cs >= 2715) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (negBalance >= 54) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (Rs >= 3274) {
                                                        if (SC < 5855) {
                                                            return 2;
                                                        }
                                                        else if (SC >= 5855) {
                                                            return 1;
                                                        }
                                                    }
                                                }
                                                else if (diffTSC >= 1) {
                                                    return 8;
                                                }
                                            }
                                        }
                                        else if (refDCval >= 119) {
                                            if (Rs < 3923) {
                                                if (Cs < 2934) {
                                                    if (tcor < 88) {
                                                        return 1;
                                                    }
                                                    else if (tcor >= 88) {
                                                        return 8;
                                                    }
                                                }
                                                else if (Cs >= 2934) {
                                                    return 8;
                                                }
                                            }
                                            else if (Rs >= 3923) {
                                                if (Rs < 4456) {
                                                    if (Cs < 1755) {
                                                        return 1;
                                                    }
                                                    else if (Cs >= 1755) {
                                                        return 4;
                                                    }
                                                }
                                                else if (Rs >= 4456) {
                                                    return 2;
                                                }
                                            }
                                        }
                                    }
                                    else if (mcTcor >= 93) {
                                        if (Cs < 2406) {
                                            if (Cs < 2257) {
                                                if (posBalance < 57) {
                                                    return 4;
                                                }
                                                else if (posBalance >= 57) {
                                                    return 2;
                                                }
                                            }
                                            else if (Cs >= 2257) {
                                                if (MVDiff < 22) {
                                                    return 8;
                                                }
                                                else if (MVDiff >= 22) {
                                                    if (Rs < 3568) {
                                                        return 2;
                                                    }
                                                    else if (Rs >= 3568) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                        }
                                        else if (Cs >= 2406) {
                                            if (refDCval < 103) {
                                                if (Rs < 3449) {
                                                    if (refDCval < 85) {
                                                        return 8;
                                                    }
                                                    else if (refDCval >= 85) {
                                                        return 1;
                                                    }
                                                }
                                                else if (Rs >= 3449) {
                                                    return 2;
                                                }
                                            }
                                            else if (refDCval >= 103) {
                                                if (negBalance < 45) {
                                                    if (Rs < 3239) {
                                                        return 8;
                                                    }
                                                    else if (Rs >= 3239) {
                                                        return 1;
                                                    }
                                                }
                                                else if (negBalance >= 45) {
                                                    return 8;
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                        }
                    }
                }
                else if (Scindex >= 6) {
                    if (CsDiff < 10) {
                        if (TSC < 7) {
                            if (ssDCval < 117) {
                                if (refDCval < 54) {
                                    if (MVDiff < 19) {
                                        if (RsCsDiff < 11) {
                                            if (MVDiff < 7) {
                                                if (Rs < 3612) {
                                                    return 8;
                                                }
                                                else if (Rs >= 3612) {
                                                    return 4;
                                                }
                                            }
                                            else if (MVDiff >= 7) {
                                                return 4;
                                            }
                                        }
                                        else if (RsCsDiff >= 11) {
                                            if (Cs < 3899) {
                                                return 4;
                                            }
                                            else if (Cs >= 3899) {
                                                return 2;
                                            }
                                        }
                                    }
                                    else if (MVDiff >= 19) {
                                        if (tcor < 93) {
                                            if (MVDiff < 20) {
                                                if (tcor < 92) {
                                                    return 4;
                                                }
                                                else if (tcor >= 92) {
                                                    return 8;
                                                }
                                            }
                                            else if (MVDiff >= 20) {
                                                return 8;
                                            }
                                        }
                                        else if (tcor >= 93) {
                                            if (gchDC < 1) {
                                                return 4;
                                            }
                                            else if (gchDC >= 1) {
                                                if (Rs < 3541) {
                                                    return 8;
                                                }
                                                else if (Rs >= 3541) {
                                                    return 4;
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (refDCval >= 54) {
                                    if (mu_mv_mag_sq < 55) {
                                        if (tcor < 93) {
                                            if (Rs < 4523) {
                                                if (posBalance < 48) {
                                                    if (SC < 7455) {
                                                        return 8;
                                                    }
                                                    else if (SC >= 7455) {
                                                        return 4;
                                                    }
                                                }
                                                else if (posBalance >= 48) {
                                                    return 8;
                                                }
                                            }
                                            else if (Rs >= 4523) {
                                                return 4;
                                            }
                                        }
                                        else if (tcor >= 93) {
                                            if (SC < 6751) {
                                                return 2;
                                            }
                                            else if (SC >= 6751) {
                                                return 4;
                                            }
                                        }
                                    }
                                    else if (mu_mv_mag_sq >= 55) {
                                        if (RsDiff < 8) {
                                            return 8;
                                        }
                                        else if (RsDiff >= 8) {
                                            return 2;
                                        }
                                    }
                                }
                            }
                            else if (ssDCval >= 117) {
                                if (Rs < 3767) {
                                    return 8;
                                }
                                else if (Rs >= 3767) {
                                    if (Cs < 2550) {
                                        if (Rs < 4135) {
                                            return 2;
                                        }
                                        else if (Rs >= 4135) {
                                            if (mcTcor < 95) {
                                                if (Rs < 5021) {
                                                    return 4;
                                                }
                                                else if (Rs >= 5021) {
                                                    return 8;
                                                }
                                            }
                                            else if (mcTcor >= 95) {
                                                return 4;
                                            }
                                        }
                                    }
                                    else if (Cs >= 2550) {
                                        return 2;
                                    }
                                }
                            }
                        }
                        else if (TSC >= 7) {
                            if (Rs < 3562) {
                                if (ssDCval < 61) {
                                    if (negBalance < 50) {
                                        if (diffRsCsdiff < 9) {
                                            if (CsDiff < 3) {
                                                return 2;
                                            }
                                            else if (CsDiff >= 3) {
                                                return 8;
                                            }
                                        }
                                        else if (diffRsCsdiff >= 9) {
                                            return 2;
                                        }
                                    }
                                    else if (negBalance >= 50) {
                                        if (tcor < 90) {
                                            if (Rs < 3523) {
                                                if (diffTSC < -1) {
                                                    return 2;
                                                }
                                                else if (diffTSC >= -1) {
                                                    return 8;
                                                }
                                            }
                                            else if (Rs >= 3523) {
                                                if (Rs < 3552) {
                                                    return 4;
                                                }
                                                else if (Rs >= 3552) {
                                                    return 2;
                                                }
                                            }
                                        }
                                        else if (tcor >= 90) {
                                            if (posBalance < 55) {
                                                if (MVDiff < 30) {
                                                    return 2;
                                                }
                                                else if (MVDiff >= 30) {
                                                    return 8;
                                                }
                                            }
                                            else if (posBalance >= 55) {
                                                return 2;
                                            }
                                        }
                                    }
                                }
                                else if (ssDCval >= 61) {
                                    if (diffMVdiffVal < 4) {
                                        if (Cs < 2900) {
                                            if (Rs < 3350) {
                                                return 4;
                                            }
                                            else if (Rs >= 3350) {
                                                return 1;
                                            }
                                        }
                                        else if (Cs >= 2900) {
                                            return 8;
                                        }
                                    }
                                    else if (diffMVdiffVal >= 4) {
                                        return 1;
                                    }
                                }
                            }
                            else if (Rs >= 3562) {
                                if (refDCval < 104) {
                                    if (RsCsDiff < 46) {
                                        if (negBalance < 50) {
                                            if (posBalance < 59) {
                                                if (MVDiff < 24) {
                                                    if (Cs < 4201) {
                                                        return 4;
                                                    }
                                                    else if (Cs >= 4201) {
                                                        return 2;
                                                    }
                                                }
                                                else if (MVDiff >= 24) {
                                                    return 2;
                                                }
                                            }
                                            else if (posBalance >= 59) {
                                                if (mcTcor < 93) {
                                                    return 8;
                                                }
                                                else if (mcTcor >= 93) {
                                                    if (Cs < 3902) {
                                                        return 2;
                                                    }
                                                    else if (Cs >= 3902) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                        }
                                        else if (negBalance >= 50) {
                                            if (TSCindex < 9) {
                                                if (Cs < 3856) {
                                                    if (posBalance < 59) {
                                                        if (posBalance < 57) {
                                                            return 4;
                                                        }
                                                        else if (posBalance >= 57) {
                                                            if (Rs < 3766) {
                                                                if (AFD < 11) {
                                                                    return 2;
                                                                }
                                                                else if (AFD >= 11) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (Rs >= 3766) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                    else if (posBalance >= 59) {
                                                        if (SC < 7317) {
                                                            return 8;
                                                        }
                                                        else if (SC >= 7317) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                                else if (Cs >= 3856) {
                                                    return 2;
                                                }
                                            }
                                            else if (TSCindex >= 9) {
                                                if (Rs < 3565) {
                                                    return 4;
                                                }
                                                else if (Rs >= 3565) {
                                                    return 2;
                                                }
                                            }
                                        }
                                    }
                                    else if (RsCsDiff >= 46) {
                                        if (mcTcor < 92) {
                                            if (diffMVdiffVal < -16) {
                                                return 8;
                                            }
                                            else if (diffMVdiffVal >= -16) {
                                                if (AFD < 15) {
                                                    if (Cs < 4354) {
                                                        return 4;
                                                    }
                                                    else if (Cs >= 4354) {
                                                        if (SC < 9142) {
                                                            return 8;
                                                        }
                                                        else if (SC >= 9142) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                                else if (AFD >= 15) {
                                                    if (RsDiff < 13) {
                                                        if (Rs < 5002) {
                                                            return 8;
                                                        }
                                                        else if (Rs >= 5002) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (RsDiff >= 13) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                        }
                                        else if (mcTcor >= 92) {
                                            if (mu_mv_mag_sq < 39) {
                                                if (refDCval < 58) {
                                                    return 4;
                                                }
                                                else if (refDCval >= 58) {
                                                    return 8;
                                                }
                                            }
                                            else if (mu_mv_mag_sq >= 39) {
                                                if (refDCval < 77) {
                                                    if (MVDiff < 70) {
                                                        if (ssDCval < 75) {
                                                            return 4;
                                                        }
                                                        else if (ssDCval >= 75) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (MVDiff >= 70) {
                                                        return 8;
                                                    }
                                                }
                                                else if (refDCval >= 77) {
                                                    if (TSC < 10) {
                                                        return 8;
                                                    }
                                                    else if (TSC >= 10) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (refDCval >= 104) {
                                    if (Rs < 3910) {
                                        if (SC < 6223) {
                                            if (AFD < 11) {
                                                return 2;
                                            }
                                            else if (AFD >= 11) {
                                                return 1;
                                            }
                                        }
                                        else if (SC >= 6223) {
                                            if (TSC < 12) {
                                                if (RsCsDiff < 205) {
                                                    if (refDCval < 135) {
                                                        if (Cs < 2865) {
                                                            return 8;
                                                        }
                                                        else if (Cs >= 2865) {
                                                            if (Rs < 3801) {
                                                                return 1;
                                                            }
                                                            else if (Rs >= 3801) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                    else if (refDCval >= 135) {
                                                        return 1;
                                                    }
                                                }
                                                else if (RsCsDiff >= 205) {
                                                    if (Rs < 3803) {
                                                        return 1;
                                                    }
                                                    else if (Rs >= 3803) {
                                                        return 2;
                                                    }
                                                }
                                            }
                                            else if (TSC >= 12) {
                                                if (Rs < 3886) {
                                                    return 4;
                                                }
                                                else if (Rs >= 3886) {
                                                    return 8;
                                                }
                                            }
                                        }
                                    }
                                    else if (Rs >= 3910) {
                                        if (ssDCval < 128) {
                                            if (RsCsDiff < 129) {
                                                if (tcor < 82) {
                                                    if (Rs < 4133) {
                                                        if (Rs < 4085) {
                                                            return 8;
                                                        }
                                                        else if (Rs >= 4085) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (Rs >= 4133) {
                                                        if (AFD < 11) {
                                                            return 8;
                                                        }
                                                        else if (AFD >= 11) {
                                                            return 1;
                                                        }
                                                    }
                                                }
                                                else if (tcor >= 82) {
                                                    if (posBalance < 48) {
                                                        if (Cs < 2359) {
                                                            if (MVDiff < 10) {
                                                                return 4;
                                                            }
                                                            else if (MVDiff >= 10) {
                                                                if (Cs < 2177) {
                                                                    return 4;
                                                                }
                                                                else if (Cs >= 2177) {
                                                                    if (Cs < 2301) {
                                                                        return 8;
                                                                    }
                                                                    else if (Cs >= 2301) {
                                                                        return 4;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (Cs >= 2359) {
                                                            if (ssDCval < 125) {
                                                                if (CsDiff < 8) {
                                                                    if (negBalance < 39) {
                                                                        return 8;
                                                                    }
                                                                    else if (negBalance >= 39) {
                                                                        if (TSC < 10) {
                                                                            return 8;
                                                                        }
                                                                        else if (TSC >= 10) {
                                                                            if (Rs < 5145) {
                                                                                return 2;
                                                                            }
                                                                            else if (Rs >= 5145) {
                                                                                return 8;
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                                else if (CsDiff >= 8) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (ssDCval >= 125) {
                                                                if (Rs < 4499) {
                                                                    return 4;
                                                                }
                                                                else if (Rs >= 4499) {
                                                                    return 2;
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (posBalance >= 48) {
                                                        if (posBalance < 49) {
                                                            if (Rs < 4547) {
                                                                return 4;
                                                            }
                                                            else if (Rs >= 4547) {
                                                                return 2;
                                                            }
                                                        }
                                                        else if (posBalance >= 49) {
                                                            if (MVDiff < 6) {
                                                                return 2;
                                                            }
                                                            else if (MVDiff >= 6) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                            else if (RsCsDiff >= 129) {
                                                if (Cs < 3207) {
                                                    if (tcor < 77) {
                                                        if (ssDCval < 107) {
                                                            return 4;
                                                        }
                                                        else if (ssDCval >= 107) {
                                                            if (AFD < 16) {
                                                                return 1;
                                                            }
                                                            else if (AFD >= 16) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                    else if (tcor >= 77) {
                                                        if (negBalance < 34) {
                                                            if (SC < 7584) {
                                                                return 4;
                                                            }
                                                            else if (SC >= 7584) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (negBalance >= 34) {
                                                            if (Rs < 5166) {
                                                                if (CsDiff < 7) {
                                                                    if (RsDiff < 12) {
                                                                        return 4;
                                                                    }
                                                                    else if (RsDiff >= 12) {
                                                                        return 8;
                                                                    }
                                                                }
                                                                else if (CsDiff >= 7) {
                                                                    if (ssDCval < 121) {
                                                                        if (Rs < 4514) {
                                                                            if (tcor < 92) {
                                                                                if (MVDiff < 24) {
                                                                                    return 4;
                                                                                }
                                                                                else if (MVDiff >= 24) {
                                                                                    return 8;
                                                                                }
                                                                            }
                                                                            else if (tcor >= 92) {
                                                                                return 2;
                                                                            }
                                                                        }
                                                                        else if (Rs >= 4514) {
                                                                            if (RsDiff < 10) {
                                                                                if (AFD < 10) {
                                                                                    return 2;
                                                                                }
                                                                                else if (AFD >= 10) {
                                                                                    return 8;
                                                                                }
                                                                            }
                                                                            else if (RsDiff >= 10) {
                                                                                return 2;
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (ssDCval >= 121) {
                                                                        return 4;
                                                                    }
                                                                }
                                                            }
                                                            else if (Rs >= 5166) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (Cs >= 3207) {
                                                    if (Cs < 3514) {
                                                        return 2;
                                                    }
                                                    else if (Cs >= 3514) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                        }
                                        else if (ssDCval >= 128) {
                                            if (Rs < 4646) {
                                                if (mu_mv_mag_sq < 75) {
                                                    if (ssDCval < 156) {
                                                        if (diffAFD < -2) {
                                                            if (TSC < 10) {
                                                                return 2;
                                                            }
                                                            else if (TSC >= 10) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (diffAFD >= -2) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (ssDCval >= 156) {
                                                        return 2;
                                                    }
                                                }
                                                else if (mu_mv_mag_sq >= 75) {
                                                    return 2;
                                                }
                                            }
                                            else if (Rs >= 4646) {
                                                if (TSC < 10) {
                                                    return 2;
                                                }
                                                else if (TSC >= 10) {
                                                    return 1;
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                        }
                    }
                    else if (CsDiff >= 10) {
                        if (Cs < 3532) {
                            if (tcor < 81) {
                                if (ssDCval < 143) {
                                    if (ssDCval < 124) {
                                        if (Rs < 4693) {
                                            if (negBalance < 54) {
                                                if (AFD < 14) {
                                                    return 4;
                                                }
                                                else if (AFD >= 14) {
                                                    if (posBalance < 64) {
                                                        if (Cs < 3470) {
                                                            return 1;
                                                        }
                                                        else if (Cs >= 3470) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (posBalance >= 64) {
                                                        if (gchDC < 4) {
                                                            return 2;
                                                        }
                                                        else if (gchDC >= 4) {
                                                            if (posBalance < 80) {
                                                                return 8;
                                                            }
                                                            else if (posBalance >= 80) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                            else if (negBalance >= 54) {
                                                if (RsCsDiff < 442) {
                                                    if (SC < 6192) {
                                                        return 4;
                                                    }
                                                    else if (SC >= 6192) {
                                                        if (Cs < 2796) {
                                                            return 4;
                                                        }
                                                        else if (Cs >= 2796) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                                else if (RsCsDiff >= 442) {
                                                    if (refDCval < 100) {
                                                        if (AFD < 30) {
                                                            return 4;
                                                        }
                                                        else if (AFD >= 30) {
                                                            return 2;
                                                        }
                                                    }
                                                    else if (refDCval >= 100) {
                                                        if (Cs < 3183) {
                                                            if (Rs < 3714) {
                                                                return 1;
                                                            }
                                                            else if (Rs >= 3714) {
                                                                return 2;
                                                            }
                                                        }
                                                        else if (Cs >= 3183) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (Rs >= 4693) {
                                            if (refDCval < 107) {
                                                if (ssDCval < 105) {
                                                    return 4;
                                                }
                                                else if (ssDCval >= 105) {
                                                    return 2;
                                                }
                                            }
                                            else if (refDCval >= 107) {
                                                if (TSC < 14) {
                                                    return 2;
                                                }
                                                else if (TSC >= 14) {
                                                    return 8;
                                                }
                                            }
                                        }
                                    }
                                    else if (ssDCval >= 124) {
                                        if (posBalance < 53) {
                                            if (tcor < 73) {
                                                if (Rs < 5086) {
                                                    return 1;
                                                }
                                                else if (Rs >= 5086) {
                                                    return 4;
                                                }
                                            }
                                            else if (tcor >= 73) {
                                                return 8;
                                            }
                                        }
                                        else if (posBalance >= 53) {
                                            if (refDCval < 127) {
                                                if (diffAFD < -2) {
                                                    if (Rs < 4111) {
                                                        return 8;
                                                    }
                                                    else if (Rs >= 4111) {
                                                        return 1;
                                                    }
                                                }
                                                else if (diffAFD >= -2) {
                                                    if (Rs < 3979) {
                                                        if (Rs < 3511) {
                                                            return 1;
                                                        }
                                                        else if (Rs >= 3511) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (Rs >= 3979) {
                                                        return 1;
                                                    }
                                                }
                                            }
                                            else if (refDCval >= 127) {
                                                if (gchDC < 2) {
                                                    return 1;
                                                }
                                                else if (gchDC >= 2) {
                                                    if (Rs < 3882) {
                                                        return 4;
                                                    }
                                                    else if (Rs >= 3882) {
                                                        return 1;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (ssDCval >= 143) {
                                    if (RsDiff < 16) {
                                        return 2;
                                    }
                                    else if (RsDiff >= 16) {
                                        if (ssDCval < 153) {
                                            if (Rs < 5113) {
                                                return 8;
                                            }
                                            else if (Rs >= 5113) {
                                                return 2;
                                            }
                                        }
                                        else if (ssDCval >= 153) {
                                            if (AFD < 22) {
                                                if (refDCval < 157) {
                                                    return 4;
                                                }
                                                else if (refDCval >= 157) {
                                                    return 2;
                                                }
                                            }
                                            else if (AFD >= 22) {
                                                return 4;
                                            }
                                        }
                                    }
                                }
                            }
                            else if (tcor >= 81) {
                                if (ssDCval < 125) {
                                    if (Rs < 4181) {
                                        if (ssDCval < 104) {
                                            if (diffRsCsdiff < 107) {
                                                if (MVDiff < 3) {
                                                    if (RsCsDiff < 256) {
                                                        return 8;
                                                    }
                                                    else if (RsCsDiff >= 256) {
                                                        return 4;
                                                    }
                                                }
                                                else if (MVDiff >= 3) {
                                                    if (MVDiff < 13) {
                                                        return 4;
                                                    }
                                                    else if (MVDiff >= 13) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                            else if (diffRsCsdiff >= 107) {
                                                return 1;
                                            }
                                        }
                                        else if (ssDCval >= 104) {
                                            if (ssDCval < 122) {
                                                return 8;
                                            }
                                            else if (ssDCval >= 122) {
                                                return 1;
                                            }
                                        }
                                    }
                                    else if (Rs >= 4181) {
                                        if (mcTcor < 90) {
                                            if (Rs < 4654) {
                                                return 4;
                                            }
                                            else if (Rs >= 4654) {
                                                return 2;
                                            }
                                        }
                                        else if (mcTcor >= 90) {
                                            return 4;
                                        }
                                    }
                                }
                                else if (ssDCval >= 125) {
                                    return 8;
                                }
                            }
                        }
                        else if (Cs >= 3532) {
                            if (Rs < 6622) {
                                if (mu_mv_mag_sq < 3) {
                                    if (ssDCval < 159) {
                                        if (CsDiff < 16) {
                                            if (tcor < 64) {
                                                return 1;
                                            }
                                            else if (tcor >= 64) {
                                                if (tcor < 77) {
                                                    if (AFD < 27) {
                                                        if (diffTSC < 1) {
                                                            return 4;
                                                        }
                                                        else if (diffTSC >= 1) {
                                                            if (RsDiff < 15) {
                                                                return 8;
                                                            }
                                                            else if (RsDiff >= 15) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                    else if (AFD >= 27) {
                                                        return 8;
                                                    }
                                                }
                                                else if (tcor >= 77) {
                                                    return 8;
                                                }
                                            }
                                        }
                                        else if (CsDiff >= 16) {
                                            if (TSC < 15) {
                                                if (Cs < 4241) {
                                                    return 4;
                                                }
                                                else if (Cs >= 4241) {
                                                    return 1;
                                                }
                                            }
                                            else if (TSC >= 15) {
                                                if (SC < 8976) {
                                                    return 1;
                                                }
                                                else if (SC >= 8976) {
                                                    return 8;
                                                }
                                            }
                                        }
                                    }
                                    else if (ssDCval >= 159) {
                                        return 2;
                                    }
                                }
                                else if (mu_mv_mag_sq >= 3) {
                                    if (ssDCval < 136) {
                                        if (CsDiff < 18) {
                                            if (posBalance < 57) {
                                                if (tcor < 78) {
                                                    if (posBalance < 52) {
                                                        return 8;
                                                    }
                                                    else if (posBalance >= 52) {
                                                        return 4;
                                                    }
                                                }
                                                else if (tcor >= 78) {
                                                    if (RsCsDiff < 699) {
                                                        if (Rs < 6379) {
                                                            return 8;
                                                        }
                                                        else if (Rs >= 6379) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (RsCsDiff >= 699) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                            else if (posBalance >= 57) {
                                                if (RsDiff < 23) {
                                                    if (negBalance < 39) {
                                                        return 4;
                                                    }
                                                    else if (negBalance >= 39) {
                                                        if (Cs < 3592) {
                                                            if (ssDCval < 95) {
                                                                return 4;
                                                            }
                                                            else if (ssDCval >= 95) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (Cs >= 3592) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                                else if (RsDiff >= 23) {
                                                    return 1;
                                                }
                                            }
                                        }
                                        else if (CsDiff >= 18) {
                                            if (mcTcor < 93) {
                                                if (tcor < 32) {
                                                    if (Rs < 5279) {
                                                        return 2;
                                                    }
                                                    else if (Rs >= 5279) {
                                                        return 8;
                                                    }
                                                }
                                                else if (tcor >= 32) {
                                                    return 8;
                                                }
                                            }
                                            else if (mcTcor >= 93) {
                                                return 1;
                                            }
                                        }
                                    }
                                    else if (ssDCval >= 136) {
                                        if (Rs < 4487) {
                                            if (posBalance < 60) {
                                                if (AFD < 23) {
                                                    return 1;
                                                }
                                                else if (AFD >= 23) {
                                                    return 8;
                                                }
                                            }
                                            else if (posBalance >= 60) {
                                                return 8;
                                            }
                                        }
                                        else if (Rs >= 4487) {
                                            if (mu_mv_mag_sq < 38) {
                                                if (diffAFD < 3) {
                                                    if (refDCval < 139) {
                                                        return 4;
                                                    }
                                                    else if (refDCval >= 139) {
                                                        if (Rs < 5613) {
                                                            return 8;
                                                        }
                                                        else if (Rs >= 5613) {
                                                            if (Cs < 4548) {
                                                                return 4;
                                                            }
                                                            else if (Cs >= 4548) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (diffAFD >= 3) {
                                                    return 8;
                                                }
                                            }
                                            else if (mu_mv_mag_sq >= 38) {
                                                if (Cs < 4122) {
                                                    if (Cs < 4114) {
                                                        return 8;
                                                    }
                                                    else if (Cs >= 4114) {
                                                        return 4;
                                                    }
                                                }
                                                else if (Cs >= 4122) {
                                                    return 8;
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                            else if (Rs >= 6622) {
                                if (MVDiff < 9) {
                                    if (tcor < 74) {
                                        if (AFD < 38) {
                                            if (ssDCval < 88) {
                                                return 8;
                                            }
                                            else if (ssDCval >= 88) {
                                                if (posBalance < 60) {
                                                    if (Rs < 15038) {
                                                        if (Rs < 7998) {
                                                            if (Cs < 7114) {
                                                                return 8;
                                                            }
                                                            else if (Cs >= 7114) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (Rs >= 7998) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (Rs >= 15038) {
                                                        return 8;
                                                    }
                                                }
                                                else if (posBalance >= 60) {
                                                    return 8;
                                                }
                                            }
                                        }
                                        else if (AFD >= 38) {
                                            return 8;
                                        }
                                    }
                                    else if (tcor >= 74) {
                                        if (Cs < 6047) {
                                            if (MVDiff < 6) {
                                                return 8;
                                            }
                                            else if (MVDiff >= 6) {
                                                return 4;
                                            }
                                        }
                                        else if (Cs >= 6047) {
                                            if (ssDCval < 72) {
                                                return 4;
                                            }
                                            else if (ssDCval >= 72) {
                                                if (CsDiff < 34) {
                                                    return 8;
                                                }
                                                else if (CsDiff >= 34) {
                                                    if (MVDiff < 2) {
                                                        return 4;
                                                    }
                                                    else if (MVDiff >= 2) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (MVDiff >= 9) {
                                    if (mu_mv_mag_sq < 38) {
                                        if (mcTcor < 85) {
                                            if (ssDCval < 137) {
                                                if (refDCval < 76) {
                                                    return 4;
                                                }
                                                else if (refDCval >= 76) {
                                                    if (Rs < 7416) {
                                                        return 8;
                                                    }
                                                    else if (Rs >= 7416) {
                                                        if (Cs < 6970) {
                                                            return 4;
                                                        }
                                                        else if (Cs >= 6970) {
                                                            if (posBalance < 59) {
                                                                return 8;
                                                            }
                                                            else if (posBalance >= 59) {
                                                                if (mcTcor < 80) {
                                                                    if (AFD < 47) {
                                                                        return 4;
                                                                    }
                                                                    else if (AFD >= 47) {
                                                                        return 8;
                                                                    }
                                                                }
                                                                else if (mcTcor >= 80) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                            else if (ssDCval >= 137) {
                                                if (SC < 12701) {
                                                    return 2;
                                                }
                                                else if (SC >= 12701) {
                                                    if (ssDCval < 147) {
                                                        if (ssDCval < 138) {
                                                            return 2;
                                                        }
                                                        else if (ssDCval >= 138) {
                                                            if (negBalance < 50) {
                                                                return 8;
                                                            }
                                                            else if (negBalance >= 50) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                    else if (ssDCval >= 147) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                        }
                                        else if (mcTcor >= 85) {
                                            if (RsDiff < 38) {
                                                if (tcor < 70) {
                                                    return 4;
                                                }
                                                else if (tcor >= 70) {
                                                    if (SC < 13119) {
                                                        if (ssDCval < 152) {
                                                            if (mcTcor < 92) {
                                                                return 4;
                                                            }
                                                            else if (mcTcor >= 92) {
                                                                return 2;
                                                            }
                                                        }
                                                        else if (ssDCval >= 152) {
                                                            return 2;
                                                        }
                                                    }
                                                    else if (SC >= 13119) {
                                                        if (mu_mv_mag_sq < 16) {
                                                            if (mcTcor < 92) {
                                                                if (diffMVdiffVal < 7) {
                                                                    if (tcor < 80) {
                                                                        return 8;
                                                                    }
                                                                    else if (tcor >= 80) {
                                                                        if (MVDiff < 11) {
                                                                            return 4;
                                                                        }
                                                                        else if (MVDiff >= 11) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                }
                                                                else if (diffMVdiffVal >= 7) {
                                                                    return 2;
                                                                }
                                                            }
                                                            else if (mcTcor >= 92) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (mu_mv_mag_sq >= 16) {
                                                            if (diffAFD < -4) {
                                                                return 2;
                                                            }
                                                            else if (diffAFD >= -4) {
                                                                if (SC < 13449) {
                                                                    return 8;
                                                                }
                                                                else if (SC >= 13449) {
                                                                    if (Rs < 10461) {
                                                                        if (gchDC < 3) {
                                                                            if (posBalance < 59) {
                                                                                if (mcTcor < 91) {
                                                                                    return 2;
                                                                                }
                                                                                else if (mcTcor >= 91) {
                                                                                    return 4;
                                                                                }
                                                                            }
                                                                            else if (posBalance >= 59) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                        else if (gchDC >= 3) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                    else if (Rs >= 10461) {
                                                                        return 8;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                            else if (RsDiff >= 38) {
                                                return 2;
                                            }
                                        }
                                    }
                                    else if (mu_mv_mag_sq >= 38) {
                                        if (refDCval < 68) {
                                            return 4;
                                        }
                                        else if (refDCval >= 68) {
                                            if (AFD < 18) {
                                                return 4;
                                            }
                                            else if (AFD >= 18) {
                                                if (mu_mv_mag_sq < 59) {
                                                    if (gchDC < 3) {
                                                        if (ssDCval < 79) {
                                                            return 4;
                                                        }
                                                        else if (ssDCval >= 79) {
                                                            if (SC < 20787) {
                                                                return 8;
                                                            }
                                                            else if (SC >= 20787) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                    else if (gchDC >= 3) {
                                                        return 4;
                                                    }
                                                }
                                                else if (mu_mv_mag_sq >= 59) {
                                                    return 8;
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                        }
                    }
                }
            }
        }
        else if (mu_mv_mag_sq >= 89) {
            if (AFD < 6) {
                if (Rs < 588) {
                    if (ssDCval < 161) {
                        if (refDCval < 95) {
                            if (ssDCval < 90) {
                                if (negBalance < 44) {
                                    if (CsDiff < 1) {
                                        if (Cs < 537) {
                                            if (gchDC < 1) {
                                                if (negBalance < 20) {
                                                    if (posBalance < 40) {
                                                        return 8;
                                                    }
                                                    else if (posBalance >= 40) {
                                                        if (posBalance < 41) {
                                                            return 4;
                                                        }
                                                        else if (posBalance >= 41) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                                else if (negBalance >= 20) {
                                                    if (Rs < 270) {
                                                        return 4;
                                                    }
                                                    else if (Rs >= 270) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                            else if (gchDC >= 1) {
                                                if (AFD < 3) {
                                                    if (MVDiff < 336) {
                                                        if (diffAFD < 1) {
                                                            return 4;
                                                        }
                                                        else if (diffAFD >= 1) {
                                                            return 1;
                                                        }
                                                    }
                                                    else if (MVDiff >= 336) {
                                                        return 1;
                                                    }
                                                }
                                                else if (AFD >= 3) {
                                                    if (SC < 637) {
                                                        if (posBalance < 97) {
                                                            return 4;
                                                        }
                                                        else if (posBalance >= 97) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (SC >= 637) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                        }
                                        else if (Cs >= 537) {
                                            return 4;
                                        }
                                    }
                                    else if (CsDiff >= 1) {
                                        if (mu_mv_mag_sq < 113) {
                                            if (Cs < 377) {
                                                return 4;
                                            }
                                            else if (Cs >= 377) {
                                                return 8;
                                            }
                                        }
                                        else if (mu_mv_mag_sq >= 113) {
                                            return 8;
                                        }
                                    }
                                }
                                else if (negBalance >= 44) {
                                    if (AFD < 5) {
                                        return 4;
                                    }
                                    else if (AFD >= 5) {
                                        return 2;
                                    }
                                }
                            }
                            else if (ssDCval >= 90) {
                                if (SC < 1357) {
                                    if (refDCval < 91) {
                                        return 8;
                                    }
                                    else if (refDCval >= 91) {
                                        return 4;
                                    }
                                }
                                else if (SC >= 1357) {
                                    return 8;
                                }
                            }
                        }
                        else if (refDCval >= 95) {
                            if (AFD < 4) {
                                if (refDCval < 161) {
                                    return 8;
                                }
                                else if (refDCval >= 161) {
                                    return 4;
                                }
                            }
                            else if (AFD >= 4) {
                                if (SC < 916) {
                                    return 8;
                                }
                                else if (SC >= 916) {
                                    if (negBalance < 40) {
                                        return 4;
                                    }
                                    else if (negBalance >= 40) {
                                        return 2;
                                    }
                                }
                            }
                        }
                    }
                    else if (ssDCval >= 161) {
                        if (refDCval < 165) {
                            if (Rs < 314) {
                                return 8;
                            }
                            else if (Rs >= 314) {
                                if (SC < 899) {
                                    return 4;
                                }
                                else if (SC >= 899) {
                                    if (Rs < 468) {
                                        return 2;
                                    }
                                    else if (Rs >= 468) {
                                        if (MVDiff < 8) {
                                            return 4;
                                        }
                                        else if (MVDiff >= 8) {
                                            return 2;
                                        }
                                    }
                                }
                            }
                        }
                        else if (refDCval >= 165) {
                            if (Rs < 443) {
                                return 4;
                            }
                            else if (Rs >= 443) {
                                if (Rs < 532) {
                                    return 2;
                                }
                                else if (Rs >= 532) {
                                    if (SC < 1049) {
                                        return 4;
                                    }
                                    else if (SC >= 1049) {
                                        return 2;
                                    }
                                }
                            }
                        }
                    }
                }
                else if (Rs >= 588) {
                    if (Rs < 3812) {
                        if (SC < 7785) {
                            if (mu_mv_mag_sq < 121) {
                                if (negBalance < 35) {
                                    if (refDCval < 32) {
                                        if (Cs < 708) {
                                            if (Cs < 648) {
                                                return 8;
                                            }
                                            else if (Cs >= 648) {
                                                return 4;
                                            }
                                        }
                                        else if (Cs >= 708) {
                                            if (refDCval < 30) {
                                                if (MVDiff < 22) {
                                                    return 8;
                                                }
                                                else if (MVDiff >= 22) {
                                                    return 4;
                                                }
                                            }
                                            else if (refDCval >= 30) {
                                                if (Cs < 1131) {
                                                    if (Rs < 1456) {
                                                        return 1;
                                                    }
                                                    else if (Rs >= 1456) {
                                                        return 8;
                                                    }
                                                }
                                                else if (Cs >= 1131) {
                                                    if (Rs < 986) {
                                                        return 8;
                                                    }
                                                    else if (Rs >= 986) {
                                                        if (Rs < 1045) {
                                                            return 4;
                                                        }
                                                        else if (Rs >= 1045) {
                                                            if (ssDCval < 32) {
                                                                return 2;
                                                            }
                                                            else if (ssDCval >= 32) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (refDCval >= 32) {
                                        if (negBalance < 16) {
                                            if (refDCval < 56) {
                                                if (Cs < 227) {
                                                    return 4;
                                                }
                                                else if (Cs >= 227) {
                                                    return 8;
                                                }
                                            }
                                            else if (refDCval >= 56) {
                                                if (SC < 4469) {
                                                    if (CsDiff < 2) {
                                                        if (TSC < 3) {
                                                            if (MVDiff < 2) {
                                                                return 8;
                                                            }
                                                            else if (MVDiff >= 2) {
                                                                if (diffMVdiffVal < -3) {
                                                                    return 1;
                                                                }
                                                                else if (diffMVdiffVal >= -3) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                        else if (TSC >= 3) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (CsDiff >= 2) {
                                                        if (negBalance < 5) {
                                                            if (mcTcor < 64) {
                                                                if (Rs < 1145) {
                                                                    return 8;
                                                                }
                                                                else if (Rs >= 1145) {
                                                                    return 1;
                                                                }
                                                            }
                                                            else if (mcTcor >= 64) {
                                                                if (negBalance < 4) {
                                                                    return 8;
                                                                }
                                                                else if (negBalance >= 4) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                        else if (negBalance >= 5) {
                                                            if (CsDiff < 3) {
                                                                if (MVDiff < 8) {
                                                                    if (MVDiff < 5) {
                                                                        if (diffMVdiffVal < -3) {
                                                                            return 1;
                                                                        }
                                                                        else if (diffMVdiffVal >= -3) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                    else if (MVDiff >= 5) {
                                                                        if (diffMVdiffVal < 0) {
                                                                            return 1;
                                                                        }
                                                                        else if (diffMVdiffVal >= 0) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                }
                                                                else if (MVDiff >= 8) {
                                                                    return 1;
                                                                }
                                                            }
                                                            else if (CsDiff >= 3) {
                                                                return 1;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (SC >= 4469) {
                                                    if (refDCval < 117) {
                                                        if (posBalance < 9) {
                                                            if (negBalance < 8) {
                                                                return 4;
                                                            }
                                                            else if (negBalance >= 8) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (posBalance >= 9) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (refDCval >= 117) {
                                                        if (Cs < 2151) {
                                                            return 8;
                                                        }
                                                        else if (Cs >= 2151) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (negBalance >= 16) {
                                            if (Rs < 3126) {
                                                if (ssDCval < 51) {
                                                    if (refDCval < 48) {
                                                        if (AFD < 5) {
                                                            return 8;
                                                        }
                                                        else if (AFD >= 5) {
                                                            if (Rs < 2245) {
                                                                return 4;
                                                            }
                                                            else if (Rs >= 2245) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                    else if (refDCval >= 48) {
                                                        if (Rs < 1576) {
                                                            return 1;
                                                        }
                                                        else if (Rs >= 1576) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                                else if (ssDCval >= 51) {
                                                    if (mcTcor < 97) {
                                                        if (diffRsCsdiff < -35) {
                                                            if (Rs < 1680) {
                                                                return 4;
                                                            }
                                                            else if (Rs >= 1680) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (diffRsCsdiff >= -35) {
                                                            if (posBalance < 18) {
                                                                return 1;
                                                            }
                                                            else if (posBalance >= 18) {
                                                                if (diffAFD < -2) {
                                                                    return 1;
                                                                }
                                                                else if (diffAFD >= -2) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (mcTcor >= 97) {
                                                        if (tcor < 96) {
                                                            if (ssDCval < 84) {
                                                                return 8;
                                                            }
                                                            else if (ssDCval >= 84) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (tcor >= 96) {
                                                            if (refDCval < 144) {
                                                                if (Rs < 1312) {
                                                                    if (ssDCval < 108) {
                                                                        return 8;
                                                                    }
                                                                    else if (ssDCval >= 108) {
                                                                        if (RsDiff < 3) {
                                                                            return 4;
                                                                        }
                                                                        else if (RsDiff >= 3) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                }
                                                                else if (Rs >= 1312) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (refDCval >= 144) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                            else if (Rs >= 3126) {
                                                if (ssDCval < 109) {
                                                    if (Cs < 1505) {
                                                        return 1;
                                                    }
                                                    else if (Cs >= 1505) {
                                                        if (RsDiff < 4) {
                                                            return 8;
                                                        }
                                                        else if (RsDiff >= 4) {
                                                            if (mu_mv_mag_sq < 117) {
                                                                return 8;
                                                            }
                                                            else if (mu_mv_mag_sq >= 117) {
                                                                if (Cs < 3064) {
                                                                    return 1;
                                                                }
                                                                else if (Cs >= 3064) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (ssDCval >= 109) {
                                                    return 1;
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (negBalance >= 35) {
                                    if (ssDCval < 73) {
                                        if (ssDCval < 55) {
                                            if (Rs < 1704) {
                                                if (Rs < 795) {
                                                    return 4;
                                                }
                                                else if (Rs >= 795) {
                                                    if (posBalance < 57) {
                                                        if (refDCval < 53) {
                                                            return 8;
                                                        }
                                                        else if (refDCval >= 53) {
                                                            if (Rs < 1109) {
                                                                return 8;
                                                            }
                                                            else if (Rs >= 1109) {
                                                                return 1;
                                                            }
                                                        }
                                                    }
                                                    else if (posBalance >= 57) {
                                                        return 1;
                                                    }
                                                }
                                            }
                                            else if (Rs >= 1704) {
                                                if (Cs < 1627) {
                                                    return 4;
                                                }
                                                else if (Cs >= 1627) {
                                                    return 8;
                                                }
                                            }
                                        }
                                        else if (ssDCval >= 55) {
                                            if (MVDiff < 7) {
                                                return 8;
                                            }
                                            else if (MVDiff >= 7) {
                                                if (negBalance < 43) {
                                                    return 8;
                                                }
                                                else if (negBalance >= 43) {
                                                    if (TSC < 5) {
                                                        return 8;
                                                    }
                                                    else if (TSC >= 5) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (ssDCval >= 73) {
                                        if (posBalance < 58) {
                                            if (Rs < 2576) {
                                                if (mu_mv_mag_sq < 114) {
                                                    return 4;
                                                }
                                                else if (mu_mv_mag_sq >= 114) {
                                                    if (tcor < 95) {
                                                        return 8;
                                                    }
                                                    else if (tcor >= 95) {
                                                        if (Rs < 1678) {
                                                            return 8;
                                                        }
                                                        else if (Rs >= 1678) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (Rs >= 2576) {
                                                if (Cs < 1364) {
                                                    return 1;
                                                }
                                                else if (Cs >= 1364) {
                                                    if (negBalance < 41) {
                                                        return 8;
                                                    }
                                                    else if (negBalance >= 41) {
                                                        if (Rs < 2790) {
                                                            return 1;
                                                        }
                                                        else if (Rs >= 2790) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (posBalance >= 58) {
                                            return 8;
                                        }
                                    }
                                }
                            }
                            else if (mu_mv_mag_sq >= 121) {
                                if (posBalance < 2) {
                                    if (mu_mv_mag_sq < 124) {
                                        if (gchDC < 1) {
                                            return 8;
                                        }
                                        else if (gchDC >= 1) {
                                            if (Rs < 1964) {
                                                return 1;
                                            }
                                            else if (Rs >= 1964) {
                                                return 4;
                                            }
                                        }
                                    }
                                    else if (mu_mv_mag_sq >= 124) {
                                        if (Cs < 2313) {
                                            return 1;
                                        }
                                        else if (Cs >= 2313) {
                                            if (Rs < 1757) {
                                                return 8;
                                            }
                                            else if (Rs >= 1757) {
                                                return 2;
                                            }
                                        }
                                    }
                                }
                                else if (posBalance >= 2) {
                                    if (SC < 7308) {
                                        if (AFD < 4) {
                                            if (mu_mv_mag_sq < 125) {
                                                if (MVDiff < 7) {
                                                    if (tcor < 98) {
                                                        if (SC < 6621) {
                                                            if (diffMVdiffVal < 1) {
                                                                return 8;
                                                            }
                                                            else if (diffMVdiffVal >= 1) {
                                                                if (Cs < 1735) {
                                                                    return 8;
                                                                }
                                                                else if (Cs >= 1735) {
                                                                    if (ssDCval < 51) {
                                                                        return 4;
                                                                    }
                                                                    else if (ssDCval >= 51) {
                                                                        return 8;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (SC >= 6621) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (tcor >= 98) {
                                                        if (refDCval < 107) {
                                                            return 8;
                                                        }
                                                        else if (refDCval >= 107) {
                                                            if (negBalance < 24) {
                                                                if (Rs < 1032) {
                                                                    return 1;
                                                                }
                                                                else if (Rs >= 1032) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (negBalance >= 24) {
                                                                if (Rs < 3149) {
                                                                    return 8;
                                                                }
                                                                else if (Rs >= 3149) {
                                                                    if (posBalance < 42) {
                                                                        if (Rs < 3227) {
                                                                            return 8;
                                                                        }
                                                                        else if (Rs >= 3227) {
                                                                            if (Rs < 3478) {
                                                                                return 1;
                                                                            }
                                                                            else if (Rs >= 3478) {
                                                                                return 8;
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (posBalance >= 42) {
                                                                        return 1;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (MVDiff >= 7) {
                                                    if (Cs < 806) {
                                                        return 4;
                                                    }
                                                    else if (Cs >= 806) {
                                                        return 1;
                                                    }
                                                }
                                            }
                                            else if (mu_mv_mag_sq >= 125) {
                                                if (diffMVdiffVal < 12) {
                                                    if (ssDCval < 104) {
                                                        return 8;
                                                    }
                                                    else if (ssDCval >= 104) {
                                                        if (posBalance < 52) {
                                                            if (Rs < 2008) {
                                                                if (refDCval < 113) {
                                                                    if (SC < 2703) {
                                                                        return 4;
                                                                    }
                                                                    else if (SC >= 2703) {
                                                                        if (SC < 3477) {
                                                                            return 8;
                                                                        }
                                                                        else if (SC >= 3477) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                }
                                                                else if (refDCval >= 113) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (Rs >= 2008) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (posBalance >= 52) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                                else if (diffMVdiffVal >= 12) {
                                                    return 4;
                                                }
                                            }
                                        }
                                        else if (AFD >= 4) {
                                            if (mcTcor < 96) {
                                                if (tcor < 95) {
                                                    return 8;
                                                }
                                                else if (tcor >= 95) {
                                                    if (refDCval < 74) {
                                                        if (ssDCval < 65) {
                                                            return 8;
                                                        }
                                                        else if (ssDCval >= 65) {
                                                            return 2;
                                                        }
                                                    }
                                                    else if (refDCval >= 74) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                            else if (mcTcor >= 96) {
                                                if (ssDCval < 88) {
                                                    if (Cs < 1734) {
                                                        if (negBalance < 44) {
                                                            return 8;
                                                        }
                                                        else if (negBalance >= 44) {
                                                            if (Rs < 1810) {
                                                                return 8;
                                                            }
                                                            else if (Rs >= 1810) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                    else if (Cs >= 1734) {
                                                        if (ssDCval < 67) {
                                                            return 8;
                                                        }
                                                        else if (ssDCval >= 67) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                                else if (ssDCval >= 88) {
                                                    if (diffMVdiffVal < -125) {
                                                        return 4;
                                                    }
                                                    else if (diffMVdiffVal >= -125) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (SC >= 7308) {
                                        if (diffRsCsdiff < 2) {
                                            if (Cs < 3849) {
                                                return 1;
                                            }
                                            else if (Cs >= 3849) {
                                                if (Rs < 3774) {
                                                    return 8;
                                                }
                                                else if (Rs >= 3774) {
                                                    if (Rs < 3777) {
                                                        return 1;
                                                    }
                                                    else if (Rs >= 3777) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                        }
                                        else if (diffRsCsdiff >= 2) {
                                            if (Rs < 3739) {
                                                return 1;
                                            }
                                            else if (Rs >= 3739) {
                                                return 8;
                                            }
                                        }
                                    }
                                }
                            }
                        }
                        else if (SC >= 7785) {
                            return 1;
                        }
                    }
                    else if (Rs >= 3812) {
                        if (negBalance < 1) {
                            if (MVDiff < 5) {
                                return 8;
                            }
                            else if (MVDiff >= 5) {
                                return 2;
                            }
                        }
                        else if (negBalance >= 1) {
                            return 8;
                        }
                    }
                }
            }
            else if (AFD >= 6) {
                if (Cs < 1675) {
                    if (mu_mv_mag_sq < 129) {
                        if (AFD < 7) {
                            if (tcor < 93) {
                                if (Rs < 715) {
                                    if (ssDCval < 130) {
                                        if (SC < 1177) {
                                            return 1;
                                        }
                                        else if (SC >= 1177) {
                                            return 4;
                                        }
                                    }
                                    else if (ssDCval >= 130) {
                                        return 8;
                                    }
                                }
                                else if (Rs >= 715) {
                                    if (ssDCval < 72) {
                                        if (diffMVdiffVal < -1) {
                                            return 8;
                                        }
                                        else if (diffMVdiffVal >= -1) {
                                            if (Cs < 1622) {
                                                return 2;
                                            }
                                            else if (Cs >= 1622) {
                                                return 8;
                                            }
                                        }
                                    }
                                    else if (ssDCval >= 72) {
                                        if (posBalance < 59) {
                                            return 8;
                                        }
                                        else if (posBalance >= 59) {
                                            if (Cs < 1609) {
                                                return 2;
                                            }
                                            else if (Cs >= 1609) {
                                                return 8;
                                            }
                                        }
                                    }
                                }
                            }
                            else if (tcor >= 93) {
                                if (negBalance < 37) {
                                    return 8;
                                }
                                else if (negBalance >= 37) {
                                    if (posBalance < 57) {
                                        if (SC < 4019) {
                                            if (mu_mv_mag_sq < 119) {
                                                return 4;
                                            }
                                            else if (mu_mv_mag_sq >= 119) {
                                                return 8;
                                            }
                                        }
                                        else if (SC >= 4019) {
                                            if (Rs < 2789) {
                                                return 1;
                                            }
                                            else if (Rs >= 2789) {
                                                return 8;
                                            }
                                        }
                                    }
                                    else if (posBalance >= 57) {
                                        if (refDCval < 74) {
                                            if (TSC < 6) {
                                                return 8;
                                            }
                                            else if (TSC >= 6) {
                                                return 2;
                                            }
                                        }
                                        else if (refDCval >= 74) {
                                            if (diffAFD < -1) {
                                                return 2;
                                            }
                                            else if (diffAFD >= -1) {
                                                if (ssDCval < 75) {
                                                    if (Rs < 1853) {
                                                        return 8;
                                                    }
                                                    else if (Rs >= 1853) {
                                                        return 2;
                                                    }
                                                }
                                                else if (ssDCval >= 75) {
                                                    if (Rs < 1766) {
                                                        return 8;
                                                    }
                                                    else if (Rs >= 1766) {
                                                        if (RsCsDiff < 46) {
                                                            return 4;
                                                        }
                                                        else if (RsCsDiff >= 46) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                        }
                        else if (AFD >= 7) {
                            if (mcTcor < 88) {
                                if (negBalance < 55) {
                                    if (MVDiff < 13) {
                                        if (Rs < 1152) {
                                            return 4;
                                        }
                                        else if (Rs >= 1152) {
                                            return 8;
                                        }
                                    }
                                    else if (MVDiff >= 13) {
                                        if (AFD < 8) {
                                            return 4;
                                        }
                                        else if (AFD >= 8) {
                                            return 1;
                                        }
                                    }
                                }
                                else if (negBalance >= 55) {
                                    if (AFD < 15) {
                                        return 8;
                                    }
                                    else if (AFD >= 15) {
                                        return 2;
                                    }
                                }
                            }
                            else if (mcTcor >= 88) {
                                if (posBalance < 53) {
                                    if (refDCval < 109) {
                                        return 4;
                                    }
                                    else if (refDCval >= 109) {
                                        return 8;
                                    }
                                }
                                else if (posBalance >= 53) {
                                    if (ssDCval < 111) {
                                        if (negBalance < 49) {
                                            if (mcTcor < 93) {
                                                if (ssDCval < 72) {
                                                    return 2;
                                                }
                                                else if (ssDCval >= 72) {
                                                    if (ssDCval < 75) {
                                                        return 8;
                                                    }
                                                    else if (ssDCval >= 75) {
                                                        if (Rs < 1734) {
                                                            return 8;
                                                        }
                                                        else if (Rs >= 1734) {
                                                            return 2;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (mcTcor >= 93) {
                                                if (MVDiff < 2) {
                                                    return 2;
                                                }
                                                else if (MVDiff >= 2) {
                                                    return 8;
                                                }
                                            }
                                        }
                                        else if (negBalance >= 49) {
                                            return 2;
                                        }
                                    }
                                    else if (ssDCval >= 111) {
                                        if (RsDiff < 12) {
                                            if (ssDCval < 119) {
                                                if (SC < 4055) {
                                                    return 4;
                                                }
                                                else if (SC >= 4055) {
                                                    return 8;
                                                }
                                            }
                                            else if (ssDCval >= 119) {
                                                return 8;
                                            }
                                        }
                                        else if (RsDiff >= 12) {
                                            if (Rs < 2579) {
                                                if (Cs < 796) {
                                                    return 8;
                                                }
                                                else if (Cs >= 796) {
                                                    return 2;
                                                }
                                            }
                                            else if (Rs >= 2579) {
                                                return 4;
                                            }
                                        }
                                    }
                                }
                            }
                        }
                    }
                    else if (mu_mv_mag_sq >= 129) {
                        if (AFD < 33) {
                            if (CsDiff < 6) {
                                if (diffMVdiffVal < -46) {
                                    if (ssDCval < 149) {
                                        return 2;
                                    }
                                    else if (ssDCval >= 149) {
                                        return 1;
                                    }
                                }
                                else if (diffMVdiffVal >= -46) {
                                    if (diffAFD < -2) {
                                        if (AFD < 18) {
                                            return 1;
                                        }
                                        else if (AFD >= 18) {
                                            return 2;
                                        }
                                    }
                                    else if (diffAFD >= -2) {
                                        return 1;
                                    }
                                }
                            }
                            else if (CsDiff >= 6) {
                                return 1;
                            }
                        }
                        else if (AFD >= 33) {
                            if (RsDiff < 13) {
                                return 8;
                            }
                            else if (RsDiff >= 13) {
                                return 2;
                            }
                        }
                    }
                }
                else if (Cs >= 1675) {
                    if (mu_mv_mag_sq < 128) {
                        if (refDCval < 94) {
                            if (Cs < 3522) {
                                if (refDCval < 86) {
                                    if (Rs < 1946) {
                                        if (tcor < 94) {
                                            if (TSC < 7) {
                                                return 8;
                                            }
                                            else if (TSC >= 7) {
                                                if (ssDCval < 76) {
                                                    if (AFD < 8) {
                                                        return 4;
                                                    }
                                                    else if (AFD >= 8) {
                                                        return 2;
                                                    }
                                                }
                                                else if (ssDCval >= 76) {
                                                    return 8;
                                                }
                                            }
                                        }
                                        else if (tcor >= 94) {
                                            if (negBalance < 44) {
                                                return 4;
                                            }
                                            else if (negBalance >= 44) {
                                                if (ssDCval < 81) {
                                                    return 8;
                                                }
                                                else if (ssDCval >= 81) {
                                                    return 2;
                                                }
                                            }
                                        }
                                    }
                                    else if (Rs >= 1946) {
                                        if (posBalance < 53) {
                                            if (Rs < 2083) {
                                                return 4;
                                            }
                                            else if (Rs >= 2083) {
                                                if (negBalance < 21) {
                                                    if (Cs < 2084) {
                                                        return 4;
                                                    }
                                                    else if (Cs >= 2084) {
                                                        if (Cs < 2475) {
                                                            return 8;
                                                        }
                                                        else if (Cs >= 2475) {
                                                            return 1;
                                                        }
                                                    }
                                                }
                                                else if (negBalance >= 21) {
                                                    return 8;
                                                }
                                            }
                                        }
                                        else if (posBalance >= 53) {
                                            if (MVDiff < 1) {
                                                if (mcTcor < 93) {
                                                    return 8;
                                                }
                                                else if (mcTcor >= 93) {
                                                    return 4;
                                                }
                                            }
                                            else if (MVDiff >= 1) {
                                                if (RsCsDiff < 94) {
                                                    if (mcTcor < 94) {
                                                        return 8;
                                                    }
                                                    else if (mcTcor >= 94) {
                                                        if (Cs < 2613) {
                                                            if (mcTcor < 95) {
                                                                if (Rs < 3091) {
                                                                    return 4;
                                                                }
                                                                else if (Rs >= 3091) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (mcTcor >= 95) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (Cs >= 2613) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                                else if (RsCsDiff >= 94) {
                                                    return 4;
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (refDCval >= 86) {
                                    if (Scindex < 6) {
                                        if (posBalance < 46) {
                                            if (MVDiff < 8) {
                                                return 1;
                                            }
                                            else if (MVDiff >= 8) {
                                                return 8;
                                            }
                                        }
                                        else if (posBalance >= 46) {
                                            if (mu_mv_mag_sq < 106) {
                                                if (refDCval < 93) {
                                                    if (tcor < 95) {
                                                        if (ssDCval < 93) {
                                                            return 8;
                                                        }
                                                        else if (ssDCval >= 93) {
                                                            return 2;
                                                        }
                                                    }
                                                    else if (tcor >= 95) {
                                                        if (SC < 4471) {
                                                            return 2;
                                                        }
                                                        else if (SC >= 4471) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                                else if (refDCval >= 93) {
                                                    if (diffMVdiffVal < 2) {
                                                        return 2;
                                                    }
                                                    else if (diffMVdiffVal >= 2) {
                                                        if (Rs < 2444) {
                                                            return 2;
                                                        }
                                                        else if (Rs >= 2444) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (mu_mv_mag_sq >= 106) {
                                                return 8;
                                            }
                                        }
                                    }
                                    else if (Scindex >= 6) {
                                        if (TSC < 13) {
                                            return 2;
                                        }
                                        else if (TSC >= 13) {
                                            return 8;
                                        }
                                    }
                                }
                            }
                            else if (Cs >= 3522) {
                                if (SC < 14578) {
                                    if (RsCsDiff < 163) {
                                        if (AFD < 7) {
                                            if (Cs < 5021) {
                                                return 8;
                                            }
                                            else if (Cs >= 5021) {
                                                if (SC < 10815) {
                                                    if (posBalance < 58) {
                                                        return 4;
                                                    }
                                                    else if (posBalance >= 58) {
                                                        return 8;
                                                    }
                                                }
                                                else if (SC >= 10815) {
                                                    return 8;
                                                }
                                            }
                                        }
                                        else if (AFD >= 7) {
                                            if (Scindex < 7) {
                                                if (diffAFD < 1) {
                                                    return 4;
                                                }
                                                else if (diffAFD >= 1) {
                                                    return 8;
                                                }
                                            }
                                            else if (Scindex >= 7) {
                                                return 8;
                                            }
                                        }
                                    }
                                    else if (RsCsDiff >= 163) {
                                        return 8;
                                    }
                                }
                                else if (SC >= 14578) {
                                    if (tcor < 77) {
                                        if (RsCsDiff < 950) {
                                            return 4;
                                        }
                                        else if (RsCsDiff >= 950) {
                                            if (diffRsCsdiff < 48) {
                                                if (diffRsCsdiff < -1) {
                                                    if (Cs < 7736) {
                                                        return 2;
                                                    }
                                                    else if (Cs >= 7736) {
                                                        return 4;
                                                    }
                                                }
                                                else if (diffRsCsdiff >= -1) {
                                                    return 2;
                                                }
                                            }
                                            else if (diffRsCsdiff >= 48) {
                                                return 4;
                                            }
                                        }
                                    }
                                    else if (tcor >= 77) {
                                        return 2;
                                    }
                                }
                            }
                        }
                        else if (refDCval >= 94) {
                            if (Cs < 3641) {
                                if (negBalance < 54) {
                                    if (mu_mv_mag_sq < 99) {
                                        if (refDCval < 134) {
                                            if (Rs < 4661) {
                                                if (tcor < 89) {
                                                    if (refDCval < 97) {
                                                        return 2;
                                                    }
                                                    else if (refDCval >= 97) {
                                                        if (MVDiff < 12) {
                                                            return 8;
                                                        }
                                                        else if (MVDiff >= 12) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                                else if (tcor >= 89) {
                                                    if (RsCsDiff < 46) {
                                                        return 8;
                                                    }
                                                    else if (RsCsDiff >= 46) {
                                                        if (mcTcor < 94) {
                                                            if (mu_mv_mag_sq < 92) {
                                                                if (Rs < 4067) {
                                                                    return 1;
                                                                }
                                                                else if (Rs >= 4067) {
                                                                    return 2;
                                                                }
                                                            }
                                                            else if (mu_mv_mag_sq >= 92) {
                                                                if (RsDiff < 10) {
                                                                    return 8;
                                                                }
                                                                else if (RsDiff >= 10) {
                                                                    if (SC < 7949) {
                                                                        return 4;
                                                                    }
                                                                    else if (SC >= 7949) {
                                                                        return 8;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (mcTcor >= 94) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (Rs >= 4661) {
                                                if (SC < 7992) {
                                                    return 2;
                                                }
                                                else if (SC >= 7992) {
                                                    if (tcor < 88) {
                                                        return 2;
                                                    }
                                                    else if (tcor >= 88) {
                                                        if (Cs < 2944) {
                                                            return 2;
                                                        }
                                                        else if (Cs >= 2944) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (refDCval >= 134) {
                                            if (Rs < 3204) {
                                                return 4;
                                            }
                                            else if (Rs >= 3204) {
                                                return 1;
                                            }
                                        }
                                    }
                                    else if (mu_mv_mag_sq >= 99) {
                                        if (refDCval < 122) {
                                            if (mu_mv_mag_sq < 123) {
                                                if (negBalance < 48) {
                                                    if (Rs < 3822) {
                                                        if (Cs < 2410) {
                                                            if (ssDCval < 109) {
                                                                if (Rs < 2525) {
                                                                    return 4;
                                                                }
                                                                else if (Rs >= 2525) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (ssDCval >= 109) {
                                                                if (RsCsDiff < 122) {
                                                                    if (AFD < 7) {
                                                                        return 4;
                                                                    }
                                                                    else if (AFD >= 7) {
                                                                        return 8;
                                                                    }
                                                                }
                                                                else if (RsCsDiff >= 122) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                        else if (Cs >= 2410) {
                                                            if (RsCsDiff < 37) {
                                                                if (ssDCval < 114) {
                                                                    if (Rs < 2712) {
                                                                        return 8;
                                                                    }
                                                                    else if (Rs >= 2712) {
                                                                        if (ssDCval < 113) {
                                                                            if (Cs < 3033) {
                                                                                return 8;
                                                                            }
                                                                            else if (Cs >= 3033) {
                                                                                if (Rs < 2946) {
                                                                                    return 4;
                                                                                }
                                                                                else if (Rs >= 2946) {
                                                                                    return 8;
                                                                                }
                                                                            }
                                                                        }
                                                                        else if (ssDCval >= 113) {
                                                                            if (Cs < 3089) {
                                                                                return 4;
                                                                            }
                                                                            else if (Cs >= 3089) {
                                                                                return 8;
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                                else if (ssDCval >= 114) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (RsCsDiff >= 37) {
                                                                if (refDCval < 99) {
                                                                    if (Cs < 2957) {
                                                                        return 2;
                                                                    }
                                                                    else if (Cs >= 2957) {
                                                                        return 8;
                                                                    }
                                                                }
                                                                else if (refDCval >= 99) {
                                                                    if (diffRsCsdiff < -78) {
                                                                        if (SC < 6583) {
                                                                            return 8;
                                                                        }
                                                                        else if (SC >= 6583) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                    else if (diffRsCsdiff >= -78) {
                                                                        return 8;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (Rs >= 3822) {
                                                        if (negBalance < 42) {
                                                            if (Rs < 4840) {
                                                                return 4;
                                                            }
                                                            else if (Rs >= 4840) {
                                                                return 2;
                                                            }
                                                        }
                                                        else if (negBalance >= 42) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                                else if (negBalance >= 48) {
                                                    if (refDCval < 109) {
                                                        if (CsDiff < 6) {
                                                            if (Rs < 2032) {
                                                                return 8;
                                                            }
                                                            else if (Rs >= 2032) {
                                                                return 2;
                                                            }
                                                        }
                                                        else if (CsDiff >= 6) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (refDCval >= 109) {
                                                        if (refDCval < 112) {
                                                            return 8;
                                                        }
                                                        else if (refDCval >= 112) {
                                                            if (Rs < 2650) {
                                                                if (posBalance < 56) {
                                                                    return 4;
                                                                }
                                                                else if (posBalance >= 56) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (Rs >= 2650) {
                                                                if (tcor < 97) {
                                                                    return 8;
                                                                }
                                                                else if (tcor >= 97) {
                                                                    if (Rs < 3035) {
                                                                        if (ssDCval < 114) {
                                                                            return 4;
                                                                        }
                                                                        else if (ssDCval >= 114) {
                                                                            if (SC < 5776) {
                                                                                if (Cs < 2983) {
                                                                                    return 8;
                                                                                }
                                                                                else if (Cs >= 2983) {
                                                                                    return 4;
                                                                                }
                                                                            }
                                                                            else if (SC >= 5776) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (Rs >= 3035) {
                                                                        return 8;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                            else if (mu_mv_mag_sq >= 123) {
                                                if (refDCval < 113) {
                                                    if (diffRsCsdiff < -8) {
                                                        if (Cs < 3073) {
                                                            return 8;
                                                        }
                                                        else if (Cs >= 3073) {
                                                            if (gchDC < 1) {
                                                                return 8;
                                                            }
                                                            else if (gchDC >= 1) {
                                                                if (Cs < 3175) {
                                                                    return 4;
                                                                }
                                                                else if (Cs >= 3175) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (diffRsCsdiff >= -8) {
                                                        return 8;
                                                    }
                                                }
                                                else if (refDCval >= 113) {
                                                    if (Rs < 4058) {
                                                        if (negBalance < 48) {
                                                            return 8;
                                                        }
                                                        else if (negBalance >= 48) {
                                                            if (diffRsCsdiff < -8) {
                                                                if (CsDiff < 5) {
                                                                    if (Rs < 2663) {
                                                                        return 8;
                                                                    }
                                                                    else if (Rs >= 2663) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (CsDiff >= 5) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (diffRsCsdiff >= -8) {
                                                                if (SC < 5738) {
                                                                    return 8;
                                                                }
                                                                else if (SC >= 5738) {
                                                                    if (Cs < 2965) {
                                                                        return 4;
                                                                    }
                                                                    else if (Cs >= 2965) {
                                                                        return 8;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (Rs >= 4058) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                        }
                                        else if (refDCval >= 122) {
                                            if (mcTcor < 86) {
                                                if (Rs < 3534) {
                                                    return 8;
                                                }
                                                else if (Rs >= 3534) {
                                                    return 1;
                                                }
                                            }
                                            else if (mcTcor >= 86) {
                                                return 8;
                                            }
                                        }
                                    }
                                }
                                else if (negBalance >= 54) {
                                    if (SC < 7302) {
                                        if (TSC < 14) {
                                            return 8;
                                        }
                                        else if (TSC >= 14) {
                                            return 2;
                                        }
                                    }
                                    else if (SC >= 7302) {
                                        return 4;
                                    }
                                }
                            }
                            else if (Cs >= 3641) {
                                if (refDCval < 106) {
                                    return 8;
                                }
                                else if (refDCval >= 106) {
                                    if (ssDCval < 109) {
                                        if (MVDiff < 1) {
                                            if (SC < 12166) {
                                                if (negBalance < 55) {
                                                    return 4;
                                                }
                                                else if (negBalance >= 55) {
                                                    if (CsDiff < 13) {
                                                        return 8;
                                                    }
                                                    else if (CsDiff >= 13) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                            else if (SC >= 12166) {
                                                return 8;
                                            }
                                        }
                                        else if (MVDiff >= 1) {
                                            return 8;
                                        }
                                    }
                                    else if (ssDCval >= 109) {
                                        if (mu_mv_mag_sq < 127) {
                                            if (tcor < 71) {
                                                if (SC < 9389) {
                                                    return 1;
                                                }
                                                else if (SC >= 9389) {
                                                    return 8;
                                                }
                                            }
                                            else if (tcor >= 71) {
                                                return 8;
                                            }
                                        }
                                        else if (mu_mv_mag_sq >= 127) {
                                            if (SC < 11020) {
                                                return 8;
                                            }
                                            else if (SC >= 11020) {
                                                return 4;
                                            }
                                        }
                                    }
                                }
                            }
                        }
                    }
                    else if (mu_mv_mag_sq >= 128) {
                        if (ssDCval < 107) {
                            if (ssDCval < 84) {
                                if (Rs < 4888) {
                                    if (Rs < 4794) {
                                        if (SC < 9343) {
                                            return 4;
                                        }
                                        else if (SC >= 9343) {
                                            if (Cs < 4874) {
                                                return 8;
                                            }
                                            else if (Cs >= 4874) {
                                                if (mcTcor < 92) {
                                                    return 8;
                                                }
                                                else if (mcTcor >= 92) {
                                                    if (ssDCval < 80) {
                                                        if (RsDiff < 8) {
                                                            return 4;
                                                        }
                                                        else if (RsDiff >= 8) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (ssDCval >= 80) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (Rs >= 4794) {
                                        if (Cs < 5082) {
                                            if (Rs < 4876) {
                                                if (negBalance < 48) {
                                                    if (CsDiff < 9) {
                                                        return 4;
                                                    }
                                                    else if (CsDiff >= 9) {
                                                        if (Cs < 5015) {
                                                            return 4;
                                                        }
                                                        else if (Cs >= 5015) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                                else if (negBalance >= 48) {
                                                    return 8;
                                                }
                                            }
                                            else if (Rs >= 4876) {
                                                return 8;
                                            }
                                        }
                                        else if (Cs >= 5082) {
                                            return 8;
                                        }
                                    }
                                }
                                else if (Rs >= 4888) {
                                    if (SC < 10175) {
                                        if (diffRsCsdiff < 26) {
                                            if (posBalance < 57) {
                                                return 8;
                                            }
                                            else if (posBalance >= 57) {
                                                return 4;
                                            }
                                        }
                                        else if (diffRsCsdiff >= 26) {
                                            if (SC < 9924) {
                                                return 4;
                                            }
                                            else if (SC >= 9924) {
                                                return 8;
                                            }
                                        }
                                    }
                                    else if (SC >= 10175) {
                                        return 4;
                                    }
                                }
                            }
                            else if (ssDCval >= 84) {
                                if (tcor < 91) {
                                    if (Rs < 7098) {
                                        return 4;
                                    }
                                    else if (Rs >= 7098) {
                                        return 2;
                                    }
                                }
                                else if (tcor >= 91) {
                                    if (ssDCval < 106) {
                                        if (SC < 7506) {
                                            if (Rs < 3375) {
                                                return 8;
                                            }
                                            else if (Rs >= 3375) {
                                                return 2;
                                            }
                                        }
                                        else if (SC >= 7506) {
                                            return 8;
                                        }
                                    }
                                    else if (ssDCval >= 106) {
                                        if (Cs < 6176) {
                                            return 4;
                                        }
                                        else if (Cs >= 6176) {
                                            return 8;
                                        }
                                    }
                                }
                            }
                        }
                        else if (ssDCval >= 107) {
                            if (SC < 7004) {
                                return 8;
                            }
                            else if (SC >= 7004) {
                                return 4;
                            }
                        }
                    }
                }
            }
        }
        return 4;
    }
    mfxU16 AGOPSelect12(mfxU32 Rs, mfxU32 CsDiff, mfxU32 TSC, mfxI32 diffAFD, mfxU32 Scindex, mfxU32 RsDiff, mfxI32 diffMVdiffVal, mfxU32 RsCsDiff, mfxU32 TSCindex, mfxI16 tcor, mfxU32 refDCval, mfxU32 MVDiff, mfxU32 AFD, mfxI32 diffTSC, mfxI32 diffRsCsdiff, mfxU32 posBalance, mfxU32 SC, mfxU32 mu_mv_mag_sq, mfxU32 negBalance, mfxI16 mcTcor, mfxU32 gchDC, mfxU32 Cs, mfxU32 ssDCval) {
        if (MVDiff < 5) {
            if (AFD < 7) {
                if (negBalance < 3) {
                    if (SC < 9345) {
                        if (SC < 7800) {
                            if (SC < 3702) {
                                if (refDCval < 30) {
                                    if (Rs < 366) {
                                        return 4;
                                    }
                                    else if (Rs >= 366) {
                                        return 8;
                                    }
                                }
                                else if (refDCval >= 30) {
                                    if (refDCval < 32) {
                                        if (Cs < 1246) {
                                            return 4;
                                        }
                                        else if (Cs >= 1246) {
                                            return 2;
                                        }
                                    }
                                    else if (refDCval >= 32) {
                                        if (posBalance < 2) {
                                            if (Rs < 1347) {
                                                if (Cs < 925) {
                                                    return 8;
                                                }
                                                else if (Cs >= 925) {
                                                    return 1;
                                                }
                                            }
                                            else if (Rs >= 1347) {
                                                return 8;
                                            }
                                        }
                                        else if (posBalance >= 2) {
                                            if (Rs < 2526) {
                                                return 8;
                                            }
                                            else if (Rs >= 2526) {
                                                return 4;
                                            }
                                        }
                                    }
                                }
                            }
                            else if (SC >= 3702) {
                                if (Cs < 4003) {
                                    if (mcTcor < 99) {
                                        if (Rs < 1894) {
                                            return 1;
                                        }
                                        else if (Rs >= 1894) {
                                            if (posBalance < 22) {
                                                if (SC < 7709) {
                                                    if (ssDCval < 116) {
                                                        return 8;
                                                    }
                                                    else if (ssDCval >= 116) {
                                                        return 1;
                                                    }
                                                }
                                                else if (SC >= 7709) {
                                                    if (AFD < 1) {
                                                        if (Cs < 3985) {
                                                            if (Cs < 3951) {
                                                                return 1;
                                                            }
                                                            else if (Cs >= 3951) {
                                                                if (Cs < 3980) {
                                                                    return 8;
                                                                }
                                                                else if (Cs >= 3980) {
                                                                    return 1;
                                                                }
                                                            }
                                                        }
                                                        else if (Cs >= 3985) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (AFD >= 1) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                            else if (posBalance >= 22) {
                                                if (negBalance < 2) {
                                                    if (Cs < 2575) {
                                                        return 4;
                                                    }
                                                    else if (Cs >= 2575) {
                                                        return 1;
                                                    }
                                                }
                                                else if (negBalance >= 2) {
                                                    return 8;
                                                }
                                            }
                                        }
                                    }
                                    else if (mcTcor >= 99) {
                                        if (SC < 7308) {
                                            return 8;
                                        }
                                        else if (SC >= 7308) {
                                            if (SC < 7541) {
                                                if (Cs < 3849) {
                                                    return 1;
                                                }
                                                else if (Cs >= 3849) {
                                                    return 8;
                                                }
                                            }
                                            else if (SC >= 7541) {
                                                if (Cs < 3984) {
                                                    return 8;
                                                }
                                                else if (Cs >= 3984) {
                                                    if (SC < 7764) {
                                                        return 8;
                                                    }
                                                    else if (SC >= 7764) {
                                                        return 1;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (Cs >= 4003) {
                                    return 1;
                                }
                            }
                        }
                        else if (SC >= 7800) {
                            return 8;
                        }
                    }
                    else if (SC >= 9345) {
                        if (Cs < 4632) {
                            return 1;
                        }
                        else if (Cs >= 4632) {
                            return 8;
                        }
                    }
                }
                else if (negBalance >= 3) {
                    if (mu_mv_mag_sq < 109) {
                        if (MVDiff < 1) {
                            if (ssDCval < 97) {
                                if (posBalance < 58) {
                                    if (negBalance < 36) {
                                        if (tcor < 97) {
                                            if (Rs < 984) {
                                                if (Cs < 403) {
                                                    return 1;
                                                }
                                                else if (Cs >= 403) {
                                                    return 4;
                                                }
                                            }
                                            else if (Rs >= 984) {
                                                return 1;
                                            }
                                        }
                                        else if (tcor >= 97) {
                                            return 4;
                                        }
                                    }
                                    else if (negBalance >= 36) {
                                        return 4;
                                    }
                                }
                                else if (posBalance >= 58) {
                                    if (Rs < 2202) {
                                        return 2;
                                    }
                                    else if (Rs >= 2202) {
                                        if (Rs < 3242) {
                                            return 4;
                                        }
                                        else if (Rs >= 3242) {
                                            return 8;
                                        }
                                    }
                                }
                            }
                            else if (ssDCval >= 97) {
                                if (Cs < 936) {
                                    if (AFD < 5) {
                                        if (diffMVdiffVal < -22) {
                                            if (Cs < 540) {
                                                return 4;
                                            }
                                            else if (Cs >= 540) {
                                                return 2;
                                            }
                                        }
                                        else if (diffMVdiffVal >= -22) {
                                            return 8;
                                        }
                                    }
                                    else if (AFD >= 5) {
                                        return 2;
                                    }
                                }
                                else if (Cs >= 936) {
                                    if (tcor < 97) {
                                        return 4;
                                    }
                                    else if (tcor >= 97) {
                                        if (TSC < 4) {
                                            if (RsDiff < 2) {
                                                if (Rs < 510) {
                                                    return 8;
                                                }
                                                else if (Rs >= 510) {
                                                    return 4;
                                                }
                                            }
                                            else if (RsDiff >= 2) {
                                                return 8;
                                            }
                                        }
                                        else if (TSC >= 4) {
                                            return 1;
                                        }
                                    }
                                }
                            }
                        }
                        else if (MVDiff >= 1) {
                            if (Rs < 750) {
                                if (refDCval < 161) {
                                    if (mu_mv_mag_sq < 43) {
                                        if (ssDCval < 70) {
                                            if (TSC < 3) {
                                                return 8;
                                            }
                                            else if (TSC >= 3) {
                                                return 4;
                                            }
                                        }
                                        else if (ssDCval >= 70) {
                                            if (Rs < 496) {
                                                return 2;
                                            }
                                            else if (Rs >= 496) {
                                                return 1;
                                            }
                                        }
                                    }
                                    else if (mu_mv_mag_sq >= 43) {
                                        if (refDCval < 92) {
                                            if (RsDiff < 1) {
                                                if (MVDiff < 2) {
                                                    if (gchDC < 1) {
                                                        return 4;
                                                    }
                                                    else if (gchDC >= 1) {
                                                        return 8;
                                                    }
                                                }
                                                else if (MVDiff >= 2) {
                                                    return 8;
                                                }
                                            }
                                            else if (RsDiff >= 1) {
                                                if (mcTcor < 96) {
                                                    return 4;
                                                }
                                                else if (mcTcor >= 96) {
                                                    if (negBalance < 34) {
                                                        return 4;
                                                    }
                                                    else if (negBalance >= 34) {
                                                        if (TSC < 3) {
                                                            return 4;
                                                        }
                                                        else if (TSC >= 3) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (refDCval >= 92) {
                                            if (SC < 1229) {
                                                return 8;
                                            }
                                            else if (SC >= 1229) {
                                                if (mu_mv_mag_sq < 92) {
                                                    return 4;
                                                }
                                                else if (mu_mv_mag_sq >= 92) {
                                                    if (AFD < 5) {
                                                        return 8;
                                                    }
                                                    else if (AFD >= 5) {
                                                        if (Rs < 618) {
                                                            return 8;
                                                        }
                                                        else if (Rs >= 618) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (refDCval >= 161) {
                                    if (Cs < 454) {
                                        return 4;
                                    }
                                    else if (Cs >= 454) {
                                        if (tcor < 99) {
                                            return 4;
                                        }
                                        else if (tcor >= 99) {
                                            return 2;
                                        }
                                    }
                                }
                            }
                            else if (Rs >= 750) {
                                if (ssDCval < 111) {
                                    if (refDCval < 54) {
                                        if (ssDCval < 39) {
                                            if (Rs < 1028) {
                                                return 8;
                                            }
                                            else if (Rs >= 1028) {
                                                return 2;
                                            }
                                        }
                                        else if (ssDCval >= 39) {
                                            if (Cs < 1191) {
                                                if (negBalance < 36) {
                                                    if (mcTcor < 96) {
                                                        if (Rs < 893) {
                                                            if (Cs < 565) {
                                                                return 4;
                                                            }
                                                            else if (Cs >= 565) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (Rs >= 893) {
                                                            if (negBalance < 35) {
                                                                return 1;
                                                            }
                                                            else if (negBalance >= 35) {
                                                                if (tcor < 90) {
                                                                    return 1;
                                                                }
                                                                else if (tcor >= 90) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (mcTcor >= 96) {
                                                        if (gchDC < 1) {
                                                            return 8;
                                                        }
                                                        else if (gchDC >= 1) {
                                                            return 1;
                                                        }
                                                    }
                                                }
                                                else if (negBalance >= 36) {
                                                    if (SC < 1917) {
                                                        return 4;
                                                    }
                                                    else if (SC >= 1917) {
                                                        if (mu_mv_mag_sq < 63) {
                                                            return 1;
                                                        }
                                                        else if (mu_mv_mag_sq >= 63) {
                                                            if (diffMVdiffVal < -2) {
                                                                if (diffMVdiffVal < -4) {
                                                                    return 4;
                                                                }
                                                                else if (diffMVdiffVal >= -4) {
                                                                    return 1;
                                                                }
                                                            }
                                                            else if (diffMVdiffVal >= -2) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                            else if (Cs >= 1191) {
                                                if (MVDiff < 2) {
                                                    return 8;
                                                }
                                                else if (MVDiff >= 2) {
                                                    if (Rs < 3510) {
                                                        if (Rs < 2328) {
                                                            return 4;
                                                        }
                                                        else if (Rs >= 2328) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (Rs >= 3510) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (refDCval >= 54) {
                                        if (negBalance < 16) {
                                            if (gchDC < 1) {
                                                if (Cs < 21) {
                                                    return 1;
                                                }
                                                else if (Cs >= 21) {
                                                    if (Cs < 1114) {
                                                        return 8;
                                                    }
                                                    else if (Cs >= 1114) {
                                                        return 1;
                                                    }
                                                }
                                            }
                                            else if (gchDC >= 1) {
                                                if (Rs < 829) {
                                                    return 1;
                                                }
                                                else if (Rs >= 829) {
                                                    return 8;
                                                }
                                            }
                                        }
                                        else if (negBalance >= 16) {
                                            if (AFD < 5) {
                                                if (posBalance < 32) {
                                                    if (SC < 3899) {
                                                        return 8;
                                                    }
                                                    else if (SC >= 3899) {
                                                        return 4;
                                                    }
                                                }
                                                else if (posBalance >= 32) {
                                                    if (posBalance < 38) {
                                                        if (refDCval < 92) {
                                                            return 1;
                                                        }
                                                        else if (refDCval >= 92) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (posBalance >= 38) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                            else if (AFD >= 5) {
                                                if (tcor < 98) {
                                                    if (mu_mv_mag_sq < 78) {
                                                        if (Cs < 566) {
                                                            if (RsDiff < 3) {
                                                                return 8;
                                                            }
                                                            else if (RsDiff >= 3) {
                                                                return 2;
                                                            }
                                                        }
                                                        else if (Cs >= 566) {
                                                            if (TSC < 4) {
                                                                if (Rs < 2325) {
                                                                    return 8;
                                                                }
                                                                else if (Rs >= 2325) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (TSC >= 4) {
                                                                if (Rs < 1544) {
                                                                    return 1;
                                                                }
                                                                else if (Rs >= 1544) {
                                                                    if (Rs < 2562) {
                                                                        return 4;
                                                                    }
                                                                    else if (Rs >= 2562) {
                                                                        return 1;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (mu_mv_mag_sq >= 78) {
                                                        if (negBalance < 46) {
                                                            if (diffAFD < -1) {
                                                                if (posBalance < 54) {
                                                                    if (Rs < 2485) {
                                                                        return 8;
                                                                    }
                                                                    else if (Rs >= 2485) {
                                                                        return 1;
                                                                    }
                                                                }
                                                                else if (posBalance >= 54) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (diffAFD >= -1) {
                                                                if (mu_mv_mag_sq < 103) {
                                                                    if (mu_mv_mag_sq < 82) {
                                                                        if (Rs < 1578) {
                                                                            return 8;
                                                                        }
                                                                        else if (Rs >= 1578) {
                                                                            if (AFD < 6) {
                                                                                return 4;
                                                                            }
                                                                            else if (AFD >= 6) {
                                                                                return 8;
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (mu_mv_mag_sq >= 82) {
                                                                        return 8;
                                                                    }
                                                                }
                                                                else if (mu_mv_mag_sq >= 103) {
                                                                    if (ssDCval < 77) {
                                                                        return 8;
                                                                    }
                                                                    else if (ssDCval >= 77) {
                                                                        return 4;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (negBalance >= 46) {
                                                            if (diffAFD < -1) {
                                                                return 2;
                                                            }
                                                            else if (diffAFD >= -1) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (tcor >= 98) {
                                                    if (SC < 3980) {
                                                        return 4;
                                                    }
                                                    else if (SC >= 3980) {
                                                        if (SC < 4222) {
                                                            return 8;
                                                        }
                                                        else if (SC >= 4222) {
                                                            if (Rs < 2472) {
                                                                return 8;
                                                            }
                                                            else if (Rs >= 2472) {
                                                                return 1;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (ssDCval >= 111) {
                                    if (diffRsCsdiff < 57) {
                                        return 8;
                                    }
                                    else if (diffRsCsdiff >= 57) {
                                        return 1;
                                    }
                                }
                            }
                        }
                    }
                    else if (mu_mv_mag_sq >= 109) {
                        if (Rs < 588) {
                            if (SC < 1260) {
                                if (Cs < 502) {
                                    if (SC < 480) {
                                        return 4;
                                    }
                                    else if (SC >= 480) {
                                        if (TSC < 1) {
                                            return 8;
                                        }
                                        else if (TSC >= 1) {
                                            if (Rs < 256) {
                                                if (TSCindex < 2) {
                                                    if (Cs < 310) {
                                                        if (Rs < 179) {
                                                            return 8;
                                                        }
                                                        else if (Rs >= 179) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (Cs >= 310) {
                                                        return 8;
                                                    }
                                                }
                                                else if (TSCindex >= 2) {
                                                    return 4;
                                                }
                                            }
                                            else if (Rs >= 256) {
                                                return 8;
                                            }
                                        }
                                    }
                                }
                                else if (Cs >= 502) {
                                    if (gchDC < 1) {
                                        if (tcor < 102) {
                                            return 4;
                                        }
                                        else if (tcor >= 102) {
                                            return 8;
                                        }
                                    }
                                    else if (gchDC >= 1) {
                                        return 8;
                                    }
                                }
                            }
                            else if (SC >= 1260) {
                                return 8;
                            }
                        }
                        else if (Rs >= 588) {
                            if (AFD < 4) {
                                if (ssDCval < 28) {
                                    if (MVDiff < 1) {
                                        return 8;
                                    }
                                    else if (MVDiff >= 1) {
                                        return 4;
                                    }
                                }
                                else if (ssDCval >= 28) {
                                    if (ssDCval < 104) {
                                        if (mu_mv_mag_sq < 118) {
                                            if (refDCval < 71) {
                                                if (negBalance < 4) {
                                                    if (RsDiff < 1) {
                                                        return 4;
                                                    }
                                                    else if (RsDiff >= 1) {
                                                        return 8;
                                                    }
                                                }
                                                else if (negBalance >= 4) {
                                                    if (diffAFD < -2) {
                                                        return 1;
                                                    }
                                                    else if (diffAFD >= -2) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                            else if (refDCval >= 71) {
                                                if (negBalance < 13) {
                                                    if (Cs < 3059) {
                                                        return 1;
                                                    }
                                                    else if (Cs >= 3059) {
                                                        return 8;
                                                    }
                                                }
                                                else if (negBalance >= 13) {
                                                    if (mu_mv_mag_sq < 117) {
                                                        return 8;
                                                    }
                                                    else if (mu_mv_mag_sq >= 117) {
                                                        if (Rs < 1751) {
                                                            return 8;
                                                        }
                                                        else if (Rs >= 1751) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (mu_mv_mag_sq >= 118) {
                                            if (tcor < 92) {
                                                if (posBalance < 34) {
                                                    return 8;
                                                }
                                                else if (posBalance >= 34) {
                                                    if (Cs < 1562) {
                                                        return 8;
                                                    }
                                                    else if (Cs >= 1562) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                            else if (tcor >= 92) {
                                                if (diffRsCsdiff < -60) {
                                                    if (SC < 4622) {
                                                        return 1;
                                                    }
                                                    else if (SC >= 4622) {
                                                        return 8;
                                                    }
                                                }
                                                else if (diffRsCsdiff >= -60) {
                                                    return 8;
                                                }
                                            }
                                        }
                                    }
                                    else if (ssDCval >= 104) {
                                        if (RsCsDiff < 12) {
                                            if (posBalance < 11) {
                                                if (tcor < 97) {
                                                    if (Rs < 1556) {
                                                        return 8;
                                                    }
                                                    else if (Rs >= 1556) {
                                                        if (ssDCval < 116) {
                                                            if (refDCval < 116) {
                                                                return 8;
                                                            }
                                                            else if (refDCval >= 116) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (ssDCval >= 116) {
                                                            if (SC < 5865) {
                                                                return 4;
                                                            }
                                                            else if (SC >= 5865) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (tcor >= 97) {
                                                    return 8;
                                                }
                                            }
                                            else if (posBalance >= 11) {
                                                if (posBalance < 49) {
                                                    if (diffMVdiffVal < -32) {
                                                        return 4;
                                                    }
                                                    else if (diffMVdiffVal >= -32) {
                                                        if (diffRsCsdiff < -4) {
                                                            if (TSCindex < 3) {
                                                                return 8;
                                                            }
                                                            else if (TSCindex >= 3) {
                                                                if (negBalance < 38) {
                                                                    if (diffAFD < -1) {
                                                                        if (SC < 6235) {
                                                                            return 8;
                                                                        }
                                                                        else if (SC >= 6235) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                    else if (diffAFD >= -1) {
                                                                        return 8;
                                                                    }
                                                                }
                                                                else if (negBalance >= 38) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                        else if (diffRsCsdiff >= -4) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                                else if (posBalance >= 49) {
                                                    if (refDCval < 106) {
                                                        return 4;
                                                    }
                                                    else if (refDCval >= 106) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                        }
                                        else if (RsCsDiff >= 12) {
                                            if (SC < 6454) {
                                                if (Cs < 2935) {
                                                    if (CsDiff < 4) {
                                                        if (SC < 2642) {
                                                            return 1;
                                                        }
                                                        else if (SC >= 2642) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (CsDiff >= 4) {
                                                        if (RsDiff < 4) {
                                                            return 4;
                                                        }
                                                        else if (RsDiff >= 4) {
                                                            if (mu_mv_mag_sq < 114) {
                                                                return 8;
                                                            }
                                                            else if (mu_mv_mag_sq >= 114) {
                                                                if (SC < 5507) {
                                                                    return 1;
                                                                }
                                                                else if (SC >= 5507) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (Cs >= 2935) {
                                                    if (mu_mv_mag_sq < 120) {
                                                        if (Cs < 3073) {
                                                            if (Rs < 3117) {
                                                                return 8;
                                                            }
                                                            else if (Rs >= 3117) {
                                                                if (ssDCval < 108) {
                                                                    return 8;
                                                                }
                                                                else if (ssDCval >= 108) {
                                                                    return 1;
                                                                }
                                                            }
                                                        }
                                                        else if (Cs >= 3073) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (mu_mv_mag_sq >= 120) {
                                                        if (Cs < 2998) {
                                                            return 1;
                                                        }
                                                        else if (Cs >= 2998) {
                                                            if (Rs < 3149) {
                                                                return 8;
                                                            }
                                                            else if (Rs >= 3149) {
                                                                if (posBalance < 39) {
                                                                    return 8;
                                                                }
                                                                else if (posBalance >= 39) {
                                                                    if (Cs < 3004) {
                                                                        return 8;
                                                                    }
                                                                    else if (Cs >= 3004) {
                                                                        return 1;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                            else if (SC >= 6454) {
                                                if (negBalance < 13) {
                                                    if (diffTSC < -1) {
                                                        return 4;
                                                    }
                                                    else if (diffTSC >= -1) {
                                                        if (Cs < 2882) {
                                                            return 8;
                                                        }
                                                        else if (Cs >= 2882) {
                                                            if (Cs < 2961) {
                                                                if (Rs < 3811) {
                                                                    return 4;
                                                                }
                                                                else if (Rs >= 3811) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (Cs >= 2961) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (negBalance >= 13) {
                                                    return 8;
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                            else if (AFD >= 4) {
                                if (Rs < 2410) {
                                    if (Cs < 2023) {
                                        if (tcor < 94) {
                                            if (TSC < 6) {
                                                return 8;
                                            }
                                            else if (TSC >= 6) {
                                                if (ssDCval < 72) {
                                                    if (tcor < 92) {
                                                        return 8;
                                                    }
                                                    else if (tcor >= 92) {
                                                        return 2;
                                                    }
                                                }
                                                else if (ssDCval >= 72) {
                                                    if (SC < 3663) {
                                                        return 8;
                                                    }
                                                    else if (SC >= 3663) {
                                                        if (SC < 3683) {
                                                            return 2;
                                                        }
                                                        else if (SC >= 3683) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (tcor >= 94) {
                                            if (TSC < 6) {
                                                if (SC < 3467) {
                                                    if (RsCsDiff < 11) {
                                                        if (diffAFD < -1) {
                                                            return 4;
                                                        }
                                                        else if (diffAFD >= -1) {
                                                            if (Rs < 1808) {
                                                                return 8;
                                                            }
                                                            else if (Rs >= 1808) {
                                                                if (negBalance < 42) {
                                                                    if (ssDCval < 73) {
                                                                        return 8;
                                                                    }
                                                                    else if (ssDCval >= 73) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (negBalance >= 42) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (RsCsDiff >= 11) {
                                                        if (diffAFD < -1) {
                                                            if (Rs < 1789) {
                                                                return 8;
                                                            }
                                                            else if (Rs >= 1789) {
                                                                if (Rs < 1815) {
                                                                    if (Rs < 1800) {
                                                                        return 2;
                                                                    }
                                                                    else if (Rs >= 1800) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (Rs >= 1815) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                        else if (diffAFD >= -1) {
                                                            if (SC < 2756) {
                                                                if (CsDiff < 4) {
                                                                    return 8;
                                                                }
                                                                else if (CsDiff >= 4) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (SC >= 2756) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (SC >= 3467) {
                                                    if (ssDCval < 79) {
                                                        if (tcor < 96) {
                                                            if (mu_mv_mag_sq < 117) {
                                                                return 4;
                                                            }
                                                            else if (mu_mv_mag_sq >= 117) {
                                                                if (SC < 3674) {
                                                                    if (ssDCval < 75) {
                                                                        if (Rs < 1862) {
                                                                            return 8;
                                                                        }
                                                                        else if (Rs >= 1862) {
                                                                            return 2;
                                                                        }
                                                                    }
                                                                    else if (ssDCval >= 75) {
                                                                        return 8;
                                                                    }
                                                                }
                                                                else if (SC >= 3674) {
                                                                    if (Rs < 2046) {
                                                                        return 4;
                                                                    }
                                                                    else if (Rs >= 2046) {
                                                                        return 8;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (tcor >= 96) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (ssDCval >= 79) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                            else if (TSC >= 6) {
                                                if (ssDCval < 74) {
                                                    if (posBalance < 46) {
                                                        return 4;
                                                    }
                                                    else if (posBalance >= 46) {
                                                        return 2;
                                                    }
                                                }
                                                else if (ssDCval >= 74) {
                                                    if (RsCsDiff < 29) {
                                                        if (refDCval < 75) {
                                                            return 2;
                                                        }
                                                        else if (refDCval >= 75) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (RsCsDiff >= 29) {
                                                        if (SC < 3289) {
                                                            return 8;
                                                        }
                                                        else if (SC >= 3289) {
                                                            if (Cs < 1725) {
                                                                if (CsDiff < 5) {
                                                                    return 4;
                                                                }
                                                                else if (CsDiff >= 5) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (Cs >= 1725) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (Cs >= 2023) {
                                        if (mcTcor < 96) {
                                            if (mu_mv_mag_sq < 121) {
                                                if (refDCval < 81) {
                                                    if (tcor < 94) {
                                                        return 8;
                                                    }
                                                    else if (tcor >= 94) {
                                                        if (Rs < 2212) {
                                                            return 4;
                                                        }
                                                        else if (Rs >= 2212) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                                else if (refDCval >= 81) {
                                                    if (Rs < 2250) {
                                                        if (Cs < 2197) {
                                                            return 4;
                                                        }
                                                        else if (Cs >= 2197) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (Rs >= 2250) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                            else if (mu_mv_mag_sq >= 121) {
                                                return 8;
                                            }
                                        }
                                        else if (mcTcor >= 96) {
                                            if (negBalance < 32) {
                                                if (posBalance < 37) {
                                                    if (TSC < 3) {
                                                        return 4;
                                                    }
                                                    else if (TSC >= 3) {
                                                        return 8;
                                                    }
                                                }
                                                else if (posBalance >= 37) {
                                                    return 1;
                                                }
                                            }
                                            else if (negBalance >= 32) {
                                                if (posBalance < 56) {
                                                    return 4;
                                                }
                                                else if (posBalance >= 56) {
                                                    if (tcor < 96) {
                                                        return 8;
                                                    }
                                                    else if (tcor >= 96) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (Rs >= 2410) {
                                    if (CsDiff < 8) {
                                        if (TSCindex < 7) {
                                            if (CsDiff < 7) {
                                                return 8;
                                            }
                                            else if (CsDiff >= 7) {
                                                if (diffRsCsdiff < 60) {
                                                    return 8;
                                                }
                                                else if (diffRsCsdiff >= 60) {
                                                    if (Cs < 3608) {
                                                        return 1;
                                                    }
                                                    else if (Cs >= 3608) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                        }
                                        else if (TSCindex >= 7) {
                                            if (negBalance < 48) {
                                                if (SC < 9752) {
                                                    if (Rs < 2894) {
                                                        if (mu_mv_mag_sq < 124) {
                                                            if (Cs < 2987) {
                                                                if (diffAFD < -2) {
                                                                    return 4;
                                                                }
                                                                else if (diffAFD >= -2) {
                                                                    if (diffRsCsdiff < -8) {
                                                                        if (ssDCval < 113) {
                                                                            return 8;
                                                                        }
                                                                        else if (ssDCval >= 113) {
                                                                            if (SC < 5608) {
                                                                                return 8;
                                                                            }
                                                                            else if (SC >= 5608) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (diffRsCsdiff >= -8) {
                                                                        return 8;
                                                                    }
                                                                }
                                                            }
                                                            else if (Cs >= 2987) {
                                                                if (refDCval < 114) {
                                                                    if (Rs < 2711) {
                                                                        return 8;
                                                                    }
                                                                    else if (Rs >= 2711) {
                                                                        if (CsDiff < 5) {
                                                                            if (posBalance < 59) {
                                                                                if (tcor < 97) {
                                                                                    return 8;
                                                                                }
                                                                                else if (tcor >= 97) {
                                                                                    return 4;
                                                                                }
                                                                            }
                                                                            else if (posBalance >= 59) {
                                                                                return 8;
                                                                            }
                                                                        }
                                                                        else if (CsDiff >= 5) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                }
                                                                else if (refDCval >= 114) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                        else if (mu_mv_mag_sq >= 124) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (Rs >= 2894) {
                                                        return 8;
                                                    }
                                                }
                                                else if (SC >= 9752) {
                                                    if (Rs < 5226) {
                                                        if (diffRsCsdiff < -39) {
                                                            return 8;
                                                        }
                                                        else if (diffRsCsdiff >= -39) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (Rs >= 5226) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                            else if (negBalance >= 48) {
                                                if (SC < 6145) {
                                                    if (Rs < 2668) {
                                                        return 8;
                                                    }
                                                    else if (Rs >= 2668) {
                                                        if (mu_mv_mag_sq < 123) {
                                                            if (ssDCval < 112) {
                                                                return 8;
                                                            }
                                                            else if (ssDCval >= 112) {
                                                                if (mcTcor < 98) {
                                                                    if (tcor < 97) {
                                                                        return 8;
                                                                    }
                                                                    else if (tcor >= 97) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (mcTcor >= 98) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                        else if (mu_mv_mag_sq >= 123) {
                                                            if (ssDCval < 113) {
                                                                if (Rs < 2861) {
                                                                    return 8;
                                                                }
                                                                else if (Rs >= 2861) {
                                                                    if (Rs < 2872) {
                                                                        return 4;
                                                                    }
                                                                    else if (Rs >= 2872) {
                                                                        return 8;
                                                                    }
                                                                }
                                                            }
                                                            else if (ssDCval >= 113) {
                                                                if (diffRsCsdiff < -5) {
                                                                    return 4;
                                                                }
                                                                else if (diffRsCsdiff >= -5) {
                                                                    if (Rs < 2796) {
                                                                        return 8;
                                                                    }
                                                                    else if (Rs >= 2796) {
                                                                        if (posBalance < 57) {
                                                                            return 8;
                                                                        }
                                                                        else if (posBalance >= 57) {
                                                                            if (Rs < 2824) {
                                                                                return 4;
                                                                            }
                                                                            else if (Rs >= 2824) {
                                                                                return 8;
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (SC >= 6145) {
                                                    return 8;
                                                }
                                            }
                                        }
                                    }
                                    else if (CsDiff >= 8) {
                                        if (posBalance < 55) {
                                            return 8;
                                        }
                                        else if (posBalance >= 55) {
                                            if (ssDCval < 82) {
                                                if (Cs < 5059) {
                                                    if (ssDCval < 80) {
                                                        if (tcor < 93) {
                                                            return 8;
                                                        }
                                                        else if (tcor >= 93) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (ssDCval >= 80) {
                                                        if (negBalance < 46) {
                                                            return 4;
                                                        }
                                                        else if (negBalance >= 46) {
                                                            if (Cs < 4915) {
                                                                return 8;
                                                            }
                                                            else if (Cs >= 4915) {
                                                                if (Rs < 4931) {
                                                                    if (Rs < 4804) {
                                                                        return 4;
                                                                    }
                                                                    else if (Rs >= 4804) {
                                                                        if (Rs < 4860) {
                                                                            if (Cs < 4921) {
                                                                                return 4;
                                                                            }
                                                                            else if (Cs >= 4921) {
                                                                                return 8;
                                                                            }
                                                                        }
                                                                        else if (Rs >= 4860) {
                                                                            if (Cs < 5055) {
                                                                                return 4;
                                                                            }
                                                                            else if (Cs >= 5055) {
                                                                                return 8;
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                                else if (Rs >= 4931) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (Cs >= 5059) {
                                                    if (RsCsDiff < 137) {
                                                        return 4;
                                                    }
                                                    else if (RsCsDiff >= 137) {
                                                        if (SC < 10230) {
                                                            return 8;
                                                        }
                                                        else if (SC >= 10230) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (ssDCval >= 82) {
                                                return 8;
                                            }
                                        }
                                    }
                                }
                            }
                        }
                    }
                }
            }
            else if (AFD >= 7) {
                if (Cs < 4177) {
                    if (mu_mv_mag_sq < 128) {
                        if (SC < 3759) {
                            if (RsDiff < 5) {
                                if (TSC < 7) {
                                    if (tcor < 91) {
                                        if (mu_mv_mag_sq < 102) {
                                            if (Rs < 1434) {
                                                if (ssDCval < 54) {
                                                    if (RsCsDiff < 33) {
                                                        return 1;
                                                    }
                                                    else if (RsCsDiff >= 33) {
                                                        return 4;
                                                    }
                                                }
                                                else if (ssDCval >= 54) {
                                                    if (Rs < 1389) {
                                                        if (diffTSC < 1) {
                                                            return 4;
                                                        }
                                                        else if (diffTSC >= 1) {
                                                            if (Cs < 684) {
                                                                return 4;
                                                            }
                                                            else if (Cs >= 684) {
                                                                return 1;
                                                            }
                                                        }
                                                    }
                                                    else if (Rs >= 1389) {
                                                        return 1;
                                                    }
                                                }
                                            }
                                            else if (Rs >= 1434) {
                                                return 8;
                                            }
                                        }
                                        else if (mu_mv_mag_sq >= 102) {
                                            return 8;
                                        }
                                    }
                                    else if (tcor >= 91) {
                                        if (AFD < 9) {
                                            if (refDCval < 102) {
                                                if (RsCsDiff < 37) {
                                                    if (refDCval < 59) {
                                                        return 4;
                                                    }
                                                    else if (refDCval >= 59) {
                                                        if (ssDCval < 75) {
                                                            return 2;
                                                        }
                                                        else if (ssDCval >= 75) {
                                                            if (tcor < 94) {
                                                                return 4;
                                                            }
                                                            else if (tcor >= 94) {
                                                                if (posBalance < 17) {
                                                                    return 4;
                                                                }
                                                                else if (posBalance >= 17) {
                                                                    return 2;
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (RsCsDiff >= 37) {
                                                    if (diffRsCsdiff < 5) {
                                                        return 8;
                                                    }
                                                    else if (diffRsCsdiff >= 5) {
                                                        return 2;
                                                    }
                                                }
                                            }
                                            else if (refDCval >= 102) {
                                                if (Rs < 679) {
                                                    return 4;
                                                }
                                                else if (Rs >= 679) {
                                                    if (ssDCval < 100) {
                                                        return 4;
                                                    }
                                                    else if (ssDCval >= 100) {
                                                        return 2;
                                                    }
                                                }
                                            }
                                        }
                                        else if (AFD >= 9) {
                                            if (SC < 1730) {
                                                return 4;
                                            }
                                            else if (SC >= 1730) {
                                                return 8;
                                            }
                                        }
                                    }
                                }
                                else if (TSC >= 7) {
                                    if (diffRsCsdiff < -16) {
                                        if (Cs < 1330) {
                                            if (SC < 2324) {
                                                return 2;
                                            }
                                            else if (SC >= 2324) {
                                                return 4;
                                            }
                                        }
                                        else if (Cs >= 1330) {
                                            return 8;
                                        }
                                    }
                                    else if (diffRsCsdiff >= -16) {
                                        if (mu_mv_mag_sq < 120) {
                                            if (ssDCval < 113) {
                                                if (tcor < 83) {
                                                    return 8;
                                                }
                                                else if (tcor >= 83) {
                                                    return 2;
                                                }
                                            }
                                            else if (ssDCval >= 113) {
                                                if (Rs < 813) {
                                                    return 2;
                                                }
                                                else if (Rs >= 813) {
                                                    return 8;
                                                }
                                            }
                                        }
                                        else if (mu_mv_mag_sq >= 120) {
                                            if (SC < 3041) {
                                                return 8;
                                            }
                                            else if (SC >= 3041) {
                                                return 2;
                                            }
                                        }
                                    }
                                }
                            }
                            else if (RsDiff >= 5) {
                                if (mcTcor < 88) {
                                    if (mu_mv_mag_sq < 83) {
                                        if (SC < 3322) {
                                            return 1;
                                        }
                                        else if (SC >= 3322) {
                                            if (Rs < 1866) {
                                                return 4;
                                            }
                                            else if (Rs >= 1866) {
                                                return 8;
                                            }
                                        }
                                    }
                                    else if (mu_mv_mag_sq >= 83) {
                                        return 8;
                                    }
                                }
                                else if (mcTcor >= 88) {
                                    if (tcor < 84) {
                                        if (Cs < 951) {
                                            return 8;
                                        }
                                        else if (Cs >= 951) {
                                            return 4;
                                        }
                                    }
                                    else if (tcor >= 84) {
                                        if (mu_mv_mag_sq < 112) {
                                            if (negBalance < 45) {
                                                if (SC < 3377) {
                                                    if (AFD < 10) {
                                                        if (TSC < 4) {
                                                            return 2;
                                                        }
                                                        else if (TSC >= 4) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (AFD >= 10) {
                                                        return 2;
                                                    }
                                                }
                                                else if (SC >= 3377) {
                                                    return 4;
                                                }
                                            }
                                            else if (negBalance >= 45) {
                                                if (MVDiff < 4) {
                                                    if (Cs < 1861) {
                                                        return 2;
                                                    }
                                                    else if (Cs >= 1861) {
                                                        return 8;
                                                    }
                                                }
                                                else if (MVDiff >= 4) {
                                                    if (SC < 3405) {
                                                        if (CsDiff < 5) {
                                                            return 4;
                                                        }
                                                        else if (CsDiff >= 5) {
                                                            if (Rs < 1740) {
                                                                return 8;
                                                            }
                                                            else if (Rs >= 1740) {
                                                                return 1;
                                                            }
                                                        }
                                                    }
                                                    else if (SC >= 3405) {
                                                        if (Rs < 1646) {
                                                            return 4;
                                                        }
                                                        else if (Rs >= 1646) {
                                                            return 2;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (mu_mv_mag_sq >= 112) {
                                            if (tcor < 94) {
                                                return 8;
                                            }
                                            else if (tcor >= 94) {
                                                return 4;
                                            }
                                        }
                                    }
                                }
                            }
                        }
                        else if (SC >= 3759) {
                            if (mu_mv_mag_sq < 96) {
                                if (SC < 5053) {
                                    if (negBalance < 48) {
                                        if (posBalance < 57) {
                                            if (MVDiff < 1) {
                                                if (Rs < 1888) {
                                                    return 1;
                                                }
                                                else if (Rs >= 1888) {
                                                    return 4;
                                                }
                                            }
                                            else if (MVDiff >= 1) {
                                                if (Rs < 3445) {
                                                    if (diffMVdiffVal < -12) {
                                                        return 8;
                                                    }
                                                    else if (diffMVdiffVal >= -12) {
                                                        if (Cs < 1055) {
                                                            return 2;
                                                        }
                                                        else if (Cs >= 1055) {
                                                            if (SC < 4964) {
                                                                if (diffAFD < -2) {
                                                                    return 4;
                                                                }
                                                                else if (diffAFD >= -2) {
                                                                    if (RsDiff < 5) {
                                                                        if (ssDCval < 81) {
                                                                            return 4;
                                                                        }
                                                                        else if (ssDCval >= 81) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                    else if (RsDiff >= 5) {
                                                                        if (SC < 4083) {
                                                                            return 8;
                                                                        }
                                                                        else if (SC >= 4083) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                            else if (SC >= 4964) {
                                                                if (Cs < 2314) {
                                                                    return 4;
                                                                }
                                                                else if (Cs >= 2314) {
                                                                    if (Rs < 2452) {
                                                                        return 2;
                                                                    }
                                                                    else if (Rs >= 2452) {
                                                                        return 8;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (Rs >= 3445) {
                                                    return 4;
                                                }
                                            }
                                        }
                                        else if (posBalance >= 57) {
                                            if (TSC < 7) {
                                                return 4;
                                            }
                                            else if (TSC >= 7) {
                                                if (ssDCval < 84) {
                                                    return 4;
                                                }
                                                else if (ssDCval >= 84) {
                                                    if (Cs < 2233) {
                                                        return 2;
                                                    }
                                                    else if (Cs >= 2233) {
                                                        if (diffMVdiffVal < -3) {
                                                            return 2;
                                                        }
                                                        else if (diffMVdiffVal >= -3) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (negBalance >= 48) {
                                        if (tcor < 86) {
                                            if (negBalance < 51) {
                                                if (Rs < 2653) {
                                                    if (AFD < 11) {
                                                        if (Scindex < 5) {
                                                            return 4;
                                                        }
                                                        else if (Scindex >= 5) {
                                                            if (ssDCval < 109) {
                                                                return 4;
                                                            }
                                                            else if (ssDCval >= 109) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                    else if (AFD >= 11) {
                                                        if (refDCval < 124) {
                                                            return 4;
                                                        }
                                                        else if (refDCval >= 124) {
                                                            return 1;
                                                        }
                                                    }
                                                }
                                                else if (Rs >= 2653) {
                                                    return 4;
                                                }
                                            }
                                            else if (negBalance >= 51) {
                                                if (mu_mv_mag_sq < 3) {
                                                    if (tcor < 84) {
                                                        return 4;
                                                    }
                                                    else if (tcor >= 84) {
                                                        if (posBalance < 59) {
                                                            if (Cs < 2270) {
                                                                return 8;
                                                            }
                                                            else if (Cs >= 2270) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (posBalance >= 59) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                                else if (mu_mv_mag_sq >= 3) {
                                                    if (AFD < 12) {
                                                        if (ssDCval < 102) {
                                                            if (Rs < 2091) {
                                                                return 8;
                                                            }
                                                            else if (Rs >= 2091) {
                                                                return 2;
                                                            }
                                                        }
                                                        else if (ssDCval >= 102) {
                                                            if (mu_mv_mag_sq < 89) {
                                                                if (diffRsCsdiff < 7) {
                                                                    if (SC < 4380) {
                                                                        return 4;
                                                                    }
                                                                    else if (SC >= 4380) {
                                                                        if (RsDiff < 8) {
                                                                            return 8;
                                                                        }
                                                                        else if (RsDiff >= 8) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                }
                                                                else if (diffRsCsdiff >= 7) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (mu_mv_mag_sq >= 89) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                    else if (AFD >= 12) {
                                                        if (SC < 4759) {
                                                            return 2;
                                                        }
                                                        else if (SC >= 4759) {
                                                            if (SC < 4912) {
                                                                return 4;
                                                            }
                                                            else if (SC >= 4912) {
                                                                return 1;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (tcor >= 86) {
                                            if (refDCval < 86) {
                                                if (SC < 4127) {
                                                    if (Rs < 2145) {
                                                        return 2;
                                                    }
                                                    else if (Rs >= 2145) {
                                                        return 1;
                                                    }
                                                }
                                                else if (SC >= 4127) {
                                                    return 4;
                                                }
                                            }
                                            else if (refDCval >= 86) {
                                                if (ssDCval < 97) {
                                                    if (negBalance < 52) {
                                                        if (mu_mv_mag_sq < 27) {
                                                            if (Cs < 1981) {
                                                                return 1;
                                                            }
                                                            else if (Cs >= 1981) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (mu_mv_mag_sq >= 27) {
                                                            if (Rs < 1953) {
                                                                return 8;
                                                            }
                                                            else if (Rs >= 1953) {
                                                                return 2;
                                                            }
                                                        }
                                                    }
                                                    else if (negBalance >= 52) {
                                                        if (gchDC < 4) {
                                                            return 2;
                                                        }
                                                        else if (gchDC >= 4) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                                else if (ssDCval >= 97) {
                                                    if (Rs < 2120) {
                                                        return 8;
                                                    }
                                                    else if (Rs >= 2120) {
                                                        if (Rs < 3036) {
                                                            return 4;
                                                        }
                                                        else if (Rs >= 3036) {
                                                            return 2;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (SC >= 5053) {
                                    if (tcor < 77) {
                                        if (Rs < 2683) {
                                            return 4;
                                        }
                                        else if (Rs >= 2683) {
                                            return 1;
                                        }
                                    }
                                    else if (tcor >= 77) {
                                        if (Cs < 2607) {
                                            if (negBalance < 55) {
                                                if (TSC < 11) {
                                                    if (negBalance < 52) {
                                                        if (tcor < 92) {
                                                            if (Rs < 3039) {
                                                                return 8;
                                                            }
                                                            else if (Rs >= 3039) {
                                                                if (Rs < 3130) {
                                                                    return 4;
                                                                }
                                                                else if (Rs >= 3130) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                        else if (tcor >= 92) {
                                                            if (Rs < 3625) {
                                                                return 4;
                                                            }
                                                            else if (Rs >= 3625) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                    else if (negBalance >= 52) {
                                                        return 4;
                                                    }
                                                }
                                                else if (TSC >= 11) {
                                                    if (Rs < 2988) {
                                                        return 8;
                                                    }
                                                    else if (Rs >= 2988) {
                                                        return 1;
                                                    }
                                                }
                                            }
                                            else if (negBalance >= 55) {
                                                if (Rs < 2947) {
                                                    return 2;
                                                }
                                                else if (Rs >= 2947) {
                                                    return 8;
                                                }
                                            }
                                        }
                                        else if (Cs >= 2607) {
                                            if (Rs < 3804) {
                                                if (MVDiff < 3) {
                                                    if (mu_mv_mag_sq < 90) {
                                                        return 8;
                                                    }
                                                    else if (mu_mv_mag_sq >= 90) {
                                                        if (CsDiff < 9) {
                                                            if (Rs < 3544) {
                                                                return 2;
                                                            }
                                                            else if (Rs >= 3544) {
                                                                return 1;
                                                            }
                                                        }
                                                        else if (CsDiff >= 9) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                                else if (MVDiff >= 3) {
                                                    if (diffRsCsdiff < -35) {
                                                        if (RsDiff < 10) {
                                                            return 4;
                                                        }
                                                        else if (RsDiff >= 10) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (diffRsCsdiff >= -35) {
                                                        if (RsDiff < 9) {
                                                            return 8;
                                                        }
                                                        else if (RsDiff >= 9) {
                                                            if (diffAFD < 4) {
                                                                if (ssDCval < 106) {
                                                                    return 4;
                                                                }
                                                                else if (ssDCval >= 106) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (diffAFD >= 4) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                            else if (Rs >= 3804) {
                                                if (Cs < 3287) {
                                                    if (mcTcor < 91) {
                                                        if (TSC < 9) {
                                                            return 8;
                                                        }
                                                        else if (TSC >= 9) {
                                                            return 2;
                                                        }
                                                    }
                                                    else if (mcTcor >= 91) {
                                                        if (posBalance < 48) {
                                                            if (Rs < 4539) {
                                                                return 4;
                                                            }
                                                            else if (Rs >= 4539) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (posBalance >= 48) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                                else if (Cs >= 3287) {
                                                    if (posBalance < 59) {
                                                        if (negBalance < 45) {
                                                            return 2;
                                                        }
                                                        else if (negBalance >= 45) {
                                                            if (refDCval < 111) {
                                                                if (Rs < 3929) {
                                                                    return 8;
                                                                }
                                                                else if (Rs >= 3929) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (refDCval >= 111) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                    else if (posBalance >= 59) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                            else if (mu_mv_mag_sq >= 96) {
                                if (refDCval < 94) {
                                    if (tcor < 92) {
                                        if (mu_mv_mag_sq < 102) {
                                            if (tcor < 88) {
                                                return 8;
                                            }
                                            else if (tcor >= 88) {
                                                if (ssDCval < 85) {
                                                    if (Cs < 2486) {
                                                        return 8;
                                                    }
                                                    else if (Cs >= 2486) {
                                                        return 4;
                                                    }
                                                }
                                                else if (ssDCval >= 85) {
                                                    return 2;
                                                }
                                            }
                                        }
                                        else if (mu_mv_mag_sq >= 102) {
                                            if (diffRsCsdiff < 67) {
                                                return 8;
                                            }
                                            else if (diffRsCsdiff >= 67) {
                                                if (Cs < 2539) {
                                                    return 4;
                                                }
                                                else if (Cs >= 2539) {
                                                    return 8;
                                                }
                                            }
                                        }
                                    }
                                    else if (tcor >= 92) {
                                        if (Rs < 3247) {
                                            if (posBalance < 58) {
                                                if (MVDiff < 2) {
                                                    if (diffTSC < -1) {
                                                        if (Rs < 2037) {
                                                            return 4;
                                                        }
                                                        else if (Rs >= 2037) {
                                                            if (SC < 4436) {
                                                                return 8;
                                                            }
                                                            else if (SC >= 4436) {
                                                                return 2;
                                                            }
                                                        }
                                                    }
                                                    else if (diffTSC >= -1) {
                                                        return 4;
                                                    }
                                                }
                                                else if (MVDiff >= 2) {
                                                    if (refDCval < 93) {
                                                        if (TSCindex < 9) {
                                                            return 8;
                                                        }
                                                        else if (TSCindex >= 9) {
                                                            if (Rs < 2823) {
                                                                return 1;
                                                            }
                                                            else if (Rs >= 2823) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                    else if (refDCval >= 93) {
                                                        if (Cs < 2218) {
                                                            return 1;
                                                        }
                                                        else if (Cs >= 2218) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (posBalance >= 58) {
                                                if (Cs < 1921) {
                                                    return 2;
                                                }
                                                else if (Cs >= 1921) {
                                                    return 8;
                                                }
                                            }
                                        }
                                        else if (Rs >= 3247) {
                                            return 2;
                                        }
                                    }
                                }
                                else if (refDCval >= 94) {
                                    if (Rs < 3815) {
                                        if (Rs < 2760) {
                                            if (Rs < 2757) {
                                                if (Rs < 2508) {
                                                    if (Rs < 2248) {
                                                        if (ssDCval < 111) {
                                                            return 4;
                                                        }
                                                        else if (ssDCval >= 111) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (Rs >= 2248) {
                                                        if (refDCval < 116) {
                                                            return 4;
                                                        }
                                                        else if (refDCval >= 116) {
                                                            if (AFD < 8) {
                                                                return 4;
                                                            }
                                                            else if (AFD >= 8) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (Rs >= 2508) {
                                                    return 8;
                                                }
                                            }
                                            else if (Rs >= 2757) {
                                                return 4;
                                            }
                                        }
                                        else if (Rs >= 2760) {
                                            if (refDCval < 99) {
                                                if (negBalance < 36) {
                                                    return 2;
                                                }
                                                else if (negBalance >= 36) {
                                                    return 8;
                                                }
                                            }
                                            else if (refDCval >= 99) {
                                                return 8;
                                            }
                                        }
                                    }
                                    else if (Rs >= 3815) {
                                        if (refDCval < 123) {
                                            if (Cs < 3198) {
                                                return 4;
                                            }
                                            else if (Cs >= 3198) {
                                                if (ssDCval < 113) {
                                                    if (SC < 7566) {
                                                        return 4;
                                                    }
                                                    else if (SC >= 7566) {
                                                        return 8;
                                                    }
                                                }
                                                else if (ssDCval >= 113) {
                                                    if (mu_mv_mag_sq < 121) {
                                                        if (Cs < 3364) {
                                                            return 4;
                                                        }
                                                        else if (Cs >= 3364) {
                                                            return 2;
                                                        }
                                                    }
                                                    else if (mu_mv_mag_sq >= 121) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                        }
                                        else if (refDCval >= 123) {
                                            if (mu_mv_mag_sq < 100) {
                                                if (Cs < 2504) {
                                                    return 8;
                                                }
                                                else if (Cs >= 2504) {
                                                    return 4;
                                                }
                                            }
                                            else if (mu_mv_mag_sq >= 100) {
                                                return 8;
                                            }
                                        }
                                    }
                                }
                            }
                        }
                    }
                    else if (mu_mv_mag_sq >= 128) {
                        if (Rs < 4309) {
                            if (refDCval < 93) {
                                return 2;
                            }
                            else if (refDCval >= 93) {
                                return 8;
                            }
                        }
                        else if (Rs >= 4309) {
                            return 4;
                        }
                    }
                }
                else if (Cs >= 4177) {
                    if (SC < 14344) {
                        if (ssDCval < 106) {
                            if (ssDCval < 81) {
                                if (mu_mv_mag_sq < 127) {
                                    return 8;
                                }
                                else if (mu_mv_mag_sq >= 127) {
                                    if (diffRsCsdiff < 42) {
                                        if (Rs < 4825) {
                                            if (Cs < 4894) {
                                                if (Rs < 4744) {
                                                    return 4;
                                                }
                                                else if (Rs >= 4744) {
                                                    return 8;
                                                }
                                            }
                                            else if (Cs >= 4894) {
                                                return 8;
                                            }
                                        }
                                        else if (Rs >= 4825) {
                                            return 4;
                                        }
                                    }
                                    else if (diffRsCsdiff >= 42) {
                                        if (SC < 9844) {
                                            return 4;
                                        }
                                        else if (SC >= 9844) {
                                            return 8;
                                        }
                                    }
                                }
                            }
                            else if (ssDCval >= 81) {
                                return 8;
                            }
                        }
                        else if (ssDCval >= 106) {
                            if (refDCval < 109) {
                                if (posBalance < 57) {
                                    if (negBalance < 48) {
                                        return 8;
                                    }
                                    else if (negBalance >= 48) {
                                        if (RsDiff < 14) {
                                            return 4;
                                        }
                                        else if (RsDiff >= 14) {
                                            return 8;
                                        }
                                    }
                                }
                                else if (posBalance >= 57) {
                                    if (mu_mv_mag_sq < 123) {
                                        return 8;
                                    }
                                    else if (mu_mv_mag_sq >= 123) {
                                        return 4;
                                    }
                                }
                            }
                            else if (refDCval >= 109) {
                                if (mu_mv_mag_sq < 127) {
                                    if (mcTcor < 83) {
                                        return 4;
                                    }
                                    else if (mcTcor >= 83) {
                                        return 8;
                                    }
                                }
                                else if (mu_mv_mag_sq >= 127) {
                                    return 4;
                                }
                            }
                        }
                    }
                    else if (SC >= 14344) {
                        if (ssDCval < 95) {
                            if (SC < 15437) {
                                if (tcor < 76) {
                                    if (CsDiff < 22) {
                                        return 4;
                                    }
                                    else if (CsDiff >= 22) {
                                        return 2;
                                    }
                                }
                                else if (tcor >= 76) {
                                    return 8;
                                }
                            }
                            else if (SC >= 15437) {
                                if (RsCsDiff < 1036) {
                                    return 2;
                                }
                                else if (RsCsDiff >= 1036) {
                                    if (Cs < 8194) {
                                        return 4;
                                    }
                                    else if (Cs >= 8194) {
                                        return 2;
                                    }
                                }
                            }
                        }
                        else if (ssDCval >= 95) {
                            if (mu_mv_mag_sq < 64) {
                                if (TSC < 25) {
                                    return 8;
                                }
                                else if (TSC >= 25) {
                                    return 4;
                                }
                            }
                            else if (mu_mv_mag_sq >= 64) {
                                return 8;
                            }
                        }
                    }
                }
            }
        }
        else if (MVDiff >= 5) {
            if (SC < 7556) {
                if (tcor < 90) {
                    if (Cs < 1938) {
                        if (AFD < 13) {
                            if (Cs < 272) {
                                if (TSC < 6) {
                                    if (negBalance < 38) {
                                        if (diffAFD < -2) {
                                            if (posBalance < 25) {
                                                if (MVDiff < 401) {
                                                    return 8;
                                                }
                                                else if (MVDiff >= 401) {
                                                    return 2;
                                                }
                                            }
                                            else if (posBalance >= 25) {
                                                if (mcTcor < 20) {
                                                    if (Cs < 1) {
                                                        return 8;
                                                    }
                                                    else if (Cs >= 1) {
                                                        return 1;
                                                    }
                                                }
                                                else if (mcTcor >= 20) {
                                                    return 1;
                                                }
                                            }
                                        }
                                        else if (diffAFD >= -2) {
                                            if (tcor < 66) {
                                                if (ssDCval < 20) {
                                                    if (mu_mv_mag_sq < 25) {
                                                        if (mcTcor < 44) {
                                                            if (refDCval < 19) {
                                                                if (MVDiff < 582) {
                                                                    if (diffMVdiffVal < -683) {
                                                                        return 2;
                                                                    }
                                                                    else if (diffMVdiffVal >= -683) {
                                                                        return 1;
                                                                    }
                                                                }
                                                                else if (MVDiff >= 582) {
                                                                    if (MVDiff < 1221) {
                                                                        return 4;
                                                                    }
                                                                    else if (MVDiff >= 1221) {
                                                                        return 2;
                                                                    }
                                                                }
                                                            }
                                                            else if (refDCval >= 19) {
                                                                if (refDCval < 20) {
                                                                    return 8;
                                                                }
                                                                else if (refDCval >= 20) {
                                                                    return 2;
                                                                }
                                                            }
                                                        }
                                                        else if (mcTcor >= 44) {
                                                            if (diffAFD < -1) {
                                                                return 8;
                                                            }
                                                            else if (diffAFD >= -1) {
                                                                if (MVDiff < 68) {
                                                                    return 8;
                                                                }
                                                                else if (MVDiff >= 68) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (mu_mv_mag_sq >= 25) {
                                                        if (MVDiff < 62) {
                                                            if (mu_mv_mag_sq < 35) {
                                                                if (posBalance < 27) {
                                                                    if (MVDiff < 61) {
                                                                        return 1;
                                                                    }
                                                                    else if (MVDiff >= 61) {
                                                                        return 8;
                                                                    }
                                                                }
                                                                else if (posBalance >= 27) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (mu_mv_mag_sq >= 35) {
                                                                return 1;
                                                            }
                                                        }
                                                        else if (MVDiff >= 62) {
                                                            if (mcTcor < 51) {
                                                                return 8;
                                                            }
                                                            else if (mcTcor >= 51) {
                                                                if (MVDiff < 66) {
                                                                    return 8;
                                                                }
                                                                else if (MVDiff >= 66) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (ssDCval >= 20) {
                                                    if (refDCval < 28) {
                                                        if (mu_mv_mag_sq < 8) {
                                                            if (TSC < 3) {
                                                                if (posBalance < 82) {
                                                                    return 2;
                                                                }
                                                                else if (posBalance >= 82) {
                                                                    return 1;
                                                                }
                                                            }
                                                            else if (TSC >= 3) {
                                                                return 2;
                                                            }
                                                        }
                                                        else if (mu_mv_mag_sq >= 8) {
                                                            if (posBalance < 99) {
                                                                if (Rs < 3) {
                                                                    if (Rs < 1) {
                                                                        if (posBalance < 85) {
                                                                            if (tcor < 43) {
                                                                                return 4;
                                                                            }
                                                                            else if (tcor >= 43) {
                                                                                if (tcor < 60) {
                                                                                    return 2;
                                                                                }
                                                                                else if (tcor >= 60) {
                                                                                    return 4;
                                                                                }
                                                                            }
                                                                        }
                                                                        else if (posBalance >= 85) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                    else if (Rs >= 1) {
                                                                        return 2;
                                                                    }
                                                                }
                                                                else if (Rs >= 3) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (posBalance >= 99) {
                                                                if (TSC < 4) {
                                                                    return 1;
                                                                }
                                                                else if (TSC >= 4) {
                                                                    return 2;
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (refDCval >= 28) {
                                                        return 1;
                                                    }
                                                }
                                            }
                                            else if (tcor >= 66) {
                                                if (mcTcor < 70) {
                                                    if (mcTcor < 47) {
                                                        if (diffAFD < -1) {
                                                            return 8;
                                                        }
                                                        else if (diffAFD >= -1) {
                                                            if (tcor < 77) {
                                                                return 4;
                                                            }
                                                            else if (tcor >= 77) {
                                                                return 2;
                                                            }
                                                        }
                                                    }
                                                    else if (mcTcor >= 47) {
                                                        return 8;
                                                    }
                                                }
                                                else if (mcTcor >= 70) {
                                                    if (posBalance < 79) {
                                                        if (Cs < 33) {
                                                            if (negBalance < 11) {
                                                                if (tcor < 74) {
                                                                    return 2;
                                                                }
                                                                else if (tcor >= 74) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (negBalance >= 11) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (Cs >= 33) {
                                                            if (MVDiff < 17) {
                                                                return 4;
                                                            }
                                                            else if (MVDiff >= 17) {
                                                                return 2;
                                                            }
                                                        }
                                                    }
                                                    else if (posBalance >= 79) {
                                                        if (mu_mv_mag_sq < 3) {
                                                            if (diffTSC < 1) {
                                                                if (MVDiff < 18) {
                                                                    if (ssDCval < 38) {
                                                                        return 4;
                                                                    }
                                                                    else if (ssDCval >= 38) {
                                                                        return 2;
                                                                    }
                                                                }
                                                                else if (MVDiff >= 18) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (diffTSC >= 1) {
                                                                return 2;
                                                            }
                                                        }
                                                        else if (mu_mv_mag_sq >= 3) {
                                                            if (diffMVdiffVal < 5) {
                                                                if (tcor < 81) {
                                                                    if (MVDiff < 101) {
                                                                        if (posBalance < 99) {
                                                                            if (AFD < 5) {
                                                                                return 4;
                                                                            }
                                                                            else if (AFD >= 5) {
                                                                                return 8;
                                                                            }
                                                                        }
                                                                        else if (posBalance >= 99) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                    else if (MVDiff >= 101) {
                                                                        return 8;
                                                                    }
                                                                }
                                                                else if (tcor >= 81) {
                                                                    if (Rs < 274) {
                                                                        return 8;
                                                                    }
                                                                    else if (Rs >= 274) {
                                                                        return 4;
                                                                    }
                                                                }
                                                            }
                                                            else if (diffMVdiffVal >= 5) {
                                                                if (SC < 485) {
                                                                    if (MVDiff < 85) {
                                                                        return 2;
                                                                    }
                                                                    else if (MVDiff >= 85) {
                                                                        return 8;
                                                                    }
                                                                }
                                                                else if (SC >= 485) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (negBalance >= 38) {
                                        if (AFD < 9) {
                                            if (RsDiff < 2) {
                                                if (posBalance < 62) {
                                                    if (MVDiff < 48) {
                                                        return 2;
                                                    }
                                                    else if (MVDiff >= 48) {
                                                        if (diffMVdiffVal < 166) {
                                                            return 1;
                                                        }
                                                        else if (diffMVdiffVal >= 166) {
                                                            if (TSC < 1) {
                                                                return 4;
                                                            }
                                                            else if (TSC >= 1) {
                                                                return 2;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (posBalance >= 62) {
                                                    return 1;
                                                }
                                            }
                                            else if (RsDiff >= 2) {
                                                return 1;
                                            }
                                        }
                                        else if (AFD >= 9) {
                                            if (MVDiff < 182) {
                                                return 2;
                                            }
                                            else if (MVDiff >= 182) {
                                                return 1;
                                            }
                                        }
                                    }
                                }
                                else if (TSC >= 6) {
                                    if (refDCval < 109) {
                                        if (ssDCval < 75) {
                                            if (Rs < 25) {
                                                return 2;
                                            }
                                            else if (Rs >= 25) {
                                                if (Rs < 354) {
                                                    return 1;
                                                }
                                                else if (Rs >= 354) {
                                                    return 2;
                                                }
                                            }
                                        }
                                        else if (ssDCval >= 75) {
                                            return 1;
                                        }
                                    }
                                    else if (refDCval >= 109) {
                                        if (Rs < 492) {
                                            return 4;
                                        }
                                        else if (Rs >= 492) {
                                            return 2;
                                        }
                                    }
                                }
                            }
                            else if (Cs >= 272) {
                                if (RsDiff < 2) {
                                    if (Cs < 384) {
                                        if (refDCval < 68) {
                                            if (SC < 727) {
                                                if (negBalance < 44) {
                                                    if (posBalance < 56) {
                                                        if (RsDiff < 1) {
                                                            return 4;
                                                        }
                                                        else if (RsDiff >= 1) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (posBalance >= 56) {
                                                        return 4;
                                                    }
                                                }
                                                else if (negBalance >= 44) {
                                                    if (SC < 636) {
                                                        return 4;
                                                    }
                                                    else if (SC >= 636) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                            else if (SC >= 727) {
                                                if (tcor < 84) {
                                                    if (Cs < 376) {
                                                        if (posBalance < 56) {
                                                            if (CsDiff < 1) {
                                                                return 1;
                                                            }
                                                            else if (CsDiff >= 1) {
                                                                if (Cs < 327) {
                                                                    return 1;
                                                                }
                                                                else if (Cs >= 327) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                        else if (posBalance >= 56) {
                                                            if (diffRsCsdiff < 1) {
                                                                return 1;
                                                            }
                                                            else if (diffRsCsdiff >= 1) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                    else if (Cs >= 376) {
                                                        if (Rs < 470) {
                                                            if (ssDCval < 38) {
                                                                if (Cs < 382) {
                                                                    return 4;
                                                                }
                                                                else if (Cs >= 382) {
                                                                    if (Rs < 438) {
                                                                        return 4;
                                                                    }
                                                                    else if (Rs >= 438) {
                                                                        return 2;
                                                                    }
                                                                }
                                                            }
                                                            else if (ssDCval >= 38) {
                                                                return 2;
                                                            }
                                                        }
                                                        else if (Rs >= 470) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                                else if (tcor >= 84) {
                                                    return 4;
                                                }
                                            }
                                        }
                                        else if (refDCval >= 68) {
                                            if (ssDCval < 88) {
                                                return 2;
                                            }
                                            else if (ssDCval >= 88) {
                                                return 4;
                                            }
                                        }
                                    }
                                    else if (Cs >= 384) {
                                        if (SC < 1057) {
                                            if (tcor < 85) {
                                                if (diffMVdiffVal < -9) {
                                                    if (MVDiff < 23) {
                                                        return 4;
                                                    }
                                                    else if (MVDiff >= 23) {
                                                        if (Cs < 428) {
                                                            return 8;
                                                        }
                                                        else if (Cs >= 428) {
                                                            return 1;
                                                        }
                                                    }
                                                }
                                                else if (diffMVdiffVal >= -9) {
                                                    if (Rs < 356) {
                                                        return 2;
                                                    }
                                                    else if (Rs >= 356) {
                                                        if (mu_mv_mag_sq < 70) {
                                                            if (ssDCval < 39) {
                                                                if (Cs < 404) {
                                                                    if (negBalance < 44) {
                                                                        return 2;
                                                                    }
                                                                    else if (negBalance >= 44) {
                                                                        if (ssDCval < 38) {
                                                                            if (Rs < 472) {
                                                                                return 4;
                                                                            }
                                                                            else if (Rs >= 472) {
                                                                                return 2;
                                                                            }
                                                                        }
                                                                        else if (ssDCval >= 38) {
                                                                            return 2;
                                                                        }
                                                                    }
                                                                }
                                                                else if (Cs >= 404) {
                                                                    if (mu_mv_mag_sq < 68) {
                                                                        return 2;
                                                                    }
                                                                    else if (mu_mv_mag_sq >= 68) {
                                                                        if (Cs < 417) {
                                                                            return 2;
                                                                        }
                                                                        else if (Cs >= 417) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                            else if (ssDCval >= 39) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (mu_mv_mag_sq >= 70) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (tcor >= 85) {
                                                if (refDCval < 39) {
                                                    return 2;
                                                }
                                                else if (refDCval >= 39) {
                                                    if (Rs < 339) {
                                                        return 8;
                                                    }
                                                    else if (Rs >= 339) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                        }
                                        else if (SC >= 1057) {
                                            if (Cs < 609) {
                                                if (posBalance < 59) {
                                                    if (negBalance < 47) {
                                                        return 4;
                                                    }
                                                    else if (negBalance >= 47) {
                                                        if (Rs < 759) {
                                                            if (diffAFD < -1) {
                                                                return 1;
                                                            }
                                                            else if (diffAFD >= -1) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (Rs >= 759) {
                                                            if (mu_mv_mag_sq < 48) {
                                                                return 1;
                                                            }
                                                            else if (mu_mv_mag_sq >= 48) {
                                                                if (Cs < 575) {
                                                                    return 1;
                                                                }
                                                                else if (Cs >= 575) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (posBalance >= 59) {
                                                    return 1;
                                                }
                                            }
                                            else if (Cs >= 609) {
                                                if (Scindex < 4) {
                                                    if (mu_mv_mag_sq < 57) {
                                                        if (MVDiff < 29) {
                                                            return 2;
                                                        }
                                                        else if (MVDiff >= 29) {
                                                            if (Rs < 692) {
                                                                return 1;
                                                            }
                                                            else if (Rs >= 692) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                    else if (mu_mv_mag_sq >= 57) {
                                                        if (ssDCval < 153) {
                                                            return 8;
                                                        }
                                                        else if (ssDCval >= 153) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                                else if (Scindex >= 4) {
                                                    if (SC < 2975) {
                                                        if (posBalance < 54) {
                                                            return 8;
                                                        }
                                                        else if (posBalance >= 54) {
                                                            return 2;
                                                        }
                                                    }
                                                    else if (SC >= 2975) {
                                                        return 1;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (RsDiff >= 2) {
                                    if (mu_mv_mag_sq < 3) {
                                        if (Rs < 629) {
                                            if (refDCval < 93) {
                                                if (SC < 1290) {
                                                    return 4;
                                                }
                                                else if (SC >= 1290) {
                                                    return 2;
                                                }
                                            }
                                            else if (refDCval >= 93) {
                                                if (posBalance < 60) {
                                                    if (MVDiff < 10) {
                                                        return 4;
                                                    }
                                                    else if (MVDiff >= 10) {
                                                        return 8;
                                                    }
                                                }
                                                else if (posBalance >= 60) {
                                                    if (AFD < 10) {
                                                        return 8;
                                                    }
                                                    else if (AFD >= 10) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                        }
                                        else if (Rs >= 629) {
                                            if (posBalance < 59) {
                                                if (ssDCval < 128) {
                                                    if (Rs < 1818) {
                                                        if (tcor < 77) {
                                                            if (Rs < 1132) {
                                                                return 4;
                                                            }
                                                            else if (Rs >= 1132) {
                                                                if (AFD < 11) {
                                                                    return 8;
                                                                }
                                                                else if (AFD >= 11) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                        else if (tcor >= 77) {
                                                            if (MVDiff < 12) {
                                                                if (mcTcor < 92) {
                                                                    if (diffMVdiffVal < 0) {
                                                                        if (CsDiff < 3) {
                                                                            return 4;
                                                                        }
                                                                        else if (CsDiff >= 3) {
                                                                            if (Rs < 1352) {
                                                                                if (Rs < 1257) {
                                                                                    return 1;
                                                                                }
                                                                                else if (Rs >= 1257) {
                                                                                    return 4;
                                                                                }
                                                                            }
                                                                            else if (Rs >= 1352) {
                                                                                return 2;
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (diffMVdiffVal >= 0) {
                                                                        if (tcor < 82) {
                                                                            if (Rs < 1177) {
                                                                                return 4;
                                                                            }
                                                                            else if (Rs >= 1177) {
                                                                                if (Rs < 1468) {
                                                                                    return 2;
                                                                                }
                                                                                else if (Rs >= 1468) {
                                                                                    return 8;
                                                                                }
                                                                            }
                                                                        }
                                                                        else if (tcor >= 82) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                }
                                                                else if (mcTcor >= 92) {
                                                                    return 1;
                                                                }
                                                            }
                                                            else if (MVDiff >= 12) {
                                                                if (RsCsDiff < 16) {
                                                                    if (Cs < 971) {
                                                                        return 1;
                                                                    }
                                                                    else if (Cs >= 971) {
                                                                        return 8;
                                                                    }
                                                                }
                                                                else if (RsCsDiff >= 16) {
                                                                    if (diffAFD < -2) {
                                                                        if (CsDiff < 5) {
                                                                            if (Cs < 802) {
                                                                                return 2;
                                                                            }
                                                                            else if (Cs >= 802) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                        else if (CsDiff >= 5) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                    else if (diffAFD >= -2) {
                                                                        return 2;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (Rs >= 1818) {
                                                        if (mcTcor < 87) {
                                                            if (ssDCval < 107) {
                                                                return 2;
                                                            }
                                                            else if (ssDCval >= 107) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (mcTcor >= 87) {
                                                            if (diffAFD < 1) {
                                                                return 4;
                                                            }
                                                            else if (diffAFD >= 1) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (ssDCval >= 128) {
                                                    if (ssDCval < 139) {
                                                        return 1;
                                                    }
                                                    else if (ssDCval >= 139) {
                                                        if (Cs < 1128) {
                                                            return 4;
                                                        }
                                                        else if (Cs >= 1128) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (posBalance >= 59) {
                                                if (negBalance < 48) {
                                                    if (negBalance < 39) {
                                                        if (RsDiff < 7) {
                                                            return 4;
                                                        }
                                                        else if (RsDiff >= 7) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (negBalance >= 39) {
                                                        return 1;
                                                    }
                                                }
                                                else if (negBalance >= 48) {
                                                    if (refDCval < 63) {
                                                        if (SC < 1965) {
                                                            return 8;
                                                        }
                                                        else if (SC >= 1965) {
                                                            return 1;
                                                        }
                                                    }
                                                    else if (refDCval >= 63) {
                                                        if (AFD < 10) {
                                                            if (diffMVdiffVal < -3) {
                                                                if (SC < 1984) {
                                                                    if (SC < 1126) {
                                                                        return 8;
                                                                    }
                                                                    else if (SC >= 1126) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (SC >= 1984) {
                                                                    if (MVDiff < 6) {
                                                                        return 4;
                                                                    }
                                                                    else if (MVDiff >= 6) {
                                                                        return 2;
                                                                    }
                                                                }
                                                            }
                                                            else if (diffMVdiffVal >= -3) {
                                                                if (refDCval < 97) {
                                                                    if (ssDCval < 93) {
                                                                        return 4;
                                                                    }
                                                                    else if (ssDCval >= 93) {
                                                                        if (diffMVdiffVal < 6) {
                                                                            if (Rs < 650) {
                                                                                return 4;
                                                                            }
                                                                            else if (Rs >= 650) {
                                                                                return 8;
                                                                            }
                                                                        }
                                                                        else if (diffMVdiffVal >= 6) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                }
                                                                else if (refDCval >= 97) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                        else if (AFD >= 10) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (mu_mv_mag_sq >= 3) {
                                        if (mu_mv_mag_sq < 41) {
                                            if (RsCsDiff < 16) {
                                                if (Rs < 646) {
                                                    if (ssDCval < 67) {
                                                        if (Rs < 401) {
                                                            return 1;
                                                        }
                                                        else if (Rs >= 401) {
                                                            return 2;
                                                        }
                                                    }
                                                    else if (ssDCval >= 67) {
                                                        if (SC < 718) {
                                                            return 4;
                                                        }
                                                        else if (SC >= 718) {
                                                            if (SC < 1136) {
                                                                return 8;
                                                            }
                                                            else if (SC >= 1136) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (Rs >= 646) {
                                                    if (mu_mv_mag_sq < 29) {
                                                        if (Cs < 392) {
                                                            return 1;
                                                        }
                                                        else if (Cs >= 392) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (mu_mv_mag_sq >= 29) {
                                                        if (gchDC < 4) {
                                                            if (diffRsCsdiff < 6) {
                                                                if (Cs < 670) {
                                                                    if (Cs < 659) {
                                                                        if (Rs < 900) {
                                                                            if (MVDiff < 57) {
                                                                                if (posBalance < 59) {
                                                                                    if (negBalance < 49) {
                                                                                        return 4;
                                                                                    }
                                                                                    else if (negBalance >= 49) {
                                                                                        if (Rs < 770) {
                                                                                            if (SC < 1305) {
                                                                                                return 1;
                                                                                            }
                                                                                            else if (SC >= 1305) {
                                                                                                return 4;
                                                                                            }
                                                                                        }
                                                                                        else if (Rs >= 770) {
                                                                                            return 1;
                                                                                        }
                                                                                    }
                                                                                }
                                                                                else if (posBalance >= 59) {
                                                                                    return 1;
                                                                                }
                                                                            }
                                                                            else if (MVDiff >= 57) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                        else if (Rs >= 900) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                    else if (Cs >= 659) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (Cs >= 670) {
                                                                    if (ssDCval < 51) {
                                                                        return 1;
                                                                    }
                                                                    else if (ssDCval >= 51) {
                                                                        return 8;
                                                                    }
                                                                }
                                                            }
                                                            else if (diffRsCsdiff >= 6) {
                                                                if (posBalance < 60) {
                                                                    if (Cs < 506) {
                                                                        return 4;
                                                                    }
                                                                    else if (Cs >= 506) {
                                                                        return 1;
                                                                    }
                                                                }
                                                                else if (posBalance >= 60) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                        else if (gchDC >= 4) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (RsCsDiff >= 16) {
                                                if (SC < 2166) {
                                                    if (Rs < 636) {
                                                        return 2;
                                                    }
                                                    else if (Rs >= 636) {
                                                        if (tcor < 85) {
                                                            if (SC < 1932) {
                                                                if (posBalance < 60) {
                                                                    if (mu_mv_mag_sq < 28) {
                                                                        if (Cs < 757) {
                                                                            return 1;
                                                                        }
                                                                        else if (Cs >= 757) {
                                                                            return 2;
                                                                        }
                                                                    }
                                                                    else if (mu_mv_mag_sq >= 28) {
                                                                        if (Cs < 815) {
                                                                            return 2;
                                                                        }
                                                                        else if (Cs >= 815) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                }
                                                                else if (posBalance >= 60) {
                                                                    if (tcor < 71) {
                                                                        return 2;
                                                                    }
                                                                    else if (tcor >= 71) {
                                                                        if (SC < 1900) {
                                                                            if (Rs < 1139) {
                                                                                return 8;
                                                                            }
                                                                            else if (Rs >= 1139) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                        else if (SC >= 1900) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                            else if (SC >= 1932) {
                                                                if (ssDCval < 83) {
                                                                    if (refDCval < 57) {
                                                                        return 4;
                                                                    }
                                                                    else if (refDCval >= 57) {
                                                                        return 1;
                                                                    }
                                                                }
                                                                else if (ssDCval >= 83) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                        else if (tcor >= 85) {
                                                            if (Cs < 705) {
                                                                return 2;
                                                            }
                                                            else if (Cs >= 705) {
                                                                if (Rs < 1199) {
                                                                    return 4;
                                                                }
                                                                else if (Rs >= 1199) {
                                                                    if (Rs < 1263) {
                                                                        return 2;
                                                                    }
                                                                    else if (Rs >= 1263) {
                                                                        return 1;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (SC >= 2166) {
                                                    if (refDCval < 114) {
                                                        if (ssDCval < 81) {
                                                            if (gchDC < 1) {
                                                                if (refDCval < 76) {
                                                                    return 1;
                                                                }
                                                                else if (refDCval >= 76) {
                                                                    if (Cs < 1331) {
                                                                        return 2;
                                                                    }
                                                                    else if (Cs >= 1331) {
                                                                        return 1;
                                                                    }
                                                                }
                                                            }
                                                            else if (gchDC >= 1) {
                                                                if (tcor < 74) {
                                                                    return 8;
                                                                }
                                                                else if (tcor >= 74) {
                                                                    if (mu_mv_mag_sq < 36) {
                                                                        return 1;
                                                                    }
                                                                    else if (mu_mv_mag_sq >= 36) {
                                                                        return 4;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (ssDCval >= 81) {
                                                            if (Rs < 1639) {
                                                                if (Rs < 1516) {
                                                                    return 4;
                                                                }
                                                                else if (Rs >= 1516) {
                                                                    return 2;
                                                                }
                                                            }
                                                            else if (Rs >= 1639) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                    else if (refDCval >= 114) {
                                                        if (Cs < 1496) {
                                                            if (Rs < 1237) {
                                                                if (negBalance < 32) {
                                                                    if (RsCsDiff < 80) {
                                                                        return 1;
                                                                    }
                                                                    else if (RsCsDiff >= 80) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (negBalance >= 32) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (Rs >= 1237) {
                                                                return 1;
                                                            }
                                                        }
                                                        else if (Cs >= 1496) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (mu_mv_mag_sq >= 41) {
                                            if (TSC < 7) {
                                                if (negBalance < 15) {
                                                    if (CsDiff < 8) {
                                                        if (RsCsDiff < 9) {
                                                            if (Cs < 667) {
                                                                return 1;
                                                            }
                                                            else if (Cs >= 667) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (RsCsDiff >= 9) {
                                                            if (AFD < 3) {
                                                                if (MVDiff < 7) {
                                                                    return 1;
                                                                }
                                                                else if (MVDiff >= 7) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (AFD >= 3) {
                                                                if (tcor < 52) {
                                                                    return 8;
                                                                }
                                                                else if (tcor >= 52) {
                                                                    if (tcor < 74) {
                                                                        return 4;
                                                                    }
                                                                    else if (tcor >= 74) {
                                                                        if (MVDiff < 13) {
                                                                            return 8;
                                                                        }
                                                                        else if (MVDiff >= 13) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (CsDiff >= 8) {
                                                        return 1;
                                                    }
                                                }
                                                else if (negBalance >= 15) {
                                                    if (mu_mv_mag_sq < 83) {
                                                        if (MVDiff < 34) {
                                                            if (Rs < 1185) {
                                                                if (mu_mv_mag_sq < 56) {
                                                                    if (ssDCval < 42) {
                                                                        if (posBalance < 56) {
                                                                            if (Rs < 735) {
                                                                                return 2;
                                                                            }
                                                                            else if (Rs >= 735) {
                                                                                return 8;
                                                                            }
                                                                        }
                                                                        else if (posBalance >= 56) {
                                                                            if (Cs < 572) {
                                                                                return 4;
                                                                            }
                                                                            else if (Cs >= 572) {
                                                                                if (Rs < 943) {
                                                                                    if (MVDiff < 28) {
                                                                                        if (Cs < 600) {
                                                                                            return 1;
                                                                                        }
                                                                                        else if (Cs >= 600) {
                                                                                            return 4;
                                                                                        }
                                                                                    }
                                                                                    else if (MVDiff >= 28) {
                                                                                        return 1;
                                                                                    }
                                                                                }
                                                                                else if (Rs >= 943) {
                                                                                    return 2;
                                                                                }
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (ssDCval >= 42) {
                                                                        if (diffAFD < -2) {
                                                                            if (Rs < 682) {
                                                                                return 4;
                                                                            }
                                                                            else if (Rs >= 682) {
                                                                                return 2;
                                                                            }
                                                                        }
                                                                        else if (diffAFD >= -2) {
                                                                            if (MVDiff < 18) {
                                                                                return 4;
                                                                            }
                                                                            else if (MVDiff >= 18) {
                                                                                return 8;
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                                else if (mu_mv_mag_sq >= 56) {
                                                                    if (tcor < 89) {
                                                                        return 4;
                                                                    }
                                                                    else if (tcor >= 89) {
                                                                        if (Cs < 851) {
                                                                            return 4;
                                                                        }
                                                                        else if (Cs >= 851) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                            else if (Rs >= 1185) {
                                                                if (ssDCval < 53) {
                                                                    if (Rs < 1366) {
                                                                        if (Cs < 832) {
                                                                            if (SC < 2018) {
                                                                                return 1;
                                                                            }
                                                                            else if (SC >= 2018) {
                                                                                return 8;
                                                                            }
                                                                        }
                                                                        else if (Cs >= 832) {
                                                                            if (mu_mv_mag_sq < 69) {
                                                                                return 1;
                                                                            }
                                                                            else if (mu_mv_mag_sq >= 69) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (Rs >= 1366) {
                                                                        if (Cs < 1561) {
                                                                            if (Rs < 1735) {
                                                                                if (MVDiff < 7) {
                                                                                    return 8;
                                                                                }
                                                                                else if (MVDiff >= 7) {
                                                                                    if (Cs < 1120) {
                                                                                        return 4;
                                                                                    }
                                                                                    else if (Cs >= 1120) {
                                                                                        return 1;
                                                                                    }
                                                                                }
                                                                            }
                                                                            else if (Rs >= 1735) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                        else if (Cs >= 1561) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                }
                                                                else if (ssDCval >= 53) {
                                                                    if (RsDiff < 4) {
                                                                        if (Rs < 1529) {
                                                                            if (Rs < 1452) {
                                                                                if (Cs < 1037) {
                                                                                    return 1;
                                                                                }
                                                                                else if (Cs >= 1037) {
                                                                                    return 8;
                                                                                }
                                                                            }
                                                                            else if (Rs >= 1452) {
                                                                                return 1;
                                                                            }
                                                                        }
                                                                        else if (Rs >= 1529) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                    else if (RsDiff >= 4) {
                                                                        if (diffAFD < -3) {
                                                                            if (Rs < 1789) {
                                                                                return 8;
                                                                            }
                                                                            else if (Rs >= 1789) {
                                                                                return 2;
                                                                            }
                                                                        }
                                                                        else if (diffAFD >= -3) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (MVDiff >= 34) {
                                                            if (mu_mv_mag_sq < 63) {
                                                                if (tcor < 76) {
                                                                    return 4;
                                                                }
                                                                else if (tcor >= 76) {
                                                                    return 2;
                                                                }
                                                            }
                                                            else if (mu_mv_mag_sq >= 63) {
                                                                if (Rs < 1261) {
                                                                    return 4;
                                                                }
                                                                else if (Rs >= 1261) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (mu_mv_mag_sq >= 83) {
                                                        if (ssDCval < 94) {
                                                            return 8;
                                                        }
                                                        else if (ssDCval >= 94) {
                                                            if (ssDCval < 141) {
                                                                if (tcor < 78) {
                                                                    return 8;
                                                                }
                                                                else if (tcor >= 78) {
                                                                    return 1;
                                                                }
                                                            }
                                                            else if (ssDCval >= 141) {
                                                                if (Rs < 804) {
                                                                    return 8;
                                                                }
                                                                else if (Rs >= 804) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                            else if (TSC >= 7) {
                                                if (RsDiff < 7) {
                                                    if (SC < 3313) {
                                                        if (ssDCval < 103) {
                                                            if (Cs < 1303) {
                                                                if (AFD < 11) {
                                                                    if (Rs < 1135) {
                                                                        return 1;
                                                                    }
                                                                    else if (Rs >= 1135) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (AFD >= 11) {
                                                                    if (posBalance < 61) {
                                                                        return 8;
                                                                    }
                                                                    else if (posBalance >= 61) {
                                                                        return 4;
                                                                    }
                                                                }
                                                            }
                                                            else if (Cs >= 1303) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (ssDCval >= 103) {
                                                            if (Rs < 1376) {
                                                                if (Rs < 340) {
                                                                    return 4;
                                                                }
                                                                else if (Rs >= 340) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (Rs >= 1376) {
                                                                if (AFD < 11) {
                                                                    return 2;
                                                                }
                                                                else if (AFD >= 11) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (SC >= 3313) {
                                                        if (mu_mv_mag_sq < 64) {
                                                            if (ssDCval < 107) {
                                                                if (RsDiff < 6) {
                                                                    if (mcTcor < 82) {
                                                                        return 8;
                                                                    }
                                                                    else if (mcTcor >= 82) {
                                                                        if (mu_mv_mag_sq < 60) {
                                                                            if (Cs < 1665) {
                                                                                return 2;
                                                                            }
                                                                            else if (Cs >= 1665) {
                                                                                return 8;
                                                                            }
                                                                        }
                                                                        else if (mu_mv_mag_sq >= 60) {
                                                                            return 2;
                                                                        }
                                                                    }
                                                                }
                                                                else if (RsDiff >= 6) {
                                                                    return 2;
                                                                }
                                                            }
                                                            else if (ssDCval >= 107) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (mu_mv_mag_sq >= 64) {
                                                            if (SC < 3980) {
                                                                return 8;
                                                            }
                                                            else if (SC >= 3980) {
                                                                if (mcTcor < 89) {
                                                                    if (negBalance < 53) {
                                                                        return 8;
                                                                    }
                                                                    else if (negBalance >= 53) {
                                                                        return 2;
                                                                    }
                                                                }
                                                                else if (mcTcor >= 89) {
                                                                    return 2;
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (RsDiff >= 7) {
                                                    if (Rs < 2918) {
                                                        if (mu_mv_mag_sq < 77) {
                                                            if (negBalance < 47) {
                                                                return 1;
                                                            }
                                                            else if (negBalance >= 47) {
                                                                if (ssDCval < 103) {
                                                                    if (SC < 3702) {
                                                                        return 8;
                                                                    }
                                                                    else if (SC >= 3702) {
                                                                        if (Rs < 2105) {
                                                                            return 2;
                                                                        }
                                                                        else if (Rs >= 2105) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                }
                                                                else if (ssDCval >= 103) {
                                                                    if (AFD < 11) {
                                                                        return 4;
                                                                    }
                                                                    else if (AFD >= 11) {
                                                                        if (Rs < 2659) {
                                                                            return 4;
                                                                        }
                                                                        else if (Rs >= 2659) {
                                                                            if (Rs < 2854) {
                                                                                return 8;
                                                                            }
                                                                            else if (Rs >= 2854) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (mu_mv_mag_sq >= 77) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (Rs >= 2918) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                        }
                        else if (AFD >= 13) {
                            if (ssDCval < 76) {
                                if (TSCindex < 9) {
                                    if (Rs < 929) {
                                        if (posBalance < 52) {
                                            if (Rs < 465) {
                                                return 4;
                                            }
                                            else if (Rs >= 465) {
                                                return 2;
                                            }
                                        }
                                        else if (posBalance >= 52) {
                                            return 2;
                                        }
                                    }
                                    else if (Rs >= 929) {
                                        if (mcTcor < 89) {
                                            if (refDCval < 50) {
                                                return 2;
                                            }
                                            else if (refDCval >= 50) {
                                                return 1;
                                            }
                                        }
                                        else if (mcTcor >= 89) {
                                            if (SC < 2060) {
                                                return 1;
                                            }
                                            else if (SC >= 2060) {
                                                return 2;
                                            }
                                        }
                                    }
                                }
                                else if (TSCindex >= 9) {
                                    if (MVDiff < 41) {
                                        if (RsCsDiff < 92) {
                                            if (Rs < 1092) {
                                                if (Cs < 722) {
                                                    return 2;
                                                }
                                                else if (Cs >= 722) {
                                                    return 1;
                                                }
                                            }
                                            else if (Rs >= 1092) {
                                                if (MVDiff < 21) {
                                                    if (mcTcor < 86) {
                                                        return 8;
                                                    }
                                                    else if (mcTcor >= 86) {
                                                        if (RsCsDiff < 67) {
                                                            return 8;
                                                        }
                                                        else if (RsCsDiff >= 67) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                                else if (MVDiff >= 21) {
                                                    return 8;
                                                }
                                            }
                                        }
                                        else if (RsCsDiff >= 92) {
                                            if (Cs < 1798) {
                                                return 4;
                                            }
                                            else if (Cs >= 1798) {
                                                return 8;
                                            }
                                        }
                                    }
                                    else if (MVDiff >= 41) {
                                        if (diffRsCsdiff < -5) {
                                            if (AFD < 17) {
                                                return 8;
                                            }
                                            else if (AFD >= 17) {
                                                return 1;
                                            }
                                        }
                                        else if (diffRsCsdiff >= -5) {
                                            if (TSC < 18) {
                                                if (posBalance < 55) {
                                                    if (CsDiff < 7) {
                                                        if (Rs < 572) {
                                                            return 2;
                                                        }
                                                        else if (Rs >= 572) {
                                                            if (RsDiff < 3) {
                                                                return 4;
                                                            }
                                                            else if (RsDiff >= 3) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                    else if (CsDiff >= 7) {
                                                        return 2;
                                                    }
                                                }
                                                else if (posBalance >= 55) {
                                                    if (diffTSC < 1) {
                                                        return 4;
                                                    }
                                                    else if (diffTSC >= 1) {
                                                        if (AFD < 22) {
                                                            if (gchDC < 7) {
                                                                return 1;
                                                            }
                                                            else if (gchDC >= 7) {
                                                                if (refDCval < 40) {
                                                                    return 1;
                                                                }
                                                                else if (refDCval >= 40) {
                                                                    return 2;
                                                                }
                                                            }
                                                        }
                                                        else if (AFD >= 22) {
                                                            if (TSC < 14) {
                                                                return 4;
                                                            }
                                                            else if (TSC >= 14) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                            else if (TSC >= 18) {
                                                if (Rs < 258) {
                                                    return 8;
                                                }
                                                else if (Rs >= 258) {
                                                    return 4;
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                            else if (ssDCval >= 76) {
                                if (Rs < 3027) {
                                    if (diffAFD < 4) {
                                        if (tcor < 65) {
                                            if (RsDiff < 17) {
                                                if (AFD < 32) {
                                                    if (CsDiff < 6) {
                                                        if (refDCval < 105) {
                                                            if (ssDCval < 92) {
                                                                if (Cs < 403) {
                                                                    return 1;
                                                                }
                                                                else if (Cs >= 403) {
                                                                    return 2;
                                                                }
                                                            }
                                                            else if (ssDCval >= 92) {
                                                                return 1;
                                                            }
                                                        }
                                                        else if (refDCval >= 105) {
                                                            if (mu_mv_mag_sq < 23) {
                                                                return 2;
                                                            }
                                                            else if (mu_mv_mag_sq >= 23) {
                                                                if (Rs < 515) {
                                                                    return 1;
                                                                }
                                                                else if (Rs >= 515) {
                                                                    if (Cs < 584) {
                                                                        return 2;
                                                                    }
                                                                    else if (Cs >= 584) {
                                                                        return 8;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (CsDiff >= 6) {
                                                        if (ssDCval < 142) {
                                                            if (RsCsDiff < 279) {
                                                                if (RsDiff < 12) {
                                                                    return 1;
                                                                }
                                                                else if (RsDiff >= 12) {
                                                                    if (AFD < 29) {
                                                                        if (TSC < 10) {
                                                                            if (CsDiff < 11) {
                                                                                return 2;
                                                                            }
                                                                            else if (CsDiff >= 11) {
                                                                                return 1;
                                                                            }
                                                                        }
                                                                        else if (TSC >= 10) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                    else if (AFD >= 29) {
                                                                        if (Cs < 1085) {
                                                                            return 1;
                                                                        }
                                                                        else if (Cs >= 1085) {
                                                                            return 2;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                            else if (RsCsDiff >= 279) {
                                                                if (TSCindex < 9) {
                                                                    if (Cs < 1505) {
                                                                        return 8;
                                                                    }
                                                                    else if (Cs >= 1505) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (TSCindex >= 9) {
                                                                    return 1;
                                                                }
                                                            }
                                                        }
                                                        else if (ssDCval >= 142) {
                                                            if (Rs < 1895) {
                                                                return 8;
                                                            }
                                                            else if (Rs >= 1895) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (AFD >= 32) {
                                                    return 2;
                                                }
                                            }
                                            else if (RsDiff >= 17) {
                                                if (ssDCval < 125) {
                                                    if (Rs < 2820) {
                                                        if (TSC < 15) {
                                                            if (AFD < 36) {
                                                                return 1;
                                                            }
                                                            else if (AFD >= 36) {
                                                                return 2;
                                                            }
                                                        }
                                                        else if (TSC >= 15) {
                                                            return 2;
                                                        }
                                                    }
                                                    else if (Rs >= 2820) {
                                                        return 2;
                                                    }
                                                }
                                                else if (ssDCval >= 125) {
                                                    if (Cs < 1215) {
                                                        return 1;
                                                    }
                                                    else if (Cs >= 1215) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                        }
                                        else if (tcor >= 65) {
                                            if (negBalance < 51) {
                                                if (Cs < 1396) {
                                                    if (RsDiff < 9) {
                                                        if (tcor < 84) {
                                                            if (mcTcor < 89) {
                                                                if (tcor < 69) {
                                                                    return 8;
                                                                }
                                                                else if (tcor >= 69) {
                                                                    return 1;
                                                                }
                                                            }
                                                            else if (mcTcor >= 89) {
                                                                if (mcTcor < 98) {
                                                                    if (Cs < 977) {
                                                                        if (diffMVdiffVal < 15) {
                                                                            if (gchDC < 2) {
                                                                                if (SC < 1743) {
                                                                                    return 4;
                                                                                }
                                                                                else if (SC >= 1743) {
                                                                                    return 2;
                                                                                }
                                                                            }
                                                                            else if (gchDC >= 2) {
                                                                                return 2;
                                                                            }
                                                                        }
                                                                        else if (diffMVdiffVal >= 15) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                    else if (Cs >= 977) {
                                                                        if (Cs < 1302) {
                                                                            if (diffRsCsdiff < 14) {
                                                                                return 4;
                                                                            }
                                                                            else if (diffRsCsdiff >= 14) {
                                                                                return 2;
                                                                            }
                                                                        }
                                                                        else if (Cs >= 1302) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                }
                                                                else if (mcTcor >= 98) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                        else if (tcor >= 84) {
                                                            if (ssDCval < 152) {
                                                                if (Rs < 1265) {
                                                                    return 2;
                                                                }
                                                                else if (Rs >= 1265) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (ssDCval >= 152) {
                                                                return 1;
                                                            }
                                                        }
                                                    }
                                                    else if (RsDiff >= 9) {
                                                        if (Cs < 434) {
                                                            return 8;
                                                        }
                                                        else if (Cs >= 434) {
                                                            if (TSC < 13) {
                                                                if (AFD < 15) {
                                                                    return 1;
                                                                }
                                                                else if (AFD >= 15) {
                                                                    return 2;
                                                                }
                                                            }
                                                            else if (TSC >= 13) {
                                                                return 1;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (Cs >= 1396) {
                                                    if (MVDiff < 21) {
                                                        if (Rs < 2840) {
                                                            if (gchDC < 2) {
                                                                if (mcTcor < 95) {
                                                                    if (mu_mv_mag_sq < 26) {
                                                                        return 1;
                                                                    }
                                                                    else if (mu_mv_mag_sq >= 26) {
                                                                        return 2;
                                                                    }
                                                                }
                                                                else if (mcTcor >= 95) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (gchDC >= 2) {
                                                                return 2;
                                                            }
                                                        }
                                                        else if (Rs >= 2840) {
                                                            return 2;
                                                        }
                                                    }
                                                    else if (MVDiff >= 21) {
                                                        if (refDCval < 100) {
                                                            if (Rs < 1676) {
                                                                return 1;
                                                            }
                                                            else if (Rs >= 1676) {
                                                                if (refDCval < 88) {
                                                                    return 4;
                                                                }
                                                                else if (refDCval >= 88) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                        else if (refDCval >= 100) {
                                                            if (RsCsDiff < 192) {
                                                                if (diffTSC < 1) {
                                                                    if (tcor < 82) {
                                                                        return 4;
                                                                    }
                                                                    else if (tcor >= 82) {
                                                                        if (Rs < 1596) {
                                                                            return 4;
                                                                        }
                                                                        else if (Rs >= 1596) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                }
                                                                else if (diffTSC >= 1) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (RsCsDiff >= 192) {
                                                                if (gchDC < 2) {
                                                                    if (Cs < 1440) {
                                                                        return 4;
                                                                    }
                                                                    else if (Cs >= 1440) {
                                                                        return 2;
                                                                    }
                                                                }
                                                                else if (gchDC >= 2) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                            else if (negBalance >= 51) {
                                                if (Cs < 1751) {
                                                    if (MVDiff < 10) {
                                                        if (CsDiff < 7) {
                                                            if (TSC < 5) {
                                                                return 8;
                                                            }
                                                            else if (TSC >= 5) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (CsDiff >= 7) {
                                                            if (Rs < 2023) {
                                                                return 2;
                                                            }
                                                            else if (Rs >= 2023) {
                                                                return 1;
                                                            }
                                                        }
                                                    }
                                                    else if (MVDiff >= 10) {
                                                        if (refDCval < 117) {
                                                            if (RsCsDiff < 94) {
                                                                if (negBalance < 54) {
                                                                    if (Cs < 758) {
                                                                        return 2;
                                                                    }
                                                                    else if (Cs >= 758) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (negBalance >= 54) {
                                                                    if (MVDiff < 59) {
                                                                        if (AFD < 14) {
                                                                            if (TSC < 7) {
                                                                                return 2;
                                                                            }
                                                                            else if (TSC >= 7) {
                                                                                return 1;
                                                                            }
                                                                        }
                                                                        else if (AFD >= 14) {
                                                                            return 2;
                                                                        }
                                                                    }
                                                                    else if (MVDiff >= 59) {
                                                                        return 1;
                                                                    }
                                                                }
                                                            }
                                                            else if (RsCsDiff >= 94) {
                                                                return 2;
                                                            }
                                                        }
                                                        else if (refDCval >= 117) {
                                                            if (refDCval < 141) {
                                                                if (refDCval < 121) {
                                                                    if (posBalance < 58) {
                                                                        if (SC < 4056) {
                                                                            return 8;
                                                                        }
                                                                        else if (SC >= 4056) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                    else if (posBalance >= 58) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (refDCval >= 121) {
                                                                    if (mcTcor < 88) {
                                                                        if (SC < 3203) {
                                                                            if (Cs < 482) {
                                                                                if (Rs < 1476) {
                                                                                    return 2;
                                                                                }
                                                                                else if (Rs >= 1476) {
                                                                                    return 4;
                                                                                }
                                                                            }
                                                                            else if (Cs >= 482) {
                                                                                return 2;
                                                                            }
                                                                        }
                                                                        else if (SC >= 3203) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                    else if (mcTcor >= 88) {
                                                                        if (diffMVdiffVal < -2) {
                                                                            return 2;
                                                                        }
                                                                        else if (diffMVdiffVal >= -2) {
                                                                            if (Rs < 2070) {
                                                                                return 4;
                                                                            }
                                                                            else if (Rs >= 2070) {
                                                                                if (AFD < 15) {
                                                                                    return 4;
                                                                                }
                                                                                else if (AFD >= 15) {
                                                                                    return 2;
                                                                                }
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                            else if (refDCval >= 141) {
                                                                if (TSC < 13) {
                                                                    return 2;
                                                                }
                                                                else if (TSC >= 13) {
                                                                    return 1;
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (Cs >= 1751) {
                                                    if (RsDiff < 12) {
                                                        if (Rs < 2160) {
                                                            if (MVDiff < 62) {
                                                                if (SC < 3955) {
                                                                    return 2;
                                                                }
                                                                else if (SC >= 3955) {
                                                                    return 1;
                                                                }
                                                            }
                                                            else if (MVDiff >= 62) {
                                                                if (AFD < 18) {
                                                                    return 4;
                                                                }
                                                                else if (AFD >= 18) {
                                                                    return 1;
                                                                }
                                                            }
                                                        }
                                                        else if (Rs >= 2160) {
                                                            if (tcor < 88) {
                                                                return 1;
                                                            }
                                                            else if (tcor >= 88) {
                                                                if (Rs < 2309) {
                                                                    return 1;
                                                                }
                                                                else if (Rs >= 2309) {
                                                                    return 2;
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (RsDiff >= 12) {
                                                        if (mcTcor < 88) {
                                                            if (Rs < 2409) {
                                                                return 2;
                                                            }
                                                            else if (Rs >= 2409) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (mcTcor >= 88) {
                                                            if (ssDCval < 127) {
                                                                if (Rs < 2455) {
                                                                    if (SC < 3808) {
                                                                        return 1;
                                                                    }
                                                                    else if (SC >= 3808) {
                                                                        if (Cs < 1875) {
                                                                            return 4;
                                                                        }
                                                                        else if (Cs >= 1875) {
                                                                            return 2;
                                                                        }
                                                                    }
                                                                }
                                                                else if (Rs >= 2455) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (ssDCval >= 127) {
                                                                return 2;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (diffAFD >= 4) {
                                        if (posBalance < 49) {
                                            if (posBalance < 45) {
                                                if (posBalance < 19) {
                                                    if (AFD < 26) {
                                                        return 8;
                                                    }
                                                    else if (AFD >= 26) {
                                                        return 4;
                                                    }
                                                }
                                                else if (posBalance >= 19) {
                                                    return 1;
                                                }
                                            }
                                            else if (posBalance >= 45) {
                                                if (AFD < 37) {
                                                    return 8;
                                                }
                                                else if (AFD >= 37) {
                                                    return 2;
                                                }
                                            }
                                        }
                                        else if (posBalance >= 49) {
                                            if (SC < 3538) {
                                                if (CsDiff < 6) {
                                                    if (Cs < 780) {
                                                        if (AFD < 54) {
                                                            return 2;
                                                        }
                                                        else if (AFD >= 54) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (Cs >= 780) {
                                                        return 1;
                                                    }
                                                }
                                                else if (CsDiff >= 6) {
                                                    if (negBalance < 42) {
                                                        if (Rs < 1251) {
                                                            if (negBalance < 34) {
                                                                if (ssDCval < 135) {
                                                                    return 4;
                                                                }
                                                                else if (ssDCval >= 135) {
                                                                    if (tcor < 7) {
                                                                        return 4;
                                                                    }
                                                                    else if (tcor >= 7) {
                                                                        return 1;
                                                                    }
                                                                }
                                                            }
                                                            else if (negBalance >= 34) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (Rs >= 1251) {
                                                            if (ssDCval < 129) {
                                                                return 2;
                                                            }
                                                            else if (ssDCval >= 129) {
                                                                if (Rs < 1318) {
                                                                    return 8;
                                                                }
                                                                else if (Rs >= 1318) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (negBalance >= 42) {
                                                        if (Rs < 1770) {
                                                            if (TSCindex < 8) {
                                                                return 8;
                                                            }
                                                            else if (TSCindex >= 8) {
                                                                return 1;
                                                            }
                                                        }
                                                        else if (Rs >= 1770) {
                                                            if (Rs < 2269) {
                                                                return 8;
                                                            }
                                                            else if (Rs >= 2269) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                            else if (SC >= 3538) {
                                                if (RsDiff < 13) {
                                                    if (diffTSC < 2) {
                                                        return 4;
                                                    }
                                                    else if (diffTSC >= 2) {
                                                        if (SC < 4468) {
                                                            return 1;
                                                        }
                                                        else if (SC >= 4468) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                                else if (RsDiff >= 13) {
                                                    if (RsCsDiff < 731) {
                                                        return 8;
                                                    }
                                                    else if (RsCsDiff >= 731) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (Rs >= 3027) {
                                    if (Cs < 1238) {
                                        if (MVDiff < 106) {
                                            return 2;
                                        }
                                        else if (MVDiff >= 106) {
                                            if (tcor < 50) {
                                                return 2;
                                            }
                                            else if (tcor >= 50) {
                                                if (Rs < 3645) {
                                                    return 1;
                                                }
                                                else if (Rs >= 3645) {
                                                    return 2;
                                                }
                                            }
                                        }
                                    }
                                    else if (Cs >= 1238) {
                                        if (ssDCval < 129) {
                                            if (RsCsDiff < 363) {
                                                if (ssDCval < 112) {
                                                    if (Cs < 1872) {
                                                        return 8;
                                                    }
                                                    else if (Cs >= 1872) {
                                                        return 4;
                                                    }
                                                }
                                                else if (ssDCval >= 112) {
                                                    if (Rs < 3785) {
                                                        return 2;
                                                    }
                                                    else if (Rs >= 3785) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                            else if (RsCsDiff >= 363) {
                                                if (TSC < 21) {
                                                    return 2;
                                                }
                                                else if (TSC >= 21) {
                                                    return 4;
                                                }
                                            }
                                        }
                                        else if (ssDCval >= 129) {
                                            if (ssDCval < 137) {
                                                if (posBalance < 60) {
                                                    return 1;
                                                }
                                                else if (posBalance >= 60) {
                                                    if (Rs < 3248) {
                                                        return 1;
                                                    }
                                                    else if (Rs >= 3248) {
                                                        return 2;
                                                    }
                                                }
                                            }
                                            else if (ssDCval >= 137) {
                                                if (RsCsDiff < 575) {
                                                    if (RsDiff < 8) {
                                                        return 4;
                                                    }
                                                    else if (RsDiff >= 8) {
                                                        if (posBalance < 43) {
                                                            return 4;
                                                        }
                                                        else if (posBalance >= 43) {
                                                            return 2;
                                                        }
                                                    }
                                                }
                                                else if (RsCsDiff >= 575) {
                                                    return 4;
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                        }
                    }
                    else if (Cs >= 1938) {
                        if (CsDiff < 10) {
                            if (AFD < 12) {
                                if (ssDCval < 108) {
                                    if (ssDCval < 58) {
                                        if (MVDiff < 17) {
                                            if (diffRsCsdiff < 17) {
                                                return 4;
                                            }
                                            else if (diffRsCsdiff >= 17) {
                                                return 8;
                                            }
                                        }
                                        else if (MVDiff >= 17) {
                                            if (mu_mv_mag_sq < 36) {
                                                if (MVDiff < 29) {
                                                    return 8;
                                                }
                                                else if (MVDiff >= 29) {
                                                    if (SC < 6559) {
                                                        return 8;
                                                    }
                                                    else if (SC >= 6559) {
                                                        return 2;
                                                    }
                                                }
                                            }
                                            else if (mu_mv_mag_sq >= 36) {
                                                if (CsDiff < 5) {
                                                    return 2;
                                                }
                                                else if (CsDiff >= 5) {
                                                    return 4;
                                                }
                                            }
                                        }
                                    }
                                    else if (ssDCval >= 58) {
                                        if (RsCsDiff < 129) {
                                            if (Rs < 2915) {
                                                if (ssDCval < 93) {
                                                    if (negBalance < 26) {
                                                        return 1;
                                                    }
                                                    else if (negBalance >= 26) {
                                                        return 8;
                                                    }
                                                }
                                                else if (ssDCval >= 93) {
                                                    if (TSC < 9) {
                                                        return 2;
                                                    }
                                                    else if (TSC >= 9) {
                                                        return 1;
                                                    }
                                                }
                                            }
                                            else if (Rs >= 2915) {
                                                if (SC < 7505) {
                                                    if (Cs < 2103) {
                                                        return 4;
                                                    }
                                                    else if (Cs >= 2103) {
                                                        return 8;
                                                    }
                                                }
                                                else if (SC >= 7505) {
                                                    return 4;
                                                }
                                            }
                                        }
                                        else if (RsCsDiff >= 129) {
                                            if (TSC < 9) {
                                                return 8;
                                            }
                                            else if (TSC >= 9) {
                                                if (ssDCval < 103) {
                                                    return 4;
                                                }
                                                else if (ssDCval >= 103) {
                                                    return 2;
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (ssDCval >= 108) {
                                    if (Cs < 2605) {
                                        if (refDCval < 128) {
                                            if (SC < 6553) {
                                                if (mu_mv_mag_sq < 77) {
                                                    if (refDCval < 108) {
                                                        return 8;
                                                    }
                                                    else if (refDCval >= 108) {
                                                        return 4;
                                                    }
                                                }
                                                else if (mu_mv_mag_sq >= 77) {
                                                    if (TSCindex < 9) {
                                                        if (Cs < 2405) {
                                                            return 8;
                                                        }
                                                        else if (Cs >= 2405) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (TSCindex >= 9) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                            else if (SC >= 6553) {
                                                return 4;
                                            }
                                        }
                                        else if (refDCval >= 128) {
                                            return 1;
                                        }
                                    }
                                    else if (Cs >= 2605) {
                                        if (SC < 6234) {
                                            return 8;
                                        }
                                        else if (SC >= 6234) {
                                            if (Rs < 4249) {
                                                return 1;
                                            }
                                            else if (Rs >= 4249) {
                                                if (ssDCval < 112) {
                                                    return 8;
                                                }
                                                else if (ssDCval >= 112) {
                                                    return 2;
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                            else if (AFD >= 12) {
                                if (mu_mv_mag_sq < 46) {
                                    if (TSC < 13) {
                                        if (negBalance < 48) {
                                            if (posBalance < 54) {
                                                if (ssDCval < 80) {
                                                    if (diffAFD < -2) {
                                                        return 1;
                                                    }
                                                    else if (diffAFD >= -2) {
                                                        return 8;
                                                    }
                                                }
                                                else if (ssDCval >= 80) {
                                                    if (mcTcor < 92) {
                                                        if (diffAFD < 2) {
                                                            if (posBalance < 46) {
                                                                if (Rs < 4268) {
                                                                    return 8;
                                                                }
                                                                else if (Rs >= 4268) {
                                                                    return 1;
                                                                }
                                                            }
                                                            else if (posBalance >= 46) {
                                                                if (CsDiff < 9) {
                                                                    if (ssDCval < 84) {
                                                                        return 8;
                                                                    }
                                                                    else if (ssDCval >= 84) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (CsDiff >= 9) {
                                                                    if (TSC < 11) {
                                                                        return 4;
                                                                    }
                                                                    else if (TSC >= 11) {
                                                                        return 1;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (diffAFD >= 2) {
                                                            if (Rs < 2277) {
                                                                return 4;
                                                            }
                                                            else if (Rs >= 2277) {
                                                                return 1;
                                                            }
                                                        }
                                                    }
                                                    else if (mcTcor >= 92) {
                                                        if (negBalance < 42) {
                                                            return 1;
                                                        }
                                                        else if (negBalance >= 42) {
                                                            if (diffRsCsdiff < -1) {
                                                                return 8;
                                                            }
                                                            else if (diffRsCsdiff >= -1) {
                                                                if (Rs < 2406) {
                                                                    return 1;
                                                                }
                                                                else if (Rs >= 2406) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                            else if (posBalance >= 54) {
                                                if (ssDCval < 107) {
                                                    if (CsDiff < 7) {
                                                        return 8;
                                                    }
                                                    else if (CsDiff >= 7) {
                                                        if (MVDiff < 11) {
                                                            if (AFD < 14) {
                                                                if (Cs < 2307) {
                                                                    return 1;
                                                                }
                                                                else if (Cs >= 2307) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (AFD >= 14) {
                                                                return 2;
                                                            }
                                                        }
                                                        else if (MVDiff >= 11) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                                else if (ssDCval >= 107) {
                                                    if (Cs < 2027) {
                                                        return 2;
                                                    }
                                                    else if (Cs >= 2027) {
                                                        return 1;
                                                    }
                                                }
                                            }
                                        }
                                        else if (negBalance >= 48) {
                                            if (Rs < 2536) {
                                                if (mcTcor < 91) {
                                                    if (negBalance < 57) {
                                                        if (RsCsDiff < 99) {
                                                            if (Rs < 2441) {
                                                                return 1;
                                                            }
                                                            else if (Rs >= 2441) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (RsCsDiff >= 99) {
                                                            if (gchDC < 3) {
                                                                return 4;
                                                            }
                                                            else if (gchDC >= 3) {
                                                                return 2;
                                                            }
                                                        }
                                                    }
                                                    else if (negBalance >= 57) {
                                                        if (Rs < 1927) {
                                                            return 1;
                                                        }
                                                        else if (Rs >= 1927) {
                                                            return 2;
                                                        }
                                                    }
                                                }
                                                else if (mcTcor >= 91) {
                                                    if (diffRsCsdiff < -18) {
                                                        if (mu_mv_mag_sq < 3) {
                                                            if (Cs < 2179) {
                                                                return 2;
                                                            }
                                                            else if (Cs >= 2179) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (mu_mv_mag_sq >= 3) {
                                                            return 1;
                                                        }
                                                    }
                                                    else if (diffRsCsdiff >= -18) {
                                                        if (Rs < 2507) {
                                                            return 2;
                                                        }
                                                        else if (Rs >= 2507) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (Rs >= 2536) {
                                                if (ssDCval < 82) {
                                                    if (RsCsDiff < 121) {
                                                        if (AFD < 13) {
                                                            if (RsCsDiff < 56) {
                                                                if (Rs < 3599) {
                                                                    return 8;
                                                                }
                                                                else if (Rs >= 3599) {
                                                                    return 2;
                                                                }
                                                            }
                                                            else if (RsCsDiff >= 56) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (AFD >= 13) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (RsCsDiff >= 121) {
                                                        if (AFD < 16) {
                                                            return 2;
                                                        }
                                                        else if (AFD >= 16) {
                                                            return 1;
                                                        }
                                                    }
                                                }
                                                else if (ssDCval >= 82) {
                                                    if (posBalance < 57) {
                                                        if (ssDCval < 97) {
                                                            return 2;
                                                        }
                                                        else if (ssDCval >= 97) {
                                                            if (TSC < 12) {
                                                                if (Rs < 3741) {
                                                                    return 1;
                                                                }
                                                                else if (Rs >= 3741) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (TSC >= 12) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                    else if (posBalance >= 57) {
                                                        if (Cs < 2063) {
                                                            if (Cs < 2017) {
                                                                return 2;
                                                            }
                                                            else if (Cs >= 2017) {
                                                                return 1;
                                                            }
                                                        }
                                                        else if (Cs >= 2063) {
                                                            if (TSC < 12) {
                                                                if (tcor < 88) {
                                                                    return 2;
                                                                }
                                                                else if (tcor >= 88) {
                                                                    if (SC < 5438) {
                                                                        return 2;
                                                                    }
                                                                    else if (SC >= 5438) {
                                                                        return 1;
                                                                    }
                                                                }
                                                            }
                                                            else if (TSC >= 12) {
                                                                if (RsCsDiff < 163) {
                                                                    if (Rs < 2795) {
                                                                        return 2;
                                                                    }
                                                                    else if (Rs >= 2795) {
                                                                        return 1;
                                                                    }
                                                                }
                                                                else if (RsCsDiff >= 163) {
                                                                    if (Rs < 2650) {
                                                                        return 8;
                                                                    }
                                                                    else if (Rs >= 2650) {
                                                                        return 2;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (TSC >= 13) {
                                        if (negBalance < 54) {
                                            if (posBalance < 59) {
                                                if (Rs < 2508) {
                                                    return 4;
                                                }
                                                else if (Rs >= 2508) {
                                                    if (diffTSC < 1) {
                                                        if (Cs < 2308) {
                                                            return 8;
                                                        }
                                                        else if (Cs >= 2308) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (diffTSC >= 1) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                            else if (posBalance >= 59) {
                                                if (RsDiff < 11) {
                                                    if (Rs < 2473) {
                                                        return 2;
                                                    }
                                                    else if (Rs >= 2473) {
                                                        return 1;
                                                    }
                                                }
                                                else if (RsDiff >= 11) {
                                                    return 8;
                                                }
                                            }
                                        }
                                        else if (negBalance >= 54) {
                                            if (ssDCval < 77) {
                                                return 8;
                                            }
                                            else if (ssDCval >= 77) {
                                                if (RsDiff < 14) {
                                                    if (Rs < 3696) {
                                                        if (CsDiff < 8) {
                                                            return 8;
                                                        }
                                                        else if (CsDiff >= 8) {
                                                            return 1;
                                                        }
                                                    }
                                                    else if (Rs >= 3696) {
                                                        return 4;
                                                    }
                                                }
                                                else if (RsDiff >= 14) {
                                                    return 8;
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (mu_mv_mag_sq >= 46) {
                                    if (mcTcor < 94) {
                                        if (refDCval < 85) {
                                            if (mu_mv_mag_sq < 63) {
                                                if (TSC < 13) {
                                                    if (Rs < 2287) {
                                                        return 8;
                                                    }
                                                    else if (Rs >= 2287) {
                                                        return 1;
                                                    }
                                                }
                                                else if (TSC >= 13) {
                                                    return 8;
                                                }
                                            }
                                            else if (mu_mv_mag_sq >= 63) {
                                                return 8;
                                            }
                                        }
                                        else if (refDCval >= 85) {
                                            if (diffRsCsdiff < 36) {
                                                if (Cs < 2436) {
                                                    if (SC < 5262) {
                                                        if (posBalance < 54) {
                                                            return 1;
                                                        }
                                                        else if (posBalance >= 54) {
                                                            if (refDCval < 89) {
                                                                return 4;
                                                            }
                                                            else if (refDCval >= 89) {
                                                                if (SC < 4934) {
                                                                    if (Cs < 2087) {
                                                                        return 1;
                                                                    }
                                                                    else if (Cs >= 2087) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (SC >= 4934) {
                                                                    return 1;
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (SC >= 5262) {
                                                        if (ssDCval < 117) {
                                                            if (negBalance < 51) {
                                                                return 4;
                                                            }
                                                            else if (negBalance >= 51) {
                                                                if (RsCsDiff < 106) {
                                                                    return 1;
                                                                }
                                                                else if (RsCsDiff >= 106) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                        else if (ssDCval >= 117) {
                                                            if (diffTSC < 1) {
                                                                return 1;
                                                            }
                                                            else if (diffTSC >= 1) {
                                                                if (Cs < 2056) {
                                                                    return 4;
                                                                }
                                                                else if (Cs >= 2056) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (Cs >= 2436) {
                                                    if (Cs < 2841) {
                                                        if (SC < 6498) {
                                                            return 1;
                                                        }
                                                        else if (SC >= 6498) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (Cs >= 2841) {
                                                        if (RsCsDiff < 145) {
                                                            return 1;
                                                        }
                                                        else if (RsCsDiff >= 145) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (diffRsCsdiff >= 36) {
                                                if (Rs < 3315) {
                                                    if (MVDiff < 8) {
                                                        return 8;
                                                    }
                                                    else if (MVDiff >= 8) {
                                                        if (RsCsDiff < 243) {
                                                            return 4;
                                                        }
                                                        else if (RsCsDiff >= 243) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                                else if (Rs >= 3315) {
                                                    if (Rs < 4433) {
                                                        return 1;
                                                    }
                                                    else if (Rs >= 4433) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (mcTcor >= 94) {
                                        if (TSC < 10) {
                                            if (SC < 5709) {
                                                return 2;
                                            }
                                            else if (SC >= 5709) {
                                                return 4;
                                            }
                                        }
                                        else if (TSC >= 10) {
                                            if (Cs < 2403) {
                                                return 4;
                                            }
                                            else if (Cs >= 2403) {
                                                return 1;
                                            }
                                        }
                                    }
                                }
                            }
                        }
                        else if (CsDiff >= 10) {
                            if (refDCval < 64) {
                                if (TSC < 9) {
                                    if (RsDiff < 12) {
                                        return 4;
                                    }
                                    else if (RsDiff >= 12) {
                                        if (CsDiff < 16) {
                                            return 8;
                                        }
                                        else if (CsDiff >= 16) {
                                            return 4;
                                        }
                                    }
                                }
                                else if (TSC >= 9) {
                                    return 8;
                                }
                            }
                            else if (refDCval >= 64) {
                                if (mcTcor < 94) {
                                    if (posBalance < 56) {
                                        if (AFD < 22) {
                                            if (gchDC < 2) {
                                                if (SC < 7416) {
                                                    if (Rs < 1221) {
                                                        return 8;
                                                    }
                                                    else if (Rs >= 1221) {
                                                        if (CsDiff < 11) {
                                                            if (Rs < 2426) {
                                                                if (ssDCval < 84) {
                                                                    return 1;
                                                                }
                                                                else if (ssDCval >= 84) {
                                                                    if (AFD < 17) {
                                                                        if (RsCsDiff < 191) {
                                                                            return 8;
                                                                        }
                                                                        else if (RsCsDiff >= 191) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                    else if (AFD >= 17) {
                                                                        return 4;
                                                                    }
                                                                }
                                                            }
                                                            else if (Rs >= 2426) {
                                                                if (MVDiff < 9) {
                                                                    if (Cs < 2551) {
                                                                        return 1;
                                                                    }
                                                                    else if (Cs >= 2551) {
                                                                        return 8;
                                                                    }
                                                                }
                                                                else if (MVDiff >= 9) {
                                                                    if (ssDCval < 83) {
                                                                        if (refDCval < 76) {
                                                                            return 1;
                                                                        }
                                                                        else if (refDCval >= 76) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                    else if (ssDCval >= 83) {
                                                                        return 1;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (CsDiff >= 11) {
                                                            if (Cs < 2823) {
                                                                if (posBalance < 36) {
                                                                    return 8;
                                                                }
                                                                else if (posBalance >= 36) {
                                                                    if (MVDiff < 8) {
                                                                        if (posBalance < 54) {
                                                                            return 1;
                                                                        }
                                                                        else if (posBalance >= 54) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                    else if (MVDiff >= 8) {
                                                                        return 1;
                                                                    }
                                                                }
                                                            }
                                                            else if (Cs >= 2823) {
                                                                if (tcor < 81) {
                                                                    if (Rs < 3040) {
                                                                        return 1;
                                                                    }
                                                                    else if (Rs >= 3040) {
                                                                        if (AFD < 19) {
                                                                            return 1;
                                                                        }
                                                                        else if (AFD >= 19) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                }
                                                                else if (tcor >= 81) {
                                                                    if (Rs < 2624) {
                                                                        if (diffAFD < 3) {
                                                                            return 8;
                                                                        }
                                                                        else if (diffAFD >= 3) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                    else if (Rs >= 2624) {
                                                                        return 1;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (SC >= 7416) {
                                                    if (Rs < 4665) {
                                                        return 4;
                                                    }
                                                    else if (Rs >= 4665) {
                                                        return 1;
                                                    }
                                                }
                                            }
                                            else if (gchDC >= 2) {
                                                if (SC < 6557) {
                                                    if (RsCsDiff < 243) {
                                                        if (MVDiff < 15) {
                                                            return 4;
                                                        }
                                                        else if (MVDiff >= 15) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (RsCsDiff >= 243) {
                                                        if (posBalance < 47) {
                                                            if (Rs < 1993) {
                                                                return 1;
                                                            }
                                                            else if (Rs >= 1993) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (posBalance >= 47) {
                                                            if (tcor < 84) {
                                                                return 1;
                                                            }
                                                            else if (tcor >= 84) {
                                                                if (Cs < 2406) {
                                                                    return 4;
                                                                }
                                                                else if (Cs >= 2406) {
                                                                    return 1;
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (SC >= 6557) {
                                                    return 8;
                                                }
                                            }
                                        }
                                        else if (AFD >= 22) {
                                            if (mcTcor < 85) {
                                                if (gchDC < 3) {
                                                    return 8;
                                                }
                                                else if (gchDC >= 3) {
                                                    if (MVDiff < 101) {
                                                        return 2;
                                                    }
                                                    else if (MVDiff >= 101) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                            else if (mcTcor >= 85) {
                                                if (mu_mv_mag_sq < 11) {
                                                    return 8;
                                                }
                                                else if (mu_mv_mag_sq >= 11) {
                                                    if (MVDiff < 14) {
                                                        if (Cs < 2482) {
                                                            return 4;
                                                        }
                                                        else if (Cs >= 2482) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (MVDiff >= 14) {
                                                        return 1;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (posBalance >= 56) {
                                        if (ssDCval < 107) {
                                            if (AFD < 18) {
                                                if (Rs < 2053) {
                                                    if (SC < 4019) {
                                                        return 2;
                                                    }
                                                    else if (SC >= 4019) {
                                                        return 1;
                                                    }
                                                }
                                                else if (Rs >= 2053) {
                                                    if (RsCsDiff < 201) {
                                                        if (diffMVdiffVal < 3) {
                                                            if (Rs < 3364) {
                                                                return 4;
                                                            }
                                                            else if (Rs >= 3364) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (diffMVdiffVal >= 3) {
                                                            if (Cs < 2451) {
                                                                return 2;
                                                            }
                                                            else if (Cs >= 2451) {
                                                                if (refDCval < 96) {
                                                                    return 1;
                                                                }
                                                                else if (refDCval >= 96) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (RsCsDiff >= 201) {
                                                        if (MVDiff < 37) {
                                                            return 4;
                                                        }
                                                        else if (MVDiff >= 37) {
                                                            return 2;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (AFD >= 18) {
                                                if (tcor < 46) {
                                                    return 4;
                                                }
                                                else if (tcor >= 46) {
                                                    if (AFD < 29) {
                                                        if (refDCval < 82) {
                                                            if (AFD < 21) {
                                                                if (Cs < 2949) {
                                                                    return 8;
                                                                }
                                                                else if (Cs >= 2949) {
                                                                    return 1;
                                                                }
                                                            }
                                                            else if (AFD >= 21) {
                                                                if (posBalance < 57) {
                                                                    return 8;
                                                                }
                                                                else if (posBalance >= 57) {
                                                                    return 1;
                                                                }
                                                            }
                                                        }
                                                        else if (refDCval >= 82) {
                                                            if (tcor < 72) {
                                                                if (Cs < 2543) {
                                                                    if (SC < 4140) {
                                                                        return 1;
                                                                    }
                                                                    else if (SC >= 4140) {
                                                                        return 2;
                                                                    }
                                                                }
                                                                else if (Cs >= 2543) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (tcor >= 72) {
                                                                if (Rs < 2684) {
                                                                    if (Cs < 2116) {
                                                                        return 1;
                                                                    }
                                                                    else if (Cs >= 2116) {
                                                                        if (CsDiff < 15) {
                                                                            return 4;
                                                                        }
                                                                        else if (CsDiff >= 15) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                }
                                                                else if (Rs >= 2684) {
                                                                    if (Rs < 4080) {
                                                                        if (MVDiff < 11) {
                                                                            return 4;
                                                                        }
                                                                        else if (MVDiff >= 11) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                    else if (Rs >= 4080) {
                                                                        return 4;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (AFD >= 29) {
                                                        return 2;
                                                    }
                                                }
                                            }
                                        }
                                        else if (ssDCval >= 107) {
                                            if (tcor < 68) {
                                                if (ssDCval < 128) {
                                                    if (RsCsDiff < 472) {
                                                        if (Cs < 2383) {
                                                            return 1;
                                                        }
                                                        else if (Cs >= 2383) {
                                                            if (AFD < 27) {
                                                                if (SC < 5248) {
                                                                    return 2;
                                                                }
                                                                else if (SC >= 5248) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (AFD >= 27) {
                                                                return 1;
                                                            }
                                                        }
                                                    }
                                                    else if (RsCsDiff >= 472) {
                                                        if (AFD < 24) {
                                                            return 1;
                                                        }
                                                        else if (AFD >= 24) {
                                                            if (Cs < 2257) {
                                                                if (AFD < 30) {
                                                                    return 2;
                                                                }
                                                                else if (AFD >= 30) {
                                                                    return 1;
                                                                }
                                                            }
                                                            else if (Cs >= 2257) {
                                                                if (Cs < 3183) {
                                                                    if (Cs < 2684) {
                                                                        return 8;
                                                                    }
                                                                    else if (Cs >= 2684) {
                                                                        return 2;
                                                                    }
                                                                }
                                                                else if (Cs >= 3183) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (ssDCval >= 128) {
                                                    if (TSC < 25) {
                                                        return 1;
                                                    }
                                                    else if (TSC >= 25) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                            else if (tcor >= 68) {
                                                if (negBalance < 54) {
                                                    if (tcor < 82) {
                                                        if (ssDCval < 124) {
                                                            if (AFD < 21) {
                                                                if (Cs < 2502) {
                                                                    return 4;
                                                                }
                                                                else if (Cs >= 2502) {
                                                                    if (negBalance < 49) {
                                                                        return 2;
                                                                    }
                                                                    else if (negBalance >= 49) {
                                                                        return 8;
                                                                    }
                                                                }
                                                            }
                                                            else if (AFD >= 21) {
                                                                if (TSC < 12) {
                                                                    return 4;
                                                                }
                                                                else if (TSC >= 12) {
                                                                    return 1;
                                                                }
                                                            }
                                                        }
                                                        else if (ssDCval >= 124) {
                                                            if (MVDiff < 21) {
                                                                if (CsDiff < 14) {
                                                                    return 1;
                                                                }
                                                                else if (CsDiff >= 14) {
                                                                    if (Rs < 3321) {
                                                                        return 8;
                                                                    }
                                                                    else if (Rs >= 3321) {
                                                                        return 1;
                                                                    }
                                                                }
                                                            }
                                                            else if (MVDiff >= 21) {
                                                                if (AFD < 21) {
                                                                    return 2;
                                                                }
                                                                else if (AFD >= 21) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (tcor >= 82) {
                                                        if (AFD < 13) {
                                                            return 4;
                                                        }
                                                        else if (AFD >= 13) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                                else if (negBalance >= 54) {
                                                    if (ssDCval < 124) {
                                                        if (AFD < 20) {
                                                            return 4;
                                                        }
                                                        else if (AFD >= 20) {
                                                            return 1;
                                                        }
                                                    }
                                                    else if (ssDCval >= 124) {
                                                        if (Rs < 3858) {
                                                            if (Rs < 2530) {
                                                                return 2;
                                                            }
                                                            else if (Rs >= 2530) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (Rs >= 3858) {
                                                            if (CsDiff < 13) {
                                                                return 2;
                                                            }
                                                            else if (CsDiff >= 13) {
                                                                return 1;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (mcTcor >= 94) {
                                    if (diffRsCsdiff < -122) {
                                        return 8;
                                    }
                                    else if (diffRsCsdiff >= -122) {
                                        if (Cs < 2954) {
                                            if (tcor < 88) {
                                                return 2;
                                            }
                                            else if (tcor >= 88) {
                                                if (Cs < 2317) {
                                                    return 2;
                                                }
                                                else if (Cs >= 2317) {
                                                    return 4;
                                                }
                                            }
                                        }
                                        else if (Cs >= 2954) {
                                            return 4;
                                        }
                                    }
                                }
                            }
                        }
                    }
                }
                else if (tcor >= 90) {
                    if (Rs < 3000) {
                        if (Rs < 2549) {
                            if (Cs < 1027) {
                                if (posBalance < 39) {
                                    if (tcor < 119) {
                                        if (posBalance < 9) {
                                            if (Cs < 900) {
                                                if (negBalance < 115) {
                                                    if (negBalance < 48) {
                                                        if (gchDC < 1) {
                                                            return 8;
                                                        }
                                                        else if (gchDC >= 1) {
                                                            if (tcor < 101) {
                                                                return 8;
                                                            }
                                                            else if (tcor >= 101) {
                                                                if (AFD < 2) {
                                                                    return 4;
                                                                }
                                                                else if (AFD >= 2) {
                                                                    return 2;
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (negBalance >= 48) {
                                                        if (ssDCval < 48) {
                                                            if (negBalance < 92) {
                                                                return 4;
                                                            }
                                                            else if (negBalance >= 92) {
                                                                if (gchDC < 3) {
                                                                    return 4;
                                                                }
                                                                else if (gchDC >= 3) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                        else if (ssDCval >= 48) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                                else if (negBalance >= 115) {
                                                    return 1;
                                                }
                                            }
                                            else if (Cs >= 900) {
                                                return 1;
                                            }
                                        }
                                        else if (posBalance >= 9) {
                                            if (mu_mv_mag_sq < 110) {
                                                if (refDCval < 114) {
                                                    if (mcTcor < 88) {
                                                        return 8;
                                                    }
                                                    else if (mcTcor >= 88) {
                                                        if (refDCval < 78) {
                                                            if (ssDCval < 29) {
                                                                if (SC < 674) {
                                                                    return 4;
                                                                }
                                                                else if (SC >= 674) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (ssDCval >= 29) {
                                                                if (diffRsCsdiff < -1) {
                                                                    if (ssDCval < 39) {
                                                                        if (Rs < 99) {
                                                                            return 4;
                                                                        }
                                                                        else if (Rs >= 99) {
                                                                            return 2;
                                                                        }
                                                                    }
                                                                    else if (ssDCval >= 39) {
                                                                        if (CsDiff < 1) {
                                                                            return 8;
                                                                        }
                                                                        else if (CsDiff >= 1) {
                                                                            if (posBalance < 37) {
                                                                                return 4;
                                                                            }
                                                                            else if (posBalance >= 37) {
                                                                                if (Rs < 1756) {
                                                                                    return 8;
                                                                                }
                                                                                else if (Rs >= 1756) {
                                                                                    if (Cs < 346) {
                                                                                        return 2;
                                                                                    }
                                                                                    else if (Cs >= 346) {
                                                                                        return 4;
                                                                                    }
                                                                                }
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                                else if (diffRsCsdiff >= -1) {
                                                                    if (SC < 2957) {
                                                                        if (RsDiff < 2) {
                                                                            return 4;
                                                                        }
                                                                        else if (RsDiff >= 2) {
                                                                            if (AFD < 5) {
                                                                                return 2;
                                                                            }
                                                                            else if (AFD >= 5) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (SC >= 2957) {
                                                                        return 2;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (refDCval >= 78) {
                                                            if (SC < 1241) {
                                                                if (ssDCval < 79) {
                                                                    if (Cs < 235) {
                                                                        return 2;
                                                                    }
                                                                    else if (Cs >= 235) {
                                                                        return 8;
                                                                    }
                                                                }
                                                                else if (ssDCval >= 79) {
                                                                    return 2;
                                                                }
                                                            }
                                                            else if (SC >= 1241) {
                                                                if (Cs < 798) {
                                                                    return 8;
                                                                }
                                                                else if (Cs >= 798) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (refDCval >= 114) {
                                                    if (diffAFD < -1) {
                                                        if (posBalance < 28) {
                                                            if (MVDiff < 8) {
                                                                if (Rs < 776) {
                                                                    return 4;
                                                                }
                                                                else if (Rs >= 776) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (MVDiff >= 8) {
                                                                return 1;
                                                            }
                                                        }
                                                        else if (posBalance >= 28) {
                                                            if (Rs < 835) {
                                                                return 2;
                                                            }
                                                            else if (Rs >= 835) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                    else if (diffAFD >= -1) {
                                                        if (MVDiff < 51) {
                                                            if (Cs < 465) {
                                                                if (Rs < 456) {
                                                                    return 4;
                                                                }
                                                                else if (Rs >= 456) {
                                                                    return 2;
                                                                }
                                                            }
                                                            else if (Cs >= 465) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (MVDiff >= 51) {
                                                            if (Rs < 544) {
                                                                return 1;
                                                            }
                                                            else if (Rs >= 544) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                            else if (mu_mv_mag_sq >= 110) {
                                                if (MVDiff < 94) {
                                                    if (diffRsCsdiff < -114) {
                                                        return 4;
                                                    }
                                                    else if (diffRsCsdiff >= -114) {
                                                        return 8;
                                                    }
                                                }
                                                else if (MVDiff >= 94) {
                                                    return 1;
                                                }
                                            }
                                        }
                                    }
                                    else if (tcor >= 119) {
                                        if (MVDiff < 125) {
                                            if (AFD < 1) {
                                                return 8;
                                            }
                                            else if (AFD >= 1) {
                                                if (mcTcor < 109) {
                                                    if (diffAFD < -1) {
                                                        return 2;
                                                    }
                                                    else if (diffAFD >= -1) {
                                                        if (refDCval < 24) {
                                                            return 2;
                                                        }
                                                        else if (refDCval >= 24) {
                                                            if (CsDiff < 3) {
                                                                return 4;
                                                            }
                                                            else if (CsDiff >= 3) {
                                                                return 2;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (mcTcor >= 109) {
                                                    if (AFD < 6) {
                                                        if (ssDCval < 22) {
                                                            return 4;
                                                        }
                                                        else if (ssDCval >= 22) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (AFD >= 6) {
                                                        return 2;
                                                    }
                                                }
                                            }
                                        }
                                        else if (MVDiff >= 125) {
                                            if (Rs < 112) {
                                                return 8;
                                            }
                                            else if (Rs >= 112) {
                                                return 1;
                                            }
                                        }
                                    }
                                }
                                else if (posBalance >= 39) {
                                    if (CsDiff < 3) {
                                        if (mu_mv_mag_sq < 84) {
                                            if (TSC < 5) {
                                                if (Cs < 854) {
                                                    if (Rs < 2060) {
                                                        if (Cs < 251) {
                                                            if (refDCval < 79) {
                                                                if (SC < 167) {
                                                                    return 4;
                                                                }
                                                                else if (SC >= 167) {
                                                                    return 2;
                                                                }
                                                            }
                                                            else if (refDCval >= 79) {
                                                                if (Rs < 771) {
                                                                    return 1;
                                                                }
                                                                else if (Rs >= 771) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                        else if (Cs >= 251) {
                                                            if (ssDCval < 60) {
                                                                if (posBalance < 52) {
                                                                    if (Rs < 944) {
                                                                        return 4;
                                                                    }
                                                                    else if (Rs >= 944) {
                                                                        if (posBalance < 47) {
                                                                            if (MVDiff < 15) {
                                                                                return 4;
                                                                            }
                                                                            else if (MVDiff >= 15) {
                                                                                return 8;
                                                                            }
                                                                        }
                                                                        else if (posBalance >= 47) {
                                                                            if (posBalance < 50) {
                                                                                if (Cs < 665) {
                                                                                    return 4;
                                                                                }
                                                                                else if (Cs >= 665) {
                                                                                    return 1;
                                                                                }
                                                                            }
                                                                            else if (posBalance >= 50) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                                else if (posBalance >= 52) {
                                                                    if (SC < 1600) {
                                                                        if (Rs < 646) {
                                                                            return 4;
                                                                        }
                                                                        else if (Rs >= 646) {
                                                                            if (TSC < 4) {
                                                                                return 8;
                                                                            }
                                                                            else if (TSC >= 4) {
                                                                                if (Rs < 856) {
                                                                                    return 4;
                                                                                }
                                                                                else if (Rs >= 856) {
                                                                                    return 8;
                                                                                }
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (SC >= 1600) {
                                                                        return 4;
                                                                    }
                                                                }
                                                            }
                                                            else if (ssDCval >= 60) {
                                                                if (Rs < 468) {
                                                                    if (Rs < 262) {
                                                                        return 2;
                                                                    }
                                                                    else if (Rs >= 262) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (Rs >= 468) {
                                                                    if (gchDC < 3) {
                                                                        return 2;
                                                                    }
                                                                    else if (gchDC >= 3) {
                                                                        return 4;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (Rs >= 2060) {
                                                        return 2;
                                                    }
                                                }
                                                else if (Cs >= 854) {
                                                    if (tcor < 93) {
                                                        if (AFD < 5) {
                                                            return 1;
                                                        }
                                                        else if (AFD >= 5) {
                                                            if (mu_mv_mag_sq < 62) {
                                                                if (MVDiff < 6) {
                                                                    return 1;
                                                                }
                                                                else if (MVDiff >= 6) {
                                                                    if (negBalance < 37) {
                                                                        return 1;
                                                                    }
                                                                    else if (negBalance >= 37) {
                                                                        if (Cs < 954) {
                                                                            return 4;
                                                                        }
                                                                        else if (Cs >= 954) {
                                                                            if (SC < 2425) {
                                                                                return 1;
                                                                            }
                                                                            else if (SC >= 2425) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                            else if (mu_mv_mag_sq >= 62) {
                                                                if (ssDCval < 50) {
                                                                    return 1;
                                                                }
                                                                else if (ssDCval >= 50) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (tcor >= 93) {
                                                        if (SC < 2879) {
                                                            return 1;
                                                        }
                                                        else if (SC >= 2879) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (TSC >= 5) {
                                                if (gchDC < 2) {
                                                    if (diffRsCsdiff < 4) {
                                                        if (ssDCval < 104) {
                                                            if (CsDiff < 2) {
                                                                return 1;
                                                            }
                                                            else if (CsDiff >= 2) {
                                                                if (Rs < 1187) {
                                                                    if (AFD < 7) {
                                                                        if (Rs < 1012) {
                                                                            return 4;
                                                                        }
                                                                        else if (Rs >= 1012) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                    else if (AFD >= 7) {
                                                                        return 8;
                                                                    }
                                                                }
                                                                else if (Rs >= 1187) {
                                                                    return 1;
                                                                }
                                                            }
                                                        }
                                                        else if (ssDCval >= 104) {
                                                            return 2;
                                                        }
                                                    }
                                                    else if (diffRsCsdiff >= 4) {
                                                        if (Rs < 810) {
                                                            return 8;
                                                        }
                                                        else if (Rs >= 810) {
                                                            if (posBalance < 53) {
                                                                return 4;
                                                            }
                                                            else if (posBalance >= 53) {
                                                                return 1;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (gchDC >= 2) {
                                                    if (mcTcor < 94) {
                                                        if (diffMVdiffVal < 6) {
                                                            if (Rs < 344) {
                                                                return 1;
                                                            }
                                                            else if (Rs >= 344) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (diffMVdiffVal >= 6) {
                                                            return 1;
                                                        }
                                                    }
                                                    else if (mcTcor >= 94) {
                                                        if (Cs < 281) {
                                                            if (Cs < 65) {
                                                                return 1;
                                                            }
                                                            else if (Cs >= 65) {
                                                                return 2;
                                                            }
                                                        }
                                                        else if (Cs >= 281) {
                                                            if (Rs < 974) {
                                                                if (MVDiff < 24) {
                                                                    if (Rs < 771) {
                                                                        return 4;
                                                                    }
                                                                    else if (Rs >= 771) {
                                                                        return 2;
                                                                    }
                                                                }
                                                                else if (MVDiff >= 24) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (Rs >= 974) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (mu_mv_mag_sq >= 84) {
                                            if (mcTcor < 98) {
                                                if (tcor < 96) {
                                                    if (diffMVdiffVal < -1) {
                                                        if (posBalance < 46) {
                                                            return 1;
                                                        }
                                                        else if (posBalance >= 46) {
                                                            if (TSC < 5) {
                                                                return 8;
                                                            }
                                                            else if (TSC >= 5) {
                                                                if (Rs < 1140) {
                                                                    return 8;
                                                                }
                                                                else if (Rs >= 1140) {
                                                                    return 1;
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (diffMVdiffVal >= -1) {
                                                        return 8;
                                                    }
                                                }
                                                else if (tcor >= 96) {
                                                    if (Cs < 755) {
                                                        return 1;
                                                    }
                                                    else if (Cs >= 755) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                            else if (mcTcor >= 98) {
                                                if (refDCval < 103) {
                                                    if (ssDCval < 83) {
                                                        if (SC < 526) {
                                                            return 8;
                                                        }
                                                        else if (SC >= 526) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (ssDCval >= 83) {
                                                        return 8;
                                                    }
                                                }
                                                else if (refDCval >= 103) {
                                                    if (MVDiff < 6) {
                                                        return 4;
                                                    }
                                                    else if (MVDiff >= 6) {
                                                        return 2;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (CsDiff >= 3) {
                                        if (Cs < 786) {
                                            if (SC < 1351) {
                                                if (AFD < 14) {
                                                    if (AFD < 8) {
                                                        return 8;
                                                    }
                                                    else if (AFD >= 8) {
                                                        if (AFD < 10) {
                                                            if (Cs < 656) {
                                                                return 4;
                                                            }
                                                            else if (Cs >= 656) {
                                                                return 2;
                                                            }
                                                        }
                                                        else if (AFD >= 10) {
                                                            return 2;
                                                        }
                                                    }
                                                }
                                                else if (AFD >= 14) {
                                                    return 1;
                                                }
                                            }
                                            else if (SC >= 1351) {
                                                return 2;
                                            }
                                        }
                                        else if (Cs >= 786) {
                                            if (ssDCval < 70) {
                                                if (mu_mv_mag_sq < 65) {
                                                    return 1;
                                                }
                                                else if (mu_mv_mag_sq >= 65) {
                                                    return 4;
                                                }
                                            }
                                            else if (ssDCval >= 70) {
                                                if (SC < 1683) {
                                                    return 2;
                                                }
                                                else if (SC >= 1683) {
                                                    if (Cs < 895) {
                                                        return 8;
                                                    }
                                                    else if (Cs >= 895) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                            else if (Cs >= 1027) {
                                if (CsDiff < 2) {
                                    if (mcTcor < 95) {
                                        if (Cs < 1549) {
                                            if (Rs < 1766) {
                                                if (MVDiff < 23) {
                                                    return 8;
                                                }
                                                else if (MVDiff >= 23) {
                                                    return 2;
                                                }
                                            }
                                            else if (Rs >= 1766) {
                                                if (Cs < 1499) {
                                                    if (MVDiff < 20) {
                                                        return 4;
                                                    }
                                                    else if (MVDiff >= 20) {
                                                        if (SC < 3324) {
                                                            return 4;
                                                        }
                                                        else if (SC >= 3324) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                                else if (Cs >= 1499) {
                                                    if (MVDiff < 15) {
                                                        if (mu_mv_mag_sq < 61) {
                                                            return 1;
                                                        }
                                                        else if (mu_mv_mag_sq >= 61) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (MVDiff >= 15) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                        }
                                        else if (Cs >= 1549) {
                                            if (diffMVdiffVal < -5) {
                                                return 1;
                                            }
                                            else if (diffMVdiffVal >= -5) {
                                                if (Cs < 1587) {
                                                    if (diffAFD < -1) {
                                                        return 1;
                                                    }
                                                    else if (diffAFD >= -1) {
                                                        if (diffMVdiffVal < 7) {
                                                            return 4;
                                                        }
                                                        else if (diffMVdiffVal >= 7) {
                                                            return 1;
                                                        }
                                                    }
                                                }
                                                else if (Cs >= 1587) {
                                                    return 1;
                                                }
                                            }
                                        }
                                    }
                                    else if (mcTcor >= 95) {
                                        if (SC < 4054) {
                                            if (negBalance < 31) {
                                                if (Cs < 1129) {
                                                    return 8;
                                                }
                                                else if (Cs >= 1129) {
                                                    if (Cs < 1491) {
                                                        if (RsDiff < 1) {
                                                            return 4;
                                                        }
                                                        else if (RsDiff >= 1) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (Cs >= 1491) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                            else if (negBalance >= 31) {
                                                return 8;
                                            }
                                        }
                                        else if (SC >= 4054) {
                                            return 2;
                                        }
                                    }
                                }
                                else if (CsDiff >= 2) {
                                    if (negBalance < 44) {
                                        if (SC < 4233) {
                                            if (mu_mv_mag_sq < 76) {
                                                if (TSC < 7) {
                                                    if (ssDCval < 55) {
                                                        if (Rs < 1058) {
                                                            return 2;
                                                        }
                                                        else if (Rs >= 1058) {
                                                            return 1;
                                                        }
                                                    }
                                                    else if (ssDCval >= 55) {
                                                        if (mcTcor < 95) {
                                                            if (Cs < 1071) {
                                                                return 4;
                                                            }
                                                            else if (Cs >= 1071) {
                                                                if (Rs < 1241) {
                                                                    return 8;
                                                                }
                                                                else if (Rs >= 1241) {
                                                                    return 1;
                                                                }
                                                            }
                                                        }
                                                        else if (mcTcor >= 95) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                                else if (TSC >= 7) {
                                                    if (gchDC < 3) {
                                                        if (posBalance < 61) {
                                                            if (Cs < 2222) {
                                                                return 8;
                                                            }
                                                            else if (Cs >= 2222) {
                                                                return 1;
                                                            }
                                                        }
                                                        else if (posBalance >= 61) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (gchDC >= 3) {
                                                        if (Rs < 1569) {
                                                            return 4;
                                                        }
                                                        else if (Rs >= 1569) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (mu_mv_mag_sq >= 76) {
                                                if (diffTSC < -3) {
                                                    if (Cs < 1199) {
                                                        return 8;
                                                    }
                                                    else if (Cs >= 1199) {
                                                        return 4;
                                                    }
                                                }
                                                else if (diffTSC >= -3) {
                                                    if (Rs < 1846) {
                                                        if (SC < 2026) {
                                                            if (AFD < 3) {
                                                                return 8;
                                                            }
                                                            else if (AFD >= 3) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (SC >= 2026) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (Rs >= 1846) {
                                                        if (Scindex < 5) {
                                                            return 4;
                                                        }
                                                        else if (Scindex >= 5) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (SC >= 4233) {
                                            if (ssDCval < 101) {
                                                if (Rs < 2136) {
                                                    return 1;
                                                }
                                                else if (Rs >= 2136) {
                                                    if (ssDCval < 93) {
                                                        if (SC < 4540) {
                                                            if (RsCsDiff < 129) {
                                                                if (diffMVdiffVal < -3) {
                                                                    return 8;
                                                                }
                                                                else if (diffMVdiffVal >= -3) {
                                                                    if (Rs < 2281) {
                                                                        return 8;
                                                                    }
                                                                    else if (Rs >= 2281) {
                                                                        if (AFD < 7) {
                                                                            return 8;
                                                                        }
                                                                        else if (AFD >= 7) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                            else if (RsCsDiff >= 129) {
                                                                return 1;
                                                            }
                                                        }
                                                        else if (SC >= 4540) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (ssDCval >= 93) {
                                                        return 1;
                                                    }
                                                }
                                            }
                                            else if (ssDCval >= 101) {
                                                if (AFD < 10) {
                                                    return 4;
                                                }
                                                else if (AFD >= 10) {
                                                    return 2;
                                                }
                                            }
                                        }
                                    }
                                    else if (negBalance >= 44) {
                                        if (AFD < 12) {
                                            if (mu_mv_mag_sq < 74) {
                                                if (refDCval < 86) {
                                                    if (Rs < 1886) {
                                                        if (Cs < 1173) {
                                                            return 1;
                                                        }
                                                        else if (Cs >= 1173) {
                                                            if (Rs < 1356) {
                                                                return 2;
                                                            }
                                                            else if (Rs >= 1356) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                    else if (Rs >= 1886) {
                                                        if (SC < 4240) {
                                                            return 1;
                                                        }
                                                        else if (SC >= 4240) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                                else if (refDCval >= 86) {
                                                    if (diffRsCsdiff < -1) {
                                                        if (mcTcor < 94) {
                                                            return 8;
                                                        }
                                                        else if (mcTcor >= 94) {
                                                            if (Rs < 2498) {
                                                                return 2;
                                                            }
                                                            else if (Rs >= 2498) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                    else if (diffRsCsdiff >= -1) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                            else if (mu_mv_mag_sq >= 74) {
                                                if (TSC < 8) {
                                                    if (tcor < 97) {
                                                        return 8;
                                                    }
                                                    else if (tcor >= 97) {
                                                        if (Cs < 1741) {
                                                            return 4;
                                                        }
                                                        else if (Cs >= 1741) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                                else if (TSC >= 8) {
                                                    if (diffMVdiffVal < 1) {
                                                        return 4;
                                                    }
                                                    else if (diffMVdiffVal >= 1) {
                                                        if (Rs < 2085) {
                                                            return 2;
                                                        }
                                                        else if (Rs >= 2085) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (AFD >= 12) {
                                            if (MVDiff < 13) {
                                                if (Cs < 2140) {
                                                    if (gchDC < 2) {
                                                        return 1;
                                                    }
                                                    else if (gchDC >= 2) {
                                                        return 4;
                                                    }
                                                }
                                                else if (Cs >= 2140) {
                                                    return 2;
                                                }
                                            }
                                            else if (MVDiff >= 13) {
                                                if (Cs < 1761) {
                                                    if (RsCsDiff < 81) {
                                                        return 2;
                                                    }
                                                    else if (RsCsDiff >= 81) {
                                                        return 1;
                                                    }
                                                }
                                                else if (Cs >= 1761) {
                                                    return 2;
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                        }
                        else if (Rs >= 2549) {
                            if (refDCval < 98) {
                                if (SC < 5062) {
                                    if (MVDiff < 22) {
                                        if (Cs < 1151) {
                                            return 4;
                                        }
                                        else if (Cs >= 1151) {
                                            return 1;
                                        }
                                    }
                                    else if (MVDiff >= 22) {
                                        if (Rs < 2687) {
                                            return 2;
                                        }
                                        else if (Rs >= 2687) {
                                            return 1;
                                        }
                                    }
                                }
                                else if (SC >= 5062) {
                                    if (Cs < 2387) {
                                        if (diffMVdiffVal < -7) {
                                            return 4;
                                        }
                                        else if (diffMVdiffVal >= -7) {
                                            if (Rs < 2826) {
                                                return 4;
                                            }
                                            else if (Rs >= 2826) {
                                                return 1;
                                            }
                                        }
                                    }
                                    else if (Cs >= 2387) {
                                        if (Cs < 2420) {
                                            return 4;
                                        }
                                        else if (Cs >= 2420) {
                                            if (AFD < 8) {
                                                return 8;
                                            }
                                            else if (AFD >= 8) {
                                                return 2;
                                            }
                                        }
                                    }
                                }
                            }
                            else if (refDCval >= 98) {
                                if (Cs < 1513) {
                                    if (Rs < 2722) {
                                        return 2;
                                    }
                                    else if (Rs >= 2722) {
                                        return 4;
                                    }
                                }
                                else if (Cs >= 1513) {
                                    if (MVDiff < 17) {
                                        return 8;
                                    }
                                    else if (MVDiff >= 17) {
                                        if (AFD < 10) {
                                            return 8;
                                        }
                                        else if (AFD >= 10) {
                                            return 2;
                                        }
                                    }
                                }
                            }
                        }
                    }
                    else if (Rs >= 3000) {
                        if (TSC < 10) {
                            if (TSC < 6) {
                                if (ssDCval < 108) {
                                    if (negBalance < 37) {
                                        if (Rs < 4385) {
                                            if (mu_mv_mag_sq < 41) {
                                                if (RsCsDiff < 29) {
                                                    return 4;
                                                }
                                                else if (RsCsDiff >= 29) {
                                                    if (MVDiff < 13) {
                                                        return 8;
                                                    }
                                                    else if (MVDiff >= 13) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                            else if (mu_mv_mag_sq >= 41) {
                                                return 8;
                                            }
                                        }
                                        else if (Rs >= 4385) {
                                            return 2;
                                        }
                                    }
                                    else if (negBalance >= 37) {
                                        if (SC < 6784) {
                                            if (Rs < 3188) {
                                                if (AFD < 7) {
                                                    if (Rs < 3159) {
                                                        return 8;
                                                    }
                                                    else if (Rs >= 3159) {
                                                        return 4;
                                                    }
                                                }
                                                else if (AFD >= 7) {
                                                    return 4;
                                                }
                                            }
                                            else if (Rs >= 3188) {
                                                return 8;
                                            }
                                        }
                                        else if (SC >= 6784) {
                                            if (MVDiff < 9) {
                                                if (Cs < 3404) {
                                                    return 4;
                                                }
                                                else if (Cs >= 3404) {
                                                    return 8;
                                                }
                                            }
                                            else if (MVDiff >= 9) {
                                                if (negBalance < 41) {
                                                    return 8;
                                                }
                                                else if (negBalance >= 41) {
                                                    return 4;
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (ssDCval >= 108) {
                                    if (mu_mv_mag_sq < 96) {
                                        if (refDCval < 109) {
                                            if (tcor < 96) {
                                                if (MVDiff < 10) {
                                                    if (Cs < 1565) {
                                                        return 4;
                                                    }
                                                    else if (Cs >= 1565) {
                                                        return 8;
                                                    }
                                                }
                                                else if (MVDiff >= 10) {
                                                    return 4;
                                                }
                                            }
                                            else if (tcor >= 96) {
                                                return 8;
                                            }
                                        }
                                        else if (refDCval >= 109) {
                                            return 4;
                                        }
                                    }
                                    else if (mu_mv_mag_sq >= 96) {
                                        return 8;
                                    }
                                }
                            }
                            else if (TSC >= 6) {
                                if (TSCindex < 8) {
                                    if (mu_mv_mag_sq < 93) {
                                        if (ssDCval < 105) {
                                            if (ssDCval < 54) {
                                                if (Cs < 3395) {
                                                    if (Rs < 3455) {
                                                        if (MVDiff < 24) {
                                                            return 4;
                                                        }
                                                        else if (MVDiff >= 24) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (Rs >= 3455) {
                                                        return 8;
                                                    }
                                                }
                                                else if (Cs >= 3395) {
                                                    return 4;
                                                }
                                            }
                                            else if (ssDCval >= 54) {
                                                if (Cs < 2056) {
                                                    return 4;
                                                }
                                                else if (Cs >= 2056) {
                                                    if (Cs < 3595) {
                                                        return 8;
                                                    }
                                                    else if (Cs >= 3595) {
                                                        if (diffAFD < 1) {
                                                            return 4;
                                                        }
                                                        else if (diffAFD >= 1) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (ssDCval >= 105) {
                                            if (Cs < 2107) {
                                                if (diffRsCsdiff < -44) {
                                                    if (Cs < 2039) {
                                                        return 4;
                                                    }
                                                    else if (Cs >= 2039) {
                                                        return 8;
                                                    }
                                                }
                                                else if (diffRsCsdiff >= -44) {
                                                    return 4;
                                                }
                                            }
                                            else if (Cs >= 2107) {
                                                if (Rs < 4084) {
                                                    if (Cs < 2689) {
                                                        return 2;
                                                    }
                                                    else if (Cs >= 2689) {
                                                        return 8;
                                                    }
                                                }
                                                else if (Rs >= 4084) {
                                                    if (SC < 7200) {
                                                        if (diffMVdiffVal < -6) {
                                                            if (Cs < 2405) {
                                                                return 2;
                                                            }
                                                            else if (Cs >= 2405) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (diffMVdiffVal >= -6) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (SC >= 7200) {
                                                        if (Rs < 4855) {
                                                            return 2;
                                                        }
                                                        else if (Rs >= 4855) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (mu_mv_mag_sq >= 93) {
                                        if (Cs < 3669) {
                                            return 8;
                                        }
                                        else if (Cs >= 3669) {
                                            return 4;
                                        }
                                    }
                                }
                                else if (TSCindex >= 8) {
                                    if (Rs < 3612) {
                                        if (CsDiff < 4) {
                                            if (ssDCval < 50) {
                                                if (TSCindex < 9) {
                                                    if (Rs < 3263) {
                                                        return 2;
                                                    }
                                                    else if (Rs >= 3263) {
                                                        return 8;
                                                    }
                                                }
                                                else if (TSCindex >= 9) {
                                                    return 2;
                                                }
                                            }
                                            else if (ssDCval >= 50) {
                                                return 2;
                                            }
                                        }
                                        else if (CsDiff >= 4) {
                                            if (posBalance < 49) {
                                                if (Rs < 3452) {
                                                    if (gchDC < 1) {
                                                        return 4;
                                                    }
                                                    else if (gchDC >= 1) {
                                                        if (Rs < 3261) {
                                                            return 2;
                                                        }
                                                        else if (Rs >= 3261) {
                                                            if (AFD < 10) {
                                                                return 2;
                                                            }
                                                            else if (AFD >= 10) {
                                                                if (Rs < 3400) {
                                                                    return 4;
                                                                }
                                                                else if (Rs >= 3400) {
                                                                    return 2;
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (Rs >= 3452) {
                                                    return 2;
                                                }
                                            }
                                            else if (posBalance >= 49) {
                                                if (AFD < 10) {
                                                    return 8;
                                                }
                                                else if (AFD >= 10) {
                                                    if (Cs < 3580) {
                                                        if (diffAFD < 1) {
                                                            if (RsCsDiff < 144) {
                                                                return 8;
                                                            }
                                                            else if (RsCsDiff >= 144) {
                                                                return 2;
                                                            }
                                                        }
                                                        else if (diffAFD >= 1) {
                                                            if (Cs < 2796) {
                                                                return 4;
                                                            }
                                                            else if (Cs >= 2796) {
                                                                if (Rs < 3386) {
                                                                    return 8;
                                                                }
                                                                else if (Rs >= 3386) {
                                                                    return 2;
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (Cs >= 3580) {
                                                        if (Rs < 3547) {
                                                            return 2;
                                                        }
                                                        else if (Rs >= 3547) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (Rs >= 3612) {
                                        if (Cs < 2190) {
                                            if (Cs < 1683) {
                                                return 2;
                                            }
                                            else if (Cs >= 1683) {
                                                if (Rs < 3728) {
                                                    return 2;
                                                }
                                                else if (Rs >= 3728) {
                                                    return 4;
                                                }
                                            }
                                        }
                                        else if (Cs >= 2190) {
                                            if (ssDCval < 128) {
                                                if (MVDiff < 13) {
                                                    if (negBalance < 34) {
                                                        if (mu_mv_mag_sq < 83) {
                                                            return 2;
                                                        }
                                                        else if (mu_mv_mag_sq >= 83) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (negBalance >= 34) {
                                                        if (negBalance < 39) {
                                                            if (ssDCval < 125) {
                                                                if (mu_mv_mag_sq < 80) {
                                                                    return 8;
                                                                }
                                                                else if (mu_mv_mag_sq >= 80) {
                                                                    if (negBalance < 36) {
                                                                        if (diffRsCsdiff < -7) {
                                                                            return 4;
                                                                        }
                                                                        else if (diffRsCsdiff >= -7) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                    else if (negBalance >= 36) {
                                                                        return 4;
                                                                    }
                                                                }
                                                            }
                                                            else if (ssDCval >= 125) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (negBalance >= 39) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                                else if (MVDiff >= 13) {
                                                    if (AFD < 11) {
                                                        if (RsCsDiff < 22) {
                                                            if (Rs < 3656) {
                                                                return 4;
                                                            }
                                                            else if (Rs >= 3656) {
                                                                return 2;
                                                            }
                                                        }
                                                        else if (RsCsDiff >= 22) {
                                                            if (refDCval < 119) {
                                                                if (Rs < 4510) {
                                                                    if (tcor < 92) {
                                                                        if (Cs < 3519) {
                                                                            return 4;
                                                                        }
                                                                        else if (Cs >= 3519) {
                                                                            return 2;
                                                                        }
                                                                    }
                                                                    else if (tcor >= 92) {
                                                                        return 2;
                                                                    }
                                                                }
                                                                else if (Rs >= 4510) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (refDCval >= 119) {
                                                                if (AFD < 8) {
                                                                    return 4;
                                                                }
                                                                else if (AFD >= 8) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (AFD >= 11) {
                                                        if (AFD < 15) {
                                                            return 4;
                                                        }
                                                        else if (AFD >= 15) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (ssDCval >= 128) {
                                                if (SC < 6681) {
                                                    return 8;
                                                }
                                                else if (SC >= 6681) {
                                                    return 2;
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                        }
                        else if (TSC >= 10) {
                            if (ssDCval < 86) {
                                if (Rs < 3074) {
                                    return 8;
                                }
                                else if (Rs >= 3074) {
                                    if (posBalance < 59) {
                                        return 4;
                                    }
                                    else if (posBalance >= 59) {
                                        if (Cs < 2782) {
                                            return 8;
                                        }
                                        else if (Cs >= 2782) {
                                            return 4;
                                        }
                                    }
                                }
                            }
                            else if (ssDCval >= 86) {
                                if (Cs < 1606) {
                                    return 2;
                                }
                                else if (Cs >= 1606) {
                                    if (MVDiff < 10) {
                                        if (TSC < 11) {
                                            return 8;
                                        }
                                        else if (TSC >= 11) {
                                            if (RsDiff < 11) {
                                                return 1;
                                            }
                                            else if (RsDiff >= 11) {
                                                if (TSC < 12) {
                                                    return 4;
                                                }
                                                else if (TSC >= 12) {
                                                    return 8;
                                                }
                                            }
                                        }
                                    }
                                    else if (MVDiff >= 10) {
                                        return 8;
                                    }
                                }
                            }
                        }
                    }
                }
            }
            else if (SC >= 7556) {
                if (mu_mv_mag_sq < 70) {
                    if (TSC < 17) {
                        if (CsDiff < 11) {
                            if (ssDCval < 106) {
                                if (RsCsDiff < 46) {
                                    if (ssDCval < 57) {
                                        if (negBalance < 52) {
                                            if (mu_mv_mag_sq < 53) {
                                                if (Rs < 3926) {
                                                    if (diffRsCsdiff < 4) {
                                                        return 4;
                                                    }
                                                    else if (diffRsCsdiff >= 4) {
                                                        return 2;
                                                    }
                                                }
                                                else if (Rs >= 3926) {
                                                    return 8;
                                                }
                                            }
                                            else if (mu_mv_mag_sq >= 53) {
                                                return 4;
                                            }
                                        }
                                        else if (negBalance >= 52) {
                                            return 2;
                                        }
                                    }
                                    else if (ssDCval >= 57) {
                                        if (TSC < 7) {
                                            if (SC < 8165) {
                                                return 8;
                                            }
                                            else if (SC >= 8165) {
                                                return 4;
                                            }
                                        }
                                        else if (TSC >= 7) {
                                            if (CsDiff < 5) {
                                                return 4;
                                            }
                                            else if (CsDiff >= 5) {
                                                return 8;
                                            }
                                        }
                                    }
                                }
                                else if (RsCsDiff >= 46) {
                                    if (ssDCval < 61) {
                                        if (MVDiff < 32) {
                                            if (TSC < 7) {
                                                return 8;
                                            }
                                            else if (TSC >= 7) {
                                                return 4;
                                            }
                                        }
                                        else if (MVDiff >= 32) {
                                            if (Rs < 4321) {
                                                return 8;
                                            }
                                            else if (Rs >= 4321) {
                                                return 4;
                                            }
                                        }
                                    }
                                    else if (ssDCval >= 61) {
                                        if (mcTcor < 91) {
                                            if (SC < 9444) {
                                                if (MVDiff < 13) {
                                                    if (Cs < 2950) {
                                                        return 2;
                                                    }
                                                    else if (Cs >= 2950) {
                                                        return 4;
                                                    }
                                                }
                                                else if (MVDiff >= 13) {
                                                    return 4;
                                                }
                                            }
                                            else if (SC >= 9444) {
                                                return 8;
                                            }
                                        }
                                        else if (mcTcor >= 91) {
                                            if (RsCsDiff < 67) {
                                                return 4;
                                            }
                                            else if (RsCsDiff >= 67) {
                                                if (Rs < 5057) {
                                                    return 8;
                                                }
                                                else if (Rs >= 5057) {
                                                    if (MVDiff < 12) {
                                                        return 8;
                                                    }
                                                    else if (MVDiff >= 12) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                            else if (ssDCval >= 106) {
                                if (tcor < 81) {
                                    if (SC < 7772) {
                                        if (refDCval < 157) {
                                            if (RsDiff < 14) {
                                                if (CsDiff < 7) {
                                                    return 4;
                                                }
                                                else if (CsDiff >= 7) {
                                                    return 1;
                                                }
                                            }
                                            else if (RsDiff >= 14) {
                                                return 4;
                                            }
                                        }
                                        else if (refDCval >= 157) {
                                            return 2;
                                        }
                                    }
                                    else if (SC >= 7772) {
                                        return 2;
                                    }
                                }
                                else if (tcor >= 81) {
                                    if (ssDCval < 107) {
                                        return 2;
                                    }
                                    else if (ssDCval >= 107) {
                                        if (AFD < 10) {
                                            return 8;
                                        }
                                        else if (AFD >= 10) {
                                            if (Cs < 3245) {
                                                return 4;
                                            }
                                            else if (Cs >= 3245) {
                                                return 8;
                                            }
                                        }
                                    }
                                }
                            }
                        }
                        else if (CsDiff >= 11) {
                            if (refDCval < 153) {
                                if (Cs < 3050) {
                                    if (Rs < 5816) {
                                        if (Rs < 4926) {
                                            return 2;
                                        }
                                        else if (Rs >= 4926) {
                                            if (RsDiff < 30) {
                                                return 4;
                                            }
                                            else if (RsDiff >= 30) {
                                                return 2;
                                            }
                                        }
                                    }
                                    else if (Rs >= 5816) {
                                        return 1;
                                    }
                                }
                                else if (Cs >= 3050) {
                                    if (AFD < 31) {
                                        if (Rs < 7572) {
                                            if (tcor < 68) {
                                                if (mu_mv_mag_sq < 3) {
                                                    return 1;
                                                }
                                                else if (mu_mv_mag_sq >= 3) {
                                                    if (Cs < 3481) {
                                                        return 1;
                                                    }
                                                    else if (Cs >= 3481) {
                                                        if (RsDiff < 16) {
                                                            return 1;
                                                        }
                                                        else if (RsDiff >= 16) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (tcor >= 68) {
                                                if (CsDiff < 17) {
                                                    if (RsCsDiff < 703) {
                                                        if (mcTcor < 82) {
                                                            if (mcTcor < 80) {
                                                                return 8;
                                                            }
                                                            else if (mcTcor >= 80) {
                                                                if (mu_mv_mag_sq < 55) {
                                                                    return 4;
                                                                }
                                                                else if (mu_mv_mag_sq >= 55) {
                                                                    if (SC < 9222) {
                                                                        return 4;
                                                                    }
                                                                    else if (SC >= 9222) {
                                                                        return 8;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (mcTcor >= 82) {
                                                            if (Rs < 6336) {
                                                                if (TSC < 15) {
                                                                    if (tcor < 77) {
                                                                        if (SC < 8624) {
                                                                            return 4;
                                                                        }
                                                                        else if (SC >= 8624) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                    else if (tcor >= 77) {
                                                                        return 8;
                                                                    }
                                                                }
                                                                else if (TSC >= 15) {
                                                                    if (Rs < 5182) {
                                                                        if (ssDCval < 108) {
                                                                            return 4;
                                                                        }
                                                                        else if (ssDCval >= 108) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                    else if (Rs >= 5182) {
                                                                        return 4;
                                                                    }
                                                                }
                                                            }
                                                            else if (Rs >= 6336) {
                                                                if (RsDiff < 14) {
                                                                    return 4;
                                                                }
                                                                else if (RsDiff >= 14) {
                                                                    if (Rs < 6861) {
                                                                        return 1;
                                                                    }
                                                                    else if (Rs >= 6861) {
                                                                        return 8;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (RsCsDiff >= 703) {
                                                        return 4;
                                                    }
                                                }
                                                else if (CsDiff >= 17) {
                                                    if (tcor < 84) {
                                                        if (mcTcor < 92) {
                                                            return 8;
                                                        }
                                                        else if (mcTcor >= 92) {
                                                            return 1;
                                                        }
                                                    }
                                                    else if (tcor >= 84) {
                                                        if (mcTcor < 93) {
                                                            if (diffRsCsdiff < -561) {
                                                                return 4;
                                                            }
                                                            else if (diffRsCsdiff >= -561) {
                                                                if (posBalance < 60) {
                                                                    return 8;
                                                                }
                                                                else if (posBalance >= 60) {
                                                                    return 1;
                                                                }
                                                            }
                                                        }
                                                        else if (mcTcor >= 93) {
                                                            if (SC < 12550) {
                                                                return 1;
                                                            }
                                                            else if (SC >= 12550) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (Rs >= 7572) {
                                            if (mu_mv_mag_sq < 64) {
                                                return 4;
                                            }
                                            else if (mu_mv_mag_sq >= 64) {
                                                return 8;
                                            }
                                        }
                                    }
                                    else if (AFD >= 31) {
                                        return 2;
                                    }
                                }
                            }
                            else if (refDCval >= 153) {
                                if (posBalance < 59) {
                                    if (Cs < 4064) {
                                        return 4;
                                    }
                                    else if (Cs >= 4064) {
                                        if (ssDCval < 166) {
                                            return 8;
                                        }
                                        else if (ssDCval >= 166) {
                                            return 4;
                                        }
                                    }
                                }
                                else if (posBalance >= 59) {
                                    if (mu_mv_mag_sq < 3) {
                                        if (CsDiff < 14) {
                                            if (SC < 9234) {
                                                return 2;
                                            }
                                            else if (SC >= 9234) {
                                                return 4;
                                            }
                                        }
                                        else if (CsDiff >= 14) {
                                            return 2;
                                        }
                                    }
                                    else if (mu_mv_mag_sq >= 3) {
                                        return 4;
                                    }
                                }
                            }
                        }
                    }
                    else if (TSC >= 17) {
                        if (Rs < 6890) {
                            if (MVDiff < 13) {
                                return 8;
                            }
                            else if (MVDiff >= 13) {
                                if (ssDCval < 161) {
                                    if (refDCval < 135) {
                                        if (Cs < 3963) {
                                            if (SC < 8964) {
                                                return 8;
                                            }
                                            else if (SC >= 8964) {
                                                return 4;
                                            }
                                        }
                                        else if (Cs >= 3963) {
                                            return 8;
                                        }
                                    }
                                    else if (refDCval >= 135) {
                                        if (Rs < 5310) {
                                            return 8;
                                        }
                                        else if (Rs >= 5310) {
                                            if (SC < 10065) {
                                                return 4;
                                            }
                                            else if (SC >= 10065) {
                                                if (posBalance < 54) {
                                                    if (negBalance < 56) {
                                                        return 4;
                                                    }
                                                    else if (negBalance >= 56) {
                                                        return 8;
                                                    }
                                                }
                                                else if (posBalance >= 54) {
                                                    return 8;
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (ssDCval >= 161) {
                                    return 2;
                                }
                            }
                        }
                        else if (Rs >= 6890) {
                            if (mcTcor < 85) {
                                if (ssDCval < 88) {
                                    return 4;
                                }
                                else if (ssDCval >= 88) {
                                    if (TSC < 25) {
                                        if (Rs < 7429) {
                                            if (Rs < 7199) {
                                                return 8;
                                            }
                                            else if (Rs >= 7199) {
                                                return 2;
                                            }
                                        }
                                        else if (Rs >= 7429) {
                                            if (Cs < 4439) {
                                                return 4;
                                            }
                                            else if (Cs >= 4439) {
                                                return 8;
                                            }
                                        }
                                    }
                                    else if (TSC >= 25) {
                                        if (SC < 24882) {
                                            if (AFD < 39) {
                                                return 4;
                                            }
                                            else if (AFD >= 39) {
                                                return 8;
                                            }
                                        }
                                        else if (SC >= 24882) {
                                            return 8;
                                        }
                                    }
                                }
                            }
                            else if (mcTcor >= 85) {
                                if (AFD < 35) {
                                    if (MVDiff < 22) {
                                        if (refDCval < 98) {
                                            if (Rs < 8326) {
                                                return 4;
                                            }
                                            else if (Rs >= 8326) {
                                                return 2;
                                            }
                                        }
                                        else if (refDCval >= 98) {
                                            if (Cs < 6100) {
                                                if (mcTcor < 87) {
                                                    if (Cs < 4998) {
                                                        if (Rs < 7713) {
                                                            return 2;
                                                        }
                                                        else if (Rs >= 7713) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (Cs >= 4998) {
                                                        return 8;
                                                    }
                                                }
                                                else if (mcTcor >= 87) {
                                                    return 4;
                                                }
                                            }
                                            else if (Cs >= 6100) {
                                                if (ssDCval < 143) {
                                                    return 8;
                                                }
                                                else if (ssDCval >= 143) {
                                                    if (mcTcor < 87) {
                                                        return 8;
                                                    }
                                                    else if (mcTcor >= 87) {
                                                        if (posBalance < 50) {
                                                            return 8;
                                                        }
                                                        else if (posBalance >= 50) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (MVDiff >= 22) {
                                        if (Cs < 6510) {
                                            return 2;
                                        }
                                        else if (Cs >= 6510) {
                                            if (Rs < 7639) {
                                                return 8;
                                            }
                                            else if (Rs >= 7639) {
                                                return 4;
                                            }
                                        }
                                    }
                                }
                                else if (AFD >= 35) {
                                    return 2;
                                }
                            }
                        }
                    }
                }
                else if (mu_mv_mag_sq >= 70) {
                    if (TSC < 11) {
                        if (Rs < 4639) {
                            if (gchDC < 1) {
                                if (TSC < 7) {
                                    return 8;
                                }
                                else if (TSC >= 7) {
                                    if (diffRsCsdiff < 30) {
                                        return 4;
                                    }
                                    else if (diffRsCsdiff >= 30) {
                                        return 8;
                                    }
                                }
                            }
                            else if (gchDC >= 1) {
                                return 8;
                            }
                        }
                        else if (Rs >= 4639) {
                            if (SC < 8675) {
                                if (Rs < 5107) {
                                    if (MVDiff < 6) {
                                        return 8;
                                    }
                                    else if (MVDiff >= 6) {
                                        if (mu_mv_mag_sq < 75) {
                                            return 8;
                                        }
                                        else if (mu_mv_mag_sq >= 75) {
                                            if (ssDCval < 111) {
                                                if (SC < 7987) {
                                                    return 2;
                                                }
                                                else if (SC >= 7987) {
                                                    return 8;
                                                }
                                            }
                                            else if (ssDCval >= 111) {
                                                return 2;
                                            }
                                        }
                                    }
                                }
                                else if (Rs >= 5107) {
                                    if (SC < 8422) {
                                        return 8;
                                    }
                                    else if (SC >= 8422) {
                                        return 2;
                                    }
                                }
                            }
                            else if (SC >= 8675) {
                                if (Rs < 7615) {
                                    return 8;
                                }
                                else if (Rs >= 7615) {
                                    return 4;
                                }
                            }
                        }
                    }
                    else if (TSC >= 11) {
                        if (ssDCval < 136) {
                            if (CsDiff < 14) {
                                if (Cs < 3247) {
                                    if (Rs < 5956) {
                                        return 2;
                                    }
                                    else if (Rs >= 5956) {
                                        return 1;
                                    }
                                }
                                else if (Cs >= 3247) {
                                    if (TSC < 13) {
                                        return 8;
                                    }
                                    else if (TSC >= 13) {
                                        if (Rs < 4627) {
                                            if (Cs < 4019) {
                                                return 4;
                                            }
                                            else if (Cs >= 4019) {
                                                return 8;
                                            }
                                        }
                                        else if (Rs >= 4627) {
                                            return 8;
                                        }
                                    }
                                }
                            }
                            else if (CsDiff >= 14) {
                                return 8;
                            }
                        }
                        else if (ssDCval >= 136) {
                            if (posBalance < 59) {
                                if (Rs < 4593) {
                                    return 1;
                                }
                                else if (Rs >= 4593) {
                                    return 8;
                                }
                            }
                            else if (posBalance >= 59) {
                                return 8;
                            }
                        }
                    }
                }
            }
        }
        return 4;
    }
    mfxU16 AGOPSelect13(mfxU32 Rs, mfxU32 CsDiff, mfxU32 TSC, mfxI32 diffAFD, mfxU32 Scindex, mfxU32 RsDiff, mfxI32 diffMVdiffVal, mfxU32 RsCsDiff, mfxU32 TSCindex, mfxI16 tcor, mfxU32 refDCval, mfxU32 MVDiff, mfxU32 AFD, mfxI32 diffTSC, mfxI32 diffRsCsdiff, mfxU32 posBalance, mfxU32 SC, mfxU32 mu_mv_mag_sq, mfxU32 negBalance, mfxI16 mcTcor, mfxU32 gchDC, mfxU32 Cs, mfxU32 ssDCval) {
        if (mu_mv_mag_sq < 89) {
            if (posBalance < 5) {
                if (Cs < 1825) {
                    if (SC < 375) {
                        if (mu_mv_mag_sq < 22) {
                            if (posBalance < 4) {
                                if (negBalance < 102) {
                                    if (negBalance < 33) {
                                        if (diffMVdiffVal < 31) {
                                            if (tcor < 67) {
                                                return 4;
                                            }
                                            else if (tcor >= 67) {
                                                if (negBalance < 10) {
                                                    return 8;
                                                }
                                                else if (negBalance >= 10) {
                                                    return 4;
                                                }
                                            }
                                        }
                                        else if (diffMVdiffVal >= 31) {
                                            return 8;
                                        }
                                    }
                                    else if (negBalance >= 33) {
                                        if (mcTcor < 78) {
                                            return 2;
                                        }
                                        else if (mcTcor >= 78) {
                                            return 4;
                                        }
                                    }
                                }
                                else if (negBalance >= 102) {
                                    if (AFD < 4) {
                                        return 1;
                                    }
                                    else if (AFD >= 4) {
                                        return 8;
                                    }
                                }
                            }
                            else if (posBalance >= 4) {
                                return 2;
                            }
                        }
                        else if (mu_mv_mag_sq >= 22) {
                            if (tcor < 128) {
                                if (negBalance < 114) {
                                    return 8;
                                }
                                else if (negBalance >= 114) {
                                    return 2;
                                }
                            }
                            else if (tcor >= 128) {
                                if (TSC < 5) {
                                    if (ssDCval < 22) {
                                        return 8;
                                    }
                                    else if (ssDCval >= 22) {
                                        return 2;
                                    }
                                }
                                else if (TSC >= 5) {
                                    return 1;
                                }
                            }
                        }
                    }
                    else if (SC >= 375) {
                        if (TSC < 6) {
                            return 8;
                        }
                        else if (TSC >= 6) {
                            if (negBalance < 115) {
                                return 8;
                            }
                            else if (negBalance >= 115) {
                                if (Rs < 794) {
                                    return 2;
                                }
                                else if (Rs >= 794) {
                                    return 4;
                                }
                            }
                        }
                    }
                }
                else if (Cs >= 1825) {
                    if (SC < 9327) {
                        if (SC < 3818) {
                            if (SC < 3696) {
                                return 8;
                            }
                            else if (SC >= 3696) {
                                return 1;
                            }
                        }
                        else if (SC >= 3818) {
                            if (negBalance < 67) {
                                if (AFD < 1) {
                                    if (SC < 7755) {
                                        if (SC < 7709) {
                                            return 8;
                                        }
                                        else if (SC >= 7709) {
                                            return 1;
                                        }
                                    }
                                    else if (SC >= 7755) {
                                        return 8;
                                    }
                                }
                                else if (AFD >= 1) {
                                    return 8;
                                }
                            }
                            else if (negBalance >= 67) {
                                return 2;
                            }
                        }
                    }
                    else if (SC >= 9327) {
                        if (Rs < 5376) {
                            return 1;
                        }
                        else if (Rs >= 5376) {
                            return 8;
                        }
                    }
                }
            }
            else if (posBalance >= 5) {
                if (SC < 7536) {
                    if (AFD < 12) {
                        if (Rs < 2991) {
                            if (refDCval < 106) {
                                if (Rs < 544) {
                                    if (AFD < 7) {
                                        if (ssDCval < 85) {
                                            if (Cs < 381) {
                                                if (mu_mv_mag_sq < 29) {
                                                    if (MVDiff < 41) {
                                                        if (SC < 496) {
                                                            if (diffMVdiffVal < -17) {
                                                                if (Cs < 59) {
                                                                    if (posBalance < 95) {
                                                                        return 4;
                                                                    }
                                                                    else if (posBalance >= 95) {
                                                                        return 8;
                                                                    }
                                                                }
                                                                else if (Cs >= 59) {
                                                                    return 1;
                                                                }
                                                            }
                                                            else if (diffMVdiffVal >= -17) {
                                                                if (tcor < 79) {
                                                                    return 2;
                                                                }
                                                                else if (tcor >= 79) {
                                                                    if (tcor < 82) {
                                                                        if (gchDC < 1) {
                                                                            return 2;
                                                                        }
                                                                        else if (gchDC >= 1) {
                                                                            if (SC < 88) {
                                                                                return 4;
                                                                            }
                                                                            else if (SC >= 88) {
                                                                                if (Rs < 186) {
                                                                                    return 8;
                                                                                }
                                                                                else if (Rs >= 186) {
                                                                                    return 4;
                                                                                }
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (tcor >= 82) {
                                                                        if (refDCval < 33) {
                                                                            if (Rs < 64) {
                                                                                return 2;
                                                                            }
                                                                            else if (Rs >= 64) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                        else if (refDCval >= 33) {
                                                                            return 2;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (SC >= 496) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (MVDiff >= 41) {
                                                        if (refDCval < 22) {
                                                            if (MVDiff < 86) {
                                                                if (negBalance < 16) {
                                                                    if (MVDiff < 83) {
                                                                        if (diffMVdiffVal < -267) {
                                                                            return 8;
                                                                        }
                                                                        else if (diffMVdiffVal >= -267) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                    else if (MVDiff >= 83) {
                                                                        return 8;
                                                                    }
                                                                }
                                                                else if (negBalance >= 16) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (MVDiff >= 86) {
                                                                if (mcTcor < 48) {
                                                                    if (mcTcor < 39) {
                                                                        if (posBalance < 35) {
                                                                            return 1;
                                                                        }
                                                                        else if (posBalance >= 35) {
                                                                            if (negBalance < 1) {
                                                                                return 2;
                                                                            }
                                                                            else if (negBalance >= 1) {
                                                                                if (diffMVdiffVal < 247) {
                                                                                    return 4;
                                                                                }
                                                                                else if (diffMVdiffVal >= 247) {
                                                                                    return 2;
                                                                                }
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (mcTcor >= 39) {
                                                                        if (diffAFD < -2) {
                                                                            return 2;
                                                                        }
                                                                        else if (diffAFD >= -2) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                }
                                                                else if (mcTcor >= 48) {
                                                                    if (mu_mv_mag_sq < 20) {
                                                                        if (posBalance < 17) {
                                                                            return 8;
                                                                        }
                                                                        else if (posBalance >= 17) {
                                                                            if (posBalance < 88) {
                                                                                return 4;
                                                                            }
                                                                            else if (posBalance >= 88) {
                                                                                if (diffAFD < -1) {
                                                                                    if (MVDiff < 210) {
                                                                                        return 8;
                                                                                    }
                                                                                    else if (MVDiff >= 210) {
                                                                                        return 4;
                                                                                    }
                                                                                }
                                                                                else if (diffAFD >= -1) {
                                                                                    return 8;
                                                                                }
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (mu_mv_mag_sq >= 20) {
                                                                        if (ssDCval < 22) {
                                                                            return 2;
                                                                        }
                                                                        else if (ssDCval >= 22) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (refDCval >= 22) {
                                                            if (tcor < 82) {
                                                                if (Cs < 235) {
                                                                    if (negBalance < 2) {
                                                                        if (posBalance < 109) {
                                                                            return 8;
                                                                        }
                                                                        else if (posBalance >= 109) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                    else if (negBalance >= 2) {
                                                                        return 1;
                                                                    }
                                                                }
                                                                else if (Cs >= 235) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (tcor >= 82) {
                                                                if (Rs < 21) {
                                                                    return 2;
                                                                }
                                                                else if (Rs >= 21) {
                                                                    return 1;
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (mu_mv_mag_sq >= 29) {
                                                    if (mcTcor < 86) {
                                                        if (posBalance < 66) {
                                                            if (SC < 727) {
                                                                if (tcor < 47) {
                                                                    if (mu_mv_mag_sq < 39) {
                                                                        if (posBalance < 28) {
                                                                            if (tcor < 43) {
                                                                                if (tcor < 37) {
                                                                                    return 8;
                                                                                }
                                                                                else if (tcor >= 37) {
                                                                                    return 1;
                                                                                }
                                                                            }
                                                                            else if (tcor >= 43) {
                                                                                return 8;
                                                                            }
                                                                        }
                                                                        else if (posBalance >= 28) {
                                                                            if (mcTcor < 43) {
                                                                                return 8;
                                                                            }
                                                                            else if (mcTcor >= 43) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (mu_mv_mag_sq >= 39) {
                                                                        return 1;
                                                                    }
                                                                }
                                                                else if (tcor >= 47) {
                                                                    if (negBalance < 4) {
                                                                        if (SC < 9) {
                                                                            return 4;
                                                                        }
                                                                        else if (SC >= 9) {
                                                                            return 2;
                                                                        }
                                                                    }
                                                                    else if (negBalance >= 4) {
                                                                        if (ssDCval < 19) {
                                                                            return 4;
                                                                        }
                                                                        else if (ssDCval >= 19) {
                                                                            if (Rs < 379) {
                                                                                if (ssDCval < 28) {
                                                                                    return 8;
                                                                                }
                                                                                else if (ssDCval >= 28) {
                                                                                    if (Cs < 155) {
                                                                                        return 4;
                                                                                    }
                                                                                    else if (Cs >= 155) {
                                                                                        return 8;
                                                                                    }
                                                                                }
                                                                            }
                                                                            else if (Rs >= 379) {
                                                                                if (Rs < 387) {
                                                                                    return 4;
                                                                                }
                                                                                else if (Rs >= 387) {
                                                                                    return 8;
                                                                                }
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                            else if (SC >= 727) {
                                                                if (Rs < 403) {
                                                                    if (Rs < 395) {
                                                                        return 8;
                                                                    }
                                                                    else if (Rs >= 395) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (Rs >= 403) {
                                                                    if (MVDiff < 25) {
                                                                        if (mu_mv_mag_sq < 74) {
                                                                            if (Rs < 458) {
                                                                                return 4;
                                                                            }
                                                                            else if (Rs >= 458) {
                                                                                return 1;
                                                                            }
                                                                        }
                                                                        else if (mu_mv_mag_sq >= 74) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                    else if (MVDiff >= 25) {
                                                                        if (diffMVdiffVal < 7) {
                                                                            if (diffRsCsdiff < 2) {
                                                                                return 1;
                                                                            }
                                                                            else if (diffRsCsdiff >= 2) {
                                                                                return 8;
                                                                            }
                                                                        }
                                                                        else if (diffMVdiffVal >= 7) {
                                                                            if (SC < 766) {
                                                                                return 1;
                                                                            }
                                                                            else if (SC >= 766) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (posBalance >= 66) {
                                                            if (diffAFD < -2) {
                                                                return 1;
                                                            }
                                                            else if (diffAFD >= -2) {
                                                                if (tcor < 66) {
                                                                    if (MVDiff < 56) {
                                                                        return 4;
                                                                    }
                                                                    else if (MVDiff >= 56) {
                                                                        return 1;
                                                                    }
                                                                }
                                                                else if (tcor >= 66) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (mcTcor >= 86) {
                                                        if (posBalance < 99) {
                                                            if (Rs < 400) {
                                                                if (tcor < 94) {
                                                                    if (MVDiff < 19) {
                                                                        return 8;
                                                                    }
                                                                    else if (MVDiff >= 19) {
                                                                        if (posBalance < 61) {
                                                                            return 2;
                                                                        }
                                                                        else if (posBalance >= 61) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                }
                                                                else if (tcor >= 94) {
                                                                    if (TSC < 1) {
                                                                        if (mu_mv_mag_sq < 72) {
                                                                            return 8;
                                                                        }
                                                                        else if (mu_mv_mag_sq >= 72) {
                                                                            if (ssDCval < 23) {
                                                                                return 8;
                                                                            }
                                                                            else if (ssDCval >= 23) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (TSC >= 1) {
                                                                        if (Cs < 325) {
                                                                            if (SC < 558) {
                                                                                return 4;
                                                                            }
                                                                            else if (SC >= 558) {
                                                                                if (SC < 587) {
                                                                                    return 8;
                                                                                }
                                                                                else if (SC >= 587) {
                                                                                    return 4;
                                                                                }
                                                                            }
                                                                        }
                                                                        else if (Cs >= 325) {
                                                                            return 2;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                            else if (Rs >= 400) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (posBalance >= 99) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (Cs >= 381) {
                                                if (mu_mv_mag_sq < 66) {
                                                    if (gchDC < 3) {
                                                        if (negBalance < 47) {
                                                            if (ssDCval < 39) {
                                                                if (diffAFD < 1) {
                                                                    if (SC < 868) {
                                                                        if (SC < 867) {
                                                                            return 2;
                                                                        }
                                                                        else if (SC >= 867) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                    else if (SC >= 868) {
                                                                        return 2;
                                                                    }
                                                                }
                                                                else if (diffAFD >= 1) {
                                                                    if (MVDiff < 29) {
                                                                        return 4;
                                                                    }
                                                                    else if (MVDiff >= 29) {
                                                                        return 2;
                                                                    }
                                                                }
                                                            }
                                                            else if (ssDCval >= 39) {
                                                                if (Rs < 278) {
                                                                    return 2;
                                                                }
                                                                else if (Rs >= 278) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                        else if (negBalance >= 47) {
                                                            if (Rs < 460) {
                                                                return 4;
                                                            }
                                                            else if (Rs >= 460) {
                                                                if (Cs < 411) {
                                                                    return 8;
                                                                }
                                                                else if (Cs >= 411) {
                                                                    return 2;
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (gchDC >= 3) {
                                                        if (Rs < 353) {
                                                            if (Rs < 322) {
                                                                return 2;
                                                            }
                                                            else if (Rs >= 322) {
                                                                return 1;
                                                            }
                                                        }
                                                        else if (Rs >= 353) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                                else if (mu_mv_mag_sq >= 66) {
                                                    if (SC < 893) {
                                                        return 4;
                                                    }
                                                    else if (SC >= 893) {
                                                        if (Rs < 494) {
                                                            if (AFD < 4) {
                                                                return 4;
                                                            }
                                                            else if (AFD >= 4) {
                                                                return 2;
                                                            }
                                                        }
                                                        else if (Rs >= 494) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (ssDCval >= 85) {
                                            if (ssDCval < 88) {
                                                return 2;
                                            }
                                            else if (ssDCval >= 88) {
                                                if (SC < 839) {
                                                    if (MVDiff < 5) {
                                                        return 2;
                                                    }
                                                    else if (MVDiff >= 5) {
                                                        return 4;
                                                    }
                                                }
                                                else if (SC >= 839) {
                                                    return 4;
                                                }
                                            }
                                        }
                                    }
                                    else if (AFD >= 7) {
                                        if (TSC < 6) {
                                            if (Cs < 282) {
                                                if (AFD < 9) {
                                                    if (RsDiff < 2) {
                                                        if (negBalance < 2) {
                                                            return 4;
                                                        }
                                                        else if (negBalance >= 2) {
                                                            if (ssDCval < 63) {
                                                                if (negBalance < 82) {
                                                                    return 1;
                                                                }
                                                                else if (negBalance >= 82) {
                                                                    return 2;
                                                                }
                                                            }
                                                            else if (ssDCval >= 63) {
                                                                return 2;
                                                            }
                                                        }
                                                    }
                                                    else if (RsDiff >= 2) {
                                                        return 1;
                                                    }
                                                }
                                                else if (AFD >= 9) {
                                                    if (MVDiff < 68) {
                                                        return 2;
                                                    }
                                                    else if (MVDiff >= 68) {
                                                        return 1;
                                                    }
                                                }
                                            }
                                            else if (Cs >= 282) {
                                                if (Cs < 730) {
                                                    if (negBalance < 50) {
                                                        if (SC < 618) {
                                                            return 2;
                                                        }
                                                        else if (SC >= 618) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (negBalance >= 50) {
                                                        return 8;
                                                    }
                                                }
                                                else if (Cs >= 730) {
                                                    return 2;
                                                }
                                            }
                                        }
                                        else if (TSC >= 6) {
                                            if (gchDC < 4) {
                                                return 1;
                                            }
                                            else if (gchDC >= 4) {
                                                if (gchDC < 6) {
                                                    return 4;
                                                }
                                                else if (gchDC >= 6) {
                                                    if (ssDCval < 48) {
                                                        return 1;
                                                    }
                                                    else if (ssDCval >= 48) {
                                                        return 2;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (Rs >= 544) {
                                    if (TSC < 7) {
                                        if (SC < 1128) {
                                            if (SC < 889) {
                                                if (ssDCval < 83) {
                                                    if (negBalance < 47) {
                                                        if (Rs < 632) {
                                                            return 2;
                                                        }
                                                        else if (Rs >= 632) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (negBalance >= 47) {
                                                        return 8;
                                                    }
                                                }
                                                else if (ssDCval >= 83) {
                                                    if (AFD < 11) {
                                                        return 1;
                                                    }
                                                    else if (AFD >= 11) {
                                                        return 2;
                                                    }
                                                }
                                            }
                                            else if (SC >= 889) {
                                                if (TSC < 6) {
                                                    if (mu_mv_mag_sq < 7) {
                                                        if (diffRsCsdiff < 2) {
                                                            return 8;
                                                        }
                                                        else if (diffRsCsdiff >= 2) {
                                                            if (SC < 1121) {
                                                                return 4;
                                                            }
                                                            else if (SC >= 1121) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                    else if (mu_mv_mag_sq >= 7) {
                                                        if (AFD < 5) {
                                                            if (Rs < 678) {
                                                                return 8;
                                                            }
                                                            else if (Rs >= 678) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (AFD >= 5) {
                                                            if (RsCsDiff < 33) {
                                                                if (ssDCval < 59) {
                                                                    if (Cs < 465) {
                                                                        return 4;
                                                                    }
                                                                    else if (Cs >= 465) {
                                                                        return 2;
                                                                    }
                                                                }
                                                                else if (ssDCval >= 59) {
                                                                    return 2;
                                                                }
                                                            }
                                                            else if (RsCsDiff >= 33) {
                                                                return 1;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (TSC >= 6) {
                                                    if (ssDCval < 83) {
                                                        if (SC < 1019) {
                                                            return 8;
                                                        }
                                                        else if (SC >= 1019) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (ssDCval >= 83) {
                                                        return 1;
                                                    }
                                                }
                                            }
                                        }
                                        else if (SC >= 1128) {
                                            if (mu_mv_mag_sq < 3) {
                                                if (Rs < 1132) {
                                                    if (refDCval < 97) {
                                                        if (Cs < 631) {
                                                            if (RsDiff < 4) {
                                                                if (posBalance < 50) {
                                                                    return 4;
                                                                }
                                                                else if (posBalance >= 50) {
                                                                    if (RsCsDiff < 7) {
                                                                        return 8;
                                                                    }
                                                                    else if (RsCsDiff >= 7) {
                                                                        if (SC < 1309) {
                                                                            if (MVDiff < 22) {
                                                                                if (negBalance < 53) {
                                                                                    return 8;
                                                                                }
                                                                                else if (negBalance >= 53) {
                                                                                    if (Rs < 613) {
                                                                                        return 8;
                                                                                    }
                                                                                    else if (Rs >= 613) {
                                                                                        if (Rs < 655) {
                                                                                            return 4;
                                                                                        }
                                                                                        else if (Rs >= 655) {
                                                                                            return 8;
                                                                                        }
                                                                                    }
                                                                                }
                                                                            }
                                                                            else if (MVDiff >= 22) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                        else if (SC >= 1309) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                            else if (RsDiff >= 4) {
                                                                return 2;
                                                            }
                                                        }
                                                        else if (Cs >= 631) {
                                                            if (posBalance < 57) {
                                                                if (SC < 1674) {
                                                                    return 2;
                                                                }
                                                                else if (SC >= 1674) {
                                                                    if (Cs < 878) {
                                                                        return 4;
                                                                    }
                                                                    else if (Cs >= 878) {
                                                                        return 8;
                                                                    }
                                                                }
                                                            }
                                                            else if (posBalance >= 57) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                    else if (refDCval >= 97) {
                                                        return 4;
                                                    }
                                                }
                                                else if (Rs >= 1132) {
                                                    if (posBalance < 48) {
                                                        if (ssDCval < 80) {
                                                            if (diffAFD < -2) {
                                                                if (Cs < 432) {
                                                                    return 2;
                                                                }
                                                                else if (Cs >= 432) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (diffAFD >= -2) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (ssDCval >= 80) {
                                                            return 2;
                                                        }
                                                    }
                                                    else if (posBalance >= 48) {
                                                        if (tcor < 81) {
                                                            if (negBalance < 53) {
                                                                if (diffRsCsdiff < 4) {
                                                                    if (Rs < 1538) {
                                                                        return 2;
                                                                    }
                                                                    else if (Rs >= 1538) {
                                                                        if (Rs < 1945) {
                                                                            return 1;
                                                                        }
                                                                        else if (Rs >= 1945) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                }
                                                                else if (diffRsCsdiff >= 4) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (negBalance >= 53) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (tcor >= 81) {
                                                            if (refDCval < 56) {
                                                                if (Cs < 944) {
                                                                    return 1;
                                                                }
                                                                else if (Cs >= 944) {
                                                                    if (RsDiff < 3) {
                                                                        return 8;
                                                                    }
                                                                    else if (RsDiff >= 3) {
                                                                        if (MVDiff < 6) {
                                                                            if (SC < 2458) {
                                                                                return 1;
                                                                            }
                                                                            else if (SC >= 2458) {
                                                                                return 8;
                                                                            }
                                                                        }
                                                                        else if (MVDiff >= 6) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                            else if (refDCval >= 56) {
                                                                if (Cs < 944) {
                                                                    return 4;
                                                                }
                                                                else if (Cs >= 944) {
                                                                    return 1;
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                            else if (mu_mv_mag_sq >= 3) {
                                                if (refDCval < 72) {
                                                    if (ssDCval < 54) {
                                                        if (MVDiff < 22) {
                                                            if (negBalance < 44) {
                                                                if (SC < 2027) {
                                                                    if (TSCindex < 4) {
                                                                        if (Rs < 661) {
                                                                            return 4;
                                                                        }
                                                                        else if (Rs >= 661) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                    else if (TSCindex >= 4) {
                                                                        if (RsCsDiff < 39) {
                                                                            if (diffAFD < -1) {
                                                                                if (ssDCval < 46) {
                                                                                    return 1;
                                                                                }
                                                                                else if (ssDCval >= 46) {
                                                                                    if (mu_mv_mag_sq < 40) {
                                                                                        return 1;
                                                                                    }
                                                                                    else if (mu_mv_mag_sq >= 40) {
                                                                                        return 4;
                                                                                    }
                                                                                }
                                                                            }
                                                                            else if (diffAFD >= -1) {
                                                                                if (AFD < 6) {
                                                                                    return 4;
                                                                                }
                                                                                else if (AFD >= 6) {
                                                                                    if (SC < 1678) {
                                                                                        if (Cs < 692) {
                                                                                            return 4;
                                                                                        }
                                                                                        else if (Cs >= 692) {
                                                                                            return 1;
                                                                                        }
                                                                                    }
                                                                                    else if (SC >= 1678) {
                                                                                        return 4;
                                                                                    }
                                                                                }
                                                                            }
                                                                        }
                                                                        else if (RsCsDiff >= 39) {
                                                                            if (SC < 1823) {
                                                                                return 1;
                                                                            }
                                                                            else if (SC >= 1823) {
                                                                                return 2;
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                                else if (SC >= 2027) {
                                                                    if (Cs < 1182) {
                                                                        if (negBalance < 19) {
                                                                            return 4;
                                                                        }
                                                                        else if (negBalance >= 19) {
                                                                            if (mcTcor < 94) {
                                                                                if (refDCval < 50) {
                                                                                    if (Rs < 1959) {
                                                                                        if (posBalance < 59) {
                                                                                            return 1;
                                                                                        }
                                                                                        else if (posBalance >= 59) {
                                                                                            if (Rs < 1279) {
                                                                                                return 8;
                                                                                            }
                                                                                            else if (Rs >= 1279) {
                                                                                                return 1;
                                                                                            }
                                                                                        }
                                                                                    }
                                                                                    else if (Rs >= 1959) {
                                                                                        return 4;
                                                                                    }
                                                                                }
                                                                                else if (refDCval >= 50) {
                                                                                    if (negBalance < 40) {
                                                                                        if (Rs < 1264) {
                                                                                            if (diffMVdiffVal < 2) {
                                                                                                return 4;
                                                                                            }
                                                                                            else if (diffMVdiffVal >= 2) {
                                                                                                return 1;
                                                                                            }
                                                                                        }
                                                                                        else if (Rs >= 1264) {
                                                                                            if (diffMVdiffVal < 1) {
                                                                                                if (Cs < 882) {
                                                                                                    return 4;
                                                                                                }
                                                                                                else if (Cs >= 882) {
                                                                                                    if (Rs < 1579) {
                                                                                                        if (TSCindex < 6) {
                                                                                                            return 1;
                                                                                                        }
                                                                                                        else if (TSCindex >= 6) {
                                                                                                            if (Rs < 1313) {
                                                                                                                return 4;
                                                                                                            }
                                                                                                            else if (Rs >= 1313) {
                                                                                                                return 1;
                                                                                                            }
                                                                                                        }
                                                                                                    }
                                                                                                    else if (Rs >= 1579) {
                                                                                                        return 4;
                                                                                                    }
                                                                                                }
                                                                                            }
                                                                                            else if (diffMVdiffVal >= 1) {
                                                                                                if (Rs < 1304) {
                                                                                                    return 1;
                                                                                                }
                                                                                                else if (Rs >= 1304) {
                                                                                                    if (posBalance < 52) {
                                                                                                        return 4;
                                                                                                    }
                                                                                                    else if (posBalance >= 52) {
                                                                                                        if (Cs < 1063) {
                                                                                                            return 8;
                                                                                                        }
                                                                                                        else if (Cs >= 1063) {
                                                                                                            return 4;
                                                                                                        }
                                                                                                    }
                                                                                                }
                                                                                            }
                                                                                        }
                                                                                    }
                                                                                    else if (negBalance >= 40) {
                                                                                        return 1;
                                                                                    }
                                                                                }
                                                                            }
                                                                            else if (mcTcor >= 94) {
                                                                                if (MVDiff < 5) {
                                                                                    if (posBalance < 50) {
                                                                                        return 1;
                                                                                    }
                                                                                    else if (posBalance >= 50) {
                                                                                        if (Rs < 1376) {
                                                                                            if (Rs < 1315) {
                                                                                                return 1;
                                                                                            }
                                                                                            else if (Rs >= 1315) {
                                                                                                return 4;
                                                                                            }
                                                                                        }
                                                                                        else if (Rs >= 1376) {
                                                                                            return 1;
                                                                                        }
                                                                                    }
                                                                                }
                                                                                else if (MVDiff >= 5) {
                                                                                    return 1;
                                                                                }
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (Cs >= 1182) {
                                                                        if (SC < 4697) {
                                                                            return 4;
                                                                        }
                                                                        else if (SC >= 4697) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                            else if (negBalance >= 44) {
                                                                if (diffMVdiffVal < 2) {
                                                                    if (diffRsCsdiff < 6) {
                                                                        if (mu_mv_mag_sq < 63) {
                                                                            if (negBalance < 45) {
                                                                                return 1;
                                                                            }
                                                                            else if (negBalance >= 45) {
                                                                                if (mcTcor < 82) {
                                                                                    return 4;
                                                                                }
                                                                                else if (mcTcor >= 82) {
                                                                                    if (Rs < 967) {
                                                                                        return 1;
                                                                                    }
                                                                                    else if (Rs >= 967) {
                                                                                        if (diffAFD < 1) {
                                                                                            if (negBalance < 46) {
                                                                                                if (Rs < 1497) {
                                                                                                    return 4;
                                                                                                }
                                                                                                else if (Rs >= 1497) {
                                                                                                    return 1;
                                                                                                }
                                                                                            }
                                                                                            else if (negBalance >= 46) {
                                                                                                return 4;
                                                                                            }
                                                                                        }
                                                                                        else if (diffAFD >= 1) {
                                                                                            return 1;
                                                                                        }
                                                                                    }
                                                                                }
                                                                            }
                                                                        }
                                                                        else if (mu_mv_mag_sq >= 63) {
                                                                            if (ssDCval < 53) {
                                                                                if (diffRsCsdiff < -5) {
                                                                                    return 8;
                                                                                }
                                                                                else if (diffRsCsdiff >= -5) {
                                                                                    if (SC < 3623) {
                                                                                        if (Cs < 1565) {
                                                                                            if (diffMVdiffVal < 0) {
                                                                                                return 4;
                                                                                            }
                                                                                            else if (diffMVdiffVal >= 0) {
                                                                                                if (Rs < 1845) {
                                                                                                    return 4;
                                                                                                }
                                                                                                else if (Rs >= 1845) {
                                                                                                    return 8;
                                                                                                }
                                                                                            }
                                                                                        }
                                                                                        else if (Cs >= 1565) {
                                                                                            if (Cs < 1573) {
                                                                                                return 1;
                                                                                            }
                                                                                            else if (Cs >= 1573) {
                                                                                                return 4;
                                                                                            }
                                                                                        }
                                                                                    }
                                                                                    else if (SC >= 3623) {
                                                                                        return 1;
                                                                                    }
                                                                                }
                                                                            }
                                                                            else if (ssDCval >= 53) {
                                                                                return 1;
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (diffRsCsdiff >= 6) {
                                                                        if (Rs < 1361) {
                                                                            if (Rs < 1177) {
                                                                                return 4;
                                                                            }
                                                                            else if (Rs >= 1177) {
                                                                                return 1;
                                                                            }
                                                                        }
                                                                        else if (Rs >= 1361) {
                                                                            if (Rs < 1828) {
                                                                                if (diffTSC < 1) {
                                                                                    return 8;
                                                                                }
                                                                                else if (diffTSC >= 1) {
                                                                                    return 1;
                                                                                }
                                                                            }
                                                                            else if (Rs >= 1828) {
                                                                                return 1;
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                                else if (diffMVdiffVal >= 2) {
                                                                    if (diffMVdiffVal < 11) {
                                                                        if (Cs < 1397) {
                                                                            if (mu_mv_mag_sq < 36) {
                                                                                return 4;
                                                                            }
                                                                            else if (mu_mv_mag_sq >= 36) {
                                                                                if (MVDiff < 8) {
                                                                                    return 1;
                                                                                }
                                                                                else if (MVDiff >= 8) {
                                                                                    if (tcor < 96) {
                                                                                        return 8;
                                                                                    }
                                                                                    else if (tcor >= 96) {
                                                                                        return 4;
                                                                                    }
                                                                                }
                                                                            }
                                                                        }
                                                                        else if (Cs >= 1397) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                    else if (diffMVdiffVal >= 11) {
                                                                        if (AFD < 6) {
                                                                            return 4;
                                                                        }
                                                                        else if (AFD >= 6) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (MVDiff >= 22) {
                                                            if (RsCsDiff < 9) {
                                                                if (TSCindex < 6) {
                                                                    if (Cs < 1545) {
                                                                        if (ssDCval < 39) {
                                                                            if (posBalance < 55) {
                                                                                if (SC < 1544) {
                                                                                    return 4;
                                                                                }
                                                                                else if (SC >= 1544) {
                                                                                    if (Rs < 845) {
                                                                                        return 1;
                                                                                    }
                                                                                    else if (Rs >= 845) {
                                                                                        return 2;
                                                                                    }
                                                                                }
                                                                            }
                                                                            else if (posBalance >= 55) {
                                                                                return 2;
                                                                            }
                                                                        }
                                                                        else if (ssDCval >= 39) {
                                                                            if (refDCval < 47) {
                                                                                return 4;
                                                                            }
                                                                            else if (refDCval >= 47) {
                                                                                return 8;
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (Cs >= 1545) {
                                                                        return 1;
                                                                    }
                                                                }
                                                                else if (TSCindex >= 6) {
                                                                    if (mu_mv_mag_sq < 41) {
                                                                        if (tcor < 75) {
                                                                            if (MVDiff < 48) {
                                                                                if (SC < 1420) {
                                                                                    if (SC < 1215) {
                                                                                        if (Cs < 503) {
                                                                                            if (TSC < 6) {
                                                                                                return 4;
                                                                                            }
                                                                                            else if (TSC >= 6) {
                                                                                                if (Rs < 703) {
                                                                                                    return 1;
                                                                                                }
                                                                                                else if (Rs >= 703) {
                                                                                                    return 4;
                                                                                                }
                                                                                            }
                                                                                        }
                                                                                        else if (Cs >= 503) {
                                                                                            return 1;
                                                                                        }
                                                                                    }
                                                                                    else if (SC >= 1215) {
                                                                                        return 1;
                                                                                    }
                                                                                }
                                                                                else if (SC >= 1420) {
                                                                                    if (SC < 1518) {
                                                                                        return 4;
                                                                                    }
                                                                                    else if (SC >= 1518) {
                                                                                        return 1;
                                                                                    }
                                                                                }
                                                                            }
                                                                            else if (MVDiff >= 48) {
                                                                                if (Rs < 652) {
                                                                                    return 8;
                                                                                }
                                                                                else if (Rs >= 652) {
                                                                                    return 4;
                                                                                }
                                                                            }
                                                                        }
                                                                        else if (tcor >= 75) {
                                                                            if (SC < 1235) {
                                                                                if (Cs < 480) {
                                                                                    return 4;
                                                                                }
                                                                                else if (Cs >= 480) {
                                                                                    return 1;
                                                                                }
                                                                            }
                                                                            else if (SC >= 1235) {
                                                                                if (mu_mv_mag_sq < 32) {
                                                                                    if (AFD < 7) {
                                                                                        return 1;
                                                                                    }
                                                                                    else if (AFD >= 7) {
                                                                                        return 4;
                                                                                    }
                                                                                }
                                                                                else if (mu_mv_mag_sq >= 32) {
                                                                                    return 1;
                                                                                }
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (mu_mv_mag_sq >= 41) {
                                                                        if (mu_mv_mag_sq < 53) {
                                                                            if (MVDiff < 29) {
                                                                                if (tcor < 89) {
                                                                                    if (tcor < 76) {
                                                                                        if (CsDiff < 2) {
                                                                                            return 4;
                                                                                        }
                                                                                        else if (CsDiff >= 2) {
                                                                                            return 1;
                                                                                        }
                                                                                    }
                                                                                    else if (tcor >= 76) {
                                                                                        return 1;
                                                                                    }
                                                                                }
                                                                                else if (tcor >= 89) {
                                                                                    return 4;
                                                                                }
                                                                            }
                                                                            else if (MVDiff >= 29) {
                                                                                if (negBalance < 47) {
                                                                                    return 2;
                                                                                }
                                                                                else if (negBalance >= 47) {
                                                                                    return 4;
                                                                                }
                                                                            }
                                                                        }
                                                                        else if (mu_mv_mag_sq >= 53) {
                                                                            if (diffMVdiffVal < 13) {
                                                                                if (diffRsCsdiff < -7) {
                                                                                    if (Cs < 1020) {
                                                                                        return 1;
                                                                                    }
                                                                                    else if (Cs >= 1020) {
                                                                                        return 4;
                                                                                    }
                                                                                }
                                                                                else if (diffRsCsdiff >= -7) {
                                                                                    return 4;
                                                                                }
                                                                            }
                                                                            else if (diffMVdiffVal >= 13) {
                                                                                return 1;
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                            else if (RsCsDiff >= 9) {
                                                                if (CsDiff < 3) {
                                                                    if (TSCindex < 7) {
                                                                        if (tcor < 89) {
                                                                            return 8;
                                                                        }
                                                                        else if (tcor >= 89) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                    else if (TSCindex >= 7) {
                                                                        return 1;
                                                                    }
                                                                }
                                                                else if (CsDiff >= 3) {
                                                                    if (TSC < 6) {
                                                                        if (Cs < 2279) {
                                                                            if (Rs < 1346) {
                                                                                return 2;
                                                                            }
                                                                            else if (Rs >= 1346) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                        else if (Cs >= 2279) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                    else if (TSC >= 6) {
                                                                        if (MVDiff < 29) {
                                                                            return 1;
                                                                        }
                                                                        else if (MVDiff >= 29) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (ssDCval >= 54) {
                                                        if (TSC < 5) {
                                                            if (ssDCval < 65) {
                                                                if (RsCsDiff < 19) {
                                                                    if (posBalance < 63) {
                                                                        if (RsDiff < 3) {
                                                                            if (mu_mv_mag_sq < 76) {
                                                                                if (SC < 1754) {
                                                                                    return 8;
                                                                                }
                                                                                else if (SC >= 1754) {
                                                                                    if (Rs < 826) {
                                                                                        return 8;
                                                                                    }
                                                                                    else if (Rs >= 826) {
                                                                                        if (Rs < 1048) {
                                                                                            return 1;
                                                                                        }
                                                                                        else if (Rs >= 1048) {
                                                                                            return 4;
                                                                                        }
                                                                                    }
                                                                                }
                                                                            }
                                                                            else if (mu_mv_mag_sq >= 76) {
                                                                                return 8;
                                                                            }
                                                                        }
                                                                        else if (RsDiff >= 3) {
                                                                            if (mu_mv_mag_sq < 75) {
                                                                                if (mu_mv_mag_sq < 59) {
                                                                                    if (SC < 2116) {
                                                                                        return 8;
                                                                                    }
                                                                                    else if (SC >= 2116) {
                                                                                        return 4;
                                                                                    }
                                                                                }
                                                                                else if (mu_mv_mag_sq >= 59) {
                                                                                    return 1;
                                                                                }
                                                                            }
                                                                            else if (mu_mv_mag_sq >= 75) {
                                                                                return 8;
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (posBalance >= 63) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (RsCsDiff >= 19) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (ssDCval >= 65) {
                                                                if (mu_mv_mag_sq < 61) {
                                                                    return 2;
                                                                }
                                                                else if (mu_mv_mag_sq >= 61) {
                                                                    if (Rs < 1696) {
                                                                        return 4;
                                                                    }
                                                                    else if (Rs >= 1696) {
                                                                        return 1;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (TSC >= 5) {
                                                            if (refDCval < 70) {
                                                                if (Rs < 2064) {
                                                                    if (tcor < 93) {
                                                                        if (SC < 2208) {
                                                                            if (negBalance < 51) {
                                                                                if (posBalance < 58) {
                                                                                    if (AFD < 7) {
                                                                                        if (ssDCval < 56) {
                                                                                            return 4;
                                                                                        }
                                                                                        else if (ssDCval >= 56) {
                                                                                            return 1;
                                                                                        }
                                                                                    }
                                                                                    else if (AFD >= 7) {
                                                                                        return 4;
                                                                                    }
                                                                                }
                                                                                else if (posBalance >= 58) {
                                                                                    if (TSC < 6) {
                                                                                        if (Cs < 747) {
                                                                                            return 4;
                                                                                        }
                                                                                        else if (Cs >= 747) {
                                                                                            return 1;
                                                                                        }
                                                                                    }
                                                                                    else if (TSC >= 6) {
                                                                                        return 2;
                                                                                    }
                                                                                }
                                                                            }
                                                                            else if (negBalance >= 51) {
                                                                                if (Rs < 839) {
                                                                                    return 2;
                                                                                }
                                                                                else if (Rs >= 839) {
                                                                                    return 8;
                                                                                }
                                                                            }
                                                                        }
                                                                        else if (SC >= 2208) {
                                                                            if (AFD < 8) {
                                                                                if (Cs < 1014) {
                                                                                    return 1;
                                                                                }
                                                                                else if (Cs >= 1014) {
                                                                                    return 8;
                                                                                }
                                                                            }
                                                                            else if (AFD >= 8) {
                                                                                if (posBalance < 54) {
                                                                                    if (CsDiff < 6) {
                                                                                        return 4;
                                                                                    }
                                                                                    else if (CsDiff >= 6) {
                                                                                        return 1;
                                                                                    }
                                                                                }
                                                                                else if (posBalance >= 54) {
                                                                                    return 1;
                                                                                }
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (tcor >= 93) {
                                                                        if (posBalance < 47) {
                                                                            return 4;
                                                                        }
                                                                        else if (posBalance >= 47) {
                                                                            if (MVDiff < 3) {
                                                                                return 2;
                                                                            }
                                                                            else if (MVDiff >= 3) {
                                                                                if (SC < 1586) {
                                                                                    return 8;
                                                                                }
                                                                                else if (SC >= 1586) {
                                                                                    return 1;
                                                                                }
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                                else if (Rs >= 2064) {
                                                                    if (Cs < 1463) {
                                                                        return 2;
                                                                    }
                                                                    else if (Cs >= 1463) {
                                                                        return 8;
                                                                    }
                                                                }
                                                            }
                                                            else if (refDCval >= 70) {
                                                                if (Rs < 1210) {
                                                                    return 4;
                                                                }
                                                                else if (Rs >= 1210) {
                                                                    return 1;
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (refDCval >= 72) {
                                                    if (mu_mv_mag_sq < 47) {
                                                        if (gchDC < 3) {
                                                            if (RsDiff < 3) {
                                                                return 4;
                                                            }
                                                            else if (RsDiff >= 3) {
                                                                if (negBalance < 52) {
                                                                    return 2;
                                                                }
                                                                else if (negBalance >= 52) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                        else if (gchDC >= 3) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (mu_mv_mag_sq >= 47) {
                                                        if (SC < 4248) {
                                                            if (mu_mv_mag_sq < 68) {
                                                                if (Rs < 835) {
                                                                    return 8;
                                                                }
                                                                else if (Rs >= 835) {
                                                                    if (posBalance < 36) {
                                                                        return 8;
                                                                    }
                                                                    else if (posBalance >= 36) {
                                                                        if (Cs < 1710) {
                                                                            return 4;
                                                                        }
                                                                        else if (Cs >= 1710) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                            else if (mu_mv_mag_sq >= 68) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (SC >= 4248) {
                                                            if (SC < 4745) {
                                                                if (CsDiff < 5) {
                                                                    return 1;
                                                                }
                                                                else if (CsDiff >= 5) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (SC >= 4745) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (TSC >= 7) {
                                        if (Rs < 2572) {
                                            if (negBalance < 47) {
                                                if (negBalance < 44) {
                                                    if (gchDC < 3) {
                                                        if (diffRsCsdiff < 30) {
                                                            if (CsDiff < 9) {
                                                                if (gchDC < 1) {
                                                                    if (SC < 4956) {
                                                                        if (mu_mv_mag_sq < 78) {
                                                                            return 8;
                                                                        }
                                                                        else if (mu_mv_mag_sq >= 78) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                    else if (SC >= 4956) {
                                                                        return 2;
                                                                    }
                                                                }
                                                                else if (gchDC >= 1) {
                                                                    if (diffRsCsdiff < 4) {
                                                                        if (MVDiff < 13) {
                                                                            if (ssDCval < 84) {
                                                                                return 8;
                                                                            }
                                                                            else if (ssDCval >= 84) {
                                                                                if (negBalance < 31) {
                                                                                    return 8;
                                                                                }
                                                                                else if (negBalance >= 31) {
                                                                                    return 1;
                                                                                }
                                                                            }
                                                                        }
                                                                        else if (MVDiff >= 13) {
                                                                            if (ssDCval < 95) {
                                                                                return 8;
                                                                            }
                                                                            else if (ssDCval >= 95) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (diffRsCsdiff >= 4) {
                                                                        if (Rs < 1702) {
                                                                            return 8;
                                                                        }
                                                                        else if (Rs >= 1702) {
                                                                            if (Cs < 2319) {
                                                                                if (mu_mv_mag_sq < 74) {
                                                                                    if (refDCval < 92) {
                                                                                        return 1;
                                                                                    }
                                                                                    else if (refDCval >= 92) {
                                                                                        return 8;
                                                                                    }
                                                                                }
                                                                                else if (mu_mv_mag_sq >= 74) {
                                                                                    return 8;
                                                                                }
                                                                            }
                                                                            else if (Cs >= 2319) {
                                                                                return 8;
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                            else if (CsDiff >= 9) {
                                                                if (Cs < 2599) {
                                                                    return 1;
                                                                }
                                                                else if (Cs >= 2599) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                        else if (diffRsCsdiff >= 30) {
                                                            if (Rs < 2250) {
                                                                return 4;
                                                            }
                                                            else if (Rs >= 2250) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                    else if (gchDC >= 3) {
                                                        if (mcTcor < 87) {
                                                            if (Rs < 1257) {
                                                                return 1;
                                                            }
                                                            else if (Rs >= 1257) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (mcTcor >= 87) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                                else if (negBalance >= 44) {
                                                    if (Cs < 1383) {
                                                        return 1;
                                                    }
                                                    else if (Cs >= 1383) {
                                                        if (diffRsCsdiff < 1) {
                                                            if (AFD < 11) {
                                                                if (Rs < 2197) {
                                                                    return 2;
                                                                }
                                                                else if (Rs >= 2197) {
                                                                    return 1;
                                                                }
                                                            }
                                                            else if (AFD >= 11) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (diffRsCsdiff >= 1) {
                                                            if (SC < 3650) {
                                                                return 2;
                                                            }
                                                            else if (SC >= 3650) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                            else if (negBalance >= 47) {
                                                if (SC < 1265) {
                                                    return 1;
                                                }
                                                else if (SC >= 1265) {
                                                    if (refDCval < 98) {
                                                        if (posBalance < 61) {
                                                            if (tcor < 78) {
                                                                if (Cs < 1213) {
                                                                    if (Rs < 969) {
                                                                        return 8;
                                                                    }
                                                                    else if (Rs >= 969) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (Cs >= 1213) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (tcor >= 78) {
                                                                if (refDCval < 86) {
                                                                    if (Cs < 1781) {
                                                                        if (gchDC < 3) {
                                                                            if (mcTcor < 88) {
                                                                                return 8;
                                                                            }
                                                                            else if (mcTcor >= 88) {
                                                                                if (Rs < 1862) {
                                                                                    if (negBalance < 56) {
                                                                                        if (tcor < 85) {
                                                                                            return 4;
                                                                                        }
                                                                                        else if (tcor >= 85) {
                                                                                            return 2;
                                                                                        }
                                                                                    }
                                                                                    else if (negBalance >= 56) {
                                                                                        return 1;
                                                                                    }
                                                                                }
                                                                                else if (Rs >= 1862) {
                                                                                    return 1;
                                                                                }
                                                                            }
                                                                        }
                                                                        else if (gchDC >= 3) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                    else if (Cs >= 1781) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (refDCval >= 86) {
                                                                    if (MVDiff < 10) {
                                                                        if (mcTcor < 87) {
                                                                            if (ssDCval < 97) {
                                                                                if (mu_mv_mag_sq < 61) {
                                                                                    if (RsDiff < 6) {
                                                                                        if (posBalance < 59) {
                                                                                            if (Cs < 1654) {
                                                                                                return 2;
                                                                                            }
                                                                                            else if (Cs >= 1654) {
                                                                                                return 8;
                                                                                            }
                                                                                        }
                                                                                        else if (posBalance >= 59) {
                                                                                            return 2;
                                                                                        }
                                                                                    }
                                                                                    else if (RsDiff >= 6) {
                                                                                        return 2;
                                                                                    }
                                                                                }
                                                                                else if (mu_mv_mag_sq >= 61) {
                                                                                    if (Rs < 2195) {
                                                                                        if (diffAFD < -1) {
                                                                                            return 2;
                                                                                        }
                                                                                        else if (diffAFD >= -1) {
                                                                                            return 8;
                                                                                        }
                                                                                    }
                                                                                    else if (Rs >= 2195) {
                                                                                        if (refDCval < 93) {
                                                                                            return 8;
                                                                                        }
                                                                                        else if (refDCval >= 93) {
                                                                                            return 2;
                                                                                        }
                                                                                    }
                                                                                }
                                                                            }
                                                                            else if (ssDCval >= 97) {
                                                                                return 2;
                                                                            }
                                                                        }
                                                                        else if (mcTcor >= 87) {
                                                                            if (gchDC < 3) {
                                                                                if (refDCval < 87) {
                                                                                    if (Cs < 1883) {
                                                                                        return 2;
                                                                                    }
                                                                                    else if (Cs >= 1883) {
                                                                                        return 8;
                                                                                    }
                                                                                }
                                                                                else if (refDCval >= 87) {
                                                                                    return 2;
                                                                                }
                                                                            }
                                                                            else if (gchDC >= 3) {
                                                                                if (Rs < 2130) {
                                                                                    return 4;
                                                                                }
                                                                                else if (Rs >= 2130) {
                                                                                    return 8;
                                                                                }
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (MVDiff >= 10) {
                                                                        if (tcor < 92) {
                                                                            return 8;
                                                                        }
                                                                        else if (tcor >= 92) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (posBalance >= 61) {
                                                            if (mu_mv_mag_sq < 70) {
                                                                if (refDCval < 89) {
                                                                    if (ssDCval < 86) {
                                                                        if (RsDiff < 7) {
                                                                            if (Cs < 841) {
                                                                                return 1;
                                                                            }
                                                                            else if (Cs >= 841) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                        else if (RsDiff >= 7) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                    else if (ssDCval >= 86) {
                                                                        return 8;
                                                                    }
                                                                }
                                                                else if (refDCval >= 89) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (mu_mv_mag_sq >= 70) {
                                                                if (RsDiff < 5) {
                                                                    return 2;
                                                                }
                                                                else if (RsDiff >= 5) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (refDCval >= 98) {
                                                        if (mcTcor < 84) {
                                                            return 2;
                                                        }
                                                        else if (mcTcor >= 84) {
                                                            if (posBalance < 58) {
                                                                if (Cs < 1288) {
                                                                    return 2;
                                                                }
                                                                else if (Cs >= 1288) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (posBalance >= 58) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (Rs >= 2572) {
                                            if (ssDCval < 87) {
                                                if (Cs < 2409) {
                                                    if (SC < 4821) {
                                                        if (Cs < 1372) {
                                                            return 2;
                                                        }
                                                        else if (Cs >= 1372) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (SC >= 4821) {
                                                        return 4;
                                                    }
                                                }
                                                else if (Cs >= 2409) {
                                                    if (Rs < 2855) {
                                                        return 2;
                                                    }
                                                    else if (Rs >= 2855) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                            else if (ssDCval >= 87) {
                                                if (posBalance < 48) {
                                                    if (Cs < 2217) {
                                                        return 2;
                                                    }
                                                    else if (Cs >= 2217) {
                                                        return 8;
                                                    }
                                                }
                                                else if (posBalance >= 48) {
                                                    if (SC < 5231) {
                                                        return 1;
                                                    }
                                                    else if (SC >= 5231) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                            else if (refDCval >= 106) {
                                if (ssDCval < 119) {
                                    if (AFD < 10) {
                                        if (Cs < 7) {
                                            if (Rs < 804) {
                                                return 1;
                                            }
                                            else if (Rs >= 804) {
                                                return 8;
                                            }
                                        }
                                        else if (Cs >= 7) {
                                            if (Rs < 2723) {
                                                if (SC < 764) {
                                                    if (MVDiff < 8) {
                                                        return 8;
                                                    }
                                                    else if (MVDiff >= 8) {
                                                        if (refDCval < 110) {
                                                            if (tcor < 86) {
                                                                return 8;
                                                            }
                                                            else if (tcor >= 86) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (refDCval >= 110) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                                else if (SC >= 764) {
                                                    if (posBalance < 32) {
                                                        if (mu_mv_mag_sq < 83) {
                                                            if (Cs < 1201) {
                                                                return 8;
                                                            }
                                                            else if (Cs >= 1201) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (mu_mv_mag_sq >= 83) {
                                                            if (SC < 1162) {
                                                                return 2;
                                                            }
                                                            else if (SC >= 1162) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                    else if (posBalance >= 32) {
                                                        if (RsDiff < 6) {
                                                            if (ssDCval < 118) {
                                                                return 8;
                                                            }
                                                            else if (ssDCval >= 118) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (RsDiff >= 6) {
                                                            if (negBalance < 51) {
                                                                if (TSC < 9) {
                                                                    return 8;
                                                                }
                                                                else if (TSC >= 9) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (negBalance >= 51) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                            else if (Rs >= 2723) {
                                                if (refDCval < 113) {
                                                    return 4;
                                                }
                                                else if (refDCval >= 113) {
                                                    if (RsDiff < 6) {
                                                        return 8;
                                                    }
                                                    else if (RsDiff >= 6) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (AFD >= 10) {
                                        if (mu_mv_mag_sq < 73) {
                                            if (SC < 794) {
                                                return 2;
                                            }
                                            else if (SC >= 794) {
                                                if (diffMVdiffVal < -8) {
                                                    return 2;
                                                }
                                                else if (diffMVdiffVal >= -8) {
                                                    if (ssDCval < 116) {
                                                        if (posBalance < 43) {
                                                            return 8;
                                                        }
                                                        else if (posBalance >= 43) {
                                                            if (negBalance < 50) {
                                                                if (Cs < 2650) {
                                                                    return 8;
                                                                }
                                                                else if (Cs >= 2650) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (negBalance >= 50) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                    else if (ssDCval >= 116) {
                                                        if (mu_mv_mag_sq < 54) {
                                                            return 4;
                                                        }
                                                        else if (mu_mv_mag_sq >= 54) {
                                                            if (AFD < 11) {
                                                                return 4;
                                                            }
                                                            else if (AFD >= 11) {
                                                                if (SC < 4452) {
                                                                    return 8;
                                                                }
                                                                else if (SC >= 4452) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (mu_mv_mag_sq >= 73) {
                                            if (negBalance < 51) {
                                                return 8;
                                            }
                                            else if (negBalance >= 51) {
                                                if (posBalance < 59) {
                                                    if (diffAFD < 1) {
                                                        return 8;
                                                    }
                                                    else if (diffAFD >= 1) {
                                                        return 4;
                                                    }
                                                }
                                                else if (posBalance >= 59) {
                                                    return 4;
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (ssDCval >= 119) {
                                    if (mcTcor < 94) {
                                        if (refDCval < 126) {
                                            if (mu_mv_mag_sq < 51) {
                                                if (negBalance < 52) {
                                                    if (Cs < 1916) {
                                                        return 4;
                                                    }
                                                    else if (Cs >= 1916) {
                                                        if (ssDCval < 122) {
                                                            return 4;
                                                        }
                                                        else if (ssDCval >= 122) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                                else if (negBalance >= 52) {
                                                    return 4;
                                                }
                                            }
                                            else if (mu_mv_mag_sq >= 51) {
                                                if (mcTcor < 89) {
                                                    if (AFD < 10) {
                                                        return 4;
                                                    }
                                                    else if (AFD >= 10) {
                                                        if (SC < 4739) {
                                                            if (RsDiff < 7) {
                                                                return 8;
                                                            }
                                                            else if (RsDiff >= 7) {
                                                                if (Rs < 2116) {
                                                                    return 1;
                                                                }
                                                                else if (Rs >= 2116) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                        else if (SC >= 4739) {
                                                            if (SC < 5034) {
                                                                return 4;
                                                            }
                                                            else if (SC >= 5034) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (mcTcor >= 89) {
                                                    if (RsCsDiff < 83) {
                                                        if (SC < 2909) {
                                                            return 8;
                                                        }
                                                        else if (SC >= 2909) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (RsCsDiff >= 83) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                        }
                                        else if (refDCval >= 126) {
                                            if (refDCval < 140) {
                                                if (diffRsCsdiff < -66) {
                                                    if (Rs < 1332) {
                                                        return 4;
                                                    }
                                                    else if (Rs >= 1332) {
                                                        return 8;
                                                    }
                                                }
                                                else if (diffRsCsdiff >= -66) {
                                                    if (MVDiff < 4) {
                                                        if (Rs < 2276) {
                                                            return 4;
                                                        }
                                                        else if (Rs >= 2276) {
                                                            return 1;
                                                        }
                                                    }
                                                    else if (MVDiff >= 4) {
                                                        return 1;
                                                    }
                                                }
                                            }
                                            else if (refDCval >= 140) {
                                                if (AFD < 7) {
                                                    return 8;
                                                }
                                                else if (AFD >= 7) {
                                                    if (posBalance < 33) {
                                                        if (ssDCval < 145) {
                                                            return 8;
                                                        }
                                                        else if (ssDCval >= 145) {
                                                            return 1;
                                                        }
                                                    }
                                                    else if (posBalance >= 33) {
                                                        if (diffRsCsdiff < 72) {
                                                            return 4;
                                                        }
                                                        else if (diffRsCsdiff >= 72) {
                                                            if (Rs < 1242) {
                                                                return 8;
                                                            }
                                                            else if (Rs >= 1242) {
                                                                return 1;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (mcTcor >= 94) {
                                        if (MVDiff < 29) {
                                            if (tcor < 95) {
                                                if (Cs < 2436) {
                                                    if (mu_mv_mag_sq < 48) {
                                                        if (mu_mv_mag_sq < 7) {
                                                            if (AFD < 9) {
                                                                return 1;
                                                            }
                                                            else if (AFD >= 9) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (mu_mv_mag_sq >= 7) {
                                                            return 2;
                                                        }
                                                    }
                                                    else if (mu_mv_mag_sq >= 48) {
                                                        if (mcTcor < 95) {
                                                            if (Cs < 1445) {
                                                                if (Rs < 680) {
                                                                    return 8;
                                                                }
                                                                else if (Rs >= 680) {
                                                                    return 1;
                                                                }
                                                            }
                                                            else if (Cs >= 1445) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (mcTcor >= 95) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                                else if (Cs >= 2436) {
                                                    return 8;
                                                }
                                            }
                                            else if (tcor >= 95) {
                                                if (ssDCval < 134) {
                                                    if (mu_mv_mag_sq < 84) {
                                                        return 8;
                                                    }
                                                    else if (mu_mv_mag_sq >= 84) {
                                                        if (Cs < 874) {
                                                            return 2;
                                                        }
                                                        else if (Cs >= 874) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                                else if (ssDCval >= 134) {
                                                    if (posBalance < 36) {
                                                        if (CsDiff < 1) {
                                                            if (SC < 1031) {
                                                                return 2;
                                                            }
                                                            else if (SC >= 1031) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (CsDiff >= 1) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (posBalance >= 36) {
                                                        if (MVDiff < 3) {
                                                            return 8;
                                                        }
                                                        else if (MVDiff >= 3) {
                                                            if (ssDCval < 165) {
                                                                return 2;
                                                            }
                                                            else if (ssDCval >= 165) {
                                                                if (TSC < 2) {
                                                                    return 2;
                                                                }
                                                                else if (TSC >= 2) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (MVDiff >= 29) {
                                            if (tcor < 98) {
                                                if (posBalance < 49) {
                                                    return 1;
                                                }
                                                else if (posBalance >= 49) {
                                                    return 2;
                                                }
                                            }
                                            else if (tcor >= 98) {
                                                if (SC < 1143) {
                                                    return 1;
                                                }
                                                else if (SC >= 1143) {
                                                    return 8;
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                        }
                        else if (Rs >= 2991) {
                            if (refDCval < 108) {
                                if (TSCindex < 8) {
                                    if (refDCval < 54) {
                                        if (SC < 6792) {
                                            if (CsDiff < 3) {
                                                if (ssDCval < 50) {
                                                    if (mcTcor < 96) {
                                                        return 8;
                                                    }
                                                    else if (mcTcor >= 96) {
                                                        return 4;
                                                    }
                                                }
                                                else if (ssDCval >= 50) {
                                                    return 8;
                                                }
                                            }
                                            else if (CsDiff >= 3) {
                                                if (SC < 6783) {
                                                    return 4;
                                                }
                                                else if (SC >= 6783) {
                                                    return 8;
                                                }
                                            }
                                        }
                                        else if (SC >= 6792) {
                                            if (posBalance < 56) {
                                                if (Cs < 3349) {
                                                    if (MVDiff < 7) {
                                                        return 8;
                                                    }
                                                    else if (MVDiff >= 7) {
                                                        return 4;
                                                    }
                                                }
                                                else if (Cs >= 3349) {
                                                    return 4;
                                                }
                                            }
                                            else if (posBalance >= 56) {
                                                if (SC < 7073) {
                                                    return 8;
                                                }
                                                else if (SC >= 7073) {
                                                    return 4;
                                                }
                                            }
                                        }
                                    }
                                    else if (refDCval >= 54) {
                                        if (tcor < 94) {
                                            if (SC < 5884) {
                                                if (negBalance < 39) {
                                                    if (refDCval < 104) {
                                                        if (Cs < 2112) {
                                                            return 4;
                                                        }
                                                        else if (Cs >= 2112) {
                                                            return 2;
                                                        }
                                                    }
                                                    else if (refDCval >= 104) {
                                                        if (ssDCval < 107) {
                                                            if (Rs < 3194) {
                                                                if (posBalance < 47) {
                                                                    return 4;
                                                                }
                                                                else if (posBalance >= 47) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (Rs >= 3194) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (ssDCval >= 107) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                                else if (negBalance >= 39) {
                                                    return 4;
                                                }
                                            }
                                            else if (SC >= 5884) {
                                                if (refDCval < 55) {
                                                    if (Cs < 3522) {
                                                        return 4;
                                                    }
                                                    else if (Cs >= 3522) {
                                                        return 8;
                                                    }
                                                }
                                                else if (refDCval >= 55) {
                                                    return 8;
                                                }
                                            }
                                        }
                                        else if (tcor >= 94) {
                                            if (MVDiff < 10) {
                                                return 8;
                                            }
                                            else if (MVDiff >= 10) {
                                                if (posBalance < 50) {
                                                    return 8;
                                                }
                                                else if (posBalance >= 50) {
                                                    if (mu_mv_mag_sq < 40) {
                                                        if (Rs < 3475) {
                                                            return 2;
                                                        }
                                                        else if (Rs >= 3475) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (mu_mv_mag_sq >= 40) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (TSCindex >= 8) {
                                    if (RsDiff < 5) {
                                        if (Rs < 3671) {
                                            if (Rs < 3478) {
                                                if (Cs < 3274) {
                                                    if (ssDCval < 50) {
                                                        if (Cs < 3182) {
                                                            return 2;
                                                        }
                                                        else if (Cs >= 3182) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (ssDCval >= 50) {
                                                        return 2;
                                                    }
                                                }
                                                else if (Cs >= 3274) {
                                                    if (diffMVdiffVal < -3) {
                                                        if (Cs < 3588) {
                                                            return 8;
                                                        }
                                                        else if (Cs >= 3588) {
                                                            return 2;
                                                        }
                                                    }
                                                    else if (diffMVdiffVal >= -3) {
                                                        if (MVDiff < 19) {
                                                            return 4;
                                                        }
                                                        else if (MVDiff >= 19) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (Rs >= 3478) {
                                                if (SC < 7228) {
                                                    if (diffAFD < 1) {
                                                        return 2;
                                                    }
                                                    else if (diffAFD >= 1) {
                                                        if (MVDiff < 21) {
                                                            return 4;
                                                        }
                                                        else if (MVDiff >= 21) {
                                                            if (Cs < 3505) {
                                                                if (Cs < 3364) {
                                                                    return 2;
                                                                }
                                                                else if (Cs >= 3364) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (Cs >= 3505) {
                                                                return 2;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (SC >= 7228) {
                                                    if (RsCsDiff < 22) {
                                                        if (SC < 7272) {
                                                            return 8;
                                                        }
                                                        else if (SC >= 7272) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (RsCsDiff >= 22) {
                                                        if (AFD < 11) {
                                                            return 2;
                                                        }
                                                        else if (AFD >= 11) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (Rs >= 3671) {
                                            if (diffRsCsdiff < -6) {
                                                return 4;
                                            }
                                            else if (diffRsCsdiff >= -6) {
                                                if (AFD < 8) {
                                                    return 4;
                                                }
                                                else if (AFD >= 8) {
                                                    return 2;
                                                }
                                            }
                                        }
                                    }
                                    else if (RsDiff >= 5) {
                                        if (Cs < 2096) {
                                            return 4;
                                        }
                                        else if (Cs >= 2096) {
                                            if (negBalance < 50) {
                                                if (mcTcor < 94) {
                                                    if (RsDiff < 9) {
                                                        if (Cs < 3333) {
                                                            if (ssDCval < 108) {
                                                                if (RsCsDiff < 107) {
                                                                    return 8;
                                                                }
                                                                else if (RsCsDiff >= 107) {
                                                                    if (diffRsCsdiff < -33) {
                                                                        return 4;
                                                                    }
                                                                    else if (diffRsCsdiff >= -33) {
                                                                        return 8;
                                                                    }
                                                                }
                                                            }
                                                            else if (ssDCval >= 108) {
                                                                return 1;
                                                            }
                                                        }
                                                        else if (Cs >= 3333) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (RsDiff >= 9) {
                                                        if (ssDCval < 103) {
                                                            return 4;
                                                        }
                                                        else if (ssDCval >= 103) {
                                                            if (Cs < 2823) {
                                                                return 2;
                                                            }
                                                            else if (Cs >= 2823) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (mcTcor >= 94) {
                                                    if (mcTcor < 95) {
                                                        return 4;
                                                    }
                                                    else if (mcTcor >= 95) {
                                                        if (posBalance < 38) {
                                                            return 4;
                                                        }
                                                        else if (posBalance >= 38) {
                                                            return 2;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (negBalance >= 50) {
                                                if (CsDiff < 7) {
                                                    return 4;
                                                }
                                                else if (CsDiff >= 7) {
                                                    return 8;
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                            else if (refDCval >= 108) {
                                if (Cs < 2515) {
                                    if (mu_mv_mag_sq < 63) {
                                        if (TSC < 9) {
                                            if (ssDCval < 109) {
                                                if (diffRsCsdiff < -15) {
                                                    return 8;
                                                }
                                                else if (diffRsCsdiff >= -15) {
                                                    if (RsDiff < 5) {
                                                        if (Rs < 3271) {
                                                            return 4;
                                                        }
                                                        else if (Rs >= 3271) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (RsDiff >= 5) {
                                                        if (tcor < 95) {
                                                            return 4;
                                                        }
                                                        else if (tcor >= 95) {
                                                            if (Rs < 3213) {
                                                                return 4;
                                                            }
                                                            else if (Rs >= 3213) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                            else if (ssDCval >= 109) {
                                                return 4;
                                            }
                                        }
                                        else if (TSC >= 9) {
                                            return 8;
                                        }
                                    }
                                    else if (mu_mv_mag_sq >= 63) {
                                        if (refDCval < 119) {
                                            if (mcTcor < 94) {
                                                if (diffTSC < -1) {
                                                    return 8;
                                                }
                                                else if (diffTSC >= -1) {
                                                    if (MVDiff < 5) {
                                                        return 8;
                                                    }
                                                    else if (MVDiff >= 5) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                            else if (mcTcor >= 94) {
                                                if (Cs < 1788) {
                                                    return 8;
                                                }
                                                else if (Cs >= 1788) {
                                                    if (mu_mv_mag_sq < 77) {
                                                        if (posBalance < 43) {
                                                            if (Cs < 2232) {
                                                                if (MVDiff < 25) {
                                                                    return 4;
                                                                }
                                                                else if (MVDiff >= 25) {
                                                                    return 2;
                                                                }
                                                            }
                                                            else if (Cs >= 2232) {
                                                                if (posBalance < 37) {
                                                                    return 4;
                                                                }
                                                                else if (posBalance >= 37) {
                                                                    return 2;
                                                                }
                                                            }
                                                        }
                                                        else if (posBalance >= 43) {
                                                            if (Rs < 3386) {
                                                                return 4;
                                                            }
                                                            else if (Rs >= 3386) {
                                                                return 2;
                                                            }
                                                        }
                                                    }
                                                    else if (mu_mv_mag_sq >= 77) {
                                                        return 2;
                                                    }
                                                }
                                            }
                                        }
                                        else if (refDCval >= 119) {
                                            if (RsCsDiff < 99) {
                                                return 4;
                                            }
                                            else if (RsCsDiff >= 99) {
                                                if (ssDCval < 124) {
                                                    if (ssDCval < 121) {
                                                        return 8;
                                                    }
                                                    else if (ssDCval >= 121) {
                                                        if (TSC < 7) {
                                                            return 4;
                                                        }
                                                        else if (TSC >= 7) {
                                                            if (Rs < 3795) {
                                                                return 4;
                                                            }
                                                            else if (Rs >= 3795) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (ssDCval >= 124) {
                                                    if (refDCval < 127) {
                                                        return 4;
                                                    }
                                                    else if (refDCval >= 127) {
                                                        return 2;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (Cs >= 2515) {
                                    if (ssDCval < 125) {
                                        if (tcor < 85) {
                                            if (RsCsDiff < 122) {
                                                return 1;
                                            }
                                            else if (RsCsDiff >= 122) {
                                                return 2;
                                            }
                                        }
                                        else if (tcor >= 85) {
                                            if (RsCsDiff < 197) {
                                                return 8;
                                            }
                                            else if (RsCsDiff >= 197) {
                                                return 4;
                                            }
                                        }
                                    }
                                    else if (ssDCval >= 125) {
                                        if (negBalance < 35) {
                                            if (refDCval < 126) {
                                                return 4;
                                            }
                                            else if (refDCval >= 126) {
                                                return 2;
                                            }
                                        }
                                        else if (negBalance >= 35) {
                                            if (Cs < 2744) {
                                                if (SC < 6674) {
                                                    return 1;
                                                }
                                                else if (SC >= 6674) {
                                                    return 2;
                                                }
                                            }
                                            else if (Cs >= 2744) {
                                                if (Rs < 3601) {
                                                    return 8;
                                                }
                                                else if (Rs >= 3601) {
                                                    return 4;
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                        }
                    }
                    else if (AFD >= 12) {
                        if (Cs < 1484) {
                            if (Rs < 1431) {
                                if (MVDiff < 12) {
                                    if (Scindex < 4) {
                                        if (RsDiff < 4) {
                                            return 4;
                                        }
                                        else if (RsDiff >= 4) {
                                            return 2;
                                        }
                                    }
                                    else if (Scindex >= 4) {
                                        if (ssDCval < 89) {
                                            if (negBalance < 50) {
                                                if (MVDiff < 11) {
                                                    return 1;
                                                }
                                                else if (MVDiff >= 11) {
                                                    return 2;
                                                }
                                            }
                                            else if (negBalance >= 50) {
                                                if (RsCsDiff < 46) {
                                                    return 4;
                                                }
                                                else if (RsCsDiff >= 46) {
                                                    return 8;
                                                }
                                            }
                                        }
                                        else if (ssDCval >= 89) {
                                            if (TSC < 5) {
                                                return 1;
                                            }
                                            else if (TSC >= 5) {
                                                return 4;
                                            }
                                        }
                                    }
                                }
                                else if (MVDiff >= 12) {
                                    if (Rs < 1255) {
                                        if (Cs < 533) {
                                            if (MVDiff < 27) {
                                                if (gchDC < 6) {
                                                    return 2;
                                                }
                                                else if (gchDC >= 6) {
                                                    if (Rs < 647) {
                                                        return 2;
                                                    }
                                                    else if (Rs >= 647) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                            else if (MVDiff >= 27) {
                                                if (tcor < 80) {
                                                    if (TSC < 27) {
                                                        if (tcor < 44) {
                                                            if (diffRsCsdiff < 25) {
                                                                if (ssDCval < 44) {
                                                                    return 2;
                                                                }
                                                                else if (ssDCval >= 44) {
                                                                    if (mcTcor < 75) {
                                                                        return 1;
                                                                    }
                                                                    else if (mcTcor >= 75) {
                                                                        if (Rs < 401) {
                                                                            return 2;
                                                                        }
                                                                        else if (Rs >= 401) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                            else if (diffRsCsdiff >= 25) {
                                                                if (Rs < 121) {
                                                                    return 1;
                                                                }
                                                                else if (Rs >= 121) {
                                                                    return 2;
                                                                }
                                                            }
                                                        }
                                                        else if (tcor >= 44) {
                                                            if (SC < 658) {
                                                                if (MVDiff < 69) {
                                                                    if (Rs < 352) {
                                                                        return 1;
                                                                    }
                                                                    else if (Rs >= 352) {
                                                                        if (Rs < 540) {
                                                                            return 2;
                                                                        }
                                                                        else if (Rs >= 540) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                }
                                                                else if (MVDiff >= 69) {
                                                                    return 2;
                                                                }
                                                            }
                                                            else if (SC >= 658) {
                                                                return 2;
                                                            }
                                                        }
                                                    }
                                                    else if (TSC >= 27) {
                                                        return 8;
                                                    }
                                                }
                                                else if (tcor >= 80) {
                                                    if (RsDiff < 3) {
                                                        if (diffRsCsdiff < -11) {
                                                            return 1;
                                                        }
                                                        else if (diffRsCsdiff >= -11) {
                                                            return 2;
                                                        }
                                                    }
                                                    else if (RsDiff >= 3) {
                                                        return 1;
                                                    }
                                                }
                                            }
                                        }
                                        else if (Cs >= 533) {
                                            if (CsDiff < 4) {
                                                if (MVDiff < 26) {
                                                    return 2;
                                                }
                                                else if (MVDiff >= 26) {
                                                    if (mu_mv_mag_sq < 34) {
                                                        return 8;
                                                    }
                                                    else if (mu_mv_mag_sq >= 34) {
                                                        if (Cs < 688) {
                                                            return 8;
                                                        }
                                                        else if (Cs >= 688) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (CsDiff >= 4) {
                                                if (mcTcor < 92) {
                                                    if (diffMVdiffVal < 92) {
                                                        if (Rs < 1089) {
                                                            if (negBalance < 105) {
                                                                if (mu_mv_mag_sq < 61) {
                                                                    return 1;
                                                                }
                                                                else if (mu_mv_mag_sq >= 61) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (negBalance >= 105) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (Rs >= 1089) {
                                                            if (negBalance < 45) {
                                                                return 1;
                                                            }
                                                            else if (negBalance >= 45) {
                                                                if (ssDCval < 66) {
                                                                    if (mu_mv_mag_sq < 19) {
                                                                        return 4;
                                                                    }
                                                                    else if (mu_mv_mag_sq >= 19) {
                                                                        if (Cs < 918) {
                                                                            return 1;
                                                                        }
                                                                        else if (Cs >= 918) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                }
                                                                else if (ssDCval >= 66) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (diffMVdiffVal >= 92) {
                                                        if (TSC < 14) {
                                                            return 2;
                                                        }
                                                        else if (TSC >= 14) {
                                                            if (ssDCval < 69) {
                                                                if (Rs < 464) {
                                                                    return 4;
                                                                }
                                                                else if (Rs >= 464) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (ssDCval >= 69) {
                                                                if (RsDiff < 14) {
                                                                    if (RsDiff < 11) {
                                                                        return 4;
                                                                    }
                                                                    else if (RsDiff >= 11) {
                                                                        return 1;
                                                                    }
                                                                }
                                                                else if (RsDiff >= 14) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (mcTcor >= 92) {
                                                    if (Rs < 986) {
                                                        if (MVDiff < 16) {
                                                            if (TSC < 6) {
                                                                return 4;
                                                            }
                                                            else if (TSC >= 6) {
                                                                return 2;
                                                            }
                                                        }
                                                        else if (MVDiff >= 16) {
                                                            if (CsDiff < 7) {
                                                                return 2;
                                                            }
                                                            else if (CsDiff >= 7) {
                                                                if (mu_mv_mag_sq < 3) {
                                                                    return 1;
                                                                }
                                                                else if (mu_mv_mag_sq >= 3) {
                                                                    return 2;
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (Rs >= 986) {
                                                        if (posBalance < 64) {
                                                            if (Cs < 890) {
                                                                return 1;
                                                            }
                                                            else if (Cs >= 890) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (posBalance >= 64) {
                                                            if (AFD < 22) {
                                                                return 8;
                                                            }
                                                            else if (AFD >= 22) {
                                                                return 2;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (Rs >= 1255) {
                                        if (Cs < 967) {
                                            if (ssDCval < 63) {
                                                return 1;
                                            }
                                            else if (ssDCval >= 63) {
                                                return 2;
                                            }
                                        }
                                        else if (Cs >= 967) {
                                            if (AFD < 14) {
                                                if (Cs < 1171) {
                                                    return 1;
                                                }
                                                else if (Cs >= 1171) {
                                                    return 8;
                                                }
                                            }
                                            else if (AFD >= 14) {
                                                if (mcTcor < 87) {
                                                    if (gchDC < 11) {
                                                        return 8;
                                                    }
                                                    else if (gchDC >= 11) {
                                                        if (Cs < 1124) {
                                                            return 8;
                                                        }
                                                        else if (Cs >= 1124) {
                                                            return 1;
                                                        }
                                                    }
                                                }
                                                else if (mcTcor >= 87) {
                                                    if (Cs < 1429) {
                                                        if (Cs < 1099) {
                                                            return 4;
                                                        }
                                                        else if (Cs >= 1099) {
                                                            return 2;
                                                        }
                                                    }
                                                    else if (Cs >= 1429) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                            else if (Rs >= 1431) {
                                if (tcor < 78) {
                                    if (RsCsDiff < 117) {
                                        if (mu_mv_mag_sq < 17) {
                                            if (negBalance < 40) {
                                                if (AFD < 39) {
                                                    return 2;
                                                }
                                                else if (AFD >= 39) {
                                                    return 4;
                                                }
                                            }
                                            else if (negBalance >= 40) {
                                                if (mcTcor < 81) {
                                                    if (Rs < 2101) {
                                                        return 2;
                                                    }
                                                    else if (Rs >= 2101) {
                                                        return 1;
                                                    }
                                                }
                                                else if (mcTcor >= 81) {
                                                    return 1;
                                                }
                                            }
                                        }
                                        else if (mu_mv_mag_sq >= 17) {
                                            return 8;
                                        }
                                    }
                                    else if (RsCsDiff >= 117) {
                                        if (mcTcor < 87) {
                                            if (refDCval < 83) {
                                                if (ssDCval < 83) {
                                                    return 4;
                                                }
                                                else if (ssDCval >= 83) {
                                                    return 8;
                                                }
                                            }
                                            else if (refDCval >= 83) {
                                                if (ssDCval < 122) {
                                                    if (ssDCval < 107) {
                                                        if (SC < 3647) {
                                                            if (SC < 2978) {
                                                                return 2;
                                                            }
                                                            else if (SC >= 2978) {
                                                                if (TSC < 16) {
                                                                    return 1;
                                                                }
                                                                else if (TSC >= 16) {
                                                                    if (diffRsCsdiff < -53) {
                                                                        return 2;
                                                                    }
                                                                    else if (diffRsCsdiff >= -53) {
                                                                        return 1;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (SC >= 3647) {
                                                            if (diffMVdiffVal < -59) {
                                                                return 1;
                                                            }
                                                            else if (diffMVdiffVal >= -59) {
                                                                if (Rs < 3279) {
                                                                    return 2;
                                                                }
                                                                else if (Rs >= 3279) {
                                                                    if (Rs < 3316) {
                                                                        return 1;
                                                                    }
                                                                    else if (Rs >= 3316) {
                                                                        return 2;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (ssDCval >= 107) {
                                                        return 2;
                                                    }
                                                }
                                                else if (ssDCval >= 122) {
                                                    if (tcor < 70) {
                                                        return 1;
                                                    }
                                                    else if (tcor >= 70) {
                                                        if (Cs < 510) {
                                                            return 4;
                                                        }
                                                        else if (Cs >= 510) {
                                                            if (Cs < 1254) {
                                                                return 2;
                                                            }
                                                            else if (Cs >= 1254) {
                                                                return 1;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (mcTcor >= 87) {
                                            if (ssDCval < 135) {
                                                if (mu_mv_mag_sq < 7) {
                                                    if (diffMVdiffVal < 2) {
                                                        return 1;
                                                    }
                                                    else if (diffMVdiffVal >= 2) {
                                                        return 2;
                                                    }
                                                }
                                                else if (mu_mv_mag_sq >= 7) {
                                                    return 2;
                                                }
                                            }
                                            else if (ssDCval >= 135) {
                                                if (mu_mv_mag_sq < 25) {
                                                    if (TSC < 13) {
                                                        return 4;
                                                    }
                                                    else if (TSC >= 13) {
                                                        if (SC < 6015) {
                                                            return 2;
                                                        }
                                                        else if (SC >= 6015) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                                else if (mu_mv_mag_sq >= 25) {
                                                    return 2;
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (tcor >= 78) {
                                    if (gchDC < 3) {
                                        if (refDCval < 76) {
                                            if (RsDiff < 8) {
                                                if (SC < 2995) {
                                                    return 1;
                                                }
                                                else if (SC >= 2995) {
                                                    return 8;
                                                }
                                            }
                                            else if (RsDiff >= 8) {
                                                return 4;
                                            }
                                        }
                                        else if (refDCval >= 76) {
                                            if (RsCsDiff < 108) {
                                                if (refDCval < 114) {
                                                    return 2;
                                                }
                                                else if (refDCval >= 114) {
                                                    if (SC < 2837) {
                                                        if (SC < 2494) {
                                                            return 2;
                                                        }
                                                        else if (SC >= 2494) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (SC >= 2837) {
                                                        if (mu_mv_mag_sq < 7) {
                                                            return 4;
                                                        }
                                                        else if (mu_mv_mag_sq >= 7) {
                                                            if (AFD < 15) {
                                                                return 2;
                                                            }
                                                            else if (AFD >= 15) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                            else if (RsCsDiff >= 108) {
                                                if (mu_mv_mag_sq < 43) {
                                                    if (MVDiff < 6) {
                                                        if (TSC < 9) {
                                                            return 1;
                                                        }
                                                        else if (TSC >= 9) {
                                                            return 2;
                                                        }
                                                    }
                                                    else if (MVDiff >= 6) {
                                                        return 2;
                                                    }
                                                }
                                                else if (mu_mv_mag_sq >= 43) {
                                                    if (AFD < 16) {
                                                        return 4;
                                                    }
                                                    else if (AFD >= 16) {
                                                        return 2;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (gchDC >= 3) {
                                        if (MVDiff < 11) {
                                            if (Rs < 2365) {
                                                return 1;
                                            }
                                            else if (Rs >= 2365) {
                                                return 8;
                                            }
                                        }
                                        else if (MVDiff >= 11) {
                                            if (refDCval < 121) {
                                                if (gchDC < 5) {
                                                    if (Rs < 1662) {
                                                        return 8;
                                                    }
                                                    else if (Rs >= 1662) {
                                                        return 2;
                                                    }
                                                }
                                                else if (gchDC >= 5) {
                                                    return 4;
                                                }
                                            }
                                            else if (refDCval >= 121) {
                                                return 2;
                                            }
                                        }
                                    }
                                }
                            }
                        }
                        else if (Cs >= 1484) {
                            if (mu_mv_mag_sq < 17) {
                                if (mcTcor < 85) {
                                    if (Rs < 3159) {
                                        if (SC < 4492) {
                                            if (refDCval < 73) {
                                                if (SC < 2518) {
                                                    return 2;
                                                }
                                                else if (SC >= 2518) {
                                                    return 8;
                                                }
                                            }
                                            else if (refDCval >= 73) {
                                                if (RsCsDiff < 395) {
                                                    if (tcor < 78) {
                                                        if (refDCval < 117) {
                                                            if (TSC < 12) {
                                                                return 4;
                                                            }
                                                            else if (TSC >= 12) {
                                                                if (Cs < 2621) {
                                                                    return 1;
                                                                }
                                                                else if (Cs >= 2621) {
                                                                    if (TSC < 18) {
                                                                        return 1;
                                                                    }
                                                                    else if (TSC >= 18) {
                                                                        return 4;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (refDCval >= 117) {
                                                            if (TSC < 16) {
                                                                return 1;
                                                            }
                                                            else if (TSC >= 16) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                    else if (tcor >= 78) {
                                                        if (SC < 3902) {
                                                            return 2;
                                                        }
                                                        else if (SC >= 3902) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                                else if (RsCsDiff >= 395) {
                                                    return 8;
                                                }
                                            }
                                        }
                                        else if (SC >= 4492) {
                                            if (posBalance < 55) {
                                                if (RsDiff < 12) {
                                                    return 1;
                                                }
                                                else if (RsDiff >= 12) {
                                                    if (negBalance < 71) {
                                                        return 8;
                                                    }
                                                    else if (negBalance >= 71) {
                                                        if (TSC < 11) {
                                                            return 8;
                                                        }
                                                        else if (TSC >= 11) {
                                                            return 2;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (posBalance >= 55) {
                                                if (negBalance < 55) {
                                                    if (ssDCval < 129) {
                                                        if (AFD < 21) {
                                                            if (posBalance < 63) {
                                                                if (mu_mv_mag_sq < 3) {
                                                                    return 4;
                                                                }
                                                                else if (mu_mv_mag_sq >= 3) {
                                                                    return 1;
                                                                }
                                                            }
                                                            else if (posBalance >= 63) {
                                                                return 1;
                                                            }
                                                        }
                                                        else if (AFD >= 21) {
                                                            return 1;
                                                        }
                                                    }
                                                    else if (ssDCval >= 129) {
                                                        return 1;
                                                    }
                                                }
                                                else if (negBalance >= 55) {
                                                    if (Rs < 2953) {
                                                        if (MVDiff < 84) {
                                                            if (diffRsCsdiff < -26) {
                                                                if (negBalance < 56) {
                                                                    if (Cs < 2466) {
                                                                        return 1;
                                                                    }
                                                                    else if (Cs >= 2466) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (negBalance >= 56) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (diffRsCsdiff >= -26) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (MVDiff >= 84) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (Rs >= 2953) {
                                                        if (TSC < 16) {
                                                            if (Cs < 3072) {
                                                                return 1;
                                                            }
                                                            else if (Cs >= 3072) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (TSC >= 16) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (Rs >= 3159) {
                                        if (negBalance < 57) {
                                            if (ssDCval < 128) {
                                                if (negBalance < 49) {
                                                    if (Rs < 4033) {
                                                        return 8;
                                                    }
                                                    else if (Rs >= 4033) {
                                                        return 1;
                                                    }
                                                }
                                                else if (negBalance >= 49) {
                                                    return 1;
                                                }
                                            }
                                            else if (ssDCval >= 128) {
                                                return 1;
                                            }
                                        }
                                        else if (negBalance >= 57) {
                                            if (diffRsCsdiff < -50) {
                                                return 2;
                                            }
                                            else if (diffRsCsdiff >= -50) {
                                                if (Rs < 3883) {
                                                    return 1;
                                                }
                                                else if (Rs >= 3883) {
                                                    return 8;
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (mcTcor >= 85) {
                                    if (posBalance < 57) {
                                        if (MVDiff < 5) {
                                            if (ssDCval < 99) {
                                                if (gchDC < 2) {
                                                    return 4;
                                                }
                                                else if (gchDC >= 2) {
                                                    return 2;
                                                }
                                            }
                                            else if (ssDCval >= 99) {
                                                if (tcor < 81) {
                                                    if (Rs < 2599) {
                                                        return 1;
                                                    }
                                                    else if (Rs >= 2599) {
                                                        return 4;
                                                    }
                                                }
                                                else if (tcor >= 81) {
                                                    return 8;
                                                }
                                            }
                                        }
                                        else if (MVDiff >= 5) {
                                            if (negBalance < 52) {
                                                if (CsDiff < 11) {
                                                    if (Rs < 2522) {
                                                        if (TSC < 10) {
                                                            if (TSC < 8) {
                                                                if (ssDCval < 88) {
                                                                    return 1;
                                                                }
                                                                else if (ssDCval >= 88) {
                                                                    if (Cs < 1727) {
                                                                        return 4;
                                                                    }
                                                                    else if (Cs >= 1727) {
                                                                        return 2;
                                                                    }
                                                                }
                                                            }
                                                            else if (TSC >= 8) {
                                                                if (Cs < 2276) {
                                                                    if (Rs < 1786) {
                                                                        if (Cs < 1866) {
                                                                            if (RsDiff < 9) {
                                                                                return 8;
                                                                            }
                                                                            else if (RsDiff >= 9) {
                                                                                return 1;
                                                                            }
                                                                        }
                                                                        else if (Cs >= 1866) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                    else if (Rs >= 1786) {
                                                                        return 8;
                                                                    }
                                                                }
                                                                else if (Cs >= 2276) {
                                                                    if (MVDiff < 10) {
                                                                        return 4;
                                                                    }
                                                                    else if (MVDiff >= 10) {
                                                                        return 1;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (TSC >= 10) {
                                                            if (MVDiff < 15) {
                                                                if (Cs < 2222) {
                                                                    if (Rs < 2236) {
                                                                        return 8;
                                                                    }
                                                                    else if (Rs >= 2236) {
                                                                        return 1;
                                                                    }
                                                                }
                                                                else if (Cs >= 2222) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (MVDiff >= 15) {
                                                                if (posBalance < 49) {
                                                                    return 8;
                                                                }
                                                                else if (posBalance >= 49) {
                                                                    if (diffTSC < 2) {
                                                                        if (tcor < 88) {
                                                                            if (refDCval < 87) {
                                                                                if (posBalance < 54) {
                                                                                    if (SC < 4457) {
                                                                                        return 8;
                                                                                    }
                                                                                    else if (SC >= 4457) {
                                                                                        return 4;
                                                                                    }
                                                                                }
                                                                                else if (posBalance >= 54) {
                                                                                    return 4;
                                                                                }
                                                                            }
                                                                            else if (refDCval >= 87) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                        else if (tcor >= 88) {
                                                                            if (diffAFD < -1) {
                                                                                return 8;
                                                                            }
                                                                            else if (diffAFD >= -1) {
                                                                                if (MVDiff < 24) {
                                                                                    return 4;
                                                                                }
                                                                                else if (MVDiff >= 24) {
                                                                                    return 8;
                                                                                }
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (diffTSC >= 2) {
                                                                        return 1;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (Rs >= 2522) {
                                                        if (posBalance < 52) {
                                                            if (AFD < 14) {
                                                                return 4;
                                                            }
                                                            else if (AFD >= 14) {
                                                                if (Cs < 1964) {
                                                                    return 2;
                                                                }
                                                                else if (Cs >= 1964) {
                                                                    return 1;
                                                                }
                                                            }
                                                        }
                                                        else if (posBalance >= 52) {
                                                            if (ssDCval < 86) {
                                                                if (refDCval < 77) {
                                                                    return 4;
                                                                }
                                                                else if (refDCval >= 77) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (ssDCval >= 86) {
                                                                if (Cs < 1631) {
                                                                    return 4;
                                                                }
                                                                else if (Cs >= 1631) {
                                                                    if (Cs < 2503) {
                                                                        return 1;
                                                                    }
                                                                    else if (Cs >= 2503) {
                                                                        if (SC < 5534) {
                                                                            return 4;
                                                                        }
                                                                        else if (SC >= 5534) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (CsDiff >= 11) {
                                                    if (refDCval < 99) {
                                                        if (refDCval < 82) {
                                                            if (mcTcor < 89) {
                                                                if (gchDC < 2) {
                                                                    return 8;
                                                                }
                                                                else if (gchDC >= 2) {
                                                                    if (TSC < 9) {
                                                                        return 2;
                                                                    }
                                                                    else if (TSC >= 9) {
                                                                        return 1;
                                                                    }
                                                                }
                                                            }
                                                            else if (mcTcor >= 89) {
                                                                return 1;
                                                            }
                                                        }
                                                        else if (refDCval >= 82) {
                                                            if (RsCsDiff < 214) {
                                                                if (SC < 4467) {
                                                                    return 1;
                                                                }
                                                                else if (SC >= 4467) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (RsCsDiff >= 214) {
                                                                if (diffMVdiffVal < -6) {
                                                                    if (Cs < 2412) {
                                                                        return 4;
                                                                    }
                                                                    else if (Cs >= 2412) {
                                                                        return 1;
                                                                    }
                                                                }
                                                                else if (diffMVdiffVal >= -6) {
                                                                    return 1;
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (refDCval >= 99) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                            else if (negBalance >= 52) {
                                                if (TSC < 11) {
                                                    if (refDCval < 83) {
                                                        if (Rs < 1820) {
                                                            return 1;
                                                        }
                                                        else if (Rs >= 1820) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (refDCval >= 83) {
                                                        if (CsDiff < 11) {
                                                            return 2;
                                                        }
                                                        else if (CsDiff >= 11) {
                                                            if (diffTSC < -1) {
                                                                if (Rs < 1870) {
                                                                    return 4;
                                                                }
                                                                else if (Rs >= 1870) {
                                                                    return 2;
                                                                }
                                                            }
                                                            else if (diffTSC >= -1) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (TSC >= 11) {
                                                    if (mcTcor < 90) {
                                                        if (tcor < 82) {
                                                            if (gchDC < 2) {
                                                                if (TSC < 12) {
                                                                    return 1;
                                                                }
                                                                else if (TSC >= 12) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (gchDC >= 2) {
                                                                return 1;
                                                            }
                                                        }
                                                        else if (tcor >= 82) {
                                                            if (Rs < 3183) {
                                                                return 2;
                                                            }
                                                            else if (Rs >= 3183) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                    else if (mcTcor >= 90) {
                                                        if (SC < 4488) {
                                                            if (AFD < 17) {
                                                                return 4;
                                                            }
                                                            else if (AFD >= 17) {
                                                                return 2;
                                                            }
                                                        }
                                                        else if (SC >= 4488) {
                                                            if (CsDiff < 11) {
                                                                return 1;
                                                            }
                                                            else if (CsDiff >= 11) {
                                                                if (Cs < 3177) {
                                                                    return 4;
                                                                }
                                                                else if (Cs >= 3177) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (posBalance >= 57) {
                                        if (Cs < 1814) {
                                            if (CsDiff < 8) {
                                                if (ssDCval < 116) {
                                                    if (MVDiff < 14) {
                                                        if (Cs < 1622) {
                                                            return 1;
                                                        }
                                                        else if (Cs >= 1622) {
                                                            return 2;
                                                        }
                                                    }
                                                    else if (MVDiff >= 14) {
                                                        if (Rs < 2149) {
                                                            return 4;
                                                        }
                                                        else if (Rs >= 2149) {
                                                            if (Cs < 1685) {
                                                                return 2;
                                                            }
                                                            else if (Cs >= 1685) {
                                                                return 1;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (ssDCval >= 116) {
                                                    if (SC < 4877) {
                                                        return 4;
                                                    }
                                                    else if (SC >= 4877) {
                                                        return 2;
                                                    }
                                                }
                                            }
                                            else if (CsDiff >= 8) {
                                                if (negBalance < 56) {
                                                    if (diffMVdiffVal < 1) {
                                                        if (diffTSC < -2) {
                                                            if (Cs < 1609) {
                                                                return 1;
                                                            }
                                                            else if (Cs >= 1609) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (diffTSC >= -2) {
                                                            if (gchDC < 1) {
                                                                return 2;
                                                            }
                                                            else if (gchDC >= 1) {
                                                                if (AFD < 19) {
                                                                    if (TSC < 10) {
                                                                        if (Cs < 1734) {
                                                                            return 1;
                                                                        }
                                                                        else if (Cs >= 1734) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                    else if (TSC >= 10) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (AFD >= 19) {
                                                                    return 2;
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (diffMVdiffVal >= 1) {
                                                        if (Rs < 1636) {
                                                            if (Rs < 1454) {
                                                                return 8;
                                                            }
                                                            else if (Rs >= 1454) {
                                                                return 2;
                                                            }
                                                        }
                                                        else if (Rs >= 1636) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                                else if (negBalance >= 56) {
                                                    if (AFD < 17) {
                                                        return 8;
                                                    }
                                                    else if (AFD >= 17) {
                                                        return 2;
                                                    }
                                                }
                                            }
                                        }
                                        else if (Cs >= 1814) {
                                            if (TSC < 12) {
                                                if (mcTcor < 87) {
                                                    if (negBalance < 51) {
                                                        return 1;
                                                    }
                                                    else if (negBalance >= 51) {
                                                        if (AFD < 19) {
                                                            return 4;
                                                        }
                                                        else if (AFD >= 19) {
                                                            return 1;
                                                        }
                                                    }
                                                }
                                                else if (mcTcor >= 87) {
                                                    if (Cs < 2441) {
                                                        if (negBalance < 46) {
                                                            if (refDCval < 86) {
                                                                return 2;
                                                            }
                                                            else if (refDCval >= 86) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (negBalance >= 46) {
                                                            if (tcor < 89) {
                                                                if (RsDiff < 15) {
                                                                    if (diffRsCsdiff < 85) {
                                                                        return 2;
                                                                    }
                                                                    else if (diffRsCsdiff >= 85) {
                                                                        if (Rs < 2586) {
                                                                            return 2;
                                                                        }
                                                                        else if (Rs >= 2586) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                }
                                                                else if (RsDiff >= 15) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (tcor >= 89) {
                                                                if (Rs < 2728) {
                                                                    if (ssDCval < 95) {
                                                                        return 2;
                                                                    }
                                                                    else if (ssDCval >= 95) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (Rs >= 2728) {
                                                                    return 1;
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (Cs >= 2441) {
                                                        if (MVDiff < 10) {
                                                            if (RsDiff < 12) {
                                                                if (Rs < 2638) {
                                                                    return 2;
                                                                }
                                                                else if (Rs >= 2638) {
                                                                    if (AFD < 13) {
                                                                        return 4;
                                                                    }
                                                                    else if (AFD >= 13) {
                                                                        return 8;
                                                                    }
                                                                }
                                                            }
                                                            else if (RsDiff >= 12) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (MVDiff >= 10) {
                                                            return 2;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (TSC >= 12) {
                                                if (refDCval < 95) {
                                                    if (RsDiff < 13) {
                                                        if (negBalance < 52) {
                                                            if (Rs < 2178) {
                                                                return 8;
                                                            }
                                                            else if (Rs >= 2178) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (negBalance >= 52) {
                                                            if (tcor < 87) {
                                                                if (diffTSC < 1) {
                                                                    if (diffMVdiffVal < -1) {
                                                                        if (Rs < 3093) {
                                                                            return 8;
                                                                        }
                                                                        else if (Rs >= 3093) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                    else if (diffMVdiffVal >= -1) {
                                                                        return 8;
                                                                    }
                                                                }
                                                                else if (diffTSC >= 1) {
                                                                    if (Rs < 2943) {
                                                                        if (Cs < 2203) {
                                                                            return 4;
                                                                        }
                                                                        else if (Cs >= 2203) {
                                                                            return 2;
                                                                        }
                                                                    }
                                                                    else if (Rs >= 2943) {
                                                                        return 8;
                                                                    }
                                                                }
                                                            }
                                                            else if (tcor >= 87) {
                                                                return 2;
                                                            }
                                                        }
                                                    }
                                                    else if (RsDiff >= 13) {
                                                        if (Rs < 3683) {
                                                            return 1;
                                                        }
                                                        else if (Rs >= 3683) {
                                                            return 2;
                                                        }
                                                    }
                                                }
                                                else if (refDCval >= 95) {
                                                    if (ssDCval < 128) {
                                                        if (posBalance < 59) {
                                                            if (RsCsDiff < 234) {
                                                                if (SC < 4379) {
                                                                    return 2;
                                                                }
                                                                else if (SC >= 4379) {
                                                                    if (TSC < 13) {
                                                                        return 4;
                                                                    }
                                                                    else if (TSC >= 13) {
                                                                        return 1;
                                                                    }
                                                                }
                                                            }
                                                            else if (RsCsDiff >= 234) {
                                                                if (Rs < 3173) {
                                                                    return 4;
                                                                }
                                                                else if (Rs >= 3173) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                        else if (posBalance >= 59) {
                                                            if (ssDCval < 125) {
                                                                if (RsDiff < 20) {
                                                                    if (gchDC < 1) {
                                                                        if (Cs < 1987) {
                                                                            return 1;
                                                                        }
                                                                        else if (Cs >= 1987) {
                                                                            return 2;
                                                                        }
                                                                    }
                                                                    else if (gchDC >= 1) {
                                                                        return 1;
                                                                    }
                                                                }
                                                                else if (RsDiff >= 20) {
                                                                    return 2;
                                                                }
                                                            }
                                                            else if (ssDCval >= 125) {
                                                                if (Rs < 3374) {
                                                                    return 4;
                                                                }
                                                                else if (Rs >= 3374) {
                                                                    return 1;
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (ssDCval >= 128) {
                                                        if (RsDiff < 13) {
                                                            if (ssDCval < 132) {
                                                                return 4;
                                                            }
                                                            else if (ssDCval >= 132) {
                                                                return 1;
                                                            }
                                                        }
                                                        else if (RsDiff >= 13) {
                                                            if (MVDiff < 20) {
                                                                if (Cs < 2968) {
                                                                    return 2;
                                                                }
                                                                else if (Cs >= 2968) {
                                                                    return 1;
                                                                }
                                                            }
                                                            else if (MVDiff >= 20) {
                                                                return 2;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                            else if (mu_mv_mag_sq >= 17) {
                                if (tcor < 90) {
                                    if (ssDCval < 76) {
                                        if (negBalance < 43) {
                                            if (CsDiff < 8) {
                                                return 8;
                                            }
                                            else if (CsDiff >= 8) {
                                                return 1;
                                            }
                                        }
                                        else if (negBalance >= 43) {
                                            if (SC < 5677) {
                                                if (negBalance < 56) {
                                                    return 8;
                                                }
                                                else if (negBalance >= 56) {
                                                    if (diffMVdiffVal < 5) {
                                                        return 8;
                                                    }
                                                    else if (diffMVdiffVal >= 5) {
                                                        if (SC < 3627) {
                                                            return 4;
                                                        }
                                                        else if (SC >= 3627) {
                                                            return 1;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (SC >= 5677) {
                                                if (ssDCval < 65) {
                                                    if (SC < 7342) {
                                                        return 8;
                                                    }
                                                    else if (SC >= 7342) {
                                                        if (RsCsDiff < 86) {
                                                            return 4;
                                                        }
                                                        else if (RsCsDiff >= 86) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                                else if (ssDCval >= 65) {
                                                    if (Cs < 3407) {
                                                        return 1;
                                                    }
                                                    else if (Cs >= 3407) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (ssDCval >= 76) {
                                        if (SC < 5207) {
                                            if (mcTcor < 94) {
                                                if (posBalance < 57) {
                                                    if (RsCsDiff < 220) {
                                                        if (Rs < 2204) {
                                                            if (RsDiff < 9) {
                                                                if (AFD < 19) {
                                                                    return 8;
                                                                }
                                                                else if (AFD >= 19) {
                                                                    return 1;
                                                                }
                                                            }
                                                            else if (RsDiff >= 9) {
                                                                return 1;
                                                            }
                                                        }
                                                        else if (Rs >= 2204) {
                                                            if (MVDiff < 19) {
                                                                if (diffMVdiffVal < 3) {
                                                                    if (ssDCval < 138) {
                                                                        if (posBalance < 56) {
                                                                            return 1;
                                                                        }
                                                                        else if (posBalance >= 56) {
                                                                            if (TSC < 11) {
                                                                                return 2;
                                                                            }
                                                                            else if (TSC >= 11) {
                                                                                return 1;
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (ssDCval >= 138) {
                                                                        return 8;
                                                                    }
                                                                }
                                                                else if (diffMVdiffVal >= 3) {
                                                                    if (AFD < 16) {
                                                                        return 4;
                                                                    }
                                                                    else if (AFD >= 16) {
                                                                        return 1;
                                                                    }
                                                                }
                                                            }
                                                            else if (MVDiff >= 19) {
                                                                if (TSC < 14) {
                                                                    if (mu_mv_mag_sq < 42) {
                                                                        if (SC < 4867) {
                                                                            return 8;
                                                                        }
                                                                        else if (SC >= 4867) {
                                                                            if (RsDiff < 8) {
                                                                                return 4;
                                                                            }
                                                                            else if (RsDiff >= 8) {
                                                                                return 8;
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (mu_mv_mag_sq >= 42) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (TSC >= 14) {
                                                                    return 1;
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (RsCsDiff >= 220) {
                                                        if (diffMVdiffVal < 4) {
                                                            if (Cs < 1746) {
                                                                if (Rs < 2059) {
                                                                    return 1;
                                                                }
                                                                else if (Rs >= 2059) {
                                                                    return 2;
                                                                }
                                                            }
                                                            else if (Cs >= 1746) {
                                                                return 1;
                                                            }
                                                        }
                                                        else if (diffMVdiffVal >= 4) {
                                                            if (Rs < 2320) {
                                                                if (Rs < 1631) {
                                                                    return 8;
                                                                }
                                                                else if (Rs >= 1631) {
                                                                    return 1;
                                                                }
                                                            }
                                                            else if (Rs >= 2320) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (posBalance >= 57) {
                                                    if (AFD < 14) {
                                                        if (Cs < 2164) {
                                                            if (Rs < 2081) {
                                                                return 4;
                                                            }
                                                            else if (Rs >= 2081) {
                                                                return 1;
                                                            }
                                                        }
                                                        else if (Cs >= 2164) {
                                                            if (ssDCval < 92) {
                                                                return 4;
                                                            }
                                                            else if (ssDCval >= 92) {
                                                                if (RsCsDiff < 146) {
                                                                    return 1;
                                                                }
                                                                else if (RsCsDiff >= 146) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (AFD >= 14) {
                                                        if (Cs < 1907) {
                                                            if (CsDiff < 9) {
                                                                if (Rs < 2341) {
                                                                    return 2;
                                                                }
                                                                else if (Rs >= 2341) {
                                                                    if (RsCsDiff < 329) {
                                                                        return 8;
                                                                    }
                                                                    else if (RsCsDiff >= 329) {
                                                                        return 2;
                                                                    }
                                                                }
                                                            }
                                                            else if (CsDiff >= 9) {
                                                                if (refDCval < 117) {
                                                                    if (SC < 3269) {
                                                                        return 1;
                                                                    }
                                                                    else if (SC >= 3269) {
                                                                        return 2;
                                                                    }
                                                                }
                                                                else if (refDCval >= 117) {
                                                                    if (AFD < 22) {
                                                                        return 4;
                                                                    }
                                                                    else if (AFD >= 22) {
                                                                        return 1;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (Cs >= 1907) {
                                                            if (ssDCval < 92) {
                                                                if (Cs < 2048) {
                                                                    if (TSC < 11) {
                                                                        return 2;
                                                                    }
                                                                    else if (TSC >= 11) {
                                                                        return 1;
                                                                    }
                                                                }
                                                                else if (Cs >= 2048) {
                                                                    if (diffMVdiffVal < -4) {
                                                                        if (Cs < 2153) {
                                                                            if (Cs < 2138) {
                                                                                return 1;
                                                                            }
                                                                            else if (Cs >= 2138) {
                                                                                return 2;
                                                                            }
                                                                        }
                                                                        else if (Cs >= 2153) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                    else if (diffMVdiffVal >= -4) {
                                                                        if (SC < 4632) {
                                                                            if (Cs < 2073) {
                                                                                return 2;
                                                                            }
                                                                            else if (Cs >= 2073) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                        else if (SC >= 4632) {
                                                                            if (diffAFD < 1) {
                                                                                return 2;
                                                                            }
                                                                            else if (diffAFD >= 1) {
                                                                                if (negBalance < 54) {
                                                                                    if (Rs < 2772) {
                                                                                        return 4;
                                                                                    }
                                                                                    else if (Rs >= 2772) {
                                                                                        return 2;
                                                                                    }
                                                                                }
                                                                                else if (negBalance >= 54) {
                                                                                    if (Rs < 2743) {
                                                                                        return 2;
                                                                                    }
                                                                                    else if (Rs >= 2743) {
                                                                                        return 1;
                                                                                    }
                                                                                }
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                            else if (ssDCval >= 92) {
                                                                if (Rs < 2904) {
                                                                    if (tcor < 89) {
                                                                        return 1;
                                                                    }
                                                                    else if (tcor >= 89) {
                                                                        if (Rs < 2740) {
                                                                            return 8;
                                                                        }
                                                                        else if (Rs >= 2740) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                }
                                                                else if (Rs >= 2904) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                            else if (mcTcor >= 94) {
                                                if (Rs < 1555) {
                                                    return 4;
                                                }
                                                else if (Rs >= 1555) {
                                                    return 2;
                                                }
                                            }
                                        }
                                        else if (SC >= 5207) {
                                            if (MVDiff < 3) {
                                                return 8;
                                            }
                                            else if (MVDiff >= 3) {
                                                if (Cs < 2892) {
                                                    if (Cs < 2436) {
                                                        if (mcTcor < 93) {
                                                            if (negBalance < 52) {
                                                                if (posBalance < 49) {
                                                                    if (Rs < 4335) {
                                                                        return 8;
                                                                    }
                                                                    else if (Rs >= 4335) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (posBalance >= 49) {
                                                                    if (RsDiff < 19) {
                                                                        if (TSC < 12) {
                                                                            if (refDCval < 126) {
                                                                                if (diffMVdiffVal < -2) {
                                                                                    return 4;
                                                                                }
                                                                                else if (diffMVdiffVal >= -2) {
                                                                                    if (refDCval < 84) {
                                                                                        return 1;
                                                                                    }
                                                                                    else if (refDCval >= 84) {
                                                                                        if (Rs < 2955) {
                                                                                            if (RsDiff < 8) {
                                                                                                return 1;
                                                                                            }
                                                                                            else if (RsDiff >= 8) {
                                                                                                return 4;
                                                                                            }
                                                                                        }
                                                                                        else if (Rs >= 2955) {
                                                                                            return 4;
                                                                                        }
                                                                                    }
                                                                                }
                                                                            }
                                                                            else if (refDCval >= 126) {
                                                                                return 1;
                                                                            }
                                                                        }
                                                                        else if (TSC >= 12) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                    else if (RsDiff >= 19) {
                                                                        return 4;
                                                                    }
                                                                }
                                                            }
                                                            else if (negBalance >= 52) {
                                                                if (posBalance < 57) {
                                                                    if (posBalance < 51) {
                                                                        if (TSC < 12) {
                                                                            return 4;
                                                                        }
                                                                        else if (TSC >= 12) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                    else if (posBalance >= 51) {
                                                                        return 8;
                                                                    }
                                                                }
                                                                else if (posBalance >= 57) {
                                                                    if (diffMVdiffVal < -5) {
                                                                        return 8;
                                                                    }
                                                                    else if (diffMVdiffVal >= -5) {
                                                                        if (diffTSC < -1) {
                                                                            return 2;
                                                                        }
                                                                        else if (diffTSC >= -1) {
                                                                            if (ssDCval < 103) {
                                                                                if (SC < 5350) {
                                                                                    return 1;
                                                                                }
                                                                                else if (SC >= 5350) {
                                                                                    return 2;
                                                                                }
                                                                            }
                                                                            else if (ssDCval >= 103) {
                                                                                if (Rs < 2980) {
                                                                                    return 4;
                                                                                }
                                                                                else if (Rs >= 2980) {
                                                                                    return 1;
                                                                                }
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (mcTcor >= 93) {
                                                            if (Rs < 3185) {
                                                                if (AFD < 15) {
                                                                    return 4;
                                                                }
                                                                else if (AFD >= 15) {
                                                                    return 1;
                                                                }
                                                            }
                                                            else if (Rs >= 3185) {
                                                                if (MVDiff < 14) {
                                                                    return 4;
                                                                }
                                                                else if (MVDiff >= 14) {
                                                                    if (Rs < 3538) {
                                                                        return 2;
                                                                    }
                                                                    else if (Rs >= 3538) {
                                                                        if (Cs < 2003) {
                                                                            return 2;
                                                                        }
                                                                        else if (Cs >= 2003) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (Cs >= 2436) {
                                                        if (mu_mv_mag_sq < 34) {
                                                            if (CsDiff < 10) {
                                                                if (Rs < 3527) {
                                                                    return 1;
                                                                }
                                                                else if (Rs >= 3527) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (CsDiff >= 10) {
                                                                if (refDCval < 116) {
                                                                    if (MVDiff < 43) {
                                                                        if (posBalance < 56) {
                                                                            return 8;
                                                                        }
                                                                        else if (posBalance >= 56) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                    else if (MVDiff >= 43) {
                                                                        return 1;
                                                                    }
                                                                }
                                                                else if (refDCval >= 116) {
                                                                    return 1;
                                                                }
                                                            }
                                                        }
                                                        else if (mu_mv_mag_sq >= 34) {
                                                            if (gchDC < 1) {
                                                                if (ssDCval < 123) {
                                                                    if (RsDiff < 10) {
                                                                        if (diffRsCsdiff < 22) {
                                                                            return 1;
                                                                        }
                                                                        else if (diffRsCsdiff >= 22) {
                                                                            if (Cs < 2488) {
                                                                                return 4;
                                                                            }
                                                                            else if (Cs >= 2488) {
                                                                                return 1;
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (RsDiff >= 10) {
                                                                        if (mcTcor < 88) {
                                                                            return 4;
                                                                        }
                                                                        else if (mcTcor >= 88) {
                                                                            if (diffRsCsdiff < 52) {
                                                                                return 1;
                                                                            }
                                                                            else if (diffRsCsdiff >= 52) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                                else if (ssDCval >= 123) {
                                                                    return 1;
                                                                }
                                                            }
                                                            else if (gchDC >= 1) {
                                                                if (ssDCval < 96) {
                                                                    if (diffAFD < 2) {
                                                                        return 1;
                                                                    }
                                                                    else if (diffAFD >= 2) {
                                                                        if (SC < 5740) {
                                                                            return 8;
                                                                        }
                                                                        else if (SC >= 5740) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                }
                                                                else if (ssDCval >= 96) {
                                                                    if (refDCval < 132) {
                                                                        if (mcTcor < 78) {
                                                                            return 1;
                                                                        }
                                                                        else if (mcTcor >= 78) {
                                                                            if (SC < 7262) {
                                                                                if (Rs < 3624) {
                                                                                    if (ssDCval < 108) {
                                                                                        return 4;
                                                                                    }
                                                                                    else if (ssDCval >= 108) {
                                                                                        return 1;
                                                                                    }
                                                                                }
                                                                                else if (Rs >= 3624) {
                                                                                    return 4;
                                                                                }
                                                                            }
                                                                            else if (SC >= 7262) {
                                                                                return 2;
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (refDCval >= 132) {
                                                                        if (negBalance < 57) {
                                                                            return 1;
                                                                        }
                                                                        else if (negBalance >= 57) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (Cs >= 2892) {
                                                    if (mcTcor < 83) {
                                                        if (ssDCval < 128) {
                                                            if (tcor < 67) {
                                                                if (MVDiff < 53) {
                                                                    return 2;
                                                                }
                                                                else if (MVDiff >= 53) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (tcor >= 67) {
                                                                if (negBalance < 42) {
                                                                    if (SC < 6970) {
                                                                        return 8;
                                                                    }
                                                                    else if (SC >= 6970) {
                                                                        return 1;
                                                                    }
                                                                }
                                                                else if (negBalance >= 42) {
                                                                    if (diffRsCsdiff < 191) {
                                                                        return 4;
                                                                    }
                                                                    else if (diffRsCsdiff >= 191) {
                                                                        return 8;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (ssDCval >= 128) {
                                                            return 1;
                                                        }
                                                    }
                                                    else if (mcTcor >= 83) {
                                                        if (ssDCval < 79) {
                                                            return 1;
                                                        }
                                                        else if (ssDCval >= 79) {
                                                            if (tcor < 79) {
                                                                if (MVDiff < 12) {
                                                                    return 1;
                                                                }
                                                                else if (MVDiff >= 12) {
                                                                    if (Cs < 2999) {
                                                                        return 1;
                                                                    }
                                                                    else if (Cs >= 2999) {
                                                                        return 8;
                                                                    }
                                                                }
                                                            }
                                                            else if (tcor >= 79) {
                                                                if (posBalance < 61) {
                                                                    if (MVDiff < 4) {
                                                                        return 4;
                                                                    }
                                                                    else if (MVDiff >= 4) {
                                                                        if (diffMVdiffVal < -10) {
                                                                            if (Rs < 2808) {
                                                                                return 1;
                                                                            }
                                                                            else if (Rs >= 2808) {
                                                                                return 8;
                                                                            }
                                                                        }
                                                                        else if (diffMVdiffVal >= -10) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                }
                                                                else if (posBalance >= 61) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (tcor >= 90) {
                                    if (CsDiff < 9) {
                                        if (SC < 5370) {
                                            if (refDCval < 88) {
                                                if (MVDiff < 13) {
                                                    if (refDCval < 81) {
                                                        return 2;
                                                    }
                                                    else if (refDCval >= 81) {
                                                        return 1;
                                                    }
                                                }
                                                else if (MVDiff >= 13) {
                                                    return 2;
                                                }
                                            }
                                            else if (refDCval >= 88) {
                                                if (RsCsDiff < 99) {
                                                    if (MVDiff < 7) {
                                                        if (Cs < 2339) {
                                                            return 2;
                                                        }
                                                        else if (Cs >= 2339) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (MVDiff >= 7) {
                                                        if (mu_mv_mag_sq < 70) {
                                                            return 1;
                                                        }
                                                        else if (mu_mv_mag_sq >= 70) {
                                                            if (TSC < 11) {
                                                                return 1;
                                                            }
                                                            else if (TSC >= 11) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (RsCsDiff >= 99) {
                                                    if (posBalance < 65) {
                                                        if (ssDCval < 106) {
                                                            if (AFD < 13) {
                                                                if (CsDiff < 7) {
                                                                    return 1;
                                                                }
                                                                else if (CsDiff >= 7) {
                                                                    if (SC < 4562) {
                                                                        return 8;
                                                                    }
                                                                    else if (SC >= 4562) {
                                                                        return 2;
                                                                    }
                                                                }
                                                            }
                                                            else if (AFD >= 13) {
                                                                return 1;
                                                            }
                                                        }
                                                        else if (ssDCval >= 106) {
                                                            if (Cs < 1866) {
                                                                if (Cs < 1786) {
                                                                    return 8;
                                                                }
                                                                else if (Cs >= 1786) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (Cs >= 1866) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                    else if (posBalance >= 65) {
                                                        if (Rs < 1819) {
                                                            return 4;
                                                        }
                                                        else if (Rs >= 1819) {
                                                            return 2;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (SC >= 5370) {
                                            if (ssDCval < 109) {
                                                if (negBalance < 41) {
                                                    return 4;
                                                }
                                                else if (negBalance >= 41) {
                                                    if (MVDiff < 8) {
                                                        if (diffAFD < 1) {
                                                            if (Rs < 3119) {
                                                                return 4;
                                                            }
                                                            else if (Rs >= 3119) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (diffAFD >= 1) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (MVDiff >= 8) {
                                                        if (ssDCval < 89) {
                                                            return 8;
                                                        }
                                                        else if (ssDCval >= 89) {
                                                            return 1;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (ssDCval >= 109) {
                                                if (refDCval < 115) {
                                                    return 1;
                                                }
                                                else if (refDCval >= 115) {
                                                    return 8;
                                                }
                                            }
                                        }
                                    }
                                    else if (CsDiff >= 9) {
                                        if (ssDCval < 109) {
                                            if (Cs < 2354) {
                                                return 2;
                                            }
                                            else if (Cs >= 2354) {
                                                if (Rs < 2216) {
                                                    return 2;
                                                }
                                                else if (Rs >= 2216) {
                                                    if (Cs < 3118) {
                                                        return 8;
                                                    }
                                                    else if (Cs >= 3118) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                        }
                                        else if (ssDCval >= 109) {
                                            return 8;
                                        }
                                    }
                                }
                            }
                        }
                    }
                }
                else if (SC >= 7536) {
                    if (ssDCval < 68) {
                        if (TSC < 7) {
                            if (tcor < 90) {
                                if (diffTSC < -1) {
                                    if (Rs < 4048) {
                                        return 8;
                                    }
                                    else if (Rs >= 4048) {
                                        return 4;
                                    }
                                }
                                else if (diffTSC >= -1) {
                                    return 8;
                                }
                            }
                            else if (tcor >= 90) {
                                if (diffTSC < -1) {
                                    if (Rs < 3836) {
                                        return 2;
                                    }
                                    else if (Rs >= 3836) {
                                        return 4;
                                    }
                                }
                                else if (diffTSC >= -1) {
                                    if (mu_mv_mag_sq < 71) {
                                        if (Cs < 3791) {
                                            return 4;
                                        }
                                        else if (Cs >= 3791) {
                                            return 8;
                                        }
                                    }
                                    else if (mu_mv_mag_sq >= 71) {
                                        return 8;
                                    }
                                }
                            }
                        }
                        else if (TSC >= 7) {
                            if (tcor < 80) {
                                if (SC < 11511) {
                                    if (mcTcor < 88) {
                                        return 8;
                                    }
                                    else if (mcTcor >= 88) {
                                        if (MVDiff < 16) {
                                            return 4;
                                        }
                                        else if (MVDiff >= 16) {
                                            return 8;
                                        }
                                    }
                                }
                                else if (SC >= 11511) {
                                    return 4;
                                }
                            }
                            else if (tcor >= 80) {
                                if (Rs < 3950) {
                                    if (gchDC < 3) {
                                        if (RsCsDiff < 46) {
                                            if (mu_mv_mag_sq < 50) {
                                                if (mcTcor < 93) {
                                                    return 2;
                                                }
                                                else if (mcTcor >= 93) {
                                                    if (refDCval < 54) {
                                                        if (Rs < 3753) {
                                                            return 2;
                                                        }
                                                        else if (Rs >= 3753) {
                                                            if (RsDiff < 4) {
                                                                return 4;
                                                            }
                                                            else if (RsDiff >= 4) {
                                                                return 2;
                                                            }
                                                        }
                                                    }
                                                    else if (refDCval >= 54) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                            else if (mu_mv_mag_sq >= 50) {
                                                return 4;
                                            }
                                        }
                                        else if (RsCsDiff >= 46) {
                                            return 4;
                                        }
                                    }
                                    else if (gchDC >= 3) {
                                        return 8;
                                    }
                                }
                                else if (Rs >= 3950) {
                                    return 4;
                                }
                            }
                        }
                    }
                    else if (ssDCval >= 68) {
                        if (Cs < 3293) {
                            if (tcor < 77) {
                                if (ssDCval < 106) {
                                    if (TSC < 15) {
                                        return 4;
                                    }
                                    else if (TSC >= 15) {
                                        return 2;
                                    }
                                }
                                else if (ssDCval >= 106) {
                                    if (ssDCval < 150) {
                                        if (SC < 8892) {
                                            if (Cs < 2586) {
                                                return 1;
                                            }
                                            else if (Cs >= 2586) {
                                                if (mcTcor < 80) {
                                                    if (Rs < 4825) {
                                                        return 1;
                                                    }
                                                    else if (Rs >= 4825) {
                                                        return 2;
                                                    }
                                                }
                                                else if (mcTcor >= 80) {
                                                    if (TSC < 15) {
                                                        if (Rs < 4841) {
                                                            return 1;
                                                        }
                                                        else if (Rs >= 4841) {
                                                            return 2;
                                                        }
                                                    }
                                                    else if (TSC >= 15) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                        }
                                        else if (SC >= 8892) {
                                            return 1;
                                        }
                                    }
                                    else if (ssDCval >= 150) {
                                        return 4;
                                    }
                                }
                            }
                            else if (tcor >= 77) {
                                if (CsDiff < 12) {
                                    if (tcor < 83) {
                                        if (Rs < 6276) {
                                            if (refDCval < 104) {
                                                return 4;
                                            }
                                            else if (refDCval >= 104) {
                                                if (RsDiff < 17) {
                                                    if (CsDiff < 8) {
                                                        if (Cs < 3155) {
                                                            return 4;
                                                        }
                                                        else if (Cs >= 3155) {
                                                            return 2;
                                                        }
                                                    }
                                                    else if (CsDiff >= 8) {
                                                        if (SC < 7543) {
                                                            return 4;
                                                        }
                                                        else if (SC >= 7543) {
                                                            if (AFD < 22) {
                                                                return 2;
                                                            }
                                                            else if (AFD >= 22) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (RsDiff >= 17) {
                                                    return 4;
                                                }
                                            }
                                        }
                                        else if (Rs >= 6276) {
                                            return 1;
                                        }
                                    }
                                    else if (tcor >= 83) {
                                        if (Rs < 5107) {
                                            if (Rs < 4657) {
                                                if (MVDiff < 10) {
                                                    return 4;
                                                }
                                                else if (MVDiff >= 10) {
                                                    return 8;
                                                }
                                            }
                                            else if (Rs >= 4657) {
                                                if (Cs < 2625) {
                                                    return 8;
                                                }
                                                else if (Cs >= 2625) {
                                                    if (CsDiff < 6) {
                                                        return 8;
                                                    }
                                                    else if (CsDiff >= 6) {
                                                        return 2;
                                                    }
                                                }
                                            }
                                        }
                                        else if (Rs >= 5107) {
                                            if (Rs < 5344) {
                                                if (refDCval < 110) {
                                                    return 4;
                                                }
                                                else if (refDCval >= 110) {
                                                    return 8;
                                                }
                                            }
                                            else if (Rs >= 5344) {
                                                return 4;
                                            }
                                        }
                                    }
                                }
                                else if (CsDiff >= 12) {
                                    if (posBalance < 56) {
                                        if (SC < 9095) {
                                            return 4;
                                        }
                                        else if (SC >= 9095) {
                                            return 1;
                                        }
                                    }
                                    else if (posBalance >= 56) {
                                        return 8;
                                    }
                                }
                            }
                        }
                        else if (Cs >= 3293) {
                            if (mu_mv_mag_sq < 63) {
                                if (refDCval < 149) {
                                    if (Rs < 6965) {
                                        if (mu_mv_mag_sq < 3) {
                                            if (Cs < 4295) {
                                                if (ssDCval < 121) {
                                                    if (Rs < 3933) {
                                                        return 8;
                                                    }
                                                    else if (Rs >= 3933) {
                                                        return 4;
                                                    }
                                                }
                                                else if (ssDCval >= 121) {
                                                    if (ssDCval < 132) {
                                                        return 8;
                                                    }
                                                    else if (ssDCval >= 132) {
                                                        if (Rs < 5370) {
                                                            if (RsDiff < 17) {
                                                                return 1;
                                                            }
                                                            else if (RsDiff >= 17) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (Rs >= 5370) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (Cs >= 4295) {
                                                if (TSC < 17) {
                                                    if (CsDiff < 17) {
                                                        return 8;
                                                    }
                                                    else if (CsDiff >= 17) {
                                                        if (TSC < 16) {
                                                            return 1;
                                                        }
                                                        else if (TSC >= 16) {
                                                            if (SC < 8986) {
                                                                return 1;
                                                            }
                                                            else if (SC >= 8986) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (TSC >= 17) {
                                                    if (Rs < 6670) {
                                                        return 8;
                                                    }
                                                    else if (Rs >= 6670) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                        }
                                        else if (mu_mv_mag_sq >= 3) {
                                            if (SC < 8259) {
                                                if (gchDC < 2) {
                                                    if (mu_mv_mag_sq < 37) {
                                                        if (RsCsDiff < 515) {
                                                            if (Cs < 3830) {
                                                                return 4;
                                                            }
                                                            else if (Cs >= 3830) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (RsCsDiff >= 515) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (mu_mv_mag_sq >= 37) {
                                                        if (tcor < 73) {
                                                            return 1;
                                                        }
                                                        else if (tcor >= 73) {
                                                            if (ssDCval < 126) {
                                                                return 4;
                                                            }
                                                            else if (ssDCval >= 126) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (gchDC >= 2) {
                                                    if (RsDiff < 20) {
                                                        if (RsDiff < 6) {
                                                            return 4;
                                                        }
                                                        else if (RsDiff >= 6) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (RsDiff >= 20) {
                                                        if (SC < 7959) {
                                                            return 2;
                                                        }
                                                        else if (SC >= 7959) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (SC >= 8259) {
                                                if (Cs < 4323) {
                                                    if (ssDCval < 140) {
                                                        if (RsDiff < 28) {
                                                            if (tcor < 89) {
                                                                if (MVDiff < 14) {
                                                                    if (Rs < 5912) {
                                                                        if (mcTcor < 82) {
                                                                            if (Cs < 4127) {
                                                                                return 4;
                                                                            }
                                                                            else if (Cs >= 4127) {
                                                                                return 8;
                                                                            }
                                                                        }
                                                                        else if (mcTcor >= 82) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                    else if (Rs >= 5912) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (MVDiff >= 14) {
                                                                    if (refDCval < 96) {
                                                                        return 8;
                                                                    }
                                                                    else if (refDCval >= 96) {
                                                                        if (ssDCval < 139) {
                                                                            return 4;
                                                                        }
                                                                        else if (ssDCval >= 139) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                            else if (tcor >= 89) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (RsDiff >= 28) {
                                                            if (Rs < 5455) {
                                                                return 2;
                                                            }
                                                            else if (Rs >= 5455) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                    else if (ssDCval >= 140) {
                                                        return 8;
                                                    }
                                                }
                                                else if (Cs >= 4323) {
                                                    if (SC < 11811) {
                                                        if (mcTcor < 93) {
                                                            return 8;
                                                        }
                                                        else if (mcTcor >= 93) {
                                                            if (RsDiff < 22) {
                                                                return 8;
                                                            }
                                                            else if (RsDiff >= 22) {
                                                                if (Cs < 5020) {
                                                                    return 1;
                                                                }
                                                                else if (Cs >= 5020) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (SC >= 11811) {
                                                        if (tcor < 80) {
                                                            return 8;
                                                        }
                                                        else if (tcor >= 80) {
                                                            if (diffRsCsdiff < -10) {
                                                                return 4;
                                                            }
                                                            else if (diffRsCsdiff >= -10) {
                                                                if (SC < 11988) {
                                                                    return 1;
                                                                }
                                                                else if (SC >= 11988) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (Rs >= 6965) {
                                        if (refDCval < 80) {
                                            if (mu_mv_mag_sq < 62) {
                                                if (tcor < 62) {
                                                    return 8;
                                                }
                                                else if (tcor >= 62) {
                                                    return 4;
                                                }
                                            }
                                            else if (mu_mv_mag_sq >= 62) {
                                                if (Cs < 6692) {
                                                    return 4;
                                                }
                                                else if (Cs >= 6692) {
                                                    return 8;
                                                }
                                            }
                                        }
                                        else if (refDCval >= 80) {
                                            if (MVDiff < 22) {
                                                if (Rs < 14727) {
                                                    if (RsDiff < 37) {
                                                        if (CsDiff < 21) {
                                                            return 4;
                                                        }
                                                        else if (CsDiff >= 21) {
                                                            if (refDCval < 89) {
                                                                if (Rs < 8403) {
                                                                    return 4;
                                                                }
                                                                else if (Rs >= 8403) {
                                                                    return 2;
                                                                }
                                                            }
                                                            else if (refDCval >= 89) {
                                                                if (Cs < 6107) {
                                                                    return 4;
                                                                }
                                                                else if (Cs >= 6107) {
                                                                    if (negBalance < 49) {
                                                                        return 4;
                                                                    }
                                                                    else if (negBalance >= 49) {
                                                                        if (gchDC < 4) {
                                                                            return 8;
                                                                        }
                                                                        else if (gchDC >= 4) {
                                                                            if (Rs < 7345) {
                                                                                return 8;
                                                                            }
                                                                            else if (Rs >= 7345) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (RsDiff >= 37) {
                                                        if (posBalance < 60) {
                                                            if (ssDCval < 107) {
                                                                return 8;
                                                            }
                                                            else if (ssDCval >= 107) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (posBalance >= 60) {
                                                            if (AFD < 47) {
                                                                if (Cs < 8915) {
                                                                    return 4;
                                                                }
                                                                else if (Cs >= 8915) {
                                                                    return 2;
                                                                }
                                                            }
                                                            else if (AFD >= 47) {
                                                                return 2;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (Rs >= 14727) {
                                                    if (RsDiff < 69) {
                                                        return 8;
                                                    }
                                                    else if (RsDiff >= 69) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                            else if (MVDiff >= 22) {
                                                if (mcTcor < 84) {
                                                    return 8;
                                                }
                                                else if (mcTcor >= 84) {
                                                    return 2;
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (refDCval >= 149) {
                                    if (negBalance < 54) {
                                        if (mcTcor < 87) {
                                            return 8;
                                        }
                                        else if (mcTcor >= 87) {
                                            if (AFD < 26) {
                                                return 8;
                                            }
                                            else if (AFD >= 26) {
                                                if (posBalance < 56) {
                                                    return 4;
                                                }
                                                else if (posBalance >= 56) {
                                                    return 2;
                                                }
                                            }
                                        }
                                    }
                                    else if (negBalance >= 54) {
                                        if (TSC < 16) {
                                            if (Rs < 5532) {
                                                if (RsCsDiff < 569) {
                                                    if (Cs < 3813) {
                                                        return 2;
                                                    }
                                                    else if (Cs >= 3813) {
                                                        return 8;
                                                    }
                                                }
                                                else if (RsCsDiff >= 569) {
                                                    return 4;
                                                }
                                            }
                                            else if (Rs >= 5532) {
                                                if (MVDiff < 31) {
                                                    return 2;
                                                }
                                                else if (MVDiff >= 31) {
                                                    return 4;
                                                }
                                            }
                                        }
                                        else if (TSC >= 16) {
                                            if (Cs < 3993) {
                                                if (ssDCval < 160) {
                                                    return 4;
                                                }
                                                else if (ssDCval >= 160) {
                                                    return 2;
                                                }
                                            }
                                            else if (Cs >= 3993) {
                                                if (diffRsCsdiff < 559) {
                                                    if (AFD < 34) {
                                                        if (refDCval < 155) {
                                                            return 8;
                                                        }
                                                        else if (refDCval >= 155) {
                                                            if (diffTSC < 2) {
                                                                return 2;
                                                            }
                                                            else if (diffTSC >= 2) {
                                                                if (mcTcor < 83) {
                                                                    return 2;
                                                                }
                                                                else if (mcTcor >= 83) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (AFD >= 34) {
                                                        return 2;
                                                    }
                                                }
                                                else if (diffRsCsdiff >= 559) {
                                                    return 4;
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                            else if (mu_mv_mag_sq >= 63) {
                                if (posBalance < 58) {
                                    if (AFD < 16) {
                                        if (diffAFD < 1) {
                                            if (ssDCval < 127) {
                                                if (Rs < 4666) {
                                                    return 4;
                                                }
                                                else if (Rs >= 4666) {
                                                    if (Rs < 6407) {
                                                        return 2;
                                                    }
                                                    else if (Rs >= 6407) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                            else if (ssDCval >= 127) {
                                                return 8;
                                            }
                                        }
                                        else if (diffAFD >= 1) {
                                            return 8;
                                        }
                                    }
                                    else if (AFD >= 16) {
                                        if (refDCval < 135) {
                                            if (negBalance < 64) {
                                                return 8;
                                            }
                                            else if (negBalance >= 64) {
                                                if (CsDiff < 21) {
                                                    return 4;
                                                }
                                                else if (CsDiff >= 21) {
                                                    return 8;
                                                }
                                            }
                                        }
                                        else if (refDCval >= 135) {
                                            if (Rs < 4975) {
                                                return 1;
                                            }
                                            else if (Rs >= 4975) {
                                                return 8;
                                            }
                                        }
                                    }
                                }
                                else if (posBalance >= 58) {
                                    if (Cs < 3624) {
                                        if (Cs < 3566) {
                                            return 8;
                                        }
                                        else if (Cs >= 3566) {
                                            return 4;
                                        }
                                    }
                                    else if (Cs >= 3624) {
                                        return 8;
                                    }
                                }
                            }
                        }
                    }
                }
            }
        }
        else if (mu_mv_mag_sq >= 89) {
            if (AFD < 7) {
                if (SC < 1223) {
                    if (ssDCval < 161) {
                        if (refDCval < 92) {
                            if (Cs < 512) {
                                if (mu_mv_mag_sq < 113) {
                                    if (ssDCval < 83) {
                                        if (tcor < 97) {
                                            if (mcTcor < 97) {
                                                if (RsDiff < 3) {
                                                    return 4;
                                                }
                                                else if (RsDiff >= 3) {
                                                    return 1;
                                                }
                                            }
                                            else if (mcTcor >= 97) {
                                                return 1;
                                            }
                                        }
                                        else if (tcor >= 97) {
                                            if (AFD < 2) {
                                                if (Rs < 61) {
                                                    return 4;
                                                }
                                                else if (Rs >= 61) {
                                                    return 8;
                                                }
                                            }
                                            else if (AFD >= 2) {
                                                return 4;
                                            }
                                        }
                                    }
                                    else if (ssDCval >= 83) {
                                        if (RsCsDiff < 2) {
                                            return 8;
                                        }
                                        else if (RsCsDiff >= 2) {
                                            return 4;
                                        }
                                    }
                                }
                                else if (mu_mv_mag_sq >= 113) {
                                    if (SC < 621) {
                                        if (diffMVdiffVal < 64) {
                                            if (negBalance < 13) {
                                                return 8;
                                            }
                                            else if (negBalance >= 13) {
                                                if (Cs < 304) {
                                                    return 8;
                                                }
                                                else if (Cs >= 304) {
                                                    if (RsDiff < 1) {
                                                        if (Cs < 310) {
                                                            return 4;
                                                        }
                                                        else if (Cs >= 310) {
                                                            if (Rs < 182) {
                                                                return 8;
                                                            }
                                                            else if (Rs >= 182) {
                                                                if (Rs < 187) {
                                                                    if (Cs < 318) {
                                                                        return 4;
                                                                    }
                                                                    else if (Cs >= 318) {
                                                                        return 8;
                                                                    }
                                                                }
                                                                else if (Rs >= 187) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (RsDiff >= 1) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                        }
                                        else if (diffMVdiffVal >= 64) {
                                            if (Cs < 1) {
                                                return 4;
                                            }
                                            else if (Cs >= 1) {
                                                return 1;
                                            }
                                        }
                                    }
                                    else if (SC >= 621) {
                                        return 8;
                                    }
                                }
                            }
                            else if (Cs >= 512) {
                                if (ssDCval < 36) {
                                    return 8;
                                }
                                else if (ssDCval >= 36) {
                                    return 4;
                                }
                            }
                        }
                        else if (refDCval >= 92) {
                            if (Cs < 658) {
                                if (negBalance < 31) {
                                    if (diffAFD < 2) {
                                        return 8;
                                    }
                                    else if (diffAFD >= 2) {
                                        if (Cs < 409) {
                                            return 8;
                                        }
                                        else if (Cs >= 409) {
                                            return 4;
                                        }
                                    }
                                }
                                else if (negBalance >= 31) {
                                    return 4;
                                }
                            }
                            else if (Cs >= 658) {
                                if (Rs < 199) {
                                    return 8;
                                }
                                else if (Rs >= 199) {
                                    return 2;
                                }
                            }
                        }
                    }
                    else if (ssDCval >= 161) {
                        if (Cs < 438) {
                            return 4;
                        }
                        else if (Cs >= 438) {
                            if (posBalance < 25) {
                                return 4;
                            }
                            else if (posBalance >= 25) {
                                if (diffMVdiffVal < -1) {
                                    return 2;
                                }
                                else if (diffMVdiffVal >= -1) {
                                    if (Cs < 452) {
                                        return 2;
                                    }
                                    else if (Cs >= 452) {
                                        return 4;
                                    }
                                }
                            }
                        }
                    }
                }
                else if (SC >= 1223) {
                    if (posBalance < 51) {
                        if (Rs < 3779) {
                            if (SC < 7764) {
                                if (diffMVdiffVal < -3) {
                                    if (ssDCval < 117) {
                                        if (RsDiff < 3) {
                                            if (tcor < 97) {
                                                if (diffRsCsdiff < 1) {
                                                    return 8;
                                                }
                                                else if (diffRsCsdiff >= 1) {
                                                    return 2;
                                                }
                                            }
                                            else if (tcor >= 97) {
                                                if (MVDiff < 1) {
                                                    if (diffMVdiffVal < -32) {
                                                        if (Rs < 2373) {
                                                            return 4;
                                                        }
                                                        else if (Rs >= 2373) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (diffMVdiffVal >= -32) {
                                                        if (SC < 1630) {
                                                            return 4;
                                                        }
                                                        else if (SC >= 1630) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                                else if (MVDiff >= 1) {
                                                    if (ssDCval < 32) {
                                                        if (ssDCval < 30) {
                                                            if (Cs < 783) {
                                                                return 4;
                                                            }
                                                            else if (Cs >= 783) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (ssDCval >= 30) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (ssDCval >= 32) {
                                                        if (refDCval < 90) {
                                                            if (Cs < 1820) {
                                                                if (RsDiff < 1) {
                                                                    return 1;
                                                                }
                                                                else if (RsDiff >= 1) {
                                                                    if (mcTcor < 98) {
                                                                        return 1;
                                                                    }
                                                                    else if (mcTcor >= 98) {
                                                                        return 8;
                                                                    }
                                                                }
                                                            }
                                                            else if (Cs >= 1820) {
                                                                return 1;
                                                            }
                                                        }
                                                        else if (refDCval >= 90) {
                                                            if (TSC < 4) {
                                                                return 8;
                                                            }
                                                            else if (TSC >= 4) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (RsDiff >= 3) {
                                            return 8;
                                        }
                                    }
                                    else if (ssDCval >= 117) {
                                        if (SC < 2821) {
                                            if (gchDC < 1) {
                                                if (Cs < 733) {
                                                    return 1;
                                                }
                                                else if (Cs >= 733) {
                                                    return 8;
                                                }
                                            }
                                            else if (gchDC >= 1) {
                                                return 1;
                                            }
                                        }
                                        else if (SC >= 2821) {
                                            if (TSC < 2) {
                                                return 8;
                                            }
                                            else if (TSC >= 2) {
                                                return 4;
                                            }
                                        }
                                    }
                                }
                                else if (diffMVdiffVal >= -3) {
                                    if (refDCval < 70) {
                                        if (MVDiff < 3) {
                                            if (refDCval < 28) {
                                                if (Cs < 713) {
                                                    return 4;
                                                }
                                                else if (Cs >= 713) {
                                                    return 8;
                                                }
                                            }
                                            else if (refDCval >= 28) {
                                                if (diffAFD < -2) {
                                                    if (SC < 2833) {
                                                        if (Rs < 1173) {
                                                            return 8;
                                                        }
                                                        else if (Rs >= 1173) {
                                                            return 1;
                                                        }
                                                    }
                                                    else if (SC >= 2833) {
                                                        return 8;
                                                    }
                                                }
                                                else if (diffAFD >= -2) {
                                                    if (posBalance < 50) {
                                                        if (diffRsCsdiff < -17) {
                                                            if (posBalance < 37) {
                                                                return 8;
                                                            }
                                                            else if (posBalance >= 37) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (diffRsCsdiff >= -17) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (posBalance >= 50) {
                                                        if (diffMVdiffVal < -1) {
                                                            return 1;
                                                        }
                                                        else if (diffMVdiffVal >= -1) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (MVDiff >= 3) {
                                            if (RsDiff < 1) {
                                                if (SC < 2303) {
                                                    if (Cs < 919) {
                                                        if (Cs < 708) {
                                                            return 4;
                                                        }
                                                        else if (Cs >= 708) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (Cs >= 919) {
                                                        if (Cs < 1052) {
                                                            return 1;
                                                        }
                                                        else if (Cs >= 1052) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                                else if (SC >= 2303) {
                                                    if (Rs < 1228) {
                                                        return 2;
                                                    }
                                                    else if (Rs >= 1228) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                            else if (RsDiff >= 1) {
                                                if (RsCsDiff < 19) {
                                                    if (diffAFD < -1) {
                                                        if (Cs < 815) {
                                                            return 8;
                                                        }
                                                        else if (Cs >= 815) {
                                                            if (Rs < 1146) {
                                                                return 8;
                                                            }
                                                            else if (Rs >= 1146) {
                                                                return 1;
                                                            }
                                                        }
                                                    }
                                                    else if (diffAFD >= -1) {
                                                        return 8;
                                                    }
                                                }
                                                else if (RsCsDiff >= 19) {
                                                    if (posBalance < 47) {
                                                        if (SC < 1857) {
                                                            return 4;
                                                        }
                                                        else if (SC >= 1857) {
                                                            if (AFD < 6) {
                                                                if (MVDiff < 24) {
                                                                    return 8;
                                                                }
                                                                else if (MVDiff >= 24) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (AFD >= 6) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                    else if (posBalance >= 47) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (refDCval >= 70) {
                                        if (posBalance < 28) {
                                            if (Rs < 3730) {
                                                if (Cs < 3716) {
                                                    if (AFD < 4) {
                                                        if (mu_mv_mag_sq < 127) {
                                                            if (Cs < 2181) {
                                                                if (negBalance < 1) {
                                                                    return 1;
                                                                }
                                                                else if (negBalance >= 1) {
                                                                    if (Cs < 1952) {
                                                                        if (diffRsCsdiff < 12) {
                                                                            if (diffAFD < -1) {
                                                                                if (mu_mv_mag_sq < 121) {
                                                                                    return 8;
                                                                                }
                                                                                else if (mu_mv_mag_sq >= 121) {
                                                                                    if (Cs < 1387) {
                                                                                        return 1;
                                                                                    }
                                                                                    else if (Cs >= 1387) {
                                                                                        return 8;
                                                                                    }
                                                                                }
                                                                            }
                                                                            else if (diffAFD >= -1) {
                                                                                return 8;
                                                                            }
                                                                        }
                                                                        else if (diffRsCsdiff >= 12) {
                                                                            if (diffAFD < 2) {
                                                                                return 1;
                                                                            }
                                                                            else if (diffAFD >= 2) {
                                                                                return 8;
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (Cs >= 1952) {
                                                                        return 1;
                                                                    }
                                                                }
                                                            }
                                                            else if (Cs >= 2181) {
                                                                if (posBalance < 11) {
                                                                    if (refDCval < 116) {
                                                                        return 8;
                                                                    }
                                                                    else if (refDCval >= 116) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (posBalance >= 11) {
                                                                    if (tcor < 86) {
                                                                        return 4;
                                                                    }
                                                                    else if (tcor >= 86) {
                                                                        if (CsDiff < 3) {
                                                                            return 8;
                                                                        }
                                                                        else if (CsDiff >= 3) {
                                                                            if (Rs < 3675) {
                                                                                if (Rs < 2705) {
                                                                                    return 1;
                                                                                }
                                                                                else if (Rs >= 2705) {
                                                                                    return 8;
                                                                                }
                                                                            }
                                                                            else if (Rs >= 3675) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (mu_mv_mag_sq >= 127) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (AFD >= 4) {
                                                        if (TSCindex < 3) {
                                                            return 4;
                                                        }
                                                        else if (TSCindex >= 3) {
                                                            if (mcTcor < 97) {
                                                                if (diffAFD < 1) {
                                                                    if (tcor < 87) {
                                                                        return 8;
                                                                    }
                                                                    else if (tcor >= 87) {
                                                                        if (diffRsCsdiff < 1) {
                                                                            if (diffRsCsdiff < -4) {
                                                                                if (negBalance < 14) {
                                                                                    return 1;
                                                                                }
                                                                                else if (negBalance >= 14) {
                                                                                    return 8;
                                                                                }
                                                                            }
                                                                            else if (diffRsCsdiff >= -4) {
                                                                                return 1;
                                                                            }
                                                                        }
                                                                        else if (diffRsCsdiff >= 1) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                }
                                                                else if (diffAFD >= 1) {
                                                                    return 1;
                                                                }
                                                            }
                                                            else if (mcTcor >= 97) {
                                                                if (diffAFD < -1) {
                                                                    return 4;
                                                                }
                                                                else if (diffAFD >= -1) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (Cs >= 3716) {
                                                    if (diffRsCsdiff < 2) {
                                                        if (SC < 7428) {
                                                            return 1;
                                                        }
                                                        else if (SC >= 7428) {
                                                            if (posBalance < 21) {
                                                                return 1;
                                                            }
                                                            else if (posBalance >= 21) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                    else if (diffRsCsdiff >= 2) {
                                                        return 1;
                                                    }
                                                }
                                            }
                                            else if (Rs >= 3730) {
                                                return 8;
                                            }
                                        }
                                        else if (posBalance >= 28) {
                                            if (refDCval < 110) {
                                                if (RsCsDiff < 11) {
                                                    if (SC < 1792) {
                                                        return 4;
                                                    }
                                                    else if (SC >= 1792) {
                                                        if (AFD < 3) {
                                                            return 8;
                                                        }
                                                        else if (AFD >= 3) {
                                                            if (Rs < 2149) {
                                                                if (Scindex < 5) {
                                                                    if (AFD < 4) {
                                                                        return 8;
                                                                    }
                                                                    else if (AFD >= 4) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (Scindex >= 5) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (Rs >= 2149) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (RsCsDiff >= 11) {
                                                    if (mu_mv_mag_sq < 125) {
                                                        if (negBalance < 38) {
                                                            if (Rs < 3436) {
                                                                if (Cs < 3076) {
                                                                    if (posBalance < 48) {
                                                                        if (SC < 6062) {
                                                                            if (mu_mv_mag_sq < 97) {
                                                                                return 1;
                                                                            }
                                                                            else if (mu_mv_mag_sq >= 97) {
                                                                                if (SC < 4233) {
                                                                                    return 8;
                                                                                }
                                                                                else if (SC >= 4233) {
                                                                                    if (ssDCval < 109) {
                                                                                        if (gchDC < 1) {
                                                                                            if (Cs < 2204) {
                                                                                                return 1;
                                                                                            }
                                                                                            else if (Cs >= 2204) {
                                                                                                if (Cs < 2892) {
                                                                                                    return 8;
                                                                                                }
                                                                                                else if (Cs >= 2892) {
                                                                                                    return 1;
                                                                                                }
                                                                                            }
                                                                                        }
                                                                                        else if (gchDC >= 1) {
                                                                                            return 8;
                                                                                        }
                                                                                    }
                                                                                    else if (ssDCval >= 109) {
                                                                                        if (AFD < 4) {
                                                                                            return 8;
                                                                                        }
                                                                                        else if (AFD >= 4) {
                                                                                            if (Cs < 2422) {
                                                                                                return 8;
                                                                                            }
                                                                                            else if (Cs >= 2422) {
                                                                                                return 1;
                                                                                            }
                                                                                        }
                                                                                    }
                                                                                }
                                                                            }
                                                                        }
                                                                        else if (SC >= 6062) {
                                                                            if (negBalance < 26) {
                                                                                if (tcor < 99) {
                                                                                    if (Rs < 3309) {
                                                                                        return 8;
                                                                                    }
                                                                                    else if (Rs >= 3309) {
                                                                                        return 1;
                                                                                    }
                                                                                }
                                                                                else if (tcor >= 99) {
                                                                                    return 1;
                                                                                }
                                                                            }
                                                                            else if (negBalance >= 26) {
                                                                                return 1;
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (posBalance >= 48) {
                                                                        return 8;
                                                                    }
                                                                }
                                                                else if (Cs >= 3076) {
                                                                    if (AFD < 3) {
                                                                        if (RsCsDiff < 16) {
                                                                            return 1;
                                                                        }
                                                                        else if (RsCsDiff >= 16) {
                                                                            if (RsDiff < 4) {
                                                                                return 8;
                                                                            }
                                                                            else if (RsDiff >= 4) {
                                                                                return 1;
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (AFD >= 3) {
                                                                        return 8;
                                                                    }
                                                                }
                                                            }
                                                            else if (Rs >= 3436) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (negBalance >= 38) {
                                                            if (MVDiff < 4) {
                                                                return 4;
                                                            }
                                                            else if (MVDiff >= 4) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                    else if (mu_mv_mag_sq >= 125) {
                                                        if (Rs < 1774) {
                                                            return 4;
                                                        }
                                                        else if (Rs >= 1774) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (refDCval >= 110) {
                                                if (SC < 2750) {
                                                    if (RsDiff < 2) {
                                                        if (AFD < 3) {
                                                            if (Cs < 1363) {
                                                                return 4;
                                                            }
                                                            else if (Cs >= 1363) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (AFD >= 3) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (RsDiff >= 2) {
                                                        if (MVDiff < 1) {
                                                            return 4;
                                                        }
                                                        else if (MVDiff >= 1) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                                else if (SC >= 2750) {
                                                    if (Rs < 3749) {
                                                        return 8;
                                                    }
                                                    else if (Rs >= 3749) {
                                                        if (Rs < 3769) {
                                                            return 4;
                                                        }
                                                        else if (Rs >= 3769) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                            else if (SC >= 7764) {
                                return 1;
                            }
                        }
                        else if (Rs >= 3779) {
                            if (mu_mv_mag_sq < 92) {
                                if (Rs < 4144) {
                                    return 8;
                                }
                                else if (Rs >= 4144) {
                                    return 4;
                                }
                            }
                            else if (mu_mv_mag_sq >= 92) {
                                return 8;
                            }
                        }
                    }
                    else if (posBalance >= 51) {
                        if (ssDCval < 84) {
                            if (Rs < 1768) {
                                if (TSC < 6) {
                                    if (mcTcor < 88) {
                                        if (SC < 1433) {
                                            return 4;
                                        }
                                        else if (SC >= 1433) {
                                            return 8;
                                        }
                                    }
                                    else if (mcTcor >= 88) {
                                        if (Cs < 1717) {
                                            if (MVDiff < 9) {
                                                return 8;
                                            }
                                            else if (MVDiff >= 9) {
                                                if (Rs < 1368) {
                                                    if (TSC < 5) {
                                                        return 8;
                                                    }
                                                    else if (TSC >= 5) {
                                                        if (RsCsDiff < 5) {
                                                            return 1;
                                                        }
                                                        else if (RsCsDiff >= 5) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                                else if (Rs >= 1368) {
                                                    return 4;
                                                }
                                            }
                                        }
                                        else if (Cs >= 1717) {
                                            if (ssDCval < 77) {
                                                if (negBalance < 43) {
                                                    return 4;
                                                }
                                                else if (negBalance >= 43) {
                                                    return 8;
                                                }
                                            }
                                            else if (ssDCval >= 77) {
                                                return 8;
                                            }
                                        }
                                    }
                                }
                                else if (TSC >= 6) {
                                    if (ssDCval < 72) {
                                        if (Cs < 988) {
                                            return 8;
                                        }
                                        else if (Cs >= 988) {
                                            return 2;
                                        }
                                    }
                                    else if (ssDCval >= 72) {
                                        if (tcor < 96) {
                                            return 8;
                                        }
                                        else if (tcor >= 96) {
                                            return 2;
                                        }
                                    }
                                }
                            }
                            else if (Rs >= 1768) {
                                if (tcor < 95) {
                                    if (SC < 9679) {
                                        if (CsDiff < 3) {
                                            return 4;
                                        }
                                        else if (CsDiff >= 3) {
                                            if (mcTcor < 95) {
                                                if (mu_mv_mag_sq < 128) {
                                                    if (mcTcor < 94) {
                                                        return 8;
                                                    }
                                                    else if (mcTcor >= 94) {
                                                        if (MVDiff < 1) {
                                                            if (CsDiff < 5) {
                                                                if (diffRsCsdiff < 4) {
                                                                    if (Rs < 1808) {
                                                                        if (Cs < 1549) {
                                                                            return 8;
                                                                        }
                                                                        else if (Cs >= 1549) {
                                                                            return 2;
                                                                        }
                                                                    }
                                                                    else if (Rs >= 1808) {
                                                                        return 8;
                                                                    }
                                                                }
                                                                else if (diffRsCsdiff >= 4) {
                                                                    if (tcor < 94) {
                                                                        return 2;
                                                                    }
                                                                    else if (tcor >= 94) {
                                                                        return 4;
                                                                    }
                                                                }
                                                            }
                                                            else if (CsDiff >= 5) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (MVDiff >= 1) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                                else if (mu_mv_mag_sq >= 128) {
                                                    if (SC < 9403) {
                                                        return 4;
                                                    }
                                                    else if (SC >= 9403) {
                                                        if (gchDC < 1) {
                                                            return 8;
                                                        }
                                                        else if (gchDC >= 1) {
                                                            if (Rs < 4706) {
                                                                return 4;
                                                            }
                                                            else if (Rs >= 4706) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                            else if (mcTcor >= 95) {
                                                if (MVDiff < 1) {
                                                    if (diffRsCsdiff < -12) {
                                                        return 8;
                                                    }
                                                    else if (diffRsCsdiff >= -12) {
                                                        if (Rs < 1831) {
                                                            if (Cs < 1740) {
                                                                return 2;
                                                            }
                                                            else if (Cs >= 1740) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (Rs >= 1831) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                                else if (MVDiff >= 1) {
                                                    return 8;
                                                }
                                            }
                                        }
                                    }
                                    else if (SC >= 9679) {
                                        if (Rs < 4974) {
                                            if (mu_mv_mag_sq < 128) {
                                                if (RsDiff < 8) {
                                                    return 4;
                                                }
                                                else if (RsDiff >= 8) {
                                                    if (diffRsCsdiff < 25) {
                                                        return 8;
                                                    }
                                                    else if (diffRsCsdiff >= 25) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                            else if (mu_mv_mag_sq >= 128) {
                                                if (ssDCval < 80) {
                                                    return 8;
                                                }
                                                else if (ssDCval >= 80) {
                                                    if (CsDiff < 9) {
                                                        if (Cs < 5124) {
                                                            if (Cs < 4932) {
                                                                if (Cs < 4921) {
                                                                    return 4;
                                                                }
                                                                else if (Cs >= 4921) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (Cs >= 4932) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (Cs >= 5124) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (CsDiff >= 9) {
                                                        if (Rs < 4900) {
                                                            return 4;
                                                        }
                                                        else if (Rs >= 4900) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (Rs >= 4974) {
                                            if (RsDiff < 9) {
                                                return 4;
                                            }
                                            else if (RsDiff >= 9) {
                                                if (Cs < 5172) {
                                                    return 8;
                                                }
                                                else if (Cs >= 5172) {
                                                    return 4;
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (tcor >= 95) {
                                    if (Cs < 1650) {
                                        if (Cs < 1217) {
                                            return 8;
                                        }
                                        else if (Cs >= 1217) {
                                            if (ssDCval < 76) {
                                                if (Cs < 1380) {
                                                    if (ssDCval < 73) {
                                                        if (Cs < 1272) {
                                                            return 8;
                                                        }
                                                        else if (Cs >= 1272) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (ssDCval >= 73) {
                                                        return 4;
                                                    }
                                                }
                                                else if (Cs >= 1380) {
                                                    if (Rs < 1898) {
                                                        if (mu_mv_mag_sq < 119) {
                                                            return 4;
                                                        }
                                                        else if (mu_mv_mag_sq >= 119) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (Rs >= 1898) {
                                                        if (RsCsDiff < 22) {
                                                            return 2;
                                                        }
                                                        else if (RsCsDiff >= 22) {
                                                            if (RsCsDiff < 29) {
                                                                return 4;
                                                            }
                                                            else if (RsCsDiff >= 29) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                            else if (ssDCval >= 76) {
                                                return 4;
                                            }
                                        }
                                    }
                                    else if (Cs >= 1650) {
                                        if (Scindex < 6) {
                                            if (diffRsCsdiff < -8) {
                                                if (posBalance < 58) {
                                                    return 4;
                                                }
                                                else if (posBalance >= 58) {
                                                    return 8;
                                                }
                                            }
                                            else if (diffRsCsdiff >= -8) {
                                                return 4;
                                            }
                                        }
                                        else if (Scindex >= 6) {
                                            if (ssDCval < 55) {
                                                return 4;
                                            }
                                            else if (ssDCval >= 55) {
                                                return 8;
                                            }
                                        }
                                    }
                                }
                            }
                        }
                        else if (ssDCval >= 84) {
                            if (negBalance < 39) {
                                if (posBalance < 58) {
                                    if (AFD < 4) {
                                        if (ssDCval < 103) {
                                            return 8;
                                        }
                                        else if (ssDCval >= 103) {
                                            if (ssDCval < 108) {
                                                return 4;
                                            }
                                            else if (ssDCval >= 108) {
                                                return 8;
                                            }
                                        }
                                    }
                                    else if (AFD >= 4) {
                                        if (SC < 4904) {
                                            if (Cs < 1804) {
                                                if (SC < 2655) {
                                                    return 4;
                                                }
                                                else if (SC >= 2655) {
                                                    return 8;
                                                }
                                            }
                                            else if (Cs >= 1804) {
                                                return 4;
                                            }
                                        }
                                        else if (SC >= 4904) {
                                            if (RsCsDiff < 121) {
                                                return 8;
                                            }
                                            else if (RsCsDiff >= 121) {
                                                return 4;
                                            }
                                        }
                                    }
                                }
                                else if (posBalance >= 58) {
                                    return 8;
                                }
                            }
                            else if (negBalance >= 39) {
                                if (Cs < 3161) {
                                    if (ssDCval < 91) {
                                        if (mu_mv_mag_sq < 99) {
                                            return 1;
                                        }
                                        else if (mu_mv_mag_sq >= 99) {
                                            if (RsDiff < 7) {
                                                return 8;
                                            }
                                            else if (RsDiff >= 7) {
                                                return 4;
                                            }
                                        }
                                    }
                                    else if (ssDCval >= 91) {
                                        if (posBalance < 58) {
                                            if (mu_mv_mag_sq < 123) {
                                                if (ssDCval < 109) {
                                                    return 4;
                                                }
                                                else if (ssDCval >= 109) {
                                                    if (Cs < 2943) {
                                                        if (RsCsDiff < 29) {
                                                            if (ssDCval < 114) {
                                                                return 4;
                                                            }
                                                            else if (ssDCval >= 114) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (RsCsDiff >= 29) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (Cs >= 2943) {
                                                        if (ssDCval < 112) {
                                                            return 8;
                                                        }
                                                        else if (ssDCval >= 112) {
                                                            if (diffRsCsdiff < 8) {
                                                                if (refDCval < 115) {
                                                                    if (negBalance < 48) {
                                                                        if (Rs < 2867) {
                                                                            if (diffRsCsdiff < -5) {
                                                                                return 4;
                                                                            }
                                                                            else if (diffRsCsdiff >= -5) {
                                                                                return 8;
                                                                            }
                                                                        }
                                                                        else if (Rs >= 2867) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                    else if (negBalance >= 48) {
                                                                        if (mu_mv_mag_sq < 121) {
                                                                            return 4;
                                                                        }
                                                                        else if (mu_mv_mag_sq >= 121) {
                                                                            if (Cs < 3113) {
                                                                                if (Rs < 2884) {
                                                                                    if (RsDiff < 5) {
                                                                                        if (SC < 5625) {
                                                                                            return 8;
                                                                                        }
                                                                                        else if (SC >= 5625) {
                                                                                            return 4;
                                                                                        }
                                                                                    }
                                                                                    else if (RsDiff >= 5) {
                                                                                        return 4;
                                                                                    }
                                                                                }
                                                                                else if (Rs >= 2884) {
                                                                                    return 8;
                                                                                }
                                                                            }
                                                                            else if (Cs >= 3113) {
                                                                                return 8;
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                                else if (refDCval >= 115) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (diffRsCsdiff >= 8) {
                                                                if (Rs < 2955) {
                                                                    return 4;
                                                                }
                                                                else if (Rs >= 2955) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                            else if (mu_mv_mag_sq >= 123) {
                                                if (SC < 5750) {
                                                    return 8;
                                                }
                                                else if (SC >= 5750) {
                                                    if (negBalance < 48) {
                                                        return 8;
                                                    }
                                                    else if (negBalance >= 48) {
                                                        if (diffRsCsdiff < -5) {
                                                            if (tcor < 98) {
                                                                if (SC < 6055) {
                                                                    return 4;
                                                                }
                                                                else if (SC >= 6055) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (tcor >= 98) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (diffRsCsdiff >= -5) {
                                                            if (SC < 5761) {
                                                                return 4;
                                                            }
                                                            else if (SC >= 5761) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (posBalance >= 58) {
                                            if (tcor < 90) {
                                                if (Cs < 2042) {
                                                    return 8;
                                                }
                                                else if (Cs >= 2042) {
                                                    return 4;
                                                }
                                            }
                                            else if (tcor >= 90) {
                                                if (ssDCval < 93) {
                                                    return 4;
                                                }
                                                else if (ssDCval >= 93) {
                                                    if (Cs < 3150) {
                                                        if (refDCval < 113) {
                                                            return 8;
                                                        }
                                                        else if (refDCval >= 113) {
                                                            if (mu_mv_mag_sq < 121) {
                                                                if (ssDCval < 114) {
                                                                    if (posBalance < 59) {
                                                                        return 4;
                                                                    }
                                                                    else if (posBalance >= 59) {
                                                                        return 8;
                                                                    }
                                                                }
                                                                else if (ssDCval >= 114) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (mu_mv_mag_sq >= 121) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                    else if (Cs >= 3150) {
                                                        if (RsCsDiff < 37) {
                                                            return 8;
                                                        }
                                                        else if (RsCsDiff >= 37) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (Cs >= 3161) {
                                    return 8;
                                }
                            }
                        }
                    }
                }
            }
            else if (AFD >= 7) {
                if (Cs < 1672) {
                    if (RsCsDiff < 33) {
                        if (posBalance < 53) {
                            if (ssDCval < 107) {
                                if (Rs < 649) {
                                    return 1;
                                }
                                else if (Rs >= 649) {
                                    return 4;
                                }
                            }
                            else if (ssDCval >= 107) {
                                return 8;
                            }
                        }
                        else if (posBalance >= 53) {
                            if (CsDiff < 4) {
                                return 2;
                            }
                            else if (CsDiff >= 4) {
                                if (mcTcor < 93) {
                                    if (refDCval < 71) {
                                        return 2;
                                    }
                                    else if (refDCval >= 71) {
                                        if (diffRsCsdiff < 4) {
                                            if (Rs < 1899) {
                                                return 8;
                                            }
                                            else if (Rs >= 1899) {
                                                return 2;
                                            }
                                        }
                                        else if (diffRsCsdiff >= 4) {
                                            return 2;
                                        }
                                    }
                                }
                                else if (mcTcor >= 93) {
                                    if (diffAFD < 1) {
                                        return 2;
                                    }
                                    else if (diffAFD >= 1) {
                                        if (TSCindex < 8) {
                                            if (mu_mv_mag_sq < 111) {
                                                return 2;
                                            }
                                            else if (mu_mv_mag_sq >= 111) {
                                                return 4;
                                            }
                                        }
                                        else if (TSCindex >= 8) {
                                            return 2;
                                        }
                                    }
                                }
                            }
                        }
                    }
                    else if (RsCsDiff >= 33) {
                        if (MVDiff < 27) {
                            if (SC < 2911) {
                                if (ssDCval < 63) {
                                    return 4;
                                }
                                else if (ssDCval >= 63) {
                                    return 8;
                                }
                            }
                            else if (SC >= 2911) {
                                if (ssDCval < 84) {
                                    if (mu_mv_mag_sq < 113) {
                                        return 2;
                                    }
                                    else if (mu_mv_mag_sq >= 113) {
                                        if (tcor < 91) {
                                            return 8;
                                        }
                                        else if (tcor >= 91) {
                                            if (Rs < 1643) {
                                                return 8;
                                            }
                                            else if (Rs >= 1643) {
                                                return 2;
                                            }
                                        }
                                    }
                                }
                                else if (ssDCval >= 84) {
                                    if (tcor < 95) {
                                        if (RsCsDiff < 98) {
                                            if (Cs < 1659) {
                                                return 8;
                                            }
                                            else if (Cs >= 1659) {
                                                return 4;
                                            }
                                        }
                                        else if (RsCsDiff >= 98) {
                                            if (tcor < 60) {
                                                if (Cs < 1414) {
                                                    return 2;
                                                }
                                                else if (Cs >= 1414) {
                                                    return 1;
                                                }
                                            }
                                            else if (tcor >= 60) {
                                                return 4;
                                            }
                                        }
                                    }
                                    else if (tcor >= 95) {
                                        if (Rs < 2429) {
                                            return 4;
                                        }
                                        else if (Rs >= 2429) {
                                            return 1;
                                        }
                                    }
                                }
                            }
                        }
                        else if (MVDiff >= 27) {
                            if (SC < 3464) {
                                if (diffTSC < 15) {
                                    if (ssDCval < 92) {
                                        if (AFD < 22) {
                                            return 4;
                                        }
                                        else if (AFD >= 22) {
                                            return 2;
                                        }
                                    }
                                    else if (ssDCval >= 92) {
                                        if (Cs < 260) {
                                            return 2;
                                        }
                                        else if (Cs >= 260) {
                                            if (mcTcor < 100) {
                                                return 1;
                                            }
                                            else if (mcTcor >= 100) {
                                                return 8;
                                            }
                                        }
                                    }
                                }
                                else if (diffTSC >= 15) {
                                    if (CsDiff < 9) {
                                        return 4;
                                    }
                                    else if (CsDiff >= 9) {
                                        return 8;
                                    }
                                }
                            }
                            else if (SC >= 3464) {
                                if (MVDiff < 145) {
                                    if (diffRsCsdiff < 94) {
                                        return 2;
                                    }
                                    else if (diffRsCsdiff >= 94) {
                                        if (ssDCval < 128) {
                                            return 2;
                                        }
                                        else if (ssDCval >= 128) {
                                            return 4;
                                        }
                                    }
                                }
                                else if (MVDiff >= 145) {
                                    return 1;
                                }
                            }
                        }
                    }
                }
                else if (Cs >= 1672) {
                    if (Cs < 4172) {
                        if (mu_mv_mag_sq < 128) {
                            if (Rs < 3410) {
                                if (refDCval < 86) {
                                    if (Rs < 1904) {
                                        if (refDCval < 81) {
                                            if (Cs < 2269) {
                                                return 8;
                                            }
                                            else if (Cs >= 2269) {
                                                return 1;
                                            }
                                        }
                                        else if (refDCval >= 81) {
                                            return 2;
                                        }
                                    }
                                    else if (Rs >= 1904) {
                                        if (negBalance < 39) {
                                            if (Rs < 2137) {
                                                return 1;
                                            }
                                            else if (Rs >= 2137) {
                                                if (TSC < 11) {
                                                    return 8;
                                                }
                                                else if (TSC >= 11) {
                                                    return 4;
                                                }
                                            }
                                        }
                                        else if (negBalance >= 39) {
                                            if (posBalance < 59) {
                                                if (mcTcor < 94) {
                                                    if (Rs < 3018) {
                                                        return 4;
                                                    }
                                                    else if (Rs >= 3018) {
                                                        return 8;
                                                    }
                                                }
                                                else if (mcTcor >= 94) {
                                                    return 4;
                                                }
                                            }
                                            else if (posBalance >= 59) {
                                                if (Cs < 1930) {
                                                    return 4;
                                                }
                                                else if (Cs >= 1930) {
                                                    return 8;
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (refDCval >= 86) {
                                    if (SC < 4709) {
                                        if (posBalance < 46) {
                                            if (mu_mv_mag_sq < 98) {
                                                return 8;
                                            }
                                            else if (mu_mv_mag_sq >= 98) {
                                                return 1;
                                            }
                                        }
                                        else if (posBalance >= 46) {
                                            if (ssDCval < 97) {
                                                if (tcor < 93) {
                                                    if (ssDCval < 93) {
                                                        return 8;
                                                    }
                                                    else if (ssDCval >= 93) {
                                                        return 2;
                                                    }
                                                }
                                                else if (tcor >= 93) {
                                                    if (negBalance < 38) {
                                                        return 8;
                                                    }
                                                    else if (negBalance >= 38) {
                                                        return 2;
                                                    }
                                                }
                                            }
                                            else if (ssDCval >= 97) {
                                                if (ssDCval < 111) {
                                                    if (Cs < 1833) {
                                                        return 2;
                                                    }
                                                    else if (Cs >= 1833) {
                                                        return 4;
                                                    }
                                                }
                                                else if (ssDCval >= 111) {
                                                    if (diffMVdiffVal < -2) {
                                                        return 4;
                                                    }
                                                    else if (diffMVdiffVal >= -2) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (SC >= 4709) {
                                        if (Cs < 3045) {
                                            return 8;
                                        }
                                        else if (Cs >= 3045) {
                                            if (diffMVdiffVal < 1) {
                                                if (ssDCval < 98) {
                                                    return 2;
                                                }
                                                else if (ssDCval >= 98) {
                                                    return 8;
                                                }
                                            }
                                            else if (diffMVdiffVal >= 1) {
                                                if (Rs < 3007) {
                                                    return 4;
                                                }
                                                else if (Rs >= 3007) {
                                                    return 8;
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                            else if (Rs >= 3410) {
                                if (refDCval < 127) {
                                    if (ssDCval < 94) {
                                        if (refDCval < 89) {
                                            if (Rs < 3988) {
                                                if (Rs < 3896) {
                                                    return 4;
                                                }
                                                else if (Rs >= 3896) {
                                                    if (Cs < 3588) {
                                                        return 4;
                                                    }
                                                    else if (Cs >= 3588) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                            else if (Rs >= 3988) {
                                                return 8;
                                            }
                                        }
                                        else if (refDCval >= 89) {
                                            if (TSC < 13) {
                                                return 2;
                                            }
                                            else if (TSC >= 13) {
                                                return 8;
                                            }
                                        }
                                    }
                                    else if (ssDCval >= 94) {
                                        if (Rs < 3991) {
                                            if (SC < 6986) {
                                                if (negBalance < 42) {
                                                    if (diffAFD < -5) {
                                                        return 4;
                                                    }
                                                    else if (diffAFD >= -5) {
                                                        return 8;
                                                    }
                                                }
                                                else if (negBalance >= 42) {
                                                    if (SC < 6443) {
                                                        return 8;
                                                    }
                                                    else if (SC >= 6443) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                            else if (SC >= 6986) {
                                                return 8;
                                            }
                                        }
                                        else if (Rs >= 3991) {
                                            if (negBalance < 48) {
                                                if (Rs < 4686) {
                                                    if (TSC < 10) {
                                                        if (Cs < 3809) {
                                                            if (SC < 6648) {
                                                                if (diffRsCsdiff < -5) {
                                                                    return 8;
                                                                }
                                                                else if (diffRsCsdiff >= -5) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (SC >= 6648) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (Cs >= 3809) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (TSC >= 10) {
                                                        if (Cs < 3374) {
                                                            return 4;
                                                        }
                                                        else if (Cs >= 3374) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                                else if (Rs >= 4686) {
                                                    if (diffTSC < -1) {
                                                        return 2;
                                                    }
                                                    else if (diffTSC >= -1) {
                                                        if (MVDiff < 3) {
                                                            return 2;
                                                        }
                                                        else if (MVDiff >= 3) {
                                                            if (diffRsCsdiff < 18) {
                                                                return 8;
                                                            }
                                                            else if (diffRsCsdiff >= 18) {
                                                                if (SC < 8465) {
                                                                    return 2;
                                                                }
                                                                else if (SC >= 8465) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                            else if (negBalance >= 48) {
                                                return 4;
                                            }
                                        }
                                    }
                                }
                                else if (refDCval >= 127) {
                                    if (tcor < 84) {
                                        if (MVDiff < 6) {
                                            return 8;
                                        }
                                        else if (MVDiff >= 6) {
                                            return 1;
                                        }
                                    }
                                    else if (tcor >= 84) {
                                        if (mu_mv_mag_sq < 90) {
                                            return 2;
                                        }
                                        else if (mu_mv_mag_sq >= 90) {
                                            return 8;
                                        }
                                    }
                                }
                            }
                        }
                        else if (mu_mv_mag_sq >= 128) {
                            if (Rs < 4309) {
                                if (ssDCval < 94) {
                                    if (Rs < 2704) {
                                        return 4;
                                    }
                                    else if (Rs >= 2704) {
                                        return 2;
                                    }
                                }
                                else if (ssDCval >= 94) {
                                    return 8;
                                }
                            }
                            else if (Rs >= 4309) {
                                return 4;
                            }
                        }
                    }
                    else if (Cs >= 4172) {
                        if (Cs < 7432) {
                            if (refDCval < 106) {
                                if (TSC < 8) {
                                    if (mcTcor < 93) {
                                        if (Rs < 4829) {
                                            if (Rs < 4728) {
                                                return 8;
                                            }
                                            else if (Rs >= 4728) {
                                                if (Cs < 5021) {
                                                    return 4;
                                                }
                                                else if (Cs >= 5021) {
                                                    return 8;
                                                }
                                            }
                                        }
                                        else if (Rs >= 4829) {
                                            if (Rs < 4935) {
                                                if (Rs < 4878) {
                                                    return 4;
                                                }
                                                else if (Rs >= 4878) {
                                                    return 8;
                                                }
                                            }
                                            else if (Rs >= 4935) {
                                                return 4;
                                            }
                                        }
                                    }
                                    else if (mcTcor >= 93) {
                                        return 8;
                                    }
                                }
                                else if (TSC >= 8) {
                                    return 8;
                                }
                            }
                            else if (refDCval >= 106) {
                                if (mu_mv_mag_sq < 122) {
                                    if (mcTcor < 70) {
                                        return 1;
                                    }
                                    else if (mcTcor >= 70) {
                                        return 8;
                                    }
                                }
                                else if (mu_mv_mag_sq >= 122) {
                                    if (posBalance < 55) {
                                        if (TSC < 11) {
                                            return 8;
                                        }
                                        else if (TSC >= 11) {
                                            return 4;
                                        }
                                    }
                                    else if (posBalance >= 55) {
                                        if (negBalance < 55) {
                                            return 4;
                                        }
                                        else if (negBalance >= 55) {
                                            if (Cs < 5313) {
                                                return 8;
                                            }
                                            else if (Cs >= 5313) {
                                                return 4;
                                            }
                                        }
                                    }
                                }
                            }
                        }
                        else if (Cs >= 7432) {
                            if (ssDCval < 95) {
                                if (AFD < 19) {
                                    return 2;
                                }
                                else if (AFD >= 19) {
                                    if (Cs < 8194) {
                                        if (RsCsDiff < 1015) {
                                            if (RsCsDiff < 950) {
                                                return 4;
                                            }
                                            else if (RsCsDiff >= 950) {
                                                if (gchDC < 1) {
                                                    return 2;
                                                }
                                                else if (gchDC >= 1) {
                                                    return 4;
                                                }
                                            }
                                        }
                                        else if (RsCsDiff >= 1015) {
                                            return 4;
                                        }
                                    }
                                    else if (Cs >= 8194) {
                                        return 2;
                                    }
                                }
                            }
                            else if (ssDCval >= 95) {
                                return 8;
                            }
                        }
                    }
                }
            }
        }
        return 4;
    }
    mfxU16 AGOPSelect14(mfxU32 Rs, mfxU32 CsDiff, mfxU32 TSC, mfxI32 diffAFD, mfxU32 Scindex, mfxU32 RsDiff, mfxI32 diffMVdiffVal, mfxU32 RsCsDiff, mfxU32 TSCindex, mfxI16 tcor, mfxU32 refDCval, mfxU32 MVDiff, mfxU32 AFD, mfxI32 diffTSC, mfxI32 diffRsCsdiff, mfxU32 posBalance, mfxU32 SC, mfxU32 mu_mv_mag_sq, mfxU32 negBalance, mfxI16 mcTcor, mfxU32 gchDC, mfxU32 Cs, mfxU32 ssDCval) {
        if (MVDiff < 6) {
            if (AFD < 7) {
                if (mu_mv_mag_sq < 117) {
                    if (posBalance < 4) {
                        if (Cs < 1772) {
                            if (Cs < 4) {
                                return 4;
                            }
                            else if (Cs >= 4) {
                                if (mu_mv_mag_sq < 116) {
                                    return 8;
                                }
                                else if (mu_mv_mag_sq >= 116) {
                                    return 1;
                                }
                            }
                        }
                        else if (Cs >= 1772) {
                            if (SC < 9327) {
                                if (SC < 4597) {
                                    return 1;
                                }
                                else if (SC >= 4597) {
                                    if (SC < 7709) {
                                        return 8;
                                    }
                                    else if (SC >= 7709) {
                                        if (SC < 7755) {
                                            return 1;
                                        }
                                        else if (SC >= 7755) {
                                            return 8;
                                        }
                                    }
                                }
                            }
                            else if (SC >= 9327) {
                                if (Cs < 5084) {
                                    return 1;
                                }
                                else if (Cs >= 5084) {
                                    return 8;
                                }
                            }
                        }
                    }
                    else if (posBalance >= 4) {
                        if (mu_mv_mag_sq < 63) {
                            if (ssDCval < 61) {
                                if (Cs < 668) {
                                    if (tcor < 93) {
                                        if (negBalance < 36) {
                                            if (tcor < 81) {
                                                return 4;
                                            }
                                            else if (tcor >= 81) {
                                                return 8;
                                            }
                                        }
                                        else if (negBalance >= 36) {
                                            if (Rs < 856) {
                                                return 4;
                                            }
                                            else if (Rs >= 856) {
                                                if (Cs < 546) {
                                                    return 8;
                                                }
                                                else if (Cs >= 546) {
                                                    return 4;
                                                }
                                            }
                                        }
                                    }
                                    else if (tcor >= 93) {
                                        if (posBalance < 50) {
                                            if (Cs < 479) {
                                                return 8;
                                            }
                                            else if (Cs >= 479) {
                                                return 4;
                                            }
                                        }
                                        else if (posBalance >= 50) {
                                            return 8;
                                        }
                                    }
                                }
                                else if (Cs >= 668) {
                                    if (Rs < 1172) {
                                        if (posBalance < 51) {
                                            if (posBalance < 47) {
                                                return 4;
                                            }
                                            else if (posBalance >= 47) {
                                                return 1;
                                            }
                                        }
                                        else if (posBalance >= 51) {
                                            return 4;
                                        }
                                    }
                                    else if (Rs >= 1172) {
                                        if (Cs < 1011) {
                                            if (diffRsCsdiff < -9) {
                                                return 8;
                                            }
                                            else if (diffRsCsdiff >= -9) {
                                                return 1;
                                            }
                                        }
                                        else if (Cs >= 1011) {
                                            if (refDCval < 53) {
                                                return 4;
                                            }
                                            else if (refDCval >= 53) {
                                                return 1;
                                            }
                                        }
                                    }
                                }
                            }
                            else if (ssDCval >= 61) {
                                if (ssDCval < 102) {
                                    if (tcor < 60) {
                                        return 4;
                                    }
                                    else if (tcor >= 60) {
                                        if (mu_mv_mag_sq < 57) {
                                            return 2;
                                        }
                                        else if (mu_mv_mag_sq >= 57) {
                                            return 8;
                                        }
                                    }
                                }
                                else if (ssDCval >= 102) {
                                    if (RsDiff < 5) {
                                        if (Rs < 852) {
                                            if (Rs < 427) {
                                                return 8;
                                            }
                                            else if (Rs >= 427) {
                                                return 1;
                                            }
                                        }
                                        else if (Rs >= 852) {
                                            return 8;
                                        }
                                    }
                                    else if (RsDiff >= 5) {
                                        if (SC < 4892) {
                                            if (Rs < 2697) {
                                                return 8;
                                            }
                                            else if (Rs >= 2697) {
                                                return 4;
                                            }
                                        }
                                        else if (SC >= 4892) {
                                            return 8;
                                        }
                                    }
                                }
                            }
                        }
                        else if (mu_mv_mag_sq >= 63) {
                            if (Rs < 576) {
                                if (mcTcor < 96) {
                                    if (TSC < 2) {
                                        return 4;
                                    }
                                    else if (TSC >= 2) {
                                        return 8;
                                    }
                                }
                                else if (mcTcor >= 96) {
                                    if (refDCval < 146) {
                                        if (refDCval < 83) {
                                            if (Rs < 144) {
                                                if (SC < 433) {
                                                    return 1;
                                                }
                                                else if (SC >= 433) {
                                                    return 4;
                                                }
                                            }
                                            else if (Rs >= 144) {
                                                return 4;
                                            }
                                        }
                                        else if (refDCval >= 83) {
                                            if (RsDiff < 1) {
                                                if (Rs < 285) {
                                                    return 8;
                                                }
                                                else if (Rs >= 285) {
                                                    if (tcor < 98) {
                                                        if (Cs < 867) {
                                                            return 2;
                                                        }
                                                        else if (Cs >= 867) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (tcor >= 98) {
                                                        if (Rs < 293) {
                                                            return 4;
                                                        }
                                                        else if (Rs >= 293) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (RsDiff >= 1) {
                                                if (Rs < 399) {
                                                    return 8;
                                                }
                                                else if (Rs >= 399) {
                                                    return 4;
                                                }
                                            }
                                        }
                                    }
                                    else if (refDCval >= 146) {
                                        if (Cs < 438) {
                                            return 4;
                                        }
                                        else if (Cs >= 438) {
                                            if (SC < 995) {
                                                if (refDCval < 163) {
                                                    return 4;
                                                }
                                                else if (refDCval >= 163) {
                                                    return 2;
                                                }
                                            }
                                            else if (SC >= 995) {
                                                return 4;
                                            }
                                        }
                                    }
                                }
                            }
                            else if (Rs >= 576) {
                                if (MVDiff < 1) {
                                    if (posBalance < 52) {
                                        if (negBalance < 35) {
                                            if (CsDiff < 4) {
                                                if (tcor < 92) {
                                                    if (posBalance < 9) {
                                                        return 4;
                                                    }
                                                    else if (posBalance >= 9) {
                                                        return 8;
                                                    }
                                                }
                                                else if (tcor >= 92) {
                                                    if (negBalance < 33) {
                                                        return 8;
                                                    }
                                                    else if (negBalance >= 33) {
                                                        if (refDCval < 105) {
                                                            return 8;
                                                        }
                                                        else if (refDCval >= 105) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (CsDiff >= 4) {
                                                if (Cs < 3073) {
                                                    if (ssDCval < 64) {
                                                        return 8;
                                                    }
                                                    else if (ssDCval >= 64) {
                                                        if (refDCval < 131) {
                                                            return 1;
                                                        }
                                                        else if (refDCval >= 131) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                                else if (Cs >= 3073) {
                                                    return 8;
                                                }
                                            }
                                        }
                                        else if (negBalance >= 35) {
                                            return 4;
                                        }
                                    }
                                    else if (posBalance >= 52) {
                                        if (Rs < 1763) {
                                            if (refDCval < 72) {
                                                if (AFD < 6) {
                                                    return 8;
                                                }
                                                else if (AFD >= 6) {
                                                    return 2;
                                                }
                                            }
                                            else if (refDCval >= 72) {
                                                if (RsCsDiff < 22) {
                                                    if (Cs < 945) {
                                                        return 8;
                                                    }
                                                    else if (Cs >= 945) {
                                                        if (posBalance < 54) {
                                                            return 4;
                                                        }
                                                        else if (posBalance >= 54) {
                                                            if (Cs < 1634) {
                                                                return 4;
                                                            }
                                                            else if (Cs >= 1634) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (RsCsDiff >= 22) {
                                                    return 8;
                                                }
                                            }
                                        }
                                        else if (Rs >= 1763) {
                                            if (Rs < 3383) {
                                                if (posBalance < 58) {
                                                    if (SC < 3808) {
                                                        if (tcor < 93) {
                                                            return 8;
                                                        }
                                                        else if (tcor >= 93) {
                                                            if (TSC < 3) {
                                                                return 8;
                                                            }
                                                            else if (TSC >= 3) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                    else if (SC >= 3808) {
                                                        return 4;
                                                    }
                                                }
                                                else if (posBalance >= 58) {
                                                    if (RsDiff < 5) {
                                                        if (Cs < 1602) {
                                                            if (SC < 3158) {
                                                                return 8;
                                                            }
                                                            else if (SC >= 3158) {
                                                                if (ssDCval < 75) {
                                                                    return 2;
                                                                }
                                                                else if (ssDCval >= 75) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                        else if (Cs >= 1602) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (RsDiff >= 5) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                            else if (Rs >= 3383) {
                                                return 8;
                                            }
                                        }
                                    }
                                }
                                else if (MVDiff >= 1) {
                                    if (refDCval < 54) {
                                        if (mu_mv_mag_sq < 99) {
                                            if (negBalance < 35) {
                                                if (Rs < 1115) {
                                                    if (diffAFD < 1) {
                                                        if (Rs < 653) {
                                                            return 8;
                                                        }
                                                        else if (Rs >= 653) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (diffAFD >= 1) {
                                                        if (AFD < 3) {
                                                            return 8;
                                                        }
                                                        else if (AFD >= 3) {
                                                            return 1;
                                                        }
                                                    }
                                                }
                                                else if (Rs >= 1115) {
                                                    return 1;
                                                }
                                            }
                                            else if (negBalance >= 35) {
                                                if (CsDiff < 2) {
                                                    if (Rs < 1164) {
                                                        return 8;
                                                    }
                                                    else if (Rs >= 1164) {
                                                        return 1;
                                                    }
                                                }
                                                else if (CsDiff >= 2) {
                                                    if (mcTcor < 96) {
                                                        return 4;
                                                    }
                                                    else if (mcTcor >= 96) {
                                                        if (Rs < 3510) {
                                                            return 8;
                                                        }
                                                        else if (Rs >= 3510) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (mu_mv_mag_sq >= 99) {
                                            if (diffMVdiffVal < 1) {
                                                if (Cs < 1251) {
                                                    if (Rs < 1110) {
                                                        if (ssDCval < 27) {
                                                            return 8;
                                                        }
                                                        else if (ssDCval >= 27) {
                                                            if (Rs < 998) {
                                                                if (posBalance < 56) {
                                                                    return 8;
                                                                }
                                                                else if (posBalance >= 56) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (Rs >= 998) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                    else if (Rs >= 1110) {
                                                        return 8;
                                                    }
                                                }
                                                else if (Cs >= 1251) {
                                                    if (ssDCval < 33) {
                                                        return 2;
                                                    }
                                                    else if (ssDCval >= 33) {
                                                        if (diffTSC < -1) {
                                                            if (Cs < 1793) {
                                                                return 8;
                                                            }
                                                            else if (Cs >= 1793) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (diffTSC >= -1) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (diffMVdiffVal >= 1) {
                                                return 8;
                                            }
                                        }
                                    }
                                    else if (refDCval >= 54) {
                                        if (negBalance < 16) {
                                            if (ssDCval < 86) {
                                                return 1;
                                            }
                                            else if (ssDCval >= 86) {
                                                if (negBalance < 8) {
                                                    if (RsDiff < 6) {
                                                        if (MVDiff < 2) {
                                                            if (AFD < 2) {
                                                                return 8;
                                                            }
                                                            else if (AFD >= 2) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (MVDiff >= 2) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (RsDiff >= 6) {
                                                        return 1;
                                                    }
                                                }
                                                else if (negBalance >= 8) {
                                                    if (diffMVdiffVal < -4) {
                                                        return 1;
                                                    }
                                                    else if (diffMVdiffVal >= -4) {
                                                        if (diffMVdiffVal < -2) {
                                                            if (ssDCval < 151) {
                                                                return 8;
                                                            }
                                                            else if (ssDCval >= 151) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (diffMVdiffVal >= -2) {
                                                            if (Rs < 1746) {
                                                                return 8;
                                                            }
                                                            else if (Rs >= 1746) {
                                                                if (Rs < 2358) {
                                                                    return 4;
                                                                }
                                                                else if (Rs >= 2358) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (negBalance >= 16) {
                                            if (mu_mv_mag_sq < 98) {
                                                if (Rs < 1063) {
                                                    return 8;
                                                }
                                                else if (Rs >= 1063) {
                                                    if (MVDiff < 2) {
                                                        if (negBalance < 32) {
                                                            return 8;
                                                        }
                                                        else if (negBalance >= 32) {
                                                            if (Cs < 1954) {
                                                                return 1;
                                                            }
                                                            else if (Cs >= 1954) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                    else if (MVDiff >= 2) {
                                                        if (Cs < 1611) {
                                                            if (SC < 2674) {
                                                                if (Rs < 1085) {
                                                                    return 4;
                                                                }
                                                                else if (Rs >= 1085) {
                                                                    if (mu_mv_mag_sq < 73) {
                                                                        if (AFD < 6) {
                                                                            return 1;
                                                                        }
                                                                        else if (AFD >= 6) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                    else if (mu_mv_mag_sq >= 73) {
                                                                        return 8;
                                                                    }
                                                                }
                                                            }
                                                            else if (SC >= 2674) {
                                                                if (ssDCval < 112) {
                                                                    if (SC < 3978) {
                                                                        return 4;
                                                                    }
                                                                    else if (SC >= 3978) {
                                                                        return 1;
                                                                    }
                                                                }
                                                                else if (ssDCval >= 112) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                        else if (Cs >= 1611) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (mu_mv_mag_sq >= 98) {
                                                if (negBalance < 46) {
                                                    if (SC < 3068) {
                                                        return 8;
                                                    }
                                                    else if (SC >= 3068) {
                                                        if (negBalance < 35) {
                                                            if (diffAFD < 2) {
                                                                return 8;
                                                            }
                                                            else if (diffAFD >= 2) {
                                                                if (Rs < 2369) {
                                                                    return 4;
                                                                }
                                                                else if (Rs >= 2369) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                        else if (negBalance >= 35) {
                                                            if (ssDCval < 80) {
                                                                if (mcTcor < 95) {
                                                                    return 8;
                                                                }
                                                                else if (mcTcor >= 95) {
                                                                    if (ssDCval < 73) {
                                                                        if (AFD < 6) {
                                                                            return 8;
                                                                        }
                                                                        else if (AFD >= 6) {
                                                                            if (Cs < 1828) {
                                                                                return 4;
                                                                            }
                                                                            else if (Cs >= 1828) {
                                                                                return 8;
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (ssDCval >= 73) {
                                                                        if (gchDC < 1) {
                                                                            return 4;
                                                                        }
                                                                        else if (gchDC >= 1) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                            else if (ssDCval >= 80) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (negBalance >= 46) {
                                                    if (Rs < 1427) {
                                                        if (ssDCval < 84) {
                                                            return 2;
                                                        }
                                                        else if (ssDCval >= 84) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (Rs >= 1427) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                        }
                    }
                }
                else if (mu_mv_mag_sq >= 117) {
                    if (posBalance < 46) {
                        if (Scindex < 4) {
                            if (ssDCval < 78) {
                                if (Cs < 633) {
                                    if (SC < 502) {
                                        if (SC < 498) {
                                            if (Rs < 181) {
                                                if (mu_mv_mag_sq < 127) {
                                                    return 4;
                                                }
                                                else if (mu_mv_mag_sq >= 127) {
                                                    if (Rs < 180) {
                                                        return 8;
                                                    }
                                                    else if (Rs >= 180) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                            else if (Rs >= 181) {
                                                return 8;
                                            }
                                        }
                                        else if (SC >= 498) {
                                            return 4;
                                        }
                                    }
                                    else if (SC >= 502) {
                                        return 8;
                                    }
                                }
                                else if (Cs >= 633) {
                                    if (Cs < 705) {
                                        if (RsDiff < 1) {
                                            return 4;
                                        }
                                        else if (RsDiff >= 1) {
                                            return 8;
                                        }
                                    }
                                    else if (Cs >= 705) {
                                        return 8;
                                    }
                                }
                            }
                            else if (ssDCval >= 78) {
                                if (mu_mv_mag_sq < 123) {
                                    if (TSC < 2) {
                                        return 8;
                                    }
                                    else if (TSC >= 2) {
                                        return 4;
                                    }
                                }
                                else if (mu_mv_mag_sq >= 123) {
                                    return 4;
                                }
                            }
                        }
                        else if (Scindex >= 4) {
                            if (refDCval < 103) {
                                if (Cs < 708) {
                                    if (Cs < 705) {
                                        return 8;
                                    }
                                    else if (Cs >= 705) {
                                        return 4;
                                    }
                                }
                                else if (Cs >= 708) {
                                    if (TSCindex < 7) {
                                        if (TSC < 3) {
                                            return 8;
                                        }
                                        else if (TSC >= 3) {
                                            if (diffAFD < -2) {
                                                if (Cs < 1546) {
                                                    return 1;
                                                }
                                                else if (Cs >= 1546) {
                                                    return 8;
                                                }
                                            }
                                            else if (diffAFD >= -2) {
                                                if (SC < 4175) {
                                                    if (Cs < 1778) {
                                                        return 8;
                                                    }
                                                    else if (Cs >= 1778) {
                                                        if (negBalance < 23) {
                                                            return 4;
                                                        }
                                                        else if (negBalance >= 23) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                                else if (SC >= 4175) {
                                                    return 8;
                                                }
                                            }
                                        }
                                    }
                                    else if (TSCindex >= 7) {
                                        if (Rs < 3481) {
                                            return 1;
                                        }
                                        else if (Rs >= 3481) {
                                            return 8;
                                        }
                                    }
                                }
                            }
                            else if (refDCval >= 103) {
                                if (negBalance < 3) {
                                    if (Rs < 3779) {
                                        if (Cs < 4001) {
                                            if (refDCval < 140) {
                                                if (posBalance < 22) {
                                                    if (diffAFD < -2) {
                                                        if (Rs < 2844) {
                                                            return 1;
                                                        }
                                                        else if (Rs >= 2844) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (diffAFD >= -2) {
                                                        if (diffRsCsdiff < 7) {
                                                            if (SC < 7302) {
                                                                return 8;
                                                            }
                                                            else if (SC >= 7302) {
                                                                if (posBalance < 21) {
                                                                    if (Rs < 3714) {
                                                                        return 1;
                                                                    }
                                                                    else if (Rs >= 3714) {
                                                                        return 8;
                                                                    }
                                                                }
                                                                else if (posBalance >= 21) {
                                                                    if (Rs < 3638) {
                                                                        return 1;
                                                                    }
                                                                    else if (Rs >= 3638) {
                                                                        if (Cs < 3976) {
                                                                            return 8;
                                                                        }
                                                                        else if (Cs >= 3976) {
                                                                            if (Rs < 3774) {
                                                                                return 8;
                                                                            }
                                                                            else if (Rs >= 3774) {
                                                                                return 1;
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (diffRsCsdiff >= 7) {
                                                            return 1;
                                                        }
                                                    }
                                                }
                                                else if (posBalance >= 22) {
                                                    if (tcor < 99) {
                                                        return 1;
                                                    }
                                                    else if (tcor >= 99) {
                                                        if (diffRsCsdiff < -4) {
                                                            if (SC < 5509) {
                                                                return 8;
                                                            }
                                                            else if (SC >= 5509) {
                                                                return 1;
                                                            }
                                                        }
                                                        else if (diffRsCsdiff >= -4) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (refDCval >= 140) {
                                                if (Rs < 1155) {
                                                    return 8;
                                                }
                                                else if (Rs >= 1155) {
                                                    return 1;
                                                }
                                            }
                                        }
                                        else if (Cs >= 4001) {
                                            return 1;
                                        }
                                    }
                                    else if (Rs >= 3779) {
                                        return 8;
                                    }
                                }
                                else if (negBalance >= 3) {
                                    if (mcTcor < 97) {
                                        if (mu_mv_mag_sq < 121) {
                                            if (Rs < 1953) {
                                                return 8;
                                            }
                                            else if (Rs >= 1953) {
                                                if (ssDCval < 116) {
                                                    return 8;
                                                }
                                                else if (ssDCval >= 116) {
                                                    return 4;
                                                }
                                            }
                                        }
                                        else if (mu_mv_mag_sq >= 121) {
                                            if (negBalance < 4) {
                                                if (Rs < 1964) {
                                                    return 1;
                                                }
                                                else if (Rs >= 1964) {
                                                    return 4;
                                                }
                                            }
                                            else if (negBalance >= 4) {
                                                return 8;
                                            }
                                        }
                                    }
                                    else if (mcTcor >= 97) {
                                        if (CsDiff < 3) {
                                            if (Cs < 729) {
                                                if (AFD < 1) {
                                                    return 8;
                                                }
                                                else if (AFD >= 1) {
                                                    return 1;
                                                }
                                            }
                                            else if (Cs >= 729) {
                                                if (diffMVdiffVal < -32) {
                                                    return 4;
                                                }
                                                else if (diffMVdiffVal >= -32) {
                                                    if (tcor < 92) {
                                                        return 4;
                                                    }
                                                    else if (tcor >= 92) {
                                                        if (refDCval < 105) {
                                                            if (AFD < 3) {
                                                                return 8;
                                                            }
                                                            else if (AFD >= 3) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (refDCval >= 105) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (CsDiff >= 3) {
                                            if (negBalance < 25) {
                                                return 8;
                                            }
                                            else if (negBalance >= 25) {
                                                if (diffMVdiffVal < -3) {
                                                    if (SC < 5401) {
                                                        return 4;
                                                    }
                                                    else if (SC >= 5401) {
                                                        return 8;
                                                    }
                                                }
                                                else if (diffMVdiffVal >= -3) {
                                                    if (posBalance < 42) {
                                                        if (RsCsDiff < 22) {
                                                            return 8;
                                                        }
                                                        else if (RsCsDiff >= 22) {
                                                            if (SC < 6344) {
                                                                return 8;
                                                            }
                                                            else if (SC >= 6344) {
                                                                return 1;
                                                            }
                                                        }
                                                    }
                                                    else if (posBalance >= 42) {
                                                        if (Rs < 3145) {
                                                            return 8;
                                                        }
                                                        else if (Rs >= 3145) {
                                                            return 1;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                        }
                    }
                    else if (posBalance >= 46) {
                        if (refDCval < 83) {
                            if (Rs < 1769) {
                                if (Rs < 256) {
                                    return 4;
                                }
                                else if (Rs >= 256) {
                                    if (mu_mv_mag_sq < 119) {
                                        if (TSC < 6) {
                                            return 8;
                                        }
                                        else if (TSC >= 6) {
                                            if (RsDiff < 4) {
                                                return 2;
                                            }
                                            else if (RsDiff >= 4) {
                                                return 8;
                                            }
                                        }
                                    }
                                    else if (mu_mv_mag_sq >= 119) {
                                        if (Cs < 1720) {
                                            return 8;
                                        }
                                        else if (Cs >= 1720) {
                                            if (Cs < 1729) {
                                                return 4;
                                            }
                                            else if (Cs >= 1729) {
                                                return 8;
                                            }
                                        }
                                    }
                                }
                            }
                            else if (Rs >= 1769) {
                                if (Cs < 4874) {
                                    if (Rs < 2422) {
                                        if (SC < 3583) {
                                            if (posBalance < 58) {
                                                if (Cs < 1579) {
                                                    if (Cs < 1576) {
                                                        if (Rs < 1824) {
                                                            return 8;
                                                        }
                                                        else if (Rs >= 1824) {
                                                            if (Cs < 1247) {
                                                                if (Rs < 1948) {
                                                                    return 4;
                                                                }
                                                                else if (Rs >= 1948) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (Cs >= 1247) {
                                                                if (Rs < 1902) {
                                                                    return 8;
                                                                }
                                                                else if (Rs >= 1902) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (Cs >= 1576) {
                                                        return 4;
                                                    }
                                                }
                                                else if (Cs >= 1579) {
                                                    if (mu_mv_mag_sq < 125) {
                                                        return 8;
                                                    }
                                                    else if (mu_mv_mag_sq >= 125) {
                                                        if (Rs < 1849) {
                                                            return 8;
                                                        }
                                                        else if (Rs >= 1849) {
                                                            return 2;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (posBalance >= 58) {
                                                if (refDCval < 75) {
                                                    if (mcTcor < 95) {
                                                        if (SC < 3406) {
                                                            if (SC < 3393) {
                                                                return 8;
                                                            }
                                                            else if (SC >= 3393) {
                                                                return 2;
                                                            }
                                                        }
                                                        else if (SC >= 3406) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (mcTcor >= 95) {
                                                        if (diffMVdiffVal < -1) {
                                                            return 8;
                                                        }
                                                        else if (diffMVdiffVal >= -1) {
                                                            return 2;
                                                        }
                                                    }
                                                }
                                                else if (refDCval >= 75) {
                                                    if (Rs < 1840) {
                                                        return 8;
                                                    }
                                                    else if (Rs >= 1840) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                        }
                                        else if (SC >= 3583) {
                                            if (tcor < 95) {
                                                if (SC < 3701) {
                                                    if (ssDCval < 76) {
                                                        return 2;
                                                    }
                                                    else if (ssDCval >= 76) {
                                                        return 4;
                                                    }
                                                }
                                                else if (SC >= 3701) {
                                                    return 8;
                                                }
                                            }
                                            else if (tcor >= 95) {
                                                return 4;
                                            }
                                        }
                                    }
                                    else if (Rs >= 2422) {
                                        if (SC < 9291) {
                                            return 8;
                                        }
                                        else if (SC >= 9291) {
                                            if (Rs < 4632) {
                                                return 4;
                                            }
                                            else if (Rs >= 4632) {
                                                return 8;
                                            }
                                        }
                                    }
                                }
                                else if (Cs >= 4874) {
                                    if (ssDCval < 80) {
                                        if (SC < 9647) {
                                            return 4;
                                        }
                                        else if (SC >= 9647) {
                                            return 8;
                                        }
                                    }
                                    else if (ssDCval >= 80) {
                                        if (Cs < 4989) {
                                            return 4;
                                        }
                                        else if (Cs >= 4989) {
                                            if (Cs < 5007) {
                                                return 8;
                                            }
                                            else if (Cs >= 5007) {
                                                if (mcTcor < 92) {
                                                    if (RsCsDiff < 137) {
                                                        if (Cs < 5124) {
                                                            return 4;
                                                        }
                                                        else if (Cs >= 5124) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (RsCsDiff >= 137) {
                                                        if (Rs < 4933) {
                                                            return 8;
                                                        }
                                                        else if (Rs >= 4933) {
                                                            if (Cs < 5172) {
                                                                if (SC < 10070) {
                                                                    return 4;
                                                                }
                                                                else if (SC >= 10070) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (Cs >= 5172) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (mcTcor >= 92) {
                                                    if (diffAFD < -2) {
                                                        return 8;
                                                    }
                                                    else if (diffAFD >= -2) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                        }
                        else if (refDCval >= 83) {
                            if (SC < 6135) {
                                if (mu_mv_mag_sq < 123) {
                                    if (ssDCval < 114) {
                                        if (posBalance < 58) {
                                            if (Cs < 3031) {
                                                if (RsCsDiff < 31) {
                                                    if (CsDiff < 3) {
                                                        if (Cs < 771) {
                                                            return 4;
                                                        }
                                                        else if (Cs >= 771) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (CsDiff >= 3) {
                                                        if (Rs < 2708) {
                                                            return 4;
                                                        }
                                                        else if (Rs >= 2708) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                                else if (RsCsDiff >= 31) {
                                                    if (ssDCval < 112) {
                                                        return 8;
                                                    }
                                                    else if (ssDCval >= 112) {
                                                        if (SC < 5639) {
                                                            return 8;
                                                        }
                                                        else if (SC >= 5639) {
                                                            if (ssDCval < 113) {
                                                                if (SC < 5826) {
                                                                    return 4;
                                                                }
                                                                else if (SC >= 5826) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (ssDCval >= 113) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                            else if (Cs >= 3031) {
                                                return 4;
                                            }
                                        }
                                        else if (posBalance >= 58) {
                                            if (SC < 5612) {
                                                if (RsCsDiff < 57) {
                                                    return 8;
                                                }
                                                else if (RsCsDiff >= 57) {
                                                    return 4;
                                                }
                                            }
                                            else if (SC >= 5612) {
                                                if (ssDCval < 113) {
                                                    if (RsDiff < 5) {
                                                        return 8;
                                                    }
                                                    else if (RsDiff >= 5) {
                                                        if (ssDCval < 111) {
                                                            return 8;
                                                        }
                                                        else if (ssDCval >= 111) {
                                                            if (Rs < 2947) {
                                                                return 8;
                                                            }
                                                            else if (Rs >= 2947) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (ssDCval >= 113) {
                                                    if (Cs < 3014) {
                                                        if (Rs < 2710) {
                                                            return 8;
                                                        }
                                                        else if (Rs >= 2710) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (Cs >= 3014) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (ssDCval >= 114) {
                                        if (posBalance < 56) {
                                            if (AFD < 5) {
                                                return 8;
                                            }
                                            else if (AFD >= 5) {
                                                return 4;
                                            }
                                        }
                                        else if (posBalance >= 56) {
                                            return 8;
                                        }
                                    }
                                }
                                else if (mu_mv_mag_sq >= 123) {
                                    if (ssDCval < 104) {
                                        return 8;
                                    }
                                    else if (ssDCval >= 104) {
                                        if (Rs < 2029) {
                                            return 4;
                                        }
                                        else if (Rs >= 2029) {
                                            if (SC < 5750) {
                                                return 8;
                                            }
                                            else if (SC >= 5750) {
                                                if (negBalance < 48) {
                                                    return 8;
                                                }
                                                else if (negBalance >= 48) {
                                                    if (SC < 5765) {
                                                        return 4;
                                                    }
                                                    else if (SC >= 5765) {
                                                        if (gchDC < 1) {
                                                            return 8;
                                                        }
                                                        else if (gchDC >= 1) {
                                                            if (RsDiff < 5) {
                                                                return 4;
                                                            }
                                                            else if (RsDiff >= 5) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                            else if (SC >= 6135) {
                                return 8;
                            }
                        }
                    }
                }
            }
            else if (AFD >= 7) {
                if (Scindex < 5) {
                    if (RsCsDiff < 33) {
                        if (TSC < 6) {
                            if (refDCval < 64) {
                                if (mu_mv_mag_sq < 35) {
                                    if (RsDiff < 4) {
                                        return 8;
                                    }
                                    else if (RsDiff >= 4) {
                                        return 1;
                                    }
                                }
                                else if (mu_mv_mag_sq >= 35) {
                                    if (Rs < 1181) {
                                        return 4;
                                    }
                                    else if (Rs >= 1181) {
                                        return 1;
                                    }
                                }
                            }
                            else if (refDCval >= 64) {
                                if (MVDiff < 1) {
                                    if (SC < 2501) {
                                        return 2;
                                    }
                                    else if (SC >= 2501) {
                                        return 4;
                                    }
                                }
                                else if (MVDiff >= 1) {
                                    if (RsDiff < 3) {
                                        if (tcor < 93) {
                                            return 4;
                                        }
                                        else if (tcor >= 93) {
                                            if (Cs < 516) {
                                                return 2;
                                            }
                                            else if (Cs >= 516) {
                                                return 8;
                                            }
                                        }
                                    }
                                    else if (RsDiff >= 3) {
                                        if (negBalance < 23) {
                                            return 8;
                                        }
                                        else if (negBalance >= 23) {
                                            if (diffAFD < -2) {
                                                if (Rs < 1468) {
                                                    return 1;
                                                }
                                                else if (Rs >= 1468) {
                                                    return 8;
                                                }
                                            }
                                            else if (diffAFD >= -2) {
                                                return 2;
                                            }
                                        }
                                    }
                                }
                            }
                        }
                        else if (TSC >= 6) {
                            if (tcor < 79) {
                                if (Cs < 1340) {
                                    return 4;
                                }
                                else if (Cs >= 1340) {
                                    return 8;
                                }
                            }
                            else if (tcor >= 79) {
                                if (mcTcor < 97) {
                                    if (posBalance < 56) {
                                        if (TSC < 7) {
                                            return 4;
                                        }
                                        else if (TSC >= 7) {
                                            return 8;
                                        }
                                    }
                                    else if (posBalance >= 56) {
                                        if (Rs < 1758) {
                                            if (tcor < 83) {
                                                if (SC < 2037) {
                                                    return 4;
                                                }
                                                else if (SC >= 2037) {
                                                    return 2;
                                                }
                                            }
                                            else if (tcor >= 83) {
                                                return 2;
                                            }
                                        }
                                        else if (Rs >= 1758) {
                                            if (AFD < 8) {
                                                if (tcor < 91) {
                                                    return 8;
                                                }
                                                else if (tcor >= 91) {
                                                    if (diffAFD < -1) {
                                                        if (Cs < 1423) {
                                                            return 1;
                                                        }
                                                        else if (Cs >= 1423) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (diffAFD >= -1) {
                                                        if (diffAFD < 1) {
                                                            if (Cs < 1402) {
                                                                return 4;
                                                            }
                                                            else if (Cs >= 1402) {
                                                                return 2;
                                                            }
                                                        }
                                                        else if (diffAFD >= 1) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (AFD >= 8) {
                                                return 2;
                                            }
                                        }
                                    }
                                }
                                else if (mcTcor >= 97) {
                                    return 4;
                                }
                            }
                        }
                    }
                    else if (RsCsDiff >= 33) {
                        if (tcor < 84) {
                            if (mu_mv_mag_sq < 80) {
                                if (mu_mv_mag_sq < 5) {
                                    if (diffTSC < -1) {
                                        if (TSC < 6) {
                                            if (AFD < 12) {
                                                return 4;
                                            }
                                            else if (AFD >= 12) {
                                                return 1;
                                            }
                                        }
                                        else if (TSC >= 6) {
                                            return 1;
                                        }
                                    }
                                    else if (diffTSC >= -1) {
                                        if (ssDCval < 43) {
                                            return 1;
                                        }
                                        else if (ssDCval >= 43) {
                                            if (CsDiff < 10) {
                                                return 4;
                                            }
                                            else if (CsDiff >= 10) {
                                                return 2;
                                            }
                                        }
                                    }
                                }
                                else if (mu_mv_mag_sq >= 5) {
                                    if (ssDCval < 99) {
                                        if (mu_mv_mag_sq < 68) {
                                            if (diffMVdiffVal < -3) {
                                                if (Rs < 1906) {
                                                    return 8;
                                                }
                                                else if (Rs >= 1906) {
                                                    return 2;
                                                }
                                            }
                                            else if (diffMVdiffVal >= -3) {
                                                return 2;
                                            }
                                        }
                                        else if (mu_mv_mag_sq >= 68) {
                                            if (TSC < 7) {
                                                return 4;
                                            }
                                            else if (TSC >= 7) {
                                                return 8;
                                            }
                                        }
                                    }
                                    else if (ssDCval >= 99) {
                                        if (refDCval < 130) {
                                            return 4;
                                        }
                                        else if (refDCval >= 130) {
                                            if (Cs < 1232) {
                                                return 8;
                                            }
                                            else if (Cs >= 1232) {
                                                return 1;
                                            }
                                        }
                                    }
                                }
                            }
                            else if (mu_mv_mag_sq >= 80) {
                                return 8;
                            }
                        }
                        else if (tcor >= 84) {
                            if (mu_mv_mag_sq < 102) {
                                if (ssDCval < 80) {
                                    if (negBalance < 46) {
                                        if (Cs < 2162) {
                                            return 4;
                                        }
                                        else if (Cs >= 2162) {
                                            return 1;
                                        }
                                    }
                                    else if (negBalance >= 46) {
                                        if (Rs < 2044) {
                                            return 2;
                                        }
                                        else if (Rs >= 2044) {
                                            if (SC < 3964) {
                                                return 4;
                                            }
                                            else if (SC >= 3964) {
                                                return 1;
                                            }
                                        }
                                    }
                                }
                                else if (ssDCval >= 80) {
                                    if (tcor < 95) {
                                        if (AFD < 10) {
                                            if (MVDiff < 3) {
                                                if (CsDiff < 6) {
                                                    if (Cs < 1488) {
                                                        return 8;
                                                    }
                                                    else if (Cs >= 1488) {
                                                        return 2;
                                                    }
                                                }
                                                else if (CsDiff >= 6) {
                                                    if (Cs < 1759) {
                                                        return 4;
                                                    }
                                                    else if (Cs >= 1759) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                            else if (MVDiff >= 3) {
                                                if (SC < 3278) {
                                                    return 8;
                                                }
                                                else if (SC >= 3278) {
                                                    if (RsCsDiff < 57) {
                                                        if (MVDiff < 4) {
                                                            return 2;
                                                        }
                                                        else if (MVDiff >= 4) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (RsCsDiff >= 57) {
                                                        if (mcTcor < 89) {
                                                            return 2;
                                                        }
                                                        else if (mcTcor >= 89) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (AFD >= 10) {
                                            if (negBalance < 52) {
                                                return 2;
                                            }
                                            else if (negBalance >= 52) {
                                                if (refDCval < 99) {
                                                    return 2;
                                                }
                                                else if (refDCval >= 99) {
                                                    if (Rs < 2680) {
                                                        return 4;
                                                    }
                                                    else if (Rs >= 2680) {
                                                        return 2;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (tcor >= 95) {
                                        if (Cs < 1999) {
                                            if (refDCval < 87) {
                                                return 2;
                                            }
                                            else if (refDCval >= 87) {
                                                if (Cs < 1089) {
                                                    return 2;
                                                }
                                                else if (Cs >= 1089) {
                                                    return 4;
                                                }
                                            }
                                        }
                                        else if (Cs >= 1999) {
                                            return 8;
                                        }
                                    }
                                }
                            }
                            else if (mu_mv_mag_sq >= 102) {
                                if (tcor < 95) {
                                    if (ssDCval < 76) {
                                        if (TSC < 7) {
                                            if (RsCsDiff < 46) {
                                                return 8;
                                            }
                                            else if (RsCsDiff >= 46) {
                                                return 2;
                                            }
                                        }
                                        else if (TSC >= 7) {
                                            return 2;
                                        }
                                    }
                                    else if (ssDCval >= 76) {
                                        if (Rs < 1742) {
                                            if (Cs < 1761) {
                                                return 2;
                                            }
                                            else if (Cs >= 1761) {
                                                return 8;
                                            }
                                        }
                                        else if (Rs >= 1742) {
                                            return 8;
                                        }
                                    }
                                }
                                else if (tcor >= 95) {
                                    return 4;
                                }
                            }
                        }
                    }
                }
                else if (Scindex >= 5) {
                    if (Cs < 4173) {
                        if (SC < 9178) {
                            if (Rs < 5226) {
                                if (mu_mv_mag_sq < 103) {
                                    if (Cs < 2650) {
                                        if (posBalance < 58) {
                                            if (ssDCval < 88) {
                                                if (ssDCval < 73) {
                                                    if (negBalance < 28) {
                                                        return 1;
                                                    }
                                                    else if (negBalance >= 28) {
                                                        if (SC < 4963) {
                                                            return 4;
                                                        }
                                                        else if (SC >= 4963) {
                                                            if (Rs < 2892) {
                                                                return 8;
                                                            }
                                                            else if (Rs >= 2892) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (ssDCval >= 73) {
                                                    if (mcTcor < 91) {
                                                        return 2;
                                                    }
                                                    else if (mcTcor >= 91) {
                                                        if (posBalance < 46) {
                                                            return 8;
                                                        }
                                                        else if (posBalance >= 46) {
                                                            if (MVDiff < 4) {
                                                                return 4;
                                                            }
                                                            else if (MVDiff >= 4) {
                                                                if (SC < 5357) {
                                                                    return 4;
                                                                }
                                                                else if (SC >= 5357) {
                                                                    if (AFD < 11) {
                                                                        if (CsDiff < 7) {
                                                                            return 8;
                                                                        }
                                                                        else if (CsDiff >= 7) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                    else if (AFD >= 11) {
                                                                        return 8;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                            else if (ssDCval >= 88) {
                                                if (ssDCval < 100) {
                                                    if (Cs < 2191) {
                                                        if (Rs < 2143) {
                                                            if (RsCsDiff < 57) {
                                                                return 4;
                                                            }
                                                            else if (RsCsDiff >= 57) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (Rs >= 2143) {
                                                            if (mcTcor < 90) {
                                                                if (ssDCval < 93) {
                                                                    return 8;
                                                                }
                                                                else if (ssDCval >= 93) {
                                                                    return 2;
                                                                }
                                                            }
                                                            else if (mcTcor >= 90) {
                                                                if (diffMVdiffVal < 2) {
                                                                    return 1;
                                                                }
                                                                else if (diffMVdiffVal >= 2) {
                                                                    if (AFD < 8) {
                                                                        return 8;
                                                                    }
                                                                    else if (AFD >= 8) {
                                                                        return 1;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (Cs >= 2191) {
                                                        if (SC < 4733) {
                                                            if (diffRsCsdiff < 25) {
                                                                return 2;
                                                            }
                                                            else if (diffRsCsdiff >= 25) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (SC >= 4733) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                                else if (ssDCval >= 100) {
                                                    if (AFD < 11) {
                                                        if (negBalance < 33) {
                                                            if (RsDiff < 8) {
                                                                return 4;
                                                            }
                                                            else if (RsDiff >= 8) {
                                                                return 2;
                                                            }
                                                        }
                                                        else if (negBalance >= 33) {
                                                            if (negBalance < 52) {
                                                                return 8;
                                                            }
                                                            else if (negBalance >= 52) {
                                                                if (Cs < 1930) {
                                                                    return 8;
                                                                }
                                                                else if (Cs >= 1930) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (AFD >= 11) {
                                                        if (ssDCval < 110) {
                                                            if (negBalance < 53) {
                                                                return 8;
                                                            }
                                                            else if (negBalance >= 53) {
                                                                if (Rs < 2343) {
                                                                    return 2;
                                                                }
                                                                else if (Rs >= 2343) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                        else if (ssDCval >= 110) {
                                                            if (diffTSC < 1) {
                                                                if (Cs < 1768) {
                                                                    return 4;
                                                                }
                                                                else if (Cs >= 1768) {
                                                                    return 1;
                                                                }
                                                            }
                                                            else if (diffTSC >= 1) {
                                                                if (diffRsCsdiff < 44) {
                                                                    if (AFD < 13) {
                                                                        return 2;
                                                                    }
                                                                    else if (AFD >= 13) {
                                                                        return 1;
                                                                    }
                                                                }
                                                                else if (diffRsCsdiff >= 44) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (posBalance >= 58) {
                                            if (tcor < 85) {
                                                if (negBalance < 50) {
                                                    return 8;
                                                }
                                                else if (negBalance >= 50) {
                                                    if (mu_mv_mag_sq < 30) {
                                                        if (TSC < 10) {
                                                            if (diffAFD < 1) {
                                                                return 4;
                                                            }
                                                            else if (diffAFD >= 1) {
                                                                if (negBalance < 52) {
                                                                    if (TSC < 8) {
                                                                        return 4;
                                                                    }
                                                                    else if (TSC >= 8) {
                                                                        return 8;
                                                                    }
                                                                }
                                                                else if (negBalance >= 52) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                        else if (TSC >= 10) {
                                                            return 1;
                                                        }
                                                    }
                                                    else if (mu_mv_mag_sq >= 30) {
                                                        if (SC < 4696) {
                                                            if (MVDiff < 3) {
                                                                return 4;
                                                            }
                                                            else if (MVDiff >= 3) {
                                                                if (Rs < 2551) {
                                                                    if (Cs < 2086) {
                                                                        return 4;
                                                                    }
                                                                    else if (Cs >= 2086) {
                                                                        if (diffTSC < -1) {
                                                                            return 4;
                                                                        }
                                                                        else if (diffTSC >= -1) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                }
                                                                else if (Rs >= 2551) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                        else if (SC >= 4696) {
                                                            if (mu_mv_mag_sq < 82) {
                                                                return 4;
                                                            }
                                                            else if (mu_mv_mag_sq >= 82) {
                                                                if (SC < 4793) {
                                                                    return 4;
                                                                }
                                                                else if (SC >= 4793) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                            else if (tcor >= 85) {
                                                if (refDCval < 86) {
                                                    if (AFD < 12) {
                                                        return 4;
                                                    }
                                                    else if (AFD >= 12) {
                                                        if (SC < 5569) {
                                                            return 4;
                                                        }
                                                        else if (SC >= 5569) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                                else if (refDCval >= 86) {
                                                    if (SC < 5316) {
                                                        if (mu_mv_mag_sq < 95) {
                                                            if (ssDCval < 97) {
                                                                if (Rs < 2951) {
                                                                    if (posBalance < 62) {
                                                                        return 2;
                                                                    }
                                                                    else if (posBalance >= 62) {
                                                                        if (SC < 4559) {
                                                                            if (SC < 4314) {
                                                                                return 2;
                                                                            }
                                                                            else if (SC >= 4314) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                        else if (SC >= 4559) {
                                                                            return 2;
                                                                        }
                                                                    }
                                                                }
                                                                else if (Rs >= 2951) {
                                                                    return 1;
                                                                }
                                                            }
                                                            else if (ssDCval >= 97) {
                                                                if (CsDiff < 5) {
                                                                    return 2;
                                                                }
                                                                else if (CsDiff >= 5) {
                                                                    if (mcTcor < 90) {
                                                                        return 8;
                                                                    }
                                                                    else if (mcTcor >= 90) {
                                                                        if (Rs < 2833) {
                                                                            return 4;
                                                                        }
                                                                        else if (Rs >= 2833) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (mu_mv_mag_sq >= 95) {
                                                            if (RsCsDiff < 62) {
                                                                return 2;
                                                            }
                                                            else if (RsCsDiff >= 62) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                    else if (SC >= 5316) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (Cs >= 2650) {
                                        if (Rs < 3253) {
                                            return 8;
                                        }
                                        else if (Rs >= 3253) {
                                            if (tcor < 72) {
                                                return 1;
                                            }
                                            else if (tcor >= 72) {
                                                if (diffMVdiffVal < -5) {
                                                    if (Cs < 4009) {
                                                        return 2;
                                                    }
                                                    else if (Cs >= 4009) {
                                                        if (Cs < 4117) {
                                                            return 8;
                                                        }
                                                        else if (Cs >= 4117) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                                else if (diffMVdiffVal >= -5) {
                                                    if (Cs < 2893) {
                                                        if (tcor < 74) {
                                                            return 8;
                                                        }
                                                        else if (tcor >= 74) {
                                                            if (refDCval < 108) {
                                                                return 4;
                                                            }
                                                            else if (refDCval >= 108) {
                                                                if (posBalance < 48) {
                                                                    return 4;
                                                                }
                                                                else if (posBalance >= 48) {
                                                                    return 1;
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (Cs >= 2893) {
                                                        if (Rs < 3854) {
                                                            if (mu_mv_mag_sq < 93) {
                                                                if (ssDCval < 54) {
                                                                    if (CsDiff < 3) {
                                                                        return 8;
                                                                    }
                                                                    else if (CsDiff >= 3) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (ssDCval >= 54) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (mu_mv_mag_sq >= 93) {
                                                                return 2;
                                                            }
                                                        }
                                                        else if (Rs >= 3854) {
                                                            if (SC < 7441) {
                                                                return 4;
                                                            }
                                                            else if (SC >= 7441) {
                                                                if (tcor < 78) {
                                                                    return 4;
                                                                }
                                                                else if (tcor >= 78) {
                                                                    if (RsCsDiff < 277) {
                                                                        return 8;
                                                                    }
                                                                    else if (RsCsDiff >= 277) {
                                                                        if (Cs < 3801) {
                                                                            return 4;
                                                                        }
                                                                        else if (Cs >= 3801) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (mu_mv_mag_sq >= 103) {
                                    if (negBalance < 55) {
                                        if (refDCval < 94) {
                                            if (mcTcor < 92) {
                                                if (mu_mv_mag_sq < 104) {
                                                    if (Rs < 2254) {
                                                        return 8;
                                                    }
                                                    else if (Rs >= 2254) {
                                                        return 2;
                                                    }
                                                }
                                                else if (mu_mv_mag_sq >= 104) {
                                                    return 8;
                                                }
                                            }
                                            else if (mcTcor >= 92) {
                                                if (Rs < 3207) {
                                                    if (MVDiff < 1) {
                                                        if (ssDCval < 85) {
                                                            return 4;
                                                        }
                                                        else if (ssDCval >= 85) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (MVDiff >= 1) {
                                                        if (diffTSC < 2) {
                                                            if (diffRsCsdiff < -19) {
                                                                return 2;
                                                            }
                                                            else if (diffRsCsdiff >= -19) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (diffTSC >= 2) {
                                                            return 1;
                                                        }
                                                    }
                                                }
                                                else if (Rs >= 3207) {
                                                    return 2;
                                                }
                                            }
                                        }
                                        else if (refDCval >= 94) {
                                            if (mu_mv_mag_sq < 113) {
                                                if (Rs < 3468) {
                                                    if (Rs < 2433) {
                                                        if (ssDCval < 113) {
                                                            return 4;
                                                        }
                                                        else if (ssDCval >= 113) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (Rs >= 2433) {
                                                        return 8;
                                                    }
                                                }
                                                else if (Rs >= 3468) {
                                                    if (negBalance < 46) {
                                                        if (negBalance < 42) {
                                                            if (ssDCval < 115) {
                                                                return 2;
                                                            }
                                                            else if (ssDCval >= 115) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (negBalance >= 42) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (negBalance >= 46) {
                                                        if (refDCval < 106) {
                                                            if (CsDiff < 12) {
                                                                return 4;
                                                            }
                                                            else if (CsDiff >= 12) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (refDCval >= 106) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (mu_mv_mag_sq >= 113) {
                                                if (diffMVdiffVal < 1) {
                                                    if (SC < 4755) {
                                                        return 4;
                                                    }
                                                    else if (SC >= 4755) {
                                                        if (tcor < 93) {
                                                            if (mu_mv_mag_sq < 121) {
                                                                return 8;
                                                            }
                                                            else if (mu_mv_mag_sq >= 121) {
                                                                if (Rs < 4267) {
                                                                    return 8;
                                                                }
                                                                else if (Rs >= 4267) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                        else if (tcor >= 93) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                                else if (diffMVdiffVal >= 1) {
                                                    if (ssDCval < 119) {
                                                        if (CsDiff < 10) {
                                                            if (Rs < 2761) {
                                                                return 4;
                                                            }
                                                            else if (Rs >= 2761) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (CsDiff >= 10) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (ssDCval >= 119) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (negBalance >= 55) {
                                        if (Rs < 4507) {
                                            return 8;
                                        }
                                        else if (Rs >= 4507) {
                                            return 4;
                                        }
                                    }
                                }
                            }
                            else if (Rs >= 5226) {
                                if (posBalance < 55) {
                                    return 8;
                                }
                                else if (posBalance >= 55) {
                                    return 4;
                                }
                            }
                        }
                        else if (SC >= 9178) {
                            if (posBalance < 57) {
                                if (Cs < 3603) {
                                    return 1;
                                }
                                else if (Cs >= 3603) {
                                    return 8;
                                }
                            }
                            else if (posBalance >= 57) {
                                return 4;
                            }
                        }
                    }
                    else if (Cs >= 4173) {
                        if (Rs < 6893) {
                            if (ssDCval < 106) {
                                if (refDCval < 82) {
                                    if (AFD < 8) {
                                        if (Rs < 4811) {
                                            if (mu_mv_mag_sq < 128) {
                                                if (gchDC < 1) {
                                                    return 4;
                                                }
                                                else if (gchDC >= 1) {
                                                    return 8;
                                                }
                                            }
                                            else if (mu_mv_mag_sq >= 128) {
                                                return 8;
                                            }
                                        }
                                        else if (Rs >= 4811) {
                                            if (diffRsCsdiff < -9) {
                                                if (Cs < 5169) {
                                                    if (gchDC < 1) {
                                                        return 8;
                                                    }
                                                    else if (gchDC >= 1) {
                                                        return 4;
                                                    }
                                                }
                                                else if (Cs >= 5169) {
                                                    return 4;
                                                }
                                            }
                                            else if (diffRsCsdiff >= -9) {
                                                return 4;
                                            }
                                        }
                                    }
                                    else if (AFD >= 8) {
                                        if (RsDiff < 22) {
                                            return 8;
                                        }
                                        else if (RsDiff >= 22) {
                                            return 1;
                                        }
                                    }
                                }
                                else if (refDCval >= 82) {
                                    return 8;
                                }
                            }
                            else if (ssDCval >= 106) {
                                if (ssDCval < 109) {
                                    if (SC < 12117) {
                                        if (Cs < 5289) {
                                            if (SC < 10899) {
                                                return 8;
                                            }
                                            else if (SC >= 10899) {
                                                return 4;
                                            }
                                        }
                                        else if (Cs >= 5289) {
                                            if (refDCval < 106) {
                                                return 8;
                                            }
                                            else if (refDCval >= 106) {
                                                return 4;
                                            }
                                        }
                                    }
                                    else if (SC >= 12117) {
                                        return 8;
                                    }
                                }
                                else if (ssDCval >= 109) {
                                    if (mcTcor < 83) {
                                        if (Rs < 4627) {
                                            return 1;
                                        }
                                        else if (Rs >= 4627) {
                                            if (SC < 11016) {
                                                return 4;
                                            }
                                            else if (SC >= 11016) {
                                                return 8;
                                            }
                                        }
                                    }
                                    else if (mcTcor >= 83) {
                                        return 8;
                                    }
                                }
                            }
                        }
                        else if (Rs >= 6893) {
                            if (negBalance < 63) {
                                if (AFD < 30) {
                                    return 8;
                                }
                                else if (AFD >= 30) {
                                    if (RsCsDiff < 2956) {
                                        return 8;
                                    }
                                    else if (RsCsDiff >= 2956) {
                                        return 4;
                                    }
                                }
                            }
                            else if (negBalance >= 63) {
                                if (AFD < 19) {
                                    return 2;
                                }
                                else if (AFD >= 19) {
                                    if (SC < 16539) {
                                        if (Cs < 7581) {
                                            return 4;
                                        }
                                        else if (Cs >= 7581) {
                                            if (Cs < 8064) {
                                                if (gchDC < 1) {
                                                    if (Cs < 7971) {
                                                        if (SC < 15194) {
                                                            return 2;
                                                        }
                                                        else if (SC >= 15194) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (Cs >= 7971) {
                                                        return 2;
                                                    }
                                                }
                                                else if (gchDC >= 1) {
                                                    return 2;
                                                }
                                            }
                                            else if (Cs >= 8064) {
                                                return 4;
                                            }
                                        }
                                    }
                                    else if (SC >= 16539) {
                                        return 2;
                                    }
                                }
                            }
                        }
                    }
                }
            }
        }
        else if (MVDiff >= 6) {
            if (Cs < 3285) {
                if (mu_mv_mag_sq < 42) {
                    if (Cs < 1466) {
                        if (RsDiff < 7) {
                            if (MVDiff < 17) {
                                if (tcor < 81) {
                                    if (posBalance < 54) {
                                        if (posBalance < 40) {
                                            if (refDCval < 115) {
                                                return 2;
                                            }
                                            else if (refDCval >= 115) {
                                                if (diffTSC < -2) {
                                                    return 1;
                                                }
                                                else if (diffTSC >= -2) {
                                                    return 4;
                                                }
                                            }
                                        }
                                        else if (posBalance >= 40) {
                                            if (MVDiff < 7) {
                                                return 4;
                                            }
                                            else if (MVDiff >= 7) {
                                                if (RsCsDiff < 73) {
                                                    return 1;
                                                }
                                                else if (RsCsDiff >= 73) {
                                                    if (Rs < 1243) {
                                                        return 1;
                                                    }
                                                    else if (Rs >= 1243) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (posBalance >= 54) {
                                        if (negBalance < 51) {
                                            if (ssDCval < 112) {
                                                if (Cs < 636) {
                                                    if (TSC < 5) {
                                                        if (AFD < 7) {
                                                            if (diffRsCsdiff < -1) {
                                                                if (Rs < 209) {
                                                                    return 2;
                                                                }
                                                                else if (Rs >= 209) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (diffRsCsdiff >= -1) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (AFD >= 7) {
                                                            return 2;
                                                        }
                                                    }
                                                    else if (TSC >= 5) {
                                                        return 2;
                                                    }
                                                }
                                                else if (Cs >= 636) {
                                                    if (posBalance < 69) {
                                                        if (diffAFD < 3) {
                                                            return 4;
                                                        }
                                                        else if (diffAFD >= 3) {
                                                            return 1;
                                                        }
                                                    }
                                                    else if (posBalance >= 69) {
                                                        return 2;
                                                    }
                                                }
                                            }
                                            else if (ssDCval >= 112) {
                                                return 1;
                                            }
                                        }
                                        else if (negBalance >= 51) {
                                            if (refDCval < 99) {
                                                if (Cs < 1059) {
                                                    if (Cs < 510) {
                                                        return 8;
                                                    }
                                                    else if (Cs >= 510) {
                                                        if (refDCval < 97) {
                                                            if (SC < 1221) {
                                                                return 8;
                                                            }
                                                            else if (SC >= 1221) {
                                                                if (ssDCval < 78) {
                                                                    if (MVDiff < 11) {
                                                                        if (Rs < 1020) {
                                                                            return 8;
                                                                        }
                                                                        else if (Rs >= 1020) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                    else if (MVDiff >= 11) {
                                                                        return 8;
                                                                    }
                                                                }
                                                                else if (ssDCval >= 78) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                        else if (refDCval >= 97) {
                                                            if (SC < 2219) {
                                                                return 4;
                                                            }
                                                            else if (SC >= 2219) {
                                                                if (Rs < 1237) {
                                                                    return 2;
                                                                }
                                                                else if (Rs >= 1237) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (Cs >= 1059) {
                                                    if (Rs < 1478) {
                                                        if (Rs < 1352) {
                                                            return 8;
                                                        }
                                                        else if (Rs >= 1352) {
                                                            return 2;
                                                        }
                                                    }
                                                    else if (Rs >= 1478) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                            else if (refDCval >= 99) {
                                                if (Rs < 657) {
                                                    return 2;
                                                }
                                                else if (Rs >= 657) {
                                                    return 4;
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (tcor >= 81) {
                                    if (AFD < 11) {
                                        if (Cs < 684) {
                                            if (AFD < 2) {
                                                return 1;
                                            }
                                            else if (AFD >= 2) {
                                                if (RsDiff < 3) {
                                                    if (Rs < 268) {
                                                        if (TSC < 4) {
                                                            if (MVDiff < 14) {
                                                                return 8;
                                                            }
                                                            else if (MVDiff >= 14) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (TSC >= 4) {
                                                            return 2;
                                                        }
                                                    }
                                                    else if (Rs >= 268) {
                                                        if (refDCval < 71) {
                                                            if (TSC < 4) {
                                                                return 4;
                                                            }
                                                            else if (TSC >= 4) {
                                                                if (Rs < 789) {
                                                                    return 2;
                                                                }
                                                                else if (Rs >= 789) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                        else if (refDCval >= 71) {
                                                            if (mu_mv_mag_sq < 16) {
                                                                return 4;
                                                            }
                                                            else if (mu_mv_mag_sq >= 16) {
                                                                if (TSC < 7) {
                                                                    return 1;
                                                                }
                                                                else if (TSC >= 7) {
                                                                    if (Rs < 340) {
                                                                        return 4;
                                                                    }
                                                                    else if (Rs >= 340) {
                                                                        return 8;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (RsDiff >= 3) {
                                                    if (TSCindex < 7) {
                                                        if (posBalance < 75) {
                                                            if (refDCval < 60) {
                                                                return 4;
                                                            }
                                                            else if (refDCval >= 60) {
                                                                return 2;
                                                            }
                                                        }
                                                        else if (posBalance >= 75) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (TSCindex >= 7) {
                                                        if (gchDC < 1) {
                                                            return 2;
                                                        }
                                                        else if (gchDC >= 1) {
                                                            return 1;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (Cs >= 684) {
                                            if (SC < 2098) {
                                                if (Rs < 819) {
                                                    return 8;
                                                }
                                                else if (Rs >= 819) {
                                                    if (posBalance < 53) {
                                                        if (TSCindex < 6) {
                                                            if (mcTcor < 95) {
                                                                return 1;
                                                            }
                                                            else if (mcTcor >= 95) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (TSCindex >= 6) {
                                                            if (SC < 1940) {
                                                                return 4;
                                                            }
                                                            else if (SC >= 1940) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                    else if (posBalance >= 53) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                            else if (SC >= 2098) {
                                                if (negBalance < 42) {
                                                    if (Rs < 2159) {
                                                        return 1;
                                                    }
                                                    else if (Rs >= 2159) {
                                                        return 4;
                                                    }
                                                }
                                                else if (negBalance >= 42) {
                                                    if (Rs < 1271) {
                                                        return 8;
                                                    }
                                                    else if (Rs >= 1271) {
                                                        if (negBalance < 44) {
                                                            return 8;
                                                        }
                                                        else if (negBalance >= 44) {
                                                            if (Cs < 1092) {
                                                                if (SC < 2332) {
                                                                    return 1;
                                                                }
                                                                else if (SC >= 2332) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (Cs >= 1092) {
                                                                return 1;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (AFD >= 11) {
                                        if (negBalance < 56) {
                                            if (gchDC < 3) {
                                                if (ssDCval < 75) {
                                                    if (Rs < 1442) {
                                                        if (negBalance < 52) {
                                                            if (diffMVdiffVal < -9) {
                                                                return 1;
                                                            }
                                                            else if (diffMVdiffVal >= -9) {
                                                                return 2;
                                                            }
                                                        }
                                                        else if (negBalance >= 52) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (Rs >= 1442) {
                                                        return 1;
                                                    }
                                                }
                                                else if (ssDCval >= 75) {
                                                    if (TSC < 7) {
                                                        if (mu_mv_mag_sq < 19) {
                                                            return 8;
                                                        }
                                                        else if (mu_mv_mag_sq >= 19) {
                                                            return 2;
                                                        }
                                                    }
                                                    else if (TSC >= 7) {
                                                        return 2;
                                                    }
                                                }
                                            }
                                            else if (gchDC >= 3) {
                                                if (Rs < 794) {
                                                    return 8;
                                                }
                                                else if (Rs >= 794) {
                                                    return 4;
                                                }
                                            }
                                        }
                                        else if (negBalance >= 56) {
                                            return 1;
                                        }
                                    }
                                }
                            }
                            else if (MVDiff >= 17) {
                                if (TSC < 5) {
                                    if (Cs < 7) {
                                        if (posBalance < 22) {
                                            if (mu_mv_mag_sq < 14) {
                                                if (AFD < 1) {
                                                    return 8;
                                                }
                                                else if (AFD >= 1) {
                                                    if (diffMVdiffVal < 36) {
                                                        return 2;
                                                    }
                                                    else if (diffMVdiffVal >= 36) {
                                                        if (MVDiff < 692) {
                                                            return 8;
                                                        }
                                                        else if (MVDiff >= 692) {
                                                            return 2;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (mu_mv_mag_sq >= 14) {
                                                if (tcor < 82) {
                                                    if (AFD < 1) {
                                                        if (posBalance < 4) {
                                                            return 4;
                                                        }
                                                        else if (posBalance >= 4) {
                                                            if (diffMVdiffVal < 25) {
                                                                return 4;
                                                            }
                                                            else if (diffMVdiffVal >= 25) {
                                                                return 2;
                                                            }
                                                        }
                                                    }
                                                    else if (AFD >= 1) {
                                                        return 1;
                                                    }
                                                }
                                                else if (tcor >= 82) {
                                                    if (tcor < 113) {
                                                        if (ssDCval < 20) {
                                                            if (refDCval < 23) {
                                                                return 4;
                                                            }
                                                            else if (refDCval >= 23) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (ssDCval >= 20) {
                                                            if (posBalance < 5) {
                                                                return 2;
                                                            }
                                                            else if (posBalance >= 5) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                    else if (tcor >= 113) {
                                                        if (posBalance < 4) {
                                                            if (MVDiff < 21) {
                                                                return 8;
                                                            }
                                                            else if (MVDiff >= 21) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (posBalance >= 4) {
                                                            if (ssDCval < 25) {
                                                                return 2;
                                                            }
                                                            else if (ssDCval >= 25) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (posBalance >= 22) {
                                            if (posBalance < 29) {
                                                if (MVDiff < 62) {
                                                    if (posBalance < 27) {
                                                        return 1;
                                                    }
                                                    else if (posBalance >= 27) {
                                                        if (MVDiff < 59) {
                                                            return 4;
                                                        }
                                                        else if (MVDiff >= 59) {
                                                            return 1;
                                                        }
                                                    }
                                                }
                                                else if (MVDiff >= 62) {
                                                    if (diffMVdiffVal < 7) {
                                                        if (diffMVdiffVal < -225) {
                                                            return 1;
                                                        }
                                                        else if (diffMVdiffVal >= -225) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (diffMVdiffVal >= 7) {
                                                        if (posBalance < 27) {
                                                            return 4;
                                                        }
                                                        else if (posBalance >= 27) {
                                                            return 1;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (posBalance >= 29) {
                                                if (ssDCval < 26) {
                                                    if (mu_mv_mag_sq < 9) {
                                                        if (ssDCval < 20) {
                                                            return 8;
                                                        }
                                                        else if (ssDCval >= 20) {
                                                            if (MVDiff < 42) {
                                                                if (TSC < 1) {
                                                                    return 4;
                                                                }
                                                                else if (TSC >= 1) {
                                                                    return 2;
                                                                }
                                                            }
                                                            else if (MVDiff >= 42) {
                                                                return 1;
                                                            }
                                                        }
                                                    }
                                                    else if (mu_mv_mag_sq >= 9) {
                                                        if (posBalance < 72) {
                                                            if (MVDiff < 96) {
                                                                return 4;
                                                            }
                                                            else if (MVDiff >= 96) {
                                                                if (ssDCval < 21) {
                                                                    if (ssDCval < 19) {
                                                                        if (MVDiff < 1221) {
                                                                            return 4;
                                                                        }
                                                                        else if (MVDiff >= 1221) {
                                                                            return 2;
                                                                        }
                                                                    }
                                                                    else if (ssDCval >= 19) {
                                                                        return 2;
                                                                    }
                                                                }
                                                                else if (ssDCval >= 21) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                        else if (posBalance >= 72) {
                                                            if (MVDiff < 134) {
                                                                return 4;
                                                            }
                                                            else if (MVDiff >= 134) {
                                                                if (posBalance < 106) {
                                                                    if (mu_mv_mag_sq < 14) {
                                                                        return 4;
                                                                    }
                                                                    else if (mu_mv_mag_sq >= 14) {
                                                                        if (mu_mv_mag_sq < 33) {
                                                                            return 8;
                                                                        }
                                                                        else if (mu_mv_mag_sq >= 33) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                }
                                                                else if (posBalance >= 106) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (ssDCval >= 26) {
                                                    if (mu_mv_mag_sq < 4) {
                                                        return 2;
                                                    }
                                                    else if (mu_mv_mag_sq >= 4) {
                                                        if (AFD < 4) {
                                                            if (Rs < 1) {
                                                                return 1;
                                                            }
                                                            else if (Rs >= 1) {
                                                                if (Cs < 1) {
                                                                    return 4;
                                                                }
                                                                else if (Cs >= 1) {
                                                                    return 1;
                                                                }
                                                            }
                                                        }
                                                        else if (AFD >= 4) {
                                                            if (AFD < 7) {
                                                                return 8;
                                                            }
                                                            else if (AFD >= 7) {
                                                                return 1;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (Cs >= 7) {
                                        if (ssDCval < 115) {
                                            if (mcTcor < 79) {
                                                if (MVDiff < 42) {
                                                    if (mu_mv_mag_sq < 29) {
                                                        return 2;
                                                    }
                                                    else if (mu_mv_mag_sq >= 29) {
                                                        if (Cs < 49) {
                                                            return 4;
                                                        }
                                                        else if (Cs >= 49) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                                else if (MVDiff >= 42) {
                                                    if (TSC < 1) {
                                                        return 2;
                                                    }
                                                    else if (TSC >= 1) {
                                                        return 1;
                                                    }
                                                }
                                            }
                                            else if (mcTcor >= 79) {
                                                if (posBalance < 49) {
                                                    if (ssDCval < 38) {
                                                        if (diffRsCsdiff < -2) {
                                                            if (AFD < 6) {
                                                                return 4;
                                                            }
                                                            else if (AFD >= 6) {
                                                                return 2;
                                                            }
                                                        }
                                                        else if (diffRsCsdiff >= -2) {
                                                            return 2;
                                                        }
                                                    }
                                                    else if (ssDCval >= 38) {
                                                        if (negBalance < 57) {
                                                            if (AFD < 4) {
                                                                return 1;
                                                            }
                                                            else if (AFD >= 4) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (negBalance >= 57) {
                                                            if (diffMVdiffVal < -9) {
                                                                return 1;
                                                            }
                                                            else if (diffMVdiffVal >= -9) {
                                                                if (CsDiff < 3) {
                                                                    return 8;
                                                                }
                                                                else if (CsDiff >= 3) {
                                                                    return 2;
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (posBalance >= 49) {
                                                    if (gchDC < 4) {
                                                        if (mu_mv_mag_sq < 7) {
                                                            if (ssDCval < 88) {
                                                                if (diffAFD < -3) {
                                                                    return 8;
                                                                }
                                                                else if (diffAFD >= -3) {
                                                                    return 2;
                                                                }
                                                            }
                                                            else if (ssDCval >= 88) {
                                                                if (Rs < 452) {
                                                                    if (posBalance < 57) {
                                                                        return 2;
                                                                    }
                                                                    else if (posBalance >= 57) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (Rs >= 452) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                        else if (mu_mv_mag_sq >= 7) {
                                                            if (diffAFD < -5) {
                                                                return 8;
                                                            }
                                                            else if (diffAFD >= -5) {
                                                                if (mcTcor < 85) {
                                                                    if (diffMVdiffVal < -6) {
                                                                        return 1;
                                                                    }
                                                                    else if (diffMVdiffVal >= -6) {
                                                                        if (MVDiff < 21) {
                                                                            return 1;
                                                                        }
                                                                        else if (MVDiff >= 21) {
                                                                            return 2;
                                                                        }
                                                                    }
                                                                }
                                                                else if (mcTcor >= 85) {
                                                                    if (diffMVdiffVal < -15) {
                                                                        if (mu_mv_mag_sq < 31) {
                                                                            return 2;
                                                                        }
                                                                        else if (mu_mv_mag_sq >= 31) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                    else if (diffMVdiffVal >= -15) {
                                                                        return 2;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (gchDC >= 4) {
                                                        if (diffAFD < 1) {
                                                            if (Cs < 21) {
                                                                return 2;
                                                            }
                                                            else if (Cs >= 21) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (diffAFD >= 1) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (ssDCval >= 115) {
                                            if (AFD < 12) {
                                                return 1;
                                            }
                                            else if (AFD >= 12) {
                                                return 2;
                                            }
                                        }
                                    }
                                }
                                else if (TSC >= 5) {
                                    if (Cs < 250) {
                                        if (ssDCval < 108) {
                                            if (ssDCval < 77) {
                                                if (refDCval < 60) {
                                                    if (tcor < 112) {
                                                        if (MVDiff < 265) {
                                                            return 1;
                                                        }
                                                        else if (MVDiff >= 265) {
                                                            return 2;
                                                        }
                                                    }
                                                    else if (tcor >= 112) {
                                                        return 2;
                                                    }
                                                }
                                                else if (refDCval >= 60) {
                                                    return 2;
                                                }
                                            }
                                            else if (ssDCval >= 77) {
                                                if (ssDCval < 106) {
                                                    if (diffMVdiffVal < 17) {
                                                        return 1;
                                                    }
                                                    else if (diffMVdiffVal >= 17) {
                                                        return 2;
                                                    }
                                                }
                                                else if (ssDCval >= 106) {
                                                    if (Rs < 602) {
                                                        return 1;
                                                    }
                                                    else if (Rs >= 602) {
                                                        return 2;
                                                    }
                                                }
                                            }
                                        }
                                        else if (ssDCval >= 108) {
                                            if (RsDiff < 3) {
                                                if (SC < 374) {
                                                    return 1;
                                                }
                                                else if (SC >= 374) {
                                                    return 4;
                                                }
                                            }
                                            else if (RsDiff >= 3) {
                                                return 2;
                                            }
                                        }
                                    }
                                    else if (Cs >= 250) {
                                        if (ssDCval < 44) {
                                            if (Cs < 469) {
                                                if (Rs < 378) {
                                                    return 1;
                                                }
                                                else if (Rs >= 378) {
                                                    if (Rs < 452) {
                                                        return 4;
                                                    }
                                                    else if (Rs >= 452) {
                                                        if (mcTcor < 77) {
                                                            return 2;
                                                        }
                                                        else if (mcTcor >= 77) {
                                                            if (CsDiff < 3) {
                                                                return 4;
                                                            }
                                                            else if (CsDiff >= 3) {
                                                                return 2;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                            else if (Cs >= 469) {
                                                if (diffMVdiffVal < 12) {
                                                    if (mu_mv_mag_sq < 32) {
                                                        if (tcor < 71) {
                                                            if (Rs < 642) {
                                                                return 2;
                                                            }
                                                            else if (Rs >= 642) {
                                                                if (SC < 1334) {
                                                                    return 4;
                                                                }
                                                                else if (SC >= 1334) {
                                                                    if (Rs < 794) {
                                                                        return 4;
                                                                    }
                                                                    else if (Rs >= 794) {
                                                                        return 1;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (tcor >= 71) {
                                                            if (Cs < 512) {
                                                                return 4;
                                                            }
                                                            else if (Cs >= 512) {
                                                                return 1;
                                                            }
                                                        }
                                                    }
                                                    else if (mu_mv_mag_sq >= 32) {
                                                        if (SC < 1599) {
                                                            if (negBalance < 47) {
                                                                return 4;
                                                            }
                                                            else if (negBalance >= 47) {
                                                                if (mcTcor < 80) {
                                                                    if (RsDiff < 2) {
                                                                        if (negBalance < 50) {
                                                                            return 4;
                                                                        }
                                                                        else if (negBalance >= 50) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                    else if (RsDiff >= 2) {
                                                                        if (diffMVdiffVal < -10) {
                                                                            return 4;
                                                                        }
                                                                        else if (diffMVdiffVal >= -10) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                }
                                                                else if (mcTcor >= 80) {
                                                                    return 1;
                                                                }
                                                            }
                                                        }
                                                        else if (SC >= 1599) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                                else if (diffMVdiffVal >= 12) {
                                                    if (diffRsCsdiff < -2) {
                                                        if (Rs < 595) {
                                                            return 2;
                                                        }
                                                        else if (Rs >= 595) {
                                                            return 1;
                                                        }
                                                    }
                                                    else if (diffRsCsdiff >= -2) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                        }
                                        else if (ssDCval >= 44) {
                                            if (posBalance < 64) {
                                                if (SC < 2133) {
                                                    if (Cs < 675) {
                                                        if (RsCsDiff < 15) {
                                                            if (Rs < 290) {
                                                                if (AFD < 15) {
                                                                    return 2;
                                                                }
                                                                else if (AFD >= 15) {
                                                                    return 1;
                                                                }
                                                            }
                                                            else if (Rs >= 290) {
                                                                if (AFD < 8) {
                                                                    return 4;
                                                                }
                                                                else if (AFD >= 8) {
                                                                    if (refDCval < 99) {
                                                                        if (diffRsCsdiff < -2) {
                                                                            if (Cs < 435) {
                                                                                return 2;
                                                                            }
                                                                            else if (Cs >= 435) {
                                                                                return 8;
                                                                            }
                                                                        }
                                                                        else if (diffRsCsdiff >= -2) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                    else if (refDCval >= 99) {
                                                                        if (SC < 1108) {
                                                                            if (refDCval < 110) {
                                                                                return 8;
                                                                            }
                                                                            else if (refDCval >= 110) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                        else if (SC >= 1108) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (RsCsDiff >= 15) {
                                                            if (MVDiff < 57) {
                                                                if (ssDCval < 98) {
                                                                    if (SC < 889) {
                                                                        if (Rs < 395) {
                                                                            return 2;
                                                                        }
                                                                        else if (Rs >= 395) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                    else if (SC >= 889) {
                                                                        return 2;
                                                                    }
                                                                }
                                                                else if (ssDCval >= 98) {
                                                                    if (AFD < 11) {
                                                                        if (Rs < 1001) {
                                                                            return 2;
                                                                        }
                                                                        else if (Rs >= 1001) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                    else if (AFD >= 11) {
                                                                        return 1;
                                                                    }
                                                                }
                                                            }
                                                            else if (MVDiff >= 57) {
                                                                if (SC < 1224) {
                                                                    if (Rs < 299) {
                                                                        return 8;
                                                                    }
                                                                    else if (Rs >= 299) {
                                                                        return 2;
                                                                    }
                                                                }
                                                                else if (SC >= 1224) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (Cs >= 675) {
                                                        if (negBalance < 60) {
                                                            if (mcTcor < 93) {
                                                                if (mu_mv_mag_sq < 8) {
                                                                    if (diffAFD < -1) {
                                                                        return 2;
                                                                    }
                                                                    else if (diffAFD >= -1) {
                                                                        if (SC < 1390) {
                                                                            return 8;
                                                                        }
                                                                        else if (SC >= 1390) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                }
                                                                else if (mu_mv_mag_sq >= 8) {
                                                                    if (Rs < 926) {
                                                                        if (Rs < 610) {
                                                                            return 2;
                                                                        }
                                                                        else if (Rs >= 610) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                    else if (Rs >= 926) {
                                                                        if (AFD < 14) {
                                                                            return 1;
                                                                        }
                                                                        else if (AFD >= 14) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                            else if (mcTcor >= 93) {
                                                                if (ssDCval < 67) {
                                                                    if (Cs < 721) {
                                                                        return 2;
                                                                    }
                                                                    else if (Cs >= 721) {
                                                                        return 1;
                                                                    }
                                                                }
                                                                else if (ssDCval >= 67) {
                                                                    return 2;
                                                                }
                                                            }
                                                        }
                                                        else if (negBalance >= 60) {
                                                            if (CsDiff < 6) {
                                                                return 8;
                                                            }
                                                            else if (CsDiff >= 6) {
                                                                return 1;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (SC >= 2133) {
                                                    if (tcor < 77) {
                                                        if (Rs < 972) {
                                                            return 1;
                                                        }
                                                        else if (Rs >= 972) {
                                                            if (Rs < 1576) {
                                                                return 8;
                                                            }
                                                            else if (Rs >= 1576) {
                                                                return 1;
                                                            }
                                                        }
                                                    }
                                                    else if (tcor >= 77) {
                                                        if (diffAFD < -4) {
                                                            if (SC < 2246) {
                                                                return 1;
                                                            }
                                                            else if (SC >= 2246) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (diffAFD >= -4) {
                                                            if (Rs < 1954) {
                                                                return 1;
                                                            }
                                                            else if (Rs >= 1954) {
                                                                return 2;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                            else if (posBalance >= 64) {
                                                if (negBalance < 39) {
                                                    if (mu_mv_mag_sq < 14) {
                                                        if (Cs < 402) {
                                                            return 2;
                                                        }
                                                        else if (Cs >= 402) {
                                                            return 1;
                                                        }
                                                    }
                                                    else if (mu_mv_mag_sq >= 14) {
                                                        if (RsDiff < 3) {
                                                            return 4;
                                                        }
                                                        else if (RsDiff >= 3) {
                                                            if (Cs < 592) {
                                                                return 8;
                                                            }
                                                            else if (Cs >= 592) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (negBalance >= 39) {
                                                    if (TSC < 7) {
                                                        if (mcTcor < 94) {
                                                            return 1;
                                                        }
                                                        else if (mcTcor >= 94) {
                                                            return 2;
                                                        }
                                                    }
                                                    else if (TSC >= 7) {
                                                        if (Rs < 235) {
                                                            return 2;
                                                        }
                                                        else if (Rs >= 235) {
                                                            return 1;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                        }
                        else if (RsDiff >= 7) {
                            if (tcor < 80) {
                                if (RsDiff < 11) {
                                    if (refDCval < 136) {
                                        if (CsDiff < 8) {
                                            if (tcor < 60) {
                                                if (tcor < 47) {
                                                    if (SC < 2242) {
                                                        if (Rs < 1251) {
                                                            return 2;
                                                        }
                                                        else if (Rs >= 1251) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (SC >= 2242) {
                                                        return 8;
                                                    }
                                                }
                                                else if (tcor >= 47) {
                                                    if (Cs < 1105) {
                                                        return 1;
                                                    }
                                                    else if (Cs >= 1105) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                            else if (tcor >= 60) {
                                                if (mcTcor < 88) {
                                                    if (gchDC < 2) {
                                                        if (MVDiff < 8) {
                                                            return 4;
                                                        }
                                                        else if (MVDiff >= 8) {
                                                            return 1;
                                                        }
                                                    }
                                                    else if (gchDC >= 2) {
                                                        if (Rs < 866) {
                                                            return 1;
                                                        }
                                                        else if (Rs >= 866) {
                                                            return 2;
                                                        }
                                                    }
                                                }
                                                else if (mcTcor >= 88) {
                                                    if (MVDiff < 16) {
                                                        if (Rs < 2110) {
                                                            return 4;
                                                        }
                                                        else if (Rs >= 2110) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (MVDiff >= 16) {
                                                        return 2;
                                                    }
                                                }
                                            }
                                        }
                                        else if (CsDiff >= 8) {
                                            if (AFD < 35) {
                                                if (Cs < 1007) {
                                                    return 2;
                                                }
                                                else if (Cs >= 1007) {
                                                    return 1;
                                                }
                                            }
                                            else if (AFD >= 35) {
                                                return 4;
                                            }
                                        }
                                    }
                                    else if (refDCval >= 136) {
                                        if (refDCval < 146) {
                                            return 8;
                                        }
                                        else if (refDCval >= 146) {
                                            if (AFD < 20) {
                                                return 4;
                                            }
                                            else if (AFD >= 20) {
                                                return 1;
                                            }
                                        }
                                    }
                                }
                                else if (RsDiff >= 11) {
                                    if (diffMVdiffVal < 105) {
                                        if (mcTcor < 87) {
                                            if (TSC < 11) {
                                                return 1;
                                            }
                                            else if (TSC >= 11) {
                                                if (refDCval < 107) {
                                                    if (AFD < 29) {
                                                        if (AFD < 26) {
                                                            if (MVDiff < 41) {
                                                                return 1;
                                                            }
                                                            else if (MVDiff >= 41) {
                                                                return 2;
                                                            }
                                                        }
                                                        else if (AFD >= 26) {
                                                            if (RsDiff < 22) {
                                                                if (diffAFD < 3) {
                                                                    return 1;
                                                                }
                                                                else if (diffAFD >= 3) {
                                                                    return 2;
                                                                }
                                                            }
                                                            else if (RsDiff >= 22) {
                                                                return 2;
                                                            }
                                                        }
                                                    }
                                                    else if (AFD >= 29) {
                                                        if (Cs < 696) {
                                                            return 1;
                                                        }
                                                        else if (Cs >= 696) {
                                                            return 2;
                                                        }
                                                    }
                                                }
                                                else if (refDCval >= 107) {
                                                    if (diffRsCsdiff < 59) {
                                                        return 2;
                                                    }
                                                    else if (diffRsCsdiff >= 59) {
                                                        if (SC < 3783) {
                                                            return 1;
                                                        }
                                                        else if (SC >= 3783) {
                                                            return 2;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (mcTcor >= 87) {
                                            if (Rs < 4677) {
                                                if (RsDiff < 15) {
                                                    return 2;
                                                }
                                                else if (RsDiff >= 15) {
                                                    if (negBalance < 59) {
                                                        return 2;
                                                    }
                                                    else if (negBalance >= 59) {
                                                        return 1;
                                                    }
                                                }
                                            }
                                            else if (Rs >= 4677) {
                                                return 4;
                                            }
                                        }
                                    }
                                    else if (diffMVdiffVal >= 105) {
                                        if (posBalance < 65) {
                                            return 4;
                                        }
                                        else if (posBalance >= 65) {
                                            if (Cs < 1147) {
                                                if (Rs < 2415) {
                                                    return 8;
                                                }
                                                else if (Rs >= 2415) {
                                                    return 2;
                                                }
                                            }
                                            else if (Cs >= 1147) {
                                                return 1;
                                            }
                                        }
                                    }
                                }
                            }
                            else if (tcor >= 80) {
                                if (TSC < 10) {
                                    if (negBalance < 41) {
                                        if (SC < 3070) {
                                            if (Rs < 1091) {
                                                return 2;
                                            }
                                            else if (Rs >= 1091) {
                                                return 1;
                                            }
                                        }
                                        else if (SC >= 3070) {
                                            return 8;
                                        }
                                    }
                                    else if (negBalance >= 41) {
                                        if (Rs < 2045) {
                                            if (refDCval < 74) {
                                                if (SC < 1966) {
                                                    return 2;
                                                }
                                                else if (SC >= 1966) {
                                                    return 4;
                                                }
                                            }
                                            else if (refDCval >= 74) {
                                                return 2;
                                            }
                                        }
                                        else if (Rs >= 2045) {
                                            if (refDCval < 123) {
                                                return 4;
                                            }
                                            else if (refDCval >= 123) {
                                                return 2;
                                            }
                                        }
                                    }
                                }
                                else if (TSC >= 10) {
                                    if (MVDiff < 114) {
                                        return 2;
                                    }
                                    else if (MVDiff >= 114) {
                                        if (Cs < 1128) {
                                            return 1;
                                        }
                                        else if (Cs >= 1128) {
                                            return 4;
                                        }
                                    }
                                }
                            }
                        }
                    }
                    else if (Cs >= 1466) {
                        if (RsCsDiff < 159) {
                            if (refDCval < 116) {
                                if (negBalance < 50) {
                                    if (SC < 4460) {
                                        if (diffRsCsdiff < -62) {
                                            if (SC < 3180) {
                                                return 2;
                                            }
                                            else if (SC >= 3180) {
                                                return 1;
                                            }
                                        }
                                        else if (diffRsCsdiff >= -62) {
                                            if (TSC < 11) {
                                                if (posBalance < 52) {
                                                    if (Rs < 1650) {
                                                        if (CsDiff < 7) {
                                                            if (Rs < 1469) {
                                                                return 4;
                                                            }
                                                            else if (Rs >= 1469) {
                                                                return 2;
                                                            }
                                                        }
                                                        else if (CsDiff >= 7) {
                                                            return 1;
                                                        }
                                                    }
                                                    else if (Rs >= 1650) {
                                                        if (diffAFD < 3) {
                                                            if (MVDiff < 8) {
                                                                return 4;
                                                            }
                                                            else if (MVDiff >= 8) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (diffAFD >= 3) {
                                                            if (Cs < 1987) {
                                                                return 8;
                                                            }
                                                            else if (Cs >= 1987) {
                                                                return 1;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (posBalance >= 52) {
                                                    if (posBalance < 61) {
                                                        if (mcTcor < 84) {
                                                            if (Rs < 1611) {
                                                                return 8;
                                                            }
                                                            else if (Rs >= 1611) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (mcTcor >= 84) {
                                                            if (mu_mv_mag_sq < 41) {
                                                                if (diffAFD < -2) {
                                                                    return 4;
                                                                }
                                                                else if (diffAFD >= -2) {
                                                                    return 1;
                                                                }
                                                            }
                                                            else if (mu_mv_mag_sq >= 41) {
                                                                if (Cs < 1552) {
                                                                    return 4;
                                                                }
                                                                else if (Cs >= 1552) {
                                                                    return 1;
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (posBalance >= 61) {
                                                        if (ssDCval < 85) {
                                                            return 2;
                                                        }
                                                        else if (ssDCval >= 85) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (TSC >= 11) {
                                                if (MVDiff < 29) {
                                                    return 4;
                                                }
                                                else if (MVDiff >= 29) {
                                                    if (Cs < 1905) {
                                                        return 4;
                                                    }
                                                    else if (Cs >= 1905) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (SC >= 4460) {
                                        if (tcor < 84) {
                                            if (refDCval < 108) {
                                                if (SC < 7086) {
                                                    if (CsDiff < 9) {
                                                        return 8;
                                                    }
                                                    else if (CsDiff >= 9) {
                                                        return 4;
                                                    }
                                                }
                                                else if (SC >= 7086) {
                                                    return 4;
                                                }
                                            }
                                            else if (refDCval >= 108) {
                                                if (Cs < 2022) {
                                                    return 2;
                                                }
                                                else if (Cs >= 2022) {
                                                    return 1;
                                                }
                                            }
                                        }
                                        else if (tcor >= 84) {
                                            if (posBalance < 59) {
                                                if (AFD < 16) {
                                                    if (ssDCval < 106) {
                                                        if (negBalance < 45) {
                                                            return 8;
                                                        }
                                                        else if (negBalance >= 45) {
                                                            if (diffMVdiffVal < 7) {
                                                                if (MVDiff < 23) {
                                                                    if (diffAFD < -1) {
                                                                        return 8;
                                                                    }
                                                                    else if (diffAFD >= -1) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (MVDiff >= 23) {
                                                                    if (diffRsCsdiff < -32) {
                                                                        if (Rs < 2447) {
                                                                            return 4;
                                                                        }
                                                                        else if (Rs >= 2447) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                    else if (diffRsCsdiff >= -32) {
                                                                        return 8;
                                                                    }
                                                                }
                                                            }
                                                            else if (diffMVdiffVal >= 7) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                    else if (ssDCval >= 106) {
                                                        if (mu_mv_mag_sq < 19) {
                                                            if (RsDiff < 8) {
                                                                if (SC < 5053) {
                                                                    if (diffMVdiffVal < 4) {
                                                                        return 4;
                                                                    }
                                                                    else if (diffMVdiffVal >= 4) {
                                                                        return 8;
                                                                    }
                                                                }
                                                                else if (SC >= 5053) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (RsDiff >= 8) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (mu_mv_mag_sq >= 19) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                                else if (AFD >= 16) {
                                                    return 4;
                                                }
                                            }
                                            else if (posBalance >= 59) {
                                                if (TSC < 7) {
                                                    if (SC < 5860) {
                                                        return 8;
                                                    }
                                                    else if (SC >= 5860) {
                                                        return 4;
                                                    }
                                                }
                                                else if (TSC >= 7) {
                                                    if (Rs < 3663) {
                                                        return 2;
                                                    }
                                                    else if (Rs >= 3663) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (negBalance >= 50) {
                                    if (refDCval < 76) {
                                        if (RsCsDiff < 92) {
                                            if (Cs < 2318) {
                                                return 8;
                                            }
                                            else if (Cs >= 2318) {
                                                if (mu_mv_mag_sq < 35) {
                                                    if (refDCval < 55) {
                                                        return 8;
                                                    }
                                                    else if (refDCval >= 55) {
                                                        return 2;
                                                    }
                                                }
                                                else if (mu_mv_mag_sq >= 35) {
                                                    if (MVDiff < 26) {
                                                        if (SC < 6437) {
                                                            return 2;
                                                        }
                                                        else if (SC >= 6437) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (MVDiff >= 26) {
                                                        return 2;
                                                    }
                                                }
                                            }
                                        }
                                        else if (RsCsDiff >= 92) {
                                            if (Rs < 2442) {
                                                if (MVDiff < 27) {
                                                    return 4;
                                                }
                                                else if (MVDiff >= 27) {
                                                    return 8;
                                                }
                                            }
                                            else if (Rs >= 2442) {
                                                if (MVDiff < 23) {
                                                    return 8;
                                                }
                                                else if (MVDiff >= 23) {
                                                    return 4;
                                                }
                                            }
                                        }
                                    }
                                    else if (refDCval >= 76) {
                                        if (TSC < 9) {
                                            if (posBalance < 56) {
                                                if (tcor < 78) {
                                                    if (Cs < 1916) {
                                                        return 2;
                                                    }
                                                    else if (Cs >= 1916) {
                                                        return 8;
                                                    }
                                                }
                                                else if (tcor >= 78) {
                                                    return 1;
                                                }
                                            }
                                            else if (posBalance >= 56) {
                                                if (CsDiff < 5) {
                                                    if (Cs < 1703) {
                                                        return 8;
                                                    }
                                                    else if (Cs >= 1703) {
                                                        return 4;
                                                    }
                                                }
                                                else if (CsDiff >= 5) {
                                                    if (Rs < 1916) {
                                                        return 4;
                                                    }
                                                    else if (Rs >= 1916) {
                                                        if (Rs < 2227) {
                                                            return 2;
                                                        }
                                                        else if (Rs >= 2227) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (TSC >= 9) {
                                            if (tcor < 74) {
                                                return 4;
                                            }
                                            else if (tcor >= 74) {
                                                if (ssDCval < 105) {
                                                    if (Cs < 2052) {
                                                        if (mcTcor < 96) {
                                                            if (posBalance < 59) {
                                                                if (diffMVdiffVal < 4) {
                                                                    if (refDCval < 81) {
                                                                        return 2;
                                                                    }
                                                                    else if (refDCval >= 81) {
                                                                        if (Rs < 2436) {
                                                                            return 2;
                                                                        }
                                                                        else if (Rs >= 2436) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                }
                                                                else if (diffMVdiffVal >= 4) {
                                                                    return 2;
                                                                }
                                                            }
                                                            else if (posBalance >= 59) {
                                                                if (gchDC < 1) {
                                                                    return 1;
                                                                }
                                                                else if (gchDC >= 1) {
                                                                    return 2;
                                                                }
                                                            }
                                                        }
                                                        else if (mcTcor >= 96) {
                                                            if (MVDiff < 12) {
                                                                if (Cs < 1817) {
                                                                    return 4;
                                                                }
                                                                else if (Cs >= 1817) {
                                                                    return 2;
                                                                }
                                                            }
                                                            else if (MVDiff >= 12) {
                                                                return 2;
                                                            }
                                                        }
                                                    }
                                                    else if (Cs >= 2052) {
                                                        if (mu_mv_mag_sq < 36) {
                                                            if (SC < 4947) {
                                                                return 2;
                                                            }
                                                            else if (SC >= 4947) {
                                                                return 1;
                                                            }
                                                        }
                                                        else if (mu_mv_mag_sq >= 36) {
                                                            return 2;
                                                        }
                                                    }
                                                }
                                                else if (ssDCval >= 105) {
                                                    if (Rs < 2985) {
                                                        return 1;
                                                    }
                                                    else if (Rs >= 2985) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                            else if (refDCval >= 116) {
                                if (tcor < 72) {
                                    if (MVDiff < 119) {
                                        return 1;
                                    }
                                    else if (MVDiff >= 119) {
                                        return 8;
                                    }
                                }
                                else if (tcor >= 72) {
                                    if (diffTSC < -1) {
                                        if (tcor < 83) {
                                            if (SC < 4024) {
                                                if (AFD < 11) {
                                                    return 1;
                                                }
                                                else if (AFD >= 11) {
                                                    return 4;
                                                }
                                            }
                                            else if (SC >= 4024) {
                                                return 1;
                                            }
                                        }
                                        else if (tcor >= 83) {
                                            if (CsDiff < 7) {
                                                if (Cs < 1672) {
                                                    return 8;
                                                }
                                                else if (Cs >= 1672) {
                                                    return 4;
                                                }
                                            }
                                            else if (CsDiff >= 7) {
                                                return 2;
                                            }
                                        }
                                    }
                                    else if (diffTSC >= -1) {
                                        if (gchDC < 3) {
                                            if (posBalance < 58) {
                                                if (RsCsDiff < 124) {
                                                    if (refDCval < 129) {
                                                        return 4;
                                                    }
                                                    else if (refDCval >= 129) {
                                                        if (Cs < 1613) {
                                                            return 1;
                                                        }
                                                        else if (Cs >= 1613) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                                else if (RsCsDiff >= 124) {
                                                    if (TSC < 10) {
                                                        return 2;
                                                    }
                                                    else if (TSC >= 10) {
                                                        return 1;
                                                    }
                                                }
                                            }
                                            else if (posBalance >= 58) {
                                                return 4;
                                            }
                                        }
                                        else if (gchDC >= 3) {
                                            return 8;
                                        }
                                    }
                                }
                            }
                        }
                        else if (RsCsDiff >= 159) {
                            if (ssDCval < 140) {
                                if (ssDCval < 134) {
                                    if (mu_mv_mag_sq < 2) {
                                        if (negBalance < 51) {
                                            if (CsDiff < 9) {
                                                if (SC < 6076) {
                                                    return 2;
                                                }
                                                else if (SC >= 6076) {
                                                    return 4;
                                                }
                                            }
                                            else if (CsDiff >= 9) {
                                                if (refDCval < 84) {
                                                    if (RsDiff < 11) {
                                                        if (ssDCval < 77) {
                                                            return 4;
                                                        }
                                                        else if (ssDCval >= 77) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (RsDiff >= 11) {
                                                        if (MVDiff < 25) {
                                                            return 1;
                                                        }
                                                        else if (MVDiff >= 25) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                                else if (refDCval >= 84) {
                                                    if (MVDiff < 7) {
                                                        if (diffMVdiffVal < 2) {
                                                            if (ssDCval < 129) {
                                                                return 4;
                                                            }
                                                            else if (ssDCval >= 129) {
                                                                return 1;
                                                            }
                                                        }
                                                        else if (diffMVdiffVal >= 2) {
                                                            if (Cs < 2514) {
                                                                return 2;
                                                            }
                                                            else if (Cs >= 2514) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                    else if (MVDiff >= 7) {
                                                        if (posBalance < 59) {
                                                            if (diffRsCsdiff < 33) {
                                                                if (AFD < 14) {
                                                                    return 4;
                                                                }
                                                                else if (AFD >= 14) {
                                                                    return 1;
                                                                }
                                                            }
                                                            else if (diffRsCsdiff >= 33) {
                                                                if (RsCsDiff < 226) {
                                                                    return 4;
                                                                }
                                                                else if (RsCsDiff >= 226) {
                                                                    return 1;
                                                                }
                                                            }
                                                        }
                                                        else if (posBalance >= 59) {
                                                            if (ssDCval < 121) {
                                                                if (Cs < 2672) {
                                                                    return 2;
                                                                }
                                                                else if (Cs >= 2672) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (ssDCval >= 121) {
                                                                return 1;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (negBalance >= 51) {
                                            if (mcTcor < 85) {
                                                if (ssDCval < 125) {
                                                    if (RsCsDiff < 173) {
                                                        if (AFD < 17) {
                                                            return 4;
                                                        }
                                                        else if (AFD >= 17) {
                                                            return 1;
                                                        }
                                                    }
                                                    else if (RsCsDiff >= 173) {
                                                        if (Rs < 3571) {
                                                            return 4;
                                                        }
                                                        else if (Rs >= 3571) {
                                                            return 2;
                                                        }
                                                    }
                                                }
                                                else if (ssDCval >= 125) {
                                                    if (Rs < 2558) {
                                                        return 2;
                                                    }
                                                    else if (Rs >= 2558) {
                                                        return 1;
                                                    }
                                                }
                                            }
                                            else if (mcTcor >= 85) {
                                                if (gchDC < 1) {
                                                    if (RsCsDiff < 172) {
                                                        if (Rs < 2611) {
                                                            return 8;
                                                        }
                                                        else if (Rs >= 2611) {
                                                            return 1;
                                                        }
                                                    }
                                                    else if (RsCsDiff >= 172) {
                                                        if (AFD < 14) {
                                                            return 8;
                                                        }
                                                        else if (AFD >= 14) {
                                                            return 2;
                                                        }
                                                    }
                                                }
                                                else if (gchDC >= 1) {
                                                    if (AFD < 23) {
                                                        if (tcor < 82) {
                                                            if (RsCsDiff < 267) {
                                                                if (AFD < 18) {
                                                                    if (ssDCval < 117) {
                                                                        return 2;
                                                                    }
                                                                    else if (ssDCval >= 117) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (AFD >= 18) {
                                                                    if (Rs < 2489) {
                                                                        return 4;
                                                                    }
                                                                    else if (Rs >= 2489) {
                                                                        if (refDCval < 80) {
                                                                            return 4;
                                                                        }
                                                                        else if (refDCval >= 80) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                            else if (RsCsDiff >= 267) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (tcor >= 82) {
                                                            if (ssDCval < 80) {
                                                                return 4;
                                                            }
                                                            else if (ssDCval >= 80) {
                                                                if (Rs < 2821) {
                                                                    if (RsCsDiff < 192) {
                                                                        if (Rs < 2441) {
                                                                            return 4;
                                                                        }
                                                                        else if (Rs >= 2441) {
                                                                            return 2;
                                                                        }
                                                                    }
                                                                    else if (RsCsDiff >= 192) {
                                                                        if (diffRsCsdiff < 60) {
                                                                            return 2;
                                                                        }
                                                                        else if (diffRsCsdiff >= 60) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                }
                                                                else if (Rs >= 2821) {
                                                                    if (TSC < 13) {
                                                                        return 2;
                                                                    }
                                                                    else if (TSC >= 13) {
                                                                        if (Cs < 2879) {
                                                                            return 1;
                                                                        }
                                                                        else if (Cs >= 2879) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (AFD >= 23) {
                                                        if (RsCsDiff < 267) {
                                                            return 1;
                                                        }
                                                        else if (RsCsDiff >= 267) {
                                                            return 2;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (mu_mv_mag_sq >= 2) {
                                        if (posBalance < 56) {
                                            if (mcTcor < 89) {
                                                if (CsDiff < 12) {
                                                    if (Cs < 2552) {
                                                        if (RsCsDiff < 192) {
                                                            return 2;
                                                        }
                                                        else if (RsCsDiff >= 192) {
                                                            return 1;
                                                        }
                                                    }
                                                    else if (Cs >= 2552) {
                                                        if (refDCval < 108) {
                                                            if (SC < 6386) {
                                                                if (Cs < 2955) {
                                                                    return 8;
                                                                }
                                                                else if (Cs >= 2955) {
                                                                    return 1;
                                                                }
                                                            }
                                                            else if (SC >= 6386) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (refDCval >= 108) {
                                                            if (Rs < 4643) {
                                                                return 1;
                                                            }
                                                            else if (Rs >= 4643) {
                                                                return 2;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (CsDiff >= 12) {
                                                    if (negBalance < 47) {
                                                        if (SC < 3739) {
                                                            return 2;
                                                        }
                                                        else if (SC >= 3739) {
                                                            return 1;
                                                        }
                                                    }
                                                    else if (negBalance >= 47) {
                                                        if (diffAFD < -3) {
                                                            if (TSC < 12) {
                                                                return 2;
                                                            }
                                                            else if (TSC >= 12) {
                                                                return 1;
                                                            }
                                                        }
                                                        else if (diffAFD >= -3) {
                                                            if (RsDiff < 10) {
                                                                return 4;
                                                            }
                                                            else if (RsDiff >= 10) {
                                                                if (RsCsDiff < 687) {
                                                                    if (gchDC < 3) {
                                                                        if (refDCval < 101) {
                                                                            return 8;
                                                                        }
                                                                        else if (refDCval >= 101) {
                                                                            if (SC < 5628) {
                                                                                return 2;
                                                                            }
                                                                            else if (SC >= 5628) {
                                                                                return 8;
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (gchDC >= 3) {
                                                                        if (posBalance < 51) {
                                                                            return 8;
                                                                        }
                                                                        else if (posBalance >= 51) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                }
                                                                else if (RsCsDiff >= 687) {
                                                                    if (posBalance < 53) {
                                                                        return 2;
                                                                    }
                                                                    else if (posBalance >= 53) {
                                                                        return 1;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                            else if (mcTcor >= 89) {
                                                if (posBalance < 55) {
                                                    if (refDCval < 96) {
                                                        if (Rs < 2553) {
                                                            if (tcor < 82) {
                                                                return 1;
                                                            }
                                                            else if (tcor >= 82) {
                                                                if (mcTcor < 94) {
                                                                    if (Cs < 2918) {
                                                                        if (MVDiff < 33) {
                                                                            if (mu_mv_mag_sq < 28) {
                                                                                return 1;
                                                                            }
                                                                            else if (mu_mv_mag_sq >= 28) {
                                                                                if (Rs < 2336) {
                                                                                    return 1;
                                                                                }
                                                                                else if (Rs >= 2336) {
                                                                                    return 8;
                                                                                }
                                                                            }
                                                                        }
                                                                        else if (MVDiff >= 33) {
                                                                            if (Cs < 2209) {
                                                                                if (Rs < 2187) {
                                                                                    return 1;
                                                                                }
                                                                                else if (Rs >= 2187) {
                                                                                    return 8;
                                                                                }
                                                                            }
                                                                            else if (Cs >= 2209) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (Cs >= 2918) {
                                                                        return 8;
                                                                    }
                                                                }
                                                                else if (mcTcor >= 94) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                        else if (Rs >= 2553) {
                                                            return 1;
                                                        }
                                                    }
                                                    else if (refDCval >= 96) {
                                                        if (TSC < 14) {
                                                            if (posBalance < 54) {
                                                                if (tcor < 85) {
                                                                    if (negBalance < 40) {
                                                                        return 8;
                                                                    }
                                                                    else if (negBalance >= 40) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (tcor >= 85) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (posBalance >= 54) {
                                                                if (Rs < 2466) {
                                                                    return 4;
                                                                }
                                                                else if (Rs >= 2466) {
                                                                    return 1;
                                                                }
                                                            }
                                                        }
                                                        else if (TSC >= 14) {
                                                            return 1;
                                                        }
                                                    }
                                                }
                                                else if (posBalance >= 55) {
                                                    if (RsCsDiff < 298) {
                                                        return 8;
                                                    }
                                                    else if (RsCsDiff >= 298) {
                                                        if (Cs < 1924) {
                                                            return 2;
                                                        }
                                                        else if (Cs >= 1924) {
                                                            return 1;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (posBalance >= 56) {
                                            if (ssDCval < 75) {
                                                if (RsDiff < 9) {
                                                    return 1;
                                                }
                                                else if (RsDiff >= 9) {
                                                    if (AFD < 22) {
                                                        if (Rs < 2086) {
                                                            return 4;
                                                        }
                                                        else if (Rs >= 2086) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (AFD >= 22) {
                                                        return 1;
                                                    }
                                                }
                                            }
                                            else if (ssDCval >= 75) {
                                                if (tcor < 80) {
                                                    if (refDCval < 108) {
                                                        if (TSC < 12) {
                                                            if (SC < 4019) {
                                                                if (MVDiff < 21) {
                                                                    return 2;
                                                                }
                                                                else if (MVDiff >= 21) {
                                                                    if (Rs < 1638) {
                                                                        return 1;
                                                                    }
                                                                    else if (Rs >= 1638) {
                                                                        return 4;
                                                                    }
                                                                }
                                                            }
                                                            else if (SC >= 4019) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (TSC >= 12) {
                                                            if (AFD < 29) {
                                                                if (RsDiff < 12) {
                                                                    if (Cs < 3199) {
                                                                        return 1;
                                                                    }
                                                                    else if (Cs >= 3199) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (RsDiff >= 12) {
                                                                    if (Rs < 3077) {
                                                                        if (MVDiff < 29) {
                                                                            return 4;
                                                                        }
                                                                        else if (MVDiff >= 29) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                    else if (Rs >= 3077) {
                                                                        if (AFD < 22) {
                                                                            return 1;
                                                                        }
                                                                        else if (AFD >= 22) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                            else if (AFD >= 29) {
                                                                if (TSC < 18) {
                                                                    return 2;
                                                                }
                                                                else if (TSC >= 18) {
                                                                    if (Cs < 2692) {
                                                                        return 2;
                                                                    }
                                                                    else if (Cs >= 2692) {
                                                                        if (SC < 5265) {
                                                                            return 4;
                                                                        }
                                                                        else if (SC >= 5265) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (refDCval >= 108) {
                                                        if (SC < 3072) {
                                                            return 8;
                                                        }
                                                        else if (SC >= 3072) {
                                                            if (diffMVdiffVal < 8) {
                                                                if (tcor < 69) {
                                                                    if (refDCval < 115) {
                                                                        if (TSC < 13) {
                                                                            if (Rs < 2923) {
                                                                                return 4;
                                                                            }
                                                                            else if (Rs >= 2923) {
                                                                                return 2;
                                                                            }
                                                                        }
                                                                        else if (TSC >= 13) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                    else if (refDCval >= 115) {
                                                                        return 1;
                                                                    }
                                                                }
                                                                else if (tcor >= 69) {
                                                                    if (Cs < 3049) {
                                                                        if (diffMVdiffVal < -51) {
                                                                            if (Rs < 3311) {
                                                                                return 8;
                                                                            }
                                                                            else if (Rs >= 3311) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                        else if (diffMVdiffVal >= -51) {
                                                                            if (refDCval < 135) {
                                                                                if (SC < 6489) {
                                                                                    return 1;
                                                                                }
                                                                                else if (SC >= 6489) {
                                                                                    if (Rs < 3881) {
                                                                                        if (AFD < 21) {
                                                                                            return 4;
                                                                                        }
                                                                                        else if (AFD >= 21) {
                                                                                            return 1;
                                                                                        }
                                                                                    }
                                                                                    else if (Rs >= 3881) {
                                                                                        return 1;
                                                                                    }
                                                                                }
                                                                            }
                                                                            else if (refDCval >= 135) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (Cs >= 3049) {
                                                                        if (RsDiff < 14) {
                                                                            return 1;
                                                                        }
                                                                        else if (RsDiff >= 14) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                            else if (diffMVdiffVal >= 8) {
                                                                if (RsDiff < 16) {
                                                                    return 4;
                                                                }
                                                                else if (RsDiff >= 16) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (tcor >= 80) {
                                                    if (Cs < 2045) {
                                                        if (mu_mv_mag_sq < 24) {
                                                            if (RsCsDiff < 183) {
                                                                return 4;
                                                            }
                                                            else if (RsCsDiff >= 183) {
                                                                if (RsDiff < 16) {
                                                                    if (diffRsCsdiff < -35) {
                                                                        if (SC < 4035) {
                                                                            if (Rs < 2402) {
                                                                                return 8;
                                                                            }
                                                                            else if (Rs >= 2402) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                        else if (SC >= 4035) {
                                                                            return 2;
                                                                        }
                                                                    }
                                                                    else if (diffRsCsdiff >= -35) {
                                                                        return 2;
                                                                    }
                                                                }
                                                                else if (RsDiff >= 16) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                        else if (mu_mv_mag_sq >= 24) {
                                                            if (Rs < 4081) {
                                                                if (SC < 4347) {
                                                                    return 1;
                                                                }
                                                                else if (SC >= 4347) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (Rs >= 4081) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                    else if (Cs >= 2045) {
                                                        if (diffRsCsdiff < 78) {
                                                            if (MVDiff < 23) {
                                                                if (TSC < 12) {
                                                                    if (Rs < 3609) {
                                                                        if (Cs < 2138) {
                                                                            return 4;
                                                                        }
                                                                        else if (Cs >= 2138) {
                                                                            return 2;
                                                                        }
                                                                    }
                                                                    else if (Rs >= 3609) {
                                                                        if (Rs < 4292) {
                                                                            return 8;
                                                                        }
                                                                        else if (Rs >= 4292) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                }
                                                                else if (TSC >= 12) {
                                                                    if (posBalance < 59) {
                                                                        if (Rs < 2508) {
                                                                            return 4;
                                                                        }
                                                                        else if (Rs >= 2508) {
                                                                            return 2;
                                                                        }
                                                                    }
                                                                    else if (posBalance >= 59) {
                                                                        return 1;
                                                                    }
                                                                }
                                                            }
                                                            else if (MVDiff >= 23) {
                                                                if (RsDiff < 16) {
                                                                    if (CsDiff < 10) {
                                                                        return 8;
                                                                    }
                                                                    else if (CsDiff >= 10) {
                                                                        if (SC < 4597) {
                                                                            if (Rs < 2322) {
                                                                                return 4;
                                                                            }
                                                                            else if (Rs >= 2322) {
                                                                                return 1;
                                                                            }
                                                                        }
                                                                        else if (SC >= 4597) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                }
                                                                else if (RsDiff >= 16) {
                                                                    return 1;
                                                                }
                                                            }
                                                        }
                                                        else if (diffRsCsdiff >= 78) {
                                                            return 1;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (ssDCval >= 134) {
                                    if (negBalance < 56) {
                                        return 1;
                                    }
                                    else if (negBalance >= 56) {
                                        if (MVDiff < 23) {
                                            if (posBalance < 48) {
                                                return 8;
                                            }
                                            else if (posBalance >= 48) {
                                                return 1;
                                            }
                                        }
                                        else if (MVDiff >= 23) {
                                            if (posBalance < 57) {
                                                if (AFD < 20) {
                                                    return 4;
                                                }
                                                else if (AFD >= 20) {
                                                    return 1;
                                                }
                                            }
                                            else if (posBalance >= 57) {
                                                return 4;
                                            }
                                        }
                                    }
                                }
                            }
                            else if (ssDCval >= 140) {
                                if (Rs < 2032) {
                                    return 8;
                                }
                                else if (Rs >= 2032) {
                                    if (mcTcor < 93) {
                                        if (tcor < 79) {
                                            if (diffAFD < 3) {
                                                if (mu_mv_mag_sq < 5) {
                                                    if (negBalance < 54) {
                                                        return 2;
                                                    }
                                                    else if (negBalance >= 54) {
                                                        if (ssDCval < 159) {
                                                            return 4;
                                                        }
                                                        else if (ssDCval >= 159) {
                                                            return 2;
                                                        }
                                                    }
                                                }
                                                else if (mu_mv_mag_sq >= 5) {
                                                    return 4;
                                                }
                                            }
                                            else if (diffAFD >= 3) {
                                                if (Rs < 3196) {
                                                    return 8;
                                                }
                                                else if (Rs >= 3196) {
                                                    return 2;
                                                }
                                            }
                                        }
                                        else if (tcor >= 79) {
                                            return 2;
                                        }
                                    }
                                    else if (mcTcor >= 93) {
                                        if (posBalance < 58) {
                                            if (RsDiff < 21) {
                                                if (Cs < 2362) {
                                                    return 2;
                                                }
                                                else if (Cs >= 2362) {
                                                    return 8;
                                                }
                                            }
                                            else if (RsDiff >= 21) {
                                                return 4;
                                            }
                                        }
                                        else if (posBalance >= 58) {
                                            if (TSC < 11) {
                                                return 2;
                                            }
                                            else if (TSC >= 11) {
                                                if (Rs < 4933) {
                                                    return 2;
                                                }
                                                else if (Rs >= 4933) {
                                                    return 4;
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                        }
                    }
                }
                else if (mu_mv_mag_sq >= 42) {
                    if (ssDCval < 103) {
                        if (mcTcor < 90) {
                            if (Rs < 878) {
                                if (MVDiff < 24) {
                                    if (Rs < 730) {
                                        if (Cs < 385) {
                                            if (mcTcor < 80) {
                                                if (Cs < 372) {
                                                    if (mcTcor < 79) {
                                                        return 1;
                                                    }
                                                    else if (mcTcor >= 79) {
                                                        return 8;
                                                    }
                                                }
                                                else if (Cs >= 372) {
                                                    return 4;
                                                }
                                            }
                                            else if (mcTcor >= 80) {
                                                if (tcor < 86) {
                                                    return 4;
                                                }
                                                else if (tcor >= 86) {
                                                    if (refDCval < 38) {
                                                        return 2;
                                                    }
                                                    else if (refDCval >= 38) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                        }
                                        else if (Cs >= 385) {
                                            if (Rs < 448) {
                                                if (Cs < 477) {
                                                    return 2;
                                                }
                                                else if (Cs >= 477) {
                                                    return 8;
                                                }
                                            }
                                            else if (Rs >= 448) {
                                                if (MVDiff < 10) {
                                                    if (diffMVdiffVal < 3) {
                                                        return 4;
                                                    }
                                                    else if (diffMVdiffVal >= 3) {
                                                        return 8;
                                                    }
                                                }
                                                else if (MVDiff >= 10) {
                                                    if (diffMVdiffVal < -5) {
                                                        if (ssDCval < 39) {
                                                            return 2;
                                                        }
                                                        else if (ssDCval >= 39) {
                                                            if (SC < 1702) {
                                                                return 4;
                                                            }
                                                            else if (SC >= 1702) {
                                                                return 2;
                                                            }
                                                        }
                                                    }
                                                    else if (diffMVdiffVal >= -5) {
                                                        if (mu_mv_mag_sq < 69) {
                                                            return 2;
                                                        }
                                                        else if (mu_mv_mag_sq >= 69) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (Rs >= 730) {
                                        if (tcor < 75) {
                                            return 1;
                                        }
                                        else if (tcor >= 75) {
                                            return 4;
                                        }
                                    }
                                }
                                else if (MVDiff >= 24) {
                                    if (Cs < 382) {
                                        if (negBalance < 53) {
                                            if (mcTcor < 86) {
                                                if (diffAFD < 1) {
                                                    if (Cs < 318) {
                                                        if (TSC < 4) {
                                                            if (Cs < 1) {
                                                                if (ssDCval < 21) {
                                                                    if (MVDiff < 48) {
                                                                        return 8;
                                                                    }
                                                                    else if (MVDiff >= 48) {
                                                                        if (negBalance < 1) {
                                                                            return 4;
                                                                        }
                                                                        else if (negBalance >= 1) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                }
                                                                else if (ssDCval >= 21) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (Cs >= 1) {
                                                                return 1;
                                                            }
                                                        }
                                                        else if (TSC >= 4) {
                                                            if (tcor < 81) {
                                                                return 8;
                                                            }
                                                            else if (tcor >= 81) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                    else if (Cs >= 318) {
                                                        if (mu_mv_mag_sq < 63) {
                                                            if (Rs < 403) {
                                                                return 4;
                                                            }
                                                            else if (Rs >= 403) {
                                                                if (SC < 787) {
                                                                    return 1;
                                                                }
                                                                else if (SC >= 787) {
                                                                    if (MVDiff < 34) {
                                                                        return 4;
                                                                    }
                                                                    else if (MVDiff >= 34) {
                                                                        return 1;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (mu_mv_mag_sq >= 63) {
                                                            return 1;
                                                        }
                                                    }
                                                }
                                                else if (diffAFD >= 1) {
                                                    if (ssDCval < 49) {
                                                        return 1;
                                                    }
                                                    else if (ssDCval >= 49) {
                                                        return 2;
                                                    }
                                                }
                                            }
                                            else if (mcTcor >= 86) {
                                                return 4;
                                            }
                                        }
                                        else if (negBalance >= 53) {
                                            if (TSC < 4) {
                                                if (refDCval < 20) {
                                                    return 4;
                                                }
                                                else if (refDCval >= 20) {
                                                    return 2;
                                                }
                                            }
                                            else if (TSC >= 4) {
                                                if (mu_mv_mag_sq < 72) {
                                                    return 1;
                                                }
                                                else if (mu_mv_mag_sq >= 72) {
                                                    return 2;
                                                }
                                            }
                                        }
                                    }
                                    else if (Cs >= 382) {
                                        if (Cs < 503) {
                                            if (ssDCval < 39) {
                                                return 2;
                                            }
                                            else if (ssDCval >= 39) {
                                                if (CsDiff < 2) {
                                                    return 4;
                                                }
                                                else if (CsDiff >= 2) {
                                                    if (Rs < 506) {
                                                        return 1;
                                                    }
                                                    else if (Rs >= 506) {
                                                        return 2;
                                                    }
                                                }
                                            }
                                        }
                                        else if (Cs >= 503) {
                                            if (Cs < 668) {
                                                if (negBalance < 47) {
                                                    if (SC < 1165) {
                                                        return 2;
                                                    }
                                                    else if (SC >= 1165) {
                                                        return 4;
                                                    }
                                                }
                                                else if (negBalance >= 47) {
                                                    if (mu_mv_mag_sq < 45) {
                                                        return 1;
                                                    }
                                                    else if (mu_mv_mag_sq >= 45) {
                                                        if (diffRsCsdiff < -4) {
                                                            return 1;
                                                        }
                                                        else if (diffRsCsdiff >= -4) {
                                                            if (mcTcor < 82) {
                                                                if (negBalance < 49) {
                                                                    return 4;
                                                                }
                                                                else if (negBalance >= 49) {
                                                                    if (Cs < 618) {
                                                                        if (TSCindex < 7) {
                                                                            return 4;
                                                                        }
                                                                        else if (TSCindex >= 7) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                    else if (Cs >= 618) {
                                                                        return 4;
                                                                    }
                                                                }
                                                            }
                                                            else if (mcTcor >= 82) {
                                                                return 1;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                            else if (Cs >= 668) {
                                                if (refDCval < 75) {
                                                    if (Cs < 1027) {
                                                        return 1;
                                                    }
                                                    else if (Cs >= 1027) {
                                                        return 2;
                                                    }
                                                }
                                                else if (refDCval >= 75) {
                                                    return 1;
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                            else if (Rs >= 878) {
                                if (refDCval < 93) {
                                    if (AFD < 13) {
                                        if (Rs < 2548) {
                                            if (SC < 3474) {
                                                if (diffRsCsdiff < -7) {
                                                    if (diffAFD < -3) {
                                                        if (Cs < 1128) {
                                                            return 4;
                                                        }
                                                        else if (Cs >= 1128) {
                                                            return 1;
                                                        }
                                                    }
                                                    else if (diffAFD >= -3) {
                                                        if (mcTcor < 82) {
                                                            if (Rs < 1291) {
                                                                return 4;
                                                            }
                                                            else if (Rs >= 1291) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (mcTcor >= 82) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                                else if (diffRsCsdiff >= -7) {
                                                    if (MVDiff < 11) {
                                                        if (SC < 2510) {
                                                            if (Cs < 814) {
                                                                if (RsCsDiff < 117) {
                                                                    return 8;
                                                                }
                                                                else if (RsCsDiff >= 117) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (Cs >= 814) {
                                                                if (AFD < 6) {
                                                                    return 1;
                                                                }
                                                                else if (AFD >= 6) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                        else if (SC >= 2510) {
                                                            if (diffMVdiffVal < -2) {
                                                                return 1;
                                                            }
                                                            else if (diffMVdiffVal >= -2) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                    else if (MVDiff >= 11) {
                                                        if (ssDCval < 39) {
                                                            if (AFD < 5) {
                                                                return 4;
                                                            }
                                                            else if (AFD >= 5) {
                                                                return 2;
                                                            }
                                                        }
                                                        else if (ssDCval >= 39) {
                                                            if (posBalance < 40) {
                                                                return 2;
                                                            }
                                                            else if (posBalance >= 40) {
                                                                if (AFD < 12) {
                                                                    if (gchDC < 3) {
                                                                        return 4;
                                                                    }
                                                                    else if (gchDC >= 3) {
                                                                        if (SC < 1708) {
                                                                            return 4;
                                                                        }
                                                                        else if (SC >= 1708) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                }
                                                                else if (AFD >= 12) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                            else if (SC >= 3474) {
                                                return 8;
                                            }
                                        }
                                        else if (Rs >= 2548) {
                                            if (diffTSC < 2) {
                                                if (mu_mv_mag_sq < 57) {
                                                    if (CsDiff < 7) {
                                                        return 4;
                                                    }
                                                    else if (CsDiff >= 7) {
                                                        return 2;
                                                    }
                                                }
                                                else if (mu_mv_mag_sq >= 57) {
                                                    return 4;
                                                }
                                            }
                                            else if (diffTSC >= 2) {
                                                return 2;
                                            }
                                        }
                                    }
                                    else if (AFD >= 13) {
                                        if (CsDiff < 6) {
                                            if (Rs < 2086) {
                                                if (SC < 2695) {
                                                    return 4;
                                                }
                                                else if (SC >= 2695) {
                                                    return 8;
                                                }
                                            }
                                            else if (Rs >= 2086) {
                                                return 1;
                                            }
                                        }
                                        else if (CsDiff >= 6) {
                                            if (refDCval < 62) {
                                                if (SC < 4287) {
                                                    return 4;
                                                }
                                                else if (SC >= 4287) {
                                                    return 8;
                                                }
                                            }
                                            else if (refDCval >= 62) {
                                                if (negBalance < 45) {
                                                    return 1;
                                                }
                                                else if (negBalance >= 45) {
                                                    if (tcor < 83) {
                                                        if (refDCval < 84) {
                                                            if (RsDiff < 10) {
                                                                if (Rs < 2089) {
                                                                    return 4;
                                                                }
                                                                else if (Rs >= 2089) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (RsDiff >= 10) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (refDCval >= 84) {
                                                            return 1;
                                                        }
                                                    }
                                                    else if (tcor >= 83) {
                                                        if (posBalance < 58) {
                                                            return 8;
                                                        }
                                                        else if (posBalance >= 58) {
                                                            return 1;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (refDCval >= 93) {
                                    if (Cs < 2581) {
                                        if (Cs < 1270) {
                                            if (mu_mv_mag_sq < 80) {
                                                return 4;
                                            }
                                            else if (mu_mv_mag_sq >= 80) {
                                                if (AFD < 33) {
                                                    if (TSC < 16) {
                                                        return 1;
                                                    }
                                                    else if (TSC >= 16) {
                                                        return 2;
                                                    }
                                                }
                                                else if (AFD >= 33) {
                                                    return 2;
                                                }
                                            }
                                        }
                                        else if (Cs >= 1270) {
                                            if (RsCsDiff < 51) {
                                                if (posBalance < 59) {
                                                    return 8;
                                                }
                                                else if (posBalance >= 59) {
                                                    if (tcor < 80) {
                                                        return 8;
                                                    }
                                                    else if (tcor >= 80) {
                                                        if (Rs < 1691) {
                                                            return 8;
                                                        }
                                                        else if (Rs >= 1691) {
                                                            return 2;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (RsCsDiff >= 51) {
                                                if (Cs < 2285) {
                                                    if (TSC < 20) {
                                                        if (Rs < 1821) {
                                                            return 8;
                                                        }
                                                        else if (Rs >= 1821) {
                                                            return 2;
                                                        }
                                                    }
                                                    else if (TSC >= 20) {
                                                        return 4;
                                                    }
                                                }
                                                else if (Cs >= 2285) {
                                                    return 8;
                                                }
                                            }
                                        }
                                    }
                                    else if (Cs >= 2581) {
                                        if (diffTSC < 2) {
                                            return 4;
                                        }
                                        else if (diffTSC >= 2) {
                                            return 8;
                                        }
                                    }
                                }
                            }
                        }
                        else if (mcTcor >= 90) {
                            if (mu_mv_mag_sq < 75) {
                                if (posBalance < 43) {
                                    if (Rs < 2319) {
                                        if (negBalance < 29) {
                                            if (Rs < 409) {
                                                if (Cs < 16) {
                                                    return 4;
                                                }
                                                else if (Cs >= 16) {
                                                    return 8;
                                                }
                                            }
                                            else if (Rs >= 409) {
                                                return 1;
                                            }
                                        }
                                        else if (negBalance >= 29) {
                                            if (posBalance < 8) {
                                                if (negBalance < 114) {
                                                    if (mu_mv_mag_sq < 72) {
                                                        if (diffTSC < -2) {
                                                            return 2;
                                                        }
                                                        else if (diffTSC >= -2) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (mu_mv_mag_sq >= 72) {
                                                        return 2;
                                                    }
                                                }
                                                else if (negBalance >= 114) {
                                                    return 1;
                                                }
                                            }
                                            else if (posBalance >= 8) {
                                                if (refDCval < 75) {
                                                    if (ssDCval < 49) {
                                                        if (mu_mv_mag_sq < 56) {
                                                            if (Cs < 35) {
                                                                return 4;
                                                            }
                                                            else if (Cs >= 35) {
                                                                return 2;
                                                            }
                                                        }
                                                        else if (mu_mv_mag_sq >= 56) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (ssDCval >= 49) {
                                                        if (Cs < 721) {
                                                            if (Cs < 388) {
                                                                return 4;
                                                            }
                                                            else if (Cs >= 388) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (Cs >= 721) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                                else if (refDCval >= 75) {
                                                    if (mcTcor < 101) {
                                                        return 8;
                                                    }
                                                    else if (mcTcor >= 101) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (Rs >= 2319) {
                                        if (posBalance < 24) {
                                            return 8;
                                        }
                                        else if (posBalance >= 24) {
                                            if (RsCsDiff < 4) {
                                                return 4;
                                            }
                                            else if (RsCsDiff >= 4) {
                                                if (Rs < 3156) {
                                                    if (negBalance < 32) {
                                                        return 4;
                                                    }
                                                    else if (negBalance >= 32) {
                                                        return 2;
                                                    }
                                                }
                                                else if (Rs >= 3156) {
                                                    return 4;
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (posBalance >= 43) {
                                    if (posBalance < 51) {
                                        if (Cs < 1469) {
                                            if (refDCval < 53) {
                                                if (Rs < 1221) {
                                                    if (diffAFD < -1) {
                                                        if (Rs < 976) {
                                                            if (Cs < 908) {
                                                                return 1;
                                                            }
                                                            else if (Cs >= 908) {
                                                                return 2;
                                                            }
                                                        }
                                                        else if (Rs >= 976) {
                                                            if (Cs < 744) {
                                                                return 4;
                                                            }
                                                            else if (Cs >= 744) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                    else if (diffAFD >= -1) {
                                                        return 4;
                                                    }
                                                }
                                                else if (Rs >= 1221) {
                                                    if (mcTcor < 93) {
                                                        if (refDCval < 51) {
                                                            return 1;
                                                        }
                                                        else if (refDCval >= 51) {
                                                            if (posBalance < 49) {
                                                                return 4;
                                                            }
                                                            else if (posBalance >= 49) {
                                                                return 1;
                                                            }
                                                        }
                                                    }
                                                    else if (mcTcor >= 93) {
                                                        return 1;
                                                    }
                                                }
                                            }
                                            else if (refDCval >= 53) {
                                                if (RsDiff < 3) {
                                                    if (ssDCval < 56) {
                                                        if (ssDCval < 53) {
                                                            return 8;
                                                        }
                                                        else if (ssDCval >= 53) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (ssDCval >= 56) {
                                                        return 1;
                                                    }
                                                }
                                                else if (RsDiff >= 3) {
                                                    if (Cs < 697) {
                                                        if (Rs < 1403) {
                                                            return 4;
                                                        }
                                                        else if (Rs >= 1403) {
                                                            return 2;
                                                        }
                                                    }
                                                    else if (Cs >= 697) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                        }
                                        else if (Cs >= 1469) {
                                            if (SC < 4165) {
                                                return 8;
                                            }
                                            else if (SC >= 4165) {
                                                if (gchDC < 1) {
                                                    if (diffTSC < -1) {
                                                        if (diffMVdiffVal < -2) {
                                                            return 1;
                                                        }
                                                        else if (diffMVdiffVal >= -2) {
                                                            if (Rs < 2180) {
                                                                return 1;
                                                            }
                                                            else if (Rs >= 2180) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                    else if (diffTSC >= -1) {
                                                        if (Rs < 2028) {
                                                            return 1;
                                                        }
                                                        else if (Rs >= 2028) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                                else if (gchDC >= 1) {
                                                    if (Rs < 2948) {
                                                        if (RsCsDiff < 73) {
                                                            return 4;
                                                        }
                                                        else if (RsCsDiff >= 73) {
                                                            if (diffAFD < -2) {
                                                                return 4;
                                                            }
                                                            else if (diffAFD >= -2) {
                                                                return 1;
                                                            }
                                                        }
                                                    }
                                                    else if (Rs >= 2948) {
                                                        if (SC < 6794) {
                                                            return 8;
                                                        }
                                                        else if (SC >= 6794) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (posBalance >= 51) {
                                        if (tcor < 72) {
                                            if (posBalance < 75) {
                                                return 2;
                                            }
                                            else if (posBalance >= 75) {
                                                return 8;
                                            }
                                        }
                                        else if (tcor >= 72) {
                                            if (TSCindex < 9) {
                                                if (ssDCval < 75) {
                                                    if (tcor < 93) {
                                                        if (refDCval < 39) {
                                                            if (Rs < 1788) {
                                                                return 2;
                                                            }
                                                            else if (Rs >= 1788) {
                                                                return 1;
                                                            }
                                                        }
                                                        else if (refDCval >= 39) {
                                                            if (Rs < 1133) {
                                                                if (TSCindex < 6) {
                                                                    if (diffMVdiffVal < -22) {
                                                                        return 8;
                                                                    }
                                                                    else if (diffMVdiffVal >= -22) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (TSCindex >= 6) {
                                                                    if (MVDiff < 14) {
                                                                        return 2;
                                                                    }
                                                                    else if (MVDiff >= 14) {
                                                                        if (AFD < 7) {
                                                                            return 1;
                                                                        }
                                                                        else if (AFD >= 7) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                            else if (Rs >= 1133) {
                                                                if (SC < 2519) {
                                                                    if (mcTcor < 92) {
                                                                        if (Cs < 953) {
                                                                            if (Cs < 920) {
                                                                                return 1;
                                                                            }
                                                                            else if (Cs >= 920) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                        else if (Cs >= 953) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                    else if (mcTcor >= 92) {
                                                                        if (Cs < 1025) {
                                                                            return 1;
                                                                        }
                                                                        else if (Cs >= 1025) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                }
                                                                else if (SC >= 2519) {
                                                                    if (Cs < 1525) {
                                                                        if (MVDiff < 16) {
                                                                            if (SC < 2967) {
                                                                                return 4;
                                                                            }
                                                                            else if (SC >= 2967) {
                                                                                if (negBalance < 44) {
                                                                                    return 4;
                                                                                }
                                                                                else if (negBalance >= 44) {
                                                                                    return 1;
                                                                                }
                                                                            }
                                                                        }
                                                                        else if (MVDiff >= 16) {
                                                                            if (Rs < 1716) {
                                                                                return 8;
                                                                            }
                                                                            else if (Rs >= 1716) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (Cs >= 1525) {
                                                                        if (diffMVdiffVal < -2) {
                                                                            return 1;
                                                                        }
                                                                        else if (diffMVdiffVal >= -2) {
                                                                            if (Cs < 2388) {
                                                                                if (refDCval < 40) {
                                                                                    if (MVDiff < 22) {
                                                                                        if (TSC < 5) {
                                                                                            if (diffMVdiffVal < 2) {
                                                                                                return 1;
                                                                                            }
                                                                                            else if (diffMVdiffVal >= 2) {
                                                                                                return 4;
                                                                                            }
                                                                                        }
                                                                                        else if (TSC >= 5) {
                                                                                            return 4;
                                                                                        }
                                                                                    }
                                                                                    else if (MVDiff >= 22) {
                                                                                        if (Cs < 1545) {
                                                                                            if (Rs < 1834) {
                                                                                                return 1;
                                                                                            }
                                                                                            else if (Rs >= 1834) {
                                                                                                return 4;
                                                                                            }
                                                                                        }
                                                                                        else if (Cs >= 1545) {
                                                                                            return 1;
                                                                                        }
                                                                                    }
                                                                                }
                                                                                else if (refDCval >= 40) {
                                                                                    return 4;
                                                                                }
                                                                            }
                                                                            else if (Cs >= 2388) {
                                                                                return 1;
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (tcor >= 93) {
                                                        if (Rs < 2145) {
                                                            if (refDCval < 43) {
                                                                if (Cs < 1544) {
                                                                    if (Cs < 1097) {
                                                                        return 8;
                                                                    }
                                                                    else if (Cs >= 1097) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (Cs >= 1544) {
                                                                    return 1;
                                                                }
                                                            }
                                                            else if (refDCval >= 43) {
                                                                if (Rs < 904) {
                                                                    return 2;
                                                                }
                                                                else if (Rs >= 904) {
                                                                    return 1;
                                                                }
                                                            }
                                                        }
                                                        else if (Rs >= 2145) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                                else if (ssDCval >= 75) {
                                                    if (diffTSC < -2) {
                                                        if (SC < 2738) {
                                                            return 1;
                                                        }
                                                        else if (SC >= 2738) {
                                                            return 2;
                                                        }
                                                    }
                                                    else if (diffTSC >= -2) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                            else if (TSCindex >= 9) {
                                                if (Rs < 2596) {
                                                    if (Cs < 1876) {
                                                        if (SC < 3817) {
                                                            if (CsDiff < 5) {
                                                                return 1;
                                                            }
                                                            else if (CsDiff >= 5) {
                                                                if (Cs < 1378) {
                                                                    return 4;
                                                                }
                                                                else if (Cs >= 1378) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                        else if (SC >= 3817) {
                                                            if (diffMVdiffVal < -2) {
                                                                if (Cs < 1803) {
                                                                    return 2;
                                                                }
                                                                else if (Cs >= 1803) {
                                                                    return 1;
                                                                }
                                                            }
                                                            else if (diffMVdiffVal >= -2) {
                                                                if (gchDC < 2) {
                                                                    return 1;
                                                                }
                                                                else if (gchDC >= 2) {
                                                                    return 2;
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (Cs >= 1876) {
                                                        if (refDCval < 81) {
                                                            if (AFD < 15) {
                                                                if (TSC < 10) {
                                                                    return 8;
                                                                }
                                                                else if (TSC >= 10) {
                                                                    return 1;
                                                                }
                                                            }
                                                            else if (AFD >= 15) {
                                                                return 1;
                                                            }
                                                        }
                                                        else if (refDCval >= 81) {
                                                            if (refDCval < 86) {
                                                                return 4;
                                                            }
                                                            else if (refDCval >= 86) {
                                                                if (mcTcor < 94) {
                                                                    if (mu_mv_mag_sq < 65) {
                                                                        if (MVDiff < 16) {
                                                                            return 2;
                                                                        }
                                                                        else if (MVDiff >= 16) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                    else if (mu_mv_mag_sq >= 65) {
                                                                        return 1;
                                                                    }
                                                                }
                                                                else if (mcTcor >= 94) {
                                                                    if (TSC < 9) {
                                                                        return 4;
                                                                    }
                                                                    else if (TSC >= 9) {
                                                                        return 2;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (Rs >= 2596) {
                                                    if (SC < 4893) {
                                                        if (tcor < 88) {
                                                            if (Cs < 2222) {
                                                                return 2;
                                                            }
                                                            else if (Cs >= 2222) {
                                                                return 1;
                                                            }
                                                        }
                                                        else if (tcor >= 88) {
                                                            return 1;
                                                        }
                                                    }
                                                    else if (SC >= 4893) {
                                                        if (negBalance < 39) {
                                                            if (Rs < 3795) {
                                                                return 2;
                                                            }
                                                            else if (Rs >= 3795) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (negBalance >= 39) {
                                                            if (Rs < 2941) {
                                                                if (negBalance < 48) {
                                                                    return 1;
                                                                }
                                                                else if (negBalance >= 48) {
                                                                    if (SC < 5235) {
                                                                        if (TSC < 11) {
                                                                            return 1;
                                                                        }
                                                                        else if (TSC >= 11) {
                                                                            if (mu_mv_mag_sq < 57) {
                                                                                return 1;
                                                                            }
                                                                            else if (mu_mv_mag_sq >= 57) {
                                                                                if (Cs < 2132) {
                                                                                    return 1;
                                                                                }
                                                                                else if (Cs >= 2132) {
                                                                                    return 4;
                                                                                }
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (SC >= 5235) {
                                                                        return 4;
                                                                    }
                                                                }
                                                            }
                                                            else if (Rs >= 2941) {
                                                                if (ssDCval < 87) {
                                                                    if (TSC < 11) {
                                                                        if (SC < 6171) {
                                                                            return 8;
                                                                        }
                                                                        else if (SC >= 6171) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                    else if (TSC >= 11) {
                                                                        if (diffAFD < 3) {
                                                                            return 1;
                                                                        }
                                                                        else if (diffAFD >= 3) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                }
                                                                else if (ssDCval >= 87) {
                                                                    if (SC < 5380) {
                                                                        if (SC < 5342) {
                                                                            return 1;
                                                                        }
                                                                        else if (SC >= 5342) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                    else if (SC >= 5380) {
                                                                        return 1;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                            else if (mu_mv_mag_sq >= 75) {
                                if (mu_mv_mag_sq < 92) {
                                    if (ssDCval < 44) {
                                        if (ssDCval < 23) {
                                            return 8;
                                        }
                                        else if (ssDCval >= 23) {
                                            if (SC < 3340) {
                                                if (tcor < 97) {
                                                    if (MVDiff < 15) {
                                                        return 8;
                                                    }
                                                    else if (MVDiff >= 15) {
                                                        if (tcor < 92) {
                                                            return 4;
                                                        }
                                                        else if (tcor >= 92) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                                else if (tcor >= 97) {
                                                    return 4;
                                                }
                                            }
                                            else if (SC >= 3340) {
                                                if (posBalance < 53) {
                                                    if (Rs < 2043) {
                                                        return 1;
                                                    }
                                                    else if (Rs >= 2043) {
                                                        return 4;
                                                    }
                                                }
                                                else if (posBalance >= 53) {
                                                    return 4;
                                                }
                                            }
                                        }
                                    }
                                    else if (ssDCval >= 44) {
                                        if (negBalance < 50) {
                                            if (ssDCval < 56) {
                                                if (Cs < 622) {
                                                    return 8;
                                                }
                                                else if (Cs >= 622) {
                                                    if (CsDiff < 4) {
                                                        return 1;
                                                    }
                                                    else if (CsDiff >= 4) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                            else if (ssDCval >= 56) {
                                                if (ssDCval < 101) {
                                                    if (TSC < 7) {
                                                        if (Rs < 1615) {
                                                            if (Cs < 668) {
                                                                return 8;
                                                            }
                                                            else if (Cs >= 668) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (Rs >= 1615) {
                                                            if (MVDiff < 16) {
                                                                return 8;
                                                            }
                                                            else if (MVDiff >= 16) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                    else if (TSC >= 7) {
                                                        if (CsDiff < 6) {
                                                            if (SC < 4132) {
                                                                return 8;
                                                            }
                                                            else if (SC >= 4132) {
                                                                return 1;
                                                            }
                                                        }
                                                        else if (CsDiff >= 6) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                                else if (ssDCval >= 101) {
                                                    if (Rs < 4052) {
                                                        return 4;
                                                    }
                                                    else if (Rs >= 4052) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                        }
                                        else if (negBalance >= 50) {
                                            if (tcor < 76) {
                                                return 2;
                                            }
                                            else if (tcor >= 76) {
                                                if (CsDiff < 1) {
                                                    return 8;
                                                }
                                                else if (CsDiff >= 1) {
                                                    if (diffRsCsdiff < 21) {
                                                        if (mcTcor < 92) {
                                                            if (SC < 5240) {
                                                                return 2;
                                                            }
                                                            else if (SC >= 5240) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (mcTcor >= 92) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (diffRsCsdiff >= 21) {
                                                        if (Rs < 1865) {
                                                            return 8;
                                                        }
                                                        else if (Rs >= 1865) {
                                                            return 2;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (mu_mv_mag_sq >= 92) {
                                    if (AFD < 16) {
                                        if (mu_mv_mag_sq < 99) {
                                            if (CsDiff < 1) {
                                                return 4;
                                            }
                                            else if (CsDiff >= 1) {
                                                if (RsCsDiff < 80) {
                                                    if (RsCsDiff < 12) {
                                                        if (MVDiff < 10) {
                                                            if (Cs < 1225) {
                                                                return 8;
                                                            }
                                                            else if (Cs >= 1225) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (MVDiff >= 10) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (RsCsDiff >= 12) {
                                                        return 8;
                                                    }
                                                }
                                                else if (RsCsDiff >= 80) {
                                                    return 4;
                                                }
                                            }
                                        }
                                        else if (mu_mv_mag_sq >= 99) {
                                            if (mcTcor < 100) {
                                                if (diffRsCsdiff < -131) {
                                                    if (Rs < 1880) {
                                                        if (Rs < 875) {
                                                            return 8;
                                                        }
                                                        else if (Rs >= 875) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (Rs >= 1880) {
                                                        return 8;
                                                    }
                                                }
                                                else if (diffRsCsdiff >= -131) {
                                                    if (tcor < 92) {
                                                        if (negBalance < 24) {
                                                            return 1;
                                                        }
                                                        else if (negBalance >= 24) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (tcor >= 92) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                            else if (mcTcor >= 100) {
                                                if (Rs < 1054) {
                                                    if (Rs < 813) {
                                                        return 1;
                                                    }
                                                    else if (Rs >= 813) {
                                                        return 4;
                                                    }
                                                }
                                                else if (Rs >= 1054) {
                                                    if (ssDCval < 39) {
                                                        return 8;
                                                    }
                                                    else if (ssDCval >= 39) {
                                                        return 1;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (AFD >= 16) {
                                        return 1;
                                    }
                                }
                            }
                        }
                    }
                    else if (ssDCval >= 103) {
                        if (mcTcor < 85) {
                            if (SC < 7757) {
                                if (ssDCval < 122) {
                                    if (TSC < 14) {
                                        if (ssDCval < 108) {
                                            if (posBalance < 46) {
                                                return 8;
                                            }
                                            else if (posBalance >= 46) {
                                                return 4;
                                            }
                                        }
                                        else if (ssDCval >= 108) {
                                            if (negBalance < 44) {
                                                return 1;
                                            }
                                            else if (negBalance >= 44) {
                                                if (diffMVdiffVal < 2) {
                                                    return 4;
                                                }
                                                else if (diffMVdiffVal >= 2) {
                                                    if (AFD < 11) {
                                                        return 8;
                                                    }
                                                    else if (AFD >= 11) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (TSC >= 14) {
                                        if (AFD < 40) {
                                            return 2;
                                        }
                                        else if (AFD >= 40) {
                                            if (SC < 2047) {
                                                return 8;
                                            }
                                            else if (SC >= 2047) {
                                                return 4;
                                            }
                                        }
                                    }
                                }
                                else if (ssDCval >= 122) {
                                    if (posBalance < 49) {
                                        if (refDCval < 137) {
                                            if (Rs < 1344) {
                                                return 4;
                                            }
                                            else if (Rs >= 1344) {
                                                if (MVDiff < 100) {
                                                    return 1;
                                                }
                                                else if (MVDiff >= 100) {
                                                    return 8;
                                                }
                                            }
                                        }
                                        else if (refDCval >= 137) {
                                            if (Rs < 449) {
                                                return 1;
                                            }
                                            else if (Rs >= 449) {
                                                return 8;
                                            }
                                        }
                                    }
                                    else if (posBalance >= 49) {
                                        if (diffAFD < 29) {
                                            if (negBalance < 53) {
                                                if (tcor < 81) {
                                                    return 1;
                                                }
                                                else if (tcor >= 81) {
                                                    if (Cs < 2141) {
                                                        return 1;
                                                    }
                                                    else if (Cs >= 2141) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                            else if (negBalance >= 53) {
                                                if (diffAFD < 1) {
                                                    return 4;
                                                }
                                                else if (diffAFD >= 1) {
                                                    if (AFD < 20) {
                                                        return 4;
                                                    }
                                                    else if (AFD >= 20) {
                                                        return 1;
                                                    }
                                                }
                                            }
                                        }
                                        else if (diffAFD >= 29) {
                                            if (RsDiff < 26) {
                                                return 4;
                                            }
                                            else if (RsDiff >= 26) {
                                                return 1;
                                            }
                                        }
                                    }
                                }
                            }
                            else if (SC >= 7757) {
                                if (ssDCval < 106) {
                                    return 4;
                                }
                                else if (ssDCval >= 106) {
                                    if (RsDiff < 10) {
                                        if (MVDiff < 10) {
                                            return 2;
                                        }
                                        else if (MVDiff >= 10) {
                                            return 8;
                                        }
                                    }
                                    else if (RsDiff >= 10) {
                                        return 2;
                                    }
                                }
                            }
                        }
                        else if (mcTcor >= 85) {
                            if (RsDiff < 6) {
                                if (ssDCval < 109) {
                                    if (diffAFD < -9) {
                                        if (SC < 6203) {
                                            return 4;
                                        }
                                        else if (SC >= 6203) {
                                            return 2;
                                        }
                                    }
                                    else if (diffAFD >= -9) {
                                        if (mu_mv_mag_sq < 48) {
                                            if (posBalance < 48) {
                                                return 4;
                                            }
                                            else if (posBalance >= 48) {
                                                return 8;
                                            }
                                        }
                                        else if (mu_mv_mag_sq >= 48) {
                                            if (diffMVdiffVal < 23) {
                                                return 8;
                                            }
                                            else if (diffMVdiffVal >= 23) {
                                                return 4;
                                            }
                                        }
                                    }
                                }
                                else if (ssDCval >= 109) {
                                    if (SC < 2346) {
                                        if (posBalance < 19) {
                                            if (Rs < 824) {
                                                if (refDCval < 156) {
                                                    return 8;
                                                }
                                                else if (refDCval >= 156) {
                                                    return 4;
                                                }
                                            }
                                            else if (Rs >= 824) {
                                                if (ssDCval < 140) {
                                                    if (Rs < 972) {
                                                        return 8;
                                                    }
                                                    else if (Rs >= 972) {
                                                        return 1;
                                                    }
                                                }
                                                else if (ssDCval >= 140) {
                                                    return 1;
                                                }
                                            }
                                        }
                                        else if (posBalance >= 19) {
                                            if (tcor < 97) {
                                                if (RsCsDiff < 7) {
                                                    if (SC < 211) {
                                                        return 2;
                                                    }
                                                    else if (SC >= 211) {
                                                        return 4;
                                                    }
                                                }
                                                else if (RsCsDiff >= 7) {
                                                    if (diffAFD < -2) {
                                                        if (Cs < 590) {
                                                            return 2;
                                                        }
                                                        else if (Cs >= 590) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (diffAFD >= -2) {
                                                        if (Cs < 392) {
                                                            if (Cs < 308) {
                                                                return 1;
                                                            }
                                                            else if (Cs >= 308) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (Cs >= 392) {
                                                            if (TSC < 5) {
                                                                return 1;
                                                            }
                                                            else if (TSC >= 5) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                            else if (tcor >= 97) {
                                                if (Cs < 472) {
                                                    if (mu_mv_mag_sq < 66) {
                                                        return 1;
                                                    }
                                                    else if (mu_mv_mag_sq >= 66) {
                                                        if (AFD < 2) {
                                                            return 4;
                                                        }
                                                        else if (AFD >= 2) {
                                                            if (CsDiff < 1) {
                                                                return 2;
                                                            }
                                                            else if (CsDiff >= 1) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (Cs >= 472) {
                                                    return 2;
                                                }
                                            }
                                        }
                                    }
                                    else if (SC >= 2346) {
                                        if (SC < 3978) {
                                            if (MVDiff < 23) {
                                                if (posBalance < 17) {
                                                    return 4;
                                                }
                                                else if (posBalance >= 17) {
                                                    return 8;
                                                }
                                            }
                                            else if (MVDiff >= 23) {
                                                return 4;
                                            }
                                        }
                                        else if (SC >= 3978) {
                                            if (Cs < 2294) {
                                                return 4;
                                            }
                                            else if (Cs >= 2294) {
                                                if (ssDCval < 117) {
                                                    return 2;
                                                }
                                                else if (ssDCval >= 117) {
                                                    if (Rs < 4016) {
                                                        return 8;
                                                    }
                                                    else if (Rs >= 4016) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                            else if (RsDiff >= 6) {
                                if (TSC < 10) {
                                    if (ssDCval < 128) {
                                        if (refDCval < 107) {
                                            if (Rs < 3479) {
                                                if (AFD < 9) {
                                                    if (Cs < 1505) {
                                                        return 4;
                                                    }
                                                    else if (Cs >= 1505) {
                                                        return 8;
                                                    }
                                                }
                                                else if (AFD >= 9) {
                                                    if (SC < 5331) {
                                                        return 2;
                                                    }
                                                    else if (SC >= 5331) {
                                                        if (diffRsCsdiff < -33) {
                                                            return 2;
                                                        }
                                                        else if (diffRsCsdiff >= -33) {
                                                            if (TSC < 9) {
                                                                return 4;
                                                            }
                                                            else if (TSC >= 9) {
                                                                return 2;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                            else if (Rs >= 3479) {
                                                if (SC < 6113) {
                                                    if (gchDC < 1) {
                                                        if (AFD < 8) {
                                                            return 8;
                                                        }
                                                        else if (AFD >= 8) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (gchDC >= 1) {
                                                        return 4;
                                                    }
                                                }
                                                else if (SC >= 6113) {
                                                    if (Rs < 4720) {
                                                        return 8;
                                                    }
                                                    else if (Rs >= 4720) {
                                                        return 2;
                                                    }
                                                }
                                            }
                                        }
                                        else if (refDCval >= 107) {
                                            if (Cs < 2069) {
                                                if (gchDC < 2) {
                                                    if (Rs < 3587) {
                                                        if (negBalance < 51) {
                                                            if (posBalance < 49) {
                                                                if (Rs < 3565) {
                                                                    return 4;
                                                                }
                                                                else if (Rs >= 3565) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (posBalance >= 49) {
                                                                if (ssDCval < 118) {
                                                                    if (Cs < 1585) {
                                                                        if (Rs < 2246) {
                                                                            return 8;
                                                                        }
                                                                        else if (Rs >= 2246) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                    else if (Cs >= 1585) {
                                                                        return 8;
                                                                    }
                                                                }
                                                                else if (ssDCval >= 118) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                        else if (negBalance >= 51) {
                                                            if (Cs < 1719) {
                                                                if (Cs < 1556) {
                                                                    return 4;
                                                                }
                                                                else if (Cs >= 1556) {
                                                                    if (Rs < 2831) {
                                                                        return 8;
                                                                    }
                                                                    else if (Rs >= 2831) {
                                                                        return 4;
                                                                    }
                                                                }
                                                            }
                                                            else if (Cs >= 1719) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                    else if (Rs >= 3587) {
                                                        return 4;
                                                    }
                                                }
                                                else if (gchDC >= 2) {
                                                    if (SC < 5425) {
                                                        if (refDCval < 118) {
                                                            return 2;
                                                        }
                                                        else if (refDCval >= 118) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (SC >= 5425) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                            else if (Cs >= 2069) {
                                                if (ssDCval < 119) {
                                                    if (RsDiff < 10) {
                                                        if (mcTcor < 95) {
                                                            if (refDCval < 110) {
                                                                if (posBalance < 42) {
                                                                    return 8;
                                                                }
                                                                else if (posBalance >= 42) {
                                                                    if (diffMVdiffVal < 2) {
                                                                        return 4;
                                                                    }
                                                                    else if (diffMVdiffVal >= 2) {
                                                                        return 8;
                                                                    }
                                                                }
                                                            }
                                                            else if (refDCval >= 110) {
                                                                if (refDCval < 112) {
                                                                    if (RsDiff < 8) {
                                                                        if (Cs < 2899) {
                                                                            return 8;
                                                                        }
                                                                        else if (Cs >= 2899) {
                                                                            return 2;
                                                                        }
                                                                    }
                                                                    else if (RsDiff >= 8) {
                                                                        if (Rs < 4339) {
                                                                            return 4;
                                                                        }
                                                                        else if (Rs >= 4339) {
                                                                            if (AFD < 10) {
                                                                                return 2;
                                                                            }
                                                                            else if (AFD >= 10) {
                                                                                return 8;
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                                else if (refDCval >= 112) {
                                                                    if (AFD < 11) {
                                                                        return 8;
                                                                    }
                                                                    else if (AFD >= 11) {
                                                                        return 2;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (mcTcor >= 95) {
                                                            if (diffRsCsdiff < 23) {
                                                                if (Rs < 4354) {
                                                                    return 2;
                                                                }
                                                                else if (Rs >= 4354) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (diffRsCsdiff >= 23) {
                                                                if (RsDiff < 9) {
                                                                    return 4;
                                                                }
                                                                else if (RsDiff >= 9) {
                                                                    return 2;
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (RsDiff >= 10) {
                                                        if (Cs < 2319) {
                                                            if (Cs < 2226) {
                                                                return 4;
                                                            }
                                                            else if (Cs >= 2226) {
                                                                return 2;
                                                            }
                                                        }
                                                        else if (Cs >= 2319) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                                else if (ssDCval >= 119) {
                                                    if (RsDiff < 10) {
                                                        if (SC < 7197) {
                                                            if (AFD < 8) {
                                                                return 4;
                                                            }
                                                            else if (AFD >= 8) {
                                                                if (diffAFD < -1) {
                                                                    return 8;
                                                                }
                                                                else if (diffAFD >= -1) {
                                                                    if (CsDiff < 6) {
                                                                        if (RsDiff < 8) {
                                                                            return 4;
                                                                        }
                                                                        else if (RsDiff >= 8) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                    else if (CsDiff >= 6) {
                                                                        if (tcor < 94) {
                                                                            return 4;
                                                                        }
                                                                        else if (tcor >= 94) {
                                                                            if (mcTcor < 95) {
                                                                                return 8;
                                                                            }
                                                                            else if (mcTcor >= 95) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (SC >= 7197) {
                                                            if (Rs < 4804) {
                                                                if (Rs < 4605) {
                                                                    return 4;
                                                                }
                                                                else if (Rs >= 4605) {
                                                                    if (Rs < 4779) {
                                                                        return 2;
                                                                    }
                                                                    else if (Rs >= 4779) {
                                                                        return 4;
                                                                    }
                                                                }
                                                            }
                                                            else if (Rs >= 4804) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                    else if (RsDiff >= 10) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (ssDCval >= 128) {
                                        if (mu_mv_mag_sq < 80) {
                                            if (TSC < 7) {
                                                if (Rs < 1475) {
                                                    if (Cs < 1170) {
                                                        return 8;
                                                    }
                                                    else if (Cs >= 1170) {
                                                        return 2;
                                                    }
                                                }
                                                else if (Rs >= 1475) {
                                                    return 4;
                                                }
                                            }
                                            else if (TSC >= 7) {
                                                if (SC < 3837) {
                                                    return 2;
                                                }
                                                else if (SC >= 3837) {
                                                    if (Cs < 1992) {
                                                        if (RsCsDiff < 218) {
                                                            return 4;
                                                        }
                                                        else if (RsCsDiff >= 218) {
                                                            if (RsDiff < 20) {
                                                                return 2;
                                                            }
                                                            else if (RsDiff >= 20) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                    else if (Cs >= 1992) {
                                                        if (diffAFD < 1) {
                                                            return 2;
                                                        }
                                                        else if (diffAFD >= 1) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (mu_mv_mag_sq >= 80) {
                                            if (mu_mv_mag_sq < 108) {
                                                return 2;
                                            }
                                            else if (mu_mv_mag_sq >= 108) {
                                                if (Cs < 1684) {
                                                    return 1;
                                                }
                                                else if (Cs >= 1684) {
                                                    return 8;
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (TSC >= 10) {
                                    if (tcor < 87) {
                                        if (RsDiff < 11) {
                                            if (SC < 5758) {
                                                return 1;
                                            }
                                            else if (SC >= 5758) {
                                                if (diffRsCsdiff < 18) {
                                                    if (Cs < 3022) {
                                                        return 4;
                                                    }
                                                    else if (Cs >= 3022) {
                                                        if (diffMVdiffVal < -2) {
                                                            return 2;
                                                        }
                                                        else if (diffMVdiffVal >= -2) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                                else if (diffRsCsdiff >= 18) {
                                                    return 2;
                                                }
                                            }
                                        }
                                        else if (RsDiff >= 11) {
                                            if (tcor < 61) {
                                                return 2;
                                            }
                                            else if (tcor >= 61) {
                                                if (RsCsDiff < 362) {
                                                    if (mcTcor < 86) {
                                                        if (Rs < 4673) {
                                                            return 4;
                                                        }
                                                        else if (Rs >= 4673) {
                                                            return 1;
                                                        }
                                                    }
                                                    else if (mcTcor >= 86) {
                                                        if (TSC < 12) {
                                                            return 2;
                                                        }
                                                        else if (TSC >= 12) {
                                                            if (SC < 6299) {
                                                                return 4;
                                                            }
                                                            else if (SC >= 6299) {
                                                                if (Rs < 3618) {
                                                                    return 8;
                                                                }
                                                                else if (Rs >= 3618) {
                                                                    return 2;
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (RsCsDiff >= 362) {
                                                    if (SC < 3937) {
                                                        return 2;
                                                    }
                                                    else if (SC >= 3937) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (tcor >= 87) {
                                        if (refDCval < 119) {
                                            if (posBalance < 55) {
                                                if (SC < 5919) {
                                                    return 4;
                                                }
                                                else if (SC >= 5919) {
                                                    return 8;
                                                }
                                            }
                                            else if (posBalance >= 55) {
                                                if (posBalance < 56) {
                                                    return 1;
                                                }
                                                else if (posBalance >= 56) {
                                                    return 8;
                                                }
                                            }
                                        }
                                        else if (refDCval >= 119) {
                                            return 8;
                                        }
                                    }
                                }
                            }
                        }
                    }
                }
            }
            else if (Cs >= 3285) {
                if (mu_mv_mag_sq < 61) {
                    if (AFD < 15) {
                        if (RsCsDiff < 46) {
                            if (TSCindex < 9) {
                                if (MVDiff < 18) {
                                    if (tcor < 89) {
                                        if (TSCindex < 8) {
                                            if (Rs < 3658) {
                                                return 4;
                                            }
                                            else if (Rs >= 3658) {
                                                return 8;
                                            }
                                        }
                                        else if (TSCindex >= 8) {
                                            return 4;
                                        }
                                    }
                                    else if (tcor >= 89) {
                                        if (TSC < 6) {
                                            if (negBalance < 44) {
                                                if (Rs < 3542) {
                                                    return 4;
                                                }
                                                else if (Rs >= 3542) {
                                                    return 8;
                                                }
                                            }
                                            else if (negBalance >= 44) {
                                                return 4;
                                            }
                                        }
                                        else if (TSC >= 6) {
                                            if (negBalance < 50) {
                                                if (RsCsDiff < 11) {
                                                    return 2;
                                                }
                                                else if (RsCsDiff >= 11) {
                                                    return 4;
                                                }
                                            }
                                            else if (negBalance >= 50) {
                                                if (Cs < 3856) {
                                                    if (MVDiff < 15) {
                                                        return 4;
                                                    }
                                                    else if (MVDiff >= 15) {
                                                        if (Rs < 3620) {
                                                            return 2;
                                                        }
                                                        else if (Rs >= 3620) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                                else if (Cs >= 3856) {
                                                    return 2;
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (MVDiff >= 18) {
                                    if (TSC < 7) {
                                        if (tcor < 92) {
                                            return 8;
                                        }
                                        else if (tcor >= 92) {
                                            if (RsCsDiff < 16) {
                                                return 8;
                                            }
                                            else if (RsCsDiff >= 16) {
                                                return 4;
                                            }
                                        }
                                    }
                                    else if (TSC >= 7) {
                                        if (mu_mv_mag_sq < 38) {
                                            if (tcor < 87) {
                                                if (Cs < 3703) {
                                                    return 4;
                                                }
                                                else if (Cs >= 3703) {
                                                    return 8;
                                                }
                                            }
                                            else if (tcor >= 87) {
                                                if (diffMVdiffVal < 8) {
                                                    return 2;
                                                }
                                                else if (diffMVdiffVal >= 8) {
                                                    return 8;
                                                }
                                            }
                                        }
                                        else if (mu_mv_mag_sq >= 38) {
                                            if (Rs < 3740) {
                                                return 8;
                                            }
                                            else if (Rs >= 3740) {
                                                return 4;
                                            }
                                        }
                                    }
                                }
                            }
                            else if (TSCindex >= 9) {
                                if (posBalance < 59) {
                                    if (Rs < 4089) {
                                        return 2;
                                    }
                                    else if (Rs >= 4089) {
                                        return 4;
                                    }
                                }
                                else if (posBalance >= 59) {
                                    if (Rs < 3659) {
                                        return 4;
                                    }
                                    else if (Rs >= 3659) {
                                        return 8;
                                    }
                                }
                            }
                        }
                        else if (RsCsDiff >= 46) {
                            if (posBalance < 57) {
                                if (mcTcor < 93) {
                                    if (diffRsCsdiff < -56) {
                                        if (Cs < 3853) {
                                            return 1;
                                        }
                                        else if (Cs >= 3853) {
                                            return 8;
                                        }
                                    }
                                    else if (diffRsCsdiff >= -56) {
                                        if (mu_mv_mag_sq < 34) {
                                            return 8;
                                        }
                                        else if (mu_mv_mag_sq >= 34) {
                                            return 4;
                                        }
                                    }
                                }
                                else if (mcTcor >= 93) {
                                    if (RsCsDiff < 57) {
                                        if (AFD < 13) {
                                            return 4;
                                        }
                                        else if (AFD >= 13) {
                                            return 8;
                                        }
                                    }
                                    else if (RsCsDiff >= 57) {
                                        return 8;
                                    }
                                }
                            }
                            else if (posBalance >= 57) {
                                if (gchDC < 3) {
                                    return 4;
                                }
                                else if (gchDC >= 3) {
                                    if (Rs < 4469) {
                                        return 8;
                                    }
                                    else if (Rs >= 4469) {
                                        return 4;
                                    }
                                }
                            }
                        }
                    }
                    else if (AFD >= 15) {
                        if (SC < 12036) {
                            if (refDCval < 153) {
                                if (AFD < 31) {
                                    if (Rs < 4123) {
                                        if (AFD < 18) {
                                            return 8;
                                        }
                                        else if (AFD >= 18) {
                                            if (refDCval < 124) {
                                                if (Rs < 3444) {
                                                    if (SC < 5705) {
                                                        return 4;
                                                    }
                                                    else if (SC >= 5705) {
                                                        if (diffMVdiffVal < -2) {
                                                            return 4;
                                                        }
                                                        else if (diffMVdiffVal >= -2) {
                                                            return 1;
                                                        }
                                                    }
                                                }
                                                else if (Rs >= 3444) {
                                                    if (MVDiff < 10) {
                                                        return 4;
                                                    }
                                                    else if (MVDiff >= 10) {
                                                        if (CsDiff < 15) {
                                                            if (Rs < 3988) {
                                                                return 8;
                                                            }
                                                            else if (Rs >= 3988) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (CsDiff >= 15) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (refDCval >= 124) {
                                                return 1;
                                            }
                                        }
                                    }
                                    else if (Rs >= 4123) {
                                        if (Cs < 4346) {
                                            if (Cs < 3475) {
                                                if (AFD < 16) {
                                                    if (Rs < 5034) {
                                                        return 1;
                                                    }
                                                    else if (Rs >= 5034) {
                                                        return 2;
                                                    }
                                                }
                                                else if (AFD >= 16) {
                                                    if (posBalance < 55) {
                                                        return 8;
                                                    }
                                                    else if (posBalance >= 55) {
                                                        if (Cs < 3393) {
                                                            return 1;
                                                        }
                                                        else if (Cs >= 3393) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (Cs >= 3475) {
                                                if (mu_mv_mag_sq < 50) {
                                                    if (posBalance < 52) {
                                                        if (AFD < 17) {
                                                            return 4;
                                                        }
                                                        else if (AFD >= 17) {
                                                            if (negBalance < 61) {
                                                                if (Cs < 3478) {
                                                                    return 4;
                                                                }
                                                                else if (Cs >= 3478) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (negBalance >= 61) {
                                                                if (Rs < 5172) {
                                                                    return 8;
                                                                }
                                                                else if (Rs >= 5172) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (posBalance >= 52) {
                                                        if (mcTcor < 84) {
                                                            if (posBalance < 61) {
                                                                if (RsDiff < 11) {
                                                                    return 8;
                                                                }
                                                                else if (RsDiff >= 11) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (posBalance >= 61) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (mcTcor >= 84) {
                                                            if (SC < 8996) {
                                                                return 8;
                                                            }
                                                            else if (SC >= 8996) {
                                                                if (MVDiff < 13) {
                                                                    return 8;
                                                                }
                                                                else if (MVDiff >= 13) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (mu_mv_mag_sq >= 50) {
                                                    return 8;
                                                }
                                            }
                                        }
                                        else if (Cs >= 4346) {
                                            if (tcor < 83) {
                                                if (MVDiff < 32) {
                                                    return 8;
                                                }
                                                else if (MVDiff >= 32) {
                                                    if (CsDiff < 20) {
                                                        return 4;
                                                    }
                                                    else if (CsDiff >= 20) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                            else if (tcor >= 83) {
                                                if (negBalance < 49) {
                                                    if (refDCval < 65) {
                                                        return 4;
                                                    }
                                                    else if (refDCval >= 65) {
                                                        return 8;
                                                    }
                                                }
                                                else if (negBalance >= 49) {
                                                    if (refDCval < 110) {
                                                        return 1;
                                                    }
                                                    else if (refDCval >= 110) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (AFD >= 31) {
                                    if (TSC < 18) {
                                        return 2;
                                    }
                                    else if (TSC >= 18) {
                                        if (tcor < 75) {
                                            if (Cs < 3598) {
                                                if (Cs < 3573) {
                                                    return 8;
                                                }
                                                else if (Cs >= 3573) {
                                                    return 2;
                                                }
                                            }
                                            else if (Cs >= 3598) {
                                                return 8;
                                            }
                                        }
                                        else if (tcor >= 75) {
                                            if (Rs < 6233) {
                                                return 4;
                                            }
                                            else if (Rs >= 6233) {
                                                return 2;
                                            }
                                        }
                                    }
                                }
                            }
                            else if (refDCval >= 153) {
                                if (posBalance < 60) {
                                    if (refDCval < 157) {
                                        return 8;
                                    }
                                    else if (refDCval >= 157) {
                                        return 4;
                                    }
                                }
                                else if (posBalance >= 60) {
                                    if (Cs < 3809) {
                                        if (ssDCval < 159) {
                                            return 4;
                                        }
                                        else if (ssDCval >= 159) {
                                            return 2;
                                        }
                                    }
                                    else if (Cs >= 3809) {
                                        return 2;
                                    }
                                }
                            }
                        }
                        else if (SC >= 12036) {
                            if (TSC < 18) {
                                if (refDCval < 149) {
                                    if (SC < 12208) {
                                        return 8;
                                    }
                                    else if (SC >= 12208) {
                                        return 4;
                                    }
                                }
                                else if (refDCval >= 149) {
                                    if (SC < 13238) {
                                        if (AFD < 28) {
                                            return 2;
                                        }
                                        else if (AFD >= 28) {
                                            return 4;
                                        }
                                    }
                                    else if (SC >= 13238) {
                                        return 8;
                                    }
                                }
                            }
                            else if (TSC >= 18) {
                                if (mu_mv_mag_sq < 9) {
                                    if (CsDiff < 39) {
                                        if (SC < 12486) {
                                            return 2;
                                        }
                                        else if (SC >= 12486) {
                                            if (Cs < 4677) {
                                                return 4;
                                            }
                                            else if (Cs >= 4677) {
                                                if (posBalance < 59) {
                                                    if (CsDiff < 23) {
                                                        if (AFD < 26) {
                                                            return 8;
                                                        }
                                                        else if (AFD >= 26) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (CsDiff >= 23) {
                                                        if (mcTcor < 79) {
                                                            if (Rs < 11671) {
                                                                return 8;
                                                            }
                                                            else if (Rs >= 11671) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (mcTcor >= 79) {
                                                            if (ssDCval < 94) {
                                                                return 4;
                                                            }
                                                            else if (ssDCval >= 94) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (posBalance >= 59) {
                                                    return 8;
                                                }
                                            }
                                        }
                                    }
                                    else if (CsDiff >= 39) {
                                        return 2;
                                    }
                                }
                                else if (mu_mv_mag_sq >= 9) {
                                    if (mcTcor < 87) {
                                        if (mu_mv_mag_sq < 33) {
                                            if (ssDCval < 144) {
                                                return 4;
                                            }
                                            else if (ssDCval >= 144) {
                                                return 2;
                                            }
                                        }
                                        else if (mu_mv_mag_sq >= 33) {
                                            if (refDCval < 124) {
                                                if (posBalance < 58) {
                                                    if (refDCval < 96) {
                                                        return 4;
                                                    }
                                                    else if (refDCval >= 96) {
                                                        return 8;
                                                    }
                                                }
                                                else if (posBalance >= 58) {
                                                    if (RsDiff < 24) {
                                                        return 8;
                                                    }
                                                    else if (RsDiff >= 24) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                            else if (refDCval >= 124) {
                                                return 8;
                                            }
                                        }
                                    }
                                    else if (mcTcor >= 87) {
                                        if (posBalance < 53) {
                                            if (AFD < 30) {
                                                return 8;
                                            }
                                            else if (AFD >= 30) {
                                                return 4;
                                            }
                                        }
                                        else if (posBalance >= 53) {
                                            if (ssDCval < 140) {
                                                return 2;
                                            }
                                            else if (ssDCval >= 140) {
                                                if (AFD < 34) {
                                                    if (RsDiff < 39) {
                                                        return 4;
                                                    }
                                                    else if (RsDiff >= 39) {
                                                        return 2;
                                                    }
                                                }
                                                else if (AFD >= 34) {
                                                    return 2;
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                        }
                    }
                }
                else if (mu_mv_mag_sq >= 61) {
                    if (refDCval < 66) {
                        if (TSC < 7) {
                            if (refDCval < 54) {
                                if (negBalance < 46) {
                                    if (RsCsDiff < 16) {
                                        return 8;
                                    }
                                    else if (RsCsDiff >= 16) {
                                        return 4;
                                    }
                                }
                                else if (negBalance >= 46) {
                                    return 4;
                                }
                            }
                            else if (refDCval >= 54) {
                                if (ssDCval < 56) {
                                    return 8;
                                }
                                else if (ssDCval >= 56) {
                                    if (diffTSC < -1) {
                                        return 4;
                                    }
                                    else if (diffTSC >= -1) {
                                        if (AFD < 7) {
                                            return 8;
                                        }
                                        else if (AFD >= 7) {
                                            return 4;
                                        }
                                    }
                                }
                            }
                        }
                        else if (TSC >= 7) {
                            if (tcor < 84) {
                                if (Cs < 3690) {
                                    return 4;
                                }
                                else if (Cs >= 3690) {
                                    return 8;
                                }
                            }
                            else if (tcor >= 84) {
                                return 4;
                            }
                        }
                    }
                    else if (refDCval >= 66) {
                        if (Cs < 3514) {
                            if (ssDCval < 115) {
                                if (CsDiff < 12) {
                                    if (Rs < 4147) {
                                        return 8;
                                    }
                                    else if (Rs >= 4147) {
                                        return 2;
                                    }
                                }
                                else if (CsDiff >= 12) {
                                    return 4;
                                }
                            }
                            else if (ssDCval >= 115) {
                                return 8;
                            }
                        }
                        else if (Cs >= 3514) {
                            if (ssDCval < 136) {
                                if (ssDCval < 75) {
                                    if (Rs < 7879) {
                                        return 8;
                                    }
                                    else if (Rs >= 7879) {
                                        return 4;
                                    }
                                }
                                else if (ssDCval >= 75) {
                                    return 8;
                                }
                            }
                            else if (ssDCval >= 136) {
                                if (Rs < 4399) {
                                    return 1;
                                }
                                else if (Rs >= 4399) {
                                    if (tcor < 70) {
                                        return 1;
                                    }
                                    else if (tcor >= 70) {
                                        return 8;
                                    }
                                }
                            }
                        }
                    }
                }
            }
        }
        return 4;
    }
    mfxU16 AGOPSelect15(mfxU32 Rs, mfxU32 CsDiff, mfxU32 TSC, mfxI32 diffAFD, mfxU32 Scindex, mfxU32 RsDiff, mfxI32 diffMVdiffVal, mfxU32 RsCsDiff, mfxU32 TSCindex, mfxI16 tcor, mfxU32 refDCval, mfxU32 MVDiff, mfxU32 AFD, mfxI32 diffTSC, mfxI32 diffRsCsdiff, mfxU32 posBalance, mfxU32 SC, mfxU32 mu_mv_mag_sq, mfxU32 negBalance, mfxI16 mcTcor, mfxU32 gchDC, mfxU32 Cs, mfxU32 ssDCval) {
        if (mu_mv_mag_sq < 83) {
            if (ssDCval < 17) {
                if (Cs < 1772) {
                    if (diffTSC < -1) {
                        return 4;
                    }
                    else if (diffTSC >= -1) {
                        return 8;
                    }
                }
                else if (Cs >= 1772) {
                    if (Rs < 4729) {
                        if (Rs < 2707) {
                            return 1;
                        }
                        else if (Rs >= 2707) {
                            if (Rs < 3770) {
                                if (Rs < 3755) {
                                    return 8;
                                }
                                else if (Rs >= 3755) {
                                    return 1;
                                }
                            }
                            else if (Rs >= 3770) {
                                return 8;
                            }
                        }
                    }
                    else if (Rs >= 4729) {
                        return 1;
                    }
                }
            }
            else if (ssDCval >= 17) {
                if (Rs < 3400) {
                    if (AFD < 12) {
                        if (Cs < 270) {
                            if (RsCsDiff < 3) {
                                if (SC < 833) {
                                    if (posBalance < 7) {
                                        if (refDCval < 47) {
                                            if (diffMVdiffVal < -18) {
                                                if (refDCval < 22) {
                                                    return 1;
                                                }
                                                else if (refDCval >= 22) {
                                                    return 2;
                                                }
                                            }
                                            else if (diffMVdiffVal >= -18) {
                                                if (diffTSC < -1) {
                                                    if (mcTcor < 120) {
                                                        if (diffAFD < -1) {
                                                            return 2;
                                                        }
                                                        else if (diffAFD >= -1) {
                                                            if (MVDiff < 39) {
                                                                return 4;
                                                            }
                                                            else if (MVDiff >= 39) {
                                                                return 2;
                                                            }
                                                        }
                                                    }
                                                    else if (mcTcor >= 120) {
                                                        return 8;
                                                    }
                                                }
                                                else if (diffTSC >= -1) {
                                                    if (TSCindex < 1) {
                                                        if (AFD < 1) {
                                                            if (MVDiff < 88) {
                                                                return 8;
                                                            }
                                                            else if (MVDiff >= 88) {
                                                                if (MVDiff < 111) {
                                                                    return 2;
                                                                }
                                                                else if (MVDiff >= 111) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                        else if (AFD >= 1) {
                                                            return 2;
                                                        }
                                                    }
                                                    else if (TSCindex >= 1) {
                                                        if (MVDiff < 608) {
                                                            if (MVDiff < 35) {
                                                                if (SC < 1) {
                                                                    return 4;
                                                                }
                                                                else if (SC >= 1) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (MVDiff >= 35) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (MVDiff >= 608) {
                                                            if (MVDiff < 974) {
                                                                return 4;
                                                            }
                                                            else if (MVDiff >= 974) {
                                                                return 2;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (refDCval >= 47) {
                                            return 8;
                                        }
                                    }
                                    else if (posBalance >= 7) {
                                        if (AFD < 8) {
                                            if (ssDCval < 99) {
                                                if (tcor < 59) {
                                                    if (ssDCval < 22) {
                                                        if (MVDiff < 62) {
                                                            if (posBalance < 27) {
                                                                return 1;
                                                            }
                                                            else if (posBalance >= 27) {
                                                                if (MVDiff < 59) {
                                                                    return 4;
                                                                }
                                                                else if (MVDiff >= 59) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                        else if (MVDiff >= 62) {
                                                            if (diffMVdiffVal < -43) {
                                                                if (TSCindex < 1) {
                                                                    if (ssDCval < 19) {
                                                                        return 1;
                                                                    }
                                                                    else if (ssDCval >= 19) {
                                                                        if (diffMVdiffVal < -129) {
                                                                            return 4;
                                                                        }
                                                                        else if (diffMVdiffVal >= -129) {
                                                                            return 2;
                                                                        }
                                                                    }
                                                                }
                                                                else if (TSCindex >= 1) {
                                                                    if (diffMVdiffVal < -323) {
                                                                        if (posBalance < 95) {
                                                                            return 2;
                                                                        }
                                                                        else if (posBalance >= 95) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                    else if (diffMVdiffVal >= -323) {
                                                                        return 2;
                                                                    }
                                                                }
                                                            }
                                                            else if (diffMVdiffVal >= -43) {
                                                                if (MVDiff < 208) {
                                                                    if (MVDiff < 94) {
                                                                        if (diffMVdiffVal < -1) {
                                                                            return 4;
                                                                        }
                                                                        else if (diffMVdiffVal >= -1) {
                                                                            if (MVDiff < 69) {
                                                                                if (MVDiff < 65) {
                                                                                    return 4;
                                                                                }
                                                                                else if (MVDiff >= 65) {
                                                                                    return 8;
                                                                                }
                                                                            }
                                                                            else if (MVDiff >= 69) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (MVDiff >= 94) {
                                                                        return 8;
                                                                    }
                                                                }
                                                                else if (MVDiff >= 208) {
                                                                    if (diffAFD < -2) {
                                                                        return 2;
                                                                    }
                                                                    else if (diffAFD >= -2) {
                                                                        if (tcor < 56) {
                                                                            if (mu_mv_mag_sq < 9) {
                                                                                return 1;
                                                                            }
                                                                            else if (mu_mv_mag_sq >= 9) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                        else if (tcor >= 56) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (ssDCval >= 22) {
                                                        if (tcor < 58) {
                                                            return 1;
                                                        }
                                                        else if (tcor >= 58) {
                                                            if (AFD < 2) {
                                                                return 4;
                                                            }
                                                            else if (AFD >= 2) {
                                                                return 1;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (tcor >= 59) {
                                                    if (mu_mv_mag_sq < 33) {
                                                        if (negBalance < 8) {
                                                            if (posBalance < 114) {
                                                                if (MVDiff < 28) {
                                                                    if (mu_mv_mag_sq < 30) {
                                                                        if (gchDC < 5) {
                                                                            return 4;
                                                                        }
                                                                        else if (gchDC >= 5) {
                                                                            if (Rs < 186) {
                                                                                return 8;
                                                                            }
                                                                            else if (Rs >= 186) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (mu_mv_mag_sq >= 30) {
                                                                        return 8;
                                                                    }
                                                                }
                                                                else if (MVDiff >= 28) {
                                                                    if (MVDiff < 56) {
                                                                        return 2;
                                                                    }
                                                                    else if (MVDiff >= 56) {
                                                                        if (mcTcor < 71) {
                                                                            return 8;
                                                                        }
                                                                        else if (mcTcor >= 71) {
                                                                            if (MVDiff < 98) {
                                                                                return 4;
                                                                            }
                                                                            else if (MVDiff >= 98) {
                                                                                if (mu_mv_mag_sq < 18) {
                                                                                    return 8;
                                                                                }
                                                                                else if (mu_mv_mag_sq >= 18) {
                                                                                    if (AFD < 1) {
                                                                                        return 2;
                                                                                    }
                                                                                    else if (AFD >= 1) {
                                                                                        return 4;
                                                                                    }
                                                                                }
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                            else if (posBalance >= 114) {
                                                                if (diffAFD < 2) {
                                                                    return 2;
                                                                }
                                                                else if (diffAFD >= 2) {
                                                                    return 1;
                                                                }
                                                            }
                                                        }
                                                        else if (negBalance >= 8) {
                                                            if (MVDiff < 37) {
                                                                return 2;
                                                            }
                                                            else if (MVDiff >= 37) {
                                                                if (posBalance < 19) {
                                                                    if (diffMVdiffVal < 19) {
                                                                        return 2;
                                                                    }
                                                                    else if (diffMVdiffVal >= 19) {
                                                                        if (AFD < 2) {
                                                                            return 8;
                                                                        }
                                                                        else if (AFD >= 2) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                }
                                                                else if (posBalance >= 19) {
                                                                    return 1;
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (mu_mv_mag_sq >= 33) {
                                                        if (tcor < 82) {
                                                            if (diffAFD < -2) {
                                                                return 1;
                                                            }
                                                            else if (diffAFD >= -2) {
                                                                if (Rs < 7) {
                                                                    return 4;
                                                                }
                                                                else if (Rs >= 7) {
                                                                    if (AFD < 1) {
                                                                        return 2;
                                                                    }
                                                                    else if (AFD >= 1) {
                                                                        return 4;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (tcor >= 82) {
                                                            if (tcor < 106) {
                                                                if (negBalance < 30) {
                                                                    if (Rs < 319) {
                                                                        if (tcor < 83) {
                                                                            if (SC < 297) {
                                                                                return 4;
                                                                            }
                                                                            else if (SC >= 297) {
                                                                                return 8;
                                                                            }
                                                                        }
                                                                        else if (tcor >= 83) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                    else if (Rs >= 319) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (negBalance >= 30) {
                                                                    if (Cs < 27) {
                                                                        return 1;
                                                                    }
                                                                    else if (Cs >= 27) {
                                                                        return 2;
                                                                    }
                                                                }
                                                            }
                                                            else if (tcor >= 106) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                            else if (ssDCval >= 99) {
                                                if (TSC < 5) {
                                                    return 1;
                                                }
                                                else if (TSC >= 5) {
                                                    return 4;
                                                }
                                            }
                                        }
                                        else if (AFD >= 8) {
                                            if (tcor < 61) {
                                                return 2;
                                            }
                                            else if (tcor >= 61) {
                                                if (MVDiff < 19) {
                                                    if (RsDiff < 1) {
                                                        return 2;
                                                    }
                                                    else if (RsDiff >= 1) {
                                                        return 4;
                                                    }
                                                }
                                                else if (MVDiff >= 19) {
                                                    if (negBalance < 111) {
                                                        return 1;
                                                    }
                                                    else if (negBalance >= 111) {
                                                        return 2;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (SC >= 833) {
                                    if (Rs < 852) {
                                        if (Rs < 826) {
                                            return 8;
                                        }
                                        else if (Rs >= 826) {
                                            return 1;
                                        }
                                    }
                                    else if (Rs >= 852) {
                                        return 8;
                                    }
                                }
                            }
                            else if (RsCsDiff >= 3) {
                                if (tcor < 67) {
                                    if (ssDCval < 73) {
                                        return 1;
                                    }
                                    else if (ssDCval >= 73) {
                                        return 2;
                                    }
                                }
                                else if (tcor >= 67) {
                                    if (negBalance < 47) {
                                        if (mu_mv_mag_sq < 26) {
                                            if (RsCsDiff < 5) {
                                                return 1;
                                            }
                                            else if (RsCsDiff >= 5) {
                                                if (Rs < 143) {
                                                    return 1;
                                                }
                                                else if (Rs >= 143) {
                                                    return 2;
                                                }
                                            }
                                        }
                                        else if (mu_mv_mag_sq >= 26) {
                                            if (Cs < 149) {
                                                return 2;
                                            }
                                            else if (Cs >= 149) {
                                                return 8;
                                            }
                                        }
                                    }
                                    else if (negBalance >= 47) {
                                        if (ssDCval < 104) {
                                            if (ssDCval < 71) {
                                                if (Rs < 462) {
                                                    return 4;
                                                }
                                                else if (Rs >= 462) {
                                                    return 2;
                                                }
                                            }
                                            else if (ssDCval >= 71) {
                                                return 1;
                                            }
                                        }
                                        else if (ssDCval >= 104) {
                                            if (Cs < 171) {
                                                return 2;
                                            }
                                            else if (Cs >= 171) {
                                                return 4;
                                            }
                                        }
                                    }
                                }
                            }
                        }
                        else if (Cs >= 270) {
                            if (refDCval < 98) {
                                if (negBalance < 50) {
                                    if (Rs < 736) {
                                        if (Cs < 383) {
                                            if (TSC < 6) {
                                                if (refDCval < 38) {
                                                    if (SC < 734) {
                                                        if (Rs < 339) {
                                                            if (SC < 527) {
                                                                return 2;
                                                            }
                                                            else if (SC >= 527) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (Rs >= 339) {
                                                            if (Cs < 292) {
                                                                return 4;
                                                            }
                                                            else if (Cs >= 292) {
                                                                if (Rs < 398) {
                                                                    if (RsCsDiff < 1) {
                                                                        return 4;
                                                                    }
                                                                    else if (RsCsDiff >= 1) {
                                                                        if (Rs < 390) {
                                                                            return 8;
                                                                        }
                                                                        else if (Rs >= 390) {
                                                                            if (Cs < 334) {
                                                                                return 1;
                                                                            }
                                                                            else if (Cs >= 334) {
                                                                                return 8;
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                                else if (Rs >= 398) {
                                                                    if (Rs < 411) {
                                                                        return 4;
                                                                    }
                                                                    else if (Rs >= 411) {
                                                                        return 1;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (SC >= 734) {
                                                        if (mcTcor < 87) {
                                                            if (Cs < 376) {
                                                                if (diffRsCsdiff < 1) {
                                                                    if (Cs < 367) {
                                                                        return 1;
                                                                    }
                                                                    else if (Cs >= 367) {
                                                                        if (MVDiff < 16) {
                                                                            return 4;
                                                                        }
                                                                        else if (MVDiff >= 16) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                }
                                                                else if (diffRsCsdiff >= 1) {
                                                                    if (Rs < 424) {
                                                                        return 8;
                                                                    }
                                                                    else if (Rs >= 424) {
                                                                        return 4;
                                                                    }
                                                                }
                                                            }
                                                            else if (Cs >= 376) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (mcTcor >= 87) {
                                                            return 2;
                                                        }
                                                    }
                                                }
                                                else if (refDCval >= 38) {
                                                    if (mu_mv_mag_sq < 45) {
                                                        if (Rs < 322) {
                                                            return 4;
                                                        }
                                                        else if (Rs >= 322) {
                                                            if (Cs < 345) {
                                                                return 1;
                                                            }
                                                            else if (Cs >= 345) {
                                                                return 2;
                                                            }
                                                        }
                                                    }
                                                    else if (mu_mv_mag_sq >= 45) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                            else if (TSC >= 6) {
                                                if (Rs < 437) {
                                                    if (SC < 573) {
                                                        return 2;
                                                    }
                                                    else if (SC >= 573) {
                                                        return 1;
                                                    }
                                                }
                                                else if (Rs >= 437) {
                                                    return 8;
                                                }
                                            }
                                        }
                                        else if (Cs >= 383) {
                                            if (Rs < 536) {
                                                if (tcor < 85) {
                                                    if (Rs < 381) {
                                                        if (diffMVdiffVal < -8) {
                                                            if (Rs < 185) {
                                                                return 2;
                                                            }
                                                            else if (Rs >= 185) {
                                                                return 1;
                                                            }
                                                        }
                                                        else if (diffMVdiffVal >= -8) {
                                                            return 2;
                                                        }
                                                    }
                                                    else if (Rs >= 381) {
                                                        if (refDCval < 64) {
                                                            if (Cs < 397) {
                                                                if (tcor < 82) {
                                                                    if (MVDiff < 25) {
                                                                        return 2;
                                                                    }
                                                                    else if (MVDiff >= 25) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (tcor >= 82) {
                                                                    if (tcor < 83) {
                                                                        return 2;
                                                                    }
                                                                    else if (tcor >= 83) {
                                                                        if (Rs < 449) {
                                                                            return 2;
                                                                        }
                                                                        else if (Rs >= 449) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                            else if (Cs >= 397) {
                                                                if (SC < 858) {
                                                                    if (TSC < 5) {
                                                                        return 2;
                                                                    }
                                                                    else if (TSC >= 5) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (SC >= 858) {
                                                                    if (MVDiff < 19) {
                                                                        if (SC < 878) {
                                                                            return 2;
                                                                        }
                                                                        else if (SC >= 878) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                    else if (MVDiff >= 19) {
                                                                        return 2;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (refDCval >= 64) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                                else if (tcor >= 85) {
                                                    if (negBalance < 25) {
                                                        return 2;
                                                    }
                                                    else if (negBalance >= 25) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                            else if (Rs >= 536) {
                                                if (ssDCval < 44) {
                                                    if (negBalance < 47) {
                                                        if (posBalance < 87) {
                                                            if (mu_mv_mag_sq < 59) {
                                                                if (Rs < 537) {
                                                                    return 4;
                                                                }
                                                                else if (Rs >= 537) {
                                                                    if (RsCsDiff < 4) {
                                                                        if (Rs < 679) {
                                                                            return 2;
                                                                        }
                                                                        else if (Rs >= 679) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                    else if (RsCsDiff >= 4) {
                                                                        return 2;
                                                                    }
                                                                }
                                                            }
                                                            else if (mu_mv_mag_sq >= 59) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (posBalance >= 87) {
                                                            if (Rs < 560) {
                                                                return 4;
                                                            }
                                                            else if (Rs >= 560) {
                                                                return 1;
                                                            }
                                                        }
                                                    }
                                                    else if (negBalance >= 47) {
                                                        return 4;
                                                    }
                                                }
                                                else if (ssDCval >= 44) {
                                                    if (ssDCval < 51) {
                                                        if (mcTcor < 99) {
                                                            return 8;
                                                        }
                                                        else if (mcTcor >= 99) {
                                                            if (Cs < 456) {
                                                                return 8;
                                                            }
                                                            else if (Cs >= 456) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                    else if (ssDCval >= 51) {
                                                        if (tcor < 95) {
                                                            if (diffMVdiffVal < -39) {
                                                                return 8;
                                                            }
                                                            else if (diffMVdiffVal >= -39) {
                                                                if (negBalance < 48) {
                                                                    return 4;
                                                                }
                                                                else if (negBalance >= 48) {
                                                                    if (TSC < 5) {
                                                                        return 2;
                                                                    }
                                                                    else if (TSC >= 5) {
                                                                        return 4;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (tcor >= 95) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (Rs >= 736) {
                                        if (TSCindex < 8) {
                                            if (Rs < 2001) {
                                                if (Rs < 1717) {
                                                    if (MVDiff < 22) {
                                                        if (SC < 1628) {
                                                            if (negBalance < 42) {
                                                                if (ssDCval < 50) {
                                                                    if (Cs < 522) {
                                                                        return 1;
                                                                    }
                                                                    else if (Cs >= 522) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (ssDCval >= 50) {
                                                                    if (posBalance < 50) {
                                                                        return 4;
                                                                    }
                                                                    else if (posBalance >= 50) {
                                                                        if (ssDCval < 55) {
                                                                            return 8;
                                                                        }
                                                                        else if (ssDCval >= 55) {
                                                                            if (AFD < 4) {
                                                                                return 8;
                                                                            }
                                                                            else if (AFD >= 4) {
                                                                                if (Rs < 1181) {
                                                                                    return 4;
                                                                                }
                                                                                else if (Rs >= 1181) {
                                                                                    return 8;
                                                                                }
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                            else if (negBalance >= 42) {
                                                                if (mcTcor < 90) {
                                                                    if (RsCsDiff < 33) {
                                                                        if (SC < 1312) {
                                                                            if (Rs < 769) {
                                                                                return 1;
                                                                            }
                                                                            else if (Rs >= 769) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                        else if (SC >= 1312) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                    else if (RsCsDiff >= 33) {
                                                                        return 1;
                                                                    }
                                                                }
                                                                else if (mcTcor >= 90) {
                                                                    if (SC < 1159) {
                                                                        return 1;
                                                                    }
                                                                    else if (SC >= 1159) {
                                                                        return 2;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (SC >= 1628) {
                                                            if (negBalance < 42) {
                                                                if (ssDCval < 53) {
                                                                    if (posBalance < 59) {
                                                                        if (Rs < 1171) {
                                                                            if (TSC < 3) {
                                                                                return 1;
                                                                            }
                                                                            else if (TSC >= 3) {
                                                                                if (ssDCval < 46) {
                                                                                    if (Cs < 943) {
                                                                                        return 1;
                                                                                    }
                                                                                    else if (Cs >= 943) {
                                                                                        return 4;
                                                                                    }
                                                                                }
                                                                                else if (ssDCval >= 46) {
                                                                                    if (Rs < 1128) {
                                                                                        return 4;
                                                                                    }
                                                                                    else if (Rs >= 1128) {
                                                                                        if (ssDCval < 51) {
                                                                                            return 1;
                                                                                        }
                                                                                        else if (ssDCval >= 51) {
                                                                                            return 4;
                                                                                        }
                                                                                    }
                                                                                }
                                                                            }
                                                                        }
                                                                        else if (Rs >= 1171) {
                                                                            if (negBalance < 35) {
                                                                                return 1;
                                                                            }
                                                                            else if (negBalance >= 35) {
                                                                                if (SC < 2308) {
                                                                                    if (mu_mv_mag_sq < 57) {
                                                                                        return 1;
                                                                                    }
                                                                                    else if (mu_mv_mag_sq >= 57) {
                                                                                        if (Rs < 1238) {
                                                                                            if (SC < 2080) {
                                                                                                return 8;
                                                                                            }
                                                                                            else if (SC >= 2080) {
                                                                                                return 4;
                                                                                            }
                                                                                        }
                                                                                        else if (Rs >= 1238) {
                                                                                            if (posBalance < 48) {
                                                                                                if (Cs < 894) {
                                                                                                    return 1;
                                                                                                }
                                                                                                else if (Cs >= 894) {
                                                                                                    return 4;
                                                                                                }
                                                                                            }
                                                                                            else if (posBalance >= 48) {
                                                                                                return 1;
                                                                                            }
                                                                                        }
                                                                                    }
                                                                                }
                                                                                else if (SC >= 2308) {
                                                                                    if (posBalance < 51) {
                                                                                        if (SC < 2336) {
                                                                                            return 4;
                                                                                        }
                                                                                        else if (SC >= 2336) {
                                                                                            return 1;
                                                                                        }
                                                                                    }
                                                                                    else if (posBalance >= 51) {
                                                                                        if (diffRsCsdiff < 15) {
                                                                                            if (diffAFD < 1) {
                                                                                                return 4;
                                                                                            }
                                                                                            else if (diffAFD >= 1) {
                                                                                                return 8;
                                                                                            }
                                                                                        }
                                                                                        else if (diffRsCsdiff >= 15) {
                                                                                            return 1;
                                                                                        }
                                                                                    }
                                                                                }
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (posBalance >= 59) {
                                                                        if (mcTcor < 91) {
                                                                            if (MVDiff < 13) {
                                                                                return 4;
                                                                            }
                                                                            else if (MVDiff >= 13) {
                                                                                if (Rs < 1279) {
                                                                                    return 8;
                                                                                }
                                                                                else if (Rs >= 1279) {
                                                                                    return 1;
                                                                                }
                                                                            }
                                                                        }
                                                                        else if (mcTcor >= 91) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                }
                                                                else if (ssDCval >= 53) {
                                                                    if (Rs < 1350) {
                                                                        if (diffMVdiffVal < -5) {
                                                                            return 1;
                                                                        }
                                                                        else if (diffMVdiffVal >= -5) {
                                                                            if (diffRsCsdiff < -1) {
                                                                                if (TSC < 6) {
                                                                                    return 4;
                                                                                }
                                                                                else if (TSC >= 6) {
                                                                                    return 8;
                                                                                }
                                                                            }
                                                                            else if (diffRsCsdiff >= -1) {
                                                                                if (Rs < 1141) {
                                                                                    if (CsDiff < 2) {
                                                                                        return 4;
                                                                                    }
                                                                                    else if (CsDiff >= 2) {
                                                                                        return 1;
                                                                                    }
                                                                                }
                                                                                else if (Rs >= 1141) {
                                                                                    if (mcTcor < 94) {
                                                                                        return 4;
                                                                                    }
                                                                                    else if (mcTcor >= 94) {
                                                                                        return 1;
                                                                                    }
                                                                                }
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (Rs >= 1350) {
                                                                        if (refDCval < 57) {
                                                                            if (refDCval < 54) {
                                                                                if (RsDiff < 4) {
                                                                                    return 1;
                                                                                }
                                                                                else if (RsDiff >= 4) {
                                                                                    if (Rs < 1570) {
                                                                                        return 4;
                                                                                    }
                                                                                    else if (Rs >= 1570) {
                                                                                        return 1;
                                                                                    }
                                                                                }
                                                                            }
                                                                            else if (refDCval >= 54) {
                                                                                if (posBalance < 52) {
                                                                                    return 1;
                                                                                }
                                                                                else if (posBalance >= 52) {
                                                                                    if (MVDiff < 13) {
                                                                                        return 8;
                                                                                    }
                                                                                    else if (MVDiff >= 13) {
                                                                                        return 4;
                                                                                    }
                                                                                }
                                                                            }
                                                                        }
                                                                        else if (refDCval >= 57) {
                                                                            if (SC < 2016) {
                                                                                return 8;
                                                                            }
                                                                            else if (SC >= 2016) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                            else if (negBalance >= 42) {
                                                                if (Cs < 968) {
                                                                    if (posBalance < 47) {
                                                                        if (AFD < 8) {
                                                                            return 4;
                                                                        }
                                                                        else if (AFD >= 8) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                    else if (posBalance >= 47) {
                                                                        if (Cs < 950) {
                                                                            return 1;
                                                                        }
                                                                        else if (Cs >= 950) {
                                                                            if (Cs < 953) {
                                                                                return 4;
                                                                            }
                                                                            else if (Cs >= 953) {
                                                                                return 1;
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                                else if (Cs >= 968) {
                                                                    if (MVDiff < 5) {
                                                                        if (Cs < 1593) {
                                                                            return 1;
                                                                        }
                                                                        else if (Cs >= 1593) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                    else if (MVDiff >= 5) {
                                                                        if (AFD < 10) {
                                                                            if (diffTSC < 1) {
                                                                                return 8;
                                                                            }
                                                                            else if (diffTSC >= 1) {
                                                                                if (Rs < 1396) {
                                                                                    return 8;
                                                                                }
                                                                                else if (Rs >= 1396) {
                                                                                    return 1;
                                                                                }
                                                                            }
                                                                        }
                                                                        else if (AFD >= 10) {
                                                                            return 2;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (MVDiff >= 22) {
                                                        if (Cs < 630) {
                                                            if (negBalance < 49) {
                                                                if (mu_mv_mag_sq < 53) {
                                                                    if (posBalance < 60) {
                                                                        if (diffAFD < 1) {
                                                                            if (RsCsDiff < 7) {
                                                                                return 4;
                                                                            }
                                                                            else if (RsCsDiff >= 7) {
                                                                                return 1;
                                                                            }
                                                                        }
                                                                        else if (diffAFD >= 1) {
                                                                            if (SC < 1424) {
                                                                                return 1;
                                                                            }
                                                                            else if (SC >= 1424) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (posBalance >= 60) {
                                                                        if (Rs < 774) {
                                                                            return 8;
                                                                        }
                                                                        else if (Rs >= 774) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                }
                                                                else if (mu_mv_mag_sq >= 53) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (negBalance >= 49) {
                                                                if (RsCsDiff < 4) {
                                                                    return 4;
                                                                }
                                                                else if (RsCsDiff >= 4) {
                                                                    return 1;
                                                                }
                                                            }
                                                        }
                                                        else if (Cs >= 630) {
                                                            if (TSCindex < 6) {
                                                                if (mu_mv_mag_sq < 54) {
                                                                    if (mcTcor < 94) {
                                                                        return 2;
                                                                    }
                                                                    else if (mcTcor >= 94) {
                                                                        if (RsDiff < 2) {
                                                                            return 8;
                                                                        }
                                                                        else if (RsDiff >= 2) {
                                                                            return 2;
                                                                        }
                                                                    }
                                                                }
                                                                else if (mu_mv_mag_sq >= 54) {
                                                                    if (mcTcor < 95) {
                                                                        if (Rs < 829) {
                                                                            return 1;
                                                                        }
                                                                        else if (Rs >= 829) {
                                                                            return 2;
                                                                        }
                                                                    }
                                                                    else if (mcTcor >= 95) {
                                                                        return 4;
                                                                    }
                                                                }
                                                            }
                                                            else if (TSCindex >= 6) {
                                                                if (posBalance < 52) {
                                                                    if (RsCsDiff < 37) {
                                                                        return 4;
                                                                    }
                                                                    else if (RsCsDiff >= 37) {
                                                                        return 8;
                                                                    }
                                                                }
                                                                else if (posBalance >= 52) {
                                                                    if (MVDiff < 36) {
                                                                        if (Rs < 1409) {
                                                                            return 1;
                                                                        }
                                                                        else if (Rs >= 1409) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                    else if (MVDiff >= 36) {
                                                                        if (Cs < 647) {
                                                                            return 4;
                                                                        }
                                                                        else if (Cs >= 647) {
                                                                            if (diffTSC < 1) {
                                                                                if (tcor < 84) {
                                                                                    return 2;
                                                                                }
                                                                                else if (tcor >= 84) {
                                                                                    return 8;
                                                                                }
                                                                            }
                                                                            else if (diffTSC >= 1) {
                                                                                return 1;
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (Rs >= 1717) {
                                                    if (ssDCval < 73) {
                                                        if (Cs < 1552) {
                                                            if (mcTcor < 92) {
                                                                if (AFD < 8) {
                                                                    return 4;
                                                                }
                                                                else if (AFD >= 8) {
                                                                    return 1;
                                                                }
                                                            }
                                                            else if (mcTcor >= 92) {
                                                                if (TSC < 6) {
                                                                    if (mu_mv_mag_sq < 60) {
                                                                        if (Rs < 1771) {
                                                                            return 2;
                                                                        }
                                                                        else if (Rs >= 1771) {
                                                                            if (SC < 3331) {
                                                                                return 4;
                                                                            }
                                                                            else if (SC >= 3331) {
                                                                                if (Rs < 1846) {
                                                                                    return 1;
                                                                                }
                                                                                else if (Rs >= 1846) {
                                                                                    if (Cs < 1540) {
                                                                                        return 1;
                                                                                    }
                                                                                    else if (Cs >= 1540) {
                                                                                        return 4;
                                                                                    }
                                                                                }
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (mu_mv_mag_sq >= 60) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (TSC >= 6) {
                                                                    return 1;
                                                                }
                                                            }
                                                        }
                                                        else if (Cs >= 1552) {
                                                            if (tcor < 77) {
                                                                return 4;
                                                            }
                                                            else if (tcor >= 77) {
                                                                if (diffRsCsdiff < 2) {
                                                                    return 1;
                                                                }
                                                                else if (diffRsCsdiff >= 2) {
                                                                    if (tcor < 91) {
                                                                        return 1;
                                                                    }
                                                                    else if (tcor >= 91) {
                                                                        return 4;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (ssDCval >= 73) {
                                                        if (mcTcor < 95) {
                                                            return 2;
                                                        }
                                                        else if (mcTcor >= 95) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (Rs >= 2001) {
                                                if (Cs < 821) {
                                                    if (refDCval < 72) {
                                                        return 2;
                                                    }
                                                    else if (refDCval >= 72) {
                                                        if (Rs < 2271) {
                                                            return 8;
                                                        }
                                                        else if (Rs >= 2271) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                                else if (Cs >= 821) {
                                                    if (Cs < 2145) {
                                                        if (ssDCval < 40) {
                                                            return 1;
                                                        }
                                                        else if (ssDCval >= 40) {
                                                            if (Scindex < 5) {
                                                                return 4;
                                                            }
                                                            else if (Scindex >= 5) {
                                                                if (Rs < 3377) {
                                                                    return 4;
                                                                }
                                                                else if (Rs >= 3377) {
                                                                    return 1;
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (Cs >= 2145) {
                                                        if (Rs < 2481) {
                                                            return 8;
                                                        }
                                                        else if (Rs >= 2481) {
                                                            if (refDCval < 50) {
                                                                return 4;
                                                            }
                                                            else if (refDCval >= 50) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (TSCindex >= 8) {
                                            if (negBalance < 34) {
                                                if (Rs < 1626) {
                                                    return 4;
                                                }
                                                else if (Rs >= 1626) {
                                                    if (CsDiff < 9) {
                                                        if (Rs < 2716) {
                                                            return 8;
                                                        }
                                                        else if (Rs >= 2716) {
                                                            return 2;
                                                        }
                                                    }
                                                    else if (CsDiff >= 9) {
                                                        return 1;
                                                    }
                                                }
                                            }
                                            else if (negBalance >= 34) {
                                                if (Cs < 2363) {
                                                    if (Rs < 2560) {
                                                        if (MVDiff < 4) {
                                                            if (Cs < 1884) {
                                                                return 2;
                                                            }
                                                            else if (Cs >= 1884) {
                                                                if (mu_mv_mag_sq < 82) {
                                                                    return 4;
                                                                }
                                                                else if (mu_mv_mag_sq >= 82) {
                                                                    return 2;
                                                                }
                                                            }
                                                        }
                                                        else if (MVDiff >= 4) {
                                                            if (RsCsDiff < 29) {
                                                                if (Rs < 2045) {
                                                                    return 1;
                                                                }
                                                                else if (Rs >= 2045) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (RsCsDiff >= 29) {
                                                                if (SC < 3880) {
                                                                    if (diffRsCsdiff < 32) {
                                                                        if (Rs < 1622) {
                                                                            if (mcTcor < 90) {
                                                                                if (Rs < 1571) {
                                                                                    return 8;
                                                                                }
                                                                                else if (Rs >= 1571) {
                                                                                    return 4;
                                                                                }
                                                                            }
                                                                            else if (mcTcor >= 90) {
                                                                                if (CsDiff < 4) {
                                                                                    return 2;
                                                                                }
                                                                                else if (CsDiff >= 4) {
                                                                                    return 4;
                                                                                }
                                                                            }
                                                                        }
                                                                        else if (Rs >= 1622) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                    else if (diffRsCsdiff >= 32) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (SC >= 3880) {
                                                                    if (refDCval < 83) {
                                                                        if (Cs < 2049) {
                                                                            return 4;
                                                                        }
                                                                        else if (Cs >= 2049) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                    else if (refDCval >= 83) {
                                                                        if (diffMVdiffVal < 3) {
                                                                            return 1;
                                                                        }
                                                                        else if (diffMVdiffVal >= 3) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (Rs >= 2560) {
                                                        if (refDCval < 86) {
                                                            if (CsDiff < 4) {
                                                                if (Cs < 1361) {
                                                                    return 2;
                                                                }
                                                                else if (Cs >= 1361) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (CsDiff >= 4) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (refDCval >= 86) {
                                                            if (MVDiff < 44) {
                                                                return 1;
                                                            }
                                                            else if (MVDiff >= 44) {
                                                                return 2;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (Cs >= 2363) {
                                                    if (CsDiff < 9) {
                                                        if (SC < 6658) {
                                                            return 8;
                                                        }
                                                        else if (SC >= 6658) {
                                                            if (Rs < 3378) {
                                                                return 8;
                                                            }
                                                            else if (Rs >= 3378) {
                                                                return 2;
                                                            }
                                                        }
                                                    }
                                                    else if (CsDiff >= 9) {
                                                        if (Rs < 2397) {
                                                            return 4;
                                                        }
                                                        else if (Rs >= 2397) {
                                                            return 2;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (negBalance >= 50) {
                                    if (mcTcor < 80) {
                                        if (CsDiff < 3) {
                                            if (MVDiff < 39) {
                                                return 1;
                                            }
                                            else if (MVDiff >= 39) {
                                                if (Rs < 721) {
                                                    return 4;
                                                }
                                                else if (Rs >= 721) {
                                                    if (tcor < 74) {
                                                        if (mu_mv_mag_sq < 28) {
                                                            return 4;
                                                        }
                                                        else if (mu_mv_mag_sq >= 28) {
                                                            return 1;
                                                        }
                                                    }
                                                    else if (tcor >= 74) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                        }
                                        else if (CsDiff >= 3) {
                                            if (SC < 2429) {
                                                return 4;
                                            }
                                            else if (SC >= 2429) {
                                                if (Cs < 1416) {
                                                    return 1;
                                                }
                                                else if (Cs >= 1416) {
                                                    return 8;
                                                }
                                            }
                                        }
                                    }
                                    else if (mcTcor >= 80) {
                                        if (TSC < 7) {
                                            if (Rs < 268) {
                                                if (SC < 468) {
                                                    return 8;
                                                }
                                                else if (SC >= 468) {
                                                    return 2;
                                                }
                                            }
                                            else if (Rs >= 268) {
                                                if (posBalance < 14) {
                                                    if (Rs < 1736) {
                                                        return 8;
                                                    }
                                                    else if (Rs >= 1736) {
                                                        return 2;
                                                    }
                                                }
                                                else if (posBalance >= 14) {
                                                    if (RsCsDiff < 19) {
                                                        if (posBalance < 61) {
                                                            if (Cs < 2344) {
                                                                if (negBalance < 52) {
                                                                    if (refDCval < 40) {
                                                                        if (Cs < 575) {
                                                                            return 4;
                                                                        }
                                                                        else if (Cs >= 575) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                    else if (refDCval >= 40) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (negBalance >= 52) {
                                                                    if (Cs < 549) {
                                                                        return 4;
                                                                    }
                                                                    else if (Cs >= 549) {
                                                                        if (mu_mv_mag_sq < 53) {
                                                                            if (refDCval < 97) {
                                                                                if (RsCsDiff < 12) {
                                                                                    return 8;
                                                                                }
                                                                                else if (RsCsDiff >= 12) {
                                                                                    if (Rs < 676) {
                                                                                        return 8;
                                                                                    }
                                                                                    else if (Rs >= 676) {
                                                                                        if (TSC < 5) {
                                                                                            return 1;
                                                                                        }
                                                                                        else if (TSC >= 5) {
                                                                                            return 4;
                                                                                        }
                                                                                    }
                                                                                }
                                                                            }
                                                                            else if (refDCval >= 97) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                        else if (mu_mv_mag_sq >= 53) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                            else if (Cs >= 2344) {
                                                                if (tcor < 93) {
                                                                    return 8;
                                                                }
                                                                else if (tcor >= 93) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                        else if (posBalance >= 61) {
                                                            if (Rs < 693) {
                                                                return 8;
                                                            }
                                                            else if (Rs >= 693) {
                                                                return 2;
                                                            }
                                                        }
                                                    }
                                                    else if (RsCsDiff >= 19) {
                                                        if (MVDiff < 14) {
                                                            if (SC < 2697) {
                                                                if (SC < 2251) {
                                                                    return 4;
                                                                }
                                                                else if (SC >= 2251) {
                                                                    if (AFD < 9) {
                                                                        return 2;
                                                                    }
                                                                    else if (AFD >= 9) {
                                                                        return 4;
                                                                    }
                                                                }
                                                            }
                                                            else if (SC >= 2697) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (MVDiff >= 14) {
                                                            if (Rs < 959) {
                                                                return 2;
                                                            }
                                                            else if (Rs >= 959) {
                                                                if (CsDiff < 4) {
                                                                    return 8;
                                                                }
                                                                else if (CsDiff >= 4) {
                                                                    return 1;
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (TSC >= 7) {
                                            if (mcTcor < 91) {
                                                if (SC < 3497) {
                                                    if (mu_mv_mag_sq < 3) {
                                                        if (mcTcor < 88) {
                                                            if (tcor < 74) {
                                                                return 4;
                                                            }
                                                            else if (tcor >= 74) {
                                                                if (CsDiff < 5) {
                                                                    if (RsCsDiff < 29) {
                                                                        return 4;
                                                                    }
                                                                    else if (RsCsDiff >= 29) {
                                                                        return 8;
                                                                    }
                                                                }
                                                                else if (CsDiff >= 5) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                        else if (mcTcor >= 88) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (mu_mv_mag_sq >= 3) {
                                                        if (AFD < 10) {
                                                            if (diffRsCsdiff < -18) {
                                                                return 2;
                                                            }
                                                            else if (diffRsCsdiff >= -18) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (AFD >= 10) {
                                                            if (Cs < 851) {
                                                                return 2;
                                                            }
                                                            else if (Cs >= 851) {
                                                                return 1;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (SC >= 3497) {
                                                    if (refDCval < 92) {
                                                        if (RsCsDiff < 117) {
                                                            if (ssDCval < 89) {
                                                                return 8;
                                                            }
                                                            else if (ssDCval >= 89) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (RsCsDiff >= 117) {
                                                            return 2;
                                                        }
                                                    }
                                                    else if (refDCval >= 92) {
                                                        if (SC < 3892) {
                                                            if (ssDCval < 96) {
                                                                if (Cs < 1731) {
                                                                    if (gchDC < 1) {
                                                                        return 2;
                                                                    }
                                                                    else if (gchDC >= 1) {
                                                                        return 8;
                                                                    }
                                                                }
                                                                else if (Cs >= 1731) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (ssDCval >= 96) {
                                                                return 2;
                                                            }
                                                        }
                                                        else if (SC >= 3892) {
                                                            return 2;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (mcTcor >= 91) {
                                                if (RsDiff < 5) {
                                                    if (diffAFD < -3) {
                                                        return 4;
                                                    }
                                                    else if (diffAFD >= -3) {
                                                        if (tcor < 100) {
                                                            return 2;
                                                        }
                                                        else if (tcor >= 100) {
                                                            if (Rs < 515) {
                                                                return 2;
                                                            }
                                                            else if (Rs >= 515) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (RsDiff >= 5) {
                                                    if (Rs < 2315) {
                                                        if (mcTcor < 98) {
                                                            if (tcor < 87) {
                                                                if (MVDiff < 19) {
                                                                    return 1;
                                                                }
                                                                else if (MVDiff >= 19) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (tcor >= 87) {
                                                                if (CsDiff < 7) {
                                                                    if (posBalance < 58) {
                                                                        if (diffMVdiffVal < 5) {
                                                                            return 2;
                                                                        }
                                                                        else if (diffMVdiffVal >= 5) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                    else if (posBalance >= 58) {
                                                                        return 1;
                                                                    }
                                                                }
                                                                else if (CsDiff >= 7) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                        else if (mcTcor >= 98) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (Rs >= 2315) {
                                                        if (CsDiff < 7) {
                                                            if (refDCval < 79) {
                                                                return 4;
                                                            }
                                                            else if (refDCval >= 79) {
                                                                if (Cs < 2079) {
                                                                    return 1;
                                                                }
                                                                else if (Cs >= 2079) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                        else if (CsDiff >= 7) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                            else if (refDCval >= 98) {
                                if (tcor < 83) {
                                    if (negBalance < 46) {
                                        if (ssDCval < 140) {
                                            if (diffMVdiffVal < -6) {
                                                return 4;
                                            }
                                            else if (diffMVdiffVal >= -6) {
                                                return 1;
                                            }
                                        }
                                        else if (ssDCval >= 140) {
                                            if (ssDCval < 141) {
                                                return 8;
                                            }
                                            else if (ssDCval >= 141) {
                                                if (MVDiff < 11) {
                                                    return 4;
                                                }
                                                else if (MVDiff >= 11) {
                                                    if (SC < 2163) {
                                                        return 4;
                                                    }
                                                    else if (SC >= 2163) {
                                                        if (refDCval < 146) {
                                                            if (CsDiff < 7) {
                                                                return 4;
                                                            }
                                                            else if (CsDiff >= 7) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (refDCval >= 146) {
                                                            return 1;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (negBalance >= 46) {
                                        if (RsCsDiff < 7) {
                                            return 8;
                                        }
                                        else if (RsCsDiff >= 7) {
                                            if (SC < 4062) {
                                                if (refDCval < 99) {
                                                    if (RsCsDiff < 29) {
                                                        if (Cs < 1064) {
                                                            if (mcTcor < 84) {
                                                                return 2;
                                                            }
                                                            else if (mcTcor >= 84) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (Cs >= 1064) {
                                                            return 2;
                                                        }
                                                    }
                                                    else if (RsCsDiff >= 29) {
                                                        if (AFD < 11) {
                                                            return 8;
                                                        }
                                                        else if (AFD >= 11) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                                else if (refDCval >= 99) {
                                                    return 4;
                                                }
                                            }
                                            else if (SC >= 4062) {
                                                if (refDCval < 119) {
                                                    if (MVDiff < 8) {
                                                        if (ssDCval < 109) {
                                                            return 4;
                                                        }
                                                        else if (ssDCval >= 109) {
                                                            if (CsDiff < 7) {
                                                                if (ssDCval < 117) {
                                                                    return 4;
                                                                }
                                                                else if (ssDCval >= 117) {
                                                                    if (Rs < 2589) {
                                                                        return 4;
                                                                    }
                                                                    else if (Rs >= 2589) {
                                                                        return 8;
                                                                    }
                                                                }
                                                            }
                                                            else if (CsDiff >= 7) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                    else if (MVDiff >= 8) {
                                                        return 4;
                                                    }
                                                }
                                                else if (refDCval >= 119) {
                                                    return 4;
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (tcor >= 83) {
                                    if (SC < 3899) {
                                        if (mu_mv_mag_sq < 38) {
                                            if (posBalance < 45) {
                                                if (ssDCval < 127) {
                                                    return 8;
                                                }
                                                else if (ssDCval >= 127) {
                                                    return 2;
                                                }
                                            }
                                            else if (posBalance >= 45) {
                                                if (TSC < 7) {
                                                    if (ssDCval < 105) {
                                                        return 2;
                                                    }
                                                    else if (ssDCval >= 105) {
                                                        if (ssDCval < 127) {
                                                            return 4;
                                                        }
                                                        else if (ssDCval >= 127) {
                                                            if (MVDiff < 33) {
                                                                return 1;
                                                            }
                                                            else if (MVDiff >= 33) {
                                                                return 2;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (TSC >= 7) {
                                                    if (Cs < 306) {
                                                        return 1;
                                                    }
                                                    else if (Cs >= 306) {
                                                        if (ssDCval < 110) {
                                                            if (RsDiff < 4) {
                                                                return 8;
                                                            }
                                                            else if (RsDiff >= 4) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (ssDCval >= 110) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (mu_mv_mag_sq >= 38) {
                                            if (ssDCval < 129) {
                                                if (Cs < 382) {
                                                    if (refDCval < 105) {
                                                        return 2;
                                                    }
                                                    else if (refDCval >= 105) {
                                                        return 4;
                                                    }
                                                }
                                                else if (Cs >= 382) {
                                                    if (diffAFD < -3) {
                                                        return 4;
                                                    }
                                                    else if (diffAFD >= -3) {
                                                        if (Cs < 422) {
                                                            if (gchDC < 1) {
                                                                return 8;
                                                            }
                                                            else if (gchDC >= 1) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (Cs >= 422) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (ssDCval >= 129) {
                                                if (Cs < 990) {
                                                    if (negBalance < 12) {
                                                        return 2;
                                                    }
                                                    else if (negBalance >= 12) {
                                                        return 4;
                                                    }
                                                }
                                                else if (Cs >= 990) {
                                                    if (posBalance < 47) {
                                                        return 8;
                                                    }
                                                    else if (posBalance >= 47) {
                                                        if (SC < 3179) {
                                                            return 2;
                                                        }
                                                        else if (SC >= 3179) {
                                                            return 1;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (SC >= 3899) {
                                        if (ssDCval < 106) {
                                            if (Rs < 2886) {
                                                if (mcTcor < 94) {
                                                    if (Cs < 2394) {
                                                        return 1;
                                                    }
                                                    else if (Cs >= 2394) {
                                                        return 8;
                                                    }
                                                }
                                                else if (mcTcor >= 94) {
                                                    if (SC < 4913) {
                                                        return 8;
                                                    }
                                                    else if (SC >= 4913) {
                                                        return 2;
                                                    }
                                                }
                                            }
                                            else if (Rs >= 2886) {
                                                if (Rs < 3159) {
                                                    return 8;
                                                }
                                                else if (Rs >= 3159) {
                                                    if (RsCsDiff < 47) {
                                                        if (Rs < 3183) {
                                                            return 4;
                                                        }
                                                        else if (Rs >= 3183) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (RsCsDiff >= 47) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                        }
                                        else if (ssDCval >= 106) {
                                            if (Cs < 2274) {
                                                if (Cs < 1946) {
                                                    if (mu_mv_mag_sq < 74) {
                                                        if (Cs < 1115) {
                                                            return 2;
                                                        }
                                                        else if (Cs >= 1115) {
                                                            if (refDCval < 106) {
                                                                if (Rs < 3173) {
                                                                    return 8;
                                                                }
                                                                else if (Rs >= 3173) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (refDCval >= 106) {
                                                                if (CsDiff < 7) {
                                                                    return 4;
                                                                }
                                                                else if (CsDiff >= 7) {
                                                                    if (MVDiff < 5) {
                                                                        if (ssDCval < 123) {
                                                                            return 4;
                                                                        }
                                                                        else if (ssDCval >= 123) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                    else if (MVDiff >= 5) {
                                                                        return 4;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (mu_mv_mag_sq >= 74) {
                                                        if (ssDCval < 117) {
                                                            return 4;
                                                        }
                                                        else if (ssDCval >= 117) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                                else if (Cs >= 1946) {
                                                    if (ssDCval < 112) {
                                                        if (SC < 5180) {
                                                            return 2;
                                                        }
                                                        else if (SC >= 5180) {
                                                            if (Cs < 2062) {
                                                                return 4;
                                                            }
                                                            else if (Cs >= 2062) {
                                                                return 2;
                                                            }
                                                        }
                                                    }
                                                    else if (ssDCval >= 112) {
                                                        if (Rs < 2633) {
                                                            if (MVDiff < 7) {
                                                                return 8;
                                                            }
                                                            else if (MVDiff >= 7) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (Rs >= 2633) {
                                                            if (MVDiff < 4) {
                                                                return 4;
                                                            }
                                                            else if (MVDiff >= 4) {
                                                                if (AFD < 10) {
                                                                    if (Cs < 2093) {
                                                                        return 4;
                                                                    }
                                                                    else if (Cs >= 2093) {
                                                                        return 8;
                                                                    }
                                                                }
                                                                else if (AFD >= 10) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                            else if (Cs >= 2274) {
                                                if (negBalance < 39) {
                                                    return 4;
                                                }
                                                else if (negBalance >= 39) {
                                                    return 8;
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                        }
                    }
                    else if (AFD >= 12) {
                        if (Cs < 1451) {
                            if (RsDiff < 5) {
                                if (Cs < 533) {
                                    if (tcor < 47) {
                                        if (diffRsCsdiff < 41) {
                                            return 1;
                                        }
                                        else if (diffRsCsdiff >= 41) {
                                            return 8;
                                        }
                                    }
                                    else if (tcor >= 47) {
                                        if (tcor < 75) {
                                            if (TSC < 12) {
                                                return 2;
                                            }
                                            else if (TSC >= 12) {
                                                return 1;
                                            }
                                        }
                                        else if (tcor >= 75) {
                                            if (TSC < 7) {
                                                if (Cs < 75) {
                                                    return 1;
                                                }
                                                else if (Cs >= 75) {
                                                    if (MVDiff < 39) {
                                                        return 2;
                                                    }
                                                    else if (MVDiff >= 39) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                            else if (TSC >= 7) {
                                                return 1;
                                            }
                                        }
                                    }
                                }
                                else if (Cs >= 533) {
                                    if (MVDiff < 15) {
                                        if (gchDC < 2) {
                                            return 4;
                                        }
                                        else if (gchDC >= 2) {
                                            if (Rs < 1104) {
                                                return 2;
                                            }
                                            else if (Rs >= 1104) {
                                                if (TSC < 6) {
                                                    return 8;
                                                }
                                                else if (TSC >= 6) {
                                                    return 4;
                                                }
                                            }
                                        }
                                    }
                                    else if (MVDiff >= 15) {
                                        if (mu_mv_mag_sq < 35) {
                                            if (Rs < 552) {
                                                if (tcor < 60) {
                                                    return 1;
                                                }
                                                else if (tcor >= 60) {
                                                    return 2;
                                                }
                                            }
                                            else if (Rs >= 552) {
                                                if (MVDiff < 138) {
                                                    if (Rs < 1377) {
                                                        if (SC < 1430) {
                                                            if (Cs < 646) {
                                                                return 8;
                                                            }
                                                            else if (Cs >= 646) {
                                                                return 1;
                                                            }
                                                        }
                                                        else if (SC >= 1430) {
                                                            return 1;
                                                        }
                                                    }
                                                    else if (Rs >= 1377) {
                                                        return 8;
                                                    }
                                                }
                                                else if (MVDiff >= 138) {
                                                    if (diffAFD < -7) {
                                                        return 4;
                                                    }
                                                    else if (diffAFD >= -7) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                        }
                                        else if (mu_mv_mag_sq >= 35) {
                                            if (posBalance < 45) {
                                                if (Rs < 716) {
                                                    return 4;
                                                }
                                                else if (Rs >= 716) {
                                                    return 8;
                                                }
                                            }
                                            else if (posBalance >= 45) {
                                                return 4;
                                            }
                                        }
                                    }
                                }
                            }
                            else if (RsDiff >= 5) {
                                if (mcTcor < 85) {
                                    if (ssDCval < 74) {
                                        if (CsDiff < 7) {
                                            if (Rs < 887) {
                                                return 2;
                                            }
                                            else if (Rs >= 887) {
                                                return 8;
                                            }
                                        }
                                        else if (CsDiff >= 7) {
                                            if (Rs < 1358) {
                                                if (mcTcor < 70) {
                                                    if (MVDiff < 174) {
                                                        return 8;
                                                    }
                                                    else if (MVDiff >= 174) {
                                                        if (Rs < 600) {
                                                            return 4;
                                                        }
                                                        else if (Rs >= 600) {
                                                            return 2;
                                                        }
                                                    }
                                                }
                                                else if (mcTcor >= 70) {
                                                    return 1;
                                                }
                                            }
                                            else if (Rs >= 1358) {
                                                return 4;
                                            }
                                        }
                                    }
                                    else if (ssDCval >= 74) {
                                        if (TSC < 8) {
                                            return 8;
                                        }
                                        else if (TSC >= 8) {
                                            if (AFD < 30) {
                                                if (tcor < 75) {
                                                    if (posBalance < 47) {
                                                        return 2;
                                                    }
                                                    else if (posBalance >= 47) {
                                                        if (CsDiff < 5) {
                                                            if (refDCval < 103) {
                                                                if (diffMVdiffVal < -56) {
                                                                    return 2;
                                                                }
                                                                else if (diffMVdiffVal >= -56) {
                                                                    return 1;
                                                                }
                                                            }
                                                            else if (refDCval >= 103) {
                                                                if (RsDiff < 12) {
                                                                    if (SC < 3031) {
                                                                        return 2;
                                                                    }
                                                                    else if (SC >= 3031) {
                                                                        return 1;
                                                                    }
                                                                }
                                                                else if (RsDiff >= 12) {
                                                                    return 2;
                                                                }
                                                            }
                                                        }
                                                        else if (CsDiff >= 5) {
                                                            return 1;
                                                        }
                                                    }
                                                }
                                                else if (tcor >= 75) {
                                                    if (ssDCval < 121) {
                                                        return 2;
                                                    }
                                                    else if (ssDCval >= 121) {
                                                        return 1;
                                                    }
                                                }
                                            }
                                            else if (AFD >= 30) {
                                                if (diffRsCsdiff < 290) {
                                                    if (negBalance < 86) {
                                                        if (RsDiff < 11) {
                                                            if (Cs < 384) {
                                                                return 2;
                                                            }
                                                            else if (Cs >= 384) {
                                                                return 1;
                                                            }
                                                        }
                                                        else if (RsDiff >= 11) {
                                                            return 2;
                                                        }
                                                    }
                                                    else if (negBalance >= 86) {
                                                        return 1;
                                                    }
                                                }
                                                else if (diffRsCsdiff >= 290) {
                                                    if (diffTSC < 21) {
                                                        if (refDCval < 102) {
                                                            return 8;
                                                        }
                                                        else if (refDCval >= 102) {
                                                            return 1;
                                                        }
                                                    }
                                                    else if (diffTSC >= 21) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (mcTcor >= 85) {
                                    if (posBalance < 73) {
                                        if (negBalance < 70) {
                                            if (MVDiff < 10) {
                                                if (mu_mv_mag_sq < 3) {
                                                    if (Rs < 1210) {
                                                        return 4;
                                                    }
                                                    else if (Rs >= 1210) {
                                                        return 1;
                                                    }
                                                }
                                                else if (mu_mv_mag_sq >= 3) {
                                                    if (Cs < 1340) {
                                                        if (RsCsDiff < 79) {
                                                            if (mu_mv_mag_sq < 21) {
                                                                if (Cs < 513) {
                                                                    return 2;
                                                                }
                                                                else if (Cs >= 513) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (mu_mv_mag_sq >= 21) {
                                                                if (Cs < 848) {
                                                                    return 2;
                                                                }
                                                                else if (Cs >= 848) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                        else if (RsCsDiff >= 79) {
                                                            return 2;
                                                        }
                                                    }
                                                    else if (Cs >= 1340) {
                                                        return 1;
                                                    }
                                                }
                                            }
                                            else if (MVDiff >= 10) {
                                                if (mu_mv_mag_sq < 70) {
                                                    if (AFD < 14) {
                                                        if (tcor < 80) {
                                                            if (Rs < 810) {
                                                                return 2;
                                                            }
                                                            else if (Rs >= 810) {
                                                                if (ssDCval < 54) {
                                                                    return 2;
                                                                }
                                                                else if (ssDCval >= 54) {
                                                                    return 1;
                                                                }
                                                            }
                                                        }
                                                        else if (tcor >= 80) {
                                                            if (Cs < 1036) {
                                                                return 2;
                                                            }
                                                            else if (Cs >= 1036) {
                                                                if (RsCsDiff < 79) {
                                                                    return 1;
                                                                }
                                                                else if (RsCsDiff >= 79) {
                                                                    return 2;
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (AFD >= 14) {
                                                        if (tcor < 64) {
                                                            if (Rs < 2452) {
                                                                if (mcTcor < 87) {
                                                                    return 1;
                                                                }
                                                                else if (mcTcor >= 87) {
                                                                    if (Rs < 1345) {
                                                                        if (CsDiff < 9) {
                                                                            return 2;
                                                                        }
                                                                        else if (CsDiff >= 9) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                    else if (Rs >= 1345) {
                                                                        return 2;
                                                                    }
                                                                }
                                                            }
                                                            else if (Rs >= 2452) {
                                                                if (mu_mv_mag_sq < 17) {
                                                                    return 4;
                                                                }
                                                                else if (mu_mv_mag_sq >= 17) {
                                                                    return 2;
                                                                }
                                                            }
                                                        }
                                                        else if (tcor >= 64) {
                                                            if (refDCval < 73) {
                                                                if (Cs < 709) {
                                                                    return 2;
                                                                }
                                                                else if (Cs >= 709) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (refDCval >= 73) {
                                                                if (ssDCval < 117) {
                                                                    if (negBalance < 35) {
                                                                        return 4;
                                                                    }
                                                                    else if (negBalance >= 35) {
                                                                        return 2;
                                                                    }
                                                                }
                                                                else if (ssDCval >= 117) {
                                                                    if (SC < 3240) {
                                                                        if (CsDiff < 3) {
                                                                            return 4;
                                                                        }
                                                                        else if (CsDiff >= 3) {
                                                                            if (diffAFD < 5) {
                                                                                if (Cs < 1379) {
                                                                                    return 2;
                                                                                }
                                                                                else if (Cs >= 1379) {
                                                                                    return 4;
                                                                                }
                                                                            }
                                                                            else if (diffAFD >= 5) {
                                                                                return 1;
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (SC >= 3240) {
                                                                        return 2;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (mu_mv_mag_sq >= 70) {
                                                    return 8;
                                                }
                                            }
                                        }
                                        else if (negBalance >= 70) {
                                            if (CsDiff < 6) {
                                                if (Cs < 960) {
                                                    return 4;
                                                }
                                                else if (Cs >= 960) {
                                                    return 8;
                                                }
                                            }
                                            else if (CsDiff >= 6) {
                                                return 1;
                                            }
                                        }
                                    }
                                    else if (posBalance >= 73) {
                                        if (Rs < 396) {
                                            return 1;
                                        }
                                        else if (Rs >= 396) {
                                            if (diffMVdiffVal < -4) {
                                                return 4;
                                            }
                                            else if (diffMVdiffVal >= -4) {
                                                return 8;
                                            }
                                        }
                                    }
                                }
                            }
                        }
                        else if (Cs >= 1451) {
                            if (mu_mv_mag_sq < 14) {
                                if (negBalance < 51) {
                                    if (MVDiff < 4) {
                                        if (Rs < 2603) {
                                            if (AFD < 17) {
                                                return 4;
                                            }
                                            else if (AFD >= 17) {
                                                return 2;
                                            }
                                        }
                                        else if (Rs >= 2603) {
                                            if (SC < 5181) {
                                                return 2;
                                            }
                                            else if (SC >= 5181) {
                                                return 8;
                                            }
                                        }
                                    }
                                    else if (MVDiff >= 4) {
                                        if (posBalance < 59) {
                                            if (MVDiff < 18) {
                                                if (ssDCval < 80) {
                                                    if (ssDCval < 50) {
                                                        return 4;
                                                    }
                                                    else if (ssDCval >= 50) {
                                                        if (Rs < 1874) {
                                                            return 1;
                                                        }
                                                        else if (Rs >= 1874) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                                else if (ssDCval >= 80) {
                                                    if (mcTcor < 88) {
                                                        return 1;
                                                    }
                                                    else if (mcTcor >= 88) {
                                                        if (ssDCval < 100) {
                                                            if (negBalance < 46) {
                                                                return 1;
                                                            }
                                                            else if (negBalance >= 46) {
                                                                if (Rs < 2716) {
                                                                    if (SC < 4274) {
                                                                        return 1;
                                                                    }
                                                                    else if (SC >= 4274) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (Rs >= 2716) {
                                                                    return 1;
                                                                }
                                                            }
                                                        }
                                                        else if (ssDCval >= 100) {
                                                            if (TSCindex < 9) {
                                                                if (Cs < 1822) {
                                                                    return 4;
                                                                }
                                                                else if (Cs >= 1822) {
                                                                    return 2;
                                                                }
                                                            }
                                                            else if (TSCindex >= 9) {
                                                                if (diffRsCsdiff < -3) {
                                                                    if (Cs < 2250) {
                                                                        return 2;
                                                                    }
                                                                    else if (Cs >= 2250) {
                                                                        return 8;
                                                                    }
                                                                }
                                                                else if (diffRsCsdiff >= -3) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                            else if (MVDiff >= 18) {
                                                if (CsDiff < 10) {
                                                    if (mcTcor < 92) {
                                                        if (refDCval < 85) {
                                                            if (Rs < 1898) {
                                                                return 4;
                                                            }
                                                            else if (Rs >= 1898) {
                                                                if (posBalance < 56) {
                                                                    if (MVDiff < 35) {
                                                                        return 8;
                                                                    }
                                                                    else if (MVDiff >= 35) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (posBalance >= 56) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                        else if (refDCval >= 85) {
                                                            if (Rs < 2443) {
                                                                return 4;
                                                            }
                                                            else if (Rs >= 2443) {
                                                                if (Cs < 2103) {
                                                                    return 8;
                                                                }
                                                                else if (Cs >= 2103) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (mcTcor >= 92) {
                                                        if (ssDCval < 84) {
                                                            return 1;
                                                        }
                                                        else if (ssDCval >= 84) {
                                                            if (ssDCval < 121) {
                                                                if (diffTSC < 2) {
                                                                    if (ssDCval < 88) {
                                                                        if (Rs < 2291) {
                                                                            return 8;
                                                                        }
                                                                        else if (Rs >= 2291) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                    else if (ssDCval >= 88) {
                                                                        return 8;
                                                                    }
                                                                }
                                                                else if (diffTSC >= 2) {
                                                                    return 1;
                                                                }
                                                            }
                                                            else if (ssDCval >= 121) {
                                                                if (Rs < 2137) {
                                                                    return 4;
                                                                }
                                                                else if (Rs >= 2137) {
                                                                    return 2;
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (CsDiff >= 10) {
                                                    if (refDCval < 93) {
                                                        if (tcor < 78) {
                                                            if (ssDCval < 78) {
                                                                if (Rs < 2193) {
                                                                    return 2;
                                                                }
                                                                else if (Rs >= 2193) {
                                                                    return 1;
                                                                }
                                                            }
                                                            else if (ssDCval >= 78) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (tcor >= 78) {
                                                            if (refDCval < 81) {
                                                                if (Rs < 3064) {
                                                                    if (ssDCval < 80) {
                                                                        if (diffRsCsdiff < 32) {
                                                                            if (SC < 5658) {
                                                                                return 1;
                                                                            }
                                                                            else if (SC >= 5658) {
                                                                                return 8;
                                                                            }
                                                                        }
                                                                        else if (diffRsCsdiff >= 32) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                    else if (ssDCval >= 80) {
                                                                        return 8;
                                                                    }
                                                                }
                                                                else if (Rs >= 3064) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (refDCval >= 81) {
                                                                if (RsDiff < 10) {
                                                                    if (posBalance < 52) {
                                                                        return 1;
                                                                    }
                                                                    else if (posBalance >= 52) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (RsDiff >= 10) {
                                                                    return 1;
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (refDCval >= 93) {
                                                        if (Rs < 2545) {
                                                            if (negBalance < 40) {
                                                                return 8;
                                                            }
                                                            else if (negBalance >= 40) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (Rs >= 2545) {
                                                            if (Cs < 2680) {
                                                                return 1;
                                                            }
                                                            else if (Cs >= 2680) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (posBalance >= 59) {
                                            if (tcor < 73) {
                                                if (diffMVdiffVal < 10) {
                                                    if (posBalance < 69) {
                                                        return 1;
                                                    }
                                                    else if (posBalance >= 69) {
                                                        return 4;
                                                    }
                                                }
                                                else if (diffMVdiffVal >= 10) {
                                                    return 8;
                                                }
                                            }
                                            else if (tcor >= 73) {
                                                if (SC < 4294) {
                                                    if (CsDiff < 10) {
                                                        if (ssDCval < 84) {
                                                            if (TSC < 10) {
                                                                return 2;
                                                            }
                                                            else if (TSC >= 10) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (ssDCval >= 84) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (CsDiff >= 10) {
                                                        return 2;
                                                    }
                                                }
                                                else if (SC >= 4294) {
                                                    if (diffRsCsdiff < 22) {
                                                        if (CsDiff < 10) {
                                                            if (tcor < 76) {
                                                                return 1;
                                                            }
                                                            else if (tcor >= 76) {
                                                                return 2;
                                                            }
                                                        }
                                                        else if (CsDiff >= 10) {
                                                            if (Rs < 2405) {
                                                                return 2;
                                                            }
                                                            else if (Rs >= 2405) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                    else if (diffRsCsdiff >= 22) {
                                                        if (ssDCval < 86) {
                                                            return 4;
                                                        }
                                                        else if (ssDCval >= 86) {
                                                            if (diffAFD < -1) {
                                                                return 2;
                                                            }
                                                            else if (diffAFD >= -1) {
                                                                if (SC < 5233) {
                                                                    if (negBalance < 49) {
                                                                        return 1;
                                                                    }
                                                                    else if (negBalance >= 49) {
                                                                        return 2;
                                                                    }
                                                                }
                                                                else if (SC >= 5233) {
                                                                    return 2;
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (negBalance >= 51) {
                                    if (tcor < 80) {
                                        if (mu_mv_mag_sq < 2) {
                                            if (negBalance < 54) {
                                                if (SC < 4795) {
                                                    if (TSC < 12) {
                                                        return 4;
                                                    }
                                                    else if (TSC >= 12) {
                                                        if (Rs < 2586) {
                                                            return 8;
                                                        }
                                                        else if (Rs >= 2586) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                                else if (SC >= 4795) {
                                                    return 1;
                                                }
                                            }
                                            else if (negBalance >= 54) {
                                                if (Cs < 2432) {
                                                    if (mcTcor < 88) {
                                                        if (RsCsDiff < 326) {
                                                            if (Rs < 2807) {
                                                                if (tcor < 68) {
                                                                    if (mcTcor < 79) {
                                                                        return 1;
                                                                    }
                                                                    else if (mcTcor >= 79) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (tcor >= 68) {
                                                                    if (CsDiff < 6) {
                                                                        return 8;
                                                                    }
                                                                    else if (CsDiff >= 6) {
                                                                        if (negBalance < 55) {
                                                                            if (Cs < 2107) {
                                                                                return 4;
                                                                            }
                                                                            else if (Cs >= 2107) {
                                                                                return 8;
                                                                            }
                                                                        }
                                                                        else if (negBalance >= 55) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                            else if (Rs >= 2807) {
                                                                return 1;
                                                            }
                                                        }
                                                        else if (RsCsDiff >= 326) {
                                                            if (TSC < 16) {
                                                                return 2;
                                                            }
                                                            else if (TSC >= 16) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                    else if (mcTcor >= 88) {
                                                        if (Rs < 2604) {
                                                            if (CsDiff < 11) {
                                                                return 1;
                                                            }
                                                            else if (CsDiff >= 11) {
                                                                return 2;
                                                            }
                                                        }
                                                        else if (Rs >= 2604) {
                                                            return 2;
                                                        }
                                                    }
                                                }
                                                else if (Cs >= 2432) {
                                                    return 4;
                                                }
                                            }
                                        }
                                        else if (mu_mv_mag_sq >= 2) {
                                            if (MVDiff < 19) {
                                                if (TSC < 11) {
                                                    if (refDCval < 119) {
                                                        if (RsCsDiff < 243) {
                                                            if (refDCval < 95) {
                                                                return 4;
                                                            }
                                                            else if (refDCval >= 95) {
                                                                return 2;
                                                            }
                                                        }
                                                        else if (RsCsDiff >= 243) {
                                                            return 2;
                                                        }
                                                    }
                                                    else if (refDCval >= 119) {
                                                        return 1;
                                                    }
                                                }
                                                else if (TSC >= 11) {
                                                    if (posBalance < 66) {
                                                        return 1;
                                                    }
                                                    else if (posBalance >= 66) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                            else if (MVDiff >= 19) {
                                                if (posBalance < 57) {
                                                    if (diffRsCsdiff < -34) {
                                                        if (AFD < 27) {
                                                            return 1;
                                                        }
                                                        else if (AFD >= 27) {
                                                            return 2;
                                                        }
                                                    }
                                                    else if (diffRsCsdiff >= -34) {
                                                        if (TSC < 13) {
                                                            if (Rs < 2033) {
                                                                if (SC < 3255) {
                                                                    return 4;
                                                                }
                                                                else if (SC >= 3255) {
                                                                    return 1;
                                                                }
                                                            }
                                                            else if (Rs >= 2033) {
                                                                if (Rs < 2605) {
                                                                    return 2;
                                                                }
                                                                else if (Rs >= 2605) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                        else if (TSC >= 13) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                                else if (posBalance >= 57) {
                                                    if (ssDCval < 130) {
                                                        if (mcTcor < 69) {
                                                            if (Rs < 2531) {
                                                                return 4;
                                                            }
                                                            else if (Rs >= 2531) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (mcTcor >= 69) {
                                                            if (AFD < 22) {
                                                                if (negBalance < 54) {
                                                                    return 4;
                                                                }
                                                                else if (negBalance >= 54) {
                                                                    if (SC < 4591) {
                                                                        return 8;
                                                                    }
                                                                    else if (SC >= 4591) {
                                                                        return 1;
                                                                    }
                                                                }
                                                            }
                                                            else if (AFD >= 22) {
                                                                if (diffTSC < -6) {
                                                                    return 8;
                                                                }
                                                                else if (diffTSC >= -6) {
                                                                    return 1;
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (ssDCval >= 130) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (tcor >= 80) {
                                        if (CsDiff < 7) {
                                            if (ssDCval < 107) {
                                                if (posBalance < 58) {
                                                    if (TSC < 10) {
                                                        return 1;
                                                    }
                                                    else if (TSC >= 10) {
                                                        return 4;
                                                    }
                                                }
                                                else if (posBalance >= 58) {
                                                    return 2;
                                                }
                                            }
                                            else if (ssDCval >= 107) {
                                                if (MVDiff < 26) {
                                                    return 4;
                                                }
                                                else if (MVDiff >= 26) {
                                                    if (MVDiff < 30) {
                                                        return 8;
                                                    }
                                                    else if (MVDiff >= 30) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                        }
                                        else if (CsDiff >= 7) {
                                            if (ssDCval < 80) {
                                                if (Cs < 2057) {
                                                    return 8;
                                                }
                                                else if (Cs >= 2057) {
                                                    return 4;
                                                }
                                            }
                                            else if (ssDCval >= 80) {
                                                if (SC < 5473) {
                                                    if (SC < 4400) {
                                                        if (AFD < 23) {
                                                            if (tcor < 81) {
                                                                if (diffRsCsdiff < 19) {
                                                                    if (Cs < 1795) {
                                                                        return 4;
                                                                    }
                                                                    else if (Cs >= 1795) {
                                                                        return 2;
                                                                    }
                                                                }
                                                                else if (diffRsCsdiff >= 19) {
                                                                    return 2;
                                                                }
                                                            }
                                                            else if (tcor >= 81) {
                                                                if (CsDiff < 9) {
                                                                    if (TSC < 11) {
                                                                        if (Rs < 2510) {
                                                                            return 2;
                                                                        }
                                                                        else if (Rs >= 2510) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                    else if (TSC >= 11) {
                                                                        if (Rs < 2351) {
                                                                            return 1;
                                                                        }
                                                                        else if (Rs >= 2351) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                }
                                                                else if (CsDiff >= 9) {
                                                                    return 2;
                                                                }
                                                            }
                                                        }
                                                        else if (AFD >= 23) {
                                                            if (Cs < 1900) {
                                                                return 2;
                                                            }
                                                            else if (Cs >= 1900) {
                                                                return 1;
                                                            }
                                                        }
                                                    }
                                                    else if (SC >= 4400) {
                                                        if (tcor < 82) {
                                                            if (ssDCval < 96) {
                                                                if (ssDCval < 88) {
                                                                    return 4;
                                                                }
                                                                else if (ssDCval >= 88) {
                                                                    return 2;
                                                                }
                                                            }
                                                            else if (ssDCval >= 96) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (tcor >= 82) {
                                                            if (diffRsCsdiff < 18) {
                                                                if (diffMVdiffVal < -6) {
                                                                    return 1;
                                                                }
                                                                else if (diffMVdiffVal >= -6) {
                                                                    if (ssDCval < 84) {
                                                                        return 4;
                                                                    }
                                                                    else if (ssDCval >= 84) {
                                                                        if (diffRsCsdiff < 16) {
                                                                            if (posBalance < 55) {
                                                                                if (SC < 5185) {
                                                                                    return 8;
                                                                                }
                                                                                else if (SC >= 5185) {
                                                                                    return 2;
                                                                                }
                                                                            }
                                                                            else if (posBalance >= 55) {
                                                                                return 2;
                                                                            }
                                                                        }
                                                                        else if (diffRsCsdiff >= 16) {
                                                                            if (Rs < 2661) {
                                                                                return 2;
                                                                            }
                                                                            else if (Rs >= 2661) {
                                                                                return 8;
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                            else if (diffRsCsdiff >= 18) {
                                                                if (posBalance < 56) {
                                                                    return 4;
                                                                }
                                                                else if (posBalance >= 56) {
                                                                    if (SC < 4530) {
                                                                        return 4;
                                                                    }
                                                                    else if (SC >= 4530) {
                                                                        return 2;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (SC >= 5473) {
                                                    if (SC < 5599) {
                                                        return 1;
                                                    }
                                                    else if (SC >= 5599) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                            else if (mu_mv_mag_sq >= 14) {
                                if (ssDCval < 68) {
                                    if (negBalance < 44) {
                                        return 1;
                                    }
                                    else if (negBalance >= 44) {
                                        return 8;
                                    }
                                }
                                else if (ssDCval >= 68) {
                                    if (Cs < 1801) {
                                        if (SC < 3682) {
                                            if (RsDiff < 12) {
                                                if (ssDCval < 75) {
                                                    if (ssDCval < 71) {
                                                        return 4;
                                                    }
                                                    else if (ssDCval >= 71) {
                                                        return 8;
                                                    }
                                                }
                                                else if (ssDCval >= 75) {
                                                    if (diffRsCsdiff < 2) {
                                                        if (Rs < 1390) {
                                                            return 1;
                                                        }
                                                        else if (Rs >= 1390) {
                                                            return 2;
                                                        }
                                                    }
                                                    else if (diffRsCsdiff >= 2) {
                                                        if (AFD < 27) {
                                                            return 4;
                                                        }
                                                        else if (AFD >= 27) {
                                                            return 2;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (RsDiff >= 12) {
                                                if (Rs < 1496) {
                                                    return 8;
                                                }
                                                else if (Rs >= 1496) {
                                                    return 1;
                                                }
                                            }
                                        }
                                        else if (SC >= 3682) {
                                            if (RsCsDiff < 180) {
                                                if (RsDiff < 7) {
                                                    return 2;
                                                }
                                                else if (RsDiff >= 7) {
                                                    if (AFD < 14) {
                                                        if (Rs < 2721) {
                                                            return 1;
                                                        }
                                                        else if (Rs >= 2721) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (AFD >= 14) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                            else if (RsCsDiff >= 180) {
                                                if (negBalance < 54) {
                                                    if (diffRsCsdiff < -32) {
                                                        return 4;
                                                    }
                                                    else if (diffRsCsdiff >= -32) {
                                                        if (ssDCval < 95) {
                                                            return 4;
                                                        }
                                                        else if (ssDCval >= 95) {
                                                            return 2;
                                                        }
                                                    }
                                                }
                                                else if (negBalance >= 54) {
                                                    if (CsDiff < 7) {
                                                        return 8;
                                                    }
                                                    else if (CsDiff >= 7) {
                                                        return 2;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (Cs >= 1801) {
                                        if (tcor < 81) {
                                            if (MVDiff < 69) {
                                                if (ssDCval < 141) {
                                                    if (RsCsDiff < 151) {
                                                        if (Rs < 2455) {
                                                            if (AFD < 13) {
                                                                return 4;
                                                            }
                                                            else if (AFD >= 13) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (Rs >= 2455) {
                                                            if (CsDiff < 7) {
                                                                return 8;
                                                            }
                                                            else if (CsDiff >= 7) {
                                                                if (diffMVdiffVal < 1) {
                                                                    return 1;
                                                                }
                                                                else if (diffMVdiffVal >= 1) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (RsCsDiff >= 151) {
                                                        if (mcTcor < 87) {
                                                            if (RsDiff < 18) {
                                                                if (RsCsDiff < 367) {
                                                                    if (MVDiff < 6) {
                                                                        return 4;
                                                                    }
                                                                    else if (MVDiff >= 6) {
                                                                        if (SC < 6670) {
                                                                            if (mu_mv_mag_sq < 16) {
                                                                                return 4;
                                                                            }
                                                                            else if (mu_mv_mag_sq >= 16) {
                                                                                if (CsDiff < 10) {
                                                                                    if (SC < 4682) {
                                                                                        return 8;
                                                                                    }
                                                                                    else if (SC >= 4682) {
                                                                                        return 1;
                                                                                    }
                                                                                }
                                                                                else if (CsDiff >= 10) {
                                                                                    return 1;
                                                                                }
                                                                            }
                                                                        }
                                                                        else if (SC >= 6670) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                }
                                                                else if (RsCsDiff >= 367) {
                                                                    if (Cs < 3074) {
                                                                        if (Rs < 3074) {
                                                                            if (ssDCval < 134) {
                                                                                return 8;
                                                                            }
                                                                            else if (ssDCval >= 134) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                        else if (Rs >= 3074) {
                                                                            if (posBalance < 55) {
                                                                                return 8;
                                                                            }
                                                                            else if (posBalance >= 55) {
                                                                                return 1;
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (Cs >= 3074) {
                                                                        if (MVDiff < 37) {
                                                                            return 1;
                                                                        }
                                                                        else if (MVDiff >= 37) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                            else if (RsDiff >= 18) {
                                                                return 2;
                                                            }
                                                        }
                                                        else if (mcTcor >= 87) {
                                                            if (diffRsCsdiff < -125) {
                                                                if (diffTSC < -1) {
                                                                    return 1;
                                                                }
                                                                else if (diffTSC >= -1) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (diffRsCsdiff >= -125) {
                                                                if (mu_mv_mag_sq < 52) {
                                                                    return 1;
                                                                }
                                                                else if (mu_mv_mag_sq >= 52) {
                                                                    if (TSC < 12) {
                                                                        return 1;
                                                                    }
                                                                    else if (TSC >= 12) {
                                                                        return 4;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (ssDCval >= 141) {
                                                    if (AFD < 18) {
                                                        return 8;
                                                    }
                                                    else if (AFD >= 18) {
                                                        if (Cs < 1957) {
                                                            return 4;
                                                        }
                                                        else if (Cs >= 1957) {
                                                            if (SC < 4973) {
                                                                return 8;
                                                            }
                                                            else if (SC >= 4973) {
                                                                return 2;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                            else if (MVDiff >= 69) {
                                                if (ssDCval < 141) {
                                                    if (Rs < 3192) {
                                                        return 8;
                                                    }
                                                    else if (Rs >= 3192) {
                                                        return 4;
                                                    }
                                                }
                                                else if (ssDCval >= 141) {
                                                    return 4;
                                                }
                                            }
                                        }
                                        else if (tcor >= 81) {
                                            if (refDCval < 124) {
                                                if (ssDCval < 84) {
                                                    if (negBalance < 53) {
                                                        if (mu_mv_mag_sq < 64) {
                                                            if (Rs < 2596) {
                                                                if (tcor < 90) {
                                                                    if (MVDiff < 8) {
                                                                        return 1;
                                                                    }
                                                                    else if (MVDiff >= 8) {
                                                                        if (mu_mv_mag_sq < 21) {
                                                                            return 1;
                                                                        }
                                                                        else if (mu_mv_mag_sq >= 21) {
                                                                            if (posBalance < 58) {
                                                                                return 8;
                                                                            }
                                                                            else if (posBalance >= 58) {
                                                                                return 1;
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                                else if (tcor >= 90) {
                                                                    if (Cs < 1861) {
                                                                        return 2;
                                                                    }
                                                                    else if (Cs >= 1861) {
                                                                        return 1;
                                                                    }
                                                                }
                                                            }
                                                            else if (Rs >= 2596) {
                                                                if (diffMVdiffVal < 6) {
                                                                    if (SC < 6217) {
                                                                        if (MVDiff < 19) {
                                                                            return 1;
                                                                        }
                                                                        else if (MVDiff >= 19) {
                                                                            return 2;
                                                                        }
                                                                    }
                                                                    else if (SC >= 6217) {
                                                                        return 8;
                                                                    }
                                                                }
                                                                else if (diffMVdiffVal >= 6) {
                                                                    if (SC < 5103) {
                                                                        return 2;
                                                                    }
                                                                    else if (SC >= 5103) {
                                                                        return 8;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (mu_mv_mag_sq >= 64) {
                                                            if (refDCval < 82) {
                                                                return 8;
                                                            }
                                                            else if (refDCval >= 82) {
                                                                return 1;
                                                            }
                                                        }
                                                    }
                                                    else if (negBalance >= 53) {
                                                        if (Cs < 2371) {
                                                            if (mcTcor < 92) {
                                                                return 1;
                                                            }
                                                            else if (mcTcor >= 92) {
                                                                if (Rs < 2384) {
                                                                    return 1;
                                                                }
                                                                else if (Rs >= 2384) {
                                                                    if (RsDiff < 8) {
                                                                        return 1;
                                                                    }
                                                                    else if (RsDiff >= 8) {
                                                                        return 2;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (Cs >= 2371) {
                                                            if (TSC < 12) {
                                                                return 4;
                                                            }
                                                            else if (TSC >= 12) {
                                                                return 1;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (ssDCval >= 84) {
                                                    if (Cs < 2690) {
                                                        if (AFD < 15) {
                                                            if (MVDiff < 25) {
                                                                if (posBalance < 54) {
                                                                    if (TSC < 11) {
                                                                        if (diffMVdiffVal < 3) {
                                                                            if (diffTSC < -1) {
                                                                                if (Rs < 2353) {
                                                                                    return 8;
                                                                                }
                                                                                else if (Rs >= 2353) {
                                                                                    return 1;
                                                                                }
                                                                            }
                                                                            else if (diffTSC >= -1) {
                                                                                return 1;
                                                                            }
                                                                        }
                                                                        else if (diffMVdiffVal >= 3) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                    else if (TSC >= 11) {
                                                                        if (SC < 4388) {
                                                                            return 4;
                                                                        }
                                                                        else if (SC >= 4388) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                }
                                                                else if (posBalance >= 54) {
                                                                    if (Rs < 2655) {
                                                                        if (MVDiff < 16) {
                                                                            if (diffAFD < -1) {
                                                                                return 2;
                                                                            }
                                                                            else if (diffAFD >= -1) {
                                                                                if (Rs < 2407) {
                                                                                    return 2;
                                                                                }
                                                                                else if (Rs >= 2407) {
                                                                                    return 1;
                                                                                }
                                                                            }
                                                                        }
                                                                        else if (MVDiff >= 16) {
                                                                            if (CsDiff < 7) {
                                                                                return 2;
                                                                            }
                                                                            else if (CsDiff >= 7) {
                                                                                if (Cs < 2130) {
                                                                                    return 4;
                                                                                }
                                                                                else if (Cs >= 2130) {
                                                                                    return 8;
                                                                                }
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (Rs >= 2655) {
                                                                        if (MVDiff < 13) {
                                                                            if (mu_mv_mag_sq < 53) {
                                                                                return 1;
                                                                            }
                                                                            else if (mu_mv_mag_sq >= 53) {
                                                                                if (RsDiff < 8) {
                                                                                    if (Cs < 2381) {
                                                                                        if (ssDCval < 92) {
                                                                                            if (TSC < 11) {
                                                                                                return 1;
                                                                                            }
                                                                                            else if (TSC >= 11) {
                                                                                                return 4;
                                                                                            }
                                                                                        }
                                                                                        else if (ssDCval >= 92) {
                                                                                            return 1;
                                                                                        }
                                                                                    }
                                                                                    else if (Cs >= 2381) {
                                                                                        if (SC < 5498) {
                                                                                            return 4;
                                                                                        }
                                                                                        else if (SC >= 5498) {
                                                                                            return 1;
                                                                                        }
                                                                                    }
                                                                                }
                                                                                else if (RsDiff >= 8) {
                                                                                    if (CsDiff < 8) {
                                                                                        if (refDCval < 85) {
                                                                                            return 8;
                                                                                        }
                                                                                        else if (refDCval >= 85) {
                                                                                            if (MVDiff < 5) {
                                                                                                return 8;
                                                                                            }
                                                                                            else if (MVDiff >= 5) {
                                                                                                if (ssDCval < 86) {
                                                                                                    return 4;
                                                                                                }
                                                                                                else if (ssDCval >= 86) {
                                                                                                    if (refDCval < 90) {
                                                                                                        return 8;
                                                                                                    }
                                                                                                    else if (refDCval >= 90) {
                                                                                                        return 4;
                                                                                                    }
                                                                                                }
                                                                                            }
                                                                                        }
                                                                                    }
                                                                                    else if (CsDiff >= 8) {
                                                                                        if (Cs < 2385) {
                                                                                            if (Cs < 2170) {
                                                                                                return 8;
                                                                                            }
                                                                                            else if (Cs >= 2170) {
                                                                                                return 4;
                                                                                            }
                                                                                        }
                                                                                        else if (Cs >= 2385) {
                                                                                            if (Rs < 3284) {
                                                                                                return 1;
                                                                                            }
                                                                                            else if (Rs >= 3284) {
                                                                                                return 4;
                                                                                            }
                                                                                        }
                                                                                    }
                                                                                }
                                                                            }
                                                                        }
                                                                        else if (MVDiff >= 13) {
                                                                            if (CsDiff < 6) {
                                                                                return 2;
                                                                            }
                                                                            else if (CsDiff >= 6) {
                                                                                if (diffAFD < -1) {
                                                                                    if (MVDiff < 16) {
                                                                                        return 1;
                                                                                    }
                                                                                    else if (MVDiff >= 16) {
                                                                                        return 4;
                                                                                    }
                                                                                }
                                                                                else if (diffAFD >= -1) {
                                                                                    return 1;
                                                                                }
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                            else if (MVDiff >= 25) {
                                                                if (Cs < 2170) {
                                                                    return 2;
                                                                }
                                                                else if (Cs >= 2170) {
                                                                    if (Cs < 2302) {
                                                                        return 4;
                                                                    }
                                                                    else if (Cs >= 2302) {
                                                                        return 2;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (AFD >= 15) {
                                                            if (mu_mv_mag_sq < 32) {
                                                                if (Cs < 2453) {
                                                                    if (MVDiff < 33) {
                                                                        return 1;
                                                                    }
                                                                    else if (MVDiff >= 33) {
                                                                        if (Cs < 2078) {
                                                                            return 8;
                                                                        }
                                                                        else if (Cs >= 2078) {
                                                                            return 2;
                                                                        }
                                                                    }
                                                                }
                                                                else if (Cs >= 2453) {
                                                                    if (Rs < 2801) {
                                                                        if (Cs < 2628) {
                                                                            return 4;
                                                                        }
                                                                        else if (Cs >= 2628) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                    else if (Rs >= 2801) {
                                                                        return 1;
                                                                    }
                                                                }
                                                            }
                                                            else if (mu_mv_mag_sq >= 32) {
                                                                if (refDCval < 87) {
                                                                    if (Rs < 2554) {
                                                                        return 4;
                                                                    }
                                                                    else if (Rs >= 2554) {
                                                                        return 1;
                                                                    }
                                                                }
                                                                else if (refDCval >= 87) {
                                                                    if (ssDCval < 93) {
                                                                        if (Rs < 3047) {
                                                                            return 2;
                                                                        }
                                                                        else if (Rs >= 3047) {
                                                                            if (Rs < 3135) {
                                                                                return 1;
                                                                            }
                                                                            else if (Rs >= 3135) {
                                                                                return 2;
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (ssDCval >= 93) {
                                                                        if (RsCsDiff < 211) {
                                                                            return 1;
                                                                        }
                                                                        else if (RsCsDiff >= 211) {
                                                                            return 2;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (Cs >= 2690) {
                                                        if (mcTcor < 91) {
                                                            if (refDCval < 91) {
                                                                return 1;
                                                            }
                                                            else if (refDCval >= 91) {
                                                                if (posBalance < 61) {
                                                                    if (Rs < 3275) {
                                                                        return 8;
                                                                    }
                                                                    else if (Rs >= 3275) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (posBalance >= 61) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                        else if (mcTcor >= 91) {
                                                            if (Cs < 2800) {
                                                                if (SC < 5362) {
                                                                    return 8;
                                                                }
                                                                else if (SC >= 5362) {
                                                                    return 1;
                                                                }
                                                            }
                                                            else if (Cs >= 2800) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                            else if (refDCval >= 124) {
                                                if (MVDiff < 31) {
                                                    return 8;
                                                }
                                                else if (MVDiff >= 31) {
                                                    if (Rs < 2887) {
                                                        if (Cs < 1947) {
                                                            return 8;
                                                        }
                                                        else if (Cs >= 1947) {
                                                            if (Rs < 2209) {
                                                                return 1;
                                                            }
                                                            else if (Rs >= 2209) {
                                                                return 2;
                                                            }
                                                        }
                                                    }
                                                    else if (Rs >= 2887) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                        }
                    }
                }
                else if (Rs >= 3400) {
                    if (CsDiff < 10) {
                        if (mu_mv_mag_sq < 26) {
                            if (ssDCval < 114) {
                                if (SC < 4975) {
                                    if (Cs < 1078) {
                                        if (mcTcor < 80) {
                                            return 2;
                                        }
                                        else if (mcTcor >= 80) {
                                            return 1;
                                        }
                                    }
                                    else if (Cs >= 1078) {
                                        return 4;
                                    }
                                }
                                else if (SC >= 4975) {
                                    if (posBalance < 56) {
                                        if (SC < 7918) {
                                            return 8;
                                        }
                                        else if (SC >= 7918) {
                                            return 2;
                                        }
                                    }
                                    else if (posBalance >= 56) {
                                        if (RsCsDiff < 154) {
                                            return 8;
                                        }
                                        else if (RsCsDiff >= 154) {
                                            if (Rs < 4488) {
                                                return 1;
                                            }
                                            else if (Rs >= 4488) {
                                                return 4;
                                            }
                                        }
                                    }
                                }
                            }
                            else if (ssDCval >= 114) {
                                if (negBalance < 45) {
                                    if (TSC < 8) {
                                        return 4;
                                    }
                                    else if (TSC >= 8) {
                                        if (Cs < 2707) {
                                            return 1;
                                        }
                                        else if (Cs >= 2707) {
                                            return 2;
                                        }
                                    }
                                }
                                else if (negBalance >= 45) {
                                    if (SC < 5972) {
                                        return 2;
                                    }
                                    else if (SC >= 5972) {
                                        if (Rs < 4010) {
                                            if (AFD < 23) {
                                                return 1;
                                            }
                                            else if (AFD >= 23) {
                                                return 4;
                                            }
                                        }
                                        else if (Rs >= 4010) {
                                            return 2;
                                        }
                                    }
                                }
                            }
                        }
                        else if (mu_mv_mag_sq >= 26) {
                            if (TSCindex < 8) {
                                if (refDCval < 107) {
                                    if (ssDCval < 55) {
                                        if (MVDiff < 17) {
                                            if (Cs < 3848) {
                                                if (AFD < 6) {
                                                    return 8;
                                                }
                                                else if (AFD >= 6) {
                                                    if (RsCsDiff < 16) {
                                                        if (mu_mv_mag_sq < 58) {
                                                            return 4;
                                                        }
                                                        else if (mu_mv_mag_sq >= 58) {
                                                            if (Cs < 3462) {
                                                                if (Cs < 3374) {
                                                                    return 8;
                                                                }
                                                                else if (Cs >= 3374) {
                                                                    if (SC < 7026) {
                                                                        return 4;
                                                                    }
                                                                    else if (SC >= 7026) {
                                                                        return 8;
                                                                    }
                                                                }
                                                            }
                                                            else if (Cs >= 3462) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                    else if (RsCsDiff >= 16) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                            else if (Cs >= 3848) {
                                                return 2;
                                            }
                                        }
                                        else if (MVDiff >= 17) {
                                            if (AFD < 9) {
                                                if (CsDiff < 3) {
                                                    return 8;
                                                }
                                                else if (CsDiff >= 3) {
                                                    return 4;
                                                }
                                            }
                                            else if (AFD >= 9) {
                                                return 8;
                                            }
                                        }
                                    }
                                    else if (ssDCval >= 55) {
                                        if (Cs < 1347) {
                                            return 4;
                                        }
                                        else if (Cs >= 1347) {
                                            if (mcTcor < 94) {
                                                if (Cs < 2103) {
                                                    return 4;
                                                }
                                                else if (Cs >= 2103) {
                                                    return 8;
                                                }
                                            }
                                            else if (mcTcor >= 94) {
                                                if (negBalance < 30) {
                                                    return 2;
                                                }
                                                else if (negBalance >= 30) {
                                                    if (negBalance < 40) {
                                                        return 8;
                                                    }
                                                    else if (negBalance >= 40) {
                                                        if (posBalance < 52) {
                                                            if (Cs < 3562) {
                                                                return 8;
                                                            }
                                                            else if (Cs >= 3562) {
                                                                if (SC < 7652) {
                                                                    return 4;
                                                                }
                                                                else if (SC >= 7652) {
                                                                    if (Rs < 4048) {
                                                                        return 8;
                                                                    }
                                                                    else if (Rs >= 4048) {
                                                                        return 4;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (posBalance >= 52) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (refDCval >= 107) {
                                    if (Cs < 2126) {
                                        if (MVDiff < 8) {
                                            if (Cs < 1711) {
                                                return 8;
                                            }
                                            else if (Cs >= 1711) {
                                                return 4;
                                            }
                                        }
                                        else if (MVDiff >= 8) {
                                            return 4;
                                        }
                                    }
                                    else if (Cs >= 2126) {
                                        if (RsDiff < 5) {
                                            return 8;
                                        }
                                        else if (RsDiff >= 5) {
                                            if (ssDCval < 119) {
                                                return 2;
                                            }
                                            else if (ssDCval >= 119) {
                                                if (Rs < 4723) {
                                                    return 4;
                                                }
                                                else if (Rs >= 4723) {
                                                    if (SC < 7334) {
                                                        return 8;
                                                    }
                                                    else if (SC >= 7334) {
                                                        return 2;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                            else if (TSCindex >= 8) {
                                if (ssDCval < 67) {
                                    if (RsCsDiff < 46) {
                                        if (mu_mv_mag_sq < 53) {
                                            if (refDCval < 52) {
                                                if (gchDC < 2) {
                                                    return 2;
                                                }
                                                else if (gchDC >= 2) {
                                                    return 8;
                                                }
                                            }
                                            else if (refDCval >= 52) {
                                                if (MVDiff < 18) {
                                                    if (Cs < 3856) {
                                                        if (RsDiff < 3) {
                                                            return 2;
                                                        }
                                                        else if (RsDiff >= 3) {
                                                            if (diffAFD < -2) {
                                                                return 2;
                                                            }
                                                            else if (diffAFD >= -2) {
                                                                if (mcTcor < 93) {
                                                                    if (Rs < 3740) {
                                                                        return 2;
                                                                    }
                                                                    else if (Rs >= 3740) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (mcTcor >= 93) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (Cs >= 3856) {
                                                        if (CsDiff < 4) {
                                                            return 4;
                                                        }
                                                        else if (CsDiff >= 4) {
                                                            return 2;
                                                        }
                                                    }
                                                }
                                                else if (MVDiff >= 18) {
                                                    if (Rs < 3737) {
                                                        if (diffAFD < -1) {
                                                            if (Rs < 3643) {
                                                                if (Rs < 3477) {
                                                                    return 8;
                                                                }
                                                                else if (Rs >= 3477) {
                                                                    return 2;
                                                                }
                                                            }
                                                            else if (Rs >= 3643) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (diffAFD >= -1) {
                                                            if (Cs < 3408) {
                                                                return 2;
                                                            }
                                                            else if (Cs >= 3408) {
                                                                if (MVDiff < 21) {
                                                                    return 2;
                                                                }
                                                                else if (MVDiff >= 21) {
                                                                    if (posBalance < 57) {
                                                                        if (Rs < 3609) {
                                                                            return 8;
                                                                        }
                                                                        else if (Rs >= 3609) {
                                                                            return 2;
                                                                        }
                                                                    }
                                                                    else if (posBalance >= 57) {
                                                                        return 8;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (Rs >= 3737) {
                                                        if (negBalance < 49) {
                                                            if (Rs < 3812) {
                                                                return 8;
                                                            }
                                                            else if (Rs >= 3812) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (negBalance >= 49) {
                                                            return 2;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (mu_mv_mag_sq >= 53) {
                                            return 4;
                                        }
                                    }
                                    else if (RsCsDiff >= 46) {
                                        if (MVDiff < 33) {
                                            if (AFD < 16) {
                                                if (TSCindex < 9) {
                                                    if (AFD < 11) {
                                                        return 4;
                                                    }
                                                    else if (AFD >= 11) {
                                                        return 8;
                                                    }
                                                }
                                                else if (TSCindex >= 9) {
                                                    return 4;
                                                }
                                            }
                                            else if (AFD >= 16) {
                                                return 8;
                                            }
                                        }
                                        else if (MVDiff >= 33) {
                                            if (TSC < 9) {
                                                return 8;
                                            }
                                            else if (TSC >= 9) {
                                                if (Cs < 3956) {
                                                    return 4;
                                                }
                                                else if (Cs >= 3956) {
                                                    return 8;
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (ssDCval >= 67) {
                                    if (Rs < 4570) {
                                        if (negBalance < 37) {
                                            if (Cs < 2595) {
                                                if (Rs < 3844) {
                                                    if (refDCval < 108) {
                                                        if (CsDiff < 9) {
                                                            return 4;
                                                        }
                                                        else if (CsDiff >= 9) {
                                                            return 2;
                                                        }
                                                    }
                                                    else if (refDCval >= 108) {
                                                        return 2;
                                                    }
                                                }
                                                else if (Rs >= 3844) {
                                                    if (MVDiff < 8) {
                                                        return 8;
                                                    }
                                                    else if (MVDiff >= 8) {
                                                        if (Cs < 2309) {
                                                            return 4;
                                                        }
                                                        else if (Cs >= 2309) {
                                                            if (diffRsCsdiff < 19) {
                                                                if (Cs < 2353) {
                                                                    return 2;
                                                                }
                                                                else if (Cs >= 2353) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (diffRsCsdiff >= 19) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                            else if (Cs >= 2595) {
                                                if (Cs < 2711) {
                                                    return 8;
                                                }
                                                else if (Cs >= 2711) {
                                                    if (CsDiff < 8) {
                                                        return 1;
                                                    }
                                                    else if (CsDiff >= 8) {
                                                        return 2;
                                                    }
                                                }
                                            }
                                        }
                                        else if (negBalance >= 37) {
                                            if (mcTcor < 81) {
                                                if (Cs < 2406) {
                                                    if (Rs < 3682) {
                                                        return 8;
                                                    }
                                                    else if (Rs >= 3682) {
                                                        return 2;
                                                    }
                                                }
                                                else if (Cs >= 2406) {
                                                    if (Rs < 4527) {
                                                        if (ssDCval < 108) {
                                                            return 8;
                                                        }
                                                        else if (ssDCval >= 108) {
                                                            return 1;
                                                        }
                                                    }
                                                    else if (Rs >= 4527) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                            else if (mcTcor >= 81) {
                                                if (posBalance < 57) {
                                                    if (RsCsDiff < 129) {
                                                        if (Cs < 3068) {
                                                            if (ssDCval < 108) {
                                                                if (Rs < 3549) {
                                                                    if (Rs < 3435) {
                                                                        return 1;
                                                                    }
                                                                    else if (Rs >= 3435) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (Rs >= 3549) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (ssDCval >= 108) {
                                                                if (Cs < 2825) {
                                                                    if (AFD < 10) {
                                                                        return 4;
                                                                    }
                                                                    else if (AFD >= 10) {
                                                                        if (CsDiff < 8) {
                                                                            return 8;
                                                                        }
                                                                        else if (CsDiff >= 8) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                }
                                                                else if (Cs >= 2825) {
                                                                    if (Cs < 2944) {
                                                                        if (Rs < 4460) {
                                                                            return 1;
                                                                        }
                                                                        else if (Rs >= 4460) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                    else if (Cs >= 2944) {
                                                                        return 4;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (Cs >= 3068) {
                                                            if (Rs < 4015) {
                                                                return 8;
                                                            }
                                                            else if (Rs >= 4015) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                    else if (RsCsDiff >= 129) {
                                                        if (diffAFD < 2) {
                                                            if (tcor < 93) {
                                                                if (diffAFD < -5) {
                                                                    return 8;
                                                                }
                                                                else if (diffAFD >= -5) {
                                                                    if (tcor < 72) {
                                                                        return 1;
                                                                    }
                                                                    else if (tcor >= 72) {
                                                                        if (SC < 7160) {
                                                                            return 4;
                                                                        }
                                                                        else if (SC >= 7160) {
                                                                            if (Cs < 2675) {
                                                                                return 2;
                                                                            }
                                                                            else if (Cs >= 2675) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                            else if (tcor >= 93) {
                                                                return 2;
                                                            }
                                                        }
                                                        else if (diffAFD >= 2) {
                                                            if (Cs < 2201) {
                                                                return 4;
                                                            }
                                                            else if (Cs >= 2201) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (posBalance >= 57) {
                                                    if (mcTcor < 86) {
                                                        return 4;
                                                    }
                                                    else if (mcTcor >= 86) {
                                                        if (SC < 5015) {
                                                            return 2;
                                                        }
                                                        else if (SC >= 5015) {
                                                            if (ssDCval < 89) {
                                                                if (Rs < 3627) {
                                                                    return 1;
                                                                }
                                                                else if (Rs >= 3627) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (ssDCval >= 89) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (Rs >= 4570) {
                                        if (ssDCval < 89) {
                                            return 8;
                                        }
                                        else if (ssDCval >= 89) {
                                            if (RsCsDiff < 183) {
                                                if (CsDiff < 7) {
                                                    if (AFD < 11) {
                                                        if (Cs < 2719) {
                                                            return 8;
                                                        }
                                                        else if (Cs >= 2719) {
                                                            if (SC < 7722) {
                                                                return 2;
                                                            }
                                                            else if (SC >= 7722) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                    else if (AFD >= 11) {
                                                        if (mu_mv_mag_sq < 76) {
                                                            return 4;
                                                        }
                                                        else if (mu_mv_mag_sq >= 76) {
                                                            if (Rs < 5111) {
                                                                return 2;
                                                            }
                                                            else if (Rs >= 5111) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (CsDiff >= 7) {
                                                    if (AFD < 11) {
                                                        if (Rs < 4856) {
                                                            return 2;
                                                        }
                                                        else if (Rs >= 4856) {
                                                            if (Cs < 2615) {
                                                                return 8;
                                                            }
                                                            else if (Cs >= 2615) {
                                                                return 2;
                                                            }
                                                        }
                                                    }
                                                    else if (AFD >= 11) {
                                                        if (mu_mv_mag_sq < 62) {
                                                            if (diffRsCsdiff < -51) {
                                                                return 2;
                                                            }
                                                            else if (diffRsCsdiff >= -51) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (mu_mv_mag_sq >= 62) {
                                                            if (Cs < 3172) {
                                                                if (SC < 7400) {
                                                                    return 2;
                                                                }
                                                                else if (SC >= 7400) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (Cs >= 3172) {
                                                                return 2;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                            else if (RsCsDiff >= 183) {
                                                if (mcTcor < 88) {
                                                    if (ssDCval < 106) {
                                                        if (negBalance < 48) {
                                                            return 4;
                                                        }
                                                        else if (negBalance >= 48) {
                                                            return 2;
                                                        }
                                                    }
                                                    else if (ssDCval >= 106) {
                                                        if (ssDCval < 125) {
                                                            return 2;
                                                        }
                                                        else if (ssDCval >= 125) {
                                                            return 1;
                                                        }
                                                    }
                                                }
                                                else if (mcTcor >= 88) {
                                                    return 4;
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                        }
                    }
                    else if (CsDiff >= 10) {
                        if (TSC < 17) {
                            if (mu_mv_mag_sq < 47) {
                                if (SC < 7512) {
                                    if (refDCval < 140) {
                                        if (mcTcor < 89) {
                                            if (refDCval < 124) {
                                                if (RsDiff < 14) {
                                                    if (AFD < 14) {
                                                        return 4;
                                                    }
                                                    else if (AFD >= 14) {
                                                        if (mu_mv_mag_sq < 3) {
                                                            return 8;
                                                        }
                                                        else if (mu_mv_mag_sq >= 3) {
                                                            return 1;
                                                        }
                                                    }
                                                }
                                                else if (RsDiff >= 14) {
                                                    if (RsDiff < 20) {
                                                        if (CsDiff < 14) {
                                                            if (Cs < 2893) {
                                                                return 2;
                                                            }
                                                            else if (Cs >= 2893) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (CsDiff >= 14) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (RsDiff >= 20) {
                                                        return 2;
                                                    }
                                                }
                                            }
                                            else if (refDCval >= 124) {
                                                if (tcor < 70) {
                                                    return 1;
                                                }
                                                else if (tcor >= 70) {
                                                    if (negBalance < 52) {
                                                        return 1;
                                                    }
                                                    else if (negBalance >= 52) {
                                                        if (Rs < 3863) {
                                                            return 4;
                                                        }
                                                        else if (Rs >= 3863) {
                                                            if (Rs < 4204) {
                                                                return 1;
                                                            }
                                                            else if (Rs >= 4204) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (mcTcor >= 89) {
                                            if (ssDCval < 102) {
                                                return 4;
                                            }
                                            else if (ssDCval >= 102) {
                                                if (Cs < 2586) {
                                                    if (Rs < 4004) {
                                                        return 1;
                                                    }
                                                    else if (Rs >= 4004) {
                                                        if (SC < 6731) {
                                                            return 4;
                                                        }
                                                        else if (SC >= 6731) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                                else if (Cs >= 2586) {
                                                    return 8;
                                                }
                                            }
                                        }
                                    }
                                    else if (refDCval >= 140) {
                                        if (posBalance < 54) {
                                            return 4;
                                        }
                                        else if (posBalance >= 54) {
                                            if (Cs < 1620) {
                                                return 4;
                                            }
                                            else if (Cs >= 1620) {
                                                return 2;
                                            }
                                        }
                                    }
                                }
                                else if (SC >= 7512) {
                                    if (refDCval < 153) {
                                        if (posBalance < 54) {
                                            if (AFD < 17) {
                                                if (SC < 8043) {
                                                    return 1;
                                                }
                                                else if (SC >= 8043) {
                                                    if (Rs < 5372) {
                                                        return 2;
                                                    }
                                                    else if (Rs >= 5372) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                            else if (AFD >= 17) {
                                                if (mu_mv_mag_sq < 43) {
                                                    if (tcor < 67) {
                                                        return 8;
                                                    }
                                                    else if (tcor >= 67) {
                                                        if (RsDiff < 20) {
                                                            return 8;
                                                        }
                                                        else if (RsDiff >= 20) {
                                                            if (AFD < 20) {
                                                                return 8;
                                                            }
                                                            else if (AFD >= 20) {
                                                                if (Cs < 3861) {
                                                                    return 4;
                                                                }
                                                                else if (Cs >= 3861) {
                                                                    if (Cs < 5169) {
                                                                        return 8;
                                                                    }
                                                                    else if (Cs >= 5169) {
                                                                        return 4;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (mu_mv_mag_sq >= 43) {
                                                    if (SC < 9006) {
                                                        if (Rs < 4925) {
                                                            return 8;
                                                        }
                                                        else if (Rs >= 4925) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (SC >= 9006) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                        }
                                        else if (posBalance >= 54) {
                                            if (Rs < 5867) {
                                                if (tcor < 67) {
                                                    if (RsCsDiff < 671) {
                                                        return 1;
                                                    }
                                                    else if (RsCsDiff >= 671) {
                                                        if (negBalance < 51) {
                                                            return 8;
                                                        }
                                                        else if (negBalance >= 51) {
                                                            return 2;
                                                        }
                                                    }
                                                }
                                                else if (tcor >= 67) {
                                                    if (mcTcor < 84) {
                                                        if (diffMVdiffVal < 5) {
                                                            if (tcor < 77) {
                                                                if (AFD < 23) {
                                                                    return 4;
                                                                }
                                                                else if (AFD >= 23) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (tcor >= 77) {
                                                                return 2;
                                                            }
                                                        }
                                                        else if (diffMVdiffVal >= 5) {
                                                            if (Cs < 3193) {
                                                                return 2;
                                                            }
                                                            else if (Cs >= 3193) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                    else if (mcTcor >= 84) {
                                                        if (ssDCval < 108) {
                                                            if (Rs < 5494) {
                                                                if (TSC < 8) {
                                                                    return 8;
                                                                }
                                                                else if (TSC >= 8) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (Rs >= 5494) {
                                                                return 1;
                                                            }
                                                        }
                                                        else if (ssDCval >= 108) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (Rs >= 5867) {
                                                if (Rs < 7498) {
                                                    if (SC < 11024) {
                                                        return 1;
                                                    }
                                                    else if (SC >= 11024) {
                                                        if (SC < 11186) {
                                                            return 8;
                                                        }
                                                        else if (SC >= 11186) {
                                                            if (TSC < 15) {
                                                                return 1;
                                                            }
                                                            else if (TSC >= 15) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (Rs >= 7498) {
                                                    return 4;
                                                }
                                            }
                                        }
                                    }
                                    else if (refDCval >= 153) {
                                        if (RsDiff < 23) {
                                            if (mu_mv_mag_sq < 3) {
                                                return 2;
                                            }
                                            else if (mu_mv_mag_sq >= 3) {
                                                if (RsDiff < 17) {
                                                    if (AFD < 22) {
                                                        return 2;
                                                    }
                                                    else if (AFD >= 22) {
                                                        return 4;
                                                    }
                                                }
                                                else if (RsDiff >= 17) {
                                                    if (CsDiff < 17) {
                                                        return 4;
                                                    }
                                                    else if (CsDiff >= 17) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                        }
                                        else if (RsDiff >= 23) {
                                            if (Rs < 8605) {
                                                return 4;
                                            }
                                            else if (Rs >= 8605) {
                                                if (SC < 13772) {
                                                    return 4;
                                                }
                                                else if (SC >= 13772) {
                                                    return 8;
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                            else if (mu_mv_mag_sq >= 47) {
                                if (refDCval < 133) {
                                    if (ssDCval < 122) {
                                        if (AFD < 16) {
                                            if (diffTSC < 2) {
                                                if (AFD < 15) {
                                                    if (refDCval < 103) {
                                                        return 4;
                                                    }
                                                    else if (refDCval >= 103) {
                                                        if (RsCsDiff < 353) {
                                                            if (diffAFD < -1) {
                                                                return 2;
                                                            }
                                                            else if (diffAFD >= -1) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (RsCsDiff >= 353) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                                else if (AFD >= 15) {
                                                    if (tcor < 87) {
                                                        if (SC < 8007) {
                                                            if (SC < 7192) {
                                                                return 8;
                                                            }
                                                            else if (SC >= 7192) {
                                                                return 2;
                                                            }
                                                        }
                                                        else if (SC >= 8007) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (tcor >= 87) {
                                                        if (diffTSC < -15) {
                                                            return 4;
                                                        }
                                                        else if (diffTSC >= -15) {
                                                            if (refDCval < 68) {
                                                                return 4;
                                                            }
                                                            else if (refDCval >= 68) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                            else if (diffTSC >= 2) {
                                                return 8;
                                            }
                                        }
                                        else if (AFD >= 16) {
                                            if (RsDiff < 24) {
                                                if (mu_mv_mag_sq < 61) {
                                                    if (Cs < 3526) {
                                                        return 4;
                                                    }
                                                    else if (Cs >= 3526) {
                                                        if (Cs < 5557) {
                                                            return 8;
                                                        }
                                                        else if (Cs >= 5557) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                                else if (mu_mv_mag_sq >= 61) {
                                                    return 8;
                                                }
                                            }
                                            else if (RsDiff >= 24) {
                                                if (Rs < 6863) {
                                                    if (TSC < 13) {
                                                        return 4;
                                                    }
                                                    else if (TSC >= 13) {
                                                        return 2;
                                                    }
                                                }
                                                else if (Rs >= 6863) {
                                                    return 4;
                                                }
                                            }
                                        }
                                    }
                                    else if (ssDCval >= 122) {
                                        if (MVDiff < 5) {
                                            return 1;
                                        }
                                        else if (MVDiff >= 5) {
                                            return 8;
                                        }
                                    }
                                }
                                else if (refDCval >= 133) {
                                    if (SC < 8661) {
                                        if (refDCval < 141) {
                                            if (posBalance < 48) {
                                                return 8;
                                            }
                                            else if (posBalance >= 48) {
                                                return 1;
                                            }
                                        }
                                        else if (refDCval >= 141) {
                                            if (Rs < 4394) {
                                                return 4;
                                            }
                                            else if (Rs >= 4394) {
                                                return 8;
                                            }
                                        }
                                    }
                                    else if (SC >= 8661) {
                                        return 8;
                                    }
                                }
                            }
                        }
                        else if (TSC >= 17) {
                            if (Cs < 6601) {
                                if (Rs < 6599) {
                                    if (CsDiff < 16) {
                                        if (RsCsDiff < 568) {
                                            if (posBalance < 60) {
                                                if (Cs < 3362) {
                                                    return 1;
                                                }
                                                else if (Cs >= 3362) {
                                                    return 8;
                                                }
                                            }
                                            else if (posBalance >= 60) {
                                                if (diffMVdiffVal < -1) {
                                                    if (SC < 7306) {
                                                        return 1;
                                                    }
                                                    else if (SC >= 7306) {
                                                        return 2;
                                                    }
                                                }
                                                else if (diffMVdiffVal >= -1) {
                                                    return 2;
                                                }
                                            }
                                        }
                                        else if (RsCsDiff >= 568) {
                                            if (AFD < 44) {
                                                if (gchDC < 4) {
                                                    return 4;
                                                }
                                                else if (gchDC >= 4) {
                                                    return 8;
                                                }
                                            }
                                            else if (AFD >= 44) {
                                                return 2;
                                            }
                                        }
                                    }
                                    else if (CsDiff >= 16) {
                                        if (Cs < 4323) {
                                            if (diffAFD < 3) {
                                                if (RsDiff < 20) {
                                                    if (tcor < 64) {
                                                        if (ssDCval < 96) {
                                                            return 8;
                                                        }
                                                        else if (ssDCval >= 96) {
                                                            return 1;
                                                        }
                                                    }
                                                    else if (tcor >= 64) {
                                                        return 8;
                                                    }
                                                }
                                                else if (RsDiff >= 20) {
                                                    if (Rs < 4965) {
                                                        if (Rs < 4499) {
                                                            return 2;
                                                        }
                                                        else if (Rs >= 4499) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (Rs >= 4965) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                            else if (diffAFD >= 3) {
                                                return 8;
                                            }
                                        }
                                        else if (Cs >= 4323) {
                                            if (AFD < 19) {
                                                return 1;
                                            }
                                            else if (AFD >= 19) {
                                                if (RsDiff < 30) {
                                                    if (posBalance < 54) {
                                                        if (MVDiff < 28) {
                                                            return 8;
                                                        }
                                                        else if (MVDiff >= 28) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (posBalance >= 54) {
                                                        return 8;
                                                    }
                                                }
                                                else if (RsDiff >= 30) {
                                                    if (AFD < 33) {
                                                        return 2;
                                                    }
                                                    else if (AFD >= 33) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (Rs >= 6599) {
                                    if (MVDiff < 29) {
                                        if (posBalance < 55) {
                                            if (refDCval < 143) {
                                                if (negBalance < 62) {
                                                    return 8;
                                                }
                                                else if (negBalance >= 62) {
                                                    return 4;
                                                }
                                            }
                                            else if (refDCval >= 143) {
                                                return 4;
                                            }
                                        }
                                        else if (posBalance >= 55) {
                                            if (SC < 12322) {
                                                if (Rs < 7397) {
                                                    if (SC < 12254) {
                                                        return 2;
                                                    }
                                                    else if (SC >= 12254) {
                                                        return 4;
                                                    }
                                                }
                                                else if (Rs >= 7397) {
                                                    return 4;
                                                }
                                            }
                                            else if (SC >= 12322) {
                                                if (SC < 14379) {
                                                    if (negBalance < 57) {
                                                        if (mu_mv_mag_sq < 17) {
                                                            return 8;
                                                        }
                                                        else if (mu_mv_mag_sq >= 17) {
                                                            if (Rs < 7122) {
                                                                return 8;
                                                            }
                                                            else if (Rs >= 7122) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                    else if (negBalance >= 57) {
                                                        if (SC < 12706) {
                                                            if (SC < 12401) {
                                                                return 8;
                                                            }
                                                            else if (SC >= 12401) {
                                                                return 2;
                                                            }
                                                        }
                                                        else if (SC >= 12706) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                                else if (SC >= 14379) {
                                                    if (AFD < 26) {
                                                        return 8;
                                                    }
                                                    else if (AFD >= 26) {
                                                        if (Rs < 8940) {
                                                            return 2;
                                                        }
                                                        else if (Rs >= 8940) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (MVDiff >= 29) {
                                        if (Cs < 4878) {
                                            return 8;
                                        }
                                        else if (Cs >= 4878) {
                                            if (mcTcor < 85) {
                                                if (AFD < 28) {
                                                    return 2;
                                                }
                                                else if (AFD >= 28) {
                                                    return 8;
                                                }
                                            }
                                            else if (mcTcor >= 85) {
                                                return 2;
                                            }
                                        }
                                    }
                                }
                            }
                            else if (Cs >= 6601) {
                                if (refDCval < 84) {
                                    return 4;
                                }
                                else if (refDCval >= 84) {
                                    if (tcor < 74) {
                                        if (negBalance < 54) {
                                            if (RsCsDiff < 6015) {
                                                return 8;
                                            }
                                            else if (RsCsDiff >= 6015) {
                                                return 4;
                                            }
                                        }
                                        else if (negBalance >= 54) {
                                            if (MVDiff < 12) {
                                                return 4;
                                            }
                                            else if (MVDiff >= 12) {
                                                if (mcTcor < 84) {
                                                    if (diffAFD < -1) {
                                                        return 8;
                                                    }
                                                    else if (diffAFD >= -1) {
                                                        return 4;
                                                    }
                                                }
                                                else if (mcTcor >= 84) {
                                                    if (Rs < 9702) {
                                                        return 8;
                                                    }
                                                    else if (Rs >= 9702) {
                                                        return 2;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (tcor >= 74) {
                                        if (diffTSC < -3) {
                                            if (Cs < 8637) {
                                                return 8;
                                            }
                                            else if (Cs >= 8637) {
                                                return 4;
                                            }
                                        }
                                        else if (diffTSC >= -3) {
                                            return 8;
                                        }
                                    }
                                }
                            }
                        }
                    }
                }
            }
        }
        else if (mu_mv_mag_sq >= 83) {
            if (negBalance < 34) {
                if (SC < 1257) {
                    if (refDCval < 161) {
                        if (refDCval < 92) {
                            if (ssDCval < 89) {
                                if (Rs < 147) {
                                    if (AFD < 7) {
                                        if (diffTSC < -1) {
                                            if (AFD < 1) {
                                                return 8;
                                            }
                                            else if (AFD >= 1) {
                                                if (AFD < 3) {
                                                    return 1;
                                                }
                                                else if (AFD >= 3) {
                                                    return 8;
                                                }
                                            }
                                        }
                                        else if (diffTSC >= -1) {
                                            if (mu_mv_mag_sq < 109) {
                                                if (MVDiff < 6) {
                                                    return 1;
                                                }
                                                else if (MVDiff >= 6) {
                                                    return 4;
                                                }
                                            }
                                            else if (mu_mv_mag_sq >= 109) {
                                                return 4;
                                            }
                                        }
                                    }
                                    else if (AFD >= 7) {
                                        return 2;
                                    }
                                }
                                else if (Rs >= 147) {
                                    if (posBalance < 60) {
                                        if (Rs < 517) {
                                            if (SC < 923) {
                                                if (negBalance < 13) {
                                                    return 8;
                                                }
                                                else if (negBalance >= 13) {
                                                    if (Cs < 308) {
                                                        return 8;
                                                    }
                                                    else if (Cs >= 308) {
                                                        if (ssDCval < 71) {
                                                            return 8;
                                                        }
                                                        else if (ssDCval >= 71) {
                                                            if (SC < 497) {
                                                                if (Cs < 310) {
                                                                    if (Rs < 177) {
                                                                        return 8;
                                                                    }
                                                                    else if (Rs >= 177) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (Cs >= 310) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (SC >= 497) {
                                                                if (Cs < 318) {
                                                                    return 4;
                                                                }
                                                                else if (Cs >= 318) {
                                                                    if (Rs < 198) {
                                                                        return 8;
                                                                    }
                                                                    else if (Rs >= 198) {
                                                                        return 4;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                            else if (SC >= 923) {
                                                if (refDCval < 41) {
                                                    return 8;
                                                }
                                                else if (refDCval >= 41) {
                                                    return 4;
                                                }
                                            }
                                        }
                                        else if (Rs >= 517) {
                                            return 8;
                                        }
                                    }
                                    else if (posBalance >= 60) {
                                        if (negBalance < 3) {
                                            return 8;
                                        }
                                        else if (negBalance >= 3) {
                                            return 4;
                                        }
                                    }
                                }
                            }
                            else if (ssDCval >= 89) {
                                if (Rs < 469) {
                                    return 8;
                                }
                                else if (Rs >= 469) {
                                    return 4;
                                }
                            }
                        }
                        else if (refDCval >= 92) {
                            if (MVDiff < 113) {
                                if (refDCval < 159) {
                                    return 8;
                                }
                                else if (refDCval >= 159) {
                                    if (Rs < 496) {
                                        return 8;
                                    }
                                    else if (Rs >= 496) {
                                        return 4;
                                    }
                                }
                            }
                            else if (MVDiff >= 113) {
                                if (TSC < 3) {
                                    if (AFD < 3) {
                                        return 8;
                                    }
                                    else if (AFD >= 3) {
                                        return 2;
                                    }
                                }
                                else if (TSC >= 3) {
                                    return 4;
                                }
                            }
                        }
                    }
                    else if (refDCval >= 161) {
                        if (Cs < 408) {
                            return 4;
                        }
                        else if (Cs >= 408) {
                            if (refDCval < 165) {
                                if (MVDiff < 5) {
                                    return 4;
                                }
                                else if (MVDiff >= 5) {
                                    if (AFD < 3) {
                                        if (mu_mv_mag_sq < 97) {
                                            if (Rs < 548) {
                                                return 2;
                                            }
                                            else if (Rs >= 548) {
                                                return 4;
                                            }
                                        }
                                        else if (mu_mv_mag_sq >= 97) {
                                            return 4;
                                        }
                                    }
                                    else if (AFD >= 3) {
                                        return 4;
                                    }
                                }
                            }
                            else if (refDCval >= 165) {
                                if (diffMVdiffVal < 1) {
                                    return 2;
                                }
                                else if (diffMVdiffVal >= 1) {
                                    if (Rs < 540) {
                                        return 2;
                                    }
                                    else if (Rs >= 540) {
                                        return 4;
                                    }
                                }
                            }
                        }
                    }
                }
                else if (SC >= 1257) {
                    if (mu_mv_mag_sq < 121) {
                        if (posBalance < 18) {
                            if (ssDCval < 122) {
                                if (SC < 1857) {
                                    if (Cs < 832) {
                                        return 4;
                                    }
                                    else if (Cs >= 832) {
                                        return 1;
                                    }
                                }
                                else if (SC >= 1857) {
                                    if (ssDCval < 117) {
                                        if (tcor < 100) {
                                            if (posBalance < 7) {
                                                return 4;
                                            }
                                            else if (posBalance >= 7) {
                                                return 8;
                                            }
                                        }
                                        else if (tcor >= 100) {
                                            return 4;
                                        }
                                    }
                                    else if (ssDCval >= 117) {
                                        return 4;
                                    }
                                }
                            }
                            else if (ssDCval >= 122) {
                                if (RsDiff < 6) {
                                    if (MVDiff < 5) {
                                        if (diffMVdiffVal < -3) {
                                            return 1;
                                        }
                                        else if (diffMVdiffVal >= -3) {
                                            return 8;
                                        }
                                    }
                                    else if (MVDiff >= 5) {
                                        if (TSCindex < 2) {
                                            return 8;
                                        }
                                        else if (TSCindex >= 2) {
                                            if (diffAFD < 2) {
                                                if (Cs < 1490) {
                                                    if (AFD < 7) {
                                                        return 1;
                                                    }
                                                    else if (AFD >= 7) {
                                                        return 8;
                                                    }
                                                }
                                                else if (Cs >= 1490) {
                                                    return 4;
                                                }
                                            }
                                            else if (diffAFD >= 2) {
                                                return 8;
                                            }
                                        }
                                    }
                                }
                                else if (RsDiff >= 6) {
                                    return 1;
                                }
                            }
                        }
                        else if (posBalance >= 18) {
                            if (ssDCval < 117) {
                                if (ssDCval < 43) {
                                    if (Cs < 1145) {
                                        if (MVDiff < 24) {
                                            if (Cs < 708) {
                                                if (Rs < 972) {
                                                    return 8;
                                                }
                                                else if (Rs >= 972) {
                                                    return 4;
                                                }
                                            }
                                            else if (Cs >= 708) {
                                                return 8;
                                            }
                                        }
                                        else if (MVDiff >= 24) {
                                            return 4;
                                        }
                                    }
                                    else if (Cs >= 1145) {
                                        if (RsDiff < 1) {
                                            if (refDCval < 33) {
                                                if (tcor < 99) {
                                                    return 4;
                                                }
                                                else if (tcor >= 99) {
                                                    if (Rs < 1035) {
                                                        return 4;
                                                    }
                                                    else if (Rs >= 1035) {
                                                        return 2;
                                                    }
                                                }
                                            }
                                            else if (refDCval >= 33) {
                                                return 8;
                                            }
                                        }
                                        else if (RsDiff >= 1) {
                                            if (AFD < 5) {
                                                return 8;
                                            }
                                            else if (AFD >= 5) {
                                                return 4;
                                            }
                                        }
                                    }
                                }
                                else if (ssDCval >= 43) {
                                    if (ssDCval < 70) {
                                        if (mu_mv_mag_sq < 109) {
                                            if (Rs < 1592) {
                                                if (Cs < 791) {
                                                    return 8;
                                                }
                                                else if (Cs >= 791) {
                                                    if (AFD < 5) {
                                                        return 1;
                                                    }
                                                    else if (AFD >= 5) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                            else if (Rs >= 1592) {
                                                if (RsCsDiff < 43) {
                                                    return 8;
                                                }
                                                else if (RsCsDiff >= 43) {
                                                    return 4;
                                                }
                                            }
                                        }
                                        else if (mu_mv_mag_sq >= 109) {
                                            if (diffAFD < -2) {
                                                if (Rs < 2010) {
                                                    return 1;
                                                }
                                                else if (Rs >= 2010) {
                                                    return 8;
                                                }
                                            }
                                            else if (diffAFD >= -2) {
                                                return 8;
                                            }
                                        }
                                    }
                                    else if (ssDCval >= 70) {
                                        if (Rs < 3369) {
                                            if (SC < 3707) {
                                                if (mu_mv_mag_sq < 84) {
                                                    return 4;
                                                }
                                                else if (mu_mv_mag_sq >= 84) {
                                                    if (Rs < 371) {
                                                        return 4;
                                                    }
                                                    else if (Rs >= 371) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                            else if (SC >= 3707) {
                                                if (ssDCval < 72) {
                                                    return 1;
                                                }
                                                else if (ssDCval >= 72) {
                                                    if (Rs < 2015) {
                                                        return 1;
                                                    }
                                                    else if (Rs >= 2015) {
                                                        if (negBalance < 25) {
                                                            return 8;
                                                        }
                                                        else if (negBalance >= 25) {
                                                            if (posBalance < 40) {
                                                                if (Cs < 1361) {
                                                                    return 1;
                                                                }
                                                                else if (Cs >= 1361) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (posBalance >= 40) {
                                                                if (MVDiff < 1) {
                                                                    if (CsDiff < 4) {
                                                                        if (Rs < 3144) {
                                                                            return 8;
                                                                        }
                                                                        else if (Rs >= 3144) {
                                                                            if (Cs < 3115) {
                                                                                return 1;
                                                                            }
                                                                            else if (Cs >= 3115) {
                                                                                return 8;
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (CsDiff >= 4) {
                                                                        if (negBalance < 26) {
                                                                            return 8;
                                                                        }
                                                                        else if (negBalance >= 26) {
                                                                            if (RsCsDiff < 46) {
                                                                                return 1;
                                                                            }
                                                                            else if (RsCsDiff >= 46) {
                                                                                if (SC < 6155) {
                                                                                    return 1;
                                                                                }
                                                                                else if (SC >= 6155) {
                                                                                    return 8;
                                                                                }
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                                else if (MVDiff >= 1) {
                                                                    if (mu_mv_mag_sq < 118) {
                                                                        if (Cs < 1524) {
                                                                            return 1;
                                                                        }
                                                                        else if (Cs >= 1524) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                    else if (mu_mv_mag_sq >= 118) {
                                                                        return 1;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (Rs >= 3369) {
                                            if (diffMVdiffVal < -53) {
                                                return 2;
                                            }
                                            else if (diffMVdiffVal >= -53) {
                                                return 8;
                                            }
                                        }
                                    }
                                }
                            }
                            else if (ssDCval >= 117) {
                                if (SC < 4408) {
                                    if (mu_mv_mag_sq < 85) {
                                        return 4;
                                    }
                                    else if (mu_mv_mag_sq >= 85) {
                                        if (negBalance < 30) {
                                            if (diffAFD < -2) {
                                                return 4;
                                            }
                                            else if (diffAFD >= -2) {
                                                return 8;
                                            }
                                        }
                                        else if (negBalance >= 30) {
                                            if (Rs < 1237) {
                                                if (Cs < 965) {
                                                    return 1;
                                                }
                                                else if (Cs >= 965) {
                                                    return 4;
                                                }
                                            }
                                            else if (Rs >= 1237) {
                                                return 8;
                                            }
                                        }
                                    }
                                }
                                else if (SC >= 4408) {
                                    if (Rs < 4706) {
                                        if (mu_mv_mag_sq < 95) {
                                            return 4;
                                        }
                                        else if (mu_mv_mag_sq >= 95) {
                                            if (negBalance < 23) {
                                                if (AFD < 7) {
                                                    return 4;
                                                }
                                                else if (AFD >= 7) {
                                                    return 8;
                                                }
                                            }
                                            else if (negBalance >= 23) {
                                                return 8;
                                            }
                                        }
                                    }
                                    else if (Rs >= 4706) {
                                        if (refDCval < 125) {
                                            return 8;
                                        }
                                        else if (refDCval >= 125) {
                                            return 2;
                                        }
                                    }
                                }
                            }
                        }
                    }
                    else if (mu_mv_mag_sq >= 121) {
                        if (refDCval < 103) {
                            if (refDCval < 22) {
                                return 2;
                            }
                            else if (refDCval >= 22) {
                                if (mu_mv_mag_sq < 125) {
                                    if (AFD < 6) {
                                        if (MVDiff < 7) {
                                            if (CsDiff < 4) {
                                                return 8;
                                            }
                                            else if (CsDiff >= 4) {
                                                if (negBalance < 18) {
                                                    return 8;
                                                }
                                                else if (negBalance >= 18) {
                                                    if (Rs < 2091) {
                                                        return 4;
                                                    }
                                                    else if (Rs >= 2091) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                        }
                                        else if (MVDiff >= 7) {
                                            if (MVDiff < 14) {
                                                return 1;
                                            }
                                            else if (MVDiff >= 14) {
                                                return 8;
                                            }
                                        }
                                    }
                                    else if (AFD >= 6) {
                                        if (RsDiff < 6) {
                                            return 1;
                                        }
                                        else if (RsDiff >= 6) {
                                            return 8;
                                        }
                                    }
                                }
                                else if (mu_mv_mag_sq >= 125) {
                                    return 8;
                                }
                            }
                        }
                        else if (refDCval >= 103) {
                            if (ssDCval < 104) {
                                if (SC < 7800) {
                                    if (tcor < 99) {
                                        if (Cs < 3891) {
                                            return 1;
                                        }
                                        else if (Cs >= 3891) {
                                            return 8;
                                        }
                                    }
                                    else if (tcor >= 99) {
                                        if (Rs < 3774) {
                                            if (SC < 7541) {
                                                if (Rs < 3596) {
                                                    return 8;
                                                }
                                                else if (Rs >= 3596) {
                                                    if (Rs < 3639) {
                                                        return 1;
                                                    }
                                                    else if (Rs >= 3639) {
                                                        if (posBalance < 21) {
                                                            return 1;
                                                        }
                                                        else if (posBalance >= 21) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (SC >= 7541) {
                                                return 8;
                                            }
                                        }
                                        else if (Rs >= 3774) {
                                            if (Cs < 3984) {
                                                return 8;
                                            }
                                            else if (Cs >= 3984) {
                                                return 1;
                                            }
                                        }
                                    }
                                }
                                else if (SC >= 7800) {
                                    return 8;
                                }
                            }
                            else if (ssDCval >= 104) {
                                if (posBalance < 2) {
                                    if (mu_mv_mag_sq < 124) {
                                        if (Rs < 2102) {
                                            return 8;
                                        }
                                        else if (Rs >= 2102) {
                                            return 4;
                                        }
                                    }
                                    else if (mu_mv_mag_sq >= 124) {
                                        if (Cs < 1581) {
                                            return 1;
                                        }
                                        else if (Cs >= 1581) {
                                            if (Rs < 1828) {
                                                return 8;
                                            }
                                            else if (Rs >= 1828) {
                                                return 1;
                                            }
                                        }
                                    }
                                }
                                else if (posBalance >= 2) {
                                    if (negBalance < 25) {
                                        if (MVDiff < 8) {
                                            if (mu_mv_mag_sq < 123) {
                                                if (Rs < 3746) {
                                                    return 8;
                                                }
                                                else if (Rs >= 3746) {
                                                    if (Cs < 3199) {
                                                        return 4;
                                                    }
                                                    else if (Cs >= 3199) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                            else if (mu_mv_mag_sq >= 123) {
                                                return 8;
                                            }
                                        }
                                        else if (MVDiff >= 8) {
                                            return 2;
                                        }
                                    }
                                    else if (negBalance >= 25) {
                                        if (negBalance < 30) {
                                            if (Rs < 3340) {
                                                if (SC < 6051) {
                                                    if (Rs < 1927) {
                                                        return 1;
                                                    }
                                                    else if (Rs >= 1927) {
                                                        return 8;
                                                    }
                                                }
                                                else if (SC >= 6051) {
                                                    if (CsDiff < 4) {
                                                        return 1;
                                                    }
                                                    else if (CsDiff >= 4) {
                                                        if (negBalance < 26) {
                                                            return 8;
                                                        }
                                                        else if (negBalance >= 26) {
                                                            return 1;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (Rs >= 3340) {
                                                return 8;
                                            }
                                        }
                                        else if (negBalance >= 30) {
                                            if (RsDiff < 2) {
                                                return 4;
                                            }
                                            else if (RsDiff >= 2) {
                                                return 8;
                                            }
                                        }
                                    }
                                }
                            }
                        }
                    }
                }
            }
            else if (negBalance >= 34) {
                if (SC < 4689) {
                    if (AFD < 7) {
                        if (mu_mv_mag_sq < 108) {
                            if (negBalance < 38) {
                                if (refDCval < 118) {
                                    if (Rs < 388) {
                                        return 4;
                                    }
                                    else if (Rs >= 388) {
                                        return 8;
                                    }
                                }
                                else if (refDCval >= 118) {
                                    if (Rs < 410) {
                                        return 2;
                                    }
                                    else if (Rs >= 410) {
                                        return 4;
                                    }
                                }
                            }
                            else if (negBalance >= 38) {
                                if (MVDiff < 1) {
                                    if (Cs < 1233) {
                                        if (SC < 1161) {
                                            return 2;
                                        }
                                        else if (SC >= 1161) {
                                            if (Rs < 946) {
                                                return 4;
                                            }
                                            else if (Rs >= 946) {
                                                return 2;
                                            }
                                        }
                                    }
                                    else if (Cs >= 1233) {
                                        return 4;
                                    }
                                }
                                else if (MVDiff >= 1) {
                                    if (SC < 3871) {
                                        if (MVDiff < 22) {
                                            if (posBalance < 57) {
                                                if (posBalance < 36) {
                                                    return 4;
                                                }
                                                else if (posBalance >= 36) {
                                                    if (refDCval < 45) {
                                                        if (CsDiff < 2) {
                                                            if (mu_mv_mag_sq < 101) {
                                                                if (negBalance < 46) {
                                                                    if (Rs < 987) {
                                                                        return 4;
                                                                    }
                                                                    else if (Rs >= 987) {
                                                                        if (SC < 3340) {
                                                                            return 8;
                                                                        }
                                                                        else if (SC >= 3340) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                }
                                                                else if (negBalance >= 46) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (mu_mv_mag_sq >= 101) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (CsDiff >= 2) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (refDCval >= 45) {
                                                        if (refDCval < 75) {
                                                            return 8;
                                                        }
                                                        else if (refDCval >= 75) {
                                                            if (refDCval < 83) {
                                                                return 4;
                                                            }
                                                            else if (refDCval >= 83) {
                                                                if (ssDCval < 112) {
                                                                    return 8;
                                                                }
                                                                else if (ssDCval >= 112) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                            else if (posBalance >= 57) {
                                                if (ssDCval < 73) {
                                                    if (Cs < 681) {
                                                        return 8;
                                                    }
                                                    else if (Cs >= 681) {
                                                        if (Cs < 791) {
                                                            return 1;
                                                        }
                                                        else if (Cs >= 791) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                                else if (ssDCval >= 73) {
                                                    if (gchDC < 1) {
                                                        return 2;
                                                    }
                                                    else if (gchDC >= 1) {
                                                        if (Rs < 321) {
                                                            return 4;
                                                        }
                                                        else if (Rs >= 321) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (MVDiff >= 22) {
                                            if (posBalance < 41) {
                                                if (TSCindex < 2) {
                                                    return 2;
                                                }
                                                else if (TSCindex >= 2) {
                                                    return 4;
                                                }
                                            }
                                            else if (posBalance >= 41) {
                                                return 1;
                                            }
                                        }
                                    }
                                    else if (SC >= 3871) {
                                        if (Rs < 2498) {
                                            return 8;
                                        }
                                        else if (Rs >= 2498) {
                                            if (mu_mv_mag_sq < 100) {
                                                return 1;
                                            }
                                            else if (mu_mv_mag_sq >= 100) {
                                                return 8;
                                            }
                                        }
                                    }
                                }
                            }
                        }
                        else if (mu_mv_mag_sq >= 108) {
                            if (SC < 3583) {
                                if (TSC < 6) {
                                    if (mu_mv_mag_sq < 119) {
                                        if (refDCval < 73) {
                                            if (Cs < 447) {
                                                return 4;
                                            }
                                            else if (Cs >= 447) {
                                                if (RsCsDiff < 56) {
                                                    return 8;
                                                }
                                                else if (RsCsDiff >= 56) {
                                                    return 4;
                                                }
                                            }
                                        }
                                        else if (refDCval >= 73) {
                                            if (MVDiff < 3) {
                                                if (tcor < 95) {
                                                    return 8;
                                                }
                                                else if (tcor >= 95) {
                                                    if (diffAFD < 1) {
                                                        if (AFD < 3) {
                                                            return 8;
                                                        }
                                                        else if (AFD >= 3) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (diffAFD >= 1) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                            else if (MVDiff >= 3) {
                                                return 4;
                                            }
                                        }
                                    }
                                    else if (mu_mv_mag_sq >= 119) {
                                        if (negBalance < 42) {
                                            if (Rs < 223) {
                                                return 4;
                                            }
                                            else if (Rs >= 223) {
                                                return 8;
                                            }
                                        }
                                        else if (negBalance >= 42) {
                                            if (refDCval < 74) {
                                                if (ssDCval < 73) {
                                                    return 8;
                                                }
                                                else if (ssDCval >= 73) {
                                                    return 2;
                                                }
                                            }
                                            else if (refDCval >= 74) {
                                                if (AFD < 6) {
                                                    return 8;
                                                }
                                                else if (AFD >= 6) {
                                                    if (ssDCval < 75) {
                                                        return 8;
                                                    }
                                                    else if (ssDCval >= 75) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (TSC >= 6) {
                                    if (ssDCval < 72) {
                                        if (tcor < 91) {
                                            return 8;
                                        }
                                        else if (tcor >= 91) {
                                            if (Rs < 846) {
                                                return 4;
                                            }
                                            else if (Rs >= 846) {
                                                return 2;
                                            }
                                        }
                                    }
                                    else if (ssDCval >= 72) {
                                        if (posBalance < 58) {
                                            if (Rs < 1932) {
                                                if (mu_mv_mag_sq < 114) {
                                                    if (RsDiff < 5) {
                                                        return 4;
                                                    }
                                                    else if (RsDiff >= 5) {
                                                        return 8;
                                                    }
                                                }
                                                else if (mu_mv_mag_sq >= 114) {
                                                    return 8;
                                                }
                                            }
                                            else if (Rs >= 1932) {
                                                return 4;
                                            }
                                        }
                                        else if (posBalance >= 58) {
                                            if (refDCval < 74) {
                                                return 2;
                                            }
                                            else if (refDCval >= 74) {
                                                if (diffTSC < 1) {
                                                    if (SC < 3280) {
                                                        if (Rs < 1697) {
                                                            return 8;
                                                        }
                                                        else if (Rs >= 1697) {
                                                            return 2;
                                                        }
                                                    }
                                                    else if (SC >= 3280) {
                                                        return 4;
                                                    }
                                                }
                                                else if (diffTSC >= 1) {
                                                    if (ssDCval < 75) {
                                                        if (Rs < 1853) {
                                                            return 8;
                                                        }
                                                        else if (Rs >= 1853) {
                                                            return 2;
                                                        }
                                                    }
                                                    else if (ssDCval >= 75) {
                                                        if (Rs < 1811) {
                                                            return 8;
                                                        }
                                                        else if (Rs >= 1811) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                            else if (SC >= 3583) {
                                if (tcor < 95) {
                                    if (refDCval < 76) {
                                        if (mcTcor < 95) {
                                            if (diffAFD < -1) {
                                                return 2;
                                            }
                                            else if (diffAFD >= -1) {
                                                if (diffAFD < 1) {
                                                    return 8;
                                                }
                                                else if (diffAFD >= 1) {
                                                    if (Cs < 1937) {
                                                        return 8;
                                                    }
                                                    else if (Cs >= 1937) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                        }
                                        else if (mcTcor >= 95) {
                                            return 4;
                                        }
                                    }
                                    else if (refDCval >= 76) {
                                        if (ssDCval < 102) {
                                            return 8;
                                        }
                                        else if (ssDCval >= 102) {
                                            if (AFD < 6) {
                                                return 8;
                                            }
                                            else if (AFD >= 6) {
                                                if (MVDiff < 1) {
                                                    return 4;
                                                }
                                                else if (MVDiff >= 1) {
                                                    return 8;
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (tcor >= 95) {
                                    if (mu_mv_mag_sq < 128) {
                                        if (refDCval < 89) {
                                            if (Cs < 2472) {
                                                return 4;
                                            }
                                            else if (Cs >= 2472) {
                                                return 8;
                                            }
                                        }
                                        else if (refDCval >= 89) {
                                            if (refDCval < 97) {
                                                return 8;
                                            }
                                            else if (refDCval >= 97) {
                                                if (Cs < 2169) {
                                                    return 8;
                                                }
                                                else if (Cs >= 2169) {
                                                    return 4;
                                                }
                                            }
                                        }
                                    }
                                    else if (mu_mv_mag_sq >= 128) {
                                        return 8;
                                    }
                                }
                            }
                        }
                    }
                    else if (AFD >= 7) {
                        if (RsDiff < 5) {
                            if (SC < 3656) {
                                if (tcor < 97) {
                                    if (CsDiff < 5) {
                                        if (mu_mv_mag_sq < 113) {
                                            if (Rs < 1097) {
                                                if (RsCsDiff < 23) {
                                                    return 2;
                                                }
                                                else if (RsCsDiff >= 23) {
                                                    if (Rs < 994) {
                                                        return 8;
                                                    }
                                                    else if (Rs >= 994) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                            else if (Rs >= 1097) {
                                                if (diffRsCsdiff < -17) {
                                                    if (Cs < 1588) {
                                                        return 2;
                                                    }
                                                    else if (Cs >= 1588) {
                                                        return 8;
                                                    }
                                                }
                                                else if (diffRsCsdiff >= -17) {
                                                    return 2;
                                                }
                                            }
                                        }
                                        else if (mu_mv_mag_sq >= 113) {
                                            if (tcor < 93) {
                                                if (posBalance < 59) {
                                                    if (Rs < 1742) {
                                                        if (negBalance < 49) {
                                                            if (Rs < 1383) {
                                                                return 8;
                                                            }
                                                            else if (Rs >= 1383) {
                                                                return 2;
                                                            }
                                                        }
                                                        else if (negBalance >= 49) {
                                                            return 2;
                                                        }
                                                    }
                                                    else if (Rs >= 1742) {
                                                        return 8;
                                                    }
                                                }
                                                else if (posBalance >= 59) {
                                                    return 2;
                                                }
                                            }
                                            else if (tcor >= 93) {
                                                if (refDCval < 76) {
                                                    if (Cs < 1642) {
                                                        return 2;
                                                    }
                                                    else if (Cs >= 1642) {
                                                        return 4;
                                                    }
                                                }
                                                else if (refDCval >= 76) {
                                                    return 4;
                                                }
                                            }
                                        }
                                    }
                                    else if (CsDiff >= 5) {
                                        if (ssDCval < 87) {
                                            if (AFD < 8) {
                                                if (Cs < 1645) {
                                                    return 8;
                                                }
                                                else if (Cs >= 1645) {
                                                    return 2;
                                                }
                                            }
                                            else if (AFD >= 8) {
                                                return 2;
                                            }
                                        }
                                        else if (ssDCval >= 87) {
                                            return 8;
                                        }
                                    }
                                }
                                else if (tcor >= 97) {
                                    if (Cs < 344) {
                                        if (Cs < 10) {
                                            return 1;
                                        }
                                        else if (Cs >= 10) {
                                            return 2;
                                        }
                                    }
                                    else if (Cs >= 344) {
                                        return 4;
                                    }
                                }
                            }
                            else if (SC >= 3656) {
                                if (ssDCval < 86) {
                                    if (Rs < 1924) {
                                        return 8;
                                    }
                                    else if (Rs >= 1924) {
                                        return 4;
                                    }
                                }
                                else if (ssDCval >= 86) {
                                    if (MVDiff < 4) {
                                        if (Rs < 2018) {
                                            return 2;
                                        }
                                        else if (Rs >= 2018) {
                                            return 8;
                                        }
                                    }
                                    else if (MVDiff >= 4) {
                                        return 1;
                                    }
                                }
                            }
                        }
                        else if (RsDiff >= 5) {
                            if (Cs < 1442) {
                                if (SC < 3657) {
                                    if (MVDiff < 24) {
                                        if (SC < 2213) {
                                            return 2;
                                        }
                                        else if (SC >= 2213) {
                                            return 8;
                                        }
                                    }
                                    else if (MVDiff >= 24) {
                                        if (diffAFD < 31) {
                                            if (SC < 780) {
                                                return 2;
                                            }
                                            else if (SC >= 780) {
                                                if (AFD < 33) {
                                                    if (MVDiff < 60) {
                                                        if (AFD < 26) {
                                                            return 2;
                                                        }
                                                        else if (AFD >= 26) {
                                                            return 1;
                                                        }
                                                    }
                                                    else if (MVDiff >= 60) {
                                                        return 1;
                                                    }
                                                }
                                                else if (AFD >= 33) {
                                                    return 2;
                                                }
                                            }
                                        }
                                        else if (diffAFD >= 31) {
                                            return 8;
                                        }
                                    }
                                }
                                else if (SC >= 3657) {
                                    if (MVDiff < 343) {
                                        if (diffRsCsdiff < 97) {
                                            return 2;
                                        }
                                        else if (diffRsCsdiff >= 97) {
                                            if (RsDiff < 21) {
                                                return 4;
                                            }
                                            else if (RsDiff >= 21) {
                                                return 2;
                                            }
                                        }
                                    }
                                    else if (MVDiff >= 343) {
                                        return 1;
                                    }
                                }
                            }
                            else if (Cs >= 1442) {
                                if (TSCindex < 8) {
                                    if (Rs < 2549) {
                                        if (mcTcor < 94) {
                                            if (Rs < 2232) {
                                                return 8;
                                            }
                                            else if (Rs >= 2232) {
                                                return 4;
                                            }
                                        }
                                        else if (mcTcor >= 94) {
                                            if (negBalance < 39) {
                                                return 8;
                                            }
                                            else if (negBalance >= 39) {
                                                if (SC < 3545) {
                                                    return 2;
                                                }
                                                else if (SC >= 3545) {
                                                    return 4;
                                                }
                                            }
                                        }
                                    }
                                    else if (Rs >= 2549) {
                                        return 1;
                                    }
                                }
                                else if (TSCindex >= 8) {
                                    if (RsCsDiff < 131) {
                                        if (tcor < 95) {
                                            if (refDCval < 102) {
                                                if (refDCval < 93) {
                                                    if (RsCsDiff < 46) {
                                                        if (Rs < 1352) {
                                                            return 8;
                                                        }
                                                        else if (Rs >= 1352) {
                                                            return 2;
                                                        }
                                                    }
                                                    else if (RsCsDiff >= 46) {
                                                        if (tcor < 91) {
                                                            if (SC < 4678) {
                                                                return 8;
                                                            }
                                                            else if (SC >= 4678) {
                                                                return 2;
                                                            }
                                                        }
                                                        else if (tcor >= 91) {
                                                            if (mu_mv_mag_sq < 100) {
                                                                return 2;
                                                            }
                                                            else if (mu_mv_mag_sq >= 100) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (refDCval >= 93) {
                                                    if (posBalance < 51) {
                                                        return 1;
                                                    }
                                                    else if (posBalance >= 51) {
                                                        return 2;
                                                    }
                                                }
                                            }
                                            else if (refDCval >= 102) {
                                                if (diffMVdiffVal < -2) {
                                                    if (Cs < 1795) {
                                                        return 8;
                                                    }
                                                    else if (Cs >= 1795) {
                                                        return 4;
                                                    }
                                                }
                                                else if (diffMVdiffVal >= -2) {
                                                    if (MVDiff < 8) {
                                                        if (ssDCval < 109) {
                                                            return 4;
                                                        }
                                                        else if (ssDCval >= 109) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (MVDiff >= 8) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                        }
                                        else if (tcor >= 95) {
                                            if (diffAFD < 1) {
                                                if (Cs < 1995) {
                                                    return 4;
                                                }
                                                else if (Cs >= 1995) {
                                                    if (ssDCval < 94) {
                                                        if (posBalance < 52) {
                                                            return 1;
                                                        }
                                                        else if (posBalance >= 52) {
                                                            return 2;
                                                        }
                                                    }
                                                    else if (ssDCval >= 94) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                            else if (diffAFD >= 1) {
                                                return 1;
                                            }
                                        }
                                    }
                                    else if (RsCsDiff >= 131) {
                                        if (refDCval < 104) {
                                            return 2;
                                        }
                                        else if (refDCval >= 104) {
                                            return 4;
                                        }
                                    }
                                }
                            }
                        }
                    }
                }
                else if (SC >= 4689) {
                    if (Cs < 4081) {
                        if (negBalance < 54) {
                            if (ssDCval < 98) {
                                if (TSCindex < 8) {
                                    if (SC < 4772) {
                                        return 4;
                                    }
                                    else if (SC >= 4772) {
                                        if (posBalance < 57) {
                                            if (mcTcor < 97) {
                                                if (diffTSC < -1) {
                                                    if (Rs < 3511) {
                                                        if (Rs < 2421) {
                                                            return 4;
                                                        }
                                                        else if (Rs >= 2421) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (Rs >= 3511) {
                                                        return 4;
                                                    }
                                                }
                                                else if (diffTSC >= -1) {
                                                    if (diffMVdiffVal < 1) {
                                                        if (mcTcor < 96) {
                                                            if (SC < 4868) {
                                                                return 4;
                                                            }
                                                            else if (SC >= 4868) {
                                                                if (mu_mv_mag_sq < 94) {
                                                                    if (diffMVdiffVal < -2) {
                                                                        return 4;
                                                                    }
                                                                    else if (diffMVdiffVal >= -2) {
                                                                        return 8;
                                                                    }
                                                                }
                                                                else if (mu_mv_mag_sq >= 94) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                        else if (mcTcor >= 96) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (diffMVdiffVal >= 1) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                            else if (mcTcor >= 97) {
                                                return 8;
                                            }
                                        }
                                        else if (posBalance >= 57) {
                                            return 8;
                                        }
                                    }
                                }
                                else if (TSCindex >= 8) {
                                    if (ssDCval < 89) {
                                        if (ssDCval < 87) {
                                            if (Rs < 3170) {
                                                if (diffRsCsdiff < -25) {
                                                    return 8;
                                                }
                                                else if (diffRsCsdiff >= -25) {
                                                    if (negBalance < 46) {
                                                        return 8;
                                                    }
                                                    else if (negBalance >= 46) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                            else if (Rs >= 3170) {
                                                return 4;
                                            }
                                        }
                                        else if (ssDCval >= 87) {
                                            return 8;
                                        }
                                    }
                                    else if (ssDCval >= 89) {
                                        if (MVDiff < 1) {
                                            return 2;
                                        }
                                        else if (MVDiff >= 1) {
                                            if (mcTcor < 94) {
                                                return 8;
                                            }
                                            else if (mcTcor >= 94) {
                                                if (Cs < 3059) {
                                                    return 4;
                                                }
                                                else if (Cs >= 3059) {
                                                    return 2;
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                            else if (ssDCval >= 98) {
                                if (tcor < 95) {
                                    if (Rs < 3992) {
                                        if (TSC < 11) {
                                            if (diffRsCsdiff < -36) {
                                                if (diffMVdiffVal < -3) {
                                                    if (Cs < 2247) {
                                                        return 8;
                                                    }
                                                    else if (Cs >= 2247) {
                                                        return 4;
                                                    }
                                                }
                                                else if (diffMVdiffVal >= -3) {
                                                    if (mcTcor < 86) {
                                                        if (SC < 6006) {
                                                            return 4;
                                                        }
                                                        else if (SC >= 6006) {
                                                            return 1;
                                                        }
                                                    }
                                                    else if (mcTcor >= 86) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                            else if (diffRsCsdiff >= -36) {
                                                if (SC < 4707) {
                                                    return 4;
                                                }
                                                else if (SC >= 4707) {
                                                    if (AFD < 11) {
                                                        return 8;
                                                    }
                                                    else if (AFD >= 11) {
                                                        if (SC < 4940) {
                                                            return 4;
                                                        }
                                                        else if (SC >= 4940) {
                                                            if (RsCsDiff < 183) {
                                                                return 8;
                                                            }
                                                            else if (RsCsDiff >= 183) {
                                                                if (Cs < 2973) {
                                                                    return 4;
                                                                }
                                                                else if (Cs >= 2973) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (TSC >= 11) {
                                            if (diffMVdiffVal < -4) {
                                                return 1;
                                            }
                                            else if (diffMVdiffVal >= -4) {
                                                if (MVDiff < 9) {
                                                    if (Rs < 3925) {
                                                        return 8;
                                                    }
                                                    else if (Rs >= 3925) {
                                                        return 4;
                                                    }
                                                }
                                                else if (MVDiff >= 9) {
                                                    if (Rs < 3329) {
                                                        return 1;
                                                    }
                                                    else if (Rs >= 3329) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (Rs >= 3992) {
                                        if (negBalance < 46) {
                                            if (Rs < 5015) {
                                                if (MVDiff < 7) {
                                                    if (diffAFD < 2) {
                                                        if (tcor < 88) {
                                                            if (Rs < 4436) {
                                                                return 8;
                                                            }
                                                            else if (Rs >= 4436) {
                                                                return 2;
                                                            }
                                                        }
                                                        else if (tcor >= 88) {
                                                            if (Rs < 4825) {
                                                                return 4;
                                                            }
                                                            else if (Rs >= 4825) {
                                                                if (Rs < 5007) {
                                                                    return 2;
                                                                }
                                                                else if (Rs >= 5007) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (diffAFD >= 2) {
                                                        return 2;
                                                    }
                                                }
                                                else if (MVDiff >= 7) {
                                                    if (Rs < 4239) {
                                                        if (TSC < 6) {
                                                            return 8;
                                                        }
                                                        else if (TSC >= 6) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (Rs >= 4239) {
                                                        if (diffAFD < 1) {
                                                            if (diffTSC < 1) {
                                                                return 2;
                                                            }
                                                            else if (diffTSC >= 1) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (diffAFD >= 1) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (Rs >= 5015) {
                                                if (diffTSC < 1) {
                                                    if (diffRsCsdiff < 16) {
                                                        return 8;
                                                    }
                                                    else if (diffRsCsdiff >= 16) {
                                                        if (Cs < 2944) {
                                                            return 2;
                                                        }
                                                        else if (Cs >= 2944) {
                                                            if (Cs < 3363) {
                                                                return 8;
                                                            }
                                                            else if (Cs >= 3363) {
                                                                return 2;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (diffTSC >= 1) {
                                                    if (Rs < 5132) {
                                                        return 2;
                                                    }
                                                    else if (Rs >= 5132) {
                                                        if (Cs < 3292) {
                                                            return 8;
                                                        }
                                                        else if (Cs >= 3292) {
                                                            return 2;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (negBalance >= 46) {
                                            if (refDCval < 122) {
                                                if (negBalance < 48) {
                                                    return 8;
                                                }
                                                else if (negBalance >= 48) {
                                                    return 4;
                                                }
                                            }
                                            else if (refDCval >= 122) {
                                                return 8;
                                            }
                                        }
                                    }
                                }
                                else if (tcor >= 95) {
                                    if (Rs < 5802) {
                                        if (refDCval < 125) {
                                            if (mu_mv_mag_sq < 123) {
                                                if (negBalance < 48) {
                                                    if (diffMVdiffVal < 3) {
                                                        if (RsCsDiff < 37) {
                                                            if (Cs < 2987) {
                                                                return 8;
                                                            }
                                                            else if (Cs >= 2987) {
                                                                if (Cs < 3080) {
                                                                    if (Cs < 3069) {
                                                                        if (refDCval < 114) {
                                                                            if (mu_mv_mag_sq < 122) {
                                                                                if (ssDCval < 113) {
                                                                                    return 8;
                                                                                }
                                                                                else if (ssDCval >= 113) {
                                                                                    return 4;
                                                                                }
                                                                            }
                                                                            else if (mu_mv_mag_sq >= 122) {
                                                                                if (Cs < 3026) {
                                                                                    return 8;
                                                                                }
                                                                                else if (Cs >= 3026) {
                                                                                    return 4;
                                                                                }
                                                                            }
                                                                        }
                                                                        else if (refDCval >= 114) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                    else if (Cs >= 3069) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (Cs >= 3080) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                        else if (RsCsDiff >= 37) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (diffMVdiffVal >= 3) {
                                                        if (Rs < 4047) {
                                                            return 4;
                                                        }
                                                        else if (Rs >= 4047) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                                else if (negBalance >= 48) {
                                                    if (refDCval < 112) {
                                                        return 8;
                                                    }
                                                    else if (refDCval >= 112) {
                                                        if (tcor < 97) {
                                                            return 8;
                                                        }
                                                        else if (tcor >= 97) {
                                                            if (Rs < 2650) {
                                                                return 8;
                                                            }
                                                            else if (Rs >= 2650) {
                                                                if (Rs < 3043) {
                                                                    if (Cs < 3002) {
                                                                        if (SC < 5527) {
                                                                            return 8;
                                                                        }
                                                                        else if (SC >= 5527) {
                                                                            if (mu_mv_mag_sq < 122) {
                                                                                return 4;
                                                                            }
                                                                            else if (mu_mv_mag_sq >= 122) {
                                                                                if (Rs < 2733) {
                                                                                    return 4;
                                                                                }
                                                                                else if (Rs >= 2733) {
                                                                                    return 8;
                                                                                }
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (Cs >= 3002) {
                                                                        if (ssDCval < 115) {
                                                                            return 4;
                                                                        }
                                                                        else if (ssDCval >= 115) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                }
                                                                else if (Rs >= 3043) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                            else if (mu_mv_mag_sq >= 123) {
                                                if (negBalance < 51) {
                                                    if (SC < 5750) {
                                                        return 8;
                                                    }
                                                    else if (SC >= 5750) {
                                                        if (negBalance < 48) {
                                                            return 8;
                                                        }
                                                        else if (negBalance >= 48) {
                                                            if (Rs < 2872) {
                                                                if (refDCval < 112) {
                                                                    return 8;
                                                                }
                                                                else if (refDCval >= 112) {
                                                                    if (Rs < 2807) {
                                                                        return 4;
                                                                    }
                                                                    else if (Rs >= 2807) {
                                                                        if (ssDCval < 112) {
                                                                            return 4;
                                                                        }
                                                                        else if (ssDCval >= 112) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                            else if (Rs >= 2872) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (negBalance >= 51) {
                                                    return 4;
                                                }
                                            }
                                        }
                                        else if (refDCval >= 125) {
                                            return 8;
                                        }
                                    }
                                    else if (Rs >= 5802) {
                                        return 4;
                                    }
                                }
                            }
                        }
                        else if (negBalance >= 54) {
                            if (RsCsDiff < 215) {
                                if (Rs < 4442) {
                                    if (refDCval < 87) {
                                        if (ssDCval < 83) {
                                            return 4;
                                        }
                                        else if (ssDCval >= 83) {
                                            return 2;
                                        }
                                    }
                                    else if (refDCval >= 87) {
                                        return 8;
                                    }
                                }
                                else if (Rs >= 4442) {
                                    return 4;
                                }
                            }
                            else if (RsCsDiff >= 215) {
                                if (Rs < 3708) {
                                    return 2;
                                }
                                else if (Rs >= 3708) {
                                    if (CsDiff < 10) {
                                        return 4;
                                    }
                                    else if (CsDiff >= 10) {
                                        return 8;
                                    }
                                }
                            }
                        }
                    }
                    else if (Cs >= 4081) {
                        if (negBalance < 64) {
                            if (mu_mv_mag_sq < 122) {
                                if (SC < 8582) {
                                    if (refDCval < 122) {
                                        if (RsDiff < 5) {
                                            return 4;
                                        }
                                        else if (RsDiff >= 5) {
                                            return 8;
                                        }
                                    }
                                    else if (refDCval >= 122) {
                                        return 1;
                                    }
                                }
                                else if (SC >= 8582) {
                                    if (tcor < 62) {
                                        if (diffAFD < 1) {
                                            return 1;
                                        }
                                        else if (diffAFD >= 1) {
                                            return 8;
                                        }
                                    }
                                    else if (tcor >= 62) {
                                        return 8;
                                    }
                                }
                            }
                            else if (mu_mv_mag_sq >= 122) {
                                if (ssDCval < 106) {
                                    if (SC < 10463) {
                                        if (SC < 9676) {
                                            if (Rs < 4623) {
                                                return 8;
                                            }
                                            else if (Rs >= 4623) {
                                                if (gchDC < 1) {
                                                    return 8;
                                                }
                                                else if (gchDC >= 1) {
                                                    if (Rs < 4639) {
                                                        return 4;
                                                    }
                                                    else if (Rs >= 4639) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                        }
                                        else if (SC >= 9676) {
                                            if (AFD < 9) {
                                                if (SC < 10015) {
                                                    if (mcTcor < 92) {
                                                        if (SC < 9828) {
                                                            return 4;
                                                        }
                                                        else if (SC >= 9828) {
                                                            if (RsCsDiff < 154) {
                                                                if (Rs < 4786) {
                                                                    return 4;
                                                                }
                                                                else if (Rs >= 4786) {
                                                                    if (mu_mv_mag_sq < 128) {
                                                                        return 8;
                                                                    }
                                                                    else if (mu_mv_mag_sq >= 128) {
                                                                        if (SC < 9931) {
                                                                            if (CsDiff < 9) {
                                                                                return 8;
                                                                            }
                                                                            else if (CsDiff >= 9) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                        else if (SC >= 9931) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                            else if (RsCsDiff >= 154) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                    else if (mcTcor >= 92) {
                                                        if (Cs < 5029) {
                                                            if (Rs < 4843) {
                                                                if (Cs < 4919) {
                                                                    if (Cs < 4895) {
                                                                        return 4;
                                                                    }
                                                                    else if (Cs >= 4895) {
                                                                        return 8;
                                                                    }
                                                                }
                                                                else if (Cs >= 4919) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (Rs >= 4843) {
                                                                if (Rs < 4903) {
                                                                    return 8;
                                                                }
                                                                else if (Rs >= 4903) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                        else if (Cs >= 5029) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                                else if (SC >= 10015) {
                                                    if (diffRsCsdiff < 40) {
                                                        return 4;
                                                    }
                                                    else if (diffRsCsdiff >= 40) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                            else if (AFD >= 9) {
                                                return 8;
                                            }
                                        }
                                    }
                                    else if (SC >= 10463) {
                                        if (refDCval < 83) {
                                            return 4;
                                        }
                                        else if (refDCval >= 83) {
                                            return 8;
                                        }
                                    }
                                }
                                else if (ssDCval >= 106) {
                                    if (refDCval < 121) {
                                        if (posBalance < 56) {
                                            return 8;
                                        }
                                        else if (posBalance >= 56) {
                                            return 4;
                                        }
                                    }
                                    else if (refDCval >= 121) {
                                        return 8;
                                    }
                                }
                            }
                        }
                        else if (negBalance >= 64) {
                            if (Rs < 7325) {
                                return 8;
                            }
                            else if (Rs >= 7325) {
                                if (tcor < 74) {
                                    return 4;
                                }
                                else if (tcor >= 74) {
                                    if (CsDiff < 23) {
                                        if (RsCsDiff < 950) {
                                            if (SC < 15482) {
                                                return 4;
                                            }
                                            else if (SC >= 15482) {
                                                if (RsDiff < 22) {
                                                    return 2;
                                                }
                                                else if (RsDiff >= 22) {
                                                    return 4;
                                                }
                                            }
                                        }
                                        else if (RsCsDiff >= 950) {
                                            if (diffMVdiffVal < -1) {
                                                return 4;
                                            }
                                            else if (diffMVdiffVal >= -1) {
                                                return 2;
                                            }
                                        }
                                    }
                                    else if (CsDiff >= 23) {
                                        if (posBalance < 52) {
                                            return 2;
                                        }
                                        else if (posBalance >= 52) {
                                            return 4;
                                        }
                                    }
                                }
                            }
                        }
                    }
                }
            }
        }
        return 4;
    }
    mfxU16 AGOPSelect16(mfxU32 Rs, mfxU32 CsDiff, mfxU32 TSC, mfxI32 diffAFD, mfxU32 Scindex, mfxU32 RsDiff, mfxI32 diffMVdiffVal, mfxU32 RsCsDiff, mfxU32 TSCindex, mfxI16 tcor, mfxU32 refDCval, mfxU32 MVDiff, mfxU32 AFD, mfxI32 diffTSC, mfxI32 diffRsCsdiff, mfxU32 posBalance, mfxU32 SC, mfxU32 mu_mv_mag_sq, mfxU32 negBalance, mfxI16 mcTcor, mfxU32 gchDC, mfxU32 Cs, mfxU32 ssDCval) {
        if (mu_mv_mag_sq < 91) {
            if (ssDCval < 17) {
                if (Rs < 1843) {
                    if (MVDiff < 388) {
                        return 8;
                    }
                    else if (MVDiff >= 388) {
                        return 4;
                    }
                }
                else if (Rs >= 1843) {
                    if (Rs < 4729) {
                        if (SC < 4448) {
                            if (SC < 3466) {
                                return 8;
                            }
                            else if (SC >= 3466) {
                                return 1;
                            }
                        }
                        else if (SC >= 4448) {
                            if (Cs < 3881) {
                                return 8;
                            }
                            else if (Cs >= 3881) {
                                if (SC < 7755) {
                                    if (Rs < 3755) {
                                        return 8;
                                    }
                                    else if (Rs >= 3755) {
                                        return 1;
                                    }
                                }
                                else if (SC >= 7755) {
                                    return 8;
                                }
                            }
                        }
                    }
                    else if (Rs >= 4729) {
                        if (SC < 10459) {
                            return 1;
                        }
                        else if (SC >= 10459) {
                            return 8;
                        }
                    }
                }
            }
            else if (ssDCval >= 17) {
                if (Cs < 3274) {
                    if (CsDiff < 10) {
                        if (RsCsDiff < 149) {
                            if (refDCval < 94) {
                                if (AFD < 8) {
                                    if (Cs < 542) {
                                        if (AFD < 7) {
                                            if (tcor < 84) {
                                                if (MVDiff < 31) {
                                                    if (posBalance < 65) {
                                                        if (refDCval < 38) {
                                                            if (Cs < 369) {
                                                                if (mcTcor < 81) {
                                                                    if (Rs < 423) {
                                                                        return 8;
                                                                    }
                                                                    else if (Rs >= 423) {
                                                                        return 1;
                                                                    }
                                                                }
                                                                else if (mcTcor >= 81) {
                                                                    if (Cs < 149) {
                                                                        return 2;
                                                                    }
                                                                    else if (Cs >= 149) {
                                                                        if (Cs < 330) {
                                                                            return 4;
                                                                        }
                                                                        else if (Cs >= 330) {
                                                                            if (diffMVdiffVal < -1) {
                                                                                return 4;
                                                                            }
                                                                            else if (diffMVdiffVal >= -1) {
                                                                                return 1;
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                            else if (Cs >= 369) {
                                                                if (mu_mv_mag_sq < 67) {
                                                                    if (SC < 829) {
                                                                        return 4;
                                                                    }
                                                                    else if (SC >= 829) {
                                                                        if (MVDiff < 24) {
                                                                            if (diffMVdiffVal < -5) {
                                                                                return 2;
                                                                            }
                                                                            else if (diffMVdiffVal >= -5) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                        else if (MVDiff >= 24) {
                                                                            return 2;
                                                                        }
                                                                    }
                                                                }
                                                                else if (mu_mv_mag_sq >= 67) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                        else if (refDCval >= 38) {
                                                            if (SC < 1143) {
                                                                if (Cs < 397) {
                                                                    if (negBalance < 49) {
                                                                        if (posBalance < 55) {
                                                                            return 4;
                                                                        }
                                                                        else if (posBalance >= 55) {
                                                                            if (AFD < 6) {
                                                                                if (mcTcor < 79) {
                                                                                    return 4;
                                                                                }
                                                                                else if (mcTcor >= 79) {
                                                                                    return 2;
                                                                                }
                                                                            }
                                                                            else if (AFD >= 6) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (negBalance >= 49) {
                                                                        if (diffMVdiffVal < -6) {
                                                                            if (Rs < 188) {
                                                                                return 4;
                                                                            }
                                                                            else if (Rs >= 188) {
                                                                                return 2;
                                                                            }
                                                                        }
                                                                        else if (diffMVdiffVal >= -6) {
                                                                            return 2;
                                                                        }
                                                                    }
                                                                }
                                                                else if (Cs >= 397) {
                                                                    return 2;
                                                                }
                                                            }
                                                            else if (SC >= 1143) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                    else if (posBalance >= 65) {
                                                        if (mu_mv_mag_sq < 23) {
                                                            if (tcor < 79) {
                                                                if (AFD < 2) {
                                                                    return 4;
                                                                }
                                                                else if (AFD >= 2) {
                                                                    if (Rs < 56) {
                                                                        return 2;
                                                                    }
                                                                    else if (Rs >= 56) {
                                                                        if (Cs < 106) {
                                                                            return 1;
                                                                        }
                                                                        else if (Cs >= 106) {
                                                                            return 2;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                            else if (tcor >= 79) {
                                                                if (posBalance < 115) {
                                                                    return 4;
                                                                }
                                                                else if (posBalance >= 115) {
                                                                    return 2;
                                                                }
                                                            }
                                                        }
                                                        else if (mu_mv_mag_sq >= 23) {
                                                            if (ssDCval < 44) {
                                                                if (ssDCval < 37) {
                                                                    return 4;
                                                                }
                                                                else if (ssDCval >= 37) {
                                                                    if (Rs < 123) {
                                                                        return 8;
                                                                    }
                                                                    else if (Rs >= 123) {
                                                                        return 4;
                                                                    }
                                                                }
                                                            }
                                                            else if (ssDCval >= 44) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (MVDiff >= 31) {
                                                    if (SC < 859) {
                                                        if (MVDiff < 865) {
                                                            if (diffAFD < -1) {
                                                                if (TSC < 1) {
                                                                    if (MVDiff < 200) {
                                                                        if (MVDiff < 158) {
                                                                            return 1;
                                                                        }
                                                                        else if (MVDiff >= 158) {
                                                                            return 2;
                                                                        }
                                                                    }
                                                                    else if (MVDiff >= 200) {
                                                                        if (MVDiff < 353) {
                                                                            return 8;
                                                                        }
                                                                        else if (MVDiff >= 353) {
                                                                            return 2;
                                                                        }
                                                                    }
                                                                }
                                                                else if (TSC >= 1) {
                                                                    if (RsDiff < 2) {
                                                                        return 1;
                                                                    }
                                                                    else if (RsDiff >= 2) {
                                                                        return 8;
                                                                    }
                                                                }
                                                            }
                                                            else if (diffAFD >= -1) {
                                                                if (negBalance < 1) {
                                                                    if (MVDiff < 55) {
                                                                        return 2;
                                                                    }
                                                                    else if (MVDiff >= 55) {
                                                                        if (posBalance < 99) {
                                                                            if (Rs < 4) {
                                                                                return 4;
                                                                            }
                                                                            else if (Rs >= 4) {
                                                                                return 2;
                                                                            }
                                                                        }
                                                                        else if (posBalance >= 99) {
                                                                            if (MVDiff < 109) {
                                                                                if (TSCindex < 2) {
                                                                                    return 8;
                                                                                }
                                                                                else if (TSCindex >= 2) {
                                                                                    return 4;
                                                                                }
                                                                            }
                                                                            else if (MVDiff >= 109) {
                                                                                if (AFD < 2) {
                                                                                    if (mu_mv_mag_sq < 12) {
                                                                                        if (AFD < 1) {
                                                                                            return 4;
                                                                                        }
                                                                                        else if (AFD >= 1) {
                                                                                            return 1;
                                                                                        }
                                                                                    }
                                                                                    else if (mu_mv_mag_sq >= 12) {
                                                                                        return 4;
                                                                                    }
                                                                                }
                                                                                else if (AFD >= 2) {
                                                                                    if (diffTSC < 1) {
                                                                                        if (Rs < 5) {
                                                                                            return 8;
                                                                                        }
                                                                                        else if (Rs >= 5) {
                                                                                            return 4;
                                                                                        }
                                                                                    }
                                                                                    else if (diffTSC >= 1) {
                                                                                        return 1;
                                                                                    }
                                                                                }
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                                else if (negBalance >= 1) {
                                                                    if (diffMVdiffVal < 40) {
                                                                        if (negBalance < 46) {
                                                                            if (negBalance < 4) {
                                                                                if (tcor < 37) {
                                                                                    if (MVDiff < 104) {
                                                                                        if (MVDiff < 66) {
                                                                                            return 8;
                                                                                        }
                                                                                        else if (MVDiff >= 66) {
                                                                                            return 4;
                                                                                        }
                                                                                    }
                                                                                    else if (MVDiff >= 104) {
                                                                                        return 8;
                                                                                    }
                                                                                }
                                                                                else if (tcor >= 37) {
                                                                                    if (tcor < 51) {
                                                                                        if (diffMVdiffVal < -53) {
                                                                                            if (MVDiff < 83) {
                                                                                                return 8;
                                                                                            }
                                                                                            else if (MVDiff >= 83) {
                                                                                                if (gchDC < 1) {
                                                                                                    return 2;
                                                                                                }
                                                                                                else if (gchDC >= 1) {
                                                                                                    return 1;
                                                                                                }
                                                                                            }
                                                                                        }
                                                                                        else if (diffMVdiffVal >= -53) {
                                                                                            if (negBalance < 3) {
                                                                                                if (posBalance < 23) {
                                                                                                    if (MVDiff < 61) {
                                                                                                        return 1;
                                                                                                    }
                                                                                                    else if (MVDiff >= 61) {
                                                                                                        if (diffMVdiffVal < -12) {
                                                                                                            return 4;
                                                                                                        }
                                                                                                        else if (diffMVdiffVal >= -12) {
                                                                                                            return 8;
                                                                                                        }
                                                                                                    }
                                                                                                }
                                                                                                else if (posBalance >= 23) {
                                                                                                    if (diffMVdiffVal < -3) {
                                                                                                        return 1;
                                                                                                    }
                                                                                                    else if (diffMVdiffVal >= -3) {
                                                                                                        if (MVDiff < 61) {
                                                                                                            return 1;
                                                                                                        }
                                                                                                        else if (MVDiff >= 61) {
                                                                                                            if (AFD < 1) {
                                                                                                                return 4;
                                                                                                            }
                                                                                                            else if (AFD >= 1) {
                                                                                                                return 1;
                                                                                                            }
                                                                                                        }
                                                                                                    }
                                                                                                }
                                                                                            }
                                                                                            else if (negBalance >= 3) {
                                                                                                return 4;
                                                                                            }
                                                                                        }
                                                                                    }
                                                                                    else if (tcor >= 51) {
                                                                                        if (Rs < 17) {
                                                                                            return 4;
                                                                                        }
                                                                                        else if (Rs >= 17) {
                                                                                            return 8;
                                                                                        }
                                                                                    }
                                                                                }
                                                                            }
                                                                            else if (negBalance >= 4) {
                                                                                if (posBalance < 55) {
                                                                                    return 4;
                                                                                }
                                                                                else if (posBalance >= 55) {
                                                                                    if (Cs < 221) {
                                                                                        return 1;
                                                                                    }
                                                                                    else if (Cs >= 221) {
                                                                                        if (Cs < 329) {
                                                                                            return 4;
                                                                                        }
                                                                                        else if (Cs >= 329) {
                                                                                            return 1;
                                                                                        }
                                                                                    }
                                                                                }
                                                                            }
                                                                        }
                                                                        else if (negBalance >= 46) {
                                                                            if (Cs < 392) {
                                                                                return 8;
                                                                            }
                                                                            else if (Cs >= 392) {
                                                                                if (AFD < 6) {
                                                                                    return 2;
                                                                                }
                                                                                else if (AFD >= 6) {
                                                                                    return 4;
                                                                                }
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (diffMVdiffVal >= 40) {
                                                                        if (negBalance < 33) {
                                                                            if (diffAFD < 1) {
                                                                                if (tcor < 76) {
                                                                                    return 8;
                                                                                }
                                                                                else if (tcor >= 76) {
                                                                                    return 2;
                                                                                }
                                                                            }
                                                                            else if (diffAFD >= 1) {
                                                                                return 8;
                                                                            }
                                                                        }
                                                                        else if (negBalance >= 33) {
                                                                            return 2;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (MVDiff >= 865) {
                                                            if (Rs < 12) {
                                                                if (mu_mv_mag_sq < 22) {
                                                                    if (posBalance < 84) {
                                                                        return 4;
                                                                    }
                                                                    else if (posBalance >= 84) {
                                                                        return 2;
                                                                    }
                                                                }
                                                                else if (mu_mv_mag_sq >= 22) {
                                                                    return 2;
                                                                }
                                                            }
                                                            else if (Rs >= 12) {
                                                                return 1;
                                                            }
                                                        }
                                                    }
                                                    else if (SC >= 859) {
                                                        if (Rs < 640) {
                                                            return 2;
                                                        }
                                                        else if (Rs >= 640) {
                                                            if (Cs < 496) {
                                                                return 4;
                                                            }
                                                            else if (Cs >= 496) {
                                                                return 1;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                            else if (tcor >= 84) {
                                                if (CsDiff < 1) {
                                                    if (MVDiff < 8) {
                                                        if (ssDCval < 74) {
                                                            if (refDCval < 39) {
                                                                return 4;
                                                            }
                                                            else if (refDCval >= 39) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (ssDCval >= 74) {
                                                            if (MVDiff < 2) {
                                                                return 4;
                                                            }
                                                            else if (MVDiff >= 2) {
                                                                if (SC < 623) {
                                                                    return 8;
                                                                }
                                                                else if (SC >= 623) {
                                                                    if (Rs < 334) {
                                                                        return 4;
                                                                    }
                                                                    else if (Rs >= 334) {
                                                                        return 8;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (MVDiff >= 8) {
                                                        if (diffAFD < -1) {
                                                            if (negBalance < 48) {
                                                                if (MVDiff < 34) {
                                                                    if (diffTSC < -1) {
                                                                        return 2;
                                                                    }
                                                                    else if (diffTSC >= -1) {
                                                                        if (diffMVdiffVal < -9) {
                                                                            return 2;
                                                                        }
                                                                        else if (diffMVdiffVal >= -9) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                }
                                                                else if (MVDiff >= 34) {
                                                                    if (MVDiff < 83) {
                                                                        return 8;
                                                                    }
                                                                    else if (MVDiff >= 83) {
                                                                        return 2;
                                                                    }
                                                                }
                                                            }
                                                            else if (negBalance >= 48) {
                                                                if (MVDiff < 93) {
                                                                    if (Rs < 121) {
                                                                        return 2;
                                                                    }
                                                                    else if (Rs >= 121) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (MVDiff >= 93) {
                                                                    return 1;
                                                                }
                                                            }
                                                        }
                                                        else if (diffAFD >= -1) {
                                                            if (tcor < 101) {
                                                                if (Rs < 5) {
                                                                    return 8;
                                                                }
                                                                else if (Rs >= 5) {
                                                                    if (Rs < 308) {
                                                                        return 2;
                                                                    }
                                                                    else if (Rs >= 308) {
                                                                        if (MVDiff < 19) {
                                                                            return 8;
                                                                        }
                                                                        else if (MVDiff >= 19) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                            else if (tcor >= 101) {
                                                                if (posBalance < 5) {
                                                                    if (Cs < 1) {
                                                                        if (posBalance < 3) {
                                                                            if (diffMVdiffVal < -12) {
                                                                                if (refDCval < 22) {
                                                                                    return 1;
                                                                                }
                                                                                else if (refDCval >= 22) {
                                                                                    return 2;
                                                                                }
                                                                            }
                                                                            else if (diffMVdiffVal >= -12) {
                                                                                if (posBalance < 2) {
                                                                                    if (MVDiff < 92) {
                                                                                        return 4;
                                                                                    }
                                                                                    else if (MVDiff >= 92) {
                                                                                        return 8;
                                                                                    }
                                                                                }
                                                                                else if (posBalance >= 2) {
                                                                                    return 4;
                                                                                }
                                                                            }
                                                                        }
                                                                        else if (posBalance >= 3) {
                                                                            return 2;
                                                                        }
                                                                    }
                                                                    else if (Cs >= 1) {
                                                                        return 8;
                                                                    }
                                                                }
                                                                else if (posBalance >= 5) {
                                                                    if (gchDC < 1) {
                                                                        if (negBalance < 11) {
                                                                            return 8;
                                                                        }
                                                                        else if (negBalance >= 11) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                    else if (gchDC >= 1) {
                                                                        if (mu_mv_mag_sq < 30) {
                                                                            if (Rs < 3) {
                                                                                return 4;
                                                                            }
                                                                            else if (Rs >= 3) {
                                                                                return 2;
                                                                            }
                                                                        }
                                                                        else if (mu_mv_mag_sq >= 30) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (CsDiff >= 1) {
                                                    if (SC < 2912) {
                                                        if (ssDCval < 38) {
                                                            if (negBalance < 42) {
                                                                if (TSC < 2) {
                                                                    return 2;
                                                                }
                                                                else if (TSC >= 2) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (negBalance >= 42) {
                                                                if (mcTcor < 99) {
                                                                    if (Cs < 382) {
                                                                        if (MVDiff < 23) {
                                                                            return 4;
                                                                        }
                                                                        else if (MVDiff >= 23) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                    else if (Cs >= 382) {
                                                                        return 2;
                                                                    }
                                                                }
                                                                else if (mcTcor >= 99) {
                                                                    return 2;
                                                                }
                                                            }
                                                        }
                                                        else if (ssDCval >= 38) {
                                                            if (tcor < 90) {
                                                                if (diffMVdiffVal < 4) {
                                                                    return 4;
                                                                }
                                                                else if (diffMVdiffVal >= 4) {
                                                                    if (posBalance < 55) {
                                                                        return 4;
                                                                    }
                                                                    else if (posBalance >= 55) {
                                                                        if (Cs < 381) {
                                                                            return 4;
                                                                        }
                                                                        else if (Cs >= 381) {
                                                                            return 2;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                            else if (tcor >= 90) {
                                                                if (gchDC < 5) {
                                                                    if (RsDiff < 2) {
                                                                        if (MVDiff < 5) {
                                                                            if (refDCval < 61) {
                                                                                if (negBalance < 38) {
                                                                                    if (Rs < 574) {
                                                                                        return 4;
                                                                                    }
                                                                                    else if (Rs >= 574) {
                                                                                        return 8;
                                                                                    }
                                                                                }
                                                                                else if (negBalance >= 38) {
                                                                                    return 4;
                                                                                }
                                                                            }
                                                                            else if (refDCval >= 61) {
                                                                                return 8;
                                                                            }
                                                                        }
                                                                        else if (MVDiff >= 5) {
                                                                            if (diffTSC < 1) {
                                                                                return 4;
                                                                            }
                                                                            else if (diffTSC >= 1) {
                                                                                if (Cs < 325) {
                                                                                    return 4;
                                                                                }
                                                                                else if (Cs >= 325) {
                                                                                    return 2;
                                                                                }
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (RsDiff >= 2) {
                                                                        if (Cs < 388) {
                                                                            if (SC < 2801) {
                                                                                return 2;
                                                                            }
                                                                            else if (SC >= 2801) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                        else if (Cs >= 388) {
                                                                            if (Rs < 643) {
                                                                                return 4;
                                                                            }
                                                                            else if (Rs >= 643) {
                                                                                return 8;
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                                else if (gchDC >= 5) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (SC >= 2912) {
                                                        return 2;
                                                    }
                                                }
                                            }
                                        }
                                        else if (AFD >= 7) {
                                            if (gchDC < 5) {
                                                if (Cs < 344) {
                                                    return 1;
                                                }
                                                else if (Cs >= 344) {
                                                    if (mu_mv_mag_sq < 30) {
                                                        if (RsDiff < 3) {
                                                            return 4;
                                                        }
                                                        else if (RsDiff >= 3) {
                                                            return 2;
                                                        }
                                                    }
                                                    else if (mu_mv_mag_sq >= 30) {
                                                        if (Rs < 1660) {
                                                            return 1;
                                                        }
                                                        else if (Rs >= 1660) {
                                                            return 2;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (gchDC >= 5) {
                                                if (Rs < 272) {
                                                    if (Rs < 104) {
                                                        return 2;
                                                    }
                                                    else if (Rs >= 104) {
                                                        return 4;
                                                    }
                                                }
                                                else if (Rs >= 272) {
                                                    return 8;
                                                }
                                            }
                                        }
                                    }
                                    else if (Cs >= 542) {
                                        if (refDCval < 75) {
                                            if (ssDCval < 39) {
                                                if (TSC < 5) {
                                                    if (mu_mv_mag_sq < 52) {
                                                        if (Cs < 1541) {
                                                            if (SC < 1376) {
                                                                if (MVDiff < 48) {
                                                                    if (SC < 1353) {
                                                                        return 8;
                                                                    }
                                                                    else if (SC >= 1353) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (MVDiff >= 48) {
                                                                    return 2;
                                                                }
                                                            }
                                                            else if (SC >= 1376) {
                                                                return 2;
                                                            }
                                                        }
                                                        else if (Cs >= 1541) {
                                                            if (Rs < 1736) {
                                                                return 8;
                                                            }
                                                            else if (Rs >= 1736) {
                                                                if (TSC < 2) {
                                                                    return 2;
                                                                }
                                                                else if (TSC >= 2) {
                                                                    return 1;
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (mu_mv_mag_sq >= 52) {
                                                        if (Cs < 668) {
                                                            return 4;
                                                        }
                                                        else if (Cs >= 668) {
                                                            if (refDCval < 31) {
                                                                return 4;
                                                            }
                                                            else if (refDCval >= 31) {
                                                                if (mu_mv_mag_sq < 63) {
                                                                    return 1;
                                                                }
                                                                else if (mu_mv_mag_sq >= 63) {
                                                                    if (RsDiff < 1) {
                                                                        return 1;
                                                                    }
                                                                    else if (RsDiff >= 1) {
                                                                        if (diffAFD < -2) {
                                                                            return 1;
                                                                        }
                                                                        else if (diffAFD >= -2) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (TSC >= 5) {
                                                    if (Rs < 860) {
                                                        if (mu_mv_mag_sq < 57) {
                                                            if (posBalance < 56) {
                                                                return 4;
                                                            }
                                                            else if (posBalance >= 56) {
                                                                if (CsDiff < 2) {
                                                                    if (diffAFD < 1) {
                                                                        if (diffMVdiffVal < -6) {
                                                                            return 1;
                                                                        }
                                                                        else if (diffMVdiffVal >= -6) {
                                                                            if (MVDiff < 35) {
                                                                                return 4;
                                                                            }
                                                                            else if (MVDiff >= 35) {
                                                                                return 1;
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (diffAFD >= 1) {
                                                                        return 1;
                                                                    }
                                                                }
                                                                else if (CsDiff >= 2) {
                                                                    return 1;
                                                                }
                                                            }
                                                        }
                                                        else if (mu_mv_mag_sq >= 57) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (Rs >= 860) {
                                                        if (mcTcor < 81) {
                                                            return 2;
                                                        }
                                                        else if (mcTcor >= 81) {
                                                            if (AFD < 7) {
                                                                return 1;
                                                            }
                                                            else if (AFD >= 7) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                            else if (ssDCval >= 39) {
                                                if (refDCval < 54) {
                                                    if (SC < 3344) {
                                                        if (Rs < 1142) {
                                                            if (Cs < 937) {
                                                                if (negBalance < 42) {
                                                                    if (negBalance < 36) {
                                                                        if (posBalance < 53) {
                                                                            if (Rs < 1054) {
                                                                                if (refDCval < 46) {
                                                                                    if (Rs < 747) {
                                                                                        return 4;
                                                                                    }
                                                                                    else if (Rs >= 747) {
                                                                                        return 1;
                                                                                    }
                                                                                }
                                                                                else if (refDCval >= 46) {
                                                                                    return 4;
                                                                                }
                                                                            }
                                                                            else if (Rs >= 1054) {
                                                                                return 1;
                                                                            }
                                                                        }
                                                                        else if (posBalance >= 53) {
                                                                            if (ssDCval < 48) {
                                                                                return 4;
                                                                            }
                                                                            else if (ssDCval >= 48) {
                                                                                if (mu_mv_mag_sq < 57) {
                                                                                    return 8;
                                                                                }
                                                                                else if (mu_mv_mag_sq >= 57) {
                                                                                    return 4;
                                                                                }
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (negBalance >= 36) {
                                                                        if (Rs < 1128) {
                                                                            return 4;
                                                                        }
                                                                        else if (Rs >= 1128) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                }
                                                                else if (negBalance >= 42) {
                                                                    if (refDCval < 51) {
                                                                        if (gchDC < 1) {
                                                                            if (negBalance < 48) {
                                                                                if (MVDiff < 7) {
                                                                                    return 1;
                                                                                }
                                                                                else if (MVDiff >= 7) {
                                                                                    return 4;
                                                                                }
                                                                            }
                                                                            else if (negBalance >= 48) {
                                                                                if (mcTcor < 80) {
                                                                                    if (SC < 1421) {
                                                                                        if (diffAFD < -1) {
                                                                                            return 4;
                                                                                        }
                                                                                        else if (diffAFD >= -1) {
                                                                                            return 1;
                                                                                        }
                                                                                    }
                                                                                    else if (SC >= 1421) {
                                                                                        if (Rs < 863) {
                                                                                            return 4;
                                                                                        }
                                                                                        else if (Rs >= 863) {
                                                                                            return 1;
                                                                                        }
                                                                                    }
                                                                                }
                                                                                else if (mcTcor >= 80) {
                                                                                    return 1;
                                                                                }
                                                                            }
                                                                        }
                                                                        else if (gchDC >= 1) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                    else if (refDCval >= 51) {
                                                                        if (RsDiff < 3) {
                                                                            if (Rs < 1107) {
                                                                                if (TSC < 5) {
                                                                                    return 8;
                                                                                }
                                                                                else if (TSC >= 5) {
                                                                                    if (SC < 1599) {
                                                                                        return 8;
                                                                                    }
                                                                                    else if (SC >= 1599) {
                                                                                        return 1;
                                                                                    }
                                                                                }
                                                                            }
                                                                            else if (Rs >= 1107) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                        else if (RsDiff >= 3) {
                                                                            if (SC < 1937) {
                                                                                return 4;
                                                                            }
                                                                            else if (SC >= 1937) {
                                                                                return 1;
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                            else if (Cs >= 937) {
                                                                if (AFD < 3) {
                                                                    return 8;
                                                                }
                                                                else if (AFD >= 3) {
                                                                    if (MVDiff < 7) {
                                                                        return 4;
                                                                    }
                                                                    else if (MVDiff >= 7) {
                                                                        return 2;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (Rs >= 1142) {
                                                            if (negBalance < 35) {
                                                                return 1;
                                                            }
                                                            else if (negBalance >= 35) {
                                                                if (ssDCval < 50) {
                                                                    if (negBalance < 40) {
                                                                        return 1;
                                                                    }
                                                                    else if (negBalance >= 40) {
                                                                        if (mu_mv_mag_sq < 47) {
                                                                            if (mcTcor < 94) {
                                                                                return 8;
                                                                            }
                                                                            else if (mcTcor >= 94) {
                                                                                if (posBalance < 27) {
                                                                                    return 8;
                                                                                }
                                                                                else if (posBalance >= 27) {
                                                                                    return 1;
                                                                                }
                                                                            }
                                                                        }
                                                                        else if (mu_mv_mag_sq >= 47) {
                                                                            if (Rs < 1741) {
                                                                                return 8;
                                                                            }
                                                                            else if (Rs >= 1741) {
                                                                                if (tcor < 92) {
                                                                                    return 4;
                                                                                }
                                                                                else if (tcor >= 92) {
                                                                                    return 8;
                                                                                }
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                                else if (ssDCval >= 50) {
                                                                    if (posBalance < 48) {
                                                                        if (tcor < 92) {
                                                                            if (negBalance < 37) {
                                                                                return 1;
                                                                            }
                                                                            else if (negBalance >= 37) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                        else if (tcor >= 92) {
                                                                            if (negBalance < 48) {
                                                                                return 1;
                                                                            }
                                                                            else if (negBalance >= 48) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (posBalance >= 48) {
                                                                        if (Rs < 1459) {
                                                                            if (Cs < 905) {
                                                                                return 1;
                                                                            }
                                                                            else if (Cs >= 905) {
                                                                                if (Rs < 1405) {
                                                                                    if (tcor < 88) {
                                                                                        return 1;
                                                                                    }
                                                                                    else if (tcor >= 88) {
                                                                                        if (mu_mv_mag_sq < 63) {
                                                                                            if (Rs < 1343) {
                                                                                                return 1;
                                                                                            }
                                                                                            else if (Rs >= 1343) {
                                                                                                return 4;
                                                                                            }
                                                                                        }
                                                                                        else if (mu_mv_mag_sq >= 63) {
                                                                                            return 4;
                                                                                        }
                                                                                    }
                                                                                }
                                                                                else if (Rs >= 1405) {
                                                                                    return 1;
                                                                                }
                                                                            }
                                                                        }
                                                                        else if (Rs >= 1459) {
                                                                            if (posBalance < 51) {
                                                                                return 1;
                                                                            }
                                                                            else if (posBalance >= 51) {
                                                                                if (Cs < 1073) {
                                                                                    if (tcor < 88) {
                                                                                        return 8;
                                                                                    }
                                                                                    else if (tcor >= 88) {
                                                                                        return 4;
                                                                                    }
                                                                                }
                                                                                else if (Cs >= 1073) {
                                                                                    return 1;
                                                                                }
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (SC >= 3344) {
                                                        if (mu_mv_mag_sq < 65) {
                                                            if (refDCval < 40) {
                                                                if (Cs < 1525) {
                                                                    return 4;
                                                                }
                                                                else if (Cs >= 1525) {
                                                                    if (Cs < 1565) {
                                                                        if (tcor < 92) {
                                                                            if (posBalance < 57) {
                                                                                if (tcor < 91) {
                                                                                    return 4;
                                                                                }
                                                                                else if (tcor >= 91) {
                                                                                    if (Rs < 1832) {
                                                                                        return 1;
                                                                                    }
                                                                                    else if (Rs >= 1832) {
                                                                                        return 4;
                                                                                    }
                                                                                }
                                                                            }
                                                                            else if (posBalance >= 57) {
                                                                                return 1;
                                                                            }
                                                                        }
                                                                        else if (tcor >= 92) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                    else if (Cs >= 1565) {
                                                                        return 1;
                                                                    }
                                                                }
                                                            }
                                                            else if (refDCval >= 40) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (mu_mv_mag_sq >= 65) {
                                                            if (SC < 3465) {
                                                                return 4;
                                                            }
                                                            else if (SC >= 3465) {
                                                                if (diffMVdiffVal < -7) {
                                                                    return 1;
                                                                }
                                                                else if (diffMVdiffVal >= -7) {
                                                                    if (SC < 3469) {
                                                                        return 1;
                                                                    }
                                                                    else if (SC >= 3469) {
                                                                        return 4;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (refDCval >= 54) {
                                                    if (SC < 1493) {
                                                        if (ssDCval < 63) {
                                                            if (TSC < 4) {
                                                                return 8;
                                                            }
                                                            else if (TSC >= 4) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (ssDCval >= 63) {
                                                            return 2;
                                                        }
                                                    }
                                                    else if (SC >= 1493) {
                                                        if (TSC < 5) {
                                                            if (RsCsDiff < 12) {
                                                                if (mu_mv_mag_sq < 76) {
                                                                    if (gchDC < 1) {
                                                                        if (AFD < 6) {
                                                                            if (MVDiff < 1) {
                                                                                return 1;
                                                                            }
                                                                            else if (MVDiff >= 1) {
                                                                                return 8;
                                                                            }
                                                                        }
                                                                        else if (AFD >= 6) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                    else if (gchDC >= 1) {
                                                                        if (SC < 1744) {
                                                                            return 8;
                                                                        }
                                                                        else if (SC >= 1744) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                }
                                                                else if (mu_mv_mag_sq >= 76) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (RsCsDiff >= 12) {
                                                                if (ssDCval < 57) {
                                                                    if (Rs < 1448) {
                                                                        if (Rs < 1394) {
                                                                            return 8;
                                                                        }
                                                                        else if (Rs >= 1394) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                    else if (Rs >= 1448) {
                                                                        return 8;
                                                                    }
                                                                }
                                                                else if (ssDCval >= 57) {
                                                                    if (CsDiff < 5) {
                                                                        return 4;
                                                                    }
                                                                    else if (CsDiff >= 5) {
                                                                        return 1;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (TSC >= 5) {
                                                            if (posBalance < 57) {
                                                                if (negBalance < 27) {
                                                                    return 1;
                                                                }
                                                                else if (negBalance >= 27) {
                                                                    if (posBalance < 47) {
                                                                        return 4;
                                                                    }
                                                                    else if (posBalance >= 47) {
                                                                        if (Rs < 1569) {
                                                                            if (Rs < 1359) {
                                                                                if (diffMVdiffVal < 0) {
                                                                                    if (diffMVdiffVal < -2) {
                                                                                        if (SC < 1775) {
                                                                                            return 8;
                                                                                        }
                                                                                        else if (SC >= 1775) {
                                                                                            return 1;
                                                                                        }
                                                                                    }
                                                                                    else if (diffMVdiffVal >= -2) {
                                                                                        return 4;
                                                                                    }
                                                                                }
                                                                                else if (diffMVdiffVal >= 0) {
                                                                                    return 1;
                                                                                }
                                                                            }
                                                                            else if (Rs >= 1359) {
                                                                                return 8;
                                                                            }
                                                                        }
                                                                        else if (Rs >= 1569) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                            else if (posBalance >= 57) {
                                                                if (diffTSC < 1) {
                                                                    if (Rs < 2025) {
                                                                        return 1;
                                                                    }
                                                                    else if (Rs >= 2025) {
                                                                        if (RsDiff < 5) {
                                                                            return 4;
                                                                        }
                                                                        else if (RsDiff >= 5) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                }
                                                                else if (diffTSC >= 1) {
                                                                    if (Rs < 1119) {
                                                                        return 4;
                                                                    }
                                                                    else if (Rs >= 1119) {
                                                                        return 8;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (refDCval >= 75) {
                                            if (Rs < 2429) {
                                                if (TSC < 7) {
                                                    return 4;
                                                }
                                                else if (TSC >= 7) {
                                                    return 2;
                                                }
                                            }
                                            else if (Rs >= 2429) {
                                                if (Cs < 2120) {
                                                    return 1;
                                                }
                                                else if (Cs >= 2120) {
                                                    return 8;
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (AFD >= 8) {
                                    if (Cs < 796) {
                                        if (SC < 818) {
                                            if (TSCindex < 7) {
                                                if (AFD < 10) {
                                                    if (TSCindex < 5) {
                                                        return 2;
                                                    }
                                                    else if (TSCindex >= 5) {
                                                        if (gchDC < 7) {
                                                            if (diffMVdiffVal < 7) {
                                                                return 1;
                                                            }
                                                            else if (diffMVdiffVal >= 7) {
                                                                if (Cs < 49) {
                                                                    return 1;
                                                                }
                                                                else if (Cs >= 49) {
                                                                    return 2;
                                                                }
                                                            }
                                                        }
                                                        else if (gchDC >= 7) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                                else if (AFD >= 10) {
                                                    if (Rs < 100) {
                                                        return 1;
                                                    }
                                                    else if (Rs >= 100) {
                                                        return 2;
                                                    }
                                                }
                                            }
                                            else if (TSCindex >= 7) {
                                                if (Cs < 227) {
                                                    if (MVDiff < 351) {
                                                        return 1;
                                                    }
                                                    else if (MVDiff >= 351) {
                                                        return 2;
                                                    }
                                                }
                                                else if (Cs >= 227) {
                                                    if (TSC < 9) {
                                                        if (Cs < 261) {
                                                            if (Rs < 312) {
                                                                return 1;
                                                            }
                                                            else if (Rs >= 312) {
                                                                return 2;
                                                            }
                                                        }
                                                        else if (Cs >= 261) {
                                                            return 2;
                                                        }
                                                    }
                                                    else if (TSC >= 9) {
                                                        if (Rs < 419) {
                                                            return 1;
                                                        }
                                                        else if (Rs >= 419) {
                                                            return 2;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (SC >= 818) {
                                            if (RsCsDiff < 23) {
                                                if (gchDC < 2) {
                                                    if (MVDiff < 59) {
                                                        if (tcor < 89) {
                                                            if (diffMVdiffVal < 7) {
                                                                if (MVDiff < 41) {
                                                                    return 1;
                                                                }
                                                                else if (MVDiff >= 41) {
                                                                    if (Cs < 559) {
                                                                        if (Rs < 772) {
                                                                            return 4;
                                                                        }
                                                                        else if (Rs >= 772) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                    else if (Cs >= 559) {
                                                                        return 1;
                                                                    }
                                                                }
                                                            }
                                                            else if (diffMVdiffVal >= 7) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (tcor >= 89) {
                                                            if (Cs < 477) {
                                                                return 8;
                                                            }
                                                            else if (Cs >= 477) {
                                                                return 2;
                                                            }
                                                        }
                                                    }
                                                    else if (MVDiff >= 59) {
                                                        if (SC < 1143) {
                                                            return 8;
                                                        }
                                                        else if (SC >= 1143) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                                else if (gchDC >= 2) {
                                                    if (Rs < 514) {
                                                        if (posBalance < 63) {
                                                            return 2;
                                                        }
                                                        else if (posBalance >= 63) {
                                                            if (Rs < 382) {
                                                                return 1;
                                                            }
                                                            else if (Rs >= 382) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                    else if (Rs >= 514) {
                                                        if (ssDCval < 69) {
                                                            if (ssDCval < 61) {
                                                                return 8;
                                                            }
                                                            else if (ssDCval >= 61) {
                                                                return 2;
                                                            }
                                                        }
                                                        else if (ssDCval >= 69) {
                                                            if (RsCsDiff < 8) {
                                                                return 4;
                                                            }
                                                            else if (RsCsDiff >= 8) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                            else if (RsCsDiff >= 23) {
                                                if (tcor < 68) {
                                                    if (mcTcor < 87) {
                                                        if (AFD < 13) {
                                                            return 1;
                                                        }
                                                        else if (AFD >= 13) {
                                                            if (Cs < 424) {
                                                                return 2;
                                                            }
                                                            else if (Cs >= 424) {
                                                                if (Rs < 1021) {
                                                                    if (TSC < 11) {
                                                                        return 2;
                                                                    }
                                                                    else if (TSC >= 11) {
                                                                        return 1;
                                                                    }
                                                                }
                                                                else if (Rs >= 1021) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (mcTcor >= 87) {
                                                        return 2;
                                                    }
                                                }
                                                else if (tcor >= 68) {
                                                    if (posBalance < 59) {
                                                        if (negBalance < 39) {
                                                            if (Rs < 2072) {
                                                                return 4;
                                                            }
                                                            else if (Rs >= 2072) {
                                                                return 2;
                                                            }
                                                        }
                                                        else if (negBalance >= 39) {
                                                            if (RsDiff < 8) {
                                                                return 2;
                                                            }
                                                            else if (RsDiff >= 8) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                    else if (posBalance >= 59) {
                                                        if (mcTcor < 92) {
                                                            if (posBalance < 60) {
                                                                return 8;
                                                            }
                                                            else if (posBalance >= 60) {
                                                                return 1;
                                                            }
                                                        }
                                                        else if (mcTcor >= 92) {
                                                            if (Rs < 1250) {
                                                                return 2;
                                                            }
                                                            else if (Rs >= 1250) {
                                                                if (Cs < 660) {
                                                                    return 2;
                                                                }
                                                                else if (Cs >= 660) {
                                                                    return 1;
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (Cs >= 796) {
                                        if (MVDiff < 5) {
                                            if (TSC < 7) {
                                                if (negBalance < 27) {
                                                    return 1;
                                                }
                                                else if (negBalance >= 27) {
                                                    if (mu_mv_mag_sq < 81) {
                                                        if (RsCsDiff < 41) {
                                                            if (Rs < 1046) {
                                                                return 8;
                                                            }
                                                            else if (Rs >= 1046) {
                                                                if (Cs < 1758) {
                                                                    return 2;
                                                                }
                                                                else if (Cs >= 1758) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                        else if (RsCsDiff >= 41) {
                                                            if (mcTcor < 93) {
                                                                if (diffRsCsdiff < 10) {
                                                                    return 4;
                                                                }
                                                                else if (diffRsCsdiff >= 10) {
                                                                    if (SC < 3070) {
                                                                        return 1;
                                                                    }
                                                                    else if (SC >= 3070) {
                                                                        return 4;
                                                                    }
                                                                }
                                                            }
                                                            else if (mcTcor >= 93) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                    else if (mu_mv_mag_sq >= 81) {
                                                        if (Rs < 2067) {
                                                            return 2;
                                                        }
                                                        else if (Rs >= 2067) {
                                                            if (Rs < 2424) {
                                                                return 4;
                                                            }
                                                            else if (Rs >= 2424) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                            else if (TSC >= 7) {
                                                if (Rs < 2401) {
                                                    if (negBalance < 45) {
                                                        if (Cs < 2014) {
                                                            return 4;
                                                        }
                                                        else if (Cs >= 2014) {
                                                            if (AFD < 10) {
                                                                return 8;
                                                            }
                                                            else if (AFD >= 10) {
                                                                return 1;
                                                            }
                                                        }
                                                    }
                                                    else if (negBalance >= 45) {
                                                        if (diffMVdiffVal < -3) {
                                                            if (negBalance < 53) {
                                                                return 4;
                                                            }
                                                            else if (negBalance >= 53) {
                                                                if (Rs < 1865) {
                                                                    return 2;
                                                                }
                                                                else if (Rs >= 1865) {
                                                                    return 1;
                                                                }
                                                            }
                                                        }
                                                        else if (diffMVdiffVal >= -3) {
                                                            if (mcTcor < 92) {
                                                                if (Rs < 1756) {
                                                                    return 4;
                                                                }
                                                                else if (Rs >= 1756) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (mcTcor >= 92) {
                                                                if (Cs < 984) {
                                                                    return 4;
                                                                }
                                                                else if (Cs >= 984) {
                                                                    return 2;
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (Rs >= 2401) {
                                                    if (diffMVdiffVal < -2) {
                                                        if (mcTcor < 95) {
                                                            if (Rs < 2671) {
                                                                return 2;
                                                            }
                                                            else if (Rs >= 2671) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (mcTcor >= 95) {
                                                            return 1;
                                                        }
                                                    }
                                                    else if (diffMVdiffVal >= -2) {
                                                        if (mu_mv_mag_sq < 84) {
                                                            return 4;
                                                        }
                                                        else if (mu_mv_mag_sq >= 84) {
                                                            if (Rs < 2729) {
                                                                return 4;
                                                            }
                                                            else if (Rs >= 2729) {
                                                                if (Rs < 3086) {
                                                                    return 1;
                                                                }
                                                                else if (Rs >= 3086) {
                                                                    return 2;
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (MVDiff >= 5) {
                                            if (refDCval < 81) {
                                                if (TSC < 11) {
                                                    if (refDCval < 61) {
                                                        if (MVDiff < 16) {
                                                            if (tcor < 82) {
                                                                if (diffRsCsdiff < 25) {
                                                                    if (negBalance < 45) {
                                                                        return 4;
                                                                    }
                                                                    else if (negBalance >= 45) {
                                                                        return 1;
                                                                    }
                                                                }
                                                                else if (diffRsCsdiff >= 25) {
                                                                    if (Cs < 969) {
                                                                        return 1;
                                                                    }
                                                                    else if (Cs >= 969) {
                                                                        if (Rs < 1100) {
                                                                            return 2;
                                                                        }
                                                                        else if (Rs >= 1100) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                            else if (tcor >= 82) {
                                                                if (posBalance < 59) {
                                                                    if (mu_mv_mag_sq < 39) {
                                                                        if (posBalance < 56) {
                                                                            return 8;
                                                                        }
                                                                        else if (posBalance >= 56) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                    else if (mu_mv_mag_sq >= 39) {
                                                                        if (mcTcor < 96) {
                                                                            if (Rs < 977) {
                                                                                return 8;
                                                                            }
                                                                            else if (Rs >= 977) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                        else if (mcTcor >= 96) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                }
                                                                else if (posBalance >= 59) {
                                                                    return 1;
                                                                }
                                                            }
                                                        }
                                                        else if (MVDiff >= 16) {
                                                            if (Rs < 3674) {
                                                                if (RsDiff < 4) {
                                                                    if (RsCsDiff < 16) {
                                                                        if (TSC < 7) {
                                                                            if (tcor < 93) {
                                                                                if (posBalance < 58) {
                                                                                    if (SC < 3979) {
                                                                                        return 4;
                                                                                    }
                                                                                    else if (SC >= 3979) {
                                                                                        return 8;
                                                                                    }
                                                                                }
                                                                                else if (posBalance >= 58) {
                                                                                    return 4;
                                                                                }
                                                                            }
                                                                            else if (tcor >= 93) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                        else if (TSC >= 7) {
                                                                            if (Rs < 2100) {
                                                                                return 4;
                                                                            }
                                                                            else if (Rs >= 2100) {
                                                                                return 2;
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (RsCsDiff >= 16) {
                                                                        if (MVDiff < 40) {
                                                                            if (MVDiff < 21) {
                                                                                if (Rs < 1181) {
                                                                                    if (Cs < 940) {
                                                                                        return 4;
                                                                                    }
                                                                                    else if (Cs >= 940) {
                                                                                        return 2;
                                                                                    }
                                                                                }
                                                                                else if (Rs >= 1181) {
                                                                                    return 8;
                                                                                }
                                                                            }
                                                                            else if (MVDiff >= 21) {
                                                                                if (diffMVdiffVal < -19) {
                                                                                    return 8;
                                                                                }
                                                                                else if (diffMVdiffVal >= -19) {
                                                                                    return 2;
                                                                                }
                                                                            }
                                                                        }
                                                                        else if (MVDiff >= 40) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                }
                                                                else if (RsDiff >= 4) {
                                                                    if (posBalance < 57) {
                                                                        if (diffRsCsdiff < -38) {
                                                                            return 2;
                                                                        }
                                                                        else if (diffRsCsdiff >= -38) {
                                                                            if (Cs < 1013) {
                                                                                if (MVDiff < 25) {
                                                                                    return 1;
                                                                                }
                                                                                else if (MVDiff >= 25) {
                                                                                    return 8;
                                                                                }
                                                                            }
                                                                            else if (Cs >= 1013) {
                                                                                return 8;
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (posBalance >= 57) {
                                                                        if (Rs < 986) {
                                                                            return 2;
                                                                        }
                                                                        else if (Rs >= 986) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                            else if (Rs >= 3674) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                    else if (refDCval >= 61) {
                                                        if (tcor < 90) {
                                                            if (negBalance < 53) {
                                                                if (TSC < 8) {
                                                                    if (tcor < 88) {
                                                                        if (ssDCval < 67) {
                                                                            return 1;
                                                                        }
                                                                        else if (ssDCval >= 67) {
                                                                            if (MVDiff < 9) {
                                                                                return 1;
                                                                            }
                                                                            else if (MVDiff >= 9) {
                                                                                return 2;
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (tcor >= 88) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (TSC >= 8) {
                                                                    if (mcTcor < 92) {
                                                                        if (posBalance < 55) {
                                                                            if (Rs < 2685) {
                                                                                return 8;
                                                                            }
                                                                            else if (Rs >= 2685) {
                                                                                if (Rs < 2804) {
                                                                                    return 1;
                                                                                }
                                                                                else if (Rs >= 2804) {
                                                                                    return 8;
                                                                                }
                                                                            }
                                                                        }
                                                                        else if (posBalance >= 55) {
                                                                            if (tcor < 81) {
                                                                                if (RsDiff < 3) {
                                                                                    return 2;
                                                                                }
                                                                                else if (RsDiff >= 3) {
                                                                                    return 8;
                                                                                }
                                                                            }
                                                                            else if (tcor >= 81) {
                                                                                if (Cs < 886) {
                                                                                    return 8;
                                                                                }
                                                                                else if (Cs >= 886) {
                                                                                    return 4;
                                                                                }
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (mcTcor >= 92) {
                                                                        if (mu_mv_mag_sq < 50) {
                                                                            return 1;
                                                                        }
                                                                        else if (mu_mv_mag_sq >= 50) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                            else if (negBalance >= 53) {
                                                                if (mcTcor < 89) {
                                                                    if (RsCsDiff < 73) {
                                                                        if (Cs < 1001) {
                                                                            if (SC < 1920) {
                                                                                return 8;
                                                                            }
                                                                            else if (SC >= 1920) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                        else if (Cs >= 1001) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                    else if (RsCsDiff >= 73) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (mcTcor >= 89) {
                                                                    if (SC < 2556) {
                                                                        return 4;
                                                                    }
                                                                    else if (SC >= 2556) {
                                                                        if (AFD < 14) {
                                                                            return 1;
                                                                        }
                                                                        else if (AFD >= 14) {
                                                                            return 2;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (tcor >= 90) {
                                                            if (SC < 4037) {
                                                                if (Rs < 1719) {
                                                                    if (SC < 2745) {
                                                                        if (SC < 1695) {
                                                                            return 2;
                                                                        }
                                                                        else if (SC >= 1695) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                    else if (SC >= 2745) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (Rs >= 1719) {
                                                                    return 1;
                                                                }
                                                            }
                                                            else if (SC >= 4037) {
                                                                return 2;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (TSC >= 11) {
                                                    if (refDCval < 78) {
                                                        if (RsCsDiff < 67) {
                                                            return 8;
                                                        }
                                                        else if (RsCsDiff >= 67) {
                                                            if (TSC < 12) {
                                                                if (Rs < 1571) {
                                                                    return 8;
                                                                }
                                                                else if (Rs >= 1571) {
                                                                    if (Rs < 2172) {
                                                                        return 4;
                                                                    }
                                                                    else if (Rs >= 2172) {
                                                                        return 8;
                                                                    }
                                                                }
                                                            }
                                                            else if (TSC >= 12) {
                                                                if (RsDiff < 9) {
                                                                    if (AFD < 20) {
                                                                        return 8;
                                                                    }
                                                                    else if (AFD >= 20) {
                                                                        return 1;
                                                                    }
                                                                }
                                                                else if (RsDiff >= 9) {
                                                                    if (mu_mv_mag_sq < 5) {
                                                                        return 4;
                                                                    }
                                                                    else if (mu_mv_mag_sq >= 5) {
                                                                        return 8;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (refDCval >= 78) {
                                                        if (RsDiff < 8) {
                                                            return 1;
                                                        }
                                                        else if (RsDiff >= 8) {
                                                            if (Rs < 2609) {
                                                                return 4;
                                                            }
                                                            else if (Rs >= 2609) {
                                                                if (Cs < 2712) {
                                                                    return 2;
                                                                }
                                                                else if (Cs >= 2712) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                            else if (refDCval >= 81) {
                                                if (posBalance < 52) {
                                                    if (Cs < 1672) {
                                                        if (refDCval < 85) {
                                                            if (AFD < 12) {
                                                                return 4;
                                                            }
                                                            else if (AFD >= 12) {
                                                                return 1;
                                                            }
                                                        }
                                                        else if (refDCval >= 85) {
                                                            if (SC < 1892) {
                                                                return 1;
                                                            }
                                                            else if (SC >= 1892) {
                                                                if (SC < 3477) {
                                                                    return 2;
                                                                }
                                                                else if (SC >= 3477) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (Cs >= 1672) {
                                                        if (refDCval < 87) {
                                                            if (MVDiff < 8) {
                                                                return 8;
                                                            }
                                                            else if (MVDiff >= 8) {
                                                                if (RsCsDiff < 67) {
                                                                    return 8;
                                                                }
                                                                else if (RsCsDiff >= 67) {
                                                                    return 1;
                                                                }
                                                            }
                                                        }
                                                        else if (refDCval >= 87) {
                                                            if (MVDiff < 9) {
                                                                if (Cs < 1920) {
                                                                    return 4;
                                                                }
                                                                else if (Cs >= 1920) {
                                                                    return 1;
                                                                }
                                                            }
                                                            else if (MVDiff >= 9) {
                                                                if (Rs < 2759) {
                                                                    if (negBalance < 46) {
                                                                        return 8;
                                                                    }
                                                                    else if (negBalance >= 46) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (Rs >= 2759) {
                                                                    return 2;
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (posBalance >= 52) {
                                                    if (mcTcor < 95) {
                                                        if (Rs < 2811) {
                                                            if (negBalance < 51) {
                                                                if (tcor < 89) {
                                                                    if (negBalance < 46) {
                                                                        if (Cs < 937) {
                                                                            return 2;
                                                                        }
                                                                        else if (Cs >= 937) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                    else if (negBalance >= 46) {
                                                                        if (mu_mv_mag_sq < 35) {
                                                                            if (refDCval < 87) {
                                                                                if (ssDCval < 86) {
                                                                                    if (Rs < 2491) {
                                                                                        return 4;
                                                                                    }
                                                                                    else if (Rs >= 2491) {
                                                                                        return 8;
                                                                                    }
                                                                                }
                                                                                else if (ssDCval >= 86) {
                                                                                    if (SC < 3685) {
                                                                                        return 1;
                                                                                    }
                                                                                    else if (SC >= 3685) {
                                                                                        return 8;
                                                                                    }
                                                                                }
                                                                            }
                                                                            else if (refDCval >= 87) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                        else if (mu_mv_mag_sq >= 35) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                }
                                                                else if (tcor >= 89) {
                                                                    if (ssDCval < 91) {
                                                                        if (Cs < 2009) {
                                                                            if (Rs < 2241) {
                                                                                if (Cs < 1712) {
                                                                                    return 4;
                                                                                }
                                                                                else if (Cs >= 1712) {
                                                                                    return 8;
                                                                                }
                                                                            }
                                                                            else if (Rs >= 2241) {
                                                                                return 2;
                                                                            }
                                                                        }
                                                                        else if (Cs >= 2009) {
                                                                            if (posBalance < 54) {
                                                                                return 8;
                                                                            }
                                                                            else if (posBalance >= 54) {
                                                                                if (ssDCval < 84) {
                                                                                    return 1;
                                                                                }
                                                                                else if (ssDCval >= 84) {
                                                                                    if (ssDCval < 88) {
                                                                                        return 4;
                                                                                    }
                                                                                    else if (ssDCval >= 88) {
                                                                                        return 1;
                                                                                    }
                                                                                }
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (ssDCval >= 91) {
                                                                        if (posBalance < 60) {
                                                                            if (Rs < 2752) {
                                                                                return 1;
                                                                            }
                                                                            else if (Rs >= 2752) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                        else if (posBalance >= 60) {
                                                                            if (SC < 3867) {
                                                                                return 4;
                                                                            }
                                                                            else if (SC >= 3867) {
                                                                                return 2;
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                            else if (negBalance >= 51) {
                                                                if (ssDCval < 88) {
                                                                    if (RsDiff < 7) {
                                                                        if (Rs < 1538) {
                                                                            return 1;
                                                                        }
                                                                        else if (Rs >= 1538) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                    else if (RsDiff >= 7) {
                                                                        if (posBalance < 59) {
                                                                            if (Cs < 1930) {
                                                                                return 2;
                                                                            }
                                                                            else if (Cs >= 1930) {
                                                                                if (AFD < 14) {
                                                                                    return 1;
                                                                                }
                                                                                else if (AFD >= 14) {
                                                                                    return 2;
                                                                                }
                                                                            }
                                                                        }
                                                                        else if (posBalance >= 59) {
                                                                            if (Rs < 2786) {
                                                                                return 1;
                                                                            }
                                                                            else if (Rs >= 2786) {
                                                                                return 2;
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                                else if (ssDCval >= 88) {
                                                                    if (posBalance < 59) {
                                                                        if (tcor < 90) {
                                                                            return 2;
                                                                        }
                                                                        else if (tcor >= 90) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                    else if (posBalance >= 59) {
                                                                        if (ssDCval < 93) {
                                                                            if (RsDiff < 9) {
                                                                                return 4;
                                                                            }
                                                                            else if (RsDiff >= 9) {
                                                                                return 2;
                                                                            }
                                                                        }
                                                                        else if (ssDCval >= 93) {
                                                                            if (Cs < 1871) {
                                                                                return 4;
                                                                            }
                                                                            else if (Cs >= 1871) {
                                                                                return 2;
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (Rs >= 2811) {
                                                            if (mu_mv_mag_sq < 70) {
                                                                if (MVDiff < 10) {
                                                                    if (mu_mv_mag_sq < 61) {
                                                                        return 1;
                                                                    }
                                                                    else if (mu_mv_mag_sq >= 61) {
                                                                        if (Rs < 2949) {
                                                                            return 4;
                                                                        }
                                                                        else if (Rs >= 2949) {
                                                                            if (diffMVdiffVal < -1) {
                                                                                if (SC < 5580) {
                                                                                    return 4;
                                                                                }
                                                                                else if (SC >= 5580) {
                                                                                    return 1;
                                                                                }
                                                                            }
                                                                            else if (diffMVdiffVal >= -1) {
                                                                                return 8;
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                                else if (MVDiff >= 10) {
                                                                    if (posBalance < 58) {
                                                                        if (AFD < 13) {
                                                                            return 1;
                                                                        }
                                                                        else if (AFD >= 13) {
                                                                            if (diffMVdiffVal < 1) {
                                                                                return 1;
                                                                            }
                                                                            else if (diffMVdiffVal >= 1) {
                                                                                if (Cs < 2293) {
                                                                                    return 1;
                                                                                }
                                                                                else if (Cs >= 2293) {
                                                                                    return 4;
                                                                                }
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (posBalance >= 58) {
                                                                        return 1;
                                                                    }
                                                                }
                                                            }
                                                            else if (mu_mv_mag_sq >= 70) {
                                                                if (mu_mv_mag_sq < 84) {
                                                                    if (AFD < 12) {
                                                                        return 8;
                                                                    }
                                                                    else if (AFD >= 12) {
                                                                        if (MVDiff < 6) {
                                                                            return 8;
                                                                        }
                                                                        else if (MVDiff >= 6) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                }
                                                                else if (mu_mv_mag_sq >= 84) {
                                                                    if (RsCsDiff < 86) {
                                                                        return 8;
                                                                    }
                                                                    else if (RsCsDiff >= 86) {
                                                                        return 4;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (mcTcor >= 95) {
                                                        if (Scindex < 5) {
                                                            if (negBalance < 46) {
                                                                return 4;
                                                            }
                                                            else if (negBalance >= 46) {
                                                                if (AFD < 11) {
                                                                    return 1;
                                                                }
                                                                else if (AFD >= 11) {
                                                                    if (diffTSC < -1) {
                                                                        return 1;
                                                                    }
                                                                    else if (diffTSC >= -1) {
                                                                        return 2;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (Scindex >= 5) {
                                                            if (mu_mv_mag_sq < 54) {
                                                                if (MVDiff < 14) {
                                                                    return 1;
                                                                }
                                                                else if (MVDiff >= 14) {
                                                                    if (RsDiff < 8) {
                                                                        return 2;
                                                                    }
                                                                    else if (RsDiff >= 8) {
                                                                        return 1;
                                                                    }
                                                                }
                                                            }
                                                            else if (mu_mv_mag_sq >= 54) {
                                                                return 1;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                            else if (refDCval >= 94) {
                                if (MVDiff < 18) {
                                    if (negBalance < 51) {
                                        if (refDCval < 121) {
                                            if (AFD < 11) {
                                                if (SC < 880) {
                                                    if (AFD < 3) {
                                                        if (ssDCval < 115) {
                                                            return 1;
                                                        }
                                                        else if (ssDCval >= 115) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (AFD >= 3) {
                                                        if (Rs < 286) {
                                                            if (Rs < 85) {
                                                                return 8;
                                                            }
                                                            else if (Rs >= 85) {
                                                                return 2;
                                                            }
                                                        }
                                                        else if (Rs >= 286) {
                                                            if (MVDiff < 16) {
                                                                return 4;
                                                            }
                                                            else if (MVDiff >= 16) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (SC >= 880) {
                                                    if (Cs < 1142) {
                                                        if (refDCval < 106) {
                                                            if (TSCindex < 6) {
                                                                return 2;
                                                            }
                                                            else if (TSCindex >= 6) {
                                                                if (TSC < 7) {
                                                                    if (RsDiff < 3) {
                                                                        return 4;
                                                                    }
                                                                    else if (RsDiff >= 3) {
                                                                        return 8;
                                                                    }
                                                                }
                                                                else if (TSC >= 7) {
                                                                    return 1;
                                                                }
                                                            }
                                                        }
                                                        else if (refDCval >= 106) {
                                                            if (ssDCval < 118) {
                                                                if (SC < 3873) {
                                                                    return 8;
                                                                }
                                                                else if (SC >= 3873) {
                                                                    return 2;
                                                                }
                                                            }
                                                            else if (ssDCval >= 118) {
                                                                if (Cs < 623) {
                                                                    if (Rs < 653) {
                                                                        return 4;
                                                                    }
                                                                    else if (Rs >= 653) {
                                                                        return 8;
                                                                    }
                                                                }
                                                                else if (Cs >= 623) {
                                                                    return 2;
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (Cs >= 1142) {
                                                        if (Cs < 2031) {
                                                            if (mu_mv_mag_sq < 48) {
                                                                if (ssDCval < 108) {
                                                                    if (posBalance < 46) {
                                                                        if (negBalance < 37) {
                                                                            if (ssDCval < 107) {
                                                                                return 8;
                                                                            }
                                                                            else if (ssDCval >= 107) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                        else if (negBalance >= 37) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                    else if (posBalance >= 46) {
                                                                        if (negBalance < 34) {
                                                                            return 4;
                                                                        }
                                                                        else if (negBalance >= 34) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                }
                                                                else if (ssDCval >= 108) {
                                                                    if (diffMVdiffVal < 4) {
                                                                        return 4;
                                                                    }
                                                                    else if (diffMVdiffVal >= 4) {
                                                                        if (Rs < 1965) {
                                                                            return 8;
                                                                        }
                                                                        else if (Rs >= 1965) {
                                                                            if (Cs < 1436) {
                                                                                return 4;
                                                                            }
                                                                            else if (Cs >= 1436) {
                                                                                if (RsCsDiff < 37) {
                                                                                    return 4;
                                                                                }
                                                                                else if (RsCsDiff >= 37) {
                                                                                    return 8;
                                                                                }
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                            else if (mu_mv_mag_sq >= 48) {
                                                                if (gchDC < 1) {
                                                                    if (mcTcor < 97) {
                                                                        if (RsCsDiff < 77) {
                                                                            if (diffMVdiffVal < -79) {
                                                                                return 4;
                                                                            }
                                                                            else if (diffMVdiffVal >= -79) {
                                                                                return 8;
                                                                            }
                                                                        }
                                                                        else if (RsCsDiff >= 77) {
                                                                            if (Cs < 1904) {
                                                                                if (RsDiff < 9) {
                                                                                    return 4;
                                                                                }
                                                                                else if (RsDiff >= 9) {
                                                                                    return 8;
                                                                                }
                                                                            }
                                                                            else if (Cs >= 1904) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (mcTcor >= 97) {
                                                                        if (diffAFD < 1) {
                                                                            if (negBalance < 35) {
                                                                                if (AFD < 5) {
                                                                                    return 4;
                                                                                }
                                                                                else if (AFD >= 5) {
                                                                                    return 8;
                                                                                }
                                                                            }
                                                                            else if (negBalance >= 35) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                        else if (diffAFD >= 1) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                }
                                                                else if (gchDC >= 1) {
                                                                    if (ssDCval < 105) {
                                                                        return 4;
                                                                    }
                                                                    else if (ssDCval >= 105) {
                                                                        if (MVDiff < 12) {
                                                                            if (TSCindex < 9) {
                                                                                if (mu_mv_mag_sq < 62) {
                                                                                    if (RsCsDiff < 29) {
                                                                                        return 8;
                                                                                    }
                                                                                    else if (RsCsDiff >= 29) {
                                                                                        return 4;
                                                                                    }
                                                                                }
                                                                                else if (mu_mv_mag_sq >= 62) {
                                                                                    return 8;
                                                                                }
                                                                            }
                                                                            else if (TSCindex >= 9) {
                                                                                return 8;
                                                                            }
                                                                        }
                                                                        else if (MVDiff >= 12) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (Cs >= 2031) {
                                                            if (refDCval < 107) {
                                                                if (SC < 7254) {
                                                                    return 8;
                                                                }
                                                                else if (SC >= 7254) {
                                                                    return 2;
                                                                }
                                                            }
                                                            else if (refDCval >= 107) {
                                                                if (MVDiff < 9) {
                                                                    if (diffMVdiffVal < -1) {
                                                                        if (tcor < 83) {
                                                                            return 4;
                                                                        }
                                                                        else if (tcor >= 83) {
                                                                            if (TSC < 9) {
                                                                                if (Cs < 2263) {
                                                                                    return 4;
                                                                                }
                                                                                else if (Cs >= 2263) {
                                                                                    return 8;
                                                                                }
                                                                            }
                                                                            else if (TSC >= 9) {
                                                                                if (negBalance < 37) {
                                                                                    return 2;
                                                                                }
                                                                                else if (negBalance >= 37) {
                                                                                    if (SC < 7581) {
                                                                                        return 4;
                                                                                    }
                                                                                    else if (SC >= 7581) {
                                                                                        return 8;
                                                                                    }
                                                                                }
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (diffMVdiffVal >= -1) {
                                                                        return 8;
                                                                    }
                                                                }
                                                                else if (MVDiff >= 9) {
                                                                    if (mcTcor < 95) {
                                                                        if (Cs < 2825) {
                                                                            if (posBalance < 42) {
                                                                                return 8;
                                                                            }
                                                                            else if (posBalance >= 42) {
                                                                                if (Rs < 4835) {
                                                                                    return 4;
                                                                                }
                                                                                else if (Rs >= 4835) {
                                                                                    return 8;
                                                                                }
                                                                            }
                                                                        }
                                                                        else if (Cs >= 2825) {
                                                                            return 2;
                                                                        }
                                                                    }
                                                                    else if (mcTcor >= 95) {
                                                                        if (ssDCval < 119) {
                                                                            if (Rs < 4274) {
                                                                                return 2;
                                                                            }
                                                                            else if (Rs >= 4274) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                        else if (ssDCval >= 119) {
                                                                            if (Cs < 2177) {
                                                                                return 4;
                                                                            }
                                                                            else if (Cs >= 2177) {
                                                                                return 8;
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                            else if (AFD >= 11) {
                                                if (negBalance < 48) {
                                                    if (MVDiff < 4) {
                                                        if (Rs < 3389) {
                                                            return 8;
                                                        }
                                                        else if (Rs >= 3389) {
                                                            return 1;
                                                        }
                                                    }
                                                    else if (MVDiff >= 4) {
                                                        if (posBalance < 52) {
                                                            if (gchDC < 1) {
                                                                if (Rs < 4460) {
                                                                    if (MVDiff < 13) {
                                                                        return 1;
                                                                    }
                                                                    else if (MVDiff >= 13) {
                                                                        if (Rs < 4104) {
                                                                            return 8;
                                                                        }
                                                                        else if (Rs >= 4104) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                }
                                                                else if (Rs >= 4460) {
                                                                    if (SC < 8219) {
                                                                        return 2;
                                                                    }
                                                                    else if (SC >= 8219) {
                                                                        return 8;
                                                                    }
                                                                }
                                                            }
                                                            else if (gchDC >= 1) {
                                                                if (ssDCval < 109) {
                                                                    if (diffRsCsdiff < -16) {
                                                                        if (Rs < 2870) {
                                                                            return 2;
                                                                        }
                                                                        else if (Rs >= 2870) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                    else if (diffRsCsdiff >= -16) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (ssDCval >= 109) {
                                                                    if (MVDiff < 9) {
                                                                        if (RsCsDiff < 115) {
                                                                            return 2;
                                                                        }
                                                                        else if (RsCsDiff >= 115) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                    else if (MVDiff >= 9) {
                                                                        return 2;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (posBalance >= 52) {
                                                            if (SC < 4620) {
                                                                if (refDCval < 103) {
                                                                    if (Rs < 2257) {
                                                                        return 4;
                                                                    }
                                                                    else if (Rs >= 2257) {
                                                                        return 1;
                                                                    }
                                                                }
                                                                else if (refDCval >= 103) {
                                                                    if (CsDiff < 3) {
                                                                        return 8;
                                                                    }
                                                                    else if (CsDiff >= 3) {
                                                                        return 2;
                                                                    }
                                                                }
                                                            }
                                                            else if (SC >= 4620) {
                                                                if (mcTcor < 94) {
                                                                    if (diffAFD < 4) {
                                                                        return 1;
                                                                    }
                                                                    else if (diffAFD >= 4) {
                                                                        return 2;
                                                                    }
                                                                }
                                                                else if (mcTcor >= 94) {
                                                                    return 2;
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (negBalance >= 48) {
                                                    if (Rs < 2609) {
                                                        if (Cs < 670) {
                                                            return 2;
                                                        }
                                                        else if (Cs >= 670) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (Rs >= 2609) {
                                                        if (Cs < 1758) {
                                                            return 8;
                                                        }
                                                        else if (Cs >= 1758) {
                                                            if (mu_mv_mag_sq < 59) {
                                                                return 1;
                                                            }
                                                            else if (mu_mv_mag_sq >= 59) {
                                                                if (Rs < 2788) {
                                                                    return 1;
                                                                }
                                                                else if (Rs >= 2788) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (refDCval >= 121) {
                                            if (Cs < 2521) {
                                                if (Rs < 2685) {
                                                    if (mcTcor < 96) {
                                                        if (tcor < 84) {
                                                            if (ssDCval < 140) {
                                                                if (mcTcor < 88) {
                                                                    return 1;
                                                                }
                                                                else if (mcTcor >= 88) {
                                                                    if (Cs < 1441) {
                                                                        return 1;
                                                                    }
                                                                    else if (Cs >= 1441) {
                                                                        if (RsDiff < 5) {
                                                                            return 1;
                                                                        }
                                                                        else if (RsDiff >= 5) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                            else if (ssDCval >= 140) {
                                                                if (posBalance < 38) {
                                                                    if (Rs < 1307) {
                                                                        if (MVDiff < 13) {
                                                                            return 4;
                                                                        }
                                                                        else if (MVDiff >= 13) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                    else if (Rs >= 1307) {
                                                                        return 1;
                                                                    }
                                                                }
                                                                else if (posBalance >= 38) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                        else if (tcor >= 84) {
                                                            if (mu_mv_mag_sq < 29) {
                                                                if (Rs < 1132) {
                                                                    return 1;
                                                                }
                                                                else if (Rs >= 1132) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (mu_mv_mag_sq >= 29) {
                                                                if (posBalance < 37) {
                                                                    if (RsDiff < 1) {
                                                                        return 4;
                                                                    }
                                                                    else if (RsDiff >= 1) {
                                                                        return 1;
                                                                    }
                                                                }
                                                                else if (posBalance >= 37) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (mcTcor >= 96) {
                                                        if (Rs < 183) {
                                                            return 8;
                                                        }
                                                        else if (Rs >= 183) {
                                                            if (SC < 2167) {
                                                                if (tcor < 98) {
                                                                    if (diffMVdiffVal < 3) {
                                                                        return 4;
                                                                    }
                                                                    else if (diffMVdiffVal >= 3) {
                                                                        if (Cs < 497) {
                                                                            return 2;
                                                                        }
                                                                        else if (Cs >= 497) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                }
                                                                else if (tcor >= 98) {
                                                                    return 2;
                                                                }
                                                            }
                                                            else if (SC >= 2167) {
                                                                if (Cs < 501) {
                                                                    return 8;
                                                                }
                                                                else if (Cs >= 501) {
                                                                    return 2;
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (Rs >= 2685) {
                                                    if (Cs < 2348) {
                                                        return 4;
                                                    }
                                                    else if (Cs >= 2348) {
                                                        if (mu_mv_mag_sq < 75) {
                                                            if (Rs < 3828) {
                                                                return 1;
                                                            }
                                                            else if (Rs >= 3828) {
                                                                return 2;
                                                            }
                                                        }
                                                        else if (mu_mv_mag_sq >= 75) {
                                                            if (TSC < 7) {
                                                                return 4;
                                                            }
                                                            else if (TSC >= 7) {
                                                                if (AFD < 9) {
                                                                    return 8;
                                                                }
                                                                else if (AFD >= 9) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                            else if (Cs >= 2521) {
                                                if (Rs < 3945) {
                                                    return 8;
                                                }
                                                else if (Rs >= 3945) {
                                                    if (refDCval < 128) {
                                                        if (posBalance < 40) {
                                                            return 2;
                                                        }
                                                        else if (posBalance >= 40) {
                                                            if (Cs < 2617) {
                                                                if (Rs < 4657) {
                                                                    return 8;
                                                                }
                                                                else if (Rs >= 4657) {
                                                                    return 2;
                                                                }
                                                            }
                                                            else if (Cs >= 2617) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                    else if (refDCval >= 128) {
                                                        return 2;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (negBalance >= 51) {
                                        if (mu_mv_mag_sq < 5) {
                                            if (ssDCval < 98) {
                                                if (SC < 3224) {
                                                    if (tcor < 75) {
                                                        if (Rs < 1317) {
                                                            return 4;
                                                        }
                                                        else if (Rs >= 1317) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (tcor >= 75) {
                                                        if (SC < 1180) {
                                                            return 8;
                                                        }
                                                        else if (SC >= 1180) {
                                                            if (Rs < 985) {
                                                                if (negBalance < 53) {
                                                                    return 8;
                                                                }
                                                                else if (negBalance >= 53) {
                                                                    if (MVDiff < 12) {
                                                                        return 4;
                                                                    }
                                                                    else if (MVDiff >= 12) {
                                                                        if (Rs < 655) {
                                                                            return 4;
                                                                        }
                                                                        else if (Rs >= 655) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                            else if (Rs >= 985) {
                                                                if (SC < 2526) {
                                                                    return 2;
                                                                }
                                                                else if (SC >= 2526) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (SC >= 3224) {
                                                    if (tcor < 87) {
                                                        if (ssDCval < 97) {
                                                            if (gchDC < 1) {
                                                                if (Rs < 3005) {
                                                                    return 2;
                                                                }
                                                                else if (Rs >= 3005) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (gchDC >= 1) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (ssDCval >= 97) {
                                                            return 2;
                                                        }
                                                    }
                                                    else if (tcor >= 87) {
                                                        if (negBalance < 52) {
                                                            return 4;
                                                        }
                                                        else if (negBalance >= 52) {
                                                            return 2;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (ssDCval >= 98) {
                                                if (tcor < 91) {
                                                    if (RsCsDiff < 133) {
                                                        if (refDCval < 99) {
                                                            if (Rs < 1269) {
                                                                if (Cs < 493) {
                                                                    return 8;
                                                                }
                                                                else if (Cs >= 493) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (Rs >= 1269) {
                                                                return 2;
                                                            }
                                                        }
                                                        else if (refDCval >= 99) {
                                                            if (Rs < 3047) {
                                                                if (SC < 4865) {
                                                                    return 4;
                                                                }
                                                                else if (SC >= 4865) {
                                                                    if (SC < 4958) {
                                                                        return 8;
                                                                    }
                                                                    else if (SC >= 4958) {
                                                                        return 4;
                                                                    }
                                                                }
                                                            }
                                                            else if (Rs >= 3047) {
                                                                return 2;
                                                            }
                                                        }
                                                    }
                                                    else if (RsCsDiff >= 133) {
                                                        if (ssDCval < 127) {
                                                            return 4;
                                                        }
                                                        else if (ssDCval >= 127) {
                                                            if (TSC < 10) {
                                                                return 1;
                                                            }
                                                            else if (TSC >= 10) {
                                                                return 2;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (tcor >= 91) {
                                                    if (Rs < 620) {
                                                        return 2;
                                                    }
                                                    else if (Rs >= 620) {
                                                        return 1;
                                                    }
                                                }
                                            }
                                        }
                                        else if (mu_mv_mag_sq >= 5) {
                                            if (RsDiff < 3) {
                                                if (tcor < 99) {
                                                    if (refDCval < 110) {
                                                        if (AFD < 12) {
                                                            return 8;
                                                        }
                                                        else if (AFD >= 12) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (refDCval >= 110) {
                                                        return 4;
                                                    }
                                                }
                                                else if (tcor >= 99) {
                                                    return 8;
                                                }
                                            }
                                            else if (RsDiff >= 3) {
                                                if (ssDCval < 98) {
                                                    if (SC < 3781) {
                                                        if (mcTcor < 90) {
                                                            if (CsDiff < 5) {
                                                                return 8;
                                                            }
                                                            else if (CsDiff >= 5) {
                                                                if (Cs < 1665) {
                                                                    return 2;
                                                                }
                                                                else if (Cs >= 1665) {
                                                                    if (Rs < 2008) {
                                                                        return 8;
                                                                    }
                                                                    else if (Rs >= 2008) {
                                                                        return 2;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (mcTcor >= 90) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (SC >= 3781) {
                                                        if (AFD < 11) {
                                                            return 2;
                                                        }
                                                        else if (AFD >= 11) {
                                                            if (Cs < 2064) {
                                                                return 1;
                                                            }
                                                            else if (Cs >= 2064) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (ssDCval >= 98) {
                                                    if (mu_mv_mag_sq < 22) {
                                                        if (ssDCval < 133) {
                                                            if (gchDC < 1) {
                                                                return 1;
                                                            }
                                                            else if (gchDC >= 1) {
                                                                if (SC < 3156) {
                                                                    return 8;
                                                                }
                                                                else if (SC >= 3156) {
                                                                    return 2;
                                                                }
                                                            }
                                                        }
                                                        else if (ssDCval >= 133) {
                                                            return 2;
                                                        }
                                                    }
                                                    else if (mu_mv_mag_sq >= 22) {
                                                        if (AFD < 16) {
                                                            if (refDCval < 110) {
                                                                return 4;
                                                            }
                                                            else if (refDCval >= 110) {
                                                                if (mu_mv_mag_sq < 79) {
                                                                    if (posBalance < 48) {
                                                                        return 8;
                                                                    }
                                                                    else if (posBalance >= 48) {
                                                                        if (mu_mv_mag_sq < 55) {
                                                                            return 4;
                                                                        }
                                                                        else if (mu_mv_mag_sq >= 55) {
                                                                            if (refDCval < 117) {
                                                                                return 4;
                                                                            }
                                                                            else if (refDCval >= 117) {
                                                                                if (RsDiff < 8) {
                                                                                    if (MVDiff < 6) {
                                                                                        if (gchDC < 1) {
                                                                                            if (diffMVdiffVal < -1) {
                                                                                                return 4;
                                                                                            }
                                                                                            else if (diffMVdiffVal >= -1) {
                                                                                                if (SC < 4715) {
                                                                                                    return 8;
                                                                                                }
                                                                                                else if (SC >= 4715) {
                                                                                                    if (SC < 5838) {
                                                                                                        return 4;
                                                                                                    }
                                                                                                    else if (SC >= 5838) {
                                                                                                        return 8;
                                                                                                    }
                                                                                                }
                                                                                            }
                                                                                        }
                                                                                        else if (gchDC >= 1) {
                                                                                            return 8;
                                                                                        }
                                                                                    }
                                                                                    else if (MVDiff >= 6) {
                                                                                        return 4;
                                                                                    }
                                                                                }
                                                                                else if (RsDiff >= 8) {
                                                                                    if (Cs < 1915) {
                                                                                        return 8;
                                                                                    }
                                                                                    else if (Cs >= 1915) {
                                                                                        return 4;
                                                                                    }
                                                                                }
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                                else if (mu_mv_mag_sq >= 79) {
                                                                    if (MVDiff < 7) {
                                                                        return 8;
                                                                    }
                                                                    else if (MVDiff >= 7) {
                                                                        return 4;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (AFD >= 16) {
                                                            if (Cs < 1376) {
                                                                return 2;
                                                            }
                                                            else if (Cs >= 1376) {
                                                                return 1;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (MVDiff >= 18) {
                                    if (CsDiff < 2) {
                                        if (Cs < 318) {
                                            if (AFD < 10) {
                                                if (TSC < 7) {
                                                    if (Cs < 267) {
                                                        return 1;
                                                    }
                                                    else if (Cs >= 267) {
                                                        return 2;
                                                    }
                                                }
                                                else if (TSC >= 7) {
                                                    if (Rs < 510) {
                                                        return 4;
                                                    }
                                                    else if (Rs >= 510) {
                                                        return 1;
                                                    }
                                                }
                                            }
                                            else if (AFD >= 10) {
                                                if (refDCval < 98) {
                                                    if (Rs < 801) {
                                                        return 2;
                                                    }
                                                    else if (Rs >= 801) {
                                                        return 1;
                                                    }
                                                }
                                                else if (refDCval >= 98) {
                                                    if (diffMVdiffVal < 4) {
                                                        if (mcTcor < 83) {
                                                            if (Rs < 293) {
                                                                return 1;
                                                            }
                                                            else if (Rs >= 293) {
                                                                return 2;
                                                            }
                                                        }
                                                        else if (mcTcor >= 83) {
                                                            return 1;
                                                        }
                                                    }
                                                    else if (diffMVdiffVal >= 4) {
                                                        if (SC < 692) {
                                                            if (refDCval < 159) {
                                                                return 1;
                                                            }
                                                            else if (refDCval >= 159) {
                                                                if (AFD < 13) {
                                                                    return 1;
                                                                }
                                                                else if (AFD >= 13) {
                                                                    return 2;
                                                                }
                                                            }
                                                        }
                                                        else if (SC >= 692) {
                                                            if (Rs < 637) {
                                                                return 2;
                                                            }
                                                            else if (Rs >= 637) {
                                                                return 1;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (Cs >= 318) {
                                            if (Rs < 413) {
                                                if (Cs < 442) {
                                                    return 4;
                                                }
                                                else if (Cs >= 442) {
                                                    return 2;
                                                }
                                            }
                                            else if (Rs >= 413) {
                                                if (RsCsDiff < 7) {
                                                    return 8;
                                                }
                                                else if (RsCsDiff >= 7) {
                                                    return 4;
                                                }
                                            }
                                        }
                                    }
                                    else if (CsDiff >= 2) {
                                        if (CsDiff < 3) {
                                            if (Cs < 425) {
                                                if (SC < 2013) {
                                                    if (AFD < 10) {
                                                        return 8;
                                                    }
                                                    else if (AFD >= 10) {
                                                        return 2;
                                                    }
                                                }
                                                else if (SC >= 2013) {
                                                    return 4;
                                                }
                                            }
                                            else if (Cs >= 425) {
                                                if (refDCval < 111) {
                                                    if (Cs < 593) {
                                                        return 8;
                                                    }
                                                    else if (Cs >= 593) {
                                                        if (Rs < 1637) {
                                                            if (refDCval < 99) {
                                                                if (AFD < 10) {
                                                                    return 2;
                                                                }
                                                                else if (AFD >= 10) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (refDCval >= 99) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (Rs >= 1637) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                                else if (refDCval >= 111) {
                                                    return 4;
                                                }
                                            }
                                        }
                                        else if (CsDiff >= 3) {
                                            if (Cs < 1314) {
                                                if (negBalance < 50) {
                                                    if (diffTSC < -16) {
                                                        if (SC < 2021) {
                                                            return 4;
                                                        }
                                                        else if (SC >= 2021) {
                                                            return 2;
                                                        }
                                                    }
                                                    else if (diffTSC >= -16) {
                                                        if (diffMVdiffVal < -3) {
                                                            if (Rs < 909) {
                                                                return 2;
                                                            }
                                                            else if (Rs >= 909) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (diffMVdiffVal >= -3) {
                                                            if (Cs < 617) {
                                                                return 8;
                                                            }
                                                            else if (Cs >= 617) {
                                                                return 1;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (negBalance >= 50) {
                                                    if (AFD < 20) {
                                                        if (posBalance < 32) {
                                                            return 8;
                                                        }
                                                        else if (posBalance >= 32) {
                                                            if (MVDiff < 21) {
                                                                if (diffMVdiffVal < -2) {
                                                                    return 2;
                                                                }
                                                                else if (diffMVdiffVal >= -2) {
                                                                    if (Cs < 1140) {
                                                                        return 8;
                                                                    }
                                                                    else if (Cs >= 1140) {
                                                                        return 2;
                                                                    }
                                                                }
                                                            }
                                                            else if (MVDiff >= 21) {
                                                                return 2;
                                                            }
                                                        }
                                                    }
                                                    else if (AFD >= 20) {
                                                        if (diffAFD < -2) {
                                                            return 1;
                                                        }
                                                        else if (diffAFD >= -2) {
                                                            if (ssDCval < 106) {
                                                                if (SC < 2515) {
                                                                    return 2;
                                                                }
                                                                else if (SC >= 2515) {
                                                                    return 1;
                                                                }
                                                            }
                                                            else if (ssDCval >= 106) {
                                                                if (refDCval < 151) {
                                                                    return 2;
                                                                }
                                                                else if (refDCval >= 151) {
                                                                    return 1;
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                            else if (Cs >= 1314) {
                                                if (Cs < 2727) {
                                                    if (TSC < 13) {
                                                        if (mu_mv_mag_sq < 55) {
                                                            if (SC < 3307) {
                                                                if (posBalance < 55) {
                                                                    if (negBalance < 41) {
                                                                        return 8;
                                                                    }
                                                                    else if (negBalance >= 41) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (posBalance >= 55) {
                                                                    if (Rs < 1411) {
                                                                        return 8;
                                                                    }
                                                                    else if (Rs >= 1411) {
                                                                        return 2;
                                                                    }
                                                                }
                                                            }
                                                            else if (SC >= 3307) {
                                                                if (SC < 5896) {
                                                                    if (RsCsDiff < 56) {
                                                                        if (Rs < 2435) {
                                                                            return 8;
                                                                        }
                                                                        else if (Rs >= 2435) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                    else if (RsCsDiff >= 56) {
                                                                        if (Cs < 1533) {
                                                                            if (tcor < 87) {
                                                                                return 4;
                                                                            }
                                                                            else if (tcor >= 87) {
                                                                                return 8;
                                                                            }
                                                                        }
                                                                        else if (Cs >= 1533) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                }
                                                                else if (SC >= 5896) {
                                                                    if (SC < 6554) {
                                                                        return 8;
                                                                    }
                                                                    else if (SC >= 6554) {
                                                                        return 4;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (mu_mv_mag_sq >= 55) {
                                                            if (diffRsCsdiff < -252) {
                                                                return 8;
                                                            }
                                                            else if (diffRsCsdiff >= -252) {
                                                                if (ssDCval < 105) {
                                                                    if (Cs < 2026) {
                                                                        if (Rs < 2143) {
                                                                            return 4;
                                                                        }
                                                                        else if (Rs >= 2143) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                    else if (Cs >= 2026) {
                                                                        return 2;
                                                                    }
                                                                }
                                                                else if (ssDCval >= 105) {
                                                                    if (Cs < 1759) {
                                                                        if (Rs < 1478) {
                                                                            return 2;
                                                                        }
                                                                        else if (Rs >= 1478) {
                                                                            if (Rs < 3602) {
                                                                                return 8;
                                                                            }
                                                                            else if (Rs >= 3602) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (Cs >= 1759) {
                                                                        if (mu_mv_mag_sq < 72) {
                                                                            if (mu_mv_mag_sq < 58) {
                                                                                return 2;
                                                                            }
                                                                            else if (mu_mv_mag_sq >= 58) {
                                                                                if (SC < 5103) {
                                                                                    if (Cs < 2045) {
                                                                                        return 4;
                                                                                    }
                                                                                    else if (Cs >= 2045) {
                                                                                        return 2;
                                                                                    }
                                                                                }
                                                                                else if (SC >= 5103) {
                                                                                    return 4;
                                                                                }
                                                                            }
                                                                        }
                                                                        else if (mu_mv_mag_sq >= 72) {
                                                                            if (ssDCval < 120) {
                                                                                if (Rs < 4073) {
                                                                                    return 2;
                                                                                }
                                                                                else if (Rs >= 4073) {
                                                                                    return 4;
                                                                                }
                                                                            }
                                                                            else if (ssDCval >= 120) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (TSC >= 13) {
                                                        if (SC < 3559) {
                                                            return 2;
                                                        }
                                                        else if (SC >= 3559) {
                                                            return 1;
                                                        }
                                                    }
                                                }
                                                else if (Cs >= 2727) {
                                                    if (Rs < 4120) {
                                                        return 8;
                                                    }
                                                    else if (Rs >= 4120) {
                                                        return 1;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                        }
                        else if (RsCsDiff >= 149) {
                            if (refDCval < 136) {
                                if (Rs < 3766) {
                                    if (ssDCval < 129) {
                                        if (Rs < 2134) {
                                            if (Cs < 1396) {
                                                if (posBalance < 66) {
                                                    return 2;
                                                }
                                                else if (posBalance >= 66) {
                                                    if (tcor < 56) {
                                                        return 2;
                                                    }
                                                    else if (tcor >= 56) {
                                                        if (Cs < 1249) {
                                                            return 8;
                                                        }
                                                        else if (Cs >= 1249) {
                                                            return 1;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (Cs >= 1396) {
                                                if (ssDCval < 88) {
                                                    if (Rs < 1542) {
                                                        return 1;
                                                    }
                                                    else if (Rs >= 1542) {
                                                        return 4;
                                                    }
                                                }
                                                else if (ssDCval >= 88) {
                                                    if (Rs < 1993) {
                                                        if (negBalance < 56) {
                                                            if (AFD < 17) {
                                                                return 8;
                                                            }
                                                            else if (AFD >= 17) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (negBalance >= 56) {
                                                            return 2;
                                                        }
                                                    }
                                                    else if (Rs >= 1993) {
                                                        if (ssDCval < 120) {
                                                            return 2;
                                                        }
                                                        else if (ssDCval >= 120) {
                                                            return 1;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (Rs >= 2134) {
                                            if (ssDCval < 84) {
                                                if (mu_mv_mag_sq < 25) {
                                                    if (AFD < 16) {
                                                        return 4;
                                                    }
                                                    else if (AFD >= 16) {
                                                        if (diffAFD < 3) {
                                                            if (MVDiff < 26) {
                                                                if (refDCval < 74) {
                                                                    return 8;
                                                                }
                                                                else if (refDCval >= 74) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (MVDiff >= 26) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (diffAFD >= 3) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                                else if (mu_mv_mag_sq >= 25) {
                                                    if (AFD < 12) {
                                                        if (Rs < 2640) {
                                                            return 8;
                                                        }
                                                        else if (Rs >= 2640) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (AFD >= 12) {
                                                        return 1;
                                                    }
                                                }
                                            }
                                            else if (ssDCval >= 84) {
                                                if (Cs < 2624) {
                                                    if (tcor < 86) {
                                                        if (Cs < 1605) {
                                                            if (refDCval < 103) {
                                                                if (Scindex < 5) {
                                                                    if (ssDCval < 101) {
                                                                        if (Rs < 2909) {
                                                                            return 2;
                                                                        }
                                                                        else if (Rs >= 2909) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                    else if (ssDCval >= 101) {
                                                                        return 1;
                                                                    }
                                                                }
                                                                else if (Scindex >= 5) {
                                                                    if (SC < 4147) {
                                                                        return 1;
                                                                    }
                                                                    else if (SC >= 4147) {
                                                                        return 2;
                                                                    }
                                                                }
                                                            }
                                                            else if (refDCval >= 103) {
                                                                return 2;
                                                            }
                                                        }
                                                        else if (Cs >= 1605) {
                                                            if (SC < 4544) {
                                                                if (AFD < 22) {
                                                                    if (ssDCval < 116) {
                                                                        if (RsDiff < 11) {
                                                                            if (SC < 4273) {
                                                                                return 2;
                                                                            }
                                                                            else if (SC >= 4273) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                        else if (RsDiff >= 11) {
                                                                            return 2;
                                                                        }
                                                                    }
                                                                    else if (ssDCval >= 116) {
                                                                        if (refDCval < 124) {
                                                                            return 4;
                                                                        }
                                                                        else if (refDCval >= 124) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                }
                                                                else if (AFD >= 22) {
                                                                    return 1;
                                                                }
                                                            }
                                                            else if (SC >= 4544) {
                                                                if (mcTcor < 85) {
                                                                    if (ssDCval < 116) {
                                                                        if (Rs < 2475) {
                                                                            return 4;
                                                                        }
                                                                        else if (Rs >= 2475) {
                                                                            if (Rs < 3058) {
                                                                                return 1;
                                                                            }
                                                                            else if (Rs >= 3058) {
                                                                                return 2;
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (ssDCval >= 116) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (mcTcor >= 85) {
                                                                    if (MVDiff < 34) {
                                                                        if (AFD < 12) {
                                                                            if (ssDCval < 110) {
                                                                                return 8;
                                                                            }
                                                                            else if (ssDCval >= 110) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                        else if (AFD >= 12) {
                                                                            if (TSC < 12) {
                                                                                return 2;
                                                                            }
                                                                            else if (TSC >= 12) {
                                                                                if (CsDiff < 9) {
                                                                                    if (TSC < 13) {
                                                                                        return 2;
                                                                                    }
                                                                                    else if (TSC >= 13) {
                                                                                        if (Rs < 2734) {
                                                                                            return 1;
                                                                                        }
                                                                                        else if (Rs >= 2734) {
                                                                                            return 8;
                                                                                        }
                                                                                    }
                                                                                }
                                                                                else if (CsDiff >= 9) {
                                                                                    if (refDCval < 99) {
                                                                                        if (Cs < 2351) {
                                                                                            return 2;
                                                                                        }
                                                                                        else if (Cs >= 2351) {
                                                                                            return 1;
                                                                                        }
                                                                                    }
                                                                                    else if (refDCval >= 99) {
                                                                                        return 1;
                                                                                    }
                                                                                }
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (MVDiff >= 34) {
                                                                        if (negBalance < 61) {
                                                                            return 8;
                                                                        }
                                                                        else if (negBalance >= 61) {
                                                                            if (Rs < 3116) {
                                                                                if (Rs < 2891) {
                                                                                    return 8;
                                                                                }
                                                                                else if (Rs >= 2891) {
                                                                                    return 2;
                                                                                }
                                                                            }
                                                                            else if (Rs >= 3116) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (tcor >= 86) {
                                                        if (RsDiff < 12) {
                                                            if (AFD < 13) {
                                                                if (TSC < 11) {
                                                                    if (Cs < 2303) {
                                                                        return 4;
                                                                    }
                                                                    else if (Cs >= 2303) {
                                                                        return 2;
                                                                    }
                                                                }
                                                                else if (TSC >= 11) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (AFD >= 13) {
                                                                if (diffAFD < 3) {
                                                                    if (posBalance < 57) {
                                                                        if (MVDiff < 4) {
                                                                            return 8;
                                                                        }
                                                                        else if (MVDiff >= 4) {
                                                                            if (ssDCval < 85) {
                                                                                return 8;
                                                                            }
                                                                            else if (ssDCval >= 85) {
                                                                                return 1;
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (posBalance >= 57) {
                                                                        if (Cs < 1950) {
                                                                            return 8;
                                                                        }
                                                                        else if (Cs >= 1950) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                }
                                                                else if (diffAFD >= 3) {
                                                                    if (CsDiff < 9) {
                                                                        if (SC < 4965) {
                                                                            return 1;
                                                                        }
                                                                        else if (SC >= 4965) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                    else if (CsDiff >= 9) {
                                                                        return 2;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (RsDiff >= 12) {
                                                            if (AFD < 18) {
                                                                if (AFD < 13) {
                                                                    return 4;
                                                                }
                                                                else if (AFD >= 13) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (AFD >= 18) {
                                                                return 2;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (Cs >= 2624) {
                                                    if (refDCval < 95) {
                                                        if (Rs < 2737) {
                                                            return 4;
                                                        }
                                                        else if (Rs >= 2737) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (refDCval >= 95) {
                                                        if (Cs < 2930) {
                                                            return 4;
                                                        }
                                                        else if (Cs >= 2930) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (ssDCval >= 129) {
                                        if (tcor < 86) {
                                            if (SC < 3043) {
                                                if (SC < 2679) {
                                                    return 2;
                                                }
                                                else if (SC >= 2679) {
                                                    return 4;
                                                }
                                            }
                                            else if (SC >= 3043) {
                                                if (posBalance < 62) {
                                                    return 1;
                                                }
                                                else if (posBalance >= 62) {
                                                    if (Rs < 2687) {
                                                        return 1;
                                                    }
                                                    else if (Rs >= 2687) {
                                                        return 2;
                                                    }
                                                }
                                            }
                                        }
                                        else if (tcor >= 86) {
                                            return 8;
                                        }
                                    }
                                }
                                else if (Rs >= 3766) {
                                    if (mu_mv_mag_sq < 68) {
                                        if (refDCval < 106) {
                                            if (Cs < 850) {
                                                return 2;
                                            }
                                            else if (Cs >= 850) {
                                                if (mu_mv_mag_sq < 66) {
                                                    return 4;
                                                }
                                                else if (mu_mv_mag_sq >= 66) {
                                                    if (Cs < 1879) {
                                                        return 4;
                                                    }
                                                    else if (Cs >= 1879) {
                                                        return 2;
                                                    }
                                                }
                                            }
                                        }
                                        else if (refDCval >= 106) {
                                            if (TSC < 10) {
                                                if (ssDCval < 109) {
                                                    return 8;
                                                }
                                                else if (ssDCval >= 109) {
                                                    return 4;
                                                }
                                            }
                                            else if (TSC >= 10) {
                                                if (Cs < 2043) {
                                                    return 2;
                                                }
                                                else if (Cs >= 2043) {
                                                    if (SC < 7727) {
                                                        if (negBalance < 45) {
                                                            if (RsCsDiff < 160) {
                                                                return 4;
                                                            }
                                                            else if (RsCsDiff >= 160) {
                                                                return 1;
                                                            }
                                                        }
                                                        else if (negBalance >= 45) {
                                                            if (Rs < 4237) {
                                                                return 4;
                                                            }
                                                            else if (Rs >= 4237) {
                                                                return 2;
                                                            }
                                                        }
                                                    }
                                                    else if (SC >= 7727) {
                                                        if (Rs < 5718) {
                                                            return 2;
                                                        }
                                                        else if (Rs >= 5718) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (mu_mv_mag_sq >= 68) {
                                        if (MVDiff < 22) {
                                            if (Cs < 2644) {
                                                if (Rs < 4379) {
                                                    return 2;
                                                }
                                                else if (Rs >= 4379) {
                                                    return 4;
                                                }
                                            }
                                            else if (Cs >= 2644) {
                                                return 2;
                                            }
                                        }
                                        else if (MVDiff >= 22) {
                                            if (Rs < 4531) {
                                                return 4;
                                            }
                                            else if (Rs >= 4531) {
                                                return 8;
                                            }
                                        }
                                    }
                                }
                            }
                            else if (refDCval >= 136) {
                                if (posBalance < 59) {
                                    if (AFD < 18) {
                                        if (refDCval < 139) {
                                            return 1;
                                        }
                                        else if (refDCval >= 139) {
                                            if (TSC < 6) {
                                                return 1;
                                            }
                                            else if (TSC >= 6) {
                                                if (Cs < 1760) {
                                                    return 4;
                                                }
                                                else if (Cs >= 1760) {
                                                    return 8;
                                                }
                                            }
                                        }
                                    }
                                    else if (AFD >= 18) {
                                        if (MVDiff < 21) {
                                            if (SC < 5507) {
                                                if (Rs < 3102) {
                                                    return 8;
                                                }
                                                else if (Rs >= 3102) {
                                                    return 1;
                                                }
                                            }
                                            else if (SC >= 5507) {
                                                return 4;
                                            }
                                        }
                                        else if (MVDiff >= 21) {
                                            if (AFD < 35) {
                                                if (CsDiff < 9) {
                                                    return 2;
                                                }
                                                else if (CsDiff >= 9) {
                                                    if (Rs < 2169) {
                                                        return 2;
                                                    }
                                                    else if (Rs >= 2169) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                            else if (AFD >= 35) {
                                                if (Rs < 2975) {
                                                    return 4;
                                                }
                                                else if (Rs >= 2975) {
                                                    return 2;
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (posBalance >= 59) {
                                    if (mcTcor < 85) {
                                        if (Rs < 3278) {
                                            if (Rs < 1805) {
                                                return 8;
                                            }
                                            else if (Rs >= 1805) {
                                                return 1;
                                            }
                                        }
                                        else if (Rs >= 3278) {
                                            return 2;
                                        }
                                    }
                                    else if (mcTcor >= 85) {
                                        return 2;
                                    }
                                }
                            }
                        }
                    }
                    else if (CsDiff >= 10) {
                        if (mcTcor < 93) {
                            if (mu_mv_mag_sq < 25) {
                                if (ssDCval < 140) {
                                    if (Rs < 2947) {
                                        if (Cs < 2192) {
                                            if (Rs < 1778) {
                                                if (TSC < 5) {
                                                    if (SC < 2542) {
                                                        return 2;
                                                    }
                                                    else if (SC >= 2542) {
                                                        return 4;
                                                    }
                                                }
                                                else if (TSC >= 5) {
                                                    if (tcor < 68) {
                                                        if (refDCval < 126) {
                                                            if (refDCval < 58) {
                                                                return 2;
                                                            }
                                                            else if (refDCval >= 58) {
                                                                if (CsDiff < 14) {
                                                                    return 1;
                                                                }
                                                                else if (CsDiff >= 14) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                        else if (refDCval >= 126) {
                                                            if (MVDiff < 27) {
                                                                return 4;
                                                            }
                                                            else if (MVDiff >= 27) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                    else if (tcor >= 68) {
                                                        if (ssDCval < 111) {
                                                            return 1;
                                                        }
                                                        else if (ssDCval >= 111) {
                                                            if (Rs < 1664) {
                                                                return 1;
                                                            }
                                                            else if (Rs >= 1664) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                            else if (Rs >= 1778) {
                                                if (AFD < 23) {
                                                    if (tcor < 82) {
                                                        if (tcor < 68) {
                                                            if (MVDiff < 44) {
                                                                return 1;
                                                            }
                                                            else if (MVDiff >= 44) {
                                                                return 2;
                                                            }
                                                        }
                                                        else if (tcor >= 68) {
                                                            if (SC < 4303) {
                                                                if (posBalance < 57) {
                                                                    if (SC < 4291) {
                                                                        return 4;
                                                                    }
                                                                    else if (SC >= 4291) {
                                                                        return 1;
                                                                    }
                                                                }
                                                                else if (posBalance >= 57) {
                                                                    if (mcTcor < 88) {
                                                                        if (AFD < 18) {
                                                                            return 2;
                                                                        }
                                                                        else if (AFD >= 18) {
                                                                            if (Cs < 1888) {
                                                                                return 4;
                                                                            }
                                                                            else if (Cs >= 1888) {
                                                                                return 2;
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (mcTcor >= 88) {
                                                                        return 2;
                                                                    }
                                                                }
                                                            }
                                                            else if (SC >= 4303) {
                                                                if (diffMVdiffVal < -10) {
                                                                    if (Rs < 2456) {
                                                                        return 1;
                                                                    }
                                                                    else if (Rs >= 2456) {
                                                                        return 8;
                                                                    }
                                                                }
                                                                else if (diffMVdiffVal >= -10) {
                                                                    if (posBalance < 55) {
                                                                        return 1;
                                                                    }
                                                                    else if (posBalance >= 55) {
                                                                        if (mcTcor < 87) {
                                                                            return 4;
                                                                        }
                                                                        else if (mcTcor >= 87) {
                                                                            if (Cs < 2160) {
                                                                                return 2;
                                                                            }
                                                                            else if (Cs >= 2160) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (tcor >= 82) {
                                                        if (Cs < 2005) {
                                                            return 2;
                                                        }
                                                        else if (Cs >= 2005) {
                                                            return 1;
                                                        }
                                                    }
                                                }
                                                else if (AFD >= 23) {
                                                    if (mu_mv_mag_sq < 11) {
                                                        if (tcor < 58) {
                                                            return 1;
                                                        }
                                                        else if (tcor >= 58) {
                                                            if (SC < 4568) {
                                                                return 2;
                                                            }
                                                            else if (SC >= 4568) {
                                                                if (TSC < 12) {
                                                                    return 2;
                                                                }
                                                                else if (TSC >= 12) {
                                                                    return 1;
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (mu_mv_mag_sq >= 11) {
                                                        if (MVDiff < 83) {
                                                            return 1;
                                                        }
                                                        else if (MVDiff >= 83) {
                                                            if (SC < 3758) {
                                                                return 4;
                                                            }
                                                            else if (SC >= 3758) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (Cs >= 2192) {
                                            if (mcTcor < 84) {
                                                if (AFD < 17) {
                                                    if (posBalance < 52) {
                                                        if (RsDiff < 12) {
                                                            if (TSC < 9) {
                                                                return 4;
                                                            }
                                                            else if (TSC >= 9) {
                                                                return 1;
                                                            }
                                                        }
                                                        else if (RsDiff >= 12) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (posBalance >= 52) {
                                                        return 4;
                                                    }
                                                }
                                                else if (AFD >= 17) {
                                                    if (RsDiff < 15) {
                                                        if (MVDiff < 9) {
                                                            return 1;
                                                        }
                                                        else if (MVDiff >= 9) {
                                                            if (diffAFD < -1) {
                                                                return 1;
                                                            }
                                                            else if (diffAFD >= -1) {
                                                                if (Cs < 2283) {
                                                                    return 1;
                                                                }
                                                                else if (Cs >= 2283) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (RsDiff >= 15) {
                                                        if (AFD < 33) {
                                                            return 4;
                                                        }
                                                        else if (AFD >= 33) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (mcTcor >= 84) {
                                                if (ssDCval < 78) {
                                                    return 8;
                                                }
                                                else if (ssDCval >= 78) {
                                                    if (SC < 4903) {
                                                        if (TSC < 10) {
                                                            if (Cs < 2488) {
                                                                return 1;
                                                            }
                                                            else if (Cs >= 2488) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (TSC >= 10) {
                                                            if (diffAFD < 3) {
                                                                return 4;
                                                            }
                                                            else if (diffAFD >= 3) {
                                                                if (gchDC < 1) {
                                                                    return 1;
                                                                }
                                                                else if (gchDC >= 1) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (SC >= 4903) {
                                                        if (ssDCval < 89) {
                                                            if (mcTcor < 89) {
                                                                if (Cs < 2900) {
                                                                    return 8;
                                                                }
                                                                else if (Cs >= 2900) {
                                                                    return 1;
                                                                }
                                                            }
                                                            else if (mcTcor >= 89) {
                                                                if (diffAFD < -4) {
                                                                    return 4;
                                                                }
                                                                else if (diffAFD >= -4) {
                                                                    return 1;
                                                                }
                                                            }
                                                        }
                                                        else if (ssDCval >= 89) {
                                                            if (MVDiff < 4) {
                                                                if (RsCsDiff < 243) {
                                                                    return 8;
                                                                }
                                                                else if (RsCsDiff >= 243) {
                                                                    return 1;
                                                                }
                                                            }
                                                            else if (MVDiff >= 4) {
                                                                if (Cs < 2480) {
                                                                    if (TSC < 11) {
                                                                        if (diffMVdiffVal < 1) {
                                                                            return 2;
                                                                        }
                                                                        else if (diffMVdiffVal >= 1) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                    else if (TSC >= 11) {
                                                                        if (mu_mv_mag_sq < 19) {
                                                                            return 1;
                                                                        }
                                                                        else if (mu_mv_mag_sq >= 19) {
                                                                            if (TSC < 14) {
                                                                                return 1;
                                                                            }
                                                                            else if (TSC >= 14) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                                else if (Cs >= 2480) {
                                                                    if (TSC < 13) {
                                                                        return 4;
                                                                    }
                                                                    else if (TSC >= 13) {
                                                                        return 1;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (Rs >= 2947) {
                                        if (posBalance < 57) {
                                            if (Cs < 2821) {
                                                return 1;
                                            }
                                            else if (Cs >= 2821) {
                                                if (Rs < 3821) {
                                                    if (CsDiff < 17) {
                                                        if (SC < 6214) {
                                                            return 8;
                                                        }
                                                        else if (SC >= 6214) {
                                                            if (Rs < 3278) {
                                                                return 1;
                                                            }
                                                            else if (Rs >= 3278) {
                                                                if (ssDCval < 129) {
                                                                    return 8;
                                                                }
                                                                else if (ssDCval >= 129) {
                                                                    return 1;
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (CsDiff >= 17) {
                                                        if (MVDiff < 38) {
                                                            return 2;
                                                        }
                                                        else if (MVDiff >= 38) {
                                                            return 1;
                                                        }
                                                    }
                                                }
                                                else if (Rs >= 3821) {
                                                    if (refDCval < 109) {
                                                        return 4;
                                                    }
                                                    else if (refDCval >= 109) {
                                                        if (negBalance < 56) {
                                                            return 1;
                                                        }
                                                        else if (negBalance >= 56) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (posBalance >= 57) {
                                            if (ssDCval < 115) {
                                                if (RsDiff < 19) {
                                                    if (AFD < 18) {
                                                        if (Cs < 2688) {
                                                            return 2;
                                                        }
                                                        else if (Cs >= 2688) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (AFD >= 18) {
                                                        return 1;
                                                    }
                                                }
                                                else if (RsDiff >= 19) {
                                                    return 2;
                                                }
                                            }
                                            else if (ssDCval >= 115) {
                                                if (Rs < 3097) {
                                                    if (RsDiff < 15) {
                                                        return 8;
                                                    }
                                                    else if (RsDiff >= 15) {
                                                        return 1;
                                                    }
                                                }
                                                else if (Rs >= 3097) {
                                                    if (AFD < 14) {
                                                        return 4;
                                                    }
                                                    else if (AFD >= 14) {
                                                        return 1;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (ssDCval >= 140) {
                                    if (SC < 4475) {
                                        return 8;
                                    }
                                    else if (SC >= 4475) {
                                        if (mcTcor < 91) {
                                            if (MVDiff < 13) {
                                                return 2;
                                            }
                                            else if (MVDiff >= 13) {
                                                return 4;
                                            }
                                        }
                                        else if (mcTcor >= 91) {
                                            if (RsDiff < 33) {
                                                return 2;
                                            }
                                            else if (RsDiff >= 33) {
                                                return 4;
                                            }
                                        }
                                    }
                                }
                            }
                            else if (mu_mv_mag_sq >= 25) {
                                if (negBalance < 51) {
                                    if (SC < 7779) {
                                        if (ssDCval < 79) {
                                            if (refDCval < 57) {
                                                if (Rs < 1230) {
                                                    return 4;
                                                }
                                                else if (Rs >= 1230) {
                                                    return 8;
                                                }
                                            }
                                            else if (refDCval >= 57) {
                                                return 1;
                                            }
                                        }
                                        else if (ssDCval >= 79) {
                                            if (ssDCval < 124) {
                                                if (ssDCval < 94) {
                                                    if (negBalance < 39) {
                                                        if (posBalance < 52) {
                                                            if (Rs < 2453) {
                                                                return 8;
                                                            }
                                                            else if (Rs >= 2453) {
                                                                return 1;
                                                            }
                                                        }
                                                        else if (posBalance >= 52) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (negBalance >= 39) {
                                                        if (mu_mv_mag_sq < 38) {
                                                            return 8;
                                                        }
                                                        else if (mu_mv_mag_sq >= 38) {
                                                            if (Rs < 2452) {
                                                                return 8;
                                                            }
                                                            else if (Rs >= 2452) {
                                                                if (Rs < 3125) {
                                                                    return 1;
                                                                }
                                                                else if (Rs >= 3125) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (ssDCval >= 94) {
                                                    if (Rs < 3894) {
                                                        if (posBalance < 57) {
                                                            return 8;
                                                        }
                                                        else if (posBalance >= 57) {
                                                            if (RsCsDiff < 255) {
                                                                if (MVDiff < 6) {
                                                                    if (Rs < 2842) {
                                                                        return 8;
                                                                    }
                                                                    else if (Rs >= 2842) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (MVDiff >= 6) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (RsCsDiff >= 255) {
                                                                if (RsDiff < 13) {
                                                                    if (ssDCval < 105) {
                                                                        return 2;
                                                                    }
                                                                    else if (ssDCval >= 105) {
                                                                        return 8;
                                                                    }
                                                                }
                                                                else if (RsDiff >= 13) {
                                                                    if (MVDiff < 215) {
                                                                        return 8;
                                                                    }
                                                                    else if (MVDiff >= 215) {
                                                                        return 4;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (Rs >= 3894) {
                                                        if (posBalance < 48) {
                                                            return 1;
                                                        }
                                                        else if (posBalance >= 48) {
                                                            if (mcTcor < 88) {
                                                                if (Rs < 4388) {
                                                                    return 2;
                                                                }
                                                                else if (Rs >= 4388) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (mcTcor >= 88) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                            else if (ssDCval >= 124) {
                                                if (ssDCval < 148) {
                                                    if (TSC < 7) {
                                                        return 8;
                                                    }
                                                    else if (TSC >= 7) {
                                                        return 1;
                                                    }
                                                }
                                                else if (ssDCval >= 148) {
                                                    if (RsDiff < 8) {
                                                        return 8;
                                                    }
                                                    else if (RsDiff >= 8) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (SC >= 7779) {
                                        if (RsCsDiff < 357) {
                                            if (ssDCval < 106) {
                                                return 4;
                                            }
                                            else if (ssDCval >= 106) {
                                                return 2;
                                            }
                                        }
                                        else if (RsCsDiff >= 357) {
                                            if (Rs < 6371) {
                                                return 4;
                                            }
                                            else if (Rs >= 6371) {
                                                return 1;
                                            }
                                        }
                                    }
                                }
                                else if (negBalance >= 51) {
                                    if (RsCsDiff < 489) {
                                        if (ssDCval < 133) {
                                            if (AFD < 17) {
                                                if (refDCval < 102) {
                                                    if (negBalance < 54) {
                                                        if (MVDiff < 12) {
                                                            return 4;
                                                        }
                                                        else if (MVDiff >= 12) {
                                                            if (Rs < 2914) {
                                                                return 8;
                                                            }
                                                            else if (Rs >= 2914) {
                                                                return 1;
                                                            }
                                                        }
                                                    }
                                                    else if (negBalance >= 54) {
                                                        if (RsDiff < 10) {
                                                            if (Cs < 2008) {
                                                                return 8;
                                                            }
                                                            else if (Cs >= 2008) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (RsDiff >= 10) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                                else if (refDCval >= 102) {
                                                    return 8;
                                                }
                                            }
                                            else if (AFD >= 17) {
                                                if (diffRsCsdiff < 257) {
                                                    if (refDCval < 89) {
                                                        return 1;
                                                    }
                                                    else if (refDCval >= 89) {
                                                        if (posBalance < 60) {
                                                            return 4;
                                                        }
                                                        else if (posBalance >= 60) {
                                                            if (Rs < 3669) {
                                                                return 1;
                                                            }
                                                            else if (Rs >= 3669) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (diffRsCsdiff >= 257) {
                                                    return 8;
                                                }
                                            }
                                        }
                                        else if (ssDCval >= 133) {
                                            return 1;
                                        }
                                    }
                                    else if (RsCsDiff >= 489) {
                                        if (refDCval < 120) {
                                            if (posBalance < 55) {
                                                if (RsCsDiff < 642) {
                                                    return 8;
                                                }
                                                else if (RsCsDiff >= 642) {
                                                    return 4;
                                                }
                                            }
                                            else if (posBalance >= 55) {
                                                return 2;
                                            }
                                        }
                                        else if (refDCval >= 120) {
                                            return 1;
                                        }
                                    }
                                }
                            }
                        }
                        else if (mcTcor >= 93) {
                            if (Rs < 2563) {
                                if (AFD < 21) {
                                    if (RsDiff < 8) {
                                        if (MVDiff < 4) {
                                            return 8;
                                        }
                                        else if (MVDiff >= 4) {
                                            return 2;
                                        }
                                    }
                                    else if (RsDiff >= 8) {
                                        if (MVDiff < 12) {
                                            if (Cs < 2416) {
                                                if (diffAFD < 1) {
                                                    return 4;
                                                }
                                                else if (diffAFD >= 1) {
                                                    return 2;
                                                }
                                            }
                                            else if (Cs >= 2416) {
                                                return 2;
                                            }
                                        }
                                        else if (MVDiff >= 12) {
                                            if (mcTcor < 95) {
                                                return 1;
                                            }
                                            else if (mcTcor >= 95) {
                                                if (SC < 4038) {
                                                    return 8;
                                                }
                                                else if (SC >= 4038) {
                                                    return 2;
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (AFD >= 21) {
                                    if (Cs < 695) {
                                        return 8;
                                    }
                                    else if (Cs >= 695) {
                                        return 2;
                                    }
                                }
                            }
                            else if (Rs >= 2563) {
                                if (tcor < 84) {
                                    if (posBalance < 59) {
                                        if (SC < 7357) {
                                            if (diffMVdiffVal < -7) {
                                                return 2;
                                            }
                                            else if (diffMVdiffVal >= -7) {
                                                return 4;
                                            }
                                        }
                                        else if (SC >= 7357) {
                                            return 8;
                                        }
                                    }
                                    else if (posBalance >= 59) {
                                        if (Rs < 5012) {
                                            return 2;
                                        }
                                        else if (Rs >= 5012) {
                                            if (Rs < 5120) {
                                                return 4;
                                            }
                                            else if (Rs >= 5120) {
                                                return 2;
                                            }
                                        }
                                    }
                                }
                                else if (tcor >= 84) {
                                    if (ssDCval < 104) {
                                        if (Rs < 3248) {
                                            if (TSC < 10) {
                                                return 4;
                                            }
                                            else if (TSC >= 10) {
                                                return 8;
                                            }
                                        }
                                        else if (Rs >= 3248) {
                                            return 2;
                                        }
                                    }
                                    else if (ssDCval >= 104) {
                                        if (negBalance < 39) {
                                            return 4;
                                        }
                                        else if (negBalance >= 39) {
                                            return 8;
                                        }
                                    }
                                }
                            }
                        }
                    }
                }
                else if (Cs >= 3274) {
                    if (ssDCval < 69) {
                        if (mcTcor < 81) {
                            return 8;
                        }
                        else if (mcTcor >= 81) {
                            if (TSCindex < 8) {
                                if (ssDCval < 55) {
                                    if (posBalance < 56) {
                                        if (Cs < 3848) {
                                            if (ssDCval < 54) {
                                                if (MVDiff < 27) {
                                                    return 4;
                                                }
                                                else if (MVDiff >= 27) {
                                                    if (refDCval < 50) {
                                                        return 8;
                                                    }
                                                    else if (refDCval >= 50) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                            else if (ssDCval >= 54) {
                                                if (RsCsDiff < 11) {
                                                    return 4;
                                                }
                                                else if (RsCsDiff >= 11) {
                                                    if (Rs < 3557) {
                                                        return 4;
                                                    }
                                                    else if (Rs >= 3557) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                        }
                                        else if (Cs >= 3848) {
                                            return 2;
                                        }
                                    }
                                    else if (posBalance >= 56) {
                                        if (SC < 7124) {
                                            if (Rs < 3357) {
                                                if (MVDiff < 18) {
                                                    return 4;
                                                }
                                                else if (MVDiff >= 18) {
                                                    return 8;
                                                }
                                            }
                                            else if (Rs >= 3357) {
                                                return 8;
                                            }
                                        }
                                        else if (SC >= 7124) {
                                            if (diffAFD < -1) {
                                                return 8;
                                            }
                                            else if (diffAFD >= -1) {
                                                return 4;
                                            }
                                        }
                                    }
                                }
                                else if (ssDCval >= 55) {
                                    if (ssDCval < 64) {
                                        if (AFD < 7) {
                                            if (diffRsCsdiff < 8) {
                                                return 8;
                                            }
                                            else if (diffRsCsdiff >= 8) {
                                                return 4;
                                            }
                                        }
                                        else if (AFD >= 7) {
                                            return 8;
                                        }
                                    }
                                    else if (ssDCval >= 64) {
                                        return 4;
                                    }
                                }
                            }
                            else if (TSCindex >= 8) {
                                if (SC < 7329) {
                                    if (RsCsDiff < 37) {
                                        if (Rs < 3372) {
                                            return 8;
                                        }
                                        else if (Rs >= 3372) {
                                            if (posBalance < 54) {
                                                if (Rs < 3456) {
                                                    return 4;
                                                }
                                                else if (Rs >= 3456) {
                                                    return 8;
                                                }
                                            }
                                            else if (posBalance >= 54) {
                                                if (MVDiff < 22) {
                                                    if (Rs < 3705) {
                                                        if (SC < 7046) {
                                                            if (posBalance < 61) {
                                                                return 2;
                                                            }
                                                            else if (posBalance >= 61) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (SC >= 7046) {
                                                            return 2;
                                                        }
                                                    }
                                                    else if (Rs >= 3705) {
                                                        return 4;
                                                    }
                                                }
                                                else if (MVDiff >= 22) {
                                                    if (ssDCval < 52) {
                                                        return 2;
                                                    }
                                                    else if (ssDCval >= 52) {
                                                        if (SC < 6961) {
                                                            if (Rs < 3460) {
                                                                return 8;
                                                            }
                                                            else if (Rs >= 3460) {
                                                                return 2;
                                                            }
                                                        }
                                                        else if (SC >= 6961) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (RsCsDiff >= 37) {
                                        if (AFD < 12) {
                                            return 4;
                                        }
                                        else if (AFD >= 12) {
                                            if (Cs < 3409) {
                                                return 4;
                                            }
                                            else if (Cs >= 3409) {
                                                return 8;
                                            }
                                        }
                                    }
                                }
                                else if (SC >= 7329) {
                                    if (mu_mv_mag_sq < 46) {
                                        if (CsDiff < 6) {
                                            if (MVDiff < 19) {
                                                if (Cs < 3797) {
                                                    return 4;
                                                }
                                                else if (Cs >= 3797) {
                                                    if (Rs < 3975) {
                                                        if (posBalance < 58) {
                                                            return 2;
                                                        }
                                                        else if (posBalance >= 58) {
                                                            if (Cs < 4092) {
                                                                return 4;
                                                            }
                                                            else if (Cs >= 4092) {
                                                                return 2;
                                                            }
                                                        }
                                                    }
                                                    else if (Rs >= 3975) {
                                                        if (Rs < 4029) {
                                                            return 8;
                                                        }
                                                        else if (Rs >= 4029) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (MVDiff >= 19) {
                                                if (posBalance < 60) {
                                                    return 2;
                                                }
                                                else if (posBalance >= 60) {
                                                    if (SC < 7744) {
                                                        return 8;
                                                    }
                                                    else if (SC >= 7744) {
                                                        return 2;
                                                    }
                                                }
                                            }
                                        }
                                        else if (CsDiff >= 6) {
                                            if (MVDiff < 73) {
                                                if (TSCindex < 9) {
                                                    if (AFD < 17) {
                                                        return 8;
                                                    }
                                                    else if (AFD >= 17) {
                                                        return 4;
                                                    }
                                                }
                                                else if (TSCindex >= 9) {
                                                    return 4;
                                                }
                                            }
                                            else if (MVDiff >= 73) {
                                                return 8;
                                            }
                                        }
                                    }
                                    else if (mu_mv_mag_sq >= 46) {
                                        if (ssDCval < 55) {
                                            if (SC < 7987) {
                                                return 8;
                                            }
                                            else if (SC >= 7987) {
                                                return 4;
                                            }
                                        }
                                        else if (ssDCval >= 55) {
                                            if (refDCval < 69) {
                                                return 4;
                                            }
                                            else if (refDCval >= 69) {
                                                if (Cs < 6701) {
                                                    return 4;
                                                }
                                                else if (Cs >= 6701) {
                                                    return 8;
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                        }
                    }
                    else if (ssDCval >= 69) {
                        if (AFD < 26) {
                            if (mu_mv_mag_sq < 3) {
                                if (ssDCval < 153) {
                                    if (refDCval < 91) {
                                        if (Rs < 6637) {
                                            if (Rs < 4953) {
                                                return 4;
                                            }
                                            else if (Rs >= 4953) {
                                                return 1;
                                            }
                                        }
                                        else if (Rs >= 6637) {
                                            return 4;
                                        }
                                    }
                                    else if (refDCval >= 91) {
                                        if (tcor < 63) {
                                            return 1;
                                        }
                                        else if (tcor >= 63) {
                                            if (posBalance < 53) {
                                                if (tcor < 70) {
                                                    return 4;
                                                }
                                                else if (tcor >= 70) {
                                                    return 8;
                                                }
                                            }
                                            else if (posBalance >= 53) {
                                                if (negBalance < 56) {
                                                    if (ssDCval < 100) {
                                                        return 1;
                                                    }
                                                    else if (ssDCval >= 100) {
                                                        if (mcTcor < 83) {
                                                            if (Rs < 3867) {
                                                                return 1;
                                                            }
                                                            else if (Rs >= 3867) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (mcTcor >= 83) {
                                                            if (ssDCval < 107) {
                                                                return 4;
                                                            }
                                                            else if (ssDCval >= 107) {
                                                                if (SC < 12818) {
                                                                    return 8;
                                                                }
                                                                else if (SC >= 12818) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (negBalance >= 56) {
                                                    return 8;
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (ssDCval >= 153) {
                                    if (SC < 13823) {
                                        if (CsDiff < 14) {
                                            if (Cs < 3592) {
                                                return 2;
                                            }
                                            else if (Cs >= 3592) {
                                                return 4;
                                            }
                                        }
                                        else if (CsDiff >= 14) {
                                            return 2;
                                        }
                                    }
                                    else if (SC >= 13823) {
                                        return 8;
                                    }
                                }
                            }
                            else if (mu_mv_mag_sq >= 3) {
                                if (mu_mv_mag_sq < 63) {
                                    if (Rs < 3791) {
                                        if (posBalance < 60) {
                                            return 1;
                                        }
                                        else if (posBalance >= 60) {
                                            if (Cs < 3378) {
                                                return 8;
                                            }
                                            else if (Cs >= 3378) {
                                                return 1;
                                            }
                                        }
                                    }
                                    else if (Rs >= 3791) {
                                        if (posBalance < 66) {
                                            if (Cs < 3782) {
                                                if (Rs < 4908) {
                                                    if (Cs < 3766) {
                                                        if (SC < 7804) {
                                                            if (diffRsCsdiff < -14) {
                                                                return 8;
                                                            }
                                                            else if (diffRsCsdiff >= -14) {
                                                                return 1;
                                                            }
                                                        }
                                                        else if (SC >= 7804) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (Cs >= 3766) {
                                                        return 4;
                                                    }
                                                }
                                                else if (Rs >= 4908) {
                                                    if (CsDiff < 10) {
                                                        if (Rs < 5186) {
                                                            return 4;
                                                        }
                                                        else if (Rs >= 5186) {
                                                            return 2;
                                                        }
                                                    }
                                                    else if (CsDiff >= 10) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                            else if (Cs >= 3782) {
                                                if (MVDiff < 7) {
                                                    return 8;
                                                }
                                                else if (MVDiff >= 7) {
                                                    if (posBalance < 58) {
                                                        if (mu_mv_mag_sq < 21) {
                                                            return 8;
                                                        }
                                                        else if (mu_mv_mag_sq >= 21) {
                                                            if (mu_mv_mag_sq < 31) {
                                                                return 4;
                                                            }
                                                            else if (mu_mv_mag_sq >= 31) {
                                                                if (diffTSC < 2) {
                                                                    if (Rs < 6677) {
                                                                        if (MVDiff < 21) {
                                                                            if (mcTcor < 78) {
                                                                                return 4;
                                                                            }
                                                                            else if (mcTcor >= 78) {
                                                                                if (Cs < 4075) {
                                                                                    if (diffRsCsdiff < 69) {
                                                                                        if (MVDiff < 9) {
                                                                                            if (SC < 8913) {
                                                                                                return 4;
                                                                                            }
                                                                                            else if (SC >= 8913) {
                                                                                                return 8;
                                                                                            }
                                                                                        }
                                                                                        else if (MVDiff >= 9) {
                                                                                            return 4;
                                                                                        }
                                                                                    }
                                                                                    else if (diffRsCsdiff >= 69) {
                                                                                        return 8;
                                                                                    }
                                                                                }
                                                                                else if (Cs >= 4075) {
                                                                                    if (RsDiff < 7) {
                                                                                        return 4;
                                                                                    }
                                                                                    else if (RsDiff >= 7) {
                                                                                        return 8;
                                                                                    }
                                                                                }
                                                                            }
                                                                        }
                                                                        else if (MVDiff >= 21) {
                                                                            if (mu_mv_mag_sq < 33) {
                                                                                if (Rs < 5962) {
                                                                                    return 8;
                                                                                }
                                                                                else if (Rs >= 5962) {
                                                                                    return 4;
                                                                                }
                                                                            }
                                                                            else if (mu_mv_mag_sq >= 33) {
                                                                                return 8;
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (Rs >= 6677) {
                                                                        if (TSC < 18) {
                                                                            return 4;
                                                                        }
                                                                        else if (TSC >= 18) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                }
                                                                else if (diffTSC >= 2) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (posBalance >= 58) {
                                                        if (Rs < 6322) {
                                                            if (diffTSC < 3) {
                                                                return 8;
                                                            }
                                                            else if (diffTSC >= 3) {
                                                                if (Cs < 4030) {
                                                                    return 8;
                                                                }
                                                                else if (Cs >= 4030) {
                                                                    return 1;
                                                                }
                                                            }
                                                        }
                                                        else if (Rs >= 6322) {
                                                            if (Cs < 5414) {
                                                                return 1;
                                                            }
                                                            else if (Cs >= 5414) {
                                                                if (Rs < 7699) {
                                                                    return 4;
                                                                }
                                                                else if (Rs >= 7699) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (posBalance >= 66) {
                                            return 4;
                                        }
                                    }
                                }
                                else if (mu_mv_mag_sq >= 63) {
                                    if (posBalance < 59) {
                                        if (mcTcor < 82) {
                                            if (negBalance < 43) {
                                                return 8;
                                            }
                                            else if (negBalance >= 43) {
                                                if (refDCval < 129) {
                                                    if (tcor < 67) {
                                                        return 8;
                                                    }
                                                    else if (tcor >= 67) {
                                                        if (ssDCval < 86) {
                                                            if (Rs < 4759) {
                                                                return 4;
                                                            }
                                                            else if (Rs >= 4759) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (ssDCval >= 86) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                                else if (refDCval >= 129) {
                                                    return 1;
                                                }
                                            }
                                        }
                                        else if (mcTcor >= 82) {
                                            if (Cs < 3706) {
                                                if (Rs < 4361) {
                                                    return 8;
                                                }
                                                else if (Rs >= 4361) {
                                                    if (Rs < 4638) {
                                                        return 4;
                                                    }
                                                    else if (Rs >= 4638) {
                                                        return 2;
                                                    }
                                                }
                                            }
                                            else if (Cs >= 3706) {
                                                return 8;
                                            }
                                        }
                                    }
                                    else if (posBalance >= 59) {
                                        if (Cs < 3591) {
                                            if (ssDCval < 83) {
                                                return 4;
                                            }
                                            else if (ssDCval >= 83) {
                                                return 8;
                                            }
                                        }
                                        else if (Cs >= 3591) {
                                            return 8;
                                        }
                                    }
                                }
                            }
                        }
                        else if (AFD >= 26) {
                            if (SC < 6727) {
                                if (mcTcor < 69) {
                                    return 4;
                                }
                                else if (mcTcor >= 69) {
                                    return 1;
                                }
                            }
                            else if (SC >= 6727) {
                                if (Rs < 6730) {
                                    if (refDCval < 102) {
                                        if (mcTcor < 68) {
                                            if (Cs < 3621) {
                                                return 4;
                                            }
                                            else if (Cs >= 3621) {
                                                return 8;
                                            }
                                        }
                                        else if (mcTcor >= 68) {
                                            if (Cs < 5079) {
                                                return 2;
                                            }
                                            else if (Cs >= 5079) {
                                                return 1;
                                            }
                                        }
                                    }
                                    else if (refDCval >= 102) {
                                        if (mcTcor < 89) {
                                            return 8;
                                        }
                                        else if (mcTcor >= 89) {
                                            if (ssDCval < 134) {
                                                if (MVDiff < 18) {
                                                    return 8;
                                                }
                                                else if (MVDiff >= 18) {
                                                    return 4;
                                                }
                                            }
                                            else if (ssDCval >= 134) {
                                                return 2;
                                            }
                                        }
                                    }
                                }
                                else if (Rs >= 6730) {
                                    if (mcTcor < 87) {
                                        if (ssDCval < 127) {
                                            if (mu_mv_mag_sq < 63) {
                                                if (Rs < 14727) {
                                                    if (tcor < 76) {
                                                        if (SC < 14100) {
                                                            if (Cs < 6238) {
                                                                return 4;
                                                            }
                                                            else if (Cs >= 6238) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (SC >= 14100) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (tcor >= 76) {
                                                        return 8;
                                                    }
                                                }
                                                else if (Rs >= 14727) {
                                                    return 8;
                                                }
                                            }
                                            else if (mu_mv_mag_sq >= 63) {
                                                return 8;
                                            }
                                        }
                                        else if (ssDCval >= 127) {
                                            if (RsCsDiff < 4958) {
                                                if (Cs < 4439) {
                                                    return 4;
                                                }
                                                else if (Cs >= 4439) {
                                                    if (Cs < 5277) {
                                                        if (Rs < 7649) {
                                                            return 2;
                                                        }
                                                        else if (Rs >= 7649) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (Cs >= 5277) {
                                                        if (posBalance < 52) {
                                                            if (Rs < 8745) {
                                                                return 8;
                                                            }
                                                            else if (Rs >= 8745) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (posBalance >= 52) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (RsCsDiff >= 4958) {
                                                if (AFD < 48) {
                                                    return 4;
                                                }
                                                else if (AFD >= 48) {
                                                    return 2;
                                                }
                                            }
                                        }
                                    }
                                    else if (mcTcor >= 87) {
                                        if (MVDiff < 22) {
                                            if (ssDCval < 129) {
                                                if (MVDiff < 12) {
                                                    return 8;
                                                }
                                                else if (MVDiff >= 12) {
                                                    return 2;
                                                }
                                            }
                                            else if (ssDCval >= 129) {
                                                if (CsDiff < 24) {
                                                    return 4;
                                                }
                                                else if (CsDiff >= 24) {
                                                    if (refDCval < 141) {
                                                        return 8;
                                                    }
                                                    else if (refDCval >= 141) {
                                                        if (RsDiff < 31) {
                                                            if (Rs < 7016) {
                                                                return 4;
                                                            }
                                                            else if (Rs >= 7016) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (RsDiff >= 31) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (MVDiff >= 22) {
                                            if (posBalance < 53) {
                                                return 8;
                                            }
                                            else if (posBalance >= 53) {
                                                return 2;
                                            }
                                        }
                                    }
                                }
                            }
                        }
                    }
                }
            }
        }
        else if (mu_mv_mag_sq >= 91) {
            if (AFD < 7) {
                if (Rs < 573) {
                    if (refDCval < 161) {
                        if (ssDCval < 92) {
                            if (Cs < 512) {
                                if (diffTSC < 1) {
                                    if (mu_mv_mag_sq < 115) {
                                        if (SC < 506) {
                                            if (Rs < 109) {
                                                return 8;
                                            }
                                            else if (Rs >= 109) {
                                                return 1;
                                            }
                                        }
                                        else if (SC >= 506) {
                                            if (Cs < 396) {
                                                if (Rs < 327) {
                                                    if (refDCval < 84) {
                                                        return 4;
                                                    }
                                                    else if (refDCval >= 84) {
                                                        return 8;
                                                    }
                                                }
                                                else if (Rs >= 327) {
                                                    return 8;
                                                }
                                            }
                                            else if (Cs >= 396) {
                                                return 2;
                                            }
                                        }
                                    }
                                    else if (mu_mv_mag_sq >= 115) {
                                        if (SC < 608) {
                                            if (MVDiff < 486) {
                                                if (ssDCval < 75) {
                                                    if (Rs < 184) {
                                                        if (AFD < 1) {
                                                            return 8;
                                                        }
                                                        else if (AFD >= 1) {
                                                            if (negBalance < 16) {
                                                                return 8;
                                                            }
                                                            else if (negBalance >= 16) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                    else if (Rs >= 184) {
                                                        if (negBalance < 14) {
                                                            if (negBalance < 13) {
                                                                return 8;
                                                            }
                                                            else if (negBalance >= 13) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (negBalance >= 14) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                                else if (ssDCval >= 75) {
                                                    return 4;
                                                }
                                            }
                                            else if (MVDiff >= 486) {
                                                return 1;
                                            }
                                        }
                                        else if (SC >= 608) {
                                            return 8;
                                        }
                                    }
                                }
                                else if (diffTSC >= 1) {
                                    if (mcTcor < 100) {
                                        return 4;
                                    }
                                    else if (mcTcor >= 100) {
                                        return 8;
                                    }
                                }
                            }
                            else if (Cs >= 512) {
                                if (posBalance < 49) {
                                    if (ssDCval < 38) {
                                        return 8;
                                    }
                                    else if (ssDCval >= 38) {
                                        if (Cs < 979) {
                                            return 4;
                                        }
                                        else if (Cs >= 979) {
                                            return 8;
                                        }
                                    }
                                }
                                else if (posBalance >= 49) {
                                    return 8;
                                }
                            }
                        }
                        else if (ssDCval >= 92) {
                            if (negBalance < 32) {
                                if (diffMVdiffVal < 7) {
                                    return 8;
                                }
                                else if (diffMVdiffVal >= 7) {
                                    return 4;
                                }
                            }
                            else if (negBalance >= 32) {
                                if (Cs < 1208) {
                                    return 4;
                                }
                                else if (Cs >= 1208) {
                                    if (RsCsDiff < 7) {
                                        return 8;
                                    }
                                    else if (RsCsDiff >= 7) {
                                        return 4;
                                    }
                                }
                            }
                        }
                    }
                    else if (refDCval >= 161) {
                        if (SC < 885) {
                            return 4;
                        }
                        else if (SC >= 885) {
                            if (refDCval < 165) {
                                if (diffRsCsdiff < -1) {
                                    return 2;
                                }
                                else if (diffRsCsdiff >= -1) {
                                    return 4;
                                }
                            }
                            else if (refDCval >= 165) {
                                return 2;
                            }
                        }
                    }
                }
                else if (Rs >= 573) {
                    if (negBalance < 35) {
                        if (Rs < 3779) {
                            if (SC < 7764) {
                                if (refDCval < 70) {
                                    if (ssDCval < 33) {
                                        if (Cs < 708) {
                                            if (Cs < 650) {
                                                return 8;
                                            }
                                            else if (Cs >= 650) {
                                                return 4;
                                            }
                                        }
                                        else if (Cs >= 708) {
                                            if (ssDCval < 31) {
                                                if (diffMVdiffVal < 21) {
                                                    return 8;
                                                }
                                                else if (diffMVdiffVal >= 21) {
                                                    return 4;
                                                }
                                            }
                                            else if (ssDCval >= 31) {
                                                if (Rs < 983) {
                                                    if (MVDiff < 6) {
                                                        return 8;
                                                    }
                                                    else if (MVDiff >= 6) {
                                                        if (AFD < 1) {
                                                            return 1;
                                                        }
                                                        else if (AFD >= 1) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                                else if (Rs >= 983) {
                                                    if (SC < 2312) {
                                                        if (RsDiff < 1) {
                                                            return 4;
                                                        }
                                                        else if (RsDiff >= 1) {
                                                            return 2;
                                                        }
                                                    }
                                                    else if (SC >= 2312) {
                                                        if (Rs < 1087) {
                                                            return 2;
                                                        }
                                                        else if (Rs >= 1087) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (ssDCval >= 33) {
                                        if (posBalance < 67) {
                                            if (MVDiff < 2) {
                                                if (RsDiff < 4) {
                                                    return 8;
                                                }
                                                else if (RsDiff >= 4) {
                                                    if (mu_mv_mag_sq < 123) {
                                                        return 8;
                                                    }
                                                    else if (mu_mv_mag_sq >= 123) {
                                                        if (posBalance < 33) {
                                                            return 8;
                                                        }
                                                        else if (posBalance >= 33) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (MVDiff >= 2) {
                                                if (diffAFD < -1) {
                                                    if (diffMVdiffVal < -1) {
                                                        if (Rs < 1063) {
                                                            return 8;
                                                        }
                                                        else if (Rs >= 1063) {
                                                            if (Rs < 2116) {
                                                                return 1;
                                                            }
                                                            else if (Rs >= 2116) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                    else if (diffMVdiffVal >= -1) {
                                                        return 8;
                                                    }
                                                }
                                                else if (diffAFD >= -1) {
                                                    return 8;
                                                }
                                            }
                                        }
                                        else if (posBalance >= 67) {
                                            return 4;
                                        }
                                    }
                                }
                                else if (refDCval >= 70) {
                                    if (posBalance < 29) {
                                        if (AFD < 2) {
                                            if (negBalance < 3) {
                                                if (Cs < 3876) {
                                                    if (diffRsCsdiff < 2) {
                                                        if (gchDC < 1) {
                                                            if (SC < 7308) {
                                                                if (diffAFD < -1) {
                                                                    if (ssDCval < 140) {
                                                                        if (refDCval < 127) {
                                                                            if (Cs < 2912) {
                                                                                return 1;
                                                                            }
                                                                            else if (Cs >= 2912) {
                                                                                return 8;
                                                                            }
                                                                        }
                                                                        else if (refDCval >= 127) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                    else if (ssDCval >= 140) {
                                                                        return 1;
                                                                    }
                                                                }
                                                                else if (diffAFD >= -1) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (SC >= 7308) {
                                                                if (SC < 7541) {
                                                                    if (Cs < 3849) {
                                                                        return 1;
                                                                    }
                                                                    else if (Cs >= 3849) {
                                                                        return 8;
                                                                    }
                                                                }
                                                                else if (SC >= 7541) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                        else if (gchDC >= 1) {
                                                            if (Rs < 2195) {
                                                                return 1;
                                                            }
                                                            else if (Rs >= 2195) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                    else if (diffRsCsdiff >= 2) {
                                                        if (ssDCval < 104) {
                                                            return 1;
                                                        }
                                                        else if (ssDCval >= 104) {
                                                            if (diffRsCsdiff < 12) {
                                                                return 8;
                                                            }
                                                            else if (diffRsCsdiff >= 12) {
                                                                return 1;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (Cs >= 3876) {
                                                    return 8;
                                                }
                                            }
                                            else if (negBalance >= 3) {
                                                if (diffMVdiffVal < 1) {
                                                    return 8;
                                                }
                                                else if (diffMVdiffVal >= 1) {
                                                    if (refDCval < 119) {
                                                        return 8;
                                                    }
                                                    else if (refDCval >= 119) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                        }
                                        else if (AFD >= 2) {
                                            if (ssDCval < 85) {
                                                if (Rs < 2151) {
                                                    return 1;
                                                }
                                                else if (Rs >= 2151) {
                                                    return 8;
                                                }
                                            }
                                            else if (ssDCval >= 85) {
                                                if (SC < 2769) {
                                                    if (posBalance < 24) {
                                                        if (diffMVdiffVal < -3) {
                                                            return 1;
                                                        }
                                                        else if (diffMVdiffVal >= -3) {
                                                            if (tcor < 92) {
                                                                if (MVDiff < 3) {
                                                                    return 8;
                                                                }
                                                                else if (MVDiff >= 3) {
                                                                    if (AFD < 5) {
                                                                        if (refDCval < 141) {
                                                                            return 1;
                                                                        }
                                                                        else if (refDCval >= 141) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                    else if (AFD >= 5) {
                                                                        return 8;
                                                                    }
                                                                }
                                                            }
                                                            else if (tcor >= 92) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                    else if (posBalance >= 24) {
                                                        return 4;
                                                    }
                                                }
                                                else if (SC >= 2769) {
                                                    if (Cs < 2416) {
                                                        if (refDCval < 116) {
                                                            return 8;
                                                        }
                                                        else if (refDCval >= 116) {
                                                            if (CsDiff < 5) {
                                                                if (diffRsCsdiff < -18) {
                                                                    if (Rs < 1253) {
                                                                        return 4;
                                                                    }
                                                                    else if (Rs >= 1253) {
                                                                        return 8;
                                                                    }
                                                                }
                                                                else if (diffRsCsdiff >= -18) {
                                                                    if (refDCval < 117) {
                                                                        if (Rs < 2392) {
                                                                            return 8;
                                                                        }
                                                                        else if (Rs >= 2392) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                    else if (refDCval >= 117) {
                                                                        return 4;
                                                                    }
                                                                }
                                                            }
                                                            else if (CsDiff >= 5) {
                                                                if (TSC < 3) {
                                                                    return 8;
                                                                }
                                                                else if (TSC >= 3) {
                                                                    return 1;
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (Cs >= 2416) {
                                                        if (tcor < 94) {
                                                            if (SC < 6578) {
                                                                return 8;
                                                            }
                                                            else if (SC >= 6578) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (tcor >= 94) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (posBalance >= 29) {
                                        if (CsDiff < 3) {
                                            if (negBalance < 12) {
                                                return 8;
                                            }
                                            else if (negBalance >= 12) {
                                                if (RsCsDiff < 1) {
                                                    if (refDCval < 90) {
                                                        return 8;
                                                    }
                                                    else if (refDCval >= 90) {
                                                        if (Cs < 1442) {
                                                            if (SC < 2703) {
                                                                if (Cs < 441) {
                                                                    return 8;
                                                                }
                                                                else if (Cs >= 441) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (SC >= 2703) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (Cs >= 1442) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                                else if (RsCsDiff >= 1) {
                                                    if (Rs < 3759) {
                                                        if (AFD < 4) {
                                                            if (RsCsDiff < 12) {
                                                                return 8;
                                                            }
                                                            else if (RsCsDiff >= 12) {
                                                                if (Cs < 2936) {
                                                                    return 8;
                                                                }
                                                                else if (Cs >= 2936) {
                                                                    if (Cs < 3025) {
                                                                        return 1;
                                                                    }
                                                                    else if (Cs >= 3025) {
                                                                        return 8;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (AFD >= 4) {
                                                            if (RsCsDiff < 7) {
                                                                return 4;
                                                            }
                                                            else if (RsCsDiff >= 7) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                    else if (Rs >= 3759) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                        }
                                        else if (CsDiff >= 3) {
                                            if (SC < 4233) {
                                                if (RsDiff < 3) {
                                                    if (ssDCval < 108) {
                                                        return 8;
                                                    }
                                                    else if (ssDCval >= 108) {
                                                        return 4;
                                                    }
                                                }
                                                else if (RsDiff >= 3) {
                                                    return 8;
                                                }
                                            }
                                            else if (SC >= 4233) {
                                                if (mu_mv_mag_sq < 125) {
                                                    if (Rs < 3393) {
                                                        if (Cs < 1469) {
                                                            return 1;
                                                        }
                                                        else if (Cs >= 1469) {
                                                            if (Rs < 3162) {
                                                                if (mcTcor < 98) {
                                                                    if (MVDiff < 1) {
                                                                        if (CsDiff < 5) {
                                                                            return 4;
                                                                        }
                                                                        else if (CsDiff >= 5) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                    else if (MVDiff >= 1) {
                                                                        if (MVDiff < 1504) {
                                                                            if (mu_mv_mag_sq < 116) {
                                                                                return 8;
                                                                            }
                                                                            else if (mu_mv_mag_sq >= 116) {
                                                                                if (Rs < 2678) {
                                                                                    return 1;
                                                                                }
                                                                                else if (Rs >= 2678) {
                                                                                    return 8;
                                                                                }
                                                                            }
                                                                        }
                                                                        else if (MVDiff >= 1504) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                }
                                                                else if (mcTcor >= 98) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (Rs >= 3162) {
                                                                if (SC < 6281) {
                                                                    if (refDCval < 108) {
                                                                        return 8;
                                                                    }
                                                                    else if (refDCval >= 108) {
                                                                        if (Cs < 2434) {
                                                                            return 8;
                                                                        }
                                                                        else if (Cs >= 2434) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                }
                                                                else if (SC >= 6281) {
                                                                    if (Rs < 3341) {
                                                                        if (posBalance < 41) {
                                                                            if (SC < 6382) {
                                                                                return 1;
                                                                            }
                                                                            else if (SC >= 6382) {
                                                                                return 8;
                                                                            }
                                                                        }
                                                                        else if (posBalance >= 41) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                    else if (Rs >= 3341) {
                                                                        return 1;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (Rs >= 3393) {
                                                        return 8;
                                                    }
                                                }
                                                else if (mu_mv_mag_sq >= 125) {
                                                    return 8;
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                            else if (SC >= 7764) {
                                return 1;
                            }
                        }
                        else if (Rs >= 3779) {
                            if (diffMVdiffVal < -8) {
                                if (AFD < 6) {
                                    return 8;
                                }
                                else if (AFD >= 6) {
                                    return 4;
                                }
                            }
                            else if (diffMVdiffVal >= -8) {
                                return 8;
                            }
                        }
                    }
                    else if (negBalance >= 35) {
                        if (refDCval < 84) {
                            if (Rs < 1762) {
                                if (TSC < 6) {
                                    if (mcTcor < 85) {
                                        return 4;
                                    }
                                    else if (mcTcor >= 85) {
                                        if (Cs < 1717) {
                                            return 8;
                                        }
                                        else if (Cs >= 1717) {
                                            if (Rs < 1708) {
                                                return 8;
                                            }
                                            else if (Rs >= 1708) {
                                                if (posBalance < 55) {
                                                    return 4;
                                                }
                                                else if (posBalance >= 55) {
                                                    return 8;
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (TSC >= 6) {
                                    if (CsDiff < 4) {
                                        if (SC < 2249) {
                                            return 8;
                                        }
                                        else if (SC >= 2249) {
                                            return 2;
                                        }
                                    }
                                    else if (CsDiff >= 4) {
                                        if (mu_mv_mag_sq < 119) {
                                            if (Cs < 1584) {
                                                if (Rs < 1679) {
                                                    return 8;
                                                }
                                                else if (Rs >= 1679) {
                                                    return 2;
                                                }
                                            }
                                            else if (Cs >= 1584) {
                                                return 8;
                                            }
                                        }
                                        else if (mu_mv_mag_sq >= 119) {
                                            return 8;
                                        }
                                    }
                                }
                            }
                            else if (Rs >= 1762) {
                                if (SC < 4920) {
                                    if (SC < 3545) {
                                        if (mu_mv_mag_sq < 108) {
                                            if (TSC < 6) {
                                                return 4;
                                            }
                                            else if (TSC >= 6) {
                                                return 2;
                                            }
                                        }
                                        else if (mu_mv_mag_sq >= 108) {
                                            if (mu_mv_mag_sq < 117) {
                                                if (mcTcor < 94) {
                                                    return 8;
                                                }
                                                else if (mcTcor >= 94) {
                                                    if (SC < 3107) {
                                                        return 8;
                                                    }
                                                    else if (SC >= 3107) {
                                                        if (ssDCval < 72) {
                                                            if (Rs < 1904) {
                                                                return 8;
                                                            }
                                                            else if (Rs >= 1904) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (ssDCval >= 72) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (mu_mv_mag_sq >= 117) {
                                                if (refDCval < 74) {
                                                    if (SC < 3532) {
                                                        if (mu_mv_mag_sq < 121) {
                                                            if (RsDiff < 3) {
                                                                if (Rs < 1838) {
                                                                    return 4;
                                                                }
                                                                else if (Rs >= 1838) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (RsDiff >= 3) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (mu_mv_mag_sq >= 121) {
                                                            if (SC < 3498) {
                                                                return 2;
                                                            }
                                                            else if (SC >= 3498) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                    else if (SC >= 3532) {
                                                        return 2;
                                                    }
                                                }
                                                else if (refDCval >= 74) {
                                                    if (SC < 3490) {
                                                        if (diffTSC < -1) {
                                                            if (diffRsCsdiff < 4) {
                                                                return 8;
                                                            }
                                                            else if (diffRsCsdiff >= 4) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (diffTSC >= -1) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (SC >= 3490) {
                                                        if (Cs < 1612) {
                                                            if (SC < 3522) {
                                                                return 4;
                                                            }
                                                            else if (SC >= 3522) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (Cs >= 1612) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (SC >= 3545) {
                                        if (negBalance < 42) {
                                            if (MVDiff < 2) {
                                                if (mcTcor < 95) {
                                                    if (posBalance < 56) {
                                                        return 4;
                                                    }
                                                    else if (posBalance >= 56) {
                                                        return 8;
                                                    }
                                                }
                                                else if (mcTcor >= 95) {
                                                    return 4;
                                                }
                                            }
                                            else if (MVDiff >= 2) {
                                                if (posBalance < 32) {
                                                    return 8;
                                                }
                                                else if (posBalance >= 32) {
                                                    if (Rs < 2722) {
                                                        return 4;
                                                    }
                                                    else if (Rs >= 2722) {
                                                        return 1;
                                                    }
                                                }
                                            }
                                        }
                                        else if (negBalance >= 42) {
                                            if (tcor < 94) {
                                                return 8;
                                            }
                                            else if (tcor >= 94) {
                                                if (Rs < 1860) {
                                                    return 8;
                                                }
                                                else if (Rs >= 1860) {
                                                    if (MVDiff < 3) {
                                                        if (Rs < 1984) {
                                                            if (AFD < 6) {
                                                                if (Rs < 1972) {
                                                                    return 4;
                                                                }
                                                                else if (Rs >= 1972) {
                                                                    return 2;
                                                                }
                                                            }
                                                            else if (AFD >= 6) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (Rs >= 1984) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (MVDiff >= 3) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (SC >= 4920) {
                                    if (RsCsDiff < 92) {
                                        if (diffAFD < -2) {
                                            if (Rs < 3383) {
                                                return 8;
                                            }
                                            else if (Rs >= 3383) {
                                                return 4;
                                            }
                                        }
                                        else if (diffAFD >= -2) {
                                            return 8;
                                        }
                                    }
                                    else if (RsCsDiff >= 92) {
                                        if (Cs < 5050) {
                                            if (mu_mv_mag_sq < 127) {
                                                return 8;
                                            }
                                            else if (mu_mv_mag_sq >= 127) {
                                                if (diffAFD < -1) {
                                                    return 8;
                                                }
                                                else if (diffAFD >= -1) {
                                                    if (Cs < 4873) {
                                                        if (Rs < 4547) {
                                                            return 4;
                                                        }
                                                        else if (Rs >= 4547) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (Cs >= 4873) {
                                                        if (diffRsCsdiff < -16) {
                                                            if (RsDiff < 8) {
                                                                return 4;
                                                            }
                                                            else if (RsDiff >= 8) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (diffRsCsdiff >= -16) {
                                                            if (Rs < 4964) {
                                                                if (posBalance < 58) {
                                                                    return 4;
                                                                }
                                                                else if (posBalance >= 58) {
                                                                    if (Rs < 4877) {
                                                                        if (RsDiff < 8) {
                                                                            return 4;
                                                                        }
                                                                        else if (RsDiff >= 8) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                    else if (Rs >= 4877) {
                                                                        return 4;
                                                                    }
                                                                }
                                                            }
                                                            else if (Rs >= 4964) {
                                                                if (Cs < 4999) {
                                                                    return 4;
                                                                }
                                                                else if (Cs >= 4999) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (Cs >= 5050) {
                                            if (Cs < 5826) {
                                                return 4;
                                            }
                                            else if (Cs >= 5826) {
                                                return 8;
                                            }
                                        }
                                    }
                                }
                            }
                        }
                        else if (refDCval >= 84) {
                            if (Rs < 1751) {
                                if (RsCsDiff < 23) {
                                    if (posBalance < 9) {
                                        return 8;
                                    }
                                    else if (posBalance >= 9) {
                                        return 4;
                                    }
                                }
                                else if (RsCsDiff >= 23) {
                                    if (SC < 3838) {
                                        return 8;
                                    }
                                    else if (SC >= 3838) {
                                        return 4;
                                    }
                                }
                            }
                            else if (Rs >= 1751) {
                                if (SC < 6135) {
                                    if (mu_mv_mag_sq < 123) {
                                        if (mu_mv_mag_sq < 98) {
                                            return 1;
                                        }
                                        else if (mu_mv_mag_sq >= 98) {
                                            if (refDCval < 115) {
                                                if (Cs < 2984) {
                                                    if (refDCval < 89) {
                                                        if (refDCval < 87) {
                                                            if (gchDC < 1) {
                                                                return 8;
                                                            }
                                                            else if (gchDC >= 1) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (refDCval >= 87) {
                                                            if (ssDCval < 87) {
                                                                return 8;
                                                            }
                                                            else if (ssDCval >= 87) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                    else if (refDCval >= 89) {
                                                        if (mu_mv_mag_sq < 102) {
                                                            return 4;
                                                        }
                                                        else if (mu_mv_mag_sq >= 102) {
                                                            if (ssDCval < 113) {
                                                                if (Rs < 2354) {
                                                                    if (Rs < 2288) {
                                                                        return 8;
                                                                    }
                                                                    else if (Rs >= 2288) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (Rs >= 2354) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (ssDCval >= 113) {
                                                                if (Rs < 2663) {
                                                                    return 8;
                                                                }
                                                                else if (Rs >= 2663) {
                                                                    if (SC < 5708) {
                                                                        if (Cs < 2867) {
                                                                            return 8;
                                                                        }
                                                                        else if (Cs >= 2867) {
                                                                            if (Rs < 2714) {
                                                                                if (Rs < 2688) {
                                                                                    return 4;
                                                                                }
                                                                                else if (Rs >= 2688) {
                                                                                    return 8;
                                                                                }
                                                                            }
                                                                            else if (Rs >= 2714) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (SC >= 5708) {
                                                                        return 8;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (Cs >= 2984) {
                                                    if (ssDCval < 111) {
                                                        return 8;
                                                    }
                                                    else if (ssDCval >= 111) {
                                                        if (SC < 5698) {
                                                            return 8;
                                                        }
                                                        else if (SC >= 5698) {
                                                            if (mcTcor < 97) {
                                                                return 8;
                                                            }
                                                            else if (mcTcor >= 97) {
                                                                if (posBalance < 59) {
                                                                    if (CsDiff < 5) {
                                                                        if (posBalance < 57) {
                                                                            if (Rs < 2866) {
                                                                                return 8;
                                                                            }
                                                                            else if (Rs >= 2866) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                        else if (posBalance >= 57) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                    else if (CsDiff >= 5) {
                                                                        if (mu_mv_mag_sq < 121) {
                                                                            return 4;
                                                                        }
                                                                        else if (mu_mv_mag_sq >= 121) {
                                                                            if (ssDCval < 114) {
                                                                                return 8;
                                                                            }
                                                                            else if (ssDCval >= 114) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                                else if (posBalance >= 59) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                            else if (refDCval >= 115) {
                                                return 8;
                                            }
                                        }
                                    }
                                    else if (mu_mv_mag_sq >= 123) {
                                        if (negBalance < 48) {
                                            if (posBalance < 42) {
                                                return 4;
                                            }
                                            else if (posBalance >= 42) {
                                                if (refDCval < 85) {
                                                    if (Rs < 2176) {
                                                        return 8;
                                                    }
                                                    else if (Rs >= 2176) {
                                                        return 4;
                                                    }
                                                }
                                                else if (refDCval >= 85) {
                                                    return 8;
                                                }
                                            }
                                        }
                                        else if (negBalance >= 48) {
                                            if (Cs < 2939) {
                                                return 8;
                                            }
                                            else if (Cs >= 2939) {
                                                if (SC < 5750) {
                                                    return 8;
                                                }
                                                else if (SC >= 5750) {
                                                    if (SC < 5769) {
                                                        return 4;
                                                    }
                                                    else if (SC >= 5769) {
                                                        if (refDCval < 112) {
                                                            return 8;
                                                        }
                                                        else if (refDCval >= 112) {
                                                            if (Cs < 3078) {
                                                                if (diffRsCsdiff < -5) {
                                                                    return 4;
                                                                }
                                                                else if (diffRsCsdiff >= -5) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (Cs >= 3078) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (SC >= 6135) {
                                    if (RsDiff < 9) {
                                        return 8;
                                    }
                                    else if (RsDiff >= 9) {
                                        if (Rs < 5452) {
                                            return 4;
                                        }
                                        else if (Rs >= 5452) {
                                            return 8;
                                        }
                                    }
                                }
                            }
                        }
                    }
                }
            }
            else if (AFD >= 7) {
                if (SC < 3668) {
                    if (RsDiff < 5) {
                        if (TSCindex < 9) {
                            if (ssDCval < 72) {
                                if (diffAFD < 5) {
                                    return 2;
                                }
                                else if (diffAFD >= 5) {
                                    return 8;
                                }
                            }
                            else if (ssDCval >= 72) {
                                if (mcTcor < 92) {
                                    if (mu_mv_mag_sq < 99) {
                                        if (MVDiff < 6) {
                                            return 4;
                                        }
                                        else if (MVDiff >= 6) {
                                            return 8;
                                        }
                                    }
                                    else if (mu_mv_mag_sq >= 99) {
                                        return 8;
                                    }
                                }
                                else if (mcTcor >= 92) {
                                    if (negBalance < 45) {
                                        if (posBalance < 58) {
                                            return 8;
                                        }
                                        else if (posBalance >= 58) {
                                            if (diffAFD < 1) {
                                                return 2;
                                            }
                                            else if (diffAFD >= 1) {
                                                return 8;
                                            }
                                        }
                                    }
                                    else if (negBalance >= 45) {
                                        if (diffAFD < 1) {
                                            if (diffTSC < -1) {
                                                if (Cs < 1666) {
                                                    return 2;
                                                }
                                                else if (Cs >= 1666) {
                                                    return 8;
                                                }
                                            }
                                            else if (diffTSC >= -1) {
                                                return 2;
                                            }
                                        }
                                        else if (diffAFD >= 1) {
                                            if (refDCval < 76) {
                                                return 2;
                                            }
                                            else if (refDCval >= 76) {
                                                if (Cs < 1514) {
                                                    return 2;
                                                }
                                                else if (Cs >= 1514) {
                                                    return 4;
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                        }
                        else if (TSCindex >= 9) {
                            if (SC < 1200) {
                                return 1;
                            }
                            else if (SC >= 1200) {
                                return 2;
                            }
                        }
                    }
                    else if (RsDiff >= 5) {
                        if (MVDiff < 13) {
                            if (tcor < 87) {
                                return 8;
                            }
                            else if (tcor >= 87) {
                                if (tcor < 94) {
                                    if (mu_mv_mag_sq < 113) {
                                        return 2;
                                    }
                                    else if (mu_mv_mag_sq >= 113) {
                                        if (Rs < 1870) {
                                            return 8;
                                        }
                                        else if (Rs >= 1870) {
                                            return 2;
                                        }
                                    }
                                }
                                else if (tcor >= 94) {
                                    if (mu_mv_mag_sq < 97) {
                                        return 2;
                                    }
                                    else if (mu_mv_mag_sq >= 97) {
                                        return 4;
                                    }
                                }
                            }
                        }
                        else if (MVDiff >= 13) {
                            if (Rs < 2534) {
                                if (diffAFD < 28) {
                                    if (mcTcor < 93) {
                                        return 1;
                                    }
                                    else if (mcTcor >= 93) {
                                        if (posBalance < 66) {
                                            return 8;
                                        }
                                        else if (posBalance >= 66) {
                                            if (negBalance < 39) {
                                                return 1;
                                            }
                                            else if (negBalance >= 39) {
                                                return 2;
                                            }
                                        }
                                    }
                                }
                                else if (diffAFD >= 28) {
                                    if (ssDCval < 123) {
                                        if (Cs < 1009) {
                                            return 2;
                                        }
                                        else if (Cs >= 1009) {
                                            return 4;
                                        }
                                    }
                                    else if (ssDCval >= 123) {
                                        return 8;
                                    }
                                }
                            }
                            else if (Rs >= 2534) {
                                return 2;
                            }
                        }
                    }
                }
                else if (SC >= 3668) {
                    if (CsDiff < 9) {
                        if (TSC < 9) {
                            if (ssDCval < 95) {
                                if (ssDCval < 89) {
                                    if (mu_mv_mag_sq < 108) {
                                        if (negBalance < 39) {
                                            if (Rs < 2016) {
                                                return 1;
                                            }
                                            else if (Rs >= 2016) {
                                                return 8;
                                            }
                                        }
                                        else if (negBalance >= 39) {
                                            if (tcor < 90) {
                                                return 8;
                                            }
                                            else if (tcor >= 90) {
                                                if (negBalance < 80) {
                                                    return 4;
                                                }
                                                else if (negBalance >= 80) {
                                                    return 8;
                                                }
                                            }
                                        }
                                    }
                                    else if (mu_mv_mag_sq >= 108) {
                                        if (SC < 3734) {
                                            return 4;
                                        }
                                        else if (SC >= 3734) {
                                            if (mu_mv_mag_sq < 112) {
                                                if (tcor < 91) {
                                                    return 8;
                                                }
                                                else if (tcor >= 91) {
                                                    return 4;
                                                }
                                            }
                                            else if (mu_mv_mag_sq >= 112) {
                                                if (refDCval < 81) {
                                                    if (RsCsDiff < 121) {
                                                        return 8;
                                                    }
                                                    else if (RsCsDiff >= 121) {
                                                        if (Rs < 4795) {
                                                            return 8;
                                                        }
                                                        else if (Rs >= 4795) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                                else if (refDCval >= 81) {
                                                    return 8;
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (ssDCval >= 89) {
                                    if (SC < 5726) {
                                        if (posBalance < 46) {
                                            return 1;
                                        }
                                        else if (posBalance >= 46) {
                                            if (ssDCval < 93) {
                                                if (tcor < 95) {
                                                    return 8;
                                                }
                                                else if (tcor >= 95) {
                                                    if (Rs < 2365) {
                                                        return 2;
                                                    }
                                                    else if (Rs >= 2365) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                            else if (ssDCval >= 93) {
                                                if (negBalance < 48) {
                                                    if (TSC < 8) {
                                                        return 8;
                                                    }
                                                    else if (TSC >= 8) {
                                                        return 4;
                                                    }
                                                }
                                                else if (negBalance >= 48) {
                                                    return 2;
                                                }
                                            }
                                        }
                                    }
                                    else if (SC >= 5726) {
                                        if (Cs < 4165) {
                                            return 2;
                                        }
                                        else if (Cs >= 4165) {
                                            return 8;
                                        }
                                    }
                                }
                            }
                            else if (ssDCval >= 95) {
                                if (negBalance < 53) {
                                    if (SC < 4566) {
                                        if (ssDCval < 111) {
                                            return 4;
                                        }
                                        else if (ssDCval >= 111) {
                                            if (posBalance < 56) {
                                                return 4;
                                            }
                                            else if (posBalance >= 56) {
                                                if (tcor < 95) {
                                                    return 8;
                                                }
                                                else if (tcor >= 95) {
                                                    return 4;
                                                }
                                            }
                                        }
                                    }
                                    else if (SC >= 4566) {
                                        if (mu_mv_mag_sq < 104) {
                                            if (Cs < 2535) {
                                                if (MVDiff < 12) {
                                                    return 8;
                                                }
                                                else if (MVDiff >= 12) {
                                                    return 4;
                                                }
                                            }
                                            else if (Cs >= 2535) {
                                                if (Rs < 4657) {
                                                    if (AFD < 8) {
                                                        return 8;
                                                    }
                                                    else if (AFD >= 8) {
                                                        if (ssDCval < 105) {
                                                            if (Rs < 3644) {
                                                                return 2;
                                                            }
                                                            else if (Rs >= 3644) {
                                                                if (Rs < 3809) {
                                                                    return 8;
                                                                }
                                                                else if (Rs >= 3809) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                        else if (ssDCval >= 105) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                                else if (Rs >= 4657) {
                                                    if (CsDiff < 6) {
                                                        return 2;
                                                    }
                                                    else if (CsDiff >= 6) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                        }
                                        else if (mu_mv_mag_sq >= 104) {
                                            if (posBalance < 62) {
                                                if (mu_mv_mag_sq < 118) {
                                                    if (ssDCval < 116) {
                                                        if (negBalance < 43) {
                                                            return 8;
                                                        }
                                                        else if (negBalance >= 43) {
                                                            if (Cs < 1983) {
                                                                return 8;
                                                            }
                                                            else if (Cs >= 1983) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                    else if (ssDCval >= 116) {
                                                        if (ssDCval < 131) {
                                                            return 8;
                                                        }
                                                        else if (ssDCval >= 131) {
                                                            if (Cs < 2638) {
                                                                return 8;
                                                            }
                                                            else if (Cs >= 2638) {
                                                                return 1;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (mu_mv_mag_sq >= 118) {
                                                    return 8;
                                                }
                                            }
                                            else if (posBalance >= 62) {
                                                if (Rs < 4412) {
                                                    return 8;
                                                }
                                                else if (Rs >= 4412) {
                                                    return 4;
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (negBalance >= 53) {
                                    if (Rs < 3947) {
                                        return 2;
                                    }
                                    else if (Rs >= 3947) {
                                        return 4;
                                    }
                                }
                            }
                        }
                        else if (TSC >= 9) {
                            if (Rs < 4482) {
                                if (AFD < 20) {
                                    if (SC < 6789) {
                                        if (SC < 5566) {
                                            if (RsCsDiff < 129) {
                                                if (mu_mv_mag_sq < 99) {
                                                    if (MVDiff < 6) {
                                                        if (diffRsCsdiff < -47) {
                                                            return 8;
                                                        }
                                                        else if (diffRsCsdiff >= -47) {
                                                            if (gchDC < 1) {
                                                                return 4;
                                                            }
                                                            else if (gchDC >= 1) {
                                                                if (Rs < 2450) {
                                                                    return 4;
                                                                }
                                                                else if (Rs >= 2450) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (MVDiff >= 6) {
                                                        return 8;
                                                    }
                                                }
                                                else if (mu_mv_mag_sq >= 99) {
                                                    if (tcor < 94) {
                                                        return 8;
                                                    }
                                                    else if (tcor >= 94) {
                                                        return 1;
                                                    }
                                                }
                                            }
                                            else if (RsCsDiff >= 129) {
                                                return 4;
                                            }
                                        }
                                        else if (SC >= 5566) {
                                            if (refDCval < 93) {
                                                return 2;
                                            }
                                            else if (refDCval >= 93) {
                                                return 8;
                                            }
                                        }
                                    }
                                    else if (SC >= 6789) {
                                        return 2;
                                    }
                                }
                                else if (AFD >= 20) {
                                    if (diffMVdiffVal < 62) {
                                        return 2;
                                    }
                                    else if (diffMVdiffVal >= 62) {
                                        return 1;
                                    }
                                }
                            }
                            else if (Rs >= 4482) {
                                if (posBalance < 58) {
                                    return 2;
                                }
                                else if (posBalance >= 58) {
                                    return 4;
                                }
                            }
                        }
                    }
                    else if (CsDiff >= 9) {
                        if (Cs < 3374) {
                            if (RsDiff < 10) {
                                if (SC < 6816) {
                                    if (Rs < 3643) {
                                        return 8;
                                    }
                                    else if (Rs >= 3643) {
                                        return 4;
                                    }
                                }
                                else if (SC >= 6816) {
                                    return 2;
                                }
                            }
                            else if (RsDiff >= 10) {
                                if (AFD < 37) {
                                    if (refDCval < 122) {
                                        if (Rs < 3967) {
                                            if (CsDiff < 12) {
                                                return 8;
                                            }
                                            else if (CsDiff >= 12) {
                                                return 4;
                                            }
                                        }
                                        else if (Rs >= 3967) {
                                            return 4;
                                        }
                                    }
                                    else if (refDCval >= 122) {
                                        if (refDCval < 134) {
                                            return 8;
                                        }
                                        else if (refDCval >= 134) {
                                            if (SC < 5660) {
                                                return 4;
                                            }
                                            else if (SC >= 5660) {
                                                return 1;
                                            }
                                        }
                                    }
                                }
                                else if (AFD >= 37) {
                                    if (AFD < 49) {
                                        return 2;
                                    }
                                    else if (AFD >= 49) {
                                        return 4;
                                    }
                                }
                            }
                        }
                        else if (Cs >= 3374) {
                            if (negBalance < 62) {
                                if (mu_mv_mag_sq < 122) {
                                    if (TSC < 10) {
                                        if (negBalance < 43) {
                                            if (Cs < 3996) {
                                                return 2;
                                            }
                                            else if (Cs >= 3996) {
                                                return 8;
                                            }
                                        }
                                        else if (negBalance >= 43) {
                                            return 8;
                                        }
                                    }
                                    else if (TSC >= 10) {
                                        if (ssDCval < 137) {
                                            return 8;
                                        }
                                        else if (ssDCval >= 137) {
                                            if (Rs < 4448) {
                                                return 1;
                                            }
                                            else if (Rs >= 4448) {
                                                return 8;
                                            }
                                        }
                                    }
                                }
                                else if (mu_mv_mag_sq >= 122) {
                                    if (posBalance < 53) {
                                        if (Rs < 8074) {
                                            return 8;
                                        }
                                        else if (Rs >= 8074) {
                                            return 2;
                                        }
                                    }
                                    else if (posBalance >= 53) {
                                        if (AFD < 8) {
                                            if (Rs < 4962) {
                                                if (SC < 8158) {
                                                    return 8;
                                                }
                                                else if (SC >= 8158) {
                                                    if (gchDC < 1) {
                                                        if (negBalance < 47) {
                                                            if (Cs < 5051) {
                                                                return 4;
                                                            }
                                                            else if (Cs >= 5051) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (negBalance >= 47) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (gchDC >= 1) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                            else if (Rs >= 4962) {
                                                return 4;
                                            }
                                        }
                                        else if (AFD >= 8) {
                                            if (SC < 10925) {
                                                if (Rs < 6024) {
                                                    if (gchDC < 1) {
                                                        return 8;
                                                    }
                                                    else if (gchDC >= 1) {
                                                        if (ssDCval < 107) {
                                                            return 8;
                                                        }
                                                        else if (ssDCval >= 107) {
                                                            if (ssDCval < 121) {
                                                                return 4;
                                                            }
                                                            else if (ssDCval >= 121) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (Rs >= 6024) {
                                                    return 4;
                                                }
                                            }
                                            else if (SC >= 10925) {
                                                if (refDCval < 106) {
                                                    return 8;
                                                }
                                                else if (refDCval >= 106) {
                                                    if (Cs < 6135) {
                                                        return 4;
                                                    }
                                                    else if (Cs >= 6135) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                            else if (negBalance >= 62) {
                                if (mcTcor < 78) {
                                    if (mu_mv_mag_sq < 111) {
                                        return 4;
                                    }
                                    else if (mu_mv_mag_sq >= 111) {
                                        if (Rs < 7908) {
                                            return 2;
                                        }
                                        else if (Rs >= 7908) {
                                            if (Cs < 8191) {
                                                return 4;
                                            }
                                            else if (Cs >= 8191) {
                                                if (Cs < 8345) {
                                                    return 2;
                                                }
                                                else if (Cs >= 8345) {
                                                    return 4;
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (mcTcor >= 78) {
                                    if (Rs < 7420) {
                                        return 8;
                                    }
                                    else if (Rs >= 7420) {
                                        return 2;
                                    }
                                }
                            }
                        }
                    }
                }
            }
        }
        return 4;
    }
    mfxU16 AGOPSelect17(mfxU32 Rs, mfxU32 CsDiff, mfxU32 TSC, mfxI32 diffAFD, mfxU32 Scindex, mfxU32 RsDiff, mfxI32 diffMVdiffVal, mfxU32 RsCsDiff, mfxU32 TSCindex, mfxI16 tcor, mfxU32 refDCval, mfxU32 MVDiff, mfxU32 AFD, mfxI32 diffTSC, mfxI32 diffRsCsdiff, mfxU32 posBalance, mfxU32 SC, mfxU32 mu_mv_mag_sq, mfxU32 negBalance, mfxI16 mcTcor, mfxU32 gchDC, mfxU32 Cs, mfxU32 ssDCval) {
        if (tcor < 95) {
            if (MVDiff < 6) {
                if (AFD < 6) {
                    if (SC < 1445) {
                        if (negBalance < 45) {
                            if (refDCval < 54) {
                                if (posBalance < 57) {
                                    if (Cs < 766) {
                                        return 4;
                                    }
                                    else if (Cs >= 766) {
                                        return 8;
                                    }
                                }
                                else if (posBalance >= 57) {
                                    if (SC < 1022) {
                                        return 4;
                                    }
                                    else if (SC >= 1022) {
                                        return 8;
                                    }
                                }
                            }
                            else if (refDCval >= 54) {
                                if (mcTcor < 96) {
                                    return 8;
                                }
                                else if (mcTcor >= 96) {
                                    if (RsDiff < 2) {
                                        return 4;
                                    }
                                    else if (RsDiff >= 2) {
                                        return 8;
                                    }
                                }
                            }
                        }
                        else if (negBalance >= 45) {
                            return 2;
                        }
                    }
                    else if (SC >= 1445) {
                        if (Cs < 4454) {
                            if (diffMVdiffVal < -3) {
                                if (diffMVdiffVal < -7) {
                                    if (AFD < 1) {
                                        return 2;
                                    }
                                    else if (AFD >= 1) {
                                        return 8;
                                    }
                                }
                                else if (diffMVdiffVal >= -7) {
                                    if (CsDiff < 3) {
                                        if (Cs < 833) {
                                            if (TSC < 4) {
                                                if (Rs < 926) {
                                                    return 4;
                                                }
                                                else if (Rs >= 926) {
                                                    return 1;
                                                }
                                            }
                                            else if (TSC >= 4) {
                                                return 8;
                                            }
                                        }
                                        else if (Cs >= 833) {
                                            if (Rs < 1374) {
                                                return 1;
                                            }
                                            else if (Rs >= 1374) {
                                                if (Rs < 1426) {
                                                    return 4;
                                                }
                                                else if (Rs >= 1426) {
                                                    if (Rs < 2172) {
                                                        return 1;
                                                    }
                                                    else if (Rs >= 2172) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (CsDiff >= 3) {
                                        if (diffRsCsdiff < -34) {
                                            return 4;
                                        }
                                        else if (diffRsCsdiff >= -34) {
                                            return 8;
                                        }
                                    }
                                }
                            }
                            else if (diffMVdiffVal >= -3) {
                                if (MVDiff < 3) {
                                    if (Cs < 1839) {
                                        if (Scindex < 5) {
                                            if (Rs < 1458) {
                                                if (Rs < 1307) {
                                                    if (negBalance < 37) {
                                                        return 8;
                                                    }
                                                    else if (negBalance >= 37) {
                                                        if (Cs < 728) {
                                                            return 8;
                                                        }
                                                        else if (Cs >= 728) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                                else if (Rs >= 1307) {
                                                    if (diffTSC < 1) {
                                                        if (mu_mv_mag_sq < 118) {
                                                            if (Rs < 1367) {
                                                                return 8;
                                                            }
                                                            else if (Rs >= 1367) {
                                                                if (TSC < 3) {
                                                                    return 8;
                                                                }
                                                                else if (TSC >= 3) {
                                                                    return 1;
                                                                }
                                                            }
                                                        }
                                                        else if (mu_mv_mag_sq >= 118) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (diffTSC >= 1) {
                                                        return 1;
                                                    }
                                                }
                                            }
                                            else if (Rs >= 1458) {
                                                if (diffRsCsdiff < -18) {
                                                    if (diffMVdiffVal < 1) {
                                                        return 8;
                                                    }
                                                    else if (diffMVdiffVal >= 1) {
                                                        if (Cs < 1649) {
                                                            return 8;
                                                        }
                                                        else if (Cs >= 1649) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                                else if (diffRsCsdiff >= -18) {
                                                    return 8;
                                                }
                                            }
                                        }
                                        else if (Scindex >= 5) {
                                            if (Rs < 2806) {
                                                return 8;
                                            }
                                            else if (Rs >= 2806) {
                                                return 4;
                                            }
                                        }
                                    }
                                    else if (Cs >= 1839) {
                                        if (ssDCval < 23) {
                                            if (Rs < 2707) {
                                                return 1;
                                            }
                                            else if (Rs >= 2707) {
                                                if (Rs < 3755) {
                                                    return 8;
                                                }
                                                else if (Rs >= 3755) {
                                                    if (Rs < 3770) {
                                                        return 1;
                                                    }
                                                    else if (Rs >= 3770) {
                                                        if (Cs < 3951) {
                                                            return 1;
                                                        }
                                                        else if (Cs >= 3951) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (ssDCval >= 23) {
                                            if (mu_mv_mag_sq < 112) {
                                                if (CsDiff < 5) {
                                                    if (gchDC < 1) {
                                                        return 8;
                                                    }
                                                    else if (gchDC >= 1) {
                                                        return 4;
                                                    }
                                                }
                                                else if (CsDiff >= 5) {
                                                    if (Rs < 2243) {
                                                        return 1;
                                                    }
                                                    else if (Rs >= 2243) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                            else if (mu_mv_mag_sq >= 112) {
                                                if (mcTcor < 96) {
                                                    if (refDCval < 116) {
                                                        if (ssDCval < 51) {
                                                            if (Cs < 2166) {
                                                                return 4;
                                                            }
                                                            else if (Cs >= 2166) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (ssDCval >= 51) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (refDCval >= 116) {
                                                        if (SC < 4984) {
                                                            if (Rs < 2333) {
                                                                return 1;
                                                            }
                                                            else if (Rs >= 2333) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (SC >= 4984) {
                                                            if (mu_mv_mag_sq < 120) {
                                                                return 4;
                                                            }
                                                            else if (mu_mv_mag_sq >= 120) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (mcTcor >= 96) {
                                                    if (CsDiff < 2) {
                                                        if (mu_mv_mag_sq < 117) {
                                                            if (gchDC < 1) {
                                                                return 8;
                                                            }
                                                            else if (gchDC >= 1) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (mu_mv_mag_sq >= 117) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (CsDiff >= 2) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (MVDiff >= 3) {
                                    if (mu_mv_mag_sq < 83) {
                                        if (ssDCval < 54) {
                                            if (ssDCval < 50) {
                                                return 4;
                                            }
                                            else if (ssDCval >= 50) {
                                                if (tcor < 93) {
                                                    if (diffMVdiffVal < 1) {
                                                        if (Cs < 918) {
                                                            if (SC < 2232) {
                                                                return 1;
                                                            }
                                                            else if (SC >= 2232) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (Cs >= 918) {
                                                            return 1;
                                                        }
                                                    }
                                                    else if (diffMVdiffVal >= 1) {
                                                        if (ssDCval < 52) {
                                                            return 1;
                                                        }
                                                        else if (ssDCval >= 52) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                                else if (tcor >= 93) {
                                                    if (diffRsCsdiff < -4) {
                                                        if (Cs < 948) {
                                                            return 1;
                                                        }
                                                        else if (Cs >= 948) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (diffRsCsdiff >= -4) {
                                                        return 1;
                                                    }
                                                }
                                            }
                                        }
                                        else if (ssDCval >= 54) {
                                            if (diffRsCsdiff < -2) {
                                                return 4;
                                            }
                                            else if (diffRsCsdiff >= -2) {
                                                return 8;
                                            }
                                        }
                                    }
                                    else if (mu_mv_mag_sq >= 83) {
                                        if (posBalance < 14) {
                                            if (Rs < 1307) {
                                                return 8;
                                            }
                                            else if (Rs >= 1307) {
                                                if (Rs < 1836) {
                                                    return 1;
                                                }
                                                else if (Rs >= 1836) {
                                                    if (refDCval < 117) {
                                                        return 8;
                                                    }
                                                    else if (refDCval >= 117) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                        }
                                        else if (posBalance >= 14) {
                                            return 8;
                                        }
                                    }
                                }
                            }
                        }
                        else if (Cs >= 4454) {
                            if (Cs < 4528) {
                                if (AFD < 1) {
                                    return 1;
                                }
                                else if (AFD >= 1) {
                                    return 8;
                                }
                            }
                            else if (Cs >= 4528) {
                                return 8;
                            }
                        }
                    }
                }
                else if (AFD >= 6) {
                    if (CsDiff < 10) {
                        if (Rs < 4829) {
                            if (mu_mv_mag_sq < 92) {
                                if (refDCval < 100) {
                                    if (TSC < 7) {
                                        if (negBalance < 48) {
                                            if (ssDCval < 72) {
                                                if (posBalance < 34) {
                                                    return 1;
                                                }
                                                else if (posBalance >= 34) {
                                                    if (Rs < 1377) {
                                                        if (ssDCval < 53) {
                                                            if (RsDiff < 3) {
                                                                if (Cs < 759) {
                                                                    return 1;
                                                                }
                                                                else if (Cs >= 759) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (RsDiff >= 3) {
                                                                if (ssDCval < 47) {
                                                                    return 4;
                                                                }
                                                                else if (ssDCval >= 47) {
                                                                    return 1;
                                                                }
                                                            }
                                                        }
                                                        else if (ssDCval >= 53) {
                                                            if (RsDiff < 3) {
                                                                if (RsDiff < 1) {
                                                                    return 8;
                                                                }
                                                                else if (RsDiff >= 1) {
                                                                    return 1;
                                                                }
                                                            }
                                                            else if (RsDiff >= 3) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                    else if (Rs >= 1377) {
                                                        if (RsCsDiff < 19) {
                                                            if (refDCval < 55) {
                                                                if (Rs < 3573) {
                                                                    if (posBalance < 58) {
                                                                        return 8;
                                                                    }
                                                                    else if (posBalance >= 58) {
                                                                        return 1;
                                                                    }
                                                                }
                                                                else if (Rs >= 3573) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (refDCval >= 55) {
                                                                if (Cs < 983) {
                                                                    return 1;
                                                                }
                                                                else if (Cs >= 983) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                        else if (RsCsDiff >= 19) {
                                                            if (TSC < 6) {
                                                                if (RsCsDiff < 141) {
                                                                    return 4;
                                                                }
                                                                else if (RsCsDiff >= 141) {
                                                                    if (SC < 3797) {
                                                                        return 1;
                                                                    }
                                                                    else if (SC >= 3797) {
                                                                        return 4;
                                                                    }
                                                                }
                                                            }
                                                            else if (TSC >= 6) {
                                                                if (ssDCval < 61) {
                                                                    if (Rs < 2142) {
                                                                        return 4;
                                                                    }
                                                                    else if (Rs >= 2142) {
                                                                        return 8;
                                                                    }
                                                                }
                                                                else if (ssDCval >= 61) {
                                                                    return 1;
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                            else if (ssDCval >= 72) {
                                                if (SC < 3635) {
                                                    if (MVDiff < 5) {
                                                        if (refDCval < 82) {
                                                            if (AFD < 8) {
                                                                return 4;
                                                            }
                                                            else if (AFD >= 8) {
                                                                return 2;
                                                            }
                                                        }
                                                        else if (refDCval >= 82) {
                                                            if (diffRsCsdiff < -11) {
                                                                return 2;
                                                            }
                                                            else if (diffRsCsdiff >= -11) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                    else if (MVDiff >= 5) {
                                                        return 4;
                                                    }
                                                }
                                                else if (SC >= 3635) {
                                                    if (Cs < 1190) {
                                                        return 1;
                                                    }
                                                    else if (Cs >= 1190) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                        }
                                        else if (negBalance >= 48) {
                                            if (ssDCval < 90) {
                                                if (Rs < 2061) {
                                                    return 2;
                                                }
                                                else if (Rs >= 2061) {
                                                    return 4;
                                                }
                                            }
                                            else if (ssDCval >= 90) {
                                                if (TSC < 4) {
                                                    return 2;
                                                }
                                                else if (TSC >= 4) {
                                                    return 4;
                                                }
                                            }
                                        }
                                    }
                                    else if (TSC >= 7) {
                                        if (Cs < 2596) {
                                            if (MVDiff < 2) {
                                                if (negBalance < 47) {
                                                    if (Cs < 1872) {
                                                        return 2;
                                                    }
                                                    else if (Cs >= 1872) {
                                                        return 4;
                                                    }
                                                }
                                                else if (negBalance >= 47) {
                                                    return 2;
                                                }
                                            }
                                            else if (MVDiff >= 2) {
                                                if (negBalance < 46) {
                                                    if (MVDiff < 4) {
                                                        return 8;
                                                    }
                                                    else if (MVDiff >= 4) {
                                                        if (Rs < 3119) {
                                                            return 1;
                                                        }
                                                        else if (Rs >= 3119) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                                else if (negBalance >= 46) {
                                                    if (Rs < 2375) {
                                                        if (posBalance < 61) {
                                                            if (ssDCval < 81) {
                                                                return 4;
                                                            }
                                                            else if (ssDCval >= 81) {
                                                                if (RsDiff < 8) {
                                                                    if (Rs < 2051) {
                                                                        if (tcor < 89) {
                                                                            if (Rs < 1849) {
                                                                                return 2;
                                                                            }
                                                                            else if (Rs >= 1849) {
                                                                                return 8;
                                                                            }
                                                                        }
                                                                        else if (tcor >= 89) {
                                                                            return 2;
                                                                        }
                                                                    }
                                                                    else if (Rs >= 2051) {
                                                                        return 2;
                                                                    }
                                                                }
                                                                else if (RsDiff >= 8) {
                                                                    if (mcTcor < 95) {
                                                                        if (Rs < 2369) {
                                                                            return 4;
                                                                        }
                                                                        else if (Rs >= 2369) {
                                                                            return 2;
                                                                        }
                                                                    }
                                                                    else if (mcTcor >= 95) {
                                                                        return 2;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (posBalance >= 61) {
                                                            if (ssDCval < 89) {
                                                                if (ssDCval < 86) {
                                                                    return 2;
                                                                }
                                                                else if (ssDCval >= 86) {
                                                                    if (SC < 3107) {
                                                                        return 1;
                                                                    }
                                                                    else if (SC >= 3107) {
                                                                        return 8;
                                                                    }
                                                                }
                                                            }
                                                            else if (ssDCval >= 89) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                    else if (Rs >= 2375) {
                                                        if (Cs < 1761) {
                                                            return 1;
                                                        }
                                                        else if (Cs >= 1761) {
                                                            if (ssDCval < 95) {
                                                                if (TSC < 10) {
                                                                    if (diffMVdiffVal < -2) {
                                                                        if (Rs < 2739) {
                                                                            if (RsDiff < 7) {
                                                                                return 4;
                                                                            }
                                                                            else if (RsDiff >= 7) {
                                                                                return 2;
                                                                            }
                                                                        }
                                                                        else if (Rs >= 2739) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                    else if (diffMVdiffVal >= -2) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (TSC >= 10) {
                                                                    if (SC < 5139) {
                                                                        return 2;
                                                                    }
                                                                    else if (SC >= 5139) {
                                                                        if (Cs < 2403) {
                                                                            return 4;
                                                                        }
                                                                        else if (Cs >= 2403) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                            else if (ssDCval >= 95) {
                                                                return 2;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (Cs >= 2596) {
                                            if (tcor < 90) {
                                                return 8;
                                            }
                                            else if (tcor >= 90) {
                                                if (AFD < 11) {
                                                    return 2;
                                                }
                                                else if (AFD >= 11) {
                                                    return 4;
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (refDCval >= 100) {
                                    if (posBalance < 58) {
                                        if (Cs < 1263) {
                                            if (mu_mv_mag_sq < 45) {
                                                if (SC < 2677) {
                                                    if (Cs < 754) {
                                                        return 2;
                                                    }
                                                    else if (Cs >= 754) {
                                                        return 4;
                                                    }
                                                }
                                                else if (SC >= 2677) {
                                                    return 2;
                                                }
                                            }
                                            else if (mu_mv_mag_sq >= 45) {
                                                if (SC < 1726) {
                                                    return 4;
                                                }
                                                else if (SC >= 1726) {
                                                    return 8;
                                                }
                                            }
                                        }
                                        else if (Cs >= 1263) {
                                            if (ssDCval < 120) {
                                                if (Rs < 3081) {
                                                    if (ssDCval < 100) {
                                                        return 1;
                                                    }
                                                    else if (ssDCval >= 100) {
                                                        return 8;
                                                    }
                                                }
                                                else if (Rs >= 3081) {
                                                    if (ssDCval < 110) {
                                                        if (AFD < 7) {
                                                            return 8;
                                                        }
                                                        else if (AFD >= 7) {
                                                            if (MVDiff < 5) {
                                                                if (Cs < 2732) {
                                                                    return 2;
                                                                }
                                                                else if (Cs >= 2732) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (MVDiff >= 5) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                    else if (ssDCval >= 110) {
                                                        if (Cs < 2268) {
                                                            return 2;
                                                        }
                                                        else if (Cs >= 2268) {
                                                            return 1;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (ssDCval >= 120) {
                                                if (posBalance < 49) {
                                                    if (negBalance < 38) {
                                                        if (SC < 7084) {
                                                            return 4;
                                                        }
                                                        else if (SC >= 7084) {
                                                            return 2;
                                                        }
                                                    }
                                                    else if (negBalance >= 38) {
                                                        return 8;
                                                    }
                                                }
                                                else if (posBalance >= 49) {
                                                    if (AFD < 11) {
                                                        if (Rs < 2216) {
                                                            return 1;
                                                        }
                                                        else if (Rs >= 2216) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (AFD >= 11) {
                                                        return 1;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (posBalance >= 58) {
                                        if (mu_mv_mag_sq < 50) {
                                            if (AFD < 12) {
                                                if (Rs < 1065) {
                                                    if (Cs < 624) {
                                                        return 2;
                                                    }
                                                    else if (Cs >= 624) {
                                                        return 1;
                                                    }
                                                }
                                                else if (Rs >= 1065) {
                                                    if (Cs < 1932) {
                                                        return 4;
                                                    }
                                                    else if (Cs >= 1932) {
                                                        if (TSC < 9) {
                                                            if (posBalance < 59) {
                                                                return 4;
                                                            }
                                                            else if (posBalance >= 59) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (TSC >= 9) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (AFD >= 12) {
                                                if (CsDiff < 6) {
                                                    if (ssDCval < 123) {
                                                        return 4;
                                                    }
                                                    else if (ssDCval >= 123) {
                                                        if (Rs < 2438) {
                                                            return 1;
                                                        }
                                                        else if (Rs >= 2438) {
                                                            return 2;
                                                        }
                                                    }
                                                }
                                                else if (CsDiff >= 6) {
                                                    if (Rs < 4032) {
                                                        if (negBalance < 46) {
                                                            return 8;
                                                        }
                                                        else if (negBalance >= 46) {
                                                            return 1;
                                                        }
                                                    }
                                                    else if (Rs >= 4032) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                        }
                                        else if (mu_mv_mag_sq >= 50) {
                                            if (ssDCval < 110) {
                                                return 4;
                                            }
                                            else if (ssDCval >= 110) {
                                                if (mu_mv_mag_sq < 79) {
                                                    if (Cs < 2490) {
                                                        if (SC < 4739) {
                                                            if (RsDiff < 6) {
                                                                return 4;
                                                            }
                                                            else if (RsDiff >= 6) {
                                                                if (SC < 4392) {
                                                                    if (mu_mv_mag_sq < 60) {
                                                                        if (Rs < 2296) {
                                                                            return 4;
                                                                        }
                                                                        else if (Rs >= 2296) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                    else if (mu_mv_mag_sq >= 60) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (SC >= 4392) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                        else if (SC >= 4739) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (Cs >= 2490) {
                                                        return 8;
                                                    }
                                                }
                                                else if (mu_mv_mag_sq >= 79) {
                                                    return 8;
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                            else if (mu_mv_mag_sq >= 92) {
                                if (CsDiff < 4) {
                                    if (posBalance < 58) {
                                        if (RsDiff < 2) {
                                            return 4;
                                        }
                                        else if (RsDiff >= 2) {
                                            return 8;
                                        }
                                    }
                                    else if (posBalance >= 58) {
                                        if (negBalance < 43) {
                                            return 8;
                                        }
                                        else if (negBalance >= 43) {
                                            return 2;
                                        }
                                    }
                                }
                                else if (CsDiff >= 4) {
                                    if (AFD < 7) {
                                        if (Rs < 1843) {
                                            if (RsCsDiff < 33) {
                                                if (TSC < 6) {
                                                    return 8;
                                                }
                                                else if (TSC >= 6) {
                                                    if (ssDCval < 72) {
                                                        return 2;
                                                    }
                                                    else if (ssDCval >= 72) {
                                                        if (MVDiff < 1) {
                                                            if (Rs < 1755) {
                                                                return 8;
                                                            }
                                                            else if (Rs >= 1755) {
                                                                return 2;
                                                            }
                                                        }
                                                        else if (MVDiff >= 1) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (RsCsDiff >= 33) {
                                                return 8;
                                            }
                                        }
                                        else if (Rs >= 1843) {
                                            if (refDCval < 78) {
                                                if (mcTcor < 95) {
                                                    if (Cs < 1586) {
                                                        if (RsCsDiff < 37) {
                                                            return 4;
                                                        }
                                                        else if (RsCsDiff >= 37) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (Cs >= 1586) {
                                                        if (TSC < 5) {
                                                            return 4;
                                                        }
                                                        else if (TSC >= 5) {
                                                            if (ssDCval < 74) {
                                                                return 8;
                                                            }
                                                            else if (ssDCval >= 74) {
                                                                if (posBalance < 59) {
                                                                    return 8;
                                                                }
                                                                else if (posBalance >= 59) {
                                                                    return 2;
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (mcTcor >= 95) {
                                                    if (SC < 3434) {
                                                        if (Rs < 1879) {
                                                            return 4;
                                                        }
                                                        else if (Rs >= 1879) {
                                                            return 2;
                                                        }
                                                    }
                                                    else if (SC >= 3434) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                            else if (refDCval >= 78) {
                                                if (CsDiff < 8) {
                                                    if (mu_mv_mag_sq < 118) {
                                                        if (Rs < 2245) {
                                                            return 8;
                                                        }
                                                        else if (Rs >= 2245) {
                                                            if (CsDiff < 7) {
                                                                return 4;
                                                            }
                                                            else if (CsDiff >= 7) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                    else if (mu_mv_mag_sq >= 118) {
                                                        if (diffRsCsdiff < -30) {
                                                            if (diffRsCsdiff < -31) {
                                                                return 8;
                                                            }
                                                            else if (diffRsCsdiff >= -31) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (diffRsCsdiff >= -30) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                                else if (CsDiff >= 8) {
                                                    if (RsCsDiff < 137) {
                                                        if (SC < 9669) {
                                                            if (negBalance < 45) {
                                                                return 4;
                                                            }
                                                            else if (negBalance >= 45) {
                                                                if (Cs < 4868) {
                                                                    return 8;
                                                                }
                                                                else if (Cs >= 4868) {
                                                                    if (Rs < 4703) {
                                                                        return 4;
                                                                    }
                                                                    else if (Rs >= 4703) {
                                                                        return 8;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (SC >= 9669) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (RsCsDiff >= 137) {
                                                        if (Rs < 3562) {
                                                            return 1;
                                                        }
                                                        else if (Rs >= 3562) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (AFD >= 7) {
                                        if (refDCval < 97) {
                                            if (mcTcor < 91) {
                                                if (RsCsDiff < 37) {
                                                    if (ssDCval < 71) {
                                                        return 2;
                                                    }
                                                    else if (ssDCval >= 71) {
                                                        if (Rs < 1180) {
                                                            return 4;
                                                        }
                                                        else if (Rs >= 1180) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                                else if (RsCsDiff >= 37) {
                                                    if (ssDCval < 93) {
                                                        if (CsDiff < 5) {
                                                            if (MVDiff < 1) {
                                                                return 2;
                                                            }
                                                            else if (MVDiff >= 1) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (CsDiff >= 5) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (ssDCval >= 93) {
                                                        if (AFD < 8) {
                                                            return 8;
                                                        }
                                                        else if (AFD >= 8) {
                                                            return 2;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (mcTcor >= 91) {
                                                if (negBalance < 43) {
                                                    if (RsCsDiff < 57) {
                                                        if (SC < 4796) {
                                                            return 4;
                                                        }
                                                        else if (SC >= 4796) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (RsCsDiff >= 57) {
                                                        if (mu_mv_mag_sq < 103) {
                                                            if (refDCval < 85) {
                                                                return 8;
                                                            }
                                                            else if (refDCval >= 85) {
                                                                return 1;
                                                            }
                                                        }
                                                        else if (mu_mv_mag_sq >= 103) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                                else if (negBalance >= 43) {
                                                    if (mu_mv_mag_sq < 113) {
                                                        if (Cs < 1979) {
                                                            return 2;
                                                        }
                                                        else if (Cs >= 1979) {
                                                            if (ssDCval < 88) {
                                                                if (diffAFD < -1) {
                                                                    return 8;
                                                                }
                                                                else if (diffAFD >= -1) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (ssDCval >= 88) {
                                                                return 2;
                                                            }
                                                        }
                                                    }
                                                    else if (mu_mv_mag_sq >= 113) {
                                                        if (refDCval < 88) {
                                                            if (ssDCval < 76) {
                                                                if (mcTcor < 93) {
                                                                    if (Rs < 1543) {
                                                                        return 2;
                                                                    }
                                                                    else if (Rs >= 1543) {
                                                                        return 8;
                                                                    }
                                                                }
                                                                else if (mcTcor >= 93) {
                                                                    if (Cs < 1679) {
                                                                        return 2;
                                                                    }
                                                                    else if (Cs >= 1679) {
                                                                        return 4;
                                                                    }
                                                                }
                                                            }
                                                            else if (ssDCval >= 76) {
                                                                if (Rs < 4728) {
                                                                    if (SC < 3409) {
                                                                        return 4;
                                                                    }
                                                                    else if (SC >= 3409) {
                                                                        return 8;
                                                                    }
                                                                }
                                                                else if (Rs >= 4728) {
                                                                    if (Rs < 4775) {
                                                                        return 4;
                                                                    }
                                                                    else if (Rs >= 4775) {
                                                                        return 8;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (refDCval >= 88) {
                                                            if (SC < 7514) {
                                                                return 2;
                                                            }
                                                            else if (SC >= 7514) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (refDCval >= 97) {
                                            if (Rs < 4045) {
                                                if (SC < 4660) {
                                                    if (refDCval < 111) {
                                                        return 4;
                                                    }
                                                    else if (refDCval >= 111) {
                                                        if (mu_mv_mag_sq < 121) {
                                                            return 8;
                                                        }
                                                        else if (mu_mv_mag_sq >= 121) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                                else if (SC >= 4660) {
                                                    if (Rs < 3529) {
                                                        return 8;
                                                    }
                                                    else if (Rs >= 3529) {
                                                        if (diffAFD < 1) {
                                                            if (Rs < 3595) {
                                                                return 4;
                                                            }
                                                            else if (Rs >= 3595) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (diffAFD >= 1) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (Rs >= 4045) {
                                                if (ssDCval < 127) {
                                                    return 4;
                                                }
                                                else if (ssDCval >= 127) {
                                                    return 8;
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                        }
                        else if (Rs >= 4829) {
                            if (mu_mv_mag_sq < 125) {
                                if (Cs < 2841) {
                                    return 4;
                                }
                                else if (Cs >= 2841) {
                                    if (diffTSC < 1) {
                                        return 8;
                                    }
                                    else if (diffTSC >= 1) {
                                        if (Cs < 3327) {
                                            return 2;
                                        }
                                        else if (Cs >= 3327) {
                                            return 8;
                                        }
                                    }
                                }
                            }
                            else if (mu_mv_mag_sq >= 125) {
                                if (negBalance < 52) {
                                    if (Rs < 5251) {
                                        if (Cs < 5126) {
                                            if (mu_mv_mag_sq < 128) {
                                                if (Cs < 5038) {
                                                    if (CsDiff < 8) {
                                                        return 4;
                                                    }
                                                    else if (CsDiff >= 8) {
                                                        return 8;
                                                    }
                                                }
                                                else if (Cs >= 5038) {
                                                    return 4;
                                                }
                                            }
                                            else if (mu_mv_mag_sq >= 128) {
                                                if (diffTSC < -1) {
                                                    if (RsDiff < 9) {
                                                        if (SC < 9881) {
                                                            return 8;
                                                        }
                                                        else if (SC >= 9881) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (RsDiff >= 9) {
                                                        return 8;
                                                    }
                                                }
                                                else if (diffTSC >= -1) {
                                                    if (diffRsCsdiff < -39) {
                                                        return 8;
                                                    }
                                                    else if (diffRsCsdiff >= -39) {
                                                        if (RsCsDiff < 154) {
                                                            return 4;
                                                        }
                                                        else if (RsCsDiff >= 154) {
                                                            if (Cs < 5004) {
                                                                return 4;
                                                            }
                                                            else if (Cs >= 5004) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (Cs >= 5126) {
                                            return 4;
                                        }
                                    }
                                    else if (Rs >= 5251) {
                                        return 8;
                                    }
                                }
                                else if (negBalance >= 52) {
                                    return 4;
                                }
                            }
                        }
                    }
                    else if (CsDiff >= 10) {
                        if (mu_mv_mag_sq < 84) {
                            if (ssDCval < 126) {
                                if (SC < 5058) {
                                    if (refDCval < 89) {
                                        if (RsCsDiff < 165) {
                                            if (Rs < 1786) {
                                                return 2;
                                            }
                                            else if (Rs >= 1786) {
                                                return 1;
                                            }
                                        }
                                        else if (RsCsDiff >= 165) {
                                            return 4;
                                        }
                                    }
                                    else if (refDCval >= 89) {
                                        if (AFD < 13) {
                                            return 8;
                                        }
                                        else if (AFD >= 13) {
                                            if (Rs < 2458) {
                                                return 2;
                                            }
                                            else if (Rs >= 2458) {
                                                return 4;
                                            }
                                        }
                                    }
                                }
                                else if (SC >= 5058) {
                                    if (RsDiff < 11) {
                                        if (ssDCval < 118) {
                                            if (refDCval < 89) {
                                                if (Cs < 2801) {
                                                    return 4;
                                                }
                                                else if (Cs >= 2801) {
                                                    return 8;
                                                }
                                            }
                                            else if (refDCval >= 89) {
                                                return 8;
                                            }
                                        }
                                        else if (ssDCval >= 118) {
                                            return 4;
                                        }
                                    }
                                    else if (RsDiff >= 11) {
                                        if (Cs < 3964) {
                                            if (Rs < 3927) {
                                                if (Cs < 2934) {
                                                    if (diffAFD < 5) {
                                                        return 4;
                                                    }
                                                    else if (diffAFD >= 5) {
                                                        return 8;
                                                    }
                                                }
                                                else if (Cs >= 2934) {
                                                    return 8;
                                                }
                                            }
                                            else if (Rs >= 3927) {
                                                return 4;
                                            }
                                        }
                                        else if (Cs >= 3964) {
                                            if (TSC < 25) {
                                                if (negBalance < 63) {
                                                    if (RsCsDiff < 339) {
                                                        if (TSC < 13) {
                                                            return 8;
                                                        }
                                                        else if (TSC >= 13) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (RsCsDiff >= 339) {
                                                        return 8;
                                                    }
                                                }
                                                else if (negBalance >= 63) {
                                                    if (TSC < 20) {
                                                        return 8;
                                                    }
                                                    else if (TSC >= 20) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                            else if (TSC >= 25) {
                                                if (mu_mv_mag_sq < 64) {
                                                    return 4;
                                                }
                                                else if (mu_mv_mag_sq >= 64) {
                                                    return 8;
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                            else if (ssDCval >= 126) {
                                if (Cs < 3351) {
                                    if (Cs < 1903) {
                                        return 4;
                                    }
                                    else if (Cs >= 1903) {
                                        return 1;
                                    }
                                }
                                else if (Cs >= 3351) {
                                    if (Rs < 4798) {
                                        if (Cs < 4148) {
                                            return 8;
                                        }
                                        else if (Cs >= 4148) {
                                            return 1;
                                        }
                                    }
                                    else if (Rs >= 4798) {
                                        return 8;
                                    }
                                }
                            }
                        }
                        else if (mu_mv_mag_sq >= 84) {
                            if (Cs < 7335) {
                                if (ssDCval < 105) {
                                    if (Cs < 3235) {
                                        if (Rs < 3551) {
                                            return 8;
                                        }
                                        else if (Rs >= 3551) {
                                            return 4;
                                        }
                                    }
                                    else if (Cs >= 3235) {
                                        if (TSCindex < 9) {
                                            if (Cs < 4972) {
                                                return 8;
                                            }
                                            else if (Cs >= 4972) {
                                                return 4;
                                            }
                                        }
                                        else if (TSCindex >= 9) {
                                            return 8;
                                        }
                                    }
                                }
                                else if (ssDCval >= 105) {
                                    if (refDCval < 109) {
                                        if (refDCval < 107) {
                                            if (diffMVdiffVal < -77) {
                                                return 4;
                                            }
                                            else if (diffMVdiffVal >= -77) {
                                                if (ssDCval < 106) {
                                                    if (TSC < 13) {
                                                        return 8;
                                                    }
                                                    else if (TSC >= 13) {
                                                        if (Rs < 4221) {
                                                            return 8;
                                                        }
                                                        else if (Rs >= 4221) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                                else if (ssDCval >= 106) {
                                                    if (Rs < 5789) {
                                                        return 8;
                                                    }
                                                    else if (Rs >= 5789) {
                                                        if (RsDiff < 16) {
                                                            return 4;
                                                        }
                                                        else if (RsDiff >= 16) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (refDCval >= 107) {
                                            if (RsCsDiff < 379) {
                                                return 4;
                                            }
                                            else if (RsCsDiff >= 379) {
                                                if (posBalance < 56) {
                                                    return 8;
                                                }
                                                else if (posBalance >= 56) {
                                                    return 4;
                                                }
                                            }
                                        }
                                    }
                                    else if (refDCval >= 109) {
                                        if (mcTcor < 79) {
                                            return 1;
                                        }
                                        else if (mcTcor >= 79) {
                                            if (TSC < 10) {
                                                if (Rs < 4711) {
                                                    if (SC < 6869) {
                                                        return 8;
                                                    }
                                                    else if (SC >= 6869) {
                                                        return 4;
                                                    }
                                                }
                                                else if (Rs >= 4711) {
                                                    if (refDCval < 115) {
                                                        if (SC < 9470) {
                                                            return 2;
                                                        }
                                                        else if (SC >= 9470) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (refDCval >= 115) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                            else if (TSC >= 10) {
                                                if (SC < 8319) {
                                                    if (Cs < 3563) {
                                                        return 8;
                                                    }
                                                    else if (Cs >= 3563) {
                                                        return 2;
                                                    }
                                                }
                                                else if (SC >= 8319) {
                                                    return 8;
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                            else if (Cs >= 7335) {
                                if (Cs < 8668) {
                                    if (posBalance < 53) {
                                        if (MVDiff < 1) {
                                            if (refDCval < 91) {
                                                return 2;
                                            }
                                            else if (refDCval >= 91) {
                                                if (mcTcor < 76) {
                                                    return 2;
                                                }
                                                else if (mcTcor >= 76) {
                                                    if (CsDiff < 23) {
                                                        return 4;
                                                    }
                                                    else if (CsDiff >= 23) {
                                                        return 2;
                                                    }
                                                }
                                            }
                                        }
                                        else if (MVDiff >= 1) {
                                            return 4;
                                        }
                                    }
                                    else if (posBalance >= 53) {
                                        if (gchDC < 1) {
                                            return 4;
                                        }
                                        else if (gchDC >= 1) {
                                            return 2;
                                        }
                                    }
                                }
                                else if (Cs >= 8668) {
                                    return 8;
                                }
                            }
                        }
                    }
                }
            }
            else if (MVDiff >= 6) {
                if (SC < 7645) {
                    if (mu_mv_mag_sq < 27) {
                        if (CsDiff < 10) {
                            if (Cs < 309) {
                                if (RsDiff < 2) {
                                    if (MVDiff < 43) {
                                        if (refDCval < 90) {
                                            if (diffMVdiffVal < -20) {
                                                if (Rs < 17) {
                                                    return 2;
                                                }
                                                else if (Rs >= 17) {
                                                    return 1;
                                                }
                                            }
                                            else if (diffMVdiffVal >= -20) {
                                                if (AFD < 2) {
                                                    if (refDCval < 22) {
                                                        return 2;
                                                    }
                                                    else if (refDCval >= 22) {
                                                        return 4;
                                                    }
                                                }
                                                else if (AFD >= 2) {
                                                    if (mcTcor < 88) {
                                                        return 2;
                                                    }
                                                    else if (mcTcor >= 88) {
                                                        if (AFD < 5) {
                                                            return 2;
                                                        }
                                                        else if (AFD >= 5) {
                                                            if (AFD < 7) {
                                                                if (SC < 248) {
                                                                    return 8;
                                                                }
                                                                else if (SC >= 248) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (AFD >= 7) {
                                                                return 2;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (refDCval >= 90) {
                                            if (Cs < 84) {
                                                return 1;
                                            }
                                            else if (Cs >= 84) {
                                                return 4;
                                            }
                                        }
                                    }
                                    else if (MVDiff >= 43) {
                                        if (TSCindex < 2) {
                                            if (ssDCval < 23) {
                                                if (tcor < 57) {
                                                    if (diffAFD < -5) {
                                                        return 1;
                                                    }
                                                    else if (diffAFD >= -5) {
                                                        if (MVDiff < 208) {
                                                            if (mu_mv_mag_sq < 18) {
                                                                if (diffMVdiffVal < 23) {
                                                                    return 4;
                                                                }
                                                                else if (diffMVdiffVal >= 23) {
                                                                    if (tcor < 24) {
                                                                        return 4;
                                                                    }
                                                                    else if (tcor >= 24) {
                                                                        return 8;
                                                                    }
                                                                }
                                                            }
                                                            else if (mu_mv_mag_sq >= 18) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (MVDiff >= 208) {
                                                            if (mu_mv_mag_sq < 10) {
                                                                if (mcTcor < 42) {
                                                                    return 8;
                                                                }
                                                                else if (mcTcor >= 42) {
                                                                    if (ssDCval < 20) {
                                                                        return 4;
                                                                    }
                                                                    else if (ssDCval >= 20) {
                                                                        return 1;
                                                                    }
                                                                }
                                                            }
                                                            else if (mu_mv_mag_sq >= 10) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (tcor >= 57) {
                                                    if (MVDiff < 109) {
                                                        return 8;
                                                    }
                                                    else if (MVDiff >= 109) {
                                                        if (tcor < 81) {
                                                            return 2;
                                                        }
                                                        else if (tcor >= 81) {
                                                            if (ssDCval < 20) {
                                                                return 1;
                                                            }
                                                            else if (ssDCval >= 20) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                            else if (ssDCval >= 23) {
                                                if (Rs < 126) {
                                                    return 1;
                                                }
                                                else if (Rs >= 126) {
                                                    return 2;
                                                }
                                            }
                                        }
                                        else if (TSCindex >= 2) {
                                            if (tcor < 52) {
                                                if (Cs < 80) {
                                                    if (AFD < 2) {
                                                        return 1;
                                                    }
                                                    else if (AFD >= 2) {
                                                        return 2;
                                                    }
                                                }
                                                else if (Cs >= 80) {
                                                    return 1;
                                                }
                                            }
                                            else if (tcor >= 52) {
                                                if (mu_mv_mag_sq < 26) {
                                                    if (diffMVdiffVal < -111) {
                                                        if (Rs < 7) {
                                                            return 1;
                                                        }
                                                        else if (Rs >= 7) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (diffMVdiffVal >= -111) {
                                                        return 1;
                                                    }
                                                }
                                                else if (mu_mv_mag_sq >= 26) {
                                                    return 8;
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (RsDiff >= 2) {
                                    if (mu_mv_mag_sq < 3) {
                                        return 1;
                                    }
                                    else if (mu_mv_mag_sq >= 3) {
                                        if (tcor < 75) {
                                            if (refDCval < 80) {
                                                if (SC < 830) {
                                                    if (SC < 587) {
                                                        return 8;
                                                    }
                                                    else if (SC >= 587) {
                                                        return 2;
                                                    }
                                                }
                                                else if (SC >= 830) {
                                                    return 1;
                                                }
                                            }
                                            else if (refDCval >= 80) {
                                                return 2;
                                            }
                                        }
                                        else if (tcor >= 75) {
                                            if (ssDCval < 79) {
                                                if (Rs < 547) {
                                                    if (SC < 376) {
                                                        return 1;
                                                    }
                                                    else if (SC >= 376) {
                                                        return 2;
                                                    }
                                                }
                                                else if (Rs >= 547) {
                                                    return 8;
                                                }
                                            }
                                            else if (ssDCval >= 79) {
                                                if (refDCval < 108) {
                                                    if (Cs < 261) {
                                                        return 1;
                                                    }
                                                    else if (Cs >= 261) {
                                                        return 8;
                                                    }
                                                }
                                                else if (refDCval >= 108) {
                                                    return 2;
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                            else if (Cs >= 309) {
                                if (RsDiff < 11) {
                                    if (refDCval < 86) {
                                        if (Cs < 980) {
                                            if (tcor < 67) {
                                                if (diffAFD < 18) {
                                                    if (TSC < 10) {
                                                        if (refDCval < 44) {
                                                            if (tcor < 45) {
                                                                return 2;
                                                            }
                                                            else if (tcor >= 45) {
                                                                return 1;
                                                            }
                                                        }
                                                        else if (refDCval >= 44) {
                                                            if (AFD < 13) {
                                                                return 4;
                                                            }
                                                            else if (AFD >= 13) {
                                                                if (Rs < 340) {
                                                                    return 1;
                                                                }
                                                                else if (Rs >= 340) {
                                                                    return 2;
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (TSC >= 10) {
                                                        return 1;
                                                    }
                                                }
                                                else if (diffAFD >= 18) {
                                                    return 4;
                                                }
                                            }
                                            else if (tcor >= 67) {
                                                if (SC < 1850) {
                                                    if (CsDiff < 3) {
                                                        if (refDCval < 64) {
                                                            if (mcTcor < 92) {
                                                                if (negBalance < 48) {
                                                                    if (Rs < 915) {
                                                                        return 2;
                                                                    }
                                                                    else if (Rs >= 915) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (negBalance >= 48) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (mcTcor >= 92) {
                                                                if (Rs < 563) {
                                                                    return 4;
                                                                }
                                                                else if (Rs >= 563) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                        else if (refDCval >= 64) {
                                                            if (Rs < 751) {
                                                                return 8;
                                                            }
                                                            else if (Rs >= 751) {
                                                                return 2;
                                                            }
                                                        }
                                                    }
                                                    else if (CsDiff >= 3) {
                                                        if (diffRsCsdiff < 19) {
                                                            if (diffAFD < -3) {
                                                                if (posBalance < 60) {
                                                                    return 8;
                                                                }
                                                                else if (posBalance >= 60) {
                                                                    return 2;
                                                                }
                                                            }
                                                            else if (diffAFD >= -3) {
                                                                if (Cs < 849) {
                                                                    return 2;
                                                                }
                                                                else if (Cs >= 849) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                        else if (diffRsCsdiff >= 19) {
                                                            if (SC < 1509) {
                                                                return 1;
                                                            }
                                                            else if (SC >= 1509) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (SC >= 1850) {
                                                    if (ssDCval < 65) {
                                                        if (SC < 2086) {
                                                            if (mu_mv_mag_sq < 13) {
                                                                if (diffMVdiffVal < -14) {
                                                                    return 1;
                                                                }
                                                                else if (diffMVdiffVal >= -14) {
                                                                    if (TSC < 5) {
                                                                        if (Rs < 1081) {
                                                                            return 4;
                                                                        }
                                                                        else if (Rs >= 1081) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                    else if (TSC >= 5) {
                                                                        return 4;
                                                                    }
                                                                }
                                                            }
                                                            else if (mu_mv_mag_sq >= 13) {
                                                                return 1;
                                                            }
                                                        }
                                                        else if (SC >= 2086) {
                                                            if (ssDCval < 49) {
                                                                if (Rs < 1679) {
                                                                    return 8;
                                                                }
                                                                else if (Rs >= 1679) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (ssDCval >= 49) {
                                                                return 1;
                                                            }
                                                        }
                                                    }
                                                    else if (ssDCval >= 65) {
                                                        if (RsDiff < 6) {
                                                            if (TSC < 7) {
                                                                return 2;
                                                            }
                                                            else if (TSC >= 7) {
                                                                return 1;
                                                            }
                                                        }
                                                        else if (RsDiff >= 6) {
                                                            return 2;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (Cs >= 980) {
                                            if (TSC < 10) {
                                                if (SC < 4282) {
                                                    if (refDCval < 74) {
                                                        if (SC < 2074) {
                                                            return 2;
                                                        }
                                                        else if (SC >= 2074) {
                                                            if (posBalance < 52) {
                                                                if (refDCval < 49) {
                                                                    return 4;
                                                                }
                                                                else if (refDCval >= 49) {
                                                                    if (AFD < 11) {
                                                                        return 1;
                                                                    }
                                                                    else if (AFD >= 11) {
                                                                        return 2;
                                                                    }
                                                                }
                                                            }
                                                            else if (posBalance >= 52) {
                                                                if (diffMVdiffVal < -2) {
                                                                    return 8;
                                                                }
                                                                else if (diffMVdiffVal >= -2) {
                                                                    if (TSC < 8) {
                                                                        if (tcor < 76) {
                                                                            if (Cs < 1046) {
                                                                                return 1;
                                                                            }
                                                                            else if (Cs >= 1046) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                        else if (tcor >= 76) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                    else if (TSC >= 8) {
                                                                        return 4;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (refDCval >= 74) {
                                                        if (RsDiff < 10) {
                                                            if (Rs < 2059) {
                                                                if (tcor < 89) {
                                                                    return 1;
                                                                }
                                                                else if (tcor >= 89) {
                                                                    if (Rs < 1716) {
                                                                        return 2;
                                                                    }
                                                                    else if (Rs >= 1716) {
                                                                        return 1;
                                                                    }
                                                                }
                                                            }
                                                            else if (Rs >= 2059) {
                                                                return 2;
                                                            }
                                                        }
                                                        else if (RsDiff >= 10) {
                                                            return 2;
                                                        }
                                                    }
                                                }
                                                else if (SC >= 4282) {
                                                    if (diffRsCsdiff < -31) {
                                                        return 4;
                                                    }
                                                    else if (diffRsCsdiff >= -31) {
                                                        if (RsDiff < 4) {
                                                            if (AFD < 10) {
                                                                return 8;
                                                            }
                                                            else if (AFD >= 10) {
                                                                return 2;
                                                            }
                                                        }
                                                        else if (RsDiff >= 4) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (TSC >= 10) {
                                                if (RsDiff < 7) {
                                                    if (SC < 2116) {
                                                        if (SC < 1842) {
                                                            return 1;
                                                        }
                                                        else if (SC >= 1842) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (SC >= 2116) {
                                                        if (AFD < 14) {
                                                            return 2;
                                                        }
                                                        else if (AFD >= 14) {
                                                            if (Cs < 1832) {
                                                                return 8;
                                                            }
                                                            else if (Cs >= 1832) {
                                                                if (Rs < 1801) {
                                                                    return 4;
                                                                }
                                                                else if (Rs >= 1801) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (RsDiff >= 7) {
                                                    if (posBalance < 57) {
                                                        if (negBalance < 46) {
                                                            if (CsDiff < 9) {
                                                                return 4;
                                                            }
                                                            else if (CsDiff >= 9) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (negBalance >= 46) {
                                                            if (Cs < 1428) {
                                                                return 2;
                                                            }
                                                            else if (Cs >= 1428) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                    else if (posBalance >= 57) {
                                                        if (ssDCval < 73) {
                                                            if (Cs < 1750) {
                                                                return 4;
                                                            }
                                                            else if (Cs >= 1750) {
                                                                if (Cs < 2353) {
                                                                    return 8;
                                                                }
                                                                else if (Cs >= 2353) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                        else if (ssDCval >= 73) {
                                                            if (CsDiff < 7) {
                                                                return 1;
                                                            }
                                                            else if (CsDiff >= 7) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (refDCval >= 86) {
                                        if (TSC < 9) {
                                            if (refDCval < 126) {
                                                if (posBalance < 59) {
                                                    if (RsCsDiff < 23) {
                                                        if (AFD < 6) {
                                                            return 2;
                                                        }
                                                        else if (AFD >= 6) {
                                                            if (diffAFD < -1) {
                                                                return 2;
                                                            }
                                                            else if (diffAFD >= -1) {
                                                                if (Cs < 931) {
                                                                    if (MVDiff < 31) {
                                                                        return 4;
                                                                    }
                                                                    else if (MVDiff >= 31) {
                                                                        return 2;
                                                                    }
                                                                }
                                                                else if (Cs >= 931) {
                                                                    return 2;
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (RsCsDiff >= 23) {
                                                        if (Rs < 2085) {
                                                            if (posBalance < 40) {
                                                                if (ssDCval < 85) {
                                                                    return 1;
                                                                }
                                                                else if (ssDCval >= 85) {
                                                                    return 2;
                                                                }
                                                            }
                                                            else if (posBalance >= 40) {
                                                                if (SC < 2637) {
                                                                    if (SC < 2413) {
                                                                        if (ssDCval < 100) {
                                                                            return 8;
                                                                        }
                                                                        else if (ssDCval >= 100) {
                                                                            return 2;
                                                                        }
                                                                    }
                                                                    else if (SC >= 2413) {
                                                                        return 2;
                                                                    }
                                                                }
                                                                else if (SC >= 2637) {
                                                                    if (AFD < 15) {
                                                                        if (SC < 3894) {
                                                                            if (RsDiff < 7) {
                                                                                return 8;
                                                                            }
                                                                            else if (RsDiff >= 7) {
                                                                                if (Rs < 1803) {
                                                                                    return 4;
                                                                                }
                                                                                else if (Rs >= 1803) {
                                                                                    return 8;
                                                                                }
                                                                            }
                                                                        }
                                                                        else if (SC >= 3894) {
                                                                            if (Rs < 2028) {
                                                                                return 8;
                                                                            }
                                                                            else if (Rs >= 2028) {
                                                                                return 2;
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (AFD >= 15) {
                                                                        return 2;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (Rs >= 2085) {
                                                            if (AFD < 12) {
                                                                if (Rs < 3314) {
                                                                    if (RsCsDiff < 46) {
                                                                        if (diffMVdiffVal < -1) {
                                                                            return 4;
                                                                        }
                                                                        else if (diffMVdiffVal >= -1) {
                                                                            if (SC < 4898) {
                                                                                return 8;
                                                                            }
                                                                            else if (SC >= 4898) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (RsCsDiff >= 46) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (Rs >= 3314) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (AFD >= 12) {
                                                                if (Rs < 2671) {
                                                                    return 8;
                                                                }
                                                                else if (Rs >= 2671) {
                                                                    return 2;
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (posBalance >= 59) {
                                                    if (RsCsDiff < 9) {
                                                        if (AFD < 9) {
                                                            if (ssDCval < 88) {
                                                                return 2;
                                                            }
                                                            else if (ssDCval >= 88) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (AFD >= 9) {
                                                            if (Rs < 678) {
                                                                if (tcor < 81) {
                                                                    if (negBalance < 54) {
                                                                        return 8;
                                                                    }
                                                                    else if (negBalance >= 54) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (tcor >= 81) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (Rs >= 678) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                    else if (RsCsDiff >= 9) {
                                                        if (MVDiff < 21) {
                                                            if (negBalance < 49) {
                                                                if (AFD < 12) {
                                                                    if (CsDiff < 4) {
                                                                        return 4;
                                                                    }
                                                                    else if (CsDiff >= 4) {
                                                                        if (Cs < 1245) {
                                                                            return 8;
                                                                        }
                                                                        else if (Cs >= 1245) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                }
                                                                else if (AFD >= 12) {
                                                                    return 2;
                                                                }
                                                            }
                                                            else if (negBalance >= 49) {
                                                                if (tcor < 64) {
                                                                    return 2;
                                                                }
                                                                else if (tcor >= 64) {
                                                                    if (Cs < 1713) {
                                                                        if (mcTcor < 81) {
                                                                            if (Cs < 1242) {
                                                                                return 4;
                                                                            }
                                                                            else if (Cs >= 1242) {
                                                                                return 8;
                                                                            }
                                                                        }
                                                                        else if (mcTcor >= 81) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                    else if (Cs >= 1713) {
                                                                        if (Cs < 1730) {
                                                                            return 2;
                                                                        }
                                                                        else if (Cs >= 1730) {
                                                                            if (mcTcor < 82) {
                                                                                return 2;
                                                                            }
                                                                            else if (mcTcor >= 82) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (MVDiff >= 21) {
                                                            if (posBalance < 66) {
                                                                if (RsCsDiff < 16) {
                                                                    return 4;
                                                                }
                                                                else if (RsCsDiff >= 16) {
                                                                    if (Rs < 1217) {
                                                                        return 2;
                                                                    }
                                                                    else if (Rs >= 1217) {
                                                                        return 4;
                                                                    }
                                                                }
                                                            }
                                                            else if (posBalance >= 66) {
                                                                if (SC < 1445) {
                                                                    return 1;
                                                                }
                                                                else if (SC >= 1445) {
                                                                    if (Rs < 1294) {
                                                                        return 2;
                                                                    }
                                                                    else if (Rs >= 1294) {
                                                                        return 8;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                            else if (refDCval >= 126) {
                                                if (mu_mv_mag_sq < 8) {
                                                    if (ssDCval < 140) {
                                                        if (mcTcor < 90) {
                                                            return 1;
                                                        }
                                                        else if (mcTcor >= 90) {
                                                            if (SC < 2584) {
                                                                return 1;
                                                            }
                                                            else if (SC >= 2584) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                    else if (ssDCval >= 140) {
                                                        if (posBalance < 53) {
                                                            if (tcor < 68) {
                                                                if (Rs < 1292) {
                                                                    return 4;
                                                                }
                                                                else if (Rs >= 1292) {
                                                                    return 1;
                                                                }
                                                            }
                                                            else if (tcor >= 68) {
                                                                return 1;
                                                            }
                                                        }
                                                        else if (posBalance >= 53) {
                                                            if (Rs < 1075) {
                                                                return 4;
                                                            }
                                                            else if (Rs >= 1075) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (mu_mv_mag_sq >= 8) {
                                                    if (negBalance < 50) {
                                                        if (Rs < 1698) {
                                                            return 1;
                                                        }
                                                        else if (Rs >= 1698) {
                                                            return 2;
                                                        }
                                                    }
                                                    else if (negBalance >= 50) {
                                                        if (diffMVdiffVal < 128) {
                                                            return 2;
                                                        }
                                                        else if (diffMVdiffVal >= 128) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (TSC >= 9) {
                                            if (Cs < 1388) {
                                                if (tcor < 81) {
                                                    if (Cs < 670) {
                                                        if (Cs < 478) {
                                                            return 2;
                                                        }
                                                        else if (Cs >= 478) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (Cs >= 670) {
                                                        if (MVDiff < 74) {
                                                            if (mcTcor < 80) {
                                                                return 1;
                                                            }
                                                            else if (mcTcor >= 80) {
                                                                if (MVDiff < 15) {
                                                                    return 1;
                                                                }
                                                                else if (MVDiff >= 15) {
                                                                    return 2;
                                                                }
                                                            }
                                                        }
                                                        else if (MVDiff >= 74) {
                                                            return 1;
                                                        }
                                                    }
                                                }
                                                else if (tcor >= 81) {
                                                    if (gchDC < 3) {
                                                        return 2;
                                                    }
                                                    else if (gchDC >= 3) {
                                                        if (Rs < 1132) {
                                                            return 1;
                                                        }
                                                        else if (Rs >= 1132) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (Cs >= 1388) {
                                                if (AFD < 18) {
                                                    if (negBalance < 40) {
                                                        if (SC < 4005) {
                                                            return 8;
                                                        }
                                                        else if (SC >= 4005) {
                                                            return 1;
                                                        }
                                                    }
                                                    else if (negBalance >= 40) {
                                                        if (MVDiff < 15) {
                                                            if (negBalance < 51) {
                                                                if (Cs < 2244) {
                                                                    return 1;
                                                                }
                                                                else if (Cs >= 2244) {
                                                                    return 2;
                                                                }
                                                            }
                                                            else if (negBalance >= 51) {
                                                                if (RsCsDiff < 163) {
                                                                    if (tcor < 90) {
                                                                        return 4;
                                                                    }
                                                                    else if (tcor >= 90) {
                                                                        return 1;
                                                                    }
                                                                }
                                                                else if (RsCsDiff >= 163) {
                                                                    return 2;
                                                                }
                                                            }
                                                        }
                                                        else if (MVDiff >= 15) {
                                                            if (posBalance < 51) {
                                                                return 1;
                                                            }
                                                            else if (posBalance >= 51) {
                                                                if (mcTcor < 92) {
                                                                    if (TSC < 13) {
                                                                        return 4;
                                                                    }
                                                                    else if (TSC >= 13) {
                                                                        return 1;
                                                                    }
                                                                }
                                                                else if (mcTcor >= 92) {
                                                                    if (mu_mv_mag_sq < 2) {
                                                                        if (RsDiff < 9) {
                                                                            return 4;
                                                                        }
                                                                        else if (RsDiff >= 9) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                    else if (mu_mv_mag_sq >= 2) {
                                                                        return 8;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (AFD >= 18) {
                                                    if (TSC < 13) {
                                                        if (negBalance < 55) {
                                                            if (Cs < 1736) {
                                                                return 8;
                                                            }
                                                            else if (Cs >= 1736) {
                                                                if (TSC < 11) {
                                                                    return 4;
                                                                }
                                                                else if (TSC >= 11) {
                                                                    return 1;
                                                                }
                                                            }
                                                        }
                                                        else if (negBalance >= 55) {
                                                            if (Cs < 2084) {
                                                                if (TSC < 11) {
                                                                    return 2;
                                                                }
                                                                else if (TSC >= 11) {
                                                                    return 1;
                                                                }
                                                            }
                                                            else if (Cs >= 2084) {
                                                                return 2;
                                                            }
                                                        }
                                                    }
                                                    else if (TSC >= 13) {
                                                        return 1;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (RsDiff >= 11) {
                                    if (mcTcor < 85) {
                                        if (MVDiff < 21) {
                                            if (Rs < 3467) {
                                                if (RsCsDiff < 205) {
                                                    if (ssDCval < 123) {
                                                        return 4;
                                                    }
                                                    else if (ssDCval >= 123) {
                                                        return 1;
                                                    }
                                                }
                                                else if (RsCsDiff >= 205) {
                                                    return 1;
                                                }
                                            }
                                            else if (Rs >= 3467) {
                                                if (SC < 6319) {
                                                    return 2;
                                                }
                                                else if (SC >= 6319) {
                                                    return 1;
                                                }
                                            }
                                        }
                                        else if (MVDiff >= 21) {
                                            if (AFD < 34) {
                                                if (diffRsCsdiff < 4) {
                                                    if (mcTcor < 82) {
                                                        if (negBalance < 57) {
                                                            return 2;
                                                        }
                                                        else if (negBalance >= 57) {
                                                            if (refDCval < 100) {
                                                                return 1;
                                                            }
                                                            else if (refDCval >= 100) {
                                                                if (negBalance < 60) {
                                                                    if (Cs < 1212) {
                                                                        return 1;
                                                                    }
                                                                    else if (Cs >= 1212) {
                                                                        return 2;
                                                                    }
                                                                }
                                                                else if (negBalance >= 60) {
                                                                    return 2;
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (mcTcor >= 82) {
                                                        return 2;
                                                    }
                                                }
                                                else if (diffRsCsdiff >= 4) {
                                                    if (mcTcor < 75) {
                                                        return 2;
                                                    }
                                                    else if (mcTcor >= 75) {
                                                        if (ssDCval < 84) {
                                                            return 8;
                                                        }
                                                        else if (ssDCval >= 84) {
                                                            if (Cs < 880) {
                                                                if (Rs < 3088) {
                                                                    return 2;
                                                                }
                                                                else if (Rs >= 3088) {
                                                                    return 1;
                                                                }
                                                            }
                                                            else if (Cs >= 880) {
                                                                return 1;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                            else if (AFD >= 34) {
                                                if (ssDCval < 89) {
                                                    return 8;
                                                }
                                                else if (ssDCval >= 89) {
                                                    return 2;
                                                }
                                            }
                                        }
                                    }
                                    else if (mcTcor >= 85) {
                                        if (Cs < 1551) {
                                            if (AFD < 12) {
                                                return 4;
                                            }
                                            else if (AFD >= 12) {
                                                if (diffAFD < 3) {
                                                    if (tcor < 60) {
                                                        if (diffTSC < -1) {
                                                            if (RsDiff < 13) {
                                                                return 2;
                                                            }
                                                            else if (RsDiff >= 13) {
                                                                return 1;
                                                            }
                                                        }
                                                        else if (diffTSC >= -1) {
                                                            return 2;
                                                        }
                                                    }
                                                    else if (tcor >= 60) {
                                                        if (ssDCval < 61) {
                                                            if (AFD < 14) {
                                                                return 4;
                                                            }
                                                            else if (AFD >= 14) {
                                                                return 2;
                                                            }
                                                        }
                                                        else if (ssDCval >= 61) {
                                                            if (diffRsCsdiff < -432) {
                                                                return 4;
                                                            }
                                                            else if (diffRsCsdiff >= -432) {
                                                                return 2;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (diffAFD >= 3) {
                                                    if (AFD < 36) {
                                                        if (AFD < 17) {
                                                            return 8;
                                                        }
                                                        else if (AFD >= 17) {
                                                            return 2;
                                                        }
                                                    }
                                                    else if (AFD >= 36) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                        }
                                        else if (Cs >= 1551) {
                                            if (Cs < 2006) {
                                                if (ssDCval < 117) {
                                                    if (CsDiff < 9) {
                                                        if (Rs < 3192) {
                                                            if (Cs < 1976) {
                                                                return 2;
                                                            }
                                                            else if (Cs >= 1976) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (Rs >= 3192) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (CsDiff >= 9) {
                                                        return 2;
                                                    }
                                                }
                                                else if (ssDCval >= 117) {
                                                    if (RsCsDiff < 453) {
                                                        if (ssDCval < 122) {
                                                            return 4;
                                                        }
                                                        else if (ssDCval >= 122) {
                                                            if (RsDiff < 17) {
                                                                if (Rs < 2969) {
                                                                    return 8;
                                                                }
                                                                else if (Rs >= 2969) {
                                                                    return 2;
                                                                }
                                                            }
                                                            else if (RsDiff >= 17) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                    else if (RsCsDiff >= 453) {
                                                        return 2;
                                                    }
                                                }
                                            }
                                            else if (Cs >= 2006) {
                                                if (TSC < 13) {
                                                    if (mcTcor < 87) {
                                                        if (ssDCval < 110) {
                                                            if (Rs < 2849) {
                                                                return 2;
                                                            }
                                                            else if (Rs >= 2849) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (ssDCval >= 110) {
                                                            return 1;
                                                        }
                                                    }
                                                    else if (mcTcor >= 87) {
                                                        return 2;
                                                    }
                                                }
                                                else if (TSC >= 13) {
                                                    if (TSC < 15) {
                                                        return 8;
                                                    }
                                                    else if (TSC >= 15) {
                                                        return 1;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                        }
                        else if (CsDiff >= 10) {
                            if (mcTcor < 94) {
                                if (negBalance < 55) {
                                    if (MVDiff < 20) {
                                        if (refDCval < 129) {
                                            if (MVDiff < 10) {
                                                if (mu_mv_mag_sq < 3) {
                                                    if (AFD < 17) {
                                                        return 4;
                                                    }
                                                    else if (AFD >= 17) {
                                                        if (SC < 5275) {
                                                            return 4;
                                                        }
                                                        else if (SC >= 5275) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                                else if (mu_mv_mag_sq >= 3) {
                                                    if (posBalance < 65) {
                                                        if (CsDiff < 11) {
                                                            if (RsDiff < 12) {
                                                                return 1;
                                                            }
                                                            else if (RsDiff >= 12) {
                                                                return 2;
                                                            }
                                                        }
                                                        else if (CsDiff >= 11) {
                                                            return 1;
                                                        }
                                                    }
                                                    else if (posBalance >= 65) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                            else if (MVDiff >= 10) {
                                                if (RsCsDiff < 379) {
                                                    if (Cs < 2511) {
                                                        return 1;
                                                    }
                                                    else if (Cs >= 2511) {
                                                        if (negBalance < 48) {
                                                            if (Rs < 4271) {
                                                                return 1;
                                                            }
                                                            else if (Rs >= 4271) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (negBalance >= 48) {
                                                            if (Cs < 2557) {
                                                                return 2;
                                                            }
                                                            else if (Cs >= 2557) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (RsCsDiff >= 379) {
                                                    if (RsCsDiff < 402) {
                                                        return 8;
                                                    }
                                                    else if (RsCsDiff >= 402) {
                                                        if (Cs < 3375) {
                                                            return 1;
                                                        }
                                                        else if (Cs >= 3375) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (refDCval >= 129) {
                                            if (AFD < 30) {
                                                return 1;
                                            }
                                            else if (AFD >= 30) {
                                                if (Cs < 2074) {
                                                    return 4;
                                                }
                                                else if (Cs >= 2074) {
                                                    return 2;
                                                }
                                            }
                                        }
                                    }
                                    else if (MVDiff >= 20) {
                                        if (mcTcor < 91) {
                                            if (SC < 5013) {
                                                if (negBalance < 45) {
                                                    if (Rs < 1262) {
                                                        if (MVDiff < 21) {
                                                            return 1;
                                                        }
                                                        else if (MVDiff >= 21) {
                                                            return 2;
                                                        }
                                                    }
                                                    else if (Rs >= 1262) {
                                                        return 8;
                                                    }
                                                }
                                                else if (negBalance >= 45) {
                                                    if (ssDCval < 79) {
                                                        if (Cs < 1417) {
                                                            return 4;
                                                        }
                                                        else if (Cs >= 1417) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (ssDCval >= 79) {
                                                        if (mu_mv_mag_sq < 7) {
                                                            if (RsDiff < 14) {
                                                                if (AFD < 16) {
                                                                    return 8;
                                                                }
                                                                else if (AFD >= 16) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (RsDiff >= 14) {
                                                                if (Rs < 2744) {
                                                                    return 2;
                                                                }
                                                                else if (Rs >= 2744) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                        else if (mu_mv_mag_sq >= 7) {
                                                            if (Rs < 2431) {
                                                                if (AFD < 18) {
                                                                    return 2;
                                                                }
                                                                else if (AFD >= 18) {
                                                                    return 1;
                                                                }
                                                            }
                                                            else if (Rs >= 2431) {
                                                                if (MVDiff < 31) {
                                                                    return 4;
                                                                }
                                                                else if (MVDiff >= 31) {
                                                                    return 1;
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                            else if (SC >= 5013) {
                                                if (refDCval < 84) {
                                                    if (RsCsDiff < 668) {
                                                        if (Scindex < 6) {
                                                            if (ssDCval < 72) {
                                                                return 1;
                                                            }
                                                            else if (ssDCval >= 72) {
                                                                if (mcTcor < 89) {
                                                                    return 8;
                                                                }
                                                                else if (mcTcor >= 89) {
                                                                    if (posBalance < 54) {
                                                                        return 1;
                                                                    }
                                                                    else if (posBalance >= 54) {
                                                                        return 8;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (Scindex >= 6) {
                                                            if (RsDiff < 17) {
                                                                return 1;
                                                            }
                                                            else if (RsDiff >= 17) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                    else if (RsCsDiff >= 668) {
                                                        if (Cs < 3121) {
                                                            return 1;
                                                        }
                                                        else if (Cs >= 3121) {
                                                            return 2;
                                                        }
                                                    }
                                                }
                                                else if (refDCval >= 84) {
                                                    if (Rs < 4747) {
                                                        if (AFD < 19) {
                                                            if (RsDiff < 13) {
                                                                return 4;
                                                            }
                                                            else if (RsDiff >= 13) {
                                                                return 1;
                                                            }
                                                        }
                                                        else if (AFD >= 19) {
                                                            return 1;
                                                        }
                                                    }
                                                    else if (Rs >= 4747) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                        }
                                        else if (mcTcor >= 91) {
                                            if (Cs < 1958) {
                                                if (SC < 3529) {
                                                    return 1;
                                                }
                                                else if (SC >= 3529) {
                                                    if (gchDC < 1) {
                                                        return 2;
                                                    }
                                                    else if (gchDC >= 1) {
                                                        if (Cs < 1690) {
                                                            if (Rs < 3184) {
                                                                return 4;
                                                            }
                                                            else if (Rs >= 3184) {
                                                                return 2;
                                                            }
                                                        }
                                                        else if (Cs >= 1690) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (Cs >= 1958) {
                                                if (refDCval < 98) {
                                                    if (diffAFD < -4) {
                                                        return 4;
                                                    }
                                                    else if (diffAFD >= -4) {
                                                        if (Cs < 2311) {
                                                            if (diffRsCsdiff < 66) {
                                                                if (Cs < 2268) {
                                                                    return 1;
                                                                }
                                                                else if (Cs >= 2268) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (diffRsCsdiff >= 66) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (Cs >= 2311) {
                                                            return 1;
                                                        }
                                                    }
                                                }
                                                else if (refDCval >= 98) {
                                                    if (RsDiff < 10) {
                                                        if (TSC < 10) {
                                                            return 8;
                                                        }
                                                        else if (TSC >= 10) {
                                                            return 1;
                                                        }
                                                    }
                                                    else if (RsDiff >= 10) {
                                                        if (SC < 5406) {
                                                            return 4;
                                                        }
                                                        else if (SC >= 5406) {
                                                            if (Rs < 3619) {
                                                                return 1;
                                                            }
                                                            else if (Rs >= 3619) {
                                                                return 2;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (negBalance >= 55) {
                                    if (mcTcor < 83) {
                                        if (posBalance < 47) {
                                            if (RsDiff < 16) {
                                                return 8;
                                            }
                                            else if (RsDiff >= 16) {
                                                return 2;
                                            }
                                        }
                                        else if (posBalance >= 47) {
                                            if (refDCval < 139) {
                                                if (RsDiff < 14) {
                                                    if (Scindex < 5) {
                                                        return 1;
                                                    }
                                                    else if (Scindex >= 5) {
                                                        if (refDCval < 80) {
                                                            return 4;
                                                        }
                                                        else if (refDCval >= 80) {
                                                            if (diffAFD < -1) {
                                                                if (Cs < 3554) {
                                                                    return 1;
                                                                }
                                                                else if (Cs >= 3554) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (diffAFD >= -1) {
                                                                if (mcTcor < 77) {
                                                                    if (TSC < 18) {
                                                                        return 1;
                                                                    }
                                                                    else if (TSC >= 18) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (mcTcor >= 77) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (RsDiff >= 14) {
                                                    if (negBalance < 58) {
                                                        return 1;
                                                    }
                                                    else if (negBalance >= 58) {
                                                        if (CsDiff < 14) {
                                                            if (Rs < 2931) {
                                                                return 8;
                                                            }
                                                            else if (Rs >= 2931) {
                                                                return 1;
                                                            }
                                                        }
                                                        else if (CsDiff >= 14) {
                                                            if (CsDiff < 15) {
                                                                return 4;
                                                            }
                                                            else if (CsDiff >= 15) {
                                                                if (Rs < 2147) {
                                                                    return 4;
                                                                }
                                                                else if (Rs >= 2147) {
                                                                    if (Rs < 3684) {
                                                                        if (TSC < 21) {
                                                                            return 2;
                                                                        }
                                                                        else if (TSC >= 21) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                    else if (Rs >= 3684) {
                                                                        return 8;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                            else if (refDCval >= 139) {
                                                return 4;
                                            }
                                        }
                                    }
                                    else if (mcTcor >= 83) {
                                        if (AFD < 23) {
                                            if (posBalance < 56) {
                                                if (MVDiff < 11) {
                                                    return 8;
                                                }
                                                else if (MVDiff >= 11) {
                                                    if (TSC < 10) {
                                                        if (SC < 3639) {
                                                            return 4;
                                                        }
                                                        else if (SC >= 3639) {
                                                            return 2;
                                                        }
                                                    }
                                                    else if (TSC >= 10) {
                                                        if (RsDiff < 15) {
                                                            return 1;
                                                        }
                                                        else if (RsDiff >= 15) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (posBalance >= 56) {
                                                if (Cs < 1963) {
                                                    if (tcor < 81) {
                                                        if (TSC < 13) {
                                                            return 2;
                                                        }
                                                        else if (TSC >= 13) {
                                                            if (RsDiff < 14) {
                                                                return 4;
                                                            }
                                                            else if (RsDiff >= 14) {
                                                                return 2;
                                                            }
                                                        }
                                                    }
                                                    else if (tcor >= 81) {
                                                        return 2;
                                                    }
                                                }
                                                else if (Cs >= 1963) {
                                                    if (AFD < 21) {
                                                        if (mcTcor < 88) {
                                                            return 4;
                                                        }
                                                        else if (mcTcor >= 88) {
                                                            return 2;
                                                        }
                                                    }
                                                    else if (AFD >= 21) {
                                                        if (MVDiff < 23) {
                                                            return 1;
                                                        }
                                                        else if (MVDiff >= 23) {
                                                            return 2;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (AFD >= 23) {
                                            if (Cs < 3086) {
                                                if (diffTSC < -2) {
                                                    if (Rs < 2237) {
                                                        return 1;
                                                    }
                                                    else if (Rs >= 2237) {
                                                        return 8;
                                                    }
                                                }
                                                else if (diffTSC >= -2) {
                                                    if (mu_mv_mag_sq < 9) {
                                                        if (tcor < 61) {
                                                            return 2;
                                                        }
                                                        else if (tcor >= 61) {
                                                            if (mcTcor < 89) {
                                                                if (SC < 3015) {
                                                                    return 1;
                                                                }
                                                                else if (SC >= 3015) {
                                                                    return 2;
                                                                }
                                                            }
                                                            else if (mcTcor >= 89) {
                                                                return 1;
                                                            }
                                                        }
                                                    }
                                                    else if (mu_mv_mag_sq >= 9) {
                                                        if (SC < 4121) {
                                                            if (CsDiff < 12) {
                                                                return 2;
                                                            }
                                                            else if (CsDiff >= 12) {
                                                                return 1;
                                                            }
                                                        }
                                                        else if (SC >= 4121) {
                                                            return 1;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (Cs >= 3086) {
                                                if (MVDiff < 16) {
                                                    return 2;
                                                }
                                                else if (MVDiff >= 16) {
                                                    return 8;
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                            else if (mcTcor >= 94) {
                                if (posBalance < 56) {
                                    if (Rs < 1880) {
                                        return 2;
                                    }
                                    else if (Rs >= 1880) {
                                        return 4;
                                    }
                                }
                                else if (posBalance >= 56) {
                                    if (tcor < 59) {
                                        return 8;
                                    }
                                    else if (tcor >= 59) {
                                        return 2;
                                    }
                                }
                            }
                        }
                    }
                    else if (mu_mv_mag_sq >= 27) {
                        if (AFD < 12) {
                            if (SC < 5057) {
                                if (RsCsDiff < 5) {
                                    if (tcor < 85) {
                                        if (mcTcor < 80) {
                                            if (MVDiff < 365) {
                                                if (TSCindex < 6) {
                                                    if (posBalance < 60) {
                                                        if (posBalance < 35) {
                                                            if (MVDiff < 63) {
                                                                if (posBalance < 27) {
                                                                    if (negBalance < 4) {
                                                                        if (MVDiff < 61) {
                                                                            return 1;
                                                                        }
                                                                        else if (MVDiff >= 61) {
                                                                            if (tcor < 44) {
                                                                                return 8;
                                                                            }
                                                                            else if (tcor >= 44) {
                                                                                return 1;
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (negBalance >= 4) {
                                                                        return 8;
                                                                    }
                                                                }
                                                                else if (posBalance >= 27) {
                                                                    if (diffMVdiffVal < -1) {
                                                                        if (posBalance < 28) {
                                                                            return 1;
                                                                        }
                                                                        else if (posBalance >= 28) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                    else if (diffMVdiffVal >= -1) {
                                                                        return 4;
                                                                    }
                                                                }
                                                            }
                                                            else if (MVDiff >= 63) {
                                                                if (SC < 77) {
                                                                    return 4;
                                                                }
                                                                else if (SC >= 77) {
                                                                    return 1;
                                                                }
                                                            }
                                                        }
                                                        else if (posBalance >= 35) {
                                                            if (SC < 727) {
                                                                if (refDCval < 38) {
                                                                    return 8;
                                                                }
                                                                else if (refDCval >= 38) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (SC >= 727) {
                                                                if (negBalance < 47) {
                                                                    if (Cs < 373) {
                                                                        if (SC < 730) {
                                                                            return 4;
                                                                        }
                                                                        else if (SC >= 730) {
                                                                            if (diffRsCsdiff < 1) {
                                                                                return 1;
                                                                            }
                                                                            else if (diffRsCsdiff >= 1) {
                                                                                return 8;
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (Cs >= 373) {
                                                                        if (Cs < 460) {
                                                                            return 4;
                                                                        }
                                                                        else if (Cs >= 460) {
                                                                            return 2;
                                                                        }
                                                                    }
                                                                }
                                                                else if (negBalance >= 47) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (posBalance >= 60) {
                                                        if (diffAFD < 2) {
                                                            if (tcor < 45) {
                                                                return 8;
                                                            }
                                                            else if (tcor >= 45) {
                                                                if (gchDC < 1) {
                                                                    return 2;
                                                                }
                                                                else if (gchDC >= 1) {
                                                                    if (posBalance < 83) {
                                                                        if (Rs < 1) {
                                                                            return 4;
                                                                        }
                                                                        else if (Rs >= 1) {
                                                                            return 2;
                                                                        }
                                                                    }
                                                                    else if (posBalance >= 83) {
                                                                        return 4;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (diffAFD >= 2) {
                                                            return 1;
                                                        }
                                                    }
                                                }
                                                else if (TSCindex >= 6) {
                                                    if (MVDiff < 32) {
                                                        return 4;
                                                    }
                                                    else if (MVDiff >= 32) {
                                                        if (Rs < 599) {
                                                            return 2;
                                                        }
                                                        else if (Rs >= 599) {
                                                            if (Cs < 588) {
                                                                return 1;
                                                            }
                                                            else if (Cs >= 588) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                            else if (MVDiff >= 365) {
                                                if (diffAFD < -3) {
                                                    return 8;
                                                }
                                                else if (diffAFD >= -3) {
                                                    return 2;
                                                }
                                            }
                                        }
                                        else if (mcTcor >= 80) {
                                            if (ssDCval < 81) {
                                                if (SC < 809) {
                                                    if (Cs < 342) {
                                                        if (ssDCval < 37) {
                                                            if (negBalance < 55) {
                                                                if (mu_mv_mag_sq < 34) {
                                                                    return 2;
                                                                }
                                                                else if (mu_mv_mag_sq >= 34) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (negBalance >= 55) {
                                                                return 1;
                                                            }
                                                        }
                                                        else if (ssDCval >= 37) {
                                                            if (ssDCval < 53) {
                                                                if (mu_mv_mag_sq < 67) {
                                                                    if (Rs < 378) {
                                                                        if (diffMVdiffVal < 6) {
                                                                            return 8;
                                                                        }
                                                                        else if (diffMVdiffVal >= 6) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                    else if (Rs >= 378) {
                                                                        if (MVDiff < 19) {
                                                                            return 1;
                                                                        }
                                                                        else if (MVDiff >= 19) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                }
                                                                else if (mu_mv_mag_sq >= 67) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (ssDCval >= 53) {
                                                                if (SC < 337) {
                                                                    return 1;
                                                                }
                                                                else if (SC >= 337) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (Cs >= 342) {
                                                        if (mcTcor < 84) {
                                                            return 1;
                                                        }
                                                        else if (mcTcor >= 84) {
                                                            if (SC < 785) {
                                                                if (Rs < 427) {
                                                                    return 4;
                                                                }
                                                                else if (Rs >= 427) {
                                                                    return 1;
                                                                }
                                                            }
                                                            else if (SC >= 785) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (SC >= 809) {
                                                    if (TSCindex < 6) {
                                                        if (mu_mv_mag_sq < 66) {
                                                            if (ssDCval < 39) {
                                                                if (posBalance < 54) {
                                                                    if (AFD < 3) {
                                                                        return 1;
                                                                    }
                                                                    else if (AFD >= 3) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (posBalance >= 54) {
                                                                    if (Rs < 448) {
                                                                        if (tcor < 82) {
                                                                            return 4;
                                                                        }
                                                                        else if (tcor >= 82) {
                                                                            return 2;
                                                                        }
                                                                    }
                                                                    else if (Rs >= 448) {
                                                                        return 2;
                                                                    }
                                                                }
                                                            }
                                                            else if (ssDCval >= 39) {
                                                                if (Cs < 294) {
                                                                    return 8;
                                                                }
                                                                else if (Cs >= 294) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                        else if (mu_mv_mag_sq >= 66) {
                                                            if (MVDiff < 24) {
                                                                return 4;
                                                            }
                                                            else if (MVDiff >= 24) {
                                                                if (mcTcor < 84) {
                                                                    return 4;
                                                                }
                                                                else if (mcTcor >= 84) {
                                                                    return 2;
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (TSCindex >= 6) {
                                                        if (MVDiff < 31) {
                                                            if (negBalance < 47) {
                                                                if (SC < 1315) {
                                                                    return 1;
                                                                }
                                                                else if (SC >= 1315) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (negBalance >= 47) {
                                                                if (mu_mv_mag_sq < 51) {
                                                                    return 1;
                                                                }
                                                                else if (mu_mv_mag_sq >= 51) {
                                                                    if (MVDiff < 25) {
                                                                        return 4;
                                                                    }
                                                                    else if (MVDiff >= 25) {
                                                                        return 1;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (MVDiff >= 31) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (ssDCval >= 81) {
                                                if (TSC < 6) {
                                                    return 2;
                                                }
                                                else if (TSC >= 6) {
                                                    return 4;
                                                }
                                            }
                                        }
                                    }
                                    else if (tcor >= 85) {
                                        if (diffMVdiffVal < -9) {
                                            if (ssDCval < 39) {
                                                if (Cs < 211) {
                                                    return 1;
                                                }
                                                else if (Cs >= 211) {
                                                    return 2;
                                                }
                                            }
                                            else if (ssDCval >= 39) {
                                                if (tcor < 87) {
                                                    return 4;
                                                }
                                                else if (tcor >= 87) {
                                                    return 1;
                                                }
                                            }
                                        }
                                        else if (diffMVdiffVal >= -9) {
                                            if (mu_mv_mag_sq < 101) {
                                                if (mcTcor < 92) {
                                                    if (mcTcor < 69) {
                                                        return 8;
                                                    }
                                                    else if (mcTcor >= 69) {
                                                        if (MVDiff < 21) {
                                                            if (SC < 715) {
                                                                return 4;
                                                            }
                                                            else if (SC >= 715) {
                                                                if (Rs < 424) {
                                                                    return 8;
                                                                }
                                                                else if (Rs >= 424) {
                                                                    if (diffRsCsdiff < -5) {
                                                                        return 8;
                                                                    }
                                                                    else if (diffRsCsdiff >= -5) {
                                                                        if (mu_mv_mag_sq < 94) {
                                                                            if (MVDiff < 15) {
                                                                                if (diffMVdiffVal < 1) {
                                                                                    return 4;
                                                                                }
                                                                                else if (diffMVdiffVal >= 1) {
                                                                                    if (Rs < 655) {
                                                                                        return 4;
                                                                                    }
                                                                                    else if (Rs >= 655) {
                                                                                        return 8;
                                                                                    }
                                                                                }
                                                                            }
                                                                            else if (MVDiff >= 15) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                        else if (mu_mv_mag_sq >= 94) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (MVDiff >= 21) {
                                                            if (SC < 890) {
                                                                if (negBalance < 25) {
                                                                    return 8;
                                                                }
                                                                else if (negBalance >= 25) {
                                                                    if (ssDCval < 38) {
                                                                        if (Cs < 182) {
                                                                            return 4;
                                                                        }
                                                                        else if (Cs >= 182) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                    else if (ssDCval >= 38) {
                                                                        return 4;
                                                                    }
                                                                }
                                                            }
                                                            else if (SC >= 890) {
                                                                if (ssDCval < 39) {
                                                                    if (Rs < 897) {
                                                                        if (Rs < 637) {
                                                                            return 2;
                                                                        }
                                                                        else if (Rs >= 637) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                    else if (Rs >= 897) {
                                                                        return 2;
                                                                    }
                                                                }
                                                                else if (ssDCval >= 39) {
                                                                    if (Cs < 1549) {
                                                                        return 4;
                                                                    }
                                                                    else if (Cs >= 1549) {
                                                                        return 1;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (mcTcor >= 92) {
                                                    if (TSCindex < 6) {
                                                        if (MVDiff < 18) {
                                                            if (negBalance < 44) {
                                                                if (diffAFD < -1) {
                                                                    if (gchDC < 1) {
                                                                        return 4;
                                                                    }
                                                                    else if (gchDC >= 1) {
                                                                        if (Cs < 652) {
                                                                            return 4;
                                                                        }
                                                                        else if (Cs >= 652) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                }
                                                                else if (diffAFD >= -1) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (negBalance >= 44) {
                                                                if (MVDiff < 16) {
                                                                    if (mu_mv_mag_sq < 55) {
                                                                        return 1;
                                                                    }
                                                                    else if (mu_mv_mag_sq >= 55) {
                                                                        if (TSC < 4) {
                                                                            return 8;
                                                                        }
                                                                        else if (TSC >= 4) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                }
                                                                else if (MVDiff >= 16) {
                                                                    if (Rs < 1844) {
                                                                        return 4;
                                                                    }
                                                                    else if (Rs >= 1844) {
                                                                        return 8;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (MVDiff >= 18) {
                                                            if (diffMVdiffVal < -2) {
                                                                if (Rs < 1845) {
                                                                    if (Rs < 977) {
                                                                        return 2;
                                                                    }
                                                                    else if (Rs >= 977) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (Rs >= 1845) {
                                                                    return 1;
                                                                }
                                                            }
                                                            else if (diffMVdiffVal >= -2) {
                                                                if (diffRsCsdiff < -1) {
                                                                    if (AFD < 6) {
                                                                        return 2;
                                                                    }
                                                                    else if (AFD >= 6) {
                                                                        return 8;
                                                                    }
                                                                }
                                                                else if (diffRsCsdiff >= -1) {
                                                                    if (mu_mv_mag_sq < 62) {
                                                                        return 4;
                                                                    }
                                                                    else if (mu_mv_mag_sq >= 62) {
                                                                        if (Rs < 1836) {
                                                                            return 1;
                                                                        }
                                                                        else if (Rs >= 1836) {
                                                                            if (negBalance < 44) {
                                                                                return 4;
                                                                            }
                                                                            else if (negBalance >= 44) {
                                                                                return 8;
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (TSCindex >= 6) {
                                                        if (MVDiff < 9) {
                                                            return 8;
                                                        }
                                                        else if (MVDiff >= 9) {
                                                            return 1;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (mu_mv_mag_sq >= 101) {
                                                return 8;
                                            }
                                        }
                                    }
                                }
                                else if (RsCsDiff >= 5) {
                                    if (refDCval < 90) {
                                        if (mu_mv_mag_sq < 74) {
                                            if (posBalance < 45) {
                                                if (CsDiff < 6) {
                                                    if (SC < 2095) {
                                                        if (mcTcor < 93) {
                                                            return 8;
                                                        }
                                                        else if (mcTcor >= 93) {
                                                            if (Cs < 1055) {
                                                                if (Cs < 549) {
                                                                    return 2;
                                                                }
                                                                else if (Cs >= 549) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (Cs >= 1055) {
                                                                return 2;
                                                            }
                                                        }
                                                    }
                                                    else if (SC >= 2095) {
                                                        if (diffMVdiffVal < 2) {
                                                            if (RsCsDiff < 46) {
                                                                if (RsCsDiff < 19) {
                                                                    return 4;
                                                                }
                                                                else if (RsCsDiff >= 19) {
                                                                    return 2;
                                                                }
                                                            }
                                                            else if (RsCsDiff >= 46) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (diffMVdiffVal >= 2) {
                                                            return 2;
                                                        }
                                                    }
                                                }
                                                else if (CsDiff >= 6) {
                                                    if (ssDCval < 81) {
                                                        return 8;
                                                    }
                                                    else if (ssDCval >= 81) {
                                                        return 1;
                                                    }
                                                }
                                            }
                                            else if (posBalance >= 45) {
                                                if (ssDCval < 39) {
                                                    if (Rs < 638) {
                                                        if (SC < 903) {
                                                            if (Cs < 267) {
                                                                return 1;
                                                            }
                                                            else if (Cs >= 267) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (SC >= 903) {
                                                            return 2;
                                                        }
                                                    }
                                                    else if (Rs >= 638) {
                                                        if (Scindex < 4) {
                                                            if (TSC < 6) {
                                                                if (tcor < 77) {
                                                                    if (diffTSC < -1) {
                                                                        return 1;
                                                                    }
                                                                    else if (diffTSC >= -1) {
                                                                        if (AFD < 7) {
                                                                            return 4;
                                                                        }
                                                                        else if (AFD >= 7) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                }
                                                                else if (tcor >= 77) {
                                                                    if (negBalance < 48) {
                                                                        if (Rs < 791) {
                                                                            return 4;
                                                                        }
                                                                        else if (Rs >= 791) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                    else if (negBalance >= 48) {
                                                                        return 1;
                                                                    }
                                                                }
                                                            }
                                                            else if (TSC >= 6) {
                                                                if (posBalance < 59) {
                                                                    return 1;
                                                                }
                                                                else if (posBalance >= 59) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                        else if (Scindex >= 4) {
                                                            if (Rs < 1767) {
                                                                return 2;
                                                            }
                                                            else if (Rs >= 1767) {
                                                                return 1;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (ssDCval >= 39) {
                                                    if (negBalance < 44) {
                                                        if (SC < 2020) {
                                                            if (diffMVdiffVal < -8) {
                                                                if (diffRsCsdiff < -12) {
                                                                    return 8;
                                                                }
                                                                else if (diffRsCsdiff >= -12) {
                                                                    return 1;
                                                                }
                                                            }
                                                            else if (diffMVdiffVal >= -8) {
                                                                if (mu_mv_mag_sq < 38) {
                                                                    if (Cs < 742) {
                                                                        return 8;
                                                                    }
                                                                    else if (Cs >= 742) {
                                                                        if (ssDCval < 53) {
                                                                            if (posBalance < 48) {
                                                                                return 2;
                                                                            }
                                                                            else if (posBalance >= 48) {
                                                                                return 1;
                                                                            }
                                                                        }
                                                                        else if (ssDCval >= 53) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                }
                                                                else if (mu_mv_mag_sq >= 38) {
                                                                    if (diffTSC < -2) {
                                                                        return 2;
                                                                    }
                                                                    else if (diffTSC >= -2) {
                                                                        if (mcTcor < 76) {
                                                                            return 8;
                                                                        }
                                                                        else if (mcTcor >= 76) {
                                                                            if (diffRsCsdiff < -6) {
                                                                                if (Cs < 769) {
                                                                                    return 4;
                                                                                }
                                                                                else if (Cs >= 769) {
                                                                                    return 2;
                                                                                }
                                                                            }
                                                                            else if (diffRsCsdiff >= -6) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (SC >= 2020) {
                                                            if (TSCindex < 6) {
                                                                if (ssDCval < 54) {
                                                                    if (negBalance < 38) {
                                                                        return 1;
                                                                    }
                                                                    else if (negBalance >= 38) {
                                                                        if (Rs < 1221) {
                                                                            if (RsDiff < 4) {
                                                                                return 4;
                                                                            }
                                                                            else if (RsDiff >= 4) {
                                                                                return 8;
                                                                            }
                                                                        }
                                                                        else if (Rs >= 1221) {
                                                                            if (negBalance < 40) {
                                                                                if (Cs < 935) {
                                                                                    if (refDCval < 51) {
                                                                                        return 4;
                                                                                    }
                                                                                    else if (refDCval >= 51) {
                                                                                        return 1;
                                                                                    }
                                                                                }
                                                                                else if (Cs >= 935) {
                                                                                    return 4;
                                                                                }
                                                                            }
                                                                            else if (negBalance >= 40) {
                                                                                return 1;
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                                else if (ssDCval >= 54) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (TSCindex >= 6) {
                                                                if (Rs < 2506) {
                                                                    if (gchDC < 2) {
                                                                        if (Cs < 1000) {
                                                                            if (negBalance < 39) {
                                                                                return 8;
                                                                            }
                                                                            else if (negBalance >= 39) {
                                                                                if (Rs < 1294) {
                                                                                    return 8;
                                                                                }
                                                                                else if (Rs >= 1294) {
                                                                                    return 1;
                                                                                }
                                                                            }
                                                                        }
                                                                        else if (Cs >= 1000) {
                                                                            if (refDCval < 51) {
                                                                                return 4;
                                                                            }
                                                                            else if (refDCval >= 51) {
                                                                                if (posBalance < 46) {
                                                                                    return 1;
                                                                                }
                                                                                else if (posBalance >= 46) {
                                                                                    if (posBalance < 54) {
                                                                                        if (refDCval < 52) {
                                                                                            if (Cs < 1046) {
                                                                                                return 8;
                                                                                            }
                                                                                            else if (Cs >= 1046) {
                                                                                                return 1;
                                                                                            }
                                                                                        }
                                                                                        else if (refDCval >= 52) {
                                                                                            if (AFD < 7) {
                                                                                                return 4;
                                                                                            }
                                                                                            else if (AFD >= 7) {
                                                                                                return 8;
                                                                                            }
                                                                                        }
                                                                                    }
                                                                                    else if (posBalance >= 54) {
                                                                                        if (Rs < 1495) {
                                                                                            if (AFD < 9) {
                                                                                                return 8;
                                                                                            }
                                                                                            else if (AFD >= 9) {
                                                                                                return 1;
                                                                                            }
                                                                                        }
                                                                                        else if (Rs >= 1495) {
                                                                                            if (MVDiff < 10) {
                                                                                                return 1;
                                                                                            }
                                                                                            else if (MVDiff >= 10) {
                                                                                                return 4;
                                                                                            }
                                                                                        }
                                                                                    }
                                                                                }
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (gchDC >= 2) {
                                                                        if (RsDiff < 6) {
                                                                            return 8;
                                                                        }
                                                                        else if (RsDiff >= 6) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                }
                                                                else if (Rs >= 2506) {
                                                                    if (Cs < 673) {
                                                                        return 2;
                                                                    }
                                                                    else if (Cs >= 673) {
                                                                        return 4;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (negBalance >= 44) {
                                                        if (tcor < 92) {
                                                            if (diffRsCsdiff < -17) {
                                                                if (TSC < 8) {
                                                                    if (Cs < 980) {
                                                                        return 2;
                                                                    }
                                                                    else if (Cs >= 980) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (TSC >= 8) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (diffRsCsdiff >= -17) {
                                                                if (posBalance < 52) {
                                                                    if (mcTcor < 88) {
                                                                        return 8;
                                                                    }
                                                                    else if (mcTcor >= 88) {
                                                                        if (negBalance < 45) {
                                                                            if (Cs < 1009) {
                                                                                return 1;
                                                                            }
                                                                            else if (Cs >= 1009) {
                                                                                return 8;
                                                                            }
                                                                        }
                                                                        else if (negBalance >= 45) {
                                                                            if (MVDiff < 17) {
                                                                                return 4;
                                                                            }
                                                                            else if (MVDiff >= 17) {
                                                                                if (Rs < 999) {
                                                                                    return 8;
                                                                                }
                                                                                else if (Rs >= 999) {
                                                                                    return 4;
                                                                                }
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                                else if (posBalance >= 52) {
                                                                    if (mu_mv_mag_sq < 44) {
                                                                        if (diffMVdiffVal < 16) {
                                                                            if (refDCval < 89) {
                                                                                if (ssDCval < 40) {
                                                                                    if (diffAFD < -1) {
                                                                                        return 4;
                                                                                    }
                                                                                    else if (diffAFD >= -1) {
                                                                                        if (MVDiff < 24) {
                                                                                            return 4;
                                                                                        }
                                                                                        else if (MVDiff >= 24) {
                                                                                            if (diffMVdiffVal < -9) {
                                                                                                return 4;
                                                                                            }
                                                                                            else if (diffMVdiffVal >= -9) {
                                                                                                if (mu_mv_mag_sq < 29) {
                                                                                                    if (Cs < 559) {
                                                                                                        return 4;
                                                                                                    }
                                                                                                    else if (Cs >= 559) {
                                                                                                        return 1;
                                                                                                    }
                                                                                                }
                                                                                                else if (mu_mv_mag_sq >= 29) {
                                                                                                    if (posBalance < 57) {
                                                                                                        return 4;
                                                                                                    }
                                                                                                    else if (posBalance >= 57) {
                                                                                                        if (diffRsCsdiff < 2) {
                                                                                                            return 1;
                                                                                                        }
                                                                                                        else if (diffRsCsdiff >= 2) {
                                                                                                            if (MVDiff < 40) {
                                                                                                                return 1;
                                                                                                            }
                                                                                                            else if (MVDiff >= 40) {
                                                                                                                return 4;
                                                                                                            }
                                                                                                        }
                                                                                                    }
                                                                                                }
                                                                                            }
                                                                                        }
                                                                                    }
                                                                                }
                                                                                else if (ssDCval >= 40) {
                                                                                    return 1;
                                                                                }
                                                                            }
                                                                            else if (refDCval >= 89) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                        else if (diffMVdiffVal >= 16) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                    else if (mu_mv_mag_sq >= 44) {
                                                                        if (Cs < 1558) {
                                                                            if (refDCval < 46) {
                                                                                return 4;
                                                                            }
                                                                            else if (refDCval >= 46) {
                                                                                if (tcor < 79) {
                                                                                    return 4;
                                                                                }
                                                                                else if (tcor >= 79) {
                                                                                    if (refDCval < 73) {
                                                                                        return 1;
                                                                                    }
                                                                                    else if (refDCval >= 73) {
                                                                                        if (Cs < 793) {
                                                                                            return 1;
                                                                                        }
                                                                                        else if (Cs >= 793) {
                                                                                            return 4;
                                                                                        }
                                                                                    }
                                                                                }
                                                                            }
                                                                        }
                                                                        else if (Cs >= 1558) {
                                                                            if (AFD < 8) {
                                                                                return 1;
                                                                            }
                                                                            else if (AFD >= 8) {
                                                                                if (Cs < 1940) {
                                                                                    return 8;
                                                                                }
                                                                                else if (Cs >= 1940) {
                                                                                    return 4;
                                                                                }
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (tcor >= 92) {
                                                            if (Rs < 909) {
                                                                return 2;
                                                            }
                                                            else if (Rs >= 909) {
                                                                if (MVDiff < 31) {
                                                                    if (MVDiff < 7) {
                                                                        if (SC < 2974) {
                                                                            return 4;
                                                                        }
                                                                        else if (SC >= 2974) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                    else if (MVDiff >= 7) {
                                                                        return 1;
                                                                    }
                                                                }
                                                                else if (MVDiff >= 31) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (mu_mv_mag_sq >= 74) {
                                            if (mu_mv_mag_sq < 105) {
                                                if (Cs < 161) {
                                                    return 1;
                                                }
                                                else if (Cs >= 161) {
                                                    if (mcTcor < 94) {
                                                        if (ssDCval < 79) {
                                                            if (RsCsDiff < 107) {
                                                                if (TSCindex < 5) {
                                                                    return 8;
                                                                }
                                                                else if (TSCindex >= 5) {
                                                                    if (negBalance < 34) {
                                                                        return 4;
                                                                    }
                                                                    else if (negBalance >= 34) {
                                                                        if (MVDiff < 14) {
                                                                            return 8;
                                                                        }
                                                                        else if (MVDiff >= 14) {
                                                                            if (Cs < 1725) {
                                                                                return 4;
                                                                            }
                                                                            else if (Cs >= 1725) {
                                                                                return 8;
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                            else if (RsCsDiff >= 107) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (ssDCval >= 79) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (mcTcor >= 94) {
                                                        if (SC < 4301) {
                                                            if (MVDiff < 7) {
                                                                if (Cs < 1294) {
                                                                    return 1;
                                                                }
                                                                else if (Cs >= 1294) {
                                                                    return 2;
                                                                }
                                                            }
                                                            else if (MVDiff >= 7) {
                                                                if (MVDiff < 11) {
                                                                    if (Cs < 908) {
                                                                        return 8;
                                                                    }
                                                                    else if (Cs >= 908) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (MVDiff >= 11) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                        else if (SC >= 4301) {
                                                            return 1;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (mu_mv_mag_sq >= 105) {
                                                if (mcTcor < 95) {
                                                    return 8;
                                                }
                                                else if (mcTcor >= 95) {
                                                    if (Rs < 1099) {
                                                        return 8;
                                                    }
                                                    else if (Rs >= 1099) {
                                                        return 1;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (refDCval >= 90) {
                                        if (ssDCval < 116) {
                                            if (negBalance < 52) {
                                                if (AFD < 11) {
                                                    if (mu_mv_mag_sq < 48) {
                                                        if (TSCindex < 8) {
                                                            if (Cs < 535) {
                                                                return 1;
                                                            }
                                                            else if (Cs >= 535) {
                                                                if (negBalance < 37) {
                                                                    if (AFD < 7) {
                                                                        return 4;
                                                                    }
                                                                    else if (AFD >= 7) {
                                                                        return 8;
                                                                    }
                                                                }
                                                                else if (negBalance >= 37) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                        else if (TSCindex >= 8) {
                                                            if (Rs < 340) {
                                                                return 4;
                                                            }
                                                            else if (Rs >= 340) {
                                                                if (MVDiff < 9) {
                                                                    return 4;
                                                                }
                                                                else if (MVDiff >= 9) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (mu_mv_mag_sq >= 48) {
                                                        if (mcTcor < 96) {
                                                            if (RsCsDiff < 141) {
                                                                return 8;
                                                            }
                                                            else if (RsCsDiff >= 141) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (mcTcor >= 96) {
                                                            if (Rs < 2999) {
                                                                if (Cs < 2112) {
                                                                    return 1;
                                                                }
                                                                else if (Cs >= 2112) {
                                                                    return 2;
                                                                }
                                                            }
                                                            else if (Rs >= 2999) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (AFD >= 11) {
                                                    if (Rs < 2518) {
                                                        if (RsCsDiff < 99) {
                                                            if (Cs < 1291) {
                                                                return 2;
                                                            }
                                                            else if (Cs >= 1291) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (RsCsDiff >= 99) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (Rs >= 2518) {
                                                        if (TSC < 8) {
                                                            return 2;
                                                        }
                                                        else if (TSC >= 8) {
                                                            return 1;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (negBalance >= 52) {
                                                if (ssDCval < 98) {
                                                    if (ssDCval < 93) {
                                                        if (tcor < 85) {
                                                            if (RsDiff < 9) {
                                                                return 2;
                                                            }
                                                            else if (RsDiff >= 9) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (tcor >= 85) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (ssDCval >= 93) {
                                                        if (MVDiff < 10) {
                                                            if (diffMVdiffVal < -4) {
                                                                return 8;
                                                            }
                                                            else if (diffMVdiffVal >= -4) {
                                                                if (mcTcor < 82) {
                                                                    if (RsDiff < 6) {
                                                                        return 8;
                                                                    }
                                                                    else if (RsDiff >= 6) {
                                                                        return 2;
                                                                    }
                                                                }
                                                                else if (mcTcor >= 82) {
                                                                    return 2;
                                                                }
                                                            }
                                                        }
                                                        else if (MVDiff >= 10) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                                else if (ssDCval >= 98) {
                                                    if (tcor < 83) {
                                                        return 4;
                                                    }
                                                    else if (tcor >= 83) {
                                                        if (TSC < 8) {
                                                            if (refDCval < 100) {
                                                                return 4;
                                                            }
                                                            else if (refDCval >= 100) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (TSC >= 8) {
                                                            return 2;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (ssDCval >= 116) {
                                            if (ssDCval < 127) {
                                                if (Cs < 1884) {
                                                    if (negBalance < 51) {
                                                        if (Rs < 2049) {
                                                            if (Rs < 1632) {
                                                                if (Cs < 1132) {
                                                                    return 8;
                                                                }
                                                                else if (Cs >= 1132) {
                                                                    return 1;
                                                                }
                                                            }
                                                            else if (Rs >= 1632) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (Rs >= 2049) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (negBalance >= 51) {
                                                        if (TSC < 9) {
                                                            return 4;
                                                        }
                                                        else if (TSC >= 9) {
                                                            if (ssDCval < 119) {
                                                                if (Cs < 1622) {
                                                                    return 8;
                                                                }
                                                                else if (Cs >= 1622) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (ssDCval >= 119) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (Cs >= 1884) {
                                                    return 4;
                                                }
                                            }
                                            else if (ssDCval >= 127) {
                                                if (mcTcor < 96) {
                                                    if (Rs < 808) {
                                                        if (Cs < 587) {
                                                            return 4;
                                                        }
                                                        else if (Cs >= 587) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (Rs >= 808) {
                                                        if (refDCval < 141) {
                                                            if (Rs < 1262) {
                                                                return 1;
                                                            }
                                                            else if (Rs >= 1262) {
                                                                if (SC < 4048) {
                                                                    return 8;
                                                                }
                                                                else if (SC >= 4048) {
                                                                    return 1;
                                                                }
                                                            }
                                                        }
                                                        else if (refDCval >= 141) {
                                                            if (diffRsCsdiff < -19) {
                                                                return 4;
                                                            }
                                                            else if (diffRsCsdiff >= -19) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (mcTcor >= 96) {
                                                    if (SC < 1196) {
                                                        return 4;
                                                    }
                                                    else if (SC >= 1196) {
                                                        return 2;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                            else if (SC >= 5057) {
                                if (tcor < 86) {
                                    if (ssDCval < 131) {
                                        if (SC < 7037) {
                                            if (posBalance < 59) {
                                                return 8;
                                            }
                                            else if (posBalance >= 59) {
                                                if (AFD < 11) {
                                                    return 8;
                                                }
                                                else if (AFD >= 11) {
                                                    return 4;
                                                }
                                            }
                                        }
                                        else if (SC >= 7037) {
                                            if (TSC < 9) {
                                                if (ssDCval < 54) {
                                                    return 4;
                                                }
                                                else if (ssDCval >= 54) {
                                                    return 8;
                                                }
                                            }
                                            else if (TSC >= 9) {
                                                return 1;
                                            }
                                        }
                                    }
                                    else if (ssDCval >= 131) {
                                        return 1;
                                    }
                                }
                                else if (tcor >= 86) {
                                    if (TSCindex < 8) {
                                        if (ssDCval < 108) {
                                            if (refDCval < 54) {
                                                if (MVDiff < 21) {
                                                    if (mu_mv_mag_sq < 90) {
                                                        if (negBalance < 53) {
                                                            return 4;
                                                        }
                                                        else if (negBalance >= 53) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (mu_mv_mag_sq >= 90) {
                                                        return 8;
                                                    }
                                                }
                                                else if (MVDiff >= 21) {
                                                    if (mu_mv_mag_sq < 36) {
                                                        return 8;
                                                    }
                                                    else if (mu_mv_mag_sq >= 36) {
                                                        if (AFD < 5) {
                                                            return 8;
                                                        }
                                                        else if (AFD >= 5) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (refDCval >= 54) {
                                                if (MVDiff < 27) {
                                                    if (AFD < 7) {
                                                        return 8;
                                                    }
                                                    else if (AFD >= 7) {
                                                        if (tcor < 94) {
                                                            if (Cs < 1997) {
                                                                if (SC < 5727) {
                                                                    if (ssDCval < 107) {
                                                                        return 8;
                                                                    }
                                                                    else if (ssDCval >= 107) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (SC >= 5727) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (Cs >= 1997) {
                                                                if (Rs < 3954) {
                                                                    return 8;
                                                                }
                                                                else if (Rs >= 3954) {
                                                                    if (Cs < 2995) {
                                                                        return 8;
                                                                    }
                                                                    else if (Cs >= 2995) {
                                                                        return 4;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (tcor >= 94) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                                else if (MVDiff >= 27) {
                                                    return 2;
                                                }
                                            }
                                        }
                                        else if (ssDCval >= 108) {
                                            if (gchDC < 1) {
                                                if (Rs < 4853) {
                                                    return 4;
                                                }
                                                else if (Rs >= 4853) {
                                                    return 8;
                                                }
                                            }
                                            else if (gchDC >= 1) {
                                                if (Cs < 2054) {
                                                    return 4;
                                                }
                                                else if (Cs >= 2054) {
                                                    if (ssDCval < 119) {
                                                        return 2;
                                                    }
                                                    else if (ssDCval >= 119) {
                                                        if (SC < 7269) {
                                                            return 4;
                                                        }
                                                        else if (SC >= 7269) {
                                                            return 2;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (TSCindex >= 8) {
                                        if (refDCval < 128) {
                                            if (Rs < 3615) {
                                                if (mu_mv_mag_sq < 76) {
                                                    if (posBalance < 47) {
                                                        if (mu_mv_mag_sq < 64) {
                                                            if (SC < 5137) {
                                                                return 2;
                                                            }
                                                            else if (SC >= 5137) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (mu_mv_mag_sq >= 64) {
                                                            if (RsDiff < 10) {
                                                                return 2;
                                                            }
                                                            else if (RsDiff >= 10) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                    else if (posBalance >= 47) {
                                                        if (Cs < 3514) {
                                                            if (RsCsDiff < 16) {
                                                                return 2;
                                                            }
                                                            else if (RsCsDiff >= 16) {
                                                                if (Rs < 3510) {
                                                                    if (diffRsCsdiff < -13) {
                                                                        return 2;
                                                                    }
                                                                    else if (diffRsCsdiff >= -13) {
                                                                        if (mu_mv_mag_sq < 36) {
                                                                            if (RsCsDiff < 22) {
                                                                                if (Rs < 3318) {
                                                                                    return 8;
                                                                                }
                                                                                else if (Rs >= 3318) {
                                                                                    if (Cs < 3387) {
                                                                                        return 2;
                                                                                    }
                                                                                    else if (Cs >= 3387) {
                                                                                        return 8;
                                                                                    }
                                                                                }
                                                                            }
                                                                            else if (RsCsDiff >= 22) {
                                                                                return 8;
                                                                            }
                                                                        }
                                                                        else if (mu_mv_mag_sq >= 36) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                }
                                                                else if (Rs >= 3510) {
                                                                    if (Rs < 3614) {
                                                                        return 2;
                                                                    }
                                                                    else if (Rs >= 3614) {
                                                                        return 8;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (Cs >= 3514) {
                                                            if (mcTcor < 95) {
                                                                return 4;
                                                            }
                                                            else if (mcTcor >= 95) {
                                                                return 2;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (mu_mv_mag_sq >= 76) {
                                                    if (Rs < 2979) {
                                                        return 4;
                                                    }
                                                    else if (Rs >= 2979) {
                                                        if (mcTcor < 94) {
                                                            return 8;
                                                        }
                                                        else if (mcTcor >= 94) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (Rs >= 3615) {
                                                if (refDCval < 51) {
                                                    return 2;
                                                }
                                                else if (refDCval >= 51) {
                                                    if (Rs < 4507) {
                                                        if (Cs < 2416) {
                                                            if (Cs < 2196) {
                                                                return 4;
                                                            }
                                                            else if (Cs >= 2196) {
                                                                if (tcor < 93) {
                                                                    if (RsDiff < 7) {
                                                                        return 8;
                                                                    }
                                                                    else if (RsDiff >= 7) {
                                                                        if (Cs < 2244) {
                                                                            if (Rs < 4065) {
                                                                                return 2;
                                                                            }
                                                                            else if (Rs >= 4065) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                        else if (Cs >= 2244) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                }
                                                                else if (tcor >= 93) {
                                                                    if (MVDiff < 22) {
                                                                        return 2;
                                                                    }
                                                                    else if (MVDiff >= 22) {
                                                                        if (Rs < 4071) {
                                                                            return 2;
                                                                        }
                                                                        else if (Rs >= 4071) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (Cs >= 2416) {
                                                            if (MVDiff < 19) {
                                                                if (posBalance < 46) {
                                                                    if (ssDCval < 125) {
                                                                        if (RsCsDiff < 131) {
                                                                            return 8;
                                                                        }
                                                                        else if (RsCsDiff >= 131) {
                                                                            if (Rs < 4090) {
                                                                                return 8;
                                                                            }
                                                                            else if (Rs >= 4090) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (ssDCval >= 125) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (posBalance >= 46) {
                                                                    if (CsDiff < 3) {
                                                                        return 2;
                                                                    }
                                                                    else if (CsDiff >= 3) {
                                                                        return 4;
                                                                    }
                                                                }
                                                            }
                                                            else if (MVDiff >= 19) {
                                                                if (Rs < 3739) {
                                                                    if (negBalance < 53) {
                                                                        return 8;
                                                                    }
                                                                    else if (negBalance >= 53) {
                                                                        return 2;
                                                                    }
                                                                }
                                                                else if (Rs >= 3739) {
                                                                    if (diffAFD < -1) {
                                                                        if (diffTSC < -2) {
                                                                            return 8;
                                                                        }
                                                                        else if (diffTSC >= -2) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                    else if (diffAFD >= -1) {
                                                                        return 2;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (Rs >= 4507) {
                                                        if (mu_mv_mag_sq < 82) {
                                                            if (mcTcor < 89) {
                                                                return 2;
                                                            }
                                                            else if (mcTcor >= 89) {
                                                                if (posBalance < 47) {
                                                                    if (Cs < 2881) {
                                                                        if (ssDCval < 124) {
                                                                            return 8;
                                                                        }
                                                                        else if (ssDCval >= 124) {
                                                                            return 2;
                                                                        }
                                                                    }
                                                                    else if (Cs >= 2881) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (posBalance >= 47) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                        else if (mu_mv_mag_sq >= 82) {
                                                            if (Rs < 4604) {
                                                                return 8;
                                                            }
                                                            else if (Rs >= 4604) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (refDCval >= 128) {
                                            if (Cs < 2503) {
                                                if (TSC < 9) {
                                                    if (Cs < 2205) {
                                                        return 4;
                                                    }
                                                    else if (Cs >= 2205) {
                                                        return 2;
                                                    }
                                                }
                                                else if (TSC >= 9) {
                                                    return 8;
                                                }
                                            }
                                            else if (Cs >= 2503) {
                                                return 2;
                                            }
                                        }
                                    }
                                }
                            }
                        }
                        else if (AFD >= 12) {
                            if (tcor < 90) {
                                if (Cs < 1831) {
                                    if (refDCval < 75) {
                                        if (Cs < 1141) {
                                            if (CsDiff < 4) {
                                                if (ssDCval < 58) {
                                                    if (Rs < 569) {
                                                        return 2;
                                                    }
                                                    else if (Rs >= 569) {
                                                        return 8;
                                                    }
                                                }
                                                else if (ssDCval >= 58) {
                                                    return 2;
                                                }
                                            }
                                            else if (CsDiff >= 4) {
                                                if (mu_mv_mag_sq < 42) {
                                                    if (Rs < 1187) {
                                                        return 1;
                                                    }
                                                    else if (Rs >= 1187) {
                                                        if (SC < 3610) {
                                                            return 8;
                                                        }
                                                        else if (SC >= 3610) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                                else if (mu_mv_mag_sq >= 42) {
                                                    if (posBalance < 48) {
                                                        return 2;
                                                    }
                                                    else if (posBalance >= 48) {
                                                        if (MVDiff < 282) {
                                                            return 4;
                                                        }
                                                        else if (MVDiff >= 282) {
                                                            return 1;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (Cs >= 1141) {
                                            if (RsCsDiff < 67) {
                                                return 8;
                                            }
                                            else if (RsCsDiff >= 67) {
                                                if (TSC < 10) {
                                                    if (Cs < 1663) {
                                                        return 1;
                                                    }
                                                    else if (Cs >= 1663) {
                                                        return 8;
                                                    }
                                                }
                                                else if (TSC >= 10) {
                                                    if (RsCsDiff < 234) {
                                                        return 4;
                                                    }
                                                    else if (RsCsDiff >= 234) {
                                                        return 1;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (refDCval >= 75) {
                                        if (tcor < 31) {
                                            if (MVDiff < 684) {
                                                if (diffAFD < 26) {
                                                    return 1;
                                                }
                                                else if (diffAFD >= 26) {
                                                    if (TSC < 29) {
                                                        if (SC < 772) {
                                                            return 2;
                                                        }
                                                        else if (SC >= 772) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (TSC >= 29) {
                                                        return 1;
                                                    }
                                                }
                                            }
                                            else if (MVDiff >= 684) {
                                                return 8;
                                            }
                                        }
                                        else if (tcor >= 31) {
                                            if (mcTcor < 87) {
                                                if (TSC < 11) {
                                                    if (diffMVdiffVal < -306) {
                                                        return 4;
                                                    }
                                                    else if (diffMVdiffVal >= -306) {
                                                        if (Cs < 1609) {
                                                            return 1;
                                                        }
                                                        else if (Cs >= 1609) {
                                                            return 2;
                                                        }
                                                    }
                                                }
                                                else if (TSC >= 11) {
                                                    if (Rs < 562) {
                                                        if (TSC < 22) {
                                                            return 1;
                                                        }
                                                        else if (TSC >= 22) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (Rs >= 562) {
                                                        if (TSC < 12) {
                                                            if (Cs < 810) {
                                                                if (mu_mv_mag_sq < 43) {
                                                                    return 2;
                                                                }
                                                                else if (mu_mv_mag_sq >= 43) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (Cs >= 810) {
                                                                return 1;
                                                            }
                                                        }
                                                        else if (TSC >= 12) {
                                                            return 2;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (mcTcor >= 87) {
                                                if (RsCsDiff < 108) {
                                                    if (Cs < 1809) {
                                                        if (mu_mv_mag_sq < 47) {
                                                            if (CsDiff < 3) {
                                                                if (Cs < 833) {
                                                                    return 8;
                                                                }
                                                                else if (Cs >= 833) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (CsDiff >= 3) {
                                                                if (RsCsDiff < 43) {
                                                                    return 4;
                                                                }
                                                                else if (RsCsDiff >= 43) {
                                                                    return 2;
                                                                }
                                                            }
                                                        }
                                                        else if (mu_mv_mag_sq >= 47) {
                                                            if (Rs < 1419) {
                                                                if (negBalance < 33) {
                                                                    return 8;
                                                                }
                                                                else if (negBalance >= 33) {
                                                                    if (ssDCval < 109) {
                                                                        return 1;
                                                                    }
                                                                    else if (ssDCval >= 109) {
                                                                        if (tcor < 86) {
                                                                            return 2;
                                                                        }
                                                                        else if (tcor >= 86) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                            else if (Rs >= 1419) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                    else if (Cs >= 1809) {
                                                        return 1;
                                                    }
                                                }
                                                else if (RsCsDiff >= 108) {
                                                    if (negBalance < 53) {
                                                        if (Cs < 1266) {
                                                            return 2;
                                                        }
                                                        else if (Cs >= 1266) {
                                                            if (AFD < 30) {
                                                                if (gchDC < 4) {
                                                                    return 2;
                                                                }
                                                                else if (gchDC >= 4) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (AFD >= 30) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                    else if (negBalance >= 53) {
                                                        if (SC < 5138) {
                                                            if (MVDiff < 166) {
                                                                return 2;
                                                            }
                                                            else if (MVDiff >= 166) {
                                                                if (SC < 3301) {
                                                                    return 8;
                                                                }
                                                                else if (SC >= 3301) {
                                                                    return 1;
                                                                }
                                                            }
                                                        }
                                                        else if (SC >= 5138) {
                                                            if (AFD < 23) {
                                                                return 8;
                                                            }
                                                            else if (AFD >= 23) {
                                                                return 2;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (Cs >= 1831) {
                                    if (refDCval < 59) {
                                        if (tcor < 84) {
                                            return 8;
                                        }
                                        else if (tcor >= 84) {
                                            if (MVDiff < 17) {
                                                return 4;
                                            }
                                            else if (MVDiff >= 17) {
                                                if (RsDiff < 5) {
                                                    return 2;
                                                }
                                                else if (RsDiff >= 5) {
                                                    return 8;
                                                }
                                            }
                                        }
                                    }
                                    else if (refDCval >= 59) {
                                        if (mcTcor < 94) {
                                            if (Rs < 3602) {
                                                if (posBalance < 57) {
                                                    if (CsDiff < 9) {
                                                        if (refDCval < 84) {
                                                            if (Rs < 2614) {
                                                                return 8;
                                                            }
                                                            else if (Rs >= 2614) {
                                                                if (SC < 5737) {
                                                                    if (RsDiff < 10) {
                                                                        return 1;
                                                                    }
                                                                    else if (RsDiff >= 10) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (SC >= 5737) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                        else if (refDCval >= 84) {
                                                            if (TSC < 11) {
                                                                if (diffMVdiffVal < -5) {
                                                                    return 2;
                                                                }
                                                                else if (diffMVdiffVal >= -5) {
                                                                    if (ssDCval < 106) {
                                                                        return 1;
                                                                    }
                                                                    else if (ssDCval >= 106) {
                                                                        return 8;
                                                                    }
                                                                }
                                                            }
                                                            else if (TSC >= 11) {
                                                                if (ssDCval < 117) {
                                                                    if (Rs < 3115) {
                                                                        return 4;
                                                                    }
                                                                    else if (Rs >= 3115) {
                                                                        return 1;
                                                                    }
                                                                }
                                                                else if (ssDCval >= 117) {
                                                                    if (ssDCval < 124) {
                                                                        return 1;
                                                                    }
                                                                    else if (ssDCval >= 124) {
                                                                        return 8;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (CsDiff >= 9) {
                                                        if (ssDCval < 79) {
                                                            if (mcTcor < 84) {
                                                                return 8;
                                                            }
                                                            else if (mcTcor >= 84) {
                                                                return 1;
                                                            }
                                                        }
                                                        else if (ssDCval >= 79) {
                                                            if (negBalance < 59) {
                                                                if (mu_mv_mag_sq < 71) {
                                                                    if (Rs < 2553) {
                                                                        if (Cs < 2316) {
                                                                            if (diffTSC < -2) {
                                                                                return 1;
                                                                            }
                                                                            else if (diffTSC >= -2) {
                                                                                if (TSC < 10) {
                                                                                    return 8;
                                                                                }
                                                                                else if (TSC >= 10) {
                                                                                    return 4;
                                                                                }
                                                                            }
                                                                        }
                                                                        else if (Cs >= 2316) {
                                                                            if (TSC < 10) {
                                                                                if (Cs < 2855) {
                                                                                    if (ssDCval < 128) {
                                                                                        return 1;
                                                                                    }
                                                                                    else if (ssDCval >= 128) {
                                                                                        return 8;
                                                                                    }
                                                                                }
                                                                                else if (Cs >= 2855) {
                                                                                    return 8;
                                                                                }
                                                                            }
                                                                            else if (TSC >= 10) {
                                                                                return 8;
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (Rs >= 2553) {
                                                                        if (Cs < 3012) {
                                                                            if (AFD < 22) {
                                                                                if (MVDiff < 7) {
                                                                                    if (Rs < 3092) {
                                                                                        return 1;
                                                                                    }
                                                                                    else if (Rs >= 3092) {
                                                                                        return 4;
                                                                                    }
                                                                                }
                                                                                else if (MVDiff >= 7) {
                                                                                    return 1;
                                                                                }
                                                                            }
                                                                            else if (AFD >= 22) {
                                                                                return 8;
                                                                            }
                                                                        }
                                                                        else if (Cs >= 3012) {
                                                                            if (SC < 6923) {
                                                                                return 8;
                                                                            }
                                                                            else if (SC >= 6923) {
                                                                                return 1;
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                                else if (mu_mv_mag_sq >= 71) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (negBalance >= 59) {
                                                                if (mu_mv_mag_sq < 47) {
                                                                    if (SC < 3879) {
                                                                        return 1;
                                                                    }
                                                                    else if (SC >= 3879) {
                                                                        return 8;
                                                                    }
                                                                }
                                                                else if (mu_mv_mag_sq >= 47) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (posBalance >= 57) {
                                                    if (mcTcor < 90) {
                                                        if (refDCval < 81) {
                                                            if (TSC < 12) {
                                                                if (Rs < 2864) {
                                                                    return 1;
                                                                }
                                                                else if (Rs >= 2864) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (TSC >= 12) {
                                                                if (Cs < 2693) {
                                                                    return 8;
                                                                }
                                                                else if (Cs >= 2693) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                        else if (refDCval >= 81) {
                                                            if (refDCval < 111) {
                                                                if (MVDiff < 10) {
                                                                    return 4;
                                                                }
                                                                else if (MVDiff >= 10) {
                                                                    if (diffMVdiffVal < 5) {
                                                                        if (AFD < 13) {
                                                                            if (Rs < 2902) {
                                                                                return 4;
                                                                            }
                                                                            else if (Rs >= 2902) {
                                                                                return 1;
                                                                            }
                                                                        }
                                                                        else if (AFD >= 13) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                    else if (diffMVdiffVal >= 5) {
                                                                        return 4;
                                                                    }
                                                                }
                                                            }
                                                            else if (refDCval >= 111) {
                                                                return 1;
                                                            }
                                                        }
                                                    }
                                                    else if (mcTcor >= 90) {
                                                        if (AFD < 14) {
                                                            if (ssDCval < 84) {
                                                                return 1;
                                                            }
                                                            else if (ssDCval >= 84) {
                                                                if (refDCval < 89) {
                                                                    if (AFD < 13) {
                                                                        return 4;
                                                                    }
                                                                    else if (AFD >= 13) {
                                                                        return 8;
                                                                    }
                                                                }
                                                                else if (refDCval >= 89) {
                                                                    if (RsCsDiff < 146) {
                                                                        if (MVDiff < 11) {
                                                                            if (Cs < 2377) {
                                                                                return 1;
                                                                            }
                                                                            else if (Cs >= 2377) {
                                                                                return 4;
                                                                            }
                                                                        }
                                                                        else if (MVDiff >= 11) {
                                                                            if (SC < 5123) {
                                                                                return 4;
                                                                            }
                                                                            else if (SC >= 5123) {
                                                                                return 1;
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (RsCsDiff >= 146) {
                                                                        return 4;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (AFD >= 14) {
                                                            if (mu_mv_mag_sq < 44) {
                                                                if (refDCval < 92) {
                                                                    if (posBalance < 59) {
                                                                        return 2;
                                                                    }
                                                                    else if (posBalance >= 59) {
                                                                        if (ssDCval < 83) {
                                                                            return 1;
                                                                        }
                                                                        else if (ssDCval >= 83) {
                                                                            return 2;
                                                                        }
                                                                    }
                                                                }
                                                                else if (refDCval >= 92) {
                                                                    if (Cs < 1906) {
                                                                        return 2;
                                                                    }
                                                                    else if (Cs >= 1906) {
                                                                        return 1;
                                                                    }
                                                                }
                                                            }
                                                            else if (mu_mv_mag_sq >= 44) {
                                                                if (negBalance < 44) {
                                                                    if (Rs < 2891) {
                                                                        return 4;
                                                                    }
                                                                    else if (Rs >= 2891) {
                                                                        return 2;
                                                                    }
                                                                }
                                                                else if (negBalance >= 44) {
                                                                    if (diffAFD < 1) {
                                                                        if (diffMVdiffVal < -5) {
                                                                            return 2;
                                                                        }
                                                                        else if (diffMVdiffVal >= -5) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                    else if (diffAFD >= 1) {
                                                                        if (Rs < 2912) {
                                                                            return 1;
                                                                        }
                                                                        else if (Rs >= 2912) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                            else if (Rs >= 3602) {
                                                if (tcor < 78) {
                                                    if (refDCval < 108) {
                                                        if (tcor < 65) {
                                                            if (SC < 7490) {
                                                                return 8;
                                                            }
                                                            else if (SC >= 7490) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (tcor >= 65) {
                                                            if (posBalance < 49) {
                                                                if (diffRsCsdiff < -10) {
                                                                    return 4;
                                                                }
                                                                else if (diffRsCsdiff >= -10) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (posBalance >= 49) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                    else if (refDCval >= 108) {
                                                        if (ssDCval < 130) {
                                                            if (ssDCval < 119) {
                                                                if (AFD < 23) {
                                                                    return 1;
                                                                }
                                                                else if (AFD >= 23) {
                                                                    return 2;
                                                                }
                                                            }
                                                            else if (ssDCval >= 119) {
                                                                if (negBalance < 48) {
                                                                    if (AFD < 18) {
                                                                        return 1;
                                                                    }
                                                                    else if (AFD >= 18) {
                                                                        return 4;
                                                                    }
                                                                }
                                                                else if (negBalance >= 48) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                        else if (ssDCval >= 130) {
                                                            if (ssDCval < 140) {
                                                                return 1;
                                                            }
                                                            else if (ssDCval >= 140) {
                                                                if (Cs < 2011) {
                                                                    return 2;
                                                                }
                                                                else if (Cs >= 2011) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (tcor >= 78) {
                                                    if (tcor < 82) {
                                                        if (diffMVdiffVal < 6) {
                                                            return 4;
                                                        }
                                                        else if (diffMVdiffVal >= 6) {
                                                            if (Cs < 2166) {
                                                                return 4;
                                                            }
                                                            else if (Cs >= 2166) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                    else if (tcor >= 82) {
                                                        if (tcor < 86) {
                                                            if (Cs < 2585) {
                                                                return 8;
                                                            }
                                                            else if (Cs >= 2585) {
                                                                return 2;
                                                            }
                                                        }
                                                        else if (tcor >= 86) {
                                                            if (refDCval < 124) {
                                                                if (Cs < 2352) {
                                                                    return 8;
                                                                }
                                                                else if (Cs >= 2352) {
                                                                    return 4;
                                                                }
                                                            }
                                                            else if (refDCval >= 124) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (mcTcor >= 94) {
                                            if (Rs < 3538) {
                                                if (refDCval < 85) {
                                                    return 1;
                                                }
                                                else if (refDCval >= 85) {
                                                    if (AFD < 13) {
                                                        if (refDCval < 105) {
                                                            return 4;
                                                        }
                                                        else if (refDCval >= 105) {
                                                            return 2;
                                                        }
                                                    }
                                                    else if (AFD >= 13) {
                                                        if (diffTSC < -2) {
                                                            return 8;
                                                        }
                                                        else if (diffTSC >= -2) {
                                                            return 2;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (Rs >= 3538) {
                                                if (Cs < 2722) {
                                                    return 4;
                                                }
                                                else if (Cs >= 2722) {
                                                    return 8;
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                            else if (tcor >= 90) {
                                if (RsDiff < 9) {
                                    if (Rs < 2595) {
                                        if (posBalance < 55) {
                                            if (CsDiff < 7) {
                                                return 1;
                                            }
                                            else if (CsDiff >= 7) {
                                                return 8;
                                            }
                                        }
                                        else if (posBalance >= 55) {
                                            if (Rs < 1899) {
                                                if (Rs < 998) {
                                                    return 1;
                                                }
                                                else if (Rs >= 998) {
                                                    return 4;
                                                }
                                            }
                                            else if (Rs >= 1899) {
                                                if (MVDiff < 14) {
                                                    if (posBalance < 58) {
                                                        return 1;
                                                    }
                                                    else if (posBalance >= 58) {
                                                        if (TSC < 10) {
                                                            return 2;
                                                        }
                                                        else if (TSC >= 10) {
                                                            return 1;
                                                        }
                                                    }
                                                }
                                                else if (MVDiff >= 14) {
                                                    if (SC < 4572) {
                                                        return 2;
                                                    }
                                                    else if (SC >= 4572) {
                                                        return 1;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (Rs >= 2595) {
                                        if (SC < 5280) {
                                            if (RsCsDiff < 107) {
                                                return 1;
                                            }
                                            else if (RsCsDiff >= 107) {
                                                return 2;
                                            }
                                        }
                                        else if (SC >= 5280) {
                                            if (Cs < 2442) {
                                                if (MVDiff < 12) {
                                                    if (MVDiff < 8) {
                                                        return 4;
                                                    }
                                                    else if (MVDiff >= 8) {
                                                        return 1;
                                                    }
                                                }
                                                else if (MVDiff >= 12) {
                                                    if (diffRsCsdiff < -9) {
                                                        return 4;
                                                    }
                                                    else if (diffRsCsdiff >= -9) {
                                                        return 1;
                                                    }
                                                }
                                            }
                                            else if (Cs >= 2442) {
                                                if (refDCval < 60) {
                                                    return 4;
                                                }
                                                else if (refDCval >= 60) {
                                                    return 8;
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (RsDiff >= 9) {
                                    if (negBalance < 38) {
                                        if (Cs < 2158) {
                                            return 1;
                                        }
                                        else if (Cs >= 2158) {
                                            return 2;
                                        }
                                    }
                                    else if (negBalance >= 38) {
                                        if (MVDiff < 10) {
                                            if (SC < 6291) {
                                                if (Rs < 3241) {
                                                    return 4;
                                                }
                                                else if (Rs >= 3241) {
                                                    return 1;
                                                }
                                            }
                                            else if (SC >= 6291) {
                                                return 8;
                                            }
                                        }
                                        else if (MVDiff >= 10) {
                                            return 8;
                                        }
                                    }
                                }
                            }
                        }
                    }
                }
                else if (SC >= 7645) {
                    if (Cs < 4310) {
                        if (tcor < 68) {
                            if (refDCval < 125) {
                                if (RsDiff < 32) {
                                    return 8;
                                }
                                else if (RsDiff >= 32) {
                                    return 2;
                                }
                            }
                            else if (refDCval >= 125) {
                                if (ssDCval < 139) {
                                    if (posBalance < 56) {
                                        if (AFD < 23) {
                                            return 1;
                                        }
                                        else if (AFD >= 23) {
                                            return 4;
                                        }
                                    }
                                    else if (posBalance >= 56) {
                                        return 1;
                                    }
                                }
                                else if (ssDCval >= 139) {
                                    if (ssDCval < 156) {
                                        if (refDCval < 143) {
                                            return 8;
                                        }
                                        else if (refDCval >= 143) {
                                            return 2;
                                        }
                                    }
                                    else if (ssDCval >= 156) {
                                        return 4;
                                    }
                                }
                            }
                        }
                        else if (tcor >= 68) {
                            if (ssDCval < 106) {
                                if (TSC < 14) {
                                    if (TSC < 7) {
                                        if (negBalance < 45) {
                                            if (Rs < 4048) {
                                                return 8;
                                            }
                                            else if (Rs >= 4048) {
                                                if (CsDiff < 4) {
                                                    return 4;
                                                }
                                                else if (CsDiff >= 4) {
                                                    return 8;
                                                }
                                            }
                                        }
                                        else if (negBalance >= 45) {
                                            if (RsDiff < 3) {
                                                return 2;
                                            }
                                            else if (RsDiff >= 3) {
                                                if (refDCval < 53) {
                                                    if (Rs < 3960) {
                                                        return 4;
                                                    }
                                                    else if (Rs >= 3960) {
                                                        return 8;
                                                    }
                                                }
                                                else if (refDCval >= 53) {
                                                    return 8;
                                                }
                                            }
                                        }
                                    }
                                    else if (TSC >= 7) {
                                        if (Rs < 4009) {
                                            if (RsDiff < 6) {
                                                if (negBalance < 51) {
                                                    if (posBalance < 58) {
                                                        return 4;
                                                    }
                                                    else if (posBalance >= 58) {
                                                        if (ssDCval < 57) {
                                                            if (Cs < 3802) {
                                                                return 2;
                                                            }
                                                            else if (Cs >= 3802) {
                                                                if (Rs < 3883) {
                                                                    return 4;
                                                                }
                                                                else if (Rs >= 3883) {
                                                                    return 2;
                                                                }
                                                            }
                                                        }
                                                        else if (ssDCval >= 57) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                                else if (negBalance >= 51) {
                                                    if (CsDiff < 6) {
                                                        return 2;
                                                    }
                                                    else if (CsDiff >= 6) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                            else if (RsDiff >= 6) {
                                                if (TSC < 9) {
                                                    return 8;
                                                }
                                                else if (TSC >= 9) {
                                                    if (RsCsDiff < 104) {
                                                        return 4;
                                                    }
                                                    else if (RsCsDiff >= 104) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                        }
                                        else if (Rs >= 4009) {
                                            if (SC < 8794) {
                                                if (mu_mv_mag_sq < 103) {
                                                    if (refDCval < 107) {
                                                        return 4;
                                                    }
                                                    else if (refDCval >= 107) {
                                                        return 2;
                                                    }
                                                }
                                                else if (mu_mv_mag_sq >= 103) {
                                                    return 8;
                                                }
                                            }
                                            else if (SC >= 8794) {
                                                if (MVDiff < 15) {
                                                    return 8;
                                                }
                                                else if (MVDiff >= 15) {
                                                    return 4;
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (TSC >= 14) {
                                    return 8;
                                }
                            }
                            else if (ssDCval >= 106) {
                                if (tcor < 83) {
                                    if (Cs < 3293) {
                                        if (AFD < 22) {
                                            if (Rs < 5447) {
                                                if (CsDiff < 8) {
                                                    if (Rs < 5187) {
                                                        return 4;
                                                    }
                                                    else if (Rs >= 5187) {
                                                        return 2;
                                                    }
                                                }
                                                else if (CsDiff >= 8) {
                                                    if (diffMVdiffVal < 6) {
                                                        return 2;
                                                    }
                                                    else if (diffMVdiffVal >= 6) {
                                                        if (Rs < 5046) {
                                                            return 2;
                                                        }
                                                        else if (Rs >= 5046) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (Rs >= 5447) {
                                                if (MVDiff < 20) {
                                                    return 1;
                                                }
                                                else if (MVDiff >= 20) {
                                                    return 4;
                                                }
                                            }
                                        }
                                        else if (AFD >= 22) {
                                            return 4;
                                        }
                                    }
                                    else if (Cs >= 3293) {
                                        if (mu_mv_mag_sq < 3) {
                                            if (ssDCval < 159) {
                                                if (Cs < 3638) {
                                                    return 8;
                                                }
                                                else if (Cs >= 3638) {
                                                    if (diffTSC < -5) {
                                                        return 8;
                                                    }
                                                    else if (diffTSC >= -5) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                            else if (ssDCval >= 159) {
                                                return 2;
                                            }
                                        }
                                        else if (mu_mv_mag_sq >= 3) {
                                            if (diffTSC < 3) {
                                                if (RsDiff < 20) {
                                                    if (ssDCval < 124) {
                                                        if (RsCsDiff < 437) {
                                                            return 4;
                                                        }
                                                        else if (RsCsDiff >= 437) {
                                                            if (Rs < 4396) {
                                                                return 4;
                                                            }
                                                            else if (Rs >= 4396) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                    else if (ssDCval >= 124) {
                                                        return 8;
                                                    }
                                                }
                                                else if (RsDiff >= 20) {
                                                    if (negBalance < 47) {
                                                        return 8;
                                                    }
                                                    else if (negBalance >= 47) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                            else if (diffTSC >= 3) {
                                                return 8;
                                            }
                                        }
                                    }
                                }
                                else if (tcor >= 83) {
                                    if (CsDiff < 12) {
                                        if (Rs < 5101) {
                                            if (SC < 7735) {
                                                return 4;
                                            }
                                            else if (SC >= 7735) {
                                                if (Rs < 4297) {
                                                    return 8;
                                                }
                                                else if (Rs >= 4297) {
                                                    if (AFD < 15) {
                                                        if (diffMVdiffVal < 4) {
                                                            return 2;
                                                        }
                                                        else if (diffMVdiffVal >= 4) {
                                                            if (SC < 7990) {
                                                                return 2;
                                                            }
                                                            else if (SC >= 7990) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                    else if (AFD >= 15) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                        }
                                        else if (Rs >= 5101) {
                                            if (RsDiff < 12) {
                                                if (ssDCval < 111) {
                                                    if (Cs < 3095) {
                                                        return 4;
                                                    }
                                                    else if (Cs >= 3095) {
                                                        if (Cs < 3206) {
                                                            return 8;
                                                        }
                                                        else if (Cs >= 3206) {
                                                            return 2;
                                                        }
                                                    }
                                                }
                                                else if (ssDCval >= 111) {
                                                    return 8;
                                                }
                                            }
                                            else if (RsDiff >= 12) {
                                                if (Rs < 5281) {
                                                    return 8;
                                                }
                                                else if (Rs >= 5281) {
                                                    return 4;
                                                }
                                            }
                                        }
                                    }
                                    else if (CsDiff >= 12) {
                                        if (RsDiff < 22) {
                                            return 8;
                                        }
                                        else if (RsDiff >= 22) {
                                            if (RsCsDiff < 979) {
                                                return 4;
                                            }
                                            else if (RsCsDiff >= 979) {
                                                return 8;
                                            }
                                        }
                                    }
                                }
                            }
                        }
                    }
                    else if (Cs >= 4310) {
                        if (Rs < 6689) {
                            if (refDCval < 66) {
                                if (AFD < 16) {
                                    return 4;
                                }
                                else if (AFD >= 16) {
                                    return 8;
                                }
                            }
                            else if (refDCval >= 66) {
                                if (mu_mv_mag_sq < 33) {
                                    if (tcor < 83) {
                                        if (refDCval < 136) {
                                            return 8;
                                        }
                                        else if (refDCval >= 136) {
                                            if (AFD < 25) {
                                                if (refDCval < 147) {
                                                    if (Cs < 4556) {
                                                        return 1;
                                                    }
                                                    else if (Cs >= 4556) {
                                                        return 4;
                                                    }
                                                }
                                                else if (refDCval >= 147) {
                                                    return 2;
                                                }
                                            }
                                            else if (AFD >= 25) {
                                                if (RsDiff < 30) {
                                                    return 8;
                                                }
                                                else if (RsDiff >= 30) {
                                                    return 2;
                                                }
                                            }
                                        }
                                    }
                                    else if (tcor >= 83) {
                                        return 1;
                                    }
                                }
                                else if (mu_mv_mag_sq >= 33) {
                                    if (AFD < 37) {
                                        if (negBalance < 52) {
                                            if (mcTcor < 70) {
                                                if (Rs < 5239) {
                                                    return 1;
                                                }
                                                else if (Rs >= 5239) {
                                                    return 8;
                                                }
                                            }
                                            else if (mcTcor >= 70) {
                                                if (Rs < 6486) {
                                                    return 8;
                                                }
                                                else if (Rs >= 6486) {
                                                    if (Rs < 6528) {
                                                        return 1;
                                                    }
                                                    else if (Rs >= 6528) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                        }
                                        else if (negBalance >= 52) {
                                            return 8;
                                        }
                                    }
                                    else if (AFD >= 37) {
                                        if (TSC < 20) {
                                            return 2;
                                        }
                                        else if (TSC >= 20) {
                                            return 8;
                                        }
                                    }
                                }
                            }
                        }
                        else if (Rs >= 6689) {
                            if (Cs < 6235) {
                                if (refDCval < 76) {
                                    return 4;
                                }
                                else if (refDCval >= 76) {
                                    if (refDCval < 153) {
                                        if (AFD < 28) {
                                            if (negBalance < 52) {
                                                if (AFD < 24) {
                                                    return 8;
                                                }
                                                else if (AFD >= 24) {
                                                    return 4;
                                                }
                                            }
                                            else if (negBalance >= 52) {
                                                if (ssDCval < 147) {
                                                    if (Cs < 6063) {
                                                        return 4;
                                                    }
                                                    else if (Cs >= 6063) {
                                                        return 8;
                                                    }
                                                }
                                                else if (ssDCval >= 147) {
                                                    return 8;
                                                }
                                            }
                                        }
                                        else if (AFD >= 28) {
                                            if (TSC < 18) {
                                                return 4;
                                            }
                                            else if (TSC >= 18) {
                                                if (ssDCval < 145) {
                                                    if (MVDiff < 15) {
                                                        if (TSC < 19) {
                                                            return 2;
                                                        }
                                                        else if (TSC >= 19) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (MVDiff >= 15) {
                                                        return 2;
                                                    }
                                                }
                                                else if (ssDCval >= 145) {
                                                    if (SC < 14608) {
                                                        return 4;
                                                    }
                                                    else if (SC >= 14608) {
                                                        return 2;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (refDCval >= 153) {
                                        if (negBalance < 52) {
                                            return 8;
                                        }
                                        else if (negBalance >= 52) {
                                            if (Cs < 5072) {
                                                if (SC < 12424) {
                                                    return 2;
                                                }
                                                else if (SC >= 12424) {
                                                    return 8;
                                                }
                                            }
                                            else if (Cs >= 5072) {
                                                return 2;
                                            }
                                        }
                                    }
                                }
                            }
                            else if (Cs >= 6235) {
                                if (AFD < 24) {
                                    return 8;
                                }
                                else if (AFD >= 24) {
                                    if (ssDCval < 88) {
                                        return 4;
                                    }
                                    else if (ssDCval >= 88) {
                                        if (negBalance < 54) {
                                            if (CsDiff < 43) {
                                                return 8;
                                            }
                                            else if (CsDiff >= 43) {
                                                return 4;
                                            }
                                        }
                                        else if (negBalance >= 54) {
                                            if (TSC < 22) {
                                                if (negBalance < 55) {
                                                    if (Cs < 6568) {
                                                        return 4;
                                                    }
                                                    else if (Cs >= 6568) {
                                                        return 8;
                                                    }
                                                }
                                                else if (negBalance >= 55) {
                                                    return 8;
                                                }
                                            }
                                            else if (TSC >= 22) {
                                                if (SC < 24995) {
                                                    if (diffMVdiffVal < -3) {
                                                        if (posBalance < 59) {
                                                            if (Rs < 10936) {
                                                                return 2;
                                                            }
                                                            else if (Rs >= 10936) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (posBalance >= 59) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (diffMVdiffVal >= -3) {
                                                        if (ssDCval < 147) {
                                                            return 4;
                                                        }
                                                        else if (ssDCval >= 147) {
                                                            if (RsDiff < 39) {
                                                                return 4;
                                                            }
                                                            else if (RsDiff >= 39) {
                                                                return 2;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (SC >= 24995) {
                                                    if (Rs < 14665) {
                                                        if (Rs < 14081) {
                                                            return 8;
                                                        }
                                                        else if (Rs >= 14081) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (Rs >= 14665) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                        }
                    }
                }
            }
        }
        else if (tcor >= 95) {
            if (MVDiff < 5) {
                if (negBalance < 53) {
                    if (mu_mv_mag_sq < 123) {
                        if (mu_mv_mag_sq < 108) {
                            if (SC < 4911) {
                                if (AFD < 6) {
                                    if (Cs < 266) {
                                        if (Cs < 12) {
                                            if (AFD < 2) {
                                                return 1;
                                            }
                                            else if (AFD >= 2) {
                                                return 8;
                                            }
                                        }
                                        else if (Cs >= 12) {
                                            return 8;
                                        }
                                    }
                                    else if (Cs >= 266) {
                                        if (ssDCval < 153) {
                                            if (MVDiff < 1) {
                                                if (diffMVdiffVal < -1) {
                                                    if (refDCval < 85) {
                                                        return 1;
                                                    }
                                                    else if (refDCval >= 85) {
                                                        if (diffRsCsdiff < 2) {
                                                            if (diffMVdiffVal < -22) {
                                                                return 2;
                                                            }
                                                            else if (diffMVdiffVal >= -22) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (diffRsCsdiff >= 2) {
                                                            if (Rs < 1153) {
                                                                return 4;
                                                            }
                                                            else if (Rs >= 1153) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (diffMVdiffVal >= -1) {
                                                    if (posBalance < 62) {
                                                        if (negBalance < 19) {
                                                            if (Cs < 441) {
                                                                return 1;
                                                            }
                                                            else if (Cs >= 441) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (negBalance >= 19) {
                                                            if (diffTSC < 2) {
                                                                return 4;
                                                            }
                                                            else if (diffTSC >= 2) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                    else if (posBalance >= 62) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                            else if (MVDiff >= 1) {
                                                if (SC < 2107) {
                                                    if (negBalance < 35) {
                                                        if (diffMVdiffVal < -2) {
                                                            if (AFD < 5) {
                                                                return 4;
                                                            }
                                                            else if (AFD >= 5) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (diffMVdiffVal >= -2) {
                                                            if (mcTcor < 96) {
                                                                return 4;
                                                            }
                                                            else if (mcTcor >= 96) {
                                                                if (SC < 531) {
                                                                    return 4;
                                                                }
                                                                else if (SC >= 531) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (negBalance >= 35) {
                                                        if (SC < 1127) {
                                                            if (Rs < 365) {
                                                                return 2;
                                                            }
                                                            else if (Rs >= 365) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (SC >= 1127) {
                                                            if (posBalance < 55) {
                                                                return 4;
                                                            }
                                                            else if (posBalance >= 55) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (SC >= 2107) {
                                                    if (SC < 3931) {
                                                        if (SC < 2336) {
                                                            if (Cs < 1037) {
                                                                return 1;
                                                            }
                                                            else if (Cs >= 1037) {
                                                                if (TSC < 1) {
                                                                    return 2;
                                                                }
                                                                else if (TSC >= 1) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                        else if (SC >= 2336) {
                                                            if (Cs < 686) {
                                                                if (TSC < 3) {
                                                                    return 8;
                                                                }
                                                                else if (TSC >= 3) {
                                                                    return 2;
                                                                }
                                                            }
                                                            else if (Cs >= 686) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                    else if (SC >= 3931) {
                                                        if (posBalance < 38) {
                                                            return 4;
                                                        }
                                                        else if (posBalance >= 38) {
                                                            if (Rs < 3013) {
                                                                return 8;
                                                            }
                                                            else if (Rs >= 3013) {
                                                                return 1;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (ssDCval >= 153) {
                                            if (Cs < 462) {
                                                if (diffMVdiffVal < 1) {
                                                    return 4;
                                                }
                                                else if (diffMVdiffVal >= 1) {
                                                    if (Rs < 462) {
                                                        if (posBalance < 31) {
                                                            return 2;
                                                        }
                                                        else if (posBalance >= 31) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (Rs >= 462) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                            else if (Cs >= 462) {
                                                if (posBalance < 27) {
                                                    if (Rs < 530) {
                                                        return 2;
                                                    }
                                                    else if (Rs >= 530) {
                                                        if (Cs < 1609) {
                                                            return 4;
                                                        }
                                                        else if (Cs >= 1609) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                                else if (posBalance >= 27) {
                                                    return 4;
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (AFD >= 6) {
                                    if (SC < 3266) {
                                        if (Cs < 1326) {
                                            if (diffMVdiffVal < -5) {
                                                return 4;
                                            }
                                            else if (diffMVdiffVal >= -5) {
                                                return 2;
                                            }
                                        }
                                        else if (Cs >= 1326) {
                                            if (MVDiff < 2) {
                                                return 2;
                                            }
                                            else if (MVDiff >= 2) {
                                                return 8;
                                            }
                                        }
                                    }
                                    else if (SC >= 3266) {
                                        if (Rs < 2573) {
                                            if (diffAFD < -1) {
                                                if (ssDCval < 88) {
                                                    if (CsDiff < 6) {
                                                        return 4;
                                                    }
                                                    else if (CsDiff >= 6) {
                                                        return 8;
                                                    }
                                                }
                                                else if (ssDCval >= 88) {
                                                    return 2;
                                                }
                                            }
                                            else if (diffAFD >= -1) {
                                                if (diffTSC < 1) {
                                                    return 4;
                                                }
                                                else if (diffTSC >= 1) {
                                                    if (Rs < 2047) {
                                                        return 4;
                                                    }
                                                    else if (Rs >= 2047) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                        }
                                        else if (Rs >= 2573) {
                                            if (SC < 4554) {
                                                if (diffAFD < 2) {
                                                    return 1;
                                                }
                                                else if (diffAFD >= 2) {
                                                    return 8;
                                                }
                                            }
                                            else if (SC >= 4554) {
                                                return 4;
                                            }
                                        }
                                    }
                                }
                            }
                            else if (SC >= 4911) {
                                if (ssDCval < 98) {
                                    if (negBalance < 43) {
                                        return 8;
                                    }
                                    else if (negBalance >= 43) {
                                        if (mu_mv_mag_sq < 89) {
                                            return 8;
                                        }
                                        else if (mu_mv_mag_sq >= 89) {
                                            return 4;
                                        }
                                    }
                                }
                                else if (ssDCval >= 98) {
                                    return 8;
                                }
                            }
                        }
                        else if (mu_mv_mag_sq >= 108) {
                            if (Cs < 3161) {
                                if (negBalance < 35) {
                                    if (CsDiff < 3) {
                                        if (Cs < 708) {
                                            if (SC < 1575) {
                                                if (Rs < 567) {
                                                    if (refDCval < 46) {
                                                        return 8;
                                                    }
                                                    else if (refDCval >= 46) {
                                                        if (posBalance < 59) {
                                                            return 4;
                                                        }
                                                        else if (posBalance >= 59) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                                else if (Rs >= 567) {
                                                    return 8;
                                                }
                                            }
                                            else if (SC >= 1575) {
                                                if (AFD < 1) {
                                                    return 4;
                                                }
                                                else if (AFD >= 1) {
                                                    return 1;
                                                }
                                            }
                                        }
                                        else if (Cs >= 708) {
                                            if (MVDiff < 3) {
                                                if (Cs < 2332) {
                                                    if (diffAFD < -2) {
                                                        if (Rs < 1748) {
                                                            return 1;
                                                        }
                                                        else if (Rs >= 1748) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (diffAFD >= -2) {
                                                        if (AFD < 4) {
                                                            return 8;
                                                        }
                                                        else if (AFD >= 4) {
                                                            if (Cs < 1419) {
                                                                return 4;
                                                            }
                                                            else if (Cs >= 1419) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (Cs >= 2332) {
                                                    if (negBalance < 16) {
                                                        if (Rs < 2789) {
                                                            return 1;
                                                        }
                                                        else if (Rs >= 2789) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (negBalance >= 16) {
                                                        if (RsDiff < 3) {
                                                            return 8;
                                                        }
                                                        else if (RsDiff >= 3) {
                                                            if (MVDiff < 1) {
                                                                return 1;
                                                            }
                                                            else if (MVDiff >= 1) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                            else if (MVDiff >= 3) {
                                                if (Cs < 1258) {
                                                    return 8;
                                                }
                                                else if (Cs >= 1258) {
                                                    if (Cs < 1299) {
                                                        return 2;
                                                    }
                                                    else if (Cs >= 1299) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (CsDiff >= 3) {
                                        if (refDCval < 110) {
                                            if (refDCval < 69) {
                                                return 8;
                                            }
                                            else if (refDCval >= 69) {
                                                if (ssDCval < 83) {
                                                    return 1;
                                                }
                                                else if (ssDCval >= 83) {
                                                    if (posBalance < 45) {
                                                        if (RsCsDiff < 29) {
                                                            if (refDCval < 108) {
                                                                return 8;
                                                            }
                                                            else if (refDCval >= 108) {
                                                                if (mu_mv_mag_sq < 116) {
                                                                    return 8;
                                                                }
                                                                else if (mu_mv_mag_sq >= 116) {
                                                                    if (Rs < 3156) {
                                                                        return 8;
                                                                    }
                                                                    else if (Rs >= 3156) {
                                                                        return 1;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (RsCsDiff >= 29) {
                                                            if (MVDiff < 2) {
                                                                if (SC < 4250) {
                                                                    return 8;
                                                                }
                                                                else if (SC >= 4250) {
                                                                    if (negBalance < 26) {
                                                                        return 8;
                                                                    }
                                                                    else if (negBalance >= 26) {
                                                                        return 1;
                                                                    }
                                                                }
                                                            }
                                                            else if (MVDiff >= 2) {
                                                                return 1;
                                                            }
                                                        }
                                                    }
                                                    else if (posBalance >= 45) {
                                                        if (negBalance < 33) {
                                                            if (refDCval < 109) {
                                                                return 8;
                                                            }
                                                            else if (refDCval >= 109) {
                                                                if (MVDiff < 1) {
                                                                    if (CsDiff < 4) {
                                                                        return 8;
                                                                    }
                                                                    else if (CsDiff >= 4) {
                                                                        return 1;
                                                                    }
                                                                }
                                                                else if (MVDiff >= 1) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                        else if (negBalance >= 33) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (refDCval >= 110) {
                                            if (Rs < 1337) {
                                                return 4;
                                            }
                                            else if (Rs >= 1337) {
                                                return 8;
                                            }
                                        }
                                    }
                                }
                                else if (negBalance >= 35) {
                                    if (SC < 3032) {
                                        if (Rs < 1367) {
                                            if (ssDCval < 81) {
                                                if (Rs < 986) {
                                                    return 8;
                                                }
                                                else if (Rs >= 986) {
                                                    return 2;
                                                }
                                            }
                                            else if (ssDCval >= 81) {
                                                return 4;
                                            }
                                        }
                                        else if (Rs >= 1367) {
                                            return 8;
                                        }
                                    }
                                    else if (SC >= 3032) {
                                        if (Rs < 2441) {
                                            if (ssDCval < 77) {
                                                if (Rs < 1806) {
                                                    if (Cs < 1807) {
                                                        if (Cs < 1241) {
                                                            return 4;
                                                        }
                                                        else if (Cs >= 1241) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (Cs >= 1807) {
                                                        return 4;
                                                    }
                                                }
                                                else if (Rs >= 1806) {
                                                    if (Cs < 1272) {
                                                        if (SC < 3077) {
                                                            return 4;
                                                        }
                                                        else if (SC >= 3077) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (Cs >= 1272) {
                                                        if (mu_mv_mag_sq < 121) {
                                                            return 4;
                                                        }
                                                        else if (mu_mv_mag_sq >= 121) {
                                                            if (AFD < 5) {
                                                                return 4;
                                                            }
                                                            else if (AFD >= 5) {
                                                                if (diffAFD < -1) {
                                                                    return 4;
                                                                }
                                                                else if (diffAFD >= -1) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                            else if (ssDCval >= 77) {
                                                if (negBalance < 44) {
                                                    if (Rs < 1671) {
                                                        return 8;
                                                    }
                                                    else if (Rs >= 1671) {
                                                        return 4;
                                                    }
                                                }
                                                else if (negBalance >= 44) {
                                                    if (SC < 4521) {
                                                        if (RsDiff < 6) {
                                                            return 2;
                                                        }
                                                        else if (RsDiff >= 6) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (SC >= 4521) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                        }
                                        else if (Rs >= 2441) {
                                            if (negBalance < 48) {
                                                if (posBalance < 59) {
                                                    if (ssDCval < 88) {
                                                        return 4;
                                                    }
                                                    else if (ssDCval >= 88) {
                                                        if (Cs < 2987) {
                                                            if (ssDCval < 113) {
                                                                return 8;
                                                            }
                                                            else if (ssDCval >= 113) {
                                                                if (ssDCval < 114) {
                                                                    if (SC < 5610) {
                                                                        return 8;
                                                                    }
                                                                    else if (SC >= 5610) {
                                                                        if (Rs < 2760) {
                                                                            return 4;
                                                                        }
                                                                        else if (Rs >= 2760) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                }
                                                                else if (ssDCval >= 114) {
                                                                    return 8;
                                                                }
                                                            }
                                                        }
                                                        else if (Cs >= 2987) {
                                                            if (Cs < 3073) {
                                                                if (Rs < 2832) {
                                                                    if (CsDiff < 5) {
                                                                        if (SC < 5716) {
                                                                            return 8;
                                                                        }
                                                                        else if (SC >= 5716) {
                                                                            if (refDCval < 115) {
                                                                                return 4;
                                                                            }
                                                                            else if (refDCval >= 115) {
                                                                                return 8;
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (CsDiff >= 5) {
                                                                        return 8;
                                                                    }
                                                                }
                                                                else if (Rs >= 2832) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (Cs >= 3073) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (posBalance >= 59) {
                                                    if (SC < 6214) {
                                                        return 8;
                                                    }
                                                    else if (SC >= 6214) {
                                                        return 2;
                                                    }
                                                }
                                            }
                                            else if (negBalance >= 48) {
                                                if (Rs < 2650) {
                                                    return 8;
                                                }
                                                else if (Rs >= 2650) {
                                                    if (refDCval < 112) {
                                                        return 8;
                                                    }
                                                    else if (refDCval >= 112) {
                                                        if (Rs < 3009) {
                                                            if (ssDCval < 115) {
                                                                if (mu_mv_mag_sq < 122) {
                                                                    return 4;
                                                                }
                                                                else if (mu_mv_mag_sq >= 122) {
                                                                    if (Cs < 3004) {
                                                                        if (ssDCval < 114) {
                                                                            if (Cs < 2975) {
                                                                                if (Rs < 2711) {
                                                                                    return 8;
                                                                                }
                                                                                else if (Rs >= 2711) {
                                                                                    return 4;
                                                                                }
                                                                            }
                                                                            else if (Cs >= 2975) {
                                                                                return 8;
                                                                            }
                                                                        }
                                                                        else if (ssDCval >= 114) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                    else if (Cs >= 3004) {
                                                                        if (tcor < 97) {
                                                                            return 8;
                                                                        }
                                                                        else if (tcor >= 97) {
                                                                            return 4;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                            else if (ssDCval >= 115) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (Rs >= 3009) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                            else if (Cs >= 3161) {
                                if (mcTcor < 96) {
                                    if (Rs < 4216) {
                                        return 2;
                                    }
                                    else if (Rs >= 4216) {
                                        return 8;
                                    }
                                }
                                else if (mcTcor >= 96) {
                                    return 8;
                                }
                            }
                        }
                    }
                    else if (mu_mv_mag_sq >= 123) {
                        if (negBalance < 3) {
                            if (SC < 7800) {
                                if (Rs < 3596) {
                                    if (refDCval < 121) {
                                        return 8;
                                    }
                                    else if (refDCval >= 121) {
                                        return 1;
                                    }
                                }
                                else if (Rs >= 3596) {
                                    if (Rs < 3779) {
                                        if (SC < 7764) {
                                            if (Rs < 3730) {
                                                if (Cs < 3876) {
                                                    if (Rs < 3686) {
                                                        if (Rs < 3641) {
                                                            return 1;
                                                        }
                                                        else if (Rs >= 3641) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (Rs >= 3686) {
                                                        return 1;
                                                    }
                                                }
                                                else if (Cs >= 3876) {
                                                    return 8;
                                                }
                                            }
                                            else if (Rs >= 3730) {
                                                return 8;
                                            }
                                        }
                                        else if (SC >= 7764) {
                                            return 1;
                                        }
                                    }
                                    else if (Rs >= 3779) {
                                        return 8;
                                    }
                                }
                            }
                            else if (SC >= 7800) {
                                return 8;
                            }
                        }
                        else if (negBalance >= 3) {
                            if (Rs < 588) {
                                if (Cs < 557) {
                                    if (Rs < 190) {
                                        if (SC < 486) {
                                            return 8;
                                        }
                                        else if (SC >= 486) {
                                            if (negBalance < 12) {
                                                return 8;
                                            }
                                            else if (negBalance >= 12) {
                                                if (diffAFD < 1) {
                                                    if (Rs < 187) {
                                                        if (posBalance < 27) {
                                                            return 4;
                                                        }
                                                        else if (posBalance >= 27) {
                                                            if (SC < 490) {
                                                                return 4;
                                                            }
                                                            else if (SC >= 490) {
                                                                if (Rs < 181) {
                                                                    return 8;
                                                                }
                                                                else if (Rs >= 181) {
                                                                    if (Rs < 184) {
                                                                        return 4;
                                                                    }
                                                                    else if (Rs >= 184) {
                                                                        return 8;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (Rs >= 187) {
                                                        return 4;
                                                    }
                                                }
                                                else if (diffAFD >= 1) {
                                                    return 4;
                                                }
                                            }
                                        }
                                    }
                                    else if (Rs >= 190) {
                                        return 8;
                                    }
                                }
                                else if (Cs >= 557) {
                                    if (Scindex < 4) {
                                        if (RsCsDiff < 1) {
                                            return 4;
                                        }
                                        else if (RsCsDiff >= 1) {
                                            if (posBalance < 45) {
                                                return 4;
                                            }
                                            else if (posBalance >= 45) {
                                                return 8;
                                            }
                                        }
                                    }
                                    else if (Scindex >= 4) {
                                        return 8;
                                    }
                                }
                            }
                            else if (Rs >= 588) {
                                if (negBalance < 35) {
                                    if (RsCsDiff < 12) {
                                        if (negBalance < 10) {
                                            return 8;
                                        }
                                        else if (negBalance >= 10) {
                                            if (AFD < 1) {
                                                if (Cs < 2140) {
                                                    if (Rs < 1194) {
                                                        return 8;
                                                    }
                                                    else if (Rs >= 1194) {
                                                        if (Cs < 1283) {
                                                            if (Rs < 1422) {
                                                                return 4;
                                                            }
                                                            else if (Rs >= 1422) {
                                                                return 8;
                                                            }
                                                        }
                                                        else if (Cs >= 1283) {
                                                            return 4;
                                                        }
                                                    }
                                                }
                                                else if (Cs >= 2140) {
                                                    return 8;
                                                }
                                            }
                                            else if (AFD >= 1) {
                                                if (Cs < 2327) {
                                                    return 8;
                                                }
                                                else if (Cs >= 2327) {
                                                    if (SC < 4647) {
                                                        return 4;
                                                    }
                                                    else if (SC >= 4647) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (RsCsDiff >= 12) {
                                        if (posBalance < 43) {
                                            if (negBalance < 25) {
                                                return 8;
                                            }
                                            else if (negBalance >= 25) {
                                                if (Cs < 3093) {
                                                    if (Rs < 3106) {
                                                        if (CsDiff < 4) {
                                                            return 8;
                                                        }
                                                        else if (CsDiff >= 4) {
                                                            return 1;
                                                        }
                                                    }
                                                    else if (Rs >= 3106) {
                                                        return 1;
                                                    }
                                                }
                                                else if (Cs >= 3093) {
                                                    return 8;
                                                }
                                            }
                                        }
                                        else if (posBalance >= 43) {
                                            return 8;
                                        }
                                    }
                                }
                                else if (negBalance >= 35) {
                                    if (refDCval < 83) {
                                        if (Rs < 1838) {
                                            if (ssDCval < 74) {
                                                if (AFD < 4) {
                                                    return 8;
                                                }
                                                else if (AFD >= 4) {
                                                    if (Rs < 1788) {
                                                        return 4;
                                                    }
                                                    else if (Rs >= 1788) {
                                                        return 2;
                                                    }
                                                }
                                            }
                                            else if (ssDCval >= 74) {
                                                if (Cs < 1709) {
                                                    return 8;
                                                }
                                                else if (Cs >= 1709) {
                                                    if (Rs < 1711) {
                                                        return 8;
                                                    }
                                                    else if (Rs >= 1711) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                        }
                                        else if (Rs >= 1838) {
                                            if (tcor < 96) {
                                                if (Rs < 1951) {
                                                    if (refDCval < 75) {
                                                        return 2;
                                                    }
                                                    else if (refDCval >= 75) {
                                                        return 8;
                                                    }
                                                }
                                                else if (Rs >= 1951) {
                                                    if (Cs < 1615) {
                                                        return 2;
                                                    }
                                                    else if (Cs >= 1615) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                            else if (tcor >= 96) {
                                                if (Cs < 1586) {
                                                    return 8;
                                                }
                                                else if (Cs >= 1586) {
                                                    if (SC < 4878) {
                                                        return 4;
                                                    }
                                                    else if (SC >= 4878) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (refDCval >= 83) {
                                        if (tcor < 96) {
                                            if (ssDCval < 106) {
                                                if (Cs < 3889) {
                                                    if (Cs < 3127) {
                                                        return 8;
                                                    }
                                                    else if (Cs >= 3127) {
                                                        return 2;
                                                    }
                                                }
                                                else if (Cs >= 3889) {
                                                    return 8;
                                                }
                                            }
                                            else if (ssDCval >= 106) {
                                                if (Rs < 5320) {
                                                    if (Rs < 2220) {
                                                        return 4;
                                                    }
                                                    else if (Rs >= 2220) {
                                                        return 8;
                                                    }
                                                }
                                                else if (Rs >= 5320) {
                                                    if (TSC < 9) {
                                                        return 8;
                                                    }
                                                    else if (TSC >= 9) {
                                                        return 4;
                                                    }
                                                }
                                            }
                                        }
                                        else if (tcor >= 96) {
                                            if (Rs < 1840) {
                                                if (Cs < 1993) {
                                                    return 8;
                                                }
                                                else if (Cs >= 1993) {
                                                    return 4;
                                                }
                                            }
                                            else if (Rs >= 1840) {
                                                if (diffRsCsdiff < -9) {
                                                    if (Cs < 3077) {
                                                        if (refDCval < 112) {
                                                            return 8;
                                                        }
                                                        else if (refDCval >= 112) {
                                                            if (negBalance < 48) {
                                                                return 8;
                                                            }
                                                            else if (negBalance >= 48) {
                                                                if (Rs < 2663) {
                                                                    return 8;
                                                                }
                                                                else if (Rs >= 2663) {
                                                                    if (posBalance < 57) {
                                                                        if (Rs < 2805) {
                                                                            return 4;
                                                                        }
                                                                        else if (Rs >= 2805) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                    else if (posBalance >= 57) {
                                                                        return 4;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (Cs >= 3077) {
                                                        return 8;
                                                    }
                                                }
                                                else if (diffRsCsdiff >= -9) {
                                                    if (mu_mv_mag_sq < 124) {
                                                        if (negBalance < 48) {
                                                            return 8;
                                                        }
                                                        else if (negBalance >= 48) {
                                                            if (negBalance < 49) {
                                                                if (diffRsCsdiff < 4) {
                                                                    if (Rs < 2736) {
                                                                        return 8;
                                                                    }
                                                                    else if (Rs >= 2736) {
                                                                        if (Rs < 2828) {
                                                                            return 4;
                                                                        }
                                                                        else if (Rs >= 2828) {
                                                                            return 8;
                                                                        }
                                                                    }
                                                                }
                                                                else if (diffRsCsdiff >= 4) {
                                                                    return 8;
                                                                }
                                                            }
                                                            else if (negBalance >= 49) {
                                                                return 8;
                                                            }
                                                        }
                                                    }
                                                    else if (mu_mv_mag_sq >= 124) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                        }
                    }
                }
                else if (negBalance >= 53) {
                    if (Rs < 5675) {
                        if (TSC < 4) {
                            if (ssDCval < 55) {
                                return 4;
                            }
                            else if (ssDCval >= 55) {
                                return 8;
                            }
                        }
                        else if (TSC >= 4) {
                            if (Cs < 3371) {
                                if (TSC < 5) {
                                    if (Rs < 714) {
                                        return 4;
                                    }
                                    else if (Rs >= 714) {
                                        return 8;
                                    }
                                }
                                else if (TSC >= 5) {
                                    if (AFD < 13) {
                                        return 4;
                                    }
                                    else if (AFD >= 13) {
                                        return 1;
                                    }
                                }
                            }
                            else if (Cs >= 3371) {
                                if (SC < 10895) {
                                    return 8;
                                }
                                else if (SC >= 10895) {
                                    if (Rs < 5651) {
                                        return 4;
                                    }
                                    else if (Rs >= 5651) {
                                        return 8;
                                    }
                                }
                            }
                        }
                    }
                    else if (Rs >= 5675) {
                        return 4;
                    }
                }
            }
            else if (MVDiff >= 5) {
                if (SC < 4975) {
                    if (AFD < 9) {
                        if (ssDCval < 150) {
                            if (negBalance < 30) {
                                if (refDCval < 31) {
                                    if (Rs < 1133) {
                                        if (posBalance < 10) {
                                            if (negBalance < 10) {
                                                return 8;
                                            }
                                            else if (negBalance >= 10) {
                                                if (negBalance < 17) {
                                                    return 4;
                                                }
                                                else if (negBalance >= 17) {
                                                    if (MVDiff < 75) {
                                                        return 8;
                                                    }
                                                    else if (MVDiff >= 75) {
                                                        return 2;
                                                    }
                                                }
                                            }
                                        }
                                        else if (posBalance >= 10) {
                                            return 4;
                                        }
                                    }
                                    else if (Rs >= 1133) {
                                        if (RsDiff < 1) {
                                            return 8;
                                        }
                                        else if (RsDiff >= 1) {
                                            return 2;
                                        }
                                    }
                                }
                                else if (refDCval >= 31) {
                                    if (TSC < 3) {
                                        if (posBalance < 56) {
                                            if (Cs < 1063) {
                                                if (negBalance < 15) {
                                                    if (mcTcor < 100) {
                                                        if (Rs < 363) {
                                                            return 4;
                                                        }
                                                        else if (Rs >= 363) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (mcTcor >= 100) {
                                                        return 1;
                                                    }
                                                }
                                                else if (negBalance >= 15) {
                                                    return 1;
                                                }
                                            }
                                            else if (Cs >= 1063) {
                                                if (diffRsCsdiff < -24) {
                                                    return 1;
                                                }
                                                else if (diffRsCsdiff >= -24) {
                                                    if (RsCsDiff < 1) {
                                                        return 4;
                                                    }
                                                    else if (RsCsDiff >= 1) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                        }
                                        else if (posBalance >= 56) {
                                            if (Rs < 678) {
                                                return 4;
                                            }
                                            else if (Rs >= 678) {
                                                return 8;
                                            }
                                        }
                                    }
                                    else if (TSC >= 3) {
                                        if (RsCsDiff < 51) {
                                            return 8;
                                        }
                                        else if (RsCsDiff >= 51) {
                                            return 4;
                                        }
                                    }
                                }
                            }
                            else if (negBalance >= 30) {
                                if (mu_mv_mag_sq < 56) {
                                    if (posBalance < 2) {
                                        if (diffMVdiffVal < 51) {
                                            if (mu_mv_mag_sq < 22) {
                                                if (refDCval < 33) {
                                                    if (diffMVdiffVal < 13) {
                                                        return 2;
                                                    }
                                                    else if (diffMVdiffVal >= 13) {
                                                        return 4;
                                                    }
                                                }
                                                else if (refDCval >= 33) {
                                                    return 8;
                                                }
                                            }
                                            else if (mu_mv_mag_sq >= 22) {
                                                return 8;
                                            }
                                        }
                                        else if (diffMVdiffVal >= 51) {
                                            if (MVDiff < 125) {
                                                return 2;
                                            }
                                            else if (MVDiff >= 125) {
                                                if (TSC < 1) {
                                                    return 1;
                                                }
                                                else if (TSC >= 1) {
                                                    return 8;
                                                }
                                            }
                                        }
                                    }
                                    else if (posBalance >= 2) {
                                        if (posBalance < 45) {
                                            if (ssDCval < 72) {
                                                if (MVDiff < 38) {
                                                    if (Rs < 380) {
                                                        if (diffAFD < -1) {
                                                            if (SC < 148) {
                                                                return 2;
                                                            }
                                                            else if (SC >= 148) {
                                                                if (MVDiff < 14) {
                                                                    return 2;
                                                                }
                                                                else if (MVDiff >= 14) {
                                                                    return 4;
                                                                }
                                                            }
                                                        }
                                                        else if (diffAFD >= -1) {
                                                            return 4;
                                                        }
                                                    }
                                                    else if (Rs >= 380) {
                                                        if (CsDiff < 1) {
                                                            return 8;
                                                        }
                                                        else if (CsDiff >= 1) {
                                                            return 2;
                                                        }
                                                    }
                                                }
                                                else if (MVDiff >= 38) {
                                                    if (MVDiff < 93) {
                                                        return 2;
                                                    }
                                                    else if (MVDiff >= 93) {
                                                        return 1;
                                                    }
                                                }
                                            }
                                            else if (ssDCval >= 72) {
                                                if (SC < 2463) {
                                                    return 8;
                                                }
                                                else if (SC >= 2463) {
                                                    if (SC < 4777) {
                                                        return 4;
                                                    }
                                                    else if (SC >= 4777) {
                                                        return 8;
                                                    }
                                                }
                                            }
                                        }
                                        else if (posBalance >= 45) {
                                            if (MVDiff < 15) {
                                                return 8;
                                            }
                                            else if (MVDiff >= 15) {
                                                if (mu_mv_mag_sq < 35) {
                                                    return 1;
                                                }
                                                else if (mu_mv_mag_sq >= 35) {
                                                    return 8;
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (mu_mv_mag_sq >= 56) {
                                    if (Rs < 371) {
                                        if (mu_mv_mag_sq < 75) {
                                            if (SC < 162) {
                                                return 2;
                                            }
                                            else if (SC >= 162) {
                                                return 8;
                                            }
                                        }
                                        else if (mu_mv_mag_sq >= 75) {
                                            if (ssDCval < 83) {
                                                if (Rs < 1) {
                                                    return 2;
                                                }
                                                else if (Rs >= 1) {
                                                    return 4;
                                                }
                                            }
                                            else if (ssDCval >= 83) {
                                                if (Cs < 933) {
                                                    if (SC < 707) {
                                                        return 8;
                                                    }
                                                    else if (SC >= 707) {
                                                        return 2;
                                                    }
                                                }
                                                else if (Cs >= 933) {
                                                    return 4;
                                                }
                                            }
                                        }
                                    }
                                    else if (Rs >= 371) {
                                        if (TSCindex < 5) {
                                            if (tcor < 96) {
                                                if (Rs < 1339) {
                                                    return 4;
                                                }
                                                else if (Rs >= 1339) {
                                                    return 8;
                                                }
                                            }
                                            else if (tcor >= 96) {
                                                if (diffTSC < -1) {
                                                    if (SC < 1977) {
                                                        return 4;
                                                    }
                                                    else if (SC >= 1977) {
                                                        return 8;
                                                    }
                                                }
                                                else if (diffTSC >= -1) {
                                                    return 8;
                                                }
                                            }
                                        }
                                        else if (TSCindex >= 5) {
                                            if (RsCsDiff < 9) {
                                                if (SC < 3081) {
                                                    if (Rs < 656) {
                                                        if (mu_mv_mag_sq < 80) {
                                                            return 8;
                                                        }
                                                        else if (mu_mv_mag_sq >= 80) {
                                                            if (diffRsCsdiff < -475) {
                                                                return 8;
                                                            }
                                                            else if (diffRsCsdiff >= -475) {
                                                                return 4;
                                                            }
                                                        }
                                                    }
                                                    else if (Rs >= 656) {
                                                        if (refDCval < 60) {
                                                            if (mu_mv_mag_sq < 79) {
                                                                return 1;
                                                            }
                                                            else if (mu_mv_mag_sq >= 79) {
                                                                if (ssDCval < 53) {
                                                                    return 8;
                                                                }
                                                                else if (ssDCval >= 53) {
                                                                    return 1;
                                                                }
                                                            }
                                                        }
                                                        else if (refDCval >= 60) {
                                                            return 8;
                                                        }
                                                    }
                                                }
                                                else if (SC >= 3081) {
                                                    return 4;
                                                }
                                            }
                                            else if (RsCsDiff >= 9) {
                                                if (Cs < 1646) {
                                                    if (mcTcor < 97) {
                                                        if (CsDiff < 4) {
                                                            if (ssDCval < 47) {
                                                                return 8;
                                                            }
                                                            else if (ssDCval >= 47) {
                                                                return 4;
                                                            }
                                                        }
                                                        else if (CsDiff >= 4) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (mcTcor >= 97) {
                                                        return 4;
                                                    }
                                                }
                                                else if (Cs >= 1646) {
                                                    if (AFD < 8) {
                                                        if (SC < 4166) {
                                                            return 4;
                                                        }
                                                        else if (SC >= 4166) {
                                                            return 8;
                                                        }
                                                    }
                                                    else if (AFD >= 8) {
                                                        return 1;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                        }
                        else if (ssDCval >= 150) {
                            if (refDCval < 168) {
                                if (posBalance < 33) {
                                    if (diffTSC < -1) {
                                        if (Cs < 1263) {
                                            return 2;
                                        }
                                        else if (Cs >= 1263) {
                                            return 4;
                                        }
                                    }
                                    else if (diffTSC >= -1) {
                                        return 4;
                                    }
                                }
                                else if (posBalance >= 33) {
                                    if (Rs < 513) {
                                        if (mcTcor < 100) {
                                            return 4;
                                        }
                                        else if (mcTcor >= 100) {
                                            return 2;
                                        }
                                    }
                                    else if (Rs >= 513) {
                                        return 2;
                                    }
                                }
                            }
                            else if (refDCval >= 168) {
                                if (SC < 1744) {
                                    return 1;
                                }
                                else if (SC >= 1744) {
                                    return 8;
                                }
                            }
                        }
                    }
                    else if (AFD >= 9) {
                        if (SC < 4463) {
                            if (Cs < 337) {
                                if (posBalance < 1) {
                                    return 8;
                                }
                                else if (posBalance >= 1) {
                                    if (diffAFD < -1) {
                                        return 2;
                                    }
                                    else if (diffAFD >= -1) {
                                        if (MVDiff < 29) {
                                            return 2;
                                        }
                                        else if (MVDiff >= 29) {
                                            if (mcTcor < 102) {
                                                return 1;
                                            }
                                            else if (mcTcor >= 102) {
                                                return 2;
                                            }
                                        }
                                    }
                                }
                            }
                            else if (Cs >= 337) {
                                if (refDCval < 85) {
                                    if (Cs < 1410) {
                                        if (SC < 1957) {
                                            return 2;
                                        }
                                        else if (SC >= 1957) {
                                            return 4;
                                        }
                                    }
                                    else if (Cs >= 1410) {
                                        return 1;
                                    }
                                }
                                else if (refDCval >= 85) {
                                    if (ssDCval < 91) {
                                        if (SC < 3081) {
                                            return 2;
                                        }
                                        else if (SC >= 3081) {
                                            return 4;
                                        }
                                    }
                                    else if (ssDCval >= 91) {
                                        if (refDCval < 99) {
                                            return 8;
                                        }
                                        else if (refDCval >= 99) {
                                            if (MVDiff < 155) {
                                                if (gchDC < 3) {
                                                    return 8;
                                                }
                                                else if (gchDC >= 3) {
                                                    return 4;
                                                }
                                            }
                                            else if (MVDiff >= 155) {
                                                return 8;
                                            }
                                        }
                                    }
                                }
                            }
                        }
                        else if (SC >= 4463) {
                            return 1;
                        }
                    }
                }
                else if (SC >= 4975) {
                    if (refDCval < 109) {
                        if (RsDiff < 1) {
                            if (Cs < 3541) {
                                return 2;
                            }
                            else if (Cs >= 3541) {
                                return 8;
                            }
                        }
                        else if (RsDiff >= 1) {
                            if (refDCval < 54) {
                                return 4;
                            }
                            else if (refDCval >= 54) {
                                if (TSCindex < 6) {
                                    return 8;
                                }
                                else if (TSCindex >= 6) {
                                    if (mu_mv_mag_sq < 73) {
                                        if (diffRsCsdiff < -5) {
                                            if (MVDiff < 10) {
                                                return 8;
                                            }
                                            else if (MVDiff >= 10) {
                                                return 4;
                                            }
                                        }
                                        else if (diffRsCsdiff >= -5) {
                                            return 8;
                                        }
                                    }
                                    else if (mu_mv_mag_sq >= 73) {
                                        if (diffAFD < -2) {
                                            if (Rs < 4789) {
                                                return 4;
                                            }
                                            else if (Rs >= 4789) {
                                                return 8;
                                            }
                                        }
                                        else if (diffAFD >= -2) {
                                            return 8;
                                        }
                                    }
                                }
                            }
                        }
                    }
                    else if (refDCval >= 109) {
                        if (mu_mv_mag_sq < 93) {
                            if (refDCval < 119) {
                                if (mu_mv_mag_sq < 71) {
                                    return 4;
                                }
                                else if (mu_mv_mag_sq >= 71) {
                                    if (Rs < 3232) {
                                        return 8;
                                    }
                                    else if (Rs >= 3232) {
                                        if (diffRsCsdiff < 21) {
                                            return 2;
                                        }
                                        else if (diffRsCsdiff >= 21) {
                                            return 4;
                                        }
                                    }
                                }
                            }
                            else if (refDCval >= 119) {
                                return 4;
                            }
                        }
                        else if (mu_mv_mag_sq >= 93) {
                            return 8;
                        }
                    }
                }
            }
        }
        return 4;
    }
    typedef std::function<bool(
        std::pair<mfxU16, mfxU16>,
        std::pair<mfxU16, mfxU16>)>
        Size_Comparator;

    mfxU16 AGOPSelectRF(
        mfxI32 diffMVdiffVal, mfxU32 RsCsDiff, mfxU32 MVDiff, mfxU32 Rs, mfxU32 AFD,
        mfxU32 CsDiff, mfxI32 diffTSC, mfxU32 TSC, mfxU32 gchDC, mfxI32 diffRsCsdiff,
        mfxU32 posBalance, mfxU32 SC, mfxU32 TSCindex, mfxU32 Scindex, mfxU32 Cs,
        mfxI32 diffAFD, mfxU32 negBalance, mfxU32 ssDCval, mfxU32 refDCval, mfxU32 RsDiff,
        mfxU32 mu_mv_mag_sq, mfxI16 mcTcor, mfxI16 tcor) {
        std::map<mfxU16, mfxU16>
            gop_freq;

        Size_Comparator Comparison = [](
            std::pair<mfxU16, mfxU16> size1,
            std::pair<mfxU16, mfxU16> size2
            )
        {
            return size1.second != size2.second ? size1.second > size2.second : size1.first < size2.first;
        };

        gop_freq[AGOPSelect1(Rs, CsDiff, TSC, diffAFD, Scindex, RsDiff, diffMVdiffVal, RsCsDiff, TSCindex, tcor, refDCval, MVDiff, AFD, diffTSC, diffRsCsdiff, posBalance, SC, mu_mv_mag_sq, negBalance, mcTcor, gchDC, Cs, ssDCval)]++;
        gop_freq[AGOPSelect2(Rs, CsDiff, TSC, diffAFD, Scindex, RsDiff, diffMVdiffVal, RsCsDiff, TSCindex, tcor, refDCval, MVDiff, AFD, diffTSC, diffRsCsdiff, posBalance, SC, mu_mv_mag_sq, negBalance, mcTcor, gchDC, Cs, ssDCval)]++;
        gop_freq[AGOPSelect3(Rs, CsDiff, TSC, diffAFD, Scindex, RsDiff, diffMVdiffVal, RsCsDiff, TSCindex, tcor, refDCval, MVDiff, AFD, diffTSC, diffRsCsdiff, posBalance, SC, mu_mv_mag_sq, negBalance, mcTcor, gchDC, Cs, ssDCval)]++;
        gop_freq[AGOPSelect4(Rs, CsDiff, TSC, diffAFD, Scindex, RsDiff, diffMVdiffVal, RsCsDiff, TSCindex, tcor, refDCval, MVDiff, AFD, diffTSC, diffRsCsdiff, posBalance, SC, mu_mv_mag_sq, negBalance, mcTcor, gchDC, Cs, ssDCval)]++;
        gop_freq[AGOPSelect5(Rs, CsDiff, TSC, diffAFD, Scindex, RsDiff, diffMVdiffVal, RsCsDiff, TSCindex, tcor, refDCval, MVDiff, AFD, diffTSC, diffRsCsdiff, posBalance, SC, mu_mv_mag_sq, negBalance, mcTcor, gchDC, Cs, ssDCval)]++;
        gop_freq[AGOPSelect6(Rs, CsDiff, TSC, diffAFD, Scindex, RsDiff, diffMVdiffVal, RsCsDiff, TSCindex, tcor, refDCval, MVDiff, AFD, diffTSC, diffRsCsdiff, posBalance, SC, mu_mv_mag_sq, negBalance, mcTcor, gchDC, Cs, ssDCval)]++;
        gop_freq[AGOPSelect7(Rs, CsDiff, TSC, diffAFD, Scindex, RsDiff, diffMVdiffVal, RsCsDiff, TSCindex, tcor, refDCval, MVDiff, AFD, diffTSC, diffRsCsdiff, posBalance, SC, mu_mv_mag_sq, negBalance, mcTcor, gchDC, Cs, ssDCval)]++;
        gop_freq[AGOPSelect8(Rs, CsDiff, TSC, diffAFD, Scindex, RsDiff, diffMVdiffVal, RsCsDiff, TSCindex, tcor, refDCval, MVDiff, AFD, diffTSC, diffRsCsdiff, posBalance, SC, mu_mv_mag_sq, negBalance, mcTcor, gchDC, Cs, ssDCval)]++;
        gop_freq[AGOPSelect9(Rs, CsDiff, TSC, diffAFD, Scindex, RsDiff, diffMVdiffVal, RsCsDiff, TSCindex, tcor, refDCval, MVDiff, AFD, diffTSC, diffRsCsdiff, posBalance, SC, mu_mv_mag_sq, negBalance, mcTcor, gchDC, Cs, ssDCval)]++;
        gop_freq[AGOPSelectA(Rs, CsDiff, TSC, diffAFD, Scindex, RsDiff, diffMVdiffVal, RsCsDiff, TSCindex, tcor, refDCval, MVDiff, AFD, diffTSC, diffRsCsdiff, posBalance, SC, mu_mv_mag_sq, negBalance, mcTcor, gchDC, Cs, ssDCval)]++;
        gop_freq[AGOPSelectB(Rs, CsDiff, TSC, diffAFD, Scindex, RsDiff, diffMVdiffVal, RsCsDiff, TSCindex, tcor, refDCval, MVDiff, AFD, diffTSC, diffRsCsdiff, posBalance, SC, mu_mv_mag_sq, negBalance, mcTcor, gchDC, Cs, ssDCval)]++;
        gop_freq[AGOPSelectC(Rs, CsDiff, TSC, diffAFD, Scindex, RsDiff, diffMVdiffVal, RsCsDiff, TSCindex, tcor, refDCval, MVDiff, AFD, diffTSC, diffRsCsdiff, posBalance, SC, mu_mv_mag_sq, negBalance, mcTcor, gchDC, Cs, ssDCval)]++;
        gop_freq[AGOPSelectD(Rs, CsDiff, TSC, diffAFD, Scindex, RsDiff, diffMVdiffVal, RsCsDiff, TSCindex, tcor, refDCval, MVDiff, AFD, diffTSC, diffRsCsdiff, posBalance, SC, mu_mv_mag_sq, negBalance, mcTcor, gchDC, Cs, ssDCval)]++;
        gop_freq[AGOPSelectE(Rs, CsDiff, TSC, diffAFD, Scindex, RsDiff, diffMVdiffVal, RsCsDiff, TSCindex, tcor, refDCval, MVDiff, AFD, diffTSC, diffRsCsdiff, posBalance, SC, mu_mv_mag_sq, negBalance, mcTcor, gchDC, Cs, ssDCval)]++;
        gop_freq[AGOPSelectF(Rs, CsDiff, TSC, diffAFD, Scindex, RsDiff, diffMVdiffVal, RsCsDiff, TSCindex, tcor, refDCval, MVDiff, AFD, diffTSC, diffRsCsdiff, posBalance, SC, mu_mv_mag_sq, negBalance, mcTcor, gchDC, Cs, ssDCval)]++;
        gop_freq[AGOPSelect10(Rs, CsDiff, TSC, diffAFD, Scindex, RsDiff, diffMVdiffVal, RsCsDiff, TSCindex, tcor, refDCval, MVDiff, AFD, diffTSC, diffRsCsdiff, posBalance, SC, mu_mv_mag_sq, negBalance, mcTcor, gchDC, Cs, ssDCval)]++;
        gop_freq[AGOPSelect11(Rs, CsDiff, TSC, diffAFD, Scindex, RsDiff, diffMVdiffVal, RsCsDiff, TSCindex, tcor, refDCval, MVDiff, AFD, diffTSC, diffRsCsdiff, posBalance, SC, mu_mv_mag_sq, negBalance, mcTcor, gchDC, Cs, ssDCval)]++;
        gop_freq[AGOPSelect12(Rs, CsDiff, TSC, diffAFD, Scindex, RsDiff, diffMVdiffVal, RsCsDiff, TSCindex, tcor, refDCval, MVDiff, AFD, diffTSC, diffRsCsdiff, posBalance, SC, mu_mv_mag_sq, negBalance, mcTcor, gchDC, Cs, ssDCval)]++;
        gop_freq[AGOPSelect13(Rs, CsDiff, TSC, diffAFD, Scindex, RsDiff, diffMVdiffVal, RsCsDiff, TSCindex, tcor, refDCval, MVDiff, AFD, diffTSC, diffRsCsdiff, posBalance, SC, mu_mv_mag_sq, negBalance, mcTcor, gchDC, Cs, ssDCval)]++;
        gop_freq[AGOPSelect14(Rs, CsDiff, TSC, diffAFD, Scindex, RsDiff, diffMVdiffVal, RsCsDiff, TSCindex, tcor, refDCval, MVDiff, AFD, diffTSC, diffRsCsdiff, posBalance, SC, mu_mv_mag_sq, negBalance, mcTcor, gchDC, Cs, ssDCval)]++;
        gop_freq[AGOPSelect15(Rs, CsDiff, TSC, diffAFD, Scindex, RsDiff, diffMVdiffVal, RsCsDiff, TSCindex, tcor, refDCval, MVDiff, AFD, diffTSC, diffRsCsdiff, posBalance, SC, mu_mv_mag_sq, negBalance, mcTcor, gchDC, Cs, ssDCval)]++;
        gop_freq[AGOPSelect16(Rs, CsDiff, TSC, diffAFD, Scindex, RsDiff, diffMVdiffVal, RsCsDiff, TSCindex, tcor, refDCval, MVDiff, AFD, diffTSC, diffRsCsdiff, posBalance, SC, mu_mv_mag_sq, negBalance, mcTcor, gchDC, Cs, ssDCval)]++;
        gop_freq[AGOPSelect17(Rs, CsDiff, TSC, diffAFD, Scindex, RsDiff, diffMVdiffVal, RsCsDiff, TSCindex, tcor, refDCval, MVDiff, AFD, diffTSC, diffRsCsdiff, posBalance, SC, mu_mv_mag_sq, negBalance, mcTcor, gchDC, Cs, ssDCval)]++;

        std::set<std::pair<mfxU16, mfxU16>, Size_Comparator>
            size_sorted(
                gop_freq.begin(),
                gop_freq.end(),
                Comparison);

        return(size_sorted.begin()->first);
    }
};//namespace aenc
#endif // MFX_ENABLE_ADAPTIVE_ENCODE
